# @(#)buildtree.make	1.2 00/06/14
#
# Usage:
#
# $(MAKE) -f buildtree.make ARCH=arch GAMMADIR=dir OS_FAMILY=os VARIANT=variant
#
# The macros ARCH, GAMMADIR, OS_FAMILY and VARIANT must be defined in the
# environment or on the command-line:
#
# ARCH		- sparc, i486, ...
# GAMMADIR	- top of workspace
# OS_FAMILY	- operating system
# VARIANT	- core, compiler1, or compiler2
#
# Builds the directory trees with 4 makefiles plus some convenience files in
# each directory:
#
# Makefile	- for "make foo"
# flags.make	- with macro settings
# vm.make	- to support making "$(MAKE) -v vm.make" in makefiles
# adlc.make	- 
# env.[ck]sh	- environment settings
# test_gamma	- script to run the Queens program
# 
# The makefiles are split this way so that "make foo" will run faster by not
# having to read the dependency files for the vm.

# 'gmake MAKE_VERBOSE=y' or 'gmake QUIETLY=' gives all the gory details.
QUIETLY$(MAKE_VERBOSE)	= @

# For now, until the compiler is less wobbly:
TESTFLAGS	= -Xbatch

PLATFORM_FILE	= $(GAMMADIR)/build/$(OS_FAMILY)/platform_$(ARCH)
PLATFORM_DIR	= $(OS_FAMILY)_$(ARCH)_$(VARIANT)

ifeq		($(VARIANT),compiler2)
EXCLUDE_DIRS	= -name SCCS -o -name adlc -o -name c1
else
# compiler1 and core use the same exclude list
EXCLUDE_DIRS	= -name SCCS -o -name adlc -o -name opto -o -name libadt
endif

# Get things from the platform file.
COMPILER	= $(shell sed -n 's/^compiler[ 	]*=[ 	]*//p' $(PLATFORM_FILE))
LIB_ARCH	= $(shell sed -n 's/^lib_arch[ 	]*=[ 	]*//p' $(PLATFORM_FILE))

SIMPLE_DIRS	= \
	$(PLATFORM_DIR)/generated/incls \
	$(PLATFORM_DIR)/generated/adfiles

SUBMAKE_DIRS	= \
	$(PLATFORM_DIR)/debug \
	$(PLATFORM_DIR)/fastdebug \
	$(PLATFORM_DIR)/jvmg \
	$(PLATFORM_DIR)/optimized \
	$(PLATFORM_DIR)/product \
	$(PLATFORM_DIR)/profiled \
	$(PLATFORM_DIR)/release

# For dependencies and recursive makes.
BUILDTREE_MAKE	= $(GAMMADIR)/build/solaris/makefiles/buildtree.make

BUILDTREE_TARGETS = Makefile flags.make vm.make adlc.make env.ksh env.csh \
	.dbxrc test_gamma

BUILDTREE_VARS	= GAMMADIR=$(GAMMADIR) OS_FAMILY=$(OS_FAMILY) \
	ARCH=$(ARCH) VARIANT=$(VARIANT)

BUILDTREE	= \
	$(MAKE) -f $(BUILDTREE_MAKE) $(BUILDTREE_TARGETS) $(BUILDTREE_VARS)

BUILDTREE_COMMENT	= echo "\# Generated by $(BUILDTREE_MAKE)"

.PHONY:  all FORCE

all:  $(SUBMAKE_DIRS)

# Run make in each subdirectory recursively.
$(SUBMAKE_DIRS): $(SIMPLE_DIRS) FORCE
	$(QUIETLY) [ -d $@ ] || { mkdir -p $@; }
	$(QUIETLY) cd $@ && $(BUILDTREE) TARGET=$(@F)
	$(QUIETLY) touch $@

$(SIMPLE_DIRS):
	$(QUIETLY) mkdir -p $@

flags.make: $(BUILDTREE_MAKE) ../shared_dirs.lst
	@echo Creating $@ ...
	$(QUIETLY) ( \
	$(BUILDTREE_COMMENT); \
	echo; \
	echo "Platform_file = $(PLATFORM_FILE)"; \
	sed -n '/=/s/^ */Platform_/p' < $(PLATFORM_FILE); \
	echo; \
	echo "GAMMADIR = $(GAMMADIR)"; \
	echo "SYSDEFS = $$(Platform_sysdefs)"; \
	echo; \
	echo "Src_Dirs = \\"; \
	sed 's/$$/ \\/' ../shared_dirs.lst; \
	echo "$(GAMMADIR)/src/cpu/$(ARCH)/vm \\"; \
	echo "$(GAMMADIR)/src/os/$(OS_FAMILY)/vm \\"; \
	echo "$(GAMMADIR)/src/os_cpu/$(OS_FAMILY)_$(ARCH)/vm"; \
	[ -n "$(CFLAGS_BROWSE)" ] && \
	    echo && echo "CFLAGS_BROWSE = $(CFLAGS_BROWSE)"; \
	echo; \
	echo include $(GAMMADIR)/build/$(OS_FAMILY)/makefiles/$(VARIANT).make; \
	[ "$(TARGET)" = profiled ] && \
	echo include $(GAMMADIR)/build/$(OS_FAMILY)/makefiles/optimized.make; \
	echo include $(GAMMADIR)/build/$(OS_FAMILY)/makefiles/$(TARGET).make; \
	echo include $(GAMMADIR)/build/$(OS_FAMILY)/makefiles/$(COMPILER).make; \
	) > $@

../shared_dirs.lst:  $(BUILDTREE_MAKE) $(GAMMADIR)/src/share/vm
	@echo Creating directory list $@
	$(QUIETLY) find $(GAMMADIR)/src/share/vm/* -prune \
	-type d \! \( $(EXCLUDE_DIRS) \) -print > $@

Makefile: $(BUILDTREE_MAKE)
	@echo Creating $@ ...
	$(QUIETLY) ( \
	$(BUILDTREE_COMMENT); \
	echo; \
	echo include flags.make; \
	echo; \
	echo include $(GAMMADIR)/build/$(OS_FAMILY)/makefiles/top.make; \
	) > $@

vm.make adlc.make: $(BUILDTREE_MAKE)
	@echo Creating $@ ...
	$(QUIETLY) ( \
	$(BUILDTREE_COMMENT); \
	echo; \
	echo include flags.make; \
	echo; \
	echo include $(GAMMADIR)/build/$(OS_FAMILY)/makefiles/$(@F); \
	) > $@

env.ksh: $(BUILDTREE_MAKE)
	@echo Creating $@ ...
	$(QUIETLY) ( \
	$(BUILDTREE_COMMENT); \
	[ -n "$$JAVA_HOME" ] && { echo ": \$${JAVA_HOME:=$${JAVA_HOME}}"; }; \
	{ \
	echo "LD_LIBRARY_PATH=.:$${LD_LIBRARY_PATH:+$$LD_LIBRARY_PATH:}\$${JAVA_HOME}/jre/lib/${LIB_ARCH}/native_threads:\$${JAVA_HOME}/jre/lib/${LIB_ARCH}"; \
	echo "CLASSPATH=$${CLASSPATH:+$$CLASSPATH:}.:\$${JAVA_HOME}/jre/lib/rt.jar:\$${JAVA_HOME}/jre/lib/i18n.jar"; \
	} | sed s:$${JAVA_HOME:--------}:\$${JAVA_HOME}:g; \
	echo "export JAVA_HOME LD_LIBRARY_PATH CLASSPATH"; \
	) > $@

env.csh: env.ksh
	@echo Creating $@ ...
	$(QUIETLY) ( \
	$(BUILDTREE_COMMENT); \
	[ -n "$$JAVA_HOME" ] && \
	{ echo "if (! \$$?JAVA_HOME) setenv JAVA_HOME \"$$JAVA_HOME\""; }; \
	sed -n 's/^\([A-Za-z_][A-Za-z0-9_]*\)=/setenv \1 /p' $?; \
	) > $@

.dbxrc:  $(BUILDTREE_MAKE)
	@echo Creating $@ ...
	$(QUIETLY) ( \
	echo "echo '# Loading $(PLATFORM_DIR)/$(TARGET)/.dbxrc'"; \
	echo "if [ -f \"\$${HOTSPOT_DBXWARE}\" ]"; \
	echo "then"; \
	echo "	source \"\$${HOTSPOT_DBXWARE}\""; \
	echo "elif [ -f \"\$$HOME/.dbxrc\" ]"; \
	echo "then"; \
	echo "	source \"\$$HOME/.dbxrc\""; \
	echo "fi"; \
	) > $@

# Skip the test for product builds (which only work when installed in a JDK), to
# avoid exiting with an error and causing make to halt.
NO_TEST_MSG	= \
	echo "$@:  skipping the test--this build must be tested in a JDK."

test_gamma:  $(BUILDTREE_MAKE)
	@echo Creating $@ ...
	$(QUIETLY) ( \
	echo '#!/bin/ksh'; \
	$(BUILDTREE_COMMENT); \
	[ "$(TARGET)" = product ] && { echo '$(NO_TEST_MSG)'; exit 0; }; \
	echo '. ./env.ksh'; \
	echo 'CLASSPATH="$(GAMMADIR)/build/solaris:$$CLASSPATH"'; \
	echo '[ -f gamma_g ] && { gamma=gamma_g; }'; \
	echo './$${gamma:-gamma} $(TESTFLAGS) Queens < /dev/null'; \
	) > $@
	$(QUIETLY) chmod +x $@

FORCE:
