// @(#)c1_CacheLocals_i486.cpp	1.28 00/04/24 10:01:06
// 
// Copyright 1993-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


# include "incls/_precompiled.incl"
# include "incls/_c1_CacheLocals_i486.cpp.incl"


//------------------caching locals---------------------



class BlockItemSetter: public BlockClosure {
 private:
  BlockItem* _item;
 public:
  BlockItemSetter(BlockItem* item): _item(item) {}

  virtual void block_do(BlockBegin* block) {
    assert(block->block_item() == NULL, "destroying information");
    block->set_block_item(_item);
  }
};


//-------------------------------------------------------------------------


// Returns an array of cpu register numbers as they are used inside
// register allocator. To map the number to RInfo use set_word_reg in RInfo
RInfoCollection* LocalCaching::collect_free_cpu_registers(Loop* loop) {

  // Collect Free registers using standard iteration
  RegAlloc reg_alloc;
  int max_spills = 0;
  { ValueGenInvariant vgi(_ir->method(), &reg_alloc, NULL);
    ValueGen visitor(NULL, &vgi, false); // no code: collect free registers only
    CodeGenerator gen(&visitor, &vgi);
    for (int n = 0; n < loop->nof_blocks(); n++) {
      BlockBegin* bb = loop->block_no(n);
      gen.block_do(bb);
    }
    max_spills = gen.max_spills();
  }

  // get registers that have not been used inside the blocks of the loop, and
  // lock them (guarantee that they are never used inside the loop)
  c1_RegMask available_registers = reg_alloc.free_registers();
  loop->set_lockout(available_registers);

  // No caching of locals if there are any spills
  RInfoCollection* free_regs = 
    max_spills == 0 ? RegAlloc::regmask_2_rinfocollection(available_registers) : new RInfoCollection();

  if (PrintLoops) {
    if (max_spills > 0) tty->print("[has spills, no caching of locals]");
    tty->print("Loop %d..%d nof free registers is %d;", loop->start()->bci(), loop->end()->bci(), free_regs->number_of_registers());
    tty->print(" lockout 0x%x; ");
    loop->lockout().print();
    tty->cr();
    for (int i = 0; i < free_regs->number_of_registers(); i++) {
      RInfo reg = free_regs->register_at(i);
      reg.print();
      tty->print(" "); 
    }
    tty->cr();
  }

  // clear items in blocks of the loop
  loop->blocks_do(CodeGenerator::clear_instruction_items);
  return free_regs;
}


// returns mapping local_index -> RInfo.number() (also marks if it is oop)
void LocalCaching::assign_registers(CachedLocals* locals_desc,
				    const ScanResult& scan,
				    GrowableArray<ALocal*>* ordered_locals,
				    RInfoCollection* free_regs,
				    Loop* loop) {
  RInfoCollection* mapping = new RInfoCollection();
  GrowableArray<bool>* oops = new GrowableArray<bool>();
  // We assign free registers to locals, starting with the locals at index 0 inside
  // array "ordered_locals" (-> contains local indces ordered by their usage)
  for (int r = 0; r < free_regs->number_of_registers(); r++) {
    if (r < ordered_locals->length()) {
      // build register - local pair
      int local_index = ordered_locals->at(r)->index();
      bool is_oop     = ordered_locals->at(r)->is_oop();
      assert(local_index >= 0, "invalid index");
      CachedLocal* cl = new CachedLocal(local_index, free_regs->register_at(r));
      loop->append_cached_local(cl);

      mapping->at_put(local_index, cl->rinfo());
      oops->at_put_grow(local_index, is_oop, false);
    }
  }
  locals_desc->set_mapping(mapping);
  locals_desc->set_oops_info(oops);
}


void LocalCaching::add_block_items(Loop* loop, const RInfoCollection* mapping, GrowableArray<bool>* oops, bool set_32bit_precision) {
  int n = loop->nof_blocks() - 1;
  BlockItem* item = new BlockItem(mapping, oops, loop->lockout(), set_32bit_precision);
  { BlockItemSetter setter(item);
    loop->iterate_blocks(&setter);
  }
}

// Currently we optimize only integer and oop locals in methods that
// do not have exception handlers and loops that do not have calls or slow cases
void LocalCaching::cache_loop_locals() {
  GrowableArray<Loop*>* loops = _ir->loops();
  if (loops == NULL || _ir->method()->has_exception_handlers()) return;
  for (int i = 0; i < loops->length(); i++) {
    Loop* loop = loops->at(i);
    loop->sort_blocks();
    ScanBlocks scan(loop->blocks());
    ScanResult scan_result;
    scan.scan(&scan_result);
    assert(!scan_result.has_calls(), "the loops may not have calls");
    if (PrintLoops) scan.print(&scan_result);
    if (scan_result.can_cache_locals()) {
      RInfoCollection* free_int_regs = collect_free_cpu_registers(loop);
      GrowableArray<ALocal*>* locals = NULL;
      locals = scan.most_used_locals();

      // locals contains indeces of locals, sorted by their usage
      // describe oops
      if (PrintLoops) {
        tty->print(" MAX[");
        for (int i = 0; i < locals->length(); i++) {
          tty->print("%d ", locals->at(i));
        }
        tty->print_cr("]");
      }
      CachedLocals* locals_desc = new CachedLocals();
      assign_registers(locals_desc, scan_result, locals, free_int_regs, loop);
      bool set_32bit_precision = scan_result.has_floats() && !scan_result.has_doubles();
      add_block_items(loop, locals_desc->_mapping, locals_desc->_is_oop, set_32bit_precision);
    }
  }
}


CachedLocals* LocalCaching::cache_locals_globally() {
  ShouldNotReachHere(); // SPARC only
  return NULL;
}


bool LocalCaching::can_cache_receiver(ScanResult* scan_result) {
  return !scan_result->has_calls() && !scan_result->has_slow_cases() && !scan_result->has_store_0() && !scan_result->has_class_init();
}


// reg_alloc has information which registers have been used in the method
void LocalCaching::cache_receiver(RegAlloc* reg_alloc) {
  c1_RegMask lockout(ValueGen::receiverRInfo(), 'x' /*dummy*/);
  RInfoCollection* mapping = new RInfoCollection();
  mapping->append(ValueGen::receiverRInfo());
  
  // Note: set the lock_out after we are done with analysis
  BlockItem* item = new BlockItem(mapping, NULL, c1_RegMask::empty_set() /*lockout*/, false); // work without lockout for the moment

  if (!reg_alloc->did_use_register(ValueGen::receiverRInfo())) {
    if (can_cache_receiver(_ir->compilation()->scan_result())) {
      // cache it !
      { BlockItemSetter setter(item);
        _ir->code()->iterate_forward(&setter);
      }
      item->set_lockout(lockout);
      return;
    }
  }
  // Note: do not do it if any JSR's in method
  // assign caching of receiver only to some blocks
  if (UseCachedReceiver2) selective_caching(item);
  item->set_lockout(lockout);
}


// Description:
// - the parameter 'item' contains the caching information for cached receiver
// - we try to set this block item into as many blocks as possible
// - a block cannot cache the receiver if:
//    - it contains calls etc. (see can_cache_receiver)
//    - one of its predecessors cannot cache the receiver (i.e., has block item NULL and was visited)
// - no exception handler entry block can cache the receiver
// - the iteration over block occurs as follow:
//    - start with the first block (set its block item and push it on worklist)
//    - iterate until the worklist is empty
//    - if the current block item is not null we check if we may cache the receiver;
//      if we cannot cache the receiver we set the block item to NULL
//    - every sucessor that was not visited is pushed on the worklist
//    - every visited successor that is being changed from having a block item to not having one
//      is marked as not visited.
void LocalCaching::selective_caching(BlockItem* item) {
  // we cannot handle methods with JSR's -> complicated correct block traversal
  if (_ir->compilation()->scan_result()->has_jsr()) return;
  assert(_ir->is_valid(), "check");
  GrowableArray<bool> mark(BlockBegin::number_of_blocks(), BlockBegin::number_of_blocks(), false);
  GrowableArray<BlockBegin*> workset;
  GrowableArray<BlockBegin*> block_holder; // used for passing the block to ScanBlocks
  block_holder.at_put_grow(0, NULL);
  workset.push(_ir->start());
  _ir->start()->set_block_item(item);

  while (workset.is_nonempty()) {
    BlockBegin* block = workset.pop();
    BlockItem* block_item = block->block_item();
    if (!mark.at(block->block_id())) {
      mark.at_put(block->block_id(), true); // mark as visited
      if (block_item != NULL) {
        // check if we have to clear block_item
        assert(block->number_of_subroutines() == 0, "may not have JSR's");
        block_holder.at_put(0, block);
        ScanBlocks scan (&block_holder);
        ScanResult scan_result;
        scan.scan(&scan_result);
        if (!can_cache_receiver(&scan_result)) {
          block_item = NULL; 
        } else {
          // check if the receiver register is being used in the current block
          RegAlloc reg_alloc;
          ValueGenInvariant vgi(_ir->method(), &reg_alloc, NULL);
          ValueGen visitor(NULL, &vgi, false); // no code for the moment
          CodeGenerator gen(&visitor, &vgi);
          gen.block_do(block);
          if (reg_alloc.did_use_register(ValueGen::receiverRInfo())) {
            block_item = NULL;
          }
          CodeGenerator::clear_instruction_items(block);
        }
      }

      block->set_block_item(block_item); // block_item may have become 

      { // visit exception handlers -> push them on worklist while setting the block item to NULL
        BlockEnd* e = block->end();
        for (int i = block->number_of_exception_handlers() - 1; i >= 0; i--) {
          BlockBegin* exc_handler = block->exception_handler_at(i);
          if (!mark.at(exc_handler->block_id())) {
            // exception handler not visited, push on workset
            exc_handler->set_block_item(NULL); 
            workset.push(exc_handler);
          }
        }
      }

      // Visit all successors:
      BlockEnd* e = block->end();
      for (int i = e->number_of_sux() - 1; i >= 0; i--) {
        BlockBegin* sux = e->sux_at(i);
        if (mark.at(sux->block_id())) {
          if (sux->block_item() == NULL) {
            // block visited and was set to NULL -> cannot cache
          } else if (block_item == NULL) {
            // clear block item of successor
            if (sux->block_item() != NULL) {
              // must propagate the change
              sux->set_block_item(block_item);
              workset.push(sux);
              mark.at_put(sux->block_id(), false);
            }
          }
        } else {
          workset.push(sux);
          sux->set_block_item(block_item);
        }
      }
    } // not marked as visited
  } // while nonempty
}
