/*
 * @(#)interpreterRT_i486.cpp	1.33 99/12/23 21:48:58
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_interpreterRT_i486.cpp.incl"


#define __ _masm->


// Fingerprinter computes a unique fingerprint for a given method. The fingerprint
// is a 32bit int value characterizing the methods signature (incl. the receiver).

class Fingerprinter: public NativeSignatureIterator {
 private:
  int _fingerprint;
  int _shift_count;

 public:
  enum {
    result_feature_size    = 4,
    parameter_feature_size = 2,
    max_size_of_parameters = (BitsPerWord - result_feature_size) / parameter_feature_size
  };

  void pass_int()                                { _fingerprint |= (1 << _shift_count); _shift_count += parameter_feature_size; }
  void pass_long()                               { _fingerprint |= (2 << _shift_count); _shift_count += parameter_feature_size; }
  void pass_object()                             { _fingerprint |= (3 << _shift_count); _shift_count += parameter_feature_size; }

  Fingerprinter(methodHandle method) : NativeSignatureIterator(method) {
    assert(method->size_of_parameters() <= max_size_of_parameters, "too many parameters for fingerprinting");
  }

  int fingerprint() {
    assert(AbstractInterpreter::number_of_result_handlers < (1 << result_feature_size), "adjust result feature size");
    _fingerprint = AbstractInterpreter::BasicType_as_index(method()->result_type());
    _shift_count = result_feature_size;
    iterate();
    return _fingerprint;
  }
};


// Implementation of SignatureHandlerGenerator

void InterpreterRuntime::SignatureHandlerGenerator::move(int from_offset, int to_offset) {
  __ movl(temp(), Address(from(), from_offset * wordSize));
  __ movl(Address(to(), to_offset * wordSize), temp());
}


void InterpreterRuntime::SignatureHandlerGenerator::box(int from_offset, int to_offset) {
  __ leal(temp(), Address(from(), from_offset * wordSize));
  __ cmpl(Address(from(), from_offset * wordSize), 0); // do not use temp() to avoid AGI
  Label L;
  __ jcc(Assembler::notZero, L);
  __ movl(temp(), 0);
  __ bind(L);
  __ movl(Address(to(), to_offset * wordSize), temp());
}


void InterpreterRuntime::SignatureHandlerGenerator::generate() {
  // generate code to handle arguments
  iterate();
  // return result handler
  __ movl(eax, (int)AbstractInterpreter::result_handler(method()->result_type()));
  // return
  __ ret(0);
  __ flush();
}


Register InterpreterRuntime::SignatureHandlerGenerator::from()       { return edi; }
Register InterpreterRuntime::SignatureHandlerGenerator::to()         { return esp; }
Register InterpreterRuntime::SignatureHandlerGenerator::temp()       { return ecx; }


// Implementation of SignatureHandlerLibrary

class SignatureHandlerLibrary: public AllStatic {
 public:
  enum   { size = 1024 };                        // the size of the temporary code buffer
 private:
  static GrowableArray<int    >* _fingerprints;  // the fingerprint collection
  static GrowableArray<address>* _handlers;      // the corresponding handlers
  static u_char                  _buffer[SignatureHandlerLibrary::size]; // the temporary code buffer

  static void initialize() {
    if (_fingerprints != NULL) return;
    _fingerprints = new(ResourceObj::C_HEAP)GrowableArray<int    >(32, true);
    _handlers     = new(ResourceObj::C_HEAP)GrowableArray<address>(32, true);
  }

 public:
  static void add(methodHandle method) {
    if (method->signature_handler() == NULL) {
      // check if we can use customized (fast) signature handler
      if (UseFastSignatureHandlers && method->size_of_parameters() <= Fingerprinter::max_size_of_parameters) {
        // use customized signature handler
        MutexLocker mu(SignatureHandlerLibrary_lock);
        // make sure data structure is initialized
        initialize();
        // lookup method signature's fingerprint
        int fingerprint = Fingerprinter(method).fingerprint();
        int index = _fingerprints->find(fingerprint);
        // create handler if necessary
        if (index < 0) {
          ResourceMark rm;
          CodeBuffer* buffer = new CodeBuffer(_buffer, SignatureHandlerLibrary::size);
          InterpreterRuntime::SignatureHandlerGenerator(method, buffer).generate();
          // copy into C-heap allocated memory location
          address handler = (address)NEW_C_HEAP_ARRAY(u_char, buffer->code_size());
          memcpy(handler, buffer->code_begin(), buffer->code_size());
          // debugging suppport
          if (PrintSignatureHandlers) {
            tty->cr();
            tty->print_cr(
              "argument handler #%d for: %s %s (fingerprint = 0x%x, %d bytes generated)",
              _handlers->length(), (method->is_static() ? "static" : "receiver"), method->signature()->as_C_string(), fingerprint, buffer->code_size()
            );
            Disassembler::decode(handler, handler + buffer->code_size());
          }
          // add handler to library
          _fingerprints->append(fingerprint);
          _handlers->append(handler);
          // set handler index
          assert(_fingerprints->length() == _handlers->length(), "sanity check");
          index = _fingerprints->length() - 1;
        }
        // set handler
        method->set_signature_handler(_handlers->at(index));
      } else {
        // use generic signature handler
        method->set_signature_handler(AbstractInterpreter::slow_signature_handler());
      }
    }
    assert(
      method->signature_handler() == AbstractInterpreter::slow_signature_handler() ||
      _handlers->find(method->signature_handler()) == _fingerprints->find(Fingerprinter(method).fingerprint()),
      "sanity check"
    );
  }
};


GrowableArray<int    >* SignatureHandlerLibrary::_fingerprints = NULL;
GrowableArray<address>* SignatureHandlerLibrary::_handlers     = NULL;
u_char                  SignatureHandlerLibrary::_buffer[SignatureHandlerLibrary::size];


IRT_ENTRY(void, InterpreterRuntime::prepare_native_call(JavaThread* thread, methodOop method))
  methodHandle m(thread, method);
  assert(m->is_native(), "sanity check");
  // make sure signature handler is installed
  SignatureHandlerLibrary::add(m);
  // lookup native function entry point if it doesn't exist
  bool in_base_library;
  if (m->native_function() == NULL) NativeLookup::lookup(m, in_base_library, CHECK);
IRT_END


class SlowSignatureHandler: public NativeSignatureIterator {
 private:
  jint* _from;
  jint* _to;

  virtual void pass_int()                        { *_to++ = *_from--; }
  virtual void pass_long()                       { _to[0] = _from[-1]; _to[1] = _from[0]; _to += 2; _from -= 2; }
  virtual void pass_object()                     { *_to++ = (*_from == 0) ? 0 : (jint)_from; _from--; }

 public:
  SlowSignatureHandler(methodHandle method, jint* from, jint* to) : NativeSignatureIterator(method) {
    _from = from;
    _to   = to;
  }
};


IRT_ENTRY(address, InterpreterRuntime::slow_signature_handler(JavaThread* thread, methodOop method, jint* from, jint* to))
  methodHandle m(thread, method);
  assert(m->is_native(), "sanity check");
  // handle arguments
  SlowSignatureHandler(m, from, to + 1).iterate();
  // return result handler
  return AbstractInterpreter::result_handler(m->result_type());
IRT_END



