/*
 * @(#)nativeInst_i486.hpp	1.50 00/07/13 09:34:14
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// We have interfaces for the following instructions:
// - NativeInstruction
// - - NativeCall
// - - NativeMovConstReg
// - - NativeMovConstRegPatching
// - - NativeMovRegMem
// - - NativeMovRegMemPatching
// - - NativeJump
// - - NativeIllegalOpCode
// - - NativeReturn   
// - - NativeReturnX (return with argument)
// - - NativePushConst

// The base class for different kinds of native instruction abstractions.
// Provides the primitive operations to manipulate code relative to this.

class NativeInstruction VALUE_OBJ_CLASS_SPEC {
  friend class Relocation;

 public:
  enum Intel_specific_constants {
    nop_instruction_code        = 0x90,
    nop_instruction_size        =    1
  };

  bool is_nop()                        { return (char_at(0)&0xFF) == nop_instruction_code; }
  inline bool is_call();
  inline bool is_illegal();
  inline bool is_return();
  inline bool is_jump();
  inline bool is_cond_jump();

 protected:
  address addr_at(int offset) const    { return address(this) + offset; }

  char char_at(int offset) const       { return  (char)*addr_at(offset); }
  int  long_at(int offset) const       { return *(jint*)addr_at(offset); }
  oop  oop_at (int offset) const       { return *(oop*) addr_at(offset); }

  void set_char_at(int offset, char c) { *addr_at(offset) = (u_char)c; wrote(offset); }
  void set_long_at(int offset, int  i) { *(jint*)addr_at(offset) = i;  wrote(offset); }
  void set_oop_at (int offset, oop  o) { *(oop*) addr_at(offset) = o;  wrote(offset); }

  // This doesn't really do anything on Intel, but it is the place where
  // cache invalidation belongs, generically:
  void wrote(int offset);

 public:

  // unit test stuff
  static void test() {}			// override for testing

  friend NativeInstruction* nativeInstruction_at(address address) {
    NativeInstruction* inst = (NativeInstruction*)address;
    #ifdef ASSERT
      //inst->verify();
    #endif
    return inst;
  }
};


// The NativeCall is an abstraction for accessing/manipulating native call imm32
// instructions (used to manipulate inline caches, primitive & dll calls, etc.).

class NativeCall: public NativeInstruction {
 public:
  enum Intel_specific_constants {
    instruction_code	        = 0xE8,
    instruction_size	        =    5,
    instruction_offset	        =    0,
    displacement_offset	        =    1,
    return_address_offset	=    5
  };

  address instruction_address() const       { return addr_at(instruction_offset); }
  address next_instruction_address() const  { return addr_at(return_address_offset); }
  int   displacement() const                { return long_at(displacement_offset); }
  address displacement_address() const      { return addr_at(displacement_offset); }
  address return_address() const            { return addr_at(return_address_offset); }
  address destination() const               { return return_address() + displacement(); }  
  void  set_destination(address dest)       { set_long_at(displacement_offset, dest - return_address()); }

  bool destination_is_compiled_verified_entry_point();

  bool is_call_to_stub() const;   // returns true if calls to stub (destination in same nmethod as call)

  void  verify();
  void  print();
  
  // Creation
  friend NativeCall* nativeCall_at(address address) {
    NativeCall* call = (NativeCall*)(address - instruction_offset);
    #ifdef ASSERT
      call->verify();
    #endif
    return call;
  }

  friend NativeCall* nativeCall_before(address return_address) {
    NativeCall* call = (NativeCall*)(return_address - return_address_offset);
    #ifdef ASSERT
      call->verify();
    #endif
    return call;
  }

  static bool is_call_at(address instr);

  // MT-safe patching of a call instruction.
  static void insert(address code_pos, address entry);
  static void replace_mt_safe(address instr_addr, address code_buffer);  
};


// An interface for accessing/manipulating native mov reg, imm32 instructions.
// (used to manipulate inlined 32bit data dll calls, etc.)
class NativeMovConstReg: public NativeInstruction {
 public:
  enum Intel_specific_constants {
    instruction_code	        = 0xB8,
    instruction_size	        =    5,
    instruction_offset	        =    0,
    data_offset	                =    1,
    next_instruction_offset	=    5,
    register_mask	        = 0x07
  };

  address instruction_address() const       { return addr_at(instruction_offset); }
  address next_instruction_address() const  { return addr_at(next_instruction_offset); }
  int   data() const                        { return long_at(data_offset); }
  void  set_data(int x)                     { set_long_at(data_offset, x); }

  void  verify();
  void  print();
  
  // unit test stuff
  static void test() {}

  // Creation
  friend NativeMovConstReg* nativeMovConstReg_at(address address) {
    NativeMovConstReg* test = (NativeMovConstReg*)(address - instruction_offset);
    #ifdef ASSERT
      test->verify();
    #endif
    return test;
  }


  friend NativeMovConstReg* nativeMovConstReg_before(address address) {
    NativeMovConstReg* test = (NativeMovConstReg*)(address - instruction_size - instruction_offset);
    #ifdef ASSERT
      test->verify();
    #endif
    return test;
  }

};



class NativeMovConstRegPatching: public NativeMovConstReg {
 private:
    friend NativeMovConstRegPatching* nativeMovConstRegPatching_at(address address) {
    NativeMovConstRegPatching* test = (NativeMovConstRegPatching*)(address - instruction_offset);
    #ifdef ASSERT
      test->verify();
    #endif
    return test;
  }
};


// An interface for accessing/manipulating native moves of the form:
//	mov [reg + offset], reg   (instruction_code_reg2mem)
//      mov reg, [reg+offset]     (instruction_code_mem2reg
//      fld_s  [reg+offset]
//      fld_d  [reg+offset]
//	fstp_s [reg + offset]
//	fstp_d [reg + offset]

class NativeMovRegMem: public NativeInstruction {
 public:
  enum Intel_specific_constants {
    instruction_code_reg2mem	        = 0x89,
    instruction_code_mem2reg	        = 0x8b,
    instruction_code_float_s	        = 0xd9,
    instruction_code_float_d	        = 0xdd,
    
    instruction_size	                = 4,
    instruction_offset	                = 0,
    data_offset	                        = 2,
    next_instruction_offset	        = 4
  };
  address instruction_address() const	        { return addr_at(instruction_offset); }
  address next_instruction_address() const	{ return addr_at(next_instruction_offset); }
  int   offset() const	                        { return long_at(data_offset); }
  void  set_offset(int x)	                { set_long_at(data_offset, x); }
  void  add_offset_in_words(int add_offset)     { set_offset ( ( offset() + add_offset ) * wordSize); }
  void  copy_instruction_to(address new_instruction_address);

  void verify();
  void print ();

  // unit test stuff
  static void test() {}

 private:
  friend NativeMovRegMem* nativeMovRegMem_at (address address) {
    NativeMovRegMem* test = (NativeMovRegMem*)(address - instruction_offset);
    #ifdef ASSERT
      test->verify();
    #endif
    return test;
  }
};


class NativeMovRegMemPatching: public NativeMovRegMem {
 private:
  friend NativeMovRegMemPatching* nativeMovRegMemPatching_at (address address) {
    NativeMovRegMemPatching* test = (NativeMovRegMemPatching*)(address - instruction_offset);
    #ifdef ASSERT
      test->verify();
    #endif
    return test;
  }
};



// An interface for accessing/manipulating native leal instruction of form:
//        leal reg, [reg + offset]

class NativeLoadAddress: public NativeMovRegMem {
 public:
  enum Intel_specific_constants {
    instruction_code	        = 0x8D
  };

  void verify();
  void print ();

  // unit test stuff
  static void test() {}

 private:
  friend NativeLoadAddress* nativeLoadAddress_at (address address) {
    NativeLoadAddress* test = (NativeLoadAddress*)(address - instruction_offset);
    #ifdef ASSERT
      test->verify();
    #endif
    return test;
  }
};

// far jump
class NativeJump: public NativeInstruction {
 public:
  enum Intel_specific_constants {
    instruction_code	        = 0xe9,
    instruction_size	        =    5,
    instruction_offset	        =    0,
    data_offset	                =    1,
    next_instruction_offset	=    5
  };

  address instruction_address() const       { return addr_at(instruction_offset); }
  address next_instruction_address() const  { return addr_at(next_instruction_offset); }  
  address jump_destination() const          { return address (long_at(data_offset)+next_instruction_address()); }
  void  set_jump_destination(address dest)  { set_long_at(data_offset, int(dest - next_instruction_address()) ); }

  // Creation
  friend NativeJump* nativeJump_at(address address) {
    NativeJump* jump = (NativeJump*)(address - instruction_offset);
    #ifdef ASSERT
      jump->verify();
    #endif
    return jump;
  }

  void verify();

  // Unit testing stuff
  static void test() {}

  // Insertion of native jump instruction
  static void insert(address code_pos, address entry);
  // MT-safe insertion of native jump at verified method entry
  static void patch_verified_entry(address entry, address verified_entry, address dest);
};

// Handles all kinds of jump on Intel. Long/far, conditional/unconditional
class NativeGeneralJump: public NativeInstruction {
 public:
  enum Intel_specific_constants {
    // Constans does not apply, since the lengths and offsets depends on the accually jump
    // used
    // Instruction codes:
    //   Unconditional jumps: 0xE9    (long), 0xEB (short)
    //   Conditional jumps:   0x0F8x  (long), 0x7x (short)
  };

  address instruction_address() const       { return addr_at(0); }  
  address jump_destination()	const;          

  // Creation
  friend NativeGeneralJump* nativeGeneralJump_at(address address) {
    NativeGeneralJump* jump = (NativeGeneralJump*)(address);
    debug_only(jump->verify();)    
    return jump;
  }

  void verify();
};


class NativePopReg : public NativeInstruction {
 public:
  enum Intel_specific_constants {
    instruction_code	        = 0x58,
    instruction_size	        =    1,
    instruction_offset	        =    0,
    data_offset	                =    1,
    next_instruction_offset	=    1
  };

  // Insert a pop instruction
  static void insert(address code_pos, Register reg);
};


class NativeIllegalInstruction: public NativeInstruction {
 public:
  enum Intel_specific_constants {
    instruction_code	        = 0x0B0F,    // Real byte order is: 0x0F, 0x0B
    instruction_size	        =    2,
    instruction_offset	        =    0,    
    next_instruction_offset	=    2
  };

  // Insert illegal opcode as specific address
  static void insert(address code_pos);
};

// return instruction that does not pop values of the stack
class NativeReturn: public NativeInstruction {
 public:
  enum Intel_specific_constants {
    instruction_code	        = 0xC3, 
    instruction_size	        =    1,
    instruction_offset	        =    0,    
    next_instruction_offset	=    1
  };
};

// return instruction that does pop values of the stack
class NativeReturnX: public NativeInstruction {
 public:
  enum Intel_specific_constants {
    instruction_code	        = 0xC2, 
    instruction_size	        =    2,
    instruction_offset	        =    0,    
    next_instruction_offset	=    2
  };
};

inline bool NativeInstruction::is_illegal()	 { return (short)long_at(0) == (short)NativeIllegalInstruction::instruction_code; }
inline bool NativeInstruction::is_call()	 { return (char_at(0)&0xFF) == NativeCall::instruction_code; }
inline bool NativeInstruction::is_return()	 { return (char_at(0)&0xFF) == NativeReturn::instruction_code ||
 							  (char_at(0)&0xFF) == NativeReturnX::instruction_code; } 
inline bool NativeInstruction::is_jump()         { return (char_at(0)&0xFF) == NativeJump::instruction_code ||
                                                          (char_at(0)&0xFF) == 0xEB; /* short jump */ }
inline bool NativeInstruction::is_cond_jump()    { return (long_at(0)&0xF0FF) == 0x800F /* long jump */ ||
                                                          (char_at(0)&0xF0)   == 0x70;  /* short jump */ }

