/*
 * @(#)c1_RegAlloc_sparc.cpp	1.25 01/02/09 14:43:17
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_c1_RegAlloc_sparc.cpp.incl"




//--------------------------------------------------------
//               c1_AllocTable
//--------------------------------------------------------


int c1_AllocTable::_reg_mask[c1_AllocTable::max];


c1_AllocTable::c1_AllocTable (int size) : _size(size) {
  assert(sizeof(_state) * BitsPerByte >= max, "bit vector must fit in _state");
  assert(max >= size && size >= 0, "too many registers");
  _state    = 0;
  _or_state = 0;
  _lockout  = 0;
  _none_available = (_size == max) ? ~0 : (1 << _size) - 1;
}


bool c1_AllocTable::is_free (int rnr) const {
  debug_only(reg_check(rnr););
  return (_state & _reg_mask[rnr]) == 0;
}


bool c1_AllocTable::did_use_register (int rnr) const {
  debug_only(reg_check(rnr););
  return (_or_state & _reg_mask[rnr]) != 0;
}


bool c1_AllocTable::are_free (int r1, int r2) const {
  return is_free(r1) && is_free(r2);
}


bool c1_AllocTable::is_pair_free (int rnr) const {
  return are_free(rnr, rnr+1);
}


bool c1_AllocTable::has_one_free () const {
  return _state != _none_available;
}


bool c1_AllocTable::has_one_free_masked (c1_RegMask m) const {
  return (_state | (~m._mask & _none_available)) != _none_available;
}


bool c1_AllocTable::has_two_free () const {   // for longs
  if (!has_one_free()) return false;
  c1_AllocTable* self = (c1_AllocTable*)this; // force cast to non-const type
  int r = self->get_free();
  self->set_locked(r);
  bool has_two = self->has_one_free();
  self->set_free(r);
  return has_two;
}


bool c1_AllocTable::has_pair_free () const {   // for doubles
  int mask = 3;
  assert(mask != 0, "too many registers");
  for (int rnr = 0; rnr < _size - 1; rnr += 2, mask <<= 2) {
    if ((_state & mask) == 0)
      return true;
  }
  return false;
}


bool c1_AllocTable::are_all_free () const {
  return _state == 0;
}


void c1_AllocTable::set_free (int rnr) {
  assert(!is_free(rnr), "not locked");     
  _state = _state & ~_reg_mask[rnr];
}


void c1_AllocTable::set_locked (int rnr) {
  assert(is_free(rnr), "already locked");
  assert(!is_lockout(rnr), "lock out check failed");
  _state    |=  _reg_mask[rnr];
  _or_state |= _state; 
}


void c1_AllocTable::set_pair_free (int rnr) {
  assert(!is_free(rnr) && !is_free(rnr+1), "not locked");
  assert(rnr % 2 == 0, "pairs start at even boundaries");
  _state = _state & ~(_reg_mask[rnr] | _reg_mask[rnr+1]);
}


void c1_AllocTable::set_pair_locked (int rnr) {
  assert(is_free(rnr) && is_free(rnr+1) , "already locked");
  assert(!is_lockout(rnr) && !is_lockout(rnr+1), "lock out check failed");
  assert(rnr % 2 == 0, "pairs start at even boundaries");
  _state = _state | _reg_mask[rnr] | _reg_mask[rnr+1];
  _or_state |= _state;
}


int c1_AllocTable::get_free () {
  assert(has_one_free(), "no free reg"); 
  int rnr;
  int mask = 1;
  assert(mask != 0, "too many registers");
  for (rnr = 0; rnr < _size; rnr++, mask <<= 1) {
    if ((_state & mask) == 0) {
      assert(!is_lockout(rnr), "lock out check failed");
      _or_state = _or_state | _reg_mask[rnr];
      return rnr;
    }
  }
  ShouldNotReachHere();
  return -1;
}


int c1_AllocTable::get_pair_free () {
  assert(has_pair_free(), "no free pair"); 
  int rnr;
  int mask = 3;
  assert(mask != 0, "too many registers");
  for (rnr = 0; rnr < _size; rnr += 2, mask <<= 2) {
    if ((_state & mask) == 0) {
      assert(!is_lockout(rnr) && !is_lockout(rnr+1), "lock out check failed");
      _or_state = _or_state | _reg_mask[rnr] | _reg_mask[rnr+1];
      return rnr;
    }
  }
  ShouldNotReachHere();
  return -1;
}


int c1_AllocTable::get_free_masked (c1_RegMask m) {
  assert(has_one_free_masked(m), "no free reg in mask");
  int tmp_state = _state | ~m._mask;
  for (int rnr = 0; rnr < _size; rnr++) {
    if ((tmp_state & _reg_mask[rnr]) == 0) {
      assert(!is_lockout(rnr), "lock out check failed");
      _or_state = _or_state | _reg_mask[rnr];
      return rnr;
    }
  }
  ShouldNotReachHere();
  return -1;
}


c1_RegMask c1_AllocTable::used_registers () const {
  c1_RegMask rm = c1_RegMask::empty_set();
  for (int rnr = 0; rnr < _size; rnr++) {
    if ((_or_state & _reg_mask[rnr]) != 0) {
      rm.add_reg(rnr);
    }
  }
  return rm;
}


c1_RegMask c1_AllocTable::free_registers () const {
  c1_RegMask rm = c1_RegMask::empty_set();
  for (int rnr = 0; rnr < _size; rnr++) {
    if ((_or_state & _reg_mask[rnr]) == 0) {
      rm.add_reg(rnr);
    }
  }
  return rm;
}


void c1_AllocTable::set_lockout (c1_RegMask lockout) {
  _lockout = lockout._mask;
}


void c1_AllocTable::init_tables () {
  for (int i = 0; i < max; i++) {
    _reg_mask[i] = nth_bit(i);
  }
}


//---------------------------------------------------------
//          RegAlloc
//--------------------------------------------------------


bool RegAlloc::_init_done = false;


void RegAlloc::init_register_allocation() {
  if (_init_done) return;
  FrameMap::init();
  c1_AllocTable::init_tables();
  _init_done = true;
}


bool RegAlloc::has_free_reg(ValueType* type) const {
  if (type->is_float()) {
    return _fpu_alloc_table.has_one_free();
  } else if (type->is_double()) {
    return _fpu_alloc_table.has_pair_free();
  } else if (type->is_long()) {
    return _cpu_alloc_table.has_two_free();
  } else {
    return _cpu_alloc_table.has_one_free();
  }
}


RInfo RegAlloc::get_free_reg (ValueType* type) {
  assert(has_free_reg(type), "no free register");
  RInfo rinfo;
  switch (type->tag()) {
  case intTag:
  case objectTag:
  case addressTag:
    rinfo = RInfo::word_reg(_cpu_alloc_table.get_free());
    break;
  case longTag:
    {
      int rnr_lo = _cpu_alloc_table.get_free(); ((RegAlloc*)this)->set_locked_cpu(rnr_lo, NULL, 1); // temp lock
      int rnr_hi = _cpu_alloc_table.get_free(); ((RegAlloc*)this)->set_free_cpu(rnr_lo);
      rinfo.set_long_reg(rnr_lo, rnr_hi);
    }
    break;  
  case floatTag:
    rinfo.set_float_reg(_fpu_alloc_table.get_free());
    break;
  case doubleTag:
    rinfo.set_double_reg(_fpu_alloc_table.get_pair_free());
    break;
  default:
    ShouldNotReachHere();
    break;
  };
  return rinfo;
}


bool RegAlloc::is_free_double (int rnr) const {
  return _fpu_alloc_table.is_pair_free(rnr);
}


void RegAlloc::set_free_double(int rnr) {
  debug_only(fpu_range_check(rnr);)
  assert(_fpu_rc[rnr] > 0, "wrong ref count");
  _fpu_rc[rnr]--;
  _fpu_rc[rnr+1]--;
  if (_fpu_rc[rnr] == 0) _fpu_alloc_table.set_pair_free(rnr);
}


void RegAlloc::set_locked_double(int rnr, Value instr, int rc) {
  debug_only(fpu_range_check(rnr);)
  assert(rc > 0, "wrong rc");
  _fpu_alloc_table.set_pair_locked(rnr);
  _fpu_rc[rnr]     = rc;
  _fpu_rc[rnr+1]   = rc;
  _fpu_values[rnr] = instr;
}


void RegAlloc::set_double_reg (int rnr, int rc, Value value) {
  assert(!is_free_double(rnr), "register not locked");
  assert(rc > 0, "would free register");
  _fpu_rc[rnr]     = rc;
  _fpu_values[rnr] = value;
}


int RegAlloc::get_double_rc (int rnr) const {
  assert(!is_free_double(rnr), "register not locked");
  return _fpu_rc[rnr];
}


Value RegAlloc::get_double_val(int rnr) const {
  assert(!is_free_double(rnr), "register not locked");
  return _fpu_values[rnr];
}


bool RegAlloc::is_locked_double_spill_count (int rnr) const {
  return _fpu_slock[rnr] != 0 && _fpu_slock[rnr+1] != 0;
}


void RegAlloc::change_double_spill_count (int rnr, int d) {
  _fpu_slock[ rnr ] += d; assert(_fpu_slock[ rnr ] >= 0, "illegal");
  _fpu_slock[rnr+1] += d; assert(_fpu_slock[rnr+1] >= 0, "illegal");
}


Value RegAlloc::get_smallest_value_to_spill(ValueType* type) const {
  Value smallest_v = NULL;

  if (type->is_float()) {
    for (int i = 0; i < nof_fpu_regs; i++) {
      Value v = NULL;
      if (!is_free_fpu(i) && !is_locked_fpu_spill_count(i)) v = _fpu_values[i];
      // searching for value with smallest root
      // Sometimes we have a value whose item is not a register: this happens when
      // we are just handling the value and therefore do not spill it
      if (v != NULL && v->item() != NULL && v->item()->is_register()) {
	if (smallest_v == NULL || smallest_v->bci() > v->bci()) {
	  smallest_v = v;
	}
      }
    }
  }
  else if (type->is_double()) {
    for (int i = 0; i < nof_fpu_regs; i += 2) {
      // Search for a double to spill
      Value v = NULL;
      if (!is_free_double(i) && !is_locked_double_spill_count(i)) v = _fpu_values[i];
      if (v == NULL || !v->type()->is_double())
        continue;
      // searching for value with smallest root
      // Sometimes we have a value whose item is not a register: this happens when
      // we are just handling the value and therefore do not spill it
      if (v->item() != NULL && v->item()->is_register()) {
	if (smallest_v == NULL || smallest_v->bci() > v->bci()) {
	  smallest_v = v;
	}
      }
    }
    if (smallest_v == NULL) {
      // No double item to spill. Spill upto two floats (one
      // per call to this function)
      for (int i = 0; i < nof_fpu_regs; i += 2) {
        // Search for a one/two floats to spill
        Value v = NULL;
        if (!is_free_fpu(i) && !is_locked_fpu_spill_count(i)) {
	  v = _fpu_values[i];
        } else if (!is_free_fpu(i + 1) && !is_locked_fpu_spill_count(i+1)) {
	  v = _fpu_values[i + 1];
	}
        // searching for value with smallest root
	// Sometimes we have a value whose item is not a register: this happens when
	// we are just handling the value and therefore do not spill it
	if (v != NULL && v->item() != NULL && v->item()->is_register()) {
	  if (smallest_v == NULL || smallest_v->bci() > v->bci()) {
	    smallest_v = v;
	  }
	}
      }
    }
  }
  else {
    for (int i = 0; i < nof_cpu_regs; i++) {
      Value v = NULL;
      if (!is_free_cpu(i) && !is_locked_cpu_spill_count(i)) v = _cpu_values[i];
      // searching for value with smallest root
      // Sometimes we have a value whose item is not a register: this happens when
      // we are just handling the value and therefore do not spill it
      if (v != NULL && v->item() != NULL && v->item()->is_register()) {
	if (smallest_v == NULL || smallest_v->bci() > v->bci()) {
	  smallest_v = v;
	}
      }
    }
  }

  assert(smallest_v != NULL, "no spillable root found");
  assert(smallest_v->item()->is_register(), "not a register!");
  assert(!is_spill_locked(smallest_v->item()->get_register()), "is spill-locked");
  return smallest_v;
}
