/*
 * @(#)frame_sparc.hpp	1.54 00/05/08 13:08:51
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// A frame represents a physical stack frame (an activation).  Frames can be
// C or Java frames, and the Java frames can be interpreted or compiled.
// In contrast, vframes represent source-level activations, so that one physical frame 
// can correspond to multiple source level frames because of inlining.
// A frame is comprised of {pc, sp, younger_sp}


// Layout of interpreter frame:
//
//  0xfffffff
//  ......
// [last  extra incoming arg,  (local # Nargs > 6 ? Nargs-1 : undef)]
// .. Note: incoming args are copied to local frame area upon entry
// [first extra incoming arg,  (local # Nargs > 6 ? 6       : undef)]
// [6 words for C-arg storage (unused)] Are this and next one really needed? 
// [C-aggregate-word (unused)] Yes, if want extra params to be  in same place as C convention
// [16 words for register saving]                                    <--- FP
// [interpreter_frame_vm_locals ] (see below)

// [interpreter-memory-locals], if any, 

// [last local, i.e. local # Nlocals-1]
// ...
// [7th local] (if nlocals >= 6), Note: locals 0-5 reside in I0-I5, Java locals include argumetns
//              Llocals points to 6 words before here, so that Llocals[index] points to mem local
//              Note: Llocals is always double-word aligned

// [monitors                 ]
// ....                       
// [monitors                 ]    <-- Lmonitors (same as Llocals + 6*4 if none)
//                                    (must be double-word aligned because monitor element size is constrained to doubleword)

//                                <-- Lesp (points 1 past TOS)
// [last word used for stack ]
// ...
// [first word used for stack]    (first word of stack is douuble-word aligned)

// [space for outgoing args (conservatively allocated as max_stack - 6 + interpreter_frame_extra_outgoing_argument_words)]
// [6 words for C-arg storage]
// [C-aggregate-word (unused)]
// [16 words for register saving]                                    <--- SP
// ...
// 0x0000000
//
// The in registers and local registers are preserved in a block at SP.
//
// The first six in registers (I0..I5) hold the first six locals.
// The locals are used as follows:
//    Lesp         first free element of expression stack
//                 (which grows towards __higher__ addresses)
//    Lbcp         is set to address of bytecode to execute
//		   It is accessed in the frame under the name "bcx".
//		   It may at times (during GC) be an index instead.
//    Lmethod      the method being interpreted
//    Llocals      the base pointer for accessing the locals array
//                 (lower-numbered locals have lower addresses)
//    Lmonitors    the base pointer for accessing active monitors
//    Lcache       a saved pointer to the method's constant pool cache
//    
//
// When calling out to another method,
// Gmethod is set to method to call, Ginline_cache_klass may be set,
// parameters are put in O registers, and also extra parameters
// must be cleverly copied from the top of stack to the outgoing param area in the frame,



// All frames: 

 public:

  enum {
    pc_return_offset      = 2, // normal sparc return is 2 words past PC

    // size of each block, in order of increasing address:
    register_save_words                          = 16,
    callee_aggregate_return_pointer_words        =  1,
    callee_register_argument_save_area_words     =  6,
    // memory_parameter_words                    = <arbitrary>,

    // offset of each block, in order of increasing address:
    register_save_words_sp_offset                = 0,
    callee_aggregate_return_pointer_sp_offset    = register_save_words_sp_offset + register_save_words,
    callee_register_argument_save_area_sp_offset = callee_aggregate_return_pointer_sp_offset + callee_aggregate_return_pointer_words,
    memory_parameter_word_sp_offset              = callee_register_argument_save_area_sp_offset + callee_register_argument_save_area_words,
    varargs_offset                               = memory_parameter_word_sp_offset
  };

  // Note: callee_register_argument_save_area_words == Assembler::n_register_parameters.

 private:
  // The _pc value is the raw return address, plus 8 (pc_return_offset*4).

  jint*   _younger_sp; // optional next-younger SP (used to locate O7)

  int     _interpreter_sp_adjustment_offset; // intepreter adjust the stack pointer
                                             // to make all locals contiguous

  // Note:  On SPARC, unlike Intel, the saved PC for a stack frame
  // is stored at a __variable__ distance from that frame's SP.
  // (In fact, it may be in the register save area of the callee frame,
  // but that fact need not bother us.)  Thus, we must store the
  // address of that saved PC explicitly.  On the other hand, SPARC
  // stores the FP for a frame at a fixed offset from the frame's SP,
  // so there is no need for a separate "frame::_fp" field.

 public:
  // Accessors

  jint* younger_sp() const {
    assert(_younger_sp != NULL, "frame must possess a younger_sp");
    return _younger_sp;
  }

  // Constructors

  // This constructor relies on the fact that the creator of a frame
  // has flushed register windows which the frame will refer to, and
  // that those register windows will not be reloaded until the frame is
  // done reading and writing the stack.  Moreover, if the "younger_pc"
  // argument points into the register save area of the next younger
  // frame (though it need not), the register window for that next
  // younger frame must also stay flushed.  (The caller is responsible
  // for ensuring this.)

  frame(jint* sp, jint* younger_sp, unsigned int pc_adjustment = 0, bool younger_frame_is_interpreted = false);
  // make a deficient frame which doesn't know where its PC is:
  enum unpatchable_t { unpatchable };
  frame(jint* sp, unpatchable_t, address pc = NULL) {
    _sp = sp;
    _younger_sp = NULL;
    _pc = pc;
    _interpreter_sp_adjustment_offset = 0;
  }

  // this method has been added to enable passing in of a flag which indicates
  // whether the pc needs to be adjusted (in case the frame is a safepoint patched
  // frame). Adding the method here avoids changes to the shared sender method
  frame sender_with_pc_adjustment(RegisterMap* map, CodeBlob *cb, bool pc_adjustment) const;

 public:
  // Given the next-younger sp for a given frame's sp, compute the frame.
  // We need the next-younger sp, because its register save area
  // holds the flushed copy of its I7, which is the PC of the frame
  // we are interested in.
  
  frame after_save() const { return frame(younger_sp(), unpatchable); }

  // Walk outward looking for old_sp, and return its predecessor.
  // Start walking at very_young_sp, or if NULL, do a register window
  // flush and start with the youngest possible.

  static jint* next_younger_sp_or_null(jint* old_sp, jint* young_sp = NULL, jint* sp = NULL);

  static jint* next_younger_sp(jint* old_sp, jint* young_sp = NULL) { 
    jint* sp = next_younger_sp_or_null(old_sp, young_sp);
    assert(sp != NULL, "missed the sp");
    return sp;
  }

  static bool is_valid_stack_pointer(jint* old_sp, jint* young_sp, jint* sp) { return (next_younger_sp_or_null(old_sp, young_sp, sp) != NULL); }

 public:
  // accessors for the instance variables
  jint*   fp() const { return (jint*) sp()[FP.sp_offset_in_saved_window()]; }

  // patching operations
  void   patch_pc(Thread* thread, address pc);

  // All frames

  jint*  fp_addr_at(int index) const   { return &fp()[index];    }
  jint*  sp_addr_at(int index) const   { return &sp()[index];    }
  int    fp_at(     int index) const   { return *fp_addr_at(index); }
  int    sp_at(     int index) const   { return *sp_addr_at(index); }

 private:
  inline address* I7_addr() const;
  inline address* O7_addr() const;

  inline address* I0_addr() const;
  inline address* O0_addr() const;

 public:
  // access to SPARC arguments and argument registers

  jint*     register_addr(Register reg) const {
    return sp_addr_at(reg.sp_offset_in_saved_window());
  }
  jint* memory_param_addr(int param_ix, bool is_in) const {
    int offset = callee_register_argument_save_area_sp_offset + param_ix;
    if (is_in)
      return fp_addr_at(offset);
    else
      return sp_addr_at(offset);
  }
  jint*        param_addr(int param_ix, bool is_in) const {
    if (param_ix >= callee_register_argument_save_area_words)
      return memory_param_addr(param_ix, is_in);
    else if (is_in)
      return register_addr(Argument(param_ix, true).as_register());
    else {
      // the registers are stored in the next younger frame
      // %%% is this really necessary?
      frame next_younger = after_save();
      return next_younger.register_addr(Argument(param_ix, true).as_register());
    }
  }


  // Interpreter frames

 public:
  enum interpreter_frame_vm_locals {
       // 2 words, also used to save float regs across  calls to C
       interpreter_frame_d_scratch_fp_offset          = -2, 
       interpreter_frame_l_scratch_fp_offset          = -4, 
       interpreter_frame_padding_offset               = -5, // for native calls only
       interpreter_frame_mirror_offset                = -6, // for native calls only
       interpreter_frame_vm_locals_fp_offset          = -6, // should be same as above, and should be zero mod 8

       interpreter_frame_vm_local_words = -interpreter_frame_vm_locals_fp_offset,


       // interpreter frame set-up needs to save 2 extra words in outgoing param area
       // for class and jnienv arguments for native stubs (see nativeStubGen_sparc.cpp_

       interpreter_frame_extra_outgoing_argument_words = 2
  };

  // the compiler frame has many of the same fields as the interpreter frame
  // %%%%% factor out declarations of the shared fields
  enum compiler_frame_fixed_locals {
       compiler_frame_d_scratch_fp_offset          = -2, 
       compiler_frame_vm_locals_fp_offset          = -2, // should be same as above

       compiler_frame_vm_local_words = -compiler_frame_vm_locals_fp_offset
  };

 private:

  // where LcpoolCache is saved:
  constantPoolCacheOop* interpreter_frame_cpoolcache_addr() const { 
    return (constantPoolCacheOop*)sp_addr_at( LcpoolCache.sp_offset_in_saved_window());
  }

  // where Lmonitors is saved:
  BasicObjectLock**  interpreter_frame_monitors_addr() const { 
    return (BasicObjectLock**) sp_addr_at( Lmonitors.sp_offset_in_saved_window());
  }
  jint** interpreter_frame_esp_addr() const { 
    return (jint**)sp_addr_at( Lesp.sp_offset_in_saved_window());
  }

  inline void interpreter_frame_set_tos_address(jint* x);


  // %%%%% Another idea: instead of defining 3 fns per item, just define one returning a ref 

  // monitors:

  // next two fns read and write Lmonitors value, 
 private:
  BasicObjectLock* interpreter_frame_monitors()           const  { return *interpreter_frame_monitors_addr(); }
  void interpreter_frame_set_monitors(BasicObjectLock* monitors) {        *interpreter_frame_monitors_addr() = monitors; }



 // Compiled frames

 public:
  // Tells if this register can hold 64 bits on V9 (really, V8+).
  static bool holds_a_doubleword(Register reg) {
    return reg.is_out() || reg.is_global();
  }



//Reconciliation History
// 1.9 97/08/04 15:12:36 frame_i486.hpp
// 1.25 97/09/26 12:49:58 frame_i486.hpp
// 1.29 97/10/17 00:46:29 frame_i486.hpp
// 1.32 97/11/24 14:07:58 frame_i486.hpp
// 1.33 97/12/11 19:04:03 frame_i486.hpp
// 1.34 98/01/21 18:42:45 frame_i486.hpp
// 1.35 98/02/04 02:16:35 frame_i486.hpp
// 1.37 98/03/12 16:11:38 frame_i486.hpp
// 1.39 98/04/30 16:37:45 frame_i486.hpp
// 1.41 98/05/06 08:19:13 frame_i486.hpp
// 1.42 98/05/07 14:18:24 frame_i486.hpp
// 1.46 98/06/20 12:04:22 frame_i486.hpp
// 1.47 98/10/02 08:58:35 frame_i486.hpp
// 1.48 99/01/06 16:18:51 frame_i486.hpp
// 1.49 99/02/19 14:11:36 frame_i486.hpp
// 1.50 99/04/12 12:56:44 frame_i486.hpp
// 1.52 99/06/29 18:05:13 frame_i486.hpp
// 1.53 99/07/16 11:12:11 frame_i486.hpp
//End
