/*
 * @(#)interpreterRT_sparc.cpp	1.33 01/02/09 14:43:24
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_interpreterRT_sparc.cpp.incl"


#define __ _masm->


// Fingerprinter computes a unique fingerprint for a given method. The fingerprint
// is a 32bit int value characterizing the methods signature (incl. the receiver).

class Fingerprinter: public NativeSignatureIterator {
 private:
  int _fingerprint;
  int _shift_count;

 public:
  enum {
    static_feature_size    = 1,
    result_feature_size    = 4,
    parameter_feature_size = 2,
    max_size_of_parameters = (BitsPerWord-1 - result_feature_size - static_feature_size) / parameter_feature_size
  };

  void pass_int()                                { _fingerprint |= (1 << _shift_count); _shift_count += parameter_feature_size; }
  void pass_long()                               { _fingerprint |= (2 << _shift_count); _shift_count += parameter_feature_size; }
  void pass_object()                             { _fingerprint |= (3 << _shift_count); _shift_count += parameter_feature_size; }
  void pass_double()                             { _fingerprint |= (0 << _shift_count); _shift_count += parameter_feature_size; }

  Fingerprinter(methodHandle method) : NativeSignatureIterator(method) {
    assert(method->size_of_parameters() <= max_size_of_parameters, "too many parameters for fingerprinting");
  }

  int fingerprint() {
    assert(AbstractInterpreter::number_of_result_handlers < (1 << result_feature_size), "adjust result feature size");
    _fingerprint = AbstractInterpreter::BasicType_as_index(method()->result_type());
    _fingerprint <<= static_feature_size;
    if (is_static())  _fingerprint |= 1;
    _shift_count = result_feature_size + static_feature_size;
    iterate();
    _fingerprint |= 1 << _shift_count;    // leading bit declares arg count
    return _fingerprint;
  }
};


// Implementation of SignatureHandlerGenerator

void InterpreterRuntime::SignatureHandlerGenerator::pass_word(int size_of_arg, int offset_in_arg) {
  Argument  jni_arg(jni_offset() + offset_in_arg, false);
  Register     Rtmp = O0;
  __ ld(Llocals, -offset() * wordSize, Rtmp);

  __ store_argument(Rtmp, jni_arg);
}

void InterpreterRuntime::SignatureHandlerGenerator::pass_long() {
  Argument  jni_arg(jni_offset(), false);
  Register  Rtmp = O0;
  __ ld(Llocals, -(offset() + 1) * wordSize, Rtmp);
  __ store_argument(Rtmp, jni_arg            );
  __ ld(Llocals, -(offset() + 0) * wordSize, Rtmp);
  __ store_argument(Rtmp, jni_arg.successor());
}


void InterpreterRuntime::SignatureHandlerGenerator::pass_double() {
  Argument  jni_arg(jni_offset(), false);
  Register  Rtmp = O0;

  __ ld(Llocals, -(offset() + 1) * wordSize, Rtmp);
  __ store_argument(Rtmp, jni_arg);
  __ ld(Llocals, -offset() * wordSize, Rtmp);
  __ store_argument(Rtmp, jni_arg.successor());
}

void InterpreterRuntime::SignatureHandlerGenerator::pass_object() {
  Argument  jni_arg(jni_offset(), false);
  Argument java_arg(    offset(), true);
  Register    Rtmp1 = O0;
  Register    Rtmp2 =  jni_arg.is_register() ?  jni_arg.as_register() : O0;


  // the handle for a receiver will never be null
  bool do_NULL_check = offset() != 0 || is_static();

  Address     h_arg = Address(Llocals, 0, -offset() * wordSize);
  __ ld(h_arg, Rtmp1);
  if (!do_NULL_check) {
    __ add(h_arg, Rtmp2);
  } else {
    if (Rtmp1 == Rtmp2)
          __ tst(Rtmp1);
    else  __ addcc(G0, Rtmp1, Rtmp2); // optimize mov/test pair
    Label L;
    __ br(Assembler::notZero, true, Assembler::pt, L);
    __ delayed()->add(h_arg, Rtmp2);
    __ bind(L);
  }
  __ store_argument(Rtmp2, jni_arg);    // this is often a no-op
}


void InterpreterRuntime::SignatureHandlerGenerator::generate() {

  // generate code to handle arguments
  iterate();

  // return result handler
  Address result_handler(Lscratch, AbstractInterpreter::result_handler(method()->result_type()));
  __ sethi(result_handler);
  __ retl();
  __ delayed()->add(result_handler, result_handler.base());

  __ flush();
}


// Implementation of SignatureHandlerLibrary

class SignatureHandlerLibrary: public AllStatic {
 public:
  enum   { size = 1024 };                        // the size of the temporary code buffer
 private:
  static GrowableArray<int    >* _fingerprints;  // the fingerprint collection
  static GrowableArray<address>* _handlers;      // the corresponding handlers
  static u_char                  _buffer[SignatureHandlerLibrary::size];  // the temporary code buffer

  static void initialize() {
    if (_fingerprints != NULL) return;
    _fingerprints = new(ResourceObj::C_HEAP)GrowableArray<int    >(32, true);
    _handlers     = new(ResourceObj::C_HEAP)GrowableArray<address>(32, true);
  }

 public:
  static void add(methodHandle method) {
    if (method->signature_handler() == NULL) {
      // check if we can use customized (fast) signature handler
      if (UseFastSignatureHandlers && method->size_of_parameters() <= Fingerprinter::max_size_of_parameters) {
        // use customized signature handler
        MutexLocker mu(SignatureHandlerLibrary_lock);
        // make sure data structure is initialized
        initialize();
        // lookup method signature's fingerprint
        int fingerprint = Fingerprinter(method).fingerprint();
        int index = _fingerprints->find(fingerprint);
        // create handler if necessary
        if (index < 0) {
          ResourceMark rm;
          CodeBuffer* buffer = new CodeBuffer(_buffer, SignatureHandlerLibrary::size);
          InterpreterRuntime::SignatureHandlerGenerator(method, buffer).generate();
          // copy into C-heap allocated memory location
          address handler = (address)NEW_C_HEAP_ARRAY(u_char, buffer->code_size());
          memcpy(handler, buffer->code_begin(), buffer->code_size());
          ICache::invalidate_range(handler, buffer->code_size());
          // debugging suppport
          if (PrintSignatureHandlers) {
            tty->cr();
            tty->print_cr(
              "argument handler #%d for: %s %s (fingerprint = 0x%x, %d bytes generated)",
              _handlers->length(), (method->is_static() ? "static" : "receiver"), method->signature()->as_C_string(), fingerprint, buffer->code_size()
            );
            Disassembler::decode(handler, handler + buffer->code_size());
            #ifndef PRODUCT
            tty->print_cr(" --- associated result handler ---");
            address rh_begin = AbstractInterpreter::result_handler(method()->result_type());
            address rh_end = rh_begin;
            while (*(long*)rh_end != 0)  rh_end += sizeof(long);
            Disassembler::decode(rh_begin, rh_end);
            #endif
          }
          // add handler to library
          _fingerprints->append(fingerprint);
          _handlers->append(handler);
          // set handler index
          assert(_fingerprints->length() == _handlers->length(), "sanity check");
          index = _fingerprints->length() - 1;
        }
        // set handler
        method->set_signature_handler(_handlers->at(index));
      } else {
        // use generic signature handler
        method->set_signature_handler(AbstractInterpreter::slow_signature_handler());
      }
    }
    assert(
      method->signature_handler() == AbstractInterpreter::slow_signature_handler() ||
      _handlers->find(method->signature_handler()) == _fingerprints->find(Fingerprinter(method).fingerprint()),
      "sanity check"
    );
  }
};


GrowableArray<int    >* SignatureHandlerLibrary::_fingerprints = NULL;
GrowableArray<address>* SignatureHandlerLibrary::_handlers     = NULL;
u_char                  SignatureHandlerLibrary::_buffer[SignatureHandlerLibrary::size];


IRT_ENTRY(void, InterpreterRuntime::prepare_native_call(JavaThread* thread, methodOop method))
  methodHandle m(thread, method);
  assert(m->is_native(), "sanity check");
  // lookup native function entry point if it doesn't exist
  bool in_base_library;
  if (m->native_function() == NULL) NativeLookup::lookup(m, in_base_library, CHECK);
  // make sure signature handler is installed
  SignatureHandlerLibrary::add(m);

  // The interpreter entry point checks the signature handler first,
  // before trying to fetch the native entry point and klass mirror.
  // We must set the signature handler last, so that multiple processors
  // preparing the same method will be sure to see non-null entry & mirror.
IRT_END


class SlowSignatureHandler: public NativeSignatureIterator {
 private:
  jint* _from;
  jint* _to;

  virtual void pass_int()                        { *_to++ = *_from--; }
  virtual void pass_object()                     { *_to++ = (*_from == 0) ? NULL : (jint)_from; _from--; }
  virtual void pass_long()                       { _to[0] = _from[-1]; _to[1] = _from[0]; _to += 2; _from -= 2; }
  virtual void pass_double()                     { _to[0] = _from[-1]; _to[1] = _from[0]; _to += 2; _from -= 2; }

 public:
  SlowSignatureHandler(methodHandle method, jint* from, jint* to) : NativeSignatureIterator(method) {
    _from = from;
    _to   = to;
  }
};


IRT_ENTRY(address, InterpreterRuntime::slow_signature_handler(JavaThread* thread, methodOop method, jint* from, jint* to))
  methodHandle m(thread, method);
  assert(m->is_native(), "sanity check");
  // handle arguments
  SlowSignatureHandler(m, from, m->is_static() ? to+1 : to ).iterate();
  // return result handler
  return AbstractInterpreter::result_handler(m->result_type());
IRT_END


//Reconciliation History
// 1.1 98/03/27 17:28:09 interpreterRT_i486.cpp
// 1.2 98/04/28 16:52:51 interpreterRT_i486.cpp
// 1.12 98/05/06 14:32:04 interpreterRT_i486.cpp
// 1.14 98/05/21 11:55:35 interpreterRT_i486.cpp
// 1.15 98/05/29 17:36:55 interpreterRT_i486.cpp
// 1.18 98/06/18 11:01:20 interpreterRT_i486.cpp
// 1.20 98/06/23 18:19:24 interpreterRT_i486.cpp
// 1.21 98/07/02 15:03:21 interpreterRT_i486.cpp
// 1.22 98/07/13 10:36:13 interpreterRT_i486.cpp
// 1.23 98/07/28 18:44:23 interpreterRT_i486.cpp
// 1.24 98/11/23 15:35:11 interpreterRT_i486.cpp
// 1.25 98/11/25 15:47:54 interpreterRT_i486.cpp
// 1.26 99/03/11 08:23:51 interpreterRT_i486.cpp
// 1.28 99/03/26 13:37:10 interpreterRT_i486.cpp
// 1.31 99/06/28 11:00:13 interpreterRT_i486.cpp
//End
