/*
 * @(#)os_linux.hpp	1.13 01/03/08
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// Linux_OS defines the interface to Linux operating systems

class Linux {
  friend class os;
  friend class atomic;
  friend void os_init();

  // These are really "C" linkage, compiler doesnt seem to notice and complain yet
  static jint (*atomic_exchange_func)              (jint, jint*);
  static void (*atomic_store_func)                 (jint, jint*);
  static jint (*atomic_compare_and_exchange_func)  (jint, jint*, jint);
  static jint (*atomic_increment_func)             (jint, jint*);

  static jint atomic_exchange_bootstrap            (jint, jint*);
  static void atomic_store_bootstrap               (jint, jint*);
  static jint atomic_compare_and_exchange_bootstrap(jint, jint*, jint);
  static jint atomic_increment_bootstrap           (jint, jint*);

  // For signal-chaining
#define MAXSIGNUM 32
  static struct sigaction sigact[MAXSIGNUM]; // saved preinstalled sigactions
  static unsigned int sigs;             // mask of signals that have
                                        // preinstalled signal handlers
  static bool libjsig_is_loaded;        // libjsig that interposes sigaction(),
                                        // __sigaction(), signal() is loaded
  static struct sigaction *(*get_signal_action)(int);
  static struct sigaction *get_preinstalled_handler(int);
  static void save_preinstalled_handler(int, struct sigaction&);

 protected:

  static int _processor_count;

  static int processor_count() {
    return _processor_count;
  }

 public:

  static void init_thread_fpu_state();

  // This boolean allows users to forward their own non-matching signals
  // to JVM_handle_linux_signal, harmlessly.
  static bool signal_handlers_are_installed;

  static void install_signal_handlers();
  static void set_signal_handler(int, bool);

  static sigset_t* unblocked_signals();

  // For signal-chaining
  static struct sigaction *get_chained_signal_action(int sig);
  static bool chained_handler(struct sigaction *actp, int sig,
                              siginfo_t *siginfo, void *context);

  // Tells whether we're running on an MP machine
  static bool is_MP() {
    return processor_count() > 1;
  };

  // Stack overflow handling

  static int  check_stack_overflow(address addr);
  static void install_alternate_signal_stack();
  static void remove_alternate_signal_stack();
  static int max_register_window_saves_before_flushing();

  // Stack repair handling

  // none present 

  // An event is a condition variable with associated mutex.
  // (A cond_t is only usable in combination with a mutex_t.)
  class Event : public CHeapObj {
   private:
    pthread_mutex_t _mutex[1];
    pthread_cond_t  _cond[1];
    volatile int    _count;
    
   public:
    Event() {
      verify();
      int status;
      status = pthread_cond_init(_cond, NULL);
      assert(status == 0, "cond_init");
      status = pthread_mutex_init(_mutex, NULL);
      assert(status == 0, "mutex_init");
      _count = 0;
    }
    ~Event() {
      int status;
      status = pthread_cond_destroy(_cond);
      assert(status == 0, "cond_destroy");
      status = pthread_mutex_destroy(_mutex);
      assert(status == 0, "mutex_destroy");
    }
    // hook to check for mutex corruption:
    void verify() PRODUCT_RETURN;
    // for use in critical sections:
    void lock() {
      verify();
      int status = pthread_mutex_lock(_mutex);
      assert(status == 0, "mutex_lock");
    }
    bool trylock() {
      verify();
      int status = pthread_mutex_trylock(_mutex);
      if (status == EBUSY) {
	return false;
      }
      assert(status == 0, "mutex_lock");
      return true;
    }
    void unlock() {
      verify();
      int status = pthread_mutex_unlock(_mutex);
      assert(status == 0, "mutex_unlock");
    }
    int timedwait(timespec* abstime) {
      verify();
      int status = pthread_cond_timedwait(_cond, _mutex, abstime);
      assert(status == 0 || status == EINTR || status == ETIME || status == ETIMEDOUT, "cond_timedwait");
      return status;
    }
    int timedwait(jlong millis) {
      timespec abst;
      Event::compute_abstime(&abst, millis);
      return timedwait(&abst);
    }
    int wait() {
      verify();
      int status = pthread_cond_wait(_cond, _mutex);
      // for some reason, under 2.7 lwp_cond_wait() may return ETIME ...
      // Treat this the same as if the wait was interrupted
      if(status == ETIME) {
	status = EINTR;
      }
      assert(status == 0 || status == EINTR, "cond_wait");
      return status;
    }
    void signal() {
      verify();
      int status = pthread_cond_signal(_cond);
      assert(status == 0, "cond_signal");
    }
    void broadcast() {
      verify();
      int status = pthread_cond_broadcast(_cond);
      assert(status == 0, "cond_broadcast");
    }
    // functions used to support monitor and interrupt
    void down() {
      verify();
      lock();
      while (_count <= 0) {
        wait();
      }
      _count = 0;
      unlock();
    }
    int down(jlong millis) {
      verify();
      int ret = OS_TIMEOUT;
      lock();
      if (_count <= 0) {
        timedwait(millis);
      }
      if (_count > 0) {
        _count = 0;
        ret = OS_OK;
      }
      unlock();
      return ret;
    }
    void set() {
      verify();
      lock();
      _count = 1;
      broadcast();
      unlock();
    }
    void reset() {
      verify();
      _count = 0;
    }
    // utility to compute the abstime argument to timedwait:
    static struct timespec* compute_abstime(timespec* abstime, jlong millis) {
      // millis is the relative timeout time
      // abstime will be the absolute timeout time
      if (millis < 0)  millis = 0;
      struct timeval now;
      int status = gettimeofday(&now, NULL);
      assert(status == 0, "gettimeofday");
      jlong seconds = millis / 1000;
      millis %= 1000;
      if (seconds > 50000000) { // see man cond_timedwait(3T)
        seconds = 50000000;
      }
      abstime->tv_sec = now.tv_sec  + seconds;
      long       usec = now.tv_usec + millis * 1000;
      if (usec >= 1000000) {
        abstime->tv_sec += 1;
        usec -= 1000000;
      }
      abstime->tv_nsec = usec * 1000;
      return abstime;
    }
  };

  // An OSMutex is an abstraction used in the implementation of
  // ObjectMonitor; needed to abstract over the different thread
  // libraries' mutexes on Solaris.
  class OSMutex : public CHeapObj {
   private:
    #ifndef PRODUCT
    debug_only(volatile pthread_t _owner;)
    debug_only(volatile bool      _is_owned;)
    #endif
    pthread_mutex_t _mutex[1];

   public:
    OSMutex() {
      verify();
      int status = pthread_mutex_init(_mutex, NULL);
      assert(status == 0, "pthread_mutex_init");
      #ifndef PRODUCT
      debug_only(_is_owned = false;)
      #endif
    }
    ~OSMutex() {
      int status = pthread_mutex_destroy(_mutex);
      assert(status == 0, "pthread_mutex_destroy");
    }
    // for use in critical sections:
    void lock() {
      verify();
      int status = pthread_mutex_lock(_mutex);
      assert(status == 0, "pthread_mutex_lock");
      #ifndef PRODUCT
      assert(_is_owned == false, "mutex_lock should not have had owner");
      debug_only(_owner = pthread_self();)
      debug_only(_is_owned = true;)
      #endif
    }
    bool trylock() {
      verify();
      int status = pthread_mutex_trylock(_mutex);
      if (status == EBUSY)
	return false;
      assert(status == 0, "pthread_mutex_trylock");
      #ifndef PRODUCT
      debug_only(_owner = pthread_self();)
      debug_only(_is_owned = true;)
      #endif
      return true;
    }
    void unlock() {
      verify();
      #ifndef PRODUCT
      debug_only(pthread_t my_id = pthread_self();)
      assert(pthread_equal(_owner, my_id), "mutex_unlock");
      debug_only(_is_owned = false;)
      #endif
      int status = pthread_mutex_unlock(_mutex);
      assert(status == 0, "pthread_mutex_unlock");
    }

    // hook to check for mutex corruption:
    void verify() PRODUCT_RETURN;
    void verify_locked() PRODUCT_RETURN;
  };
};

// Reconciliation History
// os_solaris.hpp	1.42 99/08/27 14:13:38
// End
