/*
 * @(#)osThread_solaris.cpp	1.28 01/03/05 12:01:25
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// do not include  precompiled  header file
# include "incls/_osThread_solaris.cpp.incl"


OSThread::InterruptCallback *OSThread::_vm_interrupt_callback   = NULL;


 // ***************************************************************
 // Platform dependent initialization and cleanup
 // ***************************************************************

void OSThread::pd_initialize() {
  _thread_id                         = 0;

  _interrupt_in_progress             = false;
  _interrupt_callback_cancelled_list = NULL;
  _interrupt_callback_list           = NULL;

  _interrupted                       = false;
  _interrupt_event  = new os::Solaris::Event();
  assert(_interrupt_event != NULL, "check");
  _saved_interrupt_thread_state      = _thread_new;
  _saved_interrupt_state             = NULL;
  _jump_buffer                       = NULL;
  
}

void OSThread::pd_destroy() {
  delete_cancelled_interrupt_callbacks(true);
  delete _interrupt_event;
}

void OSThread::InterruptCallback::move_to_cancelled_list() {
  _next = _osthread->interrupt_callback_cancelled_list();
  _osthread->set_interrupt_callback_cancelled_list(this);
}


// Copies the callback from the active list to the cancelled list.
// Asynchronous callbacks are cancelled by the interrupted threads
// after executing the callback.
// For synchronous ones, the signal sender thread is waiting for info 
// returned in the callback and cancels it after it is processed.
void OSThread::InterruptCallback::cancel() {
  // make sure that this callback is cancelled by the right thread
  assert((is_synchronous() && (_osthread != Thread::current()->osthread()) || 
	  (!is_synchronous() && (_osthread == Thread::current()->osthread()))), 
	 "Interrupt callback cancelled by wrong thread");
  assert((Interrupt_lock == NULL ) || Interrupt_lock->owned_by_self(), "Must lock Interrupt_lock first");

  // Ignore if already cancelled
  if (is_cancelled()) {
    return;
  }

  // remove from the active list
  if (this == _osthread->interrupt_callback_list()) {
    if(_next){
      _next->_prev = (OSThread::InterruptCallback *)NULL;
    }
    _osthread->set_interrupt_callback_list(_next);

  } else {
    assert(_prev != NULL, "just checking");
    _prev->_next = _next;
    if (_next != NULL) _next->_prev = _prev;
  }

  move_to_cancelled_list();
  set_cancelled();
}


// the signal handler should call this before making a non-local jump
// should be called from the signal handler ("at_interrupt")
void OSThread::InterruptCallback::cancel_at_interrupt() {
  // lock without safepoint check because we may be calling
  // from the signal handler.
  MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
  // If callback is asynchronous, cancel it.
  // Othewise, it is disposed of by the signal sender thread.
  if( !is_synchronous()){
    cancel();
  }
  _osthread->set_interrupt_in_progress(false);
}


// executes the interrupt callback
void OSThread::InterruptCallback::doit_at_interrupt(InterruptArguments *args) {
  execute(args);
  
  // If callback is asynchronous, cancel it. Othewise, it is disposed of by the sender thread 
  // waiting for info returned in the callback.
  if( !is_synchronous()){
    // lock without safepoint check because we may be calling
    // from the signal handler.
    MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
    cancel();
  }
}


// return true if a signal needs to be sent 
// (if there are callbacks on the active queue and an interrupt is not is progress)
// side effect -- sets _interrupt_in_progress to true if a signal is about to be sent
bool OSThread::send_interrupt() {
  // this seems like a good place to do clean up the list
  delete_cancelled_interrupt_callbacks();

  // don't use safepoint check because we might be the watcher thread.
  assert(Threads_lock->owned_by_self() ||
	 ((this == VMThread::vm_thread()->osthread()) && ProfileVM), 
	 "must have threads lock to call this");
  MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);

  int need_to_send_interrupt = false;

  if (!_interrupt_in_progress && _interrupt_callback_list != NULL) {
    _interrupt_in_progress = true;
    need_to_send_interrupt = true;
  }
  
  return need_to_send_interrupt;
}


// iterate through the cancelled interrupt list and deletes them
// should not be called from interrupt
void OSThread::delete_cancelled_interrupt_callbacks(bool called_from_destructor) {
  // don't use safepoint check because we might be the watcher thread.
  assert(called_from_destructor ||
         Threads_lock->owned_by_self() ||
	 (this == VMThread::vm_thread()->osthread()), "must have threads lock to call this");
  MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);

  InterruptCallback *cur  = _interrupt_callback_cancelled_list;
  InterruptCallback *next;

  while(cur != NULL) {
    next = cur->next();
    delete cur;
    cur = next;
  }

  _interrupt_callback_cancelled_list = NULL;
}


// add an interrupt callback to the active queue
void OSThread::add_interrupt_callback(InterruptCallback *callback) {
  // don't use safepoint check because we might be the watcher thread.
  // Let signals to the vm_thread go even if the Threads_lock is not aquired
  assert(Threads_lock->owned_by_self() || (this == VMThread::vm_thread()->osthread()), "must have threads lock to call this");
  MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);

  InterruptCallback *cur  = _interrupt_callback_list;
  InterruptCallback *prev = NULL;

  // InterruptCallback must not think it has been cancelled once added to list
  callback->clear_cancelled();

  while(cur != NULL && cur->priority() >= callback->priority()) {
    prev = cur;
    cur = cur->next();
  }

  callback->set_next(cur);
  callback->set_prev(prev);
  if (prev != NULL) 
    prev->set_next(callback);
  if (cur != NULL)
    cur->set_prev(callback);

  if (cur == _interrupt_callback_list)
    _interrupt_callback_list = callback;

  callback->set_osthread(this);
}


// used to determine the next callback to be executed
// side effect -- sets the interrupt in progress flag to false if there are no more callbacks
// should be called from the signal handler ("at_interrupt")
OSThread::InterruptCallback* OSThread::interrupt_callback_at_interrupt() {
  InterruptCallback* callback;

  // lock without safepoint check because we are calling
  // from the signal handler.
  MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);

  if ((callback = _interrupt_callback_list) == NULL)
    _interrupt_in_progress = false;

  return callback;
}


// iterate through the interrupt callbacks and execute them
// should be called from the signal handler ("at_interrupt")
void OSThread::do_interrupt_callbacks_at_interrupt(InterruptArguments *args) {
  InterruptCallback *cur;

  while ((cur = interrupt_callback_at_interrupt()) != NULL)
    cur->doit_at_interrupt(args);
}

// Execute the VM interrupt callback (GetThreadPC_Callback)
// should be called from the signal handler ("at_interrupt")
void OSThread::do_vm_interrupt_callback_at_interrupt(InterruptArguments *args) {
  // We execute a single vm interrupt callback (they are serialized and sent only
  // from the watcher thread during profiling)
  assert(args->thread()->is_VM_thread(), "Must be VM Thread");
  OSThread::_vm_interrupt_callback->execute(args);
}

// copied from synchronizer.cpp

void OSThread::handle_spinlock_contention(int tries) {
  if (NoYieldsInMicrolock) return;

  if (tries > 10) {
    os::yield_all(tries); // Yield to threads of any priority
  } else if (tries > 5) {  
    os::yield();          // Yield to threads of same or higher priority      
  }
}
