/*
 * @(#)osThread_solaris.hpp	1.36 01/02/14 09:00:10
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

 private:

  thread_t _thread_id; // Solaris thread id

 public:

  thread_t thread_id() const      { return _thread_id; }
#ifndef PRODUCT
  // Used for debugging, return a unique integer for each thread.
  int thread_identifier() const   { return _thread_id; }
#endif
  void set_thread_id(thread_t id) { _thread_id = id;   }

 // ***************************************************************
 // interrupt support.  interrupts (using signals) are used to get
 // the thread context (fetch_top_frame), to set the thread context
 // (set_thread_pc), and to implement java.lang.Thread.interrupt.
 // ***************************************************************

 public:

  class InterruptArguments : StackObj {
   private:
    Thread*     _thread;   // the thread to signal was dispatched to
    ucontext_t* _ucontext; // the machine context at the time of the signal

   public:
    InterruptArguments(Thread* thread, ucontext_t* ucontext) {
      _thread   = thread;
      _ucontext = ucontext;
    }

    Thread*     thread()   const { return _thread;   }
    ucontext_t* ucontext() const { return _ucontext; }
  };

  class InterruptCallback : public CHeapObj {
   public:

    // levels of priority for the callback
    // determines order in the os thread priority queue of active callbacks
    enum {
      low    = 0,
      normal = 5,
      high   = 9
    };

   protected:

    int                _priority; // the priority of the callback in the active callback priority queue of the os thread
    InterruptCallback* _next;     // the next InterruptCallback in the doubly-linked list
    InterruptCallback* _prev;     // the previous InterruptCallback in the doubly-linked list
    OSThread*          _osthread; // the owner of the callback
    bool               _is_synchronous;   // is true, the signal sender is responsible for cancelling 
                                          // the callback; otherwise the cancelled by the target thread
    bool               _is_cancelled;     // Interruptable I/O implementation requires
                                          // guarding against cancelling twice

    friend class OSThread;

    InterruptCallback* next() const              { return _next; }
    void set_next(InterruptCallback* next)       { _next = next; }

    InterruptCallback* prev() const              { return _prev; }
    void set_prev(InterruptCallback* prev)       { _prev = prev; }

    OSThread* osthread() const                   { return _osthread; }
    void set_osthread(OSThread* osthread)        { _osthread = osthread; }

    virtual void move_to_cancelled_list();
    void doit_at_interrupt(InterruptArguments *args);

    void set_cancelled()                         { _is_cancelled = true; }
    void clear_cancelled()                       { _is_cancelled = false; }

   public:
    InterruptCallback(int priority, bool synch) {
      _priority  = priority;
      _next      = NULL;
      _prev      = NULL;
      _osthread  = NULL;
      _is_synchronous = synch;
    }

    int priority() const                         { return _priority; }
    bool is_synchronous() const                  { return _is_synchronous; }

    void cancel_at_interrupt();
    virtual void execute(InterruptArguments *args) = 0;  // override to implement the callback
    // Subclass should call superclass implementation (or at least set_cancelled())
    virtual void cancel();
    
    bool is_cancelled() const                    { return _is_cancelled; }
  };

 private:

  InterruptCallback* _interrupt_callback_list;           // the priority queue of active callbacks
  InterruptCallback* _interrupt_callback_cancelled_list; // the list of cancelled (used) callbacks
  static InterruptCallback* _vm_interrupt_callback;      // a single instance of VMThread callback
  bool               _interrupt_in_progress;             // true if there is a signal pending on the osthread or interrupt processing is in progress

  friend class OSThread::InterruptCallback;

  InterruptCallback* interrupt_callback_list() const                                 { return _interrupt_callback_list;           }
  void               set_interrupt_callback_list(InterruptCallback* list)            { _interrupt_callback_list = list;           }

  InterruptCallback* interrupt_callback_cancelled_list() const                       { return _interrupt_callback_cancelled_list; }
  void               set_interrupt_callback_cancelled_list(InterruptCallback* list)  { _interrupt_callback_cancelled_list = list; }

  void delete_cancelled_interrupt_callbacks(bool called_from_destructor = false);
  InterruptCallback* interrupt_callback_at_interrupt();

 public:
  bool               interrupt_in_progress() const                                   { return _interrupt_in_progress;             }
  void               set_interrupt_in_progress(bool in_progress)                     { _interrupt_in_progress = in_progress;      }

  void add_interrupt_callback(InterruptCallback* callback);
  bool send_interrupt();
  void do_interrupt_callbacks_at_interrupt(InterruptArguments *args);
  void do_vm_interrupt_callback_at_interrupt(InterruptArguments *args);
  static InterruptCallback* vm_interrupt_callback()                                  { return _vm_interrupt_callback;             }
  static void set_vm_interrupt_callback(InterruptCallback* cb) { OSThread::_vm_interrupt_callback = cb; }

 // ***************************************************************
 // java.lang.Thread.interrupt state.
 // ***************************************************************

 private:

  bool               _interrupted;
  os::Solaris::Event* _interrupt_event;
  JavaThreadState    _saved_interrupt_thread_state;       // the thread state before a system call -- restored afterward
  jint*              _saved_interrupt_state;              // the post java state before a system call -- restored afterward
  void*              _jump_buffer;                        // a pointer to the jump buffer to be passed to longjmp in the event of an interrupt

 public:

  bool              interrupted() const                                         { return _interrupted;                                              }
  void              set_interrupted(bool interrupted)                           { _interrupted = interrupted;                                       }

  os::Solaris::Event* interrupt_event() const                                   { return _interrupt_event;                                          }
  void set_interrupt_event(os::Solaris::Event* ptr)                             { _interrupt_event = ptr;                                           }

  JavaThreadState   saved_interrupt_thread_state()                              { return _saved_interrupt_thread_state;                             }
  void              set_saved_interrupt_thread_state(JavaThreadState state)     { _saved_interrupt_thread_state = state;                            }

  jint*             saved_interrupt_state()                                     { return _saved_interrupt_state;                                    }
  void              set_saved_interrupt_state(jint* state)                      { _saved_interrupt_state = state;                                   }

  void*             jump_buffer()                                               { return _jump_buffer;                                              }
  void              set_jump_buffer(void* jump_buffer)                          { _jump_buffer = jump_buffer;                                       }

  static void       handle_spinlock_contention(int tries);                      // Used for thread local eden locking

  // ***************************************************************
  // Platform dependent initialization and cleanup
  // ***************************************************************

private:

  void pd_initialize();
  void pd_destroy();
