/*
 * @(#)mutex_win32.cpp	1.32 01/03/05 12:02:04
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_mutex_win32.cpp.incl"

// put OS-includes here
# include <windows.h>

// Implementation of Mutex

// A simple Mutex for VM locking: it is not guaranteed to interoperate with
// the fast object locking, so exclusively use Mutex locking or exclusively
// use fast object locking.

void Mutex::wait_for_lock_implementation() {
  DWORD dwRet = WaitForSingleObject((HANDLE)_lock_event,  INFINITE);
  assert(dwRet == WAIT_OBJECT_0, "unexpected return value from WaitForSingleObject");
}

void Mutex::wait_for_lock_blocking_implementation(JavaThread *thread) {
  ThreadBlockInVM tib(thread);
  wait_for_lock_implementation();
}

Mutex::Mutex(int prio, const char *name, bool allow_vm_block) debug_only( : _prio(prio)) {
  _lock_count = -1;       // No threads has entered the critical section
  _lock_event = CreateEvent(NULL, false, false, NULL);
  _supress_signal = false;
  _owner          = INVALID_THREAD;
  _name           = name;

#ifndef PRODUCT
  _allow_vm_block = allow_vm_block;
  debug_only(_next= NULL;)
#endif
}


Mutex::~Mutex() {
  assert(_owner == INVALID_THREAD, "Owned Mutex being deleted");
  assert(_lock_count == -1, "Mutex being deleted with non -1 lock count");
  CloseHandle((HANDLE)_lock_event);
}


void Mutex::unlock() {  
  assert(_owner == Thread::current(), "Mutex not being unlocked by owner");
  assert(_lock_count >= 0, "Mutex being unlocked without positive lock count");
  trace("unlocks");
  set_owner(INVALID_THREAD);  
  if (InterlockedDecrement(&_lock_count) >= 0) {
    // Wake a waiting thread up
    if (!_supress_signal) {
      DWORD dwRet = SetEvent((HANDLE)_lock_event);
      assert(dwRet != 0, "unexpected return value from SetEvent");
    }
    _supress_signal = false;
  }    
}

// Can be called by non-Java threads (JVM_RawMonitorExit)
void Mutex::jvm_raw_unlock() {
  assert(priority() == native, "must be called by non-VM locks");
  // Do not call set_owner, as this would break.
  _owner = INVALID_THREAD;  
  if (InterlockedDecrement(&_lock_count) >= 0) {
    // Wake a waiting thread up
    if (!_supress_signal) {
      DWORD dwRet = SetEvent((HANDLE)_lock_event);
      assert(dwRet != 0, "unexpected return value from SetEvent");
    }
    _supress_signal = false;
  }    
}


#ifndef PRODUCT
void Mutex::print() const {
  tty->print_cr("Mutex: [0x%x] %s - lock_count: %d", this, _name, _lock_count);
}
#endif


//
// Monitor
//

Monitor::Monitor(int prio, const char *name, bool allow_vm_block) : Mutex(prio, name, allow_vm_block) {
  HANDLE e = CreateEvent(NULL, true, false, NULL);
  if (e == 0) {
    fatal("Could not initialize condition variable");
  }
  _event = e;
  _counter = 0;
  _tickets = 0;
  _waiters = 0;
}


Monitor::~Monitor() {
  CloseHandle((HANDLE)_event);
}  


bool Monitor::wait(bool no_safepoint_check, long timeout) {
  assert(_owner != INVALID_THREAD, "Wait on unknown thread");
  assert(owned_by_self(), "Wait on Monitor not by owner");

  // suspension waits on Interrupt_Lock and SuspendChecker_lock
  assert(priority() > leaf ||
    ((this == Interrupt_lock || this == SuspendChecker_lock)
      && no_safepoint_check),
    "priority too low to do a wait");

  // 0 means forever. Convert to Windows specific code.
  DWORD timeout_value = (timeout == 0) ? INFINITE : timeout;
  DWORD which;
  
  long c = _counter;
  bool retry = false;  
      
  _waiters++;
  // Loop until condition variable is signaled.  The event object is
  // set whenever the condition variable is signaled, and tickets will
  // reflect the number of threads which have been notified. The counter
  // field is used to make sure we don't respond to notifications that
  // have occurred *before* we started waiting, and is incremented each
  // time the condition variable is signaled.  

  while (true) {    

    // Leave critical region
    unlock();

    // If this is a retry, let other low-priority threads have a chance
    // to run.  Make sure that we sleep outside of the critical section.
    if (retry) {
      os::yield_all();
    } else {
      retry = true;
    }
         
    Thread* thread = Thread::current();
    if (no_safepoint_check) {
      // Need to leave state as SUSPENDED for wait on Interrupt_Lock
      // OSThreadWaitState osts(thread->osthread());
      which = WaitForSingleObject((HANDLE)_event,  timeout_value);
      // Enter critical section (this will also put us in an unblocked state)    
      lock_without_safepoint_check();
    } else { 
      { 
        assert(thread->is_Java_thread(), "Must be JavaThread");
        ThreadBlockInVM tbiv((JavaThread *)thread);
        OSThreadWaitState osts(thread->osthread());
        which = WaitForSingleObject((HANDLE)_event,  timeout_value);
      }
      // Enter critical section (this will also put us in an unblocked state)    
      lock();
    }        

    if (_tickets != 0 && _counter != c) break;    
    
    if (which == WAIT_TIMEOUT) {      
      --_waiters;          
      return true;
    }    
  }
  _waiters--;

  // If this was the last thread to be notified, then we need to reset
  // the event object.
  if (--_tickets == 0) {
    ResetEvent((HANDLE)_event);
  }  

  return false;
}


// Notify a single thread waiting on this ConditionVariable
bool Monitor::notify() {
  assert(owned_by_self(), "notify on unknown thread");
  
  if (_waiters > _tickets) {
    if (!SetEvent((HANDLE)_event)) {
      return false;
    }
    _tickets++;
    _counter++;
  }

  return true;

}


// Notify all threads waiting on this ConditionVariable
bool Monitor::notify_all() {
  assert(owned_by_self(), "notify_all on unknown thread");
  
  if (_waiters > 0) {
    if (!SetEvent((HANDLE)_event)) {
      return false;
    }
    _tickets = _waiters;
    _counter++;
  }

  return true;
}



