/*
 * @(#)objectMonitor_win32.cpp	1.29 01/03/05 12:02:23
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_objectMonitor_win32.cpp.incl"

class ObjectWaiter : public StackObj {
 public:
  ObjectWaiter* _next;
  Thread*       _thread;
  HANDLE        _event;
  intptr_t      _notified;
 public:
  ObjectWaiter(Thread* thread, HANDLE event) {
    _next     = NULL;
    _thread   = thread;
    _notified = 0;
    _event    = event;
  }
};


ObjectWaiter* ObjectMonitor::first_waiter() {
  return (ObjectWaiter*)_queue;
}

ObjectWaiter* ObjectMonitor::next_waiter(ObjectWaiter* o) {
  return o->_next;
}

Thread* ObjectMonitor::thread_of_waiter(ObjectWaiter* o) {
  return o->_thread;
}

// initialize the monitor, exception the semaphore, all other fields
// are simple integers or pointers
ObjectMonitor::ObjectMonitor() {
  _header     = NULL;
  _count      = 0;
  _waiters    = 0,
  _recursions = 0;
  _object     = NULL;
  _owner      = NULL;
  _queue      = NULL;
  _semaphore  = ::CreateSemaphore(0, 0, 1, 0);
  assert(_semaphore != NULL, "CreateSemaphore() failed!");
}

ObjectMonitor::~ObjectMonitor() {
  ::CloseHandle(_semaphore);
}

inline void ObjectMonitor::enqueue(void* argwaiter) {
  ObjectWaiter* waiter = (ObjectWaiter*)argwaiter;
  waiter->_next = NULL;
  // put waiter to the end of queue (single linked list)
  if (_queue == NULL) {
    _queue = waiter;
  } else {
    ObjectWaiter* iterator = (ObjectWaiter*)_queue;
    while (iterator->_next) {
      iterator = iterator->_next;
    }
    iterator->_next = waiter;
  }
}

inline ObjectWaiter* ObjectMonitor::dequeue() {
  // dequeue the very first waiter
  ObjectWaiter* waiter = (ObjectWaiter*)_queue;
  if (waiter) {
    _queue = waiter->_next;
    waiter->_next = NULL;
  }
  return waiter;
}

inline void ObjectMonitor::dequeue2(void* argwaiter) {
  ObjectWaiter* waiter = (ObjectWaiter*) argwaiter;
  // when the waiter has waked up because of interrupt,
  // timeout or other suprious wake-up, dequeue the 
  // waiter from waiting list (use linear search)
  if (_queue == waiter) {
    _queue = waiter->_next;
    waiter->_next = NULL;
  } else {
    ObjectWaiter* iterator = (ObjectWaiter*) _queue;
    while (iterator->_next != waiter) {
      // the waiter must be in the queue, iterator could not be null
      assert(iterator->_next != NULL, "ObjectMonitor::dequeue2");
      iterator = iterator->_next;
    }
    iterator->_next = waiter->_next;
    waiter->_next = NULL;
  }
}

void ObjectMonitor::enter(TRAPS) {
  if (atomic::compare_and_exchange_ptr(&_count, 1, 0) == 0) {
    // fast path, the monitor is not owned yet
    atomic::membar();
    _owner = THREAD;
  } else if (_owner == THREAD) {
    // normal recursion case
    _recursions++;
  } else if (THREAD->is_lock_owned((address) _owner)) {
    assert(_recursions == 0, "check");
    _owner = THREAD;
    _recursions = 1;
  } else {
    if (UseSpinning) {
      for(int j = 0; j < PreBlockSpin; j++) {
        if (PreSpinYield) {
          os::yield();
        }
        // Spinning: retry the fast path again
        for (int i = 0; i < ReadSpinIterations; i++) {
          if (_count == 0) {
            if (atomic::compare_and_exchange_ptr(&_count, 1, 0) == 0) {
              // fast path, the monitor is not owned yet
              atomic::membar();
              _owner = THREAD;
              return;
            }
          }
        }
        if (PostSpinYield) {
          os::yield();  
        }
      }
    }

    // increment the lock count
    if (atomic::add_ptr(&_count, 1) > 1) {
      // the monitor is locked by some other thread
      assert(THREAD->is_Java_thread(), "Must be Java thread!");
      {
        if (jvmpi::is_event_enabled(JVMPI_EVENT_MONITOR_CONTENDED_ENTER)) {
          jvmpi::post_monitor_contended_enter_event(object());
        } 
        OSThreadContendState osts(THREAD->osthread());
        ThreadBlockInVM tbiv((JavaThread*)THREAD);

        bool threadIsSuspended;
        do {
          DWORD ret = ::WaitForSingleObject(_semaphore, INFINITE);
          assert(ret == WAIT_OBJECT_0, "WaitForSingleObject() failed!");

          //
          // We have been granted the contended monitor, but while we were
          // waiting another thread suspended us. We don't want to enter
          // the monitor while suspended because that would surprise the
          // thread that suspended us.
          //
          threadIsSuspended = ((JavaThread *)THREAD)->is_being_ext_suspended();
          if (threadIsSuspended) {
            BOOL ret = ::ReleaseSemaphore(_semaphore, 1, NULL);
            assert(ret, "ReleaseSemaphore() failed!");
            os::yield();     // give another thread a chance at the monitor
          }
        } while (threadIsSuspended);
      }

      if (jvmpi::is_event_enabled(JVMPI_EVENT_MONITOR_CONTENDED_ENTERED)) {
        jvmpi::post_monitor_contended_entered_event(object());
      }
    }
    // there should be no ownership
    assert(_owner == NULL, "Logic error with monitor owner!");
    assert(_recursions == 0, "Logic error with monitor recursions!");
    // use membar to enforce write to write order
    atomic::membar();
    _owner = THREAD;
  }
}

void ObjectMonitor::exit(TRAPS) {
  if (THREAD != _owner) {
    if (THREAD->is_lock_owned((address) _owner)) {
      _owner = THREAD;  // regain ownership of inflated monitor
    } else {
      // NOTE: we need to handle unbalanced monitor enter/exit
      // in native code by throwing an exception.
      assert(false, "Non-balanced monitor enter/exit!");
      return;
    }
  }
  if (_recursions == 0) {
    _owner = NULL;        // clear the owner ship
    atomic::membar();     // memory barrier to prevent out-of-order write
    // atomic decrement count, if it still more than zero
    // there must be someone else blocking on monitor enter
    if (atomic::add_ptr(&_count, -1) > 0) {
      BOOL ret = ::ReleaseSemaphore(_semaphore, 1, NULL);
      assert(ret, "ReleaseSemaphore() failed!");
      if (jvmpi::is_event_enabled(JVMPI_EVENT_MONITOR_CONTENDED_EXIT)) {
        jvmpi::post_monitor_contended_exit_event(object());
      }
    }
  } else {
    _recursions--;        // this is simple recursive enter
  }
}

void ObjectMonitor::wait(jlong millis, bool interruptable, TRAPS) {
  check(CHECK);
  // check interrupt event
  if (interruptable && Thread::is_interrupted(THREAD, true) && !HAS_PENDING_EXCEPTION) {
    THROW(vmSymbols::java_lang_InterruptedException());
  }
  // create a node to be put into the queue
  ObjectWaiter node(THREAD, THREAD->osthread()->interrupt_event());
  // enter the waiting queue, which is single linked list in this case
  // how it could be a priority queue, or any data structure
  enqueue(&node);

  int save = _recursions; // record the old recursion count
  _waiters++;             // increment the number of waiters
  _recursions = 0;        // set the recursion level to be 1
  exit(THREAD);           // exit the monitor
  if (jvmpi::is_event_enabled(JVMPI_EVENT_MONITOR_WAIT)) {
    jvmpi::post_monitor_wait_event((oop)object(), millis);
  }
  {
    assert(THREAD->is_Java_thread(), "Must be Java thread!");
    ThreadBlockInVM tbiv((JavaThread*) THREAD);
    OSThreadWaitState osts(THREAD->osthread());
    EventWait(node._event, millis);
    // retain CONDVAR_WAIT state until re-enter successfully completes
    raw_enter(THREAD, false);          // re-enter the monitor
  }
 
  if (jvmpi::is_event_enabled(JVMPI_EVENT_MONITOR_WAITED)) {
    jvmpi::post_monitor_waited_event((oop)object(), millis);
  }
  _recursions = save;     // restore the old recursion count
  _waiters--;             // decrement the number of waiters

  // NOTE: this line must be here, because a notification
  // may be right after the timeout, but before re-enter
  // so the event may still be in signed state
  BOOL ret = ::ResetEvent(node._event);
  assert(ret, "ResetEvent() failed!");
  // check if the notification happened
  if (node._notified == 0) {
    // no, it could be timeout or Thread.interrupt() or both
    // dequeue the the node from the single linked list
    dequeue2(&node);
    // check for interrupt event, otherwise it is timeout
    if (interruptable && Thread::is_interrupted(THREAD, true) && !HAS_PENDING_EXCEPTION) {
      THROW(vmSymbols::java_lang_InterruptedException());
    }
  }
  // NOTES: Suprious wake up will be consider as timeout.
  // Monitor notify has precedence over thread interrupt.
}

void ObjectMonitor::notify(TRAPS) {
  check(CHECK);
  ObjectWaiter* iterator;
  if ((iterator = dequeue()) != 0) {
    iterator->_notified = 1;
    BOOL ret = ::SetEvent(iterator->_event);
    assert(ret, "SetEvent() failed!");
  }
}

void ObjectMonitor::notifyAll(TRAPS) {
  check(CHECK);
  ObjectWaiter* iterator;
  while ((iterator = dequeue()) != 0) {
    iterator->_notified = 1;
    BOOL ret = ::SetEvent(iterator->_event);
    assert(ret, "SetEvent() failed!");
  }
}

// used for JVMDI/JVMPI raw monitor implementation
// Also used by wait above for re-entry
// assumes that the thread is in native state
int ObjectMonitor::raw_enter(TRAPS, bool isRawMonitor) {
  if (atomic::compare_and_exchange_ptr(&_count, 1, 0) == 0) {
    atomic::membar();
    _owner = THREAD;
  } else if (_owner == THREAD) {
    _recursions++;
  } else {
    if (isRawMonitor) {
      if (jvmpi::is_event_enabled(JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTER)) {
        jvmpi::post_raw_monitor_contended_enter_event((RawMonitor *)this);
      }
    }
    if (atomic::add_ptr(&_count, 1) > 1) {
      assert(THREAD->is_Java_thread(), "Must be Java thread!");

      bool threadIsSuspended;
      do {
        DWORD ret = ::WaitForSingleObject(_semaphore, INFINITE);
        assert(ret == WAIT_OBJECT_0, "WaitForSingleObject() failed!");

        // were we suspended while we were waiting?
        threadIsSuspended = ((JavaThread *)THREAD)->is_being_ext_suspended();
        if (threadIsSuspended) {
          //
          // This logic isn't needed for JVM/DI or JVM/PI raw monitors,
          // but doesn't hurt just in case the suspend rules change. This
          // logic is needed for the ObjectMonitor.wait() reentry phase.
          // We have reentered the contended monitor, but while we were
          // waiting another thread suspended us. We don't want to reenter
          // the monitor while suspended because that would surprise the
          // thread that suspended us.
          //
          BOOL ret = ::ReleaseSemaphore(_semaphore, 1, NULL);
          assert(ret, "ReleaseSemaphore() failed!");
          os::yield();     // give another thread a chance at the monitor
        }
      } while (threadIsSuspended);
    }
    // there should be no ownership
    assert(_owner == NULL, "Fatal error with monitor owner!");
    assert(_recursions == 0, "Fatal error with monitor recursions!");
    // prevent out of order write
    atomic::membar();
    _owner = THREAD;
    if (isRawMonitor) {
      if (jvmpi::is_event_enabled(JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTERED)) {
        jvmpi::post_raw_monitor_contended_entered_event((RawMonitor *)this);
      }
    }
  }
  return OM_OK;
}

int ObjectMonitor::raw_exit(TRAPS, bool isRawMonitor) {
  if (THREAD != _owner) {
    return OM_ILLEGAL_MONITOR_STATE;
  }
  if (_recursions == 0) {
    if (isRawMonitor) {
      ((RawMonitor*)this)->remove_from_locked_list();
    }
    _owner = NULL;      // clear the owner ship
    atomic::membar();   // memory barrier to prevent out-of-order write
    // atomic decrement count, if it still more than zero
    // there must be someone else blocking on monitor enter
    if (atomic::add_ptr(&_count, -1) > 0) {
      BOOL ret = ::ReleaseSemaphore(_semaphore, 1, NULL);
      assert(ret, "ReleaseSemaphore() failed!");
      if (isRawMonitor) {
        if (jvmpi::is_event_enabled(JVMPI_EVENT_RAW_MONITOR_CONTENDED_EXIT)) {
          jvmpi::post_raw_monitor_contended_exit_event((RawMonitor *)this);
        }
      }
    }
  } else {
    _recursions--; // this is simple recursive enter
  }
  return OM_OK;
}

// assumes that the current thread is in native state
int ObjectMonitor::raw_wait(jlong millis, bool interruptable, TRAPS) {
  if (THREAD != _owner) {
    return OM_ILLEGAL_MONITOR_STATE;
  }
  // check interrupt event
  if (interruptable && Thread::is_interrupted(THREAD, true)) {
    return OM_INTERRUPTED;
  }
  // create a node to be put into the queue
  ObjectWaiter node(THREAD, THREAD->osthread()->interrupt_event());
  // enter the waiting queue, which is single linked list in this case
  // how it could be a priority queue, or any data structure
  enqueue(&node);

  int save = _recursions; // record the old recursion count
  _waiters++;             // increment the number of waiters
  _recursions = 0;        // set the recursion level to be 1
  raw_exit(THREAD, false);// exit the monitor
  {
    EventWait(node._event, millis);
  }
  raw_enter(THREAD, false);      // re-enter the monitor
  _recursions = save;     // restore the old recursion count
  _waiters--;             // decrement the number of waiters

  // NOTE: this line must be here, because a notification
  // may be right after the timeout, but before re-enter
  // so the event may still be in signed state
  BOOL ret = ::ResetEvent(node._event);
  assert(ret, "ResetEvent() failed!");
  // check if the notification happened
  if (node._notified == 0) {
    // no, it could be timeout or Thread.interrupt() or both
    // dequeue the the node from the single linked list
    dequeue2(&node);
    // check for interrupt event, otherwise it is timeout
    if (interruptable && Thread::is_interrupted(THREAD, true)) {
      return OM_INTERRUPTED;
    }
  }
  // NOTES: Suprious wake up will be consider as timeout.
  // Monitor notify has precedence over thread interrupt.
  return OM_OK;
}

int ObjectMonitor::raw_notify(TRAPS) {
  if (THREAD != _owner) {
    return OM_ILLEGAL_MONITOR_STATE;
  }
  ObjectWaiter* iterator;
  if ((iterator = dequeue()) != 0) {
    iterator->_notified = 1;
    BOOL ret = ::SetEvent(iterator->_event);
    assert(ret, "SetEvent() failed!");
  }
  return OM_OK;
}

int ObjectMonitor::raw_notifyAll(TRAPS) {
  if (THREAD != _owner) {
    return OM_ILLEGAL_MONITOR_STATE;
  }
  ObjectWaiter* iterator;
  while ((iterator = dequeue()) != 0) {
    iterator->_notified = 1;
    BOOL ret = ::SetEvent(iterator->_event);
    assert(ret, "SetEvent() failed!");
  }
  return OM_OK;
}

// %%% should move this to objectMonitor.cpp
void ObjectMonitor::check_slow(TRAPS) {
  // called only from check()
  assert(THREAD != _owner && !THREAD->is_lock_owned((address) _owner), "must not be owner");
  THROW_MSG(vmSymbols::java_lang_IllegalMonitorStateException(), "current thread not owner");
}

#ifndef PRODUCT
void ObjectMonitor::verify() {
}

void ObjectMonitor::print() {
}
#endif
