/*
 * @(#)assembler_abs.hpp	1.21 99/07/22 15:49:07
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// This file contains platform-independant assembler declarations.


// Register is an abstraction over machine registers.

class Register; // the platform-specific one

class AbstractRegister VALUE_OBJ_CLASS_SPEC {
 protected:
  int _number;                    // corresponds to the machine register code
                                  // -1 stands for invalid register (initial value).
  
 public:
  // creation
  AbstractRegister(void): _number(-1)			{}
  AbstractRegister(int number, char f): _number(number)	{}	// f is only to make sure that
  							        // an int is not accidentially
							        // converted into a Register...

  // attributes
  // used by isValid, and also shadowed by a real const in the (sole) subclass
  static const int nofRegisters;
  // Must put this in assembler_<arch>.cpp:
  //   const int AbstractRegister::nofRegisters = Register::nofRegisters;
  
  bool isValid() const		  { return (0 <= _number) && (_number < nofRegisters); }
  int number() const		  { assert(isValid(), "not a register"); return _number; }
  
  // comparison
  friend bool operator == (AbstractRegister x, AbstractRegister y) { return x._number == y._number; }
  friend bool operator != (AbstractRegister x, AbstractRegister y) { return x._number != y._number; }
};


// Labels represent target destinations for jumps and calls.
//
// After declaration they can be freely used to denote known or (yet) unknown
// target destinations. Assembler::bind is used to bind a label to the current
// code position. A label can be bound only once.

class Label VALUE_OBJ_CLASS_SPEC {
 private:
  // _pos encodes both the binding state (via its sign)
  // and the binding position (via its value) of a label.
  //
  // _pos <  0	bound label, pos() returns the target (jump) position
  // _pos == 0	unused label
  // _pos >  0	unbound label, pos() returns the last displacement (see .cpp file) in the chain
  int _pos;

  int pos() const {
    if (_pos < 0) return -_pos - 1;
    if (_pos > 0) return  _pos - 1;
    ShouldNotReachHere();
    return 0;
  }

  void bind_to(int pos)		{ assert(pos >= 0, "illegal position"); _pos = -pos - 1; }
  void link_to(int pos)		{ assert(pos >= 0, "illegal position"); _pos =  pos + 1; }
  void unuse()			{ _pos = 0; }

 public:
  bool is_bound() const		{ return _pos <  0; }
  bool is_unbound() const	{ return _pos >  0; }
  bool is_unused() const	{ return _pos == 0; }

  Label() : _pos(0)		{}
  ~Label()			{ assert(!is_unbound(), "unbound label"); }

  friend class AbstractAssembler;
  friend class Assembler;
  friend class MacroAssembler;
  friend class Displacement;
};


// Clients of the assembler need to emit instructions (such as calls and branches)
// that refer to addresses in the same code stream BEFORE the addresses are known.
// When asked to emit such an instruction, the assembler emits a Displacement
// instead. The Displacement contains enough information to allow the assembler
// to backpatch, and replace the displacement with the right instruction afterwards.
//
// Here we define only an abstract, platform-independent framework.

class AbstractDisplacement : public ResourceObj {

  // The following must be implemented by the platform-spefific Displacement
  // class. Cannot use virtual foo() = 0; because I don't want virtuals,
  // for efficiency's sake. -- dmu 4/97

  // AbstractDisplacments are created in an implementation-specific
  // manner, when emitting an instruction with a forward reference in it.

  // Note, all members below are really abstract, just use inlining
  // to avoid virtuals.
   
  // bind is called repeatedly at each link in the chain.
  // It backpatches the instruction as appropriate, then calls
  // next to step the label to point to the next displacement in the chain.
  inline void bind(Label &L, int pos, AbstractAssembler* masm);
  
  // An iterator, next binds the label to the next displacement
  // or unused, if this is the end of the displacement chain.
  inline void next(Label& L) const;
  
  inline void print();
};



// The Abstract Assembler: Pure assembler doing NO optimizations on the instruction
// level; i.e., what you write
// is what you get. The Assembler is generating code into a CodeBuffer.

class AbstractAssembler : public ResourceObj  {
  friend class Displacement;

 protected:
  CodeBuffer*  _code;
  address      _code_begin;            // first byte of code buffer
  address      _code_limit;            // first byte after code buffer
  address      _code_pos;              // current code generation position
  address      _inst_mark;             // marked start of current instruction
  OopRecorder* _oop_recorder;          // support for relocInfo::oop_type

  // Code emission & accessing
  address addr_at(int pos) const       { return _code_begin + pos; }

  int  byte_at(int pos) const          { return *addr_at(pos); }
  void byte_at_put(int pos, int x)     { *addr_at(pos) = (unsigned char)x; }

  jint long_at(int pos)	const          { return *(jint*)addr_at(pos); }
  void long_at_put(int pos, jint x)    { *(jint*)addr_at(pos) = x; }

  bool is8bit(int x) const             { return -0x80 <= x && x < 0x80; }
  bool isByte(int x) const             { return 0 <= x && x < 0x100; }
  bool isShiftCount(int x) const       { return 0 <= x && x < 32; }

  inline void emit_byte(int x);
  inline void emit_long(jint x);

  address  inst_mark() const           { return _inst_mark; }
  void set_inst_mark(address x)        { _inst_mark = x;    }

  // Instruction boundaries (required when emitting relocatable values).
  class InstructionMark {
   private:
    AbstractAssembler* _assm;

   public:
    InstructionMark(AbstractAssembler* assm) : _assm(assm) {
      assert(assm->inst_mark() == NULL, "overlapping instructions");
      _assm->set_inst_mark(assm->pc());
    }
    ~InstructionMark() {
      _assm->set_inst_mark(NULL);
    }
  };
  friend class InstructionMark;
  friend class CodeEmitter;

  // Label functions

  void print  (Label& L);
  void bind_to(Label& L, int pos);

  // Displacement functions

  // get Displacement object in code stream at position in L
  inline Displacement disp_at(     Label& L) const;

  // setup reference from code stream at L's pos to disp
  inline void         disp_at_put( Label& L,  Displacement& disp);

 public:
  // Creation
  AbstractAssembler(CodeBuffer* code);

  // Accessors
  void		flush();               // make sure code buffer contains all code (call this before using/copying the code)
  CodeBuffer*	code() const           { return _code; }
  address       pc() const             { return _code_pos; }
  int		offset() const         { return _code_pos - _code_begin; }
  OopRecorder*  oop_recorder() const   { return _oop_recorder; }
  void      set_oop_recorder(OopRecorder* r) { _oop_recorder = r; }

  // Constants in code
  void a_byte(int x);
  void a_long(jint x);
  inline void relocate(RelocationHolder const& rspec, int format = 0);
  inline void relocate(   relocInfo::relocType rtype, int format = 0) {
    if (rtype != relocInfo::none)
      relocate(Relocation::spec_simple(rtype), format);
  }
  inline void relocate_raw(relocInfo::relocType rtype, const short* data = NULL, int datalen = 0, int format = 0);

  static int code_fill_byte();         // used to pad out odd-sized code buffers

  // Labels
  void bind(Label& L);                 // binds an unbound label L to the current code position
};



