/*
 * @(#)codeBuffer.cpp	1.48 00/06/19 13:25:15
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_codeBuffer.cpp.incl"

// The structure of the CodeBuffer:
//
//       _instsStart ->   +----------------+
//                        |                |
//                        |                |
//                        |     Code       |
//                        |                |
//                        |                |
//                        +----------------+
//    _instsOverflow ->   |                |
//    _stubsStart         |      Stubs     |
//                        +----------------+
//    _stubsOverflow ->   |                |
//    _constStart         |   FP constants |
//                        +----------------+
//    _constOverflow ->   |                |

MacroAssembler* theMacroAssm = NULL;

int cLocEntrySize = sizeof(RelocationHolder) + sizeof(address) + sizeof(int);


int CodeBuffer::insts_memory_size(int instsSize) {
  instsSize = round_to( instsSize, oopSize);
  return instsSize + wordSize*2;
}


int CodeBuffer::locs_memory_size(int locsSize) {
  locsSize = round_to( locsSize,  oopSize);
  return locsSize * (1 + cLocEntrySize );
}


// instsSize includes the stubsSize measurement
CodeBuffer::CodeBuffer(int instsSize, int locsSize, int stubsSize, int constSize, bool needs_oop_recorder,
                       address instrs_memory, address locs_memory) {
  // Keep original instSize since Stubs are not oopSize aligned.
  int instsOrig  = instsSize;
  // conservatively align allocated code size to oopSize, for later fill/copy
  instsSize      = round_to( instsSize, oopSize);
  // conservatively align relocation size to oopSize, for later fill/copy
  locsSize       = round_to( locsSize,  oopSize);
  constSize     = round_to( constSize, oopSize);

  int alignSize = max(sizeof(jdouble), CodeEntryAlignment);
  assert(is_power_of_2(alignSize), "");
  // +alignSize*2 so that we have enough space for double-word alignment of
  // insts and fp constants tables
  int instsSlop = alignSize*2;

  address insts = (instrs_memory == NULL) ?
    NEW_RESOURCE_ARRAY(u_char, instsSize + instsSlop):
    instrs_memory;

  _instsStart    = insts + alignSize;
  // align _instsStart to double word so _constStart remains aligned in codeBlob
  // Also align to CodeEntryAlignment so that assembler offsets look the same
  // (mod CodeEntryAlignment) before and after copying.
  _instsStart = (address)_instsStart - ((unsigned)_instsStart % alignSize);
  assert ((unsigned)_instsStart % alignSize == 0, "instruction start not double word aligned");
  _instsEnd      = _instsStart;
  _instsOverflow = insts + instsSize + instsSlop;
  assert (_instsOverflow - _instsStart >= instsSize, "");

  // mark stubs boundaries
  _instsEnd_before_stubs      = NULL;
  _instsOverflow_before_stubs = NULL;
  _stubsStart    = _stubsEnd = _instsStart + instsOrig - stubsSize - constSize;
  _stubsOverflow = _instsOverflow;
  _instsOverflow = _stubsStart;

  _constStart = _instsStart + instsOrig - constSize + wordSize*2;
  // align _constStart to double word (faster on INTEL)
  _constStart = (address)((unsigned)_constStart - ((unsigned)_constStart % (wordSize * 2)));
  assert ((unsigned)_constStart % (wordSize*2) == 0, "constant table not double word aligned");
  _constEnd = _constStart;
  _constOverflow = _stubsOverflow;
  _stubsOverflow = _constStart;

  _locsStart  =  (locs_memory == NULL) ?
    (relocInfo*)NEW_RESOURCE_ARRAY(u_char, locsSize * (1 + cLocEntrySize )):
    (relocInfo*)locs_memory;

  _locsEnd = _locsStart;
  _locsOverflow  = (relocInfo*) ((address) _locsStart + locsSize);
  _stubsReloc    = (char*)_locsStart + locsSize * (1 + cLocEntrySize);
  _stubs_reloc_count = 0;
  _last_reloc_offset = code_size();
  _decode_begin  = NULL;
  _oop_recorder  = needs_oop_recorder ? new OopRecorder() : NULL;
}


CodeBuffer::CodeBuffer(address code_start, int code_size) {
  _instsStart    = _instsEnd = code_start;
  _instsOverflow = _instsStart + code_size;
  _instsEnd_before_stubs      = NULL;
  _instsOverflow_before_stubs = NULL;
  _stubsStart    = _stubsEnd = NULL;
  _locsStart     = _locsEnd = NULL;
  _locsOverflow  = NULL;
  _stubsReloc    = NULL;
  _stubs_reloc_count = 0;
  _last_reloc_offset = CodeBuffer::code_size();
  _decode_begin  = NULL;
  _constStart = _constEnd = _constOverflow = NULL;
  _oop_recorder  = NULL;
}


address CodeBuffer::decode_begin() {
  return _decode_begin == NULL ? _instsStart : _decode_begin;
}


// private
// Store out-of-line stub relocation info until finished with code generation.
void CodeBuffer::add_stub_reloc(address at, RelocationHolder const &rspec, int format) {
  // Store the address and relocInfo at end of relocation information area.
  int old_stubsReloc = (int)_stubsReloc;
  _stubsReloc      -= sizeof(address);
  *((address*)_stubsReloc)           = at;
  _stubsReloc      -= sizeof(RelocationHolder);
  *((RelocationHolder*)_stubsReloc)  = rspec;
  _stubsReloc      -= sizeof(int);
  *((int*)_stubsReloc)               = format;
  ++_stubs_reloc_count;
  assert (old_stubsReloc - (int)_stubsReloc == cLocEntrySize, "");
}

// public
// Call CodeBuffer::relocate with the stub relocations,
// when finished emiting code.
void CodeBuffer::relocate_stubs() {
  assert( _stubsStart != NULL, "Do not have any stubs to relocate");
  // For each stored stub relocation
  char *stubsReloc = _stubsReloc 
    + _stubs_reloc_count * (cLocEntrySize);

  address temp = _instsEnd;
  _instsEnd  = _stubsEnd;
  
  for( int count = _stubs_reloc_count; count != 0; --count ) {
    // Get stored reloc values and add to relocation info.
    int old_stubsReloc = (int)stubsReloc;
    stubsReloc -= sizeof(address);
    address           at      = *(address*)(stubsReloc);
    stubsReloc -= sizeof(RelocationHolder);
    RelocationHolder rspec    = *(RelocationHolder*)(stubsReloc);
    stubsReloc -= sizeof(int);
    int               format  = *(int*)(stubsReloc);
    relocate( at, rspec, format );
    assert (old_stubsReloc - (int)stubsReloc == cLocEntrySize, "");
  }
  // Only insert the stub's relocation info once.
  _stubsReloc += (_stubs_reloc_count * cLocEntrySize);
  _stubs_reloc_count = 0;

  _instsEnd = temp;
}

// public inline
// Inform CodeBuffer that incoming code and relocation will be for stubs
void CodeBuffer::start_a_stub() {
  assert( _stubsStart != NULL, "Do not have any stubs to relocate");
  
  _instsEnd_before_stubs        = _instsEnd;
  _instsOverflow_before_stubs   = _instsOverflow;
  _instsEnd       = _stubsEnd;
  _instsOverflow  = _stubsOverflow;
}

// public inline
// Inform CodeBuffer that incoming code and relocation will be code
void CodeBuffer::end_a_stub() {
  assert( _stubsStart != NULL, "Do not have any stubs to relocate");
  
  _stubsEnd      = _instsEnd;
  _stubsOverflow = _instsOverflow;
  _instsEnd      = _instsEnd_before_stubs;
  _instsOverflow = _instsOverflow_before_stubs;
  _instsEnd_before_stubs      = NULL;
  _instsOverflow_before_stubs = NULL;
}

void CodeBuffer::relocate(address at, RelocationHolder const& spec, int format) {
  relocInfo::relocType rtype = (relocInfo::relocType) spec.reloc()->type();
  if (rtype == relocInfo::none)  return;
  if (rtype == relocInfo::oop_type && at == NULL)  return;    // no reloc necessary

  // The assertion below have been adjusted, to also work for
  // relocation for fixup.  Sometimes we want to put relocation
  // information for the next instruction, since it will be patched
  // with a call.
  assert( code_begin() <= at && at <= code_end()+1, 
          "cannot relocate data outside code boundaries");

  // Store stub relocation info until finished with code generation.
  if( _instsEnd_before_stubs != NULL /* doing stubs */ ) {
    add_stub_reloc( at, spec, format );
    return;
  }

  if (_locsEnd == NULL) {
    // no space for relocation information provided => code cannot be
    // relocated.  Make sure that relocate is only called with rtypes
    // that can be ignored for this kind of code.
    assert(rtype == relocInfo::none              ||
           rtype == relocInfo::runtime_call_type ||
           rtype == relocInfo::internal_word_type||
           rtype == relocInfo::external_word_type, 
           "code needs relocation information");
    return;
  }

  assert( sizeof(relocInfo) == sizeof(short), "change this code");

  int len = at - _instsStart;
  int offset = len - _last_reloc_offset;
  assert(offset >= 0, "relocation addrs may not decrease");
  _last_reloc_offset = len;

  while( offset >= relocInfo::offset_limit() ) {
    *_locsEnd++ = filler_relocInfo();
    offset -= filler_relocInfo().addr_offset();
  }

  // write the prefix data, if any
  _locsEnd = spec.pack_data(_locsEnd, at);

  assert( offset < relocInfo::offset_limit(), "Check offset" );
  *_locsEnd++ = relocInfo( rtype, offset );
  // Note: since this is called after a write, end == overflow is ok.
  // (the next relocate() will overwrite some memory after _locsOverflow,
  // but this one did not)
  assert( _locsEnd <= _locsOverflow, "Estimate of code size or relocation entries was too low");

  if (relocInfo::have_format) {
    if (format != 0)
      (_locsEnd-1)->set_format(format);
  } else {
    assert(format == 0, "bad format");
  }
}


void CodeBuffer::relocate_raw(address at, relocInfo::relocType rtype, const short* data, int datalen, int format) {
  // a debugging routine, at present
  assert(_locsEnd != NULL, "must be collecting relocation info");

  // put in a fake relocation, to shake out various corner cases
  relocate(at, runtime_call_Relocation::spec());

  _locsEnd -= 1;
  int offset = _locsEnd->addr_offset();
  *_locsEnd = prefix_relocInfo();
  short* dp = (short*)_locsEnd->data();
  for (int i = 0; i < datalen; i++)
    *dp++ = data[i];
  _locsEnd = _locsEnd->finish_prefix(dp);
  *_locsEnd++ = relocInfo(rtype, offset, format);

  assert( _locsEnd < _locsOverflow, "routine is too long to compile");
}

void CodeBuffer::copy_code(CodeBlob* cb) { 
#ifndef CORE
  // %%% make sure the new location is sufficiently aligned
  const u_char fill_code = Assembler::code_fill_byte();

#ifdef ASSERT
  if ( (_stubsEnd - _stubsStart) != 0 ) {
    // Post-method stubs are considered part of the code
    assert (_instsEnd >= _stubsEnd, "_instsEnd not set correctly");
  }

  if ( (_constEnd - _constStart) != 0 ) {
    assert (_instsEnd >= _constEnd, "_instsEnd not set correctly");
  }
#endif

  // align code
  // Copy all instructions, code and stubs
  while (code_size()  % oopSize != 0) *_instsEnd++ = fill_code;
  Memory::copy_oops((oop*)_instsStart, (oop*)cb->instructions_begin(), code_size() /oopSize);

  int alignSize = max(sizeof(jdouble), CodeEntryAlignment);
  assert( (cb->instructions_begin() - _instsStart) % alignSize == 0, "copy must preserve alignment");

  // Fix the pc relative information after the move
  int delta = (address)cb->instructions_begin() - _instsStart;
  cb->fix_relocation_at_move(delta);
#ifndef PRODUCT
  if (PrintAssembly) {
    Disassembler::decode(cb);
  }
#endif
#endif
}

void CodeBuffer::copy_relocation(CodeBlob* blob) {
#ifndef CORE
 // align relocation info
  while (locs_size() % oopSize != 0) *_locsEnd++  = filler_relocInfo();
  relocInfo* locs = RelocIterator::create_index(this, blob->relocation_begin(), blob->relocation_end());
  assert(locs >= blob->relocation_begin(), "relocs in bounds");

  assert((address)locs + locs_size() <= (address)blob->relocation_end(), "relocs in bounds");
  Memory::copy_oops((oop*)_locsStart, (oop*)locs, locs_size()/oopSize);
#endif
}

void CodeBuffer::oops_do( void f(oop*) ) {
#ifndef CORE
  RelocIterator iter(this);
  while (iter.next()) {
    if (iter.type() == relocInfo::oop_type) {
      iter.reloc()->oops_do(f);
    }
  }
#endif
}


// fp constants support

// enters the constant c into the constant table and
// returns the address of the entry
address CodeBuffer::insert_double_constant(jdouble c) {
  assert (_constEnd != NULL && _constStart != NULL && _constOverflow != NULL, "fp constant table not initialized");
  // Align _constEnd to a multiple of wordSize * 2
  address double_address;
  int alignment = wordSize * 2;
  if (((unsigned) _constEnd) % alignment != 0) {
    _constEnd += alignment - (((unsigned) _constEnd) % alignment);
  }
  double_address = _constEnd;
  assert ((int)double_address % (wordSize * 2)==0, "double constant not aligned");
  *((jdouble*)_constEnd) = c;
  _constEnd += sizeof(jdouble);
  assert (_constStart <= _constEnd && _constEnd <= _constOverflow, "fp constant table overflow");
  return double_address;
}


address CodeBuffer::insert_float_constant(jfloat c) {
  assert (_constEnd != NULL && _constStart != NULL && _constOverflow != NULL, "fp constant table not initialized");
  address float_address;
  float_address = _constEnd;
  assert ((int)float_address % wordSize == 0, "float constant not aligned");
  *((jfloat*)_constEnd) = c;
  _constEnd += sizeof(jfloat);
  assert (_constStart <= _constEnd && _constEnd <= _constOverflow, "fp constant table overflow");
  return float_address;
}

#ifndef PRODUCT

void CodeBuffer::decode() {
  Disassembler::decode(decode_begin(), code_end());
  _decode_begin = code_end();
}


void CodeBuffer::skip_decode() {
  _decode_begin = code_end();
}


void CodeBuffer::decode_all() {
  Disassembler::decode(code_begin(), code_end());
}


void CodeBuffer::print() {
  tty->print_cr("CodeBuffer:");
  tty->print_cr("Code  [0x%x <- used -> 0x%x | <- free -> 0x%x)", 
                _instsStart, _instsEnd, _instsOverflow );
  tty->print_cr("Reloc [0x%x <- used -> 0x%x | <- free -> 0x%x)", 
                _locsStart,  _locsEnd, _locsOverflow );
}

#endif // PRODUCT

