/*
 * @(#)codeBuffer.hpp	1.36 01/02/09 14:44:27
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// A CodeBuffer describes a memory space into which assembly
// code is generated. A code buffer comes in two variants:
//
// (1) A CodeBuffer referring to an already allocated piece of memory:
//     This is used to direct 'static' code generation (e.g. for interpreter
//     or stubroutine generation, etc.).  This code comes with NO relocation
//     information.
//
// (2) A CodeBuffer referring to a piece of memory allocated when the
//     CodeBuffer is allocated.  This is used for nmethod generation.

class  MacroAssembler;
class  PhaseCFG;
class  Compile;
extern MacroAssembler* theMacroAssm;

class CodeBuffer: public ResourceObj {
 private:
  address     _instsStart;	// first byte of instructions
  address     _instsEnd;	// first byte after instructions
  address     _instsOverflow;	// first byte after instructions buffer
  address     _mark;            // user mark
  int         _exception_offset;// Offset to exception handler
  // Stubs share the instruction space, up to _instsOverflow
  address     _stubsStart;      // first byte of stubs
  address     _stubsEnd;        // first byte after stubs
  address     _stubsOverflow;   // first byte after instruction buffer
  // Stub relocation info is stored until code generation is finished.
  address     _instsEnd_before_stubs; // Store _instsEnd here while doing stubs
  address     _instsOverflow_before_stubs; // store _instsOverflow while doing stubs
  char       *_stubsReloc;      // Storage for stub's relocation information
  int         _stubs_reloc_count; // Number of relocation entries in stubs

  address     _constStart;         // Storage for float and double constants
  address     _constEnd;
  address     _constOverflow;      // Storage for float and double constants
  
  relocInfo*  _locsOverflow;	// first byte after relocation information buf
  int         _last_reloc_offset;

  OopRecorder* _oop_recorder;

  address     _decode_begin;	// start address for decode
  address     decode_begin();

  // Used internally to store reloc info during stub generation
  // After code generation, call relocate_stubs()
  void        add_stub_reloc(address at, RelocationHolder const & rspec, int format);

  void        put_reloc(address at, relocInfo::relocType rt, int format = 0);

  relocInfo*  _locsStart;	// first byte of relocation information
  relocInfo*  _locsEnd;		// first byte after relocation information

 public:
  // (1) code buffer referring to pre-allocated memory
  CodeBuffer(address code_start, int code_size); 
  // (2) code buffer allocating memory for code & relocation info
  //     instsSize must include both code and stubs sizes.
  CodeBuffer(int instsSize, int locsSize, int stubsSize, int constSize, bool needs_oop_recorder,
             address instrs_memory = NULL, address locs_memory = NULL);

  static int insts_memory_size(int instsSize);
  static int locs_memory_size(int locsSize);
  
  // Properties
  address code_begin() const          { return _instsStart; };
  address code_end() const            { return _instsEnd; }
  address code_limit() const          { return _instsOverflow; };
  bool contains(address pc) const 
  { return code_begin() <= pc && pc < code_end(); }

  void    set_mark()                  { _mark = _instsEnd; }
  address mark() const                { return _mark; }
  
  // size in bytes
  int code_size() const { return _instsEnd - _instsStart; } 
  // capacity in bytes
  int capacity() const { return _instsOverflow - _instsStart; } 

  // Exception code
  int exception_offset() const { return _exception_offset; }
  void set_exception_offset( int off ) { _exception_offset = off; }

  // Access location of stubs
  address    stub_begin() const       { return _stubsStart; }
  address    stub_end() const         { return _stubsEnd; }
  // Inform codeBuffer when generating stub code and relocation info
  void       start_a_stub();
  void       end_a_stub();
  // Install relocation info stored during stub generation
  void       relocate_stubs();

  void       set_stubs_begin (address a) { _stubsStart = a; }
  void       set_stubs_end   (address a) { _stubsEnd   = a; }

  address    ctable_start() const     { return _constStart;  }
  address    ctable_end() const       { return _constEnd;    }

  int        locs_size() const        { return (_locsEnd - _locsStart)*sizeof(relocInfo); }
  relocInfo* locs_start() const       { return _locsStart; }
  relocInfo* locs_end() const         { return _locsEnd; }

  OopRecorder* oop_recorder()         const { return _oop_recorder;     }
  void     set_oop_recorder(OopRecorder* r) {        _oop_recorder = r; }

  // return the total size of relocation information, plus indexes, in nmethod
  int reloc_size() const { return RelocIterator::locs_and_index_size(code_size(), locs_size()); }

  // Code generation
  inline void set_code_end(address end);    // adjusts the code end
  inline void set_code_end_after_constants(); // join constants
  void relocate(address at, RelocationHolder const& rspec, int format = 0);
  void relocate(address at,    relocInfo::relocType rtype, int format = 0) {
    if (rtype != relocInfo::none)
      relocate(at, Relocation::spec_simple(rtype), format);
  }
  void relocate_raw(address at, relocInfo::relocType rtype, const short* data = NULL, int datalen = 0, int format = 0);

  // NMethod generation
  void    copy_relocation(CodeBlob* blob);
  void    copy_code(CodeBlob* nm);

  // GC support
  void    oops_do(void f(oop*));

  // fp constants support
  address insert_double_constant(jdouble c);
  address insert_float_constant (jfloat c);

#ifndef PRODUCT
 public:
  // Printing / Decoding
  // decodes from decode_begin() to code_end() and sets decode_begin to end
  void    decode();		
  void    decode_all();		// decodes all the code
  void    skip_decode();	// sets decode_begin to code_end();
  void    print();
#endif
};


inline void CodeBuffer::set_code_end( address end ) {
  // Note: since this is called after a write, end == overflow is ok.
  // (the next emit() will overwrite some memory after _instsOverflow,
  // but this one did not)
  assert( end >= _instsStart && end <= _instsOverflow, "CodeBuffer overflow");
  //  assert( end >= _instsEnd, "illegal set_code_end" );
  _instsEnd = end;
}


inline void CodeBuffer::set_code_end_after_constants() {
  _instsEnd = ctable_end();
  _instsOverflow = _constOverflow;
}

