/*
 * @(#)c1_CodeEmitter.hpp	1.10 01/02/09 14:44:28
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

class Item;
class Compilation;

//-----------------------------------------------------

// This class describes a patching point; a patching point
// could be a move with an offset or a simple move of a constant; 

class PatchingDesc: public StackObj {
 private:
  int _pc_start;
  int _pc_offset;
  int _field_offset; // is either 0 or 1, depending if the move is for low or high word
 public:
  PatchingDesc() : _pc_start(0), _pc_offset(0), _field_offset(0) {}
  void set(int pc_start, int pc_offset) {
    _pc_start = pc_start; _pc_offset = pc_offset;
  }

  void set_hi_word(); // marks move that moves Hi-word
  void set_lo_word(); // marks move that moves Lo-word

  int pc_start() const { return _pc_start; }
  int pc_offset() const { return _pc_offset; }
  int field_offset() const { return _field_offset; }
};


class FPUConstEntry: public ResourceObj {
 private:
  address    _pc;
  const JavaValue _fp_value;
 public:
  FPUConstEntry(jfloat f):  _fp_value (JavaValue(f))  { _pc = NULL; }
  FPUConstEntry(jdouble d): _fp_value (JavaValue(d))  { _pc = NULL; }

  void set_pc(address pc) { _pc = pc; }

  bool is_float()   const { return _fp_value._type == T_FLOAT; }
  bool is_double()  const { return _fp_value._type == T_DOUBLE; }

  jfloat  float_value()  const { assert (is_float() , "wrong type"); return _fp_value._value.f; }
  jdouble double_value() const { assert (is_double(), "wrong type"); return _fp_value._value.d; }

  address pc()           const { return _pc; }
};



class ConstantTable: public StackObj {
 private:
  GrowableArray<FPUConstEntry*>* _entries;
 public:
  ConstantTable ();
  void append_float  (jfloat f);
  void append_double (jdouble d);

  address address_of_float_constant  (jfloat f);
  address address_of_double_constant (jdouble d);

  void emit_entries(MacroAssembler* masm);

  void print() PRODUCT_RETURN;
};


class CodeEmitInfo: public ResourceObj {
 private:
  const RInfoCollection* _register_oops; // registers that contain oops
 public:
  int _bci;
  GrowableArray<int>*    _spilled_oops;  // all spill-indeces that are oops
  ValueStack*            _stack;         // used by deoptimization
  ValueStack*            _monstack;      // used by deoptimization
  
  CodeEmitInfo(int bci, GrowableArray<int>* soops, ValueStack* stack, const RInfoCollection* oops_in_regs = NULL) :
    _bci(bci), _spilled_oops(soops), _stack(stack), _monstack(NULL), _register_oops(oops_in_regs) {}

  CodeEmitInfo(CodeEmitInfo* info) :
    _bci(info->_bci), _spilled_oops(info->_spilled_oops),
    _stack(info->_stack), _monstack(info->_monstack), _register_oops(info->_register_oops) {}

  bool has_register_oops() const { return _register_oops != NULL && _register_oops->number_of_registers() > 0; }
  const RInfoCollection* register_oops() const { return _register_oops; }
};


class LineNumbersOffset;
class CodeEmitter: public StackObj {

  // some stubs need access to helper functionality
  friend class NewInstanceStub;
  friend class NewTypeArrayStub;
  friend class NewObjectArrayStub;
  friend class MonitorEnterStub;
  friend class MonitorExitStub;
  friend class CheckCastStub;
  friend class InstanceOfStub;
  friend class ArrayStoreCheckStub;
  friend class ArrayCopyStub;
  friend class ConvertStub;

 private:
  C1_MacroAssembler* _masm;
  Compilation*       _compilation;
  ciMethod*          _method;
  FrameMap*          _frame_map;
  ConstantTable*     _const_table;
  JVMCI_OffsetDesc*  _offsets;
  int                _esp_offset;                // track esp changes
  bool               _need_debug_info;           // set to true to create debug information (locals, expression stack)
  bool               _bailout;                   // set to true to bail out of compilation
  Label              _throw_entry_label;         // entry point for throw_op


  bool               _cc_locked;                 // true if CC register is locked
  GrowableArray<JVMCI_DebugInfo*>* _info;
  GrowableArray<CodeStub*>*  _slow_case_stubs; // a part of code section
  GrowableArray<CodeStub*>*  _call_stubs;      // are part of stub section
  LineNumbersOffset* _linenumbers_offset;

  // accessors

  bool need_debug_info() const { return false; }
  Compilation*  compilation() const { return _compilation; }
  ciMethod*     method () const { return _method;     }
  FrameMap*     frame  () const { return _frame_map;  }
  GrowableArray<JVMCI_DebugInfo*>* info() const { return _info; }

  // accessors

  void set_patching_desc(PatchingDesc* patch);

  // ask about FPU constants (cannot use C++ comparisions)
  bool is_zero_float(jfloat f) const;
  bool is_one_float (jfloat f) const;

  bool is_zero_double(jdouble f) const;
  bool is_one_double (jdouble f) const;

  // helpers
  bool needs_icache() const;
  bool needs_null_check_on_entry() const;
  void check_icache ();
  void build_frame  (int max_spills);
  void init_local   (int local_ix);
  void swap_reg     (Register a, Register b);
  int  stack_offset_in_bytes (int nof_spills);

  // In case where we do not install code, we do not want to have stops
  // at unimplemented code patterns
  void not_implemented()                 { if (InstallMethods) Unimplemented(); }

  void patching_epilog(PatchingDesc* patch, CodeEmitInfo* info, PatchingStub::PatchID id);
  void jobject_2_reg_with_patching (RInfo reg, ciObject* obj, CodeEmitInfo* info);
  void try_fast_ObjectHashCode();

  // two words on the stack must be adjacent (local index, index+1)
  void check_double_address(int local_index);

  // debug information
  void add_call_info  (int pc_offset, CodeEmitInfo* info);  // for calls
  void add_debug_info (int pc_offset, CodeEmitInfo* info, bool at_call = false);
  void add_debug_info_for_branch(int bci, const RInfoCollection* oop_regs);

  // item conversions
  static Register item2reg    (Item* item);
  static Register item2regLO  (Item* item);
  static Register item2regHI  (Item* item);

  // For stack objects only (respects endian conventions for stack long and double)
  Address  stack_item2addr   (const Item* item, bool for_hi_word = false);
  Address  stack_item2addrLO (const Item* item);
  Address  stack_item2addrHI (const Item* item);

  // constant item transformations
  jobject item2jobject (Item* item);
  jint    item2int     (Item* item);
  jint    item2intLO   (Item* item);
  jint    item2intHI   (Item* item);

  // method tracing
  void trace_method_entry(jint con);
  void trace_method_exit();

  void bang_stack_with_offset(int offset);

//---------------------------------------------------------------------------------
 public:
  CodeEmitter(
    Compilation* compilation,
    ConstantTable* table,
    JVMCI_OffsetDesc* offsets
  );
  C1_MacroAssembler* masm() const        { return _masm;     }

  int  frame_size(); // fp - sp

  int  esp_offset() const { return _esp_offset; } 
  void set_esp_offset(int n) { _esp_offset = n; } // used by stubs

  void set_bailout(const char* msg);
  bool must_bailout() const              { return _bailout;  }

  // call if an instruction implicitly sets an FP result (sets fpu_stack)
  void set_fpu_result      (RInfo reg);
  void remove_fpu_result   (RInfo reg);
  bool is_empty_fpu_stack  ();
  void set_fpu_stack_empty ();
  void fpop();

  void restore_cached_receiver();
  void load_receiver       (Register receiver);

  // used to mark the area where condition code may not be changed
  void lock_cc()            { _cc_locked = true;  }
  void unlock_cc()          { _cc_locked = false; }
  bool is_cc_locked()       const { return _cc_locked; }

  // will set _bailout to true if not enough code_space available
  void check_codespace();

  void add_call_info_here (CodeEmitInfo* info);   // adds call info for current pc offset
  void add_debug_info_here(CodeEmitInfo* info);   // adds debug info for current pc offset

  void bind_label (Label* label);

  void print() PRODUCT_RETURN;

  JVMCI_LineNumbersDesc* get_jvmpi_lineno_table();

  // ======================== code emission ==================================

  void bang_stack();

  void copy_fpu_item (RInfo toReg, Item* from);
  void dup_fpu_register (RInfo reg, Item* item);

  void pop_item  (Item* item);
  void push_item (Item* item);
  void push_jint (jint i);

  void push_item_with_rinfo(Item*, RInfo reg);

  void emit_slow_case_stubs();
  void emit_call_stubs();
  void emit_accessor_code(int offset, BasicType type);

  int  emit_exception_handler(int max_spills, RInfo cached_receiver);
  void emit_osr_entry(int max_spills, int number_of_locks, Label* continuation, int osr_bci);
  void emit_constants();

  void method_entry  (GrowableArray<int>* init_locals, int max_spill);

  // only for natives
  void native_method_exit (CodeEmitInfo* info);  
  void native_call   (address native_entry, CodeEmitInfo* info);

  void local_2_reg   (RInfo reg, int local_index, BasicType t = T_ILLEGAL);
  void reg_2_local   (int local_index, RInfo reg, BasicType t = T_ILLEGAL);
  void item_2_local  (int local_index, Item* item);
  void item_2_reg    (RInfo reg, Item* item      );
  void reg_2_item    (Item* item, RInfo reg      );
  void clear_local   (int local_index);

  void item_2_reg_with_patching    (RInfo reg   , Item* item, CodeEmitInfo* info);

  void store_stack_parameter (Item* item, int offset_from_sp_in_words);

  void move_item_to_cpu_register (const Item& src, RInfo dst); // this is meant to move a floating point value to a  cpu register on sparc
  void move_item_to_fpu_register (const Item& src, RInfo dst); // this is meant to move an           int value to an fpu register on sparc

  void spill(int spill_ix, Item* item);

  void move_spill    (int to_spill_ix, int from_spill_ix, ValueType* type, RInfo tmp);

  void implicit_null_check        (Item* item, CodeEmitInfo* info);
  void explicit_null_check        (Item* item, CodeEmitInfo* info);
  void explicit_div_by_zero_check (Item* item, CodeEmitInfo* info);
  void range_check                (Item* array, Item* index, CodeEmitInfo* info);

  void array_store_check (Item* array, Item* value, Item* index, RInfo tmp1, RInfo tmp2, CodeEmitInfo* info);

  void field_load    (RInfo dst, BasicType dst_type, Item* object, int offset, bool needs_patching, bool is_loaded, CodeEmitInfo* info);
  void field_store   (BasicType value_type, Item* object, int offset, Item* value, bool needs_patching, bool is_loaded, CodeEmitInfo* info, RInfo temp_reg);
  void indexed_load  (RInfo dst, BasicType dst_type, Item* array, Item* index, CodeEmitInfo* info);
  void indexed_store (BasicType val_type, Item* array, Item* index, Item* value, RInfo temp_reg, CodeEmitInfo* info);

  void negate        (RInfo dst, Item* value);
  void math_intrinsic(ciMethod::IntrinsicId id,  RInfo dst, Item* value, RInfo thread_cache = FrameMap::_noRInfo);

  void arithmetic_op_int  (Bytecodes::Code code, Item* result, Item* left, Item* right, RInfo tmp);
  void arithmetic_op_long (Bytecodes::Code code, Item* result, Item* left, Item* right);
  void arithmetic_op_fpu  (Bytecodes::Code code, Item* result, Item* left, Item* right);
  void arithmetic_idiv    (Bytecodes::Code code, Item* result, Item* left, Item* right, RInfo scratch, CodeEmitInfo* info);

  void arithmetic_call_prolog(ValueType* result);
  void arithmetic_call_epilog(ValueType* result);
  void arithmetic_call_op (Bytecodes::Code code, RInfo temp_reg);

  void long_mul();

  void shift_op   (Bytecodes::Code code, RInfo dst_reg, Item* value, Item* count, RInfo tmp);

  void logic_op   (Bytecodes::Code code, RInfo dst_reg, Item* left, Item* right);
  void compare_op (Bytecodes::Code code, RInfo dst_reg, Item* left, Item* right);

  void convert_op (Bytecodes::Code code, RInfo dst_reg, Item* value);
  void call_convert_op(Bytecodes::Code code, RInfo tmp_reg);

  void fast_call_convert_op(Bytecodes::Code code, Item* val, RInfo reg, bool is_32bit);

  void call_convert_epilog(int nof_floats_to_restore, ValueType* result_type);
  void call_convert_prolog(int nof_floats_to_save,    ValueType* result_type);

  // clean this up: either use only sig_types or size_of_arguments
  void call_op(Bytecodes::Code code, const GrowableArray<BasicType>* sig_types, int size_of_arguments, CodeEmitInfo* info, bool optimized, bool needs_null_check, Item* receiver);

  void throw_op (Item* exception, CodeEmitInfo* info);

  void monitor_enter (Item* object, RInfo lock, RInfo hdr, RInfo scratch, int monitor_no, CodeEmitInfo* info);
  void monitor_exit  (Item* object, RInfo lock, RInfo hdr, int monitor_no);

  void goto_op (BlockBegin* dst, int bci, bool is_backward_branch, const RInfoCollection* oop_regs);
  void jsr_op  (BlockBegin* dst, int bci);
  void ret_op  (int local_ix, RInfo cached_in_reg, int bci);

  If::Condition if_op  (int phase, If::Condition cond, Item* x, Item* y, BlockBegin* t_dest, BlockBegin* f_dest, BlockBegin* u_dest,
                        int bci, bool is_backward_branch, const RInfoCollection* oop_regs);
  void tableswitch_op  (Item* tag, int match, BlockBegin* dest, int bci, bool is_backward_branch, const RInfoCollection* oop_regs);
  void lookupswitch_op (Item* tag, int key,   BlockBegin* dest, int bci, bool is_backward_branch, const RInfoCollection* oop_regs);
  void lookupswitch_range_op (Item* tag, int low_key, int high_key, BlockBegin* dest, int bci, bool is_backward_branch, const RInfoCollection* oop_regs);

  void ifop_phase2     (RInfo dst, Item* tval, Item* fval, Instruction::Condition cond);
  void ifop_phase1     (Item* x, Item* y);

  void new_instance    (RInfo dst, ciInstanceKlass* klass, RInfo scratch1, RInfo scratch2, RInfo scratch3, RInfo klass_reg, CodeEmitInfo* info);
  void new_type_array  (RInfo dst, BasicType elem_type, Item* length, RInfo scratch1, RInfo scratch2, RInfo scratch3, RInfo scratch4, RInfo klass_reg, CodeEmitInfo* info);
  void new_object_array(RInfo dst, ciKlass* elem_klass, Item* length, RInfo scratch1, RInfo scratch2, RInfo scratch3, RInfo scratch4, RInfo klass_reg, CodeEmitInfo* info);
  void new_multi_array (RInfo dst, ciKlass* klass, int rank, RInfo tmp, CodeEmitInfo* info);

  void checkcast_op  (RInfo dst_reg, Item* obj, ciKlass* k, RInfo k_RInfo, RInfo klass_RInfo, CodeEmitInfo* info);
  void instanceof_op (RInfo dst_reg, Item* obj, ciKlass* k, RInfo k_RInfo, RInfo klass_RInfo, CodeEmitInfo* info);
  void instance_of_test_op(RInfo dst_reg, Item* obj_item, ciKlass* k, RInfo klass_ri, BlockBegin* is_instance, BlockBegin* not_is_instance, bool is_inverted, CodeEmitInfo* info);

  void add_nop ();
  void add_nop_with_info(CodeEmitInfo* info); 
  void align_backward_branch();

  void increment_reg(RInfo reg);
  void decrement_reg(RInfo reg);

  void increment_local(int local_index);
  void decrement_local(int local_index);

  // support for jvmpi
  void mark_offset_for_bci(int bci);

  void return_op_prolog (int monitor_no, RInfo receiver);

  void return_op        (Item* result);
  void safepoint_return (int argSize);

  // JVMPI
  void jvmpi_method_exit();
  void jvmpi_method_entry(CodeEmitInfo* info);

  // libraries
  void triglib(ciMethod::IntrinsicId id);           // java.lang.Math.sin and cos
  void string_compare();
  void arraycopy(Item* src_item, Item* src_pos_item, Item* dst_item, Item* dst_pos_item, Item* length_item, RInfo scratch1, RInfo scratch2, RInfo tmp, CodeEmitInfo* info);

  void set_32bit_fpu_precision();
  void restore_fpu_precision();

  void breakpoint();
  void handler_entry();

  #include "incls/_c1_CodeEmitter_pd.hpp.incl"
};
