// @(#)c1_CodeGenerator.cpp	1.53 01/02/09 14:44:29
// 
// Copyright 1993-2001 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_c1_CodeGenerator.cpp.incl"


//----------------------------------------------------------------------
//                        Utilities
//----------------------------------------------------------------------

// The constructor locks the specified register and the destructor unlocks it;
// Use to prevent locking of a certain register
HideReg::HideReg(ValueGen* vg, RInfo reg, bool spill /*= false*/) {
  _vg = vg;
  _reg = reg;
  if (spill && !_vg->is_free_rinfo(reg)) {
    while (!_vg->ra()->is_free_reg(reg)) {
      _vg->spill_register(reg);
    }
  }
  assert(_vg->is_free_rinfo(reg), "cannot hide it, because it is already locked");
  _vg->ra()->lock_reg(NULL, reg, 1);
  _vg->ra()->incr_spill_lock(reg);
}

// spills, locks
HideReg::HideReg(ValueGen* vg, ValueType* t) {
  _vg = vg;
  _reg = _vg->get_free_rinfo(t); // <- may spill
  _vg->ra()->lock_reg(NULL, _reg, 1);
  _vg->ra()->incr_spill_lock(_reg);
}


HideReg::HideReg(ValueGen* vg, c1_RegMask mask) {
  _vg = vg;
  _reg = _vg->get_free_rinfo(mask); // <- may spill
  _vg->ra()->lock_reg(NULL, _reg, 1);
  _vg->ra()->incr_spill_lock(_reg);
}


HideReg::~HideReg() {
  _vg->ra()->decr_spill_lock(_reg);
  _vg->ra()->free_reg(_reg);
  assert(_vg->is_free_rinfo(_reg), "must be free");
}


//--------------------------------------------------------------


CodeGenerator::CodeGenerator (ValueGen* gen, ValueGenInvariant* vgi):  _gen(gen), _max_spills(0), _vgi(vgi) {
}


void CodeGenerator::block_do_prolog(BlockBegin* block) {
  _gen->align_block(block);
  _gen->bind_label(block->label());
  // add PcDesc to debug info so we can find entry point for exception handling
  // (note: we are finding the handler entry point by searching the PcDescs - if
  //        we use another method to find the exception handler entry we don't
  //        need to emit this information anymore).
  if (block->is_set(BlockBegin::exception_entry_flag)) _gen->exception_handler_start(block->bci());
}


void CodeGenerator::block_do_epilog(BlockBegin* block) {
  if (_vgi->code_emit() != NULL && _vgi->code_emit()->must_bailout()) return;
  assert(_gen->is_fpu_stack_free() && _gen->is_cpu_stack_free(), "still some values on FPI stack");
  assert(_vgi->reg_alloc()->are_all_free(), "not everything released at block end");
  _max_spills = max(_max_spills, _vgi->reg_alloc()->max_spills());
}


void CodeGenerator::block_do(BlockBegin* block) {
  if (_vgi->code_emit() != NULL && _vgi->code_emit()->must_bailout()) return;
  _gen->set_invariant(_vgi);
  block_do_prolog(block);
  _gen->block_prolog(block);
  _gen->set_block(block);

  for (Instruction* instr = block; instr != NULL; instr = instr->next()) {
    if (instr->is_root()) _gen->do_root(instr);
  }
  _gen->set_block(NULL);
  _gen->block_epilog(block);
  block_do_epilog(block);
}


static void clear_state_items(Value* x) {
  assert((*x)->as_Phi() != NULL, "state must only contain phis");
  (*x)->clear_item();
}


// This is same as clear_block_items in c1_Compiler1.cpp; factorize!
void CodeGenerator::clear_instruction_items(BlockBegin* b) {
  // clear phi's
  b->state_values_do(clear_state_items);
  // clear remaining instructions
  for (Instruction* n = b; n != NULL; n = n->next()) n->clear_item();
}


//-------------------------ValueGen-----------------------------

bool  ValueGen::_init_done = false;


void ValueGen::init_value_gen() {
  if (_init_done) return;
  _init_done = true;
}


bool ValueGen::is_backward_branch(Value instr, BlockBegin* dest) const {
  bool is_bb =  block()->is_after_block(dest);
  if (is_bb) {
    // the backward branch flag is set so that in the code emission pass we can 
    // align blocks that are backward branch targets
    dest->set(BlockBegin::backward_branch_target_flag);
  }
  return is_bb;
}


#ifndef PRODUCT
void ValueGen::print_at_do_root(Value instr, bool before_visit) {
  if (PrintC1RegAlloc) {
    if (before_visit) {
      tty->print("ROOT bci %d before: ", instr->bci()); ra()->print();
      if (gen_code()) code_emit()->print();
    } else {
      tty->print("ROOT bci %d after:  ", instr->bci()); ra()->print();
      if (gen_code()) code_emit()->print(); 
      tty->cr();
    }
  }
}
#endif


// This is where the tree-walk starts; instr must be root;
void ValueGen::do_root(Value instr) {
  assert(instr->is_root(), "use only with roots");
  if (gen_code()) {
    // check if there is enough space in CodeBuffer
    code_emit()->check_codespace();
    if (code_emit()->must_bailout()) return;
  }
  debug_only(ra()->check_registers();)
  assert(ra()->are_all_spill_locks_free(), "no spill locks may be alive");
  assert(!gen_code() || code_emit()->esp_offset() == 0, "esp_offset must be 0 across root bytecodes");
  // Phi's are visited separately during block prolog
  if (instr->as_Phi() != NULL) {
    assert(instr->item() != NULL, "phi item must be set");
    return;
  }

  print_at_do_root(instr, true);
  _hint   = HintItem::no_hint();
  _result = new RootItem(instr);
  assert (instr->item() == NULL, "somebody already visited it?");
  instr->visit(this);
#ifdef ASSERT
  if (instr->item() != NULL && instr->item()->is_register()) {
    assert(!ra()->is_spill_locked(instr->item()->get_register()), "register may not be spill locked");
  }
#endif
  // emits code, allocates register
  finish_root(instr);

  print_at_do_root(instr, false);
  assert(!instr->has_uses() || instr->item()->is_valid(), "invalid item set");
  assert(!gen_code() || code_emit()->esp_offset() == 0, "esp_offset must be 0 across root bytecodes");
  assert(ra()->are_all_spill_locks_free(), "no spill locks may be alive");
  debug_only(ra()->check_registers();)
}


void ValueGen::release_item(Item* item) {
  if (item->is_register()) {
    if (item->type()->is_double() || item->type()->is_float()) {
      if (gen_code()) {
        code_emit()->remove_fpu_result(item->get_register());
      }
    }
    // do not release cached register
    // call rfree because no spill lock is active here
    if (!item->is_cached()) raw_rfree(result());
  } else if (item->is_spilled()) {
    sfree(item);
  }
}


// releases result register or spilling slots if use_count of instruction is zero;
// this can occur only with root instructions
void ValueGen::check_result_usage(Value instr) {
  if (result()->has_result() && !instr->has_uses()) {
    // there is a result but the use count is zero
    release_item(result());
    result()->set_no_result();
  }
}


#ifndef PRODUCT
void ValueGen::print_at_walk(Value instr, bool before_visit) {
  if (PrintC1RegAlloc) {
    if (before_visit) {
      tty->print("     BCI %d before: ", instr->bci()); 
      if (gen_code()) code_emit()->print();
      ra()->print();  
    } else {
      tty->print("     BCI %d after:  ", instr->bci());  
      if (gen_code()) code_emit()->print();
      ra()->print(); 
      tty->cr(); 
    }
  }
}
#endif


// This is called for each node in tree; the walk stops if a root is reached
void ValueGen::walk(Value instr) {
  if (gen_code() && compilation()->jvmpi_event_method_enabled()) {
    if (instr->bci() >= 0) {
      code_emit()->mark_offset_for_bci(instr->bci());
    }
  }
  //stop walk when encounter a root
  if (instr->is_root()) {
    assert(instr->item() != NULL, "this root has not yet been visited");
    // copy content of RootItem to stack Item.
    result()->set_from_item(instr->item());
    assert(result()->is_valid(), "invalid item");
    debug_only(check_item(instr->item());)
  } else {
    assert(instr->use_count() <= 1, "an instruction with use_count > 1 must be root ");
    print_at_walk(instr, true);
    instr->visit(this);
    assert(instr->use_count() > 0, "leaf instruction must have a use");
    // check_result_usage(instr);
    print_at_walk(instr, false);
  }
}


void ValueGen::item_free(Item* item) {
  if      (item->is_spilled() ) sfree(item);
  else if (item->is_register()) rfree (item);
}


// releases spill index held by item
void ValueGen::sfree(Item* item) {
  ra()->free_spill(item->get_spilled_index(), item->type());
}


// locks specified RInfo 'reg'; spill RInfo 'reg' if not free;
void ValueGen::lock_spill_rinfo(Value instr, RInfo reg, int ref_count) {
  // do a loop because of register pairs, that requires two calls to spill_register
  while (!ra()->is_free_reg(reg)) {
    spill_register(reg);
  }
  ra()->lock_reg(instr, reg, ref_count);
}


// find an instruction that can be spilled; mark that instruction for spill
// and return its RInfo; spill priority is set by bci of root instructions;
// the smaller the bci, the higher the chance for spilling
void ValueGen::spill_one(ValueType* type) {
  // we need a loop as we may need to spill 2 registers (long type)
  while (!ra()->has_free_reg(type)) {
    Value root = ra()->get_smallest_value_to_spill(type);
    assert(root != NULL, "must search expression stack for stuff to spill (not implemented yet)");
    spill_value(root);
  }
}


// search instruction that uses reg, find a spill index, and 
// spill that instruction
void ValueGen::spill_register(RInfo reg) {
  Value to_spill = ra()->get_value_for_rinfo(reg);
  assert(to_spill != NULL && to_spill->item() != NULL, "value not legal");
  spill_value(to_spill);
}



#ifndef PRODUCT
void ValueGen::print_spilling(int bci, int spill_ix) {
  if (PrintSpilling) {
    tty->print_cr("    spilling item for bci %d to spill-slot %d", bci, spill_ix);
  }
}
#endif


void ValueGen::spill_item(Item* item) {
  assert(item != NULL, "where is the item?");
  assert(item->is_root_item() || !item->value()->is_root(), "cannot spill a non-root item that points to a root instruction(inconsistency)");
  assert(item->is_register(), "why spill if item is not register?");
  assert(!item->is_cached(), "cannot spill cached item");
  int rc       = ra()->get_register_rc(item->get_register());
  int spill_ix = ra()->get_lock_spill(item->value(), rc);
  // 'spill_ix' is the locked spill slot
  for (int i = 0; i < rc; i++) {
    raw_rfree(item);
  }
  if (gen_code()) {
    print_spilling(item->value()->bci(), spill_ix);
    code_emit()->spill(spill_ix, item);
  }
  item->set_spill_ix(spill_ix);
  assert(ra()->check_spilled(item), "wrong spilled state");
}


// Value->item must reside in register: will be moved into spill area
void ValueGen::spill_value(Value value) {
  assert(value->item()->is_register(), "must be a register");
  assert(!ra()->is_spill_locked(value->item()->get_register()), "cannot spill a spill locked value");
  assert(value->item()->value() == value, "inconsistency");
  spill_item(value->item());
}


// returns true if caller register was successfuly moved to a callee register
bool ValueGen::try_caller_to_callee(Value value, RInfo hide_reg) {
  assert(value->item() != NULL && value->item()->is_register(), "must be a register");
  assert(!ra()->is_spill_locked(value->item()->get_register()), "cannot spill a spill locked value");
  assert(value->item()->value() == value, "inconsistency");
  assert(!value->item()->is_cached(), "cannot have cached value in caller saved register");
  // any callee saved register free?
  if (value->type()->is_int() || value->type()->is_object()) {
    // cannot handle other types yet
    RootItem* item = value->item()->as_root();
    assert(item != NULL, "not a root item");
    c1_RegMask mask = FrameMap::callee_save_regs();
    if (!hide_reg.is_illegal()) mask.remove_reg(hide_reg);
    if (ra()->has_free_reg(mask)) {
      RInfo to_reg = ra()->get_free_reg_from_mask(mask);
      assert(!to_reg.is_same(hide_reg), "operation failed");

      if (gen_code()) {
        code_emit()->item_2_reg(to_reg, item);
      }
      int rc = ra()->get_register_rc(item->get_register());
      for (int i = 0; i < rc; i++) {
        raw_rfree(value->item());
      }
      ra()->lock_reg(value, to_reg, rc);
      item->set_register(to_reg);
      return true;
    }
  }
  return false;
}



// This is used to preserve registers across calls; only values in
// calller-save registers are spilled(saved)
void ValueGen::spill_values_on_stack(ValueStack* stack, RInfo hide_reg, bool caller_to_callee) {
  for (int i = 0; i < stack->stack_size();) {
    Value val = stack->stack_at_inc(i);
    // roots on stack may have use_count 0, in which case we do not spill them
    if (val->is_root() && val->has_uses() && val->item()->is_valid()) {
      // Sometimes the register of the root item may already be freed->do not spill then
      if (val->item()->is_register()) {
        assert(!val->item()->is_cached(), "cannot spill cached register");
        if (is_caller_save_register(val->item()->get_register())) { 
          if (caller_to_callee) {
            bool done = try_caller_to_callee(val, hide_reg);
            if (!done ) spill_value(val);
          } else {
            spill_value(val);
          }
        }
      } else {
        // The root may not be on stack !! The reason is that we mark
        // Load locals as roots to force loading into registers as 
        // the stack location may be destroyed by a later store;
        // code example: a[i++]
        assert(val->item()->is_constant() || val->item()->is_spilled(),  "wrong item state");
      }
    }
  }
}


// Check out code in "check_result_usage" and factor it out
void ValueGen::release_roots(ValueStack* stack) {
  for (int i = 0; i < stack->stack_size();) {
    Value val = stack->stack_at_inc(i);
    if (val->is_root()) {
      release_item(val->item());
    }
  }
}



// Phi technique:
// This is about passing live values from one basic block to the other.
// In code generated with Java it is rather rare that more than one
// value is on the stack from one basic block to the other.
// We optimize our technique for efficient passing of one value
// (of type long, int, double..) but it can be extended.
// When entering or leaving a basic block, all registers and all spill
// slots are release and empty. We use the released registers
// and spill slots to pass the live values from one block
// to the other. The topmost value, i.e., the value on TOS of expression
// stack is passed in registers. All other values are stored in spilling
// area. Every Phi has an index which designates its spill slot
// At exit of a basic block, we fill the register(s) and spill slots.
// At entry of a basic block, the block_prolog sets up the content of phi nodes
// and locks necessary registers and spilling slots.


// Value v is the value that is spilled at spill_ix;
// least_spill_ix is the minimum index where the spill value at spill_ix
// has to be moved; this is used to move a spilled value out of area reserved for
// Phi-values

void ValueGen::move_spill_to(int spill_ix, int least_spill_ix, Value v) {
  for (int i = 0; i < v->type()->size(); ++i) {
    if (ra()->get_ref_count_at(spill_ix + i) > 0) {
      // release spill slot at spill_ix, while locking it after least_spill_ix
      Value to_move = ra()->value_spilled_at(spill_ix + i);
      int from_spill_ix = to_move->item()->get_spilled_index();
      int new_spill_ix = ra()->get_free_spill_after(least_spill_ix, to_move->type());
      assert(new_spill_ix >= least_spill_ix, "wrong index computation");
      HideReg hide(this, to_move->type());
      if (gen_code()) {
        code_emit()->move_spill(new_spill_ix, from_spill_ix, to_move->type(),hide.reg());
      }
      ra()->move_spills (new_spill_ix, from_spill_ix, to_move->type());
    }
  }
}


// return phi tos value; spill slots for other values are returned in 'spill_ixs' and values in 'phi_values'
// 'stack' describes the spill values
Value ValueGen::compute_phi_arrays(ValueStack* stack, GrowableArray<Value>* phi_values, 
                                   GrowableArray<int>* spill_ixs) {
  if (stack->stack_size() == 0) return NULL;
  bool  is_tos  = true;
  Value tos_val = NULL;
  // compute spill slots, spill values and tos-vallue
  for (int i = stack->stack_size(); i > 0;) {
    Value val = stack->stack_at_dec(i);
    assert(val != NULL, "wrong stack access");
    if (is_tos) {
      is_tos = false; tos_val = val;
    } else {
      int spill_ix = i;
      assert (0 <= spill_ix && spill_ix < stack->stack_size(), "wrong spill index");
      phi_values->append(val);
      spill_ixs->append(spill_ix);
    }
  }
  return tos_val;
}


// Moves all stack values into their PHI position
void ValueGen::move_to_phi(ValueStack* stack) {
  if (stack->stack_size() == 0) {
    assert(ra()->are_all_free(), "");
    assert(!gen_code() || code_emit()->is_empty_fpu_stack(), "still some values on FPU stack");
    return;
  }
  GrowableArray<Value> phi_values;
  GrowableArray<int>   spill_ixs;
  Value tos_val = compute_phi_arrays(stack, &phi_values, &spill_ixs);

  assert(tos_val != NULL, "");
  // First: move phi values into spill-slots
  for (int i = 0; i < phi_values.length(); i++) {
    Value v  = phi_values.at(i);
    int   ix = spill_ixs.at(i);
    if (v->item() != NULL && v->item()->is_spilled() && v->item()->get_spilled_index() == ix) {
      // everything is fine, because the value is already in right place
      ra()->free_spill(ix, v->type());
    } else {
      if (!ra()->is_free_spill(ix, v->type())) {
        // the spill slot for v is currently blocked; move content of that spill slot elsewhere
        move_spill_to(ix, stack->stack_size(), v);
      }
      assert(ra()->is_free_spill(ix, v->type()), "spill slot still not empty");
      // Note: we will not lock the spill slot, but we are emitting code to move it into it.
      Item item(v);
      item.handle_float_kind(); // INTEL specific
      ValueGen vg(&item, HintItem::no_hint(), this);
      load_item(&item);
      rfree(&item);
      assert(ra()->is_free_spill(ix, v->type()), "spill slot not empty");
      if (gen_code()) {
        code_emit()->spill(ix, &item);
      }
    }
  }
  // Second: now load TOS value into register
  if (tos_val != NULL) {
    RInfo reg = result_register_for(tos_val->type());
    HintItem hint(tos_val->type(), reg);
    Item item(tos_val);
    item.handle_float_kind();

    ValueGen vg(&item, &hint, this);
    // we cannot load_item_force here, because the spilling is not set up correctly (all is released)
    load_item_hint(&item, &hint);
    rfree(&item);
    if (gen_code()) {
      code_emit()->item_2_reg(reg, &item);
      code_emit()->set_fpu_stack_empty();
    }
  }
  assert(ra()->are_all_free(), "end of block: all registers ans spilling slots must be released");
  assert(!gen_code() || code_emit()->is_empty_fpu_stack(), "still some values on FPu stack");
}


// locks a register(s) that can accomodate the result of the instruction
RInfo ValueGen::rlock(Value instr) {
  while (!ra()->has_free_reg(instr->type())) {
    spill_one(instr->type()); 
  }
  return ra()->get_lock_reg(instr, instr->type());
}


// returns RInfo that corresponds to the call result;
RInfo ValueGen::rlock_call_result(Value instr) {
  RInfo reg = result_register_for(instr->type());
  while (!ra()->is_free_reg(reg)) {
    spill_register(reg);
  }
  ra()->lock_reg(instr, reg, instr->use_count());
  return reg;
}


// Try to lock using register in hint
RInfo ValueGen::rlock(Value instr, const Item* hint) {
  if (hint->has_hint() && ra()->is_free_reg(hint->get_register())) {
    if (!hint->is_cached()) ra()->lock_reg(instr, hint->get_register(), instr->use_count());
    return hint->get_register();
  } else {
    // no hint or hint register is locked
    return rlock(instr);
  }
}


// If item is register, mark item as not spillable
void ValueGen::set_maynot_spill(Item* item) {
  if (item->is_register()) {
    // increase spill lock counter in register allocator
    ra()->incr_spill_lock(item->get_register());
  }
}


// release register described by item (decrease ref count by 1)
// this should be used together with load_item, as the spill lock counter is decreased too.
void ValueGen::rfree(Item* item) {
  ra()->decr_spill_lock(item->get_register());
  if (!item->is_cached()) ra()->free_reg(item->get_register());
  // for roots, make sure that their items are not holding on a register anymore (otherwise we may have problems when spilling expression stack values)
  if (item->value() != NULL && item->value()->is_root()) {
    Item* root_item = item->value()->item();
    if (root_item->is_register() && ra()->is_free_reg(root_item->get_register())) {
      // no more use of this item -> invalidate it
      root_item->set_no_result();
    }
  }
}


// release register described by hint (decrease ref count by 1)
// this is not used together with load item
void ValueGen::raw_rfree(Item* item) {
  assert(!item->is_cached(), "cannot release cached item");
  ra()->free_reg(item->get_register());
}


bool ValueGen::is_free_rinfo(RInfo reg) {
  return ra()->is_free_reg(reg);
}


// reg must be free; locks with refcount of 1
void ValueGen::lock_rinfo(Value instr, RInfo reg) {
  assert(ra()->is_free_reg(reg), "");
  ra()->lock_reg(instr, reg, 1);
}


RInfo ValueGen::get_free_rinfo(ValueType* type) {
  if (!ra()->has_free_reg(type)) {  spill_one(type);  }
  RInfo reg = ra()->get_free_reg(type);
  return reg;
}


RInfo ValueGen::get_free_rinfo(c1_RegMask mask) {
  while (!ra()->has_free_reg(mask)) {
    RInfo r = mask.get_first_reg();
    spill_register(r);
  }
  RInfo reg = ra()->get_free_reg_from_mask(mask);
  return reg;
}


// locks a free register; sets refcount to 1
RInfo ValueGen::lock_free_rinfo(Value instr, ValueType* type) {
  if (!ra()->has_free_reg(type)) {
    spill_one(type);
  }
  //return ra()->get_lock_reg(instr, type);
  RInfo reg = ra()->get_free_reg(type);
  lock_rinfo(instr, reg);
  return reg;
}


RInfo ValueGen::lock_free_rinfo(Value instr, c1_RegMask mask) {
  while (!ra()->has_free_reg(mask)) {
    Value val = ra()->get_smallest_value_to_spill(mask);
    assert(val != NULL, "could not spill");
    spill_value(val);
  }
  RInfo reg = ra()->get_free_reg_from_mask(mask);
  lock_rinfo(instr, reg);
  return reg;
}


//----------------------------------------------------------------


void  ValueGen::set_no_result(Value x) { 
  assert(!x->has_uses(), "cannot have a use"); 
  result()->set_no_result();  
}


//---------------------------------------------------------------------

// Call this after finished generating code for root
// If the item of root is a constant or spilled, then leave it as it is
// If the item of root is a stack value  then preload it so that the following 
// bytecode sequence is executed correctly:
//          load local 0
//          load local 0
//          inc 1
//          store local 0
//          ....            <- on stack is the "old" local 0


// Call this after finished generating code for root
void ValueGen::finish_root(Value instr) {
  if (_result->has_result()) {
    if (instr->has_uses()) {
      // eventually, if use_count > 1, we may not need to load instruction into a register
      if (_result->is_constant() || (_result->is_stack() && _result->is_spilled())) {
        // leave it there
        assert(_result->is_root_item(), "");
        instr->set_item(_result);
      } else if (_result->is_stack() || _result->is_cached()) {
        bool copy_to_reg = true;
        if (_result->is_cached() && instr->as_AccessLocal() != NULL) {
          // a cached local that is not pinned by a store (see example above) 
          // must not be copied to a register
          AccessLocal* al = instr->as_AccessLocal();
          if (!al->is_pinned_by_store()) {
            copy_to_reg = false;
          }
        }
        if (copy_to_reg) {
          // copy the _result item to a register
          RInfo reg;
          // lock a register and move stack or cached element into a registers
          reg = rlock(instr);
          if (gen_code()) {
            code_emit()->item_2_reg(reg, _result);
          }
          // release old spill slot or register
          if (_result->is_spilled()) {
            int rc1 = ra()->get_ref_count_at(_result->get_spilled_index());
            assert( rc1 == instr->use_count(), "sanity check");
            ra()->free_complete_spill(_result->get_spilled_index(), instr->type());
          } else if (_result->is_register()) {
            if (!_result->is_cached()) {
              // do not need to release cached register
              int rc1 = ra()->get_register_rc(_result->get_register());
              assert(rc1 == instr->use_count(), "sanity check");
              for (int i = 0; i < rc1; i++) raw_rfree(_result);
            }
          }
          _result->set_register(reg);
          ra()->set_reg(reg, instr->use_count(), instr);
        } else { // dont copy to reg
          assert(_result->is_cached(), "implemented only for cached locals");
          assert(_result->is_root_item(), "");
          instr->set_item(_result);
        }
      } else {
        // result is register
        RInfo reg = _result->get_register();
        ra()->set_reg(reg, instr->use_count(), instr);
      }
      assert(_result->is_root_item(), "");
      instr->set_item(_result);
#ifdef ASSERT
      if (_result->is_register()) {
        assert(!ra()->is_spill_locked(_result->get_register()), "must be spillable after done visiting root");
      }
#endif
    } else {
      // Use_count is 0, do not do anything
    }
  }
  check_result_usage(instr);
}


bool ValueGen::must_copy_register(Item* item) {
  return item->destroys_register() && (item->is_cached() || ra()->get_register_rc(item->get_register()) > 1);
}


// if the item is spilled or is register, check that the information
// entered in RegAlloc cooresponds
// We search for the value entered in the register allocator and spiller.
// Several Items may point to the same Value. A Value holds only one
// Item; Non-root items are referenced only by one Item; root Values
// can be referenced by several Items
void ValueGen::check_item(Item* item) {
  assert(item->is_valid(), "invalid item");
  if (item->is_spilled()) {
    int    ix = item->get_spilled_index();
    Value val = ra()->value_spilled_at(ix);
    assert(item->value() == val, "wrong item state");
  } else if (item->is_register()) {
    if (!item->is_cached()) {
      // This test is not valid for cached items
      RInfo reg = item->get_register();
      Value val = ra()->get_value_for_rinfo(reg);
      assert(item->value() == val, "wrong item state");
    }
  }
}


// Mark items that do not need to be loaded (we must call either load_item or dont_load_item)
void ValueGen::dont_load_item(Item* item) {
  dont_load_item_nocheck(item);
  debug_only(check_item_load(item);)
  if (item->is_register() && item->destroys_register()) {
    check_float_register(item);
#ifdef ASSERT
    if (!is_free_rinfo(item->get_register())) {
      assert(ra()->get_register_rc(item->get_register()) <= 1, "we should copy it instead of not loading it");
    }
#endif
  }
}


void ValueGen::dont_load_item_nocheck(Item* item) {
  item->update();
  set_maynot_spill(item);
}


// Note about item loading( load_xxxx..):
//    - at the start, the item being modified must be sycnhronized with the item that
//      is maintained by the instruction; 
//    - The item is then being modified and may not correspond to the item being pointed
//      to by instruction

// loads item in a register
void ValueGen::load_item(Item* item) {
  item->update();
  set_maynot_spill(item);
  debug_only(check_item_load(item);)
  if (item->is_register() && !must_copy_register(item)) {
    check_float_register(item);
  } else {
    // We come here when either item is not register or if we have to 
    // copy from a register to a different one
    if (fpu_fanout_handled(item)) {
      set_maynot_spill(item);
    } else {
      RInfo reg = lock_free_rinfo(item->value(), item->type());
      if (gen_code()) {
        code_emit()->item_2_reg(reg, item);
      }
      item_free(item);
      item->set_register(reg);
      set_maynot_spill(item);
    }
  } 
}


// Used for static field access to patch the operation
//    movl reg, klass
void ValueGen::load_item_patching(Item* item) {
  item->update();
  assert(item->is_constant() &&  item->type()->is_object(), "must be an object constant");
  set_maynot_spill(item);
  const RInfoCollection* oop_regs = NULL;
  if (gen_code()) {
    // must do this before locking the destination register as an oop register
    oop_regs = ra()->oops_in_registers();
  }
  RInfo reg = lock_free_rinfo(item->value(), item->type());
  if (gen_code()) {
    CodeEmitInfo info(item->value()->bci(), ra()->oops_in_spill(), NULL, oop_regs);
    code_emit()->item_2_reg_with_patching(reg, item, &info);
  }
  item_free(item);
  item->set_register(reg);
  set_maynot_spill(item);
}


// loads item using hint if possible
void ValueGen::load_item_hint(Item* item, const Item* hint) {
  item->update();
  set_maynot_spill(item);
  debug_only(check_item_load(item);)
  if (item->is_register() && !must_copy_register(item)) {
    check_float_register(item);
    // Note: we may not have the value loaded in the correct
    //       register, but that is OK and simpler to leave it there instead
    //       of moving it into hint
  } else {
    if (fpu_fanout_handled(item)) {
      set_maynot_spill(item);
    } else {
      bool is_cached = false;
      RInfo reg;
      if (hint->has_hint() && is_free_rinfo(hint->get_register())) {
        reg = hint->get_register();
        if (hint->is_cached()) {
          is_cached = true;
        } else {
          lock_rinfo(item->value(), reg);
        }
      } else {
        reg = lock_free_rinfo(item->value(), item->type());
      }
      if (gen_code()) {
        code_emit()->item_2_reg(reg, item);
      }
      item_free(item);
      item->set_register(reg, is_cached);
      set_maynot_spill(item);
    }
  }
}

// This is the prolog for load_item_force: if registers in item are
// overlapping the 'reg", and they are not the same, then the item must be spilled
void ValueGen::check_for_spill(Item* item, RInfo reg) {
  if (item->is_register()) {
    RInfo item_reg = item->get_register();
    if (!reg.is_same(item_reg) && reg.overlaps(item_reg)) {
      spill_item(item);
    }
  }
}


// Force loading of item into the specified register (used e.g. for return)
// Note: refrain from using this way of loading the items if possible,
//       as it is very risky (needs various tests and/or spilling before calling it)
//       -> item may not contain a register that is being locked
void ValueGen::load_item_force(Item* item, RInfo force_reg) {
  check_for_spill(item, force_reg);
  item->update();
  set_maynot_spill(item);
  debug_only(check_item_load(item);)
  if (item->is_register() && force_reg.is_same(item->get_register()) && !must_copy_register(item)) {
    // everything is fine
  } else {
    // item register and force reg may be the same if item->value() use_count is > 1
    if (item->is_register() && item->get_register().is_same(force_reg)) {
      // We cannot correctly handle this case: be sure to spill
      // item before calling this function
      ShouldNotReachHere();
    } else {
      // I'm trying to prove that it just happens to be the case that the
      // instruction on behalf of which we are locking happens to have a
      // use_count of 1. If this turns out to be true, then we can replace 
      // the constant 1 in lock_spill_rinfo below with
      // "item->value()->use_count()" as in:
      //
      //       lock_spill_rinfo(item->value(), force_reg, item->value()->use_count());
      //
      // Then we can simplify a lot more the code generator interface.
      //assert(item->value() != NULL && item->value()->use_count() == 1, "This is a temporary assert, look at comments in source code");
      lock_spill_rinfo(item->value(), force_reg, 1);
    }
    item->update();
    if (gen_code()) {
      code_emit()->item_2_reg(force_reg, item);
    }
    item_free(item);
    item->set_register(force_reg);
    set_maynot_spill(item);
  }
}


void ValueGen::block_epilog(BlockBegin* block) {
  // nothing to do for now.
}



void ValueGen::align_block(BlockBegin* block) {
  if (gen_code()) {
    if (block->is_set(BlockBegin::backward_branch_target_flag)) {
      code_emit()->align_backward_branch(); 
    }
  }
}



void ValueGen::block_prolog(BlockBegin* block) {
  // setup phi's
  assert(block->stack() != NULL, "block state must exist");
  assert(ra()->are_all_free(), "");  
  assert(!gen_code() || code_emit()->is_empty_fpu_stack(), "still some values on FPU stack");

  // lock cached registers
  BlockItem* block_item = block->block_item();
  c1_RegMask lockout = (block_item != NULL ? block_item->lockout() : c1_RegMask::empty_set());
  ra()->set_cpu_lockout(lockout);

  if (block->stack()->stack_is_empty()) { 
    return; 
  }

  GrowableArray<Value> phi_values;
  GrowableArray<int>   spill_ixs;
  Value tos_val = compute_phi_arrays(block->stack(), &phi_values, &spill_ixs);
  for (int i = 0; i < phi_values.length(); i++) {
    Phi* phi = phi_values.at(i)->as_Phi();
    assert(phi != NULL, "value is not phi");
    int spill_ix = spill_ixs.at(i);
    if (phi->has_uses()) {
      assert(phi->item() == NULL, "item already set");
      RootItem* item = new RootItem(phi);
      phi->set_item(item);
      item->set_spill_ix(spill_ix);
      ra()->lock_spill(phi, spill_ix, phi->use_count());
      assert(phi->item()->is_valid(), "invalid item");
      assert(ra()->check_spilled(item), "wrong spilled state");
    } else {
      // must set and release in order to notify the register allocator of 
      // correct number of spills
      ra()->lock_spill(phi, spill_ix, 1);
      ra()->free_spill(spill_ix, phi->type());
      phi->set_item(new RootItem(phi));  // dummy
    }
  }
  // set up tos value
  assert(tos_val != NULL, "must generate tos_value");
  assert(tos_val->item() == NULL, "tos item already set");
  if (tos_val->has_uses()) {
    RInfo phi_reg = result_register_for(tos_val->type());
    RootItem* item = new RootItem(tos_val);
    item->set_register(phi_reg);
    tos_val->set_item(item);
    assert(tos_val->item()->is_valid(), "invalid_item");
    ra()->lock_reg(tos_val, phi_reg, tos_val->use_count());
    if (tos_val->type()->is_float() || tos_val->type()->is_double()) {
      if (gen_code()) {
        code_emit()->set_fpu_result(tos_val->item()->get_register());
      }
    }
  } else {
    tos_val->set_item(new RootItem(tos_val));
    if (tos_val->type()->is_float_kind()) {
      if (gen_code()) {
        code_emit()->fpop(); // restore FPU stack
      }
    }
  }
  if (PrintC1RegAlloc) { 
    tty->print("Block prolog "); 
    if (gen_code()) code_emit()->print();
    ra()->print(); 
    tty->cr();  
  }
}


void ValueGen::bind_label(Label* label) {
  if (gen_code()) {
    code_emit()->bind_label(label);
  }
}


void ValueGen::exception_handler_start(int bci) {
  if (gen_code()) {
    CodeEmitInfo info(bci, ra()->oops_in_spill(), NULL);
    code_emit()->add_debug_info_here(&info);
    // the nop at start of exception handler prevents that we emit two pc/bci 
    // mappings with identical pc's but varying bci's
    code_emit()->handler_entry();
    if(!method()->is_static() && UseCachedReceiver) {
      // restore cached receiver, just in case that we are caching it
      code_emit()->restore_cached_receiver();
    }
  }
}


bool ValueGen::cached_cpu_local(int local_index,  RInfo& r) {
  BlockItem* bi = block()->block_item();
  if (bi != NULL) {
    r = bi->map_local(local_index);
    if (!r.is_illegal()) {
      return true;
    }
  }
  return false;
}


// Returns false if the item is local 0 in a virtual method (i.e., receiver)
// and there is no store to local 0
bool ValueGen::item_may_be_null(Item* item) {
  // Experimental code
  if (!method()->is_static() && !compilation()->scan_result()->has_store_0() ) {
    LoadLocal* load = item->value()->as_LoadLocal();
    if (load != NULL && load->index() == 0) {
      return false;
    }
  }
  return true;
}










