// @(#)c1_CodeStubs.hpp	1.27 00/04/11 10:07:27
//
// Copyright 1991-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


class CodeEmit;
class CodeEmitInfo;

// CodeStubs are little 'out-of-line' pieces of code that
// usually handle slow cases of operations. All code stubs
// are collected and code is emitted at the end of the
// nmethod.

class CodeStub: public ResourceObj {
 protected:
  Label _entry;                                  // label at the stub entry point
  Label _continuation;                           // label where stub continues, if any

 public:
  // code generation
  void assert_no_unbound_labels()                { assert(!_entry.is_unbound() && !_continuation.is_unbound(), "unbound label"); }
  virtual void emit_code(CodeEmitter* ce)        = 0;

  // label access
  Label* entry()                                 { return &_entry; }
  Label* continuation()                          { return &_continuation; }
};


class StaticCallStub: public CodeStub {
 private:
  address                          _call_pc;
  const GrowableArray<BasicType>*  _sig_types;

 public:
  StaticCallStub(address call_pc, const GrowableArray<BasicType>* sig_types):
    _call_pc(call_pc), _sig_types(sig_types) {}
  virtual void emit_code(CodeEmitter* ce);
  const GrowableArray<BasicType>* sig_types() const { return _sig_types; }
};


class RangeCheckStub: public CodeStub {
 private:
  int _pc_offset;

 public:
  RangeCheckStub(int pc_offset): _pc_offset(pc_offset) {}
  virtual void emit_code(CodeEmitter* ce);
};


class DivByZeroStub: public CodeStub {
 private:
  int _pc_offset;

 public:
  DivByZeroStub(int pc_offset): _pc_offset(pc_offset) {}
  virtual void emit_code(CodeEmitter* ce);
};


// Needed only for explicit null exception checks
class NullExceptionStub: public CodeStub {
 private:
  int _pc_offset;

 public:
  NullExceptionStub(int pc_offset): _pc_offset(pc_offset) {}
  virtual void emit_code(CodeEmitter* ce);
};


class NewInstanceStub: public CodeStub {
 private:
  Register       _klass;
  CodeEmitInfo*  _info;
  int            _esp_offset;

 public:
  NewInstanceStub(Register klass, CodeEmitInfo* info, int esp_offset = 0);
  virtual void emit_code(CodeEmitter* ce);
};


class NewTypeArrayStub: public CodeStub {
 private:
  BasicType     _elt_type;
  Register      _length;
  Register      _result;
  CodeEmitInfo* _info;
  int           _esp_offset;

 public:
  NewTypeArrayStub(BasicType elt_type, Register length, Register result, CodeEmitInfo* info, int esp_offset = 0);
  virtual void emit_code(CodeEmitter* ce);
};


class NewObjectArrayStub: public CodeStub {
 private:
  Register       _klass;
  Register       _length;
  Register       _result;
  CodeEmitInfo*  _info;
  int            _esp_offset;

 public:
  NewObjectArrayStub(Register klass, Register length, Register result, CodeEmitInfo* info, int esp_offset = 0);
  virtual void emit_code(CodeEmitter* ce);
};


class MonitorAccessStub: public CodeStub {
 protected:
  Register _obj_reg;
  Register _lock_reg;

 public:
  MonitorAccessStub(Register obj_reg, Register lock_reg) {
    _obj_reg  = obj_reg;
    _lock_reg = lock_reg;
  }
};


class MonitorEnterStub: public MonitorAccessStub {
 private:
  CodeEmitInfo* _info;
  int           _esp_offset;                             // esp offset when the call/jmp to stub was made

 public:
  MonitorEnterStub(Register obj_reg, Register lock_reg, CodeEmitInfo* info, int esp_offset = 0);
  virtual void emit_code(CodeEmitter* ce);
};


class MonitorExitStub: public MonitorAccessStub {
 private:
  bool _compute_lock;
  Address _lock_addr;

 public:
  MonitorExitStub(Register obj_reg, Register lock_reg, bool compute_lock, Address lock_addr)
  : MonitorAccessStub(obj_reg, lock_reg)
  { _compute_lock = compute_lock;
    _lock_addr    = lock_addr;
  }
  virtual void emit_code(CodeEmitter* ce);
};


class PatchingStub: public CodeStub {
 public:
  enum PatchID {
    access_field_id,
    load_klass_id
  };
 private:
  int           _id;
  int           _pc_start;
  int           _esp_offset;
  int           _bytes_to_copy;
  Register      _obj_reg; // must be preserved (may be noreg)
  Register      _src_reg; // must be preserved (may be noreg)
  CodeEmitInfo* _info;
  // some oop map for the registers too!
 public:
  PatchingStub(int start, Register obj_reg, Register src_reg, int bytes_to_copy, CodeEmitInfo* info,  PatchID id,  int esp_offset = 0): 
       _pc_start(start), _esp_offset(esp_offset), _bytes_to_copy(bytes_to_copy), _info(info), _id(id), _obj_reg(obj_reg), _src_reg(src_reg)  {}

  virtual void emit_code(CodeEmitter* ce);
};


class CheckCastStub: public CodeStub {
 private:
  Register  _k;
  CodeEmitInfo* _info;
  int _esp_offset;
  Register _dst;
 public:
  CheckCastStub(Register k, CodeEmitInfo* info, Register dst, int esp_offset = 0);
  virtual void emit_code(CodeEmitter* ce);
};



class InstanceOfStub: public CodeStub {
 private:
  Register _k;
  CodeEmitInfo* _info;
  int _esp_offset;
  Register _obj_klass;
  Register _result;
 public:
  InstanceOfStub(Register obj_klass, Register result, Register k, CodeEmitInfo* info, int esp_offset = 0);
  virtual void emit_code(CodeEmitter* ce);
};


class ArrayStoreCheckStub: public CodeStub {
 private:
  Register _array;
  Register _value;
  Register _index;
  int      _index_const;
  CodeEmitInfo* _info;
  int _esp_offset;
 public:
  ArrayStoreCheckStub(Register array, Register value, Register index, int index_const, CodeEmitInfo* info, int esp_offset = 0);
  virtual void emit_code(CodeEmitter* emit);
};


class ConvertStub: public CodeStub {
 private:
  Register _result;
  address  _runtime;
 public:
  ConvertStub(address runtime, Register result): _runtime(runtime), _result(result) {}
  virtual void emit_code(CodeEmitter* ce);
};


class ArrayCopyDesc: public ResourceObj {
 private:
  Register _src;
  Register _dst;
  Register _src_pos_reg;
  Register _dst_pos_reg;
  Register _length_reg;
  Register _tmp;
  int _src_pos_const;
  int _dst_pos_const;
  int _length_const;

  friend class ArrayCopyStub;
 public:
  ArrayCopyDesc(): 
       _src(noreg), _dst(noreg), _src_pos_reg(noreg), _dst_pos_reg(noreg), _length_reg(noreg),
       _src_pos_const(-1), _dst_pos_const(-1), _length_const(-1), _tmp(noreg) {}

  void set_src(Register r) { _src = r; }
  void set_dst(Register r) { _dst = r; }
  void set_src_pos_reg(Register r) { _src_pos_reg = r; }
  void set_dst_pos_reg(Register r) { _dst_pos_reg = r; }
  void set_length_reg(Register r)  { _length_reg = r;  }
  void set_tmp_reg(Register r)     { _tmp = r;  }

  void set_src_pos_const(int c) { _src_pos_const = c; }
  void set_dst_pos_const(int c) { _dst_pos_const = c; }
  void set_length_const(int c)  { _length_const = c;  }

};


class ArrayCopyStub: public CodeStub {
 private:
  const ArrayCopyDesc* _desc;
  CodeEmitInfo*  _info;
  int _esp_offset;
  Label _noCheckEntry;

  Register src() const { return _desc->_src; }
  Register src_pos_reg() const { return _desc->_src_pos_reg; }
  Register dst() const { return _desc->_dst; }
  Register dst_pos_reg() const { return _desc->_dst_pos_reg; }
  Register length_reg() const { return _desc->_length_reg; }
  Register tmp() const { return _desc->_tmp; }
  int src_pos_const() const { return _desc->_src_pos_const; }
  int dst_pos_const() const { return _desc->_dst_pos_const; }
  int length_const() const  { return _desc->_length_const;   }
  void push_parameter (CodeEmitter* ce, int param_ix, Register r);
  void push_params    (CodeEmitter* ce);
 public:
  ArrayCopyStub(ArrayCopyDesc* desc, CodeEmitInfo* info, int esp_offset = 0);

  virtual void emit_code(CodeEmitter* ce);
  Label* noCheckEntry()  { return &_noCheckEntry; }

};


// slow cases for java.lang.Math.sin and java.lang.Math.cos
// parameter on FPU TOS and result on FPU TOS
class MathStub: public CodeStub {
 private:
  address  _rt_addr;
 public:
  MathStub(address rt_addr) : _rt_addr(rt_addr) {}

  virtual void emit_code(CodeEmitter* ce);
};


class ICacheStub: public CodeStub {
 public:
  ICacheStub() {}
  virtual void emit_code(CodeEmitter* ce);
};

