// @(#)c1_Compilation.cpp	1.36 00/04/26 12:03:27
// 
// Copyright 1993-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_c1_Compilation.cpp.incl"


// "Inherited" functionality - should go into AbstractCompilation

GrowableArray<LinePair*>* Compilation::get_method_line_numbers(ciMethod *m) {
  const int n = _env->get_method_number_of_lines(m);
  GrowableArray<LinePair*>* pairs = new GrowableArray<LinePair*>(n, n, NULL);
  for (int i = 0; i < n; i++) pairs->at_put(i, new LinePair);
  _env->fill_method_line_numbers(m, pairs);
  return pairs;
}


// Implementation of Compilation


void Compilation::initialize() {
  _info = new GrowableArray<JVMCI_DebugInfo*>();
}


void Compilation::initialize_oop_maps() {
  _oop_map = method()->all_oop_maps();
  if (PrintOopMap) {
    if (_oop_map == NULL) tty->print_cr("NULL oop_map");
    else _oop_map->print();
  }
  if (_oop_map != NULL) {
    _oop_map->init_oops_cache(method()->code_size());
  }
}


void Compilation::build_hir() {
  assert(!method()->is_native(), "should not reach here");
  if (bailed_out()) return;

  // setup ir
  _hir = new IR(this, method(), osr_bci());
  if (!_hir->is_valid()) {
    bailout("invalid parsing");
    return;
  }
  if (PrintCFG || PrintCFG0) { tty->print_cr("CFG after parsing"); _hir->print(true); }
  if (PrintIR  || PrintIR0 ) { tty->print_cr("IR after parsing"); _hir->print(false); }

  if (UseC1Optimizations) {
    NEEDS_CLEANUP
    // optimization
    //NOT_PRODUCT(if (TimeCompiler) _optimizeIR_time.start();)
    _hir->optimize();
    //NOT_PRODUCT(if (TimeCompiler) _optimizeIR_time.stop();)

    // loop detection
    // note: after this, no block introduction/elimination must happen
    //       otherwise the loop information is going to be incorrect
    //NOT_PRODUCT(if (TimeCompiler) _loopRecognition_time.start();)
    _hir->compute_loops();
    //NOT_PRODUCT(if (TimeCompiler) _loopRecognition_time.stop();)
  }

  // compute block ordering for code generation
  _hir->compute_code();
  if (PrintCFG || PrintCFG2) { tty->print_cr("CFG before code generation"); _hir->code()->print(true); }
  if (PrintIR  || PrintIR2 ) { tty->print_cr("IR before code generation"); _hir->code()->print(false); }

  // accessor function printing
  if (PrintIR  || PrintIR2 ) {
    AccessorInfo* info = _hir->accessor_info();
    if (info != NULL) {
      info->print();
      tty->cr();
    }
  }
}


void Compilation::scan_method() {
  assert(!method()->is_native(), "should not reach here");
  if (bailed_out()) return;
  ScanBlocks scan(hir()->code()->collection());
  scan.scan(scan_result());
}


void Compilation::allocate_registers() {
  assert(!method()->is_native(), "should not reach here");
  if (bailed_out()) return;
  ValueGen::init_value_gen();
  RegAlloc reg_alloc;
  ValueGenInvariant vgi(method(), &reg_alloc, NULL);
  ValueGen visitor(this, &vgi, false); // no code for the moment
  CodeGenerator gen(&visitor, &vgi);
  hir()->code()->iterate_forward(&gen);
  // clear items
  hir()->code()->blocks_do(CodeGenerator::clear_instruction_items);
  _max_spills = gen.max_spills();
  // _used_registers and _free_registers are used by global local caching
  _used_registers = reg_alloc.used_registers(); // these are unavailable for global local caching
  _free_registers = reg_alloc.free_registers(); // these are   available for global local caching

  if (CacheReceiver) {
    if (!method()->is_static()) {
      if (!method()->has_loops()) {
        LocalCaching local_caching(hir());
        local_caching.cache_receiver(&reg_alloc);
      }
    }
  }
}


const CachedLocals* Compilation::cache_locals() {
  const CachedLocals* global_cached_locals = NULL;
  assert(!method()->is_native(), "should not reach here");
  if (bailed_out()) return NULL;
  // On SPARC: at least L0 gets destroyed if exception handler exists
  // On win32: no callee save registers
  const bool can_cache = !method()->has_exception_handlers();
  if (can_cache) {
    // CacheLocalsInLoop and CacheReceiver has higher priority than CacheLocalsGlobally
    if (CacheLocalsInLoops) {
      LocalCaching local_caching(hir());
      local_caching.cache_loop_locals();
    } else if  (!CacheReceiver && CacheLocalsGlobally) {
      assert(!CacheReceiver && !CacheLocalsInLoops, "they must be off if CacheLocalsGlobally is true");
      LocalCaching local_caching(hir());
      global_cached_locals = local_caching.cache_locals_globally();
    }
  }
  return global_cached_locals;
}


void Compilation::emit_code_prolog(FrameMap* map) {
  map->set_size_arguments(method()->arg_size());
  int size_locals = max(method()->arg_size(), method()->max_locals());
  assert(sizeof(BasicLock)/wordSize == 1, "cannot handle this BasicLock");
  if (method()->is_native()) {
    // the entry for method_holder mirror is marked as spill
    // if (get_cmethod()->is_static()) size_locals++; // need room for mirror class
    map->set_size_locals(size_locals);
    int nofm = method()->is_synchronized() ? 1 : 0;
    map->set_size_monitors(nofm * (sizeof(BasicLock)/wordSize));
  } else {
    map->set_size_locals(size_locals);
    if (hir() == NULL) {
      // optimized library call: no monitors
      map->set_size_monitors(0);
    } else {
      map->set_size_monitors(hir()->number_of_locks() * (sizeof(BasicLock)/wordSize)); 
    }
  }
#ifdef SPARC
  map->set_max_stack_size(method()->max_stack());
#else
  map->set_fixed_framesize();
#endif
  _frame_map = map;
}


void Compilation::emit_code_epilog(CodeEmitter* emit, int max_spills, RInfo cached_receiver) {
  if (bailed_out()) return;
  // generate code for osr, if any
  if (UseOnStackReplacement && is_osr_compile()) {
    BlockBegin* osr_entry = hir()->osr_entry();
    assert(osr_entry != NULL && osr_entry->is_set(BlockBegin::osr_entry_flag), "osr entry block must exist");
    emit->emit_osr_entry(max_spills, osr_entry->end()->state()->locks_size(), osr_entry->label(), osr_bci());
  }
  // generate code for slow cases
  emit->emit_slow_case_stubs();
  // if the last instruction is a call (typically to do a throw which
  // is coming at the end after block reordering) the return address
  // must still point into the code area in order to avoid assertion
  // failures when searching for the corresponding bci => add a nop
  // (was bug 5/14/1999 - gri)
  emit->add_nop();
  // generate code for exception handler
  int offset = emit->emit_exception_handler(max_spills, cached_receiver);
  code()->set_exception_offset(offset);
  emit->emit_call_stubs();
  // done
  masm()->flush();
}


class CollectConstants: public BlockClosure {
 private:
  ConstantTable* _table;

 public:
  CollectConstants(ConstantTable* table): _table(table) {}

  void block_do(BlockBegin* b) {
    for (Instruction* n = b; n != NULL; n = n->next()) {
      Constant* c = n->as_Constant();
      if (c != NULL) {
        if (c->type()->is_double()) {
          _table->append_double(c->type()->as_DoubleConstant()->value());
        } else if (c->type()->is_float() ) {
          _table->append_float(c->type()->as_FloatConstant()->value());
        }
      }
    }
  }
};


GrowableArray<int>* Compilation::get_init_vars() {
  GrowableArray<int>* vars = new GrowableArray<int>();
  if (oop_map() != NULL) {
    for (int i = 0; i < oop_map()->nof_locals_to_initialize(); i++) {
      vars->append(oop_map()->local_to_initialize(i));
    }
  }
  return vars;
}


int Compilation::emit_code_body(JVMCI_OffsetDesc* offsets, const CachedLocals* cached_locals) {
  assert(!method()->is_native(), "should not reach here");
  if (bailed_out()) return no_frame_size;
  FrameMap map(max_spills(), cached_locals);
  emit_code_prolog(&map);

  // collect float and double constants

  ConstantTable ct;
  CollectConstants cc(&ct);
  if (UseFPConstTables) hir()->code()->iterate_forward(&cc);
  
  // emit code

  CodeEmitter emit(this, &ct, offsets);
  emit.emit_constants();
 
  GrowableArray<int>* init_vars = get_init_vars();
  emit.method_entry(init_vars, max_spills());

  RegAlloc reg_alloc;
  ValueGenInvariant vgi(method(), &reg_alloc, &emit);
  ValueGen visitor(this, &vgi, true); 
  CodeGenerator gen(&visitor, &vgi);
  hir()->code()->iterate_forward(&gen);

  RInfo cached_receiver =
    (cached_locals != NULL && !method()->is_static()) ? cached_locals->get_cache_reg(0) : FrameMap::_noRInfo;

  emit_code_epilog(&emit, max_spills(), cached_receiver);
  // return
  assert(max_spills() >= gen.max_spills(), "the frame is too small");
  if (emit.must_bailout()) {
    bailout("Code emission bailout");
  }
  if (jvmpi_event_method_enabled()) {
    _jvmpi_lines = emit.get_jvmpi_lineno_table();
  }
  return emit.frame_size();
}


int Compilation::emit_accessor_body(JVMCI_OffsetDesc* offsets, AccessorInfo* info) {
  if (bailed_out()) return no_frame_size;
  assert(!method()->is_static() && !method()->is_native() && !method()->is_synchronized() &&
         info != NULL && info->type() == T_FLOAT, "cannot handle other cases");
  CodeEmitter emit(this, NULL, offsets);
  // receiver is never NULL, no trap is possible in accessor body
  emit.emit_accessor_code(info->offset(), info->type());
  emit.emit_call_stubs(); // emits interpreter entru point, for example
  return no_frame_size;
}


int Compilation::compile_java_method(JVMCI_OffsetDesc& offsets) {
  assert(!method()->is_native(), "should not reach here");
  if (bailed_out()) return no_frame_size;

  initialize_oop_maps();

  //NOT_PRODUCT(if (TimeCompiler) {  _setup_time.stop(); _buildIR_time.start();  })

  build_hir();
  scan_method();

  if (OptimizeSimpleGetters && !jvmpi_event_method_enabled() && !bailed_out() &&
      !method()->is_static() && !method()->is_synchronized()) {
    // Only virtual getters currently supported
    AccessorInfo* info = hir()->accessor_info();
    if (info != NULL && info->type() == T_FLOAT) {
      // try for floats first
     int frame_size = emit_accessor_body(&offsets, info);
     return frame_size;
    }
  }

  //NOT_PRODUCT(if (TimeCompiler) { _buildIR_time.stop(); _prepass_time.start();  })

  allocate_registers();
  const CachedLocals* global_cached_locals = NULL;
  if (CacheLocalsGlobally || CacheLocalsInLoops) {
    global_cached_locals = cache_locals();
  }

  //NOT_PRODUCT(if (TimeCompiler) { _prepass_time.stop(); _codeemit_time.start();  })

  int frame_size = emit_code_body(&offsets, global_cached_locals);

  //NOT_PRODUCT(if (TimeCompiler) { _codeemit_time.stop();_codeinstall_time.start();  })
  return frame_size;
}


void Compilation::emit_code_for_native(address native_entry, JVMCI_OffsetDesc* offsets) {
  assert(method()->is_native(), "should not reach here");
  assert(!is_osr_compile(), "no OSR for natives");
  if (bailed_out()) return;

  const int nof_spills = method()->is_static() ? 1 : 0; // one spill needed for holding method holder mirror
  FrameMap map(nof_spills, NULL);
  emit_code_prolog(&map);
  CodeEmitter emit(this, NULL, offsets);

  GrowableArray<int>* spilled = NULL;
  if (method()->is_static()) {
    spilled = new GrowableArray<int>();
    spilled->append(0); // stack slot for the mirror class
  }
  CodeEmitInfo info(0, spilled, NULL, NULL);

  emit.method_entry(NULL, nof_spills);
  emit.native_call(native_entry, &info);
  emit.native_method_exit(&info);
  emit_code_epilog(&emit, nof_spills);
}


void native_not_found_warning() {
  tty->print_cr("native was not found during compilation - check this code");
}


int Compilation::compile_native_method(JVMCI_OffsetDesc& offsets) {
  address entry = method()->native_entry();
  assert(entry != NULL, "native entry must have been precomputed");
  emit_code_for_native(entry, &offsets);
  return no_frame_size;
}


ciMethod::IntrinsicId Compilation::library_method_id() const {
  ciMethod::IntrinsicId id = method()->intrinsic_id();
  // optimizable library methods
  switch (id) {
    case ciMethod::_dsin     : // fall through
    case ciMethod::_dcos     : // fall through
    case ciMethod::_compareTo: // fall through
      return id;
  }
  // default
  return ciMethod::_none;
}


int Compilation::compile_library_method(JVMCI_OffsetDesc& offsets) {
  FrameMap map(0, NULL);
  emit_code_prolog(&map);
  CodeEmitter emit(this, NULL, &offsets);
  emit.method_entry(NULL, 0);
  ciMethod::IntrinsicId id = library_method_id();
  switch (id) {
    case methodOopDesc::_dsin     : // fall through
    case methodOopDesc::_dcos     : emit.triglib(id);      break;
    case methodOopDesc::_compareTo: emit.string_compare(); break;
    default                       : ShouldNotReachHere();
  };
  emit_code_epilog(&emit, 0);
  return no_frame_size;
}


void Compilation::install_code(JVMCI_OffsetDesc& offsets, int frame_size) {
  DebugInformationRecorder*  recorder = new DebugInformationRecorder(code()->oop_recorder());
  recorder->set_oopmaps(new OopMapSet());
  for (int i = 0; i < info()->length(); i++) {
    JVMCI_DebugInfo* elem = info()->at(i);
    recorder->add_safepoint(elem->pc_offset, elem->at_call, elem->oop_map);
    DebugToken *locvals = recorder->create_scope_values  (elem->locarray);
    DebugToken *expvals = recorder->create_scope_values  (elem->exparray);
    DebugToken *monvals = recorder->create_monitor_values(elem->monarray);
    recorder->describe_scope(method(), elem->bci, locvals, expvals, monvals); 
  }
  _env->register_method(
    method(),
    osr_bci(),
    offsets._iep_offset,
    offsets._ep_offset,
    offsets._vep_offset,
    offsets._code_offset,
    offsets._osr_offset,
    recorder,
    code(),
    frame_size,
    no_return_address_offset,
    recorder->_oopmaps,
    NULL, // no exception table
    NULL  // no implicit null checks table
  );
}


void Compilation::compile_method() {
  // setup compilation
  initialize();

  // compile method
  JVMCI_OffsetDesc offsets;
  int frame_size;
  if (OptimizeLibraryCalls && !is_osr_compile() && library_method_id() != ciMethod::_none) {
    // note: we cannot produce an osr method for libray calls
    //       because these methods are hand-coded and no osr
    //       entry point is provided!
    frame_size = compile_library_method(offsets);
  } else if (method()->is_native()) {
    frame_size = compile_native_method(offsets);
  } else {
    frame_size = compile_java_method(offsets);
  }

  // bailout if method couldn't be compiled
  // Note: make sure we mark the method as not compilable!
  if (bailed_out()) return;

  // install code
  install_code(offsets, frame_size);
}


Compilation::Compilation(ciEnv* env, ciMethod* method, int osr_bci, C1_MacroAssembler* masm)
: _env(env)
, _method(method)
, _osr_bci(osr_bci)
, _hir(NULL)
, _max_spills(-1)
, _used_registers()
, _free_registers()
, _oop_map(NULL)
, _frame_map(NULL)
, _masm(masm)
, _bailout_msg(NULL)
, _info(NULL)
, _jvmpi_lines(NULL)
, _jvmpi_event_compiled_method_enabled(jvmpi::is_event_enabled(JVMPI_EVENT_COMPILED_METHOD_LOAD))
, _jvmpi_event_method_entry_enabled   (jvmpi::is_event_enabled(JVMPI_EVENT_METHOD_ENTRY))
, _jvmpi_event_method_entry2_enabled  (jvmpi::is_event_enabled(JVMPI_EVENT_METHOD_ENTRY2))
, _jvmpi_event_method_exit_enabled    (jvmpi::is_event_enabled(JVMPI_EVENT_METHOD_EXIT))
{
  compile_method();
}


// This is the main method that creates an OopMap using the spilled oops, esp offset
// and pregenerated locals oop map information; it is called from the CodeEmit;
// The oop map generated here is passed to JVMCI_DebugInfo
OopMap* Compilation::create_oop_map(int bci, GrowableArray<int>* spilled_oops, int esp_offset) {
  int arg_size = method()->arg_size();
  int frame_size = frame_map()->om_framesize(esp_offset);
  OopMap* map = frame_map()->new_oopmap(esp_offset);
  const CachedLocals* cached_locals = frame_map()->cached_locals();
  // spilled oops
  if (spilled_oops != NULL) {
    for (int i = spilled_oops->length(); i-- > 0;) {
      int spill_ix = spilled_oops->at(i);
      OptoReg::Name rn = frame_map()->spill_regname(spill_ix, esp_offset);
      map->set_oop(rn, frame_size, arg_size);
    }
  }

  if (oop_map() != NULL) {
    // find oop map index for this bci
    GrowableArray<int>* oops = oop_map()->get_oops_for_bci(bci);
    // This assert may fail for exception handlers that are not reachable
    // assert(oops != NULL, "No oop map for this bci");
    if (oops != NULL) {
      for (int i = oops->length(); i-- > 0;) {
        int local_ix = oops->at(i);
        OptoReg::Name rn;
	RInfo cache_reg;
	if (cached_locals != NULL) {
	  cache_reg = cached_locals->get_cache_reg(local_ix);
	}
        if (!cache_reg.is_illegal()) {
	  assert (cached_locals->is_oop(local_ix), "inconsistent caching information");
          assert (cache_reg.is_word(), "only words can be cached");
          rn = frame_map()->register_regname (cache_reg, esp_offset);
        } else {
          rn = frame_map()->local_regname (local_ix, esp_offset);
        }
        map->set_oop(rn, frame_size, arg_size);
      }
    }
  }
  return map;
}


void Compilation::enter_in_oop_map(OopMap* map, ArgumentLocation loc, int local_ix, int esp_offset) {
  const int arg_size = method()->arg_size();
  const int frame_size = frame_map()->om_framesize(esp_offset);
  OptoReg::Name rn;
  if (loc.is_register_arg()) {
    rn = frame_map()->register_regname (loc.incoming_reg_location(), esp_offset);
  } else {
    assert(loc.is_stack_arg(), "invalid mapping of args");
    rn = frame_map()->local_regname (local_ix, esp_offset);
  }
  map->set_oop(rn, frame_size, arg_size);
}


// Inside native stub all oops are on stack
OopMap* Compilation::create_oop_map_inside_natives(GrowableArray<int>* spilled_oops, int esp_offset) {
  assert(PassParametersInRegisters, "only for this flag");
  const int arg_size = method()->arg_size();
  const int frame_size = frame_map()->om_framesize(esp_offset);
  OopMap* map = frame_map()->new_oopmap(esp_offset);
  NEEDS_CLEANUP // factor this code out
  if (spilled_oops != NULL) {
    for (int i = spilled_oops->length() - 1; i >= 0; i--) {
      int spill_ix = spilled_oops->at(i);
      OptoReg::Name rn = frame_map()->spill_regname(spill_ix, esp_offset);
      map->set_oop(rn, frame_size, arg_size);
    }
  }
  ciSignature* sig = method()->signature();
  CallingConvention* args = FrameMap::calling_convention(method());
  // Note that sig has one entry for double-word, but args have two entries
  // Note: sig does NOT contain the receiver, args contain the receiver
  // the receiver first
  int slot_ix = 0;
  // signature oops
  if (!method()->is_static()) {
    OptoReg::Name rn = frame_map()->local_regname(0, esp_offset);
    map->set_oop(rn, frame_size, arg_size);
    slot_ix++;
  }
  for (int i = 0; i < sig->count(); i++) {
    if (sig->type_at(i) == T_OBJECT || sig->type_at(i) == T_ARRAY) {
      OptoReg::Name rn = frame_map()->local_regname(slot_ix, esp_offset);
      map->set_oop(rn, frame_size, arg_size);
    }
    slot_ix += sig->is_two_word_at(i) ? 2 : 1;
  }
  return map;
}


// Use calling convention to specify oop 
OopMap* Compilation::create_oop_map_for_own_signature(GrowableArray<int>* spilled_oops, int esp_offset) {
  const int arg_size = method()->arg_size();
  const int frame_size = frame_map()->om_framesize(esp_offset);
  OopMap* map = frame_map()->new_oopmap(esp_offset);
  // spilled oops
  if (spilled_oops != NULL) {
    for (int i = spilled_oops->length() - 1; i >= 0; i--) {
      int spill_ix = spilled_oops->at(i);
      OptoReg::Name rn = frame_map()->spill_regname(spill_ix, esp_offset);
      map->set_oop(rn, frame_size, arg_size);
    }
  }
  if (PassParametersInRegisters) {
    ciSignature* sig = method()->signature();
    CallingConvention* args = FrameMap::calling_convention(method());
    // Note that sig has one entry for double-word, but args have two entries
    // Note: sig does NOT contain the receiver, args contain the receiver
    int slot_ix = 0;
    // the receiver first
    if (!method()->is_static()) {
      enter_in_oop_map(map, args->arg_at(slot_ix), slot_ix, esp_offset);
      slot_ix = 1;
    }
    for (int n = 0; n < sig->count(); n++) {
      if (sig->is_oop_at(n)) {
        enter_in_oop_map(map, args->arg_at(slot_ix), slot_ix, esp_offset);
      }
      slot_ix += sig->is_two_word_at(n) ? 2 : 1;
    }
  } else {
    int slot_ix = 0;
    // signature oops
    if (!method()->is_static()) {
      OptoReg::Name rn = frame_map()->local_regname(0, esp_offset);
      map->set_oop(rn, frame_size, arg_size);
      slot_ix++;
    }
    ciSignature* sig = method()->signature();
    for (int i = 0; i < sig->count(); i++) {
      if (sig->type_at(i) == T_OBJECT || sig->type_at(i) == T_ARRAY) {
        OptoReg::Name rn = frame_map()->local_regname(slot_ix, esp_offset);
        map->set_oop(rn, frame_size, arg_size);
      }
      slot_ix += sig->is_two_word_at(i) ? 2 : 1;
    }
  }
  return map;
}


void Compilation::add_registers_to_oop_map(OopMap* map, const RInfoCollection* reg_oops, int esp_offset) {
  assert(reg_oops != NULL, "null check");
  int arg_size = method()->arg_size();
  int frame_size = frame_map()->om_framesize(esp_offset);
  for (int i = 0; i < reg_oops->number_of_registers(); i++) {
    RInfo rinfo = reg_oops->register_at(i);
    OptoReg::Name rn = frame_map()->register_regname(rinfo, esp_offset);
    map->set_oop(rn, frame_size, arg_size);
  }
}


// returns indices of locals that contain an address (in form of bci or pc)
GrowableArray<int>* Compilation::get_address_map(int bci) {
  ciAddressMap* map = method()->address_map_for_bci(bci);
  GrowableArray<int>* res = new GrowableArray<int>();
  // No map means no locals. Just return an empty map.
  if (map == NULL) return res;
  for (int i = 0; i < map->nof_locals(); i++) {
    if (map->is_address(i)) res->append(i);
  }
  return res;
}


void Compilation::bailout(const char* msg) {
  assert(msg != NULL, "bailout message must exist");
  if (!bailed_out()) {
    // keep first bailout message
    if (PrintBailouts) tty->print_cr("compilation bailout: %s", msg);
    _bailout_msg = msg;
  }
}


// old code to get native_entry
// this code os derived from MethodCompiler::compile
//
// NOTE: The native entry computation code should be factored
//       out and go into the CompilerBroker code!!!
/*
nmethod* C1Compiler::compile(methodHandle method, int osr_bci, TRAPS) {
  address native_entry = NULL;
  if (method->is_native()) {
    if (CompileC1Natives) {
      // check if native is available, and throw a trap here instead in compiler
      bool in_base_library; // set by lookup, we don't need it
      native_entry = NativeLookup::lookup(method, in_base_library, THREAD);
      if (HAS_PENDING_EXCEPTION) {
        if (PENDING_EXCEPTION ==  Universe::vm_exception()) CLEAR_PENDING_EXCEPTION;
        return NULL;
      }
    } else {
      return NULL;
    }
  }

  if (HAS_PENDING_EXCEPTION) {
    // bailing out, do not throw exception !! (it may be rethrown at proper
    // place during interpretation)
    if (PENDING_EXCEPTION ==  Universe::vm_exception()) CLEAR_PENDING_EXCEPTION; 
    return NULL;
  } else {
    C1::compile(NULL, NULL, osr_bci, native_entry);
    if (HAS_PENDING_EXCEPTION && PENDING_EXCEPTION == Universe::vm_exception()) {
      CLEAR_PENDING_EXCEPTION;
      return NULL;
    }
  }
  return method->code();  
}
*/
