/*
 * @(#)c1_FrameMap.hpp	1.12 00/04/23 15:11:08
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2



class ciMethod;
class CallingConvention;
class CachedLocals;


//--------------------------------------------------------
//               FrameMap
//--------------------------------------------------------

//  This class is responsible of mapping items (locals, monitors, spill
//  slots and registers to their frame location
//
//  The locals are specified by an index, same as in bytecodes (index is low endian).
//  The local index is 0.._size_locals-1.
//  The monitors are specified by a consecutive index, although each monitor entry
//  occupies two words. The monitor_index is 0.._size_monitors/2-1
//  The spill index is similar to local index; it is in range 0..(open)
//
//  The CPU registers are mapped using a fixed table; register with number 0
//  is the most used one.


class FrameMap VALUE_OBJ_CLASS_SPEC {
 public:
  enum {
    nof_cpu_regs = pd_nof_cpu_regs_frame_map,
    nof_fpu_regs = pd_nof_fpu_regs_frame_map
  };

# include "incls/_c1_FrameMap_pd.hpp.incl"  // platform dependent declarations

  friend class RInfo;
  friend class c1_AllocTable;

 public:
  enum {
    stack0 = pd_stack0
  };
  enum {
    monitor_slot_size = wordSize / sizeof(BasicLock)
  };

 private:
  static bool         _init_done;
  static Register     _cpu_regs [nof_cpu_regs];

  int                 _size_locals;
  int                 _size_monitors;
  int                 _size_arguments;
  int                 _size_spills;
  int                 _max_stack_size;
  const CachedLocals* _cached_locals;

  OptoReg::Name regname_for (int index, int esp_offset = 0) const;

  void check_local_index   (int local_index)   const { assert(local_index   >= 0 && local_index   < _size_locals, "bad index"); }
  void check_spill_index   (int spill_index)   const { assert(spill_index   >= 0 && spill_index   < _size_spills, "bad index"); }
  void check_monitor_index (int monitor_index) const { assert(monitor_index >= 0 &&
							      monitor_index < _size_monitors * monitor_slot_size, "bad index"); }

  static Register cpu_rnr2reg (int rnr);
  static int cpu_reg2rnr (Register reg);

 protected:
#ifndef PRODUCT
  static void cpu_range_check (int rnr)  { assert(0 <= rnr && rnr < nof_cpu_regs, "cpu register number is too big"); }
  static void fpu_range_check (int rnr)  { assert(0 <= rnr && rnr < nof_fpu_regs, "fpu register number is too big"); }
#endif

 public:
  FrameMap(int max_spills, const CachedLocals* cached_locals);

  static GrowableArray<BasicType>* signature_type_array_for(const ciMethod* method);
  static CallingConvention* calling_convention (bool is_static, const GrowableArray<BasicType>& signature, GrowableArray<int>* reg_args = NULL);
  static CallingConvention* calling_convention (const ciMethod* method, GrowableArray<int>* reg_args = NULL);

  int   om_framesize (int esp_offset = 0) const;
  int   size_locals                 () const { return _size_locals; }
  int   size_arguments              () const { return _size_arguments; }
  int   size_monitors               () const { return _size_monitors; }
  int   size_spills                 () const { return _size_spills; }
  int   max_stack_size              () const { return _max_stack_size; }
  const CachedLocals* cached_locals () const { return _cached_locals; }

  void set_size_locals    (int size) { assert(size >= 0, "check"); _size_locals    = size; }
  void set_size_arguments (int size) { assert(size >= 0, "check"); _size_arguments = size; }
  void set_size_monitors  (int size) { assert(size >= 0, "check"); _size_monitors  = size; }
  void set_max_stack_size (int size) { assert(size >= 0, "check"); _max_stack_size = size; }

  // mapping
  Address local_address   (int local_index, bool is_two_word, bool for_hi_word = false) const;
  Address spill_address   (int spill_index, bool is_two_word, bool for_hi_word = false) const;
  Address monitor_address (int monitor_index) const;

  OptoReg::Name local_regname    (int local_index, int esp_offset = 0) const;
  OptoReg::Name spill_regname    (int spill_index, int esp_offset = 0) const;
  OptoReg::Name register_regname (RInfo reg,       int esp_offset = 0) const;

  OopMap* new_oopmap (int esp_offset = 0);
  int esp_offset_for_local (int local_index, int esp_offset = 0) const;

  // return true if real-indeces of specified java-local-indeces are adjacent
  bool are_adjacent_indeces (int jix1, int jix2) const;

  static void init ();
};


//--------------------------------------------------------
//               ArgumentLocation
//--------------------------------------------------------

class ArgumentLocation VALUE_OBJ_CLASS_SPEC {
  friend CallingConvention;
 private:
  int _n;         // either a register number 0-5 (I0-I5 or O0-O5), or an offset (in words) from SP/FP

  ArgumentLocation (int n) : _n(n) {}
  ArgumentLocation (char k, int n) {
    switch (k) {
    case 'r':   set_register_location(n); break;
    case 's':   set_stack_location(n);    break;
    default:    ShouldNotReachHere();     break;
    }
  }

  int number () const { return _n; }

  void set_stack_location (int offset);
  void set_register_location (int number);

 public:
  ArgumentLocation () : _n(-1) {}

  bool is_illegal ()      const { return _n == -1; }
  bool is_register_arg () const;
  bool is_stack_arg ()    const;

  static ArgumentLocation new_stack_arg (int offset) {
    return ArgumentLocation('s', offset);
  }

  static ArgumentLocation new_reg_arg (int n) {
    return ArgumentLocation('r', n);
  }

  int stack_offset_in_words () const;

  Address incoming_stack_location () const;
  Address outgoing_stack_location () const;

  RInfo incoming_reg_location () const;
  RInfo outgoing_reg_location () const;

  void print (bool incoming = true) const  PRODUCT_RETURN;
};


//--------------------------------------------------------
//               CallingConvention
//--------------------------------------------------------

class CallingConvention: public ResourceObj {
  friend FrameMap;

 private:
  GrowableArray<int>* _args;

  CallingConvention () : _args(NULL)             {}
  CallingConvention (int size)                   { _args = new GrowableArray<int>(size, size); }

  void map (int arg, ArgumentLocation where)     { _args->at_put_grow(arg, where.number()); }

 public:
  ArgumentLocation arg_at(int i) const           { return ArgumentLocation(_args->at(i)); }
  int length() const                             { return _args->length(); }

#ifndef PRODUCT
  void print (bool incoming = true) const {
    for (int i = 0; i < length(); i++) {
      arg_at(i).print(incoming);
    }
  }
#endif // PRODUCT
};
