/*
 * @(#)c1_GraphBuilder.cpp	1.117 00/04/18 19:17:20
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_c1_GraphBuilder.cpp.incl"


// Implementation of BlockListBuilder

BlockBegin* BlockListBuilder::new_block_at(int bci, BlockBegin::Flag f) {
  BlockBegin* b = _bci2block->at(bci);
  if (b == NULL) {
    b = new BlockBegin(scope(), bci);
    _bci2block->at_put(bci, b);
  }
  // make sure all flags are accumulated
  b->set(f);
  return b;
}


void BlockListBuilder::set_leaders() {
  ciBytecodeStream s(method());
  while (s.next() >= 0) {
    switch (s.code()) {
      case Bytecodes::_ifeq        : // fall through
      case Bytecodes::_ifne        : // fall through
      case Bytecodes::_iflt        : // fall through
      case Bytecodes::_ifge        : // fall through
      case Bytecodes::_ifgt        : // fall through
      case Bytecodes::_ifle        : // fall through
      case Bytecodes::_if_icmpeq   : // fall through
      case Bytecodes::_if_icmpne   : // fall through
      case Bytecodes::_if_icmplt   : // fall through
      case Bytecodes::_if_icmpge   : // fall through
      case Bytecodes::_if_icmpgt   : // fall through
      case Bytecodes::_if_icmple   : // fall through
      case Bytecodes::_if_acmpeq   : // fall through
      case Bytecodes::_if_acmpne   : // fall through
      case Bytecodes::_ifnull      : // fall through
      case Bytecodes::_ifnonnull   :
        new_block_at(s.dest());
        new_block_at(s.next_bci());
        break;

      case Bytecodes::_goto        :
        new_block_at(s.dest());
        break;

      case Bytecodes::_jsr         :
        new_block_at(s.dest(), BlockBegin::subroutine_entry_flag);
        break;

      case Bytecodes::_tableswitch :
        { Bytecode_tableswitch *switch_ = Bytecode_tableswitch_at(s.bcp());
          // set block for each case
          int l = switch_->length();
          for (int i = 0; i < l; i++) new_block_at(s.bci() + switch_->dest_offset_at(i));
          // set default block
          new_block_at(s.bci() + switch_->default_offset());
        }
        break;
      
      case Bytecodes::_lookupswitch:
        { Bytecode_lookupswitch *switch_ = Bytecode_lookupswitch_at(s.bcp());
          // set block for each case
          int l = switch_->number_of_pairs();        
          for (int i = 0; i < l; i++) new_block_at(s.bci() + switch_->pair_at(i)->offset());
          // set default block
          new_block_at(s.bci() + switch_->default_offset());
        }
        break;

      case Bytecodes::_goto_w      :
        new_block_at(s.dest_w());
        break;

      case Bytecodes::_jsr_w       :
        new_block_at(s.dest_w(), BlockBegin::subroutine_entry_flag);
        break;

      default                      :
        // nothing to do
        break;
    }
  }
}


void BlockListBuilder::set_xhandler_entries() {
  XHandlers* list = xhandlers();
  const int n = list->number_of_handlers();
  for (int i = 0; i < n; i++) {
    XHandler* h = list->handler_at(i);
    h->set_entry(new_block_at(h->handler_bci(), BlockBegin::exception_entry_flag));
  }
}


BlockListBuilder::BlockListBuilder(IRScope* scope, int osr_bci) {
  _scope = scope;
  _bci2block = new GrowableArray<BlockBegin*>(method()->code_size(), method()->code_size(), NULL);
  set_leaders();
  set_xhandler_entries();
  _std_entry = new_block_at(0, BlockBegin::std_entry_flag);
  _osr_entry = (osr_bci >= 0 ? new_block_at(osr_bci, BlockBegin::osr_entry_flag) : NULL);
}


// Implementation of GraphBuilder

#define BAILOUT(msg)         { bailout(msg); return    ; }
#define BAILOUT_(msg, res)   { bailout(msg); return res; }


void GraphBuilder::load_constant() {
  ciConstant con = stream()->get_constant();
  if (con.type() == T_ILLEGAL) {
    BAILOUT("could not resolve a constant");
  } else {
    ValueType* t = illegalType;
    switch (con.type()) {
      case T_BOOLEAN: t = new IntConstant     (con.as_boolean()); break;
      case T_BYTE   : t = new IntConstant     (con.as_byte   ()); break;
      case T_CHAR   : t = new IntConstant     (con.as_char   ()); break;
      case T_SHORT  : t = new IntConstant     (con.as_short  ()); break;
      case T_INT    : t = new IntConstant     (con.as_int    ()); break;
      case T_LONG   : t = new LongConstant    (con.as_long   ()); break;
      case T_FLOAT  : t = new FloatConstant   (con.as_float  ()); break;
      case T_DOUBLE : t = new DoubleConstant  (con.as_double ()); break;
      case T_ARRAY  : t = new ArrayConstant   (con.as_object ()->as_array   ()); break;
      case T_OBJECT : t = new InstanceConstant(con.as_object ()->as_instance()); break;
      default       : ShouldNotReachHere();
    }
    push(t, append(new Constant(t)));
  }
}


void GraphBuilder::load_local(ValueType* type, int index) {
  Value x = NULL;
  if (EliminateLoads) x = state()->load_local(index);
  if (x != NULL) {
    if (PrintLoadElimination) tty->print_cr("load local %d eliminated @ %d", index, bci());
  } else {
    x = append(new LoadLocal(type, index));
    state()->store_local(index, x);
  }
  push(type, x);
}


void GraphBuilder::store_local(ValueType* type, int index) {
  Value x = pop(type);
  // because of Java FPU semantics, float and doubles must be rounded
  // when storing them => kill them instead of keeping the result if
  // the CPU does not provide correctly rounded results in general
  if (RoundFloatsWithStore && type->is_float_kind()) {
    state()->kill_local(index);
  } else {
    state()->store_local(index, x);
  }
  append(new StoreLocal(type, index, x));
  // make sure any outstanding loads are executed before
  state()->pin_stack_locals(index);
  if (type->is_double_word()) state()->pin_stack_locals(index + 1);
}


void GraphBuilder::load_indexed(BasicType type) {
  Value index = ipop();
  Value array = apop();
  push(as_ValueType(type), append(new LoadIndexed(array, index, type)));
}


void GraphBuilder::store_indexed(BasicType type) {
  Value value = pop(as_ValueType(type));
  Value index = ipop();
  Value array = apop();
  StoreIndexed* result = new StoreIndexed(array, index, type, value);
  if (value->type()->is_object()) {
    // array store check may be required => needs copy of stack
    vmap()->kill_all();      // for now, hopefully we need this only for calls eventually
    state()->clear_locals(); // for now, hopefully we need this only for calls eventually
    state()->pin_stack_all();
    result->set_state(state()->copy());
  } else {
    vmap()->kill_memory(value->type()); // invalidate all CSEs that are memory accesses
    state()->pin_stack_indexed();
  }
  append(result);
}


void GraphBuilder::stack_op(Bytecodes::Code code) {
  switch (code) {
    case Bytecodes::_pop:
      { state()->raw_pop();
      }
      break;
    case Bytecodes::_pop2:
      { state()->raw_pop();
        state()->raw_pop();
      }
      break;
    case Bytecodes::_dup:
      { Value w = state()->raw_pop();
        state()->raw_push(w);
        state()->raw_push(w);
      }
      break;
    case Bytecodes::_dup_x1:
      { Value w1 = state()->raw_pop();
        Value w2 = state()->raw_pop();
        state()->raw_push(w1);
        state()->raw_push(w2);
        state()->raw_push(w1);
      }
      break;
    case Bytecodes::_dup_x2:
      { Value w1 = state()->raw_pop();
        Value w2 = state()->raw_pop();
        Value w3 = state()->raw_pop();
        state()->raw_push(w1);
        state()->raw_push(w3);
        state()->raw_push(w2);
        state()->raw_push(w1);
      }
      break;
    case Bytecodes::_dup2:
      { Value w1 = state()->raw_pop();
        Value w2 = state()->raw_pop();
        state()->raw_push(w2);
        state()->raw_push(w1);
        state()->raw_push(w2);
        state()->raw_push(w1);
      }
      break;
    case Bytecodes::_dup2_x1:
      { Value w1 = state()->raw_pop();
        Value w2 = state()->raw_pop();
        Value w3 = state()->raw_pop();
        state()->raw_push(w2);
        state()->raw_push(w1);
        state()->raw_push(w3);
        state()->raw_push(w2);
        state()->raw_push(w1);
      }
      break;
    case Bytecodes::_dup2_x2:
      { Value w1 = state()->raw_pop();
        Value w2 = state()->raw_pop();
        Value w3 = state()->raw_pop();
        Value w4 = state()->raw_pop();
        state()->raw_push(w2);
        state()->raw_push(w1);
        state()->raw_push(w4);
        state()->raw_push(w3);
        state()->raw_push(w2);
        state()->raw_push(w1);
      }
      break;
    case Bytecodes::_swap:
      { Value w1 = state()->raw_pop();
        Value w2 = state()->raw_pop();
        state()->raw_push(w1);
        state()->raw_push(w2);
      }
      break;
    default:
      ShouldNotReachHere();
      break;
  }
}


void GraphBuilder::arithmetic_op(ValueType* type, Bytecodes::Code code) {
  Value y = pop(type);
  Value x = pop(type);
  push(type, append(new ArithmeticOp(code, x, y)));
}


void GraphBuilder::negate_op(ValueType* type) {
  push(type, append(new NegateOp(pop(type))));
}


void GraphBuilder::shift_op(ValueType* type, Bytecodes::Code code) {
  Value s = ipop();
  Value x = pop(type);
  // try to simplify
  // Note: This code should go into the canonicalizer as soon as it can
  //       can handle canonicalized forms that contain more than one node.
  if (CanonicalizeNodes && code == Bytecodes::_iushr) {
    // pattern: x >>> s
    IntConstant* s1 = s->type()->as_IntConstant();
    if (s1 != NULL) {
      // pattern: x >>> s1, with s1 constant
      ShiftOp* l = x->as_ShiftOp();
      if (l != NULL && l->op() == Bytecodes::_ishl) {
        // pattern: (a << b) >>> s1
        IntConstant* s0 = l->y()->type()->as_IntConstant();
        if (s0 != NULL) {
          // pattern: (a << s0) >>> s1
          const int s0c = s0->value() & 0x1F; // only the low 5 bits are significant for shifts
          const int s1c = s1->value() & 0x1F; // only the low 5 bits are significant for shifts
          if (s0c == s1c) {
            if (s0c == 0) {
              // pattern: (a << 0) >>> 0 => simplify to: a
              ipush(l->x());
            } else {
              // pattern: (a << s0c) >>> s0c => simplify to: a & m, with m constant
              assert(0 < s0c && s0c < BitsPerWord, "adjust code below to handle corner cases");
              const int m = (1 << (BitsPerWord - s0c)) - 1;
              Value s = append(new Constant(new IntConstant(m)));
              ipush(append(new LogicOp(Bytecodes::_iand, l->x(), s)));
            }
            return;
          }
        }
      }
    }
  }
  // could not simplify
  push(type, append(new ShiftOp(code, x, s)));
}


void GraphBuilder::logic_op(ValueType* type, Bytecodes::Code code) {
  Value y = pop(type);
  Value x = pop(type);
  push(type, append(new LogicOp(code, x, y)));
}


void GraphBuilder::compare_op(ValueType* type, Bytecodes::Code code) {
  Value y = pop(type);
  Value x = pop(type);
  ipush(append(new CompareOp(code, x, y)));
}


void GraphBuilder::convert(Bytecodes::Code op, BasicType from, BasicType to) {
  push(as_ValueType(to), append(new Convert(op, pop(as_ValueType(from)), as_ValueType(to))));
}


void GraphBuilder::increment() {
  int index = stream()->get_index();
  int delta = stream()->is_wide() ? (signed short)Bytes::get_Java_u2(stream()->bcp() + 4) : (signed char)(stream()->bcp()[2]);
  load_local(intType, index);
  ipush(append(new Constant(new IntConstant(delta))));
  arithmetic_op(intType, Bytecodes::_iadd);
  store_local(intType, index);
}


void GraphBuilder::if_node(Value x, If::Condition cond, Value y) {
  BlockBegin* tsux = block_at(stream()->dest());
  BlockBegin* fsux = block_at(stream()->next_bci());
  append(new If(x, cond, false, y, tsux, fsux));
}


void GraphBuilder::if_zero(ValueType* type, If::Condition cond) {
  Value y = append(new Constant(intZero));
  Value x = ipop();
  if_node(x, cond, y);
}


void GraphBuilder::if_null(ValueType* type, If::Condition cond) {
  Value y = append(new Constant(objectNull));
  Value x = apop();
  if_node(x, cond, y);
}


void GraphBuilder::if_same(ValueType* type, If::Condition cond) {
  Value y = pop(type);
  Value x = pop(type);
  if_node(x, cond, y);
}


void GraphBuilder::jsr(int dest) {
  // remember jsr
  _found_jsr = true;
  // add subroutine to current block
  BlockBegin* subroutine = block_at(dest);
  assert(subroutine != NULL, "subroutine begin must be control flow begin");
  assert(subroutine->is_set(BlockBegin::subroutine_entry_flag), "must be marked as subroutine entry");
  // compute subroutine sub-graph
  if (!subroutine->is_set(BlockBegin::was_visited_flag)) {
    // create temporary starting state and join with subroutine
    push(addressType, new Constant(new AddressConstant(stream()->next_bci())));
    // be carefull, bailout if bytecodes are strange
    if (!subroutine->try_join(state())) BAILOUT("jsr join failed");
    state()->raw_pop();
    // fill in subroutine sub-graph
    GraphBuilder gb(compilation(), scope(), bci2block(), subroutine);
    if (gb.bailed_out()) BAILOUT("bailout in subroutine");
  }
  // make sure control flow ends correspond
  EndList* endlist = subroutine->endlist();
  if (endlist == NULL) BAILOUT("jsr endlist could not be computed");
  // find a Ret, if any
  int i;
  for (i = endlist->length() - 1; i >= 0 && endlist->at(i)->as_Ret() == NULL; i--);
  if (i >= 0) {
    // found a Ret => compare its state with state of all other Rets
    ValueStack* ret_state = endlist->at(i)->as_Ret()->state();
    for (int j = endlist->length() - 1; j >= 0; j--) {
      Ret* xi = endlist->at(j)->as_Ret();
      if (xi != NULL) {
        // if states are not the same => bail out
        if (!ret_state->is_same(xi->state())) BAILOUT("jsr ret states incompatible");
      }
    }
    // all states correspond
    // => jsr is ok
    // => add subroutine to current block and add jsr node
    _block->add_subroutine(subroutine);
    append_split(new Jsr(subroutine));
    _last = _last->set_next(new JsrContinuation, stream()->next_bci()); // must have jsr 'return address'
    // continue with ret state
    // note: make a copy since it is modified & since we must
    //       have all locals cleared (for load elimination) -
    //       was bug - gri 6/22/99
    _state = ret_state->copy();
    // note: bailout if continuation stack is not empty
    //       since the backend can currently not deal
    //       with this situation.
    if (!state()->stack_is_empty()) BAILOUT("jsr returns with non-empty stack");
  } else {
    // jsr subgraph ended w/o a Ret
    // => control flow cannot return here
    // => push next bci and add Goto instead (this also terminates further parsing of this block)
    push(addressType, append(new Constant(new AddressConstant(stream()->next_bci()))));
    append(new Goto(block_at(dest)));
  }
}


void GraphBuilder::table_switch() {
  Bytecode_tableswitch* switch_ = Bytecode_tableswitch_at(method()->code() + bci());
  const int l = switch_->length();
  if (CanonicalizeNodes && l == 1) {
    // total of 2 successors => use If instead of switch
    // Note: This code should go into the canonicalizer as soon as it can
    //       can handle canonicalized forms that contain more than one node.
    Value key = append(new Constant(new IntConstant(switch_->low_key())));
    BlockBegin* tsux = block_at(bci() + switch_->dest_offset_at(0));
    BlockBegin* fsux = block_at(bci() + switch_->default_offset());
    append(new If(ipop(), If::eql, true, key, tsux, fsux));
  } else {
    // collect successors
    BlockList* sux = new GrowableArray<BlockBegin*>(l + 1, l + 1, NULL);
    int i;
    for (i = 0; i < l; i++) {
      sux->at_put(i, block_at(bci() + switch_->dest_offset_at(i)));
    }
    // add default successor
    sux->at_put(i, block_at(bci() + switch_->default_offset()));
    append(new TableSwitch(ipop(), sux, switch_->low_key()));
  }
}


void GraphBuilder::lookup_switch() {
  Bytecode_lookupswitch* switch_ = Bytecode_lookupswitch_at(method()->code() + bci());
  const int l = switch_->number_of_pairs();
  if (CanonicalizeNodes && l == 1) {
    // total of 2 successors => use If instead of switch
    // Note: This code should go into the canonicalizer as soon as it can
    //       can handle canonicalized forms that contain more than one node.
    // simplify to If
    LookupswitchPair* pair = switch_->pair_at(0);
    Value key = append(new Constant(new IntConstant(pair->match())));
    BlockBegin* tsux = block_at(bci() + pair->offset());
    BlockBegin* fsux = block_at(bci() + switch_->default_offset());
    append(new If(ipop(), If::eql, true, key, tsux, fsux));
  } else {
    // collect successors & keys
    BlockList* sux = new GrowableArray<BlockBegin*>(l + 1, l + 1, NULL);
    GrowableArray<int>* keys = new GrowableArray<int>(l, l, 0);
    int i;
    for (i = 0; i < l; i++) {
      LookupswitchPair* pair = switch_->pair_at(i);
      sux->at_put(i, block_at(bci() + pair->offset()));
      keys->at_put(i, pair->match());
    }
    // add default successor
    sux->at_put(i, block_at(bci() + switch_->default_offset()));
    append(new LookupSwitch(ipop(), sux, keys));
  }
}


void GraphBuilder::method_return(Value x) {
  int index = method()->is_synchronized() ? state()->unlock() : -1;
  append(new Return(x, index));
}


void GraphBuilder::access_field(Bytecodes::Code code) {
  ciField* field = stream()->get_field();
  ciInstanceKlass* holder = field->holder();
  BasicType field_type = field->field_type();
  ValueType* type = as_ValueType(field_type);
  Value obj = new Constant(new ClassConstant(holder));
  // call to will_link may cause class loading: do not call it if holder is not loaded
  const bool is_loaded = holder->is_loaded() && field->will_link(method()->holder(), code);
  const bool is_initialized = is_loaded && holder->is_initialized();
  const int offset = is_loaded ? field->offset() : -1;
  switch (code) {
    case Bytecodes::_getstatic:
      // make sure klass is initialized, bailout otherwise
      if (!LateBailout && !is_initialized) BAILOUT("klass not initialized for getstatic");
      push(type, append(new LoadField(append(obj), offset, field_type, true, is_loaded, is_initialized)));
      break;
    case Bytecodes::_putstatic:
      // make sure klass is initialized, bailout otherwise
      if (!LateBailout && !is_initialized) BAILOUT("klass not initialized for putstatic");
      { Value val = pop(type);
        append(new StoreField(append(obj), offset, field_type, val, true, is_loaded, is_initialized));
        vmap()->kill_memory(type);   // invalidate all CSEs that are memory accesses
        state()->pin_stack_fields(); // make sure any outstanding loads are executed before (conservative)
      }
      break;
    case Bytecodes::_getfield :
      push(type, append(new LoadField(apop(), offset, field_type, false, is_loaded, true)));
      break;
    case Bytecodes::_putfield :
      { Value val = pop(type);
        append(new StoreField(apop(), offset, field_type, val, false, is_loaded, true));
        vmap()->kill_memory(type);   // invalidate all CSEs that are memory accesses
        state()->pin_stack_fields(); // make sure any outstanding loads are executed before (conservative)
      }
      break;
    default                   :
      ShouldNotReachHere();
      break;
  }
}


static void complete_substitution(Value* x) {
  Value s = (*x)->subst();
  if (s != *x) {
    assert((*x)->as_LoadLocal() != NULL, "must be a LoadLocal node");
    if (PrintInlining) tty->print_cr("substituted N%d with N%d", (*x)->id(), s->id());
    *x = s;
  }
}


void GraphBuilder::invoke(Bytecodes::Code code) {
  ciMethod* target = stream()->get_method();
  // we have to make sure the argument size (incl. the receiver)
  // is correct for compilation (the call would fail later during
  // linkage anyway) - was bug (gri 7/28/99)
  if (target->is_loaded() && target->is_static() != (code == Bytecodes::_invokestatic)) BAILOUT("will cause link error");
  // check if we could do inlining
  ciInstanceKlass* klass = target->holder();
  if (!PatchALot && Inline && klass->is_loaded() && klass->is_initialized() 
      && target->will_link(klass, stream()->get_declared_method_holder(), code)) {
    // callee is known => check if we have static binding
    assert(target->is_loaded(), "callee must be known");
    if (code == Bytecodes::_invokestatic
     || code == Bytecodes::_invokespecial
     || code == Bytecodes::_invokevirtual && target->is_final_method()
    ) {
      // static binding => check if callee is ok
      Inliner inliner(this);
      if (inliner.try_inline(target)) {
        if (PrintCompilation) {
          tty->print_cr("- (%d) inlining at bci %d ", scope()->level(), bci());
        }
        return;
      }
    }
  }
  // inlining not successful => standard invoke
  bool is_static = code == Bytecodes::_invokestatic;
  ValueType* result_type = as_ValueType(target->return_type());
  Values* args = state()->pop_arguments(target->arg_size_no_receiver());
  Value recv = is_static ? NULL : apop();
  Invoke* result = new Invoke(code, result_type, recv, args, target->is_loaded() && target->is_final_method(), target->is_loaded());
  // push result
  append_split(result);
  if (result_type != voidType) push(result_type, result);
}


void GraphBuilder::new_instance(int klass_index) {
  ciKlass* klass = stream()->get_klass();
  assert(klass->is_instance_klass(), "must be an instance klass");
  apush(append_split(new NewInstance(klass->as_instance_klass())));
}


void GraphBuilder::new_type_array() {
  apush(append_split(new NewTypeArray(ipop(), (BasicType)stream()->get_index())));
}


void GraphBuilder::new_object_array() {
  ciKlass* klass = stream()->get_klass();
  apush(append_split(new NewObjectArray(klass, ipop())));
}


void GraphBuilder::check_cast(int klass_index) {
  ciKlass* klass = stream()->get_klass();
  apush(append_split(new CheckCast(klass, apop())));
}


void GraphBuilder::instance_of(int klass_index) {
  ciKlass* klass = stream()->get_klass();
  ipush(append_split(new InstanceOf(klass, apop())));
}


void GraphBuilder::monitorenter(Value x) {
  append_split(new MonitorEnter(x, state()->lock(scope(), bci())));
}


void GraphBuilder::monitorexit(Value x) {
  // Note: Potentially, the monitor state in an exception handler
  //       can be wrong due to wrong 'initialization' of the handler
  //       via a wrong asynchronous exception path. This can happen,
  //       if the exception handler range for asynchronous exceptions
  //       is too long (see also java bug 4327029, and comment in
  //       GraphBuilder::handle_exception()). This may cause 'under-
  //       flow' of the monitor stack => bailout instead.
  if (state()->locks_size() < 1) BAILOUT("monitor stack underflow");
  append_split(new MonitorExit(x, state()->unlock()));
}


void GraphBuilder::new_multi_array(int dimensions) {
  ciKlass* klass = stream()->get_klass();
  Values* dims = new GrowableArray<Value>(dimensions, dimensions, NULL);
  // fill in all dimensions
  int i = dimensions;
  while (i-- > 0) dims->at_put(i, ipop());
  // create array
  apush(append_split(new NewMultiArray(klass, dims)));
}


Instruction* GraphBuilder::append_base(Instruction* instr) {
  Instruction* i1 = Canonicalizer(instr).canonical();
  Instruction* i2 = vmap()->find(i1);
  if (i1 == i2) {
    // i1 was not eliminated => append it
    _last = _last->set_next(i2, bci());
    assert(_last == i2, "adjust code below");
    if (i2->as_StateSplit() != NULL && i2->as_BlockEnd() == NULL) {
      StateSplit* s = i2->as_StateSplit();
      assert(s != NULL, "s must exist");
      vmap()->kill_all();      // for now, hopefully we need this only for calls eventually
      state()->clear_locals(); // for now, hopefully we need this only for calls eventually
      state()->pin_stack_for_state_split();
      s->set_state(state()->copy());
    }
  }
  return i2;
}


Instruction* GraphBuilder::append(Instruction* instr) {
  assert(instr->as_StateSplit() == NULL || instr->as_BlockEnd() != NULL, "wrong append used");
  return append_base(instr);
}


Instruction* GraphBuilder::append_split(StateSplit* instr) {
  return append_base(instr);
}


void GraphBuilder::bailout(const char* msg) {
  assert(msg != NULL, "bailout msg must exist");
  _bailout_msg = msg;
  if (PrintBailouts) tty->print_cr(" Bailout reason: %s", msg);
}


void GraphBuilder::handle_exception(bool is_async) {
  // join with all potential exception handlers
  XHandlers* list = xhandlers();
  const int n = list->number_of_handlers();
  for (int i = 0; i < n; i++) {
    XHandler* h = list->handler_at(i);
    if (h->covers(bci())) {
      assert(h->entry() == block_at(h->handler_bci()), "blocks must correspond");
      // h is a potential exception handler => join it

      // Note: The code below is necessary since methods may have in-
      //       correct exception handler ranges for async. exceptions
      //       (this is the case for the current javac and some 3rd
      //       party javac's). In particular, the exception handler
      //       ranges may be too long, in which case an asynchronous
      //       exception will cause a control flow to the exception
      //       handler with the wrong monitor state (i.e., monitor
      //       already unlocked). If the graph builder tries to join
      //       the corresponding exception handler, it will fail with
      //       an assertion since the monitor stacks do not correspond,
      //       unless we do some extra checking beforehand.
      //       
      //       This problem showed up with java.lang.ref.Finalizer::add
      //       in J2SE 1.3 rc2 (see also bug 4324989: we forgot to add
      //       consider asynchronous exception handlers - after fixing
      //       that problem we encountered the javac problem 4327029).
      if (h->entry()->is_set(BlockBegin::was_visited_flag)) {
        assert(h->entry()->state() != NULL, "state must be set if node was visited before");
        if (state()->locks_size() != h->entry()->state()->locks_size()) {
          // trying to join the exception handler would fail with
          // an assertion failure since the monitor stacks do not
          // correspond
          if (is_async) {
            // assume exception handler range is wrong for
            // asynchronous exception => simply ignore it
            return;
          } else {
            // the monitor states may not match because of
            // an asynchronous exception path that set up
            // the handler monitor state wrongly before =>
            // bailout since we don't know what is correct
            BAILOUT("illegal monitor state");
          }
        }
      }

      // Note: Since locals are all NULL here & the expression
      //       stack is empty besides the dummy exception object
      //       we don't have to do a full & time consuming join
      //       if the exception handler was visited before
      //       => do it always only in debug mode
#ifndef ASSERT
      if (!h->entry()->is_set(BlockBegin::was_visited_flag)) {
#endif // ASSERT
        ValueStack* s = state()->copy();    // do not change current state => make a copy
        s->clear_stack();                   // empty expression stack after exception
        s->apush(new Constant(objectNull)); // but the exception oop is on the stack
        // Note: This join must always work (we are not looking at locals)
        //       since the exception is the only value on the stack. Also,
        //       the monitor stacks must correspond => join (not try_join).
        h->entry()->join(s);                // merge with exception handler
        // fill in exception handler subgraph
        GraphBuilder gb(compilation(), scope(), bci2block(), h->entry());
        if (gb.bailed_out()) BAILOUT("bailout in exception handler");
#ifndef ASSERT
      }
#endif // ASSERT

      // add h to the list of exception handlers of this block
      _block->add_exception_handler(h->entry());

      // stop when reaching catchall
      if (h->catch_type() == 0) return;
    }
  }
}


BlockEnd* GraphBuilder::connect_to_end(BlockBegin* beg) {
  // setup iteration
  vmap()->kill_all();
  _block = beg;
  _state = beg->state()->copy();
  _last  = beg;
  ciBytecodeStream s(method());
  s.set_start(beg->bci());
  _stream = &s;
  int prev_bci = beg->bci();
  // iterate
  Bytecodes::Code code = Bytecodes::_illegal;
  bool prev_is_monitorenter = false;
  while (last()->as_BlockEnd() == NULL && (code = stream()->next()) >= 0 && (block_at(s.bci()) == NULL || block_at(s.bci()) == beg) && !bailed_out()) {
    // handle potential exceptions thrown by current bytecode
    // note 1: exceptions must be handled before the bytecode as
    //         the (locking) state before the bytecode is relevant
    //         in the exception handler (e.g. for monitorenter)
    // note 2: if the previous bytecode was a monitorenter bytecode,
    //         assume the current bytecode throws an asynchronous
    //         exception to get at least one control flow path to
    //         the handler for asynchronous exceptions (this will
    //         hopefully also be the first control flow path to
    //         that handler which will 'initialize' it correctly;
    //         see also javac problem mentioned in handle_exception())
    if (has_handler() && (prev_is_monitorenter || can_trap(code))) {
      handle_exception(prev_is_monitorenter || is_async(code));
      if (bailed_out()) return NULL;
    }
    prev_is_monitorenter = false;

    // handle bytecode
    switch (code) {
      case Bytecodes::_nop            : /* nothing to do */ break;
      case Bytecodes::_aconst_null    : apush(append(new Constant(objectNull            ))); break;
      case Bytecodes::_iconst_m1      : ipush(append(new Constant(new IntConstant   (-1)))); break;
      case Bytecodes::_iconst_0       : ipush(append(new Constant(intZero               ))); break;
      case Bytecodes::_iconst_1       : ipush(append(new Constant(intOne                ))); break;
      case Bytecodes::_iconst_2       : ipush(append(new Constant(new IntConstant   ( 2)))); break;
      case Bytecodes::_iconst_3       : ipush(append(new Constant(new IntConstant   ( 3)))); break;
      case Bytecodes::_iconst_4       : ipush(append(new Constant(new IntConstant   ( 4)))); break;
      case Bytecodes::_iconst_5       : ipush(append(new Constant(new IntConstant   ( 5)))); break;
      case Bytecodes::_lconst_0       : lpush(append(new Constant(new LongConstant  ( 0)))); break;
      case Bytecodes::_lconst_1       : lpush(append(new Constant(new LongConstant  ( 1)))); break;
      case Bytecodes::_fconst_0       : fpush(append(new Constant(new FloatConstant ( 0)))); break;
      case Bytecodes::_fconst_1       : fpush(append(new Constant(new FloatConstant ( 1)))); break;
      case Bytecodes::_fconst_2       : fpush(append(new Constant(new FloatConstant ( 2)))); break;
      case Bytecodes::_dconst_0       : dpush(append(new Constant(new DoubleConstant( 0)))); break;
      case Bytecodes::_dconst_1       : dpush(append(new Constant(new DoubleConstant( 1)))); break;
      case Bytecodes::_bipush         : ipush(append(new Constant(new IntConstant(((signed char*)s.bcp())[1])))); break;
      case Bytecodes::_sipush         : ipush(append(new Constant(new IntConstant((short)Bytes::get_Java_u2(s.bcp()+1))))); break;
      case Bytecodes::_ldc            : // fall through
      case Bytecodes::_ldc_w          : // fall through
      case Bytecodes::_ldc2_w         : load_constant(); break;
      case Bytecodes::_iload          : load_local(intType     , s.get_index()); break;
      case Bytecodes::_lload          : load_local(longType    , s.get_index()); break;
      case Bytecodes::_fload          : load_local(floatType   , s.get_index()); break;
      case Bytecodes::_dload          : load_local(doubleType  , s.get_index()); break;
      case Bytecodes::_aload          : load_local(instanceType, s.get_index()); break;
      case Bytecodes::_iload_0        : load_local(intType   , 0); break;
      case Bytecodes::_iload_1        : load_local(intType   , 1); break;
      case Bytecodes::_iload_2        : load_local(intType   , 2); break;
      case Bytecodes::_iload_3        : load_local(intType   , 3); break;
      case Bytecodes::_lload_0        : load_local(longType  , 0); break;
      case Bytecodes::_lload_1        : load_local(longType  , 1); break;
      case Bytecodes::_lload_2        : load_local(longType  , 2); break;
      case Bytecodes::_lload_3        : load_local(longType  , 3); break;
      case Bytecodes::_fload_0        : load_local(floatType , 0); break;
      case Bytecodes::_fload_1        : load_local(floatType , 1); break;
      case Bytecodes::_fload_2        : load_local(floatType , 2); break;
      case Bytecodes::_fload_3        : load_local(floatType , 3); break;
      case Bytecodes::_dload_0        : load_local(doubleType, 0); break;
      case Bytecodes::_dload_1        : load_local(doubleType, 1); break;
      case Bytecodes::_dload_2        : load_local(doubleType, 2); break;
      case Bytecodes::_dload_3        : load_local(doubleType, 3); break;
      case Bytecodes::_aload_0        : load_local(objectType, 0); break;
      case Bytecodes::_aload_1        : load_local(objectType, 1); break;
      case Bytecodes::_aload_2        : load_local(objectType, 2); break;
      case Bytecodes::_aload_3        : load_local(objectType, 3); break;
      case Bytecodes::_iaload         : load_indexed(T_INT   ); break;
      case Bytecodes::_laload         : load_indexed(T_LONG  ); break;
      case Bytecodes::_faload         : load_indexed(T_FLOAT ); break;
      case Bytecodes::_daload         : load_indexed(T_DOUBLE); break;
      case Bytecodes::_aaload         : load_indexed(T_OBJECT); break;
      case Bytecodes::_baload         : load_indexed(T_BYTE  ); break;
      case Bytecodes::_caload         : load_indexed(T_CHAR  ); break;
      case Bytecodes::_saload         : load_indexed(T_SHORT ); break;
      case Bytecodes::_istore         : store_local(intType   , s.get_index()); break;
      case Bytecodes::_lstore         : store_local(longType  , s.get_index()); break;
      case Bytecodes::_fstore         : store_local(floatType , s.get_index()); break;
      case Bytecodes::_dstore         : store_local(doubleType, s.get_index()); break;
      case Bytecodes::_astore         : store_local(objectType, s.get_index()); break;
      case Bytecodes::_istore_0       : store_local(intType   , 0); break;
      case Bytecodes::_istore_1       : store_local(intType   , 1); break;
      case Bytecodes::_istore_2       : store_local(intType   , 2); break;
      case Bytecodes::_istore_3       : store_local(intType   , 3); break;
      case Bytecodes::_lstore_0       : store_local(longType  , 0); break;
      case Bytecodes::_lstore_1       : store_local(longType  , 1); break;
      case Bytecodes::_lstore_2       : store_local(longType  , 2); break;
      case Bytecodes::_lstore_3       : store_local(longType  , 3); break;
      case Bytecodes::_fstore_0       : store_local(floatType , 0); break;
      case Bytecodes::_fstore_1       : store_local(floatType , 1); break;
      case Bytecodes::_fstore_2       : store_local(floatType , 2); break;
      case Bytecodes::_fstore_3       : store_local(floatType , 3); break;
      case Bytecodes::_dstore_0       : store_local(doubleType, 0); break;
      case Bytecodes::_dstore_1       : store_local(doubleType, 1); break;
      case Bytecodes::_dstore_2       : store_local(doubleType, 2); break;
      case Bytecodes::_dstore_3       : store_local(doubleType, 3); break;
      case Bytecodes::_astore_0       : store_local(objectType, 0); break;
      case Bytecodes::_astore_1       : store_local(objectType, 1); break;
      case Bytecodes::_astore_2       : store_local(objectType, 2); break;
      case Bytecodes::_astore_3       : store_local(objectType, 3); break;
      case Bytecodes::_iastore        : store_indexed(T_INT   ); break;
      case Bytecodes::_lastore        : store_indexed(T_LONG  ); break;
      case Bytecodes::_fastore        : store_indexed(T_FLOAT ); break;
      case Bytecodes::_dastore        : store_indexed(T_DOUBLE); break;
      case Bytecodes::_aastore        : store_indexed(T_OBJECT); break;
      case Bytecodes::_bastore        : store_indexed(T_BYTE  ); break;
      case Bytecodes::_castore        : store_indexed(T_CHAR  ); break;
      case Bytecodes::_sastore        : store_indexed(T_SHORT ); break;
      case Bytecodes::_pop            : // fall through
      case Bytecodes::_pop2           : // fall through
      case Bytecodes::_dup            : // fall through
      case Bytecodes::_dup_x1         : // fall through
      case Bytecodes::_dup_x2         : // fall through
      case Bytecodes::_dup2           : // fall through
      case Bytecodes::_dup2_x1        : // fall through
      case Bytecodes::_dup2_x2        : // fall through
      case Bytecodes::_swap           : stack_op(code); break;
      case Bytecodes::_iadd           : arithmetic_op(intType   , code); break;
      case Bytecodes::_ladd           : arithmetic_op(longType  , code); break;
      case Bytecodes::_fadd           : arithmetic_op(floatType , code); break;
      case Bytecodes::_dadd           : arithmetic_op(doubleType, code); break;
      case Bytecodes::_isub           : arithmetic_op(intType   , code); break;
      case Bytecodes::_lsub           : arithmetic_op(longType  , code); break;
      case Bytecodes::_fsub           : arithmetic_op(floatType , code); break;
      case Bytecodes::_dsub           : arithmetic_op(doubleType, code); break;
      case Bytecodes::_imul           : arithmetic_op(intType   , code); break;
      case Bytecodes::_lmul           : arithmetic_op(longType  , code); break;
      case Bytecodes::_fmul           : arithmetic_op(floatType , code); break;
      case Bytecodes::_dmul           : arithmetic_op(doubleType, code); break;
      case Bytecodes::_idiv           : arithmetic_op(intType   , code); break;
      case Bytecodes::_ldiv           : arithmetic_op(longType  , code); break;
      case Bytecodes::_fdiv           : arithmetic_op(floatType , code); break;
      case Bytecodes::_ddiv           : arithmetic_op(doubleType, code); break;
      case Bytecodes::_irem           : arithmetic_op(intType   , code); break;
      case Bytecodes::_lrem           : arithmetic_op(longType  , code); break;
      case Bytecodes::_frem           : arithmetic_op(floatType , code); break;
      case Bytecodes::_drem           : arithmetic_op(doubleType, code); break;
      case Bytecodes::_ineg           : negate_op(intType   ); break;
      case Bytecodes::_lneg           : negate_op(longType  ); break;
      case Bytecodes::_fneg           : negate_op(floatType ); break;
      case Bytecodes::_dneg           : negate_op(doubleType); break;
      case Bytecodes::_ishl           : shift_op(intType , code); break;
      case Bytecodes::_lshl           : shift_op(longType, code); break;
      case Bytecodes::_ishr           : shift_op(intType , code); break;
      case Bytecodes::_lshr           : shift_op(longType, code); break;
      case Bytecodes::_iushr          : shift_op(intType , code); break;
      case Bytecodes::_lushr          : shift_op(longType, code); break;
      case Bytecodes::_iand           : logic_op(intType , code); break;
      case Bytecodes::_land           : logic_op(longType, code); break;
      case Bytecodes::_ior            : logic_op(intType , code); break;
      case Bytecodes::_lor            : logic_op(longType, code); break;
      case Bytecodes::_ixor           : logic_op(intType , code); break;
      case Bytecodes::_lxor           : logic_op(longType, code); break;
      case Bytecodes::_iinc           : increment(); break;
      case Bytecodes::_i2l            : convert(code, T_INT   , T_LONG  ); break;
      case Bytecodes::_i2f            : convert(code, T_INT   , T_FLOAT ); break;
      case Bytecodes::_i2d            : convert(code, T_INT   , T_DOUBLE); break;
      case Bytecodes::_l2i            : convert(code, T_LONG  , T_INT   ); break;
      case Bytecodes::_l2f            : convert(code, T_LONG  , T_FLOAT ); break;
      case Bytecodes::_l2d            : convert(code, T_LONG  , T_DOUBLE); break;
      case Bytecodes::_f2i            : convert(code, T_FLOAT , T_INT   ); break;
      case Bytecodes::_f2l            : convert(code, T_FLOAT , T_LONG  ); break;
      case Bytecodes::_f2d            : convert(code, T_FLOAT , T_DOUBLE); break;
      case Bytecodes::_d2i            : convert(code, T_DOUBLE, T_INT   ); break;
      case Bytecodes::_d2l            : convert(code, T_DOUBLE, T_LONG  ); break;
      case Bytecodes::_d2f            : convert(code, T_DOUBLE, T_FLOAT ); break;
      case Bytecodes::_i2b            : convert(code, T_INT   , T_BYTE  ); break;
      case Bytecodes::_i2c            : convert(code, T_INT   , T_CHAR  ); break;
      case Bytecodes::_i2s            : convert(code, T_INT   , T_SHORT ); break;
      case Bytecodes::_lcmp           : compare_op(longType  , code); break;
      case Bytecodes::_fcmpl          : compare_op(floatType , code); break;
      case Bytecodes::_fcmpg          : compare_op(floatType , code); break;
      case Bytecodes::_dcmpl          : compare_op(doubleType, code); break;
      case Bytecodes::_dcmpg          : compare_op(doubleType, code); break;
      case Bytecodes::_ifeq           : if_zero(intType   , If::eql); break;
      case Bytecodes::_ifne           : if_zero(intType   , If::neq); break;
      case Bytecodes::_iflt           : if_zero(intType   , If::lss); break;
      case Bytecodes::_ifge           : if_zero(intType   , If::geq); break;
      case Bytecodes::_ifgt           : if_zero(intType   , If::gtr); break;
      case Bytecodes::_ifle           : if_zero(intType   , If::leq); break;
      case Bytecodes::_if_icmpeq      : if_same(intType   , If::eql); break;
      case Bytecodes::_if_icmpne      : if_same(intType   , If::neq); break;
      case Bytecodes::_if_icmplt      : if_same(intType   , If::lss); break;
      case Bytecodes::_if_icmpge      : if_same(intType   , If::geq); break;
      case Bytecodes::_if_icmpgt      : if_same(intType   , If::gtr); break;
      case Bytecodes::_if_icmple      : if_same(intType   , If::leq); break;
      case Bytecodes::_if_acmpeq      : if_same(objectType, If::eql); break;
      case Bytecodes::_if_acmpne      : if_same(objectType, If::neq); break;
      case Bytecodes::_goto           : append(new Goto(block_at(s.dest()))); break;
      case Bytecodes::_jsr            : jsr(s.dest()); break;
      case Bytecodes::_ret            : append(new Ret(s.get_index())); break;
      case Bytecodes::_tableswitch    : table_switch(); break;
      case Bytecodes::_lookupswitch   : lookup_switch(); break;
      case Bytecodes::_ireturn        : method_return(ipop()); break;
      case Bytecodes::_lreturn        : method_return(lpop()); break;
      case Bytecodes::_freturn        : method_return(fpop()); break;
      case Bytecodes::_dreturn        : method_return(dpop()); break;
      case Bytecodes::_areturn        : method_return(apop()); break;
      case Bytecodes::_return         : method_return(NULL  ); break;
      case Bytecodes::_getstatic      : // fall through
      case Bytecodes::_putstatic      : // fall through
      case Bytecodes::_getfield       : // fall through
      case Bytecodes::_putfield       : access_field(code); break;
      case Bytecodes::_invokevirtual  : // fall through
      case Bytecodes::_invokespecial  : // fall through
      case Bytecodes::_invokestatic   : // fall through
      case Bytecodes::_invokeinterface: invoke(code); break;
      case Bytecodes::_xxxunusedxxx   : ShouldNotReachHere(); break;
      case Bytecodes::_new            : new_instance(s.get_index_big()); break;
      case Bytecodes::_newarray       : new_type_array(); break;
      case Bytecodes::_anewarray      : new_object_array(); break;
      case Bytecodes::_arraylength    : ipush(append(new ArrayLength(apop()))); break;
      case Bytecodes::_athrow         : append(new Throw(apop())); break;
      case Bytecodes::_checkcast      : check_cast(s.get_index_big()); break;
      case Bytecodes::_instanceof     : instance_of(s.get_index_big()); break;
      case Bytecodes::_monitorenter   : monitorenter(apop()); prev_is_monitorenter = true; break;
      case Bytecodes::_monitorexit    : monitorexit (apop()); break;
      case Bytecodes::_wide           : ShouldNotReachHere(); break;
      case Bytecodes::_multianewarray : new_multi_array(s.bcp()[3]); break;
      case Bytecodes::_ifnull         : if_null(objectType, If::eql); break;
      case Bytecodes::_ifnonnull      : if_null(objectType, If::neq); break;
      case Bytecodes::_goto_w         : append(new Goto(block_at(s.dest_w()))); break;
      case Bytecodes::_jsr_w          : jsr(s.dest_w()); break;
      default                         : ShouldNotReachHere(); break;
    }
    // save current bci to setup Goto at the end
    prev_bci = s.bci();
  }
  if (bailed_out()) return NULL;

  // if there are any, check if last instruction is a BlockEnd instruction
  BlockEnd* end = last()->as_BlockEnd();
  if (end == NULL) {
    // all blocks must end with a BlockEnd instruction => add a Goto
    end = new Goto(block_at(s.bci()));
    _last = _last->set_next(end, prev_bci);
  }
  assert(end == last()->as_BlockEnd(), "inconsistency");

  // if the method terminates, we don't need the stack anymore
  if (end->as_Return() != NULL || end->as_Throw() != NULL) state()->clear_stack();
  // here all expression stack values must be pinned
  // note: this only involves values that are inputs
  //       to phi nodes of other blocks - we must be
  //       sure we compute them before we compute the
  //       last expression of the block, usually a
  //       compare instruction
  state()->pin_stack_all();

  // connect to beg & set state
  beg->set_end(end);
  end->set_state(state());
  // propagate state
  for (int i = end->number_of_sux() - 1; i >= 0; i--) {
    // be careful, bailout if bytecodes are strange
    if (!end->sux_at(i)->try_join(state())) BAILOUT_("block join failed", NULL);
  }

  // done
  return end;
}


void GraphBuilder::build_transitive_closure(BlockBegin* b) {
  assert(!bailed_out(), "should not call this in case of previous bailout");
  if (!b->is_set(BlockBegin::was_visited_flag)) {
    b->set(BlockBegin::was_visited_flag);
    BlockEnd* e = connect_to_end(b);
    if (!bailed_out()) {
      for (int i = e->number_of_sux () - 1; i >= 0 && !bailed_out(); i--) {
        build_transitive_closure(e->sux_at(i));
      }
    }
  }
}


bool GraphBuilder::_is_initialized = false;
bool GraphBuilder::_can_trap      [Bytecodes::number_of_java_codes];
bool GraphBuilder::_is_async[Bytecodes::number_of_java_codes];

void GraphBuilder::initialize() {
  // make sure initialization happens only once (need a
  // lock here, if we allow the compiler to be re-entrant)
  if (is_initialized()) return;
  _is_initialized = true;

  // the following bytecodes are assumed to potentially
  // throw exceptions in compiled code - note that e.g.
  // monitorexit & the return bytecodes do not throw
  // exceptions since monitor pairing proved that they
  // succeed (if monitor pairing succeeded)
  Bytecodes::Code can_trap_list[] =
    { Bytecodes::_ldc
    , Bytecodes::_ldc_w
    , Bytecodes::_ldc2_w
    , Bytecodes::_iaload
    , Bytecodes::_laload
    , Bytecodes::_faload
    , Bytecodes::_daload
    , Bytecodes::_aaload
    , Bytecodes::_baload
    , Bytecodes::_caload
    , Bytecodes::_saload
    , Bytecodes::_iastore
    , Bytecodes::_lastore
    , Bytecodes::_fastore
    , Bytecodes::_dastore
    , Bytecodes::_aastore
    , Bytecodes::_bastore
    , Bytecodes::_castore
    , Bytecodes::_sastore
    , Bytecodes::_idiv
    , Bytecodes::_ldiv
    , Bytecodes::_irem
    , Bytecodes::_lrem
    , Bytecodes::_getstatic
    , Bytecodes::_putstatic
    , Bytecodes::_getfield
    , Bytecodes::_putfield
    , Bytecodes::_invokevirtual
    , Bytecodes::_invokespecial
    , Bytecodes::_invokestatic
    , Bytecodes::_invokeinterface
    , Bytecodes::_new
    , Bytecodes::_newarray
    , Bytecodes::_anewarray
    , Bytecodes::_arraylength
    , Bytecodes::_athrow
    , Bytecodes::_checkcast
    , Bytecodes::_instanceof
    , Bytecodes::_monitorenter
    , Bytecodes::_multianewarray
    };

  // the following bytecodes are assumed to potentially
  // throw asynchronous exceptions in compiled code due
  // to safepoints (note: these entries could be merged
  // with the can_trap_list - however, we need to know
  // which ones are asynchronous for now - see also the
  // comment in GraphBuilder::handle_exception)
  Bytecodes::Code is_async_list[] =
    { Bytecodes::_ifeq
    , Bytecodes::_ifne
    , Bytecodes::_iflt
    , Bytecodes::_ifge
    , Bytecodes::_ifgt
    , Bytecodes::_ifle
    , Bytecodes::_if_icmpeq
    , Bytecodes::_if_icmpne
    , Bytecodes::_if_icmplt
    , Bytecodes::_if_icmpge
    , Bytecodes::_if_icmpgt
    , Bytecodes::_if_icmple
    , Bytecodes::_if_acmpeq
    , Bytecodes::_if_acmpne
    , Bytecodes::_goto
    , Bytecodes::_jsr
    , Bytecodes::_ret
    , Bytecodes::_tableswitch
    , Bytecodes::_lookupswitch
    , Bytecodes::_ireturn
    , Bytecodes::_lreturn
    , Bytecodes::_freturn
    , Bytecodes::_dreturn
    , Bytecodes::_areturn
    , Bytecodes::_return
    , Bytecodes::_ifnull
    , Bytecodes::_ifnonnull
    , Bytecodes::_goto_w
    , Bytecodes::_jsr_w
    };

  // inititialize trap tables
  for (int i = 0; i < Bytecodes::number_of_java_codes; i++) {
    _can_trap[i] = false;
    _is_async[i] = false;
  }
  // set standard trap info
  for (int j = 0; j < sizeof(can_trap_list) / sizeof(can_trap_list[0]) ; j++) {
    _can_trap[can_trap_list[j]] = true;
  }
  // set asynchronous trap info
  for (int k = 0; k < sizeof(is_async_list) / sizeof(is_async_list[0]) ; k++) {
    assert(!_can_trap[is_async_list[k]], "can_trap_list and is_async_list should be disjoint");
    _can_trap[is_async_list[k]] = true;
    _is_async[is_async_list[k]] = true;
  }
}


GraphBuilder::GraphBuilder(Compilation* compilation, IRScope* scope, BlockList* bci2block, BlockBegin* start) {
  assert(is_initialized(), "GraphBuilder must have been initialized");
  _compilation = compilation;
  _scope       = scope;
  _bci2block   = bci2block;
  _vmap        = new ValueMap(29); // this should really be dynamically computed somehow
  _has_handler = xhandlers()->has_handlers();
  _found_jsr   = false;
  _bailout_msg = NULL;
  build_transitive_closure(start);
}

