// @(#)c1_IR.cpp	1.84 00/04/17 17:21:32
// 
// Copyright 1993-1999 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


# include "incls/_precompiled.incl"
# include "incls/_c1_IR.cpp.incl"


// Implementation of XHandlers
//
// Note: This code could eventually go away if we are
//       just using the ciExceptionHandlerStream.

XHandlers::XHandlers(ciMethod* method) : _list(method->exception_table_length()) {
  ciExceptionHandlerStream s(method);
  while (!s.is_done()) {
    _list.append(new XHandler(s.handler()));
    s.next();
  }
  assert(s.count() == method->exception_table_length(), "exception table lengths inconsistent");
}


// Implementation of IRScope

BlockBegin* IRScope::header_block(BlockBegin* entry, BlockBegin::Flag f) {
  if (entry == NULL) return NULL;
  assert(entry->is_set(f), "entry/flag mismatch");
  // create header block
  BlockBegin* h = new BlockBegin(entry->scope(), entry->bci());
  BlockEnd* g = new Goto(entry);
  h->set_next(g, entry->bci());
  h->set_end(g);
  h->set(f);
  // setup header block end state
  ValueStack* s = entry->state()->copy(); // can use copy since stack is empty (=> no phis)
  assert(s->stack_is_empty(), "must have empty stack at entry point");
  g->set_state(s);
  return h;
}


BlockBegin* IRScope::build_graph(Compilation* compilation, int osr_bci) {
  // determine entry points and bci2block mapping
  BlockListBuilder blb(this, osr_bci);
  assert(osr_bci < 0 || blb.osr_entry() != NULL, "osr entry must exist for osr compile");
  // create & setup initial state for std entry
  { ValueStack* state = new ValueStack(method()->max_locals(), method()->max_stack());
    if (method()->is_synchronized()) state->lock(this, SynchronizationEntryBCI);
    blb.std_entry()->set_state(state);
  }
  // complete graph
  GraphBuilder gb(compilation, this, blb.bci2block(), blb.std_entry());
  if (gb.bailed_out()) return NULL;
  assert(osr_bci < 0 || blb.osr_entry()->is_set(BlockBegin::was_visited_flag), "osr entry must have been visited for osr compile");
  // for osr compile, bailout if some requirements are not fulfilled
  if (osr_bci >= 0) {
    // check if method contains jsr's and bailout if we don't accept them
    if (gb.found_jsr() && !AcceptJsrForOSR) return NULL; // bailout
    // check if osr entry point has empty stack - we cannot handle non-empty stacks at osr entry points
    if (!blb.osr_entry()->state()->stack_is_empty()) return NULL; // bailout
  }
  // setup start block (root for the IR graph)
  BlockBegin* start = new BlockBegin(this, 0);
  { Base* base =
      new Base(
        header_block(blb.std_entry(), BlockBegin::std_entry_flag),
        header_block(blb.osr_entry(), BlockBegin::osr_entry_flag)
      );
    start->set_next(base, 0);
    start->set_end(base);
    // create & setup state for start block
    ValueStack* state = new ValueStack(method()->max_locals(), method()->max_stack());
    start->set_state(state);
    base->set_state(state->copy());
    // setup states for header blocks
    base->std_entry()->join(state);
    if (osr_bci >= 0) base->osr_entry()->join(state);
  }
  // done
  return start;
}


IRScope::IRScope(Compilation* compilation, IRScope* caller, ciMethod* method, int osr_bci) : _callees(2) {
  _level              = caller == NULL ? 0 : caller->level() + 1;
  _caller             = caller;
  _method             = method;
  _xhandlers          = new XHandlers(method);
  _number_of_locks    = 0;
  _monitor_pairing_ok = method->has_balanced_monitors();
  _start              = NULL;
  // build graph if monitor pairing is ok
  if (monitor_pairing_ok()) _start = build_graph(compilation, osr_bci);
}


// Implementation of AccessorInfo

#ifndef PRODUCT
void AccessorInfo::print() const {
  tty->print_cr("accessor info: %s @ %d", InstructionPrinter::basic_type_name(type()), offset());
}
#endif // PRODUCT


// Implementation of IR

IR::IR(Compilation* compilation, ciMethod* method, int osr_bci) {
  // initialize data structures
  ValueType::initialize();
  Instruction::initialize();
  BlockBegin::initialize();
  GraphBuilder::initialize();
  // setup IR fields
  _compilation = compilation;
  _top_scope   = new IRScope(compilation, NULL, method, osr_bci);
  _loops       = NULL;
  _code        = NULL;
}


void IR::optimize() {
  Optimizer opt(this);
  if (DoCEE) {
    opt.eliminate_conditional_expressions();
    if (PrintCFG || PrintCFG1) { tty->print_cr("CFG after CEE"); print(true); }
    if (PrintIR  || PrintIR1 ) { tty->print_cr("IR after CEE"); print(false); }
  }
  if (EliminateBlocks) {
    opt.eliminate_blocks();
    if (PrintCFG || PrintCFG1) { tty->print_cr("CFG after block elimination"); print(true); }
    if (PrintIR  || PrintIR1 ) { tty->print_cr("IR after block elimination"); print(false); }
  }
}


void IR::compute_loops() {
  if (ComputeLoops && method()->has_loops()) {
    LoopFinder l(this, BlockBegin::number_of_blocks());
    _loops = l.compute_loops();
  }
}


void IR::iterate_and_set_weight(GrowableArray<bool>& mark, BlockBegin* b, BlockClosure* closure, int d) const {
  if (!mark.at(b->block_id())) {
    mark.at_put(b->block_id(), true);
    b->set_weight(d);
    closure->block_do(b);
    BlockEnd* e = b->end();
    { for (int i = b->number_of_exception_handlers() - 1; i >= 0; i--) iterate_and_set_weight(mark, b->exception_handler_at(i), closure, d + 1); }
    { for (int i = b->number_of_subroutines       () - 1; i >= 0; i--) iterate_and_set_weight(mark, b->subroutine_at       (i), closure, d + 1); }
    { for (int i = e->number_of_sux               () - 1; i >= 0; i--) iterate_and_set_weight(mark, e->sux_at              (i), closure, d + 1); }
  }
}


class UseCountComputer: public AllStatic {
 private:
  static void update_use_count(Value* n) {
    // use n's input if not visited before
    if (!(*n)->is_pinned() && !(*n)->has_uses()) {
      // note: a) if the instruction is pinned, it will be handled by compute_use_count
      //       b) if the instruction has uses, it was touched before
      //       => in both cases we don't need to update n's values
      uses_do(n);
    }
    // use n
    (*n)->_use_count++;
  }

  static void uses_do(Value* n) {
    (*n)->input_values_do(update_use_count);
    // special handling for some instructions
    if ((*n)->as_Jsr() != NULL || (*n)->as_BlockEnd() != NULL) {
      // note on Jsr:
      //   must 'use' the state since stack values may be
      //   used in the subroutine (was bug - gri 5/25/99)
      // note on BlockEnd:
      //   must 'use' the stack only if the method doesn't
      //   terminate, however, in those cases stack is empty
      (*n)->state_values_do(update_use_count);
    }
  }

  static void basic_compute_use_count(BlockBegin* b) {
    for (Instruction* n = b; n != NULL; n = n->next()) {
      if (n->is_pinned()) uses_do(&n);
    }
  }

  static bool updated_pinning(BlockBegin* b) {
    bool updated = false;
    for (Instruction* n = b; n != NULL; n = n->next()) {
      if (!n->has_uses() && n->can_trap()) { n->pin(); updated = true; }
    }
    return updated;
  }

  static void clear_use_count(BlockBegin* b) {
    // clear use counts of phi instructions
    ValueStack* s = b->state();
    for (int i = 0; i < s->stack_size();) s->stack_at_inc(i)->_use_count = 0;
    // clear use counts of all other instructions
    for (Instruction* n = b; n != NULL; n = n->next()) n->_use_count = 0;
  }

 public:
  static void compute_use_count(BlockBegin* b) {
    // determine use count given the current pinning information
    // note: all use counts are cleared in the beginning
    basic_compute_use_count(b);
    // instructions that are never used but that can trap must
    // not be eliminated => make sure code is generated for'em
    // by pinning them explicitly (was bug - gri 11/10/99)
    if (updated_pinning(b)) {
      // set of pinned instructions was updated
      // => need to recompute use count information
      // note: must clear use counts first
      clear_use_count(b);
      basic_compute_use_count(b);
    }
  }
};


static int cmp(BlockBegin** a, BlockBegin** b)   { return (*a)->weight() - (*b)->weight(); }


class SuxAndWeightAdjuster: public BlockClosure {
 private:
  BlockBegin* _prev;

 public:
  SuxAndWeightAdjuster()                          { _prev = NULL; }

  virtual void block_do(BlockBegin* block) {
    if (_prev == NULL) {
      // set weight for first block
      block->_weight = 0;
    } else {
      // set weight for current block
      block->_weight = _prev->weight() + 1;
      // swap successors if default successor is not the block immediately following
      BlockEnd* end = _prev->end();
      if (end->number_of_sux() == 2 && end->default_sux() != block && end->sux_index(block) >= 0) {
             if (end->as_If          () != NULL) end->as_If          ()->swap_sux();
        else if (end->as_IfInstanceOf() != NULL) end->as_IfInstanceOf()->swap_sux();
      }
    }
    _prev = block;
  }
};


void IR::compute_code() {
  assert(is_valid(), "IR must be valid");
  const int n = BlockBegin::number_of_blocks();
  // collect all blocks and set weight depending on distance from start
  // note: we need a special iterator because the standard onces don't
  //       compute the distance from start.
  BlockCollection* c = new BlockCollection(n);
  { GrowableArray<bool> mark(n, n, false);
    iterate_and_set_weight(mark, start(), c, 0);
  }
  // now we have all blocks => sort them according to weight
  c->collection()->sort(cmp);
  // swap successors of If's if default successor is not the block immediately following
  // and adjust weights so they are increasing by 1 and starting with 0 for first block
  { SuxAndWeightAdjuster swa;
    c->iterate_forward(&swa);
  }
  // compute use counts
  c->blocks_do(UseCountComputer::compute_use_count);
  // done
  _code = c;
}


void IR::iterate_preorder(BlockClosure* closure) {
  assert(is_valid(), "IR must be valid");
  start()->iterate_preorder(closure);
}


void IR::iterate_postorder(BlockClosure* closure) {
  assert(is_valid(), "IR must be valid");
  start()->iterate_postorder(closure);
}


void IR::iterate_topological(BlockClosure* closure) {
  assert(is_valid(), "IR must be valid");
  start()->iterate_topological(closure);
}


AccessorInfo* IR::accessor_info() const {
  assert(is_valid(), "IR must be valid");
  // exclude OSR compilations
  if (osr_entry() != NULL) return NULL;
  // exclude graphs with more than one block (besides base block)
  BlockBegin* b = std_entry();
  assert(b != NULL, "std entry must exist");
  if (b->number_of_subroutines       () > 0) return NULL;
  if (b->number_of_exception_handlers() > 0) return NULL;
  if (b->end()->number_of_sux        () > 0) return NULL;
  // exclude basic blocks with the wrong instructions
  // 1) make sure 1st instruction loads receiver
  LoadLocal* recv = b->next()->as_LoadLocal();
  if (recv == NULL || recv->index() != 0) return NULL;
  // 2) make sure 2nd instruction is load field
  LoadField* load = recv->next()->as_LoadField();
  if (load == NULL || load->obj() != recv || !load->is_loaded() || !load->is_initialized()) return NULL;
  assert(!load->is_static(), "must not be static since obj == recv");
  // 3) make sure 3rd instruction is return
  Return* ret = load->next()->as_Return();
  if (ret == NULL || ret->result() != load) return NULL;
  assert(ret == b->end(), "return should be the block end");
  // we have an accessor function
  return new AccessorInfo(load->field_type(), load->offset());
}


#ifndef PRODUCT
class BlockPrinter: public BlockClosure {
 private:
  InstructionPrinter* _ip;
  bool                _cfg_only;

 public:
  BlockPrinter(InstructionPrinter* ip, bool cfg_only) {
    _ip       = ip;
    _cfg_only = cfg_only;
  }

  virtual void block_do(BlockBegin* block) {
    if (_cfg_only) {
      _ip->print_instr(block); tty->cr();
    } else {
      block->print_block();
    }
  }
};


void IR::print(bool cfg_only) {
  if (is_valid()) {
    InstructionPrinter ip;
    BlockPrinter bp(&ip, cfg_only);
    iterate_topological(&bp);
  } else {
    tty->print_cr("invalid IR");
  }
  tty->cr();
}


void IR::verify() {
  Unimplemented();
}
#endif // PRODUCT

