// @(#)c1_IR.hpp	1.51 00/04/17 17:21:33
// 
// Copyright 1993-1999 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


// An XHandler is a C1 internal description for an exception handler

class XHandler: public ResourceObj {
 private:
  ciExceptionHandler* _desc;
  BlockBegin*         _entry;

 public:
  // creation
  XHandler(ciExceptionHandler* desc) : _desc(desc), _entry(NULL) {}
  
  void set_entry(BlockBegin* entry) {
    assert(entry->is_set(BlockBegin::exception_entry_flag), "must be an exception handler entry");
    assert(entry->bci() == handler_bci(), "bci's must correspond");
    _entry = entry;
  }

  // accessors
  int beg_bci() const                            { return _desc->start(); }
  int end_bci() const                            { return _desc->limit(); }
  int handler_bci() const                        { return _desc->handler_bci(); }
  int catch_type() const                         { return _desc->catch_klass_index(); }
  BlockBegin* entry() const                      { return _entry; }
  bool covers(int bci) const                     { return beg_bci() <= bci && bci < end_bci(); }
};


// XHandlers is the C1 internal list of exception handlers for a method

class XHandlers: public ResourceObj {
 private:
  GrowableArray<XHandler*> _list;

 public:
  // creation
  XHandlers(ciMethod* method);

  // accessors
  int number_of_handlers() const                 { return _list.length(); }
  XHandler* handler_at(int i) const              { return _list.at(i); }
  bool has_handlers() const                      { return number_of_handlers() > 0; }
};


class Compilation;

class IRScope: public ResourceObj {
 private:
  // hierarchy
  int                     _level;                // the inlining level
  IRScope*                _caller;               // the caller scope
  ciMethod*               _method;               // the corresponding method
  GrowableArray<IRScope*> _callees;              // the inlined method scopes

  // graph
  XHandlers*              _xhandlers;            // the exception handlers
  int                     _number_of_locks;      // the number of monitor lock slots needed
  bool                    _monitor_pairing_ok;   // the monitor pairing info
  BlockBegin*             _start;                // the start block, successsors are method entries

  // helper functions
  BlockBegin* header_block(BlockBegin* entry, BlockBegin::Flag f);
  BlockBegin* build_graph(Compilation* compilation, int osr_bci);

 public:
  // creation
  IRScope(Compilation* compilation, IRScope* caller, ciMethod* method, int osr_bci);

  // accessors
  int           level() const                    { return _level; }
  IRScope*      caller() const                   { return _caller; }
  ciMethod*     method() const                   { return _method; }

  // hierarchy
  void          add_callee(IRScope* callee)      { _callees.append(callee); }
  int           number_of_callees() const        { return _callees.length(); }
  IRScope*      callee_no(int i) const           { return _callees.at(i); }

  // accessors, graph
  bool          is_valid() const                 { return start() != NULL; }
  XHandlers*    xhandlers() const                { return _xhandlers; }
  int           number_of_locks() const          { return _number_of_locks; }
  void          set_min_number_of_locks(int n)   { if (n > _number_of_locks) _number_of_locks = n; }
  bool          monitor_pairing_ok() const       { return _monitor_pairing_ok; }
  BlockBegin*   start() const                    { return _start; }
};


class AccessorInfo: public ResourceObj {
 private:
  BasicType _type;
  int       _offset;

 public:
  // creation
  AccessorInfo(BasicType type, int offset) {
    assert(type != T_ILLEGAL, "type must be valid");
    assert(0 <= offset, "offset must be positive");
    _type   = type;
    _offset = offset;
  }

  // accessors
  BasicType type  () const                       { return _type; }
  int       offset() const                       { return _offset; }

  // debugging/printing
  void print() const                             PRODUCT_RETURN;
};


class Loop;
typedef GrowableArray<Loop*> LoopList;

class IR: public ResourceObj {
 private:
  Compilation*     _compilation;                 // the current compilation
  IRScope*         _top_scope;                   // the root of the scope hierarchy
  LoopList*        _loops;                       // the 'relevant' loops of the IR
  BlockCollection* _code;                        // the blocks in code generation order w/ use counts

  void iterate_and_set_weight(GrowableArray<bool>& mark, BlockBegin* b, BlockClosure* closure, int d) const;

 public:
  // creation
  IR(Compilation* compilation, ciMethod* method, int osr_bci);

  // accessors
  bool             is_valid() const              { return top_scope()->is_valid(); }
  Compilation*     compilation() const           { return _compilation; }
  IRScope*         top_scope() const             { return _top_scope; }
  int              number_of_locks() const       { return top_scope()->number_of_locks(); }
  ciMethod*        method() const                { return top_scope()->method(); }
  BlockBegin*      start() const                 { return top_scope()->start(); }
  BlockBegin*      std_entry() const             { return start()->end()->as_Base()->std_entry(); }
  BlockBegin*      osr_entry() const             { return start()->end()->as_Base()->osr_entry(); }
  LoopList*        loops() const                 { return _loops; }
  BlockCollection* code() const                  { return _code; }

  // ir manipulation
  void optimize();
  void compute_loops();
  void compute_code();

  // iteration
  void iterate_preorder   (BlockClosure* closure);
  void iterate_postorder  (BlockClosure* closure);
  void iterate_topological(BlockClosure* closure);

  // testers
  AccessorInfo* accessor_info() const;           // returns NULL if function is not an accessor

  // debugging
  void print(bool cfg_only)                      PRODUCT_RETURN;
  void verify()                                  PRODUCT_RETURN;
};

