/*
 * @(#)c1_Inliner.cpp	1.51 00/04/26 21:33:29
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_c1_Inliner.cpp.incl"


#define BAILOUT(msg) { bailout(msg); return false; }


void Inliner::bailout(const char* msg) {
  assert(msg != NULL, "bailout msg must exist");
  _bailout_msg = msg;
  if (PrintInlining) tty->print_cr("%s: cannot inline", msg);
}


bool Inliner::try_inline_intrinsics(ciMethod* callee) {
  if (!InlineIntrinsics        ) BAILOUT("intrinsic method inlining disabled");
  if (callee->is_synchronized()) BAILOUT("intrinsic method is synchronized");
  // callee seems like a good candidate
  // determine id
  ciMethod::IntrinsicId id = callee->intrinsic_id();

#ifdef SPARC
  // As soon as arraycopy inlining is implemented on sparc we should get
  // rid of this
  NEEDS_CLEANUP
  switch (id) {
    case ciMethod::_arraycopy: if (!InlineArrayCopy) return false;  // fall through or return false
    case ciMethod::_dsin     : // fall through
    case ciMethod::_dcos     : // fall through
    case ciMethod::_dsqrt    : break;
    default                  : return false; // do not inline
  }
#else
  switch (id) {
    case ciMethod::_dsin     : // fall through
    case ciMethod::_dcos     : // fall through
    case ciMethod::_dsqrt    : // fall through
    case ciMethod::_arraycopy: break;        // do     inline
    default                  : return false; // do not inline
  }
#endif
  // create intrinsic node
  const bool has_receiver = !callee->is_static();
  ValueType* result_type = as_ValueType(callee->return_type());
  Values* args = state()->pop_arguments((has_receiver ? 1 : 0) + callee->arg_size());
  Intrinsic* result = new Intrinsic(result_type, id, args);
  // add null check if necessary
  if (has_receiver) gb()->append(new NullCheck(args->at(0)));
  // append instruction & push result
  gb()->append_split(result);
  if (result_type != voidType) gb()->push(result_type, result);
  // done
  return true;
}


class InstructionFilter: public InstructionVisitor {
 private:
  ValueStack* _state;
  int         _args_size;
  int         _args_base;
  const char* _not_ok_msg;

  void not_ok(const char* msg)                     { assert(msg != NULL, "msg must exist"); _not_ok_msg = msg; }

  void substitute_load(LoadLocal* x) {
    assert(0 <= x->index() && x->index() < _args_size, "index out of bounds");
    int i = _args_base + x->index();
    x->set_subst(_state->stack_at_inc(i));
  }

  void check_field_access(AccessField* x) {
    if (!x->is_loaded() || !x->is_initialized()) not_ok("field access");
  }

 public:
  // creation
  InstructionFilter(ValueStack* state, int args_size)
  : _state(state)
  , _args_size(args_size)
  , _args_base(state->stack_size() - args_size)
  , _not_ok_msg(NULL)
  {
    assert(state != NULL, "state must exist");
    assert(state->stack_size() >= args_size, "stack too small");
  }

  // application
  void apply(Instruction* list) {
    for (; list != NULL; list = list->next()) list->visit(this);
  }

  // accessors
  int  args_base() const                              { return _args_base; }
  Value receiver() const                              { int i = _args_base; return _state->stack_at_inc(i); }
  const char* not_ok_msg() const                      { return _not_ok_msg; }
  bool ok() const                                     { return not_ok_msg() == NULL; }

  // Note: Generally, we cannot inline any method that contains code that
  //       may cause GC or exception handling (or goes into a potentially
  //       blocking runtime call for that matter) since we don't have the
  //       correct debug information for these inlined methods! Eventually
  //       we need to fix this by providing correct debug information for
  //       inlined methods (was bug - gri 10/13/99, in particular must ex-
  //       clude Instructions that can have a slow-case such as Intrinsic
  //       or that can throw an exception such as NullCheck).

  virtual void do_Phi            (Phi*             x) {}
  virtual void do_Constant       (Constant*        x) {}
  virtual void do_LoadLocal      (LoadLocal*       x) { substitute_load(x); }
  virtual void do_StoreLocal     (StoreLocal*      x) { not_ok("StoreLocal"); }
  virtual void do_LoadField      (LoadField*       x) { check_field_access(x); }
  virtual void do_StoreField     (StoreField*      x) { check_field_access(x); }
  virtual void do_ArrayLength    (ArrayLength*     x) {}
  virtual void do_LoadIndexed    (LoadIndexed*     x) {} // incorrect stack dumps if index out of bounds (minor)
  virtual void do_StoreIndexed   (StoreIndexed*    x) {} // incorrect stack dumps if index out of bounds (minor)
  virtual void do_LoopEnter      (LoopEnter*       x) { ShouldNotReachHere(); }
  virtual void do_LoopExit       (LoopExit*        x) { ShouldNotReachHere(); }
  virtual void do_NegateOp       (NegateOp*        x) {}
  virtual void do_ArithmeticOp   (ArithmeticOp*    x) {}
  virtual void do_ShiftOp        (ShiftOp*         x) {}
  virtual void do_LogicOp        (LogicOp*         x) {}
  virtual void do_CompareOp      (CompareOp*       x) {}
  virtual void do_IfOp           (IfOp*            x) {}
  virtual void do_Intrinsic      (Intrinsic*       x) { not_ok("Intrinsic"); }
  virtual void do_Convert        (Convert*         x) {}
  virtual void do_JsrContinuation(JsrContinuation* x) { not_ok("JsrContinuation"); }
  virtual void do_NullCheck      (NullCheck*       x) { not_ok("NullCheck"); }
  virtual void do_Jsr            (Jsr*             x) { not_ok("Jsr"); }
  virtual void do_Invoke         (Invoke*          x) { not_ok("Invoke"); }
  virtual void do_NewInstance    (NewInstance*     x) { not_ok("NewInstance"); }
  virtual void do_NewTypeArray   (NewTypeArray*    x) { not_ok("NewTypeArray"); }
  virtual void do_NewObjectArray (NewObjectArray*  x) { not_ok("NewObjectArray"); }
  virtual void do_NewMultiArray  (NewMultiArray*   x) { not_ok("NewMultiArray"); }
  virtual void do_CheckCast      (CheckCast*       x) { not_ok("CheckCast"); }
  virtual void do_InstanceOf     (InstanceOf*      x) { not_ok("InstanceOf"); }
  virtual void do_MonitorEnter   (MonitorEnter*    x) { not_ok("MonitorEnter"); }
  virtual void do_MonitorExit    (MonitorExit*     x) { not_ok("MonitorExit"); }
  virtual void do_BlockBegin     (BlockBegin*      x) {}
  virtual void do_Goto           (Goto*            x) { ShouldNotReachHere(); }
  virtual void do_If             (If*              x) { ShouldNotReachHere(); }
  virtual void do_IfInstanceOf   (IfInstanceOf*    x) { ShouldNotReachHere(); }
  virtual void do_TableSwitch    (TableSwitch*     x) { ShouldNotReachHere(); }
  virtual void do_LookupSwitch   (LookupSwitch*    x) { ShouldNotReachHere(); }
  virtual void do_Ret            (Ret*             x) { not_ok("Ret"); }
  virtual void do_Return         (Return*          x) {}
  virtual void do_Throw          (Throw*           x) { not_ok("Throw"); }
  virtual void do_Base           (Base*            x) {}
};


bool Inliner::try_inline_simple(ciMethod* callee) {
  assert(!callee->is_native(), "callee must not be native");
  if (!InlineSimpleMethods                      ) BAILOUT("simple method inlining disabled");
  if (callee->has_exception_handlers()          ) BAILOUT("callee has exception handlers");
  if (callee->max_locals() != callee->arg_size()) BAILOUT("callee has locals");
  if (callee->code_size() > MaxInlineSize       ) BAILOUT("callee is too large");
  if (callee->is_synchronized()                 ) BAILOUT("callee is synchronized");

  NEEDS_CLEANUP
  // It's not clear wether this test should apply to all attempts to
  // inline (instrinsic or not), or only to the simple case (not
  // intrinsic).
  // On i486: is it correct to inline sine, for example, regardless of
  // wether the caller is strict or not?
  if (strict_fp_requires_explicit_rounding && method()->is_strict() != callee->is_strict()) return false;

  // callee seems like a good candidate
  const int args_base = state()->stack_size() - callee->arg_size();
  Value result = NULL;

  // check for empty method first
  // Note: we need this for now since we don't have real inlinining -
  //       empty methods can always be inlined and thus will enable
  //       more simple inlining because no scope level test is per-
  //       formed for empty methods.
  if (callee->code_size() == 1 && *callee->code() == Bytecodes::_return) {
    // empty method returning void => always inline
    assert(callee->return_type() == T_VOID, "inconsistent result type");
    // insert null check if necessary
    if (code() != Bytecodes::_invokestatic) {
      assert(!callee->is_static(), "callee must not be static");
      assert(callee->arg_size() > 0, "must have at least a receiver");
      int i = args_base;
      gb()->_last = gb()->_last->set_next(new NullCheck(state()->stack_at_inc(i)), bci());
    }

  } else {
    // method not empty
    if (scope()->level() > 0) BAILOUT("can only inline one level");
    // inlining possible => build subgraph
    IRScope* callee_scope = new IRScope(compilation(), scope(), callee, -1);
    assert(!callee_scope->xhandlers()->has_handlers(), "just checking");
    if (!callee_scope->is_valid()) BAILOUT("invalid IR (monitors may not match)");
    // find start & stop of method
    BlockBegin* start = NULL;
    Return*     stop  = NULL;
    { BlockBegin* s = callee_scope->start()->end()->as_Base()->std_entry();
      assert(s->next() == s->end(), "should be an empty block");
      assert(s->end()->as_Goto() != NULL, "should have a Goto at the end");
      start = s->end()->default_sux();
      stop  = start->end()->as_Return(); // generalize for Throw!
    }
    if (stop == NULL) BAILOUT("more than one block"); // cannot handle methods w/ more than one block (besides the start block)
    assert(stop->type()->tag() == as_ValueType(callee->return_type())->tag(), "types must match");
    assert(stop->prev(start), "make sure this works");
    // apply InstructionFilter
    assert(start->state()->stack_size() == 0, "just checking");
    InstructionFilter ifilter(state(), callee->arg_size());
    ifilter.apply(start->next());
    if (!ifilter.ok()) BAILOUT(ifilter.not_ok_msg());
    // method ok for inlining, LoadLocals subst'ed
    // resolve substitution
    start->resolve_substitution();
    // insert null check if necessary
    if (code() != Bytecodes::_invokestatic) {
      assert(!callee->is_static(), "callee must not be static");
      // no null check needed if null check is done implicitly
      // by the first instruction => just check for getfield
      // first (i.e., the first instruction is going to be a
      // LoadLocal, but it is not used anymore due to rewriting -
      // we don't really care about is as long as it doesn't cause
      // an exception)
      LoadLocal* loadrecv = start->next()->as_LoadLocal();
      LoadField* getfield = loadrecv == NULL ? NULL : loadrecv->next()->as_LoadField();
      if (getfield != NULL && getfield->obj() == ifilter.receiver()) {
        // next instruction is a matching getfield => no explicit null check needed
        assert(!getfield->is_static(), "must be getfield - otherwise recv couldn't match");
      } else {
        // explicit null check needed
        gb()->_last = gb()->_last->set_next(new NullCheck(ifilter.receiver()), bci());
      }
    }
    // append instructions of callee to current position & use same bci
    // (this code needs a rewrite - we need to provide a decent abstraction in BlockBegin or so...)
    { Instruction* last = gb()->_last;
      last->set_next(start->next(), bci());
      while (last->next() != stop) { last = last->next(); last->set_bci(bci()); } // must terminate!
      last->set_next(NULL, 0);
      gb()->_last = last;
    }
    // determine result if any
    if (callee->return_type() != T_VOID) result = stop->result();
  }

  // make sure all args are computed before entering the (inlined) callee
  { int i = args_base;
    while (i < state()->stack_size()) state()->stack_at_inc(i)->pin();
  }

  // make sure result is computed before leaving the (inlined) callee
  if (result != NULL) result->pin();

  // remove args from stack and push result if any
  state()->truncate_stack(args_base);
  if (callee->return_type() != T_VOID) gb()->push(result->type(), result);

  // printing
  if (PrintInlining) {
    for (int i = scope()->level(); i > 0; i--) tty->print("  ");
    tty->print_cr("*** inlined method @ %d (%d bytes)", bci(), callee->code_size());
  }

  // done
  return true;
}


bool Inliner::try_inline(ciMethod* callee) {
  if (compilation()->jvmpi_event_method_enabled()) {
    // do not inline at all
    return false;
  } else if (try_inline_intrinsics(callee)) {
    // intrinsics can be native or not
    return true;
  } else if (callee->is_native()) {
    // non-intrinsic natives cannot be inlined 
    return false;
  } else {
    return try_inline_simple(callee);
  } 
}

