// @(#)c1_Items.hpp	1.30 00/04/20 16:12:30
// 
// Copyright 1993-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2

// The classes responsible for code emission and register allocation


// class hierarchy:
//    Item
//      HintItem (instr = NULL, type explicit)
//      RootItem (asigned to root instructions, not stack allocated)

class Item;
class   HintItem;
class   RootItem;


class Item: public ResourceObj {
 protected:
  enum ItemMode {
    regMode,
    stackMode,
    constMode,
    noResult,
    noMode
  };
  bool       _cached;          // true if the item contains a cached register 
  bool       _round32;         // true if the item must be in 32 bit precision
  bool       _inreg;           // true if Item must be in reg-mode (temporary fix for SPARC)
  ItemMode   _mode;            // see enumerator
  int        _loc;             // hold RInfo or stack slot index
  Value      _instr;           // instr specifies the type of item
  enum {
    not_spilled = -1
  };
  int       _spill_ix;           // < 0 if not spilled
  bool      _destroys_register;  // true if the register specified by this item will be destroyed

 public:
  Item():
     _instr(NULL), 
     _mode(noMode), 
     _spill_ix(not_spilled), 
     _destroys_register(false), 
     _loc(0),
     _round32(false), _cached(false), _inreg(false) {}

  Item(Value instr): 
     _instr(instr), 
     _mode(noMode), 
     _spill_ix(not_spilled), 
     _destroys_register(false), 
     _loc(0),
     _round32(false), _cached(false), _inreg(false) {
    assert(instr != NULL, "");
    if (!instr->is_root()) {
      assert(instr->item() == NULL, "already set");
      instr->set_item(this);
    }
  }

  ~Item() {
    if (_instr != NULL) { // hints have NULL instructions
      if (!_instr->is_root()) {
        assert(_instr->item() == this, "wrong instr-item pairing");
        _instr->clear_item();
      }
    }
  }

  // testers
  bool is_root_item() { return as_root() != NULL; }
  bool is_hint_item() { return as_hint() != NULL; }

  virtual HintItem* as_hint() { return NULL;  }
  virtual RootItem* as_root() { return NULL;  }

  virtual bool is_same(Item* item); // slow comparision

  bool is_register () const { return _mode == regMode && _spill_ix == not_spilled;   }
  bool is_stack    () const { return _mode == stackMode || _spill_ix != not_spilled; }
  bool is_constant () const { return _mode == constMode;       }
  bool is_spilled  () const { return _spill_ix != not_spilled; }
  bool has_result  () const { return _mode != noMode; }
  bool has_hint    () const { return _mode != noMode; }
  bool is_valid    () const { return _mode != noMode; }

  // getters
  virtual ValueType* type() const { return _instr->type(); }
  bool  is_round32  ()   const { return _round32;        }
  bool  is_inreg    ()   const { return _inreg;          }
  bool  is_cached   ()   const { return _cached;         }
  RInfo get_register()   const { assert(is_register(),""); return RInfo(_loc, 'a'); }
  int   get_stack   ()   const { assert(is_stack(),   ""); return _loc; }

  ciObject* get_jobject_constant() const;
  jint      get_jint_constant() const;
  jlong     get_jlong_constant() const;
  jfloat    get_jfloat_constant() const;
  jdouble   get_jdouble_constant() const;
  jint      get_address_constant() const;

  Value value()             const { return _instr;    }
  int   get_spilled_index() const { return _spill_ix; }

  // setters
  void set_round32 ()        { _round32 = true; }
  void set_inreg(bool f)     { _inreg  = f;     }

  virtual void set_from_item(const Item* item) { 
    _mode = item->_mode; _loc = item->_loc; _spill_ix = item->_spill_ix;
    _round32 = item->_round32;
    _inreg = item->_inreg;
    _instr = item->value();
    _cached = item->_cached;
  }

  void set_register(RInfo reg, bool cached = false) { 
    _mode = regMode;   _loc = reg.number(); _spill_ix = not_spilled; 
    _cached = cached;
  }

  void set_stack   (int index) { _mode = stackMode; _loc = index; _spill_ix = not_spilled; }
  void set_constant()          { _mode = constMode; _spill_ix = not_spilled;}

  void set_no_result ()             { _mode = noMode; }
  void set_spill_ix  (int spill_ix) { _mode = regMode; _spill_ix = spill_ix; }

  void set_destroys_register()      { _destroys_register = true; }
  bool destroys_register() const    { return _destroys_register; }

  // The item stored in root is more up to date: copy state
  // of that item to yourself
  void update();

  // sets _destroys flag if item is float or double
  void handle_float_kind();

  void print() PRODUCT_RETURN;
};


class HintItem: public Item {
 private:
  static HintItem _no_hint;
  ValueType*      _type;
 public:
  HintItem(ValueType* type, bool is_no_hint=false) : Item(), _type(type) { 
    assert(is_no_hint || type != NULL, "type must be set");
  }

  HintItem(ValueType* type, const RInfo reg, bool cached = false) : Item(), _type(type) {
    assert(reg.is_valid(), "invalid register");
    set_register(reg, cached);
  }

  virtual HintItem*  as_hint()    { return this;  }
  virtual ValueType* type() const { return _type; }

  virtual bool is_same(Item* item);
  virtual void set_from_item(const Item* item);

  static HintItem* no_hint()   { return &_no_hint;   }
};


class RootItem: public Item {
 public:
  RootItem(Value instr): Item(instr) { }
  virtual RootItem* as_root() { return this; }
};


class BlockItem: public ResourceObj {
 public:
  const RInfoCollection* _mapping;
  GrowableArray<bool>*   _is_oop; // for oop maps
  c1_RegMask             _lockout;
  bool                   _32bit_precision;
 public: 
  BlockItem(const RInfoCollection* map, GrowableArray<bool>* oops, c1_RegMask lockout, bool is32bit): 
       _mapping(map), _is_oop(oops), _lockout(lockout), _32bit_precision(is32bit) { }
  RInfo map_local (int local_index) const;
  bool  is_oop    (int local_index) const;
  int        map_length ()          const  { return _mapping->number_of_registers(); }
  c1_RegMask lockout ()             const  { return _lockout; }
  bool       is_32bit_precision ()  const  { return _32bit_precision; }
  void       set_lockout (c1_RegMask l)    { _lockout = l; }
};

