// @(#)c1_Loops.cpp	1.30 00/03/21 16:16:35
// 
// Copyright 1993-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


# include "incls/_precompiled.incl"
# include "incls/_c1_Loops.cpp.incl"


//-------------------BlockLoopInfo----------------------------

BitMap* BlockLoopInfo::_all_blocks = NULL;

BlockLoopInfo::BlockLoopInfo(BlockBegin* bb, int max_blocks): 
      _block(bb), _doms_map(max_blocks), _backedge_start(false), _loop_index(-1) {
  if (_all_blocks != NULL) {
    _doms_map.set_from(*_all_blocks);
  }
  _preds = new GrowableArray<BlockBegin*>();
}


void BlockLoopInfo::remove_predecessor(BlockBegin* pred) {
  debug_only(int lng = _preds->length();)
  _preds->remove(pred);
  assert(_preds->length() == lng - 1, "removal failed");
}


#ifndef PRODUCT
void BlockLoopInfo::print() {
  tty->print_cr("BlockLoopInfo for %d",  _block->block_id());
  tty->print ("  Preds [");
  int i;
  for (i = 0; i < _preds->length(); i++) {
    tty->print(" B%d ", _preds->at(i)->block_id());
  }
  tty->print_cr("]");
  tty->print ("  Doms  [");
  for (i = 0; i < _doms_map.size(); i++) {
    if (is_dom_block(i)) {
      tty->print(" B%d ", i);
    }
  }
  tty->print_cr("]");
}
#endif


//-------------------CheckDomClosure--------------------------

class CheckDomClosure: public BlockClosure {
private:
  GrowableArray<bool>* _visited;

public:
  CheckDomClosure(GrowableArray<bool>* visited) : _visited (visited) {}
  void block_do(BlockBegin* bb) {
    if (bb->is_set(BlockBegin::subroutine_entry_flag)) {
      // jsr destination
      assert(!_visited->at(bb->block_id()), "should not have visited this node");
    }
  }
};

//-------------------CreateInfoClosure--------------------------------

class CreateInfoClosure: public BlockClosure {
 private:
  LoopFinder* _lf;
 public:
  CreateInfoClosure(LoopFinder* lf): _lf(lf) {}
  void block_do(BlockBegin* bb) {
    assert(_lf->info()->at(bb->block_id()) == NULL, "already allocated?");
    BlockLoopInfo* bli = new BlockLoopInfo(bb, _lf->max_blocks());
    _lf->info()->at_put(bb->block_id(), bli);
  }
};

//-------------------SetPredsClosure----------------------------------


class SetPredsClosure: public BlockClosure {
 private:
  LoopFinder* _lf;
 public:
  SetPredsClosure(LoopFinder* lf) : _lf(lf) {}
  void block_do(BlockBegin* bb) {
    BlockEnd* be = bb->end();
    int n = be->number_of_sux();
    for (int i = 0; i < n; i++) {
      BlockBegin* sux = be->sux_at(i);
      BlockLoopInfo* sux_bli = _lf->get_block_info(sux);
      sux_bli->add_predecessor(bb);
    }
  }
};

//-------------------PrintBlockDominators--------------------------------

#ifndef PRODUCT
class PrintBlockDominators: public BlockClosure {
 private:
  LoopFinder* _lf;
 public:
  PrintBlockDominators(LoopFinder* lf): _lf(lf) {}
  void block_do(BlockBegin* bb) {
    assert(_lf->info()->at(bb->block_id()) != NULL, "already allocated?");
    BlockLoopInfo* bli = _lf->info()->at(bb->block_id());
    tty->print (" B%d doms [", bb->block_id());
    for (int i = 0; i < _lf->max_blocks(); i++) {
      if (bli->is_dom_block(i)) {
        tty->print(" B%d ", i);
      }
    }
    tty->print_cr("]");
  }
};
#endif

//-------------------LoopFinder---------------------------------------


// The CFG has one entry node that is predecessor of all method-entries: 
// - standard method entry
// - exception handlers
// JSR-entries are not analyzed in LoopFinder; however, we have to check that no 
// block that has not been analyzed jumps into or is jumped from an analyzed block.

LoopFinder::LoopFinder(IR* ir, int max_block_id) : 
       _ir(ir), _max_blocks(max_block_id), _ok(true), _changed(false) {
  _info = new GrowableArray<BlockLoopInfo*>(max_block_id, max_block_id, NULL);
  _changed = false;
  _valid_doms = false;
}


// Compute dominators for bb and 
// walk the successors of bb in depth first order
void LoopFinder::dominator_walk_sux(BlockBegin* bb, GrowableArray<bool>* visited) {
  // we may not visit a block that is jsr-target
  if (bb->is_set(BlockBegin::subroutine_entry_flag)) set_not_ok();
  BlockEnd*      be  = bb->end();
  BlockLoopInfo* bli = get_block_info(bb);
  visited->at_put(bb->block_id(), true);

  // compute new dominators using predecessors
  BitMap map(max_blocks());
  map.set_from(*BlockLoopInfo::all_blocks_map());
  { // Compute dominators for myself (looking at predecessors)
    int nof_preds = bli->nof_preds();
    for (int i = 0; i < nof_preds; i++) {
      BlockBegin* pred     = bli->pred_no(i);
      BitMap      pred_map = get_block_info(pred)->doms_map();
      map.set_intersection(pred_map);
    }
    // add itself
    map.at_put(bb->block_id(), true);
    
    // if the computed dominators differ from the one stored,
    // then we need another iteration
    BitMap bb_map = bli->doms_map();
    if (!bb_map.is_same(map)) {
      set_changed(true);
      bb_map.set_from(map);
    }
  }
  { // Visit all successors
    int n = be->number_of_sux();
    for (int i = 0; i < n; i++) {
      BlockBegin* sux = be->sux_at(i);
      if (!visited->at(sux->block_id())) {
        dominator_walk_sux(sux, visited);
      }
    }
  }
}


void LoopFinder::compute_dominators(GrowableArray<bool>* visited) {
  // set up a bitmap that contains all blocks
  BitMap all_map(max_blocks());
  all_map.clear();
  for (int i = 0; i < max_blocks(); i++) all_map.at_put(i, true);
  BlockLoopInfo::set_all_blocks_map(&all_map);
  { // initialize block loop info and set all predecessors
    CreateInfoClosure c(this);
    ir()->iterate_preorder(&c);
    SetPredsClosure s(this);
    ir()->iterate_preorder(&s);
  }
  { // compute dominators
    // init dominators
    // set entry block (only one exists) and its dominators
    BlockBegin* root_bb = ir()->start();
    assert(!root_bb->is_set(BlockBegin::subroutine_entry_flag), "root may not be jsr target");

    { // initialize root dominators (only itself)
      BitMap root_doms(max_blocks());
      root_doms.clear();
      root_doms.at_put(root_bb->block_id(), true);
      BlockLoopInfo* bli = get_block_info(root_bb);
      bli->doms_map().set_from(root_doms);
    }

    // iterate until either iter_count exceeds or dominators stable
    int iter_count = 0;
    do {
      iter_count++;
      visited->at_put(root_bb->block_id(), true);
      set_changed(false);
      BlockEnd* be = root_bb->end();
      int n = be->number_of_sux();
      for (int i = 0; i < n; i++) {
        BlockBegin* sux = be->sux_at(i);
        if (!visited->at(sux->block_id())) {
          dominator_walk_sux(sux, visited);
        }
      }
      if (changed()) {
        for (int i = visited->length() - 1; i >= 0; i--) {
          visited->at_put(i, false);
        }
      }
    } while (changed() && iter_count <= max_nof_dom_iterations);
    if (iter_count == max_nof_dom_iterations) {
      if (PrintLoops) {
        tty->print_cr("could not computer dominators");
      }
      set_not_ok();
    }

    // if (PrintLoops) {
    //   tty->print_cr("  Dominators: %d iterations", iter_count);
    //   PrintBlockDominators p(this);
    //   ir()->iterate_topological(&p);
    // }
  }  
  BlockLoopInfo::set_all_blocks_map(NULL);


  // go through all blocks; if a block has not been analyzed, then check its
  // predecessors and successors: all must be also un-analyzed;
  // Note: the block may not be JSR blocks
  if (ok()) {
    _valid_doms = true;
#ifdef ASSERT
    CheckDomClosure cc(visited);
    ir()->iterate_preorder(&cc);
#endif
  }

}



//-------------------------Loop-----------------------------------------------

static int cmp(BlockBegin** a, BlockBegin** b)   { return (*a)->weight() - (*b)->weight(); }

void Loop::sort_blocks() {
  _blocks->sort(cmp);
}

// Factor out visiting of all entries/exits
void Loop::update_loop_blocks(BlockBegin* old_from, BlockBegin* old_to, BlockBegin* insert) {
  int n = _loop_entries->length() - 1;
  for (; n >= 0; n--) {
    BlockPair* bp = _loop_entries->at(n);
    if (bp->is_same(old_from, old_to)) {
      bp->set_from(insert);
    }
  }
  n  = _loop_exits->length() - 1;
  for (; n >= 0; n--) {
    BlockPair* bp = _loop_exits->at(n);
    if (bp->is_same(old_from, old_to)) {
      bp->set_to(insert);
    }
  }
}


#ifndef PRODUCT
void Loop::print() {
  tty->print_cr("!Loop header B%d (%d), backedge B%d (%d)", 
    start()->block_id(), start()->bci(), end()->block_id(), end()->bci());
  tty->print("   blocks[");
  if (Verbose) {
    int i;
    for (i = 0; i < _blocks->length(); i++) {
      tty->print(" B%d", _blocks->at(i)->block_id());
    }
    tty->print_cr("]");
    for (i = 0; i < _loop_entries->length(); i++) {
      tty->print_cr("   entry [B%d -> B%d]", _loop_entries->at(i)->from()->block_id(), _loop_entries->at(i)->to()->block_id());
    }
    for (i = 0; i < _loop_exits->length(); i++) {
      tty->print_cr("   exit  [B%d -> B%d]", _loop_exits->at(i)->from()->block_id(), _loop_exits->at(i)->to()->block_id());
    }
  } else {
    tty->print_cr("]");
  }
}
#endif


//-------------------------Loop-----------------------------------------------


// Gather backedges of natural loops: an edge a -> b where b dominates a
GrowableArray<Loop*>* LoopFinder::find_backedges(GrowableArray<bool>* visited) {
  GrowableArray<Loop*>* backedges = new GrowableArray<Loop*>();
  for (int i = 0; i < max_blocks(); i++) {
    if (visited->at(i)) {
      BlockLoopInfo* bli = _info->at(i);
      BlockBegin*    bb  = bli->block();
      BlockEnd*      be  = bb->end();
      int n = be->number_of_sux();
      for (int i = 0; i < n; i++) {
        BlockBegin* sux = be->sux_at(i);
        if (bli->is_dom_block(sux->block_id())) {
          bli->mark_backedge_start();
          backedges->push(new Loop(sux, bb));
        }
      }
    }
  }
  return backedges;
}


void LoopFinder::gather_loop_blocks(GrowableArray<Loop*>* loops) {
  int lng = loops->length();
  BitMap blocks_in_loop(max_blocks());
  for (int i = 0; i < lng; i++) {
    // for each loop do the following
    blocks_in_loop.clear();
    Loop* loop = loops->at(i);
    if (loop->start() != loop->end()) {
      blocks_in_loop.at_put(loop->start()->block_id(), true);
      blocks_in_loop.at_put(loop->end()->block_id()  , true);
      GrowableArray<BlockBegin*>* stack = new GrowableArray<BlockBegin*>();
      stack->push(loop->end());
      while (!stack->is_empty()) {
        BlockBegin* bb = stack->pop();
        BlockLoopInfo* bli = get_block_info(bb);
        // push all predecessors that are not yet in loop
        int npreds = bli->nof_preds();
        for (int m = 0; m < npreds; m++) {
          BlockBegin* pred = bli->pred_no(m);
          if (!blocks_in_loop.at(pred->block_id())) {
            blocks_in_loop.at_put(pred->block_id(), true);
            loop->append_node(pred);
            stack->push(pred);
          }
        }
      }
      loop->append_node(loop->start());
    }
    loop->append_node(loop->end());
  }
}


void LoopFinder::find_loop_entries(BlockBegin* bb, Loop* loop) {
  BlockLoopInfo* bli = get_block_info(bb);
  int loop_index = bli->loop_index();
  // search all predecessors and locate the ones that do not have the same loop_index
  int n = bli->nof_preds() - 1;
  assert(bli->nof_preds() >= 2, "at least two predecessors required for a loop header");
  for (; n >= 0; n--) {
    BlockBegin* pred_bb = bli->pred_no(n);
    BlockLoopInfo* pred_bli = get_block_info(pred_bb);
    if (pred_bli->loop_index() != loop_index) {
      loop->append_loop_entry(pred_bb, bb);
    }
  }
  assert(loop->nof_loopentries() > 0, "a loop must be entered from somewhere");
}


void LoopFinder::find_loop_exits(BlockBegin* bb, Loop* loop) {
  BlockLoopInfo* bli = get_block_info(bb);
  int loop_index = bli->loop_index();
  // search all successors and locate the ones that do not have the same loop_index
  BlockEnd* be = bb->end();
  int n = be->number_of_sux() - 1;
  for(; n >= 0; n--) {
    BlockBegin* sux = be->sux_at(n);
    BlockLoopInfo* sux_bli = get_block_info(sux);
    if (sux_bli->loop_index() != loop_index) {
      loop->append_loop_exit(bb, sux);
    }
  }
}



void LoopFinder::compute_loop_exits_and_entries(GrowableArray<Loop*>* loops) {
  // create loop exits for each loop
  int loop_index = loops->length() - 1;
  for (; loop_index >= 0; loop_index--) {
    Loop* loop = loops->at(loop_index);
    int n = loop->nof_blocks() - 1;
    // mark all nodes belonging to this loop 
    for (; n >= 0; n --) {
      BlockBegin* bb = loop->block_no(n);
      BlockLoopInfo* bli = get_block_info(bb);
      bli->set_loop_index(loop_index);
    }
    find_loop_entries(loop->start(), loop);
    // search for blocks that have successors outside the loop
    n = loop->nof_blocks() - 1;
    for (; n >= 0; n--) {
      BlockBegin* bb = loop->block_no(n);
      find_loop_exits(bb, loop);
    }
  }
}


// Returns inner loops that have no calls
GrowableArray<Loop*>* LoopFinder::find_inner_loops(GrowableArray<Loop*>* loops) {
  GrowableArray<Loop*>* inner = new GrowableArray<Loop*>();
  int lng = loops->length();
  // First step: find loops that have no calls and no backedges
  //             in the loop except its own
  int i;
  for (i = 0; i < lng; i++) {
    Loop* loop = loops->at(i);
    int k = loop->nof_blocks() - 1;
    for (; k >= 0; k--) {
      BlockBegin* b = loop->block_no(k);
      // Is this innermost loop:
      // - no block, except end block, may be a back edge start,
      //   otherwise we have an outer loop
      if (b != loop->end()) {
        BlockLoopInfo* bli = get_block_info(b);
        if (bli->is_backedge_start()) {
          // invalidate loop because this is an outer loop
          loop = NULL;
          break;
        }
      }
    }
    if (loop != NULL) {
      ScanBlocks scan(loop->blocks());
      ScanResult scan_result;
      scan.scan(&scan_result);
      if (!scan_result.has_calls() && !scan_result.has_slow_cases()) {
        inner->append(loop);
      }
    }
  }
  // Second step: if several loops have the same loop-end, select the one
  //              with fewer blocks.
  //              if several loops have the same loop-start, select the one
  //              with fewer blocks
  // now check for loops that have the same header and eliminate one
  for (i = 0; i < inner->length() ; i++) {
    Loop* current_loop = inner->at(i);
    BlockBegin* header = current_loop->start();
    BlockBegin* end    = current_loop->end();
    for (int n = i + 1; n < inner->length(); n++) {
      Loop*  test_loop = inner->at(n);
      BlockBegin* test = test_loop->start();
      if (header == test_loop->start() || end == test_loop->end()) {
        // discard loop with less blocks
        if (test_loop->nof_blocks() > current_loop->nof_blocks()) {
          inner->remove(test_loop);
        } else {
          inner->remove(current_loop);
        }
        // restart the computation
        i = -1;
        break;
      }
    }
  }
  return inner;
}



BlockBegin* LoopFinder::new_block(IRScope* scope, int bci) {
  BlockBegin* b = new BlockBegin(scope, bci);
  _valid_doms = false;
  assert(b->block_id() == max_blocks(), "illegal block_id");
  _max_blocks++;
  BlockLoopInfo* bli = new BlockLoopInfo(b, max_blocks());
  _info->append(bli);
  assert(_info->length() == max_blocks(), "operation failed");
  return b;
}



BlockBegin* LoopFinder::create_loop_entry(int loop_id, BlockBegin* header, int smallest_pred_bci) {
  int bci = smallest_pred_bci;
  BlockBegin* b = new_block(header->scope(), bci);
  BlockEnd* e = new Goto(header);
  b->set_end(e);
  // b->append(new LoopEnter(loop_id), bci)->append(e, -1); // set to -1 so that it does not look like an endless loop
  b->set_next(new LoopEnter(loop_id), bci)->set_next(e, bci); 
  if (PrintLoops && Verbose) {
    tty->print_cr("Added loop entry B%d (header B%d)", b->block_id(), header->block_id());
  }
  return b;
}


// Currently same as create_loop_exit
BlockBegin* LoopFinder::create_loop_exit(int loop_id, BlockBegin* destination) {
  int bci = destination->bci();
  BlockBegin* b = new_block(destination->scope(), bci);
  BlockEnd* e = new Goto(destination);
  b->set_end(e);
  // b->append(new LoopExit(loop_id), bci)->append(e, -1);
  b->set_next(new LoopExit(loop_id), bci)->set_next(e, bci);
  if (PrintLoops && Verbose) {
    tty->print_cr("Added loop exit B%d (dest B%d)", b->block_id(), destination->block_id());
  }
  return b;
}


void LoopFinder::add_predecessor(BlockBegin* b, BlockBegin* pred) {
  BlockLoopInfo* bli = get_block_info(b);
  bli->add_predecessor(pred);
}


void LoopFinder::remove_predecessor(BlockBegin* b, BlockBegin* pred) {
  BlockLoopInfo* bli = get_block_info(b);
  bli->remove_predecessor(pred);
}


void LoopFinder::insert_block(BlockBegin* from, BlockBegin* to, BlockBegin* insert) {
  BlockEnd* from_end = from->end();
  from_end->substitute_sux(to, insert);
  insert->join(from_end->state());
  assert(insert->state() != NULL, "illegal operation");
  add_predecessor(insert, from);
  remove_predecessor(to, from);
  add_predecessor(to, insert);
}


void LoopFinder::update_loops(GrowableArray<Loop*>* loops, BlockBegin* from, BlockBegin* to, BlockBegin* insert) {
  int n = loops->length() - 1;
  for (; n >= 0; n--) {
    loops->at(n)->update_loop_blocks(from, to, insert);
  }
}


// Adds one preheader for the specified loop
void LoopFinder::add_preheader(GrowableArray<Loop*>* loops, int loop_no) {
  Loop* loop = loops->at(loop_no);

  BlockBegin* header = loop->start();
  BlockLoopInfo* bli = get_block_info(header);
  assert( bli->nof_preds() > 1, "must have backedge and at least root coming in");
  int smallest_pred_bci = -1;
  {
    int p = bli->nof_preds() - 1;
    smallest_pred_bci = bli->pred_no(0)->end()->bci();
    for(; p >= 0; p--) {
      BlockBegin* pred = bli->pred_no(p);
      smallest_pred_bci = min(smallest_pred_bci, pred->end()->bci());
      assert(smallest_pred_bci >= 0, "wrong bci");
    }
  }
  assert(smallest_pred_bci >= 0, "wrong bci");

  // add preheader (make it close to predecessor with smallest bci)
  BlockBegin* loop_entry = create_loop_entry(loop_no, header, smallest_pred_bci);
  loop->append_node(loop_entry);
  // sux of entry is already set correctly; update all predecessors
  int p = bli->nof_preds() - 1;
  for(; p >= 0; p--) {
    BlockBegin* pred = bli->pred_no(p);
    smallest_pred_bci = max(smallest_pred_bci, pred->bci());
    // The backedge block continues to go directly to header
    if (pred != loop->end()) {
      insert_block(pred, header, loop_entry);
      update_loops(loops, pred, header, loop_entry);
    }
  }
  { // Setup state
    assert(loop_entry->state() != NULL, "illegal state");
    ValueStack* end_state = loop_entry->state()->copy();
    loop_entry->end()->set_state(end_state);
    header->join(end_state);
  }
  // set bci 
}


BlockPair* LoopFinder::find_blockpair_with_to(GrowableArray<BlockPair*>* a, BlockBegin* to) {
  int i = a->length() - 1;
  for (; i >=0; i--) {
    if (a->at(i)->to() == to) return a->at(i);
  }
  return NULL;
}

// All exits to the same block must be through a single loop exit block
void LoopFinder::add_loopexits(GrowableArray<Loop*>* loops, int loop_no) {
  Loop* loop = loops->at(loop_no);
  GrowableArray<BlockPair*>* new_blocks = new GrowableArray<BlockPair*>();
  int n = loop->nof_loopexits() - 1;
  for (; n >= 0; n--) {
    // allocate all loop exit blocks and store them with the appropriate destination
    BlockPair* exit = loop->loopexit_no(n);
    BlockPair* new_exit = find_blockpair_with_to(new_blocks, exit->to());
    if (new_exit == NULL) {
      BlockBegin* exit_block = create_loop_exit(loop_no, exit->to());
      loop->append_node(exit_block);
      new_exit= new BlockPair(exit_block, exit->to());
      new_blocks->append(new_exit);
    }
    insert_block(exit->from(), exit->to(), new_exit->from());
    update_loops(loops, exit->from(), exit->to(), new_exit->from());
  }
  
  { // go through all new_block entries and update states
    n = new_blocks->length() - 1;
    for (; n >= 0; n--) {
      BlockPair* p = new_blocks->at(n);
      BlockBegin* loop_exit = p->from();
      assert(loop_exit->state() != NULL, "wrong computation");
      ValueStack* end_state = loop_exit->state()->copy();
      loop_exit->end()->set_state(end_state);
      p->to()->join(end_state);
    }
  }
}

// Add preheader and loop exit blocks; after this pass,
// dominators are not valid anymore
void LoopFinder::insert_blocks(GrowableArray<Loop*>* loops) {
  int n = loops->length() - 1;
  for (; n >=0; n--) {
    add_preheader(loops, n);
    add_loopexits(loops, n);
  }
}


#ifndef PRODUCT
void LoopFinder::print_loops(const char* msg, GrowableArray<Loop*>* loops) {
  if (loops != NULL && loops->length() != 0) {
    tty->print_cr(msg);
    for (int i = 0; i < loops->length(); i++) {
      loops->at(i)->print();
    }
  }
}
#endif

// Note that OSR compilation prevents the OSR loop from being detected
// as the loop is being jumped in the middle and not at loop entry

// returns an array of call-free inner-loops
GrowableArray<Loop*>* LoopFinder::compute_loops() {
  // visited has flags set for each block that has been visited
  GrowableArray<bool> visited (max_blocks(), max_blocks(), false);
  if (ok()) compute_dominators(&visited);

  GrowableArray<Loop*>* loops = NULL;
  GrowableArray<Loop*>* inner = NULL;
  if (ok()) loops = find_backedges(&visited);
  if (ok())         gather_loop_blocks(loops);
  if (ok()) inner = find_inner_loops(loops);
  if (ok())         compute_loop_exits_and_entries(inner);
  if (ok() && PrintLoops) {
    if (Verbose) print_loops("+All Loops:", loops);
    print_loops("+Inner Loops:", inner);
  }
  if (ok() /*&& CacheLocals::can_be_optimized(inner)*/)  insert_blocks(inner);
  if (ok() && PrintLoops && Verbose) {
    print_loops("+All Loops:", loops);
    print_loops("+Inner Loops:", inner);
    ir()->print(false);
  }
  return inner;
}



