// @(#)c1_Loops.hpp	1.16 00/03/21 16:16:36
// 
// Copyright 1993-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


class BlockLoopInfo: public ResourceObj {
 private:
  static BitMap* _all_blocks;  // bitmap where all bits for all possible block id's are set

  BitMap      _doms_map;       // for each dominating block the id bit is set
  BlockBegin* _block;          // the block that is described by this BLI (BlockLoopInfo)
  bool        _backedge_start; // true if a backedge goes of from this node
  int         _loop_index;

  GrowableArray<BlockBegin*>* _preds; // list of predecessors for each block id

 public:
  BlockLoopInfo(BlockBegin* bb, int max_blocks);

  BlockBegin* block() const                  { return _block;          }

  // predecessors
  void add_predecessor(BlockBegin* pred)     { _preds->append(pred);    }
  void remove_predecessor(BlockBegin* pred);
  int  nof_preds() const                     { return _preds->length(); }
  BlockBegin* pred_no(int i) const           { return _preds->at(i);    }

  // bitmaps
  BitMap doms_map() const                    { return _doms_map; }
  bool is_dom_block(int id) const            { return _doms_map.at(id); }

  // backedges
  void mark_backedge_start()                 { _backedge_start = true;  }
  bool is_backedge_start() const             { return _backedge_start;  }

  void set_loop_index(int ix)                { _loop_index = ix;        }
  int  loop_index() const                    { return _loop_index;      }

  void print() PRODUCT_RETURN;

  static void set_all_blocks_map(BitMap* map) { _all_blocks = map;  }
  static BitMap* all_blocks_map()             { return _all_blocks; }
};


class BlockPair: public ResourceObj {
 private:
  BlockBegin* _from; 
  BlockBegin* _to;   
 public:
  BlockPair(BlockBegin* from, BlockBegin* to): _from(from), _to(to) {}
  BlockBegin* from() const { return _from; }
  BlockBegin* to() const   { return _to;   }
  bool is_same(BlockBegin* from, BlockBegin* to) const { return  _from == from && _to == to; }
  void set_to(BlockBegin* b)   { _to = b; }
  void set_from(BlockBegin* b) { _from = b; }
};


class CachedLocal: public ResourceObj {
 private:
  int   _local_index;
  RInfo _rinfo;
 public:
  CachedLocal(int index, RInfo rinfo): _local_index(index), _rinfo(rinfo) {}

  int local_index() const { return _local_index; }
  RInfo rinfo() const     { return _rinfo;  }
};


class Loop: public ResourceObj {
 private:
  BlockBegin* _loop_start;  // header
  BlockBegin* _loop_end;    // containing backedge

  GrowableArray<BlockBegin*>* _blocks; // all blocks contained inside the loop (including start and end node)
  GrowableArray<BlockPair*>*  _loop_entries; 
  GrowableArray<BlockPair*>*  _loop_exits;   

  GrowableArray<CachedLocal*>* _cached_locals;

  c1_RegMask                   _lockout; // these are register used for caching of locals in the loop

 public:
  Loop(BlockBegin* start, BlockBegin* end): _loop_start(start), _loop_end(end) {
    _blocks = new GrowableArray<BlockBegin*>();
    _loop_entries = new GrowableArray<BlockPair*>();
    _loop_exits = new GrowableArray<BlockPair*>();
    _cached_locals = new GrowableArray<CachedLocal*>();
    _lockout = c1_RegMask::empty_set();
  }

  void set_lockout(c1_RegMask lockout) { _lockout = lockout;  }
  c1_RegMask lockout() const { return _lockout; }

  BlockBegin* start() const { return _loop_start; }
  BlockBegin* end() const   { return _loop_end;   }

  void append_node(BlockBegin* b)  { _blocks->append(b); }

  int nof_blocks() const            { return _blocks->length(); }
  BlockBegin* block_no(int i) const { return _blocks->at(i);    }

  GrowableArray<BlockBegin*>* blocks() const { return _blocks; }

  GrowableArray<CachedLocal*>* cached_locals() const { return _cached_locals; }
  void append_cached_local(CachedLocal* cl) { _cached_locals->append(cl); }

  void append_loop_entry(BlockBegin* from, BlockBegin* to) { 
    _loop_entries->append(new BlockPair(from, to));
  }

  void append_loop_exit(BlockBegin* from, BlockBegin* to) { 
    _loop_exits->append(new BlockPair(from, to));
  }

  int nof_loopexits() const           { return _loop_exits->length(); }
  BlockPair* loopexit_no(int i) const { return _loop_exits->at(i);    }

  int nof_loopentries() const          { return _loop_entries->length(); }
  BlockPair* loopentry_no(int i) const { return _loop_entries->at(i);    }

  void update_loop_blocks(BlockBegin* old_from, BlockBegin* old_to, BlockBegin* insert);

  void sort_blocks();

  void blocks_do(void f(BlockBegin*)) {
    for (int n = nof_blocks() - 1; n >= 0; n--) {
      f(block_no(n));
    }
  }

  void iterate_blocks(BlockClosure* closure) {
    const int l = nof_blocks();
    for (int i = 0; i < l; i++) closure->block_do(block_no(i));
  }


#ifndef PRODUCT
  void print();
#endif
};



class LoopFinder: public StackObj {
 private:
  enum {
    max_nof_dom_iterations = 8
  };
  bool _ok;         // set to false if we encounter problems (loop analysis failed)
  bool _changed;    // used by iterative dominator algorithm
  bool _valid_doms; // true if computed dominators are valid
  int  _max_blocks; // maximum number of blocks
  IR*  _ir;
  GrowableArray<BlockLoopInfo*>* _info;

  bool ok() const                       { return _ok;  }
  void set_not_ok()                     { _ok = false; }

  bool changed() const                  { return _changed; }
  void set_changed(bool f)              { _changed = f;    }

  IR* ir() const                        { return _ir; }

  void dominator_walk_sux(BlockBegin* bb, GrowableArray<bool>* visited);
  void compute_dominators(GrowableArray<bool>* visited);

  GrowableArray<Loop*>* find_backedges(GrowableArray<bool>* visited);
  GrowableArray<Loop*>* find_inner_loops(GrowableArray<Loop*>* loops);

  void compute_loop_exits_and_entries(GrowableArray<Loop*>* loops);
  void gather_loop_blocks  (GrowableArray<Loop*>* loops);
  void insert_blocks       (GrowableArray<Loop*>* loops);
  void find_loop_exits     (BlockBegin* bb, Loop* loop);
  void find_loop_entries   (BlockBegin* bb, Loop* loop);
  
  BlockBegin* create_loop_entry (int loop_id, BlockBegin* header, int bci);
  BlockBegin* create_loop_exit  (int loop_id, BlockBegin* header);

  void add_predecessor    (BlockBegin* b, BlockBegin* pred);
  void remove_predecessor (BlockBegin* b, BlockBegin* pred);
  void insert_block       (BlockBegin* from, BlockBegin* to, BlockBegin* insert);

  void add_preheader (GrowableArray<Loop*>* loops, int n);
  void add_loopexits (GrowableArray<Loop*>* loops, int n);

  BlockPair* find_blockpair_with_to (GrowableArray<BlockPair*>* a, BlockBegin* to);
  void update_loops(GrowableArray<Loop*>* loops, BlockBegin* from, BlockBegin* to, BlockBegin* insert);

  BlockBegin* new_block(IRScope* scope, int bci);

  void print_loops(const char* msg, GrowableArray<Loop*>* loops);


 protected:
  friend class SetPredsClosure;
  friend class CreateInfoClosure;
  friend class PrintBlockDominators;

  int  max_blocks() const                             { return _max_blocks; }
  BlockLoopInfo* get_block_info(BlockBegin* bb) const { return _info->at(bb->block_id()); }
  GrowableArray<BlockLoopInfo*>* info() const         { return _info; }

 public:
  LoopFinder(IR* ir, int max_block_id);

  GrowableArray<Loop*>* compute_loops();

  void print() PRODUCT_RETURN;
};

