// @(#)c1_Optimizer.cpp	1.34 00/04/10 12:14:49
// 
// Copyright 1993-1999 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_c1_Optimizer.cpp.incl"


Optimizer::Optimizer(IR* ir) {
  assert(ir->is_valid(), "IR must be valid");
  _ir = ir;
}


class CE_Eliminator: public BlockClosure {
 private:
  int _cee_count;                                // the number of CEs successfully eliminated

  Value simple_value_copy(Value x) {
    ValueType* type = x->type();
    if (type->is_float_kind()) return NULL;      // don't do floats for now since backend doesn't support it
    if (x->as_Constant() != NULL) return new Constant(type);
    if (x->as_LoadLocal() != NULL) return new LoadLocal(type, x->as_LoadLocal()->index());
    return NULL;
  }

 public:
  CE_Eliminator() : _cee_count(0)                {}
  int cee_count() const                          { return _cee_count; }

  virtual void block_do(BlockBegin* block) {
    // 1) find conditional expression
    // check if block ends with an If
    If* if_ = block->end()->as_If();
    if (if_ == NULL) return;
    // check if If works on int or object types
    // (we cannot handle If's working on long, float or doubles yet,
    // since IfOp doesn't support them - these If's show up if cmp
    // operations followed by If's are eliminated)
    { ValueType* type = if_->x()->type();
      if (!type->is_int() && !type->is_object()) return;
    }
    // check if branches push a simple value as first instruction
    // (i.e., a constant or a local) and copy the instruction if so
    Instruction* t1 = if_->tsux()->next();
    Instruction* f1 = if_->fsux()->next();
    Value tval = simple_value_copy(t1);
    if (tval == NULL) return;
    Value fval = simple_value_copy(f1);
    if (fval == NULL) return;
    // check if both branches end with a goto
    // note: next() must exist since previous instruction
    //       was simple (and thus not a BlockEnd instruction)
    assert(t1->next() != NULL && f1->next() != NULL, "next must exist");
    Goto* t2 = t1->next()->as_Goto();
    if (t2 == NULL) return;
    Goto* f2 = f1->next()->as_Goto();
    if (f2 == NULL) return;
    // check if both gotos merge into the same block
    BlockBegin* sux = t2->default_sux();
    if (sux != f2->default_sux()) return;
    // check if both branches actually "produce" a value
    // Note: both the left and the right branch could load a value and
    //       then pop it, but the load would still be there as instruction
    //       (this could happen after inlining) => we must check that the
    //       stack size has actually changed by - was bug (gri 8/20/99).
    assert(tval->type()->size() == fval->type()->size(), "sizes must match");
    if (if_->state()->stack_size() + tval->type()->size() != sux->state()->stack_size()) return;

    // 2) substitute conditional expression
    //    with an IfOp followed by a Goto
    // cut if_ away and get node before
    Instruction* if_prev = if_->prev(block);
    if_prev->set_next(NULL, 0);
    // create new nodes and connect them
    IfOp* if_op = new IfOp(if_->x(), if_->cond(), if_->y(), tval, fval);
    Goto* goto_ = new Goto(sux);
    int bci = if_->bci();
    if_prev
      ->set_next(tval , bci)
      ->set_next(fval , bci)
      ->set_next(if_op, bci)
      ->set_next(goto_, bci);
    // set state and join sux
    ValueStack* state = if_->state()->copy();
    state->push(if_op->type(), if_op);
    goto_->set_state(state);
    sux->join(state);
    // update block end
    block->set_end(goto_);
    state->pin_stack_all();

    // 3) successfully eliminated a conditional expression
    _cee_count++;
    if (PrintCEE) {
      tty->print_cr("%d. CEE in B%d", cee_count(), block->id());
    }
  }
};


void Optimizer::eliminate_conditional_expressions() {
  // find conditional expressions & replace them with IfOps
  CE_Eliminator ce;
  ir()->iterate_preorder(&ce);
}


class PredecessorCounter: public BlockClosure {
 private:
  GrowableArray<int> _count;
  void add_pred(BlockBegin* block)               { (*_count.adr_at(block->block_id()))++; }

 public:
  PredecessorCounter(int n) : _count(n, n, 0)    {}

  int pred_count(BlockBegin* block) const        { return _count.at(block->block_id()); }

  virtual void block_do(BlockBegin* block) {
    // Note: Make sure to count subroutine and exception handler paths as well in case
    //       one of them is shared with regular control flow (was bug - gri 4/10/2000).
    for (int i = block->end()->number_of_sux        (); i-- > 0;) add_pred(block->end()->sux_at       (i));
    for (int j = block->number_of_subroutines       (); j-- > 0;) add_pred(block->subroutine_at       (j));
    for (int k = block->number_of_exception_handlers(); k-- > 0;) add_pred(block->exception_handler_at(k));
  }
};


class BlockMerger: public BlockClosure {
 private:
  PredecessorCounter* _pc;
  int                 _merge_count;              // the number of block pairs successfully merged

 public:
  BlockMerger(PredecessorCounter* pc)
  : _pc(pc)
  , _merge_count(0)
  {}

  int pred_count(BlockBegin* block) const        { return _pc->pred_count(block); }
  int merge_count() const                        { return _merge_count; }

  bool try_merge(BlockBegin* block) {
    BlockEnd* end = block->end();
    if (end->as_Goto() != NULL) {
      assert(end->number_of_sux() == 1, "end must have exactly one successor");
      // Note: It would be sufficient to check for the number of successors (= 1)
      //       in order to decide if this block can be merged potentially. That
      //       would then also include switch statements w/ only a default case.
      //       However, in that case we would need to make sure the switch tag
      //       expression is executed if it can produce observable side effects.
      //       We should probably have the canonicalizer simplifying such switch
      //       statements and then we are sure we don't miss these merge oportunities
      //       here (was bug - gri 7/7/99).
      BlockBegin* sux = end->default_sux();
      if (pred_count(sux) == 1 && !sux->is_entry_block()) {
        // merge the two blocks
        assert(end->state()->stack_size() == sux->state()->stack_size(), "stack sizes must correspond");
        // find instruction before end & append first instruction of sux block
        Instruction* prev = end->prev(block);
        Instruction* next = sux->next();
        assert(prev->as_BlockEnd() == NULL, "must not be a BlockEnd");
        prev->set_next(next, next->bci());
        block->set_end(sux->end());
        // add subroutines of deleted block, if any
        for (int j = 0; j < sux->number_of_subroutines(); j++) {
          block->add_subroutine(sux->subroutine_at(j));
        }
        // add exception handlers of deleted block, if any
        for (int k = 0; k < sux->number_of_exception_handlers(); k++) {
          block->add_exception_handler(sux->exception_handler_at(k));
        }
        // substitute phi's
        int stack_size = sux->state()->stack_size();
        if (stack_size > 0) {
          // set substitutions
          for (int i = 0; i < stack_size;) {
            int i0 = i; // stack_at_inc(...) modifies i
#ifdef ASSERT
            int i1 = i;
            assert(sux->state()->stack_at_inc(i1)->as_Phi() != NULL, "must be a phi node");
#endif // ASSERT
            sux->state()->stack_at_inc(i0)->set_subst(end->state()->stack_at_inc(i));
          }
          // resolve substitutions
          sux->resolve_substitution();
        }
        // debugging output
        _merge_count++;
        if (PrintBlockElimination) {
          tty->print_cr("%d. merged B%d & B%d (stack size = %d)", merge_count(), block->block_id(), sux->block_id(), stack_size);
        }
        return true;
      }
    }
    return false;
  }

  virtual void block_do(BlockBegin* block) {
    while (try_merge(block)); // repeat since the same block may merge again
  }
};


void Optimizer::eliminate_blocks() {
  // determine predecessor count for each block
  PredecessorCounter pc(BlockBegin::number_of_blocks());
  ir()->iterate_preorder(&pc);
  // merge blocks if possible
  BlockMerger bm(&pc);
  ir()->iterate_preorder(&bm);
}

