// @(#)c1_RInfo.hpp	1.10 00/04/24 10:01:04
// 
// Copyright 1993-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


// Forward declarations:
class Register;


// Abstract register representation

// This class represents a register that is either in CPU or in FPU.
// It also represents "long-registers", which are composed of two CPU
// registers. The type and the registers are stored in the _number.

class RInfo VALUE_OBJ_CLASS_SPEC {
  #include "incls/_c1_RInfo_pd.hpp.incl"

  // The underlying representation of RInfo is an integer stored in
  // _number. We want to keep _number private as well as any other
  // conversions between ints and RInfos. There are a few classes that
  // need to convert from int to RInfo, those are the friend classes
  // below.
  friend class RInfoCollection;
  friend class c1_RegMask;
  friend class RegAlloc;
  friend class RInfo2Reg;
  friend class Item;
  friend class FrameMap;
  
private:
  enum {
    type_bits = 2,
    reg_bits  = 5
  };
  enum RegType {
    word_reg_type   = 0,   // one cpu register
    long_reg_type   = 1,   // a pair of cpu registers
    float_reg_type  = 2,
    double_reg_type = 3
  };
  enum {
    type_shift = 0,
    reg1_shift = type_bits,
    reg2_shift = reg1_shift + reg_bits
  };
  enum {
    type_mask          = right_n_bits(type_bits),
    type_mask_in_place = type_mask << type_shift,
    reg1_mask          = right_n_bits(reg_bits),
    reg1_mask_in_place = reg1_mask << reg1_shift,
    reg2_mask          = right_n_bits(reg_bits),
    reg2_mask_in_place = reg2_mask << reg2_shift
  };
  enum {
    illegal_reg_number = -1
  };

private:
  int _number;

  RInfo (int n, char dummy) : _number(n) {}

  RegType type () const {
    assert(type_shift == 0, "assumption");
    return (RegType)mask_bits(number(), type_mask_in_place);
  }

  int number () const { return _number; }
  int reg1   () const { return mask_bits(number() >> reg1_shift, reg1_mask); }
  int reg2   () const { return mask_bits(number() >> reg2_shift, reg2_mask); }

  void set_word_reg   (int reg)            { _number = (reg  << reg1_shift) + word_reg_type; }
  void set_long_reg   (int reg1, int reg2) { _number = (reg1 << reg1_shift) + (reg2 << reg2_shift) + long_reg_type; }
  void set_float_reg  (int reg)            { _number = (reg  << reg1_shift) + float_reg_type; }
  void set_double_reg (int reg)            { _number = (reg  << reg1_shift) + double_reg_type; }

  int reg ()        const { assert(is_word(),   "type_check"); return reg1(); }
  int reg_lo ()     const { assert(is_long(),   "type_check"); return reg1(); }
  int reg_hi ()     const { assert(is_long(),   "type_check"); return reg2(); }
  int float_reg ()  const { assert(is_float(),  "type check"); return reg1(); }
  int double_reg () const { assert(is_double(), "type check"); return reg1(); }

  static RInfo word_reg (int rnr) { RInfo r; r.set_word_reg(rnr); return r; }
  
public:
  // constructors:
  RInfo () : _number(illegal_reg_number) {}

  // accessors and tests:
  void set_no_rinfo ()                    { _number = illegal_reg_number; }
  bool is_same  (const RInfo other) const { return _number == other._number; }
  bool overlaps (const RInfo other) const;

  bool is_word ()      const { return type() == word_reg_type; }
  bool is_long ()      const { return type() == long_reg_type; }
  bool is_float ()     const { return type() == float_reg_type; }
  bool is_double ()    const { return type() == double_reg_type; }
  bool is_float_kind() const { return is_float() || is_double(); }
  bool is_illegal ()   const { return _number == illegal_reg_number; }
  bool is_valid ()     const { return _number != illegal_reg_number; }

  Register as_register () const;     // for integers
  Register as_register_lo () const;  // for longs
  Register as_register_hi () const;  // for longs

  // debugging:
  void print ()          const PRODUCT_RETURN;
  void print_raw ()      const PRODUCT_RETURN;
  void print (char* msg) const PRODUCT_RETURN;
};

const RInfo norinfo = ::RInfo();

class RInfoCollection : public ResourceObj {
private:
  GrowableArray<int>* _regs;  // register as RInfo numbers
  void init()                            { if (_regs == NULL) _regs = new GrowableArray<int>(); }
public:
  RInfoCollection () : _regs(NULL) {}

  int number_of_registers () const       { return _regs != NULL ? _regs->length() : 0; }
  RInfo register_at (int index) const    { assert(_regs != NULL, "just checking"); return RInfo(_regs->at(index), 'a'); }
  void append (const RInfo r)            { init(); _regs->append(r.number()); }
  void at_put (int index, const RInfo r) { init(); _regs->at_put_grow(index, r.number(), RInfo::illegal_reg_number);}
  bool contains(RInfo r) const           { return _regs != NULL ? _regs->contains(r.number()) : false; }
  void remove(RInfo r) const             { _regs->remove(r.number()); }
};


//-------------------------------------------------------
//                 c1_RegMask
//-------------------------------------------------------

// Represents a set of registers. Implemented as an int mask, a register
// is a member of the set if the bit corresponding to its rnr is set in
// the mask.

class c1_RegMask VALUE_OBJ_CLASS_SPEC {

  friend class c1_AllocTable;
  friend class FrameMap;
  friend class RegAlloc;

 private:
  int _mask;

  static int _size;
  static const c1_RegMask _empty_set;

  c1_RegMask add_reg    (int rnr)               { check_rnr(rnr); set_nth_bit  (_mask, rnr); return *this; }
  c1_RegMask remove_reg (int rnr)               { check_rnr(rnr); clear_nth_bit(_mask, rnr); return *this; }
  bool       contains   (int rnr) const         { check_rnr(rnr); return is_set_nth_bit(_mask, rnr); }
  int        get_first () const;

  static void check_rnr (int rnr)               { assert(rnr >= 0 && rnr < _size, "wrong index"); }

 public:
  c1_RegMask () : _mask(0) {}
  c1_RegMask (RInfo r, char dummy) : _mask(0)   { assert(r.is_word(), "only cpu regs"); add_reg(r.reg()); }

  c1_RegMask add_reg    (RInfo r)               { assert(r.is_word(), "only cpu regs"); return add_reg   (r.reg()); }
  c1_RegMask remove_reg (RInfo r)               { assert(r.is_word(), "only cpu regs"); return remove_reg(r.reg()); }
  bool       contains   (RInfo r) const         { assert(r.is_word(), "only cpu regs"); return contains  (r.reg()); }
  bool       is_empty() const                   { return _mask == 0; }
  RInfo      get_first_reg() const              { return RInfo::word_reg(get_first()); }

  static void init_masks(int size);
  static const c1_RegMask empty_set()           { return _empty_set; }

  // debugging:
  void print () const PRODUCT_RETURN;
};

