/*
 * @(#)c1_ValueMap.cpp	1.8 99/11/17 15:21:12
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_c1_ValueMap.cpp.incl"


// Implementation of Bucket
//
// A bucket is a list of values and corresponding hash values. New values are added
// at the end; find searches for an equal value. 0 has values indicate that the slot
// has been killed (via a partial kill).

class Bucket: public ResourceObj {
 private:
  Values             _values;                    // the list of values hashed into the same bucket
  GrowableArray<int> _hashes;                    // the corresponding list of hash values for fast comparison
  int                _id;                        // the bucket id for printing
  int                _max_size;                  // the buckets maximum size
  int                _number_of_finds;           // the number of calls to find since the bucket was emptied
  int                _total_find_size;           // the total number of comparisons since the bucket was emptied

 public:
  // creation
  Bucket::Bucket(int id, int initial_size, int max_size)
  : _values(initial_size)
  , _hashes(initial_size)
  , _id(id)
  , _max_size(max_size)
  , _number_of_finds(0)
  , _total_find_size(0)
  {
    assert(initial_size <= max_size, "initial_size > max_size");
  }

  // accessors
  int id() const                                 { return _id; }
  int max_size() const                           { return _max_size; }
  int size() const                               { return _values.length(); }
  int number_of_finds() const                    { return _number_of_finds; }
  int total_find_size() const                    { return _total_find_size; }
  float average_find_size() const                { return (float)total_find_size() / (float)number_of_finds(); }

  // manipulation
  void kill_all();
  void kill_memory(ValueType* type);
  Value find(Value x, int hash);
};


void Bucket::kill_all() {
  _values.clear();
  _hashes.clear();
  _number_of_finds = 0;
  _total_find_size = 0;
}


void Bucket::kill_memory(ValueType* type) {
  for (int i = size() - 1; i >= 0; i--) {
    Value x = _values.at(i);
    if (x != NULL
    && (x->as_LoadField() != NULL || x->as_LoadIndexed() != NULL)
    && x->type()->tag() == type->tag()
    ) {
      // memory access => invalidate
      _values.at_put(i, NULL);
      _hashes.at_put(i, 0);
    }
  }
}


Value Bucket::find(Value x, int hash) {
  assert(x != NULL && hash != 0 && x->hash() == hash, "illegal x or hash");
  // make sure buckets don't become too large, otherwise compilation time
  // becomes too big => for now just reduce bucket if they seem to big -
  // eventually this should become more dynamic and depend on the method
  // and other factors
  if (size() >= max_size()) {
    // throw away half of the buckets content
    const int new_size = size() / 2;
    _values.trunc_to(new_size);
    _hashes.trunc_to(new_size);
    if (PrintValueNumbering) {
      tty->print_cr("warning: reduced bucket size");
    }
  }
  assert(size() < max_size(), "bucket must not exceed maximum size");
  _number_of_finds++;
  for (int i = size() - 1; i >= 0; i--) {
    _total_find_size++;
    if (_hashes.at(i) == hash) {
      Value f = _values.at(i);
      if (f->is_equal(x)) {
        if (PrintValueNumbering) {
          tty->print_cr(
            "bucket %3d [%2d] (#f = %3d, avg size = %4.1f) %s (%d -> %d)",
            id(),
            size(),
            number_of_finds(),
            average_find_size(),
            x->name(),
            x->id(),
            f->id()
          );
        }
        return f;
      }
    }
  }
  // add x to bucket
  _values.append(x);
  _hashes.append(hash);
  return x;
}


// Implementation of ValueMap
//
// A ValueMap is a list of buckets. Values are hashed into
// buckets via their hash value and an index computation.

ValueMap::ValueMap(int initial_size) : _map(initial_size) {
  if (!UseValueNumbering) return;
  const int initial_bucket_size =  4;
  const int max_bucket_size     = 15; // this should probably be more dynamic, somehow
  for (int i = 0; i < initial_size; i++) _map.append(new Bucket(i, initial_bucket_size, max_bucket_size));
}


void ValueMap::kill_all() {
  if (!UseValueNumbering) return;
  for (int i = 0; i < number_of_buckets(); i++) _map.at(i)->kill_all();
}


void ValueMap::kill_memory(ValueType* type) {
  if (!UseValueNumbering) return;
  for (int i = 0; i < number_of_buckets(); i++) _map.at(i)->kill_memory(type);
}


Value ValueMap::find(Value x) {
  if (UseValueNumbering) {
    const int h = x->hash();
    if (h != 0) {
      // 0 hash means: exclude from value numbering
      const int i = (h & 0x7FFFFFFF) % _map.length();
      return _map.at(i)->find(x, h);
    }
  }
  return x;
}


void ValueMap::print() {
  tty->print_cr("ValueMap");
}

