// @(#)c1_ValueStack.cpp	1.35 00/04/12 09:00:40
// 
// Copyright 1993-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_c1_ValueStack.cpp.incl"


// Implementation of ValueStack

ValueStack::ValueStack(int locals_size, int max_stack_size)
: _locals(locals_size, locals_size, NULL)
, _stack(max_stack_size)
, _locks(1)
{
  // nothing to do
}


ValueStack::ValueStack(ValueStack* s)
: _locals(s->locals_size(), s->locals_size(), NULL)
, _stack(s->max_stack_size())
, _locks(s->locks_size())
{
  init(s);
}


ValueStack* ValueStack::copy() {
  ValueStack* s = new ValueStack(locals_size(), max_stack_size());
  s->_stack.appendAll(&_stack);
  s->_locks.appendAll(&_locks);
  return s;
}


void ValueStack::init(ValueStack* s) {
  assert(s != NULL, "state must exist");
  assert(locals_size() == s->locals_size(), "locals sizes must correspond");
  // introduce a stack phi for each stack element of s
  for (int i = 0; i < s->stack_size();) {
    int i0 = i; // i is modified by stack_at_inc(...)
    ValueType* t = s->stack_at_inc(i)->type();
    push(t, new Phi(t, i0, s->stack_size()));
  }
  assert(stack_size() == s->stack_size(), "stack sizes must correspond");
  // introduce a lock element for each lock element of s - simply
  // copy them for now as we are only interested in the locks size
  _locks.appendAll(&s->_locks);
}


bool ValueStack::is_same(ValueStack* s) {
  assert(s != NULL, "state must exist");
  assert(locals_size() == s->locals_size(), "locals sizes must correspond");
  assert(stack_size () == s->stack_size (), "stack  sizes must correspond");
  assert(locks_size () == s->locks_size (), "locks  sizes must correspond");
  // compare each stack element with the corresponding stack element of s
  for (int i = 0; i < s->stack_size();) {
    int i0 = i; // i is modified by stack_at_inc(...)
    Value x = stack_at_inc(i0);
    Value y = s->stack_at_inc(i);
    if (x->type()->tag() != y->type()->tag()) return false;
  }
  // compare each locks element with the corresponding locks element of s
  for (int j = 0; j < s->locks_size(); j++) {
    if (_locks.at(j) != s->_locks.at(j)) return false;
  }
  return true;
}


void ValueStack::clear_locals() {
  for (int i = 0; i < _locals.length(); i++) _locals.at_put(i, NULL);
}


void ValueStack::pin_stack_locals(int index) {
  pin_stack_all();
  return;
  // for now pin all due to pinning bug (see notes.txt)
  // this function could be made more selective by taking the type into account
  for (int i = 0; i < stack_size();) {
    AccessLocal* t = stack_at_inc(i)->as_AccessLocal();
    if (t != NULL && t->index() == index) {
      t->pin();
      t->set_pinned_by_store();
    }
  }
}


void ValueStack::pin_stack_fields() {
  pin_stack_all();
  return;
  // for now pin all due to pinning bug (see notes.txt)
  // this function could be made more selective by taking the type into account
  for (int i = 0; i < stack_size();) {
    AccessField* t = stack_at_inc(i)->as_AccessField();
    if (t != NULL) t->pin();
  }
}


void ValueStack::pin_stack_indexed() {
  pin_stack_all();
  return;
  // for now pin all due to pinning bug (see notes.txt)
  // this function could be made more selective by taking the type into account
  for (int i = 0; i < stack_size();) {
    AccessIndexed* t = stack_at_inc(i)->as_AccessIndexed();
    if (t != NULL) t->pin();
  }
}


void ValueStack::pin_stack_for_state_split() {
  // extend this later to walk the trees and pin all array and field accesses
  for (int i = 0; i < stack_size();) {
    Value v = stack_at_inc(i);
    if (v->as_LoadLocal() == NULL && v->as_Constant() == NULL) v->pin();
  }
}


void ValueStack::pin_stack_all() {
  for (int i = 0; i < stack_size();) stack_at_inc(i)->pin();
}


void ValueStack::values_do(void f(Value*)) {
  for (int i = 0; i < stack_size();) {
    // apply f
    Value* ta = _stack.adr_at(i);
    Value t0 = *ta; // value before application of f
    assert(t0->as_HiWord() == NULL, "address points to hi word");
    assert(t0->type()->is_single_word() || t0 == _stack.at(i+1)->as_HiWord()->lo_word(), "stack inconsistent");
    f(ta);
    Value t1 = *ta; // value after application of f - may have changed (substitution)
    assert(t1->as_HiWord() == NULL, "address points to hi word");
    // if value changed we need to adjust HiWord if any
    if (t1 != t0) {
      assert(t1->type()->tag() == t0->type()->tag(), "types must match");
      if (t1->type()->is_double_word()) _stack.at_put(i+1, new HiWord(t1));
    }
    i += t0->type()->size();
  }
}


Values* ValueStack::pop_arguments(int argument_size) {
  assert(stack_size() >= argument_size, "stack too small or too many arguments");
  int base = stack_size() - argument_size;
  Values* args = new GrowableArray<Value>(argument_size);
  for (int i = base; i < stack_size();) args->push(stack_at_inc(i));
  truncate_stack(base);
  return args;
}


int ValueStack::lock(IRScope* scope, int bci) {
  _locks.push(bci);
  scope->set_min_number_of_locks(locks_size());
  return locks_size() - 1;
}


int ValueStack::unlock() {
  _locks.pop();
  return locks_size();
}


#ifndef PRODUCT
void ValueStack::print() {
  if (stack_is_empty()) {
    tty->print_cr("empty stack");
  } else {
    InstructionPrinter ip;
    for (int i = stack_size(); i > 0;) {
      Value t = stack_at_dec(i);
      tty->print("%2d  ", i);
      ip.print_instr(t);
      tty->cr();
    }
  }
}


void ValueStack::verify() {
  Unimplemented();
}
#endif // PRODUCT

