// @(#)c1_ValueStack.hpp	1.23 00/04/12 09:00:41
// 
// Copyright 1993-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


class ValueStack: public ResourceObj {
 private:
  Values             _locals;                    // the locals
  Values             _stack;                     // the expression stack
  GrowableArray<int> _locks;                     // the monitor stack (the monitorenter bcis)

  Value check(ValueTag tag, Value t) {
    assert(tag == t->type()->tag() || tag == objectTag && t->type()->tag() == addressTag, "types must correspond");
    return t;
  }

  Value check(ValueTag tag, Value t, Value h) {
    assert(h->as_HiWord()->lo_word() == t, "incorrect stack pair");
    return check(tag, t);
  }

 public:
  // creation
  ValueStack(int locals_size, int max_stack_size);
  ValueStack(ValueStack* s);

  // merging
  ValueStack* copy();                            // returns a copy of this w/ cleared locals
  void init(ValueStack* s);                      // initializes this w/ cleared locals & phis for stack entries of s
  bool is_same(ValueStack* s);                   // returns true if this & s's types match (w/o checking locals)

  // accessors
  int locals_size() const                        { return _locals.length(); }
  int stack_size() const                         { return _stack.length(); }
  int locks_size() const                         { return _locks.length(); }
  int max_stack_size() const                     { return _stack.capacity(); }
  bool stack_is_empty() const                    { return _stack.is_empty(); }
  bool no_active_locks() const                   { return _locks.is_empty(); }

  // locals access
  void clear_locals();                           // sets all locals to NULL

  void kill_local(int i) {
    Value x = _locals.at(i);
    if (x != NULL) {
      if (x->type()->is_double_word()) {
        HiWord* h = x->as_HiWord();
        if (h == NULL) {
          assert(_locals.at(i+1)->as_HiWord()->lo_word() == x, "locals inconsistent");
          _locals.at_put(i+1, NULL);
        } else {
          assert(_locals.at(i-1) == h->lo_word(), "stack inconsistent");
          _locals.at_put(i-1, NULL);
        }
      }
      _locals.at_put(i, NULL);
    }
  }

  Value load_local(int i) const {
    Value x = _locals.at(i);
    assert(x == NULL || x->as_HiWord() == NULL, "index points to hi word");
    assert(x == NULL || x->type()->is_single_word() || x == _locals.at(i+1)->as_HiWord()->lo_word(), "locals inconsistent");
    return x;
  }

  void store_local(int i, Value x) {
    assert(x->as_HiWord() == NULL, "x must not be HiWord");
    kill_local(i);
    if (x->type()->is_single_word()) {
      _locals.at_put(i, x);
    } else {
      kill_local(i+1);
      _locals.at_put(i, x);
      _locals.at_put(i+1, new HiWord(x));
    }
  }

  // stack access
  Value stack_at_inc(int& i) const {
    Value x = _stack.at(i);
    assert(x->as_HiWord() == NULL, "index points to hi word");
    assert(x->type()->is_single_word() || x == _stack.at(i+1)->as_HiWord()->lo_word(), "stack inconsistent");
    i += x->type()->size();
    return x;
  }

  Value stack_at_dec(int& i) const {
    Value x = _stack.at(--i);
    HiWord* h = x->as_HiWord();
    if (h != NULL) {
      x = _stack.at(--i);
      assert(x == h->lo_word(), "stack inconsistent");
    }
    return x;
  }

  // pinning support
  void pin_stack_locals(int index);
  void pin_stack_fields();
  void pin_stack_indexed();
  void pin_stack_all();
  void pin_stack_for_state_split();

  // iteration
  void values_do(void f(Value*));

  // untyped manipulation (for dup_x1, etc.)
  void clear_stack()                             { _stack.clear(); }
  void truncate_stack(int size)                  { _stack.trunc_to(size); }
  void raw_push(Value t)                         { _stack.push(t); }
  Value raw_pop()                                { return _stack.pop(); }

  // typed manipulation
  void ipush(Value t)                            { _stack.push(check(intTag    , t)); }
  void fpush(Value t)                            { _stack.push(check(floatTag  , t)); }
  void apush(Value t)                            { _stack.push(check(objectTag , t)); }
  void rpush(Value t)                            { _stack.push(check(addressTag, t)); }
  void lpush(Value t)                            { _stack.push(check(longTag   , t)); _stack.push(new HiWord(t)); }
  void dpush(Value t)                            { _stack.push(check(doubleTag , t)); _stack.push(new HiWord(t)); }

  void push(ValueType* type, Value t) {
    switch (type->tag()) {
      case intTag    : ipush(t); return;
      case longTag   : lpush(t); return;
      case floatTag  : fpush(t); return;
      case doubleTag : dpush(t); return;
      case objectTag : apush(t); return;
      case addressTag: rpush(t); return;
    }
    ShouldNotReachHere();
  }

  Value ipop()                                   { return check(intTag    , _stack.pop()); }
  Value fpop()                                   { return check(floatTag  , _stack.pop()); }
  Value apop()                                   { return check(objectTag , _stack.pop()); }
  Value rpop()                                   { return check(addressTag, _stack.pop()); }
  Value lpop()                                   { Value h = _stack.pop(); return check(longTag  , _stack.pop(), h); }
  Value dpop()                                   { Value h = _stack.pop(); return check(doubleTag, _stack.pop(), h); }

  Value pop(ValueType* type) {
    switch (type->tag()) {
      case intTag    : return ipop();
      case longTag   : return lpop();
      case floatTag  : return fpop();
      case doubleTag : return dpop();
      case objectTag : return apop();
      case addressTag: return rpop();
    }
    ShouldNotReachHere();
    return NULL;
  }

  Values* pop_arguments(int argument_size);

  // locks access
  int lock  (IRScope* scope, int bci);
  int unlock();

  // debugging
  void print()  PRODUCT_RETURN;
  void verify() PRODUCT_RETURN;
};

