/*
 * @(#)ciField.cpp	1.6 00/03/13 17:06:35
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
#include "incls/_precompiled.incl"
#include "incls/_ciField.cpp.incl"

// ciField
//
// This class represents the result of a field lookup in the VM.
// The lookup may not succeed, in which case the information in
// the ciField will be incomplete.

// ------------------------------------------------------------------
// ciField::ciField
ciField::ciField(ciInstanceKlass* klass, int index) {
  ASSERT_IN_VM;

  assert(klass->get_instanceKlass()->is_linked(), "must be linked before using its constan-pool");

  _cp_index = index;
  constantPoolHandle cpool = klass->get_instanceKlass()->constants();

  // Get the field's name, signature, and type.
  symbolHandle name = cpool->name_ref_at(index);

  int nt_index = cpool->name_and_type_ref_index_at(index);
  int sig_index = cpool->signature_ref_index_at(nt_index);
  symbolHandle signature = cpool->symbol_at(sig_index);

  _field_type = FieldType::basic_type(signature());

  // If the field is a pointer type, get the klass of the
  // field.
  if (_field_type == T_OBJECT || _field_type == T_ARRAY) {
    _field_klass = CURRENT_ENV->get_klass_by_index(klass, sig_index);
  }

  // Get the field's declared holder.
  //
  // Note: we actually create a ciInstanceKlass for this klass,
  // even though we may not need to.
  int holder_index = cpool->klass_ref_index_at(index);
  ciInstanceKlass* declared_holder =
    CURRENT_ENV->get_klass_by_index(klass, holder_index)
      ->as_instance_klass();

  // The declared holder of this field may not have been loaded.
  // Bail out with partial field information.
  if (!declared_holder->is_loaded()) {
    // _cp_index, _field_type and _field_klass have already been set.
    // The default values for _flags and _constant_value will suffice.
    // We need values for _holder, _offset,  and _is_constant,
    _holder = declared_holder;
    _offset = -1;
    _is_constant = false;
    return;
  }

  instanceKlass* loaded_decl_holder = declared_holder->get_instanceKlass();

  // Perform the field lookup.
  fieldDescriptor field_desc;
  klassOop canonical_holder = 
    loaded_decl_holder->find_field(name(), signature(), &field_desc);
  if (canonical_holder == NULL) {
    // Field lookup failed.  Will be detected by will_link.
    _holder = declared_holder;
    _offset = -1;
    _is_constant = false;
    return;
  }

  // Get the flags, offset, and canonical holder of the field.
  _flags = ciFlags(field_desc.access_flags());
  _offset = field_desc.offset();
  _holder = CURRENT_ENV->get_object(canonical_holder)->as_instance_klass();
  
  // Check to see if the field is constant.
  if (_holder->is_initialized() &&
      this->is_final() && this->is_static()) {
    // This field just may be constant.  The only cases where it will
    // not be constant are:
    //
    // 1. The field holds a non-perm-space oop.  The field is, strictly
    //    speaking, constant but we cannot embed non-perm-space oops into
    //    generated code.  For the time being we need to consider the
    //    field to be not constant.
    // 2. The field is a *special* static&final field whose value
    //    may change.  The two examples are java.lang.System.in and
    //    java.lang.System.out.
    //
    // Note: we do not yet implement the check for case 2.
    _is_constant = true;
    klassOop k = _holder->get_klassOop();
    switch(_field_type) {
    case T_BOOLEAN:
    case T_INT:
      _constant_value = ciConstant(_field_type, k->int_field(_offset));
      break;
    case T_CHAR:
      _constant_value = ciConstant(T_CHAR, k->char_field(_offset));
      break;
    case T_BYTE:
      _constant_value = ciConstant(T_BYTE, k->byte_field(_offset));
      break;
    case T_SHORT:
      _constant_value = ciConstant(T_SHORT, k->short_field(_offset));
      break;
    case T_FLOAT:
      _constant_value = ciConstant(k->float_field(_offset));
      break;
    case T_DOUBLE:
      _constant_value = ciConstant(k->double_field(_offset));
      break;
    case T_LONG:
      _constant_value = ciConstant(k->long_field(_offset));
      break;
    case T_OBJECT:
    case T_ARRAY:
      {
	oop o = k->obj_field(_offset);
	if (o == NULL) {
	  _constant_value = ciConstant(_field_type,
                                       ciNullObject::make());
	} else if (o->is_perm()) {
	  _constant_value = ciConstant(_field_type,
                                       CURRENT_ENV->get_object(o));
	} else {
	  // This object, although constant, is not yet
	  // allocated in perm space.  We choose to treat
	  // it as a non-constant.
	  _is_constant = false;
	}
      }
    }
  } else {
    _is_constant = false;
  }
}

// ------------------------------------------------------------------
// ciField::will_link
//
// Can a specific access to this field be made without causing
// link errors?
bool ciField::will_link(ciInstanceKlass* accessing_klass,
			Bytecodes::Code bc) {
  VM_ENTRY_MARK;
  FieldAccessInfo result;
  constantPoolOop c_pool =
    accessing_klass->get_instanceKlass()->constants();
  LinkResolver::resolve_field(result, c_pool, _cp_index,
                              Bytecodes::java_code(bc),
			      true, KILL_COMPILE_ON_FATAL_(false));
  return true;
}

// ------------------------------------------------------------------
// ciField::print
void ciField::print() {
  tty->print("<ciField holder=");
  _holder->print_name();
  tty->print(" offset=%d field_type=%s", _offset,
	     basictype_to_str(_field_type));
  if (_field_type == T_OBJECT || _field_type == T_ARRAY) {
    tty->print(" field_klass=");
    _field_klass->print_name();
  }
  tty->print(" is_constant=%s", bool_to_str(_is_constant));
  if (_is_constant) {
    tty->print(" constant_value=");
    _constant_value.print();
  }
  tty->print(">");
}

