/*
 * @(#)ciMethod.cpp	1.19 00/06/05 06:50:58
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
#include "incls/_precompiled.incl"
#include "incls/_ciMethod.cpp.incl"

// ciMethod
//
// This class represents a methodOop in the HotSpot virtual
// machine.

// ------------------------------------------------------------------
// ciMethod::ciMethod
//
// Loaded method.
ciMethod::ciMethod(methodOop m) : ciObject(m) {
  assert(m != NULL, "no null method");

  // These fields are always filled in in loaded methods.
  _flags = ciFlags(m->access_flags());

  // Easy to compute, so fill them in now.
  _max_stack          = m->max_stack();
  _max_locals         = m->max_locals();  
  _code_size          = m->code_size();
  _handler_count      = m->exception_table()->length() / 4;
  _uses_monitors      = m->access_flags().has_monitor_bytecodes();
  _balanced_monitors  = !_uses_monitors || m->access_flags().is_monitor_matching();
  // Lazy fields, filled in on demand.  Require allocation.
  _code               = NULL;
  _exception_handlers = NULL;
#ifdef COMPILER2
  _liveness           = NULL;
#endif

  // generating _signature may allow GC and therefore move m.
  // These fields are always filled in.
  _name = CURRENT_ENV->get_object(m->name())->as_symbol();
  _holder = CURRENT_ENV->get_object(m->method_holder())->as_instance_klass();
  ciSymbol* sig_symbol = CURRENT_ENV->get_object(m->signature())->as_symbol();
  _signature = new (CURRENT_ENV->arena()) ciSignature(_holder, sig_symbol);
}


// ------------------------------------------------------------------
// ciMethod::ciMethod
//
// Unloaded method.
ciMethod::ciMethod(ciInstanceKlass* holder,
                   ciSymbol* name,
                   ciSymbol* signature) : ciObject(ciMethodKlass::make()) {
  // These fields are always filled in.
  _name = name;
  _holder = holder;
  _signature = new (CURRENT_ENV->arena()) ciSignature(_holder, signature);
#ifdef COMPILER2
  _liveness = NULL;
#endif
}


// ------------------------------------------------------------------
// ciMethod::load_code
//
// Load the bytecodes and exception handler table for this method.
void ciMethod::load_code() {
  VM_QUICK_ENTRY_MARK;
  assert(is_loaded(), "only loaded methods have code");

  methodOop me = get_methodOop();
  Arena* arena = CURRENT_ENV->arena();

  // Load the bytecodes.
  _code = (address)arena->Amalloc(code_size());
  memcpy(_code, me->code_base(), code_size());
  
  // And load the exception table.
  typeArrayOop exc_table = me->exception_table();

  // Allocate one extra spot in our list of exceptions.  This
  // last entry will be used to represent the possibility that
  // an exception escapes the method.  See ciExceptionHandlerStream
  // for details.
  _exception_handlers =
    (ciExceptionHandler**)arena->Amalloc(sizeof(ciExceptionHandler*)
                                         * (_handler_count + 1));
  if (_handler_count > 0) {
    for (int i=0; i<_handler_count; i++) {
      int base = i*4;
      _exception_handlers[i] = new (arena) ciExceptionHandler(
                                holder(),
            /* start    */      exc_table->int_at(base),
            /* limit    */      exc_table->int_at(base+1),
            /* goto pc  */      exc_table->int_at(base+2),
            /* cp index */      exc_table->int_at(base+3));
    }
  }

  // Put an entry at the end of our list to represent the possibility
  // of exceptional exit.
  _exception_handlers[_handler_count] =
    new (arena) ciExceptionHandler(holder(), 0, code_size(), -1, 0);
}

// ------------------------------------------------------------------
// ciMethod::has_linenumber_table
//
// length unknown until decompression
bool    ciMethod::has_linenumber_table() const {
  check_is_loaded();
  VM_ENTRY_MARK;
  return get_methodOop()->has_linenumber_table();
}

// ------------------------------------------------------------------
// ciMethod::compressed_linenumber_table
u_char* ciMethod::compressed_linenumber_table() const {
  check_is_loaded();
  VM_ENTRY_MARK;
  return get_methodOop()->compressed_linenumber_table();
}

// ------------------------------------------------------------------
// ciMethod::vtable_index
//
// Get the position of this method's entry in the vtable, if any.
int ciMethod::vtable_index() {
  check_is_loaded();
  VM_ENTRY_MARK;
  return get_methodOop()->vtable_index();
}

// ------------------------------------------------------------------
// ciMethod::native_entry
//
// Get the address of this method's native code, if any.
address ciMethod::native_entry() {
  check_is_loaded();
  assert(flags().is_native(), "must be native method");
  VM_ENTRY_MARK;
  methodOop method = get_methodOop();
  address entry = method->native_function();
  assert(entry != NULL, "must be valid entry point");
  return entry;
}

// ------------------------------------------------------------------
// ciMethod::interpreter_entry
//
// Get the entry point for running this method in the interpreter.
address ciMethod::interpreter_entry() {
  check_is_loaded();
  VM_ENTRY_MARK;
  return AbstractInterpreter::entry_for_method(get_methodOop());
}


ciMethod::IntrinsicId ciMethod::intrinsic_id() const {
  VM_ENTRY_MARK;
  switch (get_methodOop()->intrinsic_id()) {
    case methodOopDesc::_hash     : return _hash;
    case methodOopDesc::_dsin     : return _dsin;
    case methodOopDesc::_dcos     : return _dcos;
    case methodOopDesc::_dsqrt    : return _dsqrt;
    case methodOopDesc::_arraycopy: return _arraycopy;
    case methodOopDesc::_compareTo: return _compareTo;
    case methodOopDesc::_none     : return _none;
  }
  ShouldNotReachHere();
  return _none;
}


// ------------------------------------------------------------------
// ciMethod::uses_balanced_monitors
//
// Does this method use monitors in a strict stack-disciplined manner?
bool ciMethod::has_balanced_monitors() {
  check_is_loaded();
  if (_balanced_monitors) return true;

  // Analyze the method to see if monitors are used properly.
  VM_ENTRY_MARK;
  methodOop method = get_methodOop();
  assert(method->has_monitor_bytecodes(), "should have checked this");

  // Check to see if a previous compilation computed the
  // monitor-matching analysis.
  if (method->guaranteed_monitor_matching()) {
    _balanced_monitors = true;
    return true;
  }

  {
    EXCEPTION_MARK;
    ResourceMark rm;
    GeneratePairingInfo gpi(method);
    gpi.compute_map();
    if (!gpi.monitor_safe()) {
      return false;
    }
    method->set_guaranteed_monitor_matching();
    _balanced_monitors = true;
  }
  return true;
}

#ifdef COMPILER2
class GenerateOopMapForBCI : public GenerateOopMap {
private:
  int          _bci;
  BitMap*      _result;
  
public:
  GenerateOopMapForBCI(methodHandle method, int bci, BitMap* result)
  : GenerateOopMap(method) {
    _bci = bci;
    _result = result;
  }

  void compute_map();
  bool possible_gc_point(BytecodeStream* bcs) { return false; }
  bool report_results() const { return false; }
  void fill_stackmap_prolog(int nof_gc_points) { /* do nothing */ }
  void fill_stackmap_epilog() { /* do nothing */ }
  void fill_stackmap_for_opcodes(BytecodeStream* bcs,
				 CellTypeState* vars,
				 CellTypeState* stack,
				 int stack_top);
  void fill_init_vars(GrowableArray<int>* init_vars) { /* do nothing */ }
};

void GenerateOopMapForBCI::compute_map() {
  ResourceMark rm;
  GenerateOopMap::compute_map();
  result_for_basicblock(_bci);
}

void GenerateOopMapForBCI::fill_stackmap_for_opcodes(BytecodeStream* bcs,
						     CellTypeState* vars,
						     CellTypeState* stack,
						     int stack_top) {
  if (bcs->bci() == _bci) {
    assert(stack_top == 0, "empty stack");
    int max_locals = _method->max_locals();
    
    for (int i=0; i<max_locals; i++) {
      _result->at_put(i, vars[i].is_reference());
    }
  }
}


// ------------------------------------------------------------------
// ciMethod::oops_at_bci
//
// Which local variables are oops at a specific bci?
//
// Implementation note: Expensive.  Runs generateOopMap for each
// call
BitMap ciMethod::oops_at_bci(int bci) {
  assert(is_loaded(), "not loaded");
  VM_ENTRY_MARK;

  Arena* arena = CURRENT_ENV->arena();
  BitMap bm((unsigned int*)arena->Amalloc((_max_locals+7)>>3), _max_locals);
  {
    EXCEPTION_MARK;
    ResourceMark rm;
    GenerateOopMapForBCI gen(get_methodOop(), bci, &bm);
    gen.compute_map();
  }
  return bm;
}
#endif // COMPILER2


#ifdef COMPILER1
ciOopMap* ciMethod::all_oop_maps() {
  assert(!is_abstract() && !is_native(), "cannot compute oop maps for abstract or native methods");
  // do nothing if possible
  if (code_size() == 0 || max_locals() == 0) return NULL;
  // compute oop maps
  check_is_loaded();
  VM_ENTRY_MARK;
  Arena* arena = CURRENT_ENV->arena();
  ciGenerateOopMap omg(arena, get_methodOop());
  omg.compute_map();  
  ciOopMap* map = omg.get_oop_maps();
  assert(map != NULL, "map must exist");
  return map;
}


ciAddressMap* ciMethod::address_map_for_bci(int bci) {
  assert(!is_abstract() && !is_native(), "cannot compute address maps for abstract or native methods");
  // First check if it is a method where the stackmap is always empty
  if (code_size() == 0 || max_locals() == 0) return NULL;
  // compute address map
  check_is_loaded();
  VM_ENTRY_MARK;
  Arena* arena = CURRENT_ENV->arena();
  ciAddressMapGenerator amg(arena, get_methodOop(), bci);
  amg.compute_map();
  ciAddressMap* map = amg.get_address_map();
  assert(map != NULL, "map must exist");
  return map;
}
#endif // COMPILER1


// ------------------------------------------------------------------
// ciMethod::liveness_at_bci
//
// Which local variables are live at a specific bci?
#ifdef COMPILER2
BitMap ciMethod::liveness_at_bci(int bci) {
  check_is_loaded();
  VM_ENTRY_MARK;
  if (_liveness == NULL) {
    // Create the liveness analyzer.
    Arena* arena = CURRENT_ENV->arena();
    _liveness = new (arena) MethodLiveness(arena, handle());
    _liveness->compute_liveness();
  }
  return _liveness->get_liveness_at(bci);
}
#endif

// ------------------------------------------------------------------
// ciMethod::init_vars
//
// Which local variables need to be initialized to zero?
#ifdef COMPILER2
BitMap ciMethod::init_vars() {
  check_is_loaded();
  Unimplemented();
  return BitMap(NULL,0);
}
#endif

// ------------------------------------------------------------------
// ciMethod::call_profile_at_bci
//
// Get the ciCallProfile for the invocation of this method.
ciCallProfile ciMethod::call_profile_at_bci(int bci) {
  check_is_loaded();
  VM_ENTRY_MARK;
  methodOop caller = get_methodOop();
  KlassHandle prof_klassh;
  methodHandle prof_methodh =
    CallProfile::resolve_method(caller, bci, prof_klassh);
  if (prof_methodh.not_null()) {
    ciMethod* m = CURRENT_ENV->get_object(prof_methodh())->as_method();
    ciKlass*  k = CURRENT_ENV->get_object(prof_klassh())->as_klass();
    return ciCallProfile(m, k);
  } else {
    return ciCallProfile(NULL, NULL);
  }
}

// ------------------------------------------------------------------
// ciMethod::null_cast_cache
u2 ciMethod::null_cast_cache() {
  check_is_loaded();
  VM_ENTRY_MARK;
#ifdef COMPILER2
  return get_methodOop()->null_cast_cache();
#else
  ShouldNotReachHere();
  return 0;
#endif
}

// ------------------------------------------------------------------
// ciMethod::null_cast_seen
bool ciMethod::null_cast_seen() {
  check_is_loaded();
  VM_ENTRY_MARK;
#ifdef COMPILER2
  return (get_methodOop()->null_cast_seen() != 0);
#else
  ShouldNotReachHere();
  return false;
#endif
}

// ------------------------------------------------------------------
bool ciMethod::is_not_reached(int bci) {
  address bcp = code() + bci;
  if (!Bytecode_at(bcp)->must_rewrite())
    return false;               // might have been reached

  // the bytecode might not be rewritten if the method is an accessor, etc.
  VM_ENTRY_MARK;
  address ientry = get_methodOop()->interpreter_entry();
  if (ientry != AbstractInterpreter::entry_for_kind(AbstractInterpreter::zerolocals) &&
      ientry != AbstractInterpreter::entry_for_kind(AbstractInterpreter::zerolocals_synchronized))
    return false;           // interpreter does not run this method!

  // otherwise, we can be sure this bytecode has never been executed
  return true;
}

// ------------------------------------------------------------------
// ciMethod::find_monomorphic_target
//
// Given a certain calling environment, find the monomorphic target
// for the call.  Return NULL if the call is not monomorphic in
// its calling environment.
ciMethod* ciMethod::find_monomorphic_target(ciKlass* caller,
					    ciKlass* callee_holder,
					    ciKlass* actual_recv) {
  check_is_loaded();
  VM_ENTRY_MARK;
  CHAResult* result =
    CHA::analyze_call(caller->get_klassOop(),
		      callee_holder->get_klassOop(),
		      actual_recv->get_klassOop(),
		      name()->get_symbolOop(),
		      signature()->get_symbolOop());
  if (!result->is_monomorphic()) {
    return NULL;
  }
  methodHandle target = result->monomorphic_target();
  if (target() == NULL) {
    return NULL;
  }
  if (target() == get_methodOop()) {
    return this;
  }
  return CURRENT_ENV->get_object(target())->as_method();
}

// ------------------------------------------------------------------
// ciMethod::interpreter_invocation_count
int ciMethod::interpreter_invocation_count() {
  VM_ENTRY_MARK;
#ifdef COMPILER2
  return get_methodOop()->interpreter_invocation_count();
#else
  ShouldNotReachHere();
  return 0;
#endif
}

// ------------------------------------------------------------------
// ciMethod::interpreter_call_site_count
int ciMethod::interpreter_call_site_count(int bci) {
  VM_ENTRY_MARK;
#ifdef COMPILER2
  methodOop method = get_methodOop();
  int bcp = (int)method->bcp_from(bci);
  bcp = bcp & right_n_bits(CallFreqTableSize);
  int cnt = SharedInfo::callTable[bcp];
  return cnt;
#else
  ShouldNotReachHere();
  return 0;
#endif
}

// ------------------------------------------------------------------
// ciMethod::will_link
//
// Will this method link in a specific calling context?
bool ciMethod::will_link(ciKlass* accessing_klass,
			 ciKlass* declared_method_holder,
			 Bytecodes::Code bc) {
  if (!is_loaded()) {
    // Method lookup failed.
    return false;
  }

  // The link checks have been front-loaded into the get_method
  // call.  This method (ciMethod::will_link()) will be removed
  // in the future.

  return true;
}

// ------------------------------------------------------------------
// ciMethod::should_exclude
//
// Should this method be excluded from compilation?
bool ciMethod::should_exclude() {
  check_is_loaded();
  VM_ENTRY_MARK;
  return CompilerOracle::should_exclude(get_methodOop());
}

// ------------------------------------------------------------------
// ciMethod::should_print_assembly
//
// Should the compiler print the generated code for this method?
bool ciMethod::should_print_assembly() {
  check_is_loaded();
  VM_ENTRY_MARK;
  return CompilerOracle::should_print(get_methodOop());
}

// ------------------------------------------------------------------
// ciMethod::break_at_execute
//
// Should the compiler insert a breakpoint into the generated code
// method?
bool ciMethod::break_at_execute() {
  check_is_loaded();
  VM_ENTRY_MARK;
  return CompilerOracle::should_break_at(get_methodOop());
}

// ------------------------------------------------------------------
// ciMethod::can_be_compiled
//
// Have previous compilations of this method succeeded?
bool ciMethod::can_be_compiled() {
  check_is_loaded();
  Unimplemented()
  return true;
}

// ------------------------------------------------------------------
// ciMethod::can_be_osr_compiled
//
// Have previous compilations of this method succeeded?
//
// Implementation note: the VM does not currently keep track
// of failed OSR compilations per bci.  The entry_bci parameter
// is currently unused.
bool ciMethod::can_be_osr_compiled(int entry_bci) {
  check_is_loaded();
  VM_ENTRY_MARK;
  return !get_methodOop()->access_flags().is_not_osr_compilable();
}

// ------------------------------------------------------------------
// ciMethod::print_bytes
//
// Print the bytecodes for this method.
void ciMethod::print_bytes() {
  check_is_loaded();
  GUARDED_VM_ENTRY(get_methodOop()->print_codes();)
}


bool ciMethod::has_loops() const {
  check_is_loaded();
  VM_ENTRY_MARK;
  return get_methodOop()->has_loops();
}


// ------------------------------------------------------------------
// ciMethod::print_bytes
//
// Print a range of the bytecodes for this method.
void ciMethod::print_bytes(int from, int to) {
  check_is_loaded();
  Unimplemented();
}

// ------------------------------------------------------------------
// ciMethod::print_name
//
// Print the name of this method, including signature and some flags.
void ciMethod::print_name() {
  check_is_loaded();
  GUARDED_VM_ENTRY(get_methodOop()->print_name(tty);)
}

// ------------------------------------------------------------------
// ciMethod::print_short_name
//
// Print the name of this method, without signature.
void ciMethod::print_short_name() {
  check_is_loaded();
  GUARDED_VM_ENTRY(get_methodOop()->print_short_name(tty);)
}

// ------------------------------------------------------------------
// ciMethod::print_impl
//
// Implementation of the print method.
void ciMethod::print_impl() {
  ciObject::print_impl();
  tty->print(" name=");
  name()->print_symbol();
  tty->print(" holder=");
  holder()->print_name();
  tty->print(" signature=");
  signature()->print_signature();
  if (is_loaded()) {
    tty->print(" loaded=true flags=");
    flags().print_member_flags();
  } else {
    tty->print(" loaded=false");
  }
}


