/*
 * @(#)ciMethod.hpp	1.17 01/02/09 14:44:35
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

#ifdef COMPILER1
class ciOopMap;
class ciAddressMap;
#endif
#ifdef COMPILER2
class MethodLiveness;
class BitMap;
#endif


// ciMethod
//
// This class represents a methodOop in the HotSpot virtual
// machine.
class ciMethod : public ciObject {
  friend class CompileBroker;
  CI_PACKAGE_ACCESS

 private:
  // General method information.
  ciFlags          _flags;
  ciSymbol*        _name;
  ciInstanceKlass* _holder;
  ciSignature*     _signature;

  // Code attributes.
  int _code_size;
  int _max_stack;
  int _max_locals;
  int _handler_count;

  bool _uses_monitors;
  bool _balanced_monitors;

  // Lazy fields, filled in on demand
  address              _code;
  ciExceptionHandler** _exception_handlers;

  // Optional liveness analyzer.
#ifdef COMPILER2
  MethodLiveness* _liveness;
#endif

  ciMethod(methodOop m);
  ciMethod(ciInstanceKlass* holder, ciSymbol* name, ciSymbol* signature);

  methodOop get_methodOop() const {
    methodOop m = (methodOop)get_oop();
    assert(m != NULL, "illegal use of unloaded method");
    return m;
  }

  oop loader() const                             { return _holder->loader(); }

  const char* type_name()                        { return "ciMethod"; }

  void print_impl();

  void load_code();

  void check_is_loaded() const                   { assert(is_loaded(), "not loaded"); }

 public:
  // Basic method information.
  ciFlags flags() const                          { check_is_loaded(); return _flags; }
  ciSymbol* name() const                         { return _name; }
  ciInstanceKlass* holder() const                { return _holder; }
  
  // Signature information.
  ciSignature* signature() const                 { return _signature; }
  BasicType    return_type() const               { return _signature->return_type(); }
  ciKlass*     return_klass() const              { return _signature->return_klass(); }
  int          arg_size_no_receiver() const      { return _signature->size(); }
  int          arg_size() const                  { return _signature->size() + (_flags.is_static() ? 0 : 1); }

  // Method code and related information.
  address code()                                 { if (_code == NULL) load_code(); return _code; }
  int code_size() const                          { check_is_loaded(); return _code_size; }
  int max_stack() const                          { check_is_loaded(); return _max_stack; }
  int max_locals() const                         { check_is_loaded(); return _max_locals; }
  bool has_exception_handlers() const            { check_is_loaded(); return _handler_count > 0; }
  int exception_table_length() const             { check_is_loaded(); return _handler_count; }

  Bytecodes::Code java_code_at_bci(int bci) {
    address bcp = code() + bci;
    return Bytecodes::java_code(Bytecodes::cast(*bcp));
  }

  // True if bytecode is never reached
  bool is_not_reached(int bci);

  bool    has_linenumber_table() const;          // length unknown until decompression
  u_char* compressed_linenumber_table() const;   // not preserved by gc

  // Runtime information.
  int           vtable_index();
  address       native_entry();
  address       interpreter_entry();

  // Miscellaneous information.
  enum IntrinsicId {
    _hash,                                       // java.lang.Object.hashCode
    _dsin,                                       // java.lang.Math.sin
    _dcos,                                       // java.lang.Math.cos
    _dsqrt,                                      // java.lang.Math.sqrt
    _arraycopy,                                  // java.lang.System.arraycopy
    _compareTo,                                  // java.lang.String.compareTo
    _none                                        // not an intrinsic (default answer)
  };
  IntrinsicId intrinsic_id() const;              // returns _none if not an intrinsic

  // Analysis and profiling.
  //
  // Usage note: liveness_at_bci and init_vars should be wrapped in ResourceMarks.
  bool          uses_monitors() const            { return _uses_monitors; } // this one should go away, it has a misleading name
  bool          has_monitor_bytecodes() const    { return _uses_monitors; }
  bool          has_balanced_monitors();
#ifdef COMPILER1
  ciOopMap*     all_oop_maps();
  ciAddressMap* address_map_for_bci(int bci);
#endif
#ifdef COMPILER2
  BitMap        liveness_at_bci(int bci);
  BitMap        init_vars();
  BitMap        oops_at_bci(int bci);
#endif
  ciCallProfile call_profile_at_bci(int bci);
  bool          null_cast_seen();
  u2            null_cast_cache();

  int           interpreter_invocation_count();
  int           interpreter_call_site_count(int bci);

  // Given a certain calling environment, find the monomorphic target
  // for the call.  Return NULL if the call is not monomorphic in
  // its calling environment.
  ciMethod* find_monomorphic_target(ciKlass* caller, ciKlass* callee_holder, ciKlass* actual_receiver);

  // Compilation directives
  bool will_link(ciKlass* accessing_klass, 
		 ciKlass* declared_method_holder,
		 Bytecodes::Code bc);
  bool should_exclude();
  bool should_print_assembly();
  bool break_at_execute();
  bool can_be_compiled();
  bool can_be_osr_compiled(int entry_bci);

  // What kind of ciObject is this?
  bool is_method()                               { return true; }

  // Java access flags
  bool is_public      () const                   { return flags().is_public(); }
  bool is_private     () const                   { return flags().is_private(); }
  bool is_protected   () const                   { return flags().is_protected(); }
  bool is_static      () const                   { return flags().is_static(); }
  bool is_final       () const                   { return flags().is_final(); }
  bool is_synchronized() const                   { return flags().is_synchronized(); }
  bool is_native      () const                   { return flags().is_native(); }
  bool is_interface   () const                   { return flags().is_interface(); }
  bool is_abstract    () const                   { return flags().is_abstract(); }
  bool is_strict      () const                   { return flags().is_strict(); }

  // Other flags
  bool is_final_method() const                   { return is_final() || holder()->is_final(); }
  bool has_loops      () const;

  // Print the bytecodes of this method.
  void print_bytes();
  void print_bytes(int from, int to);

  // Print the name of this method in various incarnations.
  void print_name();
  void print_short_name();
};

