/*
 * @(#)ciObject.hpp	1.5 00/01/13 16:04:09
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// ciObject
//
// This class represents an oop in the HotSpot virtual machine.
// Its subclasses are structured in a hierarchy which mirrors
// an aggregate of the VM's oop and klass hierarchies (see
// oopHierarchy.hpp).  Each instance of ciObject holds a handle
// to a corresponding oop on the VM side and provides routines
// for accessing the information in its oop.  By using the ciObject
// hierarchy for accessing oops in the VM, the compiler ensures
// that it is safe with respect to garbage collection; that is,
// GC and compilation can proceed independently without
// interference.
//
// Within the VM, the oop and klass hierarchies are separate.
// The compiler interface does not preserve this separation --
// the distinction between `klassOop' and `Klass' are not
// reflected in the interface and instead the Klass hierarchy
// is directly modeled as the subclasses of ciKlass.
class ciObject : public ResourceObj {
  CI_PACKAGE_ACCESS
  friend class DebugInformationRecorder;

private:
  // A JNI handle referring to an oop in the VM.  This
  // handle may, in a small set of cases, correctly be NULL.
  jobject  _handle;
  ciKlass* _klass;

protected:
  ciObject();
  ciObject(oop o);
  ciObject(ciKlass* klass);

  jobject      handle()  const { return _handle; }
  oop          get_oop() const;

  // Virtual behavior of the print() method.
  virtual void print_impl() {}

  virtual const char* type_name() { return "ciObject"; }

public:
  // The klass of this ciObject.
  ciKlass* klass();

  // Are two ciObjects equal?
  bool equals(ciObject* obj);

  // A hash value for the convenience of compilers.
  int hash();

  // The address which the compiler should embed into the
  // generated code to represent this oop.  This address
  // is not the true address of the oop -- it will get patched
  // during nmethod creation.
  //
  // Usage note: no address arithmetic allowed.  Oop must
  // be registered with the oopRecorder.
  jobject encoding();

  // What kind of ciObject is this?
  virtual bool is_null_object() const       { return false; }
  virtual bool is_instance()                { return false; }
  virtual bool is_method()                  { return false; }
  virtual bool is_array()                   { return false; }
  virtual bool is_obj_array()               { return false; }
  virtual bool is_type_array()              { return false; }
  virtual bool is_symbol()                  { return false; }
  virtual bool is_klass()                   { return false; }
  virtual bool is_instance_klass()          { return false; }
  virtual bool is_method_klass()            { return false; }
  virtual bool is_array_klass()             { return false; }
  virtual bool is_obj_array_klass()         { return false; }
  virtual bool is_type_array_klass()        { return false; }
  virtual bool is_symbol_klass()            { return false; }
  virtual bool is_klass_klass()             { return false; }
  virtual bool is_instance_klass_klass()    { return false; }
  virtual bool is_array_klass_klass()       { return false; }
  virtual bool is_obj_array_klass_klass()   { return false; }
  virtual bool is_type_array_klass_klass()  { return false; }

  // Is this ciObject a Java Language Object?  That is,
  // is the ciObject an instance or an array
  virtual bool is_java_object()             { return false; }

  // Does this ciObject represent a Java Language class?
  // That is, is the ciObject an instanceKlass or arrayKlass?
  virtual bool is_java_klass()              { return false; }

  // Is this ciObject the ciInstanceKlass representing
  // java.lang.Object()?
  virtual bool is_java_lang_Object()        { return false; }

  // Does this ciObject refer to a real oop in the VM?
  //
  // Note: some ciObjects refer to oops which have yet to be
  // created.  We refer to these as "unloaded".  Specifically,
  // there are unloaded ciMethods, ciObjArrayKlasses, and
  // ciInstanceKlasses.  By convention the ciNullObject is
  // considered loaded.
  bool is_loaded() const {
    return handle() != NULL && !is_null_object();
  }

  // Subclass casting with assertions.
  ciNullObject*            as_null_object() {
    assert(is_null_object(), "bad cast");
    return (ciNullObject*)this;
  }
  ciInstance*              as_instance() {
    assert(is_instance(), "bad cast");
    return (ciInstance*)this;
  }
  ciMethod*                as_method() {
    assert(is_method(), "bad cast");
    return (ciMethod*)this;
  }
  ciArray*                 as_array() {
    assert(is_array(), "bad cast");
    return (ciArray*)this;
  }
  ciObjArray*              as_obj_array() {
    assert(is_obj_array(), "bad cast");
    return (ciObjArray*)this;
  }
  ciTypeArray*             as_type_array() {
    assert(is_type_array(), "bad cast");
    return (ciTypeArray*)this;
  }
  ciSymbol*                as_symbol() {
    assert(is_symbol(), "bad cast");
    return (ciSymbol*)this;
  }
  ciKlass*                 as_klass() {
    assert(is_klass(), "bad cast");
    return (ciKlass*)this;
  }
  ciInstanceKlass*         as_instance_klass() {
    assert(is_instance_klass(), "bad cast");
    return (ciInstanceKlass*)this;
  }
  ciMethodKlass*           as_method_klass() {
    assert(is_method_klass(), "bad cast");
    return (ciMethodKlass*)this;
  }
  ciArrayKlass*            as_array_klass() {
    assert(is_array_klass(), "bad cast");
    return (ciArrayKlass*)this;
  }
  ciObjArrayKlass*         as_obj_array_klass() {
    assert(is_obj_array_klass(), "bad cast");
    return (ciObjArrayKlass*)this;
  }
  ciTypeArrayKlass*        as_type_array_klass() {
    assert(is_type_array_klass(), "bad cast");
    return (ciTypeArrayKlass*)this;
  }
  ciSymbolKlass*           as_symbol_klass() {
    assert(is_symbol_klass(), "bad cast");
    return (ciSymbolKlass*)this;
  }
  ciKlassKlass*            as_klass_klass() {
    assert(is_klass_klass(), "bad cast");
    return (ciKlassKlass*)this;
  }
  ciInstanceKlassKlass*    as_instance_klass_klass() {
    assert(is_instance_klass_klass(), "bad cast");
    return (ciInstanceKlassKlass*)this;
  }
  ciArrayKlassKlass*       as_array_klass_klass() {
    assert(is_array_klass_klass(), "bad cast");
    return (ciArrayKlassKlass*)this;
  }
  ciObjArrayKlassKlass*    as_obj_array_klass_klass() {
    assert(is_obj_array_klass_klass(), "bad cast");
    return (ciObjArrayKlassKlass*)this;
  }
  ciTypeArrayKlassKlass*   as_type_array_klass_klass() {
    assert(is_type_array_klass_klass(), "bad cast");
    return (ciTypeArrayKlassKlass*)this;
  }

  // Print debugging output about this ciObject.
  void print();

  // Print debugging output about the oop this ciObject represents.
  void print_oop();
};

