/*
 * @(#)ciObjectFactory.cpp	1.7 00/02/09 11:19:45
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
#include "incls/_precompiled.incl"
#include "incls/_ciObjectFactory.cpp.incl"

// ciObjectFactory
//
// This class handles requests for the creation of new instances
// of ciObject and its subclasses.  It contains a caching mechanism
// which ensures that for each oop, at most one ciObject is created.
// This invariant allows more efficient implementation of ciObject.
//
// Implementation note: the oop->ciObject mapping is represented as
// a table stored in an array.  Even though objects are moved
// by the garbage collector, the compactor preserves their relative
// order; address comparison of oops (in perm space) is safe so long
// as we prohibit GC during our comparisons.  We currently use binary
// search to find the oop in the table, and inserting a new oop
// into the table may be costly.  If this cost ends up being
// problematic the underlying data structure can be switched to some
// sort of balanced binary tree.

GrowableArray<ciObject*>* ciObjectFactory::_shared_ci_objects = NULL;

// ------------------------------------------------------------------
// ciObjectFactory::ciObjectFactory
ciObjectFactory::ciObjectFactory(Arena* arena,
                                 int expected_size) {
  if (!shared_is_initialized()) {
    init_shared_objects();
  }
  _arena = arena;
  _ci_objects = new (arena) GrowableArray<ciObject*>(arena, expected_size, 0, NULL);
  _ci_objects->appendAll(_shared_ci_objects);

  _unloaded_methods = new (arena) GrowableArray<ciMethod*>(arena, 4, 0, NULL);
  _unloaded_klasses = new (arena) GrowableArray<ciKlass*>(arena, 8, 0, NULL);
}

// ------------------------------------------------------------------
// ciObjectFactory::ciObjectFactory
void ciObjectFactory::init_shared_objects() {
  ASSERT_IN_VM;
  ciEnv* env = ciEnv::current();

  Arena* save_env_arena = env->arena();

  env->_factory = this;

  // Create an arena for the shared ciObjects.
  _arena = new Arena();
  
  env->_arena = _arena;

  _ci_objects = new (_arena) GrowableArray<ciObject*>(_arena, 64, 0, NULL);

  ciEnv::_null_object_instance = new (_arena) ciNullObject();
  ciEnv::_method_klass_instance =
    get(Universe::methodKlassObj())->as_method_klass();
  ciEnv::_symbol_klass_instance =
    get(Universe::symbolKlassObj())->as_symbol_klass();
  ciEnv::_klass_klass_instance =
    get(Universe::klassKlassObj())->as_klass_klass();
  ciEnv::_instance_klass_klass_instance =
    get(Universe::instanceKlassKlassObj())
      ->as_instance_klass_klass();
  ciEnv::_type_array_klass_klass_instance =
    get(Universe::typeArrayKlassKlassObj())
      ->as_type_array_klass_klass();
  ciEnv::_obj_array_klass_klass_instance =
    get(Universe::objArrayKlassKlassObj())
      ->as_obj_array_klass_klass();
  ciEnv::_ArrayStoreException =
    get(SystemDictionary::ArrayStoreException_klass())
      ->as_instance_klass();
  ciEnv::_Class =
    get(SystemDictionary::class_klass())
      ->as_instance_klass();
  ciEnv::_ClassCastException =
    get(SystemDictionary::ClassCastException_klass())
      ->as_instance_klass();
  ciEnv::_Object =
    get(SystemDictionary::object_klass())
      ->as_instance_klass();
  ciEnv::_OutOfMemoryError =
    get(SystemDictionary::OutOfMemoryError_klass())
      ->as_instance_klass();
  ciEnv::_String =
    get(SystemDictionary::string_klass())
      ->as_instance_klass(); 

  get(Universe::boolArrayKlassObj());
  get(Universe::charArrayKlassObj());
  get(Universe::singleArrayKlassObj());
  get(Universe::doubleArrayKlassObj());
  get(Universe::byteArrayKlassObj());
  get(Universe::shortArrayKlassObj());
  get(Universe::intArrayKlassObj());
  get(Universe::longArrayKlassObj());

  env->_arena = save_env_arena;

  _shared_ci_objects = _ci_objects;
}

// ------------------------------------------------------------------
// ciObjectFactory::get
//
// Get the ciObject corresponding to some oop.  If the ciObject has
// already been created, it is returned.  Otherwise, a new ciObject
// is created.
ciObject* ciObjectFactory::get(oop key) {
  ASSERT_IN_VM;
#ifdef ASSERT
  oop last = NULL;
  for (int j = 0; j< _ci_objects->length(); j++) {
    oop o = JNIHandles::resolve(_ci_objects->at(j)->handle());
    assert(last < o, "out of order");
    last = o;
  }
#endif ASSERT
  int len = _ci_objects->length();
  int index = find(key);
#ifdef ASSERT
  for (int i=0; i<_ci_objects->length(); i++) {
    if (JNIHandles::resolve(_ci_objects->at(i)->handle()) == key) {
      assert(index == i, " bad lookup");
    }
  }
#endif
  if (index >= _ci_objects->length() ||
      JNIHandles::resolve(_ci_objects->at(index)->handle()) != key) {
    // The ciObject does not yet exist.  Create it and insert it
    // into the cache.
    ciObject* new_object = create_new_object(key);
    if (len != _ci_objects->length()) {
      // creating the new object has recursively entered new objects
      // into the table.  We need to recompute our index.
      index = find(key);
    }
    insert(index, new_object);
    return new_object;
  }
  return _ci_objects->at(index);
}

// ------------------------------------------------------------------
// ciObjectFactory::create_new_object
//
// Create a new ciObject from an oop.
//
// Implementation note: this functionality could be virtual behavior
// of the oop itself.  For now, we explicitly marshal the object.
ciObject* ciObjectFactory::create_new_object(oop o) {
  if (o->is_symbol()) {
    return new (arena()) ciSymbol((symbolOop)o);
  } else if (o->is_klass()) {
    Klass* k = ((klassOop)o)->klass_part();
    if (k->oop_is_instance()) {
      return new (arena()) ciInstanceKlass((klassOop)o);
    } else if (k->oop_is_objArray()) {
      return new (arena()) ciObjArrayKlass((klassOop)o);
    } else if (k->oop_is_typeArray()) {
      return new (arena()) ciTypeArrayKlass((klassOop)o);
    } else if (k->oop_is_method()) {
      return new (arena()) ciMethodKlass((klassOop)o);
    } else if (k->oop_is_symbol()) {
      return new (arena()) ciSymbolKlass((klassOop)o);
    } else if (k->oop_is_klass()) {
      if (k->oop_is_objArrayKlass()) {
        return new (arena()) ciObjArrayKlassKlass((klassOop)o);
      } else if (k->oop_is_typeArrayKlass()) {
        return new (arena()) ciTypeArrayKlassKlass((klassOop)o);
      } else if (k->oop_is_instanceKlass()) {
        return new (arena()) ciInstanceKlassKlass((klassOop)o);
      } else {
        assert(o == Universe::klassKlassObj(), "bad klassKlass");
        return new (arena()) ciKlassKlass((klassOop)o);
      }
    }
  } else if (o->is_method()) {
    return new (arena()) ciMethod((methodOop)o);
  } else if (o->is_instance()) {
    return new (arena()) ciInstance((instanceOop)o);
  } else if (o->is_objArray()) {
    return new (arena()) ciObjArray((objArrayOop)o);
  } else if (o->is_typeArray()) {
    return new (arena()) ciTypeArray((typeArrayOop)o);
  }

  // The oop is of some type not supported by the compiler interface.
  ShouldNotReachHere();
  return NULL;
}

//------------------------------------------------------------------
// ciObjectFactory::get_unloaded_method
//
// Get the ciMethod representing an unloaded/unfound method.
//
// Implementation note: unloaded methods are currently stored in
// an unordered array, requiring a linear-time lookup for each
// unloaded method.  This may need to change.
ciMethod* ciObjectFactory::get_unloaded_method(ciInstanceKlass* holder,
                                               ciSymbol*        name,
                                               ciSymbol*        signature) {
  for (int i=0; i<_unloaded_methods->length(); i++) {
    ciMethod* entry = _unloaded_methods->at(i);
    if (entry->holder()->equals(holder) &&
        entry->name()->equals(name) &&
        entry->signature()->as_symbol()->equals(signature)) {
      // We've found a match.
      return entry;
    }
  }

  // This is a new unloaded method.  Create it and stick it in
  // the cache.
  ciMethod* new_method = new (arena()) ciMethod(holder, name, signature);

  _unloaded_methods->append(new_method);

  return new_method;
}

//------------------------------------------------------------------
// ciObjectFactory::get_unloaded_klass
//
// Get a ciKlass representing an unloaded klass.
//
// Implementation note: unloaded klasses are currently stored in
// an unordered array, requiring a linear-time lookup for each
// unloaded klass.  This may need to change.
ciKlass* ciObjectFactory::get_unloaded_klass(ciKlass* accessing_klass,
                                             ciSymbol* name) {
  EXCEPTION_CONTEXT;
  oop loader = NULL;
  oop domain = NULL;
  if (accessing_klass != NULL) {
    loader = accessing_klass->loader();
    domain = accessing_klass->protection_domain();
  }
  for (int i=0; i<_unloaded_klasses->length(); i++) {
    ciKlass* entry = _unloaded_klasses->at(i);
    if (entry->name()->equals(name) &&
        entry->loader() == loader &&
        entry->protection_domain() == domain) {
      // We've found a match.
      return entry;
    }
  }

  // This is a new unloaded klass.  Create it and stick it in
  // the cache.
  ciKlass* new_klass = NULL;

  // Two cases: this is an unloaded objArrayKlass or an
  // unloaded instanceKlass.  Deal with both.
  if (name->byte_at(0) == '[') {
    // Decompose the name.'
    jint dimension = 0;
    symbolOop element_name = NULL;
    BasicType element_type= FieldType::get_array_info(name->get_symbolOop(),
                                                      &dimension,
                                                      &element_name,
                                                      KILL_COMPILE_ON_ANY);
    assert(element_type != T_ARRAY, "unsuccessful decomposition");
    ciKlass* element_klass = NULL;
    if (element_type == T_OBJECT) {
      ciSymbol* ci_name = CURRENT_ENV->get_object(element_name)->as_symbol();
      element_klass =
        CURRENT_ENV->get_klass_by_name(accessing_klass, ci_name)->as_instance_klass();
    } else {
      assert(dimension > 1, "one dimensional type arrays are always loaded.");

      // The type array itself takes care of one of the dimensions.
      dimension--;

      // The element klass is a typeArrayKlass.
      element_klass = ciTypeArrayKlass::make(element_type);
    }
    new_klass = new (arena()) ciObjArrayKlass(name, element_klass, dimension);
  } else {
    jobject loader_handle = NULL;
    jobject domain_handle = NULL;
    if (accessing_klass != NULL) {
      loader_handle = accessing_klass->loader_handle();
      domain_handle = accessing_klass->protection_domain_handle();
    }
    new_klass = new (arena()) ciInstanceKlass(name, loader_handle, domain_handle);
  }
  _unloaded_klasses->append(new_klass);

  return new_klass;
}

// ------------------------------------------------------------------
// ciObjectFactory::find
//
// Use binary search to find the position of this oop in the cache.
// If there is no entry in the cache corresponding to this oop, return
// the position at which the oop should be inserted.
int ciObjectFactory::find(oop key) {
  int min = 0;
  int max = _ci_objects->length()-1;

  // print_contents();

  while (max >= min) {
    int mid = (max + min) / 2;
    oop value = JNIHandles::resolve(_ci_objects->at(mid)->handle());
    if (value < key) {
      min = mid + 1;
    } else if (value > key) {
      max = mid - 1;
    } else {
      return mid;
    }
  }
  return min;
}

// ------------------------------------------------------------------
// ciObjectFactory::insert
//
// Insert a ciObject into the table at some index.
void ciObjectFactory::insert(int index, ciObject* obj) {
  int len = _ci_objects->length();
  if (len == index) {
    _ci_objects->append(obj);
  } else {
    _ci_objects->append(_ci_objects->at(len-1));
    int pos;
    for (pos = len-2; pos >= index; pos--) {
      _ci_objects->at_put(pos+1,_ci_objects->at(pos));
    }
    _ci_objects->at_put(index, obj);
  }
#ifdef ASSERT
  oop last = NULL;
  for (int j = 0; j< _ci_objects->length(); j++) {
    oop o = JNIHandles::resolve(_ci_objects->at(j)->handle());
    assert(last < o, "out of order");
    last = o;
  }
#endif ASSERT
}

// ------------------------------------------------------------------
// ciObjectFactory::print_contents_impl
void ciObjectFactory::print_contents_impl() {
  int len = _ci_objects->length();
  tty->print_cr("ciObjectFactory (%d) oop contents:", len);
  for (int i=0; i<len; i++) {
    _ci_objects->at(i)->print();
    tty->cr();
  }
}

// ------------------------------------------------------------------
// ciObjectFactory::print_contents
void ciObjectFactory::print_contents() {
  print();
  tty->cr();
  GUARDED_VM_ENTRY(print_contents_impl();)
}

// ------------------------------------------------------------------
// ciObjectFactory::print
//
// Print debugging information about the object factory
void ciObjectFactory::print() {
  tty->print("<ciObjectFactory oops=%d unloaded_methods=%d unloaded_klasses=%d>",
             _ci_objects->length(), _unloaded_methods->length(),
             _unloaded_klasses->length());
}

