// @(#)ciOopMap.cpp	1.5 00/01/25 16:43:13
// 
// Copyright 1993-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_ciOopMap.cpp.incl"


// Implementation of ciOopMap

ciOopMap::ciOopMap(Arena* arena, int max_nof_locals, int nof_gc_points)
: _arena(arena)
, _max_nof_locals(max_nof_locals)
, _nof_gc_points(nof_gc_points)
{
  _nof_initialize = 0;
  _locals         = NULL;
  _cached_oops    = NULL;
  _nof_entries    = -1; // will be set later

  _bci = NULL;
  _map = NULL;
  _mask_size = max_nof_locals * bits_per_entry;
  _mask_words = (_mask_size + BitsPerWord - 1) / BitsPerWord;

  // One big block for all masks
  if (nof_gc_points > 0 ) {
    if (_mask_words > 0 ) {
      _map = (unsigned int*)arena->Amalloc(sizeof(unsigned int) * _mask_words * nof_gc_points);
    }
    _bci = (unsigned int*)arena->Amalloc(sizeof(unsigned int) * _nof_gc_points);
  }
}


unsigned int ciOopMap::bci_for_index(int map_index) const {
  assert(0 <= map_index && map_index < _nof_gc_points, "");
  assert(_bci != NULL, "");
  return _bci[map_index];
}


unsigned int* ciOopMap::map_for_index(int map_index) const {
  assert(0 <= map_index && map_index < _nof_gc_points, "");
  assert(_map != NULL, "");
  int start_index = map_index * _mask_words;
  return &_map[start_index];
}


void ciOopMap::set_bci_for_index(int map_index, unsigned int bci) {
  assert(0 <= map_index && map_index < _nof_gc_points, "");
  _bci[map_index] = bci;
}


void ciOopMap::print() {
  if (nof_locals_to_initialize() > 0) {
    tty->print("Locals to initialize: [");
    for (int i = 0; i < nof_locals_to_initialize(); i++) {
      tty->print(" %d", local_to_initialize(i));
    }
    tty->print_cr(" ]");
  }
  tty->print_cr("Oop Map:");
  for (int i = 0; i < _nof_gc_points; i++) {
    tty->print("  @%d: [", bci_for_index(i));
    for (int l = 0; l < _max_nof_locals; l++) {
#ifdef ENABLE_ZAP_DEAD_LOCALS
      if (is_dead(i,l))     { tty->print("D"); }
      else 
#endif
      if (is_oop(i,l)) { tty->print("O"); }
      else tty->print("x");
    }
    tty->print_cr("]");
  }
}


unsigned int ciOopMap::entry_at(int map_index, int offset) const {
  assert (0 <= map_index && map_index < _nof_gc_points, "wrong offset");
  unsigned int* bit_mask = map_for_index(map_index);
  int i = offset * bits_per_entry;
  return bit_mask[i / BitsPerWord] >> (i % BitsPerWord);
}


void ciOopMap::set_init_local(int index, int local_index) {
  assert (0<=index && index<_nof_initialize, "range check");
  _locals[index] = local_index;
}


void ciOopMap::set_nof_initialize(int n) {
  assert(_locals == NULL, "already allocated _locals: memory leak");
  if (n == 0) return;
  _nof_initialize = n;
  _locals = (unsigned int*)_arena->Amalloc(sizeof(unsigned int) * n);
}


void ciOopMap::init_oops_cache(int code_size) {
  _cached_oops = new GrowableArray<GrowableArray<int>*>(code_size, code_size, NULL);
  int index = nof_entries() - 1;
  for (; index >= 0; index--) {
    int bci = bci_for_index(index);
    GrowableArray<int>* bci_map = new GrowableArray<int>();
    _cached_oops->at_put(bci, bci_map);
    int offset = max_nof_locals() - 1;
    for (; offset >= 0; offset--) {
      if (is_oop(index, offset)) bci_map->append(offset);
    }
  }
}


GrowableArray<int>* ciOopMap::get_oops_for_bci(int bci) {
  assert(_cached_oops != NULL, "call init_oops_cache first");
  if (bci == SynchronizationEntryBCI) {
    // the map at bci 0 is identical to the incoming arguments oop map
    return _cached_oops->at(0);
  } else {
    return _cached_oops->at(bci);
  }
}


// Implementation of ciGenerateOopMap

GrowableArray<bool>* create_gc_point_array(Arena* arena, int size) {
  return new(arena)GrowableArray<bool>(arena, size, size, false);
}


void ciGenerateOopMap::find_jsr_return_points(methodHandle method) {
  BytecodeStream s(method);
  Bytecodes::Code code;
  while ((code = s.next()) >= 0) {
    if (code == Bytecodes::_jsr) {
      if (_is_gc_point == NULL) {
        _is_gc_point = create_gc_point_array(_arena, method->code_size());
      }
      _is_gc_point->at_put(s.next_bci(), true); // set flag at return point of jsr
    }
  }
}


ciGenerateOopMap::ciGenerateOopMap(Arena* arena, methodHandle method) : GenerateOopMap(method) {
  _arena        = arena;
  _oop_maps     = NULL;
  _fill_counter = 0;
  // we have to adjust the assert in jvmciMethod before 
  // allowing no oop maps at exception points without handlers.
  _has_exceptions_handlers = method->exception_table()->length() != 0;
  _is_synchronized = method->is_synchronized();
  _is_gc_point = NULL;
  if (_has_exceptions_handlers) {
    _is_gc_point = create_gc_point_array(_arena, method->code_size());
    int lng = method()->exception_table()->length() ;
    for (int i = 0; i < lng; i+=4) {
      int start = method->exception_table()->int_at(i + 2);
      _is_gc_point->at_put(start, true);
    }
  }
  find_jsr_return_points(method);
}


bool ciGenerateOopMap::possible_gc_point(BytecodeStream *bcs) {
  Bytecodes::Code code = bcs->code();
  // all exception handler starts must have an oop map
  if (_is_gc_point != NULL && _is_gc_point->at(bcs->bci())) {
    return true;
  }
  if (_is_synchronized && bcs->bci() == 0) {
    return true;
  }
  switch (code) {
    // exception points only: need gc map only if there is an exception handler 
    // in the same method
    case Bytecodes::_iaload         : 
    case Bytecodes::_laload         : 
    case Bytecodes::_faload         : 
    case Bytecodes::_daload         : 
    case Bytecodes::_aaload         : 
    case Bytecodes::_baload         : 
    case Bytecodes::_caload         : 
    case Bytecodes::_saload         : 
    case Bytecodes::_iastore        : 
    case Bytecodes::_lastore        : 
    case Bytecodes::_fastore        : 
    case Bytecodes::_dastore        : 
    case Bytecodes::_bastore        : 
    case Bytecodes::_castore        : 
    case Bytecodes::_sastore        : 

    case Bytecodes::_idiv           : 
    case Bytecodes::_ldiv           : 
    case Bytecodes::_irem           : 
    case Bytecodes::_lrem           : 

    case Bytecodes::_arraylength    : return _has_exceptions_handlers || _is_synchronized;

    case Bytecodes::_athrow         : return true;
    // possible deoptimization points
    case Bytecodes::_getstatic      :
    case Bytecodes::_putstatic      :
    case Bytecodes::_getfield       :
    case Bytecodes::_putfield       :
    case Bytecodes::_invokevirtual  : 
    case Bytecodes::_invokespecial  : 
    case Bytecodes::_invokestatic   : 
    case Bytecodes::_invokeinterface: return true;

    case Bytecodes::_xxxunusedxxx   : ShouldNotReachHere(); return false;

    // allocation
    case Bytecodes::_new            : 
    case Bytecodes::_newarray       : 
    case Bytecodes::_anewarray      : 
    case Bytecodes::_multianewarray : return true;

    // bytecodes with slow case calling into runtime
    case Bytecodes::_aastore        : 
    case Bytecodes::_checkcast      : 
    case Bytecodes::_instanceof     : 
    case Bytecodes::_monitorenter   : return true;

    // for compiler safepoints
    case Bytecodes::_jsr_w          :
    case Bytecodes::_jsr            :
    case Bytecodes::_ret            : return UseCompilerSafepoints;

    case Bytecodes::_tableswitch    :
    case Bytecodes::_lookupswitch   :
    case Bytecodes::_goto_w         :
    case Bytecodes::_goto           : 
    case Bytecodes::_ifeq           :
    case Bytecodes::_ifne           :
    case Bytecodes::_iflt           :
    case Bytecodes::_ifge           :
    case Bytecodes::_ifgt           :
    case Bytecodes::_ifle           :
    case Bytecodes::_ifnull         :
    case Bytecodes::_ifnonnull      :
    case Bytecodes::_if_icmpeq      :
    case Bytecodes::_if_icmpne      :
    case Bytecodes::_if_icmplt      :
    case Bytecodes::_if_icmpge      :
    case Bytecodes::_if_icmpgt      :
    case Bytecodes::_if_icmple      :
    case Bytecodes::_if_acmpeq      :
    case Bytecodes::_if_acmpne      : return UseCompilerSafepoints; // actually only backward branches need it

    default                         : return false;
  }
}


void ciGenerateOopMap::fill_stackmap_prolog(int nof_gc_points) {
  _nof_gc_points = nof_gc_points;
}


void ciGenerateOopMap::fill_stackmap_for_opcodes(BytecodeStream *bcs, CellTypeState* vars, CellTypeState* stack, int stackTop) {
  if (_oop_maps == NULL) {
    _oop_maps     = new ciOopMap(_arena, method()->max_locals(), _nof_gc_points);
    _fill_counter = 0;
  }
  if (possible_gc_point(bcs) && method()->max_locals() > 0) {
    assert (_fill_counter < _nof_gc_points, "too few gc points reported");
    _oop_maps->set_bci_for_index(_fill_counter, bcs->bci());

    unsigned int* bit_mask = _oop_maps->map_for_index(_fill_counter);

    // we are interested only in the vars
    int n_entries       = method()->max_locals();
    CellTypeState* cell = vars;
    unsigned int  mask  = 1;
    unsigned int  value = 0;
    int word_index      = 0;

    for (int i = 0; i < n_entries; i++, cell++,  mask <<= ciOopMap::bits_per_entry) {
      if (mask == 0) {

        bit_mask[word_index++] = value;
        value = 0;
        mask = 1;
      }
      if (cell->is_reference()) {
        value |= (mask << ciOopMap::oop_bit_number );
      }
    #ifdef ENABLE_ZAP_DEAD_LOCALS
      // set dead bit
      if (!cell->is_live()) {
        value |= (mask << ciOopMap::dead_bit_number);
        assert(!cell->is_reference(), "dead value marked as oop");
      }
    #endif
    }
    bit_mask[word_index] = value;

    _fill_counter++;
  }
}


void ciGenerateOopMap::fill_init_vars(GrowableArray<int> *init_vars) {
  _oop_maps->set_nof_initialize(init_vars->length());
  for (int i = 0; i < init_vars->length(); i++) {
    _oop_maps->set_init_local(i, init_vars->at(i));
  }
}


ciAddressMap::ciAddressMap(Arena* arena, int nof_locals) 
: _arena(arena)
, _nof_locals(nof_locals)
{
  _map = NULL;
  _mask_size = nof_locals * bits_per_entry;
  _mask_words = (_mask_size + BitsPerWord - 1) / BitsPerWord;
  _map = (unsigned int*)arena->Amalloc(sizeof(unsigned int) * _mask_words);
}


ciAddressMap::~ciAddressMap() {
  ShouldNotReachHere(); // with new arena allocation
}


// returns a value that describes the entry
unsigned int ciAddressMap::entry_at(int offset)  const { 
  unsigned int* bit_mask = _map;
  int i = offset * bits_per_entry; 
  return bit_mask[i / BitsPerWord] >> (i % BitsPerWord); 
}




//-------------------------------------------------------------------


ciAddressMapGenerator::ciAddressMapGenerator(Arena* arena, methodHandle m, int for_bci)
: GenerateOopMap(m)
, _arena(arena)
, _for_bci(for_bci)
, _address_map(NULL)
{}



bool ciAddressMapGenerator::possible_gc_point(BytecodeStream *bcs) {
  return bcs->bci() == _for_bci;
}


void ciAddressMapGenerator::fill_stackmap_prolog(int nof_gc_points) {
}


void ciAddressMapGenerator::fill_stackmap_for_opcodes (BytecodeStream *bcs,
                                           CellTypeState* vars, 
                                           CellTypeState* stack, 
                                           int stackTop) {
  if (_address_map == NULL) {
    _address_map = new ciAddressMap(_arena, method()->max_locals());
  }
  if (possible_gc_point(bcs) && method()->max_locals() > 0) {

    unsigned int* bit_mask = _address_map->get_map_raw();

    // we are interested only in the vars
    int n_entries       = method()->max_locals();
    CellTypeState* cell = vars;
    unsigned int  mask  = 1;
    unsigned int  value = 0;
    int word_index      = 0;

    for (int i = 0; i < n_entries; i++, cell++,  mask <<= ciAddressMap::bits_per_entry) {
      if (mask == 0) {
        bit_mask[word_index++] = value;
        value = 0;
        mask = 1;
      }
      if (cell->is_address()) {
        value |= (mask << ciAddressMap::address_bit_number );
      }
    }
    bit_mask[word_index] = value;
  }
}
