/*
 * @(#)codeBlob.cpp	1.70 01/04/21 22:47:00
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_codeBlob.cpp.incl"

unsigned int align_code_offset(int offset) {
  // Align the size to CodeEntryAlignment
  return ((offset + CodeHeap::header_size() + (CodeEntryAlignment-1)) & ~(CodeEntryAlignment-1))
       - CodeHeap::header_size();
}

unsigned int CodeBlob::allocation_size(CodeBuffer *cb, int header_size, int oop_size) {
  unsigned int size = header_size + round_to(cb->reloc_size(), oopSize);
  // Align the size to CodeEntryAlignment
  size = align_code_offset(size);
  size += round_to(cb->code_size(),  oopSize);
  size += oop_size;
  return size;
}

// Creates a CodeBlob from a CodeBuffer. Sets up the size of the different regions, and copy
// code and relocation info.
CodeBlob::CodeBlob(CodeBuffer *cb, int header_size, int size, int frame_size, int rao, OopMapSet* oop_maps, int oop_size) {
  assert( size == round_to(size, oopSize), "unaligned size");
  assert( header_size == round_to(header_size, oopSize), "unaligned size");

  _size                = size;                            
  _header_size         = header_size;  
  _relocation_size     = round_to(cb->reloc_size(), oopSize);
  _instructions_offset = align_code_offset(header_size + _relocation_size);
  _data_offset         = _instructions_offset + round_to(cb->code_size(), oopSize);  
  _oops_offset         = _size - round_to(oop_size, oopSize);
  _oops_length         = 0;
  assert(_oops_offset >= _data_offset, "codeBlob is too small");
  assert(_data_offset <= size, "codeBlob is too small");
				
  cb->copy_relocation(this);	// Copy relocation info
  cb->copy_code(this);		// Copy code
  set_oop_maps(oop_maps);
  _frame_size = frame_size;
  COMPILER2_ONLY(_link_offset = not_yet_computed;) // Set to Uncomputed value
  _return_address_offset = rao;
#ifdef COMPILER1
  assert(_return_address_offset == -1, "do not use return_address_offset");
#ifdef SPARC
  assert(_frame_size >= -1, "must use frame size or -1 for runtime stubs");
#else 
  assert(_frame_size == -1, "do not use _frame_size");
#endif
#endif // compiler 1
  COMPILER2_ONLY(assert(_return_address_offset < _frame_size, "offset must be in frame");)
}

void CodeBlob::set_oop_maps( OopMapSet* p) {
  // Danger Will Robinson! This method allocates a big
  // chunk of memory, its your job to free it.
  if(p != NULL) {
    // We need to allocate a chunk big enough to hold the OopMapSet and all of its OopMaps
    _oop_maps = (OopMapSet *)NEW_C_HEAP_ARRAY(unsigned char, p->heap_size());
    p->copy_to((address)_oop_maps);
  } else {
    _oop_maps = NULL;
  }
}

void CodeBlob::flush() {
  if (_oop_maps) {
    FREE_C_HEAP_ARRAY(unsigned char, _oop_maps);
    _oop_maps = NULL;
  }
}

// relocation indexes are biased by 1 (because 0 is reserved)
oop* CodeBlob::oop_addr_at(int index) const {  
  assert(index > 0 && index <= _oops_length, "must be a valid non-zero index");
  return &oops_begin()[index-1];  
}

void CodeBlob::copy_oops(jobject* array, int length) {
  for (int index = 0 ; index < length; index++) {
    oops_begin()[index] = JNIHandles::resolve(array[index]);
  }
  _oops_length = length;
}

relocInfo::relocType CodeBlob::reloc_type_for_address(address pc) {
  RelocIterator iter(this, pc, pc+1);
  while (iter.next()) {
    return (relocInfo::relocType) iter.type();  
  }
  // Not relocation info found for pc 
  ShouldNotReachHere();
  return relocInfo::none; // dummy return value
}


void CodeBlob::fix_relocation_at_move(int delta) {
  PatchingRelocIterator iter(this);
  while (iter.next()) {
    iter.reloc()->fix_relocation_at_move(delta);
  }
}


void CodeBlob::fix_oop_relocations(address begin, address end) {
  // re-patch all oop-bearing instructions, just in case some oops moved
  RelocIterator iter(this, begin, end);
  while (iter.next()) {
    iter.reloc()->fix_oop_relocation();
  }
}

void CodeBlob::fix_oop_relocations() {
  fix_oop_relocations(NULL, NULL);
}


void CodeBlob::follow_roots_or_mark_for_unloading(bool unloading_occurred, bool& marked_for_unloading) {
  ShouldNotReachHere();
}

void CodeBlob::adjust_pointers() {
  ShouldNotReachHere();
}


OopMap* CodeBlob::oop_map_for_return_address(address return_address, bool at_call) {
  address pc = return_address ;
  assert (oop_maps() != NULL, "nope");
  if (is_native_method()) {
    // do it unprecise; for natives we are using the value stored in
    // TLS and not the return address when we cook the last frame
    return oop_maps()->find_map_at_offset ((int) pc - (int) instructions_begin(), at_call, false);
  } else {
    return oop_maps()->find_map_at_offset ((int) pc - (int) instructions_begin(), at_call, true);
  }
}


// ---------------------------------------------------------------------------
// Implementation of RuntimeStub

RuntimeStub::RuntimeStub(CodeBuffer* cb, int size, int return_address_offset, int frame_size, OopMapSet* oop_maps, bool caller_must_gc_arguments) 
: CodeBlob(cb, sizeof(RuntimeStub), size, frame_size, return_address_offset, oop_maps) {  
  _name = NULL;
  _caller_must_gc_arguments = caller_must_gc_arguments;  
}


RuntimeStub* RuntimeStub::new_runtime_stub(CodeBuffer* cb, int return_address_offset, int frame_size, OopMapSet* oop_maps, bool caller_must_gc_arguments) {
  MutexLocker mu(CodeCache_lock);
  unsigned int size = allocation_size(cb, sizeof(RuntimeStub));
  return new (size) RuntimeStub(cb, size, return_address_offset, frame_size, oop_maps, caller_must_gc_arguments);
}


void* RuntimeStub::operator new(size_t s, unsigned size) {
  void* p = CodeCache::allocate(size);
  if (!p) fatal("Initial size of CodeCache is too small");
  return p;
}


//-------------------------------------------------------------------------------------------
// DeoptimizationBlob

DeoptimizationBlob::DeoptimizationBlob(CodeBuffer *cb,
                                       int size,
                                       OopMapSet  *oop_maps, 
                                       int unpack_offset,
                                       int unpack_with_exception_offset,
                                       int return_address_offset,
                                       int frame_size)
: SingletonBlob(cb, sizeof(DeoptimizationBlob), size, frame_size, return_address_offset, oop_maps) {
  _unpack_offset           = unpack_offset;  
  _unpack_with_exception   = unpack_with_exception_offset;
}


DeoptimizationBlob* DeoptimizationBlob::create(CodeBuffer *cb,
                          OopMapSet  *oop_maps, 
                          int unpack_offset,
                          int unpack_with_exception_offset,
                          int return_address_offset,
                          int frame_size) {
  MutexLocker mu(CodeCache_lock);
  unsigned int size = allocation_size(cb, sizeof(DeoptimizationBlob));
  return new (size) DeoptimizationBlob(cb, size, oop_maps, unpack_offset, unpack_with_exception_offset, return_address_offset, frame_size);
}


void* DeoptimizationBlob::operator new(size_t s, unsigned size) {
  void* p = CodeCache::allocate(size);
  if (!p) fatal("Initial size of CodeCache is too small");
  return p;
}

// Follow outgoing arguments for the deoptimized frame. This is similar to handle a compiled frame, except
// we find the signature differently
void DeoptimizationBlob::preserve_callee_argument_oops(frame fr, const RegisterMap *reg_map, void f(oop*))  {
#ifndef COMPILER2
  ShouldNotReachHere();
#else
  assert(reg_map != NULL, "must be set");
  assert(reg_map->thread() != NULL, "must be set");  
  javaVFrame* vf = (javaVFrame*)vframe::new_vframe(&fr, reg_map, reg_map->thread());
  assert(vf != NULL, "vframe must exist");
  assert(vf->is_java_frame(), "just checkking");
  if (vf->bci() != 0) {
    Bytecode_invoke* call = Bytecode_invoke_at(vf->method(), vf->bci());
    bool is_static = call->is_invokestatic();
    symbolOop signature = call->signature();
    fr.oops_compiled_arguments_do(signature, is_static, reg_map, f);
  }
#endif // COMPILER2
}

//-------------------------------------------------------------------------------------------
// UncommonTrapBlob

UncommonTrapBlob::UncommonTrapBlob(CodeBuffer* cb,
                                   int         size,
                                   OopMapSet*  oop_maps,
                                   int         return_address_offset,
                                   int         frame_size)
: SingletonBlob(cb, sizeof(UncommonTrapBlob), size, frame_size, return_address_offset, oop_maps) {
}

UncommonTrapBlob* UncommonTrapBlob::create(CodeBuffer *cb,
                                           OopMapSet  *oop_maps,
                                           int        return_address_offset,
                                           int        frame_size) {
  MutexLocker mu(CodeCache_lock);
  unsigned int size = allocation_size(cb, sizeof(UncommonTrapBlob)); 
  return new (size) UncommonTrapBlob(cb, size, oop_maps, return_address_offset, frame_size);
}

void* UncommonTrapBlob::operator new(size_t s, unsigned size) {
  void* p = CodeCache::allocate(size);
  if (!p) fatal("Initial size of CodeCache is too small");
  return p;
}

//-------------------------------------------------------------------------------------------
// ExceptionBlob

ExceptionBlob::ExceptionBlob(CodeBuffer *cb, int size, OopMapSet  *oop_maps, int return_address_offset, int frame_size) 
: SingletonBlob(cb, sizeof(ExceptionBlob), size, frame_size, return_address_offset, oop_maps) {
}


ExceptionBlob* ExceptionBlob::create(CodeBuffer *cb, OopMapSet  *oop_maps, int return_address_offset, int frame_size) {
  MutexLocker mu(CodeCache_lock);
  unsigned int size = allocation_size(cb, sizeof(ExceptionBlob)); 
  return new (size) ExceptionBlob(cb, size, oop_maps, return_address_offset, frame_size);
}

void* ExceptionBlob::operator new(size_t s, unsigned size) {
  void* p = CodeCache::allocate(size);
  if (!p) fatal("Initial size of CodeCache is too small");
  return p;
}

//-------------------------------------------------------------------------------------------
// ExceptionBlob

SafepointBlob ::SafepointBlob(CodeBuffer *cb, int size, OopMapSet  *oop_maps, int return_address_offset, int frame_size) 
: SingletonBlob(cb, sizeof(ExceptionBlob), size, frame_size, return_address_offset, oop_maps) {
}


SafepointBlob * SafepointBlob::create(CodeBuffer *cb, OopMapSet  *oop_maps, int return_address_offset, int frame_size) {
  MutexLocker mu(CodeCache_lock);
  unsigned int size = allocation_size(cb, sizeof(ExceptionBlob)); 
  return new (size) SafepointBlob(cb, size, oop_maps, return_address_offset, frame_size);
}

void* SafepointBlob::operator new(size_t s, unsigned size) {
  void* p = CodeCache::allocate(size);
  if (!p) fatal("Initial size of CodeCache is too small");
  return p;
}

bool SafepointBlob::caller_must_gc_arguments(JavaThread *thread) const {
  return thread->safepoint_state()->caller_must_gc_arguments();
}

//-------------------------------------------------------------------------------------------
// Non-product code
#ifndef PRODUCT

void CodeBlob::verify() {
  ShouldNotReachHere(); 
}

void CodeBlob::print() const {
  tty->print_cr("[CodeBlob]");
  tty->print_cr("Framesize: %d", _frame_size);
  tty->print_cr("Return address offset: %d", _return_address_offset);
}


void CodeBlob::print_value_on(outputStream* st) const {
  st->print_cr("[CodeBlob]");
}


void RuntimeStub::verify() {  
}


void RuntimeStub::print() const {
  CodeBlob::print();
  tty->print("Runtime Stub: ");
  tty->print_cr(name());
  Disassembler::decode((CodeBlob*)this);
}


void RuntimeStub::print_value_on(outputStream* st) const {
  st->print("RuntimeStub: "); st->print(name());
}


void SingletonBlob::verify() {  
}

void SingletonBlob::print() const {
  CodeBlob::print();  
  tty->print_cr(name());
  Disassembler::decode((CodeBlob*)this);
}


void SingletonBlob::print_value_on(outputStream* st) const {
  st->print_cr(name());
}

void DeoptimizationBlob::print_value_on(outputStream* st) const {
  st->print_cr("Deoptimization (frame not available)");
}

void DeoptimizationBlob::print_value_on_fancy(outputStream* st, JavaThread *thread, frame fr) const {
  int pc_off = fr.pc() - instructions_begin();
  if( pc_off == _unpack_offset ) 
    st->print_cr("Deoptimization pending for");
  else if( pc_off == _unpack_with_exception ) 
    st->print_cr("Deoptimization with exception pending for");
  else if( fr.pc() == 0 ) 
    st->print_cr("Deoptimization (frame not available)");
  else
    st->print_cr("Deoptimization in progress for");
#ifndef CORE
  if( thread ) {
    vframeArray *array = thread->vframe_array_for( fr.sp() );
    array->print_on_2(st);
  }
#endif
}

#endif

