/*
 * @(#)codeCache.cpp	1.82 01/02/09 14:44:37
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_codeCache.cpp.incl"

// Helper class for printing in CodeCache

class CodeBlob_sizes {
 private:
  int count;
  int total_size;
  int header_size;
  int constants_size;
  int code_size;
  int exception_size;
  int stub_size;
  int relocation_size;
  int scopes_oop_size;
  int scopes_data_size;
  int scopes_pcs_size;

 public:
  CodeBlob_sizes() {
    count            = 0;
    total_size       = 0;
    header_size      = 0;
    constants_size   = 0;
    code_size        = 0;
    exception_size   = 0;
    stub_size        = 0;
    relocation_size  = 0;
    scopes_oop_size  = 0;
    scopes_data_size = 0;
    scopes_pcs_size  = 0;
  }

  int total()                                    { return total_size; }
  bool is_empty()                                { return count == 0; }

  void print(const char* title, int t) {
    tty->print_cr(" #%d %s = %dK (hdr %d%%,  loc %d%%, const %d%%, code %d%%, except %d%%, stub %d%%, [oops %d%%, data %d%%, pcs %d%%])",
                  count,
                  title,
                  total() / K,
                  header_size             * 100 / total_size,
                  relocation_size         * 100 / total_size,
                  constants_size          * 100 / total_size,
                  code_size               * 100 / total_size,
                  exception_size          * 100 / total_size,
                  stub_size               * 100 / total_size,                  
                  scopes_oop_size         * 100 / total_size,
                  scopes_data_size        * 100 / total_size,
                  scopes_pcs_size         * 100 / total_size);
  }

  void add(CodeBlob* cb) {
    count++;
    total_size       += cb->size();
    header_size      += cb->header_size();
    relocation_size  += cb->relocation_size();
    scopes_oop_size  += cb->oops_size();
    if (cb->is_nmethod()) { 
      nmethod *nm = (nmethod*)cb;
      constants_size   += nm->constants_size();
      code_size        += nm->code_size();
      exception_size   += nm->exception_size();
      stub_size        += nm->stub_size();
    
      scopes_data_size += nm->scopes_data_size();
      scopes_pcs_size  += nm->scopes_pcs_size();
    } else {
      code_size        += cb->instructions_size();
    }
  }
};


// CodeCache implementation

CodeHeap CodeCache::_heap;
int CodeCache::_number_of_blobs = 0;
int CodeCache::_number_of_nmethods_with_dependencies = 0;


CodeBlob* CodeCache::first() {
  assert_lock(CodeCache_lock);
  return (CodeBlob*)_heap.first();
}


CodeBlob* CodeCache::next(CodeBlob* cb) {
  assert_lock(CodeCache_lock);
  return (CodeBlob*)_heap.next(cb);
}

CodeBlob* CodeCache::alive(CodeBlob *cb) {
  assert_lock(CodeCache_lock);
  while(cb != NULL && !cb->is_alive()) {
    cb = next(cb);
  }
  return cb;
}

CodeBlob* CodeCache::allocate(int size) {
  // Do not seize the CodeCache lock here--if the caller has not
  // already done so, we are going to lose bigtime, since the code
  // cache will contain a garbage CodeBlob until the caller can
  // run the constructor for the CodeBlob subclass he is busy
  // instantiating.
  assert_lock(CodeCache_lock);
  CodeBlob* cb = NULL;
  _number_of_blobs++;
  while (true) {
    cb = (CodeBlob*)_heap.allocate(size);
    if (cb != NULL) break;
    if (!_heap.expand_by(CodeCacheExpansionSize)) {
      // Expansion failed
      return NULL;
    }
    if (PrintCodeCacheExtension) {
      ResourceMark rm;
      tty->print_cr("code cache extended to [0x%x, 0x%x[ (%d bytes)", _heap.begin(), _heap.end(), (address)_heap.end() - (address)_heap.begin());
    }
  }
  verify_if_often();
  return cb;
}

void CodeCache::free(CodeBlob* cb) {
  assert_lock(CodeCache_lock);
  verify_if_often();

  if (cb->is_nmethod() && ((nmethod *)cb)->number_of_dependents() > 0) {
    _number_of_nmethods_with_dependencies--;
  }  
  _number_of_blobs--;  

  _heap.deallocate(cb);
  verify_if_often();  
  assert(_number_of_blobs >= 0, "sanity check");
}


void CodeCache::commit(CodeBlob* cb) {
  // this is called by nmethod::nmethod, which must already own CodeCache_lock
  assert_lock(CodeCache_lock);
  if (cb->is_nmethod() && ((nmethod *)cb)->number_of_dependents() > 0) {
    _number_of_nmethods_with_dependencies++;
  }  
  // flush the hardware I-cache
  ICache::invalidate_range(cb->instructions_begin(), cb->instructions_size());
}


void CodeCache::flush() {
  assert_lock(CodeCache_lock);
  Unimplemented();
}


// Iteration over CodeBlobs

#define FOR_ALL_BLOBS(var)       for (CodeBlob *var =       first() ; var != NULL; var =       next(var) )
#define FOR_ALL_ALIVE_BLOBS(var) for (CodeBlob *var = alive(first()); var != NULL; var = alive(next(var)))


bool CodeCache::contains(void *p) {
  // It should be ok to call contains without holding a lock
  return _heap.contains(p);
}


// This method is safe to call without holding the CodeCache_lock, as long as a dead codeblob is not
// looked up (i.e., one that has been marked for deletion). It only dependes on the _segmap to contain
// valid indices, which it will always do, as long as the CodeBlob is not in the process of being recycled.
CodeBlob* CodeCache::find_blob(void* start) {  
  CodeBlob* result = find_blob_unsafe(start);
  if (result == NULL) return NULL;
  // We could potientially look up non_entrant methods
  guarantee(!result->is_zombie() || result->is_locked_by_vm() || is_error_reported(), "unsafe access to zombie method");    
  return result;  
}

CodeBlob* CodeCache::find_blob_unsafe(void* start) {  
  CodeBlob* result = (CodeBlob*)_heap.find_start(start);
  if (result == NULL) return NULL;
  assert(result->blob_contains((address)start), "found wrong CodeBlob");     
  return result;  
}

nmethod* CodeCache::find_nmethod(void* start) {  
  CodeBlob *cb = find_blob(start);
  assert(cb == NULL || cb->is_nmethod(), "did not find an nmethod");  
  return (nmethod*)cb;
}


void CodeCache::blobs_do(void f(CodeBlob* nm)) {
  assert_lock(CodeCache_lock);
  FOR_ALL_BLOBS(p) { 
    f(p); 
  }
}


void CodeCache::nmethods_do(void f(nmethod* nm)) {
  assert_lock(CodeCache_lock);
  FOR_ALL_BLOBS(nm) { 
    if (nm->is_nmethod()) f((nmethod*)nm); 
  }
}


int CodeCache::alignment_unit() {
  return _heap.alignment_unit();
}


int CodeCache::alignment_offset() {
  return _heap.alignment_offset();
}


// This is called at the end of MarkSweep::phase1. Follow all roots in 
// in the compiled code, unless they are the only ones keeping the a
// class alive. In that case, we NULL out the roots, mark the CodeBlob
// for unloading and set the boolean flag marked_for_unloading to true.

void CodeCache::follow_roots_phase2(bool unloading_occurred, bool& marked_for_unloading) {
  assert_lock(CodeCache_lock);
  marked_for_unloading = false;
  FOR_ALL_ALIVE_BLOBS(cb) {
    cb->follow_roots_or_mark_for_unloading(unloading_occurred, marked_for_unloading);
  }
}

void CodeCache::adjust_pointers() {
  assert_lock(CodeCache_lock);
  FOR_ALL_ALIVE_BLOBS(cb) {
    cb->adjust_pointers();
  }
}

void CodeCache::gc_prologue() {
}


void CodeCache::gc_epilogue() {
  assert_lock(CodeCache_lock);
  FOR_ALL_ALIVE_BLOBS(cb) {
    cb->fix_oop_relocations();
  }
}


address CodeCache::first_address() {
  assert_lock(CodeCache_lock);
  return (address)_heap.begin();
}


address CodeCache::last_address() {
  assert_lock(CodeCache_lock);
  return (address)_heap.end();
}


void CodeCache::initialize() {
  assert(CodeCacheSegmentSize >= CodeEntryAlignment, "CodeCacheSegmentSize must be large enough to align entry points");
#ifdef COMPILER2
  assert(CodeCacheSegmentSize >= OptoLoopAlignment,  "CodeCacheSegmentSize must be large enough to align inner loops");
#endif
  assert(CodeCacheSegmentSize >= sizeof(jdouble),    "CodeCacheSegmentSize must be large enough to align constants");
  Universe::check_alignment(InitialCodeCacheSize, os::vm_page_size(), "initial code cache");
  Universe::check_alignment(ReservedCodeCacheSize, os::vm_page_size(), "maximum code cache");
  if (!_heap.reserve(ReservedCodeCacheSize, InitialCodeCacheSize, CodeCacheSegmentSize)) {
    vm_exit_during_initialization("Could not reserve enough space for code cache");
  }

}


void codeCache_init() {
  CodeCache::initialize();
}

int CodeCache::capacity() {   
  return _heap.capacity(); 
}

int CodeCache::max_capacity() {   
  return _heap.max_capacity(); 
}

//------------------------------------------------------------------------------------------------
// Non-CORE stuff

#ifndef CORE
int CodeCache::number_of_nmethods_with_dependencies() { 
  return _number_of_nmethods_with_dependencies; 
}

void CodeCache::clear_inline_caches() {
  assert_lock(CodeCache_lock);
  FOR_ALL_ALIVE_BLOBS(nm) {
    if (nm->is_nmethod()) ((nmethod *)nm)->clear_inline_caches();
  }
}

int CodeCache::mark_for_deoptimization(klassOop dependee) {
  MutexLocker mu(CodeCache_lock);
  int number_of_marked_CodeBlobs = 0;
  FOR_ALL_ALIVE_BLOBS(nm) {
    if (nm->is_nmethod() && ((nmethod *)nm)->is_dependent_on(dependee)) {
      if (TraceDependencies) {
        ResourceMark rm;
        tty->print_cr("Marked for deoptimization");
        tty->print_cr("  dependee = %s", instanceKlass::cast(dependee)->external_name());
        nm->print();
        ((nmethod *)nm)->print_dependencies();
      }
      ((nmethod *)nm)->mark_for_deoptimization();
      number_of_marked_CodeBlobs++;
    }
  }
  return number_of_marked_CodeBlobs;
}


int CodeCache::mark_all_for_deoptimization() {
  assert_lock(CodeCache_lock);
  int number_of_marked_nmethods = 0;
  FOR_ALL_ALIVE_BLOBS(nm) {
    if (nm->is_nmethod()) {
      ((nmethod *)nm)->mark_for_deoptimization();
      number_of_marked_nmethods++;
    }
  }
  return number_of_marked_nmethods;
}


void CodeCache::make_marked_nmethods_zombies() {
  assert_lock(CodeCache_lock);
  FOR_ALL_ALIVE_BLOBS(nm) {
    if (nm->is_nmethod() && ((nmethod *)nm)->is_marked_for_deoptimization()) {
      ((nmethod *)nm)->make_zombie();
    }
  }
}

void CodeCache::make_marked_nmethods_unloaded() {
  assert_lock(CodeCache_lock);
  FOR_ALL_ALIVE_BLOBS(nm) {
    if (nm->is_nmethod() && ((nmethod *)nm)->is_marked_for_unloading()) {
      ((nmethod *)nm)->make_unloaded();
    }
  }
}


#endif

//------------------------------------------------------------------------------------------------
// Non-product version

#ifndef PRODUCT

void CodeCache::verify() {
  _heap.verify();
  FOR_ALL_ALIVE_BLOBS(p) {
    p->verify();
  }
}

void CodeCache::verify_if_often() {
  if (VerifyCodeCacheOften) {
    _heap.verify();
  }
}

void CodeCache::print_internals() {
  int nmethodCount = 0;
  int runtimeStubCount = 0;
  int i2cAdapterCount = 0;
  int c2iAdapterCount = 0;
  int deoptimizationStubCount = 0;
  int uncommonTrapStubCount = 0;
  int total = 0;
  int nmethodAlive = 0;
  int nmethodNotEntrant = 0;
  int nmethodZombie = 0;
  int nmethodUnloaded = 0;
  int nmethodYoung = 0;
  int nmethodOld = 0;
  int nmethodJava = 0;
  int nmethodNative = 0;
  int maxCodeSize = 0;
  ResourceMark rm;

  CodeBlob *cb;
  for (cb = first(); cb != NULL; cb = next(cb)) {
    total++;
    if (cb->is_nmethod()) {
      nmethod* nm = (nmethod*)cb;

      nmethodCount++;
 
      if(nm->is_alive()) { nmethodAlive++; }
      if(nm->is_not_entrant()) { nmethodNotEntrant++; }
      if(nm->is_zombie()) { nmethodZombie++; }
      if(nm->is_unloaded()) { nmethodUnloaded++; }
      if(nm->is_young()) { nmethodYoung++; }
      if(nm->is_old()) { nmethodOld++; }
      if(nm->is_native_method()) { nmethodNative++; }

      if(nm->is_java_method()) { 
        nmethodJava++;
        if(nm->code_size() > maxCodeSize) {
          maxCodeSize = nm->code_size();
        }
      }
    } else if (cb->is_runtime_stub()) {
      runtimeStubCount++;
    } else if (cb->is_i2c_adapter()) {
      i2cAdapterCount++;
    } else if (cb->is_c2i_adapter()) {
      c2iAdapterCount++;
    } else if (cb->is_deoptimization_stub()) {
      deoptimizationStubCount++;
    } else if (cb->is_uncommon_trap_stub()) {
      uncommonTrapStubCount++;
    }
  }

  int bucketSize = 512;
  int bucketLimit = maxCodeSize / bucketSize + 1;
  int *buckets = NEW_C_HEAP_ARRAY(int, bucketLimit);
  memset(buckets,0,sizeof(int) * bucketLimit);

  for (cb = first(); cb != NULL; cb = next(cb)) {
    if (cb->is_nmethod()) {
      nmethod* nm = (nmethod*)cb;
      if(nm->is_java_method()) { 
        buckets[nm->code_size() / bucketSize]++;
      }
    }
  }
  tty->print_cr("Code Cache Entries (total of %d)",total);
  tty->print_cr("-------------------------------------------------");
  tty->print_cr("nmethods: %d",nmethodCount);
  tty->print_cr("\talive: %d",nmethodAlive);
  tty->print_cr("\tnot_entrant: %d",nmethodNotEntrant);
  tty->print_cr("\tzombie: %d",nmethodZombie);
  tty->print_cr("\tunloaded: %d",nmethodUnloaded);
  tty->print_cr("\tyoung: %d",nmethodYoung);
  tty->print_cr("\told: %d",nmethodOld);
  tty->print_cr("\tjava: %d",nmethodJava);
  tty->print_cr("\tnative: %d",nmethodNative);
  tty->print_cr("runtime_stubs: %d",runtimeStubCount);
  tty->print_cr("i2c_adapters: %d",i2cAdapterCount);
  tty->print_cr("c2i_adapters: %d",c2iAdapterCount);
  tty->print_cr("deoptimization_stubs: %d",deoptimizationStubCount);
  tty->print_cr("uncommon_traps: %d",uncommonTrapStubCount);
  tty->print_cr("\nnmethod size distribution (non-zombie java)");
  tty->print_cr("-------------------------------------------------");

  for(int i=0; i<bucketLimit; i++) {
    if(buckets[i] != 0) {
      tty->print("%d - %d bytes",i*bucketSize,(i+1)*bucketSize);
      tty->fill_to(40);
      tty->print_cr("%d",buckets[i]);
    }
  }

  FREE_C_HEAP_ARRAY(int, buckets);
}

void CodeCache::print() {
  CodeBlob_sizes live;
  CodeBlob_sizes dead;

  FOR_ALL_BLOBS(p) {
    if (!p->is_alive()) {
      dead.add(p);
    } else {
      live.add(p);
    }
  }

  tty->print_cr("CodeCache:");

  if (!live.is_empty()) {
    live.print("live", _heap.capacity());
  }
  if (!dead.is_empty()) {
    dead.print("dead", _heap.capacity());
  }


#ifdef ASSERT
  if (Verbose) {
     // print the oop_map usage
    int code_size = 0;
    int number_of_blobs = 0;
    int number_of_oop_maps = 0;
    int map_size = 0;
    FOR_ALL_BLOBS(p) {
      if (p->is_alive()) {
        number_of_blobs++;
        code_size += p->instructions_size();
        OopMapSet* set = p->oop_maps();
        if (set != NULL) {
          number_of_oop_maps += set->size();      
          map_size   += set->heap_size();
        }
      }
    }
    tty->print_cr("OopMaps");
    tty->print_cr("  #blobs    = %d", number_of_blobs);
    tty->print_cr("  code size = %d", code_size);
    tty->print_cr("  #oop_maps = %d", number_of_oop_maps);
    tty->print_cr("  map size  = %d", map_size);
  }
#endif

}

#endif

