/*
 * @(#)exceptionHandlerTable.cpp	1.15 99/11/17 15:21:33
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_exceptionHandlerTable.cpp.incl"


void ExceptionHandlerTable::add_entry(HandlerTableEntry entry) {
  _nesting.check();
  if (_length >= _size) {
    // not enough space => grow the table (amortized growth, double its size)
    guarantee(_size > 0, "no space allocated => cannot grow the table since it is part of nmethod");
    int new_size = _size * 2;
    _table = REALLOC_RESOURCE_ARRAY(HandlerTableEntry, _table, _size, new_size);
    _size = new_size;
  }
  assert(_length < _size, "sanity check");
  _table[_length++] = entry;
}


HandlerTableEntry* ExceptionHandlerTable::subtable_for(int catch_pco) const {
  int i = 0;
  while (i < _length) {
    HandlerTableEntry* t = _table + i;
    if (t->pco() == catch_pco) {
      // found subtable matching the catch_pco
      return t;
    } else {
      // advance to next subtable
      i += t->len() + 1; // +1 for header
    }
  }
  return NULL;
}


ExceptionHandlerTable::ExceptionHandlerTable(int initial_size) {
  guarantee(initial_size > 0, "initial size must be > 0");
  _table  = NEW_RESOURCE_ARRAY(HandlerTableEntry, initial_size);
  _length = 0;
  _size   = initial_size;
}


ExceptionHandlerTable::ExceptionHandlerTable(const nmethod* nm) {
  _table  = (HandlerTableEntry*)nm->handler_table_begin();
  _length = nm->handler_table_size() / sizeof(HandlerTableEntry);
  _size   = 0; // no space allocated by ExeptionHandlerTable!
}


void ExceptionHandlerTable::add_subtable(
  int                 catch_pco,
  GrowableArray<int>* handler_bcis,
  GrowableArray<int>* handler_pcos
) {
  assert(subtable_for(catch_pco) == NULL, "catch handlers for this catch_pco added twice");
  assert(handler_bcis->length() == handler_pcos->length(), "bci & pc table have different length");
  if (handler_bcis->length() > 0) {
    // add subtable header
    add_entry(HandlerTableEntry(handler_bcis->length(), catch_pco));
    // add individual entries
    for (int i = 0; i < handler_bcis->length(); i++) {
      add_entry(HandlerTableEntry(handler_bcis->at(i), handler_pcos->at(i)));
      assert(entry_for(catch_pco, handler_bcis->at(i))->pco() == handler_pcos->at(i), "entry not added correctly");
    }
  }
}


void ExceptionHandlerTable::copy_to(nmethod* nm) {
  assert(size_in_bytes() == nm->handler_table_size(), "size of space allocated in nmethod incorrect");
  memmove(nm->handler_table_begin(), _table, size_in_bytes());
}


HandlerTableEntry* ExceptionHandlerTable::entry_for(int catch_pco, int handler_bci) const {
  HandlerTableEntry* t = subtable_for(catch_pco);
  if (t != NULL) {
    int l = t->len();
    while (l-- > 0) {
      t++;
      if (t->bci() == handler_bci) return t;
    }
  }
  return NULL;
}


void ExceptionHandlerTable::print_subtable(HandlerTableEntry* t) const {
  int l = t->len();
  tty->print_cr("catch_pco = %d (%d entries)", t->pco(), l);
  while (l-- > 0) {
    t++;
    tty->print_cr("  bci %d -> pco %d", t->bci(), t->pco());
  }
}


void ExceptionHandlerTable::print() const {
  tty->print_cr("ExceptionHandlerTable (size = %d bytes)", size_in_bytes());
  int i = 0;
  while (i < _length) {
    HandlerTableEntry* t = _table + i;
    print_subtable(t);
    // advance to next subtable
    i += t->len() + 1; // +1 for header
  }
}

void ExceptionHandlerTable::print_subtable_for(int catch_pco) const {
  HandlerTableEntry* subtable = subtable_for(catch_pco);

  if( subtable != NULL ) { print_subtable( subtable ); }
}

// ----------------------------------------------------------------------------
// Implicit null exception tables.  Maps an exception PC offset to a
// continuation PC offset.  Table has a first length word, then pairs of
// <excp-offset, const-offset>.
void ImplicitNullCheckTable::set_size( uint size ) {
  _size = size;
  _data = NEW_RESOURCE_ARRAY(implicit_null_entry, (size*2+1));
  _data[0] = 0;                 // NOT size!
}

void ImplicitNullCheckTable::append( uint exec_off, uint cont_off ) {
  assert( (sizeof(implicit_null_entry) >= 4) || (exec_off < 65535), "" );
  assert( (sizeof(implicit_null_entry) >= 4) || (cont_off < 65535), "" );
  uint l = len();
  *(adr(l)  ) = exec_off;
  *(adr(l)+1) = cont_off;
  _data[0] = l+1;
};

void ImplicitNullCheckTable::replicate_faulting_entry( uint old_exec_off, uint new_exec_off ) {
  _nesting.check();
  uint l = len();
  for( uint i=0; i<l; i++ ) {
    if( *adr(i) == old_exec_off ) {
      uint cont_off = *(adr(i)+1); // get it before reallocation (_data may move)
      if (_size <= len()) {
        uint old_size_in_elements = _size*2 +1;
        _size *= 2;
        uint new_size_in_elements = _size*2 +1;
        _data = REALLOC_RESOURCE_ARRAY(uint, _data, old_size_in_elements, new_size_in_elements);
      }
      append(new_exec_off, cont_off);
      break;
    }
  }
};

uint ImplicitNullCheckTable::at( uint exec_off ) const {
  uint l = len();
  for( uint i=0; i<l; i++ )
    if( *adr(i) == exec_off )
      return *(adr(i)+1);
  return 0;                     // Failed to find any execption offset
}

void ImplicitNullCheckTable::print() const {
  tty->print("{");
  for( uint i=0; i<len(); i++ )
    tty->print("<%d,%d> ",*adr(i),*(adr(i)+1));
  tty->print_cr("}");
}

ImplicitNullCheckTable::ImplicitNullCheckTable(const nmethod* nm) {
  _data  = (implicit_null_entry*)nm->nul_chk_table_begin();
  _size = len();
}

void ImplicitNullCheckTable::copy_to( nmethod* nm ) {
  assert(size_in_bytes() <= nm->nul_chk_table_size(), "size of space allocated in nmethod incorrect");
  memmove( nm->nul_chk_table_begin(), _data, size_in_bytes() );
}

void ImplicitNullCheckTable::verify(nmethod *nm) const {
  for (uint i = 0; i < len(); i++) {
     if ((*adr(i) > (unsigned int)nm->code_size()) ||
         (*(adr(i)+1) > (unsigned int)nm->code_size()))
       fatal1("Invalid offset in ImplicitNullCheckTable at %lx", _data);
  }
}
