/*
 * @(#)exceptionHandlerTable.hpp	1.12 99/07/06 16:37:16
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


// A HandlerTableEntry describes an individual entry of a subtable
// of ExceptionHandlerTable. An entry consists of a pair(bci, pco),
// where bci is the exception handler bci, and pco is the pc offset
// relative to the nmethod code start for the compiled exception
// handler corresponding to the (interpreted) exception handler
// starting at bci.
//
// The first HandlerTableEntry of each subtable holds the length
// and catch_pco for the subtable (the length is the number of
// subtable entries w/o header).

class HandlerTableEntry {
 private:
  int _bci;
  int _pco;

 public:
  HandlerTableEntry(int bci, int pco) {
    assert( 0 <= pco, "pco must be positive");
    _bci = bci;
    _pco = pco;
  }

  int len() const { return _bci; } // for entry at subtable begin
  int bci() const { return _bci; }
  int pco() const { return _pco; }
};


// An ExceptionHandlerTable is an abstraction over a list of subtables
// of exception handlers for CatchNodes. Each subtable has a one-entry
// header holding length and catch_pco of the subtable, followed
// by 'length' entries for each exception handler that can be reached
// from the corresponding CatchNode. The catch_pco is the pc offset of
// the CatchNode in the corresponding nmethod. Empty subtables are dis-
// carded.
//
// Structure of the table:
//
// table    = { subtable }.
// subtable = header entry { entry }.
// header   = a pair (number of subtable entries, catch pc offset)
// entry    = a pair (handler bci, handler pc offset)
//
// An ExceptionHandlerTable can be created from scratch, in which case
// it is possible to add subtables. It can also be created from an
// nmethod (for lookup purposes) in which case the table cannot be
// modified.

class nmethod;
class ExceptionHandlerTable VALUE_OBJ_CLASS_SPEC {
 private:
  HandlerTableEntry* _table;    // the table
  int                _length;   // the current length of the table
  int                _size;     // the number of allocated entries
  ReallocMark        _nesting;  // assertion check for reallocations

  // add the entry & grow the table if needed
  void add_entry(HandlerTableEntry entry); 
  HandlerTableEntry* subtable_for(int catch_pco) const;

 public:
  // (compile-time) construction within compiler
  ExceptionHandlerTable(int initial_size = 8);

  // (run-time) construction from nmethod
  ExceptionHandlerTable(const nmethod* nm);

  // (compile-time) add entries
  void add_subtable(
    int                 catch_pco, // the pc offset for the CatchNode
    GrowableArray<int>* handler_bcis, // the exception handler entry point bcis
    GrowableArray<int>* handler_pcos // pc offsets for the compiled handlers
  );

  // nmethod support
  int  size_in_bytes() const { return _length * sizeof(HandlerTableEntry); }
  void copy_to(nmethod* nm);

  // lookup
  HandlerTableEntry* entry_for(int catch_pco, int handler_bci) const;

  // debugging
  void print_subtable(HandlerTableEntry* t) const;
  void print() const;
  void print_subtable_for(int catch_pco) const;
};


// ----------------------------------------------------------------------------
// Implicit null exception tables.  Maps an exception PC offset to a
// continuation PC offset.  Table has a first length word, then pairs of
// <excp-offset, const-offset>.
// // Use 32-bit representation for offsets
typedef  uint              implicit_null_entry;

class ImplicitNullCheckTable VALUE_OBJ_CLASS_SPEC {
  uint _size;
  implicit_null_entry *_data;
  implicit_null_entry *adr( uint idx ) const { return &_data[2*idx+1]; }
  ReallocMark          _nesting;  // assertion check for reallocations
public:
  ImplicitNullCheckTable( ) :  _data(0) { }
  // (run-time) construction from nmethod
  ImplicitNullCheckTable( const nmethod *nm );

  void set_size( uint size );
  void append( uint exec_off, uint cont_off );
  void replicate_faulting_entry( uint old_exec_off, uint new_exec_off );
  uint at( uint exec_off ) const;

  uint len() const { return _data[0]; }
  int size_in_bytes() const { return (2*len() +1) * sizeof(implicit_null_entry); }
  
  void copy_to(nmethod* nm);
  void print() const;
  void verify(nmethod *nm) const;
};
