/*
 * @(#)onStackReplacement.cpp	1.23 01/02/09 14:44:42
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

#include "incls/_precompiled.incl"
#include "incls/_onStackReplacement.cpp.incl"


#ifdef COMPILER2
// Implementation of OSRAdapter

OSRAdapter::OSRAdapter(CodeBuffer *cb, OopMapSet *oop_maps, int size, int frame_size, int sp_offset)
  :CodeBlob(cb,                  // Code buffer
            sizeof(OSRAdapter),  // Header size            
            size,                // Size
            frame_size,          // Frame size
            frame_size-1,        // return address offset
            oop_maps) {
  _sp_offset = sp_offset;
}


OSRAdapter* OSRAdapter::new_osr_adapter(CodeBuffer* cb,	OopMapSet *oop_maps, int frame_size, int sp_offset) {
  unsigned int size = allocation_size(cb, sizeof(OSRAdapter));
  MutexLocker mu(CodeCache_lock);
  return new (size) OSRAdapter(cb, oop_maps, size, frame_size, sp_offset);
}


void* OSRAdapter::operator new(size_t s, unsigned size) {
  void* p = CodeCache::allocate(size);  
  return p;
}


#ifndef PRODUCT
void OSRAdapter::verify() {
  // Nothing to verify for now
}


void OSRAdapter::print() const {
  tty->print_cr("[OSRAdapter. Framesize: %d]", frame_size());
}


void OSRAdapter::print_value_on(outputStream* st) const {
  tty->print_cr("[OSRAdapter]");
}
#endif // PRODUCT
#endif // COMPILER2


// Implementation of OnStackReplacement

#ifdef COMPILER2
GrowableArray<OSRAdapter*>* OnStackReplacement::_osr_adapters;

void OnStackReplacement::initialize() {
  if (!UseOnStackReplacement) return;
  // Initialize vector of OSRAdapter. The common used sizes gets precomputed
  // to limit fragmentation of the code cache.
  _osr_adapters = new(ResourceObj::C_HEAP)GrowableArray<OSRAdapter*>(InitialAdapterVectorSize, true);
  // Precompute common cases
  assert(MaxTypicalAdapterSize < InitialAdapterVectorSize, "wrong setup parameters");
  for(int framesize = MaxTypicalAdapterSize; framesize >= MinTypicalAdapterSize; framesize--) {
    OSRAdapter* osr = OptoRuntime::generate_osr_blob(framesize);
    assert(osr != NULL, "must generate an adapter blob");
    _osr_adapters->at_put_grow(framesize, osr);
  }
}


// The generator method is implemented in runtime_<cpu>.cpp
OSRAdapter* OnStackReplacement::get_osr_adapter(int framesize) {
  assert(UseOnStackReplacement, "on-stack replacement not used");

  OSRAdapter* osr = NULL;  
  { MutexLocker mu(AdapterCache_lock);
    // This will potentially grow the array. 
    osr = _osr_adapters->at_grow(framesize);
  }

  // Need to create OSR adapter and update list.
  if (osr == NULL) {
    osr = OptoRuntime::generate_osr_blob(framesize);
    if (osr == NULL) {
       //CodeCache is probably full
       return NULL;
    }

    OSRAdapter* check = NULL;
    { MutexLocker mu(AdapterCache_lock);
      // Grap lock and update cache. Check if someone else already updated the list
      check = _osr_adapters->at(framesize);
      if (check == NULL) {
        _osr_adapters->at_put(framesize, osr);    
      } 
    }

    // The adapter already existed
    if (check != NULL) {
      MutexLocker mu(CodeCache_lock);
      CodeCache::free(osr);     
      osr = check;                
    }      
  }
  return osr;
}
#endif // COMPILER2


  // Returns an osr nmethod if possible; returns NULL otherwise.
nmethod* OnStackReplacement::get_osr_code(methodHandle method, int bci, TRAPS) {
#ifdef ASSERT  
  assert(UseOnStackReplacement, "on-stack replacement not used");
  frame fr = JavaThread::current()->last_frame();
  assert(JavaThread::current() == THREAD, "wrong thread");
  assert(fr.is_interpreted_frame(), "must replace an interpreter frame");
  assert(fr.interpreter_frame_method() == method(), "wrong method");
  assert(fr.interpreter_frame_bcp() - method()->code_base() == bci, "wrong bci");
#endif

  // check if on-stack-replacement is possible
  if (method->is_not_osr_compilable()) {
    return NULL;
  }

  // check if on-stack replacement nmethod already exist
  nmethod* osr = method->lookup_osr_nmethod_for(bci);
  // found a match => done
  if (osr != NULL) {
    if (TraceOnStackReplacement) {
      ResourceMark rm;
      tty->print("Reused OSR entry for ");
      method->print_short_name(tty);
      tty->print_cr(" at bci %d used", bci);
    }    
    return osr;
  }

  // this method has not had an OSR nmethod compiled for it
  if (TraceOnStackReplacement) {
    ResourceMark rm;
    tty->print("Compiling OSR entry for ");
    method->print_short_name(tty);
    tty->print(" at bci %d: ", bci);    
  }

  //----------------------------------------------------------------------------
  // NOTE: The code below (which makes sure that the counter doesn't overflow again
  // too soon) was copied from recompile.cpp.  The original has additional machinery not
  // useful to on stack replacement.  Whether we want to fiddle with the counters
  // in such a way is unclear -- it means that we will end up reusing our OSR code more
  // often in preference to immediately compiling a traditional entry at the following
  // invocation.

  // Make sure invocation counter doesn't overflow again right away
  // as would be the case for native methods.
  // Note: native methods cannot be OSR, disable code
 /* method->invocation_counter()->set_carry();
  assert(
    !method->was_never_executed(),
    "don't reset to 0 -- could be mistaken for never-executed"
  );
  */
  //----------------------------------------------------------------------------

  // compile method for osr
  osr = CompileBroker::compile_method(method, bci, CHECK_0);  
  if (TraceOnStackReplacement) {
    if (osr == NULL) tty->print_cr("compilation failed");
    else tty->print_cr("nmethod 0x%x", osr);
  }
  if (osr == NULL) {
    // delay compilation
    method->invocation_counter()->decay();
  }

  // done
  return osr;
}


void onStackReplacement_init() {
  // Note: C1 doesn't need any of the osr adapter machinery
  //       in this file as there is only one osr adapter and
  //       the corresponding code is generated via c1_Runtime1
  //       => no initialization code needed for now
#ifdef COMPILER2
  OnStackReplacement::initialize();

  // OnStackReplaceALot, disables all compiles and lowers the on-stack replace threshold
  if (OnStackReplaceALot) {
#ifndef PRODUCT
    // OptoStop may only be set in non-product builds.
    CICountOSR = true;
    CIStop = 0;
#endif // PRODUCT
    OnStackReplaceThreshold = 25;
  }
#endif // COMPILER2
}
