/*
 * @(#)compileBroker.hpp	1.17 00/03/21 16:16:46
 *
 * Copyright 1991-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

#ifdef COMPILER2
class BasicAdapter;
#endif

// CompileTask
//
// An entry in the compile queue.  It represents a pending or current
// compilation.
class CompileTask : public CHeapObj {
 private:
  Monitor*     _lock;
  uint         _compile_id;
  jobject      _method;
  int          _osr_bci;
  bool         _is_complete;
  bool         _is_success;
  bool         _is_blocking;
  int          _adapter_kind;
  CompileTask* _next;

 public:
  CompileTask() {
    _lock = new Monitor(Mutex::nonleaf+2, "CompileTaskLock");
  }

  void initialize(int compile_id, methodHandle method, int osr_bci, int adapter_kind, bool is_blocking);

  void free();

  int          compile_id() const                { return _compile_id; }
  jobject      method_handle() const             { return _method; }
  int          osr_bci() const                   { return _osr_bci; }
  bool         is_complete() const               { return _is_complete; }
  bool         is_blocking() const               { return _is_blocking; }
  bool         is_success() const                { return _is_success; }

  Monitor*     lock() const                      { return _lock; }

  void         mark_complete()                   { _is_complete = true; }
  void         mark_success()                    { _is_success = true; }

  int          adapter_kind() const              { return _adapter_kind; }
  bool         is_method_compile() const         { return _adapter_kind == ciEnv::not_adapter; }

  CompileTask* next() const                      { return _next; }
  void         set_next(CompileTask* next)       { _next = next; }

  void         print();
  void         print_line();
};


// CompileQueue
//
// A list of CompileTasks.
class CompileQueue : public CHeapObj {
 private:
  const char* _name;
  Monitor*    _lock;

  CompileTask* _first;
  CompileTask* _last;

 public:
  CompileQueue(const char* name, Monitor* lock) {
    _name = name;
    _lock = lock;
    _first = NULL;
    _last = NULL;
  }

  const char*  name() const                      { return _name; }
  Monitor*     lock() const                      { return _lock; }

  void         add(CompileTask* task);

  CompileTask* get();

  bool         is_empty() const                  { return _first == NULL; }

  void         print();
};


// Compilation
//
// The broker for all compilation requests.
class CompileBroker: AllStatic {
 friend class Threads;
  friend class CompileTaskWrapper;

 public:
  enum {
    name_buffer_length = 100
  };

 private:
  static bool _initialized;

  // The installed compiler
  static AbstractCompiler* _compiler;

  // These counters are used for assigning id's to each compilation
  static uint _compilation_id;
  static uint _osr_compilation_id;
  static uint _native_compilation_id;

  // Information for print_last_compile()
  enum { no_compile, normal_compile, osr_compile, native_compile };

  static int  _last_compile_type;
  static char _last_method_compiled[name_buffer_length];

  static CompileQueue* _adapter_queue;
  static CompileQueue* _method_queue;
  static CompileTask* _task_free_list;

  static CompilerThread*                 _adapter_thread;
  static GrowableArray<CompilerThread*>* _method_threads;

  static ThreadPriority _current_compile_priority;

  // Timers and counters for generating statistics
#ifndef PRODUCT
  static elapsedTimer _t_total_compilation;
  static elapsedTimer _t_native_compilation;
  static elapsedTimer _t_osr_compilation;
  static elapsedTimer _t_standard_compilation;

  static int _total_bailout_count;
  static int _total_invalidated_count;
  static int _total_compile_count;
  static int _total_native_compile_count;
  static int _total_osr_compile_count;
  static int _total_standard_compile_count;
  
  static int _sum_osr_bytes_compiled;
  static int _sum_standard_bytes_compiled;
  static int _sum_nmethod_size;
  static int _sum_nmethod_code_size;
#endif // PRODUCT

  static int compiler_count()                    { return CICompilerCount; }

  static CompilerThread* make_compiler_thread(const char* name, CompileQueue* queue, TRAPS);
  static void init_compiler_threads(int compiler_count);
  static bool check_compilation_result(methodHandle method, int osr_bci, nmethod** result);

#ifdef COMPILER2
  static bool check_adapter_result(methodHandle method, int adapter_kind, BasicAdapter** result);
  static BasicAdapter* compile_adapter_for(methodHandle method, int adapter_kind, bool blocking);
#endif // COMPILER2

  static bool compilation_is_in_queue  (methodHandle method, int osr_bci);
  static bool compilation_is_prohibited(methodHandle method, int osr_bci);
  static bool is_not_compile_only      (methodHandle method);
  static uint assign_compile_id        (methodHandle method, int osr_bci);
  static bool is_compile_blocking      (methodHandle method, int osr_bci);
  static void preload_classes          (methodHandle method, TRAPS);

  static CompileTask* create_compile_task(CompileQueue* queue,
                                          int           compile_id,
                                          methodHandle  method,
                                          int           osr_bci,
                                          int           adapter_kind,
                                          bool          blocking);
  static CompileTask* allocate_task();
  static void free_task(CompileTask* task);
  static nmethod* wait_for_completion(CompileTask* task);

#ifdef COMPILER2
  static BasicAdapter* wait_for_adapter_completion(CompileTask* task);
  static bool eager_compile_i2c_adapters(ciEnv* ci_env, ciMethod* target, ciScope* scope, bool env_used);
  static bool eager_compile_c2i_adapters(ciEnv* ci_env, ciMethod* target, ciScope* scope, bool env_used);
  static void invoke_compiler_on_adapter(CompileTask* task);
#endif // COMPILER2

  static void invoke_compiler_on_method(CompileTask* task);
  static void set_last_compile(methodHandle method, bool is_osr, bool is_native);
  static void push_jni_handle_block();
  static void pop_jni_handle_block();
  static bool check_break_at(int compile_id, bool is_osr, bool is_native);
  static void collect_statistics(elapsedTimer time, CompileTask* task) PRODUCT_RETURN;

  static nmethod* compile_method_base(methodHandle method, int osr_bci, TRAPS);

 public:
  enum {
    // The entry bci used for non-OSR compilations.
    standard_entry_bci = InvocationEntryBci
  };

  static AbstractCompiler* compiler()            { return _compiler; }

  static void compilation_init(AbstractCompiler* compiler);
  static nmethod* compile_method(methodHandle method, int osr_bci, TRAPS);

#ifdef COMPILER2
  static C2IAdapter* compile_c2i_adapter_for(methodHandle method, bool blocking) {
    return (C2IAdapter*)compile_adapter_for(method, ciEnv::c2i, blocking);
  }
  static I2CAdapter* compile_i2c_adapter_for(methodHandle method, bool blocking) {
    return (I2CAdapter*)compile_adapter_for(method, ciEnv::i2c, blocking);
  }
#endif // COMPILER2

  static void compiler_thread_loop();

  static void set_compile_priority(ThreadPriority priority);

  static bool is_idle();

  // Print a detailed accounting of compilation time
  static void print_times()                      PRODUCT_RETURN;

  // Debugging output for failure
  static void print_last_compile();

  static void print_compiler_threads();
};
