// @(#)callProfile.hpp	1.5 99/12/21 05:12:46
//
// Copyright 1993-1998 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2


//--------------------------TypeProfileEntry----------------------------------
// Simple data structure for observing monomorphism.
// Used for virtual and interface calls, and other type checking operations.
class TypeProfileEntry VALUE_OBJ_CLASS_SPEC {
 private:
  klassOop _klass;

#ifdef ASSERT
  // infrastructure for debugging and validating this mechanism
  jint	   _mono_count;   // number of updates while in monomorphic state
  klassOop _other_klass;  // conflicting klass

  // (if needed, pad the size of this struct to a power of two)
  jint	   _pad[1];
#endif

  friend class CallProfile;

 public:
  bool  is_empty() const         { return _klass == NULL; }
  void set_empty()               { _klass = NULL; }

  bool  is_klass() const         { return _klass != NULL && !is_conflict(); }
  klassOop klass() const         { assert(is_klass(), "must be singleton"); return _klass; }
  void set_klass(klassOop klass) { _klass = klass; assert(is_klass(), "must set a klass"); }

  bool  is_conflict() const      { return (bool)((int)_klass & 1); }
  void set_conflict()            { _klass = (klassOop)( debug_only((jint)_klass |) 1 ); }

  static int size()                         { return sizeof(TypeProfileEntry) / oopSize; }  // size in words

  static int klass_offset_in_bytes()        { return (int)&((TypeProfileEntry*)NULL)->_klass; }
#ifdef ASSERT
  static int mono_count_offset_in_bytes()   { return (int)&((TypeProfileEntry*)NULL)->_mono_count; }
  static int other_klass_offset_in_bytes()  { return (int)&((TypeProfileEntry*)NULL)->_other_klass; }
#endif

  void record_klass(klassOop klass) {
    // There are harmless race conditions here.
    // Eventually, every racer will top out into the conflict state, if necessary.
    // In order for this to work, every user of type profile data must be prepared
    // to see occasional "garbage" klasses (from unrelated bcps).
    if (klass != _klass) {
      if (is_empty()) {
	set_klass(klass); // (bottom|klass1) -> klass1
      } else {
	set_conflict();   // (klass2|klass1) -> top; (top|klass1) -> top
#ifdef ASSERT
	if (_other_klass == NULL)  _other_klass = klass;
#endif
      }
    } else {
      debug_only(++_mono_count);
    }
  }

  void print();
  void verify();
};


//--------------------------BCPProfileEntry----------------------------------
// In order to debug interpreter profiles, we can collect information about
// the BCPs that generate profile information.  For example, a call count
// profile is a table of hash buckets, each containing a count.  This structure
// records, for each the bucket, the BCP (or up to two BCPs) that generated
// the profile data.
class BCPProfileEntry VALUE_OBJ_CLASS_SPEC {
 private:
  address  _bcp;           // null if empty, bcp, or -(int)bcp for bcp conflict

#ifdef ASSERT
  // infrastructure for debugging and validating this mechanism
  address  _other_bcp;     // first conflicting bcp
  jint     _bcp_count;     // number of times this entry was probed w/o conflict
#endif

  friend class CallProfile;

 public:

#ifdef ASSERT
  jint     bcp_count() const             { return _bcp_count; }
  void set_bcp_count(jint new_count)     { _bcp_count = new_count; }
#endif

  bool  is_empty() const                 { return _bcp == NULL; }
  void set_empty()                       { _bcp = NULL; debug_only(_bcp_count = 0); }

  // layout info
  static int size()                      { return sizeof(BCPProfileEntry) / oopSize; }  // size in words
  static int bcp_offset_in_bytes()       { return (int)&((BCPProfileEntry*)NULL)->_bcp; }
#ifdef ASSERT
  static int other_bcp_offset_in_bytes() { return (int)&((BCPProfileEntry*)NULL)->_other_bcp; }
  static int bcp_count_offset_in_bytes() { return (int)&((BCPProfileEntry*)NULL)->_bcp_count; }
#endif

  address bcp();   // returns a non-conflict value if possible
  bool is_bcp_conflict();
  void record_bcp(address bcp);

  void print();
  void verify();
};

//--------------------------CallProfileEntry----------------------------------
// The Interpreter collects information about hot calls, and the compiler uses
// them to guide optimizations of call sites.  Multiple unrelated call sites might
// share (via hash bucket aliasing) a single CallProfileEntry, but we expect this
// to be infrequent and hence harmless.
class CallProfileEntry VALUE_OBJ_CLASS_SPEC {
 private:
  TypeProfileEntry _receiver;
  jint             _count;

#ifdef ASSERT
  // infrastructure for debugging and validating this mechanism
  BCPProfileEntry  _key; // where does this entry come from?

  // (if needed, pad the size of this struct to a power of two)
#endif

  friend class CallProfile;

 public:
  jint     count() const             { return _count; }
  void set_count(jint new_count)     { _count = new_count; }

  TypeProfileEntry& receiver()       { return _receiver; }
#ifdef ASSERT
  BCPProfileEntry& key()             { return _key; }
#endif

  bool  is_empty() const         { return _count == 0; }
  void set_empty()               { _count = 0; }

  bool  is_monomorphic() const   { return _receiver.is_klass(); }
//@@  klassOop monomorphic_klass() const { assert(is_monomorphic(), "must be singleton"); return _receiver.klass(); }

  static int size()                         { return sizeof(CallProfileEntry) / oopSize; }  // size in words

  static int count_offset_in_bytes()        { return (int)&((CallProfileEntry*)NULL)->_count; }
  static int receiver_offset_in_bytes()     { return (int)&((CallProfileEntry*)NULL)->_receiver; }
#ifdef ASSERT
  static int key_offset_in_bytes()          { return (int)&((CallProfileEntry*)NULL)->_key; }
#endif

  // There are harmless race conditions here.
  // Every racer will either add a little noise to the count,
  // or else "top out" eventually into a conflict state.
  void record_static_call() {
    ++_count;
  }
  void record_call(klassOop receiver) {
    record_static_call();
    _receiver.record_klass(receiver);
  }

  void print();
  void verify();
};

class CallProfile: AllStatic {
 private:
  static CallProfileEntry* _table;

  static bool is_initialized() { return _table != NULL; }

 public:
  static int table_length()       { return CallProfileSize; }
  static int table_index_mask()   { assert(is_power_of_2(table_length()), "table must be indexable via bit masking");
			            return table_length() - 1; }
  static address table_base()     { return (address)_table; }

  static CallProfileEntry* entry_addr_at(address bcp) {
    return &_table[(int)bcp & table_index_mask()];
  }
  static CallProfileEntry* entry_addr_at(methodHandle method, int bci) {
    return entry_addr_at(method->bcp_from(bci));
  }

  static CallProfileEntry entry_at(address bcp) {
    return *entry_addr_at(bcp);
  }
  static CallProfileEntry entry_at(methodHandle method, int bci) {
    return *entry_addr_at(method, bci);
  }

  // Utility function.  Queries the profile and resolves the method.
  static methodHandle resolve_method(methodHandle method, int bci, KlassHandle& target_klass);

  // Called once.  Allocates the table.
  static void initialize();
  static void clear();

  // Here's where the interpreter finds out when to start profiling:
  static int unprofiled_invocations() {
      int sub_threshold = CompileThreshold - FreqCountInvocations;
      if (sub_threshold < 0)  sub_threshold = 0;
      assert(sub_threshold >= 0 && sub_threshold <= CompileThreshold, "");
      return sub_threshold;
  }

  // GC support.
  static void gc_prologue();
  static void gc_epilogue();
  static void oops_do(void f(oop*));

  static void print_summary_line();
  static void print();
  static void verify();
};
