/*
 * @(#)interpreter_abs.cpp	1.174 01/03/05 10:08:58
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

#include "incls/_precompiled.incl"
#include "incls/_interpreter_abs.cpp.incl"

# define __ _masm->


//------------------------------------------------------------------------------------------------------------------------
// Implementation of InterpreterCodelet

void InterpreterCodelet::initialize(const char* description, bool is_safepoint_safe, Bytecodes::Code bytecode) {
  _description       = description;
  _is_safepoint_safe = is_safepoint_safe;
  _bytecode          = bytecode;
}


void InterpreterCodelet::verify() {
  if (bytecode() >= 0) {
    assert(is_safepoint_safe() == Bytecodes::safepoint_safe(bytecode()), "is_safepoint_safe flag inconsistent");
  }
}


void InterpreterCodelet::print() {
  if (PrintInterpreter) {
    tty->cr();
    tty->print_cr("----------------------------------------------------------------------");
  }

  if (description() != NULL) tty->print("%s  ", description());
  if (bytecode()    >= 0   ) tty->print("%d %s  ", bytecode(), Bytecodes::name(bytecode()));
  const char* _not = is_safepoint_safe() ? "" : "not ";
  tty->print_cr("[0x%x, 0x%x[  %d bytes  %ssafepoint safe", code_begin(), code_end(), code_size(), _not);

  if (PrintInterpreter) {
    tty->cr();
    Disassembler::decode(code_begin(), code_end(), tty);
  }
}


//------------------------------------------------------------------------------------------------------------------------
// Implementation of AbstractInterpreter


// Define a prototype interface
DEF_STUB_INTERFACE(InterpreterCodelet);


void AbstractInterpreter::initialize() {
  if (_code != NULL) return;

  // assertions
  assert(Bytecodes::number_of_codes <= DispatchTable::length, "dispatch table too small");

  // make sure 'imported' classes are initialized
  if (CountBytecodes || TraceBytecodes || StopInterpreterAt) BytecodeCounter::reset();
  if (PrintBytecodeHistogram)                                BytecodeHistogram::reset();
  if (PrintBytecodePairHistogram)                            BytecodePairHistogram::reset();
  TemplateTable::initialize();
#ifndef CORE
  InvocationCounter::reinitialize(DelayCompilationDuringStartup);
#endif

#ifdef COMPILER2
  // allocate branch frequency table
  if (CountInterpBranches) {
    SharedInfo::freqTable = NEW_C_HEAP_ARRAY(FreqPair, nth_bit(BranchFreqTableSize));
    SharedInfo::flush_branch_freq_table();
  }
  // allocate call frequency table
  if (CountInterpCalls) {
    SharedInfo::callTable = NEW_C_HEAP_ARRAY(int, nth_bit(CallFreqTableSize));
    SharedInfo::flush_call_freq_table();
  }

  if (UseCallProfile || UseTypeProfile) {
    CallProfile::initialize();
  }
#endif
  // generate interpreter
  { ResourceMark rm;
    TraceTime timer("Interpreter generation", TraceStartupTime);    
    int code_size = Interpreter_Code_Size;
    NOT_PRODUCT(code_size *= 3);     // extra space for StopInterpreterAt, etc.
    _code = new StubQueue(new InterpreterCodeletInterface, code_size, NULL);
    InterpreterGenerator g(_code);
    if (PrintInterpreter) print();
  }

  // initialize dispatch table
  _active_table = _normal_table;
}


void AbstractInterpreter::print() {
  tty->cr();
  tty->print_cr("----------------------------------------------------------------------");
  tty->print_cr("Interpreter");
  tty->cr();
  tty->print_cr("code size        = %6d bytes", _code->used_space());
  tty->print_cr("total space      = %6d bytes", _code->total_space());
  tty->print_cr("wasted space     = %6d bytes", _code->available_space());
  tty->cr();
  tty->print_cr("# of codelets    = %6d"      , _code->number_of_stubs());
  tty->print_cr("avg codelet size = %6d bytes", _code->used_space() / _code->number_of_stubs());
  tty->cr();
  _code->print();
  tty->print_cr("----------------------------------------------------------------------");
  tty->cr();
}


void interpreter_init() {
  Interpreter::initialize();
  if (TraceBytecodes) BytecodeTracer::set_closure(BytecodeTracer::std_closure());
  // need to hit every safepoint in order to call zapping routine
  // register the interpreter
  VTune::register_stub(
    "Interpreter",
    AbstractInterpreter::code()->code_start(),
    AbstractInterpreter::code()->code_end()
  );
}


//------------------------------------------------------------------------------------------------------------------------
// Implementation of EntryPoint

EntryPoint::EntryPoint() {
  assert(number_of_states == 5, "check the code below");
  _entry[itos] = NULL;
  _entry[ltos] = NULL;
  _entry[ftos] = NULL;
  _entry[dtos] = NULL;
  _entry[vtos] = NULL;
}


EntryPoint::EntryPoint(address ientry, address lentry, address fentry, address dentry, address ventry) {
  assert(number_of_states == 5, "check the code below");
  _entry[itos] = ientry;
  _entry[ltos] = lentry;
  _entry[ftos] = fentry;
  _entry[dtos] = dentry;
  _entry[vtos] = ventry;
}


void EntryPoint::set_entry(TosState state, address entry) {
  if (state == atos) state = itos; // handle atos state as itos
  assert(0 <= state && state < number_of_states, "state out of bounds");
  _entry[state] = entry;
}


address EntryPoint::entry(TosState state) const {
  if (state == atos) state = itos; // handle atos state as itos
  assert(0 <= state && state < number_of_states, "state out of bounds");
  return _entry[state];
}


void EntryPoint::print() {
  tty->print("[");
  for (int i = 0; i < number_of_states; i++) {
    if (i > 0) tty->print(", ");
    tty->print("0x%x", _entry[i]);
  }
  tty->print("]");
}


bool EntryPoint::operator == (const EntryPoint& y) {
  int i = number_of_states;
  while (i-- > 0) {
    if (_entry[i] != y._entry[i]) return false;
  }
  return true;
}


//------------------------------------------------------------------------------------------------------------------------
// Implementation of DispatchTable

EntryPoint DispatchTable::entry(int i) const {
  assert(0 <= i && i < length, "index out of bounds");
  return
    EntryPoint(
      _itos_table[i],
      _ltos_table[i],
      _ftos_table[i],
      _dtos_table[i],
      _vtos_table[i]
    );
}


void DispatchTable::set_entry(int i, EntryPoint& entry) {
  assert(0 <= i && i < length, "index out of bounds");
  assert(number_of_states == 5, "check the code below");
  _itos_table[i] = entry.entry(itos);
  _ltos_table[i] = entry.entry(ltos);
  _ftos_table[i] = entry.entry(ftos);
  _dtos_table[i] = entry.entry(dtos);
  _vtos_table[i] = entry.entry(vtos);
}


address* DispatchTable::table_for(TosState state) {
  switch (state) {
    case atos: // fall through
    case itos: return _itos_table;
    case ltos: return _ltos_table;
    case ftos: return _ftos_table;
    case dtos: return _dtos_table;
    case vtos: return _vtos_table;
  }
  ShouldNotReachHere();
  return NULL;
}


bool DispatchTable::operator == (DispatchTable& y) {
  int i = length;
  while (i-- > 0) {
    EntryPoint t = y.entry(i); // for compiler compatibility (BugId 4150096)
    if (!(entry(i) == t)) return false;
  }
  return true;
}


//------------------------------------------------------------------------------------------------------------------------
// Implementation of interpreter

StubQueue* AbstractInterpreter::_code                                       = NULL;
bool       AbstractInterpreter::_notice_safepoints                          = false;

address    AbstractInterpreter::_remove_activation_entry                    = NULL;
address    AbstractInterpreter::_rethrow_exception_entry                    = NULL;

address    AbstractInterpreter::_native_entry_begin                         = NULL;
address    AbstractInterpreter::_native_entry_end                           = NULL;


address    AbstractInterpreter::_throw_ArrayIndexOutOfBoundsException_entry = NULL;
address    AbstractInterpreter::_throw_ArrayStoreException_entry            = NULL;
address    AbstractInterpreter::_throw_ArithmeticException_entry            = NULL;
address    AbstractInterpreter::_throw_NullPointerException_entry           = NULL;
address    AbstractInterpreter::_throw_StackOverflowError_entry             = NULL;
address    AbstractInterpreter::_throw_exception_entry                      = NULL;
NOT_PRODUCT(address  AbstractInterpreter::_bci_histogram_entry              = NULL;)

EntryPoint AbstractInterpreter::_trace_code;
EntryPoint AbstractInterpreter::_return_entry[AbstractInterpreter::number_of_return_entries];
#ifndef CORE
EntryPoint AbstractInterpreter::_deopt_entry [AbstractInterpreter::number_of_deopt_entries ];
#endif
EntryPoint AbstractInterpreter::_continuation_entry;
EntryPoint AbstractInterpreter::_safept_entry;

address    AbstractInterpreter::_entry_table            [AbstractInterpreter::number_of_method_entries];
address    AbstractInterpreter::_return_3_addrs_by_index[AbstractInterpreter::number_of_return_addrs];
address    AbstractInterpreter::_return_5_addrs_by_index[AbstractInterpreter::number_of_return_addrs];
address    AbstractInterpreter::_result_handler_by_index[AbstractInterpreter::number_of_result_handlers];
address    AbstractInterpreter::_slow_signature_handler;

DispatchTable AbstractInterpreter::_active_table;
DispatchTable AbstractInterpreter::_normal_table;
DispatchTable AbstractInterpreter::_safept_table;
address    AbstractInterpreter::_wentry_point[DispatchTable::length];


//------------------------------------------------------------------------------------------------------------------------
// A CodeletMark serves as an automatic creator/initializer for Codelets
// (As a subclass of ResourceMark it automatically GC's the allocated
// code buffer and assemblers).

class CodeletMark: ResourceMark {
 private:
  InterpreterCodelet*         _clet;
  InterpreterMacroAssembler** _masm;

 public:
  CodeletMark(
    InterpreterMacroAssembler*& masm,
    const char* description,
    bool is_safepoint_safe = true,
    Bytecodes::Code bytecode = Bytecodes::_illegal)
  { // request all space (add some slack for Codelet data)
#ifdef PRODUCT
    int codelet_size = 8*K;
#else
    int codelet_size = 9*K;  // extra space for VerifyThread, etc..
#endif
    _clet = (InterpreterCodelet*)AbstractInterpreter::code()->request(codelet_size);
    guarantee(_clet != NULL, "not enough space for interpreter generation");
    // initialize Codelet attributes
    _clet->initialize(description, is_safepoint_safe, bytecode);
    // create assembler for code generation
    masm  = new InterpreterMacroAssembler(new CodeBuffer(_clet->code_begin(), _clet->code_size()));
    _masm = &masm;
  }  
  
  ~CodeletMark() {
    // align so printing shows nop's instead of random code at the end (Codelets are aligned)
    (*_masm)->align(wordSize);
    // make sure all code is in code buffer
    (*_masm)->flush();


    // commit Codelet
    AbstractInterpreter::code()->commit((*_masm)->code()->code_size());
    // make sure nobody can use _masm outside a CodeletMark lifespan
    *_masm = NULL;
  }
};


//------------------------------------------------------------------------------------------------------------------------
// Generation of complete interpreter

AbstractInterpreterGenerator::AbstractInterpreterGenerator(StubQueue* _code) {
  _masm                      = NULL;
  _unimplemented_bytecode    = NULL;
  _illegal_bytecode_sequence = NULL;
}


void AbstractInterpreterGenerator::generate_all() {
  { CodeletMark cm(_masm, "error exits", false);
    _unimplemented_bytecode    = generate_error_exit("unimplemented bytecode");
    _illegal_bytecode_sequence = generate_error_exit("illegal bytecode sequence - method not verified");
  }

  { CodeletMark cm(_masm, "bytecode tracing support", false);
    Interpreter::_trace_code =
      EntryPoint(
        generate_trace_code(itos),
        generate_trace_code(ltos),
        generate_trace_code(ftos),
        generate_trace_code(dtos),
        generate_trace_code(vtos)
      );
  }

#ifndef PRODUCT
  { CodeletMark cm(_masm, "bci histogram support", false);
    Interpreter::_bci_histogram_entry = generate_bci_histogram_entry();
  }
#endif

  { CodeletMark cm(_masm, "return entry points", false);
    for (int i = 0; i < Interpreter::number_of_return_entries; i++) {
      Interpreter::_return_entry[i] =
        EntryPoint(
          generate_return_entry_for(itos, i),
          generate_return_entry_for(ltos, i),
          generate_return_entry_for(ftos, i),
          generate_return_entry_for(dtos, i),
          generate_return_entry_for(vtos, i)
        );
    }
  }

#ifndef CORE

  { CodeletMark cm(_masm, "deoptimization entry points", false);
    for (int i = 0; i < Interpreter::number_of_deopt_entries; i++) {
      Interpreter::_deopt_entry[i] =
	EntryPoint(
	  generate_deopt_entry_for(itos, i),
	  generate_deopt_entry_for(ltos, i),
	  generate_deopt_entry_for(ftos, i),
	  generate_deopt_entry_for(dtos, i),
	  generate_deopt_entry_for(vtos, i)
	);	  
    } 
  }
#endif

  { CodeletMark cm(_masm, "result handlers for native calls", false);
    const BasicType types[Interpreter::number_of_result_handlers] = {
      T_BOOLEAN,
      T_CHAR   ,
      T_BYTE   ,
      T_SHORT  ,
      T_INT    ,
      T_LONG   ,
      T_VOID   ,
      T_FLOAT  ,
      T_DOUBLE ,
      T_OBJECT
    };
    int is_generated[Interpreter::number_of_result_handlers];
    memset(is_generated, 0, sizeof(is_generated));
    for (int i = 0; i < Interpreter::number_of_result_handlers; i++) {
      BasicType type = types[i];
      if (is_generated[Interpreter::BasicType_as_index(type)]++)  continue;
      Interpreter::_result_handler_by_index[Interpreter::BasicType_as_index(type)] = generate_result_handler_for(type);
    }
  }

  { CodeletMark cm(_masm, "slow signature handler", false);
    Interpreter::_slow_signature_handler = generate_slow_signature_handler();
  }

  for (int j = 0; j < 6; j++) {
    const TosState states[] = {itos, ltos, ftos, dtos, vtos, atos};
    Interpreter::_return_3_addrs_by_index[Interpreter::TosState_as_index(states[j])] = Interpreter::return_entry(states[j], 3);
    Interpreter::_return_5_addrs_by_index[Interpreter::TosState_as_index(states[j])] = Interpreter::return_entry(states[j], 5);
  }

  { CodeletMark cm(_masm, "continuation entry points", false);
    Interpreter::_continuation_entry =
      EntryPoint(
        generate_continuation_for(itos),
        generate_continuation_for(ltos),
        generate_continuation_for(ftos),
        generate_continuation_for(dtos),
        generate_continuation_for(vtos)
      );
  }

  { CodeletMark cm(_masm, "safepoint entry points", false);
    Interpreter::_safept_entry =
      EntryPoint(
        generate_safept_entry_for(itos, CAST_FROM_FN_PTR(address, InterpreterRuntime::at_safepoint)),
        generate_safept_entry_for(ltos, CAST_FROM_FN_PTR(address, InterpreterRuntime::at_safepoint)),
        generate_safept_entry_for(ftos, CAST_FROM_FN_PTR(address, InterpreterRuntime::at_safepoint)),
        generate_safept_entry_for(dtos, CAST_FROM_FN_PTR(address, InterpreterRuntime::at_safepoint)),
        generate_safept_entry_for(vtos, CAST_FROM_FN_PTR(address, InterpreterRuntime::at_safepoint))
      );
  }

  { CodeletMark cm(_masm, "exception handling", false);
    // (Note: this is not safepoint safe because thread may return to compiled code)
    generate_throw_exception();
  }

  { CodeletMark cm(_masm, "throw exception entrypoints");
    Interpreter::_throw_ArrayIndexOutOfBoundsException_entry = generate_exception_handler("java/lang/ArrayIndexOutOfBoundsException", NULL       );
    Interpreter::_throw_ArrayStoreException_entry            = generate_exception_handler("java/lang/ArrayStoreException"           , NULL       );
    Interpreter::_throw_ArithmeticException_entry            = generate_exception_handler("java/lang/ArithmeticException"           , "/ by zero");
    Interpreter::_throw_NullPointerException_entry           = generate_exception_handler("java/lang/NullPointerException"          , NULL       );
    Interpreter::_throw_StackOverflowError_entry             = generate_StackOverflowError_handler();
  }

#define method_entry(kind)                                                                    \
  { CodeletMark cm(_masm, "method entry point (kind = " #kind ")", false);                    \
    Interpreter::_entry_table[Interpreter::kind] = generate_method_entry(Interpreter::kind);  \
  }

  // all non-native method kinds  
  method_entry(zerolocals)
  method_entry(zerolocals_synchronized)
  method_entry(empty)
  method_entry(accessor)
  method_entry(abstract)
  method_entry(java_lang_math_sin )
  method_entry(java_lang_math_cos )
  method_entry(java_lang_math_sqrt)
  
  // all native method kinds (must be one contiguous block)
  Interpreter::_native_entry_begin = Interpreter::code()->code_end();
  method_entry(native)
  method_entry(native_synchronized)
  Interpreter::_native_entry_end = Interpreter::code()->code_end();

#undef method_entry

  // Bytecodes
  set_entry_points_for_all_bytes();
  set_safepoints_for_all_bytes();
}


//------------------------------------------------------------------------------------------------------------------------

address AbstractInterpreterGenerator::generate_error_exit(const char* msg) {
  address entry = __ pc();
  __ stop(msg);
  return entry;
}


//------------------------------------------------------------------------------------------------------------------------

void AbstractInterpreterGenerator::set_entry_points_for_all_bytes() {
  for (int i = 0; i < DispatchTable::length; i++) {
    Bytecodes::Code code = (Bytecodes::Code)i;
    if (Bytecodes::is_defined(code)) {
      set_entry_points(code);
    } else {
      set_unimplemented(i);
    }
  }
}


void AbstractInterpreterGenerator::set_safepoints_for_all_bytes() {
  for (int i = 0; i < DispatchTable::length; i++) {
    Bytecodes::Code code = (Bytecodes::Code)i;
    if (Bytecodes::is_defined(code)) Interpreter::_safept_table.set_entry(code, Interpreter::_safept_entry);
  }
}


void AbstractInterpreterGenerator::set_unimplemented(int i) {
  address e = _unimplemented_bytecode;
  EntryPoint entry(e, e, e, e, e);
  Interpreter::_normal_table.set_entry(i, entry);
  Interpreter::_wentry_point[i] = _unimplemented_bytecode;
}


void AbstractInterpreterGenerator::set_entry_points(Bytecodes::Code code) {
  CodeletMark cm(_masm, NULL, Bytecodes::safepoint_safe(code), code);
  // initialize entry points
  assert(_unimplemented_bytecode    != NULL, "should have been generated before");
  assert(_illegal_bytecode_sequence != NULL, "should have been generated before");
  address iep = _illegal_bytecode_sequence;
  address lep = _illegal_bytecode_sequence;
  address fep = _illegal_bytecode_sequence;
  address dep = _illegal_bytecode_sequence;
  address vep = _unimplemented_bytecode;
  address wep = _unimplemented_bytecode;
  // code for short & wide version of bytecode
  if (Bytecodes::is_defined(code)) {
    Template* t = TemplateTable::template_for(code);
    assert(t->is_valid(), "just checking");
    set_short_entry_points(t, iep, lep, fep, dep, vep);
  }
  if (Bytecodes::wide_is_defined(code)) {
    Template* t = TemplateTable::template_for_wide(code);
    assert(t->is_valid(), "just checking");
    set_wide_entry_point(t, wep);
  }
  // set entry points
  EntryPoint entry(iep, lep, fep, dep, vep);
  Interpreter::_normal_table.set_entry(code, entry);
  Interpreter::_wentry_point[code] = wep;
}


void AbstractInterpreterGenerator::set_wide_entry_point(Template* t, address& wep) {
  assert(t->is_valid(), "template must exist");
  assert(t->tos_in() == vtos, "only vtos tos_in supported for wide instructions")
  wep = __ pc(); generate_and_dispatch(t);
}


void AbstractInterpreterGenerator::set_short_entry_points(Template* t, address& iep, address& lep, address& fep, address& dep, address& vep) {
  assert(t->is_valid(), "template must exist");
  switch (t->tos_in()) {
    case atos: // fall through
    case itos: vep = __ pc(); __ pop(itos); iep = __ pc(); generate_and_dispatch(t); break;
    case ltos: vep = __ pc(); __ pop(ltos); lep = __ pc(); generate_and_dispatch(t); break;
    case ftos: vep = __ pc(); __ pop(ftos); fep = __ pc(); generate_and_dispatch(t); break;
    case dtos: vep = __ pc(); __ pop(dtos); dep = __ pc(); generate_and_dispatch(t); break;
    case vtos: set_vtos_entry_points(t, iep, lep, fep, dep, vep);                    break;
    default  : ShouldNotReachHere();                                                 break;
  }
}


//------------------------------------------------------------------------------------------------------------------------

void AbstractInterpreterGenerator::generate_and_dispatch(Template* t, TosState tos_out) {
  if (PrintBytecodeHistogram)                                    histogram_bytecode(t);
#ifndef PRODUCT
  // debugging code
  if (CountBytecodes || TraceBytecodes || StopInterpreterAt > 0) count_bytecode();
  if (PrintBytecodePairHistogram)                                histogram_bytecode_pair(t);
#endif
  if (TraceBytecodes || EnableJVMPIInstructionStartEvent)        trace_bytecode(t);
#ifndef PRODUCT
  if (PrintBCIHistogram)                                         histogram_bci();
  if (StopInterpreterAt > 0)                                     stop_interpreter_at();
  __ verify_FPU(1, t->tos_in());
#endif
  int step;
  if (!t->does_dispatch()) { 
    step = t->is_wide() ? Bytecodes::wide_length_for(t->bytecode()) : Bytecodes::length_for(t->bytecode());
    if (tos_out == ilgl) tos_out = t->tos_out();
    // compute bytecode size
    assert(step > 0, "just checkin'");    
    // setup stuff for dispatching next bytecode 
    __ dispatch_prolog(tos_out, step);
  }
  // generate template
  t->generate(_masm);
  // advance
  if (t->does_dispatch()) {
    #ifdef ASSERT
    // make sure execution doesn't go beyond this point if code is broken
    __ should_not_reach_here();
    #endif
  } else {
    // dispatch to next bytecode
    __ dispatch_epilog(tos_out, step);
  }
}


//------------------------------------------------------------------------------------------------------------------------
// Entry points

AbstractInterpreter::MethodKind AbstractInterpreter::method_kind(methodHandle m) {
  // Abstract method?
  if (m->is_abstract()) return abstract;

  // Native method?
  // Note: This test must come _before_ the test for intrinsic
  //       methods. See also comments below.
  if (m->is_native()) {
    return m->is_synchronized() ? native_synchronized : native;
  } 

  // Synchronized?
  if (m->is_synchronized()) {
    return zerolocals_synchronized;
  } 

  // Empty method?
  if (m->is_empty_method()) {
    return empty;
  } 
  
  // Accessor method?
  if (m->is_accessor()) {
    assert(m->size_of_parameters() == 1, "fast code for accessors assumes parameter size = 1");
    return accessor;
  }
  
  // Special intrinsic method?
  // Note: This test must come _after_ the test for native methods,
  //       otherwise we will run into problems with JDK 1.2, see also
  //       AbstractInterpreterGenerator::generate_method_entry() for
  //       for details.
  switch (m->intrinsic_id()) {
    case methodOopDesc::_dsin : return java_lang_math_sin ;
    case methodOopDesc::_dcos : return java_lang_math_cos ;
    case methodOopDesc::_dsqrt: return java_lang_math_sqrt;
  }

  // Note: for now: zero locals for all non-empty methods
  return zerolocals;  
}


// Return true if the interpreter can prove that the given bytecode has
// not yet been executed (in Java semantics, not in actual operation).
bool AbstractInterpreter::is_not_reached(methodHandle method, int bci) {
  address bcp = method->bcp_from(bci);

  if (!Bytecode_at(bcp)->must_rewrite()) {
    // might have been reached
    return false;
  }

  // the bytecode might not be rewritten if the method is an accessor, etc.
  address ientry = method->interpreter_entry();
  if (ientry != entry_for_kind(AbstractInterpreter::zerolocals) &&
      ientry != entry_for_kind(AbstractInterpreter::zerolocals_synchronized))
    return false;  // interpreter does not run this method!

  // otherwise, we can be sure this bytecode has never been executed
  return true;
}


#ifndef PRODUCT
void AbstractInterpreter::print_method_kind(MethodKind kind) {
  switch (kind) {
    case zerolocals             : tty->print("zerolocals"             ); break;
    case zerolocals_synchronized: tty->print("zerolocals_synchronized"); break;
    case native                 : tty->print("native"                 ); break;
    case native_synchronized    : tty->print("native_synchronized"    ); break;
    case empty                  : tty->print("empty"                  ); break;
    case accessor               : tty->print("accessor"               ); break;
    case abstract               : tty->print("abstract"               ); break;
    case java_lang_math_sin     : tty->print("java_lang_math_sin"     ); break;
    case java_lang_math_cos     : tty->print("java_lang_math_cos"     ); break;
    case java_lang_math_sqrt    : tty->print("java_lang_math_sqrt"    ); break;
    default                     : ShouldNotReachHere();
  }
}
#endif


address AbstractInterpreter::return_entry(TosState state, int length) {
  guarantee(0 <= length && length < Interpreter::number_of_return_entries, "illegal length");
  return _return_entry[length].entry(state);
}


#ifndef CORE
address AbstractInterpreter::deopt_entry(TosState state, int length) {
  guarantee(0 <= length && length < Interpreter::number_of_deopt_entries, "illegal length");
  return _deopt_entry[length].entry(state);
}
#endif


static BasicType constant_pool_type( methodOop method, int index ) {
  constantTag tag = method->constants()->tag_at(index);
       if( tag.is_int   () ) return T_INT;
  else if( tag.is_float () ) return T_FLOAT;
  else if( tag.is_long  () ) return T_LONG;
  else if( tag.is_double() ) return T_DOUBLE;
  else if( tag.is_string() ) return T_OBJECT;
  else if( tag.is_unresolved_string() ) return T_OBJECT;
  ShouldNotReachHere();
  return T_ILLEGAL;
}

//------------------------------------------------------------------------------------------------------------------------
// Deoptimization support

// If deoptimization happens, this method returns the point where to continue in
// interpreter. For calls (invokexxxx, newxxxx) the continuation is at next
// bci and the top of stack is in eax/edx/FPU tos.
// For putfield/getfield, put/getstatic, the continuation is at the same
// bci and the TOS is on stack.

// Note: deopt_entry(type, 0) means reexecute bytecode
//       deopt_entry(type, length) means continue at next bytecode

#ifndef CORE
address AbstractInterpreter::continuation_for(methodOop method, address bcp, int callee_parameters, bool is_top_frame) {
  assert(method->contains(bcp), "just checkin'");
  Bytecodes::Code code   = Bytecodes::java_code(Bytecodes::cast(*bcp));
  int             bci    = method->bci_from(bcp);
  int             length = -1; // initial value for debugging
  // compute continuation length
  length = Bytecodes::length_at(bcp);
  // compute result type
  BasicType type = T_ILLEGAL;
  switch (code) {
#ifdef COMPILER2
    case Bytecodes::_ifnull    : // fall through
    case Bytecodes::_ifnonnull : // fall through
    case Bytecodes::_goto      : // fall through
    case Bytecodes::_goto_w    : // fall through
    case Bytecodes::_lookupswitch:
    case Bytecodes::_tableswitch:
    case Bytecodes::_fast_binaryswitch:
    case Bytecodes::_fast_linearswitch:
#endif
    // compiler safepoints must be reexecuted
    case Bytecodes::_ifeq      : // fall through
    case Bytecodes::_ifne      : // fall through
    case Bytecodes::_iflt      : // fall through
    case Bytecodes::_ifge      : // fall through
    case Bytecodes::_ifgt      : // fall through
    case Bytecodes::_ifle      : // fall through
    case Bytecodes::_if_icmpeq : // fall through
    case Bytecodes::_if_icmpne : // fall through
    case Bytecodes::_if_icmplt : // fall through
    case Bytecodes::_if_icmpge : // fall through
    case Bytecodes::_if_icmpgt : // fall through
    case Bytecodes::_if_icmple : // fall through
    case Bytecodes::_if_acmpeq : // fall through
    case Bytecodes::_if_acmpne : 
      // reexecute the operation and TOS value is on stack
      assert(is_top_frame, "must be top frame");
      return deopt_entry(vtos, 0);
      break;
    // special cases
    case Bytecodes::_getfield:
    case Bytecodes::_putfield:
    case Bytecodes::_getstatic:
    case Bytecodes::_putstatic:
    case Bytecodes::_aastore:
      // reexecute the operation and TOS value is on stack
      assert(is_top_frame, "must be top frame");
      return deopt_entry(vtos, 0);
      break;

    case Bytecodes::_invokevirtual  : // fall through
    case Bytecodes::_invokespecial  : // fall through
    case Bytecodes::_invokestatic   : // fall through
    case Bytecodes::_invokeinterface: {
      ResourceMark rm;
      type = Bytecode_invoke_at(method, bci)->result_type();
      // This is a hack needed because the cache entry might not be initialized.
      // NOT needed for the old calling convension.
      // Robert will find a beter solution soon.....
      if (!is_top_frame) {
        int index = Bytes::get_native_u2(bcp+1);
        method->constants()->cache()->entry_at(index)->set_parameter_size(callee_parameters);
      }
      break;
    }    
    case Bytecodes::_ldc:    type = constant_pool_type( method, *(bcp+1) ); break;
    case Bytecodes::_ldc_w : 
    case Bytecodes::_ldc2_w: type = constant_pool_type( method, Bytes::get_Java_u2(bcp+1) ); break;

    default:
      type = Bytecodes::result_type(code);
      break;
  }


  // return entry point for computed continuation state & bytecode length
  return is_top_frame
       ? deopt_entry(as_TosState(type), length)
       : return_entry(as_TosState(type), length);
}

// This method lays out the basic structure of an interpreted frame:
frame AbstractInterpreter::layout_activation(frame caller_frame, methodOop method, int tempcount, int moncount, int callee_params, int callee_locals) {
  layout_activation_impl(method, tempcount, moncount, &caller_frame, callee_params, callee_locals);
  return caller_frame;               // copied out as the interpreter frame
}
#endif




//------------------------------------------------------------------------------------------------------------------------
// Suport for invokes

int AbstractInterpreter::TosState_as_index(TosState state) {
  int i = 0;
  switch (state) {
    case itos: i = 0; break;
    case ftos: i = 1; break;
    case ltos: i = 2; break;
    case dtos: i = 3; break;
    case atos: i = 4; break;
    case vtos: i = 5; break;
    default  : ShouldNotReachHere();
  }
  assert(0 <= i && i < AbstractInterpreter::number_of_return_addrs, "index out of bounds");
  return i;
}


//------------------------------------------------------------------------------------------------------------------------
// Suport for native calls
// %%%% has been moved to interpreter_<arch>.cpp

//------------------------------------------------------------------------------------------------------------------------
// Safepoint suppport

static inline void copy_table(address* from, address* to, int size) {
  // Copy non-overlapping tables. The copy has to occur word wise for MT safety.
  while (size-- > 0) *to++ = *from++;
}


void AbstractInterpreter::notice_safepoints() {
  if (jvmdi::enabled()) {
    jvmdi::notice_safepoints();
  }
  if (!_notice_safepoints) {
    // switch to safepoint dispatch table
    _notice_safepoints = true;
    copy_table((address*)&_safept_table, (address*)&_active_table, sizeof(_active_table) / sizeof(address));
  }
}


void AbstractInterpreter::ignore_safepoints() {
  if (_notice_safepoints) {
    if ((!jvmdi::enabled()) || (jvmdi::ignore_safepoints())) {
      // switch to normal dispatch table
      _notice_safepoints = false;
      copy_table((address*)&_normal_table, (address*)&_active_table, sizeof(_active_table) / sizeof(address));
    }
  }
}
