/*
 * @(#)interpreter_abs.hpp	1.109 01/02/09 14:44:56
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


// This file contains the platform-independant parts
// of the interpreter and the interpreter generator.

//------------------------------------------------------------------------------------------------------------------------
// An InterpreterCodelet is a piece of interpreter code. All
// interpreter code is generated into little codelets which
// contain extra information for safepoint, debugging and
// printing purposes.

class InterpreterCodelet: public Stub {
  friend class VMStructs;
 private:
  int         _size;                             // the size in bytes
  const char* _description;                      // a description of the codelet, for debugging & printing
  bool        _is_safepoint_safe;                // true if thread cannot escape interpreter during a safepoint, false otherwise
  Bytecodes::Code _bytecode;                     // associated bytecode if any

 public:
  // Initialization/finalization
  void    initialize(int size)                   { _size = size; }
  void    finalize()                             { ShouldNotCallThis(); }

  // General info/converters
  int     size() const                           { return _size; }
  static  int code_size_to_size(int code_size)   { return sizeof(InterpreterCodelet) + code_size; }

  // Code info
  address code_begin() const                     { return (address)this + sizeof(InterpreterCodelet); }
  address code_end() const                       { return (address)this + size(); }

  // Debugging
  void    verify();
  void    print();

  // Interpreter-specific initialization
  void    initialize(const char* description, bool is_safepoint_safe, Bytecodes::Code bytecode);

  // Interpreter-specific attributes
  int         code_size() const                  { return code_end() - code_begin(); }
  const char* description() const                { return _description; }
  bool        is_safepoint_safe() const          { return _is_safepoint_safe; }
  Bytecodes::Code bytecode() const               { return _bytecode; }
};


//------------------------------------------------------------------------------------------------------------------------
// A little wrapper class to group tosca-specific entry points into a unit.
// (tosca = Top-Of-Stack CAche)

class EntryPoint VALUE_OBJ_CLASS_SPEC {
 private:
  address _entry[number_of_states];

 public:
  // Construction
  EntryPoint();
  EntryPoint(address ientry, address lentry, address fentry, address dentry, address ventry);

  // Attributes
  address entry(TosState state) const;                // return target address for a given tosca state
  void    set_entry(TosState state, address entry);   // set    target address for a given tosca state
  void    print();

  // Comparison
  bool operator == (const EntryPoint& y);             // for debugging only
};


//------------------------------------------------------------------------------------------------------------------------
// A little wrapper class to group tosca-specific dispatch tables into a unit.

class DispatchTable VALUE_OBJ_CLASS_SPEC {
 public:
  enum { length = 1 << BitsPerByte };                 // an entry point for each byte value (also for undefined bytecodes)

 private:
  address _itos_table[length];                        // the dispatch table for itos tosca state
  address _ltos_table[length];                        // the dispatch table for ltos tosca state
  address _ftos_table[length];                        // the dispatch table for ftos tosca state
  address _dtos_table[length];                        // the dispatch table for dtos tosca state
  address _vtos_table[length];                        // the dispatch table for vtos tosca state

 public:
  // Attributes
  EntryPoint entry(int i) const;                      // return entry point for a given bytecode i
  void       set_entry(int i, EntryPoint& entry);     // set    entry point for a given bytecode i
  address*   table_for(TosState state);

  // Comparison
  bool operator == (DispatchTable& y);                // for debugging only
};


//------------------------------------------------------------------------------------------------------------------------
// The C++ interface to the bytecode interpreter.

class AbstractInterpreter: AllStatic {
  friend class VMStructs;
 public:
  enum MethodKind {        
    zerolocals,                                                 // method needs locals initialization
    zerolocals_synchronized,                                    // method needs locals initialization & is synchronized
    native,                                                     // native method
    native_synchronized,                                        // native method & is synchronized
    empty,                                                      // empty method (code: _return)
    accessor,                                                   // accessor method (code: _aload_0, _getfield, _(a|i)return)
    abstract,                                                   // abstract method (throws an AbstractMethodException)
    java_lang_math_sin,                                         // implementation of java.lang.Math.sin (x)
    java_lang_math_cos,                                         // implementation of java.lang.Math.cos (x)
    java_lang_math_sqrt,                                        // implementation of java.lang.Math.sqrt(x)
    number_of_method_entries
  };

  enum SomeConstants {
    number_of_return_entries  = 6,                              // number of return entry points
#ifndef CORE
    number_of_deopt_entries   = 6,                              // number of deoptimization entry points
#endif
    number_of_return_addrs    = 6,                              // number of return addresses
    number_of_result_handlers = 10                              // number of result handlers for native calls
  };    

 protected:
  static StubQueue* _code;                                      // the interpreter code (codelets)
  
  static address    _throw_ArrayIndexOutOfBoundsException_entry;
  static address    _throw_ArrayStoreException_entry;
  static address    _throw_ArithmeticException_entry;
  static address    _throw_NullPointerException_entry;
  static address    _throw_StackOverflowError_entry;
  static address    _throw_exception_entry;

  NOT_PRODUCT(static address    _bci_histogram_entry;)
  
  static address    _entry_table[number_of_method_entries];     // entry points for a given method

  static address    _remove_activation_entry;                   // continuation address if an exception is not handled by current frame
  static address    _rethrow_exception_entry;                   // rethrows an activation in previous frame

  static address    _native_entry_begin;                        // Region for native entry code
  static address    _native_entry_end;

  static EntryPoint _trace_code;
  static EntryPoint _return_entry[number_of_return_entries];    // entry points to return to from a call
#ifndef CORE
  static EntryPoint _deopt_entry[number_of_return_entries];     // entry points to return to from a deoptimization
#endif
  static EntryPoint _continuation_entry;
  static EntryPoint _safept_entry;

  static address    _return_3_addrs_by_index[number_of_return_addrs];     // for invokevirtual   return entries
  static address    _return_5_addrs_by_index[number_of_return_addrs];     // for invokeinterface return entries
  static address    _result_handler_by_index[number_of_result_handlers];  // for native method result handlers
  static address    _slow_signature_handler;                              // the native method generic (slow) signature handler

  static DispatchTable _active_table;                           // the active    dispatch table (used by the interpreter for dispatch)
  static DispatchTable _normal_table;                           // the normal    dispatch table (used to set the active table in normal mode)
  static DispatchTable _safept_table;                           // the safepoint dispatch table (used to set the active table for safepoints)
  static address       _wentry_point[DispatchTable::length];    // wide instructions only (vtos tosca always)
  
  static bool       _notice_safepoints;                         // true if safepoints are activated
  
  friend class      TemplateTable;
  friend class      AbstractInterpreterGenerator;
  friend class              InterpreterGenerator;
  friend class      InterpreterMacroAssembler;

 public:
  // Initialization/debugging
  static void       initialize();
  static StubQueue* code()                                      { return _code; }
  static bool       contains(address pc)                        { return _code->contains(pc); }

  // Method activation
  static MethodKind method_kind(methodHandle m);
  static address    entry_for_kind(MethodKind k)                { assert(0 <= k && k < number_of_method_entries, "illegal kind"); return _entry_table[k]; }
  static address    entry_for_method(methodHandle m)            { return _entry_table[method_kind(m)]; }
  static void       print_method_kind(MethodKind kind)          PRODUCT_RETURN;

  // Runtime support
  static bool       is_not_reached(                       methodHandle method, int bci);
  static int        size_activation(                      methodOop method, int temps, int monitors, int callee_params, int callee_locals);
  static frame      layout_activation(frame caller_frame, methodOop method, int temps, int monitors, int callee_params, int callee_locals);
  static address    remove_activation_entry()                   { return _remove_activation_entry; }
  static address    rethrow_exception_entry()                   { return _rethrow_exception_entry; }
  static address    throw_exception_entry()                     { return _throw_exception_entry; }
  static address    throw_ArithmeticException_entry()           { return _throw_ArithmeticException_entry; }
  static address    throw_NullPointerException_entry()          { return _throw_NullPointerException_entry; }
  static address    throw_StackOverflowError_entry()            { return _throw_StackOverflowError_entry; }

  // Code generation
  static address    trace_code    (TosState state)              { return _trace_code.entry(state); }
  NOT_PRODUCT(static address    bci_histogram_entry()           { return _bci_histogram_entry; })

  static address    return_entry  (TosState state, int length); // length = invoke bytecode length (to advance to next bytecode)
#ifndef CORE
  static address    deopt_entry   (TosState state, int length); // length = invoke bytecode length (to advance to next bytecode)
#endif
  static address    continuation  (TosState state)              { return _continuation_entry.entry(state); }
  static address*   dispatch_table(TosState state)              { return _active_table.table_for(state); }
  static address*   normal_table(TosState state)                { return _normal_table.table_for(state); }

  // Safepoint support
  static void       notice_safepoints();                        // stops the thread when reaching a safepoint
  static void       ignore_safepoints();                        // ignores safepoints

  // Deoptimization support
#ifndef CORE
  static address    continuation_for(methodOop method, address bcp, int callee_parameters, bool is_top_frame);
#endif

  // Support for invokes
  static address*   return_3_addrs_by_index_table()             { return _return_3_addrs_by_index; }
  static address*   return_5_addrs_by_index_table()             { return _return_5_addrs_by_index; }
  static int        TosState_as_index(TosState state);          // computes index into return_3_entry_by_index table

  // Support for native calls
  static address    slow_signature_handler()                    { return _slow_signature_handler; }
  static address    result_handler(BasicType type)              { return _result_handler_by_index[BasicType_as_index(type)]; }
  static int        BasicType_as_index(BasicType type);         // computes index into result_handler_by_index table
  static bool       in_native_entry(address pc)                 { return _native_entry_begin <= pc && pc < _native_entry_end; }
  // Debugging/printing
  static InterpreterCodelet* codelet_containing(address pc)     { return (InterpreterCodelet*)_code->stub_containing(pc); }
  static void       print();                                    // prints the interpreter code

 private:
  // share implementation of size_activation and layout_activation:
#ifndef CORE
  static int        layout_activation_impl(methodOop method, int temps, int monitors, frame* caller_frame, int callee_params, int callee_locals);
#endif
};


//------------------------------------------------------------------------------------------------------------------------
// The interpreter generator.

class Template;
class AbstractInterpreterGenerator: public StackObj {
 protected:
  InterpreterMacroAssembler* _masm;

  // entry points for shared code sequence
  address _unimplemented_bytecode;
  address _illegal_bytecode_sequence;

  // shared code sequences
  address generate_error_exit(const char* msg);
  address generate_StackOverflowError_handler();
  address generate_exception_handler(const char* name, const char* message);
  address generate_continuation_for(TosState state);
  address generate_return_entry_for(TosState state, int step);
#ifndef CORE
  address generate_deopt_entry_for(TosState state, int step);
#endif
  address generate_result_handler_for(BasicType type);
  address generate_slow_signature_handler();
  address generate_safept_entry_for(TosState state, address runtime_entry);
  void    generate_throw_exception();

  // entry point generator
  address generate_method_entry(AbstractInterpreter::MethodKind kind);
  
  // Instruction generation
  void generate_and_dispatch (Template* t, TosState tos_out = ilgl);
  void set_vtos_entry_points (Template* t, address& iep, address& lep, address& fep, address& dep, address& vep);
  void set_short_entry_points(Template* t, address& iep, address& lep, address& fep, address& dep, address& vep);
  void set_wide_entry_point  (Template* t, address& wep);

  void set_entry_points(Bytecodes::Code code);
  void set_unimplemented(int i);
  void set_entry_points_for_all_bytes();
  void set_safepoints_for_all_bytes();
  void generate_all();

  // Helpers for generate_and_dispatch
  address generate_trace_code(TosState state);
  address generate_bci_histogram_entry()        PRODUCT_RETURN0;
  void count_bytecode()                         PRODUCT_RETURN;  
  void histogram_bytecode(Template* t)          PRODUCT_RETURN;
  void histogram_bytecode_pair(Template* t)     PRODUCT_RETURN;
  void trace_bytecode(Template* t);
  void histogram_bci()                          PRODUCT_RETURN;
  void stop_interpreter_at()                    PRODUCT_RETURN;

 public:
  AbstractInterpreterGenerator(StubQueue* _code);
};
