/*
 * @(#)oopMapCache.hpp	1.32 99/06/22 16:40:24
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// A Cache for storing (method, bci) -> oopMap.
// The memory management system uses the cache when locating object references in an interpreted frame.
//
// OopMapCache's are allocated lazily per instanceKlass.

// The oopMap is stored as a bit mask. If the bit_mask can fit into two words it is stored in
// _low and _high, otherwise it is allocated on the C heap.
//
// If ENABBLE_ZAP_DEAD_LOCALS is defined, two bits are used per entry instead of one. In all cases,
// the first bit is set to indicate oops as opposed to other values. If the second bit is available,
// it is set for dead values. We get the following encoding:
//
// 00 live value
// 01 live oop
// 10 dead value
// 11 <unused>                                   (we cannot distinguish between dead oops or values with the current oop map generator)


class OffsetClosure  {
 public:
  virtual void offset_do(int offset) = 0;
};


class OopMapCacheEntry {
  friend class OopMapForCacheEntry;

 public:
  enum {
    N                = 2,                        // the number of words reserved for inlined mask storage
    small_mask_limit = N * BitsPerWord,          // the maximum number of bits available for small masks (i.e., masks stored inline)

#ifdef ENABLE_ZAP_DEAD_LOCALS
    bits_per_entry   = 2,
    dead_bit_number  = 1,
#else
    bits_per_entry   = 1,
#endif
    oop_bit_number   = 0
  };


 private:
  methodOop      _method;                        // the method for which the mask is valid
  unsigned short _bci;                           // the bci    for which the mask is valid
  unsigned short _mask_size;                     // the mask size in bits
  int            _bit_mask[N];                   // the bit mask if mask_size <= small_mask_limit, ptr to bit mask otherwise

  methodOop      method() const                  { return _method; }
  int            bci() const                     { return _bci; }
  int            mask_size() const               { return _mask_size; }
  int            number_of_entries() const       { return mask_size() / bits_per_entry; }
  unsigned int*  bit_mask()                      { return (unsigned int*)(mask_size() <= small_mask_limit ? (int)_bit_mask : _bit_mask[0]); }
  unsigned int   entry_at(int offset)            { int i = offset * bits_per_entry; return bit_mask()[i / BitsPerWord] >> (i % BitsPerWord); }

  void set_empty_mask()                          { _mask_size = 0; }
  void allocate_bit_mask();                      // allocates the bit mask on C heap f necessary
  void fill_for_native();                        // fills the bit mask for native calls
  bool verify_mask(CellTypeState *vars, CellTypeState *stack, int max_locals, int stack_top);

 public:
  // Lookup
  bool match(methodOop method, int bci)          { return _method == method && _bci == bci; }
  bool is_empty()                                { return _method == NULL; }  

  // Iteration
  bool is_value(int offset)                      { return (entry_at(offset) & (1 << oop_bit_number )) == 0; }
  bool is_oop  (int offset)                      { return (entry_at(offset) & (1 << oop_bit_number )) != 0; }
  void iterate_oop(OffsetClosure* oop_closure);

#ifdef ENABLE_ZAP_DEAD_LOCALS
  bool is_dead(int offset)                       { return (entry_at(offset) & (1 << dead_bit_number)) != 0; }
  void iterate_all(OffsetClosure* oop_closure, OffsetClosure* value_closure, OffsetClosure* dead_closure);
#endif

  // Initialization
  void fill(methodOop method, int bci);
  void set_mask(CellTypeState* vars, CellTypeState* stack, int stack_top);

  void copy(OopMapCacheEntry* from)               { memcpy((void *)this, (void *)from, sizeof(OopMapCacheEntry)); }

  void print();  
  void flush();
  void initialize();  
  void oops_do(void f(oop*));
  void oop_iterate(OopClosure * blk);
  void verify();
};


class OopMapCache : public CHeapObj {
 private:
  enum { _size        = 32,     // Use fixed size for now
         _probe_depth = 3       // probe depth in case of collisions
  };                          

  OopMapCacheEntry* _array;

  unsigned int hash_value_for(methodOop method, int bci);
  OopMapCacheEntry* entry_at(int i) const                { return &_array[i% _size]; }

 public:
  OopMapCache();
  ~OopMapCache();                                // free up memory

  // Returns the oopMap for (method, bci)
  OopMapCacheEntry* lookup(methodOop method, int bci, bool fill_if_absent);

  // Helpers
  void oops_do(void f(oop*))                     { for (int i = 0; i < _size; i++) _array[i].oops_do(f); }
  void oop_iterate(OopClosure *blk)              { for (int i = 0; i < _size; i++) _array[i].oop_iterate(blk); }
  void flush()                                   { for (int i = 0; i < _size; i++) _array[i].flush(); }
  void verify()                                  { for (int i = 0; i < _size; i++) _array[i].verify(); }

  // Returns total no. of bytes allocated as part of OopMapCache's
  static long memory_usage()                     PRODUCT_RETURN0;
};

