/*
 * @(#)rewriter.cpp	1.23 00/04/23 20:13:59
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_rewriter.cpp.incl"


// Computes an index_map (new_index -> original_index) for contant pool entries
// that are referred to by the interpreter at runtime via the constant pool cache.
void Rewriter::compute_index_maps(constantPoolHandle pool, GrowableArray<int>*& index_map, GrowableArray<int>*& inverse_index_map) {
  const int length  = pool->length();
  index_map         = new GrowableArray<int>(length, length, -1);
  inverse_index_map = new GrowableArray<int>(length / 8);
  for (int i = 0; i < length; i++) {
    switch (pool->tag_at(i).value()) {
      case JVM_CONSTANT_Fieldref          : // fall through
      case JVM_CONSTANT_Methodref         : // fall through
      case JVM_CONSTANT_InterfaceMethodref: {
        index_map->at_put(i, inverse_index_map->length());
        inverse_index_map->append(i);
      }
    }
  }
}


// Creates a constant pool cache given an inverse_index_map
constantPoolCacheHandle Rewriter::new_constant_pool_cache(GrowableArray<int>* inverse_index_map, TRAPS) {
  const int length = inverse_index_map->length();
  constantPoolCacheOop cache = oopFactory::new_constantPoolCache(length, CHECK_(constantPoolCacheOop(NULL)));
  cache->initialize(inverse_index_map);
  return cache;
}


// Rewrites a method given the index_map information
methodHandle Rewriter::rewrite_method(methodHandle method, GrowableArray<int>* index_map, TRAPS) {  
  BytecodeStream s(method);
  Bytecodes::Code c;
  int nof_jsrs = 0;  
  bool has_monitor_bytecodes = false;

  while ((c = s.next()) >= 0) {   
    switch (c) {
      case Bytecodes::_lookupswitch   : {
        Bytecode_lookupswitch* bc = Bytecode_lookupswitch_at(s.bcp());        
        bc->set_code( (bc->number_of_pairs() < BinarySwitchThreshold) ? Bytecodes::_fast_linearswitch :
                                                                        Bytecodes::_fast_binaryswitch);        
        break;
      }
      case Bytecodes::_getstatic      : // fall through
      case Bytecodes::_putstatic      : // fall through
      case Bytecodes::_getfield       : // fall through
      case Bytecodes::_putfield       : // fall through
      case Bytecodes::_invokevirtual  : // fall through
      case Bytecodes::_invokespecial  : // fall through
      case Bytecodes::_invokestatic   : // fall through
      case Bytecodes::_invokeinterface: {
        address p = s.bcp() + 1;
        Bytes::put_native_u2(p, index_map->at(Bytes::get_Java_u2(p)));
        break;
      }
      case Bytecodes::_jsr            : // fall through
      case Bytecodes::_jsr_w          : nof_jsrs++;                   break;      
      case Bytecodes::_monitorenter   : // fall through
      case Bytecodes::_monitorexit    : has_monitor_bytecodes = true; break;
    }  
  }  
  
  // Update access flags
  if (has_monitor_bytecodes) {
    method->set_has_monitor_bytecodes();
  }

  // The present of a jsr bytecode implies that the method might potentially
  // have to be rewritten, so we run the oopMapGenerator on the method
  if (nof_jsrs > 0) {
    ResolveOopMapConflicts romc(method);    
    methodHandle original_method = method;
    method = romc.do_potential_rewrite(CHECK_(methodOop(NULL)));
    if (method() != original_method()) {
      // Insert invalid bytecode into original methodOop and set
      // interpreter entrypoint, so that a executing this method
      // will manifest itself in an easy recognizable form.
      address bcp = original_method->bcp_from(0);
      *bcp = (u1)Bytecodes::_shouldnotreachhere;
      address entry = AbstractInterpreter::entry_for_method(original_method);
      original_method->set_interpreter_entry(entry);
    }
       
    // Update monitor matching info.
    if (romc.monitor_safe()) {
      method->set_guaranteed_monitor_matching();
    }
  }  

  // Setup method entrypoints for compiler and interpreter
  method->link_method();
      
  return method;
}


void Rewriter::rewrite(instanceKlassHandle klass, TRAPS) {
  // gather starting points  
  ResourceMark rm;
  constantPoolHandle pool = klass->constants();
  objArrayHandle methods  = klass->methods();
  assert(pool->cache() == NULL, "constant pool cache must not be set yet");

  // determine index maps for methodOop rewriting
  GrowableArray<int>* index_map         = NULL;
  GrowableArray<int>* inverse_index_map = NULL;
  compute_index_maps(pool, index_map, inverse_index_map);

  // allocate constant pool cache
  constantPoolCacheHandle cache = new_constant_pool_cache(inverse_index_map, CHECK);
  pool->set_cache(cache());
  cache->set_constant_pool(pool());
  
  // rewrite methods
  { int i = methods->length();
    while (i-- > 0) {       
      methodHandle m(THREAD, (methodOop)methods->obj_at(i));
      m = rewrite_method(m, index_map, CHECK);
      // Method might have gotten rewritten.
      methods->obj_at_put(i, m());      
    }
  }
}
