/*
 * @(#)carRememberedSet.cpp	1.28 00/01/18 10:30:37
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_carRememberedSet.cpp.incl"


void CarRememberedSet::allocate(int capacity) {
  assert(is_power_of_2(capacity), "remembered set capacity must be power of 2");
  _capacity_mask = capacity - 1;
  _set = NEW_C_HEAP_ARRAY(jbyte*, capacity);
  clear();
}


CarRememberedSet::~CarRememberedSet() {
  FREE_C_HEAP_ARRAY(jbyte*, _set);
}


void CarRememberedSet::clear() {
  _size = 0;
  _being_processed = false;
  for (int index = 0; index <= _capacity_mask; index++) {
    _set[index] = NULL;
  }
}


void CarRememberedSet::reinitialize(bool already_empty) {
  if (capacity() != initial_capacity) {
    FreeHeap(_set);
    allocate(initial_capacity);
  } else {
    if (already_empty) {
      _being_processed = false;
      debug_only(verify_is_cleared());
    } else {
      clear();
    }
  }
}


void CarRememberedSet::grow_and_rehash(int grow_factor) {
  int old_capacity = capacity();
  jbyte** old_set = _set;
  debug_only(int old_size = _size);
  allocate(old_capacity * grow_factor);

  if (PrintScavenge && Verbose) {
    tty->print_cr("[Growing carRememberedSet 0x%x from %d to %d]", this, old_capacity, capacity());
  }

  for (int old_index = 0; old_index < old_capacity; old_index++) {
    jbyte* old_card = old_set[old_index];
    if (old_card != NULL) {
      int index = index_for(old_card);
      while (_set[index] != NULL) {
        index = next_index(index);
      }
      _set[index] = old_card;
      _size++;
    }
  }
  assert(_size == old_size, "just checking");
  FreeHeap(old_set);
}


void CarRememberedSet::add_reference(oop* from) {
  if (!_being_processed) {
    jbyte* from_card = RememberedSet::byte_for(from);
    assert(RememberedSet::is_within_array_bounds(from_card), "just checking");
    assert(!Scavenge::is_active() || !Scavenge::do_train_tick() || 
            Universe::train_gen()->first_train()->first_car()->remembered_set() != this, "just checking");
    if (_size * 2 > capacity()) {
      grow_and_rehash(2);
    }
    int index = index_for(from_card);
    while (true) {
      if (_set[index] == from_card) return;
      if (_set[index] == NULL) break;
      index = next_index(index);
    }
    _set[index] = from_card;
    _size++;
  }
}


void CarRememberedSet::scavenge_higher_train_recorded_stores() {
  _being_processed = true;
  Train* first_train = Universe::train_gen()->first_train();
  assert(first_train->first_car()->remembered_set() == this, "just checking");
  for (int index = 0; index <= _capacity_mask; index++) {
    jbyte* card = _set[index];
    if (card != NULL) {
      oop* bottom = RememberedSet::oop_for(card);
      CarTableDesc* desc = CarTable::desc_for(bottom);
      if (first_train != desc->train()) {
        OldSpace* sp = (OldSpace*) desc->space();
        oop* top = min(sp->top(), bottom + RememberedSet::card_size_in_oops);
        oop* p = (oop *)sp->fast_object_start(bottom);
        while (p < top) {
          assert(p < sp->top(), "just checking");
          p += oop(p)->scavenge_contents(bottom, top);
        }
        // clear entry for next scan        
        _set[index] = NULL;
        _size--;
      }
    }
  }
}


void CarRememberedSet::scavenge_same_train_recorded_stores() {
  Train* first_train = Universe::train_gen()->first_train();
  assert(first_train->first_car()->remembered_set() == this, "just checking");
  for (int index = 0; index <= _capacity_mask; index++) {
    jbyte* card = _set[index];
    if (card != NULL) {
      oop* bottom = RememberedSet::oop_for(card);
      CarTableDesc* desc = CarTable::desc_for(bottom);
      OldSpace* sp = (OldSpace*) desc->space();
      oop* top = min(sp->top(), bottom + RememberedSet::card_size_in_oops);
      assert(first_train == desc->train(), "just checking");
      oop* p = (oop *)sp->fast_object_start(bottom);
      while (p < top) {
        assert(p < sp->top(), "just checking");
        p += oop(p)->scavenge_contents(bottom, top);
      }
      // done, clear entry
      _set[index] = NULL;
      _size--;
    }
  }
  assert(_size == 0, "checking size");
}


void CarRememberedSet::scavenge_special_nonoop_recorded_stores(bool exclude_lowest_car, bool exclude_lowest_train) {
  // Save all cards in a separate resource array and clear this remembered set in the process.
  // For a normal car, live objects are copied out so new entries should never be inserted 
  // during scanning. For a special large typearray car, we leave the object in place if live,
  // and valid remembered cards will thus get reinserted.
  assert((int)exclude_lowest_car >= (int)exclude_lowest_train, "sanity check");
  ResourceMark rm;
  int s = size();
  jbyte** cards = NEW_RESOURCE_ARRAY(jbyte*, s);
  int index = 0;
  for (int array_index = 0; array_index <= _capacity_mask; array_index++) {
    jbyte* card = _set[array_index];
    if (card != NULL) {
      cards[index++] = card;
      // clear entry
      _set[array_index] = NULL;
      _size--;
    }
  }
  assert(_size == 0, "checking size");
  assert(index == s, "checking size");
  // Scan the saved cards
  CarTableDesc* first_desc = Universe::train_gen()->first_car_desc();
  for (index = 0; index < s; index++) {
    jbyte* card = cards[index];
    oop* bottom = RememberedSet::oop_for(card);
    CarTableDesc* desc = CarTable::desc_for(bottom);
    // Check if card belongs to first train, and if so whether we can skip all references
    // from that train or only references from the first car.
    if (first_desc->train() != desc->train()
        || (!exclude_lowest_car)
        || (!exclude_lowest_train && !desc->equals(first_desc))) {
      OldSpace* sp = (OldSpace*) desc->space();
      oop* top = min(sp->top(), bottom + RememberedSet::card_size_in_oops);
      oop* p = (oop *)sp->fast_object_start(bottom);
      while (p < top) {
        assert(p < sp->top(), "just checking");
        p += oop(p)->scavenge_contents(bottom, top);
      }
    }
  }
}


bool CarRememberedSet::has_other_train_references(Train* t) {
  for (int index = 0; index <= _capacity_mask; index++) {
    jbyte* card = _set[index];
    if (card != NULL) {
      oop* card_start = RememberedSet::oop_for(card);
      CarTableDesc* card_desc = CarTable::desc_for(card_start);
      if (card_desc->train() != t) return true;
    }
  }
  return false;
}


#ifndef PRODUCT

void CarRememberedSet::verify(CarSpace* holder) {
  guarantee(holder->remembered_set() == this, "just checking");
  for (int index = 0; index <= _capacity_mask; index++) {
    jbyte* card = _set[index];
    if (card != NULL) {
      oop* card_start = RememberedSet::oop_for(card);
      CarTableDesc* card_desc = CarTable::desc_for(card_start);
      guarantee(holder->desc()->less_than(card_desc), "invalid card in remembered set");
    }
  }
}


bool CarRememberedSet::contains_reference(oop* from) const {
  jbyte* from_card = RememberedSet::byte_for(from);
  assert(RememberedSet::is_within_array_bounds(from_card), "just checking");
  int index = index_for(from_card);
  jbyte* card;
  do {
    card = _set[index]; 
    if (card == from_card) return true;
    index = next_index(index);
  } while (card != NULL);
  return false;
}


void CarRememberedSet::verify_is_cleared() const {
  for (int index = 0; index <= _capacity_mask; index++) {
    if (_set[index] != NULL) fatal("Remembered set contains non-null entry");
  }
  if (_size != 0) fatal("Remembered set size not zero");
}


void CarRememberedSet::print(CarSpace* holder) const {
  guarantee(holder->remembered_set() == this, "just checking");
  tty->print("Remembered set 0x%x for car (", this);
  tty->print_julong(holder->train_number());
  tty->print_cr(",%u)", holder->car_number());
  for (int index = 0; index <= _capacity_mask; index++) {
    jbyte* card = _set[index];
    if (card != NULL) {
      oop* card_start = RememberedSet::oop_for(card);
      CarTableDesc* card_desc = CarTable::desc_for(card_start);
      tty->print(" card 0x%x [0x%x;0x%x[ in ", card, card_start, card_start + RememberedSet::card_size_in_oops);
      if (card_desc->is_perm()) {
        tty->print_cr("perm space");
      } else {
        card_desc->space()->print_short();
        tty->cr();
      }
    }
  }
}

#endif
