/*
 * @(#)cartable.hpp	1.24 99/06/22 16:40:43
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


// Descriptor for individual entry in car table.
// 
// Invariant: the should_scavenge flag is always true for eden and cars in
// the first train. It is always false for perm space, and depending on
// which part of scavenge we are in it will be set for higher trains.
// 
// Prior to scavenge, the flag is set for from space and cleared for to space,
// and during scavenge the flag for train_gen is updated.

class CarTableDesc VALUE_OBJ_CLASS_SPEC {
 private:
  julong _train_number;
  juint  _car_number;
  Train* _train;
  Space* _space;
  bool   _should_scavenge;

   // Constants
  enum SomeConstants {
    invalid_train_number        = (julong)  0,
    newgen_train_number         = (julong)  1,
    special_nonoop_train_number = (julong)  2,
    initial_train_number        = (julong)  3,
  //tenured_gen_train_number    = (julong) -2,  (see below)
  //perm_gen_train_number       = (julong) -1,  (see below)

    invalid_car_number          = (juint)   0,
    initial_car_number          = (juint)   1
  };
  static const julong tenuredgen_train_number; // can't do enum -- outside of int range
  static const julong permgen_train_number;    // can't do enum -- outside of int range
  friend class CarTable;
  friend class TrainGeneration;
  friend class Scavenge;

 public:
  // Accessors
  julong train_number() const         { return _train_number; }
  juint car_number() const            { return _car_number; }
  Train* train() const                { return _train; }
  Space* space() const                { return _space; }
  void set_space(Space* s)            { _space = s; }
  bool should_scavenge() const        { return _should_scavenge; }
  void set_should_scavenge(bool b)    { _should_scavenge = b; }


  // Testers
  bool is_new() const                 { return _train_number == newgen_train_number; }
  bool is_special_nonoop() const      { return _train_number == special_nonoop_train_number; }
  bool is_train() const               { return _train != NULL; }
  bool is_tenured() const             { return _train_number == tenuredgen_train_number; }
  bool is_perm() const                { return _train_number == permgen_train_number; }

  bool equals(CarTableDesc* d) const {
    return (train_number() == d->train_number()) && (car_number() == d->car_number());
  }

  bool less_than(CarTableDesc* d) const {
    return (train_number() < d->train_number()) || (train_number() == d->train_number() && car_number() < d->car_number());
  }

  // Initialization
  void initialize(julong train_number, juint car_number, Train* train, Space* sp, bool should_scavenge);
};



// The cartable maintains a fast mapping of heap address to [train number, car number, train, space].
//
// It is only necessary to map the train heap, but in order to avoid bounds checks we map the
// entire heap. The extra space overhead is neglible.

class CarTable: AllStatic {
  friend class TrainGeneration;

 private:
  static CarTableDesc* _table;            // The actual table
  static CarTableDesc* _table_base;       // Adjusted base pointer
  static int           _table_size;       // Table size

  // Updating table entries
  static void update_entry(CarSpace* car, julong train_number, juint car_number, Train* train);
  static void clear_entry(CarSpace* car);

 public:
  // Table allocation and initialization
  static void initialize();
  static void initialize_new_generation();
  // Return train number holding oop
  static julong train_number_for(oop p)       { return desc_for((oop*)p)->train_number(); }
  // Return car holding number oop
  static juint car_number_for(oop p)          { return desc_for((oop*)p)->car_number(); }
  // Return train holding oop
  static Train* train_for(oop p)              { return desc_for((oop*)p)->train(); }
  // Return space holding oop
  static Space* space_for(oop p)              { return desc_for((oop*)p)->space(); }
  // Return table descriptor for address
  static CarTableDesc* desc_for(oop* p) { 
    CarTableDesc* entry = &(_table_base[(juint) p >> LogOfCarSpaceSize]);
    assert(entry >= _table && entry < _table + _table_size, "Invalid oop");
    return entry;
  }

  // Scavenge support
  static bool should_scavenge(oop p)          { return desc_for((oop*)p)->should_scavenge(); }
  static void update_after_scavenge();

  // Debugging. Like space_for, but returns NULL if p is outside heap.
  static Space* space_for_or_null(oop p) {
    CarTableDesc* entry = &(_table_base[(juint) p >> LogOfCarSpaceSize]);
    return (entry >= _table && entry < _table + _table_size) ? entry->space() : NULL;
  }

  // Verification
  static void verify() PRODUCT_RETURN;
};
