/*
 * @(#)generation.cpp	1.115 01/03/29 15:19:51
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_generation.cpp.incl"


#ifndef PRODUCT

void Generation::print() {
  tty->print(" %-20s", name());
  tty->print(" total %uK, used %uK", capacity()/K, used()/K);
  tty->print_cr(" [%#x, %#x)", low_boundary, high_boundary);
}


void Generation::print_max_capacity() {
  tty->print_cr("[%-20s maximum capacity %uK]", name(), _max_capacity/K);
}

  
void Generation::update_max_capacity() {
  if (Universe::is_fully_initialized()) {
    size_t c = capacity();
    if (_max_capacity < c) _max_capacity = c;
  }
}


// Utility iterator classes

class GenerationContainsClosure : public SpaceClosure {
 public:
  void* _p;
  bool  _found;
  virtual void do_space(Space* s) {
    if (!_found) {
      _found = s->contains(_p);
    }
  }
  GenerationContainsClosure(void* p) { _p = p; _found = false; }
};

class GenerationObjectStartClosure : public SpaceClosure {
 public:
  oop* _p;
  oop* _start;
  virtual void do_space(Space* s) {
    if (_start == NULL && s->contains(_p)) {
      _start = (oop*) s->slow_object_start(_p);
    }
  }
  GenerationObjectStartClosure(oop* p) { _p = p; _start = NULL; }
};



bool Generation::contains(void* p) {
  GenerationContainsClosure blk(p);
  space_iterate(&blk);
  return blk._found;
}


oop Generation::slow_object_start(oop* p) {
  GenerationObjectStartClosure blk(p);
  space_iterate(&blk);
  return oop(blk._start);
}

#endif // PRODUCT


Generation::Generation() {
  NOT_PRODUCT(_max_capacity = 0;)
}


void NewGeneration::swap_spaces() {
  if (jvmpi::is_event_enabled(JVMPI_EVENT_ARENA_DELETE)) {
    jvmpi::post_arena_delete_event(Universe::addr_to_arena_id(from()->bottom()));
  }
  if (jvmpi::is_event_enabled(JVMPI_EVENT_ARENA_NEW)) {
    jvmpi::post_arena_new_event(Universe::addr_to_arena_id(from()->bottom()), "Semi");
  }
  SurvivorSpace* s = from();
  from_space       = to();
  to_space         = s;
}


void NewGeneration::initialize(ReservedSpace rs, size_t initial_size) {
  if (!virtual_space.initialize(rs, initial_size)) {
    vm_exit_during_initialization("Could not reserve enough space for object heap");
  }
  low_boundary = (oop*) virtual_space.low_boundary();
  high_boundary= (oop*) virtual_space.high_boundary();

  from_space = new SurvivorSpace();
  to_space   = new SurvivorSpace();

  compute_space_boundaries(0);
}


void NewGeneration::compute_space_boundaries(size_t minimum_eden_size) {
  // All space sizes must be multiples of car size in order for the CarTable to work.
  // Note that the CarTable is used with and without train gc (for fast lookup).
  size_t alignment = (size_t) CarSpace::car_size();

  // Compute sizes
  size_t size = virtual_space.committed_size();
  int survivor_share = SurvivorRatio + 2;
  size_t survivor_size = max(align_size_down(size/survivor_share, alignment), alignment);
  size_t eden_size = size - (2*survivor_size);
  assert(survivor_size > 0 && eden_size > 0 && survivor_size <= eden_size, "just checking");

  if (eden_size < minimum_eden_size) {
    // May happen due to 64Kb rounding, if so adjust eden size back up
    minimum_eden_size = align_size_up(minimum_eden_size, alignment);
    size_t maximum_survivor_size = (size - minimum_eden_size) / 2;
    survivor_size = max(align_size_down(maximum_survivor_size, alignment), alignment);
    eden_size = size - (2*survivor_size);
    assert(survivor_size > 0 && eden_size > 0 && survivor_size <= eden_size, "just checking");
    assert(eden_size >= minimum_eden_size, "just checking");
  }

  char *eden_start = virtual_space.low();
  char *from_start = eden_start + eden_size;
  char *to_start   = from_start + survivor_size;
  char *to_end     = to_start   + survivor_size;

  assert(to_end == virtual_space.high(), "just checking");
  assert(Space::is_aligned((oop*)eden_start), "checking alignment");
  assert(Space::is_aligned((oop*)from_start), "checking alignment");
  assert(Space::is_aligned((oop*)to_start),   "checking alignment");

  eden()->initialize((oop*) eden_start, (oop*) from_start, (minimum_eden_size == 0));
  eden()->set_proper_top(NULL);
  from()->initialize((oop*) from_start, (oop*) to_start, true);
    to()->initialize((oop*) to_start,   (oop*) to_end,   true);

  if (jvmpi::is_event_enabled(JVMPI_EVENT_ARENA_NEW)) {
    jvmpi::post_arena_new_event(Universe::addr_to_arena_id(eden_start), "Eden");
    jvmpi::post_arena_new_event(Universe::addr_to_arena_id(from_start), "Semi");
    jvmpi::post_arena_new_event(Universe::addr_to_arena_id(to_start), "Semi");
  }
}


void NewGeneration::compute_new_generation_size() {
  // This is called after a full gc, so from-space will normally be empty.
  // Note that we check both spaces, since if scavenge failed they revert roles.
  // If not we bail out (otherwise we would have to relocate the objects)
  assert(to()->is_empty(), "to-space should be empty here");
  if (!from()->is_empty()) return;

  size_t old_size = Universe::old_gen()->capacity();
  size_t new_size_before = virtual_space.committed_size();
  size_t min_new_size = Universe::new_size();
  size_t max_new_size = Universe::reserved_new_size();
  assert(min_new_size <= new_size_before && new_size_before <= max_new_size, "just checking");

  // All space sizes must be multiples of car size in order for the CarTable to work.
  // Note that the CarTable is used with and without train gc (for fast lookup).
  size_t alignment = (size_t) CarSpace::car_size();

  // Compute desired new generation size based on NewRatio and NewSizeThreadIncrease
  size_t desired_new_size = old_size/NewRatio;
  int threads_count = Threads::number_of_non_daemon_threads();
  size_t thread_increase_size = threads_count * NewSizeThreadIncrease;
  desired_new_size = align_size_up(desired_new_size + thread_increase_size, alignment);

  // Adjust new generation size
  desired_new_size = max(min(desired_new_size, max_new_size), min_new_size);
  ptrdiff_t change = desired_new_size - new_size_before;
  assert(desired_new_size > 0 && desired_new_size <= max_new_size, "just checking");
  assert(change % alignment == 0, "just checking");

  bool changed = false;
  if (change > 0) {
    changed = virtual_space.expand_by(change);
    if (!changed) {
      // Do better than this for Merlin
      vm_exit_out_of_memory(change, "heap expansion");
    }
  }
  if (change < 0 && eden()->is_empty()) { // bail out of shrinking if objects in eden
    virtual_space.shrink_by(-change);
    changed = true;
  }
  if (changed) {
    compute_space_boundaries(eden()->used());
    CarTable::initialize_new_generation();
    if (Verbose && PrintGC) {
      size_t new_size_after  = virtual_space.committed_size();
      size_t eden_size_after = eden()->capacity();
      size_t survivor_size_after = from()->capacity();
      tty->print("New generation size %uK->%uK [eden=%uK,survivor=%uK]", 
        new_size_before/K, new_size_after/K, eden_size_after/K, survivor_size_after/K);
      if (WizardMode) {
        tty->print("[allowed %uK extra for %d threads]", thread_increase_size/K, threads_count);
      }
      tty->cr();
    }
    if(UseTLE && ResizeTLE) {
      ThreadLocalEden::adjust_max_size();
    }
  }
}


void NewGeneration::prepare_for_compaction() {
  assert(UseTrainGC, "should be called only when using train gc");
  // %note same order as in compact
  eden()->prepare_for_compaction();
}


void NewGeneration::adjust_pointers() {
  assert(UseTrainGC, "should be called only when using train gc");
  eden()->adjust_pointers();
}


void NewGeneration::compact() {
  assert(UseTrainGC, "should be called only when using train gc");
  // %note same order as in prepare_for_compaction
  eden()->compact();
}


size_t NewGeneration::capacity() {
  return eden()->capacity()
       + from()->capacity();  // to() is only used during scavenge
}


size_t NewGeneration::used() {
  return eden()->used()
       + from()->used();      // to() is only used during scavenge
}


size_t NewGeneration::free() {
  return eden()->free()
       + from()->free();      // to() is only used during scavenge
}


void NewGeneration::object_iterate(ObjectClosure* blk) {
  eden()->object_iterate(blk);
  from()->object_iterate(blk);
}


void NewGeneration::space_iterate(SpaceClosure* blk) {
  blk->do_space(eden());
  blk->do_space(from());
  blk->do_space(to());
}


void NewGeneration::set_saved_marks_to_top() { 
  _saved_mark = to()->top_mark();
}


void NewGeneration::set_saved_marks_to_bottom() { 
  _saved_mark = to()->bottom_mark();
}


bool NewGeneration::top_at_saved_marks() {
  return _saved_mark == to()->top_mark();
}


bool NewGeneration::scavenge_contents_from_saved_marks() {
  return to()->scavenge_contents_from(&_saved_mark);
}


#ifdef PRODUCT
void NewGeneration::verify(bool allow_dirty) {}
void NewGeneration::print() {}
#else

void NewGeneration::verify(bool allow_dirty) {
  eden()->verify(allow_dirty);
  from()->verify(allow_dirty);
    to()->verify(allow_dirty);
}


void NewGeneration::print() {
  Generation::print();
  tty->print("  eden");
  eden()->print();
  tty->print("  from");
  from()->print();
  tty->print("  to  ");
  to()->print();
}

#endif


const char* NewGeneration::name() {
  return "new generation";
}


void OldGeneration::initialize_card_marking_array(ReservedSpace rs) {
  if (!_card_marking_space.initialize(rs, 0)) {
    vm_exit_during_initialization("Could not reserve enough space for object heap");
  }
  grow_card_marking_array();
}


void OldGeneration::grow_card_marking_array() {
  // adjust size for grown generation
  // -/+ 1 below required to avoid assertion failure in byte_for
  char* new_high_card = (char*) RememberedSet::byte_for((oop*) virtual_space.high() - 1) + 1;
  char* current_high_card = _card_marking_space.high();
  if (new_high_card > current_high_card) {
    size_t delta = ReservedSpace::page_align_size_up(new_high_card - current_high_card);
    assert(delta > 0, "can only grow with positive amount");
    bool succeeded = _card_marking_space.expand_by(delta);
    if (!succeeded) {
      // Do better than this for Merlin
      vm_exit_out_of_memory(delta, "heap expansion");
    }
    assert(_card_marking_space.high() == current_high_card + delta, "invalid expansion");
    memset(current_high_card, RememberedSet::clean_card, delta);
  }
}


void OldGeneration::shrink_card_marking_array() {
  // adjust size for shrunk generation
  // -/+ 1 below required to avoid assertion failure in byte_for
  char* new_high_card = (char*) RememberedSet::byte_for((oop*) virtual_space.high() - 1) + 1;
  char* current_high_card = _card_marking_space.high();
  if (new_high_card < current_high_card) {
    size_t delta = ReservedSpace::page_align_size_down(current_high_card - new_high_card);
    if (delta > 0) {
      _card_marking_space.shrink_by(delta);
      assert(_card_marking_space.high() == current_high_card - delta, "invalid expansion");
    }
  }
}


void OldGeneration::clear_card_marking_array() {
  jbyte* start = (jbyte*) _card_marking_space.low();
  jbyte* end   = (jbyte*) _card_marking_space.high();
  while (start < end) {
    *start++ = RememberedSet::clean_card;
  }
}


void OldGeneration::initialize_offset_array() {
  size_t size = ReservedSpace::page_align_size_up(virtual_space.reserved_size() / RememberedSet::card_size);
  ReservedSpace rs(size);
  if (!rs.is_reserved()) {
    vm_exit_during_initialization("Could not reserve enough space for heap offset array");
  }
  if (!_offset_space.initialize(rs, 0)) {
    vm_exit_during_initialization("Could not reserve enough space for heap offset array");
  }
  _offset_array = (u_char*)_offset_space.low_boundary();
  grow_offset_array();
}


void OldGeneration::grow_offset_array() {
  // adjust size for grown generation
  size_t new_size = ReservedSpace::page_align_size_up(virtual_space.committed_size() / RememberedSet::card_size);
  size_t delta    = ReservedSpace::page_align_size_up(new_size - _offset_space.committed_size());
  if (delta == 0) return;   // can happen because of page size rounding
  assert(delta > 0, "can't grow by negative amount");
  char* high = _offset_space.high();
  bool succeeded = _offset_space.expand_by(delta);
  if (!succeeded) {
    // Do better than this for Merlin
    vm_exit_out_of_memory(delta, "heap expansion");
  }
  assert(_offset_space.high() == high + delta, "invalid expansion");
  memset(high, 0, delta);
  // This method is called whenever a generation grows, so we can update the maximum heap capaticy sizes here
  if (TraceMaxHeapCapacity) Universe::update_max_heap_capatity();
}


void OldGeneration::shrink_offset_array() {
  // adjust size for grown generation
  size_t new_size = ReservedSpace::page_align_size_up(virtual_space.committed_size() / RememberedSet::card_size);
  size_t delta    = ReservedSpace::page_align_size_down(_offset_space.committed_size() - new_size);
  if (delta == 0) return;   // can happen because of page size rounding
  assert(delta > 0, "can't grow by negative amount");
  char* high = _offset_space.high();
  _offset_space.shrink_by(delta);
  assert(_offset_space.high() == high - delta, "invalid expansion");
}


oop* OneSpaceOldGeneration::safepoint_allocate(int size) {
  oop* obj = _the_space->safepoint_allocate(size);
  return (obj != NULL) ? obj : safepoint_allocate_and_expand(size);
}


oop* OneSpaceOldGeneration::allocate_large_typearray(int size, TRAPS) {
  return allocate(size, THREAD);
}


oop* OneSpaceOldGeneration::safepoint_allocate_large_typearray(int size) {
  return safepoint_allocate(size);
}


oop* OneSpaceOldGeneration::allocate_and_collect(int size, TRAPS) {
  assert(Universe::no_gc_in_progress(), "just checking");
  MarkSweep::invoke(size, gc_cause(), CHECK_0);
  oop* obj = _the_space->allocate(size);
  if (obj == NULL) {
    // allocation failed even after doing GC, can happen if GC is locked,
    // if request is really large, or if space is close to limit
    return allocate_and_expand(size, 0, THREAD);
  } else {
    return obj;
  }
}


oop* OneSpaceOldGeneration::allocate_and_expand(int size, int expand_size, TRAPS) {
  assert(Universe::no_gc_in_progress(), "just checking");
  oop* obj = NULL;
  {
    // Grab ExpandHeap_lock since we are not at a safepoint, make sure to release it again before blocking
    MutexLocker ml(ExpandHeap_lock);
    // We have the lock, try allocating again and then expand
    bool expand_succeeded = true;
    while (obj == NULL) {
      obj = _the_space->allocate(size);
      if (obj == NULL) {
        if (!expand_succeeded) {
          // Cannot allocate or cannot expand more.
          THROW_OOP_0(Universe::out_of_memory_error_instance());
        }
        // Expand and retry allocation
        expand_succeeded = expand(size*wordSize, expand_size*wordSize); // safe if expansion fails
      }
    }
  }
  return obj;
}


oop* OneSpaceOldGeneration::safepoint_allocate_and_expand(int size) {
  assert(Scavenge::is_active(), "should only be called during scavenge");
  int free_before = free(); // in bytes
  expand(size*wordSize, MinOldHeapExpansion);
  int free_after = free();
  // expand() may not have fully succeeded but still expanded enough.
  if (free_after < size*wordSize) {
    // Do better than this for Merlin
    vm_exit_out_of_memory(size*wordSize, "heap expansion");
  }
  _expansion_at_safepoint += free_after - free_before;
  oop* obj = the_space()->safepoint_allocate(size);
  assert(obj != NULL, "allocation should have succeeded");
  return obj;
}


bool OneSpaceOldGeneration::expand(size_t bytes, size_t expand_bytes) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  size_t aligned_bytes  = ReservedSpace::page_align_size_up(bytes);
  size_t aligned_expand_bytes = ReservedSpace::page_align_size_up(expand_bytes);
  bool success = false;
  if (aligned_expand_bytes > aligned_bytes) {
    success = grow_by(aligned_expand_bytes);
  }
  if (!success) {
    success = grow_by(aligned_bytes);
  }
  if (!success) {
    grow_to_reserved();
  }
  if (GC_locker::is_active()) {
    if (PrintGC && Verbose) {
      tty->print_cr("Garbage collection disabled, expanded heap instead");
    }
  }
  return success;
}


void OneSpaceOldGeneration::shrink(size_t bytes) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  size_t size = ReservedSpace::page_align_size_down(bytes);
  if (size > 0) {
    shrink_by(size);
  }
}


size_t OneSpaceOldGeneration::capacity(){
  return _the_space->capacity();
}


size_t OneSpaceOldGeneration::used(){
  return _the_space->used();
}


size_t OneSpaceOldGeneration::free(){
  return _the_space->free();
}


size_t OneSpaceOldGeneration::contiguous_available(){
  return _the_space->free() + virtual_space.uncommitted_size();
}


bool OneSpaceOldGeneration::grow_by(size_t bytes) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  bool result = virtual_space.expand_by(bytes);
  if (result) {
    _the_space->set_end((oop*) virtual_space.high());
    grow_offset_array();
    grow_card_marking_array();
    if (Verbose && PrintGC) {
      size_t new_mem_size = virtual_space.committed_size();
      size_t old_mem_size = new_mem_size - bytes;
      tty->print_cr("Expanding %s from %uK to %uK", name(), old_mem_size/K, new_mem_size/K);
    }
  }
  assert(_the_space->top() <= _the_space->end(),
    "Grow of space malfunctioned");
  return result;
}


bool OneSpaceOldGeneration::grow_to_reserved() {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  size_t remaining_bytes = virtual_space.uncommitted_size();
  if (remaining_bytes > 0) {
    return grow_by(remaining_bytes);
  } else {
    // Already at reserved size
    return true;
  }
}


void OneSpaceOldGeneration::shrink_by(size_t bytes) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  virtual_space.shrink_by(bytes);

  _the_space->set_end((oop*) virtual_space.high());
  shrink_offset_array();
  shrink_card_marking_array();
  if (Verbose && PrintGC) {
    size_t new_mem_size = virtual_space.committed_size();
    size_t old_mem_size = new_mem_size + bytes;
    tty->print_cr("Shrinking %s from %uK to %uK", name(), old_mem_size/K, new_mem_size/K);
  }
  assert(_the_space->top() <= _the_space->end(),
    "Shrink of space malfunctioned");
}


bool OneSpaceOldGeneration::scavenge_contents_from_saved_marks() {
  return _the_space->scavenge_contents_from(&_saved_mark);
}


void OneSpaceOldGeneration::prepare_for_compaction() {
  _the_space->prepare_for_compaction();
}


void OneSpaceOldGeneration::adjust_pointers() {
  _the_space->adjust_pointers();
}


void OneSpaceOldGeneration::compact() {
  _the_space->compact();
}


#ifdef PRODUCT
void OneSpaceOldGeneration::verify(bool allow_dirty) {}
void OneSpaceOldGeneration::print() {}
#else

void OneSpaceOldGeneration::verify(bool allow_dirty) {
  _the_space->verify(allow_dirty);
}

#endif


void OneSpaceOldGeneration::object_iterate(ObjectClosure* blk) {
  _the_space->object_iterate(blk);
}


void OneSpaceOldGeneration::space_iterate(SpaceClosure* blk) {
  blk->do_space(_the_space);
}


void OneSpaceOldGeneration::object_iterate_from_saved_marks(ObjectClosure* blk) {
  WaterMark m = _saved_mark;
  _the_space->object_iterate_from(&m, blk);
}


void OneSpaceOldGeneration::scavenge_recorded_stores() {
  _the_space->scavenge_recorded_stores();
}


void OneSpaceOldGeneration::update_offset_array() {
  _the_space->update_offset_array();
}


oop OneSpaceOldGeneration::fast_object_start(oop* p) { 
  return _the_space->fast_object_start(p); 
}


void OneSpaceOldGeneration::set_saved_marks_to_top() { 
  _saved_mark = top_mark();
}


void OneSpaceOldGeneration::set_saved_marks_to_bottom() { 
  _saved_mark = bottom_mark();
}


bool OneSpaceOldGeneration::top_at_saved_marks() {
  return _saved_mark == top_mark();
}


#ifndef PRODUCT
void OneSpaceOldGeneration::print() {
  Generation::print();
  tty->print("  old ");
  _the_space->print();
}
#endif


void TenuredGeneration::initialize(ReservedSpace rs, size_t initial_size) {
  if (!virtual_space.initialize(rs, initial_size)) {
    vm_exit_during_initialization("Could not reserve enough space for object heap");
  }
  low_boundary  = (oop*) virtual_space.low_boundary();
  high_boundary = (oop*) virtual_space.high_boundary();
  initialize_offset_array();
  
  oop* bottom = (oop*) virtual_space.low();
  oop* end    = (oop*) virtual_space.high();
  _the_space  = new TenuredSpace(_offset_array, bottom, end);
  _saved_mark = bottom_mark();
  _shrink_factor = 0;
  _expansion_at_safepoint = 0;

  if (jvmpi::is_event_enabled(JVMPI_EVENT_ARENA_NEW)) {
    jvmpi::post_arena_new_event(Universe::addr_to_arena_id(bottom), name());
  }
}


const char* TenuredGeneration::name() {
  return "tenured generation";
}


oop* TenuredGeneration::allocate(int size, TRAPS) {
  oop* obj = _the_space->allocate(size);
  if (obj != NULL) {
    return obj;
  } else {      
    return allocate_and_collect(size, THREAD);
  }
}


void TenuredGeneration::compute_new_generation_size() {
  assert(_shrink_factor >= 0 && _shrink_factor <= 100, "invalid shrink factor");
  int current_shrink_factor = _shrink_factor;
  _shrink_factor = 0;

  double free_percentage = ((double) free()) / capacity();
  double minimum_free_percentage = (double) MinHeapFreeRatio / 100;
  double maximum_free_percentage = (double) MaxHeapFreeRatio / 100;
  size_t minimum_desired_capacity = used() / ((double) 1 - minimum_free_percentage);
  size_t maximum_desired_capacity = used() / ((double) 1 - maximum_free_percentage);
  if (maximum_desired_capacity < Universe::old_size()) {
    // Make sure we never shrink below initial size
    maximum_desired_capacity = Universe::old_size();
  }
 
  if (capacity() < minimum_desired_capacity) {
    // Capacity too small, compute expansion size
    size_t expand_bytes = minimum_desired_capacity - capacity();
    expand(expand_bytes, 0); // safe if expansion fails
  } else {
    size_t shrink_bytes = 0;
    // We would never want to shrink below this level
    size_t max_shrink_bytes = capacity() - minimum_desired_capacity;
    if (capacity() > maximum_desired_capacity) {
      // Capacity too large, compute shrinking size
      shrink_bytes = ((double)(capacity() - maximum_desired_capacity)) * current_shrink_factor / 100;
      assert(shrink_bytes >= 0, "just checking");
      if (current_shrink_factor == 0) {
        _shrink_factor = 10;
      } else {
      _shrink_factor = min(current_shrink_factor * 4, 100);
      }
    }
    if (_expansion_at_safepoint > 0) {
      // Adjust shrink_bytes if we temporarily expanded old generation at last scavenge(s)
      shrink_bytes = min(max(shrink_bytes, _expansion_at_safepoint), max_shrink_bytes);
    }
    assert(shrink_bytes >= 0 && shrink_bytes <= max_shrink_bytes, "invalid shrink size");
    if (shrink_bytes > 0) {
      shrink(shrink_bytes);
    }
  }

  assert(_shrink_factor >= 0 && _shrink_factor <= 100, "invalid shrink factor");
  assert(used()>=0 && used()<=capacity(), "sanity check");
  _expansion_at_safepoint = 0;
}


MarkSweepInvokeCause TenuredGeneration::gc_cause() {
  return _tenured_generation_full;
}


void PermGeneration::initialize(ReservedSpace rs, size_t initial_size) {
  if (!virtual_space.initialize(rs, initial_size)) {
    vm_exit_during_initialization("Could not reserve enough space for object heap");
  }
  low_boundary  = (oop*) virtual_space.low_boundary();
  high_boundary = (oop*) virtual_space.high_boundary();
  initialize_offset_array();
  
  oop* bottom = (oop*) virtual_space.low();
  oop* end    = (oop*) virtual_space.high();
  _the_space  = new PermSpace(_offset_array, bottom, end);
  _saved_mark = bottom_mark();
  _capacity_expansion_limit = capacity() + MaxPermHeapExpansion;
  _expansion_at_safepoint = 0;

  if (jvmpi::is_event_enabled(JVMPI_EVENT_ARENA_NEW)) {
    jvmpi::post_arena_new_event(Universe::addr_to_arena_id(bottom), name());
  }
}


const char* PermGeneration::name() {
  return "permanent generation";
}


oop* PermGeneration::allocate(int size, TRAPS) {
  oop* obj = _the_space->allocate(size);
  if (obj != NULL) {
    return obj;
  } else {
    if (capacity() >= _capacity_expansion_limit) {
      return allocate_and_collect(size, THREAD);
    } else {
      int desired_expand_size = align_size_up(size, MinPermHeapExpansion/wordSize);
      return allocate_and_expand(size, desired_expand_size, THREAD);
    }
  }
}


void PermGeneration::compute_new_generation_size() {
  size_t desired_capacity = align_size_up(used(), MinPermHeapExpansion);
  if (desired_capacity < Universe::perm_size()) {
    desired_capacity = Universe::perm_size();
  }
  ptrdiff_t shrink_bytes = capacity() - desired_capacity;
  if (shrink_bytes > 0) {
    shrink(shrink_bytes);
  }
  _capacity_expansion_limit = capacity() + MaxPermHeapExpansion;
  _expansion_at_safepoint = 0;
}


MarkSweepInvokeCause PermGeneration::gc_cause() {
  return _permanent_generation_full;
}


#define FOR_EACH_TRAIN(t) \
  for (Train* t = first_train(); t != NULL; t = t->next_train())


#define FOR_EACH_TRAIN_REVERSE(t) \
  for (Train* t = last_train(); t != NULL; t = t->prev_train())


void TrainGeneration::initialize(ReservedSpace rs) {
  if (!virtual_space.initialize(rs, 0)) {
    vm_exit_during_initialization("Could not reserve enough space for object heap");
  }
  low_boundary = (oop*) virtual_space.low_boundary();
  high_boundary= (oop*) virtual_space.high_boundary();
  initialize_offset_array();

  CarTable::initialize();

  // Clear free lists
  set_car_free_list(NULL);
  set_train_free_list(NULL);

  // Make initial train consisting of one initial car
  set_first_train(retrieve_from_train_pool(CarSpace::car_size_in_oops(), CarTableDesc::initial_train_number));
  set_last_train(first_train());
  set_first_car_desc();
  set_special_nonoop_train(NULL);
}


oop* TrainGeneration::allocate(int size, TRAPS) {
  oop* obj = last_train()->allocate_within(size);
  if (obj != NULL) {
    return obj;
  } else {
    return allocate_and_expand(size, THREAD);
  }
}


oop* TrainGeneration::safepoint_allocate(int size) {
  oop* obj = last_train()->safepoint_allocate_within(size);
  return (obj != NULL) ? obj : safepoint_allocate_and_expand(size);
}


oop* TrainGeneration::allocate_large_typearray(int size, TRAPS) {
  assert(size >= Universe::large_typearray_limit(), "should be large type array");
  if (!UseSpecialLargeObjectHandling) {
    return allocate(size, CHECK_0);
  }
  oop* obj = NULL;
  {
    // Grab ExpandHeap_lock since we are not at a safepoint, make sure to release it again before blocking
    MutexLocker ml(ExpandHeap_lock);
    if (_special_nonoop_train == NULL) {
      _special_nonoop_train = retrieve_from_train_pool(size, CarTableDesc::special_nonoop_train_number);
      if (TraceSpecialLargeObjectHandling && _special_nonoop_train != NULL) {
        tty->print_cr("Allocated large type array train");
      }
    }
    if (_special_nonoop_train != NULL) {
      obj = _special_nonoop_train->allocate(size);
    }
  }
  if (obj != NULL) {
    CarTableDesc* desc = CarTable::desc_for(obj);
    assert(desc->train_number() == CarTableDesc::special_nonoop_train_number, "just checking");
    if (TraceSpecialLargeObjectHandling) {
      tty->print("Allocated large type array 0x%x (%d bytes)", obj, size*wordSize);
      desc->space()->print_short();
      tty->cr();
    }
    return obj;
  } else {
    return allocate_and_expand(size, THREAD);
  }
}


oop* TrainGeneration::safepoint_allocate_large_typearray(int size) {
  assert(size >= Universe::large_typearray_limit(), "should be large type array");
  if (!UseSpecialLargeObjectHandling) {
    return safepoint_allocate(size);
  }
  if (_special_nonoop_train == NULL) {
    _special_nonoop_train = retrieve_from_train_pool(size, CarTableDesc::special_nonoop_train_number);
    if (TraceSpecialLargeObjectHandling && _special_nonoop_train != NULL) {
      tty->print_cr("Allocated large type array train");
    }
  }
  if (_special_nonoop_train != NULL) {
    oop* obj = _special_nonoop_train->safepoint_allocate(size);
    if (obj != NULL) {
      CarTableDesc* desc = CarTable::desc_for(obj);
      assert(desc->train_number() == CarTableDesc::special_nonoop_train_number, "just checking");
      ((CarSpace*) desc->space())->set_marked(true);  // mark explicitly during scavenge
      if (TraceSpecialLargeObjectHandling) {
        tty->print("Allocated large type array 0x%x (%d bytes)", obj, size*wordSize);
        desc->space()->print_short();
        tty->cr();
      }
      return obj;
    }
  }
  return safepoint_allocate_and_expand(size);
}


void TrainGeneration::set_first_car_desc() {
  _first_car_desc = first_train()->first_car()->desc();
}


Train* TrainGeneration::add_new_highest_train(int size) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  Train* old_last_train = last_train();
  Train* new_last_train = retrieve_from_train_pool(size, last_train()->train_number() + 1);
  if (new_last_train != NULL) {
    set_last_train(new_last_train);
    old_last_train->set_next_train(new_last_train);
    new_last_train->set_prev_train(old_last_train);
  }
  return new_last_train;
}


int TrainGeneration::release_train(Train* t) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  if (t == first_train()) {
    // releasing first train
    Train* new_first_train = t->next_train();
    if (new_first_train == NULL) {
      new_first_train = retrieve_from_train_pool(CarSpace::car_size_in_oops(), CarTableDesc::initial_train_number);
      set_last_train(new_first_train);
    }
    set_first_train(new_first_train);
    new_first_train->set_prev_train(NULL);
    set_first_car_desc();
  } else if (t == last_train()) {
    // releasing last train
    Train* new_last_train = t->prev_train();
    if (new_last_train == NULL) {
      new_last_train = retrieve_from_train_pool(CarSpace::car_size_in_oops(), CarTableDesc::initial_train_number);
      set_first_train(new_last_train);
      set_first_car_desc();
    }
    set_last_train(new_last_train);
    new_last_train->set_next_train(NULL);
  } else {
    // releasing interior train
    assert(t->next_train() != NULL && t->prev_train() != NULL, "just checking");
    t->next_train()->set_prev_train(t->prev_train());
    t->prev_train()->set_next_train(t->next_train());
  }
  return add_to_train_pool(t); // will release all cars as well
}


oop* TrainGeneration::allocate_and_expand_helper(int size, TRAPS) {
  // Grab ExpandHeap_lock since we are not at a safepoint. Make sure to release it again before blocking.
  MutexLocker ml(ExpandHeap_lock);
  // We have the lock, try allocating again in case someone expanded before us
  Train* t = last_train();
  oop* obj = t->allocate_within(size);
  if (obj == NULL) {
    if (t->length() >= MinTrainLength) t = add_new_highest_train(size);
    if (t != NULL) obj = t->allocate(size);
  }
  return obj;
}


oop* TrainGeneration::allocate_and_expand(int size, TRAPS) {
  assert(!Scavenge::is_active(), "just checking");
  oop* obj = allocate_and_expand_helper(size, CHECK_0);
  if (obj == NULL) {
    // Do full GC and retry allocation
    MarkSweep::invoke(size, _train_generation_full, CHECK_0);
    obj = allocate_and_expand_helper(size, CHECK_0);
    if (obj == NULL) {
      THROW_OOP_0(Universe::out_of_memory_error_instance());
    }
  }
  return obj;
}


oop* TrainGeneration::safepoint_allocate_and_expand(int size) {
  assert(SafepointSynchronize::is_at_safepoint(), "should be at safepoint");
  Train* t = last_train();
  if (t->length() >= MinTrainLength) {
    t = add_new_highest_train(size);
  }
  return t == NULL ? NULL : t->safepoint_allocate(size);
}


int TrainGeneration::release_lowest_car() {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  CarSpace* first_car = first_train()->first_car();
  CarSpace* next_car = first_car->next_car();
  if (next_car == NULL) {
    return release_lowest_train();
  } else {
    int used = first_car->used();
    first_train()->set_first_car(next_car);
    add_to_car_pool(first_car, true);
    set_first_car_desc();
    return used;
  }
}


void TrainGeneration::coalesce_cars(CarSpace* c1, CarSpace* c2) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  assert(c1 && c2 && c1->next_car() == c2, "just checking");
  assert(c1->bottom() < c2->bottom(), "just checking");
  // merge c1 and c2 if they are adjacent
  if (c1->end() == c2->bottom()) {
    c1->set_next_car(c2->next_car());
    c1->set_blocks(c1->blocks() + c2->blocks());
    c1->set_end(c2->end());
    delete c2;
  }
}


CarSpace* TrainGeneration::split_car(CarSpace* c, int blocks) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  assert(c->blocks() > blocks, "just checking");
  // split first part of c into new car
  int new_blocks = c->blocks() - blocks;
  oop* old_bottom = c->bottom();
  oop* new_bottom = old_bottom + (CarSpace::car_size_in_oops() * blocks);
  u_char* old_offset_array = c->_offset_array;
  int index = (new_bottom - low_boundary) / RememberedSet::card_size_in_oops;
  u_char* new_offset_array = &_offset_array[index];
  // update c but leave it in free list
  c->commit();
  c->set_blocks(new_blocks);
  c->set_bottom(new_bottom);
  c->set_top(new_bottom);
  c->_offset_array = new_offset_array;
  c->initialize_threshold();
  c->set_desc(CarTable::desc_for(new_bottom));
  c->uncommit();
  // create new car
  return new CarSpace(old_offset_array, old_bottom, new_bottom, blocks);
}


void TrainGeneration::add_to_car_pool(CarSpace* c, bool remset_already_empty) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  c->clear();
  c->uncommit();
  CarTable::clear_entry(c);
  // reset remembered set, should already be empty
  c->remembered_set()->reinitialize(remset_already_empty);
  // First check free list is null
  CarSpace* p = car_free_list();
  if (p == NULL) {
    c->set_next_car(NULL);
    set_car_free_list(c);
    return;
  }
  // First check if c should go before p
  if (c->bottom() < p->bottom()) {
    c->set_next_car(p);
    set_car_free_list(c);
    coalesce_cars(c, p);
    return;
  } 
  // Iterate until c should be inserted after p
  while (p->next_car() != NULL && p->next_car()->bottom() < c->bottom()) {
    p = p->next_car();
  }
  c->set_next_car(p->next_car());
  p->set_next_car(c);
  if (c->next_car() != NULL) {
    coalesce_cars(c, c->next_car());
  }
  coalesce_cars(p, c);
}


CarSpace* TrainGeneration::retrieve_from_car_pool(int size, julong train_number, juint car_number, Train* train) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  assert(size>0, "Invalid size");
  int blocks = ((size - 1) / CarSpace::car_size_in_oops()) + 1;
  // Look for best fit in free list
  CarSpace* c = car_free_list();
  CarSpace* prev_c = NULL;
  CarSpace* best = NULL;
  CarSpace* prev_best = NULL;
  int best_blocks = Universe::reserved_old_size();    // just a big number
  while (c != NULL) {
    if (c->blocks() >= blocks && c->blocks() < best_blocks) {
      best_blocks = c->blocks();
      best = c;
      prev_best = prev_c;
    }
    prev_c = c;
    c = c->next_car();
  }
  // Check if we found existing usable car
  if (best != NULL) {
    if (best_blocks == blocks) {
      // best is exact fit, take it out of free list
      if (prev_best != NULL) {
        prev_best->set_next_car(best->next_car());
      } else {
        set_car_free_list(best->next_car());
      }
      best->set_next_car(NULL);
    } else {
      // best is to large, split it and leave remaining part in free list
      best = split_car(best, blocks);
    }
  } else {
    // No usable car found, expand train heap
    char* old_high = virtual_space.high();
    assert((int) old_high % CarSpace::car_size() == 0, "CarSpace not aligned");
    assert(prev_c == NULL || prev_c->next_car() == NULL, "prev_c should be last in free list");
    int expand_blocks = blocks;
    if (prev_c != NULL && prev_c->end() == (oop*) old_high) {
      // Last car in free list is adjacent to end of heap, extend it
      expand_blocks -= prev_c->blocks();
    }
    if (!virtual_space.expand_by(expand_blocks * CarSpace::car_size())) {
      if (Scavenge::is_active()) {
	// Before proceeding with scavenge we verified that worst case promotion would fit,
	// so the expansion should have succeeded. Do better than this for Merlin
	vm_exit_out_of_memory(expand_blocks * CarSpace::car_size(), "heap expansion");
      } else {
	// We are not scavenging, the allocation request might be really huge and larger
	// than the uncommitted area. Returning NULL is safe, the caller will throw
	// OutOfMemoryError if needed.
	return NULL;
      }
    }
    // expand offset array to cover virtual space expansion
    grow_offset_array();
    if (Universe::is_fully_initialized()) {
      // expand card marking array to cover virtual space expansion (unless bootstrapping)
      grow_card_marking_array();
    }
    if (expand_blocks == blocks) {
      // we have to create entire new car
      int index = ((oop*) old_high - low_boundary) / RememberedSet::card_size_in_oops;
      best = new CarSpace(&_offset_array[index], (oop*) old_high, (oop*) virtual_space.high(), expand_blocks);
    } else {
      // extend prev_c to cover expanded heap area
      prev_c->set_blocks(blocks);
      prev_c->set_end((oop*) virtual_space.high());
      // retry, free list allocation will succeed with best fit now
      return retrieve_from_car_pool(size, train_number, car_number, train);
    }
  }
  // best contains car
  CarTable::update_entry(best, train_number, car_number, train);
  if (TraceCarAllocation) {
    tty->print("Assigned");
    best->print();
  }
  RememberedSet::clear_for_space(best);
  best->commit();
  return best;
}


Train* TrainGeneration::retrieve_from_train_pool(int size, julong train_number) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  Train *t;
  if (train_free_list() == NULL) {
    t = new Train();
  } else {
    t = train_free_list();
    set_train_free_list(t->next_train());
  }
  CarSpace* c = retrieve_from_car_pool(size, train_number, CarTableDesc::initial_car_number, t);
  if (c == NULL) {
    t->set_first_car(NULL);
    add_to_train_pool(t);
    return NULL;
  } else {
    t->initialize(c);
    return t;
  }
}


int TrainGeneration::add_to_train_pool(Train* t) {
  assert_lock(ExpandHeap_lock); // holds lock or is at safepoint
  // Release all cars in train
  int used = 0;
  CarSpace* c = t->first_car();
  while (c != NULL) {
    used += c->used();
    CarSpace* next_c = c->next_car();
    add_to_car_pool(c, false);
    c = next_c;
  }
  // Add to train free list (threaded through next_train fields)
  if (train_free_list() == NULL) {
    set_train_free_list(t);
    t->set_next_train(NULL);
  } else {
    t->set_next_train(train_free_list());
    set_train_free_list(t);
  }
  t->set_prev_train(NULL);
  return used;
}


void TrainGeneration::scavenge_recorded_stores(bool include_first) {
  first_train()->scavenge_recorded_stores(include_first);
  Train* t = first_train()->next_train();
  while (t != NULL) {
    t->scavenge_recorded_stores(true);
    t = t->next_train();
  }
}


bool TrainGeneration::scavenge_contents_from_saved_marks() {
  // Skip _special_nonoop_train, contains no fan-out
  bool done = false;
  bool work = false;
  while (!done) {
    done = true;
    FOR_EACH_TRAIN_REVERSE(t) {
      if (t->scavenge_contents_from_saved_mark()) {   // returns true if any work was done
        done = false;
        work = true;
      }
    }
  }
  return work;
}


void TrainGeneration::scavenge_higher_train_recorded_stores() {
  first_train()->first_car()->remembered_set()->scavenge_higher_train_recorded_stores();
}


void TrainGeneration::scavenge_same_train_recorded_stores() {
  first_train()->first_car()->remembered_set()->scavenge_same_train_recorded_stores();
}


void TrainGeneration::scavenge_special_nonoop_train_recorded_stores(bool exclude_lowest_car, bool exclude_lowest_train) {
  if (!UseSpecialLargeObjectHandling) return;
  if (_special_nonoop_train != NULL) {
    CarSpace* c = _special_nonoop_train->first_car();
    while (c != NULL) {
      c->remembered_set()->scavenge_special_nonoop_recorded_stores(exclude_lowest_car, exclude_lowest_train);
      c = c->next_car();
    }
  }
}


void TrainGeneration::release_special_nonoop_train_contents() {
  if (!UseSpecialLargeObjectHandling) return;
  if (_special_nonoop_train != NULL) {
    CarSpace* prev = NULL;
    CarSpace* current = _special_nonoop_train->first_car();
    CarSpace* next = NULL;
    while (current != NULL) {
      next = current->next_car();
      if (current->marked()) {
        current->set_marked(false);
        prev = current;
      } else {
        if (prev == NULL) {
          _special_nonoop_train->set_first_car(next);
        } else {
          prev->set_next_car(next);
        }
        if (TraceSpecialLargeObjectHandling) {
          tty->print("Reclaimed large type array 0x%x (%d bytes)", current->bottom(), current->used());
          current->print_short();
          tty->cr();
        }
        add_to_car_pool(current, true);
      }
      current = next;
    }
    if (_special_nonoop_train->first_car() == NULL) {
      add_to_train_pool(_special_nonoop_train);
      _special_nonoop_train = NULL;
      if (TraceSpecialLargeObjectHandling) {
        tty->print_cr("Reclaimed large type array train");
      }
    } else {
       _special_nonoop_train->set_last_car(prev);
    }
  }
}


void TrainGeneration::tail_trains_set_should_scavenge(bool value) {
  Train* t = first_train()->next_train();
  while (t != NULL) {
    t->set_should_scavenge(value);
    t = t->next_train();
  }
}


void TrainGeneration::special_nonoop_train_set_should_scavenge(bool value) {
  if (_special_nonoop_train != NULL) {
    _special_nonoop_train->set_should_scavenge(value);
  }
}


void TrainGeneration::clear_remembered_sets() {
  if (special_nonoop_train() != NULL) {
    special_nonoop_train()->clear_remembered_sets();
  }
  FOR_EACH_TRAIN(t) {
    t->clear_remembered_sets();
  }
}


void TrainGeneration::full_gc_prologue() {
  // Collapse all consecutive single-car trains to a single train 
  // in order to improve compaction.
  Train* t = first_train(); 
  while (t != NULL) {
    Train* next_t = t->next_train();
    if (t->has_single_car()) {
      while (next_t != NULL && next_t->has_single_car()) {
        CarSpace* c = next_t->first_car();
        juint last_car_number = t->last_car()->car_number();
        t->last_car()->set_next_car(c);
        t->set_last_car(c);
        next_t->set_first_car(NULL);
        CarTable::update_entry(c, t->train_number(), last_car_number+1, t);
        release_train(next_t);
        next_t = t->next_train();
      }
    }
    t = next_t;
  }
}


void TrainGeneration::full_gc_epilogue() {
  if (_special_nonoop_train != NULL) {
    if (_special_nonoop_train->full_gc_epilogue()) {
      add_to_train_pool(_special_nonoop_train);
      _special_nonoop_train = NULL;
      if (TraceSpecialLargeObjectHandling) {
        tty->print_cr("Reclaimed large type array train");
      }
    }
  }
  Train* t = first_train(); 
  while (t != NULL) {
    Train* next_t = t->next_train();
    if (t->full_gc_epilogue()) {
      release_train(t);
    }
    t = next_t;
  }
  set_first_car_desc();
}


void TrainGeneration::object_iterate(ObjectClosure* blk) {
  if (_special_nonoop_train != NULL) {
    _special_nonoop_train->object_iterate(blk);
  }
  FOR_EACH_TRAIN(t) {
    t->object_iterate(blk);
  }
}


void TrainGeneration::object_iterate_from_saved_marks(ObjectClosure* blk) {
  if (_special_nonoop_train != NULL) {
    _special_nonoop_train->object_iterate_from_saved_mark(blk);
  }
  FOR_EACH_TRAIN(t) {
    t->object_iterate_from_saved_mark(blk);
  }
}


void TrainGeneration::space_iterate(SpaceClosure* blk) {
  if (_special_nonoop_train != NULL) {
    _special_nonoop_train->space_iterate(blk);
  }
  FOR_EACH_TRAIN(t) {
    t->space_iterate(blk);
  }
}


void TrainGeneration::set_saved_marks_to_top() {
  if (_special_nonoop_train != NULL) {
    _special_nonoop_train->set_saved_mark_to_top();
  }
  FOR_EACH_TRAIN(t) {
    t->set_saved_mark_to_top();
  }
}


void TrainGeneration::set_saved_marks_to_bottom() {
  if (_special_nonoop_train != NULL) {
    _special_nonoop_train->set_saved_mark_to_bottom();
  }
  FOR_EACH_TRAIN(t) {
    t->set_saved_mark_to_bottom();
  }
}


bool TrainGeneration::top_at_saved_marks() {
  // Skip _special_nonoop_train, it is never scanned in scavenge_contents_from_saved_marks
  FOR_EACH_TRAIN(t) {
    if (!t->top_at_saved_mark()) return false;
  }
  return true;
}


size_t TrainGeneration::free() {
  size_t size = 0;
  if (_special_nonoop_train != NULL) {
    size += _special_nonoop_train->free();
  }
  FOR_EACH_TRAIN(t) {
    size += t->free();
  }
  return size;
}


size_t TrainGeneration::used() {
  size_t size = 0;
  if (_special_nonoop_train != NULL) {
    size += _special_nonoop_train->used();
  }
  FOR_EACH_TRAIN(t) {
    size += t->used();
  }
  return size;
}


size_t TrainGeneration::capacity() {
  size_t size = 0;
  if (_special_nonoop_train != NULL) {
    size += _special_nonoop_train->capacity();
  }
  FOR_EACH_TRAIN(t) {
    size += t->capacity();
  }
  return size;
}


size_t TrainGeneration::contiguous_available() {
  // compute maximum available space in free list
  size_t max_free_list_capacity = 0;
  CarSpace* c = car_free_list();
  CarSpace* last = NULL;
  while (c != NULL) {
    size_t size = c->capacity();
    if (size > max_free_list_capacity) {
      max_free_list_capacity = size;
    }
    last = c;
    c = c->next_car();
  }
  // compute available uncommitted space
  size_t uncommitted_capacity = virtual_space.uncommitted_size();
  if (last != NULL && last->end() == (oop*) virtual_space.high()) {
    // check for the case where the last free list entry is adjacent to the uncommitted part
    uncommitted_capacity += last->capacity();
  }
  return max(max_free_list_capacity, uncommitted_capacity);
}


void TrainGeneration::update_offset_array() {
  if (_special_nonoop_train != NULL) {
    _special_nonoop_train->update_offset_array();
  }
  FOR_EACH_TRAIN(t) {
    t->update_offset_array();
  }
}


oop TrainGeneration::fast_object_start(oop* p) { 
  CarSpace* car = (CarSpace*) CarTable::space_for(oop(p));
  return car->fast_object_start(p); 
}


void TrainGeneration::compute_new_generation_size() {
  // do nothing for now
}


void TrainGeneration::prepare_for_compaction() {
  if (_special_nonoop_train != NULL) {
    _special_nonoop_train->first_car()->prepare_for_compaction();
  }
  FOR_EACH_TRAIN(t) {
    t->prepare_for_compaction();
  }
}


void TrainGeneration::adjust_pointers() {
  if (_special_nonoop_train != NULL) {
    _special_nonoop_train->first_car()->adjust_pointers();
  }
  FOR_EACH_TRAIN(t) {
    t->adjust_pointers();
  }
}


void TrainGeneration::compact() {
  if (_special_nonoop_train != NULL) {
    _special_nonoop_train->first_car()->compact();
  }
  FOR_EACH_TRAIN(t) {
    t->compact();
  }
}


const char* TrainGeneration::name() {
  return "train generation";
}


#ifndef PRODUCT

void TrainGeneration::print() {
  Generation::print();
  if (Verbose) {
    if (special_nonoop_train() != NULL) {
      special_nonoop_train()->print();
    }
    FOR_EACH_TRAIN(t) {
      t->print();
    }
  }
  if (Verbose && WizardMode) {
    tty->print_cr("Train generation car free list");
    for (CarSpace* c = car_free_list(); c != NULL; c = c->next_car()) {
      c->print();
    }
  }
}


void TrainGeneration::verify(bool allow_dirty) {
  Train* t = first_train();
  guarantee(t != NULL, "No first train in train_gen");
  guarantee(t->first_car()->desc() == first_car_desc(), "Invalid first car desc");
  while (t != NULL) {
    julong tn = t->train_number();
    guarantee(tn > 0, "Illegal train number");
    Train* next_t = t->next_train();
    if (next_t == NULL) {
      guarantee(t == last_train(), "Wrong last train in train_gen");
    } else {
      guarantee(tn < next_t->train_number(), "Train numbers in train_gen not ascending");
    }
    t->verify(allow_dirty);
    t = next_t;
  }
  if (special_nonoop_train() != NULL) {
    assert(UseSpecialLargeObjectHandling, "just checking");
    special_nonoop_train()->verify(true);
    julong tn = special_nonoop_train()->train_number();
    guarantee(tn == CarTableDesc::special_nonoop_train_number, "Illegal train number");
    CarSpace* c = special_nonoop_train()->first_car();
    guarantee(c != NULL, "First car should be present");
    while (c != NULL) {
      oop obj = (oop) (c->bottom());
      guarantee(obj->is_typeArray(), "Non-typearray found in special large typearray train");
      guarantee(obj->size() >= Universe::large_typearray_limit(), "Small typearray found in special large typearray train");
      c = c->next_car();
    }
  }
  CarSpace* c = car_free_list();
  while (c != NULL) {
    CarSpace* next_c = c->next_car();
    if (next_c != NULL) {
      guarantee(c->end() < next_c->bottom(), "free list should be in address order and coalesced")
    }
    c = next_c;
  }
}


class LowestCarRefOopClosure: public OopClosure {
 public:
  int cards;
  int refs;
  jbyte* last_card;
  CarSpace* lowest_car;

  void do_oop(oop* from) {
    assert(oop(from)->is_old_or_perm() && !lowest_car->contains(from), "just checking");
    oop to = *from;
    if (to != NULL && lowest_car->contains(to)) {
      refs++;
      jbyte* from_card = RememberedSet::byte_for(from);
      if (from_card != last_card) {
        cards++;
        last_card = from_card;
      }
    }
  }

  LowestCarRefOopClosure() {
    cards = 0;
    refs = 0;
    last_card = NULL;
  }
};

class LowestCarRefObjectClosure: public ObjectClosure {
 public:
  LowestCarRefOopClosure* oop_blk;

  void do_object(oop obj) {
    obj->oop_iterate(oop_blk);
  }
};

class LowestCarRefSpaceClosure: public SpaceClosure {
 public:
  LowestCarRefObjectClosure* obj_blk;
  CarSpace* lowest_car;

  void do_space(Space* s) {
    if (s->is_old() && s != lowest_car) {
      s->object_iterate(obj_blk);
    }
  }
};


void TrainGeneration::compute_references_into_lowest_car(int* nof_cards, int* nof_refs) {
  LowestCarRefOopClosure oop_blk;
  oop_blk.lowest_car = first_train()->first_car();
  LowestCarRefObjectClosure obj_blk;
  obj_blk.oop_blk = &oop_blk;
  LowestCarRefSpaceClosure space_blk;
  space_blk.lowest_car = first_train()->first_car();
  space_blk.obj_blk = &obj_blk;
  Universe::space_iterate(&space_blk);
  *nof_cards = oop_blk.cards;
  *nof_refs = oop_blk.refs;
}


#endif // PRODUCT


