/*
 * @(#)generation.hpp	1.78 01/02/09 14:45:01
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// A Generation models a heap area for similarly-aged objects.
// It will contain one ore more spaces holding the actual objects.
//
// The Generation class hierarchy:
//
// Generation                      - abstract base class
// - NewGeneration                 - allocation area (copy collected)
// - OldGeneration                 - abstract class adding offset array behavior
//   - OneSpaceOldGeneration       - abstract class holding a single space (mark-compact collected)
//     - TenuredGeneration         - tenured object area when not using train gc
//     - PermGeneration            - reflective object area (klasses, methods, symbols, ...)
//   - TrainGeneration             - tenured object area when using train gc (train copy collected)
//
// The system configurations currently used are:
//
//   NewGeneration + TenuredGeneration + PermGeneration
//   NewGeneration + TrainGeneration + PermGeneration


class Generation: public CHeapObj {
  friend class RememberedSet;
  friend class oopDesc;
  friend class Universe;
  friend class CarTable;
  friend class VMStructs;
 private:
  // Tracing of maximum generation capaticy during lifetime of VM
  NOT_PRODUCT(size_t _max_capacity;)
  void update_max_capacity() PRODUCT_RETURN;
 protected:
  // Minimum and maximum addresses for memory reserved (not necessarily committed) for generation.
  // Used by card marking code. Must not overlap with address ranges of other generations.
  oop* low_boundary;
  oop* high_boundary;
  // Memory area reserved for generation 
  VirtualSpace virtual_space;

 public:
  // Space enquiries (results in bytes)
  virtual size_t capacity() = 0;
  virtual size_t used() = 0;
  virtual size_t free() = 0;
 
  // Iteration - do not use for time critical operations
  virtual void object_iterate(ObjectClosure* blk) = 0;
  virtual void space_iterate(SpaceClosure* blk) = 0;

  // Scavenge support
  virtual bool scavenge_contents_from_saved_marks() = 0;

  // Accessing marks
  virtual void set_saved_marks_to_top() = 0;
  virtual void set_saved_marks_to_bottom() = 0;
  virtual bool top_at_saved_marks() = 0;

  // Constructor
  Generation();

  // GC support
  virtual void compute_new_generation_size() = 0;

  // Printing
  virtual const char* name() = 0;

  // Debugging - do not use for time critical operations
#ifndef PRODUCT
  bool contains(void* p);
  oop slow_object_start(oop* p);

  virtual void print();
  void print_max_capacity() PRODUCT_RETURN;
  virtual void verify(bool allow_dirty) = 0;
  address reserved_high() { return (address)high_boundary; }
  address reserved_low()  { return (address) low_boundary; }
#endif
};


// NewGeneration models the young generation containing eden, from- and to-space

class NewGeneration: public Generation {
  friend class RememberedSet;
  friend class Scavenge;
  friend class MarkSweep;
  friend class Universe;
  friend class VMStructs;

 protected:
  // Spaces
  EdenSpace      eden_space;
  SurvivorSpace* from_space;
  SurvivorSpace* to_space;
  WaterMark      _saved_mark;     // watermark used during scavenge and allocation profiling

 public:
  // Accessing spaces
  EdenSpace*      eden()            { return &eden_space; }
  SurvivorSpace*  from()            { return from_space;  }
  SurvivorSpace*  to()              { return to_space;    }

  // Space enquiries
  size_t capacity();
  size_t used();
  size_t free();

  // Iteration
  void object_iterate(ObjectClosure* blk);
  void space_iterate(SpaceClosure* blk);

  // Accessing marks
  void set_saved_marks_to_top();
  void set_saved_marks_to_bottom();
  bool top_at_saved_marks();

  // Printing
  const char* name();

  // GC support
  void compute_new_generation_size();

  // Debugging
  void print();
  void verify(bool allow_dirty);

 protected:
  // Initialization, called by Universe
  void initialize(ReservedSpace rs, size_t initial_size);
  void compute_space_boundaries(size_t minimum_eden_size);
  // Scavenge support
  void swap_spaces();
  bool scavenge_contents_from_saved_marks();
  // Mark sweep support phase2
  void prepare_for_compaction();
  // Mark sweep support phase3
  void adjust_pointers();
  // Mark sweep support phase4
  void compact();
};



// Class OldGeneration is an abstract base class for old (train, perm) generations.

class OldGeneration: public Generation {
  friend class RememberedSet;
  friend class VMStructs;
 protected:
  // Portion of card marking array mapped by this generation
  VirtualSpace _card_marking_space;
  void initialize_card_marking_array(ReservedSpace rs);
  void grow_card_marking_array();     // grow array after growing generation
  void shrink_card_marking_array();   // shrink array after shrinking generation
  void clear_card_marking_array();

  // Array for keeping offsets for retrieving object start given a card_page
  VirtualSpace _offset_space;   // byte array keeping backwards offsets
  u_char* _offset_array;        // start of offset_space

  void initialize_offset_array();
  void grow_offset_array();     // grow array after growing generation
  void shrink_offset_array();   // shrink array after shrinking generation

 public:
  // Allocation
  virtual oop* allocate(int size, TRAPS) = 0;
  virtual oop* allocate_large_typearray(int size, TRAPS) = 0;
  // Allocation at safepoint, not MT safe
  virtual oop* safepoint_allocate(int size) = 0;
  virtual oop* safepoint_allocate_large_typearray(int size) = 0;
  // Iteration
  virtual void object_iterate_from_saved_marks(ObjectClosure* blk) = 0;
  // Offset array
  virtual void update_offset_array() = 0;
  virtual oop fast_object_start(oop* p) = 0;
  // Space enquiries (results in bytes)
  virtual size_t contiguous_available() = 0;
};


// OneSpaceOldGeneration models a heap of old objects contained in a single
// contiguous space.
//
// Garbage collection is performed using mark-compact.

class OneSpaceOldGeneration: public OldGeneration {
  friend class VMStructs;
 protected:
  OldSpace*  _the_space;      // actual space holding objects
  WaterMark  _saved_mark;     // watermark used during scavenge and allocation profiling
  size_t     _expansion_at_safepoint; // accumulated heap expansion while at safepoint

  // Grow generation with specified size (returns false if unable to grow)
  bool grow_by(size_t bytes);
  // Grow generation to reserved size (returns true if already expanded fully)
  bool grow_to_reserved();
  // Shrink generation with specified size (returns false if unable to shrink)
  void shrink_by(size_t bytes);

  // Allocation failure
  bool expand(size_t bytes, size_t expand_bytes);
  void shrink(size_t bytes);
  oop* allocate_and_collect(int size, TRAPS);
  oop* allocate_and_expand(int size, int expand_size, TRAPS);
  oop* safepoint_allocate_and_expand(int size);

 public:
  // Accessing spaces
  OldSpace* the_space()                 { return _the_space; }
 
  // Allocation
  oop* allocate_large_typearray(int size, TRAPS);

  // Allocation at safepoint, not MT safe (return NULL if unable to expand)
  oop* safepoint_allocate(int size);
  oop* safepoint_allocate_large_typearray(int size);

  // Space enquiries
  size_t capacity();
  size_t used();
  size_t free();
  size_t contiguous_available();

  // Offset array
  void update_offset_array();
  oop fast_object_start(oop* p);

  // Accessing marks
  WaterMark top_mark()                  { return _the_space->top_mark(); }
  WaterMark bottom_mark()               { return _the_space->bottom_mark(); }
  void set_saved_marks_to_top();
  void set_saved_marks_to_bottom();
  bool top_at_saved_marks();

  // Iteration
  void object_iterate(ObjectClosure* blk);
  void object_iterate_from_saved_marks(ObjectClosure* blk);
  void space_iterate(SpaceClosure* blk);

  // Debugging
  void verify(bool allow_dirty);
  void print();

  // Scavenge support
  bool scavenge_contents_from_saved_marks();
  void scavenge_recorded_stores();            // Scavenge using remembered set

  // Mark sweep support
  void prepare_for_compaction();
  void adjust_pointers();
  void compact();
  virtual MarkSweepInvokeCause gc_cause() = 0;
};



// TenuredGeneration models the heap containing old (promoted/tenured) objects.

class TenuredGeneration: public OneSpaceOldGeneration {
  friend class VMStructs;
 protected:
  int _shrink_factor;  // current shrinking effect
 public:
  // Initialization, called by Universe
  void initialize(ReservedSpace rs, size_t initial_size);
  
  // Allocation
  oop* allocate(int size, TRAPS);

  // Printing
  const char* name();
  
  // Mark sweep support
  void compute_new_generation_size();
  MarkSweepInvokeCause gc_cause();
};


// PermGeneration models the heap containing permanent (reflective) objects.

class PermGeneration: public OneSpaceOldGeneration {
  friend class VMStructs;
 protected:
  size_t _capacity_expansion_limit;  // maximum expansion allowed without a full gc occuring
 public:
  // Initialization, called by Universe
  void initialize(ReservedSpace rs, size_t initial_size);
  
  // Allocation
  oop* allocate(int size, TRAPS);

  // Printing
  const char* name();

  // Mark sweep support
  void compute_new_generation_size();
  MarkSweepInvokeCause gc_cause();
};


// TrainGeneration models the (optional) heap containing old objects.
//
// Garbage collection is performed using the train algorithm, or alternatively, 
// a full mark-compact.

class TrainGeneration: public OldGeneration {
  friend class CarSpace;
  friend class Train;
  friend class MarkSweep;
  friend class Scavenge;
  friend class Universe;
  friend class VMStructs;
 protected:
  Train* _first_train;	                // The oldest train
  Train* _last_train;                   // The youngest train
  Train* _special_nonoop_train;         // Special train holding nonoop data (typically large arrays)

  CarTableDesc* _first_car_desc;        // Table entry for lowest car
                                        
  CarSpace* _car_free_list;	        // Free pool of cars
  Train*    _train_free_list;           // Free pool of trains
                                        
  void set_first_train(Train* t)        { _first_train = t; }
  void set_last_train(Train* t)         { _last_train = t; }
  void set_special_nonoop_train(Train* t) { _special_nonoop_train = t; }

  CarSpace* car_free_list() const       { return _car_free_list; }
  void set_car_free_list(CarSpace* c)   { _car_free_list = c; }

  Train* train_free_list() const        { return _train_free_list; }
  void set_train_free_list(Train* t)    { _train_free_list = t; }

  oop* allocate_and_expand(int size, TRAPS);
  oop* allocate_and_expand_helper(int size, TRAPS);
  oop* safepoint_allocate_and_expand(int size);

  void coalesce_cars(CarSpace* c1, CarSpace* c2);
  CarSpace* split_car(CarSpace* c, int blocks);

  int release_train(Train* t);

 public:
  // Accessors
  Train* first_train() const            { return _first_train; }
  Train* last_train() const             { return _last_train; }	
  Train* special_nonoop_train() const   { return _special_nonoop_train; }	

  // Cached block table descriptor of car to be processed
  CarTableDesc* first_car_desc() const  { return _first_car_desc; }
  void set_first_car_desc();

  // Add new highest numbered train. First car must be large enough to hold 'size' oops.
  Train* add_new_highest_train(int size);

  // Release lowest numbered train.
  int release_lowest_train()           { return release_train(first_train()); }
  
  // Release lowest numbered car.
  int release_lowest_car();

  // Allocate oop of given size highest train
  oop* allocate(int size, TRAPS);
  oop* allocate_large_typearray(int size, TRAPS);
  // Allocation at safepoint, not MT safe (return NULL if unable to expand)
  oop* safepoint_allocate(int size);
  oop* safepoint_allocate_large_typearray(int size);

  // Remember current top marks for all trains
  void set_saved_marks_to_top();
  void set_saved_marks_to_bottom();
  bool top_at_saved_marks();

  void scavenge_recorded_stores(bool include_first);
  void scavenge_higher_train_recorded_stores();
  void scavenge_same_train_recorded_stores();
  void tail_trains_set_should_scavenge(bool value);
  void special_nonoop_train_set_should_scavenge(bool value);
  void scavenge_special_nonoop_train_recorded_stores(bool exclude_lowest_car, bool exclude_lowest_train);
  void release_special_nonoop_train_contents();

  // Scavenge from saved marks to top marks and update saved marks. Returns true if any work was done.
  bool scavenge_contents_from_saved_marks();
  void clear_remembered_sets();

  // Iteration
  void object_iterate(ObjectClosure* blk);
  void object_iterate_from_saved_marks(ObjectClosure* blk);
  void space_iterate(SpaceClosure* blk);

  void initialize(ReservedSpace rs);

  // Space enquiries
  size_t capacity();
  size_t used();
  size_t free();
  size_t contiguous_available();

  // Offset array support
  void update_offset_array();
  oop fast_object_start(oop* p);

  // Pool of unused trains
  Train* retrieve_from_train_pool(int size, julong train_number);
  int add_to_train_pool(Train* t);

  // Pool of unused cars
  CarSpace* retrieve_from_car_pool(int size, julong train_number, juint car_number, Train* train);
  void add_to_car_pool(CarSpace* c, bool remset_already_empty);

  // Mark sweep support
  void compute_new_generation_size();
  void full_gc_prologue();
  void full_gc_epilogue();
  void prepare_for_compaction();
  void adjust_pointers();
  void compact();

  // Printing
  const char* name();

  // Debugging
  void verify(bool allow_dirty)   PRODUCT_RETURN;
  void print()                    PRODUCT_RETURN;
  void compute_references_into_lowest_car(int* nof_cards, int* nof_refs) PRODUCT_RETURN;
};
