/*
 * @(#)heap.cpp	1.31 01/02/09 14:45:02
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_heap.cpp.incl"


int CodeHeap::header_size() {
  return sizeof(HeapBlock);
}


// Implementation of Heap

CodeHeap::CodeHeap() {
  _number_of_committed_segments = 0;
  _number_of_reserved_segments  = 0;
  _segment_size                 = 0;
  _log2_segment_size            = 0;
  _next_segment                 = 0;
  _freelist                     = NULL;
  _free_segments                = 0;
}


void CodeHeap::mark_segmap_as_free(int beg, int end) {
  assert(0   <= beg && beg <  _number_of_committed_segments, "interval begin out of bounds");
  assert(beg <  end && end <= _number_of_committed_segments, "interval end   out of bounds");
  // setup _segmap pointers for faster indexing
  address p = (address)_segmap.low() + beg;
  address q = (address)_segmap.low() + end;
  // initialize interval
  while (p < q) *p++ = 0xFF;
}


void CodeHeap::mark_segmap_as_used(int beg, int end) {
  assert(0   <= beg && beg <  _number_of_committed_segments, "interval begin out of bounds");
  assert(beg <  end && end <= _number_of_committed_segments, "interval end   out of bounds");
  // setup _segmap pointers for faster indexing
  address p = (address)_segmap.low() + beg;
  address q = (address)_segmap.low() + end;
  // initialize interval
  int i = 0;
  while (p < q) {
    *p++ = i++;
    if (i == 0xFF) i = 1;
  }
}


static int align_to_page_size(int size) {
  const int alignment = os::vm_page_size();
  assert(is_power_of_2(alignment), "no kidding ???");
  return (size + alignment - 1) & ~(alignment - 1);
}


bool CodeHeap::reserve(int reserved_size, int committed_size, int segment_size) {
  assert(segment_size >= sizeof(FreeBlock), "segment size is too small");
  assert(is_power_of_2(segment_size), "segment_size must be a power of 2");
  _segment_size      = segment_size;
  _log2_segment_size = exact_log2(segment_size);
  // reserve space for _memory
  assert(reserved_size >= committed_size, "reserved size must be >= committed size");
  if (!_memory.initialize(align_to_page_size(reserved_size), align_to_page_size(committed_size))) {
    return false;
  }
  _number_of_committed_segments = number_of_segments(_memory.committed_size());
  _number_of_reserved_segments  = number_of_segments(_memory.reserved_size());
  assert(_number_of_reserved_segments >= _number_of_committed_segments, "just checking");
  // reserve space for _segmap
  if (!_segmap.initialize(align_to_page_size(_number_of_reserved_segments), align_to_page_size(_number_of_committed_segments))) {
    return false;  
  }
  assert(_segmap.committed_size() >= (size_t) _number_of_committed_segments, "could not commit  enough space for segment map");
  assert(_segmap.reserved_size()  >= (size_t) _number_of_reserved_segments , "could not reserve enough space for segment map");
  assert(_segmap.reserved_size()  >= _segmap.committed_size()     , "just checking");
  // initialize remaining instance variables
  clear();
  return true;
}


void CodeHeap::release() {
  Unimplemented();
}


bool CodeHeap::expand_by(int size) {
  assert(size > 0, "size must be > 0");
  // expand _memory space
  int dm = align_to_page_size(_memory.committed_size() + size) - _memory.committed_size();
  if (dm > 0) {
    if (!_memory.expand_by(dm)) return false;
    int i = _number_of_committed_segments;
    _number_of_committed_segments = number_of_segments(_memory.committed_size());
    assert(_number_of_reserved_segments == number_of_segments(_memory.reserved_size()), "number of reserved segments should not change");
    assert(_number_of_reserved_segments >= _number_of_committed_segments, "just checking");
    // expand _segmap space
    int ds = align_to_page_size(_number_of_committed_segments) - _segmap.committed_size();
    if (ds > 0) {
      if (!_segmap.expand_by(ds)) return false;
    }
    assert(_segmap.committed_size() >= (size_t) _number_of_committed_segments, "just checking");
    // initialize additional segmap entries
    mark_segmap_as_free(i, _number_of_committed_segments);
  }
  return true;
}


void CodeHeap::shrink_by(int size) {
  Unimplemented();
}


void CodeHeap::clear() {
  _next_segment = 0;
  mark_segmap_as_free(0, _number_of_committed_segments);
}


void* CodeHeap::allocate(int size) {
  int length = number_of_segments(size + sizeof(HeapBlock));
  assert(length *_segment_size >= sizeof(FreeBlock), "not enough room for FreeList");
    
  // First check if we can satify request from freelist
  debug_only(verify());
  HeapBlock* block = search_freelist(length);
  debug_only(verify());
  if (block != NULL) {
    assert(block->length() == length, "sanity check");    
    assert(!block->free(), "must be marked free");    
    return block->allocated_space();
  }

  if (_next_segment + length <= _number_of_committed_segments) {
    mark_segmap_as_used(_next_segment, _next_segment + length);
    HeapBlock* b =  block_at(_next_segment);
    b->initialize(length);    
    _next_segment += length;
    return b->allocated_space();
  } else {
    return NULL;
  }
}


void CodeHeap::deallocate(void* p) {
  assert(p == find_start(p), "illegal deallocation");
  // Find start of HeapBlock
  HeapBlock* b = (((HeapBlock *)p) - 1);
  assert(b->allocated_space() == p, "sanity check");
  add_to_freelist(b);

  debug_only(verify());
}


void* CodeHeap::find_start(void* p) const {
  if (!contains(p)) {
    return NULL;
  }
  int i = segment_for(p);
  address b = (address)_segmap.low();
  if (b[i] == 0xFF) {
    return NULL;
  }
  while (b[i] > 0) i -= (int)b[i];
  HeapBlock* h = block_at(i);
  if (h->free()) {
    return NULL;
  }
  return h->allocated_space();
}


int CodeHeap::alignment_unit() const {
  // this will be a power of two
  return _segment_size;
}


int CodeHeap::alignment_offset() const {
  // The lowest address in any allocated block will be
  // equal to alignment_offset (mod alignment_unit).
  return sizeof(HeapBlock) & (_segment_size - 1);
}

// Finds the next free heapblock. If the current one is free, that it returned
void* CodeHeap::next_free(HeapBlock *b) const {
  // Since free blocks are merged, there is max. on free block
  // between two used ones
  if (b != NULL && b->free()) b = next_block(b);
  assert(b == NULL || !b->free(), "must be in use or at end of heap");
  return (b == NULL) ? NULL : b->allocated_space();
}

// Returns the first used HeapBlock
HeapBlock* CodeHeap::first_block() const {
  if (_next_segment > 0) {    
    return block_at(0);
  } else {
    return NULL;
  }
}

HeapBlock *CodeHeap::block_start(void *q) const {
  HeapBlock* b = (HeapBlock*)find_start(q);
  if (b == NULL) return NULL;
  b -= 1;
  return b;
}

// Returns the next Heap block an offset into one
HeapBlock* CodeHeap::next_block(HeapBlock *b) const {  
  if (b == NULL) return NULL;  
  int i = segment_for(b) + b->length();
  if (i < _next_segment) {
    return block_at(i);
  } else {
    return NULL;
  }
}


// Returns current capacity
int CodeHeap::capacity() const {
  return _memory.committed_size() + _segmap.committed_size();   
}

int CodeHeap::max_capacity() const {
  return _memory.reserved_size() + _segmap.reserved_size();
}

// Free list management

FreeBlock *CodeHeap::following_block(FreeBlock *b) { 
  return (FreeBlock*)(((address)b) + _segment_size * b->length()); 
}

// Inserts block b after a
void CodeHeap::insert_after(FreeBlock* a, FreeBlock* b) {
  assert(a != NULL && b != NULL, "must be real pointers");  
  
  // Link b into the list after a
  b->set_link(a->link());
  a->set_link(b);

  // See if we can merge blocks
  merge_right(b); // Try to make b bigger
  merge_right(a); // Try to make a include b
}

// Try to merge this block with the following block
void CodeHeap::merge_right(FreeBlock *a) {
  assert(a->free(), "must be a free block");
  if (following_block(a) == a->link()) {
    assert(a->link() != NULL && a->link()->free(), "must be free too");
    // Update block a to include the following block    
    a->set_length(a->length() + a->link()->length()); 
    a->set_link(a->link()->link());
    // Update find_start map
    int beg = segment_for(a);
    mark_segmap_as_used(beg, beg + a->length());    
  }
}

void CodeHeap::add_to_freelist(HeapBlock *a) {  
  FreeBlock* b = (FreeBlock*)a;
  assert(b != _freelist, "cannot be removed twice");

  // Mark as free and update free space count
  _free_segments += b->length();
  b->set_free();

  // First element in list?
  if (_freelist == NULL) {
    _freelist = b;
    b->set_link(NULL);
    return;
  }  
      
  // Scan for right place to put into list. List
  // is sorted by increasing addresseses
  FreeBlock* prev = NULL;
  FreeBlock* cur  = _freelist;    
  while(cur != NULL && cur < b) { 
    assert(prev == NULL || prev < cur, "must be ordered");
    prev = cur;
    cur  = cur->link();
  }
    
  assert( (prev == NULL && b < _freelist) ||          
          (prev < b && (cur == NULL || b < cur)), "list must be ordered");
  
  if (prev == NULL) {
    // Insert first in list
    b->set_link(_freelist);
    _freelist = b;    
    merge_right(_freelist); 
  } else {
    insert_after(prev, b);
  }
}

// Search freelist for an entry on the list with the best fit
// Return NULL if no one was found
FreeBlock* CodeHeap::search_freelist(int length) {
  FreeBlock *best_block = NULL;
  FreeBlock *best_prev  = NULL;
  int best_length = 0;

  // Search for smallest block witch is bigger than length
  FreeBlock *prev = NULL;
  FreeBlock *cur = _freelist;
  while(cur != NULL) {
    int l = cur->length();
    if (l >= length && (best_block == NULL || best_length > l)) {
      // Remember bedst block, its previous element, and its length
      best_block = cur;
      best_prev  = prev;
      best_length = best_block->length();
    }

    // Next element in list    
    prev = cur;
    cur  = cur->link();
  }

  if (best_block == NULL) {
    // None found
    return NULL;
  }

  assert((best_prev == NULL && _freelist == best_block ) || 
         (best_prev != NULL && best_prev->link() == best_block), "sanity check");

  // Exact fit. Remove from list
  if (best_length == length) {    
    if (best_prev == NULL) {
      assert(_freelist == best_block, "sanity check");
      _freelist = _freelist->link();            
    } else {
      // Unmap element
      best_prev->set_link(best_block->link());
    }        
  } else {
    // Truncate block and return a pointer to the following block
    best_block->set_length(best_length - length);  
    best_block = following_block(best_block);
    // Set used bit and length on new block 
    int beg = segment_for(best_block);
    mark_segmap_as_used(beg, beg + length);
    best_block->set_length(length); 
  }

  best_block->set_used();
  _free_segments -= length;
  return best_block;  
}

//----------------------------------------------------------------------------
// Non-product code

#ifndef PRODUCT

void CodeHeap::print() {
  tty->print_cr("The Heap");
}


void CodeHeap::verify() {
  // Verify that freelist contains the right amount of free space
  int count = 0;
  for(FreeBlock* b = _freelist; b != NULL; b = b->link()) {
    count += b->length();
    if (count > _free_segments || count < 0) {
      fatal("wrong freelist");
    }
  }
  assert(count == _free_segments, "wrong freelist");  

  // Verify that there is the right number of live blobs
  
  count = 0;
  for(HeapBlock *h = first_block(); h != NULL; h = next_block(h)) {
    if (h->free()) count++;
    if (count > 10000) {
      fatal("can this be true?");
    }
  }
}

#endif
