/*
 * @(#)heap.hpp	1.27 01/02/09 14:45:03
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// Blocks

class HeapBlock {
  friend class VMStructs;
 public:
  struct Header {
    int  _length;                                // the length in segments    
    int  _used;                                  // Used bit
  };

 protected:
  union {
    Header _header;
    int  _padding[ALIGN_ALL_OBJECTS ? 4 : 2];    // maybe pad to 0 mod 8
  };

 public:
  // Initialization
  void initialize(int length)                    { assert(length > 0, "length must be > 0"); _header._length = length; set_used(); }

  // Accessors
  void* allocated_space() const                  { return (void*)(this + 1); }
  int   length() const                           { return _header._length; }  

  // Used/free
  void set_used()                                { _header._used = 1; }
  void set_free()                                { _header._used = 0; }
  bool free()                                    { return _header._used == 0; }
};

class FreeBlock: public HeapBlock {
  friend class VMStructs;
 protected:  
  FreeBlock* _link; 

 public:
  // Initialization
  void initialize(int length)                { HeapBlock::initialize(length); _link= NULL; }

  // Merging    
  void set_length(int l)                     { _header._length = l; }
  
  // Accessors    
  FreeBlock* link() const                    { return _link; }  
  void set_link(FreeBlock* link)             { _link = link; }  
};

class CodeHeap VALUE_OBJ_CLASS_SPEC  {
  friend class VMStructs;
 private:
  VirtualSpace _memory;                          // the memory holding the blocks
  VirtualSpace _segmap;                          // the memory holding the segment map

  int          _number_of_committed_segments;
  int          _number_of_reserved_segments;
  int          _segment_size;
  int          _log2_segment_size;

  int          _next_segment;

  FreeBlock*   _freelist; 
  int          _free_segments;                    // No. of segments in freelist

  // Helper functions
  int   number_of_segments(int size) const       { return (size + _segment_size - 1) >> _log2_segment_size; }
  int   size(int number_of_segments) const       { return number_of_segments << _log2_segment_size; }

  int    segment_for(void* p) const              { return ((char*)p - _memory.low()) >> _log2_segment_size; }
  HeapBlock* block_at(int i) const               { return (HeapBlock*)(_memory.low() + (i << _log2_segment_size)); }

  void  mark_segmap_as_free(int beg, int end);
  void  mark_segmap_as_used(int beg, int end);

  // Freelist management helpers      
  FreeBlock* following_block(FreeBlock *b);
  void insert_after(FreeBlock* a, FreeBlock* b);
  void merge_right (FreeBlock* a);
  
  
  // Toplevel freelist management
  void add_to_freelist(HeapBlock *b);
  FreeBlock* search_freelist(int length);

  // Iteration helpers
  void*      next_free(HeapBlock* b) const;
  HeapBlock* first_block() const;
  HeapBlock* next_block(HeapBlock* b) const;
  HeapBlock* block_start(void* p) const;

 public:
  CodeHeap();

  // Heap extents
  bool  reserve(int reserved_size, int committed_size, int segment_size);
  void  release();                               // releases all allocated memory
  bool  expand_by(int size);                     // expands commited memory by size
  void  shrink_by(int size);                     // shrinks commited memory by size
  void  clear();                                 // clears all heap contents

  // Memory allocation
  void* allocate  (int size);                    // allocates a block of size or returns NULL
  void  deallocate(void* p);                     // deallocates a block

  // Attributes
  void* begin() const                            { return _memory.low (); }
  void* end() const                              { return _memory.high(); }
  bool  contains(void* p) const                  { return begin() <= p && p < end(); }
  void* find_start(void* p) const;               // returns the block containing p or NULL
  int alignment_unit() const;                    // alignment of any block
  int alignment_offset() const;                  // offset of first byte of any block, within the enclosing alignment unit
  static int header_size();                      // returns the header size for each heap block


  // Iteration
  
  // returns the first block or NULL
  void* first() const       { return next_free(first_block()); }
  // returns the next block given a block p or NULL
  void* next(void* p) const { return next_free(next_block(block_start(p))); }

  // Statistics
  int   capacity() const;  
  int   max_capacity() const;
  
  // Debugging
  void  verify() PRODUCT_RETURN;
  void  print()  PRODUCT_RETURN;
};
