/*
 * @(#)markSweep.hpp	1.37 01/02/09 14:45:07
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// MarkSweep takes care of global mark-compact garbage collection
// using a four-phase pointer forwarding algorithm.
//
// Class unloading will only occur when a full gc is invoked.

// If VALIDATE_MARK_SWEEP is defined, the -XX:+ValidateMarkSweep flag will
// be operational, and will provide slow but comprehensive self-checks within
// the GC.  This is not enabled by default in product or release builds,
// since the extra call to track_adjusted_pointer() in _adjust_pointer()
// would be too much overhead, and would disturb performance measurement.
// However, debug builds are sometimes way too slow to run GC tests!
#ifdef ASSERT
#define VALIDATE_MARK_SWEEP 1
#endif
#ifdef VALIDATE_MARK_SWEEP
#define VALIDATE_MARK_SWEEP_ONLY(code) code
#else
#define VALIDATE_MARK_SWEEP_ONLY(code)
#endif



// declared at end
class PreservedMark;


class MarkSweep : AllStatic {
  friend class VM_MarkSweep;
  friend void Scavenge::invoke_at_safepoint(int, bool, bool&);
  friend void marksweep_init();
 private:
  static bool           _is_active;  
  static elapsedTimer   _accumulated_time;
  static int            _invoke_count;
  static jlong          _time_of_last_gc;
 
  // Traversal stack used during phase1
  static GrowableArray<oop>*      _marking_stack;   
  // Stack for live klasses to revisit at end of marking phase
  static GrowableArray<Klass*>*   _revisit_klass_stack;   

  // Space for storing/restoring mark word
  static GrowableArray<markOop>*         _preserved_mark_stack;
  static GrowableArray<oop>*             _preserved_oop_stack;
  static int			         _preserved_count;
  static int			         _preserved_count_max;
  static PreservedMark*                  _preserved_marks;
  
#ifdef VALIDATE_MARK_SWEEP
  static GrowableArray<oop*>*    _root_refs_stack;
  static GrowableArray<oop> *    _live_oops;
  static GrowableArray<oop> *    _live_oops_moved_to;
  static GrowableArray<int> *    _live_oops_size;
  static int                     _live_oops_index;
  static int                     _live_oops_index_at_perm;
  static GrowableArray<oop*>*    _other_refs_stack;
  static GrowableArray<oop*>*    _adjusted_pointers;
  static bool                    _pointer_tracking;
  static bool                    _root_tracking;
#endif


 public:
  // Invoke full gc at safepoint
  static void invoke(int size_to_be_allocated, MarkSweepInvokeCause cause, TRAPS);

  // Call backs for marking
  static void mark_object(oop obj);
  static void follow_root(oop* p);        // Mark pointer and follow contents. Empty marking
                                          // stack afterwards.
  static void mark_and_follow(oop* p);    // Mark pointer and follow contents.
  static void _mark_and_push(oop* p);     // Mark pointer and push obj on marking stack.
  static void mark_and_push(oop* p) {     // Check mark and maybe push on marking stack
    // assert(Universe::is_reserved_heap((oop)p), "we should only be traversing objects here");
    oop m = *p;
    if (m && !m->mark()->is_marked()) {
      _mark_and_push(p);
    }
  }

  static void follow_stack();             // Empty marking stack.
  static void preserve_mark(oop p, markOop mark);       // Save the mark word so it can be restored later
  static void adjust_marks();             // Adjust the pointers in the preserved marks table
  static void restore_marks();            // Restore the marks that we saved in preserve_mark

  static void adjust_root_pointer(oop* p) { // Adjust something which should be a root
    _adjust_pointer(p, true);
  }

  static void adjust_pointer(oop* p) {
    _adjust_pointer(p, false);
  }

  static void _adjust_pointer(oop* p, bool isroot) {
    oop obj = *p;
    VALIDATE_MARK_SWEEP_ONLY(oop saved_new_pointer = NULL);
    if (obj) {
      oop new_pointer = oop(obj->mark()->decode_pointer());
      assert(new_pointer != NULL || obj->mark() == markOopDesc::prototype(),
	     "should contain a forwarding pointer");
      if (new_pointer != NULL) {
	*p = new_pointer;
	assert(Universe::is_reserved_heap(new_pointer), "should be in object space");
	VALIDATE_MARK_SWEEP_ONLY(saved_new_pointer = new_pointer);
      }
    }
    VALIDATE_MARK_SWEEP_ONLY(track_adjusted_pointer(p, saved_new_pointer, isroot));
  }

#ifdef VALIDATE_MARK_SWEEP
  static void track_adjusted_pointer(oop* p, oop newobj, bool isroot);
  static void check_adjust_pointer(oop* p);     // Adjust this pointer
  static void track_interior_pointers(oop obj);
  static void check_interior_pointers();

  static void reset_live_oop_tracking(bool at_perm);
  static void register_live_oop(oop p, int size);
  static void validate_live_oop(oop p, int size);
  static void live_oop_moved_to(oop* q, int size, oop* compaction_top);
#endif

  // Call backs for class unloading
  static void revisit_weak_klass_link(Klass* k);  // Update subklass/sibling/implementor links at end of marking.

  // Tells whether collection is in progress (used for profiling)
  static bool is_active()                 { return _is_active; }

  // Number of collections done
  static int  invoke_count()              { return _invoke_count; }

  // Time since last full gc (in milliseconds)
  static jlong time_since_last_gc();

  // Timing
  static elapsedTimer* accumulated_time() { return &_accumulated_time; }

 private:
  // Entry point called from vm_operation
  static void invoke_at_safepoint(int size_to_be_allocated, bool deferred, bool clear_all_softrefs, bool& notify_ref_lock, MarkSweepInvokeCause cause);

  // Mark live objects
  static void mark_sweep_phase1(bool& marked_for_deopt, bool clear_all_softrefs);
  // Calculate new addresses
  static void mark_sweep_phase2();
  // Update pointers
  static void mark_sweep_phase3();
  // Move objects to new positions
  static void mark_sweep_phase4();

  // Used for java/lang/ref handling
  static bool mark_sweep_is_alive(oop p);
  static void mark_sweep_keep_alive(oop* p);

  // Temporary data structures for traversal and storing/restoring marks
  static void allocate_stacks();
  static void deallocate_stacks();

  // Class unloading. Update subklass/sibling/implementor links at end of marking phase.
  static void follow_weak_klass_links();

  // Reset time since last full gc
  static void reset_time_since_last_gc();

  // Debugging
  static void trace(const char* msg) PRODUCT_RETURN;
  static const char* invoke_cause_name(MarkSweepInvokeCause cause) PRODUCT_RETURN0;
};


class PreservedMark {
private:
  oop _obj;
  markOop _mark;

public:
  void init(oop obj, markOop mark) {
    _obj = obj;
    _mark = mark;
  }

  void adjust_pointer() {
    MarkSweep::adjust_pointer(&_obj);
  }

  void restore() {
    _obj->set_mark(_mark);
  }
};

