/*
 * @(#)rememberedSet.cpp	1.75 01/02/09 14:45:09
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_rememberedSet.cpp.incl"


jbyte* RememberedSet::_byte_map      = NULL;
jbyte* RememberedSet::byte_map_base  = NULL;


void RememberedSet::initialize() {
  // Compute virtual spaces for card marking array. Note that we let the virtual space for
  // old_gen cover new_gen as well.
  oop* low_boundary       = Universe::new_gen()->low_boundary;
  oop* perm_low_boundary  = Universe::perm_gen()->low_boundary;
  oop* high_boundary      = Universe::perm_gen()->high_boundary;
  assert((juint(low_boundary)  & (card_size - 1))  == 0, "new must start at card boundary");
  assert((juint(perm_low_boundary) & (card_size - 1))  == 0, "perm must start at card boundary");
  assert((juint(high_boundary) & (card_size - 1))  == 0, "perm must end at card boundary");

  assert(card_size <= 512, "card_size must be less than 512");
  size_t reserved_size          = ReservedSpace::page_align_size_up(((size_t) high_boundary - (size_t) low_boundary) / card_size);
  size_t oldgen_reserved_size   = ReservedSpace::page_align_size_up(((size_t) perm_low_boundary - (size_t) low_boundary) / card_size);
  size_t permgen_reserved_size  = ReservedSpace::page_align_size_up(((size_t) high_boundary - (size_t) perm_low_boundary) / card_size);
  assert(oldgen_reserved_size + permgen_reserved_size == reserved_size, "invalid card marking array sizes");

  ReservedSpace heap_rs(reserved_size);
  if (!heap_rs.is_reserved()) {
    vm_exit_during_initialization("Could not reserve enough space for card marking array");
  }
  ReservedSpace old_rs     = heap_rs.first_part(oldgen_reserved_size);
  ReservedSpace perm_rs    = heap_rs.last_part(oldgen_reserved_size);

  // The assember store_check code will do an unsigned shift of the oop, 
  // then add it to byte_map_base, i.e.
  // 
  //   _byte_map = byte_map_base + juint(low_boundary) >> card_shift
  _byte_map = (jbyte*) heap_rs.base();
  byte_map_base = _byte_map - (juint(low_boundary) >> card_shift);
  assert(byte_for(low_boundary) == _byte_map, "Checking start of map");

  Universe::old_gen()->initialize_card_marking_array(old_rs);
  Universe::perm_gen()->initialize_card_marking_array(perm_rs);

  assert(oop_for(byte_for(low_boundary)) == low_boundary, "Checking start of object heap");
  assert(oop_for(byte_for(perm_low_boundary)) == perm_low_boundary, "Checking start of perm heap");
  assert(_byte_map == (jbyte*) Universe::old_gen()->_card_marking_space.low_boundary(), "Checking start of map");
}


void RememberedSet::clear() {
  Universe::old_gen()->clear_card_marking_array();
  Universe::perm_gen()->clear_card_marking_array();
}


void RememberedSet::clear_for_space(OldSpace* s) {
  if (_byte_map != NULL) {  // check initialized (bootstrapping problem)
    jbyte* start = byte_for(s->bottom());
    jbyte* end   = byte_for(s->end() - 1);
    while (start <= end) {
      *start++ = clean_card;
    }
  }
}


void RememberedSet::invalidate_for_space(OldSpace* s) {
  jbyte* start = byte_for(s->bottom());
  jbyte* top   = byte_for(s->top() - 1);
  jbyte* end   = byte_for(s->end() - 1);
  // Invalidate used part of space and clear unused part.
  // Note that unused part is not guaranteed to be clear since
  // object sizes may have been stored there during mark-sweep.
  while (start <= top) {
    *start++ = dirty_card;
  }
  while (start <= end) {
    *start++ = clean_card;
  }
}


class RememberedSetInvalidateClosure: public SpaceClosure {
  void do_space(Space* sp) {
    if (sp->is_old()) {
      RememberedSet::invalidate_for_space((OldSpace*) sp);
    }
  }
};


void RememberedSet::invalidate() {
  RememberedSetInvalidateClosure blk;
  Universe::space_iterate(&blk);
}


void RememberedSet::record_train_gc_store(oop* from, oop to) {
  assert(UseTrainGC, "Just checking");
  assert(oop(from)->is_old_or_perm(), "Newgen cell not expected");
  assert(to == *from, "Invalid arguments");
  assert(to != NULL, "Cell containing NULL not expected");
  assert(!to->is_perm(), "Should not be in perm gen");
  CarTableDesc* to_desc = CarTable::desc_for((oop*) to);
  if (to_desc->is_new()) {
    jbyte* from_card = byte_for(from);
    *from_card = newgen_card;
  } else {
    assert(to_desc->is_train(), "Should be in train gen");
    CarTableDesc* from_desc = CarTable::desc_for(from);
    if (to_desc->less_than(from_desc)) {
      ((CarSpace*) to_desc->space())->remembered_set()->add_reference(from);
    }
  }
}


void RememberedSet::record_runtime_store(oop* from, oop to) {
  assert(oop(from)->is_old_or_perm(), "Newgen cell not expected");
  assert(to == *from, "Invalid arguments");
  assert(to != NULL, "Cell containing NULL not expected");
  if (to->is_new()) {
    jbyte* from_card = byte_for(from);
    *from_card = newgen_card;
  } else {
    if (UseTrainGC) {
      CarTableDesc* to_desc = CarTable::desc_for((oop*) to);
      if (!to_desc->is_perm()) {
        assert(to_desc->is_train(), "Should be in train gen");
        CarTableDesc* from_desc = CarTable::desc_for(from);
        if (to_desc->less_than(from_desc)) {
          ((CarSpace*) to_desc->space())->remembered_set()->add_reference(from);
        }
      }
    } else {
      assert(to->is_old_or_perm(), "Should be in perm gen");
    }
  }
}


void RememberedSet::record_gc_store(oop* from, oop to) {
  assert(!UseTrainGC, "Just checking");
  assert(oop(from)->is_old_or_perm(), "Newgen cell not expected");
  assert(to == *from, "Invalid arguments");
  assert(to != NULL, "Cell containing NULL not expected");
  if (to->is_new()) {
    jbyte* from_card = byte_for(from);
    *from_card = newgen_card;
  } else {
    assert(to->is_old_or_perm(), "Should be in old or perm gen");
  }
}


void RememberedSet::scavenge_contents(OldSpace* sp) {
  if (sp->not_empty()) {
    oop* sp_top = sp->top();
    oop* prev_top = NULL;
    jbyte* current_card = byte_for(sp->bottom());
    jbyte* end_card     = byte_for(sp_top - 1);    // sp_top is exclusive
    // scan card marking array
    while (current_card <= end_card) {
      jbyte value = *current_card;
      // skip clean cards
      if (card_is_clean(value)) {
        current_card++;
      } else {
        // we found a non-clean card
        jbyte* first_nonclean_card = current_card++;
        oop* bottom = oop_for(first_nonclean_card);
        // find object starting on card
        oop* bottom_obj = (oop*)sp->fast_object_start(bottom);
        assert(bottom_obj <= bottom, "just checking");
        // make sure we don't scan oops we already looked at
        if (bottom < prev_top) bottom = prev_top;
        // figure out when to stop scanning
        jbyte* first_clean_card;
        oop* top;
        bool restart_scanning;
        do {
          restart_scanning = false;
          // find a clean card
          while (current_card <= end_card) {
            value = *current_card;
            if (card_is_clean(value)) break;
            current_card++;
          }
          // check if we reached the end, if so we are done
          if (current_card >= end_card) {
            first_clean_card = end_card + 1;
            current_card++;
            top = sp_top;
          } else {
            // we have a clean card, find object starting on that card
            first_clean_card = current_card++;
            top = oop_for(first_clean_card);
            oop* top_obj = (oop*)sp->fast_object_start(top);
            assert(top_obj <= top, "just checking");
            if (oop(top_obj)->is_objArray() || oop(top_obj)->is_typeArray()) {
              // an arrayOop is starting on the clean card - since we do exact store
              // checks for objArrays we are done
            } else {
              // otherwise, it is possible that the object starting on the clean card
              // spans the entire card, and that the store happened on a later card.
              // figure out where the object ends
              top = top_obj + oop(top_obj)->size();
              jbyte* top_card = byte_for(top - 1);   // top is exclusive
              if (top_card > first_clean_card) {
                // object ends a different card
                current_card = top_card + 1;
                if (card_is_clean(*top_card)) {
                  // the ending card is clean, we are done
                  first_clean_card = top_card;
                } else {
                  // the ending card is not clean, continue scanning at start of do-while
                  restart_scanning = true;
                }
              } else {
                // object ends on the clean card, we are done.
                assert(first_clean_card == top_card, "just checking");
              }
            }
          }
        } while (restart_scanning);
        // we know which cards to scan, now clear them
        while (first_nonclean_card < first_clean_card) {
          *first_nonclean_card++ = clean_card;
        }
        // scan oops in objects
        do {
          bottom_obj += oop(bottom_obj)->scavenge_contents(bottom, top);
          assert(bottom_obj <= sp_top, "just checking");
        } while (bottom_obj < top);
        // remember top oop* scanned
        prev_top = top;
      }
    }
  }
}


#ifndef PRODUCT

class RememberedSetPrintClosure: public SpaceClosure {
 public:
  void do_space(Space* s) {
    RememberedSet::print_set_for_space((OldSpace*) s);
  }
};


void RememberedSet::print() {
  RememberedSetPrintClosure blk;
  Universe::space_iterate(&blk);
};


void RememberedSet::print_set_for_range(jbyte* start, jbyte* end) {
  const int max_indent = 2;
  int indent = 0;
  while (start <= end) {
    jbyte value = *start;
    char c;
    if (card_is_dirty(value)) {
      c = 'D';
    } else {
      if (card_is_newgen(value)) {
        c ='N';
      } else {
        assert(card_is_clean(value), "just checking");
        c ='C';
      }
    }
    tty->print("[0x%08x:0x%08x %c]", start, oop_for(start), c);
    indent++;
    if (indent == max_indent) {
      tty->cr();
      indent = 0;
    }
    start++;
  }
  tty->cr();
}


void RememberedSet::print_set_for_space(OldSpace* sp) {
  if (sp->not_empty()) {
    oop* start = sp->bottom();
    oop* end = sp->top() - 1;
    print_set_for_range(byte_for(start), byte_for(end));
  }
}


void RememberedSet::print_set_for_object(oop obj) {
  oop* obj_end = (oop*) obj + obj->size();
  jbyte* start = byte_for((oop*) obj);
  jbyte* end   = byte_for(obj_end - 1);
  print_set_for_range(start-1,end+1);   // add one on each side to detect bugs
}


bool RememberedSet::is_within_array_bounds(jbyte* card)  {
  if (Universe::old_gen()->_card_marking_space.contains(card)) return true;
  if (Universe::perm_gen()->_card_marking_space.contains(card)) return true;
  return false; 
}


bool RememberedSet::is_within_heap_bounds(oop* p)  { 
  // cannot call is_reserved_heap in hpp file (include dependencies)
  return Universe::is_reserved_heap(oop(p));
}


int RememberedSet::number_of_marked_cards_for_space(Space* s) {
  assert(s->is_old(), "should be old space");
  int result = 0;
  jbyte* start = byte_for(s->bottom());
  jbyte* end   = byte_for(s->top() - 1);
  while (start <= end) {
    if (!card_is_clean(*start)) result++;
    start++;
  }
  return result;
}


class RememberedSetCountClosure: public SpaceClosure {
 public:
  int result;
  void do_space(Space* s) {
    if (s->is_old()) {
      result += RememberedSet::number_of_marked_cards_for_space(s);
    }
  }
};


int RememberedSet::number_of_marked_cards() {
  RememberedSetCountClosure blk;
  blk.result = 0;
  Universe::space_iterate(&blk);
  return blk.result;
}


void RememberedSet::verify_old_oop(oop obj, oop* from, bool allow_dirty, bool precise_card) {
  if (!Universe::is_reserved_heap(oop(from))) {
    // we can hit this case when verifying e.g. OopMapCacheEntry elements
    guarantee((*from)->safepoint_is_oop_or_null(), "invalid oop");
    return;
  }
  guarantee(oop(from)->is_old_or_perm(), "illegal arguments to verify_old_precise_oop");
  const char* error_msg = NULL;
  oop to = *from;
  // Figure out which card values to use
  jbyte* from_card = byte_for(from);
  jbyte from_card_value = *from_card;

  jbyte* obj_card = byte_for((oop*) obj);
  jbyte obj_card_value;
  if (precise_card || !allow_dirty) {
    obj_card_value = clean_card;  // no effect
  } else {
    obj_card_value = *obj_card;
  }

  // Check for dirty card
  if (card_is_dirty(obj_card_value) || card_is_dirty(from_card_value)) {
    if (!allow_dirty) {
      error_msg = "Dirty card not expected";
    }
  } else {
    // Otherwise check if remset info is correct
    if (to != NULL && !to->is_perm()) {
      // from is a oldgen or permgen cell pointing to a newgen or oldgen object
      if (to->is_new()) {
        // to is a newgen object
        if (card_is_clean(obj_card_value) && card_is_clean(from_card_value)) {
          error_msg = "Newgen reference found on unmarked card";
        }
      } else {
        if (UseTrainGC) {
          // to is a traingen object
          CarTableDesc* from_desc = CarTable::desc_for(from);
          CarTableDesc* to_desc   = CarTable::desc_for((oop*)to);
          guarantee(to_desc->train() != NULL, "should be in train gen");
          if (to_desc->less_than(from_desc)) {
            if (!((CarSpace*) to_desc->space())->remembered_set()->contains_reference(from)) {
              error_msg = "Lower train reference not present";
            }
          }
        }
      }
    }
  }
  if (error_msg) {
    tty->cr();
    tty->print_cr("%s", error_msg);
    tty->print_cr("- object containing the reference (0x%x):", obj);
    obj->print();
    tty->print_cr("- the referred object (0x%x->0x%x):", from , to);
    to->print_value();
    tty->cr();
    print_set_for_object(obj);
    Universe::print();
    fatal("GC remembered set problem");
  }
};


void RememberedSet::verify() {
  // verify done by generations
}


#endif // PRODUCT
