/*
 * @(#)rememberedSet.hpp	1.50 01/02/09 14:45:10
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// Card marking array for garbage collection. Currently a byte array with one byte / card.
// Store checks are done by simply storing a dirty value into the corresponding byte.
// Non-dirty entires contain information about which pointers are present in the card, and can
// also store information about exact pointer locations (not necessarily enabled).
//
// Card size is 512 bytes, and must be <= 512 because of the offset_arrays.
//
// For regular objects the card being dirtied correspond to the object start rather than the 
// actual pointer store location (so if the object starts towards the end of a card, the updated 
// location may be on the next card). This "fuzzy" marking makes the store check faster and
// enables the compiler to emit only one store check for multiple stores into the same object.
// 
// For objArrays, the precise card corresponding to the pointer store is dirtied so we don't need 
// to scavenge the entire array.
//
// For computed entries (filled in during GC) the relevant card is always the one containing the 
// actual pointer location.
//
// During global mark-sweep the array is used to remember object sizes, see further below.


class RememberedSet: public AllStatic {
  friend class VMStructs;
 private:
  static jbyte*       _byte_map;      // the card marking array

 public:
  // Constants
  enum SomePublicConstants {
    card_shift                  = 9,
    card_size                   = 1 << card_shift,
    card_size_in_oops           = card_size / oopSize,
    clean_card                  = -1,
    newgen_card                 = 1,
    dirty_card                  = 0
  };

  // Because many machines have addressing modes that do an addition, it is cheaper
  // to do a store-check by shifting the oop then adding something, as opposed
  // to the simpler version of subtracting, shifting, then adding, as in byte_for.
  // There is some trickiness here, when you allow for object heaps anywhere in
  // the address space, even in the top half. The shift performed in the store check
  // stub must be UNSIGNED, otherwise an object heap crossing 0x80000000 would give
  // inconsistent results.

  // Card marking array base (adjusted for heap low boundary)
  static jbyte* byte_map_base;

  // Mapping from address to card marking array entry
  static jbyte* byte_for(oop* p) { 
    assert(is_within_heap_bounds(p), "out of bounds access to card marking array");
    return &byte_map_base[((juint)p) >> card_shift];
  }

  // Mapping from card marking array entry to address (first word on card)
  static oop* oop_for(jbyte* p) { 
    assert(is_within_array_bounds(p), "out of bounds access to card marking array");
    return (oop*) ((p - byte_map_base) << card_shift);
  }

  // Initialization
  static void initialize();

  // Grow remembered set to cover grown old space
  static void grow();

  // Filling of card marking array
  static void clear();
  static void clear_for_space(OldSpace* s);
  static void invalidate_for_space(OldSpace* s);
  static void invalidate();

  // Check pointer store and update set accordingly
  static void record_gc_store(oop* from, oop to);
  static void record_train_gc_store(oop* from, oop to);
  static void record_runtime_store(oop* from, oop to);

  // Same for multi-word stores (all words in [start..limit[ have been stored into)
  static void record_array_store(oop* start, oop* limit) {
    if (oop(start)->is_old_or_perm()) {
      jbyte* s = byte_for(start);
      jbyte* e = byte_for(limit - 1);
      while (s <= e) *s++ = dirty_card;
    }
  }

  // Testers for entries
  static bool card_is_dirty(int value)      { return value == dirty_card; }
  static bool card_is_newgen(int value)     { return value == newgen_card; }
  static bool card_is_clean(int value)      { return value == clean_card; }

  // Scavenge support
  static void scavenge_contents(OldSpace* s);

  // Debugging  
  static void verify() PRODUCT_RETURN;

  // Printing
  static void print() PRODUCT_RETURN;

  #ifndef PRODUCT
  // Debugging
  static void verify_old_oop(oop obj, oop* from, bool allow_dirty, bool precise_card);
  static bool is_within_array_bounds(jbyte* card);
  static bool is_within_heap_bounds(oop* p);
  static int number_of_marked_cards_for_space(Space* s);
  static int number_of_marked_cards();

  static void print_set_for_space(OldSpace* s);
  static void print_set_for_object(oop obj);
  static void print_set_for_range(jbyte* start, jbyte* end);   // end is inclusive

  #endif
};
