/*
 * @(#)scavenge.hpp	1.26 99/12/02 20:34:25
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


// All code for new generation scavenge and old generation collection using
// the train algorithm is located in class Scavenge.
//
// The code for scavenge and train tick is interleaved in order to avoid
// scanning the basic roots and the card marking array twice.

class Scavenge: AllStatic {
  friend class VM_Scavenge;
  friend class MarkSweep;
  friend void scavenge_init();
  friend oop oopDesc::copy_to_survivor_space(oop*);
  friend oop oopDesc::copy_to_train(oop*);

 private:
  // Flags/counters
  static ageTable       _age_table;                 // object age table for scavenge
  static int            _tenuring_threshold;        // tenuring threshold for next scavenge
  static int            _invoke_count;              // number of scavenges done
  static bool           _is_active;                 // flag telling whether scavenge is in progress
  static elapsedTimer   _accumulated_time;          // total time spent on scavenge
  static bool           _deferred_pending;          // tells whether an deferred scavenge is scheduled
  static int            _full_gc_revert_count;      // tells whether the next scavenge should revert to full mark-sweep
  static bool           _last_scavenge_near_limit;  // tells whether the last scavenge was unable to proceed because 
                                                    // old generation was too full

  static int            _train_invoke_count;        // number of train ticks done
  static bool           _reference_to_first_train_found;
  static int            _tick_interval;             // current train tick interval
  static int            _next_invoke_count;         // next invoke_count to do train tick
  static int            _delay_adjustment_count;    // ticks to skip before starting adaptive adjustment

  static jlong          _total_promoted;            // accumulated size of promoted objects (in words)
  static jlong          _total_processed;           // accumulated size of processed objects (in words)
  static jlong          _delta_promoted;            // saved value of _total_promoted
  static jlong          _delta_processed;           // saved value of _total_processed

  // Private accessors
  static ageTable*        age_table()                 { return &_age_table;                 }
  static int              tenuring_threshold()        { return _tenuring_threshold;         }

  static jlong            delta_promoted()            { return _delta_promoted;             }
  static jlong            delta_processed()           { return _delta_processed;            }
  static void             add_total_promoted(int n)   { _total_promoted += n;               }
  static void             add_total_processed(int n)  { _total_processed += n;              }

  static bool             last_scavenge_near_limit()  { return _last_scavenge_near_limit;   }
  static void             set_last_scavenge_near_limit(bool b) { _last_scavenge_near_limit = b; }

 public:
  // Accessors
  static int              invoke_count()              { return _invoke_count;               }
  static bool             is_active()                 { return _is_active;                  }
  static elapsedTimer*    accumulated_time()          { return &_accumulated_time;          }

  static int              train_invoke_count()        { return _train_invoke_count;         }
  static int              tick_interval()             { return _tick_interval;              }
  static bool             do_train_tick()             { return UseTrainGC && _invoke_count == _next_invoke_count; }

  static jlong            total_promoted()            { return _total_promoted;             }
  static jlong            total_processed()           { return _total_processed;            }

  // Invoke scavenge
  static void invoke(int size_to_be_allocated, TRAPS);

  // Invoke scavenge and allocate
  static oop* invoke_and_allocate(int size_to_be_allocated, bool is_large_typearray, TRAPS);

  // Test whether oop is relevant for current scavenge operation
  static inline bool should_scavenge(oop p);

  // Scavenge oop and update root. Called when should_scavenge returns true.
  static void (*scavenge)(oop*);

  // fill the new generation large dummy objects
  static void fill_newgen();

 private:
  // Scavenge functions and java/lang/ref handling
  static void scavenge_oop(oop* p);
  static void scavenge_oop_with_check(oop* p);
  static void scavenge_tenured_oop(oop* p);

  static bool scavenge_is_alive(oop p);
  static void scavenge_keep_alive(oop* p);

  static void train_scavenge_oop(oop* p);
  static void train_scavenge_oop_with_check(oop* p);
  static void train_scavenge_tenured_oop(oop* p);

  static void skip_train_scavenge_oop(oop* p);
  static void skip_train_scavenge_oop_with_check(oop* p);
  static void skip_train_scavenge_tenured_oop(oop* p);

  static bool train_scavenge_is_alive(oop p);
  static bool entire_train_scavenge_is_alive(oop p);

  static void compute_tick_adjustment();
  static void reset_tick_adjustment();

  // Computed scavenge functions (set at start of scavenge)
  static void (*_scavenge_nontenured)(oop*);
  static void (*_scavenge_nontenured_with_check)(oop*);
  static void (*_scavenge_tenured)(oop*);

  static void compute_scavenge_functions();
 
  // Scavenge entry point called from VM operation
  static void invoke_at_safepoint(int size_to_be_allocated, bool deferred, bool& notify_ref_lock);

  // Cheney scan
  static void evacuate_followers();

  // Debugging
  static void print_train_processing_statictics(julong prev_first_number) PRODUCT_RETURN;
};
