/*
 * @(#)space.cpp	1.114 00/08/10 21:58:49
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_space.cpp.incl"

// This is a helper class used during phase 3 and 4 to move all the objects
// Dead regions in a Space are linked together to keep track of the live regions
// so that the live data can be traversed quickly without having to look at each
// object.

class LiveRange {
private:
  oop* _bottom;
  oop* _top;

public:
  LiveRange(oop* bottom, oop* top) {
    set_bottom(bottom);
    set_top(top);
  }

  void set_bottom(oop* p) { _bottom = (oop*)((unsigned)p | markOopDesc::unlocked_value); }
  oop* bottom() const { return (oop*)((unsigned)_bottom & ~markOopDesc::unlocked_value); }
  void set_top(oop* p) {
    assert(p >= bottom(), "should be a non-zero range");
    _top = p;
  }
  oop* top() const { return _top; }

  // This is the size in oop*'s, not bytes
  size_t size() const { return top() - bottom(); }

  LiveRange * next() { return (LiveRange *) top(); }

  void move_to(oop* destination) {
    Memory::copy_oops_aligned_overlapping(bottom(), destination, size());
  }
};


void Space::initialize(oop* bottom, oop* end, bool clear_space) {
  assert(Universe::on_page_boundary(bottom) && Universe::on_page_boundary(end),
         "invalid space boundaries");
  set_bottom(bottom);
  set_end(end);
  if (clear_space) clear();
}


void Space::clear() {
  set_top(bottom());
  initialize_threshold();
  if (ZapUnusedHeapArea) mangle_unused_area();
}


void Space::mangle_unused_area() {
  // to-space is used for storing marks during mark-sweep
  if (this != Universe::new_gen()->to())
    Memory::set_oops(top(), free()/oopSize, badHeapOop);
}


void Space::prepare_for_compaction() {
  // Compute the new addresses for the live objects and store it in the mark
  // Used by universe::mark_sweep_phase2()

  Space* scanning_space = this;   // space currently sweeped
  Space* compaction_space = this; // space currently compacted into
  oop* compaction_top = compaction_space->bottom(); // new top in current compaction_space
  LiveRange * liveRange = NULL;
  bool skip_dead = (MarkSweep::invoke_count() % MarkSweepAlwaysCompactCount) ? true : false;

  oop* threshold = compaction_space->initialize_threshold();
  while (scanning_space != NULL) {
    int allowed_deadspace = 0;
    if (skip_dead) {
      int ratio = scanning_space->allowed_dead_ratio();
      allowed_deadspace = (scanning_space->capacity() * ratio / 100) / oopSize;
    }
    oop* q = scanning_space->bottom();
    oop* t = scanning_space->top();

    scanning_space->_first_dead = scanning_space->top();
    scanning_space->set_compaction_top(scanning_space->bottom());

    while (q < t) {
      assert(oop(q)->mark()->is_marked() || oop(q)->mark()->is_unlocked(),
             "these are the only valid states during a mark sweep");
      if (oop(q)->is_gc_marked()) {
        int size = oop(q)->size(); 

        // q is alive
        // First check if we should switch compaction space
        size_t compaction_max_size = compaction_space->end() - compaction_top;
        if ((size_t) size > compaction_max_size) {
          // switch to next compaction space
          compaction_space->set_compaction_top(compaction_top);
          compaction_space = compaction_space->next_compaction_space();
          compaction_top = compaction_space->bottom();
          compaction_space->set_compaction_top(compaction_space->bottom());
          threshold = compaction_space->initialize_threshold();
          continue;
        }

        // store the forwarding pointer into the mark word
        if (q != compaction_top) {
          oop(q)->forward_to((oop)compaction_top);
          assert(oop(q)->is_gc_marked(), "encoding the pointer should preserve the mark");
        } else {
          // if the object isn't moving we can just set the mark to the default
          // mark and handle it specially later on.  
          oop(q)->init_mark();
          assert(oop(q)->forwardee() == NULL, "should be forwarded to NULL");
        }

        debug_only(MarkSweep::register_live_oop(oop(q), size));
        q += size;
        compaction_top += size;

        // we need to update the offset table so that the beginnings of objects can be
        // found during scavenge.  Note that we are updating the offset table based on
        // where the object will be once the compaction phase finishes.
        if (compaction_top > threshold)
          threshold = compaction_space->update_offset_array_entry(compaction_top - size,
                                                      compaction_top);
      } else {
        // run over all the contiguous dead objects
        oop* end = q;
        end += oop(end)->size();
        while (end < t && !oop(end)->is_gc_marked()) {
          end += oop(end)->size();
        }

        // see if we might want to pretend this object is alive so that
        // we don't have to compact quite as often
        if (allowed_deadspace > 0) {
          int deadlength = end - q;
          allowed_deadspace -= deadlength;
          if (allowed_deadspace >= 0 && q == compaction_top) {
            oop(q)->set_mark(markOopDesc::prototype()->set_marked());
            const int min_int_array_size = typeArrayOopDesc::header_size(T_INT);
            if (deadlength >= min_int_array_size) {
                oop(q)->set_klass(Universe::intArrayKlassObj());
                typeArrayOop(q)->set_length(deadlength - min_int_array_size);
            } else {
                assert(deadlength == instanceOopDesc::header_size(), "size for smallest fake dead object doesn't match");
                oop(q)->set_klass(SystemDictionary::object_klass());
            }
            assert(deadlength == oop(q)->size(), "make sure size for fake dead object match");
            continue;
          }
          allowed_deadspace = 0;
        }

        // for the previous LiveRange, record the end of the live objects
        if (liveRange) {
          liveRange->set_top(q);
        }

        // record the current LiveRange object
        liveRange = (LiveRange *) q;
        liveRange->set_bottom(end);
        liveRange->set_top(end);

        if (scanning_space->_first_dead == scanning_space->top()) {
          scanning_space->_first_dead = q;
        }
        // move on to the next object
        q = end;
      }
    }
    assert(q == t, "just checking");
    if (liveRange != NULL) {
      liveRange->set_top(q);
      liveRange = NULL;
    }
    scanning_space = scanning_space->next_compaction_space();
  }
  if (compaction_space)
    compaction_space->set_compaction_top(compaction_top);
}


void Space::adjust_pointers() {
  // adjust all the interior pointers to point at the new locations of objects
  // Used by MarkSweep::mark_sweep_phase3()

  Space* scanning_space = this;   // space currently sweeped

  while (scanning_space != NULL) {
    oop* q = scanning_space->bottom();
    oop* t = scanning_space->top();

    if (q < t && scanning_space->_first_dead > q &&
        oop(q)->mark() == markOopDesc::prototype()) {
      // we have a chunk of the space which hasn't moved and we've reinitialized the
      // mark word during the previous pass, so we can't use is_gc_marked for the
      // traversal.
      oop* end = scanning_space->_first_dead;
      debug_only(oop* prev_q = NULL);
      
      while (q < end) {
        debug_only(MarkSweep::track_interior_pointers(oop(q)));
        int size = oop(q)->adjust_pointers();  // point all the oops to the new location
        debug_only(MarkSweep::check_interior_pointers());
        
        debug_only(prev_q = q);
        debug_only(MarkSweep::validate_live_oop(oop(q), size));
        
        q += size;
      }

      if (scanning_space->_first_dead == t) {
        q = t;
      } else {
        q = (oop*) oop(scanning_space->_first_dead)->mark()->decode_pointer();
      }
    }

    while (q < t) {
      debug_only(oop* prev_q = NULL);
      if (oop(q)->is_gc_marked()) {
        // q is alive

        debug_only(MarkSweep::track_interior_pointers(oop(q)));
        int size = oop(q)->adjust_pointers();  // point all the oops to the new location
        debug_only(MarkSweep::check_interior_pointers());

        debug_only(prev_q = q);
        debug_only(MarkSweep::validate_live_oop(oop(q), size));

        q += size;
      } else {
        // q is dead, so it's mark should point at the next live object
        debug_only(prev_q = q);
        q = (oop*) oop(q)->mark()->decode_pointer();
        assert(q > prev_q, "we should be moving forward through memory");
      }
    }
    assert(q == t, "just checking");
    scanning_space = scanning_space->next_compaction_space();
  }
}


void Space::compact() {
  // Copy all live objects to their new location
  // Used by MarkSweep::mark_sweep_phase4()

  bool jvmpi_move_event_enabled = Universe::jvmpi_move_event_enabled();
  Space* scanning_space = this;   // space currently sweeped

  while (scanning_space != NULL) {
    oop* q = scanning_space->bottom();
    oop* t = scanning_space->top();
    debug_only(oop* prev_q = NULL);

    if (q < t && scanning_space->_first_dead > q &&
        oop(q)->mark() == markOopDesc::prototype()) {
#ifdef ASSERT
      // we have a chunk of the space which hasn't moved and we've reinitialized the
      // mark word during the previous pass, so we can't use is_gc_marked for the
      // traversal.
      oop* end = scanning_space->_first_dead;
      
      while (q < end) {
        int size = oop(q)->size();
        assert(oop(q)->mark() == markOopDesc::prototype(), "should have prototype mark");
        debug_only(MarkSweep::live_oop_moved_to(q, size, q));
        debug_only(prev_q = q);
        q += size;
      }
#endif
      
      if (scanning_space->_first_dead == t) {
        q = t;
      } else {
        q = (oop*) oop(scanning_space->_first_dead)->mark()->decode_pointer();
      }
    }

    // JVMPI move notification must be done here since 'size' call goes
    // through klass, which may have been moved.
    // Loop body is unswitched/cloned since performance is critical
    if( !jvmpi_move_event_enabled ) {
      // normal path
      while (q < t) {
        if (!oop(q)->is_gc_marked()) { // mark is pointer to next marked oop
          debug_only(prev_q = q);
          q = (oop*) oop(q)->mark()->decode_pointer();
          assert(q > prev_q, "we should be moving forward through memory");
        } else {
          int size = oop(q)->size();
          oop* compaction_top = (oop*)oop(q)->forwardee();

          debug_only(MarkSweep::live_oop_moved_to(q, size, compaction_top));
          assert(q != compaction_top, "everything in this pass should be moving");

          Memory::copy_oops_aligned_overlapping(q, compaction_top, size);
          oop(compaction_top)->init_mark();
          assert(oop(compaction_top)->klass() != NULL, "should have a class");
          debug_only(prev_q = q);
          q += size;
        }
      }
    } else {
      // JVMPI -- perform object move notification
      while (q < t) {
        if (!oop(q)->is_gc_marked()) { // mark is pointer to next marked oop
          debug_only(prev_q = q);
          q = (oop*) oop(q)->mark()->decode_pointer();
          assert(q > prev_q, "we should be moving forward through memory");
        } else {
          int size = oop(q)->size();
          oop* compaction_top = (oop*)oop(q)->forwardee();

          debug_only(MarkSweep::live_oop_moved_to(q, size, compaction_top));
          assert(q != compaction_top, "everything in this pass should be moving");

          Memory::copy_oops_aligned_overlapping(q, compaction_top, size);
          oop(compaction_top)->init_mark();
          assert(oop(compaction_top)->klass() != NULL, "should have a class");
          // JVMPI move notification
          Universe::jvmpi_object_move((oop)q, (oop)compaction_top);
          debug_only(prev_q = q);
          q += size;
        }
      }
    }

    scanning_space->set_top_after_compaction(scanning_space->compaction_top());
    if (ZapUnusedHeapArea) scanning_space->mangle_unused_area();
    scanning_space = scanning_space->next_compaction_space();
  }
}


void Space::set_top_after_compaction(oop* value) {
  assert(value >= bottom() && value <= end(), "should point inside space");
  set_top(value);
}


void Space::object_iterate(ObjectClosure* blk) {
  if (is_empty()) return;
  WaterMark bm = bottom_mark();
  object_iterate_from(&bm, blk);
}


void Space::object_iterate_from(WaterMark* mark, ObjectClosure* blk) {
  assert(mark->space() == this, "Mark does not match space");
  oop* p = mark->point();
  oop* t = top();
  while (p < t) {
    blk->do_object(oop(p));
    p += oop(p)->size();
  }
  mark->set_point(p);
}


oop Space::slow_object_start(oop* p) {
  assert(bottom() <= p && p < top(), "p must be in space");
  oop* q = bottom();
  oop* t = top();
  while (q < t) {
    assert(oop(q)->is_oop(), "check for header");
    oop prev = oop(q);
    q += oop(q)->size();
    if (q > p) return prev;
  }
  ShouldNotReachHere();
  return NULL;
}


void Space::allocate_temporary_filler(int factor) {
  // allocate temporary type array decreasing free size with factor 'factor' 
  assert(factor >= 0, "just checking");
  int size = end() - top();
  if (factor > 0) {
    size -= size/factor;
  }
  size = align_object_size(size);
  int length = size - typeArrayOopDesc::header_size(T_INT);
  if (length >= 0) {
    // allocate uninitialized int array
    typeArrayOop t = (typeArrayOop) safepoint_allocate(size);
    assert(t != NULL, "allocation should succeed");
    t->set_mark();
    t->set_klass(Universe::intArrayKlassObj());
    t->set_length(length);
  }  
}


#ifndef PRODUCT

void Space::print_short() {
  tty->print(" space %6uK, %d%% used", capacity() / K, (int) ((double) used() * 100 / capacity()));
}

void Space::print() {
  print_short();
  tty->print_cr(" [%#-6lx,%#-6lx,%#-6lx)", bottom(), top(), end());
}


void Space::verify(bool allow_dirty) {
  oop* p = bottom();
  oop* t = top();
  oop* prev_p = NULL;
  while (p < t) {
    oop(p)->verify();
    prev_p = p;
    p += oop(p)->size();
  }
  guarantee(p == top(), "end of last object must match end of space");
}


bool Space::at_safepoint() const {
  return SafepointSynchronize::is_at_safepoint();
}


bool Space::no_gc_in_progress() const {
  return Universe::no_gc_in_progress();
}
#endif


void EdenSpace::ensure_proper_top() {
  if (proper_top() != NULL) {
    assert(proper_top() <= top_no_check(), "proper_top should be lower than top");
    assert(proper_top() <= end(), "proper_top should be lower than end");
    set_top(proper_top());
    set_proper_top(NULL);
  }
}


size_t EdenSpace::used() const {
  oop* t = top();
  if (t < end()) {
    return byte_size(bottom(), t);
  }
  if (proper_top() != NULL) {
    return byte_size(bottom(), proper_top());
  }
  return capacity();
}


size_t EdenSpace::free() const {
  oop* t = top_no_check();
  if (t < end()) {
    return byte_size(t, end());
  }
  if (proper_top() != NULL) {
    return byte_size(proper_top(), end());
  }
  return 0;
}


Space* EdenSpace::next_compaction_space() const {
  return Universe::new_gen()->from();
}


#ifndef PRODUCT

void EdenSpace::verify(bool allow_dirty) {
  guarantee(proper_top() == NULL, "proper_top should not be set");
  if(!UseTLE) {
    Space::verify(allow_dirty);
  }
  else {
    // TLEs verify themselves when they fill up (either naturally or due to scavenge)
  }
}

#endif


bool SurvivorSpace::scavenge_contents_from(WaterMark* mark) {
  assert(mark->point() <= top(), "scavenging past top");
  debug_only(if (TraceScavenge)
    tty->print_cr("scavenging toSpace  0x%x from 0x%x to 0x%x", bottom(), mark->point(), top())
  );

  oop* old_top = top();
  oop* p = mark->point();

  while (p < top()) {
    debug_only(oop *prev = p);
    oop m = oop(p);
    assert(m->safepoint_is_oop(), "check for header");
    p += m->scavenge_contents();
    debug_only(if (p - prev != m->size()) fatal("scavenge_contents is not returning the right size"))
  }
  mark->set_point(p);
  return old_top != top();
}


Space* SurvivorSpace::next_compaction_space() const {
  assert((const Space*) this == Universe::new_gen()->from(), "should be from_space");
  assert(Universe::new_gen()->to()->is_empty(), "to space should be empty");
  return NULL;
}
  

oop* OldSpace::initialize_threshold() {
  assert(!Universe::is_reserved_heap(oop(_offset_array)), "just checking");
  _next_offset_index                  = 0;
  _offset_array[_next_offset_index++] = 0;
  _next_offset_threshold               = bottom() + RememberedSet::card_size_in_oops;
  _last_offset_top                    = bottom();
  return _next_offset_threshold;
}


void OldSpace::set_top_after_compaction(oop* value) {
  assert(value >= bottom() && value <= end(), "should point inside space");
  set_top(value);
  _last_offset_top = value;
}


void OldSpace::check_offset_array_during_compaction(oop* begin, oop* end) {
  check_offset_array(begin, end);
}


oop* OldSpace::update_offset_array_entry(oop* begin, oop* end) {
  //  [       ][       ][       ]        cards for offset array
  //   begin^  ^t ^end
  assert(end > _next_offset_threshold, "should be past threshold");
  assert(_next_offset_threshold - begin <= RememberedSet::card_size_in_oops, "offset should be <= card_size_in_oops");
  assert(Universe::is_reserved_heap(oop(begin)), "reference must be into the heap");
  assert(Universe::is_reserved_heap(oop(end-1)), "limit must be within the heap");
  assert(_next_offset_threshold == bottom() + _next_offset_index*RememberedSet::card_size_in_oops, "index must agree with threshold");

  if (_next_offset_threshold >= begin) {
    // Common case
    _offset_array[_next_offset_index++] = _next_offset_threshold - begin;  // set current index (can be zero)
    _next_offset_threshold += RememberedSet::card_size_in_oops;            // update threshold
  } else {
    // Uncommon case. This can happen when aligning for double/long arrays, and the previous
    // end was on a card boundary. If the next object is aligned begin will be one word 
    // higher than the threshold.
    ShouldNotReachHere();
     assert(_next_offset_threshold+1 == begin, "threshold check failed");
    _offset_array[_next_offset_index++] = RememberedSet::card_size_in_oops;
    _next_offset_threshold += RememberedSet::card_size_in_oops;
  }
  while (_next_offset_threshold < end) {
    // fill in sentinel values if object spans more than one card
    _offset_array[_next_offset_index++] = RememberedSet::card_size_in_oops;
    _next_offset_threshold += RememberedSet::card_size_in_oops;
  }
  return _next_offset_threshold;
}


void OldSpace::update_offset_array() {
  // Traverse objects allocated since last update of _last_top and update offset array
  debug_only(oop *prev_last_offset_top = NULL);
  debug_only(int last_size = 0);
  while (_last_offset_top < top()) {
    assert(oop(_last_offset_top)->is_oop(), "check for header");
    debug_only(prev_last_offset_top = _last_offset_top);
    debug_only(last_size = oop(_last_offset_top)->size());
    oop* obj_end = _last_offset_top + oop(_last_offset_top)->size();
    check_offset_array(_last_offset_top, obj_end);
    _last_offset_top = obj_end;
  }
  assert(_last_offset_top == top(), "_last_top should be top");
}


oop OldSpace::fast_object_start(oop* p) {
  if (p >= _last_offset_top) update_offset_array();
  assert(p <= _last_offset_top, "offset array should be up-to-date for p");
  // Find the page start
  size_t b = (size_t) p;
  b &= ~(right_n_bits(RememberedSet::card_shift));
  oop* q = (oop*) b;
  assert(p>=q && q-p < RememberedSet::card_size_in_oops, "q must be close to p");
  assert(contains(q), "q must be in this space");
  int index  = ((size_t)q - (size_t)bottom()) / RememberedSet::card_size;
  assert(index >= 0, "just checking");

  int offset = _offset_array[index--];
  assert(offset >= 0, "just checking");
  while(offset == RememberedSet::card_size_in_oops) {
    q -= RememberedSet::card_size_in_oops;
    offset = _offset_array[index--];
    assert(offset >= 0, "just checking");
  }
  q -= offset;
  oop* n = q;

  assert(oop(n)->is_oop(), "check for header");
  while (n <= p) {
    q = n;
    n += oop(n)->size();
  }
  assert(oop(q)->is_oop(), "check for header");
  return oop(q);
}


OldSpace::OldSpace(u_char* offset_array, oop* bottom, oop* end) {
  assert(bottom < end, "Illegal space limits");
  _offset_array = offset_array;
  initialize(bottom, end, true);
}


bool OldSpace::scavenge_contents_from(WaterMark* mark) {
  assert(this == mark->space(), "Mark does not match space");
  debug_only(if (TraceScavenge)
    tty->print_cr("scavenging OldSpace 0x%x from 0x%x to 0x%x", bottom(), mark->point(), top())
  );
  oop* p = mark->point();
  oop* old_top = top();
  while (p < top()) {
    assert(oop(p)->safepoint_is_oop(),"check for header");
    oop* p_end = p + oop(p)->scavenge_contents();
    check_offset_array(p, p_end);
    p = p_end;
  }
  assert(p == top(), "p should be top");
  _last_offset_top = top();
  mark->set_point(top());
  return old_top != top();
}


void OldSpace::scavenge_recorded_stores() {
  RememberedSet::scavenge_contents(this);
}


#ifdef PRODUCT
void OldSpace::verify(bool allow_dirty) {}
#else
class VerifyOldOopClosure : public OopClosure {
 public:
  oop the_obj;
  bool allow_dirty;
  void do_oop(oop* p) {
    the_obj->verify_old_oop(p, allow_dirty);
  }
};


void OldSpace::verify(bool allow_dirty) {
  oop* p = bottom();
  oop* prev_p = NULL;
  VerifyOldOopClosure blk;
  blk.allow_dirty = allow_dirty;
  while (p < top()) {
    int size = oop(p)->size();
    guarantee(p == (oop*)fast_object_start(p + (size/2)), 
      "check offset computation");
    oop(p)->verify();
    blk.the_obj = oop(p);
    oop(p)->oop_iterate(&blk);
    prev_p = p;
    p += size;
  }
  guarantee(p == top(), "end of last object must match end of space");
  guarantee(_last_offset_top == top() || oop(_last_offset_top)->safepoint_is_oop(), "check for header");
}
#endif


TenuredSpace::TenuredSpace(u_char* offset_array, oop* bottom, oop* end) : OldSpace(offset_array, bottom, end) {
}


Space* TenuredSpace::next_compaction_space() const {
  return Universe::new_gen()->eden();
}


int TenuredSpace::allowed_dead_ratio() const {
  return MarkSweepDeadRatio;
}


PermSpace::PermSpace(u_char* offset_array, oop* bottom, oop* end) : OldSpace(offset_array, bottom, end) {
}


Space* PermSpace::next_compaction_space() const {
  return NULL;
}


int PermSpace::allowed_dead_ratio() const {
  return PermMarkSweepDeadRatio;
}


int CarSpace::_car_size = 0;
int CarSpace::_car_size_in_oops = 0;


CarSpace::CarSpace(u_char* offset_array, oop* bottom, oop* end, int blocks) : OldSpace(offset_array, bottom, end) {
  assert(((int) end - (int) bottom) % car_size() == 0, "Illegal car limits");
  assert(((int) end - (int) bottom) / car_size() == blocks, "Illegal car limits");
  set_next_car(NULL);
  set_blocks(blocks);
  set_remembered_set(new CarRememberedSet());
  set_desc(CarTable::desc_for(bottom));
  set_marked(false);
}


CarSpace::~CarSpace() {
  delete _remembered_set;
}


Space* CarSpace::next_compaction_space() const {
  return next_car();
}


void CarSpace::initialize() {
  _car_size         = 1 << LogOfCarSpaceSize;
  _car_size_in_oops = _car_size / oopSize;
  int page_size = os::vm_page_size();
  assert(page_size > 0, "page size should be set");
  if (_car_size < page_size || _car_size % os::vm_page_size() != 0) {
    vm_exit_during_initialization("Invalid car size");
  }
}

#ifndef PRODUCT

void CarSpace::verify(bool allow_dirty) {
  OldSpace::verify(allow_dirty);
  remembered_set()->verify(this);
  guarantee(desc()->space() == this, "invalid desc in CarSpace");
  guarantee(!marked(), "CarSpace should not be marked outside scavenge");
}

void CarSpace::print() {
  print_short();
  Space::print();
}

void CarSpace::print_short() {
  tty->print(" car (");
  tty->print_julong(train_number());
  tty->print(",%u,%uK)", car_number(), capacity() / K);
}

#endif // PRODUCT

void CarSpace::set_should_scavenge(bool value) {
  int size = blocks();
  CarTableDesc* d = desc();
  while (size-- > 0) {
    d->set_should_scavenge(value);
    d++;
  } 
}


bool CarSpace::commit() {
  return os::commit_memory((char*)bottom(), sizeof(*bottom()) * (end() - bottom()));
}


void CarSpace::uncommit() {
  os::uncommit_memory((char*)bottom(), sizeof(*bottom()) * (end() - bottom()));
  debug_only(os::guard_memory((char*)bottom(), sizeof(*bottom()) * (end() - bottom())));
}


void carSpace_init() {
  CarSpace::initialize();
}
