/*
 * @(#)space.hpp	1.72 00/10/07 11:46:42
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

class Space;
class Train;


// A water mark points into a space and is used during GC to keep track of progress.

class WaterMark VALUE_OBJ_CLASS_SPEC {
  friend class VMStructs;
 private:
  oop*    _point;
  Space*  _space;
 public:
  // Accessors
  Space* space() const        { return _space;  }
  void set_space(Space* s)    { _space = s;     }
  oop* point() const          { return _point;  }
  void set_point(oop* p)      { _point = p;     }

  // Constructors
  WaterMark(Space* s, oop* p) {
    set_space(s);
    set_point(p);
  }
  WaterMark() {
    set_space(NULL);
    set_point(NULL);
  }
};


inline bool operator==(const WaterMark& x, const WaterMark& y) {
  return (x.point() == y.point()) && (x.space() == y.space());
}

inline bool operator!=(const WaterMark& x, const WaterMark& y) {
  return !(x == y);
}


// A space describes a contiguous heap area. Class Space is an abstract base class.
//
// Allocation, size computation and GC support is provided.
//
// Invariant: bottom() and end() are on page_size boundaries and bottom() <= top() <= end()
// top() is inclusive and end() is exclusive.

class Space: public CHeapObj {
  friend class VMStructs;
 private:
  oop*  _compaction_top;

  // Temporary used during mark-sweep
  oop* compaction_top() const            { return _compaction_top;    }
  void set_compaction_top(oop* value)    {
    assert(value == NULL || (value >= bottom() && value <= end()),
	   "should point inside space");
    _compaction_top = value;
  }

 protected:
  oop*  _bottom;
  oop*  _top;
  oop*  _end;
  oop*  _first_dead;

  // Next space for full mark-sweep-compact
  virtual Space* next_compaction_space() const = 0;

 public:
  // Accessors
  oop* bottom() const         { return _bottom; }
  oop* top() const            { return _top;    }
  oop* end() const            { return _end;    }

  void set_bottom(oop* value) { _bottom = value; }
  void set_top(oop* value)    { _top    = value; }
  void set_end(oop* value)    { _end    = value; }

  WaterMark bottom_mark()     { return WaterMark(this, bottom()); }
  WaterMark top_mark()        { return WaterMark(this, top()); }
  
  // Initialization
  void initialize(oop* bottom, oop* end, bool clear_space);
  void clear();
  void mangle_unused_area();  // for detecting GC bugs

  // percentage of dead space allowed before we start compacting
  virtual int allowed_dead_ratio() const { return 0; }
  // Initialization of threshold used for computing object_start fast, see oldSpace 
  virtual oop* initialize_threshold() { return end(); }
  // Update offset array for object crossing threshold
  virtual oop* update_offset_array_entry(oop* begin, oop* end) {
    ShouldNotReachHere();
    return NULL;
  }

  // Testers
  bool is_empty() const           { return used() == 0; }
  bool not_empty() const          { return used() > 0; }
  bool contains(void* p) const    { return (oop*) p >= bottom() && (oop*) p < top(); }
  virtual bool is_old() const     { return false; }

  // Test whether p is aligned 0 mod 8.
  static bool is_aligned(oop* p)  { return ((int)p & (2*wordSize - 1)) == 0; }

  // Size computations
  size_t capacity() const        { return byte_size(bottom(), end()); }
  size_t used() const            { return byte_size(bottom(), top()); }
  size_t free() const            { return byte_size(top(),    end()); }

  // Scavenge support
  void allocate_temporary_filler(int factor); // factor of 0 means entire space

  // MarkSweep support phase2
  void prepare_for_compaction();
  // MarkSweep support phase3
  void adjust_pointers();
  // MarkSweep support phase4
  void compact();

  virtual void check_offset_array_during_compaction(oop* begin, oop* end) {}
  virtual void set_top_after_compaction(oop* value);

  // Iteration
  void object_iterate(ObjectClosure* blk);
  void object_iterate_from(WaterMark* mark, ObjectClosure* blk);

  // Find start of object containing p - very slow. See fast_object_start in OldSpace.
  oop slow_object_start(oop* p);

  // Allocation (return NULL if full)
  oop* allocate(int size) {
    assert(no_gc_in_progress(), "should not be called during gc");
    do {
      oop* obj = top();
      oop* new_top = obj + size;
      // The 'new_top>obj' check is needed if size is large enough to make 'obj+size' overflow
      if ((new_top > obj) && (new_top <= end())) {
        intptr_t result = atomic::compare_and_exchange_ptr((intptr_t*) top_addr(), (intptr_t) new_top, (intptr_t) obj);
        // result can be one of two:
        //  the old top value: the exchange succeeded
        //  otherwise: the new value of the top is returned.
        if (result != (intptr_t) obj) {          
          continue; // another thread beat us to the allocation, try again
        }
        assert(!ALIGN_ALL_OBJECTS || (is_aligned(obj) && is_aligned(new_top)), "checking alignment");
        return obj;
      } else {
        return NULL;
      }
    } while (true);
  }

  // Allocation at safepoint, not MT safe (return NULL if full)
  oop* safepoint_allocate(int size) {
    assert(at_safepoint(), "should be called at safepoint only");
    oop* obj = top();
    oop* new_top = obj + size;

    if (PrefetchCopyIntervalInBytes != 0) {
       atomic::prefetch_write(((char*)obj) + PrefetchCopyIntervalInBytes);
    }

    // No wrap around check needed, we assume object is already allocated in the heap
    if (new_top <= end()) {
      set_top(new_top);
      assert(!ALIGN_ALL_OBJECTS || (is_aligned(obj) && is_aligned(new_top)), "checking alignment");
      return obj;
    } else {
      return NULL;
    }
  }

  // Addresses for inlined allocation
  oop** top_addr() { return &_top; }
  oop** end_addr() { return &_end; }

  // Debugging
  virtual void print()                  PRODUCT_RETURN;
  virtual void print_short()            PRODUCT_RETURN;
  virtual void verify(bool allow_dirty) PRODUCT_RETURN;
  bool at_safepoint() const             PRODUCT_RETURN0;
  bool no_gc_in_progress() const        PRODUCT_RETURN0;
};



// Class NewSpace is an abstract base class for new generation spaces.

class NewSpace: public Space {
  friend class VMStructs;
};


// Class EdenSpace describes eden-space in new generation.

class EdenSpace : public NewSpace {
  friend class VMStructs;
 private:
  // When entering a GC safepoint the eden _top may be set past _end
  // by the inline allocation code. If this is the case, _proper_top
  // contains the correct value for _top. GC will call ensure_proper_top
  // before starting a collection.
  oop*  _proper_top;

 protected:
  Space* next_compaction_space() const;

 public:
  // Accessors
  oop* top() const            { 
    assert(_proper_top == 0, "_proper_top not reset");
    return _top;    }
  oop* top_no_check() const             { return _top;    }
  oop* proper_top() const               { return _proper_top;    }
  void set_proper_top(oop* value)       { _proper_top = value; }

  // Size computations
  size_t used() const;
  size_t free() const;

  // Update top and proper_top pointers
  void ensure_proper_top();

  // Debugging
  virtual void verify(bool allow_dirty) PRODUCT_RETURN;
};


// Class SurvivorSpace describes from- and to-space in new generation.

class SurvivorSpace : public NewSpace {
  friend class VMStructs;
 protected:
  Space* next_compaction_space() const;

 public:
  // Scavenge support
  bool scavenge_contents_from(WaterMark* mark);
};


// Class OldSpace is an abstract base class for old generation (train, tenured, perm) spaces.

class OldSpace: public Space {
 friend class TrainGeneration;
  friend class VMStructs;
 private:
  // Array for keeping offsets for retrieving object start fast given a card_page
  // The array is not updated at allocation time, but before any GC happens and during mark-sweep.
  u_char* _offset_array;         // byte array keeping backwards offsets
  oop*    _next_offset_threshold; // allocation boundary for which offset array must be updated
  int     _next_offset_index;    // index corresponding to boundary
  oop*    _last_offset_top;      // last value of _top for which offset array was up-to-date

  // Update offset array for object crossing threshold
  oop* update_offset_array_entry(oop* begin, oop* end);
  // Check if offset array should be updated for object
  void check_offset_array(oop* begin, oop* end) {
    if (end > _next_offset_threshold) update_offset_array_entry(begin, end);
  }

 public:
  // Reset offset array
  oop* initialize_threshold();
  // Compute missing offset array part prior to GC
  void update_offset_array();

  // Compute start of object containing p
  oop fast_object_start(oop* p);

  // Test
  bool is_old() const { return true; }

  // Constructor
  OldSpace(u_char* offset_array, oop* bottom, oop* end);

  // Scavenge support
  bool scavenge_contents_from(WaterMark* mark);

  void scavenge_recorded_stores();

  // MarkSweep support phase3
  void check_offset_array_during_compaction(oop* begin, oop* end);
  void set_top_after_compaction(oop* value);

  // Debugging
  void verify(bool allow_dirty);
};


// Class TenuredSpace is used by TenuredGeneration

class TenuredSpace: public OldSpace {
  friend class VMStructs;
 protected:
  // Mark sweep support
  Space* next_compaction_space() const;
  int allowed_dead_ratio() const;
 public:
  // Constructor
  TenuredSpace(u_char* offset_array, oop* bottom, oop* end);
};


// Class PermSpace is used by PermGeneration

class PermSpace: public OldSpace {
  friend class VMStructs;
 protected:
  // Mark sweep support
  Space* next_compaction_space() const;
  int allowed_dead_ratio() const;
 public:
  // Constructor
  PermSpace(u_char* offset_array, oop* bottom, oop* end);
};


// Class CarSpace describes an individual car in a train

class CarSpace: public OldSpace {
 friend class TrainGeneration;
  friend class VMStructs;
 private:
  CarSpace*         _next_car;
  CarRememberedSet* _remembered_set;
  int               _blocks;
  CarTableDesc*     _desc;
  bool              _marked;
  static int        _car_size;
  static int        _car_size_in_oops;

  void set_blocks(int n)                        { _blocks = n; }
  void set_remembered_set(CarRememberedSet* rs) { _remembered_set = rs; }
  void set_desc(CarTableDesc* desc)             { _desc = desc; }

 protected:
  Space* next_compaction_space() const;

 public:
  // Sizes
  static int car_size()                     { return _car_size; }
  static int car_size_in_oops()             { return _car_size_in_oops; }
  // Next car in train
  CarSpace* next_car() const                { return _next_car; }	
  void set_next_car(CarSpace* next)         { _next_car = next; }	
  // Local remembered set
  CarRememberedSet* remembered_set() const  { return _remembered_set; }
  // Car table description corresponding to this car
  CarTableDesc* desc() const                { return _desc; }
  // Number of 2^LogOfCarSpaceSize blocks allocated for this car
  int blocks() const                        { return _blocks; }
  // Special marking bit (support for special large object handling)
  bool marked() const                       { return _marked; }
  // Train containing this car
  Train* train() const                      { return desc()->train(); }
  // Number assigned to this car
  juint car_number() const                  { return desc()->car_number(); }
  // Number assigned to containing train
  julong train_number() const               { return desc()->train_number(); }

  // Scavenge support
  void set_should_scavenge(bool value);
  void set_marked(bool value)               { _marked = value; }

  // Constructor
  CarSpace(u_char* offset_array, oop* bottom, oop* end, int blocks);
  ~CarSpace();

  // Debugging
  void verify(bool allow_dirty) PRODUCT_RETURN;
  void print()                  PRODUCT_RETURN;
  void print_short()            PRODUCT_RETURN;
  
  // commit/uncommit
  bool commit();
  void uncommit();

  // Initialization
  static void initialize();
};
