/*
 * @(#)symbolTable.hpp	1.25 00/03/21 16:17:04
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// The symbol table holds all symbolOops and corresponding interned strings.
// symbolOops and literal strings should be canonicalized.
//
// The interned strings are created lazily.
//
// It is implemented as an open hash table with a fixed number of buckets.
//
// %note:
//  - symbolTableEntrys are allocated in blocks to reduce the space overhead.

class symbolTableBucket: public CHeapObj {
 private:
  // Instance variable
  symbolOop _entry;
 public:
  // Accessing
  void clear()                { _entry = NULL; }
  symbolOop entry() const     { return _entry; }
  void set_entry(symbolOop l) { _entry = l; }  
  symbolOop* entry_addr()     { return &_entry; }

  // Lookup name in bucket
  symbolOop lookup(const char* name, int len);

  // Debugging
  void verify(int index)  PRODUCT_RETURN;
  int  length()           PRODUCT_RETURN0;
};

class SymbolTable: public AllStatic {
 public:
  // Table size
  enum {
    symbol_table_size = 20011
  };
  // Initialization
  static void initialize();
  // Symbol lookup
  static symbolOop lookup(const char* name, int len, TRAPS);

  //%%TEMPORARY FUNCTION, needed for preloading classes in signatures when compiling. 
  // Returns the symbol is already present in symbol table, otherwise NULL
  // NO ALLOCATION IS GUARANTEED!
  static symbolOop probe(char* name, int len);

  // GC support
  static void follow_roots(); 
  static void adjust_pointers();
  // Histogram
  static void print_histogram()     PRODUCT_RETURN;
  // Debugging
  static void verify()              PRODUCT_RETURN;
  static void print()               PRODUCT_RETURN;

 private:
  // Instance variables
  static symbolTableBucket* _buckets;

  // Adding elements    
  static symbolOop basic_add(u1* name, int len, int hashValue, TRAPS);

  // Bucket handling
  static symbolTableBucket* bucketFor(int hashValue) {
    assert(hashValue >= 0 && hashValue < symbol_table_size, "illegal hash value");
    return &_buckets[hashValue]; 
  }
  static symbolTableBucket* firstBucket() { return &_buckets[0]; }
  static symbolTableBucket* lastBucket()  { return &_buckets[symbol_table_size-1]; }
};

class stringTableEntry: public CHeapObj {
 private:
  oop _literal_string;        // corresponding java.lang.String instance
  stringTableEntry* _next;    // link to next element in bucket
 public:
  stringTableEntry* next() const        { return _next; }
  void set_next(stringTableEntry* next) { _next = next; }

  stringTableEntry** next_addr()        { return &_next; }

  // Literal string
  oop literal_string() { return _literal_string; }
  void set_literal_string(oop s) { _literal_string = s; }

  void adjust_pointers() {
    MarkSweep::adjust_root_pointer((oop*) &_literal_string);
  }
  void follow_literal_string();

  // Debugging
  void verify(int index)          PRODUCT_RETURN;   
};


class stringTableBucket: public CHeapObj {
 private:
  // Instance variable
  stringTableEntry* _entry;
 public:
  // Accessing
  void clear()                        { _entry = NULL; }
  stringTableEntry* entry() const     { return _entry; }
  void set_entry(stringTableEntry* l) { _entry = l; }  

  stringTableEntry** entry_addr()     { return &_entry; }

  // Lookup name in bucket
  oop lookup(jchar* str, int len);

  // Debugging
  void verify(int index)  PRODUCT_RETURN;
};


class StringTable: public AllStatic {
 private:
  static oop intern(Handle string_or_null, jchar* chars, int length, TRAPS);
  static oop basic_add(Handle string_or_null, jchar* name, int len, int hashValue, TRAPS);

 public:
  // Table size
  enum {
    string_table_size = 1009
  };
  // Initialization
  static void initialize();

  // Probing
  static oop lookup(symbolOop symbol);

  // Interning
  static oop intern(symbolOop symbol, TRAPS);
  static oop intern(oop string, TRAPS);
  static oop intern(char *utf8_string, TRAPS);

  // GC support
  static void follow_roots(); 
  static void adjust_pointers();

  // Debugging
  static void verify()              PRODUCT_RETURN;
  static void print()               PRODUCT_RETURN;

 private:
  // Instance variables
  static stringTableBucket buckets[string_table_size];
  static stringTableEntry* free_list;
  static stringTableEntry* first_free_entry;
  static stringTableEntry* end_block;

  // Bucket handling
  static stringTableBucket* bucketFor(int hashValue) {
    assert(hashValue>=0 && hashValue<string_table_size, "illegal hash value");
    return &buckets[hashValue]; 
  }
  static stringTableBucket* firstBucket() { return &buckets[0]; }
  static stringTableBucket* lastBucket()  { return &buckets[string_table_size-1]; }
};
