/*
 * @(#)systemDictionary.cpp	1.226 01/02/09
 *
 * Copyright 1993-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_systemDictionary.cpp.incl"


class ProtectionDomainEntry :public CHeapObj {
 public: 
  ProtectionDomainEntry* _next;
  oop                    _protection_domain;

  ProtectionDomainEntry(oop protection_domain, ProtectionDomainEntry* next) {
    _protection_domain = protection_domain;
    _next              = next;
  }
};

class SystemDictionaryEntry : public CHeapObj {
 private:
  // Contains the set of approved protection domains that can access
  // this system dictionary entry.
  ProtectionDomainEntry* _pd_set;

 public:
  // Tells whether a protection is in the approved set.
  bool contains_protection_domain(oop protection_domain) const;
  // Adds a protection domain to the approved set.
  void add_protection_domain(oop protection_domain);
 
 public:
  SystemDictionaryEntry* _next;
  oop                    _klass;
  oop                    _loader;

  SystemDictionaryEntry(SystemDictionaryEntry* next, oop klass, oop loader) {
    _next   = next;
    _klass  = klass;
    _loader = loader;
    _pd_set = NULL; // Initially no protection domain has been approved to load this entry
  }

  ~SystemDictionaryEntry() {
    // avoid recursion when deleting linked list
    while (_pd_set != NULL) {
      ProtectionDomainEntry* to_delete = _pd_set;
      _pd_set = _pd_set->_next;
      delete to_delete;
    }
  }

  // Tells whether the initiating class' protection can access the this _klass
  bool is_valid_protection_domain(Handle protection_domain) {
    if (!ProtectionDomainVerification) return true;
    if (!SystemDictionary::has_checkPackageAccess()) return true;

    return protection_domain() == NULL
         ? true
         : contains_protection_domain(protection_domain());
  }


  void protection_domain_set_oops_do(void f(oop*)) {
    for (ProtectionDomainEntry* current = _pd_set; current != NULL; current = current->_next) {
      f(&(current->_protection_domain));
    }
  }

  void verify_protection_domain_set() {
    for (ProtectionDomainEntry* current = _pd_set; current != NULL; current = current->_next) {
      current->_protection_domain->verify();
    }
  }

  void print() {
    int count = 0;
    for (ProtectionDomainEntry* current = _pd_set; current != NULL; current = current->_next) {
      count++;
    }
    tty->print_cr("pd set = #%d", count);
  }
};

bool SystemDictionaryEntry::contains_protection_domain(oop protection_domain) const {
#ifdef ASSERT
  if (protection_domain == instanceKlass::cast(klassOop(_klass))->protection_domain()) {
    // Ensure this doesn't show up in the pd_set (invariant)
    bool in_pd_set = false;
    for (ProtectionDomainEntry* current = _pd_set; current != NULL; current = current->_next) {
      if (current->_protection_domain == protection_domain) {
	in_pd_set = true;
	break;
      }
    }
    if (in_pd_set) {
      assert(false, "A klass's protection domain should not show up in its sys. dict. PD set");
    }
  }
#endif /* ASSERT */

  if (protection_domain == instanceKlass::cast(klassOop(_klass))->protection_domain()) {
    // Succeeds trivially
    return true;
  }

  for (ProtectionDomainEntry* current = _pd_set; current != NULL; current = current->_next) {
    if (current->_protection_domain == protection_domain) return true;
  }
  return false;
}

void SystemDictionaryEntry::add_protection_domain(oop protection_domain) {
  if (!contains_protection_domain(protection_domain)) {
    _pd_set = new ProtectionDomainEntry(protection_domain, _pd_set);
  }
  if (TraceProtectionDomainVerification && WizardMode) {
    print();
  }
}

class LoaderConstraintEntry : public CHeapObj {
 public:
  LoaderConstraintEntry* _next;
  klassOop               _klass;                  // initially NULL
  symbolOop              _name;                   // class name
  int                    _num_loaders;
  int                    _max_loaders;
  oop*                   _loaders;                // initiating loaders
};


SystemDictionaryEntry** SystemDictionary::_buckets             = NULL;
int                     SystemDictionary::_number_of_classes   =    0;
int                     SystemDictionary::_number_of_adds      =    0;

SystemDictionaryEntry*  SystemDictionary::_current_class_entry = NULL;
int                     SystemDictionary::_current_class_index =    0;

LoaderConstraintEntry** SystemDictionary::_loader_constraints  = NULL;

oop         SystemDictionary::_system_loader_lock_obj     =  NULL;

klassOop    SystemDictionary::_object_klass               =  NULL;
klassOop    SystemDictionary::_string_klass               =  NULL;
klassOop    SystemDictionary::_class_klass                =  NULL;
klassOop    SystemDictionary::_cloneable_klass            =  NULL;
klassOop    SystemDictionary::_classloader_klass          =  NULL;
klassOop    SystemDictionary::_serializable_klass         =  NULL;

klassOop    SystemDictionary::_throwable_klass            =  NULL;
klassOop    SystemDictionary::_error_klass                =  NULL;
klassOop    SystemDictionary::_threaddeath_klass          =  NULL;
klassOop    SystemDictionary::_exception_klass            =  NULL;
klassOop    SystemDictionary::_runtime_exception_klass    =  NULL;
klassOop    SystemDictionary::_classNotFoundException_klass = NULL;
klassOop    SystemDictionary::_noClassDefFoundError_klass = NULL;
klassOop    SystemDictionary::_classCastException_klass   =  NULL;
klassOop    SystemDictionary::_arrayStoreException_klass  =  NULL;
klassOop    SystemDictionary::_outOfMemoryError_klass     =  NULL;
klassOop    SystemDictionary::_StackOverflowError_klass   =  NULL;
klassOop    SystemDictionary::_protectionDomain_klass     =  NULL;
klassOop    SystemDictionary::_AccessControlContext_klass = NULL;

klassOop    SystemDictionary::_reference_klass            =  NULL;
klassOop    SystemDictionary::_soft_reference_klass       =  NULL;
klassOop    SystemDictionary::_weak_reference_klass       =  NULL;
klassOop    SystemDictionary::_final_reference_klass      =  NULL;
klassOop    SystemDictionary::_phantom_reference_klass    =  NULL;
klassOop    SystemDictionary::_finalizer_klass            =  NULL;

klassOop    SystemDictionary::_thread_klass               =  NULL;
klassOop    SystemDictionary::_threadGroup_klass          =  NULL;
klassOop    SystemDictionary::_properties_klass           =  NULL;
klassOop    SystemDictionary::_reflect_method_klass       =  NULL;
klassOop    SystemDictionary::_reflect_constructor_klass  =  NULL;

klassOop    SystemDictionary::_vector_klass               =  NULL;
klassOop    SystemDictionary::_hashtable_klass            =  NULL;
klassOop    SystemDictionary::_stringBuffer_klass         =  NULL;


klassOop    SystemDictionary::_boolean_klass              =  NULL;
klassOop    SystemDictionary::_char_klass                 =  NULL;
klassOop    SystemDictionary::_float_klass                =  NULL;
klassOop    SystemDictionary::_double_klass               =  NULL;
klassOop    SystemDictionary::_byte_klass                 =  NULL;
klassOop    SystemDictionary::_short_klass                =  NULL;
klassOop    SystemDictionary::_int_klass                  =  NULL;
klassOop    SystemDictionary::_long_klass                 =  NULL;

oop         SystemDictionary::_int_mirror                 =  NULL;
oop         SystemDictionary::_float_mirror               =  NULL;
oop         SystemDictionary::_double_mirror              =  NULL;
oop         SystemDictionary::_byte_mirror                =  NULL;
oop         SystemDictionary::_bool_mirror                =  NULL;
oop         SystemDictionary::_char_mirror                =  NULL;
oop         SystemDictionary::_long_mirror                =  NULL;
oop         SystemDictionary::_short_mirror               =  NULL;
oop         SystemDictionary::_void_mirror                =  NULL;

oop         SystemDictionary::_java_system_loader         =  NULL;

bool        SystemDictionary::_has_loadClassInternal      =  false;
bool        SystemDictionary::_has_checkPackageAccess     =  false;

// ------------------------------------------------------------------------------------------------------------
// Java-level SystemLoader

oop SystemDictionary::java_system_loader() {
  return _java_system_loader;
}

void SystemDictionary::compute_java_system_loader(TRAPS) {
  KlassHandle system_klass(_classloader_klass);    
  JavaValue result(T_OBJECT);
  JavaCalls::call_static(&result, 
                         _classloader_klass, 
                         vmSymbolHandles::getSystemClassLoader_name(),
                         vmSymbolHandles::void_classloader_signature(),
                         CHECK);
    
  _java_system_loader = (oop)result._value.h;    
}

// ------------------------------------------------------------------------------------------------------------
// Hashing

inline unsigned int SystemDictionary::hash_value(symbolHandle class_name, Handle class_loader) {  
  return (unsigned int) class_name->identity_hash()
       ^ (unsigned int) (class_loader.is_null() ? 0 : class_loader->identity_hash());
}

inline int SystemDictionary::index_for(symbolHandle class_name, Handle class_loader) {
  int hash = hash_value(class_name, class_loader) % _nof_buckets;  
  return hash;
}

// ------------------------------------------------------------------------------------------------------------
// Methods for adding and allocation of SystemDictionary

// adds a new klass or placeholder
void SystemDictionary::add_klass_or_placeholder(symbolOop class_name, oop class_loader, oop obj) {
  assert_lock(SystemDictionary_lock);
  assert(obj != NULL, "adding NULL obj");
  debug_only(if (obj->is_symbol()) assert(class_name == obj, "sanity check on name");)
  debug_only(if (obj->is_klass()) assert(Klass::cast(klassOop(obj))->name() == class_name, "sanity check on name");)
  int index = index_for(class_name, class_loader);
  _buckets[index] = new SystemDictionaryEntry(_buckets[index], obj, class_loader);
  _number_of_classes++;
}

// replaces an entry
void SystemDictionary::remove_placeholder(symbolOop class_name, oop class_loader, klassOop klass) {
  assert_lock(SystemDictionary_lock);
  int index = index_for(class_name, class_loader);
  SystemDictionaryEntry** p = &_buckets[index];
  while (*p) {
    SystemDictionaryEntry *probe = *p;
    if (probe->_klass == class_name && probe->_loader == class_loader) {
      if (klass == NULL) {
        *p = probe->_next;
        if (probe == _current_class_entry) {
          _current_class_entry = NULL;
        }
        delete probe;
        _number_of_classes--;
        continue;
      } else {
        assert(Klass::cast(klass)->name() == class_name, "sanity check on name");
        probe->_klass = klass;
        return; 
      }
    }
    p = &(probe->_next);
  }
}

//-----------------------------------------------------------------------------------------------------------------
// debugging

#ifdef ASSERT

// return true if class_name contains no '.' (internal format is '/')
static bool is_internal_format(symbolHandle class_name) {
  if (class_name.not_null()) {
    ResourceMark rm;
    char* name = class_name->as_C_string();
    return strchr(name, '.') == NULL;
  } else {
    return true;
  }
}

#endif

//-----------------------------------------------------------------------------------------------------------------
// Resolving of classes

// Forwards to resolve_or_null

klassOop SystemDictionary::resolve_or_fail(symbolHandle class_name, Handle class_loader, Handle protection_domain, bool throw_error, TRAPS) {  
  klassOop klass = resolve_or_null(class_name, class_loader, protection_domain, THREAD);
  if (HAS_PENDING_EXCEPTION) {
    // If we have a pending exception we forward it to the caller, unless throw_error is true,
    // in which case we have to check whether the pending exception is a ClassNotFoundException,
    // and if so convert it to a NoClassDefFoundError
    if (throw_error && PENDING_EXCEPTION->is_a(SystemDictionary::classNotFoundException_klass())) {
      // Convert ClassNotFoundException to NoClassDefFoundError (by falling through)
      assert(klass == NULL, "Should not have result with exception pending");
      CLEAR_PENDING_EXCEPTION;
    } else {
      return NULL;  // Forward exception to caller
    }
  }
  // Class not found, throw appropriate error or exception depending on value of throw_error
  if (klass == NULL) {
    ResourceMark rm;
    if (throw_error) {
      THROW_MSG_0(vmSymbols::java_lang_NoClassDefFoundError(), class_name->as_C_string());
    } else {      
      THROW_MSG_0(vmSymbols::java_lang_ClassNotFoundException(), class_name->as_C_string());      
    }
  }
  return klass;
}


// Forwards to resolve_instance_class_or_null

klassOop SystemDictionary::resolve_or_null(symbolHandle class_name, Handle class_loader, Handle protection_domain, TRAPS) {  
  assert(is_internal_format(class_name), "external class name format used internally");
  if (FieldType::is_array(class_name())) {
    return resolve_array_class_or_null(class_name, class_loader, protection_domain, CHECK_0);
  } else {
    return resolve_instance_class_or_null(class_name, class_loader, protection_domain, CHECK_0);
  }
}


// Forwards to resolve_instance_class_or_null

klassOop SystemDictionary::resolve_array_class_or_null(symbolHandle class_name, Handle class_loader, Handle protection_domain, TRAPS) {  
  assert(FieldType::is_array(class_name()), "must be array");
  jint dimension;
  symbolOop object_key;
  klassOop k = NULL;  
  // dimension and object_key are assigned as a side-effect of this call
  BasicType t = FieldType::get_array_info(class_name(), &dimension, &object_key, CHECK_0);
  if (t == T_OBJECT) {
    k = SystemDictionary::resolve_instance_class_or_null(object_key, class_loader, protection_domain, CHECK_0);
    if (k != NULL) {
      k = Klass::cast(k)->array_klass(dimension, CHECK_0);
    }
  } else {
    switch (t) {
      case T_BOOLEAN: k = Universe::boolArrayKlassObj();    break;
      case T_CHAR:    k = Universe::charArrayKlassObj();    break;
      case T_FLOAT:   k = Universe::singleArrayKlassObj();  break;
      case T_DOUBLE:  k = Universe::doubleArrayKlassObj();  break;
      case T_BYTE:    k = Universe::byteArrayKlassObj();    break;
      case T_SHORT:   k = Universe::shortArrayKlassObj();   break;
      case T_INT:     k = Universe::intArrayKlassObj();     break;
      case T_LONG:    k = Universe::longArrayKlassObj();    break;
      default:        ShouldNotReachHere();
    }
    k = typeArrayKlass::cast(k)->array_klass(dimension, CHECK_0);
  }
  return k;
}

void SystemDictionary::validate_protection_domain(instanceKlassHandle klass, Handle class_loader, Handle protection_domain, TRAPS) {
  if(!has_checkPackageAccess()) return;

  // Now we have to call back to java to check if the initating class has access
  JavaValue result(T_VOID);
  if (TraceProtectionDomainVerification) {
    // Print out trace information
    tty->print_cr("Checking package access");
    tty->print(" - class loader:      "); class_loader()->print_value_on(tty);      tty->cr();
    tty->print(" - protection domain: "); protection_domain()->print_value_on(tty); tty->cr();
    tty->print(" - loading:           "); klass()->print_value_on(tty);             tty->cr();
  }
  
  assert(class_loader() != NULL, "should not have non-null protection domain for null classloader");

  JavaCalls::call_special(&result,
                         class_loader(),
                         SystemDictionary::classloader_klass(),
                         vmSymbolHandles::checkPackageAccess_name(),
                         vmSymbolHandles::class_protectiondomain_signature(), 
                         klass->java_mirror(),
                         protection_domain(),
                         THREAD);

  if (TraceProtectionDomainVerification) {
    if (HAS_PENDING_EXCEPTION) {
      tty->print_cr(" -> DENIED !!!!!!!!!!!!!!!!!!!!!");
    } else {
     tty->print_cr(" -> granted");
    }
    tty->cr();
  }

  if (HAS_PENDING_EXCEPTION) return; 
    
  // If no exception has been thrown, we have validated the protection domain
  // Insert the protection domain of the initiaing class into the set.
  {
    MutexLocker mu(SystemDictionary_lock);
    SystemDictionaryEntry* entry = get_entry(klass->name(), class_loader());
    assert(entry != NULL,"entry must be present,we just created added it");
    assert(protection_domain() != NULL, "real protection domain should be present");
    entry->add_protection_domain(protection_domain());
    assert(entry->contains_protection_domain(protection_domain()), "now protection domain should be present");
  }
}

klassOop SystemDictionary::resolve_instance_class_or_null(symbolHandle class_name, Handle class_loader, Handle protection_domain, TRAPS) {
  assert(class_name.not_null() && !FieldType::is_array(class_name()), "invalid class name");
  // First check to see if we should remove wrapping L and ;
  symbolHandle name;    
  if (FieldType::is_obj(class_name())) {
    ResourceMark rm;
    // Ignore wrapping L and ;.
    name = oopFactory::new_symbol(class_name()->as_C_string() + 1, class_name()->utf8_length() - 2, CHECK_0);    
  } else {
    name = class_name;
  }

  // Do lookup to see if class already exist and the protection domain has the right access
  klassOop probe = find(name, class_loader, protection_domain);
  if (probe != NULL) return probe;

  // Class is not is SystemDictionary so we have to do loading.
  // Make sure we are synchronized on the class loader before we proceed
  ObjectLocker ol(compute_loader_lock_object(class_loader), THREAD);

  // Check again (after locking) if class already exist in SystemDictionary
  bool class_has_been_loaded   = false;
  bool throw_circularity_error = false;
  instanceKlassHandle k        = klassOop(NULL);

  {           
    MutexLocker mu(SystemDictionary_lock);  
    oop check = find_class_or_placeholder(name, class_loader);
    if (check != NULL) {
      // Found klass or place holder
      if (check->is_klass()) {
        // Klass is already loaded, so just return it
        class_has_been_loaded = true;
        k = klassOop(check);
      } else {
        // Being loaded by this loader (since we have locked the loader object)
        assert(check->is_symbol() && check == name(), "should be symbolOop place holder");
        throw_circularity_error = true;
      }
    } else {
      // Nothing found, add place holder
      add_klass_or_placeholder(name(), class_loader(), name());
    }
  }

  // Throw error now if needed (cannot throw while holding SystemDictionary_lock)
  if (throw_circularity_error) {
   ResourceMark rm;
   THROW_MSG_0(vmSymbols::java_lang_ClassCircularityError(), name->as_C_string());
  }

  if (!class_has_been_loaded) {
    // Do actual loading
    k = load_instance_class(name, class_loader, THREAD);
    if (!HAS_PENDING_EXCEPTION && !k.is_null() && k->class_loader() != class_loader()) {
      check_constraints(k, class_loader, THREAD);
      { // Grabbing the Compile_lock prevents systemDictionary updates during compilations. 
        MutexLocker mu(Compile_lock);      
        update_dictionary(k, class_loader);
      }
      if (jvmdi::enabled() && !HAS_PENDING_EXCEPTION) {
        Thread *thread = THREAD;
        assert(thread->is_Java_thread(), "thread->is_Java_thread()");
        jvmdi::post_class_load_event((JavaThread *) thread, k());
      }
      // notify jvmpi
      if (jvmpi::is_event_enabled(JVMPI_EVENT_CLASS_LOAD) && k->class_loader() == class_loader()) {
        jvmpi::post_class_load_event(Klass::cast(k())->java_mirror());
      }
    }

    // %%Note: This jvmdi code may need some attention

  
    if (HAS_PENDING_EXCEPTION || k.is_null()) {
      // No success, remove place holder and return
      MutexLocker mu(SystemDictionary_lock);  
      remove_placeholder(name(), class_loader(), NULL);
        return NULL;
    }
  }

#ifdef ASSERT
  {
    MutexLocker mu(SystemDictionary_lock);  
    oop kk = find_class_or_placeholder(name(), k->class_loader());    
    assert(kk == k(), "should be present in dictionary");
  }
#endif

  // return if the protection domain in NULL
  if (protection_domain() == NULL) return k();

  // Check the protection domain has the right access 
  {
    MutexLocker mu(SystemDictionary_lock);  
    SystemDictionaryEntry* entry = get_entry(name(), class_loader());
    if (entry->is_valid_protection_domain(protection_domain)) return k();
  }

  // Verify protection domain. If it fails an exception is thrown
  validate_protection_domain(k, class_loader, protection_domain, CHECK_(klassOop(NULL)));

  return k();
}

klassOop SystemDictionary::find(symbolHandle class_name, Handle class_loader, Handle protection_domain) {
  MutexLocker mu(SystemDictionary_lock);
  SystemDictionaryEntry* entry = get_entry(class_name(), class_loader());
  if (entry == NULL)              return NULL;
  if (!entry->_klass->is_klass()) return NULL;
  return entry->is_valid_protection_domain(protection_domain)
       ? klassOop(entry->_klass)
       : NULL;
}

// Look for a loaded instance or array klass by name.  Do not do any loading.
// return NULL in case of error.
klassOop SystemDictionary::find_instance_or_array_klass(symbolHandle class_name,
                                                        Handle class_loader,
							Handle protection_domain,
                                                        TRAPS) {
  klassOop k = NULL;
  assert(class_name() != NULL, "class name must be non NULL");
  if (FieldType::is_array(class_name())) {
    // The name refers to an array.  Parse the name.
    jint dimension;
    symbolOop object_key;

    // dimension and object_key are assigned as a side-effect of this call
    BasicType t = FieldType::get_array_info(class_name(), &dimension,
					    &object_key, CHECK_(NULL));
    switch (t) {
      case T_BOOLEAN: k = Universe::boolArrayKlassObj();    break;
      case T_CHAR:    k = Universe::charArrayKlassObj();    break;
      case T_FLOAT:   k = Universe::singleArrayKlassObj();  break;
      case T_DOUBLE:  k = Universe::doubleArrayKlassObj();  break;
      case T_BYTE:    k = Universe::byteArrayKlassObj();    break;
      case T_SHORT:   k = Universe::shortArrayKlassObj();   break;
      case T_INT:     k = Universe::intArrayKlassObj();     break;
      case T_LONG:    k = Universe::longArrayKlassObj();    break;
      case T_OBJECT:  k = SystemDictionary::find(object_key, class_loader, protection_domain);
	              break;
      default:        ShouldNotReachHere();
    }
    if (k != NULL) {
      k = Klass::cast(k)->array_klass_or_null(dimension);
    }
  } else {
    k = find(class_name, class_loader, protection_domain);
  }
  return k;
}


// Add a klass to the system from a stream (called by jni_DefineClass)
// Note: class_name can be NULL. In that case we do not know the name of the class until we have parsed the stream.

klassOop SystemDictionary::resolve_from_stream(symbolHandle class_name, Handle class_loader, Handle protection_domain, ClassFileStream* st, TRAPS) {
  // Make sure we are synchronized on the class loader before we initiate loading
  ObjectLocker ol(compute_loader_lock_object(class_loader), THREAD);

  // Parse the stream. Note that we do this even though this klass might already be present 
  // in the SystemDictionary, otherwise we would not throw potential ClassFormatErrors
  instanceKlassHandle k = ClassFileParser(st).parseClassFile(class_name, class_loader, protection_domain, CHECK_0);

  // We are sure to know the name by now
  symbolHandle name;
  if (class_name.is_null()) {
    name = symbolHandle(THREAD, k->name());
  } else {
    name = class_name;
  }
  assert(is_internal_format(name), "external class name format used internally");
  
  // Add class just loaded
  define_instance_class(k, CHECK_0);  

  return k();
}

instanceKlassHandle SystemDictionary::load_instance_class(symbolHandle class_name, Handle class_loader, TRAPS) {
  if (class_loader.is_null()) {
    // Use VM class loader
    instanceKlassHandle k = ClassLoader::load_classfile(class_name, CHECK_(klassOop(NULL)));
    if (!k.is_null()) {
      define_instance_class(k, CHECK_(klassOop(NULL)));
    }
    return k;
  } else {
    // Use user specified class loader to load class. Call loadClass operation on class_loader.
    ResourceMark rm;
      
    oop string_oop = java_lang_String::create_from_symbol(class_name, CHECK_(klassOop(NULL)));
    // Translate to external class name format, i.e., convert '/' chars to '.'
    Handle string = java_lang_String::externalize_classname(string_oop, CHECK_(klassOop(NULL)));    

    JavaValue result(T_OBJECT);

    if (has_loadClassInternal()) {
      JavaCalls::call_special(&result, 
                              class_loader, 
                              SystemDictionary::classloader_klass(),
                              vmSymbolHandles::loadClassInternal_name(),
                              vmSymbolHandles::string_class_signature(), 
                              string,
                              CHECK_(klassOop(NULL)));      
    } else {
      JavaCalls::call_virtual(&result, 
                              class_loader, 
                              SystemDictionary::classloader_klass(),
                              vmSymbolHandles::loadClass_name(),
                              vmSymbolHandles::string_class_signature(), 
                              string,
                              CHECK_(klassOop(NULL)));      
    }

    assert(result._type == T_OBJECT, "just checking");
    oop obj = (oop) result._value.h;

    if (obj != NULL) {      
      return instanceKlassHandle(java_lang_Class::as_klassOop(obj));
    } else {      
      return instanceKlassHandle(klassOop(NULL));
    }
  }
}

void SystemDictionary::define_instance_class(instanceKlassHandle k, TRAPS) {

  // Check class-loading constraints. Throw exception if violation is detected.
  // Graps SystemDictionary_lock
  check_constraints(k, k->class_loader(), CHECK);

  // Add the new class. We need recompile lock during update of CHA.
  { MutexLocker mu_r(Compile_lock);                    

    // Add to class hierarchy, initialize vtables, and do possible deoptimizations
    add_to_hierarchy(k, CHECK); // No exception, but can block

    // Add to systemDictionary - so other classes can see it.
    // Graps SystemDictionary_lock
    update_dictionary(k, k->class_loader());
  }

  k->eager_initialize();

  // Register class just loaded with class loader (placed in Vector)
  if (k->class_loader() != NULL) {
    methodHandle m(THREAD, Universe::loader_addClass_method());
    Handle l(THREAD, k->class_loader());
    JavaValue result(T_VOID);
    JavaCallArguments args(l);
    args.push_oop(k->java_mirror());
    JavaCalls::call(&result, m, &args, CHECK);
  }

  // notify jvmdi
  if (jvmdi::enabled()) {
      assert(THREAD->is_Java_thread(), "thread->is_Java_thread()");
      jvmdi::post_class_load_event((JavaThread *) THREAD, k());

  }
  // notify jvmpi
  if (jvmpi::is_event_enabled(JVMPI_EVENT_CLASS_LOAD)) {
    jvmpi::post_class_load_event(Klass::cast(k())->java_mirror());
  }
}

Handle SystemDictionary::compute_loader_lock_object(Handle class_loader) {
  // If class_loader is NULL we synchronize on _system_loader_lock_obj
  if (class_loader.is_null()) {
    return Handle(_system_loader_lock_obj);
  } else {
    return class_loader;
  }
}


//-----------------------------------------------------------------------------------------------------------------
// debugging

#ifdef ASSERT
static int lookup_count  = 0;
static int lookup_length = 0;

static void verify_lookup_length(float load) {
  if ((float)lookup_length / lookup_count > load * 2) {
    warning("Performance bug: SystemDictionary lookup_count=%d lookup_length=%d average=%lf load=%f", 
      lookup_count, lookup_length, (double) lookup_length / lookup_count, load);
  }
}

#endif


//-----------------------------------------------------------------------------------------------------------------
// Lookup

oop SystemDictionary::find_class_or_placeholder(symbolHandle class_name, Handle class_loader) {
  assert(is_internal_format(class_name), "external class name format used internally");
  assert_lock(SystemDictionary_lock);
  int index = index_for(class_name, class_loader);
  symbolOop class_name_ = class_name();
  oop class_loader_ = class_loader();

  debug_only(lookup_count++);
  for (SystemDictionaryEntry *probe = _buckets[index]; probe; probe = probe->_next) {
    // It could be a temporary place holder
    oop this_klass = probe->_klass;
    oop this_loader = probe->_loader;
    if (this_klass->is_klass()) {
      if (instanceKlass::cast(klassOop(this_klass))->name() == class_name_ && this_loader == class_loader_)
        return this_klass;
    } else {
      assert(this_klass->is_symbol(), "must be place holder for class");
      if (this_klass == class_name_ && this_loader == class_loader_)
        return this_klass;
    }
    debug_only(lookup_length++);
  }
  return NULL;
}

klassOop SystemDictionary::try_get_next_class() {
  while (true) {
    if (_current_class_entry != NULL) {
      oop k = _current_class_entry->_klass;
      _current_class_entry = _current_class_entry->_next;
      if (k->is_klass()) {
        return klassOop(k);
      } else {
        continue;
      }
    }
    _current_class_index = (_current_class_index + 1) % _nof_buckets;
    _current_class_entry = _buckets[_current_class_index];
  }
  // never reached
}

SystemDictionaryEntry* SystemDictionary::get_entry(symbolOop class_name, oop class_loader) {
  int index = index_for(class_name, class_loader);
  for (SystemDictionaryEntry* entry = _buckets[index]; entry != NULL; entry = entry->_next) {
    if (   entry->_loader == class_loader
        && entry->_klass->is_klass()
        && instanceKlass::cast(klassOop(entry->_klass))->name() == class_name) {
      return entry;
    }
  }
  return NULL;
}

void* SystemDictionary::get_classID(klassOop k) {
  symbolOop class_name = instanceKlass::cast(k)->name();
  oop class_loader = instanceKlass::cast(k)->class_loader();
  int index = index_for(class_name, class_loader);

  for (SystemDictionaryEntry *probe = _buckets[index]; probe; probe = probe->_next) {
    if (klassOop(probe->_klass) == k) {
      return probe;
    }
  }
  return NULL;
}

//------------------------------------------------------------------------------------------------------------------
// Update hierachy. This is done before the new klass has been added to the SystemDictionary. The Recompile_lock
// is held, to ensure that the compiler is not using the class hierachy, and that deoptimization will kick in
// before a new class is used.

void SystemDictionary::add_to_hierarchy(instanceKlassHandle k, TRAPS) {
  assert(k.not_null(), "just checking");
  // Link into hierachy. Make sure the vtables are initialized before linking into 
  k->append_to_sibling_list();                    // add to superklass/sibling list
  k->process_interfaces();                        // handle all "implements" declarations  
  k->set_init_state(instanceKlass::loaded);
  // Now flush all code that depended on old class hierarchy.
  // Note: must be done *after* linking k into the hierarchy (was bug 12/9/97)
  // Also, first reinitialize vtable because it may have gotten out of synch 
  // while the new class wasn't connected to the class hierarchy.     
  NOT_CORE(Universe::flush_dependents_on(k());)
}


//-----------------------------------------------------------------------------------------------------------------
// GC support

// Following roots during mark-sweep is separated in two phases. 
//
// The first phase follows preloaded classes and all other system 
// classes, since these will never get unloaded anyway.
//
// The second phase removes (unloads) unreachable classes from the
// system dictionary and follows the remaining classes' contents.

void SystemDictionary::follow_roots_phase1() {
  // Follow preloaded classes/mirrors and system loader object
  MarkSweep::follow_root(&_java_system_loader);
  preloaded_oops_do(&MarkSweep::follow_root);
  // Follow all system classes and temporary placeholders in dictionary
  for (int index = 0; index < _nof_buckets; index++) {
    for (SystemDictionaryEntry *probe = _buckets[index]; probe; probe = probe->_next) {
      oop e = probe->_klass;
      oop class_loader = probe->_loader;            
      if (class_loader == NULL || e->is_symbol() || !ClassUnloading) {
        MarkSweep::follow_root(&(probe->_klass));
        if (class_loader != NULL) {
          MarkSweep::follow_root(&(probe->_loader));
        }
        probe->protection_domain_set_oops_do(&MarkSweep::follow_root);
      }
    }
  }
}


bool SystemDictionary::follow_roots_phase2() {
  bool purged_class = false;
  int  index = 0; // Defined here for portability! Do not move

  // Remove unloadable entries and classes from system dictionary
  for (index = 0; index < _nof_buckets; index++) {
    SystemDictionaryEntry** p = &_buckets[index];
    while (*p) {
      SystemDictionaryEntry* probe = *p;
      oop e = probe->_klass;
      oop class_loader = probe->_loader;

      if (class_loader != NULL && !e->is_symbol() && ClassUnloading) {
        // Entry was not visited in phase1 (negated test from phase1)
        assert(e->is_klass(), "should be klass");
        if (!class_loader->is_gc_marked()) {
          // If the loader is not reachable this entry should always be removed 
          // (will never be looked up again). Note that this is not the same as
          // unloading the referred class.
          instanceKlass* k = instanceKlass::cast(klassOop(e));
          if (k->class_loader() == class_loader) {
            // This is the defining entry, so the referred class is about to be unloaded.
            guarantee(!e->is_gc_marked(), "klass should not be marked if defining loader is not");
            purged_class = true;
            // notify the debugger
            if (jvmdi::enabled()) {
              jvmdi::post_class_unload_event(k->as_klassOop());
            }
            // Cannot post CLASS_UNLOAD event from here because JVM/PI 1.X
	    // requires that the event be posted with GC disabled. Since we
	    // are part of GC we just have to save the necessary info for
	    // the post.
	    if (jvmpi::is_event_enabled(JVMPI_EVENT_CLASS_UNLOAD)) {
	      jvmpi::save_class_unload_event_info(k->java_mirror());
	    }
            // Clean up C heap.
            k->release_C_heap_structures();
            if (TraceClassUnloading) {
              ResourceMark rm;
              tty->print_cr("[Unloading class %s]", k->external_name());
            }
          }
          // Purge entry
          *p = probe->_next;
          if (probe == _current_class_entry) {
            _current_class_entry = NULL;
          }
          delete probe;
          _number_of_classes--;
          continue;
        } else {
          // Update entry
          MarkSweep::follow_root(&(probe->_klass));
          MarkSweep::follow_root(&(probe->_loader));
          probe->protection_domain_set_oops_do(&MarkSweep::follow_root);
        }
      }
      p = &(probe->_next);
    }
  }

  // Remove unloaded entries from constraint table
  for (index = 0; index < _loader_constraint_size; index++) {
    LoaderConstraintEntry** p = &_loader_constraints[index];
    while(*p) {
      LoaderConstraintEntry* probe = *p;
      klassOop klass = probe->_klass;
      symbolOop name = probe->_name;
      // Remove unmarked name
      if (!name->is_gc_marked()) {
        probe->_name = NULL;
      }
      // Remove unmarked klass
      if (klass != NULL && !klass->is_gc_marked()) {
        probe->_klass = NULL;
      }
      // Remove unmarked entries from loader array
      int n = 0; 
      while (n < probe->_num_loaders) {
        if (probe->_loaders[n] != NULL) {
          if (!probe->_loaders[n]->is_gc_marked()) {
            // Compact array
            probe->_num_loaders--;
            probe->_loaders[n] = probe->_loaders[probe->_num_loaders];
            probe->_loaders[probe->_num_loaders] = NULL;
            continue;  // current element replaced, so restart without incrementing n
          }
        }
        n++;
      }
      // Check whether entry should be purged
      if (probe->_name == NULL || probe->_num_loaders < 2) {
        // Purge entry
        *p = probe->_next;
        FREE_C_HEAP_ARRAY(oop, probe->_loaders);
        delete probe;
      } else {
        // Otherwise follow contents of entry. Note that we stricly don't have to do this
        // since everything is known to be marked, but is doesn't cost anything and makes
        // the ValidateMarkSweep code work correctly.
        assert(probe->_name->is_gc_marked(), "name should be marked");
        MarkSweep::follow_root((oop*)&(probe->_name));
        if (probe->_klass != NULL) {
          assert(probe->_klass->is_gc_marked(), "klass should be marked");
          MarkSweep::follow_root((oop*)&(probe->_klass));
        }
        for (n = 0; n < probe->_num_loaders; n++) {
          if (probe->_loaders[n] != NULL) {
            assert(probe->_loaders[n]->is_gc_marked(), "loader should be marked");
            MarkSweep::follow_root(&(probe->_loaders[n]));
          }
        }
        // Go to next entry
        p = &(probe->_next);
      }
    }
  }
  return purged_class;
}


void SystemDictionary::adjust_pointers() {
  int index = 0; // Defined out here for portability

  // Adjust preloaded classes/mirrors and system loader object
  MarkSweep::adjust_root_pointer(&_java_system_loader);
  preloaded_oops_do(MarkSweep::adjust_root_pointer);
  // Adjust dictionary
  for (index = 0; index < _nof_buckets; index++) {
    for (SystemDictionaryEntry* probe = _buckets[index]; probe; probe = probe->_next) {
      MarkSweep::adjust_root_pointer(&(probe->_klass));
      if (probe->_loader != NULL) {
        MarkSweep::adjust_root_pointer(&(probe->_loader));
      }
      probe->protection_domain_set_oops_do(&MarkSweep::adjust_root_pointer);
    }
  }
  // Adjust constraint table
  for (index = 0; index < _loader_constraint_size; index++) {
    for (LoaderConstraintEntry* probe = _loader_constraints[index]; probe; probe = probe->_next) {
      MarkSweep::adjust_root_pointer((oop*)&(probe->_name));
      if (probe->_klass != NULL) {
        MarkSweep::adjust_root_pointer((oop*)&(probe->_klass));
      }
      for (int n = 0; n < probe->_num_loaders; n++) {
        if (probe->_loaders[n] != NULL) {
          MarkSweep::adjust_root_pointer(&(probe->_loaders[n]));
        }
      }
    }
  }
}


void SystemDictionary::young_oops_do(void f(oop*)) {
  int index = 0; // Defined out here for portability
  f((oop*) &_java_system_loader);
  // Iterate over loaders in dictionary
  for (index = 0; index < _nof_buckets; index++) {
    for (SystemDictionaryEntry* probe = _buckets[index]; probe; probe = probe->_next) {
      if (probe->_loader != NULL) {
        f(&(probe->_loader));
      }
      probe->protection_domain_set_oops_do(f);
    }
  }
  // Iterate over loaders in constraint table
  for (index = 0; index < _loader_constraint_size; index++) {
    for (LoaderConstraintEntry* probe = _loader_constraints[index]; probe; probe = probe->_next) {
      for (int n = 0; n < probe->_num_loaders; n++) {
        if (probe->_loaders[n] != NULL) {
          f(&(probe->_loaders[n]));
        }
      }
    }
  }
}


void SystemDictionary::preloaded_oops_do(void f(oop*)) {
  f((oop*) &_string_klass);
  f((oop*) &_object_klass);
  f((oop*) &_class_klass);
  f((oop*) &_cloneable_klass);
  f((oop*) &_classloader_klass);
  f((oop*) &_serializable_klass);

  f((oop*) &_throwable_klass);
  f((oop*) &_error_klass);
  f((oop*) &_threaddeath_klass);
  f((oop*) &_exception_klass);
  f((oop*) &_runtime_exception_klass);
  f((oop*) &_classNotFoundException_klass);
  f((oop*) &_noClassDefFoundError_klass);
  f((oop*) &_classCastException_klass);
  f((oop*) &_arrayStoreException_klass);
  f((oop*) &_outOfMemoryError_klass);
  f((oop*) &_StackOverflowError_klass);
  f((oop*) &_protectionDomain_klass);
  f((oop*) &_AccessControlContext_klass);

  f((oop*) &_reference_klass);
  f((oop*) &_soft_reference_klass);
  f((oop*) &_weak_reference_klass);
  f((oop*) &_final_reference_klass);
  f((oop*) &_phantom_reference_klass);
  f((oop*) &_finalizer_klass);
  
  f((oop*) &_thread_klass);
  f((oop*) &_threadGroup_klass);
  f((oop*) &_properties_klass);      
  f((oop*) &_reflect_method_klass);      
  f((oop*) &_reflect_constructor_klass);      

  f((oop*) &_stringBuffer_klass);
  f((oop*) &_vector_klass);
  f((oop*) &_hashtable_klass);

  f((oop*) &_boolean_klass);
  f((oop*) &_char_klass);
  f((oop*) &_float_klass);
  f((oop*) &_double_klass);
  f((oop*) &_byte_klass);
  f((oop*) &_short_klass);
  f((oop*) &_int_klass);
  f((oop*) &_long_klass);

  f((oop*) &_int_mirror);
  f((oop*) &_float_mirror);
  f((oop*) &_double_mirror);
  f((oop*) &_byte_mirror);
  f((oop*) &_bool_mirror);
  f((oop*) &_char_mirror);
  f((oop*) &_long_mirror);
  f((oop*) &_short_mirror);
  f((oop*) &_void_mirror);

  f((oop*) &_system_loader_lock_obj);
}

//   Just the classes from declaring class loaders
void SystemDictionary::classes_do(void f(klassOop)) {
  for (int index = 0; index < _nof_buckets; index++) {
    for (SystemDictionaryEntry* probe = _buckets[index]; probe; probe = probe->_next) {
      oop k = probe->_klass;
      if ((k->is_klass()) && 
          (probe->_loader == instanceKlass::cast(klassOop(k))->class_loader())) {
        f(klassOop(k));
      }
    }
  }
}

//   All classes, and their class loaders
void SystemDictionary::classes_do(void f(klassOop, oop)) {
  for (int index = 0; index < _nof_buckets; index++) {
    for (SystemDictionaryEntry* probe = _buckets[index]; probe; probe = probe->_next) {
      oop k = probe->_klass;
      if (k->is_klass()) {
        f(klassOop(k), probe->_loader);
      }
    }
  }
}

void SystemDictionary::methods_do(void f(methodOop)) {
  for (int index = 0; index < _nof_buckets; index++) {
    for (SystemDictionaryEntry* probe = _buckets[index]; probe; probe = probe->_next) {
      oop k = probe->_klass;
      if (k->is_klass()) {
        if (probe->_loader == instanceKlass::cast(klassOop(k))->class_loader()) {
          // only take klass is we have the entry with the defining class loader
          instanceKlass::cast(klassOop(k))->methods_do(f);
        }
      } else {
        // If not a klassOop it must be a placeholder
        assert(k->is_symbol(), "system dictionary inconsistency");
      }
    }
  }
}


// ------------------------------------------------------------------------------------------------------------
// Initialization

void SystemDictionary::initialize(TRAPS) {
  // Allocate array
  assert(_buckets == NULL, "SystemDictionary should only be initialized once");
  _buckets = NEW_C_HEAP_ARRAY(SystemDictionaryEntry*, _nof_buckets);
  memset(_buckets, 0, _nof_buckets * sizeof(SystemDictionaryEntry*));
  _number_of_classes = 0;
  _number_of_adds = 0;
  _loader_constraints = NEW_C_HEAP_ARRAY(LoaderConstraintEntry*, _loader_constraint_size);
  memset(_loader_constraints, 0, _loader_constraint_size * sizeof(LoaderConstraintEntry *));
  // Allocate private object used as system class loader lock
  _system_loader_lock_obj = oopFactory::new_system_objArray(0, CHECK);
  // Initialize basic classes
  initialize_preloaded_classes(CHECK);
  initialize_basic_type_mirrors(CHECK);
}


void SystemDictionary::initialize_preloaded_classes(TRAPS) {
  assert(_object_klass == NULL, "preloaded classes should only be initialized once");
  // Preload commonly used klasses
  _object_klass            = resolve_or_fail(vmSymbolHandles::java_lang_Object(),               NULL, NULL, true, CHECK);
  _string_klass            = resolve_or_fail(vmSymbolHandles::java_lang_String(),               NULL, NULL, true, CHECK);  
  _class_klass             = resolve_or_fail(vmSymbolHandles::java_lang_Class(),                NULL, NULL, true, CHECK);
  debug_only(instanceKlass::verify_class_klass_nonstatic_oop_maps(_class_klass));
  // Fixup mirrors for classes loaded before java.lang.Class
  Universe::fixup_mirrors(CHECK);

  _cloneable_klass         = resolve_or_fail(vmSymbolHandles::java_lang_Cloneable(),            NULL, NULL, true, CHECK);
  _classloader_klass       = resolve_or_fail(vmSymbolHandles::java_lang_ClassLoader(),          NULL, NULL, true, CHECK);
  _serializable_klass      = resolve_or_fail(vmSymbolHandles::java_io_Serializable(),           NULL, NULL, true, CHECK);

  _throwable_klass         = resolve_or_fail(vmSymbolHandles::java_lang_Throwable(),            NULL, NULL, true, CHECK);
  _error_klass             = resolve_or_fail(vmSymbolHandles::java_lang_Error(),                NULL, NULL, true, CHECK);
  _threaddeath_klass       = resolve_or_fail(vmSymbolHandles::java_lang_ThreadDeath(),          NULL, NULL, true, CHECK);
  _exception_klass         = resolve_or_fail(vmSymbolHandles::java_lang_Exception(),            NULL, NULL, true, CHECK);
  _runtime_exception_klass = resolve_or_fail(vmSymbolHandles::java_lang_RuntimeException(),     NULL, NULL, true, CHECK);
  _protectionDomain_klass  = resolve_or_fail(vmSymbolHandles::java_security_ProtectionDomain(), NULL, NULL, true, CHECK);
  _AccessControlContext_klass = resolve_or_fail(vmSymbolHandles::java_security_AccessControlContext(), NULL, NULL, true, CHECK);
  _classNotFoundException_klass = resolve_or_fail(vmSymbolHandles::java_lang_ClassNotFoundException(), NULL, NULL, true, CHECK);
  _noClassDefFoundError_klass   = resolve_or_fail(vmSymbolHandles::java_lang_NoClassDefFoundError(), NULL, NULL, true, CHECK);  
  _classCastException_klass = resolve_or_fail(vmSymbolHandles::java_lang_ClassCastException(),  NULL, NULL, true, CHECK);  
  _arrayStoreException_klass = resolve_or_fail(vmSymbolHandles::java_lang_ArrayStoreException(),  NULL, NULL, true, CHECK);  
  _outOfMemoryError_klass  = resolve_or_fail(vmSymbolHandles::java_lang_OutOfMemoryError(),     NULL, NULL, true, CHECK);  
  _StackOverflowError_klass = resolve_or_fail(vmSymbolHandles::java_lang_StackOverflowError(),  NULL, NULL, true, CHECK);  

  // Preload ref klasses and set reference types
  _reference_klass         = resolve_or_fail(vmSymbolHandles::java_lang_ref_Reference(),        NULL, NULL, true, CHECK);
  instanceKlass::cast(_reference_klass)->set_reference_type(REF_OTHER);
  instanceRefKlass::update_nonstatic_oop_maps(_reference_klass);

  _soft_reference_klass    = resolve_or_fail(vmSymbolHandles::java_lang_ref_SoftReference(),    NULL, NULL, true, CHECK);
  instanceKlass::cast(_soft_reference_klass)->set_reference_type(REF_SOFT);
  _weak_reference_klass    = resolve_or_fail(vmSymbolHandles::java_lang_ref_WeakReference(),    NULL, NULL, true, CHECK);
  instanceKlass::cast(_weak_reference_klass)->set_reference_type(REF_WEAK);
  _final_reference_klass   = resolve_or_fail(vmSymbolHandles::java_lang_ref_FinalReference(),   NULL, NULL, true, CHECK);
  instanceKlass::cast(_final_reference_klass)->set_reference_type(REF_FINAL);
  _phantom_reference_klass = resolve_or_fail(vmSymbolHandles::java_lang_ref_PhantomReference(), NULL, NULL, true, CHECK);
  instanceKlass::cast(_phantom_reference_klass)->set_reference_type(REF_PHANTOM);
  _finalizer_klass         = resolve_or_fail(vmSymbolHandles::java_lang_ref_Finalizer(),        NULL, NULL, true, CHECK);

  _thread_klass           = resolve_or_fail(vmSymbolHandles::java_lang_Thread(),               NULL, NULL, true, CHECK);
  _threadGroup_klass      = resolve_or_fail(vmSymbolHandles::java_lang_ThreadGroup(),          NULL, NULL, true, CHECK);
  _properties_klass       = resolve_or_fail(vmSymbolHandles::java_util_Properties(),           NULL, NULL, true, CHECK);  
  _reflect_method_klass   = resolve_or_fail(vmSymbolHandles::java_lang_reflect_Method(),       NULL, NULL, true, CHECK);  
  _reflect_constructor_klass = resolve_or_fail(vmSymbolHandles::java_lang_reflect_Constructor(),  NULL, NULL, true, CHECK);  

  _vector_klass           = resolve_or_fail(vmSymbolHandles::java_util_Vector(),               NULL, NULL, true, CHECK);  
  _hashtable_klass        = resolve_or_fail(vmSymbolHandles::java_util_Hashtable(),            NULL, NULL, true, CHECK);  
  _stringBuffer_klass     = resolve_or_fail(vmSymbolHandles::java_lang_StringBuffer(),         NULL, NULL, true, CHECK);  


  // Preload boxing klasses
  _boolean_klass           = resolve_or_fail(vmSymbolHandles::java_lang_Boolean(),              NULL, NULL, true, CHECK);
  _char_klass              = resolve_or_fail(vmSymbolHandles::java_lang_Character(),            NULL, NULL, true, CHECK);
  _float_klass             = resolve_or_fail(vmSymbolHandles::java_lang_Float(),                NULL, NULL, true, CHECK);
  _double_klass            = resolve_or_fail(vmSymbolHandles::java_lang_Double(),               NULL, NULL, true, CHECK);
  _byte_klass              = resolve_or_fail(vmSymbolHandles::java_lang_Byte(),                 NULL, NULL, true, CHECK);
  _short_klass             = resolve_or_fail(vmSymbolHandles::java_lang_Short(),                NULL, NULL, true, CHECK);
  _int_klass               = resolve_or_fail(vmSymbolHandles::java_lang_Integer(),              NULL, NULL, true, CHECK);
  _long_klass              = resolve_or_fail(vmSymbolHandles::java_lang_Long(),                 NULL, NULL, true, CHECK);

  { // Compute whether we should use loadClass or loadClassInternal when loading classes.
    methodOop method = instanceKlass::cast(classloader_klass())->find_method(vmSymbols::loadClassInternal_name(), vmSymbols::string_class_signature());
    _has_loadClassInternal = (method != NULL);
  }

  { // Compute whether we should use checkPackageAccess or NOT
    methodOop method = instanceKlass::cast(classloader_klass())->find_method(vmSymbolHandles::checkPackageAccess_name(), vmSymbolHandles::class_protectiondomain_signature());
    _has_checkPackageAccess = (method != NULL); 
  }
}

void SystemDictionary::initialize_basic_type_mirrors(TRAPS) { 
  assert(_int_mirror==NULL, "basic type mirrors already initialized");
  _int_mirror     = java_lang_Class::create_basic_type_mirror("int",    CHECK); 
  _float_mirror   = java_lang_Class::create_basic_type_mirror("float",  CHECK);
  _double_mirror  = java_lang_Class::create_basic_type_mirror("double", CHECK); 
  _byte_mirror    = java_lang_Class::create_basic_type_mirror("byte",   CHECK);
  _bool_mirror    = java_lang_Class::create_basic_type_mirror("bool",   CHECK);
  _char_mirror    = java_lang_Class::create_basic_type_mirror("char",   CHECK);
  _long_mirror    = java_lang_Class::create_basic_type_mirror("long",   CHECK);
  _short_mirror   = java_lang_Class::create_basic_type_mirror("short",  CHECK);
  _void_mirror    = java_lang_Class::create_basic_type_mirror("void",   CHECK);
}

// Constraints on class loaders. The details of the algorithm can be found in the OOPSLA'98 paper
// "Dynamic Class Loading in the Java Virtual Machine" by Sheng Liang and Gilad Bracha.
// The basic idea is that the system dictionary needs to maintain a set of contraints that must
// be satisfied by all classes in the dictionary.

void SystemDictionary::check_constraints(instanceKlassHandle k, Handle class_loader, TRAPS) {
  const char *linkage_error = NULL;
  {
    MutexLocker mu(SystemDictionary_lock);         
    symbolOop name = k->name();    

    oop loader = class_loader();
    oop check = find_class_or_placeholder(name, loader);
    if (check != NULL) {
      if (check->is_klass()) {
        linkage_error = "duplicate class definition: %s";
      } else {
        assert(check->is_symbol(), "expecting a symbol");
        assert(check == name, "invalid symbol");
      }
    }
    if (linkage_error == NULL) {
      LoaderConstraintEntry* p = *(find_loader_constraint(name, loader));
      
      if (p && p->_klass != NULL && p->_klass != k()) {
        linkage_error = "Class %s violates loader constraints";
      } else {
        if (p && p->_klass == NULL) {
          p->_klass = k();
        }
      }
    }

  }

  // Throw error now if needed (cannot throw while holding SystemDictionary_lock)
  if (linkage_error) {
    ResourceMark rm;
    char message[128];
    jio_snprintf(message, sizeof(message), linkage_error, k->name()->as_C_string());
    THROW_MSG(vmSymbols::java_lang_LinkageError(), message);
  }
}


// Update system dictionary - done after check_constraint and add_to_hierachy has been called.
void SystemDictionary::update_dictionary(instanceKlassHandle k, Handle class_loader) {    
  // Compile_lock prevents systemDictionary updates during compilations
  assert_lock(Compile_lock);
  MutexLocker mu1(SystemDictionary_lock);           
  symbolOop name = k->name();    
  oop loader = class_loader();

  oop check = find_class_or_placeholder(name, loader);
  if (check != NULL) {
    remove_placeholder(name, loader, k());
  } else {
    add_klass_or_placeholder(name, loader, k());
  }
  _number_of_adds++;
}

LoaderConstraintEntry** SystemDictionary::find_loader_constraint(symbolOop name, oop loader) {
  int index = (unsigned long)name->identity_hash() % _loader_constraint_size;
  LoaderConstraintEntry** pp = &_loader_constraints[index];
  while (*pp) {
    LoaderConstraintEntry* p = *pp;
    if (p->_name == name) {
      for (int i = p->_num_loaders - 1; i >= 0; i--) {
        if (p->_loaders[i] == loader) {
          return pp;
        }
      }
    }
    pp = &(p->_next);
  }
  return pp;
}

void SystemDictionary::ensure_loader_constraint_capacity(LoaderConstraintEntry *p, int nfree) {
    if (p->_max_loaders - p->_num_loaders < nfree) {
        int n = nfree + p->_num_loaders;
        oop* new_loaders = NEW_C_HEAP_ARRAY(oop, n);
        memcpy(new_loaders, p->_loaders, sizeof(oop) * p->_num_loaders);
        p->_max_loaders = n;
        FREE_C_HEAP_ARRAY(oop, p->_loaders);
        p->_loaders = new_loaders;
    }
}
 
void SystemDictionary::extend_loader_constraint(LoaderConstraintEntry* p, oop loader, klassOop klass) {
  ensure_loader_constraint_capacity(p, 1);
  p->_loaders[p->_num_loaders++] = loader;
  if (p->_klass == NULL) {
    p->_klass = klass;
  } else {
    assert(klass == NULL || p->_klass == klass, "constraints corrupted");
  }
}

void SystemDictionary::merge_loader_constraints(LoaderConstraintEntry** pp1, LoaderConstraintEntry** pp2, klassOop klass) {
  // make sure *pp1 has higher capacity 
  if ((*pp1)->_max_loaders < (*pp2)->_max_loaders) {
    LoaderConstraintEntry** tmp = pp2;
    pp2 = pp1;
    pp1 = tmp;
  }
  
  LoaderConstraintEntry* p1 = *pp1;
  LoaderConstraintEntry* p2 = *pp2;
  
  ensure_loader_constraint_capacity(p1, p2->_num_loaders);
  
  for (int i = 0; i < p2->_num_loaders; i++) {
    p1->_loaders[p1->_num_loaders++] = p2->_loaders[i];
  }
  
  // p1->_klass will hold NULL if klass, p2->_klass, and old p1->_klass are all NULL.
  // In addition, all three must have matching non-NULL values, otherwise either
  // the constraints would have been violated, or the constraints had been corrupted
  // (and an assertion would fail).
  if (p2->_klass) {
    assert(p2->_klass == klass, "constraints corrupted");
  }
  if (p1->_klass == NULL) {
    p1->_klass = klass;
  } else {
    assert(p1->_klass == klass, "constraints corrupted");
  }

  *pp2 = p2->_next;
  FREE_C_HEAP_ARRAY(oop, p2->_loaders);
  delete p2;
  return;
}

void SystemDictionary::add_loader_constraint(symbolHandle class_name, Handle class_loader1, Handle class_loader2, TRAPS) {
  bool linkage_error = false;
  {
    MutexLocker mu_s(SystemDictionary_lock);

    symbolOop name = class_name();
    oop loader1 = class_loader1();
    oop loader2 = class_loader2();

    oop k1 = find_class_or_placeholder(name, loader1);
    oop k2 = find_class_or_placeholder(name, loader2);

    klassOop klass1 = (k1 && k1->is_klass()) ? klassOop(k1) : NULL;
    klassOop klass2 = (k2 && k2->is_klass()) ? klassOop(k2) : NULL;    

    if (klass1 != NULL && klass2 != NULL && klass1 != klass2) {
      linkage_error = true;
    } else {
      klassOop klass = klass1 != NULL ? klass1 : klass2;
      
      LoaderConstraintEntry** pp1 = find_loader_constraint(name, loader1);
      if (*pp1 != NULL && (*pp1)->_klass != NULL) {
        if (klass != NULL) {
          if (klass != (*pp1)->_klass) {
            linkage_error = true;
          }
        } else {
          klass = (*pp1)->_klass;
        }
      }

      LoaderConstraintEntry** pp2 = find_loader_constraint(name, loader2);
      if (*pp2 != NULL && (*pp2)->_klass != NULL) {
        if (klass != NULL) {
          if (klass != (*pp2)->_klass) {
            linkage_error = true;
          }
        } else {
          klass = (*pp2)->_klass;
        }
      }

      if (!linkage_error) {
        if (*pp1 == NULL && *pp2 == NULL) {
          LoaderConstraintEntry* p = new LoaderConstraintEntry();
          p->_loaders = NEW_C_HEAP_ARRAY(oop, 2);
          p->_name = name;
          p->_loaders[0] = loader1;
          p->_loaders[1] = loader2;
          p->_num_loaders = 2;
          p->_max_loaders = 2;
          p->_klass = klass;
        
          int index = (unsigned long)name->identity_hash() % _loader_constraint_size;
          p->_next = _loader_constraints[index];
          _loader_constraints[index] = p;
        } else if (*pp1 == *pp2) {
          /* constraint already imposed */
          if ((*pp1)->_klass == NULL) {
            (*pp1)->_klass = klass;
          } else {
            assert((*pp1)->_klass == klass, "loader constraints corrupted");
          }
        } else if (*pp1 == NULL) {
          extend_loader_constraint(*pp2, loader1, klass);
        } else if (*pp2 == NULL) {
          extend_loader_constraint(*pp1, loader2, klass);
        } else {
	  merge_loader_constraints(pp1, pp2, klass);
        }
      }
    }
  }
  // Throw error now if needed (cannot throw while holding SystemDictionary_lock)
  if (linkage_error) {
    ResourceMark rm;
    char message[128];
    jio_snprintf(message, sizeof(message), "loader constraints violated when linking %s class", class_name()->as_C_string());
    THROW_MSG(vmSymbols::java_lang_LinkageError(), message);
  }
}

// Make sure all class components (including arrays) in the given
// signature will be resolved to the same class in both loaders.
//
void SystemDictionary::check_signature_loaders(symbolHandle signature, Handle loader1, Handle loader2, bool is_method, TRAPS)  {
  // Nothing to do if loaders are the same. 
  if (loader1() == loader2()) {
    return;
  }
  
  SignatureStream sig_strm(signature(), is_method);
  while (!sig_strm.is_done()) {
    symbolOop sig = sig_strm.as_symbol(CHECK);
    if (sig_strm.is_object()) {
      add_loader_constraint(sig, loader1, loader2, CHECK);
    }
    sig_strm.next();
  }
}

//---------------------------------------------------------------------------------
#ifndef PRODUCT

void SystemDictionary::print() {
  ResourceMark rm;
  HandleMark   hm;

  tty->print_cr("Java system dictionary (classes=%d)", _number_of_classes);
  tty->print_cr("^ indicates that initiating loader is different from defining loader");
  for (int index = 0; index < _nof_buckets; index++) {    
    for (SystemDictionaryEntry* probe = _buckets[index]; probe; probe = probe->_next) {
      oop e = probe->_klass;
      if (Verbose) tty->print("%4d: ", index);
      oop class_loader =  probe->_loader;
      if (e->is_klass()) {
        bool is_defining_class = (class_loader == instanceKlass::cast(klassOop(e))->class_loader());
        tty->print("%s%s", is_defining_class ? " " : "^", Klass::cast(klassOop(e))->external_name());
      } else {
        tty->print(" place holder ");
        e->print_value();
      }
      if (class_loader != NULL) {
        tty->print(", loader ");
        class_loader->print_value();
      }
      tty->cr();
    }
  }
}


void SystemDictionary::verify() {
  guarantee(_buckets != NULL,                      "Verify of system dictionary failed");
  guarantee(_number_of_classes >= 0,               "Verify of system dictionary failed");

  // Verify dictionary
  int element_count = 0;
  int index = 0; // Declared outside for portability
  for (index = 0; index < _nof_buckets; index++) {
    for (SystemDictionaryEntry* probe = _buckets[index]; probe; probe = probe->_next) {
      oop e = probe->_klass;
      oop class_loader = probe->_loader;
      if (e->is_klass()) {
        guarantee(Klass::cast(klassOop(e))->oop_is_instance(), "Verify of system dictionary failed");
      } else {
        guarantee(e->is_symbol(), "Verify of system dictionary failed");
      }
      // class loader must be present
      guarantee(class_loader == NULL || class_loader->is_instance(), "checking type of class_loader");
      e->verify();
      probe->verify_protection_domain_set();
      element_count++;  // both klasses and place holders count
    }
  }
  guarantee(_number_of_classes == element_count, "Verify of system dictionary failed");
  debug_only(verify_lookup_length((float)_number_of_classes / _nof_buckets));

  // Verify constraint table
  guarantee(_loader_constraints != NULL, "Verify of loader constraints failed");
  for (index = 0; index < _loader_constraint_size; index++) {
    for (LoaderConstraintEntry* probe = _loader_constraints[index]; probe; probe = probe->_next) {
      guarantee(probe->_name->is_symbol(), "should be symbol");
      if (probe->_klass != NULL) {
        instanceKlass* ik = instanceKlass::cast(probe->_klass); 
        guarantee(ik->name() == probe->_name, "name should match");
        klassOop k = find(ik->name(), ik->class_loader(), NULL);
        guarantee(k == probe->_klass, "klass should be in dictionary");
      }
      for (int n = 0; n< probe->_num_loaders; n++) {
        guarantee(probe->_loaders[n]->safepoint_is_oop_or_null(), "should be oop");
      }
    }
  }
}

// statistics code
class ClassStatistics: AllStatic {
 private:
  static int nclasses;    // number of classes
  static int nmethods;    // number of methods
  static int class_size;  // size of class objects in words
  static int method_size; // size of method objects in words
  static int debug_size;  // size of debug info in methods

  static void do_class(klassOop k) {
    nclasses++;
    class_size += k->size();
    if (k->klass_part()->oop_is_instance()) {
      instanceKlass* ik = (instanceKlass*)k->klass_part();
      class_size += ik->methods()->size();
      class_size += ik->constants()->size();
      class_size += ik->local_interfaces()->size();
      class_size += ik->transitive_interfaces()->size();
      // We do not have to count implementors, since we only store one!      
      class_size += ik->fields()->size();
    }
  }

  static void do_method(methodOop m) {
    nmethods++;
    method_size += m->size();
    // class loader uses same objArray for empty vectors, so don't count these
    if (m->exception_table()->length() == 0)       method_size += m->exception_table()->size();
  }

 public:
  static void print() {
    SystemDictionary::classes_do(do_class);
    SystemDictionary::methods_do(do_method);
    tty->print_cr("Class statistics:");
    tty->print_cr("%d classes (%d bytes)", nclasses, class_size * oopSize);
    tty->print_cr("%d methods (%d bytes = %d base + %d debug info)", nmethods, 
                  (method_size + debug_size) * oopSize, method_size * oopSize, debug_size * oopSize);
  }
};


int ClassStatistics::nclasses    = 0;  
int ClassStatistics::nmethods    = 0;
int ClassStatistics::class_size  = 0;
int ClassStatistics::method_size = 0;
int ClassStatistics::debug_size  = 0;

void SystemDictionary::print_class_statistics() {
  ResourceMark rm;
  ClassStatistics::print();
}


class MethodStatistics: AllStatic {
 public:
  enum {
    max_parameter_size = 10
  };
 private:

  static int _number_of_methods;
  static int _number_of_final_methods;
  static int _number_of_static_methods;
  static int _number_of_native_methods;
  static int _number_of_synchronized_methods;
  static int _number_of_bytecodes;
  static int _parameter_size_profile[max_parameter_size];
  static int _bytecodes_profile[Bytecodes::number_of_java_codes];

  static void initialize() {
    _number_of_methods        = 0;
    _number_of_final_methods  = 0;
    _number_of_static_methods = 0;
    _number_of_native_methods = 0;
    _number_of_synchronized_methods = 0;
    _number_of_bytecodes      = 0;
    for (int i = 0; i < max_parameter_size             ; i++) _parameter_size_profile[i] = 0;
    for (int j = 0; j < Bytecodes::number_of_java_codes; j++) _bytecodes_profile     [j] = 0;
  };

  static void do_method(methodOop m) {
    _number_of_methods++;
    // collect flag info
    if (m->is_final()       ) _number_of_final_methods++;
    if (m->is_static()      ) _number_of_static_methods++;
    if (m->is_native()      ) _number_of_native_methods++;
    if (m->is_synchronized()) _number_of_synchronized_methods++;
    // collect parameter size info (add one for receiver, if any)
    _parameter_size_profile[min(m->size_of_parameters() + (m->is_static() ? 0 : 1), max_parameter_size - 1)]++;
    // collect bytecodes info
    { HandleMark hm;
      BytecodeStream s(m);
      Bytecodes::Code c;
      while ((c = s.next()) >= 0) {
        _number_of_bytecodes++;
        _bytecodes_profile[c]++;
      }
    }
  }

 public:
  static void print() {
    initialize();
    SystemDictionary::methods_do(do_method);
    // generate output
    tty->cr();
    tty->print_cr("Method statistics (static):");
    // flag distribution
    tty->cr();
    tty->print_cr("%6d final        methods  %6.1f%%", _number_of_final_methods       , _number_of_final_methods        * 100.0F / _number_of_methods);
    tty->print_cr("%6d static       methods  %6.1f%%", _number_of_static_methods      , _number_of_static_methods       * 100.0F / _number_of_methods);
    tty->print_cr("%6d native       methods  %6.1f%%", _number_of_native_methods      , _number_of_native_methods       * 100.0F / _number_of_methods);
    tty->print_cr("%6d synchronized methods  %6.1f%%", _number_of_synchronized_methods, _number_of_synchronized_methods * 100.0F / _number_of_methods);
    // parameter size profile
    tty->cr();
    { int tot = 0;
      int avg = 0;
      for (int i = 0; i < max_parameter_size; i++) {
        int n = _parameter_size_profile[i];
        tot += n;
        avg += n*i;
        tty->print_cr("parameter size = %1d: %6d methods  %5.1f%%", i, n, n * 100.0F / _number_of_methods);
      }
      assert(tot == _number_of_methods, "should be the same");
      tty->print_cr("                    %6d methods  100.0%%", _number_of_methods);
      tty->print_cr("(average parameter size = %3.1f including receiver, if any)", (float)avg / _number_of_methods);
    }
    // bytecodes profile
    tty->cr();
    { int tot = 0;
      for (int i = 0; i < Bytecodes::number_of_java_codes; i++) {
        if (Bytecodes::is_defined(i)) {
          Bytecodes::Code c = Bytecodes::cast(i);
          int n = _bytecodes_profile[c];
          tot += n;
          tty->print_cr("%9d  %7.3f%%  %s", n, n * 100.0F / _number_of_bytecodes, Bytecodes::name(c));
        }
      }
      assert(tot == _number_of_bytecodes, "should be the same");
      tty->print_cr("%9d  100.000%%", _number_of_bytecodes);
    }
    tty->cr();
  }
};

int MethodStatistics::_number_of_methods;
int MethodStatistics::_number_of_final_methods;
int MethodStatistics::_number_of_static_methods;
int MethodStatistics::_number_of_native_methods;
int MethodStatistics::_number_of_synchronized_methods;
int MethodStatistics::_number_of_bytecodes;
int MethodStatistics::_parameter_size_profile[MethodStatistics::max_parameter_size];
int MethodStatistics::_bytecodes_profile[Bytecodes::number_of_java_codes];


void SystemDictionary::print_method_statistics() {
  MethodStatistics::print();
}

#endif // PRODUCT
