/*
 * @(#)systemDictionary.hpp	1.83 00/01/31 13:46:35
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


// The system dictionary stores all loaded classes and maps:
//
//   [class name,class loader] -> class   i.e.  [symbolOop,oop] -> klassOop
//
// Classes are loaded lazily. The default VM class loader is
// represented as NULL.

// The underlying data structure is an open hash table with a fixed number
// of buckets. During loading the loader object is locked, (for the VM loader 
// a private lock object is used). Class loading can thus be done concurrently, 
// but only by different loaders.
//
// During loading a placeholder (name,loader) is temporarily placed into the
// table (name is a symbolOop). This is used to detect ClassCircularityErrors.

class SystemDictionaryEntry;
class LoaderConstraintEntry;

class SystemDictionary : AllStatic {
 public:
  // Returns a class with a given class name and class loader.  Loads the
  // class if needed. If not found a NoClassDefFoundError or a
  // ClassNotFoundException is thrown, depending on the value on the
  // throw_error flag.  For most uses the throw_error argument should be set
  // to true.
  static klassOop resolve_or_fail(symbolHandle class_name, Handle class_loader, Handle protection_domain, bool throw_error, TRAPS);

  // Returns a class with a given class name and class loader.
  // Loads the class if needed. If not found NULL is returned.
  static klassOop resolve_or_null(symbolHandle class_name, Handle class_loader, Handle protection_domain, TRAPS);

  // Resolve from stream (called by jni_DefineClass and JVM_DefineClass)
  static klassOop resolve_from_stream(symbolHandle class_name, Handle class_loader, Handle protection_domain, ClassFileStream* st, TRAPS);
  
  // Lookup an already loaded class. If not found NULL is returned.
  static klassOop find(symbolHandle class_name, Handle class_loader, Handle protection_domain);

  // Lookup an already loaded instance or array class.
  // If not found NULL is returned.
  static klassOop find_instance_or_array_klass(symbolHandle class_name,
					       Handle class_loader,
					       Handle protection_domain,
					       TRAPS);
  
  // Iterate over all klasses in dictionary
  //   Just the classes from declaring class loaders
  static void classes_do(void f(klassOop));
  //   All classes, and their class loaders
  static void classes_do(void f(klassOop, oop));

  // Iterate over all methods in all klasses in dictionary
  static void methods_do(void f(methodOop));

  // Garbage collection support
  static void follow_roots_phase1();
  static bool follow_roots_phase2();
  static void adjust_pointers();
  static void young_oops_do(void f(oop*)); // traverse oops might reside in new space
  static void preloaded_oops_do(void f(oop*));

  // Printing
  static void print()                   PRODUCT_RETURN;
  static void print_class_statistics()  PRODUCT_RETURN;
  static void print_method_statistics() PRODUCT_RETURN;

  // Number of contained klasses  
  static inline int number_of_classes()     { return _number_of_classes; }

  // Monotonically increasing counter which grows as classes are loaded
  static inline int number_of_adds()     { return _number_of_adds; }

  // Verification
  static void verify();

  // Initialization
  static void initialize(TRAPS);

  // Fast access to commonly used classes (preloaded)
  static klassOop check_klass(klassOop k) {
    assert(k != NULL, "preloaded klass not initialized"); 
    return k;
  }

  static klassOop object_klass()            { return check_klass(_object_klass); }
  static klassOop string_klass()            { return check_klass(_string_klass); }
  static klassOop class_klass()             { return check_klass(_class_klass); }
  static klassOop cloneable_klass()         { return check_klass(_cloneable_klass); }
  static klassOop classloader_klass()       { return check_klass(_classloader_klass); }
  static klassOop serializable_klass()      { return check_klass(_serializable_klass); }

  static klassOop throwable_klass()         { return check_klass(_throwable_klass); }
  static klassOop error_klass()             { return check_klass(_error_klass); }
  static klassOop threaddeath_klass()       { return check_klass(_threaddeath_klass); }
  static klassOop exception_klass()         { return check_klass(_exception_klass); }
  static klassOop runtime_exception_klass() { return check_klass(_runtime_exception_klass); }
  static klassOop classNotFoundException_klass() { return check_klass(_classNotFoundException_klass); }
  static klassOop noClassDefFoundError_klass()   { return check_klass(_noClassDefFoundError_klass); }
  static klassOop ClassCastException_klass() { return check_klass(_classCastException_klass); }
  static klassOop ArrayStoreException_klass() { return check_klass(_arrayStoreException_klass); }
  static klassOop OutOfMemoryError_klass()  { return check_klass(_outOfMemoryError_klass); }
  static klassOop StackOverflowError_klass() { return check_klass(_StackOverflowError_klass); }
  static klassOop protectionDomain_klass()  { return check_klass(_protectionDomain_klass); }
  static klassOop AccessControlContext_klass() { return check_klass(_AccessControlContext_klass); }
  static klassOop reference_klass()         { return check_klass(_reference_klass); }
  static klassOop soft_reference_klass()    { return check_klass(_soft_reference_klass); }
  static klassOop weak_reference_klass()    { return check_klass(_weak_reference_klass); }
  static klassOop final_reference_klass()   { return check_klass(_final_reference_klass); }
  static klassOop phantom_reference_klass() { return check_klass(_phantom_reference_klass); }
  static klassOop finalizer_klass()         { return check_klass(_finalizer_klass); }
  
  static klassOop thread_klass()            { return check_klass(_thread_klass); }
  static klassOop threadGroup_klass()       { return check_klass(_threadGroup_klass); }
  static klassOop properties_klass()        { return check_klass(_properties_klass); }  
  static klassOop reflect_method_klass()    { return check_klass(_reflect_method_klass); }
  static klassOop reflect_constructor_klass() { return check_klass(_reflect_constructor_klass); }

  static klassOop vector_klass()            { return check_klass(_vector_klass); }
  static klassOop hashtable_klass()         { return check_klass(_hashtable_klass); }
  static klassOop stringBuffer_klass()      { return check_klass(_stringBuffer_klass); }

  static klassOop boolean_klass()           { return check_klass(_boolean_klass); }
  static klassOop char_klass()              { return check_klass(_char_klass); }
  static klassOop float_klass()             { return check_klass(_float_klass); }
  static klassOop double_klass()            { return check_klass(_double_klass); }
  static klassOop byte_klass()              { return check_klass(_byte_klass); }
  static klassOop short_klass()             { return check_klass(_short_klass); }
  static klassOop int_klass()               { return check_klass(_int_klass); }
  static klassOop long_klass()              { return check_klass(_long_klass); } 

  // Tells whether ClassLoader.loadClassInternal is present
  static bool has_loadClassInternal()       { return _has_loadClassInternal; }

  // Tells whether ClassLoader.checkPackageAccess is present
  static bool has_checkPackageAccess()      { return _has_checkPackageAccess; }

  static bool class_klass_loaded()          { return _class_klass != NULL; }
  static bool cloneable_klass_loaded()      { return _cloneable_klass != NULL; }
  
  // Returns default system loader
  static oop java_system_loader();
  // Compute the default system loader
  static void compute_java_system_loader(TRAPS);

  // Mirrors for primitive classes (created eagerly)
  static oop check_mirror(oop m) {
    assert(m != NULL, "mirror not initialized"); 
    return m;
  }

  static oop int_mirror()                   { return check_mirror(_int_mirror); }
  static oop float_mirror()                 { return check_mirror(_float_mirror); }
  static oop double_mirror()                { return check_mirror(_double_mirror); }
  static oop byte_mirror()                  { return check_mirror(_byte_mirror); }
  static oop bool_mirror()                  { return check_mirror(_bool_mirror); }
  static oop char_mirror()                  { return check_mirror(_char_mirror); }
  static oop long_mirror()                  { return check_mirror(_long_mirror); }
  static oop short_mirror()                 { return check_mirror(_short_mirror); }
  static oop void_mirror()                  { return check_mirror(_void_mirror); }

  // Check class loader constraints
  static void add_loader_constraint(symbolHandle name, Handle loader1, Handle loader2, TRAPS);
  static void check_signature_loaders(symbolHandle signature, Handle loader1, Handle loader2, bool is_method, TRAPS);

  // Basic find operation (needed by some verify code)
  static oop find_class_or_placeholder(symbolHandle class_name, Handle class_loader);

 private:

  enum Constants {
    _loader_constraint_size = 107,                     // number of entries in constraint table
    _nof_buckets            = 1009                     // number of buckets in hash table
  };

  // Static variables
  static SystemDictionaryEntry** _buckets;                  // array of buckets
  static int                     _number_of_classes;        // number of classes loaded
  static int                     _number_of_adds;           // monotonically increasing counter
                                                            //   which grows with _number_of_classes.
  static oop                     _system_loader_lock_obj;   // lock object for system class loader
  static LoaderConstraintEntry** _loader_constraints;       // constraints on class loaders

  // for VM_CounterDecay iteration support
  static int                    _current_class_index;      // current iteration index
  static SystemDictionaryEntry* _current_class_entry;      // pointer to the current hash table entry
  friend class CounterDecay;
  static klassOop try_get_next_class();

  // Returns the entry for name loader
  static SystemDictionaryEntry* get_entry(symbolOop class_name, oop class_loader);
  static void validate_protection_domain(instanceKlassHandle klass, Handle class_loader, Handle protection_domain, TRAPS);

  // unique class ID for vtune support
  friend struct MethodInfo;
  static void * get_classID(klassOop);

  // Basic loading operations
  static klassOop resolve_instance_class_or_null(symbolHandle class_name, Handle class_loader, Handle protection_domain, TRAPS);
  static klassOop resolve_array_class_or_null(symbolHandle class_name, Handle class_loader, Handle protection_domain, TRAPS);
  static void define_instance_class(instanceKlassHandle k, TRAPS);
  static instanceKlassHandle load_instance_class(symbolHandle class_name, Handle class_loader, TRAPS);
  static Handle compute_loader_lock_object(Handle class_loader);

  // Hash value computation
  static unsigned int hash_value(symbolHandle class_name, Handle class_loader);
 
  // Array index computation
  static int index_for(symbolHandle class_name, Handle class_loader); 

  // Setup link to hierarchy
  static void add_to_hierarchy(instanceKlassHandle k, TRAPS);  
    
  // Updating entry in dictionary
  static void add_klass_or_placeholder(symbolOop class_name, oop class_loader, oop obj);
  static void remove_placeholder(symbolOop class_name, oop class_loader, klassOop klass);  
  
  // Initialization
  static void initialize_preloaded_classes(TRAPS);
  static void initialize_basic_type_mirrors(TRAPS);
    
  // Class loader constraints
  static LoaderConstraintEntry** find_loader_constraint(symbolOop name, oop loader);
  static void ensure_loader_constraint_capacity(LoaderConstraintEntry *p, int nfree);
  static void extend_loader_constraint(LoaderConstraintEntry* p, oop loader, klassOop klass);
  static void merge_loader_constraints(LoaderConstraintEntry** pp1, LoaderConstraintEntry** pp2, klassOop klass);
  static void check_constraints(instanceKlassHandle k, Handle loader, TRAPS);
  static void update_dictionary(instanceKlassHandle k, Handle loader);

  // Variables holding commonly used klasses (preloaded)
  static klassOop _object_klass;
  static klassOop _string_klass;
  static klassOop _class_klass;
  static klassOop _cloneable_klass;
  static klassOop _classloader_klass;
  static klassOop _serializable_klass;
  
  static klassOop _throwable_klass;
  static klassOop _error_klass;
  static klassOop _threaddeath_klass;
  static klassOop _exception_klass;
  static klassOop _runtime_exception_klass;
  static klassOop _classNotFoundException_klass;
  static klassOop _noClassDefFoundError_klass;
  static klassOop _classCastException_klass;
  static klassOop _arrayStoreException_klass;
  static klassOop _outOfMemoryError_klass;
  static klassOop _StackOverflowError_klass;
  static klassOop _protectionDomain_klass;
  static klassOop _AccessControlContext_klass;
  static klassOop _reference_klass;
  static klassOop _soft_reference_klass;
  static klassOop _weak_reference_klass;
  static klassOop _final_reference_klass;
  static klassOop _phantom_reference_klass;
  static klassOop _finalizer_klass;

  static klassOop _thread_klass;
  static klassOop _threadGroup_klass;
  static klassOop _properties_klass;      
  static klassOop _reflect_method_klass;
  static klassOop _reflect_constructor_klass;

  static klassOop _stringBuffer_klass;
  static klassOop _vector_klass;
  static klassOop _hashtable_klass;

  static klassOop _boolean_klass;
  static klassOop _char_klass;
  static klassOop _float_klass;
  static klassOop _double_klass;
  static klassOop _byte_klass;
  static klassOop _short_klass;
  static klassOop _int_klass;
  static klassOop _long_klass;

  static oop  _java_system_loader;

  static bool _has_loadClassInternal;
  static bool _has_checkPackageAccess;

  // Primitive classes
  static oop _int_mirror;
  static oop _float_mirror;
  static oop _double_mirror;
  static oop _byte_mirror;
  static oop _bool_mirror;
  static oop _char_mirror;
  static oop _long_mirror;
  static oop _short_mirror;
  static oop _void_mirror;
};

