/*
 * @(#)train.cpp	1.36 01/02/09 14:45:15
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_train.cpp.incl"


#define FOR_EACH_CAR(c) \
  for (CarSpace* c = first_car(); c != NULL; c = c->next_car())

void Train::set_saved_mark_to_top() { 
  set_saved_top_mark(top_mark());
}

void Train::set_saved_mark_to_bottom() { 
  set_saved_top_mark(bottom_mark());
}

bool Train::top_at_saved_mark() const {
  return saved_top_mark() == top_mark();
}

WaterMark Train::top_mark() const {
  return last_car()->top_mark();
}

WaterMark Train::bottom_mark() const {
  return first_car()->bottom_mark();
}

size_t Train::free() const {     // Free chunks at the end of the first n-1 cars cannot 
  return last_car()->free();   // be used and is therefore not considered free
}                                

size_t Train::used() const {     // The first n-1 cars are considered as filled
  size_t size = 0;
  CarSpace* c = first_car();
  while (c != last_car()) {
    size += c->capacity();
    c = c->next_car();
  }
  size += c->used();
  return size;
}

size_t Train::capacity() const {
  size_t size = 0;
  FOR_EACH_CAR(c) {
    size += c->capacity();
  }
  return size;
}


int Train::length() const {
  int len = 0;
  FOR_EACH_CAR(c) {
    len++;
  }
  return len;
}


oop* Train::allocate(int size) {
  oop* p = last_car()->allocate(size);
  if (p != NULL) {
    return p;
  } else {
    CarSpace* new_last_car = add_new_last_car(size);
    if (new_last_car != NULL) {
      return new_last_car->allocate(size);
    } else {
      return NULL;
    }
  }
}


oop* Train::safepoint_allocate(int size) {
  oop* p = last_car()->safepoint_allocate(size);
  if (p != NULL) {
    return p;
  } else {
    CarSpace* new_last_car = add_new_last_car(size);
    if (new_last_car != NULL) {
      return new_last_car->safepoint_allocate(size);
    } else {
      return NULL;
    }
  }
}


void Train::initialize(CarSpace* first_car) {
  set_first_car(first_car);
  set_last_car(first_car);
  set_next_train(NULL);
  set_prev_train(NULL);
  set_saved_mark_to_bottom();
}


CarSpace* Train::add_new_last_car(int size) {
  CarSpace* c = Universe::train_gen()->retrieve_from_car_pool(size, train_number(), last_car()->car_number() + 1, this);
  if (c != NULL) {
    last_car()->set_next_car(c);
    set_last_car(c);
  }
  return c;
}


void Train::scavenge_recorded_stores(bool include_first) {
  CarSpace* c = first_car();
  if (!include_first) {
    c = c->next_car();
  }
  while (c != NULL) {
    c->scavenge_recorded_stores();
    c = c->next_car();
  }
}


bool Train::scavenge_contents_from_saved_mark() {
  bool work = false;
  while (saved_top_mark() != top_mark()) {
    work = true;
    CarSpace* sp = (CarSpace*) saved_top_mark().space();   // Could probably make scavenge_contents_from virtual ?
    sp->scavenge_contents_from(addr_saved_top_mark());
    if (sp->next_car() != NULL) {
      set_saved_top_mark(sp->next_car()->bottom_mark());
    }
  }
  return work;
}


void Train::object_iterate_from_saved_mark(ObjectClosure* blk) {
  while (saved_top_mark() != top_mark()) {
    CarSpace* sp = (CarSpace*) saved_top_mark().space();   // Could probably make train_scavenge_contents_from virtual ?
    sp->object_iterate_from(addr_saved_top_mark(), blk);
    if (sp->next_car() != NULL) {
      set_saved_top_mark(sp->next_car()->bottom_mark());
    }
  }
}


void Train::object_iterate(ObjectClosure* blk) {
  FOR_EACH_CAR(c) {
    c->object_iterate(blk);
  }
}


void Train::space_iterate(SpaceClosure* blk) {
  FOR_EACH_CAR(c) {
    blk->do_space(c);
  }
}


void Train::update_offset_array() {
  FOR_EACH_CAR(c) {
    c->update_offset_array();
  }
}


#ifndef PRODUCT

void Train::verify(bool allow_dirty) {
  julong tn = train_number();
  CarSpace* c = first_car();
  guarantee(c != NULL, "No first car in train");
  while (c != NULL) {
    guarantee(c->train() == this, "Inconsistent car table");    
    guarantee(c->train_number() == tn, "Inconsistent car table");    
    juint cn = c->car_number();
    guarantee(cn > 0, "Illegal car number");
    CarSpace* next_c = c->next_car();
    if (next_c == NULL) {
      guarantee(c == last_car(), "Wrong last car in train");
    } else {
      guarantee(cn < next_c->car_number(), "Car numbers in train not increasing");
    }
    c->verify(allow_dirty);
    c = next_c;
  }
}
  

void Train::print() {
  tty->print(" train ");
  tty->print_julong(train_number());
  tty->cr();
  FOR_EACH_CAR(c) {
    c->print();
  }
}


void Train::print_short() {
  tty->print(" train (");
  tty->print_julong(train_number());
  tty->print(",%u-%u)", first_car()->car_number(), last_car()->car_number());
}

#endif // PRODUCT


void Train::prepare_for_compaction() {
  bool prev_has_single_car = prev_train() ? prev_train()->has_single_car() : false;
  if (prev_has_single_car && has_single_car()) {
    // do nothing, our single car is already traversed
  } else {
    first_car()->prepare_for_compaction();
  }
}


void Train::adjust_pointers() {
  bool prev_has_single_car = prev_train() ? prev_train()->has_single_car() : false;
  if (prev_has_single_car && has_single_car()) {
    // do nothing, our single car is already traversed
  } else {
    first_car()->adjust_pointers();
  }
}


void Train::compact() {
  bool prev_has_single_car = prev_train() ? prev_train()->has_single_car() : false;
  if (prev_has_single_car && has_single_car()) {
    // do nothing, our single car is already traversed
  } else {
    first_car()->compact();
  }
}


void Train::set_should_scavenge(bool value) {
  FOR_EACH_CAR(c) {
    c->set_should_scavenge(value);
  }
}


void Train::clear_remembered_sets() {
  FOR_EACH_CAR(c) {
    c->remembered_set()->clear();
  }
}


bool Train::full_gc_epilogue() {
  CarSpace* c = first_car();  
  while (c->used() == 0) {
    // first car empty, release it and update first_car pointer
    CarSpace* new_first = c->next_car();
    if (new_first == NULL) return true; // entire train empty
    set_first_car(new_first);
    Universe::train_gen()->add_to_car_pool(c, false);
    c = new_first;
  }
  // first car no longer empty, cannot release train
  c->remembered_set()->reinitialize(false);
  CarSpace* prev_c = c;
  c = c->next_car();
  // iterate over remaining cars
  while (c != NULL) {
    CarSpace* next_c = c->next_car();
    if (c->used() == 0) {
      // we found an empty car, release it and update next_car pointer
      prev_c->set_next_car(next_c);
      Universe::train_gen()->add_to_car_pool(c, false);
    } else {
      // car still in use, clear remembered set
      c->remembered_set()->reinitialize(false);
      prev_c = c;
    }
    c = next_c;
  }
  // update last_car pointer
  set_last_car(prev_c);
  return false;
}


bool Train::tail_has_other_train_references() {
  CarSpace* c = first_car(); 
  assert(c != NULL, "just checking");
  c = c->next_car();
  while (c != NULL) {
    if (c->remembered_set()->has_other_train_references(this)) return true;
    c = c->next_car();
  }
  return false;
}


