/*
 * @(#)train.hpp	1.29 01/02/09 14:45:16
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// A Train instance models an individual train in the train generation
// and provides support for allocation and garbage collection.

class Train: public CHeapObj {
  friend class TrainGeneration;
 private:
  CarSpace* _first_car;                 // The oldest car in this train
  CarSpace* _last_car;                  // The youngest car in this train
  Train*    _next_train;                // Next train (or NULL)
  Train*    _prev_train;                // Previous train (or NULL)
  WaterMark _saved_top_mark;            // Watermark used for gc/allocation profiling

  // Watermark
  WaterMark saved_top_mark() const      { return _saved_top_mark; }
  void set_saved_top_mark(WaterMark m)  { _saved_top_mark = m; }
  WaterMark* addr_saved_top_mark()      { return &_saved_top_mark; }
  void set_saved_mark_to_top();
  void set_saved_mark_to_bottom();
  bool top_at_saved_mark() const;
  WaterMark top_mark() const;
  WaterMark bottom_mark() const;

  // Accessors
  void set_first_car(CarSpace* c)       { _first_car = c; }
  void set_last_car(CarSpace* c)        { _last_car = c; }
  void set_next_train(Train* t)         { _next_train = t; }
  void set_prev_train(Train* t)         { _prev_train = t; }

  // Add new car at end of train
  CarSpace* add_new_last_car(int size);

 public:
  // Accessors
  CarSpace* first_car() const           { return _first_car; }
  CarSpace* last_car() const            { return _last_car; }
  Train* next_train() const             { return _next_train; }
  Train* prev_train() const             { return _prev_train; }

  // Number of this train
  julong train_number() const           { return first_car()->train_number(); }

  // Length computations
  bool has_single_car() const           { return first_car()->next_car() == NULL; }
  int length() const;

  // Size computations (in bytes)
  size_t capacity() const;
  size_t used() const;
  size_t free() const;

  // Allocate oop of given size in train, append new car if necessary
  oop* allocate(int size);
  oop* safepoint_allocate(int size);

  // Allocate oop of given size in train, return NULL if full
  oop* allocate_within(int size) {
    return last_car()->allocate(size);
  }

  oop* safepoint_allocate_within(int size) {
    return last_car()->safepoint_allocate(size);
  }

  // Initialization.
  void initialize(CarSpace* first_car);

  // Scavenge from saved mark to top mark and update saved mark. Returns true if any work was done.
  bool scavenge_contents_from_saved_mark();

  // Iterate from saved marks to top marks and update saved marks. Returns true if any work was done.
  void object_iterate_from_saved_mark(ObjectClosure* blk);

  // Iterate over entire train
  void object_iterate(ObjectClosure* blk);
  void space_iterate(SpaceClosure* blk);

  // GC support
  void scavenge_recorded_stores(bool include_first);
  void clear_remembered_sets();
  bool tail_has_other_train_references();
  void set_should_scavenge(bool value);
  void update_offset_array();

  // Debugging
  void verify(bool allow_dirty) PRODUCT_RETURN;
  void print()                  PRODUCT_RETURN;
  void print_short()            PRODUCT_RETURN;

 private:
  // Mark sweep support
  bool full_gc_epilogue();
  void prepare_for_compaction();
  void adjust_pointers();
  void compact();
};

