/*
 * @(#)universe.inline.hpp	1.35 00/07/20 11:47:31
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


// Checkes if pointer points into heap. Does not check for
// badHeapOop, so can be used when a pointer points into a just
// allocated but not yet initialized object.

#ifndef PRODUCT

inline bool Universe::is_in_heap(oop p) {
  Space* space = CarTable::space_for_or_null(p);
  return space == NULL ? false : space->contains(p);
}


inline bool Universe::is_heap_or_null(oop p) {
  return p == NULL ? true : is_heap(p);
}


inline bool Universe::is_heap(oop p) {
  if (!is_in_heap(p)) return false;
  return *((oop*)p) != badHeapOop;
}

#endif // PRODUCT


inline bool Universe::is_reserved_heap(oop p) {
  // Checks whether p is in the area reserved (but not necessarily committed) for the heap.
  // This is much faster than the is_heap check.
  return (oop*) p >= _new_gen->low_boundary && (oop*) p < _perm_gen->high_boundary;
}


// Check whether an element of a typeArrayOop with the given type must be aligned 0 mod 8.
// The typeArrayOop itself must be aligned at least this strongly,

// This code is more convoluted than it needs be because MSVC 5.0 doesn't inline the clean version.
// It's time-critical (part of e.g. array allocation).

#ifdef ASSERT
inline bool Universe::element_type_should_be_aligned(BasicType type) {  // clean version
  if ((ALIGN_DOUBLE_ELEMENTS && type == T_DOUBLE) ||
      (ALIGN_LONG_ELEMENTS   && type == T_LONG)) {
    assert(ALIGN_ALL_OBJECTS, "aligned element must be inside aligned array object");
    return true;
  } else {
    return false;
  }
}
#else
inline bool Universe::element_type_should_be_aligned(BasicType type) {  // clean version
  return
#if ALIGN_DOUBLE_ELEMENTS
    type == T_DOUBLE ||
#endif
#if ALIGN_LONG_ELEMENTS
    type == T_LONG
#else
    false
#endif
    ;
}
#endif


// Check whether an object field (static/non-static) of the given type must be aligned 0 mod 8.

inline bool Universe::field_type_should_be_aligned(BasicType type) {
  if ((ALIGN_DOUBLE_FIELDS && type == T_DOUBLE) ||
      (ALIGN_LONG_FIELDS   && type == T_LONG)) {
    assert(ALIGN_ALL_OBJECTS, "aligned field must be inside aligned object");
    return true;
  } else {
    return false;
  }
}


// Allocation support
inline void Universe::post_allocation_setup(KlassHandle klass, oop* obj, int size) {
  ((oop)obj)->set_mark();
  ((oop)obj)->set_klass(klass());
  assert(is_bootstrapping() || !((oop)obj)->is_array(), "must not be an array");
  if (jvmpi_alloc_event_enabled()) {
    jvmpi_object_alloc((oop)obj, size * wordSize);
  }
}

inline void Universe::post_allocation_setup_array(KlassHandle klass, oop* obj, int size, int length) {
  ((oop)obj)->set_mark();
  ((oop)obj)->set_klass(klass());
  assert(((oop)obj)->is_array(), "must be an array");
  ((arrayOop)obj)->set_length(length);
  if (jvmpi_alloc_event_enabled()) {
    jvmpi_object_alloc((oop)obj, size * wordSize);
  }
}


// Normal allocation (uninitialized)  
inline oop* Universe::allocate(KlassHandle klass, int size, TRAPS) {
  debug_only(check_for_valid_allocation_state(THREAD));
  assert(no_gc_in_progress(), "Allocation during gc not allowed");
  oop* obj = _new_gen->eden()->allocate(size);
  if (obj == NULL) {
    obj = Scavenge::invoke_and_allocate(size, false, CHECK_0);
  }
  post_allocation_setup(klass, obj, size);
  return obj; 
}


inline oop* Universe::allocate_array(KlassHandle klass, int size, int length, TRAPS) {
  debug_only(check_for_valid_allocation_state(THREAD));
  assert(no_gc_in_progress(), "Allocation during gc not allowed");
  oop* obj = _new_gen->eden()->allocate(size);
  if (obj == NULL) {
    obj = Scavenge::invoke_and_allocate(size, false, CHECK_0);
  }
  post_allocation_setup_array(klass, obj, size, length);
  return obj; 
}


// Normal allocation (initialized)  
inline oop* Universe::clear_allocate_shared(KlassHandle klass, int size, TRAPS) {
  const int hs = oopDesc::header_size();
  oop* obj = allocate(klass, size, CHECK_0);
  assert(size >= hs, "unexpected object size");
  memset(obj + hs, 0, (size - hs) * wordSize);
  return obj;
}


// Normal allocation (initialized)
inline oop* Universe::clear_allocate(KlassHandle klass, int size, TRAPS) {
  if(UseTLE) {
    return clear_allocate_local (klass, size, 0, false, CATCH);
  }
  else {
    return clear_allocate_shared(klass, size, CATCH);
  }
}


inline oop* Universe::clear_allocate_array_shared(KlassHandle klass, int size, int length, TRAPS) {
  const int hs = arrayOopDesc::header_size(T_INT); // T_INT is ok for all array types
  oop* obj = allocate_array(klass, size, length, CHECK_0);
  assert(size >= hs, "unexpected object size");
  memset(obj + hs, 0, (size - hs) * wordSize);
  return obj;
}


// Normal allocation (initialized)
inline oop* Universe::clear_allocate_array(KlassHandle klass, int size, int length, TRAPS) {
  if(UseTLE) {
    return clear_allocate_local       (klass, size, length, true, CATCH);
  }
  else {
    return clear_allocate_array_shared(klass, size, length, CATCH);
  }
}


// Tenured allocation (uninitialized)  
inline oop* Universe::allocate_tenured(KlassHandle klass, int size, TRAPS) {
  debug_only(check_for_valid_allocation_state(THREAD));
  assert(no_gc_in_progress(), "Allocation during gc not allowed");
  oop* obj = _old_gen->allocate(size, CHECK_0);
  post_allocation_setup(klass, obj, size);
  return obj;
}


// Tenured allocation (initialized)  
inline oop* Universe::clear_allocate_tenured(KlassHandle klass, int size, TRAPS) {
  const int hs = oopDesc::header_size();
  oop* obj = allocate_tenured(klass, size, CHECK_0);
  assert(size >= hs, "unexpected object size");
  memset(obj + hs, 0, (size - hs) * wordSize);
  return obj;
}


// Tenured allocation of large type arrays (uninitialized)
inline oop* Universe::allocate_large_typearray(KlassHandle klass, int size, int length, TRAPS) {
  debug_only(check_for_valid_allocation_state(THREAD));
  assert(no_gc_in_progress(), "Allocation during gc not allowed");
  oop* obj = _new_gen->eden()->allocate(size);
  if (obj == NULL) {
    obj = Scavenge::invoke_and_allocate(size, true, CHECK_0);
  }
  post_allocation_setup_array(klass, obj, size, length);
  return obj; 
}


// Tenured allocation of large type arrays (initialized)
inline oop* Universe::clear_allocate_large_typearray(KlassHandle klass, int size, int length, TRAPS) {
  const int hs = arrayOopDesc::header_size(T_INT); // T_INT is ok for all array types
  oop* obj = allocate_large_typearray(klass, size, length, CHECK_0);
  assert(size >= hs, "unexpected object size");
  memset(obj + hs, 0, (size - hs) * wordSize);
  return obj;
}


// Permanent allocation (uninitialized)  
inline oop* Universe::allocate_permanent(KlassHandle klass, int size, TRAPS) { 
  debug_only(check_for_valid_allocation_state(THREAD));
  assert(no_gc_in_progress(), "Allocation during gc not allowed");
  oop* obj = _perm_gen->allocate(size, CHECK_0);
  post_allocation_setup(klass, obj, size);
  return obj; 
}


inline oop* Universe::allocate_permanent_array(KlassHandle klass, int size, int length, TRAPS) { 
  debug_only(check_for_valid_allocation_state(THREAD));
  assert(no_gc_in_progress(), "Allocation during gc not allowed");
  oop* obj = _perm_gen->allocate(size, CHECK_0);
  post_allocation_setup_array(klass, obj, size, length);
  return obj; 
}


// Permanent allocation (initialized)  
inline oop* Universe::clear_allocate_permanent(KlassHandle klass, int size, TRAPS) {
  const int hs = oopDesc::header_size();
  oop* obj = allocate_permanent(klass, size, CHECK_0);
  assert(size >= hs, "unexpected object size");
  memset(obj + hs, 0, (size - hs) * wordSize);
  return obj;
}


inline oop* Universe::clear_allocate_permanent_array(KlassHandle klass, int size, int length, TRAPS) {
  const int hs = arrayOopDesc::header_size(T_INT); // T_INT is ok for all array types
  oop* obj = allocate_permanent_array(klass, size, length, CHECK_0);
  assert(size >= hs, "unexpected object size");
  memset(obj + hs, 0, (size - hs) * wordSize);
  return obj;
}

