/*
 * @(#)constantPoolOop.cpp	1.62 00/03/27 15:37:38
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_constantPoolOop.cpp.incl"

klassOop constantPoolOopDesc::klass_at_impl(constantPoolHandle this_oop, int which, TRAPS) {   
  // A resolved constantPool entry will contain a klassOop, otherwise a symbolOop. 
  // It is not safe to rely on the tag bit's here, since we don't have a lock, and the entry and
  // tag is not updated atomicly.  
  oop entry = *(this_oop->obj_at_addr(which));
  if (entry->is_klass()) {
    // Already resolved - return entry.
    return (klassOop)entry;
  }  

  // Acquire lock on constant oop while doing update. After we get the lock, we check if another object
  // already has updated the object
  assert(Thread::current()->is_Java_thread(), "must be a Java thread");
  bool do_resolve = false;
  symbolHandle name   = symbolOop(NULL);
  Handle       loader = NULL;
  { ObjectLocker ol(this_oop, THREAD);
    do_resolve = this_oop->tag_at(which).is_unresolved_klass();
    if (do_resolve) {
      name   = this_oop->unresolved_klass_at(which);
      loader = instanceKlass::cast(this_oop->pool_holder())->class_loader();
    }
  } // unlocking constantPool
    

  if (do_resolve) {
    // this_oop must be unlocked during resolve_or_fail
    oop protection_domain = Klass::cast(this_oop->pool_holder())->protection_domain();
    KlassHandle k = SystemDictionary::resolve_or_fail(name, loader, protection_domain, true, CHECK_0);

    // Do access check for klasses
    verify_constant_pool_resolve(this_oop, k, CHECK_0);
    
    { ObjectLocker ol (this_oop, THREAD);
      // Only updated constant pool - if it is resolved.
      do_resolve = this_oop->tag_at(which).is_unresolved_klass();
      if (do_resolve) {
        this_oop->klass_at_put(which, k());       
      }
    }
  }
 
  entry = *(this_oop->obj_at_addr(which));
  assert(this_oop->tag_at(which).is_klass(), "must be resolved at this point");
  assert(entry->is_klass(), "must be resolved at this point");
  return (klassOop)entry;
}


// Does not update constantPoolOop - to avoid any exception throwing. Only used by compiler
// and exception handling. Returns NULL if the class has not been loaded or if the verification
// of constant pool failed
klassOop constantPoolOopDesc::klass_at_if_loaded(constantPoolHandle this_oop, int which) {  
  oop entry = *this_oop->obj_at_addr(which);
  if (entry->is_klass()) {
    return (klassOop)entry;
  } else {
    assert(entry->is_symbol(), "must be either symbol or klass");
    symbolOop name = (symbolOop)entry;
    oop loader = instanceKlass::cast(this_oop->pool_holder())->class_loader();
    oop protection_domain = Klass::cast(this_oop->pool_holder())->protection_domain();
    klassOop k = SystemDictionary::find(name, loader, protection_domain);    

    if (k != NULL) {
      // Make sure that resolving is legalKlassHandle klass(THREAD, k);
      EXCEPTION_MARK;
      KlassHandle klass(THREAD, k);
      // return NULL if verification fails
      verify_constant_pool_resolve(this_oop, klass, THREAD);
      if (HAS_PENDING_EXCEPTION) {
	CLEAR_PENDING_EXCEPTION;
	return NULL;
      }
      return klass();
    } else {
      return k;
    }
  }
}


klassOop constantPoolOopDesc::klass_ref_at_if_loaded(constantPoolHandle this_oop, int which) {
  return klass_at_if_loaded(this_oop, this_oop->klass_ref_index_at(which));
}


// This is an interface for the compiler that allows accessing non-resolved entries
// in the constant pool - but still performs the validations tests. Must be used
// in a pre-parse of the compiler - to determine what it can do and not do.
// Note: We cannot update the ConstantPool from the vm_thread.
klassOop constantPoolOopDesc::klass_ref_at_if_loaded_check(constantPoolHandle this_oop, int index, TRAPS) {
  int which = this_oop->klass_ref_index_at(index);
  oop entry = *this_oop->obj_at_addr(which);
  if (entry->is_klass()) {
    return (klassOop)entry;
  } else {
    assert(entry->is_symbol(), "must be either symbol or klass");
    symbolOop name = (symbolOop)entry;
    oop loader = instanceKlass::cast(this_oop->pool_holder())->class_loader();
    oop protection_domain = Klass::cast(this_oop->pool_holder())->protection_domain();
    KlassHandle k(THREAD, SystemDictionary::find(name, loader, protection_domain));
    assert(k.not_null(), "must already be loaded");  
  
    // Do access check for klasses
    verify_constant_pool_resolve(this_oop, k, CHECK_0);
    return k();
  }
}


void constantPoolOopDesc::verify_constant_pool_resolve(constantPoolHandle this_oop, KlassHandle k, TRAPS) {
 if (k->oop_is_instance() || k->oop_is_objArray()) {        
    instanceKlassHandle holder = this_oop->pool_holder();
    KlassHandle element = k->oop_is_instance() ? k() : objArrayKlass::cast(k())->bottom_klass();
    
    // The element type could be a typeArray - we only need the access check if it is
    // an reference to another class
    if (element->oop_is_instance()) {
      LinkResolver::check_klass_accessability(holder, element, CHECK);
    }        
  }
}


int constantPoolOopDesc::klass_ref_index_at(int which) {
  jint ref_index = field_or_method_at(which);
  return extract_low_short_from_int(ref_index);
}


int constantPoolOopDesc::name_and_type_ref_index_at(int which) {
  jint ref_index = field_or_method_at(which);
  return extract_high_short_from_int(ref_index);
}


int constantPoolOopDesc::name_ref_index_at(int which) {
  jint ref_index = name_and_type_at(which);
  return extract_low_short_from_int(ref_index);
}


int constantPoolOopDesc::signature_ref_index_at(int which) {
  jint ref_index = name_and_type_at(which);
  return extract_high_short_from_int(ref_index);
}


klassOop constantPoolOopDesc::klass_ref_at(int which, TRAPS) {
  return klass_at(klass_ref_index_at(which), CHECK_0);
}


symbolOop constantPoolOopDesc::name_ref_at(int which) {
  jint ref_index = name_and_type_at(name_and_type_ref_index_at(which));
  int name_index = extract_low_short_from_int(ref_index);
  return symbol_at(name_index);
}


symbolOop constantPoolOopDesc::signature_ref_at(int which) {
  jint ref_index = name_and_type_at(name_and_type_ref_index_at(which));
  int signature_index = extract_high_short_from_int(ref_index);
  return symbol_at(signature_index);
}


BasicType constantPoolOopDesc::basic_type_for_signature_at(int which) {
  return FieldType::basic_type(symbol_at(which));
}


void constantPoolOopDesc::resolve_string_constants_impl(constantPoolHandle this_oop, TRAPS) {
  for (int index = 1; index < this_oop->length(); index++) { // Index 0 is unused
    if (this_oop->tag_at(index).is_unresolved_string()) {
      this_oop->string_at(index, CHECK);
    }
  }
}

oop constantPoolOopDesc::string_at_impl(constantPoolHandle this_oop, int which, TRAPS) {
  oop entry = *(this_oop->obj_at_addr(which));
  if (entry->is_symbol()) {
    ObjectLocker ol(this_oop, THREAD);
    if (this_oop->tag_at(which).is_unresolved_string()) {
      // Intern string
      symbolOop sym = this_oop->unresolved_string_at(which);
      entry = StringTable::intern(sym, CHECK_(constantPoolOop(NULL)));
      this_oop->string_at_put(which, entry);
    } else {
      // Another thread beat us and interned string, read string from constant pool
      entry = this_oop->resolved_string_at(which);
    }
  }
  assert(java_lang_String::is_instance(entry), "must be string");
  return entry;
}

#ifndef PRODUCT

const char* constantPoolOopDesc::printable_name_at(int which) {
  if (tag_at(which).is_klass()) {
    return Klass::cast(resolved_klass_at(which))->internal_name();
  }
  if (tag_at(which).is_unresolved_klass()) {
    return unresolved_klass_at(which)->as_C_string();
  }
  if (tag_at(which).is_symbol()) {
    return symbol_at(which)->as_C_string();
  }
  if (tag_at(which).is_unresolved_string()) {
    return unresolved_string_at(which)->as_C_string();
  }
  if (tag_at(which).is_string()) {
    return java_lang_String::as_utf8_string(resolved_string_at(which));
  }
  return "";
}

#endif // PRODUCT


