/*
 * @(#)cpCacheOop.hpp	1.25 01/02/09 14:45:25
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


// A ConstantPoolCacheEntry describes an individual entry of the constant
// pool cache. There's 2 principal kinds of entries: field entries for in-
// stance & static field access, and method entries for invokes. Some of
// the entry layout is shared and looks as follows:
//
// bit number |31                0|
// bit length |-8--|-8--|---16----|
// --------------------------------
// _indices   [ b2 | b1 |  index  ]
// _f1        [  entry specific   ]
// _f2        [  entry specific   ]
// _flags     [f|unused |  hindex ] (for field entries)
// _flags     [f|unused |eidx|psze] (for method entries)       
// --------------------------------
// bit length |4|--12---|-8--|-8--]
//
// with:
// index  = original constant pool index
// b1     = bytecode 1
// b2     = bytecode 2
// psze   = parameters size (method entries only)
// eidx   = interpreter entry index (method entries only)
// hindex = original field index in method holder
// f      = flags (see below)
//
// The flags have the following interpretation for field entries:
// bit 31: S flag  true if store check is needed for put operation
// bit 30: 2 flag  true if value size is 2 words
// bit 29: F flag  true if value is float value
// bit 28: f flag  true if field is marked final
//
// The flags 31, 30, 29 together build a 3bit number 0 to 5 with the
// following mapping to the TosState states:
//
// itos: 0
// ftos: 1
// ltos: 2
// dtos: 3
// atos: 4
// vtos: 5
//
// Entry specific: field entries:
// _indices = get (b1 section) and put (b2 section) bytecodes, original constant pool index
// _f1      = field holder
// _f2      = field offset in words
// _flags   = field type information, original field index in field holder (hindex section)
//
// Entry specific: method entries:
// _indices = invoke code for f1 (b1 section), invoke code for f2 (b2 section), original constant pool index
// _f1      = method for all but virtual calls, unused by virtual calls
// _f2      = method/vtable index for virtual calls only, unused by all other calls
// _flags   = field type info (f section), interpreter entry index (eidx section), parameter size (psze section)
//
// Note: invokevirtual & invokespecial bytecodes can share the same constant pool entry and thus
//       the same constant pool cache entry. All invoke bytecodes but invokevirtual use only _f1
//       and the corresponding b1 bytecode, while invokevirtual uses only _f2 and the corresponding
//       b2 bytecode.

class ConstantPoolCacheEntry VALUE_OBJ_CLASS_SPEC {
  friend class VMStructs;
 private:
  int  _indices;                                 // constant pool index & rewrite bytecodes
  oop  _f1;                                      // entry specific oop field
  int  _f2;                                      // entry specific int/oop field
  int  _flags;                                   // flags

#ifdef ASSERT
  bool same_methodOop(oop cur_f1, oop f1);
#endif

  void set_bytecode_1(Bytecodes::Code code);
  void set_bytecode_2(Bytecodes::Code code);
  void set_f1(oop f1)                            { assert(_f1 == NULL || same_methodOop(_f1, f1), "illegal field change"); _f1 = f1; }
  void set_f2(int f2)                            { assert(_f2 == 0    || _f2 == f2, "illegal field change"); _f2 = f2; }
  void set_flags(int flags)                      { _flags = flags; }

 public:
  // Initialization
  void set_initial_state(int index);             // sets entry to initial state

  void set_field(                                // sets entry to resolved field state
    Bytecodes::Code get_code,                    // the bytecode used for reading the field
    Bytecodes::Code put_code,                    // the bytecode used for writing the field
    KlassHandle     field_holder,                // the object/klass holding the field
    int             field_index,                 // the original field index in the field holder
    int             field_offset,                // the field offset in words in the field holder
    TosState        field_type,                  // the (machine) field type    
    bool            is_final                     // the field is final 
  );

  void set_method(                               // sets entry to resolved method entry
    Bytecodes::Code invoke_code,                 // the bytecode used for invoking the method
    methodHandle    method,                      // the method/prototype if any (NULL, otherwise)
    int             vtable_index,                // the vtable index if any (-1, otherwise)
    bool            change_to_virtual = false    // set if the invokeinterface needs to be rewritten to invokevirtual
  );                                             // (see cpCacheOop.cpp for details)


  void set_interface_call(
    methodHandle method,                         // Resolved method    
    int index                                    // Method index into interface
  );               


  void set_parameter_size(int value) {
    assert(parameter_size() == 0 || parameter_size() == value, "size must not change");
    _flags = ((_flags & 0xFFFFFF00) | (value & 0xFF));
  }

  // Accessors
  int constant_pool_index() const                { return _indices & 0xFFFF; }
  Bytecodes::Code bytecode_1() const             { return Bytecodes::cast((_indices >> 16) & 0xFF); }
  Bytecodes::Code bytecode_2() const             { return Bytecodes::cast((_indices >> 24) & 0xFF); }
  oop  f1() const                                { return _f1; }
  int  f2() const                                { return _f2; }
  int  holder_index() const                      { return _flags & 0xFFFF; }
  int  parameter_size() const                    { return _flags & 0xFF; }

  // Code generation support
  static int size()                              { return sizeof(ConstantPoolCacheEntry) / wordSize; }
  static int indices_offset_in_bytes()           { return (int)&((ConstantPoolCacheEntry*)NULL)->_indices; }
  static int f1_offset_in_bytes()                { return (int)&((ConstantPoolCacheEntry*)NULL)->_f1; }
  static int f2_offset_in_bytes()                { return (int)&((ConstantPoolCacheEntry*)NULL)->_f2; }
  static int flags_offset_in_bytes()             { return (int)&((ConstantPoolCacheEntry*)NULL)->_flags; }

  // GC Support
  void oops_do(void f(oop*));
  void oop_iterate(OopClosure* blk);
  void follow_contents();
  void adjust_pointers();

  // Debugging & Printing
  void print (outputStream* st, int index) const;
  void verify(outputStream* st) const;
};


// A constant pool cache is a runtime data structure set aside to a constant pool. The cache
// holds interpreter runtime information for all field access and invoke bytecodes. The cache
// is created and initialized before a class is actively used (i.e., initialized), the indivi-
// dual cache entries are filled at resolution (i.e., "link") time (see also: rewriter.*).

class constantPoolCacheOopDesc: public arrayOopDesc {
  friend class VMStructs;
 private:
  constantPoolOop _constant_pool;                // the corresponding constant pool

  // Sizing
  static int header_size()                       { return sizeof(constantPoolCacheOopDesc) / wordSize; }
  static int object_size(int length)             { return align_object_size(header_size() + length * ConstantPoolCacheEntry::size()); }

  // Helpers
  constantPoolOop*        constant_pool_addr()   { return &_constant_pool; }
  ConstantPoolCacheEntry* base() const           { return (ConstantPoolCacheEntry*)((address)this + base_offset_in_bytes()); }

  friend class constantPoolCacheKlass;

 public:
  // Initialization
  void initialize(GrowableArray<int>* inverse_index_map);

  // Accessors
  void set_constant_pool(constantPoolOop pool)   { _constant_pool = pool; }
  constantPoolOop constant_pool() const          { return _constant_pool; }
  ConstantPoolCacheEntry* entry_at(int i) const  { assert(0 <= i && i < length(), "index out of bounds"); return base() + i; }

  // Code generation
  static int base_offset_in_bytes()              { return sizeof(constantPoolCacheOopDesc); }
};

