/*
 * @(#)instanceKlass.cpp	1.216 00/11/29 17:43:05
 *
 * Copyright 1993-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_instanceKlass.cpp.incl"

bool instanceKlass::should_be_initialized() const {
  return !is_initialized();
}

klassVtable* instanceKlass::vtable() const {
  return new klassVtable(as_klassOop(), start_of_vtable(), vtable_length() / vtableEntry::size());
}

klassItable* instanceKlass::itable() const {
  return new klassItable(as_klassOop());
}

void instanceKlass::eager_initialize() {
  if (!EagerInitialization) return;

  if (this->is_not_initialized()) {
    // abort if the the class has a class initializer
    if (this->class_initializer() != NULL) return;

    // abort if it is java.lang.Object (initialization is handled in genesis)
    klassOop super = this->super();
    if (super == NULL) return;

    // abort if the super class should be initialized
    if (!instanceKlass::cast(super)->is_initialized()) return;

    // call body to expose the this pointer
    instanceKlassHandle this_oop(this->as_klassOop());
    eager_initialize_impl(this_oop);
  }
}


void instanceKlass::eager_initialize_impl(instanceKlassHandle this_oop) {
  EXCEPTION_MARK;
  ObjectLocker ol(this_oop, THREAD);

  // abort if someone beat us to the initialization
  if (!this_oop->is_not_initialized()) return;  // note: not equivalent to is_initialized()

  ClassState old_state = this_oop->_init_state;
  link_class_impl(this_oop, THREAD);
  if (HAS_PENDING_EXCEPTION) {
    CLEAR_PENDING_EXCEPTION;
    // abort if linking the class throws an exception
    this_oop->set_init_state (old_state);
  } else {
    // linking successfull, mark class as initialized
    this_oop->set_init_state (fully_initialized);
    // trace 
    if (TraceClassInitialization) {
      ResourceMark rm;
      tty->print_cr("[Initialized %s without side effects]", this_oop->external_name());
    }
  }
}


// See "The Virtual Machine Specification" section 2.16.5 for a detailed explanation of the class initialization
// process. The step comments refers to the procedure described in that section.
// Note: implementation moved to static method to expose the this pointer.
void instanceKlass::initialize(TRAPS) {
  if (this->should_be_initialized()) {
    HandleMark hm;
    instanceKlassHandle this_oop(this->as_klassOop());
    initialize_impl(this_oop, CHECK);
    // Note: at this point the class may be initialized
    //       OR it may be in the state of being initialized
    //       in case of recursive initialization!
  } else {
    assert(is_initialized(), "sanity check");
  }
}


void instanceKlass::verify_code(instanceKlassHandle this_oop, TRAPS) {
  // 1) Verify the bytecodes
  Verifier::verify_byte_codes(this_oop, CHECK);
}

void instanceKlass::add_loader_constraints(instanceKlassHandle this_oop, TRAPS) {
  // Walk over methods looking for methods declared in super-classes (overrides)
  // or methods declared in interfaces, and insist that the names in the 
  // signatures refer to the same classes.  See JLS 5.4.2.
  objArrayHandle methods = this_oop->methods();
  int number_of_methods = methods->length();
  Handle this_loader = this_oop->class_loader();
  for (int method_index = 0; method_index < number_of_methods; method_index += 1) {
    methodHandle method = (methodOop) methods->obj_at(method_index);
    // skip private, static and <init> methods
    if ((!method->is_private()) &&
      (!method->is_static()) &&
      (method->name() != vmSymbols::object_initializer_name())) {
      
      symbolHandle name = method->name();
      symbolHandle signature = method->signature();
      { 
        // First, look through superclasses of this class for overriding methods
        instanceKlassHandle super_klass = this_oop->super();
        while (super_klass.not_null()) {
          // lookup a matching method in the super class hierarchy
          if (super_klass->lookup_method(name(), signature()) == NULL) {
            break; // didn't find any match; get out
          }
          Handle super_loader = super_klass->class_loader();
          SystemDictionary::check_signature_loaders(signature, this_loader, super_loader, true, CHECK);
          super_klass = super_klass->super(); // continue to look
        }
      }
      {
        // Next, look through interfaces for implemented methods
        objArrayHandle interfaces = this_oop->transitive_interfaces();
        int number_of_interfaces = interfaces->length();
        for (int interface_index = 0; interface_index < number_of_interfaces; interface_index += 1) {
          instanceKlassHandle interface_klass = klassOop(interfaces->obj_at(interface_index));
          if (interface_klass->lookup_method(name(), signature()) == NULL) {
            break;  // didn't find any match; get out
          }
          Handle interface_loader = interface_klass->class_loader();
          SystemDictionary::check_signature_loaders(signature, this_loader, interface_loader, true, CHECK);
        }
      }
    }
  }
  {
    // Make sure the current class links with the system Throwable class. 
    // See bug 4157312.
    klassOop throwable_klass = SystemDictionary::throwable_klass();
    Handle throwable_loader = instanceKlass::cast(throwable_klass)->class_loader();
    if (throwable_loader() != this_loader()) {
      symbolHandle throwable_name = instanceKlass::cast(throwable_klass)->name();
      SystemDictionary::add_loader_constraint(throwable_name, this_loader, throwable_loader, CHECK);
    }
  }
}

void instanceKlass::link_class(TRAPS) {    
  assert(is_loaded(), "must be loaded");
  if (!is_linked()) {
    instanceKlassHandle this_oop(this->as_klassOop());  
    link_class_impl(this_oop, CHECK);
  }
}

void instanceKlass::link_class_impl(instanceKlassHandle this_oop, TRAPS) {
  // check for error state
  if (this_oop->is_in_error_state()) {
    ResourceMark rm;
    THROW_MSG(vmSymbols::java_lang_NoClassDefFoundError(), this_oop->external_name());  
  }
  // return if already verified
  if (this_oop->is_linked()) {
    return;
  }
  
  // link super class before linking this class
  instanceKlassHandle super(this_oop->super());
  if (super.not_null()) {
    if (super->is_interface()) {  // check if super class is an interface
      ResourceMark rm;
      Exceptions::fthrow(  
	THREAD_AND_LOCATION,
	vmSymbols::java_lang_IncompatibleClassChangeError(),
	"class %s has interface %s as super class",
	this_oop->external_name(),
	super->external_name()
	);
    }

    link_class_impl(super, CHECK);
  }
  
  // link all interfaces implemented by this class before linking this class
  objArrayHandle interfaces = this_oop->local_interfaces();
  int num_interfaces = interfaces->length();
  for (int index = 0; index < num_interfaces; index++) {
    HandleMark hm;
    instanceKlassHandle ih(klassOop(interfaces->obj_at(index)));
    link_class_impl(ih, CHECK);
  }
  
  // in case the class is linked in the process of linking its superclasses
  if (this_oop->is_linked()) {
    return;
  }
   
  // verification & rewriting
  {
    ObjectLocker ol(this_oop, THREAD);
    if (!this_oop->is_linked()) {
      verify_code(this_oop, CHECK);
      add_loader_constraints(this_oop, CHECK);
      // The rewriter must be called exactly once. Rewriting must happen after
      // verification but before the first method of the class is executed.
      Rewriter::rewrite(this_oop, CHECK); // No exception can happen here

      // Initialize the vtable and interface table after
      // methods have been rewritten
      { ResourceMark rm;
        this_oop->vtable()->initialize_vtable(); // No exception can happen here
        this_oop->itable()->initialize_itable();
      }
      this_oop->set_init_state(linked);
      if (jvmdi::enabled()) {
	Thread *thread = THREAD;
	assert(thread->is_Java_thread(), "thread->is_Java_thread()");
        jvmdi::post_class_prepare_event((JavaThread *) thread, this_oop());
      }
    }
  }    
}


void instanceKlass::initialize_impl(instanceKlassHandle this_oop, TRAPS) {
  // Make sure klass is linked (verified) before initialization
  // A class could already be verified, since it has been reflected upon.
  this_oop->link_class(CHECK);

  // refer to the JVM book page 47 for description of steps
  // Step 1
  { ObjectLocker ol(this_oop, THREAD);

    Thread *self = Thread::current();        

    // Step 2
    while(this_oop->is_being_initialized() && !this_oop->is_reentrant_initialization(self)) {      
      ol.wait(CHECK);      
    }

    // Step 3
    if (this_oop->is_being_initialized() && this_oop->is_reentrant_initialization(self))
      return;

    // Step 4 
    if (this_oop->is_initialized())
      return;

    // Step 5
    if (this_oop->is_in_error_state()) {
      THROW(vmSymbols::java_lang_NoClassDefFoundError());  
    }

    // Step 6
    this_oop->set_init_state(being_initialized);
    this_oop->set_init_thread(self);
  }
  
  // Step 7
  klassOop super_klass = this_oop->super();
  if (super_klass != NULL && !this_oop->is_interface() && Klass::cast(super_klass)->should_be_initialized()) {
    Klass::cast(super_klass)->initialize(THREAD);

    if (HAS_PENDING_EXCEPTION) {
      Handle e(PENDING_EXCEPTION);
      CLEAR_PENDING_EXCEPTION;
      {
        EXCEPTION_MARK;
        this_oop->set_initialization_state_and_notify(initialization_error, THREAD); // Locks object, set state, and notify all waiting threads
        CLEAR_PENDING_EXCEPTION;   // ignore any exception thrown, superclass initialization error is thrown below
      }
      THROW_OOP(e());
    }
  }

  // Step 8  
  this_oop->call_class_initializer(THREAD);  

  // Step 9
  if (!HAS_PENDING_EXCEPTION) {    
    this_oop->set_initialization_state_and_notify(fully_initialized, CHECK);
    { ResourceMark rm;
      debug_only(this_oop->vtable()->verify(tty, true);)
    }
  }
  else {    
    // Step 10 and 11
    Handle e(THREAD, PENDING_EXCEPTION);
    CLEAR_PENDING_EXCEPTION;
    { 
      EXCEPTION_MARK;
      this_oop->set_initialization_state_and_notify(initialization_error, THREAD);
      CLEAR_PENDING_EXCEPTION;   // ignore any exception thrown, class initialization error is thrown below
    }
    if (e->is_a(SystemDictionary::error_klass())) {
      THROW_OOP(e());
    } else {
      JavaCallArguments args(e);
      THROW_ARG(vmSymbols::java_lang_ExceptionInInitializerError(),
                vmSymbols::throwable_void_signature(),
                &args);      
    }
  }
}
  

// Note: implementation moved to static method to expose the this pointer.
void instanceKlass::set_initialization_state_and_notify(ClassState state, TRAPS) {
  set_initialization_state_and_notify_impl(this->as_klassOop(), state, CHECK);
}

void instanceKlass::set_initialization_state_and_notify_impl(instanceKlassHandle this_oop, ClassState state, TRAPS) {    
  ObjectLocker ol(this_oop, THREAD);   
  this_oop->set_init_state(state);
  ol.notify_all(CHECK);  
}

void instanceKlass::add_implementor(klassOop k) {
  assert(Compile_lock->owned_by_self(), "");
  // Filter out subinterfaces because they're already in the subklass list
  if (instanceKlass::cast(k)->is_interface()) return;

  // Update number of implementors
  _nof_implementors++;

  // Set implementor, if we have one
  if (_nof_implementors == 1) {
    assert(_implementor == NULL, "should be exactly one implementor");
    _implementor = k;
  } else {
    _implementor = NULL;
  }

  // The implementor also implements the transitive_interfaces
  for (int index = 0; index < local_interfaces()->length(); index++) {
    instanceKlass::cast(klassOop(local_interfaces()->obj_at(index)))->add_implementor(k);
  }
}

void instanceKlass::init_implementor() {
  _implementor = NULL;  
  _nof_implementors = 0;
}


void instanceKlass::process_interfaces() {
  // link this class into the implementors list of every interface it implements
  KlassHandle this_as_oop = this->as_klassOop();
  for (int i = local_interfaces()->length() - 1; i >= 0; i--) {
    assert(local_interfaces()->obj_at(i)->is_klass(), "must be a klass");
    instanceKlass* interf = instanceKlass::cast(klassOop(local_interfaces()->obj_at(i)));
    assert(interf->is_interface(), "expected interface");
    interf->add_implementor(this_as_oop()); 
  }
}

bool instanceKlass::compute_is_subtype_of(klassOop k) {
  if (Klass::cast(k)->is_interface()) {
    return implements_interface(k);
  } else {
    return Klass::compute_is_subtype_of(k);
  }
}

bool instanceKlass::implements_interface(klassOop k) const {
  if (as_klassOop() == k) return true;
  assert(Klass::cast(k)->is_interface(), "should be an interface class");
  for (int i = 0; i < transitive_interfaces()->length(); i++) {
    if (transitive_interfaces()->obj_at(i) == k) {
      return true;
    }
  }
  return false;
}

objArrayOop instanceKlass::allocate_objArray(int n, int length, TRAPS) {
  if (length < 0) THROW_0(vmSymbols::java_lang_NegativeArraySizeException());
  if (length > arrayOopDesc::max_array_length(T_OBJECT)) {
    THROW_OOP_0(Universe::out_of_memory_error_instance());
  }
  int size = objArrayOopDesc::object_size(length);
  KlassHandle h_ak = array_klass(n, CHECK_0);
  objArrayOop o = (objArrayOop)Universe::clear_allocate_array(h_ak, size, length, CHECK_0);
  return o;
}

instanceOop instanceKlass::register_finalizer(instanceOop i, TRAPS) {
  if (TraceFinalizerRegistration) {
    tty->print("Registered ");
    i->print_value_on(tty);
    tty->print_cr(" (0x%x) as finalizable", i);
  }
  instanceHandle h_i(THREAD, i);
  // Pass the handle as argument, JavaCalls::call expects oop as jobjects
  JavaValue result(T_VOID);
  JavaCallArguments args(h_i);  
  JavaCalls::call(&result, Universe::finalizer_register_method(), &args, CHECK_0);
  return h_i();
}

instanceOop instanceKlass::allocate_instance(TRAPS) {
  bool has_finalizer_flag = has_finalizer(); // Query before possible GC
  KlassHandle h_k(THREAD, as_klassOop());
  instanceOop i;
 
  if (UseSpecLock
    && SpecLock::is_speclock_class(h_k())){
    i = (instanceOop)Universe::clear_allocate(h_k, size_helper(), CHECK_0);
    i->set_mark_speclock();
    i->set_klass(h_k());
    JavaThread* jth = JavaThread::current();
    int tid = jth->get_fast_thread_id();
    i->int_field_put(oopDesc::header_size(), tid);
#ifdef ASSERT
    // Occassionally seen some problems in setting the thread id. 
    // Keep this assert in to make sure.
    jth->set_fast_thread_id();
    int newtid = jth->get_fast_thread_id();
    assert(tid == newtid, " Change in fast thread id???");
#endif
   
  } else {
    i = (instanceOop)Universe::clear_allocate(h_k, size_helper(), CHECK_0);
    i->set_mark();
    i->set_klass(h_k());
  }
  if (has_finalizer_flag && RegisterFinalizers) {
    i = register_finalizer(i, CHECK_0);
  }
  return i;
}

instanceOop instanceKlass::allocate_permanent_instance(TRAPS) {
  bool has_finalizer_flag = has_finalizer(); // Query before possible GC
  KlassHandle h_k(THREAD, as_klassOop());
  instanceOop i = (instanceOop)Universe::clear_allocate_permanent(h_k, size_helper(), CHECK_0);
  if (has_finalizer_flag && RegisterFinalizers) {
    i = register_finalizer(i, CHECK_0);
  }
  return i;
}

void instanceKlass::check_valid_for_instantiation(bool throwError, TRAPS) {
  if (is_interface() || is_abstract()) {
    ResourceMark rm;
    THROW_MSG(throwError ? vmSymbols::java_lang_InstantiationError()
	      : vmSymbols::java_lang_InstantiationException(), external_name());
  }
  if (as_klassOop() == SystemDictionary::class_klass()) {
    ResourceMark rm;
    THROW_MSG(throwError ? vmSymbols::java_lang_IllegalAccessError()
	      : vmSymbols::java_lang_IllegalAccessException(), external_name());
  }
}

klassOop instanceKlass::array_klass_impl(bool or_null, int n, TRAPS) {
  instanceKlassHandle this_oop(as_klassOop());
  return array_klass_impl(this_oop, or_null, n, THREAD);
}

klassOop instanceKlass::array_klass_impl(instanceKlassHandle this_oop, bool or_null, int n, TRAPS) {    
  if (this_oop->array_klasses() == NULL) {
    if (or_null) return NULL;

    // Atomic creation of array_klasses
    MutexLocker mc(Compile_lock);   // for vtables
    MutexLocker ma(MultiArray_lock);

    // Check if update has already taken place    
    if (this_oop->array_klasses() == NULL) {
      objArrayKlassKlass* oakk = (objArrayKlassKlass*)Universe::objArrayKlassKlassObj()->klass_part();
      klassOop  k = oakk->allocate_objArray_klass(1, this_oop, CHECK_0);                  
      this_oop->set_array_klasses(k);
    }
  }
  // _this will always be set at this point
  objArrayKlass* oak = (objArrayKlass*)this_oop->array_klasses()->klass_part();
  if (or_null) {
    return oak->array_klass_or_null(n);
  }
  return oak->array_klass(n, CHECK_0);
}

klassOop instanceKlass::array_klass_impl(bool or_null, TRAPS) {
  return array_klass_impl(or_null, 1, THREAD);
}

void instanceKlass::call_class_initializer(TRAPS) {
  call_class_initializer_impl(as_klassOop(), THREAD);

}

static int call_class_initializer_impl_counter = 0;   // for debugging

methodOop instanceKlass::class_initializer() {
  return find_method(vmSymbols::class_initializer_name(), vmSymbols::void_method_signature()); 
}

void instanceKlass::call_class_initializer_impl(instanceKlassHandle this_oop, TRAPS) {  
  methodHandle h_method(THREAD, this_oop->class_initializer());
  assert(!this_oop->is_initialized(), "we cannot initialize twice");
  if (TraceClassInitialization) {
    tty->print("%d Initializing ", call_class_initializer_impl_counter++);
    this_oop->name()->print_value();
    tty->print_cr("%s (0x%x)", h_method() == NULL ? "(no method)" : "", this_oop());
  }
  if (h_method() != NULL) {  
    JavaCallArguments args; // No arguments
    JavaValue result(T_VOID);
    JavaCalls::call(&result, h_method, &args, CHECK); // Static call (no args)
  }
}


OopMapCacheEntry* instanceKlass::mask_for(methodOop method, int bci, bool fill_if_absent) {
  if (_oop_map_cache == NULL) {
    // First time use. Allocate a cache in C heap
    if (!fill_if_absent) return NULL;
    _oop_map_cache = new OopMapCache();
  }
  return _oop_map_cache->lookup(method, bci, fill_if_absent);
}


bool instanceKlass::find_local_field(symbolOop name, symbolOop sig, fieldDescriptor* fd) const {  
  const int n = fields()->length();
  for (int i = 0; i < n; i += 5) {
    int name_index = fields()->ushort_at(i + 1);
    int sig_index  = fields()->ushort_at(i + 2);
    symbolOop f_name = constants()->symbol_at(name_index);
    symbolOop f_sig  = constants()->symbol_at(sig_index);
    if (f_name == name && f_sig == sig) {
      fd->initialize(as_klassOop(), i);
      return true;
    }
  }
  return false;
}


klassOop instanceKlass::find_interface_field(symbolOop name, symbolOop sig, fieldDescriptor* fd) const {
  const int n = local_interfaces()->length();
  for (int i = 0; i < n; i++) {
    klassOop intf1 = klassOop(local_interfaces()->obj_at(i));
    assert(Klass::cast(intf1)->is_interface(), "just checking type");
    // search for field in current interface
    if (instanceKlass::cast(intf1)->find_local_field(name, sig, fd)) {
      assert(fd->is_static(), "interface field must be static");
      return intf1;
    }
    // search for field in direct superinterfaces
    klassOop intf2 = instanceKlass::cast(intf1)->find_interface_field(name, sig, fd);
    if (intf2 != NULL) return intf2;
  }
  // otherwise field lookup fails
  return NULL;
}


klassOop instanceKlass::find_field(symbolOop name, symbolOop sig, fieldDescriptor* fd) const {
  // search order according to newest JVM spec (5.4.3.2, p.167).
  // 1) search for field in current klass
  if (find_local_field(name, sig, fd)) {
    return as_klassOop();
  }
  // 2) search for field recursively in direct superinterfaces
  { klassOop intf = find_interface_field(name, sig, fd);
    if (intf != NULL) return intf;
  }
  // 3) apply field lookup recursively if superclass exists
  { klassOop supr = super();
    if (supr != NULL) return instanceKlass::cast(supr)->find_field(name, sig, fd);
  }
  // 4) otherwise field lookup fails
  return NULL;
}


klassOop instanceKlass::jni_find_field(symbolOop name, symbolOop sig, bool is_static, fieldDescriptor* fd) const {
  // search order according to newest JVM spec (5.4.3.2, p.167).
  // 1) search for field in current klass
  if (find_local_field(name, sig, fd)) {
    if (fd->is_static() == is_static) return as_klassOop();
  }
  // 2) search for field recursively in direct superinterfaces
  if (is_static) {
    klassOop intf = find_interface_field(name, sig, fd);
    if (intf != NULL) return intf;
  }
  // 3) apply field lookup recursively if superclass exists
  { klassOop supr = super();
    if (supr != NULL) return instanceKlass::cast(supr)->jni_find_field(name, sig, is_static, fd);
  }
  // 4) otherwise field lookup fails
  return NULL;
}


bool instanceKlass::jni_find_local_field_from_offset(int offset, bool is_static, fieldDescriptor* fd) const {  
  int length = fields()->length();
  for (int i = 0; i < length; i += 5) {
    if (fields()->ushort_at(i + 4) == offset) {
      fd->initialize(as_klassOop(), i);      
      if (fd->is_static() == is_static) return true;
    }
  }
  return false;
}


bool instanceKlass::jni_find_field_from_offset(int offset, bool is_static, fieldDescriptor* fd) const {
  klassOop klass = as_klassOop();
  while (klass != NULL) {
    if (instanceKlass::cast(klass)->jni_find_local_field_from_offset(offset, is_static, fd)) {
      return true;
    }
    klass = Klass::cast(klass)->super();
  }
  return false;
}


void instanceKlass::methods_do(void f(methodOop method)) {
  int len = methods()->length();
  for (int index = 0; index < len; index++) {
    methodOop m = methodOop(methods()->obj_at(index));
    assert(m->is_method(), "must be method");
    f(m);
  }
}


void instanceKlass::do_local_static_fields(void f(fieldDescriptor*, oop), oop obj) {  
  fieldDescriptor fd;
  int length = fields()->length();
  for (int i = 0; i < length; i += 5) {
    fd.initialize(as_klassOop(), i);
    if (fd.is_static()) f(&fd, obj);
  }
}


void instanceKlass::do_local_static_fields(void f(fieldDescriptor*, TRAPS), TRAPS) {
  instanceKlassHandle h_this(THREAD, as_klassOop());
  do_local_static_fields_impl(h_this, f, CHECK);
}
 

void instanceKlass::do_local_static_fields_impl(instanceKlassHandle this_oop, void f(fieldDescriptor* fd, TRAPS), TRAPS) {
  fieldDescriptor fd;
  int length = this_oop->fields()->length();
  for (int i = 0; i < length; i += 5) {
    fd.initialize(this_oop(), i);
    if (fd.is_static()) { f(&fd, CHECK); } // Do NOT remove {}! (CHECK macro expands into several statements)
  }
}


void instanceKlass::do_nonstatic_fields(void f(fieldDescriptor*, oop), oop obj) {
  fieldDescriptor fd;
  instanceKlass* super = superklass();
  if (super != NULL) {
    super->do_nonstatic_fields(f, obj);
  }
  int length = fields()->length();
  for (int i = 0; i < length; i += 5) {
    fd.initialize(as_klassOop(), i);
    if (!(fd.is_static())) f(&fd, obj);
  }  
}


void instanceKlass::array_klasses_do(void f(klassOop k)) {
  if (array_klasses() != NULL)
    arrayKlass::cast(array_klasses())->array_klasses_do(f);
}


void instanceKlass::with_array_klasses_do(void f(klassOop k)) {
  f(as_klassOop());
  array_klasses_do(f);
}

#ifdef ASSERT
static int linear_search(objArrayHandle methods, symbolHandle name, symbolHandle signature) {
  int len = methods->length();
  for (int index = 0; index < len; index++) {
    methodOop m = (methodOop)(methods->obj_at(index));
    assert(m->is_method(), "must be method");
    if (m->signature() == signature() && m->name() == name()) {
       return index;
    }
  }
  return -1;
}
#endif

methodOop instanceKlass::find_method(symbolHandle name, symbolHandle signature) const {
  objArrayHandle methods = instanceKlass::cast(as_klassOop())->methods();
  return instanceKlass::find_method(methods, name, signature);
}

methodOop instanceKlass::find_method(objArrayHandle methods, symbolHandle name, symbolHandle signature) {
  int len = methods->length();
  // methods are sorted, so do binary search
  int l = 0;
  int h = len - 1;
  while (l <= h) {
    int mid = (l + h) >> 1;
    methodOop m = (methodOop)methods->obj_at(mid);
    assert(m->is_method(), "must be method");
    int res = m->name()->fast_compare(name());
    if (res == 0) {
      // found matching name; do linear search to find matching signature
      // first, quick check for common case 
      if (m->signature() == signature()) return m;
      // search downwards through overloaded methods
      int i;
      for (i = mid - 1; i >= l; i--) {
        methodOop m = (methodOop)methods->obj_at(i);
        assert(m->is_method(), "must be method");
        if (m->name() != name()) break;
        if (m->signature() == signature()) return m;
      }
      // search upwards
      for (i = mid + 1; i <= h; i++) {
        methodOop m = (methodOop)methods->obj_at(i);
        assert(m->is_method(), "must be method");
        if (m->name() != name()) break;
        if (m->signature() == signature()) return m;
      }
      // not found
#ifdef ASSERT
      int index = linear_search(methods, name, signature);
      if (index != -1) fatal1("binary search bug: should have found entry %d", index);
#endif
      return NULL;
    } else if (res < 0) {
      l = mid + 1;
    } else {
      h = mid - 1;
    }
  }
#ifdef ASSERT
  int index = linear_search(methods, name, signature);
  if (index != -1) fatal1("binary search bug: should have found entry %d", index);
#endif
  return NULL;
} 

int instanceKlass::method_index_for(methodOop method, TRAPS) const {
  int len = methods()->length();
  for (int index = 0; index < len; index++) {
    if (methodOop(methods()->obj_at(index)) == method) {
      return index;
    }
  }
  THROW_(vmSymbols::java_lang_NoSuchMethodException(), -1);
}


methodOop instanceKlass::uncached_lookup_method(symbolOop name, symbolOop signature) const {  
  instanceKlassHandle h_klass(as_klassOop());
  symbolHandle h_name(name);
  symbolHandle h_signature(signature);
   
  while (h_klass() != NULL) {
    methodOop method = h_klass->find_method(h_name, h_signature);
    if (method != NULL) return method;
    h_klass = h_klass->super();
  }
  return NULL;
}

// lookup a method in all the interfaces that this class implements
methodOop instanceKlass::lookup_method_in_all_interfaces(symbolOop name, 
							 symbolOop signature) const {
  objArrayOop all_ifs = instanceKlass::cast(as_klassOop())->transitive_interfaces();
  int num_ifs = all_ifs->length();
  instanceKlass *ik = NULL;
  for (int i = 0; i < num_ifs; i++) {
    ik = instanceKlass::cast(klassOop(all_ifs->obj_at(i)));
    methodOop m = ik->lookup_method(name, signature);
    if (m != NULL) {
      return m;
    }
  }
  return NULL;
}


JNIid* instanceKlass::jni_id_for_impl(instanceKlassHandle this_oop, int offset) {
  MutexLocker ml(JNIIdentifier_lock);
  // Retry lookup after we got the lock
  JNIid* probe = this_oop->jni_ids() == NULL ? NULL : this_oop->jni_ids()->find(offset);
  if (probe == NULL) {
    // Slow case, allocate new static field identifier
    probe = new JNIid(this_oop->as_klassOop(), offset, this_oop->jni_ids());
    this_oop->set_jni_ids(probe);
  }
  return probe;
}


JNIid* instanceKlass::jni_id_for(int offset) {
  JNIid* probe = jni_ids() == NULL ? NULL : jni_ids()->find(offset);
  if (probe == NULL) {
    instanceKlassHandle this_oop(this->as_klassOop());
    probe = jni_id_for_impl(this_oop, offset);
  }
  return probe;
}


void instanceKlass::follow_static_fields() {
  oop* start = start_of_static_fields();
  oop* end   = start + static_oop_field_size();
  while (start < end) {
    if (*start != NULL) {
      assert(Universe::is_heap(*start), "should be in heap");
      MarkSweep::mark_and_push(start);
    }
    start++;
  }
}


void instanceKlass::adjust_static_fields() {
  oop* start = start_of_static_fields();
  oop* end   = start + static_oop_field_size();
  while (start < end) {
    MarkSweep::adjust_pointer(start);
    start++;
  }
}


void instanceKlass::oop_follow_contents(oop obj) {
  assert (obj!=NULL, "can't follow the content of NULL object");
  obj->follow_header();
  OopMapBlock* map     = start_of_nonstatic_oop_maps();
  OopMapBlock* end_map = map + nonstatic_oop_map_size();
  while (map < end_map) {
    oop* start = obj->obj_field_addr(map->offset());
    oop* end   = start + map->length();
    while (start < end) {
      if (*start != NULL) {
        assert(Universe::is_heap(*start), "should be in heap");
        MarkSweep::mark_and_push(start);
      }
      start++;
    }
    map++;
  }
}


void instanceKlass::oop_oop_iterate(oop obj, OopClosure* closure) {
  // header
  obj->oop_iterate_header(closure);
  // instance variables
  iterate_nonstatic_fields(obj, closure);
}


void instanceKlass::iterate_nonstatic_fields(oop obj, OopClosure* closure) {
  OopMapBlock* map     = start_of_nonstatic_oop_maps();
  OopMapBlock* end_map = map + nonstatic_oop_map_size();
  while (map < end_map) {
    oop* start = obj->obj_field_addr(map->offset());
    oop* end   = start + map->length();
    while (start < end) {
      assert(Universe::is_heap_or_null(*start), "should be in heap");
      closure->do_oop(start);
      start++;
    }
    map++;
  }
}


void instanceKlass::iterate_static_fields(OopClosure* closure) {
  oop* start = start_of_static_fields();
  oop* end   = start + static_oop_field_size();
  while (start < end) {
    assert(Universe::is_heap_or_null(*start), "should be in heap");
    closure->do_oop(start);
    start++;
  }
}


int instanceKlass::oop_scavenge_contents(oop obj) {
  // Compute oopmap block range. The common case is nonstatic_oop_map_size == 1.
  OopMapBlock* map     = start_of_nonstatic_oop_maps();
  OopMapBlock* end_map = map + nonstatic_oop_map_size();
  // Iterate over oopmap blocks
  while (map < end_map) {
    // Compute oop range for this block
    oop* start = obj->obj_field_addr(map->offset());
    oop* end   = start + map->length();
    // Iterate over oops
    if (PrefetchScavengeFieldsAhead != 0) {
      while (start < end) {
        oop* p = start + PrefetchScavengeFieldsAhead;
        if (p < end) {
          atomic::prefetch_read(*p);   // Prefetch of null is okay
        }
        if (Scavenge::should_scavenge(*start)) {
          assert(Universe::is_heap(*start), "should be in heap");
          Scavenge::scavenge(start);
        }
        start++;
      }
    } else {
      while (start < end) {
        if (Scavenge::should_scavenge(*start)) {
          assert(Universe::is_heap(*start), "should be in heap");
          Scavenge::scavenge(start);
        }
        start++;
      }
    }
    map++;
  }
  return size_helper();
}


int instanceKlass::oop_scavenge_contents(oop obj, oop* bottom, oop* top) {
  // Compute oopmap block range. The common case is nonstatic_oop_map_size == 1.
  OopMapBlock* map     = start_of_nonstatic_oop_maps();
  OopMapBlock* end_map = map + nonstatic_oop_map_size();
  // Iterate over oopmap blocks
  while (map < end_map) {
    // Compute oop range for this block
    oop* start = obj->obj_field_addr(map->offset());
    oop* end   = start + map->length();
    // Adjust for boundaries
    if (start < bottom) start = bottom;
    if (end > top)      end = top;
    // Iterate over oops
    while (start < end) {
      if (start >= bottom && Scavenge::should_scavenge(*start)) {
        assert(Universe::is_heap(*start), "should be in heap");
        Scavenge::scavenge(start);
      }
      start++;
    }
    map++;
  }
  return size_helper();
}


int instanceKlass::oop_adjust_pointers(oop obj) {
  int size = size_helper();

  // Compute oopmap block range. The common case is nonstatic_oop_map_size == 1.
  OopMapBlock* map     = start_of_nonstatic_oop_maps();
  OopMapBlock* end_map = map + nonstatic_oop_map_size();
  // Iterate over oopmap blocks
  while (map < end_map) {
    // Compute oop range for this block
    oop* start = obj->obj_field_addr(map->offset());
    oop* end   = start + map->length();
    // Iterate over oops
    while (start < end) {
      assert(Universe::is_heap_or_null(*start), "should be in heap");
      MarkSweep::adjust_pointer(start);
      start++;
    }
    map++;
  }

  obj->adjust_header();
  return size;
}


void instanceKlass::scavenge_static_fields() {
  // Compute oop range
  oop* start = start_of_static_fields();
  oop* end   = start + static_oop_field_size();
  // Iterate over oops
  while (start < end) {
    if (Scavenge::should_scavenge(*start)) {
      assert(Universe::is_heap(*start), "should be in heap");
      Scavenge::scavenge(start);
    }
    start++;
  }
}


void instanceKlass::scavenge_static_fields(oop* bottom, oop* top) {
  // Compute oop range
  oop* start = start_of_static_fields();
  oop* end   = start + static_oop_field_size();
  // Adjust for boundaries
  if (start < bottom) start = bottom;
  if (end > top)      end = top;
  // Iterate over oops
  while (start < end) {
    if (Scavenge::should_scavenge(*start)) {
      assert(Universe::is_heap(*start), "should be in heap");
      Scavenge::scavenge(start);
    }
    start++;
  }
}


// This klass is alive but the implementor link is not followed/updated.
// Subklass and sibling links are handled by Klass::follow_weak_klass_links
void instanceKlass::follow_weak_klass_links() {
  assert(as_klassOop()->is_gc_marked(), "just checking");
  if (ClassUnloading) {
    if (implementor() != NULL) { 
      if (!implementor()->is_gc_marked()) {
        assert(nof_implementors() == 1, "just checking");
        init_implementor();
      } else {
        MarkSweep::mark_and_push(adr_implementor());
      }
    }
  } else {
    MarkSweep::mark_and_push(adr_implementor());
  }
  Klass::follow_weak_klass_links();
}


static void clear_all_breakpoints(methodOop m) {
  m->clear_all_breakpoints();
}


void instanceKlass::release_C_heap_structures() {
  // Deallocate oop map cache
  if (_oop_map_cache != NULL) {
    delete _oop_map_cache;
    _oop_map_cache = NULL;
  }
  // Deallocate JNI identifiers
  JNIid::deallocate(jni_ids());
  set_jni_ids(NULL);

  // Deallocate breakpoint records
  if (breakpoints() != 0x0) {
    methods_do(clear_all_breakpoints);
    assert(breakpoints() == 0x0, "should have cleared breakpoints");
  }
}

const char* instanceKlass::external_name() const{
  char* str    = name()->as_C_string();
  int   length = strlen(str);
  for (int index = 0; index < length; index++) {
    if (str[index] == '/') {
      str[index] = '.';
    }
  }
  return str;
}

char* instanceKlass::signature_name() const {
  const char* src = (const char*) (name()->as_C_string());
  const int src_length = strlen(src);
  char* dest = NEW_RESOURCE_ARRAY(char, src_length + 3);
  int src_index = 0;
  int dest_index = 0;
  dest[dest_index++] = 'L';
  while (src_index < src_length) {
    dest[dest_index++] = src[src_index++];
  }
  dest[dest_index++] = ';';
  dest[dest_index] = '\0';
  return dest;
}

// different verisons of is_same_class_package
bool instanceKlass::is_same_class_package(klassOop class2) {
  klassOop class1 = as_klassOop();
  oop classloader1 = instanceKlass::cast(class1)->class_loader();
  symbolOop classname1 = Klass::cast(class1)->name();

  if (Klass::cast(class2)->oop_is_objArray()) {
    class2 = objArrayKlass::cast(class2)->bottom_klass();
  }
  oop classloader2;  
  if (Klass::cast(class2)->oop_is_instance()) {
    classloader2 = instanceKlass::cast(class2)->class_loader();
  } else {
    assert(Klass::cast(class2)->oop_is_typeArray(), "should be type array");
    classloader2 = NULL;
  }
  symbolOop classname2 = Klass::cast(class2)->name();

  return instanceKlass::is_same_class_package(classloader1, classname1,
					      classloader2, classname2);
}

bool instanceKlass::is_same_class_package(oop classloader2, symbolOop classname2) {
  klassOop class1 = as_klassOop();
  oop classloader1 = instanceKlass::cast(class1)->class_loader();
  symbolOop classname1 = Klass::cast(class1)->name();

  return instanceKlass::is_same_class_package(classloader1, classname1,
					      classloader2, classname2);
}

// return true if two classes are in the same package, classloader 
// and classname information is enough to determine a class's package
bool instanceKlass::is_same_class_package(oop class_loader1, symbolOop class_name1, 
					  oop class_loader2, symbolOop class_name2) {
  if (class_loader1 != class_loader2) {
    return false;
  } else {
    ResourceMark rm;

    // The symbolOop's are in UTF8 encoding. Since we only need to check explicitly
    // for ASCII characters ('/', 'L', '['), we can keep them in UTF8 encoding.
    // Otherwise, we just compare jbyte values between the strings.
    jbyte *name1 = class_name1->base();
    jbyte *name2 = class_name2->base();

    jbyte *last_slash1 = UTF8::strrchr(name1, class_name1->utf8_length(), '/');
    jbyte *last_slash2 = UTF8::strrchr(name2, class_name2->utf8_length(), '/');
    
    if ((last_slash1 == NULL) || (last_slash2 == NULL)) {
      // One of the two doesn't have a package.  Only return true
      // if the other one also doesn't have a package.
      return last_slash1 == last_slash2; 
    } else {
      // Skip over '['s
      if (*name1 == '[') {
        do {
          name1++;
        } while (*name1 == '[');
        if (*name1 != 'L') {
	  // Something is terribly wrong.  Shouldn't be here.
          return false;
        }
      }
      if (*name2 == '[') {
        do {
          name2++;
        } while (*name2 == '[');
        if (*name2 != 'L') {
	  // Something is terribly wrong.  Shouldn't be here.
          return false;
        }
      }

      // Check that package part is identical
      int length1 = last_slash1 - name1;
      int length2 = last_slash2 - name2;

      return UTF8::equal(name1, length1, name2, length2);      
    }
  }
}


jint instanceKlass::modifier_flags(TRAPS) const {
  klassOop k = as_klassOop();
  jint access = access_flags().as_int();

  // But check if it happens to be member class.
  typeArrayOop inner_class_list = inner_classes();
  int length = inner_class_list->length();
  assert (length % instanceKlass::inner_class_next_offset == 0, "just checking");
  if (length > 0) {
    typeArrayHandle inner_class_list_h(THREAD, inner_class_list);
    instanceKlassHandle ik(THREAD, k);
    for (int i = 0; i < length; i += instanceKlass::inner_class_next_offset) {
      int ioff = inner_class_list_h->ushort_at(i + instanceKlass::inner_class_inner_class_info_offset);

      // only look at classes that are already loaded
      // since we are looking for the flags for our self.
      klassOop inner_oop = constantPoolOopDesc::klass_at_if_loaded(ik->constants(), ioff);
      if ((inner_oop != NULL) && (ik->as_klassOop() == inner_oop)) {
        // This is really a member class.
        access = inner_class_list_h->ushort_at(i + instanceKlass::inner_class_access_flags_offset);
        break;
      }
    }
  }
  // Remember to strip ACC_SUPER bit
  return (access & (~JVM_ACC_SUPER)) & JVM_ACC_WRITTEN_FLAGS;
}

jint instanceKlass::class_status() const {
  jint result = 0;

  if (is_linked()) {
    result |= JVMDI_CLASS_STATUS_VERIFIED | JVMDI_CLASS_STATUS_PREPARED;
  }

  if (is_initialized()) {
    assert(is_linked(), "Class status is not consistent");
    result |= JVMDI_CLASS_STATUS_INITIALIZED;
  }
  if (is_in_error_state()) {
    result |= JVMDI_CLASS_STATUS_ERROR;
  }
  return result;
}


// On-stack replacement stuff
#ifndef CORE

void instanceKlass::add_osr_nmethod(nmethod* n) {
  // only one compilation can be active
  NEEDS_CLEANUP
  //COMPILER1_ONLY(Unimplemented(); /*assert(C1Compiler::is_active(), "");*/)
  // This is a short non-blocking critical region, so the no safepoint check is ok. 
  OsrList_lock->lock_without_safepoint_check();
  assert(n->is_osr_method(), "wrong kind of nmethod");  
  n->set_link(osr_nmethods_head());
  set_osr_nmethods_head(n);  
  // Remember to unlock again
  OsrList_lock->unlock();
}


void instanceKlass::remove_osr_nmethod(nmethod* n) {
  // This is a short non-blocking critical region, so the no safepoint check is ok. 
  OsrList_lock->lock_without_safepoint_check();
  assert(n->is_osr_method(), "wrong kind of nmethod");
  nmethod* last = NULL;
  nmethod* cur  = osr_nmethods_head();
  // Search for match
  while(cur != NULL && cur != n) {
    last = cur;
    cur = cur->link();    
  }   
  if (cur == n) {
    if (last == NULL) {
      // Remove first element
      set_osr_nmethods_head(osr_nmethods_head()->link());      
    } else {
      last->set_link(cur->link());
    }
  }
  n->set_link(NULL);  
  // Remember to unlock again
  OsrList_lock->unlock();
}

nmethod* instanceKlass::lookup_osr_nmethod(const methodOop m, int bci) const {
  // This is a short non-blocking critical region, so the no safepoint check is ok. 
  OsrList_lock->lock_without_safepoint_check();
  nmethod* osr = osr_nmethods_head();
  while (osr != NULL) {
    assert(osr->is_osr_method(), "wrong kind of nmethod found in chain");
    if (osr->method() == m && osr->osr_entry_bci() == bci) {
      // Found a match - return it.        
      OsrList_lock->unlock();
      return osr;
    }    
    osr = osr->link();
  }
  OsrList_lock->unlock();
  return NULL;
}
#endif

// -----------------------------------------------------------------------------------------------------

#ifndef PRODUCT

// Printing

static outputStream* printing_stream = NULL;

static void print_nonstatic_field(fieldDescriptor* fd, oop obj) {
  assert(printing_stream != NULL, "Invalid printing stream");
  printing_stream->print("   - ");
  fd->print_on_for(printing_stream, obj);
  printing_stream->cr();
}


void instanceKlass::oop_print_on(oop obj, outputStream* st) {
  Klass::oop_print_on(obj, st);
  if (WizardMode) {
    st->print("header: ");
    obj->mark()->print_on(tty);
    tty->cr();
  }

  st->print_cr("fields:");
  printing_stream = st;
  do_nonstatic_fields(print_nonstatic_field, obj);
  printing_stream = NULL;

  if (as_klassOop() == SystemDictionary::class_klass()) {
    klassOop mirrored_klass = java_lang_Class::as_klassOop(obj);
    st->print("   - fake entry for mirror: ");
    mirrored_klass->print_value_on(st);
    st->cr();
  }
}

void instanceKlass::oop_print_value_on(oop obj, outputStream* st) {
  st->print("a ");
  name()->print_value_on(st);
  obj->print_address_on(st);
}

const char* instanceKlass::internal_name() const {
  return external_name();
}



// Verification

class VerifyFieldClosure: public OopClosure {
 public:
  void do_oop(oop* p) {
    guarantee(Universe::is_heap(oop(p)), "should be in heap");
    guarantee((*p)->safepoint_is_oop_or_null(), "should be in heap");
  }
};


void instanceKlass::oop_verify_on(oop obj, outputStream* st) {
  Klass::oop_verify_on(obj, st);
  VerifyFieldClosure blk;
  iterate_nonstatic_fields(obj, &blk);
}


void instanceKlass::verify_class_klass_nonstatic_oop_maps(klassOop k) {
  // Verify that java.lang.Class instances have a fake oop field added.
  instanceKlass* ik = instanceKlass::cast(k);

  // Check that we have the right class
  static bool first_time = true;
  guarantee(k == SystemDictionary::class_klass() && first_time, "Invalid verify of maps");
  first_time = false;
  const int extra = java_lang_Class::number_of_fake_oop_fields;
  guarantee(ik->nonstatic_field_size() == extra, "just checking");
  guarantee(ik->nonstatic_oop_map_size() == 1, "just checking");
  guarantee(ik->size_helper() == align_object_size(instanceOopDesc::header_size() + extra), "just checking");

  // Check that the map is (2,extra)
  int offset = java_lang_Class::klass_offset + instanceOopDesc::header_size();
  OopMapBlock* map = ik->start_of_nonstatic_oop_maps();
  guarantee(map->offset() == offset && map->length() == extra, "just checking");
}

#endif


JNIid::JNIid(klassOop holder, int offset, JNIid* next) {
  _holder = holder;
  _offset = offset;
  _next = next;
  debug_only(_is_static_field_id = false;)
  debug_only(_is_method_id = false;)
}


JNIid* JNIid::find(int offset) {
  JNIid* current = this;
  while (current != NULL) {
    if (current->offset() == offset) return current;
    current = current->next();
  }
  return NULL;
}


methodOop JNIid::method() {
  assert(is_method_id(), "Invalid JNI method identifier");
  return methodOop(instanceKlass::cast(_holder)->methods()->obj_at(_offset));
}


void JNIid::oops_do(void f(oop*)) {
  JNIid* current = this;
  while (current != NULL) {
    f((oop*)&(current->_holder));
    current = current->next();
  }
}


void JNIid::deallocate(JNIid* current) {
  while (current != NULL) {
    JNIid* next = current->next();
    delete current;
    current = next;
  }
}

#ifndef PRODUCT

void JNIid::verify(klassOop holder) {
  int first_field_offset  = instanceKlass::cast(holder)->offset_of_static_fields();
  int end_field_offset    = first_field_offset + instanceKlass::cast(holder)->static_field_size();
  int first_method_offset = 0;
  int end_method_offset   = instanceKlass::cast(holder)->methods()->length();
  
  JNIid* current = this;
  while (current != NULL) {
    guarantee(current->holder() == holder, "Invalid klass in JNIid");
#ifdef ASSERT
    int o = current->offset();
    if (current->is_static_field_id()) {
      guarantee(o >= first_field_offset  && o < end_field_offset,  "Invalid static field offset in JNIid");
    }
    if (current->is_method_id()) {
      guarantee(o >= first_method_offset && o < end_method_offset, "Invalid method offset in JNIid");
    }
#endif
    current = current->next();
  }
}

#endif


#ifdef ASSERT
  void instanceKlass::set_init_state(ClassState state) { 
    assert(_init_state < state || state == allocated, "illegal state transition");
    _init_state = state; 
  }
#endif
