/*
 * @(#)klass.hpp	1.78 01/02/09 14:45:33
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// A Klass is the part of the klassOop that provides:
//  1: language level class object (method dictionary etc.)
//  2: provide vm dispatch behavior for the object
// Both functions are combined into one C++ class. The toplevel class "Klass"
// implements purpose 1 whereas all subclasses provide extra virtual functions 
// for purpose 2.

// One reason for the oop/klass dichotomy in the implementation is
// that we don't want a C++ vtbl pointer in every object.  Thus,
// normal oops don't have any virtual functions.  Instead, they
// forward all "virtual" functions to their klass, which does have
// a vtbl and does the C++ dispatch depending on the object's
// actual type.  (See oop.inline.hpp for some of the forwarding code.)
// ALL FUNCTIONS IMPLEMENTING THIS DISPATCH ARE PREFIXED WITH "oop_"!

//  Klass layout:
//    [header        ] klassOop
//    [klass pointer ] klassOop
//    [C++ vtbl ptr  ] (contained in Klass_vtbl)
//    [subtype cache1]
//    [subtype cache2]
//    [size helper   ] (see encoding in oop.inline.hpp)
//    [java mirror   ]
//    [super         ]
//    [name          ]
//    [access flags  ]
//    [first subklass]
//    [next sibling  ]


class klassVtable;
class KlassHandle;

// Holder (or cage) for the C++ vtable of each kind of Klass.
// We want to tightly constrain the location of the C++ vtable in the overall layout.
class Klass_vtbl {
 protected:
  // The following virtual exists only to force creation of a C++ vtable,
  // so that this class truly is the location of the vtable of all Klasses.
  virtual void unused_initial_virtual() { }

 public:
  // The following virtual makes Klass_vtbl play a second role as a
  // factory protocol for subclasses of Klass ("sub-Klasses").
  // Here's how it works....
  //
  // This VM uses metaobjects as factories for their instances.
  //
  // In order to initialize the C++ vtable of a new instance, its
  // metaobject is forced to use the C++ placed new operator to
  // allocate the instance.  In a typical C++-based system, each
  // sub-class would have its own factory routine which
  // directly uses the placed new operator on the desired class,
  // and then calls the appropriate chain of C++ constructors.
  //
  // However, this system uses shared code to performs the first
  // allocation and initialization steps for all sub-Klasses.
  // (See base_create_klass() and base_create_array_klass().)
  // This does not factor neatly into a hierarchy of C++ constructors.
  // Each caller of these shared "base_create" routines knows
  // exactly which sub-Klass it is creating, but the shared routine
  // does not, even though it must perform the actual allocation.
  //
  // Therefore, the caller of the shared "base_create" must wrap
  // the specific placed new call in a virtual function which
  // performs the actual allocation and vtable set-up.  That
  // virtual function is here, Klass_vtbl::allocate_permanent.
  //
  // The arguments to Universe::allocate_permanent() are passed
  // straight through the placed new operator, which in turn
  // obtains them directly from this virtual call.
  //
  // This virtual is called on a temporary "example instance" of the
  // sub-Klass being instantiated, a C++ auto variable.  The "real"
  // instance created by this virtual is on the VM heap, where it is
  // equipped with a klassOopDesc header.
  //
  // It is merely an accident of implementation that we use "example
  // instances", but that is why the virtual function which implements
  // each sub-Klass factory happens to be defined by the same sub-Klass
  // for which it creates instances.
  //
  // The vtbl_value() call (see below) is used to strip away the
  // accidental Klass-ness from an "example instance" and present it as
  // a factory.  Think of each factory object as a mere container of the
  // C++ vtable for the desired sub-Klass.  Since C++ does not allow
  // direct references to vtables, the factory must also be delegated
  // the task of allocating the instance, but the essential point is
  // that the factory knows how to initialize the C++ vtable with the
  // right pointer value.  All other common initializations are handled
  // by the shared "base_create" subroutines.
  //
  virtual void* allocate_permanent(KlassHandle& klass, int size, TRAPS) const = 0;

  // Every subclass on which vtbl_value is called must include this macro.
#define DEFINE_ALLOCATE_PERMANENT(thisKlass) \
  void* allocate_permanent(KlassHandle& klass, int size, TRAPS) const { \
    return new(klass, size, THREAD) thisKlass(); \
  }

 protected:
  void* operator new(size_t ignored, KlassHandle& klass, int size, TRAPS);
};


class Klass : public Klass_vtbl {
  friend class VMStructs;
 protected:
  // note: put frequently-used fields together at start of klass structure
  // for better cache behavior (may not make much of a difference but sure won't hurt)
  union {
    double _force_alignment;    // force double-word alignment
    klassOop _subtype_cache[2]; // used for fast is_a test
  };  
  int       _size_helper;       // for instances, positive equal to size.
				// for arrays, -1-log2 of element size.
  oop       _java_mirror;       // instance of java/lang/Class mirroring this class
  klassOop  _super;             // superclass
  symbolOop _name;              // class name
                                //   instance classes: java/lang/String, etc.
                                //   array classes: [I, [Ljava/lang/String;, etc.
                                //   Set to zero for all other kinds of classes.
  AccessFlags _access_flags;    // Access flags. The class/interface distinction is stored here.
  klassOop _subklass;           // first subclass (NULL if none); _subklass->next_sibling() is next one
  klassOop _next_sibling;       // sibling link (or NULL); links all subklasses of a klass

#ifndef PRODUCT
  int           _verify_count;  // to avoid redundant verifies
#endif

  juint    _alloc_count;        // allocation profiling support - update klass_size_in_bytes() if moved/deleted


 public:

  // returns the enclosing klassOop
  klassOop as_klassOop() const {
    // see klassOop.hpp for layout.
    return (klassOop) (((char*) this) - sizeof(klassOopDesc));
  }

 public:
  // Allocation
  const Klass_vtbl& vtbl_value() const { return *this; }  // used only on "example instances"
  static klassOop base_create_klass(KlassHandle& klass, int size, const Klass_vtbl& vtbl, TRAPS);

  // super
  klassOop super() const               { return _super; }
  void set_super(klassOop k)           { oop_store_without_check((oop*) &_super, (oop) k); }

  // is_a_cache
  klassOop subtype_cache_1() const     { return _subtype_cache[0]; }
  void set_subtype_cache_1(klassOop k) { oop_store_without_check((oop*) &_subtype_cache[0], (oop) k); }

  klassOop subtype_cache_2() const     { return _subtype_cache[1]; }
  void set_subtype_cache_2(klassOop k) { oop_store_without_check((oop*) &_subtype_cache[1], (oop) k); }

  // java mirror
  oop java_mirror() const              { return _java_mirror; }
  void set_java_mirror(oop m)          { oop_store((oop*) &_java_mirror, m); }

  // size helper
  int size_helper() const              { return _size_helper; }
  void set_size_helper(int size)       { _size_helper = size; }

  // Note: for instances size_helper() may include padding.
  // Use instanceKlass::contains_field_offset to classify field offsets.

  // sub/superklass links
  instanceKlass* superklass() const;
  Klass* subklass() const;
  Klass* next_sibling() const;
  void append_to_sibling_list();           // add newly created receiver to superklass' subklass list
  void remove_from_sibling_list();         // remove receiver from sibling list
 protected:                                // internal accessors
  klassOop subklass_oop() const            { return _subklass; }
  klassOop next_sibling_oop() const        { return _next_sibling; }
  void     set_subklass(klassOop s); 
  void     set_next_sibling(klassOop s);

  oop* adr_super()	     const { return (oop*)&_super;             }
  oop* adr_subtype_cache_1() const { return (oop*)&_subtype_cache[0];  }
  oop* adr_subtype_cache_2() const { return (oop*)&_subtype_cache[1];  }
  oop* adr_java_mirror()     const { return (oop*)&_java_mirror;       }
  oop* adr_name()            const { return (oop*)&_name;              }
  oop* adr_subklass()        const { return (oop*)&_subklass;          }
  oop* adr_next_sibling()    const { return (oop*)&_next_sibling;      }

 public:
  // Allocation profiling support
  juint alloc_count() const          { return _alloc_count; }
  void set_alloc_count(juint n)      { _alloc_count = n; }
  virtual juint alloc_size() const = 0;
  virtual void set_alloc_size(juint n) = 0;

  // Compiler support
  static int super_offset_in_bytes()         { return (int)&((Klass*)NULL)->_super;            }
  static int is_a_cache_1_offset_in_bytes()  { return (int)&((Klass*)NULL)->_subtype_cache[0]; }
  static int is_a_cache_2_offset_in_bytes()  { return (int)&((Klass*)NULL)->_subtype_cache[1]; }
  static int java_mirror_offset_in_bytes()   { return (int)&((Klass*)NULL)->_java_mirror;      }
  static int size_helper_offset_in_bytes()   { return (int)&((Klass*)NULL)->_size_helper;    }

  // vtables
  virtual klassVtable* vtable() const        { return NULL; }
  static int klass_size_in_bytes()           { return ((int)&((Klass*)NULL)->_alloc_count) + sizeof(juint); }  // all "visible" fields

  // subclass check
  bool is_subclass_of(klassOop k) const;
  // subtype check: true if is_subclass_of, or if k is interface and receiver implements it
  bool is_subtype_of(klassOop k) {
    return subtype_cache_1() == k // Fast check
        || subtype_cache_2() == k // Fast check
        || compute_and_set_is_subtype_of(k); // Computes the result and fills the cache
  }
  // Find LCA in class heirarchy
  Klass *LCA( Klass *k );

  // dispatch optimization: can m be called w/o dynamic dispatch?
  static bool can_be_statically_bound(methodOop m);

  // Check whether reflection/jni/jvm code is allowed to instantiate this class;
  // if not, throw either an Error or an Exception.
  virtual void check_valid_for_instantiation(bool throwError, TRAPS);

  // Casting
  static Klass* cast(klassOop k) {
    assert(k->is_klass(), "cast to Klass");
    return k->klass_part(); 
  }

  // array copying
  virtual void  copy_array(arrayOop s, int src_pos, arrayOop d, int dst_pos, int length, TRAPS);

  // tells if the class should be initialized
  virtual bool should_be_initialized() const    { return false; }
  // initializes the klass
  virtual void initialize(TRAPS);
  // lookup operation for MethodLookupCache
  friend class MethodLookupCache;
  virtual methodOop uncached_lookup_method(symbolOop name, symbolOop signature) const;
 public:  
  methodOop lookup_method(symbolOop name, symbolOop signature) const; // uses MethodLookupCache

  // array class with specific rank
  klassOop array_klass(int rank, TRAPS)         {  return array_klass_impl(false, rank, THREAD); }

  // array class with this klass as element type
  klassOop array_klass(TRAPS)                   {  return array_klass_impl(false, THREAD); }

  // these will return NULL instead of allocating on the heap:
  klassOop array_klass_or_null(int rank);
  klassOop array_klass_or_null();

  virtual oop protection_domain() { return NULL; }

 protected:
  virtual klassOop array_klass_impl(bool or_null, int rank, TRAPS);
  virtual klassOop array_klass_impl(bool or_null, TRAPS);

 public:
  // computes and sets cache for subtype check
  bool compute_and_set_is_subtype_of(klassOop k);
 protected:
  // computes the subtype relationship
  virtual bool compute_is_subtype_of(klassOop k);
 public:
  // subclass accessor (here for convenience; undefined for non-klass objects)
  virtual bool is_leaf_class() const { fatal("not a class"); return false; }
 public:
  // ALL FUNCTIONS BELOW THIS POINT ARE DISPATCHED FROM AN OOP
  // These functions describe behavior for the oop not the KLASS.

  // actual oop size of obj in memory
  virtual int oop_size(oop obj) const = 0;

  // actual oop size of this klass in memory
  virtual int klass_oop_size() const = 0;

  // Returns the Java name for a class (Resource allocated)
  // For arrays, this returns the name of the element with a leading '['.
  // For classes, this returns the name with the package separators 
  //     turned into '.'s.
  virtual const char* external_name() const;
  // Returns the name for a class (Resource allocated) as the class 
  // would appear in a signature.
  // For arrays, this returns the name of the element with a leading '['.
  // For classes, this returns the name with a leading 'L' and a trailing ';'
  //     and the package separators as '/'.
  virtual char* signature_name() const;

  // garbage collection support
  virtual int  oop_scavenge_contents(oop obj) = 0;
  virtual int  oop_scavenge_contents(oop obj, oop* bottom, oop* top) = 0;
  virtual void oop_follow_contents(oop obj) = 0;
  virtual int  oop_adjust_pointers(oop obj) = 0;
  virtual bool oop_being_unloaded(oop obj);

 public:
  // type testing operations
  virtual bool oop_is_instance()            const { return false; }
  virtual bool oop_is_instanceRef()         const { return false; }
  virtual bool oop_is_array()	            const { return false; }
  virtual bool oop_is_objArray()            const { return false; }
  virtual bool oop_is_symbol()	            const { return false; }
  virtual bool oop_is_klass()	            const { return false; }
  virtual bool oop_is_thread()	            const { return false; }
  virtual bool oop_is_method()	            const { return false; }
  virtual bool oop_is_constantPool()        const { return false; }
  virtual bool oop_is_constantPoolCache()   const { return false; }
  virtual bool oop_is_typeArray()           const { return false; }
  virtual bool oop_is_arrayKlass()          const { return false; }
  virtual bool oop_is_objArrayKlass()       const { return false; }
  virtual bool oop_is_typeArrayKlass()      const { return false; }
  virtual bool oop_is_compiledICHolder()    const { return false; }
  virtual bool oop_is_instanceKlass()       const { return false; }

  // Access flags
  AccessFlags access_flags() const         { return _access_flags;  }
  void set_access_flags(AccessFlags flags) { _access_flags = flags; }

  bool is_public() const                { return _access_flags.is_public(); }
  bool is_final() const                 { return _access_flags.is_final(); }
  bool is_interface() const             { return _access_flags.is_interface(); }
  bool is_abstract() const              { return _access_flags.is_abstract(); }
  bool is_super() const                 { return _access_flags.is_super(); }
  bool is_synthetic() const             { return _access_flags.is_synthetic(); }
  void set_is_synthetic()               { _access_flags.set_is_synthetic(); }
  bool has_finalizer() const            { return _access_flags.has_finalizer(); }
  void set_has_finalizer()              { _access_flags.set_has_finalizer(); }
  bool is_cloneable() const             { return _access_flags.is_cloneable(); }
  void set_is_cloneable()               { _access_flags.set_is_cloneable(); }
  bool has_vanilla_constructor() const  { return _access_flags.has_vanilla_constructor(); }
  void set_has_vanilla_constructor()    { _access_flags.set_has_vanilla_constructor(); }
  bool has_miranda_methods () const     { return access_flags().has_miranda_methods(); }
  void set_has_miranda_methods()        { _access_flags.set_has_miranda_methods(); }



  // garbage collection support
  virtual void follow_weak_klass_links();

  // iterators
  virtual void oop_oop_iterate(oop obj, OopClosure* blk) = 0;
  virtual void array_klasses_do(void f(klassOop k)) {}
  virtual void with_array_klasses_do(void f(klassOop k));

  // Return self, except for abstract classes with exactly 1
  // implementor.  Then return the 1 concrete implementation.
  Klass *up_cast_abstract();

  // klass name
  symbolOop name() const                   { return _name; }
  void set_name(symbolOop n)               { oop_store_without_check((oop*) &_name, (oop) n); }

  friend class klassKlass;

 public:
  // jvm support
  virtual jint modifier_flags(TRAPS) const;

 public:
  // jvmdi support
  virtual jint class_status() const;

#ifndef PRODUCT
 public:
  // Printing
  virtual void oop_print_on      (oop obj, outputStream* st);
  virtual void oop_print_value_on(oop obj, outputStream* st);
  virtual const char* internal_name() const = 0;

  // Verification
  virtual void oop_verify_on(oop obj, outputStream* st);
  virtual void oop_verify_old_oop(oop obj, oop* p, bool allow_dirty);
  // tells whether obj is partially constructed (gc during class loading)
  virtual bool oop_partially_loaded(oop obj) const { return false; }
  virtual void oop_set_partially_loaded(oop obj) {};
#endif
};
