/*
 * @(#)markOop.hpp	1.36 00/04/17 17:26:02
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// The markOop describes the header of an object.
//
// Note that the mark is not a real oop but just a word. 
// It is placed in the oop hierarchy for historical reasons.
//
// Bit-format of an object header (most significant first):
//
//  
//  speclock:1 age:6 hash:23 lock:2 = 32 bits 
//
//  - hash contains the identity hash value.
//
//  - the two lock bits are used to describe three states: locked/unlocked and monitor.
//
//    [ptr    | 00]  locked                  ptr points to real header on stack
//    [header | 01]  unlocked                regular object header
//    [ptr    | 10]  monitor                 inflated lock (header is wapped out)
//    [ptr    | 11]  marked                  used by markSweep to mark an object
//                                           not valid at any other time
//
//    We assume that stack/thread pointers have the lowest two bits cleared.

class BasicLock;
class ObjectMonitor;

class markOopDesc: public oopDesc {
 private:
  // Conversion
  intptr_t value() const { return (intptr_t) this; }

 public:
  // Constants
  enum { speclock_bits            = 1,
         age_bits                 = 6,
         lock_bits                = 2,
         hash_bits                = BitsPerWord - age_bits - lock_bits - speclock_bits
  };

  enum { lock_shift               = 0,
         hash_shift               = lock_bits,
         age_shift                = lock_bits + hash_bits,
         speclock_shift           = lock_bits + hash_bits + age_bits
  };

  enum { lock_mask                = right_n_bits(lock_bits),
         lock_mask_in_place       = lock_mask << lock_shift,
         hash_mask                = right_n_bits(hash_bits),
         hash_mask_in_place       = hash_mask << hash_shift,
         age_mask                 = right_n_bits(age_bits),
         age_mask_in_place        = age_mask << age_shift,
         speclock_mask            = right_n_bits(speclock_bits),
         speclock_mask_in_place   = speclock_mask << speclock_shift
  };

  enum { locked_value             = 0,
         unlocked_value           = 1,
         monitor_value            = 2,
         marked_value             = 3
  };

  enum { no_speclock              = 0,
         speclock                 = 1
  } ;

  enum { no_speclock_in_place     = no_speclock << speclock_shift,
         speclock_in_place        = speclock << speclock_shift
  };

  enum { speclock_pattern         = speclock_mask_in_place | unlocked_value };

  enum { no_hash                  = 0 };  // no hash value assigned

  enum { no_hash_in_place         = no_hash << hash_shift,
         no_lock_in_place         = unlocked_value
  };

  enum { max_age                  = age_mask };

  // lock accessors (note that these assume lock_shift == 0)
  bool is_locked()   const {
    return (mask_bits(value(), lock_mask_in_place) != unlocked_value);
  }
  bool is_unlocked() const {
    return (mask_bits(value(), lock_mask_in_place) == unlocked_value);
  }
  bool is_marked()   const {
    return (mask_bits(value(), lock_mask_in_place) == marked_value);
  }
  bool is_speclock() const {
    // Put in full mask
    return (is_unlocked() && (mask_bits(value(), speclock_in_place) == speclock_in_place));

  }
  
  // Should this header be preserved during GC?
  bool must_be_preserved() const {
     return (is_speclock() || !is_unlocked() || !has_no_hash());
  }

  // WARNING: The following routines are used EXCLUSIVELY by 
  // synchronization functions. They are not really gc safe.
  // They must get updated if markOop layout get changed.
  markOop set_unlocked() const {
    return markOop(value() | unlocked_value);
  }
  bool has_locker() const {
    return ((value() & lock_mask_in_place) == 0);
  }
  BasicLock* locker() const {
    assert(has_locker(), "check");
    return (BasicLock*) value();
  }
  bool has_monitor() const {
    return ((value() & monitor_value) != 0);
  }
  ObjectMonitor* monitor() const {
    assert(has_monitor(), "check");
    // Use xor instead of &~ to provide one extra tag-bit check.
    return (ObjectMonitor*) (value() ^ monitor_value);
  }
  bool has_displaced_mark_helper() const {
    return ((value() & unlocked_value) == 0);
  }
  markOop displaced_mark_helper() const {
    assert(has_displaced_mark_helper(), "check");
    intptr_t ptr = (value() & ~monitor_value);
    return *(markOop*)ptr;
  }
  void set_displaced_mark_helper(markOop m) const {
    assert(has_displaced_mark_helper(), "check");
    intptr_t ptr = (value() & ~monitor_value);
    *(markOop*)ptr = m;
  }
  markOop copy_set_hash(int hash) const {
    intptr_t tmp = value() & (~hash_mask_in_place);
    tmp |= (hash << hash_shift);
    return (markOop)tmp;
  }
  // it is only used to be stored into BasicLock as the 
  // indicator that the lock is using heavyweight monitor
  static markOop unused_mark() {
    return (markOop) marked_value;
  }
  // the following two functions create the markOop to be
  // stored into object header, it encodes monitor info
  static markOop encode(BasicLock* lock) {
    return (markOop) lock;
  }
  static markOop encode(ObjectMonitor* monitor) {
    intptr_t tmp = (intptr_t) monitor;
    return (markOop) (tmp | monitor_value);
  }
  
  // used for alignment-based marking to reuse the busy state to encode pointers
  // (see markOop_alignment.hpp)  
  markOop clear_lock_bits() { return markOop(value() & ~lock_mask_in_place); }  

  markOop clear_speclock_bits() { return markOop(value() & ~speclock_mask_in_place); }

  // age operations
  markOop set_marked()   { return markOop((value() & ~lock_mask_in_place) | marked_value); }

  int     age()               const { return mask_bits(value() >> age_shift, age_mask); }
  markOop set_age(int v) const {
    assert((v & ~age_mask) == 0, "shouldn't overflow age field");
    return markOop((value() & ~age_mask_in_place) | ((v & age_mask) << age_shift));
  }
  markOop incr_age()          const { return age() == max_age ? markOop(this) : set_age(age() + 1); }

  // hash operations
  int hash() const {     
    return mask_bits(value(), hash_mask_in_place) >> hash_shift; 
  }
  
  bool has_no_hash() const { 
    return hash() == no_hash; 
  }

  // Prototype mark for initialization
 // Prototype mark for initialization
  static markOop prototype() {
    return markOop( no_speclock_in_place | no_hash_in_place | no_lock_in_place )->set_no_pointer();
  }

  static markOop prototype_speclock() {
    return markOop( speclock_in_place | no_hash_in_place | no_lock_in_place )->set_no_pointer();
  }
 
  // Debugging
  void print_on(outputStream* st) const;

  //    (During a mark-sweep or scavenge) an object's mark is replaced be a pointer to the reference
  //    that object. That word may, in turn either contain another pointer to another reference
  //    or may contain the original mark (at the end of the list).
  //    So, it must be possible to distinguish a mark from a (possibly encoded) address
  //    of an oop.
  //
  //    There are two schemes to do this:
  //
  //    For the Intel platform, oops may be contained in nonaligned addresses
  //    (e.g. oops embedded in machine code),
  //    but oops may never reside in a region of memory with the high-order-bit
  //    set (guaranteed by NT).
  //    So, the sentinel bit is located in the high-order-bit of the mark word and is
  //    set to one in marks. Testing this bit tells the system whether it has a mark or
  //    a pointer to a location containing a reference.
  //
  //    For the SPARC platform, oops must always be in word-aligned locations,
  //    and we would prefer not to require that the stacks and object heap reside
  //    in any particular place in the address space.
  //    So for these platforms, we exploit the fact that any locked oops will be
  //    handled by a prepass (on either platform). The prepass stores the lock info
  //    elsewhere and restores the oops to the unlocked state.
  //    We therefore reuse the busy locked state (11) to encode a mark that is really
  //    a pointer to an oop reference. The sentinel is not used for this platform.
  //


  //  Here is the interface, common to both:
 public:
  // Tells whether p is a root to an oop or a markOop.
  // Used during mark-sweep and scavenging.
  inline bool has_pointer();

  // Used for asserts, to rule out marks where we expect oops
  inline bool assert_not_real_mark();
  //  and vice-versa
  inline bool assert_real_mark();

  // For initialization, setup mark to say "I am not a pointer"
  inline markOop set_no_pointer();

  // prepare address of oop for placement into mark
  inline static markOop encode_pointer_as_mark(void* p);

  // recover address of oop from encoded form used in mark
  inline void* decode_pointer();
};

# include "incls/_markOop_pd.hpp.incl"
