/*
 * @(#)objArrayKlass.cpp	1.90 01/04/23
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_objArrayKlass.cpp.incl"

int objArrayKlass::oop_size(oop obj) const {
  assert(obj->is_objArray(), "must be object array");
  return objArrayOopDesc::object_size(objArrayOop(obj)->length());
}

objArrayOop objArrayKlass::allocate(int length, TRAPS) {  
  if (length >= 0) {
    if (length <= arrayOopDesc::max_array_length(T_OBJECT)) {
      int size = objArrayOopDesc::object_size(length);
      KlassHandle h_k(THREAD, as_klassOop());
      objArrayOop a = (objArrayOop)Universe::clear_allocate_array(h_k, size, length, CHECK_0);
      return a;
    } else {
      THROW_OOP_0(Universe::out_of_memory_error_instance());
    }
  } else {
    THROW_0(vmSymbols::java_lang_NegativeArraySizeException());
  }
}

static int multi_alloc_counter = 0;

oop objArrayKlass::multi_allocate(int rank, jint* sizes, int next_dim_step, TRAPS) { 
  assert(next_dim_step == 1  ||  next_dim_step == -1,  "next_dim_step better be one of these");
  int length = *sizes;
  // Call to lower_dimension uses this pointer, so most be called before a possible GC
  KlassHandle h_lower_dimension(THREAD, lower_dimension());
  objArrayHandle h_array = allocate(length, CHECK_0);
  if (rank > 1) {
    for (int index = 0; index < length; index++) {  
       oop sub_array = arrayKlass::cast(h_lower_dimension())->multi_allocate(rank-1, &sizes[next_dim_step], next_dim_step, CHECK_0);
       h_array->obj_at_put(index, sub_array);
    }
  }
  return h_array();
}

void objArrayKlass::copy_array(arrayOop s, int src_pos, arrayOop d, int dst_pos, int length, TRAPS) {
  assert(s->is_objArray(), "must be obj array");

  if (!d->is_objArray()) {
    THROW(vmSymbols::java_lang_ArrayStoreException());
  }

  // Check is all offsets and lengths are non negative
  if (src_pos < 0 || dst_pos < 0 || length < 0) {
    THROW(vmSymbols::java_lang_ArrayIndexOutOfBoundsException());
  }
  // Check if the ranges are valid
  if  ( (((unsigned int) length + (unsigned int) src_pos) > (unsigned int) s->length())
     || (((unsigned int) length + (unsigned int) dst_pos) > (unsigned int) d->length()) ) {
    THROW(vmSymbols::java_lang_ArrayIndexOutOfBoundsException());
  }

  // Special case. Boundary cases must be checked first
  // This allows the following call: copy_array(s, s.length(), d.length(), 0).
  // This is correct, since the position is supposed to be an 'in between point', i.e., s.length(),
  // points to the right of the last element.
  if (length==0) {
    return;
  }

  oop* src = objArrayOop(s)->obj_at_addr(src_pos);
  oop* dst = objArrayOop(d)->obj_at_addr(dst_pos);

  if (s == d) {
    // since source and destination are equal we do not need conversion checks. 
    assert(length > 0, "sanity check")
    Memory::copy_oops_overlapping_atomic(src, dst, length);
    RememberedSet::record_array_store(dst, dst + length);
  } else {
    // We have to make sure all elements conform to the destination array
    klassOop bound = objArrayKlass::cast(d->klass())->element_klass();
    klassOop stype = objArrayKlass::cast(s->klass())->element_klass();
    if (stype == bound || Klass::cast(stype)->is_subtype_of(bound)) {
      // elements are guaranteed to be subtypes, so no check necessary
      Memory::copy_oops_overlapping_atomic(src, dst, length);
      RememberedSet::record_array_store(dst, dst + length);
    } else {
      // slow case: need individual subtype checks
      // do store checks first so they're guaranteed to be done even if an exception is thrown
      RememberedSet::record_array_store(dst, dst + length);
      // note: don't use obj_at_put below because it includes a redundant store check
      oop* from = src;
      oop* end = from + length;
      for (oop* to = dst; from < end; from++, to++) {
        oop element = *from;
        if (element == NULL || Klass::cast(element->klass())->is_subtype_of(bound)) {
          *to = element;
        } else {
          THROW(vmSymbols::java_lang_ArrayStoreException());
          return;
        }
      }
    }
  }
}


klassOop objArrayKlass::array_klass_impl(bool or_null, int n, TRAPS) {
  objArrayKlassHandle h_this(THREAD, as_klassOop());
  return array_klass_impl(h_this, or_null, n, CHECK_0);
}


klassOop objArrayKlass::array_klass_impl(objArrayKlassHandle this_oop, bool or_null, int n, TRAPS) {  
  
  assert(this_oop->dimension() <= n, "check order of chain");
  int dimension = this_oop->dimension();
  if (dimension == n) 
    return this_oop();

  objArrayKlassHandle ak = this_oop->higher_dimension();
  if (ak.is_null()) {    
    if (or_null)  return NULL;

    MutexLocker mc(Compile_lock);   // for vtables
    // Ensure atomic cration of higher dimensions
    MutexLocker mu(MultiArray_lock);

    // Check if another thread beat us
    ak =  this_oop->higher_dimension();
    if (this_oop->higher_dimension() == NULL) {
    
      // Create multi-dim klass object and link them together
      ak = objArrayKlassKlass::cast(Universe::objArrayKlassKlassObj())->allocate_objArray_klass(dimension + 1, this_oop, CHECK_0);      
      this_oop->set_higher_dimension(ak());    
      ak->set_lower_dimension(this_oop());
      assert(ak->oop_is_objArray(), "incorrect initialization of objArrayKlass");
    }
  }

  if (or_null) {
    return ak->array_klass_or_null(n);
  }
  return ak->array_klass(n, CHECK_0);
}

klassOop objArrayKlass::array_klass_impl(bool or_null, TRAPS) {
  return array_klass_impl(or_null, dimension() +  1, CHECK_0);
}

bool objArrayKlass::compute_is_subtype_of(klassOop k) {
  if (!k->klass_part()->oop_is_objArray())
    return arrayKlass::compute_is_subtype_of(k);

  objArrayKlass* oak = objArrayKlass::cast(k);
  return element_klass()->klass_part()->is_subtype_of(oak->element_klass());
}


void objArrayKlass::initialize(TRAPS) {
  Klass::cast(bottom_klass())->initialize(THREAD);  // dispatches to either instanceKlass or typeArrayKlass
}


void objArrayKlass::oop_follow_contents(oop obj) {
  assert (obj->is_array(), "obj must be array");
  arrayOop a = arrayOop(obj);
  unsigned int* base = (unsigned int*) (a->base_offset_in_bytes(T_OBJECT) + (int)a);
  for (int i = 0; i< a->length(); i ++) {
    if (*base != 0) 
      // we call mark_and_follow here to avoid excessive marking stack usage
      MarkSweep::mark_and_follow((oop*)base); 
    base++;
  }  
  arrayKlass::oop_follow_contents(obj);
}


void objArrayKlass::oop_oop_iterate(oop obj, OopClosure* blk) {
  assert (obj->is_array(), "obj must be array");
  arrayKlass::oop_oop_iterate(obj, blk);

  arrayOop a = arrayOop(obj);
  unsigned int* base = (unsigned int*) (a->base_offset_in_bytes(T_OBJECT) + (int)a);
  for (int i = 0; i< a->length(); i ++) {
    blk->do_oop((oop*)base); 
    base++;
  }  
}


int objArrayKlass::oop_scavenge_contents(oop obj) {
  assert (obj->is_objArray(), "obj must be obj array");
  // Compute oop range
  oop* base = objArrayOop(obj)->base();
  oop* end = base + objArrayOop(obj)->length();
  // Iterate over oops
  if (PrefetchScavengeFieldsAhead != 0) {
    while (base < end) {
      oop* p = base + PrefetchScavengeFieldsAhead;
      if (p < end) {
	atomic::prefetch_read(*p);   // Prefetching null is okay
      }
      if (Scavenge::should_scavenge(*base)) {
	Scavenge::scavenge(base); 
      }
      base++;
    }
  } else {
    while (base < end) {
      if (Scavenge::should_scavenge(*base)) {
	Scavenge::scavenge(base); 
      }
      base++;
    }
  }
  assert(align_object_size(end - (oop*)obj) == oop_size(obj), "checking size");
  return align_object_size(end - (oop*)obj);  // saves a virtual call to oop_size
}


int objArrayKlass::oop_scavenge_contents(oop obj, oop* bottom, oop* top) {
  assert (obj->is_objArray(), "obj must be obj array");
  // Compute oop range
  oop* base = objArrayOop(obj)->base();
  oop* end = base + objArrayOop(obj)->length();
    // Adjust for boundaries
  if (base < bottom) base = bottom;
  if (end > top)     end = top;
  // Iterate over oops
  while (base < end) {
    if (Scavenge::should_scavenge(*base)) {
      Scavenge::scavenge(base); 
    }
    base++;
  }
  return oop_size(obj);
}


int objArrayKlass::oop_adjust_pointers(oop obj) {
  assert (obj->is_objArray(), "obj must be obj array");
  int size = arrayKlass::oop_adjust_pointers(obj);

  arrayOop a = arrayOop(obj);
  unsigned int* base = (unsigned int*) (a->base_offset_in_bytes(T_OBJECT) + (int)a);
  for (int i = 0; i< a->length(); i ++) {
    MarkSweep::adjust_pointer((oop*)base); 
    base++;
  }  

  return size;
}


const char* objArrayKlass::external_name() const {
  char* result;
  if (dimension() == 1) {
    const char* name = instanceKlass::cast(element_klass())->external_name();
    int   name_length = strlen(name);
    result = NEW_RESOURCE_ARRAY(char, name_length + 4);
    result[0] = '[';
    result[1] = 'L';
    for (int j = 0; j < name_length; j++) {
      result[j + 2] = name[j];
    }
    result[name_length + 2] = ';';
    result[name_length + 3] = '\0';
  } else {
    const char* name = Klass::cast(element_klass())->external_name();
    int   name_length = strlen(name);
    result = NEW_RESOURCE_ARRAY(char, name_length + 2);
    result[0] = '[';
    for (int j = 0; j < name_length; j++) {
      result[j + 1] = name[j];
    }
    result[name_length + 1] = '\0';
  }
  return result;
}

// JVM support

jint objArrayKlass::modifier_flags(TRAPS) const {
  // The modifier for an objectArray is the same as its element
  assert(element_klass() != NULL, "sanity check");
  // Recurse down the element list
  jint element_flags = Klass::cast(element_klass())->modifier_flags(CHECK_0);  
  jint array_flags = 0;
  if (element_flags & (JVM_ACC_PUBLIC | JVM_ACC_PROTECTED)) {
    // The array type is public is the component type is public or protected
    array_flags = JVM_ACC_ABSTRACT | JVM_ACC_FINAL | JVM_ACC_PUBLIC;  
  } else {
    // The array type is private if the component type is private
    array_flags = JVM_ACC_ABSTRACT | JVM_ACC_FINAL;
  }  
  return array_flags;
}


#ifndef PRODUCT
// Printing

void objArrayKlass::oop_print_on(oop obj, outputStream* st) {
  arrayKlass::oop_print_on(obj, st);
  assert(obj->is_objArray(), "must be objArray");
  objArrayOop oa = objArrayOop(obj);
  int print_len = min(oa->length(), MaxElementPrintSize);
  for(int index = 0; index < print_len; index++) {
    st->print(" - %3d : ", index);
    oa->obj_at(index)->print_value_on(st);
    st->cr();
  }
  int remaining = oa->length() - print_len;
  if (remaining > 0) {
    tty->print_cr(" - <%d more elements, increase MaxElementPrintSize to print>", remaining);
  }
}


void objArrayKlass::oop_print_value_on(oop obj, outputStream* st) {
  assert(obj->is_objArray(), "must be objArray");
  element_klass()->print_value_on(st);
  st->print("a [%d] ", objArrayOop(obj)->length());
  as_klassOop()->klass()->print_value_on(st);
}

const char* objArrayKlass::internal_name() const {
  return external_name();
}

// Verification

void objArrayKlass::oop_verify_on(oop obj, outputStream* st) {
  arrayKlass::oop_verify_on(obj, st);
  guarantee(obj->is_objArray(), "must be objArray");
  objArrayOop oa = objArrayOop(obj);
  for(int index = 0; index < oa->length(); index++) {
    guarantee(oa->obj_at(index)->safepoint_is_oop_or_null(), "should be oop");
  }
}

void objArrayKlass::oop_verify_old_oop(oop obj, oop* p, bool allow_dirty) {
  RememberedSet::verify_old_oop(obj, p, allow_dirty, true);
}
#endif // PRODUCT
