/*
 * @(#)addnode.cpp	1.88 01/02/09 14:45:47
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
// Portions of code courtesy of Clifford Click

#include "incls/_precompiled.incl"
#include "incls/_addnode.cpp.incl"

#define MAXFLOAT        ((float)3.40282346638528860e+38)

// Classic Add functionality.  This covers all the usual 'add' behaviors for
// an algebraic ring.  Add-integer, add-float, add-double, and binary-or are
// all inherited from this class.  The various identity values are supplied
// by virtual functions.


//=============================================================================
//------------------------------hash-------------------------------------------
// Hash function over AddNodes.  Needs to be commutative; i.e., I swap 
// (commute) inputs to AddNodes willy-nilly so the hash function must return
// the same value in the presence of edge swapping.
uint AddNode::hash() const {
  return (uint)_in[1] + (uint)_in[2] + Opcode();
}

//------------------------------Identity---------------------------------------
// If either input is a constant 0, return the other input.  
Node *AddNode::Identity( PhaseTransform *phase ) {
  const Type *zero = add_id();	// The additive identity
  if( phase->type( _in[1] )->higher_equal( zero ) ) return _in[2];
  if( phase->type( _in[2] )->higher_equal( zero ) ) return _in[1];
  return this;
}

//------------------------------commute----------------------------------------
// Commute operands to move loads and constants to the right.
static bool commute( Node *add, int con_left, int con_right ) {
  Node *in1 = add->in(1);
  Node *in2 = add->in(2);

  // Convert "1+x" into "x+1".
  // Right is a constant; leave it
  if( con_right ) return false;
  // Left is a constant; move it right.
  if( con_left ) {
    add->set_req(1,in2);             // No need for set_req_du since just
    add->set_req(2,in1);             // ...swapping inputs; DU info is OK.
    return true;
  }

  // Convert "Load+x" into "x+Load".
  // Now check for loads
  if( in2->is_Load() ) return false;
  // Left is a Load and Right is not; move it right.
  if( in1->is_Load() ) {
    add->set_req(1,in2);             // No need for set_req_du since just
    add->set_req(2,in1);             // ...swapping inputs; DU info is OK.
    return true;
  }

  // Check for tight loop increments: Loop-phi of Add of loop-phi
  if( in1->is_Phi() && in1->in(0) && in1->in(0)->is_Loop() && in1->in(2)==add)
    return false;
  if( in2->is_Phi() && in2->in(0) && in2->in(0)->is_Loop() && in2->in(2)==add){
    add->set_req(1,in2);             // No need for set_req_du since just
    add->set_req(2,in1);             // ...swapping inputs; DU info is OK.
    return true;
  }

  // Check for making long spline calculations
  uint opa = add->Opcode();
  uint op1 = in1->Opcode();
  uint op2 = in2->Opcode();
  if( 0 && opa == op1 && opa != op2 ) {
    add->set_req(1,in2);             // No need for set_req_du since just
    add->set_req(2,in1);             // ...swapping inputs; DU info is OK.
    return true;
  } else if( 0 && opa != op1 && opa == op2 )
    return false;


  // Otherwise, sort inputs (commutativity) to help value numbering.
  if( in1->_idx > in2->_idx ) {
    add->set_req(1,in2);             // No need for set_req_du since just
    add->set_req(2,in1);             // ...swapping inputs; DU info is OK.
    return true;
  }
  return false;
}

//------------------------------Idealize---------------------------------------
// If we get here, we assume we are associative!
Node *AddNode::Ideal(PhaseGVN *phase, PhaseDefUse *du){
  const Type *t1 = phase->type( _in[1] );
  const Type *t2 = phase->type( _in[2] );
  int con_left  = t1->singleton();
  int con_right = t2->singleton();

  // Check for commutative operation desired
  if( commute(this,con_left,con_right) ) return this;

  AddNode *progress = NULL;		// Progress flag

  // Convert "(x+1)+2" into "x+(1+2)".  If the right input is a
  // constant, and the left input is an add of a constant, flatten the
  // expression tree.
  Node *add1 = in(1);
  Node *add2 = in(2);
  int add1_op = add1->Opcode();
  int this_op = Opcode();
  if( con_right && t2 != Type::TOP && // Right input is a constant?
      add1_op == this_op ) { // Left input is an Add?

    // Type of left _in right input
    const Type *t12 = phase->type( add1->in(2) ); 
    if( t12->singleton() ) {	// Left input is an add of a constant?
      if( add1 == add1->in(1) ) { // Check for rare case of closed data cycle
	// Closed data cycle can happen for unreachable loops.
	// In this case the computation is undefined.
	return new (1) ConNode( Type::TOP );
      }
      // The Add of the flattened expression
      Node *x1 = add1->in(1);
      Node *x2 = phase->makecon( add1->is_Add()->add_ring( t2, t12 ));
      PhaseIterGVN *igvn = phase->is_IterGVN();
      if( igvn ) {
        set_req_X(2,x2,igvn);
        set_req_X(1,x1,igvn);
      } else {
        set_req(2,x2);
        set_req(1,x1);
      }
      progress = this;            // Made progress
      add1 = in(1);
      add1_op = add1->Opcode();
    }
  }

  // Convert "(x+1)+y" into "(x+y)+1".  Push constants down the expression tree.
  if( add1_op == this_op && !con_right ) {
    Node *a12 = add1->in(2);
    const Type *t12 = phase->type( a12 );
    if( t12->singleton() && t12 != Type::TOP && (add1 != add1->in(1)) ) {
      add2 = add1->clone();
      add2->set_req(2,in(2));
      add2 = phase->transform(add2);
      set_req(1,add2,du);
      set_req(2,a12,du);
      progress = this;
      add2 = a12;
    }
  }

  // Convert "x+(y+1)" into "(x+y)+1".  Push constants down the expression tree.
  int add2_op = add2->Opcode();
  if( add2_op == this_op && !con_left ) {
    Node *a22 = add2->in(2);
    const Type *t22 = phase->type( a22 );
    if( t22->singleton() && t22 != Type::TOP && (add2 != add2->in(1)) ) {
      Node *addx = add2->clone();
      addx->set_req(1,in(1));
      addx->set_req(2,add2->in(1));
      addx = phase->transform(addx);
      set_req(1,addx,du);
      set_req(2,a22,du);
      progress = this;
    }
  }

  return progress;
}

//------------------------------Value-----------------------------------------
// An add node sums it's two _in.  If one input is an RSD, we must mixin
// the other input's symbols.
const Type *AddNode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  const Type *t2 = phase->type( _in[2] );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t2 == Type::TOP ) return Type::TOP;

  // Either input is BOTTOM ==> the result is the local BOTTOM
  const Type *bot = bottom_type();
  if( (t1 == bot) || (t2 == bot) ||
      (t1 == Type::BOTTOM) || (t2 == Type::BOTTOM) ) 
    return bot;

  // Check for an addition involving the additive identity
  const Type *tadd = add_of_identity( t1, t2 );
  if( tadd ) return tadd;

  return add_ring(t1,t2);		// Local flavor of type addition
}

//------------------------------add_identity-----------------------------------
// Check for addition of the identity
const Type *AddNode::add_of_identity( const Type *t1, const Type *t2 ) const {
  const Type *zero = add_id();	// The additive identity
  if( t1->higher_equal( zero ) ) return t2;
  if( t2->higher_equal( zero ) ) return t1;

  return NULL;
}


//=============================================================================
//------------------------------Idealize---------------------------------------
Node *AddINode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  int op1 = in(1)->Opcode();
  int op2 = in(2)->Opcode();
  // Fold (con1-x)+con2 into (con1+con2)-x
  if( op1 == Op_SubI ) {
    const Type *t_sub1 = phase->type( in(1)->in(1) );
    const Type *t_2    = phase->type( in(2)        );
    if( t_sub1->singleton() && t_2->singleton() && t_2 != Type::TOP )
      return new (3) SubINode(phase->makecon( add_ring( t_sub1, t_2 ) ),
			      in(1)->in(2) );
    // Convert "(a-b)+(c-d)" into "(a+c)-(b+d)"
    if( op2 == Op_SubI ) {
      // Check for dead cycle: d = (a-b)+(c-d)
      if( in(1)->in(2) == this || in(2)->in(2) == this ) 
	return new (1) ConNode(Type::TOP);
      Node *sub1 = phase->transform(new (3) AddINode(in(1)->in(1), in(2)->in(1) ) );
      Node *sub2 = phase->transform(new (3) AddINode(in(1)->in(2), in(2)->in(2) ) );
      return new (3) SubINode(sub1, sub2 );
    }
  }

  // Convert "x+(0-y)" into "(x-y)"
  if( op2 == Op_SubI && phase->type(in(2)->in(1)) == TypeInt::ZERO ) 
    return new (3) SubINode(in(1), in(2)->in(2) );

  // Convert "X+X+X+X+X...+X+Y" into "k*X+Y" or really convert "X+(X+Y)"
  // into "(X<<1)+Y" and let shift-folding happen.
  if( op2 == Op_AddI &&
      in(2)->in(1) == in(1) &&
      op1 != Op_ConI && 
      0 ) {
    Node *shift = phase->transform(new (3) LShiftINode(in(1),phase->intcon(1)));
    return new (3) AddINode(shift,in(2)->in(2));
  }

  return AddNode::Ideal( phase, du );
}


//------------------------------Identity---------------------------------------
// Fold (x-y)+y  OR  y+(x-y)  into  x
Node *AddINode::Identity( PhaseTransform *phase ) {
  if( in(1)->Opcode() == Op_SubI && phase->eqv(in(1)->in(2),in(2)) ) {
    return in(1)->in(1);
  }
  else if( in(2)->Opcode() == Op_SubI && phase->eqv(in(2)->in(2),in(1)) ) {
    return in(2)->in(1);
  }
  return AddNode::Identity(phase);
}


//------------------------------add_ring---------------------------------------
// Supplied function returns the sum of the inputs.  Guaranteed never
// to be passed a TOP or BOTTOM type, these are filtered out by
// pre-check.
const Type *AddINode::add_ring( const Type *t0, const Type *t1 ) const {
  const TypeInt *r0 = t0->is_int(); // Handy access
  const TypeInt *r1 = t1->is_int();
  int lo = r0->_lo + r1->_lo;
  int hi = r0->_hi + r1->_hi;
  if( !(r0->is_con() && r1->is_con()) ) {
    // Not both constants, compute approximate result
    if( (r0->_lo & r1->_lo) < 0 && lo >= 0 ) {
      lo = MININT32; hi = MAXINT32; // Underflow on the low side
    }
    if( (~(r0->_hi | r1->_hi)) < 0 && hi < 0 ) {
      lo = MININT32; hi = MAXINT32; // Overflow on the high side
    }
    if( lo > hi ) {               // Handle overflow
      lo = MININT32; hi = MAXINT32;
    }
  } else {
    // both constants, compute precise result using 'lo' and 'hi'
    // Semantics define overflow and underflow for integer addition
    // as expected.  In particular: 0x80000000 + 0x80000000 --> 0x0
  }
  return TypeInt::make( lo, hi, max(r0->_widen,r1->_widen) );
}


//=============================================================================
//------------------------------Idealize---------------------------------------
Node *AddLNode::Ideal( PhaseGVN *phase,PhaseDefUse *du ) {
  int op1 = in(1)->Opcode();
  int op2 = in(2)->Opcode();
  // Fold (con1-x)+con2 into (con1+con2)-x
  if( op1 == Op_SubL ) {
    const Type *t_sub1 = phase->type( in(1)->in(1) );
    const Type *t_2    = phase->type( in(2)        );
    if( t_sub1->singleton() && t_2->singleton() && t_2 != Type::TOP )
      return new (3) SubLNode(phase->makecon( add_ring( t_sub1, t_2 ) ),
			      in(1)->in(2) );
    // Convert "(a-b)+(c-d)" into "(a+c)-(b+d)"
    if( op2 == Op_SubL ) {
      // Check for dead cycle: d = (a-b)+(c-d)
      if( in(1)->in(2) == this || in(2)->in(2) == this ) 
	return new (1) ConNode(Type::TOP);
      Node *sub1 = phase->transform(new (3) AddLNode(in(1)->in(1), in(2)->in(1) ) );
      Node *sub2 = phase->transform(new (3) AddLNode(in(1)->in(2), in(2)->in(2) ) );
      return new (3) SubLNode(sub1, sub2 );
    }
  }

  // Convert "x+(0-y)" into "(x-y)"
  if( op2 == Op_SubL && phase->type(in(2)->in(1)) == TypeLong::ZERO ) 
    return new (3) SubLNode(in(1), in(2)->in(2) );

  // Convert "X+X+X+X+X...+X+Y" into "k*X+Y" or really convert "X+(X+Y)"
  // into "(X<<1)+Y" and let shift-folding happen.
  if( op2 == Op_AddL &&
      in(2)->in(1) == in(1) &&
      op1 != Op_ConL && 
      0 ) {
    Node *shift = phase->transform(new (3) LShiftLNode(in(1),phase->intcon(1)));
    return new (3) AddLNode(shift,in(2)->in(2));
  }

  return AddNode::Ideal( phase, du );
}


//------------------------------Identity---------------------------------------
// Fold (x-y)+y  OR  y+(x-y)  into  x
Node *AddLNode::Identity( PhaseTransform *phase ) {
  if( in(1)->Opcode() == Op_SubL && phase->eqv(in(1)->in(2),in(2)) ) {
    return in(1)->in(1);
  }
  else if( in(2)->Opcode() == Op_SubL && phase->eqv(in(2)->in(2),in(1)) ) {
    return in(2)->in(1);
  }
  return AddNode::Identity(phase);
}


//------------------------------add_ring---------------------------------------
// Supplied function returns the sum of the inputs.  Guaranteed never
// to be passed a TOP or BOTTOM type, these are filtered out by
// pre-check.
const Type *AddLNode::add_ring( const Type *t0, const Type *t1 ) const {
  const TypeLong *r0 = t0->is_long(); // Handy access
  const TypeLong *r1 = t1->is_long();
  jlong lo = r0->_lo + r1->_lo;
  jlong hi = r0->_hi + r1->_hi;
  if( !(r0->is_con() && r1->is_con()) ) {
    // Not both constants, compute approximate result
    if( (r0->_lo & r1->_lo) < 0 && lo >= 0 ) {
      lo = MININT64; hi = MAXINT64; // Underflow on the low side
    }
    if( (~(r0->_hi | r1->_hi)) < 0 && hi < 0 ) {
      lo = MININT64; hi = MAXINT64; // Overflow on the high side
    }
    if( lo > hi ) {               // Handle overflow
      lo = MININT64; hi = MAXINT64;
    }
  } else {
    // both constants, compute precise result using 'lo' and 'hi'
    // Semantics define overflow and underflow for integer addition
    // as expected.  In particular: 0x80000000 + 0x80000000 --> 0x0
  }
  return TypeLong::make( lo, hi, max(r0->_widen,r1->_widen) );
}


//=============================================================================
//------------------------------add_of_identity--------------------------------
// Check for addition of the identity
const Type *AddFNode::add_of_identity( const Type *t1, const Type *t2 ) const {
  // x ADD 0  should return x unless 'x' is a -zero
  // 
  // const Type *zero = add_id();     // The additive identity
  // jfloat f1 = t1->getf();
  // jfloat f2 = t2->getf();
  // 
  // if( t1->higher_equal( zero ) ) return t2;
  // if( t2->higher_equal( zero ) ) return t1;

  return NULL;
}

//------------------------------add_ring---------------------------------------
// Supplied function returns the sum of the inputs.
// This also type-checks the inputs for sanity.  Guaranteed never to
// be passed a TOP or BOTTOM type, these are filtered out by pre-check.
const Type *AddFNode::add_ring( const Type *t0, const Type *t1 ) const {
  // We must be adding 2 float constants.
  return TypeF::make( t0->getf() + t1->getf() );
}

//------------------------------Ideal------------------------------------------
Node *AddFNode::Ideal(PhaseGVN *phase, PhaseDefUse *du) {
  if( !IdealizedNumerics || EnforceStrictFP || phase->C->method()->is_strict() ) {
    // Floating point additions are not associative because of boundary conditions (infinity)
    return commute(this,
                   phase->type( in(1) )->singleton(),
                   phase->type( in(2) )->singleton() ) ? this : NULL;
  } else {
    return AddNode::Ideal(phase, du); // commutative and associative transforms
  }
}


//=============================================================================
//=============================================================================
//------------------------------add_of_identity--------------------------------
// Check for addition of the identity
const Type *AddDNode::add_of_identity( const Type *t1, const Type *t2 ) const {
  // x ADD 0  should return x unless 'x' is a -zero
  // 
  // const Type *zero = add_id();     // The additive identity
  // jfloat f1 = t1->getf();
  // jfloat f2 = t2->getf();
  // 
  // if( t1->higher_equal( zero ) ) return t2;
  // if( t2->higher_equal( zero ) ) return t1;

  return NULL;
}
//------------------------------add_ring---------------------------------------
// Supplied function returns the sum of the inputs.
// This also type-checks the inputs for sanity.  Guaranteed never to
// be passed a TOP or BOTTOM type, these are filtered out by pre-check.
const Type *AddDNode::add_ring( const Type *t0, const Type *t1 ) const {
  // We must be adding 2 double constants.
  return TypeD::make( t0->getd() + t1->getd() );
}

//------------------------------Ideal------------------------------------------
Node *AddDNode::Ideal(PhaseGVN *phase, PhaseDefUse *du) {
  if( !IdealizedNumerics || EnforceStrictFP || phase->C->method()->is_strict() ) {
    // Floating point additions are not associative because of boundary conditions (infinity)
    return commute(this,
                   phase->type( in(1) )->singleton(),
                   phase->type( in(2) )->singleton() ) ? this : NULL;
  } else {
    return AddNode::Ideal(phase, du); // commutative and associative transforms
  }
}


//=============================================================================
//------------------------------Identity---------------------------------------
// If one input is a constant 0, return the other input.
Node *AddPNode::Identity( PhaseTransform *phase ) {
  return ( phase->type( _in[Offset] )->higher_equal( TypeInt::ZERO ) ) ? in(Address) : this;
}

#ifdef ASSERT
Node *base_identity(const Node *addp, PhaseGVN *phase) {
  Node *base = NULL;
  Node *base_identity = addp->in(AddPNode::Base);
  do {
    base          = base_identity;
    base_identity = base->Identity(phase);
  } while( base != base_identity );
  return base;
}
#endif

//------------------------------Idealize---------------------------------------
Node *AddPNode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  // Bail out if dead inputs
  if( phase->type( in(Address) ) == Type::TOP ) return NULL;

  // If the left input is an add of a constant, flatten the expression tree.
  const AddPNode *addp = _in[Address]->is_AddP(); // Left input is an AddP?
  if( addp && addp->in(Base) != phase->C->_top_node ) {
#ifdef ASSERT
    // Check if bases are equivalent
    if( _in[Base] != addp->in(Base) ) {
      // go through identities for both bases to confirm they are equivalent
      Node *my_base        = base_identity(this, phase);
      Node *my_inputs_base = base_identity(addp, phase);
      assert( my_base == my_inputs_base ||
	      phase->hash_find(my_base) == phase->hash_find(my_inputs_base), 
	      "base pointers must match" );
    }
#endif
    // Type of left input's right input
    const Type *t = phase->type( addp->in(Offset) );
    if( t == Type::TOP ) return NULL;
    const TypeInt *t12 = t->is_int();
    if( t12->is_con() ) {	// Left input is an add of a constant?
      // If the right input is a constant, combine constants
      const TypeInt *t2 = phase->type( _in[Offset] )->is_int();
      if( t2->is_con() ) {
        // The Add of the flattened expression
        set_req( Address, addp->in(Address), du );
        set_req( Offset , phase->intcon( t2->get_con() + t12->get_con() ), du);
        return this;			// Made progress
      }
      // Else move the constant to the right.  ((A+con)+B) into ((A+B)+con)
      set_req( Address, phase->transform(new (4) AddPNode(_in[Base],addp->in(Address),_in[Offset])), du );
      set_req( Offset , addp->in(Offset), du );
      return this;
    }
  }

  // If the right is an add of a constant, push the offset down.
  // Convert: (ptr + (offset+con)) into (ptr+offset)+con.
  // The idea is to merge array_base+scaled_index groups together,
  // and only have different constant offsets from the same base.
  const AddNode *add = _in[Offset]->is_Add();
  if( add && (const Node*)add->in(1) != add ) {
    const Type *t22 = phase->type( add->in(2) );
    if( t22->singleton() && (t22 != Type::TOP) ) {  // Right input is an add of a constant?
      set_req( Address, phase->transform(new (4) AddPNode(_in[Base],_in[Address],add->in(1))), du );
      set_req( Offset, add->in(2), du );
      return this;		// Made progress
    }      
  }

  return NULL;			// No progress
}

//------------------------------bottom_type------------------------------------
// Bottom-type is the pointer-type with unknown offset.
const Type *AddPNode::bottom_type() const { 
  const TypePtr *tp = in(Address)->bottom_type()->isa_ptr();
  if( !tp ) return Type::TOP;   // TOP input means TOP output
  int off_op = in(Offset)->Opcode();
  assert( off_op != Op_ConP, "" );
  int offset = (off_op == Op_ConI) ? in(Offset)->get_int() : Type::OffsetBot; 
  return tp->add_offset(offset);
}

//------------------------------Value------------------------------------------
const Type *AddPNode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[Address] );
  const Type *t2 = phase->type( _in[Offset] );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t2 == Type::TOP ) return Type::TOP;

  // Left input is a pointer
  const TypePtr *p1 = t1->isa_ptr();
  // Do not constant fold raw pointers.  This allows heap_top and heap_max
  // (heap_top+4) to remain as relative values instead of two 32-bit
  // constants.  Makes Intel's allocation idiom a little shorter.
  if( p1->_base == Type::RawPtr ) return TypeRawPtr::BOTTOM;
  // Right input is an int
  const TypeInt *p2 = t2->is_int();
  // Add 'em
  return p1->add_offset( p2->is_con() ? p2->get_con() : Type::OffsetBot );
}

//------------------------------match_edge-------------------------------------
// Do we Match on this edge index or not?  Do not match base pointer edge
uint AddPNode::match_edge(uint idx) const {
  return idx > Base;
}

//---------------------------mach_bottom_type----------------------------------
// Utility function for use by ADLC.  Implements bottom_type for matched AddP.
const Type *AddPNode::mach_bottom_type(const MachNode* n) {
  Node* base = n->in(Base);
  const Type *t = base->bottom_type();
  if ( t == Type::TOP ) {
    // an untyped pointer
    return TypeRawPtr::BOTTOM;
  }
  const TypePtr* tp = t->isa_javaptr();
  if ( tp == NULL )  return t;
  if ( tp->_offset == TypePtr::OffsetBot )  return tp;

  // We must carefully add up the various offsets...
  uint offset = 0;
  const TypePtr* tptr = NULL;

  uint numopnds = n->num_opnds();
  uint index = n->oper_input_base();
  for ( uint i = 1; i < numopnds; i++ ) {
    MachOper *opnd = n->_opnds[i];
    for ( uint j = opnd->num_edges(); j > 0; j-- ) {
      Node* edge = n->in(index++);
      const Type*    et  = edge->bottom_type();
      const TypeInt* eti = et->isa_int();
      if ( eti == NULL ) {
	// there must be one pointer among the operands
	guarantee(tptr == NULL, "must be only one pointer operand");
	tptr = et->isa_javaptr();
	guarantee(tptr != NULL, "non-int operand must be pointer");
	continue;
      }
      if ( eti->_hi != eti->_lo )  goto bottom_out;
      offset += eti->_lo;
    }
    int con = opnd->constant();
    if ( con == TypePtr::OffsetBot )  goto bottom_out;
    offset += con;
  }
  guarantee(tptr != NULL, "must be exactly one pointer operand");
  return tptr->add_offset(offset);

 bottom_out:
  return tp->add_offset(TypePtr::OffsetBot);
}

//=============================================================================
//------------------------------add_ring---------------------------------------
// Supplied function returns the sum of the inputs IN THE CURRENT RING.  For
// the logical operations the ring's ADD is really a logical OR function.
// This also type-checks the inputs for sanity.  Guaranteed never to
// be passed a TOP or BOTTOM type, these are filtered out by pre-check.
const Type *OrINode::add_ring( const Type *t0, const Type *t1 ) const {
  const TypeInt *r0 = t0->is_int(); // Handy access
  const TypeInt *r1 = t1->is_int();

  // If either input is not a constant, just return all integers.
  if( !r0->is_con() || !r1->is_con() )
    return TypeInt::INT;	// Any integer, but still no symbols.

  // Otherwise just OR them bits.
  return TypeInt::make( r0->get_con() | r1->get_con() );
}

//=============================================================================
//------------------------------add_ring---------------------------------------
const Type *OrLNode::add_ring( const Type *t0, const Type *t1 ) const {
  const TypeLong *r0 = t0->is_long(); // Handy access
  const TypeLong *r1 = t1->is_long();

  // If either input is not a constant, just return all integers.
  if( !r0->is_con() || !r1->is_con() )
    return TypeLong::LONG;	// Any integer, but still no symbols.

  // Otherwise just OR them bits.
  return TypeLong::make( r0->get_con() | r1->get_con() );
}

//=============================================================================
//------------------------------add_ring---------------------------------------
// Supplied function returns the sum of the inputs IN THE CURRENT RING.  For
// the logical operations the ring's ADD is really a logical OR function.
// This also type-checks the inputs for sanity.  Guaranteed never to
// be passed a TOP or BOTTOM type, these are filtered out by pre-check.
const Type *XorINode::add_ring( const Type *t0, const Type *t1 ) const {
  const TypeInt *r0 = t0->is_int(); // Handy access
  const TypeInt *r1 = t1->is_int();

  // Complementing a boolean?
  if( r0 == TypeInt::BOOL && r1 == TypeInt::ONE )
    return TypeInt::BOOL;

  if( !r0->is_con() || !r1->is_con() ) // Not constants
    return TypeInt::INT;	// Any integer, but still no symbols.

  // Otherwise just XOR them bits.
  return TypeInt::make( r0->get_con() ^ r1->get_con() );
}

//=============================================================================
//------------------------------add_ring---------------------------------------
const Type *XorLNode::add_ring( const Type *t0, const Type *t1 ) const {
  const TypeLong *r0 = t0->is_long(); // Handy access
  const TypeLong *r1 = t1->is_long();

  // If either input is not a constant, just return all integers.
  if( !r0->is_con() || !r1->is_con() )
    return TypeLong::LONG;	// Any integer, but still no symbols.

  // Otherwise just OR them bits.
  return TypeLong::make( r0->get_con() ^ r1->get_con() );
}

//=============================================================================
//------------------------------add_ring---------------------------------------
// Supplied function returns the sum of the inputs.
const Type *MaxINode::add_ring( const Type *t0, const Type *t1 ) const {
  const TypeInt *r0 = t0->is_int(); // Handy access
  const TypeInt *r1 = t1->is_int();

  // Otherwise just MAX them bits.
  return TypeInt::make( max(r0->_lo,r1->_lo), max(r0->_hi,r1->_hi) );
}


//=============================================================================
//------------------------------Idealize---------------------------------------
// MINs show up in range-check loop limit calculations.  Look for
// "min(x+c0,min(y,x+c1))".  Pick the smaller constant: "min(x+c0,y)"
Node *MinINode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  Node *progress = NULL;
  // Force a right-spline graph
  Node *l = _in[1];
  Node *r = _in[2];
  if( l->Opcode() == Op_MinI ) {
    r = phase->transform(new (3) MinINode(l->in(2),r));
    l = l->in(1);
    set_req(1,l,du);
    set_req(2,r,du);
    return this;
  }

  // Get left input & constant
  Node *x = l;
  int x_off = 0;
  if( x->Opcode() == Op_AddI && // Check for "x+c0" and collect constant
      x->in(2)->is_Con() ) {
    x_off = x->in(2)->bottom_type()->is_int()->get_con();
    x = x->in(1);
  }

  // Scan a right-spline-tree for MINs
  Node *y = r;
  int y_off = 0;
  // Check final part of MIN tree
  if( y->Opcode() == Op_AddI && // Check for "x+c0" and collect constant
      y->in(2)->is_Con() ) {
    y_off = y->in(2)->bottom_type()->is_int()->get_con();
    y = y->in(1);
  }
  if( x->_idx > y->_idx ) {
    set_req(1,r,du);
    set_req(2,l,du);
    return this;
  }


  if( r->Opcode() == Op_MinI ) {
    y = r->in(1);
    // Check final part of MIN tree
    if( y->Opcode() == Op_AddI &&// Check for "x+c0" and collect constant
        y->in(2)->is_Con() ) {
      y_off = y->in(2)->bottom_type()->is_int()->get_con();
      y = y->in(1);
    }

    if( x->_idx > y->_idx ) 
      return new (3) MinINode(r->in(1),phase->transform(new (3) MinINode(l,r->in(2))));
    
    // See if covers: min(x+c0,x+c1)
    if( !phase->eqv(x,y) ) return NULL;
    return new (3) MinINode(phase->transform(new (3) AddINode(x,phase->intcon(min(x_off,y_off)))),r->in(2));
  } else {
    // See if covers: min(x+c0,x+c1)
    if( !phase->eqv(x,y) ) return NULL;
    return new (3) AddINode(x,phase->intcon(min(x_off,y_off)));
  }

}

//------------------------------add_ring---------------------------------------
// Supplied function returns the sum of the inputs.
const Type *MinINode::add_ring( const Type *t0, const Type *t1 ) const {
  const TypeInt *r0 = t0->is_int(); // Handy access
  const TypeInt *r1 = t1->is_int();

  // Otherwise just MIN them bits.
  return TypeInt::make( min(r0->_lo,r1->_lo), min(r0->_hi,r1->_hi) );
}
