/*
 * @(#)cfgnode.cpp	1.185 01/03/05 10:08:59
 *
 * Copyright 1991-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
// Portions of code courtesy of Clifford Click

// Optimization - Graph Style

#include "incls/_precompiled.incl"
#include "incls/_cfgnode.cpp.incl"

//=============================================================================
//------------------------------Value------------------------------------------
// Compute the type of the RegionNode.
const Type *RegionNode::Value( PhaseTransform *phase ) const {
  for( uint i=1; i<req(); ++i ) {       // For all paths in
    Node *n = _in[i];           // Get Control source
    if( !n ) continue;          // Missing inputs are TOP
    if( phase->type(n) == Type::CONTROL )
      return Type::CONTROL;
  }
  return Type::TOP;             // All paths dead?  Then so are we
}

//------------------------------Identity---------------------------------------
// Check for Region being Identity.
Node *RegionNode::Identity( PhaseTransform *phase ) {
  // Cannot have Region be an identity, even if it has only 1 input.
  // Phi users cannot have their Region input folded away for them,
  // since they need to select the proper data input
  return this;
}

//------------------------------merge_region-----------------------------------
// If a Region flows into a Region, merge into one big happy merge.  This is
// hard to do if there is stuff that has to happen
static Node *merge_region( Node *region, PhaseGVN *phase, PhaseDefUse *du ) {
  if( !du ) return NULL;        // Need DU
  if( region->Opcode() != Op_Region ) // Do not do to LoopNodes
    return NULL;
  Node *progress = NULL;        // Progress flag
  PhaseIterGVN *igvn = phase->is_IterGVN();

  uint max = du->cnt(region);

  uint rreq = region->req();
  for( uint i = 1; i < rreq; i++ ) {
    Node *r = region->in(i);
    if( r && r->Opcode() == Op_Region && // Found a region?
        r->in(0) == r &&        // Not already collapsed?
	r != region &&		// Avoid stupid situations
        du->cnt(r) == 2 ) {     // Self user and 'region' user only?
#ifdef ASSERT
      for( uint k = du->cnt(r); k > 0; k-- )
        assert( !du->out(r)[k-1]->is_Phi(), "no phi users" );
#endif
      if( !progress ) {         // No progress
        Node **outp = du->out(region);
        for( uint k = 0; k < max; k++ ) {
          Node *phi = outp[k];
          if( phi->is_Phi() ) {
            return NULL;        // Only flatten if no Phi users
            // igvn->hash_delete( phi );
          }
        }
        igvn->hash_delete( region );
        progress = region;      // Making progress
      }
      igvn->hash_delete( r );

      // Append inputs to 'r' onto 'region'
      for( uint j = 1; j < r->req(); j++ ) {
        // Move an input from 'r' to 'region'
        region->add_req( NULL );
        region->set_req( rreq, r->in(j), du );
        r->set_req( j, phase->C->_top_node, du );
        // Update phis of 'region'
        //for( uint k = 0; k < max; k++ ) {
        //  Node *phi = du->out(region)[k];
        //  if( phi->is_Phi() ) {
        //    phi->add_req( NULL );
        //    phi->set_req( rreq, phi->in(i), du );
        //  }
        //}

        rreq++;                 // One more input to Region
      } // Found a region to merge into Region
      // Clobber pointer to the now dead 'r'
      region->set_req( i, phase->C->_top_node, du );
    }
  }

  return progress;
}



//------------------------------only_phi_use-----------------------------------
// Helper function: Return the only PhiNode that uses this region or NULL
static const PhiNode * const multiple_phis = (const PhiNode *)-1;
static PhiNode *only_phi_use( const RegionNode *region, PhaseDefUse *du ) {
  // Check that only one use is a Phi
  const PhiNode *only_phi = NULL;

  uint   max  = du->cnt(region);
  Node **outp = du->out(region);
  for( uint i = 0; i < max; ++i ) {
    if( outp[i]->is_Phi() ) {   // Check for Phi users
      only_phi = (only_phi == NULL) ? (PhiNode*)outp[i] : multiple_phis;
    }
  }

  return ( (only_phi != multiple_phis) ? (PhiNode*)only_phi : NULL );
}


//------------------------------check_phi_clipping-----------------------------
// Helper function for RegionNode's identification of FP clipping
// Check inputs to the Phi
static bool check_phi_clipping( PhiNode *phi, ConNode * &min, uint &min_idx, ConNode * &max, uint &max_idx, Node * &val, uint &val_idx ) {
  min     = NULL;
  max     = NULL;
  val     = NULL;
  min_idx = 0;
  max_idx = 0;
  val_idx = 0;
  uint  phi_max = phi->req();
  if( phi_max == 4 ) {
    for( uint j = 1; j < phi_max; ++j ) {
      Node *n = phi->in(j);
      int opcode = n->Opcode();
      switch( opcode ) {
      case Op_ConI: 
        {
          if( min == NULL ) {
            min     = n->Opcode() == Op_ConI ? (ConNode*)n : NULL;
            min_idx = j;
          } else {
            max     = n->Opcode() == Op_ConI ? (ConNode*)n : NULL;
            max_idx = j;
            if( min->get_int() > max->get_int() ) {
              // Swap min and max
              ConNode *temp;
              uint     temp_idx;
              temp     = min;     min     = max;     max     = temp;
              temp_idx = min_idx; min_idx = max_idx; max_idx = temp_idx;
            }
          }
        }
        break;
      default:
        {
          val = n;
          val_idx = j;
        }
        break;
      }
    }
  }
  return ( min && max && val && (min->get_int() <= 0) && (max->get_int() >=0) );
}
  

//------------------------------check_if_clipping------------------------------
// Helper function for RegionNode's identification of FP clipping
// Check that inputs to Region come from two IfNodes, 
// 
//            If
//      False    True
//       If        |
//  False  True    |
//    |      |     |
//  RegionNode_inputs
// 
static bool check_if_clipping( const RegionNode *region, IfNode * &bot_if, IfNode * &top_if ) {
  top_if = NULL;
  bot_if = NULL;

  // Check control structure above RegionNode for (if  ( if  ) )
  ProjNode *in1 = region->in(1)->is_Proj();
  ProjNode *in2 = region->in(2)->is_Proj();
  ProjNode *in3 = region->in(3)->is_Proj();
  // Check that all inputs are projections
  if( in1 && in2 && in3 ) {
    IfNode *in10 = in1->in(0) ? in1->in(0)->is_If() : NULL;
    IfNode *in20 = in2->in(0) ? in2->in(0)->is_If() : NULL;
    IfNode *in30 = in3->in(0) ? in3->in(0)->is_If() : NULL;
    // Check that #1 and #2 are ifTrue and ifFalse from same If
    if( in10 && in20 && in30 && in10 == in20 && (in1->Opcode() != in2->Opcode()) ) {
      ProjNode *in100  = in10->in(0)  ? in10->in(0)->is_Proj() : NULL;
      IfNode   *in1000 = in100 && in100->in(0) ? in100->in(0)->is_If() : NULL;
      // Check that control for in10 comes from other branch of IF from in3
      if( in30 == in1000 && (in3->Opcode() != in100->Opcode()) ) {
        // Control pattern checks
        top_if = in1000;
        bot_if = in10;
      }
    }
  }

  return (top_if != NULL);
}


//------------------------------check_convf2i_clipping-------------------------
// Helper function for RegionNode's identification of FP clipping
// Verify that the value input to the phi comes from "ConvF2I; LShift; RShift"
static bool check_convf2i_clipping( PhiNode *phi, uint idx, ConvF2INode * &convf2i, Node *min, Node *max) {
  convf2i = NULL;

  // Check for the RShiftNode
  Node *rshift = phi->in(idx);
  assert( rshift, "Previous checks ensure phi input is present");
  if( rshift->Opcode() != Op_RShiftI )  { return false; }

  // Check for the LShiftNode
  Node *lshift = rshift->in(1);
  assert( lshift, "Previous checks ensure phi input is present");
  if( lshift->Opcode() != Op_LShiftI )  { return false; }

  // Check for the ConvF2INode
  Node *conv = lshift->in(1);
  if( conv->Opcode() != Op_ConvF2I ) { return false; }

  // Check that shift amounts are only to get sign bits set after F2I
  jint max_cutoff     = max->get_int();
  jint min_cutoff     = min->get_int();
  jint left_shift     = lshift->in(2)->get_int();
  jint right_shift    = rshift->in(2)->get_int();
  jint max_post_shift = nth_bit(BitsPerJavaInteger - left_shift - 1);
  if( left_shift != right_shift || 
      0 > left_shift || left_shift >= BitsPerJavaInteger ||
      max_post_shift < max_cutoff ||
      max_post_shift < -min_cutoff ) {
    // Shifts are necessary but current transformation eliminates them
    return false;
  }

  // OK to return the result of ConvF2I without shifting
  convf2i = (ConvF2INode*)conv;
  return true;
}


//------------------------------float_compare_path-----------------------------
// Helper function for RegionNode's identification of FP clipping
static bool check_compare_clipping( bool less_than, IfNode *iff, ConNode *limit, Node * & input ) {
  BoolNode *bool1 = iff->in(1)->is_Bool();
  if ( bool1 == NULL ) { return false; }
  if(       less_than && bool1->_test._test != BoolTest::le ) { return false; }
  else if( !less_than && bool1->_test._test != BoolTest::lt ) { return false; }
  const SubNode  *cmpI = bool1->in(1)->is_Cmp();
  if( cmpI == NULL )                                          { return false; }
  if( cmpI->in(2)->bottom_type() != TypeInt::ZERO )           { return false; }
  const SubNode *cmpF = cmpI->in(1)->is_Cmp();
  if( cmpF == NULL )                                          { return false; }
  if(       less_than && cmpF->Opcode() != Op_CmpFLess )      { return false; }
  else if( !less_than && cmpF->Opcode() != Op_CmpFGreater )   { return false; }
  if( cmpF->in(0) != NULL )  { return false; }
  // Test that the float value being compared against
  // is equivalent to the int value used as a limit
  Node *nodef = cmpF->in(2);
  if( nodef->Opcode() != Op_ConF ) { return false; }
  jfloat conf = nodef->getf();
  jint   coni = limit->get_int();
  if( ((int)conf) != coni )        { return false; }
  input = cmpF->in(1);
  return true;
}

//------------------------------Ideal------------------------------------------
// Return a node which is more "ideal" than the current node.  Must preserve
// the CFG, but we can still strip out dead paths.
Node *RegionNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  if( !_in[0] ) return NULL;    // Already degraded to a Copy

  // Check for RegionNode with no Phi users and both inputs come from either
  // arm of the same IF.  If found, then the control-flow split is useless.
  if( du ) {                    // Need DU info to check for Phi users
    uint   max  = du->cnt(this);
    Node **outp = du->out(this);
    uint i;
    for( i = 0; i < max; i++ )
      if( outp[i]->is_Phi() )   // Check for Phi users
        break;
    if( i == max ) {            // No Phi users?  Nothing merging?
      for( i=1; i<req()-1; i++ ) {
        Node *if1 = in(i);
        if( !if1 ) continue;
        Node *iff = if1->in(0);
        if( !iff || !iff->is_If() ) continue;
        for( uint j=i+1; j<req(); j++ ) {
          if( in(j) && in(j)->in(0) == iff &&
              if1->Opcode() != in(j)->Opcode() ) {
            set_req(i,iff->in(0),du);// Skip around the useless IF diamond
            set_req(j,NULL,du);
            return this;      // Record progress
          }
        }
      }
    }
  }

  // Remove TOP or NULL input paths. If only 1 input path remains, this Region 
  // degrades to a copy.
  int cnt = 0;                  // Count of values merging
  // For all inputs...
  for( uint i=1; i<req(); ++i ){// For all paths in
    Node *n = in(i);            // Get the input
    if( n ) {
      // Remove useless control copy inputs
      if( n->is_Region() && !n->in(0) ) {
        set_req(i,n->nonnull_req(),du);
        i--;
        continue;
      }
      if( n->is_Proj() ) {      // Remove useless rethrows
        CallNode *call = n->in(0)->is_Call();
        if( call && call->_entrypoint == OptoRuntime::rethrow_stub() ) {
          set_req(i,call->in(0),du);
          i--;
          continue;
        }
      }
      if( phase->type(n) == Type::TOP ) {
        set_req(i,NULL,du);     // Ignore TOP inputs
        i--;
        continue;
      }
      cnt++;                    // One more value merging

    } else if( du ) {           // Else found dead path with DU info
      PhaseIterGVN *igvn = phase->is_IterGVN();
      del_req(i);               // Yank path from self
      uint max = du->cnt(this);
      bool progress = true;
      while( progress ) {
	progress = false;
	for( uint j=0; j<max; j++ ) {
	  Node *n = du->out(this)[j];
	  if( n->req() != req() && n->is_Phi() ) {
	    assert( n->in(0) == this, "" );
	    igvn->hash_delete(n); // Yank from hash before hacking edges
	    n->set_req_X(i,NULL,igvn);// Correct DU info
	    n->del_req(i);        // Yank path from Phis
	    if( max != du->cnt(this) ) {
	      progress = true;
	      max = du->cnt(this);
	    }
	  }
	}
      }
      igvn->add_users_to_worklist(this); // Revisit collapsed Phis
      i--;
    }
  }
  
  if( cnt <= 1 ) {              // Only 1 path in?  Become a copy!
    set_req(0, NULL, du);       // Null control input for region copy
    if( cnt == 0 ) {
      if( req() > 1 ) // No inputs at all?
        set_req(1,phase->C->_top_node,du);  // Then copy from TOP control
    } else if( du ) {
      PhaseIterGVN *igvn = phase->is_IterGVN();
      Node *parent_ctrl = in(1);
      assert(parent_ctrl != NULL, "Region is a copy of some non-null control");
      if( parent_ctrl == this ) // Closed, dead, loop?
	return new (1) ConNode( Type::TOP );
      igvn->add_users_to_worklist(this); // Check for further allowed opts
      while( du->cnt(this) ) {
        // du->out inside loop since set_req() may force move of edges
        Node *n = du->out(this)[0];
        igvn->hash_delete(n); // Remove from worklist before modifying edges
        if( n->is_Phi() ) {   // Collapse all Phis to copies
          n->set_req(0,NULL,du);
        } 
        else if( n->is_Region() ) { // Update all incoming edges
          assert( n != this, "Must be removed from DefUse edges");
          for( uint k=1; k < n->req(); k++ ) 
            if( n->in(k) == this ) 
              n->set_req(k, parent_ctrl, du);
        }
        else {
          assert( n->in(0) == this, "Expect RegionNode to be control parent");
          n->set_req(0, parent_ctrl, du);
        }
#ifdef ASSERT
        for( uint k=0; k < n->req(); k++ ) {
          assert( n->in(k) != this, "All uses of RegionNode should be gone");
        }
#endif
      }
      // Remove the RegionNode itself from DefUse info
      igvn->remove_dead_node(this);
    }
    return this;                // Record progress
  }


  // If a Region flows into a Region, merge into one big happy merge.
  Node *m = merge_region( this, phase, du );
  if( m ) return m;

  // Check if this region is the root of a clipping idiom on floats
  if( ConvertFloat2IntClipping && du && req() == 4 ) {
    // Check that only one use is a Phi and that it simplifies to two constants +
    PhiNode *phi = only_phi_use( this, du );
    if( phi ) {            // One Phi user
      // Check inputs to the Phi
      ConNode *min;
      ConNode *max;
      Node    *val;
      uint     min_idx;
      uint     max_idx;
      uint     val_idx;
      if( check_phi_clipping( phi, min, min_idx, max, max_idx, val, val_idx )  ) {
        IfNode *top_if;
        IfNode *bot_if;
        if( check_if_clipping( this, bot_if, top_if ) ) {
          // Control pattern checks, now verify compares
          Node   *top_in = NULL;   // value being compared against
          Node   *bot_in = NULL;
          if( check_compare_clipping( true,  bot_if, min, bot_in ) && 
              check_compare_clipping( false, top_if, max, top_in ) ) {
            if( bot_in == top_in ) {
              PhaseIterGVN *gvn = phase->is_IterGVN();
              assert( gvn != NULL, "Only had DefUse info in IterGVN");
              // Only remaining check is that bot_in == top_in == (Phi's val + mods)

              // Check for the ConvF2INode
              ConvF2INode *convf2i;
              if( check_convf2i_clipping( phi, val_idx, convf2i, min, max ) && 
                convf2i->in(1) == bot_in ) {
                // Matched pattern, including LShiftI; RShiftI, replace with integer compares
                // max test
                Node *cmp   = gvn->register_new_node_with_optimizer(new (3) CmpINode( convf2i, min ));
                Node *boo   = gvn->register_new_node_with_optimizer(new (2) BoolNode( cmp, BoolTest::lt ));
                IfNode *iff = (IfNode*)gvn->register_new_node_with_optimizer(new (2) IfNode( top_if->in(0), boo, 0.00001f, top_if->_cnt ));
                Node *if_min= gvn->register_new_node_with_optimizer(new (1) IfTrueNode (iff));
                Node *ifF   = gvn->register_new_node_with_optimizer(new (1) IfFalseNode(iff));
                // min test
                cmp         = gvn->register_new_node_with_optimizer(new (3) CmpINode( convf2i, max ));
                boo         = gvn->register_new_node_with_optimizer(new (2) BoolNode( cmp, BoolTest::gt ));
                iff         = (IfNode*)gvn->register_new_node_with_optimizer(new (2) IfNode( ifF, boo, 0.00001f, bot_if->_cnt ));
                Node *if_max= gvn->register_new_node_with_optimizer(new (1) IfTrueNode (iff));
                ifF         = gvn->register_new_node_with_optimizer(new (1) IfFalseNode(iff));
                // update input edges to region node
                set_req_X( min_idx, if_min, gvn );
                set_req_X( max_idx, if_max, gvn );
                set_req_X( val_idx, ifF,    gvn );
                // remove unnecessary 'LShiftI; RShiftI' idiom
                gvn->hash_delete(phi);
                phi->set_req_X( val_idx, convf2i, gvn );
                gvn->hash_find_insert(phi);
                // Return transformed region node
                return this;
              }
            }
          }
        }
      }
    }
  }

  return NULL;
}



const RegMask &RegionNode::out_RegMask() const { 
  return RegMask::Empty;
}

// Find the one non-null required input.  RegionNode only
Node *Node::nonnull_req() const {
  assert( is_Region(), "" );
  for( uint i = 1; i < _cnt; i++ )
    if( _in[i] )
      return _in[i];
  ShouldNotReachHere();
  return NULL;
}


//=============================================================================
//------------------------------Value------------------------------------------
// Compute the type of the PhiNode
const Type *PhiNode::Value( PhaseTransform *phase ) const {
  Node *r = _in[0];             // RegionNode
  if( !r )                      // Copy or dead
    return in(1) ? phase->type(_in[1]) : Type::TOP;
  if( phase->type(r) == Type::TOP ) // Dead code?
    return Type::TOP;

  // Check for trip-counted loop.  If so, be smarter.
  CountedLoopNode *l = r->is_CountedLoop();
  if( l && l->in(0) && l->req() == 3 && 
      ((const Node*)l->phi() == this) ) { // Trip counted loop!
    const TypeInt *lo = l->init_trip()->bottom_type()->isa_int();
    const TypeInt *hi = l->limit()    ->bottom_type()->isa_int();
    if( lo && hi ) {		// Dying loops might have TOP here
      int stride = l->stride_con();
      if( stride < 0 ) {          // Down-counter loop
	const TypeInt *tmp = lo; lo = hi; hi = tmp;
	stride = -stride;
      }
      if( lo->_hi < hi->_lo )     // Reversed endpoints are well defined :-(
	return TypeInt::make(lo->_lo,hi->_hi);
    }
  }

  // Default case: merge all inputs
  const Type *t = Type::TOP;        // Merged type starting value
  for( uint i = 1; i < req(); ++i ) // For all paths in
    // Reachable control path?
    if( r->in(i) && phase->type(r->in(i)) == Type::CONTROL)
      t = t->meet( phase->type(in(i)) );
  // If we grew the result range, we also need to widen
  t = phase->widen(t, phase->_types[_idx]); // Widen once per Phi
  // Join with worst-case type, so we can pass a BOTTOM to a PhiNode
  // producing only INTs, for instance.
  return t->join(_type);            // Worst case type
}

//------------------------------Identity---------------------------------------
// Check for Region being Identity.
Node *PhiNode::Identity( PhaseTransform *phase ) {
  Node *r = _in[0];             // RegionNode
  if( !r ) return _in[1];       // Already degraded to a Copy 

  // Check for a loop-phi merging same value on back edge, where backedge
  // value is itself a phi merging not-null and unknown versions of the
  // same loop-phi.
  if( r->is_Loop() && in(LoopNode::LoopBackControl)->is_Phi() ) {
    Node *phi = in(LoopNode::LoopBackControl);
    uint i;
    for( i = 1; i < phi->req(); i++ )
      if( phi->in(i) != this &&
	  (phi->in(i)->Opcode() != Op_CastPP ||
	   phi->in(i)->in(1) != phi) )
	break;
    if( i == phi->req() )
      return in(LoopNode::EntryControl);
  }

  // Check for no merging going on
  Node *input = NULL;           // The one unique input
  // For all inputs...
  for( uint i=1; i<req(); ++i ){// For all paths in
    // Unreachable control path?
    Node *rc = r->in(i);
    if( !rc || phase->type(rc) == Type::TOP ) 
      continue;
    Node *n = _in[i];           // Get the input
    if( phase->type(n) == Type::TOP )
      continue;                 // Ignore TOP inputs
    if( phase->eqv(n,this) ) continue; // Ignore simple cycles
    // Ignore more complex cycles using a cast-away-nullness 
    // feeding into a Phi
    if( n->Opcode() == Op_CastPP &&
	phase->eqv(n->in(1),this) )
      continue;

    if( input && input != n )   // Merging different values?
      return this;              // Then not any identity
    input = n;                  // Capture
  }
  return input ? input : phase->C->_top_node; // Return unique thing we are equal to
}

//------------------------------is_x2logic-------------------------------------
// Check for simple convert-to-boolean pattern
// If:(C Bool) Region:(IfF IfT) Phi:(Region 0 1)
// Convert Phi to an ConvIB.
Node *is_x2logic( PhaseGVN *phase, Node *phi ) {
  int flipped = 0;
  Node *region = phi->in(0);
  if( !region ) return NULL;
  if( region->req() != 3 ) return NULL;
  Node *iffalse = region->in(1);
  Node *iftrue = region->in(2);
  if( !iffalse || !iftrue ) return NULL;
  Node *iff = iffalse->in(0);
  if( !iff || !iff->is_If() ) return NULL;
  if( iff != iftrue->in(0) ) return NULL;
  if( phi->req() != 3 ) return NULL;
  Node *zero = phi->in(1);
  const Type *tzero = phase->type( zero );
  Node *one = phi->in(2);
  const Type *tone = phase->type( one );
  BoolNode *bool = iff->in(1)->is_Bool();
  if( !bool ) return NULL;

  // Check for compare vs 0
  const SubNode *cmp = bool->in(1)->is_Cmp();
  if( !cmp ) return NULL;
  const Type *tcmp = phase->type(cmp->in(2));
  if( tcmp != TypeInt::ZERO && tcmp != TypePtr::NULL_PTR ) {
    // Allow cmp-vs-1 if the other input is bounded by 0-1
    if( !(tcmp == TypeInt::ONE && phase->type(cmp->in(1)) == TypeInt::BOOL) )
      return NULL;
    flipped = 1-flipped;	// Test is vs 1 instead of 0!
  }

  // Check for branching opposite expected
  if( iftrue->Opcode() == Op_IfTrue ) {
    assert( iffalse->Opcode() == Op_IfFalse, "" );
  } else {
    assert( iffalse->Opcode() == Op_IfTrue, "" );
    flipped = 1-flipped;
  }

  // Check for setting zero/one opposite expected
  if( tzero == TypeInt::ZERO ) {
    if( tone == TypeInt::ONE ) {
    } else return NULL;
  } else if( tzero == TypeInt::ONE ) {
    if( tone == TypeInt::ZERO ) {
      flipped = 1-flipped;
    } else return NULL;
  } else return NULL;

  // Check for boolean test backwards
  if( bool->_test._test == BoolTest::ne ) {
  } else if( bool->_test._test == BoolTest::eq ) {
    flipped = 1-flipped;
  } else return NULL;

  // Build int->bool conversion
  Node *n = new (2) Conv2BNode( cmp->in(1) );
  if( flipped ) 
    n = new (3) XorINode( phase->transform(n), phase->intcon(1) );

  return n;
}

//------------------------------is_cond_add------------------------------------
// Check for simple conditional add pattern:  "(P < Q) ? X+Y : X;"
// To be profitable the control flow has to disappear; there can be no other
// values merging here.  We replace the test-and-branch with: 
// "(sgn(P-Q))&Y) + X".  Basically, convert "(P < Q)" into 0 or -1 by
// moving the carry bit from (P-Q) into a register with 'sbb EAX,EAX'.  
// Then convert Y to 0-or-Y and finally add.
// This is a key transform for SpecJava _201_compress.
Node *is_cond_add( PhaseGVN *phase, PhaseDefUse *du, Node *phi ) {
  Node *region = phi->in(0);  
  if( region->req() != 3 ) return NULL;
  if( !du ) return NULL;
  // Make sure only merging this one phi here
  uint   max  = du->cnt(region);
  Node **outp = du->out(region);
  for( uint i = 0 ; i < max; i++ )
    if( outp[i]->is_Phi() && outp[i] != phi ) 
      return NULL;
  Node *if_t = region->in(1);
  if( !if_t ) return NULL;
  Node *if_f = region->in(2);
  if( !if_f ) return NULL;
  Node *iff = if_t->in(0);
  if( !iff || iff != if_f->in(0) ) return NULL;
  if( iff->Opcode() != Op_If ) return NULL;

  BoolNode *b = iff->in(1)->is_Bool();
  if( !b || b->_test._test != BoolTest::lt ) return NULL;

  Node *cmp = b->in(1);
  if( cmp->Opcode() != Op_CmpI ) return NULL;
  Node *p = cmp->in(1);
  Node *q = cmp->in(2);
  Node *n1 = phi->in(1);
  Node *n2 = phi->in(2);

  if( if_t->Opcode() == Op_IfFalse ) {
    Node *tmp = n1; n1 = n2; n2 = tmp;
    tmp = if_t; if_t = if_f; if_f = tmp;
  }

  int op = n1->Opcode();
  if( op != Op_AddI           // Need zero as additive identity
      /*&&op != Op_SubI &&
      op != Op_AddP &&
      op != Op_XorI &&
      op != Op_OrI*/ )
    return NULL;

  Node *x = n2;
  Node *y = n1->in(1);
  if( n2 == n1->in(1) ) {
    y = n1->in(2);
  } else if( n2 == n1->in(1) ) {
  } else return NULL;

  // Not so profitable if compare and add are constants
  if( q->is_Con() && phase->type(q) != TypeInt::ZERO && y->is_Con() ) 
    return NULL;
  
  Node *cmplt = phase->transform( new (3) CmpLTMaskNode(p,q) );
  Node *j_and   = phase->transform( new (3) AndINode(cmplt,y) );
  return new (3) AddINode(j_and,x);
}

//------------------------------split_once-------------------------------------
// Helper for split_flow_path
static void split_once( PhaseIterGVN *igvn, PhaseDefUse *du, Node *phi, Node *val, Node *n, Node *newn ) {
  igvn->hash_delete(n);		// Remove from hash before hacking edges

  uint j = 1;
  for( uint i = phi->req()-1; i > 0; i-- ) {
    if( phi->in(i) == val ) {	// Found a path with val?
      // Add to NEW Region/Phi, no DU info
      newn->set_req( j++, n->in(i) );
      // Remove from OLD Region/Phi
      n->set_req(i,NULL,du);
      n->del_req(i);
    }
  }

  // Register the new node but do not transform it.  Cannot transform until the
  // entire Region/Phi conglerate has been hacked as a single huge transform.
  igvn->register_new_node_with_optimizer( newn );
  // Now I can point to the new node.
  n->add_req(NULL);
  n->set_req( n->req()-1, newn, du );
  igvn->_worklist.push(n);
}

//------------------------------split_flow_path--------------------------------
// Check for merging identical values and split flow paths
static Node *split_flow_path( PhaseGVN *phase, PhaseDefUse *du, PhiNode *phi) {
  if( !du ) return NULL;

  BasicType bt = Type::_basic_type[phi->_type->_base];
  if( bt == T_ILLEGAL || type2size[bt] <= 0 )
    return NULL;		// Bail out on funny non-value stuff
  if( phi->req() <= 3 )		// Need at least 2 matched inputs and a
    return NULL;		// third unequal input to be worth doing

  // Scan for a constant
  uint i;
  for( i = 1; i < phi->req()-1; i++ ) {
    Node *n = phi->in(i);
    if( !n ) return NULL;
    if( phase->type(n) == Type::TOP ) return NULL;
    if( n->Opcode() == Op_ConP )
      break;
  }
  if( i >= phi->req() )		// Only split for constants
    return NULL;
  
  Node *val = phi->in(i);	// Constant to split for
  uint hit = 0;			// Number of times it occurs

  for( ; i < phi->req(); i++ ){ // Count occurances of constant
    Node *n = phi->in(i);
    if( !n ) return NULL;
    if( phase->type(n) == Type::TOP ) return NULL;
    if( phi->in(i) == val ) 
      hit++;
  }

  if( hit <= 1 ||		// Make sure we find 2 or more
      hit == phi->req()-1 )	// and not ALL the same value
    return NULL;

  // Now start splitting out the flow paths that merge the same value.
  // Split first the RegionNode.
  PhaseIterGVN *igvn = phase->is_IterGVN();
  Node *r = phi->in(0);
  RegionNode *newr = new RegionNode(hit+1);
  split_once( igvn, du, phi, val, r, newr );

  // Now split all other Phis than this one
  uint   max  = du->cnt(r);
  for( i = 0; i < max; i++ ) {
    PhiNode *phi2 = du->out(r)[i]->is_Phi();
    if( phi2 && phi2 != phi ) {
      PhiNode *newphi = new PhiNode( newr, phi2->_type );
      split_once( igvn, du, phi, val, phi2, newphi );
    }
  }      

  // Clean up this guy
  igvn->hash_delete(phi);
  for( i = phi->req()-1; i > 0; i-- ) {
    if( phi->in(i) == val ) {
      phi->set_req(i,NULL,du);
      phi->del_req(i);
    }
  }
  phi->add_req(NULL);
  phi->set_req(phi->req()-1,val,du);

  return phi;
}


//------------------------------Ideal------------------------------------------
// Return a node which is more "ideal" than the current node.  Must preserve
// the CFG, but we can still strip out dead paths.
Node *PhiNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  Node *r = _in[0];             // RegionNode
  if( r == NULL )               // Already degraded to a Copy ?
    return NULL;                // No change
  if( phase->type(r) == Type::TOP ) // Dead code?
    return NULL;                // No change

  // Check for simple convert-to-boolean pattern
  Node *xbool = is_x2logic(phase,this);
  if( xbool ) return xbool;

  // Check for absolute value
  Node *abs = AbsNode::is_absolute(phase,this);
  if( abs ) return abs;

  // Check for conditional add
  Node *add = is_cond_add(phase,du,this);
  if( add ) return add;

  // Check for merging identical values and split flow paths
  Node *foo = split_flow_path( phase, du, this );
  if( foo ) return foo;

  // Check for merging only 1 value
  Node *progress = NULL;        // Record if any progress made
  Node *u = NULL;               // First unique input
  // For all inputs...
  for( uint i=1; i<req(); ++i ){// For all paths in
    // Unreachable control path?
    Node *rc = r->in(i);
    Node *n = in(i);            // Get the input
    if( (!rc || phase->type(rc) == Type::TOP) ) {
      if( n != phase->C->_top_node ) { // Not already top?  
        set_req(i,phase->C->_top_node,du);// Nuke it down
        progress = this;        // Record progress
      }
      continue;                 // Dead input is ignored
    }
    // assert( n->Opcode() != Op_MergeMem, "Fix this in parser" );
    if( phase->eqv(n,this) ) continue; // Ignore simple cycles
    // // TOP input on a live path?  By JVM semantics must be unused!
    // Unfortunately inputs to Phi and Region may not be consistent
    // Set input to Phi to a constant,
    // If Region's input becomes TOP, then Phi's input will become TOP also
    if( phase->type(n) == Type::TOP ) {
      Node *con_in;
      switch( bottom_type()->_base ) {  // Flatten worst-case type for Phis
      case Type::Long:      con_in = phase->makecon(TypeLong::ZERO); break;
      case Type::Int:       con_in = phase->intcon(0); break;
      case Type::FloatCon:  con_in = phase->makecon(TypeF::ZERO);  break;
      case Type::DoubleCon: con_in = phase->makecon(TypeD::ZERO); break;
      case Type::AnyPtr:    con_in = phase->makecon(TypePtr::NULL_PTR); break;
      case Type::OopPtr:    // fall through
      case Type::AryPtr:    con_in=phase->makecon(TypeJavaPtr::NULL_PTR);break;
      default:              con_in = NULL; break;
      }
      if( con_in ) {            // oop-mem or some such; ignore NULLs
        n = con_in;
        set_req(i, con_in, du);
        progress = this;        // Record progress
        if( TraceDeadPhis ) {
          tty->print_cr("  Applying Conservative Gosling Property");
#ifndef PRODUCT
          tty->print("  "); r->dump();
          tty->print("  "); dump();
          tty->cr();
#endif
        }
      }
    }
    if( !u ) u = n;             // First unique input
    else if( u != n )           // Do they differ?
      u = (Node*)-1;            // Then flag as a real merge
  }

  if( u != (Node*)-1 && u ) {   // Unique value?
    if( du ) {
      PhaseIterGVN *igvn = phase->is_IterGVN();
      set_req_X(0,NULL,igvn);     // Unique value; degrade to Copy
      set_req_X(1,u,igvn);          // Copy of the one unique input
      for( uint i = 2; i < req(); i++ )// Remaining inputs are smashed
        set_req_X(i,NULL,igvn);
    } else {
      set_req(0,NULL,du);     // Unique value; degrade to Copy
      set_req(1,u,du);          // Copy of the one unique input
      for( uint i = 2; i < req(); i++ )// Remaining inputs are smashed
        set_req(i,NULL,du);
    }
    progress = this;            // Record progress
  }
  return progress;              // Return any progress
}

//------------------------------out_RegMask------------------------------------
const RegMask &PhiNode::in_RegMask(uint i) const { 
  return i ? out_RegMask() : RegMask::Empty;
}

const RegMask &PhiNode::out_RegMask() const { 
  uint ideal_reg = Matcher::base2reg[_type->_base];
  assert( ideal_reg != Node::NotAMachineReg, "invalid type at Phi" );
  if( ideal_reg == 0 ) return RegMask::Empty;
  return *(Compile::current()->_matcher->idealreg2spillmask[ideal_reg]);
}


//=============================================================================
const Type *GotoNode::Value( PhaseTransform *phase ) const {
  // If the input is reachable, then we are executed.
  // If the input is not reachable, then we are not executed.
  return phase->type(_in[0]); 
}

Node *GotoNode::Identity( PhaseTransform *phase ) {
  return _in[0];                // Simple copy of incoming control
}

const RegMask &GotoNode::out_RegMask() const { 
  return RegMask::Empty;
}


//=============================================================================
const RegMask &CProjNode::out_RegMask() const { 
  return RegMask::Empty;
}



//=============================================================================

uint PCTableNode::hash() const { return Node::hash() + _size; }
uint PCTableNode::cmp( const Node &n ) const
{ return _size == ((PCTableNode&)n)._size; }

const Type *PCTableNode::bottom_type() const {
  const Type** f = TypeTuple::fields(_size);
  for( uint i = 0; i < _size; i++ ) f[i] = Type::CONTROL;
  return TypeTuple::make(_size, f);
}

//------------------------------Value------------------------------------------
// Compute the type of the PCTableNode.  If reachable it is a tuple of 
// Control, otherwise the table targets are not reachable
const Type *PCTableNode::Value( PhaseTransform *phase ) const {
  if( phase->type(in(0)) == Type::CONTROL )
    return bottom_type();
  return Type::TOP;             // All paths dead?  Then so are we
}

//------------------------------Ideal------------------------------------------
// Return a node which is more "ideal" than the current node.  Strip out 
// control copies
Node *PCTableNode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  return remove_dead_region(phase,du) ? this : NULL;
}


//=============================================================================
//------------------------------Value------------------------------------------
// Check for being unreachable, or for coming from a Rethrow.  Rethrow's cannot
// have the default "fall_through_index" path.
const Type *CatchNode::Value( PhaseTransform *phase ) const {
  // Unreachable?  Then so are all paths from here.
  if( phase->type(in(0)) == Type::TOP ) return Type::TOP;
  // First assume all paths are reachable
  const Type** f = TypeTuple::fields(_size);
  for( uint i = 0; i < _size; i++ ) f[i] = Type::CONTROL;
  // Identify cases that will always throw an exception
  // () rethrow call
  // () virtual or interface call with NULL receiver
  // () call is a check cast with incompatible arguments
  if( in(1)->is_Proj() ) {
    CallNode *call = in(1)->in(0)->is_Call();
    if( call ) {
      // Rethrows always throw exceptions, never return
      if( call->_entrypoint == OptoRuntime::rethrow_stub() ) {
        f[CatchProjNode::fall_through_index] = Type::TOP;
      } else if( call->req() > TypeFunc::Parms ) {
        const Type *arg0 = phase->type( call->in(TypeFunc::Parms) );
        // Check for null reciever to virtual or interface calls
        if( call->is_CallDynamicJava() && 
            arg0->higher_equal(TypePtr::NULL_PTR) ) { 
          f[CatchProjNode::fall_through_index] = Type::TOP;
        }
        // Check for check-cast with incompatible arguments
        if( call->_entrypoint == OptoRuntime::check_cast_stub() ) {
          const Type *arg1_type = phase->type(call->in(TypeFunc::Parms+1));
          if( arg1_type == Type::TOP ) {
            f[CatchProjNode::fall_through_index] = Type::TOP;
          } else {
            // Now arg0 is a klass and arg1 is an oop pointer.
            const TypeOopPtr *arg1 = arg1_type->isa_oopptr();
            if( arg1 ) {
              const TypeKlassPtr *k0 = arg0->is_klassptr();
              ciKlass* cast_klass =   k0->klass();
              ciKlass*   in_klass = arg1->klass();
              // Check to see if we _must_ throw an exception
              if( !cast_klass->is_interface() && // bail on optimizing interfcs
                  !in_klass->is_interface() &&
                  !in_klass->is_subtype_of(cast_klass) &&
                  !cast_klass->is_subtype_of(in_klass) ) {
                f[CatchProjNode::fall_through_index] = Type::TOP;
              }
            }
          }
        }
      } // End of if not a runtime stub
    } // End of if have call above me
  } // End of slot 1 is not a projection
  return TypeTuple::make(_size, f);
}

//=============================================================================
uint CatchProjNode::hash() const {
  return Node::hash() + _handler_bci;
}


uint CatchProjNode::cmp( const Node &n ) const {
  return ProjNode::cmp(n) &&
    _handler_bci == ((CatchProjNode&)n)._handler_bci;
}


CatchProjNode::CatchProjNode(Node* catchnode, uint proj_no, int handler_bci)
: CProjNode(catchnode, proj_no), _handler_bci(handler_bci) {
  assert(proj_no != fall_through_index || handler_bci < 0, "fall through case must have bci < 0");
}


//------------------------------Identity---------------------------------------
// If only 1 target is possible, choose it if it is the main control
Node *CatchProjNode::Identity( PhaseTransform *phase ) {
  // If my value is control and no other value is, then treat as ID
  const TypeTuple *t = phase->type(_in[0])->is_tuple();
  if( t->_fields[_con] != Type::CONTROL ) return this;
  // If we remove the last CatchProj and elide the Catch/CatchProj, then we
  // also remove any exception table entry.  Thus we must know the call
  // feeding the Catch will not really throw an exception.  This is ok for
  // the main fall-thru control (happens when we know a call can never throw
  // an exception) or for "rethrow", because a further optimnization will
  // yank the rethrow (happens when we inline a function that can throw an
  // exception and the caller has no handler).  Not legal, e.g., for passing
  // a NULL receiver to a v-call, or passing bad types to a slow-check-cast.
  // These cases MUST throw an exception via the runtime system, so the VM
  // will be looking for a table entry.
  Node *proj = in(0)->in(1);    // Expect a proj feeding CatchNode
  CallNode *call;
  if( _con != TypeFunc::Control && // Bail out if not the main control.
      !(proj->is_Proj() &&      // AND NOT a rethrow
        (call = proj->in(0)->is_Call()) &&
        call->_entrypoint == OptoRuntime::rethrow_stub() ) )
    return this;

  // Search for any other path being control
  for( uint i = 0; i < t->_cnt; i++ )
    if( i != _con && t->_fields[i] == Type::CONTROL ) 
      return this;
  // Only my path is possible; I am identity on control to the jump
  return _in[0]->in(0);
}


#ifndef PRODUCT
void CatchProjNode::dump_spec() const { 
  ProjNode::dump_spec();
  tty->print("@bci %d ",_handler_bci);
}
#endif

//=============================================================================
//------------------------------Identity---------------------------------------
// Check for CreateEx being Identity.
Node *CreateExNode::Identity( PhaseTransform *phase ) {
  if( phase->type(in(1)) == Type::TOP ) return in(1);
  if( phase->type(in(0)) == Type::TOP ) return in(0);
  // We only come from CatchProj, unless the CatchProj goes away.
  // If the CatchProj is optimized away, then we just carry the 
  // exception oop through.
  CallNode *call = in(1)->in(0)->is_Call();
  assert( call, "" );

  return ( in(0)->is_CatchProj() && in(0)->in(0)->in(1) == in(1) ) 
    ? this
    : call->in(TypeFunc::Parms);
}

const Type* CreateExNode::bottom_type() const {
  return TypeOopPtr::NOTNULL;
}

//=============================================================================
uint RethrowLocalNode::hash() const { return Node::hash() + (uint)_method + _bci; }
uint RethrowLocalNode::cmp( const Node &n ) const
{ return _method == ((RethrowLocalNode&)n)._method &&
    _bci == ((RethrowLocalNode&)n)._bci; }

#ifndef PRODUCT
void RethrowLocalNode::dump_spec() const { 
  tty->print(" "); 
  _method->print_short_name(tty);
  tty->print("@%d",_bci);
}
#endif


//=============================================================================
#ifndef PRODUCT
void NeverBranchNode::format( PhaseChaitin *pc_ ) const {
  tty->print("%s", Name());
}
#endif

