/*
 * @(#)cfgnode.hpp	1.75 01/02/09 14:45:58
 *
 * Copyright 1991-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
// Portions of code courtesy of Clifford Click

// Optimization - Graph Style

class Matcher;
class Node;
class   RegionNode;
class   TypeNode;
class     PhiNode;
class   GotoNode;
class   MultiNode;
class     IfNode;
class       CheckTypeNode;
class       FastLockNode;
class       FastUnockNode;
class     PCTableNode;
class       JumpNode;
class       CatchNode;
class   ProjNode;
class     CProjNode;
class       IfTrueNode;
class       IfFalseNode;
class       CatchProjNode;
class PhaseIdealLoop;

//------------------------------RegionNode-------------------------------------
// The class of RegionNodes, which can be mapped to basic blocks in the
// program.  Their inputs point to Control sources.  PhiNodes (described
// below) have an input point to a RegionNode.  Merged data inputs to PhiNodes
// correspond 1-to-1 with RegionNode inputs.  The zero input of a PhiNode is
// the RegionNode, and the zero input of the RegionNode is itself.
class RegionNode : public Node {
public:
  RegionNode( uint required ) : Node(required) { set_req(0,this); }
  virtual int Opcode() const;
  virtual int pinned() const { return (const Node*)_in[0] == this; }
  virtual const RegionNode *is_Region() const { return this; }
  virtual int is_block_start() const { return (const Node*)_in[0] == this; }
  virtual bool  is_CFG   () const { return true; }
  virtual bool depends_only_on_test() const { return false; }
  virtual const Type *bottom_type() const { return Type::CONTROL; }
  virtual const Type *Value( PhaseTransform *phase ) const;
  virtual Node *Identity( PhaseTransform *phase );
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
  virtual const RegMask &out_RegMask() const;
  virtual uint ideal_reg() const { return 0; }
};

//------------------------------PhiNode----------------------------------------
// PhiNodes merge values from different Control paths.  Slot 0 points to the
// controlling RegionNode.  Other slots map 1-for-1 with incoming control flow
// paths to the RegionNode.  For speed reasons (to avoid another pass) we 
// can turn PhiNodes into copys in-place by NULL'ing out their RegionNode 
// input in slot 0.
class PhiNode : public TypeNode {
public:
  PhiNode( Node *r, const Type *t ) : TypeNode(t,r->req()) { _in[0]=r; }
  virtual int Opcode() const;
  virtual PhiNode *is_Phi () { return this; }
  virtual int pinned() const { return _in[0] != 0; }
  virtual const Type *Value( PhaseTransform *phase ) const;
  virtual Node *Identity( PhaseTransform *phase );
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
  virtual const RegMask &out_RegMask() const;
  virtual const RegMask &in_RegMask(uint) const;
};

//------------------------------GotoNode---------------------------------------
// GotoNodes perform direct branches.
class GotoNode : public Node {
public:
  GotoNode( Node *control ) : Node(control) {}
  virtual int Opcode() const;
  virtual int pinned() const { return 1; }
  virtual bool  is_CFG() const { return true; }
  virtual const Node *is_block_proj() const { return this; }
  virtual bool depends_only_on_test() const { return false; }
  virtual const Type *bottom_type() const { return Type::CONTROL; }
  virtual const Type *Value( PhaseTransform *phase ) const;
  virtual Node *Identity( PhaseTransform *phase );
  virtual const RegMask &out_RegMask() const;
  virtual uint  is_Goto() const { return 1; }
  virtual uint ideal_reg() const { return 0; }
};

//------------------------------CProjNode--------------------------------------
// control projection for node that produces multiple control-flow paths
class CProjNode : public ProjNode {
public:
  CProjNode( Node *ctrl, uint idx ) : ProjNode(ctrl,idx) {}
  virtual int Opcode() const;
  virtual bool  is_CFG() const { return true; }
  virtual bool depends_only_on_test() const { return false; }
  virtual const Node *is_block_proj() const { return _in[0]; }
  virtual const RegMask &out_RegMask() const;
  virtual uint ideal_reg() const { return 0; }
};

//------------------------------IfNode-----------------------------------------
// Output selected Control, based on a boolean test
class IfNode : public MultiNode {
  // Size is bigger to hold the probability field.  However, _prob does not
  // change the semantics so it does not appear in the hash & cmp functions.
  virtual uint size_of() const { return sizeof(*this); }
public:
  float _prob;                  // Probability of true path being taken.
  float _cnt;                   // Frequency counter
  IfNode( Node *control, Node *bool, float p, float cnt ) : MultiNode(2), _prob(p), _cnt(cnt) { set_req(0,control); set_req(1,bool); }
  virtual int Opcode() const;
  virtual int pinned() const { return 1; }
  virtual const Type *bottom_type() const { return TypeTuple::IFBOTH; }
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
  virtual const Type *Value( PhaseTransform *phase ) const;
  virtual const RegMask &out_RegMask() const;
  virtual uint ideal_reg() const { return NotAMachineReg; }
  virtual IfNode *is_If() { return this; }
  void dominated_by( Node *prev_dom, PhaseIterGVN *igvn );
#ifndef PRODUCT
  virtual void dump_spec() const;
#endif
};

class IfTrueNode : public CProjNode {
public:
  IfTrueNode( IfNode *ifnode ) : CProjNode(ifnode,1) {}
  virtual int Opcode() const;
  virtual Node *Identity( PhaseTransform *phase );  
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
};

class IfFalseNode : public CProjNode {
public:
  IfFalseNode( IfNode *ifnode ) : CProjNode(ifnode,0) {}
  virtual int Opcode() const;
  virtual Node *Identity( PhaseTransform *phase );  
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
};


//----------------------CheckTypeNode------------------------------------------
// Check that 'oop' is a subclass of 'klass_con'. Uses: fast-path/slow-path
// code for do_checkcast, do_instanceof, array-store_check, and
// inline_arraycopy.
class CheckTypeNode: public IfNode {
public:
  CheckTypeNode( Node *ctrl, Node *oop, Node *klass_con, float p, float cnt ) : IfNode(ctrl,oop,p,cnt) { add_req(klass_con); }
  virtual int Opcode() const;
  virtual const Type *Value( PhaseTransform *phase ) const;
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
};

//------------------------------PCTableNode------------------------------------
// Build an indirect branch table.  Given a control and a table index,
// control is passed to the Projection matching the table index.  Used to
// implement switch statements and exception-handling capabilities.
// Undefined behavior if passed-in index is not inside the table.
class PCTableNode : public MultiNode {
  virtual uint hash() const;    // Target count; table size
  virtual uint cmp( const Node &n ) const;
  virtual uint size_of() const { return sizeof(*this); }

public:
  const uint _size;             // Number of targets

  PCTableNode( Node *ctrl, Node *idx, uint size ) : MultiNode(2), _size(size) {
    set_req(0, ctrl);
    set_req(1, idx);
  }
  virtual int Opcode() const;
  virtual const Type *Value( PhaseTransform *phase ) const;
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
  virtual const Type *bottom_type() const;
  virtual int pinned() const { return 1; }
  virtual uint ideal_reg() const { return NotAMachineReg; }
  virtual const PCTableNode *is_PCTable() const { return this; }
};

//------------------------------JumpNode---------------------------------------
// Indirect branch.  Uses PCTable above to implement a switch statement.
// It emits as a table load and local branch.  
class JumpNode : public PCTableNode {
public:
  JumpNode( Node *ctrl, Node *idx, uint size) : PCTableNode(ctrl,idx,size){}
  virtual int Opcode() const;
};

//------------------------------CatchNode--------------------------------------
// Helper node to fork exceptions.  "Catch" catches any exceptions thrown by
// a just-prior call.  Looks like a PCTableNode but emits no code - just the
// table.  The table lookup and branch is implemented by RethrowNode.
class CatchNode : public PCTableNode {
public:
  CatchNode( Node *ctrl, Node *idx, uint size ) : PCTableNode(ctrl,idx,size){};
  virtual int Opcode() const;
  virtual const Type *Value( PhaseTransform *phase ) const;
  virtual const CatchNode *is_Catch() const { return this; }
};

// CatchProjNode controls which exception handler is targetted after a call.
// It is passed in the bci of the target handler, or no_handler_bci in case
// the projection doesn't lead to an exception handler.
class CatchProjNode : public CProjNode {
  virtual uint hash() const;
  virtual uint cmp( const Node &n ) const;
  virtual uint size_of() const { return sizeof(*this); }

private:
  const int _handler_bci;

public:
  enum {
    fall_through_index =  0,      // the fall through projection index
    no_handler_bci     = -1       // the bci for fall through or rethrow projs
  };

  CatchProjNode(Node* catchnode, uint proj_no, int handler_bci);
  virtual int Opcode() const;
  virtual Node *Identity( PhaseTransform *phase );
  virtual const Type *bottom_type() const { return Type::CONTROL; }
  virtual const CatchProjNode *is_CatchProj() const { return this; }
  int  handler_bci() const        { return _handler_bci; }
  bool is_handler_proj() const    { return _handler_bci >= 0; }
#ifndef PRODUCT
  virtual void dump_spec() const;
#endif
};


//---------------------------------CreateExNode--------------------------------
// Helper node to create the exception coming back from a call
class CreateExNode : public Node {
public:
  CreateExNode(Node *control, Node *idx) : Node(control,idx) {}
  virtual int Opcode() const;
  virtual Node *Identity( PhaseTransform *phase );
  virtual int pinned() const { return 1; }
  virtual const Type *bottom_type() const;
  uint match_edge(uint idx) const { return 0; }
  virtual uint ideal_reg() const { return Op_RegP; }
  virtual const CreateExNode *is_CreateEx() const { return this; }
};

//---------------------------------RethrowLocalNode----------------------------
// 
class RethrowLocalNode : public Node {
  virtual uint hash() const;    // Target count; table size
  virtual uint cmp( const Node &n ) const;
  virtual uint size_of() const { return sizeof(*this); }

public:
  const methodOop _method;      // Call site where exception is being caught
  const uint _bci;

  RethrowLocalNode( Node *ex_oop, methodOop method, uint bci ) : Node(0,ex_oop), _method(method), _bci(bci) {}
  virtual int Opcode() const;
  virtual bool  is_CFG() const { return true; }
  virtual bool depends_only_on_test() const { return false; }
  virtual const Type *bottom_type() const { return TypeInt::POS; }
  virtual uint ideal_reg() const { return Op_RegI; }
#ifndef PRODUCT
  virtual void dump_spec() const;
#endif
};

//------------------------------NeverBranchNode-------------------------------
// The never-taken branch.  Used to give the appearance of exiting infinite
// loops to those algorithms that like all paths to be reachable.  Encodes
// empty.
class NeverBranchNode : public MultiNode {
public:
  NeverBranchNode( Node *ctrl ) : MultiNode(1) { set_req(0,ctrl); }
  virtual int Opcode() const;
  virtual int pinned() const { return 1; };
  virtual const Type *bottom_type() const { return TypeTuple::IFBOTH; }

  virtual void emit(CodeBuffer &cbuf, PhaseChaitin *pc_) const { }
  virtual uint size(PhaseChaitin *pc_) const { return 0; }
#ifndef PRODUCT
  virtual void format( PhaseChaitin * ) const;
#endif
};

