/*
 * @(#)chaitin.cpp	1.23 10:53:15
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
#include "incls/_precompiled.incl"
#include "incls/_chaitin.cpp.incl"

//=============================================================================
static const int NodeRegsOverflowSize = 10;

#ifndef PRODUCT
void LRG::dump( ) const {
  tty->print("%d ",num_regs());
  _mask.dump();
  if( _msize_valid ) {
    if( mask_size() == compute_mask_size() ) tty->print(", #%d ",_mask_size);
    else tty->print(", #!!!_%d_vs_%d ",_mask_size,_mask.Size());
  } else {
    tty->print(", #?(%d) ",_mask.Size());
  }

  tty->print("EffDeg: ");
  if( _degree_valid ) tty->print( "%d ", _eff_degree );
  else tty->print("? ");

  if( _def == (Node*)-1 ) tty->print("MultiDef ");
  else if( _def == 0 ) tty->print("Dead ");
  else tty->print("Def: N%d ",_def->_idx);

  tty->print("Cost:%4.2g Area:%4.2g Score:%4.2g ",_cost,_area, score());
  // Flags
  if( _is_oop ) tty->print("Oop ");
  if( _is_float ) tty->print("Float ");
  if( _was_spilled1 ) tty->print("Spilled ");
  if( _was_spilled2 ) tty->print("Spilled2 ");
  if( _direct_conflict ) tty->print("Direct_conflict ");
  if( _fat_proj ) tty->print("Fat ");
  if( _was_lo ) tty->print("Lo ");
  if( _has_copy ) tty->print("Copy ");
  if( _at_risk ) tty->print("Risk ");

  if( _must_spill ) tty->print("Must_spill ");
  if( _is_bound ) tty->print("Bound ");
  if( _msize_valid ) {
    if( _degree_valid && lo_degree() ) tty->print("Trivial ");
  }

  tty->cr();
}
#endif

//------------------------------score------------------------------------------
// Compute score from cost and area.  Low score is best to spill.
static double raw_score( double cost, double area ) {
  return cost - (area*RegisterCostAreaRatio) * 1.52588e-5;
}

double LRG::score() const {
  // Scale _area by RegisterCostAreaRatio/64K then subtract from cost.
  // Bigger area lowers score, encourages spilling this live range.
  // Bigger cost raise score, prevents spilling this live range.
  // (Note: 1/65536 is the magic constant below; I dont trust the C optimizer
  // to turn a divide by a constant into a multiply by the reciprical).
  double score = raw_score( _cost, _area);

  // Account for area.  Basically, LRGs covering large areas are better 
  // to spill because more other LRGs get freed up.
  if( _area == 0.0 )            // No area?  Then no progress to spill
    return 1e35;

  if( _was_spilled2 )           // If spilled once before, we are unlikely
    return score + 1e30;	// to make progress again.

  if( _cost >= _area*3.0 )      // Tiny area relative to cost
    return score + 1e17;        // Probably no progress to spill

  if( (_cost+_cost) >= _area*3.0 ) // Small area relative to cost
    return score + 1e10;        // Likely no progress to spill

  return score;
}

//------------------------------LRG_List---------------------------------------
LRG_List::LRG_List( uint max ) : _cnt(max), _max(max), _lidxs(NEW_RESOURCE_ARRAY(uint16,max)) {
  memset( _lidxs, 0, sizeof(uint16)*max );
}

void LRG_List::extend( uint16 nidx, uint16 lidx ) {
  _nesting.check();
  if( nidx >= _max ) {
    uint size = 16; 
    while( size <= nidx ) size <<=1;
    _lidxs = REALLOC_RESOURCE_ARRAY( uint16, _lidxs, _max, size );
    _max = size;
  }
  while( _cnt <= nidx ) 
    _lidxs[_cnt++] = 0;
  _lidxs[nidx] = lidx;
}

//=============================================================================
extern "C" {
typedef int (*qsort_func)(const void *, const void *);

static int cmp_freq( const Block **bp1, const Block **bp2 ) {
  const Block *b1 = *bp1;
  const Block *b2 = *bp2;
  double f1 = b1->_freq;
  double f2 = b2->_freq;

  // Bump importance if block contains only copies.

  // Sort by frequency
  return (f1 < f2) ? 1 : ((f1 == f2) ? 0 : -1);
}

}

//------------------------------Chaitin----------------------------------------
PhaseChaitin::PhaseChaitin( uint unique, PhaseCFG &cfg, Matcher &matcher, bool save_argument_registers ) : Phase(Register_Allocation), _names(unique), _uf_map(unique), _cfg(cfg), _matcher(matcher), _maxlrg(0), _live(0), _node_oops(Thread::current()->resource_area()), _spilled_once(Thread::current()->resource_area()), _spilled_twice(Thread::current()->resource_area()), _lo_degree(0), _lo_stk_degree(0), _hi_degree(0), _simplified(0), _free_list(&C->_comp_arena), _save_argument_registers(save_argument_registers), _node_regs(0), _oldphi(unique) {
  uint i;
  // Build a list of basic blocks, sorted by frequency
  _blks = NEW_RESOURCE_ARRAY( Block *, _cfg._num_blocks );
  for( i = 0; i < _cfg._num_blocks; i++ ) {
    _blks[i] = _cfg._blocks[i];
  }
  qsort( _blks, _cfg._num_blocks, sizeof(Block*), (qsort_func)cmp_freq );
}

void PhaseChaitin::Register_Allocate() {

  // Above the OLD FP (and in registers) are the incoming arguments.  Stack
  // slots in this area are called "arg_slots".  Above the NEW FP (and in
  // registers) is the outgoing argument area; above that is the spill/temp
  // area.  These are all "frame_slots".  Arg_slots start at the zero
  // stack_slots and count up to the known arg_size.  Frame_slots start at
  // the stack_slot #arg_size and go up.  After allocation I map stack
  // slots to actual offsets.  Stack-slots in the arg_slot area are biased
  // by the frame_size; stack-slots in the frame_slot area are biased by 0.

  _trip_cnt = 0;
  _alternate = 0;
  _matcher._allocation_started = true;

  ResourceArea live_arena;      // Arena for liveness & IFG info
  ResourceMark rm(&live_arena);

  // Need live-ness for the IFG; need the IFG for coalescing.  If the
  // liveness is JUST for coalescing, then I can get some milage by renaming
  // all copy-related live ranges low and then using the max copy-related
  // live range as a cut-off for LIVE and the IFG.  In other words, I can
  // build a subset of LIVE and IFG just for copies.
  PhaseLive live(_cfg,_names,&live_arena);

  // Need IFG for coalescing and coloring
  PhaseIFG ifg( &live_arena );
  _ifg = &ifg;

  if( C->_unique > _names.Size() ) _names.extend( C->_unique-1, 0);

  // Come out of SSA world to the Named world.  Assign (virtual) registers to
  // Nodes.  Use the same register for all inputs and the output of PhiNodes
  // - effectively ending SSA form.  This requires either coalescing live
  // ranges or inserting copies.  For the moment, we insert "virtual copies"
  // - we pretend there is a copy prior to each Phi in predecessor blocks.
  // We will attempt to coalesce such "virtual copies" before we manifest
  // them for real.
  de_ssa();


  _live = NULL;                 // Mark live as being not available
  rm.reset_to_mark();           // Reclaim working storage
  IndexSet::reset_memory(&live_arena);
  ifg.init(_maxlrg);            // Empty IFG
  gather_lrg_masks( false );    // Collect LRG masks
  live.compute( _maxlrg );	// Compute liveness
  _live = &live;                // Mark LIVE as being available

  // Base pointers are currently "used" by instructions which define new
  // derived pointers.  This makes base pointers live up to the where the
  // derived pointer is made, but not beyond.  Really, they need to be live
  // across any GC point where the derived value is live.  So this code looks
  // at all the GC points, and "stretches" the live range of any base pointer
  // to the GC point.
  if( stretch_base_pointer_live_ranges(&live_arena) ) {
    // Since some live range stretched, I need to recompute live
    _live = NULL;
    rm.reset_to_mark();         // Reclaim working storage
    IndexSet::reset_memory(&live_arena);
    ifg.init(_maxlrg);
    gather_lrg_masks( false );
    live.compute( _maxlrg );
    _live = &live;
  }
  // Create the interference graph using virtual copies
  build_ifg_virtual( );  // Include stack slots this time

  // Aggressive (but pessimistic) copy coalescing.
  // This pass works on virtual copies.  Any virtual copies which are not
  // coalesced get manifested as actual copies
  { 
    // The IFG is/was triangular.  I am 'squaring it up' so Union can run 
    // faster.  Union requires a 'for all' operation which is slow on the
    // triangular adjacency matrix (quick reminder: the IFG is 'sparse' -
    // meaning I can visit all the Nodes neighbors less than a Node in time
    // O(# of neighbors), but I have to visit all the Nodes greater than a
    // given Node and search them for an instance, i.e., time O(#MaxLRG)).
    _ifg->SquareUp();

    PhaseAggressiveCoalesce coalesce( *this ); 
    coalesce.coalesce_driver( );
    // Insert un-coalesced copies.  Visit all Phis.  Where inputs to a Phi do
    // not match the Phi itself, insert a copy.
    coalesce.insert_copies(_matcher);
  }

  // After aggressive coalesce, attempt a first cut at coloring.
  // To color, we need the IFG and for that we need LIVE.
  _live = NULL;
  rm.reset_to_mark();           // Reclaim working storage
  IndexSet::reset_memory(&live_arena);
  ifg.init(_maxlrg);
  gather_lrg_masks( true );
  live.compute( _maxlrg );
  _live = &live;

  // Build physical interference graph
  uint must_spill = build_ifg_physical( &live_arena );
  // If we have a guaranteed spill, might as well spill now
  if( must_spill ) {  
    if( !_maxlrg ) return;
    // Bail out if unique gets too large (ie - unique > 64000)
    if( C->_unique + 10*must_spill > 65000 ) {
      C->_root = NULL;
      return;
    }
    _maxlrg = Split( _maxlrg );        // Split spilling LRG everywhere
    // Bail out if unique gets too large (ie - unique > 63000)
    // or we failed to split
    if( !_maxlrg || C->_unique > 63000 ) {
      C->_root = NULL;
      return;
    }
#ifdef ASSERT
    if( VerifyOpto ) {
      _cfg.verify();
      verify_base_ptrs(&live_arena);
    }
#endif

    compact();                  // Compact LRGs; return new lower max lrg

    _live = NULL;
    rm.reset_to_mark();         // Reclaim working storage
    IndexSet::reset_memory(&live_arena);
    ifg.init(_maxlrg);          // Build a new interference graph
    gather_lrg_masks( true );   // Collect intersect mask
    live.compute( _maxlrg );	// Compute LIVE
    _live = &live;
    build_ifg_physical( &live_arena );
    _ifg->SquareUp();
    _ifg->Compute_Effective_Degree();
    // Only do conservative coalescing if requested
    if( OptoCoalesce ) {
      // Conservative (and pessimistic) copy coalescing of those spills
      PhaseConservativeCoalesce coalesce( *this );
      // If max live ranges greater than cutoff, don't color the stack.
      // This cutoff can be larger than below since it is only done once.
      coalesce.coalesce_driver( );
    }
    compress_uf_map_for_nodes();

#ifdef ASSERT
    if( VerifyOpto ) _ifg->verify(this);
#endif
  } else {
    ifg.SquareUp();
    ifg.Compute_Effective_Degree();
#ifdef ASSERT
    set_was_low();
#endif
  }

  // Prepare for Simplify & Select
  cache_lrg_info();           // Count degree of LRGs

  // Simplify the InterFerence Graph by removing LRGs of low degree.
  // LRGs of low degree are trivially colorable.
  Simplify(); 

  // Select colors by re-inserting LRGs back into the IFG in reverse order.
  // Return whether or not something spills.
  uint spills = Select( ); 

  // If we spill, split and recycle the entire thing
  while( spills ) {
    if( _trip_cnt++ > 15 ) {
      assert( true, "failed spill-split-recycle sanity check" );
      prepare_to_bailout("failed spill-split-recycle sanity check");
      return;
    }
  
    if( !_maxlrg ) return;
    _maxlrg = Split( _maxlrg );        // Split spilling LRG everywhere
    // Bail out if unique gets too large (ie - unique > 63000)
    if( C->_unique > 63000 ) {
      C->_root = NULL;
      return;
    }
#ifdef ASSERT
    if( VerifyOpto ) {
      _cfg.verify();
      verify_base_ptrs(&live_arena);
    }
#endif

    compact();                  // Compact LRGs; return new lower max lrg

    // Nuke the live-ness and interference graph and LiveRanGe info
    _live = NULL;
    rm.reset_to_mark();         // Reclaim working storage
    IndexSet::reset_memory(&live_arena);
    ifg.init(_maxlrg);

    // Create LiveRanGe array. 
    // Intersect register masks for all USEs and DEFs
    gather_lrg_masks( true );
    live.compute( _maxlrg );
    _live = &live;
    must_spill = build_ifg_physical( &live_arena );
    _ifg->SquareUp();
    _ifg->Compute_Effective_Degree();

    // Only do conservative coalescing if requested
    if( OptoCoalesce ) {
      // Conservative (and pessimistic) copy coalescing
      PhaseConservativeCoalesce coalesce( *this );
      // Check for few live ranges determines how aggressive coalesce is.
      coalesce.coalesce_driver( );
    }
    compress_uf_map_for_nodes();
#ifdef ASSERT
    if( VerifyOpto ) _ifg->verify(this);
#endif

    cache_lrg_info();           // Count degree of LRGs

    // Simplify the InterFerence Graph by removing LRGs of low degree.
    // LRGs of low degree are trivially colorable.
    Simplify(); 

    // Select colors by re-inserting LRGs back into the IFG in reverse order.
    // Return whether or not something spills.
    spills = Select( );
  }
  _allocator_attempts += _trip_cnt;
  ++_allocator_successes;

  // Peephole remove copies
  post_allocate_copy_removal();

  // Collect GC mask info - where are all the OOPs?
  collect_gc_masks(_save_argument_registers);
  if( C->_root == NULL ) { return; }   // Bailout from compilation

  // Convert spill-copies to loads & stores
  bool fixed_spills = fixup_spills();

  // Bail out if unique gets too large (ie - unique > 63000)
  if( !fixed_spills ) {
    prepare_to_bailout("Spilling requires too many nodes");
    return;
  }

  // Move important info out of the live_arena to longer lasting storage.
  _node_regs_max_index = _names.Size() + (_names.Size() >> 4) + NodeRegsOverflowSize;
  _node_regs = NEW_RESOURCE_ARRAY( uint16, _node_regs_max_index );
  for( uint i=0; i < _names.Size(); i++ ) {
    if( _names[i] ) {           // Live range associated with Node?
      LRG &lrg = lrgs( _names[i] );
      // Live ranges record the highest register in their mask.
      // We want the low register for the AD file writer's convenience.
      _node_regs[i] = OptoReg::add(lrg.reg(),1-lrg.num_regs());
      if( lrg._is_oop ) _node_oops.set(i);
    } else {
      _node_regs[i] = OptoReg::Bad;
    }
  }
  for( uint i2 = _names.Size(); i2 < _node_regs_max_index; ++i2 ) {
    _node_regs[i2] = OptoReg::Bad;
  }

  // Done!  
  _live = NULL;
  _ifg = NULL;
}

// Set the register associated with the Node
void PhaseChaitin::set_register( const Node *n, OptoReg::Name reg ) {
  if( n->_idx >= _node_regs_max_index ) {
    // grow the resource array
    uint new_size = _node_regs_max_index + (_node_regs_max_index >> 4) + NodeRegsOverflowSize;
    _node_regs = REALLOC_RESOURCE_ARRAY( uint16, _node_regs, _node_regs_max_index, new_size );
    // Initialize data
    for( uint i = _node_regs_max_index; i < new_size; ++i ) {
      _node_regs[i] = OptoReg::Bad;
    }
    _node_regs_max_index = new_size;
  }
  _node_regs[n->_idx] = reg;
}

//------------------------------~Chaitin---------------------------------------
PhaseChaitin::~PhaseChaitin( ) {
}


//------------------------------record_bailout---------------------------------
void PhaseChaitin::prepare_to_bailout(const char *printed_error_msg) const {
  if( PrintOptoBailouts || PrintOpto || PrintCompilation ) {
    tty->print_cr("Bailout: %s", printed_error_msg);
  }
  C->_root = NULL;
  return;
}

//------------------------------de_ssa-----------------------------------------
void PhaseChaitin::de_ssa() {
  // Set initial Names for all Nodes.  Most Nodes get the virtual register
  // number.  A few get the ZERO live range number.  These do not
  // get allocated, but instead rely on correct scheduling to ensure that
  // only one instance is simultaneously live at a time.
  uint lr_counter = 1;
  for( uint i = 0; i < _cfg._num_blocks; i++ ) {
    Block *b = _cfg._blocks[i];
    uint cnt = b->_nodes.size();

    // Handle all the normal Nodes in the block
    for( uint j = 0; j < cnt; j++ ) {
      Node *n = b->_nodes[j];
      // Pre-color to the zero live range, or pick virtual register
      const RegMask &rm = n->out_RegMask();
      _names.map( n->_idx, rm.is_NotEmpty() ? lr_counter++ : 0 );
    }
  }
  // Reset the Union-Find mapping to be identity
  reset_uf_map(lr_counter);
}


//------------------------------gather_lrg_masks-------------------------------
// Gather LiveRanGe information, including register masks.  Modification of
// cisc spillable in_RegMasks should not be done before AggressiveCoalesce.
void PhaseChaitin::gather_lrg_masks( bool after_aggressive ) {

  // Nail down the frame pointer live range
  uint fp_lrg = n2lidx(_cfg._root->in(1)->in(TypeFunc::FramePtr));
  lrgs(fp_lrg)._cost += 1e12;   // Cost is infinite

  // Reset all live-ranges-per-inner-loops
  for( InnerLoop *il = _cfg._loops; il; il = il->_next ) {
    il->_ud_lrgs.Clear();
    il->_isolate_lrgs.Clear();
  }

  // For all blocks
  for( uint i = 0; i < _cfg._num_blocks; i++ ) {
    Block *b = _cfg._blocks[i];

    // For all instructions
    for( uint j = 1; j < b->_nodes.size(); j++ ) {
      Node *n = b->_nodes[j];
      uint input_edge_start =1; // Skip control most nodes
      MachNode *mach = n->is_Mach();
      if( mach ) input_edge_start = mach->oper_input_base();
      uint idx = n->is_Copy();

      // Get virtual register number, same as LiveRanGe index
      uint vreg = n2lidx(n);
      LRG &lrg = lrgs(vreg);
      if( vreg ) {		// No vreg means un-allocable (e.g. memory)

	// Add def'd live range to inner loop collection
	if( b->_inner_loop ) {
	  // Don't count useless Phis as def's
	  if( !(b->_inner_loop->_head == b && n->is_Phi() && n->in(2) == n) )
	    b->_inner_loop->_ud_lrgs.set(vreg);
	}

        // Collect has-copy bit
        if( idx ) {
          lrg._has_copy = 1;
          uint clidx = n2lidx(n->in(idx));
          LRG &copy_src = lrgs(clidx);
          copy_src._has_copy = 1;
        }

        // Check for float-vs-int live range (used in register-pressure
        // calculations)
        const Type *n_type = n->bottom_type();
        if( n_type->is_floatingpoint() )
          lrg._is_float = 1;

        // Check for twice prior spilling.  Once prior spilling might have
        // spilled 'soft', 2nd prior spill should have spilled 'hard' and
        // further spilling is unlikely to make progress.
        if( _spilled_once.test(n->_idx) ) {
          lrg._was_spilled1 = 1;
	  if( _spilled_twice.test(n->_idx) ) 
            lrg._was_spilled2 = 1;
        }
        
        // Check for a single def LRG; these can spill nicely
        // via rematerialization.  Flag as NULL for no def found
        // yet, or 'n' for single def or -1 for many defs.
        lrg._def = lrg._def ? (Node *)-1 : n;

        // Limit result register mask to acceptable registers
        const RegMask &rm = n->out_RegMask();
        lrg.AND( rm );
        // Check for bound register masks
        const RegMask &lrgmask = lrg.mask();
        if( lrgmask.is_bound1() || lrgmask.is_bound2() )
          lrg._is_bound = 1;

        // Check for maximum frequency value
        if( lrg._maxfreq < b->_freq )
          lrg._maxfreq = b->_freq;

        int ireg = n->ideal_reg();
        assert( !n->bottom_type()->isa_oop_ptr() || ireg == Op_RegP,
                "oops must be in Op_RegP's" );
        // Check for oop-iness, or long/double
        // Check for multi-kill projection
        switch( ireg ) {
        case MachProjNode::fat_proj:
          // Fat projections have size equal to number of registers killed
          lrg.set_num_regs(rm.Size());
          lrg._fat_proj = 1;
          lrg._is_bound = 1;
          break;
        case Op_RegP:
          lrg.set_num_regs(1);  // Size is 1 stack word 
          if( n_type->isa_oop_ptr() ) {
            lrg._is_oop = 1;
          }
          break;
        case Op_RegL:           // Check for long or double
        case Op_RegD:
          lrg.set_num_regs(2);
          // If this def of a double forces a mis-aligned double, 
          // flag as '_fat_proj' - really flag as allowing misalignment
          // AND changes how we count interferences.  A mis-aligned
          // double can interfere with TWO aligned pairs, or effectively
          // FOUR registers!
          if( rm.is_bound2() && !rm.is_aligned_Pairs() )
            lrg._fat_proj = 1;
          break;
        case Op_RegF:
        case Op_RegI:
	case Op_RegFlags:
        case 0:                 // not an ideal register
          lrg.set_num_regs(1);
          break;
        default:
          ShouldNotReachHere();
        }
      }

      // Now do the same for inputs
      uint cnt = n->req();
      // Setup for CISC SPILLING
      uint inp = AdlcVMDeps::Not_cisc_spillable;
      if( UseCISCSpill && after_aggressive ) {
        inp = n->cisc_operand();
	if( inp != AdlcVMDeps::Not_cisc_spillable )
	  // Convert operand number to edge index number
	  inp = n->is_Mach()->operand_index(inp);
      }
      // Prepare register mask for each input
      for( uint k = input_edge_start; k < cnt; k++ ) {
        uint vreg = n2lidx(n->in(k));
        if( !vreg ) continue;

	// Add use'd live range to inner loop collection
	if( b->_inner_loop ) {
	  if( !(b->_inner_loop->_head == b && n->is_Phi() && n->in(2) == n) )
	    b->_inner_loop->_ud_lrgs.set(vreg);
	}

        // If this instruction is CISC Spillable, add the flags
        // bit to its appropriate input
        if( UseCISCSpill && after_aggressive && inp == k ) {
#ifndef PRODUCT
          if( TraceCISCSpill ) {
            tty->print("  use_cisc_RegMask: ");
            n->dump();
          }
#endif
          n->is_Mach()->use_cisc_RegMask();
        }

        LRG &lrg = lrgs(vreg);
        // // Testing for floating point code shape
        // Node *test = n->in(k);
        // if( test->is_Mach() ) {
        //   MachNode *m = test->is_Mach();
        //   int  op = m->ideal_Opcode();
        //   if (n->is_Call() && (op == Op_AddF || op == Op_MulF) ) {
        //     int zzz = 1;
        //   }
        // }

        // Limit result register mask to acceptable registers.
        // Do not limit registers from uncommon uses before 
        // AggressiveCoalesce.  This effectively pre-virtual-splits 
        // around uncommon uses of common defs.
        const RegMask &rm = n->in_RegMask(k);
        if( !after_aggressive && 
          _cfg._bbs[n->in(k)->_idx]->_freq > 1000*b->_freq ) {
          // Since we are BEFORE aggressive coalesce, leave the register
          // mask untrimmed by the call.  This encourages more coalescing.
          // Later, AFTER aggressive, this live range will have to spill
          // but the spiller handles slow-path calls very nicely.
        } else {
          lrg.AND( rm );
        }
        // Check for bound register masks
        const RegMask &lrgmask = lrg.mask();
        if( lrgmask.is_bound1() || lrgmask.is_bound2() )
          lrg._is_bound = 1;
        // If this use of a double forces a mis-aligned double, 
        // flag as '_fat_proj' - really flag as allowing misalignment
        // AND changes how we count interferences.  A mis-aligned
        // double can interfere with TWO aligned pairs, or effectively
        // FOUR registers!
        if( lrg.num_regs() == 2 && !lrg._fat_proj && 
            rm.is_bound2() && !rm.is_aligned_Pairs() )
          lrg._fat_proj = 1;

        // Check for maximum frequency value
        if( lrg._maxfreq < b->_freq )
          lrg._maxfreq = b->_freq;
            
      } // End for all allocated inputs
    } // end for all instructions
  } // end for all blocks

  // Final per-liverange setup
  for( uint i2=0; i2<_maxlrg; i2++ ) {
    LRG &lrg = lrgs(i2);
    if( lrg.num_regs() == 2 && !lrg._fat_proj )
      lrg.ClearToPairs();
    lrg.compute_set_mask_size();
    if( lrg.not_free() ) {      // Handle case where we lose from the start
      lrg.set_reg(OptoReg::Name(LRG::SPILL_REG));
      lrg._direct_conflict = 1;
    }
    lrg.set_degree(0);          // no neighbors in IFG yet
  }  
}

//------------------------------set_was_low------------------------------------
// Set the was-lo-degree bit.  Conservative coalescing should not change the
// colorability of the graph.  If any live range was of low-degree before
// coalescing, it should Simplify.  This call sets the was-lo-degree bit.
// The bit is checked in Simplify.
void PhaseChaitin::set_was_low() {
#ifdef ASSERT
  for( uint i = 1; i < _maxlrg; i++ ) {
    int size = lrgs(i).num_regs();
    uint old_was_lo = lrgs(i)._was_lo;
    lrgs(i)._was_lo = 0;
    if( lrgs(i).lo_degree() ) {
      lrgs(i)._was_lo = 1;      // Trivially of low degree
    } else {                    // Else check the Brigg's assertion
      // Brigg's observation is that the lo-degree neighbors of a 
      // hi-degree live range will not interfere with the color choices
      // of said hi-degree live range.  The Simplify reverse-stack-coloring
      // order takes care of the details.  Hence you do not have to count
      // low-degree neighbors when determining if this guy colors.
      int briggs_degree = 0;
      IndexSet *s = _ifg->neighbors(i);
      IndexSetIterator elements(s);
      uint lidx;
      while((lidx = elements.next()) != 0) {
        if( !lrgs(lidx).lo_degree() ) 
          briggs_degree += max(size,lrgs(lidx).num_regs());
      }
      if( briggs_degree < lrgs(i).degrees_of_freedom() ) 
        lrgs(i)._was_lo = 1;    // Low degree via the briggs assertion
    }
    assert(old_was_lo <= lrgs(i)._was_lo, "_was_lo may not decrease");
  }
#endif
}

#define REGISTER_CONSTRAINED 16

//------------------------------cache_lrg_info---------------------------------
// Compute cost/area ratio, in case we spill.  Build the lo-degree list.
void PhaseChaitin::cache_lrg_info( ) {

  for( uint i = 1; i < _maxlrg; i++ ) {
    LRG &lrg = lrgs(i);

    // Check for being of low degree: means we can be trivially colored.
    // Low degree, dead or must-spill guys just get to simplify right away
    if( lrg.lo_degree() ||
       !lrg.alive() ||
        lrg._must_spill ) {
      // Split low degree list into those guys that must get a
      // register and those that can go to register or stack.
      // The idea is LRGs that can go register or stack color first when
      // they have a good chance of getting a register.  The register-only
      // lo-degree live ranges always get a register.
      OptoReg::Name hi_reg = lrg.mask().find_last_elem();
      if( hi_reg >= SharedInfo::stack0 ) { // Can go to stack?
        lrg._next = _lo_stk_degree;
        _lo_stk_degree = i;
      } else {
        lrg._next = _lo_degree;
        _lo_degree = i;
      }
    } else {                    // Else high degree
      lrgs(_hi_degree)._prev = i;
      lrg._next = _hi_degree;
      lrg._prev = 0;
      _hi_degree = i;
    }
  }
}

//------------------------------Pre-Simplify-----------------------------------
// Simplify the IFG by removing LRGs of low degree that have NO copies
void PhaseChaitin::Pre_Simplify( ) {

  // Warm up the lo-degree no-copy list
  int lo_no_copy = 0;
  for( uint i = 1; i < _maxlrg; i++ ) {
    if( (lrgs(i).lo_degree() && !lrgs(i)._has_copy) ||
        !lrgs(i).alive() ||
        lrgs(i)._must_spill ) {
      lrgs(i)._next = lo_no_copy;
      lo_no_copy = i;
    }
  }

  while( lo_no_copy ) {
    uint lo = lo_no_copy;
    lo_no_copy = lrgs(lo)._next; 
    int size = lrgs(lo).num_regs();

    // Put the simplified guy on the simplified list.
    lrgs(lo)._next = _simplified;
    _simplified = lo;
    
    // Yank this guy from the IFG.  
    IndexSet *adj = _ifg->remove_node( lo );
      
    // If any neighbors' degrees fall below their number of
    // allowed registers, then put that neighbor on the low degree
    // list.  Note that 'degree' can only fall and 'numregs' is
    // unchanged by this action.  Thus the two are equal at most once,
    // so LRGs hit the lo-degree worklists at most once.
    IndexSetIterator elements(adj);
    uint neighbor;
    while ((neighbor = elements.next()) != 0) {
      LRG *n = &lrgs(neighbor);
      assert( _ifg->effective_degree(neighbor) == n->degree(), "" );

      // Check for just becoming of-low-degree
      if( n->just_lo_degree() && !n->_has_copy ) {
        assert(!(*_ifg->_yanked)[neighbor],"Cannot move to lo degree twice");
        // Put on lo-degree list
        n->_next = lo_no_copy; 
        lo_no_copy = neighbor;
      } 
    }
  } // End of while lo-degree no_copy worklist not empty

  // No more lo-degree no-copy live ranges to simplify
}

//------------------------------Simplify---------------------------------------
// Simplify the IFG by removing LRGs of low degree.
void PhaseChaitin::Simplify( ) {

  while( 1 ) {                  // Repeat till simplified it all
    // May want to explore simplifying lo_degree before _lo_stk_degree.
    // This might result in more spills coloring into registers during
    // Select().
    while( _lo_degree || _lo_stk_degree ) {
      // If possible, pull from lo_stk first
      uint lo;
      if( _lo_degree ) {
        lo = _lo_degree;
        _lo_degree = lrgs(lo)._next;
      } else {
        lo = _lo_stk_degree;
        _lo_stk_degree = lrgs(lo)._next;
      }

      // Put the simplified guy on the simplified list.
      lrgs(lo)._next = _simplified;
      _simplified = lo;
      // If this guy is "at risk" then mark his current neighbors
      if( lrgs(lo)._at_risk ) {
        IndexSetIterator elements(_ifg->neighbors(lo));
        uint datum;
        while ((datum = elements.next()) != 0) {
          lrgs(datum)._risk_bias = lo;
        }
      }
      
      // Yank this guy from the IFG.  
      IndexSet *adj = _ifg->remove_node( lo );
      
      // If any neighbors' degrees fall below their number of
      // allowed registers, then put that neighbor on the low degree
      // list.  Note that 'degree' can only fall and 'numregs' is
      // unchanged by this action.  Thus the two are equal at most once,
      // so LRGs hit the lo-degree worklist at most once.
      IndexSetIterator elements(adj);
      uint neighbor;
      while ((neighbor = elements.next()) != 0) {
        LRG *n = &lrgs(neighbor);
        if( VerifyOpto ) {
          assert( _ifg->effective_degree(neighbor) == n->degree(), "" );
        }

        // Check for just becoming of-low-degree just counting registers.
        // _must_spill live ranges are already on the low degree list.
        if( n->just_lo_degree() && !n->_must_spill ) {
          assert(!(*_ifg->_yanked)[neighbor],"Cannot move to lo degree twice");
          // Pull from hi-degree list
          uint prev = n->_prev;
          uint next = n->_next;
          if( prev ) lrgs(prev)._next = next;
          else _hi_degree = next;
          lrgs(next)._prev = prev;
          n->_next = _lo_degree; 
          _lo_degree = neighbor;
        }
      }
    } // End of while lo-degree/lo_stk_degree worklist not empty

    // Check for got everything: is hi-degree list empty?
    if( !_hi_degree ) break;

    // Time to pick a potential spill guy
    uint lo_cost = _hi_degree;
    double cost = lrgs(lo_cost).score();
    
    // Find cheapest guy
    debug_only( int lo_no_simplify=0; );
    for( uint i = _hi_degree; i; i = lrgs(i)._next ) {
      assert( !(*_ifg->_yanked)[i], "" );
      // It's just vaguely possible to move hi-degree to lo-degree without
      // going through a just-lo-degree stage: If you remove a double from
      // a float live range it's degree will drop by 2 and you can skip the
      // just-lo-degree stage.  It's very rare (shows up after 5000+ methods
      // in -Xcomp of Java2Demo).  So just choose this guy to simplify next.
      if( lrgs(i).lo_degree() ) {
        lo_cost = i;
        break;
      }
      debug_only( if( lrgs(i)._was_lo ) lo_no_simplify=i; );
      double icost = lrgs(i).score();

      // Compare cost/area of i vs cost/area of lo_cost.  Smaller cost/area
      // wins.  Ties happen because all live ranges in question have spilled
      // a few times before and the spill-score adds a huge number which
      // washes out the low order bits.  We are choosing the lesser of 2
      // evils; in this case pick largest area to spill.
      if( icost < cost ) {
        lo_cost = i;
        cost = icost;
      } 
    }
    LRG *lo_lrg = &lrgs(lo_cost);
    // The live range we choose for spilling is either hi-degree, or very
    // rarely it can be low-degree.  If we choose a hi-degree live range
    // there better not be any lo-degree choices.
    assert( lo_lrg->lo_degree() || !lo_no_simplify, "Live range was lo-degree before coalesce; should simplify" );

    // Pull from hi-degree list
    uint prev = lo_lrg->_prev;
    uint next = lo_lrg->_next;
    if( prev ) lrgs(prev)._next = next;
    else _hi_degree = next;
    lrgs(next)._prev = prev;
    // Jam him on the lo-degree list, despite his high degree.
    // Maybe he'll get a color, and maybe he'll spill.
    // Only Select() will know.
    lrgs(lo_cost)._at_risk = true;
    _lo_degree = lo_cost;
    lo_lrg->_next = 0;
    
  } // End of while not simplified everything

}

//------------------------------bias_color-------------------------------------
// Choose a color using the biasing heuristic
OptoReg::Name PhaseChaitin::bias_color( LRG &lrg, int chunk ) {
  // Check for "at_risk" LRG's
  uint risk_lrg = Find(lrg._risk_bias);
  if( risk_lrg != 0 ) {
    // Walk the colored neighbors of the "at_risk" candidate
    // Choose a color which is both legal and already taken by a neighbor
    // of the "at_risk" candidate in order to improve the chances of the
    // "at_risk" candidate of coloring
    IndexSetIterator elements(_ifg->neighbors(risk_lrg));
    uint datum;
    while ((datum = elements.next()) != 0) {
      OptoReg::Name reg = lrgs(datum).reg();
      // If this LRG's register is legal for us, choose it
      if( reg >= chunk && reg < chunk+(RM_SIZE<<5) &&
          lrg.mask().Member(OptoReg::add(reg,-chunk)) )
        return reg;
    }
  }

  uint copy_lrg = Find(lrg._copy_bias);
  if( copy_lrg != 0 ) {
    // If he has a color,
    if( !(*(_ifg->_yanked))[copy_lrg] ) {
      OptoReg::Name reg = lrgs(copy_lrg).reg();
      //  And it is legal for you,
      if( reg >= chunk && reg < chunk+(RM_SIZE<<5) &&
          lrg.mask().Member(OptoReg::add(reg,-chunk)) )
        return reg;
    } else if( chunk == 0 ) {
      // Choose a color which is legal for him
      RegMask tempmask = lrg.mask();
      tempmask.AND(lrgs(copy_lrg).mask());
      OptoReg::Name reg = tempmask.find_first_elem();
      if( reg != OptoReg::Bad )
        return reg;
    }
  }

  // If no bias info exists, just go with the register selection ordering
 
  // CNC - Fun hack.  Alternate 1st and 2nd selection.  Enables post-allocate
  // copy removal to remove many more copies, by preventing a just-assigned
  // register from being repeatedly assigned.
  OptoReg::Name reg = lrg.mask().find_first_elem();
  if( ++_alternate & 1 ) {
    // This 'Remove; find; Insert' idiom is an expensive way to find the
    // SECOND element in the mask.  
    lrg.Remove(reg);
    OptoReg::Name reg2 = lrg.mask().find_first_elem();
    lrg.Insert(reg);
    if( reg2 != OptoReg::Bad &&
        reg2 < SharedInfo::stack0 )
      reg = reg2;
  }
  return OptoReg::add( reg, chunk );
}

//------------------------------choose_color-----------------------------------
// Choose a color in the current chunk
OptoReg::Name PhaseChaitin::choose_color( LRG &lrg, int chunk ) {  
  assert( C->in_preserve_stack_slots() == 0 || chunk != 0 || lrg._is_bound || lrg.mask().is_bound1() || !lrg.mask().Member(OptoReg::Name(_matcher._old_SP-1)), "must not allocate stack0 (inside preserve area)");
  assert(C->out_preserve_stack_slots() == 0 || chunk != 0 || lrg._is_bound || lrg.mask().is_bound1() || !lrg.mask().Member(OptoReg::Name(_matcher._old_SP+0)), "must not allocate stack0 (inside preserve area)");

  if( lrg.num_regs() == 1 )    // Common Case
    // Use a heuristic to "bias" the color choice
    return bias_color(lrg, chunk);

  assert( lrg.num_regs() >= 2, "dead live ranges do not color" );

  if( !lrg._fat_proj ) {        // Aligned pairs?
    assert( lrg.num_regs() == 2, "only aligned pairs (no quads yet)" );
    // Find an aligned pair
    return OptoReg::add(lrg.mask().find_first_pair(),chunk);
  } 

  // Fat-proj case or misaligned double argument. 
  assert(lrg.compute_mask_size() == lrg.num_regs() ||
         lrg.num_regs() == 2,"fat projs exactly color" );
  assert( !chunk, "always color in 1st chunk" );
  // Return the highest element in the set.
  return lrg.mask().find_last_elem();
}

//------------------------------Select-----------------------------------------
// Select colors by re-inserting LRGs back into the IFG.  LRGs are re-inserted
// in reverse order of removal.  As long as nothing of hi-degree was yanked,
// everything going back is guaranteed a color.  Select that color.  If some
// hi-degree LRG cannot get a color then we record that we must spill.
uint PhaseChaitin::Select( ) {
  uint spill_reg = LRG::SPILL_REG;
  _max_reg = OptoReg::Name(0);  // Past max register used
  while( _simplified ) {
    // Pull next LRG from the simplified list - in reverse order of removal
    uint lidx = _simplified;
    LRG *lrg = &lrgs(lidx);
    _simplified = lrg->_next;

    // Re-insert into the IFG
    _ifg->re_insert(lidx);
    if( !lrg->alive() ) continue;
    // capture allstackedness flag before mask is hacked
    const int is_allstack = lrg->mask().is_AllStack();

    // Yeah, yeah, yeah, I know, I know.  I can refactor this
    // to avoid the GOTO, although the refactored code will not
    // be much clearer.  We arrive here IFF we have a stack-based
    // live range that cannot color in the current chunk, and it
    // has to move into the next free stack chunk.
    int chunk = 0;              // Current chunk is first chunk
    retry_next_chunk:

    // Remove neighbor colors
    IndexSet *s = _ifg->neighbors(lidx);
    assert( !lrg->_fat_proj || _ifg->neighbor_cnt(lidx) == 0, "fat proj has no nbrs");
    debug_only(RegMask orig_mask = lrg->mask();)
    IndexSetIterator elements(s);
    uint neighbor;
    while ((neighbor = elements.next()) != 0) {
      // Note that neighbor might be a spill_reg.  In this case, exclusion
      // of its color will be a no-op, since the spill_reg chunk is in outer
      // space.  Also, if neighbor is in a different chunk, this exclusion
      // will be a no-op.  (Later on, if lrg runs out of possible colors in
      // its chunk, a new chunk of color may be tried, in which case
      // examination of neighbors is started again, at retry_next_chunk.)
      LRG &nlrg = lrgs(neighbor);
      OptoReg::Name nreg = nlrg.reg();
      // Only subtract masks in the same chunk
      if( nreg >= chunk && nreg < chunk + (RM_SIZE<<5) )
        lrg->SUBTRACT(nlrg.mask());
    }
    //assert(is_allstack == lrg->mask().is_AllStack(), "nbrs must not change AllStackedness");
    // Aligned pairs need aligned masks
    if( lrg->num_regs() == 2 && !lrg->_fat_proj )
      lrg->ClearToPairs();

    // Check if a color is available and if so pick the color
    OptoReg::Name reg = choose_color( *lrg, chunk );
#ifdef SPARC
    debug_only(lrg->compute_set_mask_size());
    assert(lrg->num_regs() != 2 || lrg->is_bound() || is_even(reg-1), "allocate all doubles aligned");
#endif

    //---------------
    // If we fail to color and the AllStack flag is set, trigger
    // a chunk-rollover event
    if(OptoReg::add(reg,-chunk) == OptoReg::Bad && is_allstack) {
      // Bump register mask up to next stack chunk
      chunk += (RM_SIZE<<5);
      lrg->Set_All();
      
      goto retry_next_chunk;
    }

    //---------------
    // Did we get a color?
    else if( reg != OptoReg::Bad ) {
      // Record selected register
      lrg->set_reg(reg);

      if( reg >= _max_reg )     // Compute max register limit
        _max_reg = OptoReg::add(reg,1);
      // Fold reg back into normal space
      reg = OptoReg::add(reg,-chunk);

      // If the live range is not bound, then we actually had some choices
      // to make.  In this case, the mask has more bits in it than the colors
      // choosen.  Restrict the mask to just what was picked.
      if( lrg->num_regs() == 1 ) { // Size 1 live range
        lrg->Clear();           // Clear the mask
        lrg->Insert(reg);       // Set regmask to match selected reg
        lrg->set_mask_size(1);
      } else if( !lrg->_fat_proj ) {
        // For pairs, also insert the low bit of the pair
        assert( lrg->num_regs() == 2, "unbound fatproj???" );
        lrg->Clear();           // Clear the mask
        lrg->Insert(reg);       // Set regmask to match selected reg
        lrg->Insert(OptoReg::add(reg,-1));
        lrg->set_mask_size(2);
      } else {                  // Else fatproj
        // mask must be equal to fatproj bits, by definition
      }
      // Note that reg is the highest-numbered register in the newly-bound mask.
    } // end color available case

    //---------------
    // Live range is live and no colors available
    else {   
      assert( lrg->alive(), "" );
      assert( !lrg->_fat_proj, "fat_proj cannot spill");
      assert( !orig_mask.is_AllStack(), "All Stack does not spill" );

      // Assign the special spillreg register
      lrg->set_reg(OptoReg::Name(spill_reg++));
      // Do not empty the regmask; leave mask_size lying around
      // for use during Spilling
#ifndef PRODUCT
      if( TraceSpilling ) {
        tty->print(" neighbors: ");
        s->dump();
        debug_only(tty->print(" original mask: "));
        debug_only(orig_mask.dump());
        dump_lrg(lidx);
      }
#endif
    } // end spill case

  }

  return spill_reg-LRG::SPILL_REG;      // Return number of spills
}


//------------------------------copy_was_spilled-------------------------------
// Copy 'was_spilled'-edness from the source Node to the dst Node.
void PhaseChaitin::copy_was_spilled( Node *src, Node *dst ) {
  if( _spilled_once.test(src->_idx) ) {
    _spilled_once.set(dst->_idx);
    lrgs(Find(dst))._was_spilled1 = 1;
    if( _spilled_twice.test(src->_idx) ) {
      _spilled_twice.set(dst->_idx);
      lrgs(Find(dst))._was_spilled2 = 1;
    }
  }
}

//------------------------------set_was_spilled--------------------------------
// Set the 'spilled_once' or 'spilled_twice' flag on a node.
void PhaseChaitin::set_was_spilled( Node *n ) {
  if( _spilled_once.test_set(n->_idx) )
    _spilled_twice.set(n->_idx);
}

//------------------------------unlink_dead_copy-------------------------------
// Unlink in the graph (skip around) a dead copy.
void PhaseChaitin::unlink_dead_copy( Node *n, uint inpidx ) {
  Node *def = n->in(inpidx);
  if (def == NULL) return;
  uint useidx = Find(def);
  // Check for a coalesced copy.  They must be removed before
  // Split decisions are made on a USE.
  uint idx = def->is_Copy();
  // Is this a copy to/from the same live range?
  while( idx && (Find(def->in(idx)) == useidx) ) {
    // Then skip it; the copy goes dead & will be removed later
    n->set_req(inpidx,def->in(idx));
    // Check for removing a dead prior-spilled copy.
    // If so, make sure this live range keeps a prior-spilled tag.
    copy_was_spilled( def, n );
    // Now see if there is a chain of these
    assert(def != def->in(idx),"SpillCopy of Self not legal");
    def = def->in(idx);
    idx = def->is_Copy();
  }
}

//------------------------------reg2offset-------------------------------------
int PhaseChaitin::reg2offset_unchecked( OptoReg::Name reg ) const {
  // Slots below _max_in_arg_stack_reg are offset by the entire frame.
  // Slots above _max_in_arg_stack_reg are frame_slots and are not offset.
  int slot = (reg < _matcher._new_SP) 
    ? reg - SharedInfo::stack0 + _framesize
    : reg - _matcher._new_SP;
  // Note:  We use the direct formula (reg - SharedInfo::stack0) instead of
  // SharedInfo::reg2stack(reg), in order to avoid asserts in the latter
  // function.  This routine must remain unchecked, so that dump_frame()
  // can do its work undisturbed.
  return slot*sizeof(jint);     
}

int PhaseChaitin::reg2offset( OptoReg::Name reg ) const {

  // Not allowed in the out-preserve area.
  // In-preserve area is allowed so Intel can fetch the return pc out.
  assert( reg <  _matcher._old_SP || 
          (reg >= OptoReg::add(_matcher._old_SP,C->out_preserve_stack_slots()) &&
           reg <  _matcher._in_arg_limit) ||
          reg >=  OptoReg::add(_matcher._new_SP,C->out_preserve_stack_slots()),
          "register allocated in a preserve area" );
  return reg2offset_unchecked( reg );
}

//------------------------------fixup_spills-----------------------------------
// Convert Ideal spill instructions into proper FramePtr + offset Loads and
// Stores.  Use-def chains are NOT preserved, but Node->LRG->reg maps are.
bool PhaseChaitin::fixup_spills() {
  // Grab the Frame Pointer
  Node *fp = _cfg._broot->head()->in(1)->in(TypeFunc::FramePtr);
  Node *mem= _cfg._broot->head()->in(1)->in(TypeFunc::Memory);
  // Compute a generic offset
  ConINode    *offset  = new (1) ConINode(TypeInt::ZERO);
  AddPNode    *adr     = new (4) AddPNode(C->_top_node,fp,offset);
  ConINode    *doffset = new (1) ConINode( TypeInt::ZERO);
  AddPNode    *dst_adr = new (4) AddPNode(C->_top_node,fp,doffset);
  MemMoveNode *mmove   = new (3) MemMoveNode( dst_adr, adr );

  // Dont care and shared are hosed; reset locally.
  _matcher._shared <<= fp->_idx;
  _matcher._shared >>= offset->_idx;
  _matcher._shared >>= doffset->_idx;
  _matcher._shared >>= adr->_idx;
  _matcher._shared >>= dst_adr->_idx;
  _matcher._dontcare >>= fp->_idx;
  _matcher._dontcare >>= offset->_idx;
  _matcher._dontcare >>= doffset->_idx;
  _matcher._dontcare >>= adr->_idx;
  _matcher._dontcare >>= dst_adr->_idx;
  _matcher._dontcare <<= mem->_idx;

  // For all blocks
  for( uint i = 0; i < _cfg._num_blocks; i++ ) {
    Block *b = _cfg._blocks[i];

    // For all instructions in block
    uint last_inst = b->end_idx();
    for( uint j = 1; j < last_inst; j++ ) {
      // Bailout if we might exceed node limit when spilling this instruction
      if( C->_unique > 63000 ) {
        return false;
      }
      Node *n = b->_nodes[j];

      // unlink all dead copies
      uint cnt = n->req();

      for( uint inpidx = 1; inpidx < cnt; inpidx++ )
        unlink_dead_copy( n, inpidx );

      // Check for CISC-Spill
      if( UseCISCSpill )  {
        int inp = n->cisc_operand();
        if( inp != AdlcVMDeps::Not_cisc_spillable ) {
	  // Convert operand number to edge index number
	  MachNode *mach = n->is_Mach();
	  inp = mach->operand_index(inp);
          Node *src = n->in(inp);   // Value to load or store
          LRG &lrg_cisc = lrgs( Find_const(src) );
          OptoReg::Name src_reg = lrg_cisc.reg();
          // Doubles record the HIGH register of an adjacent pair.
          src_reg = OptoReg::add(src_reg,1-lrg_cisc.num_regs());
          if( src_reg >= SharedInfo::stack0 ) { // If input is on stack
            // This is a CISC Spill, get stack offset and construct new node
#ifndef PRODUCT
            if( TraceCISCSpill ) {
              tty->print("    reg-instr:  ");
              n->dump();
            }
#endif
            int stk_offset = reg2offset(src_reg);
            offset->_type = TypeInt::make(stk_offset);
            // Transform node
            MachNode *cisc = mach->cisc_version(stk_offset)->is_Mach();
            cisc->set_req(inp,fp);          // Base register is frame pointer
            if( cisc->oper_input_base() > 1 && mach->oper_input_base() <= 1 ) {
              assert( cisc->oper_input_base() == 2, "Only adding one edge");
              cisc->ins_req(1,NULL);        // Requires a memory edge
            }
            b->_nodes.map(j,cisc);          // Insert into basic block
            // 
            ++_used_cisc_instructions;
#ifndef PRODUCT
            if( TraceCISCSpill ) {
              tty->print("    cisc-instr: ");
              cisc->dump();
            }
#endif
            // MachNode *stk_access = _matcher.match_tree( src_copy->load_make(NULL,(stk_offset ? adr : fp)) );
            // cisc->_opnds[inp] = stk_access;
          } else {
#ifndef PRODUCT
            if( TraceCISCSpill ) {
              tty->print("    using reg-instr: ");
              n->dump();
            }
#endif
            ++_unused_cisc_instructions;    // input can be on stack
          }
        }
      }

      // Check for being some flavor of copy
      int idx = n->is_Copy();
      if( !idx ) continue;

      // Grab source and destination registers
      Node *src = n->in(idx);   // Value to load or store
      LRG &lrg_src = lrgs( Find_const(src) );
      LRG &lrg_dst = lrgs( Find_const(n  ) );
      OptoReg::Name src_reg = lrg_src.reg();
      OptoReg::Name dst_reg = lrg_dst.reg();

      // Doubles record the HIGH register of an adjacent pair.
      // We want the LOW register first, then we'll add +4 to the
      // stack offset to get the HIGH part.  Doubles are NOT
      // required to be aligned, in case the calling convention
      // requires mis-aligned doubles.
      assert( lrg_src.num_regs() == lrg_dst.num_regs(), "" );
      src_reg = OptoReg::add(src_reg,1-lrg_src.num_regs());
      dst_reg = OptoReg::add(dst_reg,1-lrg_src.num_regs());

      // Check for and remove useless copies
      if( dst_reg == src_reg ){ // Copy register to same register?
        b->_nodes.remove(j);    // Yank useless copy
        last_inst--;  j--;      // Do new instruction at same iter count
        continue;
      } 

      // Check for spill copy
      const SpillCopyNode *copy = n->is_SpillCopy();
      if( !copy ) {             // Not a spill-copy
        _final_copies++;        // Must be a normal copy left in code
        _final_copy_cost += b->_freq;
        continue;
      }
      // I'll need to match the src as part of a match tree
      _matcher._shared.set(src->_idx);
      _matcher._dontcare >>= src->_idx;
      MachNode *spill;

      // Spill-copies come in 4 flavors: reg-reg, reg-stk, stk-reg and stk-stk
      if( src_reg < SharedInfo::stack0 ) {// src in register?
        if( dst_reg < SharedInfo::stack0 ) {// reg-reg case?
          _final_copies++;      // Normal copy left in code
          _final_copy_cost += b->_freq;
          spill = _matcher.match_tree(copy->copy_make( src ) );
          _names.extend(spill->_idx,n2lidx(n));

        } else {                // Else reg-stk case
          // Need to store value from register to stack here
          _final_stores++;
          _final_store_cost += b->_freq;
          int stk_offset = reg2offset(dst_reg);
          offset->_type = TypeInt::make( stk_offset );
          spill = _matcher.match_tree(copy->store_make( mem, stk_offset ? adr : fp, src ) );
          _names.extend(spill->_idx,0);
          // Update the memory operand for instruction scheduling,
          // by pointing to the previous SpillCopy's store, if any
          if (!spill->is_Con()) {
            Node *mt;
            for (mt = src; mt->is_SpillCopy(); mt = mt->in(MemNode::Memory));
            if (mt->Opcode() == Op_MachProj || mt->is_Phi())
              mt = mt->in(0);
            spill->set_req(MemNode::Memory, mt); 
          }

          // Set the original SpillCopy to point to the store
          n->set_req(MemNode::Memory, spill);
        }
      } else {                  // Else src in stk
        if( dst_reg < SharedInfo::stack0 ) {// stk-reg case?
          // Need to load a value from stack to register here
          _final_loads++;
          _final_load_cost += b->_freq;
          int stk_offset = reg2offset(src_reg);
          offset->_type = TypeInt::make( stk_offset );

          spill = _matcher.match_tree(copy->load_make( mem, stk_offset ? adr : fp ) );
          assert(Matcher::idealreg2regmask[spill->ideal_reg()]->overlap(copy->out_RegMask()), "must load to a register of the required type");
          _names.extend(spill->_idx,n2lidx(n));

          // Update the memory operand for instruction scheduling,
          // by pointing to the previous SpillCopy's store, if any
          if (!spill->is_Con()) {
            Node *mt;
            for (mt = src; mt->is_SpillCopy(); mt = mt->in(MemNode::Memory));
            if (mt->Opcode() == Op_MachProj || mt->is_Phi())
              mt = mt->in(0);
            spill->set_req(MemNode::Memory, mt); 
          }
          n->set_req(MemNode::Memory, spill);
        } else {                // Else stk-stk case
          // Need to store value from stack to stack here
          // %%%% remove this code, since mem-mem moves are not so special now?
          _final_memoves++;
          _final_memove_cost += b->_freq;
          uint src_offset = reg2offset(src_reg);
          offset->_type = TypeInt::make( src_offset );
          uint dst_offset = reg2offset(dst_reg);
          doffset->_type = TypeInt::make( dst_offset );
          spill = _matcher.match_tree( mmove );
          _names.extend(spill->_idx,0);

          // Update the memory operand for instruction scheduling,
          // by pointing to the previous SpillCopy's store, if any
          Node *mt;
          for (mt = src; mt->is_SpillCopy(); mt = mt->in(MemNode::Memory));
          if (mt->Opcode() == Op_MachProj || mt->is_Phi())
            mt = mt->in(0);
          spill->add_prec(mt); 

          // Set the original SpillCopy to point to the mem-mem copy
          n->set_req(MemNode::Memory, spill);

          switch( copy->_type->_base ) {
          case Type::DoubleBot:
          case Type::DoubleCon:
          case Type::Long:     
            // Smash over the Ideal SpillCopy with the matched load or store
            // instruction.  Use-defs are NOT corrected.
            b->_nodes.insert(j++,spill);  last_inst++;
            // Now do the other half
            offset ->_type = TypeInt::make( src_offset+4 );
            doffset->_type = TypeInt::make( dst_offset+4 );
            spill = _matcher.match_tree( new (3) MemMoveNode( dst_adr, adr ) );
            _names.extend(spill->_idx,0);
            spill->add_prec(mt); 

            // Set the original SpillCopy to point to the mem-mem copy
            n->set_req(MemNode::Memory, spill);
          }
        } // End of if dst in register or stack
      } // End of if src in register or stack

      // Smash over the Ideal SpillCopy with the matched load or store
      // instruction.  Use-defs are NOT corrected.
      b->_nodes.map(j,spill);

      // Update the basic block mapping
      _cfg._bbs.map(spill->_idx, b);
      _cfg._bbs.map(n->_idx, NULL);

    } // End of for all instructions

  } // End of for all blocks

  return true;
}


//------------------------------collect_gc_masks-------------------------------
// Collect GC mask info - where are all the OOPs?
void PhaseChaitin::collect_gc_masks(bool save_argument_registers) {

  // Push the callee-save register info around into all the LRGs that
  // carry it.  The original callee-save register might end up in a
  // stack-slot, or another register, or whatever...  Track it.
  int16 *callee = NEW_RESOURCE_ARRAY( int16, _maxlrg );
  // Flag as -1 meaning NOT CALLEE SAVE
  memset( callee, -1, _maxlrg*sizeof(uint16) );

  // ===========================
  // Alas, its a dataflow problem.  Bummer.  Iterate to a fixed point.
  int changes = 1;
  while( changes ) {
    changes = 0;
    // For all blocks
    for( uint i = 0; i < _cfg._num_blocks; i++ ) {
      Block *b = _cfg._blocks[i];

      // For all instructions
      uint last_inst = b->end_idx();
      for( uint j = 1; j<last_inst; j++ ) {
        Node *n = b->_nodes[j];

        // Find callee-save register projections
        ProjNode *proj = n->is_Proj();
        if( proj && proj->_con >= TypeFunc::Parms &&
            proj->in(0)->is_Start() != NULL ) {
          const RegMask &rm = proj->out_RegMask();
          OptoReg::Name reg = rm.find_last_elem();
          if( reg < SharedInfo::stack0 ) { // ignore stack
            // Save-on-entry register?
            if( _matcher._register_save_policy[reg] == 'E' || 
                _matcher._register_save_policy[reg] == 'A' ||
                // Also save argument registers in the trampolining stubs
                (save_argument_registers && _matcher._can_be_arg[reg]) ) {
              uint lidx = n2lidx(n);
              assert( callee[lidx] == -1 || callee[lidx] == reg, "" );
              callee[lidx] = reg; // Recorder caller register
            }
          }
        } // End of found a projection

        // Copies propagate callee-save registers
        int idx = n->is_Copy();
        if( idx ) {
          uint lidx  = n2lidx(n);
          uint lidx1 = n2lidx(n->in(idx));

          // Is input LRG callee-save?  Then we are too
          if( callee[lidx] != callee[lidx1] ) {
            // Neither can be an oop, since I do not know what the
            // caller has put in the register
            assert( !lrgs(lidx )._is_oop, "" );
            assert( !lrgs(lidx1)._is_oop, "" );
            callee[lidx] = callee[lidx1]; // Propagate callee-save info
            changes = 1;        // Flag as changes
          }
        }
      }
    }
  }

  // _max_reg is past the largest *register* used.
  // Convert that to a frame_slot number.
  if( _max_reg <= _matcher._new_SP ) 
    _framesize = C->out_preserve_stack_slots();
  else _framesize = OptoReg::add(_max_reg, -_matcher._new_SP);
  assert((int)(_matcher._new_SP+_framesize) >= (int)_matcher._out_arg_limit, "framesize must be large enough");

  // this frame must preserve the required fp alignment
  const int stack_alignment_in_words = _matcher.stack_alignment() - LogBitsPerWord;
  if (stack_alignment_in_words > 0)
    _framesize = round_to(_framesize, 1 << stack_alignment_in_words);
  assert( _framesize >= 0 && _framesize <= 1000000, "sanity check" );
  _total_framesize += _framesize;
  if( (int)_framesize > _max_framesize)
    _max_framesize = _framesize;

  // ===========================
  // Generate GC info now...

  // For all blocks (in any order) do...
  for( uint i = 0; i < _cfg._num_blocks; i++ ) {
    Block *b = _cfg._blocks[i];
    IndexSet *liveout = _live->live(b);
    // Update the liveout data via the Union-Find mapping.
    // Remove the need 
    IndexSetIterator elements(liveout);
    uint lidx;
    while ((lidx = elements.next()) != 0) {
      uint fidx = Find(lidx);
      if( lidx != fidx ) {
        liveout->remove(lidx);
        liveout->insert(fidx);

        // !!!!! Before we used to back up one or two iterations in a
        // for loop with the following code:
        //
        //        j = j ? j-2 : j-1;
        //
        // This does not fit well into the semantics of the new
        // IndexSetIterator.  As a temporary fix-up, we reset the iterator
        // to the very beginning of the set and go from there.  In practice
        // this works fine, but a more elegant solution would be nice.
        elements = IndexSetIterator(liveout);
      }
    }

    // For all instructions in reverse order, incrementally compute
    // LIVE through the block
    for( uint j = b->end_idx()+1; j > 1; j-- ) {
      Node *n = b->_nodes[j-1];
      // First Phi on a backwards pass marks the end of any chance of
      // seeing a SafePoint node.
      if( n->is_Phi() ) break;

      // Get LRG being defined
      uint r = n2lidx(n);
      assert( r == Find_const(r), "" );

      // Some special values do not allocate
      // Remove defined value from the liveout set
      if( r ) {
        // If it is not live, then this instruction is dead.  Probably caused
        // by spilling and rematerialization.  Who cares why, yank this baby.
        if( !liveout->member(r) ) {
          Node *def = n->in(0);
          if( !n->is_Proj() ||
              // Could also be a flags-projection of a dead ADD or such.
              (n2lidx(def) && !liveout->member(n2lidx(def))) ) {
            b->_nodes.remove(j - 1);
            continue;
          }
        } else
          liveout->remove(r);
      }

      // Make all inputs live
      // Need special handling for input edges which are for derived pointers
      for( uint k = 1; k < n->req(); k++ ) 
        liveout->insert( n2lidx(n->in(k)) );

      // Any call site is a GC point.
      // Collect which locations hold oops
      MachNode *mach = n->is_Mach();
      MachSafePointNode *sfn = mach ? mach->is_MachSafePoint() : NULL;
      // no map needed for leaf calls
      if (sfn != NULL && sfn->is_MachCallLeaf()) sfn = NULL;
      if( sfn )
        n->set_oop_map( build_oop_map(callee, liveout, sfn) );
#ifdef ASSERT
      // Check that debug info does not get clobbered by the call
      if( sfn && sfn->_jvms ) {
        for( uint i = sfn->_jvms->locoff(); i < sfn->req(); i++ ) {
          // Register for input edge
          uint lidx = n2lidx(n->in(i));
	  if( lidx ) {		// Ignore TOP debug inputs
	    OptoReg::Name reg = lrgs(lidx).reg(); 
	    // Now scan for kills just following the call
	    for( uint k = j; k < b->end_idx(); k++ ) {
	      ProjNode *m = b->_nodes[k]->is_Proj();
	      if( !m ) break;
	      // Cannot let our debug info be killed by a call.  If we deopt
	      // at the call site, the debug info will be missing.
	      assert( reg >= (RM_SIZE<<5) || !m->out_RegMask().Member(reg),
		      "debug info killed by a call" );
	    }
          }
        }
      }
#endif


    } // End of forall instructions in block
  } // End of forall blocks

}

#ifdef ASSERT
static bool oop_is_derived(LRG& lrg) {
  if( !lrg.alive() )  return false;
  if( lrg._def == (Node*)-1 )  return false;  // cannot really tell, but it is just an assert...
  const TypeJavaPtr *tp = lrg._def->bottom_type()->isa_javaptr();
  if( !tp )  return false;
  return tp->_offset != 0;  // a derived oop must have an offset of exactly zero
}
#endif

//------------------------------build_oop_map----------------------------------
// Figure out which oop live-ranges cross all GC points and make GC maps
// for them.
OopMap *PhaseChaitin::build_oop_map( int16 *callee, IndexSet *liveout, MachSafePointNode *sfpt ) {

  // Compute framesize after sync slots are accounted for
  int max_inarg_slot = _matcher._new_SP - SharedInfo::stack0;
  OopMap *omap = new OopMap( _framesize,  max_inarg_slot );
  MachCallNode *mcall = sfpt->is_MachCall();
  IndexSetIterator elements(liveout);
  uint lidx;
  // Before handling the liveout set, process the derived pointers
  JVMStateObj *jvms = sfpt->jvmtail();
  for( uint i = jvms->oopoff(); i < sfpt->req(); i++ ) {
    // grab derived/base pair
    Node *derived = sfpt->in(i++);
    Node *base    = sfpt->in(i);
    // get their live ranges
    uint dlidx = n2lidx(derived);
    uint blidx = n2lidx(base);
    LRG &dlrg = lrgs(dlidx);
    LRG &blrg = lrgs(blidx);
      // get the register assignment from the LRG
    OptoReg::Name dreg = dlrg.reg();
    OptoReg::Name breg = blrg.reg();
    // create the oop map entry
    omap->set_derived_oop( dreg, _framesize, max_inarg_slot, breg );
  }
  // Iterate over liveout set and add to oopmap
  while ((lidx = elements.next()) != 0) {
    LRG &lrg = lrgs(lidx);
    OptoReg::Name reg = lrg.reg();
    // Who cares that the frame pointer is live?
    if( reg == SharedInfo::c_frame_pointer ) continue;
    
    // Inspect to decide if its an oop or not
    if( lrg._is_oop ) {  // Oop?
      // Do not re-add derived oops which have already been handled
      if( omap->locs_used((int)reg) == OopMapValue::derived_oop_value )
        continue;
      if( mcall != NULL ) {      // Call-site?
        // Outgoing argument GC mask responsibility belongs to the callee, 
        // not the caller.  Inspect the inputs to the call, to see if 
        // this live-range is one of them.
        uint cnt = mcall->_func->_domain->_cnt;
        uint j;
        for( j = TypeFunc::Parms; j < cnt; j++)
          if( n2lidx(mcall->in(j)) == lidx )
            break;
        if( j >= cnt ) {
          omap->set_oop( reg, _framesize, max_inarg_slot );
          assert(!oop_is_derived(lrg),"must hook all derived oops explicitly");
        }
      }
      else {
        omap->set_oop( reg, _framesize, max_inarg_slot );
        assert(!oop_is_derived(lrg), "must hook all derived oops explicitly");
      }
    } else if( callee[lidx] != -1 ) { // Callee-save?
      omap->set_callee_saved( reg, _framesize, max_inarg_slot, OptoReg::Name(callee[lidx]) );
      if( lrg.num_regs() == 2 ) 
        omap->set_callee_saved( OptoReg::add(reg,-1), _framesize, max_inarg_slot, OptoReg::Name(callee[lidx]-1) );
    } else {                    // Else, it is a live value
      omap->set_value( reg, _framesize, max_inarg_slot );
      if( lrg.num_regs() == 2 ) // Doubles take 2 slots
        omap->set_value( OptoReg::add(reg,-1), _framesize, max_inarg_slot );
      // Live floats take 2 slots as well!  They get callee-saved as doubles,
      // so both halves have to be treated as live.  Make the other half live.
      else if( _matcher._register_save_type[reg] == Op_RegF ) { 
        omap->set_value( OptoReg::Name(reg^1), _framesize, max_inarg_slot ); 
      }
    }
  } // End of for liveout set

  if (ZapDeadCompiledLocals) {
    // Tag any dead locals so that they can be zapped.  Our idealized register
    // space breaks into the following areas:
    //
    //          Machine Registers
    //          Preserve Area
    //          Incoming Arguments
    //          Outgoing Arguments
    //          Normal Spills
    //
    // We want to tag all unused slots in the Incoming Arguments and Normal
    // Spills areas as being dead.  We currently use the _matcher._endoutargs
    // value to figure out where the start of the Outgoing Arguments area is.
    // In the future, if we relax this rigid boundary between the Outgoing
    // Arguments and the Normal Spills, then we will need to figure out
    // normal_spill_start on a per-call basis.
    int incoming_start = _matcher._old_SP +Compile::out_preserve_stack_slots();
    int incoming_limit = _matcher._in_arg_limit;
    int normal_spill_start = _matcher._out_arg_limit;
    int normal_spill_limit = _max_reg;

    omap->set_unused_as_dead(incoming_start, incoming_limit, _framesize, max_inarg_slot);
    omap->set_unused_as_dead(normal_spill_start, normal_spill_limit, _framesize, max_inarg_slot);
  }

  return omap;
}

//------------------------------find_base_for_derived--------------------------
// Helper to stretch above; recursively discover the base Node for a
// given derived Node.  Easy for AddP-related machine nodes, but needs
// to be recursive for derived Phis.
Node *PhaseChaitin::find_base_for_derived( Node **derived_base_map, Node *derived, uint &maxlrg ) {
  // See if already computed; if so return it
  if( derived_base_map[derived->_idx] )
    return derived_base_map[derived->_idx];

  // See if this happens to be a base.  
  // NOTE: we use TypePtr instead of TypeJavaPtr because we can have
  // pointers derived from NULL!  These are always along paths that
  // can't happen at run-time but the optimizer cannot deduce it so
  // we have to handle it gracefully.
  const TypePtr *tj = derived->bottom_type()->isa_ptr();
  // If its an OOP with a non-zero offset, then it is derived.
  if( tj->_offset == 0 ) {
    derived_base_map[derived->_idx] = derived;
    return derived;
  }
  // Derived is NULL+offset?  Base is NULL!
  if( derived->is_Con() ) {
    Node *base = new (1) ConPNode( TypePtr::NULL_PTR );
    uint no_lidx = 0;  // an unmatched constant in debug info has no LRG
    _names.extend(base->_idx, no_lidx);
    derived_base_map[derived->_idx] = base;
    return base;
  }

  // Check for AddP-related opcodes
  if( !derived->is_Phi() ) {
    assert( derived->is_Mach()->ideal_Opcode() == Op_AddP, "" );
    Node *base = derived->in(AddPNode::Base);
    derived_base_map[derived->_idx] = base;
    return base;
  }

  // Recursively find bases for Phis.
  // First check to see if we can avoid a base Phi here.
  Node *base = find_base_for_derived( derived_base_map, derived->in(1),maxlrg);
  uint i;
  for( i = 2; i < derived->req(); i++ )
    if( base != find_base_for_derived( derived_base_map,derived->in(i),maxlrg))
      break;
  // Went to the end without finding any different bases?
  if( i == derived->req() ) {   // No need for a base Phi here
    derived_base_map[derived->_idx] = base;
    return base;
  }

  // Now we see we need a base-Phi here to merge the bases
  base = new PhiNode( derived->in(0), base->bottom_type() );
  for( i = 1; i < derived->req(); i++ )
    base->set_req(i, find_base_for_derived(derived_base_map,derived->in(i),maxlrg));

  // Search the current block for an existing base-Phi
  Block *b = _cfg._bbs[derived->_idx];
  for( i = 1; i < b->end_idx(); i++ ) {// Search for matching Phi
    Node *phi = b->_nodes[i];
    if( !phi->is_Phi() ) {      // Found end of Phis with no match?
      b->_nodes.insert( i, base ); // Must insert created Phi here as base
      _cfg._bbs.map( base->_idx, b );
      new_lrg(base,maxlrg++);
      break;
    }
    // See if Phi matches.
    uint j;
    for( j = 1; j < base->req(); j++ )
      if( phi->in(j) != base->in(j) &&
	  !(phi->in(j)->is_Con() && base->in(j)->is_Con()) ) // allow different NULLs
        break;
    if( j == base->req() ) {    // All inputs match?
      base = phi;               // Then use existing 'phi' and drop 'base'
      break;
    }
  }


  // Cache info for later passes
  derived_base_map[derived->_idx] = base;
  return base;
}


//------------------------------stretch_base_pointer_live_ranges---------------
// At each Safepoint, insert extra debug edges for each pair of derived value/
// base pointer that is live across the Safepoint for oopmap building.  The
// edge pairs get added in after sfpt->jvmtail()->oopoff(), but are in the
// required edge set.
bool PhaseChaitin::stretch_base_pointer_live_ranges( ResourceArea *a ) {
  int must_recompute_live = false;
  uint maxlrg = _maxlrg;
  Node **derived_base_map = (Node**)a->Amalloc(sizeof(Node*)*C->_unique);
  memset( derived_base_map, 0, sizeof(Node*)*C->_unique );

  // For all blocks in RPO do...
  for( uint i=0; i<_cfg._num_blocks; i++ ) {
    Block *b = _cfg._blocks[i];
    // Note use of deep-copy constructor.  I cannot hammer the original 
    // liveout bits, because they are needed by the following coalesce pass.
    IndexSet liveout(_live->live(b));

    for( uint j = b->end_idx() + 1; j > 1; j-- ) {
      Node *n = b->_nodes[j-1];

      // Pre-split compares of loop-phis.  Loop-phis form a cycle we would
      // like to see in the same register.  Compare uses the loop-phi and so
      // extends its live range BUT cannot be part of the cycle.  If this
      // extended live range overlaps with the update of the loop-phi value
      // we need both alive at the same time -- which requires at least 1
      // copy.  But because Intel has only 2-address registers we end up with
      // at least 2 copies, one before the loop-phi update instruction and
      // one after.  Instead we split the input to the compare just after the
      // phi.
      MachNode *mach = n->is_Mach();
      if( mach && mach->ideal_Opcode() == Op_CmpI ) {
        Node *phi = n->in(1);
        if( phi->is_Phi() && phi->in(0)->is_Loop() ) {
          Block *phi_block = _cfg._bbs[phi->_idx];
          if( _cfg._bbs[phi_block->pred(2)->_idx] == b ) {
            const RegMask *mask = C->_matcher->idealreg2spillmask[Op_RegI];
            Node *spill = get_spillcopy( phi, *mask, *mask );
            insert_proj( phi_block, 1, spill, maxlrg++ );
            n->set_req(1,spill);
            must_recompute_live = true;
          }
        }
      }

      // Get value being defined
      uint lidx = n2lidx(n);
      if( lidx ) {
        // Remove from live-out set
        liveout.remove(lidx);

        // Copies do not define a new value and so do not interfere.
        // Remove the copies source from the liveout set before interfering.
        uint idx = n->is_Copy();
        if( idx ) liveout.remove( n2lidx(n->in(idx)) );
      }

      // Found a safepoint?
      JVMStateObj *jvms = n->jvms();
      if( jvms ) {
        // Now scan for a live derived pointer
        IndexSetIterator elements(&liveout);
        uint neighbor;
        while ((neighbor = elements.next()) != 0) {
          // Find reaching DEF for base and derived values
          // This works because we are still in SSA during this call.
          Node *derived = lrgs(neighbor)._def;
          const TypePtr *tj = derived->bottom_type()->isa_ptr();
          // If its an OOP with a non-zero offset, then it is derived.
          if( tj && tj->_offset != 0 ) {
            Node *base = find_base_for_derived( derived_base_map, derived, maxlrg );
            assert( base->_idx < _names.Size(), "" );
            // Add reaching DEFs of derived pointer and base pointer as a
            // pair of inputs
            n->add_req( derived );
            n->add_req( base );
          
            // See if the base pointer is already live to this point.
            // Since I'm working on the SSA form, live-ness amounts to
            // reaching def's.  So if I find the base's live range then
            // I know the base's def reaches here.
            if( (n2lidx(base) >= _maxlrg ||// (Brand new base (hence not live) or
		 !liveout.member( n2lidx(base) ) ) && // not live) AND
                 _cfg._bbs[base->_idx] != b ) {     //  base not def'd in blk)
              // Base pointer is not currently live.  Since I stretched
              // the base pointer to here and it crosses basic-block
              // boundaries, the global live info is now incorrect.
              // Recompute live.
              must_recompute_live = true;
            } // End of if base pointer is not live to debug info
          }
        } // End of scan all live data for derived ptrs crossing GC point
      } // End of if found a GC point

      // Make all inputs live
      if( !n->is_Phi() ) {      // Phi function uses come from prior block
        for( uint k = 1; k < n->req(); k++ ) {
          uint lidx = n2lidx(n->in(k));
          if( lidx < _maxlrg )
            liveout.insert( lidx );
        }
      }

    } // End of forall instructions in block
    liveout.clear();  // Free the memory used by liveout.

  } // End of forall blocks
  _maxlrg = maxlrg;  

  // If I created a new live range I need to recompute live
  if( maxlrg != _ifg->_maxlrg )
    must_recompute_live = true;

  return must_recompute_live;
}

//------------------------------get_encode-------------------------------------
int PhaseChaitin::get_encode( const Node *n ) const {
  OptoReg::Name  reg    = get_register(n);
  int                    encode = Matcher::_regEncode[reg];

  return encode;
}

//------------------------------set_oop----------------------------------------
void PhaseChaitin::set_oop( const Node *n, bool is_an_oop ) { 
  if( is_an_oop ) { 
    _node_oops.set(n->_idx);
  }
}

//------------------------------is_oop-----------------------------------------
bool PhaseChaitin::is_oop( const Node *n ) const { 
  return _node_oops.test(n->_idx); 
}

//------------------------------set_node_LRG-----------------------------------
// Extend the node to LRG mapping
void PhaseChaitin::set_node_LRG( const Node *node, const Node *old_node ) {
  _names.extend( node->_idx, n2lidx(old_node) );
}

//------------------------------dump-------------------------------------------
#ifndef PRODUCT
void PhaseChaitin::dump( const Node *n ) const {
  uint r = (n->_idx < _names.Size() ) ? Find_const(n) : 0;
  tty->print("L%d",r);
  if( r && n->Opcode() != Op_Phi ) {
    if( _node_regs ) {          // Got a post-allocation copy of allocation?
      OptoReg::Name reg = get_register(n);
      if( reg < SharedInfo::stack0 )
        tty->print("[%s]",Matcher::regName[reg]);
      else
         tty->print("[%s+%d]",SharedInfo::regName[SharedInfo::c_frame_pointer], reg2offset_unchecked(reg));
    } else
    n->out_RegMask().dump();
  }
  tty->print("/N%d\t",n->_idx);
  const MachNode *mach = ((Node*)n)->is_Mach();
  tty->print("%s <== ", n->Name());
  uint k;
  for( k = 0; k < n->req(); k++) {
    Node *m = n->in(k);
    if( !m ) tty->print("_ ");
    else {
      uint r = (m->_idx < _names.Size() ) ? Find_const(m) : 0;
      tty->print("L%d",r);
      if( r && n->Opcode() != Op_Phi ) {
        if( _node_regs ) {
          OptoReg::Name reg = get_register(n->in(k));
          if( reg < SharedInfo::stack0 )
            tty->print("[%s]",Matcher::regName[reg]);
          else
            tty->print("[%s+%d]",SharedInfo::regName[SharedInfo::c_frame_pointer],
                       reg2offset_unchecked(reg));
        } else
          n->in_RegMask(k).dump();
      }
      tty->print("/N%d ",m->_idx);
    }
  }
  if( k < n->len() && n->in(k) ) tty->print("| ");
  for( ; k < n->len(); k++ ) {
    Node *m = n->in(k);
    if( !m ) break;
    uint r = (m->_idx < _names.Size() ) ? Find_const(m) : 0;
    tty->print("L%d",r);
    tty->print("/N%d ",m->_idx);
  }
  if( mach ) mach->dump_spec();
  else n->dump_spec();
  if( _spilled_once.test(n->_idx ) ) {
    tty->print(" Spill_1");
    if( _spilled_twice.test(n->_idx ) ) 
      tty->print(" Spill_2");
  }
  tty->print("\n");
}

void PhaseChaitin::dump( const Block * b ) const {
  b->dump_head( &_cfg._bbs );
  
  // For all instructions
  for( uint j = 0; j < b->_nodes.size(); j++ ) 
    dump(b->_nodes[j]);
  // Print live-out info at end of block
  if( _live ) {
    tty->print("Liveout: ");
    IndexSet *live = _live->live(b);
    IndexSetIterator elements(live);
    tty->print("{");
    uint i;
    while ((i = elements.next()) != 0) {
      tty->print("L%d ", Find_const(i));
    }
    tty->print_cr("}");
  }
  tty->print("\n");
}

void PhaseChaitin::dump() const {
  tty->print( "--- Chaitin -- argsize: %d  framesize: %d ---\n",
              _matcher._new_SP, _framesize );

  // For all blocks
  for( uint i = 0; i < _cfg._num_blocks; i++ ) 
    dump(_cfg._blocks[i]);
  // End of per-block dump
  tty->print("\n");

  if (!_ifg) {
    tty->print("(No IFG.)\n");
    return;
  }

  // Dump LRG array
  tty->print("--- Live RanGe Array ---\n");
  for(uint i2 = 1; i2 < _maxlrg; i2++ ) {
    tty->print("L%d: ",i2);
    if( i2 < _ifg->_maxlrg ) lrgs(i2).dump( );
    else tty->print("new LRG");
  }  
  tty->print_cr("");

  // Dump lo-degree list
  tty->print("Lo degree: ");
  for(uint i3 = _lo_degree; i3; i3 = lrgs(i3)._next )
    tty->print("L%d ",i3);
  tty->print_cr("");

  // Dump lo-stk-degree list
  tty->print("Lo stk degree: ");
  for(uint i4 = _lo_stk_degree; i4; i4 = lrgs(i4)._next )
    tty->print("L%d ",i4);
  tty->print_cr("");

  // Dump lo-degree list
  tty->print("Hi degree: ");
  for(uint i5 = _hi_degree; i5; i5 = lrgs(i5)._next )
    tty->print("L%d ",i5);
  tty->print_cr("");
}

//------------------------------dump_degree_lists------------------------------
void PhaseChaitin::dump_degree_lists() const {
  // Dump lo-degree list
  tty->print("Lo degree: ");
  for( uint i = _lo_degree; i; i = lrgs(i)._next )
    tty->print("L%d ",i);
  tty->print_cr("");

  // Dump lo-stk-degree list
  tty->print("Lo stk degree: ");
  for(uint i2 = _lo_stk_degree; i2; i2 = lrgs(i2)._next )
    tty->print("L%d ",i2);
  tty->print_cr("");

  // Dump lo-degree list
  tty->print("Hi degree: ");
  for(uint i3 = _hi_degree; i3; i3 = lrgs(i3)._next )
    tty->print("L%d ",i3);
  tty->print_cr("");  
}

//------------------------------dump_simplified--------------------------------
void PhaseChaitin::dump_simplified() const {
  tty->print("Simplified: ");
  for( uint i = _simplified; i; i = lrgs(i)._next )
    tty->print("L%d ",i);
  tty->print_cr("");
}

static void print_reg( OptoReg::Name reg, const PhaseChaitin *pc, char *buf ) {
  if( reg < SharedInfo::stack0 )
    strcpy( buf, Matcher::regName[reg] );
  else
    sprintf(buf,"%s + #%d",SharedInfo::regName[SharedInfo::c_frame_pointer],
            pc->reg2offset(reg));
}

//------------------------------dump_register----------------------------------
// Dump a register name into a buffer.  Be intelligent if we get called
// before allocation is complete.
char *PhaseChaitin::dump_register( const Node *n, char *buf  ) const {
  if( !this ) {                 // Not got anything?
    sprintf(buf,"N%d",n->_idx); // Then use Node index
  } else if( _node_regs ) {
      // Post allocation, use direct mappings, no LRG info available
    print_reg( get_register(n), this, buf );
  } else {
    uint lidx = Find_const(n); // Grab LRG number
    if( !_ifg ) {
      sprintf(buf,"L%d",lidx);  // No register binding yet
    } else if( !lidx ) {        // Special, not allocated value
      strcpy(buf,"Special");
    } else if( (lrgs(lidx).num_regs() == 1) 
                ? !lrgs(lidx).mask().is_bound1()
                : !lrgs(lidx).mask().is_bound2() ) {
      sprintf(buf,"L%d",lidx); // No register binding yet
    } else {                    // Hah!  We have a bound machine register
      print_reg( lrgs(lidx).reg(), this, buf );
    }
  }
  return buf+strlen(buf);
}

//------------------------------dump_frame------------------------------------
void PhaseChaitin::dump_frame() const {
  const char *fp = SharedInfo::regName[SharedInfo::c_frame_pointer];
  const TypeTuple *domain = C->_tf->_domain;
  const int argcnt = C->argcnt();

  // Incoming arguments in registers dump
  for( int k = 0; k < argcnt; k++ ) {
    int parmreg = _matcher._parm_regs[k];
    if( parmreg < SharedInfo::stack0 )  {
      const char *reg_name = SharedInfo::regName[parmreg];
      tty->print("#r%3.3d %s   : parm %d: ", parmreg, reg_name, k);
      domain->_fields[k + TypeFunc::Parms]->dump();
      tty->print_cr("");
    }
  }

  // Check for un-owned padding above incoming args
  OptoReg::Name reg = _matcher._new_SP;
  if( reg > _matcher._in_arg_limit ) {
    reg = OptoReg::add(reg, -1);
    tty->print_cr("#r%3.3d %s+%2d: pad0, owned by CALLER", reg, fp, reg2offset_unchecked(reg));
  }

  // Incoming argument area dump
  OptoReg::Name begin_in_arg = OptoReg::add(_matcher._old_SP,C->out_preserve_stack_slots());
  while( reg > begin_in_arg ) {
    reg = OptoReg::add(reg, -1);
    tty->print("#r%3.3d %s+%2d: ",reg,fp,reg2offset_unchecked(reg));
    int j;
    for( j = 0; j < argcnt; j++) {
      if( (int)_matcher._parm_regs[j] == reg ) {
        tty->print("parm %d: ",j);
        domain->_fields[j + TypeFunc::Parms]->dump();
        tty->print_cr("");
        break;
      }
    }
    if( j >= argcnt ) 
      tty->print_cr("HOLE, owned by SELF");
  }

  // Old outgoing preserve area
  while( reg > _matcher._old_SP ) {
    reg = OptoReg::add(reg, -1);
    tty->print_cr("#r%3.3d %s+%2d: old out preserve",reg,fp,reg2offset_unchecked(reg));
  }

  // Old SP
  tty->print_cr("# -- Old %s -- Framesize: %d --",fp,
    reg2offset_unchecked(OptoReg::add(_matcher._old_SP,-1)) - reg2offset_unchecked(_matcher._new_SP)+4);

  // Preserve area dump
  while( reg > SharedInfo::stack0 ) {
    reg = OptoReg::add(reg, -1);
    tty->print("#r%3.3d %s+%2d: ",reg,fp,reg2offset_unchecked(reg));
    if( _matcher.return_addr() == reg )
      tty->print_cr("return address");
    else if( _matcher.return_addr() == OptoReg::add(reg,1) && 
             VerifyStackAtCalls )
      tty->print_cr("0xBADB100D   +VerifyStackAtCalls");
    else if( (int)SharedInfo::reg2stack(reg) < C->_sync*C->sync_stack_slots() )
      tty->print_cr("Lock %d", SharedInfo::reg2stack(reg)/C->sync_stack_slots());
    else
      tty->print_cr("pad2, in_preserve");
  }

  // Spill area dump
  reg = OptoReg::add(_matcher._new_SP, _framesize );
  while( reg > _matcher._out_arg_limit ) {
    reg = OptoReg::add(reg, -1);
    tty->print_cr("#r%3.3d %s+%2d: spill",reg,fp,reg2offset_unchecked(reg));
  }

  // Outgoing argument area dump
  while( reg > OptoReg::add(_matcher._new_SP, C->out_preserve_stack_slots()) ) {
    reg = OptoReg::add(reg, -1);
    tty->print_cr("#r%3.3d %s+%2d: outgoing argument",reg,fp,reg2offset_unchecked(reg));
  }

  // Outgoing new preserve area
  while( reg > _matcher._new_SP ) {
    reg = OptoReg::add(reg, -1);
    tty->print_cr("#r%3.3d %s+%2d: new out preserve",reg,fp,reg2offset_unchecked(reg));
  }
  tty->print_cr("#");
}

//------------------------------dump_bb----------------------------------------
void PhaseChaitin::dump_bb( uint pre_order ) const {
  tty->print_cr("---dump of B%d---",pre_order);
  for( uint i = 0; i < _cfg._num_blocks; i++ ) {
    Block *b = _cfg._blocks[i];
    if( b->_pre_order == pre_order )
      dump(b);
  }
}

//------------------------------dump_lrg---------------------------------------
void PhaseChaitin::dump_lrg( uint lidx ) const {
  tty->print_cr("---dump of L%d---",lidx);

  if( _ifg ) {
    if( lidx >= _maxlrg ) {
      tty->print("Attempt to print live range index beyond max live range.\n");
      return;
    }
    tty->print("L%d: ",lidx);
    lrgs(lidx).dump( );
  }
  if( _ifg ) {    tty->print("Neighbors: %d - ", _ifg->neighbor_cnt(lidx));
    _ifg->neighbors(lidx)->dump();
    tty->cr();
  }
  // For all blocks
  for( uint i = 0; i < _cfg._num_blocks; i++ ) {
    Block *b = _cfg._blocks[i];
    int dump_once = 0;

    // For all instructions
    for( uint j = 0; j < b->_nodes.size(); j++ ) {
      Node *n = b->_nodes[j];
      if( Find_const(n) == lidx ) {
        if( !dump_once++ ) {
          tty->cr();
          b->dump_head( &_cfg._bbs );
        }
        dump(n);
        continue;
      }
      uint cnt = n->req();
      for( uint k = 1; k < cnt; k++ ) {
        Node *m = n->in(k);
        if (!m)  continue;  // be robust in the dumper
        if( Find_const(m) == lidx ) {
          if( !dump_once++ ) {
            tty->cr();
            b->dump_head( &_cfg._bbs );
          }
          dump(n);
        }
      }
    }
  } // End of per-block dump
  tty->cr();
}
#endif // not PRODUCT

//------------------------------print_statistics-------------------------------
int PhaseChaitin::_final_loads  = 0; 
int PhaseChaitin::_final_stores = 0; 
int PhaseChaitin::_final_memoves= 0; 
int PhaseChaitin::_final_copies = 0; 
double PhaseChaitin::_final_load_cost  = 0; 
double PhaseChaitin::_final_store_cost = 0; 
double PhaseChaitin::_final_memove_cost= 0; 
double PhaseChaitin::_final_copy_cost  = 0; 
int PhaseChaitin::_total_framesize = 0;
int PhaseChaitin::_max_framesize = 0; 
int PhaseChaitin::_conserv_coalesce = 0;
int PhaseChaitin::_conserv_coalesce_pair = 0;
int PhaseChaitin::_conserv_coalesce_trie = 0;
int PhaseChaitin::_conserv_coalesce_quad = 0;
int PhaseChaitin::_post_alloc = 0;
int PhaseChaitin::_lost_opp_pp_coalesce = 0;
int PhaseChaitin::_lost_opp_cflow_coalesce = 0;
int PhaseChaitin::_used_cisc_instructions   = 0;
int PhaseChaitin::_unused_cisc_instructions = 0;
int PhaseChaitin::_allocator_attempts       = 0;
int PhaseChaitin::_allocator_successes      = 0;
uint PhaseChaitin::_high_pressure           = 0;
uint PhaseChaitin::_low_pressure            = 0;

#ifndef PRODUCT
void PhaseChaitin::print_statistics() {
  tty->print_cr("Inserted %d spill loads, %d spill stores, %d mem-mem moves and %d copies.", _final_loads, _final_stores, _final_memoves, _final_copies);
  tty->print_cr("Total load cost= %6.0f, store cost = %6.0f, mem-mem cost = %5.2f, copy cost = %5.0f.", _final_load_cost, _final_store_cost, _final_memove_cost, _final_copy_cost);
  tty->print_cr("Total frameslots = %d, Max frameslots = %d, adjusted spill cost = %7.0f.", 
                _total_framesize, _max_framesize,
                _final_load_cost*4.0 + _final_store_cost  * 2.0 + 
                _final_copy_cost*1.0 + _final_memove_cost*12.0);
  tty->print("Conservatively coalesced %d copies, %d pairs", 
                _conserv_coalesce, _conserv_coalesce_pair);
  if( _conserv_coalesce_trie || _conserv_coalesce_quad )
    tty->print(", %d tries, %d quads", _conserv_coalesce_trie, _conserv_coalesce_quad);
  tty->print_cr(", %d post alloc.", _post_alloc);
  if( _lost_opp_pp_coalesce || _lost_opp_cflow_coalesce )
    tty->print_cr("Lost coalesce opportunity, %d private-private, and %d cflow interfered.",
                  _lost_opp_pp_coalesce, _lost_opp_cflow_coalesce );
  if( _used_cisc_instructions || _unused_cisc_instructions )
    tty->print_cr("Used cisc instruction  %d,  remained in register %d",
                   _used_cisc_instructions, _unused_cisc_instructions);
  if( _allocator_successes != 0 )
    tty->print_cr("Average allocation trips %f", (float)_allocator_attempts/(float)_allocator_successes);
  tty->print_cr("High Pressure Blocks = %d, Low Pressure Blocks = %d", _high_pressure, _low_pressure);
}
#endif // not PRODUCT
