/*
 * @(#)connode.cpp	1.143 01/02/16 10:59:48
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// Optimization - Graph Style

#include "incls/_precompiled.incl"
#include "incls/_connode.cpp.incl"

//=============================================================================
//------------------------------hash-------------------------------------------
uint ConNode::hash() const {
  return (uint)in(TypeFunc::Control) + _type->hash();
}

//------------------------------make-------------------------------------------
ConNode *ConNode::make( const Type *t ) {
  switch( t->_base ) {
  case Type::Int:       return new ConINode( t->is_int() );
  case Type::Long:      return new ConLNode( t->is_long() );
  case Type::FloatCon:  return new ConFNode( t->is_float_constant() );
  case Type::DoubleCon: return new ConDNode( t->is_double_constant() );
  case Type::AnyPtr:
  case Type::RawPtr:
  case Type::AryPtr:
  case Type::KlassPtr:
  case Type::OopPtr:    return new ConPNode( t->isa_ptr() );
  default:              return new ConNode ( t );
  }
  return NULL;
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *CopyNode::Value( PhaseTransform *phase ) const {
  return phase->type( _in[1] );
}

//------------------------------Identity---------------------------------------
Node *CopyNode::Identity( PhaseTransform *phase ) {
  return in(1);
}

//------------------------------make-------------------------------------------
Node *CopyNode::make( Node *n, const Type *t ) {
  switch( t->_base ) {
  case Type::Int:       
    return t->is_int()->is_con() 
      ? (Node*)new ConINode( t->is_int() ) 
      : (Node*)new CopyINode( n );
  case Type::AnyPtr:
  case Type::RawPtr:
  case Type::AryPtr:
  case Type::KlassPtr:  // fall through
  case Type::OopPtr:    return new CopyPNode( n );
  case Type::FloatCon:  return new ConFNode( t->is_float_constant() );
  case Type::FloatBot : return new CopyFNode( n );
  case Type::DoubleBot: return new CopyDNode( n );
  case Type::DoubleCon: return new ConDNode( t->is_double_constant() );
  default:              return new CopyNode(n,t);
  }
  return NULL;
}

//=============================================================================
/*
The major change is for CMoveP and StrComp.  They have related but slightly
different problems.  They both take in TWO oops which are both null-checked
independently before the using Node.  After CCP removes the CastPP's they need
to pick up the guarding test edge - in this case TWO control edges.  I tried
various solutions, all have problems:

(1) Do nothing.  This leads to a bug where we hoist a Load from a CMoveP or a
StrComp above a guarding null check.  I've seen both cases in normal -Xcomp
testing.

(2) Plug the control edge from 1 of the 2 oops in.  Apparent problem here is
to figure out which test post-dominates.  The real problem is that it doesn't
matter which one you pick.  After you pick up, the dominating-test elider in
IGVN can remove the test and allow you to hoist up to the dominating test on
the choosen oop bypassing the test on the not-choosen oop.  Seen in testing.
Oops.

(3) Leave the CastPP's in.  This makes the graph more accurate in some sense;
we get to keep around the knowledge that an oop is not-null after some test.
Alas, the CastPP's interfere with GVN (some values are the regular oop, some
are the CastPP of the oop, all merge at Phi's which cannot collapse, etc).
This cost us 10% on SpecJVM, even when I removed some of the more trivial
cases in the optimizer.  Removing more useless Phi's started allowing Loads to 
illegally float above null checks.  I gave up on this approach.

(4) Add BOTH control edges to both tests.  Alas, too much code knows that
control edges are in slot-zero ONLY.  Many quick asserts fail; no way to do
this one.  Note that I really want to allow the CMoveP to float and add both
control edges to the dependent Load op - meaning I can select early but I
cannot Load until I pass both tests.

(5) Do not hoist CMoveP and StrComp.  To this end I added the v-call
depends_only_on_test().  No obvious performance loss on Spec, but we are
clearly conservative on CMoveP (also so on StrComp but that's unlikely to
matter ever).

*/


//------------------------------Ideal------------------------------------------
// Return a node which is more "ideal" than the current node.  
// Move constants to the right.
Node *CMoveNode::Ideal(PhaseGVN *phase, PhaseDefUse *du ) {
  if( remove_dead_region(phase,du) ) return this;
  if( in(IfFalse)->is_Con() && !in(IfTrue)->is_Con() ) {
    BoolNode *b = in(Condition)->is_Bool();
    if( b ) {
      Node *b2 = phase->transform(new (2) BoolNode( b->in(1), b->_test.negate() ));
      return make( in(Control), b2, in(IfTrue), in(IfFalse), _type );
    }
  }
  return NULL;
}

//------------------------------Identity---------------------------------------
// Conditional-move is an identity if both inputs are the same, or the test
// true or false.
Node *CMoveNode::Identity( PhaseTransform *phase ) {
  if( phase->eqv(in(IfFalse),in(IfTrue)) ) // C-moving identical inputs?
    return in(IfFalse);		// Then it doesn't matter
  if( phase->type(in(1)) == TypeInt::ZERO ) 
    return in(IfFalse);		// Always pick left(false) input
  if( phase->type(in(1)) == TypeInt::ONE ) 
    return in(IfTrue);		// Always pick right(true) input
  return this;
}

//------------------------------Value------------------------------------------
// Result is the meet of inputs
const Type *CMoveNode::Value( PhaseTransform *phase ) const {
  return phase->type(in(IfFalse))->meet(phase->type(in(IfTrue)));
}

//------------------------------make-------------------------------------------
// Make a correctly-flavored CMove.  Since _type is directly determined
// from the inputs we do not need to specify it here.
CMoveNode *CMoveNode::make( Node *c, Node *bol, Node *left, Node *right, const Type *t ) {
  switch( t->_base ) {
  case Type::DoubleBot: return new CMoveDNode( bol, left, right );
  case Type::FloatBot : return new CMoveFNode( bol, left, right );
  case Type::Int:       return new CMoveINode( bol, left, right );
  case Type::Long:      return new CMoveLNode( bol, left, right );
  case Type::AnyPtr:
  case Type::RawPtr:
  case Type::AryPtr:
  case Type::KlassPtr:  // fall through
  case Type::OopPtr:    return new CMovePNode( c, bol, left, right );
  default:		break;
  }
  ShouldNotReachHere();
  return NULL;
}

//=============================================================================
//------------------------------Ideal------------------------------------------
// Return a node which is more "ideal" than the current node.  
// Check for conversions to boolean
Node *CMoveINode::Ideal(PhaseGVN *phase, PhaseDefUse *du ) {
  // Try generic ideal's first
  Node *x = CMoveNode::Ideal(phase,du);
  if( x ) return x;

  // Now check for booleans
  int flip = 0;

  // Check for picking from zero/one
  if( phase->type(in(IfFalse)) == TypeInt::ZERO && phase->type(in(IfTrue)) == TypeInt::ONE ) {
    flip = 1 - flip;
  } else if( phase->type(in(IfFalse)) == TypeInt::ONE && phase->type(in(IfTrue)) == TypeInt::ZERO ) {
  } else return NULL;

  // Check for eq/ne test
  BoolNode *bol = in(1)->is_Bool();
  if( !bol ) return NULL;
  if( bol->_test._test == BoolTest::eq ) {
  } else if( bol->_test._test == BoolTest::ne ) {
    flip = 1-flip;
  } else return NULL;

  // Check for vs 0 or 1
  const SubNode *cmp = bol->in(1)->is_Cmp();
  if( !cmp ) return NULL;
  if( phase->type(cmp->in(2)) == TypeInt::ZERO ) {
  } else if( phase->type(cmp->in(2)) == TypeInt::ONE ) {
    // Allow cmp-vs-1 if the other input is bounded by 0-1
    if( phase->type(cmp->in(1)) != TypeInt::BOOL )
      return NULL;
    flip = 1 - flip;
  } else return NULL;
  
  // Convert to a bool (flipped)
  // Build int->bool conversion
#ifndef PRODUCT
  if( PrintOpto ) tty->print_cr("CMOV to I2B");
#endif
  Node *n = new (2) Conv2BNode( cmp->in(1) );
  if( flip ) 
    n = new (3) XorINode( phase->transform(n), phase->intcon(1) );

  return n;
}

//=============================================================================
//------------------------------Ideal------------------------------------------
// Return a node which is more "ideal" than the current node.  
// Check for absolute value
Node *CMoveFNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  // Try generic ideal's first
  Node *x = CMoveNode::Ideal(phase,du);
  if( x ) return x;

  // Abs or -Abs?  
  // Might be an issue with -0.0 vs +0.0
  int flip = 0;

  // Find the Bool
  BoolNode *bol = in(1)->is_Bool();
  if( !bol ) return NULL;
  // Check bool sense
  switch( bol->_test._test ) {
  case BoolTest::eq:
  case BoolTest::ne:
    return NULL;
  case BoolTest::gt:
  case BoolTest::ge:
    flip = 1-flip;
  case BoolTest::lt:
  case BoolTest::le:
    break;
  }

  // Check compare
  Node *cmp = bol->in(1);
  if( !cmp->is_Cmp() ) return NULL;
  // Check compare sense
  if( phase->type(cmp->in(2)) != TypeInt::ZERO ) return NULL;

  // Find CmpF vs zero
  Node *cmpf = cmp->in(1);
  if( !cmpf->is_Cmp() ) return NULL;
  if( phase->type(cmpf->in(2)) != TypeF::ZERO ) return NULL;
  Node *X = cmpf->in(1);

  // Find CmpF vs zero
  Node *neg;
  if( X == in(IfFalse) ) {
    neg = in(IfTrue);
  } else if( X == in(IfTrue) ) {
    flip = 1-flip;
    neg = in(IfFalse);
  } else return NULL;

  if( neg->Opcode() != Op_NegF ) return NULL;
  if( neg->in(1) != X ) return NULL;

  Node *abs = new (2) AbsFNode( X );
  if( flip )
    abs = new (2) NegFNode( phase->transform(abs) );

  return abs;
}

//=============================================================================
//------------------------------Ideal------------------------------------------
// Return a node which is more "ideal" than the current node.  
// Check for absolute value
Node *CMoveDNode::Ideal(PhaseGVN *phase, PhaseDefUse *du ) {
  // Try generic ideal's first
  Node *x = CMoveNode::Ideal(phase,du);
  if( x ) return x;

  // Abs or -Abs?
  // Might be an issue with -0.0 vs +0.0
  int flip = 0;

  // Find the Bool
  BoolNode *bol = in(1)->is_Bool();
  if( !bol ) return NULL;
  // Check bool sense
  switch( bol->_test._test ) {
  case BoolTest::eq:
  case BoolTest::ne:
    return NULL;
  case BoolTest::gt:
  case BoolTest::ge:
    flip = 1-flip;
  case BoolTest::lt:
  case BoolTest::le:
    break;
  }

  // Check compare
  Node *cmp = bol->in(1);
  if( !cmp->is_Cmp() ) return NULL;
  // Check compare sense
  if( phase->type(cmp->in(2)) != TypeInt::ZERO ) return NULL;

  // Find CmpD vs zero
  Node *cmpd = cmp->in(1);
  if( !cmpd->is_Cmp() ) return NULL;
  if( phase->type(cmpd->in(2)) != TypeD::ZERO ) return NULL;
  Node *X = cmpd->in(1);

  // Find CmpD vs zero
  Node *neg;
  if( X == in(IfFalse) ) {
    neg = in(IfTrue);
  } else if( X == in(IfTrue) ) {
    flip = 1-flip;
    neg = in(IfFalse);
  } else return NULL;

  if( neg->Opcode() != Op_NegD ) return NULL;
  if( neg->in(1) != X ) return NULL;

  Node *abs = new (2) AbsDNode( X );
  if( flip )
    abs = new (2) NegDNode( phase->transform(abs) );

  return abs;
}

//=============================================================================
// If input is already higher or equal to cast type, then this is an identity.
Node *CastPPNode::Identity( PhaseTransform *phase ) {
  return phase->type(_in[1])->higher_equal(_type) ? _in[1] : this;
}

//------------------------------Value------------------------------------------
// Take 'join' of input and cast-up type
const Type *CastPPNode::Value( PhaseTransform *phase ) const {
  if( in(0) && phase->type(in(0)) == Type::TOP ) return Type::TOP;
  if( phase->type(_in[1]) == TypePtr::NULL_PTR &&
      _type->isa_ptr() && _type->is_ptr()->_ptr == TypePtr::NotNull )
    return Type::TOP;
  return phase->type(_in[1])->join(_type);
}

//------------------------------Ideal------------------------------------------
// Return a node which is more "ideal" than the current node.  Strip out 
// control copies
Node *CastPPNode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  return (in(0) && remove_dead_region(phase,du)) ? this : NULL; 
}

//------------------------------Ideal_DU_postCCP-------------------------------
// If not converting int->oop, throw away cast after constant propagation
Node *CastPPNode::Ideal_DU_postCCP( PhaseCCP *ccp ) {
  const Type *t = ccp->type(in(1));
  if( t->isa_oop_ptr() ) {
    ccp->hash_delete(this);
    _type = t;   // Turn into ID function
    ccp->hash_insert(this);
    return this;
  }
  return NULL;                     // No progress
}



//=============================================================================
uint CheckCastPPNode::size_of() const { return sizeof(*this); }
uint CheckCastPPNode::hash() const { return Node::hash() + _type->hash(); }
uint CheckCastPPNode::cmp( const Node &n ) const
{ return !Type::cmp( _type, ((TypeNode&)n)._type ); }
#ifndef PRODUCT
void CheckCastPPNode::dump_spec() const { 
  if( !Verbose && !WizardMode ) {
    // standard dump does this in Verbose and WizardMode
    tty->print(" #"); _type->dump();
  }
}
#endif

//------------------------------Identity---------------------------------------
// If input is already higher or equal to cast type, then this is an identity.
Node *CheckCastPPNode::Identity( PhaseTransform *phase ) {
  // Toned down to rescue meeting at a Phi 3 different oops all implementing
  // the same interface.  CompileTheWorld starting at 502, kd12rc1.zip.
  return (phase->type(in(1)) == phase->type(this)) ? in(1) : this;
}

//------------------------------Value------------------------------------------
// Take 'join' of input and cast-up type, unless working with an Interface
const Type *CheckCastPPNode::Value( PhaseTransform *phase ) const {
  if( in(0) && phase->type(in(0)) == Type::TOP ) return Type::TOP;

  const Type *in = phase->type(_in[1]);
  if( in == Type::TOP ) return Type::TOP;  // No information yet

  const TypePtr *in_type   = in->isa_ptr();
  const TypePtr *my_type   = _type->isa_ptr();
  if( in_type != NULL && my_type != NULL ) {
    TypePtr::PTR   in_ptr    = in_type->_ptr;
    if( in_ptr == TypePtr::Null ) {
      return in_type;
    } else if( in_ptr == TypePtr::Constant ) {
      // Casting a constant oop to an interface?
      // (i.e., a String to a Comparable?)
      // Then return the interface.
      const TypeJavaPtr *jptr = my_type->isa_javaptr();
      assert( jptr, "" );
      return jptr->klass()->is_interface() 
	? my_type->cast_to_ptr_type( TypePtr::NotNull )
	: in_type;
    } else {
      return my_type->cast_to_ptr_type( in_ptr );
    }
  }
  return _type;

  // 
  // Remove this code after overnight run indicates no performance
  // loss from not performing JOIN at CheckCastPPNode
  // 
  // const TypeOopPtr *in_oop = in->isa_oopptr();
  // const TypeOopPtr *my_oop = _type->isa_oopptr();
  // // If either input is an 'interface', return destination type
  // assert (in_oop == NULL || in_oop->klass() != NULL, "");
  // assert (my_oop == NULL || my_oop->klass() != NULL, "");
  // if( (in_oop && in_oop->klass()->klass_part()->is_interface())
  //   ||(my_oop && my_oop->klass()->klass_part()->is_interface()) ) {
  //   TypePtr::PTR  in_ptr = in->isa_ptr() ? in->is_ptr()->_ptr : TypePtr::BotPTR;
  //   // Preserve cast away nullness for interfaces
  //   if( in_ptr == TypePtr::NotNull && my_oop && my_oop->_ptr == TypePtr::BotPTR ) {
  //     return my_oop->cast_to_ptr_type(TypePtr::NotNull);
  //   }
  //   return _type;
  // }
  // 
  // // Neither the input nor the destination type is an interface, 
  // 
  // // history: JOIN used to cause weird corner case bugs
  // //          return (in == TypeJavaPtr::NULL_PTR) ? in : _type;
  // // JOIN picks up NotNull in common instance-of/check-cast idioms, both oops.
  // // JOIN does not preserve NotNull in other cases, e.g. RawPtr vs OopPtr
  // const Type *join = in->join(_type);
  // // Check if join preserved NotNull'ness for pointers
  // if( join->isa_ptr() && _type->isa_ptr() ) {
  //   TypePtr::PTR join_ptr = join->is_ptr()->_ptr;
  //   TypePtr::PTR type_ptr = _type->is_ptr()->_ptr;
  //   // If there isn't any NotNull'ness to preserve
  //   // OR if join preserved NotNull'ness then return it
  //   if( type_ptr == TypePtr::BotPTR  || type_ptr == TypePtr::Null ||
  //       join_ptr == TypePtr::NotNull || join_ptr == TypePtr::Constant ) {
  //     return join;
  //   }
  //   // ELSE return same old type as before 
  //   return _type;
  // }
  // // Not joining two pointers
  // return join;
}

//------------------------------Ideal------------------------------------------
// Return a node which is more "ideal" than the current node.  Strip out 
// control copies
Node *CheckCastPPNode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  return (in(0) && remove_dead_region(phase,du)) ? this : NULL; 
}

//=============================================================================
//------------------------------Identity---------------------------------------
Node *Conv2BNode::Identity( PhaseTransform *phase ) {
  const Type *t = phase->type( _in[1] );
  if( t == Type::TOP ) return _in[1];
  if( t == TypeInt::ZERO ) return _in[1];
  if( t == TypeInt::ONE ) return _in[1];
  if( t == TypeInt::BOOL ) return _in[1];

  return this;
}

//------------------------------Value------------------------------------------
const Type *Conv2BNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( _in[1] );
  if( t == Type::TOP ) return Type::TOP;
  if( t == TypeInt::ZERO ) return TypeInt::ZERO;
  if( t == TypePtr::NULL_PTR ) return TypeInt::ZERO;
  const TypePtr *tp = t->isa_ptr();
  if( tp ) {
    if( tp->_ptr == TypePtr::AnyNull ) return Type::TOP;
    if( tp->_ptr == TypePtr::Constant) return TypeInt::ONE;
    if( tp->_ptr == TypePtr::NotNull ) return TypeInt::ONE;
    return TypeInt::BOOL;
  }
  if( t->_base != Type::Int ) return TypeInt::BOOL;
  const TypeInt *ti = t->is_int();
  if( ti->_hi < 0 || ti->_lo > 0 ) return TypeInt::ONE;
  return TypeInt::BOOL;
}


//=============================================================================
// The conversions operations are all Alpha sorted.  Please keep it that way!
//------------------------------Value------------------------------------------
const Type *ConvD2FNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( in(1) );
  if( t == Type::TOP ) return Type::TOP;
  if( t == Type::DOUBLE ) return Type::FLOAT;
  const TypeD *td = t->is_double_constant();
  return TypeF::make( (float)td->getd() );
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvD2INode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( in(1) );
  if( t == Type::TOP ) return Type::TOP;
  if( t == Type::DOUBLE ) return TypeInt::INT;
  const TypeD *td = t->is_double_constant();
  return TypeInt::make( SharedRuntime::d2i( td->getd() ) );
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvD2LNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( in(1) );
  if( t == Type::TOP ) return Type::TOP;
  if( t == Type::DOUBLE ) return TypeLong::LONG;
  const TypeD *td = t->is_double_constant();
  return TypeLong::make( SharedRuntime::d2l( td->getd() ) );
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvF2DNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( in(1) );
  if( t == Type::TOP ) return Type::TOP;
  if( t == Type::FLOAT ) return Type::DOUBLE;
  const TypeF *tf = t->is_float_constant();
  return TypeD::make( (double)tf->getf() );
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvF2INode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( _in[1] );
  if( t == Type::TOP )       return Type::TOP;
  if( t == Type::FLOAT ) return TypeInt::INT;
  const TypeF *tf = t->is_float_constant();
  return TypeInt::make( SharedRuntime::f2i( tf->getf() ) );
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvF2LNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( _in[1] );
  if( t == Type::TOP )       return Type::TOP;
  if( t == Type::FLOAT ) return TypeLong::LONG;
  const TypeF *tf = t->is_float_constant();
  return TypeLong::make( SharedRuntime::f2l( tf->getf() ) );
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvI2DNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( in(1) );
  if( t == Type::TOP ) return Type::TOP;
  const TypeInt *ti = t->is_int();
  if( ti->is_con() ) return TypeD::make( (double)ti->get_con() );
  return bottom_type();
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvI2FNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( in(1) );
  if( t == Type::TOP ) return Type::TOP;
  const TypeInt *ti = t->is_int();
  if( ti->is_con() ) return TypeF::make( (float)ti->get_con() );
  return bottom_type();
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvI2LNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( in(1) );
  if( t == Type::TOP ) return Type::TOP;
  const TypeInt *ti = t->is_int();
  if( ti->is_con() ) return TypeLong::make( (jlong)ti->get_con() );
  return bottom_type();
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvL2DNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( in(1) );
  if( t == Type::TOP ) return Type::TOP;
  const TypeLong *tl = t->is_long();
  if( tl->is_con() ) return TypeD::make( (double)tl->get_con() );
  return bottom_type();
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvL2FNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( in(1) );
  if( t == Type::TOP ) return Type::TOP;
  const TypeLong *tl = t->is_long();
  if( tl->is_con() ) return TypeF::make( (float)tl->get_con() );
  return bottom_type();
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *ConvL2INode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( in(1) );
  if( t == Type::TOP ) return Type::TOP;
  const TypeLong *tl = t->is_long();
  if( tl->is_con() ) return TypeInt::make( (jint)tl->get_con() );
  return bottom_type();
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *DoubleHalfNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( _in[1] );
  if( t == Type::TOP )       return Type::TOP;
  if( t == TypeD::ZERO )     return TypeInt::ZERO;
  // Unimplemented
  return bottom_type();
}

const RegMask &DoubleHalfNode::out_RegMask() const {
  return *Matcher::idealreg2regmask[Op_RegF];
}
const RegMask &DoubleHalfNode::in_RegMask(uint) const {
  return *Matcher::idealreg2regmask[Op_RegD];
}

//=============================================================================
//------------------------------Value------------------------------------------
const Type *LongHalfNode::Value( PhaseTransform *phase ) const {
  const Type *t = phase->type( _in[1] );
  if( t == Type::TOP )       return Type::TOP;
  if( t == TypeLong::ZERO )  return TypeInt::ZERO;
  // Unimplemented
  return bottom_type();
}

const RegMask &LongHalfNode::out_RegMask() const {
  return *Matcher::idealreg2regmask[Op_RegI];
}
const RegMask &LongHalfNode::in_RegMask(uint) const {
  return *Matcher::idealreg2regmask[Op_RegL];
}

//=============================================================================

#ifdef ASSERT
inline bool verify_output_regclass(const RegMask& rm, const Type* t) {
  // Must special-case this semi-invisible register:
  if (t->_base == Type::Return_Address && rm.Member(Compile::current()->_matcher->return_addr()))
    return true;

  uint mreg = Matcher::base2reg[t->_base];
  if (Matcher::idealreg2regmask[mreg]->overlap(rm))
    return true;
  tty->print("ERROR: spill copies type ");
  t->dump();
  tty->print(" but must produce result in register ");
  rm.dump();
  tty->cr();
  return false;
}
#endif

//------------------------------copy_make--------------------------------------
Node *SpillCopyNode::copy_make( Node *src ) const {
  assert(verify_output_regclass(*_out, _type), "bad regclass in copy");
  switch( _type->_base ) {
  case Type::FloatCon:
    if ( Matcher::rematerialize_float_constants )
      return new (1) ConFNode( _type->is_float_constant() );
  // Fall into FloatBot case
  case Type::FloatBot:  return new (2) CopyFNode( src );

  case Type::DoubleCon:  
    if ( Matcher::rematerialize_float_constants )
      return new (1) ConDNode( _type->is_double_constant() );
  // Fall into DoubleBot case
  case Type::DoubleBot: return new (2) CopyDNode( src );

  case Type::Int:       return new (2) CopyINode( src );
  case Type::Long:      return new (2) CopyLNode( src );

  case Type::KlassPtr:
  case Type::Return_Address:
  case Type::AnyPtr:
  case Type::RawPtr:
  case Type::OopPtr:
  case Type::AryPtr:    return new (2) CopyPNode( src );
  default:              Unimplemented(); 
  }
  return NULL;
}

//------------------------------load_make--------------------------------------
Node *SpillCopyNode::load_make( Node *mem, Node *stk ) const {
  assert(verify_output_regclass(*_out, _type), "bad regclass in copy");
  switch( _type->_base ) {
  case Type::FloatCon:
    if ( Matcher::rematerialize_float_constants )
      return new (1) ConFNode( _type->is_float_constant() );
  // Fall into FloatBot case
  case Type::FloatBot:  return (Node*)new (3) LoadFNode( NULL, mem, stk );

  case Type::DoubleCon:  
    assert(Matcher::idealreg2regmask[Op_RegD]->overlap(out_RegMask()), "bad regclass in copy");
    if( Matcher::rematerialize_float_constants )
      return new (1) ConDNode( _type->is_double_constant() );
    // Fall into DoubleBot case
  case Type::DoubleBot: return new (3) LoadDNode( NULL, mem, stk );

  case Type::Int:       return new (3) LoadINode( NULL, mem, stk );
  case Type::Long:      return new (3) LoadLNode( NULL, mem, stk );

  case Type::AnyPtr: 
  case Type::RawPtr: 
  case Type::OopPtr:  
  case Type::AryPtr:
  case Type::KlassPtr:
  case Type::Return_Address:
    return new (3) LoadPNode( NULL, mem, stk, _type->isa_ptr() );

  default:              Unimplemented();
  }
  return NULL;
}

//------------------------------store_make-------------------------------------
Node *SpillCopyNode::store_make( Node *mem, Node *stk, Node *val ) const {
  switch( _type->_base ) {
  case Type::FloatBot:
  case Type::FloatCon: 
    return new (4) StoreFNode(NULL, mem, stk, val, Type::MEMORY);
  case Type::DoubleBot:
  case Type::DoubleCon:
    return new (4) StoreDNode(NULL, mem, stk, val, Type::MEMORY);
  case Type::Int:      
    return new (4) StoreINode(NULL, mem, stk, val, Type::MEMORY);
  case Type::Long:     
    return new (4) StoreLNode(NULL, mem, stk, val, Type::MEMORY);
  case Type::AnyPtr: 
  case Type::RawPtr: 
  case Type::OopPtr: 
  case Type::AryPtr: 
  case Type::KlassPtr:
  case Type::Return_Address:
    return new (4) StorePNode(NULL, mem, stk, val, Type::MEMORY);
  default:            Unimplemented();
  }
  return NULL;
}

//=============================================================================
//------------------------------cmp--------------------------------------------
// Do not common AtomicAdds together.  AtomicAdd modifies machine state
// and would have to be split along control paths
uint AtomicAddNode::cmp( const Node &n ) const { 
  return (&n == this);          // Always fail except on self
}

//------------------------------match_edge-------------------------------------
// Do we Match on this edge index or not?  Do not match memory
uint AtomicAddNode::match_edge(uint idx) const {
  return idx > 1;
}

//------------------------------Ideal------------------------------------------
// Return a node which is more "ideal" than the current node.  Strip out 
// control copies
Node *AtomicAddNode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  return remove_dead_region(phase,du) ? this : NULL; 
}

//=============================================================================
const RegMask &BoxNode::in_RegMask(uint i) const {
  return Compile::current()->FIRST_STACK_mask;
}

const RegMask &BoxNode::out_RegMask() const { 
  return *Matcher::idealreg2regmask[Op_RegP];
}


//=============================================================================
//------------------------------Identity---------------------------------------
// Remove redundant roundings
Node *RoundFloatNode::Identity( PhaseTransform *phase ) {
  assert(Matcher::strict_fp_requires_explicit_rounding, "should only generate for Intel");
  // Do not round constants
  if( phase->type(in(1))->_base == Type::FloatCon ) return in(1);
  int op = in(1)->Opcode();
  // Redundant rounding
  if( op == Op_RoundFloat ) return in(1);
  // Already rounded
  if( op == Op_Parm ) return in(1);
  if( op == Op_LoadF ) return in(1);
  return this;
}

//------------------------------Value------------------------------------------
const Type *RoundFloatNode::Value( PhaseTransform *phase ) const {
  return phase->type( _in[1] );
}

//=============================================================================
//------------------------------Identity---------------------------------------
// Remove redundant roundings.  Incoming arguments are already rounded.
Node *RoundDoubleNode::Identity( PhaseTransform *phase ) {
  assert(Matcher::strict_fp_requires_explicit_rounding, "should only generate for Intel");
  // Do not round constants
  if( phase->type(in(1))->_base == Type::DoubleCon ) return in(1);
  int op = in(1)->Opcode();
  // Redundant rounding
  if( op == Op_RoundDouble ) return in(1);
  // Already rounded
  if( op == Op_Parm ) return in(1);
  if( op == Op_LoadD ) return in(1);
  if( op == Op_ConvF2D ) return in(1);
  return this;
}

//------------------------------Value------------------------------------------
const Type *RoundDoubleNode::Value( PhaseTransform *phase ) const {
  return phase->type( _in[1] );
}

//=============================================================================
// Do not allow value-numbering
uint Opaque1Node::cmp( const Node &n ) const { 
  return (&n == this);          // Always fail except on self
}

//------------------------------Identity---------------------------------------
// If _major_progress, then more loop optimizations follow.  Do NOT remove
// the opaque Node until no more loop ops can happen.  Note the timing of
// _major_progress; it's set in the major loop optimizations THEN comes the
// call to IterGVN and any chance of hitting this code.  Hence there's no
// phase-ordering problem with stripping Opaque1 in IGVN followed by some
// more loop optimizations that require it.
Node *Opaque1Node::Identity( PhaseTransform *phase ) {
  return phase->C->_major_progress ? this : in(1);
}

//=============================================================================
// A node to prevent unwanted optimizations.  Allows constant folding.  Stops
// value-numbering, most Ideal calls or Identity functions.  This Node is
// specifically designed to prevent the pre-increment value of a loop trip
// counter from being live out of the bottom of the loop (hence causing the
// pre- and post-increment values both being live and thus requiring an extra
// temp register and an extra move).  If we "accidentally" optimize through
// this kind of a Node, we'll get slightly pessimal, but correct, code.  Thus
// it's OK to be slightly sloppy on optimizations here.

// Do not allow value-numbering
uint Opaque2Node::cmp( const Node &n ) const { 
  return (&n == this);          // Always fail except on self
}
