/*
 * @(#)divnode.cpp	1.46 01/02/09 14:46:12
 *
 * Copyright 1991-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
// Portions of code courtesy of Clifford Click

// Optimization - Graph Style

#include "incls/_precompiled.incl"
#include "incls/_divnode.cpp.incl"
#include <math.h>

//=============================================================================
//------------------------------Identity---------------------------------------
// If the divisor is 1, we are an identity on the dividend.
Node *DivINode::Identity( PhaseTransform *phase ) {
  return (phase->type( _in[2] )->higher_equal(TypeInt::ONE)) ? _in[1] : this;
}

//------------------------------Idealize---------------------------------------
// Dividing by a power of 2 is a shift.
Node *DivINode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  if( in(0) && remove_dead_region(phase,du) ) return this;

  const Type *t = phase->type( _in[2] );
  if( t == TypeInt::ONE )       // Identity?
    return NULL;		// Skip it

  const TypeInt *ti = t->isa_int();
  if( !ti ) return NULL;
  if( !ti->is_con() ) return NULL;
  int i = ti->get_con();	// Get divisor
  if( i ) set_req(0,NULL,du);	// Dividing by a not-zero constant; no faulting

  // Dividing by MININT does not optimize as a power-of-2 shift.
  if( i == MININT32 ) return NULL;

  // Check for negative power of 2
  Node *dividend = _in[1];
  if( (-i & ~(-i-1)) == -i ) {
    i = -i;			// Flip divisor
    dividend = phase->transform(new (3) SubINode(phase->intcon(0),_in[1]));
  }

  // Check for power of 2
  if( (i & ~(i-1)) != i )       // Is divisor a power of 2?
    return NULL;		// Not a power of 2

  // Generate number of bits to shift
  Node *log_i = phase->intcon(log2(i));

  // An AND mask of sufficient size clears the low bits and
  // I can avoid rounding.
  if( dividend->Opcode() == Op_AndI ) {
    const TypeInt *andconi = phase->type( dividend->in(2) )->isa_int();
    if( andconi &&
        andconi->is_con() &&
        andconi->get_con() == -i ) {
      return new (3) RShiftINode(dividend->in(1), log_i);
    }
  }

  // Divide-by-power-of-2 can be made into a shift, but you have to do
  // more math for the rounding.  You need to add 0 for positive
  // numbers, and "i-1" for negative numbers.  Example: i=4, so the
  // shift is by 2.  You need to add 3 to negative dividends and 0 to
  // positive ones.  So (-7+3)>>2 becomes -1, (-4+3)>>2 becomes -1,
  // (-2+3)>>2 becomes 0, etc.
  Node *hook = new (3) Node( NULL, NULL, NULL );
  hook->set_req(1,dividend,du);
  hook->set_req(2,log_i,du);

  // Compute 0 or -1, based on sign bit
  Node *sign = phase->transform(new (3) RShiftINode(dividend,phase->intcon(31)));
  // Mask sign bit to the low sign bits
  Node *round = phase->transform(new (3) AndINode(sign,phase->intcon(i-1)));
  // Round up before shifting
  Node *sum = phase->transform(new (3) AddINode(dividend,round));
  // Shift for division
  Node *shift = new (3) RShiftINode(sum, log_i);

  hook->set_req( 1, NULL, du );
  hook->set_req( 2, NULL, du );
  return shift;
}

//------------------------------Value------------------------------------------
// A DivINode divides its inputs.  The third input is a Control input, used to
// prevent hoisting the divide above an unsafe test.
const Type *DivINode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  const Type *t2 = phase->type( _in[2] );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t2 == Type::TOP ) return Type::TOP;

  // Either input is BOTTOM ==> the result is the local BOTTOM
  const Type *bot = bottom_type();
  if( (t1 == bot) || (t2 == bot) ||
      (t1 == Type::BOTTOM) || (t2 == Type::BOTTOM) ) 
    return bot;

  // x/x == 1,except for 0
  if( phase->eqv( _in[1], _in[2] ) && t1 != TypeInt::ZERO )
    return TypeInt::ONE;

  // Divide the two numbers.  We approximate.
  // If divisor is a constant and not zero
  const TypeInt *i1 = t1->is_int();
  const TypeInt *i2 = t2->is_int();
  if( i2->is_con() && i2->get_con() ) {
    int32 d = i2->get_con(); // Divisor
    if( d < 0 ) return TypeInt::make(i1->_hi/d, i1->_lo/d);
    return TypeInt::make(i1->_lo/d, i1->_hi/d);
  }  

  // If the dividend is a constant
  if( i1->is_con() ) {
    int32 d = i1->get_con();
    if( d < 0 ) return TypeInt::make(d,-d);
    return TypeInt::make(-d,d);
  }

  // Otherwise we give up all hope
  return TypeInt::INT;
}


//=============================================================================
//------------------------------Identity---------------------------------------
// If the divisor is 1, we are an identity on the dividend.
Node *DivLNode::Identity( PhaseTransform *phase ) {
  return (phase->type( _in[2] )->higher_equal(TypeLong::ONE)) ? _in[1] : this;
}

//------------------------------Idealize---------------------------------------
// Dividing by a power of 2 is a shift.
Node *DivLNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  if( in(0) && remove_dead_region(phase,du) ) return this;

  const Type *t = phase->type( _in[2] );
  if( t == TypeLong::ONE )       // Identity?
    return NULL;		// Skip it

  const TypeLong *ti = t->isa_long();
  if( !ti ) return NULL;
  if( !ti->is_con() ) return NULL;
  jlong i = ti->get_con();	// Get divisor
  if( i ) set_req(0,NULL,du);	// Dividing by a not-zero constant; no faulting

  // Dividing by MININT does not optimize as a power-of-2 shift.
  if( i == MININT64 ) return NULL;

  // Check for negative power of 2
  Node *dividend = in(1);
  if( (-i & ~(-i-1)) == -i ) {
    i = -i;			// Flip divisor
    dividend = phase->transform(new (3) SubLNode(phase->makecon(TypeLong::ZERO),in(1)));
  }

  // Check for power of 2
  if( (i & ~(i-1)) != i )       // Is divisor a power of 2?
    return NULL;		// Not a power of 2

  // Generate number of bits to shift
  Node *log_i = phase->intcon(log2_long(i));

  // An AND mask of sufficient size clears the low bits and
  // I can avoid rounding.
  if( dividend->Opcode() == Op_AndL ) {
    const TypeLong *andconi = phase->type( dividend->in(2) )->isa_long();
    if( andconi &&
        andconi->is_con() &&
        andconi->get_con() == -i ) 
      return new (3) RShiftLNode(dividend->in(1), log_i);
  }

  // Divide-by-power-of-2 can be made into a shift, but you have to do
  // more math for the rounding.  You need to add 0 for positive
  // numbers, and "i-1" for negative numbers.  Example: i=4, so the
  // shift is by 2.  You need to add 3 to negative dividends and 0 to
  // positive ones.  So (-7+3)>>2 becomes -1, (-4+3)>>2 becomes -1,
  // (-2+3)>>2 becomes 0, etc.
  Node *hook = new (3) Node( NULL, NULL, NULL );
  hook->set_req(1,dividend,du);
  hook->set_req(2,log_i,du);

  // Compute 0 or -1, based on sign bit
  Node *sign = phase->transform(new (3) RShiftLNode(dividend,phase->intcon(63)));
  // Mask sign bit to the low sign bits
  Node *round = phase->transform(new (3) AndLNode(sign,phase->makecon(TypeLong::make(i-1))));
  // Round up before shifting
  Node *sum = phase->transform(new (3) AddLNode(dividend,round));
  // Shift for division
  Node *shift = new (3) RShiftLNode(sum, log_i);

  hook->set_req( 1, NULL, du );
  hook->set_req( 2, NULL, du );
  return shift;
}

//------------------------------Value------------------------------------------
// A DivLNode divides its inputs.  The third input is a Control input, used to
// prevent hoisting the divide above an unsafe test.
const Type *DivLNode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  const Type *t2 = phase->type( _in[2] );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t2 == Type::TOP ) return Type::TOP;

  // Either input is BOTTOM ==> the result is the local BOTTOM
  const Type *bot = bottom_type();
  if( (t1 == bot) || (t2 == bot) ||
      (t1 == Type::BOTTOM) || (t2 == Type::BOTTOM) ) 
    return bot;

  // x/x == 1,except for 0
  if( phase->eqv( _in[1], _in[2] ) && t1 != TypeLong::ZERO )
    return TypeLong::ONE;

  // Divide the two numbers.  We approximate.
  // If divisor is a constant and not zero
  const TypeLong *i1 = t1->is_long();
  const TypeLong *i2 = t2->is_long();
  if( i2->is_con() && i2->get_con() ) {
    jlong d = i2->get_con();    // Divisor
    if( d < 0 ) return TypeLong::make(i1->_hi/d, i1->_lo/d);
    return TypeLong::make(i1->_lo/d, i1->_hi/d);
  }  

  // If the dividend is a constant
  if( i1->is_con() ) {
    jlong d = i1->get_con();
    if( d < 0 ) return TypeLong::make(d,-d);
    return TypeLong::make(-d,d);
  }

  // Otherwise we give up all hope
  return TypeLong::LONG;
}


//=============================================================================
//------------------------------Value------------------------------------------
// An DivFNode divides its inputs.  The third input is a Control input, used to
// prevent hoisting the divide above an unsafe test.
const Type *DivFNode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  const Type *t2 = phase->type( _in[2] );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t2 == Type::TOP ) return Type::TOP;

  // Either input is BOTTOM ==> the result is the local BOTTOM
  const Type *bot = bottom_type();
  if( (t1 == bot) || (t2 == bot) ||
      (t1 == Type::BOTTOM) || (t2 == Type::BOTTOM) ) 
    return bot;

  // x/x == 1, we ignore 0/0.
  // Note: if t1 and t2 are zero than result is NaN (JVMS page 213)
  // does not work for variables because of NaN's
  if( phase->eqv( _in[1], _in[2] ) && t1->_base == Type::FloatCon)
    if (!g_isnan(t1->getf()) && g_isfinite(t1->getf()) && t1->getf() != 0.0) // could be negative ZERO or NaN
      return TypeF::ONE;

  if( t2 == TypeF::ONE )
    return t1;

  // If divisor is a constant and not zero, divide them numbers
  if( t1->_base == Type::FloatCon && 
      t2->_base == Type::FloatCon &&
      t2->getf() != 0.0 ) // could be negative zero
    return TypeF::make( t1->getf()/t2->getf() );

  // If the dividend is a constant zero
  // Note: if t1 and t2 are zero than result is NaN (JVMS page 213)
  // Test TypeF::ZERO is not sufficient as it could be negative zero

  if( t1 == TypeF::ZERO && !g_isnan(t2->getf()) && t2->getf() != 0.0 )
    return TypeF::ZERO;

  // Otherwise we give up all hope
  return Type::FLOAT;
}

//------------------------------isA_Copy---------------------------------------
// Dividing by self is 1.
// If the divisor is 1, we are an identity on the dividend.
Node *DivFNode::Identity( PhaseTransform *phase ) {
  return (phase->type( _in[2] ) == TypeF::ONE) ? _in[1] : this;
}


//=============================================================================
//------------------------------Value------------------------------------------
// An DivDNode divides its inputs.  The third input is a Control input, used to
// prvent hoisting the divide above an unsafe test.
const Type *DivDNode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  const Type *t2 = phase->type( _in[2] );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t2 == Type::TOP ) return Type::TOP;

  // Either input is BOTTOM ==> the result is the local BOTTOM
  const Type *bot = bottom_type();
  if( (t1 == bot) || (t2 == bot) ||
      (t1 == Type::BOTTOM) || (t2 == Type::BOTTOM) ) 
    return bot;

  // x/x == 1, we ignore 0/0.
  // Note: if t1 and t2 are zero than result is NaN (JVMS page 213)
  // Does not work for variables because of NaN's
  if( phase->eqv( _in[1], _in[2] ) && t1->_base == Type::DoubleCon)
    if (!g_isnan(t1->getd()) && g_isfinite(t1->getd()) && t1->getd() != 0.0) // could be negative ZERO or NaN
      return TypeD::ONE;

  if( t2 == TypeD::ONE )
    return t1;

  // If divisor is a constant and not zero, divide them numbers
  if( t1->_base == Type::DoubleCon && 
      t2->_base == Type::DoubleCon &&
      t2->getd() != 0.0 ) // could be negative zero
    return TypeD::make( t1->getd()/t2->getd() );

  // If the dividend is a constant zero
  // Note: if t1 and t2 are zero than result is NaN (JVMS page 213)
  // Test TypeF::ZERO is not sufficient as it could be negative zero
  if( t1 == TypeD::ZERO && !g_isnan(t2->getd()) && t2->getd() != 0.0 )
    return TypeD::ZERO;

  // Otherwise we give up all hope
  return Type::DOUBLE;
}


//------------------------------isA_Copy---------------------------------------
// Dividing by self is 1.
// If the divisor is 1, we are an identity on the dividend.
Node *DivDNode::Identity( PhaseTransform *phase ) {
  return (phase->type( _in[2] ) == TypeD::ONE) ? _in[1] : this;
}


//=============================================================================
//------------------------------Idealize---------------------------------------
Node *ModINode::Ideal(PhaseGVN *phase,PhaseDefUse *du ) {
  if( !in(0) ) return NULL;     // No control input?  Then nothing to do
  // Check for dead control input
  if( remove_dead_region(phase,du) ) return this;
  // Check for useless control input
  const Type *t = phase->type( in(2) );
  if( t == Type::TOP ) return NULL;
  const TypeInt *ti = t->is_int();
  // Check for excluding mod-zero case
  if( ti->_hi < 0 || ti->_lo > 0 ) {
    set_req( 0, 0, du );        // Yank control input
    return this;
  }
  return NULL;
}

//------------------------------Value------------------------------------------
const Type *ModINode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  const Type *t2 = phase->type( _in[2] );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t2 == Type::TOP ) return Type::TOP;

  // Either input is BOTTOM ==> the result is the local BOTTOM
  const Type *bot = bottom_type();
  if( (t1 == bot) || (t2 == bot) ||
      (t1 == Type::BOTTOM) || (t2 == Type::BOTTOM) ) 
    return bot;

  // 0 MOD X is 0
  if( t1 == TypeInt::ZERO ) return TypeInt::ZERO; 
  // X MOD X is 0
  if( phase->eqv( _in[1], _in[2] ) && t1 != TypeInt::ZERO) return TypeInt::ZERO;

  const TypeInt *i1 = t1->is_int();
  const TypeInt *i2 = t2->is_int();
  if( !i1->is_con() || !i2->is_con() ) {
    if( i1->_lo >= 0 && i2->_lo >= 0 )
      return TypeInt::POS;
    // If both numbers are not constants, we know little.
    return TypeInt::INT;
  }
  // Mod by zero?  Throw exception at runtime!
  if( !i2->get_con() ) return TypeInt::INT;
  
  // We must be modulo'ing 2 float constants.
  // Check for min_jint % '-1', result is defined to be '0'.
  if( i1->get_con() == min_jint && i2->get_con() == -1 ) 
    return TypeInt::ZERO;

  return TypeInt::make( i1->get_con() % i2->get_con() );
}


//=============================================================================
//------------------------------Idealize---------------------------------------
Node *ModLNode::Ideal(PhaseGVN *phase,PhaseDefUse *du ) {
  if( !in(0) ) return NULL;     // No control input?  Then nothing to do
  // Check for dead control input
  if( remove_dead_region(phase,du) ) return this;
  // Check for useless control input
  const Type *t = phase->type( in(2) );
  if( t == Type::TOP ) return NULL;
  const TypeLong *ti = t->is_long();
  // Check for excluding mod-zero case
  if( ti->_hi < 0 || ti->_lo > 0 ) {
    set_req( 0, NULL, du );     // Yank control input
    return this;
  }
  return NULL;
}

//------------------------------Value------------------------------------------
const Type *ModLNode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  const Type *t2 = phase->type( _in[2] );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t2 == Type::TOP ) return Type::TOP;

  // Either input is BOTTOM ==> the result is the local BOTTOM
  const Type *bot = bottom_type();
  if( (t1 == bot) || (t2 == bot) ||
      (t1 == Type::BOTTOM) || (t2 == Type::BOTTOM) ) 
    return bot;

  // 0 MOD X is 0
  if( t1 == TypeLong::ZERO ) return TypeLong::ZERO; 
  // X MOD X is 0
  if( phase->eqv( _in[1], _in[2] ) && t1 != TypeLong::ZERO) return TypeLong::ZERO;

  const TypeLong *i1 = t1->is_long();
  const TypeLong *i2 = t2->is_long();
  if( !i1->is_con() || !i2->is_con() ) {
    if( i1->_lo >= 0L && i2->_lo >= 0L )
      return TypeLong::make(0L,max_jlong);
    // If both numbers are not constants, we know little.
    return TypeLong::LONG;
  }
  // Mod by zero?  Throw exception at runtime!
  if( !i2->get_con() ) return TypeLong::LONG;
  
  // We must be modulo'ing 2 float constants.
  // Check for min_jint % '-1', result is defined to be '0'.
  if( i1->get_con() == min_jlong && i2->get_con() == -1 ) 
    return TypeLong::ZERO;

  return TypeLong::make( i1->get_con() % i2->get_con() );
}


//=============================================================================
//------------------------------Value------------------------------------------
const Type *ModFNode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  const Type *t2 = phase->type( _in[2] );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t2 == Type::TOP ) return Type::TOP;

  // Either input is BOTTOM ==> the result is the local BOTTOM
  const Type *bot = bottom_type();
  if( (t1 == bot) || (t2 == bot) ||
      (t1 == Type::BOTTOM) || (t2 == Type::BOTTOM) ) 
    return bot;

  // If either is a NaN, return an input NaN
  if( g_isnan(t1->getf()) )    return t1;
  if( g_isnan(t2->getf()) )    return t2;

  // It is not worth trying to constant fold this stuff!
  return Type::FLOAT;

  /*
  // If dividend is infinity or divisor is zero, or both, the result is NaN
  if( !g_isfinite(t1->getf()) || ((t2->getf() == 0.0) || (jint_cast(t2->getf()) == 0x80000000)) )

  // X MOD infinity = X
  if( !g_isfinite(t2->getf()) && !g_isnan(t2->getf()) ) return t1;
  // 0 MOD finite = dividend (positive or negative zero)
  // Not valid for: NaN MOD any; any MOD nan; 0 MOD 0; or for 0 MOD NaN
  // NaNs are handled previously.
  if( !(t2->getf() == 0.0) && !((int)t2->getf() == 0x80000000)) {
    if (((t1->getf() == 0.0) || ((int)t1->getf() == 0x80000000)) && g_isfinite(t2->getf()) ) {
      return t1;
    }
  }
  // X MOD X is 0
  // Does not work for variables because of NaN's
  if( phase->eqv( _in[1], _in[2] ) && t1->_base == Type::FloatCon) 
    if (!g_isnan(t1->getf()) && (t1->getf() != 0.0) && ((int)t1->getf() != 0x80000000)) {
      if(t1->getf() < 0.0) {
        float result = jfloat_cast(0x80000000);
        return TypeF::make( result );
      }
      else
        return TypeF::ZERO;
    }

  // If both numbers are not constants, we know nothing.
  if( (t1->_base != Type::FloatCon) || (t2->_base != Type::FloatCon) )
    return Type::FLOAT;
  
  // We must be modulo'ing 2 float constants.
  // Make sure that the sign of the fmod is equal to the sign of the dividend
  float result = (float)fmod( t1->getf(), t2->getf() );
  float dividend = t1->getf();
  if( (dividend < 0.0) || ((int)dividend == 0x80000000) ) {
    if( result > 0.0 )
      result = 0.0 - result;
    else if( result == 0.0 ) {
      result = jfloat_cast(0x80000000);
    }
  }
  return TypeF::make( result );
  */
}


//=============================================================================
//------------------------------Value------------------------------------------
const Type *ModDNode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  const Type *t2 = phase->type( _in[2] );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t2 == Type::TOP ) return Type::TOP;

  // Either input is BOTTOM ==> the result is the local BOTTOM
  const Type *bot = bottom_type();
  if( (t1 == bot) || (t2 == bot) ||
      (t1 == Type::BOTTOM) || (t2 == Type::BOTTOM) ) 
    return bot;

  // If either is a NaN, return an input NaN
  if( g_isnan(t1->getd()) )    return t1;
  if( g_isnan(t2->getd()) )    return t2;
  // X MOD infinity = X
  if( !g_isfinite(t2->getd())) return t1;
  // 0 MOD finite = dividend (positive or negative zero)
  // Not valid for: NaN MOD any; any MOD nan; 0 MOD 0; or for 0 MOD NaN
  // NaNs are handled previously.
  if( !(t2->getd() == 0.0) ) {
    if( t1->getd() == 0.0 && g_isfinite(t2->getd()) ) {
      return t1;
    }
  }

  // X MOD X is 0
  // does not work for variables because of NaN's
  if( phase->eqv( _in[1], _in[2] ) && t1->_base == Type::DoubleCon ) 
    if (!g_isnan(t1->getd()) && t1->getd() != 0.0)
      return TypeD::ZERO;
 

  // If both numbers are not constants, we know nothing.
  if( (t1->_base != Type::DoubleCon) || (t2->_base != Type::DoubleCon) )
    return Type::DOUBLE;
  
  // We must be modulo'ing 2 double constants.
  return TypeD::make( fmod( t1->getd(), t2->getd() ) );
}

