// @(#)loopTransform.cpp	1.39 01/02/14 09:20:41
//
//  Copyright 1997-2001 Sun Microsystems, Inc. 901 San Antonio Road,
// Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
// 
// This software is the confidential and proprietary information of Sun
// Microsystems, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Sun.
// 
// CopyrightVersion 1.2
#include "incls/_precompiled.incl"
#include "incls/_loopTransform.cpp.incl"

//------------------------------is_loop_exit-----------------------------------
// Given an IfNode, return the loop-exiting projection or NULL if both 
// arms remain in the loop.
Node *IdealLoopTree::is_loop_exit( Node *iff, PhaseIdealLoop *phase ) const {
  PhaseDefUse &du = phase->_igvn._du;
  if( du.cnt(iff) != 2 ) return NULL; // Ignore partially dead tests

  // Test is an IfNode, has 2 projections.  If BOTH are in the loop
  // we need loop unswitching instead of peeling.
  Node **outp = du.out(iff);
  if( !is_member(phase->get_loop( outp[0] )) )
    return outp[0];
  if( !is_member(phase->get_loop( outp[1] )) )
    return outp[1];
  return NULL;
}


//=============================================================================

//------------------------------policy_peeling---------------------------------
// Return TRUE or FALSE if the loop should be peeled or not.  Peel if we can
// make some loop-invariant test (usually a null-check) happen before the loop.
bool IdealLoopTree::policy_peeling( PhaseIdealLoop *phase ) const {
  PhaseDefUse &du = phase->_igvn._du;
  Node *test = ((IdealLoopTree*)this)->tail();
  while( test != _head ) {	// Scan till run off top of loop
    if( test->is_If() ) {	// Test?
      Node *ctrl = phase->get_ctrl( test->in(1)->_idx );
      if( ctrl == phase->C->_top_node ) 
        return false;           // Found dead test on live IF?  No peeling!
      // Condition is not a member of this loop?
      if( !is_member(phase->get_loop(ctrl)) &&
	  is_loop_exit(test,phase) )
	return true;	        // Found reason to peel!
    }
    // Walk up dominators to loop _head looking for test which is
    // executed on every path thru loop.
    test = phase->idom(test->_idx);
  }
  return false;
}

//------------------------------do_peeling-------------------------------------
// Peel the first iteration of the given loop.  
// Step 1: Clone the loop body.  The clone becomes the peeled iteration.
//         The pre-loop illegally has 2 control users (old & new loops).
// Step 2: Make the old-loop fall-in edges point to the peeled iteration.
//         Do this by making the old-loop fall-in edges act as if they came
//         around the loopback from the prior iteration (follow the old-loop
//         backedges) and then map to the new peeled iteration.  This leaves
//         the pre-loop with only 1 user (the new peeled iteration), but the
//         peeled-loop backedge has 2 users.
// Step 3: Cut the backedge on the clone (so its not a loop) and remove the
//         extra backedge user.
void PhaseIdealLoop::do_peeling( IdealLoopTree *loop, Node_List &old_new ) {

  C->_major_progress = true;
  // Peeling a 'main' loop in a pre/main/post situation obfuscates the
  // 'pre' loop from the main and the 'pre' can no longer have it's
  // iterations adjusted.  Therefore, we need to declare this loop as
  // no longer a 'main' loop; it will need new pre and post loops before
  // we can do further RCE.
  CountedLoopNode *cl = loop->_head->is_CountedLoop();
  if( cl && cl->is_main_loop() ) {
    cl->set_normal_loop();
#ifndef PRODUCT
    if( PrintOpto && VerifyLoopOptimizations ) {
      tty->print("Peeling a 'main' loop; resetting to 'normal' ");
      loop->dump_head();
    }
#endif
  }

  // Step 1: Clone the loop body.  The clone becomes the peeled iteration.
  //         The pre-loop illegally has 2 control users (old & new loops).
  clone_loop( loop, old_new );


  // Step 2: Make the old-loop fall-in edges point to the peeled iteration.
  //         Do this by making the old-loop fall-in edges act as if they came
  //         around the loopback from the prior iteration (follow the old-loop
  //         backedges) and then map to the new peeled iteration.  This leaves
  //         the pre-loop with only 1 user (the new peeled iteration), but the
  //         peeled-loop backedge has 2 users.
  PhaseDefUse &du = _igvn._du;
  uint j;
  uint max = du.cnt(loop->_head);
  for( j = 0; j < max; j++ ) {
    Node *old = du.out(loop->_head)[j];
    if( old->in(0) == loop->_head && old->req() == 3 &&
        (old->is_Loop() || old->is_Phi()) ) {
      Node *new_exit_value = old_new[old->in(LoopNode::LoopBackControl)->_idx];
      if( !new_exit_value )	// Backedge value is ALSO loop invariant?
	// Then loop body backedge value remains the same.
	new_exit_value = old->in(LoopNode::LoopBackControl);
      _igvn.hash_delete(old);
      old->set_req( LoopNode::EntryControl, new_exit_value, &du );
    }
  }
  

  // Step 3: Cut the backedge on the clone (so its not a loop) and remove the
  //         extra backedge user.
  Node *nnn = old_new[loop->_head->_idx];
  _igvn.hash_delete(nnn);
  nnn->set_req( LoopNode::LoopBackControl, C->_top_node, &du );
  max  = du.cnt(nnn);
  for( j = 0; j < max; j++ ) {
    Node *use = du.out(nnn)[j];
    if( use->in(0) == nnn && use->req() == 3 && use->is_Phi() ) {
      _igvn.hash_delete(use);
      use->set_req( LoopNode::LoopBackControl, C->_top_node, &du );
    }
  }


  // Step 4: Correct dom-depth info.  Set to loop-head depth.
  int dd = dom_depth(loop->_head->_idx);
  set_idom( loop->_head->_idx, loop->_head->in(1), dd );
  for( j = 0; j < loop->_body.size(); j++ ) {
    Node *old = loop->_body.at(j);
    Node *nnn = old_new[old->_idx];
    if( !has_ctrl(nnn->_idx) )
      set_idom( nnn->_idx, idom(nnn->_idx), dd-1 );
    // While we're at it, remove any SafePoints from the peeled code
    if( old->Opcode() == Op_SafePoint ) {
      Node *nnn = old_new[old->_idx];
      // Flag him as removeable
      _igvn.hash_delete(nnn);
      while( nnn->req() >= TypeFunc::Parms ) {
	nnn->set_req_X( nnn->req()-1, NULL, &_igvn );
	nnn->del_req(nnn->req()-1);
      }
      // Finish removing him in IterGVN
      _igvn.add_users_to_worklist(nnn);
    }
  }

  // Now force out all loop-invariant dominating tests.  The optimizer 
  // finds some, but we _know_ they are all useless.
  bool progress = true;
  while( progress ) {
    progress = false;	        // Reset for next iteration
    Node *prev = loop->_head->in(LoopNode::LoopBackControl);//loop->tail();
    Node *test = prev->in(0);
    while( test != loop->_head ) { // Scan till run off top of loop
      
      int p_op = prev->Opcode();
      if( (p_op == Op_IfFalse || p_op == Op_IfTrue) &&
          test->is_If() &&	// Test?
          !test->in(1)->is_Con() && // And not already obvious?
          // Condition is not a member of this loop?
          !loop->is_member(get_loop(get_ctrl(test->in(1)->_idx)))){
        // Walk loop body looking for instances of this test
        for( uint i = 0; i < loop->_body.size(); i++ ) {
          Node *n = loop->_body.at(i);
          IfNode *iff = n->is_If();
          if( iff && iff->in(1) == test->in(1) /*&& iff != loop->tail()->in(0)*/ ) {
            // IfNode was dominated by version in peeled loop body
    	    progress = true;
    	    dominated_by( old_new[prev->_idx], iff );
          }
        }
      }
      prev = test;
      test = idom(test->_idx);
    } // End of scan tests in loop

  }	// End of while( progress )

}

//------------------------------policy_maximally_unroll------------------------
// Return exact loop trip count, or 0 if not maximally unrolling
int IdealLoopTree::policy_maximally_unroll( PhaseIdealLoop *phase ) const {
  assert( _head->is_CountedLoop(), "" );
  CountedLoopNode *cl = (CountedLoopNode*)_head;
  assert( cl->is_normal_loop(), "" );

  Node *init = cl->init_trip();
  Node *limit = cl->limit();

  // Non-constant bounds
  if( !init->is_Con() || !limit->is_Con() ) return 0;
  int span = limit->get_int() - init->get_int();
  int stride = cl->stride_con();
  int trip_count = span/stride;
  assert( trip_count*stride == span, "must divide evenly" );
  // Note: A span of zero implies a trip_count of zero and this is OK.  We'll
  // return a zero (no maximally unroll) and the regular unroll/peel route
  // will make a small mess which CCP will fold away.

  // Real policy: if we maximally unroll, does it get too big?
  // Allow the unrolled mess to get larger than standard loop
  // size.  After all, it will no longer be a loop.
  if( (int)_body.size() * trip_count > LoopUnrollLimit*4 ) return 0;

  // Decide to maximally unroll
  return trip_count;
}


//------------------------------policy_unroll----------------------------------
// Return TRUE or FALSE if the loop should be unrolled or not.  Unroll if 
// the loop is a CountedLoop and the body is small enough.
bool IdealLoopTree::policy_unroll( PhaseIdealLoop *phase ) const {

  assert( _head->is_CountedLoop(), "" );
  CountedLoopNode *cl = (CountedLoopNode*)_head;
  assert( cl->is_normal_loop() || cl->is_main_loop(), "" );

  // Check for being too big
  if( _body.size() > (uint)LoopUnrollLimit ) { 
    // Key test to unroll CaffeineMark's Logic test
    int xors_in_loop = 0;
    for( uint k = 0; k < _body.size(); k++ ) 
      if( _body.at(k)->Opcode() == Op_XorI )
        xors_in_loop++;
    if( xors_in_loop >= 4 && _body.size() < (uint)LoopUnrollLimit*4) return true;
    // Normal case: loop too big
    return false;
  }
  
  // Check for stride being a small enuf constant
  if( abs(cl->stride_con()) > (1<<3) ) return false;

  // Unroll once!  (Each trip will soon do double iterations)
  return true;
}

//------------------------------policy_align-----------------------------------
// Return TRUE or FALSE if the loop should be cache-line aligned.  Gather the
// expression that does the alignment.  Note that only one array base can be
// aligned in a loop (unless the VM guarentees mutual alignment).  Note that
// if we vectorize short memory ops into longer memory ops, we may want to
// increase alignment.
bool IdealLoopTree::policy_align( PhaseIdealLoop *phase ) const {
  return false;
}

//------------------------------policy_range_check-----------------------------
// Return TRUE or FALSE if the loop should be range-check-eliminated.
// Actually we do interation-splitting, a more powerfully form of RCE.
bool IdealLoopTree::policy_range_check( PhaseIdealLoop *phase ) const {
  if( !RangeCheckElimination ) return false;

  PhaseDefUse &du = phase->_igvn._du;
  assert( _head->is_CountedLoop(), "" );
  CountedLoopNode *cl = (CountedLoopNode*)_head;
  // If we unrolled with no intention of doing RCE and we later
  // changed our minds, we got no pre-loop.  Either we need to
  // make a new pre-loop, or we gotta disallow RCE.
  if( cl->is_main_no_pre_loop() ) return false; // Disallowed for now.
  Node *trip_counter = cl->phi();

  // Check loop body for tests of trip-counter plus loop-invariant vs
  // loop-invariant.
  for( uint i = 0; i < _body.size(); i++ ) {
    Node *iff = _body[i];
    if( iff->Opcode() == Op_If ) { // Test?

      // Comparing trip+off vs limit
      Node *bol = iff->in(1);
      if( bol->req() != 2 ) continue; // dead constant test
      Node *cmp = bol->in(1);

      Node *limit = cmp->in(2);

      Node *limit_c = phase->get_ctrl( limit->_idx );
      if( limit_c == phase->C->_top_node ) 
	return false;		// Found dead test on live IF?  No RCE!
      if( is_member(phase->get_loop(limit_c) ) )
	continue;		// Loop-varying limit

      Node *add = cmp->in(1);
      if( add == trip_counter ) {
      } else {
	if( !add->is_Add() && !add->is_Sub() ) continue; // Odd compare
	Node *off;		// Try to find trip+offset
	if( add->in(1) == trip_counter ) 
	  off = add->in(2);	// Found offset
	else if( add->in(2) == trip_counter )
	  off = add->in(1);	// Found offset
	else continue;		// NOT trip+offset

	Node *off_c = phase->get_ctrl( off->_idx );
	if( off_c == phase->C->_top_node ) 
	  return false;		// Found dead test on live IF?  No RCE!
	if( is_member(phase->get_loop(off_c) ) )
	  continue;		// Offset is not really loop invariant
      }	
      // Yeah!  Found a test like 'trip+off vs limit'
      // Test is an IfNode, has 2 projections.  If BOTH are in the loop
      // we need loop unswitching instead of iteration splitting.
      if( is_loop_exit(iff,phase) )
	return true;	        // Found reason to split iterations
    } // End of is IF
  }

  return false;
}

//------------------------------policy_peel_only-------------------------------
// Return TRUE or FALSE if the loop should NEVER be RCE'd or aligned.  Useful
// for unrolling loops with NO array accesses.
bool IdealLoopTree::policy_peel_only( PhaseIdealLoop *phase ) const {

  for( uint i = 0; i < _body.size(); i++ )
    if( _body[i]->is_Mem() )
      return false;

  // No memory accesses at all!
  return true;
}

//------------------------------insert_pre_post_loops--------------------------
// Insert pre and post loops.  If peel_only is set, the pre-loop cannout have
// more iterations added.  It acts as a 'peel' only, no lower-bound RCE, no
// alignment.  Useful to unroll loops that do no array accesses.
void PhaseIdealLoop::insert_pre_post_loops( IdealLoopTree *loop, Node_List &old_new, bool peel_only ) {

  C->_major_progress = true;

  // Find common pieces of the loop being guarded with pre & post loops
  PhaseDefUse &du = _igvn._du;
  assert( loop->_head->is_CountedLoop(), "" );
  CountedLoopNode *main_head = (CountedLoopNode*)loop->_head;
  assert( main_head->is_normal_loop(), "" );
  CountedLoopEndNode *main_end = main_head->loopexit();
  assert( du.cnt(main_end->_idx) == 2, "1 true, 1 false path only" );
  uint dd = dom_depth(main_head->_idx);
  uint max = du.cnt(main_head);
  uint i;

  Node *pre_header= main_head->in(LoopNode::EntryControl);
  Node *init      = main_head->init_trip();
  Node *incr      = main_end ->incr();
  Node *limit     = main_end ->limit();
  Node *stride    = main_end ->stride();
  Node *cmp       = main_end ->cmp_node();
  BoolTest::mask b_test = main_end->test_trip();

  // Need only 1 user of 'cmp' because I will be hacking the loop bounds.
  if( du.cnt(cmp) != 1 ) {
    cmp = cmp->clone();
    register_new_node(cmp,main_end->in(CountedLoopEndNode::TestControl));
    Node *bol = main_end->in(CountedLoopEndNode::TestValue);
    _igvn.hash_delete(bol);
    bol->set_req(1,cmp,&du);
  }

  //------------------------------
  // Step A: Create Post-Loop.

  // Step A1: Clone the loop body.  The clone becomes the post-loop.  The main
  // loop pre-header illegally has 2 control users (old & new loops).
  clone_loop( loop, old_new );
  assert( old_new[main_head->_idx]->is_CountedLoop   (), "" );
  assert( old_new[main_end ->_idx]->Opcode() == Op_CountedLoopEnd, "" );
  CountedLoopNode    *post_head = (CountedLoopNode   *)old_new[main_head->_idx];
  post_head->set_post_loop(main_head);

  // Find the main-loop normal exit.
  Node *main_exit = du.out(main_end)[0];
  if( main_exit->Opcode() == Op_IfTrue )
    main_exit = du.out(main_end)[1];
  assert( main_exit->Opcode() == Op_IfFalse, "" );
  IfFalseNode *new_main_exit = new (1) IfFalseNode(main_end);
  _igvn.register_new_node_with_optimizer( new_main_exit );
  set_idom( new_main_exit->_idx, main_end, dd-1 );
  set_loop( new_main_exit->_idx, loop->_parent );

  // Step A2: Build a zero-trip guard for the post-loop.  After leaving the
  // main-loop, the post-loop may not execute at all.  We 'opaque' the incr
  // (the main-loop trip-counter exit value) because we will be changing
  // the exit value (via unrolling) so we cannot constant-fold away the zero
  // trip guard until all unrolling is done.
  Node *zer_opaq = new (2) Opaque1Node(incr);
  Node *zer_cmp  = new (3) CmpINode( zer_opaq, limit );
  Node *zer_bol  = new (2) BoolNode( zer_cmp, b_test );
  register_new_node( zer_opaq, new_main_exit );
  register_new_node( zer_cmp , new_main_exit );
  register_new_node( zer_bol , new_main_exit );

  // Build the IfNode
  IfNode *zer_iff = new (2) IfNode( new_main_exit, zer_bol, 0.50f, -1.0f );
  _igvn.register_new_node_with_optimizer( zer_iff );
  set_idom( zer_iff->_idx, new_main_exit, dd-1 );
  set_loop( zer_iff->_idx, loop->_parent );

  // Plug in the false-path, taken if we need to skip post-loop
  _igvn.hash_delete( main_exit );
  main_exit->set_req( 0, zer_iff, &du );
  _igvn._worklist.push(main_exit);
  set_idom( main_exit->_idx, zer_iff, dd );
  assert( du.cnt(main_exit) == 1, "" );
  set_idom( du.out(main_exit)[0]->_idx, zer_iff, dd );
  // Make the true-path, must enter the main loop
  Node *zer_taken = new (1) IfTrueNode( zer_iff );
  _igvn.register_new_node_with_optimizer( zer_taken );
  set_idom( zer_taken->_idx, zer_iff, dd-1 );
  set_loop( zer_taken->_idx, loop->_parent );
  // Plug in the true path
  _igvn.hash_delete( post_head );
  post_head->set_req( LoopNode::EntryControl, zer_taken, &du );
  set_idom( post_head->_idx, zer_taken, dd );

  // Step A3: Make the fall-in values to the post-loop come from the
  // fall-out values of the main-loop.
  for( i = 0; i < max; i++ ) {
    Node *main_phi = du.out(main_head)[i];
    if( main_phi->is_Phi() && main_phi->in(0) == main_head ) {
      Node *post_phi = old_new[main_phi->_idx];
      _igvn.hash_delete(post_phi);
      post_phi->set_req( LoopNode::EntryControl, main_phi->in(LoopNode::LoopBackControl), &du );
    }
  }

  // Update local caches for next stanza
  main_exit = new_main_exit;


  //------------------------------
  // Step B: Create Pre-Loop.

  // Step B1: Clone the loop body.  The clone becomes the pre-loop.  The main
  // loop pre-header illegally has 2 control users (old & new loops).
  clone_loop( loop, old_new );
  assert( old_new[main_head->_idx]->is_CountedLoop   (), "" );
  assert( old_new[main_end ->_idx]->Opcode() == Op_CountedLoopEnd, "" );
  CountedLoopNode    *pre_head = (CountedLoopNode   *)old_new[main_head->_idx];
  CountedLoopEndNode *pre_end  = (CountedLoopEndNode*)old_new[main_end ->_idx];
  pre_head->set_pre_loop(main_head);
  Node *pre_incr = old_new[incr->_idx];

  // Find the pre-loop normal exit.
  Node *pre_exit = du.out(pre_end)[0];
  if( pre_exit->Opcode() == Op_IfTrue )
    pre_exit = du.out(pre_end)[1];
  assert( pre_exit->Opcode() == Op_IfFalse, "" );
  IfFalseNode *new_pre_exit = new (1) IfFalseNode(pre_end);
  _igvn.register_new_node_with_optimizer( new_pre_exit );
  set_idom( new_pre_exit->_idx, pre_end, dd-1 );
  set_loop( new_pre_exit->_idx, loop->_parent );

  // Step B2: Build a zero-trip guard for the main-loop.  After leaving the
  // pre-loop, the main-loop may not execute at all.  Later in life this
  // zero-trip guard will become the minimum-trip guard when we unroll
  // the main-loop.
  Node *min_opaq = new (2) Opaque1Node(limit);
  Node *min_cmp  = new (3) CmpINode( pre_incr, min_opaq );
  Node *min_bol  = new (2) BoolNode( min_cmp, b_test );
  register_new_node( min_opaq, new_pre_exit );
  register_new_node( min_cmp , new_pre_exit );
  register_new_node( min_bol , new_pre_exit );

  // Build the IfNode
  IfNode *min_iff = new (2) IfNode( new_pre_exit, min_bol, 0.50f, -1.0f );
  _igvn.register_new_node_with_optimizer( min_iff );
  set_idom( min_iff->_idx, new_pre_exit, dd-1 );
  set_loop( min_iff->_idx, loop->_parent );

  // Plug in the false-path, taken if we need to skip main-loop
  _igvn.hash_delete( pre_exit );
  pre_exit->set_req( 0, min_iff, &du );
  set_idom( pre_exit->_idx, min_iff, dd );
  assert( du.cnt(pre_exit) == 1, "" );
  set_idom( du.out(pre_exit)[0]->_idx, min_iff, dd );
  // Make the true-path, must enter the main loop
  Node *min_taken = new (1) IfTrueNode( min_iff );
  _igvn.register_new_node_with_optimizer( min_taken );
  set_idom( min_taken->_idx, min_iff, dd-1 );
  set_loop( min_taken->_idx, loop->_parent );
  // Plug in the true path
  _igvn.hash_delete( main_head );
  main_head->set_req( LoopNode::EntryControl, min_taken, &du );
  set_idom( main_head->_idx, min_taken, dd );

  // Step B3: Make the fall-in values to the main-loop come from the
  // fall-out values of the pre-loop.
  for( i = 0; i < max; i++ ) {
    Node *main_phi = du.out(main_head)[i];
    if( main_phi->is_Phi() && main_phi->in(0) == main_head ) {
      Node *pre_phi = old_new[main_phi->_idx];
      _igvn.hash_delete(main_phi);
      main_phi->set_req( LoopNode::EntryControl, pre_phi->in(LoopNode::LoopBackControl), &du );
    }
  }

  // Step B4: Shorten the pre-loop to run only 1 iteration (for now).
  // RCE and alignment may change this later.
  Node *cmp_end = pre_end->cmp_node();
  assert( cmp_end->in(2) == limit, "" );
  Node *pre_limit = new (3) AddINode( init, stride );
  Node *pre_opaq  = new (2) Opaque1Node(pre_limit);
  register_new_node( pre_limit, pre_head->in(0) );
  register_new_node( pre_opaq , pre_head->in(0) );

  // Since no other users of pre-loop compare, I can hack limit directly
  assert( du.cnt(cmp_end) == 1, "no other users" );
  _igvn.hash_delete(cmp_end);
  cmp_end->set_req( 2, peel_only ? pre_limit : pre_opaq, &du );

  // Flag main loop
  main_head->set_main_loop();
  if( peel_only ) main_head->set_main_no_pre_loop();
}


//------------------------------do_unroll--------------------------------------
// Unroll the loop body one step - make each trip do 2 iterations.
void PhaseIdealLoop::do_unroll( IdealLoopTree *loop, Node_List &old_new, bool adjust_min_trip ) {
  assert( LoopUnrollLimit, "" );
  C->_major_progress = true;
#ifndef PRODUCT
  if( PrintOpto && VerifyLoopOptimizations ) {
    tty->print("Unrolling ");
    loop->dump_head();
  }
#endif
  PhaseDefUse &du = _igvn._du;
  assert( loop->_head->is_CountedLoop(), "" );
  CountedLoopNode *loop_head = (CountedLoopNode*)loop->_head;
  CountedLoopEndNode *loop_end = loop_head->loopexit();
  assert( loop_end, "" );

  Node *ctrl  = loop_head->in(LoopNode::EntryControl);
  Node *limit = loop_head->limit();
  Node *init  = loop_head->init_trip();
  Node *strid = loop_head->stride();

  // -----------
  // Step 2: Cut back the trip counter for an unroll amount of 2.
  // Loop will normally trip (limit - init)/stride_con.  Since it's a
  // CountedLoop this is exact (stride divides limit-init exactly).
  // We are going to double the loop body, so we want to knock off any
  // odd iteration: (trip_cnt & ~1).  Then back compute a new limit.
  Node *span = new (3) SubINode( limit, init );
  register_new_node( span, ctrl );
  Node *trip = new (3) DivINode( 0, span, strid );
  register_new_node( trip, ctrl );
  Node *mtwo = _igvn.intcon(-2);
  set_ctrl( mtwo->_idx, C->_root );
  Node *rond = new (3) AndINode( trip, mtwo );
  register_new_node( rond, ctrl );
  Node *spn2 = new (3) MulINode( rond, strid );
  register_new_node( spn2, ctrl );
  Node *lim2 = new (3) AddINode( spn2, init );
  register_new_node( lim2, ctrl );

  // Hammer in the new limit
  Node *ctrl2 = loop_end->in(0);
  Node *cmp2 = new (3) CmpINode( loop_head->incr(), lim2 );
  register_new_node( cmp2, ctrl2 );
  Node *bol2 = new (2) BoolNode( cmp2, loop_end->test_trip() );
  register_new_node( bol2, ctrl2 );
  _igvn.hash_delete(loop_end);
  loop_end->set_req( CountedLoopEndNode::TestValue, bol2, &du );

  // Step 3: Find the min-trip test guarenteed before a 'main' loop.
  // Make it a 1-trip test (means at least 2 trips).
  if( adjust_min_trip ) {
    assert( loop_head->is_main_loop(), "" );
    assert( ctrl->Opcode() == Op_IfTrue || ctrl->Opcode() == Op_IfFalse, "" );
    Node *iff = ctrl->in(0);
    assert( iff->Opcode() == Op_If, "" );
    Node *bol = iff->in(1);
    assert( bol->Opcode() == Op_Bool, "" );
    Node *cmp = bol->in(1);
    assert( cmp->Opcode() == Op_CmpI, "" );
    Node *opaq = cmp->in(2);
    assert( opaq->Opcode() == Op_Opaque1, "" );

    // Guard test uses an 'opaque' node which is not shared.  Hence I
    // can edit it's inputs directly.  Hammer in the new limit for the
    // minimum-trip guard.
    assert( du.cnt(opaq) == 1, "" );
    _igvn.hash_delete(opaq);
    opaq->set_req( 1, lim2, &du );
  }

  // ---------
  // Step 4: Clone the loop body.  Move it inside the loop.  This loop body 
  // represents the odd iterations; since the loop trips an even number of
  // times its backedge is never taken.  Kill the backedge.
  clone_loop( loop, old_new );

  // Make backedges of the clone equal to backedges of the original.
  // Make the fall-in from the original come from the fall-out of the clone.
  uint max    = du.cnt(loop_head);
  for( uint j = 0; j < max; j++ ) {
    Node *phi = du.out(loop_head)[j];
    if( phi->is_Phi() && phi->in(0) == loop_head ) {
      Node *newphi = old_new[phi->_idx];
      _igvn.hash_delete( phi );
      _igvn.hash_delete( newphi );

      phi   ->set_req( LoopNode::   EntryControl, newphi->in(LoopNode::LoopBackControl), &du );
      newphi->set_req( LoopNode::LoopBackControl, phi   ->in(LoopNode::LoopBackControl), &du );
      phi   ->set_req( LoopNode::LoopBackControl, C->_top_node, &du );
    }
  }  
  Node *clone_head = old_new[loop_head->_idx];
  _igvn.hash_delete( clone_head );
  loop_head ->set_req( LoopNode::   EntryControl, clone_head->in(LoopNode::LoopBackControl), &du );
  clone_head->set_req( LoopNode::LoopBackControl, loop_head ->in(LoopNode::LoopBackControl), &du );
  loop_head ->set_req( LoopNode::LoopBackControl, C->_top_node, &du );
  loop->_head = clone_head;	// New loop header

  uint dd = dom_depth(loop_head->_idx);
  set_idom( loop_head ->_idx, loop_head ->in(LoopNode::EntryControl), dd );
  set_idom( clone_head->_idx, clone_head->in(LoopNode::EntryControl), dd );

  // Kill the clone's backedge
  Node *newcle = old_new[loop_end->_idx];
  _igvn.hash_delete( newcle );
  Node *one = _igvn.intcon(1);
  set_ctrl(one->_idx,C->_root);
  newcle->set_req( 1, one, &du );
  // Force clone into same loop body
  max = loop->_body.size();
  for( uint k = 0; k < max; k++ ) {
    Node *old = loop->_body.at(k);
    Node *nnn = old_new[old->_idx];
    loop->_body.push(nnn);
    if( !has_ctrl(old->_idx) )
      set_loop( nnn->_idx, loop );
  }

}

//------------------------------do_maximally_unroll----------------------------

void PhaseIdealLoop::do_maximally_unroll( IdealLoopTree *loop, Node_List &old_new, int trip_count ) {
  assert( trip_count > 0, "");

  // If loop is tripping an odd number of times, peel odd iteration
  if( (trip_count & 1) == 1 ) {
    do_peeling( loop, old_new );
    trip_count--;
  }

  // Now its tripping an even number of times remaining.  Double loop body.
  // Do not adjust pre-guards; they are not needed and do not exist.
  if( trip_count ) 
    do_unroll( loop, old_new, false );
}

//------------------------------add_constraint---------------------------------
// Constrain the main loop iterations so the condition:
//    scale_con * I + offset  <  limit
// always holds true.  That is, either increase the number of iterations in
// the pre-loop or the post-loop until the condition holds true in the main 
// loop.  Stride, scale, offset and limit are all loop invariant.  Further, 
// stride and scale are constants (offset and limit often are).
void PhaseIdealLoop::add_constraint( int stride_con, int scale_con, Node *offset, Node *limit, Node *pre_ctrl, Node **pre_limit, Node **main_limit ) {

  // Compute "I :: (limit-offset)/|scale_con|"
  Node *con = new (3) SubINode( limit, offset );
  register_new_node( con, pre_ctrl );
  Node *scale = _igvn.intcon(scale_con);
  set_ctrl( scale->_idx, C->_root );
  Node *X = new (3) DivINode( 0, con, scale );
  register_new_node( X, pre_ctrl );

  // For positive stride, the pre-loop limit always uses a MAX function 
  // and the main loop a MIN function.  For negative stride these are
  // reversed.  
  
  // Also for positive stride*scale the affine function is increasing, so the 
  // pre-loop must check for underflow and the post-loop for overflow.
  // Negative stride*scale reverses this; pre-loop checks for overflow and
  // post-loop for underflow.
  if( stride_con*scale_con > 0 ) {
    // Compute I < (limit-offset)/scale_con
    // Adjust main-loop last iteration to be MIN/MAX(main_loop,X)
    *main_limit = (stride_con > 0) 
      ? (Node*)(new (3) MinINode( *main_limit, X ))
      : (Node*)(new (3) MaxINode( *main_limit, X ));
    register_new_node( *main_limit, pre_ctrl );

  } else {
    // Compute (limit-offset)/scale_con + 1 <= I
    // Adjust pre-loop last iteration to be MAX/MIN(pre_loop,X+1)
    Node *one = _igvn.intcon(1);
    set_ctrl( one->_idx, C->_root );
    Node *adj = new (3) AddINode( X, one );
    register_new_node( adj, pre_ctrl );
    *pre_limit = (stride_con > 0)
      ? (Node*)new (3) MaxINode( *pre_limit, adj )
      : (Node*)new (3) MinINode( *pre_limit, adj );
    register_new_node( *pre_limit, pre_ctrl );

  }
}

//------------------------------do_range_check---------------------------------
// Eliminate range-checks and other trip-counter vs loop-invariant tests.
void PhaseIdealLoop::do_range_check( IdealLoopTree *loop, Node_List &old_new ) {
#ifndef PRODUCT
  if( PrintOpto && VerifyLoopOptimizations ) {
    tty->print("Range Check Elimination ");
    loop->dump_head();
  }
#endif
  assert( RangeCheckElimination, "" );
  assert( loop->_head->is_CountedLoop(), "" );
  CountedLoopNode *cl = (CountedLoopNode*)loop->_head;
  assert( cl->is_main_loop(), "" );
  PhaseDefUse &du = _igvn._du;

  // Find the trip counter; we are iteration splitting based on it
  Node *trip_counter = cl->phi();
  // Find the main loop limit; we will trim it's iterations 
  // to not ever trip end tests
  Node *main_limit = cl->limit();
  // Find the pre-loop limit; we will expand it's iterations to
  // not ever trip low tests.
  Node *ctrl  = cl->in(LoopNode::EntryControl);
  assert( ctrl->Opcode() == Op_IfTrue || ctrl->Opcode() == Op_IfFalse, "" );
  Node *iffm = ctrl->in(0);
  assert( iffm->Opcode() == Op_If, "" );
  Node *p_f = iffm->in(0);
  assert( p_f->Opcode() == Op_IfFalse, "" );
  assert( p_f->in(0)->Opcode() == Op_CountedLoopEnd, "" );
  CountedLoopEndNode *pre_end = (CountedLoopEndNode*)p_f->in(0);
  assert( pre_end->loopnode()->is_pre_loop(), "" );
  Node *pre_opaq = pre_end->limit();
  assert( pre_opaq->Opcode() == Op_Opaque1, "" );
  Node *pre_limit = pre_opaq->in(1);

  // Where do we put new limit calculations
  Node *pre_ctrl = pre_end->loopnode()->in(LoopNode::EntryControl);

  // Must know if its a count-up or count-down loop
  int stride_con = cl->stride_con();
  Node *zero = _igvn.intcon(0); 
  Node *one  = _igvn.intcon(1);
  set_ctrl( zero->_idx, C->_root );
  set_ctrl( one ->_idx, C->_root );

  // Check loop body for tests of trip-counter plus loop-invariant vs
  // loop-invariant.
  for( uint i = 0; i < loop->_body.size(); i++ ) {
    Node *iff = loop->_body[i];
    if( iff->Opcode() == Op_If ) { // Test?

      // Test is an IfNode, has 2 projections.  If BOTH are in the loop
      // we need loop unswitching instead of iteration splitting.
      Node *exit = loop->is_loop_exit(iff,this);
      if( !exit ) continue;
      int flip = (exit->Opcode() == Op_IfTrue) ? 1 : 0;

      // Get boolean condition to test
      BoolNode *bol = iff->in(1)->is_Bool();
      if( !bol ) continue;
      BoolTest b_test = bol->_test;
      // Flip sense of test if exit condition is flipped
      if( flip )
	b_test = b_test.negate();

      // Get compare
      Node *cmp = bol->in(1);

      // Look for trip_counter + offset vs limit
      Node *offset = cmp->in(1);
      Node *limit  = cmp->in(2);
      int scale_con= 1;		// Assume trip counter not scaled

      Node *limit_c = get_ctrl( limit->_idx );
      if( loop->is_member(get_loop(limit_c) ) ) {
	// Compare might have operands swapped; commute them
	b_test = b_test.commute();
	offset = cmp->in(2);
	limit  = cmp->in(1);
	limit_c = get_ctrl( limit->_idx );
	if( loop->is_member(get_loop(limit_c) ) ) 
	  continue;		// Both inputs are loop varying; cannot RCE
      }
      // Here we know 'limit' is loop invariant
      if( limit_c == ctrl ) continue;// Bail out for being pinned below zero trip test

      // Make sure something stupid hasn't been pinned below the loop
      // guard test, where it's loop-invariant but can't be used by
      // the guard test.
      if( limit_c == ctrl ) continue;
 
      if( offset == trip_counter ) {
	offset = zero;          // Use zero offset
      } else if( offset->is_Add() ) {
        if( offset->in(1) == trip_counter ) 
	  offset = offset->in(2);	// Found offset
        else if( offset->in(2) == trip_counter )
	  offset = offset->in(1);	// Found offset
        else continue;          // Not really (trip_counter+offset)  
      } else if( offset->is_Sub() ) {
        if( offset->in(1) == trip_counter ) {
	  offset = new (3) SubINode( zero, offset->in(2) );// Found offset
	  register_new_node( offset, pre_ctrl );
        } else if( offset->in(2) == trip_counter ) {
	  offset = offset->in(1);	// Found offset
          scale_con = -1;	// Flip scale
        } else continue;        // Not really (trip_counter+offset)  
      } else continue;          // Not really (trip_counter+offset) 

      Node *offset_c = get_ctrl( offset->_idx );
      if( loop->is_member( get_loop(offset_c) ) )
        continue;		// Offset is not really loop invariant
      // Here we know 'offset' is loop invariant.

      // At this point we have the expression as:
      //   scale_con * trip_counter + offset :: limit
      // where scale_con, offset and limit are loop invariant.  Trip_counter 
      // monotonically increases by stride_con, a constant.  Both (or either) 
      // stride_con and scale_con can be negative which will flip about the 
      // sense of the test.

      // Adjust pre and main loop limits to guard the correct iteration set
      if( cmp->Opcode() == Op_CmpU ) {// Unsigned compare is really 2 tests
	if( b_test._test == BoolTest::lt ) { // Range checks always use lt
          // The overflow limit: scale*I+offset < limit
          add_constraint( stride_con, scale_con, offset, limit, pre_ctrl, &pre_limit, &main_limit );
          // The underflow limit: 0 <= scale*I+offset.
          // Some math yields: -scale*I-(offset+1) < 0
          Node *plus_one = new (3) AddINode( offset, one );
	  register_new_node( plus_one, pre_ctrl );
          Node *neg_offset = new (3) SubINode( zero, plus_one );
	  register_new_node( neg_offset, pre_ctrl );
          add_constraint( stride_con, -scale_con, neg_offset, zero, pre_ctrl, &pre_limit, &main_limit );
	} else {
#ifndef PRODUCT
	  if( PrintOpto ) 
	    tty->print_cr("missed RCE opportunity");
#endif
          continue;		// In release mode, ignore it
	}
      } else {			// Otherwise work on normal compares
        switch( b_test._test ) {
        case BoolTest::ge:	// Convert X >= Y to -X <= -Y
	  scale_con = -scale_con;
          offset = new (3) SubINode( zero, offset );
	  register_new_node( offset, pre_ctrl );
          limit  = new (3) SubINode( zero, limit  );
	  register_new_node( limit, pre_ctrl );
	  // Fall into LE case
        case BoolTest::le:	// Convert X <= Y to X < Y+1
          limit = new (3) AddINode( limit, one );
	  register_new_node( limit, pre_ctrl );
	  // Fall into LT case
        case BoolTest::lt: 
          add_constraint( stride_con, scale_con, offset, limit, pre_ctrl, &pre_limit, &main_limit );
          break;
        default:
#ifndef PRODUCT
	  if( PrintOpto ) 
	    tty->print_cr("missed RCE opportunity");
#endif
          continue;             // Unhandled case
        }
      }

      // Kill the eliminated test
      C->_major_progress = true;
      Node *kill_con = _igvn.intcon( 1-flip );
      set_ctrl( kill_con->_idx, C->_root );
      _igvn.hash_delete(iff);
      iff->set_req(1, kill_con, &du );
      _igvn._worklist.push(iff);
      // Find surviving projection
      ProjNode *dp = _igvn._du.out(iff)[0]->is_Proj();
      if( (int)dp->_con != 1-flip ) {
	dp = _igvn._du.out(iff)[1]->is_Proj();
	assert( (int)dp->_con == 1-flip, "" );
      }
      // Find loads off the surviving projection; remove their control edge
      uint    max = _igvn._du.cnt( dp );
      Node **outp = _igvn._du.out( dp );
      for( uint i = 0; i < max; i++ ) {
	Node *cd = outp[i];	// Control-dependent node
	if( cd->is_Load() ) {	// Loads can now float around in the loop
	  _igvn.hash_delete(cd);
	  // Allow the load to float around in the loop, or before it
	  // but NOT before the pre-loop.
	  cd->set_req(0,ctrl,&du); 
	  _igvn._worklist.push(cd);
	  --i;			// Rerun at same iteration
	  max = _igvn._du.cnt(dp); // Reset max
	}
      }

    } // End of is IF

  }

  // Update loop limits
  _igvn.hash_delete(pre_opaq);
  pre_opaq->set_req( 1, pre_limit, &du);

  // Note:: we are making the main loop limit no longer precise;
  // need to round based on stride.
  Node *main_cmp = cl->loopexit()->cmp_node();
  _igvn.hash_delete(main_cmp);
  main_cmp->set_req( 2, main_limit, &du);
  _igvn._worklist.push(main_cmp);
}

//------------------------------DCE_loop_body----------------------------------
// Remove simplistic dead code from loop body
void IdealLoopTree::DCE_loop_body( PhaseDefUse &du ) {
  for( uint i = 0; i < _body.size(); i++ ) 
    if( du.cnt(_body.at(i)) == 0 ) 
      _body.map( i--, _body.pop() );
}
  

//------------------------------adjust_loop_exit_prob--------------------------
// Look for loop-exit tests with my 50/50 guesses from the Parsing stage.
// Replace with a 1-in-10 exit guess.
void IdealLoopTree::adjust_loop_exit_prob( PhaseIdealLoop *phase ) {
  PhaseDefUse &du = phase->_igvn._du;
  Node *test = tail();
  while( test != _head ) {
    uint top = test->Opcode();
    if( top == Op_IfTrue || top == Op_IfFalse ) {
      IfNode *iff = test->in(0)->is_If();
      if( du.cnt(iff) == 2 ) {	// Ignore dead tests
	// Find the OTHER exit path from the IF
	Node *ex = du.out(iff)[0];
	if( ex == test )        // Oops, found self
	  ex = du.out(iff)[1];
	float p = iff->_prob;
	if( !phase->is_member( this, ex ) && iff->_cnt <= -0.9 ) {
	  if( top == Op_IfTrue ) {
	    if( p < 0.501 ) iff->_prob = 0.9f;
	  } else {
	    if( p > 0.499 ) iff->_prob = 0.1f;
	  }
	}
      }
    }
    test = phase->idom( test->_idx );
  }
}
  

//------------------------------policy_do_remove_empty_loop--------------------
// Micro-benchmark spamming.  Policy is to always remove empty loops.
// The 'DO' part is to replace the trip counter with the value it will
// have on the last iteration.  This will break the loop.
bool IdealLoopTree::policy_do_remove_empty_loop( PhaseIdealLoop *phase ) {
  // Minimum size must be empty loop
  if( _body.size() > 7 ) return false;

  CountedLoopNode *cl = _head->is_CountedLoop();
  if( !cl ) return false;
  if( !phase->is_member(this,cl->in(CountedLoopEndNode::TestValue) ) )
    return false;             // Infinite loop
#ifndef PRODUCT
  if( PrintOpto ) 
    tty->print_cr("Removing empty loop");
#endif
  assert( phase->_igvn._du.cnt(cl) == 3, "Self, CLE and phi" );
  // Replace the phi at loop head with the final value of the last
  // iteration.  Then the CountedLoopEnd will collapse (backedge never
  // taken) and all loop-invariant uses of the exit values will be correct.
  Node *phi = cl->phi();
  Node *final = new (3) SubINode( cl->limit(), cl->stride() );
  phase->register_new_node(final,cl->in(LoopNode::EntryControl));
  phase->_igvn.hash_delete(phi);
  phase->_igvn.subsume_node(phi,final);
  phase->C->_major_progress = true;
  return true;
}
  

//=============================================================================
//------------------------------iteration_split_impl---------------------------
void IdealLoopTree::iteration_split_impl( PhaseIdealLoop *phase, Node_List &old_new ) {
  // Check and remove empty loops (spam micro-benchmarks)
  if( policy_do_remove_empty_loop(phase) ) 
    return;                     // Here we removed an empty loop

  bool should_peel = policy_peeling(phase); // Should we peel?

  // Non-counted loops may be peeled; exactly 1 iteration is peeled.
  // This removes loop-invariant tests (usually null checks).
  CountedLoopNode *cl = _head->is_CountedLoop();
  if( !cl ) {                   // Non-counted loop
    if( should_peel ) 	        // Should we peel?
      phase->do_peeling(this,old_new);
    return;
  }

  if( !cl->loopexit() ) return; // Ignore various kinds of broken loops

  // Do nothing special to pre- and post- loops
  if( cl->is_pre_loop() || cl->is_post_loop() ) return;

  // Before attempting fancy unrolling, RCE or alignment, see if we want
  // to completely unroll this loop.
  if( cl->is_normal_loop() ) {
    int trip_count = policy_maximally_unroll(phase);
    if( trip_count > 0 ) {
      // Here we did some unrolling and peeling.  Eventually we will 
      // completely unroll this loop and it will no longer be a loop.
      phase->do_maximally_unroll(this,old_new,trip_count);
      return;
    }
  }


  // Counted loops may be peeled, may need some iterations run up
  // front for RCE, and may want to align loop refs to a cache
  // line.  Thus we clone a full loop up front whose trip count is
  // at least 1 (if peeling), but may be several more.
	
  // The main loop will start cache-line aligned with at least 1
  // iteration of the unrolled body (zero-trip test required) and
  // will have some range checks removed.
	
  // A post-loop will finish any odd iterations (leftover after
  // unrolling), plus any needed for RCE purposes.

  bool should_unroll = policy_unroll(phase);
  
  bool should_rce = policy_range_check(phase);

  bool should_align = policy_align(phase);

  // If not RCE'ing (iteration splitting) or Aligning, then we do not
  // need a pre-loop.  We may still need to peel an initial iteration but
  // we will not be needing an unknown number of pre-iterations.
  //
  // Basically, if may_rce_align reports FALSE first time through, 
  // we will not be able to later do RCE or Aligning on this loop.
  bool may_rce_align = !policy_peel_only(phase) || should_rce || should_align;

  // If we have any of these conditions (RCE, alignment, unrolling) met, then
  // we switch to the pre-/main-/post-loop model.  This model also covers
  // peeling.
  if( should_rce || should_align || should_unroll ) {
    if( cl->is_normal_loop() )  // Convert to 'pre/main/post' loops
      phase->insert_pre_post_loops(this,old_new, !may_rce_align);

    // Adjust the pre- and main-loop limits to let the pre and post loops run
    // with full checks, but the main-loop with no checks.  Remove said
    // checks from the main body.
    if( should_rce ) 
      phase->do_range_check(this,old_new);

    // Double loop body for unrolling.  Adjust the minimum-trip test (will do
    // twice as many iterations as before) and the main body limit (only do
    // an even number of trips).  If we are peeling, we might enable some RCE
    // and we'd rather unroll the post-RCE'd loop SO... do not unroll if
    // peeling.
    if( should_unroll && !should_peel ) 
      phase->do_unroll(this,old_new, true);

    // Adjust the pre-loop limits to align the main body
    // iterations.
    if( should_align )
      Unimplemented();

  } else {		        // Else we have an unchanged counted loop
    if( should_peel )	        // Might want to peel but do nothing else
      phase->do_peeling(this,old_new);
  }
}


//=============================================================================
//------------------------------iteration_split--------------------------------
void IdealLoopTree::iteration_split( PhaseIdealLoop *phase, Node_List &old_new ) {
  // Recursively iteration split nested loops
  if( _child ) _child->iteration_split( phase, old_new );

  // Clean out prior deadwood
  DCE_loop_body(phase->_igvn._du);


  // Look for loop-exit tests with my 50/50 guesses from the Parsing stage.
  // Replace with a 1-in-10 exit guess.
  if( _parent /*not the root loop*/ && 
      !_irreducible && 
      // Also ignore the occasional dead backedge
      tail() != phase->C->_top_node ) {
    adjust_loop_exit_prob(phase);
  }


  // Gate unrolling, RCE and peeling efforts.
  if( !_child &&		// If not an inner loop, do not split
      !_has_call &&		// Do not bother if call; fix inlining instead
      !_irreducible &&
      tail() != phase->C->_top_node && // Also ignore the occasional dead backedge
      LoopOptimizations ) {	// 
    iteration_split_impl( phase, old_new );
  }

  // Minor offset re-organization to remove loop-fallout uses of 
  // trip counter.
  CountedLoopNode *cl = _head->is_CountedLoop();
  if( cl ) phase->reorg_offsets( this );
  /*
  if( OptoReorgOffsets && !_child ) {
    for( uint i = 0; i < _body.size(); i++ ) {
      // Check for a Q+con value alive at the same time as Q
      // Reorganize to lower register pressure.
      Node *add = _body.at(i);
      if( add->Opcode() == Op_AddI && add->in(2)->is_Con() )
        reorg_offsets( add, phase );
    }
  }
  */

  if( _next ) _next->iteration_split( phase, old_new );
}
