/*
 * @(#)memnode.hpp	1.64 01/02/09 14:46:38
 *
 * Copyright 1991-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
// Portions of code courtesy of Clifford Click

class MultiNode;
class PhaseCCP;
class PhaseTransform;

//------------------------------MemNode----------------------------------------
// Load or Store, possibly throwing a NULL pointer exception
class MemNode : public Node {
  virtual uint hash() const;    // Check the type
  virtual uint cmp( const Node &n ) const;
  virtual uint size_of() const; // Size is bigger
public:
  const Type *_type;
  enum { Control,		// When is it safe to do this load?
         Memory,		// Chunk of memory is being loaded from
         Address,		// Actually address, derived from base
         ValueIn		// Value to store
  };
  MemNode( Node *c0, Node *c1, Node *c2, const Type *t ) : Node(c0,c1,c2), _type(t) {}
  MemNode( Node *c0, Node *c1, Node *c2, Node *c3, const Type *t ) : Node(c0,c1,c2,c3), _type(t) {}

  // Is this Node a MemNode or some descendent?  Default is YES.
  virtual MemNode *is_Mem() { return this; }
  virtual Node *Ideal_DU_postCCP( PhaseCCP *ccp );

  virtual uint ideal_reg() const;
  virtual const Type *bottom_type() const;
#ifndef PRODUCT
  virtual void dump_spec() const;
#endif
};

//------------------------------LoadNode---------------------------------------
// Load value; requires Memory and Address
class LoadNode : public MemNode {
public:
	 
  LoadNode( Node *c, Node *mem, Node *adr, const Type *t ) : MemNode(c,mem,adr,t) {}
  virtual int  Opcode() const;
  virtual uint hash()   const;  // Check the type

  // Handle algebraic identities here.	If we have an identity, return the Node
  // we are equivalent to.  We look for Load of a Store.
  virtual Node *Identity( PhaseTransform *phase );

  // If the load is from Field memory and the pointer is non-null, we can
  // zero out the control input.
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);

  // Compute a new Type for this node.	Basically we just do the pre-check,
  // then call the virtual add() to set the type.
  virtual const Type *Value( PhaseTransform *phase ) const;

  // Is this Node an LoadNode or some descendent?  Default is YES.
  virtual const LoadNode *is_Load() const { return this; }
  // Do not match memory edge
  virtual uint match_edge(uint idx) const;

};

//------------------------------LoadBNode--------------------------------------
// Load a byte (8bits signed) from memory
class LoadBNode : public LoadNode {
public:
  LoadBNode( Node *c, Node *mem, Node *adr ) : LoadNode(c,mem,adr,TypeInt::BYTE) {}
  virtual int Opcode() const;
  virtual uint ideal_reg() const { return Op_RegI; }
  virtual Node *Ideal( PhaseGVN *phase, PhaseDefUse *du );
};

//------------------------------LoadCNode--------------------------------------
// Load a char (16bits unsigned) from memory
class LoadCNode : public LoadNode {
public:
  LoadCNode( Node *c, Node *mem, Node *adr ) 
    : LoadNode(c,mem,adr,TypeInt::CHAR) {}
  virtual int Opcode() const;
  virtual uint ideal_reg() const { return Op_RegI; }
  virtual Node *Ideal( PhaseGVN *phase, PhaseDefUse *du );
};

//------------------------------LoadINode--------------------------------------
// Load an integer from memory
class LoadINode : public LoadNode {
public:
  LoadINode( Node *c, Node *mem, Node *adr, const TypeInt *ti = TypeInt::INT ) 
    : LoadNode(c,mem,adr,ti) {}
  virtual int Opcode() const;
  virtual uint ideal_reg() const { return Op_RegI; }
  virtual const Type *Value( PhaseTransform *phase ) const;
};

//------------------------------LoadRangeNode----------------------------------
// Load an array length from the array
class LoadRangeNode : public LoadINode {
public:
  LoadRangeNode( Node *c, Node *mem, Node *adr, const TypeInt *ti ) : LoadINode(c,mem,adr,ti) {}
  virtual int Opcode() const;
  virtual const Type *Value( PhaseTransform *phase ) const;
};

//------------------------------LoadLNode--------------------------------------
// Load a long from memory
class LoadLNode : public LoadNode {
public:
  LoadLNode(Node *c,Node *mem,Node *adr) : LoadNode(c,mem,adr,TypeLong::LONG) {}
  virtual int Opcode() const;
  virtual uint ideal_reg() const { return Op_RegL; }
};

//------------------------------LoadFNode--------------------------------------
// Load a float (64 bits) from memory
class LoadFNode : public LoadNode {
public:
  LoadFNode( Node *c, Node *mem, Node *adr ) 
    : LoadNode(c,mem,adr,Type::FLOAT) {}
  virtual int Opcode() const;
  virtual uint ideal_reg() const { return Op_RegF; }
};

//------------------------------LoadDNode--------------------------------------
// Load a double (64 bits) from memory
class LoadDNode : public LoadNode {
public:
  LoadDNode( Node *c, Node *mem, Node *adr ) 
    : LoadNode(c,mem,adr,Type::DOUBLE) {}
  virtual int Opcode() const;
  virtual uint ideal_reg() const { return Op_RegD; }
};

//------------------------------LoadPNode--------------------------------------
// Load a pointer from memory (either object or array)
class LoadPNode : public LoadNode {
public:
  LoadPNode( Node *c, Node *mem, Node *adr, const TypePtr *t ) 
    : LoadNode(c,mem,adr,t) {}
  virtual int Opcode() const;
  virtual uint ideal_reg() const { return Op_RegP; }
};

//------------------------------LoadKlassNode----------------------------------
// Load a Klass from an object
class LoadKlassNode : public LoadPNode {
public:
  LoadKlassNode( Node *c, Node *mem, Node *adr, const TypePtr *t ) : LoadPNode(c,mem,adr,t) {}
  virtual int Opcode() const;
  virtual const Type *Value( PhaseTransform *phase ) const;
};

//------------------------------LoadSNode--------------------------------------
// Load a short (16bits signed) from memory
class LoadSNode : public LoadNode {
public:
  LoadSNode( Node *c, Node *mem, Node *adr ) 
    : LoadNode(c,mem,adr,TypeInt::SHORT) {}
  virtual int Opcode() const;
  virtual uint ideal_reg() const { return Op_RegI; }
  virtual Node *Ideal( PhaseGVN *phase, PhaseDefUse *du );
};


//------------------------------StoreNode--------------------------------------
// Store value; requires Store, Address and Value
class StoreNode : public MemNode {
  virtual uint cmp( const Node &n ) const;
  virtual bool depends_only_on_test() const { return false; }
public:
  StoreNode( Node *c, Node *mem, Node *adr, Node *val, const Type *t ) : MemNode(c,mem,adr,val,t) {}
  virtual int  Opcode() const;
  virtual uint hash()   const;  // Check the type

  // If the store is to Field memory and the pointer is non-null, we can
  // zero out the control input.
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);

  // Compute a new Type for this node.	Basically we just do the pre-check,
  // then call the virtual add() to set the type.
  virtual const Type *Value( PhaseTransform *phase ) const;

  // Check for identity function on memory (Load then Store at same address)
  Node *Identity( PhaseTransform *phase );

  // Is this Node an StoreNode or some descendent?  Default is YES.
  virtual const StoreNode *is_Store() const { return this; }
  // Do not match memory edge
  virtual uint match_edge(uint idx) const;

};

//------------------------------StoreBNode-------------------------------------
// Store byte to memory
class StoreBNode : public StoreNode {
public:
  StoreBNode( Node *c, Node *mem, Node *adr, Node *val, const Type *t ) : StoreNode(c,mem,adr,val,t) {}
  virtual int Opcode() const;
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
  virtual const RegMask &out_RegMask() const;
};

//------------------------------StoreCNode-------------------------------------
// Store char/short to memory
class StoreCNode : public StoreNode {
public:
  StoreCNode( Node *c, Node *mem, Node *adr, Node *val, const Type *t ) : StoreNode(c,mem,adr,val,t) {}
  virtual int Opcode() const;
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
  virtual const RegMask &out_RegMask() const;
};

//------------------------------StoreINode-------------------------------------
// Store int to memory
class StoreINode : public StoreNode {
public:
  StoreINode( Node *c, Node *mem, Node *adr, Node *val, const Type *t ) : StoreNode(c,mem,adr,val,t) {}
  virtual int Opcode() const;
  virtual const RegMask &out_RegMask() const;
};

//------------------------------StoreLNode-------------------------------------
// Store int to memory
class StoreLNode : public StoreNode {
public:
  StoreLNode( Node *c, Node *mem, Node *adr, Node *val, const Type *t ) : 
    StoreNode(c,mem,adr,val,t) {}
  virtual int Opcode() const;
  virtual const RegMask &out_RegMask() const;
};

//------------------------------StoreFNode-------------------------------------
// Store float to memory
class StoreFNode : public StoreNode {
public:
  StoreFNode( Node *c, Node *mem, Node *adr, Node *val, const Type *t ) : StoreNode(c,mem,adr,val,t) {}
  virtual int Opcode() const;
  virtual const RegMask &out_RegMask() const;
};

//------------------------------StoreDNode-------------------------------------
// Store double to memory
class StoreDNode : public StoreNode {
public:
  StoreDNode( Node *c, Node *mem, Node *adr, Node *val, const Type *t ) : StoreNode(c,mem,adr,val,t) {}
  virtual int Opcode() const;
  virtual const RegMask &out_RegMask() const;
};

//------------------------------StorePNode-------------------------------------
// Store pointer to memory
class StorePNode : public StoreNode {
public:
  StorePNode( Node *c, Node *mem, Node *adr, Node *val, const Type *t ) : StoreNode(c,mem,adr,val,t) {}
  virtual int Opcode() const;
  virtual const RegMask &out_RegMask() const;
};

//------------------------------ClearArray-------------------------------------
class ClearArrayNode: public Node {
public:
  ClearArrayNode( Node *ctrl, Node *arymem, Node *word_cnt, Node *base ) : Node(ctrl,arymem,word_cnt,base) {}
  virtual int         Opcode() const;
  virtual const Type *bottom_type() const { return Type::MEMORY; }
  virtual Node *Identity( PhaseTransform *phase );
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
  virtual uint match_edge(uint idx) const;
};

//------------------------------StrComp-------------------------------------
class StrCompNode: public Node {
public:
  StrCompNode(Node *control,
	      Node* char_array_mem,
	      Node* value_mem,
	      Node* count_mem,
	      Node* offset_mem,
	      Node* s1, Node* s2): Node(control,
					char_array_mem,
					value_mem,
					count_mem,
					offset_mem,
					s1, s2) {};
  virtual int Opcode() const;
  virtual const Type* bottom_type() const { return TypeInt::INT; }
  virtual uint match_edge(uint idx) const;
  virtual uint ideal_reg() const { return Op_RegI; }
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
};

//------------------------------MemBar-----------------------------------------
class MemBarNode: public MultiNode {
  virtual uint cmp( const Node &n ) const ;    // Always fail, except on self
public:
  MemBarNode(Node *control, Node *memory);
  virtual int Opcode() const;
  virtual const Type *Value( PhaseTransform *phase ) const;
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
  virtual uint ideal_reg() const { return NotAMachineReg; }
  virtual uint match_edge(uint idx) const { return 0; }
  virtual const Type *bottom_type() const { return TypeTuple::MEMBAR; }
  virtual Node *match( const ProjNode *proj, const Matcher *m );
};

//------------------------------MergeMem---------------------------------------
class MergeMemNode: public Node {
  virtual uint cmp( const Node &n ) const ;    // Always fail, except on self
public:
  MergeMemNode(Node *def,uint cnt);
  virtual int Opcode() const;
  virtual Node *Identity( PhaseTransform *phase );
  virtual Node *Ideal(PhaseGVN *phase,PhaseDefUse *du);
  virtual uint ideal_reg() const { return NotAMachineReg; }
  virtual uint match_edge(uint idx) const { return 0; }
  virtual const RegMask &out_RegMask() const;
  virtual const Type *bottom_type() const { return Type::MEMORY; }
#ifndef PRODUCT
  virtual void dump_spec() const;
#endif
};
 
//------------------------------Prefetch---------------------------------------
// Non-faulting prefetch load.  Prefetch for many reads & many writes.
class PrefetchNode : public Node {
public:
  PrefetchNode(Node *abio, Node *adr) : Node(0,abio,adr) {}
  virtual int Opcode() const;
  virtual uint ideal_reg() const { return NotAMachineReg; }
  virtual uint match_edge(uint idx) const { return idx==2; }
  virtual const Type *bottom_type() const { return Type::ABIO; }
};
 
