/*
 * @(#)node.hpp	1.149 01/02/14 09:21:09
 *
 * Copyright 1991-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
// Portions of code courtesy of Clifford Click

// Optimization - Graph Style


class AddNode;
class AddPNode;
class AliasInfo;
class Block;
class Block_Array;
class BoolNode;
class BoxLockNode;
class CallNode;
class CatchNode;
class CatchProjNode;
class CMoveNode;
class CodeBuffer;
class ConNode;
class CopyNode;
class CountedLoopNode;
class FastLockNode;
class IfNode;
class JVMStateObj;
class JumpProjNode;
class LoadNode;
class LockNode;
class LoopNode;
class MachNode;
class Matcher;
class MemNode;
class MulNode;
class MultiNode;
class Node;
class Node_Array;
class Node_List;
class NullCheckNode;
class OopMap;
class PCTableNode;
class PhaseChaitin;
class PhaseCCP;
class PhaseDefUse;
class PhaseGVN;
class PhaseIterGVN;
class PhaseTransform;
class PhiNode;
class Pipeline;
class ProjNode;
class RegMask;
class RegionNode;
class RootNode;
class SafePointNode;
class SpillCopyNode;
class StartC2INode;
class StartI2CNode;
class StartNode;
class State;
class StoreNode;
class SubNode;
class TypeNode;
class UnlockNode;
class VectorSet;
class FastLockNode;
typedef void (*NFunc)(Node&,void*);
extern "C" {
  typedef int (*C_sort_func_t)(const void *, const void *); 
}


//------------------------------Node-------------------------------------------
// Nodes define actions in the program.  They create values, which have types.
// They are both vertices in a directed graph and program primitives.  Nodes
// are labeled; the label is the "opcode", the primitive function in the lambda
// calculus sense that gives meaning to the Node.  Node inputs are ordered (so
// that "a-b" is different from "b-a").  The inputs to a Node are the inputs to
// the Node's function.  These inputs also define a Type equation for the Node.
// Solving these Type equations amounts to doing dataflow analysis.
// Control and data are uniformly represented in the graph.  Finally, Nodes
// have a unique dense integer index which is used to index into side arrays
// whenever I have phase-specific information.

class Node {
  // Lots of restrictions on cloning Nodes
  Node(const Node&);            // not defined; linker error to use these
  Node &operator=(const Node &rhs);
  // Hidden complex du update
  void set_req_du( uint i, Node *n, PhaseDefUse *du );

public:
  friend class Compile;

  // Because Nodes come and go, I define an Arena of Node structures to pull
  // from.  This should allow fast access to node creation & deletion.  This
  // field is a local cache of a value defined in some "program fragment" for
  // which these Nodes are just a part of.
  void *operator new( size_t x ) { return Compile::current()->_arena.Amalloc_4(x); }
  // Version of new that also allocates Node array
  void *operator new( size_t x, uint y ) { 
    void *thsi = (char*)Compile::current()->_arena.Amalloc_4(x+y*sizeof(void*))+y*sizeof(void*);
#ifdef ASSERT
    ((void**)thsi)[0-y] = thsi; // magic cookie for assertion check
#endif
    return thsi; }
  // Delete is a NOP
  void operator delete( void *ptr ) {}
  // Fancy destructor; eagerly attempt to reclaim Node numberings and storage
  void destruct();

  // Create a new Node.  Required is the number is of inputs required for
  // semantic correctness.
  Node( uint required );

  // Create a new Node with given input edges.
  // This version requires use of the "edge-count" new.  
  // E.g.  new (C,3) FooNode( C, NULL, left, right );
  Node( Node *n0 );
  Node( Node *n0, Node *n1 );
  Node( Node *n0, Node *n1, Node *n2 );
  Node( Node *n0, Node *n1, Node *n2, Node *n3 );
  Node( Node *n0, Node *n1, Node *n2, Node *n3,
	    Node *n4, Node *n5, Node *n6 );

  // Clone an inherited Node given only the base Node type.
  virtual Node *clone( ) const;


//----------------- input edge handling
protected:
  friend class PhaseCFG;        // Access to address of _in array elements
  Node **_in;                   // Array of use-def references to Nodes

  // Input edges are split into two catagories.  Required edges are required
  // for semantic correctness; order is important and NULLs are allowed.
  // Precedence edges are used to help determine execution order and are
  // added, e.g., for scheduling purposes.  They are unordered and not
  // duplicated; they have no embedded NULLs.  Edges from 0 to in_required-1
  // are required, from in_required to in_len-1 are precedence edges.
  uint16 _cnt;			// Total number of required Node inputs.

  uint16 _max;			// Actual length of input array.

  // Grow the actual input array to the next larger power-of-2 bigger than len.
  void grow( uint len );

 public:
  // Each Node is assigned a unique small/dense number.  This number is used
  // to index into auxiliary arrays of data and bitvectors.
  uint16 _idx;
  uint16 _pad_to_16_bytes;

  // Get the (read-only) number of input edges
  uint req() const { return _cnt; }
  uint len() const { return _max; }

  // Reference to the i'th input Node.  Error if out of bounds.
  Node* in(uint i) const { assert(i < _max,"oob"); return _in[i]; }
  // Set a required input edge
  void add_req( Node *n ); // Append a NEW required input
  void del_req( uint idx ); // Delete required edge & compact
  void ins_req( uint i, Node *n ); // Insert a NEW required input
  void set_req( uint i, Node *n ) { assert( i < _cnt, "oob"); _in[i] = n; }
  // Set-required edge with optional def-use info
  void set_req( uint i, Node *n, PhaseDefUse *du ) { 
    assert( i < _cnt, "oob"); if( du ) set_req_du(i,n,du); else _in[i] = n; }
  void set_req_X( uint i, Node *n, PhaseIterGVN *igvn );
  // Remove inputs from this Node that are NULL in the region Node.
  // "this" may or may not be equal to "region".
  void pack_req( const Node *region );
  // Find the one non-null required input.  RegionNode only
  Node *nonnull_req() const;
  // Add or remove precedence edges
  void add_prec( Node *n );
  void rm_prec( uint i );
  // Set this node's index, used by cisc_version to replace current node
  void set_idx( uint idx ) { _idx = idx; }

  // Copy edge array into location passed in
  void edge_copy(Node **edges, int length) const;

  // Iterators over input Nodes for a Node X are written as:
  // for( i = 0; i < X.req(); i++ ) ... X[i] ...
  // NOTE: Required edges can contain embedded NULL pointers.

//----------------- Other Node Properties

  // Return a dense integer opcode number
  virtual int Opcode() const;
  
  // This call defines a class-unique string used to identify class instances
  virtual const char *Name() const;

  // Virtual inherited Node size
  virtual uint size_of() const;

  // Other interesting Node properties
  virtual       AddPNode      *is_AddP      ()       { return 0; }
  virtual       BoolNode      *is_Bool      ()       { return 0; }
  virtual       CallNode      *is_Call      ()       { return 0; }
  virtual       CMoveNode     *is_CMove     ()       { return 0; }
  virtual       LockNode      *is_Lock      ()       { return 0; }
  virtual       LoopNode      *is_Loop      ()       { return 0; }
  virtual       CountedLoopNode*is_CountedLoop()     { return 0; }
  virtual       MachNode      *is_Mach      ()       { return 0; }
  virtual       MemNode       *is_Mem       ()       { return 0; }
  virtual       MultiNode     *is_Multi     ()       { return 0; }
  virtual       PhiNode       *is_Phi       ()       { return 0; }
  virtual       ProjNode      *is_Proj      ()       { return 0; }
  virtual       RootNode      *is_Root      ()       { return 0; }
  virtual       SafePointNode *is_SafePoint ()       { return 0; }
  virtual       SpillCopyNode *is_SpillCopy ()       { return 0; }
  virtual       StartNode     *is_Start     ()       { return 0; }
  virtual       SubNode       *is_Sub       ()       { return 0; }
  virtual       TypeNode      *is_Type      ()       { return 0; }
  virtual       UnlockNode    *is_Unlock    ()       { return 0; }
  virtual const AddNode       *is_Add       () const { return 0; }
  virtual const BoxLockNode   *is_BoxLock   () const { return 0; }
  virtual const CatchNode     *is_Catch     () const { return 0; }
  virtual const CatchProjNode *is_CatchProj () const { return 0; }
  virtual const FastLockNode  *is_FastLock  () const { return 0; }
  virtual const bool          is_SpecLock   () const { return false; }
  virtual const bool          maybe_SpecLock() const { return false; }
  virtual       IfNode        *is_If        ()       { return 0; }
  virtual const JumpProjNode  *is_JumpProj  () const { return 0; }
  virtual const LoadNode      *is_Load      () const { return 0; }
  virtual const MulNode       *is_Mul       () const { return 0; }
  virtual const NullCheckNode *is_NullCheck () const { return 0; }
  virtual const PCTableNode   *is_PCTable   () const { return 0; }
  virtual const RegionNode    *is_Region    () const { return 0; }
  virtual const StoreNode     *is_Store     () const { return 0; }
  virtual const SubNode       *is_Cmp       () const { return 0; }

  virtual uint  is_Con   () const { return 0; }
  virtual uint  is_Copy  () const { return 0; } // return copied edge index
  virtual uint  is_Goto  () const { return 0; }
  virtual bool  is_CFG   () const { return false; }
  // If this node is control-dependent on a test, can it be
  // rerouted to a dominating equivalent test?  This is usually
  // true of non-CFG nodes, but can be false for operations which
  // depend for their correct sequencing on more than one test.
  // (In that case, hoisting to a dominating test may silently
  // skip some other important test.)
  virtual bool depends_only_on_test() const { assert(!is_CFG(), ""); return true; };

  // defined for MachNodes that match 'If' | 'Goto' | 'CountedLoopEnd'
  virtual uint  is_Branch() const { return 0; }

  // must be the last instruction in a block
  virtual uint  is_last_instr() const { return 0; }

  // When building basic blocks, I need to have a notion of block beginning
  // Nodes, next block selector Nodes (block enders), and next block
  // projections.  These calls need to work on their machine equivalents.  The
  // Ideal beginning Nodes are RootNode, RegionNode and StartNode.
  virtual int is_block_start() const;

  // The Ideal control projection Nodes are IfTrue/IfFalse, Root,
  // Goto and Return.  This call also returns the block ending Node.
  virtual const Node *is_block_proj() const;

//----------------- Optimization

  // Get the worst-case Type output for this Node.
  virtual const class Type *bottom_type() const;

  // Return an existing node which computes the same function as this node.
  // The optimistic combined algorithm requires this to return a Node which 
  // is a small number of steps away (e.g., one of my inputs).
  virtual Node *Identity( PhaseTransform *phase );

  // Return the set of values this Node can take on at runtime.
  virtual const Type *Value( PhaseTransform *phase ) const;

  // Return a node which is more "ideal" than the current node.  
  // The invariants on this call are subtle.  If in doubt, read the
  // treatise in node.cpp above the default implemention AND TEST WITH
  // +VerifyIterativeGVN!
  virtual Node *Ideal( PhaseGVN *phase, PhaseDefUse *du );
protected:
  bool remove_dead_region( PhaseGVN *phase, PhaseDefUse *du );
public:

  // Idealize graph, using DU info.  Done after constant propagation
  virtual Node *Ideal_DU_postCCP( PhaseCCP *ccp );

  // See if there is valid pipeline info
  virtual const Pipeline *pipeline() const;
  bool valid_pipeline_info() const { return (pipeline() != NULL); }

  // Compute the latency from the def to this instruction of the ith input node
  uint latency(uint i);

  // Compute the (backwards) latency from the uses of this instructions
  int latency_from_uses(PhaseDefUse &du, Block_Array &bbs, GrowableArray<uint> &node_latency) const;

  // Compute the (backwards) latency from a single use
  int latency_from_use (PhaseDefUse &du, Block_Array &bbs, GrowableArray<uint> &node_latency, const Node *def, Node *use) const;

  // Compute the (backwards) latency from the uses of this instructions
  void partial_latency_of_defs(Block_Array &bbs, GrowableArray<uint> &node_latency);

  // Compute the (backwards) local (intra-bb) latency from the uses of this instructions
  void partial_latency_of_local_defs(Block_Array &bbs, GrowableArray<uint> &node_latency);

  // Hash & compare functions, for pessimistic value numbering
  virtual uint hash() const;
  virtual uint cmp( const Node &n ) const;

//----------------- Code Generation

  // Ideal register class for Matching.  Zero means unmatched instruction
  // (these are cloned instead of converted to machine nodes).
  virtual uint ideal_reg() const;

  static const uint NotAMachineReg;   // must be > max. machine register

  // Do we Match on this edge index or not?  Generally false for Control
  // and true for everything else.  Weird for calls & returns.
  virtual uint match_edge(uint idx) const;

  // Register class output is returned in
  virtual const RegMask &out_RegMask() const;
  // Register class input is expected in
  virtual const RegMask &in_RegMask(uint) const;
  // Should we clone rather than spill this instruction?
  virtual bool rematerialize() const { return false; }

  // Return JVM State Object if this Node carries debug info, or NULL otherwise
  virtual JVMStateObj *jvms() const;
  virtual JVMStateObj *jvmtail() const;
  virtual void set_oop_map(OopMap *map);
  virtual OopMap *oop_map() const;

  // Print as assembly
  virtual void format( PhaseChaitin * ) const;
  // Emit bytes starting at parameter 'ptr'
  // Bump 'ptr' by the number of output bytes 
  virtual void emit(CodeBuffer &cbuf, PhaseChaitin *pc_) const;
  // Size of instruction in bytes
  virtual uint size(PhaseChaitin *pc_) const;

  // Convenience function to extract an integer constant from a ConstNode.
  // Returns True if it is an integer constant ConstNode and puts the constant
  // into 'con', otherwise it returns False and 'con' is untouched.
  virtual int get_int( jint *con ) const;
  // Return the constant, knowing it is an integer constant already
  jint get_int( ) const;
  jlong get_long( ) const;
  jdouble getd( ) const;
  jfloat getf( ) const;

  // Part of Global Code Motion.
  // Schedule Nodes early in their basic blocks
  bool schedule_early( VectorSet &visited, Node_List &roots, Block_Array &bbs);
  // Nodes which are pinned into basic blocks
  virtual int   pinned() const { return 0; }
  // Nodes which use memory without consuming it, hence need antidependences
  // More specifically, check_for_anti_dependence returns true iff the node
  // (a) does a load, and (b) does not perform a store (except perhaps to a
  // stack slot or some other unaliased location).
  virtual bool  check_for_anti_dependence() const { return false; }
  // Look into this node to see what memory it accesses.
  virtual Node *load() const { return NULL; }
  virtual int   may_alias( Node * ) { return 1; }
  // Return 'true' if we split a Node so it can be post-dominated by uses
  bool          split_for_uses(Block *LCA, PhaseDefUse &du, Block_Array &bbs);

  // Return which operand this instruction may cisc-spill
  virtual int   cisc_operand() const { return AdlcVMDeps::Not_cisc_spillable; }

private:
  // True if Node is post-dominated by uses [helper for split_for_uses()]
  bool check_post_dom( Block *LCA, PhaseDefUse &du, Block_Array &bbs ) const;

//----------------- Graph walking
public:
  // Walk and apply member functions recursively.
  // Supplied (this) pointer is root.
  void walk(NFunc pre, NFunc post, void *env);
  static void nop(Node &, void*); // Dummy empty function
  static void packregion( Node &n, void* );
private:
  void walk_(NFunc pre, NFunc post, void *env, VectorSet &visited);

//----------------- Printing, etc
public:
#ifndef PRODUCT
  Node* find(int idx) const;         // Search the graph for the given idx.
  Node* find_ctrl(int idx) const;    // Search control ancestors for the given idx.
  void dump() const;                 // Print this node,
  void dump(int depth) const;        // Print this node, recursively to depth d
  void dump_ctrl(int depth) const;   // Print control nodes, to depth d
  virtual void dump_req() const;     // Print required-edge info
  virtual void dump_prec() const;    // Print precedence-edge info
  virtual void dump_spec() const {}; // Print per-node info
  void verify(PhaseDefUse *du) const; // Check Def-Use info for my subgraph
  static void verify_recur(const Node *n, int verify_depth, PhaseDefUse *du, VectorSet &old_space, VectorSet &new_space);
  void dump_format(PhaseChaitin *pc) const; // debug access to MachNode::format(...)
  // RegMask Print Functions
  void dump_in_regmask(int idx) { in_RegMask(idx).dump(); }
  void dump_out_regmask() { out_RegMask().dump(); }
  void fast_dump() const {
    tty->print("%4d: %-17s", _idx, Name());
    for (uint i = 0; i < len(); i++)
      if (in(i))
        tty->print(" %4d", in(i)->_idx);
      else
        tty->print(" NULL");
    tty->print("\n");
  }
#endif
#ifdef ASSERT
  virtual void verify_construction();
  int  _debug_idx;
  int   debug_idx() const              { return _debug_idx; }
  void  set_debug_idx( int debug_idx ) { _debug_idx = debug_idx; }
#endif
};

//-----------------------------------------------------------------------------
// Map dense integer indices to Nodes.  Uses classic doubling-array trick.
// Abstractly provides an infinite array of Node*'s, initialized to NULL.
// Note that the constructor just zeros things, and since I use Arena 
// allocation I do not need a destructor to reclaim storage.
class Node_Array : public ResourceObj {
protected:
  Arena *_a;                    // Arena to allocate in
  uint   _max;
  Node **_nodes;
  void   grow( uint i );        // Grow array node to fit
public:
  Node_Array(Arena *a) : _a(a), _max(OptoNodeListSize) {
    _nodes = NEW_ARENA_ARRAY( a, Node *, OptoNodeListSize );
    for( int i = 0; i < OptoNodeListSize; i++ ) {
      _nodes[i] = NULL;
    }
  }

  Node_Array(Node_Array *na) : _a(na->_a), _max(na->_max), _nodes(na->_nodes) {}
  Node *operator[] ( uint i ) const // Lookup, or NULL for not mapped
  { return (i<_max) ? _nodes[i] : (Node*)NULL; }
  Node *at( uint i ) const { assert(i<_max,"oob"); return _nodes[i]; }
  Node **adr() { return _nodes; }
  // Extend the mapping: index i maps to Node *n.
  void map( uint i, Node *n ) { if( i>=_max ) grow(i); _nodes[i] = n; }
  void insert( uint i, Node *n );
  void remove( uint i );        // Remove, preserving order
  void sort( C_sort_func_t func);
  void reset( Arena *new_a );   // Zap mapping to empty; reclaim storage
  void clear();                 // Set all entries to NULL, keep storage
  uint Size() const { return _max; }
  void dump() const;
};

class Node_List : public Node_Array {
  uint _cnt;
public:
  Node_List() : Node_Array(Thread::current()->resource_area()), _cnt(0) {}
  Node_List(Arena *a) : Node_Array(a), _cnt(0) {}
  void insert( uint i, Node *n ) { Node_Array::insert(i,n); _cnt++; }
  void remove( uint i ) { Node_Array::remove(i); _cnt--; }
  void push( Node *b ) { map(_cnt++,b); }
  void yank( Node *n );         // Find and remove
  Node *pop() { return _nodes[--_cnt]; }
  Node *rpop() { Node *b = _nodes[0]; _nodes[0]=_nodes[--_cnt]; return b;}
  void clear() { _cnt = 0; Node_Array::clear(); } // retain storage
  uint size() const { return _cnt; }
  void dump() const;
};

//------------------------------Unique_Node_List-------------------------------
class Unique_Node_List : public Node_List {
  VectorSet _in_worklist;
  uint _clock_index;            // Index in list where to pop from next
public:
  Unique_Node_List() : Node_List(), _in_worklist(Thread::current()->resource_area()), _clock_index(0) {}
  Unique_Node_List(Arena *a) : Node_List(a), _in_worklist(a), _clock_index(0) {}

  void remove( Node *n );
  bool member( Node *n ) { return _in_worklist.test(n->_idx); }
  void push( Node *b ) { 
    if( !_in_worklist.test_set(b->_idx) ) 
      Node_List::push(b);
  }
  Node *pop() {
    if( _clock_index >= size() ) _clock_index = 0;
    Node *b = at(_clock_index);
    map( _clock_index++, Node_List::pop());
    _in_worklist >>= b->_idx;
    return b;
  }
  Node *remove( uint i ) {
    Node *b = Node_List::at(i);
    _in_worklist >>= b->_idx;
    map(i,Node_List::pop());
    return b;
  }
  void yank( Node *n ) { _in_worklist >>= n->_idx; Node_List::yank(n); }
  void  clear() {
    _in_worklist.Clear();        // Discards storage but grows automatically
    Node_List::clear();
    _clock_index = 0;
  }

#ifndef PRODUCT
  void print_set() const { _in_worklist.print(); }
#endif
};

//------------------------------TypeNode---------------------------------------
// Node with a Type constant.
class TypeNode : public Node {
  virtual uint hash() const;    // Check the type
  virtual uint cmp( const Node &n ) const;
  virtual uint size_of() const; // Size is bigger
public:
  const Type *_type;
  const Type *get_type() const { return _type; };
  TypeNode( const Type *t, uint required ) : Node(required), _type(t) {}
  virtual const Type *Value( PhaseTransform *phase ) const;
  virtual const Type *bottom_type() const;
  virtual       uint  ideal_reg() const;
  virtual   TypeNode *is_Type() { return this; }
#ifndef PRODUCT
  virtual void dump_spec() const;
#endif
};

