/*
 * @(#)output.hpp	1.8 01/02/16 11:00:42
 *
 * Copyright 1997-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
class Arena;
class Bundle;
class Block;
class Block_Array;
class Node;
class Node_Array;
class Node_List;
class PhaseCFG;
class PhaseChaitin;
class Pipeline_Use_Element;
class Pipeline_Use;

#ifndef PRODUCT
#define DEBUG_ARG(x) , x
#else
#define DEBUG_ARG(x)
#endif

//------------------------------Scheduling----------------------------------
// This class contains all the information necessary to implement instruction
// scheduling and bundling.
class Scheduling {

private:
  // Arena to use
  Arena *_arena;

  // Control-Flow Graph info
  PhaseCFG *_cfg;

  // Register Allocation info
  PhaseChaitin *_chaitin;

  // Number of nodes in the method
  uint _node_bundle_limit;

  // List of scheduled nodes. Generated in reverse order
  Node_List _scheduled;

  // List of nodes currently available for choosing for scheduling
  Node_List _available;

  // List of machine projection nodes to be scheduled
  Node_List _mach_projs;

  // Mapping from node (index) to basic block
  Block_Array& _bbs;

  // For each instruction beginning a bundle, the number of following
  // nodes to be bundled with it.
  Bundle *_node_bundling;
  
  // Latency from the beginning of the containing basic block (base 1)
  // for each node.
  unsigned short *_node_latency;

  // Number of uses of this node within the containing basic block.
  short *_uses;

  // Indicates if node has been seen
  uint *_found;

  // Indicates if node visited in scan
  uint *_visited;

  // Latency from the end of the basic block as scheduled
  unsigned short *_current_latency;

  // Mapping from register kind to instruction storing to it
  Node_Array _next_assignment;

  // Kludge for bad spillcopies
  Node_Array _spillcopy_write;

  // Remember the next call; it blows away lots of registers
  Node *_next_call;

  // Remember the next node
  Node *_next_node;

  // List of nodes that write Oops; used at Safepoints to prevent move Oops
  Node_List _writes_oop;

  // Use this for an unconditional branch delay slot
  Node *_unconditional_delay_slot;

  // Pointer to the MachProj node for the frame pointer
  Node *_frame_pointer;

  // Pointer to a Nop
  MachNopNode *_nop;

  // Number of header instructions in the first block
  uint _entry_instructions;

  // Length of the current bundle, in instructions
  uint _bundle_instr_count;

  // Current Cycle number, for computing latencies and bundling
  uint _bundle_cycle_number;

  // Size of bit vectors, in bytes
  uint _bit_vector_size;

  // Bundle information
  Pipeline_Use_Element _bundle_use_elements[resource_count];
  Pipeline_Use         _bundle_use;

  // Dump the available list
  void dump_available() {
    tty->print("#     ");
    for (uint i = 0; i < _available.size(); i++)
      tty->print(" %3d", _available[i]->_idx);
    if (_mach_projs.size() > 0) {
      tty->print(" -");
      for (uint i = 0; i < _mach_projs.size(); i++)
        tty->print(" %3d", _mach_projs[i]->_idx);
    }
    tty->print("\n");
  }

public:
  Scheduling(Arena *arena, Compile &compile);

  // Destructor
  ~Scheduling();

  // Step ahead "i" cycles
  void step(uint i);

  // Step ahead 1 cycle, and clear the bundle state (for example,
  // at a branch target)
  void step_and_clear();

  void set_visited(uint i) {
    _visited[i >> 5] |= ((uint)1) << (i & 31);
  }

  bool not_visited(uint i) {
    return ((_visited[i >> 5] & (((uint)1) << (i & 31))) == 0);
  }

  void clear_found_vector() {
    memset(_found, 0, _bit_vector_size);
  }

  void set_found(uint i) {
    _found[i >> 5] |= ((uint)1) << (i & 31);
  }

  bool not_found(uint i) {
    return ((_found[i >> 5] & (((uint)1) << (i & 31))) == 0);
  }

  bool found(uint i) {
    return ((_found[i >> 5] & (((uint)1) << (i & 31))) != 0);
  }

  Bundle* node_bundling(const Node *n) {
    return (&_node_bundling[n->_idx]);
  }

  bool valid_bundle_info(const Node *n) const {
    return (_node_bundle_limit > n->_idx);
  }

  // Do the scheduling
  void DoScheduling();

  // Compute the local latencies walking forward over the list of
  // nodes for a basic block
  void ComputeLocalLatenciesForward(const Block *bb);

  // Compute the register antidependencies within a basic block
  void ComputeRegisterAntidependencies(const Block *bb);

  // Add a node to the current bundle
  void AddNodeToBundle(Node *n);

  // Add a node to the list of available nodes
  void AddNodeToAvailableList(Node *n);

  // Compute the local use count for the nodes in a block, and compute
  // the list of instructions with no uses in the block as available
  void ComputeUseCount(const Block *bb);

  // Choose an instruction from the available list to add to the bundle
  Node * ChooseNodeToBundle();

  // See if this Node fits into the currently accumulating bundle
  bool NodeFitsInBundle(Node *n);

  // Decrement the use count for a node
 void DecrementUseCounts(Node *n);

  // Add an antidependence input
  void AddAntidependency(Node *src, Node *dst
      DEBUG_ARG(OptoReg::Name nam)
      DEBUG_ARG(const char * msg)
    );

  // Add a read of a register to dependence processing
  void ReadRegister(OptoReg::Name register, Node *n
      DEBUG_ARG(const char * comment)
    );

  // Add a write of a register to dependence processing
  void WriteRegister(OptoReg::Name register, Node *n
      DEBUG_ARG(const char * comment)
    );

  // Compute the register number from the offset
  OptoReg::Name OffsetToReg(int offset) const;

  // Add antidependencies for a spillcopy kludge
  void InsertAntidependenciesForSpillCopy(const Block *bb, uint node_idx, Node *spillcopy, Node *upper_limit);

  // Get the register written to by this instruction; include stores to
  // the stack (which the Chaitin::get_register may miss)
  OptoReg::Name get_register(Node *n) const;
  OptoReg::Name get_register(Node *n, bool &is_double) const;
  
  // Information for statistics gathering
#ifndef PRODUCT
private:
  // Gather information on size of nops relative to total
  uint _branches, _unconditional_delays;

  static uint _total_nop_size, _total_method_size;
  static uint _total_branches, _total_unconditional_delays;
  static uint _total_instructions_per_bundle[Pipeline::_max_instrs_per_bundle+1];

public:
  static void print_statistics();

  static void increment_instructions_per_bundle(uint i) {
    _total_instructions_per_bundle[i]++;
  }

  static void increment_nop_size(uint s) {
    _total_nop_size += s;
  }

  static void increment_method_size(uint s) {
    _total_method_size += s;
  }
#endif

};

