/*
 * @(#)parse.hpp	1.202 01/02/09 14:46:49
 *
 * Copyright 1997-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// CopyrightVersion 1.2
class FastLockNode;
class FastUnlockNode;
class InlineTree;
class Parse;
class RootNode;

//-----------------------------------------------------------------------------
// Supporting data structures

// Parser state at start of basic block
class Bci2Node : public StackObj {
 public:
   enum Constants {
     MaxJsrDepth = 10
   };
 private:
  int             _bci;
  int             _sp;
  SafePointNode*  _map;
  GrowableArray<int>* _jsr_stack; // NULL means empty (too)
 public:  
   Bci2Node() { clear(); }
  
  // Accessers
  int                 bci() const           { return _bci; }
  int                 sp() const            { return _sp; }
  SafePointNode*      map() const           { return _map; }
  GrowableArray<int>* jsr_stack() const     { return _jsr_stack; }

  int jsr_stack_depth() const               { return (_jsr_stack == NULL) ? 0 : _jsr_stack->length(); }

  void set_bci(int bci)                     { _bci = bci; }  
  void set_sp(int sp)                       { _sp = sp; }    
  void set_map(SafePointNode* map)          { _map = map; }  
  void set_jsr_stack(GrowableArray<int>* s) { _jsr_stack = s; }

  // Initialization
  void clear() { _bci = _sp = 0; _jsr_stack = NULL; _map = NULL; }
  
  // Copying
  void put(Bci2Node* n) { copy(this, n); }
  void get(Bci2Node* n) { copy(n, this); }
  static void copy(Bci2Node* dst, Bci2Node* src);
};

class Bci2NodeArray: public ResourceObj {
 private:
  Bci2Node *_stack;  
  int       _max_size;  
 public:
  enum Constants {
    InitialSize = 5
  };

  Bci2NodeArray();  
  
  void grow();
  int  max_size() const { return _max_size; }
  Bci2Node* at(int i) const  { 
    assert(0 <= i && i < _max_size, "wrong index"); return &_stack[i];
  }

  // Highlevel interface
  bool equals(int idx, Parse *p, int bci) const; 
  bool jsr_stack_equals(int idx, Parse *p) const;
  void copy(int dst, int src) { Bci2Node::copy(at(dst), at(src)); }  
  int  copy_to_parser(int idx, Parse *p);
  void copy_from_parser(int idx, Parse* p, int bci);
};

//------------------------------OSRParseInfo-----------------------------------
// This structure contains the information which we need when we perform a
// parse for on stack replacement.  Specifically this structure is used to
// store and retrieve the starting state of the parser.
class OSRParseInfo : public ResourceObj {
 private:
  int              _bci;         
  int              _stack_depth;
  int              _monitor_depth;
  const TypeTuple *_domain;
  BitMap           _oops;
 public:
  OSRParseInfo(ciMethod* method, int bci) :
    _bci(bci),
    _stack_depth(-1),
    _monitor_depth(-1),
    _domain(NULL),
    _oops(method->oops_at_bci(_bci)) {}

  // Have all fields been set?
  bool is_complete() { return _stack_depth >= 0 && _monitor_depth >= 0 && _domain != NULL; }

  // Accessors
  int              bci()           { return _bci; }
  int              stack_depth()   { return _stack_depth; }
  int              monitor_depth() { return _monitor_depth; }
  const TypeTuple *domain()        { return _domain; }
  BitMap           oops()          { return _oops; }

  void set_bci(int bci)                     { _bci = bci; }
  void set_stack_depth(int stack_depth)     { _stack_depth = stack_depth; }
  void set_monitor_depth(int monitor_depth) { _monitor_depth = monitor_depth; }
  void set_domain(const TypeTuple *domain)  { _domain = domain; }
};

//------------------------------Parse------------------------------------------
// Parse bytecodes, build a Graph
class Parse : public Phase {
  int _depth;                   // Inline tree depth, for debug printouts
  Node *_exit_return;           // All normal returns branch to here
  Node *_exit_rethrow;          // All exceptional returns branch to here
  InlineTree *_ilt;             // Inline tree at this level
  ciMethod* _method;           // The method here
  ciKlass*  _klass;            // Class of method being compiled
  ciEnv*    _env;              // Compilation environment
  int16  *_preds;               // Array of CFG predecessor counts
  uint8  *_bits;                // Array of bit-flags per bytecode
  uint8  *_args;                // Array of arguments used per bytecode
  bool _has_ex_handler;         // Possibly covered by ex-handler, recursively
  
#ifndef PRODUCT
  // Record if uncommon trap bailout inserted, No_deopt_in_method indicates no bailouts
  int  _deopt_cause;
  void record_deopt(int deopt_cause);

  // Parser remembers generating a deopt that can cause recompilation
  enum DeoptCauseInit {
    No_deopt_in_method = 0x80000000  // No recompilation deopt generated
  };
#endif

  enum Constants {        
    InitialMonitorList  = 4,
    DefaultJsrStackSize = 4
  };

public:
  // Bailout types
  enum BailoutConstants { 
    Bailout_too_many_args      = -4, // Too many arguments to allocate
    Bailout_throwaway_parse    = -3, // Used during on stack replacement to terminate parse
    Bailout_node_limit_exceeded= -2, // Exceeded node construction limit during parsing
    Bailout_shared_handler     = -1, // Exception handler shared in/out of jsr...ret
    Bailout_none               =  1,  // No bailout
    Bailout_exception_thrown   =  2, // Exception thrown (somewhere)
    Bailout_unhandled_bytecode =  3, // Unhandled bytecode 
    Bailout_monitor_matching   =  4, // Wrong monitor nesting
    Bailout_compile_only       =  5  // signal return values from find_block_starts
  };
  BailoutConstants _bailout;         // Record Bailout info during parse
  bool parse_succeeded()     { return _bailout == Bailout_none; }
  void show_bailout_info()   PRODUCT_RETURN;    
private:

  // Monitor Stuff
  const FastLockNode*  _synch_lock;   // FastLockNode for synchronized method

  BailoutConstants check_monitor_matching(); // Returns a Bailout_xxx constant

  // implementation of monitorenter/monitorexit
  void do_monitor_enter( int bci );
  void do_monitor_exit ( int bci );
  // Helper functions for fast/slow path codes
  IfNode *make_check_type( Node *control, Node *fieldmem, Node *cast, Node *slow_arg, float prob, float cnt );
  Node *opt_iff( Node *region, Node *iff );
  Node *make_slow_call( const TypeFunc *slow_call_type, address slow_call, const char *leaf_name, int bci, Node *slow_path, Node *parm0, Node *parm1 );
  void make_merge_mem( Node *region, Node *call, Node *fast_mem );
  void make_slow_call_ex( Node *call, int bci, ciInstanceKlass* ex_klass );

  // Helper functions to build FastLockNodes/FastUnlockNodes
  int next_monitor();
  void insert_mem_bar();
public:
  FastLockNode *shared_lock( Node *obj, uint bci );
  void shared_unlock( Node *box, Node *obj, uint bci );
private:

  // Helper Function for checking clonable blocks
  int not_obviously_loopback(ciByteCodeStream &iter, uint new_bci);

  // Some optimizations while parsing
  PhaseGVN &_gvn;

  // Merge the current mapping into the basic block starting at bci
  void merge( uint bci, uint cloning );
  Node *merge_value( Node *map, uint idx, Node *region );

  // Parse the basic block starting at bci
  void do_one_block( ciByteCodeStream &iter, uint bci, uint cloning );

  // Mappings from bci to partially merged basic blocks.  When a block is
  // first reached it goes on the reached list.  After all CFG paths in to
  // it have been visited it moves to the ready list.  After a block is
  // both ready and parsed it is removed from all lists.  
  Bci2NodeArray* _bci2node;
  int _ready, _reached, _looped;  

  // This should really be a Bci2Node. Intead of individual fields.  
  GrowableArray<int>*  _jsr_stack;    // Keeps track of jsr return address
  SafePointNode*       _map;          // Parser map from JVM to Nodes
  int                  _sp;           // JVM Expression Stack Pointer

  // Access unaliased memory
  Node *memory( uint alias_idx ) {
    Node *mem = _map->in(TypeFunc::Memory);
    assert( mem->Opcode() == Op_MergeMem, "" );
    assert( mem->in(alias_idx)->Opcode() != Op_MergeMem, "" );
    return mem->in(alias_idx);
  }
  Node *memory( const TypePtr *tp ) { return memory(C->get_alias_index(tp)); }
  Node *memory( Node *adr ) { return memory(_gvn.type(adr)->is_ptr()); }
  // Set unaliased memory
  void set_memory( Node *c, uint alias_idx ) {
    Node *mem = _map->in(TypeFunc::Memory);
    assert( mem->Opcode() == Op_MergeMem, "" );
    assert( c->Opcode() != Op_MergeMem, "" );
    mem->set_req( alias_idx, c );
  }
  void set_memory( Node *c, const TypePtr *tp ) { set_memory(c,C->get_alias_index(tp)); }
  void set_memory( Node *c, Node *adr ) { set_memory(c,_gvn.type(adr)->is_ptr()); }
  // Get the entire memory state (probably a MergeMemNode)
  Node *all_memory() const;
  // Set the entire memory state; produce a new MergeMemNode.
  void set_all_memory( Node *newmem );
  // Create a memory projection from the call, then set_all_memory.
  void set_all_memory_call( Node *call );
#ifndef PRODUCT
  void dump_map_adr_mem() const;
#endif

  // Parser mapping from JVM indices into Nodes.
  // Low slots are accessed by the StartNode::enum.
  // Then come the locals at StartNode::Parms to StartNode::Parms+max_locals();
  // Finally come JVM stack slots.
      
  Node *control  () const { return _map->control(); }
  Node *i_o      () const { return _map->i_o(); }
  Node *returnadr() const { return _map->returnadr(); }
  Node *frameptr () const { return _map->frameptr(); }
  Node *local( uint idx ) const { return _map->local(_map->_jvms, idx); }
  Node *stack( uint idx ) const { return _map->stack(_map->_jvms, idx); }
  Node *monitor_box( uint idx ) const { return _map->monitor_box(_map->_jvms, idx); }
  Node *monitor_obj( uint idx ) const { return _map->monitor_obj(_map->_jvms, idx); }

public:
  void set_control  ( Node *c ) { _map->set_control(c); }
  void set_i_o      ( Node *c ) { _map->set_i_o(c); }
  void set_local( uint idx, Node *c ) { _map->set_local(_map->_jvms,idx,c); }
  void set_stack( uint idx, Node *c ) { _map->set_stack(_map->_jvms,idx,c); }
  void push( Node *c ) { _map->set_stack(_map->_jvms,_sp++,c); }
  Node *pop( )         { return _map->stack(_map->_jvms,--_sp); }
  Node *peek(int off=0){ return _map->stack(_map->_jvms, _sp - off - 1); }

  ciEnv*    env   () const { return _env; }
  ciMethod* method() { return _method; }
  ciKlass*  klass () { return _klass; }

private:

  // Parse this bytecode, and alter the Parsers JVM->Node mapping
  void do_one_bytecode( Bytecodes::Code bc, ciByteCodeStream &iter, uint cloning );

  // Add a fixed offset to a pointer
  Node *basic_plus_adr( Node *base, Node *ptr, int offset );
  // helper function to generate array store check
  void array_store_check(int bci);
  // Helper function to compute array addressing
  Node *array_addressing( uint stride, BasicType type, int bci, int vals );
  // Helpers to compute array addressing of two word types T_DOUBLE and T_LONG
  Node *array_addressing_hi_word( Node *ptr );
  // Helper function to do a NULL pointer check or ZERO check based on type.
  Node *do_null_check( Node *value, int bci, BasicType type );
  // Cast obj to not-null on this path
  Node *cast_not_null( Node *obj );
  // Cast adr up to type tp.  Happens on pointers at getfield/putfield
  Node *cast_up( Node *adr, const TypePtr *tp );

  // Pass current map to _exit_return
  void return_current( Node *parm0, int bci );

  // Clone the existing parser map
  SafePointNode *clone_map( int bci);
  // Insert a compiler safepoint into the graph
  void add_safepoint(ciByteCodeStream &iter);

  // Helper function for athrow
  void do_athrow(int klass_index, int bci);

  // Bail out to the interpreter right now
  void uncommon_trap(int klass_index, int bci);

  // Insert a write-barrier'd store.  This is to let generational GC
  // work; we have to flag all oop-stores before the next GC point.
  void store_barrier( Node* adr, Node* val );

  //---------------- JVMPI support --------------------
  void make_jvmpi_method_entry();
  void make_jvmpi_method_exit(int bci);

  //----------------Inlining methods-------------------

  bool inline_library_methods(ciMethod* method, int nargs,
			      const TypeFunc *tf,
			      Bytecodes::Code bc, int bci);

  void generate_guard(Node *condition, RegionNode *slow_region);

  // Helper functions to inline natives
  bool inline_math_natives(ciMethod* method, int nargs);

  bool inline_string_compareTo(ciMethod* method, const TypeFunc* tf, int bci);

  // Helper function for inlining native object hash method
  bool inline_native_hashcode(ciMethod* method, const TypeFunc *tf,
                              Bytecodes::Code bc, int bci);

  // Helper function for inlining native newInstance method
  bool inline_native_newinstance(ciMethod* method, const TypeFunc *tf,
                                 Bytecodes::Code bc, int bci);

  // Helper function for inlining arraycopy
  bool inline_arraycopy(ciMethod* method, const TypeFunc *tf, int bci);

  // Helper function to setup Ideal Call nodes
  void do_call( ciByteCodeStream* stream, Bytecodes::Code bc, uint cloning );

  // rounding for strict float precision conformance
  Node *precision_rounding(Node *n);

  // rounding for strict double precision conformance
  Node *dprecision_rounding(Node *n);

  // rounding for non-strict double stores
  Node *dstore_rounding(Node *n);

  // helper functions for the fast path/slow path idioms
  Node *fast_and_slow( Node *in, int bci, const Type *result_type, Node *null_result, IfNode *fast_test, Node *fast_result, address slow_call, const TypeFunc *slow_call_type, Node *slow_arg, klassOop ex_klass, Node *slow_result );
  void  do_checkcast(ciByteCodeStream *stream);
  void  do_instanceof(ciByteCodeStream *stream);
  Node *null_check_oop( RegionNode *region, Node *oop, Node *null, bool never_see_null, int bci );

  // Helper Functions for accessing _bits array
  BasicType bits_type(int idx)  { return (BasicType)(_bits[idx]&0x0F); }
  int       bits_size(int idx)  { return (_bits[idx]&0x30)>>4; }
  bool      bits_merge(int idx) { return (bool)(_bits[idx]&0x40); }
  bool      bits_clone(int idx) { return (bool)(_bits[idx]&0x80); }

  void bits_size_set(int idx,int sz) {_bits[idx]=(_bits[idx] & ~0x30)|(sz<<4);}
  void bits_type_set( int idx, int ty )  { _bits[idx] = (_bits[idx]&~0x0F)|ty;}
  void bits_merge_set( int idx ) { _bits[idx] |=  0x40; }
  void bits_merge_clr( int idx ) { _bits[idx] &= ~0x40; }
  void bits_clone_set( int idx ) { _bits[idx] |=  0x80; }
  void bits_clone_clr( int idx ) { _bits[idx] &= ~0x80; }

  // Helper functions for shifting & arithmetic
  void modf();
  void modd();

  // helper method for unloaded and uninitialized class uncommon_trap
  bool do_unloaded                 (ciByteCodeStream* stream,
				    ciKlass* klass);
  bool do_unloaded_or_uninitialized(ciByteCodeStream* stream,
				    ciKlass* klass);

  // implementation of _get* and _put* bytecodes
  void do_get_xxx(ciByteCodeStream* stream, bool is_field);
  void do_putstatic(ciByteCodeStream* stream);
  void do_putfield (ciByteCodeStream* stream);

  // implementation of object creation bytecodes
  IfNode *test_and_bump_heap( Node* size );
  IfNode *bump_tle( Node* size, Node** newoop);
  void init_object( Node *klass_node, Node *newoop, bool use_speclock );
  Node * fast_path_allocate( RegionNode *region, Node *alloc_size,
                                    Node *klass_node, Node **slow_ctrl_p);
  void do_new      (ciByteCodeStream* stream);
  void do_newarray (BasicType elemtype, int bci);
  void do_anewarray(ciByteCodeStream* stream);
  void do_Xnewarray(Node* count_val, BasicType elem_type, const Type* etype,
                    const TypeKlassPtr* array_klass, int bci);
  void do_multianewarray(ciByteCodeStream* stream);

  // implementation of jsr/ret
  void do_jsr     (ciByteCodeStream &iter);
  void do_ret     (ciByteCodeStream &iter);

  IfNode* jump_if_fork_int(Node* a, Node* b, BoolTest::mask mask);
  Node*   jump_if_join(Node* iffalse, Node* iftrue);
  void    jump_if_true_fork(IfNode *ifNode, int dest_bci_if_true, uint cloning);
  void    jump_if_false_fork(IfNode *ifNode, int dest_bci_if_false, uint cloning);
  void    jump_if_always_fork(int dest_bci_if_true, uint cloning);

  friend class SwitchRange;
#ifndef PRODUCT
  int _max_switch_depth;
  int _est_switch_depth;
#endif
  void    jump_switch_ranges(int bci, Node* a, SwitchRange *lo, SwitchRange *hi, uint cloning, int depth = 0);

  void    set_predefined_output_for_runtime_call(Node* call);
  void    set_predefined_input_for_runtime_call(SafePointNode* call);

  // helper functions for statistics
  void increment_counter(address counter_addr);   // increment a debug counter

  JVMStateObj *_jhead;            // Pointer to head of inlining list
  JVMStateObj *_jtail;            // Pointer to tail of inlining list

  // Jsr stack manipulation
  bool  is_jsr_stack_empty() const { return _jsr_stack == NULL; }
  int   jsr_stack_depth() const    { return (_jsr_stack  != NULL) ? _jsr_stack->length() : 0; }
  int   jsr_return_at(int i) const { assert(_jsr_stack != NULL, "empty list"); return  _jsr_stack->at(i); }
  int   pop_jsr_return()           { assert(_jsr_stack != NULL, "empty list"); 
                                     int k = _jsr_stack->pop(); 
                                     if (_jsr_stack->length() == 0) _jsr_stack = NULL; 
                                     return k; }
  void  push_jsr_return(int bci)   { if (_jsr_stack == NULL) _jsr_stack = new GrowableArray<int>(DefaultJsrStackSize);
                                     _jsr_stack->push(bci); }  

  //-----------------------------------------------------------------------
  // On stack replacement routines and members
private:
  // Information needed to start a parse at the on stack replacement entry.
  OSRParseInfo *_osr_info;
  bool          _is_first_osr_parse;

public:
  void          collect_osr_info(int cloning);

  // When we are compiling an on stack replacement method, we actually parse
  // the method twice -- the first time we derive our _osr_info and bail out,
  // the second time we use the _osr_info we have collected in the first parse
  // to compile a continuation from the osr entry bci.  The routines below can
  // be used to figure out if this is a normal parse, first pass osr parse, or
  // second pass osr parse.
  bool is_normal_parse()    { return _osr_info == NULL; }
  bool is_first_osr_parse() { return _osr_info != NULL && _is_first_osr_parse;}
  bool is_second_osr_parse(){ return _osr_info != NULL &&!_is_first_osr_parse;}

  Node *fetch_interpreter_state(int index, const Type *type, Node *parameter);
  Node *fetch_monitor(int index, Node *box, Node *monitor_address);

private:

  // friends
  friend class PreserveJVMState;
  friend class Bci2NodeArray;

public:
  // Root invocations of parser
  Parse( ciEnv* env, InlineTree* ilt,
	 PhaseGVN &gvn, OSRParseInfo *osr_info,
	 Unique_Node_List &for_igvn );
  // Recursive parse invocation from do_call()
  Parse( ciEnv* env, InlineTree* ilt,
	 JVMStateObj *jhead, JVMStateObj *jtail,
	 int sync, int depth, PhaseGVN &gvn,
	 Unique_Node_List &for_igvn, Node *map, uint sp_off, bool has_ex_handler );
  // Parse native methods
  Parse( SafePointNode *native_map, PhaseGVN &gvn,
	 Unique_Node_List &for_igvn );

  void Init( InlineTree *ilt, JVMStateObj *jhead, JVMStateObj *jtail, int sync, int depth, Node *map, uint sp_off, bool has_ex_handler );
  StartNode *build_start_node();
  JVMStateObj *initialize_map( StartNode *start, Node *map, uint sp_off );
  void initialize_osr_map(StartNode *start);

  // From what bci will we begin the parse?
  int  entry_bci()  { return (is_second_osr_parse()) ? _osr_info->bci() : 0 ; }

  RootNode *_root;                // Where to hang everything
  Node *_top;                     // Undefined state
  Node *_null;                    // Handy constant for the NULL pointer
  SafePointNode *map() const { return _map; }
  int _sync_base;                 // The monitor depth of our calling parse
  int _sync_max;                  // The maximum monitor depth we have seen,
                                  // including recursive parses.
  const TypeFunc *_tf;            // My kind of function type
  Unique_Node_List &_for_igvn;    // Initial node list for Iterative GVN

  // helper functions used for stub generation
  // create an integer constant
  static Node *intcon( int con );
  // Add a fixed offset to a pointer
  Node *plus_adr( Node *ptr, int offset ) const;
  // Create a constant node
  Node *makecon( const Type *t ) { return _gvn.makecon(t); }
public:

  // Helper Function for adding debug information
  void kill_dead_locals( int bci );
  void add_safepoint_edges(SafePointNode *call, int bci);
  void add_debug_info( SafePointNode *call, JVMStateObj *jhead );
  ciMethod* method() const { return _method; } // the method here

  void rethrow(Node *, bool is_null_exception); // merge current map into correct rethrow node
  Node *catch_exceptions(ciExceptionHandlerStream*, int bci, Node *call);  // Add Catch/CatchProjs
  // Handle all exceptions thrown by the inlined method.  
  void handlers_for_rethrow( Node *exit_rethrow, int throw_bci );

  IfNode *create_and_map_if(Node *ctrl, Node *tst, float prob, float cnt) {
    IfNode *iff = new (2) IfNode( ctrl, tst, prob, cnt );// New IfNode's
    _gvn._types.map(iff->_idx, iff->Value(&_gvn)); // Value may be known at parse-time
    // Place 'if' on worklist if it will be in graph
    if( tst->is_Con() == 0 ) { _for_igvn.push(iff); } // Range-check and Null-check removal is later
    return iff;
  };

  IfNode *create_and_xform_if(Node *ctrl, Node *tst, float prob, float cnt) {
    IfNode *iff = new (2) IfNode( ctrl, tst, prob, cnt );// New IfNode's
    _gvn.transform(iff);                           // Value may be known at parse-time
    // Place 'if' on worklist if it will be in graph
    if( tst->is_Con() == 0 ) { _for_igvn.push(iff); } // Range-check and Null-check removal is later
    return iff;
  };


#ifndef PRODUCT
  static void print_statistics(); // Print some performance counters
  void dump();
  void dump_bci(int bci);
#endif
};


// Helper class to support building of control flow branches. Upon
// creation the map and sp at bci are cloned and restored upon de-
// struction. Typical use:
//
// { PreserveJVMState pjvms(this, bci);
//   // code of new branch
// }
// // here the JVM state at bci is established

class PreserveJVMState: public StackObj {
 private:
  Parse*         _parser;
  SafePointNode* _map;
  uint           _sp;

 public:
  PreserveJVMState(Parse* parser, int bci)       { _parser = parser; _map = parser->clone_map(bci); _sp = parser->_sp; }
  ~PreserveJVMState()                            { _parser->_map = _map; _parser->_sp = _sp; }
};

extern FastLockNode *native_lock( PhaseGVN &gvn, Node *obj, SafePointNode *map, Node *top, Node *null );
extern void native_unlock( PhaseGVN &gvn, FastLockNode *lock, Node *map, Node *top, Node *null );

//------------------------------InlineTree-------------------------------------
// Bytecode info for the entire inline tree.  Gathers basic block starts,
// argument counts and BasicTypes for each bytecode in the whole inline tree.
class InlineTree : public ResourceObj {
  void bits_size_set(int idx,int sz) {_bits[idx]=(_bits[idx] & ~0x30)|(sz<<4);}
  void bits_type_set( int idx, int ty )  { _bits[idx] = (_bits[idx]&~0x0F)|ty;}
  void bits_merge_set( int idx ) { _bits[idx] |=  0x40; }
  void bits_merge_clr( int idx ) { _bits[idx] &= ~0x40; }
  void bits_clone_set( int idx ) { _bits[idx] |=  0x80; }
  void bits_clone_clr( int idx ) { _bits[idx] &= ~0x80; }

  void set_type_and_size(int bci, Bytecodes::Code bc, BasicType type);  

public:
  ciEnv*    _env;
  ciScope*  _scope;                // back-pointer to original inline decision
  ciMethod* _method;      // The method here
  const int _caller_bci;        // BCI of caller; -1 for root
  // Call-site count / interpreter invocation count, scaled recursively.
  // Always between 0.0 and 1.0.  Represents the percentage of the method's
  // total execution time used at this call site.
  const float _site_invoke_ratio; 
  GrowableArray<InlineTree*> _subtrees;
  int16 *_preds;                // Array of CFG predecessor counts per bytecode
  uint8 *_bits;                 // Array of bit-flags per bytecode
  uint8 *_args;                 // Array of arguments used per bytecode
  uint _cnt_inlines;            // Count of inlined methods
  uint _cnt_inline_bcs;         // Count of inlined bytecodes

  // Build an inline tree for this method
  InlineTree( ciEnv* env, ciMethod* m, int caller_bci, float site_invoke_ratio );

  // Parse bytecodes and fill in fields
  Parse::BailoutConstants find_block_starts(uint entry_bci, TRAPS);

  // Build an InlineTree given an inline decision tree
  void inline_policy( ciScope* scope );

  // Return the subtree at this bci or NULL
  InlineTree *subtree_at_bci( uint bci ) const;

  // Copy fields for an OSR parse; no inlining.
  InlineTree *osr_copy() const;

  // Deep copy fields for cloning whilst inlining
  InlineTree *deep_copy() const;
};

