/*
 * @(#)phaseX.hpp	1.87 01/02/09 14:47:00
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

class Compile;
class ConINode;
class Node;
class Type;
class PhaseTransform;
class   PhaseGVN;
class     PhaseIterGVN;
class       PhaseCCP;
class   PhasePeephole;
class   PhaseDCE;
class   PhaseChaitin;


//-----------------------------------------------------------------------------
// Expandable closed hash-table of nodes, initialized to NULL.
// Note that the constructor just zeros things
// Storage is reclaimed when the Arena's lifetime is over.
class NodeHash : public StackObj {
protected:
  Arena *_a;                    // Arena to allocate in
  uint   _max;                  // Size of table (power of 2)
  uint   _inserts;              // For grow and debug, count of hash_inserts
  uint   _insert_limit;         // 'grow' when _inserts reaches _insert_limit
  Node **_table;                // Hash table of Node pointers
  Node  *_sentinel;             // Replaces deleted entries in hash table

public:
  NodeHash(uint est_max_size);
  NodeHash(Arena *arena, uint est_max_size);
  NodeHash(NodeHash *use_this_state);
  Node  *hash_find(const Node*);// Find an equivalent version in hash table
  Node  *hash_find_insert(Node*);// If not in table insert else return found node
  void   hash_insert(Node*);    // Insert into hash table
  bool   hash_delete(const Node*);// Replace with _sentinel in hash table
  void   check_grow() {
    _inserts++;
    if( _inserts == _insert_limit ) { grow(); }
    assert( _inserts <= _insert_limit, "hash table overflow");
    assert( _inserts < _max, "hash table overflow" );
  }
  static uint round_up(uint);   // Round up to nearest power of 2
  void   grow();                // Grow _table to next power of 2 and rehash
  // Return 75% of _max, rounded up.
  uint   insert_limit() const { return _max - (_max>>2); }

  void   clear();               // Set all entries to NULL, keep storage.
  // Size of hash table
  uint   size()         const { return _max; }
  // Return Node* at index in table
  Node  *at(uint table_index) {
    assert(table_index < _max, "Must be within table");
    return _table[table_index];
  }

  Node  *sentinel() { return _sentinel; }

#ifndef PRODUCT  
  Node  *find_index(uint idx);  // For debugging
  void   dump();                // For debugging, dump statistics
#endif
  uint   _grows;                // For debugging, count of table grow()s
  uint   _look_probes;          // For debugging, count of hash probes
  uint   _lookup_hits;          // For debugging, count of hash_finds
  uint   _lookup_misses;        // For debugging, count of hash_finds
  uint   _insert_probes;        // For debugging, count of hash probes
  uint   _delete_probes;        // For debugging, count of hash probes for deletes
  uint   _delete_hits;          // For debugging, count of hash probes for deletes
  uint   _delete_misses;        // For debugging, count of hash probes for deletes
  uint   _total_inserts;        // For debugging, total inserts into hash table
  uint   _total_insert_probes;  // For debugging, total probes while inserting
};


//------------------------------PhaseDefUse------------------------------------
class PhaseDefUse : public Phase {
  Arena  *_a;                   // Arena to allocate in
  uint    _unique;              // Max number of nodes
  Node  **_edges;               // Array of def-use edges
  uint   *_start;               // starting point in _edges: at init, running sum of user counts
  // _cnt and _max may exceed a uint16: e.g. ConP   #TopPTR    !!!!!
  uint   *_cnt;                 // _cnt[i] = Number of uses of node with index 'i'
  uint   *_max;                 // _max[i] = Maximum number of uses storable at _edges[start[i]]
  uint    _max_edges;           // Max space in _edges
  uint    _max_cnts;            // Max space in _start, _cnt, and _max
  uint    _cur_edges;           // Currently used space in _edges
  uint    _cur_cnts;            // Currently used space in _start, _cnt and _max, max(idx) + 1

  ReallocMark _nesting;         // assertion check for reallocations

  enum {
    MIN_USE_GROWTH   =   4,     // Minimum space to add for uses
    MIN_NODES_GROWTH =  12,     // Minimum space to add for new nodes
    MIN_EDGES_GROWTH = 120      // Minimum space to add for new edges;
  };
  // Construct DefUse info
  void compute_defuse_edges( uint root_set_size, Node **root_set );
  // Internal accessor functions
  uint    start(uint node_idx) const { return _start[node_idx]; }
  Node  **out(uint node_idx) const { return &_edges[_start[node_idx]]; }

public:
  PhaseDefUse( uint unique, Node *root );
  PhaseDefUse( uint unique, Node_List &root_set );
  PhaseDefUse( Arena *a, uint unique, Node *root );
  PhaseDefUse( PhaseDefUse *original );

  // Per-Node def-use edge count
  uint    cnt(uint node_idx) const { return _cnt[node_idx]; }
  uint cnt( const Node *n ) const { 
    assert( n->_idx <= _cur_cnts, "" );
    return _cnt[n->_idx]; }

  // Per-Node def-use edge array start
  Node **out( const Node *n ) const { return &_edges[_start[n->_idx]]; }

  // Current number of unique nodes, in case splitting has occurred
  uint unique() const { return _cur_cnts + 1; }

  // Return 'true' if a def-use edge exists between 'def' and 'use'
  bool def_use_exists( const Node *def, const Node *use );

  // The following methods modify def-use info
  void insert_use( Node *def, Node *use );
  void delete_use( Node *def, Node *use );
  bool delete_use_attempt( Node *def, Node *use );
  void delete_all_uses( Node *def );

  // Replace use-pointers to 'old' with use-pointers to 'nn'
  // does not update def-use info for 'nn'
  void replace_node( Node *old, Node *nn );

  // Append def-use pointers from 'load' with a pointer to 'store'
  // When 'load' may be anti-dependent on 'store'
  // The 'store' will have a precedence edge to 'load'
  void add_anti_dependence( Node *load, Node *store );

  // Grow the _cnt and _start arrays
  void grow_nodes ( uint new_node_idx );

protected:
  // Look for edge from 'def' to 'use' and replace with edge from 'def' to 'clone'
  // If necessary, insert a new edge from 'def' to 'clone'
  void update_an_edge( Node *def, Node *old, Node *clone );
  // Grow the _edges array
  void grow_edges ( uint num_additional_edges );
  // Shift entries in _edges, to insert new def-use edge; update _start
  // fill with edges from 'def' to 'clone'
  void shift_edges( Node *def, Node *clone, uint number );

#ifndef PRODUCT
public:  // Dump def-use edges per node
  void dump();
  void dump(int depth);
  void dump_node(uint index);
  void verify_uses(const Node *def);
#endif
};


//------------------------------PhaseTransform---------------------------------
// Phases that analyze, then transform.  Constructing the Phase object does any
// global or slow analysis.  The results are cached later for a fast 
// transformation pass.  When the Phase object is deleted the cached analysis
// results are deleted.
class PhaseTransform : public Phase {
protected:
  Arena     *_arena;
public:
  PhaseTransform( PhaseNumber pnum );
  PhaseTransform( Arena *arena, PhaseNumber pnum );
  PhaseTransform( PhaseTransform *phase, PhaseNumber pnum );

  Arena     *arena() { return _arena; }

  Node_Array _nodes;            // Map old node indices to new nodes.

  Type_Array _types;            // Map old node indices to Types.
  void        map_type(uint idx, const Type *t) { _types.map( idx, t ); }
  const Type *type( const Node *n ) const {return _types.fast_lookup(n->_idx);}
  
  // Return a node which computes the same function as this node, but
  // in a faster or cheaper fashion.  
  virtual Node *transform( Node *n ) = 0;

  // Return whether two Nodes are equivalent.  Must not be recursive,
  // since the recursive version is built from this non-recursive
  // version.  For pessimistic optimizations, this degrades to pointer
  // equivalence.
  virtual int eqv( const Node *n1, const Node *n2 ) const;

  // For pessimistic passes, the return type must monotonically narrow.
  // For optimistic  passes, the return type must monotonically widen.
  // In any case, only widen or narrow a few times before going to the
  // correct flavor of top or bottom.
  virtual const Type *widen(const Type *new_type,const Type *old_type) const
  { ShouldNotCallThis(); return NULL; }

#ifndef PRODUCT
  void dump_old2new_map() const;
  void dump_new( uint new_lidx ) const;
  void dump_types() const;
  void dump_nodes_and_types(const Node *root, uint depth, bool only_ctrl = true);
  void dump_nodes_and_types_recur( const Node *n, uint depth, bool only_ctrl, VectorSet &visited);

  uint   _count_progress;       // For profiling, count transforms that make progress
  void   set_progress()        { ++_count_progress; assert( allow_progress(),"No progress allowed during verification") }
  void   clear_progress()      { _count_progress = 0; }
  uint   made_progress() const { return _count_progress; }

  uint   _count_transforms;     // For profiling, count transforms performed
  void   set_transforms()      { ++_count_transforms; }
  void   clear_transforms()    { _count_transforms = 0; }
  uint   made_transforms() const{ return _count_transforms; }

  bool   _allow_progress;      // progress not allowed during verification pass
  void   set_allow_progress(bool allow) { _allow_progress = allow; }
  bool   allow_progress()               { return _allow_progress; }
#endif
};

//------------------------------PhaseValues------------------------------------
// Phase infrastructure to support values
class PhaseValues : public PhaseTransform {
protected:
  NodeHash  _table;             // Hash table for value-numbering

  ConINode *_icons[37];         // Integer constant Nodes from -4 to 32

public:
  PhaseValues( Arena *arena, uint est_max_size );
  PhaseValues( PhaseValues *pt );
  ~PhaseValues();
  virtual PhaseIterGVN *is_IterGVN() { return 0; }

  virtual int eqv( const Node *n1, const Node *n2 ) const;

  // Some Ideal and other transforms delete --> modify --> insert values
  bool   hash_delete(Node *n)     { return _table.hash_delete(n); }
  void   hash_insert(Node *n)     { _table.hash_insert(n); }
  Node  *hash_find_insert(Node *n){ return _table.hash_find_insert(n); }
  Node  *hash_find(const Node *n) { return _table.hash_find(n); }

  // Make an idealized constant, i.e., one of ConINode, ConPNode, ConFNode, etc
  virtual ConNode *makecon( const Type *t );

  // Fast integer constant. Same as "transform(new ConINode(TypeInt::make(i)))"
  virtual ConINode *intcon( int i );

  virtual const Type *widen( const Type *new_type, const Type *old_type ) const
  { return new_type; };

#ifndef PRODUCT
  uint   _count_new_values;     // For profiling, count new values produced
  void    inc_new_values()        { ++_count_new_values; }
  void    clear_new_values()      { _count_new_values = 0; }
  uint    made_new_values() const { return _count_new_values; }
#endif
};


//------------------------------PhaseGVN---------------------------------------
// Phase for performing local, pessimistic GVN-style optimizations.
class PhaseGVN : public PhaseValues {
public:
  PhaseGVN( Arena *arena, uint est_max_size ) : PhaseValues( arena, est_max_size ) {}
  PhaseGVN( PhaseGVN *gvn ) : PhaseValues( gvn ) {}

  // Return a node which computes the same function as this node, but
  // in a faster or cheaper fashion.  
  Node  *transform( Node *n );
  Node  *transform_no_reclaim( Node *n );
};

//------------------------------PhaseIterGVN-----------------------------------
// Phase for iteratively performing local, pessimistic GVN-style optimizations.
// and ideal transformations on the graph.
class PhaseIterGVN : public PhaseGVN {
  // Idealize old Node 'n' with respect to its inputs and its value
  virtual Node *transform_old( Node *a_node );
protected:

  // Idealize new Node 'n' with respect to its inputs and its value
  virtual Node *transform( Node *a_node );

  // Warm up hash table, type table and initial worklist
  void init_worklist( Node *a_root );

  virtual const Type *widen( const Type *new_type, const Type *old_type ) const
  { return new_type; };

public:
  PhaseIterGVN( PhaseIterGVN *igvn ); // Used by CCP constructor
  PhaseIterGVN( PhaseGVN *gvn ); // Used after Parser

  virtual PhaseIterGVN *is_IterGVN() { return this; } 

  Unique_Node_List _worklist;       // Iterative worklist 
  PhaseDefUse      _du;             // Def-Use info for forward propagation

  // Given def-use info and an initial worklist, apply Node::Ideal, 
  // Node::Value, Node::Identity, hash-based value numbering, Node::Ideal_DU
  // and dominator info to a fixed point.
  void optimize();

  // Make an idealized constant, i.e., one of ConINode, ConPNode, ConFNode, etc
  ConNode *makecon( const Type *t );

  // Fast integer constant. Same as "transform(new ConINode(TypeInt::make(i)))"
  ConINode *intcon( int i );

  // Register a new node with the iter GVN pass without transforming it.
  // Used when we need to restructure a Region/Phi area and all the Regions
  // and Phis need to complete this one big transform before any other
  // transforms can be triggered on the region.
  Node *register_new_node_with_optimizer( Node *n );

  // Kill a globally dead Node.   It is allowed to have uses which are
  // assumed dead and left 'in limbo'.
  void remove_globally_dead_node( Node *dead );

  // Kill all inputs to a dead node, recursively making more dead nodes.
  // The Node must be dead locally, i.e., have no uses.
  void remove_dead_node( Node *dead ) {
    assert( _du.cnt(dead) == 0, "node must be dead" );
    remove_globally_dead_node(dead);
  }

  // Subsume users of node 'old' into node 'nn'
  // If no Def-Use info existed for 'nn' it will after call.
  void subsume_node( Node *old, Node *nn );

  // Add users of 'n' to worklist
  void add_users_to_worklist0( Node *n );
  void add_users_to_worklist ( Node *n );

#ifndef PRODUCT
protected:
  // Sub-quadratic implementation of VerifyIterativeGVN.
  unsigned long _verify_counter;
  unsigned long _verify_full_passes;
  enum { _verify_window_size = 30 };
  Node* _verify_window[_verify_window_size];
  void verify_step(Node* n);
#endif
};

//------------------------------PhaseCCP---------------------------------------
// Phase for performing global Conditional Constant Propagation.
// Should be replaced with combined CCP & GVN someday.
class PhaseCCP : public PhaseIterGVN {
  // Non-recursive.  Use analysis to transform single Node.
  virtual Node *transform_once( Node *n );

public:
  PhaseCCP( PhaseIterGVN *igvn ); // Compute conditional constants
  ~PhaseCCP();

  // Worklist algorithm identifies constants
  void analyze();
  // Recursive traversal of program.  Used analysis to modify program.
  virtual Node *transform( Node *n );
  // Do any transformation after analysis
  void          do_transform();

  virtual const Type *widen( const Type *new_type, const Type *old_type ) const 
  { return new_type->widen(old_type); };

#ifndef PRODUCT
  static uint _total_invokes;    // For profiling, count invocations
  void    inc_invokes()          { ++PhaseCCP::_total_invokes; }

  static uint _total_constants;  // For profiling, count constants found
  uint   _count_constants;
  void    clear_constants()      { _count_constants = 0; }
  void    inc_constants()        { ++_count_constants; }
  uint    count_constants() const { return _count_constants; }

  static void print_statistics();
#endif
};


//------------------------------PhasePeephole----------------------------------
// Phase for performing peephole optimizations on register allocated basic blocks.
class PhasePeephole : public PhaseTransform {
  PhaseChaitin *_chaitin;
  PhaseCFG     &_cfg;
  // Recursive traversal of program.  Pure function is unused in this phase
  virtual Node *transform( Node *n );

public:
  PhasePeephole( PhaseChaitin *chaitin, PhaseCFG &cfg );
  ~PhasePeephole();

  // Do any transformation after analysis
  void          do_transform();

#ifndef PRODUCT
  static uint _total_peepholes;  // For profiling, count peephole rules applied
  uint   _count_peepholes;
  void    clear_peepholes()      { _count_peepholes = 0; }
  void    inc_peepholes()        { ++_count_peepholes; }
  uint    count_peepholes() const { return _count_peepholes; }

  static void print_statistics();
#endif
};


//------------------------------PhaseDCE---------------------------------------
// Phase for Dead Code Elimination.  Actually, DCE is done "for free" by not
// walking dead code in all other passes.  This pass just compacts the Nodes
// by doing a copying garbage collector.
class PhaseDCE : public PhaseTransform {
  // Non-recursive.  Use analysis to transform single Node.
  virtual Node *transform_once( Node *n );
public:
  PhaseDCE( PhaseNumber pnum );
  // Recursive traversal of program.  Used analysis to modify program.
  virtual Node *transform( Node *n );
  // Do any transformation after analysis
  void          do_transform();
};
