/*
 * @(#)postaloc.cpp	1.38 01/02/09 14:47:01
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
#include "incls/_precompiled.incl"
#include "incls/_postaloc.cpp.incl"

//------------------------------use_prior_register-----------------------------
// Turn this copy into a reg-reg copy.  Extend lifetime of incoming register.
void PhaseChaitin::use_prior_register( Node *n, uint idx, Node *def ) {
  // No effect?
  if( def == n->in(idx) ) return;

  // Not every pair of physical registers are assignment compatible,
  // e.g. on sparc floating point registers are not assignable to integer
  // registers.  
  OptoReg::Name def_reg = lrgs(n2lidx(def)).reg();
  const RegMask &use_mask = n->in_RegMask(idx);
  if( !use_mask.Member(def_reg) )
    return;

  // Capture the old def in case of derived pointer
  Node *old = n->in(idx);

  // Use the new def
  n->set_req(idx,def);
  _post_alloc++;

  // ----
  // Extend lifetime of oops and callee-save registers so 
  // we get correct GC masks.  Means we have to visit all blocks
  // between the updated use site and the last place the new
  // def was live.  

  // Derived pointer?  Need to extend the lifetime of a derived pointer.
  // Make sure its live at all GC sites we extend over.
  const TypeJavaPtr *tj = def->bottom_type()->isa_javaptr();
  const bool derived = (tj && tj->_offset != 0);

  // Blocks of all concerned
  Block *bd = _cfg._bbs[def->_idx];
  Block *bn = _cfg._bbs[n  ->_idx];

  // Live range index of def; we are adding this to the liveout sets
  uint lidx = n2lidx(def);

  // Worklist of blocks to have their liveout sets extended
  Block_List worklist;
  worklist.push( n->is_Phi() ? _cfg._bbs[bn->pred(idx)->_idx] : bn );

  while( worklist.size() ) {
    Block *b = worklist.pop();
    // If we have a derived pointer we need to scan the block for GC 
    // sites and add the derived pointer after the debug info.
    if( derived ) {
      // Last instruction in block to scan for a GC point
      uint last = (b == bn) ? bn->find_node(n) : b->end_idx();
      // First instruction in block to scan for a GC point
      uint first= (b == bd) ? bd->find_node(def) : 1;
      // Scan block for GC points
      for( uint i = first; i < last; i++ ) {
        Node *x = b->_nodes[i];
        JVMStateObj *jvms = x->jvmtail();
        if( jvms ) {            // GC point?
          // Scan derived pointer set for an instance of 'old'.
          // Need base from 'old'.
          uint old_idx = 0;     // Index of 'old' in derived/bast pair list
          uint j;
          for( j = jvms->oopoff(); j < x->req(); j += 2 ) {
            if( x->in(j) == def ) // Find an instance of derived already?
              break;            // Then no need to add it
            if( x->in(j) == old ) // Found an instance of old?
              old_idx = j;      // Capture index
          }
          if( j < x->req() )    // Found an instance of derived already?
            break;              // Then no need to add it!
          assert( old_idx, "must find old" );
          x->add_req(def);      // Add derived
          x->add_req(x->in(old_idx+1)); // Add same base as 'old'
        }
      }
    }

    // If block is same as DEF block, we do not need to
    // push DEF liveout any higher in CFG.
    if( b == bd ) continue;

    // Work back to predecessor block
    uint k;
    for(k = 1; k < b->num_preds(); k++ ) {
      Block *pb = _cfg._bbs[b->pred(k)->_idx];
      IndexSet *liveout = _live->live(pb);
      // Add DEF to live set; if added something then check predecessors
      if( liveout->insert(lidx) ) 
        worklist.push(pb);
    }
  }
}

//------------------------------post_allocate_copy_removal---------------------
// Post-Allocation peephole copy removal.  We do this in 1 pass over the 
// basic blocks.  We maintain a mapping of registers to Nodes (an  array of 
// Nodes indexed by machine register or stack slot number).  NULL means that a 
// register is not mapped to any Node.  We can (want to have!) have several 
// registers map to the same Node.  We walk forward over the instructions
// updating the mapping as we go.  At merge points we force a NULL if we have
// to merge 2 different Nodes into the same register.  Phi functions will give
// us a new Node if there is a proper value merging.  Since the blocks are
// arranged in some RPO, we will visit all parent blocks before visiting any
// successor blocks (except at loops).
//
// If we find a Copy we look to see if the Copy's source register is a stack 
// slot and that value has already been loaded into some machine register; if 
// so we use machine register directly.  This turns a Load into a reg-reg 
// Move.  We also look for reloads of identical constants.
//
// When we see a use from a reg-reg Copy, we will attempt to use the copy's
// source directly and make the copy go dead.
void PhaseChaitin::post_allocate_copy_removal() {
  ResourceMark rm;

  // Need a mapping from basic block Node_Lists.  We need a Node_List to
  // map from register number to value-producing Node.
  Node_List **blk2value = NEW_RESOURCE_ARRAY( Node_List *, _cfg._num_blocks+1);
  memset( blk2value, 0, sizeof(Node_List*)*(_cfg._num_blocks+1) );
  // Need a mapping from basic block Node_Lists.  We need a Node_List to
  // map from register number to register-defining Node.
  Node_List **blk2regnd = NEW_RESOURCE_ARRAY( Node_List *, _cfg._num_blocks+1);
  memset( blk2regnd, 0, sizeof(Node_List*)*(_cfg._num_blocks+1) );

  // Make original Node_Lists for the first block.  They default to mapping
  // all registers to NULL, meaning all registers are undefined.
  int p = _cfg._blocks[0]->_pre_order;
  blk2value[p] = new Node_List();
  blk2regnd[p] = new Node_List();
  blk2value[p]->map(SharedInfo::stack0,NULL);
  blk2regnd[p]->map(SharedInfo::stack0,NULL);

  // We keep unused Node_Lists on a free_list to avoid wasting
  // memory.
  GrowableArray<Node_List*> free_list = GrowableArray<Node_List*>(16);

  // For all blocks
  for( uint i = 0; i < _cfg._num_blocks; i++ ) {
    Block *b = _cfg._blocks[i];
    // Extract Node_List mappings
    Node_List &value = *blk2value[b->_pre_order];
    Node_List &regnd = *blk2regnd[b->_pre_order];
    // If any predecessor has not been visited, we do not know the state
    // of registers at the start.  Zap away.
    uint j;
    for( j = 1; j < b->num_preds(); j++ ) {
      Block *pb = _cfg._bbs[b->pred(j)->_idx];
      if( blk2value[pb->_pre_order] != (Node_List*)-1 ) {
        // Some predecessor has not yet been visited; zap map to empty
        for( uint l = 0; l < (uint)SharedInfo::stack0; l++ ) {
          value.map(l,NULL);
          regnd.map(l,NULL);
        }
        break;
      }
    }

    // For all instructions
    uint last_inst = b->end_idx();
    for( j = 1; j < last_inst; j++ ) {
      Node *n = b->_nodes[j];

      // Remove uses of useless copies
      uint cnt = n->req();
      uint k;
      for( k = 1; k < cnt; k++ ) {
        Node *def = n->in(k);
        uint useidx = n2lidx(def);
        uint idx = def->is_Copy();
        if( !idx ) continue;
        uint srcidx = n2lidx(def->in(idx));
        // Is this a copy to/from the same register?
        // Always true for same live range, but also true if post-alloc
        // has been busy on the 'def' copy.
        while( lrgs(srcidx).reg() == lrgs(useidx).reg() ) {
          // Then skip it; the copy goes dead & will be removed later
          n->set_req(k,def->in(idx));
	  // Now see if there is a chain of these
	  def = def->in(idx);
	  idx = def->is_Copy();
          if( !idx ) break;     // No longer a copy
          srcidx = n2lidx(def->in(idx));
	}
      }

      // Unallocated Nodes define no registers
      uint lidx = n2lidx(n);
      OptoReg::Name reg = lrgs(lidx).reg();

      // If this is a copy, find the value being copied
      Node *val = n;
      uint n_ideal_reg = n->ideal_reg();
      if( lidx && reg < SharedInfo::stack0 ) {
        uint idx = n->is_Copy();
        if( idx && n->is_SpillCopy() ) {
          // Skip through all copies to the _value_ being used.
          // Do not change from int to pointer
          uint x = idx;
          while( x && val->in(x)->ideal_reg() == n_ideal_reg ) {
            val = val->in(x);
            x = val->is_Copy();
          }

          // Look for load-copies of values already in registers
        
          uint val_reg = lrgs(n2lidx(val)).reg();
          // Do not remove spills of callee-save registers
          if( !(_matcher._register_save_policy[val_reg] == 'E' || 
                _matcher._register_save_policy[val_reg] == 'A' )
              ) { 
            // See if it happens to already be in the correct register!
            if( value[reg] == val &&
                // Doubles check both halves
                ((n_ideal_reg != Op_RegD && n_ideal_reg != Op_RegL) || value[reg-1] == val) ) {
              use_prior_register(n,idx,regnd[reg]);
            } else {
              // Scan all registers to see if this value is around already
              for( uint l = 0; l < (uint)SharedInfo::stack0; l++ ) {
                if( value[l] == val &&
                    // Doubles check both halves
		    ((n_ideal_reg != Op_RegD && n_ideal_reg != Op_RegL) || ((l&1) && value[l-1] == val)) ) {
                  use_prior_register(n,idx,regnd[l]);
                  break;
                } // End of if found value already loaded
              } // End of scan all machine registers
            } // End of check for already loading into proper register
          } // End of if n is a load-Copy
        } // End of if n is a Copy
      }

      // Skip any reg-reg copies I can to make things go dead
      JVMStateObj *jvms = n->jvmtail();
      // Do not elide copies of derived pointers because it
      // changes their lifetimes badly.
      uint oopoff = jvms ? jvms->tail()->oopoff() : cnt;
      for( k = 1; k < oopoff; k++ ) {
        Node *def = n->in(k);
        MachNode *mach = n->is_Mach();
        if( mach && mach->two_adr() ) continue;
        uint useidx = n2lidx(def);
        uint idx = def->is_Copy();
        if( !idx ) continue;
        uint srcidx = n2lidx(def->in(idx));
	uint def_ideal_reg = def->ideal_reg();
        // Is this a reg-reg copy and I can use the original value?
        OptoReg::Name usereg = lrgs(useidx).reg();
        OptoReg::Name srcreg = lrgs(srcidx).reg();
        if( usereg < SharedInfo::stack0 &&
            srcreg < SharedInfo::stack0 &&
            regnd[srcreg] == def->in(idx) &&
	    ((def_ideal_reg != Op_RegD && def_ideal_reg != Op_RegL) || ((srcreg&1) && regnd[srcreg-1] == def->in(idx))) ) {
          // Rewire input to avoid the copy, possibly making it go dead
          use_prior_register(n,k,def->in(idx));
        }
      }

      // Update the register defined by this instruction
      if( lidx && reg < SharedInfo::stack0 ) {
	if( n_ideal_reg != Op_RegD && n_ideal_reg != Op_RegL ) {
	  // If Node 'n' does not change the value mapped by the register,
	  // then 'n' is a useless copy.  Do not update the register->node
	  // mapping so 'n' will go dead.
	  if( value[reg] != val ) {
	    // Update the mapping: record new Node defined by the register
	    regnd.map(reg,n);
	    // Update mapping for defined *value*, which is the defined
	    // Node after skipping all copies.
	    value.map(reg,val);
	  }
	} else {
	  // If the value occupies a register pair, record same info
	  // in both registers.
	  assert( reg&1,"doubles in high half of aligned pair" );
	  if( value[reg] != val || value[reg-1] != val ) {
	    regnd.map(reg  ,n);
	    regnd.map(reg-1,n);
	    value.map(reg  ,val);
	    value.map(reg-1,val);
	  }
	}
 
        // Fat projections kill many registers
        if( n_ideal_reg == MachProjNode::fat_proj ) {
          RegMask rm = n->out_RegMask();
          // wow, what an expensive iterator...
          while( (reg = rm.find_first_elem()) != OptoReg::Bad ) {
            rm.Remove(reg);
            value.map(reg,NULL);
            regnd.map(reg,NULL);
          }
        }
      }

    } // End of for all instructions in the block

    // Now push our mappings out to successor blocks
    bool used_once = false;
    for( j = 0; j < b->_num_succs; j++ ) {
      Block *sb = b->_succs[j];
      Node_List *svalue = blk2value[sb->_pre_order];
      Node_List *sregnd = blk2regnd[sb->_pre_order];
      // Already visited successor?  (Loop back edge?)
      // Thing no need to push mapping into him.
      if( svalue == (Node_List*)-1 ) continue;
      // If there is no prior mapping, just use our mappings directly
      if( !svalue ) {
        Node_List *cvalue = &value;
        Node_List *cregnd = &regnd;
        if( !used_once ) {      // First use only, just copy pointers
          used_once = true;
        } else {                // Multiple uses; clone private copies
          cvalue = (free_list.is_empty() ? new Node_List()
                                         : free_list.pop());
          cregnd = (free_list.is_empty() ? new Node_List()
                                         : free_list.pop());
          cvalue->map(SharedInfo::stack0,NULL);
          cregnd->map(SharedInfo::stack0,NULL);
          for( uint l = 0; l < (uint)SharedInfo::stack0; l++ ) {
            cvalue->map(l,value[l]);
            cregnd->map(l,regnd[l]);
          }
        }
        blk2value[sb->_pre_order] = cvalue;
        blk2regnd[sb->_pre_order] = cregnd;
      } else { // Else merge mappings
        for( uint l = 0; l < (uint)SharedInfo::stack0; l++ ) {
          // Mappings disagree on exact Node defined by a register?
          if( regnd[l] != (*sregnd)[l] ) {
            svalue->map(l,NULL);
            sregnd->map(l,NULL);
          }
        }
      } // End of if prior mapping at successor or not
    } // End of push our mappings out to successor blocks

    // If we didn't use value and regnd in one of our successors
    // then we can put them on the free list.
    if (!used_once) {
      free_list.push(&value);
      free_list.push(&regnd);
    }

    // Flag as visited
    blk2value[b->_pre_order] = (Node_List*)-1;
    blk2regnd[b->_pre_order] = (Node_List*)-1;
  } // End for all blocks
}


