/*
 * @(#)regmask.hpp	1.47 99/07/20 11:21:10
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// Some fun naming (textual) substitutions:
//
// RegMask::get_low_elem() ==> RegMask::find_first_elem()
// RegMask::Special        ==> RegMask::Empty
// RegMask::_flags         ==> RegMask::is_AllStack()
// RegMask::operator<<=()  ==> RegMask::Insert()
// RegMask::operator>>=()  ==> RegMask::Remove()
// RegMask::Union()        ==> RegMask::OR
// RegMask::Inter()        ==> RegMask::AND
//
// OptoRegister::RegName   ==> OptoReg::Name
//
// SharedInfo::stack0      ==> _last_Mach_Reg  or ZERO in core version
//
// numregs in chaitin      ==> proper degree in chaitin


//------------------------------RegMask----------------------------------------
// The ADL file describes how to print the machine-specific registers, as well
// as any notion of register classes.  We provide a register mask, which is
// just a collection of Register numbers.  

// The ADLC defines 2 macros, RM_SIZE and FORALL_BODY.
// RM_SIZE is the size of a register mask in words.
// FORALL_BODY replicates a BODY macro once per word in the register mask.
// The usage is somewhat clumsy and limited to the regmask.[h,c]pp files.
// However, it means the ADLC can redefine the unroll macro and all loops
// over register masks will be unrolled by the correct amount.

class RegMask VALUE_OBJ_CLASS_SPEC {
  union {
    double _dummy_force_double_alignment[RM_SIZE>>1];
    // Array of Register Mask bits.  This array is large enough to cover 
    // all the machine registers and all parameters that need to be passed
    // on the stack (stack registers) up to some interesting limit.  Methods
    // that need more parameters will NOT be compiled.  On Intel, the limit
    // is something like 90+ parameters.
    int _A[RM_SIZE];
  };
public:
  // A constructor only used by the ADLC output.  All mask fields are filled
  // in directly.  Calls to this look something like RM(1,2,3,4);
  RegMask( 
#   define BODY(I) int a##I,
    FORALL_BODY   
#   undef BODY
    int dummy = 0 ) {
#   define BODY(I) _A[I] = a##I;
    FORALL_BODY
#   undef BODY
  }

  // Handy copying constructor
  RegMask( RegMask *rm ) {
#   define BODY(I) _A[I] = rm->_A[I];
    FORALL_BODY
#   undef BODY
  }

  // Construct an empty mask
  RegMask( ) { Clear(); }

  // Construct a mask with a single bit
  RegMask( OptoReg::Name reg ) { Clear(); Insert(reg); }

  // Check for register being in mask
  int Member( OptoReg::Name reg ) const {
    assert( reg < RM_SIZE*32, "" );
    return _A[reg>>5] & (1<<(reg&31));
  }

  // The last bit in the register mask indicates that the mask should repeat
  // indefinitely with ONE bits.  Returns TRUE if mask is infinite or 
  // unbounded in size.  Returns FALSE if mask is finite size.
  int is_AllStack() const { return _A[RM_SIZE-1] >> 31; }
  void set_AllStack() { _A[RM_SIZE-1] |= (1<<31); }

  // Test for being a not-empty mask.
  int is_NotEmpty( ) const {
    int tmp = 0;
#   define BODY(I) tmp |= _A[I];
    FORALL_BODY
#   undef BODY
    return tmp;
  }

  // Find lowest-numbered register from mask, or BAD if mask is empty.
  OptoReg::Name find_first_elem() const { 
    int base, bits;
#   define BODY(I) if( (bits = _A[I]) != 0 ) base = I<<5; else 
    FORALL_BODY
#   undef BODY
      { base = OptoReg::Bad; bits = 1<<0; }
    return OptoReg::Name(base + ff1(bits)); 
  }
  // Get highest-numbered register from mask, or BAD if mask is empty.
  OptoReg::Name find_last_elem() const { 
    int base, bits;
#   define BODY(I) if( (bits = _A[RM_SIZE-1-I]) != 0 ) base = (RM_SIZE-1-I)<<5; else 
    FORALL_BODY
#   undef BODY
      { base = OptoReg::Bad; bits = 1<<0; }
    return OptoReg::Name(base + fh1(bits)); 
  }

  // Find the lowest-numbered register pair in the mask.  Return the
  // HIGHEST register number in the pair, or BAD if no pairs.
  // Assert that the mask contains only bit pairs.
  OptoReg::Name find_first_pair() const;

  // Clear out partial bits; leave only aligned adjacent bit pairs.
  void ClearToPairs();
  // Smear out partial bits; leave only aligned adjacent bit pairs.
  void SmearToPairs();
  // Verify that the mask contains only aligned adjacent bit pairs
  void VerifyPairs() const { assert( is_aligned_Pairs(), "mask is not aligned, adjacent pairs" ); }
  // Test that the mask contains only aligned adjacent bit pairs
  bool is_aligned_Pairs() const;

  // Test for single register
  int is_bound1() const;
  // Test for a single adjacent pair
  int is_bound2() const;

  // Fast overlap test.  Non-zero if any registers in common.
  int overlap( const RegMask &rm ) const {
    return 
#   define BODY(I) (_A[I] & rm._A[I]) |
    FORALL_BODY
#   undef BODY
    0 ;
  }

  // Special test for register pressure based splitting
  // UP means register only, Register plus stack, or stack only is DOWN
  bool is_UP() const;

  // Clear a register mask
  void Clear( ) {
#   define BODY(I) _A[I] = 0;
    FORALL_BODY
#   undef BODY
  }

  // Fill a register mask with 1's
  void Set_All( ) {
#   define BODY(I) _A[I] = -1;
    FORALL_BODY
#   undef BODY
  }

  // Insert register into mask
  void Insert( OptoReg::Name reg ) {
    assert( reg < RM_SIZE*32, "" );
    _A[reg>>5] |= (1<<(reg&31));
  }

  // Remove register from mask
  void Remove( OptoReg::Name reg ) {
    assert( reg < RM_SIZE*32, "" );
    _A[reg>>5] &= ~(1<<(reg&31));
  }

  // OR 'rm' into 'this'
  void OR( const RegMask &rm ) {
#   define BODY(I) this->_A[I] |= rm._A[I];
    FORALL_BODY
#   undef BODY
  }

  // AND 'rm' into 'this'
  void AND( const RegMask &rm ) {
#   define BODY(I) this->_A[I] &= rm._A[I];
    FORALL_BODY
#   undef BODY
  }

  // Subtract 'rm' from 'this'
  void SUBTRACT( const RegMask &rm ) {
#   define BODY(I) _A[I] &= ~rm._A[I];
    FORALL_BODY
#   undef BODY
  }

  // Compute size of register mask: number of bits
  uint Size() const;

#ifndef PRODUCT
  void print() const { dump(); }
  void dump() const;            // Print a mask
#endif

  static const RegMask Empty;   // Common empty mask
};

/*
    // Difference
  void operator-= (const RegMask &rm) { 
    // Exclude on a best-efforts basis, but do not bother
    // to drop bits if rm is in a different chunk.
    // Any caller who needs more precision is responsible
    // for matching this->_chunk to rm._chunk explicitly.
    if( _chunk == rm._chunk )
      _regs &= ~rm._regs;
    _flags &= ~rm._flags;
  }

  // Get highest-numbered register from mask
  OptoReg::Name get_high_elem() const { 
    assert(_regs, "Empty mask"); 
    assert(!_flags, "Infinite mask"); 
    return OptoReg::Name((_chunk<<5) + fh1(_regs)); 
  }
  // Test for an even/odd pair
  int is_evenodd(const RegMask &rm) const;
  // Test whether it contains any non-stack regs.
  int has_nonstack() const {
    return (int)_chunk < (SharedInfo::stack0>>5) && _regs;
  }
  // Test whether it contains any stack slots.
  int has_stack() const {
    return _flags || (int)_chunk >= (SharedInfo::stack0>>5) && _regs;
  }
  // Return single bound register or Bad
  OptoReg::Name get_elem() const { 
    return is_bound1() ? get_low_elem() : OptoReg::Bad; }

  // Fast overlap test.  Non-zero if any registers in common.
  // Does not count _flags in the overlap computation.
  uint32 overlap( const RegMask &rm ) const {
    if( _flags && (int)_chunk >= (SharedInfo::stack0>>5) )
      return 0;
    if( _chunk == rm._chunk ) {
      int x = _regs & rm._regs;
      return x;
    }
    return 0;
  }
  // Fast overlap test.  Non-zero if any registers in common.
  // Does count _flags in the overlap computation.
  uint32 overlap_stack( const RegMask &rm ) const;

  // Fast subsumption test.  Non-zero if rm is a subset of this mask.
  uint32 contains_all( const RegMask &rm ) const {
    if( (rm._flags & ~_flags ) != 0 )
      return 0;
    else if( _chunk == rm._chunk || rm._regs == 0 )
      return (rm._regs & ~_regs) == 0;
    else if( _chunk < rm._chunk )
      return _flags && (int)(rm._chunk) >= (SharedInfo::stack0>>5);
    else
      return 0;
  }

  // Convert the RegMask rm to a infinite VectorSet and union the
  // bits into the VectorSet.
  friend VectorSet &operator <<= (VectorSet &set, const RegMask &rm );

  // Convenience built-in masks
  static const RegMask Empty;	// Empty register mask
  static const RegMask Special;	// The unallocated register
  static const RegMask Bad;	// The not-a-register mask
  static const RegMask Physical;// All physical registers


  void dump() const;	// Print a mask
#endif
  */


