/*
 * @(#)rootnode.cpp	1.60 00/03/01 11:57:06
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
#include "incls/_precompiled.incl"
#include "incls/_rootnode.cpp.incl"

//------------------------------Ideal------------------------------------------
// Remove dead inputs
Node *RootNode::Ideal(PhaseGVN *phase, PhaseDefUse *du) {
  for( uint i = 1; i < req(); i++ ) { // For all inputs
    // Check for and remove dead inputs
    if( phase->type(in(i)) == Type::TOP ) {
      set_req(i,NULL,du);
      del_req(i--);		// Delete TOP inputs
    }
  }

  // I used to do tail-splitting in the Ideal graph here, but it does not
  // work.  The tail-splitting forces values live into the Return to be
  // ready at a point which dominates the split returns.  This forces Stores
  // to be hoisted high.  The "proper" fix would be to split Stores down
  // each path, but this makes the split unprofitable.  If we want to do this
  // optimization, it needs to be done after allocation so we can count all
  // the instructions needing to be cloned in the cost metric.


  // caf-mark eliminate stupid tail recursion
  elim_tail_recursion( *phase );

  return NULL;			// No further opportunities exposed
}

//=============================================================================
//------------------------------skip_dull_nodes--------------------------------
// Helper function to skip past nodes along control paths.
static Node *skip_dull_nodes( Node *n ) {
  if( (n->Opcode() == Op_Proj) 
      || (n->Opcode() == Op_Parm) ) return skip_dull_nodes( n->in(0) );
  return n;
}

//------------------------------elim_tail_recursion----------------------------
// If this is the root of a tail-recursive method,
// transform to the loop-form, eliminating the recursion.

void RootNode::elim_tail_recursion( PhaseGVN &gvn ) {
  // This list of variables are referenced for the loop-version
  ParmNode *parm1;          // parameter for recursive call
  Node *base_val;           // base value of recursion
  Node *base_cmp_val;       // base condition test value
  Node *base_cmp;           // base comparison
  BoolNode *base_bool;      // base exit "if"
  Node *base_if;            // base condition if
  Node *rec_ops;            // recursive operations
  Node *rec_parm1;          // operations to adjust parameter 1

  // -----
  // Check control paths for early exit conditions
  // Must have two return sites and no exception exits
  if( req() != 2 ) return;
  if( in(1)->Opcode() != Op_Return ) return;
  const ReturnNode *retn = (ReturnNode*)in(1);
  // Merge 2 exit paths
  const Node *region = retn->in(0);
  if( !region->is_Region() ) return;

  // -----
  // Check that return type is <int>, extend later for longs, ...
  if( retn->req() != TypeFunc::Parms+1 ||
      retn->in(TypeFunc::Parms)->bottom_type() != TypeInt::INT ) 
    return;

  // Swap so these can be
  // (1) for base case
  // (2) for recursive case
  int base = 1;
  int rcur = 2;
  // If control flow path is dead, wait until it is eliminated.
  if( region->in(rcur) == NULL ) return;
  if( region->in(rcur)->Opcode() != Op_Proj ) {
    // Swap base/rcur if in wrong order
    base = 2;
    rcur = 1;
  }
  if( region->in(rcur) == NULL ) return;
  if( region->in(rcur)->Opcode() != Op_Proj ) return;

  // -----
  // Check that region is fully populated!
  if( !region->in(base) ) return;

  // Check that ret1 fits the base case, an Op_If
  const Node *if_ctrl = region->in(base)->is_block_proj();
  if( if_ctrl == NULL )                   return;
  if( if_ctrl->Opcode() != Op_If )        return;
  base_if = (IfNode*)if_ctrl;
  // And its return value comes from an invariant !!!!!
  // .. but check first if the number of arguments is correct
  if (retn->in( TypeFunc::Parms )->len() <= (uint)base ) return;
  base_val = retn->in( TypeFunc::Parms )->in(base);
  if( base_val->Opcode() != Op_ConI )     return;

  // -----
  // This If comes from start and compares param:1 with an invariant !!!!!
  StartNode *start = (StartNode*)skip_dull_nodes( if_ctrl->in(0) );
  if( start->is_Start() == NULL )       return;
  // Check comparison
  const BoolNode *if_bool = if_ctrl->in(1)->is_Bool();
  if( ! if_bool )                         return;
  if( if_bool->in(0) != NULL )            return;
  // bool_test = if_bool->_test;
  const Node *if_cmp  = if_bool->in(1);
  if( if_cmp->in(0)  != NULL )            return;
  if( if_cmp->in(1)->Opcode() != Op_Parm) return;
  parm1  = (ParmNode*)if_cmp->in(1);
  // if( parm1->_con != (TypeFunc::Parms + 1) )  return;
  if( if_cmp->in(2)->Opcode() != Op_ConI) return;
  base_bool    = (BoolNode*)if_bool;
  base_cmp     = (Node*)if_cmp;
  base_cmp_val = base_cmp->in(2);

  // -----
  // Check that ret2 fits the recursive case, Projection of recursive call
  Node *ctrl2 = (Node*)skip_dull_nodes( region->in(rcur) );
  if( ctrl2->is_Call() == NULL )          return;
  const CallJavaNode *call = (ctrl2->is_Call())->is_CallJava();
  if( call == NULL  )                     return;
  // Check that this is a recursive call
  if( call->_method != gvn.C->method() )      return;
  // call's control comes from the same If, before IfFalse or IfTrue
  const Node *call_ctrl = (const Node *)skip_dull_nodes( call->in(0) );
  call_ctrl = call_ctrl->is_block_proj();
  if( call_ctrl != if_ctrl )              return;

  // Required fields of ret2 come from Parameters or call's ProjNodes
  // control and return value are handled separately
  int  idx;
  int  max = TypeFunc::Parms-1;
  for( idx=1; idx < max; ++idx ) {
    Node *x = retn->in(idx);
    if( x->is_Phi() ) x = x->in(rcur);
    Node *input = skip_dull_nodes( x );
    if( ((const Node*)input != call) && (input->is_Start() == NULL) ) return;
  }

  // -----
  // Get the node that supplies the return value
  rec_ops = retn->in( TypeFunc::Parms )->in(rcur);
  // Check that all ops in this tree commute
  // !!!!! currently: one node allowed in commuting tree
  if( rec_ops->Opcode() != Op_AddI )      return;
  if( rec_ops->in(0) != NULL )            return;

  // Check which input edge comes from the recursive call...
  // (rec_in1) from parameters
  // (rec_in2) from call return value
  const Node *rec_in1 = rec_ops->in(1);
  const Node *rec_in2 = rec_ops->in(2);
  if( rec_in1->Opcode() == Op_Proj ) {
    const Node *recTmp = rec_in1;
    rec_in1 = rec_in2;  rec_in2 = recTmp;
  }
  // one input from start's parameters
  if( rec_in1->Opcode() != Op_Parm )      return;
  if( ((ParmNode*)rec_in1)->_con != (TypeFunc::Parms + 1) ) return;
  // other input from result of recursive call
  rec_in2  = skip_dull_nodes( rec_in2->in(0) );
  if( rec_in2 != call )                   return;

  // -----
  // Check the input to the recursive call, Parm:0 is instance, Parm:1 is param
  if( call->_func->_domain->_cnt != (TypeFunc::Parms + 1 + 1) ) return;
  rec_parm1            = call->in( TypeFunc::Parms + 1 );
  // Call's input op must count-up or count-down on call's param1
  const Node *call_in1 = rec_parm1->in(1);
  const Node *call_in2 = rec_parm1->in(2);
  if( call_in1->Opcode() != Op_Parm ) { // swap
    const Node *inTmp = call_in1;
    call_in1 = call_in2; call_in2 = inTmp;
  }
  if( call_in1->Opcode() != Op_Parm )     return;
  // Check that inc/dec is invariant !!!!!
  if( call_in2->Opcode() != Op_ConI )     return;
  const Node *inc_dec = call_in2;

  // Bail out if any statics, arrays or fields get modified
  if( call->in(TypeFunc::Memory)->in(0) != start ) return;


  // -------------------------- Build loop form -------------------------------
  // Reuse Start node and associated parameters

  // Reuse invariants, parm1, base_val, base_cmp_val, entry_debug
  // Reuse and Clone base_cmp
  // Clone base_ret, rec_parm1
  Node            *entry_debug = base_if->in(0);  // debug node at entry
  BoolTest::mask   bool_mask   = ((base_bool->_test)._test);
  
  // -----
  // Two regions will be needed
  RegionNode  *region_3 = new RegionNode( 3 );
  RegionNode  *region_6 = new RegionNode( 3 );
  // Need to set region_3's and region_6's inputs

  // -----
  // Compare incoming parameter to base test value
  Node        *cmp_24   = base_cmp;
  BoolNode    *bool_23  = new (2) BoolNode( cmp_24, bool_mask );
  gvn._types.map( bool_23->_idx, bool_23->bottom_type() );
  IfNode      *if_9     = new (2) IfNode( entry_debug, bool_23, (float)0.5, (float)-1.0 );
  gvn._types.map( if_9->_idx, if_9->bottom_type() );
  IfFalseNode *iff_31   = new (1) IfFalseNode( if_9 );
  gvn._types.map( iff_31->_idx, iff_31->bottom_type() );
  IfTrueNode  *ift_8    = new (1) IfTrueNode(  if_9 );
  gvn._types.map( ift_8->_idx, ift_8->bottom_type() );

  // -----
  // Compare modified parameter to base test value
  Node        *cmp_27   = base_cmp->clone();
  gvn._types.map( cmp_27->_idx, cmp_27->bottom_type() );
  cmp_27->set_req( 0, 0 );
  // cmp_27->set_req( 1, add_28 );
  cmp_27->set_req( 2, base_cmp_val );
  Node        *bool_26  = new (2) BoolNode( cmp_27, bool_mask );
  gvn._types.map( bool_26->_idx, bool_26->bottom_type() );
  IfNode      *if_5     = new (2) IfNode( region_6, bool_26, (float)0.5, (float)-1.0 );
  gvn._types.map( if_5->_idx, if_5->bottom_type() );
  IfFalseNode *iff_4    = new (1) IfFalseNode( if_5 );
  gvn._types.map( iff_4->_idx, iff_4->bottom_type() );
  IfTrueNode  *ift_7    = new (1) IfTrueNode(  if_5 );
  gvn._types.map( ift_7->_idx, ift_7->bottom_type() );

  // -----
  // Compute test value for termination condition
  Node        *add_28   = rec_parm1->clone();
  gvn._types.map( add_28->_idx, TypeInt::INT );
  cmp_27->set_req( 1, add_28 );          // pending set_req
  // add_28->set_req( 1, phi_29 );
  PhiNode     *phi_29   = new (3) PhiNode( region_6, TypeInt::INT );
  gvn._types.map( phi_29->_idx, TypeInt::INT );
  phi_29->set_req( 1, add_28 );
  phi_29->set_req( 2, parm1  );
  add_28->set_req( 1, phi_29 );          // pending set_req to close cycle

  // -----
  // Compute result value
  Node        *add_33   = add_28->clone();    // !!!!! rec_ops
  gvn._types.map( add_33->_idx, add_33->bottom_type() );
  // add_33->set_req( 2, phi_34 );
  PhiNode     *phi_34   = new (3) PhiNode( region_6, TypeInt::INT );
  gvn._types.map( phi_34->_idx, phi_34->bottom_type() );
  phi_34->set_req( 1, add_33   );
  phi_34->set_req( 2, base_val );
  add_33->set_req( 2, phi_34   );        // pending set_req


  // clone base-case return
  PhiNode     *phi_32   = new (3) PhiNode( region_3, TypeInt::INT );
  gvn._types.map( phi_32->_idx, phi_32->bottom_type() );
  phi_32->set_req( 1, add_33   );
  phi_32->set_req( 2, base_val );
  Node        *ret      = retn->clone();
  gvn._types.map( ret->_idx, ret->bottom_type() );
  ret->set_req( 0, region_3 );
  for( uint i=1; i<TypeFunc::Parms-1; i++ )
    if( ret->in(i)->is_Phi() ) 
      ret->set_req(i,ret->in(i)->in(base));
  ret->set_req( TypeFunc::Parms, phi_32 );

  // Set region's inputs
  region_3->set_req( 0, region_3 );
  region_3->set_req( 1, iff_4    );
  region_3->set_req( 2, iff_31   );
  gvn._types.map( region_3->_idx, region_3->bottom_type() );
  region_6->set_req( 0, region_6 );
  region_6->set_req( 1, ift_7    );
  region_6->set_req( 2, ift_8    );
  gvn._types.map( region_6->_idx, region_6->bottom_type() );

  this->set_req( 1, ret );

  return;
}

//=============================================================================
HaltNode::HaltNode( Node *ctrl, Node *frameptr ) : Node(TypeFunc::Parms) { 
  Node *top = Compile::current()->_top_node;
  set_req(TypeFunc::Control,  ctrl        ); 
  set_req(TypeFunc::I_O,      top);
  set_req(TypeFunc::Memory,   top);
  set_req(TypeFunc::FramePtr, frameptr    );
  set_req(TypeFunc::ReturnAdr,top);
}

const Type *HaltNode::bottom_type() const { return Type::BOTTOM; }

//------------------------------Ideal------------------------------------------
Node *HaltNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  return remove_dead_region(phase,du) ? this : NULL;
}

//------------------------------Value------------------------------------------
const Type *HaltNode::Value( PhaseTransform *phase ) const { 
  return ( phase->type(in(TypeFunc::Control)) == Type::TOP)
    ? Type::TOP
    : Type::BOTTOM;
}

const RegMask &HaltNode::out_RegMask() const { 
  return RegMask::Empty;
}


