/*
 * @(#)subnode.cpp	1.104 01/02/09 14:47:08
 *
 * Copyright 1991-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
// Portions of code courtesy of Clifford Click

// Optimization - Graph Style

#include "incls/_precompiled.incl"
#include "incls/_subnode.cpp.incl"
#include "math.h"

//=============================================================================
//------------------------------Identity---------------------------------------
// If right input is a constant 0, return the left input.  
Node *SubNode::Identity( PhaseTransform *phase ) {
  if( in(1) == this ) return phase->C->_top_node;
  if( in(2) == this ) return phase->C->_top_node;

  // Remove double negation
  const Type *zero = add_id();
  if( phase->type( _in[1] )->higher_equal( zero ) &&
      _in[2]->Opcode() == Opcode() &&
      phase->type( _in[2]->in(1) )->higher_equal( zero ) ) {
    return _in[2]->in(2);
  }

  // Convert "(X+Y) - Y" into X
  if( in(1)->Opcode() == Op_AddI ) {
    if( phase->eqv(in(1)->in(2),in(2)) )
      return in(1)->in(1);
    // Also catch: "(X + Opaque2(Y)) - Y".  In this case, 'Y' is a loop-varying
    // trip counter and X is likely to be loop-invariant (that's how O2 Nodes
    // are originally used, although the optimizer sometimes jiggers things).
    // This folding through an O2 removes a loop-exit use of a loop-varying
    // value and generally lowers register pressure in and around the loop.
    if( in(1)->in(2)->Opcode() == Op_Opaque2 &&
        phase->eqv(in(1)->in(2)->in(1),in(2)) )
      return in(1)->in(1);
  }

  return ( phase->type( _in[2] )->higher_equal( zero ) ) ? in(1) : this;
}

//------------------------------Value------------------------------------------
// A subtract node differences it's two inputs.  
const Type *SubNode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  if( t1 == Type::TOP ) return Type::TOP;
  const Type *t2 = phase->type( _in[2] );
  if( t2 == Type::TOP ) return Type::TOP;

  // Not correct for SubFnode and AddFNode (must check for infinity)
  // Equal?  Subtract is zero
  if( phase->eqv( in(1), in(2) ) ) return add_id();

  // Either input is BOTTOM ==> the result is the local BOTTOM
  if( t1 == Type::BOTTOM || t2 == Type::BOTTOM ) 
    return bottom_type();

  return sub(t1,t2);            // Local flavor of type subtraction

}

//=============================================================================
//------------------------------Ideal------------------------------------------
Node *SubINode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  Node *in1 = in(1);
  Node *in2 = in(2);
  if( in1 == this ) return NULL;
  if( in2 == this ) return NULL;
  uint op1 = in1->Opcode();
  uint op2 = in2->Opcode();

  const Type *t2 = phase->type( in2 );
  // Convert "x-c0" into "x+ -c0".
  if( t2->_base == Type::Int ){	// Might be bottom or top...
    const TypeInt *i = t2->is_int();
    if( i->is_con() )
      return new (3) AddINode(in(1), phase->intcon(-i->get_con()));
  }

  // Convert "(x+c0) - y" into (x-y) + c0"
  if( op1 == Op_AddI && in1->in(1) != in1 ) {
    // Do not collapse (x+y)-y if "+" is a loop increment, because the
    // "-" is loop invariant and collapsing extends the live-range of "x"
    // to overlap with the "+", forcing another register to be used in
    // the loop.
    const PhiNode *phi;
    CountedLoopNode *l;
    // This test will be clearer with '&&' (apply DeMorgan's rule)
    // but I like the early cutouts that happen here.
    if( !(phi=in1->in(1)->is_Phi()) ||
        !phi->in(0) ||
	!(l=phi->in(0)->is_CountedLoop()) ||
	in1 != l->incr() ) {
      const Type *tadd = phase->type( in1->in(2) );
      if( tadd->singleton() ) {
        Node *sub2 = phase->transform( new (3) SubINode( in1->in(1), in2 ));
        return new (3) AddINode( sub2, in1->in(2) );
      }
    }
  }
		       
  // Convert "x - (x+y)" into "-y"
  if( op2 == Op_AddI &&
      phase->eqv( in1, in2->in(1) ) )
    return new (3) SubINode( phase->intcon(0),in2->in(2));
  // Convert "x - (y+x)" into "-y"
  if( op2 == Op_AddI &&
      phase->eqv( in1, in2->in(2) ) )
    return new (3) SubINode( phase->intcon(0),in2->in(1));

  // Convert "0 - (x-y)" into "y-x"
  if( phase->type( in1 ) == TypeInt::ZERO && op2 == Op_SubI ) 
    return new (3) SubINode( in2->in(2), in2->in(1) );

  // Convert "(X+A) - (X+B)" into "A - B"
  if( op1 == Op_AddI && op2 == Op_AddI && in1->in(1) == in2->in(1) )
    return new (3) SubINode( in1->in(2), in2->in(2) );

  // Convert "(A+X) - (B+X)" into "A - B"
  if( op1 == Op_AddI && op2 == Op_AddI && in1->in(2) == in2->in(2) )
    return new (3) SubINode( in1->in(1), in2->in(1) );

  // Convert "A-(B-C)" into (A+C)-B", since add is commutative and generally
  // nicer to optimize than subtract.
  if( op2 == Op_SubI && in2->in(1) != this ) {
    Node *add1 = phase->transform( new (3) AddINode( in1, in2->in(2) ) );
    return new (3) SubINode( add1, in2->in(1) );
  }

  return NULL;
}

//------------------------------sub--------------------------------------------
// A subtract node differences it's two inputs.  
const Type *SubINode::sub( const Type *t1, const Type *t2 ) const {
  const TypeInt *r0 = t1->is_int(); // Handy access
  const TypeInt *r1 = t2->is_int();
  int32 lo = r0->_lo - r1->_hi;
  int32 hi = r0->_hi - r1->_lo;

  // We next check for 32-bit overflow.
  // If that happens, we just assume all integers are possible.
  if( (((r0->_lo ^ r1->_hi) >= 0) ||	// lo ends have same signs OR
       ((r0->_lo ^      lo) >= 0)) &&	// lo results have same signs AND
      (((r0->_hi ^ r1->_lo) >= 0) ||	// hi ends have same signs OR
       ((r0->_hi ^      hi) >= 0)) )	// hi results have same signs
    return TypeInt::make(lo,hi,max(r0->_widen,r1->_widen));
  else				// Overflow; assume all integers
    return TypeInt::INT;
}

//=============================================================================
//------------------------------Ideal------------------------------------------
Node *SubLNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  Node *in1 = in(1);
  Node *in2 = in(2);
  if( in1 == this ) return NULL;
  if( in2 == this ) return NULL;
  uint op1 = in1->Opcode();
  uint op2 = in2->Opcode();

  const TypeLong *i = phase->type( in2 )->isa_long();
  // Convert "x-c0" into "x+ -c0".
  if( i &&	                // Might be bottom or top...
      i->is_con() )
    return new (3) AddLNode(in(1), phase->makecon(TypeLong::make(-i->get_con())));

  // Convert "(x+c0) - y" into (x-y) + c0"
  if( op1 == Op_AddL && in1->in(1) != in1 ) {
    // Do not collapse (x+y)-y if "+" is a loop increment, because the
    // "-" is loop invariant and collapsing extends the live-range of "x"
    // to overlap with the "+", forcing another register to be used in
    // the loop.
    const PhiNode *phi;
    CountedLoopNode *l;
    if( !(phi=in1->in(1)->is_Phi()) ||
        !phi->in(0) ||
	!(l=phi->in(0)->is_CountedLoop()) ||
	in1 != l->incr() ) {
      const Type *tadd = phase->type( in1->in(2) );
      if( tadd->singleton() ) {
        Node *sub2 = phase->transform( new (3) SubLNode( in1->in(1), in2 ));
        return new (3) AddLNode( sub2, in1->in(2) );
      }
    }
  }
		       
  // Convert "x - (x+y)" into "-y"
  if( op2 == Op_AddL &&
      phase->eqv( in1, in2->in(1) ) )
    return new (3) SubLNode( phase->makecon(TypeLong::ZERO), in2->in(2));
  // Convert "x - (y+x)" into "-y"
  if( op2 == Op_AddL &&
      phase->eqv( in1, in2->in(2) ) )
    return new (3) SubLNode( phase->makecon(TypeLong::ZERO),in2->in(1));

  // Convert "0 - (x-y)" into "y-x"
  if( phase->type( in1 ) == TypeLong::ZERO && op2 == Op_SubL ) 
    return new (3) SubLNode( in2->in(2), in2->in(1) );

  // Convert "(X+A) - (X+B)" into "A - B"
  if( op1 == Op_AddL && op2 == Op_AddL && in1->in(1) == in2->in(1) )
    return new (3) SubLNode( in1->in(2), in2->in(2) );

  // Convert "(A+X) - (B+X)" into "A - B"
  if( op1 == Op_AddL && op2 == Op_AddL && in1->in(2) == in2->in(2) )
    return new (3) SubLNode( in1->in(1), in2->in(1) );

  // Convert "A-(B-C)" into (A+C)-B"
  if( op2 == Op_SubL && in2->in(1) != this ) {
    Node *add1 = phase->transform( new (3) AddLNode( in1, in2->in(2) ) );
    return new (3) SubLNode( add1, in2->in(1) );
  }

  return NULL;
}

//------------------------------sub--------------------------------------------
// A subtract node differences it's two inputs.  
const Type *SubLNode::sub( const Type *t1, const Type *t2 ) const {
  const TypeLong *r0 = t1->is_long(); // Handy access
  const TypeLong *r1 = t2->is_long();
  jlong lo = r0->_lo - r1->_hi;
  jlong hi = r0->_hi - r1->_lo;

  // We next check for 32-bit overflow.
  // If that happens, we just assume all integers are possible.
  if( (((r0->_lo ^ r1->_hi) >= 0) ||	// lo ends have same signs OR
       ((r0->_lo ^      lo) >= 0)) &&	// lo results have same signs AND
      (((r0->_hi ^ r1->_lo) >= 0) ||	// hi ends have same signs OR
       ((r0->_hi ^      hi) >= 0)) )	// hi results have same signs
    return TypeLong::make(lo,hi,max(r0->_widen,r1->_widen));
  else				// Overflow; assume all integers
    return TypeLong::LONG;
}

//=============================================================================
//------------------------------Value------------------------------------------
// A subtract node differences it's two inputs.  
const Type *SubFPNode::Value( PhaseTransform *phase ) const {
  // Either input is TOP ==> the result is TOP
  const Type *t1 = phase->type( _in[1] );
  if( t1 == Type::TOP ) return Type::TOP;
  const Type *t2 = phase->type( _in[2] );
  if( t2 == Type::TOP ) return Type::TOP;

  // if both operands are infinity of same sign, the result is NaN; do
  // not replace with zero
  if( (t1->is_finite() && t2->is_finite()) ) {
    if( phase->eqv( in(1), in(2) ) ) return add_id();
  }

  // Either input is BOTTOM ==> the result is the local BOTTOM
  const Type *bot = bottom_type();
  if( (t1 == bot) || (t2 == bot) ||
      (t1 == Type::BOTTOM) || (t2 == Type::BOTTOM) ) 
    return bot;

  return sub(t1,t2);            // Local flavor of type subtraction

}


//=============================================================================
//------------------------------Ideal------------------------------------------
Node *SubFNode::Ideal( PhaseGVN *phase,PhaseDefUse *du ) {
  const Type *t2 = phase->type( in(2) );
  // Convert "x-c0" into "x+ -c0".
  if( t2->_base == Type::FloatCon ) {	// Might be bottom or top...
    // return new (3) AddFNode(_in[1], phase->makecon( TypeF::make(-t2->getf()) ) );
  }
		       
  // Not associative because of boundary conditions (infinity)
  if( IdealizedNumerics &&
    !(EnforceStrictFP || phase->C->method()->is_strict()) ) {
    // Convert "x - (x+y)" into "-y"
    if( in(2)->is_Add() &&
        phase->eqv(in(1),in(2)->in(1) ) ) 
      return new (3) SubFNode( phase->makecon(TypeF::ZERO),in(2)->in(2));
  }

  // Cannot replace 0.0-X with -X because a 'fsub' bytecode computes
  // 0.0-0.0 as +0.0, while a 'fneg' bytecode computes -0.0.
  //if( phase->type(in(1)) == TypeF::ZERO )
  //return new (2) NegFNode(in(2));

  return NULL;
}

//------------------------------sub--------------------------------------------
// A subtract node differences it's two inputs.  
const Type *SubFNode::sub( const Type *t1, const Type *t2 ) const {
  // no folding if one of operands is infinity or NaN, do not do constant folding
  if( g_isfinite(t1->getf()) && g_isfinite(t2->getf()) ) {
    return TypeF::make( t1->getf() - t2->getf() );
  } 
  else if( g_isnan(t1->getf()) ) {
    return t1;
  } 
  else if( g_isnan(t2->getf()) ) {
    return t2;
  } 
  else {
    return Type::FLOAT;
  }
}

//=============================================================================
//------------------------------Ideal------------------------------------------
Node *SubDNode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  const Type *t2 = phase->type( _in[2] );
  // Convert "x-c0" into "x+ -c0".
  if( t2->_base == Type::DoubleCon ) {	// Might be bottom or top...
    // return new (3) AddDNode(_in[1], phase->makecon( TypeD::make(-t2->getd()) ) );
  }
		       
  // Not associative because of boundary conditions (infinity)
  if( IdealizedNumerics &&
    !(EnforceStrictFP || phase->C->method()->is_strict()) ) { 
    // Convert "x - (x+y)" into "-y"
    if( _in[2]->is_Add() &&
        phase->eqv(_in[1],_in[2]->in(1) ) ) 
      return new (3) SubDNode( phase->makecon(TypeD::ZERO),_in[2]->in(2));
  }

  // Cannot replace 0.0-X with -X because a 'dsub' bytecode computes
  // 0.0-0.0 as +0.0, while a 'dneg' bytecode computes -0.0.
  //if( phase->type(in(1)) == TypeD::ZERO )
  //return new (2) NegDNode(in(2));

  return NULL;
}

//------------------------------sub--------------------------------------------
// A subtract node differences it's two inputs.  
const Type *SubDNode::sub( const Type *t1, const Type *t2 ) const {
  // no folding if one of operands is infinity or NaN, do not do constant folding
  if( g_isfinite(t1->getd()) && g_isfinite(t2->getd()) ) {
    return TypeD::make( t1->getd() - t2->getd() );
  }
  else if( g_isnan(t1->getd()) ) {
    return t1;
  }
  else if( g_isnan(t2->getd()) ) {
    return t2;
  }
  else {
    return Type::DOUBLE;
  }
}

//=============================================================================
//------------------------------sub--------------------------------------------
// Simplify an CmpI (compare 2 integers) node, based on local information.
// If both inputs are constants, compare them.  
const Type *CmpINode::sub( const Type *t1, const Type *t2 ) const {
  const TypeInt *r0 = t1->is_int(); // Handy access
  const TypeInt *r1 = t2->is_int();

  if( r0->_hi < r1->_lo ) 	// Range is always low?
    return TypeInt::MINUS_1;
  else if( r0->_lo > r1->_hi )	// Range is always high?
    return TypeInt::ONE;

  else if( r0->is_con() && r1->is_con() ) { // comparing constants?
    assert(r0->get_con() == r1->get_con(), "must be equal");
    return TypeInt::ZERO;	// Equal results.
  } else if( r0->_hi == r1->_lo ) // Range is never high?
    return TypeInt::make(-1,0);
  else if( r0->_lo == r1->_hi )	// Range is never low?
    return TypeInt::BOOL;
  return TypeInt::CC;		// else use worst case results
}

// Simplify a CmpU (compare 2 integers) node, based on local information.
// If both inputs are constants, compare them.  
const Type *CmpUNode::sub( const Type *t1, const Type *t2 ) const {
  if (t1->isa_ptr()) {
    assert(t2->isa_ptr(), "comparing int and pointer");
    return TypeInt::CC;	
  }

  // comparing two unsigned ints
  const TypeInt *r0 = t1->is_int();   // Handy access
  const TypeInt *r1 = t2->is_int();

  // Current installed version
  // Compare ranges for non-overlap
  int lo0 = r0->_lo;
  int hi0 = r0->_hi;
  int lo1 = r1->_lo;
  int hi1 = r1->_hi;
  
  // Replacing comparison using constant with comparisons using ranges
  // Constants miss opportunities: array-bounds check of offset -1 with int+
  // 
  if( (lo0 < 0 && hi0 >= 0) || (lo1 < 0 && hi1 >= 0)) { 
    // This case needs further testing for "unsigned" compare
  } else {
    // results are reversed, '-' > '+' for unsigned compare
    if( (uint)hi0 < (uint)lo1 ) {
      return TypeInt::MINUS_1;        // smaller
    } else if( (uint)lo0 > (uint)hi1 ) {
      return TypeInt::ONE;            // greater
    } else if( ((uint)hi0 == (uint)lo1) && ((uint)lo0 == (uint)hi1) ) {
      return TypeInt::ZERO;           // Equal results
    } else if( (uint)lo0 >= (uint)hi1 ) {
      return TypeInt::make(0,1);
    } else if( (uint)hi0 <= (uint)lo1 ) {
      // Check for special case in Hashtable::get - the hash index is
      // mod'ed to the table size so the following range check is useless.
      // Check for: (X Mod Y) CmpU Y, where the mod result and Y both have
      // to be positive.
      if( lo0 >= 0 && lo1 >= 0 && 
          in(1)->Opcode() == Op_ModI &&
          in(1)->in(2) == in(2) )
        return TypeInt::MINUS_1;
      return TypeInt::make(-1,0);
    }
  }
  // Check for special case in Hashtable::get - the hash index is
  // mod'ed to the table size so the following range check is useless.
  // Check for: (X Mod Y) CmpU Y, where the mod result and Y both have
  // to be positive.
  if( lo0 >= 0 && lo1 >= 0 && 
      in(1)->Opcode() == Op_ModI &&
      in(1)->in(2) == in(2) )
    return TypeInt::MINUS_1;

  return TypeInt::CC;                 // else use worst case results
}

//------------------------------Idealize---------------------------------------
Node *CmpINode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {

  // Comparing a SUB against a zero is equal to comparing the
  // arguments directly.
  if( in(1)->Opcode() == Op_SubI &&
      phase->type( in(2) ) == TypeInt::ZERO ) 
    return new (3) CmpINode(in(1)->in(1),in(1)->in(2));
  else if( _in[2]->Opcode() == Op_SubI &&
	   phase->type( _in[1] ) == TypeInt::ZERO ) 
    return new (3) CmpINode(_in[2]->in(2),_in[2]->in(1));

  return NULL;			// No change
}

// Unlike SubNodes, subtract zero must still flatten return value to
// the range -1, 0, 1.  However, if the left input is in the range [-1,0,1]
// and the right input is zero...
Node *CmpINode::Identity( PhaseTransform *phase ) {
  return this;
}

Node *CmpUNode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  // right now, CmpUNodes are used for bounds & heap check only, no opportunity
  // for optimization
  return NULL;			// No change
}

Node *CmpUNode::Identity( PhaseTransform *phase ) {
  return this;
}


//=============================================================================
//------------------------------sub--------------------------------------------
// Simplify an CmpP (compare 2 pointers) node, based on local information.
// If both inputs are constants, compare them.  
const Type *CmpPNode::sub( const Type *t1, const Type *t2 ) const {
  const TypePtr *r0 = t1->is_ptr(); // Handy access
  const TypePtr *r1 = t2->is_ptr();
	
  // Undefined inputs makes for an undefined result
  if( r0->_ptr == TypePtr::TopPTR ) return Type::TOP;
  if( r1->_ptr == TypePtr::TopPTR ) return Type::TOP;
  if( r0->_ptr == TypePtr::AnyNull ||
      r1->_ptr == TypePtr::AnyNull ) 
    return Type::TOP;

  // See if it is 2 unrelated classes
  const TypeKlassPtr *k0 = r0->isa_klassptr();
  const TypeKlassPtr *k1 = r1->isa_klassptr();
  if( k0 && k1 ) {
    ciKlass* klass0 = k0->klass();
    ciKlass* klass1 = k1->klass();
    // Neither is an interface
    if( !klass0->is_interface() && !klass1->is_interface() ) {
      // See if neither subclasses the order, or if the class on top
      // is precise.  In either of these cases, the compare must fail.
      if( k0 == k1 ) {          // Equal types (and no subtyping)
        if( k0->singleton() )   // No subtyping check
          return TypeInt::ZERO; // Equal types are equal
        // I lose here exactly 3 times in a run of Spec... CNC 9/8/99
        // SomeDay I'll make all not-subklasses klasses precise
        // and fix this.
      } else if( klass0->equals(klass1) ) { // if types are unequal but klasses are
        // equal, then we have imprecise types.
        // Do nothing; we know nothing for imprecise types
      } else if( klass0->is_subtype_of(klass1) ) {
        // If klass1's type is PRECISE, then we can fail.
        if( k1->_ptr == TypePtr::Constant ) return TypeInt::ONE;
      } else if( klass1->is_subtype_of(klass0) ) {
        // If klass0's type is PRECISE, then we can fail.
        if( k0->_ptr == TypePtr::Constant ) return TypeInt::ONE;
      } else                    // Neither subtypes the other
        return TypeInt::ONE;    // ...so always fail
    }
  }


  // Unknown inputs makes an unknown result
  if( r0->singleton() ) {
    jint bits0 = r0->get_con();
    if( r1->singleton() ) 
      return bits0 == r1->get_con() ? TypeInt::ZERO : TypeInt::ONE;
    return ( r1->_ptr == TypePtr::NotNull && bits0==0 ) ? TypeInt::ONE : TypeInt::CC;
  } else if( r1->singleton() ) {
    jint bits1 = r1->get_con();
    return ( r0->_ptr == TypePtr::NotNull && bits1==0 ) ? TypeInt::ONE : TypeInt::CC;
  } else 
    return TypeInt::CC;
}

//------------------------------Idealize---------------------------------------
// Unlike SubNodes, compare to NULL must still flatten return value to
// the range -1, 0, 1.
Node *CmpPNode::Identity( PhaseTransform *phase ) {
  return this;
}

//=============================================================================

Node *CmpFLessNode::Identity( PhaseTransform *phase ) {
  return this;
}

Node *CmpFLessNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  return NULL;
}

// Simplify an CmpFless (compare 2 floats ) node, based on local information.
// If both inputs are constants, compare them.  
// currently not implemented
const Type *CmpFLessNode::sub( const Type *t1, const Type *t2 ) const {
  return TypeInt::CC;		// else use worst case results
}

//------------------------------Value------------------------------------------
// A subtract node differences it's two inputs.  
const Type *CmpFLessNode::Value( PhaseTransform *phase ) const {
  // because of NaN handling, must do complete coparision
  if( phase->eqv( in(1), in(2) ) ) return TypeInt::CC;
  return SubNode::Value(phase);
}

//=============================================================================

Node *CmpFGreaterNode::Identity( PhaseTransform *phase ) {
  return this;
}

Node *CmpFGreaterNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  return NULL;
}

// Simplify an CmpFless (compare 2 floats ) node, based on local information.
// If both inputs are constants, compare them.  
// currently not implemented
const Type *CmpFGreaterNode::sub( const Type *t1, const Type *t2 ) const {
  return TypeInt::CC;		// else use worst case results
}

//------------------------------Value------------------------------------------
// A subtract node differences it's two inputs.  
const Type *CmpFGreaterNode::Value( PhaseTransform *phase ) const {
  // because of NaN handling, must do complete coparision
  if( phase->eqv( in(1), in(2) ) ) return TypeInt::CC;
  return SubNode::Value(phase);
}


//=============================================================================
//------------------------------CmpD_Ideal_helper------------------------------
static Node *CmpD_Ideal_helper(PhaseGVN *phase, SubDNode *thsi) {
  assert( thsi->Opcode() == Op_CmpDLess || thsi->Opcode() == Op_CmpDGreater,
    "Must not be used for arbitrary SubDNodes");

  // Change  (CMPD (F2D (float)) (ConD value))
  // To      (CMPF      (float)  (ConF value))
  // Valid when 'value' does not lose precision as a float.
  // Benefits: eliminates conversion, does not require 24-bit mode
  Node *in1 = thsi->in(1);
  Node *in2 = thsi->in(2);
  if( ConvertCmpD2CmpF &&
      in1->Opcode() == Op_ConvF2D &&
      in2->Opcode() == Op_ConD ) {
    const TypeD *t2 = thsi->in(2)->is_Type()->get_type()->is_double_constant();
    double t2_value_as_double = t2->_d;
    float  t2_value_as_float  = (float)t2_value_as_double;
    if( t2_value_as_double == (double)t2_value_as_float ) {
      // Test value can be represented as a float
      // Eliminate the conversion to double and create new comparison
      Node  *new_in2 = phase->makecon( TypeF::make(t2_value_as_float) );
      Node  *new_in1 = thsi->in(1)->in(1);
      Node  *new_cmp = (thsi->Opcode() == Op_CmpDLess) ? 
                       (SubDNode*)new (3) CmpFLessNode( new_in1, new_in2 ) : 
                       (SubDNode*)new (3) CmpFGreaterNode( new_in1, new_in2 );
      return new_cmp;           // Changed to CmpF*Node
    }
    // Testing value required the precision of a double
  }
  return NULL;			// No change
}

//=============================================================================
//------------------------------Ideal------------------------------------------
Node *CmpDLessNode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  // Check if we can change this to a CmpF and remove a ConvD2F operation.
  return CmpD_Ideal_helper( phase, this );
}

//------------------------------Identity---------------------------------------
Node *CmpDLessNode::Identity( PhaseTransform *phase ) {
  return this;
}

// Simplify an CmpFless (compare 2 floats ) node, based on local information.
// If both inputs are constants, compare them.  
// currently not implemented
const Type *CmpDLessNode::sub( const Type *t1, const Type *t2 ) const {
  return TypeInt::CC;		// else use worst case results
}

//------------------------------Value------------------------------------------
// A subtract node differences it's two inputs.  
const Type *CmpDLessNode::Value( PhaseTransform *phase ) const {
  // because of NaN handling, must do complete coparision
  if( phase->eqv( in(1), in(2) ) ) return TypeInt::CC;
  return SubNode::Value(phase);
}

//=============================================================================
//------------------------------Ideal------------------------------------------
Node *CmpDGreaterNode::Ideal(PhaseGVN *phase,PhaseDefUse *du){
  // Check if we can change this to a CmpF and remove a ConvD2F operation.
  return CmpD_Ideal_helper( phase, this );
}

//------------------------------Identity---------------------------------------
Node *CmpDGreaterNode::Identity( PhaseTransform *phase ) {
  return this;
}

// Simplify an CmpFless (compare 2 floats ) node, based on local information.
// If both inputs are constants, compare them.  
// currently not implemented
const Type *CmpDGreaterNode::sub( const Type *t1, const Type *t2 ) const {
  return TypeInt::CC;		// else use worst case results
}

//------------------------------Value------------------------------------------
// A subtract node differences it's two inputs.  
const Type *CmpDGreaterNode::Value( PhaseTransform *phase ) const {
  // because of NaN handling, must do complete coparision
  if( phase->eqv( in(1), in(2) ) ) return TypeInt::CC;
  return SubNode::Value(phase);
}

//=============================================================================
Node *CmpLNode::Identity( PhaseTransform *phase ) {
  return this;
}

Node *CmpLNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  // Comparing a SUB against a zero is equal to comparing the
  // arguments directly.
  /* This optimization breaks MediaTracker.
     It attempts to change A-B < 0 into A < B.
     Alas, this fails if A is most-negative and B is positive.
     Then A-B wraps and becomes positive, so (A-B)<0 reports false,
     while A<B reports true. 
  if( _in[1]->Opcode() == Op_SubL &&
      phase->type( _in[2] ) == TypeLong::ZERO ) 
    return new (3) CmpLNode(_in[1]->in(1),_in[1]->in(2));
  else if( _in[2]->Opcode() == Op_SubL &&
	   phase->type( _in[1] ) == TypeLong::ZERO ) 
    return new (3) CmpLNode(_in[2]->in(2),_in[2]->in(1));
    */
  return NULL;			// No change
}

// Simplify a CmpL (compare 2 longs ) node, based on local information.
// If both inputs are constants, compare them.  
const Type *CmpLNode::sub( const Type *t1, const Type *t2 ) const {
  const TypeLong *r0 = t1->is_long(); // Handy access
  const TypeLong *r1 = t2->is_long();

  if( r0->_hi < r1->_lo ) 	// Range is always low?
    return TypeInt::MINUS_1;
  else if( r0->_lo > r1->_hi )	// Range is always high?
    return TypeInt::ONE;

  else if( r0->is_con() && r1->is_con() ) { // comparing constants?
    assert(r0->get_con() == r1->get_con(), "must be equal");
    return TypeInt::ZERO;	// Equal results.
  } else if( r0->_hi == r1->_lo ) // Range is never high?
    return TypeInt::make(-1,0);
  else if( r0->_lo == r1->_hi )	// Range is never low?
    return TypeInt::BOOL;
  return TypeInt::CC;		// else use worst case results
}

//------------------------------Value------------------------------------------
// A subtract node differences it's two inputs.  
const Type *CmpLNode::Value( PhaseTransform *phase ) const {
  if( phase->eqv( in(1), in(2) ) ) return TypeInt::ZERO;
  return SubNode::Value(phase);
}


//=============================================================================
//------------------------------cc2logical-------------------------------------
// Convert a condition code type to a logical type
const Type *BoolTest::cc2logical( const Type *CC ) const {
  if( CC == Type::TOP ) return Type::TOP;
  if( CC->_base != Type::Int ) return TypeInt::BOOL; // Bottom or worse
  const TypeInt *ti = CC->is_int();
  if( ti->is_con() ) {		// Only 1 kind of condition codes set?
    // Match low order 2 bits
    int tmp = ((ti->get_con()&3) == (_test&3)) ? 1 : 0; 
    if( _test & 4 ) tmp = 1-tmp;     // Optionally complement result
    return TypeInt::make(tmp);	     // Boolean result
  }
 
  if( CC == TypeInt::make(0,1) ) {
    if( _test == ge ) return TypeInt::ONE;
    if( _test == lt ) return TypeInt::ZERO;
  }
  if( CC == TypeInt::make(-1,0) ) {
    if( _test == le ) return TypeInt::ONE;
    if( _test == gt ) return TypeInt::ZERO;
  }

  return TypeInt::BOOL;
}

//------------------------------dump_spec-------------------------------------
// Print special per-node info
#ifndef PRODUCT
void BoolTest::dump() const {
  const char *msg[] = {"eq","gt","??","lt","ne","le","??","ge"};
  tty->print(msg[_test]);
}
#endif

//=============================================================================
uint BoolNode::hash() const { return Node::hash()+_test._test; } 
uint BoolNode::size_of() const { return sizeof(BoolNode); }

//------------------------------operator==-------------------------------------
uint BoolNode::cmp( const Node &n ) const {
  const BoolNode *b = (const BoolNode *)&n; // Cast up
  return (_test._test == b->_test._test);
}

//------------------------------clone_cmp--------------------------------------
// Clone a compare/bool tree
static Node *clone_cmp( Node *cmp, Node *cmp1, Node *cmp2, PhaseGVN *gvn, BoolTest::mask test ) {
  Node *ncmp = cmp->clone();
  ncmp->set_req(1,cmp1);
  ncmp->set_req(2,cmp2);
  ncmp = gvn->transform( ncmp );
  return new (2) BoolNode( ncmp, test );
}

//------------------------------Ideal------------------------------------------
Node *BoolNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  // Change "bool tst (cmp con x)" into "bool ~tst (cmp x con)".
  // This moves the constant to the right.  Helps value-numbering.
  Node *cmp = _in[1];
  if( !cmp->is_Sub() ) return NULL;
  Node *cmp1 = cmp->in(1);
  Node *cmp2 = cmp->in(2);
  if( !cmp1 ) return NULL;

  // Constant on left?
  Node *con = cmp1;
  uint op2 = cmp2->Opcode();
  // Move constants to the right of compare's to canonicalize.
  // Do not muck with Opaque1 nodes, as this indicates a loop
  // guard that cannot change shape.
  if( con->is_Con() && !cmp2->is_Con() && op2 != Op_Opaque1 ) {
    // Ok, commute the constant to the right of the cmp node.
    // Clone the Node, getting a new Node of the same class
    cmp = cmp->clone();
    // Swap inputs to the clone
    cmp->set_req(1,cmp2);
    cmp->set_req(2,con);
    cmp = phase->transform( cmp );
    return new (2) BoolNode( cmp, _test.commute() );
  }
  // Change "bool eq/ne (cmp (xor X 1) 0)" into "bool ne/eq (cmp X 0)".
  // The XOR-1 is an idiom used to flip the sense of a bool.  We flip the
  // test instead.
  int cmp1_op = cmp1->Opcode();
  const Type *cmp2_type = phase->type(cmp2);
  Node *j_xor = cmp1;
  if( cmp2_type == TypeInt::ZERO &&
      cmp1_op == Op_XorI &&
      j_xor->in(1) != j_xor &&          // An xor of itself is dead
      phase->type( j_xor->in(2) ) == TypeInt::ONE &&
      (_test._test == BoolTest::eq ||
       _test._test == BoolTest::ne) ) {
    Node *ncmp = phase->transform(new (3) CmpINode(j_xor->in(1),cmp2));
    return new (2) BoolNode( ncmp, _test.negate() );
  }
  
  // Change "bool eq/ne (cmp (Conv2B X) 0)" into "bool eq/ne (cmp X 0)".
  // This is a standard idiom for branching on a boolean value.
  Node *c2b = cmp1;
  if( cmp2_type == TypeInt::ZERO &&
      cmp1_op == Op_Conv2B &&
      (_test._test == BoolTest::eq ||
       _test._test == BoolTest::ne) ) {
    Node *ncmp = phase->transform(phase->type(c2b->in(1))->isa_int()
       ? (Node*)new (3) CmpINode(c2b->in(1),cmp2)
       : (Node*)new (3) CmpPNode(c2b->in(1),phase->makecon(TypePtr::NULL_PTR))
    );
    return new (2) BoolNode( ncmp, _test._test );
  }


  //  The transformation below is not valid for either signed or unsigned 
  //  comparisons due to wraparound concerns at MAX_VALUE and MIN_VALUE.  
  //  This transformation can be resurrected when we are able to  
  //  make inferences about the range of values being subtracted from
  //  (or added to) relative to the wraparound point.
  //
  //    // Remove +/-1's if possible.  
  //    // "X <= Y-1" becomes "X <  Y"
  //    // "X+1 <= Y" becomes "X <  Y"
  //    // "X <  Y+1" becomes "X <= Y"
  //    // "X-1 <  Y" becomes "X <= Y"
  //    // Do not this to compares off of the counted-loop-end.  These guys are
  //    // checking the trip counter and they want to use the post-incremented 
  //    // counter.  If they use the PRE-incremented counter, then the counter has
  //    // to be incremented in a private block on a loop backedge.
  //    if( du && du->cnt(this) && du->out(this)[0]->Opcode() == Op_CountedLoopEnd )
  //      return NULL;
  //  #ifndef PRODUCT
  //    // Do not do this in a wash GVN pass during verification.
  //    // Gets triggered by too many simple optimizations to be bothered with
  //    // re-trying it again and again.
  //    if( !phase->allow_progress() ) return NULL;
  //  #endif
  //    // Not valid for unsigned compare because of corner cases in involving zero.
  //    // For example, replacing "X-1 <u Y" with "X <=u Y" fails to throw an
  //    // exception in case X is 0 (because 0-1 turns into 4billion unsigned but
  //    // "0 <=u Y" is always true).
  //    if( cmp->Opcode() == Op_CmpU ) return NULL;
  //    int cmp2_op = cmp2->Opcode();
  //    if( _test._test == BoolTest::le ) {
  //      if( cmp1_op == Op_AddI &&
  //          phase->type( cmp1->in(2) ) == TypeInt::ONE ) 
  //        return clone_cmp( cmp, cmp1->in(1), cmp2, phase, BoolTest::lt );
  //      else if( cmp2_op == Op_AddI &&
  //  	     phase->type( cmp2->in(2) ) == TypeInt::MINUS_1 )
  //        return clone_cmp( cmp, cmp1, cmp2->in(1), phase, BoolTest::lt );
  //    } else if( _test._test == BoolTest::lt ) {
  //      if( cmp1_op == Op_AddI &&
  //          phase->type( cmp1->in(2) ) == TypeInt::MINUS_1 )
  //        return clone_cmp( cmp, cmp1->in(1), cmp2, phase, BoolTest::le );
  //      else if( cmp2_op == Op_AddI &&
  //  	     phase->type( cmp2->in(2) ) == TypeInt::ONE )
  //        return clone_cmp( cmp, cmp1, cmp2->in(1), phase, BoolTest::le );
  //    }
    
  return NULL;
}

//------------------------------Value------------------------------------------
// Simplify a Bool (convert condition codes to boolean (1 or 0)) node,
// based on local information.	 If the input is constant, do it.
const Type *BoolNode::Value( PhaseTransform *phase ) const {
  return _test.cc2logical( phase->type( _in[1] ) );
}

//------------------------------dump_spec--------------------------------------
// Dump special per-node info
#ifndef PRODUCT
void BoolNode::dump_spec() const {
  tty->print("[");
  _test.dump();
  tty->print("]");
}
#endif


//=============================================================================
//------------------------------AbsNode----------------------------------------
Node *AbsNode::is_absolute( PhaseGVN *phase, Node *phi_root ) {

  int chs = 0;			// No need to change sign of result
  int abs = 1;			// Assume ABS is being performed
  int typ = 0;			// Assume integer ABS

  // ABS ends with the merge of 2 control flow paths.  Find the merge point.
  Node *region = phi_root->in(0);
  if( region->req() != 3 ) return NULL;
  if( phi_root->req() != 3 ) return NULL;

  // Next up is the true/false control bits
  Node *iff = region->in(1);
  if( !iff ) return NULL;
  Node *ift = region->in(2);
  if( !ift ) return NULL;
  if( iff->Opcode() == Op_IfFalse &&
      ift->Opcode() == Op_IfTrue ) {
  } else if( iff->Opcode() == Op_IfTrue &&
	     ift->Opcode() == Op_IfFalse ) {
    chs = 1-chs;		// Test is reversed
  } else return NULL;		// Not from same IF
  Node *fif = iff->in(0);
  if( ift->in(0) != fif ) return NULL;	// Not from same IF

  // Roll up the predicate chain; get the Bool, CmpI, CmpX
  BoolNode *bol = fif->in(1)->is_Bool();
  if (bol == NULL) return NULL;   // test is dead (constant true or false)
  switch( bol->_test._test ) {
  case BoolTest::lt: 			   break;
  case BoolTest::le:                       break;
  case BoolTest::gt: 	      chs = 1-chs; break;
  case BoolTest::ge:          chs = 1-chs; break;
  case BoolTest::eq: abs = 0;              break; // Not an ABS function
  case BoolTest::ne: abs = 0; chs = 1-chs; break; // Not an ABS function
  }
  // Integer test is next
  Node *cmp = bol->in(1);
  if( cmp->Opcode() != Op_CmpI ) return NULL;

  // For floating point, the CmpI is of a CmpXxxx vs zero
  if( phase->type(cmp->in(2)) != TypeInt::ZERO) return NULL;
  Node *fcmp = cmp->in(1);
  const Type *tzero = NULL;
  switch( fcmp->Opcode() ) {
  case Op_CmpFGreater:
  case Op_CmpFLess:    tzero = TypeF::ZERO; typ = 1; break; // Float ABS
  case Op_CmpDGreater:
  case Op_CmpDLess:    tzero = TypeD::ZERO; typ = 2; break; // Double ABS
  default: return NULL;
  }

  // Left of float is value being ABS'd, right is a zero
  Node *x = fcmp->in(1);
  if( phase->type(fcmp->in(2)) != tzero ) return NULL;

  // Next get the 2 pieces being selected, one is the original value
  // and the other is the negated value.
  Node *neg_x = phi_root->in(2);
  if( phi_root->in(1) == x ) {
  } else if( phi_root->in(2) == x ) {
    chs = 1-chs;
    neg_x = phi_root->in(1);
  }
  // Check negated value for really negating
  if( neg_x->Opcode() != Op_SubD &&
      neg_x->Opcode() != Op_SubF )
    return NULL;
  if( phase->type(neg_x->in(1)) != tzero ) return NULL;
  if( neg_x->in(2) != x ) return NULL;

  // Yeah-ha!  A Hit!  Now emit either: abs or nothing, then
  // a chs or nothing, of type float or double

  if( abs ) {			// Doing ABS, or doing ID function
    if( typ == 1 ) x = new (2) AbsFNode(x);
    else           x = new (2) AbsDNode(x);
    if( chs )      x = phase->transform(x); // transform all "new" interior nodes
  }
  if( chs ) {			// Simply negating result?
    if( typ == 1 ) x = new (3) SubFNode(neg_x->in(1),x);
    else           x = new (3) SubDNode(neg_x->in(1),x);
  }

  return x;
}

//=============================================================================
Node *NegFNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  if( in(1)->Opcode() == Op_SubF )
    return new (3) SubFNode( in(1)->in(2), in(1)->in(1) );
  return NULL;
}

Node *NegDNode::Ideal( PhaseGVN *phase, PhaseDefUse *du ) {
  if( in(1)->Opcode() == Op_SubD )
    return new (3) SubDNode( in(1)->in(2), in(1)->in(1) );
  return NULL;
}


//=============================================================================
//------------------------------Value------------------------------------------
// A subtract node differences it's two inputs.  
const Type *SqrtDNode::Value( PhaseTransform *phase ) const {
  const Type *t1 = phase->type( in(1) );
  if( t1 == Type::TOP ) return Type::TOP;
  if( t1->_base != Type::DoubleCon ) return Type::DOUBLE;
  double d = t1->getd();
  if( d < 0.0 ) return Type::DOUBLE;
  return TypeD::make( sqrt( d ) );
}

