/*
 * @(#)type.hpp	1.103 01/02/09 14:47:12
 *
 * Copyright 1991-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
// Portions of code courtesy of Clifford Click

// Optimization - Graph Style


// This class defines a Type lattice.  The lattice is used in the constant
// propagation algorithms, and for some type-checking of the iloc code.
// Basic types include RSD's (lower bound, upper bound, stride for integers),
// float & double precision constants, sets of data-labels and code-labels.
// The complete lattice is described below.  Subtypes have no relationship to
// up or down in the lattice; that is entirely determined by the behavior of 
// the MEET/JOIN functions.

class Dict;
class Type;
class   TypeD;
class   TypeF;
class   TypeInt;
class   TypeLong;
class   TypeAry;
class   TypeTuple;
class   TypePtr;
class     TypeRawPtr;
class     TypeJavaPtr;
class       TypeOopPtr;
class       TypeAryPtr;
class       TypeKlassPtr;

//------------------------------Type-------------------------------------------
// Basic Type object, represents a set of primitive Values.
// Types are hash-cons'd into a private class dictionary, so only one of each
// different kind of Type exists.  Types are never modified after creation, so
// all their interesting fields are constant.
class Type {
public:
  enum TYPES {
    Bad=0,                      // Type check
    Control,                    // Control of code (not in lattice)
    Top,                        // Top of the lattice
    Int,                        // Integer range (lo-hi)
    Long,                       // Long integer range (lo-hi)
    Half,                       // Placeholder half of doubleword

    Tuple,                      // Method signature or object layout
    Array,                      // Array types

    AnyPtr,                     // Any old raw, oop or array pointer
    RawPtr,                     // Raw (non-oop) pointers
    OopPtr,                     // Object pointers
    AryPtr,                     // Array pointers
    KlassPtr,                   // Klass pointers

    Function,                   // Function signature
    Abio,                       // Abstract I/O
    Return_Address,             // Subroutine return address
    Memory,                     // Abstract store
    FloatTop,                   // No float value
    FloatCon,                   // Floating point constant
    FloatBot,                   // Any float value
    DoubleTop,                  // No double value
    DoubleCon,                  // Double precision constant
    DoubleBot,                  // Any double value
    Bottom,                     // Bottom of lattice
    lastype                     // Bogus ending type (not in lattice)
  };

  // Signal values for offsets from a base pointer
  enum OFFSET_SIGNALS {
    OffsetTop = -2000000000,    // undefined offset
    OffsetBot = -2000000001     // any possible offset
  };

private:
  // Dictionary of types shared among compilations.
  static Dict* _shared_type_dict;

  static int uhash( const Type *const t );
  // Structural equality check.  Assumes that cmp() has already compared
  // the _base types and thus knows it can cast 't' appropriately.
  virtual int eq( const Type *t ) const;

  // Top-level hash-table of types
  static Dict *typedict();

  // DUAL operation: reflect around lattice centerline.  Used instead of
  // join to ensure my lattice is symmetric up and down.  Dual is computed
  // lazily, on demand, and cached in _dual.
  const Type *_dual;            // Cached dual value
  // Table for efficient dualing of base types
  static const TYPES dual_type[lastype];

protected:
  Type( TYPES t ) : _base(t), _dual(NULL) {} // Simple types
  // ~Type();                   // Use fast deallocation
  const Type *hashcons();       // Hash-cons the type
public:
  void *operator new( size_t x );
  void operator delete( void *ptr );

  // Initialize the type system for a particular compilation.
  static void Initialize(Compile* compile);

  // Initialize the types shared by all compilations.
  static void Initialize_shared(Compile* compile);

  const enum TYPES _base;       // Enum of Types type

  // Create a new hash-consd type
  static const Type *make(enum TYPES);
  // Test for equivalence of types
  static int cmp( const Type *const t1, const Type *const t2 );
  // Test for higher or equal in lattice
  int higher_equal( const Type *t ) const { return !cmp(meet(t),t); }

  // MEET operation; lower in lattice.
  const Type *meet( const Type *t ) const;
  // WIDEN: 'widens' for Ints and other range types
  virtual const Type *widen( const Type *old ) const { return this; }

  // DUAL operation: reflect around lattice centerline.  Used instead of
  // join to ensure my lattice is symmetric up and down.
  const Type *dual() const { return _dual; }

  // Compute meet dependent on base type
  virtual const Type *xmeet( const Type *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.

  // JOIN operation; higher in lattice.  Done by finding the dual of the 
  // meet of the dual of the 2 inputs.
  const Type *join( const Type *t ) const { 
    return dual()->meet(t->dual())->dual(); }

  // Convenience access
  float getf() const;
  double getd() const;

  const TypeInt    *is_int() const;
  const TypeInt    *isa_int() const;             // Returns NULL if not an Int
  const TypeLong   *is_long() const;
  const TypeLong   *isa_long() const;            // Returns NULL if not a Long
  const TypeD      *is_double_constant() const;  // Asserts it is a DoubleCon
  const TypeD      *isa_double_constant() const; // Returns NULL if not a DoubleCon
  const TypeF      *is_float_constant() const;   // Asserts it is a FloatCon
  const TypeF      *isa_float_constant() const;  // Returns NULL if not a FloatCon
  const TypeTuple  *is_tuple() const;            // Collection of fields, NOT a pointer
  const TypeAry    *is_ary() const;              // Array, NOT array pointer
  const TypePtr    *is_ptr() const;              // Asserts it is a ptr type
  const TypePtr    *isa_ptr() const;             // Returns NULL if not ptr type
  const TypeRawPtr *is_rawptr() const;           // NOT Java oop
  const TypeJavaPtr*isa_javaptr() const;         // Returns NULL if not ptr type
  const TypeKlassPtr *isa_klassptr() const; // Returns NULL if not KlassPtr
  const TypeKlassPtr *is_klassptr() const; // assert if not KlassPtr
  const TypeJavaPtr*is_javaptr() const;          // Java-style GC'd pointer
  const TypeOopPtr *isa_oopptr() const;          // Returns NULL if not OopPtr
  const TypeOopPtr *is_oopptr() const;           // Oop
  const TypeAryPtr *isa_aryptr() const;          // Returns NULL if not AryPtr
  const TypeAryPtr *is_aryptr() const;           // Array oop  
  virtual bool      is_finite() const;           // Has a finite value
  virtual bool      is_nan()    const;           // Is not a number (NaN)

  // Special test for register pressure heuristic
  bool is_floatingpoint() const;        // True if Float or Double base type

  // Do you have memory, directly or through a tuple?
  int has_memory( ) const;

  // Are you a pointer type or not?
  int isa_oop_ptr() const;
  
  // TRUE if type is a singleton
  virtual int singleton(void) const;

  // TRUE if type is above the lattice centerline, and is therefore vacuous
  virtual int empty(void) const;

  // Return a hash for this type.  The hash function is public so ConNode
  // (constants) can hash on their constant, which is represented by a Type.
  virtual int hash() const;

  // Map ideal registers (machine types) to ideal types
  static const Type *mreg2type[];

  // Printing, statistics
  static const char * const msg[lastype]; // Printable strings  
#ifndef PRODUCT
  void         dump() const;
  virtual void dump2( Dict &d, uint depth ) const;
  static  void dump_stats();
  static  void verify_lastype();          // Check that arrays match type enum
#endif
  void typerr(const Type *t) const; // Mixing types error

  // Create basic type
  static const Type* get_const_basic_type(BasicType type); 

  // Convenience common pre-built types.
  static const Type *ABIO;
  static const Type *BOTTOM;
  static const Type *CONTROL;
  static const Type *DOUBLE;
  static const Type *DOUBLE_TOP;
  static const Type *FLOAT;
  static const Type *FLOAT_TOP;
  static const Type *HALF;  
  static const Type *MEMORY;
  static const Type *MULTI;
  static const Type *RETURN_ADDRESS;
  static const Type *TOP;

  // Mapping from compiler type to VM BasicType
  BasicType basic_type() const { return _basic_type[_base]; }
  static const BasicType _basic_type[];
};

//------------------------------TypeF------------------------------------------
// Class of Float-Constant Types.
class TypeF : public Type {
  TypeF( float f ) : Type(FloatCon), _f(f) {};
public:
  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing
  virtual int singleton(void) const;    // TRUE if type is a singleton
  virtual int empty(void) const;        // TRUE if type is vacuous
public:
  const float _f;               // Float constant

  static const TypeF *make(float f);

  virtual bool        is_finite() const;  // Has a finite value
  virtual bool        is_nan()    const;  // Is not a number (NaN)

  virtual const Type *xmeet( const Type *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.
  // Convenience common pre-built types.
  static const TypeF *ZERO; // positive zero only
  static const TypeF *ONE;
#ifndef PRODUCT
  virtual void dump2(Dict &d, uint depth) const;
#endif
};

//------------------------------TypeD------------------------------------------
// Class of Double-Constant Types.
class TypeD : public Type {
  TypeD( double d ) : Type(DoubleCon), _d(d) {};
public:
  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing
  virtual int singleton(void) const;    // TRUE if type is a singleton
  virtual int empty(void) const;        // TRUE if type is vacuous
public:
  const double _d;              // Double constant

  static const TypeD *make(double d);

  virtual bool        is_finite() const;  // Has a finite value
  virtual bool        is_nan()    const;  // Is not a number (NaN)

  virtual const Type *xmeet( const Type *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.
  // Convenience common pre-built types.
  static const TypeD *ZERO; // positive zero only
  static const TypeD *ONE;
#ifndef PRODUCT
  virtual void dump2(Dict &d, uint depth) const;
#endif
};

//------------------------------TypeInt----------------------------------------
// Class of integer ranges, the set of integers between a lower bound and an 
// upper bound, inclusive.
class TypeInt : public Type {
  TypeInt( jint lo, jint hi, int w );
public:
  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing
  virtual int singleton(void) const;    // TRUE if type is a singleton
  virtual int empty(void) const;        // TRUE if type is vacuous
public:
  const jint _lo, _hi;          // Lower bound, upper bound
  const short _widen;           // Limit on times we widen this sucker

  static const TypeInt *make(jint lo);
  static const TypeInt *make(jint lo, jint hi);
  static const TypeInt *make(jint lo, jint hi, int w);

  // Check for single integer
  int is_con() const { return _lo==_hi; }
  jint get_con() const { assert( is_con(), "" );  return _lo; }

  virtual bool        is_finite() const;  // Has a finite value

  virtual const Type *xmeet( const Type *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.
  virtual const Type *widen( const Type *t ) const;
  // Convenience common pre-built types.
  static const TypeInt *MINUS_1;
  static const TypeInt *ZERO;
  static const TypeInt *ONE;
  static const TypeInt *BOOL;
  static const TypeInt *CC;
  static const TypeInt *BYTE;
  static const TypeInt *CHAR;
  static const TypeInt *SHORT;
  static const TypeInt *POS;
  static const TypeInt *INT;
#ifndef PRODUCT
  virtual void dump2(Dict &d, uint depth) const;
#endif
};


//------------------------------TypeLong---------------------------------------
// Class of long integer ranges, the set of integers between a lower bound and 
// an upper bound, inclusive.
class TypeLong : public Type {
  TypeLong( jlong lo, jlong hi, int w );
public:
  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing
  virtual int singleton(void) const;    // TRUE if type is a singleton
  virtual int empty(void) const;        // TRUE if type is vacuous
public:
  const jlong _lo, _hi;         // Lower bound, upper bound
  const short _widen;           // Limit on times we widen this sucker

  static const TypeLong *make(jlong lo);
  static const TypeLong *make(jlong lo, jlong hi);
  static const TypeLong *make(jlong lo, jlong hi, int w);

  // Check for single integer
  int is_con() const { return _lo==_hi; }
  jlong get_con() const { assert( is_con(), "" ); return _lo; }

  virtual bool        is_finite() const;  // Has a finite value

  virtual const Type *xmeet( const Type *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.
  virtual const Type *widen( const Type *t ) const;
  // Convenience common pre-built types.
  static const TypeLong *MINUS_1;
  static const TypeLong *ZERO;
  static const TypeLong *ONE;
  static const TypeLong *LONG;
#ifndef PRODUCT
  virtual void dump2( Dict &d, uint ) const;// Specialized per-Type dumping
#endif
};

//------------------------------TypeTuple--------------------------------------
// Class of Tuple Types, essentially type collections for function signatures
// and class layouts.  It happens to also be a fast cache for the HotSpot
// signature types.
class TypeTuple : public Type {
  TypeTuple( uint cnt, const Type **fields ) : Type(Tuple), _cnt(cnt), _fields(fields) { }
public:
  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing
  virtual int singleton(void) const;    // TRUE if type is a singleton
  virtual int empty(void) const;        // TRUE if type is vacuous
public:  
  const uint _cnt;              // Count of fields
  const Type **_fields;         // Array of field types

  static const TypeTuple *make( uint cnt, const Type **fields );
  static const TypeTuple *make_range(ciSignature *sig);
  static const TypeTuple *make_domain(ciInstanceKlass* recv, ciSignature *sig);

  // Subroutine call type with space allocated for argument types
  static const Type **fields( uint arg_cnt );

  virtual const Type *xmeet( const Type *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.
  // Convenience common pre-built types.
  static const TypeTuple *IFBOTH;
  static const TypeTuple *IFFALSE;
  static const TypeTuple *IFTRUE;
  static const TypeTuple *IFNEITHER;
  static const TypeTuple *LOOPBODY;
  static const TypeTuple *MEMBAR;
#ifndef PRODUCT
  virtual void dump2( Dict &d, uint ) const; // Specialized per-Type dumping
#endif
};

//------------------------------TypeAry----------------------------------------
// Class of Array Types
class TypeAry : public Type {
  TypeAry( const Type *elem, const TypeInt *size) : Type(Array),
    _elem(elem), _size(size) {}
public:
  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing
  virtual int singleton(void) const;    // TRUE if type is a singleton
  virtual int empty(void) const;        // TRUE if type is vacuous
public:
  const Type *_elem;            // Element type of array
  const TypeInt *_size;         // Elements in array

  static const TypeAry *make(  const Type *elem, const TypeInt *size);

  virtual const Type *xmeet( const Type *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.
#ifndef PRODUCT
  virtual void dump2( Dict &d, uint ) const; // Specialized per-Type dumping
#endif
};

//------------------------------TypePtr----------------------------------------
// Class of machine Pointer Types: raw data, oops or arrays.
// If the _base enum is AnyPtr, then this refers to all of the above.
// Otherwise the _base will indicate which subset of pointers is affected,
// and the class will be inheriteted from.
class TypePtr : public Type {
public:
  enum PTR { TopPTR, BotPTR, Null, NotNull, AnyNull, Constant, lastPTR };
protected:
  TypePtr( TYPES t, PTR ptr, int offset ) : Type(t), _ptr(ptr), _offset(offset) {}
  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing
  static const PTR ptr_meet[lastPTR][lastPTR];
  static const PTR ptr_dual[lastPTR];
  static const char * const ptr_msg[lastPTR];
  
public:
  const int _offset;            // Offset into oop, with TOP & BOT
  const PTR _ptr;

  static const TypePtr *make( TYPES t, PTR ptr, int offset );

  // Return a 'ptr' version of this type
  virtual const Type *cast_to_ptr_type(PTR ptr) const;

  virtual jint get_con() const;

  virtual const TypePtr *add_offset( int offset ) const;

  virtual int singleton(void) const;    // TRUE if type is a singleton
  virtual int empty(void) const;        // TRUE if type is vacuous
  virtual const Type *xmeet( const Type *t ) const;
  int meet_offset( int offset ) const;
  int dual_offset( ) const;
  virtual const Type *xdual() const;    // Compute dual right now.
  PTR meet_ptr( PTR ) const;    // Meet over the PTR enum
  // Convenience common pre-built types.
  static const TypePtr *NULL_PTR;
  static const TypePtr *NOTNULL;
  static const TypePtr *BOTTOM;
#ifndef PRODUCT
  virtual void dump2( Dict &d, uint depth ) const;
#endif
};

//------------------------------TypeRawPtr-------------------------------------
// Class of raw pointers, pointers to things other than Oops.  Examples
// include the stack pointer, top of heap, card-marking area, handles, etc.
class TypeRawPtr : public TypePtr {
protected:
  TypeRawPtr( PTR ptr, address bits ) : TypePtr(RawPtr,ptr,0), _bits(bits){}
public:
  virtual int eq( const Type *t ) const;
  virtual int hash() const;     // Type specific hashing
  virtual int singleton(void) const;    // TRUE if type is a singleton
  virtual int empty(void) const;        // TRUE if type is vacuous
public:
  address _bits;                // Constant value, if applicable

  static const TypeRawPtr *make( PTR ptr );
  static const TypeRawPtr *make( address bits );

  // Return a 'ptr' version of this type
  virtual const Type *cast_to_ptr_type(PTR ptr) const;

  virtual jint get_con() const;

  virtual const TypePtr *add_offset( int offset ) const;

  virtual const Type *xmeet( const Type *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.
  // Convenience common pre-built types.
  static const TypeRawPtr *BOTTOM;
  static const TypeRawPtr *NOTNULL;
#ifndef PRODUCT
  virtual void dump2( Dict &d, uint depth ) const;
#endif
};

//------------------------------TypeJavaPtr------------------------------------
// Some kind of Java pointer, either oop or array or klass.  Abstract class.
class TypeJavaPtr : public TypePtr {
protected:
  TypeJavaPtr( TYPES t, PTR ptr, ciKlass* k, ciObject* o, int offset ) : TypePtr(t, ptr, offset), _const_oop(o), _klass(k) { }
public:
  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing
  virtual int singleton(void) const;    // TRUE if type is a singleton
  virtual int empty(void) const;        // TRUE if type is vacuous
protected:

  int xadd_offset( int offset ) const;
  // Oop is NULL, unless this is a constant oop.
  ciObject*        _const_oop;   // Constant oop
  // If _klass is NULL, then so is _sig.  This is an unloaded klass.
  ciKlass*      _klass;       // Klass object
public:
  // Creates a type given a klass. Correctly handles multi-dimensional arrays
  static const TypeJavaPtr* make_from_klass(ciKlass* klass);
  static const TypeJavaPtr* make_from_constant(ciObject* o);

  ciObject* const_oop()  const { return _const_oop; }
  virtual ciKlass* klass() const { return _klass;     } 

  virtual jint get_con() const;

  virtual const TypePtr *add_offset( int offset ) const = 0;

  virtual const Type *xmeet( const Type *t ) const = 0;
  virtual const Type *xdual() const = 0;// Compute dual right now.
#ifndef PRODUCT
  virtual void dump2( Dict &d, uint depth ) const;
#endif
};

//------------------------------TypeOopPtr-------------------------------------
// Class of Java object pointers, pointing either to non-array Java instances
// or to a klassOop (including array klasses).
class TypeOopPtr : public TypeJavaPtr {
  TypeOopPtr( PTR ptr, ciKlass* k, ciObject* o, int offset );
  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing

  ciSymbol*  _name;        // class name

 public:
  ciSymbol* name()         const { return _name; }

  bool  is_loaded() const { return _klass->is_loaded(); }

  // Make a pointer to a constant oop.
  static const TypeOopPtr *make(ciObject* o) {
    return make(TypePtr::Constant, o->klass(), o, 0);
  }

  // Make a pointer to a constant oop with offset.
  static const TypeOopPtr *make(ciObject* o, int offset) {
    return make(TypePtr::Constant, o->klass(), o, offset);
  }

  // Make a pointer to some value of type klass.
  static const TypeOopPtr *make(PTR ptr, ciKlass* klass) {
    return make(ptr, klass, NULL, 0);
  }

  // Make a pointer to some value of type klass with offset.
  static const TypeOopPtr *make(PTR ptr, ciKlass* klass, int offset) {
    return make(ptr, klass, NULL, offset);
  }

  // Make a pointer to an oop.
  static const TypeOopPtr *make(PTR ptr, ciKlass* k, ciObject* o, int offset );

  virtual const Type *cast_to_ptr_type(PTR ptr) const;

  virtual const TypePtr *add_offset( int offset ) const;

  virtual const Type *xmeet( const Type *t ) const;
  virtual const TypeOopPtr *xmeet_unloaded( const TypeOopPtr *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.

  // Convenience common pre-built types.
  static const TypeOopPtr *NOTNULL;
  static const TypeOopPtr *BOTTOM;
  static const TypeOopPtr *HEADER;
#ifndef PRODUCT
  virtual void dump2( Dict &d, uint ) const; // Specialized per-Type dumping
#endif
};

//------------------------------TypeAryPtr-------------------------------------
// Class of Java array pointers
class TypeAryPtr : public TypeJavaPtr {
  TypeAryPtr( PTR ptr, ciObject* o, const TypeAry *ary, ciKlass* k, int offset ) : TypeJavaPtr(AryPtr,ptr,k,o,offset), _ary(ary) {};
  virtual int eq( const Type *t ) const;
  virtual int hash() const;     // Type specific hashing
public:
  const TypeAry *_ary;          // Array we point into
  ciKlass* klass() const;

  static const TypeAryPtr *make( PTR ptr, const TypeAry *ary, ciKlass* k, int offset);
  // Constant pointer to array
  static const TypeAryPtr *make( PTR ptr, ciObject* o, const TypeAry *ary, ciKlass* k, int offset);

  // Convenience
  static const TypeAryPtr *make(ciObject* o);

  // Return a 'ptr' version of this type
  virtual const Type *cast_to_ptr_type(PTR ptr) const;

  virtual int empty(void) const;        // TRUE if type is vacuous
  virtual const TypePtr *add_offset( int offset ) const;

  virtual const Type *xmeet( const Type *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.

  // Convenience common pre-built types.
  static const TypeAryPtr *RANGE;
  static const TypeAryPtr *OOPS;
  static const TypeAryPtr *BYTES;
  static const TypeAryPtr *SHORTS;
  static const TypeAryPtr *CHARS;
  static const TypeAryPtr *INTS;
  static const TypeAryPtr *LONGS;
  static const TypeAryPtr *FLOATS;
  static const TypeAryPtr *DOUBLES;
#ifndef PRODUCT
  virtual void dump2( Dict &d, uint ) const; // Specialized per-Type dumping
#endif
};

//------------------------------TypeKlassPtr-----------------------------------
// Class of Java Klass pointers
class TypeKlassPtr : public TypeJavaPtr {
  TypeKlassPtr( PTR ptr, ciKlass* klass, int offset );

  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing

public:
  ciSymbol* name()  const { return _klass->name(); }

  // ptr to klass 'k'
  static const TypeKlassPtr *make( ciKlass* k ) { return make( TypePtr::Constant, k, 0); }
  // ptr to klass 'k' with offset
  static const TypeKlassPtr *make( ciKlass* k, int offset ) { return make( TypePtr::Constant, k, offset); }
  // ptr to klass 'k' or sub-klass
  static const TypeKlassPtr *make( PTR ptr, ciKlass* k, int offset);

  virtual const TypePtr *add_offset( int offset ) const;
  virtual const Type    *xmeet( const Type *t ) const;
  virtual const Type    *xdual() const;      // Compute dual right now.

  // Convenience common pre-built types.
  static const TypeKlassPtr *OBJECT; // Not-null object klass or below
#ifndef PRODUCT
  virtual void dump2( Dict &d, uint ) const; // Specialized per-Type dumping
#endif
};

//------------------------------TypeFunc---------------------------------------
// Class of Array Types
class TypeFunc : public Type {
  TypeFunc( const TypeTuple *domain, const TypeTuple *range ) : Type(Function),  _domain(domain), _range(range) {}
  virtual int eq( const Type *t ) const;
  virtual int hash() const;             // Type specific hashing
  virtual int singleton(void) const;    // TRUE if type is a singleton
  virtual int empty(void) const;        // TRUE if type is vacuous
public:
  // Constants are shared among ADLC and VM
  enum { Control    = AdlcVMDeps::Control, 
         I_O        = AdlcVMDeps::I_O, 
         Memory     = AdlcVMDeps::Memory, 
         FramePtr   = AdlcVMDeps::FramePtr, 
         ReturnAdr  = AdlcVMDeps::ReturnAdr, 
         Parms      = AdlcVMDeps::Parms
  };

  const TypeTuple *_domain;     // Domain of inputs
  const TypeTuple *_range;      // Range of results

  static const TypeFunc *make(ciMethod* method);
  static const TypeFunc *make(ciSignature signature, const Type* extra);
  static const TypeFunc *make(const TypeTuple* domain, const TypeTuple* range);

  virtual const Type *xmeet( const Type *t ) const;
  virtual const Type *xdual() const;    // Compute dual right now.
#ifndef PRODUCT
  virtual void dump2( Dict &d, uint ) const; // Specialized per-Type dumping
  void print_flattened() const; // Print a 'flattened' signature
#endif
  // Convenience common pre-built types.
};

//------------------------------accessors--------------------------------------
inline float Type::getf() const { 
  assert( _base == FloatCon, "Not a FloatCon" ); 
  return ((TypeF*)this)->_f;
}

inline double Type::getd() const { 
  assert( _base == DoubleCon, "Not a DoubleCon" ); 
  return ((TypeD*)this)->_d; 
}

inline const TypeF *Type::is_float_constant() const { 
  assert( _base == FloatCon, "Not a Float" ); 
  return (TypeF*)this; 
}

inline const TypeF *Type::isa_float_constant() const {
  return ( _base == FloatCon ? (TypeF*)this : NULL);
}

inline const TypeD *Type::is_double_constant() const { 
  assert( _base == DoubleCon, "Not a Double" ); 
  return (TypeD*)this; 
}

inline const TypeD *Type::isa_double_constant() const {
  return ( _base == DoubleCon ? (TypeD*)this : NULL);
}

inline const TypeInt *Type::is_int() const { 
  assert( _base == Int, "Not an Int" ); 
  return (TypeInt*)this; 
}

inline const TypeInt *Type::isa_int() const { 
  return ( _base == Int ? (TypeInt*)this : NULL); 
}

inline const TypeLong *Type::is_long() const { 
  assert( _base == Long, "Not a Long" ); 
  return (TypeLong*)this; 
}

inline const TypeLong *Type::isa_long() const { 
  return ( _base == Long ? (TypeLong*)this : NULL);
}

inline const TypeTuple *Type::is_tuple() const { 
  assert( _base == Tuple, "Not a Tuple" ); 
  return (TypeTuple*)this; 
}

inline const TypeAry *Type::is_ary() const { 
  assert( _base == Array , "Not an Array" ); 
  return (TypeAry*)this; 
}

inline const TypePtr *Type::is_ptr() const { 
  assert( _base==OopPtr || _base==AryPtr || _base==KlassPtr || _base == AnyPtr || _base == RawPtr, "Not a pointer");
  return (TypePtr*)this; 
}

inline const TypePtr *Type::isa_ptr() const { 
  return (_base == OopPtr || _base == AryPtr || _base==KlassPtr || _base == AnyPtr || _base == RawPtr )
    ? (TypePtr*)this : NULL;
}

inline const TypeJavaPtr *Type::isa_javaptr() const { 
  return (_base == OopPtr || _base == AryPtr ||  _base==KlassPtr ) 
    ? (TypeJavaPtr*)this : NULL;
}

inline const TypeRawPtr *Type::is_rawptr() const { 
  assert( _base == RawPtr, "Not a raw pointer" ); 
  return (TypeRawPtr*)this; 
}

inline const TypeOopPtr *Type::isa_oopptr() const { 
  return (_base == OopPtr) ? (TypeOopPtr*)this : NULL;
}

inline const TypeOopPtr *Type::is_oopptr() const { 
  assert( _base == OopPtr, "Not an object pointer" ); 
  return (TypeOopPtr*)this; 
}

inline const TypeAryPtr *Type::isa_aryptr() const { 
  return (_base == AryPtr) ? (TypeAryPtr*)this : NULL;
}

inline const TypeAryPtr *Type::is_aryptr() const { 
  assert( _base == AryPtr, "Not an array pointer" ); 
  return (TypeAryPtr*)this; 
}

inline const TypeKlassPtr *Type::isa_klassptr() const {
  return (_base == KlassPtr) ? (TypeKlassPtr*)this : NULL;
}

inline const TypeKlassPtr *Type::is_klassptr() const {
  assert( _base == KlassPtr, "Not a klass pointer" );
  return (TypeKlassPtr*)this;
}

inline bool Type::is_floatingpoint() const {
  if( (_base == FloatCon)  || (_base == FloatBot) ||
      (_base == DoubleCon) || (_base == DoubleBot) )
    return true;
  return false;
}

//-----------------------------------------------------------------------------
// Map dense integer indices to Types.  Uses classic doubling-array trick.
// Abstractly provides an infinite array of Type*'s, initialized to NULL.
// Note that the constructor just zeros things, and since I use Arena 
// allocation I do not need a destructor to reclaim storage.
class Type_Array : public StackObj {
  Arena *_a;                    // Arena to allocate in
  uint   _max;
  const Type **_types;
  void grow( uint i );          // Grow array node to fit
public:
  Type_Array(Arena *a) : _a(a), _max(0), _types(0) {}
  Type_Array(Type_Array *ta) : _a(ta->_a), _max(ta->_max), _types(ta->_types) { }
  const Type *operator[] ( uint i ) const // Lookup, or NULL for not mapped
  { return (i<_max) ? _types[i] : (Type*)NULL; }
  const Type *fast_lookup(uint i) const{assert(i<_max,"oob");return _types[i];}
  // Extend the mapping: index i maps to Type *n.
  void map( uint i, const Type *n ) { if( i>=_max ) grow(i); _types[i] = n; }
  uint Size() const { return _max; }
#ifndef PRODUCT
  void dump() const;
#endif
};

