/*
 * @(#)jvmdi_impl.hpp	1.43 01/03/08 09:42:18
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


//
// Forward Declarations
//

class JvmdiFramePops;
class JvmdiMonitor;
class JvmdiEvent;
class JvmdiBreakpoint;
class JvmdiWalkableMark;

#ifdef ASSERT

//
// JVMDITraceWrapper
//
// Wrapper to trace JVMDI functions
// Used only for debugging
// 

class JvmdiTraceWrapper : public StackObj {
public:
    JvmdiTraceWrapper(const char *str);
    JvmdiTraceWrapper(const char *parm, JvmdiMonitor *monitor);
    JvmdiTraceWrapper(const char *parm, jlong l);
    JvmdiTraceWrapper(const char *parm, const char *str);
    JvmdiTraceWrapper(const char *parm, JavaThread *thr);
    JvmdiTraceWrapper(const char *parm, int event_type, bool enabled);
    JvmdiTraceWrapper(const char *parm, oop obj);
    JvmdiTraceWrapper(const char *parm, methodOop m, jint bci);
    JvmdiTraceWrapper(const char *parm, JvmdiBreakpoint& bp);
    JvmdiTraceWrapper(const char *parm, JvmdiEvent& event);
};

// Macros for using the JvmdiTraceWrapper
#define JVMDI_TRACE_ARG2(X,Y)   if (TraceJVMDICalls) { JvmdiTraceWrapper _ignore(X,Y); }
#define JVMDI_TRACE_ARG3(X,Y,Z) if (TraceJVMDICalls) { JvmdiTraceWrapper _ignore(X,Y,Z); }

#else

  #define JVMDI_TRACE_ARG2(X,Y)
  #define JVMDI_TRACE_ARG3(X,Y,Z)  
  #define JVMDI_TRACE_EVENT(X) 

#endif


//
// class JvmdiEventEnabled
// Used internally by   : JvmdiEventDispatcher, JvmdiThreadState
// Used by JVMDI methods: not directly.
// Note: It's a helper class.
//
// A boolean array indexed by event_type, used as an internal
// data structure to track what JVMDI event types are enabled 
// both globally and on a per thread basis.
//

class JvmdiEventEnabled : public CHeapObj {
private:
  bool _event_enabled[JVMDI_MAX_EVENT_TYPE_VAL+1];

public:
  JvmdiEventEnabled(bool init_val);
  bool is_enabled(jint event_type);
  void set_enabled(jint event_type, bool enabled);
};


//
// class JvmdiEventDispatcher
// Used internally by   : No one, it's a top level class.
// Used by JVMDI methods: SetEventNotificationMode, post_*_event
//
// JVMDI events can be enabled (or disabled) either globally (for all threads)
// or for a single thread. The JvmdiEventDispatcher is a top level
// class that manages state describing what JVMDI event types are enabled both
// locally and globally.
//

class JvmdiEventDispatcher : public AllStatic {
private:
  // global array indexed by (jint) event_type  
  static JvmdiEventEnabled _global_event_enabled;

  static void JvmdiEventDispatcher::reset_thread_location(JavaThread *thread, jint event, bool enabled);

public:
  static void initialize();
  static void destroy();

  // events that can ONLY be enabled/disabled globally (can't toggle on individual threads).
  static inline bool is_global_event(jint event_type) {
    return ((event_type == JVMDI_EVENT_VM_INIT)      ||
	    (event_type == JVMDI_EVENT_VM_DEATH)     ||
	    (event_type == JVMDI_EVENT_THREAD_START) ||
	    (event_type == JVMDI_EVENT_CLASS_UNLOAD)); 
  }

  // returns true if the event is enabled globally
  static bool is_globally_enabled(jint event_type);

  // Use (thread == NULL) to enable/disable an event globally.
  // Use (thread != NULL) to enable/disable an event for a particular thread.
  // thread is ignored for events that can only be specified globally
  static void set_enabled(JavaThread *thread, jint event_type, bool enabled);

  // return true if event is enabled globally or for any thread
  static bool is_enabled(jint event_type);
  
  // returns true if event is enabled for thread.
  // events are enabled for a thread if they are enabled locally OR globally.
  // thread is ignored for events that can only be specified globally
  static bool is_enabled(JavaThread *thread,  jint event_type);

};



// 
// class GrowableCache, GrowableElement
// Used by              : JvmdiBreakpointCache, JvmdiFramePopCache
// Used by JVMDI methods: none directly.
//
// GrowableCache is a permanent CHeap growable array of <GrowableElement *>
//
// In addition, the GrowableCache maintains a NULL terminated cache array of type address
// that's created from the element array using the function: 
//     address GrowableElement::getCacheValue(). 
//
// Whenever the GrowableArray changes size, the cache array gets recomputed into a new C_HEAP allocated
// block of memory. Additionally, every time the cache changes its position in memory, the
//    void (*_listener_fun)(void *this_obj, address* cache) 
// gets called with the cache's new address. This gives the user of the GrowableCache a callback
// to update its pointer to the address cache.
//

class GrowableElement : public CHeapObj {
public:
  virtual address getCacheValue()          =0;
  virtual bool equals(GrowableElement* e)  =0;
  virtual bool lessThan(GrowableElement *e)=0;
  virtual GrowableElement *clone()         =0;
  virtual void oops_do(void f(oop *))      =0;
};

class GrowableCache VALUE_OBJ_CLASS_SPEC {

private:
  // Object pointer passed into cache & listener functions.
  void *_this_obj;

  // Array of elements in the collection
  GrowableArray<GrowableElement *> *_elements;

  // Parallel array of cached values
  address *_cache;

  // Listener for changes to the _cache field.
  // Called whenever the _cache field has it's value changed  
  // (but NOT when cached elements are recomputed).
  void (*_listener_fun)(void *, address*);

  static bool equals(void *, GrowableElement *);

  // recache all elements after size change, notify listener
  void recache();
  
public:  
   GrowableCache();
   ~GrowableCache();

  void initialize(void *this_obj, void listener_fun(void *, address*) );

  // number of elements in the collection
  int length();
  // get the value of the index element in the collection
  GrowableElement* at(int index);
  // find the index of the element, -1 if it doesn't exist
  int find(GrowableElement* e);
  // append a copy of the element to the end of the collection, notify listener
  void append(GrowableElement* e);  
  // insert a copy of the element using lessthan(), notify listener
  void insert(GrowableElement* e);  
  // remove the element at index, notify listener
  void remove (int index);
  // clear out all elements and release all heap space, notify listener
  void clear();
  // apply f to every element and update the cache
  void oops_do(void f(oop*));
};

//
// class JvmdiBreakpointCache
// Used by              : JvmdiBreakpoints
// Used by JVMDI methods: none directly.
// Note   : typesafe wrapper for GrowableCache of JvmdiBreakpoint
//

class JvmdiBreakpointCache : public CHeapObj {

private:
  GrowableCache _cache;
  
public:  
  JvmdiBreakpointCache()  {}
  ~JvmdiBreakpointCache() {}

  void initialize(void *this_obj, void listener_fun(void *, address*) ) {
    _cache.initialize(this_obj,listener_fun);
  }

  int length()                          { return _cache.length(); }
  JvmdiBreakpoint& at(int index)        { return (JvmdiBreakpoint&) *(_cache.at(index)); }
  int find(JvmdiBreakpoint& e)          { return _cache.find((GrowableElement *) &e); }
  void append(JvmdiBreakpoint& e)       { _cache.append((GrowableElement *) &e); }
  void remove (int index)               { _cache.remove(index); }
  void clear()                          { _cache.clear(); }
  void oops_do(void f(oop *))           { _cache.oops_do(f); }
};


//
// class JvmdiObj
// Used by              : JvmdiMonitor
// Used by JVMDI methods: none directly.
//
// Overrides new & delete to use JvmdiInternal::jvmdiMalloc and JvmdiInternal::jvmdiFree
//

class JvmdiObj
{
public:
  void *operator new(size_t size);
  void  operator delete(void *p);
};


// 
// class JvmdiMonitor
// Used by JVMDI methods: All RawMonitor methods (CreateRawMonitor, EnterRawMonitor, etc.)
//
// Wrapper for ObjectMonitor class that saves the Monitor's name
//

class JvmdiMonitor : public JvmdiObj {
private:
  int           _magic;
  char *        _name;
  ObjectMonitor _monitor;

public:
  JvmdiMonitor(const char *name, const int magic);
  ~JvmdiMonitor();
  int            magic()   { return _magic;  }
  ObjectMonitor& monitor() { return _monitor;  }
  void print(outputStream& out)       { out.print(_name); }
};


//
// class JvmdiBreakpoint
// Used by              : JvmdiBreakpoints
// Used by JVMDI methods: SetBreakpoint, ClearBreakpoint, ClearAllBreakpoints
// Note: Extends GrowableElement for use in a GrowableCache
//
// A JvmdiBreakpoint describes a location (class, method, bci) to break at.
//

class JvmdiBreakpoint : public GrowableElement {
private:   
  methodOop 		_method;
  int       		_bci;
  Bytecodes::Code 	_orig_bytecode;

public: 
  JvmdiBreakpoint();
  JvmdiBreakpoint(jclass clazz, jmethodID method, jlocation location);
  bool equals(JvmdiBreakpoint& bp);
  bool lessThan(JvmdiBreakpoint &bp);
  void copy(JvmdiBreakpoint& bp);
  bool is_valid();
  address getBcp();
  void set();
  void clear();
  void print();

  // GrowableElement implementation
  address getCacheValue()         { return getBcp(); }
  bool lessThan(GrowableElement* e) { Unimplemented(); return false; }
  bool equals(GrowableElement* e) { return equals((JvmdiBreakpoint&) *e); }
  void oops_do(void f(oop *))     {   f((oop *) &_method); }
  GrowableElement *clone()        {
    JvmdiBreakpoint *bp = new JvmdiBreakpoint();
    bp->copy(*this);
    return bp;
  }
};

//
// class VM_ChangeBreakpoints
// Used by              : JvmdiBreakpoints
// Used by JVMDI methods: none directly.
// Note: A Helper class.
//
// VM_ChangeBreakpoints implements a VM_Operation for ALL modifications to the JvmdiBreakpoints class.
//

class VM_ChangeBreakpoints : public VM_Operation {
private:
  JvmdiBreakpoints* _breakpoints;
  int               _operation;
  JvmdiBreakpoint*  _bp;
  
public:
  enum { SET_BREAKPOINT=0, CLEAR_BREAKPOINT=1, CLEAR_ALL_BREAKPOINT=2 };

  VM_ChangeBreakpoints(JvmdiBreakpoints* breakpoints, int operation) {
    _breakpoints = breakpoints;
    _bp = NULL;
    _operation = operation;
    assert(breakpoints != NULL, "breakpoints != NULL");
    assert(operation == CLEAR_ALL_BREAKPOINT, "unknown breakpoint operation");
  }
  VM_ChangeBreakpoints(JvmdiBreakpoints* breakpoints, int operation, JvmdiBreakpoint *bp) {
    _breakpoints = breakpoints;
    _bp = bp;
    _operation = operation;
    assert(breakpoints != NULL, "breakpoints != NULL");
    assert(bp != NULL, "bp != NULL");
    assert(operation == SET_BREAKPOINT || operation == CLEAR_BREAKPOINT , "unknown breakpoint operation");
  }    
  void doit();
  const char* name() const                       { return "change breakpoints"; }  
};
 

//
// class JvmdiBreakpoints 
// Used by              : JvmdiCurrentBreakpoints
// Used by JVMDI methods: none directly
// Note: A Helper class
//
// JvmdiBreakpoints is a GrowableCache of JvmdiBreakpoint.
// All changes to the GrowableCache occur at a safepoint using VM_ChangeBreakpoints.
//
// Because _bps is only modified at safepoints, its possible to always use the
// cached byte code pointers from _bps without doing any synchronization (see JvmdiCurrentBreakpoints).
//
// It would be possible to make JvmdiBreakpoints a static class, but I've made it
// CHeap allocated to emphasize its similarity to JvmdiFramePops. 
//

class JvmdiBreakpoints : public CHeapObj {
private:

  JvmdiBreakpointCache _bps;
  
  // These should only be used by VM_ChangeBreakpoints
  // to insure they only occur at safepoints.
  // Todo: add checks for safepoint
  friend class VM_ChangeBreakpoints;
  void set_at_safepoint(JvmdiBreakpoint& bp);
  void clear_at_safepoint(JvmdiBreakpoint& bp);
  void clearall_at_safepoint();

  static void do_element(GrowableElement *e);

public:  
  JvmdiBreakpoints(void listener_fun(void *, address *));
  ~JvmdiBreakpoints(); 

  int length();
  void oops_do(void f(oop *obj));
  void print();

  int  set(JvmdiBreakpoint& bp);
  int  clear(JvmdiBreakpoint& bp);
  void clearall(); 
};


//
// class JvmdiCurrentBreakpoints
// Used by              : JvmdiInternal, InterpreterRuntime::at_safepoint
// Used by JVMDI methods: SetBreakpoint, ClearBreakpoint, ClearAllBreakpoints
//
// A static wrapper class for the JvmdiBreakpoints that provides:
// 1. a fast inlined function to check if a byte code pointer is a breakpoint (is_breakpoint).
// 2. a function for lazily creating the JvmdiBreakpoints class (this is not strictly necessary,
//    but I'm copying the code from JvmdiThreadState which needs to lazily initialize 
//    JvmdiFramePops).
// 3. An oops_do entry point for GC'ing the breakpoint array.
//

class JvmdiCurrentBreakpoints : public AllStatic {

private:

  // Current breakpoints, lazily initialized by get_jvmdi_breakpoints();
  static JvmdiBreakpoints *_jvmdi_breakpoints;

  // NULL terminated cache of byte-code pointers corresponding to current breakpoints.
  // Updated only at safepoints (with listener_fun) when the cache is moved. 
  // It exists only to make is_breakpoint fast.
  static address          *_breakpoint_list;
  static inline void set_breakpoint_list(address *breakpoint_list) { _breakpoint_list = breakpoint_list; }
  static inline address *get_breakpoint_list()                     { return _breakpoint_list; }

  // Listener for the GrowableCache in _jvmdi_breakpoints, updates _breakpoint_list.
  static void listener_fun(void *this_obj, address *cache);

public:
  static void initialize();
  static void destroy();

  // lazily create _jvmdi_breakpoints and _breakpoint_list
  static JvmdiBreakpoints& get_jvmdi_breakpoints();

  // quickly test whether the bcp matches a cached breakpoint in the list
  static inline bool is_breakpoint(address bcp);

  static void oops_do(void f(oop *));
};

// quickly test whether the bcp matches a cached breakpoint in the list
bool JvmdiCurrentBreakpoints::is_breakpoint(address bcp) {
    address *bps = get_breakpoint_list();
    if (bps == NULL) return false;
    for ( ; (*bps) != NULL; bps++) {
      if ((*bps) == bcp) return true;
    }
    return false;
}


//
// class JvmdiFrame
// Used by              : JvmdiWalkableThread.
// Used by JVMDI methods: none directly.
// 
// Wrapper for hotspot frame object (PC & FP & SP)
//
// Usage outside JvmdiWalkableThread is extremely discouraged because of synchronization issues
//

class JvmdiFrame : public CHeapObj {
private:
  int           _magic;
  JavaThread * _thread;
  frame        _frame;

public:
  JvmdiFrame(JavaThread *thread, frame *f, const int magic)   {
    assert(thread != NULL, "thread != NULL");
    assert(f != NULL,      "f != NULL");

    _magic  = magic;
    _thread = thread;
    _frame  = *f;
  }
  ~JvmdiFrame()            { _magic = 0;     }
  JavaThread *get_thread() { return _thread; }
  frame &get_frame()       { return  _frame; }
  frame *get_frame_ptr()   { return &_frame; }
  int magic()              { return _magic;  }

  bool equals(frame *f) {
    assert(f != NULL, "f != NULL");
    return (_frame.fp() == f->fp()) && (_frame.pc() == f->pc());
  }

  // Todo: Add type-checking, always return true for now
  bool get_local(jint n, oop     *o){ *o = _frame.get_local_object(n); return true; }
  bool get_local(jint n, jint    *i){ *i = _frame.get_local_int   (n); return true; }
  bool get_local(jint n, jlong   *l){ *l = _frame.get_local_long  (n); return true; }
  bool get_local(jint n, jfloat  *f){ *f = _frame.get_local_float (n); return true; }
  bool get_local(jint n, jdouble *d){ *d = _frame.get_local_double(n); return true; }

  bool set_local(jint n, oop     o){ _frame.set_local_object(n,o); return true; }
  bool set_local(jint n, jint    i){ _frame.set_local_int   (n,i); return true; }
  bool set_local(jint n, jlong   l){ _frame.set_local_long  (n,l); return true; }
  bool set_local(jint n, jfloat  f){ _frame.set_local_float (n,f); return true; }
  bool set_local(jint n, jdouble d){ _frame.set_local_double(n,d); return true; }
};

//
// class JvmdiWalkableThread
// Used by              : JvmdiWalkableThreads
// Used by JVMDI methods: GetCurrentFrame, GetCallerFrame, GetFrameLocation, NotifyFramePop, GetLocal*, SetLocal*
//
// JVMDI calls cannot usually examine the frames in the stack of a running JavaThread, 
// except under the following two scenarios:
//
// 1. The thread has been suspended by the JVMDI hSuspendThread call.
// 2. The thread has called (and is still inside) the JVMDI event hook function.
// 
// In both of these cases the frames of the thread are examinable and modifiable using JVMDI calls.
// Threads which JVMDI can examine I'll call 'walkable.'
//
// JVMDI refers to a frame by a 32-bit jframeID value and gives the implementor full choice about what
// it represents. jframeID's are valid only as long as the thread they refer to remains walkable.
// Once a thread leaves the walkable state all its jframeID's become stale.
//
// The calls in JVMDI (GetFrameLocation) introduce another wrinkle, by requiring that a jframeID 
// map directly to its thread.
//
// This leads to the following implementation possibilities:
//
// 1. The simplest implementation of jframeID would be the frame's SP. However, this leaves no way to
// detect stale jframeID's or (for hotspot) easily map back to the appropriate thread.
//
// 2. Another simple implementation would use a pointer to a JvmdiFrame object, which could be marked stale,
// but would need to be collected at some point (this is the approach used by classic). Once the JvmdiFrame object
// was collected you would lose the ability to detect a pointer to it as stale. By dereferencing a stale pointer
// you might introduce bugs that show up much later and are difficult to detect.
//
// 3. A more complex alternative (which may not be a good idea?!?) can detect stale jframeID's
// which in the presence of buggy JVMDI user code might help licensee's (and our JBUG people) more. 
//
// 4. A combination of all of the above with some changes, of course. The
// frame's frame pointer value is used as the jframeID. This makes the
// jframeID unique for a particular frame and also allows the same jframeID
// value to be returned for the same, unmodified frame. The JvmdiFrame
// object is used as a wrapper around the frame to provide the mapping
// between frame and thread. It also contains a "magic" value to make it
// easier to detect stale JvmdiFrames.
//
// Currently, we use (2) after trying one attempt at (3) which didn't work for the case of nested
// event_hook calls.
// 

class JvmdiWalkableThread : public CHeapObj {
private:
  JvmdiWalkableThread *_next;
  JavaThread          *_thread;

  // was the walkable thread created by the request of its own thread?
  bool                _active;
  // If SuspendThread() is called, then we need to mark the walkable
  // thread so that it is not cleaned up by the JvmdiWalkableMark
  // destructor. ResumeThread() will clear the _suspend flag if there
  // is still a pending JvmdiWalkableMark out there. Otherwise,
  // ResumeThread() will clean up the walkable thread stuff. If
  // ResumeThread() is never called before the thread dies, then
  // cleanup_thread() will clean up the walkable thread stuff.
  bool                _suspend;

  // JvmdiFrame's that have been for this thread
  GrowableArray<JvmdiFrame *> *_frames;
  
  // verify jvmdi frames for this thread
  bool verify_jvmdi_frame(JvmdiFrame *jvmdi_frame);
  bool verify_jvmdi_frames(); 

  // helper method for lookup
  static bool equalFrame(void* v, JvmdiFrame *jvmdi_frame2);
  // helper method to find JvmdiFrame that contains fp
  static bool equalFp(void* fp, JvmdiFrame *jvmdi_frame);

  // Lookup a jframeID using the specified frame for the current thread.
  // A new JvmdiFrame is created as needed.
  jframeID lookup(frame* f);

  // given frame f, find a Java frame (which could be f), return false is none exists
  bool find_java_frame(frame f, frame *next);

public:
  JvmdiWalkableThread(JavaThread *java_thread, JvmdiWalkableThread *next, bool is_suspend);
  ~JvmdiWalkableThread();

  JvmdiWalkableThread *next()               { return _next; }
  void set_next(JvmdiWalkableThread *next)  { _next = next; }

  JavaThread *get_thread()                    { return _thread; }
  bool        equals(JavaThread *java_thread) { return _thread == java_thread; }

  //  true when the thread was suspended with a pointer to the last Java frame.
  bool has_last_frame()                     { return _thread->has_last_Java_frame(); }

  // walkable thread created by the request of its own thread
  bool is_active()                          { return _active;    }
  // walkable thread created/used by suspend?
  bool is_suspend()                         { return _suspend; }
  void set_suspend(bool value)              { _suspend = value; }
  

  // convert between JvmdiFrame and jframeID (verify frames in debug build)
  JvmdiFrame *from_jframeID(jframeID fid);
  jframeID to_jframeID(JvmdiFrame *jvmdi_frame);
  static jframeID invalid_jframeID();
  static jframeID unknown_jframeID();

  // Methods for accessing and traversing frames
  int nFrames();
  jframeID current_frame();
  jframeID get_caller(jframeID fid);
  bool is_opaque(jframeID fid);
  bool get_frame_location(jframeID fid, methodHandle *method_handle, jint *bci);
};


//
// class JvmdiFramePop
// Used by              : JvmdiFramePops
// Used by JVMDI methods: none directly.
//
// Wrapper class for FramePop, used in the JvmdiFramePops class.
//
// Within this class we treat frame pointers as of type address, 
// rather than as type jint*, because we want to reuse the 
// GrowableElement type.
//

class JvmdiFramePop : public GrowableElement {
 private:  
  address _fp;   
 public:
  JvmdiFramePop(frame* f)      { 
    assert(f != NULL, "f != NULL");
    assert(f->fp() != NULL, "f->fp() != NULL");
    _fp = (address)f->fp(); 
  }

  JvmdiFramePop(address fp)      { 
    assert(fp != NULL, "fp != NULL");
    _fp = fp; 
  }

  ~JvmdiFramePop()                 { _fp = NULL; }

  void copy(JvmdiFramePop &fp)     { _fp = fp._fp; }
  // To do: abstract the machine-specific attribute of frame identification
  address getFp()                  { return _fp; }

  // To do: Make Platform independent version
  bool lessThan(JvmdiFramePop &fp);
  bool equals(JvmdiFramePop &fp);  

  void print();

  // GrowableElement implementation
  address getCacheValue()         { return getFp(); }
  bool lessThan(GrowableElement* e) { return lessThan((JvmdiFramePop&) *e); }
  bool equals(GrowableElement* e) { return equals((JvmdiFramePop&) *e); }
  void oops_do(void f(oop *))     { ShouldNotCallThis(); }
  GrowableElement *clone()        { return new JvmdiFramePop(_fp); }
};


//
// class JvmdiFramePopCache
// Used by              : JvmdiFramePops
// Used by JVMDI methods: none directly.
// Note   : typesafe wrapper for GrowableCache of JvmdiFramePop
//

class JvmdiFramePopCache : public CHeapObj {

private:
  GrowableCache _cache;
  
public:  
  JvmdiFramePopCache()  {}
  ~JvmdiFramePopCache() {}

  void initialize(void *this_obj, void listener_fun(void *, address*) ) {
    _cache.initialize(this_obj,listener_fun);
  }

  int length()                          { return _cache.length(); }
  JvmdiFramePop& at(int index)          { return (JvmdiFramePop&) *(_cache.at(index)); }
  int find(JvmdiFramePop& e)            { return _cache.find((GrowableElement *) &e); }
  void insert(JvmdiFramePop& e)         { _cache.insert((GrowableElement *) &e); }
  void remove (int index)               { _cache.remove(index); }
  void clear()                          { _cache.clear(); }
};


//
// class VM_ChangeFramePops
// Used by              : JvmdiFramePops
// Used by JVMDI methods: none directly.
// Note: A Helper class.
//
// VM_ChangeFramePops implements a VM_Operation for adding a FramePop to the JvmdiFramePops class.
//

class VM_ChangeFramePops : public VM_Operation {
 private:
  JvmdiFramePops* _frame_pops;
  JvmdiFramePop*  _fp;

 public:
  VM_ChangeFramePops(JvmdiFramePops* frame_pops, JvmdiFramePop *fp) {
    _frame_pops = frame_pops;
    _fp = fp;
    assert(frame_pops != NULL, "frame_pops != NULL");
    assert(fp != NULL,         "fp != NULL");
  }
  void doit();
  const char* name() const                       { return "change frame pops"; }  
};

//
// class JvmdiFramePops
// Used by              : JvmdiThreadState
// Used by JVMDI methods: none directly.
//
// A collection of JvmdiFramePop.
// It records what frames on a threads stack should post frame_pop events when they're exited.
//

class JvmdiFramePops : public CHeapObj {
 private: 
  JvmdiFramePopCache  _fps;
  JvmdiThreadState   *_threadstate;

  // should only be used by VM_ChangeFramePops
  // to insure they only occur at safepoints.
  // Todo: add checks for safepoint
  friend class VM_ChangeFramePops;
  void set_at_safepoint(JvmdiFramePop& fp);
  
 public:
  JvmdiFramePops(JvmdiThreadState *threadstate, void listener_fun(void *,address *));
  ~JvmdiFramePops();

  // clears the stored framepop.
  // should only be called by thread that owns the frame.
  void clear(address fp);

  // clears all framepops lessThan(fp)
  // should only be called by thread that owns the frame.
  void clearTo(JvmdiFramePop& fp);

  int length() { return _fps.length(); }
  void set(JvmdiFramePop& fp);
  void print();
};


//
// class JvmdiWalkableThreads
// Used by              : JvmdiInternal
// Used by JVMDI methods: GetCurrentFrame, GetCallerFrame, GetFrameLocation, NotifyFramePop, GetLocal*, SetLocal*
//
// A collection of all walkable threads.
// JvmdiWalkableThreads provides MT-safe methods for creating, destroying and examining
// JvmdiWalkableThread's and JvmdiFrames.
// 
// All attempts by JVMDI to suspend and resume threads must go through the 
// JvmdiWalkableThreads interface, allowing the class to precisely determine 
// which jframeID's are valid (correspond to an actual suspended frame) and which are stale.
//
// Threads are suspended with JVM_Suspend, insuring that the thread gets stopped at a safepoint
//
// Threads suspended by JVMDI get added to JvmdiWalkableThreads.
// Threads resumed by JVMDI get removed from JvmdiWalkableThreads.
// add/remove_current adds or removes the current thread from JvmdiWalkableThreads.
// 
// All public methods are synchronized by JVMDI_frames_lock to guarantee: 
//  1. the set of valid jframeID's remains constant during the method's execution.
//  2. the set of JVMDI suspended threads remains constant during the method's execution.
//
// methods return true when successful and false when given an invalid jframeID 
// or an unwalkable JavaThread.
//
class JvmdiWalkableThreads : public AllStatic {
  friend class JvmdiWalkableThread;

 private:
  static JvmdiWalkableThread  *_first;

  // list of walkable threads
  static JvmdiWalkableThread *first()               { return _first; }
  static void set_first(JvmdiWalkableThread *first) { _first = first; }

  // get a JvmdiFrame from a jframeID, checking with the WalkableThread
  // to make sure it's valid
  static JvmdiFrame *get_frame(jframeID fid);

  // add/remove a thread from the walkable list
  static bool add_thread(JavaThread *java_thread, bool is_suspend);
  static bool remove_thread(JavaThread *java_thread);

 public:
  // get walkable thread from jframeID or JavaThread
  static JvmdiWalkableThread *get_thread(JavaThread *java_thread);
  static JvmdiWalkableThread *get_thread(jframeID fid);

  // add the current thread to the list of walkables
  static void add_current();
  // remove the current thread from the list of walkables
  static void remove_current();

  // suspend the thread, taking it to a safepoint and adding it to the list of walkables
  static bool suspend(JavaThread *java_thread);
  // resume the thread, removing it from the list of walkables
  static bool resume(JavaThread *java_thread);
  // is a thread suspended?
  static bool suspended(JavaThread *java_thread);

  // return the jframeID of the current Java frame in the thread.
  static jframeID current_frame(JavaThread *java_thread);
  // load fid into the current walkable thread's frame cache
  static bool load_frame(jframeID fid);
  // return the number of frames in the thread.
  static int          nFrames(JavaThread *java_thread);

  static bool is_opaque(jframeID fid);

  // get the parent's jframeID
  static bool get_caller(jframeID child, jframeID *parentID);
  // get the current klass, method and location for a jframeID
  static bool get_frame_location(jframeID frame, jclass *klass, jmethodID *method, jlocation *location);

  // add a frame pop notification to the thread for the jframeID
  static bool notify_frame_pop(jframeID fid);

  // get a local value from frame slot n
  static bool get_local(jframeID fid, jint n, Handle *object_handle);
  static bool get_local(jframeID fid, jint n, jint    *i);
  static bool get_local(jframeID fid, jint n, jlong   *l);
  static bool get_local(jframeID fid, jint n, jfloat  *f);
  static bool get_local(jframeID fid, jint n, jdouble *d);

  // set a local value in frame slot n
  static bool set_local(jframeID fid, jint n, Handle object_handle);
  static bool set_local(jframeID fid, jint n, jint    i);
  static bool set_local(jframeID fid, jint n, jlong   l);
  static bool set_local(jframeID fid, jint n, jfloat  f);
  static bool set_local(jframeID fid, jint n, jdouble d);

  static void print();
  static void print_frame_pops(jframeID frame);

};


//
// class JvmdiThreadState
// Used by              : JvmdiInternal
// Used by JVMDI methods: post_*_event
//
// The Jvmdi state for each thread:
// 1. Local table of enabled events.
// 2. Cache of pending frame_pop_events, created by NotifyFramePop
//    and lazily initialized.
// 3: Location of last executed instruction, used to filter out duplicate
//    events due to instruction rewriting.

class JvmdiThreadState : public CHeapObj {
 private:
  JavaThread        *_thread;
  JvmdiEventEnabled *_local_event_enabled;
  bool              _exception_detected;
  bool              _exception_caught;
  JNIid*            _current_method_id;
  int               _current_bci;
  bool		    _breakpoint_posted;
  bool		    _single_stepping_posted;
  bool              _hide_single_stepping;
  int               _hide_level;
  
 public:
  JvmdiThreadState(JavaThread *thread);
  ~JvmdiThreadState();

  // used by JvmdiEventDispatcher  
  bool is_enabled(jint event_type) const;           // is event_type enabled for this thread? 
  void set_enabled(jint event_type, bool enabled);  // enable/disable event_type for this thread   

  inline JavaThread *get_thread()      { return _thread;              }
  inline bool is_exception_detected()  { return _exception_detected;  }
  inline bool is_exception_caught()    { return _exception_caught;  }
  inline void set_exception_detected() { _exception_detected = true; 
                                         _exception_caught = false; }
  inline void set_exception_caught()   { _exception_caught = true;
                                         _exception_detected = false; }
  inline void set_current_location(JNIid* method_id, int bci)   {  _current_method_id = method_id; 
                                                                   _current_bci  = bci; }
  inline void get_current_location(JNIid** method_id, int* bci) { *method_id = _current_method_id; 
                                                                  *bci = _current_bci; }

  inline void set_breakpoint_posted()  { _breakpoint_posted = true; }
  inline void set_single_stepping_posted() {
    _single_stepping_posted = true;
  }
  inline bool breakpoint_posted() { return _breakpoint_posted; }
  inline bool single_stepping_posted() { 
    return _single_stepping_posted;
  }

  inline void clear_hide_single_stepping() {
    if (_hide_level > 0) {
      _hide_level--;
    } else {
      assert(_hide_single_stepping, "hide_single_stepping is out of phase");
      _hide_single_stepping = false;
    }
  }
  inline bool hide_single_stepping() { return _hide_single_stepping; }
  inline void set_hide_single_stepping() {
    if (_hide_single_stepping) {
      _hide_level++;
    } else {
      assert(_hide_level == 0, "hide_level is out of phase");
      _hide_single_stepping = true;
    }
  }

  // If the thread is in the given method at the given 
  // location just return.  Otherwise, reset the current location
  // and reset _breakpoint_posted and _single_stepping_posted.
  // _breakpoint_posted and _single_stepping_posted are only cleared
  // here.
  void compare_and_set_current_location(methodOop method, address location, jint event);

  // Todo: get rid of this!  
 private:
  bool _debuggable;
 public:
  // Should the thread be enumerated by jvmdiInternal::GetAllThreads?
  bool is_debuggable()                 { return _debuggable; }
  // If a thread cannot be suspended (has no valid last_java_frame) then it gets marked !debuggable
  void set_debuggable(bool debuggable) { _debuggable = debuggable; }

 private:
  int _walkable_markdepth;
 public:
  void set_walkable_markdepth(int walkable_markdepth) { _walkable_markdepth = walkable_markdepth; }
  int get_walkable_markdepth() { return _walkable_markdepth; }

 private:
  // Class used to store pending framepops.
  // Also updates the _frame_pop_list (with listener_fun) when necessary.
  // lazily initialized by get_jvmdi_frame_pops();
  JvmdiFramePops *_jvmdi_frame_pops;

  // NULL terminated cache of stack pointers corresponding to pending frame_pops.
  // Updated only at safepoints (with listener_fun) when the cache is moved. 
  // It exists only to make is_frame_pop fast.
  address        *_frame_pop_list;

  inline void set_frame_pop_list(address *frame_pop_list) { _frame_pop_list = frame_pop_list; }
  inline address *get_frame_pop_list()                    { return _frame_pop_list; }
  static void listener_fun(void *this_obj, address *cache);

 public:

  // lazily initialize _jvmdi_frame_pops and _frame_pop_list
  JvmdiFramePops& get_jvmdi_frame_pops();  

  // quickly test whether we should deliver a frame pop event on return from sp
  inline bool is_frame_pop(address fp);

#ifndef PRODUCT
 private:
  GrowableArray<JvmdiMonitor*> *_monitors;
  int _jvmdi_count;

 public:
  void enter_jvmdi_monitor(JvmdiMonitor *jvmdi_monitor) {
    assert(jvmdi_monitor != NULL, "jvmdi_monitor != NULL");
    assert(JavaThread::current() == _thread, "Only this thread can change _monitors");
    get_monitors()->append(jvmdi_monitor);
  }
  void exit_jvmdi_monitor(JvmdiMonitor *jvmdi_monitor) {
    assert(jvmdi_monitor != NULL, "jvmdi_monitor != NULL");
    assert(JavaThread::current() == _thread, "Only this thread can change _monitors");
    assert(get_monitors()->find(jvmdi_monitor) != -1, "Unmatched Monitor Exit");
    get_monitors()->remove(jvmdi_monitor);
  }

  GrowableArray<JvmdiMonitor*> *get_monitors() {
    if (_monitors == NULL) {
      _monitors = new (ResourceObj::C_HEAP) GrowableArray<JvmdiMonitor*>(5,true);
    }
    return _monitors;
  }

  int get_jvmdi_count()           { return _jvmdi_count; }
  void set_jvmdi_count(int count) { _jvmdi_count = count; }
#endif

  void print_jvmdi_monitors() {
#ifndef PRODUCT
    int len = get_monitors()->length();
    for (int i=0; i<len; i++) {
      tty->print("%d: ", i);
      JvmdiMonitor *jvmdi_monitor = get_monitors()->at(i);
      jvmdi_monitor->print(*tty);
      tty->print_cr("");
    }
#endif
  }

};

// It would be nice if the frame pop fp's were stored in stack order, 
// with the first fp corresponding to the topmost frame pop.
// But there seem to be some intervening frames, and things get added 
// more than once (we take care of that).
// It would be nice if the listener on this cache would reverse the list 
// (in place!), since it's actually sorted pessimally for this search.
// But this is fast enough for now.
bool JvmdiThreadState::is_frame_pop(address fp) {
    address *fps = get_frame_pop_list();
    if (fps == NULL) {
      return false;
    }
    for (address* fpsp = fps; *fpsp != NULL; fpsp++) {
      if (*fpsp == fp) {
        return true;
      }
    }
    return false;
}

//
// class DebuggerThread
// Used by: JvmdiInternal
//
// JavaThread used to wrap a thread started by the debugger dll
// using the Jvmdi method RunDebugThread.
//
class DebuggerThread : public JavaThread {

  JVMDI_StartFunction _start_fn;
  void *_start_arg;
  static void start_function_wrapper(JavaThread *thread, TRAPS);

 public:
  DebuggerThread(JVMDI_StartFunction start_fn, void *start_arg);
};


//
// class JvmdiInternal
//
// Internal helper class for jvmdi. Contains:
// 1. Jvmdi interface implementations.
// 2. Helper methods and state.
//

class JvmdiInternal : AllStatic {

  friend class jvmdi;
  friend class JvmdiObj;

 private:
  static bool              _jvmdi_initialized;
  static JVMDI_Interface_1 _jvmdi_interface;
  static JVMDI_EventHook   _event_hook;
  static JavaVM*           _vm;

  // Hooks for Allocate/DeAllocate functions, if they've been set.
  // never use these directly, always call jvmdiMalloc/Free.
  static JVMDI_AllocHook          _ahook;
  static JVMDI_DeallocHook        _dhook;

  // Convienence wrappers for calls to Allocate/DeAllocate
  // use for all memory that could escape outside the interface.
  static void   jvmdiFree(jbyte *mem);

  // wrapper functions for posting events
  static void post_event(JvmdiEvent& e);
  static void post_proxy_event(JvmdiEventProxy& e, JavaThread *jthread);

  // helper methods for creating arrays of global JNI Handles from local Handles
  static jobject *new_jobjectArray(int length, Handle *handles);

  // extra methods so we can have "C" linkage static functions in separate files
  // and yet still get the JVMDI_Interface_1 initialized in a single call.
  static void set_info_hooks(JVMDI_Interface_1 *jvmdi_interface);
  static void set_info2_hooks(JVMDI_Interface_1 *jvmdi_interface);

public:

  // Convienence wrappers for calls to Allocate/DeAllocate
  // use for all memory that could escape outside the interface.
  static jbyte *jvmdiMalloc(jlong size);
  
  static jvmdiError AllocateInternal(jlong size, jbyte** memPtr);

  static jvmdiError DeallocateInternal(jbyte* mem);

  static void  set_allocation_hooks(JVMDI_AllocHook ahook, JVMDI_DeallocHook dhook);

  // helper methods for creating arrays of global JNI Handles from local Handles
  static jthread *new_jthreadArray(int length, Handle *handles);
  static jthreadGroup *new_jthreadGroupArray(int length, Handle *handles);

  // catch and clear exceptions, storing error in result
  static bool catch_exception(int raw_result, TRAPS);

  // return a valid string no matter what state the threadObj is in
  static const char *safe_get_thread_name(oop threadObj);

  // get a field descriptor for the specified class and field
  static bool get_field_descriptor(klassOop k, jfieldID field, fieldDescriptor* fd);
  // convert from JNIHandle to JavaThread *
  static JavaThread *get_JavaThread(jthread jni_thread);  
  // convert from jclass to klassOop
  static klassOop get_klassOop(jclass clazz);
  // convert from jmethodID to methodOop
  static methodOop get_methodOop(jmethodID method, klassOop k);
  // convert to a jni jmethodID from a non-null methodOop
  static jmethodID get_jni_method_non_null(methodOop method);
  // convert to a jni jclass from a non-null klassOop
  static jclass get_jni_class_non_null(klassOop k);

  static jobject get_jni_object(oop o);
  static oop get_Oop(jobject obj);

  // Call to create and destroy internal data structures
  static void jvmdi_init();
  static void jvmdi_init_done();
  static void jvmdi_death();
  
  static void default_event_hook(JNIEnv *env, JVMDI_Event *event);  

  static JVMDI_Interface_1* get_jvmdi_interface()   { return &_jvmdi_interface; }
  static JVMDI_EventHook get_event_hook()           { return _event_hook; }
  static void set_event_hook(JVMDI_EventHook hook)  { _event_hook = hook; }
  static JavaVM* get_vm()                           { return _vm; }
  static void set_vm(JavaVM* vm)                    { _vm = vm; }
  static bool is_jvmdi_initialized()                { return _jvmdi_initialized; }

};

// These aren't part of JvmdiInternal because you don't seem able to declare a
// static function in a class and have it have "C" linkage. Since these pointers
// to these functions ultimately get assigned to a C defined struct and you can't
// cast away the linkage (silently) they must be outside of the class as much as
// we'd like them internal.
//
extern "C" {
 // Functions exported from the VM to the debugger.
  static jvmdiError JNICALL SetEventHook(JVMDI_EventHook hook);
  static jvmdiError JNICALL SetEventNotificationMode(jint mode, jint eventType, jthread jni_thread, ...);
  static jvmdiError JNICALL GetThreadStatus(jthread jni_thread, jint *threadStatusPtr, jint *suspendStatusPtr);
  static jvmdiError JNICALL GetAllThreads(jint *threadsCountPtr, jthread **threadsPtr);
  // Need to avoid collision with SuspendThread and ResumeThread defined
  // within Visual C++ environment
  static jvmdiError JNICALL hSuspendThread(jthread jni_thread);
  static jvmdiError JNICALL hResumeThread(jthread jni_thread);
  static jvmdiError JNICALL StopThread(jthread jni_thread, jobject exception);
  static jvmdiError JNICALL InterruptThread(jthread jni_thread);
  static jvmdiError JNICALL GetThreadInfo(jthread jni_thread, JVMDI_thread_info *infoPtr);
  static jvmdiError JNICALL GetOwnedMonitorInfo(jthread jni_thread, JVMDI_owned_monitor_info *infoPtr);
  static jvmdiError JNICALL GetCurrentContendedMonitor(jthread jni_thread, jobject *monitor);
  static jvmdiError JNICALL RunDebugThread(jthread jni_thread, JVMDI_StartFunction proc, void *arg, int priority);
  static jvmdiError JNICALL GetTopThreadGroups(jint *groupCountPtr, jthreadGroup **groupsPtr);
  static jvmdiError JNICALL GetThreadGroupInfo(jthreadGroup group, JVMDI_thread_group_info *infoPtr);
  static jvmdiError JNICALL GetThreadGroupChildren(jthreadGroup group, jint *threadCountPtr, jthread **threadsPtr, jint *groupCountPtr, jthreadGroup **groupsPtr);
  static jvmdiError JNICALL GetFrameCount(jthread jni_thread, jint *countPtr);
  static jvmdiError JNICALL GetCurrentFrame(jthread jni_thread, jframeID *framePtr);
  static jvmdiError JNICALL GetCallerFrame(jframeID called, jframeID *framePtr);
  static jvmdiError JNICALL GetFrameLocation(jframeID frame, jclass *classPtr, jmethodID *methodPtr, jlocation *locationPtr);
  static jvmdiError JNICALL NotifyFramePop(jframeID frame);
  static jvmdiError JNICALL GetLocalObject(jframeID frame, jint slot, jobject *valuePtr);
  static jvmdiError JNICALL GetLocalInt(jframeID frame, jint slot, jint *valuePtr);
  static jvmdiError JNICALL GetLocalLong(jframeID frame, jint slot, jlong *valuePtr);
  static jvmdiError JNICALL GetLocalFloat(jframeID frame, jint slot, jfloat *valuePtr);
  static jvmdiError JNICALL GetLocalDouble(jframeID frame, jint slot, jdouble *valuePtr);
  static jvmdiError JNICALL SetLocalObject(jframeID frame, jint slot, jobject value);
  static jvmdiError JNICALL SetLocalInt(jframeID frame, jint slot, jint value);
  static jvmdiError JNICALL SetLocalLong(jframeID frame, jint slot, jlong value);
  static jvmdiError JNICALL SetLocalFloat(jframeID frame, jint slot, jfloat value);
  static jvmdiError JNICALL SetLocalDouble(jframeID frame, jint slot, jdouble value);
  static jvmdiError JNICALL CreateRawMonitor(char *name, JVMDI_RawMonitor *monitorPtr);
  static jvmdiError JNICALL DestroyRawMonitor(JVMDI_RawMonitor monitor);
  static jvmdiError JNICALL RawMonitorEnter(JVMDI_RawMonitor monitor);
  static jvmdiError JNICALL RawMonitorExit(JVMDI_RawMonitor monitor);
  static jvmdiError JNICALL RawMonitorWait(JVMDI_RawMonitor monitor, jlong millis);
  static jvmdiError JNICALL RawMonitorNotify(JVMDI_RawMonitor monitor);
  static jvmdiError JNICALL RawMonitorNotifyAll(JVMDI_RawMonitor monitor);
  static jvmdiError JNICALL SetBreakpoint(jclass clazz, jmethodID method, jlocation location);
  static jvmdiError JNICALL ClearBreakpoint(jclass clazz, jmethodID method, jlocation location);
  static jvmdiError JNICALL ClearAllBreakpoints();
  static jvmdiError JNICALL SetFieldAccessWatch(jclass clazz, jfieldID field);
  static jvmdiError JNICALL ClearFieldAccessWatch(jclass clazz, jfieldID field);
  static jvmdiError JNICALL SetFieldModificationWatch(jclass clazz, jfieldID field);
  static jvmdiError JNICALL ClearFieldModificationWatch(jclass clazz, jfieldID field);
  static jvmdiError JNICALL SetAllocationHooks(JVMDI_AllocHook ahook, JVMDI_DeallocHook dhook);
  static jvmdiError JNICALL Allocate(jlong size, jbyte** memPtr);
  static jvmdiError JNICALL Deallocate(jbyte* mem);
  static jvmdiError JNICALL GetClassSignature(jclass clazz, char **sigPtr);
  static jvmdiError JNICALL GetClassStatus(jclass clazz, jint *statusPtr);
  static jvmdiError JNICALL GetSourceFileName(jclass clazz, char **sourceNamePtr);
  static jvmdiError JNICALL GetClassModifiers(jclass clazz, jint *modifiersPtr);
  static jvmdiError JNICALL GetClassMethods(jclass clazz, jint *methodCountPtr, jmethodID **methodsPtr);
  static jvmdiError JNICALL GetClassFields(jclass clazz, jint *fieldCountPtr, jfieldID **fieldsPtr);
  static jvmdiError JNICALL GetImplementedInterfaces(jclass clazz, jint *interfaceCountPtr, jclass **interfacesPtr);
  static jvmdiError JNICALL IsInterface(jclass clazz, jboolean *isInterfacePtr);
  static jvmdiError JNICALL IsArrayClass(jclass clazz, jboolean *isArrayClassPtr);
  static jvmdiError JNICALL GetClassLoader(jclass clazz, jobject *classloaderPtr);
  static jvmdiError JNICALL GetObjectHashCode(jobject object, jint *hashCodePtr);
  static jvmdiError JNICALL GetMonitorInfo(jobject object, JVMDI_monitor_info *infoPtr);
  static jvmdiError JNICALL GetFieldName(jclass clazz, jfieldID field, char **namePtr, char **signaturePtr);
  static jvmdiError JNICALL GetFieldDeclaringClass(jclass clazz, jfieldID field, jclass *declaringClassPtr);
  static jvmdiError JNICALL GetFieldModifiers(jclass clazz, jfieldID field, jint *modifiersPtr);
  static jvmdiError JNICALL IsFieldSynthetic(jclass clazz, jfieldID field, jboolean *isSyntheticPtr);
  static jvmdiError JNICALL GetMethodName(jclass clazz, jmethodID method, char **namePtr, char **signaturePtr);
  static jvmdiError JNICALL GetMethodDeclaringClass(jclass clazz, jmethodID method, jclass *declaringClassPtr);
  static jvmdiError JNICALL GetMethodModifiers(jclass clazz, jmethodID method, jint *modifiersPtr);
  static jvmdiError JNICALL GetMaxStack(jclass clazz, jmethodID method, jint *maxPtr);
  static jvmdiError JNICALL GetMaxLocals(jclass clazz, jmethodID method, jint *maxPtr);
  static jvmdiError JNICALL GetArgumentsSize(jclass clazz, jmethodID method, jint *sizePtr);
  static jvmdiError JNICALL GetLineNumberTable(jclass clazz, jmethodID method, jint *entryCountPtr, JVMDI_line_number_entry **tablePtr);
  static jvmdiError JNICALL GetMethodLocation(jclass clazz, jmethodID method, jlocation *startLocationPtr, jlocation *endLocationPtr);
  static jvmdiError JNICALL GetLocalVariableTable(jclass clazz, jmethodID method, jint *entryCountPtr, JVMDI_local_variable_entry **tablePtr);
  static jvmdiError JNICALL GetExceptionHandlerTable(jclass clazz, jmethodID method, jint *entryCountPtr, JVMDI_exception_handler_entry **tablePtr);
  static jvmdiError JNICALL GetThrownExceptions(jclass clazz, jmethodID method, jint *exceptionCountPtr, jclass **exceptionsPtr);
  static jvmdiError JNICALL GetBytecodes(jclass clazz, jmethodID method, jint *bytecodeCountPtr, jbyte **bytecodesPtr);
  static jvmdiError JNICALL IsMethodNative(jclass clazz, jmethodID method, jboolean *isNativePtr);
  static jvmdiError JNICALL IsMethodSynthetic(jclass clazz, jmethodID method, jboolean *isSyntheticPtr);
  static jvmdiError JNICALL GetLoadedClasses(jint *classCountPtr, jclass **classesPtr);
  static jvmdiError JNICALL GetClassLoaderClasses(jobject initiatingLoader, jint *classesCountPtr, jclass **classesPtr);
  static jvmdiError JNICALL GetVersionNumber(jint *versionPtr);
  static jvmdiError JNICALL GetCapabilities(JVMDI_capabilities *capabilitiesPtr);
}


// Utility macro that checks for NULL pointers:
#define NULL_CHECK(X, Y) if ((X) == NULL) { JVMDI_RETURN(Y); } 

// Variant on the CHECK macro for returning JVMDI_ERROR_INTERNAL
#define  JVMDI_CHECK  THREAD); if (HAS_PENDING_EXCEPTION) { CLEAR_PENDING_EXCEPTION; return JVMDI_ERROR_INTERNAL; }; (0
