/*
 * @(#)classFileParser.cpp	1.156 01/02/09 14:47:31
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

#include "incls/_precompiled.incl"
#include "incls/_classFileParser.cpp.incl"


// We generally try to create the oops directly when parsing, rather than allocating
// temporary data structures and copying the bytes twice. A temporary area is only
// needed when parsing utf8 entries in the constant pool and when parsing line number
// tables.


// Parse error

void classfile_parse_error(const char* msg, TRAPS) {
  THROW_MSG(vmSymbols::java_lang_ClassFormatError(), msg); 
}

void truncated_class_file_error(TRAPS) {
  classfile_parse_error("Truncated class file", CHECK);
}

// Verification of class file properties

void inline guarantee_property(bool b, const char* msg, TRAPS) {
  if (!b) { classfile_parse_error(msg, CHECK); }
}

// We are using the JDK class file format checker, so we do not have to repeat
// the format checks while parsing. We leave the checks in place in debug mode.

#ifdef ASSERT
void inline check_property(bool b, const char* msg, TRAPS) {
  if (!b) { fatal(msg); }
}
#else
#define check_property(b, msg, traps)
#endif


void ClassFileParser::parse_constant_pool_utf8_entry(constantPoolHandle cp, int index, TRAPS) {  
  u2  utf8_length = stream()->get_u2();
  u1* utf8_buffer = stream()->get_u1_buffer(utf8_length);
  if (utf8_buffer != NULL) {
    // Got utf8 string, set stream position forward
    stream()->skip_u1(utf8_length);
    // Optimistically assume that only 1 byte UTF format is used (common case)
    symbolOop result = oopFactory::new_symbol((char*)utf8_buffer, utf8_length, CHECK);
    cp->symbol_at_put(index, result);
  } else {
    truncated_class_file_error(CHECK);
  }
}



void ClassFileParser::parse_constant_pool_integer_entry(constantPoolHandle cp, int index) {
  u4 bytes = stream()->get_u4();
  cp->int_at_put(index, (jint) bytes);
}


void ClassFileParser::parse_constant_pool_float_entry(constantPoolHandle cp, int index) {
  u4 bytes = stream()->get_u4();
  cp->float_at_put(index, *(jfloat*)&bytes);      // %%%%% move cast into separate function
}


void ClassFileParser::parse_constant_pool_long_entry(constantPoolHandle cp, int index) {
  cp->long_at_put(index, stream()->get_u8());
}


void ClassFileParser::parse_constant_pool_double_entry(constantPoolHandle cp, int index) {
  u8 bytes = stream()->get_u8();
  cp->double_at_put(index, *(jdouble*)&bytes);    // %%%%% move cast into separate function
}


void ClassFileParser::parse_constant_pool_class_entry(constantPoolHandle cp, int index) {
  u2 name_index = stream()->get_u2();
  cp->klass_index_at_put(index, name_index);
}


void ClassFileParser::parse_constant_pool_string_entry(constantPoolHandle cp, int index) {
  u2 string_index = stream()->get_u2();
  cp->string_index_at_put(index, string_index);
}


void ClassFileParser::parse_constant_pool_fieldref_entry(constantPoolHandle cp, int index) {
  u2 class_index = stream()->get_u2();
  u2 name_and_type_index = stream()->get_u2();
  cp->field_at_put(index, class_index, name_and_type_index);
}


void ClassFileParser::parse_constant_pool_methodref_entry(constantPoolHandle cp, int index) {
  u2 class_index = stream()->get_u2();
  u2 name_and_type_index = stream()->get_u2();
  cp->method_at_put(index, class_index, name_and_type_index);
}


void ClassFileParser::parse_constant_pool_interfacemethodref_entry(constantPoolHandle cp, int index) {
  u2 class_index = stream()->get_u2();
  u2 name_and_type_index = stream()->get_u2();
  cp->interface_method_at_put(index, class_index, name_and_type_index);
}

void ClassFileParser::parse_constant_pool_nameandtype_entry(constantPoolHandle cp, int index) {
  u2 name_index = stream()->get_u2();
  u2 signature_index = stream()->get_u2();
  cp->name_and_type_at_put(index, name_index, signature_index);
}


void ClassFileParser::parse_constant_pool_entry(constantPoolHandle cp, int* index, TRAPS) {
  HandleMark hm;
  constantTag tag(stream()->get_u1());
  switch (tag.value()) {
    case JVM_CONSTANT_Class :
      parse_constant_pool_class_entry(cp, *index);
      break;
    case JVM_CONSTANT_Fieldref :
      parse_constant_pool_fieldref_entry(cp, *index);
      break;
    case JVM_CONSTANT_Methodref :
      parse_constant_pool_methodref_entry(cp, *index);
      break;
    case JVM_CONSTANT_InterfaceMethodref :
      parse_constant_pool_interfacemethodref_entry(cp, *index);
      break;
    case JVM_CONSTANT_String :
      parse_constant_pool_string_entry(cp, *index);
      break;
    case JVM_CONSTANT_Integer :
      parse_constant_pool_integer_entry(cp, *index);
      break;
    case JVM_CONSTANT_Float :
      parse_constant_pool_float_entry(cp, *index);
      break;
    case JVM_CONSTANT_Long :
      parse_constant_pool_long_entry(cp, *index);
      (*index)++;   // Skip entry following eigth-byte constant, see JVM book p. 98
      break;
    case JVM_CONSTANT_Double :
      parse_constant_pool_double_entry(cp, *index);
      (*index)++;   // Skip entry following eigth-byte constant, see JVM book p. 98
      break;
    case JVM_CONSTANT_NameAndType :
      parse_constant_pool_nameandtype_entry(cp, *index);
      break;
    case JVM_CONSTANT_Utf8 :
      parse_constant_pool_utf8_entry(cp, *index, CHECK);
      break;
    default:
      classfile_parse_error("Unknown constant tag in .class file", CHECK);
      break;
  }
}


constantPoolHandle ClassFileParser::parse_constant_pool(TRAPS) {
  u2 length = stream()->get_u2();

  constantPoolHandle cp = oopFactory::new_constantPool(length, CHECK_(constantPoolOop(NULL)));
  
  #ifndef PRODUCT
  cp->set_partially_loaded();    // Enables heap verify to work on partial constantPoolOops
  #endif

  
  int index = 1; // Declared outside for loops for portability

  // parsing
  for (index = 1; index < length; index++) {        // Index 0 is unused      
    parse_constant_pool_entry(cp, &index, CHECK_(constantPoolOop(NULL)));   // Pass &index so padding can be done after JVM_CONSTANT_Long and JVM_CONSTANT_Double, see JVM book p. 98
  }

  // validate cross references and fixup class and string constants
  for (index = 1; index < length; index++) {          // Index 0 is unused
    switch (cp->tag_at(index).value()) {
      case JVM_CONSTANT_Class :
        ShouldNotReachHere();     // Only JVM_CONSTANT_ClassIndex should be present
        break;
      case JVM_CONSTANT_Fieldref :
        // fall through
      case JVM_CONSTANT_Methodref :
        // fall through
      case JVM_CONSTANT_InterfaceMethodref :
        check_property(cp->tag_at(cp->klass_ref_index_at(index)).is_klass_reference(), 
          "Invalid constant in .class file", CHECK_(constantPoolOop(NULL)));
        check_property(cp->tag_at(cp->name_and_type_ref_index_at(index)).is_name_and_type(), 
          "Invalid constant in .class file", CHECK_(constantPoolOop(NULL)));
        break;
      case JVM_CONSTANT_String :
        ShouldNotReachHere();     // Only JVM_CONSTANT_StringIndex should be present
        break;
      case JVM_CONSTANT_Integer :
        break;
      case JVM_CONSTANT_Float :
        break;
      case JVM_CONSTANT_Long :
      case JVM_CONSTANT_Double :
        index++;
        break;
      case JVM_CONSTANT_NameAndType :
        check_property(cp->tag_at(cp->name_ref_index_at(index)).is_utf8(), 
          "Invalid constant in .class file", CHECK_(constantPoolOop(NULL)));
        check_property(cp->tag_at(cp->signature_ref_index_at(index)).is_utf8(), 
          "Invalid constant in .class file", CHECK_(constantPoolOop(NULL)));
        break;
      case JVM_CONSTANT_Utf8 :
        break;
      case JVM_CONSTANT_UnresolvedClass :
        ShouldNotReachHere();     // Only JVM_CONSTANT_ClassIndex should be present
        break;
      case JVM_CONSTANT_ClassIndex :
        {
          int class_index = cp->klass_index_at(index);
          check_property(cp->tag_at(class_index).is_utf8(), "Invalid constant in .class file", CHECK_(constantPoolOop(NULL)));
          cp->unresolved_klass_at_put(index, cp->symbol_at(class_index));
        }
        break;
      case JVM_CONSTANT_UnresolvedString :
        ShouldNotReachHere();     // Only JVM_CONSTANT_StringIndex should be present
        break;
      case JVM_CONSTANT_StringIndex :
        {
          int string_index = cp->string_index_at(index);
          check_property(cp->tag_at(string_index).is_utf8(), "Invalid constant in .class file", CHECK_(constantPoolOop(NULL)));
          symbolOop sym = cp->symbol_at(string_index);
          cp->unresolved_string_at_put(index, sym);
        }
        break;
      default:
        fatal1("bad constant pool tag value %d", cp->tag_at(index).value());
        ShouldNotReachHere();
        break;
    }
  }

  return cp;
}


objArrayHandle ClassFileParser::parse_interfaces(constantPoolHandle cp, Handle class_loader, Handle protection_domain, TRAPS) {  
  u2 length = stream()->get_u2();
  if (length == 0) {
    return Universe::the_empty_system_obj_array();
  } else {        
    objArrayHandle interfaces = oopFactory::new_system_objArray(length, CHECK_(objArrayOop(NULL)));
    for (int index = 0; index < length; index++) {
      u2 interface_index = stream()->get_u2();
      klassOop interf = SystemDictionary::resolve_or_fail(cp->unresolved_klass_at(interface_index), class_loader, protection_domain, true, CHECK_(objArrayOop(NULL)));
      if (!Klass::cast(interf)->is_interface()) {
        THROW_MSG_(vmSymbols::java_lang_IncompatibleClassChangeError(), "Implementing class", objArrayOop(NULL)); 
      }
      interfaces->obj_at_put(index, interf);
    }
    return interfaces;
  }
}


// Parse attributes for a field. Currently, only the ConstantValue attribute is recognized.
// The returned value is the constant pool index of the initial field value.
// This can be a constant integer, long, float, double or string. 
// 0 means no initial value for the field.

void ClassFileParser::parse_field_attributes(constantPoolHandle cp, u2* constantvalue_index_addr, bool* is_synthetic_addr, TRAPS) {  
  u2 attributes_count = stream()->get_u2();
  u2 constantvalue_index = 0;
  bool is_synthetic = false;
  while (attributes_count--) {
    u2 attribute_name_index = stream()->get_u2();
    u4 attribute_length = stream()->get_u4();
    check_property(cp->tag_at(attribute_name_index).is_utf8(), "Invalid attribute in .class file", CHECK);
    if (cp->symbol_at(attribute_name_index) == vmSymbols::tag_constant_value()) {
      if (constantvalue_index != 0) {
        classfile_parse_error("Invalid attribute in .class file", CHECK);
      }
      check_property(attribute_length == 2, "Invalid attribute in .class file", CHECK);
      constantvalue_index = stream()->get_u2();
    } else if (cp->symbol_at(attribute_name_index) == vmSymbols::tag_synthetic()) {
      if (attribute_length != 0) {
        classfile_parse_error("Invalid synthetic attribute in .class file", CHECK);
      }
      is_synthetic = true;
    } else {
      stream()->skip_u1(attribute_length);  // Skip unknown attributes
    }
  }
  *constantvalue_index_addr = constantvalue_index;
  *is_synthetic_addr = is_synthetic;
  return;
}


// Field allocation types. Used for computing field offsets.

enum FieldAllocationType {
  STATIC_OOP,
  STATIC_SINGLE,
  STATIC_DOUBLE,
  STATIC_ALIGNED_DOUBLE,
  NONSTATIC_OOP,
  NONSTATIC_SINGLE,
  NONSTATIC_DOUBLE,
  NONSTATIC_ALIGNED_DOUBLE
};


typeArrayHandle ClassFileParser::parse_fields(constantPoolHandle cp, int* static_oop_count, int* nonstatic_oop_count, TRAPS) {
  u2 length = stream()->get_u2();   
  // 5-tuples of shorts [access, name index, sig index, initial value index, offset]
  typeArrayHandle fields = oopFactory::new_permanent_shortArray(length*5, CHECK_(typeArrayOop(NULL)));
  int index = 0;
  for (int n = 0; n < length; n++) {
    AccessFlags access_flags;
    access_flags.set_flags(stream()->get_u2() & JVM_RECOGNIZED_FIELD_MODIFIERS);
    u2 name_index = stream()->get_u2();
    check_property(cp->tag_at(name_index).is_utf8(), "Invalid .class file", CHECK_(typeArrayOop(NULL)));
    u2 signature_index = stream()->get_u2();
    check_property(cp->tag_at(signature_index).is_utf8(), "Invalid .class file", CHECK_(typeArrayOop(NULL)));
    u2 constantvalue_index = 0;
    bool is_synthetic = false;
    parse_field_attributes(cp, &constantvalue_index, &is_synthetic, CHECK_(typeArrayOop(NULL)));
    if (is_synthetic) {
      access_flags.set_is_synthetic();
    }
    
    fields->short_at_put(index++, access_flags.as_short());
    fields->short_at_put(index++, name_index);
    fields->short_at_put(index++, signature_index);
    fields->short_at_put(index++, constantvalue_index);	

    // Remember how many oops we encountered and compute allocation type
    BasicType type = cp->basic_type_for_signature_at(signature_index);
    FieldAllocationType atype;
    bool is_static = access_flags.is_static();
    if (type == T_OBJECT || type == T_ARRAY) {
      if (is_static) {
        (*static_oop_count)++;
        atype = STATIC_OOP;
      } else {
        (*nonstatic_oop_count)++;
        atype = NONSTATIC_OOP;
      }
    } else {
      if (type == T_LONG || type == T_DOUBLE) {
        if (Universe::field_type_should_be_aligned(type)) {
          atype = is_static ? STATIC_ALIGNED_DOUBLE : NONSTATIC_ALIGNED_DOUBLE;
        } else {
          atype = is_static ? STATIC_DOUBLE : NONSTATIC_DOUBLE;
        }
      } else {
        atype = is_static ? STATIC_SINGLE : NONSTATIC_SINGLE;
      }
    }
    // The correct offset is computed later (all oop fields will be located together)
    // We temporarily store the allocation type in the offset field
    fields->short_at_put(index++, atype);
  }
  return fields;
}


static void copy_u2_with_conversion(u2* dest, u2* src, int length) {
  while (length-- > 0) {
    *dest++ = Bytes::get_Java_u2((u1*) (src++));
  }
}


// Note: the parse_method below is big and clunky because all parsing of the code and exceptions
// attribute is inlined. This is curbersome to avoid since we inline most of the parts in the
// methodOop to save footprint, so we only know the size of the resulting methodOop when the
// entire method attribute is parsed.
//
// The promoted_flags parameter is used to pass relevant access_flags
// from the method back up to the containing klass. These flag values
// are added to klass's access_flags.

methodHandle ClassFileParser::parse_method(constantPoolHandle cp, AccessFlags *promoted_flags, TRAPS) {
  ResourceMark rm;
  // Parse fixed parts
  int flags = stream()->get_u2();    
  u2 name_index = stream()->get_u2();
  check_property(cp->tag_at(name_index).is_utf8(), "Invalid .class file", CHECK_(methodOop(NULL)));
  u2 signature_index = stream()->get_u2();
  check_property(cp->tag_at(signature_index).is_utf8(), "Invalid .class file", CHECK_(methodOop(NULL)));

  AccessFlags access_flags;  
  if (cp->symbol_at(name_index) == vmSymbols::class_initializer_name()) {
    // We ignore the access flags for a class initializer. (JVM Spec. p. 116)
    flags = JVM_ACC_STATIC;
  }
  access_flags.set_flags(flags & JVM_RECOGNIZED_METHOD_MODIFIERS);
  
  // Default values for code and exceptions attribute elements
  u2 max_stack = 0;
  u2 max_locals = 0;
  u4 code_length = 0;
  u1* code_start = 0;
  u2 exception_table_length = 0;
  typeArrayHandle exception_handlers(THREAD, Universe::the_empty_int_array());
  u2 checked_exceptions_length = 0;
  u2* checked_exceptions_start = NULL;
  int compressed_linenumber_table_size = 0;
  u_char* compressed_linenumber_table = NULL;
  int localvariable_table_length = 0;
  u2* localvariable_table_start = NULL;
  bool parsed_code_attribute = false;
  bool parsed_checked_exceptions_attribute = false;

  // Parse code and exceptions attribute
  u2 method_attributes_count = stream()->get_u2();
  while (method_attributes_count--) {   
    u2 method_attribute_name_index = stream()->get_u2();
    u4 method_attribute_length = stream()->get_u4();
    check_property(cp->tag_at(method_attribute_name_index).is_utf8(), "Invalid attribute in .class file", CHECK_(methodOop(NULL)));

    if (cp->symbol_at(method_attribute_name_index) == vmSymbols::tag_code()) {
      // Parse Code attribute
      if (parsed_code_attribute) {
        classfile_parse_error("Invalid attribute in .class file", CHECK_(methodOop(NULL)));
      }
      parsed_code_attribute = true;
      // Stack and locals size
      max_stack = stream()->get_u2();
      max_locals = stream()->get_u2();
      // Code size and pointer
      code_length = stream()->get_u4();
      code_start = stream()->get_u1_buffer(code_length);
      if (code_start == NULL) truncated_class_file_error(CHECK_(methodOop(NULL)));
      stream()->skip_u1(code_length);
      // Exception handler table
      exception_table_length = stream()->get_u2();
      if (exception_table_length > 0) {
        // 4-tuples of ints [start_pc, end_pc, handler_pc, catch_type index]
        exception_handlers = oopFactory::new_permanent_intArray(exception_table_length*4, CHECK_(methodOop(NULL))); 
      }
      int index = 0;
      for (int i = 0; i < exception_table_length; i++) {
        for (int j = 0; j < 4; j++) {
          exception_handlers->int_at_put(index++, stream()->get_u2());  
        }
      }
      // Parse additional attributes in code attribute
      u2 code_attributes_count = stream()->get_u2();
      while (code_attributes_count--) {
        u2 code_attribute_name_index = stream()->get_u2();
        u4 code_attribute_length = stream()->get_u4();
        check_property(cp->tag_at(code_attribute_name_index).is_utf8(), "Invalid attribute in .class file", CHECK_(methodOop(NULL)));
        if (LoadLineNumberTables && cp->symbol_at(code_attribute_name_index) == vmSymbols::tag_line_number_table()) {
          // Parse and compress line number table
          int linenumber_table_length = stream()->get_u2();
          if (linenumber_table_length > 0) {
            CompressedLineNumberWriteStream c_stream(linenumber_table_length * sizeof(u2) * 2);  // initial_size large enough
            while (linenumber_table_length-- > 0) {
              u2 bci  = stream()->get_u2();
              u2 line = stream()->get_u2();
              c_stream.write_pair(bci, line);
            }
            c_stream.write_terminator();
            compressed_linenumber_table_size = c_stream.position();
            compressed_linenumber_table = c_stream.buffer();
          }
        } else if (LoadLocalVariableTables && cp->symbol_at(code_attribute_name_index) == vmSymbols::tag_local_variable_table()) {
          // Parse line number table
          localvariable_table_length = stream()->get_u2();
          int size = localvariable_table_length * sizeof(LocalVariableTableElement) / sizeof(u2);
          localvariable_table_start = stream()->get_u2_buffer(size);
          if (localvariable_table_start == NULL) truncated_class_file_error(CHECK_(methodOop(NULL)));
          stream()->skip_u2(size);
        } else {
          // Skip unknown attributes
          stream()->skip_u1(code_attribute_length);
        }
      }
    } else if (cp->symbol_at(method_attribute_name_index) == vmSymbols::tag_exceptions()) {
    // Parse Exceptions attribute
      if (parsed_checked_exceptions_attribute) {
        classfile_parse_error("Invalid attribute in .class file", CHECK_(methodOop(NULL)));
      }
      parsed_checked_exceptions_attribute = true;
      checked_exceptions_length = stream()->get_u2();
      int size = checked_exceptions_length * sizeof(CheckedExceptionElement) / sizeof(u2);
      checked_exceptions_start = stream()->get_u2_buffer(size);
      if (checked_exceptions_start == NULL) truncated_class_file_error(CHECK_(methodOop(NULL)));
      stream()->skip_u2(size);
    } else if (cp->symbol_at(method_attribute_name_index) == vmSymbols::tag_synthetic()) {
      if (method_attribute_length != 0) {
        classfile_parse_error("Invalid synthetic attribute in .class file", CHECK_(methodOop(NULL)));
      }
      // Should we check that there hasn't already been a synthetic attribute?
      access_flags.set_is_synthetic();
    } else {
      // Skip unknown attributes
      stream()->skip_u1(method_attribute_length);
    }
  }

  // All sizing information for a methodOop is finally available, now create it
  methodHandle m = oopFactory::new_method(code_length, access_flags, compressed_linenumber_table_size, localvariable_table_length, checked_exceptions_length, CHECK_(methodOop(NULL)));  

  // Fill in information from fixed part (access_flags already set)
  m->set_constants(cp());
  m->set_name_index(name_index);
  m->set_signature_index(signature_index);
  m->compute_size_of_parameters();
  // Fill in code atttribute information
  m->set_max_stack(max_stack);
  m->set_max_locals(max_locals);
  m->set_exception_table(exception_handlers());
  // Copy byte codes
  if (code_length > 0) {
    memcpy(m->code_base(), code_start, code_length);
  }
  // Copy line number table
  if (compressed_linenumber_table_size > 0) {
    memcpy(m->compressed_linenumber_table(), compressed_linenumber_table, compressed_linenumber_table_size);
  }
  // Copy checked exceptions
  if (checked_exceptions_length > 0) {
    int size = checked_exceptions_length * sizeof(CheckedExceptionElement) / sizeof(u2);
    copy_u2_with_conversion((u2*) m->checked_exceptions_start(), checked_exceptions_start, size);
  }
  // Copy local variable table
  if (localvariable_table_length > 0) {
    int size = localvariable_table_length * sizeof(LocalVariableTableElement) / sizeof(u2);
    copy_u2_with_conversion((u2*) m->localvariable_table_start(), localvariable_table_start, size);
    promoted_flags->set_has_localvariable_table();
  }
  return m;
}


extern "C" {
  static int method_compare(methodOop* a, methodOop* b) {
    return (*a)->name()->fast_compare((*b)->name());
  }

  typedef int (*compareFn)(const void*, const void*);
}


// The promoted_flags parameter is used to pass relevant access_flags
// from the methods back up to the containing klass. These flag values
// are added to klass's access_flags.

objArrayHandle ClassFileParser::parse_methods(constantPoolHandle cp, AccessFlags* promoted_flags, TRAPS) {
  u2 length = stream()->get_u2();
    if (length == 0) {
    return Universe::the_empty_system_obj_array();
  } else {
    objArrayHandle methods = oopFactory::new_system_objArray(length, CHECK_(objArrayOop(NULL)));  
    for (int index = 0; index < length; index++) {
      HandleMark hm;
      methodHandle method = parse_method(cp, promoted_flags, CHECK_(objArrayOop(NULL)));
      methods->obj_at_put(index, method());  
    }
    return methods;
  }
}


typeArrayHandle ClassFileParser::sort_methods(objArrayHandle methods, TRAPS) {
  int length = methods()->length();
  // If JVMDI is enabled we have to remember the original class file ordering.
  // We temporarily use the vtable_index field in the methodOop to store the
  // class file index, so we can read in after calling qsort.
  if (jvmdi::enabled()) {
    for (int index = 0; index < length; index++) {
      methodOop m = methodOop(methods->obj_at(index));
      assert(m->vtable_index() == -1, "vtable index should not be set");
      m->set_vtable_index(index);
    }
  }
  // Sort method array by ascending method name (for faster lookups & vtable construction)
  // Note that the ordering is not alfabetical, see symbolOopDesc::fast_compare
  if (length > 1) qsort(methods()->obj_at_addr(0), length, oopSize, (compareFn)method_compare);
  // If JVMDI is enabled construct int array remembering the original ordering
  if (jvmdi::enabled()) {
    typeArrayHandle method_ordering = oopFactory::new_permanent_intArray(length, CHECK_(typeArrayOop(NULL)));
    for (int index = 0; index < length; index++) {
      methodOop m = methodOop(methods->obj_at(index));
      int old_index = m->vtable_index();
      assert(old_index >= 0 && old_index < length, "invalid method index");
      method_ordering->int_at_put(index, old_index);
      m->set_vtable_index(-1);
    }
    return method_ordering;
  } else {
    return Universe::the_empty_int_array();
  }
}


void ClassFileParser::parse_classfile_sourcefile_attribute(constantPoolHandle cp, instanceKlassHandle k, TRAPS) {
  u2 sourcefile_index = stream()->get_u2();
  k->set_source_file_name(cp->symbol_at(sourcefile_index));
}



// Inner classes can be static, private or protected (classic VM does this)
#define RECOGNIZED_INNER_CLASS_MODIFIERS (JVM_RECOGNIZED_CLASS_MODIFIERS | JVM_ACC_PRIVATE | JVM_ACC_PROTECTED | JVM_ACC_STATIC)


void ClassFileParser::parse_classfile_inner_classes_attribute(constantPoolHandle cp, instanceKlassHandle k, TRAPS) {  
  u2 length = stream()->get_u2();

  // 4-tuples of shorts [inner_class_info_index, outer_class_info_index, inner_name_index, inner_class_access_flags]
  typeArrayHandle inner_classes = oopFactory::new_permanent_shortArray(length*4, CHECK);  
  int index = 0;
  for (int n = 0; n < length; n++) {
    // Inner class index
    u2 inner_class_info_index = stream()->get_u2();
    check_property(inner_class_info_index == 0 || cp->tag_at(inner_class_info_index).is_klass_reference(), "Invalid .class file", CHECK);
    // Outer class index
    u2 outer_class_info_index = stream()->get_u2();
    check_property(outer_class_info_index == 0 || cp->tag_at(outer_class_info_index).is_klass_reference(), "Invalid .class file", CHECK);
    // Inner class name
    u2 inner_name_index = stream()->get_u2();
    check_property(inner_name_index == 0 || cp->tag_at(inner_name_index).is_utf8(), "Invalid .class file", CHECK);    
    // Access flags
    AccessFlags inner_access_flags;
    jint flags = stream()->get_u2() & RECOGNIZED_INNER_CLASS_MODIFIERS;
    if ((flags & JVM_ACC_INTERFACE) != 0) {
      // Set the abstract bit explicitly for interface classes (Classic VM does this)
      flags |= JVM_ACC_ABSTRACT;
    }
    inner_access_flags.set_flags(flags);

    inner_classes->short_at_put(index++, inner_class_info_index);
    inner_classes->short_at_put(index++, outer_class_info_index);
    inner_classes->short_at_put(index++, inner_name_index);	
    inner_classes->short_at_put(index++, inner_access_flags.as_short());
  }

  // Update instanceKlass with inner class info.  
  k->set_inner_classes(inner_classes());
}

void ClassFileParser::parse_classfile_synthetic_attribute(constantPoolHandle cp, instanceKlassHandle k, TRAPS) {
  k->set_is_synthetic();
}

void ClassFileParser::parse_classfile_attributes(constantPoolHandle cp, instanceKlassHandle k, TRAPS) {
  // Set inner classes attribute to default sentinel
  k->set_inner_classes(Universe::the_empty_short_array());
  u2 attributes_count = stream()->get_u2();
  // Iterate over attributes
  while (attributes_count--) {    
    u2 attribute_name_index = stream()->get_u2();
    u4 attribute_length = stream()->get_u4();
    check_property(cp->tag_at(attribute_name_index).is_utf8(), "Invalid attribute in .class file", CHECK);
    symbolOop tag = cp->symbol_at(attribute_name_index);
    if (tag == vmSymbols::tag_source_file()) {
      // Check for SourceFile tag
      parse_classfile_sourcefile_attribute(cp, k, CHECK);
    } else if (tag == vmSymbols::tag_inner_classes()) {
      // Check for InnerClasses tag
      parse_classfile_inner_classes_attribute(cp, k, CHECK);
    } else if (tag == vmSymbols::tag_synthetic()) {
      // Check for Synthetic tag
      // Shouldn't we check that the synthetic flags wasn't already set?
      if (attribute_length != 0) {
        classfile_parse_error("Invalid synthetic attribute in .class file", CHECK);
      }
      parse_classfile_synthetic_attribute(cp, k, CHECK);
    } else {
      // Unknown attribute
      stream()->skip_u1(attribute_length);
    }
  }
}


static void initialize_static_field(fieldDescriptor* fd, TRAPS) {
  KlassHandle h_k = fd->field_holder();
  assert(h_k.not_null() && fd->is_static(), "just checking");
  if (fd->has_initial_value()) {
    BasicType t = fd->field_type();
    switch (t) {
      case T_BYTE:
      case T_CHAR:
      case T_SHORT:
      case T_BOOLEAN:
      case T_INT:
        h_k()->int_field_put(fd->offset(), fd->int_initial_value());
        break;
      case T_FLOAT:
        h_k()->float_field_put(fd->offset(), fd->float_initial_value());
        break;
      case T_DOUBLE:
        h_k()->double_field_put(fd->offset(), fd->double_initial_value());
        break;
      case T_LONG:
        h_k()->long_field_put(fd->offset(), fd->long_initial_value());
        break;
      case T_OBJECT:
        {
          #ifdef ASSERT      
          oop sym = oopFactory::new_symbol("Ljava/lang/String;", CHECK);
          assert(fd->signature() == sym, "just checking");      
          #endif
          oop string = fd->string_initial_value(CHECK);
          h_k()->obj_field_put(fd->offset(), string);
        }
        break;
      default:
        classfile_parse_error("Illegal ConstantValue attribute", CHECK);
    }
  }
}


instanceKlassHandle ClassFileParser::parseClassFile(symbolHandle name, Handle class_loader, Handle protection_domain, TRAPS) {
  // Timing
  TraceTime timer(NULL, ClassLoader::accumulated_time(), TraceClassLoadingTime, false);

  if (jvmpi::is_event_enabled(JVMPI_EVENT_CLASS_LOAD_HOOK)) {
    ClassFileStream* cfs = stream();
    unsigned char* ptr = cfs->buffer();
    unsigned char* end_ptr = cfs->buffer() + cfs->length();

    typedef void* (* malloc_t)(unsigned int);
    jvmpi::post_class_load_hook_event(&ptr, &end_ptr, (malloc_t)resource_allocate_bytes);

    if (ptr != cfs->buffer()) {
      cfs = new ClassFileStream(ptr, end_ptr - ptr, cfs->source());
      set_stream(cfs);
    }
  }

  // Verify the class file format
  Verifier::verify_class_format(name, class_loader, stream(), CHECK_(klassOop(NULL)));

  klassOop preserve_this_klass;   // for storing result across HandleMark
  
  // release all handles when parsing is done
  { HandleMark hm;

    // Magic value
    u4 magic = stream()->get_u4();
    guarantee_property(magic == 0xCAFEBABE, "Incompatible magic value in .class file", CHECK_(klassOop(NULL)));

    // Version numbers  
    u2 minor_version = stream()->get_u2();
    u2 major_version = stream()->get_u2();
    // version numbers checked by class format checker

    // Constant pool
    constantPoolHandle cp = parse_constant_pool(CHECK_(klassOop(NULL)));
  
    // Access flags
    AccessFlags access_flags;
    jint flags = stream()->get_u2() & JVM_RECOGNIZED_CLASS_MODIFIERS;
    if ((flags & JVM_ACC_INTERFACE) != 0) {
      // Set the abstract bit explicitly for interface classes (Classic VM does this)
      flags |= JVM_ACC_ABSTRACT;
    }
    access_flags.set_flags(flags);

    // This class and superclass
    instanceKlassHandle super_klass;
    u2 this_class_index = stream()->get_u2();
    check_property(cp->tag_at(this_class_index).is_unresolved_klass(), "Invalid .class file", CHECK_(klassOop(NULL)));

    symbolHandle class_name = cp->unresolved_klass_at(this_class_index);
  
    // Checks if name in class file matches requested name
    if (name.not_null() && class_name() != name()) {
      ResourceMark rm;
      Exceptions::fthrow(
        THREAD_AND_LOCATION,
        vmSymbols::java_lang_NoClassDefFoundError(), 
        "%s (wrong name: %s)", 
        name->as_C_string(), 
        class_name->as_C_string()
      );
    }

    u2 super_class_index = stream()->get_u2();
    if (super_class_index == 0) {
      check_property(class_name() == vmSymbols::java_lang_Object(), "Invalid superclass", CHECK_(klassOop(NULL)));    
    } else {
      check_property(cp->tag_at(super_class_index).is_unresolved_klass(), "Invalid superclass", CHECK_(klassOop(NULL)));    
      timer.suspend();  // do not count recursive loading twice
      super_klass = SystemDictionary::resolve_or_fail(cp->unresolved_klass_at(super_class_index), class_loader, protection_domain, true, CHECK_(klassOop(NULL)));    
      timer.resume();
      // Checks below not done by format checker
      if (super_klass->is_final()) {
        THROW_MSG_(vmSymbols::java_lang_VerifyError(), "Cannot inherit from final class", klassOop(NULL));
      }
      if (access_flags.is_interface() && super_klass() != SystemDictionary::object_klass()) {
        classfile_parse_error("Interfaces must have java.lang.Object as superclass", CHECK_(klassOop(NULL)));
      }
    }

    // Interfaces
    objArrayHandle local_interfaces = parse_interfaces(cp, class_loader, protection_domain, CHECK_(klassOop(NULL)));

    // Compute the transitive list of all unique interfaces implemented by this class
    objArrayHandle transitive_interfaces = compute_transitive_interfaces(super_klass, local_interfaces, CHECK_(klassOop(NULL)));

    // Fields (offsets are filled in later)
    int static_oop_count = 0;
    int nonstatic_oop_count = 0;
    typeArrayHandle fields = parse_fields(cp, &static_oop_count, &nonstatic_oop_count, CHECK_(klassOop(NULL)));

    // Methods
    AccessFlags promoted_flags;
    promoted_flags.set_flags(0);
    objArrayHandle methods = parse_methods(cp, &promoted_flags, CHECK_(klassOop(NULL)));
    typeArrayHandle method_ordering = sort_methods(methods, CHECK_(klassOop(NULL)));

    // promote flags from parse_methods() to the klass' flags
    access_flags.add_promoted_flags(promoted_flags.as_int());

    // Size of Java vtable (in words)
    int vtable_size = 0;    
    int itable_size = 0;
    int num_miranda_methods = 0;

    klassVtable::compute_vtable_size_and_num_mirandas(vtable_size, 
  						      num_miranda_methods, 
						      super_klass(),
						      methods(),
						      access_flags,
						      class_loader(),
						      class_name(), 
						      local_interfaces());  
       
    // Size of Java itable (in words)
    itable_size = access_flags.is_interface() ? 0 : klassItable::compute_itable_size(transitive_interfaces);  
    
    bool has_speclock_field =  UseSpecLock && class_loader.is_null() && SpecLock::has_speclock_field(class_name());

    // Field size and offset computation
    int nonstatic_field_size = super_klass() == NULL ? 0 : super_klass->nonstatic_field_size();
    int static_field_size = 0;
    if (has_speclock_field) {  
      if (Verbose) {
        tty->print_cr("Adding a dummy field for speculative locking to class");
      }
      nonstatic_field_size++; 
    }

    int next_static_oop_offset      = instanceKlass::header_size() + align_object_offset(vtable_size) + align_object_offset(itable_size);
    int next_static_type_offset     = next_static_oop_offset + static_oop_count;
    int next_nonstatic_oop_offset   = instanceOopDesc::header_size() + nonstatic_field_size;
    int next_nonstatic_type_offset  = next_nonstatic_oop_offset + nonstatic_oop_count;
    int first_nonstatic_oop_offset  = next_nonstatic_oop_offset;

    // Iterate over fields again and compute correct offsets.
    // The field allocation type was temporarily stored in the offset slot.
    // oop fields are located before non-oop fields (static and non-static).
    int len = fields->length();
    for (int i = 0; i < len; i += 5) {
      int real_offset;
      FieldAllocationType atype = (FieldAllocationType) fields->ushort_at(i+4);
      switch (atype) {
        case STATIC_OOP:
          real_offset = next_static_oop_offset++;
          static_field_size++;
          break;
        case STATIC_SINGLE:
          real_offset = next_static_type_offset++;
          static_field_size++;
          break;
        case STATIC_DOUBLE:
          real_offset = next_static_type_offset;
          next_static_type_offset += 2;
          static_field_size += 2;
          break;
        case STATIC_ALIGNED_DOUBLE:
          {
            // Ensure that the current offset for static fields is aligned
            int aligned_offset = align_size_up(next_static_type_offset, 2);
            if (aligned_offset != next_static_type_offset) {
              next_static_type_offset = aligned_offset;
              static_field_size++;
            }
          }
          real_offset = next_static_type_offset;
          next_static_type_offset += 2;
          static_field_size += 2;
          break;
        case NONSTATIC_OOP:
          real_offset = next_nonstatic_oop_offset++;
          nonstatic_field_size++;
          break;
        case NONSTATIC_SINGLE:
          real_offset = next_nonstatic_type_offset++;
          nonstatic_field_size++;
          break;
        case NONSTATIC_DOUBLE:
          real_offset = next_nonstatic_type_offset;
          next_nonstatic_type_offset += 2;
          nonstatic_field_size += 2;
          break;
        case NONSTATIC_ALIGNED_DOUBLE:
          {
            // Ensure that the current offset for nonstatic fields is aligned
            int aligned_offset = align_size_up(next_nonstatic_type_offset, 2);
            if (aligned_offset != next_nonstatic_type_offset) {
              next_nonstatic_type_offset = aligned_offset;
              nonstatic_field_size++;
            }
          }
          real_offset = next_nonstatic_type_offset;
          next_nonstatic_type_offset += 2;
          nonstatic_field_size += 2;
          break;
        default:
          ShouldNotReachHere();
      }
      fields->short_at_put(i+4, real_offset);
    }

    // Add fake fields for java.lang.Class instances
    // This is not particularly nice. We should consider adding a special Klass for this.
    if (class_name() == vmSymbols::java_lang_Class() && class_loader.is_null()) {
      assert(nonstatic_field_size == 0, "just checking"); 
      assert(nonstatic_oop_count == 0, "just checking");
      assert(next_nonstatic_type_offset == instanceOopDesc::header_size(), "just checking");
      // The values below are fake but will force two non-static oop fields and 
      // a corresponding non-static oop map block to be allocated.
      const int extra = java_lang_Class::number_of_fake_oop_fields;
      nonstatic_field_size = extra;
      nonstatic_oop_count = extra;
      next_nonstatic_type_offset += extra;
    }

    // Size of instances
    int instance_size = align_object_size(next_nonstatic_type_offset);

    // Size of non-static oop map blocks (in words) allocated at end of klass
    int nonstatic_oop_map_size = compute_oop_map_size(super_klass, nonstatic_oop_count, first_nonstatic_oop_offset);

    // Compute reference type
    ReferenceType rt;
    if (super_klass() == NULL) {
      rt = REF_NONE;
    } else {
      rt = super_klass->reference_type();
    }

    // We can now create the basic klassOop for this klass    
    instanceKlassHandle this_klass = oopFactory::new_instanceKlass(vtable_size, itable_size, 
                                                                   static_field_size, nonstatic_oop_map_size, 
                                                                   rt, CHECK_(klassOop(NULL)));    

    assert(this_klass->static_field_size() == static_field_size && 
           this_klass->nonstatic_oop_map_size() == nonstatic_oop_map_size, "sanity check");
    
    // Fill in information already parsed
    this_klass->set_access_flags(access_flags);
    this_klass->set_size_helper(instance_size);
    this_klass->set_super(super_klass());
    this_klass->set_class_loader(class_loader());    
    this_klass->set_nonstatic_field_size(nonstatic_field_size);
    this_klass->set_static_oop_field_size(static_oop_count);       
    cp->set_pool_holder(this_klass());
    this_klass->set_constants(cp());
    this_klass->set_local_interfaces(local_interfaces());
    this_klass->set_fields(fields());
    this_klass->set_methods(methods());
    this_klass->set_method_ordering(method_ordering());
    this_klass->set_name(cp->unresolved_klass_at(this_class_index));  // An unresolved klass entry contains a symbol
    this_klass->set_protection_domain(protection_domain());

    // Miranda methods
    if ((num_miranda_methods > 0) || 
	// if this class introduced new miranda methods or
	(super_klass.not_null() && (super_klass->has_miranda_methods()))
	// super class exists and this class inherited miranda methods
	) {
      this_klass->set_has_miranda_methods(); // then set a flag
    }

    // Additional attributes
    parse_classfile_attributes(cp, this_klass, CHECK_(klassOop(NULL)));

    // Initialize static fields
    this_klass->do_local_static_fields(&initialize_static_field, CHECK_(klassOop(NULL)));

    // VerifyOops believes that once this has been set, the object is completely loaded.
    // Compute transitive closure of interfaces this class implements
    this_klass->set_transitive_interfaces(transitive_interfaces());    

    // Initialize itable offset tables
    klassItable::setup_itable_offset_table(this_klass);

    // Do final class setup
    fill_oop_maps(this_klass, nonstatic_oop_count, first_nonstatic_oop_offset);
    set_precomputed_flags(this_klass);

    // the following checks should really be done by the verifier.

    // check if this class can access its super class
    check_super_class_access(this_klass, CHECK_(klassOop(NULL)));

    // check if this class can access its superinterfaces
    check_super_interface_access(this_klass, CHECK_(klassOop(NULL)));

    //check if this class overrides any final method
    check_final_method_override(this_klass, CHECK_(klassOop(NULL)));

    // check that if this class is an interface then it doesn't have static methods
    if (this_klass->is_interface()) {
      check_illegal_static_method(this_klass, CHECK_(klassOop(NULL)));
    }
	  
    if (TraceClassLoading) {
      tty->print("[Loaded %s", this_klass->external_name());
      if (stream()->source() != NULL) tty->print(" from %s", stream()->source());
      tty->print_cr("]");
    }
    
    // preserve result across HandleMark  
    preserve_this_klass = this_klass();    
  }

  // Create new handle outside HandleMark
  instanceKlassHandle this_klass = preserve_this_klass;
  debug_only(this_klass->as_klassOop()->verify();)

  // Fix a performance bug in the 1.2 class library. 
  //  
  // The default constructor is:
  //
  //   Hashtable() { super(101, 0.75f); }
  //
  // A default size of 11 is much more reasonable. This bug has been fixed in 
  // the 1.3 version of the libraries
  //
  // Bytecodes:
  //
  // 0 load_local_object #0
  // 1 push_int 101        <--- Change to: push_int 11
  // 3 push_float 0.750000
  // 5 invoke_special 42 <<init>> <(IF)V>
  // 8 return_void
  //  
  if (name() == vmSymbols::java_util_Hashtable() && JavaUtilHashTableFix && class_loader.is_null()) {    
    // Find no-arg consturctor
    methodOop m = this_klass->find_method(vmSymbolHandles::object_initializer_name(), vmSymbolHandles::void_method_signature());
    if (m != NULL) {
      address codes = m->code_base();
      // Make sure that the method has the correct layout
      if (m->code_size() == 9 && 
          codes[0] == Bytecodes::_aload_0       &&
          codes[1] == Bytecodes::_bipush        && codes[2] == 101  &&
          codes[3] == Bytecodes::_ldc           && codes[4] == 0x04 &&
          codes[5] == Bytecodes::_invokespecial && codes[6] == 0x00 && codes[7] == 0x2a &&
          codes[8] == Bytecodes::_return) {
        // Change byte to 11
        codes[2] = 11;
      }        
    }    
  }

  return this_klass;
}


int ClassFileParser::compute_oop_map_size(instanceKlassHandle super, int nonstatic_oop_count, int first_nonstatic_oop_offset) {
  int map_size = super.is_null() ? 0 : super->nonstatic_oop_map_size();
  if (nonstatic_oop_count > 0) {
    // We have oops to add to map
    if (map_size == 0) {
      map_size++;
    } else {
      // Check whether we should add a new map block or whether the last one can be extended
      OopMapBlock* first_map = super->start_of_nonstatic_oop_maps();
      OopMapBlock* last_map = first_map + map_size - 1;
      int next_offset = last_map->offset() + last_map->length();
      if (next_offset != first_nonstatic_oop_offset) {
        // Superklass didn't end with a oop field, add extra map
        assert(next_offset<first_nonstatic_oop_offset, "just checking");
        map_size++;
      }
    }
  }
  return map_size;
}


void ClassFileParser::fill_oop_maps(instanceKlassHandle k, int nonstatic_oop_count, int first_nonstatic_oop_offset) {
  OopMapBlock* this_oop_map = k->start_of_nonstatic_oop_maps();
  OopMapBlock* last_oop_map = this_oop_map + k->nonstatic_oop_map_size() - 1;
  instanceKlass* super = k->superklass();
  if (super != NULL) {
    int super_oop_map_size     = super->nonstatic_oop_map_size();
    OopMapBlock* super_oop_map = super->start_of_nonstatic_oop_maps();
    // Copy maps from superklass
    while (super_oop_map_size-- > 0) {
      *this_oop_map++ = *super_oop_map++;
    }
  }
  if (nonstatic_oop_count > 0) {
    if (this_oop_map == last_oop_map) {
      // We added a new map block, fill it
      last_oop_map->set_offset(first_nonstatic_oop_offset);
      last_oop_map->set_length(nonstatic_oop_count);
    } else {
      // We should extend the last map block copied from the superklass
      assert(this_oop_map == last_oop_map + 1, "just checking");
      last_oop_map->set_length(last_oop_map->length() + nonstatic_oop_count);
    }
  }
}


void ClassFileParser::set_precomputed_flags(instanceKlassHandle k) {
  // Check if this klass has an empty finalize method (i.e. one with return bytecode only),
  // in which case we don't have to register objects as finalizable
  methodOop m = k->lookup_method(vmSymbols::finalize_method_name(), 
                                 vmSymbols::void_method_signature());
  if (m != NULL && !m->is_empty_method()) {
    k->set_has_finalizer();
  }

  // Check if this klass supports the java.lang.Cloneable interface
  if (SystemDictionary::cloneable_klass_loaded()) {
    if (k->is_subtype_of(SystemDictionary::cloneable_klass())) {
      k->set_is_cloneable();
    }
  }

  // Check if this klass has a vanilla default constructor
  klassOop super = k->super();
  if (super == NULL) {
    // java.lang.Object has empty default constructor
    k->set_has_vanilla_constructor();
  } else {
    if (Klass::cast(super)->has_vanilla_constructor()) {
      methodOop constructor = k->find_method(vmSymbols::object_initializer_name(), vmSymbols::void_method_signature());
      if (constructor != NULL && constructor->is_vanilla_constructor()) {
        k->set_has_vanilla_constructor();
      }
    }
  }
}


// utility method for appending and array with check for duplicates

void append_interfaces(objArrayHandle result, int& index, objArrayOop ifs) {
  // iterate over new interfaces
  for (int i = 0; i < ifs->length(); i++) {
    oop e = ifs->obj_at(i);
    assert(e->is_klass() && instanceKlass::cast(klassOop(e))->is_interface(), "just checking");
    // check for duplicates
    bool duplicate = false;
    for (int j = 0; j < index; j++) {
      if (result->obj_at(j) == e) {
        duplicate = true;
        break;
      }
    }
    // add new interface
    if (!duplicate) {
      result->obj_at_put(index++, e);
    }
  }
}

objArrayHandle ClassFileParser::compute_transitive_interfaces(instanceKlassHandle super, objArrayHandle local_ifs, TRAPS) {
  // Compute maximum size for transitive interfaces
  int max_transitive_size = 0;
  int super_size = 0;
  // Add superclass transitive interfaces size
  if (super.not_null()) {
    super_size = super->transitive_interfaces()->length();
    max_transitive_size += super_size;
  }
  // Add local interfaces' super interfaces  
  int local_size = local_ifs->length();
  for (int i = 0; i < local_size; i++) {
    klassOop l = klassOop(local_ifs->obj_at(i));
    max_transitive_size += instanceKlass::cast(l)->transitive_interfaces()->length();
  }
  // Finally add local interfaces
  max_transitive_size += local_size;
  // Construct array
  objArrayHandle result;
  if (max_transitive_size == 0) {
    // no interfaces, use canonicalized array
    result = Universe::the_empty_system_obj_array();
  } else if (max_transitive_size == super_size) {
    // no new local interfaces added, share superklass' transitive interface array
    result = super->transitive_interfaces();
  } else if (max_transitive_size == local_size) {
    // only local interfaces added, share local interface array
    result = local_ifs;    
  } else {
    result = oopFactory::new_system_objArray(max_transitive_size, CHECK_((objArrayOop)NULL));
    int index = 0;
    // Copy down from superclass
    if (super.not_null()) {
      append_interfaces(result, index, super->transitive_interfaces());
    }    
    // Copy down from local interfaces' superinterfaces
    for (int i = 0; i < local_ifs->length(); i++) {
      klassOop l = klassOop(local_ifs->obj_at(i));
      append_interfaces(result, index, instanceKlass::cast(l)->transitive_interfaces());
    }
    // Finally add local interfaces
    append_interfaces(result, index, local_ifs());

    // Check if duplicates were removed
    if (index != max_transitive_size) {
      assert(index < max_transitive_size, "just checking");
      objArrayOop new_result = oopFactory::new_system_objArray(index, CHECK_((objArrayOop)NULL));
      for (int i = 0; i < index; i++) {
        oop e = result->obj_at(i);
        assert(e != NULL, "just checking");
        new_result->obj_at_put(i, e);
      }
      result = new_result;
    }
  }
  return result;  
}


void ClassFileParser::check_super_class_access(instanceKlassHandle this_klass, TRAPS) {
  klassOop super = this_klass->super();
  if ((super != NULL) &&
      (!Reflection::verify_class_access(this_klass->as_klassOop(), super, false))) {
    ResourceMark rm;
    Exceptions::fthrow(  
      THREAD_AND_LOCATION,
      vmSymbols::java_lang_IllegalAccessError(),
      "class %s cannot access its superclass %s",
      this_klass->external_name(),
      instanceKlass::cast(super)->external_name()
      );
  }
}


void ClassFileParser::check_super_interface_access(instanceKlassHandle this_klass, TRAPS) {
  objArrayHandle local_interfaces  = this_klass->local_interfaces();
  int lng = local_interfaces->length();
  for (int i = lng - 1; i >= 0; i--) {
    klassOop k = klassOop(local_interfaces->obj_at(i)); 
    assert (k != NULL && Klass::cast(k)->is_interface(), "invalid interface");
    if (!Reflection::verify_class_access(this_klass->as_klassOop(), k, false)) {
      ResourceMark rm;
      Exceptions::fthrow(  
        THREAD_AND_LOCATION,
        vmSymbols::java_lang_IllegalAccessError(),
        "class %s cannot access its superinterface %s",
        this_klass->external_name(),
        instanceKlass::cast(k)->external_name()
        );
    }
  }
}


void ClassFileParser::check_final_method_override(instanceKlassHandle this_klass, TRAPS) {
  objArrayHandle methods = this_klass->methods();
  int num_methods = methods->length();
  
  // go thru each method and check if it overrides a final method
  for (int index = 0; index < num_methods; index++) {
    methodOop m = (methodOop)methods->obj_at(index);

    // skip private, static and <init> methods
    if ((!m->is_private()) &&
	(!m->is_static()) &&
	(m->name() != vmSymbols::object_initializer_name())) {
	
      symbolOop name = m->name();
      symbolOop signature = m->signature();
      klassOop k = this_klass->super();
      methodOop super_m = NULL;
      while (k != NULL) {
	// lookup a matching method in the super class hierarchy
        super_m = instanceKlass::cast(k)->lookup_method(name, signature); 
	if (super_m == NULL) {
	  break; // didn't find any match; get out
	}

	if (super_m->is_final() &&
	    // matching method in super is final
	    (Reflection::verify_field_access(this_klass->as_klassOop(), 
					     super_m->method_holder(),
					     super_m->access_flags(),
					     false))
	    // this class can access super final method and therefore override
	    ) {
	  ResourceMark rm;
	  Exceptions::fthrow(  
	    THREAD_AND_LOCATION,
	    vmSymbols::java_lang_VerifyError(),
	    "class %s overrides final method %s.%s",
	    this_klass->external_name(),
	    name,
	    signature
	    );
	}

        k = instanceKlass::cast(super_m->method_holder())->super(); // continue to look
      }
    }
  }
}


// assumes that this_klass is an interface
void ClassFileParser::check_illegal_static_method(instanceKlassHandle this_klass, TRAPS) {
  assert(this_klass->is_interface(), "not an interface");
  objArrayHandle methods = this_klass->methods();
  int num_methods = methods->length();

  for (int index = 0; index < num_methods; index++) {
    methodOop m = (methodOop)methods->obj_at(index);
    // if m is static and not the init method, throw a verify error
    if ((m->is_static()) && (m->name() != vmSymbols::class_initializer_name())) {
      ResourceMark rm;
      Exceptions::fthrow(  
	THREAD_AND_LOCATION,
	vmSymbols::java_lang_VerifyError(),
	"Illegal static method %s in interface %s",
	m->name(),
	this_klass->external_name()
	);
    }
  }
}
