/*
 * @(#)deoptimization.cpp	1.170 01/04/21 22:47:09
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

#include "incls/_precompiled.incl"
#include "incls/_deoptimization.cpp.incl"

bool DeoptimizationMarker::_is_active = false;

Deoptimization::UnrollBlock::UnrollBlock(jint  size_of_deoptimized_frame,
                                         jint  pop_i2c_frame,
					 jint  adjust_i2c_frame,
                                         jint  number_of_frames,
                                         jint* frame_sizes,
                                         bool  returns_double) {
  _size_of_deoptimized_frame = size_of_deoptimized_frame;
  _pop_i2c_frame             = pop_i2c_frame;
  _adjust_i2c_frame          = adjust_i2c_frame;
  _number_of_frames          = number_of_frames;
  _frame_sizes               = frame_sizes;
  _register_block            = NEW_C_HEAP_ARRAY(jint, RegisterMap::reg_count * 2);
  _returns_double            = returns_double ? 1 : 0;
}

Deoptimization::UnrollBlock::~UnrollBlock() {
  FREE_C_HEAP_ARRAY(jint, _frame_sizes);
  FREE_C_HEAP_ARRAY(jint, _register_block);
}

jint* Deoptimization::UnrollBlock::value_addr_at(int register_number) const {
  assert(register_number < RegisterMap::reg_count, "checking register number");
  return &_register_block[register_number * 2];
}

jint Deoptimization::UnrollBlock::size_of_frames() const {
  int result = 0;
  for (int index = 0; index < number_of_frames(); index++) {
    result += frame_sizes()[index];
  }
  return result;
}

void Deoptimization::UnrollBlock::print() {
  tty->print_cr("UnrollBlock");
  tty->print_cr("  size_of_deoptimized_frame = %d", _size_of_deoptimized_frame);
  tty->print(   "  frame_sizes: ");
  for (int index = 0; index < number_of_frames(); index++) {
    tty->print("%d ", frame_sizes()[index]);
  }
  tty->cr();
}


JRT_LEAF(Deoptimization::UnrollBlock*,  Deoptimization::fetch_unroll_info(JavaThread* thread))
  // It is actually ok to allocate handles in a leaf method. It causes no safepoints,
  // but makes the entry a little slower. There is however a little dance we have to
  // do in debug mode to get around the NoHandleMark code in the JRT_LEAF macro
  ResetNoHandleMark rnhm; // No-op in release/product versions  
  HandleMark hm;
  ResourceMark rm;
  return fetch_unroll_info_helper(thread);
JRT_END

// This is factored, since it is both called from a JRT_LEAF (deoptimization) and a JRT_ENTRY (uncommon_trap)
Deoptimization::UnrollBlock* Deoptimization::fetch_unroll_info_helper(JavaThread* thread) {
#ifndef COMPILER2
  ShouldNotReachHere(); // deoptimization not yet implemented for C1
  return NULL;
#else
  // Since the frame to unpack is the top frame of this thread the 
  // vframe_array_head must point to the vframeArray
  // for the unpack frame.
  vframeArray* array = thread->vframe_array_head();
  frame stub_frame = thread->last_frame();

#ifdef ASSERT
  // compute caller frame base on sender sp of stub_frame
  // and stored frame sizes info
  CodeBlob* cb = CodeCache::find_blob(stub_frame.pc());
  assert(cb->is_deoptimization_stub() || cb->is_uncommon_trap_stub(), "just checking");
  // Verify we have the right vframeArray
  assert(cb->frame_size() >= 0, "Compiled by C1: do not use");
  jint* unpack_sp = stub_frame.sp() + cb->frame_size();
  assert(array->sp() == unpack_sp, "vframe_array_head must contain the vframeArray to unpack");
#endif

  // migrate the monitors from stack to off stack representation
  // MonitorArray -> MonitorChunk conversion
  array->migrate_monitors_off_stack();

  // Collect the vframes in the vframeArray
  ResourceMark rm;
  GrowableArray<deoptimizedVFrame*>* vframes = new GrowableArray<deoptimizedVFrame*>(10);
  for (deoptimizedVFrame* vf = new deoptimizedVFrame(array); vf; vf = deoptimizedVFrame::cast(vf->sender())) {
    vframes->push(vf);
  }

  // Get the adapter blob.  If none is needed, then we are also popping off a
  // redundant I2C or OSR adapter frame.
  C2IAdapter* adapter = array->adapter();

  // We need an extra adapter frame when going from compile code calling
  // convention to interpreter calling convention.
  int number_of_frames = vframes->length() + (adapter ? 1 : 0);

  // Compute the resulting array. Note that the order is reversed.
  jint* frame_sizes = NEW_C_HEAP_ARRAY(jint, number_of_frames);
  int callee_parameters = 0;
  int callee_locals = 0;

  for (int index = 0; index < vframes->length(); index++) {
    frame_sizes[number_of_frames - 1 - index] = vframes->at(index)->on_stack_size(callee_parameters, callee_locals) * wordSize;
    callee_parameters = vframes->at(index)->method()->size_of_parameters();
    callee_locals = vframes->at(index)->method()->max_locals();
  }

  // Compute if the top vframe expects a float or double as return value.
  bool returns_double;
  {
    HandleMark hm;
    deoptimizedVFrame* vf = vframes->at(0);
    Bytecode_invoke* invoke = Bytecode_invoke_at_check(vf->method(), vf->bci());
    BasicType type = invoke != NULL 
                   ? invoke->result_type()
                   : T_ILLEGAL;
    returns_double = (type == T_FLOAT) || (type == T_DOUBLE);
  }

  // The first element is for the adapter frame
  int i2c_frame_adjust;
  if( adapter ) {
    frame_sizes[0] = (adapter->frame_size() + last_frame_adjust(callee_parameters, callee_locals)) * wordSize; 
    i2c_frame_adjust = 0;	// Do not pop I2C frame
  } else {
    i2c_frame_adjust = array->i2c_frame_adjust(callee_parameters, callee_locals) * wordSize;
  }
  UnrollBlock* info = new UnrollBlock(array->frame_size() * wordSize, adapter == NULL, i2c_frame_adjust, number_of_frames, frame_sizes, returns_double);
  array->set_unroll_block(info);

  return info;
#endif // COMPILER2
}

JRT_LEAF(void, Deoptimization::unpack_frames(JavaThread* thread, int exec_mode))
#ifndef COMPILER2
  ShouldNotReachHere(); // deoptimization not yet implemented for C1
  return;
#else
  ResourceMark rm;

  // It is actually ok to allocate handles in a leaf method. It causes no safepoints,
  // but makes the entry a little slower. There is however a little dance we have to
  // do in debug mode to get around the NoHandleMark code in the JRT_LEAF macro
  ResetNoHandleMark rnhm; // No-op in release/product versions
  HandleMark hm(thread);

  frame stub_frame = thread->last_frame();

  // Since the frame to unpack is the top frame of this thread the 
  // vframe_array_head must point to the vframeArray
  // for the unpack frame.
  vframeArray* array = thread->vframe_array_head();


  if (TraceDeoptimization) {
    tty->print_cr("Deoptimization: unpacking vframeArray(0x%lx)", array);
  }

  // compute caller frame base on sender sp of stub_frame
  // and stored frame sizes info
  CodeBlob* cb = CodeCache::find_blob(stub_frame.pc());
  assert(cb->is_deoptimization_stub() || cb->is_uncommon_trap_stub(), "just checking");

  UnrollBlock* info = array->unroll_block();
  assert(cb->frame_size() >= 0, "Compiled by C1: do not use");

  C2IAdapter* adapter = array->adapter();
  frame caller_frame;
  if( adapter ) {		// Need a C2I adapter frame?
    // Unpack the adapter frame; stuff its registers, especially the frame pointer
    jint* caller_sp = (stub_frame.sp() + cb->frame_size()) + (info->size_of_frames()/wordSize);
    assert(adapter->frame_size() >= 0, "Compiled by C1: do not use");
    caller_frame = adapter->unpack_c2i_adapter(caller_sp, array);
    adapter->setup_stack_frame(caller_frame, array);
  } else {			// Else no C2I AND no I2C adapter
    // No adapter frames; instead we need to move frame bits captured from the
    // removed I2C adapter and stuff them into the top level interpreter frame.
    caller_frame = array->last_i2c_frame();
  }

  // remove array from thread
  thread->set_vframe_array_head(array->next());

  // Free the UnrollBlock
  array->set_unroll_block(NULL);
  delete info;

  // Unpack the remaining interpreter frames
  array->unpack_to_stack(stub_frame, caller_frame, exec_mode);
  delete array;

#endif // COMPILER2
JRT_END

int Deoptimization::deoptimize_dependents() {
  Threads::deoptimized_wrt_marked_nmethods();
  return 0;
}

vframeArray* Deoptimization::create_vframeArray(JavaThread* thread, frame fr, RegisterMap *reg_map) {
  assert(fr.is_interpreted_frame() || fr.is_compiled_frame(), "checking frame type");
  // Create the vframeArray, a growable array of VFrames, each VFrame 
  // represents an inlined Java frame.  This storage is allocated with the
  // usual system arena.
  GrowableArray<compiledVFrame*>* chunk = new GrowableArray<compiledVFrame*>(10);  
  vframe* vf = vframe::new_vframe(&fr, reg_map, thread);        
  while (!vf->is_top()) {    
    assert(vf->is_interpreted_frame() || vf->is_compiled_frame(), "Wrong frame type");
    chunk->push(compiledVFrame::cast(vf));
    vf = vf->sender();
  }
  assert(vf->is_interpreted_frame() || vf->is_compiled_frame(), "Wrong frame type");
  chunk->push(compiledVFrame::cast(vf));

#ifndef PRODUCT
  if (TraceDeoptimization && Verbose) {
    tty->print_cr("[Deoptimization: vframes to store in vframeArray]");
    for (int index = 0; index < chunk->length(); index++) {
      chunk->at(index)->print_activation(index);      
      if (chunk->at(index)->is_compiled_frame()) {
        int bci = chunk->at(index)->raw_bci();
        tty->print_cr("     for bci %d ", bci);
      }
    }
  }
#endif

  // Register map for next frame (used for stack crawl).  We capture
  // the state of the deopt'ing frame's caller.  Thus if we need to
  // stuff a C2I adapter we can properly fill in the callee-save
  // register locations.
  frame caller = fr.sender(reg_map);

  // See if we are uncommon-trapping with a i2c or OSR adapter just above us.
  // In this case we want to throw away the i2c/OSR frame and just flip back 
  // to being interpreted.
  frame last_i2c_frame;
  CodeBlob *ccb = CodeCache::find_blob(caller.pc());
  if( ccb && (ccb->is_osr_adapter() || ccb->is_i2c_adapter()) ) {
    // Capture the I2C frame; this holds info we need to populate the last
    // interpreted frame with for when it unwinds.  This kills callee-save
    // values for the caller BUT the caller's caller was the interpreter
    // and he has no callee-save values.
    last_i2c_frame = caller.sender(reg_map);
#ifdef COMPILER2
    NOT_PRODUCT(OptoRuntime::_compress_i2c2i_ctr++;)
#endif
  }
  
  // Now the entire array of VFrames is collected into C heap storage.
  // We have to pass the results back to the original Java thread, and that
  // thread has to adjust it's own stack and so can have no local storage.
  // The C heap is the only way to pass info to the Java thread.
  int frame_size = caller.sp() - fr.sp();
  vframeArray* array = vframeArray::allocate(thread, frame_size, chunk, reg_map, last_i2c_frame );

  // Compare the vframeArray to the collected vframes
  assert(array->structual_compare(thread, chunk), "just checking");
  Events::log("# vframes = %d", chunk->length());

  return array;
}

// ==================================================================================
//
// Deoptimize a frame!  Means let the frame continue execution in any nested calls
// it is making, but upon return to the frame revert to interpreted mode.  This is
// done by patching the return PC, so returning to this frame instead returns to
// the DeoptimizationBlob. 
//
// HotSpot does "eager" deoptimization while Self did "lazy".  Lars Bak claims this
// must be so and my (CNC) best recollection is that it's because monitor's have to
// be migrated off the stack.  Migrating the monitors requires the VM thread and
// can block; upon returning from nested calls and hitting the DeoptimizationBlob
// we are NOT GC safe and cannot block - so we must eagerly migrate the monitors at
// stack-patching time.  Once we are eager for the monitors we might as well be
// eager for everything else.
//
// Thus we capture all interesting state of the frame into a vframeArray, including
// callee-save registers and all required interpreter state.  After patching the PC
// the frame is dead, Dead, DEAD!  Oops in it are NOT GC'd!  To help enforce this we
// fill the frame (that's right: we trash the stack contents) with 0xDEADB10B.  All
// references to the original frame are satisfied from the vframeArray instead.
//
void Deoptimization::deoptimize(JavaThread* thread, frame fr, RegisterMap *reg_map) {
#ifndef COMPILER2
  ShouldNotReachHere(); // deoptimization not yet implemented for C1
#else
  ResourceMark rm;
  assert(fr.can_be_deoptimized(), "checking frame type");

  NOT_PRODUCT(_nof_deoptimizations++); // For statistics
  EventMark m("Deoptimization (pc=0x%lx, sp=0x%lx)", fr.pc(), fr.sp());
  if (TraceDeoptimization) {
    tty->print_cr("Deoptimization: deoptimize frame (pc=0x%lx, sp=0x%lx)", fr.pc(), fr.sp());
    if (Verbose) {
      tty->print_cr("[PACKING]");
      tty->print(" deopt:  "); fr.print_on(tty);
    }
  }

  vframeArray* array = create_vframeArray(thread, fr, reg_map);
#ifndef PRODUCT
  if (TraceDeoptimization && Verbose) {
    tty->print_cr("[Deoptimization: vframes in vframeArray]");
    int count = 0;
    for (javaVFrame* vf = new deoptimizedVFrame(array); vf; vf = vf->java_sender()) {
      vf->print_activation(count++);
    }
  }
#endif // PRODUCT


#ifdef ASSERT
  // For debugging safety, trash the frame.  It's not GC'd in any case and
  // we want to catch people who are illegally peeking at the dead frame.
  int *junk = (int*)fr.sp();
  int *end = junk + array->frame_size();
#ifdef SPARC
  junk += 16;			// Skip a window
#endif
  // I attempted to trash the stack here (fill with 0xDEADB10B) as an
  // assert.  Alas, monitors are migrated off the stack lazily (why?)
  // so the locks remain "live" on the stack and I can't zap the stack.
  //  for( ; junk < end; junk++ )
  //  *junk = 0xDEADB10B;
#endif

  // Convert the frames we deoptimized into one deoptimized frame
  fr.patch_pc(thread, OptoRuntime::deoptimize_repack_stack());
  array->set_sp(fr.sp());

  if (TraceDeoptimization) {
    tty->print_cr(" - frames stored in vframeArray(0x%lx)", array);
  }

  // add array to thread
  thread->insert_vframe_array(array);

#endif // COMPILER2
}


void Deoptimization::load_class_by_index(constantPoolHandle constant_pool, int index, TRAPS) {
  // in case of an unresolved klass entry, load the class.
  if (constant_pool->tag_at(index).is_unresolved_klass()) {
    klassOop tk = constant_pool->klass_at(index, CHECK);
    return;
  }

  if (!constant_pool->tag_at(index).is_symbol()) return;

  Handle class_loader = instanceKlass::cast(constant_pool->pool_holder())->class_loader();
  symbolHandle symbol = constant_pool->symbol_at(index);

  // class name?
  if (symbol->byte_at(0) != '(') {
    oop protection_domain = Klass::cast(constant_pool->pool_holder())->protection_domain();
    SystemDictionary::resolve_or_null(symbol, class_loader, protection_domain, CHECK);
    return;
  }

  // then it must be a signature!
  for (SignatureStream ss(symbol()); !ss.is_done(); ss.next()) {
    if (ss.is_object()) {
      symbolHandle class_name = ss.as_symbol(CHECK);
      oop protection_domain = Klass::cast(constant_pool->pool_holder())->protection_domain();
      SystemDictionary::resolve_or_null(class_name, class_loader, protection_domain, CHECK);
    }
  }
}

void Deoptimization::load_class_by_index(constantPoolHandle constant_pool, int index) {
  EXCEPTION_MARK;
  load_class_by_index(constant_pool, index, THREAD);
  if (PENDING_EXCEPTION) {
    // Exception happened during classloading. We ignore the exception here, since it
    // is going to be rethrown since the current activation is going to be deoptimzied and
    // the interpreter will re-execute the bytecode. 
    CLEAR_PENDING_EXCEPTION;
  }
}

void Deoptimization::deoptimize_frame(JavaThread* thread, jint* sp) {
  // Compute frame and register map based on thread and sp.
  RegisterMap reg_map(thread);
  frame fr = thread->last_frame();
  while (fr.sp() != sp) {
    fr = fr.sender(&reg_map);
  }

  // Deoptimize only if the frame comes from compile code
  if (fr.is_compiled_frame()) {        
    ResourceMark rm;
    DeoptimizationMarker dm;    
    deoptimize(thread, fr, &reg_map);
  }
}



JRT_ENTRY(Deoptimization::UnrollBlock*, Deoptimization::uncommon_trap(JavaThread* thread, jint unloaded_class_index))
#ifndef COMPILER2
  ShouldNotReachHere(); // deoptimization not yet implemented for C1
  return NULL;
#else
  ResourceMark rm;
  HandleMark hm;
  
  RegisterMap reg_map(thread, false); // Do not need to update callee-saved registers
  frame stub_frame = thread->last_frame();
  frame fr = stub_frame.sender(&reg_map);

  // Make sure the calling nmethod is not getting deoptimized and removed 
  // before we are done with it.
  nmethodLocker nl(fr.pc());

  Events::log("Uncommon trap occurred @0x%lx, unloaded_class_index = %d", fr.pc(), unloaded_class_index);
  gather_uncommon_trap_statistics(unloaded_class_index);
  
  vframe * vf  = vframe::new_vframe(&fr, &reg_map, thread);
  nmethod* nm = compiledVFrame::cast(vf)->code();

#ifndef PRODUCT
  if (TraceDeoptimization) {
    ResourceMark rm;
    tty->print("Uncommon trap occurred in");
    nm->method()->print_short_name(tty);
    tty->print(" (@0x%lx) unloaded_class_index = %d, thread = %d", fr.pc(), unloaded_class_index, Thread::current()->osthread()->thread_id());
    if (unloaded_class_index >=0) {
      constantPoolHandle constants = compiledVFrame::cast(vf)->method()->constants();
      if (constants->tag_at(unloaded_class_index).is_unresolved_klass()) {
        symbolOop symbol = constants->unresolved_klass_at(unloaded_class_index);
        tty->print(" unresolved class: ");
        symbol->print_symbol_on(tty);
      } else if (constants->tag_at(unloaded_class_index).is_symbol()) {
        tty->print(" symbol: ");
        symbolOop symbol = constants->symbol_at(unloaded_class_index);
        symbol->print_symbol_on(tty);
      }
    }
    tty->cr();
  }
#endif // PRODUCT

  if (BailoutToInterpreterForThrows && unloaded_class_index == Deopt_athrow) {
    // Keep method interpreted from now on
    nm->make_not_entrant();
    nm->method()->set_not_compilable();    
  }

  constantPoolHandle constants;
  if (unloaded_class_index >= Deoptimization::Deopt_unloaded) {
    constants = compiledVFrame::cast(vf)->method()->constants();

    // Reset invocation counter for outer most method. 
    ScopeDesc* sd = compiledVFrame::cast(vf)->scope();
    while (!sd->is_top()) sd = sd->sender();
    sd->method()->invocation_counter()->reset();
  }

  // Load class is necessary 
  if (constants.not_null()) {
    load_class_by_index(constants, unloaded_class_index);
  }

  if (UseSpecLock) {
    // Only mucks with this thread, so should not need a safepoint
    if (PrintSpecLock) { tty->print_cr("fixup all objects for deoptimize"); }
    SpecLock::fixup_all_objects_on_stack(thread);
  }

  VM_DeoptimizeFrame deopt(thread, fr.sp());
  VMThread::execute(&deopt);
  // An exception might have been installed as this 
  // point, e.g., ThreadDeath

  { // %%% temporary kludge, until further work puts a counter in the methodOop
    static int unexpected_klass_deopts = 0;
    if (unloaded_class_index == Deoptimization::Deopt_unexpected_klass) {
      if (++unexpected_klass_deopts >= 40)
	UseTypeProfile = false;
    }
  }

  // Make them not entrant, so next time they are called they get
  // recompiled.  Unloaded classes are loaded now so recompile before next
  // time they are called.  Same for uninitialized.  The interpreter will
  // load the array class.
  if( unloaded_class_index == Deoptimization::Deopt_uninitialized ||
      unloaded_class_index == Deoptimization::Deopt_unloaded_array_class ||
      unloaded_class_index == Deoptimization::Deopt_unexpected_klass ||
      unloaded_class_index >= Deoptimization::Deopt_unloaded ) {
    nm->make_not_entrant();
    if (TraceDeoptimization) {
      tty->print_cr(" - turned (nmethod*)0x%lx into a non_entrant nmethod",nm);
    }
  }

  // If we see an unexpected null at a check-cast we record it and force a
  // recompile; the offending check-cast will be compiled to handle NULLs.
  // If we see more than one offending BCI, then all checkcasts in the
  // method will be compiled to handle NULLs.
  if( unloaded_class_index == Deoptimization::Deopt_unexpected_null_cast ) {
    ScopeDesc* sd = compiledVFrame::cast(vf)->scope();
    int fail_bci = sd->bci();
    int cache_bci = sd->method()->null_cast_cache();
    if( cache_bci == methodOopDesc::Null_Cast_Cache_Empty ) { // Cache is empty
      sd->method()->set_null_cast_cache( fail_bci );
      cache_bci = fail_bci;
    } else if( cache_bci == fail_bci ) { // 2nd hit at same BCI
    } else {                    // Cache has many failures?
      sd->method()->set_null_cast_cache( methodOopDesc::Null_Cast_Cache_Full );
    }
    nm->make_not_entrant();
  }
  
  return fetch_unroll_info_helper(thread);
#endif // COMPILER2
JRT_END


//-----------------------------------------------------------------------------
#ifndef PRODUCT

long Deoptimization::_nof_deoptimizations        = 0;
long Deoptimization::_nof_unloaded               = 0;
long Deoptimization::_nof_athrow                 = 0;
long Deoptimization::_nof_null_check             = 0;
long Deoptimization::_nof_div0_check             = 0;
long Deoptimization::_nof_negative_array_length  = 0;
long Deoptimization::_nof_range_check            = 0;
long Deoptimization::_nof_cast_check             = 0;
long Deoptimization::_nof_uninitialized          = 0;
long Deoptimization::_nof_unloaded_array_class   = 0;
long Deoptimization::_nof_unexpected_klass       = 0;
long Deoptimization::_nof_unexpected_null_cast   = 0;

void Deoptimization::gather_uncommon_trap_statistics(int unloaded_class_index) {
  switch(unloaded_class_index) {
     case Deopt_athrow:                 ++_nof_athrow;                break;
     case Deopt_null_check:             ++_nof_null_check;            break;
     case Deopt_div0_check:             ++_nof_div0_check;            break;
     case Deopt_negative_array_length:  ++_nof_negative_array_length; break;
     case Deopt_range_check:            ++_nof_range_check;           break;
     case Deopt_cast_check:             ++_nof_cast_check;            break;
     case Deopt_uninitialized:          ++_nof_uninitialized;         break;
     case Deopt_unloaded_array_class:   ++_nof_unloaded_array_class;  break;
     case Deopt_unexpected_klass:       ++_nof_unexpected_klass;      break;
     case Deopt_unexpected_null_cast:   ++_nof_unexpected_null_cast;  break;
     default:
       assert(unloaded_class_index >= 0, "wrong index");
       _nof_unloaded++;
       break;
  }
}
  
int Deoptimization::print_stat_line(const char *name, long r) { 
  assert(_nof_deoptimizations != 0, "sanity check");
  if (r != 0) {
    tty->print_cr("  %s: %Ld (%d%%)", name, r, (r * 100) / _nof_deoptimizations);
  }
  return 0;
}

void Deoptimization::print_statistics() {
  if (_nof_deoptimizations != 0) {
    tty->print_cr("Total number of deoptimizations: %Ld", _nof_deoptimizations);
    print_stat_line("unloaded classes"       , _nof_unloaded);
    print_stat_line("athrow"                 , _nof_athrow);
    print_stat_line("null check"             , _nof_null_check);
    print_stat_line("div0 check"             , _nof_div0_check);
    print_stat_line("array length"           , _nof_negative_array_length);
    print_stat_line("range_check"            , _nof_range_check);
    print_stat_line("checkcast"              , _nof_cast_check);
    print_stat_line("uninitialized"          , _nof_uninitialized);
    print_stat_line("unloaded array classes" , _nof_unloaded_array_class);
    print_stat_line("unexpected classes"     , _nof_unexpected_klass);
    print_stat_line("unexpected null cast"   , _nof_unexpected_null_cast);
  }
}


#endif

