/*
 * @(#)deoptimization.hpp	1.48 00/03/21 16:17:48
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

class Deoptimization : AllStatic {
 public:
  // Interface for staring deoptimization
   enum DeoptInfo {
     Deopt_unloaded              =   0, // positive values describe a constant pool index that needs to be loaded
     Deopt_athrow                =  -1, // Bytecode _athrow,             do not recompile 
     Deopt_null_check            =  -2, // failed null check,            do not recompile
     Deopt_div0_check            =  -3, // failed div0 check,            do not recompile
     Deopt_negative_array_length =  -4, // failed check,                 do not recompile
     Deopt_range_check           =  -5, // failed range check,           do not recompile
     Deopt_cast_check            =  -6, // failed optimized checkcast,   do not recompile
     Deopt_uninitialized         =  -7, // uninitialized class,    recompile after interpreter executes
     Deopt_unloaded_array_class  =  -8, // array-class not loaded, recompile after interpreter loads
     Deopt_unexpected_klass      =  -9, // type prediction failure, invalidate profile and recompile
     Deopt_unexpected_null_cast  = -10  // cast-never-null failed; record failure and recompile
   };

   
   enum UnpackType {
     Unpack_deopt                = 0, // normal deoptimization, use pc computed in unpack_vframe_on_stack
     Unpack_exception            = 1, // exception is pending
     Unpack_uncommon_trap        = 2  // redo last byte code
   };

  // Checks all compiled methods. Invalid methods are deleted and
  // corresponding activations are deoptimized.
  static int deoptimize_dependents();
  
  // Deoptimizes a frame into a deoptimized frame.
  static void deoptimize(JavaThread* thread, frame fr, RegisterMap *reg_map);
  static vframeArray* create_vframeArray(JavaThread* thread, frame fr, RegisterMap *reg_map);
 
  // Interface used for unpacking deoptimized frames 

  // UnrollBlock is returned by the 
  class UnrollBlock : public CHeapObj {
   private:
    jint   _size_of_deoptimized_frame; // Size of current deoptimized frame
    jint   _pop_i2c_frame;	// True if popping I2C frame
    jint   _adjust_i2c_frame;	// Adjustment to SP for popping I2C frame
    jint   _number_of_frames;	// Number frames to unroll 
    jint*  _frame_sizes;	// Array of frame sized for unrolling the stack
    jint*  _register_block;	// Block for storing callee-saved registers.
    jint   _returns_double;	// Tells if we have to restore double return value
   public:
    // Constructor
    UnrollBlock(jint  size_of_deoptimized_frame,
		jint  pop_i2c_frame,
                jint  adjust_i2c_frame,
                jint  number_of_frames,
                jint* frame_sizes,
                bool  returns_double);
    ~UnrollBlock();

    // Returns where a register is located.
    jint* value_addr_at(int register_number) const;

    // Accessors
    jint* frame_sizes() const { return _frame_sizes; }
    jint number_of_frames() const { return _number_of_frames; }
   
    // Returns the total size of frames
    jint size_of_frames() const;

    // Accessors used by the code generator for the unpack stub.
    static int size_of_deoptimized_frame_offset_in_bytes() { return (int)&((UnrollBlock*)NULL)->_size_of_deoptimized_frame; }
    static int pop_i2c_frame_offset_in_bytes()             { return (int)&((UnrollBlock*)NULL)->_pop_i2c_frame;             }
    static int adjust_i2c_frame_offset_in_bytes()          { return (int)&((UnrollBlock*)NULL)->_adjust_i2c_frame;          }
    static int number_of_frames_offset_in_bytes()          { return (int)&((UnrollBlock*)NULL)->_number_of_frames;          }
    static int frame_sizes_offset_in_bytes()               { return (int)&((UnrollBlock*)NULL)->_frame_sizes;               }
    static int register_block_offset_in_bytes()            { return (int)&((UnrollBlock*)NULL)->_register_block;            }  
    static int returns_double_offset_in_bytes()            { return (int)&((UnrollBlock*)NULL)->_returns_double;            }

    void print();
  };

  //** Returns an UnrollBlock continuing information 
  // how to make room for the resulting interpreter frames. 
  // Called by assembly stub after execution has returned to 
  // deoptimized frame.
  // @argument thread.     Thread where stub_frame resides.
  // @see OptoRuntime::deoptimization_fetch_unroll_info_C
  static UnrollBlock* fetch_unroll_info(JavaThread* thread);

  //** Unpacks vframeArray onto execution stack.
  // Called by assembly stub after execution has returned to
  // deoptimized frame and after the stack unrolling.
  // @argument thread.     Thread where stub_frame resides.
  // @argument exec_mode.  Determines how execution should be continuted in top frame.
  //                       0 means continue after current byte code
  //                       1 means exception has happened, handle exception
  //                       2 means reexecute current bytecode (for uncommon traps).
  // @see OptoRuntime::deoptimization_unpack_frames_C
  static void unpack_frames(JavaThread* thread, int exec_mode);

  //** Performs an uncommon trap for compiled code.
  // The top most compiler frame is converted into interpreter frames
  static UnrollBlock* uncommon_trap(JavaThread* thread, jint unloaded_class_index);

  //** Deoptimizes the frame identified by sp.
  // Only called from VMDeoptimizeFrame
  // @argument thread.     Thread where stub_frame resides.
  // @argument sp.         Stack pointer for frame that should be deoptimized.
  static void deoptimize_frame(JavaThread* thread, jint* sp);

  // Statistics
  static int  print_stat_line(const char *name, long r)                  PRODUCT_RETURN0;
  static void gather_uncommon_trap_statistics(int unloaded_class_index)  PRODUCT_RETURN;
  static void print_statistics()                                         PRODUCT_RETURN;  

  // How much room to adjust the last frame's SP by, to make space for
  // the callee's interpreter frame (which expects locals to be next to
  // incoming arguments)
  static int last_frame_adjust(int callee_parameters, int callee_locals);

 private:
  // class loading support for uncommon trap
  static void load_class_by_index(constantPoolHandle constant_pool, int index, TRAPS);
  static void load_class_by_index(constantPoolHandle constant_pool, int index);

  static UnrollBlock* fetch_unroll_info_helper(JavaThread* thread);

#ifndef PRODUCT
  static long _nof_deoptimizations;       // Total count
  static long _nof_unloaded;
  static long _nof_athrow;
  static long _nof_null_check;
  static long _nof_div0_check;
  static long _nof_negative_array_length;
  static long _nof_range_check;
  static long _nof_cast_check;
  static long _nof_uninitialized;
  static long _nof_unloaded_array_class;
  static long _nof_unexpected_klass;
  static long _nof_unexpected_null_cast;
#endif
};

class DeoptimizationMarker : StackObj {  // for profiling
  static bool _is_active;  
public:
  DeoptimizationMarker()  { _is_active = true; }
  ~DeoptimizationMarker() { _is_active = false; }
  static bool is_active() { return _is_active; }
};


