/*
 * @(#)frame.cpp	1.156 01/04/21 22:47:11
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_frame.cpp.incl"

RegisterMap::RegisterMap(JavaThread *thread, bool update_map) {
  _thread         = thread;
  _update_map     = update_map;      
  clear(thread->not_at_call_sp());  
  debug_only(_update_for_sp = NULL;)
}

RegisterMap::RegisterMap(const RegisterMap* map) {
  assert(map != NULL, "RegisterMap must be present");
  _thread                = map->thread();
  _update_map            = map->update_map();
  _include_argument_oops = map->include_argument_oops();  
  _not_at_call_sp        = map->_not_at_call_sp;
  debug_only(_update_for_sp = map->_update_for_sp;)
#ifndef CORE
  pd_initialize_from(map);
  if (update_map()) {  
    for(int i = 0; i < (reg_count+BitsPerWord-1)/BitsPerWord; i++) {
      unsigned long bits = map->_location_valid[i];
      _location_valid[i] = bits;
      // for whichever bits are set, pull in the corresponding map->_location
      int j = i*BitsPerWord;
      while (bits != 0) {
        if ((bits & 1) != 0)
          _location[j] = map->_location[j];
        bits >>= 1;
        j += 1;
      }
    }
  }
#endif
}

void RegisterMap::clear(jint* not_at_call_sp) {  
  set_include_argument_oops(true);
  _not_at_call_sp = not_at_call_sp;
  debug_only(_update_for_sp = NULL;)
#ifndef CORE
  if (_update_map) {
    for(int i = 0; i < (reg_count+BitsPerWord-1)/BitsPerWord; i++) {
      _location_valid[i] = 0;
    }
    pd_clear();
  } else {
    pd_initialize();
  }
#endif
}

#ifndef PRODUCT

void RegisterMap::print_on(outputStream* st) const {
#ifndef CORE
  st->print_cr("Register map");
  for(int i = 0; i < reg_count; i++) {
    
    jint* src = (jint*) location(VMReg::Name(i));
    if (src != NULL) {
      tty->print_cr("  %s [0x%lx] = 0x%lx", SharedInfo::regName[i], src, *src);
    }
  }
#endif
}

void RegisterMap::print() const {
  print_on(tty);
}

#endif

// type testers

bool frame::is_first_frame() const {
  return is_entry_frame() && entry_frame_is_first(); 
}


bool frame::is_interpreted_frame() const  {
  return Interpreter::contains(pc());
}

bool frame::is_native_frame() const {
#ifndef CORE
  CodeBlob* cb = CodeCache::find_blob(pc());
  return (cb != NULL && cb->is_native_method());    
#else
  return false;
#endif
}

bool frame::is_entry_frame() const {
  return StubRoutines::returns_to_call_stub(pc());
}

bool frame::is_java_frame() const {
  if (is_interpreted_frame()) return true;
#ifndef CORE
  if (is_deoptimized_frame()) return true;
  if (is_compiled_frame())    return true;
#endif
  return false;
}

#ifdef COMPILER1
bool frame::is_deoptimized_frame() const {
  return false;
}
#endif

#ifdef COMPILER2
bool frame::is_deoptimized_frame() const {
  if (OptoRuntime::deoptimization_blob() == NULL) return false;
  return OptoRuntime::deoptimization_blob()->contains(pc());
}
#endif

#ifndef CORE
bool frame::is_compiled_frame() const {
  CodeBlob* cb = CodeCache::find_blob(pc());
  return (cb != NULL && cb->is_java_method());  
}


bool frame::is_c2runtime_frame() const {
  CodeBlob* cb = CodeCache::find_blob(pc());
  return (cb != NULL && cb->is_runtime_stub());
}

bool frame::is_c2i_frame() const {
  CodeBlob* cb = CodeCache::find_blob(pc());
  return (cb != NULL && cb->is_c2i_adapter());
}

bool frame::is_i2c_frame() const {
  CodeBlob* cb = CodeCache::find_blob(pc());
  return (cb != NULL && cb->is_i2c_adapter());
}

bool frame::is_safepoint_blob_frame() const {
  CodeBlob* cb = CodeCache::find_blob(pc());
  return (cb != NULL && cb->is_safepoint_stub());
}

bool frame::is_osr_adapter_frame() const {
#ifdef COMPILER1
  return Runtime1::returns_to_osr_adapter(pc());
#endif
#ifdef COMPILER2
  CodeBlob* cb = CodeCache::find_blob(pc());
  return (cb != NULL && cb->is_osr_adapter());
#endif
}

bool frame::is_glue_frame() const {
  CodeBlob* cb = CodeCache::find_blob(pc());
  if (cb == NULL ) return false;
       if (cb->is_i2c_adapter()) return true;
  else if (cb->is_c2i_adapter()) return true; 
  else if (cb->is_runtime_stub()) return true;
  else return false;
}
#endif // CORE


// testers

bool frame::is_first_java_frame() const {
  RegisterMap map(JavaThread::current(), false); // No update
  frame s;
  for (s = sender(&map); !(s.is_java_frame() || s.is_first_frame()); s = s.sender(&map));
  return s.is_first_frame();
}


bool frame::entry_frame_is_first() const {
  return entry_frame_call_wrapper()->last_Java_sp() == NULL;
}


bool frame::should_be_deoptimized() const {
#ifndef CORE
  if (!is_compiled_frame()) return false;
  CodeBlob *cb = CodeCache::find_blob(pc());
  assert(cb != NULL && cb->is_nmethod(), "must be an nmethod");
  nmethod* nm = (nmethod *)cb;
  if (TraceDependencies) {
    tty->print("checking (%s) ", nm->is_marked_for_deoptimization() ? "true" : "false"); 
    nm->print_value_on(tty);
    tty->cr();
  }
  return nm->is_marked_for_deoptimization();
#else
  return false;
#endif
}

bool frame::can_be_deoptimized() const {
#ifndef CORE
  if (!is_compiled_frame()) return false;
  nmethod* nm = CodeCache::find_nmethod(pc());  
  return nm->can_be_deoptimized();
#else
  return false;
#endif
}

frame frame::java_sender() const {
  RegisterMap map(JavaThread::current(), false);
  frame s;
  for (s = sender(&map); !(s.is_java_frame() || s.is_first_frame()); s = s.sender(&map)) ;
  guarantee(s.is_java_frame(), "tried to get caller of first java frame");
  return s;
}

frame frame::real_sender(RegisterMap* map) const {
#ifndef CORE
  frame result = sender(map);
  while (result.is_glue_frame()) {
    result = result.sender(map);
  }
  return result;
#else
  return sender(map);
#endif
}


// Interpreter frames

 
void frame::interpreter_frame_set_locals(jint* locs)  { 
  *interpreter_frame_locals_addr() = locs; 
}

methodOop frame::interpreter_frame_method() const {
  assert(is_interpreted_frame(), "interpreted frame expected");
  methodOop m = *interpreter_frame_method_addr();
  assert(m->is_perm(), "bad methodOop in interpreter frame");
  assert(m->is_method(), "not a methodOop");
  return m;
}

void frame::interpreter_frame_set_method(methodOop method) {
  assert(is_interpreted_frame(), "interpreted frame expected");
  *interpreter_frame_method_addr() = method;
}

// Note: The bcx usually contains the bcp; however during GC it contains the bci
//       (changed by gc_prologue() and gc_epilogue()) to be methodOop position
//       independent. These accessors make sure the correct value is returned
//       by testing the range of the bcx value. bcp's are guaranteed to be above
//       max_method_code_size, since methods are always allocated in OldSpace and
//       Eden is allocated before OldSpace.
//
//       The bcp is accessed sometimes during GC for ArgumentDescriptors; than
//       the correct translation has to be performed (was bug).

static bool is_bci(jint bcx) {
  return 0 <= bcx && bcx <= max_method_code_size;
}

jint frame::interpreter_frame_bci() const {
  assert(is_interpreted_frame(), "interpreted frame expected");
  jint bcx = interpreter_frame_bcx();
  return is_bci(bcx) ? bcx : interpreter_frame_method()->bci_from((address)bcx);
}

void frame::interpreter_frame_set_bci(jint bci) {
  assert(is_interpreted_frame(), "interpreted frame expected");
  assert(!is_bci(interpreter_frame_bcx()), "should not set bci during GC");
  interpreter_frame_set_bcx((jint)interpreter_frame_method()->bcp_from(bci));
}

address frame::interpreter_frame_bcp() const {
  assert(is_interpreted_frame(), "interpreted frame expected");
  jint bcx = interpreter_frame_bcx();
  return is_bci(bcx) ? interpreter_frame_method()->bcp_from(bcx) : (address)bcx;
}

void frame::interpreter_frame_set_bcp(address bcp) {
  assert(is_interpreted_frame(), "interpreted frame expected");
  assert(!is_bci(interpreter_frame_bcx()), "should not set bcp during GC");
  interpreter_frame_set_bcx((jint)bcp);
}

BasicObjectLock* frame::next_monitor_in_interpreter_frame(BasicObjectLock* current) const { 
#ifdef ASSERT
  interpreter_frame_verify_monitor(current);
#endif
  BasicObjectLock* next = (BasicObjectLock*) (((int*) current) + interpreter_frame_monitor_size());
  return next;
}

BasicObjectLock* frame::previous_monitor_in_interpreter_frame(BasicObjectLock* current) const { 
#ifdef ASSERT
//   // This verification needs to be checked before being enabled
//   interpreter_frame_verify_monitor(current);
#endif
  BasicObjectLock* previous = (BasicObjectLock*) (((int*) current) - interpreter_frame_monitor_size());
  return previous;
}


// (frame::interpreter_frame_sender_sp accessor is in frame_<arch>.cpp)

const char* frame::print_name() const {
  if (is_native_frame())      return "Native";
  if (is_interpreted_frame()) return "Interpreted";
#ifndef CORE
  if (is_deoptimized_frame()) return "Deoptimized";
  if (is_compiled_frame())    return "Compiled";
#endif
  if (_sp == NULL)            return "Empty";
  return "C";
}

void frame::print_value_on(outputStream* st, JavaThread *thread) const {
  NOT_PRODUCT(address begin = pc()-40;)
  NOT_PRODUCT(address end   = pc()+40;)

  st->print("%s frame (sp=%#lx", print_name(), sp());
  if (sp() != NULL)
    st->print(", fp=%#lx, pc=%#lx", fp(), pc());

  if (StubRoutines::contains(pc())) {
    StubCodeDesc* desc = StubCodeDesc::desc_for(pc());
    st->print("~Stub::%s", desc->name());
    NOT_PRODUCT(begin = desc->begin(); end = desc->end();)
  } else if (Interpreter::contains(pc())) {
    InterpreterCodelet* desc = Interpreter::codelet_containing(pc());
    if (desc != NULL) {
      st->print("~");
      desc->print();
      NOT_PRODUCT(begin = desc->code_begin(); end = desc->code_end();)
    } else {
      st->print("~interpreter"); 
    }
  }
  st->print_cr(")");

#ifndef CORE
  CodeBlob* cb = CodeCache::find_blob(pc());
  if (cb != NULL) {
    st->print("     ");
    const DeoptimizationBlob *db = cb->as_deoptimization_stub();
    if( db )
      db->print_value_on_fancy(st,thread,*this);
    else
      cb->print_value_on(st);
    st->cr();
    NOT_PRODUCT(begin = cb->instructions_begin(); end = cb->instructions_end();)
  }
#endif
  NOT_PRODUCT(if (WizardMode && Verbose) Disassembler::decode(begin, end);)
}


void frame::print_on(outputStream* st) const {
  print_value_on(st,NULL);
  if (is_interpreted_frame()) {
    interpreter_frame_print_on(st);
  }
}


void frame::interpreter_frame_print_on(outputStream* st) const {
#ifndef PRODUCT
  for (jint i = interpreter_frame_expression_stack_size() - 1;
       i >- 0;
       --i ) {
    jint x = interpreter_frame_expression_stack_at(i);
    st->print(" - stack  [%#lx]", x);
    st->fill_to(23);
    st->print_cr("; #%d", i);
  }
  // locks for synchronization
  for (BasicObjectLock* current = interpreter_frame_monitor_end();
       current < interpreter_frame_monitor_begin();
       current = next_monitor_in_interpreter_frame(current)) {
    st->print_cr(" [ - obj ");
    current->obj()->print_value_on(st);
    st->cr();
    st->print_cr(" - lock ");
    current->lock()->print_on(st);
    st->cr();
  }
  // monitor
  st->print_cr(" - monitor[%#lx]", interpreter_frame_monitor_begin());
  // bcp
  st->print(" - bcp    [%#lx]", interpreter_frame_bcp());
  st->fill_to(23);
  st->print_cr("; @%d", interpreter_frame_bci());
  // locals
  st->print_cr(" - locals [%#lx]", &interpreter_frame_local_at(0));
  // method
  st->print(" - method [%#lx]", interpreter_frame_method());
  st->fill_to(23);
  st->print("; ");
  interpreter_frame_method()->print_name(st);
  st->cr();
#endif
}


/*
  The interpreter_frame_expression_stack_at method in the case of SPARC needs the
  max_stack value of the method in order to compute the expression stack address.
  It uses the methodOop in order to get the max_stack value but during GC this
  methodOop value saved on the frame is changed by reverse_and_push and hence cannot
  be used. So we save the max_stack value in the FrameClosure object and pass it
  down to the interpreter_frame_expression_stack_at method
*/
class InterpreterFrameClosure : public OffsetClosure {
 private:
  frame* _fr;
  void   (*_f)(oop*);
  int    _max_locals;
  int    _max_stack;

 public:
  InterpreterFrameClosure(frame* fr, int max_locals, int max_stack, void f(oop*)) {
    _fr         = fr;
    _max_locals = max_locals;
    _max_stack  = max_stack;
    _f          = f;
  }

  void offset_do(int offset) {
    oop* addr;
    if (offset < _max_locals) {
      addr = (oop*) &_fr->interpreter_frame_local_at(offset);
      assert((jint*)addr >= _fr->sp(), "must be inside the frame");
      _f(addr);
    } else {
      addr = (oop*) &_fr->interpreter_frame_expression_stack_at((offset - _max_locals));
      // In case of exceptions, the expression stack is invalid and the esp will be reset to express
      // this condition. Therefore, we call f only if addr is 'inside' the stack (i.e., addr >= esp for Intel).
      bool in_stack;
      if (frame::interpreter_frame_expression_stack_direction() > 0) {
	in_stack = (jint*)addr <= _fr->interpreter_frame_tos_address();
      } else {
	in_stack = (jint*)addr >= _fr->interpreter_frame_tos_address();
      }
      if (in_stack) {
	_f(addr);
      }
    }
  }

  int max_locals()  { return _max_locals; }
  frame* fr()       { return _fr; }
};


class ArgumentOopFinder: public SignatureInfo {
 private:
  void   (*_f)(oop *); // Method to call
  int    _offset;      // TOS-relative offset, decremented with each argument
  bool   _is_static;   // true if the callee is a static method
  frame* _fr;

  void set(int size, BasicType type) {
    _offset -= size;
    if (type == T_OBJECT || type == T_ARRAY) oop_offset_do();    
  }

  void oop_offset_do() {
    oop* addr = (oop*)&_fr->interpreter_frame_tos_at(_offset);
    _f(addr);
  }

 public:
  ArgumentOopFinder(symbolOop signature, bool is_static, frame* fr, void f(oop*)) : SignatureInfo(signature) {
    // compute size of arguments
    int args_size = ArgumentSizeComputer(signature).size() + (is_static ? 0 : 1);
    assert(!fr->is_interpreted_frame() ||     
            args_size <= fr->interpreter_frame_expression_stack_size(), "args cannot be on stack anymore");
    // initialize ArgumentOopFinder
    _f         = f;
    _fr        = fr;    
    _offset    = args_size;
    _is_static = is_static;
  }

  void oops_do() {
    if (!_is_static) {
      --_offset;
      oop_offset_do();
    }
    iterate_parameters();
  }
};


// Entry frame has following form (n arguments)
//         +-----------+
//   sp -> |  last arg |
//         +-----------+
//         :    :::    :
//         +-----------+
// (sp+n)->|  first arg|
//         +-----------+

 

// visits and GC's all the arguments in entry frame
class EntryFrameOopFinder: public SignatureInfo {
 private:
  bool   _is_static;
  int    _offset;
  frame* _fr;
  void   (*_f)(oop*);

  void set(int size, BasicType type) {
    assert (_offset >= 0, "illegal offset");
    if (type == T_OBJECT || type == T_ARRAY) oop_at_offset_do(_offset);
    _offset -= size;
  }

  void oop_at_offset_do(int offset) {
    assert (offset >= 0, "illegal offset")
    oop* addr = (oop*) _fr->entry_frame_argument_at(offset);
    _f(addr);
  }

 public:
   EntryFrameOopFinder(frame* frame, symbolOop signature, bool is_static) : SignatureInfo(signature) {
     _f = NULL; // will be set later
     _fr = frame;
     _is_static = is_static;
     _offset = ArgumentSizeComputer(signature).size() - 1; // last parameter is at index 0
   }

  void arguments_do(void f(oop*)) {
    _f = f;
    if (!_is_static) oop_at_offset_do(_offset+1); // do the receiver
    iterate_parameters();
  }

};

oop* frame::interpreter_callee_receiver_addr(symbolHandle signature) {
  ArgumentSizeComputer asc(signature);
  int size = asc.size();  
  return (oop *)&interpreter_frame_tos_at(size); 
}


void frame::oops_interpreted_do(void f(oop*), const RegisterMap* map) {
  assert(map != NULL, "map must be set");
  methodOop m   = interpreter_frame_method();
  jint      bci = interpreter_frame_bci();

  assert(Universe::is_heap(m), "must be valid oop");
  assert(m->is_method(), "checking frame value");
  assert(m->is_native() && bci == 0 || bci >= 0 && bci < m->code_size(), "invalid bci value");

  // Handle the monitor elements in the activation
  for (
    BasicObjectLock* current = interpreter_frame_monitor_end();
    current < interpreter_frame_monitor_begin();
    current = next_monitor_in_interpreter_frame(current)
  ) {
#ifdef ASSERT
    interpreter_frame_verify_monitor(current);
#endif    
    current->oops_do(f);
  }

  // process fixed part
  f((oop*)interpreter_frame_method_addr());
  f((oop*)interpreter_frame_cache_addr());

  if (m->is_native() && m->is_static()) {
    f((oop*)( fp() + interpreter_frame_mirror_offset ));
  }

  int max_locals = m->is_native() ? m->size_of_parameters() : m->max_locals();
  InterpreterFrameClosure blk(this, max_locals, m->max_stack(), f);
  
  // process locals & expression stack
  OopMapCacheEntry *mask = m->mask_for(bci);
  mask->iterate_oop(&blk);

  // process a callee's arguments if we are at a call site
  // (i.e., if we are at an invoke bytecode)  
  if (map->include_argument_oops() && !m->is_native()) {
    ResourceMark rm;
    Bytecode_invoke* call = Bytecode_invoke_at_check(m, bci);
    if (call != NULL && interpreter_frame_expression_stack_size() > 0) {
      // we are at a call site & the expression stack is not empty
      // => process callee's arguments
      //
      // Note: The expression stack can be empty if an exception
      //       occured during method resolution/execution. In all
      //       cases we empty the expression stack completely be-
      //       fore handling the exception (the exception handling
      //       code in the interpreter calls a blocking runtime
      //       routine which can cause this code to be executed).
      //       (was bug gri 7/27/98)      
      oops_interpreted_arguments_do(call->signature(), call->is_invokestatic(), f);      
    }
  }
}

void frame::oops_interpreted_arguments_do(symbolOop signature, bool is_static, void f(oop*)) {  
  ArgumentOopFinder finder(signature, is_static, this, f);
  finder.oops_do();
}

void frame::oops_code_blob_do(void f(oop*), const RegisterMap* reg_map) {
#ifndef CORE
  CodeBlob* cb = CodeCache::find_blob(pc());
  assert(cb != NULL, "sanity check");
  if (cb->oop_maps() != NULL) {
    OopMapSet::oops_do(this, cb, reg_map, f);

    // Preserve potential arguments for a callee. We handle this by dispatching
    // on the codeblob. For c2i, we do
    if (reg_map->include_argument_oops()) {
      cb->preserve_callee_argument_oops(*this, reg_map, f);      
    }
  }

  // If we see an activation belonging to a non_entrant nmethod, we mark it.
  if (cb->is_nmethod() && ((nmethod *)cb)->is_not_entrant()) {
    ((nmethod*)cb)->mark_as_seen_on_stack();
  }
#endif
}

#ifndef CORE
class CompiledArgumentOopFinder: public SignatureInfo {
 protected:
  void          (*_f)(oop*);  
  int           _offset;        // the current offset, incremented with each argument
  bool          _is_static;     // true if the callee is a static method
  frame         _fr;
  RegisterMap*  _reg_map;      
  int           _arg_size;
  COMPILER2_ONLY(uint16* _regs;)          // VMReg::Name list of arguments
  
  void set(int size, BasicType type) {
    if (type == T_OBJECT || type == T_ARRAY) handle_oop_offset();
    _offset += size;
  }

  virtual void handle_oop_offset() {

#if defined(COMPILER1) && !defined(SPARC)
    ShouldNotReachHere(); // uses win32 C1 uses ArgumentOopFinder, should consolidate
#endif

    COMPILER2_ONLY(VMReg::Name reg = VMReg::Name(_regs[_offset]);)
    COMPILER1_ONLY(VMReg::Name reg = VMReg::Name(_offset);)
    oop *loc = _fr.pd_compiled_argument_to_location(reg, _reg_map, _arg_size);
    _f(loc); 
  }

 public:
  CompiledArgumentOopFinder(symbolOop signature, bool is_static, void f(oop*), frame fr,  const RegisterMap* reg_map) 
    : SignatureInfo(signature) {

    // initialize CompiledArgumentOopFinder
    _f         = f;
    _offset    = 0;
    _is_static = is_static;    
    _fr        = fr;    
    _reg_map   = (RegisterMap*)reg_map;
    _arg_size  = ArgumentSizeComputer(signature).size() + (is_static ? 0 : 1);

#ifdef COMPILER2
    int arg_size;
    _regs = Matcher::find_callee_arguments(signature, is_static, &arg_size);
    assert(arg_size == _arg_size, "wrong arg size");
#endif
  }

  void oops_do() {
    if (!_is_static) {
      handle_oop_offset(); 
      _offset++;
    }
    iterate_parameters();
  }  
};

void frame::oops_compiled_arguments_do(symbolOop signature, bool is_static, const RegisterMap* reg_map, void f(oop*)) {
  COMPILER2_ONLY(ResourceMark rm);
  CompiledArgumentOopFinder finder(signature, is_static, f, *this, reg_map);
  finder.oops_do();
}
#endif // !CORE


#ifndef CORE
oop* frame::oopmapreg_to_location(VMReg::Name reg, const RegisterMap* reg_map) const {
  if(reg < SharedInfo::stack0) {    
    // If it is passed in a register, it got spilled in the stub frame.  
    return (oop *)reg_map->location(reg);
  } else {			        
    int sp_offset = reg - SharedInfo::stack0 + pd_oop_map_offset_adjustment();
    return (oop *)&sp()[sp_offset];
  }
}
#endif

void frame::oops_entry_do(void f(oop*), const RegisterMap* map) {
  assert(map != NULL, "map must be set");
  if (map->include_argument_oops()) {
    // must collect argument oops, as nobody else is doing it
    methodHandle m = entry_frame_call_wrapper()->callee_method();
    EntryFrameOopFinder finder(this, m->signature(), m->is_static());
    finder.arguments_do(f);
  }
  // Traverse the Handle Block saved in the entry frame
  entry_frame_call_wrapper()->oops_do(f);
}


void frame::oops_do(void f(oop*), RegisterMap* map) {
         if (is_interpreted_frame())    { oops_interpreted_do(f, map);
  } else if (is_entry_frame())          { oops_entry_do      (f, map);
#ifdef COMPILER1
  // Note: this test has to come before the CodeCache::contains(pc())
  //       check since the code for osr adapter frames is contained
  //       in the code cache, too!
  } else if (is_osr_adapter_frame())    { // nothing to do
#endif
#ifndef CORE  
  } else if (CodeCache::contains(pc())) { oops_code_blob_do  (f, map);
#endif
  } else { ShouldNotReachHere();
  }  
}


void frame::follow_roots(RegisterMap* map) {
         if (is_interpreted_frame())   { oops_interpreted_do(&MarkSweep::mark_and_push, map);
  } else if (is_entry_frame())         { oops_entry_do      (&MarkSweep::mark_and_push, map);
#ifndef CORE
  } else if (is_deoptimized_frame()) {
    // We have to reverse the vframe array and all contained oops.
    // The latter part requires valid methodOops and oop maps, so
    // this has to be done before the marking stack is emptied. 
    // This is why we call mark_and_follow here, since that will
    // reverse the array itself and call mark_and_push on all
    // oops in the vframe array.
    // FIX this MarkSweep::mark_and_follow((oop*)deoptimized_frame_frame_array_addr());
  } else if (CodeCache::contains(pc())) { oops_code_blob_do  (&MarkSweep::mark_and_push, map);
#endif
  } else {
    ShouldNotReachHere();
  }
}


void frame::gc_prologue() {  
  if (is_interpreted_frame()) {
    // set bcx to bci to become methodOop position independent during GC
    interpreter_frame_set_bcx(interpreter_frame_bci());
  }
}


void frame::gc_epilogue() {
  if (is_interpreted_frame()) {
    // set bcx back to bcp for interpreter
    interpreter_frame_set_bcx((jint)interpreter_frame_bcp());
  }
  // call processor specific epilog function
  pd_gc_epilog();
}


# ifdef ENABLE_ZAP_DEAD_LOCALS

void frame::check_value(oop* p) {
  if (CheckOopishValues && Universe::is_reserved_heap(*p)) {
    warning("value @ 0x%x looks oopish (0x%x) (thread = 0x%x)", p, *p, Thread::current());
  }
}


void frame::check_oop(oop* p) {
  if (*p != NULL && !(*p)->is_oop()) {
    warning("value @ 0x%x should be an oop (0x%x) (thread = 0x%x)", p, *p, Thread::current());
 }
}


void frame::check_derived_oop(oop* base, oop* derived) {
  check_oop(base);
}


void frame::zap_dead(oop* p) { 
  if (TraceZapDeadLocals) tty->print_cr("zapping @ 0x%x containing 0x%x", p, *p);
  *p = (oop)0xbabebabe; 
}


void frame::zap_dead_locals(JavaThread* thread, const RegisterMap* map) {
  assert(thread == Thread::current(), "need to synchronize to do this to another thread");
  // Tracing - part 1
  if (TraceZapDeadLocals) {
    ResourceMark rm;
    tty->print_cr("--------------------------------------------------------------------------------");
    tty->print("Zapping dead locals in ");
    print_on(tty);
    tty->cr();
  }
  // Zapping
       if (is_entry_frame      ()) zap_dead_entry_locals      (thread, map);
  else if (is_interpreted_frame()) zap_dead_interpreted_locals(thread, map);
#ifndef CORE
  else if (is_compiled_frame   ()) zap_dead_compiled_locals   (thread, map);
  else if (is_deoptimized_frame()) zap_dead_deoptimized_locals(thread, map);
#endif

  else
    // could be is_c2i_frame, is_c2runtime_frame, is_i2c_frame
    // so remove error: ShouldNotReachHere();
    ;
  // Tracing - part 2
  if (TraceZapDeadLocals) {
    tty->cr();
  }
}


void frame::zap_dead_interpreted_locals(JavaThread *my_thread, const RegisterMap* map) {
  // get current interpreter 'pc'
  methodOop m   = interpreter_frame_method();
  int       bci = interpreter_frame_bci();

  int max_locals = m->is_native() ? m->size_of_parameters() : m->max_locals();

  // process dynamic part
  InterpreterFrameClosure value_blk(this, max_locals, m->max_stack(), check_value);
  InterpreterFrameClosure   oop_blk(this, max_locals, m->max_stack(), check_oop  );
  InterpreterFrameClosure  dead_blk(this, max_locals, m->max_stack(), zap_dead   );

  // get frame map
  OopMapCacheEntry *mask = m->mask_for(bci);
  mask->iterate_all( &oop_blk, &value_blk, &dead_blk); 
}


void frame::zap_dead_compiled_locals(JavaThread*, const RegisterMap* reg_map) {

# ifndef CORE
  ResourceMark rm;
  CodeBlob* cb = CodeCache::find_blob(pc());
  assert(cb != NULL, "sanity check");
  if (cb->oop_maps() != NULL) {
    OopMapSet::all_do(this, cb, reg_map, check_oop, check_derived_oop, check_value, zap_dead);
  }
# endif
}


void frame::zap_dead_entry_locals(JavaThread*, const RegisterMap*) {
  if (TraceZapDeadLocals) warning("frame::zap_dead_entry_locals unimplemented");
}


void frame::zap_dead_deoptimized_locals(JavaThread*, const RegisterMap*) {
  if (TraceZapDeadLocals) warning("frame::zap_dead_deoptimized_locals unimplemented");
}

# endif // ENABLE_ZAP_DEAD_LOCALS


void verify_oop(oop* p) {
  oop obj = *p;
  if (SafepointSynchronize::is_at_safepoint()) {
    // We can do strict checking on marks at safepoints
    assert(obj->safepoint_is_oop_or_null(), "not an oop pointer");
  } else {
    assert(Universe::is_heap_or_null(obj), "not an oop pointer");
  }
}   


void frame::verify(const RegisterMap* map) {
  // for now make sure receiver type is correct
  if (is_interpreted_frame()) {
    methodOop method = interpreter_frame_method();
    guarantee(method->is_method(), "method is wrong in frame::verify");
    if (!method->is_static()) {
      // fetch the receiver
      oop* p = (oop*) &interpreter_frame_local_at(0);
      // make sure we have the right receiver type
    }
  }
#ifndef PRODUCT
  COMPILER2_ONLY(assert(DerivedPointerTable::is_empty(), "must be empty before verify");)
  NOT_CORE(oops_do(verify_oop, (RegisterMap*)map);)
#endif
}


#ifdef ASSERT
bool frame::verify_return_pc(address x) {
  if (StubRoutines::returns_to_call_stub(x)) {
    return true;
  }  
#ifndef CORE
  if (CodeCache::contains(x)) {
    return true; 
  }
#endif
  if (Interpreter::contains(x)) {
    return true;
  }
  return false;
}
#endif


#ifdef ASSERT
void frame::interpreter_frame_verify_monitor(BasicObjectLock* value) const {
  // verify that the value is in the right part of the frame
  address low_mark  = (address) interpreter_frame_monitor_end();
  address high_mark = (address) interpreter_frame_monitor_begin();
  address current   = (address) value;

  const int monitor_size = frame::interpreter_frame_monitor_size();
  guarantee((high_mark - current) % monitor_size  ==  0         , "check pointer");
  guarantee( high_mark > current                                , "check pointer");

  guarantee((current - low_mark) % monitor_size  ==  0         , "check pointer");
  guarantee( current >= low_mark                               , "check pointer");
}
#endif


//-----------------------------------------------------------------------------------
// StackFrameStream implementation

StackFrameStream::StackFrameStream(JavaThread *thread, bool update) : _reg_map(thread, update) {
  assert(thread->has_last_Java_frame(), "sanity check");  
  _fr = thread->last_frame(); 
  _is_done = false;
}

