/*
 * @(#)frame.hpp	1.101 01/04/21 22:47:57
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

//
// RegisterMap
//
// A companion structure used for stack traversal. The RegisterMap contains
// misc. information needed in order to do correct stack traversal of stack
// frames.  Hence, it must always be passed in as an argument to
// frame::sender(RegisterMap*).
//
// In particular, 
//   1) It provides access to the thread for which the stack belongs.  The
//      thread object is needed in order to get sender of a deoptimized frame.
//
//   2) It is used to pass information from a callee frame to its caller 
//      frame about how the frame should be traversed.  This is used to let
//      the caller frame take care of calling oops-do of out-going
//      arguments, when the callee frame is not instantiated yet.  This
//      happens, e.g., when a compiled frame calls into
//      resolve_virtual_call.  (Hence, it is critical that the same
//      RegisterMap object is used for the entire stack walk.  Normally,
//      this is hidden by using the StackFrameStream.)  This is used when
//      doing follow_oops and oops_do.
//
//   3) The RegisterMap keeps track of the values of callee-saved registers 
//      from frame to frame (hence, the name).  For some stack traversal the
//      values of the callee-saved registers does not matter, e.g., if you
//      only need the static properies such as frame type, pc, and such.
//      Updating of the RegisterMap can be turned off by instantiating the
//      register map as: RegisterMap map(thread, false);


class RegisterMap : public StackObj { 
 public:
  enum {
    NOT_COMPILER2(reg_count = Register::nofRegisters)   // Temporary dummy value until C1 uses RegisterMap
    COMPILER2_ONLY(reg_count = REG_COUNT)
  };
 private:
  address     _location[reg_count];    // Location of registers
  jint        _location_valid[(reg_count+BitsPerWord-1)/BitsPerWord];
  bool        _include_argument_oops;   // Should include argument_oop marked locations for compiler  
  JavaThread* _thread;                  // Reference to current thread
  bool        _update_map;              // Tells if the register map need to be updated
                                        // when traversing the stack
  jint*       _not_at_call_sp;          // Location of a frame where the pc is not at a call (NULL if no frame exist)
 public:
  debug_only(jint*  _update_for_sp;)    // Assert that RegisterMap is not updated twice for same SP
  RegisterMap(JavaThread *thread, bool update_map = true);
  RegisterMap(const RegisterMap* map);
    
#ifndef CORE
  address location(VMReg::Name reg) const {
    assert(0 <= reg && reg < reg_count, "sanity check");
    if (_location_valid[reg / BitsPerWord] & (1 << (reg % BitsPerWord)))
      return _location[reg];
    else
      return pd_location(reg);
  }

  void set_location(VMReg::Name reg, address loc) {
    assert(0 <= reg && reg < reg_count, "sanity check");
    assert(_update_map, "updating map that does not need updating");
    _location[reg] = loc;
    _location_valid[reg / BitsPerWord] |= (1 << (reg % BitsPerWord));
  }
#endif

  // Called by an entry frame.
  void clear(jint* not_at_call_sp);

  bool include_argument_oops() const      { return _include_argument_oops; }
  void set_include_argument_oops(bool f)  { _include_argument_oops = f; }

  JavaThread *thread() const { return _thread; }
  bool update_map()    const { return _update_map; }

  void print_on(outputStream* st) const;
  void print() const;

  // Checks if the pc is at a call for the given sp.
  bool is_pc_at_call(jint* sp) const { return _not_at_call_sp != sp; }  

  // the following contains the definition of pd_xxx methods
# include "incls/_registerMap_pd.hpp.incl"
};


// A frame represents a physical stack frame (an activation).  Frames
// can be C or Java frames, and the Java frames can be interpreted or
// compiled.  In contrast, vframes represent source-level activations,
// so that one physical frame can correspond to multiple source level
// frames because of inlining.

class frame VALUE_OBJ_CLASS_SPEC {
 private:
  // Instance variables:
  jint*   _sp; // stack pointer (from Thread::last_Java_sp)
  address _pc; // program counter (the next instruction after the call)

 public:
  // Constructors
  frame();

  // Accessors

  // pc: Returns the pc at which this frame will continue normally.
  // It must point at the beginning of the next instruction to execute.
  address pc() const { return _pc; }
  jint*   sp() const { return _sp; }
  void set_pc( address newpc ) { _pc = newpc; }
  void set_sp( jint*   newsp ) { _sp = newsp; }

  // testers

  inline bool equal(frame other) const;

  // type testers
  bool is_interpreted_frame()    const;
  bool is_java_frame()           const;
  bool is_entry_frame()          const;             // Java frame called from C?
  bool is_native_frame()         const;  
#ifndef CORE  
  bool is_c2runtime_frame()      const;
  bool is_c2i_frame()            const;
  bool is_i2c_frame()            const;
  bool is_osr_adapter_frame()    const;
  bool is_glue_frame()           const;             // glue frame is either a i2c, c2i or c2runtime
  bool is_compiled_frame()       const;
  bool is_deoptimized_frame()    const;  
  bool is_safepoint_blob_frame() const;
#endif
  // testers
  bool is_first_frame() const;	                 // oldest frame? (has no sender)
  bool is_first_java_frame() const;              // same for Java frame

  bool is_interpreted_frame_valid() const;       // performs sanity checks on interpreted frames.

  // tells whether this frame is marked for deoptimization
  bool should_be_deoptimized() const;

  // tells whether this frame can be deoptimized
  bool can_be_deoptimized() const;

  // returns the frame size in ints
  inline int frame_size() const;

  // returns the sending frame
  frame sender(RegisterMap* map, CodeBlob *nm = NULL) const;

  // returns the sender, but skips conversion frames
  frame real_sender(RegisterMap* map) const;

  // returns the the sending Java frame, skipping any intermediate C frames 
  // NB: receiver must not be first frame
  frame java_sender() const;

  // platform dependent access to next compiled frame without adjustment of the pc
  // (i.e., the pc might point into the temp. codebuffer). Used by safepoint code.
  frame sender_for_raw_compiled_frame(RegisterMap* map) const;

  // used by profile
  frame sender_for_compiled_frame(RegisterMap* map, CodeBlob* cb, bool adjusted) const;   
 private:
  // Helper methods for better factored code in frame::sender
  inline frame sender_for_entry_frame(RegisterMap* map) const;
  inline frame sender_for_interpreter_frame(RegisterMap* map) const;
#ifndef CORE  
  inline frame sender_for_deoptimized_frame(RegisterMap* map, CodeBlob* cb) const;  
  inline frame sender_for_native_frame(RegisterMap* map) const;  
#endif

  // All frames:

  // A low-level interface for vframes:

 public:

  jint*  addr_at(int index) const                { return &fp()[index];    }
  int    at(int index) const                     { return *addr_at(index); }

  // accessors for locals
  oop obj_at(int offset) const                   { return *obj_at_addr(offset);  }
  void obj_at_put(int offset, oop value)         { *obj_at_addr(offset) = value; }
  
  jint int_at(int offset) const                  { return *int_at_addr(offset);  }
  void int_at_put(int offset, jint value)        { *int_at_addr(offset) = value; }
                                                
  oop*      obj_at_addr(int offset) const        { return (oop*)     addr_at(offset); }

  oop*      adjusted_obj_at_addr(methodOop method, int index) { return obj_at_addr(adjust_offset(method, index)); }

 private:
  jint*    int_at_addr(int offset) const         { return (jint*)    addr_at(offset); }

 private:
  //%note frame_1
  inline jint**   link_addr()        const;

 public:
  // Link (i.e., the pointer to the previous frame)
  jint*    link() const                          { return *link_addr(); }
  void set_link(jint* addr);

  // Return address
  inline address  sender_pc() const;
  inline void set_sender_pc(address addr);

  // Support for deoptimization
  void set_sender_pc_for_unpack(frame caller, int exec_mode);

  // returns the stack pointer of the calling frame
  inline jint* sender_sp() const;

  
  // Interpreter frames:

 private:
  inline jint** interpreter_frame_locals_addr() const;
  inline jint*  interpreter_frame_bcx_addr() const;

 public:
  // Locals
  inline jint& interpreter_frame_local_at(int index) const;
  void interpreter_frame_set_locals(jint* locs);

  // byte code index/pointer (use these functions for unchecked frame access only!)
  jint interpreter_frame_bcx() const                  { return *interpreter_frame_bcx_addr(); }
  void interpreter_frame_set_bcx(jint bcx)            { *interpreter_frame_bcx_addr() = bcx;  }

  // byte code index
  jint interpreter_frame_bci() const;
  void interpreter_frame_set_bci(jint bci);

  // byte code pointer
  address interpreter_frame_bcp() const;
  void    interpreter_frame_set_bcp(address bcp);

  // Find receiver for an invoke when arguments are just pushed on stack (i.e., callee stack-frame is 
  // not setup)
  oop interpreter_callee_receiver(symbolHandle signature)     { return *interpreter_callee_receiver_addr(signature); }

  oop *interpreter_callee_receiver_addr(symbolHandle signature);


  // expression stack (may go up or down, direction == 1 or -1)
 public:
  inline         jint* interpreter_frame_expression_stack() const;
  inline         jint& interpreter_frame_expression_stack_at(jint offset) const;  
  inline static  jint  interpreter_frame_expression_stack_direction();
 
  // top of expression stack
  inline jint& interpreter_frame_tos_at(jint offset) const;
  inline jint* interpreter_frame_tos_address() const;

  inline jint  interpreter_frame_expression_stack_size() const;

  jint* interpreter_frame_sender_sp() const;
  void  set_interpreter_frame_sender_sp(jint* sender_sp);

  // BasicObjectLocks:
  //
  // interpreter_frame_monitor_begin is higher in memory than interpreter_frame_monitor_end
  // Interpreter_frame_monitor_begin points to one element beyond the oldest one,
  // interpreter_frame_monitor_end   points to the youngest one, or if there are none,
  //                                 it points to one beyond where the first element will be.
  // interpreter_frame_monitor_size  reports the allocation size of a monitor in the interpreter stack.
  //                                 this value is >= BasicObjectLock::size(), and may be rounded up

  BasicObjectLock* interpreter_frame_monitor_begin() const;
  BasicObjectLock* interpreter_frame_monitor_end()   const;
  BasicObjectLock* next_monitor_in_interpreter_frame(BasicObjectLock* current) const;
  BasicObjectLock* previous_monitor_in_interpreter_frame(BasicObjectLock* current) const;
  static int interpreter_frame_monitor_size();

  void interpreter_frame_set_monitor_end(BasicObjectLock* value);
  void interpreter_frame_verify_monitor(BasicObjectLock* value) const;

  // Tells whether the current interpreter_frame frame pointer
  // corresponds to the old compiled/deoptimized fp
  // The receiver used to be a top level frame
  bool interpreter_frame_equals_unpacked_fp(jint* fp);

 public:
  // Method & constant pool cache
  methodOop interpreter_frame_method() const;
  void interpreter_frame_set_method(methodOop method);
  inline methodOop* interpreter_frame_method_addr() const;
  inline constantPoolCacheOop* interpreter_frame_cache_addr() const;

 public:
  // Entry frames
  inline JavaCallWrapper* entry_frame_call_wrapper() const;
  inline jint* entry_frame_argument_at(int offset) const;


  // tells whether there is another chunk of Delta stack above
  bool entry_frame_is_first() const;

  // Compiled frames:

 public:
  // Given the index of a local, and the number of argument words
  // in this stack frame, tell which word of the stack frame to find
  // the local in.  Arguments are stored above the ofp/rpc pair,
  // while other locals are stored below it.
  // Since monitors (BasicLock blocks) are also assigned indexes,
  // but may have different storage requirements, their presence
  // can also affect the calculation of offsets.
  inline static int local_offset_for_compiler(int local_index, int nof_args, int max_nof_locals, int max_nof_monitors);

  // Given the index of a monitor, etc., tell which word of the
  // stack frame contains the start of the BasicLock block.
  // Note that the local index by convention is the __higher__
  // of the two indexes allocated to the block.
  inline static int monitor_offset_for_compiler(int local_index, int nof_args, int max_nof_locals, int max_nof_monitors);

  // Tell the smallest value that local_offset_for_compiler will attain.
  // This is used to help determine how much stack frame to allocate.
  inline static int min_local_offset_for_compiler(int nof_args, int max_nof_locals, int max_nof_monitors);

  // Tells if this register must be spilled during a call.
  // On Intel, all registers are smashed by calls.
  inline static bool volatile_across_calls(Register reg);


  // Safepoints

 private:
  inline oop* saved_oop_result_addr() const;
  inline oop* saved_receiver_addr() const;

 public:
  inline oop saved_oop_result() const;
  inline void set_saved_oop_result(oop obj);

  inline oop saved_receiver() const;
  inline void set_saved_receiver(oop obj);

  // For debugging
 private:
  const char* print_name() const;

 public:
  void print_value() const { print_value_on(tty,NULL); }
  void print_value_on(outputStream* st, JavaThread *thread) const;
  void print_on(outputStream* st) const;
  void interpreter_frame_print_on(outputStream* st) const;

  // Conversion from an VMReg::Name to physical stack location
  NOT_CORE(oop* oopmapreg_to_location(VMReg::Name reg, const RegisterMap* regmap) const;)

  // Oops-do's
  void oops_interpreted_arguments_do(symbolOop signature, bool is_static,                             void f(oop*));
  NOT_CORE(void oops_compiled_arguments_do   (symbolOop signature, bool is_static, const RegisterMap* reg_map, void f(oop*));)

 private:
  // Iteration of oops
  void oops_interpreted_do(void f(oop*), const RegisterMap* map);  
  void oops_entry_do(void f(oop*), const RegisterMap* map);  
  void oops_code_blob_do(void f(oop*), const RegisterMap* map);    
  int adjust_offset(methodOop method, int index); // helper for above fn  
 public:
  // Memory management
  void oops_do(void f(oop*), RegisterMap* map);
  void follow_roots(RegisterMap* map);

  void gc_prologue();
  void gc_epilogue();
  void pd_gc_epilog();

# ifdef ENABLE_ZAP_DEAD_LOCALS
 private:
  static void check_value      (oop* p);
  static void check_oop        (oop* p);
  static void check_derived_oop(oop* base, oop* derived);
  static void zap_dead         (oop* p);

 public:
  // Zapping
  void zap_dead_locals            (JavaThread* thread, const RegisterMap* map);
  void zap_dead_interpreted_locals(JavaThread* thread, const RegisterMap* map);
  void zap_dead_compiled_locals   (JavaThread* thread, const RegisterMap* map);
  void zap_dead_entry_locals      (JavaThread* thread, const RegisterMap* map);
  void zap_dead_deoptimized_locals(JavaThread* thread, const RegisterMap* map);
# endif
  // Verification
  void verify(const RegisterMap* map);
  static bool verify_return_pc(address x);
  // Usage:
  // assert(frame::verify_return_pc(return_address), "must be a return pc");

  //
  // get/set typed locals from a frame.
  // respects platform dependent word-ordering.
  //
  // Currently these work only for interpreted frames.
  // Todo: make these work for compiled frames.
  //
public:
  oop     get_local_object(jint slot);
  jint    get_local_int   (jint slot);
  jlong   get_local_long  (jint slot);
  jfloat  get_local_float (jint slot);
  jdouble get_local_double(jint slot);
  
  void set_local_object(jint slot, oop     obj);
  void set_local_int   (jint slot, jint    i);
  void set_local_long  (jint slot, jlong   l);
  void set_local_float (jint slot, jfloat  f);
  void set_local_double(jint slot, jdouble d);

  inline int pd_oop_map_offset_adjustment() const;
  NOT_CORE(inline oop *pd_compiled_argument_to_location(VMReg::Name reg, RegisterMap reg_map, int arg_size) const;)

# include "incls/_frame_pd.hpp.incl"
};


//
// StackFrameStream iterates through the frames of a thread starting from
// top most frame. It automatically takes care of updating the location of
// all (callee-saved) registers. Notice: If a thread is stopped at
// a safepoint, all registers are saved, not only the callee-saved ones.
//
// Use:
//   
//   for(StackFrameStream fst(thread); !fst.is_done(); fst.next()) {
//     ...
//   }
//
class StackFrameStream : public StackObj {
 private:
  frame       _fr;
  RegisterMap _reg_map;
  bool        _is_done;  
 public:
   StackFrameStream(JavaThread *thread, bool update = true);

  // Iteration
  bool is_done()                  { return (_is_done) ? true : (_is_done = _fr.is_first_frame(), false); }
  void next()                     { if (!_is_done) _fr = _fr.sender(&_reg_map); }

  // Query
  frame *current()                { return &_fr; }
  RegisterMap* register_map()     { return &_reg_map; }
  
  // Is pc at current frame at a call
  bool is_pc_at_call() const      { return _reg_map.is_pc_at_call(_fr.sp()); }
};

