/*
 * @(#)interfaceSupport.hpp	1.129 00/05/01 23:08:25
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// Wrapper for all entry points to the virtual machine.
// The HandleMarkCleaner is a faster version of HandleMark.
class HandleMarkCleaner: public StackObj {
 private:
  Thread* _thread;
 public:
  HandleMarkCleaner(Thread* thread) {
    _thread = thread;
    _thread->last_handle_mark()->push();
  }
  ~HandleMarkCleaner() {
    _thread->last_handle_mark()->pop_and_restore();
  }

  // For JVMPI VM state transition.
  // TODO: This is extremely ugly!
  friend class ThreadInVMfromUnknown;

 private:
  inline void* operator new(size_t size, void* ptr) {
    return ptr;
  }
};

// InterfaceSupport provides functionality used by the __LEAF and __ENTRY
// macros. These macros are used to guard entry points into the VM and
// perform checks upon leave of the VM.


class InterfaceSupport: AllStatic {
# ifdef ASSERT
 public:
  static long _scavenge_alot_counter;
  static long _fullgc_alot_counter;
  static long _number_of_calls;
  static long _fullgc_alot_invocation;

  // tracing
  static void trace(const char* result_type, const char* header);

  // Helper methods used to implement +ScavengeALot and +FullGCALot
  static void check_gc_alot() { if (ScavengeALot || FullGCALot) gc_alot(); }
  static void gc_alot();

  static void walk_stack_from(vframe* start_vf);
  static void walk_stack(JavaThread* thread);

# ifdef ENABLE_ZAP_DEAD_LOCALS
  static void zap_dead_locals_old(JavaThread* thread);
# endif

  static void zombieAll();
  static void deoptimizeAll();
  static void stress_derived_pointers(JavaThread *thread);
  static void verify_stack();
  static void verify_last_frame();
# endif
};


// Basic class for all thread transition classes.  

class ThreadStateTransition : public StackObj {
 protected:
  JavaThread* _thread; 
 public:
  ThreadStateTransition(JavaThread *thread) { 
    _thread = thread; 
    assert(thread != NULL && thread->is_Java_thread(), "must be Java thread"); 
  }

  // Change threadstate in a manner, so safepoint can detect changes.
  // Time-critical: called on exit from every runtime routine
  static inline void transition(JavaThread *thread, JavaThreadState from, JavaThreadState to) {
    assert(from != _thread_in_Java, "use transition_from_java");
    assert((from & 1) == 0 && (to & 1) == 0, "odd numbers are transitions states");
    assert(thread->thread_state() == from, "coming from wrong thread state");
    // Change to transition state (assumes total store ordering!  -Urs)
    thread->set_thread_state((JavaThreadState)(from + 1)); 
    if (SafepointSynchronize::do_call_back()) SafepointSynchronize::block(thread);
    thread->set_thread_state(to); 
  }

  // Same as above, but assumes from = _thread_in_Java. This is simpler, since we
  // never block on entry to the VM. This will break the code, since e.g. preserve arguments
  // have not been setup.
  static inline void transition_from_java(JavaThread *thread, JavaThreadState to) {    
    assert(thread->thread_state() == _thread_in_Java, "coming from wrong thread state");    
    thread->set_thread_state(to); 
  }
 protected:  
   void trans(JavaThreadState from, JavaThreadState to)  { transition(_thread, from, to); }
   void trans_from_java(JavaThreadState to)              { transition_from_java(_thread, to); }
};


class ThreadInVMfromJava : public ThreadStateTransition {
 public:
  ThreadInVMfromJava(JavaThread* thread) : ThreadStateTransition(thread) { 
    trans_from_java(_thread_in_vm); 
  }
  ~ThreadInVMfromJava()  {     
    trans(_thread_in_vm, _thread_in_Java);
    // Check for pending. async. exceptions or suspends.
    if (_thread->has_special_runtime_exit_condition()) _thread->handle_special_runtime_exit_condition();
  }
};


class ThreadInVMfromUnknown {
 private:
  JavaThread* _thread;
  char        _buffer[sizeof(HandleMarkCleaner)];
 public:
  ThreadInVMfromUnknown() : _thread(NULL) {
    Thread* t = Thread::current();
    if (t->is_Java_thread()) {
      JavaThread* t2 = (JavaThread*) t;
      if (t2->thread_state() == _thread_in_native) {
	_thread = t2;
	ThreadStateTransition::transition(t2, _thread_in_native, _thread_in_vm);
	new(_buffer) HandleMarkCleaner(t2);
      }
    }
  }
  ~ThreadInVMfromUnknown()  {
    if (_thread) {
      ((HandleMarkCleaner*)_buffer)->~HandleMarkCleaner();
      ThreadStateTransition::transition(_thread, _thread_in_vm, _thread_in_native);
    }
  }
};


class ThreadInVMfromNative : public ThreadStateTransition { 
 public:
  ThreadInVMfromNative(JavaThread* thread) : ThreadStateTransition(thread) { 
    trans(_thread_in_native, _thread_in_vm); 
  }
  ~ThreadInVMfromNative() {     
    trans(_thread_in_vm, _thread_in_native);
    // Check for pending. async. exceptions or suspends.
    if (_thread->has_special_runtime_exit_condition()) _thread->handle_special_runtime_exit_condition();
  }
};


class ThreadToNativeFromVM : public ThreadStateTransition { 
 public:
  ThreadToNativeFromVM(JavaThread *thread) : ThreadStateTransition(thread) {
    // We are leaving the VM at this point and going directly to native code. 
    // Block, if we are in the middle of a safepoint synchronization.    
    assert(!thread->owns_locks(), "must release all locks when leaving VM");
    trans(_thread_in_vm, _thread_in_native);
    // Check for pending. async. exceptions or suspends.
    if (_thread->has_special_runtime_exit_condition()) _thread->handle_special_runtime_exit_condition();
  }

  ~ThreadToNativeFromVM() { 
    trans(_thread_in_native, _thread_in_vm); 
  }
};


class ThreadBlockInVM : public ThreadStateTransition {
 public:
  ThreadBlockInVM(JavaThread *thread) : ThreadStateTransition(thread) { 
    trans(_thread_in_vm, _thread_blocked); 
  }
  ~ThreadBlockInVM() { 
    trans(_thread_blocked, _thread_in_vm); 
  }
};


class ThreadInCompiledLeaf {
  JavaThread* _thread;
 public:
  ThreadInCompiledLeaf(JavaThread *thread) {     
    assert(thread->thread_state() == _thread_in_Java, "wrong thread state");    
    thread->set_thread_state(_thread_in_comp_leaf);
    _thread = thread;
  }
  ~ThreadInCompiledLeaf() { 
    assert(_thread->thread_state() == _thread_in_comp_leaf, "wrong thread state");    
    _thread->set_thread_state(_thread_in_Java);
  }
};

// This special transition class is only used for runtime monitorenter
// calls to prevent asynchronous exceptions from allowing a frame to
// exit without unlocking (bug 4324348).
class ThreadInVMfromJavaForMonitorEnter : public ThreadStateTransition {
 public:
  ThreadInVMfromJavaForMonitorEnter(JavaThread* thread) : ThreadStateTransition(thread) { 
    trans_from_java(_thread_in_vm); 
  }
  ~ThreadInVMfromJavaForMonitorEnter()  {     
    trans(_thread_in_vm, _thread_in_Java);
    // NOTE: We do not check for pending. async. exceptions or suspends.
    // If we did and moved the pending async exception over into the
    // pending exception field, we would need to deopt (currently C2
    // only). However, to do so would require that we transition back
    // to the _thread_in_vm state. Instead we postpone the handling of
    // the async exception.
  }
};


// Debug class instantiated in JRT_ENTRY and ITR_ENTRY macro.
// Can be used to verify properties on enter/exit of the VM.

#ifdef ASSERT
class VMEntryWrapper {
 public:
  VMEntryWrapper() {
    if (VerifyLastFrame) {
      InterfaceSupport::verify_last_frame();
    }
  }

  ~VMEntryWrapper() {
    InterfaceSupport::check_gc_alot();
    if (WalkStackALot) {
      InterfaceSupport::walk_stack(JavaThread::current());
    }
#ifdef ENABLE_ZAP_DEAD_LOCALS
    if (ZapDeadLocalsOld) {
      InterfaceSupport::zap_dead_locals_old(JavaThread::current());
    }
#endif
#ifndef CORE
#ifdef COMPILER2
// The following options are not used in Compiler 1
    if (DeoptimizeALot) {
      COMPILER2_ONLY(InterfaceSupport::deoptimizeAll();)
    }
    if (StressDerivedPointers) {
      COMPILER2_ONLY(InterfaceSupport::stress_derived_pointers(JavaThread::current());)
    }
#endif
    if (ZombieALot) {
      InterfaceSupport::zombieAll();
    }
#endif
    // do verification AFTER potential deoptimization
    if (VerifyStack) {
      InterfaceSupport::verify_stack();
    }    

  }
};


class VMNativeEntryWrapper {
 public:
  VMNativeEntryWrapper() {
    if (GCALotAtAllSafepoints) InterfaceSupport::check_gc_alot();
  }

  ~VMNativeEntryWrapper() {
    if (GCALotAtAllSafepoints) InterfaceSupport::check_gc_alot();
  }
};

#endif


// VM-internal runtime interface support

#ifdef ASSERT
  #define TRACE_CALL(result_type, header)                            \
    InterfaceSupport::_number_of_calls++;                            \
    if (TraceRuntimeCalls) InterfaceSupport::trace(#result_type, #header);
#else
  #define TRACE_CALL(result_type, header)                            \
    /* do nothing */
#endif


// LEAF routines do not lock, GC or throw exceptions

#define __LEAF(result_type, header)                                  \
  TRACE_CALL(result_type, header)                                    \
  debug_only(NoHandleMark __hm;)                                     \
  /* begin of body */


// ENTRY routines may lock, GC and throw exceptions

#define __ENTRY(result_type, header, thread)                         \
  TRACE_CALL(result_type, header)                                    \
  HandleMarkCleaner __hm(thread);                                    \
  Thread* THREAD = thread;                                           \
  /* begin of body */


// QUICK_ENTRY routines behave like ENTRY but without a handle mark

#define __QUICK_ENTRY(result_type, header, thread)                   \
  TRACE_CALL(result_type, header)                                    \
  debug_only(NoHandleMark __hm;)                                     \
  Thread* THREAD = thread;                                           \
  /* begin of body */


// Definitions for IRT (Interpreter Runtime)
// (thread is an argument passed in to all these routines)

#define IRT_ENTRY(result_type, header)                               \
  result_type header {                                               \
    ThreadInVMfromJava __tiv(thread);                                \
    __ENTRY(result_type, header, thread)                             \
    debug_only(VMEntryWrapper __vew;)


#define IRT_LEAF(result_type, header)                                \
  result_type header {                                               \
    __LEAF(result_type, header)                                      \
    debug_only(No_Safepoint_Verifier __nspv(true);)


#define IRT_ENTRY_FOR_MONITORENTER(result_type, header)              \
  result_type header {                                               \
    ThreadInVMfromJavaForMonitorEnter __tiv(thread);                 \
    __ENTRY(result_type, header, thread)                             \
    debug_only(VMEntryWrapper __vew;)


#define IRT_END }


// Definitions for JRT (Java (Compiler/Shared) Runtime)

#define JRT_ENTRY(result_type, header)                               \
  result_type header {                                               \
    ThreadInVMfromJava __tiv(thread);                                \
    __ENTRY(result_type, header, thread)                             \
    debug_only(VMEntryWrapper __vew;)


#define JRT_LEAF(result_type, header)                                \
  result_type header {                                               \
  __LEAF(result_type, header)                                        \
  debug_only(No_Safepoint_Verifier __nspv(true);)


#define JRT_ENTRY_FOR_MONITORENTER(result_type, header)              \
  result_type header {                                               \
    ThreadInVMfromJavaForMonitorEnter __tiv(thread);                 \
    __ENTRY(result_type, header, thread)                             \
    debug_only(VMEntryWrapper __vew;)


#define JRT_END }


// Definitions for JNI

#define JNI_ENTRY(result_type, header)                               \
extern "C" {                                                         \
  static result_type JNICALL header {                                \
    JavaThread* thread=JavaThread::thread_from_jni_environment(env); \
    assert( !VerifyJNIEnvThread || (thread == Thread::current()), "JNIEnv is only valid in same thread"); \
    ThreadInVMfromNative __tiv(thread);                              \
    debug_only(VMNativeEntryWrapper __vew;)                          \
    __ENTRY(result_type, header, thread)


// Ensure that the VMNativeEntryWrapper constructor, which can cause
// a GC, is called outside the NoHandleMark (set via __QUICK_ENTRY).
#define JNI_QUICK_ENTRY(result_type, header)                         \
extern "C" {                                                         \
  static result_type JNICALL header {                                \
    JavaThread* thread=JavaThread::thread_from_jni_environment(env); \
    assert( !VerifyJNIEnvThread || (thread == Thread::current()), "JNIEnv is only valid in same thread"); \
    ThreadInVMfromNative __tiv(thread);                              \
    debug_only(VMNativeEntryWrapper __vew;)                          \
    __QUICK_ENTRY(result_type, header, thread)


#define JNI_LEAF(result_type, header)                                \
extern "C" {                                                         \
  result_type JNICALL header {                                       \
    debug_only(JavaThread* thread = JavaThread::thread_from_jni_environment(env);) \
    assert( !VerifyJNIEnvThread || (thread == Thread::current()), "JNIEnv is only valid in same thread"); \
    __LEAF(result_type, header)


// Close the routine and the extern "C"
#define JNI_END } }



// Definitions for JVM

#define JVM_ENTRY(result_type, header)                               \
  JNIEXPORT result_type JNICALL header {                             \
    JavaThread* thread=JavaThread::thread_from_jni_environment(env); \
    ThreadInVMfromNative __tiv(thread);                              \
    debug_only(VMNativeEntryWrapper __vew;)                          \
    __ENTRY(result_type, header, thread)


#define JVM_ENTRY_NO_ENV(result_type, header)                        \
  JNIEXPORT result_type JNICALL header {                             \
    JavaThread* thread = (JavaThread*)ThreadLocalStorage::thread();  \
    ThreadInVMfromNative __tiv(thread);                              \
    debug_only(VMNativeEntryWrapper __vew;)                          \
    __ENTRY(result_type, header, thread)


#define JVM_QUICK_ENTRY(result_type, header)                         \
  JNIEXPORT result_type JNICALL header {                             \
    JavaThread* thread=JavaThread::thread_from_jni_environment(env); \
    ThreadInVMfromNative __tiv(thread);                              \
    debug_only(VMNativeEntryWrapper __vew;)                          \
    __QUICK_ENTRY(result_type, header, thread)


#define JVM_LEAF(result_type, header)                                \
  JNIEXPORT result_type JNICALL header {                             \
    __LEAF(result_type, header)


#define JVM_END }


// Definitions for JVMPI

#define JVMPI_ENTRY(result_type, header)                             \
  result_type header {                                               \
    Thread* thread = Thread::current();                              \
    ThreadInVMfromUnknown __tiv;                                     \
    TRACE_CALL(result_type, header)                                  \
    Thread* THREAD = thread;                                         \
    /* body */

#define JVMPI_END  }

#define JVMPI_RAW_ENTRY(result_type, header)                         \
  result_type header {                                               \
    Thread* thread = Thread::current();                              \
    TRACE_CALL(result_type, header)                                  \
    Thread* THREAD = thread;                                         \
    /* body */

#define JVMPI_RAW_END  }

