/*
 * @(#)javaCalls.hpp	1.60 01/02/09 14:47:42
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// A JavaCallWrapper is constructed before each JavaCall and destructed after the call.
// Its purpose is to allocate/deallocate a new handle block and to save/restore the last
// Java fp/sp. A pointer to the JavaCallWrapper is stored on the stack.

class JavaCallWrapper: StackObj {
  friend class VMStructs;
 private:
  JavaThread*      _thread;                 // the thread to which this call belongs
  JNIHandleBlock*  _handles;                // the saved handle block      
  methodOop        _callee_method;          // to be able to collect arguments if entry frame is top frame
  oop              _receiver;               // the receiver of the call (if a non-static call)

  jint*            _cached_state;           // any pd cached thread state
  jint*            _last_Java_sp;           // the saved last_Java_sp_to_vm
  jint*            _not_at_call_sp;         // location of a non_at_call frame (NULL if none exist)
  address          _last_native_pc;         // the saved last_native_sp
  bool             _is_handling_implicit_exception;   // the saved handling implicit exception bit

  JavaValue*       _result;                 // result value

 public:
  // Construction/destruction
   JavaCallWrapper(methodHandle callee_method, Handle receiver, JavaValue* result, TRAPS);
  ~JavaCallWrapper();

  // Accessors
  JavaThread*      thread() const           { return _thread; }
  JNIHandleBlock*  handles() const          { return _handles; }

  jint*            cached_state() const     { return _cached_state;   }
  jint*            last_Java_sp() const     { return _last_Java_sp;   }
  jint*            not_at_call_sp() const   { return _not_at_call_sp; } 
  
  JavaValue*       result() const           { return _result; }
  // GC support
  methodOop        callee_method()          { return _callee_method; }
  oop              receiver()               { return _receiver; }
  void             oops_do(void f(oop*));

  // Machine dependent stuff
  #include "incls/_javaCalls_pd.hpp.incl"

};


// Encapsulates arguments to a JavaCall (faster, safer, and more convinient than using var-args)
class JavaCallArguments : public StackObj {
 private:
  enum Constants {
   _default_size = 3     // Must match max. no. of arguments in JavaCalls methods
  };

  jint        _value_buffer [_default_size + 1];
  bool        _is_oop_buffer[_default_size + 1];    

  jint*       _value;     
  bool*       _is_oop;
  int        _size;
  int        _max_size;  
  bool       _start_at_zero;      // Support late setting of receiver

  void push(jint e, bool is_oop) {
    assert(_size < _max_size, "index out of range");
    _is_oop[_size] = is_oop;
    _value[_size++] = e;    
  }

  void initialize() {
    // Starts at first element to support set_receiver.
    _value    = &_value_buffer[1];
    _is_oop   = &_is_oop_buffer[1];
    _max_size = _default_size;
    _size = 0;
    _start_at_zero = false;
  }

 public:
  JavaCallArguments() { initialize(); }

  JavaCallArguments(Handle receiver) {    
    initialize();
    push_oop(receiver);
  }

  JavaCallArguments(int max_size) {
    if (max_size > _default_size) {
      _value  = NEW_RESOURCE_ARRAY(jint, max_size + 1);
      _is_oop = NEW_RESOURCE_ARRAY(bool, max_size + 1);
      // Reserve room for potential receiver
      _value++; _is_oop++;  
      _max_size = max_size;
      _size = 0;
      _start_at_zero = false;
    } else {
      initialize();
    }
  }  
  
  void push_int(int i)       { push(i, false); }
  void push_oop(Handle h)    { push((jint)h.raw_value(), true); }  

  void push_double(double d) { jdouble jd = d; _is_oop[_size] = _is_oop[_size + 1] = false;
                               JNITypes::put_double(&jd, _value, _size); } 
  
  void push_long(jlong l)    { jlong jl = l; _is_oop[_size] = _is_oop[_size + 1] = false;
                               JNITypes::put_long(&jl, _value, _size); }
  
  void push_float(float f)   { jfloat jf = f; _is_oop[_size] = false;
                               JNITypes::put_float(&jf, _value, _size); }
  
  // receiver
  Handle receiver() {    
    assert(_size > 0, "must at least be one argument");
    assert(_is_oop[0], "first argument must be an oop");
    assert(_value[0] != 0, "receiver must be not-null");
    return Handle((oop*)_value[0], false);
  }

  void set_receiver(Handle h) {
    assert(_start_at_zero == false, "can only be called once");
    _start_at_zero = true;
    _is_oop--;
    _value--;
    _size++;
    _is_oop[0] = true;
    _value[0] = (jint)h.raw_value();            
  }

  // Converts all Handles to oops, and returns a reference to parameter vector
  jint* parameters() ;
  int   size_of_parameters() const { return _size; }    

  // Verify that pushed arguments fits a given method
  void verify(methodHandle method, BasicType return_type) PRODUCT_RETURN;
};

// All calls to Java have to go via JavaCalls. Sets up the stack frame
// and makes sure that the last_Java_frame pointers are chained correctly.
//

class JavaCalls: AllStatic {
  static void call_helper(JavaValue* result, methodHandle* method, JavaCallArguments* args, TRAPS);
 public:
  // Optimized Constuctor call
  static void call_default_constructor(JavaThread* thread, methodHandle method, Handle receiver, TRAPS);

  // call_special
  // ------------
  // The receiver must be first oop in argument list
  static void call_special(JavaValue* result, KlassHandle klass, symbolHandle name, symbolHandle signature, JavaCallArguments* args, TRAPS);

  static void call_special(JavaValue* result, Handle receiver, KlassHandle klass, symbolHandle name, symbolHandle signature, TRAPS); // No args
  static void call_special(JavaValue* result, Handle receiver, KlassHandle klass, symbolHandle name, symbolHandle signature, Handle arg1, TRAPS); 
  static void call_special(JavaValue* result, Handle receiver, KlassHandle klass, symbolHandle name, symbolHandle signature, Handle arg1, Handle arg2, TRAPS); 
  
  // virtual call
  // ------------

  // The receiver must be first oop in argument list
  static void call_virtual(JavaValue* result, KlassHandle spec_klass, symbolHandle name, symbolHandle signature, JavaCallArguments* args, TRAPS);

  static void call_virtual(JavaValue* result, Handle receiver, KlassHandle spec_klass, symbolHandle name, symbolHandle signature, TRAPS); // No args
  static void call_virtual(JavaValue* result, Handle receiver, KlassHandle spec_klass, symbolHandle name, symbolHandle signature, Handle arg1, TRAPS); 
  static void call_virtual(JavaValue* result, Handle receiver, KlassHandle spec_klass, symbolHandle name, symbolHandle signature, Handle arg1, Handle arg2, TRAPS);   

  // Static call
  // -----------  
  static void call_static(JavaValue* result, KlassHandle klass, symbolHandle name, symbolHandle signature, JavaCallArguments* args, TRAPS);
 
  static void call_static(JavaValue* result, KlassHandle klass, symbolHandle name, symbolHandle signature, TRAPS);
  static void call_static(JavaValue* result, KlassHandle klass, symbolHandle name, symbolHandle signature, Handle arg1, TRAPS);
  static void call_static(JavaValue* result, KlassHandle klass, symbolHandle name, symbolHandle signature, Handle arg1, Handle arg2, TRAPS);

  // Low-level interface
  static void call(JavaValue* result, methodHandle method, JavaCallArguments* args, TRAPS);
};

