/*
 * @(#)mutex.cpp	1.32 01/03/05 12:10:53
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_mutex.cpp.incl"

// This needs to be an invalid return from Thread::current; NULL is checked
// for as invalid in its implementation. _owner == INVALID_THREAD, means that
// the lock is unlocked.

Thread* Mutex::INVALID_THREAD = (Thread*)NULL;

void Mutex::lock() { 
  Thread* thread = Thread::current();    
  debug_only(check_prelock_state(thread));

  // lock_implementation is a os-specific method
  if (lock_implementation()) {
    // Success, we now own the lock
  } else {    
    bool is_vm_thread = thread->is_VM_thread();
    bool at_safepoint = SafepointSynchronize::is_at_safepoint();
    if (is_vm_thread && at_safepoint && _owner == INVALID_THREAD) {        
      // a java thread has locked the lock but has not entered the
      // critical region -- let's just pretend we've locked the lock
      // and go on.  we note this with _supress_signal so we can also
      // pretend to unlock when the time comes.
      _supress_signal = true;
    } else {
      check_block_state(thread);
      if (!thread->is_Java_thread()) {
	wait_for_lock_implementation();
      } else {	
	wait_for_lock_blocking_implementation((JavaThread*)thread);
      }
    }
  }

  assert(owner() == Mutex::INVALID_THREAD, "Mutex lock count and owner are inconsistent");
  set_owner(thread);
  trace("locks");  
}


// Returns true if thread succeceed in grapping the lock, otherwise false.
bool Mutex::try_lock() {
  Thread* thread    = Thread::current();
  debug_only(check_prelock_state(thread));

  // Special case, where all Java threads are stopped. The count is not -1, but the owner
  // is not yet set. In that case the VM thread can safely grap the lock.
  bool is_vm_thread = thread->is_VM_thread();
  bool at_safepoint = SafepointSynchronize::is_at_safepoint();
  if (is_vm_thread && at_safepoint && _owner == INVALID_THREAD) {
    set_owner(thread); // Do not need to be atomic, since we are at a safepoint
    _supress_signal = true;
    return true;
  }

  // The try_lock_implementation is platform-specific
  if (try_lock_implementation()) {
    // We got the lock        
    assert(owner() == Mutex::INVALID_THREAD, "Mutex lock count and owner are inconsistent");
    set_owner(thread);
    trace("try_locks");
    return true;
  } else {
    return false;
  }
}

// Lock without safepoint check. Should ONLY be used by safepoint code and other code
// that is guaranteed not to block while running inside the VM. If this is called with
// thread state set to be in VM, the safepoint synchronization code will deadlock!

void Mutex::lock_without_safepoint_check() {
  // lock_implementation is platform specific
  if (lock_implementation()) {
    // Success, we now own the lock
  } else {    
    wait_for_lock_implementation();
  }

  assert(_owner == INVALID_THREAD, "Mutex lock count and owner are inconsistent");
  set_owner(Thread::current());
}


// Can be called by non-Java threads (JVM_RawMonitorEnter)
void Mutex::jvm_raw_lock() {
  assert(priority() == native, "must be called by non-VM locks");
  if (lock_implementation()) {
    // Success, we now own the lock
  } else {    
    wait_for_lock_implementation();
  }
  assert(_owner == INVALID_THREAD, "Mutex lock count and owner are inconsistent");
  // This can potentially be called by non-java Threads. Thus, the ThreadLocalStorage
  // might return NULL. Don't call set_owner since it will break on an NULL
  // owner
  _owner = ThreadLocalStorage::thread();
}


// ----------------------------------------------------------------------------------
// Non-product code


#ifndef PRODUCT
void Mutex::set_owner_implementation(Thread *new_owner) {  
  // This function is solely responsible for maintaining
  // and checking the invariant that threads and locks
  // are in a 1/N relation, with some some locks unowned.
  // It uses the Mutex::_owner, Mutex::_next, and
  // Thread::_owned_locks fields, and no other function
  // changes those fields.
  // It is illegal to set the mutex from one non-NULL
  // owner to another--it must be owned by NULL as an
  // intermediate state.

  if (new_owner != INVALID_THREAD) {
    // the thread is acquiring this lock
 
    assert(_owner == INVALID_THREAD, "setting the owner thread of an already owned mutex");
    _owner = new_owner; // set the owner

    // link "this" into the owned locks list

#ifdef ASSERT  // Thread::_owned_locks is under the same ifdef
    Mutex* locks = new_owner->_owned_locks; // Mutex::set_owner_implementation is a friend of Thread

    assert(this->priority() >= 0, "bad lock priority");

    if (LogMultipleMutexLocking && locks != NULL && this != Event_lock) {
      Events::log("thread %#x locks %s, already owns %s", new_owner, name(), locks->name());
    }

    // under Solaris, the interrupt lock gets acquired when doing
    // profiling, so any lock could be held. Also, we don't want to
    // worry about locks being used by external libraries (Mutex::native)

    // It is also ok to acquire Safepoint_lock at the very end while we
    // already hold Terminator_lock - may happen because of periodic safepoints
    if (locks != NULL && locks->priority() != Mutex::native
        && locks->priority() <= this->priority() &&
	!SafepointSynchronize::is_at_safepoint() && this != Interrupt_lock && 
	this != ProfileVM_lock && this != SuspendChecker_lock &&
	!(this == Safepoint_lock && locks == Terminator_lock && SafepointSynchronize::is_synchronizing())) { 
      new_owner->print_owned_locks();
      fatal4("acquiring lock %s/%d out of order with (at least) %s/%d -- possible deadlock", 
	     this->name(), this->priority(), locks->name(), locks->priority());
    }

    this->_next = locks;
    new_owner->_owned_locks = this;
#endif

  } else {
    // the thread is releasing this lock

    Thread* old_owner = _owner;
    debug_only(_last_owner = old_owner);
    
    assert(old_owner != INVALID_THREAD, "removing the owner thread of an unowned mutex");

    _owner = INVALID_THREAD; // set the owner

    #ifdef ASSERT
    Mutex *locks = old_owner->_owned_locks;

    if (LogMultipleMutexLocking && locks != this && this != Event_lock) {
      Events::log("thread %#x unlocks %s, still owns %s", old_owner, this->name(), locks->name());
    }

    // remove "this" from the owned locks list

    Mutex *prev = NULL;
    bool found = false;
    for (Mutex *cur = locks; cur != NULL; cur = cur->next()) {
      if (cur == this) {
	found = true;
	break;
      }
      prev = cur;
    }
    assert(found, "Removing a lock not owned");
    if (prev == NULL) {
      locks = _next;
    } else {
      prev->_next = _next;
    }
    _next = NULL;
    old_owner->_owned_locks = locks;
    #endif
  }
}


bool Mutex::owned_by_self() const { 
  bool ret = _owner == Thread::current(); 
  assert(_lock_count>=0 || !ret, "lock count must by >=0 for a locked mutex");
  return ret;
}

// Factored out common sanity checks for locking mutex'es. Used by lock() and try_lock()
void Mutex::check_prelock_state(Thread *thread) {
  assert(_lock_count >= -1, "sanity check");  
  assert((!thread->is_Java_thread() || ((JavaThread *)thread)->thread_state() == _thread_in_vm) 
         || priority() == Mutex::special, "wrong thread state for using locks");
  if (StrictSafepointChecks) { 
    if (thread->is_VM_thread() && !allow_vm_block()) {
      fatal1("VM thread using lock %s (not allowed to block on)", name());
    }
    debug_only(if (priority() != Mutex::special) \
      thread->check_for_valid_safepoint_state(false);)
  }
}

void Mutex::check_block_state(Thread *thread) {  
  if (!_allow_vm_block && thread->is_VM_thread()) {
    warning("VM thread blocked on lock");
    print();
    BREAKPOINT;
  }   
  assert(_owner != thread, "deadlock: blocking on monitor owned by current thread");    
}


void Mutex::trace(const char* operation) {
}

#endif // PRODUCT
