/*
 * @(#)mutex.hpp	1.30 01/02/09 14:47:44
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// A simple Mutex for VM locking using OS primitives.  Note that
// Mutex locking is NOT guaranteed to interoperate with the fast
// object locking, which is intentional: it reduces reliance on the
// fast locking mechanism as it is developed and tuned, and gives us
// a way out of all the recursive locking rat-holes that appear when
// we try to use a single locking mechanism.
//
// Implementation adapted from WIN32 Q&A by Jeffery Richter
// (implementation should be moved into platform spefific file)

class Mutex : public CHeapObj {
 private:
  // The following methods are machine-specific, and defined in mutex_<os>.inline.hpp or mutex_<os>.cpp
  bool lock_implementation();
  bool try_lock_implementation();
  void wait_for_lock_implementation();
  void wait_for_lock_blocking_implementation(JavaThread *thread);

 public:
  // A special lock: Is a lock where you are guarantteed not to block while you are holding it, i.e., no
  // vm operation can happen, taking other locks, etc. 
  // Notice: It is critical that the 'special' priority is the lowest for the deadlock dection to work
  // The native lock priority is only for use in Mutex's created by JVM_RawMonitorCreate.
  enum lock_types { event, special, leaf, safepoint, barrier, nonleaf, native };

 protected:     
  // Fields
  long        _lock_count; 
  void*       _lock_event;  
  bool        _supress_signal;     // Used for special safepoint handling
  Thread*     _owner;              // The owner of the lock
  const char* _name;               // Name of mutex  

  // Debugging fields for naming, deadlock detection, etc. (some only used in debug mode)
#ifndef PRODUCT
  bool      _allow_vm_block;
  debug_only(int _prio;)     	   // priority (to avoid/detect potential deadlocks)
  debug_only(Mutex* _next;)        // Used by a Thread to link up owned locks    
  debug_only(Thread* _last_owner;) // the last thread to own the lock    
#endif
  
  void set_owner_implementation(Thread* owner)                        PRODUCT_RETURN;
  void trace                   (const char* operation)                PRODUCT_RETURN;
  void check_prelock_state     (Thread* thread)                       PRODUCT_RETURN;
  void check_block_state       (Thread* thread)                       PRODUCT_RETURN;

  // platform-dependent support code can go here (in os_<os_family>.cpp)
  friend class MutexImplementation;
  friend class RawMonitor;
 public:
  enum {
    _no_safepoint_check_flag = true,
    _allow_vm_block_flag     = true
  };	

  Mutex(int prio, const char *name, bool allow_vm_block = !_allow_vm_block_flag);
  ~Mutex();
  
  void lock(); // prints out warning if VM thread blocks 
  void unlock();
  bool is_locked() const                            { return _owner != INVALID_THREAD; }

  bool try_lock(); // Like lock(), but unblocking. It returns false instead

  // Lock without safepoint check. Should ONLY be used by safepoint code and other code
  // that is guaranteed not to block while running inside the VM.
  void lock_without_safepoint_check();    

  // Current owner - not not MT-safe. Can only be used to guarantee that
  // the current running thread owns the lock
  Thread* owner() const         { return _owner; }

  // Support for JVM_RawMonitorEnter & JVM_RawMonitorExit. These can be called by
  // non-Java thread. (We should really have a RawMonitor abstraction)
  void jvm_raw_lock();
  void jvm_raw_unlock();
  const char *name() const                  { return _name; }

  #ifndef PRODUCT
    bool owned_by_self() const;
    void print() const;
    debug_only(int    priority() const      { return _prio; })
    bool   allow_vm_block()                 { return _allow_vm_block; }
    
    debug_only(Mutex *next()  const         { return _next; }) 
    debug_only(void   set_next(Mutex *next) { _next = next; })       
  #endif
  
  void set_owner(Thread* owner) {
  #ifndef PRODUCT
    set_owner_implementation(owner);
  #else  
    _owner = owner;
  #endif
  }

  static Thread* INVALID_THREAD; // Value of _owner when unowned. (i.e., lock is unlocked)
};


// A Monitor is a Mutex with a built in condition variable. It allows a thread, to
// temporarily to give up the lock and wait on the lock, until it is notified.
class Monitor : public Mutex {
 private:
  void* _event; 	// Manual-reset event for notifications
  long _counter;	// Current number of notifications
  long _waiters;	// Number of threads waiting for notification
  long _tickets;	// Number of waiters to be notified

  enum WaitResults {
    CONDVAR_EVENT,         // Wait returned because of condition variable notification
    INTERRUPT_EVENT,       // Wait returned because waiting thread was interrupted
    NUMBER_WAIT_RESULTS
  };

  friend class RawMonitor;
 public:
  Monitor(int prio, const char *name, bool allow_vm_block=false);
  ~Monitor();

  // Default time is forever (i.e, zero). Returns true if it times-out, otherwise
  // false. 
  bool wait(bool no_safepoint_check = !_no_safepoint_check_flag, long timeout = 0);
  bool notify();
  bool notify_all();
};


