/*
 * @(#)osThread.hpp	1.25 01/02/14 09:28:31
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// The OSThread class holds OS-specific thread information.  It is equivalent
// to the sys_thread_t structure of the classic JVM implementation.

// The thread states represented by the ThreadState values are platform-specific
// and are likely to be only approximate, because most OSes don't give you access 
// to precise thread state information.

// Note: the ThreadState is legacy code and is not correctly implemented.
// Uses of ThreadState need to be replaced by the state in the JavaThread.

enum ThreadState {
  ALLOCATED,                    // Memory has been allocated but not initialized
  INITIALIZED,                  // The thread has been initialized but yet started
  RUNNABLE,                     // Has been started and is runnable, but not necessarily running
  SUSPENDED,                    // Has been suspended
  MONITOR_WAIT,                 // Waiting on a contended monitor lock
  CONDVAR_WAIT,                 // Waiting on a monitor
  BREAKPOINTED,                 // Suspended at breakpoint 
  SLEEPING,                     // Thread.sleep() 
  ZOMBIE                        // All done, but not reclaimed yet 
};

// I'd make OSThread a ValueObj embedded in Thread to avoid an indirection, but
// the assembler test in java.cpp expects that it can install the OSThread of
// the main thread into its own Thread at will.


class OSThread: public CHeapObj {
  friend class VMStructs;
 private:
  //void* _start_proc;            // Thread start routine
  OSThreadStartFunc _start_proc;  // Thread start routine
  void* _start_parm;              // Thread start routine parameter
  ThreadState _state;             // Thread state *hint*
 // ***************************************************************
 // Suspension state.
 // ***************************************************************

  // The 2 most significant bits are reserved for various flags. The last 8 bits have a saved thread state
  // before suspend (the very first of possibly nested suspends). The rest can be used freely to cache other info.
  enum {
    self_suspended_mask       =        0x80000000,
    physically_suspended_mask =        0x40000000,
    state_before_suspend_mask =        0x0000000f
  };

  long               _suspend_depth;         // the number of suspensions - the number of resumes.  always >= 0
  unsigned int       _suspend_flags;         // various susension related flags

  // Methods
 public:
  void set_state(ThreadState state)                { _state = state; }
  ThreadState get_state()                          { return _state; }


  // Constructor
  OSThread(OSThreadStartFunc start_proc, void* start_parm);

  // Destructor
  ~OSThread();

  // Accessors
  OSThreadStartFunc start_proc() const              { return _start_proc; }
  void set_start_proc(OSThreadStartFunc start_proc) { _start_proc = start_proc; }
  void* start_parm() const                          { return _start_parm; }
  void set_start_parm(void* start_parm)             { _start_parm = start_parm; }
  ThreadState state() const                         { return _state; }

  void print();
  void set_state_under_suspend(ThreadState state);

  long suspend_depth() const                      { return _suspend_depth;            }
  void set_suspend_depth(long depth)              { _suspend_depth = depth;           }

  bool is_self_suspended() const                  { return _suspend_flags & self_suspended_mask; }
  void set_self_suspended()                       { _suspend_flags |= self_suspended_mask;  }
  void clear_self_suspended()                     { _suspend_flags &= ~self_suspended_mask; }

  bool is_physically_suspended() const            { return _suspend_flags & physically_suspended_mask; }
  void set_physically_suspended()                 { _suspend_flags |= physically_suspended_mask;  }
  void clear_physically_suspended()               { _suspend_flags &= ~physically_suspended_mask; }

  ThreadState state_before_suspend () const          { return (ThreadState) (_suspend_flags & state_before_suspend_mask); }
  void set_state_before_suspend(ThreadState  state)  { _suspend_flags |= (unsigned)state; }
  void clear_state_before_suspend()               { _suspend_flags &= ~state_before_suspend_mask; }

  // Platform dependent stuff
  #include "incls/_osThread_pd.hpp.incl"
};


// Utility class for use with condition variables:
class OSThreadWaitState {
  OSThread*   _osthread;
  ThreadState _old_state;
 public:
  OSThreadWaitState(OSThread* osthread) {
    _osthread  = osthread;
    _old_state = osthread->state();
    osthread->set_state(CONDVAR_WAIT);
  }
  ~OSThreadWaitState() {
    _osthread->set_state(_old_state);
  }
};


// Utility class for use with contended monitors:
class OSThreadContendState {
  OSThread*   _osthread;
  ThreadState _old_state;
 public:
  OSThreadContendState(OSThread* osthread) {
    _osthread  = osthread;
    _old_state = osthread->state();
    osthread->set_state(MONITOR_WAIT);
  }
  ~OSThreadContendState() {
    _osthread->set_state(_old_state);
  }
};
