/*
 * @(#)safepoint.cpp	1.221 01/03/12 08:36:11
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
//

# include "incls/_precompiled.incl"
# include "incls/_safepoint.cpp.incl"

// --------------------------------------------------------------------------------------------------
// Implementation of Safepoint begin/end

SafepointSynchronize::SynchronizeState volatile SafepointSynchronize::_state = SafepointSynchronize::_not_synchronized;
volatile int  SafepointSynchronize::_waiting_to_block = 0;
long SafepointSynchronize::_last_safepoint = 0;

// Roll all threads forward to a safepoint and suspend them all
void SafepointSynchronize::begin() {   
  assert( Thread::current()->is_VM_thread(), "Only the VM thread can initiate a safepoint synchronization");
  
  // By getting the Threads_lock, we ensure that no threads are about to start
  // or exit. It is released again in SafepointSynchronize::end().  
  Threads_lock->lock();
  assert( _state == _not_synchronized, "trying to safepoint synchronize with wrong state");     
  int nof_threads = Threads::number_of_threads();

  if (TraceSafepoint) {
    tty->print_cr("Safepoint synchronization initiated. (%d)", nof_threads);
  }

  if (PrintSafepointStatistics) {
    begin_statistics(nof_threads);
  }    

  MutexLocker mu(Safepoint_lock);

#ifdef COMPILER2
  Threads::compute_highest_running_priority();
#endif

  // Set number of threads to wait for, before we initiate the callbacks 
  _waiting_to_block = nof_threads;    
  int still_running = nof_threads;    

  // At this point, threads will start doing callback into the safepoint code. If it is in
  // native or blocked, the thread will simply block on the barrier. If it is the vm, it will
  // decrement the _waiting_to_block counter and then block.
  _state            = _synchronizing;

  // Make interpreter safepoint aware
  AbstractInterpreter::notice_safepoints(); 

  #ifdef ASSERT
  for(JavaThread *cur = Threads::first(); cur; cur = cur->next()) {
    assert(cur->safepoint_state()->is_running(), "Illegal initial state");
  }
  #endif
  // Iterate through all threads until it have been determined how to stop them all at a safepoint  
  unsigned int iterations = 0;
  while(still_running > 0) {
    for(JavaThread *cur = Threads::first(); cur; cur = cur->next()) {
      ThreadSafepointState *cur_state = cur->safepoint_state();
      if (cur_state->is_running()) {          
        cur_state->examine_state_of_thread(iterations);                              
        if (!cur_state->is_running()) still_running--;
        if (TraceSafepoint && Verbose) cur_state->print();
      }        
    } 
    if (still_running > 0) {
      if (iterations < (unsigned int)DeferThrSuspendLoopCount) {
        os::yield();
      } else {
        os::yield_all(iterations); // Yield to all (including low-priority) threads
      }
      iterations += 1;
    }
    assert(iterations < max_jint, "We have been iterating in the safepoint loop too long");
  }     
  assert(still_running == 0, "sanity check");      

  // wait until all threads are stopped    
  while(_waiting_to_block > 0) {
    if (TraceSafepoint) tty->print_cr("Waiting for %d thread(s) to block", _waiting_to_block);
    if (!SafepointTimeout) {
      Safepoint_lock->wait(true);  // true, means with no safepoint checks
    } else {
      // Wait for at most delay milliseconds to reach a safepoint
      bool timed_out = Safepoint_lock->wait(true, SafepointTimeoutDelay);
      static bool timeout_error_printed = false;
      if (timed_out && !timeout_error_printed) {
        timeout_error_printed = true;
        // Print out the thread IDs which didn't reach the safepoint
        // for debugging purposes (useful when there are lots of
        // threads in the debugger)
        tty->print_cr("# SafepointSynchronize::begin: Fatal error:");
        tty->print_cr("# SafepointSynchronize::begin: Timed out while attempting to reach a safepoint.");
        tty->print_cr("# SafepointSynchronize::begin: Threads which did not reach the safepoint:");
        for(JavaThread *cur = Threads::first(); cur; cur = cur->next()) {
          ThreadSafepointState *cur_state = cur->safepoint_state();
          if ((cur_state->type() == ThreadSafepointState::_call_back) &&
              (!cur->has_called_back())) {
            tty->print("# ");
            cur->osthread()->print();
            tty->print_cr("");
          }
        }
        tty->print_cr("# SafepointSynchronize::begin: (End of list)");
        if (DieOnSafepointTimeout) {
          ShouldNotReachHere();
        }
      }
    }
  }               
  assert(_waiting_to_block == 0, "sanity check");
    
  // Record state
  _state = _synchronized;

  if (TraceSafepoint) {
    VM_Operation *op = VMThread::vm_operation();     
    tty->print_cr("Entering safepoint region: %s", (op != NULL) ? op->name() : "no vm operation");
  }
 
  if (PrintSafepointStatistics) {    
    end_statistics();
  }    

  // Call stuff that needs to be run when a safepoint is just about to be completed
  do_cleanup_tasks();  
}

// Wake up all threads, so they are ready to resume execution after the safepoint
// operation has been carried out
void SafepointSynchronize::end() {        

#ifndef CORE
  // If any exception got thrown on a compiled thread in a compiled safepoint, then
  // deoptimize the frame.
  for(JavaThread *cur = Threads::first(); cur; cur = cur->next()) {      
    if (cur->has_pending_exception() && cur->safepoint_state()->type() == ThreadSafepointState::_compiled_safepoint) {    
#ifdef COMPILER2
      frame stub_fr = cur->last_frame();      
      RegisterMap map(cur, false);            
      frame caller_fr = stub_fr.sender_for_raw_compiled_frame(&map);         
      Deoptimization::deoptimize_frame(cur, caller_fr.sp());     
#endif
    }    
  }
#endif // CORE

  MutexLocker mu(Safepoint_lock);

  assert(_state == _synchronized, "must be synchronized before ending safepoint synchronization");
  assert(Thread::current()->is_VM_thread(), "Only the VM thread can end a safepoint synchronization");

  // Set to not synchronized, so the threads will not go into the signal_thread_blocked method
  // when they get restarted.
  _state = _not_synchronized;  

  if (TraceSafepoint) {
     tty->print_cr("Leaving safepoint region");
  }

  // Remove safepoint check from interpreter
  AbstractInterpreter::ignore_safepoints();

  // Start suspended threads
  for(JavaThread *current = Threads::first(); current; current = current->next()) {      
    Thread::unboost_priority(current);
    ThreadSafepointState* cur_state = current->safepoint_state();
    assert(cur_state->type() != ThreadSafepointState::_running, "Thread not suspended at safepoint");
    cur_state->restart();                  
    assert(cur_state->is_running(), "safepoint state has not been reset");
  }

  // Release threads lock, so threads can be created/destroyed again. It will also starts all threads
  // blocked in signal_thread_blocked
  Threads_lock->unlock();  

  _last_safepoint = os::javaTimeMillis();
}

bool SafepointSynchronize::is_cleanup_needed() {
#ifndef CORE
  // Need a safepoint if some inline cache buffers is non-empty
  if (!InlineCacheBuffer::is_empty()) return true;
  // do not trigger safepoints for running the CounterDecayTask it causes too
  // many safepoints resulting in a degradation in performance
  //if (UseCounterDecay && CounterDecay::is_decay_needed()) return true;
  if (pd_is_cleanups_needed()) return true;
#endif
  return false;
}

// Various cleaning tasks that should be done periodically at safepoints
void SafepointSynchronize::do_cleanup_tasks() {  
  // Update fat-monitor pool, since this is a safepoint.
  ObjectSynchronizer::deflate_idle_monitors();
#ifndef CORE  
  InlineCacheBuffer::update_inline_caches();  
  if( UseCounterDecay ) {
    CounterDecay::decay();
  }
  NMethodSweeper::sweep(); 
#endif
}


// -------------------------------------------------------------------------------------------------------
// Implementation of Safepoint callback point

void SafepointSynchronize::block(JavaThread *thread) {
  assert(thread != NULL, "thread must be set");
  assert(thread->is_Java_thread(), "not a Java thread");

  // Do nothing for a thread that is not in the thread lists (i.e., it has been terminated)
  if (thread->is_terminated()) return;

  JavaThreadState state = thread->thread_state();
  jint *saved_state = thread->save_state_before_block();
  
  // Check that we have a valid thread_state at this point
  switch(state) {
    case _thread_in_vm_trans:
    case _thread_in_Java:        // From compiled code            

      // We are highly likely to block on the Safepoint_lock. In order to avoid blocking in this case,
      // we pretend we are still in the VM.
      thread->set_thread_state(_thread_in_vm);

      // We will always be holding the Safepoint_lock when we are examine the state
      // of a thread. Hence, the instructions between the Safepoint_lock->lock() and 
      // Safepoint_lock->unlock() are happening atomic with regards to the safepoint code      
      Safepoint_lock->lock_without_safepoint_check();
      if (is_synchronizing()) {
        // Decrement the number of threads to wait for and signal vm thread      
        assert(_waiting_to_block > 0, "sanity check");
        _waiting_to_block--;                       
        thread->set_has_called_back(true);
        Safepoint_lock->notify_all();
      }
      thread->set_thread_state(_thread_blocked);
      Safepoint_lock->unlock();            
            
      // We now try to acquire the threads lock. Since this lock is hold by the VM thread during
      // the entire safepoint, the threads will all line up here during the safepoint.
      Threads_lock->lock_without_safepoint_check();
      // restore original state. This is important if the thread comes from compiled code, so it
      // will continue to execute with the _thread_in_Java state. 
      thread->set_thread_state(state);
      Threads_lock->unlock(); 
      break;                  
       
    case _thread_in_native_trans:
    case _thread_blocked_trans:
    case _thread_new_trans:
      if (thread->safepoint_state()->type() == ThreadSafepointState::_call_back) {
        address stop_pc = thread->safepoint_state()->_stop_pc;
        InterpreterCodelet* icd = Interpreter::codelet_containing(stop_pc);
        if (icd != NULL) {
          icd->print();
          fatal("Wrong safepoint info in interpreter");
        } else {
          fatal1("Deadlock in safepoint code. stopped at 0x%x", stop_pc);
        }
      }
          
      // It is not safe to suspend a thread if we discover it is in _thread_in_native_trans. Hence,
      // the safepoint code might still be waiting for it to block. We need to change the state here,
      // so it can see that it is at a safepoint. 
      // Furthermore, JVM_ThreadStop needs to be able to tell if a thread is in native, so it can figure
      // out how to deliver an async. exception to a thread. It uses the thread_state(), so we have a special
      // thread_state, so we can see we are here from native code.
      if (state == _thread_in_native_trans) thread->set_thread_state(_thread_in_native_blocked);

      // Block until the safepoint operation is completed. 
      Threads_lock->lock_without_safepoint_check();
      // Restore state
      
      // The call_VM call is going to check for pending exceptions immediatly after the return from this call.
      // Therefore we need to eagerly change the thread state to _thread_in_Java so the exception handling code
      // gets the right thread_state. This is only needed for _thread_in_native_trans. For _thread_blocked_trans,
      // and _thread_new_trans, there is no immediate safepoint check.
      if (state == _thread_in_native_trans) state = _thread_in_Java;      
      thread->set_thread_state(state);
      Threads_lock->unlock();      
      break;

    default:
     fatal1("Illegal threadstate encountered: %d", state);
  }

  thread->restore_state_after_block(saved_state);

  // Check for pending. async. exceptions or suspends - except if the thread was blocked inside
  // the VM.
  if (thread->has_special_runtime_exit_condition() && (state != _thread_blocked_trans && state != _thread_in_vm_trans)) {
    thread->handle_special_runtime_exit_condition();  
  }
}

// -------------------------------------------------------------------------------------------------------
// Exception handlers

int SafepointSynchronize::handle_illegal_instruction_exception(JavaThread *thread) { 
  assert(thread->is_Java_thread(), "illegal instruction encountered by VM thread");
  assert(thread->thread_state() == _thread_in_Java, "should come from Java code");
  assert(SafepointSynchronize::is_synchronizing(), "illegal instruction encountered outside safepoint synchronization");  

  if (ShowSafepointMsgs) {
    tty->print("handle_illegal_instruction_exception: ");  
  }  

  ThreadSafepointState* state = thread->safepoint_state();
  assert(state->type() == ThreadSafepointState::_compiled_safepoint, "called on thread without installing handler");
  assert(state->handle() != NULL, "handler not set");  

  return state->handle()->handle_illegal_instruction_exception();  
}


// -------------------------------------------------------------------------------------------------------
// Implementation of ThreadSafepointState 

ThreadSafepointState::ThreadSafepointState(JavaThread *thread) {
  _thread = thread;
  _type   = _running;
  NOT_CORE(_handle = new CompiledCodeSafepointHandler(thread);)
  NOT_CORE(_code_buffer = NULL;)
}

ThreadSafepointState::~ThreadSafepointState() {
#ifndef CORE
  assert(_handle != NULL, "must be created");
  delete _handle;
  _handle = NULL;
  assert(_code_buffer == NULL, "should already been removed");
#endif
}

void ThreadSafepointState::create(JavaThread *thread) {
  ThreadSafepointState *state = new ThreadSafepointState(thread);
  thread->set_safepoint_state(state);
}

void ThreadSafepointState::destroy(JavaThread *thread) {
  if (thread->safepoint_state()) {
    delete(thread->safepoint_state());  
    thread->set_safepoint_state(NULL);
  }
}

void ThreadSafepointState::examine_state_of_thread(int iterations) {  
  assert(is_running(), "better be running or this should not be called");
  _stop_pc    = 0; // temp. debugging
  
  // Quick check for common cases which can be done without suspending the thread  
  JavaThreadState state = _thread->thread_state();  
  _stop_state = state;
  
  switch(state) {

   case _thread_in_native:
    if (SafepointSynchronize::can_be_at_safepoint_before_suspend(_thread, state)) {
      roll_forward(_at_safepoint);       
      return;
    }
    break;

   case _thread_in_native_blocked:
   case _thread_blocked:
    if (SafepointSynchronize::can_be_at_safepoint_before_suspend(_thread, state)) {      
      roll_forward(_at_safepoint);
      return;
    }
    break;

   case _thread_new:   
     // Let new thread run until it is done getting local storage
     // (may require a malloc). Note that we have not suspended the
     // thread here, so we do not resume it.
     roll_forward(_running, NULL, true);
     return;

   case _thread_in_vm:
     roll_forward(_call_back);     
     return;
  }

  // Check for a thread that is suspended. Note that thread resume tries
  // to grab the Threads_lock which we own here, so a thread cannot be
  // resumed during safepoint synchronization. 

  bool is_suspended;
  {
    // We are in the process of bringing the JavaThreads to a safepoint.
    // However, while we are working on this thread another may be
    // trying to suspend this one. So we have to check is_suspended()
    // under the protection of the Interrupt_lock.
    MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
    is_suspended = _thread->is_suspended();
  }
  if (is_suspended) {
    // Ensure looking at current state
    state = _thread->thread_state();  
    _stop_state = state;
      // We may be in native if suspension happens in ~ThreadInVMfromNative, or if jvmdi or jvmpi is active
    assert(state == _thread_in_Java || state == _thread_new_trans || 
	   state == _thread_in_native || state == _thread_in_native_trans, 
	   "a suspended thread should have thread state in_native, in_Java, or just started");
    roll_forward(_at_safepoint); 
    return;
  }

  // Iterate through the safepoint loop without suspending threads in the
  // hope that threads would block or suspend themselves
  if (iterations < DeferThrSuspendLoopCount) {
    return;
  }

  // At this point it was either _thread_in_Java or one of the transition state
  // Suspend thread and examine state
#ifndef PRODUCT
  if (PrintSafepointStatistics) {
    SafepointSynchronize::log_suspension();
  }
#endif

  long ret = _thread->suspend_other(false);  // internal suspend request
  if (ret>1) {
    // Thread already suspended. Should only could happen if profiler is active
    SafepointSynchronize::safepoint_msg("Looking at suspended thread (is profiler active?)");    
  }

  // Reread thread_state for thread now it is blocked
  state = _thread->thread_state();  
  _stop_state = state; // temp. debugging
  
  // Normalize all xx_trans states to normal states  
  switch(state) {      
   case _thread_in_native_trans:   
     // The _last_sp might not be setup correctly, if we are just about to enter the
     // VM again, hence, we avoid counting this transition as being a safepoint.
     roll_forward(_running); 
     return;

   case _thread_in_vm_trans:
     if (SafepointSynchronize::can_be_at_safepoint_after_suspend(_thread, state)) {
       roll_forward(_at_safepoint_suspended); 
       return;
     }
     // we can't stop here because some important thread state. The last_java_frame
     // might not work, since the return address has been popped of the stack on return
     // from the native method
     roll_forward(_running);
     return;

   case _thread_blocked_trans:
   case _thread_new_trans:         
     roll_forward(_at_safepoint_suspended); 
     return;

   case _thread_in_native:
   case _thread_blocked:
   case _thread_new:   
   case _thread_in_native_blocked:
     // It is critical that we not keep a thread in native code suspended, since it
     // could hold a lock in a C library (such as malloc) and thereby deadlocking the vm
     roll_forward(_at_safepoint);
     _thread->resume(false, false);  // don't yield, internal resume request
     return;

   case _thread_in_vm:
     roll_forward(_call_back);
     _thread->resume(false, false);  // don't yield, internal resume request
     return; 
  }

  // Examine state of thread
  // get current frame for stopped thread  
  frame fr;
  if (!_thread->get_top_frame(&fr, &_addr)) {
    // get_top_frame can fail on Solaris due to a bug in the Solaris signal code. It also
    // returns false on Sparc if the current instruction is in a delay slot
    // Simply printing the following message changes the I/O profile of the VM thread's lwp
    // and on Solaris may render priority boosting useless (I/O increases lwp priority)
    // SafepointSynchronize::safepoint_msg("Thread is in unknown code. [id: %2x]", _thread->osthread()->thread_id());
    // help the thread make progress
    Thread::boost_priority(_thread, Thread::get_priority(VMThread::vm_thread()));
    roll_forward(_running);  
    return;
  }

  address pc = _addr.contained_pc();
  _stop_pc = pc;  

#ifndef CORE
  // Check if it is in a compiled leaf
  if (state == _thread_in_comp_leaf) {
    SafepointSynchronize::safepoint_msg("Thread in compiled leaf at 0x%x  [id: %2x]", pc, _thread->osthread()->thread_id()); 
    
    // Find nmethod we are called from last_java_sp
    RegisterMap reg_map(_thread, false);
    frame stub_fr   = _thread->last_frame();
    assert(stub_fr.is_c2runtime_frame(), "wrong kind of frame");
    frame caller_fr =  stub_fr.sender(&reg_map);    
    CodeBlob* cb = CodeCache::find_blob(caller_fr.pc());  
    assert(cb != NULL && cb->is_nmethod(), "must be called from a compiled method");    
    if (cb->is_java_method()) {
      // This is a special case of a compiler safepoint. We copy the code, and then patch the return
      // address in the frame given by the last_Java_sp to return to our copied code which has been
      // patched with safepoint traps.
      if(UseCompLeafSafepoints) {
	roll_forward(_compiled_safepoint, (nmethod *)cb);
      }
      else {
	// comp leaf safepoints disabled
	roll_forward(_running);
      }
      return;
    } else {
      SafepointSynchronize::safepoint_msg("skipping... it is a native method"); 
      roll_forward(_running);
      return;
    }
  }
#endif

  guarantee(state == _thread_in_Java, "must be in Java code at this point");  

  // We handle VM stub routines by restarting the thread again, and then stop it later.
  if (StubRoutines::contains(pc)) {    
    SafepointSynchronize::safepoint_msg("Thread in VM stub code: %s. pc: 0x%x  [id: %2x]", StubCodeDesc::name_for(pc), pc, _thread->osthread()->thread_id()); 
    // help the thread make progress    
    roll_forward(_running);
    return;
  } 
    
#ifndef CORE
  // Compiled code or native method stub
  // NOTE: we can not use find_blob here because there is a very small
  // window in which we may be in the inline cache check for a
  // previously zombied method (that is, before the zombie trap at the
  // verified entry point which will cause re-resolution of the call
  // site). In this situation the pc will be in the zombied method and
  // the guarantee in find_blob will fail. To solve this (very
  // uncommon) problem we temporarily resume a thread found in this
  // state.
  CodeBlob* cb = CodeCache::find_blob_unsafe(pc);
  if (cb != NULL) {    
    if (cb->is_java_method()) {            
      if (cb->is_zombie()) {
        nmethod* nm = (nmethod*) cb;
        guarantee(nm->inlinecache_check_contains(pc) || (nm->verified_entry_point() == pc),
                  "should not be running a zombie method");
        roll_forward(_running);
        return;
      } else {
        roll_forward(_compiled_safepoint, (nmethod *)cb);
        return;
      }
    } else {
      // Thread is in stub code. We run the thread for a little while again...            
      SafepointSynchronize::safepoint_msg("Thread in native stub.");
      // help the thread make progress      
      roll_forward(_running);
      return;
    }       
  }
#endif

  // In interpreter
  if (Interpreter::contains(pc)) {     
    InterpreterCodelet* clet = Interpreter::codelet_containing(pc);
    assert(clet != NULL, "code in interpreter must be in a codelet");
    if (UseCallBackInInterpreter && clet->is_safepoint_safe()) {
      roll_forward(_call_back);
// TODO - why was this calling resume, not resume_async?
// Do we really want to do a yield for it to run again?
      _thread->resume(true, false);  // yield, internal resume request
      return;
    } else {
      // help the thread make progress      
      roll_forward(_running);
      return;
    }    
  }
   
  // Thread is in an unknown (or unhandled) region of the code
  // On Solaris, merely printing a message (from within the examine state loop) will boost
  // VM thread's priority.
  // SafepointSynchronize::safepoint_msg("Thread is in unknown code. pc: 0x%x [id: %2x]", pc, _thread->osthread()->thread_id());
  
  // help the thread make progress  
  roll_forward(_running);  
}

// Returns true is thread could not be rolled forward at present position.
void ThreadSafepointState::roll_forward(suspend_type type, nmethod *nm, bool disable_resume_for_running_thread) {
  _type = type;  

  switch(_type) {
    case _at_safepoint:
    case _at_safepoint_suspended:
      SafepointSynchronize::save_thread_state_at_safepoint(thread());
      SafepointSynchronize::signal_thread_at_safepoint();
      break;

    case _running:            
      if (!disable_resume_for_running_thread) {
	Thread::boost_priority(thread(), Thread::get_priority(VMThread::vm_thread()));
	thread()->resume(false, false);  // don't yield, internal resume request
      }
      break;
    
    case _call_back:
      thread()->set_has_called_back(false);
      break;
         
    case _compiled_safepoint:      
      assert(nm != NULL, "handler must be set");
      assert(_stop_pc != NULL, "must be set");      
      assert(handle() != NULL, "must be set");
      handle()->setup(this, nm);
      break;

    default:
      ShouldNotReachHere();
  }      
}  

void ThreadSafepointState::restart() {  
  switch(type()) {        
    case _at_safepoint:               
    case _call_back:
      break;    
    
    case _at_safepoint_suspended:
      thread()->resume(false, false);  // don't yield, internal resume request
      break;
        
    case _compiled_safepoint:
      assert(handle() != NULL, "handler must be set (it's preallocated)");
      _handle->release(this);      
      break;    

    default:
      ShouldNotReachHere();
  }    
  _type = _running;  
}


bool ThreadSafepointState::caller_must_gc_arguments() const {
  assert(_handle != NULL, "does not make sense without a handler");
  return _handle->caller_must_gc_arguments();
}

void ThreadSafepointState::print() {
  const char *s;

  switch(_type) {    
    case _running                : s = "_running";              break;    
    case _at_safepoint           : s = "_at_safepoint";         break;
    case _at_safepoint_suspended : s = "_at_safepoint_suspend"; break;
    case _call_back              : s = "_call_back";            break;    
    case _compiled_safepoint     : s = "_compiled_handler";     break;
    default:
      ShouldNotReachHere();
  }
  
  tty->print_cr("Thread: 0x%lx  [0x%2x] State: %s  pc: 0x%x", _thread, _thread->osthread()->thread_id(), s, _stop_pc);

  if (_type == _compiled_safepoint) {
    assert(handle() != NULL, "handler must be set");
    handle()->print();
  }
}

#ifndef CORE

ThreadCodeBuffer* ThreadSafepointState::allocate_code_buffer(int size_in_bytes, nmethod* nm, address real_pc) {
  assert(_code_buffer == NULL, "buffer must be destroyed before allocating");
  _code_buffer = new ThreadCodeBuffer(size_in_bytes, nm, real_pc);
  return _code_buffer;
}

void ThreadSafepointState::destroy_code_buffer() {
  assert(_code_buffer != NULL, "cannot destroy non-existing buffer");
  delete _code_buffer;
  _code_buffer = NULL;
}

#endif

// ---------------------------------------------------------------------------------------------------------------------
// Implementation of general SafepointHandler

void SafepointHandler::set_state_to_running() {    
  thread()->safepoint_state()->roll_forward(ThreadSafepointState::_running);  
}


// ---------------------------------------------------------------------------------------------------------------------
// Implementation of SafepointHandler for compiled code (not in CORE Version)
#ifndef CORE  

void CompiledCodeSafepointHandler::setup(ThreadSafepointState *state, nmethod* nm) { 
  ResourceMark rm;
    
  // Bail out if compiler safepoints are disabled
  if (!UseCompilerSafepoints) {       
    set_state_to_running();
    return;
  } 
  
  _nm = nm;
  JavaThread* thread = state->thread();

  // This routine handles both the case where a thread is stopped in:
  // - an nmethod, or
  // - a leaf call from an nmethod
  //
  // Most parts of the code is the same. The difference is how we find which
  // nmethod to copy andpatch, and how to restart the thread.
  //
  address pc = NULL;
  frame caller_fr;
  if (thread->thread_state() == _thread_in_comp_leaf) {    
    RegisterMap reg_map(thread, false);
    frame stub_fr   = thread->last_frame();    
    caller_fr = stub_fr.sender(&reg_map);
    pc = caller_fr.pc();
    safepoint_msg("### Compiled leaf: %s. pc: 0x%x", nm->method()->name()->as_C_string(), pc);
  } else {
    assert(thread->thread_state() == _thread_in_Java, "wrong thread state");
    pc = state->current_address().contained_pc();
    safepoint_msg("### Compiled code: %s. pc: 0x%x", nm->method()->name()->as_C_string(), pc);
  }        
    
  // It it in body, exception, or stub
  if (nm->stub_contains(pc) || nm->exception_contains(pc)) {      
    // we restart thread in these cases for now
    safepoint_msg("exception/stub part");      
    set_state_to_running();
    return;
  }  
  assert(nm->code_contains(pc), "must be in body part");
  
  // Allocate code buffer for roll-forward. We currenty copy entire code-buffer + constants.
  int size_in_bytes = nm->instructions_size() + nm->constants_size();
  ThreadCodeBuffer *cb = state->allocate_code_buffer(size_in_bytes, nm, nm->constants_begin() );        
  if(ShowSafepointMsgs) {
    safepoint_msg("############# Original nmethod ################");
    nm->print();
  }

  // Step 1: Make copy of code and constants. Code is handles specially since it must be relocated
  // This is probably broken and works only if there are no constants in the beginning
#ifdef COMPILER1
  cb->copy_code(0, nm->instructions_begin(), nm->instructions_size());            
#else
  // This works only for constants_size = 0
  cb->copy     (0                   , nm->constants_begin()   , nm->constants_size());
  cb->copy_code(nm->constants_size(), nm->instructions_begin(), nm->instructions_size());            
#endif

  if (ShowSafepointMsgs) {
    tty->print_cr("Original code: [0x%x - 0x%x]", nm->constants_begin(), nm->instructions_end());
    tty->print_cr("Code buffer:   [0x%x - 0x%x]", cb->code_begin(),      cb->code_end());
    tty->print_cr("Orignal pc:     0x%x", pc);
  }
  
  // Step 2: Patch all calls and backwards jumps out of the code
  RelocIterator iter(nm);
  while (iter.next()) {
    address reloc_pc = iter.inst_addr();
    if (reloc_pc >= nm->code_end())
    break;
            
    address cb_pc = cb->code_begin() + (reloc_pc - nm->constants_begin());
    assert(cb->contains(cb_pc), "pc must be in temp. buffer");            

    switch(iter.type()) {	
      // Handle all patches of calls
	        
      case relocInfo::virtual_call_type:     // virtual call
      case relocInfo::opt_virtual_call_type: // optimized virtual call
      case relocInfo::static_call_type:      // static call             
        if (ShowSafepointMsgs && Verbose) { 
          tty->print_cr("Illegal instruction at 0x%x for invoke (org. pc: 0x%x)", cb_pc, reloc_pc); 
        }
        assert(((NativeInstruction*)cb_pc)->is_call(), "must be a call");
	NativeIllegalInstruction::insert(cb_pc);
        break;

      case relocInfo::return_type:
        if (ShowSafepointMsgs && Verbose) { 
          tty->print_cr("Illegal instruction at 0x%x for return (org. pc: 0x%x)", cb_pc, reloc_pc); 
        }	
        SafepointSynchronize::patch_return_instruction_md(cb_pc);
        break;

      case relocInfo::runtime_call_type:
        if (ShowSafepointMsgs && Verbose) { 
          tty->print_cr("ignoring runtime call at 0x%x (org. pc: 0x%x)", cb_pc, reloc_pc); 
        }
        // Nothing to do - just for debugging
        break;
            
      case relocInfo::safepoint_type:  
        // Could be any instruction for C1
        NOT_COMPILER1(assert(((NativeInstruction*)cb_pc)->is_nop(), "must be a nop");)
      	NativeIllegalInstruction::insert(cb_pc);
        break;      

      default:
	// skib
	break;
    } // switch
  } // while

  // Step 3: Reposition thread
  if (thread->thread_state() == _thread_in_comp_leaf) {
    if (ShowSafepointMsgs && Verbose) {
      tty->print_cr("Repositioning thread found in _comp_leaf state");
    }
    // If the thread was stopped in a leaf method called from compiled code, we patch the
    // return address to call patch code    
    address cb_pc = cb->code_begin() + (pc - nm->constants_begin());
    assert(cb->contains(cb_pc), "pc must be in temp. buffer");                
    caller_fr.patch_pc(thread, cb_pc);
  } else {
    if (ShowSafepointMsgs && Verbose) {
      tty->print_cr("Repositioning thread found in nmethod");
    }
    // If thread was stopped in the nmethod, we simply reposition it into the new code
    // This may fail in some rare cases under Solaris - back off if it does
    if(!cb->reposition_thread(thread, state->current_address())) {
      state->destroy_code_buffer();
      // we restart the thread in this case for now
      safepoint_msg("failed to reposition thread, resetting state to running");      
      set_state_to_running();
      return;
    }
  }
  thread->resume(false, false);  // don't yield, internal resume request
}  

void CompiledCodeSafepointHandler::release(ThreadSafepointState *state) {  
  assert(state->thread()->thread_state() == _thread_blocked, "thread_state should have changed, since it is blocked");

  if (!UseCompilerSafepoints) {      
    return;
  }  

  ThreadCodeBuffer* cbuf = thread()->safepoint_state()->code_buffer();
  if (cbuf != NULL) {            
    frame stub_fr = thread()->last_frame();
    // This has been changed to a guarantee - since if this break it means that
    // our compiled safepoints are not completly waterproof, i.e., a thread can escape
    // out of the method.
    CodeBlob* stub_cb = CodeCache::find_blob(stub_fr.pc());        
    COMPILER1_ONLY(bool rethrow_stub = false);
    COMPILER2_ONLY(bool rethrow_stub = stub_cb->instructions_begin() == OptoRuntime::rethrow_stub());
    guarantee(stub_cb != NULL && (stub_cb->is_safepoint_stub() || stub_cb->is_runtime_stub() || 
                                  stub_cb->is_exception_stub() || stub_cb->is_uncommon_trap_stub() ||
                                  rethrow_stub), "CompiledCodeSafepointHandler::release: wrong stub");      

    // Find caller frame. Do not adjust pc for safepoint
    RegisterMap map(thread(), false);    
    frame caller_fr = stub_fr.sender_for_raw_compiled_frame(&map);
    address return_addr = caller_fr.pc();

    if (cbuf->contains(return_addr)) {      
      safepoint_msg("adjusting return address");
      return_addr  = cbuf->compute_adjusted_pc(return_addr);
      caller_fr.patch_pc(thread(), return_addr);	
    }
    state->destroy_code_buffer();
  }  
}

// Return non zero if we are stopped at a return that must pop input
// arguments (UseNewCallingConvention = false && ret x, with x > 0)
int CompiledCodeSafepointHandler::handle_illegal_instruction_exception() {    

  // Step 1: find references to SafepointState and ThreadCodeBuffer
  ThreadSafepointState *state = thread()->safepoint_state();
  ThreadCodeBuffer     *cbuf   = state->code_buffer();  
  assert(cbuf != NULL, "for pc to be in code buffer the codebuffer must exist");
  nmethod* nm = get_nmethod();

  // Step 2: Find frame of caller  
  frame stub_fr = thread()->last_frame();
  CodeBlob* stub_cb = CodeCache::find_blob(stub_fr.pc());
  assert(stub_cb->is_safepoint_stub(), "must be a safepoint stub");      
  RegisterMap map(thread(), false);  
  frame caller_fr = stub_fr.sender_for_raw_compiled_frame(&map);   
    
  // Step 3: Calculate real return address     
  address real_return_addr = cbuf->compute_adjusted_pc(thread()->saved_exception_pc());  

  // Find relocInfo::relocType for the address where we get called, so we can figure out how to
  // preserve the oops.  

  relocInfo::relocType type = nm->reloc_type_for_address(real_return_addr);

  // Default case. We we are blocked at a call, then the caller must GC args
  _caller_must_gc_args = true;

  // set to non zero if using old calling convention and the safepoint
  // happened at a return where the return must remove its arguments
  int arg_size_in_bytes = 0; 
  
  switch(type) {	
    // Handle all patches of calls

    case relocInfo::virtual_call_type:	        // inline cache call
      safepoint_msg("ic_call_type");               
      { HandleMark hm;
        Handle receiver(NULL);
        Handle cached_klass(NULL);
#ifdef COMPILER1
        receiver = caller_fr.saved_receiver();
        oop saved_oop = caller_fr.saved_oop_result(); 
        if (Universe::is_non_oop(saved_oop)) {
          // cached klass remains NULL
        } else {
          cached_klass = saved_oop;
        }
#endif
        // Move return address after call, so exact oopmap lookup will work      
        caller_fr.patch_pc(thread(), real_return_addr + NativeCall::instruction_size);
        // Block
        SafepointSynchronize::block(thread()); 
        // Adjust return address so mov in inline cache is reexecuted      
      
        real_return_addr -= NativeMovConstReg::instruction_size;
#ifdef COMPILER1
        caller_fr.set_saved_receiver(receiver());
        if (cached_klass.not_null()) {
          // an oop has been preserved
          caller_fr.set_saved_oop_result(cached_klass());
        }
#endif
      }
      break;

    case relocInfo::static_call_type:       // static call  	        
      safepoint_msg("static_call_type");
      // Move return address after call, so exact oopmap lookup will work     
      caller_fr.patch_pc(thread(), real_return_addr + NativeCall::instruction_size);
      // Block
      SafepointSynchronize::block(thread());       
      break;

    case relocInfo::opt_virtual_call_type:  // optimized virtual call
      safepoint_msg("static_call_type");
      { HandleMark hm;
        Handle receiver(NULL);
        COMPILER1_ONLY(receiver = caller_fr.saved_receiver();)
        // Move return address after call, so exact oopmap lookup will work     
        caller_fr.patch_pc(thread(), real_return_addr + NativeCall::instruction_size);
        // Block
        SafepointSynchronize::block(thread());       
        COMPILER1_ONLY(caller_fr.set_saved_receiver(receiver());)
      }
      break;

    case relocInfo::return_type:            
      { safepoint_msg("return_type");        
        HandleMark hm;

        // All arguments are dead at the return. Caller must ignore arguments
        _caller_must_gc_args = false;

        // See if return type is an oop.
        bool return_oop = nm->method()->is_returning_oop();
        Handle return_value;
        if (return_oop) {
          // The oop result has been saved on the stack together with all
          // the other registers. In order to preserve it over GCs we need
          // to keep it in a handle.
          return_value = caller_fr.saved_oop_result();  
          assert(return_value() == NULL || Universe::is_heap(return_value()), "must be heap pointer");          
        }

        // compute the number of arguments to pop
	// Do this before blocking as nmethod could be zombie by the time
	// we unblock.
        nmethod* nm = CodeCache::find_nmethod(real_return_addr);
        assert(nm != NULL, "cannot find nmethod");
        arg_size_in_bytes = (UseC2CallingConventions) ? 0 : 
                              nm->method()->size_of_parameters() * wordSize;
        
        SafepointSynchronize::block(thread());
        
        // restore oop result, if any
        if (return_oop) {
          caller_fr.set_saved_oop_result(return_value());        
        }       


        real_return_addr = caller_fr.pc();    // No change        
      }
      break;

    case relocInfo::safepoint_type:
       safepoint_msg("safepoint_type");
       // We are not at a call
       _caller_must_gc_args = false; 
       thread()->set_pc_not_at_call_for_frame(caller_fr.sp());

       // Setup return address in stub, so stack traversal works
       caller_fr.patch_pc(thread(), real_return_addr);
       SafepointSynchronize::block(thread());              
       break;
    
    default:
      safepoint_msg("Unexpected relocType: %d", type);
      ShouldNotReachHere();
  }  

  // Deoptimize frame if exception has been thrown. 
  if (thread()->has_pending_exception()) {    
    safepoint_msg("Exception installed during safepoint");   
#ifdef COMPILER1
    if (type == relocInfo::safepoint_type) {
      thread()->set_is_handling_implicit_exception(true); // will be reset in Runtime1::exception_handler_for_pc
    }
#else
    // Deoptimize frame. The compiler might not have any safepoint state at this point, so the frame
    // must have been deoptimized. The deoptimization happended in the SafepointSynchronize::end method     
#endif // COMPILER1
  } else {
    // patch return address to point to code in nmethod: it might have been changed above;
    // e.g., for an inline cache the real_return_addr has been moved back to also include
    // the move instruction for the inline cache.
    //
    // There is one situation where patching the real return address is wrong:
    // The calling activation might have been deoptimized while being blocked.
    // If the calling activation has been deoptimized we must leave the
    // return address pointing to the deoptimization stub. In all other
    // cases the real_return_address will point into the caller frame.
    frame stub_fr = thread()->last_frame();
    RegisterMap map(thread(), false);  
    frame caller_fr = stub_fr.sender_for_raw_compiled_frame(&map);   
    CodeBlob* caller_cb = CodeCache::find_blob(caller_fr.pc());
    // For the safepoint type "return_type", we've already popped the
    // frame of the calling compiled method. If the caller to the
    // compiled method which took the safepoint was itself compiled
    // code, this patch is a no-op; see the return_type arm of the
    // switch statement above. However, if the caller to the
    // safepointed method was interpreted, then the caller frame's PC
    // will be the interpreter's return entry point and the above
    // find_blob call will fail. This is actually okay.
    if ((caller_cb != NULL) && (!caller_cb->is_deoptimization_stub())) {
      assert(caller_cb->contains(real_return_addr), "current caller frame must contain return address");
      caller_fr.patch_pc(thread(), real_return_addr);
    } else {
      if (Verbose) tty->print_cr("Deoptimizing at a safepoint type %d", type);
      if (   type == relocInfo::virtual_call_type 
          || type == relocInfo::static_call_type
          || type == relocInfo::opt_virtual_call_type ) {
        // if there is no pending exception, the call may be skipped as the 
        // deoptimization at call will cause the interpreter to execute the next instruction
        if (!thread()->has_pending_exception()) {
          if (Verbose) warning("deoptimizing at a call at a compiler safepoint, will skip call -> problem known");
#ifdef ASSERT
	  DeoptimizationBlob* deopt_blob = NULL;
	  COMPILER1_ONLY(ShouldNotReachHere());
	  COMPILER2_ONLY(deopt_blob = OptoRuntime::deoptimization_blob());
          assert(deopt_blob != NULL, "deoptimization blob must have been created");
          assert(caller_fr.pc() == deopt_blob->unpack(), "cannot handle this address");
#endif // ASSERT
          assert(arg_size_in_bytes == 0, "only return_type sets it to non-zero");
          // find the destination address
          // as we planned to reexecute the call, real_return_addr points to the start 
          // of the call instruction

          // compute the address of the call (real_return_addr may point at the IC move)
          address at_call = type == relocInfo::virtual_call_type ?
                                      real_return_addr + NativeMovConstReg::instruction_size
                                    : real_return_addr;

          NativeCall* call = nativeCall_at(at_call);
          // replace return address with the destination of the call,
          // so that the stub can use it; the stub will replace the
          // return address with the entry into the deoptimization blob.
          caller_fr.patch_pc(thread(), call->destination());

          arg_size_in_bytes = -1; // notify that compilation must be reexecuted
        }
      }
    }
  }
      
  // flush any inadvertedly cached thread state
  thread()->set_last_Java_sp(thread()->last_Java_sp()); 
  return arg_size_in_bytes;
}

#endif

// ------------------------------------------------------------------------------------------------
// Non-product code

#ifndef PRODUCT
//
// Statistics
// 

long      SafepointSynchronize::_total_safepoints = 0;
long      SafepointSynchronize::_total_threads    = 0;
long      SafepointSynchronize::_max_threads      = 0;
double    SafepointSynchronize::_total_secs       = 0;
double    SafepointSynchronize::_max_secs         = 0;
double    SafepointSynchronize::_min_secs         = 3600; // One hour (should be high enough)
TimeStamp SafepointSynchronize::_timer;
long      SafepointSynchronize::_num_of_suspends = 0;
long      SafepointSynchronize::_cumm_num_of_suspends = 0;

void SafepointSynchronize::begin_statistics(int nof_threads) {
  _total_safepoints += 1;
  _total_threads    += nof_threads;
  if (nof_threads>_max_threads)
    _max_threads = nof_threads;
  _num_of_suspends = 0;

  _timer.update();
}

void SafepointSynchronize::end_statistics() {
  double secs = _timer.seconds();
  
  _total_secs += secs;
  if (_min_secs>secs) _min_secs = secs;
  if (_max_secs<secs) _max_secs = secs;
  _cumm_num_of_suspends += _num_of_suspends;

  print_statistics();
}

void SafepointSynchronize::print_statistics() {
  tty->print_cr("Safepoint#%Ld: Threads:(avg: %3.1f, max: %3Ld), Time: (avg: %2.5f, min: %2.5f, max:%2.5f), Suspends: (current: %Ld, cummulative: %Ld)",      
    _total_safepoints,
    (double) _total_threads/_total_safepoints,
    _max_threads,
    _total_secs/_total_safepoints,
    _min_secs,
    _max_secs,
    _num_of_suspends,
    _cumm_num_of_suspends);
}

void SafepointSynchronize::print_state() {
  if (_state == _not_synchronized) {
    tty->print_cr("not synchronized");
  } else if (_state == _synchronizing || _state == _synchronized) {
    tty->print_cr("State: %s", (_state == _synchronizing) ? "synchronizing" : "synchronized");

    for(JavaThread *cur = Threads::first(); cur; cur = cur->next()) {       
       cur->safepoint_state()->print();
    }  
  } 
}

void SafepointSynchronize::safepoint_msg(const char* format, ...) {
  if (ShowSafepointMsgs) {
    va_list ap;
    va_start(ap, format);
    tty->vprint_cr(format, ap);
    va_end(ap);
  }
}

#endif
