/*
 * @(#)safepoint.hpp	1.68 01/02/09 14:47:51
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2
//
// Safepoint synchronization
////
// The VM-thread uses the SafepointSynchronzie::begin/end methods to enter/exit a
// safepoint region. The begin method will roll all JavaThreads forward to a safepoint.
//
// JavaThreads must use the Safepoint abstraction (defined in Thread.hpp) to indicate that
// that they are at a safepoint
//
// The Mutex/Condition vatiable and ObjectLocker classes calls the enter/
// exit safepoint methods, when a thread is blocked/restarted. Hence, all mutex exter/
// exit points *must* be at a safepoint.


class ThreadSafepointState;
class SnippetCache;
class nmethod;
class SafepointHandler;

//
// Implements roll-forward to safepoint (safepoint synchronization)
// 
class SafepointSynchronize : AllStatic {
 public:
  enum SynchronizeState {
      _not_synchronized = 0,                   // Threads not synchronized at a safepoint
      _synchronizing    = 1,                   // Synchronizing in progress  
      _synchronized     = 2                    // All Java threads are stopped at a safepoint. Only VM thread is running
  };
 private:
  static volatile SynchronizeState _state;     // Threads might read this flag directly, without acquireing the Threads_lock
  static volatile int _waiting_to_block;       // No. of threads we are waiting for to block.
  
  static long     _last_safepoint;             // Time of last safepoint
  // statistics
#ifndef PRODUCT
  static long       _total_safepoints;
  static long       _total_threads;
  static long       _max_threads;
  static long       _num_of_suspends;
  static long       _cumm_num_of_suspends;
  static double     _total_secs;
  static double     _max_secs;
  static double     _min_secs;
  static TimeStamp  _timer;
#endif

  static void begin_statistics(int nof_threads) PRODUCT_RETURN;
  static void end_statistics()                  PRODUCT_RETURN;
  static void print_statistics()                PRODUCT_RETURN;

  
public:

  // Main entry points
  static void begin();                  // Roll all threads forward to safepoint (must be called by VM thread)
  static void end();                    // Start all suspended threads again...

  // Machine-depended methods        
  static bool can_be_at_safepoint_before_suspend  (JavaThread *thread, JavaThreadState state);
  static bool can_be_at_safepoint_after_suspend   (JavaThread *thread, JavaThreadState state);
  static void save_thread_state_at_safepoint      (JavaThread *thread);
  static void patch_return_instruction_md         (address cb_pc);
  static bool pd_is_cleanups_needed();

  // Query
  static bool is_at_safepoint()                           { return _state == _synchronized;  }
  static bool is_synchronizing()                          { return _state == _synchronizing;  }
  static bool do_call_back()                              { return _state != _not_synchronized; }
  
  // Called when a thread volantary blocks
  static void   block(JavaThread *thread);
  static void   signal_thread_at_safepoint()              { _waiting_to_block--; }  

  // Exception handling for illegal bytecode patching
  static int  handle_illegal_instruction_exception(JavaThread *thread);

  // VM Thread interface for determining safepoint rate
  static long last_non_safepoint_interval()               { return os::javaTimeMillis() - _last_safepoint; }
  static bool is_cleanup_needed();
  static void do_cleanup_tasks();

  // debugging
  static void print_state()                                PRODUCT_RETURN;
  static void safepoint_msg(const char* format, ...)       PRODUCT_RETURN;
  #ifndef PRODUCT
  static void log_suspension()                             { _num_of_suspends += 1; }   
  #endif
  static void set_is_at_safepoint()                        { _state = _synchronized; }
  static void set_is_not_at_safepoint()                    { _state = _not_synchronized; }

  // assembly support
  static address address_of_state()                        { return (address)&_state; }  
};


// State class for a thread suspended at a safepoint
class ThreadSafepointState: public CHeapObj {
 public:  
  enum suspend_type {    
    _running                =  0, // Thread state not yet determined (i.e., not at a safepoint yet)
    _at_safepoint           =  1, // Thread at a safepoint (f.ex., when blocked on a lock)
    _at_safepoint_suspended =  2, // Thread at a safepoint and is suspended
    _call_back              =  3, // Keep executing and wait for callback (if thread is in interpreter or vm)
    _compiled_safepoint     =  4  // Compiled safepoint
  };
 private:
  suspend_type          _type;  
  JavaThread *          _thread;    
  // use an ExtendedPC to be sure we have real containment -- on SPARC the pc
  // might be at the delay slot of a branch so we don't contain the PC, but we do
  // contain the nPC
  ExtendedPC            _addr;
  SafepointHandler*     _handle;  // Used when _type == _safepoint_handler

  // Thead local code buffer
  NOT_CORE(ThreadCodeBuffer* _code_buffer;)

 public:       
  ThreadSafepointState(JavaThread *thread);
  ~ThreadSafepointState();

  // examine/roll-forward/restart   
  void examine_state_of_thread(int iterations);               
  void roll_forward(suspend_type type, nmethod* nm = NULL, bool disable_resume_for_running_thread = false);
  void restart();  
  
  // Query  
  JavaThread*         thread() const	      { return _thread; }
  suspend_type        type() const	      { return _type; }    
  bool                is_running() const      { return (_type==_running); } 
  ExtendedPC          current_address() const { return _addr; } 
  SafepointHandler*   handle() const          { return _handle; }
      
  bool caller_must_gc_arguments() const;  
#ifndef CORE
  // Local code-cache management
  ThreadCodeBuffer* allocate_code_buffer(int size_in_bytes, nmethod *code, address real_pc);
  void              destroy_code_buffer();
  ThreadCodeBuffer* code_buffer()        { return _code_buffer; }

  // adjustment of pc if it points into a thread local codebuffer (might block)
  inline address compute_adjusted_pc(address pc);

  // the inverse adjustment of pc, if it can be redirected into the codebuffer:
  inline address maybe_capture_pc(address pc);
#endif

  // debugging  
  void print(); 
  JavaThreadState _stop_state; // Usefull for debugging a deadlock (to be taken out)
  address         _stop_pc;    // Usefull for debugging a deadlock (to be taken out)

  // Initialize
  static void create(JavaThread *thread);
  static void destroy(JavaThread *thread);
};

#ifndef CORE
// Inlined code

inline address ThreadSafepointState::compute_adjusted_pc(address pc) {
  ThreadCodeBuffer *cb = code_buffer();
  if (cb != NULL && cb->contains(pc)) pc = cb->compute_adjusted_pc(pc);    
  return pc;
}

inline address ThreadSafepointState::maybe_capture_pc(address pc) {
  ThreadCodeBuffer *cb = code_buffer();
  if (cb != NULL && cb->captures(pc)) pc = cb->capture_pc(pc);
  return pc;
}

#endif
// ---------------------------------------------------------------------------------------------------------------
// Specialized safepoint handlers for different parts of the code

class SafepointHandler : public CHeapObj {
  JavaThread *            _thread;  
 public:
  SafepointHandler(JavaThread *thread)   { _thread = thread; }  

  // Main virtuals. Initiates safepoint operation for thread, and release thread after safepoint operation is
  // completed.
  virtual void setup    (ThreadSafepointState *state, nmethod* nm) = 0;
  virtual void release  (ThreadSafepointState *state) = 0;
  virtual const char *name()   = 0;  
  virtual bool is_compiled_code_safepoint_handler()     { return false; }
  

  // Call-backs from safepoint code
  virtual int  handle_illegal_instruction_exception()   { ShouldNotReachHere(); return 0;}

  virtual bool caller_must_gc_arguments() const         { ShouldNotReachHere(); return false; };  

  // query
  JavaThread* thread()  { return _thread; }

  // general helper methods
  void set_state_to_running();

  // debugging
  virtual void print()            { tty->print_cr("SafepointHandler: %s", name()); }
  void safepoint_msg(const char* format, ...) {
    if (ShowSafepointMsgs) {
      va_list ap;
      va_start(ap, format);
      tty->vprint_cr(format, ap);
      va_end(ap);
    }
  }
};

// -------------------------------------------------------------------------------------------------------------
// Specialized Safepoint handler for compiled code.
#ifndef CORE

class CompiledCodeSafepointHandler: public SafepointHandler {
  nmethod *_nm;
  bool    _caller_must_gc_args;    
public:
  CompiledCodeSafepointHandler(JavaThread *thread) : SafepointHandler(thread) { _nm = NULL; };  

  void setup  (ThreadSafepointState *state, nmethod* nm);
  void release(ThreadSafepointState *state);  
  int  handle_illegal_instruction_exception();
  
  bool is_compiled_code_safepoint_handler() { return true; }
  
  const char *name()          { return "CompiledCodeSafepointHandler"; }
  nmethod *get_nmethod()      { return _nm; }  

  // Called by SafepointBlob
  bool caller_must_gc_arguments() const { return _caller_must_gc_args; };
};

#endif
