/*
 * @(#)signature.hpp	1.27 00/04/25 18:04:15
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// SignatureIterators iterate over a Java signature (or parts of it).
// (Syntax according to: "The Java Virtual Machine Specification" by
// Tim Lindholm & Frank Yellin; section 4.3 Descriptors; p. 89ff.)
//
// Example: Iterating over ([Lfoo;D)I using
//                         0123456789
//
// iterate_parameters() calls: do_array(2, 7); do_double();
// iterate_returntype() calls:                              do_int();
// iterate()            calls: do_array(2, 7); do_double(); do_int();
//
// is_return_type()        is: false         ; false      ; true
//
// NOTE: The new optimizer has an alternate, for-loop based signature
// iterator implemented in opto/type.cpp, TypeTuple::make().

class SignatureIterator: public ResourceObj {
 protected:
  symbolHandle _signature;             // the signature to iterate over
  int          _index;                 // the current character index (only valid during iteration)
  int          _parameter_index;       // the current parameter index (0 outside iteration phase)
 
  void expect(char c);
  void skip_optional_size();
  int  parse_type();                   // returns the parameter size in words (0 for void)
  void check_signature_end();

 public:
  // Constructor
  SignatureIterator(symbolHandle signature);

  // Iteration
  void dispatch_field();               // dispatches once for field signatures
  void iterate_parameters();           // iterates over parameters only
  void iterate_returntype();           // iterates over returntype only
  void iterate();                      // iterates over whole signature

  // Returns the word index of the current parameter; returns a negative value at the return type 
  int  parameter_index() const         { return _parameter_index; }
  bool is_return_type() const          { return parameter_index() < 0; }

  // Basic types
  virtual void do_bool  ()             = 0;
  virtual void do_char  ()             = 0;
  virtual void do_float ()             = 0;
  virtual void do_double()             = 0;
  virtual void do_byte  ()             = 0;
  virtual void do_short ()             = 0;
  virtual void do_int   ()             = 0;
  virtual void do_long  ()             = 0;
  virtual void do_void  ()             = 0;

  // Object types (begin indexes the first character of the entry, end indexes the first character after the entry)
  virtual void do_object(int begin, int end) = 0;
  virtual void do_array (int begin, int end) = 0;
};


// Specialized SignatureIterators: Used to compute signature specific values.

class SignatureTypeNames : public SignatureIterator {
 protected:
  virtual void type_name(const char* name)   = 0;

  void do_bool()                       { type_name("jboolean"); }
  void do_char()                       { type_name("jchar"   ); }
  void do_float()                      { type_name("jfloat"  ); }
  void do_double()                     { type_name("jdouble" ); }
  void do_byte()                       { type_name("jbyte"   ); }
  void do_short()                      { type_name("jshort"  ); }
  void do_int()                        { type_name("jint"    ); }
  void do_long()                       { type_name("jlong"   ); }
  void do_void()                       { type_name("void"    ); }
  void do_object(int begin, int end)   { type_name("jobject" ); }
  void do_array (int begin, int end)   { type_name("jobject" ); }

 public:
  SignatureTypeNames(symbolHandle signature) : SignatureIterator(signature) {}
};


class SignatureInfo: public SignatureIterator {
 protected:
  bool      _has_iterated;             // need this because iterate cannot be called in constructor (set is virtual!)
  int       _size;
  BasicType _type;

  void lazy_iterate()                  { if (!_has_iterated) { iterate(); _has_iterated = true; } }

  virtual void set(int size, BasicType type) = 0;

  void do_bool  ()                     { set(T_BOOLEAN_size, T_BOOLEAN); }
  void do_char  ()                     { set(T_CHAR_size   , T_CHAR   ); }
  void do_float ()                     { set(T_FLOAT_size  , T_FLOAT  ); }
  void do_double()                     { set(T_DOUBLE_size , T_DOUBLE ); }
  void do_byte  ()                     { set(T_BYTE_size   , T_BYTE   ); }
  void do_short ()                     { set(T_SHORT_size  , T_SHORT  ); }
  void do_int   ()                     { set(T_INT_size    , T_INT    ); }
  void do_long  ()                     { set(T_LONG_size   , T_LONG   ); }
  void do_void  ()                     { set(T_VOID_size   , T_VOID   ); }
  void do_object(int begin, int end)   { set(T_OBJECT_size , T_OBJECT ); }
  void do_array (int begin, int end)   { set(T_ARRAY_size  , T_ARRAY  ); }

 public:
  SignatureInfo(symbolHandle signature) : SignatureIterator(signature) {
    _has_iterated = false;
    _size         = 0;
    _type         = T_ILLEGAL;
  }

  int       size()                     { lazy_iterate(); return _size; }
  BasicType type()                     { lazy_iterate(); return _type; }
};


// Specialized SignatureIterator: Used to compute the argument size.

class ArgumentSizeComputer: public SignatureInfo {
 private:
  void set(int size, BasicType type)   { if (!is_return_type()) _size += size; }
 public:
  ArgumentSizeComputer(symbolHandle signature) : SignatureInfo(signature) {} 
};


class ArgumentCount: public SignatureInfo {
 private:
  void set(int size, BasicType type)   { if (!is_return_type()) _size ++; }
 public:
  ArgumentCount(symbolHandle signature) : SignatureInfo(signature) {} 
};


// Specialized SignatureIterator: Used to compute the result type.

class ResultTypeFinder: public SignatureInfo {
 private:
  void set(int size, BasicType type)   { if (is_return_type()) _type = type; }
 public:
  ResultTypeFinder(symbolHandle signature) : SignatureInfo(signature) {}
};


// Specialized SignatureIterator: Used for native call purposes

class NativeSignatureIterator: public SignatureIterator {
 private:
  methodHandle _method;
  int          _offset;                // the current parameter offset, starting with 0
  int          _prepended;             // number of prepended JNI parameters (1 JNIEnv, plus 1 mirror if static)

  void do_bool  ()                     { pass_int();    _offset++;    }
  void do_char  ()                     { pass_int();    _offset++;    }
  void do_float ()                     { pass_int();    _offset++;    }
  void do_double()                     { pass_double(); _offset += 2; }
  void do_byte  ()                     { pass_int();    _offset++;    }
  void do_short ()                     { pass_int();    _offset++;    }
  void do_int   ()                     { pass_int();    _offset++;    }
  void do_long  ()                     { pass_long();   _offset += 2; }
  void do_void  ()                     { ShouldNotReachHere();        }
  void do_object(int begin, int end)   { pass_object(); _offset++;    }
  void do_array (int begin, int end)   { pass_object(); _offset++;    }

 public:
  methodHandle method() const          { return _method; }
  int          offset() const          { return _offset; }
  int      jni_offset() const          { return _offset + _prepended; }
  int     java_offset() const          { return method()->size_of_parameters() - _offset - 1; }
  bool      is_static() const          { return method()->is_static(); }
  virtual void pass_int()              = 0;
  virtual void pass_long()             = 0;
  virtual void pass_object()           = 0;
  virtual void pass_double()           { pass_long(); }  // may be same as long

  NativeSignatureIterator(methodHandle method) : SignatureIterator(method->signature()) {
    _method = method;
    _offset = 0;

    const int JNIEnv_words = 1;
    const int mirror_words = 1;
    _prepended = !is_static() ? JNIEnv_words : JNIEnv_words + mirror_words;
  }

  // iterate() calles the 2 virtual methods according to the following invocation syntax:
  //
  // {pass_int | pass_long | pass_object}
  //
  // Arguments are handled from left to right (receiver first, if any).
  // The offset() values increase from 0.
  // The java_offset() values count down to 0, and refer to the Java TOS.
  // The jni_offset() values increase from 1 or 2, and refer to C arguments.
  void iterate() {
    if (!is_static()) {
      // handle receiver (not handled by iterate because not in signature)
      pass_object(); _offset++;
    }
    iterate_parameters();
  }  
};


// Handy stream for iterating over signature

class SignatureStream : public StackObj {
 private:
  symbolHandle _signature;
  int          _begin;
  int          _end;
  BasicType    _type;
  bool         _at_return_type;

 public:
  bool at_return_type() const                    { return _at_return_type; }
  bool is_done() const;
  void next();

  SignatureStream(symbolOop signature, bool is_method = true); 
  bool is_object() const;	                 // True if this argument is an object
  BasicType type() const                         { return _type; }
  symbolOop as_symbol(TRAPS);
  
  // return same as_symbol except allocation of new symbols is avoided.
  symbolOop as_symbol_or_null();
};
