/*
 * @(#)sweeper.cpp	1.25 01/02/09 14:47:56
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_sweeper.cpp.incl"

CodeBlob* NMethodSweeper::_current = NULL;   // Current nmethod
int       NMethodSweeper::_seen = 0 ;        // No. of blobs we have currently processed in current pass of CodeCache
int       NMethodSweeper::_invocations = 0;  // No. of invocations left until we are completed with this pass
long      NMethodSweeper::_nof_zombies = 0;  // No. of zombies (i.e., any work to do?)
long      NMethodSweeper::_nof_not_entrants = 0; // No. of zombies (i.e., any work to do?)
long      NMethodSweeper::_nof_unloaded = 0; // No. of zombies (i.e., any work to do?)

void NMethodSweeper::sweep() {   
  assert(SafepointSynchronize::is_at_safepoint(), "must be executed at a safepoint");
  if (!MethodFlushing) return;  

  // No need to synchronize access, since this is always executed at a safepoint
  if (_nof_zombies == 0 && _nof_not_entrants == 0 && _nof_unloaded == 0) return;

  // Make sure CompiledIC_lock in unlocked, since we might update some
  // inline caches. If it is, we just bail-out and try later.
  if (CompiledIC_lock->is_locked() || Patching_lock->is_locked()) return;
  
  // Check for restart
  if (_current == NULL) {
    _seen        = 0;
    _invocations = NmethodSweepFraction;
    _current     = CodeCache::first();
  }

  if (PrintMethodFlushing && Verbose) {
    tty->print_cr("### Sweep at %d out of %d. Invocations left: %d", _seen, CodeCache::nof_blobs(), _invocations);
  }

  // We want to visit all nmethods after NmethodSweepFraction invocations. 
  // If invocation is 1 we do the rest
  int todo = CodeCache::nof_blobs();
  if (_invocations != 1) {
    todo = (CodeCache::nof_blobs() - _seen) / _invocations;    
    _invocations--;
  }
      
  for(int i = 0; i < todo && _current != NULL; i++) {      
    CodeBlob* next = CodeCache::next(_current); // Read next before we potentially delete current
    if (_current->is_nmethod()) {
      process_nmethod((nmethod *)_current);      
    }    
    _seen++;
    _current = next;
  }  
}

   
void NMethodSweeper::process_nmethod(nmethod *nm) {  
  // Skip methods that are currently referenced by the VM
  if (nm->is_locked_by_vm()) {
    // But still remember to clean-up inline caches for alive nmethods
    if (nm->is_alive()) {
      // Clean-up all inline caches that points to zombie/non-reentrant methods
      nm->cleanup_inline_caches();     
    }
    return;
  }

  if (nm->is_zombie()) {
    // If it is first time, we see nmethod then we mark it. Otherwise,
    // we reclame it. When we have seen a zombie method twice, we know that
    // there are no inline caches that referes to it.
    if (nm->is_marked_for_reclamation()) {
      assert(!nm->is_locked_by_vm(), "must not flush locked nmethods");
      nm->flush();      
      _nof_zombies--;
      assert(_nof_zombies >= 0, "cannot go negative");
    } else {
      nm->mark_for_reclamation();
    }  
  } else if (nm->is_not_entrant()) {    
    // If there is no current activations of this method on the
    // stack we can safely convert it to a zombie method
    if (nm->can_not_entrant_be_converted()) {
      nm->make_zombie();
      _nof_not_entrants--;
      assert(_nof_not_entrants >= 0, "cannot go negative");
    } else {
      // Still alive, clean up its inline caches 
      nm->cleanup_inline_caches();    
    }
  } else if (nm->is_unloaded()) {
    // Unloaded code, just make it a zombie 
    if (nm->is_osr_only_method()) {
      // No inline caches will ever point to osr methods, so we can just remove it
      nm->flush();
    } else {
      nm->make_zombie();
    }
    _nof_unloaded--;
  } else {
    assert(nm->is_alive(), "should be alive");
    // Clean-up all inline caches that points to zombie/non-reentrant methods
    nm->cleanup_inline_caches();    
  }
}

void NMethodSweeper::notify(nmethod *nm) {  
  if (nm->is_zombie()) {
    atomic::increment(&_nof_zombies);
  } else if (nm->is_not_entrant()) {
    atomic::increment(&_nof_not_entrants);
  } else if (nm->is_unloaded()) {
    atomic::increment(&_nof_unloaded);  
  } else {
    ShouldNotReachHere();
  }  
}
