/*
 * @(#)thread.cpp	1.565 01/04/25 16:57:24
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_thread.cpp.incl"

// Class hierarchy
// - Thread
//   - VMThread
//   - WatcherThread
//   - SuspendCheckerThread
//   - JavaThread

// ======= Thread ========

// Base class for all threads: VMThread, WatcherThread, SuspendCheckerThread,
// JavaThread

Thread::Thread() {
  // stack  
  _stack_base   = NULL;
  _stack_limit  = NULL;
  _stack_size   = 0;
  _last_sp_hash = -1;
  _fast_thread_id = -1;
  _osthread     = NULL;

  // allocated data structures
  set_resource_area(new ResourceArea());
  set_handle_area(new HandleArea(NULL));
  set_active_handles(NULL); 
  set_free_handle_block(NULL);
  set_last_handle_mark(NULL);
  set_osthread(NULL);

  // the handle mark links itself to last_handle_mark
  new HandleMark(this);

  // plain initialization  
  debug_only(_owned_locks = NULL;) 
  debug_only(_allow_allocation_count = 0;)
  NOT_PRODUCT(_allow_safepoint_count = 0;)  
  _highest_lock = NULL;
  _vm_operation_started_count = 0;
  _vm_operation_completed_count = 0;
  set_unboosted_priority(NoPriority);
  _pending_exception = NULL;
  _rawmonitor_list = NULL;
}

void Thread::initialize_thread_local_storage() {
  // Note: Make sure this method only calls 
  // non-blocking operations. Otherwise, it might not work
  // with the thread-startup/safepoint interaction.

  // During Java thread startup, safepoint code should allow this
  // method to complete because it may need to allocate memory to
  // store information for the new thread.

  // initialize structure dependent on thread local storage
  ThreadLocalStorage::set_thread(this);
  
  // set up any platform-specific state.
  os::initialize_thread();

  // Used for speculative locking
  set_fast_thread_id();
}

void Thread::delete_thread_and_TLS_current() {
  bool mythread = (Thread::current() == this); 
  delete this;
  // clear thread local storage
  if (mythread) {
    ThreadLocalStorage::set_thread(NULL);
  }
}

void Thread::record_stack_base_and_size() {
  set_stack_base(os::current_stack_base());
  set_stack_size(os::current_stack_size());
}


Thread::~Thread() {
  // deallocate data structures
  delete resource_area();
  // since the handle marks are using the handle area, we have to deallocated the root
  // handle mark before deallocating the thread's handle area,
  assert(last_handle_mark() != NULL, "check we have an element");
  delete last_handle_mark();
  assert(last_handle_mark() == NULL, "check we have reached the end");

  delete handle_area();

  // osthread() can be NULL, if creation of thread failed.
  if (osthread() != NULL) os::free_thread(osthread());
  
  // clear thread local storage - moved to after delete calls
  // so heap tracing would work
}

// NOTE: dummy function for assertion purpose.
void Thread::run() {
  ShouldNotReachHere();
}

#ifdef ASSERT
// Private method to check for dangling thread pointer
void check_for_dangling_thread_pointer(Thread *thread) {
 assert(!thread->is_Java_thread() || Thread::current() == thread || Threads_lock->owned_by_self(),
         "possibility of dangling Thread pointer");
}
#endif


#ifndef PRODUCT
// Tracing method for basic thread operations
void Thread::trace(const char* msg, Thread* thread) {    
  if (!TraceThreadEvents) return;
  ResourceMark rm;  
  ThreadCritical tc;
  const char *name = "non-Java thread";
  int prio = -1;
  if (thread->is_Java_thread() 
#ifndef CORE
      && !thread->is_Compiler_thread()
#endif
      ) {
    // The Threads_lock must be held to get information about
    // this thread but may not be in some situations when
    // tracing  thread events.
    bool release_Threads_lock = false;
    if (!Threads_lock->owned_by_self()) {
      Threads_lock->lock();
      release_Threads_lock = true;
    }
    JavaThread* jt = (JavaThread *)thread;
    name = (char *)jt->get_thread_name();
    oop thread_oop = jt->threadObj();
    if (thread_oop != NULL) {
      prio = java_lang_Thread::priority(thread_oop); 
    }
    if (release_Threads_lock) {
      Threads_lock->unlock();
    }
  }    
  tty->print_cr("Thread::%s 0x%x [%x] %s (prio: %d)", msg, thread, thread->osthread()->thread_id(), name, prio);
}
#endif


ThreadPriority Thread::get_priority(Thread* thread) {  
  trace("get priority", thread);  
  ThreadPriority priority;
  // Can return an error!
  (void)os::get_priority(thread, priority);  
  assert(MinPriority <= priority && priority <= MaxPriority, "non-Java priority found");
  return priority;
}

void Thread::set_priority(Thread* thread, ThreadPriority priority) {
  trace("set priority", thread);
  debug_only(check_for_dangling_thread_pointer(thread);)
  // Can return an error!
  (void)os::set_priority(thread, priority);
}

void Thread::boost_priority(Thread* thread, ThreadPriority priority) {    
  trace("boost priority", thread);
  debug_only(check_for_dangling_thread_pointer(thread);)
  
  thread->set_unboosted_priority(Thread::get_priority(thread));
  Thread::set_priority(thread, priority);
}

void Thread::unboost_priority(Thread* thread) {      
  debug_only(check_for_dangling_thread_pointer(thread);)
  if (thread->unboosted_priority() != NoPriority) {
    trace("unboost priority", thread);
    Thread::set_priority(thread, thread->unboosted_priority());
    thread->set_unboosted_priority(NoPriority);
  }
}

void Thread::force_stack_page_mapping_to(address addr) {
  // Prevent space_waster from being optimized away.
  char space_waster[512 * sizeof(int)];
  space_waster[0] = 0;
  space_waster[sizeof(space_waster)-1] = 0;
  strlen(space_waster);
  
  unsigned int current_page = (unsigned int)&current_page / os::vm_page_size();
  unsigned int target_page = (unsigned int)addr / os::vm_page_size();
  if(current_page > target_page) {
    force_stack_page_mapping_to(addr);
  }
}

void Thread::start(Thread* thread) {
  trace("start", thread);
  // Start is different from resume in that its safety is guaranteed by context or
  // being called from a Java method synchronized on the Thread object.
  if (!DisableStartThread) {
    os::start_thread(thread);
  }
}

// Enqueue a VM_Operation to do the job for us - sometime later
void Thread::send_async_exception(oop java_thread, oop java_throwable) {    
  VM_ThreadStop* vm_stop = new VM_ThreadStop(java_thread, java_throwable);
  VMThread::execute(vm_stop);
}

// Suspend another thread. Returns the current suspend count.
// Used for safepoint synchronization by the VMThread, flat profiling
// by the WatcherThread, and by both JVM/DI and JVM/PI for suspending
// threads in state _thread_in_native.
// Also used for VM_Exit
long Thread::suspend_other(bool ext_req) {
  trace("suspend_other", this);  

  assert(Thread::current() != this, "suspend_other() cannot suspend self");

  if (this->is_SuspendChecker_thread()) {
    return 0;  // ignore SuspendCheckerThread
  }
  if (is_Java_thread()) {
    // mark the thread for the appropriate type of suspension
    if (ext_req) {
      ((JavaThread *)this)->set_external_suspend();
    } else {
      ((JavaThread *)this)->set_internal_suspend();
    }
  }
  SuspendThread_Callback cb;
  return suspend_thread_impl(cb, ext_req);
}

// suspend_thread_self:
// Used for Solaris FetchTopFrame to force a thread which was already
// suspended to be resumed, its state saved and then resuspended.
// Unlike self-suspension triggered by send_java_suspend(), this form
// of self-suspension is an internal request.
// Also used on VM_Exit
long Thread::suspend_thread_self(Suspend_Callback& cb) {
  assert(Thread::current() == this, "self-suspension by another?");
  assert(!this->is_SuspendChecker_thread(), "SuspendCheckThread should not suspend self");

  if (is_Java_thread()) {
    JavaThread *java_thread = (JavaThread *)this;
    bool is_suspended;
    {
      MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
      is_suspended = java_thread->is_suspended();
    }
    assert(!is_suspended,
      "a thread that is self-suspending should not already be suspended");
    java_thread->set_internal_suspend();
  }
  return suspend_thread_impl(cb, false);  // internal suspend request
  // it is important that the callback is not touched after we resume here
  // since its parts may be cancelled when we get here from FetchTopFrame
  // or other handles for synchronous callbacks.
} 

// shutdown suspend. Used by Solaris and Linux to
// force other threads to suspend prior to exit, so
// they do not access things like CodeHeap thread 1 has freed
// Windows does not use this, with jvmpi to stdout
// can cause deadlock if jvmpi daemon thread holds write lock
// and thread 1 tries to write to stdout
long Thread::shutdown_suspend() {
  if (Thread::current() == this) {
    SuspendThreadSelf_Callback cb;
    return suspend_thread_self(cb);
  } else {
    return suspend_other(false);   // internal_suspend
  }
}

// suspend_thread_impl is the platform independent portion of the
// thread suspension implementation. When we are suspending another
// thread, we are given a SuspendThread_Callback object that gives
// us access to the platform specific portion of the implementation.
// When a JavaThread is suspending itself, we are given a
// SuspendThreadSelf_Callback object that allows us to use a condition
// variable to suspend ourself instead of using the platform specific
// implementation.
//
// The current suspend depth is returned or -1 is returned if the
// suspend operation is cancelled (while we were waiting for a lock).

long Thread::suspend_thread_impl(Suspend_Callback& cb, bool ext_req) {
  // We can be called from:
  // 1. a JavaThread; JavaThreads doing self-suspension via
  //    handle_special_runtime_exit_condition do NOT hold Threads_lock
  // 2. the WatcherThread (for flat profiling)
  // 3. the VMThread while synchronizing

  // Don't use safepoint check because we might be the WatcherThread.
  MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);

  OSThread* osthread = this->osthread();
  int depth = osthread->suspend_depth() + 1;
  ThreadState state = osthread->state();
  ThreadState state_before_suspend = (ThreadState) osthread->state_before_suspend();

  // If this is a nested suspend or if this is a baby thread,
  // then just set the updated depth.
  // Note: we leave state unchanged for baby threads.
  if (depth > 1 || state == ALLOCATED || state == INITIALIZED) {
    osthread->set_suspend_depth(depth);
    return depth;
  }

  // at this point, this is the first suspend (self-suspend or physical)

  if (is_Java_thread()) {
    JavaThread *java_thread = (JavaThread *)this;
    assert(!java_thread->is_suspended(), "thread should not be suspended");
    if (java_thread->is_suspended()) return -1;  // robustness

    if (ext_req) {
      // if external suspension was cancelled get out
      if (!java_thread->is_external_suspend()) return -1;
    } else {
      // internal suspend/resume is tightly coupled so we can't be cancelled
      assert(java_thread->is_internal_suspend(), "just checking");
      if (!java_thread->is_internal_suspend()) return -1;
    }
    // Since we are holding the Interrupt_lock, is_suspended() cannot
    // be true while osthread->state() != SUSPENDED.
    java_thread->set_is_suspended();
    assert(java_thread->is_external_suspend() ||
      java_thread->is_internal_suspend(), "suspend type must be set");
  }

  // Self suspension will do an Interrupt_lock.wait() in execute().
  // Otherwise, platform independent suspension will take place.

  if (cb.execute(this) != 0) {
    assert(false, "suspend thread failed");
    // Limited recovery in production mode - just reset the thread's
    // state and effectively drop the suspend request.
    // XXX - can we do better?
    osthread->set_state(state);
    osthread->set_state_before_suspend(state_before_suspend);
    osthread->set_suspend_depth(depth - 1);
    if (is_Java_thread()) {
      JavaThread *java_thread = (JavaThread *)this;
      if (ext_req) {
        java_thread->clear_external_suspend();
      } else {
        java_thread->clear_internal_suspend();
      }
      java_thread->clear_is_suspended();
    }
    return -1;
  }

  // When we get here, a previously self-suspended thread has been
  // resumed. Otherwise, the thread is suspended.
  assert(cb.is_self_suspend_cb() || osthread->state() == SUSPENDED,
    "thread is not suspended");

  return depth;
}


// This wrapper around resume_thread_impl provides a place for creating
// the ResumeThread_Callback to gain access to the platform specific
// resume semantics. It also provides tracing support, caller sanity
// checking and optional yield semantics.

long Thread::resume(bool do_yield, bool ext_req) {
  trace("resume", this);
  // Called by:
  // - an external resume request
  // - the VMThread (safepoint synchronization, FetchTopFrame_Callback)
  // - the SuspendCheckerThread (deadlock prevention)
  // - the WatcherThread (flat profiler, FetchTopFrame_Callback)
  assert(!Thread::current()->is_Java_thread() || Threads_lock->owned_by_self(),
    "JavaThread calling resume must hold Threads_lock");
  ResumeThread_Callback cb;
  long res = resume_thread_impl(cb, ext_req);
  if (do_yield) {
    os::yield_all(0);  // give a resumed thread a chance to be dispatched
  }
  return res;
}


// resume_thread_impl is the counter part to suspend_thread_impl.
// See the comment before suspend_thread_impl. Platform specific
// resumption mechanisms are provided by the Resume_Callback object.
// XXX - this does not have to be exposed since only resume() calls it.

long Thread::resume_thread_impl(Resume_Callback& cb, bool ext_req) {
  // %%% does the watcher have the Threads_lock / are we protected against dangling Thread pointer?

  MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);

  // When we get here, the target JavaThread may or may not be suspended.
  // We want to do the checking and clearing of is_external_suspend,
  // is_internal_suspend, is_suspended and osthread->state == SUSPENDED
  // while holding the Interrupt_lock.
  //
  // When we get here, the target JavaThread suspension state is:
  // 1) suspension no longer requested
  // 2) is_external_suspend() and/or is_internal_suspend() true, but
  //    is_suspended() not yet true
  // 3) already "physically" suspended (osthread->state == SUSPENDED)

  OSThread* osthread = this->osthread();

  if (is_Java_thread()) {
    JavaThread *java_thread = (JavaThread *)this;
    if (osthread->state() != SUSPENDED) {
      if (!java_thread->is_suspended()) {
        if (java_thread->is_external_suspend()) {
          // external request can only clear pending external suspend
          if (ext_req) {
            java_thread->clear_external_suspend();
            return 0;
          }
        } else if (java_thread->is_internal_suspend()) {
          // internal request can only clear pending internal suspend
          if (!ext_req) {
            java_thread->clear_internal_suspend();
            return 0;
          }
        }
        // if not a baby thread special case
        else if (!((osthread->state() == ALLOCATED ||
            osthread->state() == INITIALIZED) &&
            osthread->suspend_depth() > 0)) {
	  assert(ext_req, "internal resume must match internal suspend");
	  // Another external resume request got through before us so
	  // get out now.
          return 0;
        }
      }
#ifdef ASSERT
      else {
        // suspend_thread_impl() holds the Interrupt_lock while both
        // is_suspended() and osthread->state() are set so this case
        // cannot happen.
        assert(osthread->state() == SUSPENDED, "resuming non-suspended JavaThread");  
      }
#endif /* ASSERT */
    } 
  }
#ifdef ASSERT
  else {
    assert(osthread->state() == SUSPENDED, "resuming non-suspended thread");  
  }
#endif /* ASSERT */

  int depth = osthread->suspend_depth() - 1;
  ThreadState state = osthread->state();
  ThreadState state_before_suspend = (ThreadState) osthread->state_before_suspend();

  if (depth < 0) {
    assert(false, "suspend_depth should not already be zero");
    return 0;  // robustness
  }

  // If this is a nested resume or if this is a baby thread,
  // then just set the updated depth.
  // Note: we leave state unchanged for baby threads.
  else if (depth > 0 || state == ALLOCATED || state == INITIALIZED) {
    osthread->set_suspend_depth(depth);
    return depth;
  }

  // Restore the state that existed before the original suspension
  assert(osthread->state_before_suspend() != SUSPENDED, "wrong thread state");
  osthread->set_state((ThreadState) osthread->state_before_suspend());
  osthread->clear_state_before_suspend();
  osthread->set_suspend_depth(depth);

  if (osthread->is_self_suspended()) {
    osthread->clear_self_suspended();
    Interrupt_lock->notify_all();
  } else {
    assert(osthread->is_physically_suspended(),
      "Must be physically suspended before os level resume");
    if (cb.execute(this) != 0) {
      assert(false, "resume thread failed");
      // Limited recovery in production mode - just reset the thread's
      // state and effectively drop the resume request.
      // XXX - can we do better?
      osthread->set_state(state);
      osthread->set_state_before_suspend(state_before_suspend);
      osthread->set_suspend_depth(depth + 1);
      return -1;
    } 
  }
  if (is_Java_thread()) {
    JavaThread *java_thread = (JavaThread *)this;
    assert(java_thread->is_suspended(), "thread should be suspended");
    if (!java_thread->is_suspended()) return -1;  // robustness
    if (java_thread->is_external_suspend()) {
      // external request can only clear external suspend
      if (ext_req) {
        java_thread->clear_external_suspend();
      }
    } else if (java_thread->is_internal_suspend() && !ext_req) {
      // internal request can only clear internal suspend
      java_thread->clear_internal_suspend();
    }
    java_thread->clear_is_suspended();
  }

  return depth;
}

void Thread::interrupt(Thread* thread) {
  trace("interrupt", thread);
  debug_only(check_for_dangling_thread_pointer(thread);)
  os::interrupt(thread);
}

bool Thread::is_interrupted(Thread* thread, bool clear_interrupted) {
  trace("is_interrupted", thread);  
  debug_only(check_for_dangling_thread_pointer(thread);)
  return os::is_interrupted(thread, clear_interrupted);
}


// Exception handling

void Thread::set_pending_exception(oop exception, const char* file, int line) {
  assert(exception != NULL && exception->is_oop(), "invalid exception oop");
  _pending_exception = exception;
  _exception_file    = file;
  _exception_line    = line;
}


void Thread::clear_pending_exception() {
  if (TraceClearedExceptions) {
    if (_pending_exception != NULL) {
      tty->print_cr("Thread::clear_pending_exception: cleared exception:");
      _pending_exception->print();
    }
  }

  _pending_exception = NULL;
  _exception_file    = NULL;
  _exception_line    = NULL;
}


// GC Support
void Thread::oops_do(void f(oop*)) {
  active_handles()->oops_do(f);
  f((oop*)&_pending_exception);
  handle_area()->oops_do(f);
}

void Thread::print() {
  tty->print("prio=%d tid=0x%x ", get_priority(this), this);
  if (osthread() != NULL) osthread()->print();
  debug_only(if (WizardMode) print_owned_locks();)
}

#ifdef ASSERT
void Thread::print_owned_locks() {  
  Mutex *cur = _owned_locks;
  if (cur == NULL) {
    tty->print(" (no locks) ");
  } else {
    tty->print_cr(" Locks owned:");
    while(cur) {
      cur->print();
      cur = cur->next();
    }
  }
}

static int ref_use_count  = 0;

bool Thread::owns_locks_but_compiled_lock() const {
  for(Mutex *cur = _owned_locks; cur; cur = cur->next()) {
    if (cur != Compile_lock) return true;
  }
  return false;
}


#endif

#ifndef PRODUCT

// The flag: potential_vm_operation notifies if this particular safepoint state could potential
// invoke the vm-thread (i.e., and oop allocation). In that case, we also have to make sure that
// no threads which allow_vm_block's are held
void Thread::check_for_valid_safepoint_state(bool potential_vm_operation) {  
    // Check if current thread is allowed to block at a safepoint
    if (!(_allow_safepoint_count == 0))
      fatal("Possible safepoint reached by thread that does not allow it");
    if (is_Java_thread() && ((JavaThread*)this)->thread_state() != _thread_in_vm) {
      fatal("LEAF method calling lock?");
    }

#ifdef ASSERT    
    if (potential_vm_operation && is_Java_thread()) {
      // Make sure we do not hold any locks that the VM thread also uses.
      // This could potentially lead to deadlocks      
      for(Mutex *cur = _owned_locks; cur; cur = cur->next()) {
        // Threads_lock is special, since the safepoint synchronization will not start before this is
        // acquired. Hence, a JavaThread cannot be holding it at a safepoint. So is VMOperationRequest_lock,
        // since it is used to transfer control between JavaThreads and the VMThread
        // Do not *exclude* any locks unless you are absolutly sure it is correct. Ask someone else first!
        if ( (cur->allow_vm_block() && 
              cur != Threads_lock && 
              cur != Compile_lock &&               // Temporary: should not be necessary when we get spearate compilation
              cur != VMOperationRequest_lock &&               
              cur != VMOperationQueue_lock) ||
              cur->priority() == Mutex::special) { 
          warning("Thread holding lock at safepoint that vm can block on: %s", cur->name());          
        }
      }
    }

    if (GCALotAtAllSafepoints) {
      // We could enter a safepoint here and thus have a gc
      InterfaceSupport::check_gc_alot();
    }

#endif
}
#endif

bool Thread::is_in_stack(address adr) const {
  assert(Thread::current() == this, "is_in_stack can only be called from current thread");
  // High limit: highest_lock is set during thread execution
  // Low  limit: address of the local variable dummy, rounded to 4K boundary.
  // (The rounding helps finding threads in unsafe mode, even if the particular stack
  // frame has been popped already.  Correct as long as stacks are at least 4K long and aligned.)
  address end = os::current_stack_pointer();
  if (_highest_lock >= adr && adr >= end) return true;
  
  return false;
}

// We had to move these methods here, because vm threads get into ObjectSynchronizer::enter
// However, there is a note in JavaThread::is_lock_owned() about the VM threads not being
// used for compilation in the future. If that change is made, the need for these methods
// should be revisited, and they should be removed if possible.

bool Thread::is_lock_owned(address adr) const {
  if (is_in_stack(adr) ) return true;

  return false;
}

void Thread::set_as_starting_thread() {
 // NOTE: this must be called inside the main thread.
  os::create_main_thread(this);
}

static void initialize_class(symbolHandle class_name, TRAPS) {
  KlassHandle h_klass = SystemDictionary::resolve_or_fail(class_name, NULL, NULL, true, CHECK);
  instanceKlass::cast(h_klass())->initialize(CHECK);
}


// Creates the initial ThreadGroup
static oop create_initial_thread_group(TRAPS) {
  instanceKlassHandle klass = SystemDictionary::resolve_or_fail(vmSymbolHandles::java_lang_ThreadGroup(), NULL, NULL, true, CHECK_0);
  
  Handle system_instance = klass->allocate_instance(CHECK_0);
  {
    JavaValue result(T_VOID);
    JavaCalls::call_special(&result, 
			    system_instance, 
			    klass, 
			    vmSymbolHandles::object_initializer_name(), 
			    vmSymbolHandles::void_method_signature(), 
			    CHECK_0);    
  }
  Universe::set_system_thread_group(system_instance());

  Handle main_instance = klass->allocate_instance(CHECK_0);
  {
    JavaValue result(T_VOID);
    Handle string = java_lang_String::create_from_str("main", CHECK_0);
    JavaCalls::call_special(&result,
			    main_instance,
			    klass,
			    vmSymbolHandles::object_initializer_name(),
			    vmSymbolHandles::threadgroup_string_void_signature(),
			    system_instance,
			    string,
			    CHECK_0);
  }

  return main_instance();
}

// Creates the initial Thread
static oop create_initial_thread(Handle thread_group, JavaThread* thread, TRAPS) {  
  instanceKlassHandle klass = SystemDictionary::resolve_or_fail(vmSymbolHandles::java_lang_Thread(), NULL, NULL, true, CHECK_0);
  instanceHandle thread_oop = klass->allocate_instance(CHECK_0);
  
  java_lang_Thread::set_thread(thread_oop(), thread);
  java_lang_Thread::set_priority(thread_oop(), NormPriority);
  thread->set_threadObj(thread_oop());
    
  Handle string = java_lang_String::create_from_str("main", CHECK_0);
  
  JavaValue result(T_VOID);
  JavaCalls::call_special(&result, thread_oop, 
                                   klass, 
                                   vmSymbolHandles::object_initializer_name(), 
                                   vmSymbolHandles::threadgroup_string_void_signature(), 
                                   thread_group, 
                                   string, 
                                   CHECK_0);  
  return thread_oop();
}

static void call_initializeSystemClass(TRAPS) {
  instanceKlassHandle klass =  SystemDictionary::resolve_or_fail(vmSymbolHandles::java_lang_System(), NULL, NULL, true, CHECK);
  
  JavaValue result(T_VOID);
  JavaCalls::call_static(&result, klass, vmSymbolHandles::initializeSystemClass_name(), 
                                         vmSymbolHandles::void_method_signature(), CHECK);
}

static void reset_vm_info_property(TRAPS) {
  // the vm info string
  ResourceMark rm; 
  const char *vm_info = VM_Version::vm_info_string();
  
  // java.lang.System class
  instanceKlassHandle klass =  SystemDictionary::resolve_or_fail(vmSymbolHandles::java_lang_System(), NULL, NULL, true, CHECK);
  
  // setProperty arguments
  Handle key_str    = java_lang_String::create_from_str("java.vm.info", CHECK);
  Handle value_str  = java_lang_String::create_from_str(vm_info, CHECK);
  
  // return value
  JavaValue r(T_OBJECT); 

  // public static String setProperty(String key, String value);
  JavaCalls::call_static(&r,
                         klass,
                         vmSymbolHandles::setProperty_name(), 
                         vmSymbolHandles::string_string_string_signature(), 
                         key_str, 
                         value_str, 
                         CHECK);  
}


void JavaThread::allocate_threadObj(Handle thread_group, char* thread_name, TRAPS) {
  assert(thread_group.not_null(), "thread group should be specified");
  assert(threadObj() == NULL, "should only create Java thread object once");

  instanceKlassHandle klass = SystemDictionary::resolve_or_fail(vmSymbolHandles::java_lang_Thread(), NULL, NULL, true, CHECK);
  instanceHandle thread_oop = klass->allocate_instance(CHECK);

  java_lang_Thread::set_thread(thread_oop(), this);
  java_lang_Thread::set_priority(thread_oop(), NormPriority);
  set_threadObj(thread_oop());

  JavaValue result(T_VOID);
  if (thread_name != NULL) {
    oop str = java_lang_String::create_from_str(thread_name, CHECK);
    Handle name(THREAD, str);
    // Thread gets assigned specified name and null target
    JavaCalls::call_special(&result,
                            thread_oop,
                            klass, 
                            vmSymbolHandles::object_initializer_name(), 
                            vmSymbolHandles::threadgroup_string_void_signature(),
                            thread_group, // Argument 1                        
                            name,         // Argument 2
                            THREAD);
  } else {
    // Thread gets assigned name "Thread-nnn" and null target
    // (java.lang.Thread doesn't have a constructor taking only a ThreadGroup argument)
    JavaCalls::call_special(&result,
                            thread_oop,
                            klass, 
                            vmSymbolHandles::object_initializer_name(), 
                            vmSymbolHandles::threadgroup_runnable_void_signature(),
                            thread_group, // Argument 1                        
                            NULL,         // Argument 2
                            THREAD);
  }
}


// ======= WatcherThread ========

// The watcher thread exists to simulate timer interrupts.  It should
// be replaced by an abstraction over whatever native support for
// timer interrupts exists on the platform.

WatcherThread* WatcherThread::_watcher_thread   = NULL;
bool           WatcherThread::_should_terminate = false;

WatcherThread::WatcherThread() : Thread() {
  assert(watcher_thread() == NULL, "we can only allocate one WatcherThread");
  if (os::create_thread(this, os::watcher_thread)) {
    _watcher_thread = this;
    
    // Set the watcher thread to the highest OS priority which should not be 
    // used, unless a Java thread with priority java.lang.Thread.MAX_PRIORITY 
    // is created. The only normal thread using this priority is the reference 
    // handler thread, which runs for very short intervals only.
    // If the VMThread's priority is not lower than the WatcherThread profiling
    // will be inaccurate.
    os::set_priority(this, MaxPriority);
    if (!DisableStartThread) {
      os::start_thread(this);
    }
  }
}

void WatcherThread::run() {
  assert(this == watcher_thread(), "just checking");

  this->record_stack_base_and_size();
  this->initialize_thread_local_storage();
  this->set_active_handles(JNIHandleBlock::allocate_block());
  while(!_should_terminate) {    
    assert(watcher_thread() == Thread::current(),  "thread consistency check");
    assert(watcher_thread() == this,  "thread consistency check");
    os::sleep(this, WatcherThreadInterval, false);
    PeriodicTask::real_time_tick(WatcherThreadInterval);
  }
  
  // Signal that it is terminated
  {
    MutexLockerEx mu(Terminator_lock, Mutex::_no_safepoint_check_flag);
    _watcher_thread = NULL;
    Terminator_lock->notify();
  }
  
  // Thread destructor usually does this..
  ThreadLocalStorage::set_thread(NULL);
}

void WatcherThread::start() {
  if (watcher_thread() == NULL) {
    _should_terminate = false;
    // Create the single instance of WatcherThread
    new WatcherThread();   
  }
}

void WatcherThread::stop() {
  // it is ok to take late safepoints here, if needed
  MutexLocker mu(Terminator_lock);
  _should_terminate = true;  
  while(watcher_thread() != NULL) {
    Terminator_lock->wait();
  }  
}

void WatcherThread::print() {
  tty->print("\"VM Periodic Task Thread\" ");
  Thread::print();
  tty->cr();
}


// ======= SuspendCheckerThread ========

// The SuspendCheckerThread is used by the VMThread to see if
// the current suspend operation will cause a VMThread deadlock.

SuspendCheckerThread* SuspendCheckerThread::_suspend_checker_thread = NULL;
bool           SuspendCheckerThread::_should_terminate              = false;

SuspendCheckerThread::SuspendCheckerThread() : Thread() {
  assert(suspend_checker_thread() == NULL, "we can only allocate one SuspendCheckerThread");
  if (os::create_thread(this, os::suspend_checker_thread)) {
    // Set the suspend checker thread to the highest OS priority. Since
    // this thread is doing work for the VMThread, it should have a
    // higher priority.
    os::set_priority(this, MaxPriority);
    if (!DisableStartThread) {
      os::start_thread(this);
    }
  }
}

void SuspendCheckerThread::run() {
  this->record_stack_base_and_size();
  this->initialize_thread_local_storage();
  this->set_active_handles(JNIHandleBlock::allocate_block());
  {
    MutexLockerEx ml(SuspendChecker_lock, Mutex::_no_safepoint_check_flag);
    // once we own the lock we can advertise we are ready
    _suspend_checker_thread = this;

    while(!_should_terminate) {    
      assert(suspend_checker_thread() == Thread::current(),  "thread consistency check");
      assert(suspend_checker_thread() == this,  "thread consistency check");

      SuspendChecker_lock->wait(true, 0);  // wait for VMThread to need us
      { // temporarily drop the lock in case we get stuck
	MutexUnlocker mul(SuspendChecker_lock);

        // The malloc lock is only one operating system resource that causes
        // deadlock when held by a suspended thread.  Should there be a more
        // general model of operating system resources to fence off?

        char *b = (char *)malloc(1);         // see if we can allocate a byte
        if (b != NULL) free(b);              // give it back
      }
      SuspendChecker_lock->notify();       // notify VMThread we are not stuck
    }
  }
  
  // Signal that it is terminated
  {
    MutexLockerEx mu(Terminator_lock, Mutex::_no_safepoint_check_flag);
    _suspend_checker_thread = NULL;
    Terminator_lock->notify();
  }
  
  // Thread destructor usually does this..
  ThreadLocalStorage::set_thread(NULL);
}

void SuspendCheckerThread::start() {
  if (suspend_checker_thread() == NULL) {
    _should_terminate = false;
    // Create the single instance of SuspendCheckerThread
    new SuspendCheckerThread();   
  }
}

void SuspendCheckerThread::stop() {
  // Since the VMThread depends on us for deadlock detection, we
  // shouldn't block for safepoints.
  MutexLockerEx mu(Terminator_lock, Mutex::_no_safepoint_check_flag);
  _should_terminate = true;
  {
    // wake up SuspendCheckerThread so it can see that it needs to exit
    MutexLockerEx ml(SuspendChecker_lock, Mutex::_no_safepoint_check_flag);
    SuspendChecker_lock->notify();
  }
  while (suspend_checker_thread() != NULL) {
    Terminator_lock->wait(true);  // don't block for safepoints
  }  
}

void SuspendCheckerThread::print() {
  tty->print("\"Suspend Checker Thread\" ");
  Thread::print();
  tty->cr();
}


// ======= JavaThread ========

// A JavaThread is a normal Java thread

const struct JNINativeInterface_* jni_functions();

void JavaThread::initialize() {
  // Initialize fields
  set_threadObj(NULL);  
  set_cached_state(NULL);
  set_last_Java_sp(NULL);
  set_entry_point(NULL);
  _jni_environment.functions = jni_functions();
  set_vm_result(NULL);
  set_vm_result_2(NULL);
  set_vframe_array_head(NULL);
  set_monitor_chunks(NULL);
  set_monitor_arrays(NULL);  
  set_next(NULL);  
  set_thread_state(_thread_new);  
  set_is_terminated(false);  
  _privileged_stack_top = NULL;
  _array_for_gc = NULL;
  _external_suspend = false;
  _internal_suspend = false;
  _is_suspended = false;
  _in_jvm_wait = false;
  _exception_oop = NULL;
  _exception_pc  = 0;
  _exception_handler_pc = 0;
  _exception_stack_size = 0;
  _last_exception_pc = NULL;
  _last_native_pc    = NULL;
  _jvmdi_thread_state= NULL;
  _jvmdi_get_loaded_classes_closure = NULL;
  _not_at_call_sp    = NULL;
  _special_runtime_exit_condition = false;
  _is_handling_implicit_exception = false;
  _pending_async_exception = NULL;
  _is_compiling = false;
#ifdef COMPILER2
  _blocked_on_compilation = false;
#endif
  _jvmpi_data = NULL;
  _last_sum = 0;

  if(UseTLE) {
    tle().clear();  // initialize thread-local eden related fields
  }

  set_thread_profiler(NULL);
  if (FlatProfiler::is_active()) {
    // This is where we would decide to either give each thread it's own profiler
    // or use one global one from FlatProfiler,
    // or up to some count of the number of profiled threads, etc.
    ThreadProfiler* pp = new ThreadProfiler();
    pp->engage();
    set_thread_profiler(pp);
  }

#ifndef PRODUCT
  _method_tracer = NULL;
#endif

  // Setup safepoint state info for this thread
  ThreadSafepointState::create(this);  

  debug_only(_java_call_counter = 0);

  // Give JVMDI a chance to setup any per thread data structures it
  // might need.
  if (jvmdi::enabled()) {
    jvmdi::initialize_thread(this);
  }

  pd_initialize();
}

JavaThread::JavaThread() : Thread() {
  initialize();      
}

// Remove this ifdef when C1 is ported to the compiler interface.
#ifndef CORE
static void compiler_thread_entry(JavaThread* thread, TRAPS);
#endif

JavaThread::JavaThread(ThreadFunction entry_point) : Thread() {
  if (TraceThreadEvents) {
    tty->print_cr("creating thread %#lx", this);
  }
  initialize();
  set_entry_point(entry_point);
  // Create the native thread itself.
  // %note runtime_23
  // Remove this ifdef when C1 is ported to the compiler interface.
#ifndef CORE
  os::create_thread(this, ((entry_point == &compiler_thread_entry) ? os::compiler_thread : os::java_thread));
#else
  os::create_thread(this, os::java_thread);
#endif CORE

  // The _osthread may be NULL here because we ran out of memory (too many threads active).
  // We need to throw and OutOfMemoryError - however we cannot do this here because the caller
  // may hold a lock and all locks must be unlocked before throwing the exception (throwing
  // the exception consists of creating the exception object & initializing it, initialization
  // will leave the VM via a JavaCall and then all locks must be unlocked).
  //
  // The thread is still suspended when we reach here. Thread must be explicit started
  // by creator! Furthermore, the thread must also explicitly be added to the Threads list
  // by calling Threads:add. The reason why this is not done here, is because the thread 
  // object must be fully initialized (take a look at JVM_Start)  
}


JavaThread::~JavaThread() {
  if (TraceThreadEvents) {
      tty->print_cr("terminate thread %#lx", this);
  }
  
  // All Java related clean up happens in exit  
  ThreadSafepointState::destroy(this);
  if (_thread_profiler != NULL) delete _thread_profiler;

  if (jvmdi::enabled()) {
    jvmdi::cleanup_thread(this);
  }

#ifndef PRODUCT
  if (TraceMethods && (_method_tracer != NULL)) {
    Unimplemented(); 
    // gives compiler warning
    // delete _method_tracer;
  }
#endif
}


// The first routine called by a new Java thread
void JavaThread::run() {
  // used to test validitity of stack trace backs
  this->record_base_of_stack_pointer();
  
  // %%%% Record real stack base and size. We need to clean up all
  // the various stack usage items soon.
  this->record_stack_base_and_size();

  // Initialize thread local storage; set before calling MutexLocker
  this->initialize_thread_local_storage();  

  if(UseStackBanging) {
    this->enable_stack_yellow_zone();
    this->enable_stack_red_zone();
  }
  
  // Thread is now sufficient initialized to be handled by the safepoint code as being
  // in the VM. Change thread state from _thread_new to _thread_in_vm  
  ThreadStateTransition::transition(this, _thread_new, _thread_in_vm);  
  
  assert(JavaThread::current() == this, "sanity check");
  assert(!Thread::current()->owns_locks(), "sanity check");
  
  // This operation might block. We call that after all safepoint checks for a new thread has
  // been completed.
  this->set_active_handles(JNIHandleBlock::allocate_block());
  
  if (jvmdi::enabled()) {
    jvmdi::post_thread_start_event(this);
  }
  if (jvmpi::is_event_enabled(JVMPI_EVENT_THREAD_START)) {
    jvmpi::post_thread_start_event(this);
  }

  // We call another function to do the rest so we are sure that the stack addresses used
  // from there will be lower than the stack base just computed
  thread_main_inner();

  // Note, thread is no longer valid at this point!
}


void JavaThread::thread_main_inner() {  
  assert(JavaThread::current() == this, "sanity check");
  assert(this->threadObj() != NULL, "just checking");

  // Execute thread entry point. If this thread is being asked to restart, 
  // or has been stopped before starting, do not reexecute entry point.
  // Note: Due to JVM_StopThread we can have pending exceptions already!
  if (!this->has_pending_exception() && !java_lang_Thread::is_stillborn(this->threadObj())) {
    // enter the thread's entry point only if we have no pending exceptions
    HandleMark hm(this);    
    this->entry_point()(this, this);
  }
  
  this->exit(false);
  this->delete_thread_and_TLS_current();

}

  
static void ensure_join(JavaThread* thread) {
  // We do not need to grap the Threads_lock, since we are operating on ourself.
  Handle threadObj(thread, thread->threadObj());
  assert(threadObj.not_null(), "java thread object must exist");
  ObjectLocker lock(threadObj, thread);
  // Ignore pending exception (ThreadDeath), since we are exiting anyway
  thread->clear_pending_exception();
  // It is of profound importance that we set the stillborn bit and reset the thread object, 
  // before we do the notify. Since, changing these two variable will make JVM_IsAlive return
  // false. So in case another thread is doing a join on this thread , it will detect that the thread 
  // is dead when it gets notified.
  java_lang_Thread::set_stillborn(threadObj());
  java_lang_Thread::set_thread(threadObj(), NULL);
  lock.notify_all(thread);
  // Ignore pending exception (ThreadDeath), since we are exiting anyway
  thread->clear_pending_exception();
}


void JavaThread::exit(bool destroy_vm) {
  assert(this == JavaThread::current(),  "thread consistency check");
  if (!InitializeJavaLangSystem) return;

  HandleMark hm(this);
  Handle uncaught_exception(this, this->pending_exception());
  this->clear_pending_exception();
  Handle threadObj(this, this->threadObj());
  assert(threadObj.not_null(), "Java thread object should be created");
  
  if (get_thread_profiler() != NULL) {
    get_thread_profiler()->disengage();
    ResourceMark rm;
    get_thread_profiler()->print(get_thread_name());
  }


  // FIXIT: This code should be moved into else part, when reliable 1.2/1.3 check is in place
  instanceKlassHandle shutdown_klass;
  {
    EXCEPTION_MARK;
    
    CLEAR_PENDING_EXCEPTION;
  }
  // FIXIT: The is_null check is only so it works better on JDK1.2 VM's. This
  // has to be fixed by a runtime query method
  if (!destroy_vm || Universe::is_jdk12x_version()) {
    // Handle uncaught_exception if any, java.lang.ThreadGroup.uncaughtException must be called 
    if (uncaught_exception.not_null()) {
      Handle group(this, java_lang_Thread::threadGroup(threadObj()));
      Events::log("uncaught exception %#x %#x %#x", uncaught_exception(), threadObj(), group());
      { 
        EXCEPTION_MARK;
        JavaValue result(T_VOID);
        JavaCalls::call_virtual(&result, 
                                group, 
                                SystemDictionary::threadGroup_klass(), 
                                vmSymbolHandles::uncaughtException_name(), 
                                vmSymbolHandles::thread_throwable_void_signature(), 
                                threadObj,          // Arg 1
                                uncaught_exception, // Arg 2
                                THREAD);
        CLEAR_PENDING_EXCEPTION;
      }
    }
 
    // Call Thread.exit(). We try 3 times in case we got another Thread.stop during
    // the execution of the method. If that is not enough, then we don't really care. Thread.stop
    // is deprecated anyhow.
    { int count = 3;
      while (java_lang_Thread::threadGroup(threadObj()) != NULL && (count-- > 0)) {
        EXCEPTION_MARK;
        JavaValue result(T_VOID);
        JavaCalls::call_virtual(&result, 
                              threadObj, 
                              SystemDictionary::thread_klass(),
                              vmSymbolHandles::exit_method_name(), 
                              vmSymbolHandles::void_method_signature(), 
                              THREAD);  
        CLEAR_PENDING_EXCEPTION;
      }
    }      
  } else {
    // Last thread running. It calls java.lang.Shutdown.shutdown()
    EXCEPTION_MARK;
    instanceKlassHandle shutdown_klass = SystemDictionary::resolve_or_null(vmSymbolHandles::java_lang_Shutdown(), 
                                                                           NULL, NULL, CHECK);    
    assert(shutdown_klass.not_null(), "sanity check");
    JavaValue result(T_VOID);
    JavaCalls::call_static(&result,                      
                           shutdown_klass,
                           vmSymbolHandles::shutdown_method_name(),
                           vmSymbolHandles::void_method_signature(),
                           THREAD);  
    CLEAR_PENDING_EXCEPTION;
  }
  
  // notify jvmdi
  if (jvmdi::enabled()) {
    jvmdi::post_thread_end_event(this);
  }

  // notify jvmpi
  if (jvmpi::is_event_enabled(JVMPI_EVENT_THREAD_END)) {
    jvmpi::post_thread_end_event(this);
  }

  // Notify waiters on thread object. This has to be done after exit() is called
  // on the thread (if the thread is the last thread in a daemon ThreadGroup the
  // group should have the destroyed bit set before waiters are notified).
  ensure_join(this);

  // These things needs to be done while we are still a Java Thread. Make sure that thread
  // is in a consistent state, in case GC happens
  assert(_privileged_stack_top == NULL, "must be NULL when we get here");
    
  if (active_handles() != NULL) {
    JNIHandleBlock* block = active_handles();
    set_active_handles(NULL);
    JNIHandleBlock::release_block(block);
  }

  if (free_handle_block() != NULL) {
    JNIHandleBlock* block = free_handle_block();
    set_free_handle_block(NULL);
    JNIHandleBlock::release_block(block);
  }

  // These have to be removed while this is still a valid thread.
  if(UseStackBanging) {
    disable_stack_yellow_zone();
    disable_stack_red_zone();
  }

  if(UseTLE) {
    if(ParkTLE) {
      tle().clone()->park();   // park a _copy_ of the tle descripotor
    }
    else {
      tle().reset();  
    }
  }

  // Remove from list of active threads list, and notify VM thread if we are the last non-daemon thread 
  Threads::remove(this);  
}


JavaThread* JavaThread::active() {
  Thread* thread = ThreadLocalStorage::thread();
  assert(thread != NULL, "just checking");
  if (thread->is_Java_thread()) {
    return (JavaThread*) thread;
  } else {
    assert(thread->is_VM_thread(), "this must be a vm thread");
    VM_Operation* op = ((VMThread*) thread)->vm_operation();
    JavaThread *ret=op == NULL ? NULL : (JavaThread *)op->calling_thread();
    assert(ret->is_Java_thread(), "must be a Java thread");
    return ret;
  }
}

bool JavaThread::is_lock_owned(address adr) const {  
  if (is_in_stack(adr)) return true;
    
  for (MonitorChunk* chunk = monitor_chunks(); chunk != NULL; chunk = chunk->next()) {
    if (chunk->contains(adr)) return true;
  }

  return false;
}


void JavaThread::add_monitor_chunk(MonitorChunk* chunk) {
  chunk->set_next(monitor_chunks());
  set_monitor_chunks(chunk);
}

void JavaThread::remove_monitor_chunk(MonitorChunk* chunk) {
  guarantee(monitor_chunks() != NULL, "must be non empty");
  if (monitor_chunks() == chunk) {
    set_monitor_chunks(chunk->next());
  } else {
    MonitorChunk* prev = monitor_chunks();
    while (prev->next() != chunk) prev = prev->next();
    prev->set_next(chunk->next()); 
  }
}


void JavaThread::add_monitor_array(MonitorArray* array) {
#ifndef CORE
  array->set_next(monitor_arrays());
  set_monitor_arrays(array);
#endif
}

void JavaThread::remove_monitor_array(MonitorArray* array) {
#ifndef CORE
  guarantee(monitor_arrays() != NULL, "link must be present");
  if (monitor_arrays() == array) {
    set_monitor_arrays(array->next());
  } else {
    MonitorArray* prev = monitor_arrays();
    while (prev->next() != array) prev = prev->next();
    prev->set_next(array->next()); 
  }
#endif
}

// JVM support.

void JavaThread::handle_special_runtime_exit_condition() {
  // Assertion removed, clear_special done in external suspend path also
  // which allows us to get here after the special handling is no longer
  // needed.
  if (!_special_runtime_exit_condition) return;

  clear_special_runtime_exit_condition();

  // Check for pending async. exception
  if (_pending_async_exception != NULL) {
    // Only overwrite an already pending exception, if it is not a threadDeath.
    if (!has_pending_exception() || !pending_exception()->is_a(SystemDictionary::threaddeath_klass())) {
      // We cannot call Exceptions::_throw(...) here because we cannot block      
      set_pending_exception(_pending_async_exception, __FILE__, __LINE__);
      
      if (TraceExceptions) {
        ResourceMark rm;
      tty->print_cr("Async. exception installed at runtime exit of type: %s", instanceKlass::cast(_pending_async_exception->klass())->external_name());
      }
      _pending_async_exception = NULL;
    }        
  }

  if (!has_pending_exception()) {
    // Check for pending external suspend. Only external self-suspension
    // uses handle_special_runtime_exit_condition().
    // Only suspend if this thread has not already suspended
    // jdbx threads call the vm with another thread's jnienv
    // so we think we are operating on a suspended thread's behalf
    // see 4432884
    bool pending_suspend;
    {
      MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
      pending_suspend = !is_suspended() && is_external_suspend();
    }
    if (pending_suspend) {
      StateSaver sv(this);  // prepare for possible stack walks, etc
      java_suspend_self();
    }   
  }
}

void JavaThread::send_thread_stop(oop java_throwable)  {
  assert(Thread::current()->is_VM_thread(), "should be in the vm thread");
  assert(Threads_lock->is_locked(), "Threads_lock should be locked by safepoint code");
  assert(SafepointSynchronize::is_at_safepoint(), "all threads are stopped");  

  // This is a change from JDK 1.1, but JDK 1.2 will also do it:
  if (java_throwable->is_a(SystemDictionary::threaddeath_klass())) {
    java_lang_Thread::set_stillborn(threadObj());
  }

  if (thread_state() == _thread_in_native) {
    // If the thread is in native code, it just gets the exception
    if (!has_pending_exception() || !pending_exception()->is_a(SystemDictionary::threaddeath_klass())) {
      set_pending_exception(java_throwable, __FILE__, __LINE__);

      if (TraceExceptions) {
       ResourceMark rm;    
       tty->print_cr("Pending exception installed of type: %s", instanceKlass::cast(pending_exception()->klass())->external_name());
      }
    }
  } else {
    // Actually throw the Throwable against the target Thread - however
    // only if there is no thread death exception installed already.
    if (_pending_async_exception == NULL || !_pending_async_exception->is_a(SystemDictionary::threaddeath_klass())) {
#ifdef COMPILER2
      // If the topmost frame is a runtime stub, then we are calling into
      // OptoRuntime from compiled code. Some runtime stubs (new, monitor_exit..)
      // must deoptimize the caller before continuing, as the compiled  exception handler table
      // may not be valid
      if (has_last_Java_frame()) {
        frame f = last_frame();
        if (f.is_c2runtime_frame() || f.is_safepoint_blob_frame()) {          
          RegisterMap reg_map(this, false);
          frame compiled_frame = f.sender(&reg_map);
          if (compiled_frame.is_compiled_frame()) {
            Deoptimization::deoptimize_frame(this, compiled_frame.sp());
          }          
        }
      }
#endif
    
      // Set async. pending exception in thread.
      set_pending_async_exception(java_throwable);    

      if (TraceExceptions) {
       ResourceMark rm;    
       tty->print_cr("Pending Async. exception installed of type: %s", instanceKlass::cast(_pending_async_exception->klass())->external_name());
      }
    }
  }

  
  // Interrupt thread so it will wake up from a potential wait()
  Thread::interrupt(this);  
}

// External suspension mechanism.
// Tell the VM to suspend a thread when ever it knows that it does not hold on to
// any VM_locks and it is at a transition
// Self-suspension will happen on the transition out of the vm.
void JavaThread::java_suspend(bool native_now) {
  VM_ThreadSuspend* vm_suspend = new VM_ThreadSuspend(this->threadObj(), native_now);
  VMThread::execute(vm_suspend);
}

// Do NOT allow external suspension of the SuspendCheckerThread
void JavaThread::send_java_suspend(bool native_now) {
#ifdef ASSERT
  if (Thread::current() == this) {
    bool is_suspended;
    {
      MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
      is_suspended = this->is_suspended();
    }
    assert(!is_suspended, "a suspended thread should not be able to suspend itself");
  }
#endif /* ASSERT */
  if (this->is_SuspendChecker_thread()) {
    return;
  }

  if (native_now && (thread_state() == _thread_in_native)) {
    suspend_other(true);           // immediate external suspend request
  } else {
    set_external_suspend();        // let thread suspend itself
  }
}

// Part II of external suspension.
// A JavaThread that is not in state _thread_in_native calls
// java_suspend_self() on transition. Assumes that is_external_suspend()
// is still true, but if it isn't due to a racing resume, then that will
// be caught in suspend_thread_impl().
long JavaThread::java_suspend_self() {
  // we have been terminated so don't suspend
  if (is_terminated()) return (long) 0;

  assert(state_is_saved_before_block() ||
    (is_Java_thread() && !((JavaThread*)this)->has_last_Java_frame()),
    "must have called save_state_before_suspension");

#ifdef ASSERT
  bool is_suspended;
  {
    MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
    is_suspended = this->is_suspended();
  }
  assert(!is_suspended,
    "a thread trying to self-suspend should not already be suspended");
#endif /* ASSERT */

  SuspendThreadSelf_Callback cb;
//  NOT_CORE(assert(!is_Compiler_thread(),
//           "A compiler thread self-suspension?"));
  return suspend_thread_impl(cb, true);  // external suspend request
}


// We need to guarantee the Threads_lock here, since resumes are not 
// allowed during safepoint synchronization
// Can only resume from an external suspension
void JavaThread::java_resume() {
  bool found = false;

  assert_lock(Threads_lock);  
  for(JavaThread *thread = Threads::first(); thread && (found == false); thread = thread->next()) {
    if (this == thread) {
      found = true;
    }
  }
  
  // Sanity check: thread is gone, marked for termination or the thread
  // was not externally suspended.
  if (!found || is_terminated() || !is_external_suspend()) return;

  ((Thread *)this)->resume(true, true);  // yield, external resume request

  // It would be nice to have the following assertion in all the time,
  // but it is possible for a racing suspend request to have suspended
  // this thread right after we resumed it. Temporarily enable this
  // assertion if you are chasing a different kind of bug.
  //
  // assert(java_lang_Thread::thread(threadObj()) == NULL ||
  //   !is_being_ext_suspended(), "thread is suspended after resume operation");
}

void JavaThread::enable_stack_yellow_zone() {
  assert(UseStackBanging,"Inserting stack guard pages when UseStackBanging is false");

  // The base notation is from the stacks point of view, growing downward.
  // We need to adjust it to work correctly with guard_memory()
  unsigned int base = (unsigned int)stack_yellow_zone_base() - stack_yellow_zone_size();
  
  guarantee(base < (unsigned int)stack_base(),"Error calculating stack yellow zone");
  guarantee(base < (unsigned int)os::current_stack_pointer(),"Error calculating stack yellow zone");

  if(!os::guard_memory((char *)base,stack_yellow_zone_size())) {
    warning("Attempt to guard stack yellow zone failed.");
  }
}

void JavaThread::disable_stack_yellow_zone() {
  assert(UseStackBanging,"Removing stack guard pages when UseStackBanging is false");

  // The base notation is from the stacks point of view, growing downward.
  // We need to adjust it to work correctly with guard_memory()
  int base = (int)stack_yellow_zone_base() - stack_yellow_zone_size();
  
  if(!os::unguard_memory((char *)base,stack_yellow_zone_size())) {
    warning("Attempt to unguard stack yellow zone failed.");
  }
}

void JavaThread::enable_stack_red_zone() {
  assert(UseStackBanging,"Inserting stack guard pages when UseStackBanging is false");

  // The base notation is from the stacks point of view, growing downward.
  // We need to adjust it to work correctly with guard_memory()
  unsigned int base = (int)stack_red_zone_base() - stack_red_zone_size();
  
  guarantee(base < (unsigned int)stack_base(),"Error calculating stack red zone");
  guarantee(base < (unsigned int)os::current_stack_pointer(),"Error calculating stack red zone");

  if(!os::guard_memory((char *)base,stack_red_zone_size())) {
    warning("Attempt to guard stack red zone failed.");
  }
}

void JavaThread::disable_stack_red_zone() {
  assert(UseStackBanging,"Removing stack guard pages when UseStackBanging is false");

  // The base notation is from the stacks point of view, growing downward.
  // We need to adjust it to work correctly with guard_memory()
  int base = (int)stack_red_zone_base() - stack_red_zone_size();
  
  if(!os::unguard_memory((char *)base,stack_red_zone_size())) {
    warning("Attempt to unguard stack red zone failed.");
  }
}

void JavaThread::insert_vframe_array(vframeArray* value) {
#ifdef CORE
  ShouldNotReachHere();
#else
  vframeArray* temp = vframe_array_head();
  if (   (vframe_array_head() == NULL)
      || (vframe_array_head()->sp() > value->sp())) {
    set_vframe_array_head(value);
    value->set_next(temp);
  } else {
    // Find the right place to insert the vframe array
    while (   (temp->next() != NULL)
           && (temp->next()->sp() < value->sp())) {
      temp = temp->next();
    }
    value->set_next(temp->next());    
    temp->set_next(value);
  }
#endif
}

vframeArray* JavaThread::vframe_array_for(jint* sp) {
#ifdef CORE
  ShouldNotReachHere();
#else
  for (vframeArray* temp = vframe_array_head(); temp != NULL; temp = temp->next()) {
    if (temp->sp() == sp) return temp;
  }
#endif
  return NULL;
}

void JavaThread::frames_do(void f(frame*, const RegisterMap* map)) {
  // ignore is there is no stack
  if (!has_last_Java_frame()) return;
  // traverse the stack frames. Starts from top frame.  
  for(StackFrameStream fst(this); !fst.is_done(); fst.next()) {
    frame* fr = fst.current();    
    f(fr, fst.register_map());    
  }  
}


// Deoptimization
// Function for testing deoptimization
void JavaThread::deoptimize() {
#ifndef CORE
  bool deopt = false;		// Dump stack only if a deopt actually happens.
  // Iterate over all frames in the thread and deoptimize
  for(StackFrameStream fst(this); !fst.is_done(); fst.next()) {
    if( fst.current()->is_deoptimized_frame() ) break;
    if (fst.current()->can_be_deoptimized()) {
      if (DebugDeoptimization && !deopt) {
	deopt = true;		// One-time only print before deopt
	tty->print_cr("[BEFORE Deoptimization]");
	trace_frames();
	trace_stack();
      }
      // Make a private register-map for deoptimize to trash
      RegisterMap reg_map(fst.register_map());
      Deoptimization::deoptimize(this, *fst.current(), &reg_map);
    }
  }

  if (DebugDeoptimization && deopt) {
    tty->print_cr("[AFTER Deoptimization]");
    trace_frames();
  }
#endif
}


#ifndef PRODUCT
// Make zombies
void JavaThread::make_zombies() {
#ifndef CORE
  for(StackFrameStream fst(this); !fst.is_done(); fst.next()) {
    if (fst.current()->can_be_deoptimized()) {
      // it is a Java nmethod
      nmethod* nm = CodeCache::find_nmethod(fst.current()->pc());
      nm->make_not_entrant();
    }
  }
#endif
}
#endif


bool JavaThread::deoptimized_one_wrt_marked_nmethods() {
#ifndef CORE
  for(StackFrameStream fst(this); !fst.is_done(); fst.next()) {
    if (fst.current()->should_be_deoptimized()) {
      Deoptimization::deoptimize(this, *fst.current(), fst.register_map());
      return true;
    }
  }
#endif
  return false;
}


void JavaThread::deoptimized_wrt_marked_nmethods() {
#ifndef CORE
  if (!has_last_Java_frame()) return;
  // We only deoptimize one frame in each traversal to avoid 
  // manipulating a frame and the continue traversal.
  int cont = true;
  while (cont) {
    cont = deoptimized_one_wrt_marked_nmethods();
  }
#endif
}


// GC support
static void frame_gc_epilogue(frame* f, const RegisterMap* map) { f->gc_epilogue(); }

void JavaThread::gc_epilogue() {
  frames_do(frame_gc_epilogue);
}


static void frame_gc_prologue(frame* f, const RegisterMap* map) { f->gc_prologue(); }

void JavaThread::gc_prologue() {
  frames_do(frame_gc_prologue);
}


void JavaThread::oops_do(void f(oop*)) {  
  // The ThreadProfiler oops_do is done from FlatProfiler::oops_do
  // since there may be more than one thread using each ThreadProfiler.

  // Traverse the GCHandles
  Thread::oops_do(f);

  assert( (!has_last_Java_frame() && java_call_counter() == 0) ||
          (has_last_Java_frame() && java_call_counter() > 0), "wrong java_sp info!");

  if (has_last_Java_frame()) {

    // Traverse the privileged stack
    if (_privileged_stack_top != NULL) {
      _privileged_stack_top->oops_do(f);
    }

    // traverse the registered growable array
    if (_array_for_gc != NULL) {
      for (int index = 0; index < _array_for_gc->length(); index++) {
        f(_array_for_gc->adr_at(index));
      }
    }

    // Traverse the monitor chunks
    for (MonitorChunk* chunk = monitor_chunks(); chunk != NULL; chunk = chunk->next()) {
      chunk->oops_do(f);
    }

#ifndef CORE
    // Traverse the monitor arrays
    for (MonitorArray* array = monitor_arrays(); array != NULL; array = array->next()) {
      array->oops_do(f);
    }
#endif

    // Traverse the execution stack    
    for(StackFrameStream fst(this); !fst.is_done(); fst.next()) {
      fst.current()->oops_do(f, fst.register_map());
    }
  } 

#ifndef CORE
  // Traverse the vframeArray
  vframeArray* current = vframe_array_head();
  while (current) {
    current->oops_do(f);
    current = current->next();
  }
#endif

  // Traverse instance variables at the end since the GC may be moving things
  // around using this function
  f((oop*) &_threadObj);
  f((oop*) &_vm_result);
  f((oop*) &_vm_result_2);
  f((oop*) &_exception_oop);
  f((oop*) &_pending_async_exception);
}

// Printing

void JavaThread::print() {
  tty->print("\"%s\" ", get_thread_name());
  if (threadObj() != NULL && java_lang_Thread::is_daemon(threadObj()))  tty->print("daemon ");
  Thread::print();
  // print guess for valid stack memory region (assume 4K pages); helps lock debugging
  tty->print_cr("[%#x..%#x]", int(last_Java_sp()) & ~right_n_bits(12), highest_lock());
}

// Verification

#ifndef PRODUCT
static void frame_verify(frame* f, const RegisterMap *map) { f->verify(map); }

static void verify_thread_oop(oop* p) {
  guarantee((*p)->safepoint_is_oop_or_null(), "Invalid oop in thread");
}


void JavaThread::verify() {
  // Verify oops in the thread.
  oops_do(&verify_thread_oop);

  // Verify the stack frames.
  frames_do(frame_verify);
}

#ifndef CORE
void JavaThread::verify_vframes() {
 // Verify the vframeArray list in the thread matches the frames.
  vframeArray* current = vframe_array_head();

  if (has_last_Java_frame()) {
    for(StackFrameStream fst(this); !fst.is_done(); fst.next()) {
      frame* fr = fst.current();
      if (fr->is_deoptimized_frame()) {
        guarantee(current != NULL, "must have vframeArray for deoptimized frame");
        guarantee(fr->sp() == current->sp(), "must find matching sp in vframeArray");
        current = current->next();
      }
    }
  }
  guarantee(current == NULL, "should not have any vframeArrays left");
}
#endif // CORE
#endif // PRODUCT

const char* JavaThread::get_thread_name() const {
  // Early safepoints hit while JavaThread::current does not yet have TLS
  debug_only(if (!(SafepointSynchronize::is_at_safepoint() || JavaThread::current() == this)) assert_lock(Threads_lock);)
  oop thread_obj = threadObj();
  if (thread_obj != NULL) {
    typeArrayOop name = java_lang_Thread::name(thread_obj);
    const char* str = UNICODE::as_utf8((jchar*) name->base(T_CHAR), name->length());
    return str;
  } else {
    return NULL;
  }
  return NULL;
}

const char* JavaThread::get_threadgroup_name() const {
  debug_only(if (JavaThread::current() != this) assert_lock(Threads_lock);)
  oop thread_obj = threadObj();
  if (thread_obj != NULL) {
    oop thread_group = java_lang_Thread::threadGroup(thread_obj);
    if (thread_group != NULL) {
      typeArrayOop name = java_lang_ThreadGroup::name(thread_group);
      const char* str = UNICODE::as_utf8((jchar*) name->base(T_CHAR), name->length());
      return str;
    }
  }
  return NULL;
}

const char* JavaThread::get_parent_name() const {
  debug_only(if (JavaThread::current() != this) assert_lock(Threads_lock);)
  oop thread_obj = threadObj();
  if (thread_obj != NULL) {
    oop thread_group = java_lang_Thread::threadGroup(thread_obj);
    if (thread_group != NULL) {
      oop parent = java_lang_ThreadGroup::parent(thread_group);
      if (parent != NULL) {
	typeArrayOop name = java_lang_ThreadGroup::name(parent);
	const char* str = UNICODE::as_utf8((jchar*) name->base(T_CHAR), name->length());
	return str;
      }
    }
  }
  return NULL;
}

ThreadPriority JavaThread::java_priority() const {
  oop thr_oop = threadObj();
  if (thr_oop == NULL) return NormPriority; // Bootstrapping
  ThreadPriority priority = java_lang_Thread::priority(thr_oop);
  assert(MinPriority <= priority && priority <= MaxPriority, "sanity check");
  return priority;
}

void JavaThread::prepare(jthread jni_thread, ThreadPriority prio) {

  assert(Threads_lock->owner() == Thread::current(), "must have threads lock");
  // Link Java Thread object <-> C++ Thread

  // Get the C++ thread object (an oop) from the JNI handle (a jthread)
  // and put it into a new Handle.  The Handle "thread_oop" can then
  // be used to pass the C++ thread object to other methods.

  // Set the Java level thread object (jthread) field of the
  // new thread (a JavaThread *) to C++ thread object using the
  // "thread_oop" handle.
  
  // Set the thread field (a JavaThread *) of the
  // oop representing the java_lang_Thread to the new thread (a JavaThread *).

  Handle thread_oop(JNIHandles::resolve_non_null(jni_thread));
  assert(instanceKlass::cast(thread_oop->klass())->is_linked(), 
    "must be initialized");
  set_threadObj(thread_oop());
  java_lang_Thread::set_thread(thread_oop(), this);

  if (prio == NoPriority) {
    prio = java_lang_Thread::priority(thread_oop());
    assert(prio != NoPriority, "A valid priority should be present");
  }

  // Push the Java priority down to the native thread; needs Threads_lock
  Thread::set_priority(this, prio);

  // Add the new thread to the Threads list and set it in motion. 
  // We must have threads lock in order to call Threads::add. 
  // It is crucial that we do not block before the thread is 
  // added to the Threads list for if a GC happens, then the java_thread oop
  // will not be visited by GC.
  Threads::add(this);
}

frame JavaThread::cook_last_frame(frame fr) {
  
  // If thread is in native, we use the last_native_pc stored in the thread, since the return address
  // might not be on top of stack. 
  // Sometimes (ThreadToNativeFromVM( _last_native_pc is not set; therefore
  // check what to use    
  ThreadState state      = (ThreadState)_thread_state;
  address last_native_pc = (address)_last_native_pc;
  address frame_pc       = fr.pc();
  
  address pc = NULL;
  if ((state == _thread_in_native          || 
       state == _thread_in_native_trans    || 
       state == _thread_in_native_blocked) && last_native_pc != NULL) {
    pc = last_native_pc;
  } else {
    pc = frame_pc;
  }  
  assert(pc != NULL, "cannot be null");

#ifndef CORE
  // Check if frame pointer has been patched and/or pointing into temp. buffer.
  address adjusted_pc = safepoint_state()->compute_adjusted_pc(pc);
  // This should really be an assert - but to easy debugging it is currently a guarantee. 
  assert(AbstractInterpreter::contains(adjusted_pc) || CodeCache::contains(adjusted_pc) || StubRoutines::contains(adjusted_pc), "invalid pc");
#else
  // No adjustment needed for interpeter-only system
  address adjusted_pc = pc;  
#endif    
  fr.set_pc(adjusted_pc);
  return fr;
}


//
// Fabricate heavyweight monitor information for each lightweight monitor
// found in this Java thread.
//
void JavaThread::jvmpi_fab_heavy_monitors(bool query, int* index, GrowableArray<ObjectMonitor*>* fab_list) {
  assert(SafepointSynchronize::is_at_safepoint(), "must be at safepoint");

  if (!has_last_Java_frame()) return;
  ResourceMark rm;
  HandleMark   hm;
  RegisterMap reg_map(this);
  vframe* start_vf = last_java_vframe(&reg_map);
  int count = 0;
  for (vframe* f = start_vf; f != NULL; f = f->sender()) {
    if (count++ >= MaxJavaStackTraceDepth) return;  // stack too deep

    if (!f->is_java_frame()) continue;
    javaVFrame* jvf = javaVFrame::cast(f);
    jvf->jvmpi_fab_heavy_monitors(query, index, count, fab_list);
  }
}


//
// If we are trying to enter a monitor in this thread, then return a
// a pointer to it. See javaVFrame::get_pending_monitor().
//
address JavaThread::get_pending_monitor() {
  assert(SafepointSynchronize::is_at_safepoint(), "must be at safepoint");
  if (!has_last_Java_frame()) return NULL;
  ResourceMark rm;
  HandleMark   hm;
  RegisterMap reg_map(this);
  return last_java_vframe(&reg_map)->get_pending_monitor(0);
}


void JavaThread::print_stack() {  
  if (!has_last_Java_frame()) return;
  ResourceMark rm;
  HandleMark   hm;
  RegisterMap reg_map(this);
  vframe* start_vf = last_java_vframe(&reg_map);
  int count = 0;
  for (vframe* f = start_vf; f; f = f->sender() ) {
    if (f->is_java_frame()) {
      javaVFrame* jvf = javaVFrame::cast(f);
      java_lang_Throwable::print_stack_element(NULL, jvf->method(), jvf->bci());
      
      // Print out lock information
      if (JavaMonitorsInStackTrace) {
        jvf->print_lock_info(count);
      }
    } else {
      // Ignore non-Java frames
    }

    // Bail-out case for too deep stacks
    count++;
    if (MaxJavaStackTraceDepth == count) return;    
  }
}


#ifndef PRODUCT

void JavaThread::trace_frames() {
  tty->print_cr("[Describe stack]");
  int frame_no = 1;
  for(StackFrameStream fst(this); !fst.is_done(); fst.next()) {
    tty->print("  %d. ", frame_no++);
    fst.current()->print_value_on(tty,this);
    tty->cr();
  }
}


void JavaThread::trace_stack_from(vframe* start_vf) {
  ResourceMark rm;
  int vframe_no = 1;
  for (vframe* f = start_vf; f; f = f->sender() ) {
    if (f->is_java_frame()) {
      javaVFrame::cast(f)->print_activation(vframe_no++);
    } else {
      f->print();
    }
    if (vframe_no > StackPrintLimit) {
      tty->print_cr("...<more frames>...");
      return;
    }
  }
}


void JavaThread::trace_stack() {
  if (!has_last_Java_frame()) return;
  ResourceMark rm;
  HandleMark   hm;
  RegisterMap reg_map(this);
  trace_stack_from(last_java_vframe(&reg_map));
}


#endif // PRODUCT


javaVFrame* JavaThread::last_java_vframe(RegisterMap *reg_map) {
  assert(reg_map != NULL, "a map must be given");
  frame f = last_frame();
  for (vframe* vf = vframe::new_vframe(&f, reg_map, this); vf; vf = vf->sender() ) {
    if (vf->is_java_frame()) return javaVFrame::cast(vf);
  }
  return NULL;
}


klassOop JavaThread::security_get_caller_class(int depth) {  
  vframeStream vfst(this);
  vfst.security_get_caller_frame(depth);
  if (!vfst.at_end()) {
    return vfst.method()->method_holder();
  }    
  return NULL;
}

// Remove this ifdef when C1 is ported to the compiler interface.
#ifndef CORE
static void compiler_thread_entry(JavaThread* thread, TRAPS) {
  assert(thread->is_Compiler_thread(), "must be compiler thread");
  if (CIMaxPriority) {
    os::set_priority(thread, NearMaxPriority);
  } else {
    os::set_priority(thread, NormPriority);
  }
  CompileBroker::compiler_thread_loop();
}

// Create a CompilerThread
CompilerThread::CompilerThread(CompileQueue* queue)
: JavaThread(&compiler_thread_entry) {
  _queue = queue;
}
#endif


// ======= Threads ========

// The Threads class links together all active threads, and provides
// operations over all threads.  It is protected by its own Mutex
// lock, which is also used in other contexts to protect thread
// operations from having the thread being operated on from exiting
// and going away unexpectedly (e.g., safepoint synchronization)

JavaThread* Threads::_thread_list = NULL;
int	    Threads::_number_of_threads = 0;
int	    Threads::_number_of_non_daemon_threads = 0;
int	    Threads::_return_code = 0;
int         Threads::_priority_queue[MaxPriority + 1]; // Indexed from 1 to 10. Element 0 is not used.
ThreadPriority Threads::_highest_running_priority = NormPriority;
unsigned int    JavaThread::_stack_size_at_create = 0;


#define ALL_THREADS(X) for (JavaThread* X = _thread_list; X; X = X->next())

void os_init();
void os_stream();



jint Threads::create_vm(JavaVMInitArgs* args) {
  
  // Check version
  if (args->version != JNI_VERSION_1_2) return JNI_EVERSION;

  // Initialize the output stream module
  ostream_init();

  // Initialize the os module before using TLS
  os_init();

  // Parse arguments
  jint parse_result = Arguments::parse(args);
  if (parse_result != JNI_OK) return parse_result;

  // Timing (must come after argument parsing)
  TraceTime timer("Create VM", TraceStartupTime);

  // Initialize the os module after parsing the args
  jint os_init_2_result = os_init_2();
  if (os_init_2_result != JNI_OK) return os_init_2_result;

  // Initialize output stream logging
  ostream_init_log();

  // Initialize Threads state
  _thread_list = NULL;
  _number_of_threads = 0;
  _number_of_non_daemon_threads = 0;
  for(int i = 0; i <= MaxPriority; i++) _priority_queue[i] = 0;  

  // Initialize TLS
  ThreadLocalStorage::init();

  // initialize TLE
  if(ParkTLE) {
    TLEQueue::init();
  }

  // Initialize global data structures and create system classes in heap
  vm_init_globals();

  // Attach the main thread to this os thread
  JavaThread* main_thread = new JavaThread();
  main_thread->set_thread_state(_thread_in_vm);
  // must do this before set_active_handles and initialize_thread_local_storage
  main_thread->record_stack_base_and_size(); 
  main_thread->initialize_thread_local_storage();

  main_thread->set_active_handles(JNIHandleBlock::allocate_block());
  if(UseStackBanging) {
    // The call to os::initialize_thread() will map the primordial stack if needed.
    main_thread->enable_stack_yellow_zone();
    main_thread->enable_stack_red_zone();
  }

  main_thread->set_as_starting_thread();  
  HandleMark hm;

  // Initialize global modules
  init_globals();

  { MutexLocker mu(Threads_lock);
    Threads::add(main_thread);
  }

  // Create the VMThread  
  { TraceTime timer("Start VMThread", TraceStartupTime);
    VMThread::create();
    os::create_thread(VMThread::vm_thread(), os::vm_thread);

    // Wait for the VM thread to become ready.
    {
      MutexLocker ml(Notify_lock);
      os::start_thread(VMThread::vm_thread());
      Notify_lock->wait();
    }
  }
  
  assert (Universe::is_fully_initialized(), "not initialized");
  EXCEPTION_MARK;
  {
    TraceTime timer("Initialize java.lang classes", TraceStartupTime);

    if (InitializeJavaLangString) {
      initialize_class(vmSymbolHandles::java_lang_String(), CHECK_0);
    } else {
      warning("java.lang.String not initialized");
    }

    // Initialize java_lang.System (needed before creating the thread)
    if (InitializeJavaLangSystem) {      
      initialize_class(vmSymbolHandles::java_lang_System(), CHECK_0);
      initialize_class(vmSymbolHandles::java_lang_ThreadGroup(), CHECK_0);
      Handle thread_group  = create_initial_thread_group(CHECK_0);
      Universe::set_main_thread_group(thread_group());
      initialize_class(vmSymbolHandles::java_lang_Thread(), CHECK_0);
      oop thread_object = create_initial_thread(thread_group, main_thread, CHECK_0);
      main_thread->set_threadObj(thread_object);
      // The VM preresolve methods to these classes. Make sure that get initialized
      initialize_class(vmSymbolHandles::java_lang_reflect_Method(), CHECK_0);
      initialize_class(vmSymbolHandles::java_lang_ref_Finalizer(),  CHECK_0);
      call_initializeSystemClass(CHECK_0);
    } else {
      warning("java.lang.System not initialized");
    }

    // an instance of OutOfMemory exception has been allocated earlier
    if (InitializeJavaLangExceptionsErrors) {
      initialize_class(vmSymbolHandles::java_lang_OutOfMemoryError(), CHECK_0);
      initialize_class(vmSymbolHandles::java_lang_NullPointerException(), CHECK_0);
      initialize_class(vmSymbolHandles::java_lang_ClassCastException(), CHECK_0);
      initialize_class(vmSymbolHandles::java_lang_ArrayStoreException(), CHECK_0);
      initialize_class(vmSymbolHandles::java_lang_ArithmeticException(), CHECK_0);
      initialize_class(vmSymbolHandles::java_lang_StackOverflowError(), CHECK_0);      
    } else {
      warning("java.lang.OutOfMemoryError has not been initialized");
      warning("java.lang.NullPointerException has not been initialized");
      warning("java.lang.ClassCastException has not been initialized");
      warning("java.lang.ArrayStoreException has not been initialized");
      warning("java.lang.ArithmeticException has not been initialized");
      warning("java.lang.StackOverflowError has not been initialized");
    }
  }
  
  // See        : bugid 4211085.  
  // Background : the static initializer of java.lang.Compiler tries to read 
  //              property"java.compiler" and read & write property "java.vm.info".
  //              When a security manager is installed through the command line
  //              option "-Djava.security.manager", the above properties are not
  //              readable and the static initializer for java.lang.Compiler fails
  //              resulting in a NoClassDefFoundError.  This can happen in any
  //              user code which calls methods in java.lang.Compiler.  
  // Hack :       the hack is to pre-load and initialize this class, so that only
  //              system domains are on the stack when the properties are read.
  //              Currently even the AWT code has calls to methods in java.lang.Compiler.
  //              On the classic VM, java.lang.Compiler is loaded very early to load the JIT.
  // Future Fix : the best fix is to grant everyone permissions to read "java.compiler" and 
  //              read and write"java.vm.info" in the default policy file. See bugid 4211383
  //              Once that is done, we should remove this hack.
  initialize_class(vmSymbolHandles::java_lang_Compiler(), CHECK_0);
  
  // More hackery - the static initializer of java.lang.Compiler adds the string "nojit" to
  // the java.vm.info property if no jit gets loaded through java.lang.Compiler (the hotspot
  // compiler does not get loaded through java.lang.Compiler).  "java -version" with the 
  // hotspot vm says "nojit" all the time which is confusing.  So, we reset it here.
  // This should also be taken out as soon as 4211383 gets fixed.
  reset_vm_info_property(CHECK_0);
  
  // Set flag that basic initialization has completed. Used by exceptions and various 
  // debug stuff, that does not work until all basic classes have been initialized.
  set_init_completed();

  // Compute system loader. Note that this has to occur after set_init_completed, since
  // valid exceptions may be thrown in the process. 
  // Note that we do not use CHECK_0 here since we are inside an EXCEPTION_MARK and 
  // set_init_completed has just been called, causing exceptions not to be shortcut
  // anymore. We call vm_exit_during_initialization directly instead.
  SystemDictionary::compute_java_system_loader(THREAD);
  if (HAS_PENDING_EXCEPTION) {
    vm_exit_during_initialization(PENDING_EXCEPTION);
  }
  
  if (Arguments::init_libraries_at_startup()) create_vm_init_libraries();

  // Notify tool interfaces that VM initialization is complete.
  if (jvmdi::enabled()) {
    jvmdi::post_vm_initialized_event();
  }
  if (jvmpi::enabled() && UseTrainGC) {
    vm_exit_during_initialization("JMVPI not supported with option -Xincgc");
  }
  if (jvmpi::is_event_enabled(JVMPI_EVENT_JVM_INIT_DONE)) {
    // Issue class load, thread start and object allocation events for all preloaded classes
    jvmpi::post_vm_initialization_events();

    // Tell the agent that the VM is ready and events requested (e.g. heap dump)
    // from this point onwards will contains IDs that are all known to the agent
    jvmpi::post_vm_initialized_event();
  }

  Chunk::start_chunk_pool_cleaner_task();
#ifndef CORE
  RecompilationMonitor::start_recompilation_monitor_task();  
#endif
  os::signal_init();

  // initialize compiler
  COMPILER1_ONLY(CompileBroker::compilation_init(new Compiler());)
  COMPILER2_ONLY(CompileBroker::compilation_init(new C2Compiler());)

  if (Arguments::has_profile())       FlatProfiler::engage(main_thread, true);
  if (Arguments::has_alloc_profile()) AllocationProfiler::engage();
  if (MemProfiling)                   MemProfiler::engage();    

  return JNI_OK;
}

extern "C" {
  typedef jint (JNICALL *JVM_OnLoad_t)(JavaVM *, char *, void *);
}
void Threads::create_vm_init_libraries() {
  const char *on_load_symbols[] = JVM_ONLOAD_SYMBOLS;
  extern struct JavaVM_ main_vm; // Should this be fetched, instead?

  for(int i=0; i<Arguments::library_count(); i++) {
    char buffer[JVM_MAXPATHLEN];
    JVM_OnLoad_t JVM_OnLoad = NULL;
    void *library;
    char ebuf[1024];

    // Try to load the library from the standard dll directory
    hpi::dll_build_name(buffer, sizeof(buffer), os::get_system_properties()->dll_dir, Arguments::library_name(i));
    library = hpi::dll_load(buffer, ebuf, sizeof ebuf);
    if(library == NULL) { // Try the local directory
      char ns[1] = {0};
      hpi::dll_build_name(buffer, sizeof(buffer), ns, Arguments::library_name(i));
      library = hpi::dll_load(buffer, ebuf, sizeof ebuf);
      if (library == NULL) {
        // If we can't find the library, exit.
        vm_exit_during_initialization("Could not find -Xrun library", buffer);
      }
    }

    // Find the JVM_OnLoad function.
    for (int symbol_index = 0; symbol_index<sizeof(on_load_symbols) / sizeof(char*); symbol_index++) {
      JVM_OnLoad = CAST_TO_FN_PTR(JVM_OnLoad_t, hpi::dll_lookup(library, on_load_symbols[symbol_index]));
      if(JVM_OnLoad != NULL) break;
    }

    // Invoke the JVM_OnLoad function
    if (JVM_OnLoad != NULL) {
      JavaThread* thread = JavaThread::current();
      StateSaver nsv(thread); // we must have post Java state set while in native
      ThreadToNativeFromVM ttn(thread);
      HandleMark hm(thread);
      jint err = (*JVM_OnLoad)(&main_vm, Arguments::library_options(i), NULL);
      if (err != 0) {
        vm_exit_during_initialization("-Xrun library failed to init", Arguments::library_name(i));
      }
    } else {
      vm_exit_during_initialization("Could not find JVM_OnLoad function in -Xrun library", Arguments::library_name(i));
    }
  }
}

class VM_SuspendDaemons: public VM_Operation {
 private:
  JavaThread* _thread;
 public:
  VM_SuspendDaemons(JavaThread* thread) {
    _thread = thread; 
  }  
  void doit() {
    // make doubly sure that a thread doesn't wake up at an inconvenient time
    for (JavaThread *thr = Threads::first(); thr; thr = thr->next()) {
      if (thr == _thread)                  continue;  // ignore caller
      if (thr->is_VM_thread())             continue;  // ignore VMThread
      if (thr->is_SuspendChecker_thread()) continue;  // ignore SuspendCheckerThread
      if (!thr->is_Java_thread())          continue;  // ignore non JavaThreads

      // ignore JavaThreads in the process of exiting
      if (java_lang_Thread::thread(thr->threadObj()) == NULL) continue;
      // Since we are at a safepoint, we don't have to refresh thr
      // at every retry attempt.
      while (true) {
        // Request that the thread be suspended. For threads in state
        // _thread_in_native, the suspend request is immediate.
        // Solaris and Linux will force suspension for all threads that
        // are still running in VM_Exit via shutdown_suspend().
        thr->send_java_suspend(true);

        //
        // See if the VMThread is likely to deadlock due to this suspend:
        // - tell SuspendCheckerThread to check for deadlock
        // - briefly wait for it to tell us it is still okay
        // - if we time out, then we want to undo this suspend
        //
        MutexLockerEx scl(SuspendChecker_lock, Mutex::_no_safepoint_check_flag);
        // If the SuspendCheckerThread is not ready for work yet or has
        // exited during VM shutdown, then we skip checking this suspend
        // operation.
        if (SuspendCheckerThread::suspend_checker_thread() == NULL) break;
        SuspendChecker_lock->notify();
        bool timedOut = SuspendChecker_lock->wait(true, 1000);
        if (!timedOut) break;
  
        //
        // The SuspendCheckerThread is stuck. The thread we just suspended
        // may be holding some critical resource which means the VMThread
        // will deadlock if we leave the thread suspended. It is also
        // possible that some other running thread temporarily has the
        // resource and for some reason has held it longer than our timeout
        // above. The latter is the false positive case and it shouldn't
        // hurt to temporarily resume the thread.
        //
        thr->resume(true, true);  // yield, external resume request
      }
  
      // It would be nice to have the following assertion in all the
      // time, but it is possible for a racing resume request to have
      // resumed this thread right after we suspended it. Temporarily
      // enable this assertion if you are chasing a different kind of
      // bug.
      //
      // assert(java_lang_Thread::thread(thr->threadObj()) == NULL ||
      //   thr->is_being_ext_suspended(), "thread is not suspended");
    }
  }

  const char* name() const { return "suspend daemons"; }
};


bool Threads::destroy_vm() {
  JavaThread* thread = JavaThread::current();

  // Wait until we are the last thread to execute
  { MutexLocker nu(Threads_lock);
    while (Threads::number_of_non_daemon_threads() > 1 ) 
      Threads_lock->wait();
  }

  if (FlatProfiler::is_active()) {
    FlatProfiler::disengage();
  }

  // Terminate the signal thread
  os::terminate_signal_thread();

  // Terminate watcher thread
  WatcherThread::stop();

  if (Universe::is_jdk12x_version()) {
    // We are the last thread running, so check if finalizers should be run and then stop the VM thread.  
    // For 1.3 or later this is done in thread->exit(true)
    HandleMark rm;  
    Universe::run_finalizers_on_exit();               
  }
  
  before_exit();
  thread->exit(true); 

  // Make sure daemon threads are quiescent
  bool daemons = false;

  { MutexLocker nu(Threads_lock);
    // ignore VMThread and SuspendCheckerThread
    if (Threads::number_of_threads() > 2)
      daemons = true;
  }


  if (daemons) {
    // Fire off a VM_Operation to suspend all of the daemon threads
    VM_SuspendDaemons op(thread);
    VMThread::execute(&op);
  }

  // Terminate suspend checker thread. We do this after the daemon
  // threads are suspended since there is no longer any chance of a
  // VM_ThreadSuspend operation deadlocking the VMThread.
  SuspendCheckerThread::stop();

  // Stop VM thread. This must be done after thread->exit(NULL) since it calls Java code when
  // notifying the threadGroup about a thread exit.
  VMThread::wait_for_vm_thread_exit();  

  thread->delete_thread_and_TLS_current();

  // All threads have exited so it is safe to call exit_globals (will disable thread-local-storage etc.)  
  exit_globals();

  return true;
}


void Threads::add(JavaThread* p) {
  // The threads lock must be owned at this point
  assert_lock(Threads_lock);
  add_priority(p);
  p->set_next(_thread_list);  
  _thread_list = p;
  _number_of_threads++;
  oop threadObj = p->threadObj();
  // Bootstrapping problem: threadObj can be null for initial JavaThread
  if (threadObj == NULL || !java_lang_Thread::is_daemon(threadObj)) {
    _number_of_non_daemon_threads++;
  }

  // Possible GC point.
  Events::log("Thread added: 0x%x", p);
}

void Threads::remove(JavaThread* p) {  
  // Extra scope needed for Thread_lock, so we can check
  // that we do not remove thread without safepoint code notice
  { MutexLocker ml(Threads_lock);
  
    assert(includes(p), "p must be present");
    remove_priority(p);

    JavaThread* current = _thread_list; 
    JavaThread* prev    = NULL;

    while (current != p) {
      prev    = current;
      current = current->next();
    }

    if (prev) {
      prev->set_next(current->next());
    } else {
      _thread_list = p->next();
    }  
    _number_of_threads--;
    oop threadObj = p->threadObj();
    if (threadObj == NULL || !java_lang_Thread::is_daemon(threadObj)) {
      _number_of_non_daemon_threads--;
      // Only one thread left, do a notify on the Threads_lock so a thread waiting
      // on destroy_vm will wake up.
      if (number_of_non_daemon_threads() == 1) 
        Threads_lock->notify_all();
    }    

    // Make sure that safepoint code disregard this thread. This is needed since
    // the thread might mess around with locks after this point. This can cause it
    // to do callbacks into the safepoint code. However, the safepoint code is not aware
    // of this thread since it is removed from the queue.
    p->set_is_terminated(true);        
  } // unlock Threads_lock 

  // Since Events::log uses a lock, we grab it outside the Threads_lock
  Events::log("Thread exited: 0x%x", p);
}

// Threads_lock must be held when this is called (or must be called during a safepoint)
bool Threads::includes(JavaThread* p) {  
  assert(Threads_lock->is_locked(), "sanity check");
  ALL_THREADS(q) {
    if (q == p ) {
      return true;
    }
  }
  return false;
}

// Operations on the Threads list for GC.  These are not explicitly locked,
// but the garbage collector must provide a safe context for them to run.
// In particular, these things should never be called when the Threads_lock
// is held by some other thread. (Note: the Safepoint abstraction also
// uses the Threads_lock to gurantee this property. It also makes sure that  
// all threads gets blocked when exiting or starting).

void Threads::oops_do(void f(oop*)) {
  ALL_THREADS(p) {
    p->oops_do(f);
  }
  VMThread::vm_thread()->oops_do(f);
}

void Threads::gc_epilogue() {
  ALL_THREADS(p) {
    p->gc_epilogue();
  }
}

void Threads::gc_prologue() {
  ALL_THREADS(p) {    
    p->gc_prologue();
  }
}

void Threads::deoptimized_wrt_marked_nmethods() {
  ALL_THREADS(p) {
    p->deoptimized_wrt_marked_nmethods();
  }
}


//
// Fabricate heavyweight monitor information for each lightweight monitor
// found in every Java thread.
//
GrowableArray<ObjectMonitor*>* Threads::jvmpi_fab_heavy_monitors() {
  assert(SafepointSynchronize::is_at_safepoint(), "must be at safepoint");

  // In the first pass we have to count the number of monitors needed
  int index = 0;
  ALL_THREADS(p) {
    NOT_CORE(if (p->is_Compiler_thread()) continue;)

    p->jvmpi_fab_heavy_monitors(true, &index, NULL);
  }

  GrowableArray<ObjectMonitor*>* result = new GrowableArray<ObjectMonitor*>(index, index, NULL);
  //
  // In the second pass we actually get the data. We can't call append()
  // down in javaVFrame::jvmpi_fab_heavy_monitors() because there are other
  // uses of ResourceMark between here and there.
  //
  index = 0;
  ALL_THREADS(q) {
    NOT_CORE(if (q->is_Compiler_thread()) continue;)

    q->jvmpi_fab_heavy_monitors(false, &index, result);
  }
  return result;
}


// Get count Java threads that are waiting to enter the specified monitor.
GrowableArray<JavaThread*>* Threads::get_pending_threads(int count, address monitor) {
  assert(SafepointSynchronize::is_at_safepoint(), "must be at safepoint");
  GrowableArray<JavaThread*>* result = new GrowableArray<JavaThread*>(count);

  int i = 0;
  ALL_THREADS(p) {
    NOT_CORE(if (p->is_Compiler_thread()) continue;)

    address pending = p->get_pending_monitor();
    if (pending == monitor) {             // found a match
      if (i < count) result->append(p);   // save the first count matches
      i++;
    }
  }
  CORE_ONLY(assert(count == i, "Did not find the expected number of pending Java threads");)
  return result;
}


JavaThread* Threads::owning_thread_from_monitor(ObjectMonitor* mid) {
  assert(SafepointSynchronize::is_at_safepoint(), "must be at safepoint");

  address owner = (address)mid->owner();
  // NULL owner means not locked so we can skip the search
  if (owner == NULL) return NULL;

  ALL_THREADS(p) {	// first, see if owner is the address of a Java thread
    if (owner == (address)p) return p;
  }
  CORE_ONLY(assert(UseHeavyMonitors == false, "Did not find owning Java thread with UseHeavyMonitors enabled");)
  if (UseHeavyMonitors) return NULL;

  //
  // If we didn't find a matching Java thread and we didn't force use of
  // heavyweight monitors, then the owner is the stack address of the
  // Lock Word in the owning Java thread's stack.
  //
  // We can't use Thread::is_lock_owned() or Thread::is_in_stack() because
  // those routines rely on the "current" stack pointer. That would be our
  // stack pointer which is not relevant to the question. Instead we use the
  // highest lock ever entered by the thread and find the thread that is
  // higher than and closest to our target stack address.
  //
  address    least_diff = 0;
  bool       least_diff_initialized = false;
  JavaThread* the_owner = NULL;
  ALL_THREADS(q) {
    address addr = q->highest_lock();
    if (addr == NULL || addr < owner) continue;  // thread has entered no monitors or is too low
    address diff = (address)(addr - owner);
    if (!least_diff_initialized || diff < least_diff) {
      least_diff_initialized = true;
      least_diff = diff;
      the_owner = q;
    }
  }
  assert(the_owner != NULL, "Did not find owning Java thread for lock word address");
  return the_owner;
}


void Threads::print(bool print_stacks, bool internal_format) {
  tty->print_cr("Full thread dump:");
  tty->cr();
  ALL_THREADS(p) {
    NOT_CORE(if (p->is_Compiler_thread()) continue;)

    ResourceMark rm;
    p->print();
    if (print_stacks) {
      if (internal_format) {
        p->trace_stack();
      } else {
        p->print_stack();
      }
    }
    tty->cr();
  }
  VMThread::vm_thread()->print();
  tty->cr();
  WatcherThread::watcher_thread()->print();
  SuspendCheckerThread::suspend_checker_thread()->print();
  NOT_CORE(CompileBroker::print_compiler_threads();)
}


// Priority management
ThreadPriority Threads::lowest_priority() {
  int priority = MinPriority;
  while(_priority_queue[priority] == 0 && priority < MaxPriority) {
    priority++;
  }
  assert(MinPriority <= priority && priority <= MaxPriority, "sanity check");  
  return (ThreadPriority)priority;
}

void Threads::compute_highest_running_priority() {
  assert_lock(Threads_lock);
  ThreadPriority max_p = NoPriority;
  ALL_THREADS(thread) {
    bool is_suspended;
    {
      MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
      is_suspended = thread->is_suspended();
    }
    if (!thread->in_jvm_wait() && !is_suspended) {
      ThreadPriority p = thread->java_priority();
      if (p > max_p) max_p = p;
    }
  }
  _highest_running_priority = (max_p == NoPriority ? MinPriority/*NormPriority*/ : max_p);
}


void Threads::add_priority(JavaThread* thread) {
  assert_lock(Threads_lock);
  ThreadPriority priority = thread->java_priority(); // Read priority from Java object
  thread->set_priority_copy(priority);               // Save priority in JavaThread object
  assert(MinPriority <= priority && priority <= MaxPriority, "sanity check");  
  _priority_queue[priority]++;
}


void Threads::remove_priority(JavaThread* thread) {
  assert_lock(Threads_lock);
  ThreadPriority priority = thread->priority_copy(); // Read directly from Thread
  assert(MinPriority <= priority && priority <= MaxPriority, "sanity check");
  _priority_queue[priority]--;
  assert(_priority_queue[priority] >= 0, "wrong priority removal");
}


void Threads::change_priority(JavaThread* thread, ThreadPriority after) {
  assert_lock(Threads_lock);
  ThreadPriority before = thread->priority_copy();
  assert(MinPriority <= before && before <= MaxPriority, "sanity check");
  assert(MinPriority <= after  && after  <= MaxPriority, "sanity check");
  _priority_queue[before]--;
  assert(_priority_queue[before] >= 0, "wrong priority removal");
  Thread::set_priority(thread, after);
  _priority_queue[after]++;
  thread->set_priority_copy(after);  
}


//------------------------------------------------------------------------------------------------------------------
// Debugging code. Not included in PRODUCT version

#ifndef PRODUCT

void Threads::verify() {  
  ALL_THREADS(p) {
    p->verify();
  }
  VMThread* thread = VMThread::vm_thread();
  if (thread != NULL) thread->verify();
}

void Threads::print_on_error(JavaThread* current) {
  tty->print_cr("Java Threads:");
  ALL_THREADS(p) {
    ResourceMark rm;
    if (current == p) {
      tty->print("[CURRENT] ");
    }
    p->print();
    p->trace_stack();
  }
}

#endif
