/*
 * @(#)thread.hpp	1.296 01/03/15 14:47:27
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

class ThreadSafepointState;
class ThreadProfiler;

class JvmdiThreadState;
class JvmdiGetLoadedClassesClosure;

class ciEnv;
class CompileTask;
class CompileQueue;

class Suspend_Callback;
class Resume_Callback;

#ifndef PRODUCT
class MethodTracer; // used for -XX:+TraceMethods
#endif


// Class hierarchy
// - Thread
//   - VMThread
//   - JavaThread
//   - WatcherThread
//   - SuspendCheckerThread

class Thread: public ThreadShadow {
  friend class VMStructs;
 private:
  // Exception handling
  // (Note: _pending_exception is in ThreadShadow)
  //oop       _pending_exception;                // pending exception for current thread
  const char* _exception_file;                   // file information for exception (debugging only)
  int         _exception_line;                   // line information for exception (debugging only)

  RawMonitor* _rawmonitor_list;                  // list of locked raw monitors (JVMPI support)

  // Debug tracing
  static void trace(const char* msg, Thread* thread) PRODUCT_RETURN;

  // Active_handles points to a block of handles
  JNIHandleBlock* _active_handles;

  // One-element thread local free list
  JNIHandleBlock* _free_handle_block;

  // Point to the last handle mark
  HandleMark* _last_handle_mark;

  public:
   void set_last_handle_mark(HandleMark* mark)   { _last_handle_mark = mark; }
   HandleMark* last_handle_mark() const          { return _last_handle_mark; }
  private:

  // debug support for checking if code does allow safepoints or not
  // GC points in the VM can happen because of allocation, invoking a VM operation, or blocking on 
  // mutex, or blocking on an object synchronizer (Java locking).
  // If !allow_safepoint(), then an assertion failure will happen in any of the above cases
  // If !allow_allocation(), then an assertion failure will happen during allocation
  // (Hence, !allow_safepoint() => !allow_allocation()).
  //
  // The two classes No_Safepoint_Verifier and No_Allocation_Verifier are used to set these counters.
  //
  NOT_PRODUCT(int _allow_safepoint_count;)       // If 0, thread allow a safepoint to happen
  debug_only (int _allow_allocation_count;)      // If 0, the thread is allowed to allocate oops.  

  friend class No_Alloc_Verifier;
  friend class No_Safepoint_Verifier;
  friend class ThreadLocalStorage;

  // In order for all threads to be able to use the fast locking, we need to know the higest stack
  // address of where a lock is on the stack (stacks normally grows towards lower addresses). This
  // variable is initially set to NULL, indicating no locks are used by the thread. During the threads
  // execution, it will be set whenever locking can happen, i.e., when we call out to Java code or use
  // an ObjectLocker. The value is never decreased, hence, it will over the life time of a thread trying
  // to approximate the real stackbase. 
  address _highest_lock;                         // Highest stack address where a JavaLock exist
  ThreadPriority _unboosted_priority;            // Old priority if boosted, else NoPriority
  
  ThreadLocalEden _tle;                          // Thread-local eden

  long  _vm_operation_started_count;             // VM_Operation support
  long  _vm_operation_completed_count;           // VM_Operation support
  
 public: 
  enum {
    is_definitely_current_thread = true
  };	

  // Constructor
  Thread();
  virtual ~Thread();

  // initializtion
  void initialize_thread_local_storage();

  // thread entry point
  virtual void run();

  // cleanup
  void delete_thread_and_TLS_current();

  // Testers
  virtual bool is_VM_thread()       const        { return false; }
  virtual bool is_Java_thread()     const        { return false; }
  // Remove this ifdef when C1 is ported to the compiler interface.
#ifndef CORE
  virtual bool is_Compiler_thread() const        { return false; }
#endif
  virtual bool is_jvmpi_daemon_thread() const    { return false; }
  virtual bool is_Watcher_thread()  const        { return false; }
  virtual bool is_SuspendChecker_thread() const  { return false; }

  // Returns the running thread JavaThread, VMThread, WatcherThread,
  // or SuspendCheckerThread
  static Thread* current();  

  // Common thread operations
  static void set_priority(Thread* thread, ThreadPriority priority);
  static ThreadPriority get_priority(Thread* thread);
  static void boost_priority(Thread* thread, ThreadPriority priority);
  static void unboost_priority(Thread* thread);
  static void start(Thread* thread);
  static void interrupt(Thread* thr);
  static bool is_interrupted(Thread* thr, bool clear_interrupted);  

  // suspend/resume support
  long suspend_other(bool ext_req);
  long shutdown_suspend();
  long suspend_thread_self(Suspend_Callback& cb);
  long resume(bool _do_wait, bool ext_req);

  // Fast thread ids are used for speculative locking
  virtual void set_fast_thread_id()              { _fast_thread_id = -1; }
  int get_fast_thread_id()                       { return _fast_thread_id; }

  // Installs a pending exception to be inserted later
  static void send_async_exception(oop thread_oop, oop java_throwable);

  // Boosted priority
  ThreadPriority unboosted_priority() const      { return _unboosted_priority; }
  void set_unboosted_priority(ThreadPriority priority) { _unboosted_priority = priority; }

  // Resource area
  ResourceArea* resource_area() const            { return _resource_area; }
  void set_resource_area(ResourceArea* area)     { _resource_area = area; }

  OSThread* osthread() const                     { return _osthread;   }
  void set_osthread(OSThread* thread)            { _osthread = thread; }

  // JNI handle support
  JNIHandleBlock* active_handles() const         { return _active_handles; }
  void set_active_handles(JNIHandleBlock* block) { _active_handles = block; }
  JNIHandleBlock* free_handle_block() const      { return _free_handle_block; }
  void set_free_handle_block(JNIHandleBlock* block) { _free_handle_block = block; }

  // Internal handle support
  HandleArea* handle_area() const                { return _handle_area; }
  void set_handle_area(HandleArea* area)         { _handle_area = area; }

  // Exceptions
  void set_pending_exception(oop exception, const char* file, int line);   // use THROW whenever possible!
  void clear_pending_exception();                                          // use CLEAR_PENDING_EXCEPTION whenever possible!
  bool has_pending_exception() const             { return pending_exception() != NULL; }    
  oop  pending_exception() const                 { return _pending_exception; } 
  const char* exception_file() const             { return _exception_file; }
  int  exception_line() const                    { return _exception_line; }

  // Thread-Local Edens (TLE) support
  ThreadLocalEden& tle()                         { return _tle; }

  // VM operation support
  long vm_operation_ticket()                     { return ++_vm_operation_started_count; }
  long vm_operation_completed_count()            { return _vm_operation_completed_count; }
  void increment_vm_operation_completed_count()  { _vm_operation_completed_count++; }

  // JVMPI support
  RawMonitor* rawmonitor_list() const            { return _rawmonitor_list; }
  void        set_rawmonitor_list(RawMonitor* r) { _rawmonitor_list = r;    }

  // GC support
  void oops_do(void f(oop*));  
  
  // Brackets required around any block into a safepoint:
  virtual jint* save_state_before_block()              { assert(current() == this, "must save own state"); return NULL; }
  virtual void  restore_state_after_block(jint* state) { assert(current() == this, "must restore own state"); }

  // Fast-locking support
  address highest_lock() const                   { return _highest_lock; }
  void update_highest_lock(address base)         { if (base > _highest_lock) _highest_lock = base; }

  // Tells if adr belong to this thread. This is used
  // for checking if a lock is owned by the running thread.
  // Warning: the method can only be used on the running thread
  // Fast lock support uses these methods
  virtual bool is_in_stack(address adr) const;
  virtual bool is_lock_owned(address adr) const;  

  // Sets this thread as starting thread
  void set_as_starting_thread();

 protected:
  // OS data associated with the thread
  OSThread* _osthread;  // Platform-specific thread information 

  // Thread local resource area for temporary allocation within the VM
  ResourceArea* _resource_area;

  // Thread local handle area for allocation of handles within the VM
  HandleArea* _handle_area;

  // Support for stack overflow handling, get_thread, etc.  
  address          _stack_base;
  address          _stack_limit;
  int              _stack_size;
  unsigned int     _last_sp_hash;     // used by get_thread (mutable)
  int              _fast_thread_id;   // used for speculative locking

  // Internal common implementations of suspend and resume
  long suspend_thread_impl(Suspend_Callback& cb, bool ext_req);
  long resume_thread_impl(Resume_Callback& cb, bool ext_req);

 public:
  // Stack overflow support
  address  stack_base() const          { assert(_stack_base != NULL,"Sanity check"); return _stack_base; }
  
  void set_stack_base(address base)    { _stack_base = base; }
  int  stack_size() const              { assert(_stack_size > 0,"Sanity check"); return _stack_size; }
  void set_stack_size(int size)        { _stack_size = size; }

  // FIXIT: Currently used on win32-only
  address  stack_limit() const         { return _stack_limit; }
  void set_stack_limit(address limit)  { _stack_limit = limit; }
  

  void record_stack_base_and_size();
  static void force_stack_page_mapping_to(address addr);

  // Assembly code interface
  static int active_handles_offset_in_bytes()    { return (int)&((Thread*)NULL)->_active_handles; }

  // Printing
  void print();
  // Debug-only code
#ifdef ASSERT
 private:
  // Deadlock detection support for Mutex locks. List of locks own by thread.
  Mutex *_owned_locks;
  // Mutex::set_owner_implementation is the only place where _owned_locks is modified, thus the friendship
  friend void Mutex::set_owner_implementation(Thread *owner);

 public:  
  void print_owned_locks();
  bool owns_locks() const                        { return (_owned_locks != NULL); }  
  bool owns_locks_but_compiled_lock() const;

  virtual bool state_is_saved_before_block()     { return true; }

  // Deadlock detection
  bool allow_allocation()                        { return _allow_allocation_count == 0; }    

#endif
  void check_for_valid_safepoint_state(bool potential_vm_operation) PRODUCT_RETURN;

  // Code generation
  static int pending_exception_offset_in_bytes() { return (int)&((Thread*)NULL)->_pending_exception ; }
  static int exception_file_offset_in_bytes()    { return (int)&((Thread*)NULL)->_exception_file; }
  static int exception_line_offset_in_bytes()    { return (int)&((Thread*)NULL)->_exception_line; }

  static int tle_start_offset_in_bytes()         { return (int)&((Thread*)NULL)->_tle + ThreadLocalEden::start_offset_in_bytes(); }
  static int tle_end_offset_in_bytes()           { return (int)&((Thread*)NULL)->_tle + ThreadLocalEden::end_offset_in_bytes(); }
  static int tle_top_offset_in_bytes()           { return (int)&((Thread*)NULL)->_tle + ThreadLocalEden::top_offset_in_bytes(); }
  static int tle_size_offset_in_bytes()          { return (int)&((Thread*)NULL)->_tle + ThreadLocalEden::size_offset_in_bytes(); }
};


// Inline implementation of Thread::current()
inline Thread* Thread::current() {
  // Obtain thread using inlined assembler-generated code
  Thread* thread = ThreadLocalStorage::thread();
  assert(thread != NULL, "just checking");
  return thread;
}


// A single WatcherThread is used for simulating timer interrupts.
class WatcherThread: public Thread {
  friend class VMStructs;
 public:
  virtual void run();

 private:
  static WatcherThread* _watcher_thread;

  static bool _should_terminate;  
 public:
  enum SomeConstants {
    delay_interval = 10                          // interrupt delay in milliseconds
  };

  // Constructor
  WatcherThread();

  // Tester
  bool is_Watcher_thread() const                 { return true; }

  // Printing
  void print();

  // Returns the single instance of WatcherThread
  static WatcherThread* watcher_thread()         { return _watcher_thread; }

  // Create and start the single instance of WatcherThread, or stop it on shutdown
  static void start();
  static void stop();
};


// A single SuspendCheckerThread is used by the VMThread to see if
// the current suspend operation will cause a VMThread deadlock.
class SuspendCheckerThread : public Thread {
  friend class VMStructs;
 public:
  virtual void run();

 private:
  static SuspendCheckerThread* _suspend_checker_thread;

  static bool _should_terminate;  
 public:

  // Constructor
  SuspendCheckerThread();

  // Tester
  bool is_SuspendChecker_thread() const          { return true; }

  // Printing
  void print();

  // Returns the single instance of SuspendCheckerThread
  static SuspendCheckerThread* suspend_checker_thread() {
    return _suspend_checker_thread; }

  // Create and start the single instance of SuspendCheckerThread,
  // or stop it on shutdown
  static void start();
  static void stop();
};


typedef void (*ThreadFunction)(JavaThread*, TRAPS);

class JavaThread: public Thread {  
  friend class VMStructs;
 private:  
  JavaThread*    _next;                          // The next thread in the Threads list
  oop            _threadObj;                     // The Java level thread object
  ThreadPriority _priority_copy;                 // Copy of java_langThread::priority(theadObj())

#ifdef ASSERT
 private:
  int _java_call_counter;

 public:
  int  java_call_counter()                       { return _java_call_counter; }
  void inc_java_call_counter()                   { _java_call_counter++; }
  void dec_java_call_counter() {
    assert(_java_call_counter > 0, "Invalid nesting of JavaCallWrapper");
    _java_call_counter--;
  }
#endif	// ifdef ASSERT

 private:
   enum Constants {
     not_at_call_bit  = 1                        // Used to mark existent of a not_at_call frame
   };

  volatile jint* _last_Java_sp;    
  volatile jint* _not_at_call_sp;
 
  ThreadFunction _entry_point;

  JNIEnv        _jni_environment;
  
  vframeArray*  _vframe_array_head;              // Holds the heap fo the active vframeArrays  
  
  // Oop results of VM runtime calls  
  oop           _vm_result;                      // Used to pass back an oop result into Java code, GC-preserved
  oop           _vm_result_2;                    // Used to pass back an oop result into Java code, GC-preserved

  MonitorChunk* _monitor_chunks;                 // Contains the off stack monitors 
                                                 // allocated during deoptimization
                                                 // and by JNI_MonitorEnter/Exit

  MonitorArray* _monitor_arrays;                 // Contains the on stack monitors 
                                                 // allocated during deoptimization

  // Async. exceptions/thread self-suspend support
  bool          _special_runtime_exit_condition; // Flag indicating pending async. exception/self-suspend operation
  oop           _pending_async_exception;
  bool          _is_handling_implicit_exception; // setting it true additionally to not_at_call_sp
                                                 // used only by C1 code; must be reset before unwinding stack

  // Safepoint support
  volatile JavaThreadState _thread_state;
  volatile address      _last_native_pc;         // Contains address of last_Java_pc when calling out to native
  ThreadSafepointState *_safepoint_state;        // Holds information about a thread during a safepoint
  address               _saved_exception_pc;     // Saved pc of instruction where last implicit exception happened
  bool                  _terminated;             // True, when thread is removed from Thread list
  // suspend/resume support
  bool                  _external_suspend;       // External suspend request
  bool                  _internal_suspend;       // Internal suspend request
  bool                  _is_suspended;           // Thread is suspended
  bool                  _in_jvm_wait;            // Thread is blocked in java.lang.Object.wait
  bool                  _has_called_back;        // Thread has called back the safepoint code (for debugging)

  // Compiler exception handling (NOTE: The _exception_oop is *NOT* the same as _pending_exception. It is
  // used to temp. parsing values into and out of the runtime system during exception handling for compiled
  // code)
  volatile oop     _exception_oop;               // Exception thrown in compiled code
  volatile address _exception_pc;                // PC where exception happened
  volatile address _exception_handler_pc;        // PC for handler of exception
  volatile int     _exception_stack_size;        // Size of frame where exception happened
  volatile address _last_exception_pc;           // PC where last exception happened (used for unwinding)

  // support for compilation
  bool    _is_compiling;                         // is true if a compilation is active inthis thread (one compilation per thread possible)

  // support for JVMPI
  void* _jvmpi_data;
  unsigned int  _last_sum;

  friend class VMThread;
  friend class ThreadWaitTransition;

  void initialize();                             // Initialized the instance variables

 public:
  // Constructor
  JavaThread(); // for main thread
  JavaThread(ThreadFunction entry_point); 
  ~JavaThread();

  // Cleanup on thread exit
  void exit(bool destroy_vm);

  // Testers
  virtual bool is_Java_thread() const            { return true;  }

  // compilation
  void set_is_compiling(bool f)                  { _is_compiling = f; }
  bool is_compiling() const                      { return _is_compiling; }

  // Thread chain operations
  JavaThread* next() const	                 { return _next; }
  void set_next(JavaThread* p)                   { _next = p; }
  
  // Thread oop
  oop threadObj() const                          { return _threadObj; }
  void set_threadObj(oop p)                      { _threadObj = p; }

  ThreadPriority java_priority() const;          // Read from threadObj()
  ThreadPriority priority_copy()                 { return _priority_copy; }
  void set_priority_copy(ThreadPriority p)       { _priority_copy = p; }

  // Prepare thread and add to priority queue.  If a priority is
  // not specified, use the priority of the jthread. Threads_lock
  // must be held while this function is called.
  void prepare(jthread jni_thread, ThreadPriority prio=NoPriority);

  void set_saved_exception_pc(address pc)        { _saved_exception_pc = pc; }
  address saved_exception_pc()                   { return _saved_exception_pc; }
  
    
  ThreadFunction entry_point() const             { return _entry_point; }

  // Allocates a new Java level thread object for this thread. thread_name may be NULL.
  void allocate_threadObj(Handle thread_group, char* thread_name, TRAPS);

  // Thread stack walking state saving [bracketing] functions
  virtual jint* save_state_before_block() { 
    Thread::save_state_before_block();
    assert(Thread::current() == this, "this must be current thread");
    pd_cache_state(is_definitely_current_thread);
    return pd_cached_state(); 
  }
  virtual void restore_state_after_block(jint *state) { 
    Thread::restore_state_after_block(NULL);
    pd_set_cached_state(state);
  }

  // Thread cached stack walking state
  jint* cached_state() const                     { return pd_cached_state();   }
  void  set_cached_state(jint* state)            { pd_set_cached_state(state); }

  // last_Java_sp (Low-order bit is reserved for (not) at_call infomation)
  jint* last_Java_sp() const                     { return (jint*)((int)_last_Java_sp & (~(int)not_at_call_bit)); } 
  void  set_last_Java_sp(jint* sp)               { _last_Java_sp = sp;  pd_set_cached_state(NULL); }
  
  // scope-desc/oopmap lookup 
  void  set_pc_not_at_call_for_frame(jint* sp)   { _not_at_call_sp = sp; set_bits((int&)_last_Java_sp, not_at_call_bit); }
  bool  not_at_call_frame_exists() const         { return ((int)_last_Java_sp) & (int)not_at_call_bit; }
  jint* not_at_call_sp() const                   { return not_at_call_frame_exists() ? (jint*)_not_at_call_sp : NULL; }
  
  // native_last_sp
  address last_native_pc() const                 { return _last_native_pc; }
  void set_last_native_pc(address pc)            { _last_native_pc = pc;   }

  // tells whether the last Java frame is set
  bool has_last_Java_frame() const               { return _last_Java_sp != NULL; }
  
  // Safepoint support
  JavaThreadState thread_state() const           { return _thread_state; }
  void set_thread_state(JavaThreadState s)       { _thread_state=s;      }
  ThreadSafepointState *safepoint_state() const  { return _safepoint_state;  }
  void set_safepoint_state(ThreadSafepointState *state) { _safepoint_state = state; }
  bool is_in_compiled_safepoint()                { return _safepoint_state->type() == ThreadSafepointState::_compiled_safepoint; }
  bool is_terminated()                           { return _terminated; }
  void set_is_terminated(bool flag)              { _terminated = flag; }
  // Support for safepoint timeout (debugging)
  bool has_called_back()                         { return _has_called_back; }
  void set_has_called_back(bool val)             { _has_called_back = val; }
    
  
  // Suspend/resume support.
  //
  // External suspend/resume requests come from JVM_SuspendThread,
  // JVM_ResumeThread, JVM/DI SuspendThread, JVM/DI ResumeThread,
  // JVM/PI SuspendThread, and finally JVM/PI ResumeThread. External
  // suspend requests cause _external_suspend to be set and external
  // resume requests cause _external_suspend to be cleared. External
  // suspend requests can nest and external resume requests only
  // resume if the nesting level is zero.
  //
  // For external requests, JavaThreads in state _thread_in_native
  // are suspended immediately if native_now is true. Otherwise,
  // JavaThreads will suspend themselves when
  // handle_special_runtime_exit_condition() is called.
  //
  // Internal suspend/resume requests come from safepointing and
  // the flat profiler and suspends happen immediately. However,
  // JavaThreads in state _thread_in_native are resumed quickly
  // to avoid deadlocking the VM. Internal suspend requests cause
  // _internal_suspend to be set and internal resume requests cause
  // _internal_suspend to be cleared. Internal suspend requests can
  // nest and internal resume requests only resume if the nesting
  // level is zero.
  //
  // We differentiate between external and internal suspend/resume
  // requests to prevent an internal suspend/resume pair from erasing
  // an external suspend. The following rare scenario was observed:
  //
  // - external suspend request made by JVM/DI or JVM/PI
  // - thread marked with "pending suspend" to let it know that it
  //   should suspend itself on the next transition
  // - actual suspend is blocked by a safepoint operation
  // - internal suspend request made by safepoint synchronization
  // - actual suspend completes; "pending suspend" cleared
  // - internal resume request made by safepoint synchronization
  // - thread is resumed and the external suspend request is lost
  //
  void java_suspend(bool native_now);
  void send_java_suspend(bool native_now);
  long java_suspend_self();
  void java_resume();  
  bool is_external_suspend() const               { return _external_suspend; }
  bool is_internal_suspend() const               { return _internal_suspend; }
  bool is_suspended() const                      { return _is_suspended; }
  // utility methods to see if we are doing some kind of suspension
  bool is_being_any_suspended() const                {
    MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
    return (_is_suspended || _external_suspend || _internal_suspend);
  }
  bool is_being_ext_suspended() const            {
    MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
    return (_is_suspended || _external_suspend);
  }
  bool is_being_int_suspended() const            {
    MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
    return (_is_suspended || _internal_suspend);
  }
  // utility method for interactions between external resume and
  // handle_special_runtime_exit_condition().
  bool suspend_only_special_exit() const         {
    // internal suspend requests do not use
    // handle_special_runtime_exit_condition() code path.
    return (_external_suspend && !_pending_async_exception &&
      !has_pending_exception());
  }
  void set_external_suspend()                    {
    MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
    _external_suspend = true;
    _special_runtime_exit_condition = true;
  }
  void clear_external_suspend()                  {
    if (suspend_only_special_exit()) {
      _special_runtime_exit_condition = false;
    }
    _external_suspend = false;
  }
  void set_internal_suspend()                    {
    MutexLockerEx ml(Interrupt_lock, Mutex::_no_safepoint_check_flag);
    _internal_suspend = true;
  }
  void clear_internal_suspend()                  { _internal_suspend = false; }
  void set_is_suspended()                        { _is_suspended = true; };
  void clear_is_suspended()                      { _is_suspended = false; };

  bool is_handling_implicit_exception() const    { return _is_handling_implicit_exception; }
  void set_is_handling_implicit_exception(bool f){ _is_handling_implicit_exception = f; }

  // java.lang.Object.wait tracking
  bool in_jvm_wait() const                       { return _in_jvm_wait; }

  // Thread.stop support
  void send_thread_stop(oop throwable);
  bool has_special_runtime_exit_condition()      { return _special_runtime_exit_condition;  }
  void clear_special_runtime_exit_condition()    { _special_runtime_exit_condition = false; }
  void handle_special_runtime_exit_condition();

  void set_pending_async_exception(oop e)        { _pending_async_exception = e; _special_runtime_exit_condition = true; }
  
  // Fast-locking support
  bool is_lock_owned(address adr) const;  

  // Speculative locking support
  virtual void set_fast_thread_id() { _fast_thread_id = pd_get_fast_thread_id();  }

  // Accessors for vframe array top
  // The linked list of vframe arrays are sorted on sp. This means when we
  // unpack the head must contain the vframe array to unpack.
  void set_vframe_array_head(vframeArray* value) { _vframe_array_head = value; }
  vframeArray* vframe_array_head() const         { return _vframe_array_head;  }
  void insert_vframe_array(vframeArray* value);
  vframeArray* vframe_array_for(jint* sp);

  // Oop results of vm runtime calls
  oop  vm_result() const                         { return _vm_result; }
  void set_vm_result  (oop x)                    { _vm_result   = x; }

  oop  vm_result_2() const                       { return _vm_result_2; }
  void set_vm_result_2  (oop x)                  { _vm_result_2   = x; }

  // Exception handling for compiled methods
  oop      exception_oop() const                 { return _exception_oop; }
  int      exception_stack_size() const          { return _exception_stack_size; }
  address  exception_pc() const                  { return _exception_pc; }
  address  exception_handler_pc() const          { return _exception_handler_pc; }
  address  last_exception_pc() const             { return _last_exception_pc; }
  
  void set_exception_oop(oop o)                  { _exception_oop = o; }
  void set_exception_pc(address a)               { _exception_pc = a; }
  void set_exception_handler_pc(address a)       { _exception_handler_pc = a; }
  void set_exception_stack_size(int size)        { _exception_stack_size = size; }
  void set_last_exception_pc(address a)          { _last_exception_pc = a; }

  // Stack overflow support
  address stack_yellow_zone_base()               { return (address)(stack_base() - 
                                                   (stack_size() - (stack_red_zone_size() + stack_yellow_zone_size()))); }
  int stack_yellow_zone_size()                   { return StackYellowPages * os::vm_page_size(); }
  address stack_red_zone_base()                  { return (address)(stack_base() - (stack_size() - stack_red_zone_size())); }
  int stack_red_zone_size()                      { return StackRedPages * os::vm_page_size(); }
  bool in_stack_yellow_zone(address a)           { return (a <= stack_yellow_zone_base()) && (a >= stack_red_zone_base()); }
  bool in_stack_red_zone(address a)              { return (a <= stack_red_zone_base()) &&
                                                   (a >= (address)((int)stack_base() - stack_size())); }

  void enable_stack_yellow_zone();
  void disable_stack_yellow_zone();
  void enable_stack_red_zone();
  void disable_stack_red_zone();

  // For assembly stub generation
  static int jni_environment_offset_in_bytes()   { return (int)&((JavaThread*)NULL)->_jni_environment; }
  
  static int last_Java_sp_offset_in_bytes()      { return (int)&((JavaThread*)NULL)->_last_Java_sp; }
  static int vm_result_offset_in_bytes()         { return (int)&((JavaThread*)NULL)->_vm_result; }
  static int vm_result_2_offset_in_bytes()       { return (int)&((JavaThread*)NULL)->_vm_result_2; }
  static int thread_state_offset_in_bytes()      { return (int)&((JavaThread*)NULL)->_thread_state; }  
  static int saved_exception_pc_offset_in_bytes(){ return (int)&((JavaThread*)NULL)->_saved_exception_pc; }
  static int osthread_offset_in_bytes()          { return (int)&((JavaThread*)NULL)->_osthread; }
  static int exception_oop_offset_in_bytes()     {  return (int)&((JavaThread*)NULL)->_exception_oop; }
  static int exception_pc_offset_in_bytes()      { return (int)&((JavaThread*)NULL)->_exception_pc; }
  static int exception_handler_pc_offset_in_bytes() { return (int)&((JavaThread*)NULL)->_exception_handler_pc; }
  static int exception_stack_size_offset_in_bytes() { return (int)&((JavaThread*)NULL)->_exception_stack_size; }
  static int last_native_pc_offset_in_bytes()    { return (int)&((JavaThread*)NULL)->_last_native_pc; }

  static int is_handling_implicit_exception_offset_in_bytes()    { return (int)&((JavaThread*)NULL)->_is_handling_implicit_exception; }

  // Returns the jni environment for this thread
  JNIEnv* jni_environment()                      { return &_jni_environment; }

  static JavaThread* thread_from_jni_environment(JNIEnv* env) {
    JavaThread *thread_from_jni_env = (JavaThread*)((int)env - jni_environment_offset_in_bytes());
    return (thread_from_jni_env->is_terminated()? NULL:thread_from_jni_env);
  }

 private:
  void set_monitor_chunks(MonitorChunk* monitor_chunks) { _monitor_chunks = monitor_chunks; } 
  void set_in_jvm_wait(bool value)               { _in_jvm_wait = value; }

 public:
  MonitorChunk* monitor_chunks() const           { return _monitor_chunks; }
  void add_monitor_chunk(MonitorChunk* chunk);
  void remove_monitor_chunk(MonitorChunk* chunk);

 private:
  void set_monitor_arrays(MonitorArray* monitor_array) { _monitor_arrays = monitor_array; } 
  
 public:
  MonitorArray* monitor_arrays() const           { return _monitor_arrays; }
  void add_monitor_array(MonitorArray* array);
  void remove_monitor_array(MonitorArray* array);
  
 private:
  void set_entry_point(ThreadFunction entry_point) { _entry_point = entry_point; }

 public:
     
  // Frame iteration; calls the function f for all frames on the stack 
  void frames_do(void f(frame*, const RegisterMap*));
  
  // Memory operations
  void oops_do(void f(oop*));

  // Memory management operations
  void gc_epilogue();
  void gc_prologue();

  // Misc. operations
  void print();
  void print_value();
  void verify();
  void verify_vframes()                          CORE_RETURN;
  const char* get_thread_name() const;
  const char* get_threadgroup_name() const;
  const char* get_parent_name() const;

  // Accessing frames
  frame last_frame() { 
    pd_cache_state(!is_definitely_current_thread);
    return cook_last_frame(pd_last_frame()); 
  }
  frame cook_last_frame(frame fr); // adjust any patched pc  
  javaVFrame* last_java_vframe(RegisterMap* reg_map);

  // Returns method at 'depth' java or native frames down the stack  
  // Used for security checks
  klassOop security_get_caller_class(int depth);  

  // Fabricate heavyweight monitors for any lightweight monitors that this thread owns
  void jvmpi_fab_heavy_monitors(bool query, int* index, GrowableArray<ObjectMonitor*>* fab_list);

  // Get monitor (if any) that this thread is trying to enter
  address get_pending_monitor();

  // Print stack trace in external format
  void print_stack();

  // Print stack traces in various internal formats
  void trace_stack()                             PRODUCT_RETURN;
  void trace_stack_from(vframe* start_vf)        PRODUCT_RETURN;
  void trace_frames()                            PRODUCT_RETURN;

  // Returns the number of stack frames on the stack
  int depth() const;

  // Function for testing deoptimization
  void deoptimize();
  void make_zombies();

  void deoptimized_wrt_marked_nmethods();
  bool deoptimized_one_wrt_marked_nmethods();

  // Profiling and safepoint operation (see fprofile.cpp and safepoint.cpp)
 public:
   bool get_top_frame(frame* fr, ExtendedPC* addr = NULL, bool for_profile_only = false);
   bool profile_top_frame(frame* fr, ExtendedPC* addr = NULL);
   bool profile_last_Java_frame(frame* fr = NULL, ExtendedPC* addr = NULL);

  // JVMPI
 public:
  void* jvmpi_data() const                       { return _jvmpi_data; }
  void set_jvmpi_data(void* data)                { _jvmpi_data = data; }

  unsigned int last_sum() const                  { return _last_sum; }
  void set_last_sum(unsigned int s)              { _last_sum = s; }

 private:
   ThreadProfiler* _thread_profiler;
 private:
   friend class FlatProfiler;                    // uses both [gs]et_thread_profiler.
   friend class FlatProfilerTask;                // uses get_thread_profiler.
   friend class ThreadProfilerMark;              // uses get_thread_profiler.
   ThreadProfiler* get_thread_profiler()         { return _thread_profiler; }
   ThreadProfiler* set_thread_profiler(ThreadProfiler* tp) {
     ThreadProfiler* result = _thread_profiler;
     _thread_profiler = tp;
     return result;
   }

#ifndef PRODUCT
 private:
   MethodTracer *_method_tracer;

 public:
   MethodTracer *get_method_tracer()                  { return _method_tracer; }
   void          set_method_tracer(MethodTracer *mt)  { _method_tracer = mt; }
#endif

  // Static operations
 public:
  // Returns the running thread as a JavaThread
  static JavaThread* current();

  // Returns the active Java thread.  Do not use this if you know you are calling
  // from a JavaThread, as it's slower than JavaThread::current.  If called from
  // the VMThread, it also returns the JavaThread that instigated the VMThread's
  // operation.  You may not want that either.
  static JavaThread* active();

 public:
  virtual void run();
  void thread_main_inner();

 private:
  // PRIVILEGED STACK
  PrivilegedElement*  _privileged_stack_top;
  GrowableArray<oop>* _array_for_gc; 
 public:
    
  // Returns the privileged_stack information.
  PrivilegedElement* privileged_stack_top() const       { return _privileged_stack_top; }
  void set_privileged_stack_top(PrivilegedElement *e)   { _privileged_stack_top = e; }
  void register_array_for_gc(GrowableArray<oop>* array) { _array_for_gc = array; }

 public:
  // Thread local information maintained by JVMDI. 
  void set_jvmdi_thread_state(JvmdiThreadState *value)                           { _jvmdi_thread_state = value; }
  JvmdiThreadState *jvmdi_thread_state() const                                   { return _jvmdi_thread_state; }
  void set_jvmdi_get_loaded_classes_closure(JvmdiGetLoadedClassesClosure* value) { _jvmdi_get_loaded_classes_closure = value; }
  JvmdiGetLoadedClassesClosure* get_jvmdi_get_loaded_classes_closure() const     { return _jvmdi_get_loaded_classes_closure; }
 private:
  JvmdiThreadState *_jvmdi_thread_state;
  JvmdiGetLoadedClassesClosure* _jvmdi_get_loaded_classes_closure;

 private:
  static unsigned int _stack_size_at_create;
 
 public:
  static inline unsigned int stack_size_at_create(void) { 
    return _stack_size_at_create; 
  }
  static inline void set_stack_size_at_create(unsigned int value) { 
    _stack_size_at_create = value;
  }
  
  // Machine dependent stuff
  #include "incls/_thread_pd.hpp.incl"

#ifndef CORE
 public:
  void set_blocked_on_compilation(bool value) {
    _blocked_on_compilation = value;
  }

  bool blocked_on_compilation() {
    return _blocked_on_compilation;
  }
 protected:
  bool         _blocked_on_compilation;

  // This state should be only in CompilerThread.  TDT FIX!!!!!

  ciEnv*        _env;
  void*         _compile_data;
  CompileTask*  _task;
  CompileQueue* _queue;
#endif
};

// Inline implementation of JavaThread::current
inline JavaThread* JavaThread::current() {
  Thread* thread = ThreadLocalStorage::thread();
  assert(thread != NULL && thread->is_Java_thread(), "just checking");
  return (JavaThread*)thread;
}


// JVMPIDaemonThread are created when a profiling agent calls jvmpi::create_system_thread
extern "C" {
typedef void (*JVMPIDaemonFunction)(void*);
}

class JVMPIDaemonThread: public JavaThread {  
  friend class VMStructs;
 private:
  JVMPIDaemonFunction _func;
 public:
  JVMPIDaemonThread(ThreadFunction entry_point, JVMPIDaemonFunction f) : JavaThread(entry_point), _func(f) {}
  JVMPIDaemonFunction function() const           { return _func; }
  virtual bool is_jvmpi_daemon_thread() const    { return true;  }
};


#ifndef CORE
// A thread used for Compilation.
class CompilerThread : public JavaThread {
  friend class VMStructs;
 public:
  CompilerThread(CompileQueue* queue);

  bool is_Compiler_thread() const { return true; }

  CompileQueue* queue()                     { return _queue; }        

  // Get/set the thread's compilation environment.
  ciEnv*        env()                       { return _env; }
  void          set_env(ciEnv* env)         { _env = env; }

  // Get/set the thread's current task
  CompileTask*  task()                      { return _task; }
  void          set_task(CompileTask* task) { _task = task; }

  void*         data()                      { return _compile_data; }
  void          set_data(void* data)        { _compile_data = data; }
};
#endif


// The active thread queue. It also keeps track of the current used
// thread priorities.
class Threads: AllStatic {
  friend class VMStructs;
 private:
  static JavaThread* _thread_list;
  static int	     _number_of_threads;
  static int	     _number_of_non_daemon_threads;
  static int         _return_code;
  // List of the priorities of the threads waiting to perform a VM operation. Used for
  // automatically adjusting the priority of the VM thread.
  static int         _priority_queue[MaxPriority + 1];
  // Max priority of any running java thread running (computed at each safepoint)
  static ThreadPriority _highest_running_priority;

 public:
  // Thread management
  static void add(JavaThread* p);
  static void remove(JavaThread* p);
  static bool includes(JavaThread* p);
  static JavaThread* first()                     { return _thread_list; }
  
  // Initializes the vm and creates the vm thread
  static jint create_vm(JavaVMInitArgs* args);
  static void create_vm_init_libraries();
  static bool destroy_vm();

  // Garbage collection
  static void follow_other_roots(void f(oop*));
  static void oops_do(void f(oop*));
  static void convert_hcode_pointers();
  static void restore_hcode_pointers();

  static void gc_epilogue();
  static void gc_prologue();

  // Priority management
  static ThreadPriority lowest_priority();
  static void add_priority   (JavaThread* thread);
  static void remove_priority(JavaThread* thread);
  static void change_priority(JavaThread* thread, ThreadPriority after);

  static ThreadPriority highest_running_priority() { return _highest_running_priority; }
  static void compute_highest_running_priority();

  // Verification
  static void verify();
  static void print(bool print_stacks, bool internal_format);
  static void print_on_error(JavaThread* current);

  // Fabricate heavyweight monitor info for each lightweight monitor.
  static GrowableArray<ObjectMonitor*>* Threads::jvmpi_fab_heavy_monitors();

  // Get Java threads that are waiting to enter a monitor
  static GrowableArray<JavaThread*>* get_pending_threads(int count, address monitor);

  // Get owning Java thread from heavyweight Java monitor
  static JavaThread* owning_thread_from_monitor(ObjectMonitor* mid);

  // Number of threads on the active threads list
  static int number_of_threads()	         { return _number_of_threads; }
  // Number of non-daemon threads on the active threads list
  static int number_of_non_daemon_threads()	 { return _number_of_non_daemon_threads; }

  // Deoptimizes all frames tied to marked nmethods
  static void deoptimized_wrt_marked_nmethods();

};


// thread StateSaver enter/exit abstraction
// this class saves thread state which might be needed by a stack walker
// Note: see usage in os_solaris.cpp and conditionVar_solaris.cpp

class StateSaver VALUE_OBJ_CLASS_SPEC {
 private:
  Thread *_thread;
  jint   *_state;
 public:
  StateSaver(Thread *thread) : _thread(thread)   { _state = _thread->save_state_before_block(); }
  ~StateSaver()                                  { _thread->restore_state_after_block(_state);  }
  Thread* current()                              { return _thread; }
};


// utility class for setting _in_jvm_wait flag in JavaThread

class ThreadWaitTransition: public StackObj {
 private:
  JavaThread* _thread;
 public:
  ThreadWaitTransition(JavaThread* thread) : _thread(thread) {
    assert(!_thread->in_jvm_wait(), "sanity check");
    _thread->set_in_jvm_wait(true);
  }
  ~ThreadWaitTransition() {
    assert(_thread->in_jvm_wait(), "sanity check");
    _thread->set_in_jvm_wait(false);
  }
};

// Defines the shared interfaces whose implementations vary across os platforms
// Suspend a thread by one level. This method assumes that consequtive threads nest.
// Returns a value indicating the new suspension level.
// The nesting is tracked the same for all platforms, regardless
// of OS support to allow debugging and assertions.
class Suspend_Callback : public StackObj {
 public:
  virtual long execute(Thread* thread) = 0;
#ifdef ASSERT
  virtual bool is_self_suspend_cb() = 0;
#endif /* ASSERT */

  // Must call this to mark thread, after it is suspended.
  // For self-suspension, this must be called before blocking
  // on the condition variable.
  static void mark_suspended(OSThread* osthread) {
    assert(osthread->suspend_depth() == 0, "incorrect suspend depth"
);
    assert(osthread->state() != SUSPENDED, "incorrect thread state")
;

    osthread->clear_self_suspended();
    osthread->set_state_before_suspend(osthread->state());
    osthread->set_state(SUSPENDED);
    osthread->set_suspend_depth(osthread->suspend_depth() + 1);
  }
};

class Resume_Callback : public StackObj {
 public:
  virtual long execute(Thread* thread) = 0;
};


// this is for the normal thr_suspend case for non-self suspension
// The pd_suspend_thread implementations are platform dependent
class SuspendThread_Callback : public Suspend_Callback {
 bool _fence;
 public:
  SuspendThread_Callback()           { _fence = true; }
  SuspendThread_Callback(bool fence) { _fence = fence; }
  long execute(Thread* thread){
    OSThread *osthread = thread->osthread();
    long ret = os::pd_suspend_thread(thread, _fence);
    if (!osthread->is_self_suspended()) {
      mark_suspended(osthread);
    } else {
      osthread->set_suspend_depth(osthread->suspend_depth() + 1);
    }
    osthread->set_physically_suspended();
    return ret;
  }
#ifdef ASSERT
  bool is_self_suspend_cb()          { return false; }
#endif /* ASSERT */
};

// resume_thread
// The pd_resume_thread implementations are platform dependent

class ResumeThread_Callback : public Resume_Callback {
 public:
  long execute(Thread* thread){
    OSThread *osthread = thread->osthread();
    osthread->clear_physically_suspended();
    long ret = os::pd_resume_thread(thread);
    return ret;
  }
};

// thread_suspend_self
// we need to be able to suspend ourself while at the same (atomic) time
// giving up the Interrupt_lock -- we do this by using the
// Interrupt_lock to implement a suspend_self
class SuspendThreadSelf_Callback : public Suspend_Callback {
 public:
  long execute(Thread* thread) { 
    OSThread *osthread = thread->osthread();
    mark_suspended(osthread);

    if (!os::pd_pre_self_suspend(osthread)) {
      os::pd_post_self_suspend(osthread);
      return 0;
    }

    osthread->set_self_suspended();
    
    while(osthread->is_self_suspended()) {
      Interrupt_lock->wait(Mutex::_no_safepoint_check_flag);
    }

    os::pd_post_self_suspend(osthread);

    return 0;
  }
#ifdef ASSERT
  bool is_self_suspend_cb()          { return true; }
#endif /* ASSERT */
};


