/*
 * @(#)threadLocalEden.cpp	1.8 99/11/17 15:26:25
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// Thread-Local Edens support

# include "incls/_precompiled.incl"
# include "incls/_threadLocalEden.cpp.incl"

// static member initialization
int ThreadLocalEden::_log2_fragmentation_ratio = 0;
int ThreadLocalEden::_max_size = 0;


// Fills the current tle with a dummy filler array to create
// an illusion of a contiguous Eden and clears the tle info
void ThreadLocalEden::reset() {
     oop* tle_end = end();

     if(tle_end) {
       typeArrayOop filler_oop = (typeArrayOop) top();
       oop* tle_top = top();
       
       filler_oop->set_mark();
       filler_oop->set_klass(Universe::intArrayKlassObj());
       int aligned_header_size = align_object_size(typeArrayOopDesc::header_size(T_INT));
       tle_end = tle_end + aligned_header_size;       // real tle end
       int array_length = tle_end - tle_top - typeArrayOopDesc::header_size(T_INT);
       filler_oop->set_length(array_length);

#ifdef ASSERT
       oop* elt_oops = (oop *) (((oop *)filler_oop) + typeArrayOopDesc::header_size(T_INT));
       Memory::set_oops(elt_oops, array_length, (oop)0xDEAFBABE);
#endif
       
       assert(array_length >= 0, "Incorrect filler length in clear_allocate_local");
       if(PrintTLE) {
	 static int total_used;
	 static int total_unused;
	 int used, unused;
	 
	 used = top() - start();
	 unused = aligned_header_size + array_length;
	 total_used += used;
	 total_unused += unused;
	 tty->print_cr("[TLE reset: calling thread# %d  unused: %d  size: %d  Total fragmentation %f]", 
		       Thread::current()->osthread()->thread_id(), unused * wordSize, size() * wordSize, 
		       (float) total_unused / (float) (total_used + total_unused));
       }
       
       set_top(NULL);
       set_end(NULL);
       set_start(NULL);
       reset_original_size();
       // _next, _prev, and _has_terminated are irrelevant and cleared on re-initialization
       
     }

     if (VerifyBeforeScavenge && ((Scavenge::invoke_count() + 1) >= VerifyScavengeStartAt)) {
       verify();
     }

     assert(start() == NULL && end() == NULL && top() == NULL,
	    "TLE must be reset");
}

// Adjusts the size of TLEs as a function of the number of threads
void ThreadLocalEden::adjust_max_size() {
  assert(ResizeTLE, "Must not call if ResizeTLE=false");
  int eden_size = Universe::new_gen()->eden()->capacity();
  int new_max_tle_size = align_object_size(eden_size / (wordSize * MaxTLERatio));
  // for highly threaded applications, allow at least a certain fraction of threads have max tle size
  int tle_size_per_thread = align_object_size(eden_size / (wordSize * Threads::number_of_non_daemon_threads() / TLEThreadRatio + 1));
  new_max_tle_size = min(new_max_tle_size, tle_size_per_thread);
  if(PrintTLE) {
    int old_max_tle_size = max_size();
    tty->print_cr("[TLE adjusting max %dK -> %dK]", old_max_tle_size*wordSize, new_max_tle_size*wordSize);
  }
  set_max_size(new_max_tle_size);
}

void ThreadLocalEden::print() {
  tty->print_cr(" total %dK, used %dK", size() * wordSize / K, (top() - start()) * wordSize / K );
  tty->print_cr(" [%#x, %#x, %#x)", start(), top(), end());
}

void ThreadLocalEden::print_current() {
  tty->print("Thread #%d    ", Thread::current()->osthread()->thread_id());
  Thread::current()->tle().print();
}

void ThreadLocalEden::verify() {
  oop* p = start();
  oop* t = top();
  oop* prev_p = NULL;
  while (p < t) {
    oop(p)->verify();
    prev_p = p;
    p += oop(p)->size();
  }
  guarantee(p == top(), "end of last object must match end of space");
}


// TLEQueue

// statics
ThreadLocalEden TLEQueue::_dummy_tle;
unsigned TLEQueue::_size = 0;
jint TLEQueue::_lock = 0;

void TLEQueue::init() {
  // an empty queue always has 1 element with next and prev fireld
  // pointing to itself
  head()->set_next(head());
  head()->set_prev(head());
  _size = 0;
}


void TLEQueue::unpark(ThreadLocalEden* tle) {
  // If the tle is still on the queue, just dequeue 
  // if it is not, dequeue some other

  acquire();
  
  if(tle->top() != NULL) {
    // still on the queue
    assert(tle->prev() && tle->next(), "corrupt tle");
    remove_unguarded(tle);
  }
  else {
    // dequeue some other tle 
    ThreadLocalEden* stolen_tle = dequeue_unguarded();
    if(stolen_tle != NULL) {
      tle->initialize(stolen_tle->start(), stolen_tle->top(), stolen_tle->end(), stolen_tle->size());
      if(stolen_tle->has_terminated()) {
	// the host thread has terminated - we are responsible for destroying this copy of the tle descriptor
	delete stolen_tle;
      }
      else {
	stolen_tle->clear();  // signal that the tle has been stolen
      }
      assert(tle->top() != NULL, "empty tle" );
    }
  }
  
  if(VerifyTLE) verify();

  release();
}

// Invalidates the entire queue (usually prior to GC)
void TLEQueue::reset_at_safepoint() {
  assert(ParkTLE, "should not rreach here");
  if (size() == 0) return;

  ThreadLocalEden* tle;
  while((tle = TLEQueue::dequeue_unguarded()) != NULL) {
    tle->reset();
    if(tle->has_terminated()) {
      delete tle;             // the host thread has terminated - we need to delete the descriptor
    }
  }

  if(VerifyTLE) verify();
}

// Walk the queue and make sure that enties look ok
void TLEQueue::verify() {
  ThreadLocalEden* tle = head()->next();
  unsigned int s = 0;

  while(tle != head()) {
    assert((tle->top() != NULL) && tle->invariants(), "corrupt tle");
    s++;
    tle = tle->next();
  }
  assert(s == TLEQueue::size(), "TLE: wrong queue size");

  // now walk backwards 
  tle = head()->prev();
  s = 0;
  while(tle != head()) {
    assert((tle->top() != NULL) && tle->invariants(), "corrupt tle");
    s++;
    tle = tle->prev();
  }
  assert(s == TLEQueue::size(), "TLE: wrong queue size");
}



