/*
 * @(#)threadLocalEden.hpp	1.7 01/02/09 14:48:02
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

class TLEQueue;
class ThreadLocalEden;

// TLEQueue: a Concurrent Queue of TLE descriptors
//           implemented as a doubly linked circular queue
//           protected by an embedded spinlock (all operations are non-blocking)
class TLEQueue : public AllStatic {
  friend class VMStructs;
private:
  static ThreadLocalEden  _dummy_tle;             // does not count
  static unsigned int     _size;                  // current queue size
  static jint             _lock;                  // a spinlock protecting the queue
  static jint* lock_addr()                        { return &_lock; }
  static ThreadLocalEden* head()                  { return &_dummy_tle; }
  static void enqueue(ThreadLocalEden* tle);     
  static void enqueue_unguarded(ThreadLocalEden* tle);    
  static ThreadLocalEden* dequeue();
  static ThreadLocalEden* dequeue_unguarded();
  static void remove_unguarded(ThreadLocalEden* tle);
  static void acquire();                          // acquire()/relases() ensure thread-safety
  static void release();

public:
  static void init();
  static const unsigned int size()                { return _size; }
  static const bool is_empty()                    { return _size == 0; }  // just a hint, no locking
  static void park(ThreadLocalEden* tle);         // make tle available to other threads by parking (usually before blocking) 
  static void unpark(ThreadLocalEden *tle);       // attempt to re-use a tle parked on the queue
  static void reset_at_safepoint();               // invalidates the entire queue (usually before GC)

  static void verify();
};


// ThreadLocalEden: a descriptor for thread-local storage used by the threads for allocation
//            It is thread-private at any time, but maybe multiplexed over time across multiple
//            threads. The park()/unpark() pair is used to make it avaiable for such multiplexing.
class ThreadLocalEden: public CHeapObj {
  friend class VMStructs;
private:
  oop* _start;
  oop* _top;
  oop* _end;
  int  _size;                                    // full size (including padding)
  bool _has_terminated;                          // true if the host thread has terminated
  ThreadLocalEden* _next;                        // used when park on TLEQueue
  ThreadLocalEden* _prev;                        // used when park on TLEQueue
       
  static int _log2_fragmentation_ratio;          // log2 of the portion of (alive) TLE that can remain unused
  static int _max_size;                          // max allowed tle size in words (used with ResizeTLE)  


public:
  ThreadLocalEden() : _start(NULL), _top(NULL), _end(NULL), _next(NULL), _prev(NULL), _has_terminated(false), _size(0) {}
  ThreadLocalEden(oop* start, oop* top, oop* end, int size) : _start(start), _top(top), _end(end), _next(NULL), _prev(NULL), _has_terminated(false), _size(size) {}
  oop* start() const                             { return _start; }
  void set_start(oop* start)                     { _start = start; }
  oop* end() const                               { return _end; }
  void set_end(oop* end)                         { _end = end; }
  oop* top() const                               { return _top; }
  void set_top(oop* top)                         { _top = top; }
  int  size() const                              { return _size; }
  void set_size(int size)                        { _size = size; }
  bool has_terminated()                          { return _has_terminated; }
  void set_has_terminated(bool b)                { _has_terminated = b; }
  ThreadLocalEden* next() const                  { return _next; }
  void set_next(ThreadLocalEden* next)           { _next = next; }
  ThreadLocalEden* prev() const                  { return _prev; }
  void set_prev(ThreadLocalEden* prev)           { _prev = prev; }
  
  int  available_size()                          { return _end - _top; } // in words

  static int  log2_fragmentation_ratio()         { return _log2_fragmentation_ratio; }
  static void set_log2_fragmentation_ratio(int r){ _log2_fragmentation_ratio = r; }
  static void initialize_fragmentation_ratio()   { _log2_fragmentation_ratio = exact_log2(TLEFragmentationRatio); }
  static int  max_size()                         { return _max_size; }
  static void set_max_size(int size)             { _max_size = size; }
  static void adjust_max_size();

  void reset_original_size()                     { _size = TLESize / wordSize; }
  void initialize(oop* start, oop* top, oop* end, int size) { _start = start; _top = top; _end = end; _size = size; _next = NULL, _prev = NULL; _has_terminated = false; }
  void reset();
  ThreadLocalEden* clone()                       { ThreadLocalEden* tle = new ThreadLocalEden(_start, _top, _end, _size);
                                                   tle->set_has_terminated(true); // no currently associated thread
                                                   return tle; }
  bool invariants()                              { return (top() >= start()) && (top() <= end()) && (size() >= 0); }

  void ThreadLocalEden::park()                   { TLEQueue::park(this); }
  void ThreadLocalEden::unpark()                 { TLEQueue::unpark(this); }

  void print();
  static void print_current();

  // Code generation support
  static int start_offset_in_bytes()             { return (int)&((ThreadLocalEden *)NULL)->_start; }
  static int end_offset_in_bytes()               { return (int)&((ThreadLocalEden *)NULL)->_end; }
  static int top_offset_in_bytes()               { return (int)&((ThreadLocalEden *)NULL)->_top; }
  static int size_offset_in_bytes()              { return (int)&((ThreadLocalEden *)NULL)->_size; }  

  void ThreadLocalEden::clear() {
    set_start(NULL);
    set_end(NULL);
    set_top(NULL);
    set_next(NULL);
    set_prev(NULL);
    set_has_terminated(false);
    reset_original_size();
  }

  int compute_new_size(int allocation_size);
  void verify();
};



