/*
 * @(#)threadLocalEden.inline.hpp	1.2 99/09/09 16:46:18
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


// thread-safe version: equeue at the "front" of the queue
inline void TLEQueue::enqueue(ThreadLocalEden* tle) {
  assert(tle->top() != NULL, "corrupt tle");
  
  acquire();
  enqueue_unguarded(tle);
  release();
}

// unguarded version: enqueue at the front of the queue
inline void TLEQueue::enqueue_unguarded(ThreadLocalEden* tle) {
  assert(tle->top() != NULL, "corrupt tle");
  
  tle->set_next(head()->next());
  tle->set_prev(head());

  head()->next()->set_prev(tle);
  head()->set_next(tle);

  _size++;

  if(VerifyTLE) {
    verify();
  }
  
  assert(size >= 0, "TLEQueue: negative size");
}


// thread-safe version: dequeue from the front
inline ThreadLocalEden* TLEQueue::dequeue() {
  acquire();
  ThreadLocalEden* tle = dequeue_unguarded();
  release();
  return tle;
}

// unguarded version: dequeue from the front
inline ThreadLocalEden* TLEQueue::dequeue_unguarded() {
  if(head()->next() == head()) {
    return NULL;
  }

  ThreadLocalEden* tle = head()->next();
  head()->set_next(tle->next());
  tle->next()->set_prev(head());

  _size--;
  
  tle->set_next(NULL);
  tle->set_prev(NULL);
  assert(tle->top() != NULL, "empty tle" );

  if(VerifyTLE) {
    verify();
    tle->verify();
  }
  
  assert(size >= 0, "TLEQueue: negative size");
  return tle;
}

// removes tle from (the middle of) the queue 
inline void TLEQueue::remove_unguarded(ThreadLocalEden* tle) {
  assert(tle->next() && tle->prev(), "tle corruption");

  tle->prev()->set_next(tle->next());
  tle->next()->set_prev(tle->prev());
  
  // next and prev are zeroed by the caller
  _size--;

  if(VerifyTLE) {
    verify();
    tle->verify();
  }

  assert(size >= 0, "TLEQueue: negative size");

  return;
}

// "park" the tle by adding it to the TLEQueue
inline void TLEQueue::park(ThreadLocalEden* tle) {
  assert(UseTLE && ParkTLE, "should not reach here");
  if( tle->top() == NULL ) return;  // already empty

  debug_only(Memory::set_oops(tle->top(), tle->end() - tle->top(), (oop) 0xFEEDBABE));

  enqueue(tle); 
}



// note: allocation size includes all necessary padding
inline int ThreadLocalEden::compute_new_size(int allocation_size) {
  int new_tle_size; 
  
  if(ResizeTLE) {
    new_tle_size = size() << 1;                                     // double the old size
    new_tle_size = min(new_tle_size, ThreadLocalEden::max_size());  // bound the growth
  }
  else {
    new_tle_size = TLESize / wordSize;
  }
  
  int available_shared_size = Universe::new_gen()->eden()->end() - Universe::new_gen()->eden()->top();  // just a hint
  // The "last" tle may be smaller to reduce fragmentation
  if((available_shared_size < new_tle_size) && (available_shared_size >= allocation_size)) {
    new_tle_size = available_shared_size;
  }
  
  new_tle_size = max(new_tle_size, allocation_size);
  return new_tle_size;
}
