/*
 * @(#)vframe_hp.cpp	1.93 99/12/01 11:06:36
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_vframe_hp.cpp.incl"

// ------------- deoptimizedVFrame --------------

deoptimizedVFrame::deoptimizedVFrame(const frame* fr, const RegisterMap* reg_map, JavaThread* thread)
: javaVFrame(fr, reg_map, thread) {
  assert(fr, "valid frame must be present");
  _array = thread->vframe_array_for(fr->sp());
  _index = vframeArray::first_index();
}

deoptimizedVFrame::deoptimizedVFrame(const frame* fr, const RegisterMap* reg_map, JavaThread* thread, int index)
: javaVFrame(fr, reg_map, thread) {
  assert(fr, "valid frame must be present");
  _array = thread->vframe_array_for(fr->sp());
  _index = index;
}

deoptimizedVFrame::deoptimizedVFrame(vframeArray* array, int index) 
  : javaVFrame(NULL, register_map(), array->owner_thread()) {
  _array = array;
  _index = index;
}

deoptimizedVFrame::deoptimizedVFrame(vframeArray* array) 
  : javaVFrame(NULL, register_map(), array->owner_thread()) {
  _array = array;
  _index = vframeArray::first_index();
}

bool deoptimizedVFrame::is_top() const {
  int sender_index = array()->sender_index_for(index());
  return sender_index >= array()->length();
}

GrowableArray<StackValue*>* deoptimizedVFrame::locals() const {  
  int idx    = index();
  int length = array()->locals_size_for(idx);  
  assert(array()->method_for(idx)->max_locals() == length, "sanity check");

  GrowableArray<StackValue*>* result = new GrowableArray<StackValue*>(length);
  for(int i = 0; i < length; i++) {    
    StackValue *sv;
    if (array()->locals_is_obj_for(idx, i)) {
      oop value = array()->locals_obj_for(idx, i);
      Handle h(value);
      sv = new StackValue(h);
    } else {
      jint value = array()->locals_int_for(idx, i);                   
      sv = new StackValue(value);
    }              
    assert(sv != NULL, "sanity check");
    result->push(sv);
  }
  return result;
}


GrowableArray<StackValue*>* deoptimizedVFrame::expressions() const {  
  int idx    = index();
  int length = array()->expression_stack_size_for(idx);
  int nof_locals = array()->locals_size_for(idx);
  assert(array()->method_for(idx)->max_locals() == nof_locals, "sanity check");

  GrowableArray<StackValue*>* result = new GrowableArray<StackValue*>(length);

  for(int i = 0; i < length; i++) {    
    StackValue *sv;
    if (array()->expression_stack_is_obj_for(idx, i)) {
      oop value = array()->expression_stack_obj_for(idx, i);
      Handle h(value);
      sv = new StackValue(h);
    } else {
      jint value = array()->expression_stack_int_for(idx, i);
      sv = new StackValue(value);
    }              
    assert(sv != NULL, "sanity check");
    result->push(sv);
  }
  return result;
}


methodOop deoptimizedVFrame::method() const {
  return array()->method_for(index());
}

int deoptimizedVFrame::raw_bci() const {
  return array()->bci_for(index());
}

int deoptimizedVFrame::bci() const {
  int raw = raw_bci();
  return raw == SynchronizationEntryBCI ? 0 : raw;
}

GrowableArray<MonitorInfo*>* deoptimizedVFrame::monitors() const {
  MonitorArray* arr = array()->monitor_array_for(index());
  if (arr == NULL) return new GrowableArray<MonitorInfo*>(0);
  return arr->as_monitor_list();
}

vframe* deoptimizedVFrame::sender() const {
  int sender_index = array()->sender_index_for(index());
  if (sender_index < array()->length()) {
    return new deoptimizedVFrame(array(), sender_index);
  }
  if (fr().sp() == NULL) return NULL;
  return vframe::sender();
}

int deoptimizedVFrame::on_stack_size(int callee_parameters, int callee_locals) const {
  assert(method()->max_locals() == array()->locals_size_for(index()), "just checking");
  int locks = monitors()->length();
  int temps = array()->expression_stack_size_for(index());
  return Interpreter::size_activation(method(), temps + callee_parameters, locks, callee_parameters, callee_locals);
}

int unpack_counter = 0;

frame deoptimizedVFrame::unpack_vframe_on_stack(frame caller_frame, int callee_parameters, int callee_locals, bool is_top_frame, int exec_mode) {
  int frame_size  = on_stack_size(callee_parameters, callee_locals);
  int locals_size = array()->locals_size_for(index());

  // Look at bci and decide on bcp and continuation pc
  address bcp;
  address pc;
  if (raw_bci() == SynchronizationEntryBCI) {
    // We are deoptimizing while hanging in prologue code for synchronized method
    bcp = method()->bcp_from(0); // first byte code
    pc  = Interpreter::deopt_entry(vtos, 0); // step = 0 since we don't skip current bytecode
  } else {
    bcp = method()->bcp_from(bci());
    pc  = Interpreter::continuation_for(method(), bcp, callee_parameters, is_top_frame);
  }

  // Monitorenter is special! Since the bytecode must *never* be reexecuted - not even in case of
  // exceptions. In more detail:
  //   - deoptimization can only happen at the end of the slow case (hence, this is not the null check)
  //   - the lock is already held.
  // So, in case we got:
  //   - no exception: Everything is fine, we will execute the following instruction
  //   - an (async.) exception: Normal behavior is to reexecute the bytecode that caused the exception. To
  //     avoid this, we explicit increment the 'bcp' in the interpreter stack, so exception handling will
  //     think it was the following bytecode that got the async. exception. 
  if (*bcp == Bytecodes::_monitorenter && exec_mode == Deoptimization::Unpack_exception) {    
    // Advance bcp to the one following the _monitorenter
    assert(Bytecodes::length_for(Bytecodes::_monitorenter) == 1, "well, it is hopefully not going to change");
    assert(is_top_frame, "a _monitorenter must be a top frame");
    bcp++;
  }

  // Setup the interpreter frame
  GrowableArray<MonitorInfo*>* list = monitors();

  assert (!method()->is_synchronized() || list->length() > 0, "synchronized methods must have monitors");
  int temps = array()->expression_stack_size_for(index());

  frame interpreter_frame = Interpreter::layout_activation(caller_frame, method(), temps + callee_parameters, list->length(), callee_parameters, callee_locals);

  interpreter_frame.set_link(caller_frame.fp());       //point back to caller
  interpreter_frame.set_pc(pc);                        //record intended pc
  interpreter_frame.set_sender_pc(caller_frame.pc());  //write real frame state

  BasicObjectLock* top = interpreter_frame.interpreter_frame_monitor_begin();
  for (int index = 0; index < list->length(); index++) {
    top = interpreter_frame.previous_monitor_in_interpreter_frame(top);
    MonitorInfo* src = list->at(index);
    top->set_obj(src->owner());
    src->lock()->move_to(src->owner(), top->lock());
  }
  interpreter_frame.interpreter_frame_set_bcx((jint)bcp); // cannot use bcp because frame is not initialized yet

  // Unpack expression stack  
  GrowableArray<StackValue *> *exprs = expressions();
  int i;
  for(i = 0; i < exprs->length(); i++) {
    StackValue *value = exprs->at(i);
    jint*       addr  = &interpreter_frame.interpreter_frame_expression_stack_at(i);
    switch(value->type()) {
      case T_OBJECT: 
	*(oop*)addr = value->get_obj()();
        break;
      case T_INT:
        *addr = value->get_int();
        break;
      default:
        ShouldNotReachHere();
    }
  }

  // Unpack the locals
  GrowableArray<StackValue *> *locs = locals();
  for(i = 0; i < locs->length(); i++) {
    StackValue *value = locs->at(i);
    jint*       addr  = &interpreter_frame.interpreter_frame_local_at(i);
    switch(value->type()) {      
      case T_OBJECT:
        *(oop*)addr = value->get_obj()();
        break;
      case T_INT:
        *addr = value->get_int();
        break;
      case T_CONFLICT:
        *addr = 0xbadbabe;
        break;
      default:
        ShouldNotReachHere();
    }
  }

#ifndef PRODUCT
  if (TraceDeoptimization && Verbose) {
    tty->cr();
    tty->print_cr("[%d UNPACKING]", ++unpack_counter);
    interpreter_frame.print_on(tty);
    RegisterMap map(thread());
    vframe* f = vframe::new_vframe(&interpreter_frame, &map, thread());
    f->print();
    interpreter_frame.interpreter_frame_print_on(tty);
   
    tty->print_cr("locals size     %d", locs->length());
    tty->print_cr("expression size %d", exprs->length());
    
    method()->print_value();
    tty->cr();
    // method()->print_codes();
  } else if (TraceDeoptimization) {
    tty->print("      >> ");
    method()->print_value();
    tty->cr();
  }
#endif

  return interpreter_frame;
}

// ------------- compiledVFrame --------------

GrowableArray<StackValue*>* compiledVFrame::locals() const {  
  GrowableArray<ScopeValue*>*  scv_list = scope()->locals();
  if (scv_list == NULL) 
    return new GrowableArray<StackValue*>(0);

  // scv_list is the list of ScopeValues describing the JVM stack state.
  // There is one scv_list entry for every JVM stack state in use.
  int length = scv_list->length();
  GrowableArray<StackValue*>* result = new GrowableArray<StackValue*>(length);
  for( int i = 0; i < length; i++ )
    result->push( create_stack_value(scv_list->at(i)) );

  return result;
}

GrowableArray<StackValue*>* compiledVFrame::expressions() const {
  GrowableArray<ScopeValue*>*  scv_list = scope()->expressions();
  if (scv_list == NULL) 
    return new GrowableArray<StackValue*>(0);

  // scv_list is the list of ScopeValues describing the JVM stack state.
  // There is one scv_list entry for every JVM stack state in use.
  int length = scv_list->length();
  GrowableArray<StackValue*>* result = new GrowableArray<StackValue*>(length);
  for( int i = 0; i < length; i++ )
    result->push( create_stack_value(scv_list->at(i)) );

  return result;
}


StackValue *compiledVFrame::create_stack_value(ScopeValue *sv) const {
  if (sv->is_location()) {
    // Stack or register value
    Location loc = ((LocationValue *)sv)->location();
    if (loc.is_illegal()) return new StackValue();

#ifdef SPARC
    // %%%%% Callee-save floats will NOT be working on a Sparc until we
    // handle the case of a 2 floats in a single double register.
    assert( !(loc.is_register() && loc.holds_float()), "Sparc does not handle callee-save floats yet" );
#endif

    // First find address of value
    address value_addr = loc.is_register()
      // Value was in a callee-save register
      ? register_map()->location(VMReg::Name(loc.register_number()))
      // Else value was directly saved on the stack
      : ((address)_fr.sp()) + loc.stack_offset();

    // Then package it right depending on type
    if (loc.holds_float()) {    // Holds a float in a double register?
      // The callee has no clue whether the register holds a float,
      // double or is unused.  He always saves a double.  Here we know
      // a double was saved, but we only want a float back.  Narrow the
      // saved double to the float that the JVM wants.
      assert( loc.is_register(), "floats always saved to stack in 1 word" );
      jfloat value = (jfloat) *(jdouble*) value_addr;
      return new StackValue(* (jint*) &value);

    } else if( loc.holds_oop() ) {  // Holds an oop?
      Handle h(*(oop *)value_addr); // Wrap a handle around the oop
      return new StackValue(h);
    } else {                        // Else just regular bits
      return new StackValue(*(jint *)value_addr);
    }

  } else if (sv->is_constant_int()) {
    // contant int    
    jint value =  ((ConstantIntValue *)sv)->value();
    return new StackValue(value);
  } else if (sv->is_constant_oop()) {
    // constant oop        
    return new StackValue(((ConstantOopReadValue *)sv)->value());
  }
  // Unknown ScopeValue type
  ShouldNotReachHere();    
  return new StackValue((jint) 0);   // dummy  
}


BasicLock* compiledVFrame::resolve_monitor_lock(Location location) const {
  assert(location.is_stack(), "for now we only look at the stack");
  int word_offset = location.stack_offset() / wordSize;
  // (stack picture)
  // high: [     ]  word_offset + 1
  // low   [     ]  word_offset
  //       
  // sp->  [     ]  0
  // the word_offset is the distance from the stack pointer to the lowest address
  return (BasicLock*) (fr().sp() + word_offset);
}

GrowableArray<MonitorInfo*>* compiledVFrame::monitors() const {
  GrowableArray<MonitorValue*>* monitors = scope()->monitors();
  if (monitors == NULL) {
    return new GrowableArray<MonitorInfo*>(0);
  }
  GrowableArray<MonitorInfo*>* result = new GrowableArray<MonitorInfo*>(monitors->length());
  for (int index = 0; index < monitors->length(); index++) {
    MonitorValue* mv = monitors->at(index);
    StackValue *owner_sv = create_stack_value(mv->owner()); // it is an oop
    result->push(new MonitorInfo(owner_sv->get_obj()(), resolve_monitor_lock(mv->basic_lock())));
  }
  return result;
}

compiledVFrame::compiledVFrame(const frame* fr, const RegisterMap* reg_map, JavaThread* thread, ScopeDesc* scope)
: javaVFrame(fr, reg_map, thread) {
  _scope  = scope;
  guarantee(_scope != NULL, "scope must be present");
}

bool compiledVFrame::is_top() const {
  // FIX IT: Remove this when new native stubs are in place
  if (scope() == NULL) return true;
  return scope()->is_top();
}

nmethod* compiledVFrame::code() const {
  return CodeCache::find_nmethod(_fr.pc());
}

methodOop compiledVFrame::method() const {
  return scope()->method()();
}

int compiledVFrame::bci() const {
  int raw = raw_bci();
  return raw == SynchronizationEntryBCI ? 0 : raw;
}

int compiledVFrame::raw_bci() const {
  return scope()->bci();
}

vframe* compiledVFrame::sender() const {
  assert( scope(), "When new stub generator is in place, then scope() can never be NULL" );
  const frame f = fr();
  return scope()->is_top() 
    ? vframe::sender()
    : new compiledVFrame(&f, register_map(), thread(), scope()->sender());
}

#ifndef PRODUCT

void compiledVFrame::verify() const {
}

#endif
