/*
 * @(#)virtualspace.cpp	1.21 01/02/09 14:48:03
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

#include "incls/_precompiled.incl"
#include "incls/_virtualspace.cpp.incl"


// ReservedSpace
ReservedSpace::ReservedSpace(size_t size, size_t forced_base_alignment, bool ISM) {

  assert(size % os::vm_page_size() == 0, "size not page aligned");
  assert((forced_base_alignment % os::vm_page_size()) == 0, "size not page aligned");
  _base = NULL;
  _size = 0;

  char* base;
  // If ISM constructor is used, make sure UseISM or UsePermISM is on
  if (ISM) {
    assert((UseISM || UsePermISM) , "ISM constructor used although UseISM or UsePermISM is not on.");

    // ISM should always return a 4MB aligned address
    base = os::reserve_memory_special(size);
    
    if (base == NULL) return;
    
    // Check alignment constraints
    if (forced_base_alignment > 0) {
      assert((juint) base % forced_base_alignment == 0, 
	     "ISM returned a non-aligned address");
    }
  } else {
    // Optimistically assume that the OSes returns an aligned base pointer.
    // When reserving a large address range, most OSes seem to align to at 
    // least 64K.
    base = os::reserve_memory(size);

    if (base == NULL) return;
    // Check alignment constraints
    if (forced_base_alignment > 0 && ((juint) base % forced_base_alignment != 0)) {
      // Base not aligned, retry
      if (!os::release_memory(base, size)) fatal("os::release_memory failed");
      // Reserve size large enough to do manual alignment
      size_t extra_size = size + forced_base_alignment;
      char* extra_base = os::reserve_memory(extra_size);
      if (extra_base == NULL) return;
      // Do manual alignement
      base = (char*) align_size_up((juint) extra_base, forced_base_alignment);
      assert(base >= extra_base, "just checking");
      // Release unused areas
      size_t unused_bottom_size = base - extra_base;
      size_t unused_top_size = extra_size - size - unused_bottom_size;
      assert(unused_bottom_size % os::vm_page_size() == 0, "size not page aligned");
      assert(unused_top_size % os::vm_page_size() == 0, "size not page aligned");
      if (unused_bottom_size > 0) {
	os::release_memory(extra_base, unused_bottom_size);
      }
      if (unused_top_size > 0) {
	os::release_memory(base + size, unused_top_size);
      }
    }
  }
  // Done
  _base = base;
  _size = size;
  assert(markOopDesc::encode_pointer_as_mark(_base)->decode_pointer() == _base,
	 "area must be distinguisable from marks for mark-sweep");
  assert(markOopDesc::encode_pointer_as_mark(&_base[size])->decode_pointer() == &_base[size],
	 "area must be distinguisable from marks for mark-sweep");
}


ReservedSpace::ReservedSpace(char* base, size_t size) {
  assert((size % os::vm_page_size()) == 0, "size not page aligned");
  _base = base;
  _size = size;
}


ReservedSpace ReservedSpace::first_part(size_t partition_size) {
  assert(partition_size <= size(), "partition failed");
  ReservedSpace result(base(), partition_size);
  return result;
}


ReservedSpace ReservedSpace::last_part(size_t partition_size) {
  assert(partition_size <= size(), "partition failed");
  ReservedSpace result(base() + partition_size, size() - partition_size);
  return result;
}


size_t ReservedSpace::page_align_size_up(size_t size) {
  return align_size_up(size, os::vm_page_size());
}


size_t ReservedSpace::page_align_size_down(size_t size) {
  return align_size_down(size, os::vm_page_size());
}


// VirtualSpace

VirtualSpace::VirtualSpace() {
  _low_boundary  = NULL;
  _high_boundary = NULL;
  _low_to_high   = true;
  _low           = NULL;
  _high          = NULL;
}


bool VirtualSpace::initialize(ReservedSpace rs, size_t committed_size, bool low_to_high) {
  assert(rs.is_reserved(), "ReservedSpace not initialized");
  assert(_low_boundary == NULL, "VirtualSpace already initialized");
  _low_boundary  = rs.base();
  _high_boundary = low_boundary() + rs.size();
  _low_to_high   = low_to_high;

  // initialize committed area
  _low  = low_to_high ? low_boundary() : high_boundary();
  _high = low();

  // commit to initial size
  if (committed_size > 0) {
    if (!expand_by(committed_size)) {
      return false;
    }
  }
  return true;
}


VirtualSpace::~VirtualSpace() { 
  release();
}


void VirtualSpace::release() {
  os::release_memory(low_boundary(), reserved_size());
  _low_boundary  = NULL;
  _high_boundary = NULL;
  _low           = NULL;
  _high          = NULL;
}


size_t VirtualSpace::committed_size() const { 
  return high() - low();
}


size_t VirtualSpace::reserved_size() const {
  return high_boundary() - low_boundary();
}


size_t VirtualSpace::uncommitted_size()  const {
  return reserved_size() - committed_size();
}


bool VirtualSpace::contains(void* p) const {
  return low() <= (char*) p && (char*) p < high();
}


bool VirtualSpace::low_to_high() const {
  return _low_to_high;
}


bool VirtualSpace::expand_by(size_t bytes) {
  assert((bytes % os::vm_page_size()) == 0, "size not page aligned");
  if (uncommitted_size() < bytes) return false;
  if (low() == low_boundary()) {
    if (!os::commit_memory(high(), bytes)) {
      debug_only(warning("os::commit_memory failed"));
      return false;
    }
    _high += bytes;
  } else {
    if (!os::commit_memory(low() - bytes, bytes)) {
      debug_only(warning("os::commit_memory failed"));
      return false;
    }
    _low -= bytes;
  }
  return true;
}


void VirtualSpace::shrink_by(size_t size) {
  assert((size % os::vm_page_size()) == 0, "size not page aligned");
  if (committed_size() < size) fatal("Cannot shrink virtual space to negative size");
  if (low() == low_boundary()) {
    _high -= size;
    if (!os::uncommit_memory(high(), size)) fatal("os::uncommit_memory failed");
  } else {
    if (!os::uncommit_memory(low(), size)) fatal("os::uncommit_memory failed");
    _low += size;
  }
}


#ifndef PRODUCT

void VirtualSpace::print() {  
  tty->print_cr("Virtual space:");
  tty->print_cr(" - committed: %u", committed_size());
  tty->print_cr(" - reserved:  %u",  reserved_size());
  tty->print_cr(" - [low, high]:     [0x%lx, 0x%lx]",  low(), high());
  tty->print_cr(" - [low_b, high_b]: [0x%lx, 0x%lx]",  low_boundary(), high_boundary());
}

#endif

