/*
 * @(#)vm_operations.cpp	1.112 01/03/27 09:51:04
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_vm_operations.cpp.incl"


void VM_Operation::set_calling_thread(Thread* thread, ThreadPriority priority) { 
  _calling_thread = thread; 
  assert(MinPriority <= priority && priority <= MaxPriority, "sanity check");
  _priority = priority;
}  


void VM_Operation::evaluate() {
  ResourceMark rm;
  if (TraceVMOperation) { 
    tty->print("["); 
    NOT_PRODUCT(print();)
  }
  doit();
  if (TraceVMOperation) { 
    tty->print_cr("]"); 
  }
}


void VM_Operation::report_error(const char* title, char* message) {  
  tty->print_cr("# Error happened during: %s", name());
  tty->print_cr("#");
}


#ifndef PRODUCT
void VM_Operation::print() { 
  const char* mode;
  switch(evaluation_mode()) {
    case _safepoint      : mode = "safepoint";       break;
    case _no_safepoint   : mode = "no safepoint";    break;
    case _concurrent     : mode = "concurrent";      break;
    case _async_safepoint: mode = "async safepoint"; break;
    default              : ShouldNotReachHere();
  }
  tty->print("VM_Operation %s Mode: %s", name(), mode); 
}
#endif // !PRODUCT


void VM_GC_Operation::acquire_pending_list_lock() {
  // we may enter this with pending exception set
  _notify_ref_lock = false;
  PRESERVE_EXCEPTION_MARK;  // exceptions are never thrown, needed for TRAPS argument
  Handle h_lock(java_lang_ref_Reference::pending_list_lock());
  ObjectSynchronizer::fast_enter(h_lock, &_pending_list_basic_lock, THREAD);
  if (HAS_PENDING_EXCEPTION) CLEAR_PENDING_EXCEPTION;
}


void VM_GC_Operation::release_and_notify_pending_list_lock() {
  // we may enter this with pending exception set
  PRESERVE_EXCEPTION_MARK;  // exceptions are never thrown, needed for TRAPS argument
  Handle h_lock(java_lang_ref_Reference::pending_list_lock());
  if (_notify_ref_lock) {
    // Notify waiters on pending lists lock. The _notify_ref_lock flag is set by doit.
    ObjectSynchronizer::notifyall(h_lock, THREAD);
  }
  ObjectSynchronizer::fast_exit(h_lock(), &_pending_list_basic_lock, THREAD);
  if (HAS_PENDING_EXCEPTION) CLEAR_PENDING_EXCEPTION;
}


bool VM_GC_Operation::doit_prologue() {
  assert(Thread::current()->is_Java_thread(), "just checking");
  _scavenge_count = Scavenge::invoke_count();
  _mark_sweep_count  = MarkSweep::invoke_count();
  acquire_pending_list_lock();
  // If the GC count has changed someone beat us to the collection
  if (gc_count_changed()) {
    // skip collection
    release_and_notify_pending_list_lock();
    return false;
  } else {
    return true;
  }
}


void VM_GC_Operation::doit_epilogue() {
  assert(Thread::current()->is_Java_thread(), "just checking");
  release_and_notify_pending_list_lock();
}


void VM_Scavenge::doit() {
  Scavenge::invoke_at_safepoint(_size, false, _notify_ref_lock);
} 			


bool VM_Scavenge::gc_count_changed() const {
  return _scavenge_count != Scavenge::invoke_count() || _mark_sweep_count != MarkSweep::invoke_count();
}


void VM_MarkSweep::doit() {
  MarkSweep::invoke_at_safepoint(_size, false, false, _notify_ref_lock, _cause);
}


bool VM_MarkSweep::gc_count_changed() const {
  return _mark_sweep_count != MarkSweep::invoke_count();
}


void VM_ThreadStop::doit() {
  assert(SafepointSynchronize::is_at_safepoint(), "must be at a safepoint");
  JavaThread* target = java_lang_Thread::thread(target_thread());  
  // Note that this now allows multiple ThreadDeath exceptions to be
  // thrown at a thread.
  if (target != NULL) {
    // the thread has run and is not already in the process of exiting
    target->send_thread_stop(throwable());
  }  
}


void VM_ThreadSuspend::doit() {
  assert(SafepointSynchronize::is_at_safepoint(), "must be at a safepoint");
  JavaThread* target = java_lang_Thread::thread(target_thread());  
  if (target != NULL) {
    // The thread has run and is not in the process of exiting.
    // Since we are at a safepoint, we don't have to refresh target
    // at every retry attempt.
    while (true) {
      // request that the thread be suspended
      target->send_java_suspend(is_native_now());

      if (!is_native_now()) break;  // no immediate suspend so we are done

      //
      // See if the VMThread is likely to deadlock due to this suspend:
      // - tell SuspendCheckerThread to check for deadlock
      // - briefly wait for it to tell us it is still okay
      // - if we time out, then we want to undo this suspend
      //
      MutexLockerEx scl(SuspendChecker_lock, Mutex::_no_safepoint_check_flag);
      // If the SuspendCheckerThread is not ready for work yet or has
      // exited during VM shutdown, then we skip checking this suspend
      // operation.
      if (SuspendCheckerThread::suspend_checker_thread() == NULL) break;
      SuspendChecker_lock->notify();
      bool timedOut = SuspendChecker_lock->wait(true, 1000);
      if (!timedOut) break;

      //
      // The SuspendCheckerThread is stuck. The thread we just suspended
      // may be holding some critical resource which means the VMThread
      // will deadlock if we leave the thread suspended. It is also
      // possible that some other running thread temporarily has the
      // resource and for some reason has held it longer than our timeout
      // above. The latter is the false positive case and it shouldn't
      // hurt to temporarily resume the thread.
      //
      target->resume(true, true);  // yield, external resume request
    }

    // It would be nice to have the following assertion in all the
    // time, but it is possible for a racing resume request to have
    // resumed this thread right after we suspended it. Temporarily
    // enable this assertion if you are chasing a different kind of
    // bug.
    //
    // assert(java_lang_Thread::thread(target->threadObj()) == NULL ||
    //   target->is_being_ext_suspended(), "thread is not suspended");
  }
}


#ifndef CORE

VM_Deoptimize::VM_Deoptimize(KlassHandle dependee) {
  _dependee = dependee;
}


void VM_Deoptimize::doit() {
  // We do not want any GCs to happen while we are in the middle of this VM operation
  ResourceMark rm;
  DeoptimizationMarker dm;

  // Deoptimize all activations depending on marked nmethods  
  Deoptimization::deoptimize_dependents();

  // Make the dependent methods zombies
  CodeCache::make_marked_nmethods_zombies();
}


VM_DeoptimizeFrame::VM_DeoptimizeFrame(JavaThread* thread, jint* sp) {
  _thread = thread;
  _sp     = sp;
}


void VM_DeoptimizeFrame::doit() {
  Deoptimization::deoptimize_frame(_thread, _sp);
}


#ifndef PRODUCT

void VM_DeoptimizeAll::doit() {
  DeoptimizationMarker dm;
  // deoptimize all java threads in the system
  for (JavaThread* thread = Threads::first(); thread != NULL; thread = thread->next()) {
    if (thread->has_last_Java_frame()) {
      thread->deoptimize();
    }
  }
}


void VM_ZombieAll::doit() {
  JavaThread *thread = (JavaThread *)calling_thread();
  assert(thread->is_Java_thread(), "must be a Java thread");
  thread->make_zombies();
}

#endif // !PRODUCT
#endif // !CORE


#ifndef PRODUCT
void VM_Verify::doit() {
  Universe::verify();
}
#endif // !PRODUCT


void VM_GenerateOopMap::doit() {
  _mask = _method->mask_for(_bci);
}
