/*
 * @(#)vm_operations.hpp	1.81 01/04/21 22:48:10
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

// The following classes are used for operations
// initiated by a Java thread but that must
// take place in the VMThread.

class VM_Operation: public CHeapObj {
 public:
  enum Mode { _safepoint,       // Requesting thread     blocked,    safepoint, VM_Operation C-Heap allocated
              _no_safepoint,    // Requesting thread     blocked, no safepoint, VM_Operation C-Heap allocated
              _concurrent,      // Requesting thread not blocked, no safepoint, VM_Operation C-Heap allocated
              _async_safepoint  // Requesting thread not blocked,    safepoint, VM_Operation C-Heap allocated 
            };
              
 private:
  Thread*	  _calling_thread;
  ThreadPriority  _priority;
  long            _timestamp;
  VM_Operation*	  _next;  
  VM_Operation*   _prev;
  bool            _is_cheap_allocated;
 public:
  VM_Operation()  { _calling_thread = NULL; _next = NULL; _prev = NULL; _is_cheap_allocated = false; }

  // VM operation support (used by VM thread)  
  Thread* calling_thread() const                 { return _calling_thread; }  
  ThreadPriority priority()                      { return _priority; }
  void set_calling_thread(Thread* thread, ThreadPriority priority);
  
  long timestamp() const              { return _timestamp; }
  void set_timestamp(long timestamp)  { _timestamp = timestamp; }  
    
  // Called by VM thread - does in turn invoke doit(). Do not override this
  void evaluate();  
    
  // evaluate() is called by the VMThread and in turn calls doit(). 
  // If the thread invoking VMThread::execute((VM_Operation*) is a JavaThread, 
  // doit_prologue() is called in that thread before transferring control to 
  // the VMThread.
  // If doit_prologue() returns true the VM operation will proceed, and 
  // doit_epilogue() will be called by the JavaThread once the VM operation 
  // completes. If doit_prologue() returns false the VM operation is cancelled.    
  virtual void doit()                            = 0;
  virtual bool doit_prologue()                   { return true; };
  virtual void doit_epilogue()                   {}; // Note: Not called if mode is: _concurrent

  // Type test
  virtual bool is_methodCompiler() const         { return false; }
  
  // Linking
  VM_Operation *next() const			 { return _next; }
  VM_Operation *prev() const                     { return _prev; }
  void set_next(VM_Operation *next)		 { _next = next; }
  void set_prev(VM_Operation *prev)		 { _prev = prev; }
  
  // Configuration. Override these appropriatly in subclasses.
  virtual const char* name() const                = 0;  
  virtual Mode evaluation_mode() const            { return _safepoint; }  
  virtual bool allow_nested_vm_operations() const { return false; }    
  virtual bool is_cheap_allocated() const         { return false; }
  virtual void report_error(const char* title, char* message); // Called when an assert happens in a VM Operation
  virtual void oops_do(void f(oop*))              { /* do nothing */ };

  // DO NOT OVERWRITE THE FOLLOWING METHODS (overwrite evaluation_mode)  
  bool evaluate_at_safepoint() const { return evaluation_mode() == _safepoint  || evaluation_mode() == _async_safepoint; }   
  bool evaluate_concurrently() const { return evaluation_mode() == _concurrent || evaluation_mode() == _async_safepoint; }       

  // Debugging
  void print() PRODUCT_RETURN;
};


class VM_GC_Operation: public VM_Operation {
 protected:
  BasicLock   _pending_list_basic_lock;        // for refs pending list notification
  int         _scavenge_count;
  int         _mark_sweep_count;

  virtual bool gc_count_changed() const           { return false; }

  // java.lang.ref.Reference support
  void acquire_pending_list_lock();
  void release_and_notify_pending_list_lock();

  int         _size;                           // size of object to be allocated
  bool        _notify_ref_lock;                // tells whether waiters on reference lock should be notified

 public:
  // Acquire the reference synchronization lock
  bool doit_prologue();
  // Do notifyAll (if needed) and release held lock
  void doit_epilogue();

  bool allow_nested_vm_operations() const         { return true; }
};


class VM_Scavenge: public VM_GC_Operation {
 protected:
  bool gc_count_changed() const;
 public:
  VM_Scavenge(int size)                          { _size = size; }
  void doit();
  const char* name() const                       { return "scavenge"; }
};


class VM_MarkSweep: public VM_GC_Operation {
 protected:
  bool gc_count_changed() const;
  MarkSweepInvokeCause _cause;
 public:
  VM_MarkSweep(int size, MarkSweepInvokeCause cause)  { _size = size; _cause = cause; }
  void doit();
  const char* name() const                        { return "mark sweep"; }
};


class VM_ThreadStop: public VM_Operation {
 private:  
  oop     _thread;        // The Thread that the Throwable is thrown against
  oop     _throwable;     // The Throwable thrown at the target Thread  
 public:
  // All oops are passed as JNI handles, since there is no guarantee that a GC might happen before the
  // VM operation is executed.
  VM_ThreadStop(oop thread, oop throwable) {    
    _thread    = thread;  
    _throwable = throwable;
  }
  oop target_thread() const                      { return _thread; }
  oop throwable() const                          { return _throwable;}
  void doit();
  const char* name() const                       { return "thread stop"; }
  // We deoptimize if top-most frame is compiled - this might require a C2I adapter to be generated
  bool allow_nested_vm_operations() const        { return true; }
  Mode evaluation_mode() const                   { return _async_safepoint; }
  bool is_cheap_allocated() const                { return true; }

  // GC support
  void oops_do(void f(oop*))                     { f(&_thread); f(&_throwable);  }
};


class VM_ThreadSuspend: public VM_Operation {
 private:
  oop  _thread;                                  // The Thread to be suspended
  bool _native_now;                              // Requires immediate native suspension
 public:
  VM_ThreadSuspend(oop thread, bool native_now)  { _thread = thread; _native_now = native_now;}  
  
  oop target_thread() const                      { return _thread; }
  bool is_native_now() const                     { return _native_now; }

  void doit();
  const char* name() const                       { return "thread suspend"; }
  bool is_cheap_allocated() const                { return true; }
  // GC support
  void oops_do(void f(oop*))                     { f(&_thread); }
  // synchronous as far as setting is_external_suspend() before returning
  // and ensuring no additional bytecodes run.
  Mode evaluation_mode() const                   { return _safepoint; }
};

// dummy vm op, evaluated just to force a safepoint
class VM_ForceSafepoint: public VM_Operation {
 public:
  VM_ForceSafepoint() { }  
  void doit() {}
  const char* name() const                       { return "VM_ForceSafepoint"; }  
};

// dummy vm op, evaluated just to force a safepoint
class VM_ForceAsyncSafepoint: public VM_Operation {
 public:
  VM_ForceAsyncSafepoint() { }  
  void doit() {}
  const char* name() const                       { return "VM_ForceAsyncSafepoint"; }
  Mode evaluation_mode() const                   { return _async_safepoint; }
  bool is_cheap_allocated() const                { return true; }
};

#ifndef CORE

class VM_Deoptimize: public VM_Operation {
 private:
  KlassHandle _dependee;
 public:
  VM_Deoptimize(KlassHandle dependee);
  void doit();
  const char* name() const                       { return "deoptimize"; }
  bool allow_nested_vm_operations() const        { return true; }
};

class VM_DeoptimizeFrame: public VM_Operation {
 private:
  JavaThread* _thread;
  jint*       _sp;
 public:
  VM_DeoptimizeFrame(JavaThread* thread, jint* sp);
  void doit();
  const char* name() const                       { return "deoptimize frame"; }
  Mode evaluation_mode() const                   { return _no_safepoint; }
  bool allow_nested_vm_operations() const        { return true;  }
};


#ifndef PRODUCT
class VM_DeoptimizeAll: public VM_Operation {
 private:
  KlassHandle _dependee;
 public:
  VM_DeoptimizeAll() {}
  void doit();
  const char* name() const                       { return "deoptimize-all"; }
  bool allow_nested_vm_operations() const        { return true; }
};


class VM_ZombieAll: public VM_Operation {
 public:
  VM_ZombieAll() {}
  void doit();
  const char* name() const                       { return "zombie-all"; }
  bool allow_nested_vm_operations() const        { return true; }
};
#endif // PRODUCT
#endif // CORE


#ifndef PRODUCT
class VM_Verify: public VM_Operation {
 private:
  KlassHandle _dependee;
 public:
  VM_Verify() {}
  void doit();
  const char* name() const                       { return "verify"; }
};
#endif // PRODUCT


class VM_GenerateOopMap: public VM_Operation {
 private:
  methodOop         _method;
  int               _bci;
  OopMapCacheEntry* _mask;

 public:
  VM_GenerateOopMap( methodOop method, int bci) : VM_Operation() { 
    _method = method;
    _bci = bci;
  }

  void doit();
  const char* name() const                       { return "generate oop map"; }
  OopMapCacheEntry* mask()                       { return _mask; }
};
