/*
 * @(#)debug.cpp	1.110 01/04/21 22:48:12
 *
 * Copyright 1993-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_debug.cpp.incl"

#ifndef ASSERT
#  ifdef _DEBUG
   // NOTE: don't turn the lines below into a comment -- if you're getting
   // a compile error here, change the settings to define ASSERT
   ASSERT should be defined when _DEBUG is defined.  It is not intended to be used for debugging
   functions that do not slow down the system too much and thus can be left in optimized code.
   On the other hand, the code should not be included in a production version.
#  endif
#endif

#ifdef _DEBUG
#  ifndef ASSERT
     configuration error: ASSERT must be defined in debug version
#  endif
#endif

#ifdef PRODUCT
#  if -defined _DEBUG || -defined ASSERT
     configuration error: ASSERT et al. must not be defined in PRODUCT version
#  endif
#endif


// Convert the file_name and line_no into an obfuscated string.
// The algorithm does the following:
// Given, for example:
// 
//   file_name:  d:\hotspot\src\share\vm\runtime.cpp
//   line_no:    257
//
// Extract the file_name without path, e.g., runtime.cpp. Then outputs
// all the characters in 2-digit hexidecimal form where 32 has been 
// subtracted from each character. Followed by the linenumber in
// 4-digit hexidecimal.
//
// In the above example we will get something like:
//
//    4D5D827452A3A30101
//
// (Note: the above hexidecimal numbers are totally random)
//
// This only happens in PRODUCT builds. In DEBUG, and RELEASE builds
// the filename and linenuber is printed out directly.
//
void obfuscate_location(const char *file_name, int line_no, char* buf, int buflen) {
#ifndef PRODUCT
  int len = strlen(file_name);
  strncpy(buf, file_name, buflen);
  if (len + 10 < buflen) {
    sprintf(buf + len, ", %d", line_no);
  }    
#else
  int len = strlen(file_name);  
  if (len > 0) {
    // Find last part of string
    int i;
    for(i = len -1; i > 0 && (file_name[i] != '\\' && file_name[i] != '/'); i--);
    file_name = &file_name[i+1];
  }
  
  // Output filename into buffer. The 10 is so there is alway enough room
  // for both the line number, and a terminating zero, plus a little extra
  // to be safe.
  while(*file_name && buflen > 10) {
    sprintf(buf, "%02X", *file_name - ' ');
    buf += 2;
    buflen -= 2;
    file_name++;
  }  
  // Add then the line number
  sprintf(buf, "%04X", line_no);  
#endif

  pd_obfuscate_location(buf,buflen);
}


void warning(const char* format, ...) {
  tty->print("%s warning: ", VM_Version::vm_name());
  va_list ap;
  va_start(ap, format);
  tty->vprint_cr(format, ap);
  va_end(ap);
  if (BreakAtWarning) BREAKPOINT;
}


void report_vm_state() {
  if (CountBytecodes || TraceBytecodes || StopInterpreterAt) {
    tty->print("# ");
    BytecodeCounter::print();
  }
#ifndef CORE
#ifndef PRODUCT
  CompileBroker::print_last_compile();
  {
    // Print statistics as if before normal exit:
    extern void print_statistics();
    print_statistics();
  }
#endif // PRODUCT
#endif // CORE
  if (!DumpEventsAtError) return;
  static bool recursive_error = false;
  if (!recursive_error) {
    recursive_error = true;
    tty->print("# Last 10 internal VM events:\n");
    Events::print_last(10);
    recursive_error = false;
    tty->print_cr("###");
  }
}


void report_fatal(const char* file_name, int line_no, const char* format, ...) {
  char buffer[2048];  
  va_list ap;
  va_start(ap, format);
  vsprintf(buffer, format, ap);
  va_end(ap);  
  report_error(true, file_name, line_no, "Internal Error", "Fatal: %s", buffer);
}


void report_assertion_failure(const char* code_str, const char* file_name, int line_no, const char* message) {
  report_error(true, file_name, line_no,
               "assertion failure",
               "assert(%s, \"%s\")", code_str, message);
}


void report_should_not_call(const char* file_name, int line_no) {
  report_error(true, file_name, line_no,
               "Internal Error", "ShouldNotCall()");
}


void report_should_not_reach_here(const char* file_name, int line_no) {
  report_error(true, file_name, line_no,
               "Internal Error", "ShouldNotReachHere()");
}


void report_subclass_responsibility(const char* file_name, int line_no) {
  report_error(true, file_name, line_no,
               "Internal Error", "SubclassResponsibility()");
}


void report_unimplemented(const char* file_name, int line_no) {
  report_error(true, file_name, line_no,
               "Internal Error", "Unimplemented()");
}


void report_untested(const char* file_name, int line_no, const char* msg) {
#ifndef PRODUCT
  warning("Untested: %s in %s: %d\n", msg, file_name, line_no);
#endif
}


// helper function, returns the active vm operation if present.
static VM_Operation* active_vm_operation() {
  Thread* thread = ThreadLocalStorage::thread();
  if (thread == NULL)          return NULL;
  if (!thread->is_VM_thread()) return NULL;
  return VMThread::vm_operation();
}

extern "C" void ps();

static jint error_level = 1;
static bool error_reported = false;

// call this when the VM is dying--it might loosen some asserts
void set_error_reported() {
  error_reported = true;
}

bool is_error_reported() {
    return error_reported || error_level > 1;
}

jint inc_error_level() {
    return atomic::add(&error_level, 1);
}

jint dec_error_level() {
    return atomic::add(&error_level, -1);
}

void report_error(bool is_vm_internal_error, const char* file_name, int line_no, const char* title, const char* format,  ...) {  
  // Handle the recursive case  
  switch(inc_error_level()) {
   case 2:  // first time, do nothing
            break; 
   case 3:  // second time print recursive problem 
            tty->print_cr("[error occured during error reporting]"); 
            return;
   default: // otherwise just say NO.
            exit(-1);
  }
  
  // In case error happens before init
  if (tty == NULL) ostream_init();

  // Compute location string
  char loc_buf[256];
  if (file_name != NULL) {
    obfuscate_location(file_name, line_no, loc_buf, sizeof(loc_buf));
  } else {
    strcpy(loc_buf, "<unknown>");
  }

  // Compute the message
  char message[2*K];
  va_list ap;
  va_start(ap, format);
  vsprintf(message, format, ap);
  va_end(ap);

  if (is_vm_internal_error) {
    // Print error has happen
    tty->print_cr("#");
    tty->print_cr("# HotSpot Virtual Machine Error, %s", title);
    tty->print_cr("# Please report this error at");
    tty->print_cr("# http://java.sun.com/cgi-bin/bugreport.cgi");
    tty->print_cr("#");
  }
#ifdef PRODUCT
  else            // Do not print out any message in PRODUCT mode
#endif
  {
    char* begin = message;
    // print line by line
    for (;;) {
      char* end = strchr(begin, '\n');
      if (!end) {
        break;
      }
      *end = '\0';
      tty->print_raw("# ");
      tty->print_raw(begin);
      *end = '\n';
      begin = end + 1;
    }
    // print last line
    if (*begin) {
      tty->print_raw("# ");
      tty->print_raw(begin);
      tty->cr();
    }
    // print end mark
    tty->print_cr("#");
  }

#ifndef PRODUCT
  if (WizardMode || DumpEventsAtError) {
    tty->print_cr("#");
    report_vm_state();
  }
#ifdef COMPILER2
  if (DumpInlineImage) {
    // For easy diffing, re-read the image file
    // and write it back in sorted form
    InlineImage::parse_from_file();
    InlineImage::sort_by_name();
    InlineImage::dump_to_file();
  }
#endif
#endif

  if (is_vm_internal_error) {
    // printout is busted  
    Thread* current = ThreadLocalStorage::thread();
    // Dispatch the error to active vm operation if present
    VM_Operation* op = active_vm_operation();
    if (op != NULL) {
      op->report_error(title, message);
      current = op->calling_thread();
    }

    tty->print_cr("# Error ID: %s", loc_buf);
    tty->print_cr("#");
  }

  tty->print("# Problematic Thread: ");
  Thread::current()->print();
  tty->print_cr("");
  tty->print_cr("#");

#ifndef PRODUCT
  static int continue_counter = 0;
  if (Debugging && continue_counter++ < 5) {
    // don't stop/abort VM when assertion was triggered during 
    // expression evaluation in the debugger
    tty->print("Continuing...\n");
    dec_error_level();
    return;
  } 
#endif

  if (ShowMessageBoxOnError) {
    strcat(message, "\n\nDo you want to debug the problem?");
    if (!os::message_box(title, message)) os::abort();
  } else {
#ifndef PRODUCT
    if (WizardMode) {
      tty->print_cr("Dumping thread state before exit....");
      ps();
    }
#endif
    if (is_vm_internal_error) {
      os::abort();
    }
  }
  dec_error_level();
}


// ------ helper functions for debugging go here ------------

#ifndef PRODUCT
// All debug entries should be wrapped with a stack allocated
// Command object. It makes sure a resource mark is set and
// flushes the logfile to prevent file sharing problems.

class Command : public StackObj {
 private:
  ResourceMark rm;
  HandleMark   hm;
 public:
  static int level;
  Command(const char* str) {
    if (level++ > 0)  return;
    tty->cr();
    tty->print_cr("\"Executing %s\"", str);
    Debugging = true;
  }

  ~Command() { tty->flush(); Debugging = (--level > 0); }
};
int Command::level = 0;


#ifndef CORE

extern "C" void blob(CodeBlob* cb) {  
  Command c("blob");
  cb->print();  
}

extern "C" void dump_vtable(address p) {
  Command c("dump_vtable");
  klassOop k = (klassOop)p;
  instanceKlass::cast(k)->vtable()->print();
}

extern "C" void nm(int p) {
  // Actually we look through all CodeBlobs (the nm name has been kept for backwards compatability)
  Command c("nm");
  CodeBlob* cb = CodeCache::find_blob((address)p);
  if (cb == NULL) {
    tty->print_cr("NULL");
  } else {
    cb->print();
  }
}

extern "C" void disnm(int p) {
  Command c("disnm");
  CodeBlob* cb = CodeCache::find_blob((address) p);
  cb->print();
  Disassembler::decode(cb);
}

extern "C" Space* space_for(int p) {
  Space* space = CarTable::space_for(oop(p));
  if (space) space->print();
  return space;
}

#endif

extern "C" void universe() {
  Command c("universe");
  Universe::print();
}

extern "C" void verify() {
  // try to run a verify on the entire system
  // note: this may not be safe if we're not at a safepoint; for debugging,
  // this manipulates the safepoint settings to avoid assertion failures
  Command c("universe verify");
  bool safe = SafepointSynchronize::is_at_safepoint();
  if (!safe) {
    tty->print_cr("warning: not at safepoint -- verify may fail");
    SafepointSynchronize::set_is_at_safepoint();
  }
  // Ensure Eden top is correct before verification
  Universe::new_gen()->eden()->ensure_proper_top();
  Universe::update_offset_tables();
  Universe::verify(true);
  if (!safe) SafepointSynchronize::set_is_not_at_safepoint();
}


extern "C" void pp(void* p) {
  Command c("pp");
  FlagSetting fl(PrintVMMessages, true);
  if (Universe::is_heap((oop) p)) {
    oop obj = oop(p);
    obj->print();
  } else {
    tty->print("%#x", p);
  }
}


// pv: print vm-printable object
extern "C" void pa(int p)   { ((AllocatedObj*) p)->print(); }
extern "C" void findpc(int x);

extern "C" void ps() { // print stack
  Command c("ps");

  frame f = JavaThread::current_frame_guess();

  // Prints the stack of the current Java thread
  JavaThread* p = JavaThread::active();
  tty->print(" for thread: ");
  p->print();
  tty->cr();

  if (p->has_last_Java_frame()) {
    // If the last_Java_fp is set we are in C land and
    // can call the standard stack_trace function.
    p->trace_stack();
  } else {    
    RegisterMap reg_map(p);
    f = f.sender(&reg_map);
    tty->print("(guessing starting frame sp=%#x based on current fp)\n", f.sp());
    p->trace_stack_from(vframe::new_vframe(&f, &reg_map, p));
  }

  pd_ps(f);
}

extern "C" void psf() { // print stack frames
  {
    Command c("psf");
    JavaThread* p = JavaThread::active();
    tty->print(" for thread: ");
    p->print();
    tty->cr();
    if (p->has_last_Java_frame()) {
      p->trace_frames();
    }
  }
}

extern "C" void threads() {
  Command c("threads");
  Threads::print(false, true);
}

extern "C" void psd() {
  Command c("psd");
  SystemDictionary::print();
}

extern "C" void safepoints() {
  Command c("safepoints");
  SafepointSynchronize::print_state();
}

extern "C" void pss() { // print all stacks
  Command c("pss");
  Threads::print(true, true);
}

extern "C" void debug() {		// to set things up for compiler debugging
  Command c("debug");
  WizardMode = true;
  PrintVMMessages = PrintCompilation = true;
  PrintInlining = PrintAssembly = true;
  tty->flush();
}


extern "C" void ndebug() {		// undo debug()
  Command c("ndebug");
  PrintCompilation = false;
  PrintInlining = PrintAssembly = false;
  tty->flush();
}


extern "C" void flush()  {
  Command c("flush");
  tty->flush();
}


extern "C" void events() {
  Command c("events");
  Events::print_last(50);
}

#ifndef CORE

extern "C" methodOop findm(int pc) { 
  Command c("findm");
  nmethod* nm = CodeCache::find_nmethod((address)pc);
  return (nm == NULL) ? NULL : nm->method(); 
}


extern "C" nmethod* findnm(int addr) {
  Command c("find");
  return  CodeCache::find_nmethod((address)addr);  
}

#endif

static address same_page(address x, address y) {
  int page_bits = -os::vm_page_size();
  if ((jint(x) & page_bits) == (jint(y) & page_bits))
    return x;
  else if (x > y)
    return (address)(jint(y) | ~page_bits) + 1;
  else
    return (address)(jint(y) & page_bits);
}

static void find(int x, bool print_pc) {
  address addr = (address)x;

#ifndef CORE
  CodeBlob* n = CodeCache::find_blob((address)x);
  if (n != NULL) { 
    if (print_pc && n->is_nmethod()) {
      ResourceMark rm;

      tty->print("%#x: Compiled ", x);
      ((nmethod*)n)->method()->print_value_on(tty); 
      tty->print("  = (CodeBlob*)0x%x", (int)n);
      tty->cr();
      return; 
    }
    n->print(); 
    return; 
  }

  VtableStub* v = VtableStubs::stub_containing(addr);
  if (v != NULL) {
    v->print();
    return;
  }
#endif

  StubCodeDesc* d = StubCodeDesc::desc_for(addr);
  if (d != NULL) { 
    d->print(); 
    if (print_pc)  tty->cr();
    return; 
  }
  InterpreterCodelet* i = Interpreter::codelet_containing(addr);
  if (i != NULL) { 
    i->print(); 
    return; 
  }
  if (StubRoutines::contains(addr)) {
    tty->print_cr("0x%x is pointing to an (unnamed) stub routine", x);
    return;
  }
  if (Interpreter::contains(addr)) {
    tty->print_cr("0x%x is pointing into interpreter code (not bytecode specific)", x);
    return;
  }
  oop p = Universe::slow_object_start((oop*)x);
  if (p != NULL) { 
    p->print(); 
    if (p != (oop)x && p->is_method() &&
	methodOop(p)->contains(address(x)) && !methodOop(p)->is_native()) {
      tty->print_cr("bci_from(%x) = %d; print_codes():", x, methodOop(p)->bci_from(address(x)));
      methodOop(p)->print_codes();
    }
    return; 
  }
  if (JNIHandles::is_global_handle((jobject) x)) {
    tty->print_cr("0x%x is a global jni handle", x);
    return;
  }
  if (JNIHandles::is_weak_global_handle((jobject) x)) {
    tty->print_cr("0x%x is a weak global jni handle", x);
    return;
  }
  if (JNIHandleBlock::any_contains((jobject) x)) {
    tty->print_cr("0x%x is a local jni handle", x);
    return;
  }

#ifndef CORE
  for(JavaThread *thread = Threads::first(); thread; thread = thread->next()) {
    // Check for safepoint code buffer
    address cpc = thread->safepoint_state()->compute_adjusted_pc(addr);
    if (cpc != addr) {
      tty->print_cr("0x%x is pointing into safepoint codebuffer for thread: 0x%x", x, thread);
      CodeBlob* n = CodeCache::find_blob((address)cpc);
      if (n != NULL) { 
        n->print(); 
        return; 
      }
    }      
    // Check for priviledge stack
    if (thread->privileged_stack_top() != NULL && thread->privileged_stack_top()->contains(addr)) {
      tty->print_cr("0x%x is pointing into the priviledge stack for thread: 0x%x", addr, thread);
      return;
    }
  }

  if (InlineCacheBuffer::contains(addr)) {
    tty->print_cr("0x%x is pointing into InlineCacheBuffer");
    return;
  }
#endif
  
  // Try an OS specific find
  if (os::find(addr)) {
    return;
  }

  if (print_pc) {
    tty->print_cr("%#x: probably in C++ code; check debugger", x);
    Disassembler::decode(same_page(addr-40,addr),same_page(addr+40,addr));
    return;
  }
  tty->print_cr("0x%x is pointing to unknown location", x);
}

extern "C" void find(int x) {
  Command c("find");
  find(x, false);
}

extern "C" void findpc(int x) {
  Command c("findpc");
  find(x, true);
}


// int versions of all methods to avoid having to type type casts in the debugger

void pp(int p) { pp((void*)p); }
void pp(oop p) { pp((void*)p); }

void help() { 
  Command c("help");
  tty->print_cr("basic");
  tty->print_cr("  pp(void* p)   - try to make sense of p");
  tty->print_cr("  pv(int p)     - ((PrintableResourceObj*) p)->print()");
  tty->print_cr("  ps()          - print current thread stack");
  tty->print_cr("  pss()         - print all thread stacks");
  tty->print_cr("  pm(int pc)    - print methodOop given compiled PC");
  tty->print_cr("  findm(int pc) - finds methodOop");
  tty->print_cr("  find(int x)   - finds & prints nmethod/stub/bytecode/oop based on pointer into it");
 
  tty->print_cr("misc.");
  tty->print_cr("  flush()       - flushes the log file");
  tty->print_cr("  events()      - dump last 50 events");


  tty->print_cr("compiler debugging");
  tty->print_cr("  debug()       - to set things up for compiler debugging");
  tty->print_cr("  ndebug()      - undo debug");
}

#endif // PRODUCT


