/*
 * @(#)exceptions.cpp	1.58 00/03/06 14:38:46
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2


#include "incls/_precompiled.incl"
#include "incls/_exceptions.cpp.incl"


// Implementation of ThreadShadow

void check_ThreadShadow() {
  int offset1 = (int)&((ThreadShadow*)NULL)->_pending_exception;
  int offset2 = Thread::pending_exception_offset_in_bytes();
  if (offset1 != offset2) fatal("ThreadShadow::_pending_exception is not positioned correctly");
}


// Implementation of Exceptions

bool Exceptions::special_exception(Thread* thread, const char* file, int line, Handle h_exception) {
  // bootstrapping check
  if (!Universe::is_fully_initialized()) {     
   vm_exit_during_initialization(h_exception);
   ShouldNotReachHere();
  }

  if (thread->is_VM_thread()
      NOT_CORE( || thread->is_Compiler_thread() ) ) {
    // We do not care what kind of exception we get for the vm-thread or a thread which
    // is compiling.  We just install a dummy exception object
    thread->set_pending_exception(Universe::vm_exception(), file, line);
    return true;
  }

  return false;
}

bool Exceptions::special_exception(Thread* thread, const char* file, int line, symbolHandle h_name, const char* message) {
  // bootstrapping check
  if (!Universe::is_fully_initialized()) {     
   vm_exit_during_initialization(h_name, message);
   ShouldNotReachHere();
  }

  if (thread->is_VM_thread()
      NOT_CORE( || thread->is_Compiler_thread() ) ) {
    // We do not care what kind of exception we get for the vm-thread or a thread which
    // is compiling.  We just install a dummy exception object
    thread->set_pending_exception(Universe::vm_exception(), file, line);
    return true;
  }

  return false;
}

// This method should only be called from generated code, 
// therefore the exception oop should be in the oopmap.
void Exceptions::_throw_oop(Thread* thread, const char* file, int line, oop exception) {
  assert(exception != NULL, "exception should not be NULL");
  Handle h_exception = Handle(thread, exception);
  _throw(thread, file, line, h_exception);
}

void Exceptions::_throw(Thread* thread, const char* file, int line, Handle h_exception) {
  assert(h_exception() != NULL, "exception should not be NULL");

  // tracing (do this up front - so it works during boot strapping)
  if (TraceExceptions) {
    ThreadCritical tc;
    ResourceMark rm;
    tty->print_cr("Exception <%s> thrown [%s, line %d] for thread 0x%x", h_exception->print_value_string(), file, line, thread);
  }  

  // Check for special boot-strapping/vm-thread handling
  if (special_exception(thread, file, line, h_exception)) return;
 
  assert(h_exception->is_a(SystemDictionary::throwable_klass()), "exception is not a subclass of java/lang/Throwable");
    
  if (VMStackTraceAtException || VMStackTraceAtNullException) debug_trace(thread, h_exception); 
  // set the pending exception
  thread->set_pending_exception(h_exception(), file, line);
  
  // vm log
  Events::log("throw_exception %#x", h_exception());
}


void Exceptions::_throw_msg(Thread* thread, const char* file, int line, symbolHandle h_name, const char* message, Handle h_loader, Handle h_protection_domain) {
  // Check for special boot-strapping/vm-thread handling
  if (special_exception(thread, file, line, h_name, message)) return;
  // Create and throw exception
  Handle h_exception = new_exception(thread, h_name, message, h_loader, h_protection_domain);
  _throw(thread, file, line, h_exception);
}


void Exceptions::_throw_args(Thread* thread, const char* file, int line, symbolHandle h_name, symbolHandle h_signature, JavaCallArguments *args) {
  // Check for special boot-strapping/vm-thread handling
  if (special_exception(thread, file, line, h_name, NULL)) return;
  // Create and throw exception
  Handle exception = new_exception(thread, h_name, h_signature, args, NULL, NULL);
  _throw(thread, file, line, exception);
}


void Exceptions::throw_stack_overflow_exception(Thread* THREAD, const char* file, int line) {
  klassOop k = SystemDictionary::StackOverflowError_klass();
  oop exception = instanceKlass::cast(k)->allocate_instance(CHECK);
  _throw_oop(THREAD, file, line, exception);
}

void Exceptions::fthrow(Thread* thread, const char* file, int line, symbolHandle h_name, const char* format, ...) {
  const int max_msg_size = 1024;
  va_list ap;
  va_start(ap, format);  
  char msg[max_msg_size];
  vsnprintf(msg, max_msg_size, format, ap); 
  msg[max_msg_size-1] = '\0';
  va_end(ap);
  _throw_msg(thread, file, line, h_name, msg, NULL, NULL);
}

// Creates an exception oop, calls the <init> method with the given signature.
// and returns a Handle
Handle Exceptions::new_exception(Thread *thread, symbolHandle h_name, symbolHandle signature, JavaCallArguments *args, Handle h_loader, Handle h_protection_domain) {
  assert(Universe::is_fully_initialized(), 
    "cannot be called during initialization");
  assert(thread->is_Java_thread(), "can only be called by a Java thread");
  assert(!thread->has_pending_exception(), "already has exception");

  Handle h_exception;
  
  // Resolve exception klass
  instanceKlassHandle klass = SystemDictionary::resolve_or_fail(h_name, h_loader, h_protection_domain, true, thread);

  if (!thread->has_pending_exception()) {    
    assert(klass.not_null(), "klass must exist");
    // We are about the create an instance - so make sure that klass is initialized
    klass->initialize(thread);
    if (!thread->has_pending_exception()) {      
      // Allocate new exception
      h_exception = klass->allocate_instance(thread);
      if (!thread->has_pending_exception()) {      
        args->set_receiver(h_exception);
        // Call constructor
        JavaValue result(T_VOID);      
        JavaCalls::call_special(&result, klass, 
                                         vmSymbolHandles::object_initializer_name(),
                                         signature,
                                         args,
                                         thread);
      }
    }
  }
   
  // Check if another exception was thrown in the process, if so rethrow that one
  if (thread->has_pending_exception()) {
    h_exception = Handle(thread, thread->pending_exception());
    thread->clear_pending_exception();
  }
  return h_exception;  
}

// Convenience method. Calls either the <init>() or <init>(String) method when creating a new exception
Handle Exceptions::new_exception(Thread* thread, symbolHandle h_name, const char* message, Handle h_loader, Handle h_protection_domain) {
  JavaCallArguments args; 
  symbolHandle signature;
  if (message == NULL) {
    signature = vmSymbolHandles::void_method_signature();    
  } else {
    Handle msg(thread, java_lang_String::create_from_str(message, thread));
    if (thread->has_pending_exception()) {
      Handle exception = Handle(thread, thread->pending_exception());
      thread->clear_pending_exception();
      return exception;
    }
    args.push_oop(msg); 
    signature = vmSymbolHandles::string_void_signature();    
  }
  return new_exception(thread, h_name, signature, &args, h_loader, h_protection_domain);       
}

// Implementation of ExceptionMark

ExceptionMark::ExceptionMark(Thread*& thread) {
  thread     = Thread::current();
  _thread    = thread;
  if (_thread->has_pending_exception()) {
    oop exception = _thread->pending_exception();
    _thread->clear_pending_exception(); // Needed to avoid infinite recursion
    exception->print();
    fatal("ExceptionMark constructor expects no pending exceptions");
  }
}


ExceptionMark::~ExceptionMark() {
  if (_thread->has_pending_exception()) {
    Handle exception(_thread, _thread->pending_exception());
    _thread->clear_pending_exception(); // Needed to avoid infinite recursion
    if (is_init_completed()) {
      exception->print();
      fatal("ExceptionMark destructor expects no pending exceptions");
    } else {
      vm_exit_during_initialization(exception);
    }
  }
}

// ----------------------------------------------------------------------------------------

#ifndef PRODUCT

void Exceptions::debug_trace(Thread* thread, Handle exception) {
  bool trace = VMStackTraceAtException;
  if (VMStackTraceAtNullException) {
    EXCEPTION_MARK;
    klassOop k = SystemDictionary::resolve_or_fail(vmSymbolHandles::java_lang_NullPointerException(), NULL, NULL, true, CATCH);
    trace |= exception->klass() == k;
  }
  if (trace) {
    if (thread->is_Java_thread()) {
      ((JavaThread*)thread)->trace_stack();
    } else {
      tty->print_cr("*exception in non-Java thread");
    }
  }
}
#else

void Exceptions::debug_trace(Thread* thread, Handle exception) { }

#endif
