/*
 * @(#)ostream.cpp	1.28 00/04/20 01:11:11
 *
 * Copyright 1993-1999 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// CopyrightVersion 1.2

# include "incls/_precompiled.incl"
# include "incls/_ostream.cpp.incl"

extern "C" void jio_print(const char* s); // Declarationtion of jvm method

outputStream::outputStream(int width) {
  _width       = width;
  _position    = 0;
  _indentation = 0;
}

outputStream::outputStream(int width, bool has_time_stamps) {
  _width       = width;
  _position    = 0;
  _indentation = 0;
  if (has_time_stamps)  _stamp.update();
}

#define BUFLEN 2000   /* max size of output of individual print() methods */

void outputStream::print(const char* format, ...) {
  char buffer[BUFLEN];
  va_list ap;
  va_start(ap, format);
  if (vsnprintf(buffer, BUFLEN, format, ap) < 0) {
    warning("increase BUFLEN in ostream.cpp -- output truncated");
    buffer[BUFLEN - 1] = 0;
  }
  va_end(ap);
  print_raw(buffer);
}

void outputStream::print_cr(const char* format, ...) { 
  char buffer[BUFLEN];
  va_list ap;
  va_start(ap, format);
  if (vsnprintf(buffer, BUFLEN, format, ap) < 0) {
    warning("increase BUFLEN in ostream.cpp -- output truncated");
    buffer[BUFLEN - 1] = 0;
  }
  va_end(ap);
  print_raw(buffer); 
  cr();
}

void outputStream::vprint(const char *format, va_list argptr) { 
  char buffer[BUFLEN];
  if (vsnprintf(buffer, BUFLEN, format, argptr) < 0) {
    warning("increase BUFLEN in ostream.cpp -- output truncated");
    buffer[BUFLEN - 1] = 0;
  }
  print_raw(buffer);
}

void outputStream::vprint_cr(const char* format, va_list argptr) {
  vprint(format, argptr);
  cr();
}

void outputStream::fill_to(int col) {
  while (position() < col) sp();
}

void outputStream::put(char ch) {
  assert(ch != 0, "please fix call site");
  char buf[] = { ch, '\0' };
  print_raw(buf);
}

void outputStream::sp() {
  this->print_raw(" ");
}

void outputStream::cr() {
  this->print_raw("\n");
}

void outputStream::stamp() {
  print("%g", _stamp.seconds());
}

void outputStream::indent() {
  while (_position < _indentation) sp();
}

void outputStream::print_jlong(jlong value) {
  print(os::jlong_format_specifier(), value);
}

void outputStream::print_julong(julong value) {
  print(os::julong_format_specifier(), value);
}

stringStream::stringStream(int initial_size) : outputStream() {
  buffer_length = initial_size;
  buffer        = NEW_RESOURCE_ARRAY(char, buffer_length);
  buffer_pos    = 0;  
}

void stringStream::print_raw(const char* s) {
  int len = strlen(s);
  int end = buffer_pos + len;
  if (end >= buffer_length) {
    if (end < buffer_length * 2) {
      end = buffer_length * 2;
    }
    char* oldbuf = buffer;
    buffer = NEW_RESOURCE_ARRAY(char, end);
    strncpy(buffer, oldbuf, buffer_pos);
    buffer_length = end;
  }
  strncpy(buffer + buffer_pos, s, len);
  buffer_pos += len;

  for (;;) {
    char ch = *s++;
    if (ch == 0) {
      break;
    } else if (ch == '\n') {
      _position = 0;
    } else if (ch == '\t') {
      _position += 8;
    } else {
      _position += 1;
    }
  }
}

char* stringStream::as_string() {
  print_raw("\n");
  buffer[buffer_pos-1] = 0;
  char* copy = NEW_RESOURCE_ARRAY(char, buffer_pos);
  strcpy(copy, buffer);
  return copy;
}

stringStream::~stringStream() {}

outputStream* tty;

fileStream::fileStream(char* file_name) {
  _file = fopen(file_name, "w");
}

void fileStream::print_raw(const char* s) {
  fprintf(_file, "%s", s);
  for (;;) {
    char ch = *s++;
    if (ch == 0) {
      break;
    } else if (ch == '\n') {
      _position = 0;
    } else {
      _position += 1;
    }
  }
}

fileStream::~fileStream() {
  fclose(_file);
  _file = NULL;
}

void fileStream::flush() {
  fflush(_file);
}


class defaultStream : public outputStream {
 private:
  FILE* log_file;
  bool has_log_file() { 
    // lazily create log file (at startup, LogVMOutput is false even
    // if +LogVMOutput is used, because the flags haven't been parsed yet
    if (LogVMOutput && log_file == NULL) log_file = fopen("hotspot.log", "w");
    return log_file != NULL; 
  }
 public:
  defaultStream() : outputStream(80, true) {
    log_file = NULL;
  }

  ~defaultStream() { 
    if (has_log_file()) fclose(log_file);
  }

  virtual void print_raw(const char* s);

  void flush() {
    // assert(false, "can not flush buffer inside signal handler");
  }
};

void defaultStream::print_raw(const char* s) {
  // print to output stream. It can be redirected by a vfprintf hook
  jio_print(s);  

  // print to log file
  if (has_log_file()) {
    ::write(fileno(log_file), s, strlen(s));
  }

  for (;;) {
    char ch = *s++;
    if (ch == 0) {
      break;
    } else if (ch == '\n') {
      _position = 0;
    } else {
      _position += 1;
    }
  }
}

void ostream_init() {
  if (tty == NULL) {
    tty = new(ResourceObj::C_HEAP) defaultStream();
  }
}

void ostream_init_log() {
}

void ostream_exit() {
  delete tty;
  tty = NULL;
}
