/*
 * @(#)ArrayTypeImpl.c	1.15 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
#include <stdlib.h>
#include <string.h>

#include "ArrayTypeImpl.h"
#include "util.h"
#include "inStream.h"
#include "outStream.h"
#include "JDWP.h"

static jint 
getComponentClass(JNIEnv *env, jclass arrayClass, char *componentSignature, 
                jclass *componentClassPtr) 
{
    jobject arrayClassLoader;
    jclass *classes;
    jint count;
    jclass componentClass = NULL;
    jint error = JVMDI_ERROR_NONE;

    arrayClassLoader = classLoader(arrayClass);
    classes = allLoadedClasses(&count);
    if (classes == NULL) {
        error = JVMDI_ERROR_OUT_OF_MEMORY;
    } else {
        int i;
        for (i = 0; (i < count) && (componentClass == NULL); i++) {
            jobject loader;
            char *signature;
            jclass clazz = classes[i];
            loader = classLoader(clazz);
            signature = classSignature(clazz);
            if (signature == NULL) {
                (*env)->DeleteGlobalRef(env, loader);
                break;
            }

            if ((strcmp(signature, componentSignature) == 0) &&
                (loader == arrayClassLoader)) {
                componentClass = clazz;
            } else {
                (*env)->DeleteGlobalRef(env, clazz);
            }
            jdwpFree(signature);
            (*env)->DeleteGlobalRef(env, loader);
        }
        *componentClassPtr = componentClass;
        jdwpFree(classes);
    }
    (*env)->DeleteGlobalRef(env, arrayClassLoader);

    return error;
}

static void
writeNewObjectArray(JNIEnv *env, PacketOutputStream *out, 
                 jclass arrayClass, jint size, char *componentSignature) {

    jarray array;
    jclass componentClass;

    jint error = getComponentClass(env, arrayClass, 
                                       componentSignature, &componentClass);
    if (error != JVMDI_ERROR_NONE) {
        outStream_setError(out, error);
        return;
    }

    WITH_LOCAL_REFS(env, 1);

    array = (*env)->NewObjectArray(env, size, componentClass, 0);
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionClear(env);
        array = NULL;
    }

    if (array == NULL) {
        outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
    } else {
        outStream_writeByte(out, specificTypeKey(array));
        WRITE_LOCAL_REF(env, out, array);
    }

    END_WITH_LOCAL_REFS(env);
}

static void
writeNewPrimitiveArray(JNIEnv *env, PacketOutputStream *out, 
                       jclass arrayClass, jint size, char *componentSignature) {

    jarray array = NULL;

    WITH_LOCAL_REFS(env, 1);

    switch (componentSignature[0]) {
        case JDWP_Tag_BYTE:
            array = (*env)->NewByteArray(env, size);
            break;

        case JDWP_Tag_CHAR:
            array = (*env)->NewCharArray(env, size);
            break;

        case JDWP_Tag_FLOAT:
            array = (*env)->NewFloatArray(env, size);
            break;

        case JDWP_Tag_DOUBLE:
            array = (*env)->NewDoubleArray(env, size);
            break;

        case JDWP_Tag_INT:
            array = (*env)->NewIntArray(env, size);
            break;

        case JDWP_Tag_LONG:
            array = (*env)->NewLongArray(env, size);
            break;

        case JDWP_Tag_SHORT:
            array = (*env)->NewShortArray(env, size);
            break;

        case JDWP_Tag_BOOLEAN:
            array = (*env)->NewBooleanArray(env, size);
            break;

        default:
            outStream_setError(out, JVMDI_ERROR_TYPE_MISMATCH);
            break;
    }

    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionClear(env);
        array = NULL;
    }

    if (array == NULL) {
        outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
    } else {
        outStream_writeByte(out, specificTypeKey(array));
        WRITE_LOCAL_REF(env, out, array);
    }

    END_WITH_LOCAL_REFS(env);
}

static jboolean 
newInstance(PacketInputStream *in, PacketOutputStream *out)
{
    JNIEnv *env = getEnv();
    char *arraySignature;
    char *componentSignature;
    jclass arrayClass = inStream_readClassRef(in);
    jint size = inStream_readInt(in);
    if (inStream_error(in)) {
        return JNI_TRUE;
    }

    arraySignature = classSignature(arrayClass);
    componentSignature = &arraySignature[1];

    if ((componentSignature[0] == JDWP_Tag_OBJECT) || 
        (componentSignature[0] == JDWP_Tag_ARRAY)) {
        writeNewObjectArray(env, out, arrayClass, size, componentSignature);
    } else {
        writeNewPrimitiveArray(env, out, arrayClass, size, componentSignature);
    }

    jdwpFree(arraySignature);
    return JNI_TRUE;
}

void *ArrayType_Cmds[] = { (void *)0x1
                          ,(void *)newInstance};
