/*
 * @(#)EventRequestImpl.c	1.28 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
#include "EventRequestImpl.h"
#include "eventHandler.h"
#include "util.h"
#include "inStream.h"
#include "outStream.h"
#include "JDWP.h"
#include "stepControl.h"

#define JDWPEventType_to_JVMDIEventKind(eventType) (eventType)
    
#define ERROR_RETURN(func) \
{jvmdiError error = (func); if (error != JVMDI_ERROR_NONE) return error; }

static jvmdiError
readAndSetSelectors(PacketInputStream *in, HandlerNode *node, 
                    jint selectorCount, jbyte eventType)
{
    jthread threadOnly = NULL;
    JNIEnv *env = getEnv();
    jint i;

    for (i = 0; i < selectorCount; ++i) {
        jbyte modifier = inStream_readByte(in);
        switch (modifier) {
            case JDWP_REQUEST_MODIFIER(Conditional): {
                jint exprID = inStream_readInt(in);
                ERROR_RETURN(inStream_error(in));
                ERROR_RETURN(eventHandler_setConditionalSelector(node, i, exprID));
                break;
            }

            case JDWP_REQUEST_MODIFIER(Count): {
                jint count = inStream_readInt(in);
                ERROR_RETURN(inStream_error(in));
                ERROR_RETURN(eventHandler_setCountSelector(node, i, count));
                break;
            }

            case JDWP_REQUEST_MODIFIER(ThreadOnly): {
                jthread thread = inStream_readThreadRef(in);
                ERROR_RETURN(inStream_error(in));
                ERROR_RETURN(eventHandler_setThreadOnlySelector(node, i, 
                                                                thread));
                break;
            }

            case JDWP_REQUEST_MODIFIER(LocationOnly): {
                jbyte tag = inStream_readByte(in); /* not currently used */
                jclass clazz = inStream_readClassRef(in);
                jmethodID method = inStream_readMethodID(in);
                jlocation location = inStream_readLocation(in);
                ERROR_RETURN(inStream_error(in));
                ERROR_RETURN(eventHandler_setLocationOnlySelector(node, i,
                                             clazz, method, location));
                break;
            }

            case JDWP_REQUEST_MODIFIER(FieldOnly): {
                jclass clazz = inStream_readClassRef(in);
                jfieldID field = inStream_readFieldID(in);
                ERROR_RETURN(inStream_error(in));
                ERROR_RETURN(eventHandler_setFieldOnlySelector(node, i,
                                                      clazz, field));
                break;
            }

            case JDWP_REQUEST_MODIFIER(ClassOnly): {
                jclass clazz = inStream_readClassRef(in);
                ERROR_RETURN(inStream_error(in));
                ERROR_RETURN(eventHandler_setClassOnlySelector(node, i, clazz));
                break;
            }

            case JDWP_REQUEST_MODIFIER(ExceptionOnly): {
                jclass exception = inStream_readClassRef(in);
                jboolean caught = inStream_readBoolean(in);
                jboolean uncaught = inStream_readBoolean(in);
                ERROR_RETURN(inStream_error(in));
                ERROR_RETURN(eventHandler_setExceptionOnlySelector(node, i,
                                             exception, caught, uncaught));
                break;
            }

            case JDWP_REQUEST_MODIFIER(ClassMatch): {
                char *pattern = inStream_readString(in);
                ERROR_RETURN(inStream_error(in));
                ERROR_RETURN(eventHandler_setClassMatchSelector(node, i,
                                                                pattern));
                break;
            }

            case JDWP_REQUEST_MODIFIER(ClassExclude): {
                char *pattern = inStream_readString(in);
                ERROR_RETURN(inStream_error(in));
                ERROR_RETURN(eventHandler_setClassExcludeSelector(node, i,
                                                                 pattern));
                break;
            }
            case JDWP_REQUEST_MODIFIER(Step): {
                jthread thread = inStream_readThreadRef(in);
                jint size = inStream_readInt(in);
                jint depth = inStream_readInt(in);
                ERROR_RETURN(inStream_error(in));
                ERROR_RETURN(eventHandler_setStepSelector(node, i,
                                                 thread, size, depth));
                break;
            }

            default:
                return JVMDI_ERROR_ILLEGAL_ARGUMENT;
        }
    }
    return JVMDI_ERROR_NONE;
}

static jboolean 
setCommand(PacketInputStream *in, PacketOutputStream *out) 
{
    jint error;
    jbyte kind;
    HandlerNode *node;
    jbyte eventType = inStream_readByte(in);
    jbyte suspendPolicy = inStream_readByte(in);
    jint selectorCount = inStream_readInt(in);

    if (inStream_error(in)) {
        return JNI_TRUE;
    }

    kind = JDWPEventType_to_JVMDIEventKind(eventType);
    node = eventHandler_alloc(selectorCount, kind, suspendPolicy);
    if (node == NULL) {
        outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
        return JNI_TRUE;
    }

    error = readAndSetSelectors(in, node, selectorCount, eventType);
    if (error == JVMDI_ERROR_NONE) {
        error = eventHandler_insert(node);
    }


    if (error == JVMDI_ERROR_NONE) {
        outStream_writeInt(out, eventHandler_id(node));
    } else {
        eventHandler_freeInternal(node);
        outStream_setError(out, error);
    }

    return JNI_TRUE;
}

static jboolean
clearCommand(PacketInputStream *in, PacketOutputStream *out) 
{
    jint error = JVMDI_ERROR_NONE;
    jbyte kind;
    jbyte eventType = inStream_readByte(in);
    jint handlerID = inStream_readInt(in);

    if (inStream_error(in)) {
        return JNI_TRUE;
    }

    kind = JDWPEventType_to_JVMDIEventKind(eventType);

    error = eventHandler_free(kind, handlerID);

    if (error != JVMDI_ERROR_NONE) {
        outStream_setError(out, error);
    }

    return JNI_TRUE;
}

static jboolean 
clearAllBreakpoints(PacketInputStream *in, PacketOutputStream *out) 
{
    jint error = eventHandler_freeAll(JVMDI_EVENT_BREAKPOINT);
    if (error != JVMDI_EVENT_BREAKPOINT) {
        outStream_setError(out, error);
    }
    return JNI_TRUE;
}

void *EventRequest_Cmds[] = { (void *)0x3
    ,(void *)setCommand
    ,(void *)clearCommand
    ,(void *)clearAllBreakpoints};

