/*
 * @(#)VirtualMachineImpl.c	1.67 01/04/21
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
#include <string.h>
#include <stdlib.h>

#include "VirtualMachineImpl.h"
#include "commonRef.h"
#include "inStream.h"
#include "outStream.h"
#include "util.h"
#include "eventHandler.h"
#include "eventHelper.h"
#include "threadControl.h"
#include "JDWP.h"
#include "version.h"

static char *versionName = "Java Debug Wire Protocol (Reference Implementation)";
static int majorVersion = 1;
static int minorVersion = 0;

/* retrieved on init - defaults in-case of failure to init */
static char *classpath_property = "";
static char *bootclasspath_property = "";
static char path_separator_property = ';';
static char *user_dir_property = "";
static char *vm_info_property = "";
    
static jboolean 
version(PacketInputStream *in, PacketOutputStream *out)
{
    jint jvmdiMajor;
    jint jvmdiMinor;
    char buf[500];
    char *vmName;
    char *vmVersion;
    
    vmVersion = version_vmVersion();
    if (vmVersion == NULL) {
        vmVersion = "<unknown>";
    }
    vmName = version_vmName();
    if (vmName == NULL) {
        vmName = "<unknown>";
    }

    /*
     * Write the descriptive version information
     */
    jvmdiMinor = jvmdiVersion();
    jvmdiMajor = (jvmdiMinor >> 16) & 0x0fff;
    jvmdiMinor &= 0xff;
    sprintf(buf, "%s version %d.%d\nJVM Debug Interface version %d.%d\n"
                 "JVM version %s (%s, %s)",
                  versionName, majorVersion, minorVersion, 
                  jvmdiMajor, jvmdiMinor,
                  vmVersion, vmName, vm_info_property);
    outStream_writeString(out, buf);

    /*
     * Write the JDWP version numbers
     */
    outStream_writeInt(out, majorVersion);
    outStream_writeInt(out, minorVersion);

    /*
     * Write the VM version and name
     */
    outStream_writeString(out, vmVersion);
    outStream_writeString(out, vmName);

    return JNI_TRUE;
}

static jboolean 
deleteRefArray(JNIEnv *env, jobject *refs, jint startIndex, jint count) 
{
    int i;
    for (i = startIndex; i < count; i++) {
        (*env)->DeleteGlobalRef(env, refs[i]);
    }
    jdwpFree(refs);
    return JNI_TRUE;
}

static jboolean 
classesForSignature(PacketInputStream *in, PacketOutputStream *out) 
{
    JNIEnv *env = getEnv();
    jint classCount;
    jint i;
    jclass *theClasses;
    char *signature;
    char *candidate;

    signature = inStream_readString(in);
    if (signature == NULL) {
        outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
        return JNI_TRUE;
    }
    if (inStream_error(in)) {
        return JNI_TRUE;
    }

    theClasses = allLoadedClasses(&classCount);
    if (theClasses == NULL) {
        outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
    } else {
        /* Count classes in theClasses which match signature */
        int matchCount = 0;
        /* Count classes written to the JDWP connection */
        int writtenCount = 0;

        for (i=0; i<classCount; i++) {
            jclass clazz = theClasses[i];
            jint status = classStatus(clazz);

            /* Filter out unprepared classes (arrays may or
             * may not be marked as prepared) */
            if (((status & JVMDI_CLASS_STATUS_PREPARED) == 0)
                      && !isArrayClass(clazz)) {
                continue;
            }
            candidate = classSignature(clazz);
            if (candidate == NULL) {
                deleteRefArray(env, theClasses, 0, classCount);
                jdwpFree(signature);
                outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
                return JNI_TRUE;
            }
            if (strcmp(candidate, signature) == 0) {
                /* Float interesting classes (those that
                 * are matching and are prepared) to the
                 * beginning of the array. 
                 * Do a swap so that uninteresting classes
                 * (at the end of the array) can properly
                 * have their global references deleted.
                 */
                theClasses[i] = theClasses[matchCount];
                theClasses[matchCount++] = clazz;
            }
            jdwpFree(candidate);
        }
        /* At this point matching prepared classes occupy
         * indicies 0 thru matchCount-1 of theClasses.
         * Indicies matchCount thru classCount-1 of 
         * theClasses will not be written but remain there
         * so they can be deleted by deleteRefArray() below.
         */

        outStream_writeInt(out, matchCount);
        for (; writtenCount < matchCount; writtenCount++) {
            jclass clazz = theClasses[writtenCount];
            jint status = classStatus(clazz);
            jbyte tag = referenceTypeTag(clazz);
            outStream_writeByte(out, tag);
            WRITE_GLOBAL_REF(env, out, clazz);
            outStream_writeInt(out, status);

            /* No point in continuing if there's an error */
            if (outStream_error(out)) {
                break;
            }
        }
        /* All written classes have their global ref
         * freed in the WRITE_GLOBAL_REF; The rest,
         * there because of failure termination and/or
         * because they do not match or are not prepared,
         * must be freed.
         */
        deleteRefArray(env, theClasses, writtenCount, classCount);
    }
    jdwpFree(signature);
    return JNI_TRUE;
}

static jboolean 
allClasses(PacketInputStream *in, PacketOutputStream *out)
{
    JNIEnv *env = getEnv();
    jint classCount;
    jint i;
    jclass *theClasses;

    theClasses = allLoadedClasses(&classCount);
    if (theClasses == NULL) {
        outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
    } else {
        /* Count classes in theClasses which are prepared */
        int prepCount = 0;
        /* Count classes written to the JDWP connection */
        int writtenCount = 0;  

        for (i=0; i<classCount; i++) {
            jclass clazz = theClasses[i];
            jint status = classStatus(clazz);

            /* Filter out unprepared classes (arrays may or
             * may not be marked as prepared) */
            if (((status & JVMDI_CLASS_STATUS_PREPARED) != 0)
                      || isArrayClass(clazz)) {
                /* Float interesting classes (those that
                 * are prepared) to the beginning of the array. 
                 * Do a swap so that uninteresting classes
                 * (at the end of the array) can properly
                 * have their global references deleted.
                 */
                theClasses[i] = theClasses[prepCount];
                theClasses[prepCount++] = clazz;
            }
        }
        /* At this point prepared classes occupy
         * indicies 0 thru prepCount-1 of theClasses.
         * Indicies prepCount thru classCount-1 of 
         * theClasses will not be written but remain there
         * so they can be deleted by deleteRefArray() below.
         */

        outStream_writeInt(out, prepCount);
        for (; writtenCount < prepCount; writtenCount++) {
            char *signature;
            jclass clazz = theClasses[writtenCount];
            jint status = classStatus(clazz);
            jbyte tag = referenceTypeTag(clazz);

            signature = classSignature(clazz);
            if (signature == NULL) {
                outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
                break;
            }

            outStream_writeByte(out, tag);
            WRITE_GLOBAL_REF(env, out, clazz);
            outStream_writeString(out, signature);
            outStream_writeInt(out, status);
            jdwpFree(signature);

            /* No point in continuing if there's an error */
            if (outStream_error(out)) {
                break;
            }
        }
        /* All written classes have their global ref
         * freed in the WRITE_GLOBAL_REF; The rest,
         * there because of failure termination and/or
         * because they are not prepared, must be freed.
         */
        deleteRefArray(env, theClasses, writtenCount, classCount);
    }
    return JNI_TRUE;
}

static jboolean 
getAllThreads(PacketInputStream *in, PacketOutputStream *out)
{
    JNIEnv *env = getEnv();
    jint threadCount;
    jint i;
    jthread *theThreads;

    theThreads = allThreads(&threadCount);
    if (theThreads == NULL) {
        outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
    } else {
        /* Squish out all of the debugger-spawned threads */
        threadCount = filterDebugThreads(theThreads, threadCount);
        
        outStream_writeInt(out, threadCount);
        for (i = 0; i <threadCount; i++) {
            WRITE_GLOBAL_REF(env, out, theThreads[i]);
        }

        jdwpFree(theThreads);
    } 
    return JNI_TRUE;
}

static jboolean 
topLevelThreadGroups(PacketInputStream *in, PacketOutputStream *out)
{
    jint groupCount;
    jthreadGroup *groups;

    groups = topThreadGroups(&groupCount);
    if (groups == NULL) {
        outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
    } else {
        JNIEnv *env = getEnv();
        int i;

        outStream_writeInt(out, groupCount);
        for (i = 0; i < groupCount; i++) {
            WRITE_GLOBAL_REF(env, out, groups[i]);
        }

        jdwpFree(groups);
    }
    return JNI_TRUE;
}

static jboolean 
dispose(PacketInputStream *in, PacketOutputStream *out)
{
    return JNI_TRUE;
}

static jboolean 
idSizes(PacketInputStream *in, PacketOutputStream *out)
{
    outStream_writeInt(out, sizeof(jfieldID));    /* fields */
    outStream_writeInt(out, sizeof(jmethodID));   /* methods */
    outStream_writeInt(out, sizeof(jlong));       /* objects */
    outStream_writeInt(out, sizeof(jlong));       /* referent types */
    outStream_writeInt(out, sizeof(jframeID));    /* frames */
    return JNI_TRUE;
}

static jboolean 
suspend(PacketInputStream *in, PacketOutputStream *out)
{
    jint error = threadControl_suspendAll();
    if (error != JVMDI_ERROR_NONE) {
        outStream_setError(out, error);
    }
    return JNI_TRUE;
}

static jboolean 
resume(PacketInputStream *in, PacketOutputStream *out)
{
    jint error = threadControl_resumeAll();
    if (error != JVMDI_ERROR_NONE) {
        outStream_setError(out, error);
    }
    return JNI_TRUE;
}

static jboolean 
doExit(PacketInputStream *in, PacketOutputStream *out)
{
    jint exitCode = inStream_readInt(in);
    
    /* We send the reply from here because we are about to exit. */
    if (inStream_error(in)) {
        outStream_setError(out, inStream_error(in));
    } 
    outStream_sendReply(out);

    /*
     * TO DO: When JVMDI provides a better way to exit, use it here.
     */
    exit(exitCode);

    /* Shouldn't get here */
    JDI_ASSERT(JNI_FALSE);

    /* Shut up the compiler */
    return JNI_FALSE;

}

static jboolean 
createString(PacketInputStream *in, PacketOutputStream *out)
{
    JNIEnv *env = getEnv();
    jstring string;
    char *cstring;

    cstring = inStream_readString(in);
    if (cstring == NULL) {
        outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
        return JNI_TRUE;
    }
    if (inStream_error(in)) {
        return JNI_TRUE;
    }

    WITH_LOCAL_REFS(env, 1);

    string = (*env)->NewStringUTF(env, cstring);
    if ((*env)->ExceptionOccurred(env)) {
        outStream_setError(out, JVMDI_ERROR_OUT_OF_MEMORY);
    } else {
        WRITE_LOCAL_REF(env, out, string);
    } 
    END_WITH_LOCAL_REFS(env);

    jdwpFree(cstring);
    return JNI_TRUE;
}

static jboolean 
capabilities(PacketInputStream *in, PacketOutputStream *out)
{
    JVMDI_capabilities caps;

    jint error = jvmdi->GetCapabilities(&caps);
    if (error != JVMDI_ERROR_NONE) {
        outStream_setError(out, error);
        return JNI_TRUE;
    }

    outStream_writeBoolean(out, (jboolean)caps.can_watch_field_modification);
    outStream_writeBoolean(out, (jboolean)caps.can_watch_field_access);
    outStream_writeBoolean(out, (jboolean)caps.can_get_bytecodes);
    outStream_writeBoolean(out, (jboolean)caps.can_get_synthetic_attribute);
    outStream_writeBoolean(out, (jboolean)caps.can_get_owned_monitor_info);
    outStream_writeBoolean(out, (jboolean)caps.can_get_current_contended_monitor);
    outStream_writeBoolean(out, (jboolean)caps.can_get_monitor_info);
    return JNI_TRUE;
}

#define NULL_CHECK(e) if (((e) == 0) || (*env)->ExceptionOccurred(env)) return

/*
 * Set classpath, bootclasspath and path separator properties.
 * Called when back-end is initialized.
 */
void
VirtualMachine_initialize(void)
{
    char *value;

    value = getPropertyCString("java.class.path");
    if (value != NULL) {
        classpath_property = value;
    }

    value = getPropertyCString("sun.boot.class.path");
    if (value != NULL) {
        bootclasspath_property = value;
    }

    value = getPropertyCString("path.separator");
    if (value != NULL) {
        path_separator_property = value[0];
        jdwpFree(value);
    }

    value = getPropertyCString("user.dir");
    if (value != NULL) {
        user_dir_property = value;
    }

    value = getPropertyCString("java.vm.info");
    if (value != NULL) {
        vm_info_property = value;
    }
}

void
VirtualMachine_reset(void)
{
}

static int 
countPaths(char *string) {
    int cnt = 1; /* always have one */
    char *pos = string;

    while ((pos = strchr(pos, path_separator_property)) != NULL) {
        ++cnt;
        ++pos;
    }
    return cnt;
}

static void 
writePaths(PacketOutputStream *out, char *string) {
    char *pos = string;
    char *newPos;

    outStream_writeInt(out, countPaths(string));

    while ((newPos = strchr(pos, path_separator_property)) != NULL) {
        outStream_writeByteArray(out, newPos-pos, (jbyte *)pos);
        pos = newPos + 1;
    }
    outStream_writeByteArray(out, strlen(pos), (jbyte *)pos);
}



static jboolean 
classPaths(PacketInputStream *in, PacketOutputStream *out)
{
    outStream_writeString(out, user_dir_property);
    writePaths(out, classpath_property);
    writePaths(out, bootclasspath_property);
    return JNI_TRUE;
}

static jboolean 
disposeObjects(PacketInputStream *in, PacketOutputStream *out)
{
    int i;
    int refCount;
    jlong id;

    int requestCount = inStream_readInt(in);
    if (inStream_error(in)) {
        return JNI_TRUE;
    }

    for (i = 0; i < requestCount; i++) {
        id = inStream_readObjectID(in);
        refCount = inStream_readInt(in);
        if (inStream_error(in)) {
            return JNI_TRUE;
        }
        commonRef_releaseMultiple(id, refCount);
    }

    return JNI_TRUE;
}

static jboolean 
holdEvents(PacketInputStream *in, PacketOutputStream *out)
{
    eventHelper_holdEvents();
    return JNI_TRUE;
}

static jboolean 
releaseEvents(PacketInputStream *in, PacketOutputStream *out)
{
    eventHelper_releaseEvents();
    return JNI_TRUE;
}

void *VirtualMachine_Cmds[] = { (void *)16
    ,(void *)version
    ,(void *)classesForSignature
    ,(void *)allClasses
    ,(void *)getAllThreads
    ,(void *)topLevelThreadGroups
    ,(void *)dispose
    ,(void *)idSizes
    ,(void *)suspend
    ,(void *)resume
    ,(void *)doExit
    ,(void *)createString
    ,(void *)capabilities
    ,(void *)classPaths
    ,(void *)disposeObjects
    ,(void *)holdEvents
    ,(void *)releaseEvents
};

