/*
 * @(#)debugLoop.c	1.28 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
#include <stdlib.h>

#include <jvmdi.h>
#include "util.h"
#include "transport.h"
#include "debugLoop.h"
#include "debugDispatch.h"
#include "eventHandler.h"
#include "commonRef.h"
#include "inStream.h"
#include "outStream.h"
#include "JDWP.h"
#include "threadControl.h"
#include "debugInit.h"

struct PacketList {
    struct Packet packet;
    struct PacketList *next;
};

static volatile struct PacketList *cmdQueue;
static JVMDI_RawMonitor cmdQueueLock;

static jboolean
lastCommand(struct CmdPacket *cmd)
{
    if ((cmd->cmdSet == JDWP_COMMAND_SET(VirtualMachine)) &&
        ((cmd->cmd == JDWP_COMMAND(VirtualMachine, Dispose)) ||
         (cmd->cmd == JDWP_COMMAND(VirtualMachine, Exit)))) {
        return JNI_TRUE;
    } else {
        return JNI_FALSE;
    }
}
/*
 * This is where all the work gets done.
 */
  
void
debugLoop_run()
{
    jboolean shouldListen;
    struct Packet p;

    char errString[128];
    CommandHandler func;
    jboolean replyToSender;

    cmdQueue = NULL;
    cmdQueueLock = debugMonitorCreate("JDWP Command Queue Lock");
    shouldListen = JNI_TRUE;

    spawnNewThread(debugLoop_reader, NULL, "JDWP Command Reader");

    eventHandler_onConnect();
    threadControl_onConnect();

    /* Okay, start reading cmds! */
    while (shouldListen) {
        p = debugLoop_dequeue();

        if (p.type.cmd.flags & FLAGS_Reply) {
            /*
             * Its a reply packet.
             */
            
        } else {
            /*
             * Its a cmd packet.
             */
            struct CmdPacket *cmd = &p.type.cmd;

            func = debugDispatch_getHandler(cmd->cmdSet,cmd->cmd);
            if (func == NULL) {
                sprintf(errString,"Recieved unknown JDWP packet cmdSet:[%d] cmd:[%d]"
                        ,cmd->cmdSet,cmd->cmd);
                ERROR_MESSAGE_EXIT(errString);
            } else {
                PacketInputStream in;
                PacketOutputStream out;

                /* Initialize the input and output streams */
                inStream_init(&in, p);
                outStream_initReply(&out, inStream_id(&in));

                /* Call the command handler */
                replyToSender = func(&in, &out);

                /* Reply to the sender */
                if (replyToSender) {
                    if (inStream_error(&in)) {
                        outStream_setError(&out, inStream_error(&in));
                    } 
                    outStream_sendReply(&out);
                }

                inStream_destroy(&in);
                outStream_destroy(&out);
                
                shouldListen = !lastCommand(cmd);
            }
        }
    }
    threadControl_onDisconnect();
    eventHandler_onDisconnect();

    /*
     * Cut off the transport immediately. This has the effect of 
     * cutting off any events that the eventHelper thread might
     * be trying to send.
     */
    transport_close();
    debugMonitorDestroy(cmdQueueLock);
    debugInit_reset();
}

void 
debugLoop_reader(void *unused)
{
    struct Packet packet;
    struct CmdPacket *cmd;
    jboolean shouldListen = JNI_TRUE;

    while (shouldListen) {
        jint error = transport_receivePacket(&packet);

        /* Is it a valid packet? */
        if (error != 0) {
            fprintf(stderr,"Transport error, error code = %d\n", error);
            shouldListen = JNI_FALSE;
        } else {
            cmd = &packet.type.cmd;

            /* 
             * FIXME! We need to deal with high priority
             * packets and queue flushes!
             */
            debugLoop_enqueue(packet);

            shouldListen = !lastCommand(cmd);
        }
    }
}

/*
 * The current system for queueing packets is highly
 * inefficient, and should be rewritten! It'd be nice
 * to avoid any additional mallocs.
 */

void 
debugLoop_enqueue(struct Packet p)
{
    struct PacketList *pL;
    struct PacketList *walker;

    pL = jdwpAlloc((size_t)sizeof(struct PacketList));
    if (pL == NULL) {
        ALLOC_ERROR_EXIT();
    }

    pL->packet = p;
    pL->next = NULL;

    /*fprintf(stderr,"Starting packet enqueue.\n");*/

    debugMonitorEnter(cmdQueueLock);

    if (cmdQueue == NULL) {
        cmdQueue = pL;
        debugMonitorNotify(cmdQueueLock);
    } else {
        walker = (struct PacketList *)cmdQueue;
        while (walker->next != NULL)
            walker = walker->next;

        walker->next = pL;
    }

    debugMonitorExit(cmdQueueLock);

    /*fprintf(stderr,"Done with packet enqueue.\n");*/
}

struct Packet 
debugLoop_dequeue() {
    struct Packet p;
    struct PacketList *pL;

    /*fprintf(stderr,"Starting packet dequeue.\n");*/

    debugMonitorEnter(cmdQueueLock); 

    while (cmdQueue == NULL) {
        debugMonitorWait(cmdQueueLock);
        /*fprintf(stderr,"dequeue got notify!\n");*/
    }

    pL = (struct PacketList *)cmdQueue;
    cmdQueue = pL->next;

    debugMonitorExit(cmdQueueLock);

    /*fprintf(stderr,"Done with packet dequeue.\n");*/

    p = pL->packet;

    jdwpFree(pL);

    return p;
}
