/*
 * @(#)eventHandler.c	1.81 01/04/21
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
/*
 * eventHandler
 *
 * This module handles events as they come in directly from JVMDI
 * and also maps them to JDI events.  JDI events are those requested
 * at the JDI or JDWP level and seen on those levels.  Mapping is
 * one-to-many, a JVMDI event may map to several JDI events, or
 * to none.  Part of that mapping process is filteration, which 
 * this module also handles.  A JDI EventRequest corresponds to
 * handler and a JDI filter to a selector.  For example, if at
 * the JDI level the user executed:
 *   
 *   EventRequestManager erm = vm.eventRequestManager();
 *   ClassPrepareRequest req = erm.createClassPrepareRequest();
 *   req.enable();
 *   req = erm.createClassPrepareRequest();
 *   req.addClassFilter("Foo*");
 *   req.enable();
 *
 * Two handlers would be created, the second with a ClassMatch
 * selector.  When a JVMDI class prepare event for "Foobar"
 * comes in, the first handler will create one JDI event, the
 * second handler will compare the class signature, and since
 * it matchs create a second event.  There may also be internal
 * events as there are in this case, one created by the front-end
 * and one by the back-end.
 *
 * Each event kind has a handler chain, which is a doublely linked
 * list of handlers for that kind of event.
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <jvmdi.h>
#include "util.h"
#include "debugInit.h"
#include "eventHandler.h"
#include "threadControl.h"
#include "eventHelper.h"
#include "classTrack.h"
#include "commonRef.h"
#include "JDWP.h"
#include "version.h"

static jint requestIdCounter;
static jbyte currentSessionID;

/*
 * We are starting with a very simple locking scheme for event handling
 * All readers and writers of data in the handlers[] or defunct_handlers[] 
 * chains must own this lock for the duration of its use. If contention
 * becomes a problem, we can:
 * 
 * 1) create a lock per event type. 
 * 2) move to a readers/writers approach where multiple threads 
 * can access the chains simultaneously while reading (the normal activity
 * of an event hook). However, cases where events are written and 
 * made defunct (e.g. Count modifier) will need some special handling. 
 */
static JVMDI_RawMonitor handlerLock;

typedef struct ClassSelector {
    jclass clazz;
} ClassSelector;

typedef struct LocationSelector {
    jclass clazz;
    jmethodID method;
    jlocation location;
} LocationSelector;

typedef struct ThreadSelector {
    jthread thread;
} ThreadSelector;

typedef struct CountSelector {
    jint count;
} CountSelector;

typedef struct ConditionalSelector {
    jint exprID;
} ConditionalSelector;

typedef struct FieldSelector {
    jclass clazz;
    jfieldID field;
} FieldSelector;

typedef struct ExceptionSelector {
    jclass exception;
    jboolean caught;
    jboolean uncaught;
} ExceptionSelector;

typedef struct StepSelector {
    jint size;
    jint depth;
    jthread thread;
} StepSelector;

typedef struct MatchSelector {
    char *classPattern;
} MatchSelector;

typedef union RequestSelector {
    struct ClassSelector ClassOnly;
    struct LocationSelector LocationOnly;
    struct ThreadSelector ThreadOnly;
    struct CountSelector Count;
    struct ConditionalSelector Conditional;
    struct FieldSelector FieldOnly;
    struct ExceptionSelector ExceptionOnly;
    struct StepSelector Step;
    struct MatchSelector ClassMatch;
    struct MatchSelector ClassExclude;
} RequestSelector;

typedef struct Selector_ {
    jbyte modifier;
    union RequestSelector u;
} Selector;

struct HandlerChain_;

struct HandlerNode_ {
    struct HandlerNode_ *next;
    struct HandlerNode_ *prev;
    EventHandler handler;
    jint nodeID;
    struct HandlerChain_ *chain;
    jbyte kind;
    jbyte suspendPolicy;
    jint selectorCount;
    Selector selectors[1]; /* one extra accounted for in alloc */
};

typedef struct HandlerChain_ {
    HandlerNode *first;
    /* add lock here */
} HandlerChain;

/*
 * This array maps event kinds to handler chains.
 */
static HandlerChain handlers[JVMDI_MAX_EVENT_TYPE_VAL+1];

/*
 * This array holds no longer active handlers waiting for
 * disposal.
 */
static HandlerChain defunct_handlers[JVMDI_MAX_EVENT_TYPE_VAL+1];

static jboolean 
patternStringMatch(char *classname, const char *pattern);

static void becomeDefunct(HandlerNode *node);

static void
handleInvokeRequest(jthread thread)
{
}

/*
 * Convert the signature "Ljava/lang/Foo;" to a 
 * classname "java.lang.Foo" compatible with the pattern.
 * Signature is overwritten in-place.
 */
static void
convertSignatureToClassname(char *convert)
{
    char *p;

    p = convert + 1;
    while ((*p != ';') && (*p != '\0')) {
        char c = *p;
        if (c == '/') {
            *(p-1) = '.';
        } else {
            *(p-1) = c;
        }
        p++;
    }
    *(p-1) = '\0';
}

/*
 * Match the name of a class against a wildcard
 * string pattern.
 */
static jboolean 
patternMatch(jclass clazz, const char *pattern)
{
    jboolean rc;
    char *signature = classSignature(clazz);

    if (signature == NULL) {
        return JNI_FALSE;
    }
    convertSignatureToClassname(signature);
    rc = patternStringMatch(signature, pattern);
    jdwpFree(signature);

    return rc;
}

/*
 * Match a string against a wildcard
 * string pattern.
 */
static jboolean 
patternStringMatch(char *classname, const char *pattern)
{
    int pattLen = strlen(pattern);
    int compLen;
    char *start;
    int offset;

    if ((pattern[0] != '*') && (pattern[pattLen-1] != '*')) {
        /* An exact match is required when there is no *: bug 4331522 */ 
        return strcmp(pattern, classname) == 0; 
    } else {
        compLen = pattLen - 1;
        offset = strlen(classname) - compLen;
        if (offset < 0) {
            return JNI_FALSE;
        } else {
            if (pattern[0] == '*') {
                pattern++;
                start = classname + offset;
            }  else {
                start = classname;
            }
            return strncmp(pattern, start, compLen) == 0;
        }
    }
}

static void
insert(HandlerChain *chain, HandlerNode *node)
{
    HandlerNode *oldHead = chain->first;
    node->next = oldHead;
    node->prev = NULL;
    node->chain = chain;
    if (oldHead != NULL) {
        oldHead->prev = node;
    }
    chain->first = node;
}

static HandlerNode *
findInChain(HandlerChain *chain, jint handlerID)
{
    HandlerNode *node = chain->first;
    while (node != NULL) {
        if (node->nodeID == handlerID) {
            return node;
        }
        node = node->next;
    }
    return NULL;
}

static HandlerNode *
find(jint kind, jint handlerID)
{
    HandlerNode *node = findInChain(&handlers[kind], handlerID);
    if (node == NULL) {
        node = findInChain(&defunct_handlers[kind], handlerID);
    } 
    return node;
}

/**
 * Deinsert.  Safe for non-inserted nodes.
 */
static void
deinsert(HandlerNode *node)
{
    HandlerChain *chain = node->chain;

    if (chain == NULL) {
        return;
    }
    if (chain->first == node) {
        chain->first = node->next;
    }
    if (node->next != NULL) {
        node->next->prev = node->prev;
    }
    if (node->prev != NULL) {
        node->prev->next = node->next;
    }
    node->chain = NULL;
}

static Selector *
findSelector(HandlerNode *node, jint modifier)
{
    int i;
    Selector *selector;
    for (i = 0, selector = node->selectors; i <node->selectorCount; i++, selector++) {
        if (selector->modifier == modifier) {
            return selector;
        }
    }
    return NULL;
}

static jint 
countBreakpointHandlers(jclass clazz, jmethodID method, jlocation location) 
{
    HandlerChain *chain = &handlers[JVMDI_EVENT_BREAKPOINT];
    HandlerNode *node = chain->first;
    JNIEnv *env = getEnv();
    int count = 0;

    while (node != NULL) {
        Selector *select = node->selectors;
        int i;
        
        for (i = 0; i < node->selectorCount; ++i, ++select) {
            switch (select->modifier) {
                case JDWP_REQUEST_MODIFIER(LocationOnly):
                    if (select->u.LocationOnly.method == method &&
                        select->u.LocationOnly.location == location &&
                        (*env)->IsSameObject(env, 
                                   select->u.LocationOnly.clazz, clazz)) {
                        count++;
                    }
                    break;
            }
        }
        node = node->next;
    }
    return count;
}

static jint 
countWatchpointHandlers(jint kind, jclass clazz, jfieldID field) 
{
    HandlerChain *chain = &handlers[kind];
    HandlerNode *node = chain->first;
    JNIEnv *env = getEnv();
    int count = 0;

    while (node != NULL) {
        Selector *select = node->selectors;
        int i;
        
        for (i = 0; i < node->selectorCount; ++i, ++select) {
            switch (select->modifier) {
                case JDWP_REQUEST_MODIFIER(FieldOnly):
                    if (select->u.FieldOnly.field == field &&
                        (*env)->IsSameObject(env, 
                                      select->u.FieldOnly.clazz, clazz)) {
                        count++;
                    }
                    break;
            }
        }
        node = node->next;
    }
    return count;
}

static jthread 
requestThread(HandlerNode *node) 
{
    int i;
    Selector *select = node->selectors;

    for (i = 0; i < node->selectorCount; ++i, ++select) {
        switch (select->modifier) {
            case JDWP_REQUEST_MODIFIER(ThreadOnly):
                return select->u.ThreadOnly.thread;
        }
    }

    return NULL;
}

/*
 * Count the number of handlers, specific to the given thread and 
 * event kind. If the thread is null, count the number of global events
 * of the given kind.
 */
static jint 
countThreadHandlers(jint kind, jthread thread) 
{
    HandlerChain *chain = &handlers[kind];
    HandlerNode *node = chain->first;
    JNIEnv *env = getEnv();
    int count = 0;

    while (node != NULL) {
        jthread reqThread = requestThread(node);

        /*
         * If they event's thread and the passed thread are the same
         * (or both are NULL), we have a match and should increment.
         */
        if ((*env)->IsSameObject(env, reqThread, thread)) {
            count++;
        }

        node = node->next;
    }
    return count;
}

/*
 * Determine if this event is interesting to this handler.
 * Do so by checking each of the handler's selectors.
 * Return false if any of the selectors fail, 
 * true if the handler wants this event.
 * Anyone modifying this function should check 
 * unloadSelectorFilter as well.
 */
static jboolean 
selectorFilter(JNIEnv *env, JVMDI_Event *event, HandlerNode *node)
{
    jthread thread;
    jclass clazz;
    Selector *select = node->selectors;
    int i;

    eventThreadAndClass(event, &thread, &clazz);

    /*
     * Suppress most events if they happen in debug threads
     */
    if ((event->kind != JVMDI_EVENT_CLASS_PREPARE) &&
        (event->kind != JVMDI_EVENT_CLASS_UNLOAD) && 
        (event->kind != JVMDI_EVENT_CLASS_LOAD) &&
        threadControl_isDebugThread(thread)) {
        return JNI_FALSE;
    }

    for (i = 0; i < node->selectorCount; ++i, ++select) {
        switch (select->modifier) {
            case JDWP_REQUEST_MODIFIER(ThreadOnly):
                if (!(*env)->IsSameObject(env, thread,
                                          select->u.ThreadOnly.thread)) {
                    return JNI_FALSE;
                }
                break;
        
            case JDWP_REQUEST_MODIFIER(ClassOnly):
                /*
                 * Class filters catch events in the specified class
                 * and any subclass/subinterface.
                 */
                if (!(*env)->IsAssignableFrom(env, clazz,
                                         select->u.ClassOnly.clazz)) {
                    return JNI_FALSE;
                }
                break;
        
            /* This is kinda cheating assumming the JVMDI event fields */
            /* will be in the same locations, but it is true now. */
            case JDWP_REQUEST_MODIFIER(LocationOnly):
                if  (event->u.breakpoint.method != 
                          select->u.LocationOnly.method ||
                     event->u.breakpoint.location != 
                          select->u.LocationOnly.location ||
                     !(*env)->IsSameObject(env, clazz,
                                           select->u.LocationOnly.clazz)) {
                    return JNI_FALSE;
                }
                break;

            case JDWP_REQUEST_MODIFIER(FieldOnly): 
                /*
                 * Field watchpoints can be triggered from the declared class
                 * or any subclass/subinterface.
                 */
                if ((event->u.field_access.field != 
		     select->u.FieldOnly.field) ||
                    !(*env)->IsSameObject(env,
					  event->u.field_access.field_clazz,
					  select->u.FieldOnly.clazz)) {
                    return JNI_FALSE;
                }
                break;
            
            case JDWP_REQUEST_MODIFIER(ExceptionOnly): 
                /* do we want caught/uncaught exceptions */
                if (!((event->u.exception.catch_clazz == NULL)? 
                      select->u.ExceptionOnly.uncaught : 
                      select->u.ExceptionOnly.caught)) {
                    return JNI_FALSE;
                }
    
                /* do we care about exception class */
                if (select->u.ExceptionOnly.exception != NULL) {
                    jclass exception = event->u.exception.exception;
        
                    /* do we want this exception class */
                    if (!(*env)->IsInstanceOf(env, exception,
                                    select->u.ExceptionOnly.exception)) {
                        return JNI_FALSE;
                    }
                }
                break;
            
            case JDWP_REQUEST_MODIFIER(Count): {
                JDI_ASSERT(select->u.Count.count > 0);
                if (--select->u.Count.count > 0) {
                    return JNI_FALSE;
                }
                becomeDefunct(node);
                break;
            }
                
            case JDWP_REQUEST_MODIFIER(Conditional): 
/***
                if (...  select->u.Conditional.exprID ...) {
                    return JNI_FALSE;
                }
***/
                break;

        case JDWP_REQUEST_MODIFIER(ClassMatch): {              
            if (!patternMatch(clazz, select->u.ClassMatch.classPattern)) {
                return JNI_FALSE;
            }
            break;
        }

        case JDWP_REQUEST_MODIFIER(ClassExclude): {              
            if (patternMatch(clazz, select->u.ClassExclude.classPattern)) {
                return JNI_FALSE;
            }
            break;
        }

        case JDWP_REQUEST_MODIFIER(Step): 
                if (!stepControl_handleStep(env, event)) {
                    return JNI_FALSE;
                }
                break;
                
        
            default: 
                    ERROR_MESSAGE_EXIT("Invalid selector modifier");
                    return JNI_FALSE;
        }
    }
    return JNI_TRUE;
}

/*
 * Determine if this event is interesting to this handler.
 * Do so by checking each of the handler's selectors.
 * Return false if any of the selectors fail, 
 * true if the handler wants this event.
 * Special selector filter for unloads since they don't have
 * an event structure or a jclass.
 */
static jboolean 
unloadSelectorFilter(JNIEnv *env, char *classname, HandlerNode *node)
{
    Selector *select = node->selectors;
    int i;

    for (i = 0; i < node->selectorCount; ++i, ++select) {
        switch (select->modifier) {
           
            case JDWP_REQUEST_MODIFIER(Count): {
                JDI_ASSERT(select->u.Count.count > 0);
                if (--select->u.Count.count > 0) {
                    return JNI_FALSE;
                }
                becomeDefunct(node);
                break;
            }
                
            case JDWP_REQUEST_MODIFIER(ClassMatch): {              
                if (!patternStringMatch(classname, 
                                        select->u.ClassMatch.classPattern)) {
                    return JNI_FALSE;
                }
                break;
            }
            
            case JDWP_REQUEST_MODIFIER(ClassExclude): {              
                if (patternStringMatch(classname, 
                                       select->u.ClassExclude.classPattern)) {
                    return JNI_FALSE;
                }
                break;
            }
            
            default: 
                ERROR_MESSAGE_EXIT("Invalid selector modifier");
                return JNI_FALSE;
        }
    }
    return JNI_TRUE;
}

jboolean
eventHandler_predictFiltering(HandlerNode *node, jframeID frame)
{
    /*
     * This function returns true only if it is certain that all events 
     * for the given node in the given frame will be filtered. It is 
     * used to  optimize stepping. (If this function returns true the 
     * stepping algorithm does not have to step through every instruction 
     * in this frame; instead, it can use more efficient method entry/exit
     * events.
     */

    Selector *select = node->selectors;
    int i;
    jclass clazz;
    jmethodID method;
    jlocation location;
    jboolean willBeFiltered = JNI_FALSE;
    jboolean done = JNI_FALSE;
    JNIEnv *env = getEnv();
    jint error;

    /* Safe to use, despite 4215724, because returned location not used */
    error = jvmdi->GetFrameLocation(frame, &clazz, &method, &location);
    if (error != JVMDI_ERROR_NONE) {
        return JNI_FALSE;
    }

    for (i = 0; (i < node->selectorCount) && (!done); ++i, ++select) {
        switch (select->modifier) {
            case JDWP_REQUEST_MODIFIER(ClassOnly):
                if (!(*env)->IsAssignableFrom(env, clazz,
                                         select->u.ClassOnly.clazz)) {
                    willBeFiltered = JNI_TRUE;
                    done = JNI_TRUE;
                }
                break;
        
            case JDWP_REQUEST_MODIFIER(Count): {   
                /*
                 * Once a count modifier is hit, we can never say for
                 * sure that filtering will happen for the frame because
                 * we can't predict the number of events that will happen
                 * in the frame.
                 */
                done = JNI_TRUE;
                break;
            }
                
            case JDWP_REQUEST_MODIFIER(ClassMatch): {              
                if (!patternMatch(clazz, select->u.ClassMatch.classPattern)) {
                    willBeFiltered = JNI_TRUE;
                    done = JNI_TRUE;
                }
                break;
            }
    
            case JDWP_REQUEST_MODIFIER(ClassExclude): {              
                if (patternMatch(clazz, select->u.ClassExclude.classPattern)) {
                    willBeFiltered = JNI_TRUE;
                    done = JNI_TRUE;
                }
                break;
            }
        }
    }
    (*env)->DeleteGlobalRef(env, clazz);
    return willBeFiltered;
}

/** 
 * Free up global refs held by the selector. 
 * free things up at the JVMDI level if needed. 
 */
static jint
clearSelectors(HandlerNode *node) 
{
    JNIEnv *env = getEnv();
    jint i;
    jint error = JVMDI_ERROR_NONE;
    Selector *select = node->selectors;

    for (i = 0; i < node->selectorCount; ++i, ++select) {
        switch (select->modifier) {
            case JDWP_REQUEST_MODIFIER(ThreadOnly): {
                jthread threadOnly = select->u.ThreadOnly.thread;
                if (threadOnly != NULL) { 
                    (*env)->DeleteGlobalRef(env, threadOnly);
                }
                break;
            }
            case JDWP_REQUEST_MODIFIER(LocationOnly): {
                jclass clazz = select->u.LocationOnly.clazz;
                (*env)->DeleteGlobalRef(env, clazz);
                break;
            }
            case JDWP_REQUEST_MODIFIER(FieldOnly): {
                jclass clazz = select->u.FieldOnly.clazz;
                (*env)->DeleteGlobalRef(env, clazz);
                break;
            }
            case JDWP_REQUEST_MODIFIER(ExceptionOnly): {
                (*env)->DeleteGlobalRef(env, 
                                        select->u.ExceptionOnly.exception);
                break;
            }
            case JDWP_REQUEST_MODIFIER(ClassOnly): {
                (*env)->DeleteGlobalRef(env, select->u.ClassOnly.clazz);
                break;
            }
            case JDWP_REQUEST_MODIFIER(ClassMatch): {
                jdwpFree(select->u.ClassMatch.classPattern);
                break;
            }
            case JDWP_REQUEST_MODIFIER(ClassExclude): {
                jdwpFree(select->u.ClassExclude.classPattern);
                break;
            }
            case JDWP_REQUEST_MODIFIER(Step): {
                jthread thread = select->u.Step.thread;
                error = stepControl_endStep(thread);
                if (error == JVMDI_ERROR_NONE) {
                    (*env)->DeleteGlobalRef(env, thread);
                }
                break;
            }
        }
    }
    if (error == JVMDI_ERROR_NONE) {
        node->selectorCount = 0; /* blast so we don't clear again */
    }

    return error;
}

jint
eventHandler_setConditionalSelector(HandlerNode *node, jint index, jint exprID)
{
    ConditionalSelector *selector = &node->selectors[index].u.Conditional;
    if (index >= node->selectorCount) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    node->selectors[index].modifier = JDWP_REQUEST_MODIFIER(Conditional);
    selector->exprID = exprID;
    return JVMDI_ERROR_NONE;
}

jint
eventHandler_setCountSelector(HandlerNode *node, jint index, jint count)
{
    CountSelector *selector = &node->selectors[index].u.Count;
    if (index >= node->selectorCount) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    if (count <= 0) {
        return JDWP_ERROR(INVALID_COUNT);
    } else {
        node->selectors[index].modifier = JDWP_REQUEST_MODIFIER(Count);
        selector->count = count;
        return JVMDI_ERROR_NONE;
    }
}

jint 
eventHandler_setThreadOnlySelector(HandlerNode *node, jint index, jthread thread)
{
    JNIEnv *env = getEnv();
    ThreadSelector *selector = &node->selectors[index].u.ThreadOnly;
    if (index >= node->selectorCount) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    if (node->kind == JVMDI_EVENT_CLASS_UNLOAD) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    } 
    
    /* Create a thread ref that will live beyond */
    /* the end of this call */
    thread = (*env)->NewGlobalRef(env, thread);
    if (thread == NULL) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    node->selectors[index].modifier = JDWP_REQUEST_MODIFIER(ThreadOnly);
    selector->thread = thread;
    return JVMDI_ERROR_NONE;
}

jint 
eventHandler_setLocationOnlySelector(HandlerNode *node, jint index, 
                                     jclass clazz, jmethodID method, jlocation location)
{
    JNIEnv *env = getEnv();
    LocationSelector *selector = &node->selectors[index].u.LocationOnly;
    if (index >= node->selectorCount) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    if ((node->kind != JVMDI_EVENT_BREAKPOINT) && 
        (node->kind != JVMDI_EVENT_FIELD_ACCESS) && 
        (node->kind != JVMDI_EVENT_FIELD_MODIFICATION) && 
        (node->kind != JVMDI_EVENT_SINGLE_STEP) && 
        (node->kind != JVMDI_EVENT_EXCEPTION)) {

        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }

    /* Create a class ref that will live beyond */
    /* the end of this call */
    clazz = (*env)->NewGlobalRef(env, clazz);
    if (clazz == NULL) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    node->selectors[index].modifier = JDWP_REQUEST_MODIFIER(LocationOnly);
    selector->clazz = clazz;
    selector->method = method;
    selector->location = location;
    return JVMDI_ERROR_NONE;
}

jint 
eventHandler_setFieldOnlySelector(HandlerNode *node, jint index, 
                                  jclass clazz, jfieldID field)
{
    JNIEnv *env = getEnv();
    FieldSelector *selector = &node->selectors[index].u.FieldOnly;
    if (index >= node->selectorCount) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    if ((node->kind != JVMDI_EVENT_FIELD_ACCESS) && 
        (node->kind != JVMDI_EVENT_FIELD_MODIFICATION)) {

        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }

    /* Create a class ref that will live beyond */
    /* the end of this call */
    clazz = (*env)->NewGlobalRef(env, clazz);
    if (clazz == NULL) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    node->selectors[index].modifier = JDWP_REQUEST_MODIFIER(FieldOnly);
    selector->clazz = clazz;
    selector->field = field;
    return JVMDI_ERROR_NONE;
}

jint 
eventHandler_setClassOnlySelector(HandlerNode *node, jint index, 
                                  jclass clazz)
{
    JNIEnv *env = getEnv();
    ClassSelector *selector = &node->selectors[index].u.ClassOnly;
    if (index >= node->selectorCount) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    if ((node->kind == JVMDI_EVENT_USER_DEFINED) || 
        (node->kind == JVMDI_EVENT_CLASS_UNLOAD) ||
        (node->kind == JVMDI_EVENT_THREAD_START) ||
        (node->kind == JVMDI_EVENT_THREAD_END)) {

        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }

    /* Create a class ref that will live beyond */
    /* the end of this call */
    clazz = (*env)->NewGlobalRef(env, clazz);
    if (clazz == NULL) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    node->selectors[index].modifier = JDWP_REQUEST_MODIFIER(ClassOnly);
    selector->clazz = clazz;
    return JVMDI_ERROR_NONE;
}

jint 
eventHandler_setExceptionOnlySelector(HandlerNode *node, jint index, 
                                      jclass exceptionClass, jboolean caught, jboolean uncaught)
{
    JNIEnv *env = getEnv();
    ExceptionSelector *selector = &node->selectors[index].u.ExceptionOnly;
    if (index >= node->selectorCount) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    if (node->kind != JVMDI_EVENT_EXCEPTION) { 
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }

    if (exceptionClass != NULL) {
        /* Create a class ref that will live beyond */
        /* the end of this call */
        exceptionClass = (*env)->NewGlobalRef(env, exceptionClass);
        if (exceptionClass == NULL) {
            return JVMDI_ERROR_OUT_OF_MEMORY;
        }
    }
    node->selectors[index].modifier = JDWP_REQUEST_MODIFIER(ExceptionOnly);
    selector->exception = exceptionClass;
    selector->caught = caught;
    selector->uncaught = uncaught;
    return JVMDI_ERROR_NONE;
}

jint 
eventHandler_setClassMatchSelector(HandlerNode *node, jint index, 
                                   char *classPattern)
{
    MatchSelector *selector = &node->selectors[index].u.ClassMatch;
    if (index >= node->selectorCount) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    if ((node->kind == JVMDI_EVENT_USER_DEFINED) || 
        (node->kind == JVMDI_EVENT_THREAD_START) ||
        (node->kind == JVMDI_EVENT_THREAD_END)) {

        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }

    node->selectors[index].modifier = JDWP_REQUEST_MODIFIER(ClassMatch);
    selector->classPattern = classPattern;
    return JVMDI_ERROR_NONE;
}

jint 
eventHandler_setClassExcludeSelector(HandlerNode *node, jint index, 
                                     char *classPattern)
{
    MatchSelector *selector = &node->selectors[index].u.ClassExclude;
    if (index >= node->selectorCount) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    if ((node->kind == JVMDI_EVENT_USER_DEFINED) || 
        (node->kind == JVMDI_EVENT_THREAD_START) ||
        (node->kind == JVMDI_EVENT_THREAD_END)) {

        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }

    node->selectors[index].modifier = JDWP_REQUEST_MODIFIER(ClassExclude);
    selector->classPattern = classPattern;
    return JVMDI_ERROR_NONE;
}

jint 
eventHandler_setStepSelector(HandlerNode *node, jint index, 
                             jthread thread, jint size, jint depth)
{
    jint error;
    JNIEnv *env = getEnv();
    StepSelector *selector = &node->selectors[index].u.Step;
    if (index >= node->selectorCount) {
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    if (node->kind != JVMDI_EVENT_SINGLE_STEP) { 
        return JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }

    /* Create a thread ref that will live beyond */
    /* the end of this call */
    thread = (*env)->NewGlobalRef(env, thread);
    if (thread == NULL) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    error = stepControl_beginStep(thread, size,depth, node);
    if (error != JVMDI_ERROR_NONE) {
        (*env)->DeleteGlobalRef(env, thread);
        return error;
    }
    node->selectors[index].modifier = JDWP_REQUEST_MODIFIER(Step);
    selector->thread = thread;
    selector->depth = depth;
    selector->size = size;
    return JVMDI_ERROR_NONE;
}

/*jboolean 
eventHandler_getStepSelector(jbyte kind, jint handlerID, 
                             jthread *thread, jint *size, jint *depth)
{
    jboolean rc = JNI_FALSE;
    HandlerNode *node;
     
    debugMonitorEnter(handlerLock);

    node= find(kind, handlerID);
    if (node != NULL) {
        Selector *selector = findSelector(node, JDWP_REQUEST_MODIFIER(Step));
        if (selector != NULL) {
            StepSelector *stepSel = &selector->u.Step;
            *thread = stepSel->thread;
            *depth = stepSel->depth;
            *size = stepSel->size;
            rc = JNI_TRUE;
        }
    }

    debugMonitorExit(handlerLock);

    return rc;
}*/


static void 
deferringHandler(JNIEnv *env, JVMDI_Event *event, HandlerNode *node, 
                 struct bag *eventBag)
{
    /*
     * This handler is really just a way to make sure that the event
     * is handled on the next occurrence of a particular event.
     */
    eventHandler_freeInternal(node);
}

static jboolean
insertDeferringBreakpoint(JNIEnv *env, jthread thread, 
                       jclass clazz, jmethodID method, jlocation location) 
{
    HandlerNode *node;

    node = eventHandler_insertInternalBreakpoint(deferringHandler, 
                               thread, clazz, method, location);

    return (node != NULL);
}

static jboolean
deferEventReport(JNIEnv *env, JVMDI_Event *event, jthread thread)
{ 
    jboolean deferring = JNI_FALSE;

    /*
     * JVMDI event ordering must be supported without bugs in order to 
     * defer reported events.
     */
    if (version_supportsEventOrdering()) {
        switch (event->kind) {
            /*
             * Do things the easy way for now. Insert a breakpoint and 
             * defer the reporting of these events until we get
             * to the breakpoint. This could be done more efficiently
             * by deferring only if there's a breakpoint already set,
             * but that approach would also require work to make sure that
             * for any breakpoint that is set after these events are 
             * reported and before the thread continues, that
             * breakpoint is ignored the first time it is seen.
             */
            case JVMDI_EVENT_METHOD_ENTRY: {
                JVMDI_frame_event_data *entryEvent = &event->u.frame;
                if (!isMethodNative(entryEvent->clazz, entryEvent->method)) {
                    jint error;
                    jlocation start;
                    jlocation end;
                    error = jvmdi->GetMethodLocation(entryEvent->clazz, 
                                                     entryEvent->method,
                                                     &start, &end);
                    if (error == JVMDI_ERROR_NONE) {
                        deferring = insertDeferringBreakpoint(env, 
                                                   thread, entryEvent->clazz,
                                                   entryEvent->method, start);
                    }
                }
    
                break;
            }
            case JVMDI_EVENT_SINGLE_STEP: {
                JVMDI_single_step_event_data *stepEvent =
                    &event->u.single_step;
                deferring = insertDeferringBreakpoint(env,
                                                   thread, stepEvent->clazz,
                                                   stepEvent->method, 
                                                   stepEvent->location);
                break;
            }
            /*
             * TO DO: Once JVMDI supports a way to know if we're at the end
             * of a method, we should check here for break and 
             * step events which precede a method exit event.
             */
        }
    }
    return deferring;
}

static void 
reportEvents(JNIEnv *env, jbyte sessionID, jthread thread, 
             JVMDI_Event *event, struct bag *eventBag)
{
    jbyte suspendPolicy;
    jboolean invoking;

    if (bagSize(eventBag) < 1) {
        return;
    }

    /*
     * Never report events before initialization completes
     */
    if (!debugInit_isInitComplete()) {
        return;
    }

    /*
     * We delay the reporting of some events so that they can be properly
     * grouped into event sets with upcoming events. If the reporting is 
     * to be deferred, the event commands remain in the event bag until 
     * a subsequent event occurs. 
     * Event is NULL for synthetic events (e.g. unload).
     */
    if (event == NULL || !deferEventReport(env, event, thread)) {
        struct bag *completedBag = bagDup(eventBag);
        bagDeleteAll(eventBag);
        if (completedBag == NULL) {
            /*
             * TO DO: Report, but don't terminate?
             */
        } else {
            suspendPolicy = eventHelper_reportEvents(sessionID, completedBag);
            if (thread != NULL && suspendPolicy != JDWP_SuspendPolicy_NONE) {
                do {
                    /*
                     * The events have been reported and this thread is about to 
                     * continue, but it may have been started up up just to perform
                     * a requested method invocation. If so, we do the invoke now
                     * and then stop again waiting for another continue. By then
                     * another invoke request can be in place, so there is a loop
                     * around this code.
                     */
                    invoking = invoker_doInvoke(thread);
                    if (invoking) {
                        eventHelper_reportInvokeDone(sessionID, thread);
                    }
                } while (invoking);
            }
            bagDestroyBag(completedBag);
        }
    }
}

static void 
handleClassPrepare(JNIEnv *env, JVMDI_Event *event, HandlerNode *node, 
                   struct bag *eventBag)
{
    jthread thread = event->u.class_event.thread;
    jclass clazz = event->u.class_event.clazz;
    jbyte suspendPolicy;

    /*
     * We try hard to avoid class loads/prepares in debugger threads, but
     * it is still possible for them to happen (most likely for exceptions
     * that are thrown within JNI methods). If such an event
     * occurs, we must report it, but we cannot suspend the debugger thread.
     * 
     * 1) We report the thread as NULL because we don't want the application
     *    to get hold of a debugger thread object. 
     * 2) We try to do the right thing wrt to suspending threads without
     *    suspending debugger threads. If the requested suspend policy is 
     *    NONE, there's no problem. If the requested policy is ALL, we can
     *    just suspend all application threads without producing any 
     *    surprising results by leaving the debugger thread running. However,
     *    if the requested policy is EVENT_THREAD, we are forced to 
     *    do something different than requested. The most useful behavior
     *    is to suspend all application threads (just as if the policy
     *    was ALL). This allows the application to operate on the class
     *    before it gets into circulation and so it is preferable to the 
     *    other alternative of suspending no threads.
     */
    suspendPolicy = node->suspendPolicy;
    if (threadControl_isDebugThread(thread)) { 
        event->u.class_event.thread = NULL;
        if (suspendPolicy == JDWP_SuspendPolicy_EVENT_THREAD) {
            suspendPolicy = JDWP_SuspendPolicy_ALL;
        }
    }
    eventHelper_recordEvent(event, node->nodeID, 
                            suspendPolicy, eventBag);
}

static void 
handleClassPrepareTracking(JNIEnv *env, JVMDI_Event *event, HandlerNode *node, 
                   struct bag *eventBag)
{
    jclass clazz = event->u.class_event.clazz;

    /*
     * Record that the prepare occurred for class tracking purposes.
     */
    classTrack_addPreparedClass(env, clazz);
}


static void 
handleClassUnload(JNIEnv *env, JVMDI_Event *event, HandlerNode *node,
                  struct bag *eventBag)
{
    JDI_ASSERT_MSG(JNI_FALSE, "Should never call handleClassUnload");
}

static void 
handleFrameEvent(JNIEnv *env, JVMDI_Event *event, HandlerNode *node,
                  struct bag *eventBag)
{
    /*
     * The frame id that comes with this event is very transient.
     * We can't send the frameID to the helper thread because it
     * might be useless by the time the helper thread can use it 
     * (if suspend policy is NONE). So, get the needed info from 
     * the frame and then use a special command to the helper
     * thread.
     */
    
    jclass clazz;
    jmethodID method;
    jlocation location;
    JVMDI_frame_event_data *eventData = &event->u.frame;
    jint error;
    
    error = threadControl_getFrameLocation(eventData->thread, 
                                           eventData->frame,
                                           &clazz, &method, &location);
    if (error == JVMDI_ERROR_NONE) {
        (*env)->DeleteGlobalRef(env, clazz);
    } else {
        location = -1;
    }

    eventHelper_recordFrameEvent(node->nodeID, node->suspendPolicy,
                                 (jbyte)event->kind, 
                                 eventData->thread, eventData->clazz, 
                                 eventData->method, location, eventBag);
}

static void 
handleVMDeath(JNIEnv *env, JVMDI_Event *event, HandlerNode *node, 
              struct bag *eventBag)
{
    eventHelper_recordEvent(event, node->nodeID, JNI_FALSE, eventBag);
}

static void 
handleUserDefined(JNIEnv *env, JVMDI_Event *event, HandlerNode *node, 
                  struct bag *eventBag)
{
}

static void  
genericHandler(JNIEnv *env, JVMDI_Event *event, HandlerNode *node,  
               struct bag *eventBag)
{ 
    eventHelper_recordEvent(event, node->nodeID,  
                            node->suspendPolicy, eventBag); 
}

static EventHandler
defaultEventHandler(jint kind) 
{
    switch (kind) {
        case JVMDI_EVENT_BREAKPOINT:
        case JVMDI_EVENT_EXCEPTION:
        case JVMDI_EVENT_FIELD_ACCESS:
        case JVMDI_EVENT_FIELD_MODIFICATION:
        case JVMDI_EVENT_SINGLE_STEP:
        case JVMDI_EVENT_THREAD_START:
        case JVMDI_EVENT_THREAD_END:
            return genericHandler;

        case JVMDI_EVENT_USER_DEFINED: 
            return handleUserDefined;

        case JVMDI_EVENT_CLASS_PREPARE: 
            return handleClassPrepare;

        case JVMDI_EVENT_CLASS_UNLOAD: 
            return handleClassUnload;

        case JVMDI_EVENT_VM_DEATH: 
            return handleVMDeath;

        case JVMDI_EVENT_METHOD_ENTRY:
        case JVMDI_EVENT_METHOD_EXIT:
            return handleFrameEvent;

        default:
            ERROR_MESSAGE_EXIT("Attempt to install handler for invalid event");
            return NULL;
    }
}

/* 
 * A bagEnumerateFunction.
 * Create a synthetic class unload event for every class no longer present.
 * Analogous to eventHook combined with a handler in a unload specific 
 * (no event structure) kind of way.
 */
static jboolean
synthesizeUnloadEvent(void *signatureVoid, void *envVoid)
{
    JNIEnv *env = (JNIEnv *)envVoid;
    char *signature = *(char **)signatureVoid;
    char *classname;
    HandlerNode *node;
    jbyte eventSessionID;    
    struct bag *eventBag = eventHelper_createEventBag();

    if (eventBag == NULL) {
        /* TO DO: Report, but don't die
         */
        JDI_ASSERT(eventBag != NULL);
    }

    /* Signature needs to last, so convert extra copy to classname */
    classname = jdwpAlloc(strlen(signature)+1);
    strcpy(classname, signature);
    convertSignatureToClassname(classname);

    debugMonitorEnter(handlerLock);
    
    eventSessionID = currentSessionID;
    node = handlers[JVMDI_EVENT_CLASS_UNLOAD].first;
    while (node != NULL) {
        HandlerNode *temp = node->next; /* allows handlers to remove themselves */
        if (unloadSelectorFilter(env, classname, node)) {
            /*
             * There may be multiple handlers, the signature will be freed
             * when the event helper thread has written it.  So each event
             * needs a separate allocation.
             */
            char *durableSignature = jdwpAlloc(strlen(signature)+1);
            strcpy(durableSignature, signature);

            eventHelper_recordClassUnload(node->nodeID, durableSignature, eventBag);
        }
        node = temp;
    }
    
    debugMonitorExit(handlerLock);
    
    if (eventBag != NULL) {
        reportEvents(env, eventSessionID, NULL, NULL, eventBag);

        /*
         * bag was created locally, destroy it here.
         */
        bagDestroyBag(eventBag);
    }

    jdwpFree(signature);
    jdwpFree(classname);

    return JNI_TRUE;
}

/*
 * The JVMDI event hook. Each event is passed to a sequence of handlers in 
 * a chain until the chain ends or one handler consumes the event. Expensive
 * or complex events are handled in a separate thread, but for now, we wait
 * for the event thread to complete its handing before continuing.
 */
static void 
eventHook(JNIEnv *env, JVMDI_Event *event)
{
    static unsigned int eventCount = 0;
    static unsigned int classUnloadsPending = 0;
    struct bag *eventBag = NULL;
    jthread thread;
    jthrowable currentException;

    /* It is not safe to process unload events as they come in, 
     * since the VM will be locked down in GC. 
     * We note that they have occurred so that we can synthesize
     * them the next time a non-unload event comes in.
     * We don't dare grab locks during the GC, so reads and
     * writes to classUnloadsPending are not protected, see the
     * code that reads it (later in this function).
     */
    if (event->kind == JVMDI_EVENT_CLASS_UNLOAD) {
        ++classUnloadsPending;
        return;
    }

    /*
     * Workaround bug 4199400 in classic VM by avoiding class load events 
     * completely. Any of the lock grabbing below can result in a 
     * deadlock if we don't do this.
     */
    if (!version_supportsClassLoadEvents() && 
        event->kind == JVMDI_EVENT_CLASS_LOAD) {
        return;
    }

    /*
     * We want to preserve any current exception that might get wiped
     * out during event handling (e.g. JNI calls). We have to rely on
     * space for the local reference on the current frame because
     * doing a PushLocalFrame here might itself generate an exception. 
     */
    currentException = (*env)->ExceptionOccurred(env);
    (*env)->ExceptionClear(env);

    /*
     * Periodically, we want to compact the hash table of all 
     * objects sent to the front end by removing objects that 
     * have been collected. For now, this is done in a very primitive
     * fashion (every N events). In the future we may be able to use 
     * JVMDI events for garbage collection to trigger the compaction. 
     * If an unload happenned we are sure to need a compaction.     
     */
    if (((1 + (eventCount++)) % 500) == 0 || classUnloadsPending > 0) {
        commonRef_compact();
    }

    /*
     * Process any waiting unload events.
     *
     * "if" is an optimization to avoid entering the lock on every 
     * event; classUnloadsPending may be zapped before we enter
     * the lock but then this just becomes one big no-op.
     */
    if (classUnloadsPending > 0) {
        struct bag *unloadedSignatures;

        /* this lock won't do anything to prevent the
         * classUnloadsPending count from incrementing, but it
         * will prevent another event thread from blasting it
         * before it is read. It also allows classTrack to
         * analyze which classes have been unloaded without
         * class prepare events coming in.
         */
        debugMonitorEnter(handlerLock);
        /* the classUnloadsPending count could be incremented
         * here, and we would lose the count when it is set
         * to zero, however, these freshly unloaded classes 
         * will be seen when the unloaded classes are analyzed,
         * and nothing will be lost.
         */
        classUnloadsPending = 0;
        /* analyze which unloads occurred
         */
        unloadedSignatures = classTrack_processUnloads(env);
        debugMonitorExit(handlerLock);

        /* generate the synthetic events.
         */
        bagEnumerateOver(unloadedSignatures, synthesizeUnloadEvent, 
                         (void *)env);

        bagDestroyBag(unloadedSignatures);
    }

    thread = eventThread(event);

    /*
     * Record the fact that we're entering an event handler so that
     * thread operations (status, interrupt, stop) can be done correctly
     * and so that thread resources can be allocated.
     * This must be done before grabbing any locks.
     */
    if (thread != NULL) {
        eventBag = threadControl_onEventHandlerEntry(currentSessionID, 
                                                     event->kind, thread);
    } else {
        eventBag = eventHelper_createEventBag();
        if (eventBag == NULL) {
            /*
             * TO DO: Report, but don't die
             */
        }
    }

    {
        HandlerNode *node;
        jbyte eventSessionID;

        debugMonitorEnter(handlerLock);

        eventSessionID = currentSessionID;
        node = handlers[event->kind].first;
        while (node != NULL) {
            HandlerNode *temp = node->next; /* allows handlers to remove themselves */
            if (selectorFilter(env, event, node)) {
                (node->handler)(env, event, node, eventBag);
            }
            node = temp;
        }

        debugMonitorExit(handlerLock);

        if (eventBag != NULL) {
            reportEvents(env, eventSessionID, thread, event, eventBag);
        }
    }

    /*
     * If the bag was created locally, destroy it here.
     */
    if (thread == NULL) {
        bagDestroyBag(eventBag);
    }

    /*
     * Always restore any exception that was set beforehand.  If there is a
     * pending async exception, StopThread will be called from             
     * threadControl_onEventHandlerExit immediately below.  Depending on VM
     * implementation and state, the async exception might immediately     
     * overwrite the currentException, or it might be delayed until later. 
     */
    if (currentException != NULL) {
        (*env)->Throw(env, currentException);
    } else {
        (*env)->ExceptionClear(env);
    }

    /*
     * Release thread resources and perform any delayed operations.
     */
    if (thread != NULL) {
        threadControl_onEventHandlerExit(event->kind, thread, eventBag);
    }
}

HandlerNode *
eventHandler_alloc(jint selectorCount, jbyte kind, jbyte suspendPolicy) 
{
    HandlerNode *node = jdwpAlloc(sizeof(HandlerNode) +
                                 (selectorCount-1) * sizeof(Selector));
    if (node != NULL) {
        node->handler = NULL;  /* so generic handler can be found */
        node->suspendPolicy = JDWP_SuspendPolicy_ALL;
        node->selectorCount = selectorCount;
        /* make eventHandler_free() safe */
        node->kind = kind;
        node->suspendPolicy = suspendPolicy;
        node->next = NULL;
        node->prev = NULL;
        node->chain = NULL;
    }
    return node;
}

static jint
setBreakpoint(HandlerNode *node)
{
    jint error = JVMDI_ERROR_NONE;
    Selector *select = findSelector(node, JDWP_REQUEST_MODIFIER(LocationOnly));
    if (select == NULL) {
        error = JVMDI_ERROR_INTERNAL; /* bp event with no location selector */
    } else {
        jclass clazz = select->u.LocationOnly.clazz;
        jmethodID method = select->u.LocationOnly.method;
        jlocation location = select->u.LocationOnly.location;
        /* if this is the first handler for this */
        /* location, set bp at jvmdi level */
        if (countBreakpointHandlers(clazz, method, location) == 0) {
            error = jvmdi->SetBreakpoint(clazz, method, location);
        }
    }
    return error;
}

static jint
clearBreakpoint(HandlerNode *node)
{
    jint error = JVMDI_ERROR_NONE;
    Selector *select = findSelector(node, JDWP_REQUEST_MODIFIER(LocationOnly));
    if (select == NULL) {
        error = JVMDI_ERROR_INTERNAL; /* bp event with no location selector */
    } else {
        jclass clazz = select->u.LocationOnly.clazz;
        jmethodID method = select->u.LocationOnly.method;
        jlocation location = select->u.LocationOnly.location;
        /* if this is the last handler for this */
        /* location, set bp at jvmdi level */
        if (countBreakpointHandlers(clazz, method, location) == 0) {
            error = jvmdi->ClearBreakpoint(clazz, method, location);
        }
    }
    return error;
}

static jint
setWatchpoint(HandlerNode *node)
{
    jint error = JVMDI_ERROR_NONE;
    Selector *select = findSelector(node, JDWP_REQUEST_MODIFIER(FieldOnly));
    if (select == NULL) {
        error = JVMDI_ERROR_INTERNAL; /* event with no field selector */
    } else {
        jclass clazz = select->u.FieldOnly.clazz;
        jfieldID field = select->u.FieldOnly.field;
        /* if this is the first handler for this */
        /* field, set wp at jvmdi level */
        if (countWatchpointHandlers(node->kind, clazz, field) == 0) {
            error = (node->kind == JVMDI_EVENT_FIELD_ACCESS) ?
                    jvmdi->SetFieldAccessWatch(clazz, field) :
                    jvmdi->SetFieldModificationWatch(clazz, field);
        }
    }
    return error;
}

static jint
clearWatchpoint(HandlerNode *node)
{
    jint error = JVMDI_ERROR_NONE;
    Selector *select = findSelector(node, JDWP_REQUEST_MODIFIER(FieldOnly));
    if (select == NULL) {
        error = JVMDI_ERROR_INTERNAL; /* event with no field selector */
    } else {
        jclass clazz = select->u.FieldOnly.clazz;
        jfieldID field = select->u.FieldOnly.field;
        /* if this is the first handler for this */
        /* field, clear wp at jvmdi level */
        if (countWatchpointHandlers(node->kind, clazz, field) == 0) {
            error = (node->kind == JVMDI_EVENT_FIELD_ACCESS) ?
                    jvmdi->ClearFieldAccessWatch(clazz, field) :
                    jvmdi->ClearFieldModificationWatch(clazz, field);
        }
    }
    return error;
}

static jint
enableEvents(HandlerNode *node)
{
    jint kind = node->kind;
    jint error = JVMDI_ERROR_NONE;

    switch (kind) {
        /*
         * The stepping code directly enables/disables stepping as necessary
         */
        case JVMDI_EVENT_SINGLE_STEP:
        /*
         * Internal thread event handlers are always present (hardwired in
         * the event hook), so we don't change the notification mode here.
         */
        case JVMDI_EVENT_THREAD_START:
        case JVMDI_EVENT_THREAD_END:
            return error;

        case JVMDI_EVENT_FIELD_ACCESS:
        case JVMDI_EVENT_FIELD_MODIFICATION:
            error = setWatchpoint(node);
            break;

        case JVMDI_EVENT_BREAKPOINT:
            error = setBreakpoint(node);
            break;
    }

    if (error == JVMDI_ERROR_NONE) {
        jthread thread = requestThread(node);
        if (countThreadHandlers(kind, thread) == 0) {
            error = threadControl_setEventMode(JVMDI_ENABLE, kind, thread);
        }
    }
    return error;
}

static jint
disableEvents(HandlerNode *node)
{
    jint kind = node->kind;
    jint error = JVMDI_ERROR_NONE;

    switch (kind) {
        /*
         * The stepping code directly enables/disables stepping as necessary
         */
        case JVMDI_EVENT_SINGLE_STEP:
        /*
         * Internal thread event handlers are always present (hardwired in
         * the event hook), so we don't change the notification mode here.
         */
        case JVMDI_EVENT_THREAD_START:
        case JVMDI_EVENT_THREAD_END:
            return error;

        case JVMDI_EVENT_FIELD_ACCESS:
        case JVMDI_EVENT_FIELD_MODIFICATION:
            error = clearWatchpoint(node);
            break;

        case JVMDI_EVENT_BREAKPOINT:
            error = clearBreakpoint(node);
            break;
    }

    if (error == JVMDI_ERROR_NONE) {
        jthread thread = requestThread(node);
        if (countThreadHandlers(kind, thread) == 0) {
            error = threadControl_setEventMode(JVMDI_DISABLE, kind, thread);
        }
    }
    return error;
}

jint
eventHandler_insert(HandlerNode *node) 
{
    jint error;

    debugMonitorEnter(handlerLock);

    if (node->handler == NULL) {
        node->handler = defaultEventHandler(node->kind);
    }
    node->nodeID = ++requestIdCounter;
    error = enableEvents(node);
    if (error == JVMDI_ERROR_NONE) {
        insert(&handlers[node->kind], node);
    }

    debugMonitorExit(handlerLock);

    return error;
}

jint eventHandler_id(HandlerNode *node)
{
    return node->nodeID;
}

/**
 * Utility for programatically set zero-selector handlers
 */
HandlerNode * 
eventHandler_insertInternal(jbyte kind, EventHandler handler, jthread thread)
{
    JNIEnv *env = getEnv();
    jint error = JVMDI_ERROR_NONE;

    /*
     * Start with necessary allocations 
     */
    HandlerNode *node = eventHandler_alloc(thread == NULL ? 0 : 1,
                                           kind, JDWP_SuspendPolicy_NONE);
    if (node == NULL) {
        return NULL;
    }

    if (thread != NULL) {
        error = eventHandler_setThreadOnlySelector(node, 0, thread);
    }

    if (error == JVMDI_ERROR_NONE) {
        /*
         * Insert the new handler node
         */
        node->handler = handler;
        eventHandler_insert(node);
    
        /* Reset the node id assigned in eventHandler_insert to indicate
           no specific request */
        node->nodeID = 0;
    }

    if (error != JVMDI_ERROR_NONE) {
        eventHandler_freeInternal(node);
        node = NULL;
    }
    return node;
}

HandlerNode *
eventHandler_insertInternalBreakpoint(EventHandler handler, jthread thread,
                                     jclass clazz, jmethodID method, 
                                     jlocation location)
{
    JNIEnv *env = getEnv();
    jint index = 0;
    jint error = JVMDI_ERROR_NONE;

    /*
     * Start with necessary allocations 
     */
    HandlerNode *node = eventHandler_alloc((thread == NULL) ? 1 : 2,
                                            JVMDI_EVENT_BREAKPOINT,
                                            JDWP_SuspendPolicy_NONE);
    if (node == NULL) {
        return NULL;
    }

    if (thread != NULL) {
        error = eventHandler_setThreadOnlySelector(node, index++, thread);
    }

    if (error == JVMDI_ERROR_NONE) {
        error = eventHandler_setLocationOnlySelector(node, index++, clazz, 
                                             method, location);
        if (error == JVMDI_ERROR_NONE) {
            node->handler = handler;

            /*
             * Insert the new handler node
             */
            eventHandler_insert(node);
        
            /* Reset the node id assigned in eventHandler_insert to indicate
               no specific request */
            node->nodeID = 0;
        }
    }

    if (error != JVMDI_ERROR_NONE) {
        eventHandler_freeInternal(node);
        node = NULL;
    }
    return node;
}

/**
 * Deinsert and free memory.  Safe for non-inserted nodes.
 */
jint 
eventHandler_freeInternal(HandlerNode *node)
{
    jint error = JVMDI_ERROR_NONE;

    debugMonitorEnter(handlerLock);

    if (node != NULL) {
        /* 
         * deinsert the handler node before disableEvents() to make sure
         * the event will be disabled when no other event handlers are installed.
         */
        deinsert(node);
        error = disableEvents(node);
        if (error == JVMDI_ERROR_NONE) {
            clearSelectors(node);
        }
        jdwpFree(node);
    }

    debugMonitorExit(handlerLock);

    return error;
}

static void
becomeDefunct(HandlerNode *node) {
    deinsert(node);
    insert(&defunct_handlers[node->kind], node);
    if (disableEvents(node) == JVMDI_ERROR_NONE) {
        /*
         * Only clear the selectors if the disable worked.
         * This allows another try when the clear command comes
         * in and an error can be reported then.
         */
        clearSelectors(node);
    }
}

static jint
freeAll(HandlerChain *chain) 
{
    jint error = JVMDI_ERROR_NONE;
    HandlerNode *node = chain->first;
    while (node != NULL) {
        HandlerNode *temp = node->next;
        if (node->nodeID != 0) {
            error = eventHandler_freeInternal(node);
            if (error != JVMDI_ERROR_NONE) {
                break;
            }
        }
        node = temp;
    }
    return error;
}

jint
eventHandler_freeAll(jint kind) 
{
    jint error;
    debugMonitorEnter(handlerLock);
    error = freeAll(&handlers[kind]);
    if (error != JVMDI_ERROR_NONE) {
        error = freeAll(&defunct_handlers[kind]);
    }
    debugMonitorExit(handlerLock);
    return error;
}

jint
eventHandler_free(jint kind, jint handlerID)
{
    jint error;
    HandlerNode *node;

    debugMonitorEnter(handlerLock);
    node = find(kind, handlerID);
    if (node != NULL) {
        error = eventHandler_freeInternal(node);
    } else {
        error = JVMDI_ERROR_ILLEGAL_ARGUMENT;
    }
    debugMonitorExit(handlerLock);
    return error;
}

void 
eventHandler_initialize(jbyte sessionID)
{
    jint error;

    requestIdCounter = 1;
    currentSessionID = sessionID;

    handlerLock = debugMonitorCreate("JDWP Event Handler Lock");

    /*
     * Permanently enabled events.
     */
    error = threadControl_setEventMode(JVMDI_ENABLE, 
                                      JVMDI_EVENT_THREAD_START, NULL);
    if (error != JVMDI_ERROR_NONE) {
        ERROR_MESSAGE_EXIT("Unable to enable thread start events");
    }
    error = threadControl_setEventMode(JVMDI_ENABLE, 
                                       JVMDI_EVENT_THREAD_END, NULL);
    if (error != JVMDI_ERROR_NONE) {
        ERROR_MESSAGE_EXIT("Unable to enable thread end events");
    }

    error = jvmdi->SetEventHook(eventHook);
    if (error != JVMDI_ERROR_NONE) {
        ERROR_MESSAGE_EXIT("Unable to set event hook");
    }
            
    /* Notify other modules that the event hook is in place */
    threadControl_onHook();

    /* Get the event helper thread initialized */
    eventHelper_initialize(sessionID);
}

void 
eventHandler_onConnect()
{
    jboolean installed;

    debugMonitorEnter(handlerLock);
    /* always report VMDeath to a connected debugger */
    installed = (eventHandler_insertInternal(JVMDI_EVENT_VM_DEATH, 
                                     handleVMDeath, NULL) != NULL);
    if (!installed) {
        ERROR_MESSAGE_EXIT("Unable to install VM Death event handler");
    }
    /* for class tracking need internal class prepare handler */
    installed = (eventHandler_insertInternal(JVMDI_EVENT_CLASS_PREPARE, 
                               handleClassPrepareTracking, NULL) != NULL);
    if (!installed) {
        ERROR_MESSAGE_EXIT("Unable to install Class Prepare tracking event handler");
    }
    debugMonitorExit(handlerLock);
}

void 
eventHandler_onDisconnect()
{
    HandlerChain *chain;
    HandlerNode *node;

    debugMonitorEnter(handlerLock);

    chain = &handlers[JVMDI_EVENT_VM_DEATH]; 
    node = chain->first;
    while (node != NULL) {
        if (node->handler == handleVMDeath) {
            eventHandler_freeInternal(node);
            break;
        }
    }

    debugMonitorExit(handlerLock);
}



void 
eventHandler_reset(jbyte sessionID)
{
    int i;

    debugMonitorEnter(handlerLock);

    /*
     * We must do this first so that if any invokes complete, there
     * will be no attempt to send them to the front end. Waiting
     * for threadControl_reset leaves a window where the invoke completions
     * can sneak through.
     */
    threadControl_detachInvokes();

    /*
     * Reset the event helper thread, purging all queued and in-process
     * commands.
     */
    eventHelper_reset(sessionID);

    for (i = 0; i <= JVMDI_MAX_EVENT_TYPE_VAL; i++) {
        /*
         * Free all handlers 
         */
        HandlerChain *chain = &handlers[i]; 
        HandlerNode *node = chain->first;
        while (node != NULL) {
            HandlerNode *temp = node->next;
            eventHandler_freeInternal(node);
            node = temp;
        }
        JDI_ASSERT(chain->first == NULL);

        /*
         * Defunct handlers are session-specific, so they all go
         */
        freeAll(&defunct_handlers[i]);
    }

    requestIdCounter = 1;
    currentSessionID = sessionID;

    debugMonitorExit(handlerLock);
}

void eventHandler_lock()
{
    debugMonitorEnter(handlerLock);
}

void eventHandler_unlock()
{
    debugMonitorExit(handlerLock);
}


