/*
 * @(#)ArrayReference.java	1.15 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jdi;

import java.util.List;

/**
 * Provides access to an array object and its components in the target VM.
 * Each array component is mirrored by a {@link Value} object.
 * The array components, in aggregate, are placed in {@link java.util.List}
 * objects instead of arrays for consistency with the rest of the API and 
 * for interoperability with other APIs.
 *
 * @author Robert Field
 * @author Gordon Hirsch
 * @author James McIlree
 * @since  1.3
 */
public interface ArrayReference extends ObjectReference {

    /**
     * Returns the number of components in this array.
     *
     * @return the integer count of components in this array.
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    int length();

    /**
     * Returns an array component value.
     * 
     * @param index the index of the component to retrieve
     * @return the {@link Value} at the given index.
     * @throws java.lang.IndexOutOfBoundsException if 
     * the index is beyond the end of this array.
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    Value getValue(int index);

    /**
     * Returns all of the components in this array.
     *
     * @return a list of {@link Value} objects, one for each array
     * component ordered by array index.
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    List getValues();

    /**
     * Returns a range of array components.
     *
     * @param index the index of the first component to retrieve
     * @param length the number of components to retrieve, or -1 to 
     * retrieve all components to the end of this array.
     * @return a list of {@link Value} objects, one for each requested 
     * array component ordered by array index.
     * @throws java.lang.IndexOutOfBoundsException if 
     * <code>index + length</code> is an index beyond the end of this array.
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    List getValues(int index, int length);

    /**
     * Replaces an array component with another value.
     * <p>
     * Object values must be assignment compatible with the component type
     * (This implies that the component type must be loaded through the
     * enclosing class's class loader). Primitive values must be 
     * either assignment compatible with the component type or must be
     * convertible to the component type without loss of information. 
     * See JLS section 5.2 for more information on assignment
     * compatibility.
     *
     * @param value the new value 
     * @param index the index of the component to set
     * @throws java.lang.IndexOutOfBoundsException if 
     * <code>index</code> is beyond the end of this array.
     * @throws InvalidTypeException if the type <code>value</code>
     * is not compatible with the declared type of array components.
     * @throws ClassNotLoadedException if the array component type has not yet been loaded
     * through the appropriate class loader. 
     * @throws ObjectCollectedException if this object or the 
     * new value has been garbage collected.
     * @throws VMMismatchException if a {@link Mirror} argument and this object
     * do not belong to the same {@link VirtualMachine}.
     */
    void setValue(int index, Value value)
            throws InvalidTypeException,
                   ClassNotLoadedException;
    
    /**
     * Replaces all array components with other values. If the given
     * list is larger is size than the array, the values at the 
     * end of the list are ignored.
     * <p>
     * Object values must be assignment compatible with the element type
     * (This implies that the component type must be loaded through the
     * enclosing class's class loader). Primitive values must be 
     * either assignment compatible with the component type or must be
     * convertible to the component type without loss of information. 
     * See JLS section 5.2 for more information on assignment
     * compatibility.
     *
     * @param values a list of {@link Value} objects to be placed 
     * in this array
     * @throws InvalidTypeException if any of the <code>values</code>
     * is not compatible with the declared type of array components.
     * @throws java.lang.IndexOutOfBoundsException if the size of
     * <code>values</code> is larger than the length of this array.
     * @throws ClassNotLoadedException if the array component type has not yet been loaded
     * through the appropriate class loader. 
     * @throws ObjectCollectedException if this object or any 
     * of the new values has been 
     * garbage collected.
     * @throws VMMismatchException if a {@link Mirror} argument and this object
     * do not belong to the same {@link VirtualMachine}.
     */
    void setValues(List values)
            throws InvalidTypeException,
                   ClassNotLoadedException;

    /**
     * Replaces a range of array components with other values. 
     * <p>
     * Object values must be assignment compatible with the component type
     * (This implies that the component type must be loaded through the
     * enclosing class's class loader). Primitive values must be 
     * either assignment compatible with the component type or must be
     * convertible to the component type without loss of information. 
     * See JLS section 5.2 for more information on assignment
     * compatibility.
     *
     * @param index the index of the first component to set.
     * @param values a list of {@link Value} objects to be placed 
     * in this array.
     * @param srcIndex the index of the first source value to to use.
     * @param length the number of components to set, or -1 to set 
     * all components to the end of this array.
     * @throws InvalidTypeException if any element of <code>values</code>
     * is not compatible with the declared type of array components.
     * @throws java.lang.IndexOutOfBoundsException if
     * <code>srcIndex + length</code> is beyond the end of this array or if
     * <code>values</code> is smaller inside than the given range.
     * @throws ObjectCollectedException if this object or any
     * of the new values has been 
     * garbage collected.
     * @throws VMMismatchException if a {@link Mirror} argument and this object
     * do not belong to the same {@link VirtualMachine}.
     */
    void setValues(int index, List values, int srcIndex, int length)
            throws InvalidTypeException,
                   ClassNotLoadedException;
}

