/*
 * @(#)LocalVariable.java	1.16 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jdi;

/**
 * A local variable in the target VM. Each variable declared within a 
 * {@link Method} has its own LocalVariable object. Variables of the same
 * name declared in different scopes have different LocalVariable objects.
 * LocalVariables can be used alone to retrieve static information
 * about their declaration, or can be used in conjunction with a 
 * {@link StackFrame} to set and get values.
 *
 * @see StackFrame
 * @see Method
 *
 * @author Robert Field
 * @author Gordon Hirsch
 * @author James McIlree
 * @since  1.3
 */

public interface LocalVariable extends Mirror, Comparable {

    /**
     * Gets the name of the local variable.
     *
     * @return a string containing the name.
     */
    String name();

    /**
     * @return a text representation of the declared type
     * of this variable.
     */
    String typeName();

    /**
     * Returns the type of the this {@link LocalVariable}. For primitive variables,
     * an appropriate {@link PrimitiveType} is always returned. For object 
     * variables, the appropriate {@link ReferenceType} is returned if it has 
     * been loaded through the enclosing type's class loader.
     *
     * @return the {@link Type} of this local variable.
     * @throws ClassNotLoadedException if the type has not yet been loaded
     * through the appropriate class loader. 
     */
    Type type() throws ClassNotLoadedException;

    /**
     * Gets the JNI signature of the local variable.
     *
     * @see <a href="doc-files/signature.html">Type Signatures</a>
     * @return a string containing the signature.
     */
    String signature();

    /**
     * Determines whether this variable can be accessed from the given
     * {@link StackFrame}.
     * 
     * See {@link StackFrame#visibleVariables} for a complete description
     * variable visibility in this interface.
     *
     * @param frame the StackFrame querying visibility
     * @return <code>true</code> if this variable is visible; 
     * <code>false</code> otherwise.
     * @throws IllegalArgumentException if the stack frame's method
     * does not match this variable's method.
     * @throws VMMismatchException if a {@link Mirror} argument and this mirror
     * do not belong to the same {@link VirtualMachine}.
     */
    boolean isVisible(StackFrame frame);

    /**
     * Determines if this variable is an argument to its method.
     *
     * @return <code>true</code> if this variable is final; 
     * <code>false</code> otherwise.
     */
    boolean isArgument();

    /**
     * Compares the specified Object with this LocalVariable for equality.
     *
     * @return  true if the Object is a LocalVariable, if both LocalVariables
     * are contained in the same method (as determined by 
     * {@link Method#equals}), and if both LocalVariables mirror
     * the same declaration within that method
     */
    boolean equals(Object obj);

    /**
     * Returns the hash code value for this LocalVariable.
     * 
     * @return the integer hash code
     */
    int hashCode();
}

