/*
 * @(#)Location.java	1.22 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jdi;

/**
 * A point within the executing code of the target VM.
 * Locations are used to identify the current position of 
 * a suspended thread (analogous to an instruction pointer or 
 * program counter register in native programs). They are also used
 * to identify the position at which to set a breakpoint.
 * <p>
 * The availability of a line number for a location will
 * depend on the level of debugging information available from the 
 * target VM.
 * <p>
 * Several mirror interfaces have locations. Each such mirror extends a
 * {@link Locatable} interface.
 *
 * @see com.sun.jdi.request.EventRequestManager
 * @see StackFrame
 * @see com.sun.jdi.event.BreakpointEvent
 * @see com.sun.jdi.event.ExceptionEvent
 * @see Locatable
 *
 * @author Robert Field
 * @author Gordon Hirsch
 * @author James McIlree
 * @since  1.3
 */
public interface Location extends Mirror, Comparable {
    /**
     * Gets the type to which this Location belongs. Normally 
     * the declaring type is a {@link ClassType}, but executable
     * locations also may exist within the static initializer of an
     * {@link InterfaceType}.
     *
     * @return the {@link ReferenceType} containing this Location.
     */
    ReferenceType declaringType();

    /**
     * Gets the method, if any, containing this Location.
     *
     * @return the {@link Method} if this location is in a method, 
     * otherwise null.
     */
    Method method();

    /**
     * Gets the code position within this location's method.  
     * 
     * @return the long representing the position within the method, or
     * -1 if the method is native.
     */
    long codeIndex();

    /**
     * Gets an identifing name for the source corresponding to this 
     * location. Interpretation of this string is the responsibility
     * of the source repository mechanism.
     * <P>
     * In the reference implementation, the returned string is the 
     * unqualified name of the source file for this Location.
     *
     * @return a string specifying the source
     * @throws AbsentInformationException if the source name is not
     * known
     */
    String sourceName() throws AbsentInformationException;

    /**
     * The line number of this Location.  The line number is relative
     * to the source specified by {@link #sourceName}.
     * 
     * @return an int specifying the line in the source, return 
     * -1 if the information is not available.
     */
    int lineNumber();
    
    /**
     * Compares the specified Object with this Location for equality.
     *
     * @return true if the Object is a Location and if it refers to 
     * the same point in the same VM as this Location.
     */
    boolean equals(Object obj);

    /**
     * Returns the hash code value for this Location.
     *
     * @return the integer hash code
     */
    int hashCode();
}

