/*
 * @(#)Method.java	1.26 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jdi;

import java.util.List;

/**
 * A static or instance method in the target VM. See {@link TypeComponent}
 * for general information about Field and Method mirrors.
 *
 * @see ObjectReference
 * @see ReferenceType
 *
 * @author Robert Field
 * @author Gordon Hirsch
 * @author James McIlree
 * @since  1.3
 */
public interface Method extends TypeComponent, Locatable, Comparable {

    /**
     * Returns a text representation of the declared return type
     * of this method.
     *
     * @return a string containing the return type name.
     */
    String returnTypeName();

    /**
     * Returns the return type of the this {@link Method}. For methods
     * returning primitive values,
     * an appropriate {@link PrimitiveType} is always returned. For methods
     * returning object values,  
     * the appropriate {@link ReferenceType} is returned if it has 
     * been loaded through the enclosing type's (that is, the
     * type returned by {@link TypeComponent#enclosingType}) class loader.
     *
     * @return the return {@link Type} of this method.
     * @throws ClassNotLoadedException if the type has not yet been loaded
     * through the appropriate class loader. 
     */
    Type returnType() throws ClassNotLoadedException;

    /**
     * Returns a text representation of all declared argument types
     * of this method. This list is always available even if 
     * argument variables are unavailable due to absent debug 
     * information.
     *
     * @return a List of strings containing the return type name.
     */
    List argumentTypeNames();

    /**
     * Returns the list containing the type of each argument. 
     * For primitive arguments,
     * an appropriate {@link PrimitiveType} is always returned. 
     * For object arguments, 
     * the appropriate {@link ReferenceType} is returned if it has 
     * been loaded through the enclosing type's (that is, the
     * type returned by {@link TypeComponent#enclosingType}) class loader.
     *
     * @return the list of argument types for this method.
     * @throws ClassNotLoadedException if the type has not yet been loaded
     * through the appropriate class loader. 
     */
    List argumentTypes() throws ClassNotLoadedException;

    /**
     * Determine if this method is abstract.
     *
     * @return <code>true</code> if the method is declared abstract; 
     * false otherwise.
     */
    boolean isAbstract();

    /**
     * Determine if this method is synchronized.
     *
     * @return <code>true</code> if the method is declared synchronized; 
     * false otherwise.
     */
    boolean isSynchronized(); 

    /**
     * Determine if this method is native.
     *
     * @return <code>true</code> if the method is declared native; 
     * false otherwise.
     */
    boolean isNative(); 

    /**
     * Determine if this method is a constructor.
     *
     * @return <code>true</code> if the method is a constructor; 
     * false otherwise.
     */
    boolean isConstructor();

    /**
     * Determine if this method is a static initializer.
     *
     * @return <code>true</code> if the method is a static initializer; 
     * false otherwise.
     */
    boolean isStaticInitializer();
    
    /**
     * Returns the beginning {@link Location} objects for each 
     * executable source line in this method. Each location
     * maps a source line to a range of code indices. The beginning
     * of the range can be determined through {@link Location#codeIndex}.
     * The returned list is ordered by code index (from low to high).
     * The returned list may contain multiple locations for a  
     * particular line number, if the compiler and/or VM has 
     * mapped that line to two or more disjoint code index ranges.
     * <P>
     * If the method is native or abstract, an empty list is 
     * returned. 
     *
     * @return a List of all source line {@link Location} objects.
     * @throws AbsentInformationException if there is no line number
     * information for this (non-native, non-abstract) method.
     */
    List allLineLocations() throws AbsentInformationException;

    /**
     * Returns a list containing each {@link Location} that maps
     * to the given line. The returned
     * list will contain a location for each disjoint range
     * of code indices that have been assigned to the given line by the 
     * compiler and/or VM. Each returned location corresponds to the 
     * beginning of this range. 
     *
     * @param lineNumber the line number 
     * @return a List of {@link Location} objects that map to
     * the given line number.
     * @throws AbsentInformationException if there is no line number
     * information for this method.
     * @throws InvalidLineNumberException if there is no executable code
     * in this method associated with the given line.
     */
    List locationsOfLine(int lineNumber) throws AbsentInformationException;

    /**
     * Returns a {@link Location} for the given code index. 
     * 
     * @return the {@link Location} correspoding to the the 
     * given code index.
     * @throws InvalidCodeIndexException if there is no valid location
     * at the specified code index
     */
    Location locationOfCodeIndex(long codeIndex);

    /**
     * Returns a list containing each {@link LocalVariable} declared 
     * in this method. The list includes any variable declared in any
     * scope within the method. It may contain multiple variables of the
     * same name declared within disjoint scopes. Arguments are considered
     * local variables and will be present in the returned list.
     *
     * @return the list of {@link LocalVariable} objects which mirror
     * local variables declared in this method in the target VM.
     * If there are no local variables, a zero-length list is returned.
     * @throws AbsentInformationException if there is no variable
     * information for this method.
     * @throws NativeMethodException if this operation is attempted
     * for a native method. 
     */
    List variables() throws AbsentInformationException;

    /**
     * Returns a list containing each {@link LocalVariable} of a 
     * given name in this method.
     * Multiple variables can be returned
     * if the same variable name is used in disjoint 
     * scopes within the method.
     *
     * @return the list of {@link LocalVariable} objects of the given
     * name.
     * If there are no matching local variables, a zero-length list is returned.
     * @throws AbsentInformationException if there is no variable
     * information for this method.
     * @throws NativeMethodException if this operation is attempted
     * for a native method. 
     */
    List variablesByName(String name) throws AbsentInformationException;

    /**
     * Returns a list containing each {@link LocalVariable} that is
     * declared as an argument of this method.
     *
     * @return the list of {@link LocalVariable} arguments.
     * If there are no arguments, a zero-length list is returned.
     * @throws AbsentInformationException if there is no variable
     * information for this method.
     * @throws NativeMethodException if this operation is attempted
     * for a native method. 
     */
    List arguments() throws AbsentInformationException;

    /**
     * Returns an array containing the bytecodes for this method.
     * <p>
     * Not all target VMs support this query. See
     * VirtualMachine#canGetBytecodes to determine if the 
     * operation is supported.
     * 
     * @return the array of bytecodes.
     * @throws java.lang.UnsupportedOperationException if the target
     * VM does not support the retrieval of bytecodes.
     */
    byte[] bytecodes();
    
    /**
     * Compares the specified Object with this method for equality.
     *
     * @return true if the Object is a method and if both
     * mirror the same method (declared in the same class or interface, in
     * the same VM).
     */
    boolean equals(Object obj);

    /**
     * Returns the hash code value for this Method.
     * 
     * @return the integer hash code
     */
    int hashCode();
}

