/*
 * @(#)ReferenceType.java	1.28 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jdi;

import java.util.List;
import java.util.Map;

/**
 * The type of an object in a target VM. ReferenceType encompasses
 * classes, interfaces, and array types as defined in the Java 
 * Language Specification (and has corresponding subinterfaces for each).
 * The VM creates Class objects for all three, so from the VM perspective,
 * each ReferenceType maps to a distinct Class object.
 * <p>
 * ReferenceTypes can 
 * be obtained by querying a particular {@link ObjectReference} for its
 * type or by getting a list of all reference types from the 
 * {@link VirtualMachine}.
 * <p>
 * ReferenceType provides access to static type information such as 
 * methods and fields and provides access to dynamic type 
 * information such as the corresponding Class object and the classloader.
 *
 * @see ObjectReference
 * @see ObjectReference#referenceType
 * @see VirtualMachine
 * @see VirtualMachine#allClasses
 *
 * @author Robert Field
 * @author Gordon Hirsch
 * @author James McIlree
 * @since  1.3
 */
public interface ReferenceType extends Type, Comparable, Accessible {

    /**
     * Gets the fully qualified name of this type. The returned name
     * is formatted as it might appear in a Java langauge declaration
     * for objects of this type.
     * <p>
     * For primitive classes
     * the returned name is the name of the corresponding primitive
     * type; for example, "int" is returned as the name of the class
     * represented by {@link java.lang.Integer#TYPE}.
     * @return a string containing the type name.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    String name();

    /**
     * Gets the classloader object which loaded the class corresponding
     * to this type. 
     *
     * @return a {@link ClassLoaderReference} which mirrors the classloader,
     * or null if the class was loaded through the bootstrap class
     * loader.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    ClassLoaderReference classLoader();

    /**
     * Gets an identifing name for the source corresponding to the 
     * declaration of this type. Interpretation of this string is 
     * the responsibility of the source repository mechanism.
     * <P>
     * In the reference implementation, the returned string is the 
     * unqualified name of the source file containing the declaration
     * of this type.
     *
     * @return the string source file name
     * @throws AbsentInformationException if the source name is not
     * known
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    String sourceName() throws AbsentInformationException;

    /**
     * Determines if this type was declared static. Only nested types, 
     * can be declared static, so <code>false</code> is returned 
     * for any package-level type, array type, or primitive class
     * (for example, {@link java.lang.Integer#TYPE}).
     *
     * @return <code>true</code> if this type is static; false otherwise.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    boolean isStatic();

    /**
     * Determines if this type was declared abstract. 
     * Return value is undefined for arrays and for primitive classes 
     * (for example, {@link java.lang.Integer#TYPE}).
     *
     * @return <code>true</code> if this type is abstract; false otherwise.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    boolean isAbstract();

    /**
     * Determines if this type was declared final.
     * For {@link ArrayType} and for 
     * primitive classes, such as {@link java.lang.Integer#TYPE},
     * the return value is always true.
     *
     * @return <code>true</code> if this type is final; false otherwise.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    boolean isFinal();

    /**
     * Determines if this type has been prepared. See the JVM 
     * specification for a definiton of class preparation. 
     * Return value is undefined for arrays and for primitive classes 
     * (for example, {@link java.lang.Integer#TYPE}).
     *
     * @return <code>true</code> if this type is prepared; false otherwise.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    boolean isPrepared();

    /**
     * Determines if this type has been verified. See the JVM 
     * specification for a definiton of class verification.
     * Return value is undefined for arrays and for primitive classes 
     * (for example, {@link java.lang.Integer#TYPE}).
     *
     * @return <code>true</code> if this type is verified; false otherwise.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    boolean isVerified();

    /**
     * Determines if this type has been initialized. See the JVM 
     * specification for a definiton of class verification.  
     * For {@link InterfaceType}, this method always returns the
     * same value as {@link #isPrepared()}.
     * Return value is undefined for arrays and for primitive classes 
     * (for example, {@link java.lang.Integer#TYPE}).
     *
     * @return <code>true</code> if this type is initialized; false otherwise.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    boolean isInitialized();

    /**
     * Determines if initialization failed for this class. See the JVM 
     * specification for details on class initialization.
     * Return value is undefined for arrays and for primitive classes 
     * (for example, {@link java.lang.Integer#TYPE}).
     *
     * @return <code>true</code> if initialization was attempted and
     * failed; false otherwise.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    boolean failedToInitialize();

    /**
     * Returns a list containing each {@link Field} declared in this type. 
     * Inherited fields are not included. Any synthetic fields created
     * by the compiler are included in the list.
     *
     * @return a list {@link Field} objects; the list has length 0
     * if no fields exist. 
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    List fields();

    /**
     * Returns a list containing each unhidden and unambiguous {@link Field}
     * in this type.  
     * Each field that can be accessed from the class
     * or its instances with its simple name is included. Fields that
     * are ambiguously multiply inherited or fields that are hidden by
     * fields with the same name in a more recently inherited class
     * cannot be accessed
     * by their simple names and are not included in the returned
     * list. All other inherited fields are included.
     * See JLS section 8.3 for details.
     *
     * @return a List of {@link Field} objects; the list has length
     * 0 if no visible fields exist.
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    List visibleFields();

    /**
     * Returns a list containing each {@link Field} declared in this type,
     * and its superclasses, implemented interfaces, and/or superinterfaces.
     * All declared and inherited
     * fields are included, regardless of whether they are hidden or 
     * multiply inherited.
     *
     * @return a List of {@link Field} objects; the list has length
     * 0 if no fields exist.
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    List allFields();

    /**
     * Finds the visible {@link Field} with the given 
     * non-ambiguous name. This method follows the 
     * inheritance rules specified in the JLS (8.3.3) to determine
     * visibility. 
     *
     * @param a String containing the name of desired field.
     * @return a {@link Field} object which mirrors the found field, or
     * null if there is no field with the given name or if the given 
     * name is ambiguous.
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    Field fieldByName(String fieldName);

    /**
     * Returns a list containing each {@link Method} declared 
     * directly in this type. 
     * Inherited methods are not included. Constructors, 
     * the initialization method if any, and any synthetic methods created
     * by the compiler are included in the list.
     *
     * @return a list {@link Method} objects; the list has length 0
     * if no methods exist. 
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    List methods();

    /**
     * Returns a list containing each {@link Method}
     * declared or inherited by this type. Methods from superclasses
     * or superinterfaces that that have been hidden or overridden
     * are not included. 
     * <p>
     * Note that despite this exclusion, multiple inherited methods 
     * with the same signature can be present in the returned list, but
     * at most one can be a member of a {@link ClassType}.
     * See JLS section 8.4.6 for details.
     *
     * @return a List of {@link Method} objects; the list has length
     * 0 if no visible methods exist.
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    List visibleMethods();

    /**
     * Returns a list containing each {@link Method} declared in this type,
     * and its superclasses, implemented interfaces, and/or superinterfaces.
     * All declared and inherited
     * methods are included, regardless of whether they are hidden or 
     * overridden.
     *
     * @return a List of {@link Method} objects; the list has length
     * 0 if no methods exist.
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    List allMethods();

    /**
     * Returns a List containing each visible {@link Method} that 
     * has the given name.  This is most commonly used to
     * find overloaded methods.
     * <p>
     * Overridden and hidden methods are not included.
     * See JLS (8.4.6) for details.
     *
     * @param name the name of the method to find.
     * @return a List of {@link Method} objects that match the given
     * name; the list has length 0 if no matching methods are found.
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    List methodsByName(String name);

    /**
     * Returns a List containing each visible {@link Method} that 
     * has the given name and signature.
     * The signature string is the
     * JNI signature for the target method:
     * <ul>
     * <li><code>()V</code>
     * <li><code>([Ljava/lang/String;)V</code> 
     * <li><code>(IIII)Z</code>
     * </ul>
     * This method follows the inheritance rules specified
     * in the JLS (8.4.6) to determine visibility.
     * <p>
     * At most one method in the list is a concrete method and a 
     * component of {@link ClassType}; any other methods in the list
     * are abstract. Use {@link ClassType#concreteMethodByName} to 
     * retrieve only the matching concrete method.
     *
     * @param name the name of the method to find.
     * @param signature the signature of the method to find
     * @return a List of {@link Method} objects that match the given
     * name and signature; the list has length 0 if no matching methods 
     * are found.
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    List methodsByName(String name, String signature);

    /**
     * Returns a List containing each {@link ReferenceType} declared
     * within this type. Both static nested types and non-static nested
     * types (that is, inner types) are included. Local inner types
     * (declared within a code block somewhere in this reference type) are 
     * also included in the returned list.
     *
     * @return a List of nested {@link ReferenceType} objects; the list
     * has 0 length if there are no nested types.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    List nestedTypes();

    /**
     * Gets the {@link Value} of a given static {@link Field} in this type.
     * The Field must be valid for this type;
     * that is, it must be declared in this type, a superclass, a
     * superinterface, or an implemented interface.
     *
     * @param sig the field containing the requested value
     * @return the {@link Value} of the instance field.
     * @throws java.lang.IllegalArgumentException if the field is not valid for
     * this object's class.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    Value getValue(Field field);

    /**
     * Returns a map containing the {@link Value} of each 
     * static {@link Field} in the given list.
     * The Fields must be valid for this type;
     * that is, they must be declared in this type, a superclass, a
     * superinterface, or an implemented interface.
     *
     * @param fields a list of {@link Field} objects containing the
     * requested values.
     * @return a Map of the requested {@link Field} objects with 
     * their {@link Value}.
     * @throws java.lang.IllegalArgumentException if any field is not valid for
     * this object's class.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     * @throws VMMismatchException if a {@link Mirror} argument and this mirror
     * do not belong to the same {@link VirtualMachine}.
     */
    Map getValues(List fields);

    /**
     * Returns the class object that corresponds to this type in the 
     * target VM. The VM creates class objects for every kind of 
     * ReferenceType: classes, interfaces, and array types.
     * @return the {@link ClassObjectReference} for this reference type
     * in the target VM.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    ClassObjectReference classObject();

    /**
     * Returns the {@link Location} objects for each 
     * executable source line in this reference type. Each location
     * maps a source line to a range of code indices. The beginning
     * of the range can be determined through {@link Location#codeIndex}.
     * The returned list may contain multiple locations for a  
     * particular line number, if the compiler and/or VM has 
     * mapped that line to two or more disjoint code index ranges.
     * Note that it is possible for the same source line to represent
     * different code index ranges in <i>different</i> methods.
     * <p>
     * If this reference type is an {@link ArrayType}, the returned list is
     * always empty. If this reference type is an {@link InterfaceType}, the
     * returned list will be non-empty only if the interface has executable
     * code in its class initialization.
     *
     * @return a List of all source line {@link Location} objects.
     * @throws AbsentInformationException if there is no line number
     * information for this class and there are non-native, non-abstract
     * executable members of this class.
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this class has been 
     * unloaded and garbage collected.
     */
    List allLineLocations() throws AbsentInformationException;

    /**
     * Returns a List filled with all {@link Location} objects that map to
     * the given line number. 
     * <p>
     * If this reference type is an {@link ArrayType}, the returned list is
     * always empty. If this reference type is an {@link InterfaceType}, the
     * returned list will be non-empty only if the interface has executable
     * code in its class initialization at the specified line number.
     *
     * @return a List of all {@link Location} objects that map to the 
     * given line.
     * @throws AbsentInformationException if there is no line number
     * information for this class.
     * @throws ClassNotPreparedException if this class not yet been 
     * prepared.
     * @throws ObjectCollectedException if this class has been 
     * unloaded and garbage collected.
     * @throws InvalidLineNumberException if there is no executable code
     * associated with the given line.
     */
    List locationsOfLine(int lineNumber) throws AbsentInformationException;
    
    /**
     * Compares the specified Object with this ReferenceType for equality.
     *
     * @return  true if the Object is a {@link ReferenceType}, if the 
     * ReferenceTypes belong to the same VM, and if they mirror classes
     * which correspond to the same instance of java.lang.Class in that VM.
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    boolean equals(Object obj);

    /**
     * Returns the hash code value for this ObjectReference.
     *
     * @return the integer hash code
     * @throws ObjectCollectedException if this reference type has been 
     * unloaded and garbage collected.
     */
    int hashCode();

}

