/*
 * @(#)ThreadReference.java	1.26 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jdi;
import java.util.List;

/**
 * A thread object from the target VM.
 * A ThreadReference is an {@link ObjectReference} with additional
 * access to thread-specific information from the target VM.
 *
 * @author Robert Field
 * @author Gordon Hirsch
 * @author James McIlree
 * @since  1.3
 */
public interface ThreadReference extends ObjectReference
{
    /** Thread status is unknown */
    public final int THREAD_STATUS_UNKNOWN  =-1;
    /** Thread has completed execution */
    public final int THREAD_STATUS_ZOMBIE = 0;
    /** Thread is runnable */
    public final int THREAD_STATUS_RUNNING = 1;
    /** Thread is sleeping - Thread.sleep() or JVM_Sleep() was called */
    public final int THREAD_STATUS_SLEEPING = 2;
    /** Thread is waiting on a java monitor */
    public final int THREAD_STATUS_MONITOR = 3;
    /** Thread is waiting - Thread.wait() or JVM_MonitorWait() was called */
    public final int THREAD_STATUS_WAIT = 4;
    /** Thread has not yet been started */
    public final int THREAD_STATUS_NOT_STARTED = 5;

    /**
     * Returns the name of this thread.
     * 
     * @return the string containing the thread name.
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    String name();

    /**
     * Suspends this thread. The thread can be resumed through 
     * {@link #resume} or resumed with other threads through
     * {@link VirtualMachine#resume}.
     * <p>
     * Unlike {@link java.lang.Thread#suspend},
     * suspends of both the virtual machine and individual threads are 
     * counted. Before a thread will run again, it must be resumed
     * (through {@link #resume} or {@link ThreadReference#resume}) 
     * the same number of times it has been suspended.
     * <p>
     * Suspending single threads with this method has the same dangers
     * as {@link java.lang.Thread.suspend()}. If the suspended thread
     * holds a monitor needed by another running thread, deadlock is 
     * possible in the target VM (at least until the suspended thread
     * is resumed again).
     * <p>
     * The suspended thread is guaranteed to remain suspended until
     * resumed through one of the JDI resume methods mentioned above;
     * the application in the target VM cannot resume the suspended thread
     * through {@link java.lang.Thread#resume}.
     *
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    void suspend();

    /**
     * Resumes this thread. If this thread was not previously suspended
     * through {@link #suspend} or through {@link VirtualMachine#suspend},
     * invoking this method has no effect. Otherwise, the count of pending
     * suspends on this thread is decremented. If it is decremented to 0, 
     * the thread will continue to execute. 
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    void resume();

    /**
     * Resumes the number of pending suspends for this thread. See 
     * {@link #suspend} for an explanation of counted suspends.
     * @return the integer pending suspend count
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    int suspendCount();

    /**
     * Stops this thread with an asynchronous exception. 
     * A debugger thread in the target VM will stop this thread
     * with the given {@link java.lang.Throwable} object.
     *
     * @param throwable the asynchronous exception to throw.
     * @throws InvalidTypeException if <code>throwable</code> is not
     * an instance of java.lang.Throwable in the target VM.
     * @see java.lang.Thread#stop(Throwable)
     * @throws ObjectCollectedException if this object or the 
     * asynchronous exception has been 
     * garbage collected.
     */
    void stop(ObjectReference throwable) throws InvalidTypeException;

    /**
     * Interrupts this thread.
     *
     * @see java.lang.Thread#interrupt()
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    void interrupt();

    /**
     * Returns the thread's status. If the thread is not suspended the 
     * thread's current status is returned. If the thread is suspended, the
     * thread's status before the suspension is returned (or
     * {@link #THREAD_STATUS_UNKNOWN} if this information is not available.
     * {@link #isSuspended} can be used to determine if the thread has been 
     * suspended.
     *
     * @return one of 
     * {@link #THREAD_STATUS_UNKNOWN},
     * {@link #THREAD_STATUS_ZOMBIE},
     * {@link #THREAD_STATUS_RUNNING},
     * {@link #THREAD_STATUS_SLEEPING},
     * {@link #THREAD_STATUS_MONITOR},
     * {@link #THREAD_STATUS_WAIT},
     * {@link #THREAD_STATUS_NOT_STARTED},
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    int status();

    /**
     * Determines whether the thread has been suspended by the 
     * the debugger.
     *
     * @return <code>true</code> if the thread is currently suspended;
     * <code>false</code> otherwise.
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    boolean isSuspended();

    /**
     * Determines whether the thread is suspended at a breakpoint. 
     *
     * @return <code>true</code> if the thread is currently stopped at
     * a breakpoint; <code>false</code> otherwise.
     * @throws ObjectCollectedException if this object has been
     * garbage collected.
     */
    boolean isAtBreakpoint();

    /**
     * Returns this thread's thread group. 
     * @return a {@link ThreadGroupReference} that mirrors this thread's
     * thread group in the target VM.
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    ThreadGroupReference threadGroup();

    /**
     * Returns the number of stack frames in the thread's current
     * call stack. 
     * The thread must be suspended (normally through an interruption
     * to the VM) to get this information, and 
     * it is only valid until the thread is resumed again.
     *
     * @return an integer frame count
     * @throws IncompatibleThreadStateException if the thread is not suspended
     * in the target VM
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    int frameCount() throws IncompatibleThreadStateException;

    /**
     * Returns a List containing each {@link StackFrame} in the 
     * thread's current call stack. 
     * The thread must be suspended (normally through an interruption
     * to the VM) to get this information, and 
     * it is only valid until the thread is resumed again.
     *
     * @return a List of {@link StackFrame} with the current frame first
     * followed by each caller's frame.
     * @throws IncompatibleThreadStateException if the thread is not suspended
     * in the target VM
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    List frames() throws IncompatibleThreadStateException;

    /**
     * Returns the {@link StackFrame} at the given index in the 
     * thread's current call stack. Index 0 retrieves the current
     * frame; higher indices retrieve caller frames.
     * The thread must be suspended (normally through an interruption
     * to the VM) to get this information, and 
     * it is only valid until the thread is resumed again.
     *
     * @param index the desired frame
     * @return the requested {@link StackFrame} 
     * @throws IncompatibleThreadStateException if the thread is not suspended
     * in the target VM
     * @throws java.lang.IndexOutOfBoundsException if the index is greater than 
     * or equal to {@link #frameCount} or is negative.
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    StackFrame frame(int index) throws IncompatibleThreadStateException;

    /**
     * Returns a List containing a range of {@link StackFrame} mirrors 
     * from the thread's current call stack. 
     * The thread must be suspended (normally through an interruption
     * to the VM) to get this information, and 
     * it is only valid until the thread is resumed again.
     *
     * @param start the index of the first frame
     * @param length the number of frames to retrieve
     * @return a List of {@link StackFrame} with the current frame first
     * followed by each caller's frame.
     * @throws IncompatibleThreadStateException if the thread is not suspended
     * in the target VM
     * @throws IndexOutOfBoundsException if the specified range is not
     * within the range from 0 to {@link #frameCount} - 1.
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    List frames(int start, int length) throws IncompatibleThreadStateException;

    /**
     * Returns a List containing an {@link ObjectReference} for 
     * each monitor owned by the thread. 
     * A monitor is owned by a thread if it has been entered
     * (via the synchronized statement or entry into a synchronized 
     * method) and has not been relinquished through {@link Object#wait}.
     * <p>
     * Not all target VMs support this operation. See
     * VirtualMachine#canGetOwnedMonitorInfo to determine if the 
     * operation is supported.
     *
     * @return a List of {@link ObjectReference} objects. The list
     * has zero length if no monitors are owned by this thread.
     * @throws java.lang.UnsupportedOperationException if the target
     * target VM does not support this operation.
     * @throws IncompatibleThreadStateException if the thread is not suspended
     * in the target VM
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    List ownedMonitors() throws IncompatibleThreadStateException;

    /**
     * Returns an {@link ObjectReference} for the monitor, if any, 
     * for which this thread is currently waiting.
     * The thread can be waiting for a monitor through entry into a
     * synchronized method, the synchronized statement, or 
     * {@link Object#wait}.  The {@link #status} method can be used
     * to differentiate between the first two cases and the third.
     * <p>
     * Not all target VMs support this operation. See
     * VirtualMachine#canGetCurrentContendedMonitor to determine if the 
     * operation is supported.
     *
     * @return the {@link ObjectReference} corresponding to the 
     * contended monitor, or null if it is not waiting for a monitor.
     * @throws java.lang.UnsupportedOperationException if the target
     * target VM does not support this operation.
     * @throws IncompatibleThreadStateException if the thread is not suspended
     * in the target VM
     * @throws ObjectCollectedException if this object has been 
     * garbage collected.
     */
    ObjectReference currentContendedMonitor() throws IncompatibleThreadStateException;
}
