/*
 * @(#)EventRequestManager.java	1.27 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jdi.request;

import com.sun.jdi.*;

import java.util.List;

/**
 * Manages the creation and deletion of {@link EventRequest}s. A single 
 * implementor of this interface exists in a particuar VM and
 * is accessed through {@link VirtualMachine#eventRequestManager()}
 *
 * @see EventRequest
 * @see com.sun.jdi.event.Event
 * @see BreakpointRequest
 * @see com.sun.jdi.event.BreakpointEvent
 * @see VirtualMachine
 *
 * @author Robert Field
 * @since  1.3
 */

public interface EventRequestManager extends Mirror {

    /**
     * Creates a new disabled {@link ClassPrepareRequest}. 
     * The new event request is added to the list managed by this
     * EventRequestManager. Use {@link EventRequest#enable()} to
     * activate this event request.
     * 
     * @return the created {@link ClassPrepareRequest}
     */
    ClassPrepareRequest createClassPrepareRequest();

    /**
     * Creates a new disabled {@link ClassUnloadRequest}. 
     * The new event request is added to the list managed by this
     * EventRequestManager. Use {@link EventRequest#enable()} to
     * activate this event request.
     * 
     * @return the created {@link ClassUnloadRequest}
     */
    ClassUnloadRequest createClassUnloadRequest();

    /**
     * Creates a new disabled {@link ThreadStartRequest}. 
     * The new event request is added to the list managed by this
     * EventRequestManager. Use {@link EventRequest#enable()} to
     * activate this event request.
     * 
     * @return the created {@link ThreadStartRequest}
     */
    ThreadStartRequest createThreadStartRequest();

    /**
     * Creates a new disabled {@link ThreadDeathRequest}. 
     * The new event request is added to the list managed by this
     * EventRequestManager. Use {@link EventRequest#enable()} to
     * activate this event request.
     * 
     * @return the created {@link ThreadDeathRequest}
     */
    ThreadDeathRequest createThreadDeathRequest();

    /**
     * Creates a new disabled {@link ExceptionRequest}. 
     * The new event request is added to the list managed by this
     * EventRequestManager. Use {@link EventRequest#enable()} to
     * activate this event request.
     * <P>
     * A specific exception type and its subclasses can be selected
     * for exception events. Caught exceptions,  uncaught exceptions, 
     * or both can be selected. Note, however, that 
     * at the time an exception is thrown, it is not always
     * possible to determine whether it is truly caught. See 
     * {@link com.sun.jdi.event.ExceptionEvent#catchLocation} for 
     * details.
     * @param refType If non-null, specifies that exceptions which are
     *                instances of refType will be reported. Note: this
     *                will include instances of sub-types.  If null,
     *                all instances will be reported
     * @param notifyCaught If true, caught exceptions will be reported.
     * @param notifyUncaught If true, uncaught exceptions will be reported.
     *
     * @return the created {@link ExceptionRequest} 
     * @throws VMMismatchException if refType and this EventRequestManager
     * do not belong to the same {@link VirtualMachine}.
     */
    ExceptionRequest createExceptionRequest(ReferenceType refType, 
                                            boolean notifyCaught,  
                                            boolean notifyUncaught);

    /**
     * Creates a new disabled {@link MethodEntryRequest}. 
     * The new event request is added to the list managed by this
     * EventRequestManager. Use {@link EventRequest#enable()} to
     * activate this event request.
     * 
     * @return the created {@link MethodEntryRequest}
     */
    MethodEntryRequest createMethodEntryRequest();

    /**
     * Creates a new disabled {@link MethodExitRequest}. 
     * The new event request is added to the list managed by this
     * EventRequestManager. Use {@link EventRequest#enable()} to
     * activate this event request.
     * 
     * @return the created {@link MethodExitRequest}
     */
    MethodExitRequest createMethodExitRequest();

    /**
     * Creates a new disabled {@link StepRequest}. 
     * The new event request is added to the list managed by this
     * EventRequestManager. Use {@link EventRequest#enable()} to
     * activate this event request.
     * <p>
     * The returned request will control stepping only in the specified
     * <code>thread</code>; all other threads will be unaffected. 
     * A <code>size</code>value of {@link com.sun.jdi.request.StepRequest#STEP_MIN} will generate a 
     * step event each time the code index changes. It represents the
     * smallest step size available and often maps to the instruction
     * level.
     * A <code>size</code> value of {@link com.sun.jdi.request.StepRequest#STEP_LINE} will generate a 
     * step event each time the source line changes. A 
     * <code>depth</code> value of {@link com.sun.jdi.request.StepRequest#STEP_INTO} will generate
     * step events in any called methods.  A <code>depth</code> value
     * of {@link com.sun.jdi.request.StepRequest#STEP_OVER} restricts step events to the current frame
     * or caller frames. A <code>depth</code> value of {@link com.sun.jdi.request.StepRequest#STEP_OUT}
     * restricts step events to caller frames only. All depth 
     * restrictions are relative to the call stack immediately before the
     * step takes place. 
     * <p>
     * Only one pending step request is allowed per thread.
     * <p>
     * Note that a typical debugger will want to cancel stepping 
     * after the first step is detected.  Thus a next line method
     * would do the following:
     * <code>
     * <pre>
     *     EventRequestManager mgr = myVM.{@link VirtualMachine#eventRequestManager eventRequestManager}();
     *     StepRequest request = mgr.createStepRequest(myThread, 
     *                                                 StepRequest.{@link StepRequest#STEP_LINE STEP_LINE}, 
     *                                                 StepRequest.{@link StepRequest#STEP_OVER STEP_OVER});
     *     request.{@link EventRequest#addCountFilter addCountFilter}(1);  // next step only
     *     request.enable();
     *     myVM.{@link VirtualMachine#resume resume}();
     *
     * @param thread the thread in which to step
     * @param depth the step depth
     * @param size the step size
     * @return the created {@link StepRequest}
     * @throws DuplicateRequestException if there is already a pending
     * step request for the specified thread.
     * @throws ObjectCollectedException if the thread object has been 
     * garbage collected.
     * @throws VMMismatchException if thread and this EventRequestManager
     * do not belong to the same {@link VirtualMachine}.
     */
    StepRequest createStepRequest(ThreadReference thread, 
                                  int size,
                                  int depth);

    /**
     * Creates a new disabled {@link BreakpointRequest}. 
     * The given {@link Location} must have a valid 
     * (that is, non-negative) code index. The new 
     * breakpoint is added to the list managed by this
     * EventRequestManager. Multiple breakpoints at the 
     * same location are permitted. Use {@link EventRequest#enable()} to
     * activate this event request.
     * 
     * @param location the location of the new breakpoint.
     * @return the created {@link BreakpointRequest}
     * @throws NullPointerException if location is null.
     * @throws VMMismatchException if location and this EventRequestManager
     * do not belong to the same {@link VirtualMachine}.
     */
    BreakpointRequest createBreakpointRequest(Location location);
    
    /**
     * Creates a new disabled watchpoint which watches accesses to the
     * specified field. The new 
     * watchpoint is added to the list managed by this
     * EventRequestManager. Multiple watchpoints on the
     * same field are permitted. Use {@link EventRequest#enable()} to
     * activate this event request.
     *
     * @param field the field to watch
     * @return the created watchpoint
     * @throws NullPointerException if field is null.
     * @throws VMMismatchException if field and this EventRequestManager
     * do not belong to the same {@link VirtualMachine}.
     */
    AccessWatchpointRequest createAccessWatchpointRequest(Field field);
    
    /**
     * Creates a new disabled watchpoint which watches accesses to the
     * specified field. The new 
     * watchpoint is added to the list managed by this
     * EventRequestManager. Multiple watchpoints on the
     * same field are permitted. Use {@link EventRequest#enable()} to
     * activate this event request.
     *
     * @param field the field to watch
     * @return the created watchpoint
     * @throws NullPointerException if field is null.
     * @throws VMMismatchException if field and this EventRequestManager
     * do not belong to the same {@link VirtualMachine}.
     */
    ModificationWatchpointRequest createModificationWatchpointRequest(Field field);
    
    /**
     * Removes a eventRequest. The eventRequest is disabled and removed from
     * the list of eventRequests managed by this 
     * EventRequestManager. Once the eventRequest is deleted, no
     * significant operations (for example, {@link EventRequest#setEnabled})
     * are permitted. No other eventRequests are effected.
     * 
     * @param eventRequest the eventRequest to remove
     * @throws VMMismatchException if a {@link Mirror} argument and this mirror
     * do not belong to the same {@link VirtualMachine}.
     */
    void deleteEventRequest(EventRequest eventRequest);

    /**
     * Removes a list of {@link EventRequest}s. 
     * 
     * @see #deleteEventRequest(EventRequest)
     * 
     * @param eventRequests the list of eventRequests to remove
     * @throws VMMismatchException if a {@link Mirror} argument and this mirror
     * do not belong to the same {@link VirtualMachine}.
     */
    void deleteEventRequests(List eventRequests);

    /**
     * Remove all breakpoints managed by this EventRequestManager.
     *
     * @see #deleteEventRequest(EventRequest)
     */
    void deleteAllBreakpoints();

    /**
     * @return the all {@link StepRequest} objects.
     */
    List stepRequests();

    /**
     * @return the all {@link ClassPrepareRequest} objects.
     */
    List classPrepareRequests();

    /**
     * @return the all {@link ClassUnloadRequest} objects.
     */
    List classUnloadRequests();

    /**
     * @return the all {@link ThreadStartRequest} objects.
     */
    List threadStartRequests();

    /**
     * @return the all {@link ThreadDeathRequest} objects.
     */
    List threadDeathRequests();

    /**
     * @return the all {@link ExceptionRequest} objects.
     */
    List exceptionRequests();

    /**
     * @return the list of all {@link BreakpointRequest} objects.
     */
    List breakpointRequests();

    /**
     * @return the all {@link AccessWatchpointRequest} objects.
     */
    List accessWatchpointRequests();

    /**
     * @return the all {@link ModificationWatchpointRequest} objects.
     */
    List modificationWatchpointRequests();

    /**
     * @return the list of all {@link MethodEntryRequest} objects.
     */
    List methodEntryRequests();

    /**
     * @return the list of all {@link MethodExitRequest} objects.
     */
    List methodExitRequests();
}




			               
