/*
 * @(#)StepRequest.java	1.13 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jdi.request;

import com.sun.jdi.*;

/**
 * Request for notification when a step occurs in the target VM.
 * When an enabled StepRequest is satisfied, an
 * {@link com.sun.jdi.event.EventSet event set} containing a
 * {@link com.sun.jdi.event.StepEvent StepEvent} will be placed on the
 * {@link com.sun.jdi.event.EventQueue EventQueue}.
 * The collection of existing StepRequests is 
 * managed by the {@link EventRequestManager}
 *
 * @see com.sun.jdi.event.StepEvent
 * @see com.sun.jdi.event.EventQueue
 * @see EventRequestManager
 *
 * @author Robert Field
 * @since  1.3
 */
public interface StepRequest extends EventRequest {

    /** Step into any newly pushed frames */
    int STEP_INTO = 1;
    /** Step over any newly pushed frames */
    int STEP_OVER = 2;
    /** Step out of the current frame */
    int STEP_OUT = 3;

    /** Step to the next available location */
    int STEP_MIN = -1;
    /** Step to the next location on a different line */
    int STEP_LINE = -2;

    /**
     * @return the thread on which the step event is being requested.
     */
    ThreadReference thread();

    /**
     * @return the step size
     */
    int size();

    /**
     * @return the step depth
     */
    int depth();

    /**
     * Restricts the events generated by this request to those 
     * whose location is in the given reference type or any 
     * of its subtypes.
     * An event will be generated for any location in a reference type that can
     * be safely cast to the given reference type. 
     *
     * @param refType the reference type to filter on.
     * @throws InvalidRequestStateException if this request is currently
     * enabled. Filters may be added only to disabled requests.
     */
    void addClassFilter(ReferenceType refType);

    /**
     * Restricts the events generated by this request to those 
     * whose location is in a class whose name matches a restricted
     * regular expression. Regular expressions are limited 
     * to exact matches and patterns that begin with '*' or end with '*'; 
     * for example, "*.Foo" or "java.*".
     *
     * @param classPattern the pattern String to filter for.
     * @throws VMMismatchException if a {@link Mirror} argument and this mirror
     * do not belong to the same {@link VirtualMachine}.
     * @throws InvalidRequestStateException if this request is currently
     * enabled. Filters may be added only to disabled requests.
     */
    void addClassFilter(String classPattern);

    /**
     * Restricts the events generated by this request to those 
     * whose location is in a class whose name does <b>not</b> match a restricted
     * regular expression. Regular expressions are limited 
     * to exact matches and patterns that begin with '*' or end with '*'; 
     * for example, "*.Foo" or "java.*".
     *
     * @param classPattern the pattern String to filter against.
     * @throws InvalidRequestStateException if this request is currently
     * enabled. Filters may be added only to disabled requests.
     */
    void addClassExclusionFilter(String classPattern);
}


			
