/*
 * @(#)ArrayReferenceImpl.java	1.28 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.tools.jdi;

import com.sun.jdi.*;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

public class ArrayReferenceImpl extends ObjectReferenceImpl
    implements ArrayReference
{
    int length = -1;

    ArrayReferenceImpl(VirtualMachine aVm,long aRef) {
        super(aVm,aRef);
    }

    ArrayTypeImpl arrayType() {
        return (ArrayTypeImpl)type();
    }

    /**
     * Return array length.
     * Need not be synchronized since it cannot be provably stale.
     */
    public int length() {
        if(length == -1) {
            try {
                length = JDWP.ArrayReference.Length.
                    process(vm, this).arrayLength;
            } catch (JDWPException exc) {
                throw exc.toJDIException();
            }
        }
       return length;
    }

    public Value getValue(int index) {
        List list = getValues(index, 1);
        return (Value)list.get(0);
    }

    public List getValues() {
        return getValues(0, -1);
    }
    
    void validateArrayAccess(int index, int length) {
        int maxComponent = index;
        int arrayLength;

        if (length < -1) {
            throw new 
                IndexOutOfBoundsException("Invalid array range length: " + length);
        } else if ((length == -1) || (length == 0)) {
            if ((index < 0) || (index >= length())) {
                throw new IndexOutOfBoundsException("Invalid array index: " + index);
            } 
        } else {
            if ((index < 0) || (index + length > length())) {
                throw new IndexOutOfBoundsException("Invalid array range: " + 
                               index + " to " + (index + length - 1));
            } 
        }
    }

    public List getValues(int index, int length) {
        validateArrayAccess(index, length);

        List vals;
        try {
            vals = JDWP.ArrayReference.GetValues.
                process(vm, this, index, length).values;
        } catch (JDWPException exc) {
            throw exc.toJDIException();
        }

        return vals;
    }

    public void setValue(int index, Value value)
            throws InvalidTypeException,
                   ClassNotLoadedException {
        List list = new ArrayList(1);
        list.add(value);
        setValues(index, list, 0, 1);
    }

    public void setValues(List values)
            throws InvalidTypeException,
                   ClassNotLoadedException {
        setValues(0, values, 0, -1);
    }

    public void setValues(int index, List srcValues, int srcOffset, int length)
            throws InvalidTypeException,
                   ClassNotLoadedException {
        
        validateMirrorsOrNulls(srcValues);
        validateArrayAccess(index, length);

        if (length == -1) {
           length = length() - index;
        }

        if (length > 0) {
            boolean noOp = false;
	    ValueImpl[] values = new ValueImpl[length];

            for (int i = 0; i < length; i++) {
                ValueImpl value = (ValueImpl)srcValues.get(srcOffset + i);

                try {
                    // Validate and convert if necessary
                    values[i] = ValueImpl.prepareForAssignment(value, 
                                                        new Component());
                } catch (ClassNotLoadedException e) {
                    /*
                     * Since we got this exception,
                     * the component must be a reference type. The value
                     * we're trying to set is null, but if the component's
                     * class has not yet been loaded through the enclosing
                     * class loader, then setting to null is essentially a 
                     * no-op, and we should allow it without an exception.
                     */
                    if (value == null) {
                        // If we escape this loop, the set must be a no-op
                        noOp = true;
                    } else {
                        throw e;
                    }
                }
	    }
            if (!noOp) {
                try {
                    JDWP.ArrayReference.SetValues.
                                             process(vm, this, index, values);
                } catch (JDWPException exc) {
                    throw exc.toJDIException();
                }
            }
        }
    }

    public String toString() {
        return "instance of " + arrayType().componentTypeName() + 
               "[" + length() + "] (id=" + uniqueID() + ")";
    }

    byte typeValueKey() {
        return JDWP.Tag.ARRAY;
    }

    void validateAssignment(ValueContainer destination) 
                            throws InvalidTypeException, ClassNotLoadedException {
        try {
            super.validateAssignment(destination);
        } catch (ClassNotLoadedException e) {
            /*
             * An array can be used extensively without the 
             * enclosing loader being recorded by the VM as an 
             * initiating loader of the array type. In addition, the 
             * load of an array class is fairly harmless as long as 
             * the component class is already loaded. So we relax the
             * rules a bit and allow the assignment as long as the 
             * ultimate component types are assignable. 
             */
            boolean valid = false;
            JNITypeParser destParser = new JNITypeParser(
                                       destination.signature());
            JNITypeParser srcParser = new JNITypeParser(
                                       arrayType().signature());
            int destDims = destParser.dimensionCount();
            if (destDims <= srcParser.dimensionCount()) {
                /*
                 * Remove all dimensions from the destination. Remove
                 * the same number of dimensions from the source. 
                 * Get types for both and check to see if they are 
                 * compatible. 
                 */
                String destComponentSignature = 
                    destParser.componentSignature(destDims);
                Type destComponentType = 
                    destination.findType(destComponentSignature);
                String srcComponentSignature = 
                    srcParser.componentSignature(destDims);
                Type srcComponentType = 
                    arrayType().findComponentType(srcComponentSignature);
                valid = ArrayTypeImpl.isComponentAssignable(destComponentType,
                                                          srcComponentType);
            }

            if (!valid) {
                throw new InvalidTypeException("Cannot assign " +
                                               arrayType().name() + 
                                               " to " +
                                               destination.typeName());
            }
        }
    }

    /*
     * Represents an array component to other internal parts of this 
     * implementation. This is not exposed at the JDI level. Currently,
     * this class is needed only for type checking so it does not even
     * reference a particular component - just a generic component 
     * of this array. In the future we may need to expand its use.
     */
    class Component implements ValueContainer {
        public Type type() throws ClassNotLoadedException {
            return arrayType().componentType();
        }
        public String typeName() {
            return arrayType().componentTypeName();
        }
        public String signature() {
            return arrayType().componentSignature();
        }
        public Type findType(String signature) throws ClassNotLoadedException {
            return arrayType().findComponentType(signature);
        }
    }
}
