/*
 * @(#)MethodImpl.java	1.38 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.tools.jdi;

import com.sun.jdi.*;

import java.util.List;
import java.util.Map;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.lang.ref.SoftReference;

public class MethodImpl extends TypeComponentImpl implements Method {
    /*
     * A subset of the line number info that is softly cached
     */
    static private class SoftLineInfo {
        Map lineMapper = null;       /* Maps line number to location(s) */
        List lineLocations = null;   /* List of locations ordered by code index */

        SoftLineInfo(Map lineMapper, List lineLocations) {
            this.lineMapper = lineMapper;
            this.lineLocations = lineLocations;
        }
    }
    private SoftReference softLineInfoRef;
    private Location location = null;
    private SoftReference variablesRef = null;
    private boolean absentVariableInformation = false;
    private long firstIndex = -1;
    private long lastIndex = -1;
    private int argSlotCount = -1;
    private SoftReference bytecodesRef = null;
    private JNITypeParser signatureParser;

    /*
     * Note: these do not necessarily correspond to the line numbers of the first
     * and last elements in the lineLocations list. Use these only for bounds 
     * checking and with lineMapper.
     */
    private int lowestLine = -1;
    private int highestLine = -1;
    
    protected MethodImpl(VirtualMachine vm, ReferenceTypeImpl declaringType, 
                      long ref,
                      String name, String signature, int modifiers) {
        super(vm, declaringType, ref, name, signature,  modifiers);
        signatureParser = new JNITypeParser(signature);
    }

    public boolean equals(Object obj) {
        if ((obj != null) && (obj instanceof MethodImpl)) {
            MethodImpl other = (MethodImpl)obj;
            return (declaringType().equals(other.declaringType())) &&
                   (ref() == other.ref()) &&
                   super.equals(obj);
        } else {
            return false;
        }
    }

    public int hashCode() {
        return (int)ref();
    }

    public Location location() {
        if (location == null) {
            getLocations();
        }
        return location;
    }

    public List allLineLocations() throws AbsentInformationException {
        if (isNative() || isAbstract()) {
            return new ArrayList(0);
        }
        List lineLocations = getLocations().lineLocations;

        if (lineLocations.size() == 0) {
            throw new AbsentInformationException();
        }
        return lineLocations;
    }

    public List locationsOfLine(int lineNumber) throws AbsentInformationException {
        SoftLineInfo info = getLocations();

        if (info.lineLocations.size() == 0) {
            throw new AbsentInformationException();
        }

        /*
         * Find line matching the one passed in.
         */
        List list = (List)info.lineMapper.get(new Integer(lineNumber));

        return (list == null) ? 
                  new ArrayList(0) : 
                  Collections.unmodifiableList(list);
    }

    public Location locationOfCodeIndex(long codeIndex) {
        if (firstIndex == -1) {
            getLocations();
        }

        /*
         * Check for invalid code index.
         */
        if (codeIndex < firstIndex || codeIndex > lastIndex) {
            return null;
        }

        return new LocationImpl(virtualMachine(), this, codeIndex);
    }


    int codeIndexToLineNumber(long codeIndex) {
        if (firstIndex == -1) {
            getLocations();
        }

        /*
         * Check for invalid code index.
         */
        if (codeIndex < firstIndex || codeIndex > lastIndex) {
            return -1;
        }

        List lineLocations = getLocations().lineLocations;

        /*
         * Check for absent line numbers.
         */
        if (lineLocations.size() == 0) {
            return -1;
        }

        Iterator iter = lineLocations.iterator();
        int line = -1;
        while (iter.hasNext()) {
            Location current = (Location)iter.next();
            if (current.codeIndex() > codeIndex) {
                break;
            }
            line = current.lineNumber();
        }

        if (line == -1) {
            /*
             * Treat code before the beginning of the first line table
             * entry as part of the first line.
             * javac will generate code
             * like this for some local classes. This "prolog" code contains
             * assignments from locals in the enclosing scope to synthetic
             * fields in the local class. 
             */
            line = ((Location)lineLocations.get(0)).lineNumber();
        }
        return line;
    }


    public List variables() throws AbsentInformationException {
        return getVariables();
    }

    public List variablesByName(String name) throws AbsentInformationException {
        List variables = getVariables();
        
        List retList = new ArrayList(2);
        Iterator iter = variables.iterator();
        while(iter.hasNext()) {
            LocalVariable variable = (LocalVariable)iter.next();
            if (variable.name().equals(name)) {
                retList.add(variable);
            }
        }
        return retList;
    }

    public List arguments() throws AbsentInformationException {
        List variables = getVariables();
        
        List retList = new ArrayList(variables.size());
        Iterator iter = variables.iterator();
        while(iter.hasNext()) {
            LocalVariable variable = (LocalVariable)iter.next();
            if (variable.isArgument()) {
                retList.add(variable);
            }
        }
        return retList;
    }
    
    public byte[] bytecodes() {
        byte[] bytecodes = (bytecodesRef == null) ? null : 
                                     (byte[])bytecodesRef.get();
        if (bytecodes == null) {
            try {
                bytecodes = JDWP.Method.Bytecodes.
                                 process(vm, declaringType, ref).bytes;
            } catch (JDWPException exc) {
                throw exc.toJDIException();
            }
            bytecodesRef = new SoftReference(bytecodes);
        }
        /*
         * Arrays are always modifiable, so it is a little unsafe
         * to return the cached bytecodes directly; instead, we 
         * make a clone at the cost of using more memory.
         */
        return (byte[])bytecodes.clone();
    }

    int argSlotCount() throws AbsentInformationException {
        if (argSlotCount == -1) {
            getVariables();
        }
        return argSlotCount;
    }

    /**
     * @return a text representation of the declared return type
     * of this method.
     */
    public String returnTypeName() {
        return signatureParser.typeName();
    }

    String returnSignature() {
        return signatureParser.signature();
    }

    public Type returnType() throws ClassNotLoadedException {
        return findType(returnSignature());
    }

    public Type findType(String signature) throws ClassNotLoadedException {
        ReferenceTypeImpl enclosing = (ReferenceTypeImpl)declaringType(); 
        return enclosing.findType(signature);
    }


    public List argumentTypeNames() {
        return signatureParser.argumentTypeNames();
    }

    public List argumentSignatures() {
        return signatureParser.argumentSignatures();
    }

    Type argumentType(int index) throws ClassNotLoadedException {
        ReferenceTypeImpl enclosing = (ReferenceTypeImpl)declaringType(); 
        String signature = (String)argumentSignatures().get(index);
        return enclosing.findType(signature);
    }

    public List argumentTypes() throws ClassNotLoadedException {
        int size = argumentSignatures().size();
        ArrayList types = new ArrayList(size);
        for (int i = 0; i < size; i++) {
            Type type = argumentType(i);
            types.add(type);
        }

        return types;
    }

    public int compareTo(Object object) {
        Method method = (Method)object;
        ReferenceTypeImpl declaringType = (ReferenceTypeImpl)declaringType();
        int rc = declaringType.compareTo(method.declaringType());
        if (rc == 0) {
            rc = declaringType.indexOf(this) - 
                    declaringType.indexOf(method);
        }
        return rc;
    }
    
    public boolean isAbstract() {
        return isModifierSet(VMModifiers.ABSTRACT);
    }
 
    public boolean isSynchronized() {
        return isModifierSet(VMModifiers.SYNCHRONIZED);
    }

    public boolean isNative() {
        return isModifierSet(VMModifiers.NATIVE);
    }

    public boolean isConstructor() {
        return name().equals("<init>");
    }

    public boolean isStaticInitializer() {
        return name().equals("<clinit>");
    }

    /*
     * A container class for the argument to allow
     * proper type-checking.
     */
    class ArgumentContainer implements ValueContainer {
        int index;

        ArgumentContainer(int index) {
            this.index = index;
        }
        public Type type() throws ClassNotLoadedException {
            return argumentType(index);
        }
        public String typeName(){
            return (String)argumentTypeNames().get(index);
        }
        public String signature() {
            return (String)argumentSignatures().get(index);
        }
        public Type findType(String signature) throws ClassNotLoadedException {
            return MethodImpl.this.findType(signature);
        }
    }

    void prepareArgumentsForInvoke(List arguments) 
                         throws ClassNotLoadedException, InvalidTypeException {

        int size = arguments.size();
        for (int i = 0; i < size; i++) {
            Value value = (Value)arguments.get(i);
            value = ValueImpl.prepareForAssignment(value, 
                                                   new ArgumentContainer(i));
            arguments.set(i, value);
        }
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append(declaringType().name());
        sb.append(".");
        sb.append(name());
        sb.append("(");
        boolean first = true;
        for (Iterator it = argumentTypeNames().iterator(); it.hasNext();) {
            if (!first) {
                sb.append(", ");
            }
            sb.append((String)it.next());
            first = false;
        }
        sb.append(")");
        return sb.toString();
    }

    private SoftLineInfo getLocations() {
        /*
         * Unlike variables, we do not make a special check for native
         * methods. Native methods are defined to have 0-length line number
         * tables. 
         */

        SoftLineInfo info = (softLineInfoRef == null) ? null :
                                     (SoftLineInfo)softLineInfoRef.get();
        if (info != null) {
            return info;
        }

        JDWP.Method.LineTable lntab = null;
        try {
            lntab = JDWP.Method.LineTable.process(vm, declaringType, ref);
        } catch (JDWPException exc) {
            if (exc.errorCode() == JDWP.Error.NATIVE_METHOD) {
                Map lineMapper = Collections.unmodifiableMap(new HashMap());
                List lineLocations = Collections.unmodifiableList(new ArrayList(0));
                info = new SoftLineInfo(lineMapper, lineLocations);
                softLineInfoRef = new SoftReference(info);
                return info;
            }
            /*
             * Note: the absent info error shouldn't happen here
             * because the first and last index are always available.
             */
            throw exc.toJDIException();
        }

        int count  = lntab.lines.length;

        List lineLocations = new ArrayList(count);
        Map lineMapper = new HashMap();
        for (int i = 0; i < count; i++) {
            long bci = lntab.lines[i].lineCodeIndex;
            int lineNumber = lntab.lines[i].lineNumber;

            /* 
             * Some compilers will point multiple consecutive
             * lines at the same location. We need to choose
             * one of them so that we can consistently map back
             * and forth between line and location. So we choose
             * to record only the last line entry at a particular
             * location.
             */
            if ((i + 1 == count) || (bci != lntab.lines[i+1].lineCodeIndex)) {
                // Remember the largest/smallest line number
                if (lineNumber > highestLine) {
                    highestLine = lineNumber;
                }
                if ((lineNumber < lowestLine) || (lowestLine == -1)) {
                    lowestLine = lineNumber;
                }
                Location loc = new LocationImpl(virtualMachine(), this, 
                                                bci, lineNumber);

                // Add to the location list
                lineLocations.add(loc);
    
                // Add to the line -> locations map
                Integer key = new Integer(lineNumber);
                List mappedLocs = (List)lineMapper.get(key);
                if (mappedLocs == null) {
                    mappedLocs = new ArrayList(1);
                    lineMapper.put(key, mappedLocs);
                }
                mappedLocs.add(loc);
            }
        }

        /*
         * firstIndex, lastIndex, and startLocation need to be 
         * retrieved only once since they are strongly referenced.
         */
        if (location == null) {
            firstIndex = lntab.start;
            lastIndex = lntab.end;
            /*
             * The startLocation is the first one in the 
             * location list if we have one;
             * otherwise, we construct a location for a 
             * method start with no line info
             */
            if (count > 0) {
                location = (Location)lineLocations.get(0);
            } else {
                location = new LocationImpl(virtualMachine(), this, 
                                                 firstIndex); 
            }
        }

        lineMapper = Collections.unmodifiableMap(lineMapper);
        lineLocations = Collections.unmodifiableList(lineLocations);
        info = new SoftLineInfo(lineMapper, lineLocations);
        softLineInfoRef = new SoftReference(info);
        return info;
    }

    private List getVariables() throws AbsentInformationException {
        if (absentVariableInformation) {
            throw new AbsentInformationException();
        }

        if (isNative()) {
            throw new NativeMethodException();
        }

        List variables = (variablesRef == null) ? null :
                                     (List)variablesRef.get();
        if (variables != null) {
            return variables;
        }

        JDWP.Method.VariableTable vartab = null;
        try {
            vartab = JDWP.Method.VariableTable.
                                     process(vm, declaringType, ref);
        } catch (JDWPException exc) {
            if (exc.errorCode() == JDWP.Error.ABSENT_INFORMATION) {
                absentVariableInformation = true;
                throw new AbsentInformationException();
            } else {
                throw exc.toJDIException();
            }
        }

        // Get thenumber of slots used by argument variables
        argSlotCount = vartab.argCnt;
        int count = vartab.slots.length;
        variables = new ArrayList(count);
        for (int i=0; i<count; i++) {
            JDWP.Method.VariableTable.SlotInfo si = vartab.slots[i];

            /*
             * Skip "this*" entries because they are never real
             * variables from the JLS perspective.
             */
            if (!si.name.startsWith("this$") && !si.name.equals("this")) {
                Location scopeStart = new LocationImpl(virtualMachine(),
                                                       this, si.codeIndex);
                Location scopeEnd = 
                    new LocationImpl(virtualMachine(), this,
                                     si.codeIndex + si.length - 1);
                LocalVariable variable = 
                    new LocalVariableImpl(virtualMachine(), this, 
                                          si.slot, scopeStart, scopeEnd,
                                          si.name, si.signature);   
                // Add to the variable list
                variables.add(variable);
            }
        }

        variables = Collections.unmodifiableList(variables);
        variablesRef = new SoftReference(variables);
        return variables;
    }
}
