/*
 *  Copyright (c) 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "COPYRIGHT"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  SUN
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 */
package  sun.beans.ole;


import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.lang.reflect.*;
import java.util.Vector;
import java.io.*;
import sun.beanbox.*;


/**
 * Packager
 *
 * Java Application front end to trigger the Bean packaging process
 *
 * Author :
 *   Jerome Dochez
 *   version 1.1       
 */


class Packager implements java.awt.event.ActionListener, 
	  		  java.awt.event.WindowListener {

    /*
     * main function of the packaging application. If no arguments are
     * passed, open the UI otherwise parse the parameters and perform
     * a quiet generation
     */
    public static void main(String arg[]) {
      
	new Packager(arg, true);
	System.exit(0);
    }

    /**
     * <p>
     * Create the Packager wizard and display it 
     * </p>
     * @param arg[] arguments for the typelib generation
     * @param waitForCompletion true if the caller wants to wait for the 
     * completion of the packaging process before this function returns, 
     */
    public Packager(String arg[], boolean waitForCompletion) {

	
	// Check for parameters
	if (arg.length==0) {
	    targetDir = 
		new TextField(System.getProperty("java.home") + 
						 File.separator + 
						 "classes");		  

	    frame = new Frame(getMessage("frame.title")); 
	    frame.addWindowListener(this);
	    isCompleted = false;

	    openWizard();
	    while (waitForCompletion && !isCompleted)
		Thread.yield();
	}
	else {
	    userFeedback = new TextFeedback(System.out);
	    StubInformation stubInfo = extractArguments(arg);
	    // Check the stubInfo...
	    if (isValidStubInfo(stubInfo))
		generate(stubInfo);
	} 

    }

    /*
     * Check the stubInfo for completeness
     * @param stubInfo to study
     */ 
    private boolean isValidStubInfo(StubInformation stubInfo) {

	// If no stub Info, nothing we can do
	if (stubInfo==null)
	    return false;

	String jarFileName = stubInfo.getJarFileName();

	// If no jar file name present, nothing we can do
	if (jarFileName==null) {
	    userFeedback.printText(getMessage("error.jarnotfound"));
	    return false;
	}

	// Let's try to load the jar file 
	JarLoader jarLoader= loadJar(jarFileName);
	if (jarLoader==null) {
	    userFeedback.printText(getMessage("error.invalidjar"));
	    return false;
	}

	// The jar file exist, load it
	JarInfo jarInfo;
	try {
	    jarInfo = jarLoader.loadJar();
	} catch (java.io.IOException e) {
		userFeedback.printText(
		    getMessage("error.jarunreadeable") + jarFileName);
		return false;
	}

	// Get the bean name,
	String beanName = stubInfo.getSourceBeanName();
	if (beanName==null) {
	    if (jarInfo.getCount()==0) {
		userFeedback.printText(getMessage("error.nobeansinjar") +
				       jarFileName);
		return false;
	    }
	    beanName = jarInfo.getName(0);
	    stubInfo.setSourceBeanName(beanName);
	}
    
	// load the bean.
	OleBeanInfo beanInfo = loadBean(jarFileName, beanName);
	if (beanInfo==null) {
	    userFeedback.printText(beanName+getMessage("error.beannotfound")+
				   jarFileName);
	    return false;
	}
    
	// So far, so good, now we have all we need to generate
	stubInfo.setOleBeanInfo(beanInfo);
    
	// Get the ActiveX name for this JavaBeans component
	if (stubInfo.getActiveXName()==null) 
	    stubInfo.setActiveXName(getActiveXNameFromJavaBeans(beanName));

	String targetDir = stubInfo.getTargetDir();
	if (targetDir==null) {
	    java.util.Properties prop = System.getProperties();
	    targetDir = prop.getProperty("user.dir", null);
	}
	stubInfo.setTargetDir(targetDir);
	File target = new File(targetDir);
	if (target.exists())
	    return true;
	else {
	    userFeedback.printText(getMessage("error.directory") + targetDir);
	    return false;
	}
    }

    /* 
     * return an ActiveX name from a JavaBeans name
     * @param JavaBeans name, including the package
     * @return a suitable ActiveX name for the specified JavaBeans component
     */
    private String getActiveXNameFromJavaBeans(String javaBeansName) {

	int packageSep = javaBeansName.lastIndexOf('.');
	if (packageSep==-1) // No package name
	    return javaBeansName;
	else 
	    return javaBeansName.substring(packageSep+1,
					   javaBeansName.length());
    }

    /*
     * Extract all parameters from the command line arguments.
     * @param arg array of string containing the arguments
     */ 
    public StubInformation extractArguments(String arg[]) {

	int argCount = arg.length;
	StubInformation stubInfo = new StubInformation();
	stubInfo.setEventGeneration(true);
	stubInfo.setRegister(true);
    
	for (int i=0;i<argCount;i++) {
	    String stringArg = arg[i];
	    if (stringArg.equalsIgnoreCase("-help") || 
		stringArg.equalsIgnoreCase("-?")) {
		displayHelp();
		return null;
	    }
	    try {
		if (arg[i].equalsIgnoreCase("-jar"))
		    stubInfo.setJarFileName(arg[++i]);
		if (arg[i].equalsIgnoreCase("-n"))
		    stubInfo.setSourceBeanName(arg[++i]);
		if (arg[i].equalsIgnoreCase("-ax")) 
		    stubInfo.setActiveXName(arg[++i]);
		if (arg[i].equalsIgnoreCase("-o"))
		    stubInfo.setTargetDir(arg[++i]);
		if (arg[i].equalsIgnoreCase("-awt"))
		    stubInfo.setEventGeneration(false);
		if (arg[i].equalsIgnoreCase("-noreg"))
		    stubInfo.setRegister(false);
	    } catch (ArrayIndexOutOfBoundsException e) {
		System.out.println(getMessage("error.badparameters"));
		displayHelp();
		return null;
	    }
	}
	return stubInfo;      
    } 

    /*
     * Display help messages for command line usage of the 
     * packager step
     */
    private void displayHelp() {
	System.out.println(getMessage("packager.title") + 
			   getMessage("packager.version") + "" +
			   pluginVersion);
	System.out.println(getMessage("parms.title"));
	System.out.println("        -jar " + getMessage("parms.file"));
	System.out.println("         " + getMessage("parms.fileex"));
	System.out.println("        -n " +getMessage("parms.string"));
	System.out.println("         " + getMessage("parms.stringex"));
	System.out.println("        -ax " + getMessage("parms.string"));
	System.out.println("         " + getMessage("parms.activex"));
	System.out.println("        -o " + getMessage("parms.directory"));
	System.out.println("         " + getMessage("parms.directoryex"));
	System.out.println("        -noreg");
	System.out.println("         " + getMessage("parms.noregex"));
    }
    
    /*
     * Create the Packager wizard and display it 
     */
    public void openWizard() {

	centerPanel = new Panel();

	frame.setLayout(new BorderLayout());
	frame.setBackground(Color.lightGray);
	Font titleFont = new Font("TimesRoman", Font.BOLD | Font.ITALIC, 20);
	Label desc = new Label(getMessage("packager.title") +
			       getMessage("packager.version") + "" +
			       pluginVersion,
			       Label.CENTER);
	desc.setFont(titleFont);
	frame.add(desc,"North"); 
      
	// add the navigation buttons Next and Previous
	Panel navigationPanel = new Panel();
	Font buttonFont = new Font("TimesRoman", Font.PLAIN, 14);
	nextButton = new Button(getMessage("panel1.next"));
	nextButton.setFont(buttonFont);
	nextButton.addActionListener(this);
	previousButton = new Button(getMessage("panel1.back"));
	previousButton.setFont(buttonFont);
	previousButton.addActionListener(this);
	navigationPanel.add(previousButton);
	navigationPanel.add(nextButton);
      
	// Make the Wizard Panel
	currentStep = 1;
	Panel wizardPanel = new Panel();
	wizardPanel.setLayout(new BorderLayout());
	wizardPanel.add(centerPanel,"Center");
	wizardPanel.add(navigationPanel, "South");

	updateCenterPanel();
      
	frame.add(wizardPanel,"Center");
	frame.add(new Label(getMessage("panel1.copyright"),Label.CENTER),
	    "South");
	frame.pack();
	frame.setBounds(200,200,500,260);
	//setResizable(false);
	frame.show();
    }   

    /*
     * Update the center panel of the wizard depending which step we are in
     */
    private void updateCenterPanel() {

	// Set the step title
	Panel advicePanel = new Panel();
	advicePanel.setLayout(new GridLayout(2,1,10,10));
	Font subTitleFont = new Font("TimesRoman", Font.BOLD, 14);
	Label stepTitle = new Label(getMessage("panel1.beforeStep") +
	   currentStep + getMessage("panel1.afterStep"),Label.CENTER);
	stepTitle.setFont(subTitleFont);
	advicePanel.add(stepTitle);

	Panel widgetPanel = new Panel();
	// add the step information 
	switch (currentStep) {
	case 1 : 
	    // Info 
	    advicePanel.add(new Label(getMessage("panel1.description"), 
				      Label.CENTER));

	    // Widgets
	    widgetPanel.setLayout(new GridLayout(2,1,0,0));
	    Panel subWidgetPanel = new Panel();
	    subWidgetPanel.setLayout(new BorderLayout());
	    browseButton = new Button(getMessage("panel1.browse"));
	    browseButton.addActionListener(this);
	    subWidgetPanel.add(browseButton,"East");

	    widgetPanel.add(jarName);
	    widgetPanel.add(subWidgetPanel);
	    previousButton.setEnabled(false);
	    break;
	case 2:
	    advicePanel.add(new Label(getMessage("panel2.description"),
				      Label.CENTER));
	    previousButton.setEnabled(true);
	    // Load the jar and get the jar info
	    try {
		JarLoader jl = loadJar(jarName.getText());
		JarInfo ji = jl.loadJar();

		if (ji.getCount()>5)
		    beanList = new List(5,false);	
		else
		    beanList = new List(ji.getCount(), false);
	   
		// Get the list of Beans in the jar file
		for (int i=0;i<ji.getCount();i++) {
		    beanList.add(ji.getName(i));
		}
	    } catch (Throwable e) {
		System.err.println("->" + e.toString());
		e.printStackTrace();
	    }
	    beanList.select(0);
	    widgetPanel.removeAll();
	    widgetPanel.setLayout(new GridLayout(1,1,0,0));
	    widgetPanel.add(beanList);
	    break;
	case 3:
	    advicePanel.add(new Label(getMessage("panel3.description") +
				      beanList.getSelectedItem(),
				      Label.CENTER));
	    beanName = new TextField(getActiveXNameFromJavaBeans(
					      beanList.getSelectedItem()));
	    widgetPanel.setLayout(new GridLayout(1,1,0,0));
	    widgetPanel.add(beanName);	   
	    break;
	case 4:
	    advicePanel.add(new Label(getMessage("panel4.description"), 
				      Label.CENTER));
	    widgetPanel.setLayout(new GridLayout(1,1,0,0));

	    widgetPanel.add(targetDir);
	    nextButton.setEnabled(true);
	    break;
	case 5:
	    advicePanel.add(new Label(getMessage("panel5.description"),
				      Label.CENTER));

	    widgetPanel.setLayout(new GridLayout(3,1,0,0));
	    startButton = new Button(getMessage("panel5.start"));
	    startButton.addActionListener(this);
	    registerCheckbox = new Checkbox(getMessage("panel5.register"));
	    registerCheckbox.setState(true);
	    widgetPanel.add(registerCheckbox);
	    widgetPanel.add(startButton);
	    nextButton.setEnabled(false);
	    break;
	   
	}
	// Add the subPanels to the center panel of the packager
	centerPanel.removeAll();
	centerPanel.setLayout(new BorderLayout());
	centerPanel.add(advicePanel, "North");
	centerPanel.add(widgetPanel, "South");
	centerPanel.validate();
    }

    /* 
     * Check that the user has specfied correct information before moving
     * on the next step
     */
    private boolean checkInvariant() {
	switch (currentStep) {
	case 1 :
	    if (loadJar(jarName.getText())==null) {
		new InformationDialog(frame, getMessage("packager.error"), 
				      getMessage("error.nulljarfile"));
		return false;
	    }
	    break;
	case 2 :
	    if (beanList.getSelectedItem()==null) {
		new InformationDialog(frame, getMessage("packager.error"),
				      getMessage("error.nullbeanname"));
		return false;
	    }
	    break;
	case 3:
	    if (beanName.getText().length()==0) {
		new InformationDialog(frame,  getMessage("packager.error"),
				      getMessage("error.nullactivexname"));
		return false;
	    }
	    break;  
	case 4:
	    if (targetDir.getText().length()==0) {
		new InformationDialog(frame,  getMessage("packager.error"),
				      getMessage("error.directory"));
		return false;
	    } else {
		File target = new File(targetDir.getText());
		if (target.exists())
		    return true;
		else {
		    new InformationDialog(frame,  getMessage("packager.error"),
					 getMessage("error.directory"));
		    return false;	  
		}
	    }
	}
	return true;
    }

    /* 
     * Load a jar file
     * @param Jar file name (including path)
     * @return JarLoader for the jar or null if the jar cannot be loaded
     */
    private JarLoader loadJar(String jarFileName) {

	// Load the Jar file
	JarLoader jl = null;
	try {
	    jl = new JarLoader(jarFileName);
	    if (jl == null) {
		return null;
	    }
	} catch (Exception ex) {
	    return null;
	}
	return jl;
    }
  
    /*
     * Button events notification
     * @param information about the button event
     */
    public void actionPerformed(java.awt.event.ActionEvent e) {
	if (e.getSource() == nextButton) {
	    if (checkInvariant()) {
		currentStep++;
		updateCenterPanel();
	    }
	}
	if (e.getSource() == previousButton) {
	    currentStep--;
	    updateCenterPanel();
	}
	if (e.getSource() == browseButton) {
	    String jarFileName = getJarFileName();
	    if (jarFileName!=null) 
		jarName.setText(jarFileName);	  
	}
	if (e.getSource() == startButton) {
	    generate();
	}
    }

    /*
     * Window Listener implementation, just listen for the window close event
     */
    public void windowActivated(WindowEvent e) { }
    public void windowClosed(WindowEvent e) {
	isCompleted = true;
	frame.removeWindowListener(this);
    }
    public void windowClosing(WindowEvent e) {
    	frame.dispose();
    }
    public void windowDeactivated(WindowEvent e) {}
    public void windowDeiconified(WindowEvent e) {}
    public void windowIconified(WindowEvent e) {}
    public void windowOpened(WindowEvent e) {}

    /* 
     * Invoke the FileDialog to get the jar file name from the user
     * @return jar file name selected or null if nothing selected
     */
    public String getJarFileName() {
    
	FileDialog fd = new FileDialog(frame, getMessage("packager.loadfile"),
				       FileDialog.LOAD);
	
	// Added by Stanley Man-Kit Ho on 7/22/97.
	fd.setFile(jarName.getText());

	fd.show();
	String fname = fd.getFile();
	if (fname == null) {
	    return null;
	}
	String dname = fd.getDirectory();
	File file = new File(dname, fname);
	if (file.exists())
	    return dname + fname;
	else 
	    return null;
    }

    /*
     * load the bean from a jar file and return the OleBeanInfo on it
     * @param jar file name
     * @param bean name inside the jar file
     * @return OleBeanInfo for this bean or null if the bean cannot be loaded
     */
    private OleBeanInfo loadBean(String jarFileName, String beanName) {

	// Load the Jar file
	JarLoader jl = loadJar(jarFileName);
	if (jl==null)
	    return null;
	JarInfo ji;
	try {
	    ji = jl.loadJar();
	} catch (java.io.IOException e) {
		userFeedback.printText(getMessage("error.jarunreadeable")+
				       jarFileName);
		return null;
	}

	// Get the list of Beans in the jar file
	int beanIndex = -1;
	for (int i=0;i<ji.getCount();i++) {
	    if (ji.getName(i).equals(beanName)) 
		beanIndex=i;
	}
	
	// Get the BeanInfo for this guy
	if (beanIndex!=-1) {
	    OleBeanInfo oleBeanInfo;
	    try {
		Object b = ji.getInstance(beanName);
		if (b != null) {
		    Class c = b.getClass();
		    return new OleBeanInfo(c);
		}   
	    } catch (Throwable e) {
		String[] messages = { 
		    e.toString(),
		    getMessage("packager.statustitle")};
		userFeedback.warn(getMessage("error.introspection"), messages);
		userFeedback.printText(getMessage("error.introspectionex") +
				       beanName);
		return null;
	    }
	}
	return null;
    }


    /*
     * Fill the generation information from the fields.
     */
    private void generate() {
        OleBeanInfo                 beanInfo;


        EventSetDescriptor  outgoingInterfaces[];
        Vector              eventsVector;
        Vector              propertiesVector;



        // Inform the user.
	centerPanel.removeAll();
	centerPanel.setLayout(new GridLayout(2,1,0,0));
	centerPanel.add(new Label(getMessage("packager.statustitle")));
        infoLabel = new TextField(" ",80);
        infoLabel.setEditable(false);
        centerPanel.add(infoLabel);
	centerPanel.validate();
	userFeedback = new VisualFeedback(frame, infoLabel);

	Object bean = loadBean(jarName.getText(), beanList.getSelectedItem());


	// Create the Stub generation information holder
	StubInformation stubInfo = new StubInformation();
	stubInfo.setSourceBeanName(beanList.getSelectedItem());
	stubInfo.setJarFileName(jarName.getText());
	stubInfo.setOleBeanInfo(loadBean(jarName.getText(), 
					 beanList.getSelectedItem()));
	stubInfo.setActiveXName(beanName.getText());
	stubInfo.setTargetDir(targetDir.getText());
	stubInfo.setRegister(registerCheckbox.getState());


	generate(stubInfo);
    }

    /*
     * Generate the stub files and the typelib 
     */
    private void generate(StubInformation stubInfo) {
				     
	boolean res=false;

	// Save output directory
	String targetDir = stubInfo.getTargetDir();
	if (stubInfo.getOleBeanInfo()!=null) {

	    // So far, so good, proceed with type library generation
	    userFeedback.printText(getMessage("packager.statusbegin"));
	    res = stubInfo.generateTypeLib(true);
	    userFeedback.printText(getMessage("packager.statusend"));
	}
        userFeedback.warn(getMessage("packager.statustitle"),
			 (res?getMessage("packager.statussuccess"):
			      getMessage("packager.statusfailed")));
	if (frame!=null)
	    frame.dispose();
    }

    /*
     * <p>
     * Get a internationalized string form the resource file
     * </p>
     * @param key of the resource
     * @return the localized resource
     */
    public String getMessage(String key) {
	try {
	    return rb.getString(key);
	} catch (java.util.MissingResourceException ex) {
	    return key;
	}
    }

    static {
	System.loadLibrary("packager");
    }

    
			     
    private Frame  frame;
    private Panel  centerPanel;
    private Button nextButton;
    private Button previousButton;
    private Button browseButton;
    private List   beanList;  
    private Button startButton;
    private int currentStep;

    private TextField targetDir;
    private TextField jarName =  new TextField(System.getProperty("application.home"));
    private TextField beanName;
    private Checkbox  registerCheckbox;
    private TextField infoLabel;
    private UserFeedback userFeedback;
    private boolean   isCompleted; // True when the packaging is completed
    private java.util.ResourceBundle rb = 
       java.util.ResourceBundle.getBundle("sun.beans.ole.resources.Packager");

    final private String pluginVersion = "1.3.1";
}

