/*
 * @(#)ClassLoaderInfo.java	1.22 01/02/09
 *
 * Copyright (c) 1995-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * CopyrightVersion 1.1 
 *
 */

package sun.plugin;

/**
 * This class keps track of information about active applet class loaders.
 * The classloaders are identified by their codebase URL.
 *
 * We keep around a pool of recently used class loaders.
 *
 * @author Graham Hamilton
 */

import java.io.PrintStream;
import java.util.Collection;
import java.util.Iterator;
import java.util.HashMap;
import java.util.ArrayList;
import java.net.URL;
import sun.applet.*;
import sun.plugin.usability.*;

class ClassLoaderInfo 
{
    private URL codebase;
    private int references;
    private HashMap jars;
    private boolean locked;
    private boolean notcache = false;
    private static boolean initialized;

    // "infos" is a list of ClassLoaderInfos that are available for use.
    private static HashMap infos = new HashMap();

    // "zombies" is a list of ClassLoaderInfos that currently have
    // a reference count of zero.  We keep up to zombieLimit of these
    // guys available for resurrection.  The least recently used is
    // at the front, the more recently used at the end.
    // All entries in zombies are also in infos.
    private static int zombieLimit = 0;
    private static ArrayList zombies = new ArrayList();

    private static synchronized void initialize() {
	if (initialized) 
	    return;
	
	initialized = true;

	reset();
    }


    /** 
     * Reset setting.
     */
    public static synchronized void reset() 
    {
	initialized = true;

	zombieLimit = 0;

	String value = (String) java.security.AccessController.doPrivileged(
			    new sun.security.action.GetPropertyAction("javaplugin.classloader.cache.enabled"));

	if (value == null || value.equals("true"))
	{
	    // Set zombie cache sizes
	    zombieLimit = Integer.getInteger("javaplugin.classloader.cache.sizes", 10).intValue();
	}

/*
	if (zombieLimit > 0) 
	{
	    System.out.println(sun.plugin.AppletViewer.getMessage("jar_cache_enabled"));
	} else {
	    System.out.println(sun.plugin.AppletViewer.getMessage("jar_cache_disabled"));
	}
*/
    }


    /**
     * Clear ClassLoader cache.
     */
    public static synchronized void clearClassLoaderCache()
    {
	// Iterate through the zombie list - this represents 
	// all the stopped applets that have been cached.
	// Cleanup all classloaders and its information.
	//
	Iterator iter = zombies.iterator();

	while (iter.hasNext())
	{
	    ClassLoaderInfo victim = (ClassLoaderInfo) iter.next();

	    if (victim != null)
	    {	
	        infos.remove(victim.codebase.toString());
	        AppletPanel.flushClassLoader(victim.codebase);		
	    }
	}

	zombies.clear();


	// Iterate through the infos list - this represents all 
	// the running applets and stopped applets in the system.
	// Tag each of them as not-cache.
	//
	Collection collection = infos.values();

	if (collection != null)
	{
	    iter = collection.iterator();

	    while (iter.hasNext())
	    {
		ClassLoaderInfo target = (ClassLoaderInfo) iter.next();

		if (target != null)
		    target.notcache = true;
	    }
	} 

	// Flush all existing classloaders
	AppletPanel.flushClassLoaders();		
    }

    /**
     * Dump ClassLoader cache.
     */
    public static synchronized void dumpClassLoaderCache(PrintStream ps)
    {
	StringBuffer buffer = new StringBuffer();
	
	buffer.append("Dump classloader list ...\n");

	Collection collection = infos.values();

	if (collection != null)
	{
	    Iterator iter = collection.iterator();

	    while (iter.hasNext())
	    {
		ClassLoaderInfo info = (ClassLoaderInfo) iter.next();

		if (info != null)
		{
		    boolean isZombie = zombies.contains(info);

		    buffer.append("    codebase=" + info.codebase);
		    buffer.append(", zombie=" + isZombie);
		    buffer.append(", cache=" + !info.notcache);
		    buffer.append(", info=" + info);
		    buffer.append("\n");
		}
	    }
	}

	buffer.append("Done.");

	ps.println(buffer.toString());
    }

    /**
     * Find ClassLoaderInfo for the given AppletPanel.
     *
     * If we don't have any active information, a new ClassLoaderInfo is
     * created with an initial reference count of zero.
     */
    static synchronized ClassLoaderInfo find(AppletPanel panel) 
    {
	initialize();

	URL codebase = panel.getCodeBase();

	if (codebase == null)
	    return null;

	ClassLoaderInfo result = (ClassLoaderInfo)infos.get(codebase.toString());

	if (result == null) 
	{
	    // ClassLoaderInfo cannot be found, so we have never seen 
	    // this ClassLoader before. In this case, add it to the infos.
	    //
	    result = new ClassLoaderInfo(codebase);
	    infos.put(codebase.toString(), result);
	} 
	else 
	{
	    // We have seen this classloader before, so the ClassLoader is 
	    // currently in use.
	    // We make sure this loader isn't on the zombies list.
	    //
	    zombies.remove(result);
	}

	return result;
    }

    /**
     * Add a retaining reference.
     */
    synchronized void addReference() 
    {
	references++;
	Trace.println("Referencing classloader: " + this + ", refcount=" + references, 
		      TraceFilter.JAVA_CONSOLE_ONLY);
    }

    /**
     * Remove a retaining reference.  If there are no references left
     * then we put it on the zombies list.
     */
    synchronized void removeReference() 
    {
	references--;

	Trace.println("Releasing classloader: " + this + ", refcount=" + references,
		      TraceFilter.JAVA_CONSOLE_ONLY);

	if (references < 0) 
	{
	    throw new Error("negative ref count???");
	}
	if (references == 0) 
	{
	    addZombie(this);
	}
    }

    /**
     *  Add the given ClassLoaderInfo to the zomboies list.
     * If there are too many zombies we get rid of some.
     */
    private static synchronized void addZombie(ClassLoaderInfo cli) 
    {
	if (cli.notcache)
	{
	    // If the classloader info is tagged as not-cache, 
	    // user have cleared the classloader cache, so 
	    // we will not cache the classloader.
	    //
	    zombies.remove(cli);    
	    infos.remove(cli.codebase.toString());
	    AppletPanel.flushClassLoader(cli.codebase);
	}
	else
	{
	    // Classloader should be cached.

	    Trace.println("Caching classloader: " + cli,
			  TraceFilter.JAVA_CONSOLE_ONLY);

	    // Add the ClassLoaderInfo to the end of the zombies list.
	    zombies.add(cli);

	    Trace.println("Current classloader cache size: " + zombies.size(),
			  TraceFilter.JAVA_CONSOLE_ONLY);

	    // If there are too many zombies, kill unused classloader.
	    if (zombies.size() > zombieLimit) 
	    {    
		ClassLoaderInfo victim = (ClassLoaderInfo)zombies.get(0);

		Trace.println("Number of cached classloaders over " + zombieLimit + ", unreference " + victim,
			      TraceFilter.JAVA_CONSOLE_ONLY);
	    		    
		// Flesh all references to classloader
		zombies.remove(0);
		infos.remove(victim.codebase.toString());
		AppletPanel.flushClassLoader(victim.codebase);
	    }
	}
    }

    private ClassLoaderInfo(URL codebase) {
	references = 0;
	this.codebase = codebase;
	jars = new HashMap();
    }

    synchronized void addJar(String name) {
	jars.put(name, name);
    }

    synchronized boolean hasJar(String name) {
	if (jars.get(name) != null) {
	    return true;
	} else {
	    return false;
	}
    }

    /*
    * Flag and utility routines for recording whether local .jar files in lib/app/
    * have been loaded or not.  This is used as a performance optimization
    * so that hasJar() does not need to be checked against the filesystem
    * each time an applet is loaded.
    */
  
    private boolean localJarsLoaded = false;
  
    public boolean getLocalJarsLoaded()  {
        return(localJarsLoaded);
    }
  
    public void setLocalJarsLoaded(boolean loadedFlag)  {
        localJarsLoaded = loadedFlag;
    }
  
    /**
     * Acquire the lock.  This is used to prevent two AppletPanels
     * trying to classload JARs at the same time.
     */
    public final synchronized void lock() throws InterruptedException {
	while (locked) {
	    wait();
	}
	locked = true;
    }

    /**
     * Release the lock.   This allows other people do to classloading.
     */
    public final synchronized void unlock() {
	locked = false;
	notifyAll();
    }
}
