/*
 * @(#)DebugOutputStream.java	1.18 01/02/09
 *
 *  Copyright (c) 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * CopyrightVersion 1.1
 *
 */

package sun.plugin;

/*
 * This class is used to redirect the srderr and stdout of the java run
 * time environement. The output is using the normal Win32 debug output
 * APIs that all debuggers should implement. 
 *
 * @version 1.1 
 * @date 12/27/97
 * @Author Jerome Dochez
 */

import java.io.IOException;
import java.io.OutputStream;
import sun.plugin.usability.Trace;

public class DebugOutputStream extends OutputStream 
{
    private OutputStream traceOutputStream;
    /**
     * <p>
     * Creates a new output stream for debugging tracing
     * All messages will be forwarded to the console window.
     * Another hook is used for subclasses to display into
     * another tracing facility.
     * </p>
     *
     * @param console The java console to use for output
     */
    DebugOutputStream(ConsoleWindow console, OutputStream traceStream) 
    {
	this.console = console;
	this.traceOutputStream = traceStream;
	count=0;
    }

    /** Flush the internal buffer */
    private void flushBuffer() throws IOException {
        if (count > 0) {
	    // forward to the console if open
	    if (console != null)  {
		trace(buf, 0, count);
	    }
	    count = 0;
        }
    }

    private void trace(byte b[], int off, int len) {
	if (console != null) {
	    console.append(new String(b, off, len));
	}
	try {
	    traceOutputStream.write(b, off, len);
	    traceOutputStream.flush();
	} catch (Exception e) {
	    console.append(e.toString());
	}

    }

    /**
     * Writes the specified byte to this buffered output stream. 
     *
     * @param      b   the byte to be written.
     * @exception  IOException  if an I/O error occurs.
     */
    public synchronized void write(int b) throws IOException {
	if (count >= buf.length) {
	    flushBuffer();
	}
	buf[count++] = (byte)b;
	if (b == '\n' || b == '\r')
	    flushBuffer();
    }

    /**
     * Writes <code>len</code> bytes from the specified byte array 
     * starting at offset <code>off</code> to this buffered output stream.
     *
     * <p> Ordinarily this method stores bytes from the given array into this
     * stream's buffer, flushing the buffer to the underlying output stream as
     * needed.  If the requested length is at least as large as this stream's
     * buffer, however, then this method will flush the buffer and write the
     * bytes directly to the underlying output stream.  Thus redundant
     * <code>BufferedOutputStream</code>s will not copy data unnecessarily.
     *
     * @param      b     the data.
     * @param      off   the start offset in the data.
     * @param      len   the number of bytes to write.
     * @exception  IOException  if an I/O error occurs.
     */
    public synchronized void write(byte b[], int off, int len) throws IOException {
	if (len >= buf.length) {
	    /* If the request length exceeds the size of the output buffer,
    	       flush the output buffer and then write the data directly.
    	       In this way buffered streams will cascade harmlessly. */
	    flushBuffer();
	    trace(b, off, len);
	    return;
	}
	if (len > buf.length - count) {
	    flushBuffer();
	}
	System.arraycopy(b, off, buf, count, len);
	count += len;

	if (b[len-1] == '\n' || b[len-1] == '\r')
	    flushBuffer();
    }

    /**
     * Flushes this buffered output stream. This forces any buffered 
     * output bytes to be written out to the underlying output stream. 
     *
     * @exception  IOException  if an I/O error occurs.
     * @see        java.io.FilterOutputStream#out
     */
    public synchronized void flush() throws IOException {
        flushBuffer();
    }

    /**
     * <p>
     * Method to provide an alternative tracing facility beside 
     * the console window. Messages could be traced for example 
     * in a debugger window.
     * </p>
     *
     * @param str trace message
     */
    protected void printDebug(String out) {}

    private int count;
    private byte[] buf = new byte[80];
    private ConsoleWindow console = null;
}
