/*
 * @(#)ExtensionInstallationImpl.java	1.7 00/07/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.plugin;

import sun.misc.ExtensionInstallationException;
import sun.misc.ExtensionInfo;
import sun.misc.ExtensionInstallationProvider;
import javax.swing.JOptionPane;
import javax.swing.JDialog;
import javax.swing.JProgressBar;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.Box;
import javax.swing.JOptionPane;
import java.awt.event.WindowEvent;
import java.util.jar.JarFile;
import java.util.jar.Attributes;
import java.util.jar.Attributes.Name;
import java.util.jar.Manifest;
import java.util.jar.JarEntry;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLClassLoader;
import java.net.JarURLConnection;
import java.net.MalformedURLException;
import java.net.SocketException;
import java.io.IOException;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.CodeSource;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.lang.reflect.Method;
import sun.plugin.security.PluginClassLoader;
import sun.plugin.usability.Trace;

/**
 * Provide facilities to install extensions in the standard extension
 * directory of the running JDK
 *
 * @author Jerome Dochez
 * @version 1.7, 07/18/00
 */
public class ExtensionInstallationImpl implements ExtensionInstallationProvider {

    public boolean installExtension(ExtensionInfo reqExtInfo, 
				    ExtensionInfo instExtInfo) 
				    throws ExtensionInstallationException 
    {
	Trace.extPrintln("Plugin requested to install " + reqExtInfo);
	String jarFileName = makePlatformDependent(
				 extractJarFileName(reqExtInfo.url));
	String platformURL = makePlatformDependent(reqExtInfo.url);

	if (!askUserForAknowledgment(reqExtInfo, instExtInfo)) {
	    throw new ExtensionInstallationException(
			       "User denied installation of " +
			       platformURL);
	}
	// Proceed with jar file download, check it is signed first
	try {
	    ClassLoader cl = checkForSignature(new URL(platformURL));
	    if (cl == null) {
		throw new ExtensionInstallationException(platformURL +
						     ": Jar File not signed");		
	    }
	    URL url = new URL("jar:"+ platformURL +"!/");
	    Trace.extPrintln("Downloading " + platformURL + " filename " + jarFileName);
	    
	    URLConnection uc = url.openConnection();
	    JarFile jarFile = ((JarURLConnection) uc).getJarFile();
	    installJarFile(cl, platformURL, jarFileName, jarFile);
	    
	} catch(SocketException e) {
	    Trace.extPrintln("Cannot download the jar File " + 
			       platformURL);
	    return false; 
	} catch(MalformedURLException e) {
	    Trace.extPrintException(e);
	    return false;
	} catch(IOException e) {
	    Trace.extPrintln("Cannot download the jar File " + platformURL);
	    return false;	    	   
	}

	return true;
    }

    /*
     * <p>
     * Check that the URL we download for installing the extension is 
     * properly signed by one of the Signing authority we support
     * </p>
     * 
     * @return ClassLoader on the signed jar file, null if the file
     * is not properly signed
     */
    private ClassLoader checkForSignature(final URL url) 
	throws  IOException,
		ExtensionInstallationException
    {
	ClassLoader cl = (ClassLoader) AccessController.doPrivileged
	    (
	     new PrivilegedAction() {
		public Object run() {

		    URL hostURL;
		    try {
			String host = url.getHost();
			hostURL = new URL(url.getProtocol(),  host, 
					  url.getPort(), "" );
			Trace.extPrintln("Host url is " + hostURL);
		    } catch(MalformedURLException e) {
			Trace.extPrintException(e);
			return null;
		    }
		    
		    PluginClassLoader ucl = new PluginClassLoader(hostURL);
		    ucl.addMyJar(hostURL, url.toString());
		    JarFile jarFile;

		    try {
			URL jarUrl = new URL("jar:"+ url +"!/");
			URLConnection uc = jarUrl.openConnection();
			jarFile = ((JarURLConnection) uc).getJarFile();
			Trace.extPrintln("Jar file is " + jarFile.getName());
		    } catch (IOException e) {
			Trace.extPrintException(e);
			return null;
		    }

		    boolean allTrusted = true;
		    for (Enumeration e = jarFile.entries(); e.hasMoreElements();)
		    {
			JarEntry entry = (JarEntry) e.nextElement();
			Trace.extPrintln("Entry " + entry.getName());
			if (entry.getName().endsWith(".class")) {
			    Trace.extPrintln("Class " + entry.getName());
			    String className = entry.getName().substring(0,
					  entry.getName().lastIndexOf(".class"));
			    className = className.replace('/','.');
			    Trace.extPrintln("Trying to load " + className);
			    try {
				Class c = ucl.loadClass(className);
				Trace.extPrintln(c + " signers " + c.getSigners());
				if (!c.isInterface() && c.getSigners() == null) {
				    allTrusted = false;
				}
			    } catch(ClassNotFoundException ex) {
			    }			     
			}
			if (!entry.getName().startsWith("META-INF")) {
			    try {
				// normal resource, we just read 
				InputStream is = jarFile.getInputStream(entry);
				byte[] buffer = new byte[8192];
				int n;
				while ((n = is.read(buffer, 0, buffer.length)) != -1) {
				// we just read. jis will throw a SecurityException
				// if  a signature/digest check fails.
				    
				}				
				is.close();
				Certificate[] certs = entry.getCertificates();
				Trace.extPrintln("For entry " + entry.getName() +
						   " certs are: ");
				if (certs != null) {
				    for (int i=0; i<certs.length; i++)
				    Trace.extPrintln(certs[i].toString());
				} else {
				    allTrusted = false;
				}				
			    } catch(Throwable ex) {
				allTrusted = false;
			    }			    
			}
		    }
		    if (allTrusted) {
			return ucl;
		    } else {
			return null;
		    }
		}
	    });
	return cl;
    }
	

    private boolean askUserForAknowledgment(ExtensionInfo reqExtInfo, 
					    ExtensionInfo instExtInfo) {
	String dialogText;
	if (instExtInfo != null) {
	    // this extension is already installed, let's check why we need to
	    // reinstall it.
	    int reason = instExtInfo.isCompatibleWith(reqExtInfo);
	    switch(reason) {
	    case ExtensionInfo.REQUIRE_SPECIFICATION_UPGRADE:
		dialogText = 
		    "The applet requires a newer version (specification" +
		    reqExtInfo.specVersion + ") of extension " + 
		    instExtInfo.title;
		break;
	    case ExtensionInfo.REQUIRE_IMPLEMENTATION_UPGRADE:
		dialogText = 
		    "The applet requires a newer version (implementation " +
		    reqExtInfo.implementationVersion + ") of extension " + 
		    instExtInfo.title;
		break;
	    case ExtensionInfo.REQUIRE_VENDOR_SWITCH:
		dialogText = "The applet require the " + reqExtInfo.vendor + 
		    "(" + reqExtInfo.vendorId + ") extension " + 
		    instExtInfo.title;		   
		break;
	    default: 
		dialogText = "The applet require the installation of " +
		    reqExtInfo.name + " from " + reqExtInfo.url;
	    }
	} else {
	    dialogText = "The applet require the installation of " 
		+ reqExtInfo.name;
	}
	
	dialogText = dialogText + "\n from " +
	    makePlatformDependent(reqExtInfo.url);

	Object[] options = { getMessage("security_dialog.buttonYes"),
			     getMessage("security_dialog.buttonNo"), 
			     getMessage("security_dialog.buttonAlways")};
	
	int ret = JOptionPane.showOptionDialog(null, dialogText, 
		      getMessage("security_dialog.caption"),
		      JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE, 
		      null, options, options[0]);				
	if (ret==0) {
	    Trace.extPrintln("Starting download from " + reqExtInfo.url);
	} else {
	    Trace.extPrintln("Download not granted by user");
	}						 		
	return (ret==0);
    }

    private void installJarFile(ClassLoader cl, String url, String jarFileName, JarFile jarFile) 
	throws IOException 
    {
	Trace.extPrintln("Starting the installation of " + jarFile.getName());
	Manifest man = jarFile.getManifest();
	if (man != null) {

	    // Get the extension directory, we'll need it anyway
	    String extDirs = (String) AccessController.doPrivileged
		(
		 new PrivilegedAction() {
		     public Object run() {
			 return System.getProperty("java.ext.dirs");
		     }}
		 );

	    StringTokenizer st = new StringTokenizer(extDirs, File.pathSeparator);
	    String destDir = st.nextToken();
	    Trace.extPrintln("Raw Extension in " + destDir);
	    
	    Attributes attr = man.getMainAttributes();
	    if (attr != null) {
		String s = attr.getValue(Name.MAIN_CLASS);
		if (s != null) {
		    Trace.extPrintln("Require java install ");
		    launchJavaInstaller(cl, s, destDir);
		    return;
		}
		s = attr.getValue(Name.EXTENSION_INSTALLATION);
		if (s != null) {
		    Trace.extPrintln("Require native installation");
		    launchNativeInstallation(s, jarFile, destDir);
		    return;
		} 
		// Raw jar file, just copy to the lib/ext directory
		installRawExtension(url, jarFileName, destDir);
	    }
	}
    }


    /**
     * Copy the extension to the lib/ext directory "As is"
     */
    private boolean installRawExtension(final String url,
                                        String jarFileName, 
					String destDir)
	throws IOException
    {
	String destFileName = destDir + File.separatorChar + jarFileName;

	final File destFile = new File(destFileName);

	Trace.extPrintln("Dest File would be " + destFile);
	Boolean res = (Boolean) AccessController.doPrivileged
	    (
	     new PrivilegedAction() {
		public Object run() {
                    URL jarUrl = null;                    
                    URLConnection conn = null;
		    InputStream is = null;
                    OutputStream os = null;

		    try {
                        jarUrl = new URL(url);
                        conn = jarUrl.openConnection();
			is = conn.getInputStream();
			os = new FileOutputStream(destFile);
			Copy(is, os);
                        is.close();
                        os.close();
		    } catch (IOException e) {
			Trace.extPrintException(e);
			return Boolean.FALSE;
		    }

		    return Boolean.TRUE;
		}		 
	    });
	return res.booleanValue();
    }

    private boolean launchJavaInstaller(final ClassLoader ucl,
					final String mainClass,
                                        final String destDir)
    {
	
	Boolean res;
	try {
	    res = (Boolean) AccessController.doPrivileged
		(
		 new PrivilegedAction() {
		     public Object run() {
			 try {
			     Trace.extPrintln("Trying to load " + mainClass);
			     Class cl = ucl.loadClass(mainClass);
			     Trace.extPrintln("Success " + cl);
			     Class[] parms = new Class[1];
			     parms[0] = java.lang.String[].class;
			     String[] parmsValues = new String[1];
			     parmsValues[0] = destDir;
			     Method m = cl.getMethod("main", parms);
			     Object[] values = new Object[1];
			     values[0] = parmsValues;
			     m.invoke(null, values);
			     Trace.extPrintln("Done !");
			 } catch (Throwable e) {
			     Trace.extPrintException(e);
			     return Boolean.FALSE;
			 }
			 return Boolean.TRUE;
		     }
		 });
	} catch (Throwable e) {
	    Trace.extPrintException(e);
	    return false;
	}
	return res.booleanValue();
    }

    private boolean launchNativeInstallation(final String installName, 
					     final JarFile jarFile, 
					     String destDir) 
    {
	Trace.extPrintln("Launching " + installName);
	Boolean success;
	// Let's look for the %PLATFORM% substring first
	try {
	    success = (Boolean) AccessController.doPrivileged
		(
		 new PrivilegedAction() {
		     public Object run() {
			 try {
			     String userDir = System.getProperty("user.dir");
			     InputStream is = jarFile.getInputStream(jarFile.getEntry(installName));
			     final String tempInstall = userDir + File.separator + installName;
			     FileOutputStream os = new FileOutputStream(tempInstall);
			     Copy(is, os);

			     Process p = Runtime.getRuntime().exec(tempInstall);

			     // The p.waitFor seems to prevent the normal termination
			     // of subprocesses, so I poll the subprocess every second to 
			     // get notified when the process has died. I use the 
			     // return code (which is available only once the process is
			     // finished) to get notified
			     JProgressBar progressBar = new JProgressBar(0, 100);
			     JDialog dialog = new JDialog((java.awt.Frame) null, 
							  "Java Extension Installation", false);
			     JPanel panel = new JPanel();
			     
			     panel.add(new JLabel("   Waiting for installation    "));
			     panel.add(new JLabel(installName));
			     panel.add(new JLabel("       to finish        "));
			     panel.add(progressBar);
			     dialog.getContentPane().add(panel);
			     progressBar.setVisible(true);
			     dialog.setSize(250,150);
			     dialog.setLocation(200,200);
			     dialog.show();

			     int iProgressBarValue = 0; 
			     boolean bCont = true;			     
			     while (bCont) {
				 try {
				     p.exitValue();
				     bCont = false; 
				 } catch (IllegalThreadStateException e) {
				     Thread.currentThread().sleep(1000);
				     iProgressBarValue+=10;
				     if (iProgressBarValue == 110) {
					 iProgressBarValue = 0;
				     }
				     progressBar.setValue(iProgressBarValue);
				     System.out.print(" ");
				     System.out.flush();
				 }
			     }
			     dialog.dispatchEvent(new WindowEvent(dialog, 
								  WindowEvent.WINDOW_CLOSING));
			     System.out.println("");
			     /*
			     Object[] options = { "OK", "CANCEL" };
			     boolean bRes;
			     bRes = (JOptionPane.showConfirmDialog(null, 
							       "Click OK to continue", "Warning", 
							       JOptionPane.OK_CANCEL_OPTION, 
							       JOptionPane.WARNING_MESSAGE)
				     == JOptionPane.OK_OPTION);
			     */
			     File nativeInstall = new File(tempInstall);
			     nativeInstall.delete();
			     return Boolean.TRUE;
			 } catch(Throwable e) {
			     Trace.extPrintException(e);
			     return Boolean.FALSE;
			 }
		
		     }
		});
	} catch (Throwable e) {
	    Trace.extPrintException(e);
	    return false;
	}
	return success.booleanValue(); 
    }

    /**
     * @return the name of the jar file from an Jar URL
     */
    private String extractJarFileName(String url) {
	int lastIndex;
	int fromIndex = url.indexOf('#');
	if (fromIndex==-1) {
	    lastIndex=url.lastIndexOf('/');
	    if (lastIndex<url.lastIndexOf('\\'))
		lastIndex=url.lastIndexOf('\\');
	} else {
	    lastIndex=url.lastIndexOf('/', fromIndex);
	    if (lastIndex<url.lastIndexOf('\\', fromIndex));
		lastIndex = url.lastIndexOf('\\', fromIndex);
	}
	String toDecodeURL;
	if (lastIndex==-1)
	    return null;
	else 
	    return url.substring(lastIndex+1);
    }

    /*
     * <p>
     * Blindly copy the inputstream into the outputstream until no more 
     * data is available
     * </p>
     *
     * @param is is the inputstream to read and copy from
     * @param os is the OutputStream to copy to
     */
    private void Copy(InputStream is, OutputStream os) 
	throws IOException
    {
	byte[] buffer = new byte[10240];
	int nRead;
	do {
	    nRead = is.read(buffer);
	    if (nRead != -1)
		os.write(buffer, 0, nRead);
	} while (nRead!=-1);
	is.close();
	os.close();
    }

    /*
     * <p>
     * URL poiting to an extension installation jar may be system dependent
     * The Plug-in is responsible for determining depending on the value of 
     * the "os.name" system property for making this particular url platform
     * dependent.
     * Replace the system dependent portion of an url with the value of the
     * sytem property. 
     * </p>
     *
     * @param source is the url eventually containing the key ("os.name")
     * 
     * @return the system dependent url where the key has been replaced with
     * the value of the system propery key.
     */
    private String makePlatformDependent(String source) {

	final String property = "os.name";
	String key = "$(" + property + ")$";
	int platformIndex = source.indexOf(key);
	String target = source;
	if (platformIndex != -1) {
	    String osName = System.getProperty(property);
	    String platformString = osName.replace(' ', '-');

	    // Replace "$(os.name)$" (the key) with the value of
	    // the system property (same key).
	    target = source.substring(0, platformIndex) + 
		platformString + 
		source.substring(platformIndex + key.length(), 
				 source.length());
	}
	return target;
    }

    /**
     * Method to get an internationalized string from the Activator resource.
     */
    private static String getMessage(String key)  {
	return sun.plugin.AppletViewer.getMessage(key);
    }
}
