/*
 * @(#)JavaRunTime.java	1.31 01/02/09
 *
 *  Copyright (c) 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * CopyrightVersion 1.1
 *
 */

package sun.plugin;

/*
 * A loaded and running Java Virtual machine is initialized and
 * released in the JavaRunTime.
 * This is the opportunity for startup code
 *
 * @version 	1.2.1
 * @author	Jerome Dochez
 */

import java.awt.Frame;
import java.io.IOException;
import java.io.BufferedOutputStream;
import java.io.OutputStream;
import java.io.File;
import java.io.FileOutputStream;    
import java.io.PrintStream;
import javax.swing.SwingUtilities;
import sun.plugin.usability.Trace;

public class JavaRunTime 
{
    private static boolean onWindows = false;      

    /**
     * <p>
     * Method called once when the javaplugin is loaded. This is the opportunity
     * to add some initialization function. For convenience the java home is 
     * passed to be able to know where the java environement resides
     *
     * @param javaHome		java home directory
     * @param bridgeLibPath	Activator path
     * @param userHome		user home dir
     */
    protected static void initEnvironment(String javaHome, String bridgeLibPath, String userHome) 
    {
	java.util.Properties systemProps = System.getProperties();
	systemProps.put("java.home", javaHome );  
	if (userHome==null)
	    systemProps.put("user.home", javaHome);
	else
	    systemProps.put("user.home", userHome);

	systemProps.put("bridge.home", bridgeLibPath);
	
	// Load the properties
	AppletViewer.loadPropertiesFiles();
   
	try   
	{   
	    // Determine if we are on Windows
	    //
           String osName = (String) java.security.AccessController.doPrivileged(
                                    new sun.security.action.GetPropertyAction("os.name"));

	    if (osName.indexOf("Windows") != -1)
		onWindows = true;
	}
	catch (Exception e)
	{
	    Trace.printException(e);
	}
    }     

    /**
     * <p>
     * Return the console window for this running Java Virtual Machine.
     * If the console is not created yet, create a new Java Console.
     *
     * @return console window object
     * </p>
     */
    public synchronized static ConsoleWindow getJavaConsole() 
    {
	// Return console if exists
	if (console != null) 
	{
	    return console;
	}

	console = new ConsoleWindow();

	OutputStream traceStream = getTraceOutputStream();

	// Reassign stderr and stdout.
	//
	// Notice that two separate debug streams should be used
	// instead of one because each stream will do its own
	// buffering to ensure the output to the 
	// console/traceStream will look good.
	//
	DebugOutputStream myErrOS = new DebugOutputStream(console, traceStream);
	PrintStream myErrPS = new PrintStream(myErrOS, true);
	System.setErr(myErrPS);

	DebugOutputStream myOutOS = new DebugOutputStream(console, traceStream);
	PrintStream myOutPS = new PrintStream(myOutOS, true);
	System.setOut(myOutPS);

	// Display version information
	console.displayVersion();

	return console;
    }


    /**
     * <p>
     * Checks if the Java Console Window is visible.
     *
     * @return true if visible.
     * </p>
     */
    public static boolean isJavaConsoleVisible()  
    {
	ConsoleWindow f = getJavaConsole();   

	if(f != null)
	{
	    if (onWindows)
	    {
		// On Windows, visible means the windows is actually
		// visible.
		return f.isVisible();
	    }
	    else
	    {
		// On all other platforms, visible means normal
		// windows. invisible means iconified windows.
		return (f.getState() == Frame.NORMAL);
	    }
	}
	else
	{
	    return false;
	}
    }


    /**
     * <p>
     * Show or hide Java Console Window.
     *
     * @param visible true if to show, else to hide.
     * </p>
     */
    public static void showJavaConsole(final boolean visible)  
    {
	final ConsoleWindow f = getJavaConsole();

	if(f != null)
	{
	    // Notice that setVisible MUST be called from
	    // the event dispatch thread because the Swing
	    // component may have been realized. Otherwise,
	    // it may cause deadlock.
	    //
	    try
	    {
		SwingUtilities.invokeAndWait(new Runnable()
		{
		    public void run()
		    {
			if (onWindows)
			{
			    // Caution: changing the order of the following
			    // call may cause deadlock in the browser because
			    // of the way taskbar notification is handled in
			    // the platform.
			    //
			    if (f.isVisible() != visible)
			    {
				f.setVisible(visible);
			    }

			    if (visible)
			    {
				f.toFront();
			    }
			}
			else
			{
			    // On all other platforms, visible means normal
			    // windows. invisible means iconified windows.

			    if (visible)
			    {	
				f.setState(Frame.NORMAL);
			    }
			    else
			    {
				f.setState(Frame.ICONIFIED);
			    }

			    f.setVisible(true);
			}

			ConsoleStatus(visible);
		    }
		});
	    }
	    catch (Throwable e)
	    {
		e.printStackTrace();
	    }
	}	
    }


    /**
     * <p>
     * Show or hide Java Console Window later.
     *
     * @param visible true if to show, else to hide.
     * </p>
     */
    public static void showJavaConsoleLater(final boolean visible)  
    {
	final ConsoleWindow f = getJavaConsole();

	if(f != null)
	{
	    // Notice that setVisible MUST be called from
	    // the event dispatch thread because the Swing
	    // component may have been realized. Otherwise,
	    // it may cause deadlock.
	    //
	    try
	    {
		SwingUtilities.invokeLater(new Runnable()
		{
		    public void run()
		    {
			if (onWindows)
			{
			    // Caution: changing the order of the following
			    // call may cause deadlock in the browser because
			    // of the way taskbar notification is handled in
			    // the platform.
			    //
			    if (f.isVisible() != visible)
			    {
				f.setVisible(visible);
			    }

			    if (visible)
			    {
				f.toFront();
			    }
			}
			else
			{
			    // On all other platforms, visible means normal
			    // windows. invisible means iconified windows.

			    if (visible)
			    {	
				f.setState(Frame.NORMAL);
			    }
			    else
			    {
				f.setState(Frame.ICONIFIED);
			    }

			    f.setVisible(true);
			}

			ConsoleStatus(visible);
		    }
		});
	    }
	    catch (Throwable e)
	    {
		e.printStackTrace();
	    }
	}	
    }

    /**
     * <p>
     * Show Help in Java Console
     * </p>
     */
    public static void showJavaConsoleHelp()  
    {
	final ConsoleWindow f = getJavaConsole();

	if(f != null)
	{
	    f.displayHelp();
	}
    }

    /**
     * <p>
     * Show System Properties in Java Console
     * </p>
     */
    public static void showSystemProperties()
    {
        final ConsoleWindow f = getJavaConsole();

        if(f != null)
        {
	    f.displaySystemProperties();
	}
    }

    /**
     * <p>
     * Print a message to Java Console Window.
     *
     * @param msg Message to be printed.
     * </p>
     */
    public static void printToJavaConsole(String msg)   
    {
	ConsoleWindow f = getJavaConsole();

	if (f != null)
	{
	    f.append(msg);
	}
    }

    /**
     * <p>
     * Set the plugin taskbar icon window handle
     *
     * @param hWnd Plugin taskbar icon window handle
     * </p>
     */
    public static void setStatusIconWindowHandle(int hWnd)
    {
	hStatusIconWnd = hWnd;
    }

    /**  
     * <p>
     * Get the plugin taskbar icon window handle
     *
     * @return Plugin taskbar icon window handle
     * </p>
     */
    public static int getStatusIconWindowHandle()
    {
	return hStatusIconWnd;
    }

    /**
     * <p>
     * Return the trace output stream.
     *
     * @return trace output stream.
     * </p>
     */
    public static OutputStream getTraceOutputStream() 
    {  
	// Return stream if exists
	if (traceStream != null) 
	{
	    return traceStream;
	}

	try 
	{
	    // Use another trace file name if specified
	    String filename = System.getProperty("javaplugin.tracefile.path");
	    File file = null;

	    if (filename == null || filename.equalsIgnoreCase(""))
	    {
		// Trace file not specified, use default.
		//
    		String userHome = System.getProperty("user.home");
		file = new File(userHome, "plugin" + System.getProperty("javaplugin.nodotversion") + ".trace");
	    }
	    else
	    {
		file = new File(filename);
	    }

	    if (file.exists()) 
		file.delete();

	    traceStream = new BufferedOutputStream(new FileOutputStream(file));
	} catch (IOException e) {
	    e.printStackTrace();
	}

	return traceStream;
    }

    /*
     * <p>
     * the Java Console to display useful error and trace messages
     * </p>
     *
     */
    private static Object syncObject = new Object();
    private static OutputStream traceStream = null;    
    private static ConsoleWindow console = null;
    private static int hStatusIconWnd = 0;
    public static native void ConsoleStatus(boolean fStart);
}



