/*
 * @(#)SystemEventQueueUtilities.java	1.1 99/11/22
 *
 * Copyright 1995-2000 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.plugin;

import java.awt.Component;
import java.awt.EventQueue;
import java.awt.Toolkit;
import java.awt.AWTEvent;
import java.lang.reflect.InvocationTargetException;

/**
 * This class is a clone of javax.swing.SystemEventQueueUtilities but without the
 * baggage of swing.
 */

public class SystemEventQueueUtilities
{
    /**
     * Basically do the same thing as SwingUtilities.invokeLater() without Swing. 
     */

    public static void invokeLater(Runnable doRun) {
        postRunnable( doRun, null );
    }

    /**
     * Basically do the same thing as SwingUtilities.invokeAndWait() without Swing.
     */

    public static void invokeAndWait(final Runnable doRun)
        throws InterruptedException, InvocationTargetException
    {

        if ( EventQueue.isDispatchThread() ) {
            throw new Error("Cannot call invokeAndWait from the event dispatcher thread");
        }
        Object lock = new Object() {
            public String toString() {
                return "SwingUtilities.invokeAndWait() lock for " + doRun;
            }
        };
        Exception exc = null;
        synchronized(lock) {
            exc = postRunnable(doRun, lock);
            lock.wait();
        }

        if (exc != null) {
            throw new InvocationTargetException(exc);
        }
    }


    /**
     * Stolen from Swing ...
     */

    private static Exception postRunnable(Runnable doRun, Object lock) {
        EventQueue    queue = Toolkit.getDefaultToolkit().getSystemEventQueue();
        RunnableEvent event = new RunnableEvent(doRun, lock);

        queue.postEvent( event );
        return event.exception;
    }


    /**
     * Event type used for dispatching runnable objects for
     * SwingUtilities.invokeLater() and SwingUtilities.invokeAndWait().
     *
     * @see #postRunnable
     */
    private static class RunnableEvent extends AWTEvent {
        static final int EVENT_ID = AWTEvent.RESERVED_ID_MAX + 1000;
	static final Component target = new RunnableTarget();
	final Runnable doRun;
	final Object lock;
	Exception exception;

        RunnableEvent(Runnable doRun, Object lock) {
            super(target, EVENT_ID);
	    this.doRun = doRun;
            this.lock = lock;
        }
    }

    /**
     * A dummy Component subclass that (only) handles RunnableEvents.  If the
     * AWT System event queue is accessible (i.e. we're running as
     * an application or as trusted code), RunnableEvents are dispatched
     * to this component.
     *
     * @see #processRunnableEvent
     */
    private static class RunnableTarget extends Component
    {
        RunnableTarget() {
            super();
            enableEvents(RunnableEvent.EVENT_ID);
        }

        protected void processEvent(AWTEvent event) {
	    if (event instanceof RunnableEvent) {
		processRunnableEvent((RunnableEvent)event);
	    }
        }
    }


    /**
     * Calls RunnableEvent.doRun.run().  If RunnableEvent.lock is non
     * null then we synchronize the run() call and save the exception
     * (if any) in the RunnableEvent.exception field.
     */
    private static void processRunnableEvent(RunnableEvent runnableEvent)
    {
	Object lock = runnableEvent.lock;
	if (lock == null) {
	    runnableEvent.doRun.run();
	}
	else {
	    synchronized(lock) {
		try {
		    runnableEvent.doRun.run();
		}
		catch (Exception e) {
		    runnableEvent.exception = e;
		}
		finally {
		    if (runnableEvent.lock != null) {
			runnableEvent.lock.notify();
		    }
		}
	    }
	}
    }
}

