/*
 * @(#)HTMLVersionTable.java	1.8 01/02/09
 *
 * Copyright (c) 1995-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * CopyrightVersion 1.3
 *
 * Java Plug-in 
 *
 * @author Stanley Man-Kit Ho
 */

package sun.plugin.cachescheme;

import java.util.*;
import java.io.*;
import java.net.*;

/* 
 * Version table for storing installed HTML page information.
 */
public class HTMLVersionTable {

    // Separator character for version entry
    private static char separator = '|';

    // Version table filename
    private static String _filename = null;

    static
    {
        String userHome = (String) java.security.AccessController.doPrivileged(
               new sun.security.action.GetPropertyAction("user.home"));

	// Determine filename of the version table
	_filename = userHome 
		    + File.separator + PluginCacheTable.cacheHome
		    + File.separator + System.getProperty("javaplugin.version")
		    + File.separator + "html"
		    + File.separator + "version.table";
    }


    // Properties to store version information
    Properties _versionTable = new Properties();

    
    /**
     * Load version information from version table file.
     */    
    public void load()
		throws IOException
    {	
	// Open version table file
	File file = new File(_filename);

	// Load version info only if file exists	
	if (file.exists())
	{	   
	    FileInputStream fis = new FileInputStream(_filename);
	    BufferedInputStream bis = new BufferedInputStream(fis);

	    _versionTable.load(bis);

	    bis.close();
	    fis.close();
	}
    }


    /**
     * Save version information into version table file.
     */    
    public void save()
		throws IOException
    {
	// Open version table file
        FileOutputStream fos = new FileOutputStream(_filename);
	BufferedOutputStream bos = new BufferedOutputStream(fos);
    
	// Store version table information
        _versionTable.store(bos, ("This File is Machine Generated." +
		 		  "Please DO NOT CHANGE IT !! ." +
				  "Changing it will BREAK the Applet Installation Cache." +
				  " This will inturn cause the PLUGIN to break"));

	bos.close();
	fos.close();
    }

    
    /**
     * Add version information into version table file.
     *
     * @param htmlTitle HTML page title
     * @param version HTML page version
     * @param path Path to HTML page
     * @param filename Filename of HTML page
     */    
    public void add(String htmlTitle, String version, String path, String filename)
    {
	_versionTable.put(htmlTitle, (version + separator + path + separator + filename));
    }


    /**
     * Add version information into version table file.
     *
     * @param htmlTitle HTML page title
     * @param version HTML page version
     * @param path Path to HTML page
     * @param filename Filename of HTML page
     */    
    public void remove(String htmlTitle)
    {
	_versionTable.remove(htmlTitle);
    }


    /**
     * Return version of a given HTML title in the version table.
     *
     * @param htmlTitle HTML page title
     * @return version HTML page version
     */    
    public String getVersion(String htmlTitle) 
		  throws EntryNotFoundException
    {
	String entry = getVersionEntry(htmlTitle);

	StringTokenizer st = new StringTokenizer(entry, "|", false);

	if (st.countTokens() < 1)
	    throw new EntryNotFoundException("Version information is not found in the version table for " + htmlTitle);

	String version = st.nextToken().trim();

	return version;
    }
  

    /**
     * Return path of a given HTML title in the version table.
     *
     * @param htmlTitle HTML page title
     * @return path Path to HTML page
     */    
    public String getPath(String htmlTitle)
		  throws EntryNotFoundException
    {
	String entry = getVersionEntry(htmlTitle);
	
	StringTokenizer st = new StringTokenizer(entry, "|", false);

	if (st.countTokens() < 2)
	    throw new EntryNotFoundException("Path information is not found in the version table for " + htmlTitle);

	// Skip first token
	st.nextToken();

	String path = st.nextToken().trim();

	return path;
    }


    /**
     * Return filename of a given HTML title in version table.
     *
     * @param htmlTitle HTML page title
     * @return filename Filename of HTML page
     */    
    public String getFilename(String htmlTitle)
		  throws EntryNotFoundException
    {
	String entry = getVersionEntry(htmlTitle);

	StringTokenizer st = new StringTokenizer(entry, "|", false);

	if (st.countTokens() < 1)
	    throw new EntryNotFoundException("Filename information is not found in the version table for " + htmlTitle);

	// Skip first two token
	st.nextToken();
	st.nextToken();

	String filename = st.nextToken().trim();

	return filename;
    }


    /**
     * Get version entry from version table.
     *
     * @param htmlTitle HTML page title
     * @return version entry
     */    
    private String getVersionEntry(String htmlTitle)
 		   throws EntryNotFoundException
    {
	String entry = (String) _versionTable.get(htmlTitle);
	
	if (entry == null)
	    throw new EntryNotFoundException(htmlTitle + " is not found in version table");		   

	return entry;
    }
}


