/*
 * @(#)PluginJarCacheHandler.java	1.5 00/09/16
 *
 * Copyright (c) 1995-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * CopyrightVersion 1.3
 *
 * Java Plug-in 
 *
 * @author Devananda J. 
 */

package sun.plugin.cachescheme;

import java.util.*;
import java.io.*;
import java.lang.String;
import java.net.*;
import sun.plugin.usability.Trace;
import sun.plugin.usability.TraceFilter;
import sun.plugin.usability.DialogFactory;
import javax.swing.JOptionPane;

public class PluginJarCacheHandler implements JarCacheInterface{

    //For each codebase
    URL m_relativeCodeBase = null;
    String m_protocol = null;
    String m_urlString;
    PluginJarCacheTable m_cacheTable = null;

    /**
     * Check if the cache handler can handle the cache option.
     */
    public boolean isValidHandler(String jarOption)
    {
	if (jarOption != null && jarOption.equalsIgnoreCase("plugin"))
	    return true;
	else
	    return false;
    }

    public Vector getJarFilesPath(String jarArchive, String jarVersion, URL codeBase)
	throws PluginMutexException, InvalidCacheParameterException, IOException, JarCacheException
    {
	if (jarArchive == null)
	    throw new InvalidCacheParameterException("Requires attribute \'cache_archive\'");

	//return vector
	Vector jarFilesPath = new Vector();

	m_relativeCodeBase = codeBase;

	// Load cache table
	m_cacheTable = new PluginJarCacheTable();
	m_cacheTable.load();

	// Extract individual jar name from "cache_archive"
	ArrayList jarFiles = parseParameters(jarArchive);
	
	
	// Extract individual jar version from "cache_version"
	ArrayList jarVersions = null;

	if(jarVersion != null)
	{
	    jarVersions = parseParameters(jarVersion);

	    if(jarFiles.size() != jarVersions.size())
	    {
		throw new InvalidCacheParameterException("Requires number of attributes specified in \'cache_archive\' matches those in \'cache_version\'");
	    }

	    // Iterate through all the versions to ensure they are in valid form
	    for (int i=0; i < jarVersions.size(); i++)
	    {
		String ver = (String) jarVersions.get(i);

		if (ver != null)
    		    PluginJarCacheTableEntry.extractVersion(ver);
	    }
	}

	// Loop through each entry in "cache_archive" and download them
	//	
	for(int i = 0; i < jarFiles.size(); i++)
	{
	    File jarFile = null;

	    String curJarFileVer = null;
	    String curJarFileName = jarFiles.get(i).toString();

	    if(true == CheckForFullURL(curJarFileName))
		throw new InvalidCacheParameterException("Only relative URL should be used in the attribute \'cache_archive\'");

	    String urlString = null;
    	    try
	    {
		URL url = new URL(m_relativeCodeBase, curJarFileName);
		urlString = url.toString();
	    }
	    catch(MalformedURLException exc)
	    {
		throw new InvalidCacheParameterException("Invalid relative URL is specified in the attribute \'cache_archive\'");
	    }

	    if(jarVersions != null)
		curJarFileVer  = jarVersions.get(i).toString();

	    jarFile = cacheHandler(curJarFileName, curJarFileVer, urlString);
	    			
	    // If successful, add to array
	    if(jarFile != null)
		jarFilesPath.addElement(jarFile);		
	}

	return jarFilesPath;
    }


    private File cacheHandler(String curJarFileName, String curJarFileVer, 
		String urlString) throws PluginMutexException, InvalidCacheParameterException, IOException, JarCacheException
    {
	//acquire the cache inter-process lock
	PluginMutex urlMutex = null;
	File jarFile = null;

	try
	{
	    urlMutex = new PluginMutex(urlString);
	    urlMutex.waitForLock();

	    PluginURLConnection urlConn = new PluginURLConnection(m_relativeCodeBase, curJarFileName);

	    // Obtain the proper version of the jar file from cache
	    //
	    jarFile = getJarFileFromCache(curJarFileName, curJarFileVer, urlString, urlConn);

	    // If the proper version of jar doesn't exists in cache, download it
	    //
	    if(jarFile == null)
	    {
		jarFile = downloadJarFileToCache(curJarFileName, curJarFileVer, urlString, urlConn);
	    }
	}
	finally
	{
	    //remove the cache inter-process lock
	    if(urlMutex != null)
		urlMutex.release();
	}

	return jarFile;
    }


    /**
     * Parse parameters and store each elements into ArrayList.
     *
     * @param param Parameter string
     * @return Array of parsed elements.
     */
    private ArrayList parseParameters(String param)
    {
	ArrayList elems = new ArrayList();

	// Parse the string and fill it in a arraylist
	StringTokenizer st = new StringTokenizer(param, ",", false);
	while(st.hasMoreTokens()) 
        {
	    String elem = st.nextToken().trim();
	    elems.add(elem);
	}
	
	return elems;
    }


    /**
     * Check whether the current jar file name is not a relative URL,
     * in order to avoid security hole
     *
     * @param fileName jar file name
     * @return boolean whether the passed name is a URL or not
     */
    private boolean CheckForFullURL(String fileName)
    {
	boolean fURL = false;
	URL relativeURL = null;
	try
	{
	    relativeURL = new URL(fileName);
	    fURL = true;
	}
	catch(MalformedURLException exc)
	{
	    fURL = false;
	}

	return fURL;
    }

    /**
     * Download Jar file into cache and update cache table.
     *
     * @param curJarFileName Jar file name
     * @param curJarFileVer Jar file version
     * @param urlString Jar file url
     * @param urlConn URL connection.
     * @return Downloaded jar file
     */
    private File downloadJarFileToCache(String curJarFileName, String curJarFileVer, 
		String urlString, PluginURLConnection urlConn)
		throws PluginMutexException, InvalidCacheParameterException, IOException, JarCacheException
    {
	//get file with unique file name	
	File jarFile = generateCacheFile(curJarFileName, curJarFileVer);

	boolean success = false;
		
	try
	{
	    Trace.println("Downloading " + urlString + " to " + jarFile, TraceFilter.JAVA_CONSOLE_ONLY);
	    
	    // Download JAR file into cache
	    urlConn.downLoadFile(jarFile);

	
 	    Trace.println("Updating applet cache table", TraceFilter.JAVA_CONSOLE_ONLY);

	    // Create cache table entry
	    PluginJarCacheTableEntry newEntry = PluginJarCacheTableEntry.createEntry(curJarFileVer, jarFile.getName(), 
			       urlConn.getLastModified(), urlConn.getContentLength());

	    //add the entry into cachetable
	    m_cacheTable.add(urlString, newEntry);
	    m_cacheTable.save();

	    success = true;
	}
	finally
	{
	    if (success == false)
	    {
		// If there is any exception, delete cache file and remove
		// cache entry
		//
		if(jarFile != null && jarFile.exists())
		    jarFile.delete();

		m_cacheTable.remove(urlString);
	    }
	}

	return jarFile;
    }


    /**
     * Get the cache file from cache if it is still valid.
     */
    private File getJarFileFromCache(String curJarFileName, String curJarFileVer, 
		String urlString, PluginURLConnection urlConn)
		throws PluginMutexException, IOException, InvalidCacheParameterException
    {
	File jarFile = null;

	try
	{
	    Trace.println("Check if " + curJarFileName + " exists as entry in cache table", TraceFilter.JAVA_CONSOLE_ONLY);
	    
	    // Obtain cache entry from cache table
	    PluginJarCacheTableEntry entry = m_cacheTable.getEntry(urlString);

	    // Return if cache entry not found
	    if (entry == null)
	    {
		Trace.println(curJarFileName + " not found in cache table - force download", TraceFilter.JAVA_CONSOLE_ONLY);
		return null;
	    }


	    //----------------------------------------------------------------

	    Trace.println("Check if " + curJarFileName + " actually exists in cache", TraceFilter.JAVA_CONSOLE_ONLY);

	    // Obtain actual cache file accordingly to the cache entry
	    jarFile = entry.getFileInCache();

	    if (jarFile == null || jarFile.exists() == false)
	    {
		Trace.println(curJarFileName + " not found in applet cache - force download", TraceFilter.JAVA_CONSOLE_ONLY);
		return null;
	    }


	    //----------------------------------------------------------------

	    Trace.println("Check if " + curJarFileName + " is upto-date", TraceFilter.JAVA_CONSOLE_ONLY);

	    //if version is mentioned, check only the version
	    if(curJarFileVer != null)
	    {
		entry.verifyCacheVersion(curJarFileVer);
	    }
	    else
	    {
		//if version is not mentioned, check time stamp and file size

		Date cacheTimeStamp = entry.getCacheTimeStamp();
		Date remoteTimeStamp = new Date(urlConn.getLastModified());

		Trace.println("Last modified: cache=" + cacheTimeStamp + ", remote=" + remoteTimeStamp,
			      TraceFilter.JAVA_CONSOLE_ONLY);

		entry.verifyCacheTimeStamp(remoteTimeStamp);

		///

		int cacheFileSize = entry.getCacheFileSize();
		int remoteFileSize = urlConn.getContentLength();

		Trace.println("File size: cache=" + cacheFileSize + ", remote=" + remoteFileSize,
			      TraceFilter.JAVA_CONSOLE_ONLY);

		entry.verifyCacheFileSize(remoteFileSize);
	    }

	    Trace.println(curJarFileName + " is upto-date in cache", TraceFilter.JAVA_CONSOLE_ONLY);

	    return jarFile;
    	}
	catch (CacheTableEntryException exc)
	{
	    Trace.println("Unable to parse cache entry from cache table - force download", TraceFilter.JAVA_CONSOLE_ONLY);

	    // This is thrown if a cache entry is not valid. It 
	    // automatically invalidate the cache.
	    
	    // Catch the exception
	    return null;
	}
	catch (JarFileInvalidException exc)
	{
	    // This is thrown if the file version / time-stamp check fail
	    Trace.println("File is not upto-date in cache - force download", TraceFilter.JAVA_CONSOLE_ONLY);

	    // If the jar file is useless, don't keep it
	    if (jarFile != null)
		jarFile.delete();

	    return null;
	}
    }

    /**
     * Generate an unique cache filename.
     *
     * @param curJarFileName File name
     * @param curJarFileVer File version
     * @return Local file object.
     */
    private File generateCacheFile(String curJarFileName, String curJarFileVer)
		 throws JarCacheException
    {
	String jarFileName;
	String fullCodeBase;

    	if (curJarFileName.indexOf('/') != -1) 
	{
	    jarFileName = curJarFileName.substring(curJarFileName.lastIndexOf('/')+1);
	    fullCodeBase = m_relativeCodeBase + 
			   curJarFileName.substring(0, curJarFileName.lastIndexOf('/'));
	}
	else
	{
	    jarFileName = curJarFileName;
	    fullCodeBase = m_relativeCodeBase.toString();
	}      

	boolean isUnique = true;

	File file = null;
	
	// Random number generator	   
	Random randomObj = new Random();
	int i = 0;

	while (isUnique == true)
	{
	    // New filename
	    jarFileName = jarFileName +
			  fullCodeBase.hashCode() +
			  curJarFileVer +
			  String.valueOf(randomObj.nextLong()) + ".jar";

	    // Check whether such file name exists in cache
	    file = new File(PluginJarCacheTable.getCacheHomeDir() + File.separator + jarFileName);

	    isUnique = file.exists();

	    if(i++ > 255)
		Trace.println("Warning: Could not find a suitable name for the file in the cache after iteration " + i + ".", TraceFilter.JAVA_CONSOLE_ONLY);

	    // Check if we have gone through 512 iterations, or we may be in infinite loop - throw exception
	    if (i > 512)
		throw new JarCacheException("Unable to find a suitable name for the jar file in the cache.");
	}

	return file;
    }
}
