/*
 * @(#)ExtensionInstallationImpl.java	1.16 01/04/21
 *
 * Copyright 1997-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.plugin.extension;

import sun.misc.ExtensionInstallationException;
import sun.misc.ExtensionInfo;
import sun.misc.ExtensionInstallationProvider;
import javax.swing.JOptionPane;
import javax.swing.JDialog;
import javax.swing.JProgressBar;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.Box;
import javax.swing.JOptionPane;
import java.awt.event.WindowEvent;
import java.util.jar.JarFile;
import java.util.jar.Attributes;
import java.util.jar.Attributes.Name;
import java.util.jar.Manifest;
import java.util.jar.JarEntry;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLClassLoader;
import java.net.JarURLConnection;
import java.net.MalformedURLException;
import java.net.SocketException;
import java.io.IOException;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.security.PrivilegedActionException;
import java.security.CodeSource;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.CertificateParsingException;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Vector;
import sun.security.action.GetPropertyAction;
import sun.plugin.security.PluginClassLoader;
import sun.plugin.security.TrustDecider;
import sun.plugin.resources.ResourceHandler;
import sun.plugin.usability.Trace;
import sun.plugin.usability.TraceFilter;
import sun.plugin.usability.DialogFactory;

/**
 * Provide facilities to install extensions in the standard extension
 * directory of the running JDK
 *
 * @author Stanley Man-Kit Ho
 */
public class ExtensionInstallationImpl implements ExtensionInstallationProvider {

    public boolean installExtension(final ExtensionInfo reqExtInfo, 
				    final ExtensionInfo instExtInfo) 
				    throws ExtensionInstallationException 
    {
	Trace.extPrintln("Optional Package to be installed " + reqExtInfo);

	try
	{
	    // Copy the jar into temp directory
	    AccessController.doPrivileged(
		new PrivilegedExceptionAction() 
		{
		    public Object run() throws IOException, ExtensionInstallationException,
		    			       CertificateException, CertificateEncodingException,
					       CertificateExpiredException, CertificateNotYetValidException,
					       CertificateParsingException, KeyStoreException, 
					       NoSuchAlgorithmException, IOException,
					       InterruptedException
		    {	
			String jarFileName = ExtensionUtils.makePlatformDependent(
						 ExtensionUtils.extractJarFileName(reqExtInfo.url));
			String platformURL = ExtensionUtils.makePlatformDependent(reqExtInfo.url);
    
			// Ask if the user would like to install the extension
			//
			if (askUserForAcknowledgment(reqExtInfo, instExtInfo) == 1) {
			    throw new ExtensionInstallationException(
					       "User denied installation of " +
					       platformURL);
			}

			// Copy extension installer jar into temp directory
			URL url = new URL(platformURL);
			URLConnection conn = url.openConnection();
			InputStream is = conn.getInputStream();
			BufferedInputStream bis = new BufferedInputStream(is);

			String localJarName = ExtensionUtils.getTempDir() + File.separator + jarFileName;

			File f = new File(localJarName);
			FileOutputStream fos = new FileOutputStream(f);
			BufferedOutputStream bos = new BufferedOutputStream(fos);

			ExtensionUtils.copy(bis, bos);

			bis.close();
			is.close();
			bos.close();
			fos.close();

			// Verify the jar file
			verifyJar(platformURL, localJarName);

			// Install optional packages through installer
			installJarFile(platformURL, localJarName);

			// Clean up temp file
			if (f.exists())
			    f.delete();

			return null;
		    }
		});
	}
	catch (Throwable e)
	{
	    Throwable ex = e;
	    String desc = null;

	    if (e instanceof PrivilegedActionException)
	    {
		ex = ((PrivilegedActionException) e).getException();

		if (ex instanceof CertificateExpiredException)
		    desc = ResourceHandler.getMessage("optpkg.cert_expired");
		else if (ex instanceof CertificateNotYetValidException)
		    desc = ResourceHandler.getMessage("optpkg.cert_notyieldvalid");
		else if (ex instanceof CertificateException)
		    desc = ResourceHandler.getMessage("optpkg.cert_notverify");
		else
    		    desc = ResourceHandler.getMessage("optpkg.general_error");
	    }
	    else 
	    {
		desc = ResourceHandler.getMessage("optpkg.general_error");
	    }

	    Trace.extPrintException(ex, desc, ResourceHandler.getMessage("optpkg.caption"));

	    Trace.extPrintln("Optional Package Installation failed.");
	    return false;
    	}

	Trace.extPrintln("Optional Package Installation succeeded.");
	return true;
    }


    /** 
     * <P> Verify a signed jar file.
     * </P>
     *
     * @param url URL of the signed jar file.
     * @param jarName Name of the jar file.
     */
    private void verifyJar(String url, String jarName)
		 throws ExtensionInstallationException, 
			CertificateException, CertificateEncodingException,
			CertificateExpiredException, CertificateNotYetValidException,
			CertificateParsingException, KeyStoreException, 
			NoSuchAlgorithmException, IOException
    {
	try
	{
	    boolean anySigned = false;

	    JarFile jf = new JarFile(jarName, true);

	    Vector entriesVec = new Vector();
	    byte[] buffer = new byte[8192];


	    // Open the jar file, enumerate all its entries,
	    // and check the signature/digest of each entries
	    //
	    Enumeration entries = jf.entries();
	    while (entries.hasMoreElements()) 
	    {
		JarEntry je = (JarEntry)entries.nextElement();
		entriesVec.addElement(je);
		InputStream is = jf.getInputStream(je);
		int n;
		while ((n = is.read(buffer, 0, buffer.length)) != -1) 
		{
		    // we just read. this will throw a SecurityException
		    // if  a signature/digest check fails.
		}
		is.close();
	    }
	    jf.close();


	    // If everything goes right here, it only means that
	    // the signature/digest check passes. However, it
	    // still doesn't mean it can be trusted.
	    //
	    // In other to determine if the code can be trusted,
	    // check the certificate.
	    //
	    Manifest man = jf.getManifest();
	    if (man != null) 
	    {
		HashMap certMap = new HashMap();
		Enumeration e = entriesVec.elements();

		while (e.hasMoreElements()) 
		{
		    JarEntry je = (JarEntry) e.nextElement();
		    String name = je.getName();
		    Certificate[] certs = je.getCertificates();

//		    Trace.extPrintln("Certs: " + certs, TraceFilter.JAVA_CONSOLE_ONLY);

		    boolean isSigned = ((certs != null) && (certs.length > 0));
		    anySigned |= isSigned;

		    if (isSigned)
		    {
			// Determine if the certificate has been encountered
			// before
			CodeSource cs = (CodeSource) certMap.get(certs);

			if (cs == null)
			{
			    // No, we haven't seen this cert before, so
			    // popup security dialog
			    //
			    cs = new CodeSource(new URL(url), certs);
			    certMap.put(certs, cs);

			    if (TrustDecider.isAllPermissionGranted(cs) == false)
			    {
				throw new ExtensionInstallationException("User deny optional package installer"
									+ " to be launched.");
			    }
			}
		    }
		}
	    }

	    if (man == null)
		throw new ExtensionInstallationException("No manifest in the optional package installer.");

	    if (!anySigned) 
		throw new ExtensionInstallationException("Optional package installer is unsigned."
					    + " (signatures missing or not parsable)");
	}
	catch (IOException e)
	{
	    throw new ExtensionInstallationException("IO Error. Unable to verify optional package installer.");
	}
    }



    private int askUserForAcknowledgment(ExtensionInfo reqExtInfo, 
					 ExtensionInfo instExtInfo) 
    {
	StringBuffer dialogText = new StringBuffer();

	if (instExtInfo != null) {
	    // this extension is already installed, let's check why we need to
	    // reinstall it.
	    int reason = instExtInfo.isCompatibleWith(reqExtInfo);
	    switch(reason) {
	    case ExtensionInfo.REQUIRE_SPECIFICATION_UPGRADE:
		dialogText.append(ResourceHandler.getMessage("optpkg.prompt_user.new_spec.text0"));
		dialogText.append(reqExtInfo.specVersion);
		dialogText.append(ResourceHandler.getMessage("optpkg.prompt_user.new_spec.text1"));
		dialogText.append(instExtInfo.title);
		break;
	    case ExtensionInfo.REQUIRE_IMPLEMENTATION_UPGRADE:
		dialogText.append(ResourceHandler.getMessage("optpkg.prompt_user.new_impl.text0"));
		dialogText.append(reqExtInfo.implementationVersion);
		dialogText.append(ResourceHandler.getMessage("optpkg.prompt_user.new_impl.text1"));
		dialogText.append(instExtInfo.title);
		break;
	    case ExtensionInfo.REQUIRE_VENDOR_SWITCH:
		dialogText.append(ResourceHandler.getMessage("optpkg.prompt_user.new_vendor.text0"));
		dialogText.append(reqExtInfo.vendor);
		dialogText.append(ResourceHandler.getMessage("optpkg.prompt_user.new_vendor.text1"));
		dialogText.append(reqExtInfo.vendorId);
		dialogText.append(ResourceHandler.getMessage("optpkg.prompt_user.new_vendor.text2"));
		dialogText.append(instExtInfo.title);
		break;
	    default:
		dialogText.append(ResourceHandler.getMessage("optpkg.prompt_user.default.text"));
		dialogText.append(reqExtInfo.name);
		break;	
	    }
	} 
	else
	{
	    dialogText.append(ResourceHandler.getMessage("optpkg.prompt_user.default.text"));
	    dialogText.append(reqExtInfo.name);
	}

	dialogText.append(ResourceHandler.getMessage("optpkg.prompt_user.end"));
	dialogText.append(ExtensionUtils.makePlatformDependent(reqExtInfo.url));

	Object[] options = { getMessage("security_dialog.buttonYes"),
			     getMessage("security_dialog.buttonNo"), 
			     getMessage("security_dialog.buttonAlways")};

	int ret = 0;
	
	// Show dialog
	if (Trace.isAutomationEnabled() == false)
	{
	    ret = DialogFactory.showOptionDialog(dialogText,
						 getMessage("security_dialog.caption"), 
						 options, 
						 options[0]);
	}
	else
	{
	    // If automation is enabled
	    Trace.extPrintln("Automation: Accept optional package installation", TraceFilter.JAVA_CONSOLE_ONLY);
	    ret = 0;
	}


	if (ret==0) {
	    Trace.extPrintln("Optional package download granted by user, download from " + reqExtInfo.url);
	} else {
	    Trace.extPrintln("Optional package download not granted by user");
	}						 		
	return ret;
    }

    
    /** 
     * <P> Installs optional packages.
     * </P>
     *
     * @param url URL of the downloaded optional packages.
     * @param jarFileName Jar file that represents the downloaded optional packages.
     */
    private void installJarFile(String url, String jarFileName) 
	throws ExtensionInstallationException, IOException, InterruptedException 
    {
	Trace.extPrintln("Starting the installation of " + jarFileName);

	JarFile jarFile = new JarFile(jarFileName);
	Manifest man = jarFile.getManifest();

	if (man != null) {

	    // Get the extension directory, we'll need it anyway
	    String extDirs = (String) AccessController.doPrivileged(new GetPropertyAction("java.ext.dirs"));

	    StringTokenizer st = new StringTokenizer(extDirs, File.pathSeparator);
	    String destDir = st.nextToken();
	    
	    Attributes attr = man.getMainAttributes();
	    ExtensionInstaller installer = new RawExtensionInstaller();

	    if (attr != null) {
		if (attr.getValue(Name.MAIN_CLASS) != null)
		{
		    // Main attribute exists - Java installer
		    installer = new JavaExtensionInstaller();
		}
		else if (attr.getValue(Name.EXTENSION_INSTALLATION) != null)
		{
		    // Extension-installation attribute exists, native installer
		    installer = new NativeExtensionInstaller();
		}
	    }
	
	    // Make sure jarFile is set to null for GC
	    jarFile = null;
	    
	    // Launch extension installer
	    installer.install(url, jarFileName, destDir);
	}
    }


    /**
     * Method to get an internationalized string from the Activator resource.
     */
    private static String getMessage(String key)  {
	return sun.plugin.AppletViewer.getMessage(key);
    }
}
