/*
 * @(#)SecureInvocation.java	1.11 01/02/11
 *
 * Copyright (c) 1995-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * CopyrightVersion 1.1 
 *
 */

package sun.plugin.liveconnect;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.io.FilePermission;
import java.security.AccessControlContext;
import java.security.AccessController;
import java.security.CodeSource;
import java.security.Policy;
import java.security.ProtectionDomain;
import java.security.AllPermission;
import java.security.Permissions;
import java.security.PermissionCollection;
import java.security.PrivilegedAction;
import java.security.PrivilegedExceptionAction;
import java.security.PrivilegedActionException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.SocketPermission;
import java.net.URL;
import java.net.MalformedURLException;
import sun.plugin.usability.Trace;

/**
 * <P> SecureInvocation is for implementating nsISecureJNI which allows 
 * JNI to be called with a sandbox. The idea here is to perform the action
 * through reflection with a AccessControlContext. In this way, JNI is not 
 * called directly, but called through this wrapper class with a sandbox
 * instead.
 * </P>
 *
 * @version 	1.1 
 * @author	Stanley Man-Kit Ho
 */


public class SecureInvocation {

    /**
     * <P> Construct a new Java object with security context.
     * </P>
     *
     * @param constructor Constructor object 
     * @param jargs jvalueArray in native code
     * @param securityContext nsISecurityContext interface in native code
     * @param result jvalue to be returned.
     * @return true if the operation is successful
     */
    public static Object ConstructObject(Constructor constructor, Object[] args, 
		  			 int securityContext) throws Exception
    {
	// Set up virtual JavaScript Protection Domain
	ProtectionDomain[] domains = new ProtectionDomain[1];

	// Obtain proper protection domain for the call
	if (securityContext ==  0)
	    domains[0] = getDefaultProtectionDomain(null);
	else 
	    domains[0] = getTrustedProtectionDomain();

	AccessControlContext context = new AccessControlContext(domains);

	try {
	    // Perform the object constructor.
	    return AccessController.doPrivileged(new PrivilegedConstructObjectAction(constructor, args),
						 context);
	} catch (PrivilegedActionException e) {
	    throw e;
	}
    }


    /**
     * <P> Call a method on Java object with security context.
     * </P>
     *
     * @param obj Java object which method is called
     * @param method Method object 
     * @param jargs jvalueArray in native code
     * @param securityContext nsISecurityContext interface in native code
     * @param result jvalue to be returned.
     * @return true if the operation is successful
     */
    public static Object CallMethod(Object obj, Method method, Object[] args, 
		  		    int securityContext) throws Exception
    {
	// Set up virtual JavaScript Protection Domain
	ProtectionDomain[] domains = new ProtectionDomain[1];

	// Obtain proper protection domain for the call
	if (securityContext ==  0)
	    domains[0] = getDefaultProtectionDomain(null);
	else 
	    domains[0] = getTrustedProtectionDomain();

	AccessControlContext context = new AccessControlContext(domains);

	try {
	    // Perform the method invocation.
	    return AccessController.doPrivileged(new PrivilegedCallMethodAction(method, obj, args),
						 context);
	} catch (PrivilegedActionException e) {
	    throw e;
	}
    }


    /**
     * <P> Get a field on Java object with security context.
     * </P>
     *
     * @param obj Java object which field is accessed
     * @param field Field object 
     * @param securityContext nsISecurityContext interface in native code
     * @param result jvalue to be returned.
     * @return true if the operation is successful
     */
    public static Object GetField(Object obj, Field field, int securityContext) 
			 throws Exception
    {
	// Set up virtual JavaScript Protection Domain
	ProtectionDomain[] domains = new ProtectionDomain[1];

	// Obtain proper protection domain for the call
	if (securityContext ==  0)
	    domains[0] = getDefaultProtectionDomain(null);
	else 
	    domains[0] = getTrustedProtectionDomain();

	AccessControlContext context = new AccessControlContext(domains);

	try {
	    // Perform the getField
	    return AccessController.doPrivileged(new PrivilegedGetFieldAction(field, obj),
		    		    		 context);	
	} catch (PrivilegedActionException e) {
	    throw e;
	}
    }



    /**
     * <P> Set a field on Java object with security context.
     * </P>
     *
     * @param obj Java object which field is accessed
     * @param field Field object 
     * @param jval Value to set in the field
     * @param securityContext nsISecurityContext interface in native code
     * @return true if the operation is successful
     */
    public static void SetField(Object obj, Field field, Object val, int securityContext)
			        throws Exception
    {
	// Set up virtual JavaScript Protection Domain
	ProtectionDomain[] domains = new ProtectionDomain[1];

	// Obtain proper protection domain for the call
	if (securityContext ==  0)
	    domains[0] = getDefaultProtectionDomain(null);
	else 
	    domains[0] = getTrustedProtectionDomain();

	AccessControlContext context = new AccessControlContext(domains);

	try {
	    // Perform the setField
	    AccessController.doPrivileged(new PrivilegedSetFieldAction(field, obj, val),
	 		  		  context);	
	} catch (PrivilegedActionException e) {
	    throw e;
	}
    }


    /**
     * <P> Construct a new Java object with security context.
     * </P>
     *
     * @param clazz Java object class
     * @param constructor Constructor object 
     * @param jargs jvalueArray in native code
     * @param origin Origin of the call
     * @param isUniversalBrowserRead true if enabled
     * @param isUniversalJavaPermission true if enabled
     * @return resulting object
     */
    public static Object ConstructObject(Class clazz, Constructor constructor, Object[] args, 
		  			 String origin, boolean isUniversalBrowserRead,
					 boolean isUniversalJavaPermission) throws Exception
    {
	try
	{
	    // Check if the JavaScript call is allowed
	    checkLiveConnectCaller(clazz, origin, isUniversalBrowserRead);

	    // Set up virtual JavaScript Protection Domain
	    ProtectionDomain[] domains = new ProtectionDomain[1];

	    // Obtain proper protection domain for the call
	    if (isUniversalJavaPermission == false)
		domains[0] = getDefaultProtectionDomain(origin);
	    else 
		domains[0] = getTrustedProtectionDomain();

	    AccessControlContext context = new AccessControlContext(domains);

	    // Perform the object constructor.
	    return AccessController.doPrivileged(new PrivilegedConstructObjectAction(constructor, args),
						 context);
	} catch (Exception e) {
	    Trace.liveConnectPrintException(e);
	    throw e;
	}
    }


    /**
     * <P> Call a method on Java object with security context.
     * </P>
     *
     * @param clazz Java object class
     * @param obj Java object which method is called
     * @param method Method object 
     * @param jargs jvalueArray in native code
     * @param origin Origin of the call
     * @param isUniversalBrowserRead true if enabled
     * @param isUniversalJavaPermission true if enabled
     * @return resulting object
     */
    public static Object CallMethod(Class clazz, Object obj, Method method, Object[] args, 
		  		    String origin, boolean isUniversalBrowserRead,
				    boolean isUniversalJavaPermission) throws Exception
    {
	try
	{
	    // Check if the JavaScript call is allowed
	    checkLiveConnectCaller(clazz, origin, isUniversalBrowserRead);

	    // Set up virtual JavaScript Protection Domain
	    ProtectionDomain[] domains = new ProtectionDomain[1];

	    // Obtain proper protection domain for the call
	    if (isUniversalJavaPermission == false)
		domains[0] = getDefaultProtectionDomain(origin);
	    else 
		domains[0] = getTrustedProtectionDomain();

	    AccessControlContext context = new AccessControlContext(domains);

	    // Perform the method invocation.
	    return AccessController.doPrivileged(new PrivilegedCallMethodAction(method, obj, args),
						 context);
	} catch (Exception e) {
	    Trace.liveConnectPrintException(e);
	    throw e;
	}
    }


    /**
     * <P> Get a field on Java object with security context.
     * </P>
     *
     * @param clazz Java object class
     * @param obj Java object which field is accessed
     * @param field Field object 
     * @param origin Origin of the call
     * @param isUniversalBrowserRead true if enabled
     * @param isUniversalJavaPermission true if enabled
     * @return resulting object
     */
    public static Object GetField(Class clazz, Object obj, Field field, 
		   		  String origin, boolean isUniversalBrowserRead,
				  boolean isUniversalJavaPermission) throws Exception
    {
	try
	{
	    // Check if the JavaScript call is allowed
	    checkLiveConnectCaller(clazz, origin, isUniversalBrowserRead);

	    // Set up virtual JavaScript Protection Domain
	    ProtectionDomain[] domains = new ProtectionDomain[1];

	    // Obtain proper protection domain for the call
	    if (isUniversalJavaPermission == false)
		domains[0] = getDefaultProtectionDomain(origin);
	    else 
		domains[0] = getTrustedProtectionDomain();

	    AccessControlContext context = new AccessControlContext(domains);

	    // Perform the getField
	    return AccessController.doPrivileged(new PrivilegedGetFieldAction(field, obj),
		    		    		 context);	
	} catch (Exception e) {
	    Trace.liveConnectPrintException(e);
	    throw e;
	}
    }



    /**
     * <P> Set a field on Java object with security context.
     * </P>
     *
     * @param clazz Java object class
     * @param obj Java object which field is accessed
     * @param field Field object 
     * @param jval Value to set in the field
     * @param origin Origin of the call
     * @param isUniversalBrowserRead true if enabled
     * @param isUniversalJavaPermission true if enabled
     * @return resulting object
     */
    public static void SetField(Class clazz, Object obj, Field field, Object val, 
		  	        String origin, boolean isUniversalBrowserRead,
			        boolean isUniversalJavaPermission) throws Exception
    {
	try
	{
	    // Check if the JavaScript call is allowed
	    checkLiveConnectCaller(clazz, origin, isUniversalBrowserRead);

	    // Set up virtual JavaScript Protection Domain
	    ProtectionDomain[] domains = new ProtectionDomain[1];

	    // Obtain proper protection domain for the call
	    if (isUniversalJavaPermission == false)
		domains[0] = getDefaultProtectionDomain(origin);
	    else 
		domains[0] = getTrustedProtectionDomain();

	    AccessControlContext context = new AccessControlContext(domains);

	    // Perform the setField
	    AccessController.doPrivileged(new PrivilegedSetFieldAction(field, obj, val),
	 		  		  context);	
	} catch (Exception e) {
	    Trace.liveConnectPrintException(e);
	    throw e;
	}
    }


    /**
     * <P> Set up a protection domain according to the origin and
     * the other security related information.
     */
    private static ProtectionDomain getProtectionDomain(Class clazz, 
							String origin, 
							boolean byPassOrigin, 
							boolean allPermission)
				    throws OriginNotAllowedException, MalformedURLException
    {
	boolean trust = false;

	// First, check if the call is allowed
	checkLiveConnectCaller(clazz, origin, byPassOrigin);

	if (allPermission)
	{
	    return getTrustedProtectionDomain();
	}
	else
	{
	    return getDefaultProtectionDomain(origin);
	}
    }

    /**
     * <P> Check if LiveConnect call is allowed at all.
     * </P>
     */
    private static void checkLiveConnectCaller(Class clazz, 
					       String origin, 
					       boolean isUniversalBrowserRead)
			throws OriginNotAllowedException, MalformedURLException
    {
	// Check if the call is allowed:
	// 
	// The caller must either
	// a) Have the same origin of the callee, or
	// b) "UniversalBrowserRead" is enabled, or
	// c) The callee is a system code - "UniversalBrowserRead" should be true

	// If "UniversalBrowserRead" is enabled, return 
	if (isUniversalBrowserRead)
	{
	    Trace.liveConnectPrintln("JavaScript: UniversalBrowserRead enabled");
	    return;
	}

	// Obtain callee origin
	ProtectionDomain pd = clazz.getProtectionDomain();

	CodeSource cs = pd.getCodeSource();

	if (cs == null)
	{
	    // Callee is system code, anyone can call it
	    Trace.liveConnectPrintln("JavaScript: calling Java system code");
	    return;
	}	 	

	URL location = cs.getLocation();
	URL callerLocation = null;
	
	if (origin != null)
	{
	    try
	    {
		callerLocation = new URL(origin);
	    }
	    catch (MalformedURLException e)
	    {
		e.printStackTrace();
		return;
	    }
	}

	// If origin is the same, return
	if (location != null && callerLocation != null)
	{
	    // Compare only the scheme, host and port
	    if (location.getProtocol().equalsIgnoreCase(callerLocation.getProtocol())
		&& location.getHost().equalsIgnoreCase(callerLocation.getHost())
		&& location.getPort() == callerLocation.getPort())
	    {
		Trace.liveConnectPrintln("JavaScript: caller and callee have same origin");
		return;
	    }
	}
	
	throw new OriginNotAllowedException("JavaScript is not from the same origin as the Java code, caller="
					    + callerLocation + ", callee=" + location);
    }


    /**
     * Returns a protection domain that represents the default permission
     * for a give URL.
     *
     * @param urlString URL
     * @return protection domain.
     */
    private static ProtectionDomain getDefaultProtectionDomain(String urlString)
				    throws MalformedURLException
    {
	Trace.liveConnectPrintln("JavaScript: default security policy = " + urlString);

	URL url = null;

	if (urlString != null)
	{
	    try
	    {
		url = new URL(urlString);
	    }
	    catch (MalformedURLException e)
	    {
		// URL is malformed, so don't use URL to deduce
		// security policy - ignore exception
	    }
	}

	// Obtain Java policy
	Policy policy = Policy.getPolicy();
	CodeSource cs = new CodeSource(url, null);
	final PermissionCollection pc = policy.getPermissions(cs);

	if (url.getProtocol().equals("file")) {

	    pc.add(new FilePermission("<<ALL FILES>>", "read"));
	    pc.add(new SocketPermission("localhost", "connect,accept"));

	    AccessController.doPrivileged(new PrivilegedAction() {
		public Object run() {
		    try {
			String host = InetAddress.getLocalHost().getHostName();
			pc.add(new SocketPermission(host,"connect,accept"));
		    } catch (UnknownHostException uhe) {
			
		    }
		    return null;
		}
	    });
	}

	return new JavaScriptProtectionDomain(pc);
    }


    /**
     * Returns a protection domain that represents all permission.
     *
     * @return All permission protection domain.
     */
    private static ProtectionDomain getTrustedProtectionDomain()
    {
	Trace.liveConnectPrintln("JavaScript: UniversalJavaPermission enabled");

	Permissions pc = new Permissions();
	pc.add(new AllPermission());

	return new JavaScriptProtectionDomain(pc);
    }
}


/**
 * <P> PrivilegedCallMethodAction is used for executing the priviledge 
 * action for constructing a new Java object.
 * </P>
 */
class PrivilegedConstructObjectAction implements PrivilegedExceptionAction {

    Constructor constructor;
    Object[] args;

    PrivilegedConstructObjectAction(Constructor constructor, Object[] args)
    {
	this.constructor = constructor;
	this.args = args;

	// Ensure the argument is not null
	if (this.args == null)
	    this.args = new Object[0];
    }

    public Object run() throws Exception {
	return constructor.newInstance(args);
    }
}


/**
 * <P> PrivilegedCallMethodAction is used for executing the priviledge 
 * action for calling a method in an object.
 * </P>
 */
class PrivilegedCallMethodAction implements PrivilegedExceptionAction {

    Method method;
    Object obj;
    Object[] args;

    PrivilegedCallMethodAction(Method method, Object obj, Object[] args)
    {
	this.method = method;
	this.obj = obj;
	this.args = args;

	// Ensure the argument is not null
	if (this.args == null)
	    this.args = new Object[0];
    }

    public Object run() throws Exception {
	return method.invoke(obj, args);
    }
}


/**
 * <P> PrivilegedGetFieldAction is used for executing the priviledge action
 * for getting a field in an object.
 * </P>
 */
class PrivilegedGetFieldAction implements PrivilegedExceptionAction {

    Field field;
    Object obj;

    PrivilegedGetFieldAction(Field field, Object obj)
    {
	this.field = field;
	this.obj = obj;
    }

    public Object run() throws Exception {
	return field.get(obj);
    }
}


/**
 * <P> PrivilegedSetFieldAction is used for executing the priviledge action
 * for setting a field in an object.
 * </P>
 */
class PrivilegedSetFieldAction implements PrivilegedExceptionAction {

    Field field;
    Object obj;
    Object val;

    PrivilegedSetFieldAction(Field field, Object obj, Object val)
    {
	this.field = field;
	this.obj = obj;
	this.val = val;
    }

    public Object run() throws Exception {
	field.set(obj, val);
    	return null;
    }
}
