/*
 * @(#)PluginCertificateStore.java	1.13 00/08/28
 *
 * Copyright (c) 1995, 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * Java Plug-in 
 *
 * @author Stanley Man-Kit Ho
 */ 

package sun.plugin.security;

import java.io.*;
import java.security.*;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Random;
import sun.plugin.usability.Trace;


/**
 * PluginCertificateStore is a class that represents the permanent certificate 
 * store which contains all the certificates that Java Plug-in recognize. It 
 * is used in the certification verification process when signed applet is 
 * encountered.
 */
public class PluginCertificateStore implements CertificateStore
{
    private static String _filename = null;

    static
    {
        String userHome = (String) java.security.AccessController.doPrivileged(
               new sun.security.action.GetPropertyAction("user.home"));

        String nodotversion = (String) java.security.AccessController.doPrivileged(
               new sun.security.action.GetPropertyAction("javaplugin.nodotversion"));

	// Determine filename of the JPI cert store
	_filename = userHome 
		    + File.separator + ".java"
		    + File.separator + "jpicerts" + nodotversion;
    }

    // Collection of JPI certificates
    private KeyStore _jpicerts = null; 

    /**
     * Load the certificate store into memory.
     */
    public void load() throws IOException, CertificateException,
			      KeyStoreException, NoSuchAlgorithmException
    {
	Trace.securityPrintln("Loading JPI certificates from " + _filename);

	try
	{ 
	    AccessController.doPrivileged(new PrivilegedExceptionAction() {
			
		public Object run() throws IOException, CertificateException,
					   KeyStoreException, NoSuchAlgorithmException
		{
		    if (_jpicerts == null)
		    {
			// Create keystore first
			_jpicerts = KeyStore.getInstance("JKS");

			// Initialize the keystore
			_jpicerts.load(null, null);
		    }

		    File file = new File(_filename);

		    // Only load the cert store if it exists
		    if (file.exists())
		    {
			FileInputStream fis = new FileInputStream(new File(_filename));
			BufferedInputStream bis = new BufferedInputStream(fis);
	    		
			_jpicerts.load(bis, new char[0]);

			bis.close();
			fis.close();
		    }


		    return null;
		}
   	   });
	}
	catch (PrivilegedActionException e)
	{
	    Exception ex = e.getException();

	    if (ex instanceof IOException)
		throw (IOException)ex;
	    else if (ex instanceof CertificateException)
		throw (CertificateException)ex;
	    else if (ex instanceof KeyStoreException)
		throw (KeyStoreException)ex;
	    else if (ex instanceof NoSuchAlgorithmException)
		throw (NoSuchAlgorithmException)ex;
	    else
		Trace.securityPrintException(e);
	}

	Trace.securityPrintln("Loaded JPI certificates from " + _filename);
    }


    /**
     * Persist the certificate store.
     */
    public void save() throws IOException, CertificateException,
			      KeyStoreException, NoSuchAlgorithmException
    {
	Trace.securityPrintln("Saving JPI certificates in " + _filename);

	try
	{ 
	    AccessController.doPrivileged(new PrivilegedExceptionAction() {
		    
		public Object run() throws IOException, CertificateException,
					   KeyStoreException, NoSuchAlgorithmException
		{
		    FileOutputStream fos = new FileOutputStream(new File(_filename));
		    BufferedOutputStream bos = new BufferedOutputStream(fos);

	    	    _jpicerts.store(bos, new char[0]);

		    bos.close();
		    fos.close();

		    return null;
		}
    	     });
	}
	catch (PrivilegedActionException e)
	{
	    Exception ex = e.getException();
	    
	    if (ex instanceof IOException)
		throw (IOException)ex;
	    else if (ex instanceof CertificateException)
		throw (CertificateException)ex;
	    else if (ex instanceof KeyStoreException)
		throw (KeyStoreException)ex;
	    else if (ex instanceof NoSuchAlgorithmException)
		throw (NoSuchAlgorithmException)ex;
	    else
		Trace.securityPrintException(e);
	}

	Trace.securityPrintln("Saved JPI certificates in " + _filename);
    }


    /**
     * Add a certificate into the certificate store.
     * 
     * @param cert Certificate object.
     */
    public void add(Certificate cert) throws KeyStoreException 
    {
	Trace.securityPrintln("Adding certificate in JPI permanent certificate store");

	if (contains(cert) == false)
	{
    	    // Generate a unique alias for the certificate
	    Random rand = new Random();
	    boolean found = false;

	    String alias = null;

	    // Loop until we found a unique alias that is not in the store
	    do 
	    {
		alias = "jpicert" + rand.nextLong();
		
    		Certificate c = _jpicerts.getCertificate(alias);

		if (c == null)
		    found = true;
	    }
	    while (found == false);

	    _jpicerts.setCertificateEntry(alias, cert);

	    Trace.securityPrintln("Added certificate in JPI permanent certificate store as alias " + alias);
	}
    }


    /**
     * Remove a certificate from the certificate store.
     * 
     * @param cert Certificate object.
     */
    public void remove(Certificate cert) throws KeyStoreException 
    {
	Trace.securityPrintln("Removing certificate in JPI permanent certificate store");

	String alias = _jpicerts.getCertificateAlias(cert);

	if (alias != null)
	    _jpicerts.deleteEntry(alias);

	Trace.securityPrintln("Removed certificate in JPI permanent certificate store as alias " + alias);
    }


    /**
     * Check if a certificate is stored within the certificate store.
     *
     * @param cert Certificate object.
     * @return true if certificate is in the store.
     */
    public boolean contains(Certificate cert) throws KeyStoreException
    {
	Trace.securityPrintln("Checking if certificate is in JPI permanent certificate store");

	// Certificate alias returned only if there is a match
	String alias = null;
	
	alias = _jpicerts.getCertificateAlias(cert);
 
	return (alias != null);
    }


    /**
     * Verify if a certificate is issued by one of the certificate
     * in the certificate store.
     *
     * @param cert Certificate object.
     * @return true if certificate is issued by one in the store.
     */
    public boolean verify(Certificate cert)
    {
	Trace.securityPrintln("Check if certificate can be verified using certificates in JPI permanent certificate store");

	// Plug-in Certificate store is not intended to be used for verification.
	return false;
    }


    /**
     * Obtain all the certificates that are stored in this 
     * certificate store.
     *
     * @return Iterator for iterating certificates
     */
    public Iterator iterator() throws KeyStoreException
    {
	Trace.securityPrintln("Obtain certificate iterator in JPI permanent certificate store");

	HashSet certCollection = new HashSet();
	Enumeration enum = _jpicerts.aliases();
	
	while (enum.hasMoreElements())
	{
	    // Get certificate alias from iterator
	    String alias = (String) enum.nextElement();

	    // Get certificate from store
    	    Certificate cert = _jpicerts.getCertificate(alias);

	    // Add certificate into collection
	    certCollection.add(cert);
	}		

	return certCollection.iterator();
    }
}
