/*
 * @(#)PluginClassLoader.java	1.23 00/10/04
 *
 * Copyright 1995-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.plugin.security;

import java.net.URL;
import java.security.*;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.CertificateParsingException;
import sun.applet.AppletClassLoader;
import java.net.MalformedURLException;
import java.util.jar.*;
import java.util.jar.Attributes.Name;
import java.util.zip.*;
import java.util.Vector;
import java.util.HashMap;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.net.*;
import java.io.*;
import sun.misc.URLClassPath;
import sun.misc.Resource;
import sun.plugin.usability.Trace;
import sun.plugin.usability.TraceFilter;
import sun.plugin.usability.DialogFactory;
import sun.plugin.resources.ResourceHandler;

/**
 * This class defines the class loader for loading applet classes and
 * resources. It extends AppletClassLoader to search the applet code base
 * for the class or resource after checking any loaded JAR files.
 */
public class PluginClassLoader extends AppletClassLoader {

    private static RuntimePermission usePolicyPermission;
    private URL base;

    private URLClassPath  cachedUCP;

    // ProtectionDomain cache
    private HashMap pdCache = new HashMap();
    private boolean pdCacheDirty = true;

    /*
     * Creates a new PluginClassLoader for the specified base URL.
     */
    public PluginClassLoader(URL base) {
	super(base);
	this.base = base;
    }

    /**
     * Returns the permissions for the given codesource object.
     * The implementation of this method first calls super.getPermissions,
     * to get the permissions
     * granted by the super class, and then adds additional permissions
     * based on the URL and signers of the codesource.
     *
     * @param cs the codesource
     * @return the permissions granted to the codesource
     */
    protected PermissionCollection getPermissions(CodeSource cs)
    {
	// Try to obtain the PermissionCollection from cache.
	//
	// Notice that PermissionCollection is cache on a CodeSource
	// basis, so there should be no potential security hole.
	//
	PermissionCollection perms = null;

	// MUST check if the dirty bit is false. If so, permission
        // collection MUST be regenerated, bug #4454977
	if (pdCacheDirty == false) 
	{
	     perms = (PermissionCollection) pdCache.get(cs);
	    // If pc is found, just return
	    if (perms != null)
	    {
		return perms;
	    }
	}

	// Otherwise, try to determine it on the fly
	perms = super.getPermissions(cs);

	// Inject permission to read certain generic browser properties
	//
	perms.add(new java.util.PropertyPermission("browser", "read"));
	perms.add(new java.util.PropertyPermission("browser.version", "read"));
	perms.add(new java.util.PropertyPermission("browser.vendor", "read"));
	perms.add(new java.util.PropertyPermission("http.agent", "read"));

	// Inject permission to read certain Java Plug-in specific properties
	//
	perms.add(new java.util.PropertyPermission("javaplugin.version", "read"));
	perms.add(new java.util.PropertyPermission("javaplugin.proxy.config.type", "read"));
	perms.add(new java.util.PropertyPermission("javaplugin.proxy.config.list", "read"));
	perms.add(new java.util.PropertyPermission("javaplugin.proxy.config.bypass", "read"));


	if (usePolicyPermission == null)
	    usePolicyPermission = new RuntimePermission("usePolicy");

	if (!perms.implies(usePolicyPermission) &&
	    (cs.getCertificates() != null)) {
	    // code is signed, and user wants to be prompted for AllPermission  

	    try
	    {
		if (TrustDecider.isAllPermissionGranted(cs))
		    perms.add(new java.security.AllPermission());
	    }
	    catch (CertificateExpiredException e1)
	    {
		Trace.securityPrintException(e1, ResourceHandler.getMessage("rsa.cert_expired"),
					     ResourceHandler.getMessage("security_dialog.caption"));    
	    }
	    catch (CertificateNotYetValidException e2)
	    {
		Trace.securityPrintException(e2, ResourceHandler.getMessage("rsa.cert_notyieldvalid"),
					     ResourceHandler.getMessage("security_dialog.caption"));    
	    }
	    catch (Exception e3)
	    {
		Trace.securityPrintException(e3, ResourceHandler.getMessage("rsa.general_error"),
					     ResourceHandler.getMessage("security_dialog.caption"));    
	    }
	}

	if (cachedUCP != null) {
	    URL[] urls = cachedUCP.getURLs();
	    for (int i=0;i<urls.length;i++) {
		perms.add(new java.io.FilePermission(urls[i].getFile(), "read"));
	    }
	}

	// Store the protection domain in the cache
	pdCache.put(cs, perms);

	// Reset dirty bit, bug #4454977
	pdCacheDirty = false;

	return perms;
    }

    /*
     * Returned cached ProtectionDomain for the specified CodeSource.
     */
    private ProtectionDomain getProtectionDomain(CodeSource cs) {
	if (cs == null)
	    return null;

	/* 
	 * Fix for #4341616. ProtectionDomain should not be cached, because
	 * we may have mixture of signed/unsigned jar files in the 
	 * archive/cache_archive, and classes in each JAR file will have
	 * different protection domain.
	 */
	ProtectionDomain pdm = new ProtectionDomain(cs, getPermissions(cs));

	return pdm;
    }

    /*
     * Adds the specified JAR file to the search path of loaded JAR files.
     */
    public void addMyJar(URL base, String name) {
	URL url;
	try {
	    url = new URL(base, name);
	} catch (MalformedURLException e) {
	    throw new IllegalArgumentException("name");
	}
	addURL(url);
	// DEBUG
	//URL[] urls = getURLs();
	//for (int i = 0; i < urls.length; i++) {
	//    System.out.println("url[" + i + "] = " + urls[i]);
	//}
    }

  
  private Vector cachedJarFiles;
  private String className;
  
  /*
   * Sets the Vector Containing the Jar File names which have to be installed.
   * Also gets the base URL
   * @param. f. Represents the path to the jar file
   */
  public synchronized void addCacheJar(Vector f) {
    
      if (f==null)
	  return;
      
      if (f.isEmpty()) 
	  return;

      if (null == cachedUCP)
	  cachedUCP = new URLClassPath(new URL[0]);

      for (Enumeration e = f.elements(); e.hasMoreElements();) {
	  try {
	      // Let's avoid adding duplicates...
	      URL[] cachedURLs = cachedUCP.getURLs();
	      boolean found = false;
	      URL url = ((File) e.nextElement()).toURL();

	      for (int i=0;i<cachedURLs.length && found == false; i++) 
		  found = cachedURLs[i].equals(url); 

	      if (!found) {
		  cachedUCP.addURL(url);
		  // Set dirty bit, bug #4454977
		  pdCacheDirty = true;
	      }
	  } catch (MalformedURLException mue) {
	      System.err.println("Plugin cache corrupted, using remote copy");
	      return;
	  }
      }
  }
  

    /**
     * Finds and loads the class with the specified name from the URL search
     * path. Any URLs referring to JAR files are loaded and opened as needed
     * until the class is found.
     *
     * @param name the name of the class
     * @return the resulting class
     * @exception ClassNotFoundException if the class could not be found
     */
    protected Class findClass(final String name)
	 throws ClassNotFoundException
    {
	if (null == cachedUCP) 
	    return super.findClass(name);

	try {
	    return (Class)
		AccessController.doPrivileged(new PrivilegedExceptionAction() {
		    public Object run() throws ClassNotFoundException {
			String path = name.replace('.', '/').concat(".class");
			Resource res = cachedUCP.getResource(path, false);
			if (res != null) {
			    try {
				return defineClass(name, res);
			    } catch (IOException e) {
				throw new ClassNotFoundException(name, e);
			    }
			} else {
			    return PluginClassLoader.super.findClass(name);
			}
		    }
		    });
			//		}, AccessController.getContext());
	} catch (java.security.PrivilegedActionException pae) {
	    throw (ClassNotFoundException) pae.getException();
	}
    }

    /*
     * Defines a Class using the class bytes obtained from the specified
     * Resource. The resulting Class must be resolved before it can be
     * used.
     */
    private Class defineClass(String name, Resource res) throws IOException {
	int i = name.lastIndexOf('.');
	URL url = res.getCodeSourceURL();
	if (i != -1) {
	    String pkgname = name.substring(0, i);
	    // Check if package already loaded.
	    Package pkg = getPackage(pkgname);
	    Manifest man = res.getManifest();
	    if (pkg != null) {
		// Package found, so check package sealing.
		boolean ok;
		if (pkg.isSealed()) {
		    // Verify that code source URL is the same.
		    ok = pkg.isSealed(url);
		} else {
		    // Make sure we are not attempting to seal the package
		    // at this code source URL.
		    ok = (man == null) || !isSealed(pkgname, man);
		}
		if (!ok) {
		    throw new SecurityException("sealing violation");
		}
	    } else {
		if (man != null) {
		    definePackage(pkgname, man, url);
		} else {
                    definePackage(pkgname, null, null, null, null, null, null, null);
                }
	    }
	}
	// Now read the class bytes and define the class
	byte[] b = res.getBytes();
	java.security.cert.Certificate[] certs = res.getCertificates();
	CodeSource cs = new CodeSource(base, certs);
	return defineClass(name, b, 0, b.length, getProtectionDomain(cs));
    }

    /*
     * Returns true if the specified package name is sealed according to the
     * given manifest.
     */
    private boolean isSealed(String name, Manifest man) {
	String path = name.replace('.', '/').concat("/");
	Attributes attr = man.getAttributes(path);
	String sealed = null;
	if (attr != null) {
	    sealed = attr.getValue(Name.SEALED);
	}
	if (sealed == null) {
	    if ((attr = man.getMainAttributes()) != null) {
		sealed = attr.getValue(Name.SEALED);
	    }
	}
	return "true".equalsIgnoreCase(sealed);
    }

    /**
     * Finds the resource with the specified name on the URL search path.
     *
     * @param name the name of the resource
     * @return a <code>URL</code> for the resource, or <code>null</code> 
     * if the resource could not be found.
     */
    public URL findResource(String name) 
    {
	Resource res = null;
	if (null != cachedUCP) 
	    res = cachedUCP.getResource(name, false);
        return  res != null ? res.getURL() : super.findResource(name);
    }
    
    /**
     * Returns an Enumeration of URLs representing all of the resources
     * on the URL search path having the specified name.
     *
     * @param name the resource name
     * @exception if an I/O exception occurs
     * @return an <code>Enumeration</code> of <code>URL</code>s
     */
    public Enumeration findResources(String name) throws IOException 
    {
	if (null != cachedUCP) 
	{
	    final Enumeration e0 = cachedUCP.getResources(name, true);
	    final Enumeration e1 = findResourcesByURLPath(name);

	    return new Enumeration() {
		    public Object nextElement() 
		    {
			// Return resources in cache UCP first
			if (e0 != null && e0.hasMoreElements())
			    return ((Resource)e0.nextElement()).getURL();

			// Return resources in system UCP next
			if (e1 != null && e1.hasMoreElements())
			    return ((Resource)e1.nextElement()).getURL();

			throw new NoSuchElementException();
		    }
		    public boolean hasMoreElements() 
		    {
			return ((e0 != null && e0.hasMoreElements())
				|| (e1 != null && e1.hasMoreElements()));
		    }
		};	
	} else {
	    return findResourcesByURLPath(name);
	}
    }

    /* 
     * Returns an enumeration of URLs representing the resources
     * on the codebase having the specified name.
     */
    private Enumeration findResourcesByURLPath(final String name) throws IOException
    {
	// The following code is to avoid the sun.misc.Service to 
	// locate the sound service from the applet codebase.
	// Since the service is located for every sound file, this
	// has became a performance bottleneck. The workaround will
	// be to avoid looking up the sound file class from the
	// codebase. This speeds up the audio file loading through 
	// HTTP/HTTPS quite a lot. [stanleyh]
	//		
	if (name != null && name.indexOf("META-INF/services/javax.sound") != -1)
	{
	    return new Enumeration() {
		    public Object nextElement() 
		    {
			throw new NoSuchElementException();
		    }
		    public boolean hasMoreElements() 
		    {
			return false;
		    }
		};	
	}

	return super.findResources(name);
    }



    /**
     * Finalize method.
     */
    public void finalize()
    {
	Trace.println("Classloader [" + base + "] is finalized.", TraceFilter.JAVA_CONSOLE_ONLY);
    }
}
