/*
 * @(#)RootCACertificateStore.java	1.17 00/08/29
 *
 * Copyright (c) 1995, 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into 
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * Java Plug-in 
 *
 * @author Stanley Man-Kit Ho
 */

package sun.plugin.security;

import java.io.*;
import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.util.Collection;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Enumeration;
import sun.plugin.usability.Trace;
import sun.plugin.usability.TraceFilter;
import sun.plugin.usability.DialogFactory;


/**
 * RootCACertificateStore is a class that represents the certificate 
 * stores which contains all the root CA certificates. It is used in 
 * the certification verification process when signed applet is encountered.
 */
public class RootCACertificateStore implements CertificateStore
{
    private static String _filename = null;

    static
    {
        String javaHome = (String) java.security.AccessController.doPrivileged(
               new sun.security.action.GetPropertyAction("java.home"));

	// Determine filename of the root CA cert store
	_filename = javaHome 
		    + File.separator + "lib"
		    + File.separator + "security"
		    + File.separator + "cacerts";
    }

    // Collection of root CA cert keystore
    private KeyStore _cacerts = null;

    /**
     * Load the certificate store into memory.
     */
    public void load() throws IOException, CertificateException,
			      KeyStoreException, NoSuchAlgorithmException
    {
	Trace.securityPrintln("Loading Root CA certificates from " + _filename);

	try
	{     
	    AccessController.doPrivileged(new PrivilegedExceptionAction() {
    
		public Object run() throws IOException, CertificateException,
					   KeyStoreException, NoSuchAlgorithmException
		{
		    if (_cacerts == null)
		    {
			// Create keystore first
			_cacerts = KeyStore.getInstance("JKS");

			// Initialize the keystore
			_cacerts.load(null, null);
		    }

		    FileInputStream fis = new FileInputStream(new File(_filename));		    
		    BufferedInputStream bis = new BufferedInputStream(fis);

		    _cacerts.load(bis, null);

		    bis.close();
		    fis.close();

		    return null;
		}
	    });
	}
	catch (PrivilegedActionException e)
	{
	    Exception ex = e.getException();

	    if (ex instanceof IOException)
		throw (IOException)ex;
	    else if (ex instanceof CertificateException)
		throw (CertificateException)ex;
	    else if (ex instanceof KeyStoreException)
		throw (KeyStoreException)ex;
	    else if (ex instanceof NoSuchAlgorithmException)
		throw (NoSuchAlgorithmException)ex;
	    else
		Trace.securityPrintException(e);
	}
	
	Trace.securityPrintln("Loaded Root CA certificates from " + _filename);
    }

 
    /**
     * Persist the certificate store.
     */
    public void save() throws IOException, CertificateException,
			      KeyStoreException, NoSuchAlgorithmException
    {
	Trace.securityPrintln("Saving Root CA certificates in " + _filename);

	// We don't persist anything into the root CA store.

	// Do nothing
    }


    /**
     * Add a certificate into the certificate store.
     *
     * @param cert Certificate object.
     */
    public void add(Certificate cert) throws KeyStoreException 
    {
	Trace.securityPrintln("Adding certificate in Root CA certificate store");

	// We don't add any certificate into the root CA store.

	throw new KeyStoreException("Cannot add root CA certificate");
    }


    /**
     * Remove a certificate from the certificate store.
     *
     * @param cert Certificate object.
     */
    public void remove(Certificate cert) throws KeyStoreException 
    {
	Trace.securityPrintln("Removing certificate in Root CA certificate store");

	// We don't remove any certificate into the root CA store.

	throw new KeyStoreException("Cannot remove root CA certificate");
    }


    /**
     * Check if a certificate is stored within the certificate store.
     *
     * @param cert Certificate object.
     * @return true if certificate is in the store.
     */
    public boolean contains(Certificate cert) throws KeyStoreException 
    {
	Trace.securityPrintln("Checking if certificate is in Root CA certificate store");

	// We don't compare any certificate into the root CA store.

	return false;
    }


    /**
     * Verify if a certificate is issued by one of the certificate
     * in the certificate store. 
     *
     * @param cert Certificate object.
     * @return true if certificate is issued by one in the store.
     */ 
    public boolean verify(Certificate cert) throws KeyStoreException
    {
	Trace.securityPrintln("Check if certificate can be verified using certificates in Root CA certificate store");

	StringBuffer sb = new StringBuffer();

	sb.append("Certificate to be verified:\n");
	sb.append(cert.toString());
 
	Trace.securityPrintln(sb.toString(), TraceFilter.JAVA_CONSOLE_ONLY);
 
	// Enumerate each root CA certificate in the root store
	Enumeration enum = _cacerts.aliases();
 
	while (enum.hasMoreElements())
	{
	    String alias = (String) enum.nextElement();

	    Certificate rootCert = _cacerts.getCertificate(alias);

	    sb.setLength(0);	
	    sb.append("Comparing certificate against Root CA certificate:\n");	
	    sb.append(rootCert.toString());

	    Trace.securityPrintln(sb.toString(), TraceFilter.JAVA_CONSOLE_ONLY);

	    try
	    {
    		cert.verify(rootCert.getPublicKey());

		Trace.securityPrintln("Certificate has been verified with Root CA certificates sucessfully");
		return true;
	    }
	    catch (Exception e)
	    {
		// Ignore exception
	    }
	}

	Trace.securityPrintln("Certificate has failed the verification with the Root CA certificates");

	return false;
    }


    /**
     * Obtain all the certificates that are stored in this 
     * certificate store.
     *
     * @return Iterator for iterating certificates
     */
    public Iterator iterator() throws KeyStoreException
    {
	Trace.securityPrintln("Obtain certificate iterator in Root CA certificate store");

	throw new KeyStoreException("Cannot obtain certificate iterator in Root CA certificate");
    }
}
