/*
 * @(#)TrustDecider.java	1.22 01/04/21
 *
 * Copyright (c) 1995-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * CopyrightVersion 1.1
 *
 * Java Plug-in 
 *
 * @author Jan Luehe
 * @author Stanley Man-Kit Ho
 */

package sun.plugin.security;

import java.io.*;
import java.security.*;
import java.security.cert.*;
import sun.plugin.resources.ResourceHandler;
import sun.plugin.usability.DialogFactory;
import sun.plugin.usability.Trace;
import sun.plugin.usability.TraceFilter;


public class TrustDecider 
{
    // Certificate stores used for signed applet verification
    //
    private static CertificateStore rootStore = new RootCACertificateStore();
    private static CertificateStore permanentStore = new PluginCertificateStore();
    private static CertificateStore sessionStore = new SessionCertificateStore();


    public static boolean isAllPermissionGranted(CodeSource cs)
			  throws CertificateEncodingException, CertificateExpiredException, 
			         CertificateNotYetValidException, CertificateParsingException,
				 CertificateException, KeyStoreException, 
				 NoSuchAlgorithmException, IOException
    {
	java.security.cert.Certificate[] certs = cs.getCertificates();
	String url = cs.getLocation().toString();

	// If no certificate is found, simply return false.
	if (certs == null)
	{
	    if (isBrowserMozilla())
		return isAllPermissionGranted(cs, url, null, null, 0, null, 0, "");
	    else
		return false;
	}

	// determine each signer and its certificate chain
	int start = 0;
	int end = 0;	    

	while (end < certs.length) {
	    int i = start;
	    CertificateExpiredException certExpiredException = null;
	    CertificateNotYetValidException certNotYetValidException = null;

	    while ((i+1) < certs.length) {
		if ((certs[i] instanceof X509Certificate)
		    && (certs[i+1] instanceof X509Certificate)
		    && isIssuerOf((X509Certificate)certs[i],
				  (X509Certificate)certs[i+1])) {

		    // Check if the certificate is valid and has not expired.
		    //
		    try
		    {
			((X509Certificate) certs[i]).checkValidity();
			((X509Certificate) certs[i+1]).checkValidity();
		    }
		    catch (CertificateExpiredException e1)
		    {
			if (certExpiredException == null)
			    certExpiredException = e1;
		    }
		    catch (CertificateNotYetValidException e2)
		    {
			if (certNotYetValidException == null)
			    certNotYetValidException = e2;
		    }

		    i++;
		} 
		else 
		{
		    break;
		}
	    }
	    end = i + 1;


	    // If browser support certificate verification, we will delegate
	    // the action to the browser.
	    if (isBrowserMozilla())
	    {
		// Make certificate chain
		byte[][] chain = makeChain(certs, start, end);
		if (chain != null) 
		{
		    int[] certLengths = new int[chain.length];

		    for (int k=0; k<chain.length; k++)
			certLengths[k] = chain[k].length;

		    // Get fingerprint and common name
		    X509Certificate cert = (X509Certificate) certs[0];
		    byte[] sig = cert.getSignature();					
		    Principal prinSubject = cert.getSubjectDN();
		    String subjectDNName = prinSubject.getName();

		    if (isAllPermissionGranted(cs, url, chain, certLengths,
					       chain.length, sig, sig.length, 
					       subjectDNName))
			return true;
		}
	    }
	    else
	    {
		// Otherwise, the browser doesn't provide any certificate
		// verification support, and we need to popup a dialog and 
		// determine the outcome

		boolean trustDecision = false;

		rootStore.load();
		permanentStore.load();
		sessionStore.load();		
		
		// First, we need to verify if the certificate chain is
		// signed by a CA
		//
		if (rootStore.verify(certs[i]) == false)
		{
		    throw new CertificateException("Unable to verify the certificate with root CA");
		}


		// We need to determine if the certificate has been stored
		// in the permanent certificate store.
		//
		if (permanentStore.contains(certs[start]))
		    return true;

		// We need to determine if the certificate has been stored
		// in the session certificate store
		if (sessionStore.contains(certs[start]))
		    return true;
	

		// Check if the user still want to proceed if the cert is expired.
		if (certExpiredException != null)
		{
		    Object[] msgs = new Object[2];
		    msgs[0] = ResourceHandler.getMessage("rsa.cert_expired_prompt_user");
		    msgs[1] = certExpiredException;

		    int result = 0;

		    // Check if automation is enabled
		    if (Trace.isAutomationEnabled() == false)
		    {
			result = DialogFactory.showConfirmDialog(msgs, 
				 ResourceHandler.getMessage("security_dialog.caption"));    
		    }
		    else
		    {
	    		Trace.securityPrintln("Automation: Ignore certificate expiration", TraceFilter.JAVA_CONSOLE_ONLY);
			result = 0;
		    }

		    if (result != 0)
			throw certExpiredException;
		}

		// Check if the user still want to proceed if the cert is not yield valid.
		if (certNotYetValidException != null)
		{
		    Object[] msgs = new Object[2];
		    msgs[0] = ResourceHandler.getMessage("rsa.cert_notyieldvalid_prompt_user");
		    msgs[1] = certNotYetValidException;

		    int result = 0;

		    // Check if automation is enabled
		    if (Trace.isAutomationEnabled() == false)
		    {
			result = DialogFactory.showConfirmDialog(msgs, 
								 ResourceHandler.getMessage("security_dialog.caption"));    
		    }
		    else
		    {
	    		Trace.securityPrintln("Automation: Ignore certificate expiration (not yield valid)", TraceFilter.JAVA_CONSOLE_ONLY);
			result = 0;
		    }

		    if (result != 0)
			throw certNotYetValidException;
		}


		// This certificate chain has not been encountered before, popup
		// certificate dialog

		int action = showSecurityDialog(cs); 

		// Persist the action in either the permanent or session 
		// certificate store.
		if (action == TrustDeciderDialog.TrustOption_GrantThisSession)
		{
		    Trace.securityPrintln("User has granted the priviledges to the code for this session only",
					  TraceFilter.JAVA_CONSOLE_ONLY);

		    // Grant this session 
		    sessionStore.add(certs[start]);
		    sessionStore.save();
		    trustDecision = true;
		}
		else if (action == TrustDeciderDialog.TrustOption_GrantAlways)
		{
		    Trace.securityPrintln("User has granted the priviledges to the code forever",
					  TraceFilter.JAVA_CONSOLE_ONLY);

		    // Grant always 
		    permanentStore.add(certs[start]);
		    permanentStore.save(); 
		    trustDecision = true;
	 	}
		else
		{ 
		    Trace.securityPrintln("User has denied the priviledges to the code",
					  TraceFilter.JAVA_CONSOLE_ONLY);

		    // Deny
		}

		return trustDecision;
	    } 

	    start = end;
	}

	return false;
    }


    /*
     * Native method to determine if the current browser supports 
     * any type of certificate verification. e.g. Mozilla
     *
     * @return true if the browser is Mozilla.
     */
    private static native boolean isBrowserMozilla();


    /*
     * Native method to determine if permission should be granted
     * given the code source and the certificate chain.
     *
     * @param cs CodeSource
     * @param url URL
     * @param chain Byte array of certificate chain
     * @param certLengths Array of certificate length
     * @param numOfCerts Number of certificate in the chain.
     * @param fingerPrint Byte array of finger print
     * @param fpLength Length of finger print
     * @param subjectDN Subject DN name
     * @return true if permission should be granted.
     */
    private static native boolean isAllPermissionGranted(CodeSource cs,
							 String url,
							 byte[][] chain,
							 int[] certLengths,
							 int numOfCerts,
							 byte[] fingerPrint,
							 int fpLength,
							 String subjectName);

    /*
     * Concatenates the given certificates to a chain.
     *
     * @param certs the certificates.
     * @param start the index into <code>certs</code> where the chain
     * starts.
     * @param end the index into <code>certs</code> where the chain ends.
     *
     * @return the resulting chain, or null if all the certificates in
     * <code>certs</code> are non-X.509.
     */
    private static byte[][] makeChain(java.security.cert.Certificate[] certs,
				      int start, int end)
	throws CertificateEncodingException
    {
	if (start == end)
	    return null;

	byte[][] chain = new byte[end-start][];

	for (int i = start; i < end; i++) {
	    chain[i-start] = certs[i].getEncoded();
	}

	return chain;
    }

    /*
     * Returns true iff the issuer of <code>cert1</code> corresponds to the
     * subject (owner) of <code>cert2</code>.
     *
     * @return true iff the issuer of <code>cert1</code> corresponds to the
     * subject (owner) of <code>cert2</code>, false otherwise.
     */
    private static boolean isIssuerOf(X509Certificate cert1,
				      X509Certificate cert2)
    {
	Principal issuer = cert1.getIssuerDN();
	Principal subject = cert2.getSubjectDN();
	if (issuer.equals(subject))
	    return true;
	return false;
    }


    /**
     * <P> Display the security dialog about the codesource, and ask the user
     * for granting permission.
     * </P>
     *
     * @return 0 if "Grant always" option is selected from the security dialog.
     * @return 1 if "Deny" option is selected from the security dialog.
     * @return 2 if "Grant this session" option is selected from the security dialog.
     */
    static int showSecurityDialog(CodeSource cs) 
    {
	Integer ret = (Integer) AccessController.doPrivileged(new PrivilegedBlockAction(cs));	

	int result = 2;

	if (ret != null)
	    result = ret.intValue();

    	return result;
    }

    /**
     * <P> Private class for displaying security dialog and certificates in the
     * codesource.
     * </P>
     */
    private static class PrivilegedBlockAction implements PrivilegedAction {
	java.security.cert.Certificate[] certs;
	PrivilegedBlockAction(CodeSource cs)
	{
	    this.certs = cs.getCertificates();
	}

	public Object run() {

	    TrustDeciderDialog dialog = new TrustDeciderDialog(certs);
	    return new Integer(dialog.DoModal());
	}
    }
}
