/*
 * @(#)Trace.java	1.9 01/02/09
 *
 * Copyright 1997-2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 *
 * Java Plug-in 
 *
 * @author Stanley Man-Kit Ho
 */

package sun.plugin.usability;

import java.lang.ref.SoftReference;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.StringTokenizer;
import sun.awt.AppContext;
import sun.plugin.resources.ResourceHandler;


/**
 * Trace is a class that provides basic tracing capability in Java Plug-in.
 * It is enabled through two properties:
 *
 * 1) javaplugin.trace
 * 2) javaplugin.trace.option
 *
 * These two properties control how fine-grained the Java Plug-in tracing
 * should be.
 */
public class Trace 
{
    private static boolean bBasicTraceOn = false;
    private static boolean bExtTraceOn = false;
    private static boolean bSecurityTraceOn = false;
    private static boolean bNetTraceOn = false;	
    private static boolean bLiveConnectTraceOn = false;	
    private static boolean bAutomationOn = false;	

    static
    {
	// Determine if tracing is enabled;
	reset();

	// Determine if the tracing is enabled at all.
	//
        String automationOn = (String) java.security.AccessController.doPrivileged(
		    new sun.security.action.GetPropertyAction("javaplugin.automation"));

	if (automationOn != null && automationOn.equalsIgnoreCase("true"))
	    bAutomationOn = true;
    }

    
    // HashMap for Trace listeners
    private static HashMap listeners = new HashMap();


    /** 
     * Initialize the Trace facilities
     */
    public static void reset()
    {
	// Determine if the tracing is enabled at all.
	//
        String traceOn = (String) java.security.AccessController.doPrivileged(
		    new sun.security.action.GetPropertyAction("javaplugin.trace"));

	// Determine how fine-grained tracing should be.
	//
	if (traceOn != null && traceOn.equalsIgnoreCase("true"))
	{
            String traceOptions = (String) java.security.AccessController.doPrivileged(
    				  new sun.security.action.GetPropertyAction("javaplugin.trace.option"));

	    if (traceOptions == null)
	    {
		bBasicTraceOn = true;
		bNetTraceOn = true;
		bSecurityTraceOn = true;
		bExtTraceOn = true;
		bLiveConnectTraceOn = true;
	    }
	    else
	    {
    		StringTokenizer st = new StringTokenizer(traceOptions, "|");
     
		while (st.hasMoreTokens()) 
		{
		    String option = (String) st.nextToken();

		    if (option == null || option.equalsIgnoreCase("all"))
		    {
			bBasicTraceOn = true;
			bNetTraceOn = true;
			bSecurityTraceOn = true;
			bExtTraceOn = true;
			bLiveConnectTraceOn = true;

			// If "all", then there is no need to check the rest
			break;
		    }
		    else if (option.equalsIgnoreCase("basic"))
		    {
			bBasicTraceOn = true;
		    }
		    else if (option.equalsIgnoreCase("net"))
		    {
			bNetTraceOn = true;
		    }
		    else if (option.equalsIgnoreCase("security"))
		    {
			bSecurityTraceOn = true;
		    }
		    else if (option.equalsIgnoreCase("ext"))
		    {
			bExtTraceOn = true;
		    }
		    else if (option.equalsIgnoreCase("liveconnect"))
		    {
			bLiveConnectTraceOn = true;
		    }
		}
	    }
	}
    }

    
    /**
     * Determine if tracing is enabled.
     *
     * @return true if tracing is enabled
     */
    public static boolean isEnabled()
    {
	return (bBasicTraceOn || bNetTraceOn || bSecurityTraceOn || bExtTraceOn || bLiveConnectTraceOn);
    }

    
    /**
     * Turn on/off basic tracing.
     *
     * @param true if basic tracing should be turned on.
     */
    public static void setBasicTrace(boolean trace)
    {
	bBasicTraceOn = trace;
    }

    
    /**
     * Turn on/off network tracing.
     *
     * @param true if network tracing should be turned on.
     */
    public static void setNetTrace(boolean trace)
    {
	bNetTraceOn = trace;
    }


    /**
     * Turn on/off security tracing.
     *
     * @param true if security tracing should be turned on.
     */
    public static void setSecurityTrace(boolean trace)
    {
	bSecurityTraceOn = trace;
    }


    /**
     * Turn on/off extension tracing.
     *
     * @param true if extension tracing should be turned on.
     */
    public static void setExtTrace(boolean trace)
    {
	bExtTraceOn = trace;
    }


    /**
     * Turn on/off LiveConnect tracing.
     *
     * @param true if liveconnect tracing should be turned on.
     */
    public static void setLiveConnectTrace(boolean trace)
    {
	bLiveConnectTraceOn = trace;
    }


    /**
     * Determine if automation is enabled.
     *
     * @return true if automation is enabled
     */
    public static boolean isAutomationEnabled()
    {
	return bAutomationOn;
    }


    /** 
     * Output message to System.out if tracing is enabled. 
     *
     * @param msg Message to be printed
     */
    public static void println(String msg)
    {
	println(msg, null);
    }


    /** 
     * Output message to System.out if tracing is enabled. 
     *
     * @param msg Message to be printed
     * @param filter Trace filter
     */
    public static void println(String msg, int filter)
    {
	println(msg, null, filter);
    }


    /** 
     * Output message to System.out if tracing is enabled. Also,
     * output message to the trace listener that is passed in
     * the parameter.
     *
     * @param msg Message to be printed
     * @param listener Trace listener
     */
    public static void println(String msg, TraceListener listener)
    {
	println(msg, listener, TraceFilter.DEFAULT);
    }


    /** 
     * Output message to System.out if tracing is enabled. Also,
     * output message to the trace listener that is passed in
     * the parameter.
     *
     * @param msg Message to be printed
     * @param listener Trace listener
     * @param filter Trace filter
     */
    public static void println(String msg, TraceListener listener, int filter)
    {
	if (bBasicTraceOn && ((filter & TraceFilter.JAVA_CONSOLE_ONLY) == TraceFilter.JAVA_CONSOLE_ONLY))
	{
	    System.err.println(msg);
	    System.err.flush();
	}


	// Fire event to the trace listener
	firePrintlnEvent(msg, listener, filter);
    }


    /** 
     * Output exception to System.out if tracing is enabled.
     *
     * @param e exception to be printed
     */
    public static void printException(Throwable e)
    {
	// Exception should be shown all the time
	e.printStackTrace();

	// Show exception only if automation is disabled
	if (isAutomationEnabled() == false)
	{
	    Object[] msgs = new Object[2];
	    msgs[0] = ResourceHandler.getMessage("usability.general_error");
	    msgs[1] = e;
	    
	    // Show Exception dialog
	    DialogFactory.showExceptionDialog(msgs, null);
	}
    }


    /** 
     * Output exception to System.out if tracing is enabled.
     *
     * @param e exception to be printed
     * @param desc Description about the exception
     * @param caption Caption for the exception dialog
     */
    public static void printException(Throwable e, String desc, String caption)
    {
	// Exception should be shown all the time
	e.printStackTrace();

	// Show exception only if automation is disabled
	if (isAutomationEnabled() == false)
	{
	    Object[] msgs = new Object[2];
	    msgs[0] = desc;
	    msgs[1] = e;
	    
	    // Show Exception dialog
	    DialogFactory.showExceptionDialog(msgs, caption);
	}
    }


    /** 
     * Output message to System.out if network tracing is enabled.
     *
     * @param msg Message to be printed
     */
    public static void netPrintln(String msg)
    {
	netPrintln(msg, TraceFilter.DEFAULT);
    }


    /** 
     * Output message to System.out if network tracing is enabled.
     *
     * @param msg Message to be printed
     * @param filter Trace filter
     */
    public static void netPrintln(String msg, int filter)
    {
	if (bNetTraceOn && ((filter & TraceFilter.JAVA_CONSOLE_ONLY) == TraceFilter.JAVA_CONSOLE_ONLY))
	{
	    System.err.println(msg);
	    System.err.flush();
	}

	// Fire event to all trace listeners
	firePrintlnEvent(msg, null, filter);
    }


    /** 
     * Output exception to System.out if net tracing is enabled.
     *
     * @param e exception to be printed
     */
    public static void netPrintException(Throwable e)
    {
	// Exception should be shown all the time
	e.printStackTrace();

	// Show exception only if automation is disabled
	if (isAutomationEnabled() == false)
	{
	    Object[] msgs = new Object[2];
	    msgs[0] = ResourceHandler.getMessage("usability.net_error");
	    msgs[1] = e;
	    
	    // Show Exception dialog
	    //DialogFactory.showExceptionDialog(msgs, null);
	}
    }


    /** 
     * Output exception to System.out if tracing is enabled.
     *
     * @param e exception to be printed
     * @param desc Description about the exception
     * @param caption Caption for the exception dialog
     */
    public static void netPrintException(Throwable e, String desc, String caption)
    {
	// Exception should be shown all the time
	e.printStackTrace();

	Object[] msgs = new Object[2];
	msgs[0] = desc;
	msgs[1] = e;
	
	// Show Exception dialog
//	DialogFactory.showExceptionDialog(msgs, caption);
    }


    /** 
     * Output message to System.out if security tracing is enabled.
     *
     * @param msg Message to be printed
     */
    public static void securityPrintln(String msg)
    {
	securityPrintln(msg, TraceFilter.DEFAULT);
    }


    /** 
     * Output message to System.out if security tracing is enabled.
     *
     * @param msg Message to be printed
     * @param filter Trace filter
     */
    public static void securityPrintln(String msg, int filter)
    {
	if (bSecurityTraceOn && ((filter & TraceFilter.JAVA_CONSOLE_ONLY) == TraceFilter.JAVA_CONSOLE_ONLY))
	{
	    System.err.println(msg);
	    System.err.flush();
	}

	// Fire event to all trace listeners
	firePrintlnEvent(msg, null, filter);
    }


    /** 
     * Output exception to System.out if security tracing is enabled.
     *
     * @param e exception to be printed
     */
    public static void securityPrintException(Throwable e)
    {
	// Exception should be shown all the time
	e.printStackTrace();

	// Show exception only if automation is disabled
	if (isAutomationEnabled() == false)
	{
	    Object[] msgs = new Object[2];
	    msgs[0] = ResourceHandler.getMessage("usability.security_error");
	    msgs[1] = e;
	    
	    // Show Exception dialog
	    DialogFactory.showExceptionDialog(msgs, null);
	}
    }


    /** 
     * Output exception to System.out if tracing is enabled.
     *
     * @param e exception to be printed
     * @param desc Description about the exception
     * @param caption Caption for the exception dialog
     */
    public static void securityPrintException(Throwable e, String desc, String caption)
    {
	// Exception should be shown all the time
	e.printStackTrace();

	// Show exception only if automation is disabled
	if (isAutomationEnabled() == false)
	{
	    Object[] msgs = new Object[2];
	    msgs[0] = desc;
	    msgs[1] = e;
	    
	    // Show Exception dialog
	    DialogFactory.showExceptionDialog(msgs, caption);
	}
    }


    /** 
     * Output message to System.out if extension tracing is enabled.
     *
     * @param msg Message to be printed
     */
    public static void extPrintln(String msg)
    {
	extPrintln(msg, TraceFilter.DEFAULT);
    }


    /** 
     * Output message to System.out if extension tracing is enabled.
     *
     * @param msg Message to be printed
     * @param filter Trace filter
     */
    public static void extPrintln(String msg, int filter)
    {
	if (bExtTraceOn && ((filter & TraceFilter.JAVA_CONSOLE_ONLY) == TraceFilter.JAVA_CONSOLE_ONLY))
	{
	    System.err.println(msg);
	    System.err.flush();
	}

	// Fire event to all trace listeners
	firePrintlnEvent(msg, null, filter);
    }

    /** 
     * Output exception to System.out if extension tracing is enabled.
     *
     * @param e exception to be printed
     */
    public static void extPrintException(Throwable e)
    {
	// Exception should be shown all the time
	e.printStackTrace();

	// Show exception only if automation is disabled
	if (isAutomationEnabled() == false)
	{
	    Object[] msgs = new Object[2];
	    msgs[0] = ResourceHandler.getMessage("usability.ext_error");
	    msgs[1] = e;
	    
	    // Show Exception dialog
	    DialogFactory.showExceptionDialog(msgs, null);
	}
    }


    /** 
     * Output exception to System.out if tracing is enabled.
     *
     * @param e exception to be printed
     * @param desc Description about the exception
     * @param caption Caption for the exception dialog
     */
    public static void extPrintException(Throwable e, String desc, String caption)
    {
	// Exception should be shown all the time
	e.printStackTrace();

	// Show exception only if automation is disabled
	if (isAutomationEnabled() == false)
	{
	    Object[] msgs = new Object[2];
	    msgs[0] = desc;
	    msgs[1] = e;
	    
	    // Show Exception dialog
	    DialogFactory.showExceptionDialog(msgs, caption);
	}
    }


    /** 
     * Output message to System.out if LiveConnect tracing is enabled.
     *
     * @param msg Message to be printed
     */
    public static void liveConnectPrintln(String msg)
    {
	liveConnectPrintln(msg, TraceFilter.JAVA_CONSOLE_ONLY);
    }


    /** 
     * Output message to System.out if LiveConnect tracing is enabled.
     *
     * @param msg Message to be printed
     * @param filter Trace filter
     */
    public static void liveConnectPrintln(String msg, int filter)
    {
	if (bLiveConnectTraceOn && ((filter & TraceFilter.JAVA_CONSOLE_ONLY) == TraceFilter.JAVA_CONSOLE_ONLY))
	{
	    System.err.println(msg);
	    System.err.flush();
	}

	// Fire event to all trace listeners
	firePrintlnEvent(msg, null, filter);
    }


    /** 
     * Output exception to System.out if LiveConnect tracing is enabled.
     *
     * @param e exception to be printed
     */
    public static void liveConnectPrintException(Throwable e)
    {
	// Exception should be shown all the time
	e.printStackTrace();

	// Notice that we should never popup dialog in LiveConnect because
	// it may be called from the main thread, and we may end up deadlock
	// ourselves.
    }


    /**
     * Add trace listener.
     *
     * @param listener TraceListener
     */
    public static void addTraceListener(ThreadGroup tg, TraceListener l)
    {
	Trace.println("Added trace listener: " + l, TraceFilter.JAVA_CONSOLE_ONLY);

	// Soft reference is used to ensure that ThreadGroup and
	// TraceListener will not be held even if the caller 
	// fails to remove the listener for some reasons.
	//
    	listeners.put(tg.toString(), new SoftReference(l));
    }

    /**
     * Remove trace listener.
     *
     * @param l TraceListener
     */
    public static void removeTraceListener(ThreadGroup tg, TraceListener l)
    {
	Trace.println("Removed trace listener: " + l, TraceFilter.JAVA_CONSOLE_ONLY);

	listeners.remove(tg.toString());
    }


    /**
     * Fire println event.
     *
     * @param msg Message
     * @param listener Trace Listener
     * @param filter Trace Filter
     */
    private static void firePrintlnEvent(String msg, TraceListener listener, int filter)
    {
	// If trace listener is defined, fire println event to the dedicated
	// trace listener. Otherwise, find the proper trace listener from
	// the registered list accordingly to the caller thread group.
	//
	if (listener != null)
	{
	    listener.println(msg, filter);
	}
	else
	{
	    synchronized(listeners)
	    {
		Set keys = listeners.keySet();    
		Iterator iter = keys.iterator();		
		Thread t = Thread.currentThread();
		ThreadGroup tg = t.getThreadGroup();
		
		while (iter.hasNext())
		{
		    Object key = iter.next();

		    if (key.equals(tg.toString()))
		    {
			// Obtain soft reference
			SoftReference ref = (SoftReference) listeners.get(key);

			Object value = null;

			if (ref != null)
			    value = ref.get();

			if (value != null)
			{
			    TraceListener l = (TraceListener) value;
			    l.println(msg, filter);		
			}
		    }    
		}
	    }
	}
    }
}


