#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "Debug.h"
#include "utils.h"
#include <unistd.h>
#include <stropts.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <locale.h>
#include <errno.h>
#include <strings.h>
#include <dlfcn.h>
#include "plugin_defs.h"
#include "pluginversion.h"
/*
 * Internal utility routine to get size of a "C" string.
 * If the string is a null pointer we return zero.
 */
int tracing = 0;

static int init_utils_called = 0;

static void check_initialized(){
    if (!init_utils_called) init_utils();
}

static FILE *tracefile = NULL;

/* Initialization of the utility package. Should be called only once. */
void
init_utils() {
    char *name;
    if (init_utils_called) 
	return;
    init_utils_called = 1;
    name =  getenv("JAVA_PLUGIN_TRACE");
    if (name != NULL) {
	fprintf(stderr, "Turning tracing on....\n");
	tracing = 1;
	/* Parent file's name defined in plugin_defs.h */
	tracefile = fopen(PARENT_TRACE_FILE(PLUGIN_NODOTVERSION), "w");
    } else {
      tracing = 0;
    }

}

int
slen(const char *s) {
    if (s == NULL) {
	return 0;
    }
    return strlen(s);
}

int 
slenUTF(const char *s) {
    if (s == NULL)
	return 0;
    return strlen(s);
}


/*
 * Put an integer into the buffer in big-endian byte order.
 */
void
put_int(char *buff, int offset, int x) {
    buff[offset] = (x >> 24) & 0xFF;
    buff[offset+1] = (x >> 16) & 0xFF;
    buff[offset+2] = (x >> 8) & 0xFF;
    buff[offset+3] = x & 0xFF;
}

/*
 * Put as short into the buffer in big-endian byte order.
 */
void
put_short(char *buff, int offset, short x) {
    buff[offset] = (x >> 8) & 0xFF;
    buff[offset+1] = x & 0xFF;
}

/*
 * Get an integer from a buffer in big-endian byte order.
 */
int
get_int(char *buff, int offset) {
    int result = (buff[offset] << 24) |
		 ((buff[offset+1] << 16) & 0xFF0000) |
		 ((buff[offset+2] << 8) & 0xFF00) |
		 (buff[offset+3] & 0xFF);
    return result;
}

/*
 * Get a short from a buffer in big-endian byte order.
 */
short
get_short(char *buff, int offset) {
    short result = (buff[offset] << 8) |
		 (buff[offset+1] & 0xFF);
    return result;
}


void
trace_verbose(const char *format, ...) 
{
#ifdef DO_TRACE
    va_list ap;
    va_start(ap, format);
    check_initialized();
    if (verbose_tracing) {
	if (tracefile != NULL) {
	   fprintf(tracefile, "Plugin: ");
	    vfprintf(tracefile, format, ap);
	    fflush(tracefile);
	} else {
	    fprintf(stderr, "Plugin: ");
	    vfprintf(stderr, format, ap);
	}
    }
    va_end(ap);
#endif
}


void
trace(const char *format, ...) 
{
    va_list ap;
    va_start(ap, format);
    check_initialized();
    if (tracing) {
	if (tracefile != NULL) {
	  fprintf(tracefile, "Plugin: ");
	  vfprintf(tracefile, format, ap);
	  fflush(tracefile);
	} else {
	    fprintf(stdout, "Plugin: ");
	    vfprintf(stdout, format, ap);
	}
    }
    va_end(ap);
}

/* Perform a dlsym(libhandle,fn_name), with error checking for the result */
void*
load_function(void* libhandle, const char* fn_name) {
    void* result_fn;
    if (libhandle == NULL) 
	plugin_error("Cannot load %s from null library\n", fn_name);
    result_fn = (void *)dlsym(libhandle,  fn_name);
    if (result_fn == NULL) 
	plugin_error("Could not load the function %s\n", fn_name);
    return result_fn;
}

/* A formal error is one that we attempt to localize (maybe?) */
void
plugin_formal_error(const char *s) {
    fprintf(stderr, "%s\n", DGETTEXT(JAVA_PLUGIN_DOMAIN, s));
    trace("%s\n", s);
}

/* A formal error that is not localized (such as file name */
void
plugin_raw_formal_error(const char *s) {
    fprintf(stderr, "%s\n", s);
    trace("%s\n", s);
}

/* Any other plugin error */
void
plugin_error(const char *format, ...) 
{
    va_list ap;
    va_start(ap, format);
    fprintf(stderr, "INTERNAL ERROR on Browser End: ");
    vfprintf(stderr, format, ap);
    fprintf(stderr, "\n");
    perror("System error?:");
    va_end(ap);
    exit(-1);
}

/* Return true if 'subString' is a prefix of 'mainString' */
int
isPrefixOf(char* subString, char* mainString) {
    if ((mainString == NULL) || (subString == NULL)) 
        return FALSE;
    while(1) {
        subString++;
        mainString++;
        if (*subString == '\0') return TRUE;
        if (*mainString != *subString)
            return FALSE;
    }

}


#ifndef SVR4
/* File descriptor passing done manually */
int sendfd(int dest, int fd) {
    struct iovec base[1];
    struct msghdr message;
    int rc;
    base[0].iov_base = (char *)0;
    base[0].iov_len = 0;
    message.msg_iov = base;
    message.msg_iovlen = 1;
    message.msg_accrights = (caddr_t)&(fd);
    message.msg_accrightslen = sizeof(int);
    rc = sendmsg(dest, &message, 0);
    if (rc < 0) {
        fprintf(stderr, "error: %d %d\n",rc, errno);
        perror("Sending fd:");
    }
    return rc;
}


int s_pipe(int fds[2]) {
    int res = socketpair(AF_UNIX, SOCK_STREAM, 0, fds);
    if (tracing) {
	int flags0 = fcntl(fds[0], F_GETFL);
	int flags1 = fcntl(fds[1], F_GETFL);
	trace("Not SVR4 Flgs0 = %X Flgs1 = %X appnd=%X nonblck=%X sync=%X\n", 
	      flags0, flags1, O_APPEND, O_NONBLOCK, O_SYNC);
    }
    return res; 
}

#else

int sendfd(int dest, int fd) {
    int rc;
    rc = ioctl(dest, I_SENDFD, fd);
    if (rc < 0) {
        fprintf(stderr, "error: %d %d\n",rc, errno);
        perror("Sending fd:");
    }
    return rc;
}

#ifdef __linux__

/* This apparently does not seem to work on Linux. Need to use a socketpair
   insted of pipe, as the nsIPlugin <--> JPI uses a bidirectional comm
   sandeep.konchady@eng.sun.com 05/09/2000.
*/
int s_pipe(int fds[2]) {
    int res = socketpair(AF_UNIX, SOCK_STREAM, 0, fds);
    if (tracing) {
	int flags0 = fcntl(fds[0], F_GETFL);
	int flags1 = fcntl(fds[1], F_GETFL);
	trace("SVR4 Flgs0 = %X Flgs1 = %X appnd=%X nonblck=%X sync=%X\n", 
	      flags0, flags1, O_APPEND, O_NONBLOCK, O_SYNC);
    }
    return res; 
}
#else
int s_pipe(int fds[2]) {
    int res = (pipe(fds));
    if (tracing) {
	int flags0 = fcntl(fds[0], F_GETFL);
	int flags1 = fcntl(fds[1], F_GETFL);
	trace("SVR4 %d %d Flgs0 = %X Flgs1 = %X appnd=%X blk=%X sync=%X\n", 
	      fds[0], fds[1], flags0, flags1, O_APPEND, O_NONBLOCK, O_SYNC);
    }
    return res;

}
#endif
#endif





