/*
 * @(#)java_vm.c	1.23 01/03/08 
 * 
 *  Copyright (c) 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "COPYRIGHT"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  SUN
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 *
 * Author: Venkat Maddileti, Benedict Gomes, Robert Szewczyk
 */

/*
 * This process is forked from the navigator plugin to run the Java VM.
 * All the real work is done in Java in the sun.plugin.Plugin class.
 *
 *							    KGH Dec 97
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <jni.h>
#include "plugin_defs.h"
#include "pluginversion.h"
#include <assert.h>
#include <sys/param.h>
#include <dlfcn.h>

#define OLDJAVA_MODE 1
#define NEWJAVA_MODE 2

static char *find_options();
static void SetClassPath(char *s);
static void SetBootClassPath(char *s);
static jint ParseArguments( char *pargv);
static void AddOption(char *str, void *info);
static void *MemAlloc(size_t size);
static char *MemDup(char *src);
jboolean trace = JNI_FALSE;

/*
 * Code for dealing with libjvm.so is swiped from src/share/bin,
 * src/solaris/bin.  A better code sharing solution is highly desirable.
 */

typedef jint (JNICALL *CreateJavaVM_t)(JavaVM **pvm, void **env, void *args);
typedef jint (JNICALL *GetDefaultJavaVMInitArgs_t)(void *args);

typedef struct {
    CreateJavaVM_t             CreateJavaVM;
    GetDefaultJavaVMInitArgs_t GetDefaultJavaVMInitArgs;
} InvocationFunctions;


/*
 * List of VM options to be specified when the VM is created.
 */
static JavaVMOption *options;
static int numOptions;
int mode = NEWJAVA_MODE;

/* Tracing for the startup code. Write error messages to a file, rather
   than to the screen where they are lost. Use a file named
   plugin_child_startup.trace */
FILE* tracefile;

void startup_trace(char *format,...) {
    va_list arglist;
    va_start(arglist, format);
    if (trace) {
	if (tracefile != NULL) {
	    vfprintf(tracefile, format, arglist);
	    fprintf(stderr, "%s After exec, before startup:",PLUGIN_NODOTVERSION);
	    vfprintf(stderr, format, arglist);
	    fflush(tracefile);
	} else {
	    fprintf(stderr, "%s After exec, before startup:",PLUGIN_NODOTVERSION);
	    vfprintf(stderr, format, arglist);
	}
    }
    va_end(arglist);
}

void trace_env(char* envname) {
    char* env_value = getenv(envname);
    if (env_value) 
	startup_trace("Environment variable %s=%s\n",envname, env_value);
    else
	startup_trace("Environment variable %s not set\n", envname);
}


/*
 * Find the JVM entry points we need.
 */
static jboolean getJVMEntryPoints(char *java_home, InvocationFunctions *ifn) {
    void *libjvm;
    char jvmpath[MAXPATHLEN];
    const char *debug = "";

    if (getenv("JAVA_PLUGIN_DEBUG"))
	debug = "_g";

    snprintf(jvmpath, sizeof jvmpath, "%s/lib/"ARCH"/client/libjvm%s.so", java_home, debug);
    libjvm = dlopen(jvmpath, RTLD_NOW + RTLD_GLOBAL);

    if (libjvm == NULL) {
	snprintf(jvmpath, sizeof jvmpath, "%s/lib/"ARCH"/libjvm%s.so", java_home, debug);
	libjvm = dlopen(jvmpath, RTLD_NOW + RTLD_GLOBAL);
    }

    if (libjvm != NULL) {

	ifn->CreateJavaVM = (CreateJavaVM_t) dlsym(libjvm, "JNI_CreateJavaVM");
	ifn->GetDefaultJavaVMInitArgs = (GetDefaultJavaVMInitArgs_t)
	    			dlsym(libjvm, "JNI_GetDefaultJavaVMInitArgs");

	if (ifn->GetDefaultJavaVMInitArgs != NULL &&
	    ifn->CreateJavaVM != NULL)
	    return JNI_TRUE;
    }
    return JNI_FALSE;
}



int main(int argc, char **argv) {
    JavaVM *jvm;
    JNIEnv *env;
    jclass clz;
    jmethodID meth;
    char classpath[4*MAXPATHLEN], *s;
    char nodotversion[255];
    char version[255];
    char *java_home;
    char *plugin_home;
    char *optionstr;
    JavaVMInitArgs args;
    InvocationFunctions ifn;

    if (argc >1 && strcmp(argv[1], "-t") == 0) {
	trace = JNI_TRUE;
	fprintf(stderr, "Opening the trace file %s\n", EXEC_TRACE_FILE(PLUGIN_NODOTVERSION));
	tracefile = fopen(EXEC_TRACE_FILE(PLUGIN_NODOTVERSION), "w");
        startup_trace("Java process: entered main\n");

	/* Print out possible environment flags for tracing */
	trace_env("THREADS_TYPE");
	trace_env("CLASSPATH");
	trace_env("LD_LIBRARY_PATH");
	trace_env("JAVA_HOME");
	trace_env("PLUGIN_HOME");
	trace_env("_JAVA_OPTIONS");
	trace_env("JAVA_PLUGIN_AGENT");
	trace_env("JAVA_PLUGIN_ARGS");
	trace_env("JAVA_PLUGIN_DEBUG");
	trace_env("JAVA_VM_WAIT");
    }

    startup_trace("Going to sleep for debugging...\n");
    if (getenv("JAVA_VM_WAIT")) {
	    sleep(30);
    }
    startup_trace("Woke up...\n");

	/* JAVA_PLUGIN_VERSION must be set to actual plug-in version
	   and it should not be derived from the build environment
	   variable VERSION because this will be used to construct
	   the directory path name for lot of things
	   (Deva, 10/06/00)
	*/
    putenv("JAVA_PLUGIN_VERSION=" PLUGIN_VERSION);


    /* Disable DNS spoofing check */
    AddOption("-DtrustProxy=true", NULL);

    /* Verify remote code only */
    AddOption("-Xverify:remote", NULL);

    /* Set user-defined options for the Java VM */
    optionstr = find_options();
    ParseArguments(optionstr);

    java_home = getenv("JAVA_HOME");
    plugin_home = getenv("PLUGIN_HOME");
    snprintf(classpath, sizeof classpath,
	    "%s/lib/javaplugin.jar:"		/* from plugin_home */
	    "%s/lib/javaplugin_l10n.jar:",	/* from plugin_home */
	    plugin_home, plugin_home);
    startup_trace("Classpath = %s \n",classpath);
    s = classpath;

    /* Get JVM function pointers. */
    if (!getJVMEntryPoints(java_home, &ifn)) {
	fprintf(stderr, "%s\n", 
		DGETTEXT(JAVA_PLUGIN_DOMAIN, 
			 "java_vm process: could not find Java VM symbols"));
	exit(3);
    }

    /* Prepend system class path to default */
    if (mode == OLDJAVA_MODE) {
        JDK1_1InitArgs args;
        char *buf;
	int buflen;
        args.version = JNI_VERSION_1_1;
        if (ifn.GetDefaultJavaVMInitArgs(&args) != JNI_OK
                || args.classpath == 0) {
            fprintf(stderr, "Could not get default system class path.\n");
            exit(3);
        }
        startup_trace("System class path: %s \n", args.classpath);
        startup_trace("App class path: %s \n", s);
        startup_trace("OLD JAVA MODE \n");
        buflen = strlen(args.classpath) + strlen(s) + 2;
        buf = MemAlloc(buflen);
        snprintf(buf, buflen, "%s:%s", args.classpath, s);
        s = buf;
    }

    SetBootClassPath(s);

    /* Set javaplugin.home. */
    {
	int buflen = strlen(plugin_home) + 100;
	char *buf = MemAlloc(buflen);
	snprintf(buf, buflen, "-Djavaplugin.lib=%s/lib/" ARCH "/libjavaplugin_jni.so",
		plugin_home);
	AddOption(buf, NULL);
    }

    /* Workaround Mozilla bug */
    if (getenv("MOZILLA_WORKAROUND"))
      AddOption("-Dmozilla.workaround=true", NULL);

    sprintf(nodotversion, "%s%s", "-Djavaplugin.nodotversion=", PLUGIN_NODOTVERSION);
    AddOption(nodotversion, NULL);

    sprintf(version, "%s%s", "-Djavaplugin.version=", PLUGIN_VERSION);     
    AddOption(version, NULL);

    /* Set user-defined options for the Java VM */
    memset(&args, 0, sizeof(args));
    args.version  = JNI_VERSION_1_2;
    args.nOptions = numOptions;
    args.options  = options;
    args.ignoreUnrecognized = JNI_TRUE;

    startup_trace("Version is: %d \n", args.version);
    /* Now we can start up the java virtual machine */
    if (ifn.CreateJavaVM(&jvm, (void *)&env, &args) != JNI_OK) {
	fprintf(stderr, "Could not startup JVM properly!\n");
	fprintf(stderr, "%s\n", 
		DGETTEXT(JAVA_PLUGIN_DOMAIN, 
			 "java_vm process: could not start Java VM"));
	exit(3);
    }

    startup_trace("java_vm process: started Java VM OK\n");

    /* Clear any exception status */
    (*env)->ExceptionClear(env);

    /* Find the main javaplugin class */
    clz = (*env)->FindClass(env, "sun/plugin/navig/motif/Plugin");
    if (clz == 0) {
	fprintf(stderr, "%s sun/plugin/navig/motif/Plugin\n",
		DGETTEXT(JAVA_PLUGIN_DOMAIN, 
			 "java_vm process: Couldn't find class"));
	(*env)->ExceptionDescribe(env);
	exit(5);
    }

    /* Find its "start" method */
    meth = (*env)->GetStaticMethodID(env, clz, "start", "(Z)V");
    if (meth == 0) {
	fprintf(stderr, "%s\n", 
		DGETTEXT(JAVA_PLUGIN_DOMAIN, 
			 "Java process: Couldn't find Plugin.start"));
	(*env)->ExceptionDescribe(env);
	exit(6);
    }
    startup_trace("Ready to call start method\n");
    /* Call into the Java code */
    (*env)->CallStaticVoidMethod(env, clz, meth, trace);
    if ((*env)->ExceptionOccurred(env)) {
	fprintf(stderr, "%s\n",
		DGETTEXT(JAVA_PLUGIN_DOMAIN, "Java process: caught exception from sun.plugin.navig.motif.Plugin.start"));
        (*env)->ExceptionDescribe(env);
    }

    exit(0);
}

static char *
find_options()
{
    FILE *fin;
    char *home = getenv("HOME");
    char path[MAXPATHLEN];
    char line[MAXPATHLEN];
    char *result = 0;
    static char *key = "javaplugin.jre.params=";

    snprintf(path, sizeof path, "%s/.java/properties%s", home,PLUGIN_NODOTVERSION);	
    fin = fopen(path, "r");
    if (fin == NULL) {
	return 0;
    }

    while (fgets(line, sizeof(line), fin)) {
	if (strstr(line, key) == line) {
	    /* Copy the value, ommiting the key and the ending newline */
	    int len = strlen(line) - strlen(key);
	    result = MemAlloc(len);
	    memcpy(result, line+strlen(key), len-1);
	    result[len-1] = 0;
	}
    }
    fclose(fin);
    if (result && strstr(result, "-Xoldjava") != NULL) mode = OLDJAVA_MODE;
    return result;
}

/*
 * Returns a pointer to a block of at least 'size' bytes of memory.
 * Prints error message and exits if the memory could not be allocated.
 */
static void *
MemAlloc(size_t size)
{
    void *p = malloc(size);
    if (p == 0) {
        perror("malloc");
        exit(1);
    }
    return p;
}

static char *
MemDup(char *src)
{
    int len = strlen(src);
    char *dst = (char *) MemAlloc(len + 1);
    strcpy(dst, src);
    return dst;
}

static void
SetClassPath(char *s)
{
    char *classpath;

    int deflen = strlen(s) + 40;
    char *def = MemAlloc(deflen);
    if (s == NULL) {
        s = ".";
    }
    if (mode == OLDJAVA_MODE) {
        snprintf(def, deflen, "-Xbootclasspath:%s", s);
    } else {        
        snprintf(def, deflen, "-Djava.class.path=%s", s);        
    }
       
    AddOption(def, NULL);
}

static void
SetBootClassPath(char *s)
{
    int deflen = strlen(s) + 40;
    char *def = MemAlloc(deflen);
    if (mode == OLDJAVA_MODE)
        snprintf(def, deflen, "-Xbootclasspath:%s", s);
    else
        snprintf(def, deflen, "-Xbootclasspath/a:%s", s);
    AddOption(def, NULL);

}

/*
 * Parses command line arguments.
 */
static jint
ParseArguments( char *pargv)
{
    char seps[] = " -";
    char *arg = NULL;
    char targ[MAXPATHLEN];
    char *clsspth = NULL;
    int nexttok = 1;
    int ioffset = 2;
    
    if (pargv != NULL)
        arg = strtok( pargv, seps );

    while (arg != NULL) {
	if (strncmp(arg, "classpath", 9) == 0 || strncmp(arg, "cp", 2) == 0) {
	    if (strncmp(arg, "classpath", 9) == 0) {
	        ioffset = 9;
	    }
	    if (strlen(arg) > ioffset) {
	    	char *pcp = arg + ioffset;
	        if (strncmp(pcp, ".", 1) == 0 || strncmp(pcp, "/", 1) == 0) {
	            clsspth = pcp;
	        } else {
	            clsspth = ".";
	        }
	    } else {
	        arg = strtok( NULL, " " );
	        if (arg == NULL) {
	    	    clsspth = ".";
	    	} else if (strncmp(arg, ".", 1) == 0 || strncmp(arg, "/", 1) == 0) {
	    	    clsspth = arg;
	    	} else if (strncmp(arg, "-", 1) == 0) {
	    	    clsspth = ".";
	    	    ++arg;
	    	    nexttok = 0;
	        } else {
	    	    clsspth = ".";
	        }
	    }
	    SetClassPath(clsspth);
	} else if (strcmp(arg, "help") == 0 ||
		   strcmp(arg, "h") == 0 ||
		   strcmp(arg, "?") == 0) {
	} else if (strcmp(arg, "verbosegc") == 0) {
	    AddOption("-verbose:gc", NULL);
	} else if (strcmp(arg, "trace") == 0) {
	    AddOption("-Xt", NULL);
	} else if (strcmp(arg, "noclassgc") == 0) {
	    AddOption("-Xnoclassgc", NULL);
	} else if (strcmp(arg, "verify") == 0) {
	    AddOption("-Xverify:all", NULL);
	} else if (strcmp(arg, "verifyremote") == 0) {
	    AddOption("-Xverify:remote", NULL);
	} else if (strcmp(arg, "noverify") == 0) {
	    AddOption("-Xverify:none", NULL);
	} else if (strncmp(arg, "ss", 3) == 0 ||
		   strncmp(arg, "oss", 4) == 0 ||
		   strncmp(arg, "ms", 3) == 0 ||
		   strncmp(arg, "mx", 3) == 0) {
	    int tmplen = strlen(arg) + 6;
	    char *tmp = MemAlloc(tmplen);
	    snprintf(tmp, tmplen, "-X%s", arg + 1); /* skip '-' */
	    AddOption(tmp, NULL);
	} else if (strcmp(arg, "checksource") == 0 ||
		   strcmp(arg, "cs") == 0 ||
		   strcmp(arg, "noasyncgc") == 0) {
	    /* No longer supported */
	    fprintf(stderr,
		    "Warning: %s option is no longer supported.\n",
		    arg);
	} else if (strcmp(arg, "classic") == 0) {
	} else if (strcmp(arg, "client") == 0) {
        } else {
	    strcpy(targ,"-");
	    strcat(targ, arg);
	    AddOption(targ, NULL);
	}
	if (nexttok == 1) {
	    arg = strtok( NULL, seps );
	}
	nexttok = 1;
    }

    return 0;
}

/*
 * Adds a new VM option with the given given name and value.
 */
static void
AddOption(char *str, void *info)
{
    int i=0, j=0;
    int len = 0;
    char optionStr[1024];

    /*
     * Expand options array if needed to accomodate at least one more
     * VM option.
     */

    JavaVMOption *tmp = MemAlloc((numOptions+1) * sizeof(JavaVMOption));
    if ( numOptions > 0 ) {
        /* Copy the old options array to the new one */
        memcpy(tmp, options, numOptions * sizeof(JavaVMOption));
    }

    /* Release old options array */
    free(options);
    options = tmp;

    /* Add the new vm option */
    /*
     * 4355034: looks like a few callers are passing an automatic
     * variable as str; by the time we use options we no longer
     * own much of the memory in the argument list :(
     * Instead of enforcing a malloc-only string rule on all our
     * callers, let's just call strdup here.
     */
    /*
     * Removed unnecessary '\' character from the option string.
     */
    len = strlen(str);

    for (i=0, j=0; i < len; i++, j++)
      {
        if (str[i] == '\\')
          {
            i++;
          }

        if (i < len)
           optionStr[j] = str[i];
    }

    optionStr[j] = '\0';

    options[numOptions].optionString = MemDup(optionStr);
    options[numOptions].extraInfo = info;

    startup_trace("New Option[%d] = %s\n",numOptions,optionStr);
    startup_trace("New Info[%d] = %s\n",numOptions,info == NULL ? "null" : info);

    numOptions++;
}
