/*
 *  @(#)ProxySupport.cpp	1.9 00/08/04
 *
 * Copyright (c) 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "COPYRIGHT"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  SUN
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 * 
 * Author: Robert Sczewyczk, Benedict Gomes
 *
 */

/* 
 * A ProxySuport object hangs off the Plugin Factory and provides the
 * interface for retrieving proxy urls from raw URLs from the browser.
 *
 *
 * Initialization:
 *
 *     - During initialization of the plugin instance, ProxmapInitialize
 *       is called. Initialization requires a plugin instance peer, hence
 *       it has to wait until an instance is created. Use state->proxy_script
 *       to determine whether initialization has already happened.
 *     
 *     - Proxmap initialization finds the proxy URL from the 
 *       .netscape/preferences file (user_pref("network.proxy.type", ..");
 *       user_pref("network.proxy.autoconfig_url," ..");
 *     
 *     - If the proxy type is 2, use autoproxies and use the config url from
 *       the preferences file
 *
 *     - If the proxy type is not 2, set state->proxy_config_url to -1
 *
 *     - Initialize the read_buff to various javascript
 *       functions that are defined in ProxyScript.h, including
 *       FindProxyForURL  
 *
 *     - Get the stream associated with the proxy_config_url, in 
 *       a GetURL call (JAVA_PLUGIN_PROXMAP_INPUT)
 * 
 *     - When the proxy data returns, append it to the end of read_buff.
 *  
 *     - Inform the VM that we are ready for proxy queries 
 *         JAVA_PLUGIN_PROXMAP_READY, which sets up a hash table for
 *         proxy mappings on the browser side
 *
 * Calls for proxies by the browser:
 *    - The handler for proxies is first called in:
 *        sun.plugin.ActivatorProxyHandler.findProxyForURL("url")
 *     
 *    - This is a native method in native.c, which calls
 *       Worker.getProxySettings
 *  
 *    - getProxySettings sends a proxy request (JAVA_PLUGIN_FIND_PROXY)
 *      to the browser
 *    
 *    - The browser makes the proxy request on a randomly chosen instance.
 *      It calls ProxmapFindProxy on the instance's ProxySupport
 *
 *    - ProxySupport::ProxmapFindProxy makes 2 get URL calls. It first
 *       evaluates the state->proxy_script, which has various functions
 *       defined in it.
 * 
 *    - The second URL call is JAVA_PLUGIN_PROXMAP_QUERY, which issues the
 *      javascript: FindProxyForURL command defined earlier
 *
 *    - The response comes in the JavaStream, which then calls 
 *       ProxySupport::ProxmapReply 
 *       JAVA_PLUGIN_PROXY_MAPPING
 *
 *    - ProxmapReply sends the response back to the browser.
 *
 *                                       :Benedict Gomes
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include "ProxyScript.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include "Navig4.h"

extern "C" {
#ifdef __linux__
    extern int gethostname(char *, size_t);
#else
    extern int gethostname(char *, int);
#endif
}

ProxySupport::ProxySupport(CJavaVM  *javaVM) {
    m_javaVM = javaVM;
    read_buff = NULL;
    read_buff_size = 0;
}

/* 
 * Initialize the proxy support by sending a query to the browser
 * of the autoproxy URL, which results in several calls back 
 * to ProxmapInput and finally ProxmapInputDone.
 */
void ProxySupport::ProxmapInitialize(nsIPluginInstance *inst) {

    const char *cp;
    char *bp;
    char host[256];
    struct hostent *hp;
    struct in_addr in;

    trace("ProxymapInitialize: starting...\n");
    state = m_javaVM->state;
    if (state->proxy_config_url == NULL) {
        trace("ProxymapInitialize - FindingProxyConfigUrl...\n");
        FindProxyConfigUrl();	
    }

    if (state->proxy_config_url == (char *)-1) {
        trace("ProxymapInitialize - proxies not supported\n");
        return;
    }
    if (state->proxy_script != NULL){
        trace("ProxymapInitialize: proxy_script not null \n");
        return;
    }
    if (read_buff != NULL) {
        trace("ProxymapInitialize: readbuf not null %s\n", read_buff);
        return;
    }

    // We setup some standard JavaScript functions expected by
    // the user supplied configuration scripts.
    trace("ProxymapInitialize: setting up buffer...\n");
    read_buff_size = 12000;
    read_buff = (char *) malloc(read_buff_size);
    read_buff[0] = 0;
    strcat(read_buff, "javascript: ");
    strcat(read_buff, g_jsGlobal);
    strcat(read_buff, g_jsDnsDomainIs);
    strcat(read_buff, g_jsIsInNet);
    strcat(read_buff, g_jsIsPlainHostName);
    strcat(read_buff, g_jsIsResolvable);
    strcat(read_buff, g_jsLocalHostOrDomainIs);
    strcat(read_buff, g_jsDnsDomainLevels);
    strcat(read_buff, g_jsDnsResolve);

    // Find our host 
    gethostname(host, sizeof(host));

    // Find our IP address 
    hp = gethostbyname(host);

    // Verify that the IP address is ok 
    if (hp == 0 || hp->h_addr_list == 0 || hp->h_addr_list[0] == 0) {
        fprintf(stderr, DGETTEXT(JAVA_PLUGIN_DOMAIN,"Plugin: can't find IP address for current host"));
    } else {
        trace("Found host address\n");
        memcpy(&in.s_addr, hp->h_addr_list[0], sizeof (in.s_addr));
    }

    // Starting at the of the current readbuff,  append g_jsMyIpAddress,
    // after substituting  the numeric version of our host address 
    // into any place where there is a %s in g_jsMyIpAddress
    bp = read_buff + strlen(read_buff);
    cp = g_jsMyIpAddress;
    while (*cp) {
        if (cp[0] == '%' && cp[1] == 's') {
            cp += 2;
            sprintf(bp, "%s", inet_ntoa(in));
            bp = bp + strlen(bp);
            continue;
        }
        *bp++ = *cp++;
    }
    *bp++ = 0;

    // Add some other functions
    strcat(read_buff, g_jsShExpMatch);
    strcat(read_buff, g_jsEnableDateRange);
    strcat(read_buff, g_jsEnableTimeRange);
    strcat(read_buff, g_jsEnableWeekdayRange);
    
    nsIPluginManager *mgr =  m_javaVM->GetPlugin()->GetPluginManager();
    mgr->GetURL(inst, state->proxy_config_url, 0, 
                (void *)JAVA_PLUGIN_PROXMAP_INPUT);
}

/*
 * As a result of proxmap initialization this method gets called, 
 * indirectly via a stream. The URL request is JAVA_PLUGIN_PROXMAP_INPUT
 * and the response comes back through a CJavaStream. The result
 * is appended to the read_buff. When a plugin instance gets 
 * JAVA_PLUGIN_PROXMAP_INPUT, it indicates that getting input is done and
 * ProxmapInputDone is called, then the state->proxy_script is set to 
 * the read_buff
 */
void ProxySupport::ProxmapInput(char *buffer, int len) {
    int old_length = strlen(read_buff);
    int new_length = old_length + len;

    trace("Entering proxmapInput\n");
    // Store the result in the read_buff. 
    // This is efficient but confusing - BG
    if (new_length >= read_buff_size) {
        int new_size = new_length + 5000;
        char *new_buff = (char *)malloc(new_size);
        memcpy(new_buff, read_buff, old_length);
        free(read_buff);
        read_buff = new_buff;
        read_buff_size = new_size;
    }
    // Store the result in the read buffer
    memcpy(read_buff + old_length, buffer, len);
    read_buff[new_length] = 0;
    trace("Result of proxmap input in read_buf %s\n", read_buff);
}

/*
 * Called when a JAVA_PLUGIN_PROXMAP_INPUT returns with a write
 * on the JavaPluginInstance instead of creating a stream.
 * This indicates that all the initialization is over and 
 * we can store the read_buff in proxy_script
 * Called by some plugin instance.
 */
void ProxySupport::ProxmapInputDone(nsPluginReason reason) {
    char buff[4];
    trace("ProxySupport::ProxmapInputDone\n");
    if (reason != nsPluginReason_Done) {
        char *errmsg = DGETTEXT(JAVA_PLUGIN_DOMAIN, 
                                "Plugin: trouble reading Netscape proxy "
                                " configuration URL from\n");
        fprintf(stderr, "%s\n", errmsg);
        fprintf(stderr, "           %s\n", state->proxy_config_url);
        state->proxy_config_url = 0;
        return;
    }
    // Finally, we can set the proxmap in the worker, so that 
    // proxy requests can go through
    trace("Sending JAVA_PLUGIN_PROXMAP_READY to Worker with proxy\n");
    state->proxy_script = read_buff;
    
    // Tell the child process we're ready 
    put_int(buff, 0, JAVA_PLUGIN_PROXMAP_READY);
    m_javaVM->SendRequest(buff, 4, FALSE);
}
 
/* Internal use - private method Called by the proxmap initialization
 * to find out the autoproxy URL if the state->proxy_config_url is not
 * set.
 */
void ProxySupport::FindProxyConfigUrl()
{
    char *home = getenv("HOME");
    FILE *fin;
    char path[1024];
    char line[200];
    char config_url[200];
    int type, rc;

    state->proxy_config_url = (char *)-1;
    config_url[0] = 0;
    type = 0;

    trace("FindProxyConfigUrl: Looking at navigator version\n");
    if (m_javaVM->GetPlugin()->GetNavigatorVersion() <= 3) {
        // We don't support autoconfig with NS3. 
        trace("FindProxyConfigUrl: Navigator 3.0 or less\n");
        return;
    }

    sprintf(path, "%s/.netscape/preferences.js", home);	
    trace("preferences file: %s \n", path);
    fin = fopen(path, "r");
    if (fin == NULL) {
        trace("FindProxyConfigUrl: Could not open preferences %s\n", path);
        return;
    }

    // Look for the either autoconfig_url and/or the proxy type 
    while (fgets(line, sizeof(line), fin)) {
        trace("Next line of preferences: %s\n", line);
        sscanf(line, "user_pref(\"network.proxy.type\", %d);", &type);
        rc = sscanf(line,
                    "user_pref(\"network.proxy.autoconfig_url\", \"%s\");", 
                    config_url);
        if (rc == 1) {
            // Get rid of (put null before) trailing junk after ); 
            char *tail = strstr(config_url, "\");");
            if (tail) {
                tail[0] = 0;
            }
        }
    }
    fclose(fin);
    trace("Closed preferences file\n");

    // See if http (type==2 ?) uses the autoproxy.
    if (type == 2 && config_url[0] != 0) {
        state->proxy_config_url = (char *)malloc(4+strlen(config_url));
        strcpy(state->proxy_config_url, config_url);
        trace("FindProxyConfigUrl: type==2 \n\t%s \n\t%s \n ", 
              config_url, state->proxy_config_url);
    } else {
        trace("FindProxyConfigUrl:  -1 return \n ");
        state->proxy_config_url = (char *) -1;
    }
}


/*
 * This function will eventually provide the proxy for a given URL
 * by querying the browser
 */
void ProxySupport::ProxmapFindProxy(nsIPluginInstance *inst,
                                    char * url,
                                    char * host) {
    char *command;
    int len;
    len = strlen(url) + strlen(host) + 80;
    command = (char *)malloc(len);
    sprintf(command, "javascript: FindProxyForURL('%s', '%s');", url, host);
    trace("ProxmapFindProxy being called %s %s\n", url, host);
    // First issue (or reissue) the script definition.  Then issue the
    // specific mapping request * We'll get an answer later through a
    // stream write * redircted through to proxmap_reply
    nsIPluginManager *mgr = m_javaVM->GetPlugin()->GetPluginManager();

    mgr->GetURL(inst, state->proxy_script, 0 , (void *)0x333);
    mgr->GetURL(inst, command, 0, (void *) JAVA_PLUGIN_PROXMAP_QUERY);
 

}

/* 
 * Called by the browser (via a stream) with the response of a proxy request
 * from a worker thread
 */
void ProxySupport::ProxmapReply(const char *stream_url, int len, void *buffer)
{
    char *url;
    char *cp;
    char *packet;
    int l;

    // Figure out the URL that was being resolved.
    trace("Proxmap reply for an original JAVA_PLUGIN_PROXY_REQUEST");
    url = strchr((char*)stream_url, '\'') + 1;
    cp = strchr(url, '\'');
    *cp = 0;

    trace("Mapping for proxy : %s ", url);
    packet = (char *) malloc(strlen(url) + len + 8);
    put_int(packet, 0, JAVA_PLUGIN_PROXY_MAPPING);
    cp = packet + 4;
    l = strlen(url);
    put_short(cp, 0, l);
    cp += 2;
    memcpy(cp, url, l);
    cp += l;
    put_short(cp, 0, len);
    cp += 2;
    memcpy(cp, buffer, len);
    cp += len;

    m_javaVM->SendRequest(packet, l + len + 8, FALSE);

    free(packet);
}

