/*
 * @(#)CSecurityContext.cpp	1.9 01/02/09
 *
 *  Copyright (c) 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "COPYRIGHT"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  SUN
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 */

///=------------------------------------------------------------------------=
//
// CSecurityContext.cpp  by Stanley Man-Kit Ho
//
///=------------------------------------------------------------------------=
//
// These functions create object, invoke method, get/set field in JNI with
// security context.
//
 
#include "nsplugin.h"
#include "nsAgg.h"
#include "nsISecurityContext.h"
#include "remotejni.h"
#include "Navig5.h"
#include "CSecurityContext.h"


// Mozilla changes : Added a semicolon at the end
NS_DEFINE_IID(kISecurityContextIID, NS_ISECURITYCONTEXT_IID);
NS_DEFINE_IID(kISupportsIID, NS_ISUPPORTS_IID);

// This macro expands to the aggregated query interface scheme.
NS_IMPL_AGGREGATED(CSecurityContext);

NS_METHOD
CSecurityContext::AggregatedQueryInterface(const nsIID& aIID, void** aInstancePtr)
{
    if (aIID.Equals(kISupportsIID)) {
      *aInstancePtr = GetInner();
      AddRef();
      return NS_OK;
    }
    if (aIID.Equals(kISecurityContextIID)) {
        *aInstancePtr = (nsISecurityContext *)this;
        AddRef();
        return NS_OK;
    }
    return NS_NOINTERFACE;
}



// Implements the CSecurityContext object for encapsulating security context.
// This security context object encapuslates the security associated with
// a particular JS call from Java. The ctx is a pointer back into the
// vm's security.
CSecurityContext::CSecurityContext(nsISupports *aOuter, 
				   JavaPluginFactory5* factory, 
				   jobject ctx, const char* lpszURL,
                                   PRBool isAllPermission) : m_isAllPermission(isAllPermission)
{
    trace("CSecurityContext::CSecurityContext\n");

    NS_INIT_AGGREGATED(aOuter);

    /* ctx will be NULL if there is no security context */
    plugin_factory = factory;
    context = ctx;
    if (lpszURL != NULL)
        strcpy(m_szURL, lpszURL);
	
    /*  if (env)  context = env->NewGlobalRef(ctx); Do this on the VM side*/
}


// Destroy the security context. Who does this?
CSecurityContext::~CSecurityContext()  
{
    trace("CSecurityContext::~CSecurityContext\n");

    /* We retrieve the env for this thread every time since we don't
       know in which thread this code will actually be called
       (java script may be executed in any thread) */
    RemoteJNIEnv* env = plugin_factory->GetRemoteEnvForThread();

    /* We don't know when the destructor will be called, so
       we can't do this delete on the other end. Assymetric with
       constructor */
    if (env)
        env->DeleteGlobalRef(context);
}


// Create the CSecurityContext object for creating object, invoking method, 
// getting/setting field in JNI with security context.
NS_METHOD
CSecurityContext::Create(nsISupports* outer, JavaPluginFactory5* factory, 
			 jobject ctx,const char* lpszURL,PRBool isAllPermission, 
			 const nsIID& aIID, void* *aInstancePtr)
{
    /* If this is an aggregated object creation (i.e. outer exists)
       then we expect to return the special object Internal, which
       provides the nondelegating functions */

    if (outer && !aIID.Equals(kISupportsIID))
        return NS_NOINTERFACE;   // XXX right error?
    CSecurityContext* context = new CSecurityContext(outer, factory, ctx, lpszURL, isAllPermission);
    if (context == NULL)
        return NS_ERROR_OUT_OF_MEMORY;
    context->AddRef();
    *aInstancePtr = (outer != NULL)
	?(void *)context->GetInner()
	:(void *)context;
    return NS_OK;
}


////////////////////////////////////////////////////////////////////////////
// from nsISecurityContext:
//

// Get the security context to be used in LiveConnect.
// This is used for JavaScript <--> Java.
//
// @param target        -- Possible target.
// @param action        -- Possible action on the target.
// @return              -- NS_OK if the action is permitted 
//                      -- NS_FALSE otherwise.
NS_IMETHODIMP CSecurityContext::Implies(const char* target, 
					const char* action, 
					PRBool* bActionAllowed)
{
    trace("CSecurityContext::Implies\n");

    if (target == NULL || bActionAllowed == NULL)
        return NS_ERROR_NULL_POINTER;

    /* 
       Send a method back on the wire, using the remote jni env.  If
       this is a recursive call, then we are on the same thread as the
       original, and we should get the same remote jni env.  If this
       is a spontaneous call, the remote jni env that is created
       should use the spontaneous pipe.  In either case, this issue
       should be handled when creating/registering the RemoteJNIEnv
       */

/*
    RemoteJNIEnv* env = plugin_factory->GetRemoteEnvForThread();

    if (env->CSecurityContextImplies(context, target, action)) 
	*bActionAllowed = PR_TRUE;
    else
	*bActionAllowed = PR_FALSE;
*/

    *bActionAllowed = m_isAllPermission;

    return NS_OK;
}


////////////////////////////////////////////////////////////////////////////
// from nsISecurityContext:
//

// Get the origin associated with the context.
//
// @param buf        -- Result buffer (managed by the caller.)
// @param len        -- Buffer length.
// @return           -- NS_OK if the origin string was obtained.
//                   -- NS_FALSE otherwise.

NS_IMETHODIMP CSecurityContext::GetOrigin(char* buf, int len)
{
    trace("CSecurityContext::GetOrigin\n");
    
    if (buf == NULL)
        return NS_ERROR_NULL_POINTER;

    if (len <= strlen(m_szURL))
        return NS_ERROR_FAILURE;

    // Copy origin
    strcpy(buf, m_szURL);


    return NS_OK;
}


////////////////////////////////////////////////////////////////////////////
// from nsISecurityContext:
//

// Get the certificate associated with the context.
// 
//  @param buf        -- Result buffer (managed by the caller.)
//  @param len        -- Buffer length.
//  @return           -- NS_OK if the certificate string was obtained.
//                    -- NS_FALSE otherwise.

NS_IMETHODIMP CSecurityContext::GetCertificateID(char* buf, int len)

{
    trace("CSecurityContext::GetCertificateID\n");

    return NS_OK;
}

   

