/*
 *  Copyright (c) 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "COPYRIGHT"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  SUN
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 *
 * Author: Benedict Gomes, Robert Szewczyk
 */

#ifndef JAVAVM5_H
#define JAVAVM5_H

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include "jni.h"

class JavaPluginInstance5;

class JavaPluginFactory5;

enum nsJVMStatus {
    nsJVMStatus_Enabled,  // but not Running
    nsJVMStatus_Disabled, // explicitly disabled
    nsJVMStatus_Running,  // enabled and started
    nsJVMStatus_Failed    // enabled but failed to start
};

struct LongTermState {

    /* Command pipe used for applet and thread initiation reqursts and resp */
    PRFileDesc* command_pipe;

    /* Worker pipe used to receive applet related requests */
    PRFileDesc* work_pipe;

    /* Java to Browser pipe used for spontaneous JS requests/responses */
    PRFileDesc* spont_pipe;

    /* Java to Browser pipe used for printing return data. 
     XXXFIX no printing support yet */
    PRFileDesc* print_pipe;

    /* Socket for accepting attach_thread requests */
    PRFileDesc* server_socket;

    /* Port for connecting to the server socket */
    int port;

    /* The directory the Java Plug-in JRE is installed in */
    char* java_dir;

    /* The netscape version.  */
    int navigator_version;

    /* URL used for automatic proxy configuration (null if not used) */
    char* proxy_config_url;

    /* The proxy configuration script, for auto proxy configuration. */
    char* proxy_script;

    /* Process ID of child process */
    int child_pid;

    int comm_sema;
};

class JavaVM5 {
public:
    /* Constructor and destructor */
    JavaVM5(JavaPluginFactory5 *plugin);
    virtual ~JavaVM5();
    /*  
	Initialize is separate from the constructor The constructor is
	called from the plugin manager constructor. At that point, the
	version of navigator is not known.  The navigator version can
	only be obtained after the plugin factory initialization, when
	the plugin manager is provided.  At this point, we can decide
	whether or not to load libjni.so.  Hence, we have a separate
	Initialize method which is called from the PluginFactory
	initialization.
	*/
    void Initialize(int navigator_version);
    /* Fork and exec the java_vm process, after setting up 
       the child environment*/
    int StartJavaVM(const char * addClasspath);
    /* Terminate the connection to the VM process */
    void ShutdownJavaVM(int RealShutdown);
    /* Create a new applet */
    void CreateApplet(const char *appletType, int appletNumber, 
				   int argc, char **argn, char **argv);

    /* Return the plugin factory */
    JavaPluginFactory5* GetPluginFactory(void);

    /* Long term state of the plugin */
    struct LongTermState *state;

    /* Send a request on the command pipe */
    void SendRequest(char *buff, int len, int wait_for_reply);


    /* Return true if the VM is enabled */
    int GetJVMEnabled(void);

    /* Return the "real" status of the VM */
    nsJVMStatus GetJVMStatus(void);

    /* Enable the JVM */
    void SetJVMEnabled(PRBool enable);

    /* Return the class path */
    const char* GetClassPath(void);
    
    /* Add to the current class path */
    void AddToClassPath(const char *dirPath);

    /* Separate thread function that handles work requests */
    void ProcessWorkQueue(void); 

    /* Separate thread function that handles spontaneous requests */
    void ProcessSpontaneousQueue(void); 

    /* Read a string on the worker pipe i.e. length, chars */
    char* ReadWorkString(PRFileDesc* pipe);
    
    /* Read byte on the worker pipe i.e. length, chars */
    char* ReadWorkByte(PRFileDesc* pipe); 
	
    /* Sample calls that test the VM */
    void TestVM(void);

    /* Create a remote jni  for a secure jni */
    RemoteJNIEnv* CreateRemoteJNIEnv(JNIEnv* proxyjni);

    /* Get the java object for the instance numbered 'plugin_number' */
    jobject GetJavaObjectForInstance(int plugin_number);
    
    /* Receive printing data from the Plugin. 
     */
    void ReceivePrinting(FILE* fp);

protected:
    /* Get some response data over the command pipe. Use with care about
      interleaving requests/responses. */
    void GetResponseData(char *buff, int len);

    /* Allocate a dummy widget for worker pipe communication */
    void AllocateEventWidget(void);


    /* Send a blocking request on the command pipe, with an INT result */
    int SendBlockingRequest(char *buff, int len);
    
    NS_IMETHOD_(void) TerminateRequestAbruptly(char* msg);
    
    /* Actually do the work for a single request */
    void DoWork();

    /* Terminate a work request if the instance is dead */ 
    void TerminateWorkOnNullInstance(PRFileDesc* pipe, int code);

    /* Create the env variable for the class path */
    void SetUpClasspath(const char * jre);

    /* An error occurred when reading on the worker pipe. */
    void WorkError(int x);

    void Cleanup(void);

    /* Setup the child process environment variables and write them into 
     * the static variables env_*
     * LD_LIBRARY_PATH etc.
     */
    void SetupChildEnvironment(void);

    /* Find the location of the JRE, starting with either the .java/properties
     * javaplugin.jre.path= or the location of the plugin i.e. the 
     * javadirectory obtained by FindJavaDir
     */
    char* FindJRE(void);

    int FindJVMType(void);

    bool FindEnabled(void);

    /* Find the location of the directory where the plugin java code
     * might live, starting with NPX_PLUGIN_PATH and then 
     * HOME/.netscape/plugins 
     */
    void FindJavaDir(void);
 
private:


    /*
     * Hold values of environment variables. These are setup by 
     * SetupChildEnvironment before the fork, but are not actually
     * putenv'ed. After the fork they are putenv'ed into the child.
     * In this manner, the startup of the child does not involve
     * the mallocs and other code required to set up these environment
     * variables and the exec is close to the fork, to avoid
     * potential deadlocks due to locks held by other threads after
     * the fork1 and before the exec. See the fork1 man page.
     */
    char* env_CLASSPATH;
    char* env_LD_LIBRARY_PATH;
    char* env_JAVA_HOME;
    char* env_PLUGIN_HOME;
    char* env_THREADS_TYPE;
    char stateENVName[100];


    /* child_uses_green_threads is set if the child vm should use
       green threads. The default is green threads.
       It is set by examining the environment variable THREADS_FLAG to see 
       if it is set to native. Otherwise green threads are used.
       A putenv of the secret environment variable _JVM_THREADS_TYPE to 
       green_threads is done after the fork if this variable is true.
       Initially -1
       */
    int child_uses_green_threads; 

    /* Remote JNIEnv for the current spontaneous call */
    RemoteJNIEnv* spont_remotejni;

    /* Current VM status - whether the VM is actually running */
    nsJVMStatus jvm_status;

    /* Whether the browser thinks the VM is enabled or diabled */
    nsJVMStatus base_status;

    /* Pointer to plugin factory */
    JavaPluginFactory5* m_pPluginFactory;

    /* Indicates that the java process is ready to receive messages */
    int is_java_process_ready;

    /* The class path to the VM */
    char* jvm_classpath;
    /* The length of the classpath to the VM */
    int jvm_classpath_len;
    /* The amount of space allocated to the jvm_classpath buffer (usually
     *  greater than jvm_classpath_len to avoid allocation on every add). */
    int jvm_classpath_buflen;

    /* A monitor used to indicate to the main thread when the worker thread
       has terminated. This is used when the factory is being shutdown */
    PRMonitor* worker_termination;

    /* A widget with the child_event_handler used for sending worker
       messages to the browser */
    Widget event_widget;

    int event_window_id;
};

#endif
