/*
 * Copyright (c) 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "COPYRIGHT"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  SUN
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 */

//=--------------------------------------------------------------------------=
//
// TrustDecider.cpp by Jan Luehe
//
//=--------------------------------------------------------------------------=
//
// The API of the TrustDecider.
//
//
#define _WIN32_WINNT 0x0400

#include <windows.h>
#include "wincrypt.h"
#include "Debug.h"
#include "jni.h"
#include "nsAgg.h"
#include "sun_plugin_security_TrustDecider.h"
#include "nsplugin.h"
#include "nsIJVMPluginInstance.h"
#include "nsICapsManager.h"
#include "nsIPrincipal.h"
#include "nsICertPrincipal.h"
#include "nsICodebasePrincipal.h"
#include "CNetscapePlugin.h"
#include "CJavaPlugin.h"
#include "CJavaPluginApp.h"
#include "TrustDecider.h"
#include "CCertificateVerifier.h"
#include "PluginProxySetting.h"


// CJavaPluginApp object encapsulates the information about the Win32 plugin.
extern CJavaPluginApp g_App;

// CPluginProxySetting object encapsulates the mechanism for obtaining proxy
// setting in Navigator.
extern CPluginProxySetting g_NavigatorSetting;



NS_DEFINE_IID(kICAPSMANAGER_IID, NS_ICAPSMANAGER_IID);



jboolean 
IsAllPermissionGranted(const char *codebaseURL,
			const unsigned char **certChain,
			unsigned int *certLengths,
			int numOfCerts);

{
    jboolean ret = FALSE;

    const char *pCodebaseURL = NULL;
    unsigned char **pChain = NULL;
    jint *pCertLengths = NULL;
    unsigned int *pUCertLengths = NULL;
    // Clear all exceptions
    env->ExceptionClear();
    
    if (jCodeBaseURL) {
	pCodebaseURL = env->GetStringUTFChars(jCodeBaseURL,
					      (jboolean*)0);
    }
    
    if (jNumOfCerts > 0) {
	
	// Obtain Certificate length array
	pCertLengths = env->GetIntArrayElements(jCertLengths, JNI_FALSE);
	
	// Copy certificate length array over
	pUCertLengths = new unsigned int[jNumOfCerts];
	for (int i=0; i < jNumOfCerts; i++)
	    pUCertLengths[i] = pCertLengths[i];  
	
	// Create certificate array
	pChain = new unsigned char* [jNumOfCerts];
	
	// Copy certificate array over
	for (i=0; i<jNumOfCerts; i++) {
	    // Get certificate i
	    jobject jCert = env->GetObjectArrayElement(jChain, i);
	    
	    // Create buffer for certificate i
	    pChain[i] = new unsigned char[pUCertLengths[i]];
	    
	    env->GetByteArrayRegion((jbyteArray)jCert,
				    0,
				    pUCertLengths[i],
				    (jbyte*) pChain[i]); 
	}
    }

    // Use Caps Manager in Mozilla. 5.0 Specific
    ret = IsAllPermissionGranted(pCodebaseURL,
				 (const unsigned char **)pChain,
				 pUCertLengths, jNumOfCerts);
    
    if (pCodebaseURL)
	env->ReleaseStringUTFChars(jCodeBaseURL, pCodebaseURL);
    if (pCertLengths)
	env->ReleaseIntArrayElements(jCertLengths, pCertLengths, 0);
    if (pChain) {
	for (int i=0; i<jNumOfCerts; i++)
	    delete [] pChain[i];
	
	delete [] pChain;
    }
    if (pUCertLengths)
	delete [] pUCertLengths;

    return ret;
}

extern nsIPluginManager* pMgr;

//=--------------------------------------------------------------------------=
// IsAllPermissionGranted
//=--------------------------------------------------------------------------=
// IsAllPermissionGranted calls back Mozilla to check if a principal has
// already been granted or should be granted AllPermission. 
//
// parameters :
//  cont char*            [in]    codebase
//  const unsigned char** [in]    certificate chain
//  int*                  [in]    certificate lengths
//  int                   [in]    chain length (number of certificates)
//
// out:
//  jboolean              [out]   true iff AllPermission is granted, false
//                                otherwise
jboolean IsAllPermissionGranted(const char *codebaseURL,
				const unsigned char **certChain,
				unsigned int *certLengths,
				int numOfCerts)
{
    BOOL ret = FALSE;

    if (pPlugin)
    {
        nsICapsManager* pCapsMgr = NULL;

        if (!pMgr)   return FALSE;

        if (NS_SUCCEEDED(pMgr->QueryInterface(kICAPSMANAGER_IID,
				         (void**) &pCapsMgr)))
        {
            nsIPrincipal* pPrin = NULL;
            
            if ((codebaseURL != NULL) && (certChain != NULL))
                pCapsMgr->CreateCodeSourcePrincipal(certChain,
						    certLengths,
						    numOfCerts,
						    codebaseURL,
						    &pPrin);
            else if (certChain != NULL)
                pCapsMgr->CreateCertPrincipal(certChain, certLengths, numOfCerts,
                    					      &pPrin);
            else if (codebaseURL != NULL)
                pCapsMgr->CreateCodebasePrincipal(codebaseURL, &pPrin);
            else
            {
                pCapsMgr->Release();
                return FALSE;
            }

            //////////////////////////////
            // check if principal is trusted
            PRBool trusted;
            pPrin->IsTrusted(NULL, &trusted);

            //////////////////////////////

            if (trusted == PR_TRUE)
            {
                // Check if principal already is in CapsManager database
                nsPermission perm;
                pCapsMgr->GetPermission(pPrin, NS_ALL_PRIVILEGES, &perm);
                if ((perm == nsPermission_AllowedSession) ||
                    (perm == nsPermission_AllowedForever)) {
                    // principal already has NS_ALL_PRIVILEGES
                    ret = TRUE;
	            } else if ((perm == nsPermission_DeniedSession) ||
		               (perm == nsPermission_DeniedForever)) {
	                ret = FALSE;
                } else {
                    pCapsMgr->AskPermission(pPrin, NS_ALL_PRIVILEGES, &perm);

                    if ((perm == nsPermission_AllowedSession) ||
		                (perm == nsPermission_AllowedForever)) {
	                // grant code JDK_ALL_PERMISSION
    	                ret = TRUE;
                    }
                }
            }
            else
            {
                ret = FALSE;
            }

            pCapsMgr->Release();
        }
    }

    return ret;
}


    

