/*
 * Copyright (c) 2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "COPYRIGHT"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  SUN
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 */

//=--------------------------------------------------------------------------=
//
// NSRegistry.cpp  by Stanley Man-Kit Ho
//
//=--------------------------------------------------------------------------=

#include "stdafx.h"
#include <stdio.h>
#include <string.h>
#include "nsPlugin.h"
#include "NSRegistry.h"


//=---------------------------------------------------------------------------=
// NSRegistry::NSRegistry
//=---------------------------------------------------------------------------=
// Construct NSRegistry object.
//
NSRegistry::NSRegistry()
{
    m_file = NULL;
}


//=---------------------------------------------------------------------------=
// NSRegistry::getUserPreferenceFilePath
//=---------------------------------------------------------------------------=
// Extract current user profile file location from the Netscape registry.
//
// parameters :	lpszPath	- [in, out] User profile file path
//
// return :	BOOL	    - TRUE if successful
// 
// notes :
//
BOOL NSRegistry::getUserPreferenceFilePath(char* lpszPath)
{
    // Determine path for nsreg.dat
    BOOL bResult = FALSE;
    
    // Determine Windows directory
    char szWindowsDir[1024];

    if (GetWindowsDirectory(szWindowsDir, 1024) == 0)
	return FALSE;

    // Determine Netscape registry location
    char szRegFilePath[1024];
    sprintf(szRegFilePath, "%s\\nsreg.dat", szWindowsDir);

    // Open nsreg.dat
    m_file = fopen(szRegFilePath, "r");

    if (m_file == NULL)
    {
	// Cannot open registry file
	return FALSE;	    
    }

    // Read the entry from nsreg.dat
    REGDESC regRootRecord;

    // Read root record from registry file
    bResult = readRootRecord(&regRootRecord);

    if (bResult)
    {
	// Determine current user name
	char szCurrentUser[1024];
	bResult = getValue(&regRootRecord, "Common/Netscape/ProfileManager/LastNetscapeUser", szCurrentUser);

	if (bResult)
	{
	    char szRegPath[1024];
	    char szUserPath[1024];

	    // Determine registry key to obtain profile location
	    sprintf(szRegPath, "Users/%s/ProfileLocation", szCurrentUser);

	    // Get profile location from Netscape registry
	    bResult = getValue(&regRootRecord, szRegPath, szUserPath);

	    // Determine user profile file location
	    if (bResult)
		sprintf(lpszPath, "%s\\prefs.js", szUserPath);
	}
    }

    // Close the file
    fclose(m_file);

    return bResult;
}



//=---------------------------------------------------------------------------=
// NSRegistry::getValue
//=---------------------------------------------------------------------------=
// Read registry key value from Netscape registry. 
//
// parameters :	parentRecord	- [in] Parent registry record
//		lpszPathElts	- [in] Registry key
//		lpszValue	- [in, out] Registry key value
//
// return :	BOOL	    - TRUE if successful
// 
// notes :
//
BOOL NSRegistry::getValue(/*[in]*/const REGDESC* parentRecord, 
		          /*[in]*/const char* lpszPathElts, 
		          /*[out]*/char* lpszValue)
{
    if (m_file == NULL || parentRecord == NULL || lpszPathElts == NULL || lpszValue == NULL)
	return FALSE;

    // Determine to see if there is another '/' in the path
    char* next = strchr(lpszPathElts, '/');

    char szName[1024];
    strcpy(szName, lpszPathElts);

    if (next != NULL)
    {
	// If this isn't the last path element, then find a subkey with
	// a matching name
	//

	// Get the name from the path
	szName[next - lpszPathElts] = '\0';

	// Obtain subkey offset
	REGOFF childOffset = parentRecord->down;
	
	while (childOffset != NULL)
	{
	    REGDESC childRecord;

	    // Read child record
	    if (readRecord(childOffset, &childRecord) == FALSE)
		return FALSE;


	    // Check if the child record is the proper subkey
	    char szKey[1024];

	    if (readString(childRecord.name, childRecord.namelen, szKey) == FALSE)
		return FALSE;

	    if (stricmp(szName, szKey) == 0)
	    {
		// The child record is the subkey we want, recurively
		// goto next level.
		return getValue(&childRecord, next + 1, lpszValue);
	    }
	    else
	    {
		// The child record is not what the subkey we want, 
		// goto next child in the same level.
		//
		childOffset = childRecord.left;
	    }
	}

	return FALSE;
    }
    else
    {
	// This is the last path element so find an entry with
	// a matching name and return its value
	//
	REGOFF entryOffset = parentRecord->value;

	while (entryOffset != 0)
	{
	    REGDESC entry;

	    // Read record
	    if (readRecord(entryOffset, &entry) == FALSE)
		return FALSE;

	    char szKey[1024];

	    // Read the key name of the record 
	    if (readString(entry.name, entry.namelen, szKey) == FALSE)
		return FALSE;

	    // If the key names match, extract the value
	    if (stricmp(szName, szKey) == 0)
	    {
		return readString(entry.value, entry.valuelen, lpszValue);
	    }
	    else
	    {
		// This is not the correct key, get next key
		entryOffset = entry.left;
	    }	    
	}

	return FALSE;
    }
}


//=---------------------------------------------------------------------------=
// NSRegistry::readRecord
//=---------------------------------------------------------------------------=
// Read registry record from Netscape registry. 
//
// parameters :	offset	    - [in] File offset
//		regDesc	    - [in, out] Registry Record
//
// return :	BOOL	    - TRUE if successful
// 
// notes :
//
BOOL NSRegistry::readRecord(/*[in]*/REGOFF offset, 
			    /*[in, out]*/REGDESC* regDesc) 
{
    if (m_file == NULL || offset < 0 || regDesc == NULL)
	return FALSE;	    

    // Seek offset in file
    if (fseek(m_file, offset, SEEK_SET) != 0)
	return FALSE;

    // Determine size of REGDESC structure
    int size = sizeof(REGDESC);

    ::ZeroMemory(regDesc, size);

    // Generate format string to read the exact number of bytes
    char szFormat[256];
    sprintf(szFormat, "%%%dc", size);

    // Read registry record from file
    if (fscanf(m_file, szFormat, regDesc) != EOF)
	return TRUE;
    else 
	return FALSE;
}



//=---------------------------------------------------------------------------=
// NSRegistry::readHeader
//=---------------------------------------------------------------------------=
// Read registry header from Netscape registry. 
//
// parameters :	regHdr	    - [in, out] Registry header
//
// return :	BOOL	    - TRUE if successful
// 
// notes :
//
BOOL NSRegistry::readHeader(/*[in, out]*/REGHDR* regHdr)
{
    if (m_file == NULL || regHdr == NULL)
	return FALSE;

    // Seek offset in file, set to 0
    if (fseek(m_file, 0, SEEK_SET) != 0)
	return FALSE;

    // Determine size of registry header structure
    int size = sizeof(REGHDR);

    ::ZeroMemory(regHdr, size);

    // Generate format string to read the exact number of bytes
    char szFormat[256];
    sprintf(szFormat, "%%%dc", size);

    // Read registry header from file
    if (fscanf(m_file, szFormat, regHdr) == EOF)
	return FALSE;

    // Check if the registry header majic number is correct
    if (regHdr->magic != MAGIC_NUMBER)
	return FALSE;

    return TRUE;
}



//=---------------------------------------------------------------------------=
// NSRegistry::readRootRecord
//=---------------------------------------------------------------------------=
// Read root record from Netscape registry. 
//
// parameters :	regDesc	    - [in, out] Root record
//
// return :	BOOL	    - TRUE if successful
// 
// notes :
//
BOOL NSRegistry::readRootRecord(/*[in, out]*/REGDESC* regDesc) 
{
    if (m_file == NULL || regDesc == NULL)
	return FALSE;	    

    REGHDR regHdr;

    // Read header from file
    if (readHeader(&regHdr) == FALSE)
	return FALSE;

    // From the header, determine the root record offset
    REGOFF offset = regHdr.root;

    // Read root record from offset in file
    return readRecord(offset, regDesc);
}



//=---------------------------------------------------------------------------=
// NSRegistry::readString
//=---------------------------------------------------------------------------=
// Read string from Netscape registry. 
//
// parameters :	offset	    - [in] File offset
//		length	    - [in] Length to read
//		lpszValue   - [in, out] String read from Netscape registry
//
// return :	BOOL	    - TRUE if successful
// 
// notes :
//
BOOL NSRegistry::readString(/*[in]*/REGOFF offset, 
			    /*[in]*/uint32 length, 
			    /*[in, out]*/char* lpszValue) 
{
    if (m_file == NULL || offset < 0 || length < 0 || lpszValue == NULL)
	return FALSE;	    

    // Seek offset in file
    if (fseek(m_file, offset, SEEK_SET) != 0)
	return FALSE;

    // Generate format string to read the exact number of bytes
    char szFormat[256];
    sprintf(szFormat, "%%%dc", length);

    // Read string from file
    if (fscanf(m_file, szFormat, lpszValue) != EOF)
	return TRUE;
    else 
	return FALSE;
}
