/*
 * Copyright (c) 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "COPYRIGHT"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  SUN
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 */

//=--------------------------------------------------------------------------=
// ProxySupport.cpp	by Stanley Man-Kit Ho
//=--------------------------------------------------------------------------=
// Contains implementation of the ProxySupport functions.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "jni.h"
#include "ProxySupport.h"




//=--------------------------------------------------------------------------=
// indexOf
//=--------------------------------------------------------------------------=
// indexOf is a function which takes a string and a substring, and returns the
// first occurance index of the substring in the original string. If the 
// substring is not found, -1 is returned.
//
// parameters :
//	lpszString	[in]	a string to be searched
//      lpszSubstring	[in]	substring to search for
//
// out:
//	int	        [out]   index of the substring inside the original 
//                              string. If substring is not found, -1 is returned.
//
//
int indexOf(const char* lpszString, const char* lpszSubstring)
{
    char* pdest = strstr(lpszString, lpszSubstring);

    if (pdest == NULL)
	return -1;
    else
	return (pdest - lpszString);
}


//=--------------------------------------------------------------------------=
// indexOf2
//=--------------------------------------------------------------------------=
// indexOf2 is a function which takes a string, beginning search index and a 
// substring, and returns the first occurance index of the substring in the 
// original string starting from the given index. If the substring is not 
// found, -1 is returned.
//
// parameters :
//	lpszString	[in]	a string to be searched
//	start		[in]	beginning index of the search in the string
//      lpszSubstring	[in]	substring to search for
//
// out:
//	int	        [out]   index of the substring inside the original 
//                              string. If substring is not found, -1 is returned.
//
//
int indexOf2(const char* lpszString, int start, const char* lpszSubstring)
{
    int index = indexOf(lpszString + start, lpszSubstring);

    if (index != -1)
	index = index + start;

    return index;
}



//=--------------------------------------------------------------------------=
// subString
//=--------------------------------------------------------------------------=
// subString is a function which takes a string, beginning index, ending index,
// and returns the substring in string[begin, end-1].
//
// parameters :
//	lpszString	[in]	a string to be searched
//	start		[in]	beginning index of string for substring extraction.
//	end		[in]	ending index of the string for substring extraction.
//      lpszResult	[out]	substring to return
//
//
//
void subString(const char* lpszString, int start, int end, char* lpszResult)
{
    for (int i=start, j=0; i < end; i++, j++)
	lpszResult[j] = lpszString[i];

    lpszResult[j] = '\0';
}



//=--------------------------------------------------------------------------=
// subString2
//=--------------------------------------------------------------------------=
// subString2 is a function which takes a string, beginning index, and returns 
// the substring in string[begin, end of string].
//
// parameters :
//	lpszString	[in]	a string to be searched
//	start		[in]	beginning index of string for substring extraction.
//      lpszResult	[out]	substring to return
//
//
//
void subString2(const char* lpszString, int start, char* lpszResult)
{
    subString(lpszString, start, strlen(lpszString), lpszResult);
}


//=--------------------------------------------------------------------------=
// Proxy_ExtractHostName
//=--------------------------------------------------------------------------=
// Proxy_ExtractHostName is a function which takes a URL and returns the 
// hostname in the URL.
//
// parameters :
//	lpszURL		[in]	URL for hostname extraction
//	lpszHostName	[out]	hostname to return
//
//
void Proxy_ExtractHostName(const char* lpszURL, char* lpszHostName)
{
    char szBuffer[256];

    lpszHostName[0] = '\0';

    sprintf(szBuffer, lpszURL);

    int index = indexOf(lpszURL, ":");

    if (index == -1)
	return;

    index = index + 3;	// Skip "://"

    // Skip path
    int index2 = indexOf2(lpszURL, index, "/");

    if (index2 != -1)
	szBuffer[index2] = '\0';

    // Skip port number
    index2 = indexOf2(lpszURL, index, ":");

    if (index2 != -1)
	szBuffer[index2] = '\0';

    sprintf(lpszHostName, (szBuffer + index));
}




//=--------------------------------------------------------------------------=
// Proxy_GetCurrentIpAddress
//=--------------------------------------------------------------------------=
// Proxy_GetCurrentIpAddress is a function which takes a JNIEnv pointer and 
// returns the IP address of the local machine.
//
// parameters :
//	JNIEnv*		[in]	JNI pointer
//	lpszIpAddress	[out]	IP address to return.
//
//
void Proxy_GetCurrentIpAddress(JNIEnv* env, char* lpszIpAddress)
{
    jclass clazz = env->FindClass("java/net/InetAddress");

    jmethodID m = env->GetStaticMethodID(clazz, "getLocalHost", "()Ljava/net/InetAddress;");
    
    jobject inetAddress = env->CallStaticObjectMethod(clazz, m);

    jmethodID m2 = env->GetMethodID(clazz, "getHostAddress", "()Ljava/lang/String;");
    
    jstring ipAddress = (jstring) env->CallObjectMethod(inetAddress, m2);

    const char* value = env->GetStringUTFChars((jstring)ipAddress, 0);

    sprintf(lpszIpAddress, value);

    env->ReleaseStringUTFChars((jstring)ipAddress, value);
}



//=--------------------------------------------------------------------------=
// Proxy_NSCanonicalizeProxyOverride
//=--------------------------------------------------------------------------=
// Proxy_NSCanonicalizeProxyOverride is a function which takes a proxy-bypass-list
// which is read directly from the Netscape Navigator preference, and returns 
// the canonicalized version of the proxy-bypass-list which can be used in
// ActivatorProxyHandler in Java.
//
// parameters :
//      lpszProxyOverride [in,out]  a string which contains a list of domain which
//  				    proxy should not be used when a URL is requested.
//
// Notes: i) lpszProxyOverride contains all the domain which no proxy should be used
//        when a connection is made. For example, "a.eng,b.eng,c.eng". In Netscape
//        Navigator preference, no wildcard is allowed in the list. In order to use
//        the proxy-bypass-list in ActivatorProxyHandler in Java, the string must
//        be converted into "*.a.eng,*.b.eng,*.c.eng".
//
//
void Proxy_NSCanonicalizeProxyOverride(char* lpszProxyOverride)
{
    if (lpszProxyOverride == NULL)
	return;

    // Skip all whitespace
    for (int i=0, j=0, len=strlen(lpszProxyOverride); j <= len; j++)
    {
	if (lpszProxyOverride[j] != ' ')
	    lpszProxyOverride[i++] = lpszProxyOverride[j];
    }

    if (lpszProxyOverride[0] == '\0')
	return;

    char szBuffer[1024];
    i = 0;

    // Prepend '*' if the first item is valid.
    //if (lpszProxyOverride[0] == '.')
	szBuffer[i++] = '*';

    for (j=0, len=strlen(lpszProxyOverride); j <= len; j++, i++)
    {
	if (lpszProxyOverride[j] == ',' && lpszProxyOverride[j+1] != '\0')
	{
	    // Append ",*"
	    szBuffer[i] = ',';

	    //if (lpszProxyOverride[j+1] == '.')
		szBuffer[++i] = '*';
	}
	else
	    szBuffer[i] = lpszProxyOverride[j];
    }

    sprintf(lpszProxyOverride, szBuffer);
}





//=--------------------------------------------------------------------------=
// Proxy_IECanonicalizeProxyOverride
//=--------------------------------------------------------------------------=
// Proxy_IECanonicalizeProxyOverride is a function which takes a proxy-bypass-list
// which is read directly from the Internet Explorer preference, and returns 
// the canonicalized version of the proxy-bypass-list which can be used in
// ActivatorProxyHandler in Java.
//
// parameters :
//      lpszProxyOverride [in,out]  a string which contains a list of domain which
//  				    proxy should not be used when a URL is requested.
//
// Notes: i) lpszProxyOverride contains all the domain which no proxy should be used
//        when a connection is made. For example, "a.eng;b.*.eng;c.eng". In order 
//        to use the proxy-bypass-list in ActivatorProxyHandler in Java, all ';'
//        in the string must be converted into ','.
//
//
void Proxy_IECanonicalizeProxyOverride(char* lpszProxyOverride)
{
    if (lpszProxyOverride == NULL)
	return;

    char szBuffer[1024];

    // Skip all whitespace and convert all ';' into ','
    for (int i=0, j=0, len=strlen(lpszProxyOverride); j <= len; j++)
    {
	if (lpszProxyOverride[j] != ' ')
	{
	    if (lpszProxyOverride[j] == ';')
		szBuffer[i++] = ',';
	    else
		szBuffer[i++] = lpszProxyOverride[j];
	}
    }

    sprintf(lpszProxyOverride, szBuffer);
}

