/*
 * @(#)PluginConverter.java	1.24 03/08/01
 *
 * Copyright (c) 1995-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * @version 1.0
 *
 * @author Stanley Man-Kit Ho
 */

package sun.plugin.converter.engine;

import sun.plugin.converter.util.*;
import sun.plugin.converter.ResourceHandler;
import java.io.*;
import java.util.*;
import java.text.DateFormat;
import javax.swing.*;


public class PluginConverter implements Converter, Runnable 
{
    private String templateFilePath = null;
    private static String defaultTemplateFileName = "templates/default.tpl";
    private File logFile = null;
    private boolean createLog = false;
    private boolean processSubDirs = false;
    private boolean stopConversion = false;
    private static ConverterProgressEvent event = null;

    private String cabFileLocation = null;									  
    private String nsFileLocation = null;
    private String smartUpdateLocation = null;
    private String mimeType = null;

    private String version = "1.3.1";
    private String nodotVersion = "131";
    private String cabVersion = "1,3,1,0";
    private String urlString = "http://java.sun.com/products/plugin/" + version;

    private String defaultCabFileLocation = urlString + "/jinstall-" + nodotVersion + "-win32.cab#Version=" + cabVersion;
    private String defaultNsFileLocation = urlString + "/newuser.html";
    private String defaultSmartUpdateLocation = urlString + "/jre-" + nodotVersion + "-win32.jar";
    private String defaultMimeType = "application/x-java-applet;jpi-version=" + version;

    private String userPropsFileName = "converter.props";
    private String defaultPropsFileName = "default.props";
    private String defaultLogFileName = "convertlog.txt";

    private String caption = null;

    private boolean commandLineInvocation = false;

    private static final String propFileHeader = "#  Java Plug-in HTML Converter Properties";

    private ConvertSet convertSet = null;
    //  Properties that will be changable in a bean box or on a command line
    private boolean showProgressStdOut = true;
    private PrintManyPlaces printer = null;


    private OutputStream log = null;
    private PrintStream progress = null;
    private boolean debug = false;
    private Vector listeners = new Vector();

    private Thread runConvert = null;

    /*
     *Constructor used in GUI version
     */
    public PluginConverter() throws FileNotFoundException
    {
	this(false, false);
    }

    /**
      Constructor
      */
    public PluginConverter(boolean logfile, boolean showProgress) throws FileNotFoundException
    {
	setCreateLog(logfile);
	setShowProgressStdOut(showProgress);
	try 
	{
	    setTemplateFilePath(getDefaultTemplateFilePath());
	    if(createLog) 
	    {
		setLogFile(getDefaultLogFile());
	    }
	}
	catch(FileNotFoundException e) 
	{
	    e.printStackTrace();
	    throw e;
        }


	try 
	{
	    // Read default converter properties
	    //
    	    Properties p = new Properties();

    	    InputStream in = getClass().getClassLoader().getResourceAsStream(defaultPropsFileName);
	    p.load(in);
	    in.close();
	    in = null;

	    defaultCabFileLocation = p.getProperty("converter.cab.file.loc", "");
	    defaultNsFileLocation = p.getProperty("converter.plugin.file.loc", "");
	    defaultSmartUpdateLocation = p.getProperty("converter.smartupdate.file.loc", "");
	    defaultMimeType = p.getProperty("converter.template.mimetype", "");

	    // Read user customized converter properties
	    //
	    File f = new File(userPropsFileName);

	    try
	    {
		FileInputStream fis = new FileInputStream(f);

		p = new Properties();
		p.load(fis);
		fis.close();
	    }
	    catch(FileNotFoundException e)
	    {
		// User-defined prop file doesn't exists, so we create a new one

		FileOutputStream fos = new FileOutputStream(f);
		p.store(fos, propFileHeader);
		fos.close();
	    }

	    cabFileLocation = p.getProperty("converter.cab.file.loc", defaultCabFileLocation);
	    nsFileLocation = p.getProperty("converter.plugin.file.loc", defaultNsFileLocation);
	    smartUpdateLocation = p.getProperty("converter.smartupdate.file.loc", defaultSmartUpdateLocation);
	    mimeType = p.getProperty("converter.template.mimetype", defaultMimeType);

	    caption = ResourceHandler.getMessage("caption.warning");
	}
	catch(IOException e)
	{
	    final String info = ResourceHandler.getMessage("plugin_converter.propnotfound2");
	    JOptionPane.showMessageDialog(null, info, caption, JOptionPane.WARNING_MESSAGE);
	}
	finally {
	    if (cabFileLocation == null)
		cabFileLocation = defaultCabFileLocation;
	    if (nsFileLocation == null)
		nsFileLocation = defaultNsFileLocation;
	}
    }

    public void persistConverterSetting() 
    {
	Properties p = new Properties();
	File f = null;
	try
	{
	    f = new File(userPropsFileName);

	    if(nsFileLocation!=null)
		p.put("converter.plugin.file.loc",nsFileLocation);
	    if(cabFileLocation!=null)
		p.put("converter.cab.file.loc",cabFileLocation);
	    if(smartUpdateLocation!=null)
		p.put("converter.smartupdate.file.loc",smartUpdateLocation);
	    if(mimeType!=null)
		p.put("converter.template.mimetype",mimeType);
			    
	    if(f.exists())
	    {
		f.delete();
	    }
	    FileOutputStream fos = new FileOutputStream(f);
	    p.store(fos, propFileHeader);
	    fos.flush();
	    fos.close();
	}
	catch(IOException e)
	{
	    System.out.println(ResourceHandler.getMessage("plugin_converter.saveerror"));
	}
    }
	
    public void setConvertSet(ConvertSet c) { convertSet = c; }
    public ConvertSet getConvertSet() { return convertSet; }

    /*
      Do we show progress in Stdout
      */
    public void setShowProgressStdOut(boolean b) { 
	if(showProgressStdOut != b) {
	    showProgressStdOut = b;
	}
    }
    public boolean isShowProgressStdOut() { return showProgressStdOut; }

    public void setCommandLine(boolean commandLineInvocation) {
	this.commandLineInvocation = commandLineInvocation;
    }
    public boolean getCommandLine() {
	return commandLineInvocation;
    }

    /*
      Progress can be posted out via the ConverterProgressEvent object
      */
    public void addConverterProgressListener(ConverterProgressListener l) {
	listeners.addElement(l);
    }
    public void removedConverterProgressListener(ConverterProgressListener l) {
	listeners.removeElement(l);
    }
    public void fireEvent(ConverterProgressEvent e) {
	Vector list = (Vector)listeners.clone();
	for(int i = 0;i<list.size();i++) {
	    ConverterProgressListener listener = (ConverterProgressListener)list.elementAt(i);
	    listener.converterProgressUpdate(e);
	}

    }

    public void startConversion() {
	if(debug) System.out.println("Debug is on");
	if(runConvert==null) {
	    runConvert = new Thread(this);

	    runConvert.start();
	}
    }
    public void stopConversion() {
	if((runConvert != null) && runConvert.isAlive()) {
			
	    stopConversion = true;
	    if(progress!=null) {
		progress.println(ResourceHandler.getMessage("plugin_converter.cancel"));
	    }
	}
	runConvert = null;

    }
    public void run() 
    {
	try {
	    setPrintOutput();
	}
	catch(IOException e) {
	    System.out.println(ResourceHandler.getMessage("plugin_converter.logerror"));
	}

	Date now = new Date();

	if(progress!=null) progress.println(ResourceHandler.getMessage("plugin_converter.appletconv") 
					    +(DateFormat.getDateInstance(DateFormat.LONG)).format(now)+" "+
					    (DateFormat.getTimeInstance(DateFormat.LONG)).format(now));
	//  Starting to convert a convertSet (a directory - could be a sub directory)
	event = new ConverterProgressEvent(this);

	stopConversion = false;
	try
	{
	    runConversion(convertSet);
	}
	catch (IOException e)
	{
	    e.printStackTrace();
	}

	event.setStatus(ConverterProgressEvent.ALL_DONE);
	fireEvent(event);
	if(progress!=null) 
	    progress.println(ResourceHandler.getMessage("plugin_converter.done") + event.getFilesProcessed()
			     + ResourceHandler.getMessage("plugin_converter.appletfound") 
			     + event.getAppletsFound());

	//
	// Workaround for hang whenever a JOptionPane is shown.
	//

	if ( getCommandLine() ) {
	    System.exit(0);
	}
    }

    private void runConversion(ConvertSet convertSet) 
		 throws IOException
    {
	//  Get the Files in the ConvertSet passed in
	Enumeration files = convertSet.getFilesEnumeration();
	while(files.hasMoreElements() && !stopConversion) 
	{
	    //  Get the current name to operation on, this could be a filename or directory
	    String fileName = (String)files.nextElement();
	    File destFile = new File(convertSet.getDestinationPath(), "temp.conv");
	    File sourceFile = new File(convertSet.getSourcePath(), fileName);

	    //  If this is a directory, and we are supposed to process sub dirs, then we do so.
	    if(sourceFile.isDirectory())
	    {
		if(processSubDirs)
		{
		    ConvertSet subSet = new DefaultSet();
		    boolean filesOnly = ((FlexFilter)convertSet.getFilenameFilter()).isFilesOnly();
		    ((FlexFilter)((DefaultSet)subSet).getFilenameFilter()).setFilesOnly(filesOnly);
		    try
		    {
			subSet.setSourcePath(sourceFile,convertSet.getFilenameFilter());
		    }
		    catch(NotDirectoryException e)
		    {
			e.printStackTrace();
		    }

		    subSet.setDestinationPath(new File(convertSet.getDestinationPath(),fileName));
		    subSet.setBackupPath(new File(convertSet.getBackupPath(),fileName));
		    runConversion(subSet);
		}
		//If sub-directory, skip processing
		continue;
	    }
	    
	    //If the file doesn't exist or doesn't have write permissions, simply return
	    if(sourceFile.exists())
	    {
		if(!sourceFile.canWrite())
		{
		    JOptionPane.showMessageDialog(null, ResourceHandler.getMessage("file_unwritable.info") + sourceFile,
						  caption, JOptionPane.ERROR_MESSAGE );
		    return;						      
		}
	    }
	    else
	    {
		JOptionPane.showMessageDialog(null, ResourceHandler.getMessage("file_notexists.info") + sourceFile,
					      caption, JOptionPane.ERROR_MESSAGE );
		return;
	    }


	    InputStream is = getClass().getClassLoader().getResourceAsStream(getTemplateFilePath());
	    if (is == null)
	    {
		File f = new File(getTemplateFilePath());
		is = new FileInputStream(f);
	    }

	    //  Get a ConverFile object with a template
	    ConvertFile conFiler = new ConvertFile(is);
	    //we need to set the sourcefile so we can guessencoding
	    conFiler.setSource(sourceFile);
	    //we need to guess the encoding so we can us it for countwords
	    conFiler.guessEncoding();
	 if(StdUtils.countWords(sourceFile,"<APPLET", true, conFiler.getEncoding())>0) 

	    {

		if(progress!=null) 
		    progress.print(new File(convertSet.getSourcePath(),fileName)
				   + ResourceHandler.getMessage("plugin_converter.processing"));

		//  Send out progress information
       		event.setSourcePath(convertSet.getSourcePath().getAbsolutePath());
		event.setDestinationPath(convertSet.getDestinationPath().getAbsolutePath());
		event.setBackupPath(convertSet.getBackupPath().getAbsolutePath());
		event.setCurrentFile(fileName);
		event.setStatus(ConverterProgressEvent.CONVERTING);
		fireEvent(event);  //  Fire converting event to preceed processing
		
		conFiler.setSource(sourceFile);
		conFiler.setDestination(destFile);
		conFiler.setCabFileLocation(getCabFileLocation());
		conFiler.setNSFileLocation(getNSFileLocation());
		conFiler.setSmartUpdateLocation(getSmartUpdateLocation());
		conFiler.setMimeType(getMimeType());
		
		try
		{
		    boolean gotApplets = conFiler.convert();  //  Convert file
		    if(!gotApplets)
		    {
			if(progress!=null)
			    progress.print(ResourceHandler.getMessage("plugin_converter.converted"));
			destFile.delete();
		    }
		    else
		    { 
			//Create the backup directory
			if(!StdUtils.createDirs(convertSet.getBackupPath()))
			{
			    if(progress!=null) 
				progress.println(ResourceHandler.getMessage("plugin_converter.nobackuppath"));
			    event.setStatus(ConverterProgressEvent.DEST_DIR_NOT_CREATED);
			    fireEvent(event);

			    JOptionPane.showMessageDialog(null, ResourceHandler.getMessage("file_unwritable.info") 
							    + convertSet.getBackupPath().getAbsolutePath(),
							    caption, JOptionPane.ERROR_MESSAGE );    
			    return;
			}

			System.gc();
			System.runFinalization();

			//
			// Save a backup copy, and update the original with the converted html.
			//

			File backupFile = new File(convertSet.getBackupPath().getAbsolutePath(),fileName);
			File originalFile = new File(destFile.getParent() + File.separator + fileName);

			if ( !updateFiles(backupFile,originalFile,destFile) ) {
			    JOptionPane.showMessageDialog(null,
							  ResourceHandler.getMessage("plugin_converter.failure") +
							  destFile.getParent() + File.separator + fileName, 
							  caption, JOptionPane.ERROR_MESSAGE);
			}
		    }

		    event.setStatus(ConverterProgressEvent.DONE);  //  Done converting file

		    //Update summary information
		    event.setFilesProcessed(event.getFilesProcessed()+1);
		    event.setAppletsFound(event.getAppletsFound()+conFiler.getAppletsFound());
		    event.setErrorsFound(event.getErrorsFound()+conFiler.getErrors());
		    fireEvent(event);

		    if(progress!=null) 
			progress.println(ResourceHandler.getMessage("plugin_converter.donefound") 
					 + conFiler.getAppletsFound());
		}
		catch (Exception e)
		{
		    event.setStatus(ConverterProgressEvent.ERROR);  //  Done converting file

		    //  Update summary information
		    event.setFilesProcessed(event.getFilesProcessed());
		    event.setAppletsFound(event.getAppletsFound()+conFiler.getAppletsFound());
		    event.setErrorsFound(event.getErrorsFound()+1);
		    fireEvent(event);

		    if(progress!=null) 
		    {
			progress.println(ResourceHandler.getMessage("plugin_converter.seetrace"));
			e.printStackTrace(progress);
		    } 
		}							
	    }
	    else
	    {
		if(progress!=null) 
		    progress.println(ResourceHandler.getMessage("plugin_converter.noapplet") +sourceFile);
	    }
	}
    }


    /**
     * Update the backup and original files.
     *
     * @param backupFile The backup file object.
     * @param originalFile The original file object.
     * @param convertedFile The converted file object.
     * @return True if the files were successfully updated, false otherwise.
     */

    public boolean updateFiles(File backupFile, File originalFile, File convertedFile) {

	//
	// Make sure we do not overwrite the backup copy without permission.
	// First we move the original source to the backup copy.  Then
	// we rename the converted file to the same name as the original.
	//

	return ( backupFileIsWritable(backupFile,originalFile) && 
		 originalFile.renameTo(backupFile) &&
		 convertedFile.renameTo(originalFile) );
    }


    /**
     * Return true if the backup file is available for writing.
     *
     * @param backupFile The backupFile object.
     * @return True if the backup file is available for writing.
     */

    public boolean backupFileIsWritable(File backupFile, File originalFile) {
	if ( backupFile.exists() ) {

	    //
	    // A backup copy already exists.  Let the user make a choice to overwrite.
	    //
	    	    
	    if ( JOptionPane.showConfirmDialog( null, 
						ResourceHandler.getMessage("plugin_converter.overwrite1") +
						originalFile.getAbsolutePath() +
						ResourceHandler.getMessage("plugin_converter.overwrite2"), 
						caption,
						JOptionPane.YES_NO_OPTION ) == JOptionPane.YES_OPTION) {
		
		//
		// The user said it's ok to overwrite.
		//
		
		backupFile.delete();
	    }
	    else {
		return false;
	    }
	}
	return true;
    }

  
    public static String getDefaultTemplateFileName() 
    {
	return defaultTemplateFileName;
    }
    
    
    /*
      Name of template file to be used
      */
    public void setTemplateFilePath(String fileName) throws FileNotFoundException 
    {
	try
	{
    	    // First, try to load it from JAR file directly
	    InputStream is = getClass().getClassLoader().getResourceAsStream(fileName);
	    
	    if (is != null)
	    {
		templateFilePath = fileName;
		return;
	    }

	    // Second, try to load it locally
	    File f = new File(fileName);

	    if (f.isFile()) 
	    {
		templateFilePath = fileName;
		return;
	    }
	}
	catch (Throwable e)
	{
	    e.printStackTrace();
	}

	throw new FileNotFoundException(fileName);
    }

    public String getTemplateFilePath() { return templateFilePath; }
    public String getDefaultTemplateFilePath() throws FileNotFoundException
    {
	// Look for the template file from classloader
	ClassLoader loader = getClass().getClassLoader();

	if (loader == null) 
	    throw new FileNotFoundException(defaultTemplateFileName);

	InputStream is = loader.getResourceAsStream(defaultTemplateFileName);

	if (is == null)
	    throw new FileNotFoundException(defaultTemplateFileName);
	else
	    return defaultTemplateFileName;
    }

    /*
      Stream to direct log information
      */
    public void setLogFile(java.io.File f) throws IllegalArgumentException
    {
	if(f.exists())
	{
	    if(progress!=null) 
	    progress.println(ResourceHandler.getMessage("plugin_converter.writelog"));
	}
	else
	{
	    if(f.getParent() == null)
	    {
		f = new File(System.getProperty("user.dir"),f.getName());
	    }
	    else
	    {
		File dirs = new File(f.getParent());
		if(!dirs.exists())
		dirs.mkdirs();
	    }
	}		
	
	logFile = f;
    }

    public java.io.File getLogFile() { return logFile; }


    public java.io.File getDefaultLogFile() {
	return new File(System.getProperty("user.dir"),defaultLogFileName);
    }
	
    public void setPrintOutput() throws IOException {
	if(printer!=null) printer.closeAll();
	printer = new PrintManyPlaces();
	if(showProgressStdOut || debug) {
	    printer.addPlace(System.out);
	}
		
	if(createLog) {	    	  
	    printer.addPlace(new FileOutputStream(getLogFile()));
	}
			
	if(printer.countPlaces()>0) {
	    progress = new PrintStream(printer);
	}
	else {
	    progress = null;
	}
    }

    /*
      Do we create a log?
      */
    public void setCreateLog(boolean b) { 
	if(b != createLog) {
	    createLog = b; 
	}
    }

    public boolean isCreateLog() { return createLog; }

    /*
      Do we process into subdirectories
      */
    public void setRecurse(boolean b) { 
	((FlexFilter)convertSet.getFilenameFilter()).setFilesOnly(!b);
	processSubDirs = b;
    }
    public boolean isRecurse() { return processSubDirs; }

    public String getCabFileLocation()  {
	return cabFileLocation;
    }
    public String getDefaultCabFileLocation()  {
	return defaultCabFileLocation;
    }
    public void setCabFileLocation(String location)   {
	cabFileLocation = location;
    }

    public String getNSFileLocation()  {
	return nsFileLocation;
    }
    public String getDefaultNSFileLocation()  {
	return defaultNsFileLocation;
    }
    public void setNSFileLocation(String location)   {
	nsFileLocation = location;
    }

    public String getSmartUpdateLocation()  {
	return smartUpdateLocation;
    }
    public String getDefaultSmartUpdateLocation() {
	return defaultSmartUpdateLocation;
    }
    public void setSmartUpdateLocation(String location)  {
	smartUpdateLocation = location;
    }


    public String getMimeType()  {
	return mimeType;
    }
    public String getDefaultMimeType()  {
	return defaultMimeType;
    }
    public void setMimeType(String type)  {
	mimeType = type;
    }


    public String toString() {
	String str = "";

	str += "\nShow Progress StdOut:  "+showProgressStdOut;
	str += "\ndebug:  "+debug;
	str += "\nTemplate File Path:  "+templateFilePath.toString();
	str += "\nCount Listeners:  "+listeners.size();
	str += "\nLog File Path:  "+logFile.toString();
	str += "\nCreate Log:  "+createLog;
	str += "\nConvert Set:  "+convertSet.toString();
	str += "\nProcess Sub Dirs:  "+processSubDirs;

	return str;
    }
}
