/*
 * @(#)TransientNameServer.java	1.24 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.corba.se.internal.CosNaming;

import java.util.Properties;
import java.net.InetAddress;

import org.omg.CORBA.ORB;

import org.omg.CosNaming.NamingContext;

import com.sun.corba.se.internal.CosNaming.TransientNameService;
import com.sun.corba.se.internal.CosNaming.BootstrapServer;

/**
 * Class TransientNameServer is a standalone application which
 * implements a transient name service. It uses the TransientNameService
 * class for the name service implementation, and the BootstrapServer
 * for implementing bootstrapping, i.e., to get the initial NamingContext.
 * <p>
 * The BootstrapServer uses a Properties object specify the initial service
 * object references supported; such as Properties object is created containing
 * only a "NameService" entry together with the stringified object reference
 * for the initial NamingContext. The BootstrapServer's listening port
 * is set by first checking the supplied arguments to the name server
 * (-ORBInitialPort), and if not set, defaults to the standard port number.
 * The BootstrapServer is created supplying the Properties object, using no
 * external File object for storage, and the derived initial port number.
 * @see TransientNameService
 * @see BootstrapServer
 */
public class TransientNameServer
{
    static private boolean debug = false ;

    static public void trace( String msg )
    {
	if (debug)
	    System.out.println( msg ) ;
    }

    static public void initDebug( String[] args ) 
    {
	// If debug was compiled to be true for testing purposes,
	// don't change it.
	if (debug)
	    return ;
	    
	for (int ctr=0; ctr<args.length; ctr++)
	    if (args[ctr].equalsIgnoreCase( "-debug" )) {
		debug = true ;
		return ;
	    }

	debug = false ;
    }

    /**
     * Main startup routine. It instantiates a TransientNameService
     * object and a BootstrapServer object, and then allows invocations to
     * happen.
     * @param args an array of strings representing the startup arguments.
     */ 
    public static void main(String args[])
    {
	initDebug( args ) ;

	try {
	    trace( "Transient name server started with args " + args ) ;

	    // Create an ORB object
	    Properties props = System.getProperties() ;

	    // check if PortableServer package is available, then set property to POA ORB
	    try {
	        Class t = Class.forName("com.sun.corba.se.internal.POA.POAORB");

	        // found the PortableServer class, therefore set the property to POAORB
	        props.put( "org.omg.CORBA.ORBClass", "com.sun.corba.se.internal.POA.POAORB" ) ;
	    } catch(java.lang.ClassNotFoundException e) {
	        // PortableServer package is not available, therefore set the property to 
	        // com.sun.corba.se.internal.iiop.ORB
	        props.put( "org.omg.CORBA.ORBClass", "com.sun.corba.se.internal.iiop.ORB" ) ;
            }

	    org.omg.CORBA.ORB corb = ORB.init( args, props ) ;
	    trace( "ORB object returned from init: " + corb ) ;
	    com.sun.corba.se.internal.iiop.ORB orb = (com.sun.corba.se.internal.iiop.ORB) corb ;

	    // Create a name service
	    TransientNameService nameService = new TransientNameService(orb);
	    trace( "name service created" ) ;

	    // Get its object reference
	    NamingContext initialRef = nameService.initialNamingContext();

	    // Stringify
	    String stringifiedIOR = orb.object_to_string(initialRef) ;

	    // Create a properties object with the initial references
	    // we know about
	    Properties initialProps = new Properties();
	    initialProps.put("NameService",stringifiedIOR);

	    // Determine the initial bootstrap port to use
	    int initialPort = 900; // by convention
	    try {
		// Try environment
		String ips = System.getProperty("org.omg.CORBA.ORBInitialPort");
		if (ips != null && ips.length() > 0)
		    initialPort = java.lang.Integer.parseInt(ips);
	    } catch (java.lang.NumberFormatException e) {
		// do nothing
	    }
		
	    // Let arguments override
	    for (int i=0;i<args.length;i++) {
		// Was the initial port specified?
		if (args[i].equals("-ORBInitialPort") &&
		    i < args.length-1) {
		    initialPort = java.lang.Integer.parseInt(args[i+1]);
		}
	    }
		
	    // Create a bootstrap server object with no associated file
	    // and the properties object just created
	    BootstrapServer initialBoot = new BootstrapServer(
							      orb, initialPort, null, initialProps);
	    trace( "bootstrap server created on port " + initialPort ) ;

	    try {
		// Start the bootstrap server
		initialBoot.start();
	    } catch (org.omg.CORBA.SystemException e) {
		NamingUtils.errprint("TransientNameServer: caught an exception " +
				     "while starting the bootstrap service on port " + initialPort + ".");
		NamingUtils.errprint("TransientNameServer: try using a different " +
				     "port with commandline arguments -ORBInitialPort <portno>");
		return;
	    }

	    // Print status...

	    // This is used for handshaking by the IBM test framework!
	    // Do not modify, unless another synchronization protocol is 
	    // used to replace this hack!
	    System.out.println("Initial Naming Context:\n" + stringifiedIOR) ;
	    System.out.println("TransientNameServer: setting port for " +
			       "initial object references to: " + initialPort);
	    System.out.println("Ready.");

	    // Serve objects.
	    java.lang.Object sync = new java.lang.Object();
	    synchronized (sync) {sync.wait();}

	} catch (Exception e) {
	    NamingUtils.errprint("TransientNameServer: Caught exception: ");
	    NamingUtils.printException(e);
	}
    }

    /**
     * Private constructor since no object of this type should be instantiated.
     */ 
    private TransientNameServer() {}
}
