/*
 * @(#)ConnectionTable.java	1.70 00/02/11
 *
 * Copyright 1995-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.corba.se.internal.iiop;

import java.io.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.BindException;
import java.net.ConnectException;
import java.net.NoRouteToHostException;

import org.omg.CORBA.COMM_FAILURE;
import org.omg.CORBA.CompletionStatus;

import com.sun.corba.se.internal.core.EndPoint;
import com.sun.corba.se.internal.core.ServerGIOP;
import com.sun.corba.se.internal.util.MinorCodes;
import com.sun.corba.se.internal.util.Utility;

/**
 * One instance of this class is created per ORB object. 
 */
public class ConnectionTable {
    protected ORB orb;
    protected Hashtable connectionCache = new Hashtable();

    protected long globalCounter=0;

    private int MAX_SOCKET_RETRIES=5;

    protected ServerGIOP server;

    public ConnectionTable(ORB orb, ServerGIOP server) {
	this.orb = orb;
	this.server = server;
    }

    private void dprint(String msg) {
        Utility.dprint(this, msg);
    }


    /**
     * Get a connection from the connection table, or create a new one.
     * Called only on client side.
     * @param key The Endpoint of the connection.
     * @return The connection, pre-existing or new.
     */
    public Connection get(EndPoint key)
    {
	if (orb.transportDebugFlag)
            dprint( "Client get called: host = " + key.getHostName() +
                    " port = " + key.getPort() ) ;

        // Try to get a connection from the cache.
	// note: Hashtable.get() is synchronized
        Connection c = (Connection) connectionCache.get(key);
        if (c != null) {
	    if (orb.transportDebugFlag)
		dprint( "Returning connection " + c + " from table" ) ;
            return c;
	}

        // Create a new connection object. Even though Hashtable.get/put are
        // synchronized, we need to synchronize on "this" to keep the
        // get+put atomic and prevent two threads from creating connections
        // to the same destination at the same time.
        synchronized(this) {
            // Always recheck the entry condition after getting the lock !!
            c = (Connection) connectionCache.get(key);
            if (c != null) {
		if (orb.transportDebugFlag)
		    dprint( "Returning connection " + c + " from table" ) ;
                return c;
	    }

            // Create an entry in the connection table
            c = new IIOPConnection(orb, server, this, key);
            stampTime(c);
            connectionCache.put(key, c);

	    if (orb.transportDebugFlag)
		dprint( "Creating new connection " + c ) ;
        }

        // Try to open a socket to the server
        int tries = 0;
        while (true)
	    {
		try {
                    Socket socket = Connection.newSocket(orb, key);
		    // set socket to disable Nagle's algorithm (always send immediately)
		    try {
			socket.setTcpNoDelay(true);
		    } catch (Exception e) {
		    }
		    c.setConnection(socket, this);
		    checkConnectionTable();
		    break;
		}
		catch (SocketException ex) {
		    if (orb.transportDebugFlag)
			dprint( "SocketException " + ex + 
				" while creating socket for new connection" ) ;

		    if ( ex instanceof BindException
			 || ex instanceof ConnectException
			 || ex instanceof NoRouteToHostException )
			{
			    if (orb.transportDebugFlag)
				dprint( "Serious error: aborting connection" ) ;

			    c.abortConnection();
			    throw new COMM_FAILURE(MinorCodes.CONNECT_FAILURE,
						   CompletionStatus.COMPLETED_NO);
			}

		    if (orb.transportDebugFlag)
			dprint( "Attempting resource cleanup and retry on socket creation" ) ;

		    // Probably the system's file descriptor limit has been reached
		    // Try cleaning up some connections and retry.
		    if ((tries == MAX_SOCKET_RETRIES) || (!cleanUp()))
			{
			    if (orb.transportDebugFlag)
				dprint( "Out of resources: aborting connection" ) ;

			    c.abortConnection();
			    throw new COMM_FAILURE(MinorCodes.CONNECT_FAILURE,
						   CompletionStatus.COMPLETED_NO);
			}
		    tries++;
		}
		catch ( Exception ex ) {
		    if (orb.transportDebugFlag)
			dprint( "Exception " + ex + 
				" while creating socket for new connection: aborting connection" ) ;
		    c.abortConnection();
		    throw new COMM_FAILURE(MinorCodes.CONNECT_FAILURE,
					   CompletionStatus.COMPLETED_NO);
		}
	    }

	if (orb.transportDebugFlag)
	    dprint( "Succesfully created socket for new connection" ) ;

        return c;
    }

    /**
     * Called only on server side when a connection has been accepted.
     * Always creates a new Connection instance, which starts the upcall.
     * @param sock The socket for the connection.
     */
    public synchronized Connection get(Socket sock)
    {
        try {
	    if (orb.transportDebugFlag)
		dprint( "Server get: sockect " + sock ) ;

	    InputStream in = null;
	    OutputStream out = null;
	    try {
		in = sock.getInputStream();
		out = sock.getOutputStream();
	    } catch ( Exception ex ) {
		throw new COMM_FAILURE(MinorCodes.CONNECT_FAILURE,
				       CompletionStatus.COMPLETED_NO);
	    }

	    String host = sock.getInetAddress().getHostName();
	    int port = sock.getPort();
            EndPoint key = new EndPointImpl(EndPoint.CLEAR_TEXT, port, host);

	    if (orb.transportDebugFlag) 
		dprint( "host = " + host + " port = " + port ) ;
	    
	    // Create a new Connection instance, and start the upcall.
	    // We explicitly give the input/output streams to IIOPConnection
	    // so that there is the flexibility of substituting the streams
	    // e.g. for handling multiple protocols later.
	    Connection c = new IIOPConnection(orb, server, key, sock, 
					      in, out, this);

	    connectionCache.put(key, c);

	    stampTime(c);

	    if (orb.transportDebugFlag) 
		dprint( "Created connection " + c ) ;
	    return c;

        } catch ( Exception ex ) {
	    if (orb.transportDebugFlag) 
		dprint( "Exception " + ex + " on creating connection" ) ;

	    // We should not throw an exception back to the ListenerThread.
	    // Just close the socket and return silently.
	    try {
		sock.close();
	    } catch ( Exception ex2 ) {}
	    return null;
        }
    }


    /**
     * Delete a connection from the connection table.
     * @param key The Endpoint of the connection.
     * @see Connection
     */
    public synchronized void deleteConn(EndPoint key) {
	if (orb.transportDebugFlag) 
	    dprint( "DeleteConn called: host = " + key.getHostName() +
                    " port = " + key.getPort() ) ;

	connectionCache.remove(key);
    }

    /**
     * CleanUp by discarding least recently used Connections that
     * are not busy
     */
    private static final int LOW_WATER_MARK = 100;
    // On Solaris there seems to be a max of 256 connections per process.
    private static final int HIGH_WATER_MARK = 240;
    // How many connections to clean at a time.
    private static final int NCLEAN = 5;

    public boolean cleanUp()
    {
	if (orb.transportDebugFlag) 
	    dprint( "Cleanup called" ) ;

	if (connectionCache.size() < LOW_WATER_MARK) {
	    if (orb.transportDebugFlag) 
		dprint( "Cleanup returns false: not enough connections open to start cleanup" ) ;
	    return false;
	}

	for ( int i=0; i<NCLEAN; i++ ) {
	    Connection toClean = null;
	    long lru = java.lang.Long.MAX_VALUE;
	    Enumeration e = connectionCache.elements();

	    // Find least recently used and not busy connection in cache
	    while ( e.hasMoreElements() )
		{
		    Connection c = (Connection) e.nextElement();
		    if ( !c.isBusy() && c.timeStamp < lru )
			{
			    toClean = c; 
			    lru = c.timeStamp;
			}
		}

	    if ( toClean == null ) {
		if (orb.transportDebugFlag) 
		    dprint( "Cleanup returns false: all connections busy" ) ;
		return false;
	    }

	    // Clean connection
	    try {
		if (orb.transportDebugFlag) 
		    dprint( "Cleanup is cleaning connection " + toClean ) ;
		toClean.cleanUp();
	    } catch (Exception ex) {}
	}

	// XXX is necessary to do a GC to reclaim closed network connections ??
        // java.lang.System.gc();

	return true;
    }

    /**
     * Check if number of connections has exceeded high watermark
     */
    public void checkConnectionTable()
    {
	// note: Hashtable.size() is not synchronized
        if (connectionCache.size() > HIGH_WATER_MARK)
            cleanUp();
    }

    // Need to worry about wrap around some day
    public synchronized void stampTime(Connection c)
    {
	// _REVISIT_ Need to worry about wrap around some day
        c.timeStamp = globalCounter++;
    }


    public synchronized void print()
    {
        System.out.println("***ConnectionTable***");
	int siz = connectionCache.size();
        System.out.println("  SIZE=" + siz);
	if ( siz < 10 ) {
            Enumeration e = connectionCache.elements(); 
            while ( e.hasMoreElements() )
		{
		    Connection c = (Connection) e.nextElement();
		    c.print();
		}
	}
    }
 
}
