/*
 * @(#)IIOPConnection.java	1.89 01/02/09
 *
 * Copyright 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

/*
 * Licensed Materials - Property of IBM
 * RMI-IIOP v1.0
 * Copyright IBM Corp. 1998 1999  All Rights Reserved
 * 
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */

package com.sun.corba.se.internal.iiop;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.InetAddress;
import java.net.SocketException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.COMM_FAILURE;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.OBJECT_NOT_EXIST;
import org.omg.CORBA.MARSHAL;
import org.omg.CORBA.Object;
import org.omg.CORBA.UNKNOWN;
import com.sun.org.omg.CORBA.ValueDefPackage.FullValueDescription;

import com.sun.corba.se.internal.core.ServerGIOP;
import com.sun.corba.se.internal.core.EndPoint;
import com.sun.corba.se.internal.core.RequestHandler;
import com.sun.corba.se.internal.core.IOR;
import com.sun.corba.se.internal.core.ServerRequest;
import com.sun.corba.se.internal.core.MarshalOutputStream;
import com.sun.corba.se.internal.util.MinorCodes;
import com.sun.corba.se.internal.util.Utility;
import com.sun.corba.se.internal.orbutil.Condition;
import com.sun.corba.se.internal.orbutil.Lock;

import com.sun.org.omg.SendingContext.CodeBase;
import com.sun.org.omg.SendingContext.CodeBaseHelper; //d11638

import java.security.AccessController;
import java.security.PrivilegedAction;

final class OutCallDesc
{
    Condition done = new Condition();
    Thread thd;
    SystemException exc;
    IIOPInputStream s;
}

/**
 * A thread class to handle reading incoming messages off the wire
 */
final class ReaderThread extends Thread
{
    private boolean debug ;

    private void dprint( String msg ) 
    {
	Utility.dprint( this, msg ) ;
    }

    private IIOPConnection c;

    public ReaderThread(ThreadGroup g, IIOPConnection c, String name, boolean debug )
    {
	super(g, name);
        this.c = c;
    }

    public ReaderThread(IIOPConnection c, String name, boolean debug )
    {
	super(name);
        this.c = c;
    }

    public IIOPConnection getCurrentConnection(){
        return c;
    }

    /**
     */
    public void run() {
        for (;;) {
            try {
		// Create a new IIOPInputStream buffer to hold each incoming
		// IIOP message, capture the entire message, and process it.

		IIOPInputStream s = null;

                try {
		    // creates and prefills the mb from wire
		    s = c.createInputStream();
                } catch (IOException ex) {
		    if (debug) {
			dprint( "IOException in createInputStream: " + ex ) ;
			ex.printStackTrace() ;
		    }

		    // Close the connection, inform all threads
		    // and stop this thread.
                    c.purge_calls(Connection.CONN_ABORT, true, false);
                }

		c.processInput(s);
            } catch (DeleteConn dc) {
		if (debug) {
		    dprint( "DeleteConn thrown while reading request: " + dc ) ;
		    dc.printStackTrace() ;
		}

		// Close the connection, inform all threads
		// and stop this thread.
                c.purge_calls (dc.minorCode, true, false);
            } catch (ThreadDeath td) {
		if (debug) {
		    dprint( "ThreadDeath thrown while reading request: " + td ) ;
		    td.printStackTrace() ;
		}

                try {
		    // Close the connection and inform all threads.
                    c.purge_calls(Connection.CONN_ABORT, false, false);
                } finally {
                    throw td;
                }
            } catch (Exception ex) {
		if (debug) {
		    dprint( "Exception thrown while reading request: " + ex ) ;
		    ex.printStackTrace() ;
		}

		break;
            }
        }
    }
}


/** This is a generic worker thread that dispatches a GIOP message
    to the ORB and sends a reply back.
    It may be used by multiple protocols (e.g. IIOP and HTTP)
 */

final class Work {
    private RequestHandler handler;
    private IIOPConnection conn;
    private IIOPInputStream request;
    Work(RequestHandler h, IIOPConnection c, IIOPInputStream is) {
        handler = h;
        conn = c;
        request = is;
    }
    RequestHandler getRequestHandler() { return handler; }
    IIOPConnection getConnection() { return conn; }
    IIOPInputStream getIIOPInputStream() { return request; }
}

final class WorkQueue {

    private static Vector workToDo = new Vector();
    private static int workerCount = 0;

    synchronized void addWorkToDo(Work w)
    {
        workToDo.addElement(w);
        if (workerCount == 0) {
            //
            // Create a worker thread and start the upcall.
            //
	    CachedWorkerThread wt = new CachedWorkerThread();

	    // Catch exceptions since setDaemon can cause a
	    // security exception to be thrown under netscape
	    // in the Applet mode
	    try {
		wt.setDaemon(true);
	    }
	    catch (Exception e) {}

	    wt.start();
        } else
            notify();
    }

    synchronized Work getWorkToDo(Thread t)
    {
        workerCount++;
        while (workToDo.isEmpty()) {
            try {
                wait();
            } catch (InterruptedException e) {
            }
        }
        workerCount--;
        Work w = (Work) workToDo.firstElement();
        workToDo.removeElementAt(0);

        return w;
    }
}

final class CachedWorkerThread extends Thread {

    private static WorkQueue workQueue = new WorkQueue();
    private Work currentWork;
    public CachedWorkerThread() {
    }

    public void run()
    {
        for (;;) {
            currentWork = workQueue.getWorkToDo(this);
            doWork(currentWork.getRequestHandler(), currentWork.getConnection(), currentWork.getIIOPInputStream());
        }
    }

    public IIOPConnection getCurrentConnection(){
	return currentWork.getConnection();
    }

    static void addWorkToDo(RequestHandler h, IIOPConnection c, IIOPInputStream is)
    {
        Work w = new Work(h, c, is);
        workQueue.addWorkToDo(w);
    }

    void doWork(RequestHandler handler, Connection conn, IIOPInputStream request)
    {
        IIOPOutputStream response = null;

	Message msg = request.getMessage();

	ORB orb = conn.getORB() ;

        conn.requestBegins();

	switch ( msg.getType() ) {
        case Message.GIOPRequest: {
	    ServerRequest srequest = (ServerRequest)request;
	    try {
	        // Here is the actual dispatch to the ORB which
	        // dispatches to the subcontract which dispatches to the
		// skeleton which finally dispatches to the user's servant.

		response = (IIOPOutputStream)handler.process(srequest);

		if (srequest.isOneWay()) {
                    conn.requestEnds();
	            Thread.currentThread().stop();
	        }
            } catch (SystemException ex) {
	        // Caught a system exception marshal it back to the client
	        //
	        // ex.printStackTrace();
                // conn.dprint("run: got exception: " + ex);
                try {
		    response = (IIOPOutputStream)
			srequest.createSystemExceptionResponse(ex, null);
                } catch (Exception e) {
                    conn.requestEnds();
	            return; // marshaling exception error; return;
                }

	    } catch (ThreadDeath ex) {
		throw ex;

	    } catch (Throwable ex) {
	        //
	        // Caught an unknown Exception or Error, create a
		// SystemException and marshal it back to the client
	        //

                // conn.dprint("run: got exception: " + ex);
                // ex.printStackTrace();
                SystemException exc = new UNKNOWN(MinorCodes.UNKNOWN_SERVER_ERROR,
						  CompletionStatus.COMPLETED_MAYBE);
                try {
		    response = (IIOPOutputStream)
			srequest.createSystemExceptionResponse(exc, null);
                } catch (Exception e) {
                    conn.requestEnds();
                    return; // marshaling exception error; return;
                }
            }
	    break;
	}
        case Message.GIOPLocateRequest: {
	    LocateRequestMessage lmsg = (LocateRequestMessage)msg;
	    LocateReplyMessage reply;
	    int requestId = msg.getRequestId();

	    IOR ior = null;
	    try {
		ior = handler.locate(lmsg.getObjectKey());

		if ( ior == null ) {
		    reply = new LocateReplyMessage(requestId,
						   Message.GIOPLocateReplyObjectHere, 
						   orb.transportDebugFlag);
		}
		else {
		    reply = new LocateReplyMessage(requestId,
						   Message.GIOPLocateReplyObjectForward,
						   orb.transportDebugFlag );
	 	}
	    }
	    catch ( Exception ex ) {
		// This handles OBJECT_NOT_EXIST exceptions thrown in
		// the subcontract or obj manager. Send back UNKNOWN_OBJECT.
		reply = new LocateReplyMessage(requestId,
					       Message.GIOPLocateReplyUnknownObject,
					       orb.transportDebugFlag);
	    }
	    response = new IIOPOutputStream(request.getConnection().getORB(),
				    	    request.getConnection());
	    reply.write(response);
	    response.setMessage(reply);
	    if (ior != null)
		ior.write(response);
	    break;
	}
        case Message.GIOPCancelRequest: {
            // Do nothing for now (actually legal IIOP).
            // Just return so the WorkerThread ends.
            conn.requestEnds();
            return;
        }

        default:
            //
            // Any other message type is unexpected and is a fatal
            // error for this connection ??
            //
            conn.requestEnds();
	    return;

	} // end switch (msgType)


	// Send the reply on the wire
	// Only RequestMessages and LocateRequestMessages have replies
        try {
	    conn.sendReply(response);
            conn.requestEnds();
        }
        catch(Exception e) {
            conn.requestEnds();
	    return;
        }
    }
}


final class DeleteConn extends java.lang.Throwable {
    int minorCode;

    DeleteConn (int code) {
        minorCode = code;
    }
}




/**
 * A network connection which processes IIOP messages.
 */
public final class IIOPConnection extends Connection {
    //
    // Connection status
    //
    private static final int OPENING = 1;
    private static final int ESTABLISHED = 2;
    private static final int CLOSE_SENT = 3;
    private static final int CLOSE_RECVD = 4;
    private static final int ABORT = 5;

    //
    // Table of pending invocations on this connection indexed by
    // Integer(requestId).
    //
    private Hashtable out_calls = new Hashtable();

    //
    // Remote address that we're talking to.
    //
    private String threadName;
    protected EndPoint endpoint;

    protected int requestCount = 0;

    private ServerGIOP server;

    // Cache data used by FVDCodeBaseDelegate
    private Hashtable
	implementations = null,
	fvds = null,
	bases = null;

    // CodeBase reference - may be null for late binding
    private CodeBase codeBaseRef;

    // This is a flag associated per connection telling us if the initial set of
    // sending contexts were sent to the receiver already...
    private boolean postInitialContexts = false;

    // Remote reference to CodeBase server (supplies FullValueDescription, among other things)
    private IOR codeBaseServerIOR = null;

    private String getStateString( int state ) 
    {
	switch (state) {
	case OPENING : return "OPENING" ;
	case ESTABLISHED : return "ESTABLISHED" ;
	case CLOSE_SENT : return "CLOSE_SENT" ;
	case CLOSE_RECVD : return "CLOSE_RECVD" ;
	case ABORT : return "ABORT" ;
	default : return "???" ;
	}
    }

    public String toString()
    {
	return "Connection[remote_host=" + endpoint.getHostName() +
            " remote_port=" + endpoint.getPort() +
	    " state=" + getStateString( state ) + "]" ;
    }

    //
    // Various connection state.
    //
    Thread reader;
    int state;
    private Condition event;
    private Lock lock;
    private boolean writeLocked;

    // These I/O streams are the ONLY ONES that should be used.
    // i.e. Do not directly use the input/output streams that socket gives.
    // This restriction is to allow connections to service multiple
    // protocols (e.g. http for tunneling), which may require the
    // socket's streams to be wrapped in other streams.
    InputStream inputStream;
    OutputStream outputStream;

    /**
     * Called after client creates a connection to server
     * or after server accepts an incoming connection.
     * @param host The remote host pointed to by this connection.
     * @param port The remote port used by this connection.
     */
    public IIOPConnection(ORB orb, ServerGIOP server,
                          ConnectionTable ctab, EndPoint ep)
    {
	this.orb = orb;
	this.server = server;
	this.connectionTable = ctab;
        this.endpoint = ep;
        String host = endpoint.getHostName();
        int port = endpoint.getPort();
        threadName = "JavaIDL Reader for " + host + ":" + port;

        final ThreadGroup finalThreadGroup = orb.threadGroup;
        final String finalThreadName = threadName;
        final IIOPConnection finalThis = this;
        final boolean finalTransportDebugFlag = orb.transportDebugFlag;
        try {
            AccessController.doPrivileged(new PrivilegedAction() {
                public java.lang.Object run() {
                    reader = new ReaderThread(finalThreadGroup, finalThis, finalThreadName, finalTransportDebugFlag);
                    return null;
                }
            });
        } catch (SecurityException e) {
            //
            // For some reason we're not allowed to create a new thread
            // in the same thread group that the ORB was initialized in.
            // Fall back on creating the thread in the calling thread's
            // group.
            //
            AccessController.doPrivileged(new PrivilegedAction() {
                public java.lang.Object run() {
                    reader = new ReaderThread(finalThis, finalThreadName, finalTransportDebugFlag);
                    return null;
                }
            });
        }

	state = OPENING;
	event = new Condition();
	lock = new Lock();
    }

    /**
     * Called only from ConnectionTable.get() after server accepts an
     * incoming connection.
     * @param sock The socket for this connection.
     * @param inputStream The inputstream to use. It may be different
     from a socket's inputstream.
     * @param outputStream The outputstream to use. It may be different
     from a socket's outputstream.
    */
    public IIOPConnection(ORB orb, ServerGIOP server,
                          EndPoint ep, Socket sock, 
			  InputStream inputStream, OutputStream outputStream,
                          ConnectionTable ct)
{
    this(orb, server, ct, ep); 

    this.socket = sock;
    this.inputStream = inputStream;
    this.outputStream = outputStream;
    this.connectionTable = ct;

    isServer = true;

    state = ESTABLISHED;

    // Catch exceptions since setDaemon can cause a
    // security exception to be thrown under netscape
    // in the Applet mode
    try {
        AccessController.doPrivileged(new PrivilegedAction() {
            public java.lang.Object run() {
                reader.setDaemon(true);
                return null;
            }
        });
    } catch (Exception e) {}
    reader.start();
}


public synchronized boolean isPostInitialContexts(){
    return postInitialContexts;
}

// Can never be unset...
public synchronized void setPostInitialContexts(){
    postInitialContexts = true;
}

public synchronized void setCodeBaseRef(IOR ior){
    codeBaseServerIOR = ior;
}

public synchronized IOR getCodeBaseRef(){
    return codeBaseServerIOR;
}

public static IIOPConnection getCurrentReaderConnection(){
    try {
	return ((com.sun.corba.se.internal.iiop.CachedWorkerThread)Thread.currentThread()).getCurrentConnection();
    }
    catch(ClassCastException cce){
	return ((ReaderThread)Thread.currentThread()).getCurrentConnection();
    }
}

public java.io.InputStream getInputStream() {
    return inputStream;
}

public ServerGIOP getServerGIOP() {
    return server;
}

String getHost() {
    return endpoint.getHostName();
}

int getPort() {
    return endpoint.getPort();
}

EndPoint getEndpoint() {
    return endpoint;
}

/**
 * Read in the IIOP message from the network's InputStream and
 * create an IIOPInputStream object.
 * Called from ReaderThread only.
 */
public IIOPInputStream createInputStream() throws Exception
{
    byte[] buf = new byte[1024];

    // Read in the message header and create the appropriate type
    // of IIOP message.
    Message msg = Message.createFromStream(orb, buf, inputStream);

    // Create the stream object and read in the entire message.
    switch (msg.getType()) {
    case Message.GIOPRequest:
	if (orb.transportDebugFlag)
	    dprint( "Handling GIOP request" ) ;

	return new ServerRequestImpl(this, buf, (RequestMessage)msg);

    case Message.GIOPReply:
	if (orb.transportDebugFlag)
	    dprint( "Handling GIOP reply" ) ;

	return new ClientResponseImpl(this, buf, (ReplyMessage)msg);

    case Message.GIOPLocateRequest:
    case Message.GIOPLocateReply:
    case Message.GIOPCancelRequest:
    case Message.GIOPCloseConnection:
    case Message.GIOPMessageError:
	if (orb.transportDebugFlag)
	    dprint( "Handling other GIOP message" ) ;

	return new IIOPInputStream(this, buf, msg);

    default:
	if (orb.transportDebugFlag)
	    dprint( "Unknown GIOP message" ) ;

	throw new INTERNAL(MinorCodes.BAD_GIOP_REQUEST_TYPE,
			   CompletionStatus.COMPLETED_NO);
    }
}


/**
 * Process the IIOP message contained in the IIOPInputStream object.
 * Called from ReaderThread only.
 */
void processInput(IIOPInputStream s)
    throws DeleteConn
{
    int msgType = s.getMessage().getType();

    switch (msgType) {

    case Message.GIOPLocateRequest:
    case Message.GIOPRequest: {
	if (orb.transportDebugFlag)
	    dprint("Connection:processInput: got incoming request");

	    CachedWorkerThread.addWorkToDo(server.getRequestHandler(), this, s);

	break;
    }

    case Message.GIOPReply:
    case Message.GIOPLocateReply: {
	if (orb.transportDebugFlag)
	    dprint("Connection:processInput: got RESPONSE");

	// Remove this call from the list of outstanding calls
	Integer index = new Integer(s.getMessage().getRequestId());
	OutCallDesc call = (OutCallDesc) out_calls.get(index);
	out_calls.remove(index);

	// Set the reply IIOPInputStream and signal the client thread
	// that the reply has been received.
	lock.lock();
	call.s = s;
	call.done.signal();
	lock.unlock();

	break;
    }

    case Message.GIOPCancelRequest: {
	if (orb.transportDebugFlag)
	    dprint("Connection:processInput: got cancel");

	//
	// Do nothing for now (actually legal IIOP).
	//
	break;
    }

    case Message.GIOPCloseConnection: {
	//
	// For any outstanding requests on the connection, arrange
	// for the sending thread to receive a rebind exception.
	//
	if (orb.transportDebugFlag)
	    dprint("Connection.processInput: got CloseConn, purging");

	purge_calls(Connection.CONN_REBIND, true, false);

	break;
    }

    case Message.GIOPMessageError:
    default: {
	//
	// Any other message type is unexpected and is a fatal
	// error for this connection.
	//
	if (orb.transportDebugFlag)
	    dprint("Connection: bad message type "
		   + String.valueOf(msgType));
	purge_calls(Connection.CONN_ABORT, true, false);
    }
    }
}

/**
     * Wake up the outstanding requests on the connection, and hand them
     * COMM_FAILURE exception with a given minor code. Also, delete connection
     * from connection table and stop the reader thread. Note that this should only
     * ever be called by the Reader thread for this connection.
     * @param minor_code The minor code for the COMM_FAILURE major code.
     * @param die Kill the reader thread (this thread) before exiting.
     */
void purge_calls(int minor_code, boolean die, boolean lockHeld)
{
    OutCallDesc call;

    if (orb.transportDebugFlag)
	dprint("purge_calls: starting: code = " + minor_code
	       + " die = " + die);

    //
    // If this invocation is a result of ThreadDeath caused
    // by a previous execution of this routine, just exit.
    //
    if ((state == ABORT) || (state == CLOSE_RECVD)) {
	if (orb.transportDebugFlag)
	    dprint("purge_calls: exiting duplicate invocation");
	return;
    }

    //
    // Grab the writeLock (freeze the calls)
    //
    try {
	if (!lockHeld)
	    writeLock();
    } catch (SystemException ex) {
	if (orb.transportDebugFlag)
	    dprint("purge_calls: caught exception " + ex + "; continuing");
    }

    //
    // Mark the state of the connection and determine the request status
    //
    org.omg.CORBA.CompletionStatus completion_status;

    lock.lock();
    if (minor_code ==  Connection.CONN_REBIND) {
	state = CLOSE_RECVD;
	completion_status = CompletionStatus.COMPLETED_NO;
    } else {
	state = ABORT;
	completion_status = CompletionStatus.COMPLETED_MAYBE;
    }
    lock.unlock();

    //
    // Close the socket (if its not already closed)
    //
    try {
	// if theres no socket/connection, this is just ignored
	inputStream.close();
	outputStream.close();
	socket.close();
    } catch (Exception ex) {
    }

    SystemException comm_failure_exc =
	new COMM_FAILURE(minor_code, completion_status);

    // Signal all threads with outstanding requests on this
    // connection and give them the COMM_FAILURE exception.
    java.util.Enumeration e = out_calls.elements();
    while(e.hasMoreElements()) {
	call = (OutCallDesc) e.nextElement();

	lock.lock();
	call.exc = comm_failure_exc;
	call.done.signal();
	lock.unlock();
    }


    //  clear out the OutCallDesc table
    out_calls.clear ();

    //
    // delete connection from cache and stop the reader thread
    //
    connectionTable.deleteConn(endpoint);

    //
    // Signal all the waiters of the writeLock.
    // There are 4 types of writeLock waiters:
    // 1. Send waiters:
    // 2. SendReply waiters:
    // 3. cleanUp waiters:
    // 4. purge_call waiters:
    //
    writeUnlock();

    // Stop the current thread if needed.
    if (die) {
	Thread.currentThread().stop();
    }
}


/**
 * Sets up an established connection
 */
public void setConnection(Socket _socket, ConnectionTable ctab)
    throws Exception
{
    socket = _socket;
    inputStream = socket.getInputStream();
    outputStream = socket.getOutputStream();
    connectionTable = ctab;

    lock.lock();

    state = ESTABLISHED;
    event.signalAll();

    // Catch exceptions since setDaemon can cause a
    // security exception to be thrown under netscape
    // in the Applet mode
    try {
        AccessController.doPrivileged(new PrivilegedAction() {
            public java.lang.Object run() {
                reader.setDaemon(true);
                return null;
            }
        });
    }
    catch (Exception e) {}

    reader.start();

    lock.unlock();
}

/**
 * Changes state of connection to aborted, notifying waiters
 */
public void abortConnection()
{
    lock.lock();

    state = ABORT;
    event.signalAll();

    lock.unlock();
}

/**
 * Sets the writeLock for this connection.
 * If the writeLock is already set by someone else, block till the
 * writeLock is released and can set by us.
 * IMPORTANT: this connection's lock must be acquired before
 * setting the writeLock and must be unlocked after setting the writeLock.
 */
private boolean writeLock()
{
    lock.lock();

    // Keep looping till we can set the writeLock.
    while ( true ) {

	switch ( state ) {

	case OPENING:
	    // Wait for someone else to change the state to ESTABLISHED
	    event.wait(lock);

	    // Loop back
	    break;

	case ESTABLISHED:
	    if (!writeLocked) {
		writeLocked = true;
		lock.unlock();
		return true;
	    }

	    // Wait for the connection lock.
	    event.wait(lock);

	    // Loop back
	    break;

	    //
	    // XXX
	    // Need to distinguish between client and server roles
	    // here probably.
	    //
	case ABORT:
	case CLOSE_RECVD:
	    // the connection has been closed or closing
	    // ==> throw rebind exception
	    lock.unlock();
	    throw new COMM_FAILURE( MinorCodes.CONN_CLOSE_REBIND,
				    CompletionStatus.COMPLETED_NO);

	default:
	    if (orb.transportDebugFlag)
		dprint("Connection:writeLock: weird state");
	    lock.unlock();
	    delete(Connection.CONN_ABORT);
	    return false;
	}
    }
}

/**
 * Release the write lock on this connection.
 */
private void writeUnlock()
{
    lock.lock();
    writeLocked = false;
    event.signalAll();
    lock.unlock();
}


public void delete()
{
    delete(Connection.CONN_ABORT);
}

void delete(int code)
{
    DeleteConn dc = new DeleteConn(code);
    reader.stop(dc);
}


/** Send a two-way IIOP message to the server.
 */
public IIOPInputStream invoke(IIOPOutputStream s)
    throws SystemException
{
    return send(s, false);
}



/** Send an IIOP message to the server.
 *  If not oneway, wait for the reply.
 */
public IIOPInputStream send(IIOPOutputStream s, boolean oneWay)
{
    OutCallDesc call;
    int id;
    boolean unlock = false;

    call = new OutCallDesc();
    call.thd = Thread.currentThread();
    id = s.getMessage().getRequestId();

    try {
	writeLock();
	unlock = true;
	out_calls.put(new Integer(id), call);
	try {
	    s.writeTo(outputStream);
	    outputStream.flush();
	} catch (java.io.IOException exc) {
	    throw new COMM_FAILURE(MinorCodes.WRITE_ERROR_SEND,
				   CompletionStatus.COMPLETED_NO);
	}

	writeUnlock();
	unlock = false;

	if (oneWay)
	    return null;

	lock.lock();
	while (call.s == null && call.exc == null) {
	    // Wait for the reply from the server.
	    // The ReaderThread reads in the reply IIOP message
	    // and signals us.
	    call.done.wait(lock);
	}
	lock.unlock();

	if (call.exc != null)
	    throw call.exc;
    }
    finally {
	if (unlock)
	    writeUnlock();
    }

    return call.s;
}

public void sendReply(MarshalOutputStream os)
    throws Exception
{
    boolean unlock = false;

    try {
	writeLock();
	unlock = true;
	os.writeTo(outputStream);
	outputStream.flush();
	writeUnlock();
	unlock = false;
    } catch ( Exception e ) {
	if (unlock)
	    writeUnlock();
    }
}

// Begin CacheEntry methods

public String getImplementation(String x) {
    String result = null;

    if (implementations == null){
	implementations = new Hashtable();
    } else result = (String)implementations.get(x);

    if (result == null) {
	if (codeBaseRef == null)
	    codeBaseRef = getCodeBase(orb, codeBaseServerIOR); //d11638
	result = codeBaseRef.implementation(x);
	if (result != null)
	    implementations.put(x, result);
    }

    return result;
}


//d11638 Contains iors as keys to remote CodeBase servants
private static Hashtable cbs = new Hashtable();

//d11638 replacement for calling FVDCodeBaseImpl.connect
private static CodeBase getCodeBase(ORB orb, IOR ior) {
    CodeBase cb = (CodeBase)cbs.get(ior);
    if (cb != null) {
        return cb;
    }

    org.omg.CORBA.Object obj = orb.string_to_object(ior.stringify());
    cb = CodeBaseHelper.narrow(obj);

    cbs.put(ior, cb);
    return cb;
}


public FullValueDescription getFVD(String x) {

    FullValueDescription result = null;

    if (fvds == null){
	fvds = new Hashtable();
    } else result = (FullValueDescription)fvds.get(x);

    if (result == null) {
	if (codeBaseRef == null)
	    codeBaseRef = getCodeBase(orb, codeBaseServerIOR); //d11638
	result = codeBaseRef.meta(x);

	if (result != null)
	    fvds.put(x, result);
    }

    return result;
}

public String[] getBases(String x) {
    String result[] = null;

    if (bases == null){
	bases = new Hashtable();
    } else result = (String[])bases.get(x);

    if (result == null) {
	if (codeBaseRef == null)
	    codeBaseRef = getCodeBase(orb, codeBaseServerIOR); //d11638
	result = codeBaseRef.bases(x);
	if (result != null)
	    bases.put(x, result);
    }

    return result;
}
//// End CacheEntry methods

/** public IOR locate(byte[] key) is now in Connection.java */





/***************************************************************************
									    The following methods are for dealing with Connection cleaning for
									    better scalability of servers in high network load conditions.
***************************************************************************/

/**
 * Sends CloseConnection packet assumes write lock already held
 */
private void sendCloseConnection()
{
    Message msg = new Message( orb.transportDebugFlag );
    msg.setType(Message.GIOPCloseConnection);
    IIOPOutputStream os;
    IIOPInputStream is;

    os = orb.newOutputStream(this);
    msg.write(os);
    os.setMessage(msg);

    try {
	os.writeTo(outputStream);
	outputStream.flush();
    } catch (java.io.IOException exc) {}
}


public boolean isBusy()
{
    // Note: Hashtable.size() is not synchronized
    if (requestCount > 0 || out_calls.size() > 0)
	return true;
    else
	return false;
}


/**
 * Cleans up this Connection.
 * This is called from ConnectionTable, from the ListenerThread.
 * Note:it is possible for this to be called more than once
 */
public synchronized void cleanUp() throws Exception
{
    writeLock();

    if (requestCount > 0 || out_calls.size() > 0) { // we are busy!
	writeUnlock();
	throw new Exception();
    }

    try {
	if (isServer) {
	    sendCloseConnection();
	    lock.lock();
	    state = CLOSE_SENT;
	    lock.unlock();
	}

	// stop the reader without causing it to do purge_calls
	Exception ex = new Exception();
	reader.stop(ex);

	// this also does writeUnlock();
	purge_calls(Connection.CONN_REBIND, false, true);

    } catch (Exception ex) {}
}

/** It is possible for a Close Connection to have been
 ** sent here, but we will not check for this. A "lazy"
 ** Exception will be thrown in the Worker thread after the
 ** incoming request has been processed even though the connection
 ** is closed before the request is processed. This is o.k because
 ** it is a boundary condition. To prevent it we would have to add
 ** more locks which would reduce performance in the normal case.
 **/
public synchronized void requestBegins()
{
    requestCount++;
}

public synchronized void requestEnds()
{
    requestCount--;
}

public void print()
{
    System.out.println("Connection for " + endpoint.getHostName() +
                       " @ " + endpoint.getPort());
    System.out.println("    Time stamp = " + timeStamp);
    boolean alive = reader.isAlive();
    if (alive)
	System.out.println(" Reader is Alive");
    else
	System.out.println(" Reader is not Alive");
}

CodeBase getCodeBaseReference() {
    // If the codebase IOR was never even set, we
    // can't do anything.  Fall back to using the
    // FVDCodeBaseDelegate which uses a local
    // FVD.  This is the old behavior and is
    // somewhat questionable.  REVISIT
    if (codeBaseServerIOR == null)
        return FVDCodeBaseDelegate.singleton;
    
    if (codeBaseRef != null)
        return codeBaseRef;
    
    codeBaseRef = IIOPConnection.getCodeBase(orb, codeBaseServerIOR);
    
    return codeBaseRef;
}
}
