/*
 * @(#)Message.java	1.57 01/02/09
 *
 * Copyright 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.corba.se.internal.iiop;

import java.io.IOException;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.portable.*;
import com.sun.corba.se.internal.util.MinorCodes;
import com.sun.corba.se.internal.util.Utility;
import com.sun.corba.se.internal.core.GIOPVersion;

//
// This class encapsulates GIOP/IIOP messages and acts as a factory
// for various IIOP specific message subclasses which it recognizes.
// Instances of this class are used by the stream
// to hold the various message instances that make up a particular
// invocation marshal buffer. In IIOP 1.0 there will only be a single
// message instance per stream. IIOP 1.1 supports fragmenting
// a stream's data among multiple messages. This allows a connection
// implementation to choose to buffer the fragments and write them all
// at once rather than writing each fragment immediately. Likewise, this also
// supports reading all fragments at once. Doing this can simplify
// connection management.
//

public class Message
{
    protected boolean debug = false;

    static final int defaultBufferSize = 1024;
    static final int GIOPBigEndian = 0;
    static final int GIOPLittleEndian = 1;
    static final int GIOPBigMagic =    0x47494F50;
    static final int GIOPLittleMagic = 0x504F4947;
    static final int GIOPMajor = 1;
    static final int GIOPMinor = 0;

    public static final int GIOPMessageHeaderLength = 12;

    public static final int GIOPRequest = 0;
    public static final int GIOPReply = 1;
    public static final int GIOPCancelRequest = 2;
    public static final int GIOPLocateRequest = 3;
    public static final int GIOPLocateReply = 4;
    public static final int GIOPCloseConnection = 5;
    public static final int GIOPMessageError = 6;

    public static final int GIOPLocateReplyUnknownObject = 0;
    public static final int GIOPLocateReplyObjectHere = 1;
    public static final int GIOPLocateReplyObjectForward = 2;

    //
    // Parsed message header info.
    //
    private int magic;
    private byte major;
    private byte minor;
    private byte byteOrder;
    private byte type;
    private int size;

    //
    // This is common to several message types so I've
    // moved it here for pragmatic reasons. It is access
    // only.
    //
    protected int requestId;

    //
    // This creates a raw message buffer. It won't parse or add any
    // IIOP-specific data to the buffer.
    //
    protected
	Message( boolean debug )
    {
	this.debug = debug ;
    }

    public int getRequestId() {
	return requestId;
    }

    public final int
	getType()
    {
	return type;
    }

    protected void
	setType(int type)
    {
	this.type = (byte)type;
    }

    public int
	getSize() {
	return size;
    }


    public boolean
	isFragment()
    {
	//
	// Always false for IIOP 1.0.
	//
	return false;
    }

    public boolean
	isLittleEndian()
    {
	return byteOrder == GIOPLittleEndian;
    }

    byte getByteOrder() {
	return byteOrder;
    }
	

    public void
	setSize(byte[] buf, int size)
    {
	this.size = size ;

	//
	// Patch the size field in the header.
	//
	int patch = size - GIOPMessageHeaderLength;
        if (byteOrder == GIOPBigEndian) {
            buf[8] = (byte)((patch >>> 24) & 0xFF);
            buf[9] = (byte)((patch >>> 16) & 0xFF);
            buf[10] = (byte)((patch >>> 8) & 0xFF);
            buf[11] = (byte)((patch >>> 0) & 0xFF);
        } else {
            buf[8] = (byte)((patch >>> 0) & 0xFF);
            buf[9] = (byte)((patch >>> 8) & 0xFF);
            buf[10] = (byte)((patch >>> 16) & 0xFF);
            buf[11] = (byte)((patch >>> 24) & 0xFF);
        }
    }

    public static Message
	createFromStream(ORB orb, byte[] buf, java.io.InputStream is)
	throws SystemException, IOException
    {
	Message msg;
	boolean debug = orb.transportDebugFlag ;

	readFully(is, buf, 0, GIOPMessageHeaderLength);
	if (debug)
	    Utility.dprint("Message", "createFromStream: type is " + buf[7]);

	switch (buf[7]) {
	    //
	    // Requests from Clients.
	    //

	case GIOPRequest:
	    if (debug)
		Utility.dprint( "Message", 
				"createFromStream: creating RequestMessage" ) ;
	    msg = new RequestMessage( debug );
	    break;

	case GIOPLocateRequest:
	    msg = new LocateRequestMessage( debug );
	    break;

	case GIOPCancelRequest:
	    msg = new CancelRequestMessage( debug );
	    break;

	    //
	    // Replies from servers.
	    //

	case GIOPReply:
	    msg = new ReplyMessage( debug );
	    break;

	case GIOPLocateReply:
	    msg = new LocateReplyMessage( debug );
	    break;

	    //
	    // These 2 have no data outside the GIOP header so they
	    // are created using just the base Message class.
	    //
	case GIOPCloseConnection:
	case GIOPMessageError:
	    msg = new Message( debug );
	    break;

	default:
	    if (debug)
	        Utility.dprint("Message", "createFromStream: fell off end!");
	    throw new INTERNAL( MinorCodes.BAD_GIOP_REQUEST_TYPE,
				CompletionStatus.COMPLETED_NO );
	}

	//
	// Initialize the generic GIOP header instance variables.
	//

	//
	// Read "GIOP" characters in as a big-endian integer
	// since that's easier to compare against.
	//
	int b1, b2, b3, b4;
	b1 = (buf[0] << 24) & 0xFF000000;
	b2 = (buf[1] << 16) & 0x00FF0000;
	b3 = (buf[2] << 8) & 0x0000FF00;
	b4 = (buf[3] << 0) & 0x000000FF;
	msg.magic = (b1 | b2 | b3 | b4);
	msg.major = buf[4];
	msg.minor = buf[5];
	msg.byteOrder = buf[6];
	msg.type = buf[7];

	//
	// Hand-coded getLong() to get the size of the
	// message based on the byte-order flag.
	//
	if (msg.byteOrder == GIOPBigEndian) {
	    b1 = (buf[8] << 24) & 0xFF000000;
	    b2 = (buf[9] << 16) & 0x00FF0000;
	    b3 = (buf[10] << 8) & 0x0000FF00;
	    b4 = (buf[11] << 0) & 0x000000FF;
	} else {
	    b1 = (buf[11] << 24) & 0xFF000000;
	    b2 = (buf[10] << 16) & 0x00FF0000;
	    b3 = (buf[9] << 8) & 0x0000FF00;
	    b4 = (buf[8] << 0) & 0x000000FF;
	}
        msg.size = (b1 | b2 | b3 | b4) + GIOPMessageHeaderLength;

	if (debug)
	    Utility.dprint( "Message", 
			    "createFromStream: message construction complete." ) ;	
	return msg;
    }

    public static void
	readFully(java.io.InputStream is, byte[] buf, int offset, int size)
	throws IOException
    {
	int n = 0;
	while (n < size) {
	    int bytecount=0;
	    int itns=0;

	    // The while loop here is to workaround a possible bug
	    // in the Solaris JVM where an "interrupted system call"
	    // error is thrown as a SocketException instead of being
	    // retried inside SocketInputStream.read().
	    while ( true ) {
	        try {
	            bytecount = is.read(buf, offset + n, size - n);
		    break;
	        } catch ( java.io.IOException ex ) {
		    if ( itns++ >= 5 )
			throw ex;
	        }
	    }
		
	    if (bytecount < 0)
		throw new IOException();
	    n += bytecount;
	}
    }


    public void
	write(OutputStream s)
	throws SystemException
    {
	s.write_long(GIOPBigMagic);
	s.write_octet((byte)GIOPMajor);
	s.write_octet((byte)GIOPMinor);
	s.write_boolean(false); // BigEndian
	s.write_octet(type);
	s.write_long(0);
    }

    //
    // No-op. We've already unmarshaled the GIOP header
    // in createFromStream.
    //
    public void
	read(InputStream s)
	throws SystemException
    {
    }

    /**
     * Get the GIOPVersion from the Message based on Major and Minor numbers.
     */
    public GIOPVersion getGIOPVersion( ) {
        if( major != 0 ) {
            return new GIOPVersion( major, minor );
        }
        // If major = 0, then it means this is a new Message created for
        // a request, So return a 1.0 GIOPVersion object.
        return new GIOPVersion( GIOPMajor, GIOPMinor );
    }

}




