/*
 * @(#)Connection.java	1.12 01/02/09
 *
 * Copyright 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jndi.ldap;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.InterruptedIOException;
import java.io.IOException;
import java.io.OutputStream;
import java.io.InputStream;
import java.net.Socket;
import java.util.Vector;
import java.util.Hashtable;

import javax.naming.CommunicationException;
import javax.naming.AuthenticationException;
import javax.naming.AuthenticationNotSupportedException;
import javax.naming.ServiceUnavailableException;
import javax.naming.NamingException;
import javax.naming.InterruptedNamingException;

import javax.naming.ldap.Control;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
//import javax.net.SocketFactory;

/**
  * A thread that creates a connection to an LDAP server.
  * After the connection, the thread reads from the connection.
  * A caller can invoke methods on the instance to read LDAP responses
  * and to send LDAP requests.
  * <p>
  * There is a one-to-one correspondence between an LdapClient and
  * a Connection. Access to Connection and its methods is only via
  * LdapClient with one exception: SASL authentication. SASL needs
  * to access Connection's socket IO streams (in order to do encryption
  * of the security layer). [StartTLS might need to do something
  * similar.] The code for SASL authentication can be treated as being the
  * same as from LdapClient because the SASL code is only ever called
  * from LdapClient, from inside LdapClient's synchronized authenticate() method.
  * <p>
  * In terms of synchronization, worry about data structures
  * used by the Connection thread because that usage might contend
  * with calls by the main threads (i.e., those that call LdapClient). 
  * Main threads need to worry about contention with each other.
  * Fields that Connection thread uses:
  *     inStream - no sync; initialized in constructor; update synced 
  * 		by streamLock during 'quiet' period by main threads
  *     traceFile, traceTagIn, traceTagOut - no sync; debugging only
  *     parent - no sync; initialized in constructor; no updates
  *     pendingRequests - sync
  *     streamLock - per-instance lock; 
  *     updateStream - sync via streamLock
  * Members used by main threads (LdapClient):
  *     host, port - unsync; read-only access for messages in debug
  *     setBound(), setV3() - no sync; called only by LdapClient.authenticate(),
  * 		which is a sync method called only when connection is "quiet"
  *     getMsgId() - sync
  *     writeRequest(), removeRequest(),findRequest(), abandonOutstandingReqs() - 
  * 		access to shared pendingRequests is sync
  * 	writeRequest(),  abandonRequest(), ldapUnbind() - access to outStream sync
  *     cleanup() - sync
  *     readReply() - access to sock sync
  * Members used by SASL auth:
  *     inStream, outStream - no sync; used to construct new stream; accessed
  * 		only when conn is "quiet" and not shared
  *     replaceInputStream() - sync on streamLock for updates to ensure changes
  * 		are visible to interrupted Connection thread
  *     replaceOutputStream() - no sync; updates outStream when conn is "quiet"
  * 		and not shared; when LdapClient.authenticate() releases
  *		LdapClient lock, update to outStream will be visible
  * <p>
  * Handles anonymous, simple, and SASL bind for v3; anonymous and simple
  * for v2.
  * %%% made public for access by LdapSasl %%%
  * 
  * @author Vincent Ryan
  * @author Rosanna Lee
  * @author Jagane Sundar
  */ 
public final class Connection implements Runnable {

    private static final boolean debug = false;
    private static final int dump = 0; // > 0 r, > 1 rw

    // Initialized in constructor and accessed in replaceInputStream()
    // inside sync streamLock.
    private Thread worker;

    private boolean v3 = true;       // Set in setV3()

    String host;   // used by LdapClient for generating exception messages
    int port;      // used by LdapClient for generating exception messages

    private boolean bound = false;   // Set in setBound()

    // All three are initialized in constructor and read-only afterwards
    private OutputStream traceFile = null;
    private String traceTagIn = null;
    private String traceTagOut = null;

    // Initialized in constructor; read and used externally (LdapSasl); 
    // Updated in replaceInputStream() during "quiet", unshared, period
    public InputStream inStream;   // must be public; used by LdapSasl

    // Initialized in constructor; read and used externally (LdapSasl); 
    // Updated in replaceOutputStream() during "quiet", unshared, period
    public OutputStream outStream; // must be public; used by LdapSasl

    // Initialized in constructor; read and used externally (TLS) to
    // get new IO streams; closed during cleanup
    public Socket sock;		   // for TLS

    // For processing "disconnect" unsolicited notification
    // Initialized in constructor
    private LdapClient parent = null; 

    // Incremented and returned in sync getMsgId()
    private int outMsgId = 0;

    //
    // The list of ldapRequests pending on this binding
    //
    // Accessed only within sync methods
    private LdapRequest pendingRequests = null;

    // true means v3; false means v2
    // Called in LdapClient.authenticate() (which is synchronized)
    // when connection is "quiet" and not shared; no need to synchronize
    void setV3(boolean v) {
	v3 = v;
    }

    // A BIND request has been successfully made on this connection
    // When cleaning up, remember to do an UNBIND
    // Called in LdapClient.authenticate() (which is synchronized)
    // when connection is "quiet" and not shared; no need to synchronize
    void setBound() {
	bound = true;
    }

    ////////////////////////////////////////////////////////////////////////////
    //
    // Create an LDAP Binding object and bind to a particular server
    //
    ////////////////////////////////////////////////////////////////////////////

    Connection(LdapClient parent, String host, int port, String socketFactory, 
	OutputStream trace) throws NamingException {

	this.host = host;
	this.port = port;
	this.parent = parent;

	if (trace != null) {
	    traceFile = trace;
	    traceTagIn = "<- " + host + ":" + port + "\n\n";
	    traceTagOut = "-> " + host + ":" + port + "\n\n";
	}

	//
	// Connect to server
	//
	try {
	    if (socketFactory != null) {
		/*
		 * Use the Obj.helper.loadClass method to load the SSL
		 * socket factory class (from the jre/lib/ext directory).
		 * The application's class loader is used.
		 *
		 * Class.forName() cannot be called directly to load
		 * classes from the jre/lib/ext directory when the
		 * bootstrap class loader is active because they are
		 * not visible to that class loader.
		 */
		Class factoryClass = Obj.helper.loadClass(socketFactory);

		// factory = SocketFactory.getDefault();
		Method getDefault =
		    factoryClass.getMethod("getDefault", new Class[]{});

		// invoke the static method SocketFactory.getDefault()
		Object factory = getDefault.invoke(null, new Object[]{});

		// sock = (Socket)factory.createSocket(host, port);
		Method createSocket = 
		    factoryClass.getMethod("createSocket",
			new Class[]{String.class, int.class});

		sock = (Socket) createSocket.invoke(factory,
		    new Object[]{host, new Integer(port)});

	    } else {
		sock = new Socket(host, port);
	    }

	    if (debug) {
		System.err.println("Connection: opening socket: " + host + "," + port);
	    }

	    inStream = new BufferedInputStream(sock.getInputStream());

	    // Workaround for JSSE 1.0 race condition problem
	    OutputStream rawOutStream = sock.getOutputStream();
	    rawOutStream.write(new byte[0]);

	    outStream = new BufferedOutputStream(rawOutStream);

	} catch (InvocationTargetException e) {
	    Throwable realException = e.getTargetException();
	    // realException.printStackTrace();

            CommunicationException ce = 
		new CommunicationException(host + ":" + port);
            ce.setRootCause(realException);
	    throw ce;
	} catch (Exception e) {
	    // Class.forName() seems to do more error checking
	    // and will throw IllegalArgumentException and such.
	    // That's why we need to have a catch all here and
	    // ignore generic exceptions.
	    // Also catches all IO errors generated by socket creation.
            CommunicationException ce = 
		new CommunicationException(host + ":" + port);
            ce.setRootCause(e);
	    throw ce;
	}

	worker = Obj.helper.createThread(this);
	worker.setDaemon(true);
	worker.start();
    }


    ////////////////////////////////////////////////////////////////////////////
    //
    // Methods to IO to the LDAP server
    //
    ////////////////////////////////////////////////////////////////////////////

    synchronized int getMsgId() {
	return ++outMsgId;
    }

    LdapRequest writeRequest(BerEncoder ber, int msgId) throws IOException {

	LdapRequest req = new LdapRequest(msgId);
	addRequest(req);

	if (traceFile != null) {
	    Ber.dumpBER(traceFile, traceTagOut, ber.getBuf(), 0, ber.getDataLen());
	}

	synchronized (this) {
	    outStream.write(ber.getBuf(), 0, ber.getDataLen());
	    outStream.flush();
	}
	return req;
    }

    /**
     * Reads a reply; waits until one is ready.
     */
    BerDecoder readReply(LdapRequest ldr) throws IOException, NamingException {
	BerDecoder rber;
	while ((rber = ldr.getReplyBer()) == null) {
	    try {
		// If socket closed, don't even try
		synchronized (this) {
		    if (sock == null) {
			throw new ServiceUnavailableException(host + ":" + port +
			    "; socket closed");
		    }
		}
		synchronized (ldr) {
		    // check if condition has changed since our last check
		    rber = ldr.getReplyBer();
		    if (rber == null) {
			ldr.wait(15 * 1000); // 15 second timeout
		    } else {
			break;
		    }
		}
	    } catch (InterruptedException ex) {
		throw new InterruptedNamingException(
		    "Interrupted during LDAP operation");
	    }
	}
	return rber;
    }
    

    ////////////////////////////////////////////////////////////////////////////
    //
    // Methods to add, find, delete, and abandon requests made to server
    //
    ////////////////////////////////////////////////////////////////////////////

    private synchronized void addRequest(LdapRequest ldapRequest) {

	LdapRequest ldr = pendingRequests;
	if (ldr == null) {
	    pendingRequests = ldapRequest;
	    ldapRequest.next = null;
	} else {
	    ldapRequest.next = pendingRequests;
	    pendingRequests = ldapRequest;
	}
    }

    synchronized LdapRequest findRequest(int msgId) {

	LdapRequest ldr = pendingRequests;
	while (ldr != null) {
	    if (ldr.msgId == msgId) {
		return ldr;
	    }
	    ldr = ldr.next;
	}
	return null;

    }

    synchronized void removeRequest(LdapRequest req) {
	LdapRequest ldr = pendingRequests;
	LdapRequest ldrprev = null;

	while (ldr != null) {
	    if (ldr == req) {
		ldr.cancel();

		if (ldrprev != null) {
		    ldrprev.next = ldr.next;
		} else {
		    pendingRequests = ldr.next;
		}
		ldr.next = null;
	    }
	    ldrprev = ldr;
	    ldr = ldr.next;
	}
    }

    void abandonRequest(LdapRequest ldr, Control[] reqCtls) {
	// Remove from queue
	removeRequest(ldr);

	BerEncoder ber = new BerEncoder(256);
	int abandonMsgId = getMsgId();

	//
	// build the abandon request.
	//
	try {
	    ber.beginSeq(Ber.ASN_SEQUENCE | Ber.ASN_CONSTRUCTOR);
	        ber.encodeInt(abandonMsgId);
		ber.encodeInt(ldr.msgId, LdapClient.LDAP_REQ_ABANDON);

		if (v3) {
		    LdapClient.encodeControls(ber, reqCtls);
		}
	    ber.endSeq();

	    if (traceFile != null) {
		Ber.dumpBER(traceFile, traceTagOut, ber.getBuf(), 0, 
		    ber.getDataLen());
	    }

	    synchronized (this) {
		outStream.write(ber.getBuf(), 0, ber.getDataLen());
		outStream.flush();
	    }

	} catch (IOException ex) {
	    //System.err.println("ldap.abandon: " + ex);
	}

	// Dont expect any response for the abandon request.
    }

    synchronized void abandonOutstandingReqs(Control[] reqCtls) {
        LdapRequest ldr = pendingRequests;

	while (ldr != null) {
	    abandonRequest(ldr, reqCtls);
	    pendingRequests = ldr = ldr.next;
	}
    }

    ////////////////////////////////////////////////////////////////////////////
    //
    // Methods to unbind from server and clear up resources when object is
    // destroyed.
    //
    ////////////////////////////////////////////////////////////////////////////

    private void ldapUnbind(Control[] reqCtls) {

	BerEncoder ber = new BerEncoder(256);
	int unbindMsgId = getMsgId();

	//
	// build the unbind request.
	//

	try {

	    ber.beginSeq(Ber.ASN_SEQUENCE | Ber.ASN_CONSTRUCTOR);
	        ber.encodeInt(unbindMsgId);
		// IMPLICIT TAGS
	        ber.encodeByte(LdapClient.LDAP_REQ_UNBIND);
	        ber.encodeByte(0);

		if (v3) {
		    LdapClient.encodeControls(ber, reqCtls);
		}
	    ber.endSeq();

	    if (traceFile != null) {
		Ber.dumpBER(traceFile, traceTagOut, ber.getBuf(),
		    0, ber.getDataLen());
	    }

	    synchronized (this) {
		outStream.write(ber.getBuf(), 0, ber.getDataLen());
		outStream.flush();
	    }

	} catch (IOException ex) {
	    //System.err.println("ldap.unbind: " + ex);
	}

	// Dont expect any response for the unbind request.
    }

/*  
No need to have finalize method because run() has a finally clause that
calls cleanup(null).
  
   protected void finalize() {
      if (debug) {
         System.err.println("Connection: finalize called: " + host + "," + 
	     port + " " + this);
      }
      cleanup(null);
   }
*/

    synchronized void cleanup(Control[] reqCtls) {

	if (sock != null) {
	    if (debug) {
		System.err.println("Connection: closing socket: " + host + "," + port);
	    }
	    try {
		abandonOutstandingReqs(reqCtls);
		if (bound) {
		    ldapUnbind(reqCtls);
		}
	    } finally {
		try {
		    outStream.flush();
		    sock.close();
		} catch (java.io.IOException ie) {
		    if (debug)
			System.err.println("Connection: problem closing socket: " + ie);
		}
		LdapRequest ldr = pendingRequests;
		while (ldr != null) {
		    ldr.cancel();
		    ldr = ldr.next;
		}
		sock = null;
	    } 
	}
    }

    // lock for replacer to wait on for thread to switch to new input stream
    private Object streamLock = new Object();

    // flag for thread to check when it gets an IOException
    private boolean updateStream = false;
    
    // Assume everything is "quiet"
    // "synchronize" might lead to deadlock so don't synchronize method
    // Use streamLock instead for synchronizing update to stream

    public void replaceInputStream(InputStream newIn) throws IOException {
	if (debug) {
	    System.err.println("Replacing " + inStream + " with: " + newIn);
	}

	synchronized (streamLock) {
	    if (inStream == newIn) {
		return;  // nothing to do
	    }

	    // Do assignment inside sync block
	    // to ensure flush when streamLock is released in wait()

	    // Replace stream
	    inStream = newIn;
	
	    // Interrupt thread that's currently reading from old stream
	    // so that it can start reading from the new stream
	    if (debug) {
		System.err.println("Interrupting worker");
	    }

	    updateStream = true;
	    worker.interrupt();

	    try {
		while (updateStream) {
		    streamLock.wait();
		}
	    } catch (InterruptedException e) {
		throw new InterruptedIOException(
		    "Cannot replace input stream");
	    } finally {
		updateStream = false;
	    }
	}
    }

    public void replaceOutputStream(OutputStream newOut) throws IOException {
	if (debug) {
	    System.err.println("Replacing " + outStream + " with: " + newOut);
	}

	// Cleanup old stream
	try {
	    outStream.flush();
	} catch (java.io.IOException ie) {
	    if (debug)
		System.err.println("Connection: cannot flush outstream: " + ie);
	}
	
	// Replace stream
	outStream = newOut;
    }

    ////////////////////////////////////////////////////////////////////////////
    //
    // The LDAP Binding thread. It does the mux/demux of multiple requests
    // on the same TCP connection.
    //
    ////////////////////////////////////////////////////////////////////////////


    public void run() {
	byte inbuf[];	// Buffer for reading incoming bytes
	int inMsgId;	// Message id of incoming response
	int bytesread;	// Number of bytes in inbuf
	int bytesleft;	// Number of bytes that need to read for completing resp
	int br;		// Temp; number of bytes read from stream
	int offset;     // Offset of where to store bytes in inbuf
	int seqlen;     // Length of ASN sequence
	int seqlenlen;  // Number of sequence length bytes
	boolean eos;    // End of stream
	BerDecoder retBer;    // Decoder for ASN.1 BER data from inbuf

	try {
	    while (true) {
		try {
		    inbuf = new byte[2048];

		    offset = 0;
		    seqlen = 0;
		    seqlenlen = 0;

		    // check that it is the beginning of a sequence
		    bytesread = inStream.read(inbuf, offset, 1);
		    if (bytesread < 0)
			break;
		    if (inbuf[offset++] != (Ber.ASN_SEQUENCE | Ber.ASN_CONSTRUCTOR))
			continue;

		    // get length of sequence
		    bytesread = inStream.read(inbuf, offset, 1);
		    if (bytesread < 0)
			break;
		    seqlen = inbuf[offset++];

		    // if high bit is on, length is encoded in the 
		    // subsequent length bytes and the number of length bytes 
		    // is equal to & 0x80 (i.e. length byte with high bit off).
		    if ((seqlen & 0x80) == 0x80) {
			seqlenlen = seqlen & 0x7f;  // number of length bytes

			bytesread = 0;
			eos = false;

			// Read all length bytes
			while (bytesread < seqlenlen) {
			    br = inStream.read(inbuf, offset+bytesread,      
				seqlenlen-bytesread);
			    if (br < 0) {
				eos = true;
				break;
			    }
			    bytesread += br;
			}

			// end-of-stream reached before length bytes are read
			if (eos) 
			    break; 

			// Add contents of length bytes to determine length
			seqlen = 0;
			for( int i = 0; i < seqlenlen; i++) {
			    seqlen = (seqlen << 8) + (inbuf[offset+i] & 0xff);
			}
			offset += bytesread;
		    }

		    // read in seqlen bytes
		    bytesleft = seqlen;
		    if ((offset + bytesleft) > inbuf.length) {
			byte nbuf[] = new byte[offset + bytesleft];
			System.arraycopy(inbuf, 0, nbuf, 0, offset);
			inbuf = nbuf;
		    }
		    while (bytesleft > 0) {
			bytesread = inStream.read(inbuf, offset, bytesleft);
			if (bytesread < 0)
			    break;
			offset += bytesread;
			bytesleft -= bytesread;
		    }
/* 
if (dump > 0) {
System.err.println("seqlen: " + seqlen);
System.err.println("bufsize: " + offset);
System.err.println("bytesleft: " + bytesleft);
System.err.println("bytesread: " + bytesread);
}
*/

		    try {
			retBer = new BerDecoder(inbuf, 0, offset);

			if (traceFile != null) {
			    Ber.dumpBER(traceFile, traceTagIn, inbuf, 0, offset);
			}

			retBer.parseSeq(null);
			inMsgId = retBer.parseInt();
			retBer.reset();	// reset offset

			if (inMsgId == 0) {
			    // Unsolicited Notification
			    parent.processUnsolicited(retBer);
			} else {
			    LdapRequest ldr = findRequest(inMsgId);
			    if (ldr != null) {
				ldr.addReplyBer(retBer);
			    } else {
	    // System.err.println("Cannot find LdapRequest for " + inMsgId);
			    }
			}
		    } catch (Ber.DecodeException e) {
			//System.err.println("Cannot parse Ber");
		    }
		} catch (IOException ie) {
		    // Thread interrupted (due to stream change)
		    // Need to catch IOException because it isn't always
		    // InterruptedIOException
		    
		    if (debug) {
			System.err.println("Connection: Inside Caught " + ie);
			ie.printStackTrace();
		    }

		    synchronized (streamLock) {
			if (!updateStream) {
			    if (debug) {
				System.err.println("Connection: rethrowing " + ie);
			    }
			    throw ie;
			}
		    
			// Let replacer know that we're ready
			updateStream = false;
			streamLock.notify();
		    }

		    if (debug) {
			System.err.println("Connection: using new stream: " +
			    inStream);
		    }
		    
		    // Go to top of loop and continue
		}
	    }
	} catch (java.io.IOException ex) {
	    if (debug) {
		System.err.println("Connection: Caught " + ex);
	    }
	} finally {
	    cleanup(null); // cleanup
	}
	if (debug) {
	    System.err.println("Connection: Thread Exiting");
	}
    }


    // This code must be uncommented to run the LdapAbandonTest.
    /*public void sendSearchReqs(String dn, int numReqs) {
        int i;
	String attrs[] = null;
        for(i = 1; i <= numReqs; i++) {
	    BerEncoder ber = new BerEncoder(2048);

	    try {
	    ber.beginSeq(Ber.ASN_SEQUENCE | Ber.ASN_CONSTRUCTOR);
		ber.encodeInt(i);
		ber.beginSeq(LdapClient.LDAP_REQ_SEARCH);
		    ber.encodeString(dn == null ? "" : dn);
		    ber.encodeInt(0, LdapClient.LBER_ENUMERATED);
		    ber.encodeInt(3, LdapClient.LBER_ENUMERATED);
		    ber.encodeInt(0);
		    ber.encodeInt(0);
		    ber.encodeBoolean(true);
		    LdapClient.encodeFilter(ber, "");
		    ber.beginSeq(Ber.ASN_SEQUENCE | Ber.ASN_CONSTRUCTOR);
			ber.encodeStringArray(attrs);
		    ber.endSeq();
		ber.endSeq();
	    ber.endSeq();
	    writeRequest(ber, i);
	    //System.err.println("wrote request " + i);
	    } catch (Exception ex) {
	    //System.err.println("ldap.search: Caught " + ex + " building req");
	    }

	}
    } */

}
