/*
 * @(#)LdapCtxFactory.java	1.3 00/02/02
 *
 * Copyright 1999, 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jndi.ldap;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.net.MalformedURLException;

import javax.naming.*;
import javax.naming.directory.*;
import javax.naming.spi.ObjectFactory;
import javax.naming.spi.InitialContextFactory;
import javax.naming.ldap.Control;

import com.sun.jndi.url.ldap.ldapURLContextFactory;

final public class LdapCtxFactory implements ObjectFactory, InitialContextFactory {
    /**
     * The type of each address in an LDAP reference.
     */
    public final static String ADDRESS_TYPE = "URL";

    // ----------------- ObjectFactory interface --------------------

    public Object getObjectInstance(Object ref, Name name, Context nameCtx,
	Hashtable env) throws Exception {

	if (!isLdapRef(ref)) {
	    return null;
	}
	ObjectFactory factory = new ldapURLContextFactory();
	String[] urls = getURLs((Reference)ref);
	return factory.getObjectInstance(urls, name, nameCtx, env);
    }

    // ----------------- InitialContext interface  --------------------

    public Context getInitialContext(Hashtable envprops) throws NamingException {
	try {
	    String url = (envprops != null) ?
		(String)envprops.get(Context.PROVIDER_URL) : null;

	    String DN, host;
	    int port;

	    // If URL not in environment, use defaults
	    if (url == null) {
		host = LdapCtx.DEFAULT_HOST;
		port = LdapCtx.DEFAULT_PORT;
		DN = "";
	    } else {
		LdapURL ldapUrl = new LdapURL(url);
		String scheme = ldapUrl.getScheme();

		if (scheme == null) {
		    throw new ConfigurationException("URL scheme is missing");
		}
		if (! scheme.equals("ldap")) {
		    throw new ConfigurationException(
			"URL scheme '" + scheme + "'is unsupported");
		}

		port = ldapUrl.getPort();
		if (port == -1)
		    port = 0;

		DN = ldapUrl.getDN();
		host = ldapUrl.getHost();
	    }
	    return new LdapCtx(DN, host, port, envprops);

	} catch (LdapReferralException e) {

	    if (envprops != null && 
		"throw".equals(envprops.get(Context.REFERRAL))) {
		throw e;
	    }

	    Control[] bindCtls = (envprops != null)? 
		(Control[])envprops.get(LdapCtx.BIND_CONTROLS) : null;

	    return (LdapCtx)e.getReferralContext(envprops, bindCtls);
	}
    }

    /**
     * Returns true if argument is an LDAP reference.
     */
    private static boolean isLdapRef(Object obj) {

	if (!(obj instanceof Reference)) {
	    return false;
	}
	String thisClassName = LdapCtxFactory.class.getName();
	Reference ref = (Reference)obj;

	return thisClassName.equals(ref.getFactoryClassName());
    }

    /**
     * Returns the URLs contained within an LDAP reference.
     */
    private static String[] getURLs(Reference ref) throws NamingException {

	int size = 0;	// number of URLs
	String[] urls = new String[ref.size()];

	Enumeration addrs = ref.getAll();
	while (addrs.hasMoreElements()) {
	    RefAddr addr = (RefAddr)addrs.nextElement();

	    if ((addr instanceof StringRefAddr) &&
		addr.getType().equals(ADDRESS_TYPE)) {

		urls[size++] = (String)addr.getContent();
	    }
	}
	if (size == 0) {
	    throw (new ConfigurationException(
		    "Reference contains no valid addresses"));
	}

	// Trim URL array down to size.
	if (size == ref.size()) {
	    return urls;
	}
	String[] urls2 = new String[size];
	System.arraycopy(urls, 0, urls2, 0, size);
	return urls2;
    }

    // ------------ Utilities used by other classes ----------------
    /**
     * Used by Obj and obj/RemoteToAttrs too so must be public
     */
    public static Attribute createTypeNameAttr(Class cl) {
        Vector v = new Vector(10);
	String[] types = getTypeNames(cl, v);
	if (types.length > 0) {
	    BasicAttribute tAttr = 
		new BasicAttribute(Obj.JAVA_ATTRIBUTES[Obj.TYPENAME]);
	    for (int i = 0; i < types.length; i++) {
		tAttr.add(types[i]);
	    }
	    return tAttr;
	}
	return null;
    }

    private static String[] getTypeNames(Class currentClass, Vector v) {

	getClassesAux(currentClass, v);
	Class[] members = currentClass.getInterfaces();
	for (int i = 0; i < members.length; i++) {
	    getClassesAux(members[i], v);
	}
        String[] ret = new String[v.size()];
        int i = 0;
        for (java.util.Enumeration e = v.elements(); e.hasMoreElements();) {
            ret[i++] = (String)e.nextElement();
        }
        return ret;
    }

    private static void getClassesAux(Class currentClass, Vector v) {
	if (!v.contains(currentClass.getName())) {
	    v.addElement(currentClass.getName());
	}
	currentClass = currentClass.getSuperclass();

	while (currentClass != null) {
	    getTypeNames(currentClass, v);
	    currentClass = currentClass.getSuperclass();
	}
    }
}
