/*
 * @(#)LdapSchemaCtx.java	1.4 00/02/02
 *
 * Copyright 1999, 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jndi.ldap;

import javax.naming.*;
import javax.naming.directory.*;
import java.util.*;
import com.sun.jndi.toolkit.dir.HierMemDirCtx;

/**
 * This is the class used to implement LDAP's GetSchema call.
 *
 * It subclasses HierMemDirContext for most of the functionality. It
 * overrides functions that cause the schema definitions to change.
 * In such a case, it write the schema to the LdapServer and (assuming
 * there are no errors), calls it's superclass's equivalent function.
 * Thus, the schema tree and the LDAP server's schema attributes are
 * always in sync.
 */

final class LdapSchemaCtx extends HierMemDirCtx {

    // In a particular schema tree, everyone has a pointer to an instance of
    // this object. It's used to share info that all members need shared access
    // to.
    final static private class SchemaInfo {
	LdapSchemaCtx root;
	DirContext originatingEntry;
	LdapSchemaParser parser;
    }

    static private final boolean debug = false;

    private static final int LEAF = 0;          // schema object (e.g. attribute type defn)
    private static final int SCHEMA_ROOT = 1;   // schema tree root
    static final int OBJECTCLASS_ROOT = 2;   // root of object class subtree
    static final int ATTRIBUTE_ROOT = 3;     // root of attribute type subtree
    static final int SYNTAX_ROOT = 4;        // root of syntax subtree
    static final int MATCHRULE_ROOT = 5;     // root of matching rule subtree

    private SchemaInfo info= null;
    private boolean setupMode = true;
    private String stringRepr = null;

    private int objectType;

    static DirContext createSchemaTree(
	Hashtable env, DirContext schemaEntry, Attributes schemaAttrs, 
	boolean netscapeBug)
	throws NamingException {
	    LdapSchemaParser parser = new LdapSchemaParser(netscapeBug);

	    LdapSchemaCtx root = new LdapSchemaCtx(env, schemaEntry, parser);
	    parser.LDAP2JNDISchema(schemaAttrs, root);

	    return root;
    }


    // Called by LdapSchemaParser to create schema root
    private LdapSchemaCtx(Hashtable environment, DirContext originatingEntry, 
	LdapSchemaParser parser) {
	this(SCHEMA_ROOT, environment, originatingEntry, parser);
    }

    // Called by createNewCtx
    private LdapSchemaCtx(int objectType, Hashtable environment, 
	DirContext originatingEntry, LdapSchemaParser parser) {
	super(environment, LdapClient.caseIgnore);

	this.objectType = objectType;

	// since we are being constructed, we are the schema root
	// and need to create the info object for others to use.
	info = new SchemaInfo();
	info.root = this;
	info.originatingEntry = originatingEntry;
	info.parser = parser;
    }

    // override HierMemDirCtx.close to prevent premature GC of shared data
    public void close() throws NamingException {
    }

    // override to ignore obj and use attrs
    // treat same as createSubcontext
    final public void bind(Name name, Object obj, Attributes attrs)
	throws NamingException {
	    String repr = null;
	if (!setupMode) {
	    if (obj != null) {
		throw new IllegalArgumentException("obj must be null");
	    }

	    // Update server
	    repr = addServerSchema(attrs);
	}

	// Update in-memory copy
	LdapSchemaCtx newEntry = 
	    (LdapSchemaCtx)super.doCreateSubcontext(name, attrs);
	newEntry.stringRepr = repr;
    }

    final protected void doBind(Name name, Object obj) throws NamingException {
	if (!setupMode) {
	    throw new SchemaViolationException(
		"Cannot bind arbitrary object; use createSubcontext()");
	} else {
	    super.doBind(name, obj);
	}
    }	

    // override to use bind() instead
    final public void rebind(Name name, Object obj, Attributes attrs)
	throws NamingException {
        try {
	    lookup(name);
	    throw new SchemaViolationException(
		"Cannot replace existing schema object");
	} catch (NameNotFoundException e) {
	    bind(name, obj, attrs);
	}
    }

    final protected void doRebind(Name name, Object obj) throws NamingException {
	if (!setupMode) {
	    throw new SchemaViolationException(
		"Cannot bind arbitrary object; use createSubcontext()");
	} else {
	    super.doRebind(name, obj);
	}
    }

    final protected void doUnbind(Name name) throws NamingException {
	if (!setupMode) {
	    // Update server
	    try {
		// Lookup entry from memory
		LdapSchemaCtx target = (LdapSchemaCtx)lookup(name);

		// Delete using its string representation
		deleteServerSchema(target.stringRepr);
	    } catch (NameNotFoundException e) {
		return;
	    }
	}
	// Update in-memory copy
	super.doUnbind(name);
    }

    final protected void doRename(Name oldname, Name newname)
	throws NamingException {
        if (!setupMode) {
	    throw new SchemaViolationException("Cannot rename a schema object");
	} else {
	    doRename(oldname, newname);
	}
    }

    final protected void doDestroySubcontext(Name name) throws NamingException {
	if (!setupMode) {
	    // Update server
	    try {
		// Lookup entry from memory
		LdapSchemaCtx target = (LdapSchemaCtx)lookup(name);

		// Delete using its string representation
		deleteServerSchema(target.stringRepr);
	    } catch (NameNotFoundException e) {
		return;
	    }
	}

	// Update in-memory copy
	super.doDestroySubcontext(name);
     }

    // Called to create oc, attr, syntax or matching rule roots
    final LdapSchemaCtx setup(int objectType, String name, Attributes attrs)
	throws NamingException{
	    try {
		setupMode = true;
		LdapSchemaCtx answer = 
		    (LdapSchemaCtx) super.doCreateSubcontext(
			new CompositeName(name), attrs);

		answer.objectType = objectType;
		answer.setupMode = false;
		return answer;
	    } finally {
		setupMode = false;
	    }
    }

    // Called to create LEAF entries
    final LdapSchemaCtx setup(String name, Attributes attrs, String repr) 
	throws NamingException {
	try {
	    setupMode = true;
	    LdapSchemaCtx answer = (LdapSchemaCtx) super.doCreateSubcontext(
		new CompositeName(name), attrs);
	    answer.setupMode = false;
	    answer.stringRepr = repr;
	    return answer;
	} finally {
	    setupMode = false;
	}
    }

    final protected DirContext doCreateSubcontext(Name name, Attributes attrs)
	throws NamingException {
	String repr = null;

	if (attrs == null || attrs.size() == 0) {
	    throw new SchemaViolationException(
		"Must supply attributes describing schema");
	}

        if (!setupMode) {
	    // Update server
	    repr = addServerSchema(attrs);
	}

	// Update in-memory copy
	LdapSchemaCtx newEntry = 
	    (LdapSchemaCtx) super.doCreateSubcontext(name, attrs);
	newEntry.stringRepr = repr;
	return newEntry;
    }

    final protected void doModifyAttributes(ModificationItem[] mods) 
	throws NamingException {
	if (setupMode) {
	    super.doModifyAttributes(mods);
	} else {
	    throw new SchemaViolationException("Cannot modify a schema object");
	}
    }

    // we override this so the superclass creates the right kind of contexts
    // Default is to create LEAF objects; caller will change after creation
    // if necessary
    final protected HierMemDirCtx createNewCtx() {
	LdapSchemaCtx ctx = new LdapSchemaCtx(LEAF, myEnv, info.originatingEntry,
	    info.parser);
	ctx.setInfo(this.info);
	return ctx;
    }

    final private void setInfo(SchemaInfo info){
	this.info = info;
    }

    final private String addServerSchema(Attributes attrs) 
	throws NamingException {
	Attribute schemaAttr;

	switch (objectType) {
	case OBJECTCLASS_ROOT:
	    schemaAttr = info.parser.stringifyObjDesc(attrs);
	    break;

	case ATTRIBUTE_ROOT:
	    schemaAttr = info.parser.stringifyAttrDesc(attrs);
	    break;
	    
	case SYNTAX_ROOT:
	    schemaAttr = info.parser.stringifySyntaxDesc(attrs);
	    break;
	
	case MATCHRULE_ROOT:
	    schemaAttr = info.parser.stringifyMatchRuleDesc(attrs);
	    break;

	case SCHEMA_ROOT:
	    throw new SchemaViolationException(
		"Cannot create new entry under schema root");

	case LEAF:
	default:
	    throw new SchemaViolationException(
		"Cannot create child of schema object");
	}

	Attributes holder = new BasicAttributes(true);
	holder.put(schemaAttr);
	//System.err.println((String)schemaAttr.get());

	this.info.originatingEntry.modifyAttributes("", DirContext.ADD_ATTRIBUTE, holder);


	return (String)schemaAttr.get();
    }

    /**
      * When we delete an entry, we use the original to make sure that
      * any formatting inconsistencies are eliminated.
      * This is because we're just deleting a value from an attribute
      * on the server and there might not be any checks for extra spaces
      * or parens.
      */
    final private void deleteServerSchema(String repr)
	throws NamingException {
	String attrName;

	switch (objectType) {
	case OBJECTCLASS_ROOT:
	    attrName = LdapSchemaParser.OBJECTCLASSDESC_ATTR_ID;
	    break;

	case ATTRIBUTE_ROOT:
	    attrName = LdapSchemaParser.ATTRIBUTEDESC_ATTR_ID;
	    break;
	    
	case SYNTAX_ROOT:
	    attrName = LdapSchemaParser.SYNTAXDESC_ATTR_ID;
	    break;

	case MATCHRULE_ROOT:
	    attrName = LdapSchemaParser.MATCHRULEDESC_ATTR_ID;
	    break;

	case SCHEMA_ROOT:
	    throw new SchemaViolationException(
		"Cannot create new entry under schema root");

	case LEAF:
	default:
	    throw new SchemaViolationException(
		"Cannot create child of schema object");
	}

	Attributes holder = new BasicAttributes(attrName, repr, true);
	this.info.originatingEntry.modifyAttributes("", DirContext.REMOVE_ATTRIBUTE, holder);
    }
}
