/*
 * @(#)LdapURL.java	1.5 00/02/02
 *
 * Copyright 1999, 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.jndi.ldap;

import javax.naming.*;
import javax.naming.directory.*;
import javax.naming.spi.*;
import java.net.URL;
import java.net.MalformedURLException;
import com.sun.jndi.toolkit.url.UrlUtil;

/*
 * Extract components of an LDAP URL.
 *
 * The format of an LDAP URL is defined in RFC 2255 as follows:
 *
 *     ldapurl    = scheme "://" [hostport] ["/"
 *                  [dn ["?" [attributes] ["?" [scope]
 *                  ["?" [filter] ["?" extensions]]]]]]
 *     scheme     = "ldap"
 *     attributes = attrdesc *("," attrdesc)
 *     scope      = "base" / "one" / "sub"
 *     dn         = distinguishedName from Section 3 of [1]
 *     hostport   = hostport from Section 5 of RFC 1738 [5]
 *     attrdesc   = AttributeDescription from Section 4.1.5 of [2]
 *     filter     = filter from Section 4 of [4]
 *     extensions = extension *("," extension)
 *     extension  = ["!"] extype ["=" exvalue]
 *     extype     = token / xtoken
 *     exvalue    = LDAPString from section 4.1.2 of [2]
 *     token      = oid from section 4.1 of [3]
 *     xtoken     = ("X-" / "x-") token
 *
 * For example,
 *
 *     ldap://ldap.itd.umich.edu/o=University%20of%20Michigan,c=US
 *
 *     ldap://host.com:6666/o=IMC,c=US??sub?(cn=Babs%20Jensen)
 */

final public class LdapURL {

    private String url = null;

    private String scheme = null;

    private String host = null;

    private int port = -1;

    private String DN = null;

    private String attributes = null;

    private String scope = null;

    private String filter = null;

    private String extensions = null;

    public LdapURL(String url) throws NamingException {
	this.url = url;

	try {
	    parse();
	} catch (MalformedURLException e) {
	    NamingException ne = new NamingException();
	    ne.setRootCause(e);
	    throw ne;
	}
    }

    void parse() throws MalformedURLException {
	// extract URL scheme

	int index = url.indexOf("://");
        if (index > -1)
            scheme = url.substring(0, index);

	if (!"ldap".equals(scheme)) {
	    throw new MalformedURLException("Not an LDAP URL: " + url);
	}

	// extract host and port

        index += 3; // skip over "://"

	int slash = url.indexOf('/', index);
	int colon = url.indexOf(':', index);
        if (slash < 0) {
            if (colon < 0) {
		host = url.substring(index);
	    } else {
		host = url.substring(index, colon);
		port = Integer.parseInt(url.substring(colon + 1));
	    }
	    return;	// done
	} else {
            if (colon < 0)
		host = url.substring(index, slash);
	    else {
		host = url.substring(index, colon);
		port = Integer.parseInt(url.substring(colon + 1, slash));
	    }
	}

	// extract DN

	int qmark = url.indexOf('?', slash);
        if (qmark < 0)
            DN = url.substring(slash + 1);
	else
            DN = url.substring(slash + 1, qmark);

	if ((DN != null) && (DN.length() > 0))
	    DN = UrlUtil.decode(DN);

	if (qmark < 0)
	    return;

	// extract attributes

	int qmark2 = url.indexOf('?', qmark + 1);
        if (qmark2 < 0) {
            attributes = url.substring(qmark + 1);
	    return;	// done
	} else if (qmark2 != (qmark + 1)) {
            attributes = url.substring(qmark + 1, qmark2);
	}

	// extract scope

	int qmark3 = url.indexOf('?', qmark2 + 1);
        if (qmark3 < 0) {
            scope = url.substring(qmark2 + 1);
	    return;	// done
	} else if (qmark3 != (qmark2 + 1)) {
            scope = url.substring(qmark2 + 1, qmark3);
	}

	// extract filter

	int qmark4 = url.indexOf('?', qmark3 + 1);
        if (qmark4 < 0)
            filter = url.substring(qmark3 + 1);
	else if (qmark4 != (qmark3 + 1))
            filter = url.substring(qmark3 + 1, qmark4);

	if ((filter != null) && (filter.length() > 0))
	    filter = UrlUtil.decode(filter);

	// extract extensions

	if (qmark4 != -1)
            extensions = url.substring(qmark4 + 1);

	if ((extensions != null) && (extensions.length() > 0))
	    extensions = UrlUtil.decode(extensions);
    }

    public String getScheme() {
	return scheme;
    }

    public String getHost() {
	return host;
    }

    public int getPort() {
	return port;
    }

    public String getDN() {
	return DN;
    }

    public String getAttributes() {
	return attributes;
    }

    public String getScope() {
	return scope;
    }

    public String getFilter() {
	return filter;
    }

    public String getExtensions() {
	return extensions;
    }

    public String toString() {
	return url;
    }

    static String toUrlString(String host, int port, String dn) {
	return "ldap://" + host + ":" + port + "/" + dn;
    }

    public static boolean hasQueryComponents(String url) {
	return (url.lastIndexOf('?') != -1);
    }

/*
    public static void main(String[] args) throws Exception {

	LdapURL url = new LdapURL(
	    "ldap://host.com:6666/o=IMC,c=US??sub?(cn=Babs%20Jensen)");

	System.out.println("Example LDAP URL: " + url.toString());
	System.out.println("  scheme: " + url.getScheme());
	System.out.println("    host: " + url.getHost());
	System.out.println("    port: " + url.getPort());
	System.out.println("      DN: " + url.getDN());
	System.out.println("   attrs: " + url.getAttributes());
	System.out.println("   scope: " + url.getScope());
	System.out.println("  filter: " + url.getFilter());
	System.out.println("  extens: " + url.getExtensions());
	System.out.println("");
    }
*/
}
