/*
 * @(#)ContextEnumerator.java	1.3 00/02/02
 *
 * Copyright 1999, 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
package com.sun.jndi.toolkit.dir;

import javax.naming.*;
import javax.naming.directory.SearchControls;
import java.util.*;

/**
  * A class for recursively enumerating the contents of a Context;
  *
  * @author Jon Ruiz
  */
final public class ContextEnumerator implements NamingEnumeration {

    private static boolean debug = false;
    private NamingEnumeration children = null;
    private Binding currentChild = null;
    private Context root;
    private ContextEnumerator currentChildEnum = null;
    private boolean currentChildExpanded = false;
    private boolean rootProcessed = false;
    private int scope = SearchControls.SUBTREE_SCOPE;
    private String contextName = "";

    public ContextEnumerator(Context context) throws NamingException {
	this(context, SearchControls.SUBTREE_SCOPE);
    }

    public ContextEnumerator(Context context, int scope) 
        throws NamingException {
	    // return this object except when searching single-level
	this(context, scope, "", scope != SearchControls.ONELEVEL_SCOPE);
   }

    public ContextEnumerator(Context context, int scope, String contextName,
			     boolean returnSelf) 
	throws NamingException {
	if(context == null) {
	    throw new IllegalArgumentException("null context passed");
	}

	root = context;

	// No need to list children if we're only searching object
	if (scope != SearchControls.OBJECT_SCOPE) {
	    children = context.listBindings("");
	}
	this.scope = scope;
	this.contextName = contextName;
	// pretend root is processed, if we're not supposed to return ourself
	rootProcessed = !returnSelf;
	prepNextChild();
    }

    public boolean hasMore() throws NamingException {
	if(scope!= SearchControls.OBJECT_SCOPE && hasMoreDescendants()) {
	    return true;
	} else {
	    return !rootProcessed;
	}
    }

    public boolean hasMoreElements() {
	try {
	    return hasMore();
	} catch (NamingException e) {
	    return false;
	}
    }

    public Object nextElement() {
	try {
	    return next();
	} catch (NamingException e) {
	    throw new NoSuchElementException(e.toString());
	}
    }

    public Object next() throws NamingException {

	if(scope != SearchControls.OBJECT_SCOPE && hasMoreDescendants()) {
	    return getNextDescendant();
	}

	if(!rootProcessed) {
	    rootProcessed = true;
	    return new Binding("", root.getClass().getName(),
			       root, true);
	}

	throw new NoSuchElementException();
    }

    public void close() throws NamingException {
	root = null;
    }
    
    private boolean hasMoreChildren() throws NamingException {
	return children != null && children.hasMore();
    }

    private Binding getNextChild() throws NamingException {
	Binding oldBinding = ((Binding)children.next());
	Binding newBinding = null;

	// if the name is relative, we need to add it to the name of this
	// context to keep it relative w.r.t. the root context we are
	// enumerating
	if(oldBinding.isRelative() && !this.contextName.equals("")) {
	    NameParser parser = this.root.getNameParser("");
	    Name newName = parser.parse(this.contextName);
	    newName.add(oldBinding.getName());
	    if(debug) {
		System.out.println("ContextEnumerator: adding " + newName);
	    }
	    newBinding = new Binding(newName.toString(),
				     oldBinding.getClassName(),
				     oldBinding.getObject(),
				     oldBinding.isRelative());
	} else {
	    if(debug) {
		System.out.println("ContextEnumerator: using old binding");
	    }
	    newBinding = oldBinding;
	}
		
	return newBinding;
    }

    private boolean hasMoreDescendants() throws NamingException {
	// if the current child is expanded, see if it has more lements
	if (currentChildExpanded && currentChildEnum.hasMore()) {
	    if(debug) {System.out.println("hasMoreDescendants returning " +
					  "true" ); }
	    return true;
	} else {
	    if(debug) {System.out.println("hasMoreDescendants returning " +
					  (currentChild != null) ); }
	    return currentChild != null;
	}
    }

    private Binding getNextDescendant() throws NamingException {
        Binding nextDesc = null;

	// if the current child is expanded, use it's enumerator
	if (scope == SearchControls.SUBTREE_SCOPE &&
	    currentChildExpanded) {
	    if(currentChildEnum.hasMore()) {
		nextDesc = (Binding)currentChildEnum.next();
	    } else {
		nextDesc = currentChild;
		prepNextChild();
	    }
	} else {

	    // simple case
	    if(debug) {System.out.println("getNextDescedant: simple case");}

	    nextDesc = currentChild;
	    prepNextChild();
	}

	return nextDesc;
    }

    private void prepNextChild() throws NamingException {
	if(hasMoreChildren()) {
	    try {
		currentChild = getNextChild();
	    } catch (NamingException e){
		if (debug) System.out.println(e);
		if (debug) e.printStackTrace();
	    }
	} else {
	    currentChild = null;
	    return;
	}

	if(scope == SearchControls.SUBTREE_SCOPE && 
	   currentChild.getObject() instanceof Context) {
	    currentChildEnum = new ContextEnumerator(
					  (Context)(currentChild.getObject()),
					  scope, currentChild.getName(),
					  false);
	    currentChildExpanded = true;
	    if(debug) {System.out.println("prepNextChild: expanded");}
	} else {
	    currentChildExpanded = false;
	    currentChildEnum = null;
	    if(debug) {System.out.println("prepNextChild: normal");}
	}
    }
}
