/*
 * @(#)LazySearchEnumerationImpl.java	1.3 00/02/02
 *
 * Copyright 1999, 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

/**
  * Given an enumeration of candidates, check whether each
  * item in enumeration satifies the given filter.
  * Each item is a Binding and the following is used to get its
  * attributes for used by the filter:
  *
  *   ((DirContext)item.getObject()).getAttributes("").
  * If item.getObject() is not an DirContext, the item is skipped
  *
  * The items in the enumeration are obtained one at a time as
  * items from the search enumeration are requested. 
  *
  * @author Rosanna Lee
  */

package com.sun.jndi.toolkit.dir;

import javax.naming.*;
import javax.naming.directory.*;
import javax.naming.spi.NamingManager;

import java.util.NoSuchElementException;
import java.util.Hashtable;

final public class LazySearchEnumerationImpl implements NamingEnumeration {
    private NamingEnumeration candidates;
    private SearchResult nextMatch = null;
    private SearchControls cons;
    private AttrFilter filter;
    private Context context;
    private Hashtable env;

    public LazySearchEnumerationImpl(NamingEnumeration candidates,
	AttrFilter filter, SearchControls cons) throws NamingException {
	    this.candidates = candidates;
	    this.filter = filter;

	    if(cons == null) {
		this.cons = new SearchControls();
	    } else {
		this.cons = cons;
	    }
    }

    public LazySearchEnumerationImpl(NamingEnumeration candidates,
	AttrFilter filter, SearchControls cons,
	Context ctx, Hashtable env) throws NamingException {
	    this.candidates = candidates;
	    this.filter = filter;
	    this.env = env;
	    this.context = ctx;

	    if(cons == null) {
		this.cons = new SearchControls();
	    } else {
		this.cons = cons;
	    }
    }
    
    public boolean hasMore() throws NamingException {
	// find and do not remove from list
	return findNextMatch(false) != null;
    }

    public boolean hasMoreElements() {
	try {
	    return hasMore();
	} catch (NamingException e) {
	    return false;
	}
    }

    public Object nextElement() {
	try {
	    return findNextMatch(true);
	} catch (NamingException e) {
	    throw new NoSuchElementException(e.toString());
	}
    }

    public Object next() throws NamingException {
	// find and remove from list
	return (findNextMatch(true));
    }

    public void close() throws NamingException {
	if (candidates != null) {
	    candidates.close();
	}
    }

    private SearchResult findNextMatch(boolean remove) throws NamingException {
	SearchResult answer;
	if (nextMatch != null) {
	    answer = nextMatch;
	    if (remove) {
		nextMatch = null;
	    }
	    return answer;
	} else {
	    // need to find next match
	    Binding next;
	    Object obj;
	    Attributes targetAttrs;
	    while (candidates.hasMore()) {
		next = (Binding)candidates.next();
		obj = next.getObject();
		if (obj instanceof DirContext) {
		    targetAttrs = ((DirContext)(obj)).getAttributes("");
		    if (filter.check(targetAttrs)) {
			if (cons.getReturningObjFlag()) {
			    try {
				// Give name only if context non-null,
				// otherewise, name will be interpreted relative
				// to initial context (not what we want)
				Name nm = (context != null ? 
				    new CompositeName(next.getName()) : null);
				obj = NamingManager.getObjectInstance(obj,
				    nm, context, env);
			    } catch (NamingException e) {
				throw e;
			    } catch (Exception e) {
				NamingException e2 = new NamingException(
		    "problem generating object using object factory");
				e2.setRootCause(e);
				throw e2;
			    }
			} else {
			    obj = null;
			}
			answer = new SearchResult(next.getName(),
			    next.getClassName(), obj,
			    SearchFilter.selectAttributes(targetAttrs, 
				cons.getReturningAttributes()),
			    true);
			if (!remove) 
			    nextMatch = answer;
			return answer;
		    }
		}
	    }
	    return null;
	}
    }
}
