/*
 * @(#)AbstractMixer.java	1.20 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;

import java.util.Vector;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Control;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.Line;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;

/**
 * Abstract Mixer.  Implements Mixer (with abstract methods) and specifies
 * some other common methods for use by our implementation.
 *
 * @version 1.20, 00/02/02
 * @author Kara Kytle
 */
abstract class AbstractMixer extends AbstractDataLine implements Mixer {

//	STATIC VARIABLES

	protected static final int PCM	= 0;
	protected static final int ULAW = 1;
	protected static final int ALAW = 2;


// IMMUTABLE PROPERTIES


	/**
	 * Info object describing this mixer.
	 */
	private final Mixer.Info mixerInfo;

	protected Line.Info[] sourceLineInfo;
	protected Line.Info[] targetLineInfo; 

	/**
	 * Supported formats for the mixer.
	 */
	protected Vector formats								= new Vector();


// STATE VARIABLES


	/**
	 * Source lines (ports) currently open
	 */
	protected Vector sourceLines = new Vector();


	/**
	 * Target lines currently open.
	 */
	protected Vector targetLines = new Vector();

	
		
	/**
	 * Constructs a new AbstractMixer.
	 * @param mixer the mixer with which this line is associated
	 * @param controls set of supported controls
	 */
	protected AbstractMixer(DataLine.Info info, 
							Control[] controls, 
							AudioFormat format,
							int bufferSize,
							Mixer.Info mixerInfo, 
							Line.Info[] sourceLineInfo, 
							Line.Info[] targetLineInfo) {

		super(info, null, controls, format, bufferSize);

		// setup the line part

		this.mixer = this;
		if (controls == null) {
			controls = new Control[0];
		}
	
		// setup the mixer part

		this.mixerInfo = mixerInfo;
		this.sourceLineInfo = sourceLineInfo;
		this.targetLineInfo = targetLineInfo;
	}


// MIXER METHODS


	public Mixer.Info getMixerInfo() {

		return mixerInfo;
	}


	public Line.Info[] getSourceLineInfo() {

		Line.Info[] localArray = new Line.Info[sourceLineInfo.length];
		System.arraycopy(sourceLineInfo, 0, localArray, 0, sourceLineInfo.length);
		return localArray;
	}
	

	public Line.Info[] getTargetLineInfo() {

		Line.Info[] localArray = new Line.Info[targetLineInfo.length];
		System.arraycopy(targetLineInfo, 0, localArray, 0, targetLineInfo.length);
		return localArray;
	}

	
	public Line.Info[] getSourceLineInfo(Line.Info info) {

		int i;
		Vector vec = new Vector();
		
		for (i = 0; i < sourceLineInfo.length; i++) {

			if (info.matches(sourceLineInfo[i])) {
				vec.addElement(sourceLineInfo[i]);
			}
		}

		Line.Info[] returnedArray = new Line.Info[vec.size()];
		for (i = 0; i < returnedArray.length; i++) {
			returnedArray[i] = (Line.Info)vec.elementAt(i);
		}

		return returnedArray;
	}

	
	public Line.Info[] getTargetLineInfo(Line.Info info) {

		int i;
		Vector vec = new Vector();
		
		for (i = 0; i < targetLineInfo.length; i++) {

			if (info.matches(targetLineInfo[i])) {
				vec.addElement(targetLineInfo[i]);
			}
		}

		Line.Info[] returnedArray = new Line.Info[vec.size()];
		for (i = 0; i < returnedArray.length; i++) {
			returnedArray[i] = (Line.Info)vec.elementAt(i);
		}

		return returnedArray;
	}


	public boolean isLineSupported(Line.Info info) {

		int i;

		for (i = 0; i < sourceLineInfo.length; i++) {

			if (info.matches(sourceLineInfo[i])) {
				return true;
			}
		}

		for (i = 0; i < targetLineInfo.length; i++) {

			if (info.matches(targetLineInfo[i])) {
				return true;
			}
		}

		return false;
	}

	
	public abstract Line getLine(Line.Info info) throws LineUnavailableException;

	public abstract int getMaxLines(Line.Info info);


	public Line[] getSourceLines() {

		Line[] localLines;

		synchronized(sourceLines) {

			localLines = new Line[sourceLines.size()];

			for (int i = 0; i < localLines.length; i++) {
				localLines[i] = (Line)sourceLines.elementAt(i);
			}
		}	

		return localLines;
	}


	public Line[] getTargetLines() {

		Line[] localLines;

		synchronized(targetLines) {

			localLines = new Line[targetLines.size()];

			for (int i = 0; i < localLines.length; i++) {
				localLines[i] = (Line)targetLines.elementAt(i);
			}
		}	

		return localLines;
	}


	/**
	 * Default implementation always throws an exception.
	 */
	public void synchronize(Line[] lines, boolean maintainSync) {
		throw new IllegalArgumentException("Synchronization not supported by this mixer.");
	}


	/**
	 * Default implementation always throws an exception.
	 */
	public void unsynchronize(Line[] lines) {
		throw new IllegalArgumentException("Synchronization not supported by this mixer.");
	}


	/**
	 * Default implementation always returns false.
	 */
	public boolean isSynchronizationSupported(Line[] lines, boolean maintainSync) {
		return false;
	}


// OVERRIDES OF ABSTRACT DATA LINE METHODS

	/**
	 * This implementation imposes the requirement that all mixers have the same
	 * format, for matching input and output device formats.
	 */
	public synchronized void open(AudioFormat format, int bufferSize) throws LineUnavailableException {

		if (Printer.trace) Printer.trace(">> AbstractMixer: open(format = " + format + "), bufferSize = " + bufferSize);

		AudioFormat systemFormat = getSystemFormat();

		// if it doesn't match the format currently established format for an open device.
		// throw an exception
		if ( (systemFormat != null) && (!format.matches(systemFormat)) ) {
			throw new LineUnavailableException("Requested format incompatible with already established device format: " + systemFormat);
		}

		// otherwise to for it
		super.open(format, bufferSize);
		if (Printer.trace) Printer.trace("<< AbstractMixer: open(format = " + format + "), bufferSize = " + bufferSize + " succeeded");
	}


	/**
	 * This implementation tries to open the mixer with its current format and buffer size settings.
	 * If this fails, it tries to open with the current "system format."
	 */
	public synchronized void open() throws LineUnavailableException {

		if (Printer.trace) Printer.trace(">> AbstractMixer: open()");

		try {

			// try with the current settings
			open(format, bufferSize);

		} catch (LineUnavailableException e) {

			AudioFormat systemFormat = null;

			// try opening with the system format, if set.  
			// this may still throw a LineUnavailableException
			if ((systemFormat = getSystemFormat()) != null) {		
				open(systemFormat, AudioSystem.NOT_SPECIFIED);
			} else {
				throw e;
			}
		}

		if (Printer.trace) Printer.trace("<< AbstractMixer: open() succeeded");
	}


	
// METHOD FOR INTERNAL IMPLEMENTATION USE

	
	/**
	 * The default implementation of this method just determines whether
	 * this line is a source or target line, calls open(no-arg) on the 
	 * mixer, and adds the line to the appropriate vector.  
	 * The mixer may be opened at a format different than the line's 
	 * format if it is a DataLine.
	 */
	protected void open(Line line) throws LineUnavailableException {

		if (Printer.trace) Printer.trace(">> AbstractMixer: open(line = " + line + ")");

		// $$kk: 06.11.99: ignore ourselves for now
		if (this.equals(line)) {
			return;
		}

		// source line?
		if (isSourceLine(line.getLineInfo())) {

			if (! sourceLines.contains(line) ) {
				
				// call the no-arg open method for the mixer; it should open at its
				// default format if it is not open yet
				open();

				// we opened successfully!  add the line to the list
				sourceLines.addElement(line);
			}
		} else {
			
			// target line?
			if(isTargetLine(line.getLineInfo())) {

				if (! targetLines.contains(line) ) {
					
					// call the no-arg open method for the mixer; it should open at its
					// default format if it is not open yet
					open();

					// we opened successfully!  add the line to the list
					targetLines.addElement(line);
				}
			} else {
				if (Printer.err) Printer.err("Unknown line received for AbstractMixer.open(Line): " + line);
			}
		} 

		if (Printer.trace) Printer.trace("<< AbstractMixer: open(" + line + ") completed");
	}


	
	/**
	 * The default implementation of this method just calls 
	 * open(Line line) with the DataLine parameter.
	 * The mixer may be opened at a format different than the DataLine's format.
	 */
	protected void open(DataLine line, AudioFormat format, int bufferSize) throws LineUnavailableException {

		if (Printer.trace) Printer.trace(">> AbstractMixer: open(line = " + line + ")");
		open(line);
		if (Printer.trace) Printer.trace("<< AbstractMixer: open(line, format, bufferSize) completed");
	}


	/**
	 * Removes this line from the list of open source lines and
	 * open target lines, if it exists in either.  
	 * If the list is now empty, closes the mixer.
	 */
	protected void close(Line line) {

		if (Printer.trace) Printer.trace(">> AbstractMixer: close(" + line + ")");

		// $$kk: 06.11.99: ignore ourselves for now
		if (this.equals(line)) {
			return;
		}

		sourceLines.removeElement(line);
		targetLines.removeElement(line);

		if (Printer.debug) Printer.debug("AbstractMixer: close(line): sourceLines.size() now: " + sourceLines.size());
		if (Printer.debug) Printer.debug("AbstractMixer: close(line): targetLines.size() now: " + targetLines.size());


		if ( sourceLines.isEmpty() && targetLines.isEmpty() ) {
			close();
		}

		if (Printer.trace) Printer.trace("<< AbstractMixer: close(" + line + ") succeeded");
	}


	/**
	 * Starts the mixer.
	 */
	protected void start(Line line) {

		if (Printer.trace) Printer.trace(">> AbstractMixer: start(" + line + ")");

		// $$kk: 06.11.99: ignore ourselves for now
		if (this.equals(line)) {
			return;
		}

		// we just start the mixer regardless of anything else here.
		if (!isStartedRunning()) {
			if (Printer.debug) Printer.debug("AbstractMixer: start(line): starting the mixer");
			start();
		}

		if (Printer.trace) Printer.trace("<< AbstractMixer: start(" + line + ") succeeded");
	}


	/**
	 * Stops the mixer if this was the last running line. 
	 */
	protected void stop(Line line) {

		if (Printer.trace) Printer.trace(">> AbstractMixer: stop(" + line + ")");

		// $$kk: 06.11.99: ignore ourselves for now
		if (this.equals(line)) {
			return;
		}

		Vector localSourceLines = (Vector)sourceLines.clone();
		for (int i = 0; i < localSourceLines.size(); i++) {

			// if any other open line is running, return

			// this covers clips and source data lines
			if (localSourceLines.elementAt(i) instanceof AbstractDataLine) {
				AbstractDataLine sourceLine = (AbstractDataLine)localSourceLines.elementAt(i);	
				if ( sourceLine.isStartedRunning() && (!sourceLine.equals(line)) ) {
					return;
				}
			}
		}

		Vector localTargetLines = (Vector)targetLines.clone();

		for (int i = 0; i < localTargetLines.size(); i++) {

			// if any other open line is running, return
			AbstractDataLine targetLine = (AbstractDataLine)localTargetLines.elementAt(i);	
			if ( targetLine.isStartedRunning() && (!targetLine.equals(line)) ) {
				if (Printer.trace) Printer.trace("<< AbstractMixer: stop(" + line + ") found running targetLine: " + targetLine);
				return;
			}
		}

		// otherwise, stop
		if (Printer.debug) Printer.debug("AbstractMixer: stop(line): stopping the mixer");
		stop();

		if (Printer.trace) Printer.trace("<< AbstractMixer: stop(" + line + ") succeeded");
	}


	/**
	 * Determines whether this is a source line for this mixer.
	 * Right now this just checks whether it's supported, but should
	 * check whether it actually belongs to this mixer....
	 */
	boolean isSourceLine(Line.Info info) {

		for (int i = 0; i < sourceLineInfo.length; i++) {
			if (info.matches(sourceLineInfo[i])) {
				return true;
			}
		}

		return false;
	}


	/**
	 * Determines whether this is a target line for this mixer.
	 * Right now this just checks whether it's supported, but should
	 * check whether it actually belongs to this mixer....
	 */
	boolean isTargetLine(Line.Info info) {

		for (int i = 0; i < targetLineInfo.length; i++) {
			if (info.matches(targetLineInfo[i])) {
				return true;
			}
		}

		return false;
	}


	/**
	 * Returns the first complete Line.Info object it finds that
	 * matches the one specified, or null if no matching Line.Info
	 * object is found.
	 */
	Line.Info getLineInfo(Line.Info info) {

		int i;

		// $$kk: 05.31.99: need to change this so that
		// the format and buffer size get set in the 
		// returned info object for data lines??
		
		for (i = 0; i < sourceLineInfo.length; i++) {

			if (info.matches(sourceLineInfo[i])) {
				return sourceLineInfo[i];
			}
		}

		for (i = 0; i < targetLineInfo.length; i++) {

			if (info.matches(targetLineInfo[i])) {
				return targetLineInfo[i];
			}
		}

		return null;
	}


	/**
	 * Get the current system format: the format of any open device.
	 * Used to match capture and playback device formats.
	 *
  	 * $$kk: 08.17.99: we will at some point need to change / improve on
	 * this model to support full-duplex when available on the host system.
	 */
	static AudioFormat getSystemFormat() {

		Mixer currentMixer;
		Mixer.Info[] mixerInfoList = AudioSystem.getMixerInfo();

		for (int i = 0; i < mixerInfoList.length; i++) {
			currentMixer = AudioSystem.getMixer(mixerInfoList[i]);
			if (currentMixer.isOpen()) {				
				if (currentMixer instanceof DataLine) {
					return ((DataLine)currentMixer).getFormat();
				}
			}	
		}

		return null;		
	}
}
