/*
 * @(#)AiffFileReader.java	1.24 00/02/02
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;	  	 

import java.util.Vector;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.EOFException;
import java.net.URL;
import java.net.MalformedURLException;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.SequenceInputStream;

import javax.sound.sampled.AudioFileFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.UnsupportedAudioFileException;


/**
 * AIFF file reader and writer.
 *
 * @version 1.24 00/02/02
 * @author Kara Kytle
 * @author Jan Borgersen
 */
public class AiffFileReader extends SunFileReader  {


	/**
	 * AIFF parser type
	 */
	//public static final FileStream.FileType type = new FileStream.FileType("AIFF");
	public static final AudioFileFormat.Type types[] = {
		AudioFileFormat.Type.AIFF
	};

	
	/**
	 * Constructs a new AiffParser object.
	 */
	public AiffFileReader() {
	}




// METHODS TO IMPLEMENT AudioFileReader

	/**
	 * Obtains the audio file format of the input stream provided.  The stream must
	 * point to valid audio file data.  In general, audio file providers may 
	 * need to read some data from the stream before determining whether they
	 * support it.  These parsers must
	 * be able to mark the stream, read enough data to determine whether they 
	 * support the stream, and, if not, reset the stream's read pointer to its original 
	 * position.  If the input stream does not support this, this method may fail
	 * with an IOException. 
	 * @param stream the input stream from which file format information should be
	 * extracted
	 * @return an <code>AudioFileFormat</code> object describing the audio file format
	 * @throws UnsupportedAudioFileException if the stream does not point to valid audio
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 * @see InputStream#markSupported
	 * @see InputStream#mark
	 */
	public AudioFileFormat getAudioFileFormat(InputStream stream) throws UnsupportedAudioFileException, IOException {

		AudioFormat format = null;
		AiffFileFormat fileFormat = null;
		int maxReadLength = 8;
		int magic = 0;
		int length = 0;
		int totallength = 0;

		DataInputStream dis = new DataInputStream( stream );
		
		dis.mark(maxReadLength);

		magic = dis.readInt();
		length = dis.readInt();

		if( magic!=AiffFileFormat.AIFF_MAGIC ) {

			// not AIFF, rewind stream, throw exception
			dis.reset();
			throw new UnsupportedAudioFileException("not an AIFF file");
		}

		if(length <= 0 ) {
		    length = AudioSystem.NOT_SPECIFIED;
		    totallength = AudioSystem.NOT_SPECIFIED;
		} else {
		    totallength = length + 8;
		}

		// we can't determine the full format without reading a lot further
		// to find the 'fmt' chunk, and possibly messing up our ability to 
		// reset the input stream.  so, fill in default values here.

		format = new AudioFormat( AudioSystem.NOT_SPECIFIED,    // sampleRate
					AudioSystem.NOT_SPECIFIED,      // sampleSizeInBits
					AudioSystem.NOT_SPECIFIED,      // channels
					true,                           // signed
					true );                         // bigEndian

		fileFormat = new AiffFileFormat( AudioFileFormat.Type.AIFF,
								totallength,
								format,
								AudioSystem.NOT_SPECIFIED);
		dis.reset(); // Throws IOException
		return fileFormat;
	}


	/**
	 * Obtains the audio file format of the URL provided.  The URL must
	 * point to valid audio file data.
	 * @param url the URL from which file format information should be
	 * extracted
	 * @return an <code>AudioFileFormat</code> object describing the audio file format
	 * @throws UnsupportedAudioFileException if the URL does not point to valid audio
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 */
	public AudioFileFormat getAudioFileFormat(URL url) throws UnsupportedAudioFileException, IOException {

		InputStream				urlStream = null;
		BufferedInputStream		bis = null;
		DataInputStream			dis = null;
		AudioFileFormat			fileFormat = null;
		AudioFormat				format = null;

		urlStream = url.openStream();	// throws IOException

		bis = new BufferedInputStream( urlStream, bisBufferSize );

		fileFormat = getAudioFileFormat( bis );		// throws UnsupportedAudioFileException

		dis = new DataInputStream( bis );
		fileFormat = getCOMM( dis, fileFormat );

		
		bis.close();

		return fileFormat;
	}


	/**
	 * Obtains the audio file format of the File provided.  The File must
	 * point to valid audio file data.
	 * @param file the File from which file format information should be
	 * extracted
	 * @return an <code>AudioFileFormat</code> object describing the audio file format
	 * @throws UnsupportedAudioFileException if the File does not point to valid audio
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 */
	public AudioFileFormat getAudioFileFormat(File file) throws UnsupportedAudioFileException, IOException {

		FileInputStream			fis = null;
		BufferedInputStream		bis = null;
		DataInputStream			dis = null;
		AudioFileFormat			fileFormat = null;
		AudioFormat				format = null;

		fis = new FileInputStream( file );	// throws IOException

		bis = new BufferedInputStream( fis, bisBufferSize );

		fileFormat = getAudioFileFormat( bis );		// throws UnsupportedAudioFileException

		dis = new DataInputStream( bis );
		fileFormat = getCOMM( dis, fileFormat );
		
		bis.close();

		return fileFormat;
	}




	/**
	 * Obtains an audio stream from the input stream provided.  The stream must
	 * point to valid audio file data.  In general, audio file providers may 
	 * need to read some data from the stream before determining whether they
	 * support it.  These parsers must
	 * be able to mark the stream, read enough data to determine whether they 
	 * support the stream, and, if not, reset the stream's read pointer to its original 
	 * position.  If the input stream does not support this, this method may fail
	 * with an IOException. 
	 * @param stream the input stream from which the <code>AudioInputStream</code> should be
	 * constructed
	 * @return an <code>AudioInputStream</code> object based on the audio file data contained
	 * in the input stream.
	 * @throws UnsupportedAudioFileException if the stream does not point to valid audio
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 * @see InputStream#markSupported
	 * @see InputStream#mark
	 */
	public AudioInputStream getAudioInputStream(InputStream stream) throws UnsupportedAudioFileException, IOException {

		DataInputStream dis = null;
		int nread = 0;
		int length;
		int datahdr;
		int i;
		int blocksize;
		AudioFileFormat fileFormat = null;


		fileFormat = getAudioFileFormat( stream ); // throws UnsupportedAudioFileException, IOException

		// if we passed this call, we have an AIFF file.  look for the 'COMM' chunk
		// getCOMM leaves the input stream at the beginning of the audio data

		dis = new DataInputStream( stream );

		fileFormat = getCOMM( dis, fileFormat );	// throws UnsupportedAudioFileException, IOException
		
		// we've got everything, and the stream is at the 
		// beginning of the audio data, so return an AudioInputStream.

		// $$jb: GET THE LENGTH!!!!
		return new AudioInputStream(dis, fileFormat.getFormat(), fileFormat.getFrameLength() );
	}


	/**
	 * Obtains an audio stream from the URL provided.  The URL must
	 * point to valid audio file data.
	 * @param url the URL for which the <code>AudioInputStream</code> should be
	 * constructed
	 * @return an <code>AudioInputStream</code> object based on the audio file data pointed
	 * to by the URL
	 * @throws UnsupportedAudioFileException if the URL does not point to valid audio
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 */
	public AudioInputStream getAudioInputStream(URL url) throws UnsupportedAudioFileException, IOException {

		InputStream				urlStream = null;
		BufferedInputStream		bis = null;
		AudioFileFormat			fileFormat = null;

		urlStream = url.openStream();	// throws IOException

		bis = new BufferedInputStream( urlStream, bisBufferSize );

		return getAudioInputStream( (InputStream)bis );
	}


	/**
	 * Obtains an audio stream from the File provided.  The File must
	 * point to valid audio file data.
	 * @param file the File for which the <code>AudioInputStream</code> should be
	 * constructed
	 * @return an <code>AudioInputStream</code> object based on the audio file data pointed
	 * to by the File
	 * @throws UnsupportedAudioFileException if the File does not point to valid audio
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 */
	public AudioInputStream getAudioInputStream(File file) throws UnsupportedAudioFileException, IOException {

		FileInputStream			fis = null;
		BufferedInputStream		bis = null;
		AudioFileFormat			fileFormat = null;

		fis = new FileInputStream( file );	// throws IOException

		bis = new BufferedInputStream( fis, bisBufferSize );

		return getAudioInputStream( (InputStream)bis );
	}

//--------------------------------------------------------------------

	private AudioFileFormat getCOMM(DataInputStream dis, AudioFileFormat fileFormat) throws UnsupportedAudioFileException, IOException {

		// assumes a stream at the beginning of the file which has already 
		// passed the magic number test...
		// leaves the input stream at the beginning of the audio data

		AudioFormat format;
		AudioFormat.Encoding encoding = null;
		float	sampleRate			= -1;
		int		sampleSizeInBits	= -1;
		int		channels			= -1;
		int		frameSize			= -1;
		float	frameRate			= -1;
		int		bigendian			= 1;	// default to big for AIFF
		int length;
		int offset;
		int nread = 0;

		boolean aifc = false;
		byte bytes[] = new byte[4];
		String s;
		int chunksize, i;

		// Read the magic number
		int magic = dis.readInt();
		nread += 4;
	 
		length = dis.readInt(); nread += 4;
		dis.read(bytes, 0, 4); nread += 4;
		s = new String( bytes );

		// Is this an AIFC or just plain AIFF file.
		if (s.startsWith("AIFC")) {
			aifc = true;
		}

		// Loop through the AIFF chunks until
		// we get to the COMM chunk.
		while (true) {
	     
			// $$jb: 04.22.99:  This solves the problem of not being able to play
			// AIFF's on jdk1.1.x .  It seems that in 1.1.x, dis.read(bytes,0,4) will 
			// only read 4 bytes if those 4 bytes don't span a buffer boundary as
			// defined above in BufferedInputStream.  There were cases when we would
			// only read the first 2 bytes, and none of the additional parsing would
			// work.  dis.readInt() seems to work without problems.

			// Read the chunk name
			//dis.read(bytes, 0, 4 ); nread += 4;
			//s = new String( bytes );
			int temp = dis.readInt();
			bytes[0] = (byte)((temp>>24) & 0xFF);
			bytes[1] = (byte)((temp>>16) & 0xFF);
			bytes[2] = (byte)((temp>>8) & 0xFF);
			bytes[3] = (byte)(temp & 0xFF);
			s = new String( bytes );


			// Switch on the chunk name.
			if (s.startsWith("FVER")) {
				// Ignore format version for now.
				dis.readLong(); nread += 8;
		 
			} else if ( s.startsWith("COMM")) {
		 
				chunksize = dis.readInt(); nread += 4;
		 
				// AIFF vs. AIFFC
				if (( chunksize != 18) && (chunksize != 34 )) {
					throw new UnsupportedAudioFileException("Invalid AIFF chunksize");
				}

				// Read header info.
				channels = dis.readShort(); nread += 2;
				dis.readInt(); nread += 4;
				sampleSizeInBits = dis.readShort(); nread += 2;
				sampleRate = (float) read_ieee_extended (dis); nread += 10;

				// If this is not AIFC then we assume it's
				// a linearly encoded file.
				//encoding = AudioFormat.PCM_SIGNED_BIG_ENDIAN;
				encoding = AudioFormat.Encoding.PCM_SIGNED;

				if (aifc) {
					dis.read(bytes, 0, 4 ); nread += 4;
					s = new String (bytes);
				 
					if (s.startsWith("NONE")) 
						//encoding = AudioFormat.PCM_SIGNED_BIG_ENDIAN;
						encoding = AudioFormat.Encoding.PCM_SIGNED;
				 	else if ( s.startsWith("ULAW"))
						encoding = AudioFormat.Encoding.ULAW;
					else if ( s.startsWith("ALAW"))
						encoding = AudioFormat.Encoding.ALAW;
					else {
						throw new UnsupportedAudioFileException("Invalid AIFF encoding");
					}
				}
			} else if ( s.startsWith("SSND")) {
				// Data chunk.
				// we are getting *weird* numbers for chunksize here sometimes;
				// this really should be the size of the data chunk....
				chunksize = dis.readInt(); nread += 4;

				// Data offset
				i = dis.readInt(); nread += 4;
				// Block size
				int blocksize = dis.readInt(); nread += 4;

                 // okay, now we are done reading the header.  we need to set the size
                 // of the data segment.  we know that sometimes the value we get for
                 // the chunksize is absurd.  this is the best i can think of:if the
                 // value seems okay, use it.  otherwise, we get our value of 
                 // length by assuming that everything left is the data segment;
                 // its length should be our original length (for all AIFF data chunks)
                 // minus what we've read so far.  
                 // $$kk: we should be able to get length for the data chunk right after
                 // we find "SSND."  however, some aiff files give *weird* numbers.  what
                 // is going on??

                 if (chunksize < length) { 
                    
					// $$kk: 11.03.98: need to subtract the 8 bytes we just read
					length = chunksize - 8;

				} else {

					// $$kk: 11.03.98: this seems dangerous!
					length -= nread;
				}

				break;
			} else {
		 
				// We're in no-mans land. Just read past
				// the chunk.

				chunksize = dis.readInt(); nread += 4;
				nread += dis.skipBytes(chunksize);
			} // if
		} // while


		// Set the sample size & signed/unsigned.
		// $$kk: what's going on with the signed/unsigned bit here??
		// $$kk: 09.01.98: PCM samples in AIFF files are always *signed* (see Kientzle p. 229). 
		// $$jb: 03.22.99: signed is part of AudioFormat.Encoding now.  took care of this up
		//       above, when we determine the encoding.
		// signed = AudioFormat.SIGNED;
		switch (sampleSizeInBits) {
			case 8:
			 break;
			 case 16:
			 break;
			 default:
			 // $jdr: What should we do here?
		}			 

		// set the offset; data should begin this number of bytes from the beginning of the file								
		offset = nread;

		frameSize = (sampleSizeInBits/8)*channels;
		frameRate = sampleRate;

		if( sampleSizeInBits == 8 )
			encoding = AudioFormat.Encoding.PCM_SIGNED;

		format =  new AudioFormat(encoding, sampleRate, sampleSizeInBits, channels, frameSize, frameRate, true);

		return new AiffFileFormat(AudioFileFormat.Type.AIFF, 
				fileFormat.getByteLength(), format, length / frameSize);


	}
	
// HELPER METHODS
   /** write_ieee_extended(DataOutputStream dos, double f) throws IOException {
     * Extended precision IEEE floating-point conversion routine.
	 * @argument DataOutputStream
	 * @argument double
	 * @return void
	 * @exception IOException
	 */
	private void write_ieee_extended(DataOutputStream dos, double f) throws IOException {

		int exponent = 16398;
		double highMantissa = f;

		// For now write the integer portion of f
		// $$jb: 03.30.99: stay in synch with JMF on this!!!!
		while (highMantissa < 44000) {
			highMantissa *= 2;
			exponent--;
		}
		dos.writeShort(exponent);
		dos.writeInt( ((int) highMantissa) << 16);
		dos.writeInt(0); // low Mantissa
	}


   /**
     * read_ieee_extended
     * Extended precision IEEE floating-point conversion routine.
     * @argument DataInputStream
     * @return double
     * @exception IOException
     */
    private double read_ieee_extended(DataInputStream dis) throws IOException {

	double f = 0;
        int expon = 0;
        long hiMant = 0, loMant = 0;
        long t1, t2;
	double HUGE = ((double)3.40282346638528860e+38);


	expon = dis.readUnsignedShort();
	
	t1 = (long)dis.readUnsignedShort();
	t2 = (long)dis.readUnsignedShort();
	hiMant = t1 << 16 | t2; 
	
	t1 = (long)dis.readUnsignedShort();
	t2 = (long)dis.readUnsignedShort();
	loMant = t1 << 16 | t2; 

	if (expon == 0 && hiMant == 0 && loMant == 0) {
	    f = 0;
	} else {
	    if (expon == 0x7FFF)
		f = HUGE;
	    else {
		expon -= 16383;
		expon -= 31;
		f = (hiMant * Math.pow(2, expon));
		expon -= 32;
		f += (loMant * Math.pow(2, expon));
	    }
	}

        return f;
    }



}

