/*
 * @(#)EventDispatcher.java	1.15 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;

import java.util.EventObject;
import java.util.Vector;

import javax.sound.sampled.Line;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;

import javax.sound.midi.MetaMessage;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.MetaEventListener;
import javax.sound.midi.ControllerEventListener;

/**
 * EventDispatcher.  Used by various classes used in the Java Sound implementation
 * to send events.  
 * 
 * $$kk: 10.30.98: Should change this to send events on another thread!
 *
 * @version 1.15 00/02/02
 * @author David Rivas
 * @author Kara Kytle
 */
class EventDispatcher extends Thread {

	
	/**
	 * List of events
	 */
    private Vector eventQueue = new Vector();
    

	/**
	 * True if thread should exit
	 */
	private boolean done = false;


	/**
	 * Invoked when there is at least one event in the queue.
	 * Implement this as a callback to process one event.
	 */
    protected void processEvent(EventInfo eventInfo) {

		// process an LineEvent

		if (eventInfo.getEvent() instanceof LineEvent) {

			LineEvent event = (LineEvent)eventInfo.getEvent();
			Vector currentListeners = eventInfo.getListeners();

			for (int i = 0; i < currentListeners.size(); i++) {
				((LineListener)currentListeners.elementAt(i)).update(event);
			}

			return;
		}

		// process a MetaMessage
		
		if (eventInfo.getEvent() instanceof MetaMessage) {

			MetaMessage event = (MetaMessage)eventInfo.getEvent();
			Vector currentListeners = eventInfo.getListeners();

			for (int i = 0; i < currentListeners.size(); i++) {
				((MetaEventListener)currentListeners.elementAt(i)).meta(event);
			}

			return;
		}
		
		// process a Controller or Mode Event

		if (eventInfo.getEvent() instanceof ShortMessage) {

			ShortMessage event = (ShortMessage)eventInfo.getEvent();
			int status = event.getStatus();

			// Controller and Mode events have status byte 0xBc, where
			// c is the channel they are sent on.  

			if( (status>>4)==11 ) { 

				Vector currentListeners = eventInfo.getListeners();

				for (int i = 0; i < currentListeners.size(); i++) {
					((ControllerEventListener)currentListeners.elementAt(i)).controlChange(event);
				}
			}
			return;
		}


/*
		// process a MidiDeviceEvent
		
		if (eventInfo.getEvent() instanceof MidiDeviceEvent) {

			MidiDeviceEvent event = (MidiDeviceEvent)eventInfo.getEvent();
			Vector currentListeners = eventInfo.getListeners();

			for (int i = 0; i < currentListeners.size(); i++) {
				((MidiDeviceListener)currentListeners.elementAt(i)).update(event);
			}

			return;
		}
*/

		Printer.err("Unknown event type: " + eventInfo.getEvent());
	}


	/**
	 * Wait until there is something in the event queue to process.  Then
	 * dispatch the event to the listeners.The entire method does not
	 * need to be synchronized since this includes taking the event out
	 * from the queue and processing the event. We only need to provide
	 * exclusive access over the code where an event is removed from the
	 *queue. 
     */
	protected void dispatchEvents() {

		EventInfo eventInfo = null;
		
		synchronized (this) {
			
			// Wait till there is an event in the event queue.
			try {

				while ( (!done) && (eventQueue.size() == 0) ) {
					wait();
				}

			} catch (InterruptedException e) {

				return;
			}

			if (!done) {

				// Remove the event from the queue and dispatch it to the listeners.
				eventInfo = (EventInfo)eventQueue.elementAt(0);
				eventQueue.removeElementAt(0);
			}

		} // end of synchronized

		if (eventInfo != null) {

			processEvent(eventInfo);
		}
	}
    
	/**
	 * Queue the given event in the event queue.
	 */
	synchronized void postEvent(EventInfo eventInfo) {
	
		eventQueue.addElement(eventInfo);
		notifyAll();
	}


    /**
     * Stop the thread.
	 * $$kk: 12.21.98: we never call this -- but we should!!
     */
    synchronized void kill() {
		
		done = true;
		notifyAll();
    }

    
	/**
     * A loop to dispatch events.
     */
	public void run() {
		
		while (!done) {

			dispatchEvents();
		}
	}


	/**
	 * Send audio events.
	 */
	void sendAudioEvents(EventObject event, Vector listeners) {

		sendAudioEvents( (Object) event, listeners );
	}

	void sendAudioEvents(Object event, Vector listeners) {

		Vector currentListeners = (Vector)listeners.clone();

		EventInfo eventInfo = new EventInfo(event, currentListeners);
		postEvent(eventInfo);
	}


	/**
	 * Container for an event and a set of listeners to deliver it to.
	 */
	class EventInfo {

		private Object event;
		private Vector listeners;

		EventInfo(Object event, Vector listeners) {

			this.event = event;
			this.listeners = listeners;
		}

		Object getEvent() {

			return event;
		}

		Vector getListeners() {

			return listeners;
		}
	} // class EventInfo

} // class EventDisapatcher
