/*
 * @(#)HeadspaceMixer.java	1.47 00/02/02
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;	  

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import java.util.Vector;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.Control;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.EnumControl;
import javax.sound.sampled.Line;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.ReverbType;
import javax.sound.sampled.SourceDataLine;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiUnavailableException;


/**
 * Represents the Headspace mixer.  This mixer supports the
 * Java Sound Mixer interface, and a set of mixer-specific
 * configuration methods.
 *
 * @version 1.47, 00/02/02
 * @author Kara Kytle
 */
class HeadspaceMixer extends AbstractMixer {


// INTERPOLATION DEFINES
// (note that these match the headspace TerpMode values.)

	/**
	 * Represents drop-sample interpolation.
	 */
	public static final int INTERPOLATION_DROP_SAMPLE			= 0;


	/**
	 * Represents two-point interpolation.
	 */
	public static final int INTERPOLATION_2_POINT				= 1;


	/**
	 * Represents linear interpolation.
	 */
	public static final int INTERPOLATION_LINEAR				= 2;


// FINAL PROPERTIES
	
	/**
	 * Total voices (midi + sampled)
	 * $$kk: 03.06.98: should maybe get this value from the engine via a native method
	 */
	private static int TOTAL_VOICES;

	/**
	 * This is the default (and for now unchangeable) buffer size in SAMPLE FRAMES 
	 * for the mixer.  it's the (slice size) * (number of slices).  
	 */
	private static int DEFAULT_BUFFER_SIZE = AudioSystem.NOT_SPECIFIED;

	/**
	 * This is the maximum length of the data we can handle for a Clip, in sample frames
	 */
	static final int MAX_SAMPLES = 1048576;

	/**
	 * max bytes per frame
	 */
	static final int MAX_FRAME_SIZE = 4;


	/**
	 * 8-bit encoding for output
	 */
	//private static final AudioFormat.Encoding encoding8 = (Platform.isSigned8() ? AudioFormat.Encoding.PCM_SIGNED : AudioFormat.Encoding.PCM_UNSIGNED);
	private static final AudioFormat.Encoding encoding8 = (Platform.isSigned8() ? AudioFormat.Encoding.PCM_SIGNED : AudioFormat.Encoding.PCM_UNSIGNED);

	/**
	 * 16-bit encoding for output
	 */
	//private static final AudioFormat.Encoding encoding16 = (Platform.isBigEndian() ? AudioFormat.Encoding.PCM_SIGNED_BIG_ENDIAN : AudioFormat.Encoding.PCM_SIGNED_LITTLE_ENDIAN);	
	private static final AudioFormat.Encoding encoding16 = AudioFormat.Encoding.PCM_SIGNED;

	
	/**
	 * Supported mix formats
	 */
	private static final AudioFormat[] mixFormats		= { 

															// $$kk: 06.08.99: should have 8k here??

															// 11k
															
															(new AudioFormat(encoding8, 11025, 8, 1, 1, 11025, false)),
															(new AudioFormat(encoding8, 11025, 8, 2, 2, 11025, false)),
															(new AudioFormat(encoding16, 11025, 16, 1, 2, 11025, false)),
															(new AudioFormat(encoding16, 11025, 16, 2, 4, 11025, false)),

															(new AudioFormat(encoding8, 11025, 8, 1, 1, 11025, true)),
															(new AudioFormat(encoding8, 11025, 8, 2, 2, 11025, true)),
															(new AudioFormat(encoding16, 11025, 16, 1, 2, 11025, true)),
															(new AudioFormat(encoding16, 11025, 16, 2, 4, 11025, true)),

															// 22k

															(new AudioFormat(encoding8, 22050, 8, 1, 1, 22050, false)),
															(new AudioFormat(encoding8, 22050, 8, 2, 2, 22050, false)),
															(new AudioFormat(encoding16, 22050, 16, 1, 2, 22050, false)),
															(new AudioFormat(encoding16, 22050, 16, 2, 4, 22050, false)),

															(new AudioFormat(encoding8, 22050, 8, 1, 1, 22050, true)),
															(new AudioFormat(encoding8, 22050, 8, 2, 2, 22050, true)),
															(new AudioFormat(encoding16, 22050, 16, 1, 2, 22050, true)),
															(new AudioFormat(encoding16, 22050, 16, 2, 4, 22050, true)),

															// 44k

															(new AudioFormat(encoding8, 44100, 8, 1, 1, 44100, false)),
															(new AudioFormat(encoding8, 44100, 8, 2, 2, 44100, false)),
															(new AudioFormat(encoding16, 44100, 16, 1, 2, 44100, false)),
															(new AudioFormat(encoding16, 44100, 16, 2, 4, 44100, false)),

															(new AudioFormat(encoding8, 44100, 8, 1, 1, 44100, true)),
															(new AudioFormat(encoding8, 44100, 8, 2, 2, 44100, true)),
															(new AudioFormat(encoding16, 44100, 16, 1, 2, 44100, true)),
															(new AudioFormat(encoding16, 44100, 16, 2, 4, 44100, true))

															// $$kk: 03.24.99: should have 48k here??
															  
														   };

	
	/**
	 * Supported source line (source data line and clip) formats
	 */
	private static final AudioFormat[] sourceLineFormats	= { 
															(new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, 
														  (float)AudioSystem.NOT_SPECIFIED, 8, 1, 1, (float)AudioSystem.NOT_SPECIFIED, false)),
															(new AudioFormat(AudioFormat.Encoding.PCM_UNSIGNED, 
														  (float)AudioSystem.NOT_SPECIFIED, 8, 1, 1, (float)AudioSystem.NOT_SPECIFIED, false)),
															(new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, 
														  (float)AudioSystem.NOT_SPECIFIED, 8, 2, 2, (float)AudioSystem.NOT_SPECIFIED, false)),
															(new AudioFormat(AudioFormat.Encoding.PCM_UNSIGNED, 
														  (float)AudioSystem.NOT_SPECIFIED, 8, 2, 2, (float)AudioSystem.NOT_SPECIFIED, false)),							
															
															(new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, 
														  (float)AudioSystem.NOT_SPECIFIED, 16, 1, 2, (float)AudioSystem.NOT_SPECIFIED, true)),
															(new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, 
														  (float)AudioSystem.NOT_SPECIFIED, 16, 1, 2, (float)AudioSystem.NOT_SPECIFIED, false)),
															(new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, 
														  (float)AudioSystem.NOT_SPECIFIED, 16, 2, 4, (float)AudioSystem.NOT_SPECIFIED, true)),
															(new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, 
														  (float)AudioSystem.NOT_SPECIFIED, 16, 2, 4, (float)AudioSystem.NOT_SPECIFIED, false))															  
															};

	private DataLine.Info sourceDataLineInfo = new DataLine.Info(SourceDataLine.class, 
												sourceLineFormats, 
												0, 
												AudioSystem.NOT_SPECIFIED);

	// the engine supports clips with up to 1M *samples* (not bytes) of data.
	// our buffer size is measured in bytes.  i am using a max buffer size of
	// ((MAX_SAMPLES == 1048576) * 2 channels * 2 bytes) because this is the
	// largest buffer size we can ever support.  unfortunately, we can generate
	// IllegalArgumentExceptions if the buffer has too many *samples* but not
	// too many *bytes*!!
	// (when we update code in the engine, we will get away from this restriction.) 
	private final DataLine.Info clipInfo = new DataLine.Info(Clip.class,
												sourceLineFormats, 
												0, 
												(MAX_SAMPLES * MAX_FRAME_SIZE));

	private final DataLine.Info[] supportedInGroup = { sourceDataLineInfo, clipInfo }; 

	// $$kk: 05.31.99: need to figure this out!
	private final DataLine.Info targetDeviceLineInfo = new DataLine.Info(DataLine.class,
												mixFormats,
												0,
												AudioSystem.NOT_SPECIFIED);


// DEFAULT STATE

	private static final AudioFormat defaultFormat		= (mixFormats.length > 0) ? mixFormats[mixFormats.length-1] : null;

	private static final int defaultMixLevel			= 6;

	// $$kk: 03.11.99: check this default
	private static final int defaultInterpolationMode	= INTERPOLATION_LINEAR;


// CURRENT STATE
	
	// number of voices allocated to sampled audio streams
	private int sampledVoices;

	// number of voices allocated to MIDI voices
	private int midiVoices;
																	
	// mix level
	private int mixLevel;

	// interpolation mode
	private int terpMode;  

	// list of lines open on the mixer, to manage auto-opening and closing.
	Vector openMidiDevices = new Vector();

// DEFAULT MIXER INSTANCE (right now we only support one instance)

	private static final HeadspaceMixer mixerInstance					= new HeadspaceMixer();


	/**
 	 * Private constructor.
	 */
	private HeadspaceMixer() { 
			
		// $$kk: 05.31.99: we unfortunately cannot fill the mixer and
		// control values in until after calling the superclass constructor....
		// $$kk: 05.31.99: need to set the info object value
		super(null,
			new Control[1],
			defaultFormat,
			DEFAULT_BUFFER_SIZE,
			new MixerInfo(),
			null,
			null);

		//this.mixer = mixerInstance;

		// set the AbstractLine.info value.  this describes the internal Line 
		// $$kk: 06.03.99: should rename!!
		info = new DataLine.Info(DataLine.class, 
									mixFormats, 
									DEFAULT_BUFFER_SIZE, 
									DEFAULT_BUFFER_SIZE);

		if (Printer.trace) Printer.trace("HeadspaceMixer: constructor");

		// initialize platform-specific values, and load the native library
		Platform.initialize();

		TOTAL_VOICES = nGetTotalVoices();

		// $$kk: 06.05.99: this won't work 'cause we haven't initialized 
		// the mixer yet.
		DEFAULT_BUFFER_SIZE = nGetDefaultBufferSize();
		if (DEFAULT_BUFFER_SIZE == 0) {
			DEFAULT_BUFFER_SIZE = AudioSystem.NOT_SPECIFIED;
		} 

		// set up the controls
		controls[0] = new MixerReverbControl(); 

		
		// set up the mixer source line info
		sourceLineInfo = new Line.Info[2];
		sourceLineInfo[0] = sourceDataLineInfo;
		sourceLineInfo[1] = clipInfo; 
		
		// set up the mixer target line info
		targetLineInfo = new DataLine.Info[1];
		targetLineInfo[0] = targetDeviceLineInfo;

		sampledVoices = TOTAL_VOICES / 2;
		midiVoices = TOTAL_VOICES / 2;
		mixLevel = defaultMixLevel;
		terpMode = defaultInterpolationMode;

		if (Printer.trace) Printer.trace("HeadspaceMixer: constructor completed");
	}


// ABSTRACT MIXER: ABSTRACT METHOD IMPLEMENTATIONS

	
	public Line getLine(Line.Info info) throws LineUnavailableException{

		Line.Info fullInfo = getLineInfo(info);

		if (fullInfo == null) {
			throw new IllegalArgumentException("Line unsupported: " + info);
		}
	 
		if (fullInfo instanceof DataLine.Info) {
			
			DataLine.Info dataLineInfo = (DataLine.Info)fullInfo;
			
			AudioFormat lineFormat;
			int lineBufferSize = AudioSystem.NOT_SPECIFIED;

			// if a format is specified by the info class passed in, use it.
			// otherwise use a format from fullInfo.

			AudioFormat[] supportedFormats = null;
				
			if (info instanceof DataLine.Info) {
				supportedFormats = ((DataLine.Info)info).getFormats();
				lineBufferSize = ((DataLine.Info)info).getMaxBufferSize();
			}

			if ((supportedFormats == null) || (supportedFormats.length == 0)) {				
				supportedFormats = dataLineInfo.getFormats();
			}

			lineFormat = supportedFormats[supportedFormats.length-1];

			// if a rate is not specified, match the mixer sample rate
			if (lineFormat.getSampleRate() == (float)AudioSystem.NOT_SPECIFIED) {
				lineFormat = new AudioFormat(lineFormat.getEncoding(), getFormat().getSampleRate(), lineFormat.getSampleSizeInBits(), lineFormat.getChannels(), 
								lineFormat.getFrameSize(), lineFormat.getFrameRate(), lineFormat.isBigEndian());
			}

			if (lineBufferSize == AudioSystem.NOT_SPECIFIED) {
				lineBufferSize = dataLineInfo.getMaxBufferSize();
			}
			
			if (dataLineInfo.getLineClass().isAssignableFrom(MixerSourceLine.class)) {
				return new MixerSourceLine(sourceDataLineInfo, this, lineFormat, lineBufferSize);
			}
			
			if (dataLineInfo.getLineClass().isAssignableFrom(MixerClip.class)) {
				return new MixerClip(clipInfo, this, lineFormat, lineBufferSize);
			}
	
		}

		// $$kk: 06.08.99: the only thing left is the target line.  
		// we *should* implement this!!  for now just refuse access....
		throw new SecurityException("Access to line not granted: " + info);
	}


	public int getMaxLines(Line.Info info) {

		Line.Info fullInfo = getLineInfo(info);

		// if it's not a supported line, the number is automatically 0
		if (fullInfo == null) {
			return 0;
		}

		// this covers both source data lines and clips
		if ( (fullInfo instanceof DataLine.Info) && isSourceLine(fullInfo) ) {
			return sampledVoices;
		}

		// this is our one target line
		if ( (fullInfo instanceof DataLine.Info) && isTargetLine(fullInfo) ) {
			return 1;
		}

		// this shouldn't happen, but the answer's still 0....
		return 0;
	}


// ABSTRACT LINE: ABSTRACT METHOD IMPLEMENTATIONS


	public void implOpen(AudioFormat format, int bufferSize) throws LineUnavailableException {

		if (Printer.trace) Printer.trace(">> HeadspaceMixer: implOpen");

		// initialize the mixer: InitGeneralSound
		nOpenMixer(format.getSampleSizeInBits(), format.getChannels(), 
					(int)format.getSampleRate(),  terpMode, midiVoices, 
					sampledVoices, mixLevel);

		// this reconnects to hardware and starts the engine thread.
		// $$kk: 06.08.99: we will want to reconsider how to handle this.
		nResume();

	
		// $$kk: 06.05.99: this will work now.
		// $$kk: 06.06.99: uh, no it doesn't....

		this.format = format;
		this.bufferSize = nGetDefaultBufferSize();
		if (this.bufferSize <= 0) this.bufferSize = AudioSystem.NOT_SPECIFIED;

		if (Printer.trace) Printer.trace("<< HeadspaceMixer: implOpen succeeded");
	}


	public void implClose() {

		if (Printer.trace) Printer.trace(">> HeadspaceMixer: implClose");

		// this disconnects from hardware and stops the engine thread.
		// $$kk: 06.08.99: we will want to reconsider how to handle this.
		nPause();

		// shut down the mixer: FinisGeneralSound
		nCloseMixer();	

		if (Printer.trace) Printer.trace("<< HeadspaceMixer: implClose succeeded");
	}


// ABSTRACT DATA LINE: ABSTRACT METHOD IMPLEMENTATIONS

	/**
	 * Start the mixer
	 */
	void implStart() {

		if (Printer.trace) Printer.trace(">> HeadspaceMixer: implStart");
/*
		// this reconnects to hardware and starts the engine thread.
		// $$kk: 06.08.99: we will want to reconsider how to handle this.
		// $$kk: 06.11.99: i have the native method throwing the appropriate
		// exception since it really is opening the device here.  we need to
		// change how this is handled at a higher level....
		try {
			nResume();
		} catch (LineUnavailableException e) {
			if (Printer.err) Printer.err("\nFAILED TO START MIXER: " + e + "\n");
		}
*/
		// $$kk: 06.06.99: this is not entirely accurate, but....
		setActive(true);
		setStarted(true);

		if (Printer.trace) Printer.trace("<< HeadspaceMixer: implStart succeeded");
	}

	
	/**
	 * Stop the mixer
	 */
	void implStop() {

		if (Printer.trace) Printer.trace(">> HeadspaceMixer: implStop");
/*
		// this disconnects from hardware and stops the engine thread.
		// $$kk: 06.08.99: we will want to reconsider how to handle this.
		nPause();
*/
		// $$kk: 06.06.99: this is not entirely accurate, but....
		setActive(false);
		setStarted(false);

		if (Printer.trace) Printer.trace("<< HeadspaceMixer: implStop succeeded");
	}



// PUBLIC CONFIGURATION METHODS	


	/**
	 * Obtains the total number of voices available on the system.
	 * The voices are allocated between MIDI and sampled voices.
	 * @return total number of voices
	 */
	public int getSystemVoices() {
		return TOTAL_VOICES;
	}

	
	/**
	 * Obtains the number of voices dedicated to the MIDI synthesizer.
	 * @return number of MIDI voices
	 */
	public int getMidiVoices() {
		return midiVoices;
	}

	
	/**
	 * Obtains the number of voices dedicated to sampled audio streams and 
	 * clips.
	 * @return number of sampled voices
	 */
	public int getSampledVoices() {
		return sampledVoices;
	}

	
	/**
	 * Allocates the system voices between MIDI and sampled voices.
	 * @param midi desired number of MIDI voices
	 * @param sampled desired number of sampled audio voices
	 */
	public boolean allocateVoices(int midi, int sampled) {

		// $$kk: 03.11.99: need error case here!
		if (! (nAllocateVoices(midi, sampled)) ) {
			return false;
		}

		midiVoices = midi;
		sampledVoices = sampled;			
		return true;
	}


	
	/**
	 * Sets the mix level for the system.
	 * @param mixLevel desired mix level
	 */
	public boolean setMixLevel(int mixLevel) {

		// $$kk: 03.11.99: need error case here!

		if (! (nSetMixLevel(mixLevel)) ) {
			return false;
		} 

		this.mixLevel = mixLevel;
		return true;
	}


	
	/**
	 * Obtains the mix level for the system.
	 * @return current mix level
	 */
	public int getMixLevel() {
		return mixLevel;		
	}


	/**
	 * Sets the interpolation mode for the system.
	 * @param mode desired interpolation mode
	 * @see #INTERPOLATION_DROP_SAMPLE
	 * @see #INTERPOLATION_2_POINT
	 * @see #INTERPOLATION_LINEAR
	 */
	public boolean setInterpolationMode(int terpMode) {

		if (! (nSetInterpolation(terpMode)) ) {
			return false;
		}

		this.terpMode = terpMode;
		return true;	
	}


	/**
	 * Obtains the interpolation mode for the system.
	 * @return current interpolation mode
	 */
	public int getInterpolationMode() {
		return terpMode;
	}


	/**
	 * Obtains the current CPU load imposed by the audio system on the
	 * machine as a fraction between 0 and 1.
	 * @return current load
	 */
	public float getCpuLoad() {
		return nGetCpuLoad();
	}
	



// ABSTRACT DATA LINE METHOD OVERRIDES
// $$kk: 06.08.99: we will want to implement some of these at some point

/*
	public float getLevel() {
		return nGetLevel();
	}


	public void drain() {
		nDrain();
	}

	
	public void flush() {		
		nFlush();
	}


	public long getFramePosition() {
		return nGetPosition();
	}
*/


// METHODS FOR INTERNAL IMPLEMENTATION USE


	/**
	 * This lets us handle MIDI devices on the mixer.  
	 * $$kk: 06.08.99: we may want these to become lines....
	 */
	void open(MidiDevice mididevice) throws MidiUnavailableException {

		if (! openMidiDevices.contains(mididevice) ) {
			openMidiDevices.addElement(mididevice);
			try {
				open();
				start();
			} catch (LineUnavailableException e) {
				openMidiDevices.removeElement(mididevice);
				throw new MidiUnavailableException(e.getMessage());
			}
		}
	}

	/**
	 * This lets us handle MIDI devices on the mixer.  
	 * $$kk: 06.08.99: we may want these to become lines....
	 */
	void close(MidiDevice mididevice) {

		openMidiDevices.removeElement(mididevice);
		
		// $$kk: 06.03.99: this should stop and close the mixer if the list
		// of source lines is also empty
		DataLine line = null;
		stop(line);
		close(line);
	}


	/**
	 * Override of the AbstractMixer method for handling MIDI objects.
	 * $$kk: 06.08.99: this is not elegant!	 this is almost all code
	 * duplicated from the AbstractMixer implementation with the
	 * special case for midi....
	 */
	protected void close(Line line) {

		// $$kk: 06.11.99: ignore ourselves for now
		if (this.equals(line)) {
			return;
		}

		sourceLines.removeElement(line);
		targetLines.removeElement(line);

		if ( sourceLines.isEmpty() && targetLines.isEmpty() && openMidiDevices.isEmpty()) {
			close();
		}
	}



	/**
	 * Override of the AbstractMixer method for handling MIDI objects.
	 * $$kk: 06.08.99: this is not elegant!	 this is almost all code
	 * duplicated from the AbstractMixer implementation with the
	 * special case for midi....
	 */
	protected void stop(Line line) {

		// $$kk: 06.11.99: ignore ourselves for now
		if (this.equals(line)) {
			return;
		}

		synchronized(sourceLines) {

			for (int i = 0; i < sourceLines.size(); i++) {

				// if any other open line is running, return

				// this covers clips and source data lines
				if (sourceLines.elementAt(i) instanceof AbstractDataLine) {
					AbstractDataLine sourceLine = (AbstractDataLine)sourceLines.elementAt(i);	
					if ( sourceLine.isStartedRunning() && (!sourceLine.equals(line)) ) {
						return;
					}
				}
			}
		}

		synchronized(targetLines) {

			for (int i = 0; i < targetLines.size(); i++) {

				// if any other open line is running, return
				AbstractDataLine targetLine = (AbstractDataLine)targetLines.elementAt(i);	
				if ( targetLine.isStartedRunning() && (!targetLine.equals(line)) ) {
					return;
				}
			}
		}

		if (! openMidiDevices.isEmpty() ) {
			return;
		}

		// otherwise, stop
		stop();
	}



	/**
	 * Returns the static mixer instance.  i want to get rid of this!!
	 */
	static HeadspaceMixer getMixerInstance() {
	
		return mixerInstance;
	}


// INNER CLASSES

	/**
	 * Private Mixer.Info subclass.
	 */
	private static class MixerInfo extends Mixer.Info {

		/**
		 * Mixer name.
		 */
		private static final String name		= "Java Sound Audio Engine";

		/**
		 * Mixer vendor.
		 */
		private static final String vendor		= "Sun Microsystems";

		/**
		 * Mixer description.
		 */
		private static final String description	= "Software mixer and synthesizer";

		/**
		 * Mixer version.
		 */
		private static final String version	= "1.0";
		
		/**
		 * Constructs a new device info object.
		 * @param name name of the device
		 */
		private MixerInfo() {

			super(name, vendor, description, version);
		}

	} // class MixerInfo


	private static class MixerReverbControl extends EnumControl {

		private static final MixerReverbType NO_REVERB = new MixerReverbType("No Reverb", 0, 0.0f, 0, 0.0f, 0, 1);
		private static final MixerReverbType CLOSET = new MixerReverbType("Closet", 600, -7.0f, 2500, -10.0f, 150000, 2);
		private static final MixerReverbType GARAGE = new MixerReverbType("Garage", 3900, -4.0f, 14700, -6.0f, 900000, 3);
		private static final MixerReverbType ACOUSTIC_LAB = new MixerReverbType("Acoustic Lab", 2000, -2.0f, 8000, -3.0f, 280000, 4);
		private static final MixerReverbType CAVERN = new MixerReverbType("Cavern", 10300, -1.4f, 41300, -2.0f, 2250000, 5);
		private static final MixerReverbType DUNGEON = new MixerReverbType("Dungeon", 2600, -0.7f, 10300, -1.0f, 1600000, 6);

		/**
		 * Supported reverb types
		 */
		private static final ReverbType[] types = { NO_REVERB, CLOSET, GARAGE, ACOUSTIC_LAB, CAVERN, DUNGEON };

		/**
		 * Default reverb type
		 */
		private static final ReverbType defaultReverb = ACOUSTIC_LAB;


		// STATE VARIABLES
				
		/**
		 * Set the engine reverb to the default value.  If this
		 * fails, set the reverb type to ReverbControl.NO_REVERB.
		 */
		private MixerReverbControl() {
			super(EnumControl.Type.REVERB, types, defaultReverb);
		}
		

// $$kk: 09.30.99: new controls stuff sorta screws this up!!m		
/*
		public ReverbControl.ReverbType getReverbType(int earlyReflectionDelay, float earlyReflectionIntensity, int lateReflectionDelay, float lateReflectionIntensity, int decayTime) {

			// $$kk: 06.09.99: implemented according to warren dale's reccomendations

			if (decayTime <= 0) {
				return NO_REVERB;
			}

			if (decayTime > 1000) {

				if (lateReflectionIntensity <= -1.0f) {
					return CAVERN;
				} else {
					return DUNGEON;
				}

			} else if (decayTime >= 300) {

				return GARAGE;

			} else {

				if (lateReflectionIntensity < -1.0f) {
					return ACOUSTIC_LAB;
				} else {
					return CLOSET;
				}
			}
		}
*/

		public void setValue(Object value) {

			if (Printer.trace) Printer.trace("HeadspaceMixer: MixerReverbControl: setValue: " + value.toString());
			
			if (!(value instanceof ReverbType)) {
				throw new IllegalArgumentException("Value must be of type ReverbType");
			}

			ReverbType reverbType = (ReverbType)value;
			int reverbMode;

			if (reverbType instanceof MixerReverbType) {
				reverbMode = ((MixerReverbType)reverbType).getReverbMode();
			} else if (value == NO_REVERB) {
				reverbMode = 1;	// pass-through
			} else {
				throw new IllegalArgumentException("Unknown reverb value: " + value);
			}

			int reverbSet = nSetReverb(reverbMode);	
			
			if (reverbSet == reverbMode) {

				if (Printer.trace) Printer.trace("HeadspaceMixer: setReverb succeeded: " + value.toString());
				super.setValue(value);

			} else {
				
				if (Printer.err) Printer.err("HeadspaceMixer: MixerReverbControl: setValue failed: value: " + value.toString() + " set mode: " + reverbMode + " returned mode: " + reverbSet);
				if (reverbSet == 1) {
					/* pass-through */
				} else {

					for (int i = 0; i < types.length; i++) {

						if (types[i] instanceof MixerReverbType) {
							if ( ((MixerReverbType)types[i]).getReverbMode() == reverbSet ) {
								super.setValue(types[i]);
							}
						}
					}
				}
			}
		}


		/**
		 * Set the engine reverb to the default value.  If this
		 * fails, set the reverb type to NO_REVERB.
		 */
		private void initializeReverb() {

			try {
			
				setValue(defaultReverb);
				if (Printer.trace) Printer.trace("Initialized reverb to: " + getValue());
			
			} catch (IllegalArgumentException e) {

				setValue(NO_REVERB);
				if (Printer.err) Printer.err("Failed to initialize reverb to: " + defaultReverb + "; set reverb to: " + getValue());
			}
		}
		
		
		private static class MixerReverbType extends ReverbType {
			
			// ReverbMode value used in the Headspace engine
			private final int reverbMode;
			
			private MixerReverbType(String name, int earlyReflectionDelay, float earlyReflectionIntensity, int lateReflectionDelay, float lateReflectionIntensity, int decayTime, int reverbMode) {

				super(name, earlyReflectionDelay, earlyReflectionIntensity, lateReflectionDelay, lateReflectionIntensity, decayTime);
				this.reverbMode = reverbMode;
			}

			private int getReverbMode() {
				return reverbMode;
			}
		}
				
	} // class MixerReverbControl


// NATIVE METHODS

	// GM_InitGeneralSound and GM_SetAudioTask
	private native void nOpenMixer(int sampleSizeInBits, int channels, int sampleRate, 
									int terpMode, int midiVoices, int sampleVoices, 
									int mixLevel) throws LineUnavailableException;
	
	// GM_SetAudioTask(NULL), GM_StopHardwareSoundManager, and GM_FinisGeneralSound
	private native void nCloseMixer();

	// GM_PauseGeneralSound
	private native void nPause();
	
	// GM_ResumeGeneralSound
	// $$kk 06.11.99: we are using this for "start" which shouldn't throw this exception.
	// however, the truth is that this method opens the device and can fail!  the native
	// code can throw the exception; this will screw the program.
	private native void nResume() throws LineUnavailableException;

	// not implemented
	private native void nDrain();
	private native void nFlush();
	private native long nGetPosition();
	private native float nGetLevel();	

	
	// GM_ChangeAudioModes
	private native boolean nSetMixerFormat(int sampleSizeInBits, int channels, int sampleRate);
	
	// GM_ChangeSystemVoices
	private native boolean nAllocateVoices(int midiVoices, int sampledVoices);	
	
	// GM_ChangeSystemVoices
	private native boolean nSetMixLevel(int mixLevel);	
	
	// GM_ChangeAudioModes
	private native boolean nSetInterpolation(int terpMode);
	
	// not implemented
	private native float nGetCpuLoad();

	// $$kk: 03.15.99: we need to figure out how to handle the
	// device!!!!  for now i am using these methods....
	//private native void nOpenDevice() throws LineUnavailableException;
	//private native void nCloseDevice();

	private static native int nGetTotalVoices();

	
	// LINKED STREAMS

	private native int nCreateLinkedStreams(int[] idArray);
	private native boolean nStartLinkedStreams(int linkRef);
	private native void nStopLinkedStreams(int linkRef);

	// STATICS

	// in SAMPLE FRAMES.  
	// this should be HAE_GetMaxSamplePerSlice() * HAE_GetAudioBufferCount()
	private static native int nGetDefaultBufferSize();	

	// GM_SetReverbType
	private static native int nSetReverb(int reverbMode);
}
