/*
 * @(#)HeadspaceSoundbank.java	1.12 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;

import java.net.URL;
import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

import java.util.Vector;

import javax.sound.midi.*;


/**
 * HeadspaceSoundbank
 *
 * @version 1.12, 00/02/02
 * @author David Rivas
 * @author Kara Kytle
 */
class HeadspaceSoundbank implements Soundbank {


// STATE VARIABLES

	String name;
	String version;
	String vendor;
	String description;
	Vector instruments = new Vector();
	Vector samples = new Vector();
	Vector sequences = new Vector();


// IMPLEMENTATION VARIABLES

	private int id = 0;		// resource id for engine


// CONSTRUCTOR

	HeadspaceSoundbank(URL url) throws IOException {

		if(Printer.trace) Printer.trace("HeadspaceSoundbank: constructor: url: " + url);

		String protocol = url.getProtocol();
		if ( ! (protocol.equals("file")) ) { 

			initialize( url.openStream() );
			
			//throw new IllegalArgumentException("Non-local files not supported: " + url);
		} else {
			
			String path = url.getFile();
			initialize(path);
		}

		if(Printer.trace) Printer.trace("HeadspaceSoundbank: constructor: url: " + url + " completed");
	}		


	HeadspaceSoundbank(String path) {

		initialize(path);
		if(Printer.trace) Printer.trace("HeadspaceSoundbank: constructor: path: " + path + " completed");
	}

	HeadspaceSoundbank(InputStream stream) throws IOException {
		
		if(Printer.trace) Printer.trace("HeadspaceSoundbank: constructor: stream: " + stream);

		initialize(stream);
	}

	HeadspaceSoundbank(File file) throws IOException {

		if(Printer.trace) Printer.trace("HeadspaceSoundbank: constructor: file: " + file);

		FileInputStream fis = new FileInputStream( file );

		initialize( (InputStream) fis);
	}




// SOUNDBANK METHODS

	public String getName() {
		return name;
	}


	public String getVersion() {
		return version;
	}


	public String getVendor() {
		return vendor;
	}


	public String getDescription() {
		return description;
	}


	public SoundbankResource[] getResources() {
		
		SoundbankResource[] sampleArray;
		
		synchronized(samples) {
	
			sampleArray = new SoundbankResource[samples.size()];

			for (int i = 0; i < sampleArray.length; i++) {
				sampleArray[i] = (SoundbankResource)samples.elementAt(i);
			}
		}

		/* $$kk: 09.27.99: need to implement for embedded sequences, too! */

		return sampleArray;
	}


	public Instrument[] getInstruments() {
		
		Instrument[] instrumentArray;
		
		synchronized(instruments) {
	
			instrumentArray = new Instrument[instruments.size()];

			for (int i = 0; i < instrumentArray.length; i++) {
				instrumentArray[i] = (Instrument)instruments.elementAt(i);
			}
		}

		return instrumentArray;
	}


	public Instrument getInstrument(Patch patch) {

		Instrument returnedInstrument;
		
		synchronized(instruments) {
	
			for (int i = 0; i < instruments.size(); i++) {

				returnedInstrument = (Instrument)instruments.elementAt(i);
				if ( (returnedInstrument.getPatch().getBank() == patch.getBank()) &&
					(returnedInstrument.getPatch().getProgram() == patch.getProgram()) )

					return returnedInstrument;
			}
		}

		return null;		
	}


// HELPER METHODS

	private void setupInformation() {
	
		// set up all the information now

		name = nGetName(id);
		if (name == null) {
			name = "Untitled Headspace Soundbank";
		}

		int versionMajor = nGetVersionMajor(id);
		int versionMinor = nGetVersionMinor(id);
		int versionSubMinor = nGetVersionSubMinor(id);

		version = new String(versionMajor + "." + versionMinor + "." + versionSubMinor);

		vendor = "Sun Microsystems, Headspace Corporation";
		description = "Soundbank for use with Headspace Mixer.";

		nGetInstruments(id, instruments);
		nGetSamples(id, samples);
		//nGetSequences(id, sequences);

	}
	
	private void initialize( InputStream stream ) throws IOException {

		if(Printer.trace) Printer.trace("HeadspaceSoundbank: initialize: stream: " + stream);

		// $$kk: 04.12.99: this is sort of a hack.  we need to make sure the library
		// gets loaded and (potentially) that the mixer gets initialized.
		HeadspaceMixer mixer = HeadspaceMixer.getMixerInstance();

		// load sounbank into a local byte array

		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		byte buffer[] = new byte[16384];
		byte loadedbank[] = null;
		int bytesRead = 0;
		
		// this loop can throw an IOException
		while(true) {
			bytesRead = stream.read(buffer,0,buffer.length);
			if(bytesRead <= 0) {
				stream.close();
				break;
			}
			baos.write(buffer, 0, bytesRead);
		}
		loadedbank = baos.toByteArray();

		// bring the byte array down to native code and open the resource

		if(Printer.debug) Printer.debug("> calling nOpenResourceFromByteArray with length: " + loadedbank.length );
		id = nOpenResourceFromByteArray(loadedbank, loadedbank.length);
		if(Printer.debug) Printer.debug("> returned from nOpenResourceFromByteArray; id = " + id);

		if (id == 0) {
			throw new IllegalArgumentException("Failed to get id");
		}
	
		setupInformation();
		
		if(Printer.trace) Printer.trace("HeadspaceSoundbank: initialize: stream: " + stream + " completed");


	}

	private void initialize(String path) {
	
		if(Printer.trace) Printer.trace("HeadspaceSoundbank: initialize: path: " + path);

		// $$kk: 04.12.99: this is sort of a hack.  we need to make sure the library
		// gets loaded and (potentially) that the mixer gets initialized.
		HeadspaceMixer mixer = HeadspaceMixer.getMixerInstance();


		if(Printer.debug) Printer.debug("> calling nOpenResource with path: " + path);
		id = nOpenResource(path);
		if(Printer.debug) Printer.debug("> returned from nOpenResource; id = " + id);

		if (id == 0) {
			throw new IllegalArgumentException("Failed to get id");
		}

		setupInformation();

		if(Printer.trace) Printer.trace("HeadspaceSoundbank: initialize: path: " + path + " completed");
	}		


// NATIVE METHODS

	private native int nOpenResource(String path);
	private native int nOpenResourceFromByteArray(byte[] soundbankData, int length);
	
	// $$kk: 04.11.99: we are never calling this, but we should!!
	private native boolean nCloseResource(int id);
	
	// $$kk: 04.12.99: this seems not to be working
	private native String nGetName(int id);

	private native int nGetVersionMajor(int id);
	private native int nGetVersionMinor(int id);
	private native int nGetVersionSubMinor(int id);

	private native void nGetInstruments(int id, Vector instruments);
	private native void nGetSamples(int id, Vector samples);

	//protected native void nGetSequences();

	// $$kk: 04.11.99: this is *not* the right way to accomplish
	// what we're doing here....  i am using this to set this
	// soundbank as the first resource file so that the following
	// calls operate on it....
	//private native boolean nSetCurrentResource(int id);
}
