/*
 * @(#)JavaSoundAudioClip.java	1.21 00/02/02
 *
 * Copyright 1999, 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;

import java.io.IOException;
import java.io.InputStream;
import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.applet.AudioClip;
import java.lang.InterruptedException;

import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.Clip;
import javax.sound.sampled.Control;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;


import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiFileFormat;
import javax.sound.midi.MetaMessage;
import javax.sound.midi.Sequencer;
import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.MetaEventListener;

/**
 * Java Sound audio clip;
 *
 * @version 1.21 00/02/02
 * @author Arthur van Hoff, Kara Kytle, Jan Borgersen
 */

public class JavaSoundAudioClip implements AudioClip, MetaEventListener, LineListener {

	private Clip clip = null;
	private Sequencer sequencer = null;
	private SourceDataLine sourceline = null;
	private DataPusher datapusher = null;
	private boolean sequencerloop = false;
	private boolean streamloop = false;

	private byte loadedAudio[] = null;

	/**
	 * used for determining how many samples is the 
	 * threshhold between playing as a Clip and streaming
	 * from the file. 
	 *
	 * $$jb: 11.07.99: the engine has a limit of 1M
	 * samples to play as a Clip, so compare this number
	 * with the number of samples in the stream.
	 *
	 */
	private long clipthresh = 1048576;

    public JavaSoundAudioClip(InputStream in) throws IOException {

		BufferedInputStream bis = null;
		int bufferSize = 1024;
		AudioInputStream as = null;
		AudioFormat format = null;

		if(Printer.debug)Printer.debug("Instantiating com.sun.media.sound.JavaSoundAudioClip");

		bis = new BufferedInputStream( in, bufferSize );
		bis.mark(bufferSize);

		try {
			as = AudioSystem.getAudioInputStream(bis);
			format = as.getFormat();

			// if small enough, open as a clip.  
			// otherwise, stream

			long tmplength = as.getFrameLength();

			if ( (tmplength > 0) && (tmplength < clipthresh) ) {
			    if( openAsClip(as) < 0 ) {
				throw new IOException("Unable to create AudioClip from input stream");
			    }
			} else {
			    if( openAsStream(as) < 0 ) {
				throw new IOException("Unable to create AudioClip from input stream");
		  	    }
			}
    

		} catch( UnsupportedAudioFileException e ) {

			try {

				MidiFileFormat mff = MidiSystem.getMidiFileFormat(bis);

				if( openAsSequence(bis) < 0 ) {
					throw new IOException("Unable to create AudioClip from input stream");
				}

			} catch( InvalidMidiDataException e1 ) {

				throw new IOException("Unable to create AudioClip from input stream");
			}
		}
    }


    public synchronized void play() {

	if(Printer.debug)Printer.debug("JavaSoundAudioClip->play()");

	if (clip != null) {

            
            stop();
	    clip.start();

            if(Printer.debug)Printer.debug("Clip should be playing");

	} else if (datapusher != null) {


	    stop();
	    streamloop = false;
	    sourceline.start();
	    datapusher.start();

            if(Printer.debug)Printer.debug("Clip should be playing");

        } else if (sequencer != null) {
	

	    stop();
	    sequencerloop = false;
	    sequencer.start();


            if(Printer.debug)Printer.debug("Clip should be playing");
	}
    }

    
    public synchronized void loop() {

	if(Printer.debug)Printer.debug("JavaSoundAudioClip->loop()");

	if (clip != null) {

	    stop();
	    clip.loop(Clip.LOOP_CONTINUOUSLY);
            
	    if(Printer.debug)Printer.debug("Clip should be looping");

	} else if (datapusher != null ) {


	    stop();
	    streamloop = true;
	    sourceline.start();
	    datapusher.start();

            if(Printer.debug)Printer.debug("Clip should be playing");

        } else if (sequencer != null) {
	
	    stop();
	    sequencerloop = true;

	    sequencer.start();

            if(Printer.debug)Printer.debug("Clip should be looping");
	}
    }

    public synchronized void stop() {

	if(Printer.debug)Printer.debug("JavaSoundAudioClip->stop()");

	if (clip != null) {
			
	    clip.stop();
	    clip.setFramePosition(0);

	    if(Printer.debug)Printer.debug("Clip should be stopped");

	} else if (datapusher != null) {

	    streamloop = false;
	    datapusher.stop();

	} else if (sequencer != null) {

	    sequencer.stop();
	    sequencer.setMicrosecondPosition(0);

	    if(Printer.debug)Printer.debug("Clip should be stopped");
	}
    }

	// Event handlers

	// handle line events for looping from a byte array

	public synchronized void update( LineEvent event ) {

	    if(Printer.debug) Printer.debug("LINE EVENT RECEIVED!!!!!");

	    if( streamloop ) {
	        if( (datapusher != null) && (event.getType() == LineEvent.Type.STOP) ) {

		    loop();
	        }
	    }

	}
	
	// handle MIDI track end meta events for looping

	public synchronized void meta( MetaMessage message ) {
		
		if(Printer.debug)Printer.debug("META EVENT RECEIVED!!!!! ");

		if( message.getType() == 47 ) {
			if( sequencerloop ){

				//notifyAll();
				sequencer.setMicrosecondPosition(0);
				loop();
			} else {
				stop();
			}
		}
	}


    public String toString() {
        return getClass().toString();
    }

    
    protected void finalize() {

		if (clip != null) {
			clip.close();
		}
    }

// FILE LOADING METHODS

	private int openAsStream(AudioInputStream as) throws IOException, UnsupportedAudioFileException {

		if(Printer.debug)Printer.debug("JavaSoundAudioClip->openAsStream()");

		DataLine.Info info = null;
		AudioFormat format = as.getFormat();

		as = convertToPlayableStream( as );
		if ( as == null ) {
		    return -1;
		}

		if( loadStreamIntoMemory( as ) <= 0 ) {
		    return -1;
		}

		try {

			info = new DataLine.Info(SourceDataLine.class, as.getFormat());

			if (! (AudioSystem.isLineSupported(info)) ) {
				// fail silently
				return -1;
			}
			sourceline = (SourceDataLine)AudioSystem.getLine(info);
			sourceline.addLineListener( (LineListener) this );
			datapusher = new DataPusher( sourceline, loadedAudio );
			sourceline.open();

		} catch (LineUnavailableException e) {
			// fail silently
			return -1;
		}

		if( datapusher==null ) {
			// fail silently
			return -1;
		}


		if(Printer.debug)Printer.debug("Loaded as stream.");
		return 0;

	}

	private int loadStreamIntoMemory( AudioInputStream as ) throws IOException {

		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		byte buffer[] = new byte[16834];
		int bytesRead = 0;
		int totalBytesRead = 0;

		// this loop may throw an IOException

		while( true ) { 
		    bytesRead = as .read(buffer, 0, buffer.length);
		    if (bytesRead <= 0) {
		        as.close();
		        break;
		    }
		    totalBytesRead += bytesRead;
		    baos.write(buffer, 0, bytesRead);
		}
		loadedAudio = baos.toByteArray();
		return totalBytesRead;
	}

	private int openAsClip(AudioInputStream as) throws IOException, UnsupportedAudioFileException { 

		if(Printer.debug)Printer.debug("JavaSoundAudioClip->openAsClip()");

		DataLine.Info info = null;

		as = convertToPlayableStream( as );
		if ( as == null ) {
		    return -1;
		}

/*	        // we can't yet open the device for ALAW/ULAW playback, so we have
		// to convert ALAW/ULAW to PCM here.

		if( (as.getFormat().getEncoding() == AudioFormat.Encoding.ULAW) ||
			(as.getFormat().getEncoding() == AudioFormat.Encoding.ALAW)) {

			AudioFormat af = as.getFormat();
			AudioInputStream asold = as;
			as = AudioSystem.getAudioInputStream( 
					new AudioFormat( AudioFormat.Encoding.PCM_SIGNED,
										af.getSampleRate(),
										af.getSampleSizeInBits() * 2,
										af.getChannels(),
										af.getFrameSize() * 2,
										af.getFrameRate(),
										true )
									, asold );
			if( as==null ) {
				// could not convert
				return -1;
			}
		}
*/

		try {

			info = new DataLine.Info(Clip.class, as.getFormat(), (int)as.getFrameLength());

			if (! (AudioSystem.isLineSupported(info)) ) {
				// fail silently
				return -1;
			}
			clip = (Clip)AudioSystem.getLine(info);
			clip.open( as );

		} catch (LineUnavailableException e) {
			// fail silently
			return -1;
		}

		if( clip==null ) {
			// fail silently
			return -1;
		}


		if(Printer.debug)Printer.debug("Loaded as clip.");
		return 0;
	}

	private AudioInputStream convertToPlayableStream( AudioInputStream as ) {

		// we can't yet open the device for ALAW/ULAW playback, so we have
		// to convert ALAW/ULAW to PCM here.

		if( (as.getFormat().getEncoding() == AudioFormat.Encoding.ULAW) ||
			(as.getFormat().getEncoding() == AudioFormat.Encoding.ALAW)) {

			AudioFormat af = as.getFormat();
			AudioInputStream asold = as;
			as = AudioSystem.getAudioInputStream( 
					new AudioFormat( AudioFormat.Encoding.PCM_SIGNED,
										af.getSampleRate(),
										af.getSampleSizeInBits() * 2,
										af.getChannels(),
										af.getFrameSize() * 2,
										af.getFrameRate(),
										true )
									, asold );
		}

		return as;
	}

	private int openAsSequence(BufferedInputStream in) throws IOException {

		if(Printer.debug)Printer.debug("JavaSoundAudioClip->openAsSequence()");

		// get the sequencer
		//sequencer = MidiSystem.getSequencer( null );
		try {
		    sequencer = MidiSystem.getSequencer( );
		} catch(MidiUnavailableException me) {
		    return -1;
		}
		if( sequencer==null ) {
			return -1;
		}
		sequencer.addMetaEventListener( (MetaEventListener) this );

		try {
			sequencer.open();
			sequencer.setSequence( in );

		} catch (InvalidMidiDataException e) {
			e.printStackTrace();
			return -1;
		} catch (MidiUnavailableException e1) {
			e1.printStackTrace();
			return -1;
		}

		if(Printer.debug)Printer.debug("Loaded as sequence.");

		return 0;
	}

}

class DataPusher implements Runnable {

	SourceDataLine source;
	ByteArrayInputStream is; 
	byte [] audio;

	Thread pushThread = null;
	boolean stopping;
	boolean stopped;
			
	DataPusher(SourceDataLine source, byte audio[] ) {

	   this.source = source;
	   this.audio = audio;
	}

	protected void start() {

	    ByteArrayInputStream bais = null;

	    if (pushThread != null) {
					
		if (Printer.debug) Printer.debug("Thread is not null");
	        return;
	    }

	    if (Printer.debug) Printer.debug("Starting push");

	    stopping = false;
	    stopped = false;

	    is = new ByteArrayInputStream( audio );

	    pushThread = new Thread(this);
	    pushThread.start();
	}


	protected void stop() {

	    if (pushThread == null) {

	 	return;
	    }

	    if (Printer.debug) Printer.debug("Stopping push");
			
	    stopping = true;
	    while (!stopped) {
		try {
		    Thread.sleep(10);
		} catch (InterruptedException e) {  }
	    }
	    pushThread = null;
	}



	/**
	 * Write data to the source data line.
	 */
	public void run() {

	    byte[] dataArray = new byte[1024];

	    int bytesRead;
	    while(stopping == false) {

		try {
						
		    bytesRead = is.read(dataArray);

		    if (bytesRead == -1) {

			if (Printer.debug) Printer.debug("Found end of stream");
			source.drain();
			source.stop();
			break;
		    }		
		    if (Printer.debug) Printer.debug("> Writing " + bytesRead + " bytes");
		    int bytesRemaining = bytesRead;

		    while (bytesRemaining > 0) {
			bytesRemaining -= source.write(dataArray, 0, bytesRead);
		    }

		    if (Printer.debug) Printer.debug("< Wrote " + bytesRead + " bytes");

		} catch (java.io.IOException ioe) {

		    break;
		}
	    }

	    source.flush();
	    source.stop();
	    stopped = true;

	}

} // class DataPusher

