/*
 * @(#)MidiOutDeviceProvider.java	1.7 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;

import java.util.Vector;

import javax.sound.midi.MidiDevice;
import javax.sound.midi.Receiver;
import javax.sound.midi.spi.MidiDeviceProvider;


/**
 * MIDI output device provider.  
 *
 * @version 1.7, 00/02/02
 * @author Kara Kytle
 */
public class MidiOutDeviceProvider extends MidiDeviceProvider {


// STATIC VARIABLES
			
	
	/**
	 * Set of info objects for all MIDI output devices on the system.
	 */
	private static MidiOutDeviceInfo[] infos;
	
	/**
	 * Set of all MIDI output devices on the system.
	 */
	private static MidiOutDevice[] devices;

	

// STATIC

	/**
	 * Create objects representing all MIDI output devices on the system.
	 */
	static {

		if (Printer.trace) Printer.trace("MidiOutDeviceProvider: static");
		
		// initialize 
		Platform.initialize();
		
		// get the number of output devices
		int numDevices = nGetNumDevices();

		// initialize the arrays
		infos = new MidiOutDeviceInfo[numDevices];
		devices = new MidiOutDevice[numDevices];

		// fill in the info objects now.
		// we'll fill in the device objects as they're requested.

		String name;
		String vendor;
		String description; 
		String version;

		for (int i = 0; i < infos.length; i++) {
			
			name = nGetName(i);
			vendor = nGetVendor(i);
			description = nGetDescription(i);
			version = nGetVersion(i);			

			infos[i] = new MidiOutDeviceInfo(name, vendor, description, version, i, MidiOutDeviceProvider.class);
		}

		if (Printer.trace) Printer.trace("MidiOutDeviceProvider: static: found numDevices: " + numDevices);
	}


// CONSTRUCTOR


	/**
	 * Required public no-arg constructor.
	 */ 
	public MidiOutDeviceProvider() {
		if (Printer.trace) Printer.trace("MidiOutDeviceProvider: constructor");
	}


	public MidiDevice.Info[] getDeviceInfo() {

		MidiDevice.Info[] localArray = new MidiDevice.Info[infos.length];
		System.arraycopy(infos, 0, localArray, 0, infos.length);
		return localArray;
	}	


	public MidiDevice getDevice(MidiDevice.Info info) {

		if (! (info instanceof MidiOutDeviceInfo) ) {
			throw new IllegalArgumentException("MidiDevice " + info.toString() + " not supported by this provider.");
		}

		MidiOutDeviceInfo midiOutInfo = (MidiOutDeviceInfo)info;

		for (int i = 0; i < infos.length; i++) {

			if (midiOutInfo == infos[i]) { 

				int index = midiOutInfo.getIndex();

				if (devices[index] == null) {
					devices[index] = new MidiOutDevice(midiOutInfo);
				}

				return devices[index];
			}
		}

		throw new IllegalArgumentException("MidiDevice " + info.toString() + " not supported by this provider.");
	}


// INNER CLASSES


	/**
	 * Info class for MidiOutDevices.  Adds an index value for
	 * making native references to a particular device and a the
	 * provider's Class to keep the provider class from being 
	 * unloaded.  Otherwise, at least on JDK1.1.7 and 1.1.8,
	 * the provider class can be unloaded.  Then, then the provider
	 * is next invoked, the static block is executed again and a new
	 * instance of the device object is created.  Even though the 
	 * previous instance may still exist and be open / in use / etc.,
	 * the new instance will not reflect that state...
	 */
	static class MidiOutDeviceInfo extends MidiDevice.Info {

		private int index;
		private Class providerClass;

		private MidiOutDeviceInfo(String name, String vendor, String description, String version, int index, Class providerClass) {

			super(name, vendor, description, version);
			this.index = index;
			this.providerClass = providerClass;
		}


		int getIndex() {
			return index;
		}		

	} // class MidiOutDeviceInfo


// NATIVE METHODS  

	private static native int nGetNumDevices();
	private static native String nGetName(int index);
	private static native String nGetVendor(int index);
	private static native String nGetDescription(int index);
	private static native String nGetVersion(int index);
}





