/*
 * @(#)Platform.java	1.17 00/02/02
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;	  

import java.lang.reflect.Method;
import java.lang.reflect.Constructor;

/**
 * Audio configuration class for exposing attributes specific to the platform or system.
 *
 * @version 1.17 00/02/02
 * @author Kara Kytle
 */
class Platform {


// STATIC FINAL CHARACTERISTICS

	// $$kk: 03.06.98: should maybe get this value from the engine via a native method
	// private static final int TOTAL_VOICES						= 64;
 
	// native library we need to load
	private static final String[] libraries						= { "jsound" };


// SYSTEM CHARACTERISTICS
// vary according to hardware architecture
	
	// signed8 (use signed 8-bit values) is true for everything we support except for 
	// the solaris sbpro card.
	// we'll leave it here as a variable; in the future we may need this in java.
	// wait, is that true?  i'm not sure.  i think solaris takes unsigned data?
	// $$kk: 03.11.99: i think solaris takes unsigned 8-bit or signed 16-bit data....
	private static boolean signed8;

	// intel is little-endian.  sparc is big-endian.
	private static boolean bigEndian;

	// this is the value of the "java.home" system property.  i am looking it up here
	// for use when trying to load the soundbank, just so
	// that all the privileged code is localized in this file....
	private static String javahome;

	// this is the value of the "java.class.path" system property
	private static String classpath;

// SECURITY 

	private static JSSecurity jsSecurity = null;
	private static boolean securityPrivilege = false;

	static {

		Method m[] = new Method[1];
		Class cl[] = new Class[1];
		Object args[][] = new Object[1][0];

		if(Printer.trace)Printer.trace(">> Platform.java: static");

		try {
			jsSecurity = JSSecurityManager.getJSSecurity();
			securityPrivilege = true;
		} catch (SecurityException e) {
			if(Printer.err)Printer.err("Platform.java: Security Exception: " + e);
		}

		if(Printer.debug)Printer.debug("jsSecurity: " + jsSecurity);
		if(Printer.debug)Printer.debug("securityPrivilege: " + securityPrivilege);

		if ( securityPrivilege && (jsSecurity != null ) ) {

			if( jsSecurity instanceof DisabledSecurity ) {

				// do nothing...

			} else if( jsSecurity instanceof JDK12Security ) {

				loadLibraries();
				readProperties();

			} else {

				// run privileged code with non-1.2-style security
				try {
					jsSecurity.requestPermission(m, cl, args, JSSecurity.READ_PROPERTY);
					m[0].invoke(cl[0], args[0]);

					jsSecurity.requestPermission(m, cl, args, JSSecurity.LINK);
					m[0].invoke(cl[0], args[0]);

				} catch (Exception e) {

					if(Printer.err)Printer.err("Unable to get read property privilege: " + e);
				}

				loadLibraries();
				readProperties();
			}
		} else {

			// run privileged code without security

			loadLibraries();
			readProperties();
		}
	}


	/**
 	 * Private constructor.
	 */
	private Platform() { 
	}
	

// METHODS FOR INTERNAL IMPLEMENTATION USE


	/**
	 * Dummy method for forcing initialization.
	 */
	static void initialize() {

		if(Printer.trace)Printer.trace("Platform: initialize()");
	}

	
	/**
	 * Determine whether the system is big-endian.
	 */
	static boolean isBigEndian() {
		
		return bigEndian;
	}


	/**
	 * Determine whether the system takes signed 8-bit data.
	 */
	static boolean isSigned8() {
		
		return signed8;
	}


	/**
	 * Obtain javahome.
	 * $$kk: 04.16.99: this is *bad*!!
	 */
	static String getJavahome() {

		return javahome;
	}

	/**
	 * Obtain classpath.
	 * $$jb: 04.21.99: this is *bad* too!!
	 */
	static String getClasspath() {

		return classpath;
	}


// PRIVATE METHODS

	/**
	 * Load the native library or libraries.
	 */
	private static void loadLibraries() {

		if(Printer.trace)Printer.trace(">>Platform.loadLibraries");

		try {

			// load the libraries
			for (int i = 0; i < libraries.length; i++) {

				try {
					if( securityPrivilege ) {
						if( jsSecurity != null ) {
							if(Printer.debug)Printer.debug("using security manager to load libraries");
							jsSecurity.loadLibrary(libraries[i]);
						} else {
							if(Printer.debug)Printer.debug("not using security manager to load libraries");
							System.loadLibrary(libraries[i]);
						}
						if(Printer.debug)Printer.debug("loaded library " + libraries[i]);
					} else {
						if(Printer.debug)Printer.debug("no security privilege, did not load library " + libraries[i]);
					}
				} catch (UnsatisfiedLinkError e2) {
					if(Printer.err)Printer.err("UnsatisfiedLinkError loading native library " + libraries[i]);
					throw(e2);
				}
			}

		} catch (SecurityException e) {
			if(Printer.err)Printer.err("Security exception loading native libraries.  JavaSound requires access to these resources.");
			throw(e);
		}
	}

						  
	/**
	 * Read the required system properties.
	 */
	private static void readProperties() {	

		try {
		
			// check out the hardware architecture and determine its endianism

			// first, set our default values
			bigEndian = false;
			signed8 = false;

			String arch = null;

			// get the system properties here.  use security manager if present.
			if ( securityPrivilege && (jsSecurity != null ) ) {

				if( jsSecurity instanceof JDK12Security ) {

					if(Printer.debug)Printer.debug("using 12 security to get properties");
					try{
						Constructor cons = JDK12PropertyAction.cons;
						arch = (String) JDK12.doPrivM.invoke(
									JDK12.ac,
									new Object[] {
									cons.newInstance( new Object[] { "os.arch" } )
								});
						javahome = (String) JDK12.doPrivM.invoke(
									JDK12.ac,
									new Object[] {
									cons.newInstance( new Object[] { "java.home" } )
								});
						classpath = (String) JDK12.doPrivM.invoke(
									JDK12.ac,
									new Object[] {
									cons.newInstance( new Object[] { "java.class.path" } )
								});
					} catch( Exception e ) {
						if(Printer.debug)Printer.debug("not using security to get properties");
						arch = System.getProperty("os.arch");
						javahome = System.getProperty("java.home");
						classpath = System.getProperty("java.class.path");
					}
				} else {
					if(Printer.debug)Printer.debug("using security to get properties");
					arch = jsSecurity.readProperty("os.arch");
					javahome = jsSecurity.readProperty("java.home");
					classpath = jsSecurity.readProperty("java.class.path");
				}
			} else {
				if(Printer.debug)Printer.debug("not using security to get properties");
				arch = System.getProperty("os.arch");
				javahome = System.getProperty("java.home");
				classpath = System.getProperty("java.class.path");
			}

			if (arch != null) {

				if (arch.equals("x86")) {
					if(Printer.debug)Printer.debug("found x86 hardware");
					bigEndian = false;
					signed8 = false;
				} else if (arch.equals("sparc")) {
					if(Printer.debug)Printer.debug("found sparc hardware");
					bigEndian = true;
					signed8 = true;
				} else {
					if(Printer.err)Printer.err("could not determine hardware architecture");
				}
			}

		} catch (SecurityException e) {
			if(Printer.err)Printer.err("Security exception getting system properties.  JavaSound requires access to these resources.");
			throw(e);
		}
	}  
}
