/*
 * @(#)StandardMidiFileReader.java	1.16 00/02/02
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.SequenceInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PushbackInputStream;
import java.io.RandomAccessFile;
import java.io.BufferedInputStream;
import java.net.URL;
import java.net.MalformedURLException;



import javax.sound.midi.MidiFileFormat;	
import javax.sound.midi.InvalidMidiDataException;	
import javax.sound.midi.MetaMessage;	
import javax.sound.midi.MidiEvent;	
import javax.sound.midi.MidiMessage;	
import javax.sound.midi.Sequence;	
import javax.sound.midi.ShortMessage;	
import javax.sound.midi.SysexMessage;	
import javax.sound.midi.Track;	



/**
 * MIDI file reader.
 *
 * 1.16 00/02/02
 * @author Kara Kytle
 * @author Jan Borgersen
 */

public class StandardMidiFileReader extends SunMidiFileReader {

	private static final int MThd_MAGIC = 0x4d546864;			// 'MThd'
	private static final int MTrk_MAGIC = 0x4d54726b;			// 'MTrk'

	private static final int ONE_BYTE	= 1;
	private static final int TWO_BYTE	= 2;
	private static final int SYSEX		= 3;
	private static final int META		= 4;

	private static final int MIDI_TYPE_0 = 0;
	private static final int MIDI_TYPE_1 = 1;

	private InputStream stream;									// input stream
	private DataInputStream dis;								// data input stream
	private Sequence sequence;									// sequence object

	private short type;											// MIDI file type (0, 1, or 2)
	private short tracks;										// number of tracks

	private long tick;											// current tick during parsing
	private int status;											// current status byte during parsing

	private int bytesRemaining = 0;								// bytes remaining in current chunk
																// we don't recurse, so this should be okay

	private static final int bufferSize	= 16384;				// buffersize for write
	private DataOutputStream tddos;								// data output stream for track writing


	private static final int bisBufferSize = 1024;

	/**
	 * MIDI parser types
	 */
	public static final int types[] = {

		MIDI_TYPE_0,
		MIDI_TYPE_1
	};

	/**
	 * Obtains the MIDI file format of the input stream provided.  The stream must
	 * point to valid MIDI file data.  In general, MIDI file providers may 
	 * need to read some data from the stream before determining whether they
	 * support it.  These parsers must
	 * be able to mark the stream, read enough data to determine whether they 
	 * support the stream, and, if not, reset the stream's read pointer to its original 
	 * position.  If the input stream does not support this, this method may fail
	 * with an IOException. 
	 * @param stream the input stream from which file format information should be
	 * extracted
	 * @return an <code>MidiFileFormat</code> object describing the MIDI file format
	 * @throws InvalidMidiDataException if the stream does not point to valid MIDI
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 * @see InputStream#markSupported
	 * @see InputStream#mark
	 */
	public MidiFileFormat getMidiFileFormat(InputStream stream) throws InvalidMidiDataException, IOException {


		int maxReadLength = 16;
		int magic;
		int length;
		short filetype;
		short numtracks;
		short timing;
		float divisionType;
		int resolution;
		int duration = MidiFileFormat.UNKNOWN_LENGTH;
		MidiFileFormat format = null;
		int type;

		DataInputStream dis = new DataInputStream( stream );

		dis.mark(maxReadLength);

		magic = dis.readInt();

		if( !(magic == MThd_MAGIC) ) {

			// not MIDI, build and return exception
			dis.reset();
			throw new InvalidMidiDataException("not a valid MIDI file");
			
		} 

		length = dis.readInt();

		length += 8;
		filetype = dis.readShort();
		numtracks = dis.readShort();
		timing = dis.readShort();

		type = (int)filetype;

		/*
		switch(filetype){
		case 0:
			type = MIDI_TYPE_0;
			break;
		case 1:
			type = MIDI_TYPE_1;
			break;
		case 2:
			type = MIDI_TYPE_2;
			break;
		default:
			dis.reset();
		throw new InvalidMidiDataException("not a valid MIDI file");
		}
		*/	

		// decipher the timing code
		if (timing > 0) {
			// tempo based timing.  value is ticks per beat.
			divisionType = Sequence.PPQ;
			resolution = timing;
		} else {
			// SMPTE based timing.  first decipher the frame code.

			int frameCode = (-1 * timing) >> 8;

			switch(frameCode) {

				case 24:	
					divisionType = Sequence.SMPTE_24;
					break;
				case 25:	
					divisionType = Sequence.SMPTE_25;
					break;
				case 29:	
					divisionType = Sequence.SMPTE_30DROP;
					break;
				case 30:	
					divisionType = Sequence.SMPTE_30;
					break;
				default:
					throw new InvalidMidiDataException("Unknown frame code: " + frameCode);
			}

			// now determine the timing resolution in ticks per frame.
			resolution = timing & 0xFF;
		}

		dis.reset();

		// our previous value for length was the length of the header, 
		// not the length of the file.
		length=MidiFileFormat.UNKNOWN_LENGTH;

		format = new MidiFileFormat( type, divisionType, resolution, length, duration );

		return format;
	}


	/**
	 * Obtains the MIDI file format of the URL provided.  The URL must
	 * point to valid MIDI file data.
	 * @param url the URL from which file format information should be
	 * extracted
	 * @return an <code>MidiFileFormat</code> object describing the MIDI file format
	 * @throws InvalidMidiDataException if the URL does not point to valid MIDI
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 */
	public MidiFileFormat getMidiFileFormat(URL url) throws InvalidMidiDataException, IOException {

		InputStream				urlStream = null;
		BufferedInputStream		bis = null;
		MidiFileFormat			fileFormat = null;

		urlStream = url.openStream();	// throws IOException

		bis = new BufferedInputStream( urlStream, bisBufferSize );

		fileFormat = getMidiFileFormat( bis );		// throws InvalidMidiDataException
		
		bis.close();

		return fileFormat;
	}


	/**
	 * Obtains the MIDI file format of the File provided.  The File must
	 * point to valid MIDI file data.
	 * @param file the File from which file format information should be
	 * extracted
	 * @return an <code>MidiFileFormat</code> object describing the MIDI file format
	 * @throws InvalidMidiDataException if the File does not point to valid MIDI
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 */
	public MidiFileFormat getMidiFileFormat(File file) throws InvalidMidiDataException, IOException {

		FileInputStream			fis = null;
		BufferedInputStream		bis = null;
		MidiFileFormat			fileFormat = null;

		fis = new FileInputStream( file );	// throws IOException

		bis = new BufferedInputStream( fis, bisBufferSize );

		fileFormat = getMidiFileFormat( bis );		// throws UnsupportedAudioFileException
		
		bis.close();

		return fileFormat;
	}


	/**
	 * Obtains a MIDI sequence from the input stream provided.  The stream must
	 * point to valid MIDI file data.  In general, MIDI file providers may 
	 * need to read some data from the stream before determining whether they
	 * support it.  These parsers must
	 * be able to mark the stream, read enough data to determine whether they 
	 * support the stream, and, if not, reset the stream's read pointer to its original 
	 * position.  If the input stream does not support this, this method may fail
	 * with an IOException. 
	 * @param stream the input stream from which the <code>Sequence</code> should be
	 * constructed
	 * @return an <code>Sequence</code> object based on the MIDI file data contained
	 * in the input stream.
	 * @throws InvalidMidiDataException if the stream does not point to valid MIDI
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 * @see InputStream#markSupported
	 * @see InputStream#mark
	 */
	public Sequence getSequence(InputStream stream) throws InvalidMidiDataException, IOException {

		MidiFileFormat format = getMidiFileFormat( stream ); // throws InvalidMidiDataException
		
		this.stream = stream;
		
		// construct the sequence object
		sequence = constructSequence();	// can throw InvalidMidiDataException

		for (int i = 0; i < tracks; i++) {

			findTrack();
			readTrack();	// can throw InvalidMidiDataException
			//Track track = readTrack();	// can throw InvalidMidiDataException
			//sequence.add(track);
		}

		return sequence;
	}


	/**
	 * Obtains a MIDI sequence from the URL provided.  The URL must
	 * point to valid MIDI file data.
	 * @param url the URL for which the <code>Sequence</code> should be
	 * constructed
	 * @return an <code>Sequence</code> object based on the MIDI file data pointed
	 * to by the URL
	 * @throws InvalidMidiDataException if the URL does not point to valid MIDI
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 */
	public  Sequence getSequence(URL url) throws InvalidMidiDataException, IOException {

		InputStream				urlStream = null;
		BufferedInputStream		bis = null;

		urlStream = url.openStream();	// throws IOException

		bis = new BufferedInputStream( urlStream, bisBufferSize );

		return getSequence( (InputStream)bis );
	}


	/**
	 * Obtains a MIDI sequence from the File provided.  The File must
	 * point to valid MIDI file data.
	 * @param file the File for which the <code>Sequence</code> should be
	 * constructed
	 * @return an <code>Sequence</code> object based on the MIDI file data pointed
	 * to by the File
	 * @throws InvalidMidiDataException if the File does not point to valid MIDI
	 * file data recognized by the system
	 * @throws IOException if an I/O exception occurs
	 */
	public  Sequence getSequence(File file) throws InvalidMidiDataException, IOException {

		FileInputStream			fis = null;
		BufferedInputStream		bis = null;

		fis = new FileInputStream( file );	// throws IOException

		bis = new BufferedInputStream( fis, bisBufferSize );

		return getSequence( (InputStream)bis );
	}





//=============================================================================================================



	private Sequence constructSequence() throws IOException, InvalidMidiDataException {

		String MThdString = new String("MThd");		// chunk ID string to compare against
		float divisionType;							// timing division type
		int resolution;								// timing resolution
		
		
		PushbackInputStream pis = new PushbackInputStream(stream);
		dis = new DataInputStream(pis);

		// read the chunk name, which should be MThdString
		String chunkString = getChunkName();

		if (! (chunkString.equals(MThdString)) ) {
			
			throw new InvalidMidiDataException("Invalid chunk name: "  + chunkString + ": not a MIDI file"); 
		}

		// read the chunk length
		bytesRemaining = dis.readInt();

		// read the file type
		type = dis.readShort();
		bytesRemaining -= 2;

		// must be MIDI Type 0 or Type 1
		if ( (type != 0) && (type != 1) ) {

			throw new InvalidMidiDataException("Invalid or unsupported file type: "  + type); 
		}

		// read the number of tracks
		tracks = dis.readShort();
		bytesRemaining -= 2;

		// read the timing code
		short timing = dis.readShort();
		bytesRemaining -= 2;

		// decipher the timing code
		if (timing > 0) {

			// tempo based timing.  value is ticks per beat.
			divisionType = Sequence.PPQ;
			resolution = timing;

		} else {

			// SMPTE based timing.  first decipher the frame code.

			int frameCode = (-1 * timing) >> 8;

			switch(frameCode) {

				case 24:	
					divisionType = Sequence.SMPTE_24;
					break;
				case 25:	
					divisionType = Sequence.SMPTE_25;
					break;
				case 29:	
					divisionType = Sequence.SMPTE_30DROP;
					break;
				case 30:	
					divisionType = Sequence.SMPTE_30;
					break;
				default:
					throw new InvalidMidiDataException("Unknown frame code: " + frameCode);
			}

			// now determine the timing resolution in ticks per frame.
			resolution = timing & 0xFF;
		}


		// skip any remaining bytes in the sequence
		dis.skip(bytesRemaining);

		Sequence sequence = new Sequence(divisionType, resolution);		// can throw InvalidMidiDataException			
		return sequence;
	}

	
	private void findTrack() throws IOException, InvalidMidiDataException {

		String MTrkString = new String("MTrk");		// chunk ID string to compare against
		String chunkString;							// chunk string read


		do {
		
			// read the chunk name, which should be MTrkString
			chunkString = getChunkName();

			if (! (chunkString.equals(MTrkString)) ) {
				

				// read the chunk length
				bytesRemaining = dis.readInt();

				dis.skip(bytesRemaining);
			}
		} while (! (chunkString.equals(MTrkString)) );

	}


	private Track readTrack() throws IOException, InvalidMidiDataException {


		// reset current tick to 0
		tick = 0;

		// reset current status byte to 0 (invalid value).
		// this should cause us to throw an InvalidMidiDataException if we don't
		// get a valid status byte from the beginning of the track. 
		status = 0;

		//Track track = new Track();
		Track track = sequence.createTrack();

		// read the chunk length
		bytesRemaining = dis.readInt();

		while (bytesRemaining > 0) {
			track.add(readEvent());
		}

		return track;
	}


	private MidiEvent readEvent() throws IOException, InvalidMidiDataException {
		
		MidiEvent event;

		int data1 = -1;		// initialize to invalid value
		int data2 = 0;


		// each event has a tick delay and then the event data.

		// first read the delay (a variable-length int) and update our tick value
		tick += readVarInt();

		// check for new status
		int byteValue = dis.readUnsignedByte();
		bytesRemaining--;


		// $$kk: 01.22.99: does running status work with meta and sysex event?
		// this code will not work if it does...

		if (byteValue >= 0x80) {

			status = byteValue;

		} else {

			data1 = byteValue;
		}

		int type = getType(status);

		switch(type) {
			
		case ONE_BYTE: 
	
			if (data1 == -1) {
			
				data1 = dis.readUnsignedByte();
				bytesRemaining--;
			}
	
			ShortMessage oneByteMessage = new ShortMessage();
			oneByteMessage.setMessage(status, data1, data2);
			event = new MidiEvent(oneByteMessage, tick);
			return event;

		case TWO_BYTE: 


			if (data1 == -1) {
			
				data1 = dis.readUnsignedByte();
				bytesRemaining--;
			}

			data2 = dis.readUnsignedByte();
			bytesRemaining--;

			ShortMessage typeByteMessage = new ShortMessage();
			typeByteMessage.setMessage(status, data1, data2);
			event = new MidiEvent(typeByteMessage, tick);
			return event;

		case SYSEX:

			int sysexLength = readVarInt();

			byte[] sysexData = new byte[sysexLength];
			bytesRemaining -= dis.read(sysexData); 

			SysexMessage sysexMessage = new SysexMessage();
			sysexMessage.setMessage(status, sysexData, sysexLength);
			event = new MidiEvent(sysexMessage, tick);
			return event;
			
		case META:
			
			int metaType = dis.readUnsignedByte();
			bytesRemaining--;
			int metaLength = readVarInt();


			byte[] metaData = new byte[metaLength];
			bytesRemaining -= dis.read(metaData); 

			MetaMessage metaMessage = new MetaMessage();
			metaMessage.setMessage(metaType, metaData, metaLength);
			event = new MidiEvent(metaMessage, tick);
			return event;

		default:

			throw new InvalidMidiDataException("Internal parser error");

		} // switch
	}

	/**
	 * Returns ONE_BYTE, TWO_BYTE, SYSEX, or META
	 */
	private int getType(int byteValue) throws InvalidMidiDataException {
	
		if ( (byteValue & 0xF0) == 0xF0 ) {

			switch(byteValue) {

			case 0xF0:
			case 0xF7:
				return SYSEX;
			case 0xFF:
				return META;
			default:
				throw new InvalidMidiDataException("Invalid status byte: " + byteValue);				 
			}
		}

		switch(byteValue & 0xF0) {

		case 0x80:
		case 0x90:
		case 0xA0:
		case 0xB0:
		case 0xE0:
			return TWO_BYTE;
		case 0xC0:
		case 0xD0:
			return ONE_BYTE;
		default:
			throw new InvalidMidiDataException("Invalid status byte: " + byteValue);				 
		}
	}


	private String getChunkName() throws IOException {

		byte[] chunkBytes = new byte[4];			// chunk ID bytes (four characters)		

		// read the chunk name
		dis.read(chunkBytes);
		return new String(chunkBytes);
	}


	private int readVarInt() throws IOException, InvalidMidiDataException {

		int value = 0;			// the variable-lengh int value
		int currentByte = 0;	// current byte
		int MAX_LENGTH = 6;		// max number of bytes comprising a variable-length int
								// from kientzle p. 277; is this right??
								
		for (int i = 0; i < MAX_LENGTH; i++) {

			currentByte = dis.readUnsignedByte();
			bytesRemaining--;

			value = (value << 7) + (currentByte & 0x7F);


			if ((currentByte & 0x80) == 0)
				break;
		}									
	
		// make sure we terminated properly
		if ((currentByte & 0x80) != 0) {
		
			throw new InvalidMidiDataException("Unterminated variable-length integer");
		}								
		
		return value;		
	}


}

