/*
 * @(#)StandardMidiFileWriter.java	1.15 00/02/02
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.SequenceInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.lang.IllegalArgumentException;
import java.io.OutputStream;
import java.util.Vector;

import javax.sound.midi.MidiFileFormat;	
import javax.sound.midi.InvalidMidiDataException;	
import javax.sound.midi.MidiEvent;	
import javax.sound.midi.MetaMessage;	
import javax.sound.midi.MidiMessage;	
import javax.sound.midi.Sequence;	
import javax.sound.midi.ShortMessage;	
import javax.sound.midi.SysexMessage;	
import javax.sound.midi.Track;	
import javax.sound.midi.spi.MidiFileWriter;


/**
 * MIDI file writer.
 *
 * @version 1.15 00/02/02
 * @author Kara Kytle
 * @author Jan Borgersen
 */
public class StandardMidiFileWriter extends SunMidiFileWriter {

	private static final int MThd_MAGIC = 0x4d546864;			// 'MThd'
	private static final int MTrk_MAGIC = 0x4d54726b;			// 'MTrk'

	private static final int ONE_BYTE	= 1;
	private static final int TWO_BYTE	= 2;
	private static final int SYSEX		= 3;
	private static final int META		= 4;

	private static final int MIDI_TYPE_0 = 0;
	private static final int MIDI_TYPE_1 = 1;
	private static final int MIDI_TYPE_2 = 2;

	private InputStream stream;									// input stream
	private DataInputStream dis;								// data input stream
	private Sequence sequence;									// sequence object

	private short type;											// MIDI file type (0, 1, or 2)
	private short tracks;										// number of tracks

	private long tick;											// current tick during parsing
	private int status;											// current status byte during parsing

	private int bytesRemaining = 0;								// bytes remaining in current chunk
																// we don't recurse, so this should be okay

	private static final int bufferSize	= 16384;				// buffersize for write
	private DataOutputStream tddos;								// data output stream for track writing



	/**
	 * MIDI parser types
	 */
	public static final int types[] = {

		MIDI_TYPE_0,
		MIDI_TYPE_1
	};


	/**
	 * new
	 */
	public int[] getMidiFileTypes() {

		int[] localArray = new int[types.length];
		System.arraycopy(types, 0, localArray, 0, types.length);
		return localArray;
	}
	
	/**
	 * Obtains the file types that this provider can write from the
	 * sequence specified.
	 * @param sequence the sequence for which midi file type support
	 * is queried
	 * @return array of file types.  If no file types are supported, 
	 * returns an array of length 0.
	 */
	public  int[] getMidiFileTypes(Sequence sequence){
		
		int typesArray[];
		Track tracks[] = sequence.getTracks();

		if( tracks.length==1 ) {
			typesArray = new int[2];
			typesArray[0] = MIDI_TYPE_0;
			typesArray[1] = MIDI_TYPE_1;
		} else {
			typesArray = new int[1];
			typesArray[0] = MIDI_TYPE_1;
		}

		return typesArray;
	}
	/**
	 * new
	 */
	public boolean isFileTypeSupported(int type) {

		for(int i=0; i<types.length; i++) {
			if( type == types[i] ) {
				return true;
			}
		}
		return false;
	}


	/*
	 * deleted
	 */
	/**
	 * Obtains the file formats of the indicated type that this provider 
	 * can write from the sequence specified.
	 * @param type desired midi file type
	 * @param sequence the sequence for which midi file format support
	 * is queried
	 * @return array of midi file formats of the type indicated.  If no 
	 * file formats of this type are supported, returns an array of length 0.
	 */
/*	public  MidiFileFormat[] getMidiFileFormats(int type, Sequence sequence) { 
		
		MidiFileFormat format = null;
		MidiFileFormat formatsArray[] = new MidiFileFormat[1];
		boolean flag = false;

		for(int i=0; i<types.length; i++) {
			if( type == types[i] ) {
				flag = true;
				break;
			}
		}
		if( !flag ) {
			return new MidiFileFormat[0];
		}

		format = new MidiFileFormat( type, 
							sequence.getDivisionType(), 
							sequence.getResolution(),
							MidiFileFormat.UNKNOWN_LENGTH,
							sequence.getMicrosecondLength() );

		formatsArray[0] = format;

		return formatsArray;
	}
*/

	/*
	 * changed -- concrete in spi class now
	 */
	/**
	 * Indicates whether an midi file of the file format specified can be written
	 * from the sequence indicated.
	 * @param fileFormat description of the file for which write capabilities are queried
	 * @param sequence  the sequence for which file writing support is queried
	 * @return <code>true</code> if the file format is supported for this sequence, 
	 * otherwise <code>false</code>
	 */
/*	public  boolean isFileFormatSupported(MidiFileFormat fileFormat, Sequence sequence) {


		Track tracks[] = sequence.getTracks();
		
		if( (fileFormat.getType()==MIDI_TYPE_1) ||
			((fileFormat.getType()==MIDI_TYPE_0) && (tracks.length==1)) ) {

			if( fileFormat.getResolution()==sequence.getResolution() ) {

				if( fileFormat.getDivisionType()==sequence.getDivisionType() ) {

					if( (fileFormat.getMicrosecondLength()==MidiFileFormat.UNKNOWN_LENGTH) ||
						(fileFormat.getMicrosecondLength()==sequence.getMicrosecondLength()) ) {

						// don't check the length field...

						return true;
					}
				}
			}
		} 
		
		// no matches ... return false

		return false;
	}
*/
	/**
	 * Writes a stream of bytes representing an midi file of the file format
	 * indicated to the output stream provided. 
	 * @param in sequence containing midi data to be written to the file
	 * @param format format of the file to be written to the output stream
	 * @param out stream to which the file data should be written
	 * @return the number of bytes written to the output stream
	 * @throws IOException if an I/O exception occurs
	 * @throws IllegalArgumentException if the file format is not supported by
	 * the system  
	 * @see #isFileFormatSupported
	 * @see	#getMidiFileFormats
	 */
	public int write(Sequence in, int type, OutputStream out) throws IOException {

		InputStream fileStream;

		byte [] buffer = null;

		long bytesRead = 0;
		long bytesWritten = 0;

		if( !isFileTypeSupported(type,in) ) {
			throw new IllegalArgumentException("Could not write MIDI file");
		}
		// First get the fileStream from this sequence
		if( (fileStream=getFileStream(type,in))==null ) {
			throw new IllegalArgumentException("Could not write MIDI file");
		}
		buffer = new byte[bufferSize];

		while( (bytesRead = fileStream.read( buffer )) >= 0 ) {
			out.write( buffer, 0, (int)bytesRead );
			bytesWritten += bytesRead;
		}
		// Done....return bytesWritten
		return (int) bytesWritten;

	}
	 
/*	public int write(Sequence in, MidiFileFormat format, OutputStream out) throws IOException { 
	
		InputStream fileStream;

		byte [] buffer = null;

		long bytesRead = 0;
		long bytesWritten = 0;

		int maxLength = (int) format.getByteLength();

		if( !isFileFormatSupported(format,in) ) {
			throw new IllegalArgumentException("invalid MidiFileFormat");
		}


		// First get the fileStream from this sequence
		// $$jb: change getFileStream to throw IllegalArgumentExceptions!!!!
		if ( (fileStream=getFileStream(format.getType(),in))==null) {

			throw new IllegalArgumentException("could not write MIDI file");
		}
	
		buffer = new byte[bufferSize];

		while( (bytesRead = fileStream.read( buffer )) >=0 ) {
			if (maxLength>0) {
				if( bytesRead < maxLength ) {
					out.write( buffer, 0, (int)bytesRead );
					bytesWritten += bytesRead;
					maxLength -= bytesRead;
				} else {
					out.write( buffer, 0, (int)maxLength );
					bytesWritten += maxLength;
					maxLength = 0;
					break;
				}
			} else {
				out.write( buffer, 0, (int)bytesRead );
				bytesWritten += bytesRead;
			}
		}

		// Done....return bytesWritten.
		return (int) bytesWritten;

	}
*/
	/**
	 * Writes a stream of bytes representing an midi file of the file format
	 * indicated to the external file provided.
	 * @param in sequence containing midi data to be written to the file
	 * @param format format of the file to be written to the output stream
	 * @param out external file to which the file data should be written
	 * @return the number of bytes written to the file
	 * @throws IOException if an I/O exception occurs
	 * @throws IllegalArgumentException if the file format is not supported by
	 * the system  
	 * @see #isFileFormatSupported
	 * @see	#getMidiFileFormats
	 */
/*	public  int write(Sequence in, MidiFileFormat format, File out) throws IOException { */
	public int write(Sequence in, int type, File out) throws IOException {

		FileOutputStream		fos = null;
		BufferedOutputStream	bos = null;

		int bytesWritten = 0;
		
		fos = new FileOutputStream( out );	// throws IOException

		bytesWritten = write( in, type, fos );

		fos.close();

		return bytesWritten;

	}




	//=================================================================================




	/**
	 * Obtains a file stream from the sequence provided.
	 * @param type type of file stream to get from sequence, or null for default
	 * @param sequence sequence containing events to be written to a file
	 * @return stream from which file data may be read, or null if a file stream 
	 * cannot be obtained from the sequence provided.
	 * @throws IOException if an I/O error occurs
	 */
	private InputStream getFileStream(int type, Sequence sequence) throws IOException {

		Track tracks[] = sequence.getTracks();
		int bytesBuilt = 0;
		int headerLength = 14;
		int length = 0;
		int timeFormat;
		float divtype;

		PipedOutputStream		hpos = null;
		DataOutputStream		hdos = null;
		PipedInputStream		headerStream = null;

		InputStream				trackStreams [] = null;
		InputStream				trackStream = null;
		SequenceInputStream		fStream = null;


		// Determine the filetype to write
		if( type==MIDI_TYPE_0 ) {
			if (tracks.length != 1) {
				return null;
			}
		} else if( type==MIDI_TYPE_1 ) {
			if (tracks.length < 1) { // $$jb: 05.31.99: we _can_ write TYPE_1 if tracks.length==1
				return null;
			}
		} else {
			if(tracks.length==1) {
				type = MIDI_TYPE_0;
			} else if(tracks.length>1) {
				type = MIDI_TYPE_1;
			} else {
				return null;
			}
		}

		// Now build the file one track at a time
		// Note that above we made sure that MIDI_TYPE_0 only happens
		// if tracks.length==1
		
		trackStreams = new InputStream[tracks.length];
		for(int i=0; i<tracks.length; i++) {


			try {
				trackStreams[i] = writeTrack( tracks[i], type );
			} catch (InvalidMidiDataException e) {
				// $$jb: 04.08.99: What to do here?
				if(Printer.err) Printer.err("Exception in write: " + e.getMessage());
			}
			//bytesBuilt += trackStreams[i].getLength();

		}



		// Now seqence the track streams
		if( tracks.length == 1 ) {
			trackStream = (InputStream) trackStreams[0];
		} else if( tracks.length > 1 ){

			trackStream = (InputStream) trackStreams[0];

			for(int i=1; i<tracks.length; i++) {

				trackStream = (InputStream) (new SequenceInputStream( trackStream, trackStreams[i]));
			}
		}
		

		// Now build the header...
		hpos = new PipedOutputStream();
		hdos = new DataOutputStream(hpos);
		headerStream = new PipedInputStream(hpos);

		// Write the magic number
		hdos.writeInt( MThd_MAGIC );

		// Write the header length
		hdos.writeInt( headerLength - 8 );

		// Write the filetype
		if(type==MIDI_TYPE_0) {
			hdos.writeShort( 0 );
		} else {
			// MIDI_TYPE_1
			hdos.writeShort( 1 );
		}

		// Write the number of tracks
		hdos.writeShort( (short) tracks.length );


		// Determine and write the timing format
		divtype = sequence.getDivisionType();
		if( divtype == Sequence.PPQ ) {
			timeFormat = sequence.getResolution();

		} else if( divtype == Sequence.SMPTE_24) {
			timeFormat = (24<<8) * -1;
			timeFormat += (sequence.getResolution() & 0xFF);

		} else if( divtype == Sequence.SMPTE_25) {
			timeFormat = (25<<8) * -1;
			timeFormat += (sequence.getResolution() & 0xFF);
		
		} else if( divtype == Sequence.SMPTE_30DROP) {
			timeFormat = (29<<8) * -1;
			timeFormat += (sequence.getResolution() & 0xFF);
		
		} else if( divtype == Sequence.SMPTE_30) {
			timeFormat = (30<<8) * -1;
			timeFormat += (sequence.getResolution() & 0xFF);
		
		} else {
			// $$jb: 04.08.99: What to really do here?
			return null;
		}
		hdos.writeShort( timeFormat );

		// now construct an InputStream to become the FileStream
		fStream = new SequenceInputStream(headerStream, trackStream);
		hdos.close();

		length = bytesBuilt + headerLength;
		
		return fStream;
	}



	/**
	 * Returns ONE_BYTE, TWO_BYTE, SYSEX, or META
	 */
	private int getType(int byteValue) throws InvalidMidiDataException {
	
		if ( (byteValue & 0xF0) == 0xF0 ) {

			switch(byteValue) {

			case 0xF0:
			case 0xF7:
				return SYSEX;
			case 0xFF:
				return META;
			default:
				throw new InvalidMidiDataException("Invalid status byte: " + byteValue);				 
			}
		}

		switch(byteValue & 0xF0) {

		case 0x80:
		case 0x90:
		case 0xA0:
		case 0xB0:
		case 0xE0:
			return TWO_BYTE;
		case 0xC0:
		case 0xD0:
			return ONE_BYTE;
		default:
			throw new InvalidMidiDataException("Invalid status byte: " + byteValue);				 
		}
	}


	private int writeVarInt(int value ) throws IOException, InvalidMidiDataException {

		int MAX_LENGTH = 6;
		byte bytes[] = new byte[MAX_LENGTH];
		int length = 0;
		int currentByte = 0;

		// first make sure our array is zeroed
		for(int i=0; i<MAX_LENGTH; i++) bytes[i] = 0;

		// we fill this array up from the end so that bytes will be
		// written to the stream msb first
		for(int i=MAX_LENGTH-1; i >= 0; i--) {

			bytes[i] = (byte) (value & 0x7F);
			value = value >>> 7;
			length++;

			if( length>1 ) {
				// this is not the last byte to be written to the stream,
				// so change the msb to 1
				bytes[i] |= 0x80;
			} 
			if( value==0 ) {
				// we don't need any more bytes to store this integer
				break;
			}
		}
		// make sure we terminated properly
		if ( (bytes[MAX_LENGTH-1] & 0x80) != 0 ) {

			// $$jb: is this what we want to do here?  will this ever happen?
			throw new InvalidMidiDataException("Unable to create variable-length integer");
		}

		// now write our bytes, msb first, to the stream
		tddos.write(bytes, (MAX_LENGTH-length), length);

		return length;
	}

	private InputStream writeTrack( Track track, int type ) throws IOException, InvalidMidiDataException {

		int bytesWritten = 0;
		int lastBytesWritten = 0;
		int size = track.size();
		PipedOutputStream thpos = new PipedOutputStream();
		DataOutputStream  thdos = new DataOutputStream(thpos);
		PipedInputStream  thpis = new PipedInputStream(thpos);
		
		ByteArrayOutputStream tdbos = new ByteArrayOutputStream();
		tddos = new DataOutputStream(tdbos);
		ByteArrayInputStream tdbis = null;

		SequenceInputStream  fStream = null;

		int currentTick = 0;
		int deltaTick = 0;
		int eventTick = 0;
		int runningStatus = -1;

		// -----------------------------
		// Write each event in the track
		// -----------------------------


		currentTick = 0;
		for(int i=0; i<size; i++) {

			MidiEvent event = track.get(i);
			
			int status;
			int eventtype;
			int metatype;
			int data1, data2;
			int length;
			byte data[] = null;
			ShortMessage shortMessage = null;
			MetaMessage  metaMessage  = null;
			SysexMessage sysexMessage = null;
			
			
			// get the tick
			// $$jb: this gets easier if we change all system-wide time to delta ticks
			eventTick = (int)event.getTick();
			deltaTick = (int)event.getTick() - currentTick;
			currentTick = (int)event.getTick();


			// get the status byte
			status = event.getMessage().getStatus();
			eventtype = getType( status );


			switch( eventtype ) {
			case ONE_BYTE:
				shortMessage = (ShortMessage) event.getMessage();
				data1 = shortMessage.getData1();
				bytesWritten += writeVarInt( deltaTick );
				
				if(status!=runningStatus) {
					runningStatus=status;
					tddos.writeByte(status);  bytesWritten += 1;
				} 
				tddos.writeByte(data1);   bytesWritten += 1;
				break;

			case TWO_BYTE:
				shortMessage = (ShortMessage) event.getMessage();
				data1 = shortMessage.getData1();
				data2 = shortMessage.getData2();
	
				bytesWritten += writeVarInt( deltaTick );
				if(status!=runningStatus) {
					runningStatus=status;
					tddos.writeByte(status);  bytesWritten += 1;
				} 
				tddos.writeByte(data1);   bytesWritten += 1;
				tddos.writeByte(data2);   bytesWritten += 1;
				break;

			case SYSEX:
				sysexMessage = (SysexMessage) event.getMessage();
				length     = sysexMessage.getLength();
				//data       = sysexMessage.getData();
				data       = sysexMessage.getMessage();

				bytesWritten += writeVarInt( deltaTick );

				// $$jb: 04.08.99: always write status for sysex
				runningStatus=status;
				tddos.writeByte( data[0] ); bytesWritten += 1;
				//tddos.writeByte(status);  bytesWritten += 1;

				// $$jb: 10.18.99: we don't maintain length in
				// the message data for SysEx (it is not transmitted
				// over the line), so write the calculated length
				// minus the status byte
				bytesWritten += writeVarInt( (data.length-1) );

				// $$jb: 10.18.99: now write the rest of the
				// message
				tddos.write( data, 1, (data.length-1) );
				bytesWritten += (data.length-1);
				break;

			case META:
				metaMessage = (MetaMessage) event.getMessage();
				//metatype  = metaMessage.getType();
				length    = metaMessage.getLength();
				//data      = metaMessage.getData();
				data      = metaMessage.getMessage();

				bytesWritten += writeVarInt( deltaTick );

				// $$jb: 10.18.99: getMessage() returns the
				// entire valid midi message for a file, 
				// including the status byte and the var-length-int 
				// length value, so we can just write the data
				// here.  note that we must _always_ write the
				// status byte, regardless of runningStatus.

				runningStatus=status;
				//tddos.writeByte(data[0]); bytesWritten += 1;
				//tddos.writeByte(status);  bytesWritten += 1;
			
				//tddos.writeByte(data[1]); bytesWritten += 1;
				//tddos.writeByte(metatype); 
				//bytesWritten += 1;
			
				//bytesWritten += writeVarInt( (data.length-2) );
			
				tddos.write( data, 0, data.length );
				bytesWritten += data.length;

				break;

			default:
				// $$jb: 04.08.99: What to do here?
				throw new InvalidMidiDataException("Invalid midi event");
			}
		} 
		// ---------------------------------
		// End write each event in the track
		// ---------------------------------

		// Build Track header now that we know length
		thdos.writeInt(MTrk_MAGIC);
		thdos.writeInt(bytesWritten);
		bytesWritten += 8;
	
		// Now sequence them
		tdbis = new ByteArrayInputStream( tdbos.toByteArray() );
		fStream = new SequenceInputStream(thpis,tdbis);
		thdos.close();
		tddos.close();

		return fStream;
	}

}

