/*
 * @(#)UlawCodec.java	1.38 00/03/15
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package com.sun.media.sound;	  	 

import java.io.InputStream;
import java.io.IOException;

import java.util.Vector;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.AudioInputStream;

	
/**
 * U-law encodes linear data, and decodes u-law data to linear data. 
 *
 * @version 1.38 00/03/15
 * @author Kara Kytle
 */					  
public class UlawCodec extends SunCodec {

	/* Tables used for U-law decoding */

	final static byte ULAW_TABH[] = new byte[256];
	final static byte ULAW_TABL[] = new byte[256];

	private static final boolean ulawEncodes = true;
	private static final boolean ulawDecodes = true;

	private static final AudioFormat.Encoding[] ulawEncodings = { AudioFormat.Encoding.ULAW, AudioFormat.Encoding.PCM_SIGNED };

	private static final short seg_end [] = {0xFF, 0x1FF, 0x3FF, 
		0x7FF, 0xFFF, 0x1FFF, 0x3FFF, 0x7FFF};

	private static final int tempBufferSize = 64;
	private byte tempBuffer [] = null;
	
	/**
	 * Initializes the decode tables
	 */
	static {
		for (int i=0;i<256;i++) {
			int ulaw = ~i;
			int t;

			ulaw &= 0xFF;
			t = ((ulaw & 0xf)<<3) + 132;
			t <<= ((ulaw & 0x70) >> 4);
			t = ( (ulaw&0x80) != 0 ) ? (132-t) : (t-132);
			
			ULAW_TABL[i] = (byte) (t&0xff);
			ULAW_TABH[i] = (byte) ((t>>8) & 0xff);
		}
	}
	

	/**
	 * Constructs a new ULAW codec object.
	 */
	public UlawCodec() {
	
		super(ulawEncodings, ulawEncodings);
	}


// NEW CODE


	/**
	 */
	public AudioFormat.Encoding[] getTargetEncodings(AudioFormat sourceFormat){

		if( sourceFormat.getEncoding().equals( AudioFormat.Encoding.PCM_SIGNED )) {

			if( sourceFormat.getSampleSizeInBits() == 16 ) {

				AudioFormat.Encoding enc[] = new AudioFormat.Encoding[1];
				enc[0] = AudioFormat.Encoding.ULAW;
				return enc;

			} else {
				return new AudioFormat.Encoding[0];
			}
		} else if( sourceFormat.getEncoding().equals( AudioFormat.Encoding.ULAW ) ) {

			if( sourceFormat.getSampleSizeInBits() == 8 ) {

				AudioFormat.Encoding enc[] = new AudioFormat.Encoding[1];
				enc[0] = AudioFormat.Encoding.PCM_SIGNED;
				return enc;

			} else {
				return new AudioFormat.Encoding[0];
			}

		} else {
			return new AudioFormat.Encoding[0];
		}
	}


	/**
	 */
	public AudioFormat[] getTargetFormats(AudioFormat.Encoding targetEncoding, AudioFormat sourceFormat){


		if( (targetEncoding.equals( AudioFormat.Encoding.PCM_SIGNED ) && sourceFormat.getEncoding().equals( AudioFormat.Encoding.ULAW)) ||
			(targetEncoding.equals( AudioFormat.Encoding.ULAW) && sourceFormat.getEncoding().equals( AudioFormat.Encoding.PCM_SIGNED)) ) {
			return getOutputFormats( sourceFormat );
		} else {
			return new AudioFormat[0];
		}
	}

	/**
	 */
	public AudioInputStream getAudioInputStream(AudioFormat.Encoding targetEncoding, AudioInputStream sourceStream){

		AudioFormat sourceFormat = sourceStream.getFormat();
		AudioFormat.Encoding sourceEncoding = sourceFormat.getEncoding();

		if( sourceEncoding.equals( targetEncoding ) ) {

			return sourceStream;

		} else {

			AudioFormat targetFormat = null;

			if( !isConversionSupported(targetEncoding,sourceStream.getFormat()) ) {
				throw new IllegalArgumentException("Unsupported conversion: " + sourceStream.getFormat().toString() + " to " + targetEncoding.toString());
			}
			if( sourceEncoding.equals( AudioFormat.Encoding.ULAW ) &&
				targetEncoding.equals( AudioFormat.Encoding.PCM_SIGNED ) ) {

				targetFormat = new AudioFormat( targetEncoding,
									sourceFormat.getSampleRate(),
									sourceFormat.getSampleSizeInBits() * 2,
									sourceFormat.getChannels(),
									sourceFormat.getFrameSize() * 2,
									sourceFormat.getFrameRate(),
									sourceFormat.isBigEndian() );

			} else if( sourceEncoding.equals( AudioFormat.Encoding.PCM_SIGNED ) &&
				targetEncoding.equals( AudioFormat.Encoding.ULAW ) ) {

				targetFormat = new AudioFormat( targetEncoding,
									sourceFormat.getSampleRate(),
									sourceFormat.getSampleSizeInBits() / 2,
									sourceFormat.getChannels(),
									sourceFormat.getFrameSize() / 2,
									sourceFormat.getFrameRate(),
									sourceFormat.isBigEndian() );

			} else {
				throw new IllegalArgumentException("Unsupported conversion: " + sourceStream.getFormat().toString() + " to " + targetEncoding.toString());
			}

			return getAudioInputStream( targetFormat, sourceStream );
		}
	}
	/**
	 * use old code...
	 */
	public AudioInputStream getAudioInputStream(AudioFormat targetFormat, AudioInputStream sourceStream){
	
		return getConvertedStream( targetFormat, sourceStream );
	}
	




// OLD CODE

	/**
	 * Opens the codec with the specified parameters.
	 * @param stream stream from which data to be processed should be read
	 * @param outputFormat desired data format of the stream after processing
	 * @return stream from which processed data may be read
	 * @throws IllegalArgumentException if the format combination supplied is
	 * not supported.
	 */
/*	public AudioInputStream getConvertedStream(AudioFormat outputFormat, AudioInputStream stream) { */
	private AudioInputStream getConvertedStream(AudioFormat outputFormat, AudioInputStream stream) { 
	
		AudioInputStream cs = null;

		AudioFormat inputFormat = stream.getFormat();

		if( inputFormat.matches(outputFormat) ) {

			cs = stream;
		} else {

			cs = (AudioInputStream) (new UlawCodecStream(stream, outputFormat));
			tempBuffer = new byte[tempBufferSize];
		}
		return cs;
	}


	/**
	 * Determines whether the codec supports conversion from one
	 * particular format to another.
	 * @param inputFormat format of the incoming data
	 * @param desired format of outgoing data
	 * @return true if the conversion is supported, otherwise false
	 */
/*	public boolean isConversionSupported(AudioFormat inputFormat, AudioFormat outputFormat) {

		AudioFormat outputFormat = targetFormat;
		AudioFormat inputFormat = sourceFormat;

		AudioFormat encodeFormat;
		AudioFormat decodeFormat;

		// determine which is the encode and decode format

		if( inputFormat.matches(outputFormat) ) {

			return true;
		}

		if (inputFormat.getEncoding() == AudioFormat.Encoding.ULAW) {
			
			encodeFormat = inputFormat;
			decodeFormat = outputFormat;

		} else {

			encodeFormat = outputFormat;
			decodeFormat = inputFormat;			
		}
			
		// now test whether we can handle the conversion
		
		if ( ! (encodeFormat.getEncoding() == AudioFormat.Encoding.ULAW) ) 
			return false;
			
		if ( ! (decodeFormat.getEncoding() == AudioFormat.Encoding.PCM_SIGNED) ) 
			return false;

		if (encodeFormat.getSampleSizeInBits() != 8) 
			return false;

		if (encodeFormat.getFrameSize() != 1)
			return false;
			
		if (decodeFormat.getSampleSizeInBits() != 16)
			return false;

		if (decodeFormat.getFrameSize() != 2)
			return false;

		if ( (encodeFormat.getChannels() != 1) || (decodeFormat.getChannels() != 1) ) 
			return false;

		if ( (encodeFormat.getSampleRate() != encodeFormat.getFrameRate()) || 
			(decodeFormat.getSampleRate() != decodeFormat.getFrameRate()) ||
			(encodeFormat.getSampleRate() != decodeFormat.getSampleRate()) ) 
			return false;

		return true;
	}
*/
	/**
	 * Obtains the set of output encodings supported by the codec,
	 * given a particular input encoding.
	 * If none are supported, returns an array of length 0.
	 * @return array of supported output encodings. 
	 */
/*	public AudioFormat.Encoding[] getOutputEncodings(AudioFormat.Encoding inputEncoding) {

		Vector encodings = new Vector();

		if ( inputEncoding == AudioFormat.Encoding.PCM_SIGNED ) {

			encodings.addElement(AudioFormat.Encoding.ULAW);
		}
		if ( inputEncoding == AudioFormat.Encoding.ULAW ) {

			encodings.addElement(AudioFormat.Encoding.PCM_SIGNED);
		}

		AudioFormat.Encoding[] encodingArray;

		synchronized(encodings) {

			encodingArray = new AudioFormat.Encoding[encodings.size()];
			
			for (int i = 0; i < encodingArray.length; i++) {
				
				encodingArray[i] = (AudioFormat.Encoding)(encodings.elementAt(i));
			}
		}			

		return encodingArray;
	
	}
*/
	/**
	 * Obtains the set of output formats supported by the codec 
	 * given a particular input format.  
	 * If no output formats are supported for this input format, 
	 * returns an array of length 0.
	 * @return array of supported output formats. 
	 */
/*	public AudioFormat[] getOutputFormats(AudioFormat inputFormat) { */
	private AudioFormat[] getOutputFormats(AudioFormat inputFormat) {

		Vector formats = new Vector();
		AudioFormat format;

		if ( (inputFormat.getSampleSizeInBits() == 16) && (inputFormat.getEncoding() == AudioFormat.Encoding.PCM_SIGNED) ) {
			
			format = new AudioFormat(AudioFormat.Encoding.ULAW, 
										inputFormat.getSampleRate(), 
										8,
										inputFormat.getChannels(),
										inputFormat.getFrameSize()/2,
										inputFormat.getFrameRate(),
										false );
			
			formats.addElement(format); 
		}

		if (inputFormat.getEncoding() == AudioFormat.Encoding.ULAW) {

			format = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, 
										inputFormat.getSampleRate(), 
										16,
										inputFormat.getChannels(),
										inputFormat.getFrameSize()*2,
										inputFormat.getFrameRate(),
										false );

			formats.addElement(format); 

			format = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, 
										inputFormat.getSampleRate(), 
										16,
										inputFormat.getChannels(),
										inputFormat.getFrameSize()*2,
										inputFormat.getFrameRate(),
										true );

			formats.addElement(format); 	
		}

		AudioFormat[] formatArray;

		formatArray = new AudioFormat[formats.size()];
			
		for (int i = 0; i < formatArray.length; i++) {
				
			formatArray[i] = (AudioFormat)(formats.elementAt(i));
		}




		return formatArray;		
	}


	class UlawCodecStream extends AudioInputStream {

		/**
		 * True to encode to u-law, false to decode to linear
		 */
		boolean encode = false;

		AudioFormat encodeFormat;
		AudioFormat decodeFormat;

		byte tabByte1[] = null;
		byte tabByte2[] = null;
		int highByte = 0;
		int lowByte  = 1;
		
		UlawCodecStream(AudioInputStream stream, AudioFormat outputFormat) {

			super(stream, outputFormat, -1);
			
			AudioFormat inputFormat = stream.getFormat();
			
			// throw an IllegalArgumentException if not ok
			if ( ! (isConversionSupported(outputFormat, inputFormat)) ) {

				throw new IllegalArgumentException("Unsupported conversion: " + inputFormat.toString() + " to " + outputFormat.toString());
			}

			// determine whether we are encoding or decoding
			if (inputFormat.getEncoding() == AudioFormat.Encoding.ULAW) {

				encode = false;
				encodeFormat = inputFormat;
				decodeFormat = outputFormat;
			
			} else {

				encode = true;
				encodeFormat = outputFormat;
				decodeFormat = inputFormat;
			}

			// setup tables according to byte order
			if ( decodeFormat.isBigEndian() ) {			

				tabByte1 = ULAW_TABH;
				tabByte2 = ULAW_TABL;
			
			} else {

				tabByte1 = ULAW_TABL;
				tabByte2 = ULAW_TABH;
			}
			// setup encoding bytes according to byte order
			if (encodeFormat.isBigEndian()) {			

				highByte = 0;
				lowByte  = 1;
			
			} else {

				highByte = 1;
				lowByte  = 0;
			}

			// set the AudioInputStream length in frames if we know it
			if (stream instanceof AudioInputStream) {

				long inputLength = ((AudioInputStream)stream).getFrameLength();

				// $$jb: 06.07.99: duh... length is in frames now, not bytes,
				// so it stays the same.

				frameLength = inputLength;

/*				if (inputLength > 0) {

					if (encode) {
						frameLength = inputLength / 2;
					} else {
						frameLength = inputLength * 2;
					}
				}
*/			}
			// set framePos to zero
			framePos = 0;
			frameSize = inputFormat.getFrameSize();
			if(frameSize==AudioSystem.NOT_SPECIFIED) {
				frameSize = 1;
			}
		}

		
		/*
		 * $$jb 2/23/99
		 * Used to determine segment number in uLaw encoding
		 */
		private short search(short val, short table[], short size) {
			for(short i = 0; i < size; i++) {
				if (val <= table[i]) { return i; }
			}
			return size;
		}

		/**
		 * Note that this won't actually read anything; must read in
		 * two-byte units.
		 */ 
		public int read() throws IOException {

			byte[] b = new byte[1];
			return (int)read(b, 0, b.length);
		}


		public int read(byte[] b) throws IOException {

			return read(b, 0, b.length);
		}

		public int read(byte[] b, int off, int len) throws IOException {
		
			// don't read fractional frames
			if( len%frameSize != 0 ) {
				len -= (len%frameSize);
			}
			if (encode) {

				short BIAS = 0x84;
				short mask;
				short seg;
				int i;

				short sample;
				byte enc;

				int readCount = 0;
				int currentPos = off;
				int readLeft = len*2;
				int readLen = ( (readLeft>tempBufferSize) ? tempBufferSize : readLeft );

				while ((readCount = super.read(tempBuffer,0,readLen))>0) {

					for(i = 0; i < readCount; i+=2) {

						/* Get the sample from the tempBuffer */
						sample = (short)(( (tempBuffer[i + highByte]) << 8) & 0xFF00);
						sample |= (short)( (short) (tempBuffer[i + lowByte]) & 0xFF);

						/* Get the sign and the magnitude of the value. */
						if(sample < 0) {
							sample = (short) (BIAS - sample);
							mask = 0x7F;
						} else {
							sample += BIAS;
							mask = 0xFF;
						}
						/* Convert the scaled magnitude to segment number. */
						seg = search(sample, seg_end, (short) 8);
						/*
						 * Combine the sign, segment, quantization bits;
						 * and complement the code word.
						 */
						if (seg >= 8) {  /* out of range, return maximum value. */
							enc = (byte) (0x7F ^ mask);
						} else {
							enc = (byte) ((seg << 4) | ((sample >> (seg+3)) & 0xF));
							enc ^= mask;
						}
						/* Now put the encoded sample where it belongs */
						b[currentPos] = enc;
						currentPos++;
					}
					/* And update pointers and counters for next iteration */
					readLeft -= readCount;
					readLen = ( (readLeft>tempBufferSize) ? tempBufferSize : readLeft );
				}
				if( currentPos==off && readCount<0 ) { 	// EOF or error on read
					return readCount;
				}
				
				return (currentPos - off);  /* Number of bytes written to new buffer */

			} else {

				int i;
				int readLen = len/2;
				int readOffset = off + len/2;
				int readCount = super.read(b, readOffset, readLen);

				if(readCount<0) {		// EOF or error
					return readCount;
				}
				for (i = off; i < (off + (readCount*2)); i+=2) {

					b[i]	= (byte)tabByte1[b[readOffset] & 0xFF];
					b[i+1]	= (byte)tabByte2[b[readOffset] & 0xFF];
					readOffset++;
				}

				return (i - off);
			}
		} 
	} // end class UlawCodecStream

} // end class ULAW
