/*
 * @(#)ShortMessage.java	1.17 00/02/02
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package javax.sound.midi;	

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;


/**
 * A <code>ShortMessage</code> contains a MIDI message that has at most
 * two data bytes following its status byte.  The types of MIDI message
 * that satisfy this criterion are channel voice, channel mode, system common,
 * and system real-time--in other words, everything except system exclusive
 * and meta-events.  The <code>ShortMessage</code> class provides methods
 * for getting and setting the contents of the MIDI message.
 * <p>
 * A number of <code>ShortMessage</code> methods have integer parameters by which
 * you specify a MIDI status or data byte.  If you know the numeric value, you
 * can express it directly.  For system common and system real-time messages,
 * you can often use the corresponding fields of <code>ShortMessage</code>, such as
 * {@link #SYSTEM_RESET SYSTEM_RESET}.  For channel messages,
 * the upper four bits of the status byte are specified by a command value and
 * the lower four bits are specified by a MIDI channel number. To
 * convert incoming MIDI data bytes that are in the form of Java's signed bytes, 
 * you can use the <A HREF="MidiMessage.html#integersVsBytes">conversion code</A>
 * given in the <code>{@link MidiMessage}</code> class description.
 *
 * @see SysexMessage
 * @see MetaMessage
 *
 * @version 1.17, 00/02/02
 * @author David Rivas
 * @author Kara Kytle
 */

public class ShortMessage extends MidiMessage {


// Status byte defines


	// System common messages

	/**
	 * Status byte for MIDI Time Code Quarter Frame message (0xF1, or 241). 
	 * @see MidiMessage#getStatus
	 */
	public static final int MIDI_TIME_CODE				= 0xF1; // 241

	/**
	 * Status byte for Song Position Pointer message (0xF2, or 242). 
	 * @see MidiMessage#getStatus
	 */
	public static final int SONG_POSITION_POINTER		= 0xF2;	// 242

	/**
	 * Status byte for MIDI Song Select message (0xF3, or 243). 
	 * @see MidiMessage#getStatus
	 */
	public static final int SONG_SELECT					= 0xF3; // 243

	/**
	 * Status byte for Tune Request message (0xF6, or 246). 
	 * @see MidiMessage#getStatus
	 */
	public static final int TUNE_REQUEST				= 0xF6; // 246

	/**
	 * Status byte for End of System Exclusive message (0xF7, or 247). 
	 * @see MidiMessage#getStatus
	 */
	public static final int END_OF_EXCLUSIVE			= 0xF7; // 247


	// System real-time messages

	/**
	 * Status byte for Timing Clock messagem (0xF8, or 248). 
	 * @see MidiMessage#getStatus
	 */
	public static final int TIMING_CLOCK				= 0xF8; // 248

	/**
	 * Status byte for Start message (0xFA, or 250). 
	 * @see MidiMessage#getStatus
	 */
	public static final int START						= 0xFA; // 250

	/**
	 * Status byte for Continue message (0xFB, or 251). 
	 * @see MidiMessage#getStatus
	 */
	public static final int CONTINUE					= 0xFB; // 251

	/**
	 * Status byte for Stop message (0xFC, or 252). 
	 * @see MidiMessage#getStatus
	 */
	public static final int STOP						= 0xFC; //252

	/**
	 * Status byte for Active Sensing message (0xFE, or 254). 
	 * @see MidiMessage#getStatus
	 */
	public static final int ACTIVE_SENSING				= 0xFE; // 254

	/**
	 * Status byte for System Reset message (0xFF, or 255). 
	 * @see MidiMessage#getStatus
	 */
	public static final int SYSTEM_RESET				= 0xFF; // 255


// Channel voice message upper nibble defines

	/**
	 * Command value for Note Off message (0x80, or 128)
	 */
	public static final int NOTE_OFF					= 0x80;  // 128

	/**
	 * Command value for Note On message (0x90, or 144)
	 */
	public static final int NOTE_ON						= 0x90;  // 144

	/**
	 * Command value for Polyphonic Key Pressure (Aftertouch) message (0xA0, or 128)
	 */
	public static final int POLY_PRESSURE				= 0xA0;  // 160

	/**
	 * Command value for Control Change message (0xB0, or 176)
	 */
	public static final int CONTROL_CHANGE				= 0xB0;  // 176

	/**
	 * Command value for Program Change message (0xC0, or 192)
	 */
	public static final int PROGRAM_CHANGE				= 0xC0;  // 192

	/**
	 * Command value for Channel Pressure (Aftertouch) message (0xD0, or 208)
	 */
	public static final int CHANNEL_PRESSURE			= 0xD0;  // 208

	/**
	 * Command value for Pitch Bend message (0xE0, or 224)
	 */
	public static final int PITCH_BEND					= 0xE0;  // 224


// Default message data: NOTE_ON on Channel 0 with max volume

	private static byte[] defaultMessage				= { (byte)(NOTE_ON|0), (byte)255, (byte)0 }; 


// Instance variables    

	/**
	 * The first data byte in the MIDI message.
	 */
	//protected int data1 = 0;
    

	/**
	 * The second data byte in the MIDI message.
	 */
	//protected int data2 = 0;


	/**
	 * Constructs a new <code>ShortMessage</code>.  The
	 * contents of the new message are guaranteed to specify
	 * a valid MIDI message.  Subsequently, you may set the 
	 * contents of the message using one of the <code>setMessage</code> 
	 * methods.
	 * @see #setMessage
	 */
	public ShortMessage() {		
		//super(defaultMessage);
		this(defaultMessage);
	}


	/**
	 * Constructs a new <code>ShortMessage</code>. 
	 * @param data an array of bytes containing the complete message.
	 * The message data may be changed using the <code>setMessage</code>
	 * method.
	 * @see #setMessage
	 */
	protected ShortMessage(byte[] data) {
		super(data);
	}
	

	/**
	 * Sets the parameters for a MIDI message that takes no data bytes.
	 * @param status	the MIDI status byte 
	 * @throws  <code>InvalidMidiDataException</code> if <code>status</code> does not 
	 * specify a valid MIDI status byte for a message that requires no data bytes.
	 * @see #setMessage(int, int, int)
	 * @see #setMessage(int, int, int, int)
	 */
	public void setMessage(int status) throws InvalidMidiDataException {

		// check for valid values
		int dataLength = getDataLength(status); // can throw InvalidMidiDataException
		if (dataLength != 0) {
			throw new InvalidMidiDataException("Status byte; " + status + " requires " + dataLength + " data bytes");
		}
		
		try {

			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			DataOutputStream dos = new DataOutputStream(bos);

			dos.writeByte(status);

			this.data = bos.toByteArray();
			this.length = this.data.length;

		} catch (IOException e) {
		}


		/*
		// set the length
		length = dataLength + 1;
		
		// enlarge the byte array if necessary
		if (data.length < length) {
			data = new byte[length];
		}

		// set the data
		data[0] = (byte)status;
		*/
	}


	/**
	 * Sets the  parameters for a MIDI message that takes one or two data 
	 * bytes.  If the message takes only one data byte, the value passed in
	 * for the second data byte should be 0.
	 * 
	 * @param status	the MIDI status byte
	 * @param data1		the first data byte
	 * @param data2		the second data byte
	 * @throws		<code>InvalidMidiDataException</code> if the 
	 * the status or data bytes do not specify a valid MIDI message
	 * @see #setMessage(int, int, int, int)
	 * @see #setMessage(int)
	 */
	public void setMessage(int status, int data1, int data2) throws InvalidMidiDataException {

		// $$kk: 01.20.99: need to check for valid values

		// check for valid values
		int dataLength = getDataLength(status); // can throw InvalidMidiDataException

		try {

			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			DataOutputStream dos = new DataOutputStream(bos);

			dos.writeByte(status);

			if (dataLength >= 1) {
				dos.writeByte(data1);
			}

			if (dataLength >= 2) {
				dos.writeByte(data2);
			}

			this.data = bos.toByteArray();
			this.length = this.data.length;

		} catch (IOException e) {
		}


		/*
		// set the length
		length = dataLength + 1;

		// enlarge the byte array if necessary
		if (data.length < length) {
			data = new byte[length];
		}

		// set the data
		//data[0] = (byte)( (command & 0xF0) | (channel & 0x0F) );
		data[0] = (byte)status;
		data[1] = (dataLength >= 1) ? (byte)data1 : (byte)0;
		data[2] = (dataLength >= 2) ? (byte)data2 : (byte)0;
		*/
	}


	/**
	 * Sets the short message parameters for a  channel message
	 * which takes one or two data bytes.  If the message
	 * only takes one data byte, the second data byte value 
	 * passed should be 0.
	 *
	 * @param command	the MIDI command represented by this message
	 * @param channel	the channel associated with the message
	 * @param data1		the first data byte
	 * @param data2		the second data byte
	 * @throws		<code>InvalidMidiDataException</code> if the 
	 * parameter values do not specify a valid MIDI message
	 * @see #setMessage(int, int, int)
	 * @see #setMessage(int)
	 * @see #getCommand
	 * @see #getChannel
	 * @see #getData1
	 * @see #getData2
	 */
	public void setMessage(int command, int channel, int data1, int data2) throws InvalidMidiDataException {

		int status = (command & 0xF0) | (channel & 0x0F);
				
		// $$kk: 01.20.99: need to check for valid values

		// check for valid values
		int dataLength = getDataLength(status); // can throw InvalidMidiDataException

		try {

			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			DataOutputStream dos = new DataOutputStream(bos);
			dos.writeByte(status);

			if (dataLength >= 1) {
				dos.writeByte(data1);
			}

			if (dataLength >= 2) {
				dos.writeByte(data2);
			}

			this.data = bos.toByteArray();
			this.length = this.data.length;

		} catch (IOException e) {
		}

		
		/*
		// check for valid values
		int dataLength = getDataLength(status); // can throw InvalidMidiDataException

		// set the length
		length = dataLength + 1;

		// enlarge the byte array if necessary
		if (data.length < length) {
			data = new byte[length];
		}

		// set the data
		//data[0] = (byte)( (command & 0xF0) | (channel & 0x0F) );
		data[0] = (byte)status;
		data[1] = (dataLength >= 1) ? (byte)data1 : (byte)0;
		data[2] = (dataLength >= 2) ? (byte)data2 : (byte)0;
		*/
	}


	/**
	 * Obtains the MIDI channel associated with this event.  This method
	 * assumes that the event is a MIDI channel message; if not, the return
	 * value will not be meaningful.
	 * @return MIDI channel associated with the message.
	 * @see #setMessage(int, int, int, int)
	 */
	public int getChannel() {

		// $$kk: 01.20.99: need to check for valid values?

		int status = getStatus();
		return (status & 0x0F);
	}


	/**
	 * Obtains the MIDI command associated with this event.  This method
	 * assumes that the event is a MIDI channel message; if not, the return
	 * value will not be meaningful.
	 * @see #setMessage(int, int, int, int)
	 */
	public int getCommand() {

		// $$kk: 01.20.99: need to check for valid values?

		int status = getStatus();
		return (status & 0xF0);
	}


	/**
	 * Obtains the first data byte in the message.
	 * @return the value of the <code>data1</code> field
	 * @see #setMessage(int, int, int)
	 */
	public int getData1() {

		int data1 = 0;

		try {

			// check for valid values
			int status = getStatus();
			int dataLength = getDataLength(status); // can throw InvalidMidiDataException
			if (dataLength < 1) {
				//throw new InvalidMidiDataException("Status byte; " + status + " has " + dataLength + " data bytes");
			}
		
			ByteArrayInputStream bas = new ByteArrayInputStream(data, 0, length);
			DataInputStream das = new DataInputStream(bas);
			das.skip(1);
			data1 = das.readUnsignedByte();

		} catch (InvalidMidiDataException iedm) {
		} catch (IOException ioe) {			
		}

		return data1;
	}


	/**
	 * Obtains the second data byte in the message.
	 * @return the value of the <code>data2</code> field
	 * @see #setMessage(int, int, int)
	 */
	public int getData2() {

		int data2 = 0;

		try {

			// check for valid values
			int status = getStatus();
			int dataLength = getDataLength(status); // can throw InvalidMidiDataException
			if (dataLength < 2) {
				//throw new InvalidMidiDataException("Status byte; " + status + " has " + dataLength + " data bytes");
			}
		
			ByteArrayInputStream bas = new ByteArrayInputStream(data, 0, length);
			DataInputStream das = new DataInputStream(bas);
			das.skip(2);
			data2 = das.readUnsignedByte();

		} catch (InvalidMidiDataException iedm) {
		} catch (IOException ioe) {			
		}

		return data2;
	}

	
	/**
	 * Creates a new object of the same class and with the same contents
	 * as this object.
	 * @return a clone of this instance.
	 */
	public Object clone() {

		byte[] newData = new byte[length];
		System.arraycopy(data, 0, newData, 0, newData.length);

		ShortMessage event = new ShortMessage(newData);
		return event;
	}

	
	/**
	 * Retrieves the type of a short message.
	 * @param statusVal status byte value
	 * @return the message type: <code>CHANNEL_MODE_MESSAGE</code>, 
	 * <code>CHANNEL_VOICE_MESSAGE</code>, <code>SYSTEM_COMMON_MESSAGE</code>,
	 * <code>SYSTEM_REALTIME_MESSAGE</code>, or <code>SYSTEM_EXCLUSIVE_MESSAGE</code>
	 * @throws <code>InvalidMidiDataException</code> if  
	 * <code>statusVal</code> does not represent the status byte for any 
	 * short message
	 */
/*
	protected int getType(int statusVal, int data1) throws InvalidMidiDataException {

		// channel voice or mode message?  

		if ( (statusVal >= 0x80) && (statusVal <= 0xEF) ) {

			if ( (data1 >= 121) && (data1 <= 127) ) {
				return CHANNEL_MODE_MESSAGE;
			}

			return CHANNEL_VOICE_MESSAGE;
		}


		// system common message?
		
		if ( (statusVal >= 0xF1) && (statusVal <= 0xF7) )
			return SYSTEM_COMMON_MESSAGE;

		// system real-time message?
		
		if ( (statusVal >= 0xF8) && (statusVal <= 0xFF) )
			return SYSTEM_REALTIME_MESSAGE;

		// system exclusive?
		
		if ( statusVal == 0xF0 )
			return SYSTEM_EXCLUSIVE_MESSAGE;

		throw new InvalidMidiDataException("Invalid status value: " + statusVal);
	}
*/


	/**
	 * Retrieves the number of data bytes associated with a particular 
	 * status byte value.  
	 * @param status status byte value, which must represent a short MIDI message
	 * @return data length in bytes (0, 1, or 2)
	 * @throws <code>InvalidMidiDataException</code> if the  
	 * <code>status</code> argument does not represent the status byte for any 
	 * short message
	 */
	protected final int getDataLength(int status) throws InvalidMidiDataException {

		// system common and system real-time messages

		switch(status) {

		case 0xF6:			// Tune Request		
		case 0xF7:			// EOX
		// System real-time messages
		case 0xF8:			// Timing Clock		
		case 0xF9:			// Undefined		
		case 0xFA:			// Start
		case 0xFB:			// Continue		
		case 0xFC:			// Stop
		case 0xFD:			// Undefined
		case 0xFE:			// Active Sensing
		case 0xFF:			// System Reset
			return 0;
		case 0xF1:			// MTC Quarter Frame
		case 0xF3:			// Song Select
			return 1;
		case 0xF2:			// Song Position Pointer
			return 2;
		default:
		}

		// channel voice and mode messages
		
		switch(status & 0xF0) {

		case 0x80:
		case 0x90:
		case 0xA0:
		case 0xB0:
		case 0xE0:
			return 2;
		case 0xC0:
		case 0xD0:
			return 1;
		default:
			throw new InvalidMidiDataException("Invalid status byte: " + status);				 
		}
	}
}
