/*
 * @(#)AudioDevice.java	1.12 00/02/02
 *
 * Copyright 1995-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.audio;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;
import java.io.InputStream;
import java.io.BufferedInputStream;
import java.io.OutputStream;

import javax.sound.sampled.*;
import javax.sound.midi.*;

/**
 * This class provides an interface to the Headspace Audio engine through
 * the Java Sound API.
 *
 * This class emulates systems with multiple audio channels, mixing
 * multiple streams for the workstation's single-channel device.
 *
 * @see AudioData
 * @see AudioDataStream
 * @see AudioStream
 * @see AudioStreamSequence
 * @see ContinuousAudioDataStream
 * @author David Rivas
 * @author Kara Kytle
 * @author Jan Borgersen
 * @version 1.12 00/02/02
 */

public class
AudioDevice {

    private boolean DEBUG = false  /*true*/ ;

    /** Hashtable of audio clips / input streams. */
    private Hashtable clipStreams;

	private Vector infos;

    /** Are we currently playing audio? */
    private boolean playing = false;

    /** Handle to the JS audio mixer. */
    private Mixer mixer = null;



    /**
     * The default audio player. This audio player is initialized
     * automatically.
     */
    public static final AudioDevice device = new AudioDevice();

    /**
     * Create an AudioDevice instance.
     */
	private AudioDevice() {

		clipStreams = new Hashtable();
		infos = new Vector();
	}


	private synchronized void startSampled( AudioInputStream as, 
		InputStream in ) throws UnsupportedAudioFileException, 
		LineUnavailableException {

		Info info = null;
		DataPusher datapusher = null;
		DataLine.Info lineinfo = null;
		SourceDataLine sourcedataline = null;

		// if ALAW or ULAW, we must convert....

		if( (as.getFormat().getEncoding() == AudioFormat.Encoding.ULAW) ||
			(as.getFormat().getEncoding() == AudioFormat.Encoding.ALAW)) {

			AudioFormat af = as.getFormat();
			AudioInputStream asold = as;
			as = AudioSystem.getAudioInputStream( 
						new AudioFormat( AudioFormat.Encoding.PCM_SIGNED,
											af.getSampleRate(),
											af.getSampleSizeInBits() * 2,
											af.getChannels(),
											af.getFrameSize() * 2,
											af.getFrameRate(),
											true )
									, asold );
			if( as==null ) {
				// could not convert
				return;
			}
		}

//		lineinfo = new DataLine.Info(SourceDataLine.class, null, null, 
//							new Class[0], as.getFormat(),
//							(int)as.getLength() );
		lineinfo = new DataLine.Info(SourceDataLine.class, 
						as.getFormat(),
						(int)as.getFrameLength() );
		if( !(AudioSystem.isLineSupported(lineinfo))) {

			return;
		}
		sourcedataline = (SourceDataLine)AudioSystem.getLine(lineinfo);
//		sourcedataline.open( as.getFormat(), sourcedataline.getBufferSize()  );
		sourcedataline.open(as.getFormat(), (int)as.getFrameLength() );

		sourcedataline.start();

		datapusher = new DataPusher( sourcedataline, as);

		info = new Info( null, in, datapusher );
		infos.addElement( info );

		datapusher.start();

	}

	private synchronized void startMidi( InputStream bis, 
		InputStream in ) throws InvalidMidiDataException, 
		MidiUnavailableException  {

		Sequencer sequencer = null;
		Info info = null;
		
		//sequencer = MidiSystem.getSequencer( null );
		sequencer = MidiSystem.getSequencer( );
		//sequencer.addMetaEventListener( (MetaEventListener) this );
		sequencer.open();
		try {
			sequencer.setSequence( bis );
		} catch( IOException e ) {
			throw new InvalidMidiDataException( e.getMessage() );
		}

		info = new Info( sequencer, in, null );

		infos.addElement( info );

		sequencer.start();

	}



    /**
     *  Open an audio channel.
     */
    public synchronized void openChannel(InputStream in) {


		if(DEBUG) {
			System.out.println("AudioDevice: openChannel");
			System.out.println("input stream =" + in); 
		}

		Info info = null;

		// is this already playing?  if so, then just return
		for(int i=0; i<infos.size(); i++) {
			info = (AudioDevice.Info)infos.elementAt(i);
			if( info.in == in ) {

				return;
			}
		}


		AudioInputStream as = null;

		if( in instanceof AudioStream ) {

			if ( ((AudioStream)in).midiformat != null ) {

				// it's a midi file
				try {
					startMidi( ((AudioStream)in).stream, in );
				} catch (Exception e) { 
					return;
				}


			} else if( ((AudioStream)in).ais != null ) {

				// it's sampled audio
				try {
					startSampled( ((AudioStream)in).ais, in );
				} catch (Exception e) { 
					return;
				}

			}
		} else {
			BufferedInputStream bis = new BufferedInputStream( in, 1024 );

			try {

				try {
					as = AudioSystem.getAudioInputStream(bis);
				} catch(IOException ioe) {
					return;
				}

				startSampled( as, in );

			} catch( UnsupportedAudioFileException e ) {

				try {
					try {
						MidiFileFormat mff = 
								MidiSystem.getMidiFileFormat( bis );
					} catch(IOException ioe1) {
						return;
					}

					startMidi( bis, in );


				} catch( InvalidMidiDataException e1 ) {

					// $$jb:08.01.99: adding this section to make some of our other
					// legacy classes work.....
				    // not MIDI either, special case handling for AudioStreamSequence
					if( in instanceof AudioStreamSequence ) {

						AudioFormat defformat = new AudioFormat( AudioFormat.Encoding.ALAW,
											8000, 8, 1, 1, 8000, true );
						try {
							AudioInputStream defaif = new AudioInputStream( in,
								defformat, AudioSystem.NOT_SPECIFIED);
							startSampled( defaif, in );
						} catch (UnsupportedAudioFileException es) {
							return;
						} catch (LineUnavailableException es2) {
							return;
						}
					} else {


					// could not create stream or sequence
					return;
					}

				} catch( MidiUnavailableException e2 ) {

					// could not open sequence
					return;
				}

			} catch( LineUnavailableException e ) {

				return;
			}
		}
	
		// don't forget adjust for a new stream.
		notify();
    }


    /**
     *  Close an audio channel.
     */
    public synchronized void closeChannel(InputStream in) {


		if(DEBUG) {
			System.out.println("AudioDevice.closeChannel");
		}

		if (in == null) return;		// can't go anywhere here!

		Info info;

		for(int i=0; i<infos.size(); i++) {

			info = (AudioDevice.Info)infos.elementAt(i);

			if( info.in == in ) {

				if( info.sequencer != null ) {
	
					info.sequencer.stop();
					//info.sequencer.close();
					infos.removeElement( info );

				} else if( info.datapusher != null ) {

					info.datapusher.stop();
					infos.removeElement( info );
				}
			}
		}
		notify();
    }


    /**
     * Open the device (done automatically)
     */
    public synchronized void open() {

		// $$jb: 06.24.99: This is done on a per-stream
		// basis using the new JS API now. 
	}


    /**
     * Close the device (done automatically)
     */
	public synchronized void close() {

		// $$jb: 06.24.99: This is done on a per-stream
		// basis using the new JS API now.

	}


    /**
     * Play open audio stream(s)
     */
    public void play() {

		// $$jb: 06.24.99:  Holdover from old architechture ...
		// we now open/close the devices as needed on a per-stream
		// basis using the JavaSound API.

		if (DEBUG) {
			System.out.println("exiting play()");
		}
	}

    /**
     * Close streams
     */
    public synchronized void closeStreams() {

		Info info;

		for(int i=0; i<infos.size(); i++) {

			info = (AudioDevice.Info)infos.elementAt(i);

			if( info.sequencer != null ) {
	
				info.sequencer.stop();
				info.sequencer.close();
				infos.removeElement( info );

			} else if( info.datapusher != null ) {

				info.datapusher.stop();
				infos.removeElement( info );
			}
		}


		if (DEBUG) {
		    System.err.println("Audio Device: Streams all closed.");
		}
		// Empty the hash table.
		clipStreams = new Hashtable();
		infos = new Vector();
	}

    /**
     * Number of channels currently open.
     */
    public int openChannels() {
		return infos.size();
    }

    /**
     * Make the debug info print out.
     */
    void setVerbose(boolean v) {
		DEBUG = v;
    }






	// INFO CLASS

	class Info {

		Sequencer   sequencer;
		InputStream in;
		DataPusher  datapusher;

		Info( Sequencer sequencer, InputStream in, DataPusher datapusher ) {

			this.sequencer  = sequencer;
			this.in         = in;
			this.datapusher = datapusher;
		}
	}




	// DATA PUSHER CLASS

	class DataPusher implements Runnable {


		SourceDataLine target;
		AudioInputStream is; 


		Thread pushThread = null;
		boolean stopping;
		boolean stopped;
			
		DataPusher(SourceDataLine target, AudioInputStream is) {

			this.target = target;
			this.is = is;
		}


		public void start() {

			if (pushThread != null) {
					
				if(DEBUG) {
					System.out.println("Thread is not null");
				}
				return;
			}

			if(DEBUG) {
				System.out.println("Starting push");
			}

			stopping = false;
			stopped = false;

			pushThread = new Thread(this);
			pushThread.start();
		}


		public void stop() {

			if (pushThread == null) {

				return;
			}

			if(DEBUG) {
				System.out.println("Stopping push");
			}
			
			stopping = true;
			while (!stopped) {
				try {
					Thread.sleep(10);
				} catch (InterruptedException e) {  }
			}
			pushThread = null;
		}



		/**
		 * Write data to the source data line.
		 */
		public void run() {

			byte[] dataArray = new byte[target.getBufferSize()/2];

			int frameSize = target.getFormat().getFrameSize();
			int bytesRead;
			while(stopping == false) {

				try {
						
					bytesRead = is.read(dataArray);

					if (bytesRead == -1) {

						if(DEBUG) {
							System.out.println("\n\nFOUND END OF STREAM\n\n");
						}
						target.drain();
						target.stop();
						break;
					}
					if(DEBUG) {
						System.out.println("> WRITING " + bytesRead + 
											" BYTES\n");
					}
					int bytesRemaining = bytesRead;

					while (bytesRemaining > 0) {
						bytesRemaining -= target.write(dataArray, 0, bytesRead);
					}

					if(DEBUG) {
						System.out.println("< WROTE " + bytesRead + 
							" BYTES\n");
					}

				} catch (IOException ioe) {

					break;
				}
			}

			target.flush();
			target.stop();
			stopped = true;

		}

	} // class DataPusher

}


