/*
 * @(#)DataTransferer.java	1.4 01/02/09
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.FlavorMap;
import java.awt.datatransfer.Transferable;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * Provides a set of functions to be shared among the DataFlavor class and
 * platform-specific data transfer implementations.
 *
 * @author David Mendenhall
 * @version 1.4, 02/09/01
 *
 * @since 1.3.1
 */
public abstract class DataTransferer {

    /**
     * Encodings that we know are or are not supported go in this Map. It's 
     * much faster than actually encoding some text. 
     */ 
    private static final Map knownEncodings;
    static {
        Map tempMap = new HashMap(11);
        tempMap.put("ASCII", Boolean.TRUE); // US-ASCII
        tempMap.put("ISO8859_1", Boolean.TRUE); // ISO-8859-1
        tempMap.put("UTF8", Boolean.TRUE); // UTF-8
        tempMap.put("UnicodeBigUnmarked", Boolean.TRUE); // UTF-16BE
        tempMap.put("UnicodeLittleUnmarked", Boolean.TRUE); // UTF-16LE
        tempMap.put("UTF16", Boolean.TRUE); // UTF-16
        knownEncodings = Collections.synchronizedMap(tempMap);
    }

    private static String defaultEncoding;

    /**
     * If a DataFlavor's primary MIME type is "text", then this method
     * returns the "charset" parameter, or the default charset if no
     * parameter is specified. For non-text DataFlavors ( ** including
     * DataFlavor.stringFlavor, whose MIME type is actually
     * application/x-java-serialized-object ** ), this method returns
     * null.
     */
    public static String getTextCharset(DataFlavor flavor) {
        if (!"text".equals(flavor.getPrimaryType())) {
            return null;
        }

        String charset = flavor.getParameter("charset");
        if (charset != null) {
            return charset;
        }

        return getDefaultTextCharset();
    }

    /**
     * Returns the platform's default character encoding. This code is
     * based on code available in sun.io.Converters.getDefaultConverterClass.
     */
    public static String getDefaultTextCharset() {
        if (defaultEncoding != null) {
            return defaultEncoding;
        }

        java.security.PrivilegedAction a =
            new sun.security.action.GetPropertyAction("file.encoding");
        String enc = (String)java.security.AccessController.doPrivileged(a);
        if (enc != null) {
            defaultEncoding = (isEncodingSupported(enc)) ? enc : "ISO8859_1";
            return defaultEncoding;
        } else {
            /* Property not yet set, so do not set default encoding and
               just use ISO8859_1 for now */
            return "ISO8859_1";
        }
    }

    /**
     * Determines whether the JDK supports a given encoding.
     */
    public static boolean isEncodingSupported(String encoding) {
        if (encoding == null) { 
            return false; 
        } 
        String canonicalName = sun.io.CharacterEncoding.aliasName(encoding); 
        if (canonicalName == null) { 
            canonicalName = encoding;
        } 
 
        Boolean bool = (Boolean)knownEncodings.get(canonicalName);

        if (bool == null) {
            // keep looking - try to encode something and see if io complains 
            try { 
                new String("abc".getBytes(canonicalName), canonicalName);
                bool = Boolean.TRUE;
            } catch (UnsupportedEncodingException encodingException) {
                bool = Boolean.FALSE;
            }

            knownEncodings.put(canonicalName, bool);
        }

        return (bool == Boolean.TRUE);
    }

    /**
     * Returns a Map whose keys are all of the possible formats into which the
     * Transferable's transfer data flavors can be translated. The value of
     * each key is the DataFlavor in which the Transferable's data should be
     * requested when converting to the format.
     */
    public Map getFormatsForTransferable(Transferable contents,
						FlavorMap map) {
        return getFormatsForFlavors(contents.getTransferDataFlavors(), map);
    }

    /**
     * Returns a Map whose keys are all of the possible formats into which data
     * in the specified DataFlavor can be translated. The value of each key
     * is the DataFlavor in which a Transferable's data should be requested
     * when converting to the format.
     */
    public Map getFormatsForFlavor(DataFlavor flavor,
					  FlavorMap map) {
        return getFormatsForFlavors(new DataFlavor[] { flavor },
				    map);
    }

    /**
     * Returns a Map whose keys are all of the possible formats into which data
     * in the specified DataFlavors can be translated. The value of each key
     * is the DataFlavor in which the Transferable's data should be requested
     * when converting to the format.
     */
    public abstract Map getFormatsForFlavors(DataFlavor[] flavors,
					     FlavorMap map);

    /**
     * Reduces the Map output for the root function to an array of the
     * Map's keys.
     */
    public int[] getFormatsForTransferableAsArray(Transferable contents,
							 FlavorMap map) {
        return keysToIntArray(getFormatsForTransferable(contents, map));
    }

    public int[] getFormatsForFlavorAsArray(DataFlavor flavor,
						   FlavorMap map) {
        return keysToIntArray(getFormatsForFlavor(flavor, map));
    }

    public int[] getFormatsForFlavorsAsArray(DataFlavor[] flavors,
						    FlavorMap map) {
        return keysToIntArray(getFormatsForFlavors(flavors, map));
    }

    /**
     * Returns a Map whose keys are all of the possible DataFlavors into which
     * data in the specified formats can be translated. The value of each key
     * is the format in which the Clipboard or dropped data should be requested
     * when converting to the DataFlavor.
     */
    public abstract Map getFlavorsForFormats(int[] formats, FlavorMap map);

    public Map getFlavorsForFormat(int format, FlavorMap map) {
        return getFlavorsForFormats(new int[] { format }, map);
    }

    /**
     * Reduces the Map output for the root function to an array of the
     * Map's keys.
     */
    public DataFlavor[] getFlavorsForFormatsAsArray(int[] formats,
							   FlavorMap map) {
        return keysToDataFlavorArray(getFlavorsForFormats(formats, map));
    }

    public DataFlavor[] getFlavorsForFormatAsArray(int format,
							  FlavorMap map) {
      return keysToDataFlavorArray(getFlavorsForFormat(format, map));
    }

    public int[] keysToIntArray(Map map) {
        Set keySet = map.keySet();
	int[] retval = new int[keySet.size()];
	int i = 0;
	for (Iterator iter = keySet.iterator(); iter.hasNext(); i++) {
	    retval[i] = ((Integer)iter.next()).intValue();
	}
	return retval;
    }

    /**
     * Returns an array of DataFlavors mapped to some native formats 
     * in the specified flavor-to-native map. 
     * The array is sorted with <code>DataFlavorComparator</code>.
     * @param map mapping between DataFlavor and native formats.
     * @return an array of DataFlavors mapped to some native formats with the
     *         specified map.
     */
    public DataFlavor[] keysToDataFlavorArray(Map map) {
        return setToSortedDataFlavorArray(map.keySet(), map);
    }

    /**
     * Attempts to discern the encoding format given the specified
     * information.
     */
    protected abstract String getEncoding(DataFlavor flavor, int format,
					  String sourceEncoding);

    /**
     * Primary translation function for translating a Transferable into
     * a byte array, given a source DataFlavor and target format.
     */
    public byte[] translateTransferable(Transferable contents,
                                               DataFlavor flavor,
                                               int format) throws IOException
    {
        return translateTransferable(contents, flavor, format, null);
    }

    protected abstract byte[] translateTransferable(Transferable contents,
					   DataFlavor flavor,
					   int format,
					   String sourceEncoding)
        throws IOException ;

    public Object translateBytes(byte[] bytes, DataFlavor flavor,
					int format) throws IOException {
        return translateBytesOrStream(null, bytes, flavor, format);
    }

    public Object translateStream(InputStream str, DataFlavor flavor,
					 int format) throws IOException {
        return translateBytesOrStream(str, null, flavor, format);
    }

    /**
     * Primary translation function for translating either a byte array or
     * an InputStream into an Object, given a source format and a target
     * DataFlavor.
     *
     * One of str/bytes is non-null; the other is null.
     * The conversion from byte[] to InputStream is cheap, so do that
     * immediately if necessary. The opposite conversion is expensive,
     * so avoid it if possible.
     */
    protected abstract Object translateBytesOrStream(InputStream str,
						     byte[] bytes,
						     DataFlavor flavor,
						     int format)
        throws IOException;

    /**
     * Convert an InputStream to a ByteArray
     */
    protected byte[] inputStreamToByteArray(InputStream str)
        throws IOException
    {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
	int len = 0;
        byte[] buf = new byte[64];

	while ((len = str.read(buf)) != -1) {
	    baos.write(buf, 0, len);
	}
        
	return baos.toByteArray();
    }

    private static DataFlavorComparator defaultComparator = 
        new DataFlavorComparator(DataFlavorComparator.REVERSE_ORDER);

    /**
     * Returns an array containing all of the DataFlavors in this set.
     * The array is sorted with <code>DataFlavorComparator</code>.
     * @param flavorsSet set of data flavors.
     * @return a sorted array containing all of the DataFlavors in this set.
     */
    public DataFlavor[] setToSortedDataFlavorArray(Set flavorsSet) {
        DataFlavor[] flavors = new DataFlavor[flavorsSet.size()];
        flavorsSet.toArray(flavors);
        Arrays.sort(flavors, defaultComparator);
        return flavors;
    }

    /**
     * Returns an array containing all of the DataFlavors in this set.
     * The array is sorted with <code>DataFlavorComparator</code> 
     * created with the specified flavor-to-native map as an argument.
     * <p>
     * Flavor-to-native map is used for comparison of DataFlavors which
     * otherwise are considered equal. The one mapped to the native 
     * format with greater atom value is considered better.
     *
     * @param flavorsSet set of data flavors.
     * @param flavorToNativeMap map which maps some of these flavors 
     *        to native formats.
     * @return a sorted array containing all of the DataFlavors in this set.
     */
    public DataFlavor[] setToSortedDataFlavorArray(Set flavorsSet,
                                                   Map flavorToNativeMap) {
        DataFlavor[] flavors = new DataFlavor[flavorsSet.size()];
        flavorsSet.toArray(flavors);
        Comparator comparator = 
            new DataFlavorComparator(flavorToNativeMap,
                                     DataFlavorComparator.REVERSE_ORDER);
        Arrays.sort(flavors, comparator);
        return flavors;
    }

    public static class DataFlavorComparator implements Comparator {

        public static final boolean FORWARD_ORDER = true;
        public static final boolean REVERSE_ORDER = false;

        protected final Map flavorToFormatMap;
        private final boolean order;
        
        private static final Map primaryTypes;
        private static final Map subtypes;
        private static final Map charsets;
        private static String defaultEncoding;

        private static final Integer UNKNOWN_TYPE =
	    new Integer(Integer.MAX_VALUE);

        private static final Integer DEFAULT_CHARSET_INDEX = new Integer(2);
        private static final Integer OTHER_CHARSET_INDEX = new Integer(1);
        private static final Integer WORST_CHARSET_INDEX = new Integer(0);
        private static final Integer UNSUPPORTED_CHARSET_INDEX =
	    new Integer(Integer.MIN_VALUE);

        private static final String UNSUPPORTED_CHARSET = "UNSUPPORTED";
        
        private static final Integer REMOTE_CLASS_INDEX = new Integer(3);
        private static final Integer SERIALIZABLE_CLASS_INDEX = new Integer(2);
        private static final Integer INPUT_STREAM_CLASS_INDEX = new Integer(1);
        private static final Integer OTHER_CLASS_INDEX = new Integer(0);
        
        static {
            {
                HashMap primaryTypesMap = new HashMap(2);
                
                primaryTypesMap.put("text", new Integer(0));
                primaryTypesMap.put("application", new Integer(1));
                
                primaryTypes = Collections.unmodifiableMap(primaryTypesMap);
            }

            {
                HashMap subtypesMap = new HashMap(17);
                
                // plain text
                subtypesMap.put("plain", new Integer(0));
                
                // misc text subtypes
                subtypesMap.put("calendar", new Integer(1));
                subtypesMap.put("css", new Integer(2));
                subtypesMap.put("directory", new Integer(3));
                subtypesMap.put("rfc822-headers", new Integer(4));
                subtypesMap.put("tab-separated-values", new Integer(5));
                subtypesMap.put("uri-list", new Integer(6));
                
                // enriched subtypes
                subtypesMap.put("richtext", new Integer(7));
                subtypesMap.put("enriched", new Integer(8));
                subtypesMap.put("rtf", new Integer(9));
                
                // markup subtypes
                subtypesMap.put("html", new Integer(10));
                subtypesMap.put("xml", new Integer(11));
                subtypesMap.put("sgml", new Integer(12));
                
                // x-java subtypes
                subtypesMap.put("x-java-file-list", new Integer(13));
                subtypesMap.put("x-java-serialized-object", new Integer(14));
                subtypesMap.put("x-java-jvm-local-objectref", new Integer(15));
                subtypesMap.put("x-java-remote-object", new Integer(16));
                
                subtypes = Collections.unmodifiableMap(subtypesMap);
            }

            {
                HashMap charsetsMap = new HashMap(6);
                
                // we prefer Unicode charsets
                charsetsMap.put("UnicodeLittleUnmarked", new Integer(3));
                charsetsMap.put("UnicodeBigUnmarked", new Integer(4));
                charsetsMap.put("UTF8", new Integer(5));
                charsetsMap.put("UTF16", new Integer(6));
                
                // ASCII is the worst charset supported
                charsetsMap.put("ASCII", WORST_CHARSET_INDEX);

                String defEncoding = DataTransferer.getDefaultTextCharset();
                defaultEncoding =
		    sun.io.CharacterEncoding.aliasName(defEncoding);
                if (defaultEncoding == null) {
                    defaultEncoding = defEncoding;
                }

                if (charsetsMap.get(defaultEncoding) == null) {
                    charsetsMap.put(defaultEncoding, DEFAULT_CHARSET_INDEX);
                }
                charsetsMap.put(UNSUPPORTED_CHARSET,
				UNSUPPORTED_CHARSET_INDEX);
                
                charsets = Collections.unmodifiableMap(charsetsMap);
            }
        }
        
        DataFlavorComparator() {
            this(FORWARD_ORDER); 
        }
        
        DataFlavorComparator(boolean order) {
            this.order = order;

            flavorToFormatMap = Collections.EMPTY_MAP;
        }

        DataFlavorComparator(Map map) {
            this(map, FORWARD_ORDER);
        }
        
        DataFlavorComparator(Map map, boolean order) {
            this.order = order;

            HashMap hashMap = new HashMap(map.size());
            hashMap.putAll(map);
            flavorToFormatMap = Collections.unmodifiableMap(hashMap);
        }

        /**
         * Compares two DataFlavor objects.  Returns a negative integer, zero, 
         * or a positive integer as the first charset is worse than, equal to,
         * or better than the second.
         * <p>
         * DataFlavors are ordered according to the following rules: 
         * <ul>
         * <li>The data flavor with better primary MIME type is considered
	 *     better.
         * <li>If data flavors have equal primary MIME types the one with
	 *     better secondary MIME type is considered better.
         * <li>If data flavors have equal secondary MIME types and their
	 *     primary MIME type is "text" the one with better charset is
	 *     considered better.
         * <li>If data flavors have equal secondary MIME types and their
	 *     primary MIME type is not "text" or their primary MIME type is
	 *     "text" and their charsets are equal the one with better
	 *     representation class is considered better.
         * <li>If data flavors have equal representation classes the one mapped
         *     to the native format with greater atom value is considered
	 *     better.
         * </ul>
         *
         * @param obj1 the first data flavor to be compared.
         * @param obj2 the second data flavor to be compared.
         * @return a negative integer, zero, or a positive integer as the
         * 	       first argument is worse, equal to, or better than the
         *	       second. 
         * @throws ClassCastException if either of the arguments is not
	 *         instance of DataFlavor.
         * @throws NullPointerException if either of the arguments is
	 *         <code>null</code>.
         */
        public int compare(Object obj1, Object obj2) {
            int comp = 0;
            DataFlavor flavor1 = null;
            DataFlavor flavor2 = null;
            if (order == FORWARD_ORDER) {
                flavor1 = (DataFlavor)obj1;
                flavor2 = (DataFlavor)obj2;
            } else {
                flavor1 = (DataFlavor)obj2;
                flavor2 = (DataFlavor)obj1;
            }
            
            comp = comparePrimaryTypes(flavor1.getPrimaryType(),
                                       flavor2.getPrimaryType());
            if (comp != 0) {
                return comp;
            }
            
            comp = compareSubtypes(flavor1.getSubType(),
                                   flavor2.getSubType());
            if (comp != 0) {
                return comp;
            }
            
            if ("text".equals(flavor1.getPrimaryType())) { 
                comp = compareCharsets(DataTransferer.getTextCharset(flavor1),
                                       DataTransferer.getTextCharset(flavor2));
                if (comp != 0) {
                    return comp;
                }
            }

            comp =
		compareRepresentationClasses(flavor1.getRepresentationClass(),
					     flavor2.getRepresentationClass());
            if (comp != 0) {
                return comp;
            }
            
            return compareByNativeFormats(flavor1, flavor2);
        }
        
        /**
         * Compares two primary MIME types.  Returns a negative integer, zero, 
         * or a positive integer as the first MIME type is worse than, equal
	 * to, or better than the second.
	 * <p>
         * Primary MIME types are sorted in the following order: "text" is
	 * worse than "application", and "application" is worse than any other
	 * primary MIME type. All primary MIME types except for "text" and
	 * "application" are considered equal.
         *
         * @param subtype1 the first MIME subtype to be compared.
         * @param subtype1 the second MIME subtype to be compared.
         * @return a negative integer, zero, or a positive integer as the
         * 	       first argument is worse, equal to, or better than the
         *	       second. 
         */
        protected int comparePrimaryTypes(String primaryType1,
					  String primaryType2) {
            return compareIndices(primaryTypes, primaryType1, primaryType2,
				  UNKNOWN_TYPE);
        }
        
        /**
         * Compares two secondary MIME types.  Returns a negative integer,
	 * zero, or a positive integer as the first argument is worse than,
	 * equal to, or better than the second.
         * <p>
         * Secondary MIME types are sorted in the following order: "plain", 
         * "calendar", "css", "directory", "rfc822-header",
	 * "tab-separated-values", "uri-list", "richtext", "enriched", "rtf",
	 * "html", "xml", "sgml". For example, "text/sgml" will be selected
	 * over "text/html", and "text/enriched" will be selected over
	 * "text/richtext". 
         * <p>
         * All other secondary MIME types are considered equal and are better
	 * than any of the listed above.
         *
         * @param subtype1 the first MIME subtype to be compared.
         * @param subtype1 the second MIME subtype to be compared.
         * @return a negative integer, zero, or a positive integer as the
         * 	       first argument is worse, equal to, or better than the
         *	       second. 
         */
        protected int compareSubtypes(String subtype1, String subtype2) {
            return compareIndices(subtypes, subtype1, subtype2, UNKNOWN_TYPE);
        }
        
        /**
         * Compares charsets.  Returns a negative integer, zero, 
         * or a positive integer as the first charset is worse than, equal to,
         * or better than the second.
         * <p>
         * Charsets are ordered according to the following rules: 
         * <ul>
         * <li>If either of the two charsets is an alias name, the
	 *     corresponding canonical name will be compared instead.
         * <li>All unsupported charsets are equal.
         * <li>Any unsupported charset is worse than any supported charsets.
         * <li>Unicode charsets, such as "UTF-16", "UTF-8", "UTF-16BE" and
	 *     "UTF-16LE", are considered best.
         * <li>After them, platform default charset is selected.
         * <li>"US-ASCII" is the worst of supported charsets.
         * <li>For all other supported charsets the lexicographically less one
	 *     is considered the better.
         * </ul>
         *
         * @param charset1 the first MIME subtype to be compared.
         * @param charset2 the second MIME subtype to be compared.
         * @return a negative integer, zero, or a positive integer as the
         * 	       first argument is worse, equal to, or better than the
         *	       second. 
         */
        protected int compareCharsets(String charset1, String charset2) {
            charset1 = getEncoding(charset1);
            charset2 = getEncoding(charset2);
            
            int comp = compareIndices(charsets, charset1, charset2,
                                      OTHER_CHARSET_INDEX);
            
            if (comp == 0) {
                return charset2.compareTo(charset1);
            }
            
            return comp;
        }
        
        /**
         * Returns encoding for the specified charset according to the
	 * following rules:
         * <ul>
         * <li>If the charset is <code>null</code> the <code>null</code> will
	 *     be returned.
         * <li>Iff the charset specifies encoding unsupported by this
	 *     implementation of the Java platform,
	 *     <code>UNSUPPORTED_CHARSET</code> will be returned.
         * <li>If the charset specifies an alias name, the corresponding
	 *     canonical name will be returned.
         * </ul>
         *
         * @param charset the charset.
         * @return an encoding for this charset. 
         */
        protected static String getEncoding(String charset) {
            String encoding = null;
            
            if (charset == null) {
                return null;
            } else if (DataTransferer.isEncodingSupported(charset)) {
                String canonicalName =
		    sun.io.CharacterEncoding.aliasName(charset);
                if (canonicalName != null) {
                    encoding = canonicalName;
                } else {
                    encoding = charset;
                }
            } else {
                encoding = UNSUPPORTED_CHARSET;
            }
            
            return encoding;
        }
        
        /**
         * Helper method to compare two objects by their Integer indices in the
	 * given map. If the map doesn't contain index for either of the
	 * objects the fallback index will be used for this object instead. 
         *
         * @param indexMap the map which maps objects into Integer indexes. 
         * @param obj1 the first object to be compared.
         * @param obj2 the second object to be compared.
         * @param fallbackIndex the Integer to be used as a fallback index.
         * @return a negative integer, zero, or a positive integer as the
         * 	       first object is mapped to a less, equal to, or greater
	 *             index than the second. 
         */
        protected static int compareIndices(Map indexMap, Object obj1,
					    Object obj2,
                                            Integer fallbackIndex) {
            Integer index1 = (Integer)indexMap.get(obj1);
            Integer index2 = (Integer)indexMap.get(obj2);

            if (index1 == null) {
                index1 = fallbackIndex;
            }
            if (index2 == null) {
                index2 = fallbackIndex;
            }

            return index1.compareTo(index2);
        }

        /**
         * Compares representation classes. Returns a negative integer, zero,
         * or a positive integer as the first representation class is worse
	 * than, equal to, or better than the second.
         * <p>
         * Representation classes are ordered according to the following rules:
         * <ul>
         * <li>All classes assignable to <code>java.rmi.Remote</code> are
	 *     considered equal and better than any other class.
         * <li>Among other classes those assignable to
	 *     <code>java.io.Serializable</code> are considered equal and
	 *     better than any other class.
         * <li>Among other classes those assignable to
	 *     <code>java.io.InputStream</code> are considered equal and better
	 *     than any other class.
         * <li>All other classes are considered equal.
         * </ul>
         *
         * @param class1 the first representation class to be compared.
         * @param class2 the second representation class to be compared.
         * @return a negative integer, zero, or a positive integer as the
         * 	       first argument is worse, equal to, or better than the
         *	       second. 
         */
        protected int compareRepresentationClasses(Class class1, Class class2)
	{
            Integer clsIndex1 = getRepresentationClassIndex(class1);
            Integer clsIndex2 = getRepresentationClassIndex(class2);
            
            return clsIndex1.compareTo(clsIndex2);
        }
        
        protected static Integer getRepresentationClassIndex(Class cls) {
            if (java.rmi.Remote.class.isAssignableFrom(cls)) {
                return REMOTE_CLASS_INDEX;
            } else if (java.io.Serializable.class.isAssignableFrom(cls)) {
                return SERIALIZABLE_CLASS_INDEX;
            } else if (java.io.InputStream.class.isAssignableFrom(cls)) {
                return INPUT_STREAM_CLASS_INDEX;
            } else {
                return OTHER_CLASS_INDEX;
            }
        }
        
        /**
         * Compares DataFlavors by native formats they are mapped into by the
         * flavor-to-native map of this comparator. 
         * Returns a negative integer, zero, or a positive integer as the 
         * native format associated with first data flavor is worse than, 
         * equal to, or better than the one associated with the second.
         * <p>
         * The ordering on the set of native formats is induced by the
         * order on the set of associated integer atoms. That is the native 
         * format with greater atom value is considered better. 
         * <p>
         * Data flavors which are not mapped to any native format with the 
         * flavor-to-native map of this comparator are considered equal and 
         * worse than any "mapped" data flavor.
         *
         * @param flavor1 the first data flavor to be compared.
         * @param flavor2 the second data flavor to be compared.
         * @return a negative integer, zero, or a positive integer as the
         * 	       first argument is worse, equal to, or better than the
         *	       second.
         */
        protected int compareByNativeFormats(DataFlavor flavor1,
					     DataFlavor flavor2) {
            return compareIndices(flavorToFormatMap, flavor1, flavor2, 
                                  new Integer(Integer.MIN_VALUE));
        }   
    }
}
