/*
 * @(#)GlobalCursorManager.java	1.11 01/02/09
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt;

import java.awt.*;
import java.awt.event.InputEvent;
import java.awt.event.InvocationEvent;

/**
 * A stateless class which responds to native mouse moves, Component resizes,
 * Component moves, showing and hiding of Components, minimizing and
 * maximizing of top level Windows, addition and removal of Components,
 * and calls to setCursor().
 */
public abstract class GlobalCursorManager {
    static class CursorEvent extends InvocationEvent {
        static boolean pending = false;

        public CursorEvent() {
	    super(Toolkit.getDefaultToolkit(), new Runnable() {
	        public void run() {
		    boolean shouldUpdate = false;
		    synchronized (this) {
		        if (pending) {
			    pending = false;
			    shouldUpdate = true;
			}
		    }
		    if (shouldUpdate) {
			GlobalCursorManager.getInstance()._updateCursor(false);
		    }
		}
	    });
	}
        public void postSelfIfNotPending(Component heavy) {
	    boolean shouldPost = false;
	    synchronized (this) {
	        if (!pending) {
		    pending = shouldPost = true;
		}
	    }
	    if (shouldPost) {
		SunToolkit.postEvent(SunToolkit.targetToAppContext(heavy),
				     this);
	    }
	}
    }

    /**
     * The GlobalCursorManager is a singleton.
     */
    private static GlobalCursorManager manager;

    /**
     * Use a singleton CursorEvent for better performance.
     */
    private static final CursorEvent curEvent = new CursorEvent();

    /**
     * The last time the cursor was updated, in milliseconds.
     */
    private static long lastUpdateMillis;

    /**
     * Locking object for synchronizing access to lastUpdateMillis. The VM
     * does not guarantee atomicity of longs.
     */
    private static final Object lastUpdateLock = new Object();

    /**
     * The AWT Tree Lock.
     */
    private static final Object treeLock = new Container().getTreeLock();

    /**
     * Should be called for any activity at the Java level which may affect
     * the global cursor, except for Java MOUSE_MOVED events.
     */
    public static void updateCursorImmediately() {
        synchronized (curEvent) {
	    CursorEvent.pending = false;
	}
	GlobalCursorManager.getInstance()._updateCursor(false);
    }

    /**
     * Should be called in response to Java MOUSE_MOVED events. The update
     * will be discarded if the InputEvent is outdated.
     *
     * @param   e the InputEvent which triggered the cursor update.
     */
    public static void updateCursorImmediately(InputEvent e) {
	boolean shouldUpdate;
        synchronized (lastUpdateLock) {
	    shouldUpdate = (e.getWhen() >= lastUpdateMillis);
	}
	if (shouldUpdate) {
	    GlobalCursorManager.getInstance()._updateCursor(true);
	}
    }

    /**
     * Should be called in response to a native mouse enter or native mouse
     * button released message. Should not be called during a mouse drag.
     */
    public static void updateCursorLater(Component heavy) {
        curEvent.postSelfIfNotPending(heavy);
    }

    private static GlobalCursorManager getInstance() {
        if (GlobalCursorManager.manager == null) {
	    GlobalCursorManager.manager = ((SunToolkit)
	        (Toolkit.getDefaultToolkit())).getGlobalCursorManager();
	}
	return manager;
    }

    protected GlobalCursorManager() {}

    /**
     * Set the global cursor to the specified cursor. The component over
     * which the Cursor current resides is provided as a convenience. Not
     * all platforms may require the Component.
     */
    protected abstract void setCursor(Component comp, Cursor cursor,
				      boolean useCache);
    /**
     * Returns the global cursor position, in screen coordinates.
     */
    protected abstract void getCursorPos(Point p);

    protected abstract Component findComponentAt(Container con, int x, int y);
    protected abstract Point getLocationOnScreen(Component com);

    /**
     * Returns the most specific, visible, enabled, heavyweight Component
     * under the cursor. This method should return null iff the cursor is
     * not over any Java Window.
     *
     * @param   useCache If true, the implementation is free to use caching
     * mechanisms because the Z-order, visibility, and enabled state of the
     * Components has not changed. If false, the implementation should not
     * make these assumptions.
     */
    protected abstract Component findHeavyweightUnderCursor(boolean useCache);

    /**
     * Updates the global cursor. We apply a three-step scheme to cursor
     * updates:<p>
     *
     * (1) InputEvent updates which are outdated are discarded by
     * <code>updateCursorImmediately(InputEvent)</code>.<p>
     *
     * (2) If 'useCache' is true, the native code is free to use a cached
     * value to determine the most specific, visible, enabled heavyweight
     * because this update is occuring in response to a mouse move. If
     * 'useCache' is false, the native code must perform a new search given
     * the current mouse coordinates.<p>
     *
     * (3) Once we have determined the most specific, visible, enabled
     * heavyweight, we use findComponentAt to find the most specific, visible,
     * enabled Component.
     */
    private void _updateCursor(boolean useCache) {

        synchronized (lastUpdateLock) {
	    lastUpdateMillis = System.currentTimeMillis();
	}

        Point queryPos = null, p = null;
	Component comp;

	try {
	    comp = findHeavyweightUnderCursor(useCache);
	    if (comp == null) {
		// Cursor is not over a Java Window. Do nothing.
		return;
	    }

	    if (comp instanceof Window) {
		p = comp.getLocation();
	    } else if (comp instanceof Container) {
		p = getLocationOnScreen(comp);
	    }
	    if (p != null) {
		queryPos = new Point();
		getCursorPos(queryPos);
		Component c = findComponentAt((Container)comp,
					      queryPos.x - p.x,
					      queryPos.y - p.y);
		// If findComponentAt returns null, then something bad has
		// happened. For example, the heavyweight Component may
		// have been hidden or disabled by another thread. In that
		// case, we'll just use the originial heavyweight.
		if (c != null) {
		    comp = c;
		}
	    }
	    
	    setCursor(comp, comp.getCursor(), useCache);

	} catch (IllegalComponentStateException e) {
	    // Shouldn't happen, but if it does, abort.
	}
    }	
}
