/*
 * @(#)RepaintArea.java	1.6 00/02/02
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle; 
import java.util.LinkedList;
import java.util.ListIterator;


/**
 * The <code>RepaintArea</code> is a geometric construct created for the 
 * purpose of holding the geometry of several coalesced paint events.  
 * This geometry is accessed synchronously, although it is written such 
 * that painting may still be executed asynchronously.  
 *
 * @author 	Eric Hawkes
 * @version 	1.6 02/02/00
 * @since 	1.3
 */
public class RepaintArea extends Rectangle {

    /**
     * The <code>LinkedList</code> of Rectangles used in 
     * determining intersection with this RepaintArea
     */
    private LinkedList rectangleList; 

    /**
     * Constructs a new <code>RepaintArea</code> whose top-left corner 
     * is at (0,&nbsp;0) in the coordinate space, and whose width and 
     * height are both zero. 
     * @since 	1.3
     */ 
    public RepaintArea() {
	super();
        rectangleList = new LinkedList(); 
    }

    /**
     * Constructs a new <code>RepaintArea</code> initialized to match 
     * the values of the specified RepaintArea.  
     * @param   ra  the <code>RepaintArea</code> from which to copy initial 
     *              values to a newly constructed RepaintArea
     * @since 	1.3
     */ 
    private RepaintArea(RepaintArea ra) {
        // This constructor is private because it should only be called 
	// from the clone method
	super(ra);
        rectangleList = (LinkedList) ra.rectangleList.clone(); 
    }

    /**
     * Adds a <code>Rectangle</code> to this <code>RepaintArea</code>. 
     * The resulting RepaintArea has geometry corresponding to the 
     * union of the RepaintArea and the new rectangle.  
     * @param   r   the specified <code>Rectangle</code>
     * @since 	1.3
     */
    public synchronized void add(Rectangle r) {
        // Make sure this new rectangle has positive dimensions 
        if (r.isEmpty()) {
	    return;
	}

	// Find out if this new rectangle is already contained in our 
	// current RepaintArea.  If so, just return - no need to update 
	// anything or add it to the rectangleList.  
	if (this.contains(r)) {
            return;
        }

        // The new rectangle was not contained, so we must add it to the 
	// rectangleList and update the dimensions of the RepaintArea.  
        rectangleList.add(r);

        if (this.isEmpty()) {
	    // We want to replace the old data in this case
            setBounds(r);
	} else {
            super.add(r);
        }
    }

    /**
     * Tests whether a specifed set of coordinates lies inside the geometry of
     * this <code>RepaintArea</code>.
     * @param     x,&nbsp;y the specified coordinates
     * @return    <code>true</code> if the coordinates lie completely within the 
     *            interior of the <code>RepaintArea</code>;
     *            <code>false</code> otherwise.
     * @since 	  1.3
     */
    public boolean contains(int x, int y) {
	if (super.contains(x, y)) {
	    // The point is contained in the bounding rectangle, 
	    // so iterate through the rectangleList to see if it is 
	    // contained in any of those rectangles. 
	    ListIterator rlIterator = rectangleList.listIterator(0);
	    while (rlIterator.hasNext()) {
                if (((Rectangle) rlIterator.next()).contains(x, y)) {
		    return true;
	        }
	    }
	}

        return false;
    }

    /**
     * Tests whether a specified {@link Point} is inside the boundary
     * of this <code>RepaintArea</code>.
     * @param   p a specified <code>Point</code>
     * @return  <code>true</code> if the specified <code>Point</code> lies 
     *          inside the boundary of the <code>RepaintArea</code>;
     *		<code>false</code> otherwise.
     * @see     #contains(int, int)
     * @since 	1.3
     */
    public boolean contains(Point p) {
	return this.contains(p.x, p.y);
    }

    /**
     * Tests whether or not the interior of this <code>RepaintArea</code> 
     * object completely contains the specified rectangular area.
     * @param x,&nbsp;y the coordinates of the upper left corner of
     *          the specified rectangular area
     * @param w the width of the specified rectangular area
     * @param h the height of the specified rectangular area
     * @return  <code>true</code> if the specified rectangular area
     *          lies completely within the interior of the 
     *          <code>RepaintArea</code>; <code>false</code> otherwise.
     * @since 	1.3
     */
    public boolean contains(int x, int y, int w, int h) {
        // Note that containment is computed using the rectangleList, 
	// not the bounding dimensions of the RepaintArea.  

        if (super.contains(x, y, w, h)) {
	    // The rectangle is contained in the bounding rectangle, 
	    // so iterate through the rectangleList to see if it is 
	    // completely enclosed by any of those rectangles. 
	    ListIterator rlIterator = rectangleList.listIterator(0);
	    while (rlIterator.hasNext()) {
                if (((Rectangle) rlIterator.next()).contains(x, y, w, h)) {
		    return true;
	        }
	    }
	}

        return false; 
    }

    /**
     * Tests whether or not the interior of this <code>RepaintArea</code> 
     * object completely contains the specified <code>Rectangle</code>.
     * @param x,&nbsp;y the coordinates of the upper left corner of
     *          the specified Rectangle
     * @param w the width of the specified Rectangle
     * @param h the height of the specified Rectangle
     * @return  <code>true</code> if the specified Rectangle
     *          lies completely within the interior of the 
     *          <code>RepaintArea</code>; <code>false</code> otherwise.
     * @see     #contains(int, int, int, int)
     * @since 	1.3
     */
    public boolean contains(Rectangle r) {
        return this.contains(r.x, r.y, r.width, r.height);
    }

    /**
     * Determines whether this <code>RepaintArea</code> intersects the 
     * interior of a specified rectangular area. 
     * @param x,&nbsp;y the coordinates of the specified rectangular area
     * @param w the width of the specified rectangular area
     * @param h the height of the specified rectangular area
     * @return  <code>true</code> if this RepaintArea and the interior
     * 	        of the rectangular area intersect; <code>false</code> 
     *          otherwise.  
     * @since 	1.3
     */
    public boolean intersects(int x, int y, int w, int h) {
        // Note that the intersection is computed using the rectangleList, 
	// not the bounding dimensions of the RepaintArea.  

        if (super.intersects(x, y, w, h)) {
	    // The rectangular area intersects the bounding rectangle, 
	    // so iterate through the rectangleList to see if it intersects 
	    // any of those rectangles. 
	    ListIterator rlIterator = rectangleList.listIterator(0);
	    while (rlIterator.hasNext()) { 
                if (((Rectangle) rlIterator.next()).intersects(x, y, w, h)) {
		    return true;
	        }
	    }
	} 

        // The new rectangle did not intersect any of the rectangles in the 
	// rectangleList.  
        return false; 
    }

    /**
     * Determines whether this <code>RepaintArea</code> intersects the 
     * interior of a specified Rectangle. 
     * @param x,&nbsp;y the coordinates of the specified Rectangle
     * @param w the width of the specified Rectangle
     * @param h the height of the specified Rectangle
     * @return  <code>true</code> if this RepaintArea and the interior
     * 	        of the Rectangle intersect; <code>false</code> 
     *          otherwise.  
     * @see     #intersects(int, int, int, int)
     * @since 	1.3
     */
    public boolean intersects(Rectangle r) {
        return intersects(r.x, r.y, r.width, r.height); 
    }

    /**
     * Returns an exact copy of this <code>RepaintArea</code> object.
     * @return  the Created clone object
     * @since 	1.3
     */
    public Object clone() {
        // We must override clone() because of the rectangleList
	return new RepaintArea(this);
    }

    /**
     * Creates a new <code>RepaintArea</code> with the same geometry as this 
     * RepaintArea, then removes all of the geometry from this 
     * RepaintArea and restores it to an empty RepaintArea.  
     * @return	ra a new <code>RepaintArea</code> having the same geometry as 
     *          this RepaintArea.  
     * @since 	1.3
     */
    private synchronized RepaintArea cloneAndReset() { 
        RepaintArea ra = (RepaintArea) this.clone();
	rectangleList.clear(); 
	this.setBounds(0, 0, 0, 0); 
        return ra;
    }

    public void paint(Object target, boolean shouldClearRectBeforePaint) {
        Graphics g = ((Component)target).getGraphics();
	if (g == null) { 
	    return; 
        } 
	try {
	    RepaintArea ra = this.cloneAndReset(); 

	    int numRects = ra.rectangleList.size();

	    if (numRects == 0) {
	        return;
	    }
		
	    if (numRects == 2) {
	        // if the two rectangles do not intersect, send two separate 
	        // paint events instead of painting the bounding rectangle

	        Rectangle r1 = (Rectangle) ra.rectangleList.removeFirst();
		Rectangle r2 = (Rectangle) ra.rectangleList.removeFirst();
            
		if (!r1.intersects(r2)) {
		    g.setClip(r1);
		    if (shouldClearRectBeforePaint) {
		        g.clearRect(r1.x, r1.y, r1.width, r1.height);
		    }
		    ((Component)target).paint(g);

		    g.setClip(r2);
		    if (shouldClearRectBeforePaint) {
		        g.clearRect(r2.x, r2.y, r2.width, r2.height);
		    }
		    ((Component)target).paint(g);

		    return;
		}
	    }

	    g.setClip(ra);
	    if (shouldClearRectBeforePaint) {
	        g.clearRect(ra.x, ra.y, ra.width, ra.height);
	    }
	    ((Component)target).paint(g);
	} finally {
	    g.dispose();
	}
    }

    public void update(Object target) {
        Graphics g = ((Component)target).getGraphics();
	if (g == null) { 
	    return; 
        }
	try {
	    RepaintArea ra = this.cloneAndReset(); 

	    int numRects = ra.rectangleList.size();

	    if (numRects == 0) {
	        return;
	    }
	
	    if (numRects == 2) {
	        // if the two rectangles do not intersect, send two separate 
	        // paint events instead of painting the bounding rectangle

	      Rectangle r1 = (Rectangle) ra.rectangleList.removeFirst();
	      Rectangle r2 = (Rectangle) ra.rectangleList.removeFirst();
            
	      if (!r1.intersects(r2)) {
	          g.setClip(r1);
		  ((Component)target).update(g); 

		  g.setClip(r2);
		  ((Component)target).update(g);

		  return;
	      }
	    }

	    g.setClip(ra);
	    ((Component)target).update(g);
	} finally {
	    g.dispose();
	}
    }
}

