/*
 * @(#)SunToolkit.java	1.53 01/02/09
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.WindowEvent;
import java.awt.im.spi.InputMethodDescriptor;
import java.awt.image.*;
import java.awt.geom.AffineTransform;
import java.io.*;
import java.net.URL;
import java.net.JarURLConnection;
import java.util.*;
import sun.misc.SoftCache;
import sun.awt.font.FontDesignMetrics;
import sun.awt.im.InputContext;
import sun.awt.im.SimpleInputMethodWindow;
import sun.awt.image.*;
import sun.awt.font.NativeFontWrapper;

public abstract class SunToolkit extends Toolkit
    implements WindowClosingSupport, WindowClosingListener {
    /* Force the debug helper classes to initialize */
    {
	DebugHelper.init();
    }

    private static boolean hotjavaUrlCache = false; // REMIND: UGH!!!!!

    /* The key to put()/get() the PostEventQueue into/from the AppContext.
     */
    private static final String POST_EVENT_QUEUE_KEY = "PostEventQueue";
    
    public SunToolkit() {
	EventQueue eventQueue;
        String eqName = Toolkit.getProperty("AWT.EventQueueClass",
                                            "java.awt.EventQueue");
        try {
            eventQueue = (EventQueue)Class.forName(eqName).newInstance();
        } catch (Exception e) {
            System.err.println("Failed loading " + eqName + ": " + e);
            eventQueue = new EventQueue();
        }
	AppContext appContext = AppContext.getAppContext();
	appContext.put(AppContext.EVENT_QUEUE_KEY, eventQueue);

	PostEventQueue postEventQueue = new PostEventQueue(eventQueue);
	appContext.put(POST_EVENT_QUEUE_KEY, postEventQueue);
    }

    /*
     * Create a new AppContext, along with its EventQueue, for a
     * new ThreadGroup.  Browser code, for example, would use this
     * method to create an AppContext & EventQueue for an Applet.
     */
    public static AppContext createNewAppContext() {
	ThreadGroup threadGroup = Thread.currentThread().getThreadGroup();
	EventQueue eventQueue;
        String eqName = Toolkit.getProperty("AWT.EventQueueClass",
                                            "java.awt.EventQueue");
        try {
            eventQueue = (EventQueue)Class.forName(eqName).newInstance();
        } catch (Exception e) {
            System.err.println("Failed loading " + eqName + ": " + e);
            eventQueue = new EventQueue();
        }
	AppContext appContext = new AppContext(threadGroup);
	appContext.put(AppContext.EVENT_QUEUE_KEY, eventQueue);

	PostEventQueue postEventQueue = new PostEventQueue(eventQueue);
	appContext.put(POST_EVENT_QUEUE_KEY, postEventQueue);

	return appContext;
    }

    private static native Object getPrivateKey(Object o);

    // mapping of components to peers, Hashtable<Component,Peer>
    protected static final Hashtable peerMap = new Hashtable();  

    /*
     * Fetch the peer associated with the given target (as specified
     * in the peer creation method).  This can be used to determine 
     * things like what the parent peer is.  If the target is null
     * or the target can't be found (either because the a peer was
     * never created for it or the peer was disposed), a null will
     * be returned.
     */
    protected static Object targetToPeer(Object target) {
	if (target != null) {
	    return peerMap.get(getPrivateKey(target));
	}
	return null;
    }

    protected static void targetCreatedPeer(Object target, Object peer) {
        if (target != null && peer != null) {
            peerMap.put(getPrivateKey(target), peer);
        }
    }

    protected static void targetDisposedPeer(Object target, Object peer) {
	if (target != null && peer != null) {
            Object key = getPrivateKey(target); 
	    if (peerMap.get(key) == peer) {
		peerMap.remove(key);
	    }
	}
    }

    // mapping of Components to AppContexts, WeakHashMap<Component,AppContext>
    private static final Map appContextMap = 
        Collections.synchronizedMap(new WeakHashMap());

    /*
     * Fetch the AppContext associated with the given target.
     * This can be used to determine things like which EventQueue
     * to use for posting events to a Component.  If the target is
     * null or the target can't be found, a null with be returned.
     */
    public static AppContext targetToAppContext(Object target) {
	if (target != null) {
	    return (AppContext)appContextMap.get(getPrivateKey(target));
	}
	return null;
    }

    /*
     * Insert a mapping from target to AppContext, for later retrieval
     * via targetToAppContext() above.
     */
    public static void insertTargetMapping(Object target, AppContext appContext) {
	appContextMap.put(getPrivateKey(target), appContext);
    }

    /*
     * Post an AWTEvent to the Java EventQueue, using the PostEventQueue
     * to avoid possibly calling client code (EventQueueSubclass.postEvent())
     * on the toolkit (AWT-Windows/AWT-Motif) thread.
     */
    public static void postEvent(AppContext appContext, AWTEvent event) {
	PostEventQueue postEventQueue =
	    (PostEventQueue)appContext.get(POST_EVENT_QUEUE_KEY);
        postEventQueue.postEvent(event);
    }

    /*
     * Flush any pending events which haven't been posted to the AWT
     * EventQueue yet.
     */
    public void flushPendingEvents()  {
	AppContext appContext = AppContext.getAppContext();
	PostEventQueue postEventQueue =
	    (PostEventQueue)appContext.get(POST_EVENT_QUEUE_KEY);
        postEventQueue.flush();
    }

    /*
     * Execute a chunk of code on the Java event handler thread for the
     * given target.  Does not wait for the execution to occur before
     * returning to the caller.
     */
    public static void executeOnEventHandlerThread(Object target,
						   Runnable runnable) {
        executeOnEventHandlerThread(new PeerEvent(target, runnable, PeerEvent.PRIORITY_EVENT));
    }

    /*
     * Execute a chunk of code on the Java event handler thread for the
     * given target.  Does not wait for the execution to occur before
     * returning to the caller.
     */
    public static void executeOnEventHandlerThread(PeerEvent peerEvent) {
        postEvent(targetToAppContext(peerEvent.getSource()), peerEvent);
    }

    public Dimension getScreenSize() {
	return new Dimension(getScreenWidth(), getScreenHeight());
    }
    protected abstract int getScreenWidth();
    protected abstract int getScreenHeight();

    public static FontMetrics[] lastMetrics = new FontMetrics[5];

    public FontMetrics getFontMetrics(Font font) {
	for (int i = 0; i < lastMetrics.length; i++) {
	    FontMetrics lm = lastMetrics[i];
	    if (lm == null) {
		break;
	    }
	    if (lm.getFont() == font) {
		return lm;
	    }
	}
	FontMetrics lm = new FontDesignMetrics(font);
	System.arraycopy(lastMetrics, 0, lastMetrics, 1, lastMetrics.length-1);
	lastMetrics[0] = lm;
	return lm;
    }
    
    public String[] getFontList() {
	String[] hardwiredFontList = {
	    "Dialog", "SansSerif", "Serif", "Monospaced", "DialogInput"

	    // -- Obsolete font names from 1.0.2.  It was decided that
	    // -- getFontList should not return these old names:
	    // , "Helvetica", "TimesRoman", "Courier", "ZapfDingbats"
	};
	return hardwiredFontList;
    }

    static SoftCache imgCache = new SoftCache();

    static synchronized java.awt.Image getImageFromHash(Toolkit tk, URL url) {
	SecurityManager sm = System.getSecurityManager();
	if (sm != null) {
	    try {
		java.security.Permission perm = 
		    url.openConnection().getPermission();
		if (perm != null) {
		    try {
			sm.checkPermission(perm);
		    } catch (SecurityException se) {
			// fallback to checkRead/checkConnect for pre 1.2
			// security managers
			if ((perm instanceof java.io.FilePermission) &&
			    perm.getActions().indexOf("read") != -1) {
			    sm.checkRead(perm.getName());
			} else if ((perm instanceof 
			    java.net.SocketPermission) &&
			    perm.getActions().indexOf("connect") != -1) {
			    sm.checkConnect(url.getHost(), url.getPort());
			} else {
			    throw se;
			}
		    }
		}
	    } catch (java.io.IOException ioe) {
		    sm.checkConnect(url.getHost(), url.getPort());
	    }
	}
	java.awt.Image img = (java.awt.Image)imgCache.get(url);
	if (img == null) {
	    try {	
		img = tk.createImage(new URLImageSource(url));
		imgCache.put(url, img);
	    } catch (Exception e) {
	    }
	}
	return img;
    }

    static synchronized java.awt.Image getImageFromHash(Toolkit tk,
                                                        String filename) {
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkRead(filename);
	}
	java.awt.Image img = (java.awt.Image)imgCache.get(filename);
	if (img == null) {
	    try {	
		img = tk.createImage(new FileImageSource(filename));
		imgCache.put(filename, img);
	    } catch (Exception e) {
	    }
	}
	return img;
    }

    public java.awt.Image getImage(String filename) {
	return getImageFromHash(this, filename);
    }

    public java.awt.Image getImage(URL url) {
	return getImageFromHash(this, url);
    }

    public java.awt.Image createImage(String filename) {
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkRead(filename);
	}
	return createImage(new FileImageSource(filename));
    }

    public java.awt.Image createImage(URL url) {
	SecurityManager sm = System.getSecurityManager();
	if (sm != null) {
	    try {
		java.security.Permission perm = 
		    url.openConnection().getPermission();
		if (perm != null) {
		    try {
			sm.checkPermission(perm);
		    } catch (SecurityException se) {
			// fallback to checkRead/checkConnect for pre 1.2
			// security managers
			if ((perm instanceof java.io.FilePermission) &&
			    perm.getActions().indexOf("read") != -1) {
			    sm.checkRead(perm.getName());
			} else if ((perm instanceof 
			    java.net.SocketPermission) &&
			    perm.getActions().indexOf("connect") != -1) {
			    sm.checkConnect(url.getHost(), url.getPort());
			} else {
			    throw se;
			}
		    }
		}
	    } catch (java.io.IOException ioe) {
		    sm.checkConnect(url.getHost(), url.getPort());
	    }
	}
	return createImage(new URLImageSource(url));
    }

    public java.awt.Image createImage(byte[] data, int offset, int length) {
	return createImage(new ByteArrayImageSource(data, offset, length));
    }

    protected EventQueue getSystemEventQueueImpl() {
	AppContext appContext = AppContext.getAppContext();
	EventQueue theEventQueue =
		(EventQueue)appContext.get(AppContext.EVENT_QUEUE_KEY);
        return theEventQueue;
    }

    /**
     * Creates this SunToolkits's implementation of <code>Robot</code> using
     * the specified peer interface.
     * @param     target      the Robot object to be implemented
     * @param     screen      specifies coordinate space for the Robot object
     * @return    this toolkit's implementation of <code>Robot</code>.
     * @see       java.awt.Robot
     * @see       java.awt.peer.RobotPeer
     */
    public abstract RobotPeer createRobot(Robot target, GraphicsDevice screen) throws AWTException;

    /**
     * Give native peers the ability to query the native container 
     * given a native component (eg the direct parent may be lightweight).
     */
    public static Container getNativeContainer(Component c) {
	return Toolkit.getNativeContainer(c);
    }

    /**
     * Returns a new input method adapter descriptor for native input methods.
     */
    public abstract InputMethodDescriptor getInputMethodAdapterDescriptor() throws AWTException;

    /**
     * Returns a new input method window, with behavior as specified in
     * {@link java.awt.im.spi.InputMethodContext#createInputMethodWindow}.
     * If the inputContext is not null, the window should return it from its
     * getInputContext() method. The window needs to implement
     * sun.awt.im.InputMethodWindow, and if or its components can obtain the focus,
     * it needs to call sun.awt.im.InputContext.suppressFocusChange from its
     * setVisible method if the <code>visible</code> argument value is true.
     * <p>
     * SunToolkit subclasses can override this method to return better input
     * method windows.
     */
    public Window createInputMethodWindow(String title, InputContext context) {
        return new sun.awt.im.SimpleInputMethodWindow(title, context, false);
    }

    /**
     * Returns whether enableInputMethods should be set to true for peered
     * TextComponent instances on this platform. False by default.
     */
    public boolean enableInputMethodsForTextComponent() {
        return false;
    }

    /*
     * Get the java.awt.Component object which corresponds to the native
     * window handle argument.  If no Component is found, null is returned.
     * 
     * This method is used by the accessibility software on Win32.
     */
    public Component getComponentFromNativeWindowHandle(int nativeHandle) {
    	return null; // Default implementation returns null -- this is
                     // overloaded by platform-specific code.
    }

    /*
     * Get the native window handle which corresponds to the
     * java.awt.Component object.  If the Component is lightweight,
     * its native container's native window handle is returned.
     * 
     * This method is used by the accessibility software on Win32.
     */
    public int getNativeWindowHandleFromComponent(Component target) {
    	return 0; // Default implementation returns null -- this is
                  // overloaded by platform-specific code.
    }

    /**
     * Gets the default encoding used on the current platform
     * to transfer unicode character data. This is often used 
     * as part of a MIME type.
     */
    public abstract String getDefaultUnicodeEncoding();

    // Support for window closing event notifications
    private transient WindowClosingListener windowClosingListener = null;
    /**
     * @see sun.awt.WindowClosingSupport#getWindowClosingListener
     */
    public WindowClosingListener getWindowClosingListener() {
        return windowClosingListener;
    }
    /**
     * @see sun.awt.WindowClosingSupport#setWindowClosingListener
     */
    public void setWindowClosingListener(WindowClosingListener wcl) {
        windowClosingListener = wcl;
    }

    /**
     * @see sun.awt.WindowClosingListener#windowClosingNotify
     */
    public RuntimeException windowClosingNotify(WindowEvent event) {
        if (windowClosingListener != null) {
            return windowClosingListener.windowClosingNotify(event);
        } else {
            return null;
        }
    }
    /**
     * @see sun.awt.WindowClosingListener#windowClosingDelivered
     */
    public RuntimeException windowClosingDelivered(WindowEvent event) {
        if (windowClosingListener != null) {
            return windowClosingListener.windowClosingDelivered(event);
        } else {
            return null;
        }
    }
    
    public abstract GlobalCursorManager getGlobalCursorManager();

} // class SunToolkit



/*
 * PostEventQueue is a Thread that runs in the same AppContext as the
 * Java EventQueue.  It is a queue of AWTEvents to be posted to the
 * Java EventQueue.  The toolkit Thread (AWT-Windows/AWT-Motif) posts
 * events to this queue, which then calls EventQueue.postEvent().
 *
 * We do this because EventQueue.postEvent() may be overridden by client
 * code, and we mustn't ever call client code from the toolkit thread.
 */
class PostEventQueue extends Thread {
    static private int threadNum = 0;
    private EventQueueItem queueHead = null;
    private EventQueueItem queueTail = null;
    private boolean keepGoing = true;
    private final EventQueue eventQueue;

    PostEventQueue(EventQueue eq) {
	super("SunToolkit.PostEventQueue-"+threadNum);
	synchronized (PostEventQueue.class) { threadNum++; }
	eventQueue = eq;
        setPriority(Thread.NORM_PRIORITY+1);
	start();
    }

    /*
     * Continually post pending AWTEvents to the Java EventQueue.
     */
    public void run() {
	while (keepGoing && !isInterrupted()) {
	    try {
		EventQueueItem item;
		synchronized(this) {
		    while (keepGoing && (queueHead == null)) {
                        notifyAll();
			wait();
		    }
		    if (!keepGoing)
			break;
		    item = queueHead;
                }
		eventQueue.postEvent(item.event);
		synchronized(this) {
		    queueHead = queueHead.next;
		    if (queueHead == null)
			queueTail = null;
		}
	    } catch (InterruptedException e) {
		keepGoing = false; // Exit gracefully when interrupted
		synchronized(this) {
		    notifyAll();
		}
	    }
	}
    }

    /*
     * Enqueue an AWTEvent to be posted to the Java EventQueue.
     */
    synchronized void postEvent(AWTEvent event) {
	EventQueueItem item = new EventQueueItem(event);

	if (queueHead == null) {
	    queueHead = queueTail = item;
	    notifyAll();
	} else {
	    queueTail.next = item;
	    queueTail = item;
	}
    }

    /*
     * Wait for all pending events to be processed before returning.
     * If other events are posted before the queue empties, those
     * will also be processed before this method returns.
     */
    void flush()  {
        if (Thread.currentThread() == this)  {
            return; // Won't wait for itself
        }
        synchronized(this)  {
            if (queueHead != null)  {
                try {
                    wait();
                } catch (InterruptedException e) {
                }
            }
        }
    }

    /*
     * Notifies this PostEventQueue to stop running.
     */
    synchronized void quitRunning() {
	keepGoing = false;
	notifyAll();
    }
} // class PostEventQueue

class EventQueueItem {
    AWTEvent event;
    EventQueueItem next;

    EventQueueItem(AWTEvent evt) {
        event = evt;
    }
} // class EventQueueItem
