/*
 * @(#)SunDragSourceContextPeer.java	1.3 01/02/09
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.dnd;

import java.awt.Component;
import java.awt.Cursor;
import java.awt.Image;
import java.awt.Point;

import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceContext;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.InvalidDnDOperationException;

import java.awt.dnd.peer.DragSourceContextPeer;

import java.util.Map;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import sun.awt.Mutex;
import sun.awt.SunToolkit;

/**
 * @version 1.3, 02/09/01
 * @since 1.3.1
 */
public abstract class SunDragSourceContextPeer implements DragSourceContextPeer, Runnable {

    protected DragSource 		     dragSource;

    protected Component  		     component;
    protected DragGestureEvent		     trigger;
    protected Cursor     		     cursor;
    protected int        		     actions;

    protected long     			     nativeCtxt;

    protected DragSourceContext		     dragSourceContext;

    /*
     * dispatch constants
     */

    final static int DISPATCH_NONE    = 0;
    final static int DISPATCH_ENTER   = 1;
    final static int DISPATCH_MOTION  = 2;
    final static int DISPATCH_CHANGED = 3;
    final static int DISPATCH_EXIT    = 4;
    final static int DISPATCH_FINISH  = 5;

    // follow are used to pass context from platform upcall thread to
    // callback on EventQueue

    protected int		      messagePending = DISPATCH_NONE;
    protected int		      currentA;   // target actions
    protected int		      currentDA;  // current drop
    protected int		      currentMods;  // current modifiers
    protected boolean		      currentSuccess;
    protected int		      currentOperations;

    protected Mutex		      syncLock	  	  = new Mutex();

    /**
     * construct a new SunDragSourceContextPeer
     */

    public SunDragSourceContextPeer(DragGestureEvent dge) {
    	super();

	trigger    = dge;
	component  = dge.getComponent();
	dragSource = dge.getDragSource(); 
    }


    /**
     * initiate a DnD operation ...
     */

    abstract public void startDrag(DragSourceContext dsc, Cursor c, Image di, 
				   Point ioff) throws InvalidDnDOperationException;

    /**
     * return cursor
     */

    public Cursor getCursor() { return cursor; }

    /**
     *
     */

    abstract public void setCursor(Cursor c) throws InvalidDnDOperationException ;

    /**
     * notify peer that the Transferables DataFlavors have changed
     *
     * No longer usefull as teh transferables are determined at the time
     * of the drag.
     */

    public void transferablesFlavorsChanged() {
    }


    /**
     * called from EventDispatchThread via EventQueue
     */

    public void run() {
	int mp;

	syncLock.lock();
	mp = messagePending;
	syncLock.unlock();

	switch (mp) {
	    case DISPATCH_ENTER:
		processEnterMessage();
		break;

	    case DISPATCH_CHANGED:
	        processChangedMessage();
		break;

	    case DISPATCH_MOTION:
		processMotionMessage();
		break;

	    case DISPATCH_EXIT:
		processExitMessage();
		break;

	    case DISPATCH_FINISH:
		processFinishMessage();
		break;
	}
	
	syncLock.lock();
	if (messagePending != DISPATCH_NONE) {
	    messagePending = DISPATCH_NONE;
	    quitSecondaryDispatch();
	}
	syncLock.unlock();
    }

    /**
     * upcall from native code
     */

    private void dragEnter(int dropAction, int targetActions, int modifiers) {
	syncLock.lock();

	// Make sure the previous upcall from a native thread finished
	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}
	    
	// setup context
	currentDA = dropAction;
	currentA = targetActions;
	currentMods = modifiers;

	//schedule callback
	messagePending = DISPATCH_ENTER;
	SunToolkit.executeOnEventHandlerThread(component, this);

	// Spin while thread completes
	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}

	syncLock.unlock();
    }

    private void processEnterMessage() {
	try {
	    dragSourceContext.dragEnter( new DragSourceDragEvent(dragSourceContext, 
								 currentDA, 
								 currentA, 
								 currentMods));
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    /**
     * upcall from native code
     */
 
    private void dragMotion(int dropAction, int targetActions, int modifiers) {
	syncLock.lock();

	// Make sure the previous upcall from a native thread finished
	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}
	    
	// setup context
	currentDA = dropAction;
	currentA = targetActions;
	currentMods = modifiers;

	//schedule callback
	messagePending = DISPATCH_MOTION;
	SunToolkit.executeOnEventHandlerThread(component, this);

	// Spin while thread completes
	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}

	syncLock.unlock();
    }

    private void processMotionMessage() {
	try {
	    dragSourceContext.dragOver( new DragSourceDragEvent(dragSourceContext, 
								currentDA, 
								currentA, 
								currentMods));
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }



    /**
     * upcall from native code
     */

    private void operationChanged(int dropAction, int targetActions, int modifiers) {
	syncLock.lock();

	// Make sure the previous upcall from a native thread finished
	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}
	    
	// setup context
	currentDA = dropAction;
	currentA = targetActions;
	currentMods = modifiers;

	//schedule callback
	messagePending = DISPATCH_CHANGED;
	SunToolkit.executeOnEventHandlerThread(component, this);

	// Spin while thread completes
	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}

	syncLock.unlock();
    }

    private void processChangedMessage() {
	try {
	    dragSourceContext.dropActionChanged( new DragSourceDragEvent(dragSourceContext, 
									 currentDA, 
									 currentA, 
									 currentMods));
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    /**
     * upcall from native code
     */

    private void dragExit() {
	syncLock.lock();

	// Make sure the previous upcall from a native thread finished
	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}
	    
	//schedule callback
	messagePending = DISPATCH_EXIT;
	SunToolkit.executeOnEventHandlerThread(component, this);

	// Spin while thread completes
	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}

	syncLock.unlock();
    }

    private void processExitMessage() {
	try {
	    dragSourceContext.dragExit( new DragSourceEvent(dragSourceContext));
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    /**
     * upcall from native code via implemented class (do)
     */

    protected void dodragDropFinished(boolean success, int operations) {
	syncLock.lock();

	// Make sure the previous upcall from a native thread finished
	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}
	    
	// setup context
	currentSuccess = success;
	currentOperations = operations;

	//schedule callback
	messagePending = DISPATCH_FINISH;
	SunToolkit.executeOnEventHandlerThread(component, this);

	// Spin while thread completes
	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}

	syncLock.unlock();
    }

    private void processFinishMessage() {
	try {

	    DragSourceDropEvent dsde = new DragSourceDropEvent(dragSourceContext,
							       currentOperations,
							       currentSuccess);
	    /*
	     * for platform compatibility we call dragExit here
	     * before we notify the end of the drop
	     */
	    dragSourceContext.dragExit((DragSourceEvent)dsde);

	    dragSourceContext.dragDropEnd(dsde);	

	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    /**
     * Start executing a secondary dispatch. This method cannot be executed
     * in a synchronized block. On platforms where the event thread
     * can continue processing native events when the thread is blocked
     * this method yields the current thread to allow the processing of another
     * thread to be completed. Platforms where the event thread cannot continue
     * processing native events when the thread is blocked this method should
     * be overriden in the platform specific implementation of this class. For 
     * instance WDragSourceContextPeer will override this method by adding a 
     * call to WToolkit.startSecondaryEventLoop();
     * 
     * startSecondaryDispatch is complementary to the method 
     * quitSecondaryDispatch. Whatever is done to start the secondary dispatch
     * must have a complimentary closure of the dispatch. For instance in the
     * default implementation of startSecondaryDispatch() is yield(). The
     * complementary quitSecondaryDispatch is a noop as the thread will 
     * relinquish control when finished. In WDragSourceContextPeer
     * quitSecondaryDispatch calls WToolkit.quitSecondaryEventLoop();
     */
    protected void startSecondaryDispatch() {
	Thread.yield();
    }
    
    /**
     * Quit executing a secondary dispatch. Notifys other threads that thread
     * has completed processing. startSecondaryDispatch is complementary to the
     * method quitSecondaryDispatch. Whatever is done to start the secondary 
     * dispatch must have a complimentary closure of the dispatch. For instance
     * in the default implementation of startSecondaryDispatch() is yield().
     * The complementary quitSecondaryDispatch is a noop as the thread will 
     * relinquish control when finished. In WDragSourceContextPeer
     * quitSecondaryDispatch calls WToolkit.quitSecondaryEventLoop() 
     * complementing startSecondaryDispatch's call to 
     * WToolkit.startSecondaryEventLoop().
     */
    protected void quitSecondaryDispatch() {
    }
}
