/*
 * @(#)SunDropTargetContextPeer.java	1.5 01/04/30
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.dnd;

import java.awt.AWTPermission;
import java.awt.Component;
import java.awt.Container;
import java.awt.Point;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;

import java.awt.dnd.DnDConstants;

import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetContext;
import java.awt.dnd.DropTargetListener;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.InvalidDnDOperationException;

import java.awt.dnd.peer.DropTargetContextPeer;

import java.security.AccessController;
import java.security.AccessControlContext;

import java.util.Map;

import java.io.IOException;
import sun.awt.Mutex;
import sun.awt.SunToolkit;

/**
 * The SunDropTargetContextPeer class is the generic class responsible for
 * handling the interaction between a windowing system's DnD system and Java.
 *
 * @version 1.5, 04/30/01
 * @since 1.3.1
 */
public abstract class SunDropTargetContextPeer implements DropTargetContextPeer, Transferable, Runnable {

    protected Component	      currentC;
    protected DropTarget        currentDT;
    protected DropTargetContext currentDTC;
    protected int[]	      currentT;
    protected int		      currentA;   // target actions
    protected int		      currentDA;  // current drop
    protected int		      prevDA;  // previous drop

    protected long      	      nativeDragContext;	  

    protected Transferable      local;

    // follow are used to pass context from platform upcall thread to
    // callback on EventQueue

    protected int		      messagePending = DISPATCH_NONE;

    protected Component		      currentNC;// current native Container
    protected int		      currentX; // current X coord
    protected int		      currentY; // current Y coord
    protected int		      currentSA;// source actions

    // ...

    
    protected int	              dropStatus   = STATUS_NONE;
    protected boolean		      dropComplete = false;

    protected Object	      	      lock                = new Object();
    protected Mutex		      syncLock	  	  = new Mutex();

    /*
     * dispatch constants
     */

    final static int DISPATCH_NONE    = 0;
    final static int DISPATCH_ENTER   = 1;
    final static int DISPATCH_MOTION  = 2;
    final static int DISPATCH_CHANGED = 3;
    final static int DISPATCH_EXIT    = 4;
    final static int DISPATCH_DROP    = 5;

    /*
     * global lock
     */

    protected static final Object _globalLock = new Object();

    /*
     * a primitive mechanism for advertising intra-JVM Transferables
     */

    protected static Transferable         currentJVMLocalSourceTransferable = null;
    protected static AccessControlContext currentJVMLocalSourceACtxt        = null;

    public static void setCurrentJVMLocalSourceTransferable(Transferable t) throws InvalidDnDOperationException {
	synchronized(_globalLock) {
	    if (t != null && currentJVMLocalSourceTransferable != null) {
	    	throw new InvalidDnDOperationException();
	    } else {
	        currentJVMLocalSourceTransferable = t;
		currentJVMLocalSourceACtxt        = t != null ? AccessController.getContext() : null;
	    }
	} 
    }

    /**
     * obtain the transferable iff the operation is in the same VM
     */

    private static Transferable getJVMLocalSourceTransferable() {
	Transferable t = null;

	synchronized(_globalLock) {

	    /*
	     * If the Transferable is local then we should compare the 
	     * AccessContext of the initiator of the Drag with
	     * that of the receipient of the Drop, if they are equal
	     * or if the receipient is prepared to accept cross ctxt
	     *
	     */

		if (currentJVMLocalSourceTransferable != null) {
		    if (!currentJVMLocalSourceACtxt.equals(AccessController.getContext()))
 		        AccessController.checkPermission(new AWTPermission("acceptDropBetweenAccessControlContexts"));
		}

	    t = currentJVMLocalSourceTransferable;
	}

	return t;
    }

    /*
     * constants used by dropAccept() or dropReject()
     */

    protected final static int STATUS_NONE   =  0; // none pending
    protected final static int STATUS_WAIT   =  1; // drop pending
    protected final static int STATUS_ACCEPT =  2;
    protected final static int STATUS_REJECT = -1;

    /**
     * create the peer
     */

    public SunDropTargetContextPeer() {
	super();
    }

    /**
     * @return the DropTarget associated with this peer
     */

    public DropTarget getDropTarget() { return currentDT; }

    /**
     * @param actions set the current actions
     */

    public void setTargetActions(int actions) {
	synchronized(lock) {
	    currentA = actions & (DnDConstants.ACTION_COPY_OR_MOVE | DnDConstants.ACTION_LINK);
	}
    }
  
    /**
     * @return the current target actions
     */

    public int getTargetActions() { synchronized(lock) { return currentA; } }
  
    /**
     * get the Transferable associated with the drop
     */

    public Transferable getTransferable() {
	return this;
    }

    /**
     * @return current DataFlavors available
     */
    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    abstract public DataFlavor[] getTransferDataFlavors();

    /**
     * @return if the flavor is supported
     */

    abstract public boolean isDataFlavorSupported(DataFlavor df);

    /**
     * @return the data 
     */

    abstract public Object getTransferData(DataFlavor df) 
	throws UnsupportedFlavorException, IOException, 
	    InvalidDnDOperationException; 
    /**
     * @return if the transfer is a local one
     */
  
    public boolean isTransferableJVMLocal() {
	return local != null || getJVMLocalSourceTransferable() != null;
    }

    /**
     * called from EventDispatchThread via EventQueue
     */

    public void run() {
	int mp;

	syncLock.lock();
	mp = messagePending;
	syncLock.unlock();

	switch (mp) {
	    case DISPATCH_ENTER:
		processEnterMessage();
		break;

	    case DISPATCH_CHANGED:
	    case DISPATCH_MOTION:
		processMotionMessage(mp);
		break;

	    case DISPATCH_EXIT:
		processExitMessage();
		break;

	    case DISPATCH_DROP:
                messagePending = DISPATCH_NONE;
		processDropMessage();
		return; // messagePending already cleared
	}
	
	syncLock.lock();
	if (messagePending != DISPATCH_NONE) {
	    messagePending = DISPATCH_NONE;
	    quitSecondaryDispatch();
	}
	syncLock.unlock();
    }

    /**
     * upcall to handle enter messages
     */
  
    private int handleEnterMessage(Component component, int x, int y, int dropAction, int actions, int[] formats, long nativeCtxt) {
	syncLock.lock();

	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}
	    
	// setup context

	nativeDragContext = nativeCtxt;
	currentNC	  = component;
        currentT          = formats;
	currentX	  = x;
	currentY	  = y;
	currentSA         = actions;
	currentDA         = dropAction;
        prevDA            = dropAction;

 	// schedule callback 
	
	messagePending = DISPATCH_ENTER;
	SunToolkit.executeOnEventHandlerThread(component, this);

	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}

	int retval = currentDA;
	syncLock.unlock();
	return retval; // return target's response
    }

    /**
     * actual processing on EventQueue Thread
     */

    private void processEnterMessage() {
	Component  c;
	DropTarget dt   = null;
	boolean    hit;
	Point	   hots = null;

	// we do hit detection and co-ord translation "locked" of course
	local = getJVMLocalSourceTransferable();

	if (currentNC instanceof Container) {
	    c = findComponentAt((Container)currentNC, currentX, currentY);
	} else {
	    c = currentNC.locate(currentX, currentY);
	}

	hit = (c != null    		    &&
	       c.isShowing()		    &&
	       (dt = c.getDropTarget()) != null &&
	       dt.isActive()
	       );

	if (hit) {
	    hots = translate(currentNC, currentX, currentY, c);
	}

	synchronized(lock) {
	    if (currentDTC != null) { // some wreckage from last time
	        currentDTC.removeNotify();
	        currentDTC = null;
	    }
	}

	if (hit) { 
	    synchronized(lock) {
	        currentC   = c;
	        currentDT  = dt;
	        currentDTC = currentDT.getDropTargetContext();

	        currentDTC.addNotify(this);

	        currentA   = dt.getDefaultActions();
	    }

	    try {
		if (currentDA == DnDConstants.ACTION_NONE) {
		    currentDA = mapOperation(currentDT, currentA & currentSA);
		}

	        ((DropTargetListener)dt).dragEnter(new DropTargetDragEvent(currentDTC, 
									   hots, 
									   currentDA, 
									   currentSA));
	    } catch (Exception e) {
		e.printStackTrace();
		synchronized(lock) { currentDA = DnDConstants.ACTION_NONE; }
	    }
	} else synchronized(lock) {
	    currentDT  = null;
	    currentDTC = null;
	    currentDA   = DnDConstants.ACTION_NONE;
	    currentSA   = DnDConstants.ACTION_NONE;
	    currentA   = DnDConstants.ACTION_NONE;
	}

    }

    /**
     * upcall to handle exit messages
     */
  
    private synchronized void handleExitMessage(Component component, long nativeCtxt) {
        syncLock.lock();

	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}

	nativeDragContext = nativeCtxt;

	messagePending = DISPATCH_EXIT;
	SunToolkit.executeOnEventHandlerThread(component, this);

	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}

	syncLock.unlock();
    }

    /**
     * 
     */ 

    private void processExitMessage() {
	DropTarget        dt  = null;
	DropTargetContext dtc = null;

	synchronized(lock) { 
	    if (currentC != null) {
	        DropTarget t = currentC.getDropTarget();

	    	if (t == null) {
		    currentC  = null;
		    currentDT = null;
		    currentT  = null;

		    if (currentDTC != null) currentDTC.removeNotify();
		    
		    currentDTC = null;

		    return;
	        }

		if (t != currentDT) {

		    if (currentDTC != null) currentDTC.removeNotify();

		    dt  = currentDT  = t;
		    dtc = currentDTC = t.getDropTargetContext();

		    currentDTC.addNotify(this);
	        } else {
		    dt  = currentDT;
		    dtc = currentDTC;
		}
	    }
	}

	if (dt != null && dt.isActive()) try {
		((DropTargetListener)dt).dragExit(new DropTargetEvent(dtc));
	} catch (Exception e) {
		e.printStackTrace();
	} finally {
	    synchronized(lock) {
		currentA  = DnDConstants.ACTION_NONE;
		currentSA = DnDConstants.ACTION_NONE;
		currentDA = DnDConstants.ACTION_NONE;
		currentC  = null;
		currentDT = null;
		currentT  = null;

		currentDTC.removeNotify();
		currentDTC = null;

		local = null;
	    }
	}
    }

    /**
     * upcall to handle motion
     */
  
    private int handleMotionMessage(Component component, int x, int y, 
				    int dropAction, int actions, int[] formats, long nativeCtxt) {
        boolean operationChanged = false;

	syncLock.lock();

	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}

	nativeDragContext = nativeCtxt;
	currentNC	  = component;
	currentT          = formats;
	currentX          = x;
	currentY          = y;
	currentSA         = actions;
	currentDA	  = dropAction;
        operationChanged  = prevDA != currentDA;
        prevDA            = currentDA;

	// schedule callback 

        messagePending = operationChanged ? DISPATCH_CHANGED : DISPATCH_MOTION;
	SunToolkit.executeOnEventHandlerThread(component, this);

	while (messagePending != DISPATCH_NONE) {
	    syncLock.unlock();
	    startSecondaryDispatch();
	    syncLock.lock();
	}

	int retval = currentDA;
	syncLock.unlock();
	return retval; // return target's response
    }

    /**
     *
     */

    private void processMotionMessage(int dispatch) {
	Component  	  c     = null;
	Point      	  hots  = null;
	Point      	  nhots = null;
	boolean    	  hit   = false;
	boolean    	  nhit  = false;
	DropTarget 	  dt    = null;
	DropTargetContext dtc   = null;
	DropTarget        ndt   = null;

	if (currentC != null) {
	    boolean same = false;

	    try {
		c = findComponentAt((Container)currentNC, currentX, currentY);
	    } catch (ClassCastException cce) {
		c = currentNC.locate(currentX, currentY);
	    }

	    hots = translate(currentNC, currentX, currentY, currentC);

	    same = currentC.equals(c);

	    if (!same && c != null) {
		nhots = translate(currentNC, currentX, currentY, c);
		nhit  = c.isShowing()		      &&
		    (ndt = c.getDropTarget()) != null &&
		    ndt.isActive();
	    }

	    synchronized(lock) {
		if ((dt = currentC.getDropTarget()) == null) {
		    currentDTC.removeNotify();
		    currentC   = null;
		    currentDTC = null;
		    currentT   = null;
		} else {
		    if (currentDT != dt) {
			if (currentDTC != null) currentDTC.removeNotify();
    
			currentDT  = dt;
			currentDTC = null;
		    }

		    dtc = currentDT.getDropTargetContext();
		    if (dtc != currentDTC) {
			if (currentDTC != null) currentDTC.removeNotify();
			(currentDTC = dtc).addNotify(this);
		    }
		    
		    hit = currentC.isShowing() && dt.isActive();
		}
	    }

	    if (hit) {
	        if (!same) {
		    try {
			((DropTargetListener)dt).dragExit(new DropTargetEvent(dtc));
		    } catch (Exception e) {
			e.printStackTrace();
		    } finally {
			synchronized(lock) {
		            currentDTC.removeNotify();
		            currentDTC = null;
		            currentC   = null;
		            currentT   = null;
			    currentDA   = DnDConstants.ACTION_NONE;
			}
		    }
	        } else {
		    try {
			if (currentDA == DnDConstants.ACTION_NONE) {
			    currentDA = mapOperation(currentDT, currentA & currentSA); 
			}
			if (dispatch == DISPATCH_MOTION)
		            ((DropTargetListener)dt).dragOver(new DropTargetDragEvent(dtc, hots, currentDA, currentSA));
			else
		            ((DropTargetListener)dt).dropActionChanged(new DropTargetDragEvent(dtc, hots, currentDA, currentSA));
		    } catch (Exception e) {
			e.printStackTrace();
			synchronized (lock) { currentDA = DnDConstants.ACTION_NONE; }
		    } finally {
	                return;
		    }
		}
	    } else synchronized(lock) { currentDA = DnDConstants.ACTION_NONE; }
	}

	// lookup possibly new component

	if (c == null) {
	    try {
	        c = findComponentAt((Container)currentNC, currentX, currentY);
	    } catch (ClassCastException cce) {
	        c = currentNC.locate(currentX, currentY);
	    }

	    nhit = c != null			     &&
		   c.isShowing()		     &&
		   (ndt = c.getDropTarget()) != null &&
		   ndt.isActive();

	    if (nhit) nhots = translate(currentNC, currentX, currentY, c);
	}

	if (nhit) {
	    synchronized(lock) {
	        currentC   = c;

	        dt  = currentDT  = ndt;
	        dtc = currentDTC = ndt.getDropTargetContext();

	        currentDTC.addNotify(this);

	        currentA   = ndt.getDefaultActions();
	    }

	    try {
		if (currentDA == DnDConstants.ACTION_NONE) {
		    currentDA = mapOperation(currentDT, currentA & currentSA);
		}
		((DropTargetListener)dt).dragEnter(new DropTargetDragEvent(dtc, 
									   nhots, 
									   currentDA, 
									   currentSA));
	    } catch (Exception e) {
		e.printStackTrace();
	        synchronized(lock) { currentDA = DnDConstants.ACTION_NONE; }
	    }
        } else synchronized(lock) {
	    currentC   = null;
	    currentDT  = null;
	    currentDTC = null;

	    currentDA = DnDConstants.ACTION_NONE;
	}
    }

    /**
     * upcall to handle the Drop message
     */
  
    private void handleDropMessage(Component component, int x, int y, int dropAction, 
				   int actions, int[] formats, long nativeCtxt) {
	// setup context

	nativeDragContext = nativeCtxt;
	currentNC	  = component;
	currentT          = formats;
	currentX          = x;
	currentY          = y;
	currentSA         = actions;
	currentDA	  = dropAction;

	// schedule callback 

	messagePending = DISPATCH_DROP;
	SunToolkit.executeOnEventHandlerThread(component, this);
    }

    /**
     * 
     */
  
    private void processDropMessage() {
	Component  c    = null;
	Point      hots = null;
	boolean    hit  = false;
	DropTarget dt   = null;

	try {
	    c = findComponentAt((Container)currentNC, currentX, currentY);
	} catch (ClassCastException cce) {
	    c = currentNC.locate(currentX, currentY);
	}
	
	hit = (c != null    		    &&
	       c.isShowing()		    &&
	       (dt = c.getDropTarget()) != null &&
	       dt.isActive()
	       );

	if (hit) {
	    hots = translate(currentNC, currentX, currentY, c);
	}

	synchronized(lock) {
	    dropStatus   = STATUS_WAIT; // drop pending ACK
	    dropComplete = false;
	}

	if (hit) {
	    DropTargetContext dtc;
	    boolean	      exception = false;

	    synchronized(lock) {
	        currentC  = c;
	        currentDT = dt;

	        dtc = dt.getDropTargetContext();

	        if (currentDTC != null) currentDTC.removeNotify();

	        (currentDTC = dtc).addNotify(this);
	    }

	    synchronized(_globalLock) {
	        if ((local = getJVMLocalSourceTransferable()) != null)
	            setCurrentJVMLocalSourceTransferable(null);
	    }

	    try {

		/*
		 * for platform constistency we call dragExit before
		 * drop ...
		 */

	        ((DropTargetListener)dt).dragExit(new DropTargetEvent(dtc));

		// now process the drop.

		if (currentDA == DnDConstants.ACTION_NONE) {
		    currentDA = mapOperation(currentDT, currentA & currentSA);
		}

	        ((DropTargetListener)dt).drop(new DropTargetDropEvent(dtc, 
								      hots, 
								      currentDA, 
								      currentSA, 
								      local != null));
	    } catch (Exception e) {
	 	e.printStackTrace();
		exception = true;
	    } finally {
	        synchronized(lock) {
		    if (dropStatus == STATUS_WAIT || exception) {
		        rejectDrop();
		    }
		}
	    }
	} else {
	    rejectDrop();
	}
    }

    /**
     * acceptDrag
     */

    public void acceptDrag(int dragOperation) {
	synchronized(lock) {
	    if (currentDT == null) {
		throw new InvalidDnDOperationException("No Drag pending");
	    }

	    currentDA = mapOperation(currentDT, dragOperation);
	}
    }

    /**
     * rejectDrag
     */

    public void rejectDrag() {
	synchronized(lock) {
	    if (currentDT == null) {
		throw new InvalidDnDOperationException("No Drag pending");
	    }

	    currentDA = DnDConstants.ACTION_NONE;
	}
    }

    /**
     * acceptDrop
     */

    public void acceptDrop(int dropOperation) {
	if (dropOperation == DnDConstants.ACTION_NONE)
	    throw new IllegalArgumentException("invalid acceptDrop() action");

	synchronized (lock) {
	    if (dropStatus != STATUS_WAIT) {
		throw new InvalidDnDOperationException("invalid acceptDrop()");
	    }

	    currentDA = currentA = mapOperation(currentDT, dropOperation & currentSA);

	    dropStatus   = STATUS_ACCEPT;
	    dropComplete = false;
	}
    }

    /**
     * reject Drop
     */

    public void rejectDrop() {
	synchronized(lock) {
	    if (dropStatus != STATUS_WAIT) {
		throw new InvalidDnDOperationException("invalid rejectDrop()");
	    }

	    dropStatus = STATUS_REJECT;
	}

	dropComplete(false);
    }

    /**
     * mapOperation
     */

    private int mapOperation(DropTarget dt, int operation) {
	int[] operations = { 
		DnDConstants.ACTION_MOVE,
		DnDConstants.ACTION_COPY,
		DnDConstants.ACTION_LINK,
	};
	int   ret = DnDConstants.ACTION_NONE;

	for (int i = 0; i < operations.length; i++) {
	    if ((operation & operations[i]) == operations[i]) {
		    ret = operations[i];
		    break;
	    }
	}

	return ret;
    }

    /**
     * signal drop complete
     */

    public abstract void dropComplete(boolean success);

    /**
     * get the Status of the drop
     */

    int getDropStatus() { synchronized(lock) { return dropStatus; } }

    /**
     * translate co-ordinates - called with Component Tree Lock held
     */

    private Point translate(Component anc, int x, int y, Component desc) {
	Component c  = desc;
	int 	  dx = 0;
	int	  dy = 0;

	while (!c.equals(anc)) {
	    Point locn = c.getLocation();
	    dx += locn.x;
	    dy += locn.y;
	    c = c.getParent();
	}

	return new Point(x - dx, y - dy);
    }

    /**
     * Start executing a secondary dispatch. This method cannot be executed
     * in a synchronized block. On platforms where the event thread
     * can continue processing native events when the thread is blocked
     * this method yields the current thread to allow the processing of another
     * thread to be completed. Platforms where the event thread can not continue
     * processing native events when the thread is blocked this method should be 
     * overriden in the platform specific implementation of this class. For 
     * instance WDragSourceContextPeer will override this method by adding a 
     * call to WToolkit.startSecondaryEventLoop();
     * 
     * startSecondaryDispatch is complementary to the method 
     * quitSecondaryDispatch. Whatever is done to start the secondary dispatch
     * must have a complimentary closure of the dispatch. For instance in the
     * default implementation of startSecondaryDispatch() is yield(). The
     * complementary quitSecondaryDispatch is a noop as the thread will 
     * relinquish control when finished. In WDragSourceContextPeer
     * quitSecondaryDispatch calls WToolkit.quitSecondaryEventLoop();
     *
     */
    protected void startSecondaryDispatch() {
	Thread.yield();
    }
    
    /**
     * Quit executing a secondary dispatch. Notifys other threads that thread has
     * completed processing. startSecondaryDispatch is complementary to the 
     * method quitSecondaryDispatch. Whatever is done to start the secondary 
     * dispatch must have a complimentary closure of the dispatch. For instance
     * in the default implementation of startSecondaryDispatch() is yield(). The
     * complementary quitSecondaryDispatch is a noop as the thread will 
     * relinquish control when finished. In WDragSourceContextPeer
     * quitSecondaryDispatch calls WToolkit.quitSecondaryEventLoop() 
     * complementing startSecondaryDispatch's call to 
     * WToolkit.startSecondaryEventLoop().

     */
    protected void quitSecondaryDispatch() {
    }

    /**
     * Special version of Container.findComponentAt which bypasses a
     * JRootPane's glass pane. This is a temporary fix until DnD is
     * changed to use the normal event dispatching model in java.awt.
     */
    private native Component findComponentAt(Container container,
					     int x, int y);

}

