/*
 * @(#)ExtendedTextSourceLabel.java	1.20 00/02/02
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
/*
 * @(#)ExtendedTextSourceLabel.java	1.20 00/02/02
 *
 * (C) Copyright IBM Corp. 1998 - All Rights Reserved
 */

package sun.awt.font;

import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Shape;

import java.awt.font.FontRenderContext;
import java.awt.font.GlyphJustificationInfo;
import java.awt.font.GlyphMetrics;
import java.awt.font.LineMetrics;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.AffineTransform;

import sun.awt.font.NativeFontWrapper;

/**
 * Default implementation of ExtendedTextLabel.
 */

// !!! this implementation does not perform indic reordering

public class ExtendedTextSourceLabel extends ExtendedTextLabel {
  TextSource source;

  // caches
  Rectangle2D lb;
  Rectangle2D ab;
  Rectangle2D vb;
  StandardGlyphVector gv;
  Segment shapedSeg;
  float[] charinfo;

  /**
   * Create from a TextSource.
   */
  public ExtendedTextSourceLabel(TextSource source) {
    this.source = source;
  }

  /**
   * Create from a TextSource, optionally using cached data from oldLabel starting at the offset.
   * If present oldLabel must have been created from a run of text that includes the text used in
   * the new label.  Start in source corresponds to logical character offset in oldLabel.
   */
  public ExtendedTextSourceLabel(TextSource source, ExtendedTextSourceLabel oldLabel, int offset) {
    // currently no optimization.
    this.source = source;
  }

  // TextLabel API

  public Rectangle2D getLogicalBounds(float x, float y) {
    if (lb == null) {
      lb = createLogicalBounds();
    }
    return new Rectangle2D.Float((float)(lb.getX() + x),
                                 (float)(lb.getY() + y),
                                 (float)lb.getWidth(),
                                 (float)lb.getHeight());
  }

  public Rectangle2D getVisualBounds(float x, float y) {
    if (vb == null) {
      vb = createVisualBounds();
    }
    return new Rectangle2D.Float((float)(vb.getX() + x),
                                 (float)(vb.getY() + y),
                                 (float)vb.getWidth(),
                                 (float)vb.getHeight());
  }

  public Rectangle2D getAlignBounds(float x, float y) {
    if (ab == null) {
      ab = createAlignBounds();
    }
    return new Rectangle2D.Float((float)(ab.getX() + x),
                                 (float)(ab.getY() + y),
                                 (float)ab.getWidth(),
                                 (float)ab.getHeight());

  }

  public Shape getOutline(float x, float y) {
    return getGV().getOutline(x, y);
  }

  public void draw(Graphics2D g, float x, float y) {
    g.drawGlyphVector(getGV(), x, y);
  }

  protected Rectangle2D createLogicalBounds() {
    return getGV().getLogicalBounds();
  }

  protected Rectangle2D createVisualBounds() {
    return getGV().getVisualBounds();
  }

  // like createLogicalBounds except ignore leading and logically trailing white space
  // this assumes logically trailing whitespace is also visually trailing

  protected Rectangle2D createAlignBounds() {
    LineMetrics lm = source.getLineMetrics();

    float[] info = getCharinfo();

    float al = 0f;
    float at = -lm.getAscent();
    float aw = 0f;
    float ah = lm.getAscent() + lm.getDescent();

    boolean lineIsLTR = source.lineIsLTR();
    int rn = info.length - numvals;
    while (rn > 0 && ((info[rn+advx] == 0) || (lineIsLTR && info[rn+visw] == 0))) {
      rn -= numvals;
    }

    if (rn >= 0) {
      int ln = 0;
      while (ln < rn && ((info[ln+advx] == 0) || (!lineIsLTR && info[ln+visw] == 0))) {
        ln += numvals;
      }

      al = Math.max(0f, info[ln+posx]);
      aw = info[rn+posx] + info[rn+advx] - al;
    }

    return new Rectangle2D.Float(al, at, aw, ah);
  }

  private final StandardGlyphVector getGV() {
    if (gv == null) {
      gv = createGV();
    }

    return gv;
  }

  private final Segment getShapedSegment() {
    if (shapedSeg == null) {
      shapedSeg = createShapedSegment();
    }
    return shapedSeg;
  }

  protected Segment createShapedSegment() {
    Segment seg = new Segment();
    Shaping.getShapedSegment(source, seg);
    return seg;
  }

  protected StandardGlyphVector createGV() {
    Font f = source.getFont();
    FontRenderContext frc = source.getFRC();
    Segment seg = getShapedSegment();

    return new StandardGlyphVector(f, seg.text, seg.start, seg.length, frc);
  }

  // ExtendedTextLabel API

  private static final int posx = 0,
    posy = 1,
    advx = 2,
    advy = 3,
    visx = 4,
    visy = 5,
    visw = 6,
    vish = 7;
  private static final int numvals = 8;

  public int getNumCharacters() {
    return source.getLength();
  }

  public float getItalicAngle() {
    return source.getFont().getItalicAngle();
  }

  public LineMetrics getLineMetrics() {
    return source.getLineMetrics();
  }

  public float getCharX(int index) {
    validate(index);
    return getCharinfo()[l2v(index) * numvals + posx];
  }

  public float getCharY(int index) {
    validate(index);
    return getCharinfo()[l2v(index) * numvals + posy];
  }

  public float getCharAdvance(int index) {
    validate(index);
    return getCharinfo()[l2v(index) * numvals + advx];
  }

  public Rectangle2D getCharVisualBounds(int index, float x, float y) {
    validate(index);
    float[] charinfo = getCharinfo();
    index = l2v(index) * numvals;
    return new Rectangle2D.Float(
      charinfo[index + visx] + x,
      charinfo[index + visy] + y,
      charinfo[index + visw],
      charinfo[index + vish]);
  }

  private void validate(int index) {
    if (index < 0) {
      throw new IllegalArgumentException("index " + index + " < 0");
    } else if (index >= source.getLength()) {
      throw new IllegalArgumentException("index " + index + " < " + source.getLength());
    }
  }

  /*
  public int hitTestChar(float x, float y) {
    // !!! return index of char hit, for swing
    // result is negative for trailing-edge hits
    // no italics so no problem at margins.
    // for now, ignore y since we assume horizontal text

    // find non-combining char origin to right of x
    float[] charinfo = getCharinfo();

    int n = 0;
    int e = source.getLength();
    while (n < e && charinfo[n + advx] != 0 && charinfo[n + posx] > x) {
      n += numvals;
    }
    float rightx = n < e ? charinfo[n+posx] : charinfo[e - numvals + posx] + charinfo[e - numvals + advx];

    // find non-combining char to left of that char
    n -= numvals;
    while (n >= 0 && charinfo[n+advx] == 0) {
      n -= numvals;
    }
    float leftx = n >= 0 ? charinfo[n+posx] : 0;
    float lefta = n >= 0 ? charinfo[n+advx] : 0;

    n /= numvals;

    boolean left = true;
    if (x < leftx + lefta / 2f) {
      // left of prev char
    } else if (x < (leftx + lefta + rightx) / 2f) {
      // right of prev char
      left = false;
    } else {
      // left of follow char
      n += 1;
    }

    if (!source.dataIsLTR()) {
      n = getNumCharacters() - 1 - n;
      left = !left;
    }

    return left ? n : -n;
  }
  */

  public int logicalToVisual(int logicalIndex) {
    validate(logicalIndex);
    return l2v(logicalIndex);
  }

  public int visualToLogical(int visualIndex) {
    validate(visualIndex);
    return v2l(visualIndex);
  }

  public int getLineBreakIndex(int start, float width) {
    float[] charinfo = getCharinfo();
    int length = source.getLength();
    --start;
    while (width >= 0 && ++start < length) {
      width -= charinfo[l2v(start) * numvals + advx];
    }

    return start;
  }

  public float getCharAdvanceBetween(int start, int limit) {
    float a = 0f;

    float[] charinfo = getCharinfo();
    --start;
    while (++start < limit) {
      a += charinfo[l2v(start) * numvals + advx];
    }

    return a;
  }

  public boolean caretAtOffsetIsValid(int offset) {
      // REMIND: improve this implementation

      if (offset == 0 || offset == source.getLength()) {
          return true;
      }
      char c = source.getChars()[source.getStart() + offset];
      if (c == '\t' || c == '\n' || c == '\r') { // hack
          return true;
      }
      int v = l2v(offset);
      if (!source.dataIsLTR()) {
          v += 1;
          if (v == source.getLength()) {
              return true;
          }
      }
      return getCharinfo()[v * numvals + advx] != 0;
  }

  private final float[] getCharinfo() {
    if (charinfo == null) {
      charinfo = createCharinfo();
    }
    return charinfo;
  }

  protected float[] createCharinfo() {
    StandardGlyphVector gv = getGV();
    return gv.getGlyphInfo(); // assume 1-1 chars to glyphs, consistent ordering
  }

  protected int l2v(int index) {
    return source.dataIsLTR() ? index : source.getLength() - 1 - index;
  }

  protected int v2l(int index) {
    return source.dataIsLTR() ? index : source.getLength() - 1 - index;
  }

  public ExtendedTextLabel getSubset(int start, int limit, int dir) {
    return new ExtendedTextSourceLabel(source.getSubSource(start, limit-start, dir));
  }

  public String toString() {
    StringBuffer buf = new StringBuffer();
    buf.append(super.toString());
    buf.append("[source:");
    buf.append(source.toString(source.WITHOUT_CONTEXT));
    buf.append(", lb:");
    buf.append(lb);
    buf.append(", ab:");
    buf.append(ab);
    buf.append(", vb:");
    buf.append(vb);
    buf.append(", gv:");
    buf.append(gv);
    buf.append(", seg:");
    buf.append(shapedSeg);
    buf.append(", ci:");
    buf.append(charinfo == null ? "null" : Integer.toString(charinfo.length));
    buf.append("]");

    return buf.toString();
  }

  public static ExtendedTextLabel create(TextSource source) {
    return new ExtendedTextSourceLabel(source);
  }

  public int getNumJustificationInfos() {
    return getGV().getNumGlyphs();
  }


  public void getJustificationInfos(GlyphJustificationInfo[] infos, int infoStart, int charStart, int charLimit) {
    // This simple implementation only uses spaces for justification.
    // Since regular characters aren't justified, we don't need to deal with
    // special infos for combining marks or ligature substitution glyphs.
    // added character justification for kanjii only 2/22/98
    
    StandardGlyphVector gv = getGV();
    
    float[] charinfo = getCharinfo();

    float size = gv.getFont().getSize2D();
    
    GlyphJustificationInfo nullInfo = 
      new GlyphJustificationInfo(0, 
                                 false, GlyphJustificationInfo.PRIORITY_NONE, 0, 0,
                                 false, GlyphJustificationInfo.PRIORITY_NONE, 0, 0);

    GlyphJustificationInfo spaceInfo =
      new GlyphJustificationInfo(size,
                                 true, GlyphJustificationInfo.PRIORITY_WHITESPACE, 0, size,
                                 true, GlyphJustificationInfo.PRIORITY_WHITESPACE, 0, size / 4f);

    GlyphJustificationInfo kanjiInfo =
      new GlyphJustificationInfo(size,
				 true, GlyphJustificationInfo.PRIORITY_INTERCHAR, size, size,
				 false, GlyphJustificationInfo.PRIORITY_NONE, 0, 0);

    char[] chars = source.getChars();
    int offset = source.getStart();

    // assume data is 1-1 and either all rtl or all ltr, for now

    int numGlyphs = gv.getNumGlyphs();
    int minGlyph = 0;
    int maxGlyph = numGlyphs;
    boolean ltr = source.dataIsLTR();
    if (charStart != 0 || charLimit != source.getLength()) {
      if (ltr) {
        minGlyph = charStart;
        maxGlyph = charLimit;
      } else {
        minGlyph = numGlyphs - charLimit;
        maxGlyph = numGlyphs - charStart;
      }
    }

    for (int i = 0; i < numGlyphs; ++i) {
      GlyphJustificationInfo info = null;
      if (i >= minGlyph && i < maxGlyph) {
	if (charinfo[i * numvals + advx] == 0) { // combining marks don't justify
	  info = nullInfo;
	} else {
	  int ci = v2l(i); // 1-1 assumption again
	  char c = chars[offset + ci];
	  if (Character.isWhitespace(c)) {
	    info = spaceInfo;
	    // CJK, Hangul, CJK Compatibility areas
	  } else if (c >= 0x4e00 &&
		     (c < 0xa000) ||
		     (c >= 0xac00 && c < 0xd7b0) ||
		     (c >= 0xf900 && c < 0xfb00)) {
	    info = kanjiInfo;
	  } else {
	    info = nullInfo;
	  }
	}
      }
      infos[infoStart + i] = info;
    }
  }

  public ExtendedTextLabel applyJustificationDeltas(float[] deltas, int deltaStart, boolean[] flags) {

    // when we justify, we need to adjust the charinfo since spaces
    // change their advances.  preserve the existing charinfo.

    float[] newCharinfo = (float[])getCharinfo().clone();

    // we only push spaces, so never need to rejustify
    flags[0] = false;

    // preserve the existing gv.

    StandardGlyphVector newgv = (StandardGlyphVector)getGV().clone();
    float[] newPositions = newgv.getGlyphPositions(null);
    int numGlyphs = newgv.getNumGlyphs();

    /*
    System.out.println("oldgv: " + getGV() + ", newgv: " + newgv);
    System.out.println("newpositions: " + newPositions);
    for (int i = 0; i < newPositions.length; i += 2) {
      System.out.println("[" + (i/2) + "] " + newPositions[i] + ", " + newPositions[i+1]);
    }

    System.out.println("deltas: " + deltas + " start: " + deltaStart);
    for (int i = deltaStart; i < deltaStart + numGlyphs; i += 2) {
      System.out.println("[" + (i/2) + "] " + deltas[i] + ", " + deltas[i+1]);
    }
    */

    char[] chars = source.getChars();
    int offset = source.getStart();

    // accumulate the deltas to adjust positions and advances.
    // handle whitespace by modifying advance,
    // handle everything else by modifying position before and after
    
    float deltaPos = 0;
    for (int i = 0; i < numGlyphs; ++i) {
      if (Character.isWhitespace(chars[offset + v2l(i)])) {
        newPositions[i*2] += deltaPos;

        float deltaAdv = deltas[deltaStart + i*2] + deltas[deltaStart + i*2 + 1];

        newCharinfo[i * numvals + posx] += deltaPos;
        newCharinfo[i * numvals + visx] += deltaPos;
        newCharinfo[i * numvals + advx] += deltaAdv;

        deltaPos += deltaAdv;
      } else {
        deltaPos += deltas[deltaStart + i*2];
        
        newPositions[i*2] += deltaPos;
        newCharinfo[i * numvals + posx] += deltaPos;
        newCharinfo[i * numvals + visx] += deltaPos;

        deltaPos += deltas[deltaStart + i*2 + 1];
      }
    }
    newPositions[numGlyphs * 2] += deltaPos;

    newgv.setGlyphPositions(newPositions);

    /*
    newPositions = newgv.getGlyphPositions(null);
    System.out.println(">> newpositions: " + newPositions);
    for (int i = 0; i < newPositions.length; i += 2) {
      System.out.println("[" + (i/2) + "] " + newPositions[i] + ", " + newPositions[i+1]);
    }
    */

    ExtendedTextSourceLabel result = new ExtendedTextSourceLabel(source);
    result.gv = newgv;
    result.charinfo = newCharinfo;

    return result;
  }
}
