/*
 * @(#)StandardGlyphVector.java	1.31 00/02/02
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.awt.font;

import java.awt.Graphics2D;
import java.awt.Font;
import java.awt.Polygon;        // remind - need a floating point version
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.AffineTransform;
import java.text.CharacterIterator;
import java.awt.font.FontRenderContext;
import java.awt.Shape;
import java.awt.font.GlyphMetrics;
import java.awt.font.GlyphVector;
import java.awt.font.GlyphJustificationInfo;

/**
*   GlyphVector is a collection of Glyphs containing geometric information
*   for the placement the GlyphVector and of each Glyph in a transformed
*   coordinate space which should correspond to the device on which the
*   GlyphVector is ultimately to be displayed.  The GlyphVector does not
*   attempt any interpretation of the sequence of Glyphs it contains.
*   Relationships between adjacent Glyphs in sequence are solely used
*   in determination of placement of the Glyphs in the visual coordinate
*   space.
*   <p>
*   Instances of GlyphVector are created by a Font.
*   <p>
*   In a text processing application which can cache intermediate
*   representations of text, creation and subsequent caching of a GlyphVector
*   for use during rendering is the fastest method to present the visual
*   representation of characters to a user.
*   <p>
*   A GlyphVector is associated with exactly one Font, and can provide data
*   useful only in relation to the Font specified for the GlyphVector.  In
*   addition, metrics obtained from a GlyphVector are not generally
*   geometrically scaleable since the pixelization and spacing are
*   dependent on grid-fitting algorithms within a Font.  To facilitate
*   accurate measurement of a GlyphVector and its component Glyphs, you must
*   specify a scaling transform, anti-alias mode, and fractional metrics
*   mode at the time the GlyphVector is created.  (These characteristics can
*   be derived from the destination device.)
*   <p>
*   For each Glyph in the GlyphVector, you may obtain:
*       1) the position of the Glyph
*       2) the transform associated with the Glyph
*       3) the metrics of the Glyph in the context of the GlyphVector
*          (the metrics of the Glyph may be different under different
*           transforms, application specified rendering hints, and
*           the specific instance of the Glyph within the GlyphVector)
*   <p>
*   GlyphVectors are immutable.  There are no means provided to alter the
*   state of the GlyphVector.  Altering the data used to create the GlyphVector
*   will not alter the state of the GlyphVector.
*   <p>
*   Methods are provided to create new GlyphVectors which are the result of
*   editing editing operations on the GlyphVector such as Glyph insertion and
*   deletion.  These methods are most appropriate to applications which
*   are forming combinations such as ligatures from existing Glyphs or are
*   breaking such combinations into their component parts for visual
*   presentation.
*   <p>
*   Methods are provided to create new GlyphVectors which are the result of
*   specifying new positions for the Glyphs within the GlyphVector.  These
*   methods are most appropriate to applications which are performing
*   justification operations for the presentation of the Glyphs.
*   <p>
*   Methods are provided to return both the visual and logical bounds
*   of the entire GlyphVector or of individual Glyphs within the GlyphVector.
*   <p>
*   Methods are provided to return a Shape for the GlyphVector, and for
*   individual Glyphs within the GlyphVector.
*
*   @see Font
*   @see GlyphMetrics
*   @see TextLayout
*   @version 30 Mar 1998
*   @author Charlton Innovations, Inc.
*/

public class StandardGlyphVector extends GlyphVector {

    private Font font;
    private int [] glyphs;          // name used by native code
    private float [] positions;     // name used by native code
    private FontRenderContext frc;

    // transform matrices for:
    // fontSize * fontTX * frcTX
    private double [] glyphTX;

    public StandardGlyphVector(Font theFont, String unicodeStr,
                               FontRenderContext frc)
    {
        this.font = theFont;
        this.glyphs = null;
        this.positions = null;
        this.frc = frc;
        double [] tx = getGlyphTX();
        NativeFontWrapper.populateGlyphVector(
            font, unicodeStr.toCharArray(), 0, unicodeStr.length(),
            tx,
            frc.isAntiAliased(),
            frc.usesFractionalMetrics(),
            this);
    }

    public StandardGlyphVector(Font theFont, char[] unicodes,
                               FontRenderContext frc)
    {
        this.font = theFont;
        this.glyphs = null;
        this.positions = null;
        this.frc = frc;
        double [] tx = getGlyphTX();
        NativeFontWrapper.populateGlyphVector(
            font, unicodes, 0, unicodes.length,
            tx,
            frc.isAntiAliased(),
            frc.usesFractionalMetrics(),
            this);
    }

  /**
   * Create a StandardGlyphVector from a range of unicodes.
   * @throws IndexOutOfBoundsException if start < 0 or count < 0 or start + count > unicodes.length.
   */
    public StandardGlyphVector(Font theFont, char[] unicodes, int start, int count,
                               FontRenderContext frc)
    {
        if (start < 0 || count < 0 || start + count > unicodes.length) {
            throw new ArrayIndexOutOfBoundsException("start or count out of bounds");
        }
        this.font = theFont;
        this.glyphs = null;
        this.positions = null;
        this.frc = frc;
        double [] tx = getGlyphTX();
        NativeFontWrapper.populateGlyphVector(
            font, unicodes, start, count,
            tx,
            frc.isAntiAliased(),
            frc.usesFractionalMetrics(),
            this);
    }

    public StandardGlyphVector(Font theFont, CharacterIterator ciUnicode,
                               FontRenderContext frc)
    {
        char [] unicodes = new char [ciUnicode.getEndIndex()
                                            - ciUnicode.getBeginIndex()];
        for(char c = ciUnicode.first();
                c != CharacterIterator.DONE;
                c = ciUnicode.next()) {
            unicodes[ciUnicode.getIndex() - ciUnicode.getBeginIndex()] = c;
        }
        this.font = theFont;
        this.glyphs = null;
        this.positions = null;
        this.frc = frc;
        double [] tx = getGlyphTX();
        NativeFontWrapper.populateGlyphVector(
            font, unicodes, 0, unicodes.length,
            tx,
            frc.isAntiAliased(),
            frc.usesFractionalMetrics(),
            this);
    }

    public StandardGlyphVector(Font theFont, int [] glyphCodes,
                               FontRenderContext frc)
    {
        this.font = theFont;
        this.glyphs = glyphCodes;
        this.positions = null;
        this.frc = frc;
    }

    private double[] getGlyphTX()
    {
        if (null != glyphTX) {
            return glyphTX;
        }
        glyphTX = new double[6];
        float ptSize = font.getSize2D();
        AffineTransform tx = (AffineTransform)frc.getTransform().clone();
        tx.scale(ptSize, ptSize);
        tx.preConcatenate(font.getTransform());
    //  dest2D.draw and dest2D.fill will add the following transform...
    //    tx.preConcatenate(dest2D.transform);
        tx.getMatrix(glyphTX);
        return glyphTX;
    }

    public void performDefaultLayout()
    {
        double [] tx = getGlyphTX();
        NativeFontWrapper.layoutGlyphVector(
            font, tx,
            frc.isAntiAliased(),
            frc.usesFractionalMetrics(),
            (float)tx[4], (float)tx[5],
            this);
    }

    // methods associated with creation-time state
    /**
    *   Return the Font associated with this GlyphVector.
    *   @return Font used to create this GlyphVector
    *   @see Font
    */
    public Font getFont()
    {
        return this.font;
    }

    public FontRenderContext getFontRenderContext()
    {
        return this.frc;
    }

    // methods associated with the GlyphVector as a whole
    /**
    *   Return the number of Glyphs in this GlyphVector, you should
    *   Use this information to create arrays which are to be
    *   filled with results of other information retreival
    *   operations.
    *   @return number of Glyphs in this GlyphVector
    */
    public int getNumGlyphs()
    {
        return glyphs.length;
    }

    /**
    *   Return the glyphcode of the specified Glyph.
    *   This return value is meaningless to anything other
    *   than a Font and can be used to ask the Font
    *   about the existing of ligatures and other context
    *   sensitive information.
    *   @see Font#getLigature
    */
    public int getGlyphCode(int glyphIndex)
    {
      //        if (glyphIndex < 0 || glyphIndex >= glyphs.length) {
      //	    throw new IndexOutOfBoundsException("glyphIndex = " + glyphIndex);
      //        }
        return glyphs[glyphIndex];
    }

    /**
    *   Return and array of glyphcodes for the the specified Glyphs.
    *   The contents of this return value are meaningless to anything other
    *   than a Font and can be used to ask the Font about the existence of
    *   ligatures and other context sensitive information.  Use this method
    *   for convenience and performance in processing of glyphcodes.
    *   If no array is passed in, one will be created for you.
    *   @see Font#getLigature
    */
    public int[] getGlyphCodes(int beginGlyphIndex, int numEntries,
                               int[] codeReturn)
    {
        if (numEntries < 0) {
            throw new IllegalArgumentException("numEntries = " + numEntries);
        }
        if (beginGlyphIndex < 0) {
            throw new IndexOutOfBoundsException("beginGlyphIndex = " + beginGlyphIndex);
        }
        if (beginGlyphIndex > glyphs.length - numEntries) { // watch out for overflow if index + count overlarge
            throw new IndexOutOfBoundsException("beginGlyphIndex + numEntries = " + (beginGlyphIndex + numEntries));
        }

        int [] retval = codeReturn;
        if (retval == null) {
            retval = new int[numEntries];
        }
        for (int i=0; i < numEntries; i++) {
            retval[i] = glyphs[i + beginGlyphIndex];
        }

        return retval;
    }

    // methods associated with the GlyphVector as a whole,
    // after layout.
    /**
     *   Return the logical bounds of this GlyphVector.
     *   Use when positioning this GlyphVector in relation
     *   to visually adjacent GlyphVectors.
     */
    public Rectangle2D getLogicalBounds()
    {
      // still broken, need to record ascent, descent, leading in
      // case they differ for different ranges of text in the font.
      // for now assume fonts only have one value for these.

        float [] metrics = new float[4];
        NativeFontWrapper.getFontMetrics(font, getGlyphTX(),
                                        frc.isAntiAliased(),
                                        frc.usesFractionalMetrics(),
                                        metrics);
        float minX = 0;
        float minY = -metrics[0];
        float maxX = 0;
        float maxY = metrics[1] + metrics[2];

        if (glyphs.length > 0) {
          // assume horizontal

          initPositions();
          maxX = positions[positions.length - 2];
        }
        
        return new Rectangle2D.Float(minX, minY, maxX - minX, maxY - minY);     
    }

    /**
    *   Return the visual bounds of this GlyphVector
    *   The visual bounds is tightest Rectangle
    *   enclosing all non-background pixels in the
    *   rendered representation of the GlyphVector.
    */
    public Rectangle2D getVisualBounds()
    {
        Shape theShape = getOutline();
        return (Rectangle2D)theShape.getBounds();
    }

    /**
    *   Return a Shape whose interior corresponds to the
    *   visual representation of the GlyphVector.
    */
    public Shape getOutline() {
        return getOutline(0, 0);
    }

    /**
    *   Return a Shape whose interior corresponds to the
    *   visual representation of the GlyphVector. Offset
    *   to x, y
    */
    public Shape getOutline(float x, float y)
    {
        GeneralPath dstShape = new GeneralPath(GeneralPath.WIND_NON_ZERO);
        double tx[] = getGlyphTX();
        dstShape = (GeneralPath)NativeFontWrapper.getGlyphVectorOutline(
                                        this, font, tx,
                                        frc.isAntiAliased(),
                                        frc.usesFractionalMetrics(),
                                        (float)tx[4]+x, (float)tx[5]+y,
                                        dstShape);
        return dstShape;
    }

    /**
    *   Return a Shape whose interior corresponds to the
    *   visual representation of the specified glyph
    *   within the GlyphVector.
    */
    public Shape getGlyphOutline(int glyphIndex)
    {
        if (glyphIndex < 0 || glyphIndex >= glyphs.length) {
            throw new IndexOutOfBoundsException("glyphIndex = " + glyphIndex);
        }

        GeneralPath dstShape = new GeneralPath(GeneralPath.WIND_NON_ZERO);
        dstShape = (GeneralPath)NativeFontWrapper.getGlyphOutline(
                                        this, font, getGlyphTX(),
                                        frc.isAntiAliased(),
                                        frc.usesFractionalMetrics(),
                                        glyphIndex, dstShape);
        return dstShape;
    }

    /**
    *   Return the position of the specified glyph within the GlyphVector.
    *   This position corresponds to the leading edge of the baseline for
    *   the glyph.
    */
    public Point2D getGlyphPosition(int glyphIndex)
    {
      //        if (glyphIndex < 0 || glyphIndex > glyphs.length) { // note == glyphs.length is ok
      //            throw new IndexOutOfBoundsException("glyphIndex = " + glyphIndex);
      //        }

        initPositions();
        return new Point2D.Float(   positions[glyphIndex * 2],
                                    positions[glyphIndex * 2 + 1]);
    }

    /**
    *   Set the position of the specified glyph within the GlyphVector.
    *   This position corresponds to the leading edge of the baseline for
    *   the glyph.
    */
    public void setGlyphPosition(int glyphIndex, Point2D position)
    {
      //      if (glyphIndex < 0 || glyphIndex > glyphs.length) { // note == glyphs.length is ok
      //              throw new IndexOutOfBoundsException("glyphIndex = " + glyphIndex);
      //        }

        initPositions();
        positions[glyphIndex * 2] =     (float)position.getX();
        positions[glyphIndex * 2 + 1] = (float)position.getY();
    }

    /**
    *   Ensure that positions vector is initialized
    */
    private void initPositions()
    {
        if (positions == null) {
            performDefaultLayout();
        }
    }

    /**
    *   Return an array of glyph positions for the the specified Glyphs.
    *   The position of each glyph corresponds to the leading edge of the
    *   baseline for that glyph.
    *   Use this method for convenience and performance in processing of glyph
    *   positions.
    *   If no array is passed in, one will be created for you.
    *   Even numbered array entries beginning with position zero are the X
    *   coordinates of the glyph numbered beginGlyphIndex + position/2
    *   Odd numbered array entries beginning with position one are the Y
    *   coordinates of the glyph numbered beginGlyphIndex + (position-1)/2
    */
    public float[] getGlyphPositions(int beginGlyphIndex, int numEntries,
                                              float[] positionReturn)
    {
        if (numEntries < 0) {
            throw new IllegalArgumentException("numEntries = " + numEntries);
        }
        if (beginGlyphIndex < 0) {
            throw new IndexOutOfBoundsException("beginGlyphIndex = " + beginGlyphIndex);
        }
        if (beginGlyphIndex > glyphs.length + 1 - numEntries) { // watch for overflow
            throw new IndexOutOfBoundsException("beginGlyphIndex + numEntries = " + (beginGlyphIndex + numEntries));
        }

        return internalGetGlyphPositions(beginGlyphIndex, numEntries, positionReturn);
    }

    /** 
     * Return the positions of all the glyphs, plus the position at the end
     * of the glyph vector.  The length of the returned array is 
     * <code>getNumGlyphs() * 2 + 2</code>.
     */
    public float[] getGlyphPositions(float[] positionReturn) {
        return internalGetGlyphPositions(0, glyphs.length + 1, positionReturn);
    }

    float[] internalGetGlyphPositions(int beginGlyphIndex, int numEntries, float[] positionReturn) {
        float[] retval = positionReturn;
        if (retval == null) {
            retval = new float[numEntries * 2];
        }

        initPositions();
        for (int i=0; i < numEntries; i++) {
            retval[i*2] = positions[ (i + beginGlyphIndex) * 2];
            retval[i*2+1] = positions[ (i + beginGlyphIndex) * 2 + 1];
        }

        return retval;
    }

    /**
    *   Set some or all of the positions of each glyph.
    *   The position of each glyph corresponds to the leading edge of the
    *   baseline for that glyph.
    *   Use this method for convenience and performance in processing of glyph
    *   positions.
    *   Even numbered array entries beginning with position zero are the X
    *   coordinates of the glyph numbered beginGlyphIndex + position/2
    *   Odd numbered array entries beginning with position one are the Y
    *   coordinates of the glyph numbered beginGlyphIndex + (position-1)/2
    */
    public void setGlyphPositions(float[] srcPositions, int srcStart, 
                                  int beginGlyphIndex, int numEntries)
    {
        if (numEntries < 0) {
	  throw new IllegalArgumentException("numEntries = " + numEntries);
        }

        initPositions();
        System.arraycopy(srcPositions, srcStart, positions, beginGlyphIndex * 2, numEntries * 2);
    }

    /**
     * Set the positions of the glyphvector to those in the provided array.
     * The array must contain exactly <code>getNumGlyphs() * 2 + 2</code> values.
     */
    public void setGlyphPositions(float[] srcPositions) {
        int requiredLength = glyphs.length * 2 + 2;
        if (srcPositions.length != requiredLength) {
            throw new IllegalArgumentException("srcPositions.length != " + requiredLength);
        }

        if (positions == null) {
            positions = (float[])srcPositions.clone();
        } else {
            System.arraycopy(srcPositions, 0, positions, 0, requiredLength);
        }
    }

    /**
    *   Return the logical bounds of the specified Glyph within the GlyphVector.
    *   These logical bounds have a total of four edges, with two edges
    *   parallel to the baseline under the glyph's transform and the other two
    *   edges are shared with adjacent glyphs if they are present. Useful for
    *   hit-testing of the specified glyph, positioning of a caret at the
    *   leading or trailing edge of a glyph, and for drawing a highlight
    *   region around the specified glyph.
    *   @see #getGlyphVisualBounds
    */
    public Shape getGlyphLogicalBounds(int glyphIndex)
    {
        if (glyphIndex < 0 || glyphIndex >= glyphs.length) {
          throw new IndexOutOfBoundsException("glyphIndex = " + glyphIndex);
        }

        float [] metrics = new float[4];
        NativeFontWrapper.getFontMetrics(font, getGlyphTX(),
                                        frc.isAntiAliased(),
                                        frc.usesFractionalMetrics(),
                                        metrics);
        initPositions();
        
        float minX = positions[glyphIndex * 2];
        float minY = positions[glyphIndex * 2 + 1] - metrics[0];
        float maxX = minX;
        float maxY = minY + metrics[1] + metrics[2];

        GlyphMetrics gm = getGlyphMetrics(glyphIndex);
        maxX += gm.getAdvance();

        // !!! don't try to expand bounds to share edges
        // would need to deal with positioned accent marks, etc.
        // since italic carets depend on line height and adjacent
        // glyphvectors, this api is not all that useful

        return new Rectangle2D.Float(minX, minY, maxX - minX, maxY - minY);
    }

    /**
    *   Return the visual bounds of the specified Glyph within the GlyphVector.
    *   These visual bounds have a total of four edges, representing the
    *   tightest polygon enclosing non-background pixels in the rendered
    *   representation of the Glyph whose edges are parallel to the edges
    *   of the logical bounds.  Useful for hit-testing of the specified glyph.
    *   @see #getGlyphLogicalBounds
    */
    public Shape getGlyphVisualBounds(int glyphIndex)
    {
        if (glyphIndex < 0 || glyphIndex >= glyphs.length) {
            throw new IndexOutOfBoundsException("glyphIndex = " + glyphIndex);
        }

        initPositions();
        Shape theShape = getGlyphOutline(glyphIndex);
        return (Shape)theShape.getBounds();
    }

    public GlyphMetrics getGlyphMetrics(int glyphIndex)
    {
        if (glyphIndex < 0 || glyphIndex >= glyphs.length) {
            throw new IndexOutOfBoundsException("glyphIndex = " + glyphIndex);
        }

        GlyphMetrics gm = new GlyphMetrics(0,
                                           new Rectangle2D.Float(),
                                           GlyphMetrics.STANDARD);

        NativeFontWrapper.getGlyphMetrics(
                            font,
                            glyphs[glyphIndex], getGlyphTX(),
                            frc.isAntiAliased(),
                            frc.usesFractionalMetrics(),
                            gm);
        return gm;
    }

    /**
    *   Get the transform of the specified glyph within the GlyphVector.
    */
    public AffineTransform getGlyphTransform(int glyphIndex) {
        if (glyphIndex < 0 || glyphIndex >= glyphs.length) {
            throw new IndexOutOfBoundsException("glyphIndex = " + glyphIndex);
        }

        // REMIND: finish me
        return new AffineTransform();
    }

    /**
    *   Set the transform of the specified glyph within the GlyphVector.
    *   This can be used to rotate, mirror, translate and scale the
    *   glyph.  Adding a transform can result in signifant performance changes
    */
    public void setGlyphTransform(int glyphIndex, AffineTransform newTX) {
        if (glyphIndex < 0 || glyphIndex >= glyphs.length) {
            throw new IndexOutOfBoundsException("glyphIndex = " + glyphIndex);
        }

        // REMIND: finish me
    }


    public Object clone() {
        // positions is mutable so we have to clone it.
        try {
            StandardGlyphVector result = (StandardGlyphVector)super.clone();
            if (positions != null) {
              result.positions = (float[])positions.clone();
            }
	    return result;
        }
        catch (CloneNotSupportedException e) {
        }

        return this;
    }

  /**
   * For each glyph return posx, posy, advx, advy, visx, visy, visw, vish.
   */
    public float[] getGlyphInfo() {
        float[] result = new float[glyphs.length * 8];

	double[] matrix = getGlyphTX();

	NativeFontWrapper.getGlyphInfo(font, glyphs, positions, glyphs.length, 
				       matrix, frc.usesFractionalMetrics(), frc.isAntiAliased(), 
				       result);
	return result;
    }

    public GlyphJustificationInfo getGlyphJustificationInfo(int glyphIndex) {
        if (glyphIndex < 0 || glyphIndex >= glyphs.length) {
            throw new IndexOutOfBoundsException("glyphIndex = " + glyphIndex);
        }

        // REMIND: finish me
        return null;
    }

    // general utility methods

    /**
     * Hashcode
     */
    public int hashCode() {
      return font.hashCode() ^ glyphs.length;
    }

    /**
     * Test for complete equality.
     */
    public boolean equals(GlyphVector set) {
      if (this == set) {
	return true;
      }
      if (set == null) {
	return false;
      }
      try {
	StandardGlyphVector other = (StandardGlyphVector)set;

	if (glyphs.length != other.glyphs.length) {
	  return false;
	}

	for (int i = 0; i < glyphs.length; ++i) {
	  if (glyphs[i] != other.glyphs[i]) {
	    return false;
	  }
	}

	if (!font.equals(other.font)) {
	  return false;
	}

	if (!frc.equals(other.frc)) {
	  return false;
	}

	if ((other.positions == null) != (positions == null)) {
	  if (positions == null) {
	    performDefaultLayout();
	  } else {
	    other.performDefaultLayout();
	  }
	}

	if (positions != null) {
	  for (int i = 0; i < positions.length; ++i) {
	    if (positions[i] != other.positions[i]) {
	      return false;
	    }
	  }
	}
      }
      catch (ClassCastException e) {
	// assume they are different simply by virtue of the class difference

	return false;
      }

      return true;
    }
}
