/*
 * @(#)CompositionAreaHandler.java	1.6 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.im;

import java.awt.Component;
import java.awt.Rectangle;
import java.awt.event.InputMethodEvent;
import java.awt.event.InputMethodListener;
import java.awt.font.TextAttribute;
import java.awt.font.TextHitInfo;
import java.awt.im.InputMethodRequests;
import java.text.AttributedCharacterIterator;
import java.text.AttributedCharacterIterator.Attribute;
import java.text.AttributedString;

/**
 * A composition area handler handles events and input method requests for
 * the composition area. Typically each input method context has its own
 * composition area handler if it supports passive clients or below-the-spot
 * input, but all handlers share a single composition area.
 *
 * @version 1.6 02/02/00
 * @author JavaSoft International
 */

class CompositionAreaHandler implements InputMethodListener,
						 InputMethodRequests {

    private static CompositionArea compositionArea;
    private static CompositionAreaHandler compositionAreaOwner; // synchronized through compositionArea

    private AttributedCharacterIterator composedText;
    private TextHitInfo caret = null;
    private Component clientComponent = null;
    private InputMethodContext inputMethodContext;

    /**
     * Constructs the composition area handler, and also creates a composition
     * area if this is the first composition area handler.
     */
    CompositionAreaHandler(InputMethodContext context) {
    
        synchronized (CompositionAreaHandler.class) {
            if (compositionArea == null) {
                compositionArea = new CompositionArea();
            }
        }

        inputMethodContext = context;
    }

    void setClientComponent(Component clientComponent) {
        this.clientComponent = clientComponent;
    }

    /**
     * Grabs the composition area, makes this handler its owner, and installs
     * the handler and its input context into the composition area for event
     * and input method request handling.
     * If doUpdate is true, updates the composition area with previously sent
     * composed text.
     */
    void grabCompositionArea(boolean doUpdate) {
        synchronized (compositionArea) {
            if (compositionAreaOwner != this) {
                compositionAreaOwner = this;
                compositionArea.setHandlerInfo(this, inputMethodContext);
                if (doUpdate) {
                    compositionArea.setText(composedText, caret);
                }
            }
        }
    }

    /**
     * Releases and closes the composition area if it is currently owned by
     * this composition area handler.
     */
    void releaseCompositionArea() {
        synchronized (compositionArea) {
            if (compositionAreaOwner == this) {
                compositionAreaOwner = null;
                compositionArea.setHandlerInfo(null, null);
                compositionArea.setText(null, null);
            }
        }
    }

    /**
     * Releases and closes the composition area if it has been created,
     * independent of the current owner.
     */
    static void closeCompositionArea() {
        if (compositionArea != null) {
            synchronized (compositionArea) {
                compositionAreaOwner = null;
                compositionArea.setHandlerInfo(null, null);
                compositionArea.setText(null, null);
            }
        }
    }

    /**
     * Returns whether the composition area is currently visible
     */
    boolean isCompositionAreaVisible() {
        // use parent's isVisible since that's the one that we toggle
        // when showing or hiding the input window
        return compositionArea.getParent().isVisible();
    }

    void processInputMethodEvent(InputMethodEvent event) {
        if (event.getID() == InputMethodEvent.INPUT_METHOD_TEXT_CHANGED) {
            inputMethodTextChanged(event);
        } else {
            caretPositionChanged(event);
        }
    }

    //
    // InputMethodListener methods
    //

    private static final Attribute[] IM_ATTRIBUTES =
            { TextAttribute.INPUT_METHOD_HIGHLIGHT };

    public void inputMethodTextChanged(InputMethodEvent event) {
        AttributedCharacterIterator text = event.getText();
        int committedCharacterCount = event.getCommittedCharacterCount();

        // extract composed text and prepare it for display
        composedText = null;
        caret = null;
        if (text != null
                && committedCharacterCount < text.getEndIndex() - text.getBeginIndex()) {
            // copy the composed text
            AttributedString composedTextString;
            composedTextString = new AttributedString(text,
                    text.getBeginIndex() + committedCharacterCount, // skip over committed text
                    text.getEndIndex(), IM_ATTRIBUTES);
            composedTextString.addAttribute(TextAttribute.FONT, compositionArea.getFont());
            composedText = composedTextString.getIterator();
	    caret = event.getCaret();
	}
	compositionArea.setText(composedText, caret);

        // send any committed text to the text component
        if (committedCharacterCount > 0) {
            inputMethodContext.dispatchCommittedText(((Component) event.getSource()),
						     text, committedCharacterCount);

	    // this may have changed the text location, so reposition the window
	    if (isCompositionAreaVisible()) {
	        compositionArea.updateWindowLocation();
	    }
        }
        
        // event has been handled, so consume it
        event.consume();
    }

    public void caretPositionChanged(InputMethodEvent event) {
        compositionArea.setCaret(event.getCaret());
        
        // event has been handled, so consume it
        event.consume();
    }

    //
    // InputMethodRequests methods
    //

    /**
     * Returns the input method request handler of the client component.
     * When using the composition window for an active client (below-the-spot
     * input), input method requests that do not relate to the display of
     * the composed text are forwarded to the client component.
     */
    InputMethodRequests getClientInputMethodRequests() {
        return (InputMethodRequests) clientComponent.getInputMethodRequests();
    }

    public Rectangle getTextLocation(TextHitInfo offset) {
        synchronized (compositionArea) {
            if (compositionAreaOwner == this && isCompositionAreaVisible()) {
                return compositionArea.getTextLocation(offset);
            } else if (composedText != null) {
                // there's composed text, but it's not displayed, so fake a rectangle
                return new Rectangle(0, 0, 0, 10);
            } else {
                InputMethodRequests requests = getClientInputMethodRequests();
                if (requests != null) {
                    return requests.getTextLocation(offset);
                } else {
                    // passive client, no composed text, so fake a rectangle
                    return new Rectangle(0, 0, 0, 10);
                }
            }
        }
    }

    public TextHitInfo getLocationOffset(int x, int y) {
        synchronized (compositionArea) {
            if (compositionAreaOwner == this && isCompositionAreaVisible()) {
                return compositionArea.getLocationOffset(x, y);
            } else {
                return null;
            }
        }
    }

    public int getInsertPositionOffset() {
        InputMethodRequests req = getClientInputMethodRequests();
        if (req != null) {
            return req.getInsertPositionOffset();
	}

        // we don't have access to the client component's text.
	return 0;
    }

    private static final AttributedCharacterIterator EMPTY_TEXT =
            (new AttributedString("")).getIterator();

    public AttributedCharacterIterator getCommittedText(int beginIndex,
						       int endIndex,
						       Attribute[] attributes) {
        InputMethodRequests req = getClientInputMethodRequests();
        if(req != null) {
            return req.getCommittedText(beginIndex, endIndex, attributes);
	}

        // we don't have access to the client component's text.
	return EMPTY_TEXT;
    }

    public int getCommittedTextLength() {
        InputMethodRequests req = getClientInputMethodRequests();
        if(req != null) {
            return req.getCommittedTextLength();
	}

        // we don't have access to the client component's text.
	return 0;
    }


    public AttributedCharacterIterator cancelLatestCommittedText(Attribute[] attributes) {
        InputMethodRequests req = getClientInputMethodRequests();
        if(req != null) {
            return req.cancelLatestCommittedText(attributes);        
	}

        // we don't have access to the client component's text.
        return null;
    }

    public AttributedCharacterIterator getSelectedText(Attribute[] attributes) {
        InputMethodRequests req = getClientInputMethodRequests();
        if(req != null) {
            return req.getSelectedText(attributes);
	}

        // we don't have access to the client component's text.
        return EMPTY_TEXT;
    }

}
