/*
 * @(#)ProbeRobot.java	1.6 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.robot;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;
import java.lang.reflect.InvocationTargetException;
import sun.awt.*;
import sun.awt.robot.probe.*;
import java.io.*;

/**
 * Alternative class interface from Robot used to generate native events for test automation.
 *
 *<p> The scope of theRobot API is to cover automated testing of GUI
 * behavior. The infrastructure within the Robot provides these
 * capabilities:
 * <ul>
 *   <li> Creation of mouse and keyboard gestures in a
 * platform-native manner. This provides the greatest assurance that
 * Java will see these gestures in the same manner as if there were a
 * real person driving the mouse and keyboard.  It also provides the
 * greatest coverage of the platform specific code paths that handle
 * the platform-native events prior to dispatching them into the Java
 * code.
 *   <li> Events are generated into a buffer that allows specifiable
 * delays between event dispatches. This provides the ability to mimic
 * different speeds of interaction, testing response time of the GUI, etc.
 *   <li> A general system for recording activity from several important
 * sources within Java. This is meant to be the basis for verifying
 * proper behavior. With a given set of events against a given set
 * of GUI components, a  certain set of events should be generated.
 * </ul><p>
 *
 * ProbeRobot adds these features to the Robot environment
 * <ul>
 *
 *  <li><b>Extensive activity monitoring and reporting capabilities</b>:
 *     Uses the Probe classes to
 *     hook into activity within the Java environment.  One part attaches to
 *     the AWT Event Queue and another attaches to all the listener interfaces
 *     (using Reflection to find them) in a tree of Components.  All Probe
 *     information is directed to a centralized listening and reporting interface
 *     that can direct Probe information to multiple listeners.  Two listeners
 *     are provided, one that records everything into a buffer, and the other
 *     that simply prints everything it receives.<p>
 *
 *     This greatly simplifies gathering
 *     information about activity that happened within the application.  With this
 *     mechanism the programmer need do little instrumenting of the application
 *     because, at least for GUI testing, almost everything is known via the firing
 *     of listener interfaces.  Using the <i>wrap</i> and <i>unwrap</i> methods
 *     it is trivially easy to gather detailed tracings of what happened within
 *     the application.
 *
 *  <li><b>Extensive list of basic gesture methods</b>:
 *     In addition to the basic <i>move</i> and <i>key/button press/release</i>
 *     gestures, are also some obvious higher level abstractions such
 *     as <i>click</i> and <i>type</i>.<p>
 *
 *     Having these higher level gestures already coded and tested simplifies
 *     the task of making the gestures you want.  For example, a drag&drop sequence
 *     with this set of gesture methods is simply <i>move</i> to the source,
 *     <i>buttonPress</i> to initiate the drag, <i>glide</i> to move the mouse
 *     stepwise to the destination (ensuring the drag is recognized), and
 *     a <i>buttonRelease</i> to conclude the drag.
 *
 *  <li><b>Specifiable delays to give the opportunity to mimic different
 *  styles of user behavior, from fast to slow</b>:
 *     The delays are specified in many methods, with alternative methods
 *     that use a standardised delay.
 *
 *   <li><b>A simplistic command parser, letting one specify Robot actions
 *   in Strings</b>:
 *     This is the basis for a hokey scripting language.  It is not anything
 *     great or wonderful, but it is simple and might prove useful.  What would
 *     be more useful, if people want a scripting language, is to use Jacl
 *     as the basis for scripting.  (Jacl is a version of
 *     <a href="http://www.scriptics.com">TCL</a> that works inside
 *     Java interpreters.)
 * </ul><p>
 *
 * There are two jewels in ProbeRobot: the extensive list of gestures, and the
 * Probe mechanism.
 *
 * @version 1.6, 02/02/00
 * @author David Herron
 *
 */
public class ProbeRobot {
    // can't imagine why you'd want to wait any longer...
    private static final int MAX_DELAY = 5000;

    private Robot	robot;
    private RobotPeer	peer;
    private Random      generator;

  // These are public (but undocumented) so that
  // interested developers can use them without
  // tremendous gymnastics.
    public ProbeMonitor monitor;
    public EventProbe   evProbe;
    public ProbeBuffer  buffer;
    public WrapComponent wrapper;
    public ProbePrinter printer;

    /**
     * Initialize a <i>ProbeRobot</i>, an alternative API for using <i>Robot</i>.
     * Note that the Robot must already be initialized.
     * @param robot The Robot that has already been created.
     */
    public ProbeRobot(Robot robot) {
        this.robot = robot;
	peer     = robot.getPeer();
	//	toolkit  = robotToolkit;
	generator = new Random();
	// In case the peer cannot be built ...
	if (robot != null) {
	  monitor  = new ProbeMonitor();
	  monitor.dispatch = true;
	  evProbe  = new EventProbe(monitor);
	  buffer   = new ProbeBuffer(monitor);
	  printer  = new ProbePrinter(monitor, false);
	  wrapper  = new WrapComponent(monitor);
	}
	else {
	  monitor  = null;
	  evProbe  = null;
	  buffer   = null;
	  printer  = null;
	  wrapper  = null;
	}
    }

    /**
     * Access the RobotPeer interface (might be useful).
     */
    public RobotPeer getPeer() { return peer; }

    /**
     * Parses the String for a command and executes that command.
     * This allows us to have a simple scripting facility.<p>
     *
     * The commands are:
     * <table>
     * <tr><th>Command</th><th>Description</th></tr>
     * <tr><th><tt>move <i>x y</i></tt></th>
     *     <td>Move mouse to root coordinate.</td>
     *     </tr>
     * <tr><th><tt>print</tt></th><td>Turn on printing.</td></tr>
     * <tr><th><tt>noprint</tt></th><td>Turn off printing.</td></tr>
     * <tr><th><tt>setspeed <i>n</i></tt></th>
     *     <td>Change the default speed of event dispatching.</td></tr>
     * <tr><th><tt>setwait <i>n</i></tt></th>
     *     <td>Change the default speed of event dispatching.</td></tr>
     * <tr><th><tt>bpress <i>n</i></tt></th>
     *     <td>Cause the mouse button to be pressed.</td></tr>
     * <tr><th><tt>brelease <i>n</i></tt></th>
     *     <td>Cause the mouse button to be released.</td></tr>
     * <tr><th><tt>click <i>n</i></tt></th>
     *     <td>Cause the mouse button to be clicked.</td></tr>
     * <tr><th><tt>glide <i>x1 y1 x2 y2 n</i></tt></th>
     *     <td>Cause the mouse to move for <i>n</i> steps from 
     *         the root coordinate (x1,y1) to (x2,y2).</td></tr>
     * </table>
     *
     * @param command The Command to execute.
     * @see processInput
     */
    public void doCommand(String command) {
      StringTokenizer t = new StringTokenizer(command);
      String verb;

      try {
	verb = t.nextToken();
      } catch (Exception e) {
	System.err.println("Could not get command in " + command);
	System.err.println(e);
	return;
      }

      if (verb.equals("move")) {
	try {
	  Integer x = new Integer(t.nextToken());
	  Integer y = new Integer(t.nextToken());
	  mouseMove(x.intValue(), y.intValue());
	} catch (Exception e) {
	  System.err.println("USAGE: move x y");
	  System.err.println(e);
	}
	return;
      }
      if (verb.equals("print")) {
	try {
	  printerStart();
	} catch (Exception e) {
	  System.err.println("USAGE: print");
	  System.err.println(e);
	}
	return;
      }
      if (verb.equals("noprint")) {
	try {
	  printerStop();
	} catch (Exception e) {
	  System.err.println("USAGE: noprint");
	  System.err.println(e);
	}
	return;
      }
      if (verb.equals("setspeed")) {
	try {
	  defaultSpeed = new Integer(t.nextToken()).intValue();
	} catch (Exception e) {
	  System.err.println("USAGE: setspeed newspeed");
	  System.err.println(e);
	}
	return;
      }
      if (verb.equals("setwait")) {
	try {
	  defaultWait = new Integer(t.nextToken()).intValue();
	} catch (Exception e) {
	  System.err.println("USAGE: setwait newwait");
	  System.err.println(e);
	}
	return;
      }
      if (verb.equals("bpress")) {
	try {
	  Integer b = new Integer(t.nextToken());
	  buttonPress(b.intValue());
	} catch (Exception e) {
	  System.err.println("USAGE: bpress button");
	  System.err.println(e);
	}
	return;
      }
      if (verb.equals("brelease")) {
	try {
	  Integer b = new Integer(t.nextToken());
	  buttonRelease(b.intValue());
	} catch (Exception e) {
	  System.err.println("USAGE: brelease button");
	  System.err.println(e);
	}
	return;
      }
      if (verb.equals("click")) {
	try {
	  Integer b = new Integer(t.nextToken());
	  click((short)b.intValue());
	} catch (Exception e) {
	  System.err.println("USAGE: click button");
	  System.err.println(e);
	}
	return;
      }
      if (verb.equals("glide")) {
	try {
	  Integer xSrc   = new Integer(t.nextToken());
	  Integer ySrc   = new Integer(t.nextToken());
	  Integer xDest  = new Integer(t.nextToken());
	  Integer yDest  = new Integer(t.nextToken());
	  Integer nSteps = new Integer(t.nextToken());
	  glide( new Point(xSrc.intValue(), ySrc.intValue()),
		 new Point(xDest.intValue(), yDest.intValue()),
		 (short)nSteps.intValue(),
		 defaultSpeed);
	} catch (Exception e) {
	  System.err.println("USAGE: glide x1 y1 x2 y2 nSteps");
	  System.err.println(e);
	}
	return;
      }
    }

    /**
     * Dispatch all lines in the input stream to <i>doCommand</i>.
     *
     * @param input The input stream to process.
     */
    public void processInput(InputStream input) {
      BufferedReader in = new BufferedReader(new InputStreamReader(input));

      String cmd = null;

      try {
	while ((cmd = in.readLine()) != null) {
	  doCommand(cmd);
	}
      }
      catch (Exception e) {
	System.err.println("Error in command loop ... exiting ");
	e.printStackTrace();
      }
    }

    ///////////////////////////
    // Manipulate EventProbe //
    ///////////////////////////

    /**
     * Stop tracing all AWT events through probe mechanism.
     */
    public void awtEventsQuiet() {
      if (evProbe != null) evProbe.disconnect();
    }

    /**
     * Trace all AWT events through probe mechanism.
     * By default AWT event tracing is turned on.
     */
    public void awtEventsVerbose() {
      if (evProbe != null) evProbe.connect();
    }

    /////////////////////////////
    // Manipulate ProbeMonitor //
    /////////////////////////////

    /**
     * Post a ProbeObject to the probe monitoring system.
     * This method is used to post other probes than the ones
     * already built in to Robot.
     */
    public void monitorPost(ProbeObject evt) {
      if (monitor != null) monitor.post(evt);
    }

    ////////////////////////////
    // Manipulate ProbeBuffer //
    ////////////////////////////

    /**
     * Have the ProbeBuffer stop recording.  By default the recorder
     * is running, and we might want to turn it off for some reason.
     */
    public void recorderStop() {
      if (buffer != null) buffer.stopRecording();
    }

    /**
     * Have the ProbeBuffer start recording.
     */
    public void recorderStart() {
      if (buffer != null) buffer.startRecording();
    }

    /**
     * Empty the ProbeBuffer's buffer.
     */
    public void recorderClear() {
      if (buffer != null) buffer.clear();
    }


    /**
     * Get the Iterator for the recorder.
     */
    public Iterator recorderIterator() {
      if (buffer != null)
	return buffer.listIterator();
      else
	return null;
    }

    /**
     * Return the number of events in the buffer.
     */
    public int recorderSize() {
      if (buffer != null)
	return buffer.size();
      else
	return -1;
    }

    /**
     * Return a specific element in the buffer.
     */
    public ProbeObject recorderElement(int index) {
      if (buffer != null)
	return buffer.elementAt(index);
      else
	return null;
    }

    ///////////////////////////
    // WrapComponent support //
    ///////////////////////////

    /**
     * Connect all the listeners of <i>c</i> to send events into
     * the recorder.
     */
    public void wrap(Component c) {
      if (wrapper != null) wrapper.wrap(c);
    }

    /**
     * Connect all the listeners of <i>c</i> to send events into
     * the recorder.  As a Container, this also descends the tree
     * of all child Components, and connects to all their listeners
     * as well.
     */
    public void wrap(Container c) {
      if (wrapper != null) wrapper.wrap(c);
    }

    /**
     * Disconnect all the listeners of <i>c</i> from sending events into
     * the recorder.
     */
    public void unwrap(Component c) {
      if (wrapper != null) wrapper.unwrap(c);
    }

    /**
     * Disconnect all the listeners of <i>c</i> from send events into
     * the recorder.  As a Container, this also descends the tree
     * of all child Components, and disconnects all their listeners
     * as well.
     */
    public void unwrap(Container c) {
      if (wrapper != null) wrapper.unwrap(c);
    }

    ///////////////////////////////
    // ProbePrinter manipulation //
    ///////////////////////////////

    /**
     * Start the ProbePrinter printing things it receives.
     *
     * By default the printer is turned off.  The information received
     * by the printer is tremendous and at the same time is useful
     * for tracing activity of the application.
     */
    public void printerStart() {
      if (printer != null) printer.startPrinting();
    }

    /**
     * Stop the ProbePrinter from printing things it receives.
     */
    public void printerStop() {
      if (printer != null) printer.stopPrinting();
    }

    /**
     * Change the output for the printer.
     */
    public void printerOutput(PrintStream out) {
      printer.setOutput(out);
    }

    //////////////////////////////////////
    // Basic Gesture generation methods //
    //////////////////////////////////////


    /**
     * Default amount of delay before event sequence is played.
     * The 'wait' delay is the time specified on the first
     * event of a sequence of events.
     */
    public int   defaultWait   = 0;

    /**
     * Default amount of delay before events within an event sequence.
     */
    public int   defaultSpeed  = 20;

    /**
     * The default mouse button to press.
     */
    public short defaultButton = 1;

    /*
     * Give the MouseEvent mask for the given button#.
     */
    private int mouseButton(short button) {
      switch (button) {
      case 1: return MouseEvent.BUTTON1_MASK;
      case 2: return MouseEvent.BUTTON2_MASK;
      case 3: return MouseEvent.BUTTON3_MASK;
      }
      return 0;
    }

    private int mouseButton(int button) { return mouseButton((short)button); }

    /**
     * Calculate a Point that is the absolute coordinate of
     * the center of the Component.<p>
     *
     * <b>WARNING</b>: Do not call this method in the middle of event
     * synthesis of Drag & Drop gestures.  This method calls
     * Component.getLocationOnScreen which may cause a deadlock
     * situation with AWT native code.
     *
     * @param target Component to calculate against.
     */
    public Point targetCenter(Component target) {
      int w = target.getWidth();
      int h = target.getHeight();
      Point loc = target.getLocationOnScreen();
      loc.x += (w/2);
      loc.y += (h/2);
      return loc;
    }

    /**
     * Calculate a Point that is the absolute coordinate of
     * the upper-left corner of the Component.<p>
     *
     * <b>WARNING</b>: Do not call this method in the middle of event
     * synthesis of Drag & Drop gestures.  This method calls
     * Component.getLocationOnScreen which may cause a deadlock
     * situation with AWT native code.
     *
     * @param target Component to calculate against.
     */
    public Point targetOrigin(Component target) {
      return target.getLocationOnScreen();
    }

    /**
     * Generate a mouse button Press event,
     * specifying the mouse button and the delay.
     *
     * @param delay The time to wait before posting the event.
     * @param button The mouse button to press
     */
    public void buttonPress(int delay, short button) {
      robot.mousePress(mouseButton(button));
      pause(delay);
    }

    /**
     * Generate a mouse button Press event, specifying the mouse
     * button and with a default delay.
     *
     * @param button The mouse button to press */
    public void buttonPress(short button) {
      buttonPress(defaultWait, button);
    }

    /**
     * Generate a mouse button Press event,
     * specifying the mouse button 1 and the delay.
     *
     * @param delay The time to wait before posting the event.
     */
    public void buttonPress(int delay) {
      buttonPress(delay, defaultButton);
    }

    /**
     * Generate a mouse button Press event,
     * specifying the mouse button 1 and with a default delay.
     */
    public void buttonPress() {
      buttonPress(defaultWait, defaultButton);
    }

    /**
     * Presses one or more mouse buttons (does not override
     * any previous mousePress calls).  Included for API compatibility
     * with sun.awt.robot.Robot.
     * @param buttons	Button mask (combination of InputEvent.BUTTON_MASK1/2/3)
     */
    public void mousePress(long buttons) {
      robot.mousePress(buttons);
      pause(defaultWait);
    }

    /**
     * Generate a mouse button Release event,
     * specifying the mouse button and the delay.
     *
     * @param delay The time to wait before posting the event.
     * @param button The mouse button to release
     */
    public void buttonRelease(int delay, short button) {
      robot.mouseRelease(mouseButton(button));
      pause(delay);
    }

    /**
     * Generate a mouse button Release event,
     * specifying the mouse button and with a default delay.
     *
     * @param button The mouse button to release
     */
    public void buttonRelease(short button) {
      buttonRelease(defaultWait, button);
    }

    /**
     * Generate a mouse button Release event,
     * specifying the mouse button 1 and the delay.
     *
     * @param delay The time to wait before posting the event.
     */
    public void buttonRelease(int delay) {
      buttonRelease(delay, defaultButton);
    }

    /**
     * Generate a mouse button Release event,
     * specifying the mouse button 1 and with a default delay.
     */
    public void buttonRelease() {
      buttonRelease(defaultWait, defaultButton);
    }

    /**
     * Releases one or more mouse buttons (does not override
     * any previous mouseRelease calls).  Included for API compatibility
     * with sun.awt.robot.Robot.
     * @param buttons	Button mask (combination of InputEvent.BUTTON_MASK1/2/3)
     */
    public void mouseRelease(long buttons) {
      robot.mouseRelease(buttons);
      pause(defaultWait);
    }

    /**
     * Generate a mouse-move event
     * to the given absolute screen location.  (no delay)
     *
     * @param loc The screen location to move to.
     */
    public void mouseMove(Point loc) {
      mouseMove(defaultWait, loc.x, loc.y);
    }

    /**
     * Generate a mouse-move event to the given
     * absolute screen location.  (delay in milliseconds)
     *
     * @param delay The time to wait before posting the event.
     * @param loc The screen location to move to.
     */
    public void mouseMove(int delay, Point loc) {
      mouseMove(delay, loc.x, loc.y);
    }

    /**
     * Generate a mouse-move event
     * to the given absolute screen location.  (default delay)
     *
     * @param x The screen <i>x</i> coordinate to move to
     * @param y The screen <i>y</i> coordinate to move to
     */
    public void mouseMove(int x, int y) {
      mouseMove(defaultWait, x, y);
    }

    /**
     * Generate a mouse-move event to the given
     * absolute screen location.  (delay in milliseconds)
     *
     * @param delay The time to wait before posting the event.
     * @param x The screen <i>x</i> coordinate to move to
     * @param y The screen <i>y</i> coordinate to move to
     */
    public void mouseMove(int delay, int x, int y) {
      robot.mouseMove(x, y);
      pause(delay);
    }

    /**
     * Generate a keypress event for the given key.
     * Default delay.
     *
     * @param keycode The keycode of the key to press.
     */
    public void keyPress(int keycode) {
      keyPress(defaultWait, keycode);
    }

    /**
     * Generate a keypress event for the given key.
     *
     * @param delay The time to wait before posting the event.
     * @param keycode The keycode of the key to press.
     */
    public void keyPress(int delay, int keycode) {
      robot.keyPress(keycode);
      pause(delay);
    }

    /**
     * Generate a keyrelease event for the given key.
     * Default delay.
     *
     * @param keycode The keycode of the key to release.
     */
    public void keyRelease(int keycode) {
      keyRelease(defaultWait, keycode);
    }

    /**
     * Generate a keyrelease event for the given key.
     *
     * @param delay The time to wait before posting the event.
     * @param keycode The keycode of the key to release.
     */
    public void keyRelease(int delay, int keycode) {
      robot.keyPress(keycode);
      pause(delay);
    }

    /**
     * Wait a specifiable (in milliseconds) time.
     *
     * @param delay The time to wait before posting the event.
     */
    public void pause(int delay) {
      robot.delay(delay);
    }

    /**
     * Wait for the default time.
     */
    public void pause() { pause(defaultWait); }

    /////////////////////////////////
    // Higher level mouse gestures //
    /////////////////////////////////

    /**
     * Aid in generating chorded mouse presses.
     * Some interactions let you press multiple buttons at the same moment.
     * This method lets you specify which of the buttons to press
     * at the given moment.
     * @param b1 Indicate to press mouse button 1.
     * @param b2 Indicate to press mouse button 2.
     * @param b3 Indicate to press mouse button 3.
     */
    public void buttonPresser(boolean b1, boolean b2, boolean b3) {
      long bMask = 0;
      bMask |= b1 ? mouseButton(1) : 0;
      bMask |= b2 ? mouseButton(2) : 0;
      bMask |= b3 ? mouseButton(3) : 0;
      mousePress(bMask);
    }

    /**
     * Aid in generating chorded mouse releases.
     * Some interactions let you release multiple buttons at the same moment.
     * This method lets you specify which of the buttons to release
     * at the given moment.
     * @param b1 Indicate to release mouse button 1.
     * @param b2 Indicate to release mouse button 2.
     * @param b3 Indicate to release mouse button 3.
     */
    public void buttonReleaser(boolean b1, boolean b2, boolean b3) {
      long bMask = 0;
      bMask |= b1 ? mouseButton(1) : 0;
      bMask |= b2 ? mouseButton(2) : 0;
      bMask |= b3 ? mouseButton(3) : 0;
      mouseRelease(bMask);
    }

    /**
     * Click mouse button 1 with the default pause between Press and Release.
     */
    public void click() { click(defaultSpeed, defaultButton); }

    /**
     * Click the given mouse button with the default pause.
     */
    public void click(short button) { click(defaultSpeed, button); }

    /**
     * Click the given mouse button with the given (milliseconds) pause
     * between Press and Release.
     */
    public void click(int speed, short button) {
      buttonPress(defaultWait, button);
      buttonRelease(speed, button);
    }

    /**
     * Click the given mouse button with precise control of delays.
     * @param wait   The number of milliseconds before the buttonPress
     * @param delay  The number of milliseconds between press and release
     * @param button Which button to click.
     */
    public void click(int wait, int delay, short button) {
      buttonPress(wait, button);
      buttonRelease(delay, button);
    }

    /**
     * Move the mouse in multiple steps from where it is
     * now to the destination.<p>
     *
     * Think of this as walking the mouse in nStep substeps of the path
     * between src-->dest.  To do it perfectly, it should calculate the
     * hypotenus(?sp?), the angle, and from that calculate the perfect dx/dy
     * to get there.  The code here contains a good approximation and
     * is gauranteed (through brute force) to get to the right place.
     *
     * @param src The beginning location (root coords) of the move.
     * @param dest The destination location (root coords) of the move.
     * @param nSteps The number of move events generated in the sequence.
     * @param speed The amount of delay per step in the event sequence.
     */
    public void glide(Point src, Point dest, short nSteps, int speed) {
      Point cur;
      double tDx, tDy;
      double dx,  dy;
      double x, y;
      int stepnum;

      tDx = dest.x - src.x;   // Total change in coordinates
      tDy = dest.y - src.y;

      // This may not be the best way, but is the simplest, to
      // compute the path to the destination location.
      // Note calculations are in double's to calculate properly.
      dx = tDx / nSteps;      // The size of each step
      dy = tDy / nSteps;

      // Walk the mouse to the destination one step at a time
      mouseMove(speed, src);
      for (x = src.x, y = src.y, cur = new Point(), stepnum = 0;
	   stepnum < nSteps;
	   stepnum++) {
	x += dx;
	y += dy;
	cur.x = (int)x;  // Make int to store in the Point.
	cur.y = (int)y;
	mouseMove(speed, cur);
      }
      // Ensure the mouse moves to the right destination.
      // The steps may have led the mouse to a slightly wrong place.
      mouseMove(speed, dest);
    }

    /**
     * Do glide with a default speed.
     * @param src The beginning location (root coords) of the move.
     * @param dest The destination location (root coords) of the move.
     * @param nSteps The number of move events generated in the sequence.
     */
    public void glide(Point src, Point dest, short nSteps) {
      glide(src, dest, nSteps, defaultSpeed);
    }

    /**
     * Do glide in 10 steps and default speed.
     * @param src The beginning location (root coords) of the move.
     * @param dest The destination location (root coords) of the move.
     */
    public void glide(Point src, Point dest) {
      glide(src, dest, (short)10, defaultSpeed);
    }

    /**
     * Moves the mouse back and forth between two points a
     * number of times.  This is to emulate moving the mouse
     * back and forth on the screen around a point.  Motion
     * begins at <i>src</i>, then moves between <i>p1</i>
     * and <i>p2</i> for <i>nSteps</i> gestures, and when finished
     * the mouse returns to <i>src</i>.
     *
     * @param src The beginning and ending mouse location of the gesture.
     * @param p1  The first point of the wiggle area.
     * @param p2  The second point of the wiggle area.
     * @param nSteps The number of times to wiggle.
     * @param delay The number of milliseconds between each move.
     */
    public void wiggle(Point src, Point p1, Point p2, int nSteps, int delay) {
      int i = 0;
      mouseMove(delay, src);
      for (; i < nSteps; i++) {
	mouseMove(delay, p1);
	mouseMove(delay, p2);
      }
      mouseMove(delay, src);
    }

    /**
     * Move the mouse along a set of points.  Each entry in the
     * array is taken to have two items in the row: x,y coordinate.
     *
     * @param points Array of x,y coordinates.
     */
    public void walkMouse(int [][] points) {
      int nPoints = points.length;
      int pNum    = 0;
      for (; pNum < nPoints; pNum++) {
	mouseMove(points[pNum][0], points[pNum][1]);
      }
    }

    /**
     * Move the mouse randomly around a set of points.  Each entry in the
     * array is taken to have two items in the row: x,y coordinate.
     *
     * @param points Array of x,y coordinates.
     * @param nSteps The number of mouseMove's to generate.
     */
    public void walkDrunkenMouse(int [][] points, int nSteps) {
      int nPoints = points.length;
      int stepNum    = 0;
      for (; stepNum < nSteps; stepNum++) {
	int p = generator.nextInt(nPoints);
	mouseMove(points[p][0], points[p][1]);
      }
    }

    ///////////////////////
    // Keyboard gestures //
    ///////////////////////
  
    /**
     * Type (press&release) the keycode.  Generates events
     * for the sequence <i>Press(keycode)</i>..<i>Release(keycode)</i>.
     */
    public void type(int kc) {
      keyPress(defaultWait, kc);
      keyRelease(defaultSpeed, kc);
    }

    /**
     * Type (press&release) the keycode.  Generates events
     * for the sequence <i>Press(keycode)</i>..<i>Release(keycode)</i>.
     *
     * <i>speed</i> is the delay, in milliseconds, between Press and Release.
     */
    public void type(int speed, int kc) {
      keyPress(defaultWait, kc);
      keyRelease(speed, kc);
    }

    /**
     * Type (press&release) the keycode.  Generates events
     * for the sequence <i>Press(keycode)</i>..<i>Release(keycode)</i>.
     *
     * <i>wait</i> is the delay, in milliseconds, before the Press.
     * <i>speed</i> is the delay, in milliseconds, between Press and Release.
     */
    public void type(int wait, int speed, int kc) {
      keyPress(wait, kc);
      keyRelease(speed, kc);
    }

    /**
     * Generate keyboard events specified by an array of information.<p>
     *
     * The intended use of this is something like this:<p>
     *
     * <pre><code>
     *   robot.type(new int[][3] {
     *     { 10, 20, KeyEvent.VK_H },
     *     { 10, 20, KeyEvent.VK_I },
     *     { 10, 20, KeyEvent.VK_SPACE },
     *     { 10, 20, KeyEvent.VK_T },
     *     { 10, 20, KeyEvent.VK_H },
     *     { 10, 20, KeyEvent.VK_E },
     *     { 10, 20, KeyEvent.VK_R },
     *     { 10, 20, KeyEvent.VK_E }
     *    });
     * </code></pre>
     *
     * @param events The list of key codes to type.
     */
    public void type(int [][] events) {
      int nEvents = events.length;
      int evNum = 0;
      
      for (evNum = 0; evNum < nEvents; evNum++) {
	type(events[evNum][0], events[evNum][1], events[evNum][2]);
      }
    }

    /**
     * Generate keyboard events specified by an array of information.<p>
     *
     * The intended use of this is something like this:<p>
     *
     * <pre><code>
     *  robot.type(new int[] {
     *     KeyEvent.VK_H,
     *     KeyEvent.VK_I,
     *     KeyEvent.VK_SPACE,
     *     KeyEvent.VK_T,
     *     KeyEvent.VK_H,
     *     KeyEvent.VK_E,
     *     KeyEvent.VK_R,
     *     KeyEvent.VK_E });
     * </code></pre>
     *
     * @param events The list of key codes to type.
     */
    public void type(int [] events) {
      int nEvents = events.length;
      int evNum = 0;
      
      for (evNum = 0; evNum < nEvents; evNum++) {
	type(events[evNum]);
      }
    }

  /**
   * Send keyboard events as specified by a String.
   * The String is assumed to contain ASCII, and different case
   * in the string converts to interspersed VK_SHIFT events.
   *
   * @param s The string to type.
   */
  public void type(String s) {

    // NOTE: I (David Herron) did not write this function originally, one of
    // the India team wrote this.  While the code looks correct there are
    // two things I am worried over;
    //
    // 1- How are non ASCII character sets to be supported here?
    // 2- What does the '55' mean?  Where did that value come from?
    //
    // The code does work ...

    for (int j=0; j < s.length(); j++) {
      if (Character.isLowerCase(s.charAt(j))) {
	type((Character.getNumericValue(s.charAt(j))) + 55);  
      }
      else {
	keyPress(KeyEvent.VK_SHIFT);
	type((Character.getNumericValue(s.charAt(j))) + 55);  
	keyRelease(KeyEvent.VK_SHIFT);
      }
    }
  }
 


    /**
     * Generate an event to <i>press</i> the ALT key.
     * One does this to turn on the ALT modifier.
     */
    public void pressAlt() {
      keyPress(defaultWait, KeyEvent.VK_ALT);
    }

    /**
     * Generate an event to <i>release</i> the ALT key.
     * One does this to turn off the ALT modifier.
     */
    public void releaseAlt() {
      keyRelease(defaultWait, KeyEvent.VK_ALT);
    }

    /**
     * Generate an event to <i>press</i> the ALT_GRAPH key.
     * One does this to turn on the ALT_GRAPH modifier.
     */
    public void pressAltGraph() {
      keyPress(defaultWait, KeyEvent.VK_ALT_GRAPH);
    }

    /**
     * Generate an event to <i>release</i> the ALT_GRAPH key.
     * One does this to turn off the ALT_GRAPH modifier.
     */
    public void releaseAltGraph() {
      keyRelease(defaultWait, KeyEvent.VK_ALT_GRAPH);
    }

    /**
     * Generate an event to <i>press</i> the CTRL key.
     * One does this to turn on the CTRL modifier.
     */
    public void pressControl() {
      keyPress(defaultWait, KeyEvent.VK_CONTROL);
    }

    /**
     * Generate an event to <i>release</i> the CTRL key.
     * One does this to turn off the CTRL modifier.
     */
    public void releaseControl() {
      keyRelease(defaultWait, KeyEvent.VK_CONTROL);
    }

    /**
     * Generate an event to <i>press</i> the META key.
     * One does this to turn on the META modifier.
     */
    public void pressMeta() {
      keyPress(defaultWait, KeyEvent.VK_META);
    }

    /**
     * Generate an event to <i>release</i> the META key.
     * One does this to turn off the META modifier.
     */
    public void releaseMeta() {
      keyRelease(defaultWait, KeyEvent.VK_META);
    }

    /**
     * Generate an event to <i>press</i> the SHIFT key.
     * One does this to turn on the SHIFT modifier.
     */
    public void pressShift() {
      keyPress(defaultWait, KeyEvent.VK_SHIFT);
    }

    /**
     * Generate an event to <i>release</i> the SHIFT key.
     * One does this to turn off the SHIFT modifier.
     */
    public void releaseShift() {
      keyRelease(defaultWait, KeyEvent.VK_SHIFT);
    }

    /**
     * Generate an event to <i>press</i>
     * the CAPS_LOCK key.  One does this to toggle the SHIFT_LOCK modifier.
     */
    public void pressCapsLock() {
      keyPress(defaultWait, KeyEvent.VK_CAPS_LOCK);
    }

    /**
     * Generate an event to <i>release</i>
     * the CAPS_LOCK key.  One does this to toggle the SHIFT_LOCK modifier.
     */
    public void releaseCapsLock() {
      keyRelease(defaultSpeed, KeyEvent.VK_CAPS_LOCK);
    }

    /////////////////////////
    // Screen shot support //
    /////////////////////////

    /**
     * Returns the color of a pixel at the given screen coordinates
     * @param	x	X position of pixel
     * @param	y	Y position of pixel
     * @return  Color	Color of the pixel
     */
    public Color getPixelColor(int x, int y) {
      return robot.getPixelColor(x, y);
    }

}
