/*
 * @(#)ProbeBuffer.java	1.4 00/02/02
 *
 * Copyright 1995-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.robot.probe;

import java.util.*;

/**
 * Data store Listener for ProbeObjects.  This buffer can be
 * told to be recording or not recording.  Callers can also
 * consult the buffer and manipulate it as a Vector.
 *
 * @see ProbeMonitor
 * @see ProbeObject
 *
 * @author David Herron <tt>&lt;David.Herron@eng.sun.com&gt;</tt>
 */
public class ProbeBuffer implements ProbeObjectListener, Cloneable {

  Vector events = null;
  boolean bRecording = false; // Recording mode
  ProbeMonitor monitor;

  /**
   * Initialize the ProbeBuffer.
   */
  public ProbeBuffer(ProbeMonitor monitor) {
    this(monitor, new Vector());
    connect();
    startRecording();
  }

  protected ProbeBuffer(ProbeMonitor monitor, Vector vec) {
    this.monitor = monitor;
    events = vec;
  }

  protected void finalize() {
    disconnect();
  }

  /**
   * Connect to the ProbeMonitor.  Events are available to be recorded
   * only when the buffer is connected to the monitor.
   */
  public void connect() {
    monitor.addProbeObjectListener(this);
  }

  /**
   * Disconnect from the ProbeMonitor.  Events can no longer be
   * recorded after disconnection from the monitor.
   */
  public void disconnect() {
    monitor.removeProbeObjectListener(this);
  }


  /**
   * Enable recording of probes.  This does not cause connection with the
   * monitor, hence recording events requires both this state setting
   * and being connected.
   */
  public void startRecording() { bRecording = true; }

  /**
   * Disable recording of probes.
   */
  public void stopRecording() { bRecording = false; }

  /**
   * Indicate whether we are recording probes.
   */
  public boolean isRecording() { return bRecording; }

  /**
   * Method to receive the ProbeObject's via the probe monitor Listener
   * interface.  Ignore the event (do not record) if
   * <ul>
   * <li> we are not recording (<tt>bRecording</tt>), or
   * <li> it is an ProbeAWTEvent and the source of the event is an
   *      object contained within this panel.
   * </ul>
   */
  public void probeTriggered(ProbeObject evt) {
    if (!bRecording)
      return;

    // This may be useful.  It is left over from
    // the ProbeViewer implementation.  There this was
    // need for events in the ProbeViewer pane
    // to not be recorded.
    //
    // XXX: There is a general need to filter events
    // in order to maybe record or not record them.
    //
    //    if (evt.type == ProbeObject.EVENT) {
    //      if (this.isAncestorOf((Component)evt.getSource()) == true
    //       || evt.getSource() == this)
    //	return;
    //    }

    events.add(evt);
  }

  /**
   * Get the Iterator for the buffer.
   */
  public Iterator listIterator() { return events.listIterator(); }

  /**
   * Empty out the recorded events list.
   */
  public void clear() { events.removeAllElements(); }

  /**
   * Access the n'th object in the buffer.
   */
  public ProbeObject elementAt(int index) { return (ProbeObject) events.elementAt(index); }

  /**
   * Return the number of objects in the buffer.
   */
  public int size() { return events.size(); }

  /**
   * Returns a clone of this buffer.
   */
  public ProbeBuffer duplicate() { return new ProbeBuffer(monitor, (Vector)events.clone()); }
}
