/*
 * @(#)ProbeMonitor.java	1.5 00/02/02
 *
 * Copyright 1995-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.robot.probe;

import java.util.Vector;

/**
 * The <i>ProbeMonitor</i> exists soley to be the distribution point
 * for ProbeObjects.  Any object interested in receiving ProbeObjects
 * must implement the <i>ProbeObjectListener</i> interface and add
 * themselves as a listener.  Any object interested in sending Probe
 * information must create an object subclassed from <i>ProbeObject</i>
 * and <i>post</i> it to a ProbeMonitor.
 *
 * If the variable <i>dispatch</i> is <tt>false</tt> then any
 * ProbeObjects received are discarded.
 *
 * The <i>probe</i> system has three parts (probe sources, the ProbeMonitor,
 * and the probe listeners).  The ProbeMonitor is the central switching
 * point to this system.
 *
 * @see ProbeObjectListener
 * @see ProbeObject
 * 
 */
public class ProbeMonitor {

  private Vector listeners = new Vector();

  /**
   * Control whether <i>ProbeMonitor</i> is to be dispatching events.
   * If <tt>true</tt> then events received on the <tt>post</tt> interface
   * are sent to listeners.
   */
  public boolean dispatch = true;

  /**
   * Initialize the ProbeMonitor.
   */
  public ProbeMonitor() {
    dispatch = true;
  }

  /**
   * Post a ProbeObject into the monitor.  It gets distributed
   * to the listener(s).  It is synchronized to avoid having
   * multiple post operations active at the same exact moment,
   * which should increase the serial-ness of delivering these events.
   */
  public void post(ProbeObject evt) {

    if (!dispatch)
      return;

    Vector t1;
    synchronized(listeners) {
      t1 = (Vector) listeners.clone();
    }

    int size = t1.size();
    if (size == 0)
      return;

    // NOTE: This is probably not the best way to proceed
    // with this task.  It should be better to have ProbeMonitor
    // create another thread when initialized, and have post
    // send the events to that thread.  The thread then handles
    // this dispatch as it sees fit.
    //
    // Also we synchronize on this whole method to help ensure there
    // is not multiple post's delivering at the same time.
    for (int i = 0; i < size; i++) {
      ProbeObjectListener l = (ProbeObjectListener)t1.elementAt(i);
      l.probeTriggered(evt);
    }
  }

  //  public void post(ProbeComponentMethodCall evt) {
  //    post((ProbeObject)evt);
  //  }

  /**
   * Add a ProbeObject listener.
   */
  public void addProbeObjectListener(ProbeObjectListener l) {
    synchronized(listeners) {
      if (listeners.contains(l))
	return;
      listeners.addElement(l);
    }
  }

  /**
   * Remove a ProbeObject listener.
   */
  public void removeProbeObjectListener(ProbeObjectListener l) {
    synchronized(listeners) {
      listeners.removeElement(l);
    }
  }
}
