/*
 * @(#)Converters.java	1.7 00/07/07
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.io;

import java.io.UnsupportedEncodingException;
import java.lang.ref.SoftReference;


/**
 * Package-private utility class that caches the default converter classes and
 * provides other logic common to both the ByteToCharConverter and
 * CharToByteConverter classes.
 *
 * @author   Mark Reinhold
 * @version  1.7, 00/07/07
 * @since    JDK1.2
 */

class Converters {

    private Converters() { }	/* To prevent instantiation */

    /* Lock for all fields in this class */
    private static Object lock = Converters.class;

    /* Cached values of system properties */
    private static String converterPackageName = null;  /* file.encoding.pkg */
    private static String defaultEncoding = null;       /* file.encoding */

    /* Converter type constants and names */
    public static final int BYTE_TO_CHAR = 0;
    public static final int CHAR_TO_BYTE = 1;
    private static final String[] converterPrefix = { "ByteToChar",
						      "CharToByte" };

    /* Cached converter classes, indexed by above constants */
    private static SoftReference[] classCache = new SoftReference[2];

    /** Get the name of the converter package */
    private static String getConverterPackageName() {
	String cp = converterPackageName;
	if (cp != null) return cp;
	java.security.PrivilegedAction pa =
	    new sun.security.action.GetPropertyAction("file.encoding.pkg");
	cp = (String)java.security.AccessController.doPrivileged(pa);
	if (cp != null) {
	    /* Property is set, so take it as the true converter package */
	    converterPackageName = cp;
	} else {
	    /* Fall back to sun.io */
	    cp = "sun.io";
	}
	return cp;
    }

    /**
     * Get the class that implements the given type of converter for the named
     * encoding, or throw an UnsupportedEncodingException if no such class can
     * be found
     */
    private static Class getConverterClass(int type, String encoding)
	throws UnsupportedEncodingException
    {
	String enc = null;

	/* "ISO8859_1" is the canonical name for the ISO-Latin-1 encoding.
	   Native code in the JDK commonly uses the alias "8859_1" instead of
	   "ISO8859_1".  We hardwire this alias here in order to avoid loading
	   the full alias table just for this case. */
	if (!encoding.equals("ISO8859_1")) {
	    if (encoding.equals("8859_1")) {
		enc = "ISO8859_1";
	    /*
	     * On Solaris with nl_langinfo() called in GetJavaProperties():
	     *
	     *   locale undefined -> NULL -> hardcoded default
	     *   "C" locale       -> "" -> hardcoded default	(on 2.6)
	     *   "C" locale       -> "646" 			(on 2.7)
	     *   "en_US" locale -> "ISO8859-1"
	     *   "en_GB" locale -> "ISO8859-1"			(on 2.7)
	     *   "en_UK" locale -> "ISO8859-1"			(on 2.6)
	     */
	    } else if (encoding.equals("ISO8859-1")) {
		enc = "ISO8859_1";
	    } else if (encoding.equals("646")) {
		enc = "ASCII";
	    } else {
		enc = CharacterEncoding.aliasName(encoding);
	    }
	}
	if (enc == null) {
	    enc = encoding;
	}

	try {
	    return Class.forName(getConverterPackageName()
				 + "." + converterPrefix[type] + enc);
	} catch(ClassNotFoundException e) {
	    throw new UnsupportedEncodingException(enc);
	}

    }

    /**
     * Instantiate the given converter class, or throw an
     * UnsupportedEncodingException if it cannot be instantiated
     */
    private static Object newConverter(String enc, Class c)
	throws UnsupportedEncodingException
    {
	try {
	    return c.newInstance();
	} catch(InstantiationException e) {
	    throw new UnsupportedEncodingException(enc);
	} catch(IllegalAccessException e) {
	    throw new UnsupportedEncodingException(enc);
	}
    }

    /**
     * Create a converter object that implements the given type of converter
     * for the given encoding, or throw an UnsupportedEncodingException if no
     * appropriate converter class can be found and instantiated
     */
    static Object newConverter(int type, String enc)
	throws UnsupportedEncodingException
    {
	Class c;
	synchronized (lock) {
	    c = getConverterClass(type, enc);
	}
	return newConverter(enc, c);
    }

    /**
     * Find the class that implements the given type of converter for the
     * default encoding.  If the default encoding cannot be determined or is
     * not yet defined, return a class that implements the fallback default
     * encoding, which is just ISO 8859-1.
     */
    private static Class getDefaultConverterClass(int type) {
	boolean fillCache = false;

	/* First check the class cache */
	SoftReference sr = classCache[type];
	if (sr != null) {
	    Class c = (Class)sr.get();
	    if (c != null) {
		return c;
	    }
	    /* The soft reference has been cleared, so clear the cache */
	    classCache[type] = null;
	}

	/* Do we have a default encoding yet? */
	String enc = defaultEncoding;
	if (defaultEncoding != null) {
	    fillCache = true;
	} else {
	    /* Try to look up the property */
	    java.security.PrivilegedAction a =
		new sun.security.action.GetPropertyAction("file.encoding");
	    enc = (String)java.security.AccessController.doPrivileged(a);
	    if (enc != null) {
		/* Property is set, so take it as the default encoding */
		defaultEncoding = enc;
		fillCache = true;
	    } else {
		/* Property not yet set, so do not set default encoding and
		   just use ISO8859_1 for now */
		enc = "ISO8859_1";
	    }
	}

	/* We have an encoding name; try to find its class */
	Class c;
	try {
	    c = getConverterClass(type, enc);
	    if (fillCache) {
		classCache[type] = new SoftReference(c);
	    }
	} catch (UnsupportedEncodingException x) {
	    /* Can't find the default class, so fall back to ISO 8859-1 */
	    try {
		c = getConverterClass(type, "ISO8859_1");
	    } catch (UnsupportedEncodingException y) {
		throw new InternalError("Cannot find default "
                                        + converterPrefix[type]
					+ " converter class");
	    }
	}
	return c;

    }

    /**
     * Create a converter object that implements the given type of converter
     * for the default encoding, falling back to ISO 8859-1 if the default
     * encoding cannot be determined.
     */
    static Object newDefaultConverter(int type) {
	Class c;
	synchronized (lock) {
	    c = getDefaultConverterClass(type);
	}
	try {
	    return newConverter("", c);
	} catch (UnsupportedEncodingException x) {
	    throw new InternalError("Cannot instantiate default converter"
				    + " class " + c.getName());
	}
    }

}
