/*
 * @(#)SunGraphicsEnvironment.java	1.86 01/02/09
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.java2d;

import java.awt.GraphicsEnvironment;
import java.awt.GraphicsDevice;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.Font;
import java.text.AttributedCharacterIterator;
import java.awt.print.PrinterJob;
import java.util.Map;
import java.util.Hashtable;
import java.util.Locale;
import java.util.Vector;
import java.util.StringTokenizer;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import sun.awt.image.BufferedImageGraphics2D;
import sun.awt.font.FontNameAliases;
import java.io.IOException;
import java.io.FilenameFilter;
import java.io.File;
import java.util.NoSuchElementException;
import sun.awt.font.NativeFontWrapper;
import java.util.TreeMap;
import java.util.Set;
import java.awt.font.TextAttribute;
import sun.awt.FontProperties;

/**
 * This is an implementation of a GraphicsEnvironment object for the
 * default local GraphicsEnvironment used by the JavaSoft JDK.
 *
 * @see GraphicsDevice
 * @see GraphicsConfiguration
 * @version 10 Feb 1997
 */

public abstract class SunGraphicsEnvironment extends GraphicsEnvironment {

    private static final boolean debugMapping = false;

    private static String [] logicalFontNames = {
	"default",
	"serif",
	"sansserif",
	"monospaced",
	"dialog",
	"dialoginput",
    };

    protected FontProperties fprops;
    private TreeMap terminalNames;
    private boolean loadedAllFonts;
    protected String fontPath;
    protected TreeMap registeredFonts;
    protected String osVersion;
    private Hashtable mapFamilyCache;
    protected boolean loadNativeFonts = false;
    
    public SunGraphicsEnvironment() {
        loadedAllFonts = false;
        registeredFonts = new TreeMap();
        osVersion = getOsVersion();

	java.security.AccessController.doPrivileged(
				    new java.security.PrivilegedAction() {
	    public Object run() {

		fontPath = System.getProperty("java.awt.fonts");

                if (fontPath.length() == 0) {
		    boolean noType1Font = false;
		    String prop = System.getProperty ("sun.java2d.noType1Font");
		    if (prop == null) {
		      noType1Font = NativeFontWrapper.getType1FontVar();
		    }
		    if ("true".equals(prop)) {
		      noType1Font = true;
		    }
                    fontPath = NativeFontWrapper.getFontPath(noType1Font);
		    appendJavaDefaultFontPath();
		    loadNativeFonts = true;
                }
		initializeProps();
		initCompositeFonts();
		// loadFonts();
		return null;
	    }
	});
   }

    GraphicsDevice[] screens;

    /* Fix: 4175044
    *  allow platform to override version extension
    *  for font.properties files
    */
    protected String getOsVersion() {
        return System.getProperty("os.version");
    }

    protected synchronized void loadFonts() {
        if (loadedAllFonts) {
            return;
        }
	java.security.AccessController.doPrivileged(
				    new java.security.PrivilegedAction() {
	    public Object run() {
                if (fontPath == null) {
		    fontPath = System.getProperty("java.awt.fonts");

                    if (fontPath.length() == 0) {
		        boolean noType1Font = false;
			String prop = System.getProperty ("sun.java2d.noType1Font");
			if (prop == null) {
			    noType1Font = NativeFontWrapper.getType1FontVar();
			}
			if ("true".equals(prop)) {
			    noType1Font = true;
			}
			fontPath = NativeFontWrapper.getFontPath(noType1Font);
			appendJavaDefaultFontPath();
                        loadNativeFonts = true;
                    }
                }

		boolean foundFonts = registerFonts (fontPath);
		boolean foundNativeFonts = false;
                if (loadNativeFonts) {
                    foundNativeFonts = registerNativeFonts ();
                }
		if ((!foundFonts) && (!foundNativeFonts)) { 
		    //- REMIND adg: need to throw an exception
		    System.out.println("\nNo fonts were found in '"
                                            + fontPath + "'.\n");
		    System.exit(2);
		}
		
                loadedAllFonts = true;
		return null;
	    }
	});
    }

    /**
     * Returns an array of all of the screen devices.
     */
    public synchronized GraphicsDevice[] getScreenDevices() {
	GraphicsDevice[] ret = screens;
	if (ret == null) {
	    int num = getNumScreens();
	    ret = new GraphicsDevice[num];
	    for (int i = 0; i < num; i++) {
		ret[i] = makeScreenDevice(i);
	    }
	    screens = ret;
	}
	return ret;
    }

    protected abstract int getNumScreens();
    protected abstract GraphicsDevice makeScreenDevice(int screennum);

    /**
     * Returns the default screen graphics device.
     */
    public GraphicsDevice getDefaultScreenDevice() {
	return getScreenDevices()[0];
    }

    /**
     * Returns a Graphics2D object for rendering into the
     * given BufferedImage.
     */
    public Graphics2D createGraphics(BufferedImage img) {
	return new BufferedImageGraphics2D(img, null);
    }

    private Font[] allFonts;
   
    /**
     * Returns all fonts available in this environment.
     */
    public Font[] getAllFonts() {
        if (allFonts != null) {
            return allFonts;
        }
        loadFonts();
        Font [] fonts = null;
        String [] fontNames = null;
        int count = NativeFontWrapper.getNumFonts();
        if (count > 0) {
            TreeMap fontMapNames = new TreeMap();
            for (int i=0; i < count; i++) {
                String name = NativeFontWrapper.getFullNameByIndex(i);
                fontMapNames.put(name, null);
            }
            if (fontMapNames.size() > 0) {
                fontNames = new String[fontMapNames.size()];
                Object [] keyNames = fontMapNames.keySet().toArray();
                for (int i=0; i < keyNames.length; i++) {
                    fontNames[i] = (String)keyNames[i];
                }
            }
        }
        if (fontNames != null) {
            fonts = new Font[fontNames.length];
            for (int i=0; i < fontNames.length; i++) {
                fonts[i] = new Font(fontNames[i], Font.PLAIN, 1);
            }
        }
        allFonts = fonts;
        return allFonts;
    }

    public String[] getAvailableFontFamilyNames(Locale theLocale) {
        loadFonts();
        String [] retval = null;
        int count = NativeFontWrapper.getNumFonts();

	int localeID = getLcidFromLocale(theLocale);

        if (count > 0) {
            TreeMap familyNames = new TreeMap();
            for (int i=0; i < count; i++) {
                String name = NativeFontWrapper.getFamilyNameByIndex(i, localeID);
		String cmpName = name.toLowerCase();
		if ( cmpName.endsWith( ".bold" ) || 
		     cmpName.endsWith( ".bolditalic" ) || 
		     cmpName.endsWith ( ".italic" )){
		}
		else {
		  familyNames.put(cmpName, name);
		}
            }
            String str;

            // compatibility
            str = "Serif";          familyNames.put(str.toLowerCase(), str);
            str = "SansSerif";      familyNames.put(str.toLowerCase(), str);
            str = "Monospaced";     familyNames.put(str.toLowerCase(), str);
            str = "Dialog";         familyNames.put(str.toLowerCase(), str);
            str = "DialogInput";    familyNames.put(str.toLowerCase(), str);

//            str = "TimesRoman";     familyNames.put(str.toLowerCase(), str);
//            str = "Helvetica";      familyNames.put(str.toLowerCase(), str);
//            str = "Courier";        familyNames.put(str.toLowerCase(), str);
//            str = "ZapfDingbats";   familyNames.put(str.toLowerCase(), str);

            str = "Default";        familyNames.put(str.toLowerCase(), str);

            if (familyNames.size() > 0) {
                retval = new String[familyNames.size()];
                Object [] keyNames = familyNames.keySet().toArray();
                for (int i=0; i < keyNames.length; i++) {
                    retval[i] = (String)familyNames.get(keyNames[i]);
                }
            }
        }
        return retval;
    }

    public String[] getAvailableFontFamilyNames() {
        return getAvailableFontFamilyNames(Locale.getDefault());
    }

    /** Convenience method to locate font familynames 
        Not part of GraphicsEnvironment
     */
    public String mapFamilyName(String originalName, int style )
    {
      String newName;

      String tempName = originalName.toLowerCase(Locale.ENGLISH);
      // check for the .plain ending for some fontproperties files in the CJK
      // have these names - we just trim the font name for the .plain, so
      // we do not trigger a full search of the fonts - performance issue
      int index;
      if ( (index = tempName.lastIndexOf(".plain" )) != -1 ) {
	 tempName = tempName.substring ( 0, index );
      }

      if ( ! terminalNames.containsKey ( tempName ) ) {
	  loadFonts();
	   // System.out.println ( "Going thru the load  fonts case" );
      }
      newName = mapFamilyNameAll ( originalName, style );

      return newName;
    }

    /**
    * not part of GraphicsEnvironment methods.
    * here as convenience to Font
    * maps family name base on font.properties file and/or default algorithms.
    */
    public String mapFamilyNameAll(String originalName, int style) {
        String lcName = originalName.toLowerCase();
        if (null != mapFamilyCache) {
            String family = (String)mapFamilyCache.get(
                                                lcName + styleStr(style));
            if (null != family) {
                return family;
            }
        }

        if (fprops == null) {
            return originalName;
        }
        if (debugMapping) {
            System.out.println("Originalname[" + originalName + "."
                                            + styleStr(style) + "]");
        }
        String familyAlias = fprops.getProperty("alias" + "." + lcName);
        if (familyAlias == null) {
            familyAlias = originalName;
        }

        Object keynames[] = fprops.keySet().toArray();
        int maxIter = keynames.length;
        int iterCount = 0;

        // cheap way to detect cycles in the mapping
        do {
            String familyStr = familyAlias.toLowerCase();
            if (terminalNames.containsKey(familyStr + "." + styleStr(style))) {
                familyAlias = familyStr + "." + styleStr(style);
                break;  // no further resolution permitted
            }
            if ((style == 0) && terminalNames.containsKey(familyStr)) {
                break;  // no further resolution permitted
            }

            String propStr = familyStr + "." + styleStr(style) + ".0";
            String newName = fprops.getProperty(propStr);
            if (debugMapping) {
                System.out.println("  trying [" + propStr + "]");
            }
            if (newName == null) {
                propStr = familyStr + ".0";
                newName = fprops.getProperty(propStr);
                if (debugMapping) {
                    System.out.println("  trying [" + propStr + "]");
                }
            }
            if (newName == null) {
                break;  // fully resolved
            }
            if (debugMapping) {
                System.out.println("  found [" + newName + "]");
            }
            familyAlias = parseFamilyNameProperty(newName);
        } while (++iterCount< maxIter);
        if (debugMapping) {
            System.out.println("    resolved to [" + familyAlias + "]");
        }
        if (null == mapFamilyCache) {
            mapFamilyCache = new Hashtable(5, (float)0.9);
        }
        if (null != mapFamilyCache) {
            mapFamilyCache.put(lcName + styleStr(style), familyAlias);
        }
        return familyAlias;
    }

    // can have platform-specific overrides - see X11GraphicsConfiguration
    protected String parseFamilyNameProperty(String name) {
        int separator = name.indexOf(",");
        if (separator == -1) {
            separator = name.length();
        }
        return name.substring(0, separator);
    }

    // can have platform-specific overrides - see X11GraphicsConfiguration
    protected String getFontPropertyFD(String name) {
        return parseFamilyNameProperty(name);
    }

    // can have platform-specific overrides - see X11GraphicsConfiguration
    protected String getFileNameFromPlatformName(String platName) {
        if (fprops == null) {
            return null;
        }
        platName = platName.replace(' ','_');
	return fprops.getProperty("filename" + "." + platName);
    }

    /**
     * Gets a <code>PrintJob2D</code> object suitable for the
     * the current platform.
     * @return    a <code>PrintJob2D</code> object.
     * @see       java.awt.PrintJob2D
     * @since     JDK1.2
     */
    public PrinterJob getPrinterJob() {
        new Exception().printStackTrace();
	return null;
    }

    /* MACPORTING NOTE.needs to do file type on the Macintosh */
    // adg: ttc files are now handled by the ttf code
    public class TTFilter implements FilenameFilter{
        public boolean accept(File dir,String name) {
            return(name.endsWith(".ttf") ||
                   name.endsWith(".TTF") ||
                   name.endsWith(".ttc") ||
                   name.endsWith(".TTC"));
        }
    }			 

    public class T2KFilter implements FilenameFilter{
        public boolean accept(File dir,String name) {
            return(name.endsWith(".t2k") ||
                   name.endsWith(".T2K"));
        }
    }			 

    public class T1Filter implements FilenameFilter{
        public boolean accept(File dir,String name) {
            return(name.endsWith(".ps") ||
                   name.endsWith(".PS") ||
                   name.endsWith(".pfb") ||
                   name.endsWith(".PFB") ||
                   name.endsWith(".pfa") ||
                   name.endsWith(".PFA"));
        }
    }

    private boolean registerFonts(String pathName) {
        boolean retval = false;
        StringTokenizer parser = new StringTokenizer(pathName, 
                                                     File.pathSeparator);
        try {
            while (parser.hasMoreTokens()) {
                String newPath = parser.nextToken();
                registerFontPath(newPath);
                retval |= addPathFonts(newPath, new TTFilter(), FontNameAliases.FONTFORMAT_TRUETYPE);
                retval |= addPathFonts(newPath, new T1Filter(), FontNameAliases.FONTFORMAT_TYPE1);
                retval |= addPathFonts(newPath, new T2KFilter(), FontNameAliases.FONTFORMAT_T2K);
            }
        } catch (NoSuchElementException e) {
            System.err.println(e);
        }
        return retval;
    }

    // can have platform specific override
    protected void registerFontFile(String fontFileName, String platName) {
//        System.out.println("Registering: " + fontFileName);
        // REMIND: case compare depends on platform
        if (registeredFonts.containsKey(fontFileName)) {
//            System.out.println("   !Already done!: " + fontFileName);
            return;
        }
        int fontFormat;
        if (new TTFilter().accept(null, fontFileName)) {
            fontFormat = FontNameAliases.FONTFORMAT_TRUETYPE;
        } else if (new T1Filter().accept(null, fontFileName)) {
            fontFormat = FontNameAliases.FONTFORMAT_TYPE1;
        } else if (new T2KFilter().accept(null, fontFileName)) {
            fontFormat = FontNameAliases.FONTFORMAT_T2K;
        } else {
	    registerNative (fontFileName);
            return;
        }

        StringTokenizer parser = new StringTokenizer(fontPath, 
                                                     File.pathSeparator);
        try {
            while (parser.hasMoreTokens()) {
                String newPath = parser.nextToken();
		
                File theFile =  new File(newPath, fontFileName);
                if (theFile.canRead()) {
                    Vector fontNames = new Vector(20, 10);
		    Vector platNames = new Vector(20, 10);
		    platNames.addElement(platName);
                    fontNames.addElement(theFile.getAbsolutePath());
            	    registeredFonts.put(fontFileName, 
					theFile.getAbsolutePath ());
                    NativeFontWrapper.registerFonts(fontNames, fontNames.size(), platNames, fontFormat);
                    break;
                }
            }
        } catch (NoSuchElementException e) {
            System.err.println(e);
        }
    }

    // can have platform specific override
    protected void registerFontPath(String path) {
    }
    protected void registerNative (String fontFileName) {
    }
    protected String getPropString (String fontFileName) {
	return fontFileName;
    }

    protected boolean registerNativeFonts () {
	return false;
    }

    /*
    * helper function for registerFonts
    */
    private boolean addPathFonts(String path, FilenameFilter filter, int fontFormat) {
        boolean retval = false;
        Vector fontNames = new Vector(20, 10);
	Vector propVals = new Vector(20,10);
        File f1 = new File(path);
        String[] ls = f1.list(filter);
        if (ls == null) {
            return retval;
        }
        for (int i=0; i < ls.length; i++ ) {
            File theFile =  new File(f1, ls[i]);
//            System.out.println(ls[i]);
            // REMIND: case compare depends on platform
            if (registeredFonts.containsKey(ls[i])) {
//                System.out.println("   !Already done!: " + ls[i]);
                continue;
            }
            String fullName = theFile.getAbsolutePath();
//            System.out.println(fullName);
            registeredFonts.put(ls[i], fullName);
            fontNames.addElement(fullName);
	    propVals.addElement (getPropString (fullName));
            retval = true;
        }
        // REMIND - native code might not register everything which we
        // pass into it.
        NativeFontWrapper.registerFonts(fontNames, fontNames.size(), propVals,
			 fontFormat);
        return retval;  // REMIND: get status of registration from native
    }

    /**
     * Resolve styles on the character at start into an instance of Font
     * that can best render the text between start and limit.
     * REMIND jk. Move it to graphics environment.
     */
    public static Font getBestFontFor(AttributedCharacterIterator text,
                                      int start, int limit) {

        /*
         * choose the first font that can display the first character
	 * first iterate through the styles in the range of text we were
         * passed.  If none of them work, iterate through font families
         * using the attributes on the first character.  If this also
         * fails, use the first font.
         */
        char c = text.setIndex(start);
        Map ff = text.getAttributes();
        Font font = Font.getFont(ff);

        while (!font.canDisplay(c) && (text.getRunLimit() < limit)) {
	        text.setIndex(text.getRunLimit());
	        font = Font.getFont(text.getAttributes());
        }

	if (!font.canDisplay(c)) {
	    text.setIndex(start);
	    String[] families = 
                GraphicsEnvironment.getLocalGraphicsEnvironment(
                                        ).getAvailableFontFamilyNames();
	    for (int i = 0; i < families.length; ++i) {
			Hashtable ht = new Hashtable();
			ht.putAll(ff);
	        ht.put(TextAttribute.FAMILY, families[i]);
	        font = Font.getFont((Map)ht);
	        if (font.canDisplay(c)) {
	            break;
	        }
	    }

	    if (!font.canDisplay(c)) {
	        font = Font.getFont(ff);
	    }
	}

	return font;
    }

    protected boolean platformReady() {
        return true;
    }

    private void appendJavaDefaultFontPath() {
	String defaultPathName =
	    System.getProperty("java.home") + File.separator +
	    "lib" + File.separator +
	    "fonts";
	if (defaultPathName != null) {
	    fontPath = fontPath + File.pathSeparator + defaultPathName;
	}
    }



    protected void initializeProps() {
        if (!platformReady()) {
            return;
        }
	fprops = new FontProperties(osVersion);
    }

    // can have platform-specific overrides - see X11GraphicsConfiguration
    protected void initCompositeFonts() {
        if (!platformReady()) {
            return;
        }
        initTerminalNames();
        Object [] terminalKeys = terminalNames.keySet().toArray();
        for (int i=0; i < terminalKeys.length; i++) {
            String compositeFontName = (String)terminalKeys[i];
            Integer maxEntryInt = (Integer)terminalNames.get(terminalKeys[i]);
            int numEntries = maxEntryInt.intValue();
            String names[] = new String[numEntries];
            int exclusionMaxIndex[] = new int[numEntries];
            int exclusionRanges[] = new int[0];
            for (int j=0; j < numEntries; j++) {
                names[j] = parseFamilyNameProperty(
                                fprops.getProperty(
                                        compositeFontName + "." + j));
                exclusionRanges = 
                    appendExclusions(compositeFontName, j, exclusionRanges);
                exclusionMaxIndex[j] = exclusionRanges.length;
            }

            if (debugMapping) {
                System.out.println("initCompositeFonts compositeFontName="+
                                   compositeFontName);
            }
            NativeFontWrapper.registerCompositeFont(
                compositeFontName, names,
                exclusionRanges, exclusionMaxIndex);
        }
        terminalNames.put("default",new Integer(0));
    }

    private int[] appendExclusions(String name, int slot, int [] ranges) {
        String keyname = "exclusion." + name + "." + slot;
        // REMIND: invent exclusion ranges for dingbats and symbols
        // since no properties files specify them
        // (or fix all properties files --- better)
        if (fprops.containsKey(keyname)) {
            String exclusions = fprops.getProperty(keyname);
	    /*
	     * range format is xxxx-XXXX,yyyy-YYYY,.....
	     */
            int numExclusions = (exclusions.length() + 1) / 10;
            if (numExclusions > 0) {
                int newRanges[] = new int[numExclusions * 2];
                for (int i = 0; i < numExclusions; i++) {
                    String lower = exclusions.substring(i*10    , i*10 + 4);
                    String upper = exclusions.substring(i*10 + 5, i*10 + 9);
                    newRanges[i*2  ] = Integer.parseInt(lower, 16);
                    newRanges[i*2+1] = Integer.parseInt(upper, 16);
                }
                int totalRanges = ranges.length + newRanges.length;
                int tempRanges[]  = new int[totalRanges];
                System.arraycopy(   ranges, 0,
                                    tempRanges, 0,
                                    ranges.length);
                System.arraycopy(   newRanges, 0,
                                    tempRanges, ranges.length,
                                    newRanges.length);
                ranges = tempRanges;
            }
        }
        return ranges;
    }

    private void initTerminalNames() {
        TreeMap predefinedNames = new TreeMap();
        TreeMap registeredFileNames = new TreeMap();
        String str;

        this.terminalNames = new TreeMap();

        // compatibility
        str = "Serif";          predefinedNames.put(str.toLowerCase(Locale.ENGLISH), str);
        str = "SansSerif";      predefinedNames.put(str.toLowerCase(Locale.ENGLISH), str);
        str = "Monospaced";     predefinedNames.put(str.toLowerCase(Locale.ENGLISH), str);
        str = "Dialog";         predefinedNames.put(str.toLowerCase(Locale.ENGLISH), str);
        str = "DialogInput";    predefinedNames.put(str.toLowerCase(Locale.ENGLISH), str);

	if (fprops == null)
	    throw new Error("no font properties file found.");

        Object [] propKeys = fprops.keySet().toArray();

        for (int i=0; i < propKeys.length; i++) {
            // discard keys which aren't predefined font family names
            String property = (String)propKeys[i];
            int separator = property.indexOf(".");
            if (separator == -1) {
                separator = property.length();
            }
            String propFamily = property.substring(0, separator);
            if (!predefinedNames.containsKey(propFamily)) {
                continue;   // discard, not predefined
            }

            // find out how many entries for this key
            separator = property.lastIndexOf(".");
            if (separator == -1) {
                continue;   // discard, invalid format
            }
            String familyStyle = property.substring(0, separator);
            if (terminalNames.containsKey(familyStyle)) {
                continue;   // discard, already analyzed
            }
            if (!fprops.containsKey(familyStyle + ".0")) {
                continue;   // discard, invalid file format
            }
            int maxEntry = 0;
            while (fprops.containsKey(familyStyle + "." + maxEntry)) {
                maxEntry++;
            }
            if (maxEntry == 0) {
                continue;   // discard, want direct mapping
            }
            terminalNames.put(familyStyle, new Integer(maxEntry));

            if (debugMapping) {
                System.out.println("FamilyStyle: " + familyStyle);
                System.out.println("NumSlots: " + maxEntry);
                System.out.println("Key: " + (String)propKeys[i]);
            }

            for (int j=0; j < maxEntry; j++) {
                String platName = getFontPropertyFD(
                                    fprops.getProperty(
                                                    familyStyle + "." + j));
                String fontFileName = getFileNameFromPlatformName(platName);
//                System.out.println("FS: [" + familyStyle + "." + j
//                                    + "] PN: [" + platName + "] FN: ["
//                                    + fontFileName + "]");
                if (fontFileName == null) {
                    // invalid configuration file(s)
                    System.err.println("Font specified in font.properties not found [" +
                        platName + "]");
                    loadFonts();
                    break;
                } else if (!registeredFileNames.containsKey(fontFileName)) {
//                    System.out.println("PN: [" + platName + "] FN: ["
//                                    + fontFileName + "]");
                    registeredFileNames.put(fontFileName, null);
                    registerFontFile(fontFileName, platName);
                }
            }
        }
    }

    /*
     * return String representation of style
     */
    public static String styleStr(int num){
	switch(num){
	  case Font.BOLD:
	    return "bold";
	  case Font.ITALIC:
	    return "italic";
	  case Font.ITALIC | Font.BOLD:
	    return "bolditalic";
	  default:
	    return "plain";
	}
    }

    public static boolean isLogicalFont(Font f) {
	String name = f.getName().toLowerCase();

	for (int i=0; i<logicalFontNames.length; i++) {
	    if (name.equals(logicalFontNames[i])) {
		return true;
	    }
	}
	return false;
    }

  public static String createFont ( File fontFile ) {

  String fontName = NativeFontWrapper.createFont ( fontFile.getAbsolutePath(), FontNameAliases.FONTFORMAT_TRUETYPE);
 
  return fontName;

  }

  // Return a Microsoft LCID from the given Locale.
  // Used when getting localized font data.
  private static final String systemBundle = 
			"java.text.resources.LocaleElements";
  private static short getLcidFromLocale(Locale l) {

    short lcid = 0x0409;  // US English - default

    // optimize for common case:
    if (l.equals(Locale.US)) {
      return lcid;
    }

    String lcidAsString;
    try {
      ResourceBundle bundle = ResourceBundle.getBundle(systemBundle, l);
      lcidAsString = bundle.getString("LocaleID");
    }
    catch(MissingResourceException e) {
      return lcid;
    }

    try {
      lcid = (short) Integer.parseInt(lcidAsString, 16);
    }
    catch(NumberFormatException e) {
    }

    return lcid;
  }


    /**
     * Return the current font properties.
     */
    public FontProperties getFontProperties() {
	return fprops;
    }
}
