/*
 * @(#)URIName.java	1.9 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.security.x509;

import java.io.IOException;

import sun.security.util.*;

/**
 * This class implements the URIName as required by the GeneralNames
 * ASN.1 object.
 *
 * @author Amit Kapoor
 * @author Hemma Prafullchandra
 * @version 1.9
 * @see GeneralName
 * @see GeneralNames
 * @see GeneralNameInterface
 */
public class URIName implements GeneralNameInterface {
    private String name;

    /**
     * Create the URIName object from the passed encoded Der value.
     *
     * @param derValue the encoded DER URIName.
     * @exception IOException on error.
     */
    public URIName(DerValue derValue) throws IOException {
        name = derValue.getIA5String();
    }

    /**
     * Create the URIName object with the specified name.
     *
     * @param name the URIName.
     */
    public URIName(String name) {
        this.name = name;
    }

    /**
     * Return the type of the GeneralName.
     */
    public int getType() {
        return (GeneralNameInterface.NAME_URI);
    }

    /**
     * Encode the URI name into the DerOutputStream.
     *
     * @param out the DER stream to encode the URIName to.
     * @exception IOException on encoding errors.
     */
    public void encode(DerOutputStream out) throws IOException {
        out.putIA5String(name);
    }

    /**
     * Convert the name into user readable string.
     */
    public String toString() {
        return ("URIName: " + name);
    }

    /**
     * Returns this URI name.
     */
    public String getName() {
        return name;
    }

    /**
     * Compares this name with another, for equality.
     *
     * @return true iff the names are equivalent
     * according to RFC2459.
     */
    public boolean equals(Object obj) {
	if (this == obj)
	    return true;

	if (!(obj instanceof URIName))
	    return false;

	URIName other = (URIName)obj;

	// URI names consist of a scheme name, a host name, and a remainder,
	// e.g. "http://www.sun.com/java/page.html" consists of a scheme name
	// ("http"), a host name ("www.sun.com") and a remainder
	// ("java/page.html").
	// RFC2459 mandates that the scheme name and host name are not
	// case-sensitive, but the remainder is case-sensitive.
	if (name.equalsIgnoreCase(other.name)){

	    // Strip off the scheme name and host name from this.name
	    String remainder = getRemainder(name);

	    // Strip off the scheme name and host name from other.name
	    String otherRemainder = getRemainder(other.name);

	    if ((remainder == null) ^ (otherRemainder == null))
		return false;
	    
	    if (remainder != null && otherRemainder != null)
		return remainder.equals(otherRemainder);

	    // both are null
	    return true;

	} else {
	    return false;
	}
    }

    private String getRemainder(String name) {
	String remainder = null;
	int index = name.indexOf("//");
	if (index > 0) {
	    index = name.indexOf('/', index+2);
	    if ((index > 0) && ((index+1) < name.length())) {
		remainder = name.substring(index+1);
	    }
	}
	return remainder;
    }

    /**
     * Returns the hash code value for this object.
     * 
     * @return a hash code value for this object.
     */
    public int hashCode() {
	return name.toUpperCase().hashCode();
    }
}
