/*
 * @(#)LoadEnvironment.java	1.9 99/02/01
 *
 * Copyright 1999 Sun Microsystems, Inc. All rights reserved.
 * Copyright 1999 Sun Microsystems, Inc. Tous droits rservs.
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * 
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.util;

import java.io.*;
import java.util.Hashtable;
import sun.tools.java.*;
import sun.tools.java.Package;


/**
 * Environment for loading from binary files
 * @version 1.3, 97/05/23
 * @author Robert Field
 */

public 
class LoadEnvironment extends Environment implements Constants {

    /**
     * Flags for Compiler classes
     */
    public int flags = 0;
    
    /**
     * The class path
     */
    public ClassPath path;
    
    /**
     * A hashtable of resource contexts.
     */
    Hashtable packages = new Hashtable(31);

    /**
     * The classes, this is only a cache, not a
     * complete list.
     */
    Hashtable classes = new Hashtable(351);

	
    /**
     * Constructor
     */
    public LoadEnvironment() {
	super();
    }

    /**
     * Constructor
     */
    public LoadEnvironment(ClassPath path) {
	super();
        this.path = path;
    }
    
    /**
     * Default flags for loadFile.  Load all attributes.
     */
    protected int loadFileFlags() { 
	return ATT_ALL; 
    }
    
    /**
     * Return flags
     */
    public int getFlags() {
	return flags;
    }

    /**
     * Check if a class exists
     * Applies only to package members (non-nested classes).
     */
    public boolean classExists(Identifier nm) {
	if (nm.isInner()) {
	    nm = nm.getTopName();	// just in case
	}
	Type t = Type.tClass(nm);
	try {
	    ClassDeclaration c = (ClassDeclaration)classes.get(t);
            if (c == null) {
                Package pkg = getPackage(nm.getQualifier());
                Identifier cn = nm.getName();
                ClassFile cf = pkg.getBinaryFile(cn);
                return cf != null;
            } else {
                return c.getName().equals(nm);
            }
	} catch (IOException e) {
	    return false;
	}
    }

    /**
     * Get a class, given the fully qualified class name
     */
    public ClassDeclaration getClassDeclaration(Identifier nm) {
	return getClassDeclaration(Type.tClass(nm));
    }

    /**
     * Get a class, given a type
     */
    public ClassDeclaration getClassDeclaration(Type t) {
	ClassDeclaration c = (ClassDeclaration)classes.get(t);
	if (c == null) {
	    classes.put(t, c = new ClassDeclaration(t.getClassName()));
	}
	return c;
    }

    /**
     * Get the package path for a package
     */
    public Package getPackage(Identifier pkg) throws IOException {
	Package p = (Package)packages.get(pkg);
	if (p == null) {
	    packages.put(pkg, p = new Package(path, pkg));
	}
	return p;
    }
    
    /**
     * Load a binary file
     */
    BinaryClass loadFile(ClassFile file) throws IOException {
	long tm = System.currentTimeMillis();
	InputStream input = file.getInputStream();
	BinaryClass c = null;

	try {
	    DataInputStream is = 
		new DataInputStream(new BufferedInputStream(input));
	    c = BinaryClass.load(new Environment(this, file), is, 
				 loadFileFlags());
	} catch (ClassFormatError e) {
	    throw new IllegalArgumentException("Error: bad class format" + file.getPath() + e.getMessage());
	} catch (Exception e) {
	    e.printStackTrace();
	}

	input.close();

	return c;
    }
    
    /**
     * Load the definition of a class
     */
    public void loadDefinition(ClassDeclaration c) {
	switch (c.getStatus()) {
	  case CS_UNDEFINED: {
	    Identifier nm = c.getName();
	    Package pkg;
	    try {
		pkg = getPackage(nm.getQualifier());
	    } catch (IOException e) {
	    	throw new IllegalArgumentException(
                    "Error: I/O Exception - " + e.getMessage());
	    }
	    ClassFile binfile = pkg.getBinaryFile(nm.getName());
	    if (binfile == null) {
		throw new IllegalArgumentException(
                    "Error: No binary file '" + nm.getName() + "'");
	    }

	    BinaryClass bc = null;
	    try {
		bc = loadFile(binfile);
	    } catch (IOException e) {
	    	throw new IllegalArgumentException(
                    "Error: I/O Exception - " + e.getMessage() + 
                    " in " + binfile);
	    }
	    if (bc == null) {
		throw new IllegalArgumentException(
                    "Error: Class not found in '" + binfile + "'");
	    }
	    if (!bc.getName().equals(nm)) {
		throw new IllegalArgumentException(
                    "Error: Binary file '" + nm.getName() + 
                    "'  contains " + bc.getName());
	    }

	    c.setDefinition(bc, CS_BINARY);
	    bc.loadNested(this);
	    return;
	  }
	    
	  case CS_UNDECIDED: 
	  case CS_SOURCE: 
	  default: {
	    throw new IllegalArgumentException("Error: No binary file");
	  }
	}
    }
}

