package com.sun.tools.javac.v8;
import java.io.*;

import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.tree.*;

import com.sun.tools.javac.v8.parser.*;

import com.sun.tools.javac.v8.comp.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.tree.Tree.*;


/**
 * This class could be the main entry point for GJC when GJC is used as a
 *  component in a larger software system. It provides operations to
 *  construct a new compiler, and to run a new compiler on a set of source
 *  files.
 */
public class JavaCompiler implements ClassReader.SourceCompleter {

    /**
     * The current version number as a string.
     */
    public static String version() {
        return System.getProperty("java.version");
    }

    /**
      * The last modification date as a string.
      */
    public static String date() {
        return "DD-MMMMM-YY";
    }

    /**
      * The log to be used for error reporting.
      */
    Log log;

    /**
     * The symbol table module.
     */
    Symtab syms;

    /**
     * The module for attribution checking.
     */
    Check chk;

    /**
     * The module for type inference.
     */
    Infer infer;

    /**
     * The module for identifier resolution.
     */
    Resolve rs;

    /**
     * The tree factory module.
     */
    TreeMaker make;

    /**
     * The module for the symbol table entry phases.
     */
    Enter enter;

    /**
     * The module for attribution.
     */
    Attr attr;

    /**
     * The module for code generation.
     */
    Gen gen;

    /**
     * Construct a new compiler from a log, a symbol table and a options table.
     */
    public JavaCompiler(Log log, Symtab syms, Hashtable options) {
        super();
        this.log = log;
        this.syms = syms;
        this.syms.reader.sourceCompleter = this;
        this.chk = new Check(log, syms, options);
        this.rs = new Resolve(log, syms, chk, infer);
        this.make = new TreeMaker();
        this.enter = new Enter(log, syms, rs, chk, make, null, todo);
        this.attr = new Attr(log, syms, rs, chk, infer, make, this.enter, options);
        this.gen = new Gen(log, syms, chk, rs, make, options);
        verbose = options.get("-verbose") != null;
        sourceOutput = options.get("-s") != null;
        classOutput = options.get("-retrofit") == null;
        printFlat = options.get("-printflat") != null;
        deprecation = options.get("-deprecation") != null;
        warnunchecked = options.get("-warnunchecked") != null;
        gj = options.get("-gj") != null;
        encoding = (String) options.get("-encoding");
    }

    /**
      * Construct a new compiler from a log and an options table.
      *  This will create a new symbol table module.
      */
    public static JavaCompiler make(Log log, Hashtable options) {
        try {
            {
                return new JavaCompiler(log,
                        new Symtab(new ClassReader(options),
                        new ClassWriter(options)), options);
            }
        } catch (CompletionFailure ex) {
            {
                log.error(Position.NOPOS, ex.getMessage());
                return null;
            }
        }
    }

    /**
      * Construct a new compiler from an options table.
      *  This will create a new log and symbol table module.
      */
    public static JavaCompiler make(Hashtable options) {
        return make( new Log(options.get("-prompt") != null,
                options.get("-nowarn") == null), options);
    }

    /**
      * Construct a new compiler from an empty options table.
      *  This will create a new log and symbol table module.
      */
    public static JavaCompiler make() {
        return make(Hashtable.make());
    }

    /**
      * Verbose output.
      */
    public boolean verbose;

    /**
     * Emit plain Java source files rather than class files.
     */
    public boolean sourceOutput;

    /**
     * Emit class files. This switch is always set, except for the first
     *  phase of retrofitting, where signatures are parsed.
     */
    public boolean classOutput;

    /**
     * Debug switch: Emit Java sources after inner class flattening.
     */
    public boolean printFlat;

    /**
     * Give detailed deprecation warnings.
     */
    public boolean deprecation;

    /**
     * Give detailed unchecked warnings.
     */
    public boolean warnunchecked;

    /**
     * Accept GJ input.
     */
    public boolean gj;

    /**
     * The encoding to be used for source input.
     */
    public String encoding;

    /**
     * A queue of all as yet unattributed classes.
     */
    ListBuffer todo = new ListBuffer();

    /**
     * The set of currently compiled inputfiles, needed to ensure
     *  we don't accidentally overwrite an input file when -s is set.
     *  initialized by `compile'.
     */
    Set inputFiles = Set.make();

    /**
     * The number of errors reported so far.
     */
    public int errorCount() {
        return log.nerrors;
    }

    /**
      * Try to open input stream with given name.
      *  Report an error if this fails.
      *  @param filename   The file name of the input stream to be opened.
      */
    public InputStream openSource(String filename) {
        try {
            {
                File f = new File(filename);
                inputFiles.put(f);
                return new FileInputStream(f);
            }
        } catch (IOException e) {
            {
                log.error(Position.NOPOS, "cant.read.file", filename);
                return null;
            }
        }
    }

    /**
      * Parse contents of input stream.
      *  @param filename     The name of the file from which input stream comes.
      *  @param input        The input stream to be parsed.
      */
    public TopLevel parse(String filename, InputStream input) {
        long msec = System.currentTimeMillis();
        Name prev = log.useSource(Name.fromString(filename));
        TopLevel tree = make.TopLevel(null, Tree.emptyList);
        if (input != null) {
            if (verbose) {
                printVerbose("parsing.started", filename);
            }
            try {
                {
                    Scanner scanner = new Scanner(input, log, encoding);
                    input.close();
                    Parser parser = new Parser(scanner, make, log, gj, sourceOutput);
                    tree = parser.compilationUnit();
                    if (verbose) {
                        printVerbose("parsing.done",
                                Long.toString(System.currentTimeMillis() - msec));
                    }
                }
            } catch (IOException e) {
                {
                    log.error(Position.NOPOS, "error.reading.file", filename,
                            e.toString());
                }
            }
        }
        log.useSource(prev);
        tree.sourcefile = Name.fromString(filename);
        return tree;
    }

    /**
      * Parse contents of file.
      *  @param filename     The name of the file to be parsed.
      */
    public Tree.TopLevel parse(String filename) {
        return parse(filename, openSource(filename));
    }

    /**
      * Emit plain Java source for a class.
      *  @param env    The attribution environment of the outermost class
      *                containing this class.
      *  @param cdef   The class definition to be printed.
      */
    void printSource(Env env, ClassDef cdef) throws IOException {
        File outFile = syms.writer.outputFile(cdef.sym, ".java");
        if (inputFiles.contains(outFile)) {
            log.error(cdef.pos, "source.cant.overwrite.input.file",
                    outFile.toString());
        } else {
            PrintStream out = new PrintStream(
                    new BufferedOutputStream(new FileOutputStream(outFile)));
            try {
                {
                    new Pretty(out).printUnit(env.toplevel, cdef);
                    if (verbose)
                        printVerbose("wrote.file", outFile.getPath());
                }
            }
            finally { { out.close();
                      } } }
    }

    /**
      * Generate code and emit a class file for a given class
      *  @param env    The attribution environment of the outermost class
      *                containing this class.
      *  @param cdef   The class definition from which code is generated.
      */
    void genCode(Env env, ClassDef cdef) throws IOException {
        gen.genClass(env, cdef);
        writeClass(cdef.sym);
    }

    /**
      * Emit a class file for a given class.
      *  @param c      The class from which a class file is generated.
      */
    public void writeClass(ClassSymbol c) throws IOException {
        File outFile = syms.writer.outputFile(c, ".class");
        OutputStream out = new FileOutputStream(outFile);
        try {
            {
                syms.writer.writeClassFile(out, c);
                if (verbose)
                    printVerbose("wrote.file", outFile.getPath());
            }
        }
        finally { { out.close();
                  } } }

    /**
      * Complete compiling a source file that has been accessed
      *  by the class file reader.
      *  @param c          The class the source file of which needs to be compiled.
      *  @param filename   The name of the source file.
      *  @param f          An input stream that reads the source file.
      */
    public void complete(ClassSymbol c, String filename,
            InputStream f) throws CompletionFailure {
        Tree tree = parse(filename, f);
        enter.main(List.make(tree));
        if (c.members() == null) {
            throw syms.reader.new BadClassFile(c, filename,
                    log.getLocalizedString("file.doesnt.contain.class",
                    c.fullname.toJava()));
        }
    }

    /**
      * Main method: compile a list of files, return all compiled classes
      *  @param filenames     The names of all files to be compiled.
      */
    public List compile(List filenames) throws Throwable {
        long msec = System.currentTimeMillis();
        ListBuffer classes = new ListBuffer();
        try {
            {
                inputFiles.reset();
                Symbol.reset();
                chk.compiled.reset();
                ListBuffer trees = new ListBuffer();
                for (List l = filenames; l.nonEmpty(); l = l.tail)
                    trees.append(parse((String) l.head));
                List roots = trees.toList();
                enter.main(roots);
                List rootClasses = null;
                if (sourceOutput) {
                    ListBuffer cdefs = new ListBuffer();
                    for (List l = roots; l.nonEmpty(); l = l.tail) {
                        for (List defs = ((TopLevel) l.head).defs;
                                defs.nonEmpty(); defs = defs.tail) {
                            if (defs.head instanceof ClassDef)
                                cdefs.append((ClassDef) defs.head);
                        }
                    }
                    rootClasses = cdefs.toList();
                }
                while (todo.nonEmpty()) {
                    Env env = (Env) todo.first();
                    todo.remove();
                    Tree untranslated = env.tree;
                    if (verbose)
                        printVerbose("checking.attribution",
                                env.enclClass.sym.toJava());
                    Name prev = log.useSource(env.enclClass.sym.sourcefile);
                    attr.attribClass(env.enclClass.sym);
                    if (errorCount() == 0) {
                        new Flow(log, syms, chk).analyzeDef(env.tree);
                    }
                    TreeMaker make = new TreeMaker(env.toplevel);
                    if (errorCount() == 0) {
                        env.tree = new TransTypes(log, syms,
                                make).translateTopLevelClass(env.tree);
                    }
                    if (errorCount() == 0) {
                        ClassDef cdef = null;
                        try {
                            {
                                if (sourceOutput) {
                                    cdef = (ClassDef) env.tree;
                                    if (rootClasses.contains(untranslated)) {
                                        printSource(env, cdef);
                                    }
                                } else {
                                    List cdefs = new TransInner(log, syms, rs, chk,
                                            attr, make).translateTopLevelClass(
                                            env, env.tree);
                                    for (List l = cdefs;
                                            errorCount() == 0 && l.nonEmpty();
                                            l = l.tail) {
                                        cdef = (ClassDef) l.head;
                                        if (printFlat)
                                            printSource(env, cdef);
                                        else if (classOutput)
                                            genCode(env, cdef);
                                        classes.append(cdef.sym);
                                    }
                                }
                            }
                        } catch (IOException ex) {
                            {
                                log.error(cdef.pos, "class.cant.write",
                                        cdef.sym.toJava(), ex.getMessage());
                            }
                        }
                    }
                    log.useSource(prev);
                }
            }
        } catch (Abort ex) {
            {
            }
        }
        if (verbose)
            printVerbose("total", Long.toString(System.currentTimeMillis() - msec));
        if (chk.deprecatedSource != null && !deprecation)
            noteDeprecated(chk.deprecatedSource.toString());
        if (chk.uncheckedSource != null && !warnunchecked)
            noteUnchecked(chk.uncheckedSource.toString());
        int errCount = errorCount();
        if (errCount == 1)
            printCount("error", errCount);
        else
            printCount("error.plural", errCount);
        if (log.nwarnings == 1)
            printCount("warn", log.nwarnings);
        else
            printCount("warn.plural", log.nwarnings);
        return classes.toList();
    }

    /**
      * Output for "-verbose" option.
      *  @param key The key to look up the correct internationalized string.
      *  @param arg An argument for substitution into the output string.
      */
    private void printVerbose(String key, String arg) {
        System.err.println(log.getLocalizedString("verbose." + key, arg));
    }

    /**
      * Print note that deprecated API's are used.
      */
    private void noteDeprecated(String input) {
        if (input.equals("*"))
            log.note("deprecated.plural");
        else
            log.note("deprecated.filename", input);
        log.note("deprecated.recompile");
    }

    /**
      * Print note that unchecked operations are used.
      */
    void noteUnchecked(String input) {
        if (input.equals("*"))
            log.note("unchecked.plural");
        else
            log.note("unchecked.filename", input);
        log.note("unchecked.recompile");
    }

    /**
      * Print numbers of errors and warnings.
      */
    void printCount(String kind, int count) {
        if (count != 0)
            log.println( log.getLocalizedString("count." + kind,
                    Integer.toString(count)));
    }
}
