package com.sun.tools.javac.v8;
import java.io.IOException;

import java.util.ResourceBundle;

import java.util.MissingResourceException;

import java.text.MessageFormat;

import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.tree.*;

import com.sun.tools.javac.v8.comp.*;


/**
 * This class provides a command-line interface to the GJC compiler.
 */
public class Main {

    /**
     * For testing: enter any options you want to be set implicitly
     *  here.
     */
    static String[] forcedOpts = {};

    /**
     * The name of the compiler, for use in diagnostics.
     */
    String ownName;

    /**
     * Result codes.
     */
    static final int EXIT_OK = 0;

    /**
     * Result codes.
     */
    static final int EXIT_ERROR = 1;

    /**
     * Result codes.
     */
    static final int EXIT_CMDERR = 2;

    /**
     * Result codes.
     */
    static final int EXIT_SYSERR = 3;

    /**
     * Result codes.
     */
    static final int EXIT_ABNORMAL = 4;

    /**
     * A three-column matrix describing all possible options.
     *  First column:   The option's name
     *  Second column:  The key to locate the option's argument, "" if there
     *                  is none.
     *  Third column:   The key to locate the internationalized string which
     *                  describes the option.
     *
     *  The first and second columns are important internally, since they
     *  are used to construct the table of actual options that's passed
     *  as an argument to the JavaCompiler constructor.
     */
    static String[][] standardOptions = {{"-g", "", "opt.g"}, {"-g:none", "",
    "opt.g.none"}, {"-g:{lines,vars,source}", "",
    "opt.g.lines.vars.source"}, {"-O", "", "opt.O"}, {"-nowarn", "",
    "opt.nowarn"}, {"-verbose", "", "opt.verbose"}, {"-deprecation", "",
    "opt.deprecation"}, {"-classpath", "opt.arg.path", "opt.classpath"}, {"-sourcepath", "opt.arg.path", "opt.sourcepath"}, {"-bootclasspath",
    "opt.arg.path", "opt.bootclasspath"}, {"-extdirs", "opt.arg.dirs",
    "opt.extdirs"}, {"-d", "opt.arg.directory", "opt.d"}, {"-encoding",
    "opt.arg.encoding", "opt.encoding"}, {"-target", "opt.arg.release",
    "opt.target"}};

    /**
     * Command-line options currently available.
     */
    String[][] enabledOptions;

    /**
     * Construct a compiler instance.
     */
    public Main(String name, boolean gjEnabled) {
        super();
        ownName = name;
        enabledOptions = standardOptions;
    }

    /**
      * A table of all options that's passed to the JavaCompiler constructor.
      *  If an option has an argument, the option name is mapped to the argument.
      *  If a set option has no argument, it is mapped to itself.
      */
    public Hashtable options = Hashtable.make();

    /**
     * Print a string that explains usage.
     */
    void help() {
        System.err.println(getLocalizedString("msg.usage.header", ownName));
        for (int i = 0; i < enabledOptions.length; i++) {
            if (!enabledOptions[i][0].startsWith("-X")) {
                String s = "  " + enabledOptions[i][0] + " ";
                if (!enabledOptions[i][1].equals(""))
                    s += getLocalizedString(enabledOptions[i][1]);
                System.err.print(s);
                for (int j = s.length(); j < 28; j++)
                    System.err.print(" ");
                System.err.println(getLocalizedString(enabledOptions[i][2]));
            }
        }
        System.err.println();
    }

    /**
      * Report a usage error.
      */
    void error(String key, String arg) {
        System.err.println(ownName + ": " + getLocalizedString(key, arg));
        help();
    }

    /**
      * Does argument string match option pattern?
      *  @param pattern    The option pattern, as found in the first column of
      *                    the enabledOptions matrix.
      *  @param arg        The command line argument string.
      */
    private boolean matches(String pattern, String arg) {
        int plen = pattern.length();
        int alen = arg.length();
        int ppos = Util.pos(pattern, ':', 0);
        int apos = Util.pos(arg, ':', 0);
        if (pattern.substring(0, ppos).equals(arg.substring(0, apos))) {
            if (ppos == plen && apos == alen) {
                return true;
            } else if (ppos < plen && apos < alen) {
                apos++;
                ppos++;
                if (ppos < plen && pattern.charAt(ppos) == '{') {
                    ppos++;
                    pattern = pattern.substring(0, pattern.length() - 1).concat(",");
                }
                int apos1 = apos;
                boolean found = true;
                while (found & apos1 < alen) {
                    int apos2 = Util.pos(arg, ',', apos1);
                    String option = arg.substring(apos1, apos2);
                    int ppos1 = ppos;
                    found = false;
                    while (!found && ppos1 < plen) {
                        int ppos2 = Util.pos(pattern, ',', ppos1);
                        if (pattern.substring(ppos1, ppos2).equals(option))
                            found = true;
                        ppos1 = ppos2 + 1;
                    }
                    apos1 = apos2 + 1;
                }
                return found;
            }
        }
        return false;
    }

    /**
      * Process command line arguments: store all command line options
      *  in `options' table and return all source filenames.
      *  @param args    The array of command line arguments.
      */
    public List processArgs(String[] args) {
        ListBuffer filenames = new ListBuffer();
        int ac = 0;
        while (ac < args.length) {
            String arg = args[ac];
            ac++;
            if (arg.startsWith("-")) {
                int j = 0;
                while (j < enabledOptions.length &&
                        !matches(enabledOptions[j][0], arg))
                    j++;
                if (j == enabledOptions.length) {
                    error("err.invalid.flag", arg);
                    return null;
                }
                String operand = enabledOptions[j][1];
                if (operand.length() != 0) {
                    if (ac == args.length) {
                        error("err.req.arg", arg);
                        return null;
                    }
                    operand = args[ac];
                    ac++;
                }
                if (arg.equals("-moreinfo")) {
                    Type.moreInfo = true;
                } else if (arg.equals("-version")) {
                    System.err.println(ownName + " " + JavaCompiler.version() +
                            ", " + JavaCompiler.date());
                } else {
                    if (arg.equals("-target")) {
                        if (!ClassFile.isValidTargetRelease(operand)) {
                            error("err.invalid.target", operand);
                            return null;
                        }
                    } else if (arg.length() >= 3 && arg.charAt(2) == ':') {
                        operand = arg.substring(3);
                        arg = arg.substring(0, 3);
                    }
                    options.put(arg, operand);
                }
            } else if (arg.endsWith(".java")) {
                filenames.append(arg);
            } else {
                error("err.invalid.arg", arg);
                return null;
            }
        }
        return filenames.toList();
    }

    /**
      * Programmatic interface for main function.
      * @param args    The command line parameters.
      */
    public int compile(String[] args) {
        JavaCompiler comp = null;
        try {
            {
                if (args.length == 0) {
                    help();
                    return EXIT_CMDERR;
                }
                processArgs(forcedOpts);
                List filenames = processArgs(CommandLine.parse(args));
                if (filenames == null)
                    return EXIT_CMDERR;
                comp = JavaCompiler.make(options);
                if (comp == null)
                    return EXIT_SYSERR;
                List cs = comp.compile(filenames);
                if (Util.statistics)
                    printStats();
                System.err.flush();
                if (comp.errorCount() != 0)
                    return EXIT_ERROR;
            }
        } catch (IOException ex) {
            {
                ioMessage(ex);
                return EXIT_SYSERR;
            }
        }
        catch (OutOfMemoryError ex) {
            {
                resourceMessage(ex);
                return EXIT_SYSERR;
            }
        }
        catch (FatalError ex) {
            {
                feMessage(ex);
                return EXIT_SYSERR;
            }
        }
        catch (Throwable ex) {
            {
                bugMessage(ex);
                return EXIT_ABNORMAL;
            }
        }
        return EXIT_OK;
    }

    /**
      * Print a message reporting an internal error.
      */
    void bugMessage(Throwable ex) {
        System.err.println(getLocalizedString("msg.bug", JavaCompiler.version()));
        ex.printStackTrace();
    }

    /**
      * Print a message reporting an fatal error.
      */
    void feMessage(Throwable ex) {
        System.err.println(ex.getMessage());
    }

    /**
      * Print a message reporting an input/output error.
      */
    void ioMessage(Throwable ex) {
        System.err.println(getLocalizedString("msg.io"));
        ex.printStackTrace();
    }

    /**
      * Print a message reporting an out-of-resources error.
      */
    void resourceMessage(Throwable ex) {
        System.err.println(getLocalizedString("msg.resource"));
        ex.printStackTrace();
    }

    /**
      * Print statistic information.
      */
    private static void printStats() {
        String[][] stats = {{"stats.tree", Integer.toString(Tree.count)}, {"stats.symbol",
        Integer.toString(Symbol.count)}, {"stats.type",
        Integer.toString(Type.count)}, {"stats.scope",
        Integer.toString(Scope.count)}, {"stats.scope.entry",
        Integer.toString(Scope.Entry.count)}, {"stats.item",
        Integer.toString(Items.count)}};
        if (Util.statistics) {
            for (int i = 0; i < stats.length; i++) {
                System.err.println(getLocalizedString(stats[i][0], stats[i][1]));
            }
        }
    }

    /**
      * Find a localized string in the resource bundle.
      *  @param key     The key for the localized string.
      */
    private static String getLocalizedString(String key) {
        return getText("javac." + key, null);
    }

    private static String getLocalizedString(String key, String arg0) {
        return getText("javac." + key, arg0);
    }
    private static final String javacRB = "com.sun.tools.javac.v8.resources.javac";
    private static ResourceBundle messageRB;

    /**
     * Initialize ResourceBundle.
     */
    private static void initResource() {
        try {
            {
                messageRB = ResourceBundle.getBundle(javacRB);
            }
        } catch (MissingResourceException e) {
            {
                System.out.println(e.getMessage());
                System.out.println(e.getClassName());
                throw new FatalError("Fatal Error: Resource for javac is missing");
            }
        }
    }

    /**
      * Get and format message string from resource.
      */
    private static String getText(String key, String arg0) {
        if (messageRB == null)
            initResource();
        try {
            {
                String[] args = {arg0};
                return MessageFormat.format(messageRB.getString(key), args);
            }
        } catch (MissingResourceException e) {
            {
                if (arg0 == null)
                    arg0 = "null";
                String[] args = {key, arg0};
                String msg = "javac message file broken: key={0} arguments={1}";
                return MessageFormat.format(msg, args);
            }
        }
    }
}
