package com.sun.tools.javac.v8.code;
import com.sun.tools.javac.v8.util.*;

import java.io.File;


/**
 * Generic Java classfiles have one additional attribute for classes,
 * methods and fields:
 *
 *   "Signature" (u4 attr-length, u2 signature-index)
 *
 *   A signature gives the full Java type of a method or field. When used as a
 *   class attribute, it indicates type parameters, followed by supertype,
 *   followed by all interfaces.
 *
 *     methodOrFieldSignature ::= type
 *     classSignature         ::= [ typeparams ] supertype { interfacetype }
 *
 *   The type syntax in signatures is extended as follows:
 *
 *     type       ::= ... | classtype | methodtype | typevar
 *     classtype  ::= classsig { '.' classsig }
 *     classig    ::= 'L' name [typeargs] ';'
 *     methodtype ::= [ typeparams ] '(' { type } ')' type
 *     typevar    ::= 'T' name ';'
 *     typeargs   ::= '<' type { type } '>'
 *     typeparams ::= '<' typeparam { typeparam } '>'
 *     typeparam  ::= name ':' type
 *
 *  This class defines constants used in class files as well
 *  as routines to convert between internal ``.'' and external ``/''
 *  separators in class names.
 */
public class ClassFile {

    public ClassFile() {
        super();
    }
    public static final int JAVA_MAGIC = -889275714;
    public static final int JAVA_MAJOR_VERSION = 47;
    public static final int JAVA_MINOR_VERSION = 0;
    public static final int JAVA_MIN_MAJOR_VERSION = 45;
    public static final int JAVA_MIN_MINOR_VERSION = 3;
    public static final int DEFAULT_MAJOR_VERSION = 45;
    public static final int DEFAULT_MINOR_VERSION = 3;
    public static final int CONSTANT_Utf8 = 1;
    public static final int CONSTANT_Unicode = 2;
    public static final int CONSTANT_Integer = 3;
    public static final int CONSTANT_Float = 4;
    public static final int CONSTANT_Long = 5;
    public static final int CONSTANT_Double = 6;
    public static final int CONSTANT_Class = 7;
    public static final int CONSTANT_String = 8;
    public static final int CONSTANT_Fieldref = 9;
    public static final int CONSTANT_Methodref = 10;
    public static final int CONSTANT_InterfaceMethodref = 11;
    public static final int CONSTANT_NameandType = 12;

    /**
     * What major and minor version numbers to use for the -target flag.
     * This should grow every time the minor version number accepted by
     * the VM is incremented.
     */
    static final String[] releases = {"1.1", "1.2", "1.3"};
    static final short[] majorVersions = {45, 46, 47};
    static final short[] minorVersions = {3, 0, 0};

    public static boolean isValidTargetRelease(String target) {
        for (int i = 0; i < releases.length; i++) {
            if (releases[i].equals(target))
                return true;
        }
        return false;
    }

    public static short classMajorVersion(String target) {
        for (int i = 0; i < releases.length; i++) {
            if (releases[i].equals(target))
                return majorVersions[i];
        }
        return DEFAULT_MAJOR_VERSION;
    }

    public static short classMinorVersion(String target) {
        for (int i = 0; i < releases.length; i++) {
            if (releases[i].equals(target))
                return minorVersions[i];
        }
        return DEFAULT_MINOR_VERSION;
    }

    /**
      * Return internal representation of buf[offset..offset+len-1],
      *  converting '/' to '.'.
      */
    public static byte[] internalize(byte[] buf, int offset, int len) {
        byte[] translated = new byte[len];
        for (int j = 0; j < len; j++) {
            byte b = buf[offset + j];
            if (b == '/')
                translated[j] = (byte)'.';
            else
                translated[j] = b;
        }
        return translated;
    }

    /**
      * Return internal representation of given name,
      *  converting '/' to '.'.
      */
    public static byte[] internalize(Name name) {
        return internalize(Name.names, name.index, name.len);
    }

    /**
      * Return external representation of buf[offset..offset+len-1],
      *  converting '.' to '/'.
      */
    public static byte[] externalize(byte[] buf, int offset, int len) {
        byte[] translated = new byte[len];
        for (int j = 0; j < len; j++) {
            byte b = buf[offset + j];
            if (b == '.')
                translated[j] = (byte)'/';
            else
                translated[j] = b;
        }
        return translated;
    }

    /**
      * Return external representation of given name,
      *  converting '/' to '.'.
      */
    public static byte[] externalize(Name name) {
        return externalize(Name.names, name.index, name.len);
    }

    /**
      * Return internal representation of file name s,
      *  converting File.separatorChar to '.'.
      */
    public static Name internalizeFileName(String s) {
        return Name.fromString(s.replace(File.separatorChar, '.'));
    }

    /**
      * Return external representation of file name s,
      *  converting '.' to File.separatorChar.
      */
    public static String externalizeFileName(Name n) {
        return n.toString().replace('.', File.separatorChar);
    }

    /**
      * A class for the name-and-type signature of a method or field.
      */
    public static class NameAndType {
        Name name;
        Type type;

        NameAndType(Name name, Type type) {
            super();
            this.name = name;
            this.type = type;
        }

        public boolean equals(Object other) {
            return other instanceof NameAndType &&
                    name == ((NameAndType) other).name &&
                    type.equals(((NameAndType) other).type);
        }

        public int hashCode() {
            return name.hashCode() * type.hashCode();
        }
    }
}
