package com.sun.tools.javac.v8.code;
import com.sun.tools.javac.v8.util.*;

import java.io.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.code.Type.*;


/**
 * This class provides operations to map an internal symbol table graph
 *  rooted in a ClassSymbol into a classfile.
 */
public class ClassWriter extends ClassFile implements Flags, Kinds, TypeTags {

    /**
     * Can be reassigned from outside: the output directory.
     *  If outDir = null, files are written into same directory as the sources
     *  the were generated from.
     */
    public File outDir = null;

    /**
     * Switch: verbose output.
     */
    public boolean verbose;

    /**
     * Switch: scrable private names.
     */
    public boolean scramble;

    /**
     * Switch: scrable private names.
     */
    public boolean scrambleAll;

    /**
     * Switch: retrofit mode.
     */
    public boolean retrofit;

    /**
     * Switch: emit source file attribute.
     */
    public boolean emitSourceFile;

    /**
     * Switch: target JVM version (1.1, 1.2, 1.3, ...).
     */
    public String target;

    /**
     * Target class version, derived from target JVM version.
     */
    public short majorVersion;
    public short minorVersion;

    /**
     * The initial sizes of the data and constant pool buffers.
     *  sizes are increased when buffers get full.
     */
    static final int DATA_BUF_SIZE = 65520;
    static final int POOL_BUF_SIZE = 131056;

    /**
     * An output buffer for member info.
     */
    ByteBuffer databuf = new ByteBuffer(DATA_BUF_SIZE);

    /**
     * An output buffer for the constant pool.
     */
    ByteBuffer poolbuf = new ByteBuffer(POOL_BUF_SIZE);

    /**
     * An output buffer for type signatures.
     */
    ByteBuffer sigbuf = new ByteBuffer();

    /**
     * The constant pool.
     */
    Pool pool;

    /**
     * The inner classes to be written, as a set.
     */
    Set innerClasses;

    /**
     * The inner classes to be written, as a queue where
     *  enclosing classes come first.
     */
    ListBuffer innerClassesQueue;

    /**
     * Construct a class writer, given an options table.
     */
    public ClassWriter(Hashtable options) {
        super();
        verbose = options.get("-verbose") != null;
        scramble = options.get("-scramble") != null;
        scrambleAll = options.get("-scrambleAll") != null;
        retrofit = options.get("-retrofit") != null;
        target = (String) options.get("-target");
        majorVersion = ClassFile.classMajorVersion(target);
        minorVersion = ClassFile.classMinorVersion(target);
        String detailedDebugOptions = (String) options.get("-g:");
        if (detailedDebugOptions != null) {
            emitSourceFile = Util.contains(detailedDebugOptions, "source", ',');
        } else {
            emitSourceFile = true;
        }
        String od = (String) options.get("-d");
        if (od != null)
            outDir = new File(od);
    }

    /**
      * Value of system property 'javac.dump.modifiers' is a string
      *  indicating which modifiers should be dumped for debugging:
      *    'c' -- classes
      *    'f' -- fields
      *    'i' -- innerclass attributes
      *    'm' -- methods
      *  For example, to dump everything:
      *    javac -J-Djavac.dump.modifiers=cifm MyProg.java
      */
    private static final String dumpModFlags = System.getProperty("javac.dump.modifiers");
    private static final boolean dumpClassModifiers =
            dumpModFlags != null && dumpModFlags.indexOf('c') != -1;
    private static final boolean dumpFieldModifiers =
            dumpModFlags != null && dumpModFlags.indexOf('f') != -1;
    private static final boolean dumpInnerClassModifiers =
            dumpModFlags != null && dumpModFlags.indexOf('i') != -1;
    private static final boolean dumpMethodModifiers =
            dumpModFlags != null && dumpModFlags.indexOf('m') != -1;

    /**
     * Return flags as a string, separated by " ".
     */
    public static String flagNames(int flags) {
        StringBuffer sbuf = new StringBuffer();
        int i = 0;
        int f = flags & StandardFlags;
        while (f != 0) {
            if ((f & 1) != 0)
                sbuf.append(" " + flagName[i]);
            f = f >> 1;
            i++;
        }
        return sbuf.toString();
    }
    private static final String[] flagName = {"PUBLIC", "PRIVATE", "PROTECTED", "STATIC",
    "FINAL", "SUPER", "VOLATILE", "TRANSIENT", "NATIVE", "INTERFACE", "ABSTRACT",
    "STRICTFP"};

    /**
     * Write a character into given byte buffer;
     *  byte buffer will not be grown.
     */
    void putChar(ByteBuffer buf, int op, int x) {
        buf.elems[op] = (byte)(x >> 8 & 255);
        buf.elems[op + 1] = (byte)(x & 255);
    }

    /**
      * Write an integer into given byte buffer;
      *  byte buffer will not be grown.
      */
    void putInt(ByteBuffer buf, int adr, int x) {
        buf.elems[adr] = (byte)(x >> 24 & 255);
        buf.elems[adr + 1] = (byte)(x >> 16 & 255);
        buf.elems[adr + 2] = (byte)(x >> 8 & 255);
        buf.elems[adr + 3] = (byte)(x & 255);
    }

    /**
      * Assemble signature of given type in string buffer.
      */
    void assembleSig(Type type) {
        switch (type.tag) {
        case BYTE:
            sigbuf.appendByte('B');
            break;

        case SHORT:
            sigbuf.appendByte('S');
            break;

        case CHAR:
            sigbuf.appendByte('C');
            break;

        case INT:
            sigbuf.appendByte('I');
            break;

        case LONG:
            sigbuf.appendByte('J');
            break;

        case FLOAT:
            sigbuf.appendByte('F');
            break;

        case DOUBLE:
            sigbuf.appendByte('D');
            break;

        case BOOLEAN:
            sigbuf.appendByte('Z');
            break;

        case VOID:
            sigbuf.appendByte('V');
            break;

        case CLASS:
            ClassType ct = (ClassType) type;
            ClassSymbol c = (ClassSymbol) ct.tsym;
            enterInner(c);
            if (ct.outer().allparams().nonEmpty()) {
                assembleSig(ct.outer());
                sigbuf.appendByte('.');
            }
            sigbuf.appendByte('L');
            sigbuf.appendBytes(externalize(c.flatname));
            sigbuf.appendByte(';');
            break;

        case ARRAY:
            ArrayType at = (ArrayType) type;
            sigbuf.appendByte('[');
            assembleSig(at.elemtype);
            break;

        case METHOD:
            MethodType mt = (MethodType) type;
            sigbuf.appendByte('(');
            assembleSig(mt.argtypes);
            sigbuf.appendByte(')');
            assembleSig(mt.restype);
            break;

        default:
            throw new InternalError("typeSig" + type.tag);

        }
    }

    void assembleSig(List types) {
        for (List ts = types; ts.nonEmpty(); ts = ts.tail)
            assembleSig((Type) ts.head);
    }

    /**
      * Return signature of given type
      */
    Name typeSig(Type type) {
        if (Util.checks)
            Util.assert(sigbuf.length == 0);
        assembleSig(type);
        Name n = sigbuf.toName();
        sigbuf.reset();
        return n;
    }

    /**
      * Given a type t, return the extended class name of its erasure in
      *  external representation.
      */
    public Name xClassName(Type t) {
        if (t.tag == CLASS) {
            return Name.fromUtf(externalize(t.tsym.flatName()));
        } else if (t.tag == ARRAY) {
            return typeSig(t.erasure());
        } else {
            throw new InternalError("xClassName");
        }
    }

    /**
      * Write constant pool to pool buffer.
      *  Note: during writing, constant pool
      *  might grow since some parts of constants still need to be entered.
      */
    void writePool(Pool pool) {
        int poolCountIdx = poolbuf.length;
        poolbuf.appendChar(0);
        int i = 1;
        while (i < pool.pp) {
            Object value = pool.pool[i];
            if (Util.checks)
                Util.assert(value != null);
            if (value instanceof Name) {
                poolbuf.appendByte(CONSTANT_Utf8);
                byte[] bs = ((Name) value).toUtf();
                poolbuf.appendChar(bs.length);
                poolbuf.appendBytes(bs, 0, bs.length);
            } else if (value instanceof ClassSymbol) {
                ClassSymbol c = (ClassSymbol) value;
                if (c.owner.kind == TYP)
                    pool.put(c.owner);
                poolbuf.appendByte(CONSTANT_Class);
                poolbuf.appendChar(pool.put(Name.fromUtf(externalize(c.flatname))));
                enterInner(c);
            } else if (value instanceof MethodSymbol) {
                MethodSymbol m = (MethodSymbol) value;
                poolbuf.appendByte((m.owner.flags() & INTERFACE) != 0 ?
                        CONSTANT_InterfaceMethodref : CONSTANT_Methodref);
                poolbuf.appendChar(pool.put(m.owner));
                poolbuf.appendChar(pool.put(nameType(m)));
            } else if (value instanceof VarSymbol) {
                VarSymbol v = (VarSymbol) value;
                poolbuf.appendByte(CONSTANT_Fieldref);
                poolbuf.appendChar(pool.put(v.owner));
                poolbuf.appendChar(pool.put(nameType(v)));
            } else if (value instanceof NameAndType) {
                NameAndType nt = (NameAndType) value;
                poolbuf.appendByte(CONSTANT_NameandType);
                poolbuf.appendChar(pool.put(nt.name));
                poolbuf.appendChar(pool.put(typeSig(nt.type)));
            } else if (value instanceof Integer) {
                poolbuf.appendByte(CONSTANT_Integer);
                poolbuf.appendInt(((Integer) value).intValue());
            } else if (value instanceof Long) {
                poolbuf.appendByte(CONSTANT_Long);
                poolbuf.appendLong(((Long) value).longValue());
                i++;
            } else if (value instanceof Float) {
                poolbuf.appendByte(CONSTANT_Float);
                poolbuf.appendFloat(((Float) value).floatValue());
            } else if (value instanceof Double) {
                poolbuf.appendByte(CONSTANT_Double);
                poolbuf.appendDouble(((Double) value).doubleValue());
                i++;
            } else if (value instanceof String) {
                poolbuf.appendByte(CONSTANT_String);
                poolbuf.appendChar(pool.put(Name.fromString((String) value)));
            } else if (value instanceof Type) {
                poolbuf.appendByte(CONSTANT_Class);
                poolbuf.appendChar(pool.put(xClassName((Type) value)));
            } else {
                throw new InternalError("writePool " + value);
            }
            i++;
        }
        putChar(poolbuf, poolCountIdx, pool.pp);
    }

    /**
      * Given a field, return its name.
      */
    Name fieldName(Symbol sym) {
        if (scramble && (sym.flags() & PRIVATE) != 0 || scrambleAll &&
                (sym.flags() & (PROTECTED | PUBLIC)) == 0)
            return Name.fromString("_$" + sym.name.index);
        else
            return sym.name;
    }

    /**
      * Given a symbol, return its name-and-type.
      */
    NameAndType nameType(Symbol sym) {
        return new NameAndType(fieldName(sym),
                retrofit ? sym.erasure() : sym.externalType());
    }

    /**
      * Write header for an attribute to data buffer and return
      *  position past attribute length index.
      */
    int writeAttr(Name attrName) {
        databuf.appendChar(pool.put(attrName));
        databuf.appendInt(0);
        return databuf.length;
    }

    /**
      * Fill in attribute length.
      */
    void endAttr(int index) {
        putInt(databuf, index - 4, databuf.length - index);
    }

    /**
      * Leave space for attribute count and return index for
      *  number of attributes field.
      */
    int beginAttrs() {
        databuf.appendChar(0);
        return databuf.length;
    }

    /**
      * Fill in number of attributes.
      */
    void endAttrs(int index, int count) {
        putChar(databuf, index - 2, count);
    }

    /**
      * Write flag attributes; return number of attributes written.
      */
    int writeFlagAttrs(int flags) {
        int acount = 0;
        if ((flags & DEPRECATED) != 0) {
            int alenIdx = writeAttr(Names.Deprecated);
            endAttr(alenIdx);
            acount++;
        }
        if ((flags & SYNTHETIC) != 0) {
            int alenIdx = writeAttr(Names.Synthetic);
            endAttr(alenIdx);
            acount++;
        }
        return acount;
    }

    /**
      * Write member (field or method) attributes;
      *  return number of attributes written.
      */
    int writeMemberAttrs(Symbol sym) {
        int acount = writeFlagAttrs(sym.flags());
        return acount;
    }

    /**
      * Enter an inner class into the `innerClasses' set/queue.
      */
    void enterInner(ClassSymbol c) {
        if (pool != null && c.owner.kind != PCK &&
                (innerClasses == null || !innerClasses.contains(c))) {
            if (c.owner.kind == TYP)
                enterInner((ClassSymbol) c.owner);
            pool.put(c);
            pool.put(c.name);
            if (innerClasses == null) {
                innerClasses = Set.make();
                innerClassesQueue = new ListBuffer();
                pool.put(Names.InnerClasses);
            }
            innerClasses.put(c);
            innerClassesQueue.append(c);
        }
    }

    /**
      * Write "inner classes" attribute.
      */
    void writeInnerClasses() {
        int alenIdx = writeAttr(Names.InnerClasses);
        databuf.appendChar(innerClassesQueue.length());
        for (List l = innerClassesQueue.toList(); l.nonEmpty(); l = l.tail) {
            ClassSymbol inner = (ClassSymbol) l.head;
            if (dumpInnerClassModifiers) {
                System.out.println("INNERCLASS  " + inner.name);
                System.out.println("---" + flagNames(inner.flags_field));
            }
            databuf.appendChar(pool.get(inner));
            databuf.appendChar(inner.owner.kind == TYP ? pool.get(inner.owner) : 0);
            databuf.appendChar(inner.name.len != 0 ? pool.get(inner.name) : 0);
            databuf.appendChar(inner.flags_field);
        }
        endAttr(alenIdx);
    }

    /**
      * Write field symbol, entering all references into constant pool.
      */
    void writeField(VarSymbol v) {
        databuf.appendChar(v.flags());
        if (dumpFieldModifiers) {
            System.out.println("FIELD  " + fieldName(v));
            System.out.println("---" + flagNames(v.flags()));
        }
        databuf.appendChar(pool.put(fieldName(v)));
        databuf.appendChar(pool.put(typeSig(v.erasure())));
        int acountIdx = beginAttrs();
        int acount = 0;
        if (v.constValue != null) {
            int alenIdx = writeAttr(Names.ConstantValue);
            databuf.appendChar(pool.put(v.constValue));
            endAttr(alenIdx);
            acount++;
        }
        acount += writeMemberAttrs(v);
        endAttrs(acountIdx, acount);
    }

    /**
      * Write method symbol, entering all references into constant pool.
      */
    void writeMethod(MethodSymbol m) {
        databuf.appendChar(m.flags());
        if (dumpMethodModifiers) {
            System.out.println("METHOD  " + fieldName(m));
            System.out.println("---" + flagNames(m.flags()));
        }
        databuf.appendChar(pool.put(fieldName(m)));
        databuf.appendChar(pool.put(typeSig(m.externalType())));
        int acountIdx = beginAttrs();
        int acount = 0;
        if (m.code != null) {
            int alenIdx = writeAttr(Names.Code);
            writeCode(m.code);
            m.code = null;
            endAttr(alenIdx);
            acount++;
        }
        List thrown = m.type.thrown();
        if (thrown.nonEmpty()) {
            int alenIdx = writeAttr(Names.Exceptions);
            databuf.appendChar(thrown.length());
            for (List l = thrown; l.nonEmpty(); l = l.tail)
                databuf.appendChar(pool.put(l.head));
            endAttr(alenIdx);
            acount++;
        }
        acount += writeMemberAttrs(m);
        endAttrs(acountIdx, acount);
    }

    /**
      * Write code attribute of method.
      */
    void writeCode(Code code) {
        databuf.appendChar(code.max_stack);
        databuf.appendChar(code.max_locals);
        databuf.appendInt(code.cp);
        databuf.appendBytes(code.code, 0, code.cp);
        databuf.appendChar(code.catchInfo.length());
        for (List l = code.catchInfo.toList(); l.nonEmpty(); l = l.tail) {
            for (int i = 0; i < ((char[]) l.head).length; i++)
                databuf.appendChar(((char[]) l.head)[i]);
        }
        int acountIdx = beginAttrs();
        int acount = 0;
        if (code.lineInfo.nonEmpty()) {
            int alenIdx = writeAttr(Names.LineNumberTable);
            databuf.appendChar(code.lineInfo.length());
            for (List l = code.lineInfo.reverse(); l.nonEmpty(); l = l.tail)
                for (int i = 0; i < ((char[]) l.head).length; i++)
                    databuf.appendChar(((char[]) l.head)[i]);
            endAttr(alenIdx);
            acount++;
        }
        if (code.nvars > 0) {
            int alenIdx = writeAttr(Names.LocalVariableTable);
            int nvars = code.nvars;
            int i = 0;
            int nvars1 = 0;
            while (nvars > 0) {
                if (code.lvar[i] != null) {
                    if (code.lvar_start_pc[i] != Character.MAX_VALUE)
                        nvars1++;
                    else
                        code.lvar[i] = null;
                    nvars--;
                }
                i++;
            }
            databuf.appendChar(nvars1);
            i = 0;
            while (nvars1 > 0) {
                if (code.lvar[i] != null) {
                    databuf.appendChar(code.lvar_start_pc[i]);
                    databuf.appendChar(code.lvar_length[i]);
                    databuf.appendChar(pool.put(code.lvar[i].name));
                    databuf.appendChar(pool.put(typeSig(code.lvar[i].erasure())));
                    databuf.appendChar(code.lvar_reg[i]);
                    nvars1--;
                }
                i++;
            }
            endAttr(alenIdx);
            acount++;
        }
        endAttrs(acountIdx, acount);
    }

    void writeFields(Scope.Entry e) {
        if (e != null) {
            writeFields(e.sibling);
            if (e.sym.kind == VAR)
                writeField((VarSymbol) e.sym);
        }
    }

    void writeMethods(Scope.Entry e) {
        if (e != null) {
            writeMethods(e.sibling);
            if (e.sym.kind == MTH)
                writeMethod((MethodSymbol) e.sym);
        }
    }

    /**
      * Write class `c' to outstream `out'.
      */
    public void writeClassFile(OutputStream out, ClassSymbol c) throws IOException {
        databuf.reset();
        poolbuf.reset();
        sigbuf.reset();
        pool = c.pool;
        innerClasses = null;
        innerClassesQueue = null;
        Type supertype = c.type.supertype();
        List interfaces = c.type.interfaces();
        int flags = c.flags();
        if ((flags & PROTECTED) != 0)
            flags |= PUBLIC;
        flags = flags & ClassFlags & ~STRICTFP;
        if ((flags & INTERFACE) == 0)
            flags |= ACC_SUPER;
        if (dumpClassModifiers) {
            System.out.println();
            System.out.println("CLASSFILE  " + c.fullName());
            System.out.println("---" + flagNames(flags));
        }
        databuf.appendChar(flags);
        databuf.appendChar(pool.put(c));
        databuf.appendChar(supertype.tag == CLASS ? pool.put(supertype.tsym) : 0);
        databuf.appendChar(interfaces.length());
        for (List l = interfaces; l.nonEmpty(); l = l.tail)
            databuf.appendChar(pool.put(((Type) l.head).tsym));
        int fieldsCount = 0;
        int methodsCount = 0;
        for (Scope.Entry e = c.members().elems; e != null; e = e.sibling) {
            switch (e.sym.kind) {
            case VAR:
                fieldsCount++;
                break;

            case MTH:
                methodsCount++;
                break;

            case TYP:
                enterInner((ClassSymbol) e.sym);
                break;

            default:
                throw new InternalError();

            }
        }
        databuf.appendChar(fieldsCount);
        writeFields(c.members().elems);
        databuf.appendChar(methodsCount);
        writeMethods(c.members().elems);
        int acountIdx = beginAttrs();
        int acount = 0;
        if (c.sourcefile != null && emitSourceFile) {
            int alenIdx = writeAttr(Names.SourceFile);
            String filename = c.sourcefile.toString();
            int sepIdx = filename.lastIndexOf(File.separatorChar);
            int slashIdx = filename.lastIndexOf('/');
            if (slashIdx > sepIdx)
                sepIdx = slashIdx;
            if (sepIdx >= 0)
                filename = filename.substring(sepIdx + 1);
            databuf.appendChar(c.pool.put(Name.fromString(filename)));
            endAttr(alenIdx);
            acount++;
        }
        acount += writeFlagAttrs(c.flags());
        poolbuf.appendInt(JAVA_MAGIC);
        poolbuf.appendChar(minorVersion);
        poolbuf.appendChar(majorVersion);
        writePool(c.pool);
        if (c.pool.pp > 65535)
            throw new IOException(Log.getLocalizedString("too.many.constants"));
        if (innerClasses != null) {
            writeInnerClasses();
            acount++;
        }
        endAttrs(acountIdx, acount);
        poolbuf.appendBytes(databuf.elems, 0, databuf.length);
        out.write(poolbuf.elems, 0, poolbuf.length);
        pool = c.pool = null;
    }

    /**
      * Create output file with given extension for given class.
      */
    public File outputFile(ClassSymbol c, String extension) throws IOException {
        if (outDir == null) {
            String filename = Convert.shortName(c.flatname) + extension;
            if (c.sourcefile == null)
                return new File(filename);
            String sourcedir = new File(c.sourcefile.toString()).getParent();
            if (sourcedir == null)
                return new File(filename);
            else
                return new File(sourcedir, filename);
        } else {
            return outputFile(outDir, c.flatname.toString(), extension);
        }
    }

    /**
      * open file given by a fully qualified name from root directory `outdir'.
      *  create intermediate directories if they don't exist already
      */
    File outputFile(File outdir, String name, String extension) throws IOException {
        int start = 0;
        int end = name.indexOf('.');
        while (end >= start) {
            outdir = new File(outdir, name.substring(start, end));
            if (!outdir.exists())
                outdir.mkdir();
            start = end + 1;
            end = name.indexOf('.', start);
        }
        return new File(outdir, name.substring(start) + extension);
    }
}
