package com.sun.tools.javac.v8.comp;
import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.tree.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.tree.Tree.*;

import com.sun.tools.javac.v8.code.Type.*;


/**
 * This is the main context-dependent analysis phase in GJC. It
 *  encompasses name resolution, type checking and constant folding as
 *  subtasks. Some subtasks involve auxiliary classes.
 *  @see Check
 *  @see Resolve
 *  @see ConstFold
 *  @see Infer
 */
public class Attr extends Tree.Visitor implements Flags, Kinds, TypeTags {
    Log log;
    Symtab syms;
    Resolve rs;
    Check chk;
    Infer infer;
    TreeMaker make;
    public Enter enter;
    public ConstFold cfolder;

    public Attr(Log log, Symtab syms, Resolve rs, Check chk, Infer infer,
            TreeMaker make, Enter enter, Hashtable options) {
        super();
        this.log = log;
        this.syms = syms;
        this.rs = rs;
        this.chk = chk;
        this.infer = infer;
        this.make = make;
        this.enter = enter;
        if (enter != null)
            enter.attr = this;
        this.cfolder = new ConstFold(log, syms);
        gj = options.get("-gj") != null;
        retrofit = options.get("-retrofit") != null;
    }

    /**
      * Switch: retrofit mode?
      */
    boolean retrofit;

    /**
     * Switch: GJ mode?
     */
    boolean gj;

    /**
     * Enable/disable generation of Miranda methods.
     */
    boolean miranda = true;

    /**
     * Check kind and type of given tree against protokind and prototype.
     *  If check succeeds, store type in tree and return it.
     *  If check fails, store Type.errType in tree and return it.
     *  No checks are performed if the prototype is a method type.
     *  Its not necessary in this case since we know that kind and type
     *  are correct.
     *
     *  @param tree     The tree whose kind and type is checked
     *  @param owntype  The computed type of the tree
     *  @param ownkind  The computed kind of the tree
     *  @param pkind    The expected kind (or: protokind) of the tree
     *  @param pt       The expected type (or: prototype) of the tree
     */
    Type check(Tree tree, Type owntype, int ownkind, int pkind, Type pt) {
        if (owntype.tag != ERROR && pt.tag != METHOD) {
            if ((ownkind & ~pkind) == 0) {
                if (pt.tag != NONE)
                    owntype = chk.checkType(tree.pos, owntype, pt);
            } else {
                log.error(tree.pos, "unexpected.type", Resolve.kindNames(pkind),
                        Resolve.kindName(ownkind));
                owntype = Type.errType;
            }
        }
        tree.type = owntype;
        return owntype;
    }

    /**
      * Is given blank final variable assignable, i.e. in a scope where it
      *  may be assigned to even though it is final?
      *  @param v      The blank final variable.
      *  @param env    The current environment.
      */
    static boolean isAssignableAsBlankFinal(VarSymbol v, Env env) {
        Symbol owner = ((AttrContext) env.info).scope.owner;
        return v.owner == owner || (owner.name == Names.init || owner.kind == VAR ||
                (owner.flags() & BLOCK) != 0) && v.owner == owner.owner &&
                (v.flags() & STATIC) != 0 == Resolve.isStatic(env);
    }

    /**
      * Check that variable can be assigned to.
      *  @param pos    The current source code position.
      *  @param v      The assigned varaible
      *  @param base   If the variable is referred to in a Select, the part
      *                to the left of the `.', null otherwise.
      *  @env          The current environment.
      */
    void checkAssignable(int pos, VarSymbol v, Tree base, Env env) {
        if ((v.flags() & FINAL) != 0 && ((v.flags() & HASINIT) != 0 ||
                !((base == null || base.tag == Tree.IDENT &&
                TreeInfo.name(base) == Names._this) &&
                isAssignableAsBlankFinal(v, env)))) {
            log.error(pos, "cant.assign.val.to.final.var", v.toJava());
        }
    }

    /**
      * Does tree represent a static reference to an identifier?
      *  It is assumed that tree is either a SELECT or an IDENT.
      *  We have to weed out selects from non-type names here.
      *  @param tree    The candidate tree.
      */
    boolean isStaticReference(Tree tree) {
        if (tree.tag == Tree.SELECT) {
            Symbol lsym = TreeInfo.symbol(((Select) tree).selected);
            if (lsym == null || lsym.kind != TYP) {
                return false;
            }
        }
        return true;
    }

    /**
      * The current `this' symbol.
      *  @param env    The current environment.
      */
    Symbol thisSym(Env env) {
        return rs.resolveSelf(Position.NOPOS, env, env.enclClass.sym,
                Names._this, true);
    }

    /**
      * Add an abstract method definition to a class definition
      *  which implicitly implements a method defined in some interface
      *  implemented by the class. These methods are called "Miranda methods".
      *  Enter the newly created method into its enclosing class scope.
      *
      *  @param cd     The class definition to which the Miranda method is added.
      *  @param m      The interface method symbol for which a Miranda method
      *                is added.
      *  @param env    The current environment.
      */
    private void addAbstractMethod(ClassDef cd, MethodSymbol m, Env env) {
        MethodSymbol absMeth = new MethodSymbol(m.flags() | IPROXY, m.name,
                cd.sym.type.memberType(m), cd.sym);
        Tree def = make.at(cd.pos).MethodDef(absMeth, null);
        cd.defs = cd.defs.prepend(def);
        enter.phase2.memberEnter(def, env);
    }

    /**
      * Add abstract method definitions for all methods defined in one of
      *  the interfaces of a given class,
      *  provided they are not already implemented in the class.
      *
      *  @param c      The class whose interfaces are searched for methods
      *                for which Miranda methods should be added.
      *  @env          The environment current at the class definition.
      *                This points (via env.tree) to the class definition
      *                to which Miranda methods are added.
      */
    void implementInterfaceMethods(ClassSymbol c, Env env) {
        ClassDef cd = (ClassDef) env.tree;
        for (List l = c.type.interfaces(); l.nonEmpty(); l = l.tail) {
            ClassSymbol i = (ClassSymbol)((Type) l.head).tsym;
            for (Scope.Entry e = i.members().elems; e != null; e = e.sibling) {
                if (e.sym.kind == MTH && (e.sym.flags() & STATIC) == 0) {
                    MethodSymbol absMeth = (MethodSymbol) e.sym;
                    MethodSymbol implMeth = absMeth.implementation(cd.sym);
                    if (implMeth == null)
                        addAbstractMethod(cd, absMeth, env);
                }
            }
            implementInterfaceMethods(i, env);
        }
    }

    /**
      * Visitor argument: the current environment.
      */
    Env env;

    /**
     * Visitor argument: the currently expected proto-kind.
     */
    int pkind;

    /**
     * Visitor argument: the currently expected proto-type.
     */
    Type pt;

    /**
     * Visitor result: the computed type.
     */
    Type result;

    /**
     * Visitor method: attribute a tree, catching any completion failure
     *  exceptions. Return the tree's type.
     *
     *  @param tree    The tree to be visited.
     *  @param env     The environment visitor argument.
     *  @param pkind   The protokind visitor argument.
     * @  @param pt      The prototype visitor argument.
     */
    Type attribTree(Tree tree, Env env, int pkind, Type pt) {
        Env prevEnv = this.env;
        int prevPkind = this.pkind;
        Type prevPt = this.pt;
        try {
            {
                this.env = env;
                this.pkind = pkind;
                this.pt = pt;
                tree.visit(this);
                this.env = prevEnv;
                this.pkind = prevPkind;
                this.pt = prevPt;
                return result;
            }
        } catch (CompletionFailure ex) {
            {
                this.env = prevEnv;
                this.pkind = prevPkind;
                this.pt = prevPt;
                return chk.completionError(tree.pos, ex);
            }
        }
    }

    /**
      * Derived visitor method: attribute an expression tree.
      */
    Type attribExpr(Tree tree, Env env, Type pt) {
        return attribTree(tree, env, VAL, pt);
    }

    /**
      * Derived visitor method: attribute an expression tree with
      *  no constraints on the computed type.
      */
    Type attribExpr(Tree tree, Env env) {
        return attribTree(tree, env, VAL, Type.noType);
    }

    /**
      * Derived visitor method: attribute a type tree.
      */
    Type attribType(Tree tree, Env env) {
        return attribTree(tree, env, TYP, Type.noType);
    }

    /**
      * Derived visitor method: attribute a statement or definition tree.
      */
    Type attribStat(Tree tree, Env env) {
        return attribTree(tree, env, NIL, Type.noType);
    }

    /**
      * Attribute a list of expressions, returning a list of types.
      */
    List attribExprs(List trees, Env env, Type pt) {
        ListBuffer ts = new ListBuffer();
        for (List l = trees; l.nonEmpty(); l = l.tail)
            ts.append(attribExpr((Tree) l.head, env, pt));
        return ts.toList();
    }

    /**
      * Attribute a list of statements, returning nothing.
      */
    void attribStats(List trees, Env env) {
        for (List l = trees; l.nonEmpty(); l = l.tail)
            attribStat((Tree) l.head, env);
    }

    /**
      * Attribute the arguments in a method call, returning a list of types.
      */
    List attribArgs(List trees, Env env) {
        ListBuffer argtypes = new ListBuffer();
        for (List l = trees; l.nonEmpty(); l = l.tail)
            argtypes.append( chk.checkNonVoid(((Tree) l.head).pos,
                    attribTree((Tree) l.head, env, VAL, Type.noType)));
        return argtypes.toList();
    }

    /**
      * Attribute type reference in an `extends' or `implements' clause.
      *
      *  @param tree             The tree making up the type reference.
      *  @param env              The environment current at the reference.
      *  @param interfaceFlag    This is INTERFACE if we expect to reference an
      *                          interface, 0 if we expect to reference a class.
      */
    Type attribBase(Tree tree, Env env, int interfaceFlag) {
        Type t = chk.checkClassType(tree.pos, attribType(tree, env));
        if ((t.tsym.flags() & INTERFACE) != interfaceFlag) {
            if (interfaceFlag == 0)
                log.error(tree.pos, "no.intf.expected.here");
            else
                log.error(tree.pos, "intf.expected.here");
            return Type.errType;
        }
        if ((t.tsym.flags() & FINAL) != 0) {
            log.error(tree.pos, "cant.inherit.from.final", t.tsym.toJava());
        }
        return t;
    }

    /**
      * Visitor method for class definitions.
      */
    public void _case(ClassDef tree) {
        if ((((AttrContext) env.info).scope.owner.kind & (VAR | MTH)) != 0)
            enter.classEnter(tree, env);
        ClassSymbol c = tree.sym;
        if (c == null) {
            result = null;
        } else {
            c.complete();
            if (((AttrContext) env.info).isSelfCall &&
                    env.tree.tag == Tree.NEWCLASS &&
                    ((NewClass) env.tree).encl == null) {
                c.flags_field |= NOOUTERTHIS;
            }
            attribClass(c);
            result = tree.type = c.type;
        }
    }

    /**
      * Visitor method for method definitions.
      */
    public void _case(MethodDef tree) {
        MethodSymbol m = tree.sym;
        chk.checkOverride(tree.pos, m);
        Env localEnv = enter.methodEnv((MethodDef) tree, env);
        for (List l = tree.params; l.nonEmpty(); l = l.tail)
            attribStat((Tree) l.head, localEnv);
        chk.validate(tree.restype);
        for (List l = tree.thrown; l.nonEmpty(); l = l.tail)
            chk.checkType(((Tree) l.head).pos, ((Tree) l.head).type,
                    syms.throwableType);
        ClassSymbol owner = env.enclClass.sym;
        if (tree.body == null) {
            if ((owner.flags() & INTERFACE) == 0 &&
                    (tree.flags & (ABSTRACT | NATIVE)) == 0 && !retrofit)
                log.error(tree.pos, "missing.meth.body.or.decl.abstract");
        } else if ((owner.flags() & INTERFACE) != 0) {
            log.error(tree.pos, "intf.meth.cant.have.body");
        } else if ((tree.flags & ABSTRACT) != 0) {
            log.error(tree.pos, "abstract.meth.cant.have.body");
        } else if ((tree.flags & NATIVE) != 0) {
            log.error(tree.pos, "native.meth.cant.have.body");
        } else {
            if (tree.name == Names.init && owner.type != syms.objectType) {
                Block body = tree.body;
                if (body.stats.isEmpty() ||
                        !TreeInfo.isSelfCall((Tree) body.stats.head)) {
                    body.stats = body.stats.prepend(
                            Enter.SuperCall(make.at(body.pos), VarDef.emptyList,
                            false));
                }
            }
            attribStat(tree.body, localEnv);
        }
        ((AttrContext) localEnv.info).scope.leave();
        result = tree.type = m.type;
    }

    /**
      * Visitor method for variable definitions.
      */
    public void _case(VarDef tree) {
        if (((AttrContext) env.info).scope.owner.kind == MTH)
            enter.phase2.memberEnter(tree, env);
        chk.validate(tree.vartype);
        VarSymbol v = tree.sym;
        if (tree.init != null) {
            v.pos = Position.MAXPOS;
            if ((v.flags_field & FINAL) != 0) {
                evalInit(v);
            } else {
                Type itype = attribExpr(tree.init, enter.initEnv((VarDef) tree, env),
                        v.type);
            }
            v.pos = tree.pos;
        }
        result = tree.type = v.type;
    }

    /**
      * Visitor method for blocks.
      */
    public void _case(Block tree) {
        Env localEnv = env.dup(tree,
                ((AttrContext) env.info).dup(((AttrContext) env.info).scope.dup()));
        if (((AttrContext) env.info).scope.owner.kind == TYP) {
            ((AttrContext) localEnv.info).scope.owner =
                    new MethodSymbol(tree.flags | BLOCK, Names.empty, null,
                    ((AttrContext) env.info).scope.owner);
            if ((tree.flags & STATIC) != 0)
                ((AttrContext) localEnv.info).staticLevel++;
        }
        attribStats(tree.stats, localEnv);
        ((AttrContext) localEnv.info).scope.leave();
        result = null;
    }

    /**
      * Visitor method for do-loops.
      */
    public void _case(DoLoop tree) {
        attribStat(tree.body, env.dup(tree));
        attribExpr(tree.cond, env, Type.booleanType);
        result = null;
    }

    /**
      * Visitor method for while-loops.
      */
    public void _case(WhileLoop tree) {
        attribExpr(tree.cond, env, Type.booleanType);
        attribStat(tree.body, env.dup(tree));
        result = null;
    }

    /**
      * Visitor method for for-loops.
      */
    public void _case(ForLoop tree) {
        Env loopEnv = env.dup(env.tree,
                ((AttrContext) env.info).dup(((AttrContext) env.info).scope.dup()));
        attribStats(tree.init, loopEnv);
        if (tree.cond != null)
            attribExpr(tree.cond, loopEnv, Type.booleanType);
        loopEnv.tree = tree;
        attribStats(tree.step, loopEnv);
        attribStat(tree.body, loopEnv);
        ((AttrContext) loopEnv.info).scope.leave();
        result = null;
    }

    /**
      * Visitor method for labelled statements.
      */
    public void _case(Labelled tree) {
        Env env1 = env;
        while (env1 != null) {
            if (env1.tree.tag == Tree.LABELLED &&
                    ((Labelled) env1.tree).label == tree.label) {
                log.error(tree.pos, "label.already.in.use", tree.label.toJava());
                break;
            }
            env1 = env1.next;
        }
        attribStat(tree.body, env.dup(tree));
        result = null;
    }

    /**
      * Visitor method for switch statements.
      */
    public void _case(Switch tree) {
        Type seltype = attribExpr(tree.selector, env, Type.intType);
        Env switchEnv = env.dup(tree,
                ((AttrContext) env.info).dup(((AttrContext) env.info).scope.dup()));
        Set labels = Set.make();
        boolean hasDefault = false;
        for (List l = tree.cases; l.nonEmpty(); l = l.tail) {
            Case c = (Case) l.head;
            if (c.pat != null) {
                Type pattype = attribExpr(c.pat, switchEnv, Type.intType);
                if (pattype.tag != ERROR) {
                    if (pattype.constValue == null) {
                        log.error(c.pat.pos, "const.expr.req");
                    } else if (labels.contains(pattype.constValue)) {
                        log.error(c.pos, "duplicate.case.label");
                    } else {
                        chk.checkType(c.pat.pos, pattype, seltype);
                        labels.put(pattype.constValue);
                    }
                }
            } else if (hasDefault) {
                log.error(c.pos, "duplicate.default.label");
            } else {
                hasDefault = true;
            }
            attribStats(c.stats, switchEnv);
        }
        ((AttrContext) switchEnv.info).scope.leave();
        result = null;
    }

    /**
      * Visitor method for synchronized statements.
      */
    public void _case(Synchronized tree) {
        chk.checkRefType(tree.pos, attribExpr(tree.lock, env, syms.objectType));
        attribStat(tree.body, env);
        result = null;
    }

    /**
      * Visitor method for try statements.
      */
    public void _case(Try tree) {
        attribStat(tree.body, env.dup(tree, ((AttrContext) env.info).dup()));
        for (List l = tree.catchers; l.nonEmpty(); l = l.tail) {
            Catch c = (Catch) l.head;
            Env catchEnv = env.dup(c,
                    ((AttrContext) env.info).dup(
                    ((AttrContext) env.info).scope.dup()));
            Type ctype = attribStat(c.param, catchEnv);
            chk.checkType(c.param.vartype.pos, ctype, syms.throwableType);
            attribStat(c.body, catchEnv);
            ((AttrContext) catchEnv.info).scope.leave();
        }
        if (tree.finalizer != null)
            attribStat(tree.finalizer, env);
        result = null;
    }

    /**
      * Visitor method for conditional statements and expressions.
      */
    public void _case(Conditional tree) {
        attribExpr(tree.cond, env, Type.booleanType);
        attribExpr(tree.thenpart, env, pt);
        if (tree.elsepart != null)
            attribExpr(tree.elsepart, env, pt);
        if (tree.tag == Tree.CONDSTAT)
            result = null;
        else
            result = check(tree,
                    condType(tree.pos, tree.cond.type, tree.thenpart.type,
                    tree.elsepart.type), VAL, pkind, pt);
    }

    /**
      * Compute the type of a conditional expression, after
      *  checking that it exists. See Spec ???.
      *
      *  @param pos      The source position to be used for error diagnostics.
      *  @param condtype The type of the expression's condition.
      *  @param thentype The type of the expression's then-part.
      *  @param elsetype The type of the expression's else-part.
      */
    private Type condType(int pos, Type condtype, Type thentype, Type elsetype) {
        Type ctype = condType1(pos, condtype, thentype, elsetype);
        if (condtype.constValue != null && thentype.constValue != null &&
                elsetype.constValue != null)
            return ((Number) condtype.constValue).intValue() != 0 ? thentype :
                    elsetype;
        else
            return ctype;
    }

    /**
      * Compute the type of a conditional expression, after
      *  checking that it exists.  Does not take into
      *  account the special case where condition and both arms are constants.
      *
      *  @param pos      The source position to be used for error diagnostics.
      *  @param condtype The type of the expression's condition.
      *  @param thentype The type of the expression's then-part.
      *  @param elsetype The type of the expression's else-part.
      */
    private Type condType1(int pos, Type condtype, Type thentype, Type elsetype) {
        if (thentype.tag < INT && elsetype.tag == INT &&
                elsetype.isAssignable(thentype))
            return thentype.baseType();
        else if (elsetype.tag < INT && thentype.tag == INT &&
                thentype.isAssignable(elsetype))
            return elsetype.baseType();
        else if (thentype.tag <= DOUBLE && elsetype.tag <= DOUBLE)
            for (int i = BYTE; i <= DOUBLE; i++) {
                Type candidate = Type.typeOfTag[i];
                if (thentype.isSubType(candidate) && elsetype.isSubType(candidate))
                    return candidate;
            }
        if (thentype.tsym == syms.stringType.tsym &&
                elsetype.tsym == syms.stringType.tsym)
            return syms.stringType;
        else if (thentype.isSubType(elsetype))
            return elsetype.baseType();
        else {
            chk.checkType(pos, elsetype, thentype);
            return thentype.baseType();
        }
    }

    /**
      * Visitor method for expression statements.
      */
    public void _case(Exec tree) {
        attribExpr(tree.expr, env);
        result = null;
    }

    /**
      * Visitor method for break statements.
      */
    public void _case(Break tree) {
        tree.target = findJumpTarget(tree.pos, tree.tag, tree.label, env);
        result = null;
    }

    /**
      * Visitor method for continue statements.
      */
    public void _case(Continue tree) {
        tree.target = findJumpTarget(tree.pos, tree.tag, tree.label, env);
        result = null;
    }

    /**
      * Return the target of a break or continue statement, if it exists,
      *  report an error if not.
      *  Note: The target of a labelled break or continue is the
      *  (non-labelled) statement tree referred to by the label,
      *  not the tree representing the labelled statement itself.
      *
      *  @param pos     The position to be used for error diagnostics
      *  @param tag     The tag of the jump statement. This is either
      *                 Tree.BREAK or Tree.CONTINUE.
      *  @param label   The label of the jump statement, or null if no
      *                 label is given.
      *  @param env     The environment current at the jump statement.
      */
    private Tree findJumpTarget(int pos, int tag, Name label, Env env) {
        Env env1 = env;
        LOOP:
        while (env1 != null) {
            switch (env1.tree.tag) {
            case Tree.LABELLED:
                Labelled labelled = (Labelled) env1.tree;
                if (label == labelled.label) {
                    Tree target = TreeInfo.referencedStatement(labelled);
                    if (tag == Tree.CONTINUE && target.tag != Tree.DOLOOP &&
                            target.tag != Tree.WHILELOOP &&
                            target.tag != Tree.FORLOOP)
                        log.error(pos, "not.loop.label", label.toJava());
                    return target;
                }
                break;

            case Tree.DOLOOP:

            case Tree.WHILELOOP:

            case Tree.FORLOOP:
                if (label == null)
                    return env1.tree;
                break;

            case Tree.SWITCH:
                if (label == null && tag == Tree.BREAK)
                    return env1.tree;
                break;

            case Tree.METHODDEF:
                break LOOP;

            default:

            }
            env1 = env1.next;
        }
        if (label != null)
            log.error(pos, "undef.label", label.toJava());
        else if (tag == Tree.CONTINUE)
            log.error(pos, "cont.outside.loop");
        else
            log.error(pos, "break.outside.switch.loop");
        return null;
    }

    /**
      * Visitor method for return statements.
      */
    public void _case(Return tree) {
        if (env.enclMethod == null || env.enclMethod.sym.owner != env.enclClass.sym) {
            log.error(tree.pos, "ret.outside.meth");
        } else {
            Symbol m = env.enclMethod.sym;
            if (m.type.restype().tag == VOID) {
                if (tree.expr != null)
                    log.error(tree.expr.pos, "cant.ret.val.from.meth.decl.void");
            } else if (tree.expr == null) {
                log.error(tree.pos, "missing.ret.val");
            } else {
                attribExpr(tree.expr, env, m.type.restype());
            }
        }
        result = null;
    }

    /**
      * Visitor method for throw statements.
      */
    public void _case(Throw tree) {
        Type t = attribExpr(tree.expr, env, syms.throwableType);
        result = null;
    }

    /**
      * Visitor method for method invocations.
      *  NOTE: The method part of an application will have in its type field
      *        the return type of the method, not the method's type itself!
      */
    public void _case(Apply tree) {
        Type owntype = Type.errType;
        Env localEnv = env;
        List argtypes;
        Name methName = TreeInfo.name(tree.meth);
        boolean isConstructorCall =
                methName == Names._this || methName == Names._super;
        if (isConstructorCall) {
            if (checkFirstConstructorStat(tree, env)) {
                localEnv = env.dup(env.tree, ((AttrContext) env.info).dup());
                ((AttrContext) localEnv.info).isSelfCall = true;
                argtypes = attribArgs(tree.args, localEnv);
                Type site = env.enclClass.sym.type;
                if (methName == Names._super)
                    site = site.supertype();
                if (site.tag == CLASS) {
                    if (tree.meth.tag == Tree.SELECT) {
                        Tree qualifier = ((Select) tree.meth).selected;
                        Symbol qsym = TreeInfo.symbol(qualifier);
                        if (qsym != null && qsym.kind == TYP)
                            tree.meth = make.at(tree.pos).Ident(methName);
                        else if (site.outer().tag == CLASS)
                            attribExpr(qualifier, localEnv, site.outer());
                        else
                            log.error(qualifier.pos, "illegal.qual.not.icls",
                                    site.tsym.toJava());
                    }
                    boolean selectSuperPrev =
                            ((AttrContext) localEnv.info).selectSuper;
                    ((AttrContext) localEnv.info).selectSuper = true;
                    Symbol sym = rs.resolveConstructor(tree.meth.pos, localEnv, site,
                            argtypes);
                    ((AttrContext) localEnv.info).selectSuper = selectSuperPrev;
                    if (sym == env.enclMethod.sym) {
                        log.error(tree.pos, "recursive.ctor.invocation");
                    }
                    TreeInfo.setSymbol(tree.meth, sym);
                    checkId(tree.meth, site, sym, env, MTH, anyMethodType);
                }
            }
            result = Type.voidType;
        } else {
            argtypes = attribArgs(tree.args, localEnv);
            List saved = methTemplateSupply.elems;
            Type mpt = newMethTemplate(argtypes);
            owntype = attribExpr(tree.meth, localEnv, mpt);
            methTemplateSupply.elems = saved;
            result = check(tree, owntype, VAL, pkind, pt);
        }
    }

    /**
      * Check that given application node appears as first statement
      *  in a constructor call.
      *  @param tree   The application node
      *  @param env    The environment current at the application.
      */
    boolean checkFirstConstructorStat(Apply tree, Env env) {
        MethodDef enclMethod = env.enclMethod;
        if (enclMethod != null && enclMethod.name == Names.init) {
            Block body = (Block) enclMethod.body;
            if (((Tree) body.stats.head).tag == Tree.EXEC &&
                    ((Exec) body.stats.head).expr == tree)
                return true;
        }
        log.error(tree.pos, "call.must.be.first.stmt.in.ctor",
                TreeInfo.name(tree.meth).toJava());
        return false;
    }

    /**
      * A type constant standing for any method type. Passed as a proto-type
      *  to checkId.
      */
    Type anyMethodType = new MethodType(null, null, null);

    /**
     * Optimization: To avoid allocating a new methodtype for every
     *  attribution of an Apply node, we use a reservoir.
     */
    ListBuffer methTemplateSupply = new ListBuffer();

    /**
     * Obtain an otherwise unused method type with given argument types.
     *  Take it from the reservoir if non-empty, or create a new one.
     */
    Type newMethTemplate(List argtypes) {
        if (methTemplateSupply.elems == methTemplateSupply.last)
            methTemplateSupply.append(new MethodType(null, null, null));
        MethodType mt = (MethodType) methTemplateSupply.elems.head;
        methTemplateSupply.elems = methTemplateSupply.elems.tail;
        mt.argtypes = argtypes;
        return mt;
    }

    /**
      * Visitor method for class instance creation expressions.
      */
    public void _case(NewClass tree) {
        Type owntype = Type.errType;
        ClassDef cdef = tree.def;
        Type encltype = null;
        Tree clazz = tree.clazz;
        Tree clazzid = clazz;
        Tree clazzid1 = clazzid;
        if (tree.encl != null) {
            encltype = attribExpr(tree.encl, env);
            if (encltype.tag == CLASS) {
                clazzid1 = make.at(clazz.pos).Select(make.Type(encltype),
                        ((Ident) clazzid).name);
                clazz = clazzid1;
            }
        }
        Type clazztype = chk.checkClassType(tree.clazz.pos, attribType(clazz, env));
        chk.validate(clazz);
        if (tree.encl != null) {
            tree.clazz.type = clazztype;
            TreeInfo.setSymbol(clazzid, TreeInfo.symbol(clazzid1));
            clazzid.type = ((Ident) clazzid).sym.type;
        } else if ((clazztype.tsym.flags() & INTERFACE) == 0 &&
                clazztype.outer().tag == CLASS) {
            rs.resolveSelf(tree.pos, env, clazztype.outer().tsym, Names._this, false);
        }
        List argtypes = attribArgs(tree.args, env);
        if (clazztype.tag == CLASS) {
            if (cdef == null &&
                    (clazztype.tsym.flags() & (ABSTRACT | INTERFACE)) != 0) {
                log.error(tree.pos, "abstract.cant.be.instantiated",
                        clazztype.tsym.toJava());
            } else if (cdef != null && (clazztype.tsym.flags() & INTERFACE) != 0) {
                if (argtypes.nonEmpty()) {
                    log.error(tree.pos, "anon.class.impl.intf.no.args");
                    argtypes = Type.emptyList;
                } else if (tree.encl != null) {
                    log.error(tree.pos, "anon.class.impl.intf.no.qual.for.new");
                }
            } else {
                boolean selectSuperPrev = ((AttrContext) env.info).selectSuper;
                if (cdef != null)
                    ((AttrContext) env.info).selectSuper = true;
                tree.constructor =
                        rs.resolveConstructor(tree.pos, env, clazztype, argtypes);
                ((AttrContext) env.info).selectSuper = selectSuperPrev;
            }
            if (cdef != null) {
                if (Resolve.isStatic(env))
                    cdef.flags |= STATIC;
                if ((clazztype.tsym.flags() & INTERFACE) != 0) {
                    cdef.implementing = List.make(clazz);
                } else {
                    cdef.extending = clazz;
                }
                attribStat(cdef, env.dup(tree));
                if (tree.encl != null) {
                    tree.args = tree.args.prepend(tree.encl);
                    argtypes = argtypes.prepend(encltype);
                    tree.encl = null;
                    encltype = null;
                }
                clazztype = cdef.sym.type;
                tree.constructor =
                        rs.resolveConstructor(tree.pos, env, clazztype, argtypes);
            } else if (((AttrContext) env.info).isSelfCall &&
                    clazztype.tsym.hasOuterInstance() && tree.encl == null) {
                chk.earlyRefError(tree.pos, thisSym(env));
            }
            if (tree.constructor != null && tree.constructor.kind == MTH)
                owntype = clazztype;
        }
        result = check(tree, owntype, VAL, pkind, pt);
    }

    /**
      * Visitor method for array creation expressions.
      */
    public void _case(NewArray tree) {
        Type owntype = Type.errType;
        Type elemtype;
        if (tree.elemtype != null) {
            elemtype = attribType(tree.elemtype, env);
            chk.validate(tree.elemtype);
            owntype = elemtype;
            for (List l = tree.dims; l.nonEmpty(); l = l.tail) {
                attribExpr((Tree) l.head, env, Type.intType);
                owntype = new ArrayType(owntype);
            }
        } else {
            if (pt.tag == ARRAY) {
                elemtype = pt.elemtype();
            } else {
                if (pt.tag != ERROR) {
                    log.error(tree.pos, "illegal.initializer.for.type", pt.toJava());
                }
                elemtype = Type.errType;
            }
        }
        if (tree.elems != null) {
            attribExprs(tree.elems, env, elemtype);
            owntype = new ArrayType(elemtype);
        }
        result = check(tree, owntype, VAL, pkind, pt);
    }

    /**
      * Visitor method for assignment expressions.
      */
    public void _case(Assign tree) {
        Type owntype = attribTree(tree.lhs, env.dup(tree), VAR, pt);
        attribExpr(tree.rhs, env, owntype);
        result = check(tree, owntype, VAL, pkind, pt);
    }

    /**
      * Visitor method for assignment operation expressions.
      */
    public void _case(Assignop tree) {
        List argtypes = List.make(attribTree(tree.lhs, env, VAR, Type.noType),
                attribExpr(tree.rhs, env));
        Symbol operator = tree.operator =
                rs.resolveOperator(tree.pos, tree.tag - Tree.ASGOffset, env,
                argtypes);
        Type owntype = (Type) argtypes.head;
        if (operator.kind == MTH) {
            if (owntype.tag <= DOUBLE)
                chk.checkCastable(tree.rhs.pos, operator.type.restype(), owntype);
            else
                chk.checkType(tree.rhs.pos, operator.type.restype(), owntype);
        }
        result = check(tree, owntype, VAL, pkind, pt);
    }

    /**
      * Visitor method for (unary and binary) operation expressions.
      */
    public void _case(Operation tree) {
        List argtypes;
        if (Tree.PREINC <= tree.tag && tree.tag <= Tree.POSTDEC)
            argtypes = Type.emptyList.prepend(
                    attribTree((Tree) tree.args.head, env, VAR, Type.noType));
        else {
            argtypes = attribArgs(tree.args, env);
        }
        Symbol operator = tree.operator =
                rs.resolveOperator(tree.pos, tree.tag, env, argtypes);
        Type owntype = Type.errType;
        if (operator.kind == MTH) {
            owntype = operator.type.restype();
            int opc = ((OperatorSymbol) operator).opcode;
            List l = argtypes;
            while (l.nonEmpty() && ((Type) l.head).constValue != null)
                l = l.tail;
            if (l.isEmpty()) {
                Type ctype = cfolder.fold(opc, argtypes);
                if (ctype != null) {
                    owntype = cfolder.coerce(ctype, owntype);
                    for (List args = tree.args; args.nonEmpty(); args = args.tail) {
                        if (((Tree) args.head).type.tsym == syms.stringType.tsym) {
                            ((Tree) args.head).type = syms.stringType;
                        }
                    }
                }
            }
            if (opc == ByteCodes.if_acmpeq || opc == ByteCodes.if_acmpne) {
                if (!((Type) argtypes.head).isCastable(
                        ((Type) argtypes.tail.head).erasure()) &&
                        !((Type) argtypes.tail.head).isCastable(
                        ((Type) argtypes.head).erasure())) {
                    log.error(tree.pos, "incomparable.types",
                            ((Type) argtypes.head).toJava(),
                            ((Type) argtypes.tail.head).toJava());
                } else {
                    chk.checkCompatible(tree.pos, (Type) argtypes.head,
                            (Type) argtypes.tail.head);
                }
            }
        }
        result = check(tree, owntype, VAL, pkind, pt);
    }

    /**
      * Visitor method for type cast expressions.
      */
    public void _case(TypeCast tree) {
        Type clazztype = attribType(tree.clazz, env);
        Type exprtype = attribExpr(tree.expr, env);
        Type owntype = chk.checkCastable(tree.expr.pos, exprtype, clazztype);
        if (exprtype.constValue != null)
            owntype = cfolder.coerce(exprtype, owntype);
        result = check(tree, owntype, VAL, pkind, pt);
    }

    /**
      * Visitor method for type test expressions.
      */
    public void _case(TypeTest tree) {
        Type exprtype = attribExpr(tree.expr, env);
        Type clazztype = chk.checkClassOrArrayType(tree.clazz.pos,
                attribType(tree.clazz, env));
        chk.checkCastable(tree.expr.pos, exprtype, clazztype);
        result = check(tree, Type.booleanType, VAL, pkind, pt);
    }

    /**
      * Visitor method for indexed expressions.
      */
    public void _case(Indexed tree) {
        Type owntype = Type.errType;
        Type atype = attribExpr(tree.indexed, env);
        attribExpr(tree.index, env, Type.intType);
        if (atype.tag == ARRAY)
            owntype = atype.elemtype();
        else if (atype.tag != ERROR)
            log.error(tree.pos, "array.req.but.found", atype.toJava());
        result = check(tree, owntype, VAR, pkind, pt);
    }

    /**
      * Visitor method for simple identifiers.
      */
    public void _case(Ident tree) {
        Symbol sym;
        if (pt.tag == METHOD) {
            sym = rs.resolveMethod(tree.pos, env, tree.name, Type.emptyList,
                    pt.argtypes());
        } else {
            sym = rs.resolveIdent(tree.pos, env, tree.name, pkind);
        }
        tree.sym = sym;
        Env symEnv = env;
        boolean noOuterThisPath = false;
        if (env.enclClass.sym.owner.kind != PCK &&
                (sym.kind & (VAR | MTH | TYP)) != 0 && sym.owner.kind == TYP &&
                tree.name != Names._this && tree.name != Names._super) {
            while (symEnv.outer != null &&
                    !symEnv.enclClass.sym.isSubClass(sym.owner)) {
                if ((symEnv.enclClass.sym.flags() & NOOUTERTHIS) != 0)
                    noOuterThisPath = true;
                symEnv = symEnv.outer;
            }
            if (symEnv.enclClass.sym != sym.owner) {
                Env env1 = symEnv;
                do {
                    env1 = env1.outer;
                } while (env1 != null && (((AttrContext) env1.info).scope == null ||
                        ((AttrContext) env1.info).scope.owner.kind != PCK &&
                        checkNotHiding(tree.pos, sym,
                        ((AttrContext) env1.info).scope)) &&
                        (env1.enclClass == null || checkNotHiding(tree.pos, sym,
                        env1.enclClass.sym.members())))
                    ;
            }
        }
        if (sym.kind == VAR) {
            VarSymbol v = (VarSymbol) sym;
            checkInit(tree, env, v);
            if (v.owner.kind == MTH &&
                    v.owner != ((AttrContext) env.info).scope.owner &&
                    (v.flags_field & CAPTURED) == 0) {
                v.flags_field |= CAPTURED;
                if ((v.flags_field & FINAL) == 0) {
                    log.error(tree.pos, "local.var.accessed.from.icls.needs.final",
                            v.toJava());
                }
            }
            if (pkind == VAR)
                checkAssignable(tree.pos, v, null, env);
        }
        if ((((AttrContext) symEnv.info).isSelfCall || noOuterThisPath) &&
                (sym.kind & (VAR | MTH)) != 0 && sym.owner.kind == TYP &&
                (sym.flags() & STATIC) == 0) {
            chk.earlyRefError(tree.pos, sym.kind == VAR ? sym : thisSym(env));
        }
        result = checkId(tree, env.enclClass.sym.type, sym, env, pkind, pt);
    }

    /**
      * Visitor method for selections.
      */
    public void _case(Select tree) {
        int skind = 0;
        if (tree.name == Names._this || tree.name == Names._super ||
                tree.name == Names._class) {
            skind = TYP;
        } else {
            if ((pkind & PCK) != 0)
                skind = skind | PCK;
            if ((pkind & TYP) != 0)
                skind = skind | TYP | PCK;
            if ((pkind & (VAL | MTH)) != 0)
                skind = skind | VAL | TYP;
        }
        Type site = attribTree(tree.selected, env, skind, Type.noType);
        Symbol sitesym = TreeInfo.symbol(tree.selected);
        boolean selectSuperPrev = ((AttrContext) env.info).selectSuper;
        ((AttrContext) env.info).selectSuper = sitesym != null &&
                (sitesym.name == Names._super || sitesym.kind == TYP);
        Symbol sym = selectSym(tree, site, env, pt, pkind);
        tree.sym = sym;
        if (sym.kind == VAR) {
            VarSymbol v = (VarSymbol) sym;
            evalInit(v);
            if (pkind == VAR)
                checkAssignable(tree.pos, v, tree.selected, env);
        }
        if (((AttrContext) env.info).selectSuper) {
            if ((sym.flags() & STATIC) == 0 && sym.name != Names._this &&
                    sym.name != Names._super) {
                if (sitesym.name == Names._super) {
                    rs.checkNonAbstract(tree.pos, sym);
                } else if (sym.kind == VAR || sym.kind == MTH) {
                    rs.access(new Resolve.StaticError(sym), tree.pos, site, sym.name);
                }
                Type site1 = env.enclClass.sym.type.asSuper(site.tsym);
                if (site1 != null)
                    site = site1;
            }
            if (((AttrContext) env.info).isSelfCall && tree.name == Names._this &&
                    site.tsym == env.enclClass.sym) {
                chk.earlyRefError(tree.pos, sym);
            }
        }
        ((AttrContext) env.info).selectSuper = selectSuperPrev;
        result = checkId(tree, site, sym, env, pkind, pt);
    }

    /**
      * Determine symbol referenced by a Select expression,
      *
      *  @param pos    Position to be used for error reporting.
      *  @param site   The type of the selected expression,
      *  @param name   The selector name.
      *  @param env    The current environment.
      *  @param pt     The current prototype.
      *  @param pkind  The expected kind(s) of the Select expression.
      */
    private Symbol selectSym(Select tree, Type site, Env env, Type pt, int pkind) {
        int pos = tree.pos;
        Name name = tree.name;
        switch (site.tag) {
        case PACKAGE:
            return rs.access(rs.findIdentInPackage(env, site.tsym, name, pkind),
                    pos, site, name);

        case ARRAY:

        case CLASS:
            if (pt.tag == METHOD) {
                return rs.resolveQualifiedMethod(pos, env, site, name,
                        Type.emptyList, pt.argtypes());
            } else if (name == Names._this || name == Names._super) {
                return rs.resolveSelf(pos, env, site.tsym, name, true);
            } else if (name == Names._class) {
                return new VarSymbol(STATIC | PUBLIC, Names._class,
                        syms.classType, site.tsym);
            } else {
                return rs.access(rs.findIdentInType(env, site, name, pkind), pos,
                        site, name);
            }

        case ERROR:
            return Symbol.errSymbol;

        default:
            if (name == Names._class) {
                return new VarSymbol(STATIC | PUBLIC, Names._class,
                        syms.classType, site.tsym);
            } else {
                log.error(pos, "cant.deref", site.toJava());
                return Symbol.errSymbol;
            }

        }
    }

    /**
      * Determine type of identifier or select expression and check that
      *  (1) the referenced symbol is not deprecated
      *  (2) the symbol's type is safe (@see checkSafe)
      *  (3) if symbol is a variable, check that its type and kind are
      *      compatible with the prototype and protokind.
      *  (4) if symbol is an instance field of a raw type,
      *      which is being assigned to, issue an unchecked warning if its
      *      type changes under erasure.
      *  (5) if symbol is an instance method of a raw type, issue an
      *      unchecked warning if its argument types change under erasure.
      *  If checks succeed:
      *    If symbol is a constant, return its constant type
      *    else if symbol is a method, return its result type
      *    otherwise return its type.
      *  Otherwise return errType.
      *
      *  @param tree       The syntax tree representing the identifier
      *  @param site       If this is a select, the type of the selected
      *                    expression, otherwise the type of the current class.
      *  @param sym        The symbol representing the identifier.
      *  @param env        The current environment.
      *  @param pkind      The set of expected kinds.
      *  @param pt         The expected type.
      */
    Type checkId(Tree tree, Type site, Symbol sym, Env env, int pkind, Type pt) {
        Type owntype;
        switch (sym.kind) {
        case TYP:
            owntype = sym.type;
            if (owntype.tag == CLASS) {
                Type ownOuter = owntype.outer();
                if (ownOuter.tag == CLASS && site != ownOuter) {
                    Type normSite = site;
                    if (normSite.tag == CLASS)
                        normSite = site.asOuterSuper(ownOuter.tsym);
                    if (normSite != ownOuter)
                        owntype = new ClassType(normSite, Type.emptyList,
                                owntype.tsym);
                }
            }
            break;

        case VAR:
            VarSymbol v = (VarSymbol) sym;
            owntype = sym.owner.kind == TYP && sym.name != Names._this &&
                    sym.name != Names._super ? site.memberType(sym) : sym.type;
            if (v.constValue != null && isStaticReference(tree))
                owntype = owntype.constType(v.constValue);
            break;

        case MTH:
            if (sym.name == Names.init) {
                owntype = Type.voidType;
            } else {
                owntype = rs.instantiate(site, sym, Type.emptyList, pt.argtypes());
                if (owntype == null && Type.isRaw(pt.argtypes())) {
                    owntype = rs.instantiate(site, sym, Type.emptyList,
                            Type.unerasure(pt.argtypes()));
                }
                if (owntype == null) {
                    log.error(tree.pos, "internal.error.cant.instantiate",
                            sym.toJava(), site.toJava(),
                            Type.toJavaList(pt.argtypes()));
                }
            }
            break;

        case PCK:

        case ERR:
            owntype = sym.type;
            break;

        default:
            new Pretty().printExpr(tree);
            throw new InternalError("unexpected kind: " + sym.kind);

        }
        if ((sym.flags() & DEPRECATED) != 0)
            chk.warnDeprecated(tree.pos, sym);
        return check(tree, owntype, sym.kind, pkind, pt);
    }

    /**
      * Check that variable is initialized and evaluate the variable's
      *  initializer, if not yet done. Also check that variable is not
      *  referenced before it is defined.
      *  @param tree    The tree making up the variable reference.
      *  @param env     The current environment.
      *  @param v       The variable's symbol.
      */
    private void checkInit(Tree tree, Env env, VarSymbol v) {
        if (v.pos > tree.pos && (v.owner.kind == TYP &&
                canOwnInitializer(((AttrContext) env.info).scope.owner) &&
                v.owner == ((AttrContext) env.info).scope.owner.enclClass() &&
                (v.flags() & STATIC) != 0 == Resolve.isStatic(env)))
            log.error(tree.pos, "illegal.forward.ref");
        evalInit(v);
    }

    /**
      * Can the given symbol be the owner of code which forms part
      *  if class initialization? This is the case if the symbol is
      *  a type or field, or if the symbol is the synthetic method.
      *  owning a block.
      */
    private boolean canOwnInitializer(Symbol sym) {
        return (sym.kind & (VAR | TYP)) != 0 || sym.kind == MTH &&
                (sym.flags() & BLOCK) != 0;
    }

    /**
      * Evaluate a final variable's initializer, unless this has already been
      *  done, and set variable's constant value, if the initializer is
      *  constant.
      */
    void evalInit(VarSymbol v) {
        if (v.constValue instanceof Env) {
            Env evalEnv = (Env) v.constValue;
            Name prev = log.useSource(evalEnv.toplevel.sourcefile);
            v.constValue = null;
            Type itype = attribExpr(((VarDef) evalEnv.tree).init, evalEnv, v.type);
            if (itype.constValue != null)
                v.constValue = cfolder.coerce(itype, v.type).constValue;
            log.useSource(prev);
        }
    }

    /**
      * Check that given symbol does not hide any symbols of the same name and
      *  kind in given scope.
      *  @param pos     Position to be used for error reporting.
      *  @param sym     The given symbol.
      *  @param scope   The given scope.
      */
    boolean checkNotHiding(int pos, Symbol sym, Scope scope) {
        Scope.Entry e = scope.lookup(sym.name);
        while (e.scope != null) {
            if (e.sym.owner != sym.owner && e.sym.kind == sym.kind &&
                    e.sym.owner == scope.owner) {
                log.error(pos, "inherit.hides", sym.toJava(), sym.owner.toJava(),
                        Resolve.kindName(sym.kind), e.sym.javaLocation());
                return false;
            }
            e = e.next();
        }
        return true;
    }

    /**
      * Visitor method for literals.
      */
    public void _case(Literal tree) {
        result = check(tree, litType(tree.typetag).constType(tree.value), VAL,
                pkind, pt);
    }

    /**
      * Return the type of a literal with given type tag.
      */
    Type litType(int tag) {
        return tag == CLASS ? syms.stringType : Type.typeOfTag[tag];
    }

    /**
      * Visitor method for basic type identifiers.
      */
    public void _case(TypeIdent tree) {
        result = check(tree, Type.typeOfTag[tree.typetag], TYP, pkind, pt);
    }

    /**
      * Visitor method for array types.
      */
    public void _case(TypeArray tree) {
        Type etype = attribType(tree.elemtype, env);
        result = check(tree, new ArrayType(etype), TYP, pkind, pt);
    }

    /**
      * Visitor method for the error tree.
      */
    public void _case(Erroneous tree) {
        result = tree.type = Type.errType;
    }

    /**
      * Default visitor method for all other trees.
      */
    public void _case(Tree tree) {
        throw new InternalError();
    }

    /**
      * Main method: attribute class definition associated with given class symbol.
      *  @param c   The class symbol whose definition will be attributed.
      */
    public void attribClass(ClassSymbol c) {
        Type st = c.type.supertype();
        if (st.tag == CLASS)
            attribClass((ClassSymbol) st.tsym);
        if (c.owner.kind == TYP)
            attribClass((ClassSymbol) c.owner);
        if ((c.flags_field & UNATTRIBUTED) != 0) {
            Name prev = log.useSource(c.sourcefile);
            c.flags_field &= ~UNATTRIBUTED;
            Env env = (Env) enter.classEnvs.get(c);
            enter.classEnvs.remove(c);
            ClassDef tree = (ClassDef) env.tree;
            if ((c.flags() & INTERFACE) == 0) {
                if (miranda)
                    implementInterfaceMethods(c, env);
            } else
                chk.checkCompatibleInterfaces(tree.pos, c);
            chk.validate(tree.extending);
            chk.validate(tree.implementing);
            if ((c.flags() & (ABSTRACT | INTERFACE)) == 0 && !retrofit)
                chk.checkAllDefined(tree.pos, c);
            tree.type = c.type;
            chk.checkImplementations(tree.pos, c);
            for (List l = tree.defs; l.nonEmpty(); l = l.tail) {
                attribStat((Tree) l.head, env);
                if (c.owner.kind != PCK &&
                        ((c.flags() & STATIC) == 0 || c.name == Names.empty) &&
                        (TreeInfo.flags((Tree) l.head) & (STATIC | INTERFACE)) != 0) {
                    Symbol sym = null;
                    if (((Tree) l.head).tag == Tree.VARDEF)
                        sym = ((VarDef) l.head).sym;
                    if (sym == null || sym.kind != VAR ||
                            ((VarSymbol) sym).constValue == null) {
                        log.error(((Tree) l.head).pos, "icls.cant.have.static.decl");
                    }
                }
            }
            tree.type = c.type;
            log.useSource(prev);
        }
    }
}
