package com.sun.tools.javac.v8.comp;
import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.tree.*;

import com.sun.tools.javac.v8.tree.Tree.*;

import com.sun.tools.javac.v8.code.Type.*;

import com.sun.tools.javac.v8.code.Symbol.*;


/**
 * Type checking helper class for the attribution phase.
 */
public class Check implements Kinds, Flags, TypeTags {
    Log log;
    Symtab syms;

    public Check(Log log, Symtab syms, Hashtable options) {
        super();
        this.log = log;
        this.syms = syms;
        gj = options.get("-gj") != null;
        warnunchecked = options.get("-warnunchecked") != null;
        deprecation = options.get("-deprecation") != null;
    }

    /**
      * Switch: -gj option set?
      */
    boolean gj;

    /**
     * Switch: -warnunchecked option set?
     */
    boolean warnunchecked;

    /**
     * Switch: -deprecation option set?
     */
    boolean deprecation;

    /**
     * A table mapping flat names of all compiled classes in this run to their
     *  symbols; maintained from outside.
     */
    public Hashtable compiled = Hashtable.make();

    /**
     * Output variable: the source file where deprecated symbols were
     *  encountered. If deprecated symbols were encountered in more than one
     *  input file, the variable is set to "*".
     */
    public Name deprecatedSource;

    /**
     * Output variable: the source file where deprecated symbols were
     *  encountered. If deprecated symbols were encountered in more than one
     *  input file, the variable is set to "*".
     */
    public Name uncheckedSource;

    /**
     * Warn about deprecated symbol if symbol is declared in
     *  a currently compiled file.
     *  @param pos        Position to be used for error reporting.
     *  @param sym        The deprecated symbol.
     */
    void warnDeprecated(int pos, Symbol sym) {
        if (compiled.get(sym.enclClass().flatname) == null) {
            if (deprecatedSource == null)
                deprecatedSource = log.currentSource();
            else if (deprecatedSource != log.currentSource())
                deprecatedSource = Names.asterisk;
            if (deprecation)
                log.warning(pos, "has.been.deprecated", sym.toJava(),
                        sym.javaLocation());
        }
    }

    /**
      * Warn about unchecked operation.
      *  @param pos        Position to be used for error reporting.
      *  @param msg        A string describing the problem.
      */
    void warnUnchecked(int pos, String msg) {
        warnUnchecked(pos, msg, null, null, null, null);
    }

    void warnUnchecked(int pos, String msg, String arg0) {
        warnUnchecked(pos, msg, arg0, null, null, null);
    }

    void warnUnchecked(int pos, String msg, String arg0, String arg1) {
        warnUnchecked(pos, msg, arg0, arg1, null, null);
    }

    void warnUnchecked(int pos, String msg, String arg0, String arg1, String arg2) {
        warnUnchecked(pos, msg, arg0, arg1, arg2, null);
    }

    void warnUnchecked(int pos, String msg, String arg0, String arg1,
            String arg2, String arg3) {
        if (uncheckedSource == null)
            uncheckedSource = log.currentSource();
        else if (uncheckedSource != log.currentSource())
            uncheckedSource = Names.asterisk;
        if (warnunchecked)
            log.warning(pos, msg, arg0, arg1, arg2, arg3);
    }

    /**
      * Report a failure to complete a class.
      *  @param pos        Position to be used for error reporting.
      *  @param ex         The failure to report.
      */
    Type completionError(int pos, CompletionFailure ex) {
        log.error(pos, "cant.access", ex.sym.toJava(), ex.errmsg);
        if (ex instanceof ClassReader.BadClassFile)
            throw new Abort();
        else
            return Type.errType;
    }

    /**
      * Report a type error.
      *  @param pos        Position to be used for error reporting.
      *  @param problem    A string describing the error.
      *  @param found      The type that was found.
      *  @param req        The type that was required.
      */
    Type typeError(int pos, String problem, Type found, Type req) {
        log.error(pos, "prob.found.req", problem, found.toJava(), req.toJava());
        return Type.errType;
    }

    /**
      * Report an error that wrong type tag was found.
      *  @param pos        Position to be used for error reporting.
      *  @param required   An internationalized string describing the type tag
      *                    required.
      *  @param found      The type that was found.
      */
    Type typeTagError(int pos, String required, Type found) {
        log.error(pos, "type.found.req", found.toJava(), required);
        return Type.errType;
    }

    /**
      * Report an error that symbol cannot be referenced before super
      *  has been called.
      *  @param pos        Position to be used for error reporting.
      *  @param sym        The referenced symbol.
      */
    void earlyRefError(int pos, Symbol sym) {
        log.error(pos, "cant.ref.before.ctor.called", sym.toJava());
    }

    /**
      * Report an error that type parameter is not within its bound.
      *  @param pos                Position to be used for error reporting.
      *  @param t                  The type parameter.
      *  @param bound              Its bound.
      *  @param explanation        An internationalized string giving an
      *                            additional explanation.
      */
    void boundError(int pos, Type t, Type bound, String explanation) {
        if (explanation.equals(""))
            log.error(pos, "not.within.bounds", t.toJava());
        else
            log.error(pos, "not.within.bounds.explain", t.toJava(), explanation);
    }

    /**
      * Return name of local class.
      *  This is of the form    <topname> $ n $ <classname>
      *  where
      *    topname is the fully qualified name of the enclosing toplevel class,
      *    classname is the simple name of the local class
      *    n the minimum number >= 1, such that no other class with the same
      *      name is being compiled.
      */
    Name localClassName(ClassSymbol c, int n) {
        ClassSymbol topClass = c.outermostClass();
        Name flatname = Name.fromString(topClass.flatname + "$" + n +
                (c.name.len == 0 ? "" : "$" + c.name));
        if (compiled.get(flatname) != null)
            return localClassName(c, n + 1);
        else
            return flatname;
    }

    /**
      * Check that a given type is assignable to a given proto-type.
      *  If it is, return the type, otherwise return Type.errType.
      *  @param pos        Position to be used for error reporting.
      *  @param found      The type that was found.
      *  @param req        The type that was required.
      */
    Type checkType(int pos, Type found, Type req) {
        if (req.tag == ERROR) {
            return req;
        } else if (found.isAssignable(req)) {
            return found;
        } else if (found.isRaw() && found.unerasure().isAssignable(req)) {
            warnUnchecked(pos, "unchecked assignment", found.toJava(), req.toJava());
            return req;
        } else {
            String problem;
            if (found.tag <= DOUBLE && req.tag <= DOUBLE)
                problem = log.getLocalizedString("possible.loss.of.precision");
            else
                problem = log.getLocalizedString("incompatible.types");
            return typeError(pos, problem, found, req);
        }
    }

    /**
      * Check that a given type can be cast to a given target type.
      *  Return the result of the cast.
      *  @param pos        Position to be used for error reporting.
      *  @param found      The type that is being cast.
      *  @param req        The target type of the cast.
      */
    Type checkCastable(int pos, Type found, Type req) {
        if (found.isCastable(req)) {
            checkCompatible(pos, found, req);
            return req;
        } else {
            return typeError(pos, log.getLocalizedString("inconvertible.types"),
                    found, req);
        }
    }

    /**
      * Check that type is different from 'void'.
      *  @param pos           Position to be used for error reporting.
      *  @param t             The type to be checked.
      */
    Type checkNonVoid(int pos, Type t) {
        if (t.tag == VOID) {
            log.error(pos, "void.not.allowed.here");
            return Type.errType;
        } else {
            return t;
        }
    }

    /**
      * Check that type is a class or interface type.
      *  @param pos           Position to be used for error reporting.
      *  @param t             The type to be checked.
      */
    Type checkClassType(int pos, Type t) {
        if (t.tag != CLASS && t.tag != ARRAY && t.tag != ERROR)
            return typeTagError(pos, log.getLocalizedString("type.req.class"), t);
        else
            return t;
    }

    /**
      * Check that type is a class, interface or array type.
      *  @param pos           Position to be used for error reporting.
      *  @param t             The type to be checked.
      */
    Type checkClassOrArrayType(int pos, Type t) {
        if (t.tag != CLASS && t.tag != ARRAY && t.tag != ERROR)
            return typeTagError(pos,
                    log.getLocalizedString("type.req.class.array"), t);
        else
            return t;
    }

    /**
      * Check that type is a reference type, i.e. a class, interface or array type
      *  or a type variable.
      *  @param pos           Position to be used for error reporting.
      *  @param t             The type to be checked.
      */
    Type checkRefType(int pos, Type t) {
        if (t.tag != CLASS && t.tag != ARRAY && t.tag != TYPEVAR && t.tag != ERROR)
            return typeTagError(pos, log.getLocalizedString("type.req.ref"), t);
        else
            return t;
    }

    /**
      * Check that flag set does not contain elements of two conflicting sets. s
      *  Return true if it doesn't.
      *  @param pos           Position to be used for error reporting.
      *  @param flags         The set of flags to be checked.
      *  @param set1          Conflicting flags set #1.
      *  @param set2          Conflicting flags set #2.
      */
    boolean checkDisjoint(int pos, int flags, int set1, int set2) {
        if ((flags & set1) != 0 && (flags & set2) != 0) {
            log.error(pos, "illegal.combination.of.modifiers",
                    TreeInfo.flagNames(TreeInfo.firstFlag(flags & set1)),
                    TreeInfo.flagNames(TreeInfo.firstFlag(flags & set2)));
            return false;
        } else
            return true;
    }

    /**
      * Check that given modifiers are legal for given symbol and
      *  return modifiers together with any implicit modififiers for that symbol.
      *  Warning: we can't use flags() here since this method
      *  is called during class enter, when flags() would cause a premature
      *  completion.
      *  @param pos           Position to be used for error reporting.
      *  @param flags         The set of modifiers given in a definition.
      *  @param sym           The defined symbol.
      */
    int checkFlags(int pos, int flags, Symbol sym) {
        int mask;
        int implicit = 0;
        switch (sym.kind) {
        case VAR:
            if (sym.owner.kind != TYP)
                mask = LocalVarFlags;
            else if ((sym.owner.flags_field & INTERFACE) != 0)
                mask = implicit = InterfaceVarFlags;
            else
                mask = VarFlags;
            break;

        case MTH:
            if (sym.name == Names.init)
                mask = ConstructorFlags;
            else if ((sym.owner.flags_field & INTERFACE) != 0)
                mask = implicit = InterfaceMethodFlags;
            else {
                mask = MethodFlags;
            }
            implicit |= sym.owner.flags_field & STRICTFP;
            break;

        case TYP:
            if (sym.isLocal()) {
                mask = LocalClassFlags;
                if (sym.name.len == 0)
                    mask |= STATIC;
            } else if (sym.owner.kind == TYP) {
                mask = MemberClassFlags;
                if (sym.owner.owner.kind == PCK ||
                        (sym.owner.flags_field & STATIC) != 0)
                    mask |= STATIC;
                if ((flags & INTERFACE) != 0)
                    implicit = STATIC;
            } else {
                mask = ClassFlags;
            }
            if ((flags & INTERFACE) != 0)
                implicit |= ABSTRACT;
            implicit |= sym.owner.flags_field & STRICTFP;
            break;

        default:
            throw new InternalError();

        }
        int illegal = flags & StandardFlags & ~mask;
        if (illegal != 0)
            log.error(pos, "mod.not.allowed.here", TreeInfo.flagNames(illegal));
        else if ( (sym.kind == TYP ||
                checkDisjoint(pos, flags, ABSTRACT, PRIVATE | STATIC)) &&
                checkDisjoint(pos, flags, ABSTRACT | INTERFACE,
                FINAL | NATIVE | SYNCHRONIZED) &&
                checkDisjoint(pos, flags, PUBLIC, PRIVATE | PROTECTED) &&
                checkDisjoint(pos, flags, PRIVATE, PUBLIC | PROTECTED) &&
                checkDisjoint(pos, flags, FINAL, VOLATILE) && (sym.kind == TYP ||
                checkDisjoint(pos, flags, ABSTRACT | NATIVE, STRICTFP))) {
        }
        return flags & (mask | ~StandardFlags) | implicit;
    }

    /**
      * Validate a type expression. That is,
      *  check that all type arguments of a parametric type are within
      *  their bounds. This must be done in a second phase after type attributon
      *  since a class might have a subclass as type parameter bound. E.g:
      *
      *  class B<A extends C> { ... }
      *  class C extends B<C> { ... }
      *
      *  and we can't make sure that the bound is already attributed because
      *  of possible cycles.
      */
    private Validator validator = new Validator();

    /**
     * Visitor method: Validate a type expression, if it is not null, catching
     *  and reporting any completion failures.
     */
    void validate(Tree tree) {
        try {
            {
                if (tree != null)
                    tree.visit(validator);
            }
        } catch (CompletionFailure ex) {
            {
                completionError(tree.pos, ex);
            }
        }
    }

    /**
      * Visitor method: Validate a list of type expressions.
      */
    void validate(List trees) {
        for (List l = trees; l.nonEmpty(); l = l.tail)
            validate((Tree) l.head);
    }

    /**
      * A visitor class for type validation.
      */
    class Validator extends Visitor {

        Validator() {
            super();
        }

        /**
          * Visitor method for array types.
          */
        public void _case(TypeArray tree) {
            validate(tree.elemtype);
        }

        /**
          * Visitor method for qualified types.
          */
        public void _case(Select tree) {
            if (tree.type.tag == CLASS) {
                if (tree.type.outer().tag == CLASS)
                    validate(tree.selected);
                else if (tree.selected.type.isParameterized())
                    log.error(tree.pos, "cant.select.static.class.from.param.type");
                if (tree.type.isRaw() && tree.type.allparams().nonEmpty())
                    log.error(tree.pos, "improperly.formed.type.param.missing");
            }
        }

        /**
          * Default visitor method: do nothing.
          */
        public void _case(Tree tree) {
        }
    }

    /**
      * Is given class a subclass of some of the classes in given list?
      */
    static boolean subset(ClassSymbol c, List cs) {
        for (List l = cs; l.nonEmpty(); l = l.tail)
            if (c.isSubClass((Symbol) l.head))
                return true;
        return false;
    }

    /**
      * Is given class a subclass or superclass of
      *  some of the classes in given list?
      */
    static boolean intersects(ClassSymbol c, List cs) {
        for (List l = cs; l.nonEmpty(); l = l.tail)
            if (c.isSubClass((Symbol) l.head) || ((ClassSymbol) l.head).isSubClass(c))
                return true;
        return false;
    }

    /**
      * Add class set to given class list, unless it is a subclass of some class
      *  in the list.
      */
    static List incl(ClassSymbol c, List cs) {
        return subset(c, cs) ? cs : excl(c, cs).prepend(c);
    }

    /**
      * Remove class set from class set list.
      */
    static List excl(ClassSymbol c, List cs) {
        if (cs.isEmpty()) {
            return cs;
        } else {
            List cs1 = excl(c, cs.tail);
            if (((ClassSymbol) cs.head).isSubClass(c))
                return cs1;
            else if (cs1 == cs.tail)
                return cs;
            else
                return cs1.prepend(cs.head);
        }
    }

    /**
      * Form the union of two class set lists.
      */
    static List union(List cs1, List cs2) {
        List cs = cs1;
        for (List l = cs2; l.nonEmpty(); l = l.tail)
            cs = incl((ClassSymbol) l.head, cs);
        return cs;
    }

    /**
      * Form the difference of two class set lists.
      */
    static List diff(List cs1, List cs2) {
        List cs = cs1;
        for (List l = cs2; l.nonEmpty(); l = l.tail)
            cs = excl((ClassSymbol) l.head, cs);
        return cs;
    }

    /**
      * Form the intersection of two class set lists.
      */
    static List intersect(List cs1, List cs2) {
        List cs = ClassSymbol.emptyList;
        for (List l = cs1; l.nonEmpty(); l = l.tail)
            if (subset((ClassSymbol) l.head, cs2))
                cs = incl((ClassSymbol) l.head, cs);
        for (List l = cs2; l.nonEmpty(); l = l.tail)
            if (subset((ClassSymbol) l.head, cs1))
                cs = incl((ClassSymbol) l.head, cs);
        return cs;
    }

    /**
      * Is exc an exception that need not be declared?
      */
    boolean isUnchecked(ClassSymbol exc) {
        return exc == Type.errType.tsym || exc.isSubClass(syms.errorType.tsym) ||
                exc.isSubClass(syms.runtimeExceptionType.tsym);
    }

    /**
      * Is exc handled by given exception list?
      */
    boolean isHandled(ClassSymbol exc, List handled) {
        return isUnchecked(exc) || subset(exc, handled);
    }

    /**
      * Return all exceptions in thrown list that are not in handled list.
      *  @param thrown     The list of thrown exceptions.
      *  @param handled    The list of handled exceptions.
      */
    List unHandled(List thrown, List handled) {
        List unhandled = ClassSymbol.emptyList;
        for (List l = thrown; l.nonEmpty(); l = l.tail)
            if (!isHandled((ClassSymbol) l.head, handled))
                unhandled = unhandled.prepend(l.head);
        return unhandled;
    }

    /**
      * Check that given thrown exception is reported.
      *  @param pos        Position to be used for error reporting.
      *  @param exc        The thrown exception.
      *  @param reported   The list of all reported exceptions.
      */
    void checkHandled(int pos, ClassSymbol exc, List reported) {
        if (!isHandled(exc, reported))
            log.error(pos, "unreported.exception.need.to.catch.or.throw",
                    exc.type.toJava());
    }

    /**
      * Check that all exceptions in given list are reported.
      *  @param pos        Position to be used for error reporting.
      *  @param exc        The list of thrown exceptions.
      *  @param reported   The list of reported exceptions.
      */
    void checkHandled(int pos, List thrown, List reported) {
        List unhandled = unHandled(thrown, reported);
        if (unhandled.nonEmpty())
            checkHandled(pos, (ClassSymbol) unhandled.head, reported);
    }

    /**
      * The level of access protection given by a flag set,
      *  where PRIVATE is highest and PUBLIC is lowest.
      */
    static int protection(int flags) {
        switch (flags & AccessFlags) {
        case PRIVATE:
            return 3;

        case PROTECTED:
            return 1;

        case PUBLIC:
            return 0;

        default:
            return 2;

        }
    }

    /**
      * A string describing the access permission given by a flag set.
      *  This always returns a space-separated list of Java Keywords.
      */
    private static String protectionString(int flags) {
        int flags1 = flags & AccessFlags;
        return flags1 == 0 ? "package" : TreeInfo.flagNames(flags1);
    }

    /**
      * A customized "cannot override" error message.
      *  @param m      The overriding method.
      *  @param other  The overridden method.
      *  @return       An internationalized string.
      */
    static String cannotOverride(MethodSymbol m, MethodSymbol other) {
        String msg;
        if ((other.flags() & IPROXY | other.owner.flags() & INTERFACE) == 0)
            msg = Log.getLocalizedString("cant.override", m.toJava(),
                    m.javaLocation(), other.toJava(), other.javaLocation());
        else if ((m.flags() & IPROXY | m.owner.flags() & INTERFACE) == 0)
            msg = Log.getLocalizedString("cant.implement", m.toJava(),
                    m.javaLocation(), other.toJava(), other.javaLocation());
        else
            msg = Log.getLocalizedString("clashes.with", m.toJava(),
                    m.javaLocation(), other.toJava(), other.javaLocation());
        return msg;
    }

    /**
      * Check that this method conforms with overridden method 'other'.
      *  where `origin' is the class where checking started.
      *  Complications:
      *  (1) Do not check overriding of synthetic methods
      *      (reason: they might be final).
      *      todo: check whether this is still necessary.
      *  (2) Admit the case where an interface proxy throws fewer exceptions
      *      than the method it implements. Augment the proxy methods with the
      *      undeclared exceptions in this case.
      *  (3) In GJ, admit the case where an interface proxy has a result type
      *      extended by the result type of the method it implements.
      *      Change the proxies result type to the smaller type in this case.
      *
      *  @param pos          Position to be used for error reporting.
      *  @param m            The overriding method.
      *  @param other        The overridden method.
      *  @param origin       The class of which the overriding method is a member.
      */
    void checkOverride(int pos, MethodSymbol m, MethodSymbol other,
            ClassSymbol origin) {
        if ((other.flags() & SYNTHETIC) != 0) {
        } else if ((m.flags() & STATIC) != 0 && (other.flags() & STATIC) == 0) {
            log.error(pos, "override.static", cannotOverride(m, other),
                    m.toJava(), other.toJava());
        } else if ((other.flags() & FINAL) != 0 || (m.flags() & STATIC) == 0 &&
                (other.flags() & STATIC) != 0) {
            log.error(pos, "override.meth", cannotOverride(m, other),
                    TreeInfo.flagNames(other.flags() & (FINAL | STATIC)));
        } else if ((origin.flags() & INTERFACE) == 0 &&
                protection(m.flags()) > protection(other.flags())) {
            log.error(pos, "override.weaker.access", cannotOverride(m, other),
                    protectionString(other.flags()));
        } else {
            Type mt = origin.type.memberType(m);
            Type ot = origin.type.memberType(other);
            List mtvars = mt.typarams();
            List otvars = ot.typarams();
            Type otres = ot.restype().subst(otvars, mtvars);
            if (!(gj && mt.restype().isSubType(otres) ||
                    mt.restype().isSameType(otres))) {
                if (gj && (m.flags() & IPROXY) != 0 &&
                        otres.isSubType(mt.restype())) {
                    ((MethodType) mt).restype = otres;
                } else {
                    typeError(pos,
                            log.getLocalizedString("override.incompatible.ret",
                            cannotOverride(m, other)), mt.restype(),
                            ot.restype().subst(otvars, mtvars));
                }
            } else if ((origin.flags() & INTERFACE) == 0) {
                List unhandled = unHandled(mt.thrown(), ot.thrown());
                if (unhandled.nonEmpty()) {
                    if ((m.flags() & IPROXY) != 0) {
                        ((MethodType) mt).thrown = mt.thrown().prepend(unhandled);
                    } else {
                        log.error(pos, "override.meth.doesnt.throw",
                                cannotOverride(m, other),
                                ((ClassSymbol) unhandled.head).toJava());
                    }
                }
            }
            if ((other.flags() & DEPRECATED) != 0)
                warnDeprecated(pos, other);
        }
    }

    /**
      * Check that a given method conforms with any method it overrides.
      *  @param pos          Position to be used for error reporting.
      *  @param m            The overriding method.
      */
    void checkOverride(int pos, MethodSymbol m) {
        ClassSymbol origin = (ClassSymbol) m.owner;
        for (Type t = origin.type.supertype(); t.tag == CLASS; t = t.supertype()) {
            TypeSymbol c = t.tsym;
            Scope.Entry e = c.members().lookup(m.name);
            while (e.scope != null) {
                if (m.overrides(e.sym, origin)) {
                    checkOverride(pos, m, (MethodSymbol) e.sym, origin);
                    return;
                }
                e = e.next();
            }
        }
    }

    /**
      * Return the first method which is defined with same args
      *  but different return types in two given interfaces, or null if none exists.
      *  @param t1     The first interface.
      *  @param t2     The second interface.
      */
    public Symbol firstIncompatibility(Type t1, Type t2) {
        return firstIncompatibility(t1, t2, t2);
    }

    /**
      * Return the first method which is defined with same args
      *  but different return types in a interface `t1' and a superinterface
      *  `t2super' of an interface `t2', or null if none exists.
      *  @param t1       The first interface.
      *  @param t2       The second interface.
      *  @param t2super  The second interface's superinterface.
      */
    public Symbol firstIncompatibility(Type t1, Type t2, Type t2super) {
        for (Scope.Entry e = t2super.tsym.members().elems; e != null; e = e.sibling) {
            if (e.sym.kind == MTH && !isCompatible(t1, t1, t2, e.sym))
                return e.sym;
        }
        for (List is = t2super.interfaces(); is.nonEmpty(); is = is.tail) {
            Symbol sym = firstIncompatibility(t1, t2, (Type) is.head);
            if (sym != null)
                return sym;
        }
        return null;
    }

    /**
      * Check that superinterface `t1super' of interface `t1' does not define
      *  a method with with same name and argument types but a different return
      *  type than its definition in interface `t2'.
      *  @param t1       The first interface.
      *  @param t1super  The first interface's superinterface.
      *  @param t2       The second interface.
      *  @param sym      The method symbol.
      */
    private boolean isCompatible(Type t1, Type t1super, Type t2, Symbol sym) {
        for (Scope.Entry e = t1super.tsym.members().lookup(sym.name);
                e.scope != null; e = e.next()) {
            if (e.sym.kind == MTH && e.sym.type.hasSameArgs(sym.type)) {
                return e.sym.type.restype().isSameType(sym.type.restype());
            }
        }
        for (List is = t1super.interfaces(); is.nonEmpty(); is = is.tail) {
            if (!isCompatible(t1, (Type) is.head, t2, sym))
                return false;
        }
        return true;
    }

    /**
      * Check that (arrays of) interfaces do not each define a method
      *  with same name and arguments but different return types.
      *  If either argument type is not an (array of) interface type, do nothing.
      *  @param pos          Position to be used for error reporting.
      *  @param t1           The first argument type.
      *  @param t2           The second argument type.
      */
    public boolean checkCompatible(int pos, Type t1, Type t2) {
        if (t1.tag == ARRAY && t2.tag == ARRAY) {
            checkCompatible(pos, t1.elemtype(), t2.elemtype());
        } else if (t1.tag == CLASS && (t1.tsym.flags() & INTERFACE) != 0 &&
                t2.tag == CLASS && (t2.tsym.flags() & INTERFACE) != 0) {
            Symbol sym = firstIncompatibility(t1, t2);
            if (sym != null) {
                log.error(pos, "intf.incompatible.diff.ret", t1.toJava(),
                        t2.toJava(), sym.toJava());
                return false;
            }
        }
        return true;
    }

    /**
      * Check that all abstract members of given class have definitions.
      *  @param pos          Position to be used for error reporting.
      *  @param c            The class.
      */
    void checkAllDefined(int pos, ClassSymbol c) {
        MethodSymbol undef = firstUndef(c, c);
        if (undef != null) {
            MethodSymbol undef1 = new MethodSymbol(undef.flags(), undef.name,
                    c.type.memberType(undef), undef.owner);
            log.error(pos, "should.be.abstract.doesnt.def", c.toJava(),
                    undef1.toJava(), undef1.javaLocation());
        }
    }

    /**
      * Return first abstract member of class `c' that is not defined in `impl',
      *  null if there is none.
      */
    private MethodSymbol firstUndef(ClassSymbol impl, ClassSymbol c) {
        MethodSymbol undef = null;
        if (c == impl || (c.flags() & (ABSTRACT | INTERFACE)) != 0) {
            Scope s = c.members();
            for (Scope.Entry e = s.elems; undef == null && e != null; e = e.sibling) {
                if (e.sym.kind == MTH && (e.sym.flags() & ABSTRACT) != 0) {
                    MethodSymbol absmeth = (MethodSymbol) e.sym;
                    MethodSymbol implmeth = absmeth.implementation(impl);
                    if (implmeth == null || implmeth == absmeth)
                        undef = absmeth;
                }
            }
            if (undef == null) {
                Type st = c.type.supertype();
                if (st.tag == CLASS)
                    undef = firstUndef(impl, (ClassSymbol) st.tsym);
            }
            for (List l = c.type.interfaces(); undef == null && l.nonEmpty();
                    l = l.tail) {
                undef = firstUndef(impl, (ClassSymbol)((Type) l.head).tsym);
            }
        }
        return undef;
    }

    /**
      * Check that all methods which implement some interface
      *  method conform to the method they implement.
      *  @param pos          Position to be used for error reporting.
      *  @param origin       The class whose members are checked.
      */
    void checkImplementations(int pos, ClassSymbol origin) {
        for (List l = origin.type.interfaces(); l.nonEmpty(); l = l.tail)
            checkImplementations(pos, origin, (ClassSymbol)((Type) l.head).tsym);
    }

    /**
      * Check that all methods which implement some interface
      *  method in one of the interfaces of `ic' conform to the method
      *  they implement.
      */
    void checkImplementations(int pos, ClassSymbol origin, ClassSymbol ic) {
        for (Scope.Entry e = ic.members().elems; e != null; e = e.sibling) {
            if (e.sym.kind == MTH && (e.sym.flags() & STATIC) == 0) {
                MethodSymbol absmeth = (MethodSymbol) e.sym;
                MethodSymbol implmeth = absmeth.implementation(origin);
                if (implmeth != null) {
                    checkOverride(pos, implmeth, absmeth, origin);
                }
            }
        }
        for (List l = ic.type.interfaces(); l.nonEmpty(); l = l.tail)
            checkImplementations(pos, origin, (ClassSymbol)((Type) l.head).tsym);
    }

    /**
      * Check that all interfaces implemented by a class are mutually compatible.
      *  @param pos          Position to be used for error reporting.
      *  @param c            The class whose interfaces are checked.
      */
    void checkCompatibleInterfaces(int pos, ClassSymbol c) {
        for (List l = c.type.interfaces(); l.nonEmpty(); l = l.tail)
            for (List m = c.type.interfaces(); m != l; m = m.tail)
                if (!checkCompatible(pos, (Type) l.head, (Type) m.head))
                    return;
    }
}
