package com.sun.tools.javac.v8.comp;
import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.code.Type.*;


/**
 * Helper class for constant folding, used by the attribution phase
 */
class ConstFold implements TypeTags, ByteCodes {
    Log log;
    Symtab syms;

    ConstFold(Log log, Symtab syms) {
        super();
        this.log = log;
        this.syms = syms;
    }
    static Integer minusOne = new Integer(-1);
    static Integer zero = new Integer(0);
    static Integer one = new Integer(1);

    /**
     * Convert boolean to integer (true = 1, false = 0).
     */
    private static Integer b2i(boolean b) {
        return b ? one : zero;
    }

    private static int intValue(Object x) {
        return ((Number) x).intValue();
    }

    private static long longValue(Object x) {
        return ((Number) x).longValue();
    }

    private static float floatValue(Object x) {
        return ((Number) x).floatValue();
    }

    private static double doubleValue(Object x) {
        return ((Number) x).doubleValue();
    }

    /**
      * Fold binary or unary operation, returning constant type reflecting the
      *  operations result. Return null if fold failed due to an
      *  arithmetic exception.
      *  @param opcode    The operation's opcode instruction (usually a byte code),
      *                   as entered by class Symtab.
      *  @param argtypes  The operation's argument types (a list of length 1 or 2).
      *                   Argument types are assumed to have non-null constValue's.
      */
    Type fold(int opcode, List argtypes) {
        int argCount = argtypes.length();
        if (argCount == 1)
            return fold1(opcode, (Type) argtypes.head);
        else if (argCount == 2)
            return fold2(opcode, (Type) argtypes.head, (Type) argtypes.tail.head);
        else
            throw new InternalError();
    }

    /**
      * Fold unary operation.
      *  @param opcode    The operation's opcode instruction (usually a byte code),
      *                   as entered by class Symtab.
      *                   opcode's ifeq to ifge are for postprocessing
      *                   xcmp; ifxx pairs of instructions.
      *  @param operand   The operation's operand type.
      *                   Argument types are assumed to have non-null constValue's.
      */
    Type fold1(int opcode, Type operand) {
        try {
            {
                Object od = operand.constValue;
                switch (opcode) {
                case nop:
                    return operand;

                case ineg:
                    return Type.intType.constType(new Integer(-intValue(od)));

                case ixor:
                    return Type.intType.constType(new Integer(~intValue(od)));

                case bool_not:
                    return Type.booleanType.constType(
                            new Integer(~intValue(od) & 1));

                case ifeq:
                    return Type.booleanType.constType(b2i(intValue(od) == 0));

                case ifne:
                    return Type.booleanType.constType(b2i(intValue(od) != 0));

                case iflt:
                    return Type.booleanType.constType(b2i(intValue(od) < 0));

                case ifgt:
                    return Type.booleanType.constType(b2i(intValue(od) > 0));

                case ifle:
                    return Type.booleanType.constType(b2i(intValue(od) <= 0));

                case ifge:
                    return Type.booleanType.constType(b2i(intValue(od) >= 0));

                case lneg:
                    return Type.longType.constType(new Long(-longValue(od)));

                case lxor:
                    return Type.longType.constType(new Long(~longValue(od)));

                case fneg:
                    return Type.floatType.constType(new Float(-floatValue(od)));

                case dneg:
                    return Type.doubleType.constType(new Double(-doubleValue(od)));

                default:
                    return null;

                }
            }
        } catch (ArithmeticException e) {
            {
                return null;
            }
        }
    }

    /**
      * Fold binary operation.
      *  @param opcode    The operation's opcode instruction (usually a byte code),
      *                   as entered by class Symtab.
      *                   opcode's ifeq to ifge are for postprocessing
      *                   xcmp; ifxx pairs of instructions.
      *  @param left      The type of the operation's left operand.
      *  @param right     The type of the operation's right operand.
      */
    Type fold2(int opcode, Type left, Type right) {
        try {
            {
                if (opcode > ByteCodes.preMask) {
                    Type t1 = fold2(opcode >> ByteCodes.preShift, left, right);
                    return t1.constValue == null ? t1 :
                            fold1(opcode & ByteCodes.preMask, t1);
                } else {
                    Object l = left.constValue;
                    Object r = right.constValue;
                    switch (opcode) {
                    case iadd:
                        return Type.intType.constType(
                                new Integer(intValue(l) + intValue(r)));

                    case isub:
                        return Type.intType.constType(
                                new Integer(intValue(l) - intValue(r)));

                    case imul:
                        return Type.intType.constType(
                                new Integer(intValue(l) * intValue(r)));

                    case idiv:
                        return Type.intType.constType(
                                new Integer(intValue(l) / intValue(r)));

                    case imod:
                        return Type.intType.constType(
                                new Integer(intValue(l) % intValue(r)));

                    case iand:
                        return Type.intType.constType(
                                new Integer(intValue(l) & intValue(r)));

                    case bool_and:
                        return Type.booleanType.constType(
                                new Integer(intValue(l) & intValue(r)));

                    case ior:
                        return Type.intType.constType(
                                new Integer(intValue(l) | intValue(r)));

                    case bool_or:
                        return Type.booleanType.constType(
                                new Integer(intValue(l) | intValue(r)));

                    case ixor:
                        return Type.intType.constType(
                                new Integer(intValue(l) ^ intValue(r)));

                    case ishl:

                    case ishll:
                        return Type.intType.constType(
                                new Integer(intValue(l)<< intValue(r)));

                    case ishr:

                    case ishrl:
                        return Type.intType.constType(
                                new Integer(intValue(l)>> intValue(r)));

                    case iushr:

                    case iushrl:
                        return Type.intType.constType(
                                new Integer(intValue(l)>>> intValue(r)));

                    case if_icmpeq:
                        return Type.booleanType.constType(
                                b2i(intValue(l) == intValue(r)))
                               ;

                    case if_icmpne:
                        return Type.booleanType.constType(
                                b2i(intValue(l) != intValue(r)))
                               ;

                    case if_icmplt:
                        return Type.booleanType.constType(
                                b2i(intValue(l) < intValue(r)))
                               ;

                    case if_icmpgt:
                        return Type.booleanType.constType(
                                b2i(intValue(l) > intValue(r)))
                               ;

                    case if_icmple:
                        return Type.booleanType.constType(
                                b2i(intValue(l) <= intValue(r)))
                               ;

                    case if_icmpge:
                        return Type.booleanType.constType(
                                b2i(intValue(l) >= intValue(r)))
                               ;

                    case ladd:
                        return Type.longType.constType(
                                new Long(longValue(l) + longValue(r)));

                    case lsub:
                        return Type.longType.constType(
                                new Long(longValue(l) - longValue(r)));

                    case lmul:
                        return Type.longType.constType(
                                new Long(longValue(l) * longValue(r)));

                    case ldiv:
                        return Type.longType.constType(
                                new Long(longValue(l) / longValue(r)));

                    case lmod:
                        return Type.longType.constType(
                                new Long(longValue(l) % longValue(r)));

                    case land:
                        return Type.longType.constType(
                                new Long(longValue(l) & longValue(r)));

                    case lor:
                        return Type.longType.constType(
                                new Long(longValue(l) | longValue(r)));

                    case lxor:
                        return Type.longType.constType(
                                new Long(longValue(l) ^ longValue(r)));

                    case lshl:

                    case lshll:
                        return Type.longType.constType(
                                new Long(longValue(l)<< intValue(r)));

                    case lshr:

                    case lshrl:
                        return Type.longType.constType(
                                new Long(longValue(l)>> intValue(r)));

                    case lushr:
                        return Type.longType.constType(
                                new Long(longValue(l)>>> intValue(r)));

                    case lcmp:
                        if (longValue(l) < longValue(r))
                            return Type.intType.constType(minusOne);
                        else if (longValue(l) > longValue(r))
                            return Type.intType.constType(one);
                        else
                            return Type.intType.constType(zero);

                    case fadd:
                        return Type.floatType.constType(
                                new Float(floatValue(l) + floatValue(r)));

                    case fsub:
                        return Type.floatType.constType(
                                new Float(floatValue(l) - floatValue(r)));

                    case fmul:
                        return Type.floatType.constType(
                                new Float(floatValue(l) * floatValue(r)));

                    case fdiv:
                        return Type.floatType.constType(
                                new Float(floatValue(l) / floatValue(r)));

                    case fmod:
                        return Type.floatType.constType(
                                new Float(floatValue(l) % floatValue(r)));

                    case fcmpg:

                    case fcmpl:
                        if (floatValue(l) < floatValue(r))
                            return Type.intType.constType(minusOne);
                        else if (floatValue(l) > floatValue(r))
                            return Type.intType.constType(one);
                        else if (floatValue(l) == floatValue(r))
                            return Type.intType.constType(zero);
                        else if (opcode == fcmpg)
                            return Type.intType.constType(one);
                        else
                            return Type.intType.constType(minusOne);

                    case dadd:
                        return Type.doubleType.constType(
                                new Double(doubleValue(l) + doubleValue(r)));

                    case dsub:
                        return Type.doubleType.constType(
                                new Double(doubleValue(l) - doubleValue(r)));

                    case dmul:
                        return Type.doubleType.constType(
                                new Double(doubleValue(l) * doubleValue(r)));

                    case ddiv:
                        return Type.doubleType.constType(
                                new Double(doubleValue(l) / doubleValue(r)));

                    case dmod:
                        return Type.doubleType.constType(
                                new Double(doubleValue(l) % doubleValue(r)));

                    case dcmpg:

                    case dcmpl:
                        if (doubleValue(l) < doubleValue(r))
                            return Type.intType.constType(minusOne);
                        else if (doubleValue(l) > doubleValue(r))
                            return Type.intType.constType(one);
                        else if (doubleValue(l) == doubleValue(r))
                            return Type.intType.constType(zero);
                        else if (opcode == dcmpg)
                            return Type.intType.constType(one);
                        else
                            return Type.intType.constType(minusOne);

                    case if_acmpeq:
                        return Type.booleanType.constType(b2i(l.equals(r)))
                               ;

                    case if_acmpne:
                        return Type.booleanType.constType(b2i(!l.equals(r)))
                               ;

                    case string_add:
                        return syms.stringType.constType(left.stringValue() +
                                right.stringValue());

                    default:
                        return null;

                    }
                }
            }
        } catch (ArithmeticException e) {
            {
                return null;
            }
        }
    }

    /**
      * Coerce constant type to target type.
      *  @param etype      The source type of the coercion,
      *                    which is assumed to be a constant type compatble with
      *                    ttype.
      *  @param ttype      The target type of the coercion.
      */
    Type coerce(Type etype, Type ttype) {
        if (etype.baseType() == ttype.baseType())
            return etype;
        if (etype.tag <= DOUBLE) {
            Object n = etype.constValue;
            switch (ttype.tag) {
            case BYTE:
                return Type.byteType.constType(new Integer((byte) intValue(n)));

            case CHAR:
                return Type.charType.constType(new Integer((char) intValue(n)));

            case SHORT:
                return Type.shortType.constType(new Integer((short) intValue(n)));

            case INT:
                return Type.intType.constType(new Integer(intValue(n)));

            case LONG:
                return Type.longType.constType(new Long(longValue(n)));

            case FLOAT:
                return Type.floatType.constType(new Float(floatValue(n)));

            case DOUBLE:
                return Type.doubleType.constType(new Double(doubleValue(n)));

            }
        }
        return ttype;
    }
}
