package com.sun.tools.javac.v8.comp;
import java.io.File;

import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.tree.*;

import com.sun.tools.javac.v8.code.Type.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.tree.Tree.*;


/**
 * This class enters symbols for all encountered definitions into
 *  the symbol table. The pass consists of two phases, organized as
 *  follows:
 *
 *  In the first phase, all class symbols are intered into their enclosing scope,
 *  descending recursively down the tree for classes which are members of other
 *  classes. The class symbols are given a CompleteEnter object as completer.
 *
 *  In the second phase classes are completed using CompleteEnter.complete().
 *  Completion might occur on demand, but any classes that are not completed that
 *  way will be eventually completed by processing the `uncompleted' queue.
 *  Completion entails (1) determination of a classe's parameters, supertype and
 *  interfaces, as well as (2) entering all symbols defined in the class into its
 *  scope,  with the exception of class symbols which have been entered in phase 1.
 *  (2) depends on (1) having been completed for a class and all its superclasses
 *  and enclosing classes. That's why, after doing (1), we put classes in a
 *  `halfcompleted' queue. Only when we have performed (1) for a class and all it's
 *  superclasses and enclosing classes, we proceed to (2).
 *
 *  Whereas the first phase is organized as a sweep through all compiled
 *  syntax trees,  the second phase is demand. Members of a class are entered
 *  when the contents of a class are first accessed. This is accomplished
 *  by installing completer objects in class symbols for compiled classes
 *  which invoke the member-enter phase for the corresponding class tree.
 *
 *  Classes migrate from one phase to the next via queues:
 *
 *  class enter -> (uncompleted)   --> member enter (1)
 *              -> (halfcompleted) --> member enter (2)
 *              -> (todo)          --> attribute // only for toplevel classes
 */
public class Enter extends Tree.Visitor implements Flags, Kinds, TypeTags {
    Log log;
    Symtab syms;
    Resolve rs;
    Check chk;
    TreeMaker make;
    Attr attr;
    MemberEnter phase2;

    /**
     * A queue of all as yet unattributed toplevel classes
     */
    ListBuffer todo;

    public Enter(Log log, Symtab syms, Resolve rs, Check chk, TreeMaker make,
            Attr attr, ListBuffer todo) {
        super();
        this.log = log;
        this.syms = syms;
        this.rs = rs;
        this.chk = chk;
        this.make = make;
        this.attr = attr;
        if (attr != null)
            attr.enter = this;
        this.todo = todo;
        phase2 = new MemberEnter();
        predefClassDef = new ClassDef(PUBLIC, syms.predefClass.name, null, null, null,
                null, syms.predefClass);
    }

    /**
      * A switch to determine whether we check for package/class conflicts
      */
    static final boolean checkClash = true;

    /**
     * A hashtable mapping classes to the environments current at the point
     *  of the class declaration's local definitions.
     */
    Hashtable classEnvs = Hashtable.make();

    /**
     * The queue of all classes that might still need to be completed;
     *  saved and initialized by main().
     */
    ListBuffer uncompleted;

    /**
     * A queue for classes whose members still need to be entered into the
     *  symbol table.
     */
    ListBuffer halfcompleted = new ListBuffer();

    /**
     * A dummy class to serve as enclClass for toplevel environments.
     */
    private ClassDef predefClassDef;

    /**
     * A flag to enable/disable class completion. This is
     *  necessary to prevent false cyclic dependencies involving imports.
     *  Example (due to Todd Turnbridge): Consider the following three files:
     *
     *  A.java: public class A extends B {}
     *  B.java: public class B extends C {}
     *  C.java: import A; public class C {}
     *
     *  Now compile B.java. The (import A;) is not allowed to go beyond enter
     *  phase, or a false cycle will occur.
     */
    private boolean completionEnabled = true;

    /**
     * Generate call to superclass constructor. This is:
     *
     *    super(id_0, ..., id_n)
     *
     * or, if based == true
     *
     *    id_0.super(id_1,...,id_n)
     *
     *  where id_0, ..., id_n are the names of the given parameters.
     *
     *  @param make    The tree factory
     *  @param params  The parameters that need to be passed to super
     *  @param based   Is first parameter a this$n?
     */
    static Tree SuperCall(TreeMaker make, List params, boolean based) {
        Tree meth;
        if (based) {
            meth = make.Select(make.Ident((VarDef) params.head), Names._super);
            params = params.tail;
        } else {
            meth = make.Ident(Names._super);
        }
        return make.Exec(make.Apply(meth, make.Idents(params)));
    }

    /**
      * Generate default constructor for given class. For classes different
      *  from java.lang.Object, this is:
      *
      *    c(argtype_0 x_0, ..., argtype_n x_n) throws thrown {
      *      super(x_0, ..., x_n)
      *    }
      *
      *  or, if based == true:
      *
      *    c(argtype_0 x_0, ..., argtype_n x_n) throws thrown {
      *      x_0.super(x_1, ..., x_n)
      *    }
      *
      *  @param make     The tree factory.
      *  @param c        The class owning the default constructor.
      *  @param argtypes The parameter types of the constructor.
      *  @param thrown   The thrown exceptions of the constructor.
      *  @param based    Is first parameter a this$n?
      */
    Tree DefaultConstructor(TreeMaker make, ClassSymbol c, List argtypes,
            List thrown, boolean based) {
        List params = make.Params(argtypes, Symbol.noSymbol);
        List stats = Tree.emptyList;
        if (c.type != syms.objectType)
            stats = stats.prepend(SuperCall(make, params, based));
        ListBuffer thrownbuf = new ListBuffer();
        for (List l = thrown; l.nonEmpty(); l = l.tail)
            thrownbuf.append(make.Ident((Symbol) l.head));
        Tree result = make.MethodDef(c.flags() & Check.AccessFlags, Names.init, null,
                TypeParameter.emptyList, params, thrownbuf.toList(),
                make.Block(0, stats));
        return result;
    }

    /**
      * Report duplicate declaration error.
      */
    private void duplicateError(int pos, Symbol sym) {
        if (!sym.type.isErroneous()) {
            log.error(pos, "already.defined", sym.toJava(), sym.javaLocation());
        }
    }

    /**
      * Check that symbol is unique in given scope.
      *  @param pos           Position for error reporting.
      *  @param sym           The symbol.
      *  @param s             The scope.
      */
    private boolean checkUnique(int pos, Symbol sym, Scope s) {
        for (Scope.Entry e = s.lookup(sym.name); e.scope == s; e = e.next()) {
            if (sym != e.sym && sym.kind == e.sym.kind && sym.name != Names.error &&
                    (sym.kind != MTH || sym.type.hasSameArgs(e.sym.type))) {
                duplicateError(pos, e.sym);
                return false;
            }
        }
        return true;
    }

    /**
      * Check that variable does not hide variable with same name in
      *  immediately enclosing local scope.
      *  @param pos           Position for error reporting.
      *  @param sym           The symbol.
      *  @param s             The scope.
      */
    private void checkTransparentVar(int pos, VarSymbol v, Scope s) {
        if (s.next != null) {
            for (Scope.Entry e = s.next.lookup(v.name);
                    e.scope != null && e.sym.owner == v.owner; e = e.next()) {
                if (e.sym.kind == VAR && (e.sym.owner.kind & (VAR | MTH)) != 0 &&
                        v.name != Names.error) {
                    duplicateError(pos, e.sym);
                    return;
                }
            }
        }
    }

    /**
      * Check that a class or interface does not hide a class or
      *  interface with same name in immediately enclosing local scope.
      *  @param pos           Position for error reporting.
      *  @param sym           The symbol.
      *  @param s             The scope.
      */
    private void checkTransparentClass(int pos, ClassSymbol c, Scope s) {
        if (s.next != null) {
            for (Scope.Entry e = s.next.lookup(c.name);
                    e.scope != null && e.sym.owner == c.owner; e = e.next()) {
                if (e.sym.kind == TYP && (e.sym.owner.kind & (VAR | MTH)) != 0 &&
                        c.name != Names.error) {
                    duplicateError(pos, e.sym);
                    return;
                }
            }
        }
    }

    /**
      * Check that class does not have the same name as one of
      *  its enclosing classes, or as a class defined in its enclosing scope.
      *  return true if class is unique in its enclosing scope.
      *  @param pos           Position for error reporting.
      *  @param name          The class name.
      *  @param s             The enclosing scope.
      */
    private boolean checkUniqueClassName(int pos, Name name, Scope s) {
        for (Scope.Entry e = s.lookup(name); e.scope == s; e = e.next()) {
            if (e.sym.kind == TYP && e.sym.name != Names.error) {
                duplicateError(pos, e.sym);
                return false;
            }
        }
        for (Symbol sym = s.owner; sym != null; sym = sym.owner) {
            if (sym.kind == TYP && sym.name == name && sym.name != Names.error) {
                duplicateError(pos, sym);
                return true;
            }
        }
        return true;
    }

    /**
      * Check that name does not clash with internal names used in the
      *  translation process.
      */
    private void checkNotReserved(int pos, Name name) {
        if (name == Names.classDollar || name.startsWith(Names.thisDollar)) {
            log.error(pos, "name.reserved.for.internal.use", name.toJava());
        }
    }

    /**
      * Create a fresh environment for method bodies.
      *  @param tree     The method definition.
      *  @param env      The environment current outside of the method definition.
      */
    Env methodEnv(MethodDef tree, Env env) {
        Env localEnv = env.dup(tree,
                ((AttrContext) env.info).dup(
                ((AttrContext) env.info).scope.dupUnshared()));
        localEnv.enclMethod = tree;
        ((AttrContext) localEnv.info).scope.owner = tree.sym;
        if ((tree.flags & STATIC) != 0)
            ((AttrContext) localEnv.info).staticLevel++;
        return localEnv;
    }

    /**
      * Create a fresh environment for class bodies.
      *  This will create a fresh scope for local symbols of a class, referred
      *  to by the environments info.scope field.
      *  This scope will contain
      *    - symbols for this and super
      *    - symbols for any type parameters
      *  In addition, it serves as an anchor for scopes of methods and initializers
      *  which are nested in this scope via Scope.dup().
      *  This scope should not be confused with the members scope of a class.
      *
      *  @param tree     The class definition.
      *  @param env      The environment current outside of the class definition.
      */
    public Env classEnv(ClassDef tree, Env env) {
        Env localEnv =
                env.dup(tree, ((AttrContext) env.info).dup(new Scope(tree.sym)));
        localEnv.enclClass = tree;
        localEnv.outer = env;
        ((AttrContext) localEnv.info).isSelfCall = false;
        return localEnv;
    }

    /**
      * Create a fresh environment for toplevels.
      *  @param tree     The toplevel tree.
      */
    Env topLevelEnv(TopLevel tree) {
        Env localEnv = new Env(tree, new AttrContext());
        localEnv.toplevel = tree;
        localEnv.enclClass = predefClassDef;
        tree.namedImportScope = new Scope(tree.packge);
        tree.starImportScope = new Scope(tree.packge);
        ((AttrContext) localEnv.info).scope = tree.namedImportScope;
        return localEnv;
    }

    /**
      * Create a fresh environment a variable's initializer.
      *  If the variable is a field, the owner of the environment's scope
      *  is be the variable itself, otherwise the owner is the method
      *  enclosing the variable definition.
      *
      *  @param tree     The variable definition.
      *  @param env      The environment current outside of the variable definition.
      */
    Env initEnv(VarDef tree, Env env) {
        Env localEnv = env.dup(tree, ((AttrContext) env.info).dup());
        if (tree.sym.owner.kind == TYP) {
            ((AttrContext) localEnv.info).scope =
                    ((AttrContext) env.info).scope.dup();
            ((AttrContext) localEnv.info).scope.owner = tree.sym;
        }
        if ((tree.flags & STATIC) != 0 ||
                (env.enclClass.sym.flags() & INTERFACE) != 0)
            ((AttrContext) localEnv.info).staticLevel++;
        return localEnv;
    }

    /**
      * The scope in which a member definition in environment env is to be entered
      *  This is usually the environment's scope, except for class environments,
      *  where the local scope is for type variables, and the this and super symbol
      *  only, and members go into the class member scope.
      */
    Scope enterScope(Env env) {
        return env.tree.tag == Tree.CLASSDEF ?
                ((ClassDef) env.tree).sym.members_field :
                ((AttrContext) env.info).scope;
    }

    /**
      * Visitor argument: the current environment.
      */
    Env env;

    /**
     * Visitor result: the computed type.
     */
    Type result;

    /**
     * Visitor method: enter all classes in given tree, catching any
     *  completion failure exceptions. Return the tree's type.
     *
     *  @param tree    The tree to be visited.
     *  @param env     The environment visitor argument.
     */
    Type classEnter(Tree tree, Env env) {
        Env prevEnv = this.env;
        try {
            {
                this.env = env;
                tree.visit(this);
                this.env = prevEnv;
                return result;
            }
        } catch (CompletionFailure ex) {
            {
                this.env = prevEnv;
                return chk.completionError(tree.pos, ex);
            }
        }
    }

    /**
      * Visitor method: enter classes of a list of trees, returning a list of types.
      */
    List classEnter(List trees, Env env) {
        ListBuffer ts = new ListBuffer();
        for (List l = trees; l.nonEmpty(); l = l.tail)
            ts.append(classEnter((Tree) l.head, env));
        return ts.toList();
    }

    /**
      * Class enter visitor method for toplevels.
      */
    public void _case(TopLevel tree) {
        Name prev = log.useSource(tree.sourcefile);
        if (tree.pid != null) {
            tree.packge = syms.reader.enterPackage(TreeInfo.fullName(tree.pid));
        } else {
            tree.packge = Symbol.emptyPackage;
        }
        tree.packge.complete();
        classEnter(tree.defs, topLevelEnv(tree));
        log.useSource(prev);
        result = null;
    }

    /**
      * Class enter visitor method for class definitions:
      */
    public void _case(ClassDef tree) {
        Symbol owner = ((AttrContext) env.info).scope.owner;
        Scope enclScope = enterScope(env);
        ClassSymbol c;
        if (owner.kind == PCK) {
            PackageSymbol packge = (PackageSymbol) owner;
            c = syms.reader.enterClass(tree.name, packge);
            packge.members().enterIfAbsent(c);
            if ((tree.flags & PUBLIC) != 0 && !classNameMatchesFileName(c, env)) {
                log.error(tree.pos, "class.public.should.be.in.file",
                        tree.name.toJava());
            }
        } else {
            if (tree.name.len != 0 &&
                    !checkUniqueClassName(tree.pos, tree.name, enclScope)) {
                result = null;
                return;
            }
            if (owner.kind == TYP) {
                c = syms.reader.enterClass(tree.name, (TypeSymbol) owner);
                if ((owner.flags_field & INTERFACE) != 0) {
                    tree.flags |= PUBLIC | STATIC;
                }
            } else {
                c = syms.reader.defineClass(tree.name, owner);
                c.flatname = chk.localClassName(c, 1);
                if (c.name.len != 0)
                    checkTransparentClass(tree.pos, c,
                            ((AttrContext) env.info).scope);
            }
        }
        if (chk.compiled.get(c.flatname) != null) {
            log.error(tree.pos, "duplicate.class", c.fullname.toJava());
            result = null;
            return;
        }
        chk.compiled.put(c.flatname, c);
        enclScope.enter(c);
        tree.sym = c;
        Env localEnv = classEnv(tree, env);
        classEnvs.put(c, localEnv);
        c.completer = new CompleteEnter();
        c.flags_field = chk.checkFlags(tree.pos, tree.flags, c);
        c.sourcefile = env.toplevel.sourcefile;
        c.members_field = new Scope(c);
        ClassType ct = (ClassType) c.type;
        if (owner.kind != PCK && ((c.flags_field & STATIC) == 0 ||
                (c.flags_field & INTERFACE) != 0)) {
            Symbol owner1 = owner;
            while ((owner1.kind & (VAR | MTH)) != 0 &&
                    (owner1.flags_field & STATIC) == 0) {
                owner1 = owner1.owner;
            }
            if (owner1.kind == TYP) {
                ct.outer_field = owner1.type;
            }
        }
        ct.typarams_field = classEnter(tree.typarams, localEnv);
        if (!c.isLocal() && uncompleted != null)
            uncompleted.append(c);
        classEnter(tree.defs, localEnv);
        result = c.type;
    }

    /**
      * Does class have the same name as the file it appears in?
      */
    private static boolean classNameMatchesFileName(ClassSymbol c, Env env) {
        String fname = env.toplevel.sourcefile.toString();
        String cname = c.name + ".java";
        try {
            {
                return endsWith(fname, cname) ||
                        endsWith(new File(fname).getCanonicalPath(), cname);
            }
        } catch (java.io.IOException ex) {
            {
                return false;
            }
        }
    }

    /**
      * Does path name have file name as last component?
      */
    private static boolean endsWith(String pathname, String filename) {
        return pathname.endsWith(filename) &&
                (pathname.length() == filename.length() ||
                pathname.charAt(pathname.length() - filename.length() - 1) ==
                File.separatorChar);
    }

    /**
      * Default class enter visitor method: do nothing.
      */
    public void _case(Tree tree) {
        result = null;
    }

    /**
      * Main method: enter all classes in a list of toplevel trees and
      *  add to todo list for later attribution.
      *  @param trees      The list of trees to be processed.
      */
    public void main(List trees) {
        ListBuffer prevUncompleted = uncompleted;
        uncompleted = new ListBuffer();
        classEnter(trees, null);
        List tocomplete = uncompleted.toList();
        if (completionEnabled) {
            for (List l = tocomplete; l.nonEmpty(); l = l.tail)
                ((ClassSymbol) l.head).complete();
        }
        uncompleted = prevUncompleted;
    }

    /**
      * Enter all members of a class. This is done in a second phase
      *  after the classes themselves have been entered.
      */
    class MemberEnter extends Tree.Visitor {

        MemberEnter() {
            super();
        }

        /**
          * Given a class, and an (import) scope, is there already a
          *  class with same fully qualified name in this scope?
          */
        private boolean isIncluded(Symbol c, Scope scope) {
            for (Scope.Entry e = scope.lookup(c.name); e.scope == scope;
                    e = e.next()) {
                if (e.sym.kind == c.kind && e.sym.fullName() == c.fullName())
                    return true;
            }
            return false;
        }

        /**
          * Import all classes of a class or package on demand.
          *  @param pos       Position to be used for error reporting.
          *  @param tsym      The class or package the members of which are imported.
          *  @param toScope   The (import) scope in which imported classes
          *                   are entered.
          */
        private void importAll(int pos, TypeSymbol tsym, Scope toScope) {
            if (tsym.kind == PCK && !tsym.exists())
                if (((PackageSymbol) tsym).fullname.equals(Names.java_lang)) {
                    String msg = Log.getLocalizedString("fatal.err.no.java.lang");
                    throw new FatalError(msg);
                } else {
                    log.error(pos, "doesnt.exist", tsym.toJava());
                }
            Scope fromScope = tsym.members();
            for (Scope.Entry e = fromScope.elems; e != null; e = e.sibling)
                if (e.sym.kind == TYP && !isIncluded(e.sym, toScope))
                    toScope.enter(e.sym, fromScope);
        }

        /**
          * Import given class.
          *  @param pos       Position to be used for error reporting.
          *  @param tsym      The class to be imported.
          *  @param toScope   The (import) scope in which the imported class
          *                   is entered.
          */
        private void importNamed(int pos, Symbol tsym, Scope toScope) {
            if (tsym.kind == TYP && checkUnique(pos, tsym, toScope))
                toScope.enter(tsym, tsym.owner.members());
        }

        /**
          * Construct method type from method signature.
          *  @param typarams    The method's type parameters.
          *  @param params      The method's value parameters.
          *  @param res         The method's result type,
          *                     null if it is a constructor.
          *  @param thrown      The method's thrown exceptions.
          *  @param env         The method's (local) environment.
          */
        Type signature(List typarams, List params, Tree res, List thrown, Env env) {
            ListBuffer argbuf = new ListBuffer();
            for (List l = params; l.nonEmpty(); l = l.tail)
                argbuf.append(attr.attribType(((VarDef) l.head).vartype, env));
            Type restype = res == null ? Type.voidType : attr.attribType(res, env);
            ListBuffer thrownbuf = new ListBuffer();
            for (List l = thrown; l.nonEmpty(); l = l.tail)
                thrownbuf.append(
                        (ClassSymbol) chk.checkClassType(((Tree) l.head).pos,
                        attr.attribType((Tree) l.head, env)).tsym);
            Type mtype = new MethodType(argbuf.toList(), restype, thrownbuf.toList());
            return mtype;
        }

        /**
          * Visitor argument: the current environment
          */
        Env env;

        /**
         * Visitor method: enter field and method definitions and process import
         *  clauses, catching any completion failure exceptions.
         */
        void memberEnter(Tree tree, Env env) {
            Env prevEnv = this.env;
            try {
                {
                    this.env = env;
                    tree.visit(this);
                    this.env = prevEnv;
                }
            } catch (CompletionFailure ex) {
                {
                    this.env = prevEnv;
                    chk.completionError(tree.pos, ex);
                }
            }
        }

        /**
          * Visitor method: enter members in a list of trees.
          */
        void memberEnter(List trees, Env env) {
            for (List l = trees; l.nonEmpty(); l = l.tail)
                memberEnter((Tree) l.head, env);
        }

        /**
          * Member enter visitor method for toplevels.
          */
        public void _case(TopLevel tree) {
            if (checkClash && tree.pid != null) {
                Symbol p = tree.packge;
                while (p.owner != Symbol.rootPackage) {
                    p.owner.complete();
                    if (syms.reader.classes.get(p.fullName()) != null) {
                        log.error(tree.pos, "pkg.clashes.with.class.of.same.name",
                                p.toJava());
                    }
                    p = p.owner;
                }
            }
            importAll(tree.pos, syms.reader.enterPackage(Names.java_lang),
                    env.toplevel.starImportScope);
            memberEnter(tree.defs, env);
        }

        /**
          * Member enter visitor method for import clauses.
          */
        public void _case(Import tree) {
            Tree imp = tree.qualid;
            Name name = TreeInfo.name(imp);
            TypeSymbol p;
            completionEnabled = false;
            if (imp.tag == Tree.SELECT) {
                p = attr.attribTree(((Select) imp).selected, env, TYP | PCK,
                        Type.noType).tsym;
                if (name == Names.asterisk) {
                    importAll(tree.pos, p, env.toplevel.starImportScope);
                } else {
                    TypeSymbol c = attr.attribType(imp, env).tsym;
                    importNamed(tree.pos, c, env.toplevel.namedImportScope);
                }
            } else {
                importNamed(tree.pos,
                        rs.access(
                        rs.findIdentInPackage(env, Symbol.emptyPackage, name,
                        TYP), tree.pos, env.enclClass.sym.type, name),
                        env.toplevel.namedImportScope);
            }
            completionEnabled = true;
        }

        /**
          * Member enter visitor method for method definitions.
          */
        public void _case(MethodDef tree) {
            Scope enclScope = enterScope(env);
            MethodSymbol m = new MethodSymbol(0, tree.name, null, enclScope.owner);
            m.flags_field = chk.checkFlags(tree.pos, tree.flags, m);
            tree.sym = m;
            Env localEnv = methodEnv(tree, env);
            m.type = signature(tree.typarams, tree.params, tree.restype,
                    tree.thrown, localEnv);
            ((AttrContext) localEnv.info).scope.leave();
            if (checkUnique(tree.pos, m, enclScope)) {
                checkNotReserved(tree.pos, tree.name);
                enclScope.enter(m);
            }
        }

        /**
          * Member enter visitor method for variable definitions.
          */
        public void _case(VarDef tree) {
            attr.attribType(tree.vartype, env);
            Scope enclScope = enterScope(env);
            VarSymbol v =
                    new VarSymbol(0, tree.name, tree.vartype.type, enclScope.owner);
            v.flags_field = chk.checkFlags(tree.pos, tree.flags, v);
            tree.sym = v;
            if (tree.init != null) {
                v.flags_field |= HASINIT;
                if ((v.flags_field & FINAL) != 0)
                    v.constValue = initEnv(tree, env);
            }
            if (checkUnique(tree.pos, v, enclScope)) {
                checkNotReserved(tree.pos, tree.name);
                checkTransparentVar(tree.pos, v, enclScope);
                enclScope.enter(v);
            }
            v.pos = tree.pos;
        }

        /**
          * Default member enter visitor method: do nothing
          */
        public void _case(Tree tree) {
        }
    }

    /**
      * A completer class for source classes
      */
    class CompleteEnter implements Completer {

        CompleteEnter() {
            super();
        }

        /**
          * Complete entering a class.
          *  @param sym     The symbol of the class to be completed.
          */
        public void complete(Symbol sym) throws CompletionFailure {
            if (!completionEnabled) {
                sym.completer = this;
                return;
            }
            ClassSymbol c = (ClassSymbol) sym;
            ClassType ct = (ClassType) c.type;
            Env env = (Env) classEnvs.get(c);
            ClassDef tree = (ClassDef) env.tree;
            Name prev = log.useSource(env.toplevel.sourcefile);
            boolean isFirst = halfcompleted.isEmpty();
            try {
                {
                    halfcompleted.append(env);
                    if (c.owner.kind == PCK) {
                        phase2.memberEnter(env.toplevel,
                                env.enclosing(Tree.TOPLEVEL));
                        todo.append(env);
                    }
                    c.flags_field |= LOCKED | UNATTRIBUTED;
                    if (c.owner.kind == TYP) {
                        c.owner = checkNonCyclic(tree.pos, c.owner.type, env).tsym;
                        c.owner.complete();
                    }
                    Type st;
                    if (c.fullname == Names.java_lang_Object)
                        st = Type.noType;
                    else if (tree.extending == null)
                        st = syms.objectType;
                    else
                        st = attribBase(tree.extending, env, 0);
                    ct.supertype_field = st;
                    ListBuffer interfaces = new ListBuffer();
                    Set interfaceSet = Set.make();
                    for (List l = tree.implementing; l.nonEmpty(); l = l.tail) {
                        Type i = attribBase((Tree) l.head, env, INTERFACE);
                        if (i.tag == CLASS) {
                            interfaces.append(i);
                            if (interfaceSet.contains(i))
                                log.error(((Tree) l.head).pos, "repeated.interface");
                            else
                                interfaceSet.put(i);
                        }
                    }
                    ct.interfaces_field = interfaces.toList();
                    c.flags_field &= ~LOCKED;
                    attr.attribStats(tree.typarams, env);
                    if ((c.flags() & INTERFACE) == 0 &&
                            !TreeInfo.hasConstructors(tree.defs)) {
                        List argtypes = Type.emptyList;
                        List thrown = ClassSymbol.emptyList;
                        boolean based = false;
                        if (c.name.len == 0) {
                            NewClass nc = (NewClass) env.next.tree;
                            if (nc.constructor != null) {
                                Type superConstrType =
                                        c.type.memberType(nc.constructor);
                                argtypes = superConstrType.argtypes();
                                if (nc.encl != null) {
                                    argtypes = argtypes.prepend(nc.encl.type);
                                    based = true;
                                }
                                thrown = superConstrType.thrown();
                            }
                        }
                        Tree constrDef = DefaultConstructor(make.at(tree.pos), c,
                                argtypes, thrown, based);
                        tree.defs = tree.defs.prepend(constrDef);
                    }
                    if ((c.flags_field & INTERFACE) == 0) {
                        VarSymbol thisSym =
                                new VarSymbol(FINAL | HASINIT, Names._this,
                                c.type, c);
                        thisSym.pos = Position.FIRSTPOS;
                        ((AttrContext) env.info).scope.enter(thisSym);
                        if (st.tag == CLASS) {
                            VarSymbol superSym =
                                    new VarSymbol(FINAL | HASINIT, Names._super,
                                    st, c);
                            superSym.pos = Position.FIRSTPOS;
                            ((AttrContext) env.info).scope.enter(superSym);
                        }
                    }
                }
            } catch (CompletionFailure ex) {
                {
                    chk.completionError(tree.pos, ex);
                }
            }
            if (checkClash && c.owner.kind == PCK &&
                    syms.reader.enterPackage(c.fullname).exists()) {
                log.error(tree.pos, "clash.with.pkg.of.same.name", c.toJava());
            }
            log.useSource(prev);
            if (isFirst) {
                while (halfcompleted.nonEmpty()) {
                    finish((Env) halfcompleted.first());
                    halfcompleted.remove();
                }
            }
        }

        /**
          * Enter member fields and methods of a class
          *  @param env    the environment current for the class block.
          */
        private void finish(Env env) {
            Name prev = log.useSource(env.toplevel.sourcefile);
            ClassDef tree = (ClassDef) env.tree;
            ClassSymbol c = tree.sym;
            phase2.memberEnter(tree.defs, env);
            log.useSource(prev);
        }

        /**
          * Attribute extended or implemented type reference, and check
          *  for cyclic references.
          *  @param tree          The tree expressing the type reference.
          *  @param env           The environment current at the type reference.
          *  @param interfaceFlag This is INTERFACE if we expect to reference an
          *                       interface, 0 if we expect to reference a class.
          */
        private Type attribBase(Tree tree, Env env, int interfaceFlag) {
            Type t = attr.attribBase(tree, env, interfaceFlag);
            return checkNonCyclic(tree.pos, t, env);
        }

        /**
          * Check for cyclic references. Issue an error if the
          *  symbol of the type referred to has a LOCKED flag set.
          *
          *  @param pos      Position to be used for error reporting.
          *  @param t        The type referred to.
          *  @param env      The environment current at the reference.
          */
        private Type checkNonCyclic(int pos, Type t, Env env) {
            if ((t.tsym.flags() & LOCKED) != 0) {
                log.error(pos, "cyclic.inheritance", t.tsym.toJava());
                return Type.errType;
            } else {
                return t;
            }
        }
    }
}
