package com.sun.tools.javac.v8.comp;
import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.tree.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.tree.Tree.*;

import com.sun.tools.javac.v8.code.Type.*;


/**
 * This pass implements the control- and dataflow analysis for Java
 *  programs which is mandated by the ``reachable statements'' and ``definite
 *  assignments'' sections in the Java specification.
 */
public class Flow extends Tree.Visitor implements Flags, Kinds, TypeTags {
    Log log;
    Symtab syms;
    Check chk;

    public Flow(Log log, Symtab syms, Check chk) {
        super();
        this.log = log;
        this.syms = syms;
        this.chk = chk;
    }

    /**
      * The set of definitely assigned variables.
      */
    Bits inits = new Bits();

    /**
     * The set of definitely unassigned variables.
     */
    Bits uninits = new Bits();

    /**
     * The set of variables that are definitely unassigned anywhere
     *  in current try block. This variable is maintained lazily
     *  It is updated only when something gets added to uninits,
     *  (typically via an inclRange(firstAdr,nextAdr) issued because code is
     *  unreachable). To obtain the correct set of variables
     *  which are definitely unassigned anywhere
     *  in current try block, intersect uninitsTry and uninits.
     */
    Bits uninitsTry = new Bits();

    /**
     * When analyzing a condition, inits and uninits are null.
     *  Instead we have:
     */
    Bits initsWhenTrue;
    Bits initsWhenFalse;
    Bits uninitsWhenTrue;
    Bits uninitsWhenFalse;

    /**
     * A mapping from addresses to variable symbols.
     */
    VarSymbol[] vars = new VarSymbol[32];

    /**
     * The list of possibly thrown declarable exceptions.
     */
    List thrown;

    /**
     * The list of exceptions that are either caught or declared to be
     *  thrown.
     */
    List reported;

    /**
     * A flag that indicates whether the last statement could complete normally.
     */
    boolean alive;

    /**
     * The first variable sequence number in this class definition.
     */
    int firstadr = 0;

    /**
     * The next available variable sequence number.
     */
    int nextadr = 0;

    /**
     * A mapping from trees to the combined inits/uninits of their
     *  break statements.
     */
    Hashtable breakTargetStates;

    /**
     * A mapping from trees to the combined inits/uninits of their
     *  continue statements
     */
    Hashtable continueTargetStates;

    /**
     * The enclosing class symbol
     */
    ClassSymbol enclClass;

    /**
     * Do we need to track init/uninit state of this symbol?
     *  I.e. is symbol either a local or a blank final variable?
     */
    boolean trackable(Symbol sym) {
        return sym != null && sym.kind == VAR &&
                (sym.owner.kind == MTH || sym.owner == enclClass &&
                (sym.flags() & (FINAL | HASINIT)) == FINAL);
    }

    /**
      * Initialize new trackable variable by setting its address field
      *  to the next available sequence number and entering it under that
      *  index into the vars array.
      */
    void newVar(VarSymbol sym) {
        if (nextadr == vars.length) {
            VarSymbol[] newvars = new VarSymbol[nextadr * 2];
            System.arraycopy(vars, 0, newvars, 0, nextadr);
            vars = newvars;
        }
        sym.adr = nextadr;
        vars[nextadr] = sym;
        nextadr++;
    }

    /**
      * Record an initialization of a trackable variable.
      */
    void letInit(int pos, VarSymbol sym) {
        if (inits == null) {
            Bits initsWhenTrue1 = initsWhenTrue.dup();
            Bits initsWhenFalse1 = initsWhenFalse.dup();
            Bits uninitsWhenTrue1 = uninitsWhenTrue.dup();
            Bits uninitsWhenFalse1 = uninitsWhenFalse.dup();
            merge();
            Bits uninits1 = uninits.dup();
            letInit(pos, sym);
            initsWhenTrue = initsWhenTrue1.orSet(inits);
            initsWhenFalse = initsWhenFalse1.orSet(inits);
            Bits diff = uninits1.diffSet(uninits);
            uninitsWhenTrue = uninitsWhenTrue1.diffSet(diff);
            uninitsWhenFalse = uninitsWhenFalse1.diffSet(diff);
            inits = null;
            uninits = null;
        } else {
            inits.incl(sym.adr);
            if ((sym.flags() & FINAL) != 0 && !uninits.isMember(sym.adr)) {
                log.error(pos, "var.might.already.be.assigned", sym.toJava());
            }
            uninits.excl(sym.adr);
        }
    }

    /**
      * If tree is either a simple name or of the form this.name or C.this.name, and
      *  tree represents a trackable variable,
      *  record an initialization of the variable.
      */
    void letInit(Tree tree) {
        if (tree.tag == Tree.IDENT || tree.tag == Tree.SELECT &&
                TreeInfo.name(((Select) tree).selected) == Names._this) {
            Symbol sym = TreeInfo.symbol(tree);
            if (trackable(sym))
                letInit(tree.pos, (VarSymbol) sym);
        }
    }

    /**
      * Check that trackable variable is initialized.
      */
    void checkInit(int pos, VarSymbol sym) {
        if (!inits.isMember(sym.adr)) {
            log.error(pos, "var.might.not.have.been.initialized", sym.toJava());
            inits.incl(sym.adr);
        }
    }

    /**
      * Check that branch back to previous init state does not
      *  assign any blank finals.
      *  @param pos     Position tom be used for error reporting.
      *  @param uninitsEntry    The set of uninitialized variables on loop entry.
      *  @param uninitsEntry    The set of uninitialized variables at end of loop.
      */
    void checkBackBranch(int pos, Bits uninitsEntry, Bits uninitsLoop) {
        for (int i = firstadr; i < nextadr; i++) {
            VarSymbol sym = vars[i];
            if (sym != null && (sym.flags() & FINAL) != 0 &&
                    uninitsEntry.isMember(i) && !uninitsLoop.isMember(i)) {
                log.error(pos, "var.might.be.assigned.in.loop", sym.toJava());
            }
        }
    }

    /**
      * Is tree a constant false expression?
      */
    boolean isFalse(Tree tree) {
        return tree.type.tag == BOOLEAN && tree.type.constValue != null &&
                ((Integer) tree.type.constValue).intValue() == 0;
    }

    /**
      * Is tree a constant true expression?
      */
    boolean isTrue(Tree tree) {
        return tree.type.tag == BOOLEAN && tree.type.constValue != null &&
                ((Integer) tree.type.constValue).intValue() != 0;
    }

    /**
      * Merge current init/uninit state with entry for target of break or continue,
      *  or create a new entry if nothing was entered for target yet.
      *  `initTable' is one of `breakTargetStates' or `continueTargetStates',
      *  depending on whether tree is a break or a continue.
      */
    void jump(Tree target, Hashtable initTable) {
        Pair initState = (Pair) initTable.get(target);
        if (initState == null) {
            initTable.put(target, new Pair(inits.dup(), uninits.dup()));
        } else {
            ((Bits) initState.fst).andSet(inits);
            ((Bits) initState.snd).andSet(uninits);
        }
    }

    /**
      * Resolve all jumps to target of break or continue.
      *  `initTable' is one of `breakTargetStates' or `continueTargetStates',
      *  depending on whether tree is a break or a continue.
      */
    void resolve(Tree target, Hashtable initTable) {
        Pair initState = (Pair) initTable.get(target);
        if (initState != null) {
            inits.andSet((Bits) initState.fst);
            uninits.andSet((Bits) initState.snd);
            alive = true;
        }
    }

    /**
      * Record that exception is potentially thrown and check that it is reported.
      */
    void markThrown(int pos, ClassSymbol exc) {
        if (!chk.isUnchecked(exc)) {
            chk.checkHandled(pos, exc, reported);
            thrown = chk.incl(exc, thrown);
        }
    }

    /**
      * Record that statement is unreachable.
      */
    void markDead() {
        inits.inclRange(firstadr, nextadr);
        uninitsTry.andSet(uninits);
        uninits.inclRange(firstadr, nextadr);
        alive = false;
    }

    /**
      * Split (duplicate) inits/uninits into WhenTrue/WhenFalse sets
      */
    void split() {
        initsWhenFalse = inits.dup();
        uninitsWhenFalse = uninits.dup();
        initsWhenTrue = inits;
        uninitsWhenTrue = uninits;
    }

    /**
      * Merge (intersect) inits/uninits from WhenTrue/WhenFalse sets.
      */
    void merge() {
        inits = initsWhenFalse.andSet(initsWhenTrue);
        uninits = uninitsWhenFalse.andSet(uninitsWhenTrue);
    }

    /**
      * Mode tags.
      */
    static final int STATmode = 0;

    /**
     * Mode tags.
     */
    static final int EXPRmode = 1;

    /**
     * Mode tags.
     */
    static final int VARmode = 2;

    /**
     * Mode tags.
     */
    static final int CONDmode = 3;

    /**
     * Visitor argument: a mode
     */
    int mode;

    /**
     * Visitor method: Perform flow analysis on tree. (JLS)
     */
    void analyze(Tree tree, int mode) {
        if (tree != null) {
            int prevMode = this.mode;
            this.mode = mode;
            tree.visit(this);
            this.mode = prevMode;
        }
    }

    /**
      * Analyze an assigned variable (i.e. OK if not initialized)
      */
    void analyzeVar(Tree tree) {
        analyze(tree, VARmode);
    }

    /**
      * Analyze an expression. Make sure to set (un)inits rather than
      *  (un)initsWhenTrue(WhenFalse) on exit.
      */
    void analyzeExpr(Tree tree) {
        if (tree != null) {
            analyze(tree, EXPRmode);
            if (inits == null)
                merge();
        }
    }

    /**
      * Analyze a condition. Make sure to set (un)initsWhenTrue(WhenFalse)
      *  rather than (un)inits on exit.
      */
    void analyzeCond(Tree tree) {
        if (isFalse(tree)) {
            initsWhenTrue = inits.dup();
            initsWhenTrue.inclRange(firstadr, nextadr);
            uninitsTry.andSet(uninits);
            uninitsWhenTrue = uninits.dup();
            uninitsWhenTrue.inclRange(firstadr, nextadr);
            initsWhenFalse = inits;
            uninitsWhenFalse = uninits;
        } else if (isTrue(tree)) {
            initsWhenFalse = inits.dup();
            initsWhenFalse.inclRange(firstadr, nextadr);
            uninitsTry.andSet(uninits);
            uninitsWhenFalse = uninits.dup();
            uninitsWhenFalse.inclRange(firstadr, nextadr);
            initsWhenTrue = inits;
            uninitsWhenTrue = uninits;
        } else {
            analyze(tree, CONDmode);
            if (inits != null)
                split();
        }
        inits = null;
        uninits = null;
    }

    /**
      * Analyze a definition.
      */
    public void analyzeDef(Tree tree) {
        analyze(tree, STATmode);
        if (tree != null && tree.tag == Tree.BLOCK && !alive) {
            log.error(tree.pos, "initializer.must.be.able.to.complete.normally");
        }
    }

    /**
      * Analyze a statement. Check that statement is reachable.
      */
    void analyzeStat(Tree tree) {
        if (!alive && tree != null &&
                (tree.tag != Tree.BLOCK || ((Block) tree).stats.nonEmpty())) {
            log.error(tree.pos, "unreachable.stmt");
            alive = true;
        }
        analyze(tree, STATmode);
        resolve(tree, breakTargetStates);
    }

    /**
      * Analyze list of expressions.
      */
    void analyzeExprs(List trees) {
        if (trees != null)
            for (List l = trees; l.nonEmpty(); l = l.tail)
                analyzeExpr((Tree) l.head);
    }

    /**
      * Analyze list of statements.
      */
    void analyzeStats(List trees) {
        if (trees != null)
            for (List l = trees; l.nonEmpty(); l = l.tail)
                analyzeStat((Tree) l.head);
    }

    public void _case(ClassDef tree) {
        if (tree.sym == null)
            return;
        List thrownPrev = thrown;
        List reportedPrev = reported;
        boolean alivePrev = alive;
        int firstadrPrev = firstadr;
        ClassSymbol enclClassPrev = enclClass;
        Hashtable breakTargetStatesPrev = breakTargetStates;
        Hashtable continueTargetStatesPrev = continueTargetStates;
        thrown = ClassSymbol.emptyList;
        reported = ClassSymbol.emptyList;
        for (List l = tree.defs; l.nonEmpty(); l = l.tail) {
            if (TreeInfo.isInitialConstructor((Tree) l.head)) {
                List thrown = ((MethodDef) l.head).sym.type.thrown();
                if (reported == null)
                    reported = thrown;
                else
                    reported = chk.intersect(thrown, reported);
            }
        }
        firstadr = nextadr;
        enclClass = tree.sym;
        breakTargetStates = Hashtable.make();
        continueTargetStates = Hashtable.make();
        for (List l = tree.defs; l.nonEmpty(); l = l.tail) {
            alive = true;
            if (((Tree) l.head).tag != Tree.METHODDEF &&
                    (TreeInfo.flags((Tree) l.head) & STATIC) != 0)
                analyzeDef((Tree) l.head);
        }
        for (List l = tree.defs; l.nonEmpty(); l = l.tail) {
            alive = true;
            if (((Tree) l.head).tag != Tree.METHODDEF &&
                    (TreeInfo.flags((Tree) l.head) & STATIC) == 0)
                analyzeDef((Tree) l.head);
        }
        for (List l = tree.defs; l.nonEmpty(); l = l.tail) {
            alive = true;
            if (((Tree) l.head).tag == Tree.METHODDEF)
                analyzeDef((Tree) l.head);
        }
        thrown = thrownPrev;
        reported = reportedPrev;
        alive = alivePrev;
        nextadr = firstadr;
        firstadr = firstadrPrev;
        enclClass = enclClassPrev;
        breakTargetStates = breakTargetStatesPrev;
        continueTargetStates = continueTargetStatesPrev;
    }

    public void _case(MethodDef tree) {
        if (tree.body != null) {
            Bits initsPrev = inits.dup();
            Bits uninitsPrev = uninits.dup();
            List reportedPrev = reported;
            int nextadrPrev = nextadr;
            List thrown = tree.sym.type.thrown();
            boolean isInitialConstr = TreeInfo.isInitialConstructor(tree);
            reported = isInitialConstr ? reported.prepend(thrown) : thrown;
            breakTargetStates.reset();
            continueTargetStates.reset();
            if (!isInitialConstr)
                for (int i = firstadr; i < nextadr; i++) {
                    inits.incl(i);
                    uninits.excl(i);
                }
            for (List l = tree.params; l.nonEmpty(); l = l.tail) {
                analyzeDef((Tree) l.head);
                letInit(((VarDef) l.head).pos, ((VarDef) l.head).sym);
            }
            analyzeStat(tree.body);
            int pos = tree.body.pos;
            if (alive && tree.sym.type.restype().tag != VOID)
                log.error(pos, "missing.ret.stmt");
            if (isInitialConstr)
                for (int i = firstadr; i < nextadr; i++)
                    checkInit(pos, vars[i]);
            inits = initsPrev;
            uninits = uninitsPrev;
            reported = reportedPrev;
            nextadr = nextadrPrev;
        }
    }

    public void _case(VarDef tree) {
        boolean track = trackable(tree.sym);
        if (track)
            newVar(tree.sym);
        if (tree.init != null) {
            analyzeExpr(tree.init);
            if (track) {
                inits.incl(tree.sym.adr);
                uninits.excl(tree.sym.adr);
            }
        } else if (track) {
            inits.excl(tree.sym.adr);
            uninits.incl(tree.sym.adr);
        }
    }

    public void _case(Block tree) {
        int nextadrPrev = nextadr;
        analyzeStats(tree.stats);
        nextadr = nextadrPrev;
    }

    public void _case(DoLoop tree) {
        Bits uninitsEntry = uninits.dup();
        analyzeStat(tree.body);
        resolve(tree, continueTargetStates);
        analyzeCond(tree.cond);
        checkBackBranch(tree.pos, uninitsEntry, uninitsWhenTrue);
        alive = alive && !isTrue(tree.cond);
        inits = initsWhenFalse;
        uninits = uninitsWhenFalse;
    }

    public void _case(WhileLoop tree) {
        Bits uninitsEntry = uninits.dup();
        analyzeCond(tree.cond);
        Bits initsCond = initsWhenFalse;
        Bits uninitsCond = uninitsWhenFalse;
        inits = initsWhenTrue;
        uninits = uninitsWhenTrue;
        boolean aliveWhile = alive;
        alive = alive && !isFalse(tree.cond);
        analyzeStat(tree.body);
        checkBackBranch(tree.pos, uninitsEntry, uninits);
        alive = aliveWhile && !isTrue(tree.cond);
        inits = initsCond;
        uninits = uninitsCond;
    }

    public void _case(ForLoop tree) {
        int nextadrPrev = nextadr;
        analyzeStats(tree.init);
        Bits uninitsEntry = uninits.dup();
        Bits initsCond;
        Bits uninitsCond;
        boolean aliveFor;
        if (tree.cond != null) {
            analyzeCond(tree.cond);
            initsCond = initsWhenFalse;
            uninitsCond = uninitsWhenFalse;
            inits = initsWhenTrue;
            uninits = uninitsWhenFalse;
            aliveFor = alive;
            alive = alive & !isFalse(tree.cond);
        } else {
            initsCond = inits.dup();
            initsCond.inclRange(firstadr, nextadr);
            uninitsTry.andSet(uninits);
            uninitsCond = uninits.dup();
            uninitsCond.inclRange(firstadr, nextadr);
            aliveFor = alive;
        }
        analyzeStat(tree.body);
        resolve(tree, continueTargetStates);
        analyzeExprs(tree.step);
        checkBackBranch(tree.pos, uninitsEntry, uninits);
        alive = aliveFor && tree.cond != null && !isTrue(tree.cond);
        inits = initsCond;
        uninits = uninitsCond;
        nextadr = nextadrPrev;
    }

    public void _case(Labelled tree) {
        analyzeStat(tree.body);
    }

    public void _case(Switch tree) {
        int nextadrPrev = nextadr;
        analyzeExpr(tree.selector);
        Bits initsSwitch = inits;
        Bits uninitsSwitch = uninits;
        boolean aliveSwitch = alive;
        boolean hasDefault = false;
        for (List l = tree.cases; l.nonEmpty(); l = l.tail) {
            inits = initsSwitch.dup();
            alive = aliveSwitch;
            analyzeStats(((Case) l.head).stats);
            if (((Case) l.head).pat == null)
                hasDefault = true;
        }
        if (!hasDefault) {
            alive |= aliveSwitch;
            inits.andSet(initsSwitch);
        }
        nextadr = nextadrPrev;
    }

    public void _case(Synchronized tree) {
        analyzeExpr(tree.lock);
        analyzeStat(tree.body);
    }

    public void _case(Try tree) {
        List reportedPrev = reported;
        List thrownPrev = thrown;
        Bits uninitsTryPrev = uninitsTry;
        thrown = ClassSymbol.emptyList;
        for (List l = tree.catchers; l.nonEmpty(); l = l.tail) {
            ClassSymbol exc = (ClassSymbol)((Catch) l.head).param.type.tsym;
            reported = chk.incl(exc, reported);
        }
        Bits initsTry = inits.dup();
        uninitsTry = uninits.dup();
        boolean aliveTry = alive;
        analyzeStat(tree.body);
        uninitsTry.andSet(uninits);
        Bits initsEnd = inits;
        Bits uninitsEnd = uninits;
        boolean aliveEnd = alive;
        List thrownInTry = thrown;
        thrown = thrownPrev;
        List caught = ClassSymbol.emptyList;
        int nextadrCatch = nextadr;
        reported = reportedPrev;
        for (List l = tree.catchers; l.nonEmpty(); l = l.tail) {
            VarDef param = ((Catch) l.head).param;
            ClassSymbol exc = (ClassSymbol) param.type.tsym;
            if (chk.subset(exc, caught)) {
                log.error(((Catch) l.head).pos, "except.already.caught",
                        exc.fullname.toJava());
            } else if (!chk.isUnchecked(exc) && exc != syms.throwableType.tsym &&
                    exc != syms.exceptionType.tsym &&
                    !chk.intersects(exc, thrownInTry)) {
                log.error(((Catch) l.head).pos, "except.never.thrown.in.try",
                        exc.fullname.toJava());
            }
            alive = true;
            caught = chk.incl(exc, caught);
            inits = initsTry.dup();
            uninits = uninitsTry.dup();
            analyzeDef(param);
            letInit(param.pos, param.sym);
            analyzeStat(((Catch) l.head).body);
            initsEnd.andSet(inits);
            uninitsEnd.andSet(uninits);
            aliveEnd |= alive;
            nextadr = nextadrCatch;
        }
        if (tree.finalizer != null) {
            inits = initsTry.dup();
            uninits = uninitsTry.dup();
            alive = aliveTry;
            analyzeStat(tree.finalizer);
            inits.orSet(initsEnd);
            uninits.andSet(uninitsEnd);
            alive = alive & aliveEnd;
        } else {
            inits = initsEnd;
            uninits = uninitsEnd;
            alive = aliveEnd;
        }
        thrown = chk.union(thrown, chk.diff(thrownInTry, caught));
        uninitsTry.andSet(uninitsTryPrev).andSet(uninits);
    }

    public void _case(Conditional tree) {
        analyzeCond(tree.cond);
        if (tree.tag == Tree.CONDSTAT) {
            Bits initsBeforeElse = initsWhenFalse;
            Bits uninitsBeforeElse = uninitsWhenFalse;
            boolean aliveBeforeElse = alive;
            inits = initsWhenTrue;
            uninits = uninitsWhenTrue;
            analyzeStat(tree.thenpart);
            if (tree.elsepart != null) {
                Bits initsAfterThen = inits.dup();
                Bits uninitsAfterThen = uninits.dup();
                boolean aliveAfterThen = alive;
                inits = initsBeforeElse;
                uninits = uninitsBeforeElse;
                alive = aliveBeforeElse;
                analyzeStat(tree.elsepart);
                inits.andSet(initsAfterThen);
                uninits.andSet(uninitsAfterThen);
                alive = alive | aliveAfterThen;
            } else {
                inits.andSet(initsBeforeElse);
                uninits.andSet(uninitsBeforeElse);
                alive = true;
            }
        } else {
            Bits initsBeforeElse = initsWhenFalse;
            Bits uninitsBeforeElse = uninitsWhenFalse;
            inits = initsWhenTrue;
            uninits = uninitsWhenTrue;
            analyzeExpr(tree.thenpart);
            Bits initsAfterThen = inits.dup();
            Bits uninitsAfterThen = uninits.dup();
            inits = initsBeforeElse;
            uninits = uninitsBeforeElse;
            analyzeExpr(tree.elsepart);
            inits.andSet(initsAfterThen);
            uninits.andSet(uninitsAfterThen);
        }
    }

    public void _case(Exec tree) {
        analyzeExpr(tree.expr);
    }

    public void _case(Break tree) {
        jump(tree.target, breakTargetStates);
        markDead();
    }

    public void _case(Continue tree) {
        jump(tree.target, continueTargetStates);
        markDead();
    }

    public void _case(Return tree) {
        analyzeExpr(tree.expr);
        markDead();
    }

    public void _case(Throw tree) {
        analyzeExpr(tree.expr);
        markThrown(tree.pos, (ClassSymbol) tree.expr.type.tsym);
        markDead();
    }

    public void _case(Apply tree) {
        analyzeExpr(tree.meth);
        analyzeExprs(tree.args);
        for (List l = TreeInfo.symbol(tree.meth).type.thrown(); l.nonEmpty();
                l = l.tail)
            markThrown(tree.meth.pos, (ClassSymbol) l.head);
    }

    public void _case(NewClass tree) {
        analyzeExpr(tree.encl);
        analyzeExprs(tree.args);
        for (List l = tree.constructor.type.thrown(); l.nonEmpty(); l = l.tail)
            markThrown(tree.pos, (ClassSymbol) l.head);
        analyzeDef(tree.def);
    }

    public void _case(NewArray tree) {
        analyzeExprs(tree.dims);
        analyzeExprs(tree.elems);
    }

    public void _case(Assign tree) {
        analyzeVar(tree.lhs);
        analyzeExpr(tree.rhs);
        letInit(tree.lhs);
    }

    public void _case(Assignop tree) {
        analyzeExpr(tree.lhs);
        analyzeExpr(tree.rhs);
        letInit(tree.lhs);
    }

    public void _case(Operation tree) {
        switch (tree.tag) {
        case Tree.AND:
            analyzeCond((Tree) tree.args.head);
            Bits initsWhenFalseLeft = initsWhenFalse;
            Bits uninitsWhenFalseLeft = uninitsWhenFalse;
            inits = initsWhenTrue;
            uninits = uninitsWhenTrue;
            analyzeCond((Tree) tree.args.tail.head);
            initsWhenFalse.andSet(initsWhenFalseLeft);
            uninitsWhenFalse.andSet(uninitsWhenFalseLeft);
            break;

        case Tree.OR:
            analyzeCond((Tree) tree.args.head);
            Bits initsWhenTrueLeft = initsWhenTrue;
            Bits uninitsWhenTrueLeft = uninitsWhenTrue;
            inits = initsWhenFalse;
            uninits = uninitsWhenFalse;
            analyzeCond((Tree) tree.args.tail.head);
            initsWhenTrue.andSet(initsWhenTrueLeft);
            uninitsWhenTrue.andSet(uninitsWhenTrueLeft);
            break;

        case Tree.NOT:
            analyzeCond((Tree) tree.args.head);
            Bits t = initsWhenFalse;
            initsWhenFalse = initsWhenTrue;
            initsWhenTrue = t;
            t = uninitsWhenFalse;
            uninitsWhenFalse = uninitsWhenTrue;
            uninitsWhenTrue = t;
            break;

        case Tree.PREINC:

        case Tree.POSTINC:

        case Tree.PREDEC:

        case Tree.POSTDEC:
            analyzeExpr((Tree) tree.args.head);
            letInit((Tree) tree.args.head);
            break;

        default:
            for (List l = tree.args; l.nonEmpty(); l = l.tail)
                analyzeExpr((Tree) l.head);

        }
    }

    public void _case(TypeCast tree) {
        analyzeExpr(tree.expr);
    }

    public void _case(TypeTest tree) {
        analyzeExpr(tree.expr);
    }

    public void _case(Indexed tree) {
        analyzeExpr(tree.indexed);
        analyzeExpr(tree.index);
    }

    public void _case(Ident tree) {
        if (trackable(tree.sym) && mode != VARmode) {
            checkInit(tree.pos, (VarSymbol) tree.sym);
        }
    }

    public void _case(Select tree) {
        analyzeExpr(tree.selected);
    }

    public void _case(Tree tree1) {
    }
}
