package com.sun.tools.javac.v8.comp;
import java.io.IOException;

import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.tree.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.code.Type.*;

import com.sun.tools.javac.v8.code.Code.*;

import com.sun.tools.javac.v8.tree.Tree.*;

import com.sun.tools.javac.v8.comp.Items.*;


/**
 * This pass maps flat Java (i.e. without inner classes) to bytecodes.
 */
public class Gen extends Tree.Visitor implements Flags, Kinds, TypeTags, ByteCodes {
    Log log;
    Symtab syms;
    Check chk;
    Resolve rs;
    TreeMaker make;

    public Gen(Log log, Symtab syms, Check chk, Resolve rs, TreeMaker make,
            Hashtable options) {
        super();
        this.log = log;
        this.syms = syms;
        this.chk = chk;
        this.rs = rs;
        this.make = make;
        switchCheck = options.get("-switchcheck") != null;
        String detailedDebugOptions = (String) options.get("-g:");
        if (detailedDebugOptions != null) {
            lineDebugInfo = Util.contains(detailedDebugOptions, "lines", ',');
            varDebugInfo = Util.contains(detailedDebugOptions, "vars", ',');
        } else {
            lineDebugInfo = true;
            varDebugInfo = options.get("-g") != null;
        }
    }

    /**
      * Switches
      */
    boolean switchCheck;
    boolean lineDebugInfo;
    boolean varDebugInfo;
    Pool pool = new Pool();

    /**
     * Code buffer, set by genMethod.
     */
    Code code;

    /**
     * Items structure, set by genMethod.
     */
    Items items;

    /**
     * Environment for symbol lookup, set by genClass
     */
    Env attrEnv;

    /**
     * Class version numbers relevant to code generation, i.e., which
     * enable bug fixes or optimizations.
     */
    static short MAJOR_VERSION_1_2 = 46;

    /**
     * Generate code to load an integer constant.
     *  @param n     The integer to be loaded.
     */
    void loadIntConst(int n) {
        items.makeImmediateItem(Type.intType, new Integer(n)).load();
    }

    /**
      * The opcode that loads a zero constant of a given type code.
      *  @param tc   The given type code (@see ByteCode).
      */
    public static int zero(int tc) {
        switch (tc) {
        case INTcode:

        case BYTEcode:

        case SHORTcode:

        case CHARcode:
            return iconst_0;

        case LONGcode:
            return lconst_0;

        case FLOATcode:
            return fconst_0;

        case DOUBLEcode:
            return dconst_0;

        default:
            throw new InternalError("zero");

        }
    }

    /**
      * The opcode that loads a one constant of a given type code.
      *  @param tc   The given type code (@see ByteCode).
      */
    public static int one(int tc) {
        return zero(tc) + 1;
    }

    /**
      * Generate code to load -1 of the given type code (either int or long).
      *  @param tc   The given type code (@see ByteCode).
      */
    void emitMinusOne(int tc) {
        if (tc == LONGcode) {
            items.makeImmediateItem(Type.longType, new Long(-1)).load();
        } else {
            code.emitop(iconst_m1);
        }
    }

    /**
      * For target == 1.2: If qualified variable or method is accessed via
      *  a subclass (not a subinterface or an array!), clone it with the qualifier
      *  class as owner.
      *
      *  For target == 1.1: If qualified variable or method is defined in a
      *  non-accessible class, clone it with the qualifier class as owner.
      *
      *  @param sym    The accessed symbol
      *  @param site   The qualifier's type.
      */
    Symbol makeAccessible(Symbol sym, Type site) {
        if (syms.writer.majorVersion >= MAJOR_VERSION_1_2)
            if (sym.owner == site.tsym || site.tag != CLASS ||
                    sym.owner.kind == TYP && ((ClassSymbol) sym.owner).fullname ==
                    Names.java_lang_Object || sym.kind == VAR &&
                    (sym.owner.flags() & INTERFACE) != 0)
                return sym;
            else
                return sym.clone(site.tsym);
        else if (rs.isAccessible(attrEnv, (TypeSymbol) sym.owner))
            return sym;
        else
            return sym.clone(site.tsym);
    }

    /**
      * Insert a reference to given type in the constant pool;
      *  return the reference's index.
      *  @param type   The type for which a reference is inserted.
      */
    int makeRef(Type type) {
        return pool.put(type.tag == CLASS ? (Object) type.tsym : (Object) type);
    }

    /**
      * Create a tempory variable.
      *  @param type   The variable's type.
      */
    LocalItem makeTemp(Type type) {
        return (LocalItem) items.makeLocalItem(type, code.newLocal(type));
    }

    /**
      * Generate code to call a non-private method or constructor.
      *  @param pos         Position to be used for error reporting.
      *  @param site        The type of which the method is a member.
      *  @param name        The method's name.
      *  @param argtypes    The method's argument types.
      *  @param isStatic    A flag that indicates whether we call a
      *                     static or instance method.
      */
    void callMethod(int pos, Type site, Name name, List argtypes, boolean isStatic) {
        Symbol msym = rs.resolveQualifiedMethod(pos, attrEnv, site, name,
                Type.emptyList, argtypes);
        if (msym.kind == MTH) {
            if (isStatic)
                items.makeStaticItem(msym).invoke();
            else
                items.makeMemberItem(msym, name == Names.init).invoke();
        } else {
            throw new FatalError(
                    Log.getLocalizedString("cant.locate.meth", name.toJava()));
        }
    }

    /**
      * Generate code to invoke the finalizer associated with given environment.
      *  Any calls to finalizers are appended to the environments `cont' chain.
      */
    void callFinalizer(Env env) {
        if (code.isAlive()) {
            if (env.tree.tag == Tree.TRY) {
                Tree finalizer = ((Try) env.tree).finalizer;
                if (finalizer != null) {
                    ((Gen.GenContext) env.info).cont =
                            new Chain(code.emitJump(jsr), code.stacksize + 1,
                            ((Gen.GenContext) env.info).cont);
                }
            } else if (env.tree.tag == Tree.SYNCHRONIZED) {
                ((Gen.GenContext) env.info).finalize.gen();
            }
        }
    }

    /**
      * Generate code to call all finalizers of structures aborted by a non-local
      *  exit.  Return target environment of the non-local exit.
      *  @param target      The tree representing the structure that's aborted
      *  @param env         The environment current at the non-local exit.
      */
    Env unwind(Tree target, Env env) {
        Env env1 = env;
        while (env1 != null && env1.tree != target) {
            env1 = env1.next;
        }
        Tree lastTree = null;
        Env lastEnv = null;
        while (lastEnv != env1) {
            if (env.tree != lastTree) {
                callFinalizer(env);
                lastTree = env.tree;
            }
            lastEnv = env;
            env = env.next;
        }
        return env1;
    }

    /**
      * Do any of the structurs aborted by a non-local exit have finalizers?
      *  @param target      The tree representing the structure that's aborted
      *  @param env         The environment current at the non-local exit.
      */
    boolean hasFinalizers(Tree target, Env env) {
        while (env.tree != target) {
            if (env.tree.tag == Tree.TRY && ((Try) env.tree).finalizer != null ||
                    env.tree.tag == Tree.SYNCHRONIZED)
                return true;
            env = env.next;
        }
        return false;
    }

    /**
      * Distribute member initializer code into constructors and <clinit> method.
      *  @param defs         The list of class member declarations.
      *  @param c            The enclosing class.
      */
    List normalizeDefs(List defs, ClassSymbol c) {
        ListBuffer initCode = new ListBuffer();
        ListBuffer clinitCode = new ListBuffer();
        ListBuffer methodDefs = new ListBuffer();
        for (List l = defs; l.nonEmpty(); l = l.tail) {
            Tree def = (Tree) l.head;
            switch (def.tag) {
            case Tree.BLOCK:
                Block block = (Block) def;
                if ((block.flags & STATIC) != 0)
                    clinitCode.append(def);
                else
                    initCode.append(def);
                break;

            case Tree.METHODDEF:
                methodDefs.append(def);
                break;

            case Tree.VARDEF:
                VarDef vdef = (VarDef) def;
                if (vdef.init != null) {
                    if ((vdef.sym.flags() & STATIC) == 0) {
                        Tree init = make.at(vdef.pos).Assignment(vdef.sym, vdef.init);
                        initCode.append(init);
                    } else if (vdef.sym.constValue == null) {
                        Tree init = make.at(vdef.pos).Assignment(vdef.sym, vdef.init);
                        clinitCode.append(init);
                    }
                }
                break;

            default:
                throw new InternalError();

            }
        }
        if (initCode.length() != 0) {
            List inits = initCode.toList();
            for (Enumeration e = methodDefs.elements(); e.hasMoreElements();) {
                normalizeMethod((MethodDef) e.nextElement(), inits);
            }
        }
        if (clinitCode.length() != 0) {
            MethodSymbol clinit = new MethodSymbol(STATIC, Names.clinit,
                    new MethodType(Type.emptyList, Type.voidType,
                    ClassSymbol.emptyList), c);
            c.members().enter(clinit);
            List clinitStats = clinitCode.toList();
            make.at(((Tree) clinitStats.head).pos);
            methodDefs.append(make.MethodDef(clinit, make.Block(0, clinitStats)));
        }
        return methodDefs.toList();
    }

    /**
      * Insert instance initializer code into initial constructor.
      *  @param md        The tree potentially representing a
      *                   constructor's definition.
      *  @param initCode  The list of instance initializer statements.
      */
    void normalizeMethod(MethodDef md, List initCode) {
        if (md.name == Names.init && TreeInfo.isInitialConstructor(md)) {
            List stats = md.body.stats;
            ListBuffer newstats = new ListBuffer();
            if (stats.nonEmpty()) {
                newstats.append(stats.head);
                stats = stats.tail;
                while (stats.nonEmpty() &&
                        TreeInfo.isSyntheticInit((Tree) stats.head)) {
                    newstats.append(stats.head);
                    stats = stats.tail;
                }
                newstats.appendList(initCode);
                while (stats.nonEmpty()) {
                    newstats.append(stats.head);
                    stats = stats.tail;
                }
            }
            md.body.stats = newstats.toList();
        }
    }

    /**
      * Visitor argument: The current environment.
      */
    Env env;

    /**
     * Visitor argument: The expected type (prototype).
     */
    Type pt;

    /**
     * Visitor result: The item representing the computed value.
     */
    Item result;

    /**
     * Visitor method: generate code for a definition, catching and reporting
     *  any completion failures.
     *  @param tree    The definition to be visited.
     *  @param env     The environment current at the definition.
     */
    public void genDef(Tree tree, Env env) {
        Env prevEnv = this.env;
        try {
            {
                this.env = env;
                tree.visit(this);
                this.env = prevEnv;
            }
        } catch (CompletionFailure ex) {
            {
                this.env = prevEnv;
                chk.completionError(tree.pos, ex);
            }
        }
    }

    /**
      * Derived visitor method: generate code for a statement.
      */
    public void genStat(Tree tree, Env env) {
        if (code.isAlive()) {
            code.statBegin(tree.pos);
            genDef(tree, env);
        }
    }

    /**
      * Derived visitor method: generate code for a list of statements.
      */
    public void genStats(List trees, Env env) {
        for (List l = trees; l.nonEmpty(); l = l.tail)
            genStat((Tree) l.head, env);
    }

    /**
      * Visitor method: generate code for a definition, catching and reporting
      *  any completion failures.
      *  @param tree    The expression to be visited.
      *  @param pt      The expression's expected type (proto-type).
      */
    public Item genExpr(Tree tree, Type pt) {
        Type prevPt = this.pt;
        try {
            {
                if (tree.type.constValue != null) {
                    result = items.makeImmediateItem(tree.type, tree.type.constValue);
                } else {
                    this.pt = pt;
                    tree.visit(this);
                    this.pt = prevPt;
                }
                return result.coerce(pt);
            }
        } catch (CompletionFailure ex) {
            {
                this.pt = prevPt;
                chk.completionError(tree.pos, ex);
                return items.makeStackItem(pt);
            }
        }
    }

    /**
      * Derived visitor method: generate code for a list of method arguments.
      *  @param trees    The argument expressions to be visited.
      *  @param pts      The expression's expected types (i.e. the formal parameter
      *                  types of the invoked method).
      */
    public void genArgs(List trees, List pts) {
        for (List l = trees; l.nonEmpty(); l = l.tail) {
            genExpr((Tree) l.head, (Type) pts.head).load();
            pts = pts.tail;
        }
    }

    /**
      * Visitor method for method definitions.
      */
    public void _case(MethodDef tree) {
        Env localEnv = env.dup(tree);
        localEnv.enclMethod = tree;
        this.pt = tree.sym.erasure().restype();
        genMethod(tree, localEnv, false);
    }

    /**
      * Generate code for a method.
      *  @param tree     The tree representing the method definition.
      *  @param env      The environment current for the method body.
      *  @param fatcode  A flag that indicates whether all jumps are within 32K.
      *                  We first invoke this method under the assumption
      *                  that fatcode == false, i.e. all jumps are within 32K.
      *                  If this fails, fatcode is set to true and we try again.
      */
    void genMethod(MethodDef tree, Env env, boolean fatcode) {
        MethodSymbol meth = tree.sym;
        if (tree.body != null) {
            meth.code = code = new Code(fatcode, lineDebugInfo, varDebugInfo);
            items = new Items(pool, code);
            if (Code.debugCode)
                System.err.println(meth);
            if ((tree.flags & STATIC) == 0)
                code.setDefined( code.newLocal(
                        new VarSymbol(FINAL, Names._this, meth.owner.type,
                        meth.owner)));
            for (List l = tree.params; l.nonEmpty(); l = l.tail)
                code.setDefined(code.newLocal(((VarDef) l.head).sym));
            genStat(tree.body, env);
            if (Util.checks)
                Util.assert(code.stacksize == 0);
            if (code.isAlive()) {
                code.statBegin(TreeInfo.endPos(tree.body));
                code.emitop(return_);
            }
            code.endScopes(0);
            if (!fatcode && code.fatcode)
                genMethod(tree, env, true);
        }
    }

    /**
      * Visitor method for variable definitions.
      */
    public void _case(VarDef tree) {
        VarSymbol v = tree.sym;
        code.newLocal(v);
        if (tree.init != null && (v.constValue == null || varDebugInfo)) {
            genExpr(tree.init, v.erasure()).load();
            items.makeLocalItem(v).store();
        }
    }

    /**
      * Visitor method for blocks.
      */
    public void _case(Block tree) {
        int limit = code.nextadr;
        genStats(tree.stats, env);
        code.endScopes(limit);
    }

    /**
      * Visitor method for do-loops.
      */
    public void _case(DoLoop tree) {
        genLoop(tree, tree.body, tree.cond, Tree.emptyList, null);
    }

    /**
      * Visitor method for while-loops.
      */
    public void _case(WhileLoop tree) {
        genLoop(tree, tree.body, tree.cond, Tree.emptyList, code.branch(goto_));
    }

    /**
      * Visitor method for for-loops.
      */
    public void _case(ForLoop tree) {
        genStats(tree.init, env);
        genLoop(tree, tree.body, tree.cond, tree.step, code.branch(goto_));
    }

    /**
      * Generate code for a loop.
      *  @param loop       The tree representing the loop.
      *  @param body       The loop's body.
      *  @param cond       The loop's controling condition.
      *  @param step       "Step" statements to be inserted at end of
      *                    each iteration.
      *  @param enterChain A chain for entries that jump into loop.
      */
    private void genLoop(Tree loop, Tree body, Tree cond, List step,
            Chain enterChain) {
        Env loopEnv = env.dup(loop, new GenContext());
        int startpc = code.entryPoint();
        genStat(body, loopEnv);
        code.resolve(((Gen.GenContext) loopEnv.info).cont);
        genStats(step, loopEnv);
        code.resolve(enterChain);
        CondItem c;
        if (cond != null) {
            code.statBegin(cond.pos);
            c = genExpr(cond, Type.booleanType).mkCond();
        } else {
            c = items.makeCondItem(goto_);
        }
        code.resolve(c.jumpTrue(), startpc);
        code.resolve(c.falseJumps);
        code.resolve(((Gen.GenContext) loopEnv.info).exit);
    }

    /**
      * Visitor method for labelled statements.
      */
    public void _case(Labelled tree) {
        Env localEnv = env.dup(TreeInfo.referencedStatement(tree), new GenContext());
        genStat(tree.body, localEnv);
        code.resolve(((Gen.GenContext) localEnv.info).exit);
    }

    /**
      * Visitor method for switch statements.
      */
    public void _case(Switch tree) {
        Item sel = genExpr(tree.selector, Type.intType);
        List cases = tree.cases;
        if (cases.isEmpty()) {
            sel.drop();
        } else {
            sel.load();
            Env switchEnv = env.dup(tree, new GenContext());
            int lo = Integer.MAX_VALUE;
            int hi = Integer.MIN_VALUE;
            int nlabels = 0;
            int[] labels = new int[cases.length()];
            int defaultIndex = -1;
            List l = cases;
            for (int i = 0; i < labels.length; i++) {
                if (((Case) l.head).pat != null) {
                    int val =
                            ((Number)((Case) l.head).pat.type.constValue).intValue();
                    labels[i] = val;
                    if (val < lo)
                        lo = val;
                    if (hi < val)
                        hi = val;
                    nlabels++;
                } else {
                    if (Util.checks)
                        Util.assert(defaultIndex == -1);
                    defaultIndex = i;
                }
                l = l.tail;
            }
            long table_space_cost = 4 + ((long) hi - lo + 1);
            long table_time_cost = 3;
            long lookup_space_cost = 3 + 2 * (long) nlabels;
            long lookup_time_cost = nlabels;
            int opcode = nlabels > 0 && table_space_cost + 3 * table_time_cost <=
                    lookup_space_cost + 3 * lookup_time_cost ? tableswitch :
                    lookupswitch;
            int startpc = code.curPc();
            code.emitop(opcode);
            code.align(4);
            int tableBase = code.curPc();
            int[] offsets = null;
            code.emit4(-1);
            if (opcode == tableswitch) {
                code.emit4(lo);
                code.emit4(hi);
                for (int i = lo; i <= hi; i++) {
                    code.emit4(-1);
                }
            } else {
                code.emit4(nlabels);
                for (int i = 0; i < nlabels; i++) {
                    code.emit4(-1);
                    code.emit4(-1);
                }
                offsets = new int[labels.length];
            }
            code.markDead();
            l = cases;
            for (int i = 0; i < labels.length; i++) {
                Case c = (Case) l.head;
                l = l.tail;
                int pc = code.entryPoint();
                if (i != defaultIndex) {
                    if (opcode == tableswitch) {
                        code.put4(tableBase + 4 * (labels[i] - lo + 3), pc - startpc);
                    } else {
                        offsets[i] = pc - startpc;
                    }
                } else {
                    code.put4(tableBase, pc - startpc);
                }
                genStats(c.stats, switchEnv);
                if (switchCheck && code.isAlive() && c.stats.nonEmpty() &&
                        i < labels.length - 1)
                    log.warning(c.pos, "possible.fall-through.into.case");
            }
            code.resolve(((Gen.GenContext) switchEnv.info).exit);
            if (code.get4(tableBase) == -1) {
                code.put4(tableBase, code.entryPoint() - startpc);
            }
            if (opcode == tableswitch) {
                int defaultOffset = code.get4(tableBase);
                for (int i = lo; i <= hi; i++) {
                    if (code.get4(tableBase + 4 * (i - lo + 3)) == -1)
                        code.put4(tableBase + 4 * (i - lo + 3), defaultOffset);
                }
            } else {
                if (defaultIndex >= 0)
                    for (int i = defaultIndex; i < labels.length - 1; i++) {
                        labels[i] = labels[i + 1];
                        offsets[i] = offsets[i + 1];
                    }
                if (nlabels > 0)
                    qsort2(labels, offsets, 0, nlabels - 1);
                for (int i = 0; i < nlabels; i++) {
                    int caseidx = tableBase + 8 * (i + 1);
                    code.put4(caseidx, labels[i]);
                    code.put4(caseidx + 4, offsets[i]);
                }
            }
        }
    }

    /**
      * Sort (int) arrays of keys and values
      */
    static void qsort2(int[] keys, int[] values, int lo, int hi) {
        int i = lo;
        int j = hi;
        int pivot = keys[(i + j) / 2];
        do {
            while (keys[i] < pivot)
                i++;
            while (pivot < keys[j])
                j--;
            if (i <= j) {
                int temp1 = keys[i];
                keys[i] = keys[j];
                keys[j] = temp1;
                int temp2 = values[i];
                values[i] = values[j];
                values[j] = temp2;
                i++;
                j--;
            }
        } while (i <= j)
            ;
        if (lo < j)
            qsort2(keys, values, lo, j);
        if (i < hi)
            qsort2(keys, values, i, hi);
    }

    /**
      * Visitor method for synchronized statements:
      *  synchronized (lock) body.
      */
    public void _case(Synchronized tree) {
        final Item lockVar = makeTemp(syms.objectType);
        genExpr(tree.lock, tree.lock.type).load();
        lockVar.store();
        lockVar.load();
        code.emitop(monitorenter);
        Env syncEnv = env.dup(tree, new GenContext());
        ((Gen.GenContext) syncEnv.info).finalize = new GenFinalizer() {


                    void gen() {
                        lockVar.load();
                        code.emitop(monitorexit);
                    }
                };
        genTry(tree.body, Catch.emptyList, syncEnv);
        code.resolve(((Gen.GenContext) syncEnv.info).exit);
    }

    /**
      * Visitor method for try statements.
      */
    public void _case(final Try tree) {
        Env tryEnv = env.dup(tree, new GenContext());
        if (tree.finalizer != null) {
            ((Gen.GenContext) tryEnv.info).finalize = new GenFinalizer() {


                        void gen() {
                            genStat(tree.finalizer, env);
                        }
                    };
        }
        genTry(tree.body, tree.catchers, tryEnv);
        code.resolve(((Gen.GenContext) tryEnv.info).exit);
    }

    /**
      * Generate code for a try or synchronized statement
      *  @param body      The body of the try or synchronized statement.
      *  @param catchers  The lis of catch clauses.
      *  @param env       the environment current for the body.
      */
    void genTry(Tree body, List catchers, Env env) {
        int limit = code.nextadr;
        int startpc = code.curPc();
        genStat(body, env);
        int endpc = code.curPc();
        callFinalizer(env);
        Chain exitChain = code.branch(goto_);
        if (startpc != endpc) {
            for (List l = catchers; l.nonEmpty(); l = l.tail) {
                code.entryPoint(1);
                genCatch((Catch) l.head, env, startpc, endpc);
                callFinalizer(env);
                exitChain = code.mergeChains(exitChain, code.branch(goto_));
            }
        }
        if (((Gen.GenContext) env.info).finalize != null) {
            code.newRegSegment();
            int pc = code.entryPoint(1);
            registerCatch(body.pos, startpc, pc, pc, 0);
            code.statBegin(TreeInfo.finalizerPos(env.tree));
            code.markStatBegin();
            Item excVar = makeTemp(syms.throwableType);
            excVar.store();
            callFinalizer(env);
            excVar.load();
            code.emitop(athrow);
            if (((Gen.GenContext) env.info).cont != null) {
                code.entryPoint(1);
                code.resolve(((Gen.GenContext) env.info).cont);
                code.statBegin(TreeInfo.finalizerPos(env.tree));
                code.markStatBegin();
                LocalItem retVar = makeTemp(syms.throwableType);
                retVar.store();
                ((Gen.GenContext) env.info).finalize.gen();
                code.emitop1w(ret, code.regOf(retVar.adr));
                code.markDead();
            }
        }
        code.resolve(exitChain);
        code.endScopes(limit);
    }

    /**
      * Generate code for a catch clause.
      *  @param tree     The catch clause.
      *  @param env      The environment current in the enclosing try.
      *  @param startpc  Start pc of try-block.
      *  @param endpc    End pc of try-block.
      */
    void genCatch(Catch tree, Env env, int startpc, int endpc) {
        if (startpc != endpc) {
            registerCatch(tree.pos, startpc, endpc, code.curPc(),
                    makeRef(tree.param.type));
            code.setDefined(code.newLocal(tree.param.sym));
            code.statBegin(TreeInfo.firstStatPos(tree.body));
            code.markStatBegin();
            items.makeLocalItem(tree.param.sym).store();
            genStat(tree.body, env);
        }
    }

    /**
      * Register a catch clause in the "Exceptions" code-atttribute.
      */
    void registerCatch(int pos, int start_pc, int end_pc, int handler_pc,
            int catch_type) {
        if (start_pc != end_pc) {
            char start_pc1 = (char) start_pc;
            char end_pc1 = (char) end_pc;
            char handler_pc1 = (char) handler_pc;
            if (start_pc1 == start_pc && end_pc1 == end_pc &&
                    handler_pc1 == handler_pc) {
                code.addCatch(start_pc1, end_pc1, handler_pc1, (char) catch_type);
            } else {
                log.error(pos, "code.too.large.for.try.stmt");
            }
        }
    }

    /**
      * Visitor method for conditional expressions or statements.
      */
    public void _case(Conditional tree) {
        Chain thenExit = null;
        CondItem c = genExpr(tree.cond, Type.booleanType).mkCond();
        Chain elseChain = c.jumpFalse();
        if (c.trueJumps != null || c.opcode != dontgoto) {
            code.resolve(c.trueJumps);
            if (tree.tag == Tree.CONDSTAT)
                genStat(tree.thenpart, env);
            else
                genExpr(tree.thenpart, pt).load();
            thenExit = code.branch(goto_);
        }
        if (tree.elsepart != null && elseChain != null) {
            code.resolve(elseChain);
            if (tree.tag == Tree.CONDSTAT)
                genStat(tree.elsepart, env);
            else
                genExpr(tree.elsepart, pt).load();
            code.resolve(thenExit);
        } else {
            code.resolve(thenExit);
            code.resolve(elseChain);
        }
        if (tree.tag == Tree.CONDEXPR)
            result = items.makeStackItem(pt);
    }

    /**
      * Visitor method for statement expressions.
      */
    public void _case(Exec tree) {
        if (tree.expr.tag == Tree.POSTINC)
            tree.expr.tag = Tree.PREINC;
        else if (tree.expr.tag == Tree.POSTDEC)
            tree.expr.tag = Tree.PREDEC;
        genExpr(tree.expr, tree.expr.type).drop();
    }

    /**
      * Visitor method for break statements.
      */
    public void _case(Break tree) {
        ((Gen.GenContext) unwind(tree.target, env).info).addExit(code.branch(goto_));
    }

    /**
      * Visitor method for continue statements.
      */
    public void _case(Continue tree) {
        ((Gen.GenContext) unwind(tree.target, env).info).addCont(code.branch(goto_));
    }

    /**
      * Visitor method for return statements.
      */
    public void _case(Return tree) {
        if (tree.expr != null) {
            Item r = genExpr(tree.expr, pt);
            if (hasFinalizers(env.enclMethod, env)) {
                r.load();
                r = makeTemp(pt);
                r.store();
            }
            unwind(env.enclMethod, env);
            r.load();
            code.emitop(ireturn + Code.truncate(Code.typecode(pt)));
        } else {
            unwind(env.enclMethod, env);
            code.emitop(return_);
        }
    }

    /**
      * Visitor method for throw statements.
      */
    public void _case(Throw tree) {
        genExpr(tree.expr, tree.expr.type).load();
        code.emitop(athrow);
    }

    /**
      * Visitor method for method invocations.
      */
    public void _case(Apply tree) {
        Item m = genExpr(tree.meth, methodType);
        if (tree.args.length() !=
                TreeInfo.symbol(tree.meth).externalType().argtypes().length()) {
            throw new InternalError("argument length mismatch: " +
                    TreeInfo.symbol(tree.meth));
        }
        genArgs(tree.args, TreeInfo.symbol(tree.meth).externalType().argtypes());
        result = m.invoke();
    }

    /**
      * A type that serves as the expected type for all method expressions.
      */
    static final Type methodType = new MethodType(null, null, null);

    /**
     * Visitor method for class instance creation expressions.
     */
    public void _case(NewClass tree) {
        if (Util.checks)
            Util.assert(tree.encl == null && tree.def == null);
        code.emitop2(new_, makeRef(tree.type));
        code.emitop(dup);
        genArgs(tree.args, tree.constructor.externalType().argtypes());
        items.makeMemberItem(tree.constructor, true).invoke();
        result = items.makeStackItem(tree.type);
    }

    /**
      * Visitor method for array creation expressions.
      */
    public void _case(NewArray tree) {
        if (tree.elems != null) {
            Type elemtype = tree.type.elemtype();
            loadIntConst(tree.elems.length());
            Item arr = makeNewArray(tree.type, 1);
            int i = 0;
            for (List l = tree.elems; l.nonEmpty(); l = l.tail) {
                arr.duplicate();
                loadIntConst(i);
                i++;
                genExpr((Tree) l.head, elemtype).load();
                items.makeIndexedItem(elemtype).store();
            }
            result = arr;
        } else {
            for (List l = tree.dims; l.nonEmpty(); l = l.tail) {
                genExpr((Tree) l.head, Type.intType).load();
            }
            result = makeNewArray(tree.type, tree.dims.length());
        }
    }

    /**
      * Generate code to create an array with given element type and number
      *  of dimensions.
      */
    Item makeNewArray(Type type, int ndims) {
        Type elemtype = type.elemtype();
        int elemcode = Code.arraycode(elemtype);
        if (elemcode == 0 || elemcode == 1 && ndims == 1) {
            code.emitop2(anewarray, makeRef(elemtype));
        } else if (elemcode == 1) {
            code.emitop(multianewarray, 1 - ndims);
            code.emit2(makeRef(type));
            code.emit1(ndims);
        } else {
            code.emitop1(newarray, elemcode);
        }
        return items.makeStackItem(type);
    }

    /**
      * Visitor method for assignment expressions.
      */
    public void _case(Assign tree) {
        Item l = genExpr(tree.lhs, tree.lhs.type);
        genExpr(tree.rhs, tree.lhs.type).load();
        result = items.makeAssignItem(l);
    }

    /**
      * Visitor method for assignment operation expressions.
      */
    public void _case(Assignop tree) {
        OperatorSymbol operator = (OperatorSymbol) tree.operator;
        Item l;
        if (operator.opcode == string_add) {
            makeStringBuffer(tree.pos);
            l = genExpr(tree.lhs, tree.lhs.type);
            if (l.width() > 0) {
                code.emitop(dup_x1 + 3 * (l.width() - 1));
            }
            l.load();
            appendString(tree.lhs);
            appendStrings(tree.rhs);
            bufferToString(tree.pos);
        } else {
            l = genExpr(tree.lhs, tree.lhs.type);
            if ((tree.tag == Tree.PLUS_ASG || tree.tag == Tree.MINUS_ASG) &&
                    l instanceof LocalItem && tree.lhs.type.tag <= INT &&
                    tree.rhs.type.tag <= INT && tree.rhs.type.constValue != null) {
                int ival = ((Number) tree.rhs.type.constValue).intValue();
                if (tree.tag == Tree.MINUS_ASG)
                    ival = -ival;
                if (-128 <= ival && ival <= 127) {
                    ((LocalItem) l).incr(ival);
                    result = l;
                    return;
                }
            }
            l.duplicate();
            l.coerce((Type) operator.type.argtypes().head).load();
            completeBinop(tree.lhs, tree.rhs, operator).coerce(tree.lhs.type);
        }
        result = items.makeAssignItem(l);
    }

    /**
      * Visitor method for (unary and binary) operation expressions.
      */
    public void _case(Operation tree) {
        OperatorSymbol operator = (OperatorSymbol) tree.operator;
        if (operator.opcode == string_add) {
            makeStringBuffer(tree.pos);
            appendStrings(tree);
            bufferToString(tree.pos);
            result = items.makeStackItem(syms.stringType);
        } else {
            Item od = genExpr((Tree) tree.args.head,
                    (Type) operator.type.argtypes().head);
            switch (tree.tag) {
            case Tree.POS:
                result = od.load();
                break;

            case Tree.NEG:
                result = od.load();
                code.emitop(operator.opcode);
                break;

            case Tree.NOT:
                result = od.mkCond().negate();
                break;

            case Tree.COMPL:
                result = od.load();
                emitMinusOne(od.typecode);
                code.emitop(operator.opcode);
                break;

            case Tree.PREINC:

            case Tree.PREDEC:
                od.duplicate();
                if (od instanceof LocalItem &&
                        (operator.opcode == iadd || operator.opcode == isub)) {
                    ((LocalItem) od).incr(tree.tag == Tree.PREINC ? 1 : -1);
                    result = od;
                } else {
                    od.load();
                    code.emitop(one(od.typecode));
                    code.emitop(operator.opcode);
                    if (od.typecode != INTcode &&
                            Code.truncate(od.typecode) == INTcode)
                        code.emitop(int2byte + od.typecode - BYTEcode);
                    result = items.makeAssignItem(od);
                }
                break;

            case Tree.POSTINC:

            case Tree.POSTDEC:
                od.duplicate();
                if (od instanceof LocalItem && operator.opcode == iadd) {
                    Item res = od.load();
                    ((LocalItem) od).incr(tree.tag == Tree.POSTINC ? 1 : -1);
                    result = res;
                } else {
                    Item res = od.load();
                    od.stash(od.typecode);
                    code.emitop(one(od.typecode));
                    code.emitop(operator.opcode);
                    if (od.typecode != INTcode &&
                            Code.truncate(od.typecode) == INTcode)
                        code.emitop(int2byte + od.typecode - BYTEcode);
                    od.store();
                    result = res;
                }
                break;

            case Tree.OR:
                {
                    CondItem lcond = od.mkCond();
                    if (lcond.falseJumps != null || lcond.opcode != goto_) {
                        Chain trueJumps = lcond.jumpTrue();
                        code.resolve(lcond.falseJumps);
                        CondItem rcond = genExpr((Tree) tree.args.tail.head,
                                ((Tree) tree.args.tail.head).type).mkCond();
                        result = items.makeCondItem(rcond.opcode,
                                code.mergeChains(trueJumps, rcond.trueJumps),
                                rcond.falseJumps);
                    } else {
                        result = lcond;
                    }
                    break;
                }

            case Tree.AND:
                {
                    CondItem lcond = od.mkCond();
                    if (lcond.trueJumps != null || lcond.opcode != dontgoto) {
                        Chain falseJumps = lcond.jumpFalse();
                        code.resolve(lcond.trueJumps);
                        CondItem rcond = genExpr((Tree) tree.args.tail.head,
                                ((Tree) tree.args.tail.head).type).mkCond();
                        result = items.makeCondItem(rcond.opcode, rcond.trueJumps,
                                code.mergeChains(falseJumps, rcond.falseJumps));
                    } else {
                        result = lcond;
                    }
                    break;
                }

            default:
                od.load();
                result = completeBinop((Tree) tree.args.head,
                        (Tree) tree.args.tail.head, operator);

            }
        }
    }

    /**
      * Make a new string buffer.
      */
    void makeStringBuffer(int pos) {
        code.emitop2(new_, makeRef(syms.stringBufferType));
        code.emitop(dup);
        callMethod(pos, syms.stringBufferType, Names.init, Type.emptyList, false);
    }

    /**
      * Append value (on tos) to string buffer (on tos - 1).
      */
    void appendString(Tree tree) {
        Type t = tree.type;
        if (t.tag > lastBaseTag && t.tsym != syms.stringType.tsym) {
            t = syms.objectType;
        }
        callMethod(tree.pos, syms.stringBufferType, Names.append,
                Type.emptyList.prepend(t), false);
    }

    /**
      * Add all strings in tree to string buffer.
      */
    void appendStrings(Tree tree) {
        if (tree.tag == Tree.PLUS && tree.type.constValue == null) {
            Operation op = (Operation) tree;
            if (op.operator.kind == MTH &&
                    ((OperatorSymbol) op.operator).opcode == string_add) {
                appendStrings((Tree) op.args.head);
                appendStrings((Tree) op.args.tail.head);
                return;
            }
        }
        genExpr(tree, tree.type).load();
        appendString(tree);
    }

    /**
      * Convert string buffer on tos to string.
      */
    void bufferToString(int pos) {
        callMethod(pos, syms.stringBufferType, Names.toString, Type.emptyList, false);
    }

    /**
      * Complete generating code for operation, with left operand
      *  already on stack.
      *  @param lhs       The tree representing the left operand.
      *  @param rhs       The tree representing the right operand.
      *  @param operator  The operator symbol.
      */
    Item completeBinop(Tree lhs, Tree rhs, OperatorSymbol operator) {
        MethodType optype = (MethodType) operator.type;
        int opcode = operator.opcode;
        if (opcode >= if_icmpeq && opcode <= if_icmple &&
                rhs.type.constValue instanceof Number &&
                ((Number) rhs.type.constValue).intValue() == 0) {
            opcode = opcode + (ifeq - if_icmpeq)
                     ;
        } else if (opcode >= if_acmpeq && opcode <= if_acmpne &&
                TreeInfo.symbol(rhs) == syms.nullConst) {
            opcode = opcode + (if_acmp_null - if_acmpeq)
                     ;
        } else {
            Type rtype = (Type) operator.erasure().argtypes().tail.head;
            if (opcode >= ishll && opcode <= lushrl) {
                opcode = opcode + (ishl - ishll);
                rtype = Type.intType;
            }
            genExpr(rhs, rtype).load();
            if (opcode >= 1 << preShift) {
                code.emitop(opcode >> preShift);
                opcode = opcode & 255;
            }
        }
        if (opcode >= ifeq && opcode <= if_acmpne || opcode == if_acmp_null ||
                opcode == if_acmp_nonnull) {
            return items.makeCondItem(opcode);
        } else {
            code.emitop(opcode);
            return items.makeStackItem(optype.restype);
        }
    }

    /**
      * Visitor method for type cast expressions.
      */
    public void _case(TypeCast tree) {
        result = genExpr(tree.expr, tree.clazz.type).load();
        if (tree.clazz.type.tag > lastBaseTag &&
                tree.expr.type.asSuper(tree.clazz.type.tsym) == null) {
            code.emitop2(checkcast, makeRef(tree.clazz.type));
        }
    }

    /**
      * Visitor method for type test expressions.
      */
    public void _case(TypeTest tree) {
        genExpr(tree.expr, tree.expr.type).load();
        code.emitop2(instanceof_, makeRef(tree.clazz.type));
        result = items.makeStackItem(Type.booleanType);
    }

    /**
      * Visitor method for indexed expressions.
      */
    public void _case(Indexed tree) {
        genExpr(tree.indexed, tree.indexed.type).load();
        genExpr(tree.index, Type.intType).load();
        result = items.makeIndexedItem(tree.type);
    }

    /**
      * Visitor method for simple identifiers.
      */
    public void _case(Ident tree) {
        Symbol sym = tree.sym;
        if (tree.name == Names._this || tree.name == Names._super) {
            Item res = tree.name == Names._this ? items.makeThisItem() :
                    items.makeSuperItem();
            if (sym.kind == MTH) {
                res.load();
                res = items.makeMemberItem(sym, true);
            }
            result = res;
        } else if (tree.name == Names._null) {
            code.emitop(aconst_null);
            result = items.makeStackItem(tree.type);
        } else if (sym.kind == VAR && sym.owner.kind == MTH) {
            result = items.makeLocalItem((VarSymbol) sym);
        } else if ((sym.flags() & STATIC) != 0) {
            result = items.makeStaticItem(sym);
        } else {
            items.makeThisItem().load();
            sym = makeAccessible(sym, env.enclClass.type);
            result = items.makeMemberItem(sym, (sym.flags() & PRIVATE) != 0);
        }
    }

    /**
      * Visitor method for selections.
      */
    public void _case(Select tree) {
        Symbol sym = tree.sym;
        if (tree.name == Names._class) {
            throw new InternalError();
        } else {
            Symbol ssym = TreeInfo.symbol(tree.selected);
            boolean selectSuper =
                    ssym != null && (ssym.kind == TYP || ssym.name == Names._super);
            boolean accessSuper = isOddAccessName(env.enclMethod.name);
            Item base = selectSuper ? items.makeSuperItem() :
                    genExpr(tree.selected, tree.selected.type);
            if (sym.kind == VAR && ((VarSymbol) sym).constValue != null) {
                base.drop();
                result = items.makeImmediateItem(sym.type,
                        ((VarSymbol) sym).constValue);
            } else {
                sym = makeAccessible(sym, tree.selected.type);
                if ((sym.flags() & STATIC) != 0) {
                    base.drop();
                    result = items.makeStaticItem(sym);
                } else {
                    base.load();
                    if (sym == syms.lengthVar) {
                        code.emitop(arraylength);
                        result = items.makeStackItem(Type.intType);
                    } else {
                        result = items.makeMemberItem(sym,
                                (sym.flags() & PRIVATE) != 0 || selectSuper ||
                                accessSuper);
                    }
                }
            }
        }
    }

    /**
      * Does given name start with "access$" and end in an odd digit?
      */
    boolean isOddAccessName(Name name) {
        return name.startsWith(Names.accessDollar) &&
                (name.byteAt(name.len - 1) & 1) == 1;
    }

    /**
      * Visitor method for literals.
      */
    public void _case(Literal tree) {
        result = items.makeImmediateItem(tree.type, tree.value);
    }

    /**
      * Generate code for a class definition.
      *  @param env   The attribution environment that belongs to the
      *               outermost class containing this class definition.
      *               We need this for resolving some additional symbols.
      *  @param cdef  The tree representing the class definition.
      */
    public void genClass(Env env, ClassDef cdef) {
        attrEnv = env;
        ClassSymbol c = cdef.sym;
        cdef.defs = normalizeDefs(cdef.defs, c);
        c.pool = pool;
        pool.reset();
        Env localEnv = new Env(cdef, new GenContext());
        localEnv.toplevel = env.toplevel;
        localEnv.enclClass = cdef;
        for (List l = cdef.defs; l.nonEmpty(); l = l.tail) {
            genDef((Tree) l.head, localEnv);
        }
    }

    /**
      * An abstract class for finalizer generation.
      */
    abstract class GenFinalizer {

        GenFinalizer() {
            super();
        }

        abstract void gen();
    }

    /**
      * GJC code generation contexts,
      *  to be used as type parameter for environments.
      */
    class GenContext {

        GenContext() {
            super();
        }

        /**
          * A chain for all unresolved jumps that exit the current environment.
          */
        Chain exit = null;

        /**
         * A chain for all unresolved jumps that continue in the
         *  current environment.
         */
        Chain cont = null;

        /**
         * A closure that generates the finalizer of the current environment.
         *  Only set for Synchronized and Try contexts.
         */
        GenFinalizer finalize = null;

        /**
         * Add given chain to exit chain.
         */
        void addExit(Chain c) {
            exit = Code.mergeChains(c, exit);
        }

        /**
          * Add given chain to cont chain.
          */
        void addCont(Chain c) {
            cont = Code.mergeChains(c, cont);
        }
    }
}
