package com.sun.tools.javac.v8.comp;
import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.tree.*;

import com.sun.tools.javac.v8.code.Type.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.tree.Tree.*;


/**
 * Helper class for name resolution, used mostly by the attribution phase.
 */
public class Resolve implements Flags, Kinds, TypeTags {

    /**
     * Error kinds that complement the constants in Kinds
     */
    static final int AMBIGUOUS = 256;
    static final int HIDDEN = 257;
    static final int ABSENT_VAR = 258;
    static final int WRONG_MTHS = 259;
    static final int WRONG_MTH = 260;
    static final int ABSENT_MTH = 261;
    static final int ABSENT_TYP = 262;
    Log log;
    Symtab syms;
    Check chk;
    Infer infer;

    public Resolve(Log log, Symtab syms, Check chk, Infer infer) {
        super();
        this.syms = syms;
        this.log = log;
        this.chk = chk;
        this.infer = infer;
    }

    /**
      * An environment is "static" if its static level is greater than the one of
      *  its outer environment
      */
    static boolean isStatic(Env env) {
        return ((AttrContext) env.info).staticLevel >
                ((AttrContext) env.outer.info).staticLevel;
    }

    /**
      * The static level of a symbol is the number of prefixing static flags
      *  for the symbol and any enclosing class or member declarations.
      */
    static int staticLevel(Symbol sym) {
        int level = 0;
        do {
            if ((sym.flags() & STATIC) != 0)
                level++;
            sym = sym.owner;
        } while (sym.kind != PCK)
            ;
        return level;
    }

    /**
      * Is class accessible in given evironment?
      *  @param env    The current environment.
      *  @param c      The class whose accessibility is checked.
      */
    boolean isAccessible(Env env, TypeSymbol c) {
        switch (c.flags() & AccessFlags) {
        case PRIVATE:
            return env.enclClass.sym.outermostClass() == c.owner.outermostClass();

        case 0:
            return env.toplevel.packge == c.owner ||
                    env.toplevel.packge == c.packge();

        case PUBLIC:
            return true;

        case PROTECTED:
            return env.toplevel.packge == c.owner ||
                    env.toplevel.packge == c.packge() ||
                    isInnerSubClass(env.enclClass.sym, c.owner);

        default:
            return true;

        }
    }

    /**
      * Is given class a subclass of given base class, or an inner class
      *  of a subclass?
      *  Return null if no such class exists.
      *  @param c     The class which is the subclass or is contained in it.
      *  @param base  The base class
      */
    private boolean isInnerSubClass(ClassSymbol c, Symbol base) {
        while (c != null && !c.isSubClass(base)) {
            c = c.owner.enclClass();
        }
        return c != null;
    }

    boolean isAccessible(Env env, Type t) {
        return t.tag == ARRAY ? isAccessible(env, t.elemtype()) :
                isAccessible(env, t.tsym);
    }

    /**
      * Is symbol accessible as a member of given type in given evironment?
      *  @param env    The current environment.
      *  @param site   The type of which the tested symbol is regarded
      *                as a member.
      *  @param sym    The symbol.
      */
    boolean isAccessible(Env env, Type site, Symbol sym) {
        if (sym.name == Names.init && sym.owner != site.tsym)
            return false;
        ClassSymbol sub;
        switch (sym.flags() & AccessFlags) {
        case PRIVATE:
            return (env.enclClass.sym == sym.owner ||
                    env.enclClass.sym.outermostClass() ==
                    sym.owner.outermostClass()) && sym.isMemberOf(site.tsym);

        case 0:
            return (env.toplevel.packge == sym.owner.owner ||
                    env.toplevel.packge == sym.packge()) && sym.isMemberOf(site.tsym);

        case PROTECTED:
            return env.toplevel.packge == sym.owner.owner ||
                    env.toplevel.packge == sym.packge() ||
                    isProtectedAccessible(sym, env.enclClass.sym, site) ||
                    ((AttrContext) env.info).selectSuper &&
                    (sym.flags() & STATIC) == 0 && sym.kind != TYP;

        default:
            return isAccessible(env, site);

        }
    }

    /**
      * Is given protected symbol accessible if it is selected from given site
      *  and the selection takes place in given class?
      *  @param sym     The symbol with protected access
      *  @param c       The class where the access takes place
      *  @site          The type of the qualifier
      */
    private boolean isProtectedAccessible(Symbol sym, ClassSymbol c, Type site) {
        while (c != null && !(c.isSubClass(sym.owner) &&
                ((sym.flags() & STATIC) != 0 || sym.kind == TYP ||
                site.tsym.isSubClass(c))))
            c = c.owner.enclClass();
        return c != null;
    }

    /**
      * Try to instantiate the type of a method so that it fits
      *  given type parameters and argument types. If succesful, return
      *  the method's instantiated result type, else return null.
      *  The instantiation will take into account an additional leading
      *  formal parameter if the method is an instance method seen as a member
      *  of un underdetermined site In this case, we treat site as an additional
      *  parameter and the parameters of the class containing the method as
      *  additional type variables that get instantiated.
      *
      *  @param site        The type of which the method is a member.
      *  @param m           The method symbol.
      *  @param typarams    The invocation's given type parameters (These are always
      *                     empty in the current version of GJ, even though this is
      *                     not assumed here).
      *  @param argtypes    The invocation's given value parameters.
      */
    Type instantiate(Type site, Symbol m, List typarams, List argtypes) {
        Type mt0 = site.memberType(m);
        MethodType mt = (MethodType) mt0;
        List formals = mt.argtypes;
        Type restype = m.name == Names.init ? site : mt.restype;
        if (Type.isSubTypes(argtypes, formals))
            return restype;
        else
            return null;
    }

    /**
      * Is method (or error) symbol `a' at least as good as method symbol `b'
      *  when both are seen as members of type `site',
      *  accessed from environment `env'?
      *  Used to get the best-matching overloaded instance during object lookup.
      */
    boolean isAsGood(Env env, Type site, Symbol a, Symbol b) {
        return a.kind < b.kind || a.kind == b.kind &&
                (!isAccessible(env, site, b) || isAccessible(env, site, a) &&
                (((a.flags() | b.flags()) & STATIC) != 0 ||
                (b.owner.flags() & INTERFACE) != 0 || a.owner.isSubClass(b.owner)) &&
                instantiate(site, b, Type.emptyList,
                site.memberType(a).argtypes()) != null);
    }

    /**
      * Find field. Synthetic fields are always skipped.
      *  @param env     The current environment.
      *  @param site    The original type from where the selection takes place.
      *  @param name    The name of the field.
      *  @param c       The class to search for the field. This is always
      *                 a superclass or implemented interface of site's class.
      */
    Symbol findField(Env env, Type site, Name name, TypeSymbol c) {
        Symbol bestSoFar = ResolveError.varNotFound;
        Symbol sym;
        Scope.Entry e = c.members().lookup(name);
        while (e.scope != null) {
            if (e.sym.kind == VAR && (e.sym.flags_field & SYNTHETIC) == 0) {
                if (isAccessible(env, site, e.sym))
                    return e.sym;
                else
                    return new AccessError(e.sym);
            }
            e = e.next();
        }
        Type st = c.type.supertype();
        if (st != null && st.tag == CLASS) {
            sym = findField(env, site, name, st.tsym);
            if (sym.kind < bestSoFar.kind)
                bestSoFar = sym;
        }
        for (List l = c.type.interfaces();
                bestSoFar.kind != AMBIGUOUS && l.nonEmpty(); l = l.tail) {
            sym = findField(env, site, name, ((Type) l.head).tsym);
            if (bestSoFar.kind < AMBIGUOUS && sym.kind < AMBIGUOUS &&
                    sym.owner != bestSoFar.owner)
                bestSoFar = new AmbiguityError(bestSoFar, sym);
            else if (sym.kind < bestSoFar.kind)
                bestSoFar = sym;
        }
        return bestSoFar;
    }

    /**
      * Find unqualified variable or field with given name.
      *  Synthetic fields always skipped.
      *  @param env     The current environment.
      *  @param name    The name of the variable or field.
      */
    Symbol findVar(Env env, Name name) {
        Symbol bestSoFar = ResolveError.varNotFound;
        Symbol sym;
        Env env1 = env;
        boolean staticOnly = false;
        while (env1.outer != null) {
            if (isStatic(env1))
                staticOnly = true;
            Scope.Entry e = ((AttrContext) env1.info).scope.lookup(name);
            while (e.scope != null &&
                    (e.sym.kind != VAR || (e.sym.flags_field & SYNTHETIC) != 0))
                e = e.next();
            sym = e.scope != null ? e.sym :
                    findField(env1, env1.enclClass.sym.type, name,
                    env1.enclClass.sym);
            if (sym.kind <= AMBIGUOUS) {
                if (staticOnly && sym.kind == VAR && sym.owner.kind == TYP &&
                        (sym.flags() & STATIC) == 0)
                    return new StaticError(sym);
                else
                    return sym;
            } else if (sym.kind < bestSoFar.kind) {
                bestSoFar = sym;
            }
            if ((env1.enclClass.sym.flags() & STATIC) != 0)
                staticOnly = true;
            env1 = env1.outer;
        }
        sym = findField(env, syms.predefClass.type, name, syms.predefClass);
        if (sym.kind <= AMBIGUOUS)
            return sym;
        else
            return bestSoFar;
    }

    /**
      * Find method in one of site's interfaces, which is a better match
      *  than `bestSoFar', or return `bestSoFar' if none found.
      *  @param env       The current environment.
      *  @param site      The original type from where the selection takes place.
      *  @param name      The method's name.
      *  @param typarams  The method's type parameters.
      *  @param argtypes  The method's value parameters.
      *  @param c         The class to search for the method. This is always
      *                   an implemented interface of site's class.
      *  @param bestSoFar The symbol that represents the best match so far.
      */
    Symbol findInterfaceMethod(Env env, Type site, Name name, List typarams,
            List argtypes, ClassSymbol c, Symbol bestSoFar) {
        for (List l = c.type.interfaces(); l.nonEmpty(); l = l.tail) {
            ClassSymbol i = (ClassSymbol)((Type) l.head).tsym;
            Scope.Entry e = i.members().lookup(name);
            while (e.scope != null) {
                if (e.sym.kind == MTH) {
                    if (instantiate(site, e.sym, typarams, argtypes) != null &&
                            !isAsGood(env, site, bestSoFar, e.sym))
                        bestSoFar = e.sym;
                    else if (bestSoFar.kind > WRONG_MTH)
                        bestSoFar = ResolveError.wrongMethod.setWrongSym(e.sym);
                    else if (bestSoFar.kind == WRONG_MTH)
                        bestSoFar = ResolveError.wrongMethods;
                }
                e = e.next();
            }
            bestSoFar = findInterfaceMethod(env, site, name, typarams, argtypes, i,
                    bestSoFar);
        }
        return bestSoFar;
    }

    /**
      * Check that `bestSoFar' is better than any method in one of
      *  site's interfaces;
      *  return `bestSoFar' if it is, an AmbiguityError if not.
      *  @param env       The current environment.
      *  @param site      The original type from where the selection takes place.
      *  @param name      The method's name.
      *  @param typarams  The method's type parameters.
      *  @param argtypes  The method's value parameters.
      *  @param c         The class to search for the method. This is always
      *                   an implemented interface of site's class.
      *  @param bestSoFar The symbol that represents the best match so far.
      */
    Symbol checkBestInterfaceMethod(Env env, Type site, Name name, List typarams,
            List argtypes, ClassSymbol c, Symbol bestSoFar) {
        for (List l = c.type.interfaces(); bestSoFar.kind == MTH && l.nonEmpty();
                l = l.tail) {
            ClassSymbol i = (ClassSymbol)((Type) l.head).tsym;
            Scope.Entry e = i.members().lookup(name);
            while (bestSoFar.kind == MTH && e.scope != null) {
                if (e.sym.kind == MTH &&
                        instantiate(site, e.sym, typarams, argtypes) != null &&
                        !isAsGood(env, site, bestSoFar, e.sym))
                    bestSoFar = new AmbiguityError(bestSoFar, e.sym);
                e = e.next();
            }
            bestSoFar = checkBestInterfaceMethod(env, site, name, typarams,
                    argtypes, i, bestSoFar);
        }
        return bestSoFar;
    }

    /**
      * Find best qualified method matching given name, type and value parameters.
      *  @param env       The current environment.
      *  @param site      The original type from where the selection takes place.
      *  @param name      The method's name.
      *  @param typarams  The method's type parameters.
      *  @param argtypes  The method's value parameters.
      */
    Symbol findMethod(Env env, Type site, Name name, List typarams, List argtypes) {
        Symbol bestSoFar = ResolveError.methodNotFound;
        Type ct = site.tsym.type;
        boolean isAbstract = true;
        while (ct.tag == CLASS) {
            ClassSymbol c = (ClassSymbol) ct.tsym;
            if ((c.flags() & (ABSTRACT | INTERFACE)) == 0)
                isAbstract = false;
            Scope.Entry e = c.members().lookup(name);
            while (e.scope != null) {
                if (e.sym.kind == MTH && (e.sym.flags_field & SYNTHETIC) == 0) {
                    if (instantiate(site, e.sym, typarams, argtypes) != null &&
                            !isAsGood(env, site, bestSoFar, e.sym))
                        bestSoFar = e.sym;
                    else if (bestSoFar.kind > WRONG_MTH)
                        bestSoFar = ResolveError.wrongMethod.setWrongSym(e.sym);
                    else if (bestSoFar.kind == WRONG_MTH)
                        bestSoFar = ResolveError.wrongMethods;
                }
                e = e.next();
            }
            if (isAbstract && !((AttrContext) env.info).selectSuper)
                bestSoFar = findInterfaceMethod(env, site, name, typarams, argtypes,
                        c, bestSoFar);
            ct = ct.supertype();
        }
        if (bestSoFar.kind == MTH && !isAccessible(env, site, bestSoFar))
            bestSoFar = new AccessError(bestSoFar);
        ct = site.tsym.type;
        isAbstract = true;
        while (bestSoFar.kind == MTH && ct.tag == CLASS) {
            ClassSymbol c = (ClassSymbol) ct.tsym;
            if ((c.flags() & (ABSTRACT | INTERFACE)) == 0)
                isAbstract = false;
            Scope.Entry e = c.members().lookup(name);
            while (bestSoFar.kind == MTH && e.scope != null) {
                if (bestSoFar != e.sym && e.sym.kind == MTH &&
                        (e.sym.flags_field & SYNTHETIC) == 0 &&
                        instantiate(site, e.sym, typarams, argtypes) != null &&
                        !isAsGood(env, site, bestSoFar, e.sym)) {
                    bestSoFar = new AmbiguityError(bestSoFar, e.sym);
                }
                e = e.next();
            }
            if (bestSoFar.kind == MTH && isAbstract &&
                    !((AttrContext) env.info).selectSuper)
                bestSoFar = checkBestInterfaceMethod(env, site, name, typarams,
                        argtypes, c, bestSoFar);
            ct = ct.supertype();
        }
        return bestSoFar;
    }

    /**
      * Find unqualified method matching given name, type and value parameters.
      *  @param env       The current environment.
      *  @param name      The method's name.
      *  @param typarams  The method's type parameters.
      *  @param argtypes  The method's value parameters.
      */
    Symbol findFun(Env env, Name name, List typarams, List argtypes) {
        Symbol bestSoFar = ResolveError.methodNotFound;
        Symbol sym;
        Env env1 = env;
        boolean staticOnly = false;
        while (env1.outer != null) {
            if (isStatic(env1))
                staticOnly = true;
            sym = findMethod(env1, env1.enclClass.sym.type, name, typarams, argtypes);
            if (sym.kind <= AMBIGUOUS || sym.kind == WRONG_MTHS ||
                    sym.kind == WRONG_MTH) {
                if (staticOnly && sym.kind == MTH && sym.owner.kind == TYP &&
                        (sym.flags() & STATIC) == 0)
                    return new StaticError(sym);
                else
                    return sym;
            } else if (sym.kind < bestSoFar.kind) {
                bestSoFar = sym;
            }
            if ((env1.enclClass.sym.flags() & STATIC) != 0)
                staticOnly = true;
            env1 = env1.outer;
        }
        sym = findMethod(env, syms.predefClass.type, name, typarams, argtypes);
        if (sym.kind <= AMBIGUOUS)
            return sym;
        else
            return bestSoFar;
    }

    /**
      * Load toplevel or member class with given fully qualified name and
      *  verify that it is accessible.
      *  @param env       The current environment.
      *  @param name      The fully qualified name of the class to be loaded.
      */
    Symbol loadClass(Env env, Name name) {
        try {
            {
                ClassSymbol c = syms.reader.loadClass(name);
                if (isAccessible(env, c))
                    return c;
                else
                    return new AccessError(c);
            }
        } catch (ClassReader.BadClassFile err) {
            {
                throw err;
            }
        }
        catch (CompletionFailure ex) {
            {
                return ResolveError.typeNotFound;
            }
        }
    }

    /**
      * Find qualified member type.
      *  @param env       The current environment.
      *  @param site      The original type from where the selection takes place.
      *  @param name      The type's name.
      *  @param c         The class to search for the member type. This is always
      *                   a superclass or implemented interface of site's class.
      */
    Symbol findMemberType(Env env, Type site, Name name, TypeSymbol c) {
        Symbol bestSoFar = ResolveError.typeNotFound;
        Symbol sym;
        Scope.Entry e = c.members().lookup(name);
        while (e.scope != null) {
            if (e.sym.kind == TYP) {
                if (isAccessible(env, site, e.sym))
                    return e.sym;
                else
                    return new AccessError(e.sym);
            }
            e = e.next();
        }
        Type st = c.type.supertype();
        if (st != null && st.tag == CLASS) {
            sym = findMemberType(env, site, name, st.tsym);
            if (sym.kind < bestSoFar.kind)
                bestSoFar = sym;
        }
        for (List l = c.type.interfaces();
                bestSoFar.kind != AMBIGUOUS && l.nonEmpty(); l = l.tail) {
            sym = findMemberType(env, site, name, ((Type) l.head).tsym);
            if (bestSoFar.kind < AMBIGUOUS && sym.kind < AMBIGUOUS &&
                    sym.owner != bestSoFar.owner)
                bestSoFar = new AmbiguityError(bestSoFar, sym);
            else if (sym.kind < bestSoFar.kind)
                bestSoFar = sym;
        }
        return bestSoFar;
    }

    /**
      * Find a global type in given scope and load corresponding class.
      *  @param env       The current environment.
      *  @param scope     The scope in which to look for the type.
      *  @param name      The type's name.
      */
    Symbol findGlobalType(Env env, Scope scope, Name name) {
        Symbol bestSoFar = ResolveError.typeNotFound;
        Scope.Entry e = scope.lookup(name);
        while (e.scope != null) {
            if (e.scope.owner == e.sym.owner) {
                Symbol sym = loadClass(env, e.sym.flatName());
                if (bestSoFar.kind == TYP && sym.kind == TYP && bestSoFar != sym)
                    return new AmbiguityError(bestSoFar, sym);
                else if (sym.kind < bestSoFar.kind)
                    bestSoFar = sym;
            }
            e = e.next();
        }
        return bestSoFar;
    }

    /**
      * Find an unqualified type symbol.
      *  @param env       The current environment.
      *  @param name      The type's name.
      */
    Symbol findType(Env env, Name name) {
        Symbol bestSoFar = ResolveError.typeNotFound;
        Symbol sym;
        Env env1 = env;
        boolean staticOnly = false;
        while (env1.outer != null) {
            if (isStatic(env1))
                staticOnly = true;
            Scope.Entry e = ((AttrContext) env1.info).scope.lookup(name);
            while (e.scope != null) {
                if (e.sym.kind == TYP) {
                    if (staticOnly && e.sym.type.tag == TYPEVAR &&
                            e.sym.owner.kind == TYP)
                        return new StaticError(e.sym);
                    else
                        return e.sym;
                }
                e = e.next();
            }
            sym = findMemberType(env1, env1.enclClass.sym.type, name,
                    env1.enclClass.sym);
            if (sym.kind <= AMBIGUOUS)
                return sym;
            else if (sym.kind < bestSoFar.kind)
                bestSoFar = sym;
            int flags = env1.enclClass.sym.flags();
            if ((flags & STATIC) != 0 && (flags & INTERFACE) == 0)
                staticOnly = true;
            env1 = env1.outer;
        }
        sym = findGlobalType(env, env.toplevel.namedImportScope, name);
        if (sym.kind <= AMBIGUOUS)
            return sym;
        else if (sym.kind < bestSoFar.kind)
            bestSoFar = sym;
        sym = findGlobalType(env, env.toplevel.packge.members(), name);
        if (sym.kind <= AMBIGUOUS)
            return sym;
        else if (sym.kind < bestSoFar.kind)
            bestSoFar = sym;
        sym = findGlobalType(env, env.toplevel.starImportScope, name);
        if (sym.kind <= AMBIGUOUS)
            return sym;
        else if (sym.kind < bestSoFar.kind)
            bestSoFar = sym;
        return bestSoFar;
    }

    /**
      * Find an unqualified identifier which matches a specified kind set.
      *  @param env       The current environment.
      *  @param name      The indentifier's name.
      *  @param kind      Indicates the possible symbol kinds
      *                   (a subset of VAL, TYP, PCK).
      */
    Symbol findIdent(Env env, Name name, int kind) {
        Symbol bestSoFar = ResolveError.typeNotFound;
        Symbol sym;
        if ((kind & VAR) != 0) {
            sym = findVar(env, name);
            if (sym.kind <= AMBIGUOUS)
                return sym;
            else if (sym.kind < bestSoFar.kind)
                bestSoFar = sym;
        }
        if ((kind & TYP) != 0) {
            sym = findType(env, name);
            if (sym.kind <= AMBIGUOUS)
                return sym;
            else if (sym.kind < bestSoFar.kind)
                bestSoFar = sym;
        }
        if ((kind & PCK) != 0)
            return syms.reader.enterPackage(name);
        else
            return bestSoFar;
    }

    /**
      * Find an identifier in a package which matches a specified kind set.
      *  @param env       The current environment.
      *  @param name      The identifier's name.
      *  @param kind      Indicates the possible symbol kinds
      *                   (a subset of TYP, PCK).
      */
    Symbol findIdentInPackage(Env env, TypeSymbol pck, Name name, int kind) {
        Name fullname = TypeSymbol.formFullName(name, pck);
        Symbol bestSoFar = ResolveError.typeNotFound;
        Symbol sym;
        if ((kind & TYP) != 0) {
            sym = loadClass(env, fullname);
            if (sym.kind <= AMBIGUOUS)
                return sym;
            else if (sym.kind < bestSoFar.kind)
                bestSoFar = sym;
        }
        if ((kind & PCK) != 0)
            return syms.reader.enterPackage(fullname);
        else
            return bestSoFar;
    }

    /**
      * Find an identifier among the members of a given type `site'.
      *  @param env       The current environment.
      *  @param site      The type containing the symbol to be found.
      *  @param name      The identifier's name.
      *  @param kind      Indicates the possible symbol kinds
      *                   (a subset of VAL, TYP).
      */
    Symbol findIdentInType(Env env, Type site, Name name, int kind) {
        Symbol bestSoFar = ResolveError.typeNotFound;
        Symbol sym;
        if ((kind & VAR) != 0) {
            sym = findField(env, site, name, site.tsym);
            if (sym.kind <= AMBIGUOUS)
                return sym;
            else if (sym.kind < bestSoFar.kind)
                bestSoFar = sym;
        }
        if ((kind & TYP) != 0) {
            sym = findMemberType(env, site, name, site.tsym);
            if (sym.kind <= AMBIGUOUS)
                return sym;
            else if (sym.kind < bestSoFar.kind)
                bestSoFar = sym;
        }
        return bestSoFar;
    }

    /**
      * If `sym' is a bad symbol: report error and return errSymbol
      *  else pass through unchanged,
      *  additional arguments duplicate what has been used in trying to find the
      *  symbol (--> flyweight pattern). This improves performance since we expect
      *  misses to happen frequently.
      *
      *  @param sym       The symbol that was found, or a ResolveError.
      *  @param pos       The position to use for error reporting.
      *  @param site      The original type from where the selection took place.
      *  @param name      The symbol's name.
      *  @param typarams  The invocation's type parameters,
      *                   if we looked for a method.
      *  @param argtypes  The invocation's value parameters,
      *                   if we looked for a method.
      */
    Symbol access(Symbol sym, int pos, Type site, Name name, List typarams,
            List argtypes) {
        if (sym.kind >= AMBIGUOUS) {
            if (!site.isErroneous() && !Type.isErroneous(argtypes))
                ((ResolveError) sym).report(log, pos, site, name, typarams, argtypes);
            return ((ResolveError) sym).sym;
        } else {
            return sym;
        }
    }

    /**
      * Same as above, but without type parameters and arguments.
      */
    Symbol access(Symbol sym, int pos, Type site, Name name) {
        if (sym.kind >= AMBIGUOUS)
            return access(sym, pos, site, name, Type.emptyList, Type.emptyList);
        else
            return sym;
    }

    /**
      * Check that sym is not an abstract method.
      */
    void checkNonAbstract(int pos, Symbol sym) {
        if ((sym.flags() & ABSTRACT) != 0)
            log.error(pos, "abstract.cant.be.accessed.directly",
                    kindName(sym.kind), sym.toJava());
    }

    /**
      * print all scopes starting with scope s and proceeding outwards.
      *  used for debugging.
      */
    public static void printscopes(Scope s) {
        while (s != null) {
            if (s.owner != null)
                System.err.print(s.owner + ": ");
            for (Scope.Entry e = s.elems; e != null; e = e.sibling) {
                if ((e.sym.flags() & ABSTRACT) != 0)
                    System.err.print("abstract ");
                System.err.print(e.sym + " ");
            }
            System.err.println();
            s = s.next;
        }
    }

    static void printscopes(Env env) {
        while (env.outer != null) {
            System.err.println("------------------------------");
            printscopes(((AttrContext) env.info).scope);
            env = env.outer;
        }
    }

    public static void printscopes(Type t) {
        while (t.tag == CLASS) {
            printscopes(t.tsym.members());
            t = t.supertype();
        }
    }

    /**
      * Warn about an unchecked method invocation.
      */
    void warnUncheckedInvocation(int pos, Symbol sym, List argtypes) {
        if (sym.kind < AMBIGUOUS) {
            chk.warnUnchecked(pos, "unchecked.meth.invocation.applied",
                    sym.toJava(), sym.javaLocation(), Type.toJavaList(argtypes));
        }
    }

    /**
      * Resolve an unqualified (non-method) identifier.
      *  @param pos       The position to use for error reporting.
      *  @param env       The environment current at the identifier use.
      *  @param name      The identifier's name.
      *  @param kind      The set of admissible symbol kinds for the identifier.
      */
    Symbol resolveIdent(int pos, Env env, Name name, int kind) {
        return access(findIdent(env, name, kind), pos, env.enclClass.sym.type, name);
    }

    /**
      * Resolve an unqualified method identifier.
      *  @param pos       The position to use for error reporting.
      *  @param env       The environment current at the method invocation.
      *  @param name      The identifier's name.
      *  @param typarams  The invocation's type parameters (this is currently
      *                   always empty).
      *  @param argtypes  The types of the invocation's value parameters.
      */
    Symbol resolveMethod(int pos, Env env, Name name, List typarams, List argtypes) {
        Symbol sym = findFun(env, name, typarams, argtypes);
        if (sym.kind >= WRONG_MTHS && Type.isRaw(argtypes)) {
            sym = findFun(env, name, typarams, Type.unerasure(argtypes));
            warnUncheckedInvocation(pos, sym, argtypes);
        }
        if (sym.kind >= AMBIGUOUS) {
            sym = access(sym, pos, env.enclClass.sym.type, name, typarams, argtypes);
        }
        return sym;
    }

    /**
      * Resolve a qualified method identifier
      *  @param pos       The position to use for error reporting.
      *  @param env       The environment current at the method invocation.
      *  @param site      The type of the qualifying expression, in which identifier
      *                   is searched.
      *  @param name      The identifier's name.
      *  @param typarams  The invocation's type parameters (this is currently
      *                   always empty).
      *  @param argtypes  The types of the invocation's value parameters.
      */
    Symbol resolveQualifiedMethod(int pos, Env env, Type site, Name name,
            List typarams, List argtypes) {
        Symbol sym = findMethod(env, site, name, typarams, argtypes);
        if (sym.kind >= WRONG_MTHS && Type.isRaw(argtypes)) {
            sym = findMethod(env, site, name, typarams, Type.unerasure(argtypes));
            warnUncheckedInvocation(pos, sym, argtypes);
        }
        if (sym.kind >= AMBIGUOUS) {
            sym = access(sym, pos, site, name, typarams, argtypes);
        }
        return sym;
    }

    /**
      * Resolve a qualified method identifier, throw a fatal error if not found.
      *  @param pos       The position to use for error reporting.
      *  @param env       The environment current at the method invocation.
      *  @param site      The type of the qualifying expression, in which identifier
      *                   is searched.
      *  @param name      The identifier's name.
      *  @param typarams  The invocation's type parameters (this is currently
      *                   always empty).
      *  @param argtypes  The types of the invocation's value parameters.
      */
    Symbol resolveInternalMethod(int pos, Env env, Type site, Name name,
            List typarams, List argtypes) {
        Symbol sym = resolveQualifiedMethod(pos, env, site, name, typarams, argtypes);
        if (sym.kind == MTH)
            return sym;
        else
            throw new FatalError( Log.getLocalizedString("fatal.err.cant.locate.meth",
                    name.toJava()));
    }

    /**
      * Resolve constructor.
      *  @param pos       The position to use for error reporting.
      *  @param env       The environment current at the constructor invocation.
      *  @param site      The type of class for which a constructor is searched.
      *  @param argtypes  The types of the constructor invocation's value parameters.
      */
    Symbol resolveConstructor(int pos, Env env, Type site, List argtypes) {
        Symbol sym = resolveQualifiedMethod(pos, env, site, Names.init,
                Type.emptyList, argtypes);
        if ((sym.flags() & DEPRECATED) != 0)
            chk.warnDeprecated(pos, sym);
        return sym;
    }

    /**
      * Resolve operator.
      *  @param pos       The position to use for error reporting.
      *  @param optag     The tag of the operation tree.
      *  @param env       The environment current at the operation.
      *  @param argtypes  The types of the operands.
      */
    Symbol resolveOperator(int pos, int optag, Env env, List argtypes) {
        Name name = TreeInfo.operatorName(optag);
        return access( findMethod(env, syms.predefClass.type, name, Type.emptyList,
                argtypes), pos, env.enclClass.sym.type, name, Type.emptyList,
                argtypes);
    }

    /**
      * Resolve `c.name' where name == this or name == super.
      *  @param pos           The position to use for error reporting.
      *  @param env           The environment current at the expression.
      *  @param site          The qualifier.
      *  @param name          The identifier's name.
      *  @param preciseMatch  A flag that indicates whether `c' must be an
      *                       enclosing class, or just a superclass of an enclosing
      *                       class.
      */
    Symbol resolveSelf(int pos, Env env, TypeSymbol c, Name name,
            boolean preciseMatch) {
        Env env1 = env;
        boolean staticOnly = false;
        while (env1.outer != null) {
            if (isStatic(env1))
                staticOnly = true;
            if (env1.enclClass.sym == c || !preciseMatch &&
                    env1.enclClass.sym.isSubClass(c)) {
                Symbol sym = ((AttrContext) env1.info).scope.lookup(name).sym;
                if (sym != null) {
                    if (staticOnly)
                        sym = new StaticError(sym);
                    return access(sym, pos, env.enclClass.sym.type, name);
                }
            }
            if ((env1.enclClass.sym.flags() & STATIC) != 0)
                staticOnly = true;
            env1 = env1.outer;
        }
        log.error(pos, "not.encl.class", c.toJava());
        return Symbol.errSymbol;
    }

    /**
      * A localized string describing a given kind.
      */
    static String kindName(int kind) {
        switch (kind) {
        case PCK:
            return Log.getLocalizedString("kindname.package");

        case TYP:
            return Log.getLocalizedString("kindname.class");

        case VAR:
            return Log.getLocalizedString("kindname.variable");

        case VAL:
            return Log.getLocalizedString("kindname.value");

        case MTH:
            return Log.getLocalizedString("kindname.method");

        default:
            return Log.getLocalizedString("kindname.default", Integer.toString(kind));

        }
    }

    /**
      * A localized string describing a given set of kinds.
      */
    static String kindNames(int kind) {
        String[] s = new String[4];
        int i = 0;
        if ((kind & VAL) != 0)
            s[i++] = (kind & VAL) == VAR ?
                    Log.getLocalizedString("kindname.variable") :
                    Log.getLocalizedString("kindname.value");
        if ((kind & MTH) != 0)
            s[i++] = Log.getLocalizedString("kindname.method");
        if ((kind & TYP) != 0)
            s[i++] = Log.getLocalizedString("kindname.class");
        if ((kind & PCK) != 0)
            s[i++] = Log.getLocalizedString("kindname.package");
        String names = "";
        for (int j = 0; j < i - 1; j++)
            names = names + s[j] + ", ";
        if (i >= 1)
            names = names + s[i - 1];
        else
            names = Log.getLocalizedString("kindname.default",
                    Integer.toString(kind));
        return names;
    }

    /**
      * A localized string describing the kind -- either class or interface --
      *  of a given type.
      */
    static String typeKindName(Type t) {
        if (t.tag == TYPEVAR)
            return Log.getLocalizedString("kindname.type.variable");
        else if (t.tag == PACKAGE)
            return Log.getLocalizedString("kindname.package");
        else if ((t.tsym.flags_field & INTERFACE) != 0)
            return Log.getLocalizedString("kindname.interface");
        else
            return Log.getLocalizedString("kindname.class");
    }

    /**
      * A localized string describing the kind of a missing symbol, given an
      *  error kind.
      */
    static String absentKindName(int kind) {
        switch (kind) {
        case ABSENT_VAR:
            return Log.getLocalizedString("kindname.variable");

        case WRONG_MTHS:

        case WRONG_MTH:

        case ABSENT_MTH:
            return Log.getLocalizedString("kindname.method");

        case ABSENT_TYP:
            return Log.getLocalizedString("kindname.class");

        default:
            return Log.getLocalizedString("kindname.identifier");

        }
    }

    /**
      * Root class for resolve errors.
      *  Instances of this class indicate "Symbol not found".
      *  Instances of subclass indicate other errors.
      */
    static class ResolveError extends Symbol implements TypeTags {
        static final ResolveError varNotFound = new ResolveError(ABSENT_VAR);
        static final ResolveError wrongMethod = new ResolveError(WRONG_MTH);
        static final ResolveError wrongMethods = new ResolveError(WRONG_MTHS);
        static final ResolveError methodNotFound = new ResolveError(ABSENT_MTH);
        static final ResolveError typeNotFound = new ResolveError(ABSENT_TYP);

        ResolveError(int kind) {
            super(kind, 0, null, null, null);
        }

        /**
          * The symbol that was determined by resolution, or errSymbol if none
          *  was found.
          */
        Symbol sym = Symbol.errSymbol;

        /**
         * The symbol that was a close mismatch, or null if none was found.
         *  wrongSym is currently set if a simgle method with the correct name, but
         *  the wrong parameters was found.
         */
        Symbol wrongSym;

        /**
         * Update wrongSym and return this
         */
        ResolveError setWrongSym(Symbol sym) {
            this.wrongSym = sym;
            return this;
        }

        /**
          * Report error.
          *  @param log       The error log to be used for error reporting.
          *  @param pos       The position to be used for error reporting.
          *  @param site      The original type from where the selection took place.
          *  @param name      The name of the symbol to be resolved.
          *  @param typarams  The invocation's type parameters,
          *                   if we looked for a method.
          *  @param argtypes  The invocation's value parameters,
          *                   if we looked for a method.
          */
        void report(Log log, int pos, Type site, Name name, List typarams,
                List argtypes) {
            if (name != Names.error) {
                String kindname = absentKindName(kind);
                String idname = name.toJava();
                String targs = "";
                String args = "";
                if (kind >= WRONG_MTHS && kind <= ABSENT_MTH) {
                    if (isOperator(name)) {
                        log.error(pos, "operator.cant.be.applied", name.toJava(),
                                Type.toJavaList(argtypes));
                        return;
                    }
                    if (name == Names.init) {
                        kindname = log.getLocalizedString("kindname.constructor");
                        idname = site.tsym.name.toJava();
                    }
                    if (typarams.nonEmpty())
                        targs = "<" + Type.toJavaList(typarams) + ">";
                    args = "(" + Type.toJavaList(argtypes) + ")";
                }
                if (kind == WRONG_MTH) {
                    log.error(pos, "cant.apply.symbol",
                            wrongSym.asMemberOf(site).toJava(),
                            wrongSym.javaLocation(site), Type.toJavaList(argtypes));
                } else {
                    if (site.tsym.name.len != 0) {
                        log.error(pos, "cant.resolve.location", kindname, idname,
                                targs, args, typeKindName(site), site.toJava());
                    } else {
                        log.error(pos, "cant.resolve", kindname, idname, targs, args);
                    }
                }
            }
        }

        /**
          * A name designates an operator if it consists
          *  of a non-empty sequence of operator symbols +-~!/*%&|^<>=
          */
        boolean isOperator(Name name) {
            int i = 0;
            while (i < name.len && "+-~!*/%&|^<>=".indexOf(name.byteAt(i)) >= 0)
                i++;
            return i > 0 && i == name.len;
        }
    }

    /**
      * Resolve error class indicating that a symbol is not accessible.
      */
    static class AccessError extends ResolveError {

        AccessError(Symbol sym) {
            super(HIDDEN);
            this.sym = sym;
        }

        /**
          * Report error.
          *  @param log       The error log to be used for error reporting.
          *  @param pos       The position to be used for error reporting.
          *  @param site      The original type from where the selection took place.
          *  @param name      The name of the symbol to be resolved.
          *  @param typarams  The invocation's type parameters,
          *                   if we looked for a method.
          *  @param argtypes  The invocation's value parameters,
          *                   if we looked for a method.
          */
        void report(Log log, int pos, Type site, Name name, List typarams,
                List argtypes) {
            if (sym.owner.type.tag != ERROR) {
                if (sym.name == Names.init && sym.owner != site.tsym)
                    ResolveError.methodNotFound.report(log, pos, site, name,
                            typarams, argtypes);
                if ((sym.flags() & PUBLIC) != 0)
                    log.error(pos, "not.def.public.class.intf.cant.access",
                            sym.toJava(), sym.javaLocation());
                else if ((sym.flags() & (PRIVATE | PROTECTED)) != 0)
                    log.error(pos, "report.access", sym.toJava(),
                            TreeInfo.flagNames(sym.flags() &
                            (PRIVATE | PROTECTED)), sym.javaLocation());
                else
                    log.error(pos, "not.def.public.cant.access", sym.toJava(),
                            sym.javaLocation());
            }
        }
    }

    /**
      * Resolve error class indicating that an instance member was accessed
      *  from a static context.
      */
    static class StaticError extends ResolveError {

        StaticError(Symbol sym) {
            super(HIDDEN);
            this.sym = sym;
        }

        /**
          * Report error.
          *  @param log       The error log to be used for error reporting.
          *  @param pos       The position to be used for error reporting.
          *  @param site      The original type from where the selection took place.
          *  @param name      The name of the symbol to be resolved.
          *  @param typarams  The invocation's type parameters,
          *                   if we looked for a method.
          *  @param argtypes  The invocation's value parameters,
          *                   if we looked for a method.
          */
        void report(Log log, int pos, Type site, Name name, List typarams,
                List argtypes) {
            log.error(pos, "non-static.cant.be.ref", kindName(sym.kind),
                    sym.toJava());
        }
    }

    /**
      * Resolve error class indicating an ambiguous reference.
      */
    static class AmbiguityError extends ResolveError {
        Symbol sym1;
        Symbol sym2;

        AmbiguityError(Symbol sym1, Symbol sym2) {
            super(AMBIGUOUS);
            this.sym1 = sym1;
            this.sym2 = sym2;
        }

        /**
          * Report error.
          *  @param log       The error log to be used for error reporting.
          *  @param pos       The position to be used for error reporting.
          *  @param site      The original type from where the selection took place.
          *  @param name      The name of the symbol to be resolved.
          *  @param typarams  The invocation's type parameters,
          *                   if we looked for a method.
          *  @param argtypes  The invocation's value parameters,
          *                   if we looked for a method.
          */
        void report(Log log, int pos, Type site, Name name, List typarams,
                List argtypes) {
            Name sname = sym1.name;
            if (sname == Names.init)
                sname = sym1.owner.name;
            log.error(pos, "ref.ambiguous", sname.toJava(), kindName(sym1.kind),
                    sym1.toJava(), sym1.javaLocation(site), kindName(sym2.kind),
                    sym2.toJava(), sym2.javaLocation(site));
        }
    }
}
