package com.sun.tools.javac.v8.comp;
import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.tree.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.tree.Tree.*;

import com.sun.tools.javac.v8.code.Type.*;


/**
 * This pass maps Java with inner classes to flat Java without
 *  inner classes.
 */
public class TransInner extends TreeTranslator implements Flags, Kinds, TypeTags,
ByteCodes {
    Log log;
    Symtab syms;
    Resolve rs;
    Check chk;
    Attr attr;
    TreeMaker make;

    public TransInner(Log log, Symtab syms, Resolve rs, Check chk, Attr attr,
            TreeMaker make) {
        super();
        this.log = log;
        this.syms = syms;
        this.rs = rs;
        this.chk = chk;
        this.attr = attr;
        this.make = make;
    }

    /**
      * The currently enclosing class.
      */
    ClassSymbol currentClass;

    /**
     * A queue of all translated classes.
     */
    ListBuffer translated;

    /**
     * Environment for symbol lookup, set by translateTopLevelClass.
     */
    Env attrEnv;

    /**
     * A hash table mapping local classes to their definitions.
     */
    Hashtable classdefs;

    /**
     * A hash table mapping virtual accessed symbols in outer subclasses
     *  to the actually referred symbol in superclasses.
     */
    Hashtable actualSymbols;

    /**
     * The currently enclosing outermost class definition.
     */
    Tree outermostClassDef;

    /**
     * The currently enclosing outermost member definition.
     */
    Tree outermostMemberDef;

    /**
     * A navigator class for assembling a mapping from local class symbols
     *  to class definition trees.
     *  There is only one case; all other cases simply traverse down the tree.
     */
    class ClassMap extends TreeTranslator {

        ClassMap() {
            super();
        }

        /**
          * All encountered class defs are entered into classdefs table.
          */
        public void _case(ClassDef tree) {
            classdefs.put(tree.sym, tree);
            super._case(tree);
        }
    }

    /**
      * Map a class symbol to its definition.
      *  @param c    The class symbol of which we want to determine the definition.
      */
    ClassDef classDef(ClassSymbol c) {
        ClassDef def = (ClassDef) classdefs.get(c);
        if (def == null && outermostMemberDef != null) {
            new ClassMap().translate(outermostMemberDef);
            def = (ClassDef) classdefs.get(c);
        }
        if (def == null) {
            new ClassMap().translate(outermostClassDef);
            def = (ClassDef) classdefs.get(c);
        }
        return def;
    }

    /**
      * A hash table mapping class symbols to lists of free variables.
      *  accessed by them. Only free variables of the method immediately containing
      *  a class are associated with that class.
      */
    Hashtable freevarCache;

    /**
     * A navigator class for collecting the free variables accessed
     *  from a local class.
     *  There is only one case; all other cases simply traverse down the tree.
     */
    class FreeVarCollector extends TreeTranslator {

        /**
         * The owner of the local class.
         */
        Symbol owner;

        /**
         * The list of owner's variables accessed from within the local class,
         *  without any duplicates.
         */
        List fvs;

        FreeVarCollector(Symbol owner) {
            super();
            this.owner = owner;
            this.fvs = VarSymbol.emptyList;
        }

        /**
          * Add free variable to fvs list unless it is already there.
          */
        private void addFreeVar(VarSymbol v) {
            for (List l = fvs; l.nonEmpty(); l = l.tail)
                if (l.head == v)
                    return;
            fvs = fvs.prepend(v);
        }

        /**
          * Add all free variables of class c to fvs list
          *  unless they are already there.
          */
        private void addFreeVars(ClassSymbol c) {
            List fvs = (List) freevarCache.get(c);
            if (fvs != null) {
                for (List l = fvs; l.nonEmpty(); l = l.tail) {
                    addFreeVar((VarSymbol) l.head);
                }
            }
        }

        /**
          * If tree refers to a variable in owner of local class, add it to
          *  free variables list.
          */
        public void _case(Ident tree) {
            result = tree;
            if (tree.sym.kind == VAR && tree.sym.owner == owner) {
                VarSymbol v = (VarSymbol) tree.sym;
                if (v.constValue == null && (v.flags_field & CAPTURED) != 0)
                    addFreeVar(v);
            }
        }

        /**
          * If tree refers to a class instance creation expression
          *  add all free variables of the freshly created class.
          */
        public void _case(NewClass tree) {
            addFreeVars((ClassSymbol) tree.constructor.owner);
            super._case(tree);
        }

        /**
          * If tree refers to a superclass constructor call
          *  add all free variables of the superclass.
          */
        public void _case(Apply tree) {
            if (TreeInfo.name(tree.meth) == Names._super)
                addFreeVars((ClassSymbol) TreeInfo.symbol(tree.meth).owner);
            super._case(tree);
        }
    }

    /**
      * Return the variables accessed from within a local class, which
      *  are declared in the local class' owner.
      *  (in reverse order of first access).
      */
    List freevars(ClassSymbol c) {
        if ((c.owner.kind & (VAR | MTH)) != 0) {
            List fvs = (List) freevarCache.get(c);
            if (fvs == null) {
                FreeVarCollector collector = new FreeVarCollector(c.owner);
                collector.translate(classDef(c));
                fvs = collector.fvs;
                freevarCache.put(c, fvs);
            }
            return fvs;
        } else {
            return VarSymbol.emptyList;
        }
    }

    /**
      * Make an attributed tree representing a literal. This will be an
      *  Ident node in the case of boolean literals, a Literal node in all
      *  other cases.
      *  @param type       The literal's type.
      *  @param value      The literal's value.
      */
    Tree makeLit(Type type, Object value) {
        if (type.tag == BOOLEAN) {
            return make.Ident(((Integer) value).intValue() == 0 ?
                    syms.falseConst : syms.trueConst);
        } else {
            return make.Literal(type.tag, value).setType(type.constType(value));
        }
    }

    /**
      * Make an attributed class instance creation expression.
      *  @param ctype    The class type.
      *  @param args     The constructor arguments.
      */
    NewClass makeNewClass(Type ctype, List args) {
        NewClass tree = make.NewClass(null, make.QualIdent(ctype.tsym), args, null);
        tree.constructor = rs.resolveConstructor(make.pos, attrEnv, ctype,
                TreeInfo.types(args));
        tree.type = ctype;
        return tree;
    }

    /**
      * Make an attributed operation expression.
      *  @param optag    The operators tree tag.
      *  @param args     The constructor arguments.
      */
    Operation makeOperation(int optag, List args) {
        Operation tree = make.Operation(optag, args);
        tree.operator =
                rs.resolveOperator(make.pos, optag, attrEnv, TreeInfo.types(args));
        tree.type = tree.operator.type.restype();
        return tree;
    }

    /**
      * Convert tree into string object, unless it has already a reference type..
      */
    Tree makeString(Tree tree) {
        if (tree.type.tag >= CLASS) {
            return tree;
        } else {
            Symbol valueOfSym =
                    rs.resolveInternalMethod(tree.pos, attrEnv, syms.stringType,
                    Names.valueOf, Type.emptyList, Type.emptyList.prepend(tree.type));
            return make.App(make.QualIdent(valueOfSym), List.make(tree));
        }
    }

    /**
      * Create an empty anonymous class definition and enter and complete
      *  its symbol. Return the class definition's symbol.
      *  and create
      *  @param flags    The class symbol's flags
      *  @param owner    The class symbol's owner
      */
    ClassSymbol makeEmptyClass(int flags, ClassSymbol owner) {
        ClassSymbol c = syms.reader.defineClass(Names.empty, owner);
        c.flatname = chk.localClassName(c, 1);
        c.sourcefile = owner.sourcefile;
        c.completer = null;
        c.members_field = new Scope(c);
        ClassType ctype = (ClassType) c.type;
        ctype.supertype_field = syms.objectType;
        ctype.interfaces_field = Type.emptyList;
        owner.members().enter(c);
        chk.compiled.put(c.flatname, c);
        ClassDef cdef =
                make.ClassDef(flags, Names.empty, TypeParameter.emptyList, null,
                Tree.emptyList, Tree.emptyList);
        cdef.sym = c;
        ClassDef odef = classDef(owner);
        odef.defs = odef.defs.prepend(cdef);
        return c;
    }

    /**
      * Access codes for dereferencing, assignment,
      *  and pre/post increment/decrement.
      *  Access codes for assignment operations are determined by method accessCode
      *  below.
      *
      *  All access codes for accesses to the current class are even.
      *  If a member of the superclass should be accessed instead (because
      *  access was via a qualified super), add one to the corresponding code
      *  for the current class, making the number odd.
      *  This numbering scheme is used by the backend to decide whether
      *  to issue an invokevirtual or invokespecial call.
      *
      *  @see Gen._case(Select tree)
      */
    private static final int DEREFcode = 0;

    /**
     * Access codes for dereferencing, assignment,
     *  and pre/post increment/decrement.
     *  Access codes for assignment operations are determined by method accessCode
     *  below.
     *
     *  All access codes for accesses to the current class are even.
     *  If a member of the superclass should be accessed instead (because
     *  access was via a qualified super), add one to the corresponding code
     *  for the current class, making the number odd.
     *  This numbering scheme is used by the backend to decide whether
     *  to issue an invokevirtual or invokespecial call.
     *
     *  @see Gen._case(Select tree)
     */
    private static final int ASSIGNcode = 2;

    /**
     * Access codes for dereferencing, assignment,
     *  and pre/post increment/decrement.
     *  Access codes for assignment operations are determined by method accessCode
     *  below.
     *
     *  All access codes for accesses to the current class are even.
     *  If a member of the superclass should be accessed instead (because
     *  access was via a qualified super), add one to the corresponding code
     *  for the current class, making the number odd.
     *  This numbering scheme is used by the backend to decide whether
     *  to issue an invokevirtual or invokespecial call.
     *
     *  @see Gen._case(Select tree)
     */
    private static final int PREINCcode = 4;

    /**
     * Access codes for dereferencing, assignment,
     *  and pre/post increment/decrement.
     *  Access codes for assignment operations are determined by method accessCode
     *  below.
     *
     *  All access codes for accesses to the current class are even.
     *  If a member of the superclass should be accessed instead (because
     *  access was via a qualified super), add one to the corresponding code
     *  for the current class, making the number odd.
     *  This numbering scheme is used by the backend to decide whether
     *  to issue an invokevirtual or invokespecial call.
     *
     *  @see Gen._case(Select tree)
     */
    private static final int PREDECcode = 6;

    /**
     * Access codes for dereferencing, assignment,
     *  and pre/post increment/decrement.
     *  Access codes for assignment operations are determined by method accessCode
     *  below.
     *
     *  All access codes for accesses to the current class are even.
     *  If a member of the superclass should be accessed instead (because
     *  access was via a qualified super), add one to the corresponding code
     *  for the current class, making the number odd.
     *  This numbering scheme is used by the backend to decide whether
     *  to issue an invokevirtual or invokespecial call.
     *
     *  @see Gen._case(Select tree)
     */
    private static final int POSTINCcode = 8;

    /**
     * Access codes for dereferencing, assignment,
     *  and pre/post increment/decrement.
     *  Access codes for assignment operations are determined by method accessCode
     *  below.
     *
     *  All access codes for accesses to the current class are even.
     *  If a member of the superclass should be accessed instead (because
     *  access was via a qualified super), add one to the corresponding code
     *  for the current class, making the number odd.
     *  This numbering scheme is used by the backend to decide whether
     *  to issue an invokevirtual or invokespecial call.
     *
     *  @see Gen._case(Select tree)
     */
    private static final int POSTDECcode = 10;

    /**
     * Access codes for dereferencing, assignment,
     *  and pre/post increment/decrement.
     *  Access codes for assignment operations are determined by method accessCode
     *  below.
     *
     *  All access codes for accesses to the current class are even.
     *  If a member of the superclass should be accessed instead (because
     *  access was via a qualified super), add one to the corresponding code
     *  for the current class, making the number odd.
     *  This numbering scheme is used by the backend to decide whether
     *  to issue an invokevirtual or invokespecial call.
     *
     *  @see Gen._case(Select tree)
     */
    private static final int FIRSTASGOPcode = 12;

    /**
     * Number of access codes
     */
    private static final int NCODES = accessCode(ByteCodes.lushrl) + 2;

    /**
     * A mapping from symbols to their access numbers.
     */
    private Hashtable accessNums;

    /**
     * A mapping from symbols to an array of access symbols, indexed by
     *  access code.
     */
    private Hashtable accessSyms;

    /**
     * A mapping from (constructor) symbols to access constructor symbols.
     */
    private Hashtable accessConstrs;

    /**
     * A queue for all accessed symbols.
     */
    private ListBuffer accessed;

    /**
     * Map bytecode of binary operation to access code of corresponding
     *  assignment operation. This is always an even number.
     */
    private static int accessCode(int bytecode) {
        if (ByteCodes.iadd <= bytecode && bytecode <= ByteCodes.lxor)
            return (bytecode - iadd) * 2 + FIRSTASGOPcode;
        else if (bytecode == ByteCodes.string_add)
            return (ByteCodes.lxor + 1 - iadd) * 2 + FIRSTASGOPcode;
        else if (ByteCodes.ishll <= bytecode && bytecode <= ByteCodes.lushrl)
            return (bytecode - ishll + ByteCodes.lxor + 2 - iadd) * 2 +
                    FIRSTASGOPcode;
        else
            return -1;
    }

    /**
      * return access code for identifier,
      *  @param tree     The tree representing the identifier use.
      *  @param enclOp   The closest enclosing operation node of tree,
      *                  null if tree is not a subtree of an operation.
      */
    private static int accessCode(Tree tree, Tree enclOp) {
        if (enclOp == null)
            return DEREFcode;
        else if (enclOp.tag == Tree.ASSIGN && tree == ((Assign) enclOp).lhs)
            return ASSIGNcode;
        else if (Tree.PREINC <= enclOp.tag && enclOp.tag <= Tree.POSTDEC &&
                tree == ((Operation) enclOp).args.head)
            return (enclOp.tag - Tree.PREINC) * 2 + PREINCcode;
        else if (Tree.BITOR_ASG <= enclOp.tag && enclOp.tag <= Tree.MOD_ASG &&
                tree == ((Assignop) enclOp).lhs)
            return accessCode(((OperatorSymbol)((Assignop) enclOp).operator).opcode);
        else
            return DEREFcode;
    }

    /**
      * Return binary operator that corresponds to given access code.
      */
    private OperatorSymbol binaryAccessOperator(int acode) {
        for (Scope.Entry e = syms.predefClass.members().elems; e != null;
                e = e.sibling) {
            if (e.sym instanceof OperatorSymbol) {
                OperatorSymbol op = (OperatorSymbol) e.sym;
                if (accessCode(op.opcode) == acode)
                    return op;
            }
        }
        return null;
    }

    /**
      * Return tree tag for assignment operation corresponding
      *  to given binary operator.
      */
    private static int treeTag(OperatorSymbol operator) {
        switch (operator.opcode) {
        case ByteCodes.ior:

        case ByteCodes.lor:
            return Tree.BITOR_ASG;

        case ByteCodes.ixor:

        case ByteCodes.lxor:
            return Tree.BITXOR_ASG;

        case ByteCodes.iand:

        case ByteCodes.land:
            return Tree.BITAND_ASG;

        case ByteCodes.ishl:

        case ByteCodes.lshl:

        case ByteCodes.ishll:

        case ByteCodes.lshll:
            return Tree.SL_ASG;

        case ByteCodes.ishr:

        case ByteCodes.lshr:

        case ByteCodes.ishrl:

        case ByteCodes.lshrl:
            return Tree.SR_ASG;

        case ByteCodes.iushr:

        case ByteCodes.lushr:

        case ByteCodes.iushrl:

        case ByteCodes.lushrl:
            return Tree.USR_ASG;

        case ByteCodes.iadd:

        case ByteCodes.ladd:

        case ByteCodes.fadd:

        case ByteCodes.dadd:

        case ByteCodes.string_add:
            return Tree.PLUS_ASG;

        case ByteCodes.isub:

        case ByteCodes.lsub:

        case ByteCodes.fsub:

        case ByteCodes.dsub:
            return Tree.MINUS_ASG;

        case ByteCodes.imul:

        case ByteCodes.lmul:

        case ByteCodes.fmul:

        case ByteCodes.dmul:
            return Tree.MUL_ASG;

        case ByteCodes.idiv:

        case ByteCodes.ldiv:

        case ByteCodes.fdiv:

        case ByteCodes.ddiv:
            return Tree.DIV_ASG;

        case ByteCodes.imod:

        case ByteCodes.lmod:

        case ByteCodes.fmod:

        case ByteCodes.dmod:
            return Tree.MOD_ASG;

        default:
            throw new InternalError();

        }
    }

    /**
      * The name of the access method with number `anum' and access code `acode'.
      */
    static Name accessName(int anum, int acode) {
        return Name.fromString("access$" + anum + acode / 10 + acode % 10);
    }

    /**
      * Return access symbol for a private or protected symbol from an inner class.
      *  @param sym        The accessed private symbol.
      *  @param tree       The accessing tree.
      *  @param enclOp     The closest enclosing operation node of tree,
      *                    null if tree is not a subtree of an operation.
      *  @param protAccess Is access to a protected symbol in another
      *                    package?
      */
    MethodSymbol accessSymbol(Symbol sym, Tree tree, Tree enclOp,
            boolean protAccess) {
        ClassSymbol accOwner = accessClass(sym, protAccess);
        Symbol vsym = sym;
        if (sym.owner != accOwner) {
            vsym = sym.clone(accOwner);
            actualSymbols.put(vsym, sym);
        }
        Integer anum = (Integer) accessNums.get(vsym);
        if (anum == null) {
            anum = new Integer(accessed.length());
            accessNums.put(vsym, anum);
            accessSyms.put(vsym, new MethodSymbol[NCODES]);
            accessed.append(vsym);
        }
        int acode;
        List argtypes;
        Type restype;
        List thrown;
        if (vsym.kind == VAR) {
            acode = accessCode(tree, enclOp);
            if (acode >= FIRSTASGOPcode) {
                OperatorSymbol operator = binaryAccessOperator(acode);
                if (operator.opcode == string_add)
                    argtypes = List.make(syms.objectType);
                else
                    argtypes = operator.type.argtypes().tail;
            } else if (acode == ASSIGNcode)
                argtypes = Type.emptyList.prepend(vsym.erasure());
            else
                argtypes = Type.emptyList;
            restype = vsym.erasure();
            thrown = ClassSymbol.emptyList;
        } else if (vsym.kind == MTH) {
            acode = DEREFcode;
            argtypes = vsym.erasure().argtypes();
            restype = vsym.erasure().restype();
            thrown = vsym.type.thrown();
        } else {
            throw new InternalError();
        }
        if (protAccess)
            acode++;
        if ((vsym.flags() & STATIC) == 0) {
            argtypes = argtypes.prepend(vsym.owner.erasure());
        }
        MethodSymbol[] accessors = (MethodSymbol[]) accessSyms.get(vsym);
        MethodSymbol accessor = accessors[acode];
        if (accessor == null) {
            accessor = new MethodSymbol(STATIC | SYNTHETIC,
                    accessName(anum.intValue(), acode),
                    new MethodType(argtypes, restype, thrown), accOwner);
            accOwner.members().enter(accessor);
            accessors[acode] = accessor;
        }
        return accessor;
    }

    /**
      * The qualifier to be used for accessing a symbol in an outer class.
      *  This is either C.sym or C.this.sym, depending on whether or not
      *  sym is static.
      *  @param sym   The accessed symbol.
      */
    Tree accessBase(int pos, Symbol sym) {
        return (sym.flags() & STATIC) != 0 ?
                access(make.at(pos).QualIdent(sym.owner)) : makeOwnerThis(pos, sym);
    }

    /**
      * Do we need an access method to reference private symbol?
      */
    boolean needsPrivateAccess(Symbol sym) {
        return (sym.flags() & PRIVATE) != 0 && sym.owner != currentClass;
    }

    /**
      * Do we need an access method to reference symbol in other package?
      */
    boolean needsProtectedAccess(Symbol sym) {
        return (sym.flags() & PROTECTED) != 0 &&
                sym.owner.owner != currentClass.owner &&
                sym.packge() != currentClass.packge() &&
                !currentClass.isSubClass(sym.owner);
    }

    /**
      * The class in which an access method for given symbol goes.
      *  @param sym        The access symbol
      *  @param protAccess Is access to a protected symbol in another
      *                    package?
      */
    ClassSymbol accessClass(Symbol sym, boolean protAccess) {
        if (protAccess) {
            ClassSymbol c = currentClass;
            while (!c.isSubClass(sym.owner)) {
                c = c.owner.enclClass();
            }
            return c;
        } else {
            return sym.owner.enclClass();
        }
    }

    /**
      * Ensure that identifier is accessible, return tree accessing the identifier.
      *  @param sym      The accessed symbol.
      *  @param tree     The tree referring to the symbol.
      *  @param enclOp   The closest enclosing operation node of tree,
      *                  null if tree is not a subtree of an operation.
      *  @param refSuper Is access via a (qualified) C.super?
      */
    Tree access(Symbol sym, Tree tree, Tree enclOp, boolean refSuper) {
        if (sym.kind == VAR && (sym.flags() & CAPTURED) != 0 &&
                sym.owner.enclClass() != currentClass) {
            make.at(tree.pos);
            Object cv = ((VarSymbol) sym).constValue;
            if (cv != null)
                return makeLit(sym.type, cv);
            sym = proxies.lookup(proxyName(sym.name)).sym;
            tree = access(make.at(tree.pos).Ident(sym));
        }
        Tree base = tree.tag == Tree.SELECT ? ((Select) tree).selected : null;
        switch (sym.kind) {
        case TYP:
            if (sym.owner.kind != PCK) {
                Name flatname = Convert.shortName(sym.flatName());
                while (base != null && TreeInfo.symbol(base).kind != PCK) {
                    base = base.tag == Tree.SELECT ? ((Select) base).selected : null;
                }
                if (tree.tag == Tree.IDENT) {
                    ((Ident) tree).name = flatname;
                } else if (base == null) {
                    tree = make.at(tree.pos).Ident(sym);
                    ((Ident) tree).name = flatname;
                } else {
                    ((Select) tree).selected = base;
                    ((Select) tree).name = flatname;
                }
            }
            break;

        case MTH:

        case VAR:
            if (sym.owner.kind == TYP) {
                boolean protAccess = refSuper || needsProtectedAccess(sym);
                boolean accReq = protAccess || needsPrivateAccess(sym);
                boolean baseReq = base == null && sym.owner != syms.predefClass &&
                        !currentClass.isSubClass(sym.owner);
                if (accReq || baseReq) {
                    make.at(tree.pos);
                    if (sym.kind == VAR) {
                        Object cv = ((VarSymbol) sym).constValue;
                        if (cv != null)
                            return makeLit(sym.type, cv);
                    }
                    if (accReq) {
                        List args = Tree.emptyList;
                        if ((sym.flags() & STATIC) == 0) {
                            if (base == null)
                                base = makeOwnerThis(tree.pos, sym);
                            args = args.prepend(base);
                            base = null;
                        }
                        Tree receiver = make.Select(base != null ? base :
                                make.QualIdent(sym.owner),
                                accessSymbol(sym, tree, enclOp, protAccess));
                        return make.App(receiver, args);
                    } else if (baseReq) {
                        return make.at(tree.pos).Select(
                                accessBase(tree.pos, sym), sym);
                    }
                }
            }

        }
        return tree;
    }

    /**
      * Ensure that identifier is accessible, return tree accessing the identifier.
      *  @param tree     The identifier tree.
      */
    Tree access(Tree tree) {
        Symbol sym = TreeInfo.symbol(tree);
        return sym == null ? tree : access(sym, tree, null, false);
    }

    /**
      * Return access constructor for a private constructor,
      *  or the constructor itself, if no access constructor is needed.
      *  @param constr    The private constructor.
      */
    Symbol accessConstructor(Symbol constr) {
        if (needsPrivateAccess(constr)) {
            ClassSymbol accOwner = accessClass(constr, false);
            MethodSymbol aconstr = (MethodSymbol) accessConstrs.get(constr);
            if (aconstr == null) {
                aconstr = new MethodSymbol(SYNTHETIC, Names.init,
                        new MethodType( constr.type.argtypes().append(
                        accessConstructorTag().type), constr.type.restype(),
                        constr.type.thrown()), accOwner);
                accOwner.members().enter(aconstr);
                accessConstrs.put(constr, aconstr);
                accessed.append(constr);
            }
            return aconstr;
        } else {
            return constr;
        }
    }

    /**
      * Return an anonymous class nested in this toplevel class.
      */
    ClassSymbol accessConstructorTag() {
        ClassSymbol topClass = currentClass.outermostClass();
        Name flatname = Name.fromString(topClass.fullName() + "$" + 1);
        ClassSymbol ctag = (ClassSymbol) chk.compiled.get(flatname);
        if (ctag == null)
            ctag = makeEmptyClass(STATIC | INTERFACE, topClass);
        return ctag;
    }

    /**
      * Add all required access methods for a private symbol to enclosing class.
      *  @param sym       The symbol.
      */
    void makeAccessible(Symbol sym) {
        ClassDef cdef = classDef(sym.owner.enclClass());
        Util.assert(cdef != null, "class def not found: " + sym + " in " + sym.owner);
        if (sym.name == Names.init) {
            cdef.defs = cdef.defs.prepend( accessConstructorDef(cdef.pos, sym,
                    (MethodSymbol) accessConstrs.get(sym)));
        } else {
            MethodSymbol[] accessors = (MethodSymbol[]) accessSyms.get(sym);
            for (int i = 0; i < NCODES; i++) {
                if (accessors[i] != null)
                    cdef.defs = cdef.defs.prepend(
                            accessDef(cdef.pos, sym, accessors[i], i));
            }
        }
    }

    /**
      * Construct definition of an access method.
      *  @param pos        The source code position of the definition.
      *  @param sym        The private or protected symbol.
      *  @param accessor   The access method for the symbol.
      *  @param acode      The access code.
      */
    Tree accessDef(int pos, Symbol vsym, MethodSymbol accessor, int acode) {
        currentClass = vsym.owner.enclClass();
        make.at(pos);
        MethodDef md = make.MethodDef(accessor, null);
        Symbol sym = (Symbol) actualSymbols.get(vsym);
        if (sym == null)
            sym = vsym;
        Tree ref;
        List args;
        if ((sym.flags() & STATIC) != 0) {
            ref = make.Ident(sym);
            args = make.Idents(md.params);
        } else {
            ref = make.Select(make.Ident((VarDef) md.params.head), sym);
            args = make.Idents(md.params.tail);
        }
        Tree stat;
        if (sym.kind == VAR) {
            int acode1 = acode - (acode & 1);
            Tree expr;
            switch (acode1) {
            case DEREFcode:
                expr = ref;
                break;

            case ASSIGNcode:
                expr = make.Assign(ref, (Tree) args.head);
                break;

            case PREINCcode:

            case POSTINCcode:

            case PREDECcode:

            case POSTDECcode:
                expr = makeOperation((acode1 - PREINCcode >> 1) + Tree.PREINC,
                        List.make(ref));
                break;

            default:
                expr = make.Assignop(treeTag(binaryAccessOperator(acode1)), ref,
                        (Tree) args.head);
                ((Assignop) expr).operator = binaryAccessOperator(acode1);

            }
            stat = make.Return(expr.setType(sym.type));
        } else {
            stat = make.Call(make.App(ref, args));
        }
        md.body = make.Block(0, List.make(stat));
        for (List l = md.params; l.nonEmpty(); l = l.tail)
            ((VarDef) l.head).vartype = access(((VarDef) l.head).vartype);
        md.restype = access(md.restype);
        for (List l = md.thrown; l.nonEmpty(); l = l.tail)
            l.head = access((Tree) l.head);
        return md;
    }

    /**
      * Construct definition of an access constructor.
      *  @param pos        The source code position of the definition.
      *  @param constr     The private constructor.
      *  @param accessor   The access method for the constructor.
      */
    Tree accessConstructorDef(int pos, Symbol constr, MethodSymbol accessor) {
        make.at(pos);
        MethodDef md = make.MethodDef(accessor, accessor.externalType(), null);
        Ident callee = make.Ident(Names._this);
        callee.sym = constr;
        callee.type = constr.type;
        md.body = make.Block(0,
                List.make( make.Call(
                make.App(callee, make.Idents(md.params.reverse().tail.reverse())))));
        return md;
    }

    /**
      * A scope containing all free variable proxies for currently translated
      *  class, as well as its this$n symbol (if needed).
      *  Proxy scopes are nested in the same way classes are.
      *  Inside a constructor, proxies and any this$n symbol are duplicated
      *  in an additional innermost scope, where they represent the constructor
      *  parameters.
      */
    Scope proxies;

    /**
     * A stack containing the this$n field of the currently translated
     *  classes (if needed) in innermost first order.
     *  Inside a constructor, proxies and any this$n symbol are duplicated
     *  in an additional innermost scope, where they represent the constructor
     *  parameters.
     */
    List outerThisStack;

    /**
     * The name of a free variable proxy.
     */
    static Name proxyName(Name name) {
        return Name.fromString("val$" + name);
    }

    /**
      * Proxy definitions for all free variables in given list, in reverse order.
      *  @param pos        The source code position of the definition.
      *  @param freevars   The free variables.
      *  @param owner      The class in which the definitions go.
      */
    List freevarDefs(int pos, List freevars, Symbol owner) {
        int flags = FINAL | SYNTHETIC;
        if (owner.kind == TYP)
            flags |= PRIVATE;
        List defs = VarDef.emptyList;
        for (List l = freevars; l.nonEmpty(); l = l.tail) {
            VarSymbol v = (VarSymbol) l.head;
            VarSymbol proxy =
                    new VarSymbol(flags, proxyName(v.name), v.erasure(), owner);
            proxies.enter(proxy);
            VarDef vd = make.at(pos).VarDef(proxy, null);
            vd.vartype = access(vd.vartype);
            defs = defs.prepend(vd);
        }
        return defs;
    }

    /**
      * The name of a this$n field
      *  @param target   The class referenced by the this$n field
      */
    Name outerThisName(Type target) {
        Type t = target.outer();
        int nestingLevel = 0;
        while (t.tag == CLASS) {
            t = t.outer();
            nestingLevel++;
        }
        return Name.fromString("this$" + nestingLevel);
    }

    /**
      * Definition for this$n field.
      *  @param pos        The source code position of the definition.
      *  @param owner      The class in which the definition goes.
      */
    VarDef outerThisDef(int pos, Symbol owner) {
        int flags = FINAL | SYNTHETIC;
        if (owner.kind == TYP)
            flags |= PRIVATE;
        Type target = owner.enclClass().type.outer().erasure();
        VarSymbol outerThis =
                new VarSymbol(flags, outerThisName(target), target, owner);
        outerThisStack = outerThisStack.prepend(outerThis);
        VarDef vd = make.at(pos).VarDef(outerThis, null);
        vd.vartype = access(vd.vartype);
        return vd;
    }

    /**
      * Return a list of trees that load the free variables in given list,
      *  in reverse order.
      *  @param pos          The source code position to be used for the trees.
      *  @param freevars     The list of free variables.
      */
    List loadFreevars(int pos, List freevars) {
        List args = Tree.emptyList;
        for (List l = freevars; l.nonEmpty(); l = l.tail)
            args = args.prepend(loadFreevar(pos, (VarSymbol) l.head));
        return args;
    }

    Tree loadFreevar(int pos, VarSymbol v) {
        return access(v, make.at(pos).Ident(v), null, false);
    }

    /**
      * Construct a tree simulating the expression <C.this>.
      *  @param pos           The source code position to be used for the tree.
      *  @param c             The qualifier class.
      *  @param preciseMatch  A flag that indicates whether `c' must be an
      *                       enclosing class, or just a superclass of an enclosing
      *                       class.
      */
    Tree makeThis(int pos, TypeSymbol c, boolean preciseMatch) {
        if (currentClass == c || !preciseMatch && currentClass.isSubClass(c)) {
            return make.at(pos).This(c.erasure());
        } else {
            return makeOuterThis(pos, c, preciseMatch);
        }
    }

    /**
      * Construct a tree that represents the outer instance
      *  <C.this>. Never pick the current `this'.
      *  @param pos           The source code position to be used for the tree.
      *  @param c             The qualifier class.
      *  @param preciseMatch  A flag that indicates whether `c' must be an
      *                       enclosing class, or just a superclass of an enclosing
      *                       class.
      */
    Tree makeOuterThis(int pos, TypeSymbol c, boolean preciseMatch) {
        List ots = outerThisStack;
        if (ots.isEmpty()) {
            log.error(pos, "no.encl.instance.of.type.in.scope", c.toJava());
            return make.Ident(syms.nullConst);
        }
        VarSymbol ot = (VarSymbol) ots.head;
        Tree tree = access(make.at(pos).Ident(ot));
        TypeSymbol otc = ot.type.tsym;
        while (!(otc == c || !preciseMatch && otc.isSubClass(c))) {
            do {
                ots = ots.tail;
                if (ots.isEmpty()) {
                    log.error(pos, "no.encl.instance.of.type.in.scope", c.toJava());
                    return tree;
                }
                ot = (VarSymbol) ots.head;
            } while (ot.owner != otc)
                ;
            tree = access(make.at(pos).Select(tree, ot));
            otc = ot.type.tsym;
        }
        return tree;
    }

    /**
      * Construct a tree that represents the closest outer instance
      *  <C.this> such that the given symbol is a member of C.
      *  @param pos           The source code position to be used for the tree.
      *  @param sym           The accessed symbol.
      */
    Tree makeOwnerThis(int pos, Symbol sym) {
        Symbol c = sym.owner;
        if (currentClass.isSubClass(c) && sym.isMemberOf(currentClass)) {
            return make.at(pos).This(c.erasure());
        } else {
            List ots = outerThisStack;
            if (ots.isEmpty()) {
                log.error(pos, "no.encl.instance.of.type.in.scope", c.toJava());
                return make.Ident(syms.nullConst);
            }
            VarSymbol ot = (VarSymbol) ots.head;
            Tree tree = access(make.at(pos).Ident(ot));
            TypeSymbol otc = ot.type.tsym;
            while (!(otc.isSubClass(c) && sym.isMemberOf(otc))) {
                do {
                    ots = ots.tail;
                    if (ots.isEmpty()) {
                        log.error(pos, "no.encl.instance.of.type.in.scope",
                                c.toJava());
                        return tree;
                    }
                    ot = (VarSymbol) ots.head;
                } while (ot.owner != otc)
                    ;
                tree = access(make.at(pos).Select(tree, ot));
                otc = ot.type.tsym;
            }
            return tree;
        }
    }

    /**
      * Return tree simulating the the assignment <this.name = name>, where
      *  name is the name of a free variable.
      */
    Tree initField(int pos, Name name) {
        Scope.Entry e = proxies.lookup(name);
        Symbol rhs = e.sym;
        if (Util.checks)
            Util.assert(rhs.owner.kind == MTH);
        Symbol lhs = e.next().sym;
        if (Util.checks)
            Util.assert(rhs.owner.owner == lhs.owner);
        make.at(pos);
        return make.Exec(
                make.Assign(make.Select(make.This(lhs.owner.erasure()), lhs),
                make.Ident(rhs)).setType(lhs.erasure()));
    }

    /**
      * Return tree simulating the the assignment <this.this$n = this$n>.
      */
    Tree initOuterThis(int pos) {
        VarSymbol rhs = (VarSymbol) outerThisStack.head;
        if (Util.checks)
            Util.assert(rhs.owner.kind == MTH);
        VarSymbol lhs = (VarSymbol) outerThisStack.tail.head;
        if (Util.checks)
            Util.assert(rhs.owner.owner == lhs.owner);
        make.at(pos);
        return make.Exec(
                make.Assign(make.Select(make.This(lhs.owner.erasure()), lhs),
                make.Ident(rhs)).setType(lhs.erasure()));
    }

    /**
      * Return symbol of current class that can contain a class$ method.
      *  If we are translating a true class, this is its class symbol.
      *  Otherwise, if we are translating an interface, create an
      *  anonymous class in the interface unless one exists and return
      *  its symbol.
      */
    private ClassSymbol currentTrueClassSym() {
        if ((currentClass.flags() & INTERFACE) == 0)
            return currentClass;
        Scope s = currentClass.members();
        for (Scope.Entry e = s.elems; e != null; e = e.sibling)
            if (e.sym.kind == TYP && e.sym.name == Names.empty &&
                    (e.sym.flags() & INTERFACE) == 0)
                return (ClassSymbol) e.sym;
        return makeEmptyClass(PUBLIC | STATIC, currentClass);
    }

    /**
      * Return symbol for "class$" method. If there is no method definition
      *  for class$, construct one as follows:
      *
      *    class class$(String x0) {
      *      try {
      *        return Class.forName(x0);
      *      } catch (ClassNotFoundException x1) {
      *        throw new NoClassDefFoundError(x1.getMessage());
      *      }
      *    }
      */
    private MethodSymbol classDollarSym() {
        ClassSymbol currentTrueClass = currentTrueClassSym();
        MethodSymbol classDollarSym =
                (MethodSymbol) currentTrueClass.members().lookup(
                Names.classDollar).sym;
        if (classDollarSym == null) {
            classDollarSym = new MethodSymbol(STATIC | SYNTHETIC, Names.classDollar,
                    new MethodType(Type.emptyList.prepend(syms.stringType),
                    syms.classType, ClassSymbol.emptyList), currentTrueClass);
            currentTrueClass.members().enter(classDollarSym);
            MethodDef md = make.MethodDef(classDollarSym, null);
            Symbol forNameSym =
                    rs.resolveInternalMethod(make.pos, attrEnv, syms.classType,
                    Names.forName, Type.emptyList,
                    Type.emptyList.prepend(syms.stringType));
            VarSymbol catchParam = new VarSymbol(0, make.paramName(1),
                    syms.classNotFoundExceptionType, classDollarSym);
            Symbol getMessageSym = rs.resolveInternalMethod(make.pos, attrEnv,
                    syms.classNotFoundExceptionType, Names.getMessage,
                    Type.emptyList, Type.emptyList);
            md.body = make.Block(0,
                    Tree.emptyList.prepend( make.Try( make.Call(
                    make.App(make.QualIdent(forNameSym),
                    List.make(make.Ident(((VarDef) md.params.head).sym)))),
                    List.make( make.Catch(make.VarDef(catchParam, null),
                    make.Throw( makeNewClass(syms.noClassDefFoundErrorType,
                    List.make(
                    make.App(make.Select(make.Ident(catchParam), getMessageSym),
                    Tree.emptyList)))))), null)));
            ClassDef currentClassDef = classDef(currentTrueClass);
            currentClassDef.defs = currentClassDef.defs.prepend(md);
        }
        return classDollarSym;
    }

    /**
      * The name of the variable to cache T.class values.
      *  @param sig      The signature of type T.
      */
    private Name cacheName(String sig) {
        StringBuffer buf = new StringBuffer();
        if (sig.startsWith("[")) {
            buf = buf.append("array");
            while (sig.startsWith("[")) {
                buf = buf.append("$");
                sig = sig.substring(1);
            }
            if (sig.startsWith("L")) {
                sig = sig.substring(0, sig.length() - 1);
            }
        } else {
            buf = buf.append("class$");
        }
        buf = buf.append(sig.replace('.', '$'));
        return Name.fromString(buf.toString());
    }

    /**
      * The variable symbol that caches T.class values.
      *  If none exists yet, create a definition.
      *  @param sig      The signature of type T.
      */
    private VarSymbol cacheSym(String sig) {
        ClassSymbol currentTrueClass = currentTrueClassSym();
        Name cname = cacheName(sig);
        VarSymbol cacheSym = (VarSymbol) currentTrueClass.members().lookup(cname).sym;
        if (cacheSym == null) {
            cacheSym = new VarSymbol(STATIC | SYNTHETIC, cname, syms.classType,
                    currentTrueClass);
            currentTrueClass.members().enter(cacheSym);
            VarDef cacheDef = make.VarDef(cacheSym, null);
            ClassDef currentClassDef = classDef(currentTrueClass);
            currentClassDef.defs = currentClassDef.defs.prepend(cacheDef);
        }
        return cacheSym;
    }

    /**
      * The tree simulating a T.class expression.
      *  @param clazz      The tree identifying type T.
      */
    private Tree classOf(Tree clazz) {
        switch (clazz.type.tag) {
        case BYTE:

        case SHORT:

        case CHAR:

        case INT:

        case LONG:

        case FLOAT:

        case DOUBLE:

        case BOOLEAN:

        case VOID:
            Name bname = Type.boxedName[clazz.type.tag];
            ClassSymbol c = syms.reader.enterClass(bname);
            Symbol typeSym = rs.access(
                    rs.findIdentInType(attrEnv, c.type, Names.TYPE, VAR),
                    clazz.pos, c.type, Names.TYPE);
            if (typeSym.kind == VAR)
                attr.evalInit((VarSymbol) typeSym);
            return make.QualIdent(typeSym);

        case CLASS:

        case ARRAY:
            String sig =
                    syms.writer.xClassName(clazz.type).toString().replace('/', '.');
            Symbol cs = cacheSym(sig);
            return make.Conditional(Tree.CONDEXPR,
                    makeOperation(Tree.EQ,
                    List.make(make.Ident(cs), make.Ident(syms.nullConst))),
                    make.Assign(make.Ident(cs),
                    make.App(make.Ident(classDollarSym()),
                    List.make(
                    make.Literal(Type.CLASS, sig).setType(syms.stringType)))).
                    setType(syms.classType), make.Ident(cs)).setType(syms.classType);

        default:
            throw new InternalError();

        }
    }

    /**
      * Visitor argument: enclosing operator node.
      */
    private Tree enclOp;

    /**
     * Visitor method: Translate tree.
     */
    public Tree translate(Tree tree, Tree enclOp) {
        Tree prevEnclOp = this.enclOp;
        this.enclOp = enclOp;
        Tree res = translate(tree);
        this.enclOp = prevEnclOp;
        return res;
    }

    /**
      * Visitor method: Translate list of trees.
      */
    public List translate(List trees, Tree enclOp) {
        Tree prevEnclOp = this.enclOp;
        this.enclOp = enclOp;
        List res = translate(trees);
        this.enclOp = prevEnclOp;
        return res;
    }

    /**
      * Visitor method for class definitions.
      */
    public void _case(ClassDef tree) {
        ClassSymbol currentClassPrev = currentClass;
        currentClass = tree.sym;
        classdefs.put(currentClass, tree);
        proxies = proxies.dup();
        List prevOuterThisStack = outerThisStack;
        VarDef otdef = null;
        if (currentClass.hasOuterInstance())
            otdef = outerThisDef(tree.pos, currentClass);
        List fvdefs = freevarDefs(tree.pos, freevars(currentClass), currentClass);
        tree.extending = translate(tree.extending);
        tree.implementing = translate(tree.implementing);
        List seen = Tree.emptyList;
        while (tree.defs != seen) {
            List unseen = tree.defs;
            for (List l = unseen; l.nonEmpty() && l != seen; l = l.tail) {
                Tree outermostMemberDefPrev = outermostMemberDef;
                if (outermostMemberDefPrev == null)
                    outermostMemberDef = (Tree) l.head;
                l.head = translate((Tree) l.head);
                outermostMemberDef = outermostMemberDefPrev;
            }
            seen = unseen;
        }
        if ((tree.flags & PROTECTED) != 0)
            tree.flags |= PUBLIC;
        tree.flags &= ClassFlags;
        tree.name = Convert.shortName(currentClass.flatName());
        for (List l = fvdefs; l.nonEmpty(); l = l.tail) {
            tree.defs = tree.defs.prepend(l.head);
            currentClass.members().enter(((VarDef) l.head).sym);
        }
        if (currentClass.hasOuterInstance()) {
            tree.defs = tree.defs.prepend(otdef);
            currentClass.members().enter(otdef.sym);
        }
        proxies = proxies.leave();
        outerThisStack = prevOuterThisStack;
        translated.append(tree);
        currentClass = currentClassPrev;
        result = make.at(tree.pos).Block(0, Tree.emptyList);
    }

    /**
      * Visitor method for method definitions.
      */
    public void _case(MethodDef tree) {
        if (tree.name == Names.init && (currentClass.isInner() ||
                (currentClass.owner.kind & (VAR | MTH)) != 0)) {
            MethodSymbol m = tree.sym;
            proxies = proxies.dup();
            List prevOuterThisStack = outerThisStack;
            List fvs = freevars(currentClass);
            VarDef otdef = null;
            if (currentClass.hasOuterInstance())
                otdef = outerThisDef(tree.pos, m);
            List fvdefs = freevarDefs(tree.pos, fvs, m);
            super._case(tree);
            tree.params = tree.params.prepend(fvdefs);
            if (currentClass.hasOuterInstance())
                tree.params = tree.params.prepend(otdef);
            Tree selfCall = (Tree) tree.body.stats.head;
            List rest = tree.body.stats.tail;
            if (fvs.nonEmpty()) {
                List argtypes = m.erasure().argtypes();
                for (List l = fvs; l.nonEmpty(); l = l.tail) {
                    if (TreeInfo.isInitialConstructor(tree))
                        rest = rest.prepend( initField(tree.body.pos,
                                proxyName(((VarSymbol) l.head).name)));
                    argtypes = argtypes.prepend(((VarSymbol) l.head).erasure());
                }
                m.erasure_field = new MethodType(argtypes, m.erasure().restype(),
                        m.erasure().thrown());
            }
            if (currentClass.hasOuterInstance() &&
                    TreeInfo.isInitialConstructor(tree)) {
                rest = rest.prepend(initOuterThis(tree.body.pos));
            }
            tree.body.stats = rest.prepend(selfCall);
            proxies = proxies.leave();
            outerThisStack = prevOuterThisStack;
        } else {
            super._case(tree);
        }
        result = tree;
    }

    /**
      * Visitor method for class instance creation expressions.
      */
    public void _case(NewClass tree) {
        ClassSymbol c = (ClassSymbol) tree.constructor.owner;
        tree.args = translate(tree.args);
        if ((c.owner.kind & (VAR | MTH)) != 0) {
            tree.args = tree.args.prepend(loadFreevars(tree.pos, freevars(c)));
        }
        Symbol constructor = accessConstructor(tree.constructor);
        if (constructor != tree.constructor) {
            tree.args = tree.args.append(make.Ident(syms.nullConst));
            tree.constructor = constructor;
        }
        if (c.hasOuterInstance()) {
            Tree thisArg;
            if (tree.encl != null) {
                thisArg = translate(tree.encl);
            } else {
                thisArg = makeThis(tree.pos, c.type.outer().tsym, false);
            }
            tree.args = tree.args.prepend(thisArg);
        }
        tree.encl = null;
        if (tree.def != null) {
            translate(tree.def);
            tree.clazz = access(make.at(tree.clazz.pos).Ident(tree.def.sym));
            tree.def = null;
        } else {
            tree.clazz = access(c, tree.clazz, enclOp, false);
        }
        result = tree;
    }

    /**
      * Visitor method for method invocation expressions.
      */
    public void _case(Apply tree) {
        tree.args = translate(tree.args);
        Name methName = TreeInfo.name(tree.meth);
        if (methName == Names._this || methName == Names._super) {
            Symbol constructor = accessConstructor(TreeInfo.symbol(tree.meth));
            if (constructor != TreeInfo.symbol(tree.meth)) {
                tree.args = tree.args.append(make.Ident(syms.nullConst));
                TreeInfo.setSymbol(tree.meth, constructor);
            }
            ClassSymbol c = (ClassSymbol) constructor.owner;
            if ((c.owner.kind & (VAR | MTH)) != 0) {
                tree.args = tree.args.prepend(loadFreevars(tree.pos, freevars(c)));
            }
            if (c.hasOuterInstance()) {
                Tree thisArg;
                if (tree.meth.tag == Tree.SELECT) {
                    thisArg = translate(((Select) tree.meth).selected);
                    tree.meth = make.Ident(constructor);
                    ((Ident) tree.meth).name = methName;
                } else {
                    thisArg = makeOuterThis(tree.pos, c.type.outer().tsym, false);
                }
                tree.args = tree.args.prepend(thisArg);
            }
        } else {
            tree.meth = translate(tree.meth);
            if (tree.meth.tag == Tree.APPLY) {
                Apply app = (Apply) tree.meth;
                app.args = tree.args.prepend(app.args);
                result = app;
                return;
            }
        }
        result = tree;
    }

    /**
      * Visitor method for assignment expressions.
      */
    public void _case(Assign tree) {
        tree.lhs = translate(tree.lhs, tree);
        tree.rhs = translate(tree.rhs);
        if (tree.lhs.tag == Tree.APPLY) {
            Apply app = (Apply) tree.lhs;
            app.args = List.make(tree.rhs).prepend(app.args);
            result = app;
        } else {
            result = tree;
        }
    }

    /**
      * Visitor method for assignment operation expressions.
      */
    public void _case(Assignop tree) {
        tree.lhs = translate(tree.lhs, tree);
        tree.rhs = translate(tree.rhs);
        if (tree.lhs.tag == Tree.APPLY) {
            Apply app = (Apply) tree.lhs;
            Tree rhs = ((OperatorSymbol) tree.operator).opcode == string_add ?
                    makeString(tree.rhs) : tree.rhs;
            app.args = List.make(rhs).prepend(app.args);
            result = app;
        } else {
            result = tree;
        }
    }

    /**
      * Visitor method for operations.
      */
    public void _case(Operation tree) {
        tree.args = translate(tree.args, tree);
        if (Tree.PREINC <= tree.tag && tree.tag <= Tree.POSTDEC &&
                ((Tree) tree.args.head).tag == Tree.APPLY) {
            result = (Tree) tree.args.head;
        } else {
            result = tree;
        }
    }

    /**
      * Visitor method for simple identifiers.
      */
    public void _case(Ident tree) {
        result = access(tree.sym, tree, enclOp, false);
    }

    /**
      * Visitor method for selections.
      */
    public void _case(Select tree) {
        boolean qualifiedSuperAccess = tree.selected.tag == Tree.SELECT &&
                TreeInfo.name(tree.selected) == Names._super;
        tree.selected = translate(tree.selected);
        if (tree.name == Names._class)
            result = classOf(tree.selected);
        else if (tree.name == Names._this || tree.name == Names._super) {
            result = makeThis(tree.pos, tree.selected.type.tsym, true);
        } else
            result = access(tree.sym, tree, enclOp, qualifiedSuperAccess);
    }

    /**
      * Translate a toplevel class and return a list consisting of
      *  the translated class and translated versions of all inner classes.
      *  @param env   The attribution environment current at the class definition.
      *               We need this for resolving some additional symbols.
      *  @param cdef  The tree representing the class definition.
      */
    public List translateTopLevelClass(Env env, Tree cdef) {
        attrEnv = env;
        currentClass = null;
        outermostClassDef = cdef;
        outermostMemberDef = null;
        translated = new ListBuffer();
        classdefs = Hashtable.make();
        actualSymbols = Hashtable.make();
        freevarCache = Hashtable.make();
        proxies = new Scope(null);
        outerThisStack = VarSymbol.emptyList;
        accessNums = Hashtable.make();
        accessSyms = Hashtable.make();
        accessConstrs = Hashtable.make();
        accessed = new ListBuffer();
        translate(cdef, null);
        for (List l = accessed.toList(); l.nonEmpty(); l = l.tail)
            makeAccessible((Symbol) l.head);
        return translated.toList();
    }
}
