package com.sun.tools.javac.v8.comp;
import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.tree.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.tree.Tree.*;

import com.sun.tools.javac.v8.code.Type.*;


/**
 * This pass translates Generic Java to conventional Java.
 */
public class TransTypes extends TreeTranslator implements Flags, Kinds, TypeTags {
    Log log;
    Symtab syms;
    TreeMaker make;

    public TransTypes(Log log, Symtab syms, TreeMaker make) {
        super();
        this.log = log;
        this.syms = syms;
        this.make = make;
    }

    /**
      * A hashtable mapping bridge methods to the methods they override after
      *  type erasure.
      */
    Hashtable overridden = Hashtable.make();

    /**
     * Construct an attributed tree for a cast of expression to target type,
     *  unless it already has precisely that type.
     *  @param tree    The expression tree.
     *  @param target  The target type.
     */
    Tree cast(Tree tree, Type target) {
        int oldpos = make.pos;
        make.at(tree.pos);
        if (!tree.type.isSameType(target)) {
            tree = make.TypeCast(make.Type(target), tree).setType(target);
        }
        make.pos = oldpos;
        return tree;
    }

    /**
      * Construct an attributed tree to coerce an expression to some erased
      *  target type, unless the expression is already assignable to that type.
      *  If target type is a constant type, use its base type instead.
      *  @param tree    The expression tree.
      *  @param target  The target type.
      */
    Tree coerce(Tree tree, Type target) {
        Type btarget = target.baseType();
        return tree.type.isAssignable(btarget) ? tree : cast(tree, btarget);
    }

    /**
      * Given an erased reference type, assume this type as the tree's type.
      *  Then, coerce to some given target type unless target type is null.
      *  This operation is used in situations like the following:
      *
      *  class Cell<A> { A value; }
      *  ...
      *  Cell<Integer> cell;
      *  Integer x = cell.value;
      *
      *  Since the erasure of Cell.value is Object, but the type
      *  of cell.value in the assignment is Integer, we need to
      *  adjust the original type of cell.value to Object, and insert
      *  a cast to Integer. That is, the last assignment becomes:
      *
      *  Integer x = (Integer)cell.value;
      *
      *  @param tree       The expression tree whose type might need adjustment.
      *  @param erasedType The expression's type after erasure.
      *  @param target     The target type, which is usually the erasure of the
      *                    expression's original type.
      */
    Tree retype(Tree tree, Type erasedType, Type target) {
        if (erasedType.tag > lastBaseTag) {
            tree.type = erasedType;
            if (target != null)
                return coerce(tree, target);
        }
        return tree;
    }

    /**
      * Translate method argument list, casting each argument
      *  to its corresponding type in a list of target types.
      *  @param trees     The method argument list.
      *  @param targets   The list of target types.
      */
    List translateArgs(List trees, List targets) {
        for (List l = trees; l.nonEmpty(); l = l.tail) {
            l.head = translate((Tree) l.head, (Type) targets.head);
            targets = targets.tail;
        }
        return trees;
    }

    /**
      * Does method override some other symbol after all types are erased?
      *  @param meth     The method symbol.
      *  @param other    The other symbol.
      */
    boolean erasedOverrides(Symbol meth, Symbol other) {
        Type mt = meth.erasure();
        Type ot = other.erasure();
        return meth.name != Names.init && other.kind == MTH &&
                other.isMemberOf(meth.owner) && mt.isSameType(ot);
    }

    /**
      * The implementation of a given method in a given class, after all types
      *  are erased, null if method does not have an implementation in class.
      *  @param meth     The method symbol.
      *  @param origin   The class of which the implementation is a member.
      */
    public MethodSymbol erasedImplementation(Symbol meth, ClassSymbol origin) {
        for (TypeSymbol c = origin; c != null; c = c.type.supertype().tsym) {
            for (Scope.Entry e = c.members().lookup(meth.name); e.scope != null;
                    e = e.next()) {
                if (e.sym.kind == MTH && erasedOverrides(e.sym, meth))
                    return (MethodSymbol) e.sym;
            }
        }
        return null;
    }

    /**
      * Visitor argument: proto-type.
      */
    private Type pt;

    /**
     * Visitor method: perform a type translation on tree.
     */
    public Tree translate(Tree tree, Type pt) {
        Type prevPt = this.pt;
        this.pt = pt;
        if (tree == null)
            result = null;
        else
            tree.visit(this);
        this.pt = prevPt;
        return result;
    }

    /**
      * Visitor method: perform a type translation on list of trees.
      */
    public List translate(List trees, Type pt) {
        Type prevPt = this.pt;
        this.pt = pt;
        List res = translate(trees);
        this.pt = prevPt;
        return res;
    }

    /**
      * Visitor method for class definitions.
      */
    public void _case(ClassDef tree) {
        tree.typarams = TypeParameter.emptyList;
        super._case(tree);
        make.at(tree.pos);
        tree.type = tree.type.erasure();
        result = tree;
    }

    /**
      * Visitor method for method definitions.
      */
    public void _case(MethodDef tree) {
        tree.restype = translate(tree.restype, null);
        tree.typarams = TypeParameter.emptyList;
        tree.params = translateVarDefs(tree.params);
        tree.thrown = translate(tree.thrown, null);
        tree.body = (Block) translate(tree.body, tree.sym.erasure().restype());
        tree.type = tree.type.erasure();
        result = tree;
        for (Scope.Entry e = tree.sym.owner.members().lookup(tree.name);
                e.sym != null; e = e.next()) {
            if (e.sym != tree.sym && e.sym.type.erasure().isSameType(tree.type)) {
                log.error(tree.pos, "name.clash.same.erasure", tree.sym.toJava(),
                        e.sym.toJava());
                return;
            }
        }
    }

    /**
      * Visitor method for variable definitions.
      */
    public void _case(VarDef tree) {
        tree.vartype = translate(tree.vartype, null);
        tree.init = translate(tree.init, tree.sym.erasure());
        tree.type = tree.type.erasure();
        result = tree;
    }

    /**
      * Visitor method for do-loops.
      */
    public void _case(DoLoop tree) {
        tree.body = translate(tree.body);
        tree.cond = translate(tree.cond, null);
        result = tree;
    }

    /**
      * Visitor method for while-loops.
      */
    public void _case(WhileLoop tree) {
        tree.cond = translate(tree.cond, null);
        tree.body = translate(tree.body);
        result = tree;
    }

    /**
      * Visitor method for for-loops.
      */
    public void _case(ForLoop tree) {
        tree.init = translate(tree.init, null);
        tree.cond = translate(tree.cond, null);
        tree.step = translate(tree.step, null);
        tree.body = translate(tree.body);
        result = tree;
    }

    /**
      * Visitor method for switch statements.
      */
    public void _case(Switch tree) {
        tree.selector = translate(tree.selector, null);
        tree.cases = translateCases(tree.cases);
        result = tree;
    }

    /**
      * Visitor method for switch cases.
      */
    public void _case(Case tree) {
        tree.pat = translate(tree.pat, null);
        tree.stats = translate(tree.stats);
        result = tree;
    }

    /**
      * Visitor method for synchronized statements.
      */
    public void _case(Synchronized tree) {
        tree.lock = translate(tree.lock, null);
        tree.body = translate(tree.body);
        result = tree;
    }

    /**
      * Visitor method for conditional statements and expressions.
      */
    public void _case(Conditional tree) {
        tree.cond = translate(tree.cond, null);
        tree.thenpart = translate(tree.thenpart);
        tree.elsepart = translate(tree.elsepart);
        result = tree;
    }

    /**
      * Visitor method for expression statements.
      */
    public void _case(Exec tree) {
        tree.expr = translate(tree.expr, null);
        result = tree;
    }

    /**
      * Visitor method for return statements.
      */
    public void _case(Return tree) {
        tree.expr = translate(tree.expr);
        result = tree;
    }

    /**
      * Visitor method for throw statements.
      */
    public void _case(Throw tree) {
        tree.expr = translate(tree.expr, tree.expr.type.erasure());
        result = tree;
    }

    /**
      * Visitor method for method invocation expressions.
      */
    public void _case(Apply tree) {
        tree.meth = translate(tree.meth, null);
        Type mt = TreeInfo.symbol(tree.meth).erasure();
        tree.args = translateArgs(tree.args, mt.argtypes());
        result = retype(tree, mt.restype(), pt);
    }

    /**
      * Visitor method for class instance creation expressions.
      */
    public void _case(NewClass tree) {
        if (tree.encl != null)
            tree.encl = translate(tree.encl, tree.encl.type.erasure());
        tree.clazz = translate(tree.clazz, null);
        tree.args = translateArgs(tree.args, tree.constructor.erasure().argtypes());
        tree.def = (ClassDef) translate(tree.def, null);
        tree.type = tree.type.erasure();
        result = tree;
    }

    /**
      * Visitor method for class array creation expressions.
      */
    public void _case(NewArray tree) {
        tree.elemtype = translate(tree.elemtype, null);
        tree.dims = translate(tree.dims, null);
        tree.elems = translate(tree.elems, tree.type.elemtype().erasure());
        tree.type = tree.type.erasure();
        result = tree;
    }

    /**
      * Visitor method for class assignment expressions.
      */
    public void _case(Assign tree) {
        tree.lhs = translate(tree.lhs, null);
        tree.rhs = translate(tree.rhs, tree.lhs.type.erasure());
        tree.type = tree.type.erasure();
        result = tree;
    }

    /**
      * Visitor method for assignment operation expressions.
      */
    public void _case(Assignop tree) {
        tree.lhs = translate(tree.lhs, null);
        tree.rhs = translate(tree.rhs, null);
        tree.type = tree.type.erasure();
        result = tree;
    }

    /**
      * Visitor method for operation expressions.
      */
    public void _case(Operation tree) {
        tree.args = translateArgs(tree.args, tree.operator.type.argtypes());
        result = tree;
    }

    /**
      * Visitor method for type cast expressions.
      */
    public void _case(TypeCast tree) {
        tree.clazz = translate(tree.clazz, null);
        tree.expr = translate(tree.expr, null);
        tree.type = tree.type.erasure();
        result = tree;
    }

    /**
      * Visitor method for type test expressions.
      */
    public void _case(TypeTest tree) {
        tree.expr = translate(tree.expr, null);
        tree.clazz = translate(tree.clazz, null);
        result = tree;
    }

    /**
      * Visitor method for indexed expressions.
      */
    public void _case(Indexed tree) {
        tree.indexed = translate(tree.indexed, tree.indexed.type.erasure());
        tree.index = translate(tree.index, null);
        result = retype(tree, tree.indexed.type.elemtype(), pt);
    }

    /**
      * Visitor method for simple identifiers.
      */
    public void _case(Ident tree) {
        Type et = tree.sym.erasure();
        if (tree.sym.kind == TYP && tree.sym.type.tag == TYPEVAR) {
            result = make.at(tree.pos).Type(et);
        } else if (tree.type.constValue != null) {
            result = tree;
        } else if (tree.sym.kind == VAR) {
            result = retype(tree, et, pt);
        } else {
            tree.type = tree.type.erasure();
            result = tree;
        }
    }

    /**
      * Visitor method for selections.
      */
    public void _case(Select tree) {
        tree.selected = translate(tree.selected, tree.selected.type.erasure());
        if (tree.type.constValue != null) {
            result = tree;
        } else if (tree.sym.kind == VAR) {
            result = retype(tree, tree.sym.erasure(), pt);
        } else {
            tree.type = tree.type.erasure();
            result = tree;
        }
    }

    /**
      * Visitor method for array types.
      */
    public void _case(TypeArray tree) {
        tree.elemtype = translate(tree.elemtype, null);
        tree.type = tree.type.erasure();
        result = tree;
    }

    /**
      * Translate a toplevel class definition.
      *  @param cdef    The definition to be translated.
      */
    public Tree translateTopLevelClass(Tree cdef) {
        return translate(cdef, null);
    }
}
