package com.sun.tools.javac.v8.tree;
import java.io.*;

import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.tree.Tree.*;


/**
 * Prints out a tree as an indented Java source program.
 */
public class Pretty extends Tree.Visitor implements Flags {

    public Pretty(PrintStream out) {
        super();
        this.out = out;
    }

    public Pretty() {
        this(System.out);
    }

    /**
      * The output stream on which trees are printed.
      */
    PrintStream out;

    /**
     * Indentation width (can be reassigned from outside).
     */
    public int width = 3;

    /**
     * The current left margin.
     */
    int lmargin = 0;

    /**
     * The enclosing class name.
     */
    Name enclClassName = Names.empty;

    /**
     * A hashtable mapping trees to their documentation comments
     *  (can be null)
     */
    Hashtable docComments = null;

    /**
     * Align code to be indented to left margin.
     */
    void align() {
        for (int i = 0; i < lmargin; i++)
            out.print(" ");
    }

    /**
      * Increase left margin by indentation width.
      */
    void indent() {
        lmargin = lmargin + width;
    }

    /**
      * Decrease left margin by indentation width.
      */
    void undent() {
        lmargin = lmargin - width;
    }

    /**
      * Enter a new precedence level. Emit a `(' if new precedence level
      *  is less than precedence level so far.
      *  @param contextPrec    The precedence level in force so far.
      *  @param ownPrec        The new precedence level.
      */
    void open(int contextPrec, int ownPrec) {
        if (ownPrec < contextPrec)
            out.print("(");
    }

    /**
      * Leave precedence level. Emit a `(' if inner precedence level
      *  is less than precedence level we revert to.
      *  @param contextPrec    The precedence level we revert to.
      *  @param ownPrec        The inner precedence level.
      */
    void close(int contextPrec, int ownPrec) {
        if (ownPrec < contextPrec)
            out.print(")");
    }

    /**
      * Print string, replacing all non-ascii character with unicode escapes.
      */
    public void print(String s) {
        out.print(Convert.escapeUnicode(s));
    }

    /**
      * Print new line.
      */
    public void println() {
        out.println();
    }

    /**
      * Visitor argument: the current precedence level.
      */
    int prec;

    /**
     * Visitor method: print expression tree.
     *  @param prec  The current precedence level.
     */
    public void printExpr(Tree tree, int prec) {
        if (tree == null) {
            print("/*missing*/");
        } else {
            int prevPrec = this.prec;
            this.prec = prec;
            tree.visit(this);
            this.prec = prevPrec;
        }
    }

    /**
      * Derived visitor method: print expression tree at minimum precedence level
      *  for expression.
      */
    public void printExpr(Tree tree) {
        printExpr(tree, TreeInfo.noPrec);
    }

    /**
      * Derived visitor method: print statement tree.
      */
    public void printStat(Tree tree) {
        printExpr(tree, TreeInfo.notExpression);
    }

    /**
      * Derived visitor method: print list of expression trees, separated by commas.
      */
    public void printExprs(List trees) {
        if (trees.nonEmpty()) {
            printExpr((Tree) trees.head);
            for (List l = trees.tail; l.nonEmpty(); l = l.tail) {
                print(", ");
                printExpr((Tree) l.head);
            }
        }
    }

    /**
      * Derived visitor method: print list of statements, each on a separate line.
      */
    public void printStats(List trees) {
        for (List l = trees; l.nonEmpty(); l = l.tail) {
            align();
            printStat((Tree) l.head);
            println();
        }
    }

    /**
      * Print a set of modifiers.
      */
    public void printFlags(int flags) {
        if ((flags & SYNTHETIC) != 0)
            print("/*synthetic*/ ");
        print(TreeInfo.flagNames(flags));
        if ((flags & StandardFlags) != 0)
            print(" ");
    }

    /**
      * Print documentation comment, if it exists
      *  @param tree    The tree for which a documentation comment should be printed.
      */
    public void printDocComment(Tree tree) {
        String dc = (String) docComments.get(tree);
        if (dc != null) {
            print("/**");
            println();
            int pos = 0;
            int endpos = lineEndPos(dc, pos);
            while (pos < dc.length()) {
                align();
                print(" *");
                if (pos < dc.length() && dc.charAt(pos) > ' ')
                    print(" ");
                print(dc.substring(pos, endpos));
                println();
                pos = endpos + 1;
                endpos = lineEndPos(dc, pos);
            }
            align();
            print(" */");
            println();
            align();
        }
    }

    static int lineEndPos(String s, int start) {
        int pos = s.indexOf('\n', start);
        if (pos < 0)
            pos = s.length();
        return pos;
    }

    /**
      * Print a block.
      */
    public void printBlock(List stats) {
        print("{");
        println();
        indent();
        printStats(stats);
        undent();
        align();
        print("}");
    }

    /**
      * Print unit consisting of package clause and import statements in toplevel,
      *  followed by class definition. if class definition == null,
      *  print all definitions in toplevel.
      *  @param tree     The toplevel tree
      *  @param cdef     The class definition, which is assumed to be part of the
      *                  toplevel tree.
      */
    public void printUnit(TopLevel tree, ClassDef cdef) {
        docComments = tree.docComments;
        printDocComment(tree);
        if (tree.pid != null) {
            print("package ");
            printExpr(tree.pid);
            print(";");
            println();
        }
        for (List l = tree.defs; l.nonEmpty() &&
                (cdef == null || ((Tree) l.head).tag == Tree.IMPORT); l = l.tail) {
            printStat((Tree) l.head);
            println();
        }
        if (cdef != null) {
            printStat(cdef);
            println();
        }
    }

    /**
      * Visitor methods
      */
    public void _case(TopLevel tree) {
        printUnit(tree, null);
    }

    public void _case(Import tree) {
        print("import ");
        printExpr(tree.qualid);
        print(";");
        println();
    }

    public void _case(ClassDef tree) {
        println();
        align();
        printDocComment(tree);
        printFlags(tree.flags & ~INTERFACE);
        Name enclClassNamePrev = enclClassName;
        enclClassName = tree.name;
        if ((tree.flags & INTERFACE) != 0) {
            print("interface " + tree.name);
            if (tree.implementing.nonEmpty()) {
                print(" extends ");
                printExprs(tree.implementing);
            }
        } else {
            print("class " + tree.name);
            if (tree.extending != null) {
                print(" extends ");
                printExpr(tree.extending);
            }
            if (tree.implementing.nonEmpty()) {
                print(" implements ");
                printExprs(tree.implementing);
            }
        }
        print(" ");
        printBlock(tree.defs);
        enclClassName = enclClassNamePrev;
    }

    public void _case(MethodDef tree) {
        if (tree.name != Names.init || enclClassName != null) {
            println();
            align();
            printDocComment(tree);
            printFlags(tree.flags);
            if (tree.name == Names.init) {
                print(enclClassName.toString());
            } else {
                printExpr(tree.restype);
                print(" " + tree.name);
            }
            print("(");
            printExprs(tree.params);
            print(")");
            if (tree.thrown.nonEmpty()) {
                print(" throws ");
                printExprs(tree.thrown);
            }
            if (tree.body != null) {
                print(" ");
                printStat(tree.body);
            } else {
                print(";");
            }
        }
    }

    public void _case(VarDef tree) {
        if (docComments.get(tree) != null) {
            println();
            align();
        }
        printDocComment(tree);
        printFlags(tree.flags);
        printExpr(tree.vartype);
        print(" " + tree.name);
        if (tree.init != null) {
            print(" = ");
            printExpr(tree.init);
        }
        if (prec == TreeInfo.notExpression)
            print(";");
    }

    public void _case(Block tree) {
        printFlags(tree.flags);
        printBlock(tree.stats);
    }

    public void _case(DoLoop tree) {
        print("do ");
        printStat(tree.body);
        align();
        print(" while (");
        printExpr(tree.cond);
        print(");");
    }

    public void _case(WhileLoop tree) {
        print("while (");
        printExpr(tree.cond);
        print(") ");
        printStat(tree.body);
    }

    public void _case(ForLoop tree) {
        print("for (");
        if (tree.init.nonEmpty()) {
            if (((Tree) tree.init.head).tag == Tree.VARDEF) {
                printExpr((Tree) tree.init.head);
                for (List l = tree.init.tail; l.nonEmpty(); l = l.tail) {
                    VarDef vdef = (VarDef) l.head;
                    print(", " + vdef.name + " = ");
                    printExpr(vdef.init);
                }
            } else {
                printExprs(tree.init);
            }
        }
        print("; ");
        if (tree.cond != null)
            printExpr(tree.cond);
        print("; ");
        printExprs(tree.step);
        print(") ");
        printStat(tree.body);
    }

    public void _case(Labelled tree) {
        print(tree.label + ": ");
        printStat(tree.body);
    }

    public void _case(Switch tree) {
        print("switch (");
        printExpr(tree.selector);
        print(") {\n");
        printStats(tree.cases);
        align();
        print("}");
    }

    public void _case(Case tree) {
        if (tree.pat == null) {
            print("default");
        } else {
            print("case ");
            printExpr(tree.pat);
        }
        print(": \n");
        indent();
        printStats(tree.stats);
        undent();
        align();
    }

    public void _case(Synchronized tree) {
        print("synchronized (");
        printExpr(tree.lock);
        print(") ");
        printStat(tree.body);
    }

    public void _case(Try tree) {
        print("try ");
        printBlock(Tree.emptyList.prepend(tree.body));
        print(" ");
        for (List l = tree.catchers; l.nonEmpty(); l = l.tail) {
            printStat((Tree) l.head);
        }
        if (tree.finalizer != null) {
            print(" finally ");
            printBlock(Tree.emptyList.prepend(tree.finalizer));
        }
    }

    public void _case(Catch tree) {
        print(" catch (");
        printExpr(tree.param);
        print(") ");
        printBlock(Tree.emptyList.prepend(tree.body));
    }

    public void _case(Conditional tree) {
        if (tree.tag == Tree.CONDSTAT) {
            print("if (");
            printExpr(tree.cond);
            print(") ");
            printStat(tree.thenpart);
            if (tree.elsepart != null) {
                print(" else ");
                printStat(tree.elsepart);
            }
        } else {
            open(prec, TreeInfo.condPrec);
            printExpr(tree.cond, TreeInfo.condPrec);
            print(" ? ");
            printExpr(tree.thenpart, TreeInfo.condPrec);
            print(" : ");
            printExpr(tree.elsepart, TreeInfo.condPrec);
            close(prec, TreeInfo.condPrec);
        }
    }

    public void _case(Exec tree) {
        printExpr(tree.expr);
        if (prec == TreeInfo.notExpression)
            print(";");
    }

    public void _case(Break tree) {
        print("break");
        if (tree.label != null)
            print(" " + tree.label);
        print(";");
    }

    public void _case(Continue tree) {
        print("continue");
        if (tree.label != null)
            print(" " + tree.label);
        print(";");
    }

    public void _case(Return tree) {
        print("return");
        if (tree.expr != null) {
            print(" ");
            printExpr(tree.expr);
        }
        print(";");
    }

    public void _case(Throw tree) {
        print("throw ");
        printExpr(tree.expr);
        print(";");
    }

    public void _case(Apply tree) {
        printExpr(tree.meth);
        print("(");
        printExprs(tree.args);
        print(")");
    }

    public void _case(NewClass tree) {
        if (tree.encl != null) {
            printExpr(tree.encl);
            print(".");
        }
        print("new ");
        printExpr(tree.clazz);
        print("(");
        printExprs(tree.args);
        print(")");
        if (tree.def != null) {
            Name enclClassNamePrev = enclClassName;
            enclClassName = null;
            printBlock(((ClassDef) tree.def).defs);
            enclClassName = enclClassNamePrev;
        }
    }

    public void _case(NewArray tree) {
        if (tree.elemtype != null) {
            print("new ");
            int n = 0;
            Tree elemtype = tree.elemtype;
            while (elemtype.tag == Tree.TYPEARRAY) {
                n++;
                elemtype = ((TypeArray) elemtype).elemtype;
            }
            printExpr(elemtype);
            for (List l = tree.dims; l.nonEmpty(); l = l.tail) {
                print("[");
                printExpr((Tree) l.head);
                print("]");
            }
            for (int i = 0; i < n; i++) {
                print("[]");
            }
            if (tree.elems != null) {
                print("[]");
            }
        }
        if (tree.elems != null) {
            print("{");
            printExprs(tree.elems);
            print("}");
        }
    }

    public void _case(Assign tree) {
        open(prec, TreeInfo.assignPrec);
        printExpr(tree.lhs, TreeInfo.assignPrec + 1);
        print(" = ");
        printExpr(tree.rhs, TreeInfo.assignPrec);
        close(prec, TreeInfo.assignPrec);
    }

    public void _case(Assignop tree) {
        open(prec, TreeInfo.assignopPrec);
        printExpr(tree.lhs, TreeInfo.assignopPrec + 1);
        print(" " + TreeInfo.operatorName(tree.tag - Tree.ASGOffset) + "= ");
        printExpr(tree.rhs, TreeInfo.assignopPrec);
        close(prec, TreeInfo.assignopPrec);
    }

    public void _case(Operation tree) {
        int ownprec = TreeInfo.opPrec(tree.tag);
        String opname = TreeInfo.operatorName(tree.tag).toString();
        open(prec, ownprec);
        if (tree.tag <= Tree.PREDEC) {
            print(opname);
            printExpr((Tree) tree.args.head, ownprec);
        } else if (tree.tag <= Tree.POSTDEC) {
            printExpr((Tree) tree.args.head, ownprec);
            print(opname);
        } else {
            printExpr((Tree) tree.args.head, ownprec);
            print(" " + opname + " ");
            printExpr((Tree) tree.args.tail.head, ownprec + 1);
        }
        close(prec, ownprec);
    }

    public void _case(TypeCast tree) {
        open(prec, TreeInfo.prefixPrec);
        print("(");
        printExpr(tree.clazz);
        print(")");
        printExpr(tree.expr, TreeInfo.prefixPrec);
        close(prec, TreeInfo.prefixPrec);
    }

    public void _case(TypeTest tree) {
        open(prec, TreeInfo.ordPrec);
        printExpr(tree.expr, TreeInfo.ordPrec);
        print(" instanceof ");
        printExpr(tree.clazz, TreeInfo.ordPrec + 1);
        close(prec, TreeInfo.ordPrec);
    }

    public void _case(Indexed tree) {
        printExpr(tree.indexed, TreeInfo.postfixPrec);
        print("[");
        printExpr(tree.index);
        print("]");
    }

    public void _case(Select tree) {
        printExpr(tree.selected, TreeInfo.postfixPrec);
        print("." + tree.name);
    }

    public void _case(Ident tree) {
        print(tree.name.toString());
    }

    public void _case(Literal tree) {
        switch (tree.typetag) {
        case Type.INT:
            print(tree.value.toString());
            break;

        case Type.LONG:
            print(tree.value.toString() + "L");
            break;

        case Type.FLOAT:
            print(tree.value.toString() + "F");
            break;

        case Type.DOUBLE:
            print(tree.value.toString());
            break;

        case Type.CHAR:
            print("\'" + Convert.quote(
                    String.valueOf((char)((Number) tree.value).intValue())) + "\'");
            break;

        case Type.CLASS:
            print("\"" + Convert.quote((String) tree.value) + "\"");
            break;

        default:
            print(tree.value.toString());

        }
    }

    public void _case(TypeIdent tree) {
        print(Type.typeOfTag[tree.typetag].tsym.name.toString());
    }

    public void _case(TypeArray tree) {
        printExpr(tree.elemtype);
        print("[]");
    }

    public void _case(Erroneous tree) {
        print("(ERROR)");
    }

    public void _case(Tree tree) {
        print("(UNKNOWN: " + tree + ")");
        println();
    }
}
