package com.sun.tools.javac.v8.tree;
import java.io.File;

import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.code.*;

import com.sun.tools.javac.v8.code.Type.*;

import com.sun.tools.javac.v8.code.Symbol.*;

import com.sun.tools.javac.v8.tree.Tree.*;


/**
 * Utility class containing inspector methods for trees.
 */
public class TreeInfo implements Flags {

    public TreeInfo() {
        super();
    }

    /**
      * Is tree a constructor declaration?
      */
    public static boolean isConstructor(Tree tree) {
        return tree.tag == Tree.METHODDEF && ((MethodDef) tree).name == Names.init;
    }

    /**
      * Is there a constructor declaration in the given list of trees?
      */
    public static boolean hasConstructors(List trees) {
        for (List l = trees; l.nonEmpty(); l = l.tail)
            if (isConstructor((Tree) l.head))
                return true;
        return false;
    }

    /**
      * Is statement an initializer for a synthetic field?
      */
    public static boolean isSyntheticInit(Tree stat) {
        if (stat.tag == Tree.EXEC) {
            Exec exec = (Exec) stat;
            if (exec.expr.tag == Tree.ASSIGN) {
                Assign assign = (Assign) exec.expr;
                if (assign.lhs.tag == Tree.SELECT) {
                    Select select = (Select) assign.lhs;
                    if ((select.sym.flags() & SYNTHETIC) != 0 &&
                            name(select.selected) == Names._this) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    /**
      * Is tree an expression statement which calls a method with given name?
      */
    public static boolean isMethodCall(Tree tree, Name name) {
        if (tree.tag == Tree.EXEC) {
            Exec exec = (Exec) tree;
            if (exec.expr.tag == Tree.APPLY) {
                Name mname = TreeInfo.name(((Apply) exec.expr).meth);
                if (mname == name)
                    return true;
            }
        }
        return false;
    }

    /**
      * Is this a call to this or super?
      */
    public static boolean isSelfCall(Tree tree) {
        return isMethodCall(tree, Names._this) || isMethodCall(tree, Names._super);
    }

    /**
      * Is this a constructor whose first statement is not of the form this(...)?
      */
    public static boolean isInitialConstructor(Tree tree) {
        if (tree.tag != Tree.METHODDEF) {
            return false;
        } else {
            MethodDef md = (MethodDef) tree;
            return (md.name == Names.init || md.name == Names.clinit) &&
                    (md.body == null || md.body.stats.isEmpty() ||
                    !isMethodCall((Tree) md.body.stats.head, Names._this));
        }
    }

    /**
      * The position of the first statement in a block, or the position of
      *  the block itself if it is empty.
      */
    public static int firstStatPos(Tree tree) {
        if (tree.tag == Tree.BLOCK && ((Block) tree).stats.nonEmpty())
            return ((Tree)((Block) tree).stats.head).pos;
        else
            return tree.pos;
    }

    /**
      * The end position of given tree, if it is a block with
      *  defined endpos, it's start position otherwise.
      */
    public static int endPos(Tree tree) {
        if (tree.tag == Tree.BLOCK && ((Block) tree).endpos != Position.NOPOS)
            return ((Block) tree).endpos;
        else
            return tree.pos;
    }

    /**
      * The position of the finalizer of given try/synchronized statement,
      *  or the position of the statement itself if it does not have a finalizer.
      */
    public static int finalizerPos(Tree tree) {
        if (tree.tag == Tree.TRY) {
            Try t = (Try) tree;
            if (t.finalizer != null)
                return firstStatPos(t.finalizer);
            else
                return endPos(t.body);
        } else if (tree.tag == Tree.SYNCHRONIZED) {
            return endPos(((Synchronized) tree).body);
        } else {
            return tree.pos;
        }
    }

    /**
      * Return the statement referenced by a label.
      *  If the label refers to a loop or switch, return that switch
      *  otherwise return the labelled statement itself
      */
    public static Tree referencedStatement(Labelled tree) {
        Tree t = tree;
        do
            t = ((Labelled) t).body;
        while (t.tag == Tree.LABELLED)
            ;
        switch (t.tag) {
        case Tree.DOLOOP:

        case Tree.WHILELOOP:

        case Tree.FORLOOP:

        case Tree.SWITCH:
            return t;

        default:
            return tree;

        }
    }

    /**
      * Return the types of a list of trees.
      */
    public static List types(List trees) {
        ListBuffer ts = new ListBuffer();
        for (List l = trees; l.nonEmpty(); l = l.tail)
            ts.append(((Tree) l.head).type);
        return ts.toList();
    }

    /**
      * If this tree is an identifier or a field or a parameterized type,
      *  return its name, otherwise return null.
      */
    public static Name name(Tree tree) {
        switch (tree.tag) {
        case Tree.IDENT:
            return ((Ident) tree).name;

        case Tree.SELECT:
            return ((Select) tree).name;

        default:
            return null;

        }
    }

    /**
      * If this tree is a qualified identifier, its return fully qualified name,
      *  otherwise return null.
      */
    public static Name fullName(Tree tree) {
        switch (tree.tag) {
        case Tree.IDENT:
            return ((Ident) tree).name;

        case Tree.SELECT:
            Name sname = fullName(((Select) tree).selected);
            return sname == null ? null : Name.fromString(sname + "." + name(tree));

        default:
            return null;

        }
    }

    /**
      * If this tree is an identifier or a field, return its symbol,
      *  otherwise return null.
      */
    public static Symbol symbol(Tree tree) {
        switch (tree.tag) {
        case Tree.IDENT:
            return ((Ident) tree).sym;

        case Tree.SELECT:
            return ((Select) tree).sym;

        default:
            return null;

        }
    }

    /**
      * If this tree is an identifier or a field, set its symbol, otherwise skip.
      */
    public static void setSymbol(Tree tree, Symbol sym) {
        switch (tree.tag) {
        case Tree.IDENT:
            ((Ident) tree).sym = sym;
            break;

        case Tree.SELECT:
            ((Select) tree).sym = sym;
            break;

        default:

        }
    }

    /**
      * If this tree is a declaration or a block, return its flags field,
      *  otherwise return 0.
      */
    public static int flags(Tree tree) {
        switch (tree.tag) {
        case Tree.VARDEF:
            return ((VarDef) tree).flags;

        case Tree.METHODDEF:
            return ((MethodDef) tree).flags;

        case Tree.CLASSDEF:
            return ((ClassDef) tree).flags;

        case Tree.BLOCK:
            return ((Block) tree).flags;

        default:
            return 0;

        }
    }

    /**
      * Return first (smallest) flag in `flags':
      *  pre: flags != 0
      */
    public static int firstFlag(int flags) {
        int flag = 1;
        while ((flag & StandardFlags) != 0 && (flag & flags) == 0)
            flag = flag << 1;
        return flag;
    }

    /**
      * Return flags as a string, separated by " ".
      */
    public static String flagNames(int flags) {
        StringBuffer sbuf = new StringBuffer();
        int i = 0;
        int f = flags & StandardFlags;
        while (f != 0) {
            if ((f & 1) != 0) {
                if (sbuf.length() != 0)
                    sbuf.append(" ");
                sbuf.append(flagName[i]);
            }
            f = f >> 1;
            i++;
        }
        return sbuf.toString();
    }
    private static final String[] flagName = {"public", "private", "protected", "static",
    "final", "synchronized", "volatile", "transient", "native", "interface",
    "abstract", "strictfp"};

    /**
     * Return name of operator with given tree tag.
     */
    public static Name operatorName(int tag) {
        return opname[tag - Tree.POS];
    }
    private static Name[] opname = new Name[Tree.MOD - Tree.POS + 1];
    static {
        opname[Tree.POS - Tree.POS] = Name.fromString("+");
        opname[Tree.NEG - Tree.POS] = Name.fromString("-");
        opname[Tree.NOT - Tree.POS] = Name.fromString("!");
        opname[Tree.COMPL - Tree.POS] = Name.fromString("~");
        opname[Tree.PREINC - Tree.POS] = Name.fromString("++");
        opname[Tree.PREDEC - Tree.POS] = Name.fromString("--");
        opname[Tree.POSTINC - Tree.POS] = Name.fromString("++");
        opname[Tree.POSTDEC - Tree.POS] = Name.fromString("--");
        opname[Tree.OR - Tree.POS] = Name.fromString("||");
        opname[Tree.AND - Tree.POS] = Name.fromString("&&");
        opname[Tree.EQ - Tree.POS] = Name.fromString("==");
        opname[Tree.NE - Tree.POS] = Name.fromString("!=");
        opname[Tree.LT - Tree.POS] = Name.fromString("<");
        opname[Tree.GT - Tree.POS] = Name.fromString(">");
        opname[Tree.LE - Tree.POS] = Name.fromString("<=");
        opname[Tree.GE - Tree.POS] = Name.fromString(">=");
        opname[Tree.BITOR - Tree.POS] = Name.fromString("|");
        opname[Tree.BITXOR - Tree.POS] = Name.fromString("^");
        opname[Tree.BITAND - Tree.POS] = Name.fromString("&");
        opname[Tree.SL - Tree.POS] = Name.fromString("<<");
        opname[Tree.SR - Tree.POS] = Name.fromString(">>");
        opname[Tree.USR - Tree.POS] = Name.fromString(">>>");
        opname[Tree.PLUS - Tree.POS] = Name.fromString("+");
        opname[Tree.MINUS - Tree.POS] = Name.fromString("-");
        opname[Tree.MUL - Tree.POS] = Name.fromString("*");
        opname[Tree.DIV - Tree.POS] = Name.fromString("/");
        opname[Tree.MOD - Tree.POS] = Name.fromString("%");
    }

    /**
      * Operator precedences values.
      */
    public static final int notExpression = -1;

    /**
     * Operator precedences values.
     */
    public static final int noPrec = 0;

    /**
     * Operator precedences values.
     */
    public static final int assignPrec = 1;

    /**
     * Operator precedences values.
     */
    public static final int assignopPrec = 2;

    /**
     * Operator precedences values.
     */
    public static final int condPrec = 3;

    /**
     * Operator precedences values.
     */
    public static final int orPrec = 4;

    /**
     * Operator precedences values.
     */
    public static final int andPrec = 5;

    /**
     * Operator precedences values.
     */
    public static final int bitorPrec = 6;

    /**
     * Operator precedences values.
     */
    public static final int bitxorPrec = 7;

    /**
     * Operator precedences values.
     */
    public static final int bitandPrec = 8;

    /**
     * Operator precedences values.
     */
    public static final int eqPrec = 9;

    /**
     * Operator precedences values.
     */
    public static final int ordPrec = 10;

    /**
     * Operator precedences values.
     */
    public static final int shiftPrec = 11;

    /**
     * Operator precedences values.
     */
    public static final int addPrec = 12;

    /**
     * Operator precedences values.
     */
    public static final int mulPrec = 13;

    /**
     * Operator precedences values.
     */
    public static final int prefixPrec = 14;

    /**
     * Operator precedences values.
     */
    public static final int postfixPrec = 15;

    /**
     * Operator precedences values.
     */
    public static final int precCount = 16;

    /**
     * Map operators to their precedence levels.
     */
    public static int opPrec(int op) {
        switch (op) {
        case Tree.POS:
            return prefixPrec;

        case Tree.NEG:
            return prefixPrec;

        case Tree.NOT:
            return prefixPrec;

        case Tree.COMPL:
            return prefixPrec;

        case Tree.PREINC:
            return prefixPrec;

        case Tree.PREDEC:
            return prefixPrec;

        case Tree.POSTINC:
            return postfixPrec;

        case Tree.POSTDEC:
            return postfixPrec;

        case Tree.ASSIGN:
            return assignPrec;

        case Tree.BITOR_ASG:

        case Tree.BITXOR_ASG:

        case Tree.BITAND_ASG:

        case Tree.SL_ASG:

        case Tree.SR_ASG:

        case Tree.USR_ASG:

        case Tree.PLUS_ASG:

        case Tree.MINUS_ASG:

        case Tree.MUL_ASG:

        case Tree.DIV_ASG:

        case Tree.MOD_ASG:
            return assignopPrec;

        case Tree.OR:
            return orPrec;

        case Tree.AND:
            return andPrec;

        case Tree.EQ:
            return eqPrec;

        case Tree.NE:
            return eqPrec;

        case Tree.LT:
            return ordPrec;

        case Tree.GT:
            return ordPrec;

        case Tree.LE:
            return ordPrec;

        case Tree.GE:
            return ordPrec;

        case Tree.BITOR:
            return bitorPrec;

        case Tree.BITXOR:
            return bitxorPrec;

        case Tree.BITAND:
            return bitandPrec;

        case Tree.SL:
            return shiftPrec;

        case Tree.SR:
            return shiftPrec;

        case Tree.USR:
            return shiftPrec;

        case Tree.PLUS:
            return addPrec;

        case Tree.MINUS:
            return addPrec;

        case Tree.MUL:
            return mulPrec;

        case Tree.DIV:
            return mulPrec;

        case Tree.MOD:
            return mulPrec;

        case Tree.TYPETEST:
            return ordPrec;

        default:
            throw new InternalError();

        }
    }
}
