package com.sun.tools.javac.v8.tree;
import com.sun.tools.javac.v8.util.*;

import com.sun.tools.javac.v8.tree.Tree.*;


/**
 * A subclass of Tree.Visitor, this class defines
 *  a general tree translator pattern. Translation proceeds recursively in
 *  left-to-right order down a tree, constructing translated nodes by
 *  overwriting existing ones. There is one visitor method in this class
 *  for every possible kind of tree node.  To obtain a specific
 *  translator, it suffices to override those visitor methods which
 *  do some interesting work. The translator class itself takes care of all
 *  navigational aspects.
 */
public class TreeTranslator extends Tree.Visitor {

    public TreeTranslator() {
        super();
    }

    /**
      * Visitor result field: a tree
      */
    public Tree result;

    /**
     * Visitor method: Translate a single node.
     */
    public Tree translate(Tree tree) {
        if (tree == null) {
            return null;
        } else {
            tree.visit(this);
            return result;
        }
    }

    /**
      * Visitor method: translate a list of nodes.
      */
    public List translate(List trees) {
        if (trees == null)
            return null;
        for (List l = trees; l.nonEmpty(); l = l.tail)
            l.head = translate((Tree) l.head);
        return trees;
    }

    /**
      * Visitor method: translate a list of variable definitions.
      */
    public List translateVarDefs(List trees) {
        for (List l = trees; l.nonEmpty(); l = l.tail)
            l.head = (VarDef) translate((Tree) l.head);
        return trees;
    }

    /**
      * Visitor method: translate a list of case parts of switch statements.
      */
    public List translateCases(List trees) {
        for (List l = trees; l.nonEmpty(); l = l.tail)
            l.head = (Case) translate((Tree) l.head);
        return trees;
    }

    /**
      * Visitor method: translate a list of catch clauses in try statements.
      */
    public List translateCatchers(List trees) {
        for (List l = trees; l.nonEmpty(); l = l.tail)
            l.head = (Catch) translate((Tree) l.head);
        return trees;
    }

    public void _case(TopLevel tree) {
        tree.pid = translate(tree.pid);
        tree.defs = translate(tree.defs);
        result = tree;
    }

    public void _case(Import tree) {
        tree.qualid = translate(tree.qualid);
        result = tree;
    }

    public void _case(ClassDef tree) {
        tree.extending = translate(tree.extending);
        tree.implementing = translate(tree.implementing);
        tree.defs = translate(tree.defs);
        result = tree;
    }

    public void _case(MethodDef tree) {
        tree.restype = translate(tree.restype);
        tree.params = translateVarDefs(tree.params);
        tree.thrown = translate(tree.thrown);
        tree.body = (Block) translate(tree.body);
        result = tree;
    }

    public void _case(VarDef tree) {
        tree.vartype = translate(tree.vartype);
        tree.init = translate(tree.init);
        result = tree;
    }

    public void _case(Block tree) {
        tree.stats = translate(tree.stats);
        result = tree;
    }

    public void _case(DoLoop tree) {
        tree.body = translate(tree.body);
        tree.cond = translate(tree.cond);
        result = tree;
    }

    public void _case(WhileLoop tree) {
        tree.cond = translate(tree.cond);
        tree.body = translate(tree.body);
        result = tree;
    }

    public void _case(ForLoop tree) {
        tree.init = translate(tree.init);
        tree.cond = translate(tree.cond);
        tree.step = translate(tree.step);
        tree.body = translate(tree.body);
        result = tree;
    }

    public void _case(Labelled tree) {
        tree.body = translate(tree.body);
        result = tree;
    }

    public void _case(Switch tree) {
        tree.selector = translate(tree.selector);
        tree.cases = translateCases(tree.cases);
        result = tree;
    }

    public void _case(Case tree) {
        tree.pat = translate(tree.pat);
        tree.stats = translate(tree.stats);
        result = tree;
    }

    public void _case(Synchronized tree) {
        tree.lock = translate(tree.lock);
        tree.body = translate(tree.body);
        result = tree;
    }

    public void _case(Try tree) {
        tree.body = translate(tree.body);
        tree.catchers = translateCatchers(tree.catchers);
        tree.finalizer = translate(tree.finalizer);
        result = tree;
    }

    public void _case(Catch tree) {
        tree.param = (VarDef) translate(tree.param);
        tree.body = translate(tree.body);
        result = tree;
    }

    public void _case(Conditional tree) {
        tree.cond = translate(tree.cond);
        tree.thenpart = translate(tree.thenpart);
        tree.elsepart = translate(tree.elsepart);
        result = tree;
    }

    public void _case(Exec tree) {
        tree.expr = translate(tree.expr);
        result = tree;
    }

    public void _case(Break tree) {
        result = tree;
    }

    public void _case(Continue tree) {
        result = tree;
    }

    public void _case(Return tree) {
        tree.expr = translate(tree.expr);
        result = tree;
    }

    public void _case(Throw tree) {
        tree.expr = translate(tree.expr);
        result = tree;
    }

    public void _case(Apply tree) {
        tree.meth = translate(tree.meth);
        tree.args = translate(tree.args);
        result = tree;
    }

    public void _case(NewClass tree) {
        tree.encl = translate(tree.encl);
        tree.clazz = translate(tree.clazz);
        tree.args = translate(tree.args);
        tree.def = (ClassDef) translate(tree.def);
        result = tree;
    }

    public void _case(NewArray tree) {
        tree.elemtype = translate(tree.elemtype);
        tree.dims = translate(tree.dims);
        tree.elems = translate(tree.elems);
        result = tree;
    }

    public void _case(Assign tree) {
        tree.lhs = translate(tree.lhs);
        tree.rhs = translate(tree.rhs);
        result = tree;
    }

    public void _case(Assignop tree) {
        tree.lhs = translate(tree.lhs);
        tree.rhs = translate(tree.rhs);
        result = tree;
    }

    public void _case(Operation tree) {
        tree.args = translate(tree.args);
        result = tree;
    }

    public void _case(TypeCast tree) {
        tree.clazz = translate(tree.clazz);
        tree.expr = translate(tree.expr);
        result = tree;
    }

    public void _case(TypeTest tree) {
        tree.expr = translate(tree.expr);
        tree.clazz = translate(tree.clazz);
        result = tree;
    }

    public void _case(Indexed tree) {
        tree.indexed = translate(tree.indexed);
        tree.index = translate(tree.index);
        result = tree;
    }

    public void _case(Select tree) {
        tree.selected = translate(tree.selected);
        result = tree;
    }

    public void _case(Ident tree) {
        result = tree;
    }

    public void _case(Literal tree) {
        result = tree;
    }

    public void _case(TypeIdent tree) {
        result = tree;
    }

    public void _case(TypeArray tree) {
        tree.elemtype = translate(tree.elemtype);
        result = tree;
    }

    public void _case(Erroneous tree) {
        result = tree;
    }

    public void _case(Tree tree) {
        throw new InternalError();
    }
}
