package com.sun.tools.javac.v8.util;

/**
 * A class for extensible, mutable bit sets.
 */
public class Bits {
    private static final int wordlen = 32;
    private static final int wordshift = 5;
    private static final int wordmask = wordlen - 1;
    private int[] bits;

    /**
     * Construct an initially empty set.
     */
    public Bits() {
        this(new int[1]);
    }

    /**
      * Construct a set consisting initially of given bit vector.
      */
    public Bits(int[] bits) {
        super();
        this.bits = bits;
    }

    /**
      * Construct a set consisting initially of given range.
      */
    public Bits(int start, int limit) {
        this();
        inclRange(start, limit);
    }

    private void sizeTo(int len) {
        if (bits.length < len) {
            int[] newbits = new int[len];
            System.arraycopy(bits, 0, newbits, 0, bits.length);
            bits = newbits;
        }
    }

    /**
      * This set = {}.
      */
    public void clear() {
        for (int i = 0; i < bits.length; i++)
            bits[i] = 0;
    }

    /**
      * Return a copy of this set.
      */
    public Bits dup() {
        int[] newbits = new int[bits.length];
        System.arraycopy(bits, 0, newbits, 0, bits.length);
        return new Bits(newbits);
    }

    /**
      * Include x in this set.
      */
    public void incl(int x) {
        if (Util.checks)
            Util.assert(x >= 0);
        sizeTo((x >>> wordshift) + 1);
        bits[x >>> wordshift] = bits[x >>> wordshift] | 1 << (x & wordmask);
    }

    /**
      * Include 0..x in this set.
      */
    public void inclRange(int start, int limit) {
        sizeTo((limit >>> wordshift) + 1);
        for (int x = start; x < limit; x++)
            bits[x >>> wordshift] = bits[x >>> wordshift] | 1 << (x & wordmask);
    }

    /**
      * Exclude x from this set.
      */
    public void excl(int x) {
        if (Util.checks)
            Util.assert(x >= 0);
        sizeTo((x >>> wordshift) + 1);
        bits[x >>> wordshift] = bits[x >>> wordshift] & ~(1 << (x & wordmask));
    }

    /**
      * Is x an element of this set?
      */
    public boolean isMember(int x) {
        return 0 <= x && x < bits.length << wordshift &&
                (bits[x >>> wordshift] & 1 << (x & wordmask)) != 0;
    }

    /**
      * this set = this set & xs.
      */
    public Bits andSet(Bits xs) {
        sizeTo(xs.bits.length);
        for (int i = 0; i < xs.bits.length; i++)
            bits[i] = bits[i] & xs.bits[i];
        return this;
    }

    /**
      * this set = this set | xs.
      */
    public Bits orSet(Bits xs) {
        sizeTo(xs.bits.length);
        for (int i = 0; i < xs.bits.length; i++)
            bits[i] = bits[i] | xs.bits[i];
        return this;
    }

    /**
      * this set = this set \ xs.
      */
    public Bits diffSet(Bits xs) {
        for (int i = 0; i < bits.length; i++) {
            if (i < xs.bits.length) {
                bits[i] = bits[i] & ~xs.bits[i];
            }
        }
        return this;
    }

    /**
      * a string representation of this set.
      */
    public String toString() {
        char[] digits = new char[bits.length * wordlen];
        for (int i = 0; i < bits.length * wordlen; i++)
            digits[i] = isMember(i) ? '1' : '0';
        return new String(digits);
    }
}
