package com.sun.tools.javac.v8.util;

/**
 * An abstraction for internal compiler strings. For efficiency reasons,
 *  GJC uses hashed strings that are stored in a common large buffer.
 *
 *  Names represent unique hashable strings. Two names are equal
 *  if their indices are equal. Utf8 representation is used
 *  for storing names internally.
 */
public class Name {

    public Name() {
        super();
    }

    /**
      * The index where the bytes of this name are stored in the global name
      *  buffer `names'.
      */
    public int index;

    /**
     * The number of bytes in this name.
     */
    public int len;

    /**
     * The next name occupying the same hash bucket.
     */
    Name next;

    /**
     * The size and mask of the hash table for names.
     */
    private static final int HASH_SIZE = 32768;
    private static final int HASH_MASK = 32767;

    /**
     * The initial size of the array holding all encountered names.
     */
    private static final int NAME_SIZE = 131072;

    /**
     * The hash table for names.
     */
    private static Name[] hashtable = new Name[HASH_SIZE];

    /**
     * The array holding all encountered names.
     */
    public static byte[] names = new byte[NAME_SIZE];

    /**
     * The number of filled bytes in `names'.
     */
    private static int nc = 0;

    /**
     * The hashcode of a name.
     */
    private static int hashValue(byte[] cs, int start, int len) {
        if (len > 0)
            return len * (41 * 41 * 41) + cs[start] * (41 * 41) +
                    cs[start + len - 1] * 41 + cs[start + (len >> 1)];
        else
            return 0;
    }

    /**
      * Is (the utf8 representation of) name equal to
      *  cs[start..start+len-1]?
      */
    private static boolean equals(int index, byte[] cs, int start, int len) {
        int i = 0;
        while (i < len && names[index + i] == cs[start + i])
            i++;
        return i == len;
    }

    /**
      * Create a name from the bytes in cs[start..start+len-1].
      *  Assume that bytes are in utf8 format.
      */
    public static Name fromUtf(byte[] cs, int start, int len) {
        int h = hashValue(cs, start, len) & HASH_MASK;
        Name n = hashtable[h];
        while (n != null && (n.len != len || !equals(n.index, cs, start, len)))
            n = n.next;
        if (n == null) {
            while (nc + len > names.length) {
                byte[] newnames = new byte[names.length * 2];
                System.arraycopy(names, 0, newnames, 0, names.length);
                names = newnames;
            }
            System.arraycopy(cs, start, names, nc, len);
            n = new Name();
            n.index = nc;
            n.len = len;
            n.next = hashtable[h];
            hashtable[h] = n;
            nc = nc + len;
            if (len == 0)
                nc++;
        }
        return n;
    }

    /**
      * Create a name from the bytes in array cs.
      *  Assume that bytes are in utf8 format.
      */
    public static Name fromUtf(byte[] cs) {
        return fromUtf(cs, 0, cs.length);
    }

    /**
      * Create a name from the characters in cs[start..start+len-1].
      */
    public static Name fromChars(char[] cs, int start, int len) {
        while (nc + len * 3 >= names.length) {
            byte[] newnames = new byte[names.length * 2];
            System.arraycopy(names, 0, newnames, 0, names.length);
            names = newnames;
        }
        int nbytes = Convert.chars2utf(cs, start, names, nc, len) - nc;
        int h = hashValue(names, nc, nbytes) & HASH_MASK;
        Name n = hashtable[h];
        while (n != null && (n.len != nbytes || !equals(n.index, names, nc, nbytes)))
            n = n.next;
        if (n == null) {
            n = new Name();
            n.index = nc;
            n.len = nbytes;
            n.next = hashtable[h];
            hashtable[h] = n;
            nc = nc + nbytes;
            if (nbytes == 0)
                nc++;
        }
        return n;
    }

    /**
      * Create a name from the characters in string s.
      */
    public static Name fromString(String s) {
        char[] cs = s.toCharArray();
        return fromChars(cs, 0, cs.length);
    }

    /**
      * Return the Utf8 representation of this name.
      */
    public byte[] toUtf() {
        byte[] bs = new byte[len];
        System.arraycopy(names, index, bs, 0, len);
        return bs;
    }

    /**
      * Return the string representation of this name.
      */
    public String toString() {
        return Convert.utf2string(names, index, len);
    }

    /**
      * Return the string representation of this name.  This method is provided
      *  for consistency in internationalization.
      */
    public String toJava() {
        return toString();
    }

    /**
      * Copy all bytes of this name to buffer cs, starting at start.
      */
    public void getBytes(byte[] cs, int start) {
        System.arraycopy(names, index, cs, start, len);
    }

    /**
      * Return the hash value of this name.
      */
    public int hashCode() {
        return index;
    }

    /**
      * Is this name equal to other?
      */
    public boolean equals(Object other) {
        if (other instanceof Name)
            return index == ((Name) other).index;
        else
            return false;
    }

    /**
      * Returns the length of this name.
      */
    public int length() {
        return len;
    }

    /**
      * Returns i'th byte of this name.
      */
    public byte byteAt(int i) {
        return names[index + i];
    }

    /**
      * Returns first occurrence of byte b in this name, len if not found.
      */
    public int indexOf(byte b) {
        int i = 0;
        while (i < len && names[index + i] != b)
            i++;
        return i;
    }

    /**
      * Returns last occurrence of byte b in this name, -1 if not found.
      */
    public int lastIndexOf(byte b) {
        int i = len - 1;
        while (i >= 0 && names[index + i] != b)
            i--;
        return i;
    }

    /**
      * Does this name start with prefix?
      */
    public boolean startsWith(Name prefix) {
        int i = 0;
        while (i < prefix.len && i < len &&
                names[index + i] == names[prefix.index + i])
            i++;
        return i == prefix.len;
    }

    /**
      * Does this name end with suffix?
      */
    public boolean endsWith(Name suffix) {
        int i = len - 1;
        int j = suffix.len - 1;
        while (j >= 0 && i >= 0 && names[index + i] == names[suffix.index + j]) {
            i--;
            j--;
        }
        return j < 0;
    }

    /**
      * Returns the sub-name starting at position start, up to and
      *  excluding position end.
      */
    public Name subName(int start, int end) {
        if (end < start)
            end = start;
        return fromUtf(names, index + start, end - start);
    }

    /**
      * Replace all `from' bytes in this name with `to' bytes.
      */
    public Name replace(byte from, byte to) {
        int i = 0;
        while (i < len) {
            if (names[index + i] == from) {
                byte[] bs = new byte[len];
                System.arraycopy(names, index, bs, 0, i);
                bs[i] = to;
                i++;
                while (i < len) {
                    byte b = names[index + i];
                    bs[i] = b == from ? to : b;
                    i++;
                }
                return fromUtf(bs, 0, len);
            }
            i++;
        }
        return this;
    }

    /**
      * Return the concatenation of this name and name `n'.
      */
    public Name append(Name n) {
        byte[] bs = new byte[len + n.len];
        getBytes(bs, 0);
        n.getBytes(bs, len);
        return fromUtf(bs, 0, bs.length);
    }

    /**
      * Return the concatenation of all names in the array `ns'.
      */
    public static Name concat(Name[] ns) {
        int len = 0;
        for (int i = 0; i < ns.length; i++)
            len = len + ns[i].len;
        byte[] bs = new byte[len];
        len = 0;
        for (int i = 0; i < ns.length; i++) {
            ns[i].getBytes(bs, len);
            len = len + ns[i].len;
        }
        return fromUtf(bs, 0, len);
    }
}
