/*
 * @(#)ObjectStreamClass.c	1.37 00/02/02
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

#include <stdlib.h>

#include "jni.h"
#include "jvm.h"
#include "jlong.h"
#include "jni_util.h"

#include "java_io_ObjectStreamClass.h"


/* cached field IDs */
static jfieldID osf_field_id;		/* ObjectStreamField.field */
static jfieldID osf_type_id;		/* ObjectStreamField.type */


/*
 * Class:     java_io_ObjectStreamClass
 * Method:    initNative
 * Signature: ()V
 * 
 * Obtain and cache needed field IDs.
 */
JNIEXPORT void JNICALL 
Java_java_io_ObjectStreamClass_initNative(JNIEnv *env, jclass this)
{
    jclass osf_class;
    
    osf_class = (*env)->FindClass(env, "java/io/ObjectStreamField");
    if (osf_class == NULL)	/* exception thrown */
	return;

    osf_field_id = (*env)->GetFieldID(env, osf_class, "field",
	    "Ljava/lang/reflect/Field;");
    if (osf_field_id == NULL)	/* exception thrown */
	return;
    
    osf_type_id = (*env)->GetFieldID(env, osf_class, "type", "C");
    if (osf_type_id == NULL)	/* exception thrown */
	return;
}


/*
 * Class:     java_io_ObjectStreamClass
 * Method:    getFieldIDs
 * Signature: ([Ljava/io/ObjectStreamField;[J[J)V
 * 
 * Get the field IDs associated with the given fields.
 */
JNIEXPORT void JNICALL 
Java_java_io_ObjectStreamClass_getFieldIDs(JNIEnv *env, 
					   jclass this, 
					   jobjectArray fields, 
					   jlongArray primFieldIDs, 
					   jlongArray objFieldIDs)
{
    jsize nfields, pi, oi, i;
    jlong *primIDs = NULL, *objIDs = NULL;

    /* get number of fields */
    if (fields == NULL) {
	JNU_ThrowNullPointerException(env, NULL);
	goto end;
    }
    nfields = (*env)->GetArrayLength(env, fields);
    
    /* loop through fields, fetching and storing field IDs */
    pi = oi = 0;
    for (i = 0; i < nfields; i++) {
	jobject field, rfield;
	jfieldID fid;
	jchar tcode;
	
	/* get field descriptor */
	field = (*env)->GetObjectArrayElement(env, fields, i);
	if (field == NULL) {
	    JNU_ThrowNullPointerException(env, NULL);
	    goto end;
	}
	
	/* get reflective field */
	rfield = (*env)->GetObjectField(env, field, osf_field_id);
	fid = (rfield != NULL) ? 
	    (*env)->FromReflectedField(env, rfield) : NULL;

	/* typecode indicates whether field has primitive or object type */
	tcode = (*env)->GetCharField(env, field, osf_type_id);
	switch (tcode) {
	    case 'Z':
	    case 'B':
	    case 'S':
	    case 'C':
	    case 'I':
	    case 'J':
	    case 'F':
	    case 'D':
		if (primIDs == NULL) {		/* get prim field ID array */
		    if (primFieldIDs == NULL) {
			JNU_ThrowNullPointerException(env, NULL);
			goto end;
		    }
		    primIDs = 
			(*env)->GetLongArrayElements(env, primFieldIDs, NULL);
		    if (primIDs == NULL)	/* exception thrown */
			goto end;
		}
		primIDs[pi++] = (jlong) fid;
		break;
		
	    case '[':
	    case 'L':
		if (objIDs == NULL) {		/* get obj field ID array */
		    if (objFieldIDs == NULL) {
			JNU_ThrowNullPointerException(env, NULL);
			goto end;
		    }
		    objIDs = 
			(*env)->GetLongArrayElements(env, objFieldIDs, NULL);
		    if (objIDs == NULL)		/* exception thrown */
			goto end;
		}
		objIDs[oi++] = (jlong) fid;
		break;
		
	    default:
		JNU_ThrowIllegalArgumentException(env, "illegal typecode");
		goto end;
	}
	
	/* free local refs before next loop around */
	(*env)->DeleteLocalRef(env, field);
	(*env)->DeleteLocalRef(env, rfield);
    }
    
end:
    /* cleanup */
    if (primIDs != NULL)
	(*env)->ReleaseLongArrayElements(env, primFieldIDs, primIDs, 0);
    if (objIDs != NULL)
	(*env)->ReleaseLongArrayElements(env, objFieldIDs, objIDs, 0);
}


/*
 * Class:     java_io_ObjectStreamClass
 * Method:    hasStaticInitializer
 * Signature: (Ljava/lang/Class;)Z
 *
 * If the method <clinit> ()V is defined true is returned.
 * Otherwise, false is returned.
 */
JNIEXPORT jboolean JNICALL
Java_java_io_ObjectStreamClass_hasStaticInitializer(JNIEnv *env, jclass this,
						    jclass clazz)
{
    jclass superclazz = NULL;
    jmethodID superclinit = NULL;

    jmethodID clinit = (*env)->GetStaticMethodID(env, clazz,
						 "<clinit>", "()V");
    if (clinit == NULL || (*env)->ExceptionOccurred(env)) {
	(*env)->ExceptionClear(env);
	return 0;
    }

    /* Ask the superclass the same question
     * If the answer is the same then the constructor is from a superclass.
     * If different, it's really defined on the subclass.
     */
    superclazz = (*env)->GetSuperclass(env, clazz);
    if ((*env)->ExceptionOccurred(env)) {
	return 0;
    }

    if (superclazz == NULL)
	return 1;

    superclinit = (*env)->GetStaticMethodID(env, superclazz,
					    "<clinit>", "()V");
    if ((*env)->ExceptionOccurred(env)) {
	(*env)->ExceptionClear(env);
	superclinit = NULL;
    }

    return (superclinit != clinit);
}
