/*
 * @(#)GlyphVector.cpp	1.35 00/03/09
 *
 * Copyright 1998-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include "GlyphVector.h"
#include "Strike.h"

// REMIND: speedup if we write a "hsDesc" which can live on the stack

static jclass   g_gvClass       = 0;
static jfieldID g_gvPositions   = 0;
static jfieldID g_gvGlyphs      = 0;

static jclass    g_shapeClass       = 0;
static jmethodID g_shapeMoveTo      = 0;
static jmethodID g_shapeLineTo      = 0;
static jmethodID g_shapeQuadTo      = 0;
static jmethodID g_shapeClosePath   = 0;
static jmethodID g_shapeCurveTo     = 0;

// The invisible glyph has no image, no outline, and no advance.
// This is done here rather than at a lower level in order to localize this
// change and keep it consistent across different font types and platforms.
// Use the value ffff instead of 1 (TrueType convention) in order to avoid
// taking over a glyphID that has another use (i.e. in type1 fonts).

#define INVISIBLE_GLYPH_ID 0xffff

static Boolean initGVIDs(JNIEnv *env, jobject gv) {
    if (!g_gvClass) {
        g_gvClass = env->GetObjectClass(gv);
        if (!g_gvClass) {
            JNU_ThrowClassNotFoundException(env, "No GlyphVector class");
            return false;
        }
    }
    if (g_gvPositions && g_gvGlyphs) {
        return true;
    }
    g_gvPositions   = env->GetFieldID(g_gvClass, "positions", "[F");
    g_gvGlyphs      = env->GetFieldID(g_gvClass, "glyphs", "[I");
    if (!g_gvPositions || !g_gvGlyphs) {
        JNU_ThrowNoSuchFieldException(env,
                                    "GlyphVector positions or glyphs");
        return false;
    }
    return true;
}

static Boolean initShapeIDs(JNIEnv *env, jobject shape) {
    if (!g_shapeClass) {
        g_shapeClass = env->GetObjectClass(shape);
        if (!g_shapeClass) {
            JNU_ThrowClassNotFoundException(env, "No Shape class");
            return false;
        }
    }
    if (g_shapeMoveTo && g_shapeLineTo && g_shapeQuadTo && g_shapeClosePath &&
        g_shapeCurveTo) 
    {
        return true;
    }
    g_shapeMoveTo       = env->GetMethodID(g_shapeClass, "moveTo", "(FF)V");
    g_shapeLineTo       = env->GetMethodID(g_shapeClass, "lineTo", "(FF)V");
    g_shapeQuadTo       = env->GetMethodID(g_shapeClass, "quadTo", "(FFFF)V");
    g_shapeCurveTo      = env->GetMethodID(g_shapeClass, "curveTo", 
                                           "(FFFFFF)V");
    g_shapeClosePath    = env->GetMethodID(g_shapeClass, "closePath", "()V");
    if (!g_shapeMoveTo || !g_shapeLineTo
            || !g_shapeQuadTo || !g_shapeCurveTo || !g_shapeClosePath)
    {
        JNU_ThrowNoSuchMethodException(env,
                                    "Shape methods missing");
        return false;
    }
    return true;
}

void GlyphVector::allocateGlyphs(void)
{
    if (fNumGlyphs <= eDefaultStorage) {
        fGlyphs = fBaseGlyphs;
    } else {
        fGlyphs = new UInt32[fNumGlyphs];
    }
}

void GlyphVector::allocatePositions(void)
{
    if (fNumGlyphs < eDefaultStorage) {
        fPositions = fBasePositions;
    } else {
        fPositions = new float[fNumGlyphs * 2 + 2];
    }
}

void GlyphVector::allocateImageRefs(void)
{
    if (fNumGlyphs <= eDefaultStorage) {
        fImageRefs = fBaseImageRefs;
    } else {
        fImageRefs = new ImageRef[fNumGlyphs];
    }
}

static inline int inRange(int c, int begin, int end) {
    return (c>=begin) && (c<=end);
}

void GlyphVector::initGlyphs(const jchar *theChars, FontTransform& tx)
{
    fontObject *myFO = (fontObject *)fFont;
    allocateGlyphs();
    Strike *theStrike;
    theStrike = &myFO->getStrike(tx, fIsAntiAliased, fUsesFractionalMetrics);

    if (fGlyphs) {
        for (unsigned int i=0; i<fNumGlyphs; i++) {
            jchar theChar = theChars[i];
            fGlyphs[i] = theChar == 0xffff ? INVISIBLE_GLYPH_ID : theStrike->CharToGlyph(theChar);
            if ( !(theChar & 0xfe00)) {
                // if roman assume no shaping, BIDI
                continue;
            }
            if (inRange(theChar, 0x0590, 0x05ff)) {
                // if Hebrew, assume shaping for presentation forms
                fNeedShaping = true;
                continue;
            }
            if (inRange(theChar, 0x0600, 0x06ff)) {
                // if Arabic, assume shaping for presentation forms
                fNeedShaping = true;
                continue;
            }
            if (inRange(theChar, 0x202a, 0x202e)) {
                // if directional control, assume shaping needed
                fNeedShaping = true;
                continue;
            }
            if (inRange(theChar, 0x206a, 0x206f)) {
                // if shaping control, assume shaping needed
                fNeedShaping = true;
                continue;
            }
        }
    }
}

GlyphVector::GlyphVector(JNIEnv *env, jstring theString,
                jdoubleArray matrix,
                jboolean isAntiAliased, jboolean usesFractionalMetrics,
                fontObject *fo)
    : fFont(fo),  fEnv(env), fNeedShaping(false),
      fNumGlyphs(0), fGlyphs(NULL), fPositions(NULL),
      fImageRefs(NULL), fMatrix(matrix),
      fIsAntiAliased(isAntiAliased),
      fUsesFractionalMetrics(usesFractionalMetrics)
{
    fDevTX[0] = 1;
    fDevTX[1] = 0;
    fDevTX[2] = 0;
    fDevTX[3] = 1;

    if (JNU_IsNull(fEnv, theString) ) {
        JNU_ThrowNullPointerException(fEnv,"");
        return;
    }

    fNumGlyphs = env->GetStringLength(theString);
    FontTransform tx(fEnv, fMatrix);
    const jchar *theChars = env->GetStringCritical(theString, NULL);
    if (theChars) {
      jchar *tempChars = new jchar [fNumGlyphs];
      memcpy(tempChars, theChars, sizeof(jchar)*fNumGlyphs);
      env->ReleaseStringCritical(theString, theChars);
      initGlyphs(tempChars, tx);
      delete tempChars;
    }
}

GlyphVector::GlyphVector(JNIEnv *env,
                         jcharArray theCharArray, jint offset, jint count,
                         jdoubleArray matrix,
                         jboolean isAntiAliased,
                         jboolean usesFractionalMetrics,
                         fontObject *fo)
    : fFont(fo), fEnv(env), fNeedShaping(false),
      fNumGlyphs(0), fGlyphs(NULL), fPositions(NULL),
      fImageRefs(NULL), fMatrix(matrix),
      fIsAntiAliased(isAntiAliased),
      fUsesFractionalMetrics(usesFractionalMetrics)
{
    fDevTX[0] = 1;
    fDevTX[1] = 0;
    fDevTX[2] = 0;
    fDevTX[3] = 1;

    if (JNU_IsNull(fEnv, theCharArray) ) {
        JNU_ThrowNullPointerException(fEnv,"");
        return;
    }

    if (offset + count > env->GetArrayLength(theCharArray)) {
        JNU_ThrowArrayIndexOutOfBoundsException(fEnv,
                                        "chars [offset + count]");
        return;
    }
    fNumGlyphs = count;
    FontTransform tx(fEnv, fMatrix);
    const jchar *theChars =
        (const jchar *)env->GetPrimitiveArrayCritical(theCharArray, NULL);
    if (theChars) {
#ifdef __linux__
//	this code is needed as memory is not correctly allocated on linux 
        jchar *tempChars = new jchar[fNumGlyphs];
        memcpy(tempChars, theChars + offset, sizeof(jchar) * fNumGlyphs);
#else
        initGlyphs( theChars + offset, tx);
#endif
        env->ReleasePrimitiveArrayCritical(
                    theCharArray, (void *)theChars, JNI_ABORT);
#ifdef __linux__
        initGlyphs(tempChars, tx);
        delete tempChars;
#endif
    }
}

GlyphVector::GlyphVector(JNIEnv *env,
                         jbyteArray theByteArray, jint offset, jint count,
                         jdoubleArray matrix,
                         jboolean isAntiAliased,
                         jboolean usesFractionalMetrics,
                         fontObject *fo)
    : fFont(fo), fEnv(env), fNeedShaping(false),
      fNumGlyphs(0), fGlyphs(NULL), fPositions(NULL),
      fImageRefs(NULL), fMatrix(matrix),
      fIsAntiAliased(isAntiAliased),
      fUsesFractionalMetrics(usesFractionalMetrics)
{
    fDevTX[0] = 1;
    fDevTX[1] = 0;
    fDevTX[2] = 0;
    fDevTX[3] = 1;

    if (JNU_IsNull(fEnv, theByteArray) ) {
        JNU_ThrowNullPointerException(fEnv,"");
        return;
    }

    if (offset + count > env->GetArrayLength(theByteArray)) {
        JNU_ThrowArrayIndexOutOfBoundsException(fEnv,
                                        "chars [offset + count]");
        return;
    }
    fNumGlyphs = count;
    Strike *theStrike;
    {
        FontTransform tx(fEnv, matrix);
        theStrike = &fo->getStrike(tx, isAntiAliased, usesFractionalMetrics);
    }
    const jbyte *theBytes =
        (const jbyte *)env->GetPrimitiveArrayCritical(theByteArray, NULL);
    if (theBytes) {
        allocateGlyphs();
        if (fGlyphs) {
            for (int i=0; i<count; i++) {
                fGlyphs[i] = theStrike->CharToGlyph(theBytes[i+offset]);
            }
        }
        env->ReleasePrimitiveArrayCritical(
                    theByteArray, (void *)theBytes, JNI_ABORT);
    }
}

GlyphVector::GlyphVector(JNIEnv *env, jintArray theGlyphArray,
                         jdoubleArray matrix,
                         jboolean isAntiAliased,
                         jboolean usesFractionalMetrics,
                         fontObject *fo)
    : fFont(fo), fEnv(env), fNeedShaping(false),
      fNumGlyphs(0), fGlyphs(NULL), fPositions(NULL),
      fImageRefs(NULL), fMatrix(matrix),
      fIsAntiAliased(isAntiAliased),
      fUsesFractionalMetrics(usesFractionalMetrics)
{
    fDevTX[0] = 1;
    fDevTX[1] = 0;
    fDevTX[2] = 0;
    fDevTX[3] = 1;

    if (JNU_IsNull(fEnv, theGlyphArray)) {
        JNU_ThrowNullPointerException(fEnv,"");
        return;
    }

    fNumGlyphs = env->GetArrayLength(theGlyphArray); 
    Strike *theStrike;
    {
        FontTransform tx(fEnv, matrix);
        theStrike = &fo->getStrike(tx, isAntiAliased, usesFractionalMetrics);
    }
    const jint *theGlyphs =
        (const jint *)env->GetPrimitiveArrayCritical(theGlyphArray, NULL);
    if (theGlyphs) {
        allocateGlyphs();
        int missingGlyph = theStrike->getMissingGlyphCode();
        if (fGlyphs) {
            for (unsigned int i=0; i<fNumGlyphs; i++) {
                fGlyphs[i] = theGlyphs[i];
            }
        }
        env->ReleasePrimitiveArrayCritical(
                    theGlyphArray, (void *)theGlyphs, JNI_ABORT);
    }
}

GlyphVector::GlyphVector(JNIEnv *env,
                         jdoubleArray matrix,
                         jboolean isAntiAliased,
                         jboolean usesFractionalMetrics,
                         fontObject *fo)
    : fFont(fo), fEnv(env), fNeedShaping(false),
      fNumGlyphs(0), fGlyphs(NULL), fPositions(NULL),
      fImageRefs(NULL), fMatrix(matrix),
      fIsAntiAliased(isAntiAliased),
      fUsesFractionalMetrics(usesFractionalMetrics)
{
    fDevTX[0] = 1;
    fDevTX[1] = 0;
    fDevTX[2] = 0;
    fDevTX[3] = 1;
}

GlyphVector::~GlyphVector() {
    if (fGlyphs && (fGlyphs != fBaseGlyphs)) {
        delete [] fGlyphs;
    }
    fGlyphs = NULL;
    if (fPositions && (fPositions != fBasePositions)) {
        delete [] fPositions;
    }
    fPositions = NULL;
    if (fImageRefs && (fImageRefs != fBaseImageRefs)) {
        delete [] fImageRefs;
    }
    fImageRefs = NULL;
}

Boolean GlyphVector::needShaping()
{
    return fNeedShaping;
}

void GlyphVector::setDevTX(jdoubleArray devTX) {
    FontTransform tx(fEnv, devTX);
    tx.getMatrixInto(fDevTX, 4);
}

void GlyphVector::setPositions(jfloatArray newPos)
{
//      return; // REMIND: workaround for 4143573
    if (!JNU_IsNull(fEnv, newPos)) {
        JFloatBuffer positionBuffer(fEnv, newPos);
        UInt32 numPositions = positionBuffer.getNumElements();
        const jfloat *thePositions = positionBuffer.buffer();
        if (thePositions) {
          if (numPositions == (fNumGlyphs * 2) || numPositions == (fNumGlyphs * 2 + 2)) {
                allocatePositions();
            } else {
                JNU_ThrowIllegalArgumentException(fEnv,
                                                "Wrong size positions array");
            }
            if (fPositions) {
                for (unsigned int i=0; i<numPositions; i++) {
                    fPositions[i] = thePositions[i];
                }
            }
        }
    }
}

void GlyphVector::positionGlyphsWithImages(
        jfloat x, jfloat y,
        jdoubleArray matrix,
        Boolean isAntiAliased, Boolean usesFractionalMetrics)
{
  /*
    here's the problem.
    when you draw a glyph vector, it uses positions from the glyphvector.  these
    are based on the advances of a font that doesn't include the graphics transform.
    if fractional metrics is off, they have been pushed to the next greatest integral
    value.  in this routine, the graphics transform is applied to each point to get
    the effective origin of the glyphs under the transform.  These values are not
    necessarily integral

    when you draw a character array, you start immediately with a font that does
    include the graphics transform.  if fractional metrics is off, the advances
    are again pushed to the next greatest integral value.  These advances are
    then summed to successively produce the positions of each glyph.

    there are two problems.
    1) 'integralization' is happening in different places with glyph vectors and 
    with strings.  In the one case, it happens before the graphics transform is
    applied, in the other case, it happens after.

    Especially with rotation, 'integralization' should never be applied until the
    moment of rendering, in particular it cannot be applied to delta values like
    advances.  This leads to compression or expansion of the string, or in the 
    case of rotation, 'fading off' of the baseline.

    2) I think it is not guaranteed that the advance of a rotated glyph is the same 
    as the advance of the unrotated glyph passed through the transform.  This is 
    because the advance can come from control points in the glyph.  In the unrotated
    glyph, these can be hinted, in the rotated glyph, either hinting is turned off,
    or has a different effect.  Thus the advances may be subtly different.
   */

    Boolean loadPositions = false;
    FontTransform tx(fEnv, matrix);
    Strike& theStrike = fFont->getStrike(tx, isAntiAliased, usesFractionalMetrics);
    // load positions if they are not already specified
    if (!fPositions) {
        allocatePositions();
        loadPositions = true;
    }
    if (!fImageRefs) {
        allocateImageRefs();
    }
    if (!fPositions || !fImageRefs) {
        JNU_ThrowInternalError(fEnv, "Insufficient Memory");
        return;
    }

    theStrike.prepareCache(fNumGlyphs);
    Boolean refcheck = false;
    unsigned int i;
    for (i=0; i < fNumGlyphs; i++) {
      if (fGlyphs[i] == INVISIBLE_GLYPH_ID) {
        fImageRefs[i].rowBytes = 0;
        fImageRefs[i].width = 0;
        fImageRefs[i].height = 0;
        fImageRefs[i].pixels = 0;
        if (loadPositions) {
          fPositions[i*2] = x;
          fPositions[i*2+1] = y;
        } else {
          float tempX = fPositions[i*2];
          float tempY = fPositions[i*2+1];
	  fPositions[i*2] = tempX * fDevTX[0] + tempY * fDevTX[2] + x;
          fPositions[i*2+1] = tempX * fDevTX[1] + tempY * fDevTX[3] + y;
        }
      } else {
        hsGGlyph glyphRef;
        hsFixedPoint2 advanceXY;
        theStrike.getMetricsWithImage(fGlyphs[i], glyphRef, advanceXY, &refcheck);
        fImageRefs[i].rowBytes  = glyphRef.fRowBytes;
        fImageRefs[i].width     = glyphRef.fWidth;
        fImageRefs[i].height    = glyphRef.fHeight;
        fImageRefs[i].pixels    = glyphRef.fImage;

	/*
	 * With the per-glyph cache, the image ptr in the glyphRef
	 * can go stale, so we always recheck it.  The test that used
	 * to be done here is now in getImage.
	 *
	 * To really fix this, the glyphref should be an object, and
	 * a method on the glyphref should be used to load or ref
	 * the image, rather than exposing the pixel pointer.  A lot
	 * of the headspin stuff should be reworked since some of it
	 * doesn't quite fit in this environment and it seems like
	 * it was slapped in a bit hastily.

        if (!glyphRef.fImage && glyphRef.fWidth && glyphRef.fHeight) {
            fImageRefs[i].pixels = theStrike.getImage(fGlyphs[i]);
        }
	*/

        jfloat fX = hsFixedToFloat(glyphRef.fTopLeft.fX);
        if (i == 0 && fX < 0) {
            float txMatrix[4];
            tx.getMatrixInto(txMatrix, 4);
            if (txMatrix[0] > 0 && txMatrix[1] == 0 && txMatrix[2] == 0) {
                /* horizontal left->right layout */
                x -= fX;
            }
        }

        if (loadPositions) {
            fPositions[i*2] = x + hsFixedToFloat(glyphRef.fTopLeft.fX);
            fPositions[i*2+1] = y + hsFixedToFloat(glyphRef.fTopLeft.fY);
            x += hsFixedToFloat(advanceXY.fX);
            y += hsFixedToFloat(advanceXY.fY);
        } else {
            // this modifies the internal state in a fashion which
            // should not be returned to JAVA land...
            float tempX = fPositions[i*2];
            float tempY = fPositions[i*2 + 1];
            fPositions[i*2]   = tempX * fDevTX[0] + tempY * fDevTX[2]
                                    + x + hsFixedToFloat(glyphRef.fTopLeft.fX);
            fPositions[i*2+1] = tempX * fDevTX[1] + tempY * fDevTX[3]
                                    + y + hsFixedToFloat(glyphRef.fTopLeft.fY);
        }
      }
    }

  /*
   * PositionGlyphsWithImages is made only by TextRendering.cpp to prepare the glyphvector
   * before a lock is obtained on the buffer (or before entering code where a Critical JNI
   * call is to be made).  So we want to preload all the glyphs that we are going 
   * to render, which we did above.
   *
   * With the introduction of a per-glyph cache, it's possible that some of the glyphs
   * we load will be bumped by subsequent loads for the same glyphvector, and thus
   * that memory addresses will be invalidated.  So after we load the glyphs, we
   * pass through a second time to revalidate the addresses.
   *
   * We could do this in the rendering loop after clipping, but since the glyphs are not
   * objects, we'd have to expose the cache access to the rendering code.  By doing it
   * here we can hide it in the strike, while we have the strike, which is at least a
   * little better.
   */

    if (refcheck) {
      for (i=0; i < fNumGlyphs; i++) {
	if (fGlyphs[i] != INVISIBLE_GLYPH_ID) {
	  fImageRefs[i].pixels = theStrike.refImage(fGlyphs[i]);
	}
      }
    }
}

void GlyphVector::positionGlyphs(
        jfloat x, jfloat y,
        jdoubleArray matrix,
        Boolean isAntiAliased, Boolean usesFractionalMetrics)
{
    FontTransform tx(fEnv, matrix);
    Strike& theStrike = fFont->getStrike(tx, isAntiAliased, usesFractionalMetrics);
    // load positions if they are not already specified
    if (!fPositions) {
        allocatePositions();
    }
    if (!fPositions) {
        JNU_ThrowInternalError(fEnv, "Insufficient Memory");
        return;
    }
    for (unsigned int i=0; i < fNumGlyphs; i++) {
        fPositions[i*2] = x;
        fPositions[i*2+1] = y;
        if (fGlyphs[i] != INVISIBLE_GLYPH_ID) {
          hsGGlyph glyphRef;
          hsFixedPoint2 advanceXY;
          theStrike.getMetrics(fGlyphs[i], glyphRef, advanceXY);
          x += hsFixedToFloat(advanceXY.fX);
          y += hsFixedToFloat(advanceXY.fY);
        }
    }
    fPositions[fNumGlyphs * 2] = x;
    fPositions[fNumGlyphs * 2 + 1] = y;
}

void GlyphVector::getGlyphCodes(jobject gvTarget)
{
    if (initGVIDs(fEnv, gvTarget)) {
        jintArray glyphs =
            (jintArray)fEnv->GetObjectField(gvTarget, g_gvGlyphs);
        UInt32Buffer gb(fEnv, glyphs);
        fNumGlyphs = gb.getNumElements();
        allocateGlyphs();
        if (fGlyphs) {
            UInt32 *glyphCodes = gb.buffer();
            if (glyphCodes) {
                for (unsigned int i=0; i < fNumGlyphs; i++) {
                    fGlyphs[i] = glyphCodes[i];
                }
            }
        } else {
          fNumGlyphs = 0;
        }
    }
}

void GlyphVector::getPositions(jobject gvTarget, float x, float y,
                        jdoubleArray matrix,
                        Boolean isAntiAliased, Boolean usesFractionalMetrics)
{
    if (initGVIDs(fEnv, gvTarget)) {
        jfloatArray positions =
            (jfloatArray)fEnv->GetObjectField(gvTarget, g_gvPositions);
        setPositions(positions);
        if (!fPositions) {
            // position in absolute coordinates
            positionGlyphs(x,y, matrix, isAntiAliased, usesFractionalMetrics);
        } else {
            // caller's coordinates are relative to x, y.
            for (int i=0; i <= fNumGlyphs; i++) {
                fPositions[i*2] += x;
                fPositions[i*2+1] += y;
            }
        }
    }
}

void GlyphVector::writeGlyphCodes(jobject gvTarget)
{
    if (initGVIDs(fEnv, gvTarget)) {
        jintArray glyphsArray = fEnv->NewIntArray(fNumGlyphs);
        if (glyphsArray) {
            UInt32Buffer glyphBuffer(fEnv, glyphsArray);
            UInt32 *glyphs = glyphBuffer.buffer();
            if (glyphs) {
                if (fGlyphs) {
                    for (unsigned int i=0; i < fNumGlyphs; i++) {
                        glyphs[i] = fGlyphs[i];
                    }
                } else {
                    for (unsigned int i=0; i < fNumGlyphs; i++) {
                        glyphs[i] = 0;
                    }
                }
            }
        }
        if (glyphsArray) {
            fEnv->SetObjectField(gvTarget, g_gvGlyphs, glyphsArray);
        }
    }
}

void GlyphVector::writePositions(jobject gvTarget)
{
    if (initGVIDs(fEnv, gvTarget)) {
        jfloatArray positions = fEnv->NewFloatArray(fNumGlyphs * 2 + 2);
        if (positions) {
            JFloatBuffer floatBuffer(fEnv, positions);
            jfloat *floats = floatBuffer.buffer();
            if (floats) {
                if (fPositions) {
                    for (int i=0; i < fNumGlyphs*2 + 2; i++) {
                        floats[i] = fPositions[i];
                    }
                } else {
                    for (int i=0; i < fNumGlyphs*2 + 2; i++) {
                        floats[i] = 0;
                    }
                }
            }
        }
        if (positions) {
            fEnv->SetObjectField(gvTarget, g_gvPositions, positions);
        }
    }
}

void
GlyphVector::scanGlyph(jobject shapeTarget,
    Strike& theStrike,
    hsPathSpline& path,
    CurveWalker& walker,
    float xPos, float yPos,
    jint glyphIndex)
{
    if (fGlyphs[glyphIndex] == INVISIBLE_GLYPH_ID) {
      return;
    }

    path.Init(&path);
    theStrike.GetGlyphOutline(fGlyphs[glyphIndex], path);

    for (unsigned int j = 0; j < path.fContourCount; j++) {
        Boolean firstTime = true;
        if (theStrike.GetPathType (fGlyphs [glyphIndex]) == hsPathSpline::kQuadType) {
            walker.Init(&path.fContours[j], true);
            while (walker.NextQuad()) {
                if (firstTime) {
                    fEnv->CallVoidMethod(shapeTarget, g_shapeMoveTo,
                                    hsScalarToFloat(walker.a.fX) + xPos,
                                    hsScalarToFloat(walker.a.fY) + yPos);
                    firstTime = false;
                }
                if (walker.drawLine) {
                    fEnv->CallVoidMethod(shapeTarget, g_shapeLineTo,
                                    hsScalarToFloat(walker.b.fX) + xPos,
                                    hsScalarToFloat(walker.b.fY) + yPos);
                } else {
                    fEnv->CallVoidMethod(shapeTarget, g_shapeQuadTo,
                                    hsScalarToFloat(walker.b.fX) + xPos,
                                    hsScalarToFloat(walker.b.fY) + yPos,
                                    hsScalarToFloat(walker.c.fX) + xPos,
                                    hsScalarToFloat(walker.c.fY) + yPos);
                }
            }
            fEnv->CallVoidMethod(shapeTarget, g_shapeClosePath);
        } else {        /* is kCubicType */
            walker.Init(&path.fContours[j], false);
            while (walker.NextCubic ()) {
                if (firstTime) {
                    fEnv->CallVoidMethod(shapeTarget, g_shapeMoveTo,
                                    hsScalarToFloat(walker.a.fX) + xPos,
                                    hsScalarToFloat(walker.a.fY) + yPos);
                    firstTime = false;
                }
                if (walker.drawLine) {
                    fEnv->CallVoidMethod(shapeTarget, g_shapeLineTo,
                                    hsScalarToFloat(walker.b.fX) + xPos,
                                    hsScalarToFloat(walker.b.fY) + yPos);
                } else {
                    fEnv->CallVoidMethod(shapeTarget, g_shapeCurveTo,
                                    hsScalarToFloat(walker.b.fX) + xPos,
                                    hsScalarToFloat(walker.b.fY) + yPos,
                                    hsScalarToFloat(walker.c.fX) + xPos,
                                    hsScalarToFloat(walker.c.fY) + yPos,
                                    hsScalarToFloat(walker.d.fX) + xPos,
                                    hsScalarToFloat(walker.d.fY) + yPos);
                }
            }
            fEnv->CallVoidMethod(shapeTarget, g_shapeClosePath);
        }
    }
    path.Delete(&path);
}

void GlyphVector::scanShapeInto(jobject shapeTarget, jdoubleArray matrix,
                    jboolean isAntiAliased, jboolean usesFractionalMetrics)
{
    if (!fGlyphs || !fPositions || !initShapeIDs(fEnv, shapeTarget)) {
        return;
    }
    FontTransform tx(fEnv, matrix);
    Strike& theStrike = fFont->getStrike(tx, isAntiAliased, usesFractionalMetrics);

    hsPathSpline path;
    CurveWalker walker;
    for (unsigned int i=0; i < fNumGlyphs; i++) {
        float xPos = fPositions[i*2];
        float yPos = fPositions[i*2+1];

        scanGlyph(shapeTarget, theStrike, path, walker, xPos, yPos, i);
    }
    return;
}

void GlyphVector::scanShapeInto(jobject shapeTarget, jint glyphIndex,
                    jdoubleArray matrix,
                    jboolean isAntiAliased, jboolean usesFractionalMetrics)
{
    if (!fGlyphs || !fPositions || !initShapeIDs(fEnv, shapeTarget)) {
        return;
    }
    FontTransform tx(fEnv, matrix);
    Strike& theStrike = fFont->getStrike(tx, isAntiAliased, usesFractionalMetrics);
    hsPathSpline path;
    CurveWalker walker;
    scanGlyph(shapeTarget, theStrike, path, walker, 0, 0, glyphIndex);
    return;
}

// REMIND: lame.  the C++ compiler/linker complain if there
// is no implementation
void GlyphVector::operator delete(void *ptr)
{
}
