/*
 * @(#)fontObject.h	1.32 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

#ifndef fontObjectDefined
#define fontObjectDefined

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <jni.h>

#include "HeadSpin.h"

class Strike;           // forward referenced
class FontTransform;    // forward referenced

typedef unsigned char byte;

class fontObject;
typedef UInt16 Unicode;
enum FontFormats {
    kTrueTypeFontFormat,
    kType1FontFormat,
    kT2KFontFormat,
    kTTCFontFormat,
    kCompositeFontFormat,
    kNRFontFormat,
    kInvalidFontFormat = -1
};

#define kFullName 4
#define kFamilyName 1

#ifdef __cplusplus

typedef struct {
    char *fontName;
    char *fullName;
    char *familyName;
} t1FontInfo;

typedef struct {

  char   *fontName;
  char   *fullName;
  char   *familyName;

} fontNameRecord;

typedef struct {
    FILE *fd;
    char *buffer;
    size_t buflen;
    size_t boffs;
    char ch;
    int eof;
    int curToken;
    int prevToken;
} sDataRecord;

class CharToGlyphMapper {
public:
    virtual ~CharToGlyphMapper() {}

    virtual void CharsToGlyphs(
        int count, const Unicode unicodes[], UInt32 glyphs[]) const = 0;
    virtual Boolean canDisplay(Unicode ch) const = 0;
    virtual int getMissingGlyphCode() const { return 0; }
    UInt32 CharToGlyph(const Unicode unicode) const {
        UInt32 glyphCode;
        CharsToGlyphs(1, &unicode, &glyphCode);
        return glyphCode;
    }
};

struct FontStrikeDesc {
    fontObject      *fFontObj;
    float           fMatrix[4];
    Int32           fDoAntiAlias;
    Int32           fDoFracEnable;
    int             fStyle;
};
#define kFontStrikeDescTag  'Fsdt'

class fontObject {

private:
    Strike *fStrike;        // cached for getStrike

    // cached state for getStrike
    FontTransform *fStrikeTx;
    jboolean fStrikeIsAntiAliased;
    jboolean fStrikeUsesFractionalMetrics;

protected:
    enum FontFormats fFormat;
    Unicode *fFontname;
    int  fFontNameLen;

    void setStrike( Strike *theStrike, FontTransform& tx,
                    jboolean isAntiAliased,
                    jboolean usesFractionalMetrics);

 public:
    int fCurFont;                       //- current logical font
    int fFontCount;                     //- number of logical sub-fonts

    Boolean isValid;                    //- For lazy font verification; if false, font is not usable

    Boolean fFontNamesValid;
    fontNameRecord theFontNames;  
    // transient ... REMIND: change communication between FontWrapper & Strike
    int m_currentStyle;

    fontObject() :
	fStrike(NULL),
	fStrikeTx(NULL),
	fFormat(kInvalidFontFormat),
	fFontname(0),
	fFontNameLen(0),
	fCurFont(0),
	fFontCount(0),
	isValid(true),
	fFontNamesValid(false),
	m_currentStyle(0)
    {
    }
    
    virtual ~fontObject();

    virtual const void* ReadChunk(UInt32 offset,
                                  UInt32 length, void* chunk = nil) = 0;
    virtual void readBlock(size_t offs, size_t size, char *dest) = 0;
    virtual void ReleaseChunk(const void* chunk) = 0;
    virtual Boolean MatchName(int nameID, const UInt16 *name, int nameLen) = 0;
    virtual int GetName(UInt16& platformID, UInt16& scriptID, UInt16& languageID, UInt16& nameID, UInt16 *name) = 0;
    virtual const Unicode *GetFontName(int& nameLen) = 0;
    virtual const char *GetFontLocalPlatformName() const { return NULL;}
    virtual const char *GetFontNativeName () const { return NULL;}
    virtual const Unicode *GetFileName(int& nameLen) const { nameLen = 0; return NULL; }
 
    enum FontFormats GetFormat() const { return fFormat; }

    //	These are for TrueType
    virtual UInt32 FindTableSize(UInt32 tableTag, UInt32* offset);
    const void* ReadTable(UInt32 tableTag, UInt32* tableLength = nil);

    //	These are for T2K
    virtual const void* GetData() { return this->ReadChunk(0, 0x7fffffff); }
    virtual UInt32 GetDataLength() = 0;

    // informational methods
    virtual enum FontFormats getScalerID(
                                FontTransform& tx,
                                jboolean isAntiAliased,
                                jboolean usesFractionalMetrics)
    {
        return kT2KFontFormat;
    }

    Strike& getStrike(  FontTransform& tx,
                        jboolean isAntiAliased,
                        jboolean usesFractionalMetrics);
};

// Classes derived from fontObject
class fileFontObject : public fontObject {
 private:
    // can be multibyte in locale
    const char *fFontNativeName;
    const char *fFontLocalPlatformName;

    // always UNICODE
    Unicode * fFileName; 
    int fFileNameLen;
    int fUseCount;
    FILE* fFile;
    UInt32 fFileSize;
    
 public:
    fileFontObject() :
	fFontNativeName(0),
	fFontLocalPlatformName(0),
	fFileName(0),
	fFileNameLen(0),
	fUseCount(0),
	fFile(0),
	fFileSize(0) {}
    ~fileFontObject() {
        delete[] fFileName;
        if (fFontLocalPlatformName) {
            free((void *)fFontLocalPlatformName);
            fFontLocalPlatformName = NULL;
        }
        if (fFontNativeName) {
            free((void *)fFontNativeName);
            fFontNativeName = NULL;
        }
        assert(fUseCount == 0);
    }
    

    virtual const void* ReadChunk(UInt32 offset, UInt32 length,
                                  void* chunk = nil);

    virtual void readBlock(size_t offset, size_t size, char *dest);

    virtual void ReleaseChunk(const void* chunk);
    virtual UInt32 GetDataLength();

    virtual Boolean Init(const Unicode *fileName, int nameLen, 
			 const char *localPlatformName,
			 const char *nativeName,
                         enum FontFormats format, int logicalNumber = 0) {
        fFileName = new Unicode[nameLen];
        fFileNameLen = nameLen;
        memcpy(fFileName, fileName, nameLen * sizeof(Unicode));
        fFontLocalPlatformName = strdup(localPlatformName);
	if (nativeName != 0)
	    fFontNativeName = strdup (nativeName);
        fFormat = format;
        return true;
    }
    virtual const Unicode * GetFileName(int& nameLen) const { nameLen = fFileNameLen; return fFileName; }
    virtual const char * GetFontLocalPlatformName() const { return fFontLocalPlatformName; }
    virtual const char * GetFontNativeName () const 
			{ return fFontNativeName ; }
    
};

class sfntFileFontObject : public fileFontObject {
  protected:
    UInt32 *offsets;
  public:
    sfntFileFontObject() : offsets(0), fFontNameTable(NULL), fHintsTableChecked(false) {};
    ~sfntFileFontObject();

    virtual Boolean Init(const Unicode *fileName, int nameLen, 
			 const char *localPlatformName,
			 const char *nativeName,
                         enum FontFormats format, int logicalNumber = 0);
    virtual Boolean MatchName(int nameID, const UInt16 *name, int nameLen);
    virtual int GetName(UInt16& platformID, UInt16& scriptID, UInt16& languageID, UInt16& nameID, UInt16 *name);
    virtual const Unicode *GetFontName(int& nameLen);
    virtual UInt32 FindTableSize(UInt32 tableTag, UInt32* offset);
    inline int getLogicalFontCount() {return fFontCount;};
    // informational methods
    virtual enum FontFormats getScalerID(
                                FontTransform& tx,
                                jboolean isAntiAliased,
                                jboolean usesFractionalMetrics);
    char *fFontNameTable;
    Boolean fHintsTableChecked;
    Boolean fContainsHintsTable;

};
    
class type1FileFontObject : public fileFontObject {
public:
    virtual Boolean MatchName(int nameID, const UInt16 *name, int nameLen);
    virtual int GetName(UInt16& platformID, UInt16& scriptID, UInt16& languageID, UInt16& nameID, UInt16 *name);
    virtual const Unicode *GetFontName(int& nameLen);
    void DeleteFontInfo(t1FontInfo *fontInfo);
    t1FontInfo* GetFontInfo();
    virtual enum FontFormats getScalerID(
                                FontTransform& tx,
                                jboolean isAntiAliased,
                                jboolean usesFractionalMetrics);
    

private:
    int OpenFont(const char *fileName, sDataRecord *sData);
    void CloseFont(sDataRecord *sData);
    int NextToken(sDataRecord *sData, char *token, int toklen);
    int SkipSpaces(sDataRecord *sData);
    int NextChar(sDataRecord *sData);
};

void registerFont(enum FontFormats format, fontObject *cf);
fontObject *GetFontObject(const Unicode *fileName, int nameLength);
fontObject *GetFontObject(unsigned index);

/* Needed for native rasterizer support */
fileFontObject *GetNativeFontObject ();
void MakeNativeMapper (const char *fontString);


#endif

#ifdef __cplusplus
extern "C" {
#endif

    int CountFontObjects();
    fontObject *FindFontObject(const UInt16 * name, unsigned int nameLen, int style);

    const void *FOReadChunk(fontObject *fo, UInt32 offset, UInt32 length);
    void FOReleaseChunk(fontObject *fo, const void *chunk);
    long FOFindTableSize(fontObject *fo, UInt32 tableTag, UInt32 *offset);

    // These are need by T2K
    const void* FOGetData(fontObject *fo);	// return the entire font
    UInt32 FOGetDataLength(fontObject *fo);
    const Unicode *CreateTrueTypeFont ( Unicode *fontFileName, int fontNameLen, char *localName, int *nameLen );
#ifdef __cplusplus
}
#endif

#endif
