/*
 * @(#)hsGGlyphCache.cpp	1.21 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#if !defined(HS_BUILD_FOR_MAC)
#include <new.h>	// For _set_new_handler
#endif
#include <assert.h>
#include "hsGGlyphCache.h"
#include "hsTemplates.h"
#include "hsGFontScaler.h"

///////////////////////////////////////////////////////////////////////////////////
#define HS_BUILD_FOR_WIN32 1
typedef void (*NewHandlerProc)();
typedef void (*new_handler)(); 
// new_handler set_new_handler(new_handler); 

/*
 * A glyph cache. Relatively straightforward to implement, and efficient
 * to access, although it does require more bookkeeping of the clients.
 * The cache limits both total memory consumed and number of glyphs.
 * Here's a general description of the process omitting some details.
 *
 * Glyph memory is stored in entries held in a circular buffer.  Each entry
 * identifies the owner of the memory and the size of the memory, as well
 * as the memory pointer.  Owners refer to the entry by buffer index.  The
 * owner id must be unique for all users of the cache.  (In this use, the
 * owner is a combination of the strike id and the glyph index.  If a
 * strike id is reused then all glyphs for that strike must be purged,
 * this is similar to the purging of strikes currently.)
 *
 * Owners requesting new memory provide their id and the size of the
 * memory required.  The memory is allocated and returned, along with
 * the index.  This will purge the oldest entries in the cache to make 
 * room if necessary.
 *
 * Owners requesting old memory provide their id and the old index.
 * If the owner of the slot at the index matches the requestor, we have
 * a cache hit.  If the slot is near the 'top' of the buffer (most 
 * recently allocated slot) no more work needs to be done.  If it is 
 * near the bottom, then the top of the buffer is incremented, any
 * current memory there is freed, and the owner info is copied into the
 * new slot and erased from the old slot.  This keeps frequently accessed
 * entries near the top of the ring where they are  less likely to get
 * bumped, and also minimizes unnecessary bumping of other entries when
 * the requested entry is already near the top.
 *
 * Owners can revalidate the memory pointer without risking bumping other
 * objects from the cache.  This lets you update a number of pointers (or
 * set them to null) and be sure that at the end of the process all of the
 * pointers are still good.
 */

struct GlyphEntry {
  UInt32 fOwnerID; // id of owner of this memory
  void*  fMemPtr;  // memory owned
  UInt32 fMemSize; // size of memory allocated

  GlyphEntry() : fOwnerID(0), fMemPtr(0), fMemSize(0) {}

  ~GlyphEntry() { 
    if (fMemPtr) { 
      HSMemory::Delete(fMemPtr); 
      fMemPtr = 0; 
    }
  }

  // dangerous calls, make sure you don't reset fMemPtr by accident
  void set(UInt32 ownerID, void* memPtr, UInt32 memSize) {
    fOwnerID = ownerID;
    fMemPtr = memPtr;
    fMemSize = memSize;
  }

  void move(GlyphEntry& rhs) {
    set(rhs.fOwnerID, rhs.fMemPtr, rhs.fMemSize);
    rhs.reset();
  }

  void reset() {
    set(0, 0, 0);
  }

private:
  GlyphEntry& operator=(GlyphEntry&) { return *this; }
};

class GlyphMemCache {
  UInt16 fNumEntries;   // number of entries currently in cache
  UInt16 fMaxEntries;   // maximum entries in cache
  UInt32 fMem;          // amount of memory currently used by cache
  UInt32 fMaxMem;       // maximum memory to be used by cache (may be exceeded)
  UInt16 fNearLimit;    // number of entries considered to be near the top
  UInt16 fTop;          // current top
  GlyphEntry* fEntries; // the buffer of entries
  Boolean fPurge;       // if true, purge cache if needed, if false only purge on request.

  friend class GlyphEntry;

public:
  /** 
   * Construct a cache that can hold numEntries glyphs, with a maximum memory
   * usage of maxMem, moving entries to the top of the cache if one is 
   * located further than nearLimit from the top of the cache.  If purgeOn is
   * true, enable automatic cache management.  If it is false, entries are
   * only flushed when all glyph entries are in use, and aging entries are
   * not refetched to the top of the queue when requested.
   *
   * MaxEntries must be < 0x7fffu.
   */
  GlyphMemCache(UInt16 maxEntries, UInt16 nearLimit, UInt32 maxMem, Boolean purgeOn);

  /**
   * Free the cache.
   */
  ~GlyphMemCache();

  /**
   * The index for requests where no previous index exists.
   */
  static const UInt16 kNoIndex;

  /**
   * Turn cache management on or off.  When it is off, nothing gets purged from
   * the cache except by explicit request or when all glyph entries are in use.
   */
  void setPurgeOn(Boolean onOff) {
    fPurge = onOff;
  }

  /**
   * Reserve space for numEntries glyphs and memSize memory in the cache.  This
   * will purge glyphs if required to free up the necessary space.  If purged
   * is not null, it will return true if glyphs were purged.
   */
  void reserveSpace(UInt16 numEntries, UInt32 memSize, Boolean* purged);

  /**
   * Assign new memory of size bytes to cache, and associate with the 
   * ownerID.  The cache index is returned.  The memory can later be
   * freed using HSMemory::Delete.
   *
   * This may cause purging of memory in the cache.  If purged is not
   * null, the value it points to will be set to true if memory was purged.
   *
   * The memory is only guaranteed available until the next call to
   * assignMemory or getMemory. 
   *
   * OwnerID may be any 32-bit value.
   * Size must be < 0x7fffffffu.
   */
   UInt16 assignMemory(UInt32 ownerID, UInt32 size, void* memory, Boolean* purged = 0);

  /**
   * Return a pointer to the previously allocated memory associated with 
   * the ownerID and index.  The pointer can be null if the cache no 
   * longer holds memory for this owner at this index.
   *
   * This may cause purging of memory associated with other indexes.  If
   * purged is not null, the value it points to will be set to true if
   * memory was purged.
   *
   * The memory is only guaranteed available until the next call to
   * assignMemory or getMemory. 
   */
  void* getMemory(UInt32 ownerID, UInt16 &index, Boolean* purged = 0);

  /**
   * Return a pointer to the previously allocated memory associated with
   * the ownerID and index.  The pointer can be null if the cache no
   * longer holds memory for this owner at this index.
   *
   * This will not cause purging of memory in the cache.
   *
   * The memory is only guaranteed available until the next call to
   * assignMemory or getMemory. 
   */
  void* refMemory(UInt32 ownerID, UInt16 index);

  /**
   * Free the memory at index if it belongs to owner.
   */
  void releaseMemory(UInt32 ownerID, UInt16 index, Boolean* purged = 0);

  /**
   * Free all memory held by the cache.
   */
  void releaseAll();

private:
  UInt16 next(UInt16 index) const { 
    if (++index >= fMaxEntries) {
      index = 0; 
    }
    return index; 
  }

  void releaseEntry(GlyphEntry& e, Boolean* purged = 0) {
    if (e.fMemPtr) {
      fNumEntries--; 
      fMem -= e.fMemSize;
      HSMemory::Delete(e.fMemPtr);
      if (purged) {
	*purged = true;
      }
      e.reset();
    }
  } 

  void setEntry(GlyphEntry& e, UInt32 ownerID, void* memPtr, UInt32 memSize, Boolean* purged = 0) {
    releaseEntry(e, purged);
    e.set(ownerID, memPtr, memSize);
    if (memPtr) {
      fNumEntries++;
      fMem += memSize;
    }
  }

  void moveEntry(GlyphEntry& lhs, GlyphEntry& rhs, Boolean* purged = 0) {
    if (&lhs != &rhs) {
      releaseEntry(lhs, purged);
      lhs.move(rhs);
    }
  }
};

const UInt16 GlyphMemCache::kNoIndex = 0xffffu;

// --------------

GlyphMemCache::GlyphMemCache(UInt16 numEntries, UInt16 nearLimit, UInt32 maxMem, Boolean purgeOn)
  : fNumEntries(0)
  , fMaxEntries(numEntries)
  , fMem(0)
  , fMaxMem(maxMem)
  , fNearLimit(nearLimit)
  , fTop(numEntries - 1)
  , fEntries(new GlyphEntry[numEntries])
  , fPurge(purgeOn)
{
}

GlyphMemCache::~GlyphMemCache() {
  delete[] fEntries;
  fEntries = 0;
}

UInt16
GlyphMemCache::assignMemory(UInt32 ownerID, UInt32 memSize, void* mem, Boolean* purged) {
  if (mem) {
    // try to ensure that after allocation mem is still < maxmem
    if (fPurge && (fMem + memSize > fMaxMem)) {
      UInt16 bottom = fTop;
      UInt32 memLimit = fMaxMem > memSize ? fMaxMem - memSize : 0;
      while (fMem > memLimit && fNumEntries > 0) {
	bottom = next(bottom);
	releaseEntry(fEntries[bottom], purged); // adjusts fMem, fNumEntries
      }
    }

    fTop = next(fTop);
    if (!fPurge && fNumEntries < fMaxEntries) {
      while (fEntries[fTop].fMemPtr) {
	fTop = next(fTop);
      }
    }

    setEntry(fEntries[fTop], ownerID, mem, memSize, purged);

    return fTop;
  }

  return kNoIndex;
}

void*
GlyphMemCache::getMemory(UInt32 ownerID, UInt16& index, Boolean* purged) {

  if (index < fMaxEntries) {
    GlyphEntry& p = fEntries[index];
    if (p.fMemPtr && p.fOwnerID == ownerID) { // cache hit
      void* mem = p.fMemPtr;

      if (fPurge) {
	UInt16 dist = (fTop < index ? fTop + fMaxEntries : fTop) - index;
	if (dist > fNearLimit) { // move entry to top
	  fTop = next(fTop);

	  index = fTop; // set index
	  moveEntry(fEntries[index], p, purged); // test in move avoids copy to self
	}
      }

      return mem;
    }
  }

  // cache miss
  index = kNoIndex;
  return 0;
}

void*
GlyphMemCache::refMemory(UInt32 ownerID, UInt16 index) {
  if (index < fMaxEntries) {
    GlyphEntry& p = fEntries[index];
    if (p.fOwnerID == ownerID) {
      return p.fMemPtr; // if null that's ok
    }
  }

  return 0;
}

void
GlyphMemCache::releaseMemory(UInt32 ownerID, UInt16 index, Boolean* purged) {
  if (index < fMaxEntries) {
    GlyphEntry& p = fEntries[index];
    if (p.fOwnerID == ownerID) {
      releaseEntry(p, purged);
    }
  }
}

void 
GlyphMemCache::releaseAll() {
  for (UInt16 i = 0; i < fMaxEntries; ++i) {
    releaseEntry(fEntries[i]);
  }
}

void
GlyphMemCache::reserveSpace(UInt16 numEntries, UInt32 memSize, Boolean* purged) {
  UInt16 bottom = fTop;
  UInt16 entryLimit = fMaxEntries > numEntries ? fMaxEntries - numEntries : 0;
  while (fNumEntries > entryLimit) {
    bottom = next(bottom);
    releaseEntry(fEntries[bottom], purged); // adjusts fMem, fNumEntries
  }

  UInt32 memLimit = fMaxMem > memSize ? fMaxMem - memSize : 0;
  while (fMem > memLimit && fNumEntries > 0) {
    bottom = next(bottom);
    releaseEntry(fEntries[bottom], purged); // adjusts fMem, fNumEntries
  }
}

// ------------------------------------------------------------------------------

#define MAX_GLYPHS 768
#define MOVE_LIMIT 512
#define MAX_MEMORY (MAX_GLYPHS * 256)
#define PURGE_OFF 0
#define PURGE_ON 1

class hsGGlyphCache {
	static void		New_Handler();
	static NewHandlerProc	gPrevHandler;
	static hsGGlyphCache*	gCache;

	hsDynamicArray<hsGGlyphStrike*>	fStrikeList;
        GlyphMemCache           fMemCache;

	Boolean			Purge();

			        hsGGlyphCache(): fStrikeList(), fMemCache(MAX_GLYPHS, MOVE_LIMIT, MAX_MEMORY, PURGE_OFF) {};

	virtual			~hsGGlyphCache();
public:
	hsGGlyphStrike*	FindStrike(hsConstDescriptor desc);
	void				AddStrike(hsGGlyphStrike* strike);

        GlyphMemCache&          getMemCache() { return fMemCache; }

        void resetStrikeIDs();

	static void		Kill();
	static hsGGlyphCache* GetGlobalCache()
	{
		if (gCache == nil)
		{	gCache = new hsGGlyphCache();
	#if HS_BUILD_FOR_WIN32
		//	gPrevHandler = _set_new_handler(hsGGlyphCache::New_Handler);
	#else
//                		gPrevHandler = set_new_handler(hsGGlyphCache::New_Handler);
	#endif
		}
		return gCache;
	}
};

NewHandlerProc hsGGlyphCache::gPrevHandler;
hsGGlyphCache* hsGGlyphCache::gCache = nil;

///////////////////////////////////////////////////////////////////////////////////

struct hsGGlyphStrikeEntry {
	hsGGlyph		fGlyph;
	hsFixedPoint2	fAdvance;
        UInt16 fCacheIndex; // used for GlyphMemCache
};

#define kNoMetrics_Width		0xFFFF	// marks the fWidth field to indicate that there are no valid metrics yet
#define kNoImage_Image		(void*)-1L		// marks the fImage field to indicate that there is no valid image yet

UInt32 hsGGlyphStrike::gID = 0; // strike ID needs synchronized access

hsGGlyphStrike::hsGGlyphStrike(hsConstDescriptor desc)
{
	fDesc = hsDescriptor_Copy(desc);
	fID = ++gID << 16;
	fSizeHint = 10;

        UInt32 scalerID = hsDescriptor_Find32(desc, kScalerID_FontScalerDesc);
        hsGFontScaler*	scaler = hsGFontScaler::Find(scalerID);
	hsThrowIfNilParam(scaler);

	fScalerContext	= scaler->CreateContext(desc);
	fGlyphCount	= fScalerContext->CountGlyphs();
	fEntries	= new hsGGlyphStrikeEntry[fGlyphCount];

	for (unsigned i = 0; i < fGlyphCount; i++)
	{	fEntries[i].fGlyph.fWidth = kNoMetrics_Width;
		fEntries[i].fGlyph.fImage = kNoImage_Image;
		fEntries[i].fCacheIndex = GlyphMemCache::kNoIndex;
	}
        if (fGlyphCount) {
            fScalerContext->GetLineHeight(&fAscent, &fDescent, &fBaseline,
                                          &fLeading, &fMaxAdvance);
        }
}

hsGGlyphStrike::~hsGGlyphStrike()
{
  /* Don't need anymore, cache manages memory now.
	for (unsigned i = 0; i < fGlyphCount; i++)
		if (fEntries[i].fGlyph.fImage != kNoImage_Image)
			HSMemory::Delete((void*)fEntries[i].fGlyph.fImage);
			*/

	delete[] fEntries;	delete fScalerContext;
	
	hsDescriptor_Delete(fDesc);
}

void hsGGlyphStrike::resetID() {
  fID = ++gID << 16;
}

Boolean hsGGlyphStrike::canRotate ()
{
 	return fScalerContext->canRotate ();
}

hsPathSpline::Type hsGGlyphStrike::GetPathType() const
{
	return fScalerContext->GetPathType();
}

CharToGlyphMapper *
hsGGlyphStrike::getMapper()
{
    return fScalerContext->getMapper();
}

void hsGGlyphStrike::GetMetrics(UInt16 index, hsGGlyph* glyph, hsFixedPoint2* advance)
{
	hsAssert(index < fGlyphCount, "bad index");
        if (index >= fGlyphCount) {
            if (glyph) {
                memset(glyph, 0, sizeof(*glyph));
            }
            if (advance) {
                memset(advance, 0, sizeof(*advance));
            }
            return;
        }

	hsGGlyphStrikeEntry* entry = &fEntries[index];
	
	if (entry->fGlyph.fWidth == kNoMetrics_Width) {
            fScalerContext->GenerateMetrics(index, &entry->fGlyph, &entry->fAdvance);
            hsAssert(entry->fGlyph.fWidth != kNoMetrics_Width, "bad width flag");
	}
	if (glyph)
		*glyph = entry->fGlyph;
	if (advance)
		*advance = entry->fAdvance;
}

void hsGGlyphStrike::getEntryImage(hsGGlyphStrikeEntry* entry, Boolean* purged) {
  UInt16 glyphIndex = entry - fEntries;
  UInt32 ownerID = fID | glyphIndex;
  void* image = hsGGlyphCache::GetGlobalCache()->getMemCache().getMemory(ownerID, entry->fCacheIndex, purged);

  // signal that the image is not available in the cache
  entry->fGlyph.fImage = image ? image : kNoImage_Image;
}

void hsGGlyphStrike::refEntryImage(hsGGlyphStrikeEntry* entry) {
  UInt16 glyphIndex = entry - fEntries;
  UInt32 ownerID = fID | glyphIndex;
  
  // return null if image not available in the cache
  entry->fGlyph.fImage = hsGGlyphCache::GetGlobalCache()->getMemCache().refMemory(ownerID, entry->fCacheIndex);
}

void hsGGlyphStrike::assignEntryImage(hsGGlyphStrikeEntry* entry, Boolean* purged) {
  UInt16 glyphIndex = entry - fEntries;
  UInt32 ownerID = fID | glyphIndex;
  UInt32 memSize = entry->fGlyph.fHeight * entry->fGlyph.fRowBytes;
  fSizeHint = fSizeHint - fSizeHint >> 4 + memSize;

  entry->fCacheIndex = hsGGlyphCache::GetGlobalCache()->getMemCache().
    assignMemory(ownerID, memSize, (void*)entry->fGlyph.fImage, purged); // cast away const on void* memory
}

void hsGGlyphStrike::GetMetricsWithImage(UInt16 index, hsGGlyph* glyph, hsFixedPoint2* advance, Boolean* refcheck)
{
	hsAssert(index < fGlyphCount, "bad index");
        if (index >= fGlyphCount) {
            if (glyph) {
                memset(glyph, 0, sizeof(*glyph));
            }
            if (advance) {
                memset(advance, 0, sizeof(*advance));
            }
            return;
        }

	hsGGlyphStrikeEntry* entry = &fEntries[index];
	getEntryImage(entry, refcheck);

	if (        (entry->fGlyph.fWidth == kNoMetrics_Width)
                ||  (entry->fGlyph.fImage == kNoImage_Image)) {

            fScalerContext->GenerateMetricsWithImage(index, &entry->fGlyph, &entry->fAdvance);
	    assignEntryImage(entry, refcheck);

            hsAssert(entry->fGlyph.fWidth != kNoMetrics_Width, "bad width flag");
            hsAssert(entry->fGlyph.fImage != kNoImage_Image, "bad image");
	}
	if (glyph)
		*glyph = entry->fGlyph;
	if (advance)
		*advance = entry->fAdvance;
}

void hsGGlyphStrike::GetItalicAngle(hsFixedPoint2* italicAngle) {
    fScalerContext->GetCaretAngle(italicAngle);
}

const void* hsGGlyphStrike::RefImage(UInt16 index) {
	hsAssert(index < fGlyphCount, "bad index");
        if (index >= fGlyphCount) {
            return nil;
        }

	hsGGlyphStrikeEntry* entry = &fEntries[index];
	refEntryImage(entry);
	return entry->fGlyph.fImage;
}

void hsGGlyphStrike::PrepareCache(UInt32 numGlyphs, Boolean* purged) {
  hsGGlyphCache::GetGlobalCache()->getMemCache().reserveSpace(numGlyphs, numGlyphs * fSizeHint >> 4, purged);
}

const void* hsGGlyphStrike::GetImage(UInt16 index, Boolean* refcheck)
{
	hsAssert(index < fGlyphCount, "bad index");
        if (index >= fGlyphCount) {
            return nil;
        }

	hsGGlyphStrikeEntry* entry = &fEntries[index];
	getEntryImage(entry, refcheck);
	if (entry->fGlyph.fImage == kNoImage_Image ||
	    entry->fGlyph.fWidth == kNoMetrics_Width) {
	    entry->fGlyph.fImage = nil;

	    if (entry->fGlyph.fWidth == kNoMetrics_Width) {
	      fScalerContext->GenerateMetricsWithImage(index, &entry->fGlyph, &entry->fAdvance);
	    } else {
            UInt32 imageSize = entry->fGlyph.fHeight * entry->fGlyph.fRowBytes;
            void* image = HSMemory::SoftNew(imageSize);

            if (image) {
#ifndef CC_NOEX
                try {
#endif // !CC_NOEX
                    fScalerContext->GenerateImage(index, &entry->fGlyph, image);
		    entry->fGlyph.fImage = image;
		    assignEntryImage(entry, refcheck);
#ifndef CC_NOEX
                }
                catch (...) {
                    HSMemory::Delete(image);
                    image = kNoImage_Image;
                }
#endif // !CC_NOEX
            }
	    }
	}
	return entry->fGlyph.fImage;
}

void hsGGlyphStrike::GetPath(UInt16 index, hsPathSpline* path, hsFixedPoint2* advance)
{
    hsAssert(index < fGlyphCount, "bad index");
    if (index >= fGlyphCount) {
        return;
    }
    fScalerContext->GeneratePath(index, path, advance);
}

void hsGGlyphStrike::GetLineHeight(hsFixedPoint2* ascent, hsFixedPoint2* descent,
                                   hsFixedPoint2* baseline, hsFixedPoint2 *leading,
                                   hsFixedPoint2* maxAdvance)
{
	if (ascent)
		*ascent = fAscent;
	if (descent)
		*descent = fDescent;
	if (baseline)
		*baseline = fBaseline;
        if (leading)
                *leading = fLeading;
        if (maxAdvance)
                *maxAdvance = fMaxAdvance;
}

hsGGlyphStrike* hsGGlyphStrike::RefStrike(hsConstDescriptor desc)
{
	hsGGlyphCache* cache = hsGGlyphCache::GetGlobalCache();
	hsGGlyphStrike* strike = cache->FindStrike(desc);

	if (strike == nil) {
            strike = new hsGGlyphStrike(desc);
            cache->AddStrike(strike);
	    
	    if (gID > 0x7fff) { // empty glyph mem cache, reset ids of all current strikes
	      gID = 0;
	      cache->resetStrikeIDs();
	    }

	}
        // this is balanced by UnRef() the client when they're done using it
	strike->Ref();	
	
        return strike;
}

void hsGGlyphStrike::KillGlyphCache()
{
	hsGGlyphCache::Kill();
}

Boolean hsGGlyphStrike::isStyleSupported(int aStyle)
{
    if (fScalerContext) {
        return fScalerContext->isStyleSupported(aStyle);
    }
    return false;
}

///////////////////////////////////////////////////////////////////////////////////

static Boolean unrefStrike(hsGGlyphStrike*& strike)
{
	hsAssert(strike->RefCnt() == 1, "too many refs on strike");
	strike->UnRef();
	return false;
}

hsGGlyphCache::~hsGGlyphCache()
{
	(void)fStrikeList.ForEach(unrefStrike);
}

hsGGlyphStrike* hsGGlyphCache::FindStrike(hsConstDescriptor desc)
{
    hsGGlyphStrike**	strikeList = fStrikeList.AcquireArray();
    hsGGlyphStrike**	strike = strikeList;
    hsGGlyphStrike**	stop = strike + fStrikeList.GetCount();
    
    while (strike != stop) {
        if (hsDescriptor_Equal((*strike)->GetDesc(), desc)) {
            fStrikeList.ReleaseArray(strikeList);
            return *strike;
        }
        strike += 1;
    }
    fStrikeList.ReleaseArray(strikeList);
    return nil;
}

void hsGGlyphCache::resetStrikeIDs() {
    fMemCache.releaseAll();

    hsGGlyphStrike**	strikeList = fStrikeList.AcquireArray();
    hsGGlyphStrike**	strike = strikeList;
    hsGGlyphStrike**	stop = strike + fStrikeList.GetCount();
    
    while (strike != stop) {
        (**strike).resetID();
        strike++;
    }
    fStrikeList.ReleaseArray(strikeList);
}

#define kStrikeLimit		256

void hsGGlyphCache::AddStrike(hsGGlyphStrike* strike)
{
        int count = fStrikeList.GetCount();
	if (count > kStrikeLimit)
	{	Boolean	result = this->Purge();
		hsAssert(result, "cache is locked");
		count = 0;
	}
	fStrikeList.Append(strike);
}

Boolean hsGGlyphCache::Purge()
{
	Boolean didPurge = false;

	for (int i = fStrikeList.GetCount() - 1; i >= 0; i--)
	{	hsGGlyphStrike* strike = fStrikeList.Get(i);
	
		if (strike->RefCnt() == 1)
		{	delete strike;
			fStrikeList.Remove(i);
			didPurge = true;
		}
	}
	return didPurge;
}

void hsGGlyphCache::Kill()
{
	if (gCache)
	{	delete gCache;
		gCache = nil;
#if HS_BUILD_FOR_WIN32
//		set_new_handler(gPrevHandler);
#else
//                	set_new_handler(gPrevHandler);
#endif
	}
}

void hsGGlyphCache::New_Handler()
{
	if (gCache && gCache->Purge())
		return;

	if (gPrevHandler)
		gPrevHandler();
#ifndef CC_NOEX
	else
		throw "new failed";
#endif // !CC_NOEX
}

#undef HS_BUILD_FOR_WIN32
