/*
 * @(#)mlib_ImageConvMxNFunc.c	1.4 00/02/02
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

/*
 * @(#)mlib_ImageConvMxNFunc.c	4.2	99/04/28 SMI
 *
 * Copyright (c) 1996 - 1999, Sun Microsystems, Inc.
 */

/* Copyright (C) 1996, 1999, Sun Microsystems, Inc. */

#ifdef __SUNPRO_C
#pragma ident	"@(#)mlib_ImageConvMxNFunc.c	1.2	99/04/26 SMI"
#endif /* __SUNPRO_C */

/*
 * FUNCTION
 *      mlib_ImageConvMxN - image convolution with edge condition
 *
 * SYNOPSIS
 *      mlib_status mlib_ImageConvMxN(mlib_image *dst,
 *                                    mlib_image *src,
 *                                    mlib_s32   *kernel,
 *                                    mlib_s32   m,
 *                                    mlib_s32   n,
 *                                    mlib_s32   dm,
 *                                    mlib_s32   dn,
 *                                    mlib_s32   scale,
 *                                    mlib_s32   cmask,
 *                                    mlib_edge  edge)
 *
 * ARGUMENTS
 *      dst       Pointer to destination image.
 *      src       Pointer to source image.
 *      m         Kernel width (m must be larger than 1).
 *      n         Kernel height (n must be larger than 1).
 *      dm, dn    Position of key element in convolution kernel.
 *      kernel    Pointer to convolution kernel.
 *      scale     The scaling factor to convert the input integer
 *                coefficients into floating-point coefficients:
 *                floating-point coefficient = integer coefficient * 2^(-scale)
 *      cmask     Channel mask to indicate the channels to be convolved.
 *                Each bit of which represents a channel in the image. The
 *                channels corresponded to 1 bits are those to be processed.
 *      edge      Type of edge condition.
 *
 * DESCRIPTION
 *      2-D convolution, MxN kernel.
 *
 *      The center of the source image is mapped to the center of the
 *      destination image.
 *      The unselected channels are not overwritten. If both src and dst have
 *      just one channel, cmask is ignored.
 *
 *      The edge condition can be one of the following:
 *              MLIB_EDGE_DST_NO_WRITE  (default)
 *              MLIB_EDGE_DST_FILL_ZERO
 *              MLIB_EDGE_DST_COPY_SRC
 *              MLIB_EDGE_SRC_EXTEND
 *
 * RESTRICTION
 *      The src and the dst must be the same type and have same number
 *      of channels (1, 2, 3, or 4).
 *      m > 1.   n > 1.
 *      0 <= dm < m. 0 <= dn < n.
 *      For data type MLIB_BYTE - 16 <= scale <= 31 (to be compatible with VIS
 *      version).
 *      For data type MLIB_SHORT - 17 <= scale <= 32 (to be compatible with VIS
 *      version).
 *      For data type MLIB_INT - scale >= 0.
 */

/***************************************************************/

#include <stdlib.h>
#include <mlib_image.h>
#include <mlib_ImageCheck.h>
#include <mlib_SysMath.h>

/***************************************************************/

#define MLIB_D2_16 65536.0f
#define MLIB_D2_24 16777216.0f
extern mlib_f32 mlib_U82F32[];


/***************************************************************/

#ifdef MLIB_USE_FTOI_CLAMPING

#define FLOAT2INT_CLAMP(X) ((mlib_s32)(X))

#else

#define FLOAT2INT_CLAMP(X) (((X) >= MLIB_S32_MAX) ? MLIB_S32_MAX :  \
  (((X) <= MLIB_S32_MIN) ? MLIB_S32_MIN : ((mlib_s32)(X))))

#endif

/***************************************************************/

static void mlib_ImageGradientMxNMulAdd_U8(mlib_f32 *hdst,
                                           mlib_f32 *vdst,
                                           mlib_u8  *src,
                                           mlib_d64 *hfilter,
                                           mlib_d64 *vfilter,
                                           mlib_s32 n,
                                           mlib_s32 m,
                                           mlib_s32 nch)
{
  mlib_f32 *hdst1 = hdst + 1, *vdst1 = vdst + 1;
  mlib_s32 i, j;
  
  for(j = 0; j < m; j += 3, src += 3*nch, hfilter += 3, vfilter += 3) {
    mlib_u8  *src2  = src  + 2*nch;
    mlib_f32 hval0  = (mlib_f32) hfilter[0];
    mlib_f32 vval0  = (mlib_f32) vfilter[0];
    mlib_f32 hval1  = (mlib_f32) hfilter[1];
    mlib_f32 vval1  = (mlib_f32) vfilter[1];
    mlib_f32 hval2  = (mlib_f32) hfilter[2];
    mlib_f32 vval2  = (mlib_f32) vfilter[2];
    mlib_f32 val0   = mlib_U82F32[src[0]];
    mlib_f32 val1   = mlib_U82F32[src[nch]];
    mlib_f32 hdvl   = hdst[0];
    mlib_f32 vdvl   = vdst[0];
    
    if (j == m - 2) {
      hval2 = 0.f;
      vval2 = 0.f;
    } else if (j == m - 1) {
      hval1 = 0.f;
      vval1 = 0.f;
      hval2 = 0.f;
      vval2 = 0.f;
    }
 
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for(i = 0; i < n; i++) {
      mlib_f32 hdvl0 = val0*hval0 + hdvl;
      mlib_f32 vdvl0 = val0*vval0 + vdvl;
      mlib_f32 val2  = mlib_U82F32[src2[i*nch]];
 
      hdvl      = hdst1[i];
      vdvl      = vdst1[i];
      hdvl0    += val1*hval1;
      vdvl0    += val1*vval1;
      hdvl0    += val2*hval2;
      vdvl0    += val2*vval2;
      val0      = val1;
      val1      = val2;
 
      hdst[i]  = hdvl0;
      vdst[i]  = vdvl0;
    }
  }
}

/***************************************************************/

static void mlib_ImageConvMxNMulAdd_U8(mlib_f32 *dst,
                                       mlib_u8  *src,
                                       mlib_d64 *dkernel,
                                       mlib_s32 n,
                                       mlib_s32 m,
                                       mlib_s32 nch)
{
  mlib_f32 *dst1 = dst + 1;
  mlib_s32 i, j;
  
  for(j = 0; j < m; j += 3, src += 3*nch, dkernel += 3) {
    mlib_u8  *src2  = src  + 2*nch;
    mlib_f32 hval0  = (mlib_f32) dkernel[0];
    mlib_f32 hval1  = (mlib_f32) dkernel[1];
    mlib_f32 hval2  = (mlib_f32) dkernel[2];
    mlib_f32 val0   = mlib_U82F32[src[0]];
    mlib_f32 val1   = mlib_U82F32[src[nch]];
    mlib_f32 dval   = dst[0];
    
    if (j == m - 2) {
      hval2 = 0.f;
    } else if (j == m - 1) {
      hval1 = 0.f;
      hval2 = 0.f;
    }

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for(i = 0; i < n; i++) {
      mlib_f32 dval0 = val0*hval0 + dval;
      mlib_f32 val2  = mlib_U82F32[src2[i*nch]];

      dval      = dst1[i];
      dval0    += val1*hval1;
      dval0    += val2*hval2;
      val0      = val1;
      val1      = val2;

      dst[i]    = dval0;
    }
  }
}

/***************************************************************/

static void mlib_ImageConvMxNMedian_U8(mlib_u8  *dst,
                                       mlib_f32 *src,
                                       mlib_s32 n,
                                       mlib_s32 nch)
{
  mlib_s32 i;

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
  for(i = 0; i < n; i++) 
    ((mlib_s32*)src)[i] = FLOAT2INT_CLAMP(src[i]*MLIB_D2_24);

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
  for(i = 0; i < n; i++) {
    mlib_s32 detl = 128 + (((mlib_s32*)src)[i] >> 24);
    
    src[i]    = -128.f;
    dst[i*nch] = detl;
  }
}

/***************************************************************/

void mlib_ImageConvMxN_U8(mlib_image *dst,
                          mlib_image *src,
                          mlib_d64   *dkernel,
                          mlib_s32   m,
                          mlib_s32   n,
                          mlib_u8    *acmask,
                          mlib_d64   *dsa)
{
  mlib_f32   *fsa = (mlib_f32*)dsa, *fsh, *fsv;
  mlib_u8    *da = mlib_ImageGetData(dst);
  mlib_u8    *sa = mlib_ImageGetData(src);
  mlib_s32   dlb = mlib_ImageGetStride(dst);
  mlib_s32   slb = mlib_ImageGetStride(src);
  mlib_s32   dw  = mlib_ImageGetWidth(dst);
  mlib_s32   dh  = mlib_ImageGetHeight(dst);
  mlib_s32   nch = mlib_ImageGetChannels(dst);
  mlib_s32   i, j, j1, k;

  fsh  = fsa + dw + m;
  fsv  = fsh + dw;

  for(i = 0; i < dw; i++) {
    fsh[i] = -128.f;
    fsv[i] = -128.f;
  }

  for(j = 0; j < dh; j+=2, da += 2*dlb, sa += 2*slb) 
  for(k = 0; k < nch; k++) 
  if (acmask[k]) {
    mlib_u8  *sa1      = sa + k;
    mlib_d64 *dkernel1 = dkernel;
    
    if (j != dh - 1) {
      mlib_ImageConvMxNMulAdd_U8(fsh, sa1, dkernel1, dw, m, nch);
      sa1 += slb;

      for(j1 = 1; j1 < n; j1++, sa1 += slb, dkernel1 += m)
        mlib_ImageGradientMxNMulAdd_U8(fsh, fsv, sa1, dkernel1+m, dkernel1, dw,
          m, nch);

      mlib_ImageConvMxNMulAdd_U8(fsv, sa1, dkernel1, dw, m, nch);
      mlib_ImageConvMxNMedian_U8(da + dlb + k, fsv, dw, nch);
      mlib_ImageConvMxNMedian_U8(da + k, fsh, dw, nch);
    } else {
      for(j1 = 0; j1 < n; j1++, sa1 += slb, dkernel1 += m)
        mlib_ImageConvMxNMulAdd_U8(fsh, sa1, dkernel1, dw, m, nch);
      mlib_ImageConvMxNMedian_U8(da + k, fsh, dw, nch);
    }
  }
}

/***************************************************************/

static void mlib_ImageConvMxNU82U8_ext(mlib_u8  *dst, 
                                       mlib_u8  *src, 
                                       mlib_s32 n,
                                       mlib_s32 nch,
                                       mlib_s32 dx_l,
                                       mlib_s32 dx_r)
{
  mlib_s32 i;
  mlib_s32 val = src[0];
  
  for(i = 0; i < dx_l; i++) dst[i] = val;
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
  for(; i < n - dx_r; i++) dst[i] = src[nch*(i-dx_l)];
  val = dst[n - dx_r - 1];
  for(; i < n; i++) dst[i] = val;
}

/***************************************************************/

void mlib_ImageConvMxN_U8_ext(mlib_image *dst,
                              mlib_image *src,
                              mlib_d64   *dkernel,
                              mlib_s32   m,
                              mlib_s32   n,
                              mlib_s32   dx_l,
                              mlib_s32   dx_r,
                              mlib_s32   dy_t,
                              mlib_s32   dy_b,
                              mlib_u8    *acmask,
                              mlib_d64   *dsa)
{
  mlib_f32   *fsa = (mlib_f32*)dsa, *fsh, *fsv;
  mlib_u8    *usa = (mlib_u8*)dsa;
  mlib_u8    *da = mlib_ImageGetData(dst);
  mlib_u8    *sa = mlib_ImageGetData(src);
  mlib_s32   dlb = mlib_ImageGetStride(dst);
  mlib_s32   slb = mlib_ImageGetStride(src);
  mlib_s32   dw  = mlib_ImageGetWidth(dst);
  mlib_s32   dh  = mlib_ImageGetHeight(dst);
  mlib_s32   nch = mlib_ImageGetChannels(dst);
  mlib_s32   i, j, j1, k;

  fsh  = fsa + dw + m;
  fsv  = fsh + dw;

  for(i = 0; i < dw; i++) {
    fsh[i] = -128.f;
    fsv[i] = -128.f;
  }
  
  for(j = 0; j < dh; j++, da += dlb) {
    for(k = 0; k < nch; k++) 
    if (acmask[k]) {
      mlib_u8  *sa1      = sa + k;
      mlib_d64 *dkernel1 = dkernel;

      for(j1 = 0; j1 < n; j1++, dkernel1 += m) {
        mlib_ImageConvMxNU82U8_ext(usa, sa1, dw + m - 1, nch, dx_l, dx_r);
        mlib_ImageConvMxNMulAdd_U8(fsh, usa, dkernel1, dw, m, 1);
        if ((j + j1 >= dy_t) && (j + j1 < dh + n - dy_b - 2)) sa1 += slb;
      }

      mlib_ImageConvMxNMedian_U8(da + k, fsh, dw, nch);
    }
    if ((j >= dy_t) && (j < dh + n - dy_b - 2)) sa += slb;
  }
}

/***************************************************************/

static void mlib_ImageConvMxNS162S32(mlib_s32 *dst, 
                                     mlib_s16 *src, 
                                     mlib_s32 n,
                                     mlib_s32 nch)
{
  mlib_s32 i;
  
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
  for(i = 0; i < n; i++) dst[i] = src[nch*i];
}

/***************************************************************/

static void mlib_ImageGradientMxNMulAdd_S16(mlib_f32 *hdst,
                                            mlib_f32 *vdst,
                                            mlib_s32 *src,
                                            mlib_d64 *hfilter,
                                            mlib_d64 *vfilter,
                                            mlib_s32 n,
                                            mlib_s32 m)
{
  mlib_f32 *hdst1 = hdst + 1, *vdst1 = vdst + 1;
  mlib_s32 i, j;
  
  for(j = 0; j < m; j += 3, src += 3, hfilter += 3, vfilter += 3) {
    mlib_s32 *src2  = src  + 2;
    mlib_f32 hval0  = (mlib_f32) hfilter[0];
    mlib_f32 vval0  = (mlib_f32) vfilter[0];
    mlib_f32 hval1  = (mlib_f32) hfilter[1];
    mlib_f32 vval1  = (mlib_f32) vfilter[1];
    mlib_f32 hval2  = (mlib_f32) hfilter[2];
    mlib_f32 vval2  = (mlib_f32) vfilter[2];
    mlib_f32 val0   = (mlib_f32) src[0];
    mlib_f32 val1   = (mlib_f32) src[1];
    mlib_f32 hdvl   = hdst[0];
    mlib_f32 vdvl   = vdst[0];
    
    if (j == m - 2) {
      hval2 = 0.f;
      vval2 = 0.f;
    } else if (j == m - 1) {
      hval1 = 0.f;
      vval1 = 0.f;
      hval2 = 0.f;
      vval2 = 0.f;
    }

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for(i = 0; i < n; i++) {
      mlib_f32 hdvl0 = val0*hval0 + hdvl;
      mlib_f32 vdvl0 = val0*vval0 + vdvl;
      mlib_f32 val2  = (mlib_f32) src2[i];

      hdvl      = hdst1[i];
      vdvl      = vdst1[i];
      hdvl0    += val1*hval1;
      vdvl0    += val1*vval1;
      hdvl0    += val2*hval2;
      vdvl0    += val2*vval2;
      val0      = val1;
      val1      = val2;

      hdst[i]  = hdvl0;
      vdst[i]  = vdvl0;
    }
  }
}

/***************************************************************/

static void mlib_ImageConvMxNMulAdd_S16(mlib_f32 *dst,
                                        mlib_s32 *src,
                                        mlib_d64 *dkernel,
                                        mlib_s32 n,
                                        mlib_s32 m)
{
  mlib_f32 *dst1 = dst + 1;
  mlib_s32 i, j;
  
  for(j = 0; j < m; j += 3, src += 3, dkernel += 3) {
    mlib_s32 *src2  = src  + 2;
    mlib_f32 hval0  = (mlib_f32) dkernel[0];
    mlib_f32 hval1  = (mlib_f32) dkernel[1];
    mlib_f32 hval2  = (mlib_f32) dkernel[2];
    mlib_f32 val0   = (mlib_f32) src[0];
    mlib_f32 val1   = (mlib_f32) src[1];
    mlib_f32 dval   = dst[0];
    
    if (j == m - 2) {
      hval2 = 0.f;
    } else if (j == m - 1) {
      hval1 = 0.f;
      hval2 = 0.f;
    }

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for(i = 0; i < n; i++) {
      mlib_f32 dval0 = val0*hval0 + dval;
      mlib_f32 val2  = (mlib_f32) src2[i];

      dval      = dst1[i];
      dval0    += val1*hval1;
      dval0    += val2*hval2;
      val0      = val1;
      val1      = val2;

      dst[i]    = dval0;
    }
  }
}

/***************************************************************/

static void mlib_ImageConvMxNMedian_S16(mlib_s16 *dst,
                                        mlib_f32 *src,
                                        mlib_s32 n,
                                        mlib_s32 nch)
{
  mlib_s32 i;

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
  for(i = 0; i < n; i++) 
    ((mlib_s32*)src)[i] = FLOAT2INT_CLAMP(src[i]*MLIB_D2_16);

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
  for(i = 0; i < n; i++) {
    mlib_s32 detl = ((mlib_s32*)src)[i];

    src[i]    = 0.5f;
    dst[i*nch] = detl >> 16;
  }
}

/***************************************************************/

void mlib_ImageConvMxN_S16(mlib_image *dst,
                           mlib_image *src,
                           mlib_d64   *dkernel,
                           mlib_s32   m,
                           mlib_s32   n,
                           mlib_u8    *acmask,
                           mlib_d64   *dsa)
{
  mlib_f32   *fsa = (mlib_f32*)dsa, *fsh, *fsv;
  mlib_s32   *isa = (mlib_s32*)dsa;
  mlib_s16   *da = mlib_ImageGetData(dst);
  mlib_s16   *sa = mlib_ImageGetData(src);
  mlib_s32   dlb = mlib_ImageGetStride(dst) >> 1;
  mlib_s32   slb = mlib_ImageGetStride(src) >> 1;
  mlib_s32   dw  = mlib_ImageGetWidth(dst);
  mlib_s32   dh  = mlib_ImageGetHeight(dst);
  mlib_s32   nch = mlib_ImageGetChannels(dst);
  mlib_s32   i, j, j1, k;

  fsh  = fsa + dw + m;
  fsv  = fsh + dw;

  for(i = 0; i < dw; i++) {
    fsh[i] = 0.5f;
    fsv[i] = 0.5f;
  }
  
  for(j = 0; j < dh; j += 2, da += 2*dlb, sa += 2*slb) 
  for(k = 0; k < nch; k++) 
  if (acmask[k]) {
    mlib_s16 *sa1      = sa + k;
    mlib_d64 *dkernel1 = dkernel;
    
    if (j != dh - 1) {
      mlib_ImageConvMxNS162S32(isa, sa1, dw + m - 1, nch);
      mlib_ImageConvMxNMulAdd_S16(fsh, isa, dkernel1, dw, m);
      sa1 += slb;

      for(j1 = 1; j1 < n; j1++, sa1 += slb, dkernel1 += m) {
        mlib_ImageConvMxNS162S32(isa, sa1, dw + m - 1, nch);
        mlib_ImageGradientMxNMulAdd_S16(fsh, fsv, isa, dkernel1+m, dkernel1, dw,
          m);
      }

      mlib_ImageConvMxNS162S32(isa, sa1, dw + m - 1, nch);
      mlib_ImageConvMxNMulAdd_S16(fsv, isa, dkernel1, dw, m);
      mlib_ImageConvMxNMedian_S16(da + dlb + k, fsv, dw, nch);
      mlib_ImageConvMxNMedian_S16(da + k, fsh, dw, nch);
    } else {
      for(j1 = 0; j1 < n; j1++, sa1 += slb, dkernel1 += m) {
        mlib_ImageConvMxNS162S32(isa, sa1, dw + m - 1, nch);
        mlib_ImageConvMxNMulAdd_S16(fsh, isa, dkernel1, dw, m);
      }
      mlib_ImageConvMxNMedian_S16(da + k, fsh, dw, nch);
    }
  }
}

/***************************************************************/

static void mlib_ImageConvMxNS162S32_ext(mlib_s32 *dst, 
                                         mlib_s16 *src, 
                                         mlib_s32 n,
                                         mlib_s32 nch,
                                         mlib_s32 dx_l,
                                         mlib_s32 dx_r)
{
  mlib_s32 i;
  mlib_s32 val = src[0];
  
  for(i = 0; i < dx_l; i++) dst[i] = val;
  for(; i < n - dx_r; i++) dst[i] = src[nch*(i-dx_l)];
  val = dst[n - dx_r - 1];
  for(i = n - dx_r; i < n; i++) dst[i] = val;
}

/***************************************************************/

void mlib_ImageConvMxN_S16_ext(mlib_image *dst,
                               mlib_image *src,
                               mlib_d64   *dkernel,
                               mlib_s32   m,
                               mlib_s32   n,
                               mlib_s32   dx_l,
                               mlib_s32   dx_r,
                               mlib_s32   dy_t,
                               mlib_s32   dy_b,
                               mlib_u8    *acmask,
                               mlib_d64   *dsa)
{
  mlib_f32   *fsa = (mlib_f32*)dsa, *fsh, *fsv;
  mlib_s32   *isa = (mlib_s32*)dsa;
  mlib_s16   *da = mlib_ImageGetData(dst);
  mlib_s16   *sa = mlib_ImageGetData(src);
  mlib_s32   dlb = mlib_ImageGetStride(dst) >> 1;
  mlib_s32   slb = mlib_ImageGetStride(src) >> 1;
  mlib_s32   dw  = mlib_ImageGetWidth(dst);
  mlib_s32   dh  = mlib_ImageGetHeight(dst);
  mlib_s32   nch = mlib_ImageGetChannels(dst);
  mlib_s32   i, j, j1, k;

  fsh  = fsa + dw + m;
  fsv  = fsh + dw;

  for(i = 0; i < dw; i++) {
    fsh[i] = 0.5f;
    fsv[i] = 0.5f;
  }
  
  for(j = 0; j < dh; j++, da += dlb) {
    for(k = 0; k < nch; k++)
    if (acmask[k]) {
      mlib_s16 *sa1      = sa + k;
      mlib_d64 *dkernel1 = dkernel;

      for(j1 = 0; j1 < n; j1++, dkernel1 += m) {
        mlib_ImageConvMxNS162S32_ext(isa, sa1, dw + m - 1, nch, dx_l, dx_r);
        mlib_ImageConvMxNMulAdd_S16(fsh, isa, dkernel1, dw, m);
        if ((j + j1 >= dy_t) && (j + j1 < dh + n - dy_b - 2)) sa1 += slb;
      }

      mlib_ImageConvMxNMedian_S16(da + k, fsh, dw, nch);
    }
    if ((j >= dy_t) && (j < dh + n - dy_b - 2)) sa += slb;
  }
}

/***************************************************************/

static void mlib_ImageGradientMxNMulAdd_S32(mlib_d64 *hdst,
                                            mlib_d64 *vdst,
                                            mlib_s32 *src,
                                            mlib_d64 *hfilter,
                                            mlib_d64 *vfilter,
                                            mlib_s32 n,
                                            mlib_s32 m,
                                            mlib_s32 nch)
{
  mlib_d64 *hdst1 = hdst + 1, *vdst1 = vdst + 1;
  mlib_s32 i, j;
  
  for(j = 0; j < m; j += 3, src += 3*nch, hfilter += 3, vfilter += 3) {
    mlib_s32 *src2  = src  + 2*nch;
    mlib_d64 hval0  = hfilter[0];
    mlib_d64 vval0  = vfilter[0];
    mlib_d64 hval1  = hfilter[1];
    mlib_d64 vval1  = vfilter[1];
    mlib_d64 hval2  = hfilter[2];
    mlib_d64 vval2  = vfilter[2];
    mlib_d64 val0   = src[0];
    mlib_d64 val1   = src[nch];
    mlib_d64 hdvl   = hdst[0];
    mlib_d64 vdvl   = vdst[0];
    
    if (j == m - 2) {
      hval2 = 0.;
      vval2 = 0.;
    } else if (j == m - 1) {
      hval1 = 0.;
      vval1 = 0.;
      hval2 = 0.;
      vval2 = 0.;
    }

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for(i = 0; i < n; i++) {
      mlib_d64 hdvl0 = val0*hval0 + hdvl;
      mlib_d64 vdvl0 = val0*vval0 + vdvl;
      mlib_d64 val2  = src2[i*nch];

      hdvl      = hdst1[i];
      vdvl      = vdst1[i];
      hdvl0    += val1*hval1;
      vdvl0    += val1*vval1;
      hdvl0    += val2*hval2;
      vdvl0    += val2*vval2;
      val0      = val1;
      val1      = val2;

      hdst[i]  = hdvl0;
      vdst[i]  = vdvl0;
    }
  }
}

/***************************************************************/

static void mlib_ImageConvMxNMulAdd_S32(mlib_d64 *dst,
                                        mlib_s32 *src,
                                        mlib_d64 *dkernel,
                                        mlib_s32 n,
                                        mlib_s32 m,
                                        mlib_s32 nch)
{
  mlib_d64 *dst1 = dst + 1;
  mlib_s32 i, j;
  
  for(j = 0; j < m; j += 3, src += 3*nch, dkernel += 3) {
    mlib_s32 *src2  = src  + 2*nch;
    mlib_d64 hval0  = dkernel[0];
    mlib_d64 hval1  = dkernel[1];
    mlib_d64 hval2  = dkernel[2];
    mlib_d64 val0   = src[0];
    mlib_d64 val1   = src[nch];
    mlib_d64 dval   = dst[0];
    
    if (j == m - 2) {
      hval2 = 0.f;
    } else if (j == m - 1) {
      hval1 = 0.f;
      hval2 = 0.f;
    }

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for(i = 0; i < n; i++) {
      mlib_d64 dval0 = val0*hval0 + dval;
      mlib_d64 val2  = src2[i*nch];

      dval      = dst1[i];
      dval0    += val1*hval1;
      dval0    += val2*hval2;
      val0      = val1;
      val1      = val2;

      dst[i]    = dval0;
    }
  }
}

/***************************************************************/

static void mlib_ImageConvMxNMedian_S32(mlib_s32 *dst,
                                        mlib_d64 *src,
                                        mlib_s32 n,
                                        mlib_s32 nch)
{
  mlib_s32 i;

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
  for(i = 0; i < n; i++) {
    mlib_s32 detl = FLOAT2INT_CLAMP(src[i]);

    src[i]    = 0.5;
    dst[i*nch] = detl;
  }
}

/***************************************************************/

void mlib_ImageConvMxN_S32(mlib_image *dst,
                           mlib_image *src,
                           mlib_d64   *dkernel,
                           mlib_s32   m,
                           mlib_s32   n,
                           mlib_u8    *acmask,
                           mlib_d64   *dsa)
{
  mlib_d64   *dsh, *dsv;
  mlib_s32   *da = mlib_ImageGetData(dst);
  mlib_s32   *sa = mlib_ImageGetData(src);
  mlib_s32   dlb = mlib_ImageGetStride(dst) >> 2;
  mlib_s32   slb = mlib_ImageGetStride(src) >> 2;
  mlib_s32   dw  = mlib_ImageGetWidth(dst);
  mlib_s32   dh  = mlib_ImageGetHeight(dst);
  mlib_s32   nch = mlib_ImageGetChannels(dst);
  mlib_s32   i, j, j1, k;

  dsh  = dsa + dw + m;
  dsv  = dsh + dw;

  for(i = 0; i < dw; i++) {
    dsh[i] = 0.5;
    dsv[i] = 0.5;
  }

  for(j = 0; j < dh; j += 2, da += 2*dlb, sa += 2*slb) 
  for(k = 0; k < nch; k++) 
  if (acmask[k]) {
    mlib_s32 *sa1      = sa + k;
    mlib_d64 *dkernel1 = dkernel;
    
    if (j != dh - 1) {
      mlib_ImageConvMxNMulAdd_S32(dsh, sa1, dkernel1, dw, m, nch);
      sa1 += slb;

      for(j1 = 1; j1 < n; j1++, sa1 += slb, dkernel1 += m)
        mlib_ImageGradientMxNMulAdd_S32(dsh, dsv, sa1, dkernel1+m, dkernel1, dw,
          m, nch);

      mlib_ImageConvMxNMulAdd_S32(dsv, sa1, dkernel1, dw, m, nch);
      mlib_ImageConvMxNMedian_S32(da + dlb + k, dsv, dw, nch);
      mlib_ImageConvMxNMedian_S32(da + k, dsh, dw, nch);
    } else {
      for(j1 = 0; j1 < n; j1++, sa1 += slb, dkernel1 += m)
        mlib_ImageConvMxNMulAdd_S32(dsh, sa1, dkernel1, dw, m, nch);

      mlib_ImageConvMxNMedian_S32(da + k, dsh, dw, nch);
    }
  }
}

/***************************************************************/

static void mlib_ImageConvMxNS322S32_ext(mlib_s32 *dst, 
                                         mlib_s32 *src, 
                                         mlib_s32 n,
                                         mlib_s32 nch,
                                         mlib_s32 dx_l,
                                         mlib_s32 dx_r)
{
  mlib_s32 i;
  mlib_d64 val = src[0];
  
  for(i = 0; i < dx_l; i++) dst[i] = (mlib_s32) val;
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
  for(; i < n - dx_r; i++) dst[i] = (mlib_s32) src[nch*(i-dx_l)];
  val = dst[n - dx_r - 1];
  for(; i < n; i++) dst[i] = (mlib_s32) val;
}

/***************************************************************/

void mlib_ImageConvMxN_S32_ext(mlib_image *dst,
                               mlib_image *src,
                               mlib_d64   *dkernel,
                               mlib_s32   m,
                               mlib_s32   n,
                               mlib_s32   dx_l,
                               mlib_s32   dx_r,
                               mlib_s32   dy_t,
                               mlib_s32   dy_b,
                               mlib_u8    *acmask,
                               mlib_d64   *dsa)
{
  mlib_d64   *dsh, *dsv;
  mlib_s32   *isa = (mlib_s32*)dsa;
  mlib_s32   *da = mlib_ImageGetData(dst);
  mlib_s32   *sa = mlib_ImageGetData(src);
  mlib_s32   dlb = mlib_ImageGetStride(dst) >> 2;
  mlib_s32   slb = mlib_ImageGetStride(src) >> 2;
  mlib_s32   dw  = mlib_ImageGetWidth(dst);
  mlib_s32   dh  = mlib_ImageGetHeight(dst);
  mlib_s32   nch = mlib_ImageGetChannels(dst);
  mlib_s32   i, j, j1, k;

  dsh  = dsa + dw + m;
  dsv  = dsh + dw;

  for(i = 0; i < dw; i++) {
    dsh[i] = 0.5;
    dsv[i] = 0.5;
  }
  
  for(j = 0; j < dh; j++, da += dlb) {
    for(k = 0; k < nch; k++) 
    if (acmask[k]) {
      mlib_s32 *sa1      = sa + k;
      mlib_d64 *dkernel1 = dkernel;

      for(j1 = 0; j1 < n; j1++, dkernel1 += m) {
        mlib_ImageConvMxNS322S32_ext(isa, sa1, dw + m - 1, nch, dx_l, dx_r);
        mlib_ImageConvMxNMulAdd_S32(dsh, isa, dkernel1, dw, m, 1);
        if ((j + j1 >= dy_t) && (j + j1 < dh + n - dy_b - 2)) sa1 += slb;
      }

      mlib_ImageConvMxNMedian_S32(da + k, dsh, dw, nch);
    }
    if ((j >= dy_t) && (j < dh + n - dy_b - 2)) sa += slb;
  }
}

/***************************************************************/
