/*
 * @(#)jcov.c	1.6 01/04/21
 *
 * Copyright 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

#include <stdlib.h>

#include "jvmpi.h"
#include "jcov.h"
#include "jcov_htables.h"
#include "jcov_error.h"
#include "jcov_setup.h"
#include "jcov_events.h"

JVMPI_Interface *jvmpi_interface;
static JavaVM *jvm;

#ifndef JVMPI_VERSION_1_1
#define JVMPI_VERSION_1_1 ((jint)0x10000002)
#endif



#define def_event_name(e,v) case e: v=#e ; break
static char*
get_event_name(int event_no) {
    char* s = "UNKNOWN_EVENT";
    switch (event_no) {
        def_event_name(JVMPI_EVENT_METHOD_ENTRY, s);
        def_event_name(JVMPI_EVENT_METHOD_ENTRY2, s);
        def_event_name(JVMPI_EVENT_METHOD_EXIT, s);
        def_event_name(JVMPI_EVENT_OBJECT_ALLOC, s);
        def_event_name(JVMPI_EVENT_OBJECT_FREE, s);
        def_event_name(JVMPI_EVENT_OBJECT_MOVE, s);
        def_event_name(JVMPI_EVENT_COMPILED_METHOD_LOAD, s);
        def_event_name(JVMPI_EVENT_COMPILED_METHOD_UNLOAD, s);
        def_event_name(JVMPI_EVENT_THREAD_START, s);
        def_event_name(JVMPI_EVENT_THREAD_END, s);
        def_event_name(JVMPI_EVENT_INSTRUCTION_START, s);
        def_event_name(JVMPI_EVENT_CLASS_LOAD_HOOK, s);
        def_event_name(JVMPI_EVENT_HEAP_DUMP, s);
        def_event_name(JVMPI_EVENT_JNI_GLOBALREF_ALLOC, s);
        def_event_name(JVMPI_EVENT_JNI_GLOBALREF_FREE, s);
        def_event_name(JVMPI_EVENT_JNI_WEAK_GLOBALREF_ALLOC, s);
        def_event_name(JVMPI_EVENT_JNI_WEAK_GLOBALREF_FREE, s);
        def_event_name(JVMPI_EVENT_CLASS_LOAD, s);
        def_event_name(JVMPI_EVENT_CLASS_UNLOAD, s);
        def_event_name(JVMPI_EVENT_DATA_DUMP_REQUEST, s);
        def_event_name(JVMPI_EVENT_DATA_RESET_REQUEST, s);
        def_event_name(JVMPI_EVENT_JVM_INIT_DONE, s);
        def_event_name(JVMPI_EVENT_JVM_SHUT_DOWN, s);
        def_event_name(JVMPI_EVENT_ARENA_NEW, s);
        def_event_name(JVMPI_EVENT_ARENA_DELETE, s);
        def_event_name(JVMPI_EVENT_OBJECT_DUMP, s);
        def_event_name(JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTER, s);
        def_event_name(JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTERED, s);
        def_event_name(JVMPI_EVENT_RAW_MONITOR_CONTENDED_EXIT, s);
        def_event_name(JVMPI_EVENT_MONITOR_CONTENDED_ENTER, s);
        def_event_name(JVMPI_EVENT_MONITOR_CONTENDED_ENTERED, s);
        def_event_name(JVMPI_EVENT_MONITOR_CONTENDED_EXIT, s);
        def_event_name(JVMPI_EVENT_MONITOR_WAIT, s);
        def_event_name(JVMPI_EVENT_MONITOR_WAITED, s);
        def_event_name(JVMPI_EVENT_MONITOR_DUMP, s);
        def_event_name(JVMPI_EVENT_GC_START, s);
        def_event_name(JVMPI_EVENT_GC_FINISH, s);
    }
    return s;
}

static void
jcov_notify_event(JVMPI_Event *event)
{
    cov_item_t *cov_table, *ct_entry;
    int hash;
    int diff;
    jint instr_pc;
    jcov_method_t *meth;

    switch (event->event_type) {
    case JVMPI_EVENT_METHOD_ENTRY: 
	/* find the method in the hash table, inc exec counter */
	jcov_method_entry_event(event);
	return;
    case JVMPI_EVENT_CLASS_LOAD:
	/* if the class being loaded has previously been unloaded
	 *  then use data structures allocated on the first loading
	 *  of this class (having updated the class' and its methods' IDs;
	 *  otherwise - create new ones;
	 *  class' compare key : (source name, class name, timestamp)
	 */
        jcov_class_load_event(event);
	return;
    case JVMPI_EVENT_CLASS_LOAD_HOOK:
	/* parse the class' bytecode, create jcov_method_t structures
	 *  for each method and put them into temporary array for later
	 *  use by JVMPI_EVENT_CLASS_LOAD event handler; also extract
	 *  compare key for this class (source name, class name, timestamp)
	 */
	jcov_class_load_hook_event(event);
	return;
    case JVMPI_EVENT_CLASS_UNLOAD:
	/* lookup the class in the class table and mark it as unloaded */
        jcov_class_unload_event(event);
	return;
    case JVMPI_EVENT_OBJECT_MOVE:
        /* update ID of moved class and rehash it */
        jcov_object_move_event(event);
	return;
    case JVMPI_EVENT_INSTRUCTION_START:
        /* Increment counter of corresponding jcov_item_t.
        * For the sake of efficiency, this code is not
        * a separate function
        */
        meth = lookup_method0(event->u.instruction.method_id);
        if (!meth || meth->covtable_size < 2) {
            return;
        }
        instr_pc = event->u.instruction.offset;
        cov_table = meth->covtable;
        hash = meth->pc_cache[instr_pc];
        if (!hash) {
            return;
        }
        ct_entry = cov_table + hash - 1;
        switch (ct_entry->instr_type) {
        case INSTR_ANY:
            while (ct_entry->pc == instr_pc) {
                ct_entry->count++;
                if (ct_entry == cov_table) {
                    return;
                }
                ct_entry--;
            }
            return;
        case INSTR_IF:
            event->u.instruction.u.if_info.is_true ? 
                (ct_entry - 1)->count++ : ct_entry->count++;
            return;
        case INSTR_LOOKUPSW:
#define BASE event->u.instruction.u.lookupswitch_info
            (ct_entry - BASE.pairs_total + BASE.chosen_pair_index)->count++;
#undef BASE
            return;
        case INSTR_TABLESW:
#define BASE event->u.instruction.u.tableswitch_info
            diff = BASE.key - BASE.hi;
            if (diff > 0 || BASE.key - BASE.low < 0) {
                ct_entry->count++;
            } else {
                (ct_entry + diff - 1)->where_line ? 
                    (ct_entry + diff - 1)->count++ : ct_entry->count++;
            }
#undef BASE
            return;
        default:
            jcov_internal_error(12);
        }
    case JVMPI_EVENT_GC_START:
        /* lock access to hash tables */
        jcov_gc_start_event();
        return;
    case JVMPI_EVENT_GC_FINISH:
        /* unlock access to hash tables */
        jcov_gc_finish_event();
        return;
    case JVMPI_EVENT_THREAD_START:
        /* allocate new jcov_thread_t object, initialize it and put
        * it into the 'thread_table' hash table
        */
        jcov_thread_start_event(event);
        return;
    case JVMPI_EVENT_THREAD_END:
        /* remove corresponding jcov_thread_t object from thread_table
        * and free it
        */
        jcov_thread_end_event(event);
        return;
    case JVMPI_EVENT_JVM_SHUT_DOWN:
        /* save jcov data */
        jcov_jvm_shut_down_event();
        return;
    default:
        return;
    }
}

/* jcov initialization */
JNIEXPORT jint JNICALL
JVM_OnLoad(JavaVM *vm, char *options, void *reserved)
{
    char err_message[256];
    int res;
    jint err_code;
    jvm = vm;
    res = (*jvm)->GetEnv(jvm, (void**)&jvmpi_interface, JVMPI_VERSION_1_1);
    if (res < 0) {
        res = (*jvm)->GetEnv(jvm, (void**)&jvmpi_interface, JVMPI_VERSION_1);
        if (res < 0) {
            return JNI_ERR;
        }
    }
    jvmpi_interface->NotifyEvent = jcov_notify_event;
    
#define ENABLE_JVMPI_EVENT(event_type) \
    err_code = jvmpi_interface->EnableEvent(event_type, NULL); \
    if (err_code != JVMPI_SUCCESS) { \
        sprintf(err_message, "cannot enable JVMPI event : %s", \
                get_event_name(event_type)); \
        jcov_error(err_message); \
        return JNI_ERR; \
    }

    jcov_init(options);

    ENABLE_JVMPI_EVENT(JVMPI_EVENT_CLASS_LOAD);
    ENABLE_JVMPI_EVENT(JVMPI_EVENT_CLASS_LOAD_HOOK);
    ENABLE_JVMPI_EVENT(JVMPI_EVENT_CLASS_UNLOAD);
    ENABLE_JVMPI_EVENT(JVMPI_EVENT_OBJECT_MOVE);
    ENABLE_JVMPI_EVENT(JVMPI_EVENT_GC_START);
    ENABLE_JVMPI_EVENT(JVMPI_EVENT_GC_FINISH);
    ENABLE_JVMPI_EVENT(JVMPI_EVENT_THREAD_START);
    ENABLE_JVMPI_EVENT(JVMPI_EVENT_THREAD_END);
    ENABLE_JVMPI_EVENT(JVMPI_EVENT_METHOD_ENTRY);
    ENABLE_JVMPI_EVENT(JVMPI_EVENT_JVM_SHUT_DOWN);

    if (jcov_data_type == JCOV_DATA_B) {
        ENABLE_JVMPI_EVENT(JVMPI_EVENT_INSTRUCTION_START);
    }

    if (verbose_mode > 0) {
        sprintf(err_message, "jcov data type is set to \'%c\'", jcov_data_type);
        jcov_info(err_message);
    }

    return JNI_OK;
#undef ENABLE_JVMPI_EVENT
}
