/*
 * @(#)jcov_events.c	1.9 01/04/21
 *
 * Copyright 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
#include <stdlib.h>
#include <string.h>

#include "typedefs.h"
#include "jvmpi.h"
#include "jvm.h"
#include "opcodes.h"

#include "jcov.h"
#include "jcov_error.h"
#include "jcov_util.h"
#include "jcov_htables.h"
#include "jcov_setup.h"
#include "jcov_file.h"

#define METH_CACHE_SIZE 256 /* default size of cache for methods in a hooked class */

JVMPI_RawMonitor jcov_meth_table_lock;
JVMPI_RawMonitor jcov_thrd_table_lock;
JVMPI_RawMonitor jcov_clss_table_lock;

#define LOCK_THREADS (jvmpi_interface->RawMonitorEnter)(jcov_thrd_table_lock)
#define LOCK_CLASSES (jvmpi_interface->RawMonitorEnter)(jcov_clss_table_lock)
#define LOCK_METHODS (jvmpi_interface->RawMonitorEnter)(jcov_meth_table_lock)
#define UNLOCK_THREADS (jvmpi_interface->RawMonitorExit)(jcov_thrd_table_lock)
#define UNLOCK_CLASSES (jvmpi_interface->RawMonitorExit)(jcov_clss_table_lock)
#define UNLOCK_METHODS (jvmpi_interface->RawMonitorExit)(jcov_meth_table_lock)

#define CALL_DEPTH 2
#define JDK11_CT_ENTRY_PACK_BITS 18
#define JDK11_CT_ENTRY_POS_MASK  0x3FFFF

static int is_hook_first;

static void jcov_thread_new(JNIEnv *thread_id);
static jcov_hooked_class_t *jcov_hooked_class_new(void);

void jcov_class_load_hook_event(JVMPI_Event *event) {
    jint class_data_len;
    jint *len;
    int count, entry_count, pos, attr_count, where_line, non_abstract_met, i, j, k, m;
    unsigned int bytes2;
    long attr_len, skip, bytes4;
    char info[MAX_PATH_LEN];
    int cp_count = 0, meth_count = 0;
    unsigned char opcode = 0;
    unsigned char err_told = 0;
    unsigned char *class_data = NULL;
    unsigned char *code = NULL;
    unsigned char **ptr = NULL;
    cp_entry_t **cp = NULL;
    cp_entry_t *cp_entry = NULL;
    jcov_method_t *meth = NULL;
    jcov_thread_t *this_thread = NULL;
    jcov_hooked_class_t *hooked_class = NULL;
    cov_item_t *cov_item = NULL;
    int cov_table_entry_size;
    char *path = NULL;
    int err_code = 0;

    if (is_hook_first) {
        jcov_thread_new(event->env_id);
        is_hook_first = 0;
    }
    LOCK_THREADS;
    this_thread = lookup_thread(event->env_id);
    UNLOCK_THREADS;
    CHK(!this_thread, 1);

    hooked_class = jcov_hooked_class_new();
    class_data_len = event->u.class_load_hook.class_data_len;
    class_data = event->u.class_load_hook.class_data;
    event->u.class_load_hook.new_class_data = 
        (unsigned char*)event->u.class_load_hook.malloc_f(class_data_len);
    memcpy(event->u.class_load_hook.new_class_data, class_data, class_data_len);
    event->u.class_load_hook.new_class_data_len = class_data_len;

    ptr = &class_data;
    len = &class_data_len;
#define ASSRT(cond, n) if (!(cond)) { \
        if (hooked_class != NULL && hooked_class->name != NULL) \
            sprintf(info, "assertion failure #%d in class: %s", n, hooked_class->name); \
        else \
            sprintf(info, "assertion failure #%d (class unknown)", n); \
        jcov_error(info); \
        jcov_free_hooked_class(hooked_class); \
        jcov_free_constant_pool(cp, cp_count); \
        return; \
    }

#define READ_AND_CHECK(dest, bytes_total) \
    dest = read##bytes_total##bytes(ptr, len, &err_code); \
    if (err_code) { \
        if (hooked_class != NULL && hooked_class->name != NULL) \
            sprintf(info, "bad class format : %s", hooked_class->name); \
        else \
            sprintf(info, "bad class format"); \
        jcov_error(info); \
        jcov_free_hooked_class(hooked_class); \
        jcov_free_constant_pool(cp, cp_count); \
        return; \
    }
#define ERROR_VERBOSE(err_msg) if (verbose_mode > 0) jcov_error(err_msg)

    READ_AND_CHECK(bytes4, 4);
    if (bytes4 != 0xCAFEBABE) {
        jcov_free_hooked_class(hooked_class); \
        ERROR_VERBOSE("wrong magic number");
        return;
    }

#define SKIP(n) *ptr += n ; *len -= n ;
#define ASSRT_CP_ENTRY(x, n) if (!cp_entry || cp_entry->tag != x) { \
        if (hooked_class != NULL && hooked_class->name != NULL) \
            sprintf(info, "bad constant pool entry in : %s (assrt: %d)", hooked_class->name, n); \
        else \
            sprintf(info, "bad constant pool entry (assrt: %d)", n); \
        jcov_error(info); \
        jcov_free_hooked_class(hooked_class); \
        jcov_free_constant_pool(cp, cp_count); \
        return; \
    }
#define GET_CP_ENTRY(index) \
    if (index >= cp_count) { \
        if (hooked_class != NULL && hooked_class->name != NULL) \
            sprintf(info, "invalid constant pool index in %s : %d", hooked_class->name, index); \
        else \
            sprintf(info, "invalid constant pool index"); \
        jcov_error(info); \
        jcov_free_hooked_class(hooked_class); \
        jcov_free_constant_pool(cp, cp_count); \
        return; \
    } \
    cp_entry = cp[index]

    SKIP(4); /* class file version number */
    READ_AND_CHECK(cp_count, 2); /* number of items in constant pool */
    cp = (cp_entry_t**)jcov_calloc(sizeof(cp_entry_t*) * cp_count);
    for (i = 1; i < cp_count; i++) { /* read the whole cp */
        cp[i] = read_next_cp_entry(ptr, len, &err_code);
        if (err_code) {
            jcov_free_hooked_class(hooked_class);
            jcov_free_constant_pool(cp, cp_count);
            jcov_error("bad constant pool format");
            return;
        }
        if (cp[i]->tag == JVM_CONSTANT_Long || cp[i]->tag == JVM_CONSTANT_Double) {
            i++;
        }
    }
    READ_AND_CHECK(bytes2, 2);
    hooked_class->access_flags = bytes2 & JVM_RECOGNIZED_CLASS_MODIFIERS & ~JVM_ACC_SUPER;
    if (hooked_class->access_flags & JVM_ACC_INTERFACE) {
        hooked_class->access_flags &= ~JVM_ACC_ABSTRACT;
    }
    READ_AND_CHECK(bytes2, 2);
    GET_CP_ENTRY(bytes2); /*this_class*/
    ASSRT_CP_ENTRY(JVM_CONSTANT_Class, 0);
    GET_CP_ENTRY(cp_entry->u.Class.name_index);
    ASSRT_CP_ENTRY(JVM_CONSTANT_Utf8, 1);

    if (string_suits_filter(class_filter, cp_entry->u.Utf8.bytes) || caller_filter != NULL) {
        hooked_class->name = jcov_strdup(cp_entry->u.Utf8.bytes);
    } else {
        jcov_free_hooked_class(hooked_class);
        jcov_free_constant_pool(cp, cp_count);
        return;
    }
    SKIP(2); /* super_class */
    READ_AND_CHECK(skip, 2);
    skip *= 2;
    SKIP(skip); /* skip interfaces */

    READ_AND_CHECK(count, 2); /* fields_count */
    for (i = 0; i < count; i++) {
        SKIP(6);
        /* attributes_count */
        READ_AND_CHECK(attr_count, 2);
        for (j = 0; j < attr_count; j++) {
            SKIP(2); /* attribute_name_index */
            READ_AND_CHECK(attr_len, 4);
            SKIP(attr_len);
        }
    }
    /*-- we're at methods section at last --*/
    READ_AND_CHECK(hooked_class->methods_total, 2); /* methods_count */
    hooked_class->method_cache = (jcov_method_t**)jcov_calloc(sizeof(jcov_method_t*) *
                                                              hooked_class->methods_total);
    non_abstract_met = 0;
    for (i = 0; i < hooked_class->methods_total; i++) {
        meth = (jcov_method_t*)jcov_calloc(sizeof(jcov_method_t));
        meth->covtable_size = 0;
        meth->covtable = NULL;
        meth->pc_cache = NULL;
        meth->pc_cache_size = 0;
        READ_AND_CHECK(meth->access_flags, 2);
        meth->access_flags &= JVM_RECOGNIZED_METHOD_MODIFIERS;
        non_abstract_met |= !(meth->access_flags & JVM_ACC_ABSTRACT);
        where_line = 0;
        READ_AND_CHECK(bytes2, 2);
        GET_CP_ENTRY(bytes2); /* method name */
        ASSRT_CP_ENTRY(JVM_CONSTANT_Utf8, 2);
        meth->name = jcov_strdup(cp_entry->u.Utf8.bytes);
        READ_AND_CHECK(bytes2, 2);
        GET_CP_ENTRY(bytes2); /* method signature */
        ASSRT_CP_ENTRY(JVM_CONSTANT_Utf8, 3);
        meth->signature = jcov_strdup(cp_entry->u.Utf8.bytes);
        /* read the method's attributes */
        READ_AND_CHECK(count, 2); /* attribute count */
        for (j = 0; j < count; j++) {
            READ_AND_CHECK(bytes2, 2);
            GET_CP_ENTRY(bytes2);
            ASSRT_CP_ENTRY(JVM_CONSTANT_Utf8, 4);
            READ_AND_CHECK(attr_len, 4);
            if (strcmp(cp_entry->u.Utf8.bytes, CODE_ATTR_NAME)) {
                SKIP(attr_len);
                continue;
            }
            meth->pc_cache_size = attr_len;
            /* read coverage table */
            SKIP(4); /* max_stack and max_locals */
            READ_AND_CHECK(skip, 4); /* code_count */
            code = *ptr;
            SKIP(skip);
            READ_AND_CHECK(skip, 2);
            skip *= 8; /* handlers_count * handler_size */
            SKIP(skip);
            READ_AND_CHECK(attr_count, 2); /* attributes_count */
            for (k = 0; k < attr_count; k++) {
                READ_AND_CHECK(bytes2, 2);
                GET_CP_ENTRY(bytes2);
                ASSRT_CP_ENTRY(JVM_CONSTANT_Utf8, 5);
                READ_AND_CHECK(attr_len, 4);
                if (!strcmp(cp_entry->u.Utf8.bytes, LINE_NUMBER_TABLE_ATTR_NAME)) {
                    READ_AND_CHECK(bytes2, 2);
                    if (bytes2 < 1) { /* lines_count */
                        skip = attr_len - 2;
                        SKIP(skip);
                    } else {
                        SKIP(2); /* 1st start_pc */
                        READ_AND_CHECK(where_line, 2);
                        skip = attr_len - 6;
                        SKIP(skip);
                    }
                    continue;
                }
                if (jcov_data_type == JCOV_DATA_M ||
                    strcmp(cp_entry->u.Utf8.bytes, COV_TABLE_ATTR_NAME)) {
                    SKIP(attr_len);
                    continue;
                }

                READ_AND_CHECK(entry_count, 2); /* number of entries in the coverage table */
                ASSRT(entry_count > 0, 0);
                ASSRT((attr_len - 2) % entry_count == 0, 1);
                cov_table_entry_size = (attr_len - 2) / entry_count;
                ASSRT(cov_table_entry_size == 8 /* JDK1.1 */ || cov_table_entry_size == 12 /* >= JDK1.2 */, 2);
                meth->covtable_size = entry_count;
                meth->covtable = (cov_item_t*)jcov_calloc(entry_count * sizeof(cov_item_t));
                meth->pc_cache = (int*)jcov_calloc(meth->pc_cache_size * sizeof(int));
                err_told = 0;
                for (m = 0; m < entry_count; m++) {
                    cov_item = &(meth->covtable[m]);
                    READ_AND_CHECK(cov_item->pc, 2);
                    if (cov_item->pc >= meth->pc_cache_size) {
                        if (!err_told) {
                            if (hooked_class != NULL && hooked_class->name != NULL && meth->name != NULL)
                                sprintf(info,
                                        "invalid CoverageTable attribute in class %s method %s",
                                        hooked_class->name,
                                        meth->name);
                            else
                                sprintf(info, "invalid CoverageTable attribute met");
                            jcov_error(info);
                            err_told = 1;
                        }
                        continue;
                    }
                    opcode = code[cov_item->pc];
                    cov_item->instr_type = INSTR_ANY;
                    if (opcode >= opc_ifeq && opcode <= opc_if_acmpne ||
                        opcode == opc_ifnull || opcode == opc_ifnonnull) {
                        cov_item->instr_type = INSTR_IF;
                    } else if (opcode == opc_tableswitch) {
                        cov_item->instr_type = INSTR_TABLESW;
                    } else if (opcode == opc_lookupswitch) {
                        cov_item->instr_type = INSTR_LOOKUPSW;
                    }
                    READ_AND_CHECK(cov_item->type,       2);
                    if (cov_table_entry_size == 8) {
                        READ_AND_CHECK(pos, 4);
                        cov_item->where_line = pos >> JDK11_CT_ENTRY_PACK_BITS;
                        cov_item->where_pos = pos & JDK11_CT_ENTRY_POS_MASK;
                    } else { /* cov_table_entry_size == 12 */
                        READ_AND_CHECK(cov_item->where_line, 4);
                        READ_AND_CHECK(cov_item->where_pos,  4);
                    }
                    cov_item->count = 0;
                    meth->pc_cache[cov_item->pc] = m+1;
                }
            }
        }
        if (meth->covtable == NULL) {
            meth->covtable = (cov_item_t*)jcov_calloc(sizeof(cov_item_t));
            meth->covtable_size = 1;
            meth->covtable[0].where_line = where_line;
            meth->covtable[0].where_pos = 0;
            meth->covtable[0].type = CT_METHOD;
            meth->covtable[0].count = 0;
        }
        hooked_class->method_cache[i] = meth;
    }
    if (!non_abstract_met) {  /* only abstarct methods encountered */
        jcov_free_hooked_class_full(hooked_class);
        jcov_free_constant_pool(cp, cp_count);
        return;
    }
    /* read the class' attributes */
    READ_AND_CHECK(count, 2); /* class attrs total */
    for (i = 0; i < count; i++) {
        READ_AND_CHECK(bytes2, 2);
        GET_CP_ENTRY(bytes2); /* attr name index */
        ASSRT_CP_ENTRY(JVM_CONSTANT_Utf8, 6);
        READ_AND_CHECK(attr_len, 4); /* bytes_count */
        if (!strcmp(cp_entry->u.Utf8.bytes, ABS_SRC_PATH_ATTR_NAME)) {
            READ_AND_CHECK(bytes2, 2);
            GET_CP_ENTRY(bytes2);
            ASSRT_CP_ENTRY(JVM_CONSTANT_Utf8, 7);
            path = jcov_strdup(cp_entry->u.Utf8.bytes);
        } else if (!strcmp(cp_entry->u.Utf8.bytes, SRC_FILE_ATTR_NAME)) {
            READ_AND_CHECK(bytes2, 2);
            GET_CP_ENTRY(bytes2);
            ASSRT_CP_ENTRY(JVM_CONSTANT_Utf8, 8);
            hooked_class->src_name = jcov_strdup(cp_entry->u.Utf8.bytes);
        } else if (!strcmp(cp_entry->u.Utf8.bytes, TIMESTAMP_ATTR_NAME)) {
            union {
                struct {
                    unsigned long hi;
                    unsigned long lo;
                } parts;
                int64_t timestamp;
            } tstamp;
            
            READ_AND_CHECK(tstamp.parts.hi, 4);
            READ_AND_CHECK(tstamp.parts.lo, 4);
            sprintf(hooked_class->timestamp, "%lld", tstamp.timestamp);
        } else {
    	    SKIP(attr_len);
        }
    }
    if (path) {
        jcov_free(hooked_class->src_name);
        hooked_class->src_name = path;
    } else {
        if (jcov_data_type == JCOV_DATA_B) {
            jcov_free_hooked_class_full(hooked_class);
            jcov_free_constant_pool(cp, cp_count);
            return;
        }
        hooked_class->data_type = JCOV_DATA_M;
    }
    if (hooked_class->src_name == NULL) {
        hooked_class->src_name = dummy_src_name(hooked_class->name);
    }
    /* constant pool may be freed now */
    jcov_free_constant_pool(cp, cp_count);
    if (verbose_mode > 1) {
        sprintf(info, "CLASS_LOAD_HOOK : %s", hooked_class->name);
        jcov_info(info);
    }

    put_hooked_class(&(this_thread->hooked_class_table), &hooked_class);
	
#undef SKIP
#undef ASSRT_CP_ENTRY
}

void jcov_class_load_event(JVMPI_Event *event) {
    char info[MAX_PATH_LEN];
    jcov_class_t *class;
    jcov_hooked_class_t *hooked_class;
    jcov_class_t *found_class;
    jcov_method_t **meth_ptr;
    jcov_method_t *found_method;
    jcov_method_t **class_methods;
    JVMPI_Method *meth;
    jcov_thread_t *this_thread;
    int i, ind;
    long mem;
    char *name, *tmp;
    int last_matched = 0;
    
    LOCK_THREADS;
    this_thread = lookup_thread(event->env_id);
    UNLOCK_THREADS;
    CHK(!this_thread, 3);

    name = jcov_strdup(event->u.class_load.class_name);
    while (tmp = (char*)strchr(name, '.')) {
        *tmp = '/';
    }

    hooked_class = lookup_hooked_class(&(this_thread->hooked_class_table), name);

    if (hooked_class == NULL) {
        if (verbose_mode > 1) {
            sprintf(info, "class will not be profiled : %s", name);
            jcov_info(info);
        }
        return;
    }
    
    class = (jcov_class_t*)jcov_calloc(sizeof(jcov_class_t));
    class->name = name;
    if (verbose_mode > 1) {
        sprintf(info, "CLASS_LOAD : %s", name);
        jcov_info(info);
    }
    
    class->id = event->u.class_load.class_id;
    class->num_methods = event->u.class_load.num_methods;
    class->unloaded = 0;
    mem = class->num_methods * sizeof(JVMPI_Method);
    class->methods = (JVMPI_Method*)jcov_calloc(mem);
    memcpy(class->methods, event->u.class_load.methods, mem);
    strcpy(class->timestamp, hooked_class->timestamp);
    class->src_name = jcov_strdup(hooked_class->src_name);
    class->access_flags = hooked_class->access_flags;
    class->data_type = hooked_class->data_type;

    for (i = 0; i < event->u.class_load.num_methods; i++) {
        class->methods[i].method_name = jcov_strdup(class->methods[i].method_name);
        class->methods[i].method_signature = jcov_strdup(class->methods[i].method_signature);
    }

    LOCK_CLASSES;
    found_class = lookup_class_by_key(class);
    if (found_class) {
        if (!(found_class->unloaded)) {
            if (verbose_mode > 0) {
                sprintf(info, "class is loaded twice : %s", found_class->name);
                jcov_error(info);
            }
            UNLOCK_CLASSES;
            return;
        } 

        CHK(found_class->num_methods != class->num_methods, 5);
        remove_class_by_id(found_class->id);
        found_class->id = class->id;
        for (i = 0; i < class->num_methods; i++) {
            meth = &(class->methods[i]);
            ind = find_method_in_class(found_class, meth);
            CHK(ind == -1, 6);
            LOCK_METHODS;
            found_method = lookup_method((found_class->methods[ind]).method_id);
            CHK(!found_method, 7);
            remove_method(found_method->id);
            found_method->id = meth->method_id;
            (found_class->methods[ind]).method_id = meth->method_id;
            put_method(&found_method);
            UNLOCK_METHODS;
        }
        put_class_by_id(&found_class);
        found_class->unloaded = 0;
        UNLOCK_CLASSES;
            
        jcov_free(class->methods);
        jcov_free(class->name);
        jcov_free(class->src_name);
        jcov_free(class);
        return;
    }

    put_class_by_id(&class);
    put_class_by_key(&class);
    UNLOCK_CLASSES;

    class_methods = hooked_class->method_cache;
    LOCK_METHODS;
    for (i = 0; i < class->num_methods; i++) {
        ind = array_lookup_method(&(class->methods[i]), class_methods, 
                                  hooked_class->methods_total, last_matched);
        CHK(ind == -1, 8);
        last_matched = ind;
        class_methods[ind]->id = class->methods[i].method_id;
        class_methods[ind]->class = class;
        CHK(lookup_method(class_methods[ind]->id), 9);
        put_method(&(class_methods[ind]));
    }
    UNLOCK_METHODS;

    remove_hooked_class(&(this_thread->hooked_class_table), hooked_class);
    jcov_free_hooked_class(hooked_class);
}

void jcov_class_unload_event(JVMPI_Event *event) {
    jcov_class_t *found_class;
    char info[MAX_PATH_LEN];

    LOCK_CLASSES;
    found_class = lookup_class_by_id(event->u.class_unload.class_id);
    if (!found_class) return;
    found_class->unloaded = 1;
    UNLOCK_CLASSES;
    if (verbose_mode > 0) {
        sprintf(info, "CLASS_UNLOAD : %s", found_class->name);
        jcov_info(info);
    }
}

void jcov_object_move_event(JVMPI_Event *event) {
    jcov_class_t *found_class;
    char info[MAX_PATH_LEN];

    LOCK_CLASSES;
    found_class = lookup_class_by_id(event->u.obj_move.obj_id);
    if (found_class) {
        remove_class_by_id(found_class->id);
        found_class->id = event->u.obj_move.new_obj_id;
        put_class_by_id(&found_class);
        if (verbose_mode > 0) {
            sprintf(info, "OBJECT_MOVE : class moved on heap: %s", found_class->name);
            jcov_info(info);
        }
    }
    UNLOCK_CLASSES;
}

void jcov_gc_start_event(void) {
    if (verbose_mode > 0) {
        jcov_info("GC_START");
    }
    LOCK_THREADS;
    LOCK_CLASSES;
    LOCK_METHODS;
}

void jcov_gc_finish_event(void) {
    if (verbose_mode > 0) {
        jcov_info("GC_FINISH");
    }
    UNLOCK_METHODS;
    UNLOCK_CLASSES;
    UNLOCK_THREADS;
}

static jcov_hooked_class_t *jcov_hooked_class_new(void) {
    jcov_hooked_class_t *res = (jcov_hooked_class_t*)jcov_calloc(sizeof(jcov_hooked_class_t));
    res->data_type = jcov_data_type;
    strcpy(res->timestamp, "0");
    return res;
}

static void jcov_thread_new(JNIEnv *thread_id) {
    jcov_thread_t *thread;

    thread = (jcov_thread_t*)jcov_calloc(sizeof(jcov_thread_t));
    thread->id = thread_id;
    jcov_hash_init(&(thread->hooked_class_table), 37,
                   hash_hooked_class, size_hooked_class, compare_hooked_classes);
    LOCK_THREADS;
    put_thread(&thread);
    UNLOCK_THREADS;
}

void jcov_thread_start_event(JVMPI_Event *event) {
    char info[MAX_PATH_LEN];
    
    if (verbose_mode > 0) {
        sprintf(info, "THREAD_START : %lx", event->u.thread_start.thread_env_id);
        jcov_info(info);
    }
    LOCK_THREADS;
    if (!lookup_thread(event->u.thread_start.thread_env_id)) {
        jcov_thread_new(event->u.thread_start.thread_env_id);
    }
    UNLOCK_THREADS;
}

void jcov_thread_end_event(JVMPI_Event *event) {
    jcov_thread_t *thread;
    char info[MAX_PATH_LEN];
    
    if (verbose_mode > 0) {
        sprintf(info, "THREAD_END : %lx", event->env_id);
        jcov_info(info);
    }
    LOCK_THREADS;
    thread = lookup_thread(event->env_id);
    CHK(!thread, 11);
    remove_thread(thread->id);
    jcov_free_thread(thread);
    UNLOCK_THREADS;
}

void jcov_method_entry_event(JVMPI_Event *event) {
    jcov_class_t *c;
    jobjectID id;
    jcov_method_t *m;
    char info[MAX_PATH_LEN];
    JVMPI_CallTrace call_trace;
    JVMPI_CallFrame call_frames[CALL_DEPTH];
    jmethodID method_id = event->u.method.method_id;
    jobjectID class_id;

    LOCK_METHODS;

    m = lookup_method0(method_id);
    if (m == NULL) {
        if (caller_filter == NULL) {
            UNLOCK_METHODS;
            return;
        }
        call_trace.frames = call_frames;
        call_trace.env_id = event->env_id;
        CALL(GetCallTrace)(&call_trace, CALL_DEPTH);
        if (method_id != call_frames[0].method_id)
            jcov_error_stop("event's method_id and GetCallTrace's method_id don't match");
        m = lookup_method(call_frames[1].method_id);
        if (m == NULL) {
            UNLOCK_METHODS;
            return;
        }
        if (string_suits_filter(caller_filter, m->class->name)) {
            m = lookup_method(method_id);
            if (m == NULL) {
                UNLOCK_METHODS;
                return;
            }
            put_method0(&m);
        } else {
            UNLOCK_METHODS;
            return;
        }
    }
    if (verbose_mode > 1) {
        sprintf(info, "METHOD_ENTRY : %s%s", m->name, m->signature);
        jcov_info(info);
    }

    if (m->covtable_size == 1 || jcov_data_type == JCOV_DATA_M) {
        m->covtable[0].count++;
    }
    UNLOCK_METHODS;
}

void jcov_jvm_shut_down_event() {
    save_jcov_data(jcov_file);
}


void jcov_events_init() {
    jcov_meth_table_lock = CALL(RawMonitorCreate)("jcov_meth_table_lock");
    jcov_thrd_table_lock = CALL(RawMonitorCreate)("jcov_thrd_table_lock");
    jcov_clss_table_lock = CALL(RawMonitorCreate)("jcov_clss_table_lock");
    is_hook_first = 1;
}
