/*
 * @(#)jcov_setup.c	1.6 01/04/21
 *
 * Copyright 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

#include <stdio.h>
#include <string.h>

#include "jvmpi.h"
#include "jcov.h"
#include "jcov_setup.h"
#include "jcov_error.h"
#include "jcov_htables.h"
#include "jcov_hash.h"
#include "jcov_events.h"
#include "jcov_util.h"
#include "jcov_file.h"

#define OPT_DELIM     ','
#define OPT_VAL_DELIM '='

char jcov_data_type;
char *jcov_file;
jcov_filter_t *class_filter;
jcov_filter_t *caller_filter;
int verbose_mode;
int include_abstracts;

static FILE *opt_file;
static char *cur_opt;

const char *COV_TABLE_ATTR_NAME;
const char *SRC_FILE_ATTR_NAME;
const char *ABS_SRC_PATH_ATTR_NAME;
const char *TIMESTAMP_ATTR_NAME;
const char *LINE_NUMBER_TABLE_ATTR_NAME;
const char *CODE_ATTR_NAME;

const char *OPT_HELP;
const char *OPT_INCLUDE;
const char *OPT_EXCLUDE;
const char *OPT_TYPE;
const char *OPT_FILE;
const char *OPT_CALLER_INCL;
const char *OPT_CALLER_EXCL;
const char *OPT_ABSTR_METH;
const char *OPT_VERBOSITY;
const char *OPT_OPTIONS_FILE;
const char *OPT_ON;

static void jcov_usage(void)
{
    jcov_close(&opt_file);
    printf("Jcov usage:\n");
    printf("\t-Xrunjcov[:help]|[<option>=<value>, ...]\n\n");
    printf("Where option is one of the following:\n");
    printf("\t%s=<class name prefix>\n", OPT_INCLUDE);
    printf("\t\tonly profile classes whose names start with the specified prefix\n");
    printf("\t%s=<class name prefix>\n", OPT_EXCLUDE);
    printf("\t\tdon't profile classes whose names start with the specified prefix\n");
    printf("\t%s=['B'|'M']\n", OPT_TYPE);
    printf("\t\tspecifies save file data format, default - B\n");
    printf("\t%s=<filename>\n", OPT_FILE);
    printf("\t\tspecifies savefile name, default - 'java.jcov'\n");
    printf("\t%s=<class name prefix>\n", OPT_CALLER_INCL);
    printf("\t\tgather coverage data only for methods invoked from the specified\n");
    printf("\t\tclass or classes beloning to the specified package;\n");
    printf("\t\tthe option can be used multiple times;\n");
    printf("\t\tclass or package names should use '.' as a package separator\n");
    printf("\t%s=<class name prefix>\n", OPT_CALLER_EXCL);
    printf("\t\tgather coverage data only for methods invoked *not* from the specified\n");
    printf("\t\tclass or classes *not* beloning to the specified package;\n");
    printf("\t\tthe option can be used multiple times;\n");
    printf("\t\tclass or package names should use '.' as a package separator\n");
    printf("\t%s=<on|off>\n", OPT_ABSTR_METH);
    printf("\t\tstore/do not store jcov data for abstract methods; default is 'off'\n");
    printf("\t%s=<filename>\n", OPT_OPTIONS_FILE);
    printf("\t\tspecifies where to read options from; (caller_)include/(caller_)exclude\n");
    printf("\t\toptions values read from the file are added to those, specified in the\n");
    printf("\t\tcommand line\n");
    printf("Example:\n");
    printf("\tjava -Xrunjcov:%s=java,%s=java.lang,", OPT_INCLUDE, OPT_EXCLUDE);
    printf("%s=java.io,%s=M,%s=sys.jcov\n", OPT_EXCLUDE, OPT_TYPE, OPT_FILE);

    CALL(ProfilerExit)((jint)0);
}
             
#define ERROR_AND_EXIT(msg) \
    jcov_close(&opt_file); \
    jcov_error_stop(msg)

static int read_line(char *buf, FILE *f)
{
    int ch;
    char *tmp = buf;

    do {
        ch = getc(f);
    } while (ch != EOF && (ch == '\n' || ch == '\r'));

    if (ch == EOF) {
	return 0;
    }
    while (ch != '\n' && ch != '\r' && ch != EOF) {
        if (tmp - buf >= MAX_PATH_LEN - 1) {
            ERROR_AND_EXIT("Too long option encountered");
        }
	*(tmp++) = (char)ch;
	ch = getc(f);
    }
    *(tmp++) = '\0';
    return 1;
}

static char *extract_option(char *opt_name, char *opt_val, char *src)
{
    char *tmp;

    tmp = strchr(src, OPT_VAL_DELIM);
    if (tmp == NULL) {
        printf("Invalid option : %s\n", src);
        jcov_usage();
    }
    strncpy(opt_name, src, tmp - src);
    opt_name[tmp - src] = '\0';
    src = tmp + 1;
    tmp = strchr(src, OPT_DELIM);
    if (tmp == NULL) {
        strcpy(opt_val, src);
        return src + strlen(src);
    } else {
        strncpy(opt_val, src, tmp - src);
        opt_val[tmp - src] = '\0';
        return tmp + 1;
    }
}

static int read_option_cmdline(char *opt_name, char *opt_val)
{
    if (strlen(cur_opt) == 0)
        return 0;
    cur_opt = extract_option(opt_name, opt_val, cur_opt);
    return 1;
}

static int read_option_disk(char *opt_name, char *opt_val)
{
    char buf[MAX_PATH_LEN*2];
    if (!read_line(buf, opt_file))
        return 0;
    extract_option(opt_name, opt_val, buf);
    return 1;
}

#define ADD_FILTER(filter, op) \
    if (filter == NULL) \
        filter = jcov_filter_new(); \
    while (tmp = (char*)strchr(opt_val, '.')) { \
        *tmp = '/'; \
    } \
    filter_add_##op(filter, opt_val)

static void parse_options(char *options)
{
    char *tmp;
    int (*read_option)(char *, char *);
    char opt_name[MAX_PATH_LEN];
    char opt_val[MAX_PATH_LEN];
    int i;

    if (options == NULL)
        return;

    if ((strcmp(options, OPT_HELP)) == 0) {
        jcov_usage();
    }
    
    cur_opt = options;
    read_option = read_option_cmdline;
    
    while (1) {
        if (read_option(opt_name, opt_val) == 0) {
            jcov_close(&opt_file);
            if (read_option == read_option_disk)
                read_option = read_option_cmdline;
            else
                break;
        } else if (strcmp(opt_name, OPT_OPTIONS_FILE) == 0) {
            if ((opt_file = fopen(opt_val, "rb")) == NULL) {
                sprintf(opt_name,"cannot open file : %s\n", opt_val);
                jcov_error_stop(opt_name);
            }
            read_option = read_option_disk;
        } else if (strcmp(opt_name, OPT_FILE) == 0) {
            jcov_file = (char*)jcov_calloc(strlen(opt_val) + 1);
            strcpy(jcov_file, opt_val);
        } else if (strcmp(opt_name, OPT_TYPE) == 0) {
            jcov_data_type = opt_val[0];
            if (jcov_data_type == 'M' || jcov_data_type == 'm')
                jcov_data_type = JCOV_DATA_M;
            else if (jcov_data_type != 'B' && jcov_data_type != 'b') {
                printf("Invalid data type : %c\n", jcov_data_type);
                jcov_usage();
            }
        } else if (strcmp(opt_name, OPT_ABSTR_METH) == 0) {
            if (strcmp(opt_val, OPT_ON) == 0)
                include_abstracts = 1;
        } else if (strcmp(opt_name, OPT_VERBOSITY) == 0) {
            if (strlen(opt_val) > 1) {
		jcov_usage();
	    }
	    i = opt_val[0] - '0';
	    if (i < 0 || i > 9) {
		jcov_usage();
	    }
	    verbose_mode = i;
        } else if (strcmp(opt_name, OPT_INCLUDE) == 0) {
            ADD_FILTER(class_filter, incl);
        } else if (strcmp(opt_name, OPT_EXCLUDE) == 0) {
            ADD_FILTER(class_filter, excl);
        } else if (strcmp(opt_name, OPT_CALLER_INCL) == 0) {
            ADD_FILTER(caller_filter, incl);
        } else if (strcmp(opt_name, OPT_CALLER_EXCL) == 0) {
            ADD_FILTER(caller_filter, excl);
        } else {
            printf("Unrecognized option : %s\n", opt_name);
            jcov_usage();
        }
    }
}

void jcov_init(char *jcov_options)
{
    int i;
    jcov_data_type = JCOV_DATA_B;
    jcov_file                   = "java.jcov";
    COV_TABLE_ATTR_NAME         = "CoverageTable";
    SRC_FILE_ATTR_NAME          = "SourceFile";
    ABS_SRC_PATH_ATTR_NAME      = "AbsoluteSourcePath";
    TIMESTAMP_ATTR_NAME         = "TimeStamp";
    LINE_NUMBER_TABLE_ATTR_NAME = "LineNumberTable";
    CODE_ATTR_NAME              = "Code";

    OPT_HELP         = "help";
    OPT_INCLUDE      = "include";
    OPT_EXCLUDE      = "exclude";
    OPT_TYPE         = "type";
    OPT_FILE         = "file";
    OPT_CALLER_INCL  = "caller_include";
    OPT_CALLER_EXCL  = "caller_exclude";
    OPT_ABSTR_METH   = "abstract_methods";
    OPT_VERBOSITY    = "verbosity";
    OPT_OPTIONS_FILE = "options_file";
    OPT_ON           = "on";

    class_filter = NULL;
    caller_filter = NULL;
    verbose_mode = 0;
    include_abstracts = 0;
    opt_file = NULL;

    parse_options(jcov_options);
    
    filter_finalize(class_filter);
    filter_finalize(caller_filter);

    jcov_htables_init();
    jcov_util_init();
    jcov_events_init();
    jcov_file_init();
}
