/*
 * @(#)jcov_util.c	1.5 01/04/21
 *
 * Copyright 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

#include <stdlib.h>
#include <string.h>
#include "jvm.h"
#include "jvmpi.h"
#include "jcov.h"
#include "jcov_error.h"
#include "jcov_htables.h"
#include "jcov_hash.h"
#include "jcov_util.h"
#include "jcov_setup.h"

const char *DUMMY_SRC_PREF;
const char *DUMMY_SRC_SUFF;

void *jcov_calloc(unsigned int size)
{
    void *p;
    p = (void *)malloc(size);
    if (p == NULL) {
        jcov_error("***out of memory***");
	CALL(ProfilerExit)((jint)1);
    }
    memset(p, 0, size);
    return p;
}

void jcov_dummy_free(void *p) {
    printf("free : %lx\n", (unsigned long)p);
    return;
}

unsigned char read1bytes(unsigned char **buf, jint *len, int *err_code)
{
    *len = *len - 1;
    if (*err_code = (*len < 0)) return 0;
    (*buf)++;
    return (*buf-1)[0];
}

int read2bytes(unsigned char **buf, jint *len, int *err_code)
{
    long res;
    unsigned char *buf1 = *buf;
    *len = *len - 2;
    if (*err_code = (*len < 0)) return 0;
    res = (buf1[0] << 8) + buf1[1];
    *buf += 2;
    return res;
}

long read4bytes(unsigned char **buf, jint *len, int *err_code)
{
    long res;
    unsigned char *buf1 = *buf;
    *len = *len - 4;
    if (*err_code = (*len < 0)) return 0;
    res = (buf1[0] << 24) + (buf1[1] << 16) + (buf1[2] << 8) + buf1[3];
    *buf += 4;
    return res;
}

#define CHK_UTF(cond) if (cond) { \
    if (verbose_mode > 0) jcov_error("bad UTF string"); \
    *err_code = 1; \
    return NULL; \
}

char *readUTF8(unsigned char **buf, jint *len, int utflen, int *err_code)
{
    char *str = (char*)jcov_calloc(utflen + 1);
    int count = 0;
    int strlen = 0;
    unsigned char *sav_buf = *buf;
    jint sav_len = *len;

    while (count < utflen) {
        char char2, char3;
        char c = read1bytes(buf, len, err_code);
        if (*err_code) return NULL;
        switch (c >> 4) {
        default:
            /* 0xxx xxxx */
            count++;
            str[strlen++] = c;
            break;
        case 0xC: case 0xD:
            /* 110x xxxx   10xx xxxx */
            count += 2;
            CHK_UTF(count > utflen);
            char2 = read1bytes(buf, len, err_code);
            CHK_UTF((char2 & 0xC0) != 0x80);
            str[strlen++] = '?';
            break;
        case 0xE:
            /* 1110 xxxx  10xx xxxx  10xx xxxx */
            count += 3;
            CHK_UTF(count > utflen);
            char2 = read1bytes(buf, len, err_code);
            char3 = read1bytes(buf, len, err_code);
            CHK_UTF(((char2 & 0xC0) != 0x80) || ((char3 & 0xC0) != 0x80));
            str[strlen++] = '?';
            break;
        case 0x8: case 0x9: case 0xA: case 0xB: case 0xF:
            /* 10xx xxxx,  1111 xxxx */
            CHK_UTF(1);
        }
        if (*err_code) return NULL;
    }
    str[strlen] = '\0';
    return str;
}

#ifdef DEBUG

#define def_name(e,v) case e: v=#e ; break
static char *get_cp_entry_type_name(int tag) {
    char* s = "UNKNOWN_TYPE";
    switch (tag) {
        def_name(JVM_CONSTANT_Utf8, s);
        def_name(JVM_CONSTANT_Class, s);
        def_name(JVM_CONSTANT_Methodref, s);
        def_name(JVM_CONSTANT_InterfaceMethodref, s);
        def_name(JVM_CONSTANT_NameAndType, s);
        def_name(JVM_CONSTANT_Integer, s);
        def_name(JVM_CONSTANT_Float, s);
        def_name(JVM_CONSTANT_Long, s);
        def_name(JVM_CONSTANT_Double, s);
        def_name(JVM_CONSTANT_String, s);
        def_name(JVM_CONSTANT_Fieldref, s);
    }
    return s;
}

void print_cp_entry(cp_entry_t *cp_entry) {
    if (cp_entry == NULL) {
        printf("NULL\n");
        return;
    }
    printf("%s ", get_cp_entry_type_name(cp_entry->tag));
    switch (cp_entry->tag) {
    case JVM_CONSTANT_Utf8:
        printf("\"%s\"\n", cp_entry->u.Utf8.bytes);
        break;
    case JVM_CONSTANT_Class:
        printf("#%d\n", cp_entry->u.Class.name_index);
        break;
    case JVM_CONSTANT_Methodref:
        printf("class: #%d, name_and_type: #%d\n",
               cp_entry->u.Methodref.class_index,
               cp_entry->u.Methodref.name_and_type_index);
        break;
    case JVM_CONSTANT_InterfaceMethodref:
        printf("class: #%d, name_and_type: #%d\n",
               cp_entry->u.InterfaceMethodref.class_index,
               cp_entry->u.InterfaceMethodref.name_and_type_index);
        break;
    case JVM_CONSTANT_NameAndType:
        printf("name: #%d, type: #%d\n",
               cp_entry->u.NameAndType.name_index,
               cp_entry->u.NameAndType.type_index);
        break;
    case JVM_CONSTANT_Integer:
    case JVM_CONSTANT_Float:
            /* we're not interested in the rest of tags */
            printf("0x%x\n", cp_entry->u.Integer.val);
            break;
    case JVM_CONSTANT_Long:
    case JVM_CONSTANT_Double:
        printf("hi: 0x%x, lo: 0x%x\n", cp_entry->u.Long.hi, cp_entry->u.Long.lo);
        break;
    case JVM_CONSTANT_String:
        printf("#%d\n", cp_entry->u.String.cp_index);
        break;
    case JVM_CONSTANT_Fieldref:
        printf("class: #%d, name_and_type: #%d\n",
               cp_entry->u.Field.class_index,
               cp_entry->u.Field.name_and_type_index);
        break;
    }
}

void print_cp(cp_entry_t **cp, int size) {
    int i;
    for (i = 0; i < size; i++) {
        printf("#%d ", i+1);
        print_cp_entry(cp[i]);
    }
}

#endif /* DEBUG */

void jcov_free_cp_entry(cp_entry_t *cp_entry);
cp_entry_t *read_next_cp_entry(unsigned char **buf, jint *len, int *err_code)
{
    int utflen;
    cp_entry_t *cp_buf = (cp_entry_t *)jcov_calloc(sizeof(cp_entry_t));
    cp_buf->tag = read1bytes(buf, len, err_code);
    if (!*err_code) {
        switch (cp_buf->tag) {
        case JVM_CONSTANT_Utf8:
            utflen = read2bytes(buf, len, err_code);
            if (*err_code) break;
            cp_buf->u.Utf8.bytes = readUTF8(buf, len, utflen, err_code);
            break;
        case JVM_CONSTANT_Class:
            cp_buf->u.Class.name_index = read2bytes(buf, len, err_code);
            break;
        case JVM_CONSTANT_Methodref:
            cp_buf->u.Methodref.class_index = read2bytes(buf, len, err_code);
            if (*err_code) break;
            cp_buf->u.Methodref.name_and_type_index = read2bytes(buf, len, err_code);
            break;
        case JVM_CONSTANT_InterfaceMethodref:
            cp_buf->u.InterfaceMethodref.class_index = read2bytes(buf, len, err_code);
            if (*err_code) break;
            cp_buf->u.InterfaceMethodref.name_and_type_index = read2bytes(buf, len, err_code);
            break;
        case JVM_CONSTANT_NameAndType:
            cp_buf->u.NameAndType.name_index = read2bytes(buf, len, err_code);
            if (*err_code) break;
            cp_buf->u.NameAndType.type_index = read2bytes(buf, len, err_code);
            break;
        case JVM_CONSTANT_Integer:
        case JVM_CONSTANT_Float:
            /* we're not interested in the rest of tags */
#ifndef DEBUG
            read4bytes(buf, len, err_code);
#else
            cp_buf->u.Integer.val = read4bytes(buf, len, err_code);
#endif
            break;
        case JVM_CONSTANT_Long:
        case JVM_CONSTANT_Double:
#ifndef DEBUG
            read4bytes(buf, len, err_code);
            if (*err_code) break;
            read4bytes(buf, len, err_code);
#else
            cp_buf->u.Long.hi = read4bytes(buf, len, err_code);
            if (*err_code) break;
            cp_buf->u.Long.lo = read4bytes(buf, len, err_code);
#endif
            break;
        case JVM_CONSTANT_String:
#ifndef DEBUG
            read2bytes(buf, len, err_code);
#else
            cp_buf->u.String.cp_index = read2bytes(buf, len, err_code);
#endif
            break;
        case JVM_CONSTANT_Fieldref:
#ifndef DEBUG
            read4bytes(buf, len, err_code);
#else
            cp_buf->u.Field.class_index = read2bytes(buf, len, err_code);
            if (*err_code) break;
            cp_buf->u.Field.name_and_type_index = read2bytes(buf, len, err_code);
#endif
            break;
        default:
            jcov_error("unrecognized constant pool entry encountered");
            *err_code = 1;
        }
    }
    if (*err_code) {
        jcov_free_cp_entry(cp_buf);
        cp_buf = NULL;
    }
    return cp_buf;
}

void jcov_free_cp_entry(cp_entry_t *cp_entry) {
    if (cp_entry != NULL) {
	if (cp_entry->tag == JVM_CONSTANT_Utf8) {
	    jcov_free(cp_entry->u.Utf8.bytes);
	}
        jcov_free(cp_entry);
    }
}

void jcov_free_constant_pool(cp_entry_t **cp, int cp_count) {
    int i;
    if (*cp == NULL) return;
    for (i = 1; i < cp_count; i++) {
	jcov_free_cp_entry(cp[i]);
    }
    jcov_free(cp);
}

void jcov_free_method(jcov_method_t *meth) {
    if (meth == NULL) return;
    jcov_free(meth->name);
    jcov_free(meth->signature);
    jcov_free(meth->covtable);
    jcov_free(meth->pc_cache);
    jcov_free(meth);
}

void jcov_free_hooked_class(jcov_hooked_class_t *c) {
    jcov_free(c->name);
    jcov_free(c->src_name);
    jcov_free(c->method_cache);
    jcov_free(c);
}

void jcov_free_hooked_class_full(jcov_hooked_class_t *c) {
    int i;
    jcov_free(c->name);
    jcov_free(c->src_name);
    for (i = 0; i < c->methods_total; i++) {
        jcov_free_method(c->method_cache[i]);
    }
    jcov_free(c->method_cache);
    jcov_free(c);
}


static void *f(void *c, void *arg) {
    jcov_free_hooked_class((jcov_hooked_class_t*)c);
    return NULL;
}

void jcov_free_thread(jcov_thread_t *thread) {
    void *dummy = NULL;
    jcov_hash_iterate(&(thread->hooked_class_table), f, dummy);
    jcov_free(thread);
}

char *jcov_strdup(const char *str)
{
    char *res = jcov_calloc(strlen(str)+1);
    strcpy(res, str);
    return res;
}

static int bit_flags[12];
static int lengths[12];
static char *string_flags[12];

void decode_modifiers(char *buf, int access_flags)
{
    char *ptr = buf;
    int i = 0;
    for (i = 0; i < 12; i++) {
	if (access_flags & bit_flags[i]) {
	    memcpy(ptr, string_flags[i], lengths[i]);
	    ptr += lengths[i];
	    *ptr = ' ';
	    ptr++;
	}
    }
    if (ptr == buf) {
	ptr++;
    }
    *(ptr-1) = '\0';
}

char *dummy_src_name(char *class_name)
{
    char *res;
    int len = strlen(class_name) + strlen(DUMMY_SRC_PREF) + strlen(DUMMY_SRC_SUFF) + 1;
    res = (char*)jcov_calloc(len);
    sprintf(res, "%s%s%s", DUMMY_SRC_PREF, class_name, DUMMY_SRC_SUFF);
    res[len] = '\0';
    return res;
}

static int my_strcmp(const void *s1, const void *s2)
{
	return strcmp(*(char**)s1, *(char**)s2);
}

static int starts_with(const void *str1, const void *str2)
{
    return strncmp(*(char**)str1, *(char**)str2, strlen(*(char**)str2));
}

jcov_filter_t *jcov_filter_new()
{
    jcov_filter_t *res = (jcov_filter_t*)jcov_calloc(sizeof(jcov_filter_t));
    res->incl = (char**)jcov_calloc(sizeof(char*) * MAX_FILTER_SIZE);
    res->excl = (char**)jcov_calloc(sizeof(char*) * MAX_FILTER_SIZE);
    return res;
}

void filter_add_incl(jcov_filter_t *filter, char *str)
{
    if (filter->incl_size >= MAX_FILTER_SIZE)
        jcov_error_stop("too many filters");
    filter->incl[filter->incl_size++] = jcov_strdup(str);
}

void filter_add_excl(jcov_filter_t *filter, char *str)
{
    if (filter->excl_size >= MAX_FILTER_SIZE)
        jcov_error_stop("too many filters");
    filter->excl[filter->excl_size++] = jcov_strdup(str);
}

void filter_finalize(jcov_filter_t *filter)
{
    if (filter == NULL)
        return;
    qsort(filter->incl, filter->incl_size, sizeof(char*), my_strcmp);
    qsort(filter->excl, filter->excl_size, sizeof(char*), my_strcmp);
}

void debug_print_filter(jcov_filter_t *filter)
{
    int i;

    printf("FILTER\n");
    if (filter == NULL)
        return;
    printf("INCL:\n");
    for (i = 0; i < filter->incl_size; i++)
        printf("\t%s\n", filter->incl[i]);
    printf("EXCL:\n");
    for (i = 0; i < filter->excl_size; i++)
        printf("\t%s\n", filter->excl[i]);
}

int string_suits_filter(jcov_filter_t *filter, char *str)
{
    if (filter == NULL)
        return 1;
    if (bsearch(&str, filter->excl, filter->excl_size, sizeof(char*), starts_with) != NULL)
        return 0;
    if (filter->incl_size == 0)
        return 1;
    if (bsearch(&str, filter->incl, filter->incl_size, sizeof(char*), starts_with) != NULL)
        return 1;
    return 0;
}

void jcov_close(FILE **f) {
    if (*f == NULL)
        return;
    if (fclose(*f) != 0) {
        jcov_error("cannot close file");
    }
    *f = NULL;
}

void jcov_remove(const char *filename) {
    char buf[MAX_PATH_LEN];

    if (remove(filename) == -1) {
        sprintf(buf,"cannot remove file : %s\n", filename);
        jcov_error(buf);
    }
}

void jcov_util_init(void)
{
    int i;

    bit_flags[0] = JVM_ACC_PUBLIC;
    bit_flags[1] = JVM_ACC_PRIVATE;
    bit_flags[2] = JVM_ACC_PROTECTED;
    bit_flags[3] = JVM_ACC_STATIC;
    bit_flags[4] = JVM_ACC_FINAL;
    bit_flags[5] = JVM_ACC_SYNCHRONIZED;
    bit_flags[6] = JVM_ACC_VOLATILE;
    bit_flags[7] = JVM_ACC_TRANSIENT;
    bit_flags[8] = JVM_ACC_NATIVE;
    bit_flags[9] = JVM_ACC_INTERFACE;
    bit_flags[10] = JVM_ACC_ABSTRACT;
    bit_flags[11] = JVM_ACC_STRICT;
	
    string_flags[0] = "public";
    string_flags[1] = "private";
    string_flags[2] = "protected";
    string_flags[3] = "static";
    string_flags[4] = "final";
    string_flags[5] = "synchronized";
    string_flags[6] = "volatile";
    string_flags[7] = "transient";
    string_flags[8] = "native";
    string_flags[9] = "interface";
    string_flags[10] = "abstract";
    string_flags[11] = "strictfp";

    for (i = 0; i < 12; lengths[i++] = strlen(string_flags[i]));

    DUMMY_SRC_PREF = "<UNKNOWN_SOURCE/";
    DUMMY_SRC_SUFF = ".java>";
}
