/*
 * @(#)MDataTransferer.java	1.4 01/02/09
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.motif;

import java.awt.Toolkit;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.FlavorMap;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;

import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;

import java.rmi.MarshalledObject;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sun.awt.DataTransferer;
import sun.awt.SunToolkit;

/**
 * Provides a rich set of functions for the transfer of data via the
 * Clipboard and Drag & Drop.
 * 
 * The concept of "flavors" and "natives" is extended to include "formats",
 * which are the numeric values X11 uses to express particular data types
 * otherwise known as "atoms".
 * Like FlavorMap, which provides getNativesForFlavors(DataFlavor[]) and
 * getFlavorsForNatives(String[]) functions, WDataTransferer provides a set
 * of getFormatsForFlavors and getFlavorsForFormats functions.
 *
 * Also provided are functions for translating a Transferable into a byte
 * array, given a source DataFlavor and a target format, and for translating
 * a byte array or InputStream into an Object, given a source format and
 * a target DataFlavor.
 *
 * @author Roger Brinkley
 * @version 1.4, 02/09/01
 *
 * @since 1.3.1
 */
public class MDataTransferer extends DataTransferer{
    static int StringAtom;
    static int TextAtom;
    static int FileNameAtom;
    static int DTNetFileAtom;

    static {
	// 4154170:  Need to ensure the the toolkit is initialized prior
	// to executing this initializer
	MToolkit toolkit = (MToolkit)Toolkit.getDefaultToolkit();
	StringAtom = getAtomForTarget("STRING");
	TextAtom = getAtomForTarget("TEXT");
	FileNameAtom = getAtomForTarget("FILE_NAME");
	DTNetFileAtom = getAtomForTarget("_DT_NETFILE");
    }

    /**
     * Singleton constructor
     */
    private MDataTransferer () {
    }

    private static MDataTransferer transferer;

    public static MDataTransferer getInstance() {
	if (transferer == null) {
	    synchronized(MDataTransferer.class) {
		if (transferer == null) {
		    transferer = new MDataTransferer();
		}
	    }
	}
	return transferer;
    }

    /**
     * Gets an atom for a format name. 
     */
    static native int getAtomForTarget(String name);

    /**
     * Gets an format name for a given format (atom)
     */
    private native String getTargetNameForAtom(int atom);

    /**
     * Returns a Map whose keys are all of the possible formats into which data
     * in the specified DataFlavors can be translated. The value of each key
     * is the DataFlavor in which the Transferable's data should be requested
     * when converting to the format.
     */
    public Map getFormatsForFlavors(DataFlavor[] flavors,
					   FlavorMap map) {
        Map nativeMap = map.getNativesForFlavors(flavors),
	    formatMap = new HashMap(flavors.length);

	// Iterate backwards so that preferred DataFlavors are used over
	// other DataFlavors. (See javadoc for
	// Transferable.getTransferDataFlavors.)
	for (int i = flavors.length - 1; i >= 0; i--) {
	    // Don't explicitly test for String, since it is just a special
	    // case of Serializable
	    if (flavors[i].isRepresentationClassInputStream() ||
		flavors[i].isRepresentationClassSerializable() ||
		flavors[i].isRepresentationClassRemote() ||
		DataFlavor.javaFileListFlavor.equals(flavors[i]))
	    {
	        // Since flavormap.properties only supports 1-to-1
	        // mappings, we need to be a bit more aggressive in
	        // exporting text/plain data. This allows StringSelection
	        // to function correctly, for example. We will probably
	        // be able to remove this code once the flavormap.
	        // properties restriction is relaxed.
                if ("text".equals(flavors[i].getPrimaryType()) &&
                    "plain".equals(flavors[i].getSubType())) {
		    formatMap.put(new Integer(StringAtom),
				  flavors[i]);
		    formatMap.put(new Integer(TextAtom),
				  flavors[i]);
		} 

		String str = (String)nativeMap.get(flavors[i]);
		if (str != null) {
		    // Just get the atom. If it has already been retrived
		    // once, we'll get a copy so this should be very fast.
		    int atom = getAtomForTarget(str);
		    if (atom > 0) {
			formatMap.put(new Integer(atom), flavors[i]);
		    }
		}
	    }
	}

	return formatMap;
    }

    /**
     * Returns a Map whose keys are all of the possible DataFlavors into which
     * data in the specified formats can be translated. The value of each key
     * is the format in which the Clipboard or dropped data should be requested
     * when converting to the DataFlavor.
     */
    public Map getFlavorsForFormats(int[] formats, FlavorMap map) {
        String[] natives = new String[formats.length];

	for (int i = 0; i < formats.length; i++) {
	    natives[i] = getTargetNameForAtom(formats[i]);
	}

	Map nativeMap = map.getFlavorsForNatives(natives),
	    flavorMap = new HashMap(nativeMap.size());

	for (int i = 0; i < natives.length; i++) {
	    DataFlavor flavor = (DataFlavor)nativeMap.get(natives[i]);
	    // Don't explicitly test for String, since it is just a special
	    // case of Serializable
	    if (flavor != null &&
		(flavor.isRepresentationClassInputStream() ||
		 flavor.isRepresentationClassSerializable() ||
		 flavor.isRepresentationClassRemote() ||
		 DataFlavor.javaFileListFlavor.equals(flavor)))
	    {
	        flavorMap.put(flavor, new Integer(formats[i]));
	    }
	}

	return flavorMap;
    }

    /**
     * Attempts to discern the encoding format given the specified
     * information.
     */
    protected String getEncoding(DataFlavor flavor, int format,
				      String sourceEncoding) {
        if (format == TextAtom) {
	    return DataTransferer.getDefaultTextCharset();
	} else if (format == StringAtom) {
	    return "ISO8859_1";
	} else if (sourceEncoding != null) {
	    return sourceEncoding;
	} else {
	    return DataTransferer.getTextCharset(flavor);
	}
    }

    protected byte[] translateTransferable(Transferable contents,
					   DataFlavor flavor,
					   int format,
					   String sourceEncoding)
	throws IOException {
	// Obtain the transfer data in the source DataFlavor.
        //
        // Note that we special case DataFlavor.plainTextFlavor because 
        // StringSelection supports this flavor incorrectly -- instead of 
        // returning an InputStream as the DataFlavor representation class
        // states, it returns a Reader. Instead of using this broken
        // functionality, we request the data in stringFlavor (the other
        // DataFlavor which StringSelection supports) and use the String
        // translator.
	Object obj;
	boolean plainTextFlavorHack;
	try {
            obj = contents.getTransferData(flavor);
            if (flavor.equals(DataFlavor.plainTextFlavor) &&
                !(obj instanceof InputStream)) {

                obj = contents.getTransferData(DataFlavor.stringFlavor);
                plainTextFlavorHack = true;
            } else {
                plainTextFlavorHack = false;
            }
	} catch (UnsupportedFlavorException e) {
	    throw new IOException(e.getMessage());
	}

	// Source data is a String
	//
	// We will hit this code if the DataFlavor has a primary MIME
	// type of "text" (such as text/plain, text/enriched, text/html),
	// and representation class is java.lang.String. This is actually
	// rather uncommon. (We will also hit this code if we trigger
	// the StringSelection special case described above.)
	//
	// Note: we will not trigger this code for DataFlavor.stringFlavor
	// because stringFlavor's primary MIME type is "application".
        if (plainTextFlavorHack ||
            (String.class.equals(flavor.getRepresentationClass()) &&
             "text".equals(flavor.getPrimaryType())))
        {
	    String str = (String)obj;

	    String targetEncoding =
	        getEncoding(flavor, format, sourceEncoding);

	    byte[] bytes = str.getBytes(targetEncoding);
	    return bytes;
	}

	ByteArrayOutputStream bos = new ByteArrayOutputStream();

	// Target data is a X11 file list. Source data must be a
	// java.util.List which contains java.io.File or String instances.
	if (format == FileNameAtom ||
	    format == DTNetFileAtom) {
	    if (!DataFlavor.javaFileListFlavor.equals(flavor)) {
	        throw new IOException("data translation failed");
	    }
	    final List list = (List)obj;
	    int nFiles = 0;
	    for (int i = 0; i < list.size(); i++) {
	        Object o = list.get(i);
		if (o instanceof File || o instanceof String) {
		    nFiles++;
		}
	    }
	    final String[] files = new String[nFiles];

	    try {
	        AccessController.doPrivileged(new PrivilegedExceptionAction() {
		    public Object run() throws IOException {
		        for (int i = 0, j = 0; i < list.size(); i++) {
			    Object o = list.get(i);
			    if (o instanceof File) {
			        files[j++] = ((File)list.get(i)).
				    getCanonicalPath();
			    } else if (o instanceof String) {
			        files[j++] = (String)o;
			    }
			}
			return null;
		    }
		});
	    } catch (PrivilegedActionException pae) {
	        throw new IOException(pae.getMessage());
	    }

	    for (int i = 0; i < files.length; i++) {
	        byte[] bytes = files[i].getBytes();
		bos.write(bytes, 0, bytes.length);
		bos.write(0);
	    }
	    bos.write(0);

	// Source data is an InputStream
	//
	// Just grab the bytes and dump them into a byte array.
        } else if (flavor.isRepresentationClassInputStream()) {
	    InputStream is = (InputStream)obj;
	    boolean eof = false;
	    int avail = is.available();
	    byte[] tmp = new byte[avail > 8192 ? avail : 8192];
	    do {
	        int ret;
		if (!(eof = (ret = is.read(tmp, 0, tmp.length)) == -1)) {
		    bos.write(tmp, 0, ret);
		}
	    } while (!eof);
            is.close();

            if ("text".equals(flavor.getPrimaryType())) {
                // Reencode according to the requested charset.
	        byte[] bytes = bos.toByteArray();
		bos.close();
		sourceEncoding = DataTransferer.getTextCharset(flavor);
		return translateTransferable(new StringSelection(
                    new String(bytes, sourceEncoding)),
                    DataFlavor.plainTextFlavor, format, sourceEncoding);
	    }

	// Source data is an RMI object
        } else if (flavor.isRepresentationClassRemote()) {
	    MarshalledObject mo = new MarshalledObject(obj);
	    ObjectOutputStream oos = new ObjectOutputStream(bos);
	    oos.writeObject(mo);
	    oos.close();

	// Source data is Serializable
        } else if (flavor.isRepresentationClassSerializable()) {
	    ObjectOutputStream oos = new ObjectOutputStream(bos);
	    oos.writeObject(obj);
	    oos.close();
	} else {
	    throw new IOException("data translation failed");
	}

	byte[] ret = bos.toByteArray();
	bos.close();
	return ret;
    }

    /**
     * Primary translation function for translating either a byte array or
     * an InputStream into an Object, given a source format and a target
     * DataFlavor.
     *
     * One of str/bytes is non-null; the other is null.
     * The conversion from byte[] to InputStream is cheap, so do that
     * immediately if necessary. The opposite conversion is expensive,
     * so avoid it if possible.
     */
    protected Object translateBytesOrStream(InputStream str, byte[] bytes,
					    DataFlavor flavor, int format)
        throws IOException
    {
        if (str == null) {
	    str = new ByteArrayInputStream(bytes);
	}

	// Source data is a X11 file list. Use the dragQueryFile native
	// function to do most of the decoding. Then wrap File objects around
	// the String filenames and return a List
        if (format == FileNameAtom ||
	    format == DTNetFileAtom) {
	    if (!DataFlavor.javaFileListFlavor.equals(flavor)) {
	        throw new IOException("data translation failed");
	    }
	    if (bytes == null) {
	        bytes = inputStreamToByteArray(str);
	    }

	    String[] filenames = dragQueryFile(bytes);

	    // Make sure that filenames has something in it.
	    // Otherwise just return.
	    if (filenames == null) {
	        str.close();
	        return null;
	    }

	    // Convert the strings to File objects
	    File[] files = new File[filenames.length];
	    for (int i = 0; i < filenames.length; i++) {
	        files[i] = new File(filenames[i]);
	    }
	    str.close();
	    
	    // Turn the list of Files into a List and return
	    return Arrays.asList(files);
        }

	// Target representation is a String
	//
	// We will hit this code if the DataFlavor has a primary MIME
	// type of "text" (such as text/plain, text/enriched, text/html),
	// and representation class is java.lang.String. This is actually
	// rather uncommon.
	//
	// Note: we will not trigger this code for DataFlavor.stringFlavor
	// because stringFlavor's primary MIME type is "application".
        if (String.class.equals(flavor.getRepresentationClass()) &&
            "text".equals(flavor.getPrimaryType()))
        {
	    // A String holds all of its data in memory at one time, so
	    // we can't avoid reading the entire InputStream at this point.
	    if (bytes == null) {
	        bytes = inputStreamToByteArray(str);
	    }
	    str.close();

	    String sourceEncoding = getEncoding(flavor, format, null);

	    return new String(bytes, sourceEncoding);
	}

	// Target flavor is an InputStream
	//
	// We support representations which are exactly InputStreams, and
	// also arbitrary Objects which have a constructor which takes an
	// InputStream as its sole parameter.
	if (flavor.isRepresentationClassInputStream()) {
            if ("text".equals(flavor.getPrimaryType())) {
                // Yuck. We need to decode the bytes so that we can
                // reencode before we return them.

                if (bytes == null) {
                    bytes = inputStreamToByteArray(str);
                }
                str.close();

                String sourceEncoding = getEncoding(flavor, format, null);
                String targetEncoding = DataTransferer.getTextCharset(flavor);

                String string = new String(bytes, sourceEncoding);
                bytes = string.getBytes(targetEncoding);
                str = new ByteArrayInputStream(bytes);

                return str;
	    }
            
	    final Class dfrc = flavor.getRepresentationClass();

            if (InputStream.class.equals(dfrc)) {
	        return str; // simple case
            } else {
                Constructor[] constructors = null;

                try {
                    constructors = (Constructor[])
		        AccessController.doPrivileged(new PrivilegedAction() {
                            public Object run() {
                                return dfrc.getConstructors();
                            }
                        });
                } catch (SecurityException se) {
                    throw new IOException(se.getMessage());
                }

                Constructor constructor = null;

                for (int j = 0; j < constructors.length; j++) {
                    if (!Modifier.isPublic(constructors[j].getModifiers())) {
		        continue;
		    }

                    Class[] ptypes = constructors[j].getParameterTypes();

                    if (ptypes != null && ptypes.length == 1 &&
			    InputStream.class.equals(ptypes[0])) {
                        constructor = constructors[j];
                        break;
                    }
                }

                if (constructor == null) {
		    throw new IOException("can't find <init>(Ljava/io/InputStream;)V for class: " + dfrc.getName());
		}

                try {
		    return constructor.newInstance(new Object[] { str } );
                } catch (Exception e) {
                    throw new IOException(e.getMessage());
                }
	    }

	// Target flavor is an RMI object
	} else if (flavor.isRepresentationClassRemote()) {
	    try {
	        ObjectInputStream ois = new ObjectInputStream(str);
		Object ret = ((MarshalledObject)(ois.readObject())).get();
                ois.close();
		str.close();
                return ret;
	    } catch (Exception e) {
	        throw new IOException(e.getMessage());
	    }

	// Target flavor is Serializable
	} else if (flavor.isRepresentationClassSerializable()) {
	    try {
	        ObjectInputStream ois = new ObjectInputStream(str);
		Object ret = ois.readObject();
		ois.close();
                str.close();
                return ret;
	    } catch (Exception e) {
	        throw new IOException(e.getMessage());
	    }
	}
	    
	throw new IOException("data translation failed");
    }

    /**
     * Decodes a byte array into a set of String filenames.
     */
    private native String[] dragQueryFile(byte[] bytes);

}


