/*
 * @(#)PSPaperSize.java	1.9 00/02/02
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.motif;

import java.awt.Dimension;
import java.awt.PageAttributes.MediaType;
import sun.awt.DebugHelper;

/**
 * PSPaperSize is an object that encapsulates standard paper sizes
 * as described in the Adobe PostScript Printer Description specification
 *
 * @author Richard Berlin
 */
class PSPaperSize {
    String paperSizeName;
    Dimension paperSize;
    int tolerance;

    /**
     * Returns the name associated with this paper size
     */
    public String name() {
	return paperSizeName;
    }

    /**
     * Returns the width and height of this paper size, in points,
     * as a Dimension object
     */
    public Dimension size() {
	return paperSize;
    }

    /**
     * Returns the width of this paper size, in points
     */
    public int width() {
        return paperSize.width;
    }

    /**
     * Returns the height of this paper size, in points
     */
    public int height() {
	return paperSize.height;
    }

    /**
     * Returns a String containing PostScript code which defines
     * initClipName to be a procedure which initializes the clip
     * region, and then sets the region to a margin of margin
     * points in the default coordinate system. Generates Level
     * 2 code.
     */
    public String toInitClip(String initClipName, int margin) {
        return
	      "mark {\n"
	    + "  currentpagedevice /PageSize get\n"
	    + "  dup 0 get /" + initClipName + "W exch def\n"
	    + "  1 get /" + initClipName + "H exch def\n"
	    + "} stopped {\n"
	    + "  /" + initClipName + "W " + width() + " def\n"
	    + "  /" + initClipName + "H " + height() + " def\n"
	    + "} if cleartomark\n"
	    + "/" + initClipName + " {\n"
	    + "  initclip matrix currentmatrix initmatrix\n"
	    + "  " + margin + " " + margin + " moveto\n"
	    + "  " + margin + " " + initClipName + "H " + margin + " sub lineto\n"
	    + "  " + initClipName + "W " + margin + " sub " + initClipName + "H " + margin + " sub lineto\n"
	    + "  " + initClipName + "W " + margin + " sub " + margin + " lineto\n"
	    + "  closepath clip setmatrix newpath\n"
	    + "} bind def";
    }

    /**
     * Returns a String containing PostScript code to select this
     * paper size on a printer.
     *
     * @param PSLevel (1 or 2) if Level 2, emits much less code, but will
     * not work correctly on a printer which does not support PostScript
     * Level 2
     * @param manualFeed if true, request manual feed paper.  On level 1
     * interpreters this generates a different paper size request.
     */
    public String toPostScriptCode(int PSLevel, boolean manualFeed) {
        if (PSLevel == 2) {
            return 
	          "mark {\n"
	        + "%%BeginFeature: *ManualFeed "
	        + ((manualFeed) ? "True" : "False") + "\n"
	        + "<< /ManualFeed " + manualFeed + "\n"
	        + "   /Policies << /ManualFeed 1 >>\n"
	        + ">> setpagedevice\n"
	        + "%%EndFeature\n"
	        + "%%BeginFeature: *PageSize " + paperSizeName + "\n"
	        + "<< /PageSize [" + width() + " " + height() + "]\n"
	        + "   /ImagingBBox null\n"
	        + "   /Policies << /PageSize 3 >>\n"
	        + ">> setpagedevice\n"
	        + "%%EndFeature\n"
	        + "} stopped cleartomark";
        } else {
            String trayCmd = paperSizeName.toLowerCase();
            if (!manualFeed) trayCmd = trayCmd + "tray";

            return
                  "mark {\n"
                + "%%BeginFeature: ManualFeed "
	        + ((manualFeed) ? "True" : "False") + "\n"
                + "1 dict dup /ManualFeed " + manualFeed + " put setpagedevice\n"
                + "%%EndFeature\n"
                + "} stopped {\n"
                + "%%BeginFeature: *ManualFeed "
                + ((manualFeed) ? "True" : "False") + "\n"
                + "  statusdict /manualfeed " + manualFeed + " put\n"
                + "%%EndFeature\n"
                + "} if cleartomark mark {\n"
                + "%%BeginFeature: PageSize *" + paperSizeName + "\n"
                + "3 dict begin\n"
                + "   /PageSize [" + width() + " " + height() + "] def\n"
                + "   /ImagingBBox null def\n"
                + "   /Policies 1 dict dup/PageSize 3 put def\n"
                + " currentdict end setpagedevice\n"
                + "%%EndFeature\n"
                + "} stopped {\n"
                + "  cleartomark statusdict begin mark {\n"
                + "%%BeginFeature: *PageSize " + paperSizeName + "\n"
                + "  " + trayCmd + "\n"
                + "%%EndFeature\n"
                + "  } stopped end\n"
                + "} if cleartomark\n"
                + "% now make sure we got what we requested!\n"
                + "gsave initgraphics clippath pathbbox grestore\n"
                + "5 dict begin\n"
                + "  3 -1 roll /yoff exch def 3 -1 roll /xoff exch def\n"
                + "  yoff sub /h exch def xoff sub /w exch def\n"
                + "  w " + (paperSize.width-36) + " lt h "
                         + (paperSize.height-36) + " lt or {\n"
                + "    w 2 div xoff add h 2 div yoff add translate\n"
                + "    w xoff 2 mul add " + paperSize.width + " div\n"
                + "    h yoff 2 mul add " + paperSize.height + " div\n"
                + "    2 copy gt {exch} if pop dup scale\n"
                + "    " + (-paperSize.width/2) + " " + (-paperSize.height/2)
                     + " translate\n"
                + "  } if\n"
                + "end\n";
        }
    }

    /**
     * Returns a String containing PostScript code to select this
     * paper size on a printer.
     */
    public String toPostScriptCode() {
	return toPostScriptCode(1, false);
    }

    /**
     * Returns a String object representing the value of this PSPaperSize.
     */
    public String toString() {
	return "PSPaperSize[" + paperSizeName + ": "
	    + paperSize.width + "x" + paperSize.height + "]";
    }

    /**
     * Creates a new PSPaperSize with the given name, width and height.
     * @param name the name of this paper size
     * @param the width (in points) of this paper size
     * @param the height (in points) of this paper size
     */
    public PSPaperSize(String name, int width, int height) {
	this.paperSizeName = name;
	this.tolerance = 0;
	this.paperSize = new Dimension(width, height);
    }

    /**
     * Creates a new PSPaperSize with the given name, width and height.
     * Also specifies a matching tolerance; this is required for certain
     * "standard" paper sizes (e.g. "Executive") for which the standard
     * specification is approximate.
     *
     * @param name the name of this paper size
     * @param tolerance the matching tolerance for this paper size
     * @param width the width (in points) of this paper size
     * @param height the height (in points) of this paper size
     */
    public PSPaperSize(String name, int tolerance, int width, int height) {
	this.paperSizeName = name;
	this.tolerance = tolerance;
	this.paperSize = new Dimension(width, height);
    }

    private static final String PSSIZES[] = {
        "4A0", "2A0", "A0", "A1", "A2", "A3", "A4", "A5", "A6", "A7", "A8",
	"A9", "A10", "B0", "B1", "B2", "B3", "B4", "B5", "B6", "B7", "B8",
	"B9", "B10", "JIS.B0", "JIS.B1", "JIS.B2", "JIS.B3", "JIS.B4",
	"JIS.B5", "JIS.B6", "JIS.B7", "JIS.B8", "JIS.B9", "JIS.B10", "C0",
	"C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8", "C9", "C10",
	"ISO.Designated.Long", "Executive", "Folio", "Invoice",	"Ledger",
	"Letter", "Legal", "Quarto", "A", "B", "C", "D", "E", "10x15",
	"10x14", "10x13", "9x12", "9x11", "7x9", "6x9", "#9.Envelope",
	"#10.Envelope",	"#11.Envelope", "#12.Envelope", "#14.Envelope",
	"Invite.Envelope", "Italy.Envelope", "Monarch.Envelope",
	"Personal.Envelope"
    };

    private static final DebugHelper dbg =
        DebugHelper.create(PSPaperSize.class);

    static PSPaperSize standardSizeTable[];

    static {
        if (dbg.on) {
	    dbg.assert(PSSIZES.length ==
		       sun.awt.print.PrintControl.SIZES.length);
	}

        standardSizeTable =
	    new PSPaperSize[sun.awt.print.PrintControl.SIZES.length];
	int executiveIndex = MediaType.EXECUTIVE.hashCode();
        for (int i = 0; i < sun.awt.print.PrintControl.SIZES.length; i++) {
	    if (i == executiveIndex) {
	        // Special case for "Executive" because it has a 1/2 in
	        // tolerance
		standardSizeTable[i] =
		    new PSPaperSize(PSSIZES[i], 36, 540, 720);
	    } else {
	        standardSizeTable[i] =
		    new PSPaperSize(PSSIZES[i],
				    sun.awt.print.PrintControl.WIDTHS[i],
				    sun.awt.print.PrintControl.LENGTHS[i]);
	    }
	}
    }

    /**
     * Returns a standard paper size with the given name
     * @param name the name of the desired paper size
     */
    public static PSPaperSize find(String name) {
	if (name == null)
	    return null;

	for (int i = 0; i < standardSizeTable.length; ++i)
	    if (name.equalsIgnoreCase(standardSizeTable[i].paperSizeName))
		return(standardSizeTable[i]);

	return null;
    }

    /**
     * Returns a standard paper size whose width and height differ from the
     * specified width and height by no more than dMax in each dimension
     *
     * @param width the width of the desired paper size
     * @param height the height of the desired paper size
     * @param dMax a matching tolerance
     */
    public static PSPaperSize find(int width, int height, int dMax) {
	int dX = dMax + 1;
	int dY = dMax + 1;
	PSPaperSize closest = null;
    
	for (int i = 0; i < standardSizeTable.length; ++i) {
	    PSPaperSize p = standardSizeTable[i];

	    int dW = Math.abs(p.paperSize.width - width);
	    int dH = Math.abs(p.paperSize.height - height);

	    if ((dW <= p.tolerance) && (dH <= p.tolerance)) {
		return p;
	    }
	    else if ((dW <= dMax) && (dH <= dMax) && ((dW + dH) < (dX + dY))) {
		dX = dW;
		dY = dH;
		closest = p;
	    }
	}

	return(closest);
    }

    /**
     * Returns the standard paper size with the given width and height
     *
     * @param width the width of the desired paper size
     * @param height the height of the desired paper size
     */
    public static PSPaperSize find(int width, int height) {
	return find(width, height, 0);
    }
}
