/*
 * @(#)PSPrintControl.java	1.25 00/07/13
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.motif;

import java.awt.*;
import java.io.*;
import java.security.*;
import java.util.*;
import java.awt.JobAttributes.*;
import java.awt.PageAttributes.*;
import sun.awt.print.PrinterCapabilities;
import sun.awt.print.PrinterListUpdatable;

/**
 * A class which presents a modal print dialog to the user and
 * returns an appropriately configured print job.
 *
 * @version 	1.25 07/13/00
 * @author 	Amy Fowler
 * @author      David Mendenhall
 */
public class PSPrintControl extends sun.awt.print.PrintControl {

    private static ResourceBundle messageRB;
  
    private static String[] sortedPrinterList = null;
    private static long lastPrinterListQueryMillis = 0;
    private static final long QUERY_INTERVAL = 900000L; // 15 minutes

    // The following Strings are maintained for backward-compatibility with
    // Properties based print control.
    private final static String DEST_PROP = "awt.print.destination";
    private final static String PRINTER = "printer";
    private final static String FILE = "file";

    private final static String PRINTER_PROP = "awt.print.printer";

    private final static String FILENAME_PROP = "awt.print.fileName";

    private final static String NUMCOPIES_PROP = "awt.print.numCopies";

    private final static String OPTIONS_PROP = "awt.print.options";

    private final static String ORIENT_PROP = "awt.print.orientation";
    private final static String PORTRAIT = "portrait";
    private final static String LANDSCAPE = "landscape";

    private final static String PAPERSIZE_PROP = "awt.print.paperSize";
    private final static String LETTER = "letter";
    private final static String LEGAL = "legal";
    private final static String EXECUTIVE = "executive";
    private final static String A4 = "a4";

    /**
     * Initialize ResourceBundle
     */
    static void initResource() {
        AccessController.doPrivileged(new PrivilegedAction() {
	    public Object run() {
	        try {
		    messageRB = ResourceBundle.getBundle(
                        "sun.awt.motif.resources.printcontrol");
		    return null;
		} catch (MissingResourceException e) {
		    throw new Error("Fatal: Resource for PSPrintControl is "+
				    "missing.");
		}
	    }
	});
    }

    /**
     * get message string from resource
     */
    static synchronized String getMsg(String key) {
        if (messageRB == null) {
	    initResource();
	}
	try {
	    return messageRB.getString(key);
	} catch (MissingResourceException e) {
	    throw new Error("Fatal: Resource for PSPrintControl is broken. "+
			    "There is no " + key + " key in resource");
	}
    }

    private Hashtable procTable = new Hashtable();
    private sun.awt.print.PrintDialog dialog;
    private Properties props;

    public PSPrintControl(Frame dialogOwner, String doctitle,
			  JobAttributes jobAttributes,
			  PageAttributes pageAttributes) {
        super(dialogOwner, doctitle, jobAttributes, pageAttributes);
    }
    public PSPrintControl(Frame dialogOwner, String doctitle,
			  Properties props) {
        super(dialogOwner, doctitle, null, null);
	this.props = props;
	translateInputProps();
    }

    public PrintJob initJob() {
        PrintJob job = super.initJob();
	if (job != null) {
	    translateOutputProps();
	}
	return job;
    }

    public String getDefaultPrinterName() {
        String defaultPrinter = "lp";

        try {
	    final String[] cmd = new String[3];

            if(System.getProperty("os.name").equals("Linux")) {
	        cmd[0] = "/bin/sh";
	        cmd[1] = "-c";
	        cmd[2] = "/usr/sbin/lpc status | grep : | sed -ne '1,1 s/://p'";
            } else {
                cmd[0] = "/usr/bin/sh";
                cmd[1] = "-c";
                cmd[2] = "/usr/bin/lpstat -d|/usr/bin/expand|/usr/bin/cut -f4 -d' '";
            }
	    Process lpstat;
	    try {
	        lpstat = (Process)AccessController.doPrivileged(
                    new PrivilegedExceptionAction() {
		        public Object run() throws IOException {
		            return Runtime.getRuntime().exec(cmd);
			}
		});
	    } catch (PrivilegedActionException e) {
	        throw (IOException)e.getException();
	    }
	    lpstat.waitFor();
	    if (lpstat.exitValue() == 0) {
	        InputStream stream = lpstat.getInputStream();
		byte[] streamBytes = new byte[stream.available()];
		stream.read(streamBytes);
		defaultPrinter = new String(streamBytes).trim();
	    }
	} catch (IOException e) {
	    // just return "lp"
	} catch (InterruptedException e) {
	    // just return "lp"
	}
	return defaultPrinter;
    }

    public boolean getCapabilities(PrinterCapabilities capabilities) {
        capabilities.setCapabilities(PrinterCapabilities.COLOR |
				     PrinterCapabilities.DUPLEX);
	return true;
    }

    public void getPrinterList(final PrinterListUpdatable updatable) {
        long currentTimeMillis = System.currentTimeMillis();
	if (sortedPrinterList != null &&
	    currentTimeMillis - lastPrinterListQueryMillis < QUERY_INTERVAL) {
	    updatable.updatePrinterList(sortedPrinterList);
	    return;
	}

	lastPrinterListQueryMillis = currentTimeMillis;

        new Thread(new Runnable() {
	    public void run() {
	        String[] sortedNames = null;

		try {
		    final String[] cmd = new String[3];
                    if(System.getProperty("os.name").equals("Linux")) {
		        cmd[0] = "/bin/sh";
		        cmd[1] = "-c";
	                cmd[2] = "/usr/sbin/lpc status | grep : | sed -e 's/://'";
                    } else {
                        cmd[0] = "/usr/bin/sh";
                        cmd[1] = "-c";
                        cmd[2] = "/usr/bin/lpstat -v|/usr/bin/expand|/usr/bin/cut -f3 -d' '|/usr/bin/cut -f1 -d':'|/usr/bin/sort -u";
                    }
		    Process lpstat;
		    try {
		        lpstat = (Process)AccessController.doPrivileged(
			    new PrivilegedExceptionAction() {
		                public Object run() throws IOException {
				    return Runtime.getRuntime().exec(cmd);
				}
			});
		    } catch (PrivilegedActionException e) {
		        throw (IOException)e.getException();
		    }
		    lpstat.waitFor();
		    if (lpstat.exitValue() == 0) {
		        InputStream stream = lpstat.getInputStream();
			byte[] streamBytes;
			String stdout = "";
			while (stream.available() > 0) {
			    streamBytes = new byte[stream.available()];
			    stream.read(streamBytes);
			    stdout = stdout + new String(streamBytes);
			}
			StringTokenizer tokens =
			    new StringTokenizer(stdout);
			int total = tokens.countTokens();
			sortedNames = new String[total];
			for (int i = 0; i < total; i++) {
			    sortedNames[i] = tokens.nextToken();
			}
		    }
		}
		catch (IOException e) {
		    // just return null
		} catch (InterruptedException e) {
		    // just return null
		}

		sortedPrinterList = sortedNames;
		updatable.updatePrinterList(sortedPrinterList);
	    }
	}).start();
    }

    private void translateInputProps() {
        if (props == null) {
	    return;
	}

        String str;

	str = props.getProperty(DEST_PROP);
	if (str != null) {
	    if (str.equals(PRINTER)) {
	        setDestAttrib(DestinationType.PRINTER);
	    } else if (str.equals(FILE)) {
	        setDestAttrib(DestinationType.FILE);
	    }
	}
	str = props.getProperty(PRINTER_PROP);
	if (str != null) {
	    setPrinterAttrib(str);
	}
	str = props.getProperty(FILENAME_PROP);
	if (str != null) {
	    setFileNameAttrib(str);
	}
	str = props.getProperty(NUMCOPIES_PROP);
	if (str != null) {
	    setCopiesAttrib(Integer.parseInt(str));
	}
	str = props.getProperty(OPTIONS_PROP);
	if (str != null) {
	    setOptions(str);
	}
	str = props.getProperty(ORIENT_PROP);
	if (str != null) {
	    if (str.equals(PORTRAIT)) {
	        setOrientAttrib(OrientationRequestedType.PORTRAIT);
	    } else if (str.equals(LANDSCAPE)) {
	        setOrientAttrib(OrientationRequestedType.LANDSCAPE);
	    }
	}
	str = props.getProperty(PAPERSIZE_PROP);
	if (str != null) {
	    if (str.equals(LETTER)) {
	        setMediaAttrib(MediaType.LETTER.hashCode());
	    } else if (str.equals(LEGAL)) {
	        setMediaAttrib(MediaType.LEGAL.hashCode());
	    } else if (str.equals(EXECUTIVE)) {
	        setMediaAttrib(MediaType.EXECUTIVE.hashCode());
	    } else if (str.equals(A4)) {
	        setMediaAttrib(MediaType.A4.hashCode());
	    }
	}
    }
    private void translateOutputProps() {
        if (props == null) {
	    return;
	}

	String str;

	props.setProperty(DEST_PROP, 
            (getDestAttrib() == DestinationType.PRINTER) ? PRINTER : FILE);
	str = getPrinterAttrib();
	if (str != null && !str.equals("")) {
	    props.setProperty(PRINTER_PROP, getPrinterAttrib());
	}
	str = getFileNameAttrib();
	if (str != null && !str.equals("")) {
	    props.setProperty(FILENAME_PROP, str);
	}
	int copies = getCopiesAttrib();
	if (copies > 0) {
	    props.setProperty(NUMCOPIES_PROP, "" + copies);
	}
	str = getOptions();
	if (str != null && !str.equals("")) {
	    props.setProperty(OPTIONS_PROP, str);
	}
	props.setProperty(ORIENT_PROP,
	    (getOrientAttrib() == OrientationRequestedType.PORTRAIT)
	        ? PORTRAIT : LANDSCAPE);
	MediaType media = sun.awt.print.PrintControl.SIZES[getMediaAttrib()];
	if (media == MediaType.LETTER) {
	    str = LETTER;
	} else if (media == MediaType.LEGAL) {
	    str = LEGAL;
	} else if (media == MediaType.EXECUTIVE) {
	    str = EXECUTIVE;
	} else if (media == MediaType.A4) {
	    str = A4;
	} else {
	    str = media.toString();
	}
	props.setProperty(PAPERSIZE_PROP, str);
    }

    protected PrintJob createPrintJob() {
	OutputStream os = null;
	Process process = null;
	PSPrintJob returnJob;
	String[] execCmd = new String[1];
        String errorStr = null;

	if (getDestAttrib() == DestinationType.PRINTER) {
            try {
		execCmd = printExecCmd(getPrinterAttrib(), getOptions(),
				       doctitle);
		final String[] cmd = execCmd;
		try {
		    process = (Process)AccessController.doPrivileged(
		        new PrivilegedExceptionAction() {
                            public Object run() throws IOException {
			        return Runtime.getRuntime().exec(cmd);
			    }
                    });
		} catch (PrivilegedActionException e) {
		    throw (IOException)e.getException();
		}
		os = process.getOutputStream();
            } catch (IOException e) {
                errorStr = getMsg("error_msg.exec_print") + execCmd[0];
            }
	} else /* if (getDestAttrib() == DestinationType.FILE) */ {
            final File destFile = new File(getFileNameAttrib());
            String canonPath = "";
            try {
	        try {
		    canonPath = (String)AccessController.doPrivileged(
		        new PrivilegedExceptionAction() {
		            public Object run() throws IOException {
			        return destFile.getCanonicalPath();
			    }
		    });
		} catch (PrivilegedActionException e) {
		    throw (IOException)e.getException();
		}
                if (getDialogAttrib() != DialogType.NONE &&
		        destFile.exists()) {
                    String cancelCmd = getMsg("warning_button.cancel");
                    String overwriteCmd = getMsg("warning_button.overwrite");
                    sun.awt.print.PrintStatusDialog d = 
		        new sun.awt.print.PrintStatusDialog(dialog,
					      getMsg("warning_dialog.title"),
					      getMsg("warning_msg.file_exists")
                                                  + getFileNameAttrib(),
					      overwriteCmd,
                                              cancelCmd);
		    d.setLocationRelativeTo(dialog);
                    d.setVisible(true);
                    if (d.getCommand().equals(cancelCmd)) {
                        return null;
                    }
                }
		try {
		    os = (FileOutputStream)AccessController.doPrivileged(
                        new PrivilegedExceptionAction() {
		            public Object run() throws IOException {
			        return new FileOutputStream(destFile);
			    }
		    });
		} catch (PrivilegedActionException e) {
		    throw (IOException)e.getException();
		}
            } catch (IOException e) {
                errorStr = getMsg("error_msg.create_file") + canonPath;
            }
	}
        if (errorStr != null) {
	    if (getDialogAttrib() == DialogType.NONE) {
	        throw new InternalError(errorStr);
	    } else {
	        sun.awt.print.PrintStatusDialog d = 
		    new sun.awt.print.PrintStatusDialog(dialog,
					  getMsg("error_dialog.title"),
					  errorStr,
					  getMsg("error_button.ok"));
		d.setLocationRelativeTo(dialog);
		d.setVisible(true);
		return null;
	    }
	}

	Dimension paperDim =
	    PSPaperSize.standardSizeTable[getMediaAttrib()].size();
	int pageWidth = paperDim.width;
	int pageHeight = paperDim.height;
	int pageRes;

	if (getOrientAttrib() == OrientationRequestedType.LANDSCAPE) {
	    int temp = pageWidth;
	    pageWidth = pageHeight;
	    pageHeight = temp;
	    paperDim = new Dimension(pageWidth, pageHeight);
	}
	if (getOriginAttrib() == OriginType.PRINTABLE) {
	    pageWidth -= 36; // 1/4 inch left and right
	    pageHeight -= 36; // 1/4 inch top and bottom
	}
	pageRes = getResAttrib();
	if (pageRes != 72) {
	    pageWidth *= (pageRes / 72.0);
	    pageHeight *= (pageRes / 72.0);
	}

	returnJob = new PSPrintJob(new PrintStream(
				       new BufferedOutputStream(os), false),
				   (doctitle != null) ? doctitle : "",
				   new Dimension(pageWidth, pageHeight),
				   pageRes, paperDim, this);
	procTable.put(returnJob, (process == null) ? new NullProcess()
                                                   : process);
	return returnJob;
    }

    protected PrintJob createPrintJob(sun.awt.print.PrintDialog dialog) {
        this.dialog = dialog;
	PrintJob pjob = super.createPrintJob(dialog);
	this.dialog = null;
	return pjob;
    }

    public void finishedPrinting(PSPrintJob job) {
        Process proc = (Process) (procTable.get(job));
	new Thread(new CheckExitCodeRunnable(proc, this, dialogOwner)).start();
    }

    private String[] printExecCmd(String printer, String options, 
				  String banner) {
	int PRINTER = 0x1;
        int OPTIONS = 0x2;
	int BANNER  = 0x4;
	int NOBANNER = 0x8;
	int pFlags = 0;
	String execCmd[];
	int ncomps = 1;
	int n = 0;

	if (printer != null && !printer.equals("")) {
	    pFlags |= PRINTER;
	    ncomps+=1;
	}
	if (options != null && !options.equals("")) {
	    pFlags |= OPTIONS;
	    ncomps+=1;
	}
	if (banner != null && !banner.equals("")) {
	    pFlags |= BANNER;
	    ncomps+=1;
	} else if (banner == null) {
	    pFlags |= NOBANNER;
	    ncomps+=1;
	}
	execCmd = new String[ncomps];
	if (System.getProperty("os.name").equals("Linux")) {
	    execCmd[n++] = "/usr/bin/lpr";
	    if ((pFlags & PRINTER) != 0) {
		execCmd[n++] = new String("-P" + printer);
	    }
	    if ((pFlags & BANNER) != 0) {
		execCmd[n++] = new String("-T"  + escapeString(banner));
	    }
	} else {
	    execCmd[n++] = "/usr/bin/lp";
	    if ((pFlags & PRINTER) != 0) {
		execCmd[n++] = new String("-d" + printer);
	    }
	    if ((pFlags & BANNER) != 0) {
		execCmd[n++] = new String("-t" + escapeString(banner));
	    }
	    if ((pFlags & OPTIONS) != 0) {
		    // multiple options currently don't work because /usr/bin/lp
		    // (Solaris 2.6) doesn't handle them correctly
		execCmd[n++] = new String("-o" + options);
	    }
	    if ((pFlags & NOBANNER) != 0) {
		execCmd[n++] = new String("-onobanner");
	    }
        }
        return execCmd;
    }


    /**
     * Adjusts a string for the two special characters which confuse the
     * print command: '\\' and '"'. An empirical check showed that no other
     * non-alphanumeric characters posed a problem, although not all of them
     * printed correctly (those that didn't were coverted to '_').
     */
    private String escapeString(String s) {
        StringTokenizer tokens = new StringTokenizer(s, "\"\\", true);
 	if (tokens.countTokens() == 1) {
 	    return s;
 	}

 	StringBuffer buf = new StringBuffer(s.length() * 2);
 	String token;
 	char c;

 	while (tokens.hasMoreTokens()) {
 	    token = tokens.nextToken();
 	    if (token.length() == 1) {
 	        c = token.charAt(0);
 		if (c == '"' || c == '\\') {
 		    buf.append('\\');
 		}
 		buf.append(c);
 	    }
 	    else {
 	        buf.append(token);
 	    }
 	}

 	return new String(buf);
    }
}

class NullProcess extends Process {
    public OutputStream getOutputStream() {
        return null;
    }
    public InputStream getInputStream() {
        return null;
    }
    public InputStream getErrorStream() {
        return null;
    }
    public int waitFor() {
        return 0;
    }
    public int exitValue() {
        return 0;
    }
    public void destroy() {
    }
}

class CheckExitCodeRunnable implements Runnable {
    private Process process;
    private sun.awt.print.PrintControl control;
    private Frame dialogOwner;

    public CheckExitCodeRunnable(Process process,
				 sun.awt.print.PrintControl control,
				 Frame dialogOwner) {
	this.process = process;
	this.control = control;
	this.dialogOwner = dialogOwner;
    }

    public void run() {
	int returnVal = -1;
	String errorStr = null;

	try {
	    returnVal = process.waitFor();
	}
	catch (InterruptedException e) {
	    errorStr = PSPrintControl.getMsg("error_msg.interrupted");
	}

	if (returnVal != 0 && errorStr == null) {
	    try {
	        InputStream errorStream = process.getErrorStream();
		byte[] errorText = new byte[errorStream.available()];
	        process.getErrorStream().read(errorText);
		errorStr = PSPrintControl.getMsg("error_msg.error_code") +
		    ' ' + returnVal + ": " + new String(errorText).trim();
	    }
	    catch (IOException e) {
	        errorStr = PSPrintControl.getMsg("error_msg.unknown");
	    }
	}

	if (errorStr != null) {
	    if (control.getDialogAttrib() == DialogType.NONE) {
	        throw new InternalError(errorStr);
	    } else {
	        sun.awt.print.PrintStatusDialog d =
		    new sun.awt.print.PrintStatusDialog(
                                  dialogOwner,
				  PSPrintControl.getMsg("error_dialog.title"),
				  errorStr,
				  PSPrintControl.getMsg("error_button.ok"));
		d.setLocationRelativeTo(dialogOwner);
		d.setVisible(true);
	    }
	}
    }
}
