/*
 * @(#)PSPrintStream.java	1.26 00/02/02
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.motif;

import java.awt.*;
import java.awt.image.*;
import java.awt.JobAttributes.*;
import java.awt.PageAttributes.*;
import java.io.*;
import sun.io.CharToByteConverter;
import java.util.Properties;

import sun.awt.CharsetString;
import sun.awt.FontDescriptor;
import sun.awt.PlatformFont;

/**
 * PSPrintStream is an object that encapsulates the current state of a
 * PostScript output stream.  This object is shared among many PSGraphics
 * objects so that they can manage multiplexed access to the single
 * PostScript execution context.
 * 
 * @version 1.26 02/02/00
 * @author Jim Graham
 * @author Amy Fowler
 * @author Richard Berlin
 */
class PSPrintStream {
    PrintStream		ps;
    PSPrintJob		job;
    public String	title;
    public Dimension	paperDim;
    Properties		fontProps;

    int			curPage = -1;

    Rectangle		curClip;
    Color		curColor;
    Font		curFont;
    int[]		curColormap;

    boolean		prologSent = false;
    boolean		lastFirst = false;
    boolean             mono;

    private static final String SetColorName = "SC";
    private static final String SetFontName = "SF";
    private static final String RectClipName = "RC";
    private static final String InitClipName = "IC";
    private static final String RectFillName = "RF";
    private static final String RectStrokeName = "RS";
    private static final String RRectFillName = "RRF";
    private static final String RRectStrokeName = "RRS";
    private static final String ArcFillName = "AF";
    private static final String ArcStrokeName = "AS";
    private static final String PolygonFillName = "PYF";
    private static final String PolygonStrokeName = "PYS";
    private static final String DrawStringName = "DS";
    private static final String DrawLineName = "DL";
    private static final String ScaleName = "S";
    private static final String StartPageName = "SP";
    private static final String EndPageName = "EP";
    private static final String GsaveName = "GS";
    private static final String GrestoreName = "GR";
    private static final String SetColormapName = "CM";
    private static final String DrawImageName = "DI";
    private static final String DrawIndexImageName = "DII";

    private static final String testString =
	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ 1234567890";

    private static final int MARGIN = 18; // (1/72) points

    PSPrintStream(PSPrintJob job, PrintStream ps, String title,
		  Dimension paper, boolean lastFirst) {
	this.job = job;
	this.ps = ps;
	this.title = title;
	this.paperDim = paper;
	this.lastFirst = lastFirst;
	this.mono = (job.control.getColorAttrib() == ColorType.MONOCHROME);
	//enable privs so initProps can access system properties,
	// open the property file, etc.
	java.security.AccessController.doPrivileged(
			    new java.security.PrivilegedAction() {
	    public Object run() {
		fontProps = initProps();
		return null;
	    }
	});
    }

    /**
     * Initialize PostScript font properties
     */
    private Properties initProps() {
	// search psfont.propeties for fonts
	// and create and initialize fontProps if it exist.

	String jhome = System.getProperty("java.home");
	if (jhome != null){
	    String ulocale = System.getProperty("user.language", "en");
	    try {
		File f = new File(jhome + File.separator +
				  "lib" + File.separator + 
				  "psfont.properties." + ulocale);
		if (!f.canRead()){
		    f = new File(jhome + File.separator + "lib" +
				 File.separator + "psfont.properties");
		    if (!f.canRead()){
			return (Properties)null;
		    }
		}
		// Load property file
		InputStream in =
		    new BufferedInputStream(new FileInputStream(f.getPath()));
		Properties props = new Properties();
		props.load(in);
		in.close();
		return props;
	    } catch (Exception e){
		return (Properties)null;
	    }
	}
	return (Properties)null;
    }

    private int rgbToGray(int r, int g, int b) {
        // use the NTSC luminance conversion
        return (int)(.3 * r + .59 * g + .11 * b);
    }
    private int argbToAGGG(int argb) {
        int alpha = (argb >>> 24) & 0xff;
        int red = (argb >>> 16) & 0xff;
	int green = (argb >>> 8) & 0xff;
	int blue = (argb >>> 0) & 0xff;
	int gray = rgbToGray(red, green, blue);
	return (alpha << 24) | (gray << 16) | (gray << 8) | gray;
    }

    void startPage(int pageNumber) {
	curClip = null;
	curColor = null;
	curFont = null;
	curColormap = null;
	curPage = pageNumber;
	ps.println("\n%%Page:  "+pageNumber+" "+pageNumber);
	ps.println(StartPageName);
    }

    void endPage() {
	ps.println(EndPageName);
	curPage = -1;
	ps.flush();
    }

    void close(int numPages) {
	ps.println("%%Trailer");
	ps.println("%%Pages: "+numPages);
	ps.println("%%EOF");
	ps.flush();
    }

    private void outPoint(PSGraphics psg, int x, int y) {
	ps.print((psg.adjustX + x) + " " +
		 (psg.adjustY - y) + " ");
    }

    private void outRect(PSGraphics psg, int x, int y, int w, int h) {
	outPoint(psg, x, y);
	ps.print(w + " " + (-h) + " ");
    }

    private void outClip(PSGraphics psg) {
	// Note: This assumes that clip Rectangle objects are unique...
	Rectangle clip = psg.clipRect;
	if (curClip != clip) {
	    ps.println(InitClipName);
	    if (clip != null) {
		// Note: The clip is stored in absolute coordinates so we
		// need to adjust it back relative to the current origin
		// so that the outPoint adjustments do the right thing.
		outRect(psg,
			clip.x - psg.originX, clip.y - psg.originY,
			clip.width, clip.height);
		ps.println(RectClipName);
	    }
	    curClip = clip;
	}
    }

    private void outColor(Color c) {
	if (curColor != c) {
	    int red = c.getRed();
	    int green = c.getGreen();
	    int blue = c.getBlue();
	    if (mono) {
	        red = green = blue = rgbToGray(red, green, blue);
	    }
	    ps.println(red + " " +
		       green + " " +
		       blue + " " +
		       SetColorName);
	    curColor = c;
	}
    }

    private void outFont(PSGraphics psg) {
	Font font = psg.font;
	// REMIND: all we check is the size and style for now...
	if (curFont != font) {
	    curFont = font;
	    FontMetrics curMetrics = psg.curMetrics;
	    if (psg.curMetrics == null) {
		curMetrics = psg.getFontMetrics(font);
		psg.curMetrics = curMetrics;
	    }
	    int desiredWidth = curMetrics.stringWidth(testString);
	    int fontIndex = font.getStyle();
            // NB leaving helvetica and times in here, but in fact we do
            // not support these and they won't work
	    if ((font.getFamily()).toLowerCase().startsWith("helvetica") ||
		(font.getFamily()).toLowerCase().startsWith("sansserif")) {
		fontIndex = 0;
	    } else if ((font.getFamily()).toLowerCase().startsWith("dialog")) {
		fontIndex = 0;
	    } else if ((font.getFamily()).toLowerCase().startsWith("times") ||
		       (font.getFamily()).toLowerCase().startsWith("serif")) {
		fontIndex = 4;
	    } else {
		fontIndex = 8;
	    }
	    if (font.isBold())
		fontIndex += 1;
	    if (font.isItalic())
	        fontIndex += 2;
	    ps.println("% Font["
		       +font.getFamily()+", "
		       +font.getStyle()+", "
		       +font.getSize()
		       +"]");
	    ps.println(desiredWidth+" "+font.getSize()+" "+fontIndex+" "
		       +SetFontName);
	}
    }

    private void outStr(String str) {
	StringBuffer outbuf = new StringBuffer();
	outbuf.append('(');
	for (int i = 0; i < str.length(); i++) {
	    char ch = str.charAt(i);
	    if (ch == '(' || ch == ')' || ch == '\\')
		outbuf.append('\\');
	    outbuf.append(ch);
	}
	outbuf.append(')');
	ps.print(outbuf.toString());
    }

    synchronized void sendLine(PSGraphics psg,
			       int x1, int y1, int x2, int y2) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(psg.foreground);
	outPoint(psg, x1, y1);
	outPoint(psg, x2, y2);
	ps.println(DrawLineName);
    }

    synchronized void sendRect(PSGraphics psg,
			       int x, int y, int w, int h,
			       Color color, boolean fill) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(color);
	outRect(psg, x, y, w, h);
	ps.println(fill ? RectFillName : RectStrokeName);
    }

    void sendRect(PSGraphics psg, int x, int y, int w, int h, boolean fill) {
	sendRect(psg, x, y, w, h, psg.foreground, fill);
    }

    void fillRect(PSGraphics psg, int x, int y, int w, int h, Color color) {
	sendRect(psg, x, y, w, h, color, true);
    }

    synchronized void sendRoundRect(PSGraphics psg,
				    int x, int y, int w, int h,
				    float xRad, float yRad,
				    boolean fill) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(psg.foreground);
	outRect(psg, x, y, -w, -h); // width/height used opposite for roundrect
	ps.print(xRad + " " + yRad + " ");
	ps.println(fill ? RRectFillName : RRectStrokeName);
    }

    synchronized void sendArc(PSGraphics psg,
			      int x, int y, int xRad, int yRad,
			      int startA, int arcA,
			      boolean fill) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(psg.foreground);
	outPoint(psg, x, y);
	ps.print(xRad + " " + yRad + " " +
		 startA + " " + arcA + " ");
	ps.println(fill ? ArcFillName : ArcStrokeName);
    }

    synchronized void sendPolygon(PSGraphics psg,
				  int xp[], int yp[], int npoints,
				  boolean close, boolean fill) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(psg.foreground);
	ps.print(close+" ");
	for(int i = 0; i < npoints; i++) {
	    outPoint(psg, xp[i], yp[i]);
	}
	ps.print(npoints+" ");
	ps.println(fill ? PolygonFillName : PolygonStrokeName);
    }	

    synchronized void sendString(PSGraphics psg,
				 String str, int x, int y) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(psg.foreground);
	Font font = psg.font;
	FontMetrics curMetrics = psg.curMetrics;
	if (curMetrics == null) {
	    curMetrics = psg.getFontMetrics(font);
	    psg.curMetrics = curMetrics;
	}

        if (fontProps == null) {
	    outFont(psg);
	    outStr(str);
	    ps.print(" " + curMetrics.stringWidth(str) + " ");
	    outPoint(psg, x, y);
	    ps.println(DrawStringName);
	    return;
	}

	CharsetString[] acs =
	    ((MFontPeer)
	     (font.getPeer())).makeMultiCharsetString(str);

	for (int i = 0; i < acs.length; i++){
	    // out Font

	    CharsetString cs = acs[i];
	    CharToByteConverter fontCS = cs.fontDescriptor.fontCharset;

	    //FontMetrics curMetrics = psg.getFontMetrics(font);
	    int desiredWidth = curMetrics.charsWidth(cs.charsetChars, cs.offset, cs.length);
	    String charset = makeCharsetName(fontCS.getCharacterEncoding());
	    String style = PlatformFont.styleStr(font.getStyle());
	    String name =
		fontProps.getProperty(font.getName().toLowerCase(),
				      "serif");
	    String fontName =
		fontProps.getProperty(name + "." + charset + "." + style,
				      "Courier ISOF");
	    int fontIndex;
	    try {
		fontIndex =
		    Integer.parseInt(fontProps.getProperty(fontName));
	    } catch(NumberFormatException e){
		fontIndex = 0;
	    }

	    StringBuffer nativeStr = new StringBuffer();
	    byte[] strSeg = new byte[cs.length * 2];
	    int len;
	    try {
		len = fontCS.convert(cs.charsetChars, cs.offset,
                                     cs.offset + cs.length,
				     strSeg, 0, strSeg.length);
	    } catch(CharConversionException e){
		continue;
	    }
	    nativeStr.append('<');
	    for (int j = 0; j < len; j++){
		byte b = strSeg[j];
		// to avoid encoding conversion with println()
		String hexS = Integer.toHexString(b);
		int length = hexS.length();
		if (length > 2) {
		    hexS = hexS.substring(length - 2, length);
		} else if (length == 1) {
		    hexS = "0" + hexS;
		} else if (length == 0) {
		    hexS = "00";
		}
		nativeStr.append(hexS);
	    }
	    nativeStr.append('>');
	    ps.println("/testString " + nativeStr.toString() + " D");
	    ps.println("% Font[" + font.getName() + ", " +
		       PlatformFont.styleStr(font.getStyle()) + ", "
		       + font.getSize() + "]");
	    ps.println(desiredWidth + " " + font.getSize() + " " +
		       fontIndex + " " + SetFontName);

	    // out String
	    ps.println(nativeStr.toString());

	    // draw string
	    int width = curMetrics.charsWidth(cs.charsetChars, cs.offset, cs.length);
	    ps.print(" " + width + " ");
	    outPoint(psg, x, y);
	    ps.println(DrawStringName);
	    x += width;
	}
    }

    /*
     * Currently CharToByteConverter.getCharacterEncoding() return values are
     * not fixed yet. These are used as the part of the key of
     * psfont.propeties. When those name are fixed this routine can
     * be erased.
     */
    private String makeCharsetName(String name){
	if (name.indexOf("8859_1") != -1) {
	    return "latin1";
	} else {
	    return name.toLowerCase();
	}
	
    }

    private final static int blend(int alpha, int fg, int bg) {
	return ((alpha * fg) + ((255 - alpha) * bg)) / 255;
    }

    static private byte hexDigits[] = {
	(byte)'0', (byte)'1', (byte)'2', (byte)'3',
	(byte)'4', (byte)'5', (byte)'6', (byte)'7',
	(byte)'8', (byte)'9', (byte)'A', (byte)'B',
	(byte)'C', (byte)'D', (byte)'E', (byte)'F'
    };

    private int sendColors(byte[] colors, int start, int len,
			   byte outbuf[], int linepos) {
	while (len-- > 0) {
	    if (linepos + 3 > outbuf.length) {
		outbuf[linepos++] = (byte) '\n';
		ps.write(outbuf, 0, linepos);
		linepos = 0;
	    }

	    byte pixel = colors[start++];
	    outbuf[linepos++] = hexDigits[(pixel >>> 4) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 0) & 0xf];
	}
	return linepos;
    }

    private int sendColors(int[] colors, int start, int len,
			   byte outbuf[], int linepos) {
	while (len-- > 0) {
	    if (linepos + 7 > outbuf.length) {
		outbuf[linepos++] = (byte) '\n';
		ps.write(outbuf, 0, linepos);
		linepos = 0;
	    }

	    int pixel = colors[start++];
	    outbuf[linepos++] = hexDigits[(pixel >>> 20) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 16) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 12) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 8) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 4) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 0) & 0xf];
	}
	return linepos;
    }

    private void sendSubPixels(PSGraphics psg,
			       byte[] pixels, int pixscan,
			       int x1, int y1, int x2, int y2,
			       int x, int y, float scalew, float scaleh) {
	int w = x2 - x1;
	int h = y2 - y1;
	int index = y1 * pixscan + x1;
	int linepos = 0;
	byte outbuf[] = new byte[73];

	ps.print(x1 + " " + y1 + " ");
	ps.print(w + " " + h + " ");
	ps.print(scalew*w + " " + scaleh*h + " ");
	outPoint(psg, x, y);
	ps.println(DrawIndexImageName);
	for (int j = y1; j < y2; j++) {
	    linepos = sendColors(pixels, index, w, outbuf, linepos);
	    index += pixscan;
	}
	if (linepos > 0) {
	    outbuf[linepos++] = (byte) '\n';
	    ps.write(outbuf, 0, linepos);
	}
    }

    private void sendSubPixels(PSGraphics psg,
			       int[] pixels, int pixscan,
			       int x1, int y1, int x2, int y2,
			       int x, int y, float scalew, float scaleh) {
	int w = x2 - x1;
	int h = y2 - y1;
	int index = y1 * pixscan + x1;
	int linepos = 0;
	byte outbuf[] = new byte[73];

	ps.print(x1 + " " + y1 + " ");
	ps.print(w + " " + h + " ");
	ps.print(scalew*w + " " + scaleh*h + " ");
	outPoint(psg, x, y);
	ps.println(DrawImageName);
	for (int j = y1; j < y2; j++) {
	    linepos = sendColors(pixels, index, w, outbuf, linepos);
	    index += pixscan;
	}
	if (linepos > 0) {
	    outbuf[linepos++] = (byte) '\n';
	    ps.write(outbuf, 0, linepos);
	}
    }

    private void sendColormap(PSGraphics psg, int cmap[], int size) {
	if (curColormap != null) {
	    boolean matches = true;
	    for (int i = 0; i < size; i++) {
		if (cmap[i] != curColormap[i]) {
		    matches = false;
		    break;
		}
	    }
	    if (matches) {
		return;
	    }
	}

	curColormap = cmap;

	byte outbuf[] = new byte[73];

	ps.println(size + " " + SetColormapName);
	int linepos = sendColors(cmap, 0, size, outbuf, 0);
	if (linepos > 0) {
	    outbuf[linepos++] = (byte) '\n';
	    ps.write(outbuf, 0, linepos);
	}
    }

    synchronized void sendPixels(PSGraphics psg,
				 byte pixels[], IndexColorModel model,
				 int dx1, int dy1, int dx2, int dy2,
				 int sx1, int sy1, int sx2, int sy2,
				 Color bgcolor)
    {
	if (psg.pageNumber != curPage) {return;}
	int cmap[] = new int[256];
	int index = model.getMapSize();
	boolean trans[] = new boolean[256];
	boolean transpixels = false;
	int bgred, bggreen, bgblue;
	if (bgcolor != null) {
	    bgred = bgcolor.getRed();
	    bggreen = bgcolor.getGreen();
	    bgblue = bgcolor.getBlue();
	    if (mono) {
	        bgred = bggreen = bgblue = rgbToGray(bgred, bggreen, bgblue);
	    }
	} else {
	    bgred = bggreen = bgblue = 0;
	}
	while (--index >= 0) {
	    int rgb = model.getRGB(index);
	    if (mono) {
	        rgb = argbToAGGG(rgb);
	    }
	    int alpha = (rgb >>> 24);
	    if (alpha != 0xff) {
		if (bgcolor != null) {
		    int red = blend(alpha, (rgb >>> 16) & 0xff, bgred);
		    int green = blend(alpha, (rgb >>> 8) & 0xff, bggreen);
		    int blue = blend(alpha, (rgb >>> 0) & 0xff, bgblue);
		    rgb = (red << 16) | (green << 8) | blue;
		} else if (alpha < 128) {
		    trans[index] = true;
		    transpixels = true;
		}
	    }
	    cmap[index] = rgb;
	}
	outClip(psg);
	sendColormap(psg, cmap, model.getMapSize());

	int w = sx2 - sx1;
	int h = sy2 - sy1;
	float scalew = ((float) (dx2 - dx1)) / ((float) w);
	float scaleh = ((float) (dy2 - dy1)) / ((float) h);
	if (transpixels) {
	    index = 0;
	    for (int j = 0; j < h; j++) {
		int startx = -1;
	    	for (int i = 0; i < w; i++) {
		    if (trans[pixels[index++] & 0xff]) { 
		    	if (startx >= 0) {
			    sendSubPixels(psg, pixels, w,
					  startx, j, i, j + 1,
					  dx1, dy1, scalew, scaleh);
			    startx = -1;
		    	}
		    } else if (startx < 0) {
			startx = i;
		    }
	    	}
	    	if (startx >= 0) {
		    sendSubPixels(psg, pixels, w,
				  startx, j, w, j + 1,
				  dx1, dy1, scalew, scaleh);
	    	}
	    }
	} else {
	    sendSubPixels(psg, pixels, w,
			  0, 0, w, h,
			  dx1, dy1, scalew, scaleh);
	}
    }

    synchronized void sendPixels(PSGraphics psg,
				 int pixels[], ColorModel model,
				 int dx1, int dy1, int dx2, int dy2,
				 int sx1, int sy1, int sx2, int sy2,
				 Color bgcolor)
    {
	if (psg.pageNumber != curPage) {return;}
	int w = sx2 - sx1;
	int h = sy2 - sy1;
	float scalew = ((float) (dx2 - dx1)) / ((float) w);
	float scaleh = ((float) (dy2 - dy1)) / ((float) h);
	int bgred, bggreen, bgblue;
	if (bgcolor != null) {
	    bgred = bgcolor.getRed();
	    bggreen = bgcolor.getGreen();
	    bgblue = bgcolor.getBlue();
	    if (mono) {
	        bgred = bggreen = bgblue = rgbToGray(bgred, bggreen, bgblue);
	    }
	} else {
	    bgred = bggreen = bgblue = 0;
	}

	outClip(psg);
	int index = 0;
	int starty = -1;
	for (int j = 0; j < h; j++) {
	    int startx = -1;
	    for (int i = 0; i < w; i++) {
		int rgb = model.getRGB(pixels[index]);
		if (mono) {
		    rgb = argbToAGGG(rgb);
		}
		int alpha = (rgb >>> 24);
		if (alpha != 0xff) { 
		    if (bgcolor != null) {
			int red = blend(alpha, (rgb >>> 16) & 0xff, bgred);
			int green = blend(alpha, (rgb >>> 8) & 0xff, bggreen);
			int blue = blend(alpha, (rgb >>> 0) & 0xff, bgblue);
			pixels[index] = (red << 16) | (green << 8) | blue;
			if (startx < 0) {
			    startx = i;
			}
		    } else if (startx >= 0) {
			if (starty >= 0) {
			    sendSubPixels(psg, pixels, w,
					  0, starty, w, j,
					  dx1, dy1, scalew, scaleh);
			    starty = -1;
			}
			sendSubPixels(psg, pixels, w,
				      startx, j, i, j + 1,
				      dx1, dy1, scalew, scaleh);
			startx = -1;
		    }
		} else {
		    pixels[index] = rgb;
		    if (startx < 0) {
			startx = i;
		    }
		}
		index++;
	    }
	    if (startx == 0) {
		if (starty < 0) {
		    starty = j;
		}
	    } else if (startx > 0) {
		sendSubPixels(psg, pixels, w,
			      startx, j, w, j + 1,
			      dx1, dy1, scalew, scaleh);
	    }
	}
	if (starty >= 0) {
	    sendSubPixels(psg, pixels, w,
			  0, starty, w, h,
			  dx1, dy1, scalew, scaleh);
	}
    }

    void sendProlog() {
	if (prologSent) {
	    return;
	}
	PSPaperSize paperSize
	    = PSPaperSize.find(paperDim.width, paperDim.height, 1); 
	if (paperSize == null) {
	    paperSize = PSPaperSize.find(paperDim.height, paperDim.width, 1); 
	}

	ps.println("%!PS-Adobe-3.0");
	ps.println("%%Title: " + title);
	ps.println("%%Pages: (atend)");
	ps.println("%%PageOrder: " + (lastFirst ? "Special" : "Ascend"));
	ps.println("%%DocumentFonts: Helvetica Helvetica-Bold");
	ps.println("%%+Helvetica-Oblique Helvetica-BoldOblique");
	ps.println("%%+Times-Roman Times-Bold Times-Italic Times-BoldItalic");
	ps.println("%%+Courier Courier-Bold Courier-Oblique Courier-BoldOblique");
	ps.println("%%EndComments");
	ps.println("%%BeginProlog");
	ps.println("/BD {bind def} bind def");
	ps.println("/D {def} BD");
	ps.println("/ISOF {");
	ps.println("	dup findfont dup length 1 add dict begin {");
	ps.println("		1 index /FID eq {pop pop} {D} ifelse");
	ps.println("	} forall /Encoding ISOLatin1Encoding D");
	ps.println("	currentdict end definefont");
	ps.println("} BD");
	ps.println("/F [");
	if (fontProps == null){
	    ps.println("	/Helvetica ISOF");
	    ps.println("	/Helvetica-Bold ISOF");
	    ps.println("	/Helvetica-Oblique ISOF");
	    ps.println("	/Helvetica-BoldOblique ISOF");
	    ps.println("	/Times-Roman ISOF");
	    ps.println("	/Times-Bold ISOF");
	    ps.println("	/Times-Italic ISOF");
	    ps.println("	/Times-BoldItalic ISOF");
	    ps.println("	/Courier ISOF");
	    ps.println("	/Courier-Bold ISOF");
	    ps.println("	/Courier-Oblique ISOF");
	    ps.println("	/Courier-BoldOblique ISOF");
	} else {
	    int count = Integer.parseInt
		(fontProps.getProperty("font.num", "9"));
	    for (int i = 0; i < count; i++){
		ps.println("    /" + fontProps.getProperty
			   ("font." + String.valueOf(i), "Courier ISOF"));
	    }
	}
	ps.println("] D");
	ps.println("/R {4 2 roll moveto 1 index 0 rlineto");
	ps.println("	0 exch rlineto neg 0 rlineto closepath} BD");
	ps.println("/Adict 8 dict D");
	ps.println("Adict /mtrx matrix put");
	ps.println("/A { Adict begin");
	ps.println("    /endangle exch D /startangle exch D");
	ps.println("    /yrad exch D /xrad exch D");
	ps.println("	/y exch D /x exch D");
	ps.println("    /savematrix mtrx currentmatrix D");
	ps.println("    x y translate xrad yrad scale");
	ps.println("    0 0 1 startangle endangle arc");
	ps.println("    savematrix setmatrix");
	ps.println("    end } BD");
	ps.println("/PY {");
	ps.println("    dup 1 ge {3 1 roll moveto} if");
	ps.println("    -1 2 {pop lineto} for");
	ps.println("    {closepath} if");
	ps.println("} BD");
	ps.println("/CL { dup 0 exch -1 1 arct} BD");
	ps.println("/RCL { dup 0 exch 1 1 arct} BD");
	ps.println("/RRdict 8 dict D");
 	ps.println("RRdict /mtrx matrix put");
	ps.println("/RR {RRdict begin");
	ps.println("    /yrad exch D /xrad exch D");
	ps.println("    /height exch D /width exch D");
	ps.println("    /y exch D /x exch D");
	ps.println("	/width width xrad div D /height height yrad div D");
	ps.println("	/savematrix mtrx currentmatrix D");
	ps.println(" 	x y translate xrad yrad scale -90 rotate");
	ps.println("	1 0 moveto height RCL height 0 translate -90 rotate");
	ps.println("    width CL width 0 translate -90 rotate");
	ps.println("	height RCL height 0 translate -90 rotate width CL closepath");
	ps.println("	savematrix setmatrix");
	ps.println("	end } BD");	 
	ps.println("/"+SetColorName
		   +" {3 {255 div 3 1 roll} repeat setrgbcolor} BD");
	ps.println("/"+SetFontName +" {");
	ps.println("	F exch get exch scalefont setfont");
	if (fontProps == null) {
	    ps.println("	("+testString+") stringwidth pop div");
	} else {
	    ps.println(" testString stringwidth pop div");
	}
	ps.println("	currentfont exch scalefont setfont} BD");
	ps.println("/"+RectClipName+" {R clip newpath} BD");
	ps.println("/"+RectFillName+" {R fill} BD");
	ps.println("/"+RectStrokeName+" {R stroke} BD");
	ps.println("/"+RRectFillName+" {RR fill} BD");
	ps.println("/"+RRectStrokeName+" {RR stroke} BD");
	ps.println("/"+ArcStrokeName+" {A stroke} BD");
	ps.println("/"+ArcFillName+" {6 -2 roll 2 copy moveto 6 2 roll A fill} BD");
	ps.println("/"+PolygonStrokeName+" {PY stroke} BD");
	ps.println("/"+PolygonFillName+" {PY eofill} BD");
	ps.println("/NZ {dup 1 lt {pop 1} if} BD");
	ps.println("/"+DrawStringName +" {");
	ps.println("	moveto 1 index stringwidth pop NZ sub");
	ps.println("	1 index length 1 sub NZ div 0");
	ps.println("	3 2 roll ashow newpath} BD");
	ps.println("/"+DrawLineName+" {moveto lineto stroke} BD");
	ps.println("/"+ScaleName+" {scale} BD");
	ps.println("/"+StartPageName+" {/P save D} BD");
	ps.println("/"+EndPageName+" {showpage P restore} BD");
	ps.println("/"+GsaveName+" {gsave} BD");
	ps.println("/"+GrestoreName+" {grestore} BD");
	ps.println("/"+SetColormapName+" {");
	ps.println("	/cmapr 256 array D");
	ps.println("	/cmapg 256 array D");
	ps.println("	/cmapb 256 array D");
	ps.println("	/cmapgray null D");
	ps.println("	3 string exch 0 exch 1 exch 1 sub {");
	ps.println("		currentfile 2 index readhexstring pop");
	ps.println("		cmapr 2 index 2 index 0 get put");
	ps.println("		cmapg 2 index 2 index 1 get put");
	ps.println("		cmapb 2 index 2 index 2 get put");
	ps.println("	pop pop } for pop");
	ps.println("} BD");
	ps.println("/"+DrawIndexImageName+" {");
	ps.println("	GS");
	ps.println("	/imgsave save D");	// NeWSprint leaves files open!
	ps.println("	translate scale");
	ps.println("	/imgh exch D");
	ps.println("	/imgw exch D");
	ps.println("	/imgy exch neg D");
	ps.println("	/imgx exch neg D");
	ps.println("	/imgstr imgw string D");
	ps.println("	/colorimage where {");
	ps.println("		pop");
	ps.println("		/cimgstr imgw 3 mul string D");
	ps.println("		imgw imgh 8 [imgw 0 0 imgh neg imgx imgy] {");
	ps.println("			currentfile imgstr readhexstring pop");
	ps.println("			0 1 imgw 1 sub {");
	ps.println("				2 copy get");
	ps.println("				cmapr 1 index get");
	ps.println("				cimgstr exch 3 index");
	ps.println("				3 mul exch put");
	ps.println("				cmapg 1 index get");
	ps.println("				cimgstr exch 3 index");
	ps.println("				3 mul 1 add exch put");
	ps.println("				cmapb 1 index get");
	ps.println("				cimgstr exch 3 index");
	ps.println("				3 mul 2 add exch put");
	ps.println("			pop pop } for pop cimgstr");
	ps.println("		} false 3 colorimage");
	ps.println("	} {");
	ps.println("		cmapgray null eq {");
	ps.println("			/cmapgray cmapr length array D");
	ps.println("			0 1 cmapr length 1 sub {");
	ps.println("				cmapgray exch");
	ps.println("				cmapr 1 index get .299 mul");
	ps.println("				cmapg 2 index get .587 mul");
	ps.println("				cmapb 3 index get .114 mul");
	ps.println("				add add cvi put");
	ps.println("			} for");
	ps.println("		} if");
	ps.println("		imgw imgh 8 [imgw 0 0 imgh neg 0 0] {");
	ps.println("			currentfile imgstr readhexstring pop");
	ps.println("			0 1 imgw 1 sub {");
	ps.println("				imgstr exch 2 copy get");
	ps.println("				cmapgray exch get put");
	ps.println("			} for");
	ps.println("		} image");
	ps.println("	} ifelse");
	ps.println("	imgsave restore");
	ps.println("	GR");
	ps.println("} BD");
	ps.println("/"+DrawImageName+" {");
	ps.println("	GS");
	ps.println("	/imgsave save D");	// NeWSprint leaves files open!
	ps.println("	translate scale");
	ps.println("	/imgh exch D");
	ps.println("	/imgw exch D");
	ps.println("	/imgy exch neg D");
	ps.println("	/imgx exch neg D");
	ps.println("	/imgstr imgw 3 mul string D");
	ps.println("	imgw imgh 8 [imgw 0 0 imgh neg imgx imgy] {");
	ps.println("		currentfile imgstr readhexstring pop");
	ps.println("	} bind false 3 colorimage");
	ps.println("    imgsave restore");
	ps.println("    GR");		
	ps.println("} BD");
	ps.println("%%EndProlog");
	ps.println("%%BeginSetup");

	// Only generate sides information is we are printing directly to
	// a printer. Inserting these commands into a Postscript file limits
	// the embeddability of the file.
	if (job.control.getDestAttrib() == DestinationType.PRINTER) {
	    SidesType sides = job.control.getSidesAttrib();
	    boolean duplex = (sides == SidesType.TWO_SIDED_LONG_EDGE ||
			      sides == SidesType.TWO_SIDED_SHORT_EDGE);
	    ps.println("mark {");
	    ps.println("%%BeginFeature: *Duplex " + 
		       ((duplex) ? "Duplex" : "Simplex"));
	    ps.println("2 dict begin");
	    ps.println("  /Duplex " + duplex + " def");
	    ps.println("  /Policies");
	    ps.println("    1 dict begin");
	    ps.println("      /Duplex 1 def");
	    ps.println("    currentdict end def");
	    ps.println("currentdict end setpagedevice");
	    ps.println("%%EndFeature");
	    ps.println("} stopped cleartomark");
	    if (duplex) {
	        boolean tumble = (sides == SidesType.TWO_SIDED_SHORT_EDGE);
		ps.println("mark {");
		ps.println("%%BeginFeature: *Tumble " + 
			   ((tumble) ? "True" : "False"));
		ps.println("2 dict begin");
		ps.println("  /Tumble " + tumble + " def");
		ps.println("  /Policies");
		ps.println("    1 dict begin");
		ps.println("      /Tumble 1 def");
		ps.println("    currentdict end def");
		ps.println("currentdict end setpagedevice");
		ps.println("%%EndFeature");
		ps.println("} stopped cleartomark");
	    }
	}

	if (paperSize != null) {
	    ps.println(paperSize.toPostScriptCode());
	}
        if (paperDim.width > paperDim.height) {
            ps.println("90 rotate 0 "+-(paperDim.height)+" translate");
	}
	if (job.control.getOriginAttrib() == OriginType.PRINTABLE) {
	    // Just do the translate here. We set the margin clipping region
	    // in the InitClipName (/IC) routine. The calls to initclip would
	    // cancel out any settings we made here.
	    ps.println(paperSize.toInitClip(InitClipName, MARGIN));
	    ps.println(MARGIN + " -" + MARGIN + " translate");
	} else {
	    ps.println("/"+InitClipName+" {initclip} BD");
	}
	int res = job.control.getResAttrib();
	if (res != 72) {
	    ps.println((72.0 / res) + " " + (72.0 / res) + " scale");
	    ps.println(paperDim.height + " dup " + (res / 72.0));
	    ps.println("mul exch sub 0 exch translate");
	}
	ps.println("%%EndSetup");
	ps.flush();
	prologSent = true;
    }
}
