/*
 * @(#)PSPrinterJob.java	1.27 01/02/09
 *
 *  Copyright 2000-2001 Sun Microsystems, Inc. All rights reserved.
 *  Copyright 2000-2001 Sun Microsystems, Inc. Tous droits rservs.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 */

package sun.awt.motif;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.Toolkit;

import java.awt.image.BufferedImage;

import java.awt.geom.AffineTransform;
import java.awt.geom.PathIterator;

import java.awt.image.BufferedImage;

import java.awt.print.Pageable;
import java.awt.print.PageFormat;
import java.awt.print.Paper;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.awt.print.PrinterIOException;
import java.awt.print.PrinterJob;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.CharConversionException;
import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.io.PrintStream;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.ResourceBundle;

import sun.awt.CharToByteSymbol;
import sun.awt.CharsetString;
import sun.awt.FontDescriptor;
import sun.awt.PlatformFont;

import sun.io.CharToByteConverter;

import sun.java2d.PeekGraphics;
import sun.java2d.PeekMetrics;
import sun.java2d.RasterPrinterJob;

/** 
 * A class which initiates and executes a PostScript printer job.
 *
 * @version 1.0 01 Feb 1998
 * @author Richard Blanchard
 */
public class PSPrinterJob extends RasterPrinterJob {

 /* Class Constants */

    /** 
     * Passed to the <code>setFillMode</code>
     * method this value forces fills to be
     * done using the even-odd fill rule.
     */
    protected static final int FILL_EVEN_ODD = 1;

    /** 
     * Passed to the <code>setFillMode</code>
     * method this value forces fills to be
     * done using the non-zero winding rule.
     */
    protected static final int FILL_WINDING = 2;

    /* PostScript has a 64K maximum on its strings.
     */
    private static final int MAX_PSSTR = (1024 * 64 - 1);

    private static final int RED_MASK = 0x00ff0000;
    private static final int GREEN_MASK = 0x0000ff00;
    private static final int BLUE_MASK = 0x000000ff;

    private static final int RED_SHIFT = 16;
    private static final int GREEN_SHIFT = 8;
    private static final int BLUE_SHIFT = 0;

    private static final int LOWNIBBLE_MASK = 0x0000000f;
    private static final int HINIBBLE_MASK =  0x000000f0;
    private static final int HINIBBLE_SHIFT = 4;
    private static final byte hexDigits[] = {
	(byte)'0', (byte)'1', (byte)'2', (byte)'3',
	(byte)'4', (byte)'5', (byte)'6', (byte)'7',
	(byte)'8', (byte)'9', (byte)'A', (byte)'B',
	(byte)'C', (byte)'D', (byte)'E', (byte)'F'
    };

    private static final int PS_XRES = 300;
    private static final int PS_YRES = 300;

    /**
     * Useful mainly for debugging, this system property
     * can be used to force the printing code to print
     * using a particular pipeline. The two currently
     * supported values are FORCE_RASTER and FORCE_PDL.
     */
    private static final String FORCE_PIPE_PROP = "sun.java2d.print.pipeline";

    /**
     * When the system property FORCE_PIPE_PROP has this value
     * then each page of a print job will be rendered through
     * the raster pipeline.
     */
    private static final String FORCE_RASTER = "raster";

    /**
     * When the system property FORCE_PIPE_PROP has this value
     * then each page of a print job will be rendered through
     * the PDL pipeline.
     */
    private static final String FORCE_PDL = "pdl";

    /**
     * When the system property SHAPE_TEXT_PROP has this value
     * then text is always rendered as a shape, and no attempt is made
     * to match the font through GDI
     */
    private static final String SHAPE_TEXT_PROP = "sun.java2d.print.shapetext";

    private static final String ADOBE_PS_STR =	"%!PS-Adobe-3.0";
    private static final String EOF_COMMENT =	"%%EOF";
    private static final String PAGE_COMMENT =	"%%Page: ";

    private static final String READIMAGEPROC =	"/imStr 0 def /heximageSrc"
					      + "{currentfile imStr"
					      + " readhexstring pop} def";

    private static final String COPIES =	"/#copies exch def";
    private static final String PAGE_SAVE =     "/pgSave save def";
    private static final String PAGE_RESTORE =  "pgSave restore";
    private static final String SHOWPAGE =	"showpage";
    private static final String IMAGE_SAVE =	"/imSave save def";
    private static final String IMAGE_STR =	" string /imStr exch def";
    private static final String IMAGE_RESTORE =	"imSave restore";

    private static final String COORD_PREP =	" 0 exch translate "
					      + "1 -1 scale"
					      + "[72 " + PS_XRES + " div "
					      + "0 0 "
					      + "72 " + PS_YRES + " div "
					      + "0 0]concat";

    private static final String SetFontName = "SF";

    private static final String DrawStringName = "DS";

    /**
     * The PostScript invocation to fill a path using the
     * even-odd rule.
     */
    private static final String EVEN_ODD_FILL_STR = "eofill";

    /**
     * The PostScript invocation to fill a path using the
     * non-zero winding rule.
     */
    private static final String WINDING_FILL_STR = "fill";

    /**
     * The PostScript to set the clip to be the current path
     * using the even odd rule.
     */
    private static final String EVEN_ODD_CLIP_STR = "eoclip";

    /**
     * The PostScript to set the clip to be the current path
     * using the non-zero winding rule.
     */
    private static final String WINDING_CLIP_STR = "clip";

    /**
     * Expecting two numbers on the PostScript stack, this
     * invocation moves the current pen position.
     */
    private static final String MOVETO_STR = " moveto";
    /**
     * Expecting two numbers on the PostScript stack, this
     * invocation draws a PS line from the current pen
     * position to the point on the stack.
     */
    private static final String LINETO_STR = " lineto";

    /**
     * This PostScript operator takes two control points
     * and an ending point and using the current pen
     * position as a starting point adds a bezier
     * curve to the current path.
     */
    private static final String CURVETO_STR = " curveto";

    /**
     * The PostScript to pop a state off of the printer's
     * gstate stack.
     */
    private static final String GRESTORE_STR = "grestore";
    /**
     * The PostScript to push a state on to the printer's
     * gstate stack.
     */
    private static final String GSAVE_STR = "gsave";

    /**
     * Make the current PostScript path an empty path.
     */
    private static final String NEWPATH_STR = "newpath";

    /**
     * Close the current subpath by generating a line segment
     * from the current position to the start of the subpath.
     */
    private static final String CLOSEPATH_STR = "closepath";

    /**
     * Use the three numbers on top of the PS operator
     * stack to set the rgb color.
     */
    private static final String SETRGBCOLOR_STR = " setrgbcolor";

    /**
     * Use the top number on the stack to set the printer's
     * current gray value.
     */
    private static final String SETGRAY_STR = " setgray";

    private static final String testString =
        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ 1234567890";

    /**
     * values obtained from System properties in static initialiser block
     */
    private static boolean forcePDL = false;
    private static boolean forceRaster = false;
    static boolean shapeTextProp = false;

 /* Instance Variables */

   private static ResourceBundle messageRB;
   
   int mDestType;

   String mDestination = "lp";

   String mOptions;

   Font mLastFont;

   Color mLastColor;

   Shape mLastClip;

   AffineTransform mLastTransform;

   /**
    * The metrics for the font currently set.
    */
   FontMetrics mCurMetrics;

   /**
    * The output stream to which the generated PostScript
    * is written.
    */
   PrintStream mPSStream;

   /* The temporary file to which we spool before sending to the printer  */

   File spoolFile;

   /**
    * This string holds the PostScript operator to
    * be used to fill a path. It can be changed
    * by the <code>setFillMode</code> method.
    */
   private String mFillOpStr = WINDING_FILL_STR;

   /**
    * This string holds the PostScript operator to
    * be used to clip to a path. It can be changed
    * by the <code>setFillMode</code> method.
    */
   private String mClipOpStr = WINDING_CLIP_STR;

   /**
    * A stack that represents the PostScript gstate stack.
    */
   ArrayList mGStateStack = new ArrayList();

   /**
    * The x coordinate of the current pen position.
    */
   private float mPenX;
   
   /**
    * The y coordinate of the current pen position.
    */
   private float mPenY;

   /**
    * The x coordinate of the starting point of
    * the current subpath.
    */
   private float mStartPathX;
 
   /**
    * The y coordinate of the starting point of
    * the current subpath.
    */
   private float mStartPathY;

   /**
    * An optional mapping of fonts to PostScript names.
    */
   private static Properties mFontProps = null;

    /* Class static initialiser block */
    static {
       //enable priviledges so initProps can access system properties,
        // open the property file, etc.
        java.security.AccessController.doPrivileged(
                            new java.security.PrivilegedAction() {
            public Object run() {
                mFontProps = initProps();
                return null;
            }
        });

       forcePDL = false;
       forceRaster = false;
       shapeTextProp = false;

        /* The system property FORCE_PIPE_PROP
         * can be used to force the printing code to
         * use a particular pipeline. Either the raster
         * pipeline or the pdl pipeline can be forced.
         */
        Properties props = System.getProperties();
        if (props != null) {
            String forceStr = props.getProperty(FORCE_PIPE_PROP);
            if (forceStr != null) {
                if (forceStr.equalsIgnoreCase(FORCE_PDL)) {
                    forcePDL = true;
                    //System.err.println("PDL printer output selected");
                } else if (forceStr.equalsIgnoreCase(FORCE_RASTER)) {
                    forceRaster = true;
                    //System.err.println("Raster printer output selected");
                }
            }

           String shapeTextStr = props.getProperty(SHAPE_TEXT_PROP);

           if (shapeTextStr != null) {
                  shapeTextProp = true;
           }
        }
    }

    /*
     * Initialize PostScript font properties.
     * Copied from PSPrintStream
     */
    private static Properties initProps() {
        // search psfont.propeties for fonts
        // and create and initialize fontProps if it exist.

        String jhome = System.getProperty("java.home");

        if (jhome != null){
            String ulocale = System.getProperty("user.language", "en");
            try {

                File f = new File(jhome + File.separator +
                                  "lib" + File.separator +
                                  "psfontj2d.properties." + ulocale);

                if (!f.canRead()){

                    f = new File(jhome + File.separator +
                                      "lib" + File.separator +
                                      "psfont.properties." + ulocale);
                    if (!f.canRead()){

                        f = new File(jhome + File.separator + "lib" +
                                     File.separator + "psfontj2d.properties");

                        if (!f.canRead()){

                            f = new File(jhome + File.separator + "lib" +
                                         File.separator + "psfont.properties");

                            if (!f.canRead()){
                                return (Properties)null;
                            }
                        }
                    }
                }

                // Load property file
                InputStream in =
                    new BufferedInputStream(new FileInputStream(f.getPath()));
                Properties props = new Properties();
                props.load(in);
                in.close();
                return props;
            } catch (Exception e){
                return (Properties)null;
            }
        }
        return (Properties)null;
    }

 /* Constructors */

    public PSPrinterJob()
    {
    }

 /* Instance Methods */
 
    /**
     * Display a dialog to the user allowing the modification of a
     * PageFormat instance.
     * The <code>page</code> argument is used to initialize controls
     * in the page setup dialog.
     * If the user cancels the dialog, then the method returns the
     * original <code>page</code> object unmodified.
     * If the user okays the dialog then the method returns a new
     * PageFormat object with the indicated changes.
     * In either case the original <code>page</code> object will
     * not be modified.
     * @param     page    the default PageFormat presented to the user
     *                    for modification
     * @return    the original <code>page</code> object if the dialog
     *            is cancelled, or a new PageFormat object containing
     *            the format indicated by the user if the dialog is
     *            acknowledged
     * @since     JDK1.2
     */
    public PageFormat pageDialog(final PageFormat page) {

        PageFormat newPage =
            (PageFormat) java.security.AccessController.doPrivileged(
            new java.security.PrivilegedAction() {

            public Object run() {

                Frame frame = new Frame();
                UPageDialog dialog = new UPageDialog(frame, page);
                dialog.show();

                if (dialog.getSetupStatus() != dialog.OK) {
                    return page;
                } else { 
                   PageFormat page2 = (PageFormat)page.clone();
                   page2.setOrientation(dialog.getOrientation());
                   page2.setPaper(dialog.getPaper());
                   return page2;
                }
            }
        });

        return newPage;
    }
       
    /**
     * Presents the user a dialog for changing properties of the
     * print job interactively. 
     * @returns false if the user cancels the dialog and
     *          true otherwise.
     */
    public boolean printDialog() {

	// A security check has already been performed in the
        // java.awt.print.printerJob.getPrinterJob method.
	// So by the time we get here, it is OK for the current thread
	// to print either to a file (froma  Dialog we control!) or
	// to a chosen printer.


	// We raise privilege when we put up the dialog, to avoid
	// the "warning applet window" banner.
	Boolean doPrint = (Boolean) java.security.AccessController.doPrivileged(
	    new java.security.PrivilegedAction() {
	    public Object run() {

		Frame frame = new Frame();
		UPrinterDialog dialog = new UPrinterDialog(frame, "Print",
                                                  checkAllowedToPrintToFile());

		dialog.setCopies(getCopies());
		dialog.setBannerString(getJobName());

		dialog.show();

	 	/* If the user hit print then we want to gather the print
		 * dialog settings.
		 */
		if (dialog.getPrintStatus() != dialog.PRINT) {
		    return Boolean.FALSE;
		}
    	        mDestType = dialog.getDestType();
		mDestination = dialog.getDestString(mDestType);
		mOptions = dialog.getOptionsString();
		setJobName(dialog.getBannerString());
		setCopies(dialog.getCopies());
		setCollated(false);

		if (mDestType == PrintControl.FILE) {
		    return new Boolean(okayToCreateFile(frame, mDestination));
		} else {
		    return Boolean.TRUE;
		}
	    }
	});

	return doPrint.booleanValue();
    }


    public void setCollated(boolean collate) {
	super.setCollated(collate);
    }

    /**
     * Invoked by the RasterPrinterJob super class
     * this method is called to mark the start of a
     * document.
     */
    protected void startDoc() throws PrinterException {

	// A security check has been performed in the
        // java.awt.print.printerJob.getPrinterJob method.
	// We use an inner class to execute the privilged open operations.
	// Note that we only open a file if it has been nominated by
	// the end-user in a dialog that we ouselves put up.

	PrinterOpener po = new PrinterOpener();
	java.security.AccessController.doPrivileged(po);
	if (po.pex != null) {
	    throw po.pex;
	}
	OutputStream output = po.result;

	mPSStream = new PrintStream(new BufferedOutputStream(output));

	mPSStream.println(ADOBE_PS_STR);
        mPSStream.println("%%BeginProlog");
	mPSStream.println(READIMAGEPROC);
        mPSStream.println("/BD {bind def} bind def");
        mPSStream.println("/D {def} BD");
        mPSStream.println("/ISOF {");
        mPSStream.println("     dup findfont dup length 1 add dict begin {");
        mPSStream.println("             1 index /FID eq {pop pop} {D} ifelse");
        mPSStream.println("     } forall /Encoding ISOLatin1Encoding D");
        mPSStream.println("     currentdict end definefont");
        mPSStream.println("} BD");
        mPSStream.println("/NZ {dup 1 lt {pop 1} if} BD");
        mPSStream.println("/"+DrawStringName +" {");
        mPSStream.println("     moveto 1 index stringwidth pop NZ sub");
        mPSStream.println("     1 index length 1 sub NZ div 0");
        mPSStream.println("     3 2 roll ashow newpath} BD");
        mPSStream.println("/F [");
        if (mFontProps == null){
            mPSStream.println(" /Helvetica ISOF");
            mPSStream.println(" /Helvetica-Bold ISOF");
            mPSStream.println(" /Helvetica-Oblique ISOF");
            mPSStream.println(" /Helvetica-BoldOblique ISOF");
            mPSStream.println(" /Times-Roman ISOF");
            mPSStream.println(" /Times-Bold ISOF");
            mPSStream.println(" /Times-Italic ISOF");
            mPSStream.println(" /Times-BoldItalic ISOF");
            mPSStream.println(" /Courier ISOF");
            mPSStream.println(" /Courier-Bold ISOF");
            mPSStream.println(" /Courier-Oblique ISOF");
            mPSStream.println(" /Courier-BoldOblique ISOF");
        } else {
            int cnt = Integer.parseInt(mFontProps.getProperty("font.num", "9"));
            for (int i = 0; i < cnt; i++){
                mPSStream.println("    /" + mFontProps.getProperty
                           ("font." + String.valueOf(i), "Courier ISOF"));
            }
        }
        mPSStream.println("] D");

        mPSStream.println("/"+SetFontName +" {");
        mPSStream.println("     F exch get exch scalefont setfont");
        if (mFontProps == null) {
            mPSStream.println(" ("+testString+") stringwidth pop div");
        } else {
            mPSStream.println(" testString stringwidth pop div");
        }
        mPSStream.println(
         "     currentfont [1 0 0 -1 0 0]makefont exch scalefont setfont} BD");

        mPSStream.println("%%EndProlog");

        mPSStream.println("%%BeginSetup");
	/* PostScript printers can always generate uncollated copies.
	 */
	int copies = isCollated() ? 1 : getCopies();
	mPSStream.println(copies + COPIES);

        mPSStream.println("%%EndSetup");
    }

    // Inner class to run "privileged" to open the printer output stream.

    private class PrinterOpener implements java.security.PrivilegedAction {
	PrinterException pex;
	OutputStream result;

	public Object run() {
            try {
	        if (mDestType == PrintControl.FILE) {
		    spoolFile = new File(mDestination);
	        } else {
		    /* Write to a temporary file which will be spooled to 
                     * the printer then deleted. In the case that the file
                     * is not removed for some reason, request that it is
                     * removed when the VM exits.
		     */
                    spoolFile = File.createTempFile("javaprint", ".ps", null);
                    spoolFile.deleteOnExit();
		}
                result = new FileOutputStream(spoolFile);
                return result;
	    } catch (IOException ex) {
	        // If there is an IOError we subvert it to a PrinterException.
                pex = new PrinterIOException(ex);
	    }
	    return null;
        }
    }


    // Inner class to run "privileged" to invoke the system print command

    private class PrinterSpooler implements java.security.PrivilegedAction {
        PrinterException pex;

        public Object run() {
            try {
                /**
                 * Spool to the printer.
                 */
                if (spoolFile == null || !spoolFile.exists()) {
                   pex = new PrinterException("No spool file");
                   return null;
                }
                String fileName = spoolFile.getAbsolutePath();
                String execCmd[] = printExecCmd(mDestination, mOptions,
                               getJobName(), getCollatedCopies(), fileName);

                Process process = Runtime.getRuntime().exec(execCmd);
                process.waitFor();
                spoolFile.delete();

            } catch (IOException ex) {
                pex = new PrinterIOException(ex);
            } catch (InterruptedException ie) {
                pex = new PrinterException(ie.toString());
            }
            return null;
        }
    }


    /**
     * Invoked if the application cancelled the printjob.
     */
    protected void abortDoc() {
        if (mPSStream != null) {
            mPSStream.close();
        }
        java.security.AccessController.doPrivileged(
            new java.security.PrivilegedAction() {

            public Object run() {
               if (spoolFile != null && spoolFile.exists()) {
                   spoolFile.delete();
               }
               return null;
            }
        });
    }

    /**
     * Invoked by the RaterPrintJob super class
     * this method is called after that last page
     * has been imaged.
     */
    protected void endDoc() throws PrinterException {
	if (mPSStream != null) {
	    mPSStream.println(EOF_COMMENT);
	    mPSStream.close();
	}
        if (mDestType == PrintControl.PRINTER) {
	    PrinterSpooler spooler = new PrinterSpooler();
	    java.security.AccessController.doPrivileged(spooler);
	    if (spooler.pex != null) {
	        throw spooler.pex;
	    }
        }
    }

    /**
     * The RastePrintJob super class calls this method
     * at the start of each page.
     */
    protected void startPage(PageFormat pageFormat, Printable painter,
			     int index)
	throws PrinterException
    {
	    
	double paperHeight = pageFormat.getPaper().getHeight();
	int pageNumber = index + 1;

        /* Place an initial gstate on to our gstate stack.
         * It will have the default PostScript gstate
         * attributes.
         */
        mGStateStack = new ArrayList();
        mGStateStack.add(new GState());

	mPSStream.println(PAGE_COMMENT + pageNumber + " " + pageNumber);
        mPSStream.println(PAGE_SAVE);
	mPSStream.println(paperHeight + COORD_PREP);
    }

    /**
     * The RastePrintJob super class calls this method
     * at the end of each page.
     */
    protected void endPage(PageFormat format, Printable painter,
                           int index)
	throws PrinterException
    {
        mPSStream.println(PAGE_RESTORE);
	mPSStream.println(SHOWPAGE);
    }

   /**
     * Convert the 24 bit BGR image buffer represented by
     * <code>image</code> to PostScript. The image is drawn at
     * <code>(destX, destY)</code> in device coordinates.
     * The image is scaled into a square of size
     * specified by <code>destWidth</code> and
     * <code>destHeight</code>. The portion of the
     * source image copied into that square is specified
     * by <code>srcX</code>, <code>srcY</code>,
     * <code>srcWidth</code>, and srcHeight.
     */
    protected void drawImageBGR(byte[] data,
                                   float destX, float destY,
                                   float destWidth, float destHeight,
                                   float srcX, float srcY,
                                   float srcWidth, float srcHeight,
                                   int srcBitMapWidth, int srcBitMapHeight) {

	/* We draw images at device resolution so we probably need
	 * to change the current PostScript transform.
	 */
	setTransform(new AffineTransform());
	prepDrawing(false);

	int intSrcWidth = (int) srcWidth;
	int intSrcHeight = (int) srcHeight;

	mPSStream.println(IMAGE_SAVE);

	/* Create a PS string big enough to hold a row of pixels.
	 */
	int psBytesPerRow = 3 * (int) intSrcWidth;
	while (psBytesPerRow > MAX_PSSTR) {
	    psBytesPerRow /= 2;
	}

	mPSStream.println(psBytesPerRow + IMAGE_STR);

	/* Scale and translate the unit image.
	 */
	mPSStream.println("[" + destWidth + " 0 "
			  + "0 " + destHeight
			  + " " + destX + " " + destY
			  +"]concat");

	/* Color Image invocation.
	 */
	mPSStream.println(intSrcWidth + " " + intSrcHeight + " " + 8 + "["
			  + intSrcWidth + " 0 "
			  + "0 " + intSrcHeight
			  + " 0 " + 0 + "]"
			  + "/heximageSrc load false 3 colorimage");

	/* Image data.
	 */
	int index = 0;
	byte[] psData = new byte[intSrcWidth * 6];

        try {
	    /* Skip the parts of the image that are not part
	     * of the source rectangle.
	     */
	    index = (int) srcY * srcBitMapWidth;

	    for(int i = 0; i < intSrcHeight; i++) {
		
		/* Skip the left part of the image that is not
		 * part of the source rectangle.
		 */
		index += (int) srcX;

		index = toHex(data, index, psData);
		mPSStream.write(psData);
		mPSStream.println("");
	    }

	    /*
	     * If there is an IOError we subvert it to a PrinterException.
	     * Fix: There has got to be a better way, maybe define
	     * a PrinterIOException and then throw that? 
	     */
        } catch (IOException e) {
            //throw new PrinterException(e.toString());
        }

	mPSStream.println(IMAGE_RESTORE);
    }

    /**
     * Prints the contents of the array of ints, 'data'
     * to the current page. The band is placed at the
     * location (x, y) in device coordinates on the
     * page. The width and height of the band is
     * specified by the caller. Currently the data
     * is 24 bits per pixel in BGR format.
     */
    protected void printBand(byte[] data, int x, int y,
			     int width, int height)
	throws PrinterException
    {

	mPSStream.println(IMAGE_SAVE);

	/* Create a PS string big enough to hold a row of pixels.
	 */
	int psBytesPerRow = 3 * width;
	while (psBytesPerRow > MAX_PSSTR) {
	    psBytesPerRow /= 2;
	}

	mPSStream.println(psBytesPerRow + IMAGE_STR);

	/* Scale and translate the unit image.
	 */
	mPSStream.println("[" + width + " 0 "
			  + "0 " + height
			  + " " + x + " " + y
			  +"]concat");

	/* Color Image invocation.
	 */
	mPSStream.println(width + " " + height + " " + 8 + "["
			  + width + " 0 "
			  + "0 " + -height
			  + " 0 " + height + "]"
			  + "/heximageSrc load false 3 colorimage");

	/* Image data.
	 */
	int index = 0;
	byte[] psData = new byte[width * 6];

        try {
	    for(int i = 0; i < height; i++) {
		index = toHex(data, index, psData);
		mPSStream.write(psData);
		mPSStream.println("");
	    }

        } catch (IOException e) {
            throw new PrinterIOException(e);
        }

	mPSStream.println(IMAGE_RESTORE);
    }

    /**
     * Examine the metrics captured by the
     * <code>PeekGraphics</code> instance and
     * if capable of directly converting this
     * print job to the printer's control language
     * or the native OS's graphics primitives, then
     * return a <code>PSPathGraphics</code> to perform
     * that conversion. If there is not an object
     * capable of the conversion then return
     * <code>null</code>. Returning <code>null</code>
     * causes the print job to be rasterized.
     */
    
    protected Graphics2D createPathGraphics(PeekGraphics peekGraphics,
					    PrinterJob printerJob,
                                            Printable painter,
                                            PageFormat pageFormat,
                                            int pageIndex) {

	PSPathGraphics pathGraphics;
	PeekMetrics metrics = peekGraphics.getMetrics();

	/* If the application has drawn anything that
	 * out PathGraphics class can not handle then
	 * return a null PathGraphics.
	 */
        if (forcePDL == false && (forceRaster == true 
	                || metrics.hasNonSolidColors()
			|| metrics.hasCompositing())) {

	    pathGraphics = null;
	} else {

	    BufferedImage bufferedImage = new BufferedImage(8, 8,
					    BufferedImage.TYPE_INT_RGB);
	    Graphics2D bufferedGraphics = bufferedImage.createGraphics();

	    pathGraphics =  new PSPathGraphics(bufferedGraphics, printerJob,
                                               painter, pageFormat, pageIndex);
	}

	return pathGraphics;
    }

    /**
     * Intersect the gstate's current path with the
     * current clip and make the result the new clip.
     */
    protected void selectClipPath() {

	mPSStream.println(mClipOpStr);
    }

    protected void setClip(Shape clip) {

        mLastClip = clip;
    }

    protected void setTransform(AffineTransform transform) {
        mLastTransform = transform;
    }

    /**
     * Set the current PostScript font.
     * Taken from outFont in PSPrintStream.
     */
     protected boolean setFont(Font font) {
        mLastFont = font;
        return true;
    }

    /**
     * Given an array of CharsetStrings that make up a run
     * of text, this routine converts each CharsetString to
     * an index into our PostScript font list. If one or more
     * CharsetStrings can not be represented by a PostScript
     * font, then this routine will return a null array.
     */
     private int[] getPSFontIndexArray(Font font, CharsetString[] charSet) {
	int[] psFont = null;

	if (mFontProps != null) {
	    psFont = new int[charSet.length];
	}

	for (int i = 0; i < charSet.length && psFont != null; i++){

	    /* Get the encoding of the run of text.
	     */
	    CharsetString cs = charSet[i];
	    CharToByteConverter fontCS = cs.fontDescriptor.fontCharset;
	    String charset;
            /*
             * CharToByteSymbol perhaps should return "symbol" for encoding.
             * Similarly CharToByteX11Dingbats should return "dingbats"
             */
            if (fontCS instanceof CharToByteSymbol) {
                charset = "symbol";
            } else if (fontCS instanceof CharToByteX11Dingbats) {
                charset = "dingbats";
            } else {
	        charset = makeCharsetName(fontCS.getCharacterEncoding());
            }

	    String style = PlatformFont.styleStr(font.getStyle());

	    /* First we map the font name through the properties file.
	     * This mapping provides alias names for fonts, for example,
	     * "timesroman" is mapped to "serif".
	     */
            String fontName = font.getName().toLowerCase();
            fontName = fontName.replace(' ', '_');
	    String name = mFontProps.getProperty(fontName, "");

	    /* Now map the alias name, character set name, and style
	     * to a PostScript name.
	     */
	    String psName =
		mFontProps.getProperty(name + "." + charset + "." + style,
				      null);

	    if (psName != null) {

		/* Get the PostScript font index for the PostScript font.
		 */
		try {
		    psFont[i] =
			Integer.parseInt(mFontProps.getProperty(psName));

		/* If there is no PostScript font for this font name,
		 * then we want to termintate the loop and the method
		 * indicating our failure. Setting the array to null
		 * is used to indicate these failures.
		 */
		} catch(NumberFormatException e){
		    psFont = null;
		}

	    /* There was no PostScript name for the font, character set,
	     * and style so give up.
	     */
	    } else {
		psFont = null;
	    }
	}

	 return psFont;
     }


     protected boolean textOut(String str, float x, float y) {
	boolean didText = true;


	if (mFontProps == null) {
	    prepDrawing(true);

     	    mPSStream.println(x + " " + y + " moveto(" + str + ")show");
	} else {

	    prepDrawing(false);

	    FontMetrics curMetrics = Toolkit.getDefaultToolkit().getFontMetrics(mLastFont);

	    CharsetString[] acs =
		((MFontPeer)
		 (mLastFont.getPeer())).makeMultiCharsetString(str);

	    /* Get an array of indices into our PostScript name
	     * table. If all of the runs can not be converted
	     * to PostScript fonts then null is returned and
	     * we'll want to fall back to printing the text
	     * as shapes.
	     */
	    int[] psFonts = getPSFontIndexArray(mLastFont, acs);
	    if (psFonts != null) {

		for (int i = 0; i < acs.length; i++){
		    CharsetString cs = acs[i];
		    CharToByteConverter fontCS = cs.fontDescriptor.fontCharset;

		    StringBuffer nativeStr = new StringBuffer();
		    byte[] strSeg = new byte[cs.length * 2];
		    int len;
		    try {
			len = fontCS.convert(cs.charsetChars, cs.offset,
					     cs.offset + cs.length,
					     strSeg, 0, strSeg.length);

		    } catch(CharConversionException e){
			continue;
		    }

		    int desiredWidth = curMetrics.charsWidth(cs.charsetChars,
							     cs.offset,
							     cs.length);

		    nativeStr.append('<');
		    for (int j = 0; j < len; j++){
			byte b = strSeg[j];
			// to avoid encoding conversion with println()
			String hexS = Integer.toHexString(b);
			int length = hexS.length();
			if (length > 2) {
			    hexS = hexS.substring(length - 2, length);
			} else if (length == 1) {
			    hexS = "0" + hexS;
			} else if (length == 0) {
			    hexS = "00";
			}
			nativeStr.append(hexS);
		    }
		    nativeStr.append('>');
		    mPSStream.println("/testString " + nativeStr.toString() + " D");
		    mPSStream.println("% Font[" + mLastFont.getName() + ", " +
			       PlatformFont.styleStr(mLastFont.getStyle()) + ", "
			       + mLastFont.getSize() + "]");
		    getGState().emitPSFont(psFonts[i], mLastFont.getSize(), desiredWidth);

		    // out String
		    mPSStream.println(nativeStr.toString());

		    // draw string
		    int width = curMetrics.charsWidth(cs.charsetChars, cs.offset, cs.length);
		    mPSStream.println(" " + width + " " + x + " " + y);
		    mPSStream.println(DrawStringName);
		    x += width;
		}
	    } else {
		didText = false;
	    }
	}

	return didText;
     }
    /**
     * Set the current path rule to be either
     * <code>FILL_EVEN_ODD</code> (using the
     * even-odd file rule) or <code>FILL_WINDING</code>
     * (using the non-zero winding rule.)
     */
    protected void setFillMode(int fillRule) {
    
	switch (fillRule) {
	 
	 case FILL_EVEN_ODD:
	    mFillOpStr = EVEN_ODD_FILL_STR;
	    mClipOpStr = EVEN_ODD_CLIP_STR;
	    break;

	 case FILL_WINDING:
	     mFillOpStr = WINDING_FILL_STR;
	     mClipOpStr = WINDING_CLIP_STR;
	     break;
	 
	 default:
	     throw new IllegalArgumentException();
	}

    }

    /**
     * Set the printer's current color to be that
     * defined by <code>color</color>
     */
    protected void setColor(Color color) {
        mLastColor = color;
    }

    /**
     * Fill the current path using the current fill mode
     * and color.
     */
    protected void fillPath() {
	
	mPSStream.println(mFillOpStr);
    }

    /**
     * Called to mark the start of a new path.
     */
    protected void beginPath() {
	
        prepDrawing(false);
	mPSStream.println(NEWPATH_STR);

	mPenX = 0;
	mPenY = 0;
    }

    /**
     * Close the current subpath by appending a straight
     * line from the current point to the subpath's
     * starting point.
     */
    protected void closeSubpath() {
	
	mPSStream.println(CLOSEPATH_STR);

	mPenX = mStartPathX;
	mPenY = mStartPathY;
    }


    /**
     * Generate PostScript to move the current pen
     * position to <code>(x, y)</code>.
     */
    protected void moveTo(float x, float y) {

	mPSStream.println(x + " " + y + MOVETO_STR);

	/* moveto marks the start of a new subpath
	 * and we need to remember that starting
	 * position so that we know where the
	 * pen returns to with a close path.
	 */
	mStartPathX = x;
	mStartPathY = y;

	mPenX = x;
	mPenY = y;
    }
    /**
     * Generate PostScript to draw a line from the
     * current pen position to <code>(x, y)</code>.
     */
    protected void lineTo(float x, float y) {

	mPSStream.println(x + " " + y + LINETO_STR);

	mPenX = x;
	mPenY = y;
    }

    /**
     * Add to the current path a bezier curve formed
     * by the current pen position and the method parameters
     * which are two control points and an ending
     * point.
     */
    protected void bezierTo(float control1x, float control1y,
			        float control2x, float control2y,
				float endX, float endY) {

	mPSStream.println(control1x + " " + control1y
			  + " " + control2x + " " + control2y
			  + " " + endX + " " + endY
			  + CURVETO_STR);

	
	mPenX = endX;
	mPenY = endY;
    }

    
    /**
     * Return the x coordinate of the pen in the
     * current path.
     */
    protected float getPenX() {
	
	return mPenX;
    }
    /**
     * Return the y coordinate of the pen in the
     * current path.
     */
    protected float getPenY() {

	return mPenY;
    }

    /**
     * Return the x resolution of the coordinates
     * to be rendered.
     */
    protected double getXRes() {
	return PS_XRES;
    }
    /**
     * Return the y resolution of the coordinates
     * to be rendered.
     */
    protected double getYRes() {
	return PS_YRES;
    }

    /**
     * For PostScript the origin is in the upper-left of the
     * paper not at the imageable area corner.
     */
    protected double getPhysicalPrintableX(Paper p) {
        return 0;

    }

    /**
     * For PostScript the origin is in the upper-left of the
     * paper not at the imageable area corner.
     */
    protected double getPhysicalPrintableY(Paper p) {
        return 0;
    }

    protected double getPhysicalPrintableWidth(Paper p) {
        return p.getImageableWidth();
    }

    protected double getPhysicalPrintableHeight(Paper p) {
        return p.getImageableHeight();
    }

    protected double getPhysicalPageWidth(Paper p) {
        return p.getWidth();
    }

    protected double getPhysicalPageHeight(Paper p) {
        return p.getHeight();
    }

   /**
     * Returns how many times each page in the book
     * should be consecutively printed by PrintJob.
     * If the printer makes copies itself then this
     * method should return 1.
     */
    protected int getNoncollatedCopies() {
	return 1;
    }

    private String[] printExecCmd(String printer, String options,
				  String banner, int copies, String spoolFile) {
	int PRINTER = 0x1;
	int OPTIONS = 0x2;
	int BANNER  = 0x4;
	int COPIES  = 0x8;
	int pFlags = 0;
	String execCmd[];
	int ncomps = 2; // minimum number of print args
	int n = 0;

	if (printer != null && !printer.equals("") && !printer.equals("lp")) {
	    pFlags |= PRINTER;
	    ncomps+=1;
	}
	if (options != null && !options.equals("")) {
	    pFlags |= OPTIONS;
	    ncomps+=1;
	}
	if (banner != null && !banner.equals("")) {
	    pFlags |= BANNER;
	    ncomps+=1;
	}
	if (copies > 1) {
	    pFlags |= COPIES;
	    ncomps+=1;
	}
        if(System.getProperty("os.name").equals("Linux")) {
 	    execCmd = new String[ncomps];
            execCmd[n++] = "/usr/bin/lpr";
            if ((pFlags & PRINTER) != 0) {
                execCmd[n++] = new String("-P" + printer);
            }
            if ((pFlags & BANNER) != 0) {
                execCmd[n++] = new String("-T"  + banner);
            }
            if ((pFlags & COPIES) != 0) {
                execCmd[n++] = new String("-#" + new Integer(copies).toString());
            }
            if ((pFlags & OPTIONS) != 0) {
                execCmd[n++] = new String(options);
            }
        } else {
            ncomps+=1; //add 1 arg for lp
 	    execCmd = new String[ncomps];
            execCmd[n++] = "/usr/bin/lp";
            execCmd[n++] = "-c";           // make a copy of the spool file
            if ((pFlags & PRINTER) != 0) {
                execCmd[n++] = new String("-d" + printer);
            }
            if ((pFlags & BANNER) != 0) {
                execCmd[n++] = new String("-t"  + banner);
            }
            if ((pFlags & COPIES) != 0) {
                execCmd[n++] = new String("-n" + new Integer(copies).toString());
            }
            if ((pFlags & OPTIONS) != 0) {
                execCmd[n++] = new String("-o" + options);
            }
        }
        execCmd[n++] = spoolFile;
	return execCmd;
    }


    /**
     * See if 'destFile' exists and if it does confirm with
     * the user that the file should be replaced. Return true
     * if the file is okay to write to and false otherwise.
     */
    private boolean okayToCreateFile(Frame parentFrame, String filename) {
	boolean createFile = true;
	File destFile = new File(filename);

	if (destFile.exists()) {
	    String cancelCmd = getMsg("warning_button.cancel");
	    String overwriteCmd = getMsg("warning_button.overwrite");
	    PrintStatusDialog d =
		new PrintStatusDialog(parentFrame, 
				      getMsg("warning_dialog.title"),
				      getMsg("warning_msg.file_exists")
				      + filename, 
				      overwriteCmd,
				      cancelCmd);
	    d.show();
	    if (d.getCommand().equals(cancelCmd)) {
		createFile = false;
	    }
	}

	return createFile;
    }

    /**
     * Initialize ResourceBundle
     */
    static void initResource() {
	try {
	    messageRB =
		ResourceBundle.getBundle("sun.awt.motif.resources.printcontrol");
	} catch (MissingResourceException e) {
	    throw new Error("Fatal: Resource for PSPrintControl is missing.");
	}
    }

    /**
     * Get message string from resource 
     */
    static String getMsg(String key) {
	if (messageRB == null) {
	    initResource();
	}
	try {
	    return messageRB.getString(key);
	} catch (MissingResourceException e) {
	    throw new Error("Fatal: Resource for PSPrinterJob is broken, "+
			    "There is no " + key + " key in resource");
	}
    }


    /**
     * Take the 3 byte BGR values out of 'image' and turn them
     * into hexified RGB values in 'dest'. 'index' specified where
     * in 'image' to start reading data. The caller must make sure
     * that dest.length is a multiple of 6 (2 bytes per R, G, and B)
     * and that image.length is a multiple of length / 6 so that
     * dest is completely filled each time this routine returns.
     */
    private static int toHex(byte[] image, int index, byte[] dest) {
	int red, green, blue;
	int destIndex = 0;

	while(index < image.length && destIndex < dest.length) {
	    blue  = image[index++] & 0xff;
	    green = image[index++] & 0xff;
	    red   = image[index++] & 0xff;

	    // Red
	    dest[destIndex++] = hexDigits[(red & HINIBBLE_MASK)
					   >>> HINIBBLE_SHIFT];
	    dest[destIndex++] = hexDigits[red & LOWNIBBLE_MASK];

	    // Green
	    dest[destIndex++] = hexDigits[(green & HINIBBLE_MASK)
					   >>> HINIBBLE_SHIFT];
	    dest[destIndex++] = hexDigits[green & LOWNIBBLE_MASK];

	    // Blue
	    dest[destIndex++] = hexDigits[(blue & HINIBBLE_MASK)
					   >>> HINIBBLE_SHIFT];
	    dest[destIndex++] = hexDigits[blue & LOWNIBBLE_MASK];

	}

	return index;
    }

    /*
     * Currently CharToByteConverter.getCharacterEncoding() return values are
     * not fixed yet. These are used as the part of the key of
     * psfont.propeties. When those name are fixed this routine can
     * be erased.
     */
    private String makeCharsetName(String name){
	if (name.indexOf("8859_1") != -1) {
	    return "latin1";
	} else {
	    return name.toLowerCase();
	}
	
    }

    private void prepDrawing(boolean setFont) {

	/* Pop gstates until we can set the needed clip
	 * and transform or until we are at the outer most
	 * gstate.
	 */
	while (isOuterGState() == false
	       && (getGState().canSetClip(mLastClip) == false
		   || getGState().mTransform.equals(mLastTransform) == false)) {


	    grestore();
	}

	/* Set the color. This can push the color to the
	 * outer most gsave which is often a good thing.
	 */
	getGState().emitPSColor(mLastColor);

	/* We do not want to change the outermost
	 * transform or clip so if we are at the
	 * outer clip the generate a gsave.
	 */
	if (isOuterGState()) {
	    gsave();
	    getGState().emitTransform(mLastTransform);
	    getGState().emitPSClip(mLastClip);
	}

	/* Set the font if we have been asked to. It is
	 * important that the font is set after the
	 * transform in order to get the font size
	 * correct.
	 */
	if (setFont) {
	    getGState().emitPSFont(mLastFont);
	}

    }

    /**
     * Return the GState that is currently on top
     * of the GState stack. There should always be
     * a GState on top of the stack. If there isn't
     * then this method will throw an IndexOutOfBounds
     * exception.
     */
    private GState getGState() {
	int count = mGStateStack.size();
	return (GState) mGStateStack.get(count - 1);;
    }

    /**
     * Emit a PostScript gsave command and add a
     * new GState on to our stack which represents
     * the printer's gstate stack.
     */
    private void gsave() {
	GState oldGState = getGState();
	mGStateStack.add(new GState(oldGState));
	mPSStream.println(GSAVE_STR);
    }

    /**
     * Emit a PostScript grestore command and remove
     * a GState from our stack which represents the
     * printer's gstate stack.
     */
    private void grestore() {
	int count = mGStateStack.size();
	mGStateStack.remove(count - 1);
	mPSStream.println(GRESTORE_STR);
    }

    /**
     * Return true if the current GState is the
     * outermost GState and therefore should not
     * be restored.
     */
    private boolean isOuterGState() {
	return mGStateStack.size() == 1;
    }

    /**
     * A stack of GStates is maintained to model the printer's
     * gstate stack. Each GState holds information about
     * the current graphics attributes.
     */
    private class GState{
	Color mColor;
	Shape mClip;
	Font mFont;
	AffineTransform mTransform;

	GState() {
	    mColor = Color.black;
	    mClip = null;
	    mFont = null;
	    mTransform = new AffineTransform();
	}

	GState(GState copyGState) {
	    mColor = copyGState.mColor;
	    mClip = copyGState.mClip;
	    mFont = copyGState.mFont;
	    mTransform = copyGState.mTransform;
	}

	boolean canSetClip(Shape clip) {

	    return mClip == null || mClip.equals(clip);
	}


	void emitPSClip(Shape clip) {
	    if (clip != null
		&& (mClip == null || mClip.equals(clip) == false)) {
	
		convertToPSPath(clip.getPathIterator(new AffineTransform()));
		selectClipPath();
		mClip = clip;
	    }
	}

	void emitTransform(AffineTransform transform) {

	    if (transform != null && transform.equals(mTransform) == false) {
		double[] matrix = new double[6];
		transform.getMatrix(matrix);
		mPSStream.println("[" + matrix[0]
				  + " " + matrix[1]
				  + " " + matrix[2]
				  + " " + matrix[3]
				  + " " + matrix[4]
				  + " " + matrix[5]
				  + "] concat");

		mTransform = transform;
	    }
	}

	void emitPSColor(Color color) {
	    if (color != null && color.equals(mColor) == false) {
		float[] rgb = color.getColorComponents(null);

		/* If the color is a gray value then use
		 * setgray.
		 */
		if (rgb[0] == rgb[1] && rgb[1] == rgb[2]) {
		    mPSStream.println(rgb[0] + SETGRAY_STR);
		
		/* It's not gray so use setrgbcolor.
		 */
		} else {
		    mPSStream.println(rgb[0] + " "
				      + rgb[1] + " "
				      + rgb[2] + " "
				      + SETRGBCOLOR_STR);
		}

		mColor = color;

	    }
	}

	void emitPSFont(Font font) {

	    if ((mFont == null && font != null)
		|| mFont.equals(font) == false) {

		mCurMetrics = Toolkit.getDefaultToolkit().getFontMetrics(font);

		int desiredWidth = mCurMetrics.stringWidth(testString);
		int fontIndex = font.getStyle();

		// NB leaving helvetica and times in here, but in fact we do
		// not support these and they won't work
		if ((font.getFamily()).toLowerCase().startsWith("helvetica") ||
		    (font.getFamily()).toLowerCase().startsWith("sansserif")) {
		    fontIndex = 0;
		} else if ((font.getFamily()).toLowerCase().startsWith("dialog")) {
		    fontIndex = 0;
		} else if ((font.getFamily()).toLowerCase().startsWith("times") ||
			   (font.getFamily()).toLowerCase().startsWith("serif")) {
		    fontIndex = 4;
		} else {
		    fontIndex = 8;
		}
		if (font.isBold())
		    fontIndex += 1;
		if (font.isItalic())
		    fontIndex += 2;
		mPSStream.println("% Font["
			   +font.getFamily()+", "
			   +font.getStyle()+", "
			   +font.getSize()
			   +"]");
		getGState().emitPSFont(fontIndex, font.getSize(), desiredWidth);
	    }

	    mFont = font;

	}
 
	void emitPSFont(int psFontIndex, int fontSize, int desiredWidth) {
	    mPSStream.println(desiredWidth + " " + fontSize + " " +
			      psFontIndex + " " + SetFontName);
	}

       /**
	* Given a Java2D <code>PathIterator</code> instance,
	* this method translates that into a PostScript path..
	*/
	private void convertToPSPath(PathIterator pathIter) {

	    float[] segment = new float[6];
	    int segmentType;

	    /* Map the PathIterator's fill rule into the PostScript
	     * fill rule.
	     */
	    int fillRule;
	    if (pathIter.getWindingRule() == PathIterator.WIND_EVEN_ODD) {
		fillRule = FILL_EVEN_ODD;
	    } else {
		fillRule = FILL_WINDING;
	    }
	    setFillMode(fillRule);

	    beginPath();

	    while (pathIter.isDone() == false) {
		segmentType = pathIter.currentSegment(segment);

		switch (segmentType) {
		 case PathIterator.SEG_MOVETO:
		    moveTo(segment[0], segment[1]);
		    break;
		 
		 case PathIterator.SEG_LINETO:
		    lineTo(segment[0], segment[1]);
		    break;

		/* Convert the quad path to a bezier.
		 */
		 case PathIterator.SEG_QUADTO:
		    float lastX = getPenX();
		    float lastY = getPenY();
		    float c1x = lastX + (segment[0] - lastX) * 2 / 3;
		    float c1y = lastY + (segment[1] - lastY) * 2 / 3;
		    float c2x = segment[2] - (segment[2] - segment[0]) * 2/ 3;
		    float c2y = segment[3] - (segment[3] - segment[1]) * 2/ 3;
		    bezierTo(c1x, c1y,
					  c2x, c2y,
					  segment[2], segment[3]);
		    break;

		 case PathIterator.SEG_CUBICTO:
		    bezierTo(segment[0], segment[1],
					  segment[2], segment[3],
					  segment[4], segment[5]);
		    break;

		 case PathIterator.SEG_CLOSE:
		    closeSubpath();
		    break;
		}


		pathIter.next();
	    }
	}

    }



}
