/*
 * @(#)X11Clipboard.java	1.17 01/02/09
 *
 * Copyright 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.motif;

import java.awt.datatransfer.*;
import java.io.IOException;
import java.io.StringReader;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.HashSet;
import java.util.List;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import sun.awt.AppContext;
import sun.awt.SunClipboard;

/**
 * A class which interfaces with the X11 selection service in order
 * to support data transfer via clipboard operations.
 */
public class X11Clipboard extends SunClipboard implements X11SelectionHolder {

    private X11Selection clipboardSelection = new X11Selection("CLIPBOARD");
    private X11Selection primarySelection = new X11Selection("PRIMARY");

    /**
     * Creates a system clipboard object.
     */
    public X11Clipboard(String name) {
        super(name);
        contents = null;	
    }

    public synchronized void setContents(Transferable contents,
					 ClipboardOwner owner) {
	if ((this.owner != null) && (this.owner != owner)) {
	    this.owner.lostOwnership(this, this.contents);
            clipboardSelection.clearContentsData();
	    primarySelection.clearContentsData();
	}
	this.owner = owner;
	this.contents = contents;

        initContext(contents);

	if (clipboardSelection.getSelectionOwnership(contents, this)) {
            // Post the same contents to the PRIMARY selection and proceed
            // even if it is not posted successfully.
            primarySelection.getSelectionOwnership(contents, null);
	} else {
	    // Need to figure out how to inform owner the request failed...
	    this.owner = null;
            this.contents = null;
	}
    }

    public synchronized Transferable getContents(Object requestor) {
        if (contents != null) {
            return contents;
        }

        ArrayList list = new ArrayList(2);
        list.add(new FullyRenderedTransferable(clipboardSelection));
        list.add(new FullyRenderedTransferable(primarySelection));
        Transferable compositeSelection =
            new CompositeTransferable(Collections.unmodifiableList(list));

        return compositeSelection;
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void lostSelectionOwnership() {

        // NOTE: this code makes sense until we support multiple system 
        // clipboards. We should forcibly clear the PRIMARY selection 
        // contents, since we are going to notify the owner that the 
        // ownership is lost and so it may free the resouces associated 
        // with the contents.
        primarySelection.clearContentsData();

        lostOwnershipImpl();
    }
    protected  synchronized void lostOwnership() {
        if (this.owner != null) {
	    this.owner.lostOwnership(this, this.contents);
	    this.owner = null;
	}

	// fix 4180148: when losing ownership always set contents to the
	// current clipboard selection, even if the owner is null.

	this.contents = null;
    }
}

class CompositeTransferable implements Transferable {
    private final List transferables;
    private final DataFlavor[] flavors;

    public CompositeTransferable(List transferables) {
	this.transferables = transferables;

	HashSet flavorsSet = new HashSet();

	for (Iterator iter = transferables.iterator(); iter.hasNext(); ) {
	    Transferable transferable = (Transferable)iter.next();
	    DataFlavor[] flavors = transferable.getTransferDataFlavors();
            flavorsSet.addAll(Arrays.asList(flavors));
	}

        flavors = MDataTransferer.getInstance().setToSortedDataFlavorArray(flavorsSet);
    }
    public DataFlavor[] getTransferDataFlavors() {
        return flavors;
    }
    public boolean isDataFlavorSupported(DataFlavor flavor) {
	for (Iterator iter = transferables.iterator(); iter.hasNext(); ) {
	    Transferable transferable = (Transferable)iter.next();
	    if (transferable.isDataFlavorSupported(flavor)) {
		return true;
	    }
	}

	return false;
    }
    public Object getTransferData(DataFlavor flavor)
        throws UnsupportedFlavorException, IOException
    {
	IOException io = null;

	for (Iterator iter = transferables.iterator(); iter.hasNext(); ) {
	    Transferable transferable = (Transferable)iter.next();
	    if (transferable.isDataFlavorSupported(flavor)) {
		try {
		    return transferable.getTransferData(flavor);
		} catch (IOException e) {
		    // If a transferable throws an IOException, then try again
		    // with the next transferable. However, if no transferable
		    // successfully returns transfer data, then rethrow the
		    // first IOException we saw.
		    //
		    // This algorithm allows us to return the PRIMARY selection
		    // if the native process which owned CLIPBOARD has been
		    // terminated.
		    if (io == null) {
			io = e;
		    }
		}
	    }
	}

	if (io != null) {
	    throw io;
	} else {
	    throw new UnsupportedFlavorException(flavor);
	}
    }
}

/**
 * Since there is no way for client code to synchronize access to the
 * system Clipboard, we cannot provide a Transferable which creates
 * transfer data for a particular DataFlavor on demand. Instead, we have
 * to pre-render data for all possible DataFlavors.
 */
final class FullyRenderedTransferable implements Transferable {
    private final HashMap flavorsToData = new HashMap();
    private final DataFlavor[] flavors;

    static class DataFactory {
        final int format;
        final byte[] data;
        DataFactory(int format, byte[] data) {
            this.format = format;
            this.data   = data;
        }

        public Object getTransferData(DataFlavor flavor) throws IOException { 
            return MDataTransferer.getInstance().
                translateBytes(data, flavor, format);
        }
    }

    private static DataFlavor plainTextStringFlavor;
    static {
        try {
            plainTextStringFlavor =
                new DataFlavor("text/plain;class=java.lang.String");
        } catch (ClassNotFoundException e) {
            // not possible
        }
    }

    public FullyRenderedTransferable(X11Selection selection) {
        int[] formats = selection.getSelectionTargets();
        
        // Special hack to maintain backwards-compatibility with the
        // brokenness of StringSelection. This hack is now codified in
        // the javadoc for Toolkit.getSystemClipboard().
        //
        // If plain text data is available on the Clipboard, make sure
        // that we export DataFlavor.stringFlavor (representation: a
        // String) and DataFlavor.plainTextFlavor (representation: a
        // Reader; particularly, a StringReader).
        int textFormatTranslated = 0;
        for (int i = 0; i < formats.length; i++) {
            if ((formats[i] == MDataTransferer.StringAtom ||
                 formats[i] == MDataTransferer.TextAtom) &&
                textFormatTranslated < formats[i])
            {
                try {
                    byte[] data = selection.getSelectionData(formats[i]);
                    if (data != null) {
                        Object val = MDataTransferer.getInstance().
                            translateBytes(data, 
                                           plainTextStringFlavor,
                                           formats[i]);
                        DataFactory factory = new DataFactory(formats[i], data);
                        flavorsToData.put(DataFlavor.stringFlavor, factory);
                        flavorsToData.put(DataFlavor.plainTextFlavor, factory);
                        textFormatTranslated = formats[i];
                    }
                } catch (IOException e) {
                    if (textFormatTranslated == 0) {
                        flavorsToData.put(DataFlavor.stringFlavor, e);
                        flavorsToData.put(DataFlavor.plainTextFlavor, e);
                    }
                } catch (Throwable e) {
                    e.printStackTrace();
                }
            }
        }


        // Get all of the target DataFlavors into which 
        // the Clipboard's data can be translated. Then, 
        // for each flavor, translate the data and store 
        // the resulting Object. If an exception is thrown, 
        // store the exception instead and rethrow it if 
        // the transfer data is ever requested in that flavor.
        Map flavorsForFormats =
            MDataTransferer.getInstance().
            getFlavorsForFormats(formats, X11Selection.flavorMap);
        for (Iterator iter = flavorsForFormats.keySet().iterator();
             iter.hasNext(); )
        {
            DataFlavor flavor = (DataFlavor)iter.next();
            
            if (!flavorsToData.containsKey(flavor)) {
                int format = ((Integer)flavorsForFormats.get(flavor)).
                    intValue();
                try {
                    byte[] data = selection.getSelectionData(format);
                    DataFactory factory = new DataFactory(format, data);
                    flavorsToData.put(flavor, factory);
                } catch (IOException e) {
                    flavorsToData.put(flavor, e);
                } catch (Throwable e) {
                    e.printStackTrace();
                }
            }
        }

        flavors = MDataTransferer.getInstance().
            setToSortedDataFlavorArray(flavorsToData.keySet(),
                                       flavorsForFormats);
    }

    public DataFlavor[] getTransferDataFlavors() {
        return flavors;
    }
    public boolean isDataFlavorSupported(DataFlavor flavor) {        
        return flavorsToData.containsKey(flavor);
    }

    /**
     * If an exception was thrown during the data conversion, rethrow that
     * exception here.
     */
    public Object getTransferData(DataFlavor flavor)
        throws UnsupportedFlavorException, IOException
    {
        if (!isDataFlavorSupported(flavor)) {
            throw new UnsupportedFlavorException(flavor);
        }
        Object ret = flavorsToData.get(flavor);
        if (ret instanceof IOException) {
            throw (IOException)ret;
        } else if (ret instanceof DataFactory) {
            DataFactory factory = (DataFactory)ret;
            if (DataFlavor.stringFlavor.equals(flavor) ||
                DataFlavor.plainTextFlavor.equals(flavor)) {
                ret = factory.getTransferData(plainTextStringFlavor);
                if (DataFlavor.plainTextFlavor.equals(flavor) &&
                    ret instanceof String) {
                    ret = new StringReader((String)ret);
                }
            } else {
                ret = factory.getTransferData(flavor);
            }
        }
        return ret;
    }
}
