/*
 * @(#)X11Graphics.java	1.162 01/02/09
 *
 * Copyright 1995-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
package sun.awt.motif;

import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.awt.geom.AffineTransform;
import java.awt.BasicStroke;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import sun.awt.image.ByteComponentRaster;
import sun.awt.image.IntegerComponentRaster;
import java.awt.image.ColorModel;
import java.awt.color.ColorSpace;
import java.awt.image.ColorConvertOp;
import java.awt.Composite;
import java.awt.FontMetrics;
import java.awt.GraphicsConfiguration;
import java.awt.font.GlyphVector;
import java.awt.font.FontRenderContext;
import java.awt.font.LineMetrics;
import java.awt.image.Raster;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.awt.image.WritableRaster;
import java.awt.image.ImageObserver;
import sun.awt.DebugHelper;
import sun.awt.PlatformFont;
import sun.java2d.SunGraphics2D;
import sun.java2d.OutputManager;
import sun.awt.image.ImageRepresentation;
import sun.awt.X11GraphicsDevice;
import sun.awt.X11GraphicsConfig;
import sun.java2d.pipe.SimpleRenderPipe;
import sun.java2d.pipe.PixelDrawPipe;
import sun.java2d.pipe.PixelFillPipe;
import sun.java2d.pipe.ShapeDrawPipe;
import sun.java2d.pipe.TranslateablePipe;
import sun.java2d.pipe.LineStylePipe;
import sun.java2d.pipe.PixelToShapeConverter;
import sun.java2d.pipe.ShapeToPolyConverter;
import sun.java2d.pipe.DrawToFillConverter;
import sun.java2d.pipe.TextPipe;
import sun.java2d.pipe.SpanIterator;
import sun.java2d.loops.GraphicsPrimitiveMgr;
import sun.java2d.loops.GraphicsPrimitive;
import sun.java2d.loops.ImageData;
import sun.java2d.loops.RasterOutputManager;
import sun.java2d.DeviceClipping;
import sun.java2d.HackWorkAround2DIndexed;
import sun.java2d.loops.DrawString;

/**
 * X11Graphics is an object that encapsulates a graphics context for a
 * particular canvas. 
 * 
 * @version 1.119 12/16/97
 * @author Sami Shaio
 */

public class X11Graphics
    extends SunGraphics2D
    implements PixelDrawPipe,
	       PixelFillPipe,
	       TranslateablePipe,
	       LineStylePipe,
	       SimpleRenderPipe,
               TextPipe,
               DeviceClipping
{
    private static final DebugHelper dbg =
        DebugHelper.create(X11Graphics.class);

    static String myDevID = "X11Graphics";
    static String myDevIDMono = "X11MonoPixmap";
    long pData;
    MComponentPeer peer;
    private ImageData cachedID;
    private ImageData cachedMonoID;

    private boolean disposed = false;

    private native void createFromComponent(MComponentPeer w);
    private native void createFromGraphics(X11Graphics g);
    private native void pSetFont(Font f);
    private native void pSetForeground(Color c);

    private native static void cacheInit();
    private static boolean usesPlatformFont;

    static {
        // 4154170:  Need to ensure the the toolkit is initialized prior
        // to executing this initializer
        MToolkit toolkit = (MToolkit)Toolkit.getDefaultToolkit();

        cacheInit();
	// This line triggers the RasterOutputManager class to generate
	// its singleton instance so that the general loops mentioned
	// in its constructor will be inserted into the list first.
	// REMIND: There are better ways to trigger this registration...
        OutputManager manager = RasterOutputManager.getManager();
        GraphicsPrimitiveMgr.register(new X11GIFAcceleratorLoops(), myDevID);
        GraphicsPrimitiveMgr.register(new X11OneBitMono(), myDevIDMono);
        usesPlatformFont = RasterOutputManager.usesPlatformFont();
    }

    private Throwable constructionStackTrace;

    {
        if (dbg.on) {
	    constructionStackTrace = new Exception().fillInStackTrace();
	}
    }

    X11Graphics(MComponentPeer comp) {
        this.peer = comp;
        int types [] = {GraphicsPrimitive.TYPE_1BIT_MONO};
        drawStringClass = (DrawString) GraphicsPrimitiveMgr.locate(
                DrawString.getMethodSignature(),types,myDevIDMono);
	createFromComponent(comp);
	backgroundColor = comp.target.getBackground();
        Dimension size = comp.target.getSize();
	setDevClip(0, 0, size.width, size.height);
        this.serialNum = comp.serialNum;
    }

    /**
     * Create a new X11Graphics Object based on this one.
     */
    public Object clone() {
	X11Graphics g = (X11Graphics) super.clone();
	g.pData = 0;
	g.createFromGraphics(this);
	return g;
    }

    public ColorModel getDeviceColorModel() {
	return peer.getColorModel();
    }

    public void validatePipe() {
        RasterOutputManager.getManager().validatePipe(this);
    }

    private int serialNum = -1;

    /*
     * Intersect usrClip and devClip to form the composite clip
     */
    public Rectangle getCompBounds() {
	if (this.serialNum != peer.serialNum) {
	    Dimension size = peer.target.getSize();
	    setDevClip(0, 0, size.width, size.height);
	    this.serialNum = peer.serialNum;
	}
	return super.getCompBounds();
    }

    public native void setOrigin(int tx, int ty);

    public void setLineAttrs(float scale, float linewidth,
			     float dashes[], float phase,
			     int endcaps, int linejoins) {
	int width = Math.round(linewidth * scale);
	byte xdashes[];
	int dashoff;
	int index = 0;
	if (dashes != null) {
	    int len = dashes.length;
	    xdashes = new byte[len];
	    float rem = 0.0f;
	    do {
		for (int i = 0; i < len; i++) {
		    rem += dashes[i] * scale;
		    while (rem > 0.5f) {
			int need = (rem > 255) ? 2 : 1;
			if (index + need >= xdashes.length) {
			    byte newbuf[] = new byte[xdashes.length * 2];
			    System.arraycopy(xdashes, 0, newbuf, 0, index);
			    xdashes = newbuf;
			}
			int dash = (rem > 255) ? 255 : Math.round(rem);
			if (dash == 0) break;  // prevent infinite loops?
			xdashes[index++] = (byte) dash;
			if (rem > 255) {
			    xdashes[index++] = 0;
			}
			rem -= dash;
		    }
		}
	    } while (Math.abs(rem) > 0.1f && index < 100);
	    dashoff = Math.round(phase * scale);
	} else {
	    xdashes = null;
	    dashoff = 0;
	}
	if (width == 1 && xdashes == null) {
	    width = 0;
	}
	setLineAttrs(width, endcaps, linejoins, xdashes, index, dashoff);
    }

    /**
     * Return the device configuration associated with this Graphics2D.
     */
    public GraphicsConfiguration getDeviceConfiguration() {
        if (peer != null) {
	    GraphicsConfiguration gc = peer.getGraphicsConfiguration();
	    if (gc != null) {
	        return gc;
	    }
	}
        return MToolkit.config;
    }

    /**
     * Return the OutputManager object assigned to manage this Graphics2D.
     */
    public OutputManager getOutputManager() {
        return RasterOutputManager.getManager();
    }

    /*
     * Subclasses should override disposeImpl() instead of dispose(). Client
     * code should always invoke dispose(), never disposeImpl().
     *
     * An X11Graphics cannot be used after being disposed.
     */
    protected native void disposeImpl();
    public final void dispose() {
        boolean call_disposeImpl = false;

        if (!disposed) {
            synchronized (this) {
                if (!disposed) {
                    disposed = call_disposeImpl = true;
                }
            }
        }

        if (call_disposeImpl) {
            disposeImpl();
        }
    }

    public void finalize() {
        if (dbg.on && !disposed) {
            System.err.println("Warning: X11Graphics disposal relied on finalization.");
	    constructionStackTrace.printStackTrace();
        }
        dispose();
    }

    public void setFont(Font font) {
	if ((font != null) && (this.font != font)) {
            super.setFont(font);
	    pSetFont(font);
	}
    }
    
    /**
     * Gets font metrics for the given font.
     */
    public FontMetrics getFontMetrics(Font font) {
	if (textpipe == invalidpipe) {
	    validatePipe();
	}
	if (textpipe == this && usesPlatformFont) {
            return X11FontMetrics.getFontMetrics(font);
	} else {
            return super.getFontMetrics(font);
	}
    }

    /**
     * Sets the foreground color.
     */
    public void setColor(Color c) {
	if (c != null) {
	    super.setColor(c);
	    pSetForeground(c);
	}
    }

    /**
     * Sets the foreground color.
     */
    public void setPaint(Paint p) {
	super.setPaint(p);
	if (p instanceof Color) {	// instanceof protects against null
	    pSetForeground((Color) p);
	}
    }

    /**
     * Sets the paint mode to overwrite the destination with the
     * current color. This is the default paint mode.
     */
    public native void pSetPaintMode();

    /**
     * Sets the paint mode to alternate between the current color
     * and the given color.
     */
    public native void pSetXORMode(Color c1);

    public void setPaintMode() {
        super.setPaintMode();
        pSetPaintMode();
    }
    
    public void setXORMode(Color xorColor) {
        super.setXORMode(xorColor);
        pSetXORMode(xorColor);
    }

    public native void changeClip(int X, int Y, int W, int H, boolean set);
    public native void removeClip();

    public native void setLineAttrs(int width, int linecap, int linejoin,
				    byte[] dashes, int numdashes, int dashoff);

    /** Clears the rectangle indicated by x,y,w,h. */
    public native void devClearRect(int x, int y, int w, int h);
    public void clearRect(int x, int y, int w, int h) {
	if (fillpipe == invalidpipe) {
	    validatePipe();
	}
	if (fillpipe == this) { 
	    devClearRect(x, y, w, h);
	} else {
	    super.clearRect(x, y, w, h);
	}
    }

    /** Fills the given rectangle with the foreground color. */
    public native void devFillRect(int X, int Y, int W, int H);

    public void fillRect(SunGraphics2D sg,
			 int x, int y, int width, int height) {
	devFillRect(x, y, width, height);
    }

    /** Draws the given rectangle. */
    public native void devDrawRect(int X, int Y, int W, int H);

    public void drawRect(SunGraphics2D sg,
			 int x, int y, int width, int height) {
	devDrawRect(x, y, width, height);
    }

// implementation of textpipe for platform fonts
    public Object drawStringPlatform(SunGraphics2D g2d, String s,
                            float x, float y, Object cs)
    {
	PlatformFont pf = ((PlatformFont) font.getPeer());
	if (pf.mightHaveMultiFontMetrics()) {
	    drawMFChars(s.toCharArray(), 0, s.length(), (int)x, (int)y, font);
	} else {
            drawStringWidth(s, (int)x, (int)y);
	}
        return null;
    }

    public Object drawGlyphVectorPlatform(
            SunGraphics2D g2d, GlyphVector g, float x, float y, Object cs)
    {
        return super.drawGlyphVector(g2d, g, x, y, cs);
    }

    public Object drawCharsPlatform(SunGraphics2D g2d,
            char data[], int offset, int length, int x, int y, Object cs) {
	PlatformFont pf = ((PlatformFont) font.getPeer());
	if (pf.mightHaveMultiFontMetrics()) {
	    drawMFChars(data, offset, length, x, y, font);
	} else {
	    drawSFChars(data, offset, length, x, y);
	}
        return null;
    }

    public Object drawBytesPlatform(SunGraphics2D g2d,
            byte data[], int offset, int length, int x, int y, Object cs) {
	devDrawBytes(data, offset, length, x, y);
        return null;
    }
// end of implementation of textpipe for platform fonts

    private native void drawSFChars(char[] chars, int offset, int length, int x, int y);
    
    private native void drawMFChars(char chars[], int offset, int length, int x, int y, Font font);

    /** Draws the given byte array. */
    public native void devDrawBytes(byte data[], int offset, int length,
				  int x, int y);

    /** Draws the given string and returns the length of the drawn
      string in pixels.  If font is not set then returns -1. */
    public native int drawStringWidth(String str, int x, int y);

    /** Draws the given line. */
    public native void devDrawLine(int x1, int y1, int x2, int y2);

    public void drawLine(SunGraphics2D sg,
			 int x1, int y1, int x2, int y2) {
	devDrawLine(x1, y1, x2, y2);
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     ImageObserver observer) {

	if (width == 0 || height == 0) {
	    return true;
	}

        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            return super.drawImage((BufferedImage)img, x, y, width, height,
                                   null, observer);
        }
        
	X11Image ximg = (X11Image) img;
	if (ximg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
        
	ImageRepresentation ir = ximg.getImageRep();
	return ir.drawToBufImage(this, (sun.awt.image.Image) img, x, y,
                                 width, height, null, observer);
    }

    /**
     * Draws an image at x,y in nonblocking mode with a solid background
     * color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, Color bg,
			     ImageObserver observer) {
        if (img instanceof BufferedImage) {
            super.drawImage((BufferedImage)img, x, y, bg, observer);
            return true;
        }
                                     
	X11Image ximg = (X11Image) img;
	if (ximg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = ximg.getImageRep();
        return ir.drawToBufImage(this, (sun.awt.image.Image) img, x, y,
                                 bg, observer);
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * solid background color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
 			     Color bg, ImageObserver observer) {
	if (width == 0 || height == 0) {
	    return true;
	}

        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            return super.drawImage((BufferedImage)img, x, y, width, height,
                                   bg, observer);
        }
        
	X11Image ximg = (X11Image) img;
	if (ximg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = ximg.getImageRep();
	return ir.drawToBufImage(this, (sun.awt.image.Image) img, x, y,
                                 width, height, bg, observer);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle
     * in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     ImageObserver observer) {
        return drawImage(img, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2,
                         null, observer);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle in
     * nonblocking mode with a solid background color and a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     Color bgcolor, ImageObserver observer) {
        if (img instanceof BufferedImage) {
            return super.drawImage((BufferedImage)img, dx1, dy1, dx2, dy2,
                                   sx1, sy1, sx2, sy2, bgcolor, observer);
        }
        // REMIND: Guess we don't draw a line?
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}

        // REMIND: Guess we don't draw a line?
        if (sx1 == sx2 || sy1 == sy2) {
            return true;
        }
        
	X11Image ximg = (X11Image) img;
	if (ximg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = ximg.getImageRep();
	return ir.drawToBufImage(this, (sun.awt.image.Image) img,
                                 dx1, dy1, dx2, dy2,
                                 sx1, sy1, sx2, sy2,
                                 bgcolor, observer);
    }

    /**
     * Draw an image, applying a transform from image space into user space
     * before drawing.
     * The transformation from user space into device space is done with
     * the current transform in the Graphics2D.
     * The given transformation is applied to the image before the
     * transform attribute in the Graphics2D state is applied.
     * The rendering attributes applied include the clip, transform,
     * paint or color and composite attributes. Note that the result is
     * undefined, if the given transform is non-invertible.
     * @param img The image to be drawn.
     * @param xform The transformation from image space into user space.
     * @param observer The image observer to be notified on the image producing
     * progress.
     * @see #transform
     * @see #setComposite
     * @see #setClip
     */
    public boolean drawImage(Image img,
                             AffineTransform xform,
                             ImageObserver observer) {
        if (img instanceof BufferedImage) {
            super.drawImage((BufferedImage)img, xform, 0, 0, null, observer);
            return true;
        }
        
	X11Image ximg = (X11Image) img;
	if (ximg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
                return false;
	    }
	}
	ImageRepresentation ir = ximg.getImageRep();
        return ir.drawToBufImage(this, (sun.awt.image.Image) img, xform,
                                 observer);
    }
    
    /**
     * Draws an image at x,y in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img, int x, int y, ImageObserver observer) {
        if (img instanceof BufferedImage) {
            return super.drawImage((BufferedImage)img, x, y, null, observer);
        }
                                     
	X11Image ximg = (X11Image) img;
	if (ximg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = ximg.getImageRep(); 
        return ir.drawToBufImage(this, (sun.awt.image.Image) img, x, y,
                                 null, observer);
    }

    
    /**
     * Copies an area of the canvas that this graphics context paints to.
     * @param X the x-coordinate of the source.
     * @param Y the y-coordinate of the source.
     * @param W the width.
     * @param H the height.
     * @param dx the x-coordinate of the destination.
     * @param dy the y-coordinate of the destination.
     */
    public native void devCopyArea(int X, int Y, int W, int H, int dx, int dy);
    public void copyArea(int X, int Y, int W, int H, int dx, int dy) {
	if (fillpipe == invalidpipe) {
	    validatePipe();
	}
	if (fillpipe == this) {
	    devCopyArea(X, Y, W, H, dx, dy);
	} else {
	    super.copyArea(X, Y, W, H, dx, dy);
	}
    }


    /** Draws a rounded rectangle. */
    public native void devDrawRoundRect(int x, int y, int w, int h,
					int arcWidth, int arcHeight);

    public void drawRoundRect(SunGraphics2D sg,
			      int x, int y, int width, int height,
			      int arcWidth, int arcHeight) {
	devDrawRoundRect(x, y, width, height, arcWidth, arcHeight);
    }

    /** Draws a filled rounded rectangle. */
    public native void devFillRoundRect(int x, int y, int w, int h,
					int arcWidth, int arcHeight);

    public void fillRoundRect(SunGraphics2D sg,
			      int x, int y, int width, int height,
			      int arcWidth, int arcHeight) {
	devFillRoundRect(x, y, width, height, arcWidth, arcHeight);
    }

    /** Draws a bunch of lines defined by an array of x points and y points */
    private native void devDrawPoly(int xPoints[], int yPoints[],
				    int nPoints, boolean close);

    public void drawPolys(SunGraphics2D sg,
			  int[] xPoints, int[] yPoints, int nPoints[],
			  int numpolys, boolean close) {
	for (int i = 0; i < numpolys; i++) {
	    int numpts = nPoints[i];
	    devDrawPoly(xPoints, yPoints, numpts, close);
	    System.arraycopy(xPoints, numpts, xPoints, 0,
			     xPoints.length - numpts);
	    System.arraycopy(yPoints, numpts, yPoints, 0,
			     yPoints.length - numpts);
	}
    }

    public void drawPolyline(SunGraphics2D sg,
			     int[] xPoints, int[] yPoints, int nPoints) {
	devDrawPoly(xPoints, yPoints, nPoints, false);
    }

    public void drawPolygon(SunGraphics2D sg,
			    int[] xPoints, int[] yPoints, int nPoints) {
	devDrawPoly(xPoints, yPoints, nPoints, true);
    }

    /** Fills a polygon with the current fill mask */
    public native void devFillPolygon(int xPoints[], int yPoints[],
				      int nPoints, boolean evenodd);

    public void fillPolys(SunGraphics2D sg,
			  int[] xPoints, int[] yPoints, int nPoints[],
			  int numpolys, boolean evenodd) {
	if (numpolys < 1) {
	    return;
	}
	int numpts;
	if (numpolys > 1) {
	    Polygon pg = new Polygon();
	    int index = 0;
	    int x = 0, y = 0;
	    int mx, my;
	    for (int i = 0; i < numpolys; i++) {
		numpts = nPoints[i];
		if (numpts < 2) {
		    index += numpts;
		} else {
		    pg.addPoint(0, 0);
		    mx = xPoints[index];
		    my = yPoints[index];
		    for (int j = 0; j < numpts; j++) {
			x = xPoints[index];
			y = yPoints[index];
			pg.addPoint(x, y);
			index++;
		    }
		    if (x != mx || y != my) {
			pg.addPoint(mx, my);
		    }
		    pg.addPoint(0, 0);
		}
	    }
	    xPoints = pg.xpoints;
	    yPoints = pg.ypoints;
	    numpts = pg.npoints;
	} else {
	    numpts = nPoints[0];
	}
	devFillPolygon(xPoints, yPoints, numpts, evenodd);
    }

    /** Fills a set of spans with the current fill */
    public native void devFillSpans(SpanIterator si, long iterator);

    public void fillSpans(SunGraphics2D sg, SpanIterator si) {
	devFillSpans(si, si.getNativeIterator());
    }

    public void fillPolygon(SunGraphics2D sg,
			    int[] xPoints, int[] yPoints, int nPoints) {
	devFillPolygon(xPoints, yPoints, nPoints, true);
    }

    /** Draws an oval to fit in the given rectangle */
    public native void devDrawOval(int x, int y, int w, int h);

    public void drawOval(SunGraphics2D sg,
			 int x, int y, int width, int height) {
	devDrawOval(x, y, width, height);
    }

    /** Fills an oval to fit in the given rectangle */
    public native void devFillOval(int x, int y, int w, int h);

    public void fillOval(SunGraphics2D sg,
			 int x, int y, int width, int height) {
	devFillOval(x, y, width, height);
    }

    /**
     * Draws an arc bounded by the given rectangle from startAngle to
     * endAngle. 0 degrees is a vertical line straight up from the
     * center of the rectangle. Positive angles indicate clockwise
     * rotations, negative angle are counter-clockwise.
     */
    public native void devDrawArc(int x, int y, int w, int h,
				  int startAngle, int endAngle);

    public void drawArc(SunGraphics2D sg,
			int x, int y, int width, int height,
			int startAngle, int arcAngle) {
	devDrawArc(x, y, width, height, startAngle, arcAngle);
    }

    /** fills an arc. arguments are the same as drawArc. */
    public native void devFillArc(int x, int y, int w, int h,
				  int startAngle, int endAngle);

    public void fillArc(SunGraphics2D sg,
			int x, int y, int width, int height,
			int startAngle, int arcAngle) {
	devFillArc(x, y, width, height, startAngle, arcAngle);
    }

    public String toString() {	
	return getClass().getName();
    }

    // REMIND: DOCS !!

    public String getDevID() {
        return myDevID;
    }

    private native int X11getMonoImageData(ImageData id);
    private native int X11putMonoImageData(ImageData id);

    public ImageData getMonoImageData(int x, int y, int w, int h)
    {
     ImageData mid = cachedMonoID;
     if ((null == mid) || mid.amILocked()) {
         mid = new ImageData();
         mid.setColorModel( null);
      }
      int xView = (x/8)*8;
      int dxView = w + (x-xView);
      mid.setViewArea(xView, y, dxView, h);

      mid.setLockMethod(3);
      mid.incLockCount();
      int status = X11getMonoImageData(mid);
      if (status == SunGraphics2D.LOCK_OK) {
          cachedMonoID = mid;
          return mid;
      }
      return null;
    }

    public int putMonoImageData(ImageData idData)
    {
      int nRC = LOCK_DATA_STRUCT_PROBLEM;
      if( null != idData )
      {
        nRC = LOCK_NOT_LOCKED;
        if( true == idData.amILocked() )
        {
          if( UNLOCK_OK == (nRC = X11putMonoImageData (idData)) )
            {
              idData.decLockCount();
            }
          }
      }
      return nRC;
    }


    private native int X11LockViewResources(ImageData idData);
    private native int X11UnLockViewResources(ImageData idData);

    /**
     * The method prepares locks for a SunGraphics2D to make it accessible
     * for direct operation on graphics content by GraphicsPrimitives
     */
    public ImageData lock(
             int xView, int yView, int dxView, int dyView, int lockFlags)
    {
        ImageData imgData = cachedID;
        if ((null == imgData) || imgData.amILocked()) {
            imgData = new ImageData();
            imgData.setColorModel(getDeviceColorModel());
        }
        imgData.setViewArea(xView, yView, dxView, dyView);
        imgData.setLockMethod(lockFlags);
        imgData.incLockCount();
        int status = X11LockViewResources(imgData);
        if (SunGraphics2D.LOCK_OK == status) {
            cachedID = imgData;
            return imgData;
        }
        return null;
    }

    public int unlock(ImageData idData)
    {
	int nRC = LOCK_DATA_STRUCT_PROBLEM;
	if( null != idData )
	{
	    nRC = LOCK_NOT_LOCKED;
	    if( true == idData.amILocked() )
	    {
		if( UNLOCK_OK == (nRC = X11UnLockViewResources(idData)) )
		{
		    idData.decLockCount();
		}
	    }
	}
	return nRC;
    }

 
    /* implementation of textPipe */

    class X11TextCachedState {
        double [] matrix;
        double [] devTX;
        Font    theFont;
        Color   theColor;
        FontRenderContext frc;
        SunGraphics2D the2D;
 
        X11TextCachedState( SunGraphics2D   g2d )
        {
            the2D = g2d;
            theFont = the2D.getFont();
            float ptSize = theFont.getSize2D();
            frc = the2D.getFontRenderContext();
            AffineTransform tx = (AffineTransform)frc.getTransform();
            tx.scale(ptSize, ptSize);
            tx.preConcatenate(theFont.getTransform());
            tx.preConcatenate(the2D.transform);
            matrix = new double[6];
            tx.getMatrix(matrix);
            devTX = new double[6];
            the2D.transform.getMatrix(devTX);
            theColor = the2D.getColor();
        }
   
        double [] getOrigin(float x, float y) {
            double origin[] = {x, y};
            the2D.transform.deltaTransform(origin, 0, origin, 0, 1);
            origin[0] += matrix[4];
            origin[1] += matrix[5];
            return origin;
        }
    }
    
    /**
       This method will call super.drawString for the cases it can not
       handle below. The super class (in this case ROM) must not
       redelegate to this method, if its fast lock fails. See also the
       comment int ROM drawString().
    */
    public Object drawString(SunGraphics2D g2d,
            String s, float x, float y, Object cachedState) {

       if ( usesPlatformFont ) {
           return drawStringPlatform(g2d, s, x, y, cachedState);
       }

       X11TextCachedState tcs = null;

       if (cachedState != null) {
           if (cachedState instanceof X11TextCachedState) {
               tcs = (X11TextCachedState) cachedState;
           }
       }

       if (tcs == null) {
           tcs = new X11TextCachedState(g2d);
       }

       X11Graphics g = (X11Graphics) g2d;
        
       if (!g2d.simpleText) {
           if (cachedState instanceof X11TextCachedState) {
               cachedState = null; // don't pass manager X11TextCachedState
           }
           System.out.println("not simple text");
           return super.drawString(g2d, s, x, y, cachedState);
       }

       if (tcs.frc.isAntiAliased() || tcs.frc.usesFractionalMetrics()) {
           if (cachedState instanceof X11TextCachedState) {
               cachedState = null; // don't pass manager X11TextCachedState
           }
           System.out.println("isAntialias||usesFractionalMetrics");
           return super.drawString(g2d, s, x, y, cachedState);
       }

       DrawString ds = g.drawStringClass;

       Rectangle bounds = getBoundingRect(g2d, s, tcs.theFont, tcs.frc, x, y);

       ImageData imageData = g.getMonoImageData(bounds.x, bounds.y,
                             bounds.width, bounds.height);
       try {
            double origin[] = tcs.getOrigin(x, y);
            x = (float)origin[0];
            y = (float)origin[1];
            ds.DrawString(imageData, s, x, y, tcs.theColor, 
                          tcs.theFont, tcs.matrix,
                          false, tcs.frc.usesFractionalMetrics());
       } catch (Exception e) {
           // System.out.println("??? we got exception: "+e);
       }

       g.putMonoImageData(imageData);

       return tcs;
    }

    public Object drawGlyphVector(SunGraphics2D g2d,
            GlyphVector gv, float x, float y, Object cachedState) {
    
        if ( usesPlatformFont ) {
            return drawGlyphVectorPlatform(g2d, gv, x, y, cachedState);
        }
        return super.drawGlyphVector(g2d, gv, x, y, cachedState);
    }

    public Object drawChars(SunGraphics2D g2d,
            char data[], int offset, int length, int x, int y,
            Object cachedState) {
        if ( usesPlatformFont ) {
            return drawCharsPlatform(g2d, data, offset, length, 
                                     x, y, cachedState);
        }
        return drawString(g2d, new String(data, offset, length),
                          x, y, cachedState);
    }

    public Object drawBytes(SunGraphics2D g2d,
            byte data[], int offset, int length, int x, int y,
            Object cachedState) {
        if ( usesPlatformFont ) {
            return drawBytesPlatform(g2d, data, offset, length, 
                                     x, y, cachedState);
        }
        return drawString(g2d, new String(data, 0, offset, length),
                          x, y, cachedState);
    }


    // Returns a quick, rough estimation of the bounding rectangle of a 
    // given string in the given GC with the given origin (x, y).  Used
    // in drawString to determine how much screen to lock.
    // In certain cases this bounding rectangle is inaccurate due to a bug
    // in the t2k font rasterizer (see bug 4192418).  The bounding rect will
    // be inaccurate when the GC transform includes a rotation and either
    // Text Antialiasing or Fractional Metrics are off (both are off by
    // default)
    private Rectangle getBoundingRect(SunGraphics2D g2d, String s,
                                      Font font, FontRenderContext frc,
                                      double x, double y) {

      // estimate character width using getMaxCharBounds
      Rectangle2D charRect = font.getMaxCharBounds(frc);
      LineMetrics flm = font.getLineMetrics(s, frc);
      // estimate line width
      double width = charRect.getWidth() * s.length();
      // estimate line height, remember to add descent
      double height = charRect.getHeight() + flm.getDescent();
      // compute bounding rectangle of the string
      double tl_x = x;                    // top left x value
      double tl_y = y - flm.getAscent();  // top left y value
      
      // Transform the bounding rectangle by the matrix in g2d
      if (g2d.transformState != g2d.TRANSFORM_ISIDENT) {
      double [] srcPts = { tl_x, tl_y,
                           tl_x + width, tl_y,
                           tl_x, tl_y + height,
                           tl_x + width, tl_y + height };
      double [] dstPts = new double[8];
      g2d.transform.transform(srcPts, 0, dstPts, 0, 4);
      
      // Get the min and max x and y values of the transformed rect
      // and create a new rectangle
      int minx = Integer.MAX_VALUE; 
      int miny = Integer.MAX_VALUE;
      int maxx = Integer.MIN_VALUE;
      int maxy = Integer.MIN_VALUE;
      for (int i = 0; i < 8; i += 2) {
        if (dstPts[i] < minx) minx = (int) dstPts[i];
        if (dstPts[i] > maxx) maxx = (int) dstPts[i];
        if (dstPts[i+1] < miny) miny = (int) dstPts[i+1];
        if (dstPts[i+1] > maxy) maxy = (int) dstPts[i+1];
        }
      
      return new Rectangle(minx, miny, maxx - minx, maxy - miny);
      } else {
      return new Rectangle ((int) tl_x,(int) tl_y,(int) width,(int) height);
      }
    }      
    /* end of implementation of text pipe and supporting routines */
}
