/*
 * @(#)X11Selection.java	1.24 01/02/09
 *
 * Copyright 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.motif;

import java.awt.Toolkit;
import java.awt.Component;

import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.FlavorMap;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.datatransfer.SystemFlavorMap;

import java.io.InputStream;
import java.io.IOException;

import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import sun.awt.Mutex;
import sun.awt.SunToolkit;
import sun.awt.AppContext;

/*
 * Implements a general interface to the X11 selection mechanism.
 * This class contains the intelligence to do the format translation
 * between java DataFlavor namespace and the X11 TARGET namespace.
 */
public class X11Selection {

    static FlavorMap flavorMap = SystemFlavorMap.getDefaultFlavorMap();

    static Vector selections;

    int atom;

    X11SelectionHolder holder;
    Transferable contents;

    int targetArray[];
    Object nativeData;
    int format;

    private boolean disposed = false;
    private byte[] data = null;
    private boolean dataAvailable = false;
    private int currentFormat;
    private static Object dummyObj = new Object();
    private static Object lock = new Object();

    static {
	// 4154170:  Need to ensure the the toolkit is initialized prior
	// to executing this initializer
	MToolkit toolkit = (MToolkit)Toolkit.getDefaultToolkit();

	selections = new Vector();

        initIDs();
	init();

    }

    private static native void initIDs();
    static native void init();

    public X11Selection(String name) {
	atom = MDataTransferer.getAtomForTarget(name);
        selections.addElement(this);
    }

    protected void finalize() throws Throwable {
        clearContentsData();
        dispose();
	super.finalize();
    }

    private static Object[] getSelectionsArray() {
	return selections.toArray();
    }
  
    private int[] getFormatsForFlavors() {
        final Stack stack = new Stack();

        Runnable getFormats = new Runnable () {
                public void run() {
                    int[] formats = null;
                    try {
                        formats = MDataTransferer.getInstance().
                            getFormatsForFlavorsAsArray
                            (contents.getTransferDataFlavors(), flavorMap);
                    } finally {
                        AWTLockAccess.awtLock();
                        stack.push(formats);
                        AWTLockAccess.awtNotifyAll();
                        AWTLockAccess.awtUnlock();
                    }
                }
            };
            
        SunToolkit.executeOnEventHandlerThread(dummyObj, getFormats);
        
        while (stack.empty()) {
            AWTLockAccess.awtWait();
        }
        
        return (int[])stack.pop();
    }
  
    private byte[] convertData(int format) {
	synchronized (lock) {
	    dataAvailable = false;
	    data = null;
	    currentFormat = format;
	    Runnable getData = new Runnable () {
		public void run() {
		    synchronized (lock) {
			try {
			    Map formatMap = MDataTransferer.getInstance().
				getFormatsForFlavors
				    (contents.getTransferDataFlavors(),
				     flavorMap);
			    DataFlavor flavor = (DataFlavor)
				formatMap.get(new Integer(currentFormat));

                            // Flavor can be null if some application   
                            // requested data in unsupported format.
                            if (flavor != null) {
                                data = MDataTransferer.getInstance().
                                    translateTransferable(contents, flavor, 
                                                          currentFormat);
                            }
			} catch (Exception e) {
			    e.printStackTrace();
			    data = null;
			}
			dataAvailable = true;
			lock.notify();
		    }
		}
	    };
 
	    SunToolkit.executeOnEventHandlerThread(dummyObj, getData);
 
	    while (!dataAvailable) {
		try {
		    lock.wait();
		} catch (InterruptedException ie) {
		    // do nothing
		}
	    }
	}
	return data;
    }

    /*
     * methods for acting as selection provider
     */
    native boolean pGetSelectionOwnership();

    boolean getSelectionOwnership(Transferable contents,
				  X11SelectionHolder holder) {
        this.holder = holder;
        this.contents = contents;
	if (pGetSelectionOwnership()) {
            SunToolkit.insertTargetMapping(dummyObj,
                                           AppContext.getAppContext());
	    return true;
	}
        clearContentsData();
	return false;
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    void lostSelectionOwnership() {
	if (holder != null) {
	    holder.lostSelectionOwnership();
	}
        clearContentsData();
    }

    void clearContentsData() {
	holder = null;
	contents = null;
    }

   /*
    * methods for acting as selection consumer
    */
    native void pGetTransferTargets();
    native void pGetTransferData(int atoms[]);

    int[] getSelectionTargets() {
        targetArray = null;

	SelectionThread st = (SelectionThread)
	    java.security.AccessController.doPrivileged(
					new java.security.PrivilegedAction() {
		public Object run() {
		    return new SelectionThread(X11Selection.this, null);
		}
	    });
	synchronized(st) {
	    try {
	        st.start();
		st.wait();
	    } catch (InterruptedException e) {

	    }
	}

        return targetArray;
    }

    /* Retrieves the X11 selection value for this target
     * (currently only supports string/text flavors)
     */
    public byte[] getSelectionData(int target) throws IOException {
        nativeData = null;
        final int targets[] = new int[] { target };
	SelectionThread st = (SelectionThread)
	    java.security.AccessController.doPrivileged(
					new java.security.PrivilegedAction() {
		public Object run() {
		    return new SelectionThread(X11Selection.this, targets);
		}
	    });
	synchronized(st) {
	    try {
	        st.start();
		st.wait();
	    } catch (InterruptedException e) {

	    }
	}

	if (!(nativeData instanceof byte[])) {
	    throw new IOException("could not get transfer data");
	}

        return (byte[])nativeData;
    }

    /*
     * Subclasses should override disposeImpl() instead of dispose(). Client
     * code should always invoke dispose(), never disposeImpl().
     */
    protected void disposeImpl() {
        selections.removeElement(this);
    }

    public final void dispose() {
        boolean call_disposeImpl = false;

        if (!disposed) {
            synchronized (this) {
                if (!disposed) {
                    disposed = call_disposeImpl = true;
                }
            }
        }

        if (call_disposeImpl) {
            disposeImpl();
        }
    }
}

/*
 * Implements the thread required to deal with X11's asynchronous
 * selection API.  This thread will be responsible for requesting
 * the selection data and not returning until that data has been
 * provided to a native selection callback.
 */
class SelectionThread extends Thread {
    X11Selection s;
    int[] targets;

    public SelectionThread(X11Selection s, int[] targets) {
	super("AWT-Selection");
	this.s = s;
	this.targets = targets;
    }
    
    /*
     * Fix for bug id 4182518. Avoid possible deadlock when two separate 
     * threads call pGetTansferTargets and pGetTransferData in 
     * SelectionThread.run(). Use a global lock (selectionLock) to 
     * synchronize against this possibility.
     */
    private static Object selectionLock = new Object();

    public void run() {
	synchronized(selectionLock) {
	    synchronized(this) {
		if (targets == null) {
		    s.pGetTransferTargets();
		} else {
		    s.pGetTransferData(targets);
		}
		notifyAll();
	    }
	}
    }
}
