/*
 * @(#)context.c	1.69 00/03/16
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

/*
 * Initialize the platform dependent parts of a thread context.
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <unistd.h>

#include "hpi_impl.h"

#include "threads_md.h"
#include "interrupt.h"
#include "itimer_md.h"
#include "sparc_md.h"
#include "context.h"
#include "schedule.h"

/*
 * The list on which we put freed thread stacks pending reclamation.
 */
typedef struct free_stack_t {
    struct free_stack_t *next;
    size_t   stack_size;
} free_stack_t;

static free_stack_t *stackFreeList = (free_stack_t *) 0;
static int stackFreeCount = 0;

static int redZoneSize = 0;	/* One page */

/*
 * Free tid and stack space for dead threads.  We do this whenever
 * we do sysThreadCreate() and there are more than FREESTACKTHRESHOLD
 * free stacks.  We only free stacks in excess of FREESTACKTHRESHOLD.
 * Note that we have to play some games here because we need to make
 * stackFreeList and stackFreeCount access Green thread-safe, but can't
 * call free() with SCHED_LOCKED().
 *
 * The value for FREESTACKTHRESHOLD probably isn't too critical, at
 * least for current uses of Java in HotJava.  The choice of 5 means
 * that the majority of stack allocations when browsing HTML pages
 * including applets will be satisfied from the free lists, and
 * only when pages are left are there sometimes enough stacks freed
 * to flush them back to the O/S.  This may need to be more sophisti-
 * cated in the future.
 */

#define FREESTACKTHRESHOLD 5

static void freeStacks(void) 
{
    free_stack_t* stack = 0;
    free_stack_t* next;
    long size;

    if (stackFreeCount > FREESTACKTHRESHOLD) {
	SCHED_LOCK();
	/* Set up to stacks in excess of FREESTACKTHRESHOLD */
	while (stackFreeCount > FREESTACKTHRESHOLD) {
	    free_stack_t *freeStack;
	    freeStack = stackFreeList;
	    stackFreeList = freeStack->next;
	    freeStack->next = stack;
	    stack = freeStack;
	    stackFreeCount--;
	}
	SCHED_UNLOCK();
	while (stack) {
	    size_t junk;
	    next = stack->next;
	    Log3(2,
    "Over free stack threshold: freeing (0x%p:0x%p), new free count %d\n",
		 (char *) stack+stack->stack_size,
		 stack->stack_size, stackFreeCount);
	    /* Don't forget the redzone... */
	    size = stack->stack_size + redZoneSize;
	    /* Free the ExecEnv */
	    {
	        sys_thread_t **ptid = 
		    (sys_thread_t **)((char *)stack + stack->stack_size); 
		freeThreadBlock(*(ptid - 1));
	    }
	    /* REMIND: Ignoring return value here... */
	    sysUnmapMem((char *)stack - redZoneSize, size, &junk);
	    stack = next;
	}
    }
}

/*
 * Allocate a single thread stack of specified size.
 *
 * Note that "stack base" here and elsewhere means the base of the
 * logical stack, growing from high addresses to low addresses, and
 * *not* the way you'd think of, say, a malloc'ed chunk.
 *
 * We let the low level stack allocator take care of any rounding or
 * other system-specific sizing of the stack request.  But if no stack
 * size is given we take the default one. 
 */

static int 
allocateStack(gstack_t *stack, size_t size)
{
    free_stack_t *freeStack;
    free_stack_t **freeStackPrev;

    freeStacks();		/* Free any stacks for dead threads */

    if (size == 0) {
        /* Allocate conservatively if we don't know better */
        size = 128 * 1024;
    }

    /* Take stack off the stackFreeList if one's available */
    SCHED_LOCK();
    freeStackPrev = &stackFreeList;
    while ((freeStack = *freeStackPrev) && freeStack->stack_size != size) {
	freeStackPrev = &freeStack->next;
    }

    if (freeStack) {		/* Got one */
	*freeStackPrev = freeStack->next;
	stackFreeCount--;
	SCHED_UNLOCK();
	stack->base = (stackp_t) ((char *) freeStack + freeStack->stack_size);
	stack->size = freeStack->stack_size;
	Log3(2, "New stack from free list (0x%x:0x%x), new free count %d\n",
	     stack->base, stack->size, stackFreeCount);
	/* Free the ExecEnv */
	{
	    freeThreadBlock(*((sys_thread_t **)stack->base - 1));
	}
	return 1;
    } else {			/* Didn't get one */
	SCHED_UNLOCK();		/* Can't allocate with SCHED_LOCKED() */
	if (redZoneSize == 0) {
	    redZoneSize = sysconf(_SC_PAGESIZE);	/* one page */
	}
	size += redZoneSize;	/* For redzone */
	stack->base = (stackp_t) sysMapMem(size, &size);
	/*
	 * On Solaris we want to demand-page stack memory, so we
	 * don't sysCommitMem() here...
	 */
	if (!stack->base) {
	    return 0;		/* Allocation failed */
	} else {
	    if (mprotect((char *)stack->base, redZoneSize, PROT_NONE)) {
		return 0;
	    }
	    stack->base = (stackp_t) ((char *) stack->base + size);
	    /* gstack_t.size does NOT include the redzone */
	    stack->size = size - redZoneSize;
	    Log3(2, "New stack (0x%x:0x%x), free count %d\n",
		 stack->base, stack->size, stackFreeCount);
	    return 1;
	}
    }
}

/*
 * Returns the stack pointer for any threads but the current one.
 */
void *
currentStackPointer(sys_thread_t *tp)
{
    return (void *)CONTEXT(tp)->current_sp;
}

#if defined(__solaris__) && defined(i386)

/* Solaris/x86 version */

/*
 * This is the stack layout, as defined by the Intel ABI
 *
 * In the Intel world, a stack frame looks like this:
 *
 * %fp0->|                               |
 *       |-------------------------------|
 *       |  Args to next subroutine      |
 *       |-------------------------------|-\
 * %sp0->|  One word struct-ret address  | |
 *       |-------------------------------|  > minimum stack frame (8 bytes)
 *       |  Previous frame pointer (%fp0)| |
 * %fp1->|-------------------------------|-/
 *       |  Local variables              |
 * %sp1->|-------------------------------|
 */

void
initContext(ucontext_t *uc, gstack_t *stack, unsigned int pc,
         void (*death_func)(void), unsigned int arg)
{
    unsigned char *sp, *limit;
    int *p;

    getcontext(uc);
    limit = (unsigned char *)stack->base - stack->size;
    sp = (unsigned char *) stack->base;
    sp -= sizeof(sys_thread_t *);
    /* Force sp to be double aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & 0xfffffff8);
    /* for x86 the args to death_func are passed on the stack */
    p = (int *)sp;
    *--p = (int)arg;	/* arg 2 */
    *--p = (int)pc;	/* arg 1 */
    *--p = (int)0;	/* return PC */

    uc->uc_stack.ss_sp = (char *) p;
    uc->uc_stack.ss_size = (unsigned char *)p - limit;

    uc->uc_mcontext.gregs[USP] = (unsigned int) p;
    uc->uc_mcontext.gregs[PC] = (unsigned int) death_func;
}

#elif defined(__solaris__) && defined(sparc)

/* Solaris/SPARC version */
 
/*
 * This is the stack layout, as defined by the SPARC ABI:
 *
 *      HIGH ADDRESSES
 *	|-----------------------|
 *	| ThreadPrivate data	|
 *      |-----------------------|
 *      | 16 words for in and   |  death_func() stack frame
 *      | local save area       |
 *      |-----------------------|  --------
 *      | arguments past the 6th|
 *      |-----------------------|  startup func stack frame
 *	| 6 words for locals    |
 *	|-----------------------|
 *      | 1 word for aggregate  |
 *      | return value          |
 *      |-----------------------|
 *      | %i7 = startaddr       |
 *      | %i6 = sp              |
 *      | %i0-%i1 --> args 1-5  |
 *      |-----------------------|
 *      | 8 words for saved %l's|
 *      |-----------------------|       <--- sp(double aligned)
 *      LOW ADDRESSES
 */
void
initContext(ucontext_t *uc, gstack_t *stack, unsigned int pc,
	     void (*death_func)(void), unsigned int arg)
{
    unsigned char *sp, *limit;

    getcontext(uc);
    limit = (unsigned char *)stack->base - stack->size;
    sp = (unsigned char *) stack->base;
    sp -= (sizeof(sys_thread_t *) + WINDOWSIZE + SA(MINFRAME));

    /* Force sp to be double aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & 0xfffffff8); 

    uc->uc_stack.ss_sp = (char *) sp;
    uc->uc_stack.ss_size = sp - limit;
    uc->uc_stack.ss_flags = 0; /* ? */

    uc->uc_mcontext.gregs[REG_SP] = (unsigned int) sp;
    uc->uc_mcontext.gregs[REG_PC] = (unsigned int) death_func;
    uc->uc_mcontext.gregs[REG_nPC] = (unsigned int) death_func+4;
    uc->uc_mcontext.gregs[REG_O0] = (int) pc;
    uc->uc_mcontext.gregs[REG_O1] = arg; 
}

#elif defined(__solaris__) && defined(ppc)

/* Solaris/ppc version */

#define	MINFRAME	8
#define STACK_ALIGN	16
#define SA(X)		(((X)+(STACK_ALIGN-1)) & ~(STACK_ALIGN-1))
#define WINDOWSIZE	SA(MINFRAME+(20*4))

void
initContext(ucontext_t *uc, gstack_t *stack, unsigned int pc,
         void (*death_func)(void), unsigned int arg)
{
    unsigned char *sp, *limit;

    getcontext(uc);
    limit = (unsigned char *)stack->base - stack->size;
    sp = (unsigned char *) stack.base;
    sp -= (sizeof(sys_thread_t *) + WINDOWSIZE + SA(MINFRAME));
    /* Force sp to be 16 byte aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & 0xfffffff0);

    uc->uc_stack.ss_sp = (char *) sp;
    uc->uc_stack.ss_size = sp - limit;

    uc->uc_mcontext.gregs[R_R1] = (unsigned int) sp;

    uc->uc_mcontext.gregs[R_R3] = (unsigned int) pc;
    uc->uc_mcontext.gregs[R_R4] = (unsigned int) arg;
    uc->uc_mcontext.gregs[R_PC] = (unsigned int) death_func;
    uc->uc_mcontext.gregs[R_LR] = (unsigned int) death_func;
}

#elif   defined(__linux__) && defined(__sparc__)
void    thread_start_sparc();
void
initContext(ucontext_t *uc, gstack_t *stack, unsigned int pc,
         void (*death_func)(void), unsigned int arg)
{
    unsigned char *sp;
unsigned long *base;

    /* clear the context structure */
    memset(uc, 0, sizeof(ucontext_t));

    sp = (unsigned char *) stack->base;
    /* Note: WINDOWSIZE stuff is for SPARC, not really needed for x86 */
    sp -= (sizeof(sys_thread_t) + WINDOWSIZE + SA(MINFRAME));
    /* Force sp to be double aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & ~7L);
    /* for alpha the args to death_func are passed in the registers */
    /* however, we don't have direct access to these registers,
       so we have to make a new start function and pass it
       in these odd places */
        base = sp-64;
/* prepare stack for thread-start */
        base[0]=(unsigned long) death_func;
        base[1]=arg;
        base[2]=pc;
        base[15]=sp;
    uc->jmpbuf->__jmpbuf[JB_FP] = base;
    uc->jmpbuf->__jmpbuf[JB_PC] = (unsigned long) thread_start_sparc;

}

#elif defined(__linux__) && defined(__i386__)

void
initContext(gcontext_t *uc, gstack_t *stack, unsigned int pc,
         void (*death_func)(void), unsigned int arg)
{
    unsigned char *sp, *limit;
    int *p;
    char *fdata;

    getcontext(uc);

    fdata = (char*) uc->floatbuf;
    __asm__ ("frstor %0"::"m" (*fdata));

    limit = (unsigned char *)stack->base - stack->size;
    sp = (unsigned char *) stack->base;
    sp -= sizeof(sys_thread_t *);
    /* Force sp to be double aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & 0xfffffff8);

    /* for x86 the args to death_func are passed on the stack */
    p = (int *)sp;
    *p-- = (int ) arg;
    *p-- = (int ) pc;

    uc->uc_stack.ss_sp = (char *) p;
    uc->uc_stack.ss_size = (char *) ((long) p - (long) limit);

#ifdef __GLIBC__
    uc->jmpbuf->__jmpbuf[JB_BP] = (int) p;
    uc->jmpbuf->__jmpbuf[JB_SP] = (int) p;
    uc->jmpbuf->__jmpbuf[JB_PC] = (int) death_func;
#else
    uc->jmpbuf->__jmpbuf->__bp = (char *) p;
    uc->jmpbuf->__jmpbuf->__sp = (char *) p;
    uc->jmpbuf->__jmpbuf->__pc = death_func;
#endif
}

#elif defined(__linux__) && defined(__alpha__)

void thread_start_alpha();

void
initContext(gcontext_t *uc, gstack_t *stack, unsigned int pc,
         void (*death_func)(void), unsigned int arg)
{
    context_t *mdcontext = CONTEXT(tid);
    genunix_ucontext_t *uc = &(mdcontext->ucontext);
    unsigned char *sp;

    /* clear the context structure */
    memset(uc, 0, sizeof(genunix_ucontext_t));

    sp = (unsigned char *) tid->stack_base;
    /* Note: WINDOWSIZE stuff is for SPARC, not really needed for x86 */
    sp -= (sizeof(sys_thread_t) + WINDOWSIZE + SA(MINFRAME));
    /* Force sp to be double aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & ~7L);
    /* for alpha the args to death_func are passed in the registers */
       so we have to make a new start function and pass it
       in these odd places */
    uc->jmpbuf->__jmpbuf->__9  = (long int) death_func;
    uc->jmpbuf->__jmpbuf->__10 = (long int) pc;
    uc->jmpbuf->__jmpbuf->__11 = (long int) arg;
    uc->jmpbuf->__jmpbuf->__fp = (long int *) sp;
    uc->jmpbuf->__jmpbuf->__sp = (long int *) sp;
    uc->jmpbuf->__jmpbuf->__pc = (long int *) thread_start_alpha;
}

#elif defined(__linux__) && defined(__powerpc__)
#define MINFRAME    8
#define STACK_ALIGN 16
#define SA(X)       (((X)+(STACK_ALIGN-1)) & ~(STACK_ALIGN-1))
#define WINDOWSIZE  SA(MINFRAME+(20*4))

void
initContext(gcontext_t *uc, gstack_t *stack, unsigned int pc,
         void (*death_func)(void), unsigned int arg)
{
    unsigned char *sp, *limit;
    int *p;

    getcontext(uc);
    limit = (unsigned char *)stack->base - stack->size;
    sp = (unsigned char *) stack->base;
    sp -= (sizeof(sys_thread_t *) + WINDOWSIZE + SA(MINFRAME));
    /* Force sp to be 16 byte aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & 0xfffffff0);

    uc->uc_stack.ss_sp = (char *) sp;
    uc->uc_stack.ss_size = sp - limit;

    /* for ppc the args to death_func are passed in the ucontext structure */
    p = (int *)sp;
    uc->pc = pc;
    uc->arg = arg;

    uc->jmpbuf->__jmpbuf->__misc[0] = (long) p;
    uc->jmpbuf->__jmpbuf->__misc[2] = (long ) death_func;
}

#elif defined(__linux__) && defined(__mc68000__)

void
initContext(gcontext_t *uc, gstack_t *stack, unsigned int pc,
         void (*death_func)(void), unsigned int arg)
{
    unsigned char *sp, *limit;
    int *p;

    getcontext(uc);
    limit = (unsigned char *)stack->base - stack->size;
    sp = (unsigned char *) stack->base;
    sp -= sizeof(sys_thread_t *);
    /* Force sp to be double aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & 0xfffffff8);
    /* for m68k the args to death_func are passed on the stack */
    p = (int *)sp;
    *--p = (int)arg;    /* arg 2 */
    *--p = (int)pc;     /* arg 1 */
    *--p = (int)0;      /* return PC */

    uc->uc_stack.ss_sp = (char *) p;
    uc->uc_stack.ss_size = (unsigned char *)p - limit;

    uc->jmpbuf->__jmpbuf[0].__fp = p;
    uc->jmpbuf->__jmpbuf[0].__sp = p;
    uc->jmpbuf->__jmpbuf[0].__aregs[0] = (int *) death_func;
}

#else
#error you need an initContext here!

#endif

/*
 * This routine allocates enought memory for the thread context
 * and the stack.  This implementation puts the thread context
 * at the top of the stack.
 *
 * Must zero memory and initialize tid->PrivateInfo and
 * tid->PrivateInfo->stack.
 */
int
allocateContextAndStack(sys_thread_t *tid, long requested_size)
{
    gstack_t stack;

    if (!allocateStack(&stack, (size_t)requested_size)) {
	return 0;
    }
    *((sys_thread_t **)stack.base - 1) = tid;
    tid->stack.base = stack.base;
    tid->stack.size = stack.size;  /* Actual size allocated */
    return 1;
}

/*
 * Delete the specified context, and the stack associated with it.
 * This does *not* free the memory associated with the stack: that
 * is handled elsewhere!
 *
 * WARNING: Called when SCHED_LOCKED(), so may not enter any monitors!
 */
void
deleteContextAndStack(sys_thread_t *tid)
{
    /*
     * If tid->primordial_thread (Unix main()), we don't want to free
     * the C stack, but we have malloc'ed the sys_thread_t.  This is
     * NOT the normal case!  But we don't want this going back to the
     * stack allocator, so we drop the sys_thread_t on the floor here.
     */
    if (!tid->primordial_thread) {
        free_stack_t *tmp;
	/*
	 * The normal case: queue the stack and the thread private data
	 * to be freed when safe.
	 */
	sysAssert(0 <= tid->stack.size);
	/* Note that free stacks are pointed to *above* the redzone */
	tmp = (free_stack_t *)((char *) tid->stack.base - tid->stack.size);
	tmp->stack_size = tid->stack.size;
	tmp->next = stackFreeList;
	stackFreeList = tmp;
	stackFreeCount++;
	Log3(2, "Freed stack (0x%p:0x%p), new free count %d\n",
	     tid->stack.base, tid->stack.size, stackFreeCount);
    }
    return;
}

