/*
 * @(#)interrupt_md.c	1.71 00/04/18
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

/*
 * Solaris 2.X dependent interrupt handling code.
 */


/*
 * Header files.
 */
#include <stdio.h>
#include <signal.h>
#ifdef __solaris__
#include <sys/trap.h>
#endif
#include <stdlib.h>
#include <string.h>

#include "hpi_impl.h"

#include "interrupt.h"
#include "itimer_md.h"
#include "internal_md.h"
#include "threads_md.h"
#include "schedule.h"
#include "signals.h"
#if defined(__linux__) && defined(__powerpc__)
#include <asm/sigcontext.h>
#endif

static int pending_signals[N_INTERRUPTS];

#ifdef __linux__
static int recreate_io = 1;
#endif
/*
 * intrLock() -- Disable all interrupts.
 */
void
intrLock()
{
    sys_thread_t *cth = greenThreadSelf();
    machdep_t *machdepP;
    sigset_t newMask;
    
    if (cth) {
        machdepP = MACHDEP(cth);

        /* Sanity check */
	sysAssert(machdepP->intrLockCount >= 0);
    
	if (machdepP->intrLockCount++ == 0) {
	    sigfillset(&newMask);
	    green_sigprocmask(SIG_BLOCK, &newMask, &machdepP->intrLockOldMask);
	}
    } else {
	sigfillset(&newMask);
	green_sigprocmask(SIG_BLOCK, &newMask, NULL);
    }
}

/*
 * intrUnlock() -- Enable all interrupts.
 */
void
intrUnlock()
{
    sys_thread_t *cth = greenThreadSelf();
    machdep_t *machdepP;
    sigset_t newMask;
    
    if (cth) {
        machdepP = MACHDEP(cth);

        /* Sanity check */
	sysAssert(machdepP->intrLockCount > 0);
    
	if (--machdepP->intrLockCount == 0) {
	    green_sigprocmask(SIG_SETMASK, &machdepP->intrLockOldMask, NULL);
	}
    } else {
	sigemptyset(&newMask);
	green_sigprocmask(SIG_SETMASK, &newMask, NULL);
    }
}

/*
 * intrInitMD() -- Target-specific initialization.
 */
extern void
intrInitMD()
{
    memset(pending_signals, 0, sizeof(pending_signals));
#if   defined(__solaris__)
    sigignore(SIGPIPE);
#elif defined(__linux__)
    signal(SIGPIPE, SIG_IGN);
#else
#error What does your OS want today?
#endif
}

/*-
 * intrDispatchMD() -- Turn our signal into an intrDispatch().
 *
 * Porting note: In this function we treat SA_SIGINFO and HAVE_UCONTEXT
 * synonymously.  Maybe we should clean this up.
 */
void
#if defined(SA_SIGINFO)
intrDispatchMD(int sig, siginfo_t *info, void *uc)
#else
intrDispatchMD(int sig)
#endif
{
    int	old_errno = errno;
    sys_thread_t *self = greenThreadSelf();
#if defined(SA_SIGINFO)
#ifdef __linux__
    gcontext_t *ucp = (gcontext_t *)uc;
#else
    ucontext_t *ucp = (ucontext_t *)uc;
#endif
#endif

    Log1(1, "signalHandlerDispatch(sig=%d)\n", sig);

    SCHED_LOCK();

#ifdef SA_SIGINFO
    intrDispatch(sig, info, uc);
#else
    intrDispatch(sig, 0, 0);
#endif

#ifdef __linux__
     if (recreate_io && (sig == SIGALRM)) {
        intrDispatch(SIGIO, 0, 0);
     }
#endif

    errno = old_errno;

    /*
     * We are not preempting the current thread.  Just return
     * and let the signal trampoline restore the state of
     * the interrupted thread.
     */
    sysAssert(!(SCHED_LOCKED() > 1 && _needReschedule));
    if (_needReschedule) {

	_needReschedule = 0;
	sysAssert(SCHED_LOCKED() == 1);

#if defined(__solaris__)
	/* DEBUG check that saved mask is correct */
	sysAssert(self->sigmaskdirty ||
	    memcmp(&ucp->uc_sigmask, &MACHDEP(self)->signalMask,
	    sizeof ucp->uc_sigmask) == 0);
#endif

	/* Force reload of current signal mask */
	self->sigmaskdirty = 1;

	YIELD();
    }
    SCHED_UNLOCK();

    if (SCHED_LOCKED() == 0) {
	/*
	 * Restore cached sigmask to the saved value, so that it will
	 * be correct when we return from the signal handler.
	 *
	 * It is important that all signals be masked until we return,
	 * which is currently true.  If this changes, block all signals
	 * before setting signalMask.
	 */
#ifndef __linux__
#ifdef SA_SIGINFO
	MACHDEP(self)->signalMask = ucp->uc_sigmask;
#endif
#endif
    }
}

bool_t intrInUse(int sig)
{
    /* Signals used in green threads implementation. */
    return sig == SIGPIPE ||
      sig == SIGIO ||
      sig == SIGCHLD ||
      sig == SIGALRM;
}

void sysSignalNotify(int sig)
{
    sys_mon_t *mon = asyncMon(SYS_ASYNC_MON_USRSIG);
    pending_signals[sig]++;
    if (asyncEventNotify(mon)) {
	_needReschedule = 1;
    }
}

static int lookupSignal(void)
{
    int i;
    for (i = 0; i < N_INTERRUPTS; i++) {
        if (pending_signals[i]) {
	    pending_signals[i]--;
	    return i;
	}
    }
    return -1;
}

int sysSignalWait()
{
    sys_mon_t *mon = asyncMon(SYS_ASYNC_MON_USRSIG);
    sys_thread_t *self = sysThreadSelf();
    int sig;
    sysMonitorEnter(self, mon);
    while ((sig = lookupSignal()) == -1) {
        sysMonitorWait(self, mon, SYS_TIMEOUT_INFINITY);
    }
    sysMonitorExit(self, mon);
    return sig;
}
