/*
 * @(#)HAE_API_SolarisOS_Capture.c	1.19 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

/*****************************************************************************/
/*
**	HAE_API_SolarisOS_Capture.c
**
**	This provides platform specfic functions for Windows 95/NT. This interface
**	for HAE is for Windows 95/NT and uses the waveIn API to capture audio
**	buffer slices through the multimedia system.
**
**	Overview:
**
**	History	-
**
**	JAVASOFT
**	10.12.98	$$kk: Created
**
**	05.12.99:	$$kk: added very simple AUDIODEV environment variable support.
**  06.28.99:   $$jb: merged in changes to provide dynamic capture buffer size,
**              changes API for HAE_AquireAudioCapture
*/
/*****************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/audioio.h>

// for I_FLUSH
#include <stropts.h>

// $$kk: 08.12.98 merge 
// $$kk: 04.28.98: we *need* to use string.h, NOT strings.h; 
// the latter is not available on earlier versions of solaris
#include <string.h>

#include "HAE_API.h"

#ifndef TRUE
	#define TRUE		1
#endif
#ifndef FALSE
	#define FALSE		0
#endif


#define HAE_SOLARIS_NUM_CAPTURE_BUFFERS		1			// number of capture buffers

// $$jb: 03.22.99 Changing buffer size from 16384 to 1024
// #define HAE_SOLARIS_FRAMES_PER_BUFFER		1024;		// how many frames do we read (capture) at a time?
// $$kk: 03.31.99: increasing a bit on solaris.
// need to revisit this!
#define HAE_SOLARIS_FRAMES_PER_BUFFER		4096		// how many frames do we read (capture) at a time?
														// $$kk: 10.13.98: need to decide whether / how to let user configure this;
														// what is a good default??

#define HAE_SOLARIS_SOUND_CAPTURE_PERIOD	10			// sleep time in milliseconds, between position checks
														// $$kk: 10.13.98: need to implement position checks!

#define HAE_SOLARIS_DEFAULT_BUFFERSIZE_IN_MS	150		// default size of buffer in ms
#define HAE_SOLARIS_MIN_BUFFERSIZE_IN_MS		50		// minimum size of buffer in ms (can't go beneath
														// hardware buffersize, however)


static void				*g_captureBufferBlock = NULL;	// data buffer for capture
														// $$kk: 10.13.98: do we need to double buffer?

static unsigned long	g_captureByteBufferSize = 0;	// data buffer size in bytes


static long				g_captureSound = NULL;			// the audio device
static int				g_captureShutdown;				// false if capture active, otherwise true

static short int		g_bitSize;						// bit size of current capture format
static short int		g_channels;						// number of channels in current capture format
static short int		g_sampleRate;					// sample rate of current capture format

static short int		g_soundDeviceIndex = 0;			// if non zero then use this device to open

static unsigned long	g_encoding;						// audio encoding

// number of frames to read; right now this is fixed as HAE_SOLARIS_FRAMES_PER_BUFFER 
// $$kk: 10.13.98: need to figure out how to configure this
static long				g_audioFramesToRead = HAE_SOLARIS_FRAMES_PER_BUFFER;

// how long to sleep at a time
static unsigned int		g_audioCaptureSleepTime;		// setup upon runtime

static HAE_CaptureDone	g_captureDoneProc;				// capture callback

// magically reduces latency 
// $$kk: 08.06.99: revisit!!!
static int				g_audioCaptureBufferSizeDivisor = 2; 



// $$kk: 04.23.99: added this variable
// 1 if we are paused, otherwise 0.  using this to knock us out of the sleep loop
// when the device is paused.  otherwise, erratically, we may get stuck there forever
// when the device is resumed, so that the device is not paused but never becomes
// active.
int						g_paused = 0;			


// This proc drives audio capture.  This proc both feeds buffers to the device, into 
// which data is captured, and makes callbacks to deliver the captured data.  
// It may be run in a separate thread.  In the case of Solaris, a separate thread is
// required  because read() is (supposed to be) a blocking call.  
// When used with a Java thread, the context needs to be the  JNI environment pointer 
// valid for the thread running the proc.
void PV_AudioWaveInFrameThread(void* context)
{
    audio_info_t	sunAudioHeader;
	uint_t			deviceByteBufferSize;
    char			*pFillBuffer;
	int				currentBytesRead;
	int				totalBytesRead;
	long			currentPos, lastPos;
	long			count;

	long			framesToRead;			// number of sample frames to read per callback
	long			bytesToRead;			// number of bytes to read per callback
	long			buffersToRead;			// number of buffers to read from the device per callback
	long			bytesToReadPerBuffer;	// number of bytes to read from the device per buffer


	// $$kk: 10.14.98: need to make sure our capture buffer isn't larger than the system
	// one; otherwise we will overflow and miss data

	//fprintf(stderr, "> PV_AudioWaveInFrameThread, context: %d\n", context);

	ioctl(g_captureSound, AUDIO_GETINFO, &sunAudioHeader);
	deviceByteBufferSize = sunAudioHeader.record.buffer_size;	
	buffersToRead = g_captureByteBufferSize / deviceByteBufferSize;

	if (buffersToRead == 0)
	{
		buffersToRead = 1;		
		bytesToReadPerBuffer = g_captureByteBufferSize;
	}
	else
	{
		bytesToReadPerBuffer = deviceByteBufferSize;
	}

	bytesToRead = buffersToRead * bytesToReadPerBuffer; 
	framesToRead = bytesToRead / (g_channels * g_bitSize / 8);


	// flush anything we might have accumulated in the capture queue before starting the capture thread
	ioctl(g_captureSound, I_FLUSH, FLUSHR);

	currentPos = sunAudioHeader.record.samples;

	// $$kk: 11.03.98: this is a bit of a hack. 
	// we're not keeping track of our position very well.
	// here, if the device is not newly opened, we make sure to back off
	// our currentPos enough that we can reach our lastPos.  otherwise, 
	// when the device is paused and resumed (when the capture stream is
	// stopped and restarted), we get stuck in the while (currentPos < lastPos)
	// loop below....
	if ( (currentPos - framesToRead) >= 0 )
	{
		currentPos -= framesToRead;
	}

	lastPos = currentPos + framesToRead;

	while (g_captureShutdown == FALSE)
	{	
		pFillBuffer = (char *)g_captureBufferBlock;	
		totalBytesRead = 0;
		currentBytesRead = 0;


		// $$kk: 08.12.99: i'm taking this code out for now.  it's deadlock
		// prone and increases latency.  without it, we make a lot more 
		// callbacks.  is the performance hit worth it?  perhaps!
/*
		// wait for the device to record enough data to fill our capture buffer
		while (currentPos < lastPos)
		{
			if (g_captureShutdown == TRUE)
			{
				// do this here, when we can't call it anymore.
				g_captureDoneProc = NULL;

				// release read buffer
				HAE_Deallocate(g_captureBufferBlock);
				return;
			}

			// $$kk: 04.23.99: if we are paused, update the pos, drop out of this loop, 
			// and block down below on the read instead.  otherwise we might get stuck
			// here forever even once the device is resumed.

			if (g_paused) 
			{
				lastPos = currentPos;
				break;
			}

			HAE_SleepFrameThread(context, HAE_SOLARIS_SOUND_CAPTURE_PERIOD);		// in ms
				
			ioctl(g_captureSound, AUDIO_GETINFO, &sunAudioHeader);
			currentPos = sunAudioHeader.record.samples;
		}

		lastPos += framesToRead;
*/

		// now read the data from the device record buffer
		
		for (count = 0; count < buffersToRead; count++)
		{
			// This is a blocking call on Solaris unless the device is set to a non-blocking mode.  
			// It returns the number of bytes read or an error code.
			// $$kk: 10.13.98: this is not blocking for me, even when i don't open the device
			// with O_NONBLOCK....  what is the difference between O_NONBLOCK and O_NDELAY?
			currentBytesRead = read(g_captureSound, pFillBuffer, bytesToReadPerBuffer);
			pFillBuffer += currentBytesRead;
			totalBytesRead += currentBytesRead;
		}
		
		// callback to deliver the captured data
		
		if (totalBytesRead > 0)
		{
			// now callback with captured data
			(*g_captureDoneProc)(context, DATA_READY_CAPTURE, &g_captureBufferBlock, &totalBytesRead);		
		}
	}
}


// $$kk: 10.27.99: determine what formats are supported by the device

#define sbpro		0
#define sb16		1
#define audiocs		2
#define dbri		3
#define audioamd	4

static char sbproStr[] = "SUNW,sbpro";
static char sb16Str[] = "SUNW,sb16";
static char audiocsStr[] = "SUNW,CS4231";
static char dbriStr[] = "SUNW,dbri";
static char audioamdStr[] = "SUNW,am79c30";	


static int HAE_GetDriver()
{
	static int driver = -1;

	if (driver == -1)
	{
		int					pseudoDevice = 0;
		long				error;
		audio_device_t		deviceInfo;

		// $$kk: 05.12.99: need to handle AUDIODEV environment variable here!!
		pseudoDevice = open("/dev/audioctl", O_RDONLY);

		if (pseudoDevice == -1)
		{
			//fprintf(stderr, "could not get pseudoDevice\n");
		}
		else
		{
			error = ioctl(pseudoDevice, AUDIO_GETDEV, &deviceInfo);
			close(pseudoDevice);

			if (strcmp(sbproStr, deviceInfo.name) == 0)
			{
				driver = sbpro;
			}

			if (strcmp(sb16Str, deviceInfo.name) == 0)
			{
				driver = sb16;
			}

			if (strcmp(audiocsStr, deviceInfo.name) == 0)
			{
				driver = audiocs;
			}

			if (strcmp(dbriStr, deviceInfo.name) == 0)
			{
				driver = dbri;
			}

			if (strcmp(audioamdStr, deviceInfo.name) == 0)
			{
				driver = audioamd;
			}
		}
	}

	return driver;
}


int HAE_MaxCaptureFormats(long deviceID)
{
	int theDriver = HAE_GetDriver();
	int maxFormats = 0;

	switch(theDriver)
	{
	case sbpro:
		maxFormats = 16;
		break;
	case sb16:
		maxFormats = 6 * 2 * 2 * 2;
		break;
	case audiocs:
		maxFormats = 3 * 10 * 2 * 2;
		break;
	case dbri:
		maxFormats = 3 * 10 * 2 * 2;		
		break;
	case audioamd:
		maxFormats = 2;
		break;
	default:
		// unknown driver;
		break;
	}

	//return (maxFormats * 2 * 2);
	return (maxFormats);
}


int HAE_GetSupportedCaptureFormats(long deviceID, unsigned long *encodings, unsigned long *sampleRates, unsigned long *channels, unsigned long *bits, int maxFormats)
{
	char sbproStr[] = "SUNW,sbpro";
	char sb16Str[] = "SUNW,sb16";
	char audiocsStr[] = "SUNW,CS4231";
	char dbriStr[] = "SUNW,dbri";
	char audioamdStr[] = "SUNW,am79c30";	

	int encodingCount; int rateCount; int channelCount; int bitCount;
	int encodingIndex, rateIndex, channelIndex, bitIndex;
	int numSupportedFormats = 0;

	int					pseudoDevice = 0;
	long				error;
    audio_device_t		deviceInfo;


	// $$kk: 05.12.99: need to handle AUDIODEV environment variable here!!
	pseudoDevice = open("/dev/audioctl", O_RDONLY);

	if (pseudoDevice == -1)
	{
		//fprintf(stderr, "could not get pseudoDevice\n");
		return numSupportedFormats;
	}

	error = ioctl(pseudoDevice, AUDIO_GETDEV, &deviceInfo);
	close(pseudoDevice);

	
	// max formats: 16
	if (strcmp(sbproStr, deviceInfo.name) == 0)
	{
		// 8 bit linear unsigned data
		// mono: 4000Hz - 44100Hz (ugh; should just choose some here or what??
		// stereo: 11025Hz and 22050Hz only
		// AUDIO_ENCODING_LINEAR or AUDIO_ENCODING_ULAW; some also AUDIO_ENCODING_ALAW

		unsigned long sampleRateArray[] = { 8000, 11025, 16000, 22050, 32000, 44100 };
		rateCount = 6;

		// do 8-bit mono PCM case
		for (rateIndex = 0; rateIndex < rateCount; rateIndex++)
		{
			encodings[numSupportedFormats] = PCM;
			channels[numSupportedFormats] = 1;
			bits[numSupportedFormats] = 8;
			sampleRates[numSupportedFormats] = sampleRateArray[rateIndex];
			numSupportedFormats++;
			if (numSupportedFormats >= maxFormats) return numSupportedFormats;
		}

		// do 8-bit mono ULAW case
		for (rateIndex = 0; rateIndex < rateCount; rateIndex++)
		{
			encodings[numSupportedFormats] = ULAW;
			channels[numSupportedFormats] = 1;
			bits[numSupportedFormats] = 8;
			sampleRates[numSupportedFormats] = sampleRateArray[rateIndex];
			numSupportedFormats++;
			if (numSupportedFormats >= maxFormats) return numSupportedFormats;
		}

		// do 8-bit stereo PCM case: 11025Hz and 22050Hz

		// 11025Hz
		encodings[numSupportedFormats] = PCM;
		channels[numSupportedFormats] = 2;
		bits[numSupportedFormats] = 8;
		sampleRates[numSupportedFormats] = 11025;
		numSupportedFormats++;
		if (numSupportedFormats >= maxFormats) return numSupportedFormats;

		// 22050Hz
		encodings[numSupportedFormats] = PCM;
		channels[numSupportedFormats] = 2;
		bits[numSupportedFormats] = 8;
		sampleRates[numSupportedFormats] = 22050;
		numSupportedFormats++;
		if (numSupportedFormats >= maxFormats) return numSupportedFormats;

		// do 8-bit stereo ULAW case: 11025Hz and 22050Hz

		// 11025Hz
		encodings[numSupportedFormats] = ULAW;
		channels[numSupportedFormats] = 2;
		bits[numSupportedFormats] = 8;
		sampleRates[numSupportedFormats] = 11025;
		numSupportedFormats++;
		if (numSupportedFormats >= maxFormats) return numSupportedFormats;

		// 22050Hz
		encodings[numSupportedFormats] = ULAW;
		channels[numSupportedFormats] = 2;
		bits[numSupportedFormats] = 8;
		sampleRates[numSupportedFormats] = 22050;
		numSupportedFormats++;
		if (numSupportedFormats >= maxFormats) return numSupportedFormats;

		return numSupportedFormats;
	}


	// max formats: 48
	if (strcmp(sb16Str, deviceInfo.name) == 0)
	{
		// 8 bit linear unsigned data or 16 bit linear signed data
		// mono or stereo, 5000Hz - 44100Hz; some up to 48000Hz
		// just PCM or ULAW too?  presumable ULAW as well...?

		unsigned long encodingsArray[] = { PCM, ULAW };
		unsigned long sampleRateArray[] = { 8000, 11025, 16000, 22050, 32000, 44100 };
		unsigned long channelsArray[] = { 1, 2 };
		unsigned long bitsArray[] = { 8, 16 };

		encodingCount = 2;
		rateCount = 6;
		channelCount = 2;
		bitCount = 2;

		for (encodingIndex = 0; encodingIndex < encodingCount; encodingIndex++)
		{
			for (rateIndex = 0; rateIndex < rateCount; rateIndex++)
			{
				for (channelIndex = 0; channelIndex < channelCount; channelIndex++)
				{
					for (bitIndex = 0; bitIndex < bitCount; bitIndex++)
					{
						encodings[numSupportedFormats] = encodingsArray[encodingIndex];
						sampleRates[numSupportedFormats] = sampleRateArray[rateIndex];
						channels[numSupportedFormats] = channelsArray[channelIndex];
						bits[numSupportedFormats] = bitsArray[bitIndex];
						numSupportedFormats++;					
						if (numSupportedFormats >= maxFormats) return numSupportedFormats;
					}
				}
			}		
		} // for

		return numSupportedFormats;
	}

	// max formats: 40 (PCM: 10 * 2; ALAW: 10; ULAW: 10)
	if (strcmp(audiocsStr, deviceInfo.name) == 0)
	{
		// same as dbri...

		// not entirely sure about the signed / unsigned stuff here....
		// all sample rates for PCM, ULAW, and ALAW.  
		// only 8-bit mono for ULAW and ALAW

		unsigned long encodingsArray[] = { PCM, ULAW, ALAW };
		unsigned long sampleRateArray[] = { 8000, 9600, 11025, 16000, 18900, 22050, 32000, 37800, 44100, 48000 };
		unsigned long channelsArray[] = { 1, 2 };
		unsigned long bitsArray[] = { 8, 16 };

		encodingCount = 3;
		rateCount = 10;
		channelCount = 2;
		bitCount = 2;

		for (encodingIndex = 0; encodingIndex < encodingCount; encodingIndex++)
		{
			for (rateIndex = 0; rateIndex < rateCount; rateIndex++)
			{
				for (channelIndex = 0; channelIndex < channelCount; channelIndex++)
				{
					for (bitIndex = 0; bitIndex < bitCount; bitIndex++)
					{
						if ( ((encodingsArray[encodingIndex] == PCM) && (bitsArray[bitIndex] == 16)) || (((encodingsArray[encodingIndex] == ALAW) || (encodingsArray[encodingIndex] == ULAW)) && (channelsArray[channelIndex] == 1) && (bitsArray[bitIndex] == 8)))
						{
							encodings[numSupportedFormats] = encodingsArray[encodingIndex];
							sampleRates[numSupportedFormats] = sampleRateArray[rateIndex];
							channels[numSupportedFormats] = channelsArray[channelIndex];
							bits[numSupportedFormats] = bitsArray[bitIndex];
							numSupportedFormats++;					
							if (numSupportedFormats >= maxFormats) return numSupportedFormats;
						}
					}
				}
			}		
		} // for

		return numSupportedFormats;
	}

	// max formats: 40 (PCM: 10 * 2; ALAW: 10; ULAW: 10)
	if (strcmp(dbriStr, deviceInfo.name) == 0)
	{
		// same as audiocs...

		// not entirely sure about the signed / unsigned stuff here....
		// all sample rates for PCM, ULAW, and ALAW.  
		// only 8-bit mono for ULAW and ALAW

		unsigned long encodingsArray[] = { PCM, ULAW, ALAW };
		unsigned long sampleRateArray[] = { 8000, 9600, 11025, 16000, 18900, 22050, 32000, 37800, 44100, 48000 };
		unsigned long channelsArray[] = { 1, 2 };
		unsigned long bitsArray[] = { 8, 16 };

		encodingCount = 3;
		rateCount = 10;
		channelCount = 2;
		bitCount = 2;

		for (encodingIndex = 0; encodingIndex < encodingCount; encodingIndex++)
		{
			for (rateIndex = 0; rateIndex < rateCount; rateIndex++)
			{
				for (channelIndex = 0; channelIndex < channelCount; channelIndex++)
				{
					for (bitIndex = 0; bitIndex < bitCount; bitIndex++)
					{
						if ( ((encodingsArray[encodingIndex] == PCM) && (bitsArray[bitIndex] == 16)) || (((encodingsArray[encodingIndex] == ALAW) || (encodingsArray[encodingIndex] == ULAW)) && (channelsArray[channelIndex] == 1) && (bitsArray[bitIndex] == 8)))
						{
							encodings[numSupportedFormats] = encodingsArray[encodingIndex];
							sampleRates[numSupportedFormats] = sampleRateArray[rateIndex];
							channels[numSupportedFormats] = channelsArray[channelIndex];
							bits[numSupportedFormats] = bitsArray[bitIndex];
							numSupportedFormats++;					
							if (numSupportedFormats >= maxFormats) return numSupportedFormats;
						}
					}
				}
			}		
		} // for

		return numSupportedFormats;
	}

	// max formats: 2
	if (strcmp(audioamdStr, deviceInfo.name) == 0)
	{

		// 8000Hz 8-bit mono ULAW and ALAW only

		// ULAW case
		encodings[numSupportedFormats] = ULAW;
		channels[numSupportedFormats] = 1;
		bits[numSupportedFormats] = 8;
		sampleRates[numSupportedFormats] = 8000;
		numSupportedFormats++;
		if (numSupportedFormats >= maxFormats) return numSupportedFormats;

		// ALAW case
		encodings[numSupportedFormats] = ALAW;
		channels[numSupportedFormats] = 1;
		bits[numSupportedFormats] = 8;
		sampleRates[numSupportedFormats] = 8000;
		numSupportedFormats++;
		if (numSupportedFormats >= maxFormats) return numSupportedFormats;

		return numSupportedFormats;
	}

	// fprintf(stderr, "unrecognized audio driver; format support unknown\n");
	return numSupportedFormats;
}


// Aquire and enabled audio card
// return 0 if ok, -1 if failed
int HAE_AquireAudioCapture(void *context, unsigned long encoding, unsigned long sampleRate, unsigned long channels, unsigned long bits,
							unsigned long audioFramesPerBuffer, unsigned long *pCaptureHandle)
{
    audio_info_t	sunAudioHeader;
	long			error = -1;
	char			*pAudioDev = getenv("AUDIODEV");
	long			minFramesPerBuffer;

	if (pAudioDev == NULL) 
	{
		pAudioDev = "/dev/audio";
	}

	g_encoding = encoding;
	g_bitSize = bits;
	g_channels = channels;
	g_sampleRate = sampleRate;	

	// $$jb: 05.19.99: Setting the buffer size
	// $$kk: 08.06.99: i got rid of the fixed minimum value for buffersize
	// minFramesPerBuffer = sampleRate * HAE_SOLARIS_MIN_BUFFERSIZE_IN_MS / 1000;

	if( audioFramesPerBuffer == 0 ) {
		audioFramesPerBuffer = sampleRate * HAE_SOLARIS_DEFAULT_BUFFERSIZE_IN_MS / 1000;
	}


	// $$kk: 08.06.99: got rid of fixed minimum buffer size
	/*
	if ( audioFramesPerBuffer >= minFramesPerBuffer ) {
		g_audioFramesToRead = audioFramesPerBuffer;
	} else {
		g_audioFramesToRead = minFramesPerBuffer;
	}
	*/
	g_audioFramesToRead = audioFramesPerBuffer / g_audioCaptureBufferSizeDivisor;


	if (pCaptureHandle)
	{
		*pCaptureHandle = 0L;
	}

	// try to open wave device for recording
	// $$kk: 12.17.97: need O_NONBLOCK flag to be compatible with windows

	// $$kk: 10.13.98: we want O_NONBLOCK so that we return failure immediately if the
	// device is busy (or absent or whatever) rather than blocking.  however, i think that
	// this same O_NONBLOCK flag dictates whether the read() calls should block.  even 
	// without the O_NONBLOCK flag set, read() does *not* block for me, so i'm keeping 
	// the flag for now....
	
	g_captureSound = open(pAudioDev,O_RDONLY|O_NONBLOCK);

	if (g_captureSound > 0)
	{
		AUDIO_INITINFO(&sunAudioHeader);
		
		// Set capture format of the sun device.
		sunAudioHeader.record.sample_rate = sampleRate;
		sunAudioHeader.record.precision = bits;
		sunAudioHeader.record.channels = channels;
		sunAudioHeader.record.buffer_size = g_audioFramesToRead * channels * bits / 8;
		
		sunAudioHeader.record.encoding = AUDIO_ENCODING_LINEAR;
		if (g_encoding == ULAW)
		{
			sunAudioHeader.record.encoding = AUDIO_ENCODING_ULAW;			
		} 
		else if (g_encoding == ALAW)
		{
			sunAudioHeader.record.encoding = AUDIO_ENCODING_ALAW;			
		} 



		// start out paused so we don't overflow the device driver buffers
		sunAudioHeader.record.pause = 1;
		error = ioctl(g_captureSound, AUDIO_SETINFO, &sunAudioHeader);

		if (error == 0)
		{
			// flush anything we might have accumulated in the capture queue before pausing
			error = ioctl(g_captureSound, I_FLUSH, FLUSHR);

			error = ioctl(g_captureSound, AUDIO_GETINFO, &sunAudioHeader);
			g_audioFramesToRead = sunAudioHeader.record.buffer_size / (channels * bits / 8);


			if (error == 0)
			{
				if (pCaptureHandle)
				{
					*pCaptureHandle = (unsigned long)g_captureSound;
				}
			} 
		}
	}

	if (error != 0)
	{	// something failed
		HAE_ReleaseAudioCapture(context);
	}

	//fprintf(stderr, "<< HAE_API_SolarisOS_Capture: HAE_AquireAudioCapture() returning %d\n", error);
	return (error == 0 ? 0 : -1);
}


// Given the capture hardware is working, fill a buffer with some data. This call is
// asynchronous. When this buffer is filled, the function done will be called.
// returns 0 for success, -1 for failure
int HAE_StartAudioCapture(HAE_CaptureDone done, void *callbackContext)
{	
    audio_info_t	sunAudioHeader;
	long			error = -1;

	//fprintf(stderr, ">> HAE_API_SolarisOS_Capture: HAE_StartAudioCapture()\n");
	
	if (g_captureSound)
	{	
		// start capture
		g_captureDoneProc = done;
		
		// allocate the capture buffer
		// $$jb: 05.19.99:  This is set in HAE_AquireAudioCapture
		//g_audioFramesToRead = HAE_SOLARIS_FRAMES_PER_BUFFER;	// our read buffer will hold this many frames of sampled audio data

		// we're going to build this many buffers at a time
		if (g_bitSize == 8)
		{
			g_captureByteBufferSize = (sizeof(char) * g_audioFramesToRead);
		}
		else
		{
			g_captureByteBufferSize = (sizeof(short int) * g_audioFramesToRead);
		}
		g_captureByteBufferSize *= g_channels;

		// allocate read buffer
		g_captureBufferBlock = HAE_Allocate(g_captureByteBufferSize);

		if (g_captureBufferBlock)
		{
			// $$kk: 10.12.98: added this so we can restart capture
			// $$kk: 11.03.98: mark ourselves active *before* creating frame thread
			g_captureShutdown = FALSE;
			
			// create thread to manage audio capture
	
			error = HAE_CreateFrameThread(callbackContext, PV_AudioWaveInFrameThread);

			if (error == 0)
			{
				error = HAE_ResumeAudioCapture();
			}
		}

		if (error != 0) 
		{
			g_captureShutdown = TRUE;
		}
	}

	//fprintf(stderr, "<< HAE_API_SolarisOS_Capture: HAE_StartAudioCapture() returning %d\n", error);
	return (error == 0) ? 0 : -1;
}


// stop the capture hardware
// returns 0 for success, -1 for failure
int HAE_StopAudioCapture(void)
{
    audio_info_t	sunAudioHeader;
	long			error = -1;

	//fprintf(stderr, ">> HAE_API_SolarisOS_Capture: HAE_StopAudioCapture()\n");
	
	if (g_captureSound)
	{
		g_captureShutdown = TRUE;

	    // stop streaming data
		error = HAE_PauseAudioCapture();
	}

	// $$kk: 04.13.99: should do this regardless of error value??
	if (error == 0)
	{
		// destroy the audio capture thread
		error = HAE_DestroyFrameThread(NULL);
	}

	//fprintf(stderr, "<< HAE_API_SolarisOS_Capture: HAE_StopAudioCapture() returning %d\n", error);
	return (error == 0) ? 0 : -1;
}


// returns 0 for success, -1 for failure
int HAE_PauseAudioCapture(void)
{
    audio_info_t	sunAudioHeader;
	long			error = -1;

	//fprintf(stderr, ">> HAE_API_SolarisOS_Capture: HAE_PauseAudioCapture()\n");
	
	if (g_captureSound)
	{
		AUDIO_INITINFO(&sunAudioHeader);
		error = ioctl(g_captureSound, AUDIO_GETINFO, &sunAudioHeader);
		
		if (error == 0)
		{
			// pause capture
			sunAudioHeader.record.pause = 1;
			error = ioctl(g_captureSound, AUDIO_SETINFO, &sunAudioHeader);
		}

		// $$kk: 04.23.99: added this
		if (error == 0)
		{
			g_paused = 1;
		}
	}

	//fprintf(stderr, "<< HAE_API_SolarisOS_Capture: HAE_PauseAudioCapture() returning %d\n", error);
	return (error == 0) ? 0 : -1;
}


// returns 0 for success, -1 for failure
int HAE_ResumeAudioCapture(void)
{
    audio_info_t	sunAudioHeader;
	long			error = -1;

	//fprintf(stderr, ">> HAE_API_SolarisOS_Capture: HAE_ResumeAudioCapture()\n");
	
	if (g_captureSound)
	{	
		AUDIO_INITINFO(&sunAudioHeader);
		error = ioctl(g_captureSound, AUDIO_GETINFO, &sunAudioHeader);
		
		if (error == 0)
		{
			// unpause capture
			sunAudioHeader.record.pause = 0;
			error = ioctl(g_captureSound, AUDIO_SETINFO, &sunAudioHeader);
		}

		// $$kk: 04.23.99: added this
		if (error == 0)
		{
			g_paused = 0;
		}
	}

	//fprintf(stderr, "<< HAE_API_SolarisOS_Capture: HAE_ResumeAudioCapture() returning %d\n", error);	
	return (error == 0) ? 0 : -1;
}



// Release and free audio card.
// returns 0 for success, -1 for failure
int HAE_ReleaseAudioCapture(void *context)
{
	//fprintf(stderr, ">> HAE_API_SolarisOS_Capture: HAE_ReleaseAudioCapture()\n");
	
	if (g_captureSound)
	{
		close(g_captureSound);
		g_captureSound = 0;
	}

	//fprintf(stderr, "<< HAE_API_SolarisOS_Capture: HAE_ResumeAudioCapture() returning 0\n");	
	return 0;
}


// number of devices. ie different versions of the HAE connection. DirectSound and waveOut
// return number of devices. ie 1 is one device, 2 is two devices.
// NOTE: This function needs to function before any other calls may have happened.
long HAE_MaxCaptureDevices(void)
{
	return 1;	// only AUDIODEV (defaulting to /dev/audio)
}


// set the current device. device is from 0 to HAE_MaxDevices()
// NOTE:	This function needs to function before any other calls may have happened.
//			Also you will need to call HAE_ReleaseAudioCard then HAE_AquireAudioCard
//			in order for the change to take place.
void HAE_SetCaptureDeviceID(long deviceID, void *deviceParameter)
{
	if (deviceID < HAE_MaxCaptureDevices())
	{
		g_soundDeviceIndex = deviceID;
	}
	deviceParameter;
}


// return current device ID
// NOTE: This function needs to function before any other calls may have happened.
long HAE_GetCaptureDeviceID(void *deviceParameter)
{
	return g_soundDeviceIndex;
}


// get deviceID name 
// NOTE:	This function needs to function before any other calls may have happened.
//			The names returned in this function are platform specific.
void HAE_GetCaptureDeviceName(long deviceID, char *cName, unsigned long cNameLength)
{
	char		*data;
	static char *names[] =
			{	"SolarisOS,dev/audio,multi threaded",
			};
	if (cName && cNameLength)
	{
		if (deviceID < HAE_MaxDevices())
		{
			// can't do string functions here, it might be bad to do.
			data = names[deviceID];
			while (*data && (cNameLength > 0))
			{
				*cName = *data;
				cName++;
				data++;
				cNameLength--;
			}
			*cName = 0;
		}
	}
}


// return the number of frames in the capture buffer
// (should make this settable?)
unsigned long HAE_GetCaptureBufferSizeInFrames()
{
	return g_audioFramesToRead * g_audioCaptureBufferSizeDivisor;
}


// return the number of buffers used.
int HAE_GetCaptureBufferCount()
{
	return HAE_SOLARIS_NUM_CAPTURE_BUFFERS;
}

// return the number of samples captured at the device
// $$kk: 10.15.98: need to implement!
unsigned long HAE_GetDeviceSamplesCapturedPosition()
{
	return 0L;
}

void HAE_FlushAudioCapture() 
{
	ioctl(g_captureSound, I_FLUSH, FLUSHR);
}

// EOF of HAE_API_SolarisOS_Capture.c
