/*
 * @(#)awt_Cursor.c	1.11 01/02/09
 *
 * Copyright 1993-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

#include "awt_Component.h"
#include "awt_Cursor.h"
#include "java_awt_Cursor.h"
#include <X11/cursorfont.h>

#include "jni.h"
#include "jni_util.h"

/* fieldIDs for Cursor fields that may be accessed from C */
struct CursorIDs cursorIDs;
extern struct MComponentPeerIDs mComponentPeerIDs;

static jweak curComp = 0;

/*
 * Class:     java_awt_Cursor
 * Method:    initIDs
 * Signature: ()V
 */
/* 
 * This function gets called from the static initializer for Cursor.java
 * to initialize the fieldIDs for fields that may be accessed from C 
 */
JNIEXPORT void JNICALL
Java_java_awt_Cursor_initIDs(JNIEnv *env, jclass cls)
{
    cursorIDs.type = (*env)->GetFieldID(env, cls, "type", "I");
    cursorIDs.pData = (*env)->GetFieldID(env, cls, "pData", "J");
}

/*
 * A utility to retrieve cursor from java.awt.Cursor
 * Create and save the cursor first if it is not yet 
 */
Cursor getCursor(JNIEnv *env, jobject jCur)
{
    int cursorType = 0;
    Cursor xcursor;
    
    xcursor = (Cursor)JNU_GetLongFieldAsPtr(env, jCur, cursorIDs.pData);
    
    if (xcursor != 0) {
	return xcursor;
    }
    
    cursorType = (*env)->GetIntField(env, jCur, cursorIDs.type);
    
    DASSERT(cursorType != java_awt_Cursor_CUSTOM_CURSOR);
    
    switch (cursorType) {
    case java_awt_Cursor_DEFAULT_CURSOR:
	cursorType = XC_left_ptr;
        break;
    case java_awt_Cursor_CROSSHAIR_CURSOR:
	cursorType = XC_crosshair;
        break;
    case java_awt_Cursor_TEXT_CURSOR:
	cursorType = XC_xterm;
        break;
    case java_awt_Cursor_WAIT_CURSOR:
	cursorType = XC_watch;
        break;
    case java_awt_Cursor_SW_RESIZE_CURSOR:
	cursorType = XC_bottom_left_corner;
        break;
    case java_awt_Cursor_NW_RESIZE_CURSOR:
	cursorType = XC_top_left_corner;
        break;
    case java_awt_Cursor_SE_RESIZE_CURSOR:
	cursorType = XC_bottom_right_corner;
        break;
    case java_awt_Cursor_NE_RESIZE_CURSOR:
	cursorType = XC_top_right_corner;
        break;
    case java_awt_Cursor_S_RESIZE_CURSOR:
	cursorType = XC_bottom_side;
        break;
    case java_awt_Cursor_N_RESIZE_CURSOR:
	cursorType = XC_top_side;
        break;
    case java_awt_Cursor_W_RESIZE_CURSOR:
	cursorType = XC_left_side;
        break;
    case java_awt_Cursor_E_RESIZE_CURSOR:
	cursorType = XC_right_side;
        break;
    case java_awt_Cursor_HAND_CURSOR:
	cursorType = XC_hand2;
        break;
    case java_awt_Cursor_MOVE_CURSOR:
	cursorType = XC_fleur;
        break;
    }
    xcursor = XCreateFontCursor(awt_display, cursorType);
    
    JNU_SetLongFieldFromPtr(env, jCur, cursorIDs.pData, xcursor);
    return xcursor;
}

/*
 * Class:     java_awt_Cursor
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_awt_Cursor_finalize(JNIEnv *env, jobject this)
{
    Cursor xcursor;
    
    xcursor = (Cursor)JNU_GetLongFieldAsPtr(env, this, cursorIDs.pData);
    if (xcursor != 0) {
        AWT_LOCK();
	JNU_SetLongFieldFromPtr(env, this, cursorIDs.pData, 0);
	XFreeCursor(awt_display, xcursor);
        AWT_UNLOCK();
    }
}

/*
 *  normal replace : CACHE_UDPATE  => update curComp and updateCursor
 *  not replace    : UPDATE_ONLY   => intact curComp but updateCursor
 *  only replace   : CACHE_ONLY    => update curComp only, not updateCursor 
 */
void updateCursor(XPointer client_data, int replace) { 
    
    static jclass globalCursorManagerClass = NULL;
    static jmethodID updateCursorID = NULL;
    
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject peer = (jobject) client_data;
    jobject target; 
    
    if ((*env)->PushLocalFrame(env, 16) < 0)
        return;
      
    target = (*env)->GetObjectField(env, peer, mComponentPeerIDs.target);
    if (replace != UPDATE_ONLY) {
	if (!JNU_IsNull(env, curComp)) {
	    (*env)->DeleteWeakGlobalRef(env, curComp);
        }
	curComp = (*env)->NewWeakGlobalRef(env, target);
	if (replace == CACHE_ONLY) {
	    (*env)->PopLocalFrame(env, 0);
	    return;
	}
    }
    
    /* Initialize our java identifiers once. Checking before locking
     * is a huge performance win.
     */
    if (globalCursorManagerClass == NULL) {
	jobject sysClass = (*env)->FindClass(env, "sun/awt/GlobalCursorManager");
        if (sysClass != NULL) {
	    /* Make this class 'sticky', we don't want it GC'd */
            globalCursorManagerClass = (*env)->NewGlobalRef(env, sysClass);
	    
            updateCursorID = (*env)->GetStaticMethodID(env,
						       globalCursorManagerClass,
						       "updateCursorLater",
						       "(Ljava/awt/Component;)V"
						       );
        }
	if (JNU_IsNull(env, globalCursorManagerClass) || updateCursorID == 0) {
            JNU_ThrowClassNotFoundException(env, "sun/awt/GlobalCursorManager");
	    (*env)->PopLocalFrame(env, 0);
            return;
        }
    } /* globalCursorManagerClass == NULL*/
    
    (*env)->CallStaticVoidMethod(env, globalCursorManagerClass,
				 updateCursorID, target);
    DASSERT(!((*env)->ExceptionOccurred(env)));
    (*env)->PopLocalFrame(env, 0);
}

jobject getCurComponent() {
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    return (*env)->NewLocalRef(env, curComp);
}
