/*
 * @(#)mlib_v_ImageConv3x3_32.c	1.4 00/02/02
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

#pragma ident  "@(#)mlib_v_ImageConv3x3_32.c	1.4  98/10/20 SMI"

/*
 * FUNCTION
 *      mlib_v_conv3x3nw_32.c  - convolve an 32-bit image with a 3x3 kernel
 *      mlib_v_conv3x3ext_32.c - convolve an 32-bit image with a 3x3 kernel
 *
 *
 * SYNOPSIS
 *  mlib_status
 *  mlib_v_conv3x3nw_32( mlib_image *dst,
 *                       mlib_image *src,
 *                       mlib_s32 *kernel,
 *                       mlib_s32 scalef_expon
 *                       mlib_s32 cmask)
 *
 *  mlib_status
 *  mlib_v_conv3x3ext_32( mlib_image *dst,
 *                        mlib_image *src,
 *                        mlib_s32 dx_l,
 *                        mlib_s32 dx_r,
 *                        mlib_s32 dy_t,
 *                        mlib_s32 dy_b,
 *                        mlib_s32 *kernel,
 *                        mlib_s32 scalef_expon
 *                        mlib_s32 cmask)
 * ARGUMENT
 *      src           Pointer to structure of source image
 *      dst           Pointer to structure of destination image
 *      kernel        Convolution kernel
 *      scalef_expons The scaling factor to convert the input integer
 *                    coefficients into floating-point coefficients:
 *                    floating-point coefficient = integer coefficient * 2^(-scale)
 *      cmask         Channel mask to indicate the channels to be convolved.
 *                    Each bit of which represents a channel in the image. The
 *                    channels corresponded to 1 bits are those to be processed.
 *
 * DESCRIPTION
 *      A 2-D convolution (3x3 kernel) for 32-bit images.
 */

#include <mlib_image.h>

/*****************************************************************************************************/
#define LEFT   1
#define RIGHT  1
#define TOP    1
#define BOTTOM 1
/*****************************************************************************************************/
#define SAT_S32(x) (mlib_s32)(x)
/*****************************************************************************************************/
#define DEFINE_IDENTIFICATORS(type)                                                                   \
  type     *sl, *adr_src;                       /* pointers for pixel and line of source */           \
  type     *da, *dl, *adr_dst;                  /* pointers for pixel and line of destination */      \
  mlib_s32 row, col, block;                     /* indices */                                         \
  mlib_s32 *row0, *row1, *row2;                                                                       \
  mlib_d64 scalef;                              /* divide coeffs by scalefactor=2**scalef_expon */    \
  mlib_d64 sum1, sum2, sum3;                    /* accumulators */                                    \
  mlib_d64 k0, k1, k2, k3, k4, k5, k6, k7, k8;  /* coeff's */                                         \
  mlib_d64 p00, p01, p02, p03, p04,             /* row 0 pixel values */                              \
           p10, p11, p12, p13, p14,             /* row 1 pixel values */                              \
           p20, p21, p22, p23, p24;             /* row 2 pixel values */                              \
  mlib_s32 sll,                                 /* line length (in pixels) of source image */         \
           dll,                                 /* line length (in pixels) of destination image */    \
           wid,                                 /* src width in pixels */                             \
           hgt;                                 /* src height in pixels */                            \
  mlib_s32 num_blocks;                          /* do output pts 3 at a time; loop count for this */  \
  mlib_s32 leftover_cols;                       /* what can't be done in blocks */                    \
  mlib_s32 chan, testchan, nchannel;
/*****************************************************************************************************/
#define LOAD_KERNEL_INTO_DOUBLE()                                                                     \
  while (scalef_expon > 30) {                                                                         \
    scalef /= (1 << 30);                                                                              \
    scalef_expon -= 30;                                                                               \
  }                                                                                                   \
                                                                                                      \
  scalef /= (1 << scalef_expon);                                                                      \
                                                                                                      \
  /* keep kernel in regs */                                                                           \
  k0 = scalef * kern[0];  k1 = scalef * kern[1];  k2 = scalef * kern[2];                              \
  k3 = scalef * kern[3];  k4 = scalef * kern[4];  k5 = scalef * kern[5];                              \
  k6 = scalef * kern[6];  k7 = scalef * kern[7];  k8 = scalef * kern[8];
/*****************************************************************************************************/
#define GET_SRC_DST_PARAMETERS(type)                                                                  \
  hgt = mlib_ImageGetHeight(src);                                                                     \
  wid = mlib_ImageGetWidth(src);                                                                      \
  nchannel = mlib_ImageGetChannels(src);                                                              \
  sll = mlib_ImageGetStride(src) / sizeof(type);                                                      \
  dll = mlib_ImageGetStride(dst) / sizeof(type);                                                      \
  adr_src = (type *)mlib_ImageGetData(src);                                                           \
  adr_dst = (type *)mlib_ImageGetData(dst);
/*****************************************************************************************************/
#define INI_THREE_PIXELS_S32() sum1 = sum2 = sum3 = 0.;
#define INI_ONE_PIXEL_S32() sum1 = 0.;
/*****************************************************************************************************/
#define CONVOLVE_THREE_PIXELS(dst, shift, format)                                                     \
  p02 = row0[col]; p12 = row1[col]; p22 = row2[col]; col += shift;                                    \
  p03 = row0[col]; p13 = row1[col]; p23 = row2[col]; col += shift;                                    \
  p04 = row0[col]; p14 = row1[col]; p24 = row2[col]; col += shift;                                    \
                                                                                                      \
  INI_THREE_PIXELS_##format()                                                                         \
                                                                                                      \
  sum1 += p00 * k0; sum2 += p01 * k0; sum3 += p02 * k0;                                               \
  sum1 += p01 * k1; sum2 += p02 * k1; sum3 += p03 * k1;                                               \
  sum1 += p02 * k2; sum2 += p03 * k2; sum3 += p04 * k2;                                               \
  sum1 += p10 * k3; sum2 += p11 * k3; sum3 += p12 * k3;                                               \
  sum1 += p11 * k4; sum2 += p12 * k4; sum3 += p13 * k4;                                               \
  sum1 += p12 * k5; sum2 += p13 * k5; sum3 += p14 * k5;                                               \
  sum1 += p20 * k6; sum2 += p21 * k6; sum3 += p22 * k6;                                               \
  sum1 += p21 * k7; sum2 += p22 * k7; sum3 += p23 * k7;                                               \
  sum1 += p22 * k8; sum2 += p23 * k8; sum3 += p24 * k8;                                               \
                                                                                                      \
  *dst = SAT_##format(sum1); dst += shift;                                                            \
  *dst = SAT_##format(sum2); dst += shift;                                                            \
  *dst = SAT_##format(sum3); dst += shift;                                                            \
                                                                                                      \
  p00 = p03; p01 = p04;                    /* shift last 2 row0 elements */                           \
  p10 = p13; p11 = p14;                    /* shift last 2 row1 elements */                           \
  p20 = p23; p21 = p24;                    /* shift last 2 row2 elements */
/*****************************************************************************************************/
#define CONVOLVE_ONE_PIXEL(dst, shift, format)                                                        \
  p02 = row0[col];                                                                                    \
  p12 = row1[col];                                                                                    \
  p22 = row2[col]; col += shift;                                                                      \
  INI_ONE_PIXEL_##format()                                                                            \
  sum1 += p00 * k0;                        /* sum of prods */                                         \
  sum1 += p01 * k1;                                                                                   \
  sum1 += p02 * k2;                                                                                   \
  sum1 += p10 * k3;                                                                                   \
  sum1 += p11 * k4;                                                                                   \
  sum1 += p12 * k5;                                                                                   \
  sum1 += p20 * k6;                                                                                   \
  sum1 += p21 * k7;                                                                                   \
  sum1 += p22 * k8;                                                                                   \
  *dst = SAT_##format(sum1);                                                                          \
  dst += shift;                                                                                       \
  p00 = p01; p01 = p02;                    /* shift last 2 row0 elements */                           \
  p10 = p11; p11 = p12;                    /* shift last 2 row1 elements */                           \
  p20 = p21; p21 = p22;                    /* shift last 2 row2 elements */
/*****************************************************************************************************/

mlib_status mlib_v_conv3x3_32nw( mlib_image *dst,        /* source image */
                                 mlib_image *src,        /* destination image */
                                 mlib_s32 *kern,         /* convolution kernel */
                                 mlib_s32 scalef_expon,  /* scalefactor=2**N */
                                 mlib_s32 cmask)
{
  DEFINE_IDENTIFICATORS(mlib_s32)

  scalef = 1.0;

  LOAD_KERNEL_INTO_DOUBLE()

  GET_SRC_DST_PARAMETERS(mlib_s32)

  num_blocks = (wid - 2)/3;                      /* (wid-2): no output on 1st, last columns */
  leftover_cols = (wid - 2) - (num_blocks * 3);
  hgt -= 2;                       /* boundaries: no output for first & last rows */

  testchan = 1;
  for ( chan = nchannel-1; chan >= 0; chan--) {
    if ((cmask & testchan) == 0) {
      testchan <<= 1;
      continue;
    }
    testchan <<= 1;
    sl = adr_src + chan;
    dl = da = adr_dst + chan;
    dl = da += dll + nchannel;

  /*
   * set up rowbufs 1 and 2;
   * rowbuf 0 is always reloaded inside the loop:
   */

    row1 = sl;
    sl += sll;
    row2 = sl;
    sl += sll;

    for ( row = 0; row < hgt; row++ ) {

      row0 = row1;
      row1 = row2;
      row2 = sl;

      col = 0;

      p00 = row0[col]; p10 = row1[col]; p20 = row2[col]; col += nchannel;
      p01 = row0[col]; p11 = row1[col]; p21 = row2[col]; col += nchannel;

      /* Process along row, getting 3 output points per loop iteration: */

#pragma pipeloop(0)
      for ( block = 0; block < num_blocks; block++ ) {
        CONVOLVE_THREE_PIXELS(da, nchannel, S32)
      }

      if (leftover_cols) {
        CONVOLVE_ONE_PIXEL(da, nchannel, S32)
        if (leftover_cols > 1) {
          CONVOLVE_ONE_PIXEL(da, nchannel, S32)
        }
      }

      sl = sl + sll;                             /* set src ptrs to next row */
      dl = da = dl + dll;                        /* set dst ptrs to next row */
    }
  }

  return( MLIB_SUCCESS );
}

/*****************************************************************************************************/

mlib_status mlib_v_conv3x3_32ext( mlib_image *dst,        /* source image */
                                  mlib_image *src,        /* destination image */
                                  mlib_s32   dx_l,
                                  mlib_s32   dx_r,
                                  mlib_s32   dy_t,
                                  mlib_s32   dy_b,
                                  mlib_s32   *kern,         /* convolution kernel */
                                  mlib_s32   scalef_expon,  /* scalefactor=2**N */
                                  mlib_s32   cmask)
{
  DEFINE_IDENTIFICATORS(mlib_s32)
  mlib_s32 i, dend;

  scalef = 1.0;

  LOAD_KERNEL_INTO_DOUBLE()

  GET_SRC_DST_PARAMETERS(mlib_s32)

  dend = (dx_r > wid) ? wid : dx_r;
  num_blocks = (wid - dend)/3;
  leftover_cols = (wid - dend) - (num_blocks * 3);

  testchan = 1;
  for ( chan = nchannel-1; chan >= 0; chan--) {
    if ((cmask & testchan) == 0) {
      testchan <<= 1;
      continue;
    }
    testchan <<= 1;
    sl = adr_src + chan;
    dl = da = adr_dst + chan;

  /*
   * set up rowbufs 1 and 2;
   * rowbuf 0 is always reloaded inside the loop:
   */

    row = 0;
    row1 = sl; row++;
    if ((row > dy_t) && (row < hgt + TOP + BOTTOM - dy_b)) sl += sll;
    row2 = sl; row++;
    if ((row > dy_t) && (row < hgt + TOP + BOTTOM - dy_b)) sl += sll;

    for ( row = 0; row < hgt; row++ ) {

      row0 = row1;
      row1 = row2;
      row2 = sl;

      col = 0;
      i = 0;

      p00 = row0[col]; p10 = row1[col]; p20 = row2[col]; i++;
      if ((i > dx_l) && (i < wid + LEFT + RIGHT - dx_r)) col += nchannel;
      p01 = row0[col]; p11 = row1[col]; p21 = row2[col]; col += nchannel;

      /* Process along row, getting 3 output points per loop iteration: */

#pragma pipeloop(0)
      for ( block = 0; block < num_blocks; block++ ) {
        CONVOLVE_THREE_PIXELS(da, nchannel, S32)
      }

      if (leftover_cols) {
        CONVOLVE_ONE_PIXEL(da, nchannel, S32)
        if (leftover_cols > 1) {
          CONVOLVE_ONE_PIXEL(da, nchannel, S32)
        }
      }

      for (i = 0; i < dend; i++) {
        col -= nchannel;
        CONVOLVE_ONE_PIXEL(da, nchannel, S32)
      }

      if (row < hgt - dy_b - 1) sl = sl + sll;   /* set src ptrs to next row */
      dl = da = dl + dll;                        /* set dst ptrs to next row */
    }
  }

  return( MLIB_SUCCESS );
}

/*****************************************************************************************************/

