/*
 * @(#)mlib_v_ImageConv5x5_32.c	1.4 00/02/02
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

#pragma ident  "@(#)mlib_v_ImageConv5x5_32.c	1.2  98/10/20 SMI"

/*
 * FUNCTION
 *      mlib_v_conv5x5_32nw.c  - convolve an 32-bit image with a 5x5 kernel
 *      mlib_v_conv5x5_32ext.c - convolve an 32-bit image with a 5x5 kernel
 *
 *
 * SYNOPSIS
 *  mlib_status
 *  mlib_v_conv5x5_32nw( mlib_image *dst,
 *                       mlib_image *src,
 *                       mlib_s32 *kernel,
 *                       mlib_s32 scalef_expon
 *                       mlib_s32 cmask)
 *
 *  mlib_status
 *  mlib_v_conv5x5_32ext( mlib_image *dst,
 *                        mlib_image *src,
 *                        mlib_s32   dx_l,
 *                        mlib_s32   dx_r,
 *                        mlib_s32   dy_t,
 *                        mlib_s32   dy_b,
 *                        mlib_s32 *kernel,
 *                        mlib_s32 scalef_expon
 *                        mlib_s32 cmask)
 * ARGUMENT
 *      src           Pointer to structure of source image
 *      dst           Pointer to structure of destination image
 *      kernel        Convolution kernel
 *      scalef_expons The scaling factor to convert the input integer
 *                    coefficients into floating-point coefficients:
 *                    floating-point coefficient = integer coefficient * 2^(-scale)
 *      cmask         Channel mask to indicate the channels to be convolved.
 *                    Each bit of which represents a channel in the image. The
 *                    channels corresponded to 1 bits are those to be processed.
 *
 * DESCRIPTION
 *      A 2-D convolution (5x5 kernel) for 32-bit images.
 */

#include <mlib_image.h>
#include <stdlib.h>

/*****************************************************************************************************/
#define LEFT   2
#define RIGHT  2
#define TOP    2
#define BOTTOM 2
/*****************************************************************************************************/
#define SAT_S32(x) (mlib_s32)(x)
/*****************************************************************************************************/
#define DEFINE_IDENTIFICATORS(type)                                                                   \
  type     *sl, *adr_src;                       /* pointers for pixel and line of source */           \
  type     *da, *dl, *adr_dst;                  /* pointers for pixel and line of destination */      \
  mlib_s32 row, col, block;                     /* indices */                                         \
  mlib_s32 *row0, *row1, *row2, *row3, *row4;                                                         \
  mlib_d64 scalef;                              /* divide coeffs by scalefactor=2**scalef_expon */    \
  mlib_d64 sum1, sum2, sum3;                    /* accumulators */                                    \
  mlib_d64 k0, k1, k2, k3, k4, k5, k6, k7, k8;   /* coeff's */                                        \
  mlib_d64 k9, k10, k11, k12, k13, k14, k15;     /* coeff's */                                        \
  mlib_d64 k16, k17, k18, k19, k20, k21, k22;    /* coeff's */                                        \
  mlib_d64 k23, k24;                             /* coeff's */                                        \
  mlib_d64 p00, p01, p02, p03, p04, p05, p06,    /* row 0 pixel values */                             \
           p10, p11, p12, p13, p14, p15, p16,    /* row 1 pixel values */                             \
           p20, p21, p22, p23, p24, p25, p26;    /* row 2 pixel values */                             \
  mlib_s32 sll,                                 /* line length (in pixels) of source image */         \
           dll,                                 /* line length (in pixels) of destination image */    \
           wid,                                 /* src width in pixels */                             \
           hgt;                                 /* src height in pixels */                            \
  mlib_s32 num_blocks;                          /* do output pts 3 at a time; loop count for this */  \
  mlib_s32 leftover_cols;                       /* what can't be done in blocks */                    \
  mlib_s32 chan, testchan, nchannel;                                                                  \
  mlib_d64 *doutbuf, *d_outbuf;
/*****************************************************************************************************/
#define LOAD_KERNEL_INTO_DOUBLE()                                                                     \
  while (scalef_expon > 30) {                                                                         \
    scalef /= (1 << 30);                                                                              \
    scalef_expon -= 30;                                                                               \
  }                                                                                                   \
                                                                                                      \
  scalef /= (1 << scalef_expon);                                                                      \
                                                                                                      \
  /* keep kernel in regs */                                                                           \
  k0 = scalef * kern[0];  k1 = scalef * kern[1];  k2 = scalef * kern[2];                              \
  k3 = scalef * kern[3];  k4 = scalef * kern[4];  k5 = scalef * kern[5];                              \
  k6 = scalef * kern[6];  k7 = scalef * kern[7];  k8 = scalef * kern[8];                              \
  k9 = scalef * kern[9];  k10 = scalef * kern[10];  k11 = scalef * kern[11];                          \
  k12 = scalef * kern[12];  k13 = scalef * kern[13];  k14 = scalef * kern[14];                        \
  k15 = scalef * kern[15];  k16 = scalef * kern[16];  k17 = scalef * kern[17];                        \
  k18 = scalef * kern[18];  k19 = scalef * kern[19];  k20 = scalef * kern[20];                        \
  k21 = scalef * kern[21];  k22 = scalef * kern[22];  k23 = scalef * kern[23];                        \
  k24 = scalef * kern[24];
/*****************************************************************************************************/
#define GET_SRC_DST_PARAMETERS(type)                                                                  \
  hgt = mlib_ImageGetHeight(src);                                                                     \
  wid = mlib_ImageGetWidth(src);                                                                      \
  nchannel = mlib_ImageGetChannels(src);                                                              \
  sll = mlib_ImageGetStride(src) / sizeof(type);                                                      \
  dll = mlib_ImageGetStride(dst) / sizeof(type);                                                      \
  adr_src = (type *)mlib_ImageGetData(src);                                                           \
  adr_dst = (type *)mlib_ImageGetData(dst);
/*****************************************************************************************************/
#define INI_THREE_PIXELS_S32() sum1 = sum2 = sum3 = 0.;
#define INI_ONE_PIXEL_S32() sum1 = 0.;
/*****************************************************************************************************/
#define FIRST_CONVOLVE_THREE_PIXELS(shift, format)                                                    \
  p04 = row0[col]; p14 = row1[col]; p24 = row2[col]; col += shift;                                    \
  p05 = row0[col]; p15 = row1[col]; p25 = row2[col]; col += shift;                                    \
  p06 = row0[col]; p16 = row1[col]; p26 = row2[col]; col += shift;                                    \
                                                                                                      \
  INI_THREE_PIXELS_##format()                                                                         \
                                                                                                      \
  sum1 += p00 * k0;  sum2 += p01 * k0;  sum3 += p02 * k0;                                             \
  sum1 += p01 * k1;  sum2 += p02 * k1;  sum3 += p03 * k1;                                             \
  sum1 += p02 * k2;  sum2 += p03 * k2;  sum3 += p04 * k2;                                             \
  sum1 += p03 * k3;  sum2 += p04 * k3;  sum3 += p05 * k3;                                             \
  sum1 += p04 * k4;  sum2 += p05 * k4;  sum3 += p06 * k4;                                             \
  sum1 += p10 * k5;  sum2 += p11 * k5;  sum3 += p12 * k5;                                             \
  sum1 += p11 * k6;  sum2 += p12 * k6;  sum3 += p13 * k6;                                             \
  sum1 += p12 * k7;  sum2 += p13 * k7;  sum3 += p14 * k7;                                             \
  sum1 += p13 * k8;  sum2 += p14 * k8;  sum3 += p15 * k8;                                             \
  sum1 += p14 * k9;  sum2 += p15 * k9;  sum3 += p16 * k9;                                             \
  sum1 += p20 * k10; sum2 += p21 * k10; sum3 += p22 * k10;                                            \
  sum1 += p21 * k11; sum2 += p22 * k11; sum3 += p23 * k11;                                            \
  sum1 += p22 * k12; sum2 += p23 * k12; sum3 += p24 * k12;                                            \
  sum1 += p23 * k13; sum2 += p24 * k13; sum3 += p25 * k13;                                            \
  sum1 += p24 * k14; sum2 += p25 * k14; sum3 += p26 * k14;                                            \
                                                                                                      \
  *d_outbuf++ = sum1;                                                                                 \
  *d_outbuf++ = sum2;                                                                                 \
  *d_outbuf++ = sum3;                                                                                 \
                                                                                                      \
  p00 = p03; p01 = p04; p02 = p05; p03 = p06; /* shift 4 last row0 element */                         \
  p10 = p13; p11 = p14; p12 = p15; p13 = p16; /* shift 4 last row1 element */                         \
  p20 = p23; p21 = p24; p22 = p25; p23 = p26; /* shift 4 last row2 element */
/*****************************************************************************************************/
#define SECOND_CONVOLVE_THREE_PIXELS(dst, shift, format)                                              \
  p04 = row3[col]; p14 = row4[col]; col += shift;                                                     \
  p05 = row3[col]; p15 = row4[col]; col += shift;                                                     \
  p06 = row3[col]; p16 = row4[col]; col += shift;                                                     \
                                                                                                      \
  sum1 = *d_outbuf++;                                                                                 \
  sum2 = *d_outbuf++;                                                                                 \
  sum3 = *d_outbuf++;                                                                                 \
                                                                                                      \
  sum1 += p00 * k15;  sum2 += p01 * k15;  sum3 += p02 * k15;                                          \
  sum1 += p01 * k16;  sum2 += p02 * k16;  sum3 += p03 * k16;                                          \
  sum1 += p02 * k17;  sum2 += p03 * k17;  sum3 += p04 * k17;                                          \
  sum1 += p03 * k18;  sum2 += p04 * k18;  sum3 += p05 * k18;                                          \
  sum1 += p04 * k19;  sum2 += p05 * k19;  sum3 += p06 * k19;                                          \
  sum1 += p10 * k20;  sum2 += p11 * k20;  sum3 += p12 * k20;                                          \
  sum1 += p11 * k21;  sum2 += p12 * k21;  sum3 += p13 * k21;                                          \
  sum1 += p12 * k22;  sum2 += p13 * k22;  sum3 += p14 * k22;                                          \
  sum1 += p13 * k23;  sum2 += p14 * k23;  sum3 += p15 * k23;                                          \
  sum1 += p14 * k24;  sum2 += p15 * k24;  sum3 += p16 * k24;                                          \
                                                                                                      \
  *dst = SAT_##format(sum1); dst += shift;                                                            \
  *dst = SAT_##format(sum2); dst += shift;                                                            \
  *dst = SAT_##format(sum3); dst += shift;                                                            \
                                                                                                      \
  p00 = p03; p01 = p04; p02 = p05; p03 = p06; /* shift 4 last row3 element */                         \
  p10 = p13; p11 = p14; p12 = p15; p13 = p16; /* shift 4 last row4 element */                         \

/*****************************************************************************************************/
#define FIRST_CONVOLVE_ONE_PIXEL(shift, format)                                                       \
  p04 = row0[col];                                                                                    \
  p14 = row1[col];                                                                                    \
  p24 = row2[col]; col += shift;                                                                      \
  INI_ONE_PIXEL_##format()                                                                            \
  sum1 += p00 * k0;                         /* sum of prods */                                        \
  sum1 += p01 * k1;                                                                                   \
  sum1 += p02 * k2;                                                                                   \
  sum1 += p03 * k3;                                                                                   \
  sum1 += p04 * k4;                                                                                   \
  sum1 += p10 * k5;                                                                                   \
  sum1 += p11 * k6;                                                                                   \
  sum1 += p12 * k7;                                                                                   \
  sum1 += p13 * k8;                                                                                   \
  sum1 += p14 * k9;                                                                                   \
  sum1 += p20 * k10;                                                                                  \
  sum1 += p21 * k11;                                                                                  \
  sum1 += p22 * k12;                                                                                  \
  sum1 += p23 * k13;                                                                                  \
  sum1 += p24 * k14;                                                                                  \
  *d_outbuf++ = sum1;                                                                                 \
  p00 = p01; p01 = p02; p02 = p03; p03 = p04; /* shift 4 last row0 element */                         \
  p10 = p11; p11 = p12; p12 = p13; p13 = p14; /* shift 4 last row1 element */                         \
  p20 = p21; p21 = p22; p22 = p23; p23 = p24; /* shift 4 last row2 element */
/*****************************************************************************************************/
#define SECOND_CONVOLVE_ONE_PIXEL(dst, shift, format)                                                 \
  p04 = row3[col];                                                                                    \
  p14 = row4[col]; col += shift;                                                                      \
  sum1 = *d_outbuf++;                                                                                 \
  sum1 += p00 * k15;                        /* sum of prods */                                        \
  sum1 += p01 * k16;                                                                                  \
  sum1 += p02 * k17;                                                                                  \
  sum1 += p03 * k18;                                                                                  \
  sum1 += p04 * k19;                                                                                  \
  sum1 += p10 * k20;                                                                                  \
  sum1 += p11 * k21;                                                                                  \
  sum1 += p12 * k22;                                                                                  \
  sum1 += p13 * k23;                                                                                  \
  sum1 += p14 * k24;                                                                                  \
                                                                                                      \
  *dst = SAT_##format(sum1); dst += shift;                                                                                 \
  p00 = p01; p01 = p02; p02 = p03; p03 = p04; /* shift 4 last row3 element */                         \
  p10 = p11; p11 = p12; p12 = p13; p13 = p14; /* shift 4 last row4 element */
/*****************************************************************************************************/

mlib_status mlib_v_conv5x5_32nw( mlib_image *dst,        /* source image */
                                 mlib_image *src,        /* destination image */
                                 mlib_s32 *kern,         /* convolution kernel */
                                 mlib_s32 scalef_expon,  /* scalefactor=2**N */
                                 mlib_s32 cmask)
{
  DEFINE_IDENTIFICATORS(mlib_s32)

  scalef = 1.0;

  LOAD_KERNEL_INTO_DOUBLE()

  GET_SRC_DST_PARAMETERS(mlib_s32)

  /* allocate row bufs: */

  doutbuf = (double *)mlib_malloc(wid*sizeof(double));
  if (doutbuf == NULL) {
    return( MLIB_FAILURE );
  }

  num_blocks = (wid - 4)/3;                      /* (wid-4): no output on 2 first, last columns */
  leftover_cols = (wid - 4) - (num_blocks * 3);
  hgt -= 4;                                      /* boundaries: no output for 2 first & last rows */

  testchan = 1;
  for ( chan = nchannel-1; chan >= 0; chan--) {
    if ((cmask & testchan) == 0) {
      testchan <<= 1;
      continue;
    }
    testchan <<= 1;
    sl = adr_src + chan;
    dl = da = adr_dst + chan;
    dl = da += (dll + nchannel) * 2;

  /*
   * set up rowbufs 1, 2, 3, 4;
   * rowbuf 0 is always reloaded inside the loop:
   */

    row1 = sl;
    sl += sll;
    row2 = sl;
    sl += sll;
    row3 = sl;
    sl += sll;
    row4 = sl;
    sl += sll;

    for ( row = 0; row < hgt; row++ ) {

      row0 = row1;                               /* former row 1 is now row 0 */
      row1 = row2;                               /* former row 2 is now row 1 */
      row2 = row3;                               /* former row 3 is now row 2 */
      row3 = row4;                               /* former row 4 is now row 3 */
      row4 = sl;

      col = 0;
      d_outbuf = doutbuf;

      p00 = row0[col]; p10 = row1[col]; p20 = row2[col]; col += nchannel;
      p01 = row0[col]; p11 = row1[col]; p21 = row2[col]; col += nchannel;
      p02 = row0[col]; p12 = row1[col]; p22 = row2[col]; col += nchannel;
      p03 = row0[col]; p13 = row1[col]; p23 = row2[col]; col += nchannel;

      /* Process along row, getting 3 output points per loop iteration: */

#pragma pipeloop(0)
      for ( block = 0; block < num_blocks; block++ ) {
        FIRST_CONVOLVE_THREE_PIXELS(nchannel, S32)
      }

      if (leftover_cols) {
        FIRST_CONVOLVE_ONE_PIXEL(nchannel, S32)
        if (leftover_cols > 1) {
          FIRST_CONVOLVE_ONE_PIXEL(nchannel, S32)
        }
      }

      col = 0;
      d_outbuf = doutbuf;

      p00 = row3[col]; p10 = row4[col]; col += nchannel;
      p01 = row3[col]; p11 = row4[col]; col += nchannel;
      p02 = row3[col]; p12 = row4[col]; col += nchannel;
      p03 = row3[col]; p13 = row4[col]; col += nchannel;

      /* Process along row, getting 3 output points per loop iteration: */

#pragma pipeloop(0)
      for ( block = 0; block < num_blocks; block++ ) {
        SECOND_CONVOLVE_THREE_PIXELS(da, nchannel, S32)
      }

      if (leftover_cols) {
        SECOND_CONVOLVE_ONE_PIXEL(da, nchannel, S32)
        if (leftover_cols > 1) {
          SECOND_CONVOLVE_ONE_PIXEL(da, nchannel, S32)
        }
      }

      sl = sl + sll;                             /* set src ptrs to next row */
      dl = da = dl + dll;                        /* set dst ptrs to next row */
    }
  }

  mlib_free( doutbuf );
  return( MLIB_SUCCESS );
}

/*****************************************************************************************************/

mlib_status mlib_v_conv5x5_32ext( mlib_image *dst,        /* source image */
                                  mlib_image *src,        /* destination image */
                                  mlib_s32   dx_l,
                                  mlib_s32   dx_r,
                                  mlib_s32   dy_t,
                                  mlib_s32   dy_b,
                                  mlib_s32   *kern,         /* convolution kernel */
                                  mlib_s32   scalef_expon,  /* scalefactor=2**N */
                                  mlib_s32   cmask)
{
  DEFINE_IDENTIFICATORS(mlib_s32)
  mlib_s32 i, dend;

  scalef = 1.0;

  LOAD_KERNEL_INTO_DOUBLE()

  GET_SRC_DST_PARAMETERS(mlib_s32)

  /* allocate row bufs: */

  doutbuf = (mlib_d64 *)mlib_malloc(wid*sizeof(mlib_d64));
  if (doutbuf == NULL) {
    return( MLIB_FAILURE );
  }

  dend = (dx_r > wid) ? wid : dx_r;
  num_blocks = (wid - dend)/3;
  leftover_cols = (wid - dend) - (num_blocks * 3);

  testchan = 1;
  for ( chan = nchannel-1; chan >= 0; chan--) {
    if ((cmask & testchan) == 0) {
      testchan <<= 1;
      continue;
    }
    testchan <<= 1;
    sl = adr_src + chan;
    dl = da = adr_dst + chan;

  /*
   * set up rowbufs 1, 2, 3, 4;
   * rowbuf 0 is always reloaded inside the loop:
   */

    row = 0;
    row1 = sl; row++;
    if ((row > dy_t) && (row < hgt + TOP + BOTTOM - dy_b)) sl += sll;
    row2 = sl; row++;
    if ((row > dy_t) && (row < hgt + TOP + BOTTOM - dy_b)) sl += sll;
    row3 = sl; row++;
    if ((row > dy_t) && (row < hgt + TOP + BOTTOM - dy_b)) sl += sll;
    row4 = sl; row++;
    if ((row > dy_t) && (row < hgt + TOP + BOTTOM - dy_b)) sl += sll;

    for ( row = 0; row < hgt; row++ ) {

      row0 = row1;                               /* former row 1 is now row 0 */
      row1 = row2;                               /* former row 2 is now row 1 */
      row2 = row3;                               /* former row 3 is now row 2 */
      row3 = row4;                               /* former row 4 is now row 3 */
      row4 = sl;

      col = 0;
      d_outbuf = doutbuf;
      i = 0;

      p00 = row0[col]; p10 = row1[col]; p20 = row2[col]; i++;
      if ((i > dx_l) && (i < wid + LEFT + RIGHT - dx_r)) col += nchannel;
      p01 = row0[col]; p11 = row1[col]; p21 = row2[col]; i++;
      if ((i > dx_l) && (i < wid + LEFT + RIGHT - dx_r)) col += nchannel;
      p02 = row0[col]; p12 = row1[col]; p22 = row2[col]; i++;
      if ((i > dx_l) && (i < wid + LEFT + RIGHT - dx_r)) col += nchannel;
      p03 = row0[col]; p13 = row1[col]; p23 = row2[col]; col += nchannel;

      /* Process along row, getting 3 output points per loop iteration: */

#pragma pipeloop(0)
      for ( block = 0; block < num_blocks; block++ ) {
        FIRST_CONVOLVE_THREE_PIXELS(nchannel, S32)
      }

      if (leftover_cols) {
        FIRST_CONVOLVE_ONE_PIXEL(nchannel, S32)
        if (leftover_cols > 1) {
          FIRST_CONVOLVE_ONE_PIXEL(nchannel, S32)
        }
      }

      for (i = 0; i < dend; i++) {
        col -= nchannel;
        FIRST_CONVOLVE_ONE_PIXEL(nchannel, S32)
      }

      col = 0;
      d_outbuf = doutbuf;
      i = 0;

      p00 = row3[col]; p10 = row4[col]; i++;
      if ((i > dx_l) && (i < wid + LEFT + RIGHT - dx_r)) col += nchannel;
      p01 = row3[col]; p11 = row4[col]; i++;
      if ((i > dx_l) && (i < wid + LEFT + RIGHT - dx_r)) col += nchannel;
      p02 = row3[col]; p12 = row4[col]; i++;
      if ((i > dx_l) && (i < wid + LEFT + RIGHT - dx_r)) col += nchannel;
      p03 = row3[col]; p13 = row4[col]; p23 = row2[col]; col += nchannel;

      /* Process along row, getting 3 output points per loop iteration: */

#pragma pipeloop(0)
      for ( block = 0; block < num_blocks; block++ ) {
        SECOND_CONVOLVE_THREE_PIXELS(da, nchannel, S32)
      }

      if (leftover_cols) {
        SECOND_CONVOLVE_ONE_PIXEL(da, nchannel, S32)
        if (leftover_cols > 1) {
          SECOND_CONVOLVE_ONE_PIXEL(da, nchannel, S32)
        }
      }

      for (i = 0; i < dend; i++) {
        col -= nchannel;
        SECOND_CONVOLVE_ONE_PIXEL(da, nchannel, S32)
      }

      if (row < hgt - dy_b - 1) sl = sl + sll;   /* set src ptrs to next row */
      dl = da = dl + dll;                        /* set dst ptrs to next row */
    }
  }

  mlib_free( doutbuf );
  return( MLIB_SUCCESS );
}

/*****************************************************************************************************/

