#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <time.h>
#include <signal.h>
#include <pthread.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

/*******************************************************************
 * This file implements a workaround for the preemptive-close
 * problem on Linux (bug #4344135); A thread currently waiting in
 * a I/O operation will not wake up if one the file involved
 * is actually closed (last close - the file is no longer accessible,
 * but the thread is still waiting in the kernel).
 * This file is stand alone and doesn't require any particular
 * initialization; It just needs to be linked with the application.
 * The implementation is split in two parts; a C++ set of classes
 * that handle the files table management and a C part that
 * implements the system call wrappers.
 *******************************************************************/
 
/*
 * Missing weak symbols declaration from glibc.
 */
    extern "C" int 
__libc_recv(int, void*, size_t, int);
    extern "C" int 
__libc_recvfrom(int, void*, size_t, int, struct sockaddr*, socklen_t*);
    extern "C" int
__libc_accept(int, struct sockaddr*, socklen_t*);
    extern "C" int
__libc_sendto(int, const void*, size_t, int, const struct sockaddr*, socklen_t);
 
class FileDesc;
 
static rlim_t    fdMaxGet(void);
static rlim_t    fdMax     = fdMaxGet();
static FileDesc* fdTable   = new FileDesc[fdMax];
static int       sigWakeup = (__SIGRTMAX - 2);
 
//
// dlink - Generic link class for various list objects.
//
// In particular, it is used to avoid wasting space in the
// FileDesc structure since there are allocated in the
// heap of the process and there is one per available file.
//
class dlink {

friend class dlist;
 
     dlink* dnext;       // next list's element
     dlink* dprev;       // prev list's element
 
 public:
         int
     isEmpty() const // Return true if not in any list.
     {
         return dnext == this;
     }
         void
     reset() // Initialize.
     {
         dprev = this;
         dnext = this;
     }
         void
     append(dlink* l) // Add the element l as the "prev".
     {
         l->dprev     = dprev;
         l->dnext     = this;
         dprev->dnext = l;
         dprev        = l;
     }
         void
     insert(dlink* l) // Add the element l as the "next".
     {
         l->dprev     = this;
         l->dnext     = dnext;
         dnext->dprev = l;
         dnext        = l;
     }
         void
     del() // Remove the element (idempotent op).
     {
         dprev->dnext = dnext;
         dnext->dprev = dprev;
     }
         dlink*
     nextl() // Get the next element.
     {
         return dnext;
     }
         dlink*
     prevl() // Get the prev element.
     {
         return dprev;
     }
         dlink*
     getFirst() // remove, reset and return first element
     {
         dlink* work  = dnext;
         dnext        = work->dnext;
         dnext->dprev = this;
         work->reset();
         return work;
     }
         int
     isInList(dlink* l) const // true if the element l is not the list's head
     {
         return l != this;
     }
};
 
//
// dlist - double-linked list heads.
//
class dlist : public dlink {
 
public:
    dlist()
    {
        reset();
    }
};

//
// FileDesc - file descriptor table.
//
// There is one instance of this object allocated per process.
// It handles the file table management and implements wrappers
// for the system calls poll(), select() and close().
//
class FileDesc : public dlist {
 
    pthread_mutex_t lock;

        void
    lockGet()
    {
        int res;
        res = pthread_mutex_lock(&lock);
        assert(res == 0);
    }
        void
    lockRel()
    {
        int res;
        res = pthread_mutex_unlock(&lock);
        assert(res == 0);
    }

public:

    FileDesc()
    {
        int res;
        res = pthread_mutex_init(&lock, NULL);
        assert(res == 0);
    }
        void
    linkThread(dlink* l)
    {
        lockGet();
        insert(l);
        lockRel();
    }
        void
    unlinkThread(dlink* l)
    {
        lockGet();
        l->del();
        lockRel();
    }
        static FileDesc*
    fileGet(int fd)
    {
        if (fd < 0 || fd >= fdMax) {
            return NULL;
        }
        return &fdTable[fd];
    }
        static int
    sigGet()
    {
        return sigWakeup;
    }

        friend inline int
    Poll(struct pollfd* ufds, long unsigned int nfds, int timeout);
        friend inline int
    Select(int, fd_set*, fd_set*, fd_set*, struct timeval*);
        friend inline ssize_t
    Read(int, void*, size_t);
        friend inline int
    Recv(int, void*, size_t, int);
        friend inline int
    RecvFrom(int, void*, size_t, int, struct sockaddr*, socklen_t*);
        friend inline int
    Write(int, const void*, size_t);
        friend inline int
    Send(int, const void* msg, size_t len, int);
        friend inline int
    SendTo(int, const void*, size_t, int, const struct sockaddr*, socklen_t);
        friend inline int
    Connect(int, const struct sockaddr*, socklen_t);
        friend inline int
    Accept(int, struct sockaddr*, socklen_t*);
        friend inline int
    Close(int);
};
 
//
// ThreadLnk - thread link structure.
//
// This object implements the thread link used to
// record a thread in every file desc involved during
// a poll/select operation.
//
class ThreadLnk : public dlink {

    pthread_t thId;
    FileDesc* fDesc;

#define fDescInterrupted ((FileDesc*) -1L)

public:
        void
    init(FileDesc* fdesc, pthread_t thread)
    {
        thId  = thread;
        fDesc = fdesc;
        if (fDesc) {
            fDesc->linkThread(this);
        } else {
            reset();
        }
    }
        void
    del()
    {
        // Note that this is ok because del() is idempotent.
        if (!isEmpty()) {
            fDesc->unlinkThread(this);
        }
    }
        void
    wakeup()
    {
        fDesc = fDescInterrupted;
        pthread_kill(thId, FileDesc::sigGet());
    }
        unsigned int
    interrupted()
    {
        return fDesc == fDescInterrupted;
    }
};

//
// StackLinks - thread link table.
//
// Used to build dynamically a table of thread links
// on the stack of the calling thread during an interruptible
// IO operation.
//
class StackLinks {

    unsigned int nblnk;
    ThreadLnk*   lnks;
    int          canceltype;

public:
        void*
    operator new(size_t)
    {
        assert(0); // This is a stack object.
	return NULL;
    }

    ~StackLinks()
    {
        unsigned int intr = 0;
        for (unsigned int i = 0; i < nblnk; i++) {
            intr |= lnks[i].interrupted();
            lnks[i].del();
        }
        if (intr) {
            errno = EBADF;
        }
        pthread_setcanceltype(canceltype, NULL);
    }

    StackLinks(int fd, ThreadLnk* link)
    {
        lnks  = link;
        nblnk = 1;
        pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &canceltype);
        link->init(FileDesc::fileGet(fd), pthread_self());
    }

    StackLinks(struct pollfd*, ThreadLnk*, unsigned int);
    StackLinks(fd_set* [3], ThreadLnk*, unsigned int);
};
 
//
// StackLinks constructor - poll version
//
// Build all the thread links on the current stack
// of the caller.
//
StackLinks::StackLinks(struct pollfd* ufds,
                       ThreadLnk*     links,
                       unsigned int   nb)
{
    lnks  = links;
    nblnk = nb;

    pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &canceltype);

    for (unsigned int i = 0, thread = pthread_self(); i < nblnk; i++) {
        lnks[i].init(FileDesc::fileGet(ufds[i].fd), thread);
    }
}

//
// StackLinks constructor - select flavor.
//
// Build all the thread links on the current stack
// of the caller.
//
StackLinks::StackLinks(fd_set*      fds[3],
                       ThreadLnk*   links,
                       unsigned int nb)
{
    lnks  = links;
    nblnk = nb;

    pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &canceltype);

    for (unsigned int i = 0, thread = pthread_self(); i < nblnk; i++) {
        unsigned int used = 0;
        for (unsigned int j = 0; j < 3; j++) {
            if (fds[j] != NULL) used |= FD_ISSET(i, fds[j]);
        }
        lnks[i].init( (used)? FileDesc::fileGet(i) : NULL, thread);
    }
}

//
// Poll system call wrapper.
//
//
    inline int
Poll(struct pollfd*    ufds,
     long unsigned int nfds,
     int               timeout)
{
    //
    // Allocate the thread link table on the
    // stack of the current thread and link this
    // thread to every file desc involved in the
    // operation.
    //
    ThreadLnk  links[nfds];
    StackLinks obj = StackLinks(ufds, links, nfds);

    return __poll(ufds, nfds, timeout);
}

//
// Select system call wrapper.
//
//
    inline int
Select(int             nfds,
       fd_set*         readfds,
       fd_set*         writefds,
       fd_set*         exceptfds,
       struct timeval* timeout)
{
    fd_set* fds[3] = { readfds, writefds, exceptfds };
    //
    // Allocate the thread link table on the
    // stack of the current thread and link this
    // thread to every file desc involved in the
    // operation.
    //
    ThreadLnk  links[nfds];
    StackLinks obj = StackLinks(fds, links, nfds);

    return __select(nfds, readfds, writefds, exceptfds, timeout);
}

//
// Read system call wrapper.
//
    inline ssize_t
Read(int fd, void* buf, size_t count)
{
    ThreadLnk  lnk;
    StackLinks obj = StackLinks(fd, &lnk);

    return __read(fd, buf, count);
}

//
// Recv system call wrapper.
//
    inline int
Recv(int fd, void* buf, size_t len, int flags)
{
    ThreadLnk  lnk;
    StackLinks obj = StackLinks(fd, &lnk);

    return __libc_recv(fd, buf, len, flags);
}

//
// RecvFrom system call wrapper.
//
    inline int
RecvFrom(int              fd, 
         void*            buf, 
         size_t           len, 
         int              flags, 
         struct sockaddr* from, 
         socklen_t*       fromlen)
{
    ThreadLnk  lnk;
    StackLinks obj = StackLinks(fd, &lnk);

    return __libc_recvfrom(fd, buf, len, flags, from, fromlen);
}

//
// Write system call wrapper.
//
    inline ssize_t
Write(int fd, const void* buf, size_t count)
{
    ThreadLnk  lnk;
    StackLinks obj = StackLinks(fd, &lnk);

    return __write(fd, buf, count);
}

//
// Send system call wrapper.
//
    inline int
Send(int s, const void* msg, size_t len, int flags)
{
    ThreadLnk  lnk;
    StackLinks obj = StackLinks(s, &lnk);

    return __send(s, msg, len, flags);
}

//
// Connect system call wrapper.
//
    inline int
Connect(int s, const struct sockaddr* addr, socklen_t len)
{
    ThreadLnk  lnk;
    StackLinks obj = StackLinks(s, &lnk);

    return __connect(s, addr, len);
}

//
// Accept system call wrapper.
//
    inline int
Accept(int s, struct sockaddr* addr, socklen_t* len)
{
    ThreadLnk  lnk;
    StackLinks obj = StackLinks(s, &lnk);

    return __libc_accept(s, addr, len);
}

//
// SendTo system call wrapper.
//
    inline int
SendTo(int                    s,
       const void*            buf,
       size_t                 size,
       int                    flags,
       const struct sockaddr* addr,
       socklen_t              len)
{ 
    ThreadLnk  lnk;
    StackLinks obj = StackLinks(s, &lnk);

    return __libc_sendto(s, buf, size, flags, addr, len);
}

//
// Close system call wrapper.
//
//
    inline int
Close(int fd)
{
    FileDesc* fDesc = FileDesc::fileGet(fd);

    if (fDesc == NULL) {
        errno = EBADF;
        return -1;
    }

    fDesc->lockGet();

    int res = __close(fd);
    if (res == 0 && !fDesc->isEmpty() && dup2(fd, fd) == -1) {
        //
        // This file desc is no more accessible (last close)
        // and there is still some threads blocked in poll/select
        // using this file desc.
        //
        assert(errno == EBADF);
        do {
            ((ThreadLnk*) fDesc->getFirst())->wakeup();
        } while (!fDesc->isEmpty());
        errno = 0;
    }

    fDesc->lockRel();

    return res;
}

/*********************************************************************/

/*
 * wakeup signal handler -
 *
 * This is a dummy signal handler that forces the
 * interrupted thread to wakeup from a poll/select
 * system calls.
 */
    static void
hdlWakeup(int sig)
{
    assert(sig == FileDesc::sigGet());
}

/*
 * File table initialization -
 *
 * This function is called during the initialization
 * of the process (prior calling main) to bootstrap
 * the initialization of the FileDesc table.
 */
    static rlim_t
fdMaxGet()
{
    struct rlimit    nbr_files;
    sigset_t         sigset;
    struct sigaction sa;
    int              res;

    res = getrlimit(RLIMIT_NOFILE, &nbr_files);
    if (res != 0) {
        /*
         * Something is terribly wrong - abort.
         * An another possibility is to make the
         * glibc wrapper do nothing in this case.
         */
        abort();
    }

    sa.sa_handler = hdlWakeup;
    sa.sa_flags   = 0;
    sigemptyset(&sa.sa_mask);
    res = sigaction(FileDesc::sigGet(), &sa, NULL);
    assert(res == 0);

    sigemptyset(&sigset);
    sigaddset(&sigset, FileDesc::sigGet());
    res = sigprocmask(SIG_UNBLOCK, &sigset, NULL);
    assert(res == 0);

    return nbr_files.rlim_max;
}

/*
 * Wrapper for the poll system call -
 *
 */
    extern "C" int
poll(struct pollfd*    ufds,
     long unsigned int nfds,
     int               timeout)
{
    if (timeout == 0) {
         /* The thread won't wait - bypass the wrapper. */
        return __poll(ufds, nfds, timeout);
    }
    return Poll(ufds, nfds, timeout);
}

/*
 * Wrapper for the select system call -
 *
 */
    extern "C" int
select(int             nfds,
       fd_set*         readfds,
       fd_set*         writefds,
       fd_set*         exceptfds,
       struct timeval* tv)
{
    if (tv && tv->tv_sec == 0 && tv->tv_usec == 0) {
        /* The thread won't wait - bypass the wrapper. */
        return __select(nfds, readfds, writefds, exceptfds, tv);
    }
    return Select(nfds, readfds, writefds, exceptfds, tv);
}

/*
 * Wrapper for the read system call -
 */
    extern "C" ssize_t
read(int fd, void* buf, size_t count)
{
    return Read(fd, buf, count);
}

/*
 * Wrapper for the recv system call -
 */
    extern "C" int
recv(int s, void* buf, size_t len, int flags)
{
    return Recv(s, buf, len, flags);
}

/*
 * Wrapper for the recvfrom system call -
 */
    extern "C" int
recvfrom(int              s, 
         void*            buf, 
         size_t           len, 
         int              flags, 
         struct sockaddr* from, 
         socklen_t*       fromlen)
{
    return RecvFrom(s, buf, len, flags, from, fromlen);
}

/*
 * Wrapper for the write system call -
 */
    extern "C" ssize_t
write(int fd, const void* buf, size_t count)
{
    return Write(fd, buf, count);
}

/*
 * Wrapper for the send system call -
 */
    extern "C" int
send(int s, const void* msg, size_t len, int flags)
{
    return Send(s, msg, len, flags);
}

/*
 * Wrapper for the connect system call -
 *
 */
    extern "C" int
connect(int s, const struct sockaddr* addr, socklen_t len)
{
    return Connect(s, addr, len);
}

/*
 * Wrapper for the accept system call -
 *
 */
    extern "C" int
accept(int s, struct sockaddr* addr, socklen_t* len)
{
    return Accept(s, addr, len);
}

/*
 * Wrapper for the sendto system call -
 *
 */
    extern "C" int
sendto(int                    s, 
       const void*            buf, 
       size_t                 size, 
       int                    flags, 
       const struct sockaddr* addr, 
       socklen_t              len)
{
    return SendTo(s, buf, size, flags, addr, len);
}

/*
 * Wrapper for the close system call -
 *
 */
    extern "C" int
close(int fd)
{
    return Close(fd);
}

