/*
 * @(#)Win32GraphicsConfig.java	1.22 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt;

import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.Toolkit;
import java.awt.Transparency;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.geom.AffineTransform;
import sun.java2d.loops.RasterOutputManager;
import java.awt.Rectangle;

import java.awt.color.ColorSpace;
import java.awt.image.DataBuffer;

/**
 * This is an implementation of a GraphicsConfiguration object for a
 * single Win32 visual.
 *
 * @see GraphicsEnvironment
 * @see GraphicsDevice
 * @version 10 Feb 1997
 */
public class Win32GraphicsConfig extends GraphicsConfiguration {
    Win32GraphicsDevice screen;
    int visual;  //PixelFormatID
    ColorModel colorModel;

	private static native void initIDs();

	static {
		initIDs();
	}
	
	/**
	 * Returns a Win32GraphicsConfiguration object with the given device
	 * and PixelFormat.  Note that this method does NOT check to ensure that
	 * the returned Win32GraphicsConfig will correctly support rendering into a
	 * Java window.  This method is provided so that client code can do its
	 * own checking as to the appropriateness of a particular PixelFormat.  
	 * Safer access to Win32GraphicsConfigurations is provided by 
	 * Win32GraphicsDevice.getConfigurations().
	 */
	public static Win32GraphicsConfig getConfig(Win32GraphicsDevice device,
	 int pixFormatID) {
		return new Win32GraphicsConfig(device, pixFormatID);	
	}

    /**
     * @deprecated as of JDK version 1.3
     * replaced by <code>getConfig()</code>
     */
    public Win32GraphicsConfig(GraphicsDevice device, int visualnum) {
	this.screen = (Win32GraphicsDevice)device;
	this.visual = visualnum;
    }

    /**
     * Return the graphics device associated with this configuration.
     */
    public GraphicsDevice getDevice() {
	return screen;
    }

	/**
	 * Return the PixelFormatIndex this GraphicsConfig uses
	 */
	public int getVisual() {
	return visual;
	}

    /**
     * Returns a BufferedImage with raster and color model
     * compatible with this graphics configuration.  This method
     * has nothing to do with memory-mapping
     * a device.  This BufferedImage has
     * a raster and color model
     * that is closest to this native device configuration and thus
     * can be optimally blitted to this device.
     */
    public BufferedImage createCompatibleImage(int width, int height) {
	// REMIND: need to take into account the visual type
	ColorModel model = getColorModel();
	WritableRaster raster =
            model.createCompatibleWritableRaster(width, height);
	return new BufferedImage(model, raster, model.isAlphaPremultiplied(),
                                 null);
    }

    /**
     * Returns a BufferedImage that supports the specified transparency
     * and has a channel layout and color model
     * compatible with this graphics configuration.  This method
     * has nothing to do with memory-mapping
     * a device. This BufferedImage has a layout and
     * color model that can be optimally blitted to a device
     * with this configuration.
     * @see Transparency#OPAQUE
     * @see Transparency#BITMASK
     * @see Transparency#TRANSLUCENT
     */
    public BufferedImage createCompatibleImage(int width, int height,
					       int transparency) {
	// REMIND: need to take into account the visual type
        switch (transparency) {
        case Transparency.OPAQUE:
	    return createCompatibleImage(width, height);
            
        case Transparency.BITMASK:
        case Transparency.TRANSLUCENT:
            ColorModel cm = getColorModel(transparency);
	    WritableRaster wr = cm.createCompatibleWritableRaster(width,
                                                                  height);
            return new BufferedImage(cm, wr, cm.isAlphaPremultiplied(), null);
            
        default:
            throw new IllegalArgumentException("Unknown transparency type "+
                                               transparency);
        }
    }

    /**
     * Returns the color model associated with this configuration.
     */
    public synchronized ColorModel getColorModel() {
        if (colorModel == null)  {
	        colorModel = makeColorModel(screen.getScreen());
	    }
	    return colorModel;    
	}

    private native ColorModel makeColorModel (int screen);
	
    /**
     * Returns the color model associated with this configuration that
     * supports the specified transparency.
     */
    public ColorModel getColorModel(int transparency) {
        if (transparency == Transparency.OPAQUE) {
            return getColorModel();
        }
        else if (transparency == Transparency.BITMASK) {
            return new DirectColorModel(25, 0xff0000, 0xff00, 0xff, 0x1000000);
        }
        else if (transparency == Transparency.TRANSLUCENT) {
            return ColorModel.getRGBdefault();
        }
        else {
            return null;
        }
    }

    /**
     * Returns the default Transform for this configuration.  This
     * Transform is typically the Identity transform for most normal
     * screens.  Device coordinates for screen and printer devices will
     * have the origin in the upper left-hand corner of the target region of
     * the device, with X coordinates
     * increasing to the right and Y coordinates increasing downwards.
     * For image buffers, this Transform will be the Identity transform.
     */
    public AffineTransform getDefaultTransform() {
	return new AffineTransform();
    }

    /**
     *
     * Returns a Transform that can be composed with the default Transform
     * of a Graphics2D so that 72 units in user space will equal 1 inch
     * in device space.
     * Given a Graphics2D, g, one can reset the transformation to create
     * such a mapping by using the following pseudocode:
     * <pre>
     *      GraphicsConfiguration gc = g.getGraphicsConfiguration();
     *
     *      g.setTransform(gc.getDefaultTransform());
     *      g.transform(gc.getNormalizingTransform());
     * </pre>
     * Note that sometimes this Transform will be identity (e.g. for
     * printers or metafile output) and that this Transform is only
     * as accurate as the information supplied by the underlying system.
     * For image buffers, this Transform will be the Identity transform,
     * since there is no valid distance measurement.
     */
    public AffineTransform getNormalizingTransform() {
	double xscale = getXResolution() / 72.0;
	double yscale = getYResolution() / 72.0;
	return new AffineTransform(xscale, 0.0, 0.0, yscale, 0.0, 0.0);
    }

    private double getXResolution() {
        return 72.0;
    }
    private double getYResolution() {
        return 72.0;
    }

    public String toString() {
	return ("Win32GraphicsConfig[dev="+screen+",pixfmt="+visual+"]");
    }

    private native Rectangle getBounds(int screen);
    
    public Rectangle getBounds() {
        return getBounds(screen.getScreen());
    }
}
