/*
 * @(#)Win32GraphicsDevice.java	1.15 01/02/09
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt;

import java.awt.GraphicsDevice;
import java.awt.GraphicsConfiguration;
import java.awt.Component;
import java.lang.ref.WeakReference;
import java.util.Enumeration;
import java.util.Vector;
import java.awt.peer.ComponentPeer;

/**
 * This is an implementation of a GraphicsDevice object for a single
 * Win32 screen.
 *
 * @see GraphicsEnvironment
 * @see GraphicsConfiguration
 * @version 10 Feb 1997
 */
public class Win32GraphicsDevice extends GraphicsDevice {
    int screen;
    GraphicsConfiguration[] configs;
    GraphicsConfiguration defaultConfig;
    int defaultPixID = -1;  //Used to set defaultConfig in getConfigurations
    private Vector nonDefaultGCs;

    public Win32GraphicsDevice(int screennum) {
        this.screen = screennum;
    }

    /**
     * Returns the type of the graphics device.
     * @see #TYPE_RASTER_SCREEN
     * @see #TYPE_PRINTER
     * @see #TYPE_IMAGE_BUFFER
     */
    public int getType() {
	return TYPE_RASTER_SCREEN;
    }

    /**
     * Returns the Win32 screen of the device.
     */
    public int getScreen() {
        return screen;
    }
    
    /**
     * Returns the identification string associated with this graphics
     * device.
     */
    public String getIDstring() {
	return "\\Display"+screen;
    }


    /**
     * Returns all of the graphics
     * configurations associated with this graphics device.
     */
    public GraphicsConfiguration[] getConfigurations() {
        if (configs==null) {
            int max = getMaxConfigs(screen);
            defaultPixID = getDefaultPixID(screen);
            Vector v = new Vector( max );
            if (defaultPixID == 0) {
            //Workaround for failing GDI calls
                defaultConfig = Win32GraphicsConfig.getConfig(this,
                 defaultPixID);
                v.addElement(defaultConfig);
            }
            else {
                for (int i = 1; i <= max; i++) {
                    if (isPixFmtSupported(i, screen)) {
                        if (i == defaultPixID) {
                            defaultConfig = Win32GraphicsConfig.getConfig(
                             this, i);
                            v.addElement(defaultConfig);
                        }
                        else {
                            v.addElement(Win32GraphicsConfig.getConfig(
                             this, i));
                        }
                    }
                }
            }
            configs = new GraphicsConfiguration[v.size()];
            v.copyInto(configs);
        }
        return configs;
    }

    private static native void initIDs();

    static {
        initIDs();
    }

    /**
     * Returns the maximum number of graphics configurations available.
     * This number is less than or equal to the number of graphics
     * configurations supported.
     */
    protected native int getMaxConfigs(int screen);

    /**
     * Returns whether or not the PixelFormat indicated by index is 
     * supported.  Supported PixelFormats support drawing to a Window
     * (PFD_DRAW_TO_WINDOW), support GDI (PFD_SUPPORT_GDI), and in the
     * case of an 8-bit format (cColorBits <= 8) uses indexed colors 
     * (iPixelType == PFD_TYPE_COLORINDEX).
     * @param index a PixelFormat index
     */
    protected native boolean isPixFmtSupported(int index, int screen);

    /**
     * Returns the PixelFormatID of the default graphics configuration
     * associated with this graphics device.
     */
    protected native int getDefaultPixID(int screen);
    
    /**
     * Returns the default graphics configuration
     * associated with this graphics device.
     */
    public GraphicsConfiguration getDefaultConfiguration() {
        if (defaultConfig == null) {
            defaultPixID = getDefaultPixID(screen);
            defaultConfig = Win32GraphicsConfig.getConfig(this, defaultPixID);
        }
        return defaultConfig;
    }

    public String toString() {
        return ("Win32GraphicsDevice[screen=" + screen + "]");
    }

    /* This method is called from the Toolkit thread when Display Settings
     * are changed.  Do not execute client code here
     */
    public void displayModeChanged() {
        defaultConfig = null;
        defaultPixID = getDefaultPixID(screen);
        configs = null;    
        invalidateGCs();
    }

    public synchronized void addNonDefaultGC(WeakReference ref) {
        if (ref.get() instanceof Component) {
            if (nonDefaultGCs == null) {
                nonDefaultGCs = new Vector(1);
            }
            nonDefaultGCs.addElement(ref);
        }
    }

    private void invalidateGCs() {
        WeakReference ref;
        Object obj;
        if (nonDefaultGCs != null) {
            Enumeration gcs = nonDefaultGCs.elements();
            while (gcs.hasMoreElements()) {
                ref = (WeakReference)(gcs.nextElement());
                //Remove defunct weak references?
                obj = ref.get();
                if (obj != null && obj instanceof Component) {
                    final Component comp = (Component) obj;
                    SunToolkit.executeOnEventHandlerThread(comp,new Runnable() {
                        public void run() {
                            resetComponentGC(comp);
                        }
                    });
                    final ComponentPeer peer = comp.getPeer();
                    if (peer != null) {
                        SunToolkit.executeOnEventHandlerThread(comp, new Runnable() {
                            public void run() {
                                resetPeerGC(peer);
                            }
                        });
                    }
                }
            }
        }
    }

    /**
     * Reset the Component's GraphicsConfiguration to a default.
     * This is done as a result of a change in the user's Display Settings
     */
    private native void resetComponentGC(Component obj);

    /**
     * Reset the ComponentPeer's GraphicsConfiguration to a default.
     * This is done as a result of a change in the user's Display Settings
     */
    private native void resetPeerGC(ComponentPeer obj);
}

