/*
 * @(#)WClipboard.java	1.17 01/02/09
 *
 * Copyright 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.windows;

import java.awt.datatransfer.*;
import java.io.*;
import java.util.*;
import sun.awt.AppContext;
import sun.awt.SunClipboard;

/**
 * A class which interfaces with the Windows clipboard in order to support
 * data transfer via clipboard operations. Most of the work is provided by
 * WDataTransferer.
 *
 * @author Tom Ball
 * @author David Mendenhall
 *
 * @version 1.17, 02/09/01
 *
 * @since JDK1.1
 */
public class WClipboard extends SunClipboard {

    public static final FlavorMap flavorMap =
        SystemFlavorMap.getDefaultFlavorMap();

    public WClipboard() {
        super("System");
    }

    public synchronized void setContents(Transferable contents,
                                         ClipboardOwner owner) {
	if (this.owner != null && this.owner != owner) {
	    this.owner.lostOwnership(this, this.contents);
	}
	this.owner = owner;
	this.contents = contents;

        initContext(contents);

	// Don't use delayed Clipboard rendering for the Transferable's data.
	// If we did that, we would call Transferable.getTransferData on
	// the Toolkit thread, which is a security hole.
	//
	// Get all of the target formats into which the Transferable can be
	// translated. Then, for each format, translate the data and post
	// it to the Clipboard.
	Map formatMap = WDataTransferer.getInstance().
	    getFormatsForTransferable(contents, flavorMap);
	WClipboard.openClipboard(this);
        try {
            for (Iterator iter = formatMap.keySet().iterator();
		 iter.hasNext(); )
	    {
                Integer iFormat = (Integer)iter.next();
                int format = iFormat.intValue();
                DataFlavor flavor = (DataFlavor)formatMap.get(iFormat);
                
                try {
                    byte[] bytes = WDataTransferer.getInstance().
                        translateTransferable(contents, flavor, format);
                    publishClipboardData(format, bytes);
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        } finally {
            WClipboard.closeClipboard();
        }
    }

    public synchronized Transferable getContents(Object requestor) {
        if (contents != null) {
            return contents;
        }

	return new FullyRenderedTransferable();
    }

    private void lostSelectionOwnershipImpl() {
        lostOwnershipImpl();
    }
          
    public synchronized void lostOwnership() {
        if (this.owner != null) {
	    this.owner.lostOwnership(this, this.contents);
	    this.owner = null;
	}

	// fix 4177171 : always clear cached contents when losing ownership
	this.contents = null;
    }

    /**
     * Call the Win32 OpenClipboard function. If newOwner is non-null,
     * we also call EmptyClipboard and take ownership.
     */
    public static native void openClipboard(WClipboard newOwner);
    /**
     * Call the Win32 CloseClipboard function.
     */
    public static native void closeClipboard();
    /**
     * Call the Win32 SetClipboardData function.
     */
    private native void publishClipboardData(int format, byte[] bytes);

    private static native void init();
    static {
        init();
    }
}

/**
 * Since there is no way for client code to synchronize access to the
 * system Clipboard, we cannot provide a Transferable which creates
 * transfer data for a particular DataFlavor on demand. Instead, we have
 * to pre-render data for all possible DataFlavors.
 */
final class FullyRenderedTransferable implements Transferable {
    private final HashMap flavorsToData = new HashMap();
    private final DataFlavor[] flavors;

    static class DataFactory {
        final int format;
        final byte[] data;
        DataFactory(int format, byte[] data) {
            this.format = format;
            this.data   = data;
        }

        public Object getTransferData(DataFlavor flavor) throws IOException { 
            return WDataTransferer.getInstance().
                translateBytes(data, flavor, format);
        }
    }

    private static DataFlavor plainTextStringFlavor;
    static {
        try {
	    plainTextStringFlavor =
	        new DataFlavor("text/plain;class=java.lang.String");
	} catch (ClassNotFoundException e) {
	    // not possible
	}
    }

    public FullyRenderedTransferable() {
        WClipboard.openClipboard(null);

        try {

            int[] formats = getClipboardFormats();
            
            // Special hack to maintain backwards-compatibility with the
            // brokenness of StringSelection. This hack is now codified in
            // the javadoc for Toolkit.getSystemClipboard().
            //
            // If plain text data is available on the Clipboard, make sure
            // that we export DataFlavor.stringFlavor (representation: a
            // String) and DataFlavor.plainTextFlavor (representation: a
            // Reader; particularly, a StringReader).
            boolean addedCfUnicodeText = false;
            for (int i = 0; i < formats.length; i++) {
                if (formats[i] == WDataTransferer.CF_UNICODETEXT ||
                    (formats[i] == WDataTransferer.CF_TEXT &&
                     !addedCfUnicodeText))
                {
                    try {
                        byte[] data = getClipboardBytes(formats[i]);
                        if (data != null) {
                            Object val = WDataTransferer.getInstance().
                                translateBytes(data, 
                                               plainTextStringFlavor,
                                               formats[i]);
                            DataFactory factory = new DataFactory(formats[i],
								  data);
                            flavorsToData.put(DataFlavor.stringFlavor,
					      factory);
                            flavorsToData.put(DataFlavor.plainTextFlavor,
					      factory);
                            if (formats[i] == WDataTransferer.CF_UNICODETEXT) {
                                addedCfUnicodeText = true;
                            }
                        }
                    } catch (IOException e) {
                        flavorsToData.put(DataFlavor.stringFlavor, e);
                        flavorsToData.put(DataFlavor.plainTextFlavor, e);
                    } catch (Throwable e) {
                        e.printStackTrace();
                    }
                }
            }
            
            // Get all of the target DataFlavors into which 
            // the Clipboard's data can be translated. Then, 
            // for each flavor, translate the data and store 
            // the resulting Object. If an exception is thrown, 
            // store the exception instead and rethrow it if 
            // the transfer data is ever requested in that flavor.
            Map flavorsForFormats =
                WDataTransferer.getInstance().
                getFlavorsForFormats(formats, WClipboard.flavorMap);
            for (Iterator iter = flavorsForFormats.keySet().iterator();
                 iter.hasNext(); )
            {
                DataFlavor flavor = (DataFlavor)iter.next();
                
                if (!flavorsToData.containsKey(flavor)) {
                    int format = ((Integer)flavorsForFormats.get(flavor)).
                        intValue();
                    try {
                        byte[] data = getClipboardBytes(format);
                        DataFactory factory = new DataFactory(format, data);
                        flavorsToData.put(flavor, factory);
                    } catch (IOException e) {
                        flavorsToData.put(flavor, e);
                    } catch (Throwable e) {
                        e.printStackTrace();
                    }
                }
            }

            flavors = WDataTransferer.getInstance().
                setToSortedDataFlavorArray(flavorsToData.keySet(),
                                           flavorsForFormats);
        } finally {
            WClipboard.closeClipboard();
        }
    }

    public DataFlavor[] getTransferDataFlavors() {
	return flavors;
    }
    public boolean isDataFlavorSupported(DataFlavor flavor) {
        return flavorsToData.containsKey(flavor);
    }

    /**
     * If an exception was thrown during the data conversion, rethrow that
     * exception here.
     */
    public Object getTransferData(DataFlavor flavor)
        throws UnsupportedFlavorException, IOException
    {
        if (!isDataFlavorSupported(flavor)) {
	    throw new UnsupportedFlavorException(flavor);
	}
	Object ret = flavorsToData.get(flavor);
	if (ret instanceof IOException) {
	    throw (IOException)ret;
        } else if (ret instanceof DataFactory) {
            DataFactory factory = (DataFactory)ret;
            if (DataFlavor.stringFlavor.equals(flavor) ||
                DataFlavor.plainTextFlavor.equals(flavor)) {
                ret = factory.getTransferData(plainTextStringFlavor);
                if (DataFlavor.plainTextFlavor.equals(flavor) &&
                    ret instanceof String) {
                    ret = new StringReader((String)ret);
                }
            } else {
                ret = factory.getTransferData(flavor);
            }
        }
        return ret;
    }

    /**
     * Repeatedly call the Win32 EnumClipboardFormat function to generate
     * an integer array of all formats.
     */
    private native int[] getClipboardFormats();
    /**
     * Call the Win32 GetClipboardData function and dump the data into a
     * new byte array.
     */
    private native byte[] getClipboardBytes(int format) throws IOException;
}
