/*
 * @(#)WComponentPeer.java	1.100 01/02/14
 *
 * Copyright 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
package sun.awt.windows;

import java.awt.*;
import java.awt.peer.*;
import java.awt.image.MemoryImageSource;
import sun.awt.RepaintArea;
import sun.awt.AppContext;
import sun.awt.ScreenUpdater;
import sun.awt.UpdateClient;
import sun.awt.image.ImageRepresentation;
import sun.awt.image.OffScreenImage;
import java.awt.image.BufferedImage;
import java.awt.image.ImageProducer;
import java.awt.image.ImageObserver;
import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.awt.event.PaintEvent;
import sun.awt.Win32GraphicsConfig;

import java.awt.dnd.DropTarget;
import java.awt.dnd.peer.DropTargetPeer;

import java.security.AccessController;
import sun.security.action.GetPropertyAction;

import sun.awt.DebugHelper;

//public
abstract class WComponentPeer extends WObjectPeer implements ComponentPeer,
    UpdateClient, DropTargetPeer {

    private static final DebugHelper dbg = DebugHelper.create(WComponentPeer.class);

    // ComponentPeer implementation

    private boolean repaintPending;
    private Rectangle repaintRect;
    private RepaintArea paintArea;
    private RepaintArea updateArea;
    protected Win32GraphicsConfig winGraphicsConfig;

    /**
     * Get the GraphicsConfiguration member from the target Component.
     */
    native GraphicsConfiguration getTargetGC();

    // DropTarget support

    int nDropTargets;
    int nativeDropTargetContext;

    public synchronized native void show();
    public synchronized native void hide();
    public synchronized native void enable();
    public synchronized native void disable();

    /* New 1.1 API */
    public native Point getLocationOnScreen();

    /* New 1.1 API */
    public void setVisible(boolean b) {
	if (b) {
    	    show();
    	} else {
    	    hide();
    	}
    }

    /* New 1.1 API */
    public void setEnabled(boolean b) {
    	if (b) {
	    enable();
	} else {
	    disable();
	}
    }

    public int serialNum = 0;

    /* New 1.1 API */
    public void setBounds(int x, int y, int width, int height) {
	reshape(x, y, width, height);
	serialNum++;
    }

    /*
     * Paints any portion of the component that needs updating
     * before the call returns (similar to the Win32 API UpdateWindow)
     */
    void paintDamagedAreaImmediately() {
	// force Windows to send any pending WM_PAINT events so
	// the damage area is updated on the Java side
	updateWindow();
	// make sure paint events are transferred to main event queue
	// for coalescing
	WToolkit.getWToolkit().flushPendingEvents();
	// paint the damaged area
	paintArea.paint(target, shouldClearRectBeforePaint());
    }

    native synchronized void updateWindow();

    public void paint(Graphics g) {
        ((Component)target).paint(g);
    }

    public void repaint(long tm, int x, int y, int width, int height) {
	addRepaintArea(x, y, width, height);
	ScreenUpdater.updater.notify(this, tm);
    }

    private static final double BANDING_DIVISOR = 4.0;
    private native int[] createPrintedPixels(int srcX, int srcY,
					     int srcW, int srcH);
    public void print(Graphics g) {
        Component comp = (Component)target;

        if (g instanceof WGraphics) {
	    ((WGraphics)g).print(this);
	} else if (g instanceof WPrintGraphicsWrapper) {
	    ((WGraphics)((WPrintGraphicsWrapper)g).getTarget()).print(this);
        } else {
	    // To conserve memory usage, we will band the image.

	    int totalW = comp.getWidth();
	    int totalH = comp.getHeight();

	    int hInc = (int)(totalH / BANDING_DIVISOR);

	    for (int startY = 0; startY < totalH; startY += hInc) {
	        int endY = startY + hInc - 1;
		if (endY >= totalH) {
		    endY = totalH - 1;
		}
		int h = endY - startY + 1;

		int[] pix = createPrintedPixels(0, startY, totalW, h);
		if (pix != null) {
		    BufferedImage bim = new BufferedImage(totalW, h,
						  BufferedImage.TYPE_INT_RGB);
		    bim.setRGB(0, 0, totalW, h, pix, 0, totalW);
		    g.drawImage(bim, 0, startY, null);
		    bim.flush();
		}
	    }
	}

	comp.print(g);
    }

    public void coalescePaintEvent(PaintEvent e) {
        Rectangle r = e.getUpdateRect();
        int id = e.getID();

        switch(id) {
          case PaintEvent.PAINT:
            paintArea.add(r);

            if (dbg.on) {
                dbg.println("WCP coalescePaintEvent : PAINT : add : x = " +
                  r.x + ", y = " + r.y + ", width = " + r.width + ",height = " + r.height);
            }

            return;

          case PaintEvent.UPDATE:
            updateArea.add(r);

            if (dbg.on) {
                dbg.println("WCP coalescePaintEvent : UPDATE : add : x = " +
                  r.x + ", y = " + r.y + ", width = " + r.width + ",height = " + r.height);
            }

            return;
        }
    }

    public synchronized native void reshape(int x, int y, int width, int height);

    native void nativeHandleEvent(AWTEvent e);

    public void handleEvent(AWTEvent e) {
        int id = e.getID();

        switch(id) {
          case PaintEvent.PAINT:
            paintArea.paint(target, shouldClearRectBeforePaint());
            return;

          case PaintEvent.UPDATE:
            updateArea.update(target);
            return;

          default:
            break;
        }

        // Call the native code
        nativeHandleEvent(e);
    }

    // We must do a clearRect() before paint for most Components.
    // This is overloaded to return false for native components.
    public boolean shouldClearRectBeforePaint() {
        return true;
    }

    public Dimension getMinimumSize() {
	return ((Component)target).getSize();
    }

    public Dimension getPreferredSize() {
	return getMinimumSize();
    }

    public boolean isFocusTraversable() {
	return false;
    }
    public GraphicsConfiguration getGraphicsConfiguration() {
        if (winGraphicsConfig != null) {
            return winGraphicsConfig;
        }
        else {
            synchronized (((Component)target).getTreeLock()) {
                return ((Component)target).getGraphicsConfiguration();
            }
        }
    }

	/**
	 * Do stuff necessary for management of Components using non-default
	 * GraphicsConfigurations.
	 */
	protected void initNonDefaultGC() {
		return;
	}

    /**
     * Set winGraphicsConfig back to a default value, null for most
     * Components.  Called from the Toolkit thread, so no client code!
     */

    void resetLocalGC() {
        synchronized (((Component)target).getTreeLock()) {
            winGraphicsConfig = null;
        }
    }

    //This will return null for Components not yet added to a Container
    public ColorModel getColorModel() {
        GraphicsConfiguration gc = getGraphicsConfiguration();
        if (gc != null) {
            return gc.getColorModel();
        }
        else {
            return null;
        }
    }

    //Returns null for Components not yet added to a Container
    public ColorModel getColorModel(int transparency) {
//	return WToolkit.config.getColorModel(transparency);
        GraphicsConfiguration gc = getGraphicsConfiguration();
        if (gc != null) {
            return gc.getColorModel(transparency);
        }
        else {
            return null;
        }
    }
    public java.awt.Toolkit getToolkit() {
	return Toolkit.getDefaultToolkit();
    }
    public synchronized Graphics getGraphics() {
	if (!isDisposed()) {
	    Graphics g = new WGraphics(this);
	    g.setColor(((Component)target).getForeground());
	    g.setFont(((Component)target).getFont());
	    return g;
	}

	return null;
    }
    public FontMetrics getFontMetrics(Font font) {
	return WFontMetrics.getFontMetrics(font);
    }

    private synchronized native void _dispose();
    protected void disposeImpl() {
        ScreenUpdater.updater.removeClient(this);
        // remove from updater before calling targetDisposedPeer
	WToolkit.targetDisposedPeer(target, this);
	_dispose();
    }

    public synchronized void setForeground(Color c) {_setForeground(c.getRGB());}
    public synchronized void setBackground(Color c) {_setBackground(c.getRGB());}

    public native void _setForeground(int rgb);
    public native void _setBackground(int rgb);

    public synchronized native void setFont(Font f);
    public synchronized native void pRequestFocus();

    public synchronized void requestFocus() {
	WWindowPeer.cancelFocusOnActivate((Component)this.target);
	pRequestFocus();
    }

    public final void setCursor(Cursor c) {
        // Do nothing -- now handled by GlobalCursorManager
    }
    public Image createImage(ImageProducer producer) {
	return new WImage(producer);
    }
    public Image createImage(int width, int height) {
	ColorModel model = getColorModel(Transparency.OPAQUE);
	return new OffScreenImage((Component)target, model,
			  model.createCompatibleWritableRaster(width, height),
			  model.isAlphaPremultiplied());
    }
    public boolean prepareImage(Image img, int w, int h, ImageObserver o) {
	return WToolkit.prepareScrImage(img, w, h, o);
    }
    public int checkImage(Image img, int w, int h, ImageObserver o) {
	return WToolkit.checkScrImage(img, w, h, o);
    }

    // UpdateClient implementation

    public void updateClient(Object arg) {
       	Rectangle updateRect = null;

       	synchronized (this)
       	{
       		if (repaintPending)
       		{
       			updateRect = repaintRect;
       			repaintPending = false;
       		}
       	}
        if (updateRect != null) {
            postEvent(new PaintEvent((Component)target, PaintEvent.UPDATE,
                                     updateRect));
       	}
    }

    // Object overrides

    public String toString() {
	return getClass().getName() + "[" + target + "]";
    }

    // Toolkit & peer internals

    private int updateX1, updateY1, updateX2, updateY2;

    WComponentPeer(Component target) {
	this.target = target;
        this.repaintRect = new Rectangle();
        this.repaintPending = false;
        this.paintArea = new RepaintArea();
        this.updateArea = new RepaintArea();
	Container parent = WToolkit.getNativeContainer(target);
	WComponentPeer parentPeer = (WComponentPeer) WToolkit.targetToPeer(parent);
	create(parentPeer);
	initialize();
	start();  // Initialize enable/disable state, turn on callbacks
	initNonDefaultGC();
    }
    abstract void create(WComponentPeer parent);

    synchronized native void start();

    void initialize() {
        initZOrderPosition();

	if (((Component)target).isVisible()) {
	    show();  // the wnd starts hidden
	}
	Color fg = ((Component)target).getForeground();
	if (fg != null) {
	    setForeground(fg);
	}
        // Set background color in C++, to avoid inheriting a parent's color.
	Font  f = ((Component)target).getFont();
	if (f != null) {
	    setFont(f);
	}
	if (! ((Component)target).isEnabled()) {
	    disable();
	}
	Rectangle r = ((Component)target).getBounds();
	setBounds(r.x, r.y, r.width, r.height);
    }

    private synchronized void addRepaintArea(int x, int y, int w, int h) {
        if (repaintPending == false) {
            repaintPending = true;
            repaintRect = new Rectangle(x, y, w, h);
        } else {
            /* expand the repaint area */
            repaintRect = repaintRect.union(new Rectangle(x, y, w, h));
        }
    }

    // Callbacks for window-system events to the frame

    // Invoke a update() method call on the target
    void handleRepaint(int x, int y, int w, int h) {
        // Repaints are posted from updateClient now...
    }

    // Invoke a paint() method call on the target, after clearing the
    // damaged area.
    void handleExpose(int x, int y, int w, int h) {
        // Bug ID 4081126 & 4129709 - can't do the clearRect() here,
        // since it interferes with the java thread working in the
        // same window on multi-processor NT machines.

        postEvent(new PaintEvent((Component)target, PaintEvent.PAINT,
                                 new Rectangle(x, y, w, h)));
    }

    /* Invoke a paint() method call on the target, without clearing the
     * damaged area.  This is normally called by a native control after
     * it has painted itself. 
     *
     * NOTE: This is called on the privileged toolkit thread. Do not
     *       call directly into user code using this thread!
     */
    void handlePaint(int x, int y, int w, int h) {
        postEvent(new PaintEvent((Component)target, PaintEvent.PAINT,
                                  new Rectangle(x, y, w, h)));
    }

    /*
     * Post an event. Queue it for execution by the callback thread.
     */
    void postEvent(AWTEvent event) {
        WToolkit.postEvent(WToolkit.targetToAppContext(target), event);
    }

    // Routines to support deferred window positioning.
    public void beginValidate() {

      // Fix for oracle repaint bugs #4045627, #4045617 and #4040638
      // Removed call to clearRect that was formerly here.

        _beginValidate();
    }
    public native void _beginValidate();
    public native void endValidate();

    public void initZOrderPosition() {
        Container p = ((Component)target).getParent();
        WComponentPeer peerAbove = null;

        if (p != null) {
            Component children[] = p.getComponents();
            for (int i = 0; i < children.length; i++) {
                if (children[i] == target) {
                    break;
                } else {
                    Object cpeer = WToolkit.targetToPeer(children[i]);
                    if (cpeer != null &&
                        !(cpeer instanceof java.awt.peer.LightweightPeer)) {
                        peerAbove = (WComponentPeer)cpeer;
                    }
                }
            }

        }
        setZOrderPosition(peerAbove);
    }

    native void setZOrderPosition(WComponentPeer compAbove);


    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	return getMinimumSize();
    }

    /**
     * DEPRECATED
     */
    public Dimension preferredSize() {
	return getPreferredSize();
    }

    /**
     * register a DropTarget with this native peer
     */

    public synchronized void addDropTarget(DropTarget dt) {
	if (nDropTargets == 0) {
	    nativeDropTargetContext = addNativeDropTarget();
	}
	nDropTargets++;
    }

    /**
     * unregister a DropTarget with this native peer
     */

    public synchronized void removeDropTarget(DropTarget dt) {
	nDropTargets--;
	if (nDropTargets == 0) {
	    removeNativeDropTarget();
	    nativeDropTargetContext = 0;
	}
    }

    /**
     * add the native peer's AwtDropTarget COM object
     * @return reference to AwtDropTarget object
     */

    native int addNativeDropTarget();

    /**
     * remove the native peer's AwtDropTarget COM object
     */

    native void removeNativeDropTarget();
    
    private static final boolean isFocusFixEnabled = 
        AccessController.doPrivileged(new GetPropertyAction("sun.awt.activationFocusFix", 
	    "")).equals("true");
}
