/*
 * @(#)WDataTransferer.java	1.4 01/02/09
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.windows;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.FlavorMap;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Reader;

import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;

import java.rmi.MarshalledObject;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sun.awt.DataTransferer;

import sun.io.CharToByteConverter;

/**
 * Provides a rich set of functions for the transfer of data via the
 * Clipboard and Drag & Drop.
 * 
 * The concept of "flavors" and "natives" is extended to include "formats",
 * which are the numeric values Win32 uses to express particular data types.
 * Like FlavorMap, which provides getNativesForFlavors(DataFlavor[]) and
 * getFlavorsForNatives(String[]) functions, WDataTransferer provides a set
 * of getFormatsFor(Transferable|Flavor|Flavors) and
 * getFlavorsFor(Format|Formats) functions.
 *
 * Also provided are functions for translating a Transferable into a byte
 * array, given a source DataFlavor and a target format, and for translating
 * a byte array or InputStream into an Object, given a source format and
 * a target DataFlavor.
 *
 * @author David Mendenhall
 * @version 1.4, 02/09/01
 *
 * @since 1.3.1
 */
public class WDataTransferer extends DataTransferer {
    private static final String[] predefinedClipboardNames = {
        "",
        "TEXT",
        "BITMAP",
        "METAFILEPICT",
        "SYLK",
        "DIF",  
        "TIFF",
        "OEM TEXT",
        "DIB",
        "PALETTE",
        "PENDATA",
        "RIFF",
        "WAVE",
        "UNICODE TEXT",
        "ENHMETAFILE",
        "HDROP",
        "LOCALE"
    };

    private static final Map predefinedClipboardNameMap;
    static {
        Map tempMap = new HashMap(23);
        for (int i = 1; i < predefinedClipboardNames.length; i++) {
	    tempMap.put(predefinedClipboardNames[i], new Integer(i));
	}
	predefinedClipboardNameMap = Collections.synchronizedMap(tempMap);
    }

    /**
     * from winuser.h
     */
    public static final int CF_TEXT = 1;
    public static final int CF_UNICODETEXT = 13;
    public static final int CF_HDROP = 15;
    
    public static final Integer I_CF_TEXT = (Integer)
      predefinedClipboardNameMap.get(predefinedClipboardNames[CF_TEXT]);
    public static final Integer I_CF_UNICODETEXT = (Integer)
      predefinedClipboardNameMap.get(predefinedClipboardNames[CF_UNICODETEXT]);

    /**
     * Singleton constructor
     */
    private WDataTransferer () {
    }

    private static WDataTransferer transferer;

    public static WDataTransferer getInstance() {
	if (transferer == null) {
	    synchronized(WDataTransferer.class) {
		if (transferer == null) {
		    transferer = new WDataTransferer();
		}
	    }
	}
	return transferer;
    }

    /**
     * Returns a Map whose keys are all of the possible formats into which data
     * in the specified DataFlavors can be translated. The value of each key
     * is the DataFlavor in which the Transferable's data should be requested
     * when converting to the format.
     */
    public Map getFormatsForFlavors(DataFlavor[] flavors,
					   FlavorMap map) {
        Map nativeMap = map.getNativesForFlavors(flavors),
	    formatMap = new HashMap(flavors.length);

	// Iterate backwards so that preferred DataFlavors are used over
	// other DataFlavors. (See javadoc for
	// Transferable.getTransferDataFlavors.)
	for (int i = flavors.length - 1; i >= 0; i--) {
	    // Don't explicitly test for String, since it is just a special
	    // case of Serializable
	    if (flavors[i].isRepresentationClassInputStream() ||
		flavors[i].isRepresentationClassSerializable() ||
		flavors[i].isRepresentationClassRemote() ||
		DataFlavor.javaFileListFlavor.equals(flavors[i]))
	    {
	        // Since flavormap.properties only supports 1-to-1
	        // mappings, we need to be a bit more aggressive in
	        // exporting text/plain data. This allows StringSelection
	        // to function correctly, for example. We will probably
	        // be able to remove this code once the flavormap.
	        // properties restriction is relaxed.
	        if ("text".equals(flavors[i].getPrimaryType()) &&
		    "plain".equals(flavors[i].getSubType()))
		{
		    // There is no reason to case these exports based on
		    // OS. On NT, we have to export CF_TEXT because the
		    // Windows DnD subsystem does not implicitly convert
		    // CF_UNICODETEXT to CF_TEXT like the Clipboard does.
		    // On 95, while the Clipboard does not support
		    // CF_UNICODETEXT, a native application may support it
		    // as a DnD format.
		    formatMap.put(I_CF_UNICODETEXT, flavors[i]);
		    formatMap.put(I_CF_TEXT, flavors[i]);
		} 

		String str = (String)nativeMap.get(flavors[i]);
		if (str != null) {
		    Integer format = (Integer)
		        predefinedClipboardNameMap.get(str);
		    if (format == null) {
		        format = new Integer(registerClipboardFormat(str));
		    }
		    formatMap.put(format, flavors[i]);
		}
	    }
	}

	return formatMap;
    }

    /**
     * Returns a Map whose keys are all of the possible DataFlavors into which
     * data in the specified formats can be translated. The value of each key
     * is the format in which the Clipboard or dropped data should be requested
     * when converting to the DataFlavor.
     */
    public Map getFlavorsForFormats(int[] formats, FlavorMap map) {
        String[] natives = new String[formats.length];

	for (int i = 0; i < formats.length; i++) {
	    natives[i] = (formats[i] < predefinedClipboardNames.length)
	        ? predefinedClipboardNames[formats[i]]
	        : getClipboardFormatName(formats[i]);
	}

	Map nativeMap = map.getFlavorsForNatives(natives),
	    flavorMap = new HashMap(nativeMap.size());

	for (int i = 0; i < natives.length; i++) {
	    DataFlavor flavor = (DataFlavor)nativeMap.get(natives[i]);
	    // Don't explicitly test for String, since it is just a special
	    // case of Serializable
	    if (flavor != null &&
		(flavor.isRepresentationClassInputStream() ||
		 flavor.isRepresentationClassSerializable() ||
		 flavor.isRepresentationClassRemote() ||
		 DataFlavor.javaFileListFlavor.equals(flavor)))
	    {
	        flavorMap.put(flavor, new Integer(formats[i]));
	    }
	}

	return flavorMap;
    }

    /**
     * Attempts to discern the encoding format given the specified
     * information.
     */
    protected String getEncoding(DataFlavor flavor, int format,
				 String sourceEncoding) {
        if (format == CF_TEXT) {
	    return DataTransferer.getDefaultTextCharset();
	} else if (format == CF_UNICODETEXT) {
	    return DataTransferer.getTextCharset(
                DataFlavor.getTextPlainUnicodeFlavor());
	} else if (sourceEncoding != null) {
	    return sourceEncoding;
	} else {
	    return DataTransferer.getTextCharset(flavor);
	}
    }

    protected byte[] translateTransferable(Transferable contents,
					   DataFlavor flavor,
					   int format,
					   String sourceEncoding)
        throws IOException
    {
        // Obtain the transfer data in the source DataFlavor.
        //
        // Note that we special case DataFlavor.plainTextFlavor because 
        // StringSelection supports this flavor incorrectly -- instead of 
        // returning an InputStream as the DataFlavor representation class
        // states, it returns a Reader. Instead of using this broken
        // functionality, we request the data in stringFlavor (the other
        // DataFlavor which StringSelection supports) and use the String
        // translator.
	Object obj;
	boolean plainTextFlavorHack;
	try {
            obj = contents.getTransferData(flavor);
            if (flavor.equals(DataFlavor.plainTextFlavor) &&
                !(obj instanceof InputStream)) {

                obj = contents.getTransferData(DataFlavor.stringFlavor);
                plainTextFlavorHack = true;
            } else {
                plainTextFlavorHack = false;
            }
	} catch (UnsupportedFlavorException e) {
	    throw new IOException(e.getMessage());
	}

	// Source data is a String
	//
	// We will hit this code if the DataFlavor has a primary MIME
	// type of "text" (such as text/plain, text/enriched, text/html),
	// and representation class is java.lang.String. This is actually
	// rather uncommon. (We will also hit this code if we trigger
	// the StringSelection special case described above.)
	//
	// Note: we will not trigger this code for DataFlavor.stringFlavor
	// because stringFlavor's primary MIME type is "application".
        if (plainTextFlavorHack ||
	    (String.class.equals(flavor.getRepresentationClass()) &&
	     "text".equals(flavor.getPrimaryType())))
	{
	    String str = (String)obj;

	    // Win32ism: Convert '\n' to "\r\n"
	    StringBuffer buffer = new StringBuffer(str.length() * 2);
	    for (int i = 0; i < str.length(); i++) {
	        char c = str.charAt(i);
		if (c == '\n') {
		    buffer.append("\r\n");
		} else {
		    buffer.append(c);
		}
	    }
	    str = buffer.toString();

	    String targetEncoding =
	        getEncoding(flavor, format, sourceEncoding);
	    String targetEncodingLower = targetEncoding.toLowerCase();

	    byte[] nonTerminatedBytes = str.getBytes(targetEncoding);

	    // Win32 likes character strings to be NUL terminated. We
	    // use one NUL byte for single-byte encodings; two NUL bytes
	    // for double-byte encodings. We would theoretically use
	    // four NUL bytes for UTF-32, but the JDK currently doesn't
	    // support UTF-32.
	    int terminators;
	    if (format == CF_TEXT ||
		targetEncodingLower.equals("utf-8") ||
		targetEncodingLower.equals("unicode-1-1-utf-8")) {
	        terminators = 1;
	    } else if (format == CF_UNICODETEXT ||
		       targetEncodingLower.startsWith("unicode") ||
		       targetEncodingLower.startsWith("utf") ||
		       targetEncodingLower.startsWith("iso-10646") ||
		       targetEncodingLower.startsWith("x-utf")) {
	        terminators = 2;
	    } else {
	        terminators = 1;
	    }

	    // Append the terminating NUL bytes
	    byte[] bytes = new byte[nonTerminatedBytes.length + terminators];
	    System.arraycopy(nonTerminatedBytes, 0, bytes, 0,
			     nonTerminatedBytes.length);
	    for (int i = nonTerminatedBytes.length; i < bytes.length; i++) {
	        bytes[i] = (byte)0;
	    }

	    return bytes;
	}

	ByteArrayOutputStream bos = new ByteArrayOutputStream();

	// Target data is a Win32 file list. Source data must be a
	// java.util.List which contains java.io.File or String instances.
	if (format == CF_HDROP) {
	    if (!DataFlavor.javaFileListFlavor.equals(flavor)) {
	        throw new IOException("data translation failed");
	    }
	    final List list = (List)obj;
	    int nFiles = 0;
	    for (int i = 0; i < list.size(); i++) {
	        Object o = list.get(i);
		if (o instanceof File || o instanceof String) {
		    nFiles++;
		}
	    }
	    final String[] files = new String[nFiles];

	    try {
	        AccessController.doPrivileged(new PrivilegedExceptionAction() {
		    public Object run() throws IOException {
		        for (int i = 0, j = 0; i < list.size(); i++) {
			    Object o = list.get(i);
			    if (o instanceof File) {
			        files[j++] = ((File)o).getCanonicalPath();
			    } else if (o instanceof String) {
			        files[j++] = (String)o;
			    }
			}
			return null;
		    }
		});
	    } catch (PrivilegedActionException pae) {
	        throw new IOException(pae.getMessage());
	    }

	    for (int i = 0; i < files.length; i++) {
	        byte[] bytes = files[i].getBytes();
		bos.write(bytes, 0, bytes.length);
		bos.write(0);
	    }
	    bos.write(0);

	    // Note: DROPFILES structure not prepended here. Must implement
	    // separately in native code.

	// Source data is an InputStream
	//
	// For the general case, we just grab the bytes and dump them into
	// a byte array. For text, we need to decode the stream back to
	// characters so we can convert '\n' --> "\r\n".
        } else if (flavor.isRepresentationClassInputStream()) {
	    InputStream is = (InputStream)obj;
	    boolean eof = false;
	    int avail = is.available();
	    byte[] tmp = new byte[avail > 8192 ? avail : 8192];
	    do {
	        int ret;
		if (!(eof = (ret = is.read(tmp, 0, tmp.length)) == -1)) {
		    bos.write(tmp, 0, ret);
		}
	    } while (!eof);
            is.close();

	    if ("text".equals(flavor.getPrimaryType())) {
	        // Yuck. The text has already been encoded, but we haven't
	        // done the '\n' --> "\r\n" conversion. Decode back to a
	        // String and recur.
	        byte[] bytes = bos.toByteArray();
		bos.close();
		sourceEncoding = DataTransferer.getTextCharset(flavor);
		return translateTransferable(new StringSelection(
                    new String(bytes, sourceEncoding)),
                    DataFlavor.plainTextFlavor, format, sourceEncoding);
	    }

	// Source data is an RMI object
        } else if (flavor.isRepresentationClassRemote()) {
	    MarshalledObject mo = new MarshalledObject(obj);
	    ObjectOutputStream oos = new ObjectOutputStream(bos);
	    oos.writeObject(mo);
	    oos.close();

	// Source data is Serializable
        } else if (flavor.isRepresentationClassSerializable()) {
	    ObjectOutputStream oos = new ObjectOutputStream(bos);
	    oos.writeObject(obj);
	    oos.close();
	} else {
	    throw new IOException("data translation failed");
	}

	byte[] ret = bos.toByteArray();
	bos.close();
	return ret;
    }

    /**
     * Primary translation function for translating either a byte array or
     * an InputStream into an Object, given a source format and a target
     * DataFlavor.
     *
     * One of str/bytes is non-null; the other is null.
     * The conversion from byte[] to InputStream is cheap, so do that
     * immediately if necessary. The opposite conversion is expensive,
     * so avoid it if possible.
     */
    protected Object translateBytesOrStream(InputStream str, byte[] bytes,
					    DataFlavor flavor, int format)
        throws IOException
    {
        if (str == null) {
	    str = new ByteArrayInputStream(bytes);
	}

	// Source data is a Win32 file list. Use the dragQueryFile native
	// function to do most of the decoding. Then wrap File objects around
	// the String filenames and return a List.
        if (format == CF_HDROP) {
	    if (!DataFlavor.javaFileListFlavor.equals(flavor)) {
	        throw new IOException("data translation failed");
	    }
	    if (bytes == null) {
	        bytes = inputStreamToByteArray(str);
	    }
	    String[] filenames = dragQueryFile(bytes);
	    if (filenames == null) {
	        str.close();
	        return null;
	    }
	    File[] files = new File[filenames.length];
	    for (int i = 0; i < filenames.length; i++) {
	        files[i] = new File(filenames[i]);
	    }
	    str.close();
	    return Arrays.asList(files);
        }

	// Target representation is a String
	//
	// We will hit this code if the DataFlavor has a primary MIME
	// type of "text" (such as text/plain, text/enriched, text/html),
	// and representation class is java.lang.String. This is actually
	// rather uncommon.
	//
	// Note: we will not trigger this code for DataFlavor.stringFlavor
	// because stringFlavor's primary MIME type is "application".
	if (String.class.equals(flavor.getRepresentationClass()) &&
	    "text".equals(flavor.getPrimaryType()))
	{
	    // A String holds all of its data in memory at one time, so
	    // we can't avoid reading the entire InputStream at this point.
	    if (bytes == null) {
	        bytes = inputStreamToByteArray(str);
	    }
	    str.close();

	    String converted;
	    int count = 0;
	    int terminators;

	    String sourceEncoding = getEncoding(flavor, format, null);
	    String sourceEncodingLower = sourceEncoding.toLowerCase();

	    // Win32 likes character strings to be NUL terminated. We
	    // strip one NUL byte for single-byte encodings; two NUL bytes
	    // for double-byte encodings. We would theoretically strip
	    // four NUL bytes for UTF-32, but the JDK currently doesn't
	    // support UTF-32.
	    if (format == CF_TEXT ||
		sourceEncodingLower.equals("utf-8") ||
		sourceEncodingLower.equals("unicode-1-1-utf-8")) {
	        terminators = 1;
	    } else if (format == CF_UNICODETEXT ||
		       sourceEncodingLower.startsWith("unicode") ||
		       sourceEncodingLower.startsWith("utf") ||
		       sourceEncodingLower.startsWith("iso-10646") ||
		       sourceEncodingLower.startsWith("x-utf")) {
	        terminators = 2;
	    } else {
	        terminators = 1;
	    }

	    // Strip the terminating NUL bytes
	    if (terminators == 1) {
	        for (; count < bytes.length; count++) {
		    if (bytes[count] == (byte)0) {
		        break;
		    }
		}
		// Note that if we don't find a NUL termination, we
		// convert all of the bytes.
	    } else /* if (terminators == 2) */ {
	        for (; count < bytes.length - 1; count += 2) {
		    if (bytes[count] == (byte)0 && bytes[count + 1] == (byte)0)
		    {
		        break;
		    }
		}
		// Note that if we don't find a NUL termination, we
		// convert all of the bytes.
	    }

	    converted = new String(bytes, 0, count, sourceEncoding);

	    // Win32ism: Convert all '\r' characters
	    int length = converted.length();
	    StringBuffer buffer = new StringBuffer(length);
	    for (int i = 0; i < length; i++) {
	        char c = converted.charAt(i);
		if (c != '\r') {
		    buffer.append(c);
		}
	    }
	    return buffer.toString();
	}

	// Target flavor is an InputStream
	//
	// First, if we are converting text, we need a special InputStream
	// which will decode the stream in the source encoding, strip
	// '\r' characters and trailing NUL bytes, and then reencode in
	// the target encoding. See ReencodingInputStream for details.
	//
	// We support representations which are exactly InputStreams, and
	// also arbitrary Objects which have a constructor which takes an
	// InputStream as its sole parameter.
	if (flavor.isRepresentationClassInputStream()) {
	    if ("text".equals(flavor.getPrimaryType())) {
	        // Yuck. We need to decode the bytes so that we can strip
	        // '\r', and then reencode before we return them.
		str = new ReencodingInputStream(str,
						getEncoding(flavor, format,
							    null),
						DataTransferer.getTextCharset(flavor));
	    }

	    final Class dfrc = flavor.getRepresentationClass();

            if (InputStream.class.equals(dfrc)) {
	        return str; // simple case
            } else {
                Constructor[] constructors = null;

                try {
                    constructors = (Constructor[])
		        AccessController.doPrivileged(new PrivilegedAction() {
                            public Object run() {
                                return dfrc.getConstructors();
                            }
                        });
                } catch (SecurityException se) {
                    throw new IOException(se.getMessage());
                }

                Constructor constructor = null;

                for (int j = 0; j < constructors.length; j++) {
                    if (!Modifier.isPublic(constructors[j].getModifiers())) {
		        continue;
		    }

                    Class[] ptypes = constructors[j].getParameterTypes();

                    if (ptypes != null && ptypes.length == 1 &&
			    InputStream.class.equals(ptypes[0])) {
                        constructor = constructors[j];
                        break;
                    }
                }

                if (constructor == null) {
		    throw new IOException("can't find <init>(Ljava/io/InputStream;)V for class: " + dfrc.getName());
		}

                try {
		    return constructor.newInstance(new Object[] { str } );
                } catch (Exception e) {
                    throw new IOException(e.getMessage());
                }
	    }

	// Target flavor is an RMI object
	} else if (flavor.isRepresentationClassRemote()) {
	    try {
	        ObjectInputStream ois = new ObjectInputStream(str);
		Object ret = ((MarshalledObject)(ois.readObject())).get();
                ois.close();
		str.close();
                return ret;
	    } catch (Exception e) {
	        throw new IOException(e.getMessage());
	    }

	// Target flavor is Serializable
	} else if (flavor.isRepresentationClassSerializable()) {
	    try {
	        ObjectInputStream ois = new ObjectInputStream(str);
		Object ret = ois.readObject();
		ois.close();
                str.close();
                return ret;
	    } catch (Exception e) {
	        throw new IOException(e.getMessage());
	    }
	}
	    
	throw new IOException("data translation failed");
    }

    /**
     * Calls the Win32 RegisterClipboardFormat function to register
     * a non-standard format.
     */
    private native int registerClipboardFormat(String str);

    /**
     * Calls the Win32 GetClipboardFormatName function which is
     * the reverse operation of RegisterClipboardFormat.
     */
    private native String getClipboardFormatName(int format);
    /**
     * Decodes a byte array into a set of String filenames.
     */
    private native String[] dragQueryFile(byte[] bytes);
}

/**
 * If we are converting text, we need a special InputStream which will
 * decode the stream in the source encoding, strip '\r' characters and
 * trailing NUL bytes, and then reencode in the target encoding. We don't
 * want to just read the InputStream into a byte array and then convert
 * as if we were building a String. Win32 DnD (although not the Clipboard)
 * supports on-demand data read. A native app may provide a multiple MB
 * file as the transfer data and it would be very difficult for us to
 * load all of the data into memory.
 *
 * We use an InputStreamReader to decode the bytes, but we have to use
 * a CharToByteConverter to reencode. Public encoding APIs, such as
 * String.getBytes(String encoding), don't support incremental decoding,
 * which is a definite requirement. For example, say the target encoding
 * is "unicode" or any other Unicode encoding which begins with the
 * byte-ordering word. If we used getBytes to encode one character at a
 * time, we would return the byte-ordering word once for every character
 * in the stream.
 */
final class ReencodingInputStream extends InputStream {
    private Reader wrapped;
    private CharToByteConverter converter;
    private boolean eos;
    private byte[] cached;
    private int index, limit;

    public ReencodingInputStream(InputStream bytestream,
                                 String sourceEncoding, String targetEncoding)
        throws IOException
    {
        wrapped = new BufferedReader(
            new InputStreamReader(bytestream, sourceEncoding));
	converter = CharToByteConverter.getConverter(targetEncoding);
	cached = new byte[converter.getMaxBytesPerChar()];
    }

    public int read() throws IOException {
        if (eos) {
	    return -1;
	}

	if (index >= limit) {
	    int c;
	    do {
	        c = wrapped.read();
	    } while (((char)c) == '\r'); // Win32ism: strip '\r' characters
	    if (c == -1 || c == 0) { // -1 is EOS; 0 is a Win32 trailing NUL
	        eos = true;
		return -1;
	    }
	    limit = converter.convert(new char[] { (char)c }, 0, 1,
	                              cached, 0, cached.length);
	    index = 0;
	}

	// If we don't & with 0xFF, then the high-bit is sign-extended.
        // 255 thus becomes -1, prematurely signalling end of stream.
	return (0x000000FF & cached[index++]);
    }

    public int available() throws IOException {
        return ((eos) ? 0 : (limit - index));
    }

    public void close() throws IOException {
        wrapped.close();
    }
}
