/*
 * @(#)WDropTargetContextPeer.java	1.26 01/02/09
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.windows;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;

import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.InvalidDnDOperationException;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.Map;

import sun.awt.dnd.SunDropTargetContextPeer;

/**
 * <p>
 * The WDropTargetContextPeer class is the class responsible for handling
 * the interaction between the win32 DnD system and Java.
 * </p>
 *
 * @version 1.26, 02/09/01
 *
 */

final class WDropTargetContextPeer extends SunDropTargetContextPeer {

    /**
     * create the peer typically upcall from C++
     */

    static WDropTargetContextPeer getWDropTargetContextPeer() {
	return new WDropTargetContextPeer();
    }

    /**
     * create the peer
     */

    private WDropTargetContextPeer() {
	super();
    }

    /**
     * upcall to encapsulate file transfer
     */

    private static FileInputStream getFileStream(String file, int stgmedium)
        throws IOException
    {
	return new WDropTargetContextPeerFileStream(file, stgmedium);
    }

    /**
     * upcall to encapsulate IStream buffer transfer
     */

    private static Object getIStream(int istream) throws IOException {
	return new WDropTargetContextPeerIStream(istream);
    }

    /**
     * diff on lock
     * @return the DropTarget associated with this peer
     */

    public DropTarget getDropTarget() { synchronized(lock) { return currentDT; } }

    /**
     * @return current DataFlavors available
     */

    public DataFlavor[] getTransferDataFlavors() {
	DataFlavor[]  dfs;

	if (local != null) {  
	    return local.getTransferDataFlavors();
	} else {
	    synchronized(lock) {
	        return WDataTransferer.getInstance().
		    getFlavorsForFormatsAsArray(currentT,
						currentDT.getFlavorMap());
	    }
	}
    }

    /**
     * @return if the Flavor is supported
     */

    public boolean isDataFlavorSupported(DataFlavor df) {
	if (local != null) {
	    return local.isDataFlavorSupported(df);
	}

	synchronized(lock) {
	    return WDataTransferer.getInstance().
	        getFlavorsForFormats(currentT, currentDT.getFlavorMap()).
	        containsKey(df);
	}
    }

    /**
     * @return the data 
     */
    public Object getTransferData(DataFlavor df) throws UnsupportedFlavorException, IOException, InvalidDnDOperationException {

	if (local != null) {
	    return local.getTransferData(df);
	}
	
	Map flavorMap;
	
	synchronized(lock) {
	    if (dropStatus != STATUS_ACCEPT || dropComplete) {
		throw new InvalidDnDOperationException("No drop current");
	    }

	    flavorMap = WDataTransferer.getInstance().
	        getFlavorsForFormats(currentT, currentDT.getFlavorMap());
	}

	Integer iFormat = (Integer)flavorMap.get(df);
	if (iFormat == null) {
	    throw new UnsupportedFlavorException(df);
	}
	int format = iFormat.intValue();

	if (df.isRepresentationClassRemote() &&
	        currentDA != DnDConstants.ACTION_LINK) {
	    throw new InvalidDnDOperationException("only ACTION_LINK is permissable for transfer of java.rmi.Remote objects");
	}

	Object ret = getData(nativeDragContext, format);

	if (ret instanceof byte[]) {
	    try {
	        return WDataTransferer.getInstance().
		    translateBytes((byte[])ret, df, format);
	    } catch (IOException e) {
	        throw new InvalidDnDOperationException(e.getMessage());
	    }
	} else if (ret instanceof InputStream) {
	    try {
	        return WDataTransferer.getInstance().
		    translateStream((InputStream)ret, df, format);
	    } catch (IOException e) {
	        throw new InvalidDnDOperationException(e.getMessage());
	    }
	} else {
	    throw new IOException("no native data was transfered");
	}
    }

    public void dropComplete(boolean success) {
	synchronized(lock) {
	    if (dropStatus == STATUS_NONE) {
		throw new InvalidDnDOperationException("No Drop pending");
	    }

	    if (currentDTC != null) currentDTC.removeNotify();

	    currentC   = null;
	    currentDT  = null;
	    currentDTC = null;
	    currentT   = null;
	    currentA   = DnDConstants.ACTION_NONE;

	    synchronized(_globalLock) {
	        currentJVMLocalSourceTransferable = null;
	    }

	    dropStatus   = STATUS_NONE;
	    dropComplete = true;
	}

	dropDone(nativeDragContext, success ? currentDA : DnDConstants.ACTION_NONE);

	currentDA = DnDConstants.ACTION_NONE;

	nativeDragContext = 0;
    }

 
    /**
     * downcall to bind transfer data.
     */

     private native Object getData(long nativeContext, int format);

    /**
     * downcall to notify that drop is complete
     */

     private native void dropDone(long nativeContext, int actions);

   /**
     * native code initialization
     */

    static native void Initialize();

    /**
     * Allows extensions of this class to 
     * perform additional operations in beginning a dispatch by overriding 
     * this method.
     *
     * For instance WDragSourceContextPeerDispatcher will override
     * this method by adding a call to WToolkit.startSecondaryEventLoop();
     */
    protected void startSecondaryDispatch() {
	WToolkit.startSecondaryEventLoop();
    }
    
    /**
     * called by run. Allows extensions of this class to 
     * perform additional operations in ending a dispatch by overriding 
     * this method.
     *
     * For instance WDragSourceContextPeerDispatcher will override
     * this method by adding a call to WToolkit.quitSecondaryEventLoop();
     */
    protected void quitSecondaryDispatch() {
	WToolkit.quitSecondaryEventLoop();
    }
    
}

/**
 * package private class to handle file transfers
 */

class WDropTargetContextPeerFileStream extends FileInputStream {

    /**
     * construct file input stream
     */

    WDropTargetContextPeerFileStream(String name, int stgmedium)
        throws FileNotFoundException
    {
	super(name);

	this.stgmedium  = stgmedium;
    }

    /**
     * close
     */

    public void close() throws IOException {
        if (stgmedium != 0) {
            super.close();
            freeStgMedium(stgmedium);
            stgmedium = 0;
        }
    }

    /**
     * free underlying windows data structure
     */

    private native void freeStgMedium(int stgmedium);

    /*
     * fields
     */

    private int    stgmedium;
}

/**
 * Package private class to access IStream objects
 */

class WDropTargetContextPeerIStream extends InputStream {

    /**
     * construct a WDropTargetContextPeerIStream wrapper
     */

    WDropTargetContextPeerIStream(int istream) throws IOException {
	super();

	if (istream == 0) throw new IOException("No IStream");

	this.istream    = istream;
    }

    /**
     * @return bytes available
     */

    public int available() throws IOException {
	if (istream == 0) throw new IOException("No IStream");
	
	return Available(istream);
    }

    private native int Available(int istream);

    /**
     * read
     */

    public int read() throws IOException {
	if (istream == 0) throw new IOException("No IStream");
	return Read(istream);
    }

    private native int Read(int istream) throws IOException;

    /**
     * read into buffer
     */

    public int read(byte[] b, int off, int len) throws IOException {
	if (istream == 0) throw new IOException("No IStream");
	return ReadBytes(istream, b, off, len);
    }

    private native int ReadBytes(int istream, byte[] b, int off, int len) throws IOException;

    /**
     * close
     */

    public void close() throws IOException {
        if (istream != 0) {
            super.close();
            Close(istream);
            istream = 0;
        }
    }

    private native void Close(int istream);

    /*
     * fields
     */

    private int	   istream;
}
