/*
 * @(#)WScrollPanePeer.java	1.22 00/02/02
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
package sun.awt.windows;

import sun.awt.DebugHelper;
import sun.awt.PeerEvent;
import java.awt.*;
import java.awt.peer.*;

class WScrollPanePeer extends WPanelPeer implements ScrollPanePeer {
    private static final DebugHelper dbg = DebugHelper.create(WScrollPanePeer.class);

    static final int SCROLL_UP = 0;
    static final int SCROLL_DOWN = 1;
    static final int SCROLL_BLOCK_UP = 2;
    static final int SCROLL_BLOCK_DOWN = 3;
    static final int SCROLL_ABSOLUTE = 4;
    static final int ORIENT_VERT = 0;
    static final int ORIENT_HORZ = 1;

    int scrollbarWidth;
    int scrollbarHeight;
    int prevx;
    int prevy;

    static {
	initIDs();
    }

    static native void initIDs();
    native void create(WComponentPeer parent);	    
    native int getOffset(int orient);

    WScrollPanePeer(Component target) {
	super(target);
        scrollbarWidth = _getVScrollbarWidth();
        scrollbarHeight = _getHScrollbarHeight();
    }

    void initialize() {
        super.initialize();
        setInsets();
	Insets i = getInsets();
	setScrollPosition(-i.left,-i.top);
    }

    public void setUnitIncrement(Adjustable adj, int p) {
        // The unitIncrement is grabbed from the target as needed.
    }

    public Insets insets() {
        return getInsets();
    }
    private native void setInsets();

    public native synchronized void setScrollPosition(int x, int y);

    public int getHScrollbarHeight() {
        return scrollbarHeight;
    }
    private native int _getHScrollbarHeight();

    public int getVScrollbarWidth() {
        return scrollbarWidth;
    }
    private native int _getVScrollbarWidth();

    public Point getScrollOffset() {
	int x, y;
	x = getOffset(Scrollbar.HORIZONTAL);
	y = getOffset(Scrollbar.VERTICAL);
	return new Point(x, y);
    }

    /**
     * The child component has been resized.  The scrollbars must be
     * updated with the new sizes.  At the native level the sizes of
     * the actual windows may not have changed yet, so the size
     * information from the java-level is passed down and used.
     */
    public void childResized(int width, int height) {
	ScrollPane sp = (ScrollPane)target;
	Dimension vs = sp.getSize();
	setSpans(vs.width, vs.height, width, height);
        setInsets();
    }

    native synchronized void setSpans(int viewWidth, int viewHeight, 
                                      int childWidth, int childHeight);

    /**
     * Called by ScrollPane's internal observer of the scrollpane's adjustables.
     * This is called whenever a scroll position is changed in one
     * of adjustables, whether it was modified externally or from the
     * native scrollbars themselves.  
     */
    public void setValue(Adjustable adj, int v) {
	Component c = getScrollChild();
	Point p = c.getLocation();
	switch(adj.getOrientation()) {
	case Adjustable.VERTICAL:
	    setScrollPosition(-(p.x), v);
	    break;
	case Adjustable.HORIZONTAL:
	    setScrollPosition(v, -(p.y));
	    break;
	}
    }
	    
    private native Component getScrollChild();

    /*
     * Called from Windows in response to WM_VSCROLL/WM_HSCROLL message
     */
    private void postScrollEvent(final int orient, final int scrollType, final int pos) {
	Runnable run = new ScrollEventRunnable(orient, scrollType, pos);
	WToolkit.executeOnEventHandlerThread(new ScrollEvent(target, run));
    }

    /*
     * Event that executes on the Java dispatch thread to move the scroll bar thumbs
     * and paint the exposed area in one synchronous operation.
     */
    class ScrollEvent extends PeerEvent {
	ScrollEvent(Object source, Runnable runnable) {
	    super(source, runnable, 0L);
	}

	public PeerEvent coalesceEvents(PeerEvent newEvent) {
	    if (dbg.on) dbg.println("ScrollEvent coalesced "+newEvent);
	    if (newEvent instanceof ScrollEvent) {
		return newEvent;
	    }
	    return null;
	}
    }

    class ScrollEventRunnable implements Runnable {
	int orient, scrollType, pos;

	ScrollEventRunnable(int orient, int scrollType, int pos) {
	    this.orient = orient;
	    this.scrollType = scrollType;
	    this.pos = pos;
	}

	public void run() {
	    Adjustable adj = null;
	    ScrollPane sp = (ScrollPane)WScrollPanePeer.this.target;

	    if (orient == ORIENT_VERT) {
		adj = sp.getVAdjustable();
	    } else if (orient == ORIENT_HORZ) {
		adj = sp.getHAdjustable();
	    } else {
		if (dbg.on) dbg.assert(false);
	    }

	    if (adj == null) {
		return;
	    }

	    int		unitInc = adj.getUnitIncrement();
	    int		blockInc = adj.getBlockIncrement();
	    int		newpos = adj.getValue();

	    switch (scrollType) {
		case SCROLL_UP:
		    newpos -= unitInc;
		    break;
		case SCROLL_DOWN:
		    newpos += unitInc;
		    break;
		case SCROLL_BLOCK_UP:
		    newpos -= blockInc;
		    break;
		case SCROLL_BLOCK_DOWN:
		    newpos += blockInc;
		    break;
		case SCROLL_ABSOLUTE:
		    newpos = this.pos;
		    break;
		default:
		    if (dbg.on) {
			dbg.assert(false);
		    }
		    break;
	    }

	    // keep scroll position in acceptable range
	    newpos = Math.max(adj.getMinimum(), newpos);
	    newpos = Math.min(adj.getMaximum(), newpos);

	    adj.setValue(newpos);
	    
	    //
	    // paint the exposed area right away
	    //
	    Component c = getScrollChild(); 
	    // find the heavyweight ancestor of the scroll child
	    Component hwAncestor = c;
	    while ( !(hwAncestor.getPeer() instanceof WComponentPeer) ) {
		hwAncestor = hwAncestor.getParent();
	    }
	    if (dbg.on) {
		dbg.assert(hwAncestor != null, "couldn't find hw ancestor of scroll pane child");
	    }
	    WComponentPeer wpeer = (WComponentPeer)hwAncestor.getPeer();
	    wpeer.paintDamagedAreaImmediately();
	}
    }
}
