/*
 * @(#)WToolkit.java	1.104 01/02/09
 *
 * Copyright 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package sun.awt.windows;

import java.awt.*;
import java.awt.im.InputMethodHighlight;
import java.awt.im.spi.InputMethodDescriptor;
import java.awt.image.*;
import java.awt.peer.*;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.datatransfer.Clipboard;
import java.io.*;
import java.lang.reflect.*;
import java.net.URL;
import java.util.Hashtable;
import java.util.Map;
import java.util.Properties;
import java.util.Vector;
import java.util.Enumeration;
import java.awt.print.PageFormat;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.security.AccessController;
import java.security.PrivilegedAction;
import sun.awt.image.ByteArrayImageSource;
import sun.awt.image.FileImageSource;
import sun.awt.image.URLImageSource;
import sun.awt.image.ImageRepresentation;
import sun.awt.print.PrintControl;
import sun.awt.AppContext;
import sun.awt.EmbeddedFrame;
import sun.awt.GlobalCursorManager;
import sun.awt.SunToolkit;
import sun.awt.Win32GraphicsDevice;

import sun.java2d.loops.RasterOutputManager;


import java.awt.dnd.DragSource;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureRecognizer;
import java.awt.dnd.MouseDragGestureRecognizer;
import java.awt.dnd.InvalidDnDOperationException;
import java.awt.dnd.peer.DragSourceContextPeer;
 
import sun.awt.windows.WDragSourceContextPeer;

public class WToolkit extends SunToolkit implements Runnable {

    static GraphicsConfiguration config;

    // System clipboard.
    WClipboard clipboard;

    // cache of font peers
    private Hashtable cacheFontPeer;

    // Windows properties
    private WDesktopProperties	wprops;

    /**
     * Initialize JNI field and method IDs
     */
    private static native void initIDs();

    static {
	resetGC();
	java.security.AccessController.doPrivileged(
		    new sun.security.action.LoadLibraryAction("awt"));
	initIDs();
    }

    /* This method is called from the Toolkit thread when Display Settings
     * are changed.  Do not execute client code here.
     */
    private static void resetGC() {
        config = (GraphicsEnvironment.
		  getLocalGraphicsEnvironment().
		  getDefaultScreenDevice().
		  getDefaultConfiguration());
    }

    /*
     * New hook functions for using the AWT as an embedded service. These
     * functions replace the global C function AwtInit() which was previously
     * exported by awt.dll.
     *
     * When used as an embedded service, the AWT does NOT have its own
     * message pump. It instead relies on the parent application to provide
     * this functionality. embeddedInit() assumes that the thread on which it
     * is called is the message pumping thread. Violating this assumption
     * will lead to undefined behavior.
     *
     * embeddedInit must be called before the WToolkit() constructor.
     * embeddedDispose should be called before the applicaton terminates the
     * Java VM. It is currently unsafe to reinitialize the toolkit again
     * after it has been disposed. Instead, awt.dll must be reloaded and the
     * class loader which loaded WToolkit must be finalized before it is
     * safe to reuse AWT. Dynamic reusability may be added to the toolkit in
     * the future.
     */

    /**
     * Initializes the Toolkit for use in an embedded environment.
     *
     * @return true if the the initialization succeeded; false if it failed.
     *         The function will fail if the Toolkit was already initialized.
     * @since 1.3
     */
    public static native boolean embeddedInit();
    /**
     * Disposes the Toolkit in an embedded environment. This method should
     * not be called on exit unless the Toolkit was constructed with
     * embeddedInit.
     *
     * @return true if the disposal succeeded; false if it failed. The
     *         function will fail if the calling thread is not the same
     *         thread which called embeddedInit(), or if the Toolkit was
     *         already disposed.
     * @since 1.3
     */
    public static native boolean embeddedDispose();

    public WToolkit() {
        // Startup toolkit threads
        synchronized (this) {
	    // Fix for bug #4046430 -- Race condition
	    // where notifyAll can be called before
	    // the "AWT-Windows" thread's parent thread is 
	    // waiting, resulting in a deadlock on startup.
            Thread thread = new Thread(this, "AWT-Windows");
            thread.setPriority(Thread.NORM_PRIORITY + 1);
            thread.start();
                                                  
	    try {
	        wait();
	    }
	    catch (InterruptedException x) {
	    }
        }

	WDropTargetContextPeer.Initialize();
	WDragSourceContextPeer.Initialize();
    }

    public void run() {
        boolean startPump = init();

	if (startPump) {
	    ThreadGroup mainTG = (ThreadGroup)AccessController.doPrivileged(
	        new PrivilegedAction() {
	            public Object run() {
		        ThreadGroup currentTG =
		            Thread.currentThread().getThreadGroup();
			ThreadGroup parentTG = currentTG.getParent();
			while (parentTG != null) {
			    currentTG = parentTG;
			    parentTG = currentTG.getParent();
			}
			return currentTG;
		    }
	    });

	    Runtime.getRuntime().addShutdownHook(
	        new Thread(mainTG, new Runnable() {
	            public void run() {
		        shutdown();
		    }
		})
	    );
	}

        synchronized(this) {
            notifyAll();
        }

	if (startPump) {
	    eventLoop(); // will Dispose Toolkit when shutdown hook executes
	}
    }

    /* 
     * eventLoop() begins the native message pump which retrieves and processes
     * native events.
     *
     * When shutdown() is called by the ShutdownHook added in run(), a
     * WM_QUIT message is posted to the Toolkit thread indicating that
     * eventLoop() should Dispose the toolkit and exit.
     */
    private native boolean init();
    private native void eventLoop();
    private native void shutdown();
    protected native void finalize(); // only called if runFinalizersOnExit

    /*
     * Instead of blocking the "AWT-Windows" thread uselessly on a semaphore,
     * use these functions. startSecondaryEventLoop() corresponds to wait()
     * and quitSecondaryEventLoop() corresponds to notify.
     *
     * These functions simulate blocking while allowing the AWT to continue
     * processing native events, eliminating a potential deadlock situation
     * with SendMessage.
     *
     * WARNING: startSecondaryEventLoop must only be called from the "AWT-
     * Windows" thread.
     */
    public static native void startSecondaryEventLoop();
    public static native void quitSecondaryEventLoop();

    /*
     * Create peer objects.
     */

    public ButtonPeer createButton(Button target) {
	ButtonPeer peer = new WButtonPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public TextFieldPeer createTextField(TextField target) {
	TextFieldPeer peer = new WTextFieldPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public LabelPeer createLabel(Label target) {
	LabelPeer peer = new WLabelPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public ListPeer createList(List target) {
	ListPeer peer = new WListPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public CheckboxPeer createCheckbox(Checkbox target) {
	CheckboxPeer peer = new WCheckboxPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public ScrollbarPeer createScrollbar(Scrollbar target) {
	ScrollbarPeer peer = new WScrollbarPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public ScrollPanePeer createScrollPane(ScrollPane target) {
	ScrollPanePeer peer = new WScrollPanePeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public TextAreaPeer createTextArea(TextArea target) {
	TextAreaPeer peer = new WTextAreaPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public ChoicePeer createChoice(Choice target) {
	ChoicePeer peer = new WChoicePeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public FramePeer  createFrame(Frame target) {
	FramePeer peer = new WFramePeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public CanvasPeer createCanvas(Canvas target) {
	CanvasPeer peer = new WCanvasPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public PanelPeer createPanel(Panel target) {
	PanelPeer peer = new WPanelPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public WindowPeer createWindow(Window target) {
	WindowPeer peer = new WWindowPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public DialogPeer createDialog(Dialog target) {
	DialogPeer peer = new WDialogPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public FileDialogPeer createFileDialog(FileDialog target) {
	FileDialogPeer peer = new WFileDialogPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public MenuBarPeer createMenuBar(MenuBar target) {
	MenuBarPeer peer = new WMenuBarPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public MenuPeer createMenu(Menu target) {
	MenuPeer peer = new WMenuPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public PopupMenuPeer createPopupMenu(PopupMenu target) {
	PopupMenuPeer peer = new WPopupMenuPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public MenuItemPeer createMenuItem(MenuItem target) {
	MenuItemPeer peer = new WMenuItemPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public CheckboxMenuItemPeer createCheckboxMenuItem(CheckboxMenuItem target) {
	CheckboxMenuItemPeer peer = new WCheckboxMenuItemPeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    public RobotPeer createRobot(Robot target, GraphicsDevice screen) {
	// (target is unused for now)
	// Robot's don't need to go in the peer map since
	// they're not Component's
        return new WRobotPeer(screen);
    }

    public WEmbeddedFramePeer createEmbeddedFrame(WEmbeddedFrame target) {
	WEmbeddedFramePeer peer = new WEmbeddedFramePeer(target);
	targetCreatedPeer(target, peer);
	return peer;
    }

    WPrintDialogPeer createWPrintDialog(WPrintDialog target) {
        WPrintDialogPeer peer = new WPrintDialogPeer(target);
        targetCreatedPeer(target, peer);
        return peer;
    }

    static boolean prepareScrImage(Image img, int w, int h, ImageObserver o) {
	if (w == 0 || h == 0) {
	    return true;
	}
	
	// Must be an OffScreenImage
	if (!(img instanceof WImage)) {
	   return true;
	}

	WImage ximg = (WImage) img;
	if (ximg.hasError()) {
	    if (o != null) {
		o.imageUpdate(img, ImageObserver.ERROR|ImageObserver.ABORT,
			      -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = ximg.getImageRep();
	return ir.prepare(o);
    }

    static int checkScrImage(Image img, int w, int h, ImageObserver o) {
        if (!(img instanceof WImage)) {
            return ImageObserver.ALLBITS;
        }
	WImage ximg = (WImage) img;
	int repbits;
	if (w == 0 || h == 0) {
	    repbits = ImageObserver.ALLBITS;
	}
        else {
	    repbits = ximg.getImageRep().check(o);
	}
	return ximg.check(o) | repbits;
    }

    public int checkImage(Image img, int w, int h, ImageObserver o) {
	return checkScrImage(img, w, h, o);
    }

    public boolean prepareImage(Image img, int w, int h, ImageObserver o) {
	return prepareScrImage(img, w, h, o);
    }

    public Image createImage(ImageProducer producer) {
	return new WImage(producer);
    }

    static native ColorModel makeColorModel();
    static ColorModel screenmodel;

    static ColorModel getStaticColorModel() {
        return config.getColorModel();
    }

    public ColorModel getColorModel() {
	return getStaticColorModel();
    }

    public native int getScreenResolution();
    protected native int getScreenWidth();
    protected native int getScreenHeight();

    public FontMetrics getFontMetrics(Font font) {
        if (!RasterOutputManager.usesPlatformFont()) {
            return super.getFontMetrics(font);
        } else {
            return WFontMetrics.getFontMetrics(font);
        }
    }
    
    public FontPeer getFontPeer(String name, int style) {
        FontPeer retval = null;
        String lcName = name.toLowerCase();
        if (null != cacheFontPeer) {
            retval = (FontPeer)cacheFontPeer.get(lcName + style);
            if (null != retval) {
                return retval;
            }
        }
	/*----------*
        if (!RasterOutputManager.usesPlatformFont()) {
	    return new FontPeer() {};
        } else {
            return new WFontPeer(name, style);
	}
	*----------*/
        retval = new WFontPeer(name, style);
        if (retval != null) {
            if (null == cacheFontPeer) {
                cacheFontPeer = new Hashtable(5, (float)0.9);
            }
            if (null != cacheFontPeer) {
                cacheFontPeer.put(lcName + style, retval);
            }
        }
        return retval;
    }

    public native void sync();

    public PrintJob getPrintJob(Frame frame, String doctitle,
				Properties props) {
        return getPrintJob(frame, doctitle, null, null);
    }

    private Constructor printControlConstructor;

    private class GetConstructorAction implements 
        java.security.PrivilegedAction
    {
        private Class[] args;
        public GetConstructorAction(Class[] args) {
	    this.args = args;
	}
        public Object run() {
	    Constructor constructor = null;
	    try {
	        String nm = System.getProperty("awt.printcontrol");
		if (nm != null) {
		    Class clazz = Class.forName(nm, true,
		        ClassLoader.getSystemClassLoader());
		    constructor = clazz.getConstructor(args);
		} else {
		    constructor = WPrintControl.class.getConstructor(args);
		}
	    } catch (ClassNotFoundException e) {
	        try {
		    constructor = WPrintControl.class.getConstructor(args);
		} catch (NoSuchMethodException ee) {
		    throw new InternalError();
		}
	    } catch (NoSuchMethodException e) {
	        try {
		    constructor = WPrintControl.class.getConstructor(args);
		} catch (NoSuchMethodException ee) {
		    throw new InternalError();
		}
	    }
	    return constructor;
	}
    }

    public PrintJob getPrintJob(Frame frame, String doctitle,
				JobAttributes jobAttributes,
				PageAttributes pageAttributes) {
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkPrintJobAccess();
	}
	if (printControlConstructor == null) {
	    printControlConstructor = (Constructor)
	        java.security.AccessController.doPrivileged(
		    new GetConstructorAction(new Class[] { Frame.class,
							   String.class,
							   JobAttributes.class,
							   PageAttributes.class
							 }));
	}			
	PrintControl printControl = null;
	try {
	    printControl = (PrintControl)
	        printControlConstructor.newInstance(
                    new Object[] { frame, doctitle, jobAttributes,
			           pageAttributes });
	} catch (InstantiationException e) {
	    printControl = new WPrintControl(frame, doctitle, jobAttributes,
					     pageAttributes);
	} catch (IllegalAccessException e) {
	    printControl = new WPrintControl(frame, doctitle, jobAttributes,
					     pageAttributes);
	} catch (InvocationTargetException e) {
	    Throwable xcp = e.getTargetException();
	    if (xcp instanceof Error) {
	        throw (Error)xcp;
	    } else if (xcp instanceof RuntimeException) {
	        throw (RuntimeException)xcp;
	    } else {
	        printControl = new WPrintControl(frame, doctitle,
						 jobAttributes,
						 pageAttributes);
	    }
	}
	return printControl.initJob();
    }

    public native void beep();

    public boolean getLockingKeyState(int key) {
        if (! (key == KeyEvent.VK_CAPS_LOCK || key == KeyEvent.VK_NUM_LOCK ||
               key == KeyEvent.VK_SCROLL_LOCK || key == KeyEvent.VK_KANA_LOCK)) {
            throw new IllegalArgumentException("invalid key for Toolkit.getLockingKeyState");
        }
        return getLockingKeyStateNative(key);
    }

    public native boolean getLockingKeyStateNative(int key);

    public void setLockingKeyState(int key, boolean on) {
        if (! (key == KeyEvent.VK_CAPS_LOCK || key == KeyEvent.VK_NUM_LOCK ||
               key == KeyEvent.VK_SCROLL_LOCK || key == KeyEvent.VK_KANA_LOCK)) {
            throw new IllegalArgumentException("invalid key for Toolkit.setLockingKeyState");
        }
        setLockingKeyStateNative(key, on);
    }

    public native void setLockingKeyStateNative(int key, boolean on);

    public Clipboard getSystemClipboard() {
        SecurityManager security = System.getSecurityManager();
        if (security != null) {
	  security.checkSystemClipboardAccess();
	}
        if (clipboard == null) {
	    clipboard = new WClipboard();
	}
	return clipboard;
    }

    protected native void loadSystemColors(int[] systemColors);

    public static final Object targetToPeer(Object target) {
        return SunToolkit.targetToPeer(target);
    }

    public static final void targetDisposedPeer(Object target, Object peer) {
        SunToolkit.targetDisposedPeer(target, peer);
    }

    /**
     * Returns a new input method adapter descriptor for native input methods.
     */
    public InputMethodDescriptor getInputMethodAdapterDescriptor() {
        return new WInputMethodDescriptor();
    }

    /**
     * Returns a style map for the input method highlight.
     */
    public Map mapInputMethodHighlight(InputMethodHighlight highlight) {
	return WInputMethod.mapInputMethodHighlight(highlight);
    }

    /**
     * Returns whether enableInputMethods should be set to true for peered
     * TextComponent instances on this platform.
     */
    public boolean enableInputMethodsForTextComponent() {
        return true;
    }

    /**
     * Returns a new custom cursor.
     */
    public Cursor createCustomCursor(Image cursor, Point hotSpot, String name)
        throws IndexOutOfBoundsException {
        return new WCustomCursor(cursor, hotSpot, name);
    }

    /**
     * Returns the supported cursor size (Win32 only has one).
     */
    public Dimension getBestCursorSize(int preferredWidth, int preferredHeight) {
        return new Dimension(WCustomCursor.getCursorWidth(), 
                             WCustomCursor.getCursorHeight());
    }

    public native int getMaximumCursorColors();

    /* Called from Toolkit thread - NO CLIENT CODE! */
    static void displayChange() {
        //Reset the static GC
        resetGC();
        //NOTE: for multiscreen, this should only reset the device that was changed
        GraphicsDevice screens[] = GraphicsEnvironment.
         getLocalGraphicsEnvironment().getScreenDevices(); 
        for(int s = 0; s < screens.length; s++) {
            if (screens[s] instanceof Win32GraphicsDevice) {
                ((Win32GraphicsDevice)screens[s]).displayModeChanged();
            }
        }
    }

    /**
     * create the peer for a DragSourceContext
     */
 
    public DragSourceContextPeer createDragSourceContextPeer(DragGestureEvent dge) throws InvalidDnDOperationException {
	return new WDragSourceContextPeer(dge);
    }

    public DragGestureRecognizer createDragGestureRecognizer(Class abstractRecognizerClass, DragSource ds, Component c, int srcActions, DragGestureListener dgl) {
        if (MouseDragGestureRecognizer.class.equals(abstractRecognizerClass))
            return new WMouseDragGestureRecognizer(ds, c, srcActions, dgl);
        else
            return null;
    }

    /**
     *
     */

    private static final String prefix  = "DnD.Cursor.";
    private static final String postfix = ".32x32";

    protected Object lazilyLoadDesktopProperty(String name) {
	if (name.startsWith(prefix)) {
	    String cursorName = name.substring(prefix.length(), name.length()) + postfix;

	    try {
	        return Cursor.getSystemCustomCursor(cursorName);
	    } catch (AWTException awte) {
	 	throw new RuntimeException("cannot load system cursor: " + cursorName);
	    }
	} else if ( WDesktopProperties.isWindowsProperty(name) ) {
        synchronized(this) {
            if (wprops == null) {
                wprops = new WDesktopProperties();
            }
            Object prop = wprops.getProperty(name);
            return prop;
        }
	}

	return super.lazilyLoadDesktopProperty(name);
    }

    /*
     * Called from native toolkit code when WM_SETTINGCHANGE message received
     */
    private void windowsSettingChange() {
        //wprops created lazily, so may be null
        if (wprops != null) {
            wprops.firePropertyChanges();
        }
    }

    public synchronized void addPropertyChangeListener(String name, PropertyChangeListener pcl) {
	if ( WDesktopProperties.isWindowsProperty(name) ) {
        if (wprops == null) {
            wprops = new WDesktopProperties();
        }
	    wprops.addPropertyChangeListener(name, pcl);
	} else {
	    super.addPropertyChangeListener(name, pcl);
	}
    }

    public synchronized void removePropertyChangeListener(String name, PropertyChangeListener pcl) {
	if ( WDesktopProperties.isWindowsProperty(name) ) {
        //wprops created lazily, so may be null
        if (wprops != null) {
            wprops.removePropertyChangeListener(name, pcl);
        }
	} else {
	    super.removePropertyChangeListener(name, pcl);
	}
    }

    protected void initializeDesktopProperties() {
        desktopProperties.put("DnD.Autoscroll.initialDelay",     new Integer(50));
        desktopProperties.put("DnD.Autoscroll.interval",         new Integer(50));
        //desktopProperties.put("DnD.Autoscroll.cursorHysteresis", new Integer(5));
	// DnD uses one value for x and y drag diff, but Windows provides separate ones.
	// For now, just use the x value - rnk

        wprops = new WDesktopProperties();
        
        desktopProperties.put("DnD.Autoscroll.cursorHysteresis", wprops.getProperty("win.drag.x"));
    }

    /*
     * Get the java.awt.Component object which corresponds to the native
     * window handle argument.  If no Component is found, null is returned.
     * 
     * This method is used by the accessibility software on Win32.
     */
    public native Component getComponentFromNativeWindowHandle(int hwnd);

    /*
     * Get the native window handle which corresponds to the given peer.
     */
    private native int getNativeWindowHandleFromPeer(WComponentPeer peer);

    /*
     * Get the native window handle which corresponds to the
     * java.awt.Component object.  If the Component is lightweight,
     * its native container's native window handle is returned.
     * 
     * This method is used by the accessibility software on Win32.
     */
    public int getNativeWindowHandleFromComponent(Component target) {
        // Find the native component's WComponentPeer, and call it for this
        java.awt.peer.ComponentPeer peer = target.getPeer();
        while ((peer != null) && !(peer instanceof WComponentPeer)) {
            target = target.getParent();
            if (target == null)
                return 0;   // No non-lightweight parent found
            peer = target.getPeer();
        }
        if (peer == null)
            return 0;   // No suitable peer found -- maybe not created yet
        return getNativeWindowHandleFromPeer((WComponentPeer)peer);
    }

    public String getDefaultUnicodeEncoding() {
	return "utf-16le";
    }

    public GlobalCursorManager getGlobalCursorManager() {
        return new WGlobalCursorManager();
    }

    ///////////////////////////////////////////////////////////////////////////
    //
    // The following is used by the Java Plug-in to coordinate dialog modality
    // between containing applications (browsers, ActiveX containers etc) and
    // the AWT.
    //
    ///////////////////////////////////////////////////////////////////////////

    private ModalityListenerList	modalityListeners = new ModalityListenerList();

    public static WToolkit getWToolkit() {
	WToolkit toolkit = (WToolkit)Toolkit.getDefaultToolkit();
	return toolkit;
    }

    /**
     * Pushes AWT into another level of modality. Used by Java Plug-in when
     * the containing application puts up a modal dialog of it's own
     */
    public native void pushModality();

    /**
     * Called by Java plug-in when containing app dismisses one of it's
     * own dialogs
     */
    public native void popModality();

    /**
     * Adds a listener to be notified of changes in global modality
     */
    public void addModalityListener(ModalityListener listener) {
	modalityListeners.add(listener);
    }
    
    /**
     * Removes modality listener
     */
    public void removeModalityListener(ModalityListener listener) {
	modalityListeners.remove(listener);
    }
    
    /**
     * Call to send modality events to interested parties
     */
    final void notifyModalityChange(int id) {
	ModalityEvent ev = new ModalityEvent(this, modalityListeners, id);
	ev.dispatch();
    }
    	
    class ModalityListenerList implements ModalityListener {
	Vector	listeners = new Vector();
	
	void add(ModalityListener listener) {
	    listeners.addElement(listener);
	}
	
	void remove(ModalityListener listener) {
	    listeners.removeElement(listener);
	}
	
	//
	// ModalityListener implementation
	//
	public void modalityPushed(ModalityEvent ev) {
	    Enumeration enum = listeners.elements();
	    while (enum.hasMoreElements()) {
		((ModalityListener)enum.nextElement()).modalityPushed(ev);
	    }
	}
	
	public void modalityPopped(ModalityEvent ev) {
	    Enumeration enum = listeners.elements();
	    while (enum.hasMoreElements()) {
		((ModalityListener)enum.nextElement()).modalityPopped(ev);
	    }
	}
    }

    ///////////////////////////////////////////////////////////////////////////
    // End Plug-in code
    ///////////////////////////////////////////////////////////////////////////
}
