/*
 * @(#)WWindowPeer.java	1.29 01/02/09
 *
 * Copyright 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
package sun.awt.windows;

import java.util.Vector;
import java.awt.*;
import java.awt.peer.*;
import java.awt.event.*;
import sun.awt.Win32GraphicsDevice;
import sun.awt.Win32GraphicsConfig;
import java.lang.ref.WeakReference;

class WWindowPeer extends WPanelPeer implements WindowPeer {

    /**
     * Initialize JNI field IDs
     */
    private static native void initIDs();
    
    // keeps track of which focus-on-activate events
    // are still valid vs. cancelled
    Vector pendingFocusEvents = new Vector();

    // grants setDefaultFocus request only to those valid
    // focus-on-activate events
    boolean grantDefaultFocus = true;

    // WComponentPeer overrides

    protected void disposeImpl() {
	pendingFocusEvents.removeAllElements();
	allWindows.removeElement(this);
	super.disposeImpl();
    }

    // WindowPeer implementation

    public native void toFront();
    public native void toBack();

    /* Handle TAB and Shift-TAB events. */
    public int handleFocusTraversalEvent(KeyEvent e) {

	if (e.getKeyCode() != KeyEvent.VK_TAB ||
            (e.getSource() instanceof TextArea))
	    return IGNORE_EVENT;

 	if ((e.getModifiers() & ~InputEvent.SHIFT_MASK) > 0)
 	    return IGNORE_EVENT;

	int id = e.getID();
 	if (id == KeyEvent.KEY_RELEASED || id == KeyEvent.KEY_TYPED)
 	    return CONSUME_EVENT;

 	if (e.isShiftDown())
 	    return FOCUS_PREVIOUS;
	else
 	    return FOCUS_NEXT;
    }

    // FramePeer & DialogPeer partial shared implementation

    public void setTitle(String title) {
        // allow a null title to pass as an empty string.
        if (title == null) {
            title = new String("");
        }
        _setTitle(title);
    }
    native void _setTitle(String title);

    public void setResizable(boolean resizable) {
        _setResizable(resizable);
    }
    public native void _setResizable(boolean resizable);

    // Toolkit & peer internals

    static Vector allWindows = new Vector();  //!CQ for anchoring windows, frames, dialogs

    WWindowPeer(Window target) {
        super(target);
    }

    void initialize() {
        super.initialize();

        updateInsets(insets_);
	allWindows.addElement(this);

        Font f = ((Window)target).getFont();
	if (f == null) {
	    f = new Font("Dialog", Font.PLAIN, 12);
	    ((Window)target).setFont(f);
	    setFont(f);
	}
    }

    native void createAwtWindow(WComponentPeer parent);
    void create(WComponentPeer parent) {
        createAwtWindow(parent);
    }
        

    // Synchronize the insets members (here & in helper) with actual window
    // state.
    native void updateInsets(Insets i);

    // If someone explicitly requests the focus we clear the flag vector
    // to ignore any currently pending focus-on-activate events
    static void cancelFocusOnActivate(Component comp) {
	Window  window = null;
	// determine window that owns the component
	for ( ; comp != null; comp = comp.getParent() ) {
	    if (comp instanceof Window) {
		window = (Window)comp;
	    }
	}
	if ( window != null) {
	    // cancel pending FocusOnActivate events
	    ((WWindowPeer)(window.getPeer())).pendingFocusEvents.removeAllElements();
	}
    }

    WComponentPeer getFocusPeer() {
        Component c = ((Window)target).getFocusOwner();
        while (c != null) {
	    //im --- bug fix
	    // If c is LightWeightComponet, getPeer() returns java.awt.LightweightPeer
	    // which is not subclass of WComponentPeer.
	    // Irregal cast exception will be caused by original code.
            ComponentPeer peer = c.getPeer();
	    if (peer != null && peer instanceof WComponentPeer) {
                return (WComponentPeer)peer;
	    //im --- end
            } else {
                // peerless sub-component, try parent.
                c = c.getParent();
            }
        }
        return null;  // no child has focus
    }

    // This method is called from C++ to set focus to a Motif-like "reasonable"
    // default.  It's only called if focus was not explicitly set by either the
    // Java app or the user.
    void setDefaultFocus() {
        if (!grantDefaultFocus) {
            // event was cancelled due to an explicit requestFocus
            return;
        }
        // Find first focus-traversable component.
        if (setDefaultFocus((Container)target)) {
            return;
        }

        // No focus-traversable component was found, so set focus to
        // the first leaf node.
        Component c = (Component)target;
        while (c instanceof Container &&
               ((Container)c).getComponentCount() > 0) {
            Component child = getContainerElement((Container)c, 0);
            if (child.isVisible() && child.isEnabled()) {
                c = child;
            } else {
                break;        // Bug fix for bug #4038896 (Oracle deadlock bug) - Jonathan Locke
            }
        }
        c.requestFocus();
    }

    // This is similar to Window.activateFocus(), but won't trip over any locks
    // held by client code.
    private boolean setDefaultFocus(Container cont) {
        int ncomponents = cont.getComponentCount();
        for (int i = 0; i < ncomponents; i++) {
            Component c = getContainerElement(cont, i);
            if (c.isVisible() && c.isEnabled() && c.isFocusTraversable()) {
                c.requestFocus();
                return true;
            }
            if (c instanceof Container && c.isVisible() && c.isEnabled()) {
                if (setDefaultFocus((Container)c)) {
                    return true;
                }
            }
        }
        return false;
    }

    private native Component getContainerElement(Container c, int i);

    static native int getSysMinWidth();
    static native int getSysMinHeight();

    synchronized native void reshapeFrame(int x, int y, int width, int height);

	protected void initNonDefaultGC() {
        GraphicsConfiguration gc;
        gc = ((Window)target).getGraphicsConfiguration();
        //Always add Window to list of Components to be updated when
        //Display Settings are changed, as they are top-level Containers.
        ((Win32GraphicsDevice)gc.getDevice()).addNonDefaultGC(
         new WeakReference(target));
	}
	
    /**
     * Set winGraphicsConfig to the static GC.  Called as the result of
     * a Display Settings change.  Called from the Toolkit thread, so 
     * NO CLIENT CODE!
     */
    void resetLocalGC() {
        synchronized (((Component)target).getTreeLock()) {
            winGraphicsConfig = (Win32GraphicsConfig) (GraphicsEnvironment.
                getLocalGraphicsEnvironment().
                getDefaultScreenDevice().
                getDefaultConfiguration());
        }
    }
}
