/*
 * @(#)PlainSocketImpl.c	1.25 01/02/09
 *
 * Copyright 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

#include <windows.h>
#include <winsock.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/types.h>

#include "java_net_SocketOptions.h"
#include "java_net_PlainSocketImpl.h"
#include "java_net_SocketImpl.h"
#include "java_net_InetAddress.h"
#include "java_io_FileDescriptor.h"
#include "java_lang_Integer.h"

#include "jvm.h"
#include "net_util.h"
#include "jni_util.h"

/************************************************************************
 * PlainSocketImpl
 */

static jfieldID IO_fd_fdID;

jfieldID psi_fdID;
jfieldID psi_addressID;
jfieldID psi_portID;
jfieldID psi_localportID;
jfieldID psi_timeoutID;

/*
 * the level of the TCP protocol for setsockopt and getsockopt
 * we only want to look this up once, from the static initializer
 * of PlainSocketImpl
 */
static int tcp_level = -1;

/*
 * the timeout value for connection. 0 means no timeout
 */
static int preferredConnectionTimeout;

/*
 * The initProto function is called whenever PlainSocketImpl is
 * loaded, to cache fieldIds for efficiency. This is called everytime
 * the Java class is loaded.
 *
 * Class:     java_net_PlainSocketImpl
 * Method:    initProto

 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_initProto(JNIEnv *env, jclass cls) {

    struct protoent *proto = getprotobyname("TCP");
    jfieldID pctID =  (*env)->GetStaticFieldID(env, cls, "preferredConnectionTimeout", "I");

    preferredConnectionTimeout = (*env)->GetStaticIntField(env, cls, pctID);
    if (preferredConnectionTimeout < 0) {
	preferredConnectionTimeout = 0;
    }
    tcp_level = (proto == 0 ? IPPROTO_TCP: proto->p_proto);

    psi_fdID = (*env)->GetFieldID(env, cls , "fd", "Ljava/io/FileDescriptor;");
    psi_addressID = (*env)->GetFieldID(env, cls, "address",
					  "Ljava/net/InetAddress;");
    psi_portID = (*env)->GetFieldID(env, cls, "port", "I");
    psi_localportID = (*env)->GetFieldID(env, cls, "localport", "I");
    psi_timeoutID = (*env)->GetFieldID(env, cls, "timeout", "I");
    IO_fd_fdID = NET_GetFileDescriptorID(env);
}

/*
 * Class:     java_net_PlainSocketImpl
 * Method:    socketCreate
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketCreate(JNIEnv *env, jobject this,
					   jboolean stream) {
    jobject fdObj;
    int fd;

    fdObj = (*env)->GetObjectField(env, this, psi_fdID);

    if (IS_NULL(fdObj)) {
        JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"null fd object");
	return;

    }
    fd = JVM_Socket(AF_INET, (stream ? SOCK_STREAM: SOCK_DGRAM), 0);
    if (fd == -1) {
        NET_ThrowCurrent(env, "create");
	return;
    } else {
	(*env)->SetIntField(env, fdObj, IO_fd_fdID, fd);
    }
}

/*
 * inetAddress is the address object passed to the socket connect
 * call.
 *
 * Class:     java_net_PlainSocketImpl
 * Method:    socketConnect
 * Signature: (Ljava/net/InetAddress;I)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketConnect(JNIEnv *env, jobject this,
					    jobject iaObj, jint port) {
   jint localport = (*env)->GetIntField(env, this, psi_localportID);

    /* address, family and localport are int fields of iaObj */
    int address, family;

    /* fdObj is the FileDescriptor field on this */
    jobject fdObj = (*env)->GetObjectField(env, this, psi_fdID);
    /* fd is an int field on iaObj */
    jint fd;

    struct sockaddr_in him;
    /* The result of the connection */
    int connect_res;

    /* Timeout value for the connection */
    int timeout = preferredConnectionTimeout;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"Socket closed");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (IS_NULL(iaObj)) {
	JNU_ThrowNullPointerException(env, "inet address argument is null.");
	return;
    } else {
	address = (*env)->GetIntField(env, iaObj, ia_addressID);
	family = (*env)->GetIntField(env, iaObj, ia_familyID);
    }

    /* connect */
    memset((char *) &him, 0, sizeof(him));
    him.sin_port = htons((short) port);
    him.sin_addr.s_addr = (unsigned long) htonl(address);
    him.sin_family = family;

    /* check if a positive timeout value is specified */
    if (timeout > 0) {
        int optval;
             
        /* make socket non-blocking */ 
        optval = 1;
        (*ioctlsocket_fn)(fd, FIONBIO, &optval);

        /* initiate the connect */
        connect_res = JVM_Connect(fd, (struct sockaddr *) &him, sizeof(him));
        if(connect_res == SOCKET_ERROR) {
            if (WSAGetLastError() != WSAEWOULDBLOCK) {
                connect_res = WSAGetLastError();
            } else {
                fd_set wset, eset;
                struct timeval t;

                t.tv_sec = timeout;
                t.tv_usec = 0;
                FD_ZERO(&wset); 
                FD_SET(fd, &wset);
                FD_ZERO(&eset);
                FD_SET(fd, &eset);

                /* wait for connect to complete/fail or timeout */ 
                connect_res = (*select_fn)(fd+1, NULL, &wset, &eset, &t);
                if (connect_res == 0) {            /* timed-out */
		    JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
                                    "operation timed out");
                    /* make socket blocking again - just in case */
                    optval = 0;
                    (*ioctlsocket_fn)(fd, FIONBIO, &optval);
                    JVM_SocketShutdown(fd, 2);
                    return;
		} else if (connect_res == SOCKET_ERROR) {
		    connect_res = WSAGetLastError();
		} else {
                    if (!FD_ISSET(fd, &eset)) {    /* connection successful */
                        connect_res = 0;
                    } else {
                        /*  
                         * The connection could not be established so we try
                         * to pick up the failure reason. On W95/winsock1.1
			 * there is a bug and getsockopt( SO_ERROR ) will
			 * probably return 0. In this case we have to return
			 * a default failure reason. 
                         */          
                        int retry;
                        int optlen = sizeof(connect_res);

                        for (retry = 0; retry < 3; retry++) {
                            JVM_GetSockOpt(fd, SOL_SOCKET, SO_ERROR,
                                       (char*)&connect_res, &optlen);
                            if (connect_res) {
                                break;
                            }
                            Sleep(0);
                        }

                        if (connect_res == 0) {
                            connect_res = WSAENETDOWN;
                        }
                    }
                }
            }
        }
             
        /* make socket blocking again */
        optval = 0;
        (*ioctlsocket_fn)(fd, FIONBIO, &optval);
    } else {
        connect_res = JVM_Connect(fd, (struct sockaddr *) &him, sizeof(him));
        if (connect_res == SOCKET_ERROR) {
	    connect_res = WSAGetLastError();
        }
    }
    if (connect_res) {
        NET_ThrowNew(env, connect_res, "connect");
        return;
    }
    (*env)->SetIntField(env, fdObj, IO_fd_fdID, fd);

    /* set the remote peer address and port */
    (*env)->SetObjectField(env, this, psi_addressID, iaObj);
    (*env)->SetIntField(env, this, psi_portID, port);

    /*
     * we need to initialize the local port field if bind was called
     * previously to the connect (by the client) then localport field
     * will already be initialized
     */
    if (localport == 0) {
	/* Now that we're a connected socket, let's extract the port number
	 * that the system chose for us and store it in the Socket object.
  	 */
	int len = sizeof(him);
	if (JVM_GetSockName(fd, (struct sockaddr *)&him, &len) == -1) {
	    NET_ThrowCurrent(env, "JVM_GetSockName failed");
	    return;
	}
	(*env)->SetIntField(env, this, psi_localportID, ntohs(him.sin_port));
    }
}

/*
 * Class:     java_net_PlainSocketImpl
 * Method:    socketBind
 * Signature: (Ljava/net/InetAddress;I)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketBind(JNIEnv *env, jobject this,
					 jobject iaObj, jint localport) {

    /* fdObj is the FileDescriptor field on this */
    jobject fdObj = (*env)->GetObjectField(env, this, psi_fdID);
    /* fd is an int field on fdObj */
    int fd;

    /* address and family are int fields on iaObj */
    int address, family;

    struct sockaddr_in him;
    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"Socket closed");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (IS_NULL(iaObj)) {
	JNU_ThrowNullPointerException(env, "inet address argument");
	return;
    } else {
	address = (*env)->GetIntField(env, iaObj, ia_addressID);
	family = (*env)->GetIntField(env, iaObj, ia_familyID);
    }
    /* bind */
    memset((char *) &him, 0, sizeof(him));
    him.sin_port = htons((short)localport);
    him.sin_addr.s_addr = (unsigned long) htonl(address);
    him.sin_family = family;

    if (JVM_Bind(fd, (struct sockaddr *)&him, sizeof(him)) == -1) {
	NET_ThrowCurrent(env, "JVM_Bind");
	return;
    }

    /* set the address */
    (*env)->SetObjectField(env, this, psi_addressID, iaObj);

    /* intialize the local port */
    if (localport == 0) {
	/* Now that we're a connected socket, let's extract the port number
	 * that the system chose for us and store it in the Socket object.
  	 */
	int len = sizeof(him);
	if (JVM_GetSockName(fd, (struct sockaddr *)&him, &len) == -1) {
	    NET_ThrowCurrent(env, "JVM_GetSockName in plain socketBind");
	    return;
	}
	(*env)->SetIntField(env, this, psi_localportID,
			    ntohs(him.sin_port));
    } else {
	(*env)->SetIntField(env, this, psi_localportID, localport);
    }
}

/*
 * Class:     java_net_PlainSocketImpl
 * Method:    socketListen
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketListen (JNIEnv *env, jobject this,
					    jint count)
{
    /* this FileDescriptor fd field */
    jobject fdObj = (*env)->GetObjectField(env, this, psi_fdID);
    /* fdObj's int fd field */
    int fd;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"socket closed");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (JVM_Listen(fd, count) == -1) {
	NET_ThrowCurrent(env, "listen failed");
    }
}

/*
 * Class:     java_net_PlainSocketImpl
 * Method:    socketAccept
 * Signature: (Ljava/net/SocketImpl;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketAccept(JNIEnv *env, jobject this,
					   jobject socket)
{
    /* fields on this */
    jint port;
    jint timeout = (*env)->GetIntField(env, this, psi_timeoutID);
    jobject fdObj = (*env)->GetObjectField(env, this, psi_fdID);

    /* the FileDescriptor field on socket */
    jobject socketFdObj;
    /* the InetAddress field on socket */
    jobject socketAddressObj;

    /* the fd int field on fdObj */
    jint fd;

    jthrowable error;

    struct sockaddr_in him;
    jint len = sizeof(him);

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"Socket closed");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (IS_NULL(socket)) {
	JNU_ThrowNullPointerException(env, "socket is null");
	return;
    } else {
	socketFdObj = (*env)->GetObjectField(env, socket, psi_fdID);
	socketAddressObj = (*env)->GetObjectField(env, socket, psi_addressID);
    }
    if ((IS_NULL(socketAddressObj)) || (IS_NULL(socketFdObj))) {
	JNU_ThrowNullPointerException(env, "socket address or fd obj");
	return;
    }
    if (timeout) {
        int ret = JVM_Timeout(fd, timeout);
	if (ret == 0) {
	    JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "Accept timed out");
	    return;
	} else if (ret == -1) {
	    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "socket closed");
	/* REMIND: SOCKET CLOSED PROBLEM */
/*	      NET_ThrowCurrent(env, "Accept failed"); */
	    return;
	} else if (ret == -2) {
	    JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "operation interrupted");
	    return;
	}
    }
    fd = JVM_Accept(fd, (struct sockaddr *)&him, &len);
    if (fd < 0) {
	/* REMIND: SOCKET CLOSED PROBLEM */
        if (fd == -2) {
            JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "operation interrupted");
        } else {
            JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", 
                            "socket closed");
        }
	return;
    }
    (*env)->SetIntField(env, socketFdObj, IO_fd_fdID, fd);

    /*
     * fill up the remote peer port and address in the new socket structure
     */
    (*env)->SetIntField(env, socket, psi_portID, ntohs(him.sin_port));
    (*env)->SetIntField(env, socketAddressObj, ia_familyID, him.sin_family);
    (*env)->SetIntField(env, socketAddressObj, ia_addressID,
			ntohl(him.sin_addr.s_addr));
    /* also fill up the local port information */
    port = (*env)->GetIntField(env, this, psi_localportID);
    (*env)->SetIntField(env, socket, psi_localportID, port);
}

/*
 * Class:     java_net_PlainSocketImpl
 * Method:    socketAvailable
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_java_net_PlainSocketImpl_socketAvailable(JNIEnv *env, jobject this) {

    jint available = -1;
    jint res;
    jobject fdObj = (*env)->GetObjectField(env, this, psi_fdID);
    jint fd;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "Socket closed");
	return -1;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    res = JVM_SocketAvailable(fd, &available);
    /* if result isn't 0, it means an error */
    if (res != 0) {
	NET_ThrowNew(env, res, "socket available");
    }
    return available;
}

/*
 * Class:     java_net_PlainSocketImpl
 * Method:    socketClose
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketClose(JNIEnv *env, jobject this) {

    jobject fdObj = (*env)->GetObjectField(env, this, psi_fdID);
    jint fd;

    /*
     * WARNING: THIS NEEDS LOCKING. ALSO: SHOULD WE CHECK for fd being
     * -1 already?
     */
    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"socket already closed");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    (*env)->SetIntField(env, fdObj, IO_fd_fdID, -1);
    JVM_SocketClose(fd);
}

/*
 * Socket options for plainsocketImpl
 *
 *
 * Class:     java_net_PlainSocketImpl
 * Method:    socketSetOption
 * Signature: (IZLjava/lang/Object;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketSetOption(JNIEnv *env, jobject this,
					      jint cmd, jboolean on,
					      jobject value) {
    jobject fdObj = (*env)->GetObjectField(env, this, psi_fdID);
    jint fd, buflen;

    jclass iCls;
    jfieldID i_valueID;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "Socket closed");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (fd < 0) {

	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "Socket closed");
	return;
    }
    if (cmd == java_net_SocketOptions_TCP_NODELAY) {
	long onl = (long)on;

	if (JVM_SetSockOpt(fd, tcp_level, TCP_NODELAY,
		       (char *)&onl, sizeof(long)) < 0) {
		NET_ThrowCurrent(env, "JVM_SetSockOpt() TCP_NODELAY");
	}
	return;
    } else if(cmd == java_net_SocketOptions_SO_LINGER) {
	struct linger arg;
	jobject intObj;
	arg.l_onoff = on;

	if(on) {
	    if (IS_NULL(value)) {

		NET_ThrowCurrent(env, "setsockopt() SO_LINGER");
		return;
	    }
	    iCls = (*env)->FindClass(env, "java/lang/Integer");
	    i_valueID = (*env)->GetFieldID(env, iCls, "value", "I");
	    arg.l_linger = (*env)->GetIntField(env, value, i_valueID);
	    if(JVM_SetSockOpt(fd, SOL_SOCKET, SO_LINGER,
			  (char*)&arg, sizeof(arg)) < 0) {
		NET_ThrowCurrent(env, "JVM_SetSockOpt() SO_LINGER");
		return;
	    }
	} else {

	    if (JVM_SetSockOpt(fd, SOL_SOCKET, SO_LINGER,
			   (char*)&arg, sizeof(arg)) < 0) {
		NET_ThrowCurrent(env, "JVM_SetSockOpt() SO_LINGER");
		return;
	    }
	}
    } else if (cmd == java_net_SocketOptions_SO_SNDBUF) {
	iCls = (*env)->FindClass(env, "java/lang/Integer");
	i_valueID = (*env)->GetFieldID(env, iCls, "value", "I");
	buflen = (*env)->GetIntField(env, value, i_valueID);
	if (JVM_SetSockOpt(fd, SOL_SOCKET, SO_SNDBUF,
		       (char *)&buflen, sizeof(buflen)) < 0) {
	    NET_ThrowCurrent(env, "JVM_SetSockOpt() SO_SNDBUF");
	    return;

	}
    } else if (cmd == java_net_SocketOptions_SO_RCVBUF) {
	iCls = (*env)->FindClass(env, "java/lang/Integer");
	i_valueID = (*env)->GetFieldID(env, iCls, "value", "I");
	buflen = (*env)->GetIntField(env, value, i_valueID);
	if (JVM_SetSockOpt(fd, SOL_SOCKET, SO_RCVBUF,
		       (char *)&buflen, sizeof(buflen)) < 0) {
	    NET_ThrowCurrent(env, "JVM_SetSockOpt() SO_RCVBUF");
	    return;

	}
    } else if (cmd == java_net_SocketOptions_SO_KEEPALIVE) {
	long onl = (long)on;

	if (JVM_SetSockOpt(fd, SOL_SOCKET, SO_KEEPALIVE,
		       (char *)&onl, sizeof(long)) < 0) {
		NET_ThrowCurrent(env, "setsockopt() SO_KEEPALIVE");
	}
	return;
    } else {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
		  "Socket option unsupported");
	return;
    }
    return;
}

/*
 * Class:     java_net_PlainSocketImpl
 * Method:    socketGetOption
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL
Java_java_net_PlainSocketImpl_socketGetOption(JNIEnv *env, jobject this,
					      jint opt) {

    jobject fdObj = (*env)->GetObjectField(env, this, psi_fdID);
    int fd;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "Socket closed");
	return -1;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if(fd < 0) {

	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "Socket closed");
	return -1;
    }
    if (opt == java_net_SocketOptions_TCP_NODELAY) {

	int result, len;
	len = sizeof(int);
	if (JVM_GetSockOpt(fd, tcp_level, TCP_NODELAY, (char*)&result, &len) < 0) {
	    NET_ThrowCurrent(env, "JVM_GetSockOpt() TCP_NODELAY");
	    return -1;
	}
	return (result == 0 ? -1 : 1);

    } else if (opt == java_net_SocketOptions_SO_BINDADDR) {
	struct sockaddr_in him;
	int len = sizeof(struct sockaddr_in);
	memset(&him, 0, len);
	if (JVM_GetSockName(fd, (struct sockaddr *)&him, &len) == -1) {
	    NET_ThrowCurrent(env, "JVM_GetSockOpt() SO_BINDADDR");
	    return -1;
	}
	return (him.sin_addr.s_addr == INADDR_ANY) ? -1: htonl(him.sin_addr.s_addr);

    } else if (opt == java_net_SocketOptions_SO_LINGER) {
	struct linger arg;
	int len = sizeof(struct linger);
	memset(&arg, 0, len);

	if(JVM_GetSockOpt(fd, SOL_SOCKET, SO_LINGER, (char*)&arg, &len) < 0) {
	    NET_ThrowCurrent(env, "JVM_GetSockOpt() SO_LINGER");
	    return -1;
	}
	return (arg.l_onoff ? arg.l_linger: -1);
    } else if (opt == java_net_SocketOptions_SO_SNDBUF) {
	int buflen = -1;
	int len = sizeof(buflen);
	if (JVM_GetSockOpt(fd, SOL_SOCKET, SO_SNDBUF,
		       (char *)&buflen, &len) < 0) {
	    NET_ThrowCurrent(env, "JVM_GetSockOpt() SO_SNDBUF");
	    return -1;
	}
	return buflen;
    } else if (opt == java_net_SocketOptions_SO_RCVBUF) {
	int buflen = -1;
	int len = sizeof(buflen);
	if (JVM_GetSockOpt(fd, SOL_SOCKET, SO_RCVBUF,
		       (char *)&buflen, &len) < 0) {
	    NET_ThrowCurrent(env, "JVM_GetSockOpt() SO_RCVBUF");
	    return -1;
	}
	return buflen;
    } else if (opt == java_net_SocketOptions_SO_KEEPALIVE) {

	int result, len;
	len = sizeof(int);
	if (JVM_GetSockOpt(fd, SOL_SOCKET, SO_KEEPALIVE, (char*)&result, &len) < 0) {
	    NET_ThrowCurrent(env, "getsockopt() SO_KEEPALIVE");
	    return -1;
	}
	return (result == 0 ? -1 : 1);

    }

    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
		    "invalid option");
    return -1;
}

/*
 * Class:     java_net_PlainSocketImpl
 * Method:    socketShutdown
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketShutdown(JNIEnv *env, jobject this,
					     jint howto) 
{

    jobject fdObj = (*env)->GetObjectField(env, this, psi_fdID);
    jint fd;

    /*
     * WARNING: THIS NEEDS LOCKING. ALSO: SHOULD WE CHECK for fd being
     * -1 already?
     */
    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"socket already closed");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    JVM_SocketShutdown(fd, howto);
}


