/*
 * @(#)awt_Image.cpp	1.44 01/02/09
 *
 * Copyright 1996-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

#include "awt.h"
#include "awt_Toolkit.h"
#include "awt_Component.h"
#include "awt_Graphics.h"
#include <awt_image.h>
#include "awt_colors.h"
#include "jni.h"
#include <math.h>
#include "java_awt_color_ColorSpace.h"
#include "java_awt_Transparency.h"
#include "java_awt_image_DataBuffer.h"

CriticalSection imageLock;

/************************************************************************
 * AwtImage fields
 */

jmethodID AwtImage::getBackgroundMID;

// Initialize static variables
HPALETTE AwtImage::m_hPal = NULL;
ImgColorData AwtImage::m_clrdata;

static void make_dither_arrays(int cmapsize, ColorData *cData);
static sgn_ordered_dither_array std_img_oda_red;
static sgn_ordered_dither_array std_img_oda_green;
static sgn_ordered_dither_array std_img_oda_blue;
static int std_odas_computed = 0;

#define ERROR_GRAY (-1)
#define NON_GRAY 0
#define LINEAR_STATIC_GRAY 1
#define NON_LINEAR_STATIC_GRAY 2
#define DYNAMIC_GRAY 3


extern "C" {

#include "imageInitIDs.h"
const char *cubemapArray = NULL;

uns_ordered_dither_array img_oda_alpha;

void initInverseGrayLut(int* prgb, int rgbsize, ColorData* cData);

// This is data structure holds the common colormap that is used to paint
// all images.  When an image is painted, the BITMAPINFOHEADER of 
// the image is copied into gpBitmapInfo and then gpBitmapInfo is used to
// paint the image.  This strategy eliminates an alloc/free when painting
// an image.

// fix for 4302507: we need two BITMAPINFOs: 1 for 8-bit, and one for > 8-bit
BITMAPINFO *gpBitmapInfo8 = NULL;
BITMAPINFO *gpBitmapInfo8plus = NULL;

BITMAPINFO *gpTransMaskInfo = NULL;

const int COLOR4 = 0;
const int COLOR8 = 1;

ColorInfo g_colorInfo[2];

} /* extern "C" */

/************************************************************************
 * AwtImage methods
 */

#define SetRGBQUAD(p, r, g, b)	\
    {				\
	(p)->rgbRed = r;	\
	(p)->rgbGreen = g;	\
	(p)->rgbBlue = b;	\
	(p)->rgbReserved = 0;	\
    }

int AwtImage::FetchPaletteEntries(HDC hDC, PALETTEENTRY* pPalEntries)
{
    LOGPALETTE* pLogPal = 0;
    HPALETTE hPal = 0;
    HPALETTE hPalOld = 0;

    pLogPal = (LOGPALETTE*) new char[sizeof(LOGPALETTE) 
				    + 256*sizeof(PALETTEENTRY)];
    if (pLogPal == NULL) {
	return 0;
    }

    pLogPal->palVersion = 0x300;
    pLogPal->palNumEntries = 256;
    int iEntry;
    PALETTEENTRY* pEntry;
    for (iEntry = 0; iEntry < 256; iEntry++) {
	pEntry = pLogPal->palPalEntry + iEntry;
	pEntry->peRed = iEntry;
	pEntry->peGreen = pEntry->peBlue = 0;
	pEntry->peFlags = PC_EXPLICIT;
    }
    hPal = ::CreatePalette(pLogPal);
    if ( hPal == 0 ) {
	delete pLogPal;
	return 0;
    }

    hPalOld = ::SelectPalette(hDC, hPal, 1);
    if (hPalOld == 0) {
	delete pLogPal;
        ::DeleteObject(hPal);
	return 0;
    }
    ::RealizePalette(hDC);

    COLORREF rgb;
    for (iEntry = 0; iEntry < 256; iEntry++) {
    	rgb = ::GetNearestColor(hDC, PALETTEINDEX(iEntry));
	pPalEntries[iEntry].peRed = GetRValue(rgb);
	pPalEntries[iEntry].peGreen = GetGValue(rgb);
	pPalEntries[iEntry].peBlue = GetBValue(rgb);
    }

    delete pLogPal;
    ::DeleteObject(hPal);
    ::SelectPalette(hDC, hPalOld, 0 );
    ::RealizePalette(hDC);

    return 256;
}

static int GetGSType(HDC hDC, PALETTEENTRY* pPalEntries)
{
    int isGray = 1;
    int isLinearStaticGray = 1;
    int isNonLinearStaticGray = 1;
    int isDynamicGray = 0;
    int iEntry;
    char bUsed[256]; memset(bUsed, 0, sizeof(bUsed));
    BYTE r, g, b;

    AwtImage::FetchPaletteEntries(hDC, pPalEntries);
    for (iEntry = 0; iEntry < 256; iEntry++) {
	r = pPalEntries[iEntry].peRed;
	g = pPalEntries[iEntry].peGreen;
	b = pPalEntries[iEntry].peBlue;
	if (r != g || r != b) {
	    if (iEntry > 9 && iEntry < 246 ) {
		// not gray at all
		isGray = 0;
		break;
	    } else {
		// Dynamic Gray: 0..9 & 246..255 reserved
		// and may be non-gray, others do not matter
		isDynamicGray = 1;
	    }
	} else {
	    // the values are gray
	    if (r != iEntry) {
		// it's not linear
		// but it could be non-linear static gray
		isLinearStaticGray = 0;
	    }
	    bUsed[r] = 1;
	}
    }

    if (isGray && !isDynamicGray && !isLinearStaticGray) {
	// check if all 256 grays are there
	// if that's the case, it's non-linear static gray
	for (iEntry = 0; iEntry < 256; iEntry++ ) {
	    if (!bUsed[iEntry]) {
		// not non-linear (not all 256 colors are used)
		isNonLinearStaticGray = 0;
		break;
	    }
	}
    }

    if (!isGray) { return NON_GRAY; }
    if (isLinearStaticGray) { return LINEAR_STATIC_GRAY; }
    if (isNonLinearStaticGray) { return NON_LINEAR_STATIC_GRAY; }
    if (isDynamicGray) { return DYNAMIC_GRAY; }
    
    // not supposed to be here, error
    return ERROR_GRAY;
}

/* Builds the 8 and 4-bit palettes for
 * the given 8-bit screen.
 * Code mostly copied from AwtImage::InitializePalette
 */
void AwtImage::BuildPalettes(HDC hDC) {
    ColorInfo *pColorInfo;
    int bitsperpixel = ::GetDeviceCaps(hDC, BITSPIXEL);

    if (bitsperpixel > 8) {
        return;
    }

    pColorInfo = &g_colorInfo[COLOR8];
    DASSERT(pColorInfo);

    pColorInfo->m_pPaletteColors = (RGBQUAD*)((LPSTR)gpBitmapInfo8 +
                                                      sizeof(BITMAPINFOHEADER));

	// Create and initialize a palette
	int nEntries = max(256, ::GetDeviceCaps(hDC, SIZEPALETTE));
 	char *buf = NULL;
  	try {
  	    buf = new char[sizeof(LOGPALETTE) + nEntries *
  			  sizeof(PALETTEENTRY)];
  	} catch (std::bad_alloc&) {
  	    free(gpTransMaskInfo);
  	    gpTransMaskInfo = NULL;
  	    free(gpBitmapInfo8);
  	    gpBitmapInfo8 = NULL;
  	    throw;
  	}
  
  	LOGPALETTE *pLogPal = (LOGPALETTE*)buf;
  	PALETTEENTRY *pPalEntries = (PALETTEENTRY *)(&(pLogPal->palPalEntry[0])); 
  
  	if (nEntries == 256) {
              int cmapsize = 256;
              int staticGrayType;
              PALETTEENTRY *pPal;
              char ebuf[4];
              int i;
  
	      staticGrayType = GetGSType(hDC, pPalEntries);
	      if (staticGrayType < 0) {
#ifdef DEBUG
		  fprintf(stderr,"Error determining GS type. Assuming GS_NONGRAY\n");
#endif
		  staticGrayType = NON_GRAY;
	      }

	      if (staticGrayType == LINEAR_STATIC_GRAY) {
                  m_clrdata.grayscale = GS_STATICGRAY;
              } else if (GetEnvironmentVariable("FORCEGRAY", ebuf, 0)) {
                  float m = 255.0f / 235.0f;
                  float g = 0.5f;
                  m_clrdata.grayscale = GS_INDEXGRAY;
                  for (i = 0; i < 10; i++) {
                      pPalEntries[i].peFlags = PC_NOCOLLAPSE;
                  }
                  pPal = &pPalEntries[10];
                  for (i = 10; i < 246; i++, pPal++) {
                      pPal->peRed = pPal->peGreen = pPal->peBlue =
                          (int) g;
                      g += m;
                      pPal->peFlags = PC_NOCOLLAPSE;
                  }
                  for (i = 246; i < 256; i++) {
                      pPalEntries[i].peFlags = PC_NOCOLLAPSE;
                  }
              } else if (staticGrayType == NON_LINEAR_STATIC_GRAY) {
                  m_clrdata.grayscale = GS_INDEXGRAY;
		  // use palette entried which were retrieved
                  for (i = 0; i < 256; i++) {
                      pPalEntries[i].peFlags = PC_NOCOLLAPSE;
                  }
	      } else { // NON_GRAY or DYNAMIC_GRAY
                  for (i = 0; i < 10; i++) {
                      pPalEntries[i].peFlags = PC_NOCOLLAPSE;
                  }
                  pPal = &pPalEntries[10];
                  for (i = 10; i < 246; i++, pPal++) {
                      pPalEntries[i].peRed   = awt_reds[i];
                      pPalEntries[i].peGreen = awt_greens[i];
                      pPalEntries[i].peBlue  = awt_blues[i];
                      pPalEntries[i].peFlags = PC_NOCOLLAPSE;
                  }
                  for (i = 246; i < 256; i++) {
                      pPalEntries[i].peFlags = PC_NOCOLLAPSE;
                  }
              }
  
              {
              ColorData colorData;
  
              colorData.img_oda_red = &(std_img_oda_red[0][0]);
              colorData.img_oda_green = &(std_img_oda_green[0][0]);
              colorData.img_oda_blue = &(std_img_oda_blue[0][0]);
              make_dither_arrays(cmapsize, &colorData);
              std_odas_computed = 1;
              }
  
              // Initialize m_paletteColors8.
              for (i = 0; i < nEntries; i++) {
                  pColorInfo->m_pPaletteColors[i].rgbRed =
                  pPalEntries[i].peRed;
                  pColorInfo->m_pPaletteColors[i].rgbGreen =
                  pPalEntries[i].peGreen,
                  pColorInfo->m_pPaletteColors[i].rgbBlue =
                  pPalEntries[i].peBlue;
                  pColorInfo->m_pPaletteColors[i].rgbReserved = 0;
              }
  
              // Now clear flags of system colors.
              for (i=0; i<10; i++) {
                  pPalEntries[i].peFlags = 0;
                  pPalEntries[255-i].peFlags = 0;
              }
  	    pLogPal->palNumEntries = 256;
  	    pLogPal->palVersion = 0x300;
  	    m_hPal = ::CreatePalette(pLogPal);
  	}
  	delete [] buf;
  
  	// Now allocate the 4-bit color palette.
  	pColorInfo = &g_colorInfo[COLOR4];
  	try {
  	    pColorInfo->m_pPaletteColors = (RGBQUAD*)
  	        safe_Malloc(16*sizeof(RGBQUAD));
  	} catch (std::bad_alloc&) {
  	    free(gpTransMaskInfo);
  	    gpTransMaskInfo = NULL;
  	    free(gpBitmapInfo8);
  	    gpBitmapInfo8 = NULL;
  	    throw;
  	}
  
  	/* 9: 166, 202, 240 	246: 255 251 240 -- non-VGA colors */	
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[0]), 0, 0, 0);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[1]), 128, 0, 0);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[2]), 0, 128, 0);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[3]), 128, 128, 0);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[4]), 0, 0, 128);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[5]), 128, 0, 128);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[6]), 0, 128, 128);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[7]), 192, 192, 192);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[8]), 128, 128, 128);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[9]), 255, 0, 0);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[10]), 0, 255, 0);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[11]), 255, 255, 0);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[12]), 0, 0, 255);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[13]), 255, 0,255);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[14]), 0, 255, 255);
  	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[15]), 255, 255, 255);
}

void AwtImage::InitializePalette(HDC hDC)
{
    unsigned int ri, gi, bi;
		
    m_clrdata.bitsperpixel = ::GetDeviceCaps(hDC, BITSPIXEL);
    if (m_clrdata.bitsperpixel < 8) {
		//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
		// Bugs 4032109, 4047966, and 4071991
		// creighto@eng.sun.com 1997-10-07
		//
		// if less than 4, gracefully end
		if (m_clrdata.bitsperpixel < 4) {
			// Windows 95 special case
		    int colorbits = ::GetDeviceCaps(hDC, BITSPIXEL) *
							::GetDeviceCaps(hDC, PLANES);
			if (colorbits < 4) {
				return;
			}
			else {
				m_clrdata.bitsperpixel = 8;
			}
		}
		else {
			// If 4 bit color mode, use 8 bit palette
			m_clrdata.bitsperpixel = 8;
		}
		//////////////////////////////////////////////////
    }

    // Initialize the dithering array used for transparency masks
    make_uns_ordered_dither_array(img_oda_alpha, 256);

    // Create a BitmapInfo object for transparency masks
    gpTransMaskInfo = (BITMAPINFO *)
	safe_Malloc(sizeof(BITMAPINFOHEADER) + 2 * sizeof(RGBQUAD));

    gpTransMaskInfo->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    gpTransMaskInfo->bmiHeader.biWidth = 0;
    gpTransMaskInfo->bmiHeader.biHeight = 0;
    gpTransMaskInfo->bmiHeader.biPlanes = 1;
    gpTransMaskInfo->bmiHeader.biBitCount = 1;
    gpTransMaskInfo->bmiHeader.biCompression = BI_RGB;
    gpTransMaskInfo->bmiHeader.biSizeImage = 0;
    gpTransMaskInfo->bmiHeader.biXPelsPerMeter = 0;
    gpTransMaskInfo->bmiHeader.biYPelsPerMeter = 0;
    gpTransMaskInfo->bmiHeader.biClrUsed = 2;
    gpTransMaskInfo->bmiHeader.biClrImportant = 2;
    SetRGBQUAD(&(gpTransMaskInfo->bmiColors[0]), 0, 0, 0);
    SetRGBQUAD(&(gpTransMaskInfo->bmiColors[1]), 255, 255, 255);

    // Create a BitmapInfo object for color data
    try {
        gpBitmapInfo8 = (BITMAPINFO *)
        safe_Malloc(sizeof(BITMAPINFOHEADER) + 256 * sizeof(RGBQUAD));
        gpBitmapInfo8plus = (BITMAPINFO *)
        safe_Malloc(sizeof(BITMAPINFOHEADER) + 256 * sizeof(RGBQUAD));
    } catch (std::bad_alloc&) {
	free(gpTransMaskInfo);
	gpTransMaskInfo = NULL;
	throw;
    }

    gpBitmapInfo8->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    gpBitmapInfo8->bmiHeader.biBitCount = 0;
    gpBitmapInfo8plus->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    gpBitmapInfo8plus->bmiHeader.biBitCount = 0;
    HDC hBMDC = ::CreateDC("DISPLAY", NULL, NULL, NULL);
    HBITMAP hBM = ::CreateCompatibleBitmap(hBMDC, 1, 1);
    VERIFY(::GetDIBits(hBMDC, hBM, 0, 1, NULL, gpBitmapInfo8, DIB_RGB_COLORS));
    VERIFY(::GetDIBits(hBMDC, hBM, 0, 1, NULL, gpBitmapInfo8plus,
                       DIB_RGB_COLORS));

    m_clrdata.grayscale = GS_NOTGRAY;
    if (m_clrdata.bitsperpixel > 8) {
	if (m_clrdata.bitsperpixel != 24) {
	    int foo;
        gpBitmapInfo8plus->bmiHeader.biCompression = BI_BITFIELDS;
	    VERIFY(::GetDIBits(hBMDC, hBM, 0, 1, &foo,
                           gpBitmapInfo8plus, DIB_RGB_COLORS));
        ri = ((unsigned int *)gpBitmapInfo8plus->bmiColors)[0];
	    m_clrdata.rOff = 0;
	    while ((ri & 1) == 0) {
		m_clrdata.rOff++;
		ri >>= 1;
	    }
	    m_clrdata.rScale = 0;
	    while (ri < 0x80) {
		m_clrdata.rScale++;
		ri <<= 1;
	    }
        gi = ((unsigned int *)gpBitmapInfo8plus->bmiColors)[1];
	    m_clrdata.gOff = 0;
	    while ((gi & 1) == 0) {
		m_clrdata.gOff++;
		gi >>= 1;
	    }
	    m_clrdata.gScale = 0;
	    while (gi < 0x80) {
		m_clrdata.gScale++;
		gi <<= 1;
	    }
        bi = ((unsigned int *)gpBitmapInfo8plus->bmiColors)[2];
	    m_clrdata.bOff = 0;
	    while ((bi & 1) == 0) {
		m_clrdata.bOff++;
		bi >>= 1;
	    }
	    m_clrdata.bScale = 0;
	    while (bi < 0x80) {
		m_clrdata.bScale++;
		bi <<= 1;
	    }
            if (       (0 == m_clrdata.bOff)
                    && (5 == m_clrdata.gOff)
                    && (10 == m_clrdata.rOff)
                    && (3 == m_clrdata.bScale)
                    && (3 == m_clrdata.gScale)
                    && (3 == m_clrdata.rScale)) {
                m_clrdata.bitsperpixel = 15;
        gpBitmapInfo8plus->bmiHeader.biCompression = BI_RGB;
            }
	} else {
	    m_clrdata.bitsperpixel = 24;
        gpBitmapInfo8plus->bmiHeader.biBitCount = 24;
	    gpBitmapInfo8plus->bmiHeader.biCompression = BI_RGB;

	    // Fill these values in as a convenience for the screen
	    // ColorModel construction code below (see getColorModel())
        ((int *)gpBitmapInfo8plus->bmiColors)[0] = 0x0000ff;
 	    ((int *)gpBitmapInfo8plus->bmiColors)[1] = 0x00ff00;
 	    ((int *)gpBitmapInfo8plus->bmiColors)[2] = 0xff0000;
	}
    } else {
    	gpBitmapInfo8->bmiHeader.biBitCount = 8;
        gpBitmapInfo8->bmiHeader.biCompression = BI_RGB;
        gpBitmapInfo8->bmiHeader.biClrUsed = 256;
        gpBitmapInfo8->bmiHeader.biClrImportant = 256;
        // Code that was here has been moved into BuildPalettes
        try {
            BuildPalettes(hDC);
    //        AwtToolkit::GetInstance().adjustPalette(hDC);
        } catch (std::bad_alloc&) {
            VERIFY(::DeleteObject(hBM));
            VERIFY(::DeleteDC(hBMDC));
            throw;
        }
        VERIFY(::DeleteObject(hBM));
        VERIFY(::DeleteDC(hBMDC));
    }
}

void AwtImage::DisposePalette()
{
	free(gpBitmapInfo8);
	free(gpBitmapInfo8plus);
	free(gpTransMaskInfo);
	if (m_hPal != NULL) {
	    ::DeleteObject(m_hPal);
	}
        if (g_colorInfo[COLOR4].m_pPaletteColors != NULL) {
            free(g_colorInfo[COLOR4].m_pPaletteColors);
        }
}

jobject AwtImage::getColorModel(JNIEnv *env)
{
    jobject awt_colormodel;
    int i;
    if (m_clrdata.bitsperpixel == 24) {
        awt_colormodel =
	    JNU_NewObjectByName(env, "sun/awt/Win32ColorModel24", "()V");
    } else if (m_clrdata.bitsperpixel > 8) {
	int *masks = (int *)gpBitmapInfo8plus->bmiColors;
	int numbits = 0;
	unsigned int bits = (masks[0] | masks[1] | masks[2]);
	while (bits) {
	    numbits++;
	    bits >>= 1;
	}
        awt_colormodel = JNU_NewObjectByName(env, 
                                             "java/awt/image/DirectColorModel",
					     "(IIII)V", numbits,
					     masks[0], masks[1], masks[2]);
    } else if (m_clrdata.grayscale == GS_STATICGRAY) {
        jclass clazz;
        jclass clazz1;
        jmethodID mid;
        jobject cspace = NULL;
        jint bits[1];
        jintArray bitsArray;
        jboolean falseboolean = JNI_FALSE;
 
        clazz1 = env->FindClass("java/awt/color/ColorSpace");
        mid = env->GetStaticMethodID(clazz1, "getInstance",
              "(I)Ljava/awt/color/ColorSpace;");
        if (mid == 0) {
            return NULL;
        }
        cspace = env->CallStaticObjectMethod(clazz1, mid,
            java_awt_color_ColorSpace_CS_GRAY);
        if (cspace == 0) {
            return NULL;
        }
 
        bits[0] = 8;
        bitsArray = env->NewIntArray(1);
        if (bitsArray == 0) {
            return NULL;
        } else {
            env->SetIntArrayRegion(bitsArray, 0, 1, bits);
        }
 
        clazz = env->FindClass("java/awt/image/ComponentColorModel");

        mid = env->GetMethodID(clazz,"<init>",
            "(Ljava/awt/color/ColorSpace;[IZZII)V");
     
        if (mid == 0) {
            return NULL;
        }
     
        awt_colormodel = env->NewObject(clazz, mid,
                                        cspace,
                                        bitsArray,
                                        falseboolean,
                                        falseboolean,
                                        java_awt_Transparency_OPAQUE,
                                        java_awt_image_DataBuffer_TYPE_BYTE);        
    } else {
        jintArray hRGB = env->NewIntArray(256);
        unsigned int *rgb = NULL, *rgbP = NULL;
	RGBQUAD *c, *cBase;
        jboolean hasalpha = JNI_FALSE;

        /* Create the LUT from the color map */
	cBase = g_colorInfo[COLOR8].m_pPaletteColors;
	try {
	    rgb = (unsigned int *) env->GetPrimitiveArrayCritical(hRGB, 0);
	    if (rgb == NULL) {
	        return NULL;
	    }
	    rgbP = rgb;
	    if (m_clrdata.grayscale == GS_INDEXGRAY) {
	        /* For IndexColorModel, pretend first 10 colors and last
		   10 colors are transparent black.  This makes 
		   ICM.allgrayopaque true.
		*/
	        hasalpha = JNI_TRUE;
		for (i=0; i < 10; i++) {
		    *rgbP++ = 0x00000000;
		}
		for (i=10, c = cBase + 10; i < 246; c++, i++) {
		    *rgbP++ = 0xff000000 |
		        ((c->rgbRed&0xff)<<16)  |
		      ((c->rgbGreen&0xff)<<8) |
		      (c->rgbBlue&0xff);
		}
		for (i=246; i < 256; i++) {
		    *rgbP++ = 0x00000000;
		}
	    } else {
	        for (i=0, c = cBase; i < 256; c++, i++) {
		    *rgbP++ = 0xff000000 |
		        ((c->rgbRed&0xff)<<16)  |
		      ((c->rgbGreen&0xff)<<8) |
		      (c->rgbBlue&0xff);
		}
	    }
	} catch (...) {
	    env->ReleasePrimitiveArrayCritical(hRGB, rgb, JNI_COMMIT);
	    throw;
	}

	env->ReleasePrimitiveArrayCritical(hRGB, rgb, JNI_COMMIT);

        /* Construct a new color model */
        awt_colormodel = JNU_NewObjectByName(env, 
                                             "java/awt/image/IndexColorModel",
                                             "(II[IIZII)V",
                                             8, 256,
                                             hRGB, 0, 
                                             hasalpha,
                                             -1,
                                             java_awt_image_DataBuffer_TYPE_BYTE);
    }
    return awt_colormodel;
}

/************************************************************************
 * sun.awt.image.Image native methods
 */

extern "C" {

/*
 * Class:     sun_awt_image_Image
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_image_Image_initIDs(JNIEnv *env, jclass cls)
{
//    TRY;
//    AwtImage::getBackgroundMID = env->GetMethodID(cls, "getBackgroundColor","()Ljava/awt/Color;");
//    DASSERT(AwtImage::getBackgroundMID != NULL);
//    CATCH_BAD_ALLOC;
}


} /* extern "C" */

extern "C"
void color_init()
{
    TRY;

    static const int cmapsize = 256;
    static int initialized = 0;

    cubemapArray = (const char *)(AwtToolkit::GetInstance().getILUT());

    if (initialized) {
        return;
    } else {
        initialized = 1;
    }

    CATCH_BAD_ALLOC;
}

extern "C"
void initInverseGrayLut(int* prgb, int rgbsize, ColorData *cData) {
    TRY;

    int *idata;
    int g;

    if (!cData) {
        return;
    }

    idata = (int *) safe_Calloc(256, sizeof(int));
    cData->pGrayInverseLutData = idata;
    for (g = 0; g < 256; g++) {
        int i, mindist, d, besti;
        int *pg;
        mindist = 256;
        besti = 0;
        pg = prgb;
        for (i = 0; i < rgbsize; i++, pg++) {
            if (*pg == 0x0) {
                /* ignore transparent black */
                continue;
            }
            d = (*pg & 0xff) - g;
            if (d < 0) d = -d;
            if (d < mindist) {
                besti = i;
                if (d == 0) {
                    break;
                }
                mindist = d;
            }
        }
        *idata++ = besti;
    }

    CATCH_BAD_ALLOC;
}

extern "C"
void freeICMColorData(ColorData *pData) {
    TRY_NO_VERIFY;

    if (pData) {
        if (pData->img_clr_tbl) {
            free(pData->img_clr_tbl);
        }
        if (pData->pGrayInverseLutData) {
            free(pData->pGrayInverseLutData);
        }
	free(pData);
    }

    CATCH_BAD_ALLOC;
}

#include <malloc.h>

/*
 * state info needed for breadth-first recursion of color cube from
 * initial palette entries within the cube
 */

typedef struct {
    unsigned int depth;
    unsigned int maxDepth;

    unsigned char *usedFlags;
    unsigned int  activeEntries;
    unsigned short *rgb;
    unsigned char *indices;
    unsigned char *iLUT;
} CubeStateInfo;

#define INSERTNEW(state, rgb, index) do {                           \
        if (!state.usedFlags[rgb]) {                                \
            state.usedFlags[rgb] = 1;                               \
            state.iLUT[rgb] = index;                                \
            state.rgb[state.activeEntries] = rgb;                   \
            state.indices[state.activeEntries] = index;             \
            state.activeEntries++;                                  \
        }                                                           \
} while (0);
    

#define ACTIVATE(code, mask, delta, state, index) do {              \
    if (((rgb & mask) + delta) <= mask) {                           \
        rgb += delta;                                               \
        INSERTNEW(state, rgb, index);                               \
        rgb -= delta;                                               \
    }                                                               \
    if ((rgb & mask) >= delta) {                                    \
        rgb -= delta;                                               \
        INSERTNEW(state, rgb, index);                               \
        rgb += delta;                                               \
    }                                                               \
} while (0);

/* REMIND: does not deal well with bifurcation which happens when two
 * palette entries map to the same cube vertex
 */

static int
recurseLevel(CubeStateInfo *priorState) {
    int i;
    CubeStateInfo currentState;
    memcpy(&currentState, priorState, sizeof(CubeStateInfo));
    currentState.rgb = (unsigned short *)
        safe_Malloc(6 * sizeof(unsigned short) * priorState->activeEntries);
    
    try {
        currentState.indices = (unsigned char *)
	    safe_Malloc(6 * sizeof(unsigned char) * priorState->activeEntries);
    } catch (std::bad_alloc&) {
        free(currentState.rgb);
	currentState.rgb = NULL;
	throw;
    }
    
    currentState.depth++;
    if (currentState.depth > priorState->maxDepth) {
        priorState->maxDepth = currentState.depth;
    }
    currentState.activeEntries = 0;
    for (i=priorState->activeEntries - 1; i >= 0; i--) {
        unsigned short rgb = priorState->rgb[i];
        unsigned char  index = priorState->indices[i];
        ACTIVATE(rgb, 0x7c00, 0x0400, currentState, index);
        ACTIVATE(rgb, 0x03e0, 0x0020, currentState, index);
        ACTIVATE(rgb, 0x001f, 0x0001, currentState, index);
    }
    if (currentState.activeEntries) {
        try {
	  recurseLevel(&currentState);
	} catch (std::bad_alloc&) {
	    free(currentState.rgb);
	    free(currentState.indices);
	    throw;
        }
    }
    if (currentState.maxDepth > priorState->maxDepth) {
        priorState->maxDepth = currentState.maxDepth;
    }
    free(currentState.rgb);
    free(currentState.indices);
    return  1;
}

/*
 * REMIND: take core inversedLUT calculation to the shared tree and
 * recode the functions (Win32)awt_Image:initCubemap(),
 * (Win32)awt_Image:make_cubemap(), (Win32)AwtToolkit::GenerateInverseLUT(),
 * (Solaris)color:initCubemap() to call the shared codes.
 */
unsigned char*
make_cubemap(int* cmap,
             int  cmap_len,
             int  cube_dim) {
    int i;
    CubeStateInfo currentState;
    int cubesize = cube_dim * cube_dim * cube_dim;
    unsigned char *useFlags;
    unsigned char *newILut;

#ifdef DEBUG
    try {
#endif
        newILut = (unsigned char*)safe_Malloc(cubesize);
#ifdef DEBUG
    } catch (std::bad_alloc&) {
	fprintf(stderr, "Out of memory in awt_Image:make_cubemap()1\n");
	throw;
    }
#endif

    try {
        useFlags = (unsigned char *)safe_Calloc(cubesize, 1);
    } catch (std::bad_alloc&) {
        free(newILut);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:make_cubemap()2\n");
#endif
	throw;
    }
        
    /* use existing palette as seed values */
    currentState.depth          = 0;
    currentState.maxDepth       = 0;
    currentState.usedFlags      = useFlags;
    currentState.activeEntries  = 0;
    currentState.iLUT           = newILut;
    try {
        currentState.rgb = (unsigned short *)
	    safe_Malloc(256 * sizeof(unsigned short));
    } catch (std::bad_alloc&) {
        free(newILut);
	free(useFlags);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:make_cubemap()3\n");
#endif
	throw;
    }
        
    try {
        currentState.indices = (unsigned char *)
	    safe_Malloc(256 * sizeof(unsigned char));
    }
    catch (std::bad_alloc&) {
        free(newILut);
	free(useFlags);
	free(currentState.rgb);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:make_cubemap()4\n");
#endif
	throw;
    }
        
    for (i = 0; i < 128; i++) {
        unsigned short rgb;
	int pixel = cmap[i];
	rgb = (pixel & 0x00f80000) >> 9;
	rgb |= (pixel & 0x0000f800) >> 6;
	rgb |=  (pixel & 0xf8) >> 3;
	INSERTNEW(currentState, rgb, i);
	pixel = cmap[255-i];
	rgb = (pixel & 0x00f80000) >> 9;
	rgb |= (pixel & 0x0000f800) >> 6;
	rgb |=  (pixel & 0xf8) >> 3;
	INSERTNEW(currentState, rgb, 255-i);
    }
    try {
        recurseLevel(&currentState);
    } catch (std::bad_alloc&) {
        free(newILut);
	free(useFlags);
	free(currentState.rgb);
	free(currentState.indices);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:make_cubemap()5\n");
#endif
	throw;
    }
        
    free(useFlags);
    free(currentState.rgb);
    free(currentState.indices);
    return newILut;
}

/*
 * REMIND: take core inversedLUT calculation to the shared tree and
 * recode the functions (Win32)awt_Image:initCubemap(),
 * (Win32)awt_Image:make_cubemap(), (Win32)AwtToolkit::GenerateInverseLUT(),
 * (Solaris)color:initCubemap() to call the shared codes.
 */
extern "C"
unsigned char*
initCubemap(int* cmap,
            int  cmap_len,
            int  cube_dim) {
    // This function is called from 2D code which holds the JNI lock. We
    // must give up the ability to raise a java.lang.OutOfMemory error
    // because of this.
    TRY_NO_JNI;

    int i;
    CubeStateInfo currentState;
    int cubesize = cube_dim * cube_dim * cube_dim;
    unsigned char *useFlags;
    unsigned char *newILut;

#ifdef DEBUG
    try {
#endif
        newILut = (unsigned char*)safe_Malloc(cubesize);
#ifdef DEBUG
    } catch (std::bad_alloc&) {
	fprintf(stderr, "Out of memory in awt_Image:initCubemap()1\n");
        throw;
    }
#endif

    try {
        useFlags = (unsigned char *)safe_Calloc(cubesize, 1);
    } catch (std::bad_alloc&) {
        free(newILut);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:initCubemap()2\n");
#endif
	throw;
    }
    /* use existing palette as seed values */
    currentState.depth          = 0;
    currentState.maxDepth       = 0;
    currentState.usedFlags      = useFlags;
    currentState.activeEntries  = 0;
    currentState.iLUT           = newILut;
    
    try {
        currentState.rgb = (unsigned short *)
	    safe_Malloc(256 * sizeof(unsigned short));
    } catch (std::bad_alloc&) {
        free(newILut);
	free(useFlags);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:initCubemap()3\n");
#endif
	throw;
    }
        
    try {
        currentState.indices = (unsigned char *)
	    safe_Malloc(256 * sizeof(unsigned char));
    } catch (std::bad_alloc&) {
        free(newILut);
	free(useFlags);
	free(currentState.rgb);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:initCubemap()4\n");
#endif
	throw;
    }
        
    for (i = 0; i < 128; i++) {
        unsigned short rgb;
	int pixel = cmap[i];
	rgb = (pixel & 0x00f80000) >> 9;
	rgb |= (pixel & 0x0000f800) >> 6;
	rgb |=  (pixel & 0xf8) >> 3;
	INSERTNEW(currentState, rgb, i);
	pixel = cmap[255-i];
	rgb = (pixel & 0x00f80000) >> 9;
	rgb |= (pixel & 0x0000f800) >> 6;
	rgb |=  (pixel & 0xf8) >> 3;
	INSERTNEW(currentState, rgb, 255-i);
    }
    try {
        recurseLevel(&currentState);
    } catch (std::bad_alloc&) {
        free(newILut);
	free(useFlags);
	free(currentState.rgb);
	free(currentState.indices);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:initCubemap()5\n");
#endif
	throw;
    }
        
    free(useFlags);
    free(currentState.rgb);
    free(currentState.indices);
    return newILut;

    CATCH_BAD_ALLOC_RET_NO_JNI(NULL);
}

extern "C"
void initDitherTables(ColorData* cData) {
    TRY;

    if(std_odas_computed) {
        cData->img_oda_red   = &(std_img_oda_red[0][0]);
        cData->img_oda_green = &(std_img_oda_green[0][0]);
        cData->img_oda_blue  = &(std_img_oda_blue[0][0]);
    } else {
        cData->img_oda_red   = &(std_img_oda_red[0][0]);
        cData->img_oda_green = &(std_img_oda_green[0][0]);
        cData->img_oda_blue  = &(std_img_oda_blue[0][0]);
        make_dither_arrays(256, cData);
        std_odas_computed = 1;
    }

    CATCH_BAD_ALLOC;
}

static void make_dither_arrays(int cmapsize, ColorData *cData) {
    int i, j, k;

    /*
     * Initialize the per-component ordered dithering arrays
     * Choose a size based on how far between elements in the
     * virtual cube.  Assume the cube has cuberoot(cmapsize)
     * elements per axis and those elements are distributed
     * over 256 colors.
     * The calculation should really divide by (#comp/axis - 1)
     * since the first and last elements are at the extremes of
     * the 256 levels, but in a practical sense this formula
     * produces a smaller error array which results in smoother
     * images that have slightly less color fidelity but much
     * less dithering noise, especially for grayscale images.
     */
    i = (int) (256 / pow(cmapsize, 1.0/3.0));
    make_sgn_ordered_dither_array(cData->img_oda_red, -i / 2, i / 2);
    make_sgn_ordered_dither_array(cData->img_oda_green, -i / 2, i / 2);
    make_sgn_ordered_dither_array(cData->img_oda_blue, -i / 2, i / 2);

    /*
     * Flip green horizontally and blue vertically so that
     * the errors don't line up in the 3 primary components.
     */
    for (i = 0; i < 8; i++) {
        for (j = 0; j < 4; j++) {
            k = cData->img_oda_green[(i<<3)+j];
            cData->img_oda_green[(i<<3)+j] = cData->img_oda_green[(i<<3)+7 - j];
            cData->img_oda_green[(i<<3) + 7 - j] = k;
            k = cData->img_oda_blue[(j<<3)+i];
            cData->img_oda_blue[(j<<3)+i] = cData->img_oda_blue[((7 - j)<<3)+i];
            cData->img_oda_blue[((7 - j)<<3) + i] = k;
        }
    }
}

